// FluidLinksXMLHandler.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/annotation/FluidLinksXMLHandler.java 11    5/20/99 12:14p Bchang $

package annotation;

import org.xml.sax.*;
import java.io.IOException;
import java.util.Stack;
import java.net.URL;
import fluid.FluidUI;

public class FluidLinksXMLHandler extends HandlerBase {

  private boolean debug = false;

  protected Stack stack = new Stack();

  protected boolean startElement = false;
  protected boolean endElement = false;
  protected boolean keepAdjacent = false;
  private boolean lastCharIsWhitespace = false;
  
  protected boolean keepWhitespace = false;
  
  protected String href = "";
  
  protected URL docbase = null;
  protected FluidUI canvas = null;

  public FluidLinksXMLHandler(TextBlock tb, URL db) {
    enterTextBlock(tb);
    docbase = db;
    canvas = tb.getCanvas();
  }

  protected void enterTextBlock(TextBlock tb) {
    stack.push(tb);
  }

  protected TextBlock exitTextBlock() {
    return (TextBlock)stack.pop();
  }

  protected TextBlock currentTextBlock() {
    return (TextBlock)stack.peek();
  }

  public void startElement(String name, AttributeList attributes) {
    endElement = false;
    startElement = true;
    if (debug) System.out.println("<" + name + ">");
    if (name.equals("fluiddoc")) {
      currentTextBlock().startDocument();
    } else if (name.equals("picture")) {
      String source = attributes.getValue("source");
      String caption = attributes.getValue("caption");
      currentTextBlock().startPicture(source, caption);
    } else if (name.equals("title")) {
      currentTextBlock().startTitle();
      keepAdjacent = true;
    } 
    // BEGIN MOD
    else if (name.equals("keyword")) {
      //currentTextBlock().startTitle();
      currentTextBlock().startKeyword();
      keepAdjacent = true;
    } 
    // END MOD
    else if (name.equals("para")) {
      currentTextBlock().startParagraph();
      keepAdjacent = true;
    } else if (name.equals("link")) {
      href = "";
      for (int i = 0; i < attributes.getLength(); i++) {
        if (attributes.getName(i).equalsIgnoreCase("href")) {
          // instead, could use attributes.getValue("href") and catch errors?
          href = attributes.getValue(i);
        }
      }
    } else if (name.equals("anchor")) {
      currentTextBlock().startAnchor();
    } else if (name.equals("gloss")) {
      GlossAttributes ga = new GlossAttributes(attributes);
      TextBlock tb = currentTextBlock().startGloss(docbase, href, ga, canvas);
      enterTextBlock(tb);
      keepAdjacent = true;
    } else if (name.equals("ws")) {
      keepWhitespace = true;    
    } else {
      System.out.println("Unknown tag: " + name);
    }

    /*
    // extract attributes
    for (int i = 0; i < attributes.getLength(); i++) {
      System.out.println("Attribute: " + attributes.getName(i) + ' ' +
 			                   attributes.getType(i) + " \"" +
			                   attributes.getValue(i) + '"');
        
      }
    }
    */
    
  }

  public void endElement(String name) {
    startElement = false;
    endElement = true;
    if (debug) System.out.println("</" + name + ">");
    if (name.equals("fluiddoc")) {
      currentTextBlock().endDocument();
    } else if (name.equals("picture")) {
      currentTextBlock().endPicture();
    } else if (name.equals("title")) {
      currentTextBlock().endTitle();
    } else if (name.equals("para")) {
      currentTextBlock().endParagraph();
    } else if (name.equals("link")) {
      // do nothing
    } else if (name.equals("anchor")) {
      currentTextBlock().endAnchor();
    } else if (name.equals("gloss")) {
      currentTextBlock().endGloss();
      exitTextBlock();
    } else if (name.equals("ws")) {
        keepWhitespace = false;
    }
    // BEGIN MOD
    else if (name.equals("keyword")) {
        currentTextBlock().endKeyword();
    }
    // END MOD
  }

  public void characters(char ch[], int start, int length) {
    boolean adjacentToPrevious = false;
    if (debug) System.out.println("[chars]");
    if (startElement) {
      adjacentToPrevious = !lastCharIsWhitespace && !isWhitespace(ch[start]);
      startElement = false;
    } else if (endElement) {
      adjacentToPrevious = !isWhitespace(ch[start]);
      endElement = false;
    }
    if (keepAdjacent) {
      adjacentToPrevious = true;
      keepAdjacent = false;
    }
    String str;
    if (keepWhitespace) {
      str = new String(ch, start, length);
    } else {
      str = removeExcessWhitespace(ch, start, length);
    }
    if (str.equals("")) return;
    currentTextBlock().addToText(str, adjacentToPrevious);
    lastCharIsWhitespace = isWhitespace(ch[start + length - 1]);
  }

  // convenience methods
  protected boolean isWhitespace(char c) {
    return (c == ' ') || (c == '\t') || (c == '\n');
  }

  protected String removeExcessWhitespace(char ch[], int start, int length) {
    StringBuffer buf = new StringBuffer();
    boolean ws = false;
    boolean front = true;
    for (int i = start; i < start + length; i++) {
      switch (ch[i]) {
      case ' ':
      case '\n':
      case '\t':
        ws = true;
        break;  
      default:
        if (ws) {
          if (!front) buf.append(' ');
          ws = false;
        }
        buf.append(ch[i]);
        front = false;
      }
    }
    return buf.toString();
  }

  // exceptions
  public void warning(SAXParseException exception) {
    System.out.println("warning");
    System.out.println("Warning: " + exception.getMessage() +
		                   " (" + exception.getSystemId() + ':' +
		                   exception.getLineNumber() + ',' +
		                   exception.getColumnNumber() + ')');
  }

  public void error(SAXParseException exception) {
    System.out.println("recoverable error");
    System.out.println("Recoverable Error: " + exception.getMessage() +
		                   " (" + exception.getSystemId() + ':' +
		                   exception.getLineNumber() + ',' +
		                   exception.getColumnNumber() + ')');
  }

  public void fatalError(SAXParseException exception) {
    System.out.println("fatal error");
    System.out.println("Fatal Error: " + exception.getMessage() +
                       " (" + exception.getSystemId() + ':' +
		                   exception.getLineNumber() + ',' +
		                   exception.getColumnNumber() + ')');
  }

}

/* $Log: /FluidDocs/annotation/FluidLinksXMLHandler.java $
 * 
 * 11    5/20/99 12:14p Bchang
 * removed textblock, computed, and ctag instance variables -- no longer
 * used here. also added startDocument().
 * 
 * 10    5/12/99 11:30a Bchang
 * Moved the processing of <gloss> attributes to its own class,
 * GlossAttributes.
 * 
 * 9     1/26/99 11:58p Bchang
 * added support for width & height attributes for image glosses.
 * also made images work as nested glosses: had to cache docbase and
 * canvas in the xmlhandler, because nested glosses don't have a path
 * up the component hierarchy during parsing
 * 
 * 8     1/22/99 3:00p Bchang
 * add support for Picture pages, which have the form:
 * <picture source="image.gif" caption="my dog"/>
 * 
 * 7     1/21/99 4:40p Bchang
 * add support for image glosses.
 * 
 * 6     1/21/99 3:27p Pollez
 * added computed glosses.
 *  allow new attributes of gloss: type="computed", tag="whatever"
 *       ("title" is default)
 * allow new tag: "title", currently same as "para"
 * 
 * 5     1/19/99 3:53p Bchang
 * (Added comment for version 4.)
 * 
 * 4     1/19/99 3:45p Bchang
 * Added support for respect-whitespace: <ws>  </ws>.
 * Right now, only works for spaces, not tabs or newlines.
 * 
 * 3     1/13/99 5:31p Bchang
 * Added $Log at bottom of file
 */
