// GroupedGlossParser.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/annotation/GroupedGlossParser.java 2     6/30/99 1:11p Bchang $

// Grouped gloss format is as follows:
// Write the primary text. Whitespace is honored -- in particular, you should
// only put a newline at the end of a paragraph.
// Anchors are delimited on both sides by a string (by default, this is the
// one character symbol "^").
// At any point, you can begin writing the glosses for all the anchors seen
// thus far that haven't had glosses written for them yet. At this point, you
// must write enough glosses to "catch up" to all the anchors seen thus far.
// Glosses are begun with a string (by default, this is the one character symbol
// "$"). They may be of multiple lines -- again, whitespace is significant, so
// newlines will cause line breaks. The last line of the gloss ends with
// the filename to jump to, surrounded by a delimiter string (by default, this
// is the one character symbol "%"). For example, "%nextpage.txt%".
// Here's an example of a grouped gloss file:
//
//   Log in to the system. ^View example^.
//   $Login: root99
//   Password: ADMIN%%
//   Once in this account, you can use ^unix commands^ to do stuff.
//   $You know Unix, don't you?%unixcommand.txt%

package annotation;

import java.io.*;
import java.util.*;
import java.net.URL;
import fluid.FluidUI;

import com.oroinc.text.regex.*;

public class GroupedGlossParser {

  protected final String anchorDelimiter = "^";
  protected final String glossStartDelimiter = "$";
  protected final String glossEndDelimiter = "%";

  protected String anchorPattern;
  protected String glossPattern;
  protected String filenamePattern;

  protected URL docbase = null;
  protected FluidUI canvas = null;

  protected boolean adjacentToPrevious = false;

  protected PatternCompiler compiler = new Perl5Compiler();
  protected PatternMatcher matcher = new Perl5Matcher();

  // stack of text blocks
  protected Stack textBlockStack = new Stack();
  protected void enterTextBlock(TextBlock tb) {
    textBlockStack.push(tb);
    adjacentToPrevious = true;
  }
  protected TextBlock exitTextBlock() {
    return (TextBlock)textBlockStack.pop();
  }
  protected TextBlock currentTextBlock() {
    return (TextBlock)textBlockStack.peek();
  }

  protected String escapeDelimiter(String delim) {
    StringBuffer result = new StringBuffer();
    for (int i = 0; i < delim.length(); i++) {
      char c = delim.charAt(i);
      result.append("\\").append(c);
    }
    return result.toString();
  }

  /*
  protected void createPatterns() {
    String anchorStartPattern = escapeDelimiter(anchorDelimiter);
    String anchorEndPattern = anchorStartPattern;
    String glossStartPattern = escapeDelimiter(glossStartDelimiter);
    String glossEndPattern = escapeDelimiter(glossEndDelimiter);
    String filenameStartPattern = escapeDelimiter(filenameStartDelimiter);
    String filenameEndPattern = escapeDelimiter(filenameEndDelimiter);

    String anyCharsPattern = "(.*)"; // the parens create a regex group

    anchorPattern = anchorStartPattern + anyCharsPattern + anchorEndPattern;
    glossPattern = glossStartPattern + anyCharsPattern + glossEndPattern;
    filenamePattern = filenameStartPattern + anyCharsPattern + filenameEndPattern;
  }

    protected MatchResult result;
    protected PatternMatcherInput input;

    final String wsPat = "\\s*";

     Pattern pattern = compiler.compile(anchorPattern);
     input = new PatternMatcherInput(str);
     int startchar = 0;
     while(matcher.contains(input, pattern)) {
      result = matcher.getMatch();
       } catch (MalformedPatternException e) {
      System.out.println("Malformed pattern: " + e);
    }
    return mat;
  }
  */

  public GroupedGlossParser(TextBlock tb, URL db) {
    enterTextBlock(tb);
    docbase = db;
    canvas = tb.getCanvas();
  }

  protected void addText(String str) {
    if (str.equals("")) return;
    try {
      Pattern newlinePattern = compiler.compile("\n");
      Vector textlines = com.oroinc.text.regex.Util.split(matcher, newlinePattern, str);
      currentTextBlock().addToText((String)textlines.elementAt(0), adjacentToPrevious);
      for (int i = 1; i < textlines.size(); i++) {
        currentTextBlock().endParagraph();
        currentTextBlock().startParagraph();
        currentTextBlock().addToText((String)textlines.elementAt(i), adjacentToPrevious);
      }
    } catch (MalformedPatternException e) {
      System.out.println("malformed pattern in addText()");
    }
  }

  protected void addAnchor(String str) {
    currentTextBlock().startAnchor();
    currentTextBlock().addToText(str, adjacentToPrevious);
    currentTextBlock().endAnchor();
  }

  protected void addGloss(String str) {
    GlossAttributes ga = new GlossAttributes();
    TextBlock tb = currentTextBlock().startGloss(docbase, currentGlossFilename, ga, canvas);
    enterTextBlock(tb);
    addText(str);
    currentTextBlock().endGloss();
    exitTextBlock();
  }

  protected Vector lines = new Vector();
  protected int n = 0;

  public void parse(InputStreamReader isr) {
    BufferedReader reader = new BufferedReader(isr);
    currentTextBlock().startDocument();
    currentTextBlock().startParagraph();
    try {
      for (String str = reader.readLine(); str != null; str = reader.readLine()) {
        lines.addElement(str);
      }
    } catch (IOException e) {
      System.out.println("ioexception reading grouped gloss file");
    }
    try {
      Pattern anchorPattern = compiler.compile(escapeDelimiter(anchorDelimiter));
      // for now, don't handle nested annos
      while (n < lines.size()) {
        String text = getText();
        Vector textsplit = com.oroinc.text.regex.Util.split(matcher, anchorPattern, text);
        // split() will add null strings to beginning and end of vector if necessary
        int nglosses = (textsplit.size() - 1) / 2;
        for (int i = 0; i < nglosses; i++) {
          int j = i * 2;
          addText((String)textsplit.elementAt(j));
          addAnchor((String)textsplit.elementAt(j + 1));
          addGloss(getGloss());
        }
        addText((String)textsplit.lastElement());
      }
    } catch (MalformedPatternException e) {
      System.out.println("malformed pattern in parse()");
    }
    currentTextBlock().endParagraph();
    currentTextBlock().endDocument();
  }

  protected String getText() {
    StringBuffer text = new StringBuffer("");
    boolean newline = false;
    while (n < lines.size()) {
      if (newline) {
        text.append('\n');
      } else {
        newline = true;
      }
      String line = (String)lines.elementAt(n);
      String trimmed = (new String(line)).trim();
      if (trimmed.startsWith(glossStartDelimiter)) {
        break;
      } else {
        n++;
        text.append(line);
      }
    }
    return text.toString();
  }

  protected String currentGlossFilename = "";


  protected String getGloss() {
    StringBuffer gloss = new StringBuffer("");
    boolean newline = false;
    currentGlossFilename = "";
    while (n < lines.size()) {
      if (newline) {
        gloss.append('\n');
      } else {
        newline = true;
      }
      String line = (String)lines.elementAt(n++);
      String trimmed = (new String(line)).trim();
      boolean useTrimmed = false;
      if (trimmed.startsWith(glossStartDelimiter)) {
        trimmed = trimmed.substring(glossStartDelimiter.length());
        useTrimmed = true;
      }
      if (trimmed.endsWith(glossEndDelimiter)) {
        int j = trimmed.indexOf(glossEndDelimiter);
        int k = trimmed.lastIndexOf(glossEndDelimiter);
        if (k > j + 1) {
          currentGlossFilename = trimmed.substring(j + 1, k);
        }
        gloss.append(trimmed.substring(0, j));
        break;
      }
      if (useTrimmed) {
        gloss.append(trimmed);
      } else {
        gloss.append(line);
      }
    }
    return gloss.toString();
  }

}

/* $Log: /FluidDocs/annotation/GroupedGlossParser.java $
 * 
 * 2     6/30/99 1:11p Bchang
 * added comment describing the format
 * 
 * 1     6/28/99 10:49a Bchang
 */
