// OutlineParser.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/annotation/OutlineParser.java 3     6/30/99 1:12p Bchang $

// OutlineParser parses the "outline" format of fluid documents.
// Outline format supports a subset of XML format;
// it also makes some assumptions about the kind of links and placement of links.
// In particular, it is optimized for "hypersentence" structures.

package annotation;

import java.io.*;
import java.util.*;
import java.net.URL;
import fluid.FluidUI;

public class OutlineParser {
  protected URL docbase = null;
  protected FluidUI canvas = null;
  
  protected boolean adjacentToPrevious = false;

  // stack of text blocks
  protected Stack textBlockStack = new Stack();
  protected void enterTextBlock(TextBlock tb) {
    textBlockStack.push(tb);
    adjacentToPrevious = true;
  }
  protected TextBlock exitTextBlock() {
    return (TextBlock)textBlockStack.pop();
  }
  protected TextBlock currentTextBlock() {
    return (TextBlock)textBlockStack.peek();
  }
  
  
  protected void addBody(Line line) {
    if (!line.body.equals("")) {
      currentTextBlock().addToText(line.body, adjacentToPrevious);
      adjacentToPrevious = false;
    }
  }
  protected void addTail(Line line) {
    currentTextBlock().addToText(line.tail, adjacentToPrevious);
    adjacentToPrevious = false;
  }
  protected void addTailAsAnchor(Line line) {
    currentTextBlock().startAnchor();
    currentTextBlock().addToText(line.tail, adjacentToPrevious);
    adjacentToPrevious = false;
    currentTextBlock().endAnchor();
  }
    

  public OutlineParser(TextBlock tb, URL db) {
    enterTextBlock(tb);
    docbase = db;
    canvas = tb.getCanvas();
  }
  
  protected Dictionary lines = new Hashtable();
  
  public void parse(InputStreamReader isr) {
    BufferedReader reader = new BufferedReader(isr);
    currentTextBlock().startDocument();
    currentTextBlock().startParagraph();
    try {
      for (String str = reader.readLine(); str != null; str = reader.readLine()) {
        if (!str.equals("")) {
          Line line = new Line(str);
          lines.put(line.level, line);
        }
      }
    } catch (IOException e) {
      System.out.println("ioexception reading outline file");
    }
    int n = 1;
    while (true) {
      Line line = (Line)lines.get(new Level(n));
      if (line == null) break;
      processLine(line);
      n++;
    }
    currentTextBlock().endParagraph();
    currentTextBlock().endDocument();
  }
  
  protected void processLine(Line line) {
    addBody(line);
    Level level = line.level;
    Line nextline = (Line)lines.get(level.nextGlossLevel());
    if (nextline != null) { // there is a gloss hanging off of this line
      addTailAsAnchor(line);
      GlossAttributes ga = new GlossAttributes();
      TextBlock tb = currentTextBlock().startGloss(docbase, "", ga, canvas);
      enterTextBlock(tb);
      processLine(nextline);
    } else { // there is no gloss hanging off of this line
      addTail(line);
      while (!level.isTopLevel()) {
        if (level.inGloss()) {
          currentTextBlock().endGloss();
          exitTextBlock();
          nextline = (Line)lines.get(level.continuationLevel());
          processLine(nextline);
          break;
        } else {
          level = level.popLevel();
        }
      }
    }
  }
  
  protected class Line {
    public Level level;
    public String body;
    public String tail;
    public Line(String str) {
      String trimmed = str.trim();
      int firstSpace = trimmed.indexOf(' ');
      int lastSpace = trimmed.lastIndexOf(' ');
      level = new Level(trimmed.substring(0, firstSpace).trim());
      if (firstSpace == lastSpace) {
        body = "";
      } else {
        body = trimmed.substring(firstSpace + 1, lastSpace).trim();
      }
      tail = trimmed.substring(lastSpace + 1).trim();
    }
  }
  
  protected class Level {
    public String number;
    
    public Level(String lev) {
      number = lev;
    }
    
    // creates a top level
    public Level(int n) {
      number = String.valueOf(n) + ".";
    }
    
    public String toString() {
      return number;
    }
    
    public boolean isTopLevel() {
      return number.length() <= 2;
    }
    
    public Level nextGlossLevel() {
      return new Level(number + "2.");
    }
    
    public Level continuationLevel() {
      StringBuffer sb = new StringBuffer(number);
      sb.setCharAt(number.length() - 2, '1');
      return new Level(sb.toString());
    }
    
    public boolean inGloss() {
      return number.endsWith(".2.");
    }
    
    public Level popLevel() {
      if (number.length() <= 2) {
        return new Level("");
      }
      return new Level(number.substring(0, number.length() - 2));
    }
    
    public boolean equals(Object lev) {
      if (lev == null) return false;
      return number.equals(((Level)lev).number);
    }
    
    public int hashCode() {
      return number.hashCode();
    }
  }

  /*
rules:
you got the current guy
print its body
add .2 to the level
look for such a line
if found, // means there's a gloss hanging off here
  make last tail ("current guy") an anchor
  enter gloss
  recurse to "rules"
if not found, // means that there is no gloss hanging off here
  add current guy's tail as normal text
  look at current last element
  if it is 2, // means that we are in a gloss, so exit
    exit gloss 
    change last level element to 1
    get that line
    recurse to rules
  else it is 1, // means that we are in a mainline branch
    pop off this level
    if prev level is .2,
      change it to .1
      recurse to rules
    else if prev level is .1
      go to pop off this level
    else if prev level is top-level sentence
      stop this sentence.
*/  

}


/* $Log: /FluidDocs/annotation/OutlineParser.java $
 * 
 * 3     6/30/99 1:12p Bchang
 * removed a println that was used for debugging
 * 
 * 2     6/09/99 12:29p Bchang
 * fixed inGloss() to check for ".2." rather than "2." (the latter was a
 * bug which was causing items like "12." to return true)
 * 
 * 1     5/20/99 12:16p Bchang
 *
 */