// Animation.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/fluid/Animation.java 10    7/23/99 9:11a Bchang $

package fluid;

import java.util.Date;
import java.util.Vector;

/*
Animation objects are abstract (except for AnimationGroups).
To instantiate an Animation object that invokes someMethod
in YourObject with each value along the animation, create
an anonymous class from the Animation class (in this example,
AnimInt) and add a stepTarget() method and some boilerplate
methods for accessing the current animation, like this:

     Animation myAnim = null;
     AnimInt animation = new AnimInt(0, 10) {
       public void stepTarget(double fraction) {
         someMethod(currentValue(fraction));
       }
       protected Animation getCurrentAnimation() { return myAnim; }
       protected void setCurrentAnimation(Animation anim) { myAnimation = anim; }
     };

Alternatively, you could create an inner class.

AnimationGroups guarantee coordinated animation. You can add()
Animations to an AnimationGroup, and then simply start the
AnimationGroup instead of starting each Animation individually.
*/

public abstract class Animation implements TimerListener {
  // instances of Animation must implement stepTarget()
	protected abstract void stepTarget(double fraction);
  protected abstract void setCurrentAnimation(Animation anim);
  protected abstract Animation getCurrentAnimation();

  // subclasses of Animation must implement currentValue(),
  // returning whatever object type they animate

	protected int duration;
  protected long startTime;
  
  protected int delay = 0;
  public void setDelay(int d) { delay = d; }
  public int getDelay() { return delay; }
  
  protected Animator controllingAnimator;
  
  private boolean done = false;

  Object start, end;
  
  protected boolean gen = false;

  protected Animation() {}

  protected Animation(Object s, Object e) {
    start = s;
    end = e;
  }

  public void setDuration(int dur) {
    duration = dur;
  }
  
  protected void setStart() {}
  protected void setEnd() {}
    
  protected void establishStartAndEnd() {
    setStart();
    setEnd();
  }

  protected long currentTime() {
    return new Date().getTime();
  }

  void start(Animator animator) {
    if (delay <= 0) {
      startNoDelay(animator);
    } else {
      new Timer(this, delay, 0, animator);
    }
  }
    
  public void timerExpired(int id, Object o) {
    if (!isDone()) {
      startNoDelay((Animator)o);
    }
  }
  
  private void startNoDelay(Animator animator) {
    start(currentTime(), duration, animator);
  }  

  protected void start(long stime, Animator animator) {
    start(stime, duration, animator);
  }

  protected void start(long stime, int dur, Animator animator) {
    establishStartAndEnd();
    startTime = stime;
    setDuration(dur);
    Animation prevanim = getCurrentAnimation();
    controllingAnimator = animator;
    if ((prevanim != null) && !prevanim.isDone()) {
      prevanim.stop();
    }
    if (!isDone()) {  // in case it was stopped before it even got started
      setCurrentAnimation(this);
      animator.add(this);
      setStarted();
    }
  }

  boolean step(long currentTime) {
    stepFraction = (((double)(currentTime - startTime)) / duration);
    stepFraction = Math.min(stepFraction, 1.0);
    stepTarget(stepFraction);
    if (stepFraction == 1.0) {
      setDone();
      follow();
    }
    return (stepFraction == 1.0);
	}

  protected double stepFraction = 0.0;
  public double getStepFraction() { return stepFraction; }

  // completed is true if animation ran to completion
  public boolean completed() { return stepFraction == 1.0; }

  // done is true if animation is no longer running (may have been interrrupted with stop())
  // don't do follow() in setDone() because we don't want the follow animation to
  //   go if this animation was interrupted.
  protected synchronized void setDone() { done = true; triggerDone(); }
  public boolean isDone() { return done; }

  public synchronized boolean addAnimationListenerIfNotDone(AnimationListener al) {
    if (!isDone()) {
      addAnimationListener(al);
      return true;
    }
    return false;
  }


  private boolean started = false;
  protected void setStarted() { started = true; }
  public boolean hasStarted() { return started; }

  public void stop() {
    if (!isDone()) {
      if (hasStarted()) {
        controllingAnimator.remove(this);
      }
      setDone();
    }
  }

  protected void follow() {
    if (follower != null) {
      follower.start(controllingAnimator);
    }
  }
  public void setFollower(Animation anim) { follower = anim; }
  protected Animation follower = null;

  Vector doneListeners = new Vector();
  public void addAnimationListener(AnimationListener al) { doneListeners.addElement(al); }
  protected void triggerDone() {
    for (int i = 0; i < doneListeners.size(); i++) {
      ((AnimationListener) doneListeners.elementAt(i)).animationDone(this);
    }
  }

  protected int interpolateInt(int start, int end, double fraction) {
    return (int)Math.round(start + (((double)(end - start)) * fraction));
  }

}

/* $Log: /FluidDocs/fluid/Animation.java $
 * 
 * 10    7/23/99 9:11a Bchang
 * check if isDone before actually starting the animation
 *
 * 9     1/27/99 6:47p Bchang
 * delayed animations check to see whether they've been stopped before
 * they start
 * 
 * 8     1/27/99 2:40p Bchang
 * stop() the previous animation rather than just remove it from the
 * animator queue. this ensures that any AnimationListeners are alerted.
 * 
 * 7     1/27/99 3:02a Bchang
 * added addAhimationListenerIfNotDone for nested annotations
 * 
 * 6     1/26/99 6:47p Bchang
 * in stop(), check if animation has started before asking for
 * controllingAnimator.
 * 
 * 5     1/26/99 5:39p Bchang
 * moved follow() out of setDone() and into step() -- this ensures
 * that follow animations are not triggered if this animation is
 * interrupted.
 * 
 * 4     1/25/99 4:12p Bchang
 * support for triggering events when an animation completes:
 * call addAnimationListener(yourObj) and implement the AnimationListener
 * interface
 * 
 * 3     1/13/99 5:49p Bchang
 * Added $Log at bottom of file.
 * Added ability to set a delay before starting the animation: setDelay().
 * Enabled chained animation: setFollower().
 */
