// CalloutLine.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/fluid/CalloutLine.java 5     9/20/99 8:37a Bchang $

package fluid;


import java.awt.*;

// currently supports horizontal callout lines that have a single segment stem line
// and a platform line.
// future enhancements could include vertical callouts; multiple segment stems;
// optional platform; asymmetrical platform

// for now, don't allow freezing of callout glosses

public class CalloutLine extends FluidComponent implements AnimationListener {
  protected Point startPt, endPt, stemStart, stemEnd, platformStart, platformEnd;
  protected int fullPlatformHeight;
  protected int platformHeight;
  protected int fullPlatformY;

  public void resetPlatform() {
    platformHeight = 0;
    platformStart.y = endPt.y;
  }


  public CalloutLine(int x1, int x2, int y0, int y1, int h, Color col) {
    startPt = new Point(x1, y0);
    endPt = new Point(x2, y0);

    fullPlatformHeight = h;
    platformHeight = 0;

    fullPlatformY = y1;

    stemStart = new Point(startPt);
    stemEnd = new Point(stemStart);
    platformStart = new Point(endPt);
    platformHeight = 0;

    setColor(col);
    setSize(new Dimension(0, 0));
  }

  /*
  // this code is for maintaining a constant line growth rate for the
  // animation. surprisingly, this doesn't look right! the simpler
  // solution of having fixed and equal durations for each part of the
  // callout line seems to satisfy the need to visually comprehend
  // each portion of the two-part animation
  
  public int lineAnimationRatePer100 = 200;

  protected int stemAnimationDuration() {
    return (Math.abs(startPt.x - endPt.x) / 100) * lineAnimationRatePer100;
  }

  protected int platformAnimationDuration() {
    return ((platformHeight / 2) / 100) * lineAnimationRatePer100;
  }
  */
  
  protected int stemAnimationDuration() { return FluidUI.animduration / 2; }
  protected int platformAnimationDuration() { return FluidUI.animduration / 2; }
  
  public void drawSelf(Graphics g) {
    g.setColor(getColor());
    drawLine(stemStart, stemEnd, g);
    Point platformEnd = new Point(platformStart.x, platformStart.y + platformHeight);
    drawLine(platformStart, platformEnd, g);

    if (!stemStart.equals(stemEnd)) {
      String stm = "STEM: " + stemStart.x + " " + stemStart.y + " " + stemEnd.x + " " + stemEnd.y;
      String plat = "PLATFORMXYH: " + platformStart.x + " " + platformStart.y + " " + platformHeight;
      Color c = getColor();
      String rgb = "RGB: " + c.getRed() + " " + c.getGreen() + " " + c.getBlue();
      String details = stm + " " + plat + " " + rgb;

  	  FluidUI.fluidLog.addLog("CALLOUTLINE", details, false);
    }
  }

  protected void drawLine(Point start, Point end, Graphics g) {
    if (!start.equals(end)) {
      g.drawLine(start.x, start.y, end.x, end.y);
    }
  }

  public void grow(Animation annoanim) {
    Animation stemanim = changeStemEnd(endPt.x);
    stemanim.setDuration(stemAnimationDuration());
    Animation plathanim = changePlatformHeight(fullPlatformHeight);
    Animation platyanim = changePlatformY(fullPlatformY);
    AnimationGroup ag = new AnimationGroup();
    ag.add(plathanim);
    ag.add(platyanim);
    ag.add(annoanim);
    ag.setDuration(platformAnimationDuration());
    stemanim.setFollower(ag);
    startAnimation(stemanim);
  }

  public void shrink(Animation annoanim) {
    AnimationGroup ag = new AnimationGroup();
    Animation plathanim = changePlatformHeight(0);
    Animation platyanim = changePlatformY(endPt.y);
    ag.add(plathanim);
    ag.add(platyanim);
    ag.add(annoanim);
    ag.setDuration(platformAnimationDuration());
    Animation stemanim = changeStemEnd(startPt.x);
    stemanim.setDuration(stemAnimationDuration());
    ag.setFollower(stemanim);
    stemanim.addAnimationListener(this);
    startAnimation(ag);
  }

  public void animationDone(Animation anim) {
    // workaround for the "orphaned platform" bug
    // (the bug is (I believe) a race condition in stopping an animation
    // that hasn't quite started yet. I haven't been able to isolate it.

    // but anyway, we want to remove this calloutline from its parent (Page)
    // because it gets added anew on each grow()
    getParent().remove(this);
  }

  protected Animation stemEndAnimation = null;
  public AnimInt changeStemEnd(final int tox) {
    return new AnimInt() {
      protected void setStart() { start = stemEnd.x; }
      protected void setEnd() { end = tox; }
      protected void stepTarget(double fraction) {
        stemEnd.x = currentValue(fraction);
      }
      protected Animation getCurrentAnimation() { return stemEndAnimation; }
      protected void setCurrentAnimation(Animation anim) { stemEndAnimation = anim; }
    };
  }

  protected Animation platformHeightAnimation = null;
  public AnimInt changePlatformHeight(final int toh) {
    return new AnimInt() {
      protected void setStart() { start = platformHeight; }
      protected void setEnd() { end = toh; }
      protected void stepTarget(double fraction) {
        platformHeight = currentValue(fraction);
      }
      protected Animation getCurrentAnimation() { return platformHeightAnimation; }
      protected void setCurrentAnimation(Animation anim) { platformHeightAnimation = anim; }
    };
  }

  protected Animation platformYAnimation = null;
  public AnimInt changePlatformY(final int toh) {
    return new AnimInt() {
      protected void setStart() { start = platformStart.y; }
      protected void setEnd() { end = toh; }
      protected void stepTarget(double fraction) {
        platformStart.y = currentValue(fraction);
      }
      protected Animation getCurrentAnimation() { return platformYAnimation; }
      protected void setCurrentAnimation(Animation anim) { platformYAnimation = anim; }
    };
  }
}

/* $Log: /FluidDocs/fluid/CalloutLine.java $
 * 
 * 5     9/20/99 8:37a Bchang
 * changed stem & platform animation duration to be half of overall
 * animduration (used to be fixed to 150 msec each).
 * 
 * 4     7/26/99 9:38a Bchang
 * introduce resetPlatform() to workaround the orphaned callout platform
 * bug.
 * when shrink animation finishes, remove this callout line from its
 * parent. this should be done anyway, because
 * annotation.growmargincallout() adds it to Page, but
 * shrinkmargincallout() can't remove it because we want to wait until the
 * animation completes.
 * as a side effect, removing after shrink works around the orphaned
 * callout platform bug. calling resetPlatform() before grow ensures
 * platform is not left in orphaned size.
 * 
 * 3     7/23/99 1:42p Bchang
 * add logging
 * 
 * 2     7/16/99 3:19p Bchang
 * now "platform" (the vertical line) does not have to be centered on the
 * stem line. constructor takes the top y of the platform.
 *
 * 1     1/21/99 4:44p Bchang
 * Initial version.
 */