// FluidComponent.java
// Copyright (c) 1999 by Xerox Corporation
// bchang@parc.xerox.com

// $Header: /FluidDocs/fluid/FluidComponent.java 14    7/26/99 9:37a Bchang $

package fluid;

import java.awt.*;
import java.util.Vector;

public abstract class FluidComponent implements TimerListener {

  // "globals"
  public Animator getAnimator() { return getParent().getAnimator(); }
  public FluidUI getCanvas() { return getParent().getCanvas(); }
  public Graphics getGraphics() { return getParent().getGraphics(); }
  public FluidPage getPage() { 
    return getParent().getPage(); }

  // hierarchy
  protected FluidComponent parent = null;
  public FluidComponent getParent() { return parent; }
  public void setParent(FluidComponent par) { parent = par; }

  protected Vector components = new Vector();
  public Vector getComponents() { return components; }
  public void setComponents(Vector comps) { components = comps; }

  public void add(FluidComponent comp) {
    components.addElement(comp);
    comp.setParent(this);
  }
  public int getComponentCount() { return components.size(); }
  public FluidComponent getComponent(int i) { return (FluidComponent)components.elementAt(i); }

  public void remove(FluidComponent comp) {
    components.removeElement(comp);
    // should we do comp.setParent(null)? does it matter?
  }

  public FluidComponent getComponentAt(Point p) {
    for (int i = 0; i < components.size(); i++) {
      FluidComponent comp = (FluidComponent)components.elementAt(i);
      if (comp.inBounds(p)) {
        return comp.getComponentAt(p);
      }
    }
    return this;
  }

  public void removeAllComponents() {
    components.removeAllElements();
  }

  // location
  protected Point location = new Point(0, 0);
  public Point getLocation() { return location; }
  public int getX() { return location.x; }
  public int getY() { return location.y; }
  public int getLeft() { return getX(); }
  public int getRight() { return getX() + getWidth(); }
  public int getTop() { return getY(); }
  public int getBottom() { return getY() + getHeight(); }
  public Point getMiddle() {
    return new Point((getRight() - getLeft()) / 2, (getBottom() - getTop()) / 2);
  }
  public void setLocation(Point newloc) { location = newloc; }
  public void setX(int newx) { setLocation(new Point(newx, getY())); }
  public void setY(int newy) { setLocation(new Point(getX(), newy)); }

  public int absoluteX() { return parent.absoluteX() + getX(); }
  public int absoluteY() { return parent.absoluteY() + getY(); }
  public int absoluteLeft() { return absoluteX(); }
  public int absoluteRight() { return parent.absoluteX() + getRight(); }
  public int absoluteTop() { return absoluteY(); }
  public int absoluteBottom() { return parent.absoluteY() + getBottom(); }
  public Point absoluteLocation() { return new Point(absoluteX(), absoluteY()); }
	public String getLocationStr () {
		String str = absoluteLeft() + " " +
								 absoluteRight() + " " +
								 absoluteTop() + " " +
								 absoluteBottom();
		return str;
	}


  // size
  protected Dimension size = null;
  public Dimension getSize() { return size; }
  public int getWidth() { return size.width; }
  public int getHeight() { return size.height; }
  public void setSize(Dimension newsize) { size = newsize; }
  public void setWidth(int neww) { size.width = neww; }
  public void setHeight(int newh) { size.height = newh; }

  public boolean inBounds(Point p) {
    return ((absoluteLeft() <= p.x) && (p.x <= absoluteRight()) &&
            (absoluteTop() <= p.y) && (p.y <= absoluteBottom()));
  }

  // color
  protected Color color = Color.black;
  public Color getColor() { return color; }
  public void setColor(Color c) { color = c; }

	// logID
	public String getID() { return ""; }
	public String getLogID() {
		FluidComponent parent = getParent();
		if (parent == null) {
			return getID();
		}
		return parent.getLogID() + getID();
	}

  // drawing
  protected boolean topDrawn = false;
  public boolean isTopDrawn() { return topDrawn; }
  public void setTopDrawn(boolean b) {
    topDrawn = b;
    if (topDrawn) {
      getPage().addTopDrawnComponent(this);
    } else {
      getPage().removeTopDrawnComponent(this);
    }
  }

  // this first checks to see if component should delay drawing
  // to appear on top of all other components
  public void draw(Graphics g) {
    if (!isTopDrawn()) {
      doDraw(g);
    }
  }

  // this does not check to see if component should delay drawing
  // to appear on top of all other components
  public void doDraw(Graphics g) {
    int x = getX();
    int y = getY();
    g.translate(x, y);
    try {
      drawSelf(g);
      drawComponents(g);
//      drawHighlight(g);
    } catch (Exception e) {
      // in case doDraw or drawSelf etc throws an Exception,
      // we still do the "un"translate.
      // why would this happen? it happens when an animation finishes or is interrupted
      // and an object no longer exists or something, and you get a null pointer exception
      System.out.println("Caught an exception during FluidComponent draw.");
      e.printStackTrace();
    }
    g.translate(-x, -y);
  }

  protected void drawSelf(Graphics g) {} // do nothing

  public void drawComponents(Graphics g) {
    for (int i = 0; i < components.size(); i++) {
      ((FluidComponent)components.elementAt(i)).draw(g);
    }
  }

  public void repaint() { getParent().repaint(); }
  
  protected void drawHighlight(Graphics g) {
    if (highlight) {
      g.setColor(highlightColor);
      g.drawRect(0, 0, getWidth(), getHeight());
    }
  }

  protected boolean highlight = false;
  protected Color highlightColor;
  public void highlight(Color c) {
    highlight = true;
    highlightColor = c;
  }


  // called when the structure is completely loaded
  // components can do whatever they wish in structureLoaded()
  public void notifyStructureLoaded() {
    structureLoaded();
    for (int i = 0; i < getComponentCount(); i++) {
      FluidComponent comp = (FluidComponent)components.elementAt(i);
      comp.notifyStructureLoaded();
    }
  }
  public void structureLoaded() {}

  // initialization -- no longer used ?
  void initialize(Graphics g) {
    initializeSelf(g);
    initializeComponents(g);
  }

  protected void initializeSelf(Graphics g) {}

  protected void initializeComponents(Graphics g) {
    for (int i = 0; i < components.size(); i++) {
      ((FluidComponent)components.elementAt(i)).initialize(g);
    }
  }

  // events
  public void mouseDown(Event evt, int x, int y) {
    getParent().mouseDown(evt, x, y);
  }
  
  public void rightMouseDown(Event evt, int x, int y) {
    getParent().rightMouseDown(evt, x, y);
  }
  
  public void middleMouseDown(Event evt, int x, int y) {
    getParent().middleMouseDown(evt, x, y);
  }

  public void mouseUp(Event evt, int x, int y) {
    getParent().mouseUp(evt, x, y);
  }

  // someone else got the mouseup
  public void mouseUpNotice() {}

  public void middleMouseUp(Event evt, int x, int y) {
    getParent().middleMouseUp(evt, x, y);
  }

  public void rightMouseUp(Event evt, int x, int y) {
    getParent().rightMouseUp(evt, x, y);
  }

  public void mouseMove(Event evt, int x, int y) {
    getParent().mouseMove(evt, x, y);
  }

  // mouseEnter and mouseLeave do not propagate to parents
  // mouseEnter should pass along the component which actually was entered, like mouseLeave does
  public void mouseEnter(Event evt, int x, int y) {}
  public void mouseLeave(Event evt, int x, int y,
                         FluidComponent originatingComp, FluidComponent enteredComp) {}

  // from conversation w/ Jock:
  // mouseEnter() and mouseLeave() pass the object that they entered or left as a parameter
  // this message gets sent to all parents who are entered/left
  // but not to a parent you are still within
  // if the mouse enters/leaves a component, the component's parent gets notified with a
  // subcomponentEntered/Left message in lieu of a mouseMove
  // only the immediate parent gets the message, not all ancestors.

  // gets mouseMove event from AWT and determines
  // whether to fire a mouseEnter, mouseLeave, or mouseMove
  // also initiates a mouseDwell if possible
  void delegateMouseMotion(Event evt, int x, int y) {
    incrementSerial();
    FluidComponent comp = getComponentAtMouse(); // gets component at previous mouse location
    if (this != comp) {
      cascadeLeave(comp, evt, x, y, comp);
      setComponentAtMouse(this);
      mouseEnter(evt, x, y);
    } else {
      mouseMove(evt, x, y);
    }
    if (dwellCheckPermitted && isDwellable(x, y)) {
      pendingDwell(x, y);
    }
  }

  protected void cascadeLeave(FluidComponent comp, Event evt, int x, int y,
                              FluidComponent originatingComp) {
    // if comp is an ancestor of 'this' ("entered" object) (or comp == this),
    // then don't signal a mouseLeave on comp.
    if (comp.isAncestorOf(this) || (comp == this)) {
      return;
    } else {
      comp.mouseLeave(evt, x, y, originatingComp, this);
      // for now, we re-enable the dwell check here, on leave
      // but this doesn't take care of entering & leaving a subcomponent; in that
      // case you would like that dwell check to be re-enabled (I think). not sure.
      // --
      // this should probably be moved to enter, but right now there isn't cascaded enter
      comp.allowDwellCheck();
      if (comp.getParent() != null) {
        cascadeLeave(comp.getParent(), evt, x, y, originatingComp);
      }
    }
  }


  protected boolean dwellCheckPermitted = true;
  public void allowDwellCheck() { dwellCheckPermitted = true; }
  public void disallowDwellCheck() { dwellCheckPermitted = false; }

  public FluidComponent getComponentAtMouse() {
    return getPage().getComponentAtMouse();
  }
  public void setComponentAtMouse(FluidComponent comp) {
    getPage().setComponentAtMouse(comp);
  }

  public boolean isDescendentOf(FluidComponent comp) {
    return comp.isAncestorOf(this);
  }

  public boolean isAncestorOf(FluidComponent comp) {
    if (isParentOf(comp)) {
      return true;
    } else if (comp.getParent() == null) {
      return false;
    } else {
      return isAncestorOf(comp.getParent());
    }
  }

  public boolean isParentOf(FluidComponent comp) {
    return equals(comp.getParent());
  }

  // dwell event
  static protected int serial = 0;
  static public void incrementSerial() {
    serial = (serial + 1) % 10000;
  }

  protected boolean dwellable = false;
  public void setDwellable(boolean b) { dwellable = b; }
  public boolean getDwellable() { return dwellable; }

  public boolean isDwellable(int x, int y) {
    if (getDwellable()) {
      return true;
    } else {
      return getParent().isDwellable(x, y);
    }
  }

  public void pendingDwell(int x, int y) {
//    incrementSerial();
    new Timer(this, FluidUI.dwelltime, serial, new Point(x, y));
  }

  public void timerExpired(int id, Object o) {
    if (id != serial) {
      // this means that another dwell has begun pending
      // so this one shouldn't fire
      return;
    }
    Point p = (Point)o;
    if (dwellCheckPermitted) {
      mouseDwell(p.x, p.y);
      disallowDwellCheck();
    }
  }

  public void mouseDwell(int x, int y) {
    FluidComponent par = getParent();
    if (par.dwellCheckPermitted) {
      par.mouseDwell(x, y);
      par.disallowDwellCheck();
    }
  }

  // negotiation
  public void negotiationSetGuidelines(Negotiation ngt) {}
  public void negotiationSelectPresentation(Negotiation ngt) {}
  public void negotiationSelectSpaceStrategy(Negotiation ngt) {}

  public void negotiationRevert(Negotiation ngt, double frac) {}

  public FluidComponent getPrimary() { return this; }

  protected String graphicalProtocol = "fluid.GChar";
  public void setGraphicalProtocol(String prot) { graphicalProtocol = prot; }
  public String getGraphicalProtocol() { return graphicalProtocol; }

  // animation
  public void startAnimation(Animation anim) {
    getAnimator().startAnimation(anim);
  }

  public void startAnimation(Animation anim1, Animation anim2) {
    getAnimator().startAnimation(anim1, anim2);
  }

  public void startAnimation(Vector animvec) {
    getAnimator().startAnimation(animvec);
  }

  protected Animation sizeAnimation = null;
  public AnimDimension changeSize(final Dimension tosize) {
    return new AnimDimension() {
      protected void setStart() { start = getSize(); }
      protected void setEnd() { end = tosize; }
      protected void stepTarget(double fraction) {
        setSize(currentValue(fraction));
      }
      protected Animation getCurrentAnimation() { return sizeAnimation; }
      protected void setCurrentAnimation(Animation anim) { sizeAnimation = anim; }
    };
  }

  protected Animation locationAnimation = null;
  public AnimPoint changeLocation(final Point toloc) {
    return new AnimPoint() {
      protected void setStart() { start = getLocation(); }
      protected void setEnd() { end = toloc; }
      protected void stepTarget(double fraction) {
        setLocation(currentValue(fraction));
      }
      protected Animation getCurrentAnimation() { return locationAnimation; }
      protected void setCurrentAnimation(Animation anim) { locationAnimation = anim; }
    };
  }

  protected Animation xAnimation = null;
  public AnimInt changeX(final int tox) {
    return new AnimInt() {
      protected void setStart() { start = getX(); }
      protected void setEnd() { end = tox; }
      protected void stepTarget(double fraction) {
        setX(currentValue(fraction));
      }
      protected Animation getCurrentAnimation() { return xAnimation; }
      protected void setCurrentAnimation(Animation anim) { xAnimation = anim; }
    };
  }

  protected Animation yAnimation = null;
  public AnimInt changeY(final int toy) {
    return new AnimInt() {
      protected void setStart() { start = getY(); }
      protected void setEnd() { end = toy; }
      protected void stepTarget(double fraction) {
        setY(currentValue(fraction));
      }
      protected Animation getCurrentAnimation() { return yAnimation; }
      protected void setCurrentAnimation(Animation anim) { yAnimation = anim; }
    };
  }

  protected Animation colorAnimation = null;
  public AnimColor changeColor(final Color tocolor) {
    return new AnimColor() {
      protected void setStart() { start = getColor(); }
      protected void setEnd() { end = tocolor; }
      protected void stepTarget(double fraction) {
        setColor(currentValue(fraction));
      }
      protected Animation getCurrentAnimation() { return colorAnimation; }
      protected void setCurrentAnimation(Animation anim) { colorAnimation = anim; }
    };
  }

}

/* $Log: /FluidDocs/fluid/FluidComponent.java $
 * 
 * 14    7/26/99 9:37a Bchang
 * add remove() to remove a subcomponent
 * 
 * 13    7/16/99 10:41a Bchang
 * add notifyStructureLoaded() and structureLoaded() which can be called
 * when a page has finished building itself. this is used by Annotations
 * which consult their Page's history db to see whether to color
 * themselves visited
 * 
 * 12    7/14/99 8:44a Bchang
 * added mouseUpNotice() which is sent to a component which registered
 * interest during mouseDown() - indicates that mouseUp occurred on some
 * other component. See annotation.java for an example use.
 *
 * 11    7/13/99 5:44p Mackinla
 * Logging - basic functionality
 * 
 * 10    7/08/99 2:40p Mackinla
 * Logging v1
 *
 * 9     3/31/99 10:57a Bchang
 * put in a comment in doDraw explaining how g.translate() is always
 * called, even when an exception occurs
 * 
 * 8     1/27/99 2:41p Bchang
 * added support for highlighting the bbox of a component; this support is
 * disabled right now in the draw loop.
 * 
 * 7     1/26/99 5:41p Bchang
 * added preliminary support for starting multiple animations at once,
 * with startAnimation(anim1, anim2) and startAnimation(animvec).
 * Need to do more in order to actually ensure that the startTimes of
 * the animations are synchronized.
 * 
 * 6     1/26/99 10:42a Bchang
 * Wrapped drawSelf/drawComponents in doDraw() method
 * with a try/catch block in order to
 * protect against the "shifted world" bug: if the draw fails, the
 * graphics
 * context didn't get a chance to un-translate. With the catch, the
 * un-translate should always occur.
 * 
 * 5     1/21/99 4:45p Bchang
 * Support for middle and right mouse buttons.
 * 
 * 4     1/13/99 6:19p Bchang
 * Moved changeX() from TextLine.
 * Added setStart(), setEnd() to the change*() animation methods to allow
 * for chained animations.
 * 
 * 3     1/13/99 5:48p Bchang
 * Added $Log at bottom of file
 */
