;; jde-ant.el --- Use Apache Ant to build your JDE projects

;; $Revision: 1.27 $ $Date: 2001/11/28 13:22:14 $ 

;;
;; Author: Jason Stell | jason.stell@globalone.net
;; Created: 19 Oct 2000
;; Version 1.3   19 June 2001
;;

;; This file is not part of Emacs

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;; Commentary:
;; This file defines jde-ant-build and some helper functions.
;; jde-ant-build uses the specified ant program/shell script to 
;; execute a specified build file (in the project root). 
;; 
;;
;; To do: 
;; -- allow interactive ant target/argument specification. 
;;    {DONE - version 1.1}
;;
;; -- consider using jde-run-java-vm to execute Ant instead of
;;    requiring a shell script.
;;
;;
;; Notes:
;; -- The JDE (Java Development Environment for Emacs) can be
;;    downloaded at http://sunsite.auc.dk/jde/
;;
;; -- Apache Ant is a Java & XML build system that can be downloaded 
;;    at http://jakarta.apache.org/ant/
;;
;; Version History:
;;
;; -- Version 1.3 (19 June 2001)
;;    : Addition of jde-ant-projecthelp to display list of targets for
;;      the current buildfile.
;; -- Version 1.2 (4 June 2001) 
;;    : Addition of jde-ant-read-buildfile option to prompt for the buildfile 
;;      name -- contributed by Rob Shaw <shaw@servidium.com>
;;    : Various Bug fixes contributed by Rob Shaw <shaw@servidium.com> 
;;        - The setting of the system property in a format other
;;          than -Dname=value had the side effect of negating the -emacs 
;;          command line argument.
;;        - The setting of the current directory to the location of the 
;;          JDE project file is now taking place when
;;          jde-ant-enable-find is nil.
;;        - The ant target is now the last thing to be appended to
;;          ant command to avoid any possible confusion for ANT
;;          as to what is the desired target.
;; -- Version 1.2b2 (25 May 2001) 
;;    : Fix to properly use the -find <buildfile> Ant switch--contributed
;;      by Rob Shaw <shaw@servidium.com>.
;; -- Version 1.2b1 (23 May 2001) 
;;    : Added jde-ant-enable-find custom flag to use the -find switch
;;      available in Ant. This overrides the requirement for a JDE
;;      project file
;;    : Fixed minor bug missing whitespace before -buildfile switch
;;      when building the ant compile command
;; -- Version 1.1 (20 October 2000)
;;    : Added interactive prompts (optional, based on customizable
;;      toggles) for Ant target and additional args. Removed the 
;;      jde-ant-target custom variable, since this is really
;;      represented by the default target in the build file.
;;    : The -f switch seems to be causing problems. Removed it from
;;      the default jde-ant-args.
;;    : Basic changes to the way the ant command is assembled.
;;
;; -- Version 1.0 (19 October 2000)
;;    Initial Version

;; Define XEmacs- and Emacs-only variables 
;; and functions to avoid compiler warnings.
(eval-when (compile)
  (if jde-xemacsp
      (progn
	(defvar compilation-enter-directory-regexp-alist)
	(defvar compilation-leave-directory-regexp-alist)
	(defvar compilation-file-regexp-alist)
	(defvar compilation-nomessage-regexp-alist)
	(defvar compilation-scroll-output)
	(defvar compilation-process-setup-function)
	(defvar last-nonmenu-event))))

(defgroup jde-ant nil
  "JDE Ant"
  :group 'jde
  :prefix "jde-ant-")

;;; start of Douglas WF Acheson mod
;;; The next three custom variables and defun added by Douglas WF Acheson to
;;; allow a user to select how ant is invoked, either by script or via Java

(defcustom jde-ant-invocation-method (list "Script")
  "*Specifies how to invoke ant. Ant can be invoked in one of three
ways. The first is via the ant script/program that comes with ant.
The second is via java and the third is via the Ant Server." 
  :group 'jde-ant
  :type '(list
           (radio-button-choice
	     (const "Script")
	     (const "Java")
             (const "Ant Server"))))

(defcustom jde-ant-home ""
  "*Directory where ant is install."
  :group 'jde-ant
  :type  'file)

(defcustom jde-ant-user-jar-files nil
  "*Specifies jar files that hold user-defined tasks."
  :group 'jde-ant
  :type '(repeat (file :tag "Path")))


(defcustom jde-ant-program "ant"
  "*Specifies name of ant program/script."
 :group 'jde-ant
 :type 'string)

(defcustom jde-ant-args "-emacs"
  "*Specifies arguments to be passed to make program."
  :group 'jde-ant
  :type 'string)

(defcustom jde-ant-buildfile "build.xml"
  "*Specifies the default buildfile to use."
  :group 'jde-ant
  :type 'string)

(defcustom jde-ant-read-buildfile nil
"*Specify whether to prompt for a buildfile. If non-nil, the 
jde-ant-build command prompts you for an ant buildfile."
  :group 'jde-ant
  :type 'boolean)

(defcustom jde-ant-read-target nil
"*Specify whether to prompt for a build target. If non-nil, the 
jde-ant-build command prompts you for an ant target."
  :group 'jde-ant
  :type 'boolean)

(defvar jde-ant-interactive-target ""
"Target to use for the build in place of the default target.")

(defvar jde-ant-interactive-target-arg-history nil
"History of targets entered in the minibuffer.")

(defcustom jde-ant-read-args nil
"*Specify whether to prompt for additional arguments to pass to ant. If non-nil, the 
jde-ant-build command prompts you for the additional arguments."
  :group 'jde-ant
  :type 'boolean)

(defvar jde-ant-interactive-args ""
"Additional arguments for the ant command.")

(defvar jde-ant-interactive-args-arg-history nil
"History of targets entered in the minibuffer.")

(defvar jde-ant-buildfile-arg-history nil
 "History of targets entered in the minibuffer.")
 
(defcustom jde-ant-enable-find nil
"*Specify whether the -find option should be specified in Ant. If non-nil, Ant
will search up the directory hierarchy from the current directory for the build
definition file. Also note that, if non-nil, this will relax the requirement for
an explicit jde project file."
   :group 'jde-ant
   :type 'boolean)

(defcustom jde-ant-complete-target t
  "*Specify whether to enable completion of build target names in the
minibuffer.

If non-nil, the jde-ant-build command allows you to use tab completion
in the minibuffer to specify the build target name.  This list of
valid build targets is determined by parsing the Ant build file.  This
option has no effect if jde-ant-read-target is nil."
  :group 'jde-ant
  :type 'boolean)

(defcustom jde-ant-target-regexp "<\\s-*target\\s-[^...]*?name\\s-*=\\s-*\"\\s-*\\([^\"]+\\)"
  "*Regular expression used to match target names in Ant build files."
  :group 'jde-ant
  :type 'string)

(defun jde-build-ant-command (target more-args)
  "Constructs the java ant command. The variable `jde-ant-home' is used
if it is set, otherwise it gets the ant home from the environment
variable ANT_HOME."
  (let* ((ant-home (if (string= "" jde-ant-home)
                       (jde-normalize-path (getenv "ANT_HOME"))
                     (jde-normalize-path jde-ant-home)))
         (ant-command
         (concat 
          (if (string= (car jde-ant-invocation-method) "Java")
              (concat 
               "java -classpath " 
               (if (and (eq system-type 'windows-nt)
                        (string-match "sh$" shell-file-name))
                   "\"")
               (jde-build-classpath
                (append
                 (list 
                  (expand-file-name "lib/ant.jar" ant-home)
                  (expand-file-name "lib/crimson.jar" ant-home)
                  (expand-file-name "lib/jaxp.jar" ant-home)
                  (expand-file-name "lib/optional.jar" ant-home)
                  (expand-file-name "lib/tools.jar" (jde-get-jdk-dir)))
                 jde-ant-user-jar-files))
               (if (and (eq system-type 'windows-nt)
                        (string-match "sh$" shell-file-name))
                   "\"")))
          " -Dant.home="
          ant-home
          (if (string= (car jde-ant-invocation-method) "Java")
              (concat
               " "
               "org.apache.tools.ant.Main")))))
          
    (if (not jde-ant-enable-find)
        (if (not (string= jde-ant-buildfile ""))
            (setq ant-command 
		  (concat ant-command 
			  " -buildfile \""
                          (jde-normalize-path jde-ant-buildfile)
                          "\""))))

    (if (not (string= jde-ant-args ""))
        (setq ant-command (concat ant-command " " jde-ant-args)))

    (if (not (string= more-args ""))
        (setq ant-command (concat ant-command " " more-args)))

    (if jde-ant-enable-find
        (setq ant-command 
	      (concat ant-command 
		      " -find \"" jde-ant-buildfile "\"")))

    (if (not (string= target ""))
        (setq ant-command (concat ant-command " " target " ")))
    
    ant-command))

;;;###autoload
(defun jde-ant-build ()
  "Build the current project using Ant."
  (interactive)

  (if jde-ant-read-buildfile
      (setq jde-ant-buildfile
	      (read-from-minibuffer 
	       "Buildfile: "
	       jde-ant-buildfile
	       nil nil
	       '(jde-ant-buildfile-arg-history . 1))))

  (if jde-ant-read-target
      (if jde-ant-complete-target
          (setq jde-ant-interactive-target
                (completing-read 
                 "Target to build: "
                 (jde-ant-get-target-alist)
                 nil
                 t
                 jde-ant-interactive-target
                 '(jde-ant-interactive-target-arg-history . 1)))
        (setq jde-ant-interactive-target
	      (read-from-minibuffer 
	       "Target to build: "
	       jde-ant-interactive-target
	       nil nil
	       '(jde-ant-interactive-target-arg-history . 1)))))

  (if jde-ant-read-args
      (setq jde-ant-interactive-args
	      (read-from-minibuffer 
	       "Additional build args: "
	       jde-ant-interactive-args
	       nil nil
	       '(jde-ant-interactive-args-arg-history . 1))))


  (let ((compile-command (jde-build-ant-command
                          jde-ant-interactive-target
                          jde-ant-interactive-args)))
    (when compile-command
      ;; Force save-some-buffers to use the minibuffer
      ;; to query user about whether to save modified buffers.
      ;; Otherwise, when user invokes the command from
      ;; menu, save-some-buffers tries to popup a menu
      ;; which seems not to be supported--at least on
      ;; the PC.
      (if (and (eq system-type 'windows-nt)
	       (not jde-xemacsp))
          (let ((temp last-nonmenu-event))
            ;; The next line makes emacs think that the command
            ;; was invoked from the minibuffer, even when it
            ;; is actually invoked from the menu-bar.
            (setq last-nonmenu-event t)
            (save-some-buffers (not compilation-ask-about-save) nil)
            (setq last-nonmenu-event temp))
        (save-some-buffers (not compilation-ask-about-save) nil))

      (if (string= (car jde-ant-invocation-method) "Ant Server")
          (progn
            (while (string-match "\"" compile-command)
              (setq compile-command (replace-match "" nil nil compile-command)))
            (jde-ant-compile-internal compile-command 
                                      (bsh-eval
                                       (concat "jde.util.AntServer.start(\""
                                               compile-command "\");"))
                                      "No more errors"))
        (compile-internal compile-command "No more errors")))))

(defun jde-ant-compile-internal (command output error-message) 
  "This method displays ant output in a compilation buffer.
error-message is a string to print if the user asks to see another error
and there are no more errors. "
  (let* (error-regexp-alist
         enter-regexp-alist
         leave-regexp-alist
         file-regexp-alist
         nomessage-regexp-alist
         (parser compilation-parse-errors-function)
         (name-of-mode "Compilation") outbuf
         (len (length output)))
        
    (save-excursion				       ;;getting or creating
      (setq outbuf (get-buffer-create "*compilation*"));;the compilation buffer
      (set-buffer outbuf) ;;setting the compilation buffer
      
      ;; In case the compilation buffer is current, make sure we get the global
      ;; values of compilation-error-regexp-alist, etc.
      (kill-all-local-variables))
	(setq error-regexp-alist compilation-error-regexp-alist)
	(setq enter-regexp-alist (if (not jde-xemacsp) compilation-enter-directory-regexp-alist))
	(setq leave-regexp-alist (if (not jde-xemacsp) compilation-leave-directory-regexp-alist))
	(setq file-regexp-alist (if (not jde-xemacsp) compilation-file-regexp-alist))
	(setq nomessage-regexp-alist (if (not jde-xemacsp) compilation-nomessage-regexp-alist))
    
    (let ((thisdir default-directory)
	  outwin)
      (save-excursion
	;; Clear out the compilation buffer and make it writable.
	(set-buffer outbuf)
        (compilation-mode)
	(setq buffer-read-only nil)
	(buffer-disable-undo (current-buffer))
	(erase-buffer)
	(buffer-enable-undo (current-buffer))
        (insert "CompileServer output:\n")
	(insert command "\n")
	(insert "\n" output "\n")
        (set-buffer-modified-p nil))
      ;; If we're already in the compilation buffer, go to the end
      ;; of the buffer, so point will track the compilation output.
      (if (eq outbuf (current-buffer))
	  (goto-char (point-max)))
      ;; Pop up the compilation buffer.
      (setq outwin (display-buffer outbuf))
      (save-excursion
	(set-buffer outbuf)
	(compilation-mode name-of-mode)
	;; (setq buffer-read-only t)  ;;; Non-ergonomic.
	(set (make-local-variable 'compilation-parse-errors-function) parser)
	(set (make-local-variable 'compilation-error-message) error-message)
	(set (make-local-variable 'compilation-error-regexp-alist)
	     error-regexp-alist)
        (if (not jde-xemacsp)
            (progn
              (set (make-local-variable 'compilation-enter-directory-regexp-alist)
                   enter-regexp-alist)
              (set (make-local-variable 'compilation-leave-directory-regexp-alist)
                   leave-regexp-alist)
              (set (make-local-variable 'compilation-file-regexp-alist)
                   file-regexp-alist)
              (set (make-local-variable 'compilation-nomessage-regexp-alist)
                   nomessage-regexp-alist)))
;; 	(set (make-local-variable 'compilation-arguments)
;; 	     (list output error-message))
	(setq default-directory thisdir
	      compilation-directory-stack (list default-directory))
	(set-window-start outwin (point-min))
	(or (eq outwin (selected-window))
	    (set-window-point outwin (point-min)))
	(compilation-set-window-height outwin)
	
        (if (not jde-xemacsp)
            (if compilation-process-setup-function
                (funcall compilation-process-setup-function))))
     
      (if (not jde-xemacsp)
          (jde-ant-scroll-output outwin))
      )
    ;; Make it so the next C-x ` will use this buffer.
    (setq compilation-last-buffer outbuf)))

(defun jde-ant-scroll-output (window)
  (if compilation-scroll-output
      (save-selected-window
        (select-window window)
        (goto-char (point-max))
        (let* ((result (count-lines-page));;Getting the number of lines
               (lines (substring result (+ (string-match "+" result) 1)
                                 (string-match ")" result)))
               ;;number to offsett the scroll up.(formatting reasons)
               (offset (if compilation-window-height
                           (abs (- compilation-window-height 5))
                         (/ (window-height) 2))))
          (scroll-up (- (string-to-number lines);;reducing the number of 
                        offset)))))) ;;lines to scroll

;;;###autoload
(defun jde-ant-projecthelp ()
  "Display Ant project help for the current project.
This will execute the Ant program with the `-projecthelp'
switch to output available targets with their descriptions
for the current buildfile. This function uses the same 
rules as `jde-ant-build' for finding the buildfile."
  (interactive)

  (if jde-ant-read-buildfile
      (setq jde-ant-buildfile
	      (read-from-minibuffer 
	       "Buildfile: "
	       (jde-normalize-path jde-ant-buildfile)
	       nil nil
	       '(jde-ant-buildfile-arg-history . 1))))

  (let ((projecthelp-command (jde-build-ant-command nil " -projecthelp")))
    (when projecthelp-command
      (if (not jde-ant-enable-find)
          (shell-command projecthelp-command "*Ant Project Help*")
	(shell-command projecthelp-command "*Ant Project Help*")))))

(defun jde-ant-find-build-file (dir)
  "Find the next Ant build file upwards in the directory tree from DIR.

Returns nil if it cannot find a project file in DIR or an ascendant directory."
  (let ((file (find (cond ((string= jde-ant-buildfile "") "build.xml")
                          (t jde-ant-buildfile))
		    (directory-files dir) :test 'string=)))
    (if file
	(expand-file-name file dir)
      (if (not (jde-root-dir-p dir))
	  (jde-ant-find-build-file (concat dir "../"))))))

(defun jde-ant-get-target-alist ()
  "Returns asociation list of valid Ant project targets.
Returns nil if the build file can not be found."
  (let (targets
        (build-file (if jde-ant-enable-find (jde-ant-find-build-file default-directory)
                      jde-ant-buildfile)))
    (when build-file
      (let ((temp-buf (get-buffer-create "*jde-ant-get-target-list-temp-buffer*")))
        (unwind-protect
            (save-excursion
              (set-buffer temp-buf)
              (erase-buffer)
              (insert-file-contents build-file)
              (goto-char (point-min))
              (while (re-search-forward jde-ant-target-regexp (point-max) t)
                (setq targets (append targets (list (list (match-string 1)))))))
          (kill-buffer temp-buf))))
    targets))

(provide 'jde-ant)

;; Change History 

;;
;; $Log: jde-ant.el,v $
;; Revision 1.27  2001/11/28 13:22:14  jslopez
;; JDE ant uses jde-normalize-path with
;; jde-ant-home and jde-ant-buildfile.
;;
;; Revision 1.26  2001/11/27 14:04:43  jslopez
;; XEmacs compatability fix.
;;
;; Revision 1.25  2001/11/27 02:23:08  jslopez
;; Adds jde-ant-buildfile-arg-history variable.
;; Fixes jde-ant-target-regexp to match the name attribute
;; when it is in a different ine than the target.
;; i.e. <target description="foo"
;;              name="compile">
;; Removes unnecessary messages.
;; XEmacs compatibility fix.
;;
;; Revision 1.24  2001/11/21 15:35:56  jslopez
;; Modifies the jde-ant-target-regexp to
;; match the targets name anywhere in the line.
;; It stills fails if the target name is in a different
;; line.
;;
;; Revision 1.23  2001/11/18 21:42:06  jslopez
;; Updating documentation.
;;
;; Revision 1.22  2001/11/18 21:39:07  jslopez
;; Modifies jde-build-ant-command to get the
;; ant home value from the
;; environment if jde-ant-home is not set.
;;
;; Revision 1.21  2001/11/11 23:27:41  jslopez
;; Fixed bug in jde-ant-build-command, causing jde-ant-args getting
;; appending without a space to the ant command if
;; jde-ant-buildfile is not specified.
;; Update the documentation for the methods jde-ant-invocation-method
;; and jde-ant-home.
;;
;; Revision 1.20  2001/11/09 02:05:36  jslopez
;; Fixing the output's display.
;;
;; Revision 1.19  2001/11/05 02:15:28  jslopez
;; Adds an option to use AntServer for builds.
;;
;; Revision 1.18  2001/11/04 19:22:33  jslopez
;; Fixes bug using Ant target completion
;; cause when jde-ant-enable-find is not enabled.
;;
;; Revision 1.17  2001/10/30 15:20:41  jslopez
;; Replaces the custom variable jde-ant-dir with jde-ant-home.
;; Enclose the classpath with double quotes for Windows machines
;; using cygwin as a shell.
;; Include the flag ant.home as an argument to ant.
;;
;; Revision 1.16  2001/10/29 12:50:29  jslopez
;; Removed the installation instructions from the top of the file
;; since it was obsolete. Jde-ant can be used out of the box with
;; JDE without having to any other installation.
;; Finished moving all the ant customization variables to the
;; customization group jde-ant.
;;
;; Revision 1.15  2001/10/29 06:27:40  paulk
;; Fixed minor bug in last submission.
;;
;; Revision 1.14  2001/10/29 06:21:28  paulk
;; Added support for invoking Ant directly from java or via the ant script.
;; Thanks to Douglas WF Acheson <dwfa@yahoo.com> for some help on this.
;;
;; Revision 1.13  2001/10/21 14:36:56  jslopez
;; Added jde-ant customization group and
;; moved all its customization variables under
;; that group.
;;
;; Revision 1.12  2001/10/21 05:52:24  paulk
;; Added support auto completion of Ant targets in mini-buffer.
;; Added the following defuns and variables:
;;     defun jde-ant-find-build-file
;;     defun jde-ant-get-target-alist
;;     defcustom jde-ant-complete-target
;;     defcustom jde-ant-target-regexp
;; Modified defun jde-ant-read-target to use auto-completion, if
;; jde-ant-complete-target is non-nil. Thanks to Stephen Molitor.
;;
;; Revision 1.11  2001/10/21 05:17:54  paulk
;; Restored changes from Revison 1.9.
;;
;; Revision 1.9  2001/10/05 15:04:32  jslopez
;; Removed calls that were setting the defaul-directory in
;; jde-ant-projecthelp and jde-ant-build. There are not needed.
;;
;; Revision 1.8  2001/10/04 04:33:54  paulk
;; Start change history.
;;
;;

;; End of jde-ant.el
