/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
	AMC 12/08/2002	Now pass classpath to build event trace
	ASC 13/08/2002  Call compiler monitor to output outstanding problems
**********************************************************************/
package org.eclipse.ajdt.internal.builder;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.aspectj.ajde.Ajde;
import org.eclipse.ajdt.internal.core.AJDTEventTrace;
import org.eclipse.ajdt.internal.ui.AspectJPreferences;
import org.eclipse.ajdt.internal.ui.ajde.CompilerMonitor;
import org.eclipse.ajdt.internal.ui.ajde.ProjectProperties;
import org.eclipse.ajdt.internal.ui.editor.AspectJContentOutlinePage;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.swt.widgets.Shell;


/**
 * This Builder selects the appropriate .lst (build config file) for
 * the project and then invokes the ajc compiler.
 * No incremental compilation support as yet - recommend turning
 * the "perform build automatically on resource modification" option
 * off under workbench->preferences.
 */
public class Builder extends IncrementalProjectBuilder {

	/**
	 * The name of the default build config file for an AspectJ project
	 */
	public static final String DEFAULT_CONFIG_FILE = "default.lst";

	/**
	 * Code returned when a failure occurs creating a build config file
	 */
	private static final int CONFIG_FILE_WRITE_ERROR = -10;

	/**
	 * Default timeout value for an AspectJ compile
	 */
	private static final int DEFAULT_TIMEOUT = 120;

	/**
	 * Preference setting for compile timeout
	 */
	private static final String COMPILE_TIMEOUT =
		"org.aspectj.ajdt.ui.compile-timeout";

	/**
	 * The last project we did a build for, needed by content outline
	 * view to decide which updates to accept.
	 */
	private static IProject lastBuiltProject = null;

	/**
	 * The progress monitor used for this build
	 */
	private IProgressMonitor monitor;

	/**
	 * Constructor for AspectJBuilder
	 */
	public Builder() {}

	/**
	 * What did we last build?
	 */
	public static IProject getLastBuildTarget( ) {
		return lastBuiltProject;
	}

	/**
	 * @see IncrementalProjectBuilder#build(int, Map, IProgressMonitor)
	 * kind is one of: FULL_BUILD, INCREMENTAL_BUILD or AUTO_BUILD
	 * currently we do a full build in every case!
	 */
	protected IProject[] build (
		int kind,
		Map args,
		IProgressMonitor progressMonitor)
		throws CoreException {
		try {
			if ( kind == AUTO_BUILD && AspectJPreferences.isAutobuildSuppressed() ) {
				return null;
			}

			monitor = progressMonitor;			
			AspectJPlugin ajPlugin = AspectJPlugin.getDefault();

			IProject project = getProject();
			ajPlugin.setCurrentProject(project);
			AJDTEventTrace.build( project, AspectJPlugin.getBuildConfigurationFile( project ),
								  ajPlugin.getAjdtProjectProperties().getClasspath() );

			ProjectProperties props = ajPlugin.getAjdtProjectProperties();
			List projectFiles = props.getProjectSourceFiles( project );
			
			// AJDE bug workaround
			if ( projectFiles.isEmpty() ) { return null; }
			
			props.clearMarkers();

			CompilerMonitor compilerMonitor = ajPlugin.getCompilerMonitor();
			compilerMonitor.prepare(projectFiles, progressMonitor);

			updateBuildConfigIfNecessary(project, projectFiles);
			lastBuiltProject = project;
			try {
				Ajde.getDefault().getBuildManager().build(getBuildFilePath(project));
			} catch (Throwable t) {
				t.printStackTrace();
			}

			waitForBuildCompletion(compilerMonitor);

            // Bug22258: Get the compiler monitor to display any issues with that compile.
            CompilerMonitor.showOutstandingProblems();
            
            // Tell the outline view to perform any marker updates pending from
            // the compile.
            AspectJContentOutlinePage.performPendingAdviceMarkerUpdates();
            
		} catch (Throwable t) {
			Ajde.getDefault().getErrorHandler().handleError("Compile failed.", t);
		}

		IProject[] requiredResourceDeltasOnNextInvocation = null;
		return requiredResourceDeltasOnNextInvocation;
	}

	/**
	 * Wait up to COMPILE_WAIT_TIME seconds for a build to finish
	 */
	private void waitForBuildCompletion(CompilerMonitor monitor) {
		int timesTried = 0; 
		
		// we learn the appropriate timeout over time...
		int timeout = DEFAULT_TIMEOUT;
		IPreferenceStore store = AspectJPlugin.getDefault().getPreferenceStore();
		if ( store.contains( COMPILE_TIMEOUT ) ) {
			timeout = store.getInt( COMPILE_TIMEOUT );
		} else {
			store.setValue( COMPILE_TIMEOUT, timeout );
		}
		
		while (!monitor.finished() && timesTried < timeout ) {
			timesTried++;
			try {
				Thread.sleep(1000);
			} catch (Exception e) {}

		}
		if (timesTried == timeout) {
			if (continueCompilation( monitor )) {
					waitForBuildCompletion( monitor );
			} 
		}
	}

	/**
	 * The timeout's fired waiting for a compile to finish...
	 * Could be a BIG project, could be a compiler bug.
	 * Let's see what the user wants to do.
	 */
	private boolean continueCompilation(final CompilerMonitor monitor) {
		
		final String title = AspectJPlugin.getResourceString( "suspiciouslyLongCompileDialog");
		final String message = AspectJPlugin.getResourceString( "isYourProjectReallyBig" );
		
		AspectJPlugin.getDefault().getDisplay().syncExec(new Runnable() {
			public void run() {
				boolean keepWaiting = false;
				Shell[] s = AspectJPlugin.getDefault().getDisplay().getShells();
				if ( s!= null && s.length > 0 ) {
					Shell shell = s[0];
					keepWaiting = MessageDialog.openQuestion( shell, title, message );
				}	
		
				if (keepWaiting) {
					//exponential growth to avoid bugging user!
					IPreferenceStore store = AspectJPlugin.getDefault().getPreferenceStore();
					int timeout = store.getInt( COMPILE_TIMEOUT );
					timeout *= 2;
					store.setValue( COMPILE_TIMEOUT, timeout );		
				} else {
					monitor.finish( );					
				}
			}
		});
		
		return !monitor.finished();
	}


	/**
	 * Get the fully qualified path to the build configuation file specified
	 * for this project.
	 */
	private String getBuildFilePath(IProject project) {
		String buildfile = AspectJPlugin.getBuildConfigurationFile(project);
		return buildfile;
	}

	/**
	 * If the project uses a default build file, then it will be regenerated
	 * to ensure that all project files are included.
	 */
	private void updateBuildConfigIfNecessary(IProject project, List projectFiles)
		throws CoreException {
		if (getBuildFilePath(project).endsWith(DEFAULT_CONFIG_FILE)) {
			writeBuildConfigFile(projectFiles, project);
		}
	}

	/**
	 * Create a full build configuration file for this project
	 */
	private void writeBuildConfigFile(List projectFiles, IProject project)
		throws CoreException {
		String configurationFilename = getBuildFilePath(project);

		try {
			FileWriter fw = new FileWriter(configurationFilename);
			BufferedWriter bw = new BufferedWriter(fw);
			for (Iterator it = projectFiles.iterator(); it.hasNext();) {
				File jf = (File) it.next();
				String fileName = jf.toString();
				if (fileName.endsWith(".java") || fileName.endsWith(".aj")) {
					bw.write(fileName);
					bw.write("\n");
				}
			}
			bw.flush();
			fw.flush();
			fw.close();
			// now tell eclipse the file has been updated
			IResource res = project.findMember(DEFAULT_CONFIG_FILE);
			if (res != null) {
				res.refreshLocal(IResource.DEPTH_ZERO, monitor );
			}
		} catch (Exception e) {
			Status status =
				new Status(
					Status.ERROR,
					AspectJPlugin.PLUGIN_ID,
					CONFIG_FILE_WRITE_ERROR,
					AspectJPlugin.getResourceString("configFileCreateError"),
					e);
			throw new CoreException(status);
		}

	}
}