/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
**********************************************************************/
package org.eclipse.ajdt.internal.core;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.aspectj.ajde.ui.AbstractIcon;
import org.aspectj.ajde.ui.StructureViewNode;
import org.aspectj.asm.InheritanceAssociation;
import org.aspectj.asm.IntroductionAssociation;
import org.aspectj.asm.LinkNode;
import org.aspectj.asm.ProgramElementNode;
import org.aspectj.asm.ReferenceAssociation;
import org.aspectj.asm.Relation;
import org.aspectj.asm.RelationNode;
import org.aspectj.asm.StructureNode;
import org.aspectj.bridge.ISourceLocation;
import org.eclipse.ajdt.internal.core.resources.AJDTIcon;
import org.eclipse.ajdt.internal.core.resources.AspectJImages;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.ui.model.IWorkbenchAdapter;

public class AJDTStructureViewNode implements StructureViewNode, IAdaptable {

	private static final String JDT_IMPORTS_LABEL = "import declarations";
	private static final String FWD_INTRO_REL_NAME = "declares member on";

	private List children;
	private AJDTIcon icon;
	private StructureNode node;
	private AJDTStructureViewNode parent = null;
	private IMarker marker = null;
	private String label = null;

	/**
	 * Constructor for AJDTSructureViewNode.
	 */
	private AJDTStructureViewNode() {
		super();
	}

	/**
	 * Constructor used only to add dummy root node at top of tree
	 */
	public AJDTStructureViewNode( AJDTStructureViewNode child ) {
		this.node = child.node;
		this.icon = AJDTIcon.MISSING_ICON;
		this.children = new ArrayList( );
		children.add( child );
		child.setParent( this );
		this.label = "<root>";	
	}

	public AJDTStructureViewNode( StructureNode node, AbstractIcon icon, 
									List children) 
	{
		this.node = node;
		this.icon = (AJDTIcon) icon;
		this.children = children;
		this.label = null;
		
		// tell all the children who their parent is			
		if ( children != null ) {
			for ( Iterator it = children.iterator(); it.hasNext(); ) {
				AJDTStructureViewNode child = (AJDTStructureViewNode)
					it.next();
				child.setParent( this );	
			}				
		}
		
//		System.out.println( "Created node " + node.getName());
//		if (node.getChildren() != null) {
//		System.out.println( "Children: " );
//			for (Iterator it = node.getChildren().iterator(); it.hasNext();){
//				Object next = it.next();
//				if ( next instanceof StructureNode ) {
//					System.out.println( "    SN " + ((StructureNode)next).getName() + " " + next.getClass());
//					if (next instanceof ProgramElementNode) {
//						ProgramElementNode pe = (ProgramElementNode) next;
//						System.out.println( "       Kind: " + pe.getProgramElementKind());
//						System.out.println( "       Member Kind?: " + pe.isMemberKind());					
//					}				
//				} else if ( node instanceof LinkNode ) {
//					System.out.println( "    LN " + ((LinkNode)next).getName());				
//				} else if ( node instanceof RelationNode ) {
//					System.out.println( "    RN " + ((RelationNode)next).getName());				
//				} else {
//					System.out.println( "    " + next.getClass());	
//				}
//			}
//		} 
//		System.out.println( "End Node.");
	}

	/**
	 * @see StructureViewNode#getStructureNode()
	 */
	public StructureNode getStructureNode() {
		return node;
	}

	/**
	 * @see StructureViewNode#getIcon()
	 */
	public AbstractIcon getIcon() {
		return icon;
	}

	/**
	 * @see StructureViewNode#getChildren()
	 */
	public List getChildren() {
		return children;
	}

	/**
	 * Return the parent of this node in the tree
	 */
	public AJDTStructureViewNode getParent( ) {
		return parent;
	}

	/**
	 * Get the label to use to describe this node
	 */
	public String getLabel( ) {
		if ( label == null ) {
			label = node.getName();
			if ( node instanceof LinkNode ) {
				label = ((LinkNode)node).toString();
//				label = getLinkLabel((LinkNode)node);
			} else if ( node instanceof RelationNode ) {
				Relation r = ((RelationNode)node).getRelation();
				if ( r.equals(ReferenceAssociation.IMPORTS_RELATION) ) {
					label = JDT_IMPORTS_LABEL;
				}
			} 
		}
		return label; 	
	}

//  	private String getLinkLabel(LinkNode node) {
//        String name = "";
//        ProgramElementNode programElementNode = node.getProgramElementNode();
//        if (programElementNode.getProgramElementKind().equals(ProgramElementNode.Kind.ADVICE) ||
//            programElementNode.getProgramElementKind().equals(ProgramElementNode.Kind.INTRODUCTION) ||
//            programElementNode.getProgramElementKind().equals(ProgramElementNode.Kind.CODE)) {
//            name = programElementNode.getParent().toString() + ": " + programElementNode.getName();
//			StructureNode grandparent = programElementNode.getParent().getParent();
//            if ( grandparent instanceof ProgramElementNode ) {
//            	ProgramElementNode pe_grandparent = (ProgramElementNode)grandparent;
//            	if ( pe_grandparent.getProgramElementKind() == ProgramElementNode.Kind.CLASS ||  
//            		 pe_grandparent.getProgramElementKind() == ProgramElementNode.Kind.INTERFACE
//            		) {
//            		name = grandparent.toString() + "." + name;	
//            	}
//            }
//            
//        } else if (programElementNode.isMemberKind()) {
//            name = programElementNode.getParent().toString() + '.' + programElementNode.getName(); 
//        } else {
//            name = programElementNode.toString();
//        }
//        return name;
//    }

	/**
	 * Set the parent of this node - called from parent's
	 * constructor
	 */
	private void setParent( AJDTStructureViewNode parent ) {
		this.parent = parent;
		if ( isImportsBranch() ) {
			icon = AspectJImages.JDT_IMPORTED;
		}
//		if ( node instanceof ProgramElementNode  &&
//		     parent.node instanceof ProgramElementNode ) {
//			ProgramElementNode pNode = (ProgramElementNode) node;			
//			if ( pNode.isRunnable() && pNode.getProgramElementKind() == ProgramElementNode.Kind.METHOD) {
//				pNode.setRunnable( false );
//				((ProgramElementNode)parent.node).setRunnable(true);
//			}
//		}
	}
	
	/**
	 * @see IAdaptable#getAdapter(Class)
	 */
	public Object getAdapter(Class adapter) {
		if (adapter == IWorkbenchAdapter.class) {
			return AJDTStructureViewNodeAdapter.getDefault( );
		} else {
			return null;
		}
	}


	/**
	 * Return an IMarker for the resource represented by this node
	 */
	public IMarker getMarker( ) {
		if ( marker != null ) return marker;
		
		ProgramElementNode targetNode = null;
		
		if ( node instanceof LinkNode ) {
			targetNode = ((LinkNode)node).getProgramElementNode();
		} else if (node instanceof ProgramElementNode) {
			targetNode = (ProgramElementNode) node;
		}
		
		if ( targetNode != null ) {		
			ISourceLocation sLoc = targetNode.getSourceLocation();
			if (sLoc != null && sLoc.getSourceFile().getAbsolutePath() != null) {
				IResource ir =
					AspectJPlugin.getDefault().getAjdtProjectProperties().findResource( sLoc.getSourceFile().getAbsolutePath() );
				if (ir != null) {
					try {
						marker = ir.createMarker(IMarker.MARKER);
						marker.setAttribute(IMarker.LINE_NUMBER, sLoc.getLine());
						marker.setAttribute(IMarker.CHAR_START, sLoc.getColumn());
					} catch (CoreException coreEx) {
						System.err.println( coreEx );
						AspectJPlugin.getDefault().getLog().log(coreEx.getStatus());
					}
				}
			}		
		}
		
		return marker;	
	}
	
	
	/**
	 * Determine whether this node is part of an "import declarations"
	 * sub-branch
	 */
	private boolean isImportsBranch(  ) {
		boolean retVal = false;
		AJDTStructureViewNode parentNode = getParent();
		if ( parentNode != null ) {
			StructureNode sNode = parentNode.getStructureNode();
			if ( sNode instanceof RelationNode ) {
				RelationNode rNode = (RelationNode) sNode;
				Relation r = rNode.getRelation();
				if ( r.equals(ReferenceAssociation.IMPORTS_RELATION) ) {
					retVal = true;
				}
			}
		}
		return retVal;	
	}

	/**
	 * Check to see if this node os part of an "implements" sub-branch
	 * @return boolean
	 */
	private boolean isImplementsBranch() {
		boolean retVal = false;
		AJDTStructureViewNode parentNode = getParent();
		if ( parentNode != null ) {
			StructureNode sNode = parentNode.getStructureNode();
			if ( sNode instanceof RelationNode ) {
				RelationNode rNode = (RelationNode) sNode;
				Relation r = rNode.getRelation();
				if ( r.equals(InheritanceAssociation.IMPLEMENTS_RELATION) ) {
					retVal = true;
				}
			}
		}
		return retVal;	
	}


	/**
	 * Check to see if this node os part of an introduction relationship sub-branch
	 * @return boolean
	 */
	private boolean isFwdIntroduction() {
		boolean retVal = false;
		AJDTStructureViewNode parentNode = getParent();
		if ( parentNode != null ) {
			StructureNode sNode = parentNode.getStructureNode();
			if ( sNode instanceof RelationNode ) {
				RelationNode rNode = (RelationNode) sNode;
				Relation r = rNode.getRelation();
				if ( r.equals(IntroductionAssociation.INTRODUCES_RELATION) ) {
					if ( sNode.getName( ).equals( FWD_INTRO_REL_NAME ) ) {
						retVal = true;
					}
				}
			}
		}
		return retVal;	
	}

	/**
	 * @see org.aspectj.ajde.ui.StructureViewNode#add(StructureViewNode)
	 */
	public void add(StructureViewNode child) {
		children.add( child );
		((AJDTStructureViewNode)child).setParent( this );
	}

	/**
	 * @see org.aspectj.ajde.ui.StructureViewNode#remove(StructureViewNode)
	 */
	public void remove(StructureViewNode child) {
		children.remove( child );
	}


	// -------------------------------------------------------------------------
	// support for outline view filtering and sorting
	
	// the order of categories for alphabetic sorting
	public interface Category {
			public final int PACKAGE      = 0;
			public final int IMPORTS      = 1;
			public final int CLASS        = 2;
			public final int INTERFACE    = 3;
			public final int ASPECT       = 4;
			public final int INTRODUCTION = 5;
			public final int FIELD        = 6;
			public final int POINTCUT     = 7;
			public final int DECLARATION  = 8;
			public final int ADVICE       = 9;	
			public final int CONSTRUCTOR  = 10;
			public final int METHOD       = 11;	
			public final int OTHER        = 12;
	};	


	private boolean computedCategory = false;
	private int category = Category.OTHER;
	
	public int category( ) {
		if ( !computedCategory ) {
			computedCategory = true;
			if ( node instanceof ProgramElementNode ) {
				ProgramElementNode pNode = (ProgramElementNode) node;
				ProgramElementNode.Kind kind = pNode.getProgramElementKind();
				if ( kind.equals( ProgramElementNode.Kind.PACKAGE ) ) {
					category = Category.PACKAGE;
				} else if ( kind.equals( ProgramElementNode.Kind.CLASS ) ) {
					category = Category.CLASS;
				} else if ( kind.equals( ProgramElementNode.Kind.INTERFACE ) ) {
					category = Category.INTERFACE;
				} else if ( kind.equals( ProgramElementNode.Kind.ASPECT ) ) {
					category = Category.ASPECT;
				} else if ( kind.equals( ProgramElementNode.Kind.INTRODUCTION ) ) {
					category = Category.INTRODUCTION;
				} else if ( kind.equals( ProgramElementNode.Kind.POINTCUT ) ) {
					category = Category.POINTCUT;
				} else if ( kind.equals( ProgramElementNode.Kind.DECLARE_ERROR ) ||
							 kind.equals( ProgramElementNode.Kind.DECLARE_WARNING ) ||
							 kind.equals( ProgramElementNode.Kind.DECLARE_PARENTS ) ||
							 kind.equals( ProgramElementNode.Kind.DECLARE_SOFT )
					 ) {
						category = Category.DECLARATION;
				} else if ( kind.equals( ProgramElementNode.Kind.ADVICE ) ) {
					category = Category.ADVICE;
				} else if ( kind.equals( ProgramElementNode.Kind.FIELD ) ) {
					category = Category.FIELD;
				} else if ( kind.equals( ProgramElementNode.Kind.CONSTRUCTOR ) ) {
					category = Category.CONSTRUCTOR;
				} else if ( kind.equals( ProgramElementNode.Kind.METHOD ) ) {
					category = Category.METHOD;
				}
			}	
			if ( getLabel( ).equals( JDT_IMPORTS_LABEL )) {
				category = Category.IMPORTS;
			}
		}		
		return category;
	}


	public boolean isPublic( ) {
		boolean isPublic = true;
		if ( node instanceof ProgramElementNode ) {
			ProgramElementNode pNode = (ProgramElementNode) node;
			ProgramElementNode.Accessibility acc = pNode.getAccessibility();
			if ( acc != null ) {
				if ( !acc.equals( ProgramElementNode.Accessibility.PUBLIC ) ) {
					isPublic = false;
				 }
			}
		}
		return isPublic;
	}
	
	public boolean isStatic( ) {
		boolean isStatic = false;
		if ( node instanceof ProgramElementNode ) {
			ProgramElementNode pNode = (ProgramElementNode) node;
			List modifiers = pNode.getModifiers();
			if ( modifiers != null ) {
				for( Iterator it = modifiers.iterator(); it.hasNext(); ) {
					if ( it.next( ).equals( ProgramElementNode.Modifiers.STATIC ) ) {
						isStatic = true;
						break;	
					}		
				}
			}
		}
		return isStatic;			
	}
	
}
