/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
**********************************************************************/
package org.eclipse.ajdt.internal.core;

import java.util.ArrayList;
import java.util.Hashtable;

import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.ui.JavaElementImageDescriptor;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.graphics.Point;

/**
 * A utility class to capture all those little functions that
 * keep cropping up. Converting IPaths to fully-qualified native
 * filenames being one of the most common!!
 */
public class AJDTUtils {

	// icon sizes
	private static final Point SMALL_SIZE= new Point(16, 16);
	private static final Point BIG_SIZE= new Point(22, 16);
	private static final int SMALL_ICONS_MASK = 0x020;

    private static Hashtable imageDescriptorCache = new Hashtable();

	/**
	 * Return the fully-qualifed native OS path of the workspace.
	 * e.g. D:\eclipse\workspace
	 */
	public static String getWorkspacePath( ) {
		return AspectJPlugin.getWorkspace().getRoot().getLocation().toOSString();
	}
	
	/**
	 * Return the fully-qualified name of the root directory for a
	 * project.
	 */
	public static String getProjectRootDirectory( IProject project ) {
		return project.getLocation().toOSString();	
	}
	
	/**
	 * Return the fully-qualified native OS path of a project resource
	 */
	public static String getResourcePath( IResource resource ) {
		return resource.getLocation( ).toOSString();
	}

	/**
	 * decorate an icon given a set of adornment flags.
	 * Constants for flag values are defined in JavaElementImageDescriptor
	 */
	public static ImageDescriptor decorate( ImageDescriptor base, int decorations ) {
		
		Point size= useSmallSize(decorations) ? SMALL_SIZE : BIG_SIZE;
		// Check the image descriptor cache
		String key = new String(base.toString()+":::"+decorations+":::"+size.toString());
		// Example key is "URLImageDescriptor(platform:/plugin/org.aspectj.ajde_1.1.0/icons/structure/file-lst.gif):::0:::Point {22, 16}"
		// or "URLImageDescriptor(platform:/plugin/org.eclipse.ui_2.0.2/icons/full/obj16/fldr_obj.gif):::0:::Point {22, 16}"
		if (imageDescriptorCache.get(key)!=null) {
			//System.err.println("IDCache hit for "+key);
			return (ImageDescriptor)imageDescriptorCache.get(key);
		} else {
			ImageDescriptor imageDescriptor = new JavaElementImageDescriptor( base, decorations, size);	
			imageDescriptorCache.put(key,imageDescriptor);
			//System.err.println("IDCache mis for "+key);
			return imageDescriptor;
		}
	}

	/**
	 * determine which icon size to use for a given set of decorations
	 */
	private static boolean useSmallSize(int flags) {
		return (flags & SMALL_ICONS_MASK) != 0;
	}	
	
	/**
	 * Attempt to update the project's build classpath with the AspectJ
	 * runtime library.
	 * @param project	
	 */
	public static void addAjrtToBuildPath(IProject project)
	{
		// Locate the aspectjrt.jar file.
		String ajrtPath =
			AspectJPlugin.
			getDefault().
			getAjdtProjectProperties().
			getAspectjrtClasspath();

		if (ajrtPath != null)
		{
			IJavaProject javaProject = JavaCore.create(project);
			try
			{
				IClasspathEntry[] originalCP = javaProject.getRawClasspath();
				IClasspathEntry ajrtCP =
					JavaCore.newLibraryEntry(
						new Path(ajrtPath), // library location
						null, 				// no source
						null 				// no source
						);
				// Update the raw classpath with the new ajrtCP entry.
				int originalCPLength = originalCP.length;
				IClasspathEntry[] newCP =
					new IClasspathEntry[originalCPLength + 1];
				System.arraycopy(originalCP, 0, newCP, 0, originalCPLength);
				newCP[originalCPLength] = ajrtCP;
				javaProject.setRawClasspath(newCP, new NullProgressMonitor());
			}
			catch (JavaModelException e)
			{
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Adds the AspectJ Nature to the project
	 * @param project
	 * @throws CoreException
	 */
	public static void addAspectJNature(IProject project)
		throws CoreException
	{
		// add the AspectJ Nature
		IProjectDescription description = project.getDescription();
		String[] prevNatures = description.getNatureIds();
		String[] newNatures = new String[prevNatures.length + 1];
		System.arraycopy(prevNatures, 0, newNatures, 0, prevNatures.length);
		newNatures[prevNatures.length] = AspectJPlugin.ID_NATURE;
		description.setNatureIds(newNatures);
		project.setDescription(description, null);

		// Update the build classpath to have the aspectjrt.jar		
		addAjrtToBuildPath(project);			
	}
	
	/**
	 * Removes the AspectJ Nature from an existing AspectJ project.
	 * @param project
	 * @throws CoreException
	 */
	public static void removeAspectJNature(IProject project)
		throws CoreException
	{
		// remove the AspectJ Nature
		IProjectDescription description = project.getDescription();
		String[] prevNatures = description.getNatureIds();
		String[] newNatures = new String[prevNatures.length - 1];
		int newPosition = 0;
		for (int i = 0; i < prevNatures.length; i++)
		{
			if (!prevNatures[i].equals(AspectJPlugin.ID_NATURE))
			{
				// guard against array out of bounds which will occur if we
				// get to here in a project that DOES NOT have the aj nature
				// (should never happen).
				if (newPosition < newNatures.length)
				{
					newNatures[newPosition++] = prevNatures[i];
				}
				else
				{
					// exception... atempt to remove ajnature from a project that
					// doesn't have it. Leave the project natures unchanged.
					newNatures = prevNatures;
					break;
				}// end else
			}// end if
		}// end for
		description.setNatureIds(newNatures);
		project.setDescription(description, null);
		
		// Update the build classpath to try and remove the aspectjrt.jar
		removeAjrtFromBuildPath(project);		
	}
	
	/**
	 * Attempt to update the project's build classpath by removing any 
	 * occurance of the AspectJ runtime library.
	 * @param project
	 */
	public static void removeAjrtFromBuildPath(IProject project)
	{
		IJavaProject javaProject = JavaCore.create(project);
		try
		{
			IClasspathEntry[] originalCP = javaProject.getRawClasspath();
			ArrayList tempCP = new ArrayList();
			
			// Go through each current classpath entry one at a time. If it
			// is not a reference to the aspectjrt.jar then do not add it 
			// to the collection of new classpath entries.
			for (int i=0 ; i<originalCP.length ; i++)
			{
				IPath path = originalCP[i].getPath();
				if (!path.toOSString().endsWith("aspectjrt.jar"))
				{
					tempCP.add(originalCP[i]);
				}
			}// end for
			
			// Set the classpath with only those elements that survived the
			// above filtration process.
			if (originalCP.length != tempCP.size())
			{
				IClasspathEntry[] newCP = 
					(IClasspathEntry[])tempCP.toArray(
						new IClasspathEntry[tempCP.size()]);
				javaProject.setRawClasspath(newCP, new NullProgressMonitor());
			}// end if at least one classpath element removed
		}
		catch (JavaModelException e)
		{
			// Thrown if attempted to add a duplicate classpath entry. 
			AJDTEventTrace.generalEvent(e.getMessage());
		}
	}
}
