/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
**********************************************************************/
package org.eclipse.ajdt.internal.ui;

/**
 * ToDos
 * 1) Work out how to operate under the AspectJ menu
 * 
 */
import java.io.File;
import java.util.List;
import java.util.ListIterator;

import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowPulldownDelegate;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.core.runtime.IAdaptable;

/**
 * This pulldown menu displays a list of the .lst files available in the current
 * project.  It highlights the current 'selected' .lst file with a checkbox.  By
 * selecting one of the other options, the user can change the selected .lst
 * file for the project.  If the button in the toolbar is operated, rather than
 * the pulldown arrow next to it, then it displays the current selected build
 * file for the project in a message dialog.
 */
public class PulldownBuildselectorMenu
	implements IWorkbenchWindowPulldownDelegate, SelectionListener {

	private IWorkbenchWindow window;
	private ISelection selection;

	/**
	 * Project for which the pulldown menu instance is being built.
	 */
	private IProject project = null;

	public void dispose() {
	}

	public void init(IWorkbenchWindow window) {
		this.window = window;
	}

	/**
	 * Return the location of the workspace, for example c:\eclipse\workspace
	 */
	public String getWorkspaceDirectory() {
		IProject currentProject = AspectJPlugin.getDefault().getCurrentProject();
		// If there are no AspectJ projects in the workspace, return null 
		if (currentProject == null) return null;
		IPath trimPath = currentProject.getLocation();
		trimPath = trimPath.removeLastSegments(1);
		return new String(trimPath.toOSString() + File.separator);
	}

	/**
	 * Called when the pulldown tab on the menu is clicked, this produces
	 * a menu that contains an up to date list of the current .lst files.
	 * 'this' is registered as the selection listener for each of those
	 * .lst files - so that if the user clicks any of the options on the
	 * menu then we are called back.
	 */
	public Menu getMenu(Control c) {
		Menu m = new Menu(c);
		boolean defaultShouldBeTicked = false;
		boolean oneChoiceSelected = false;
		// What is the current project?  Ask the plugin
		project = AspectJPlugin.getDefault().getCurrentProject();

		if (project == null) {
			// There are no AspectJ projects, so make a special
			// menu to inform the user.
			MenuItem defaultLstItem = new MenuItem(m, SWT.CHECK);
			defaultLstItem.setText("No AspectJ projects exist"); //ASCFIXME-toNLS
			defaultLstItem.setData(null);
			return m;
		}
		
		String trimPathString = getWorkspaceDirectory();

		// Retrieve a list of all the .lst files that exist in that project
		List lstFiles =
			AspectJPlugin.getDefault().getListOfConfigFilesForCurrentProject();

		if (lstFiles != null) {
			ListIterator li = lstFiles.listIterator();

			// Add the 'all project files' entry which means default.lst in the filesystem
			MenuItem defaultLstItem = new MenuItem(m, SWT.CHECK);
			defaultLstItem.setText("<all project files>"); //ASCFIXME-toNLS
			defaultLstItem.setData(null);
			defaultLstItem.addSelectionListener(this);

			// For each .lst file, add an item to the menu
			MenuItem mi = null;
			while (li.hasNext()) {
				IResource iresource = (IResource) li.next();
				if (!iresource.getName().equals("default.lst")) {
					mi = new MenuItem(m, SWT.CHECK);
					String absolutePath = iresource.getLocation().toOSString();
					mi.setText(absolutePath.substring(trimPathString.length()));
					mi.addSelectionListener(this);
					// Remember the IResource in the 'data' part of the menuitem
					// for easy retrieval later on menuitem selection.
					mi.setData(iresource);
					String s = AspectJPlugin.getBuildConfigurationFile(project);
				}
				// If this menu item is the currently selected .lst file
				// for this project, then mark it with a checkbox.
				if (iresource
					.getLocation()
					.toOSString()
					.equals(AspectJPlugin.getBuildConfigurationFile(project))) {
						oneChoiceSelected = true;
					if (iresource.getName().equals("default.lst"))
						defaultShouldBeTicked = true;
					else
						mi.setSelection(true);
				}
			}

			if (defaultShouldBeTicked || !oneChoiceSelected )
				defaultLstItem.setSelection(true);
		}

		return m;
	}


	public void run(IAction action) {
		//ASCFIXME - toNLS all this bit		
		String workspaceDir = getWorkspaceDirectory();

		// Fixes Bug 25958.  Now builds selected project.		
		// Test whether the active window has a IResource selected, if so get its project.
		IResource selectedRes = null;
		if (selection instanceof IStructuredSelection) {			
			IStructuredSelection sel = (IStructuredSelection)selection;
			if (sel.getFirstElement() instanceof IAdaptable) {
				Object resource = ((IAdaptable)sel.getFirstElement()).getAdapter(IResource.class);
			   	if (resource != null) {
			    	selectedRes = (IResource)resource;
		  	   }
			}
		} 
		
		// If we can't get the selected project, try the instance var from the plugin.
		IProject project;
		if (selectedRes == null) {
			project = AspectJPlugin.getDefault().getCurrentProject();	
		} else {
			project = selectedRes.getProject();
		}
		
		
		// Fix bug 25958 to build selceted project
		
		MessageDialog dialog;
		
		if (workspaceDir == null) {
		  // There is no AspectJ project so tell the user.
		  
		  dialog =
			new MessageDialog(
				window.getShell(),
				"Build Configuration Selection", // ASCFIXME-NONLS
				null,
				"There are no AspectJ projects in the workspace",// ASCFIXME-NONLS
				MessageDialog.INFORMATION,
				new String[] { "OK" },
				0);
				
		  dialog.open();
				
		} else {
		
//		  dialog =
//			new MessageDialog(
//				window.getShell(),
//				"Build Configuration Selection for project: " + project.getName(),// ASCFIXME-NONLS
//				null,
//				AspectJPlugin.getBuildConfigurationFile(project).substring(
//					workspaceDir.length()),// ASCFIXME-NONLS
//				MessageDialog.INFORMATION,
//				new String[] { "OK" },
//				0);

			build( project );
		}
	}

	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

	/**
	 * Handle selection of an item in the menu.
	 */
	public void widgetDefaultSelected(SelectionEvent se) {
		chooseBuild(se);
	}

	/**
	 * Handle selection of an item in the menu.
	 */
	public void widgetSelected(SelectionEvent se) {
		chooseBuild(se);

	}

	/**
	 * When an item is selected from the pulldown menu, this function pulls the
	 * IResource from the menu item selected and makes that IResource the
	 * current build config for the project (remembered in the AspectJPlugin).
	 */
	public void chooseBuild(SelectionEvent se) {
		Object src = se.getSource();
		if (src instanceof MenuItem) {
			MenuItem w = (MenuItem) src;
			IResource ir = (IResource) w.getData();		
			AspectJPlugin.setBuildConfigurationFile(project, ir);
			build( project );
//			if (ir != null) 
//							System.err.println(
//							"Changing build config to "
//									+ ir.toString()
//									+ " for project "
//									+ project.toString());
//									else 
//									
//									System.err.println(
//							"Changing build config to default.lst for project "+project.toString());
									
		}
	}
	
	/**
	 * Once the button has been clicked, or a config selected, then perform a
	 * build immediately
	 */
	private void build( final IProject project ) {
		Shell activeShell = AspectJPlugin.getDefault().getActiveWorkbenchWindow().getShell();
		IRunnableWithProgress op = new IRunnableWithProgress( ) {
			public void run( IProgressMonitor pm ) {
				try {
					project.build(
						IncrementalProjectBuilder.FULL_BUILD, 
						AspectJPlugin.ID_BUILDER,
						null,
						pm );
				} catch ( CoreException cEx ) {
					AspectJPlugin.getDefault().getErrorHandler().handleError( 
						"Build pulldown error", cEx );	
				} catch ( Exception e ) {
					System.out.println( e );
					e.printStackTrace();
				}
			}
		};
		try { new ProgressMonitorDialog( activeShell).run( true, true, op ); }
		catch ( Exception e  ) {
			AspectJPlugin.getDefault( ).getErrorHandler().handleError(
			 "Auto build on select failed", e );
		}		
	}

}