/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
**********************************************************************/
package org.eclipse.ajdt.internal.ui.ajde;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.aspectj.ajde.Ajde;
import org.aspectj.ajde.BuildProgressMonitor;
import org.aspectj.ajde.TaskListManager;
import org.aspectj.bridge.IMessage;
import org.aspectj.bridge.ISourceLocation;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.swt.widgets.Shell;

/**
 * The compiler monitor interface is used by the AspectJ builder.  Just before
 * delegating the compilation of the source files to the 'external' ajc compiler,
 * the builder prepares a compiler monitor for this perspective.  The compiler
 * monitor is then called back through the CompilerMessages and CompileProgressMonitor
 * interfaces.  When the final message comes from the AJC compiler (it should be
 * finishProgress()) then the compiler monitor sets the finished flag which is
 * used by the builder to determine when it is safe to exit the build function.
 * This syncrhonization appears to be necessary because the compilation is
 * done asynchronously once the AJC compiler is started.
 */
public class CompilerMonitor implements TaskListManager, BuildProgressMonitor {

	private static final int MONITOR_MAX = 100;

	/**
	 *  Which Eclipse IProgressMonitor should this CompilerMonitor keep updating?
	 */
	private IProgressMonitor monitor = null;

	/**
	 *  Is this CompilerMonitor instance currently 'in use' ?
	 */
	private boolean monitoringInProgress = false;

	/**
	 *  problems for task list
	 */
	private static List problems = new ArrayList();

	/**
	 * Ratio of max value used by Ajde to MONITOR_MAX
	 */
	private float ajdeMonitorMaxRatio = 1.6f;

	/**
	 * Monitor progress against MONITOR_MAX
	 */
	private int currentProgress = 0;

	/**
	 * Monitor progress against Ajde max
	 */
	private int currentAjdeProgress;

	/**
	 * Called from the Builder to set up the compiler for a new build.
	 */
	public void prepare(List buildList, IProgressMonitor eclipseMonitor) {
		if (AspectJPlugin.DEBUG_COMPILER) {
			System.out.println(
				"CompilerMonitor.prepare called: IPM is "
					+ (eclipseMonitor == null ? "Null" : "Not-Null"));
		}

		monitor = eclipseMonitor;
		if (monitor != null) {
			monitor.beginTask(
				AspectJPlugin.getResourceString("ajCompilation"),
				MONITOR_MAX);

			monitoringInProgress = true;
		}
	}

	/********* BuildProgressMonitor interface implementation ************* START */

	/* Methods are called from Ajde in the following sequence:
	 *  - start( )
	 *  - clearTasks( )
	 *  - setProgressBarMax( max )
	 *  - followed by any number of setProgressBarVal(), incrementProgressBarVal( ) 
	 *    and setProgressText( ) calls in any order
	 *  - finish( )
	 */

	/**
	 * Ajde has started a compile
	 */
	public void start( final String configFile ) {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("CompileProgressMonitor.start() called");
		currentProgress = 0;
		currentAjdeProgress = 0;
		if (monitor != null) {
			AspectJPlugin.getDefault().getDisplay().asyncExec(new Runnable() {
				public void run() {
					if (monitor!=null) 
					monitor.subTask( "Building " + removePrefix(configFile) );
				}
			});
		}
	}

	/**
	 * Ajde wishes to display information about the progress of the 
	 * compilation.
	 */
	public void setProgressText(String text) {
		final String amendedText = removePrefix(text);
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("setProgressText(" + text + ")");
		if (monitor != null) {
			AspectJPlugin.getDefault().getDisplay().asyncExec(new Runnable() {
				public void run() {
					if (monitor!=null) 
					monitor.subTask(amendedText);
				}
			});
		}

	}

	/**
	 * Ajde wishes to advance the progress bar to this absolute value
	 */
	public void setProgressBarVal(int newVal) {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("setProgressBarVal(" + newVal + ")");
		int adjustedVal = Math.round(((float) newVal) / ajdeMonitorMaxRatio);
		while (currentProgress < adjustedVal) {
			incrementProgressBarVal();
		}
	}

	/**
	 * Ajde informs us that the compiler is assuming a maximum progress
	 * bar value of <code>maxVal</code>
	 */
	public void setProgressBarMax(int maxVal) {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("setProgressBarMax(" + maxVal + ")");
		ajdeMonitorMaxRatio = ((float) maxVal) / MONITOR_MAX;
	}

	/**
	 * Ajde asks us to increment the progress bar by one
	 */
	public void incrementProgressBarVal() {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("incrementProgressBarVal()");
		currentAjdeProgress++;
		// Bug 22258 - Reworked internals of run() method, it used to be responsible for
		// increasing currentProgress but I've changed it since if anything causes the execution
		// of the thread to hang then currentProgress doesn't increase and setProgressBarVal()
		// just loops calling this routine.
		final int copy_of_currentProgress = currentProgress;
		if (monitor != null) {
			AspectJPlugin.getDefault().getDisplay().asyncExec(new Runnable() {
				public void run() {
					int adjustedAjdeProgress =
						Math.round((float) currentAjdeProgress / ajdeMonitorMaxRatio);
					for (int i=copy_of_currentProgress;i<adjustedAjdeProgress;i++) 
						if (monitor!=null) monitor.worked(1);
				}
			});
		} 
		
		// Bug 22258
		// Ensure when we leave this function that currentProgress is set correctly to satisfy the
		// condition check in setProgressBarVal() - we can't rely on the updates scheduled on the
		// secondary thread happening right away.
		int adjustedAjdeProgress =
			Math.round((float) currentAjdeProgress / ajdeMonitorMaxRatio);			
		currentProgress=adjustedAjdeProgress;


	}

	/**
	 * Ajde asks our preference for the maximum value of the progress bar
	 */
	public int getProgressBarMax() {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("getProgressBarMax()");
		return MONITOR_MAX;
	}

	/**
	 * Ajde informs us that a compilation has finished.
	 * We may under some circumstances get multiple calls to finish.
	 * This method is marked synchronized to let one finish finish before
	 * another finish gets in!
	 */
	public synchronized void finish() {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("finish()");		
		// AMC - moved this next monitor var set outside of thread - 
		// this status change must be instantly visible
		monitoringInProgress = false;
		AspectJPlugin.getDefault().getDisplay().asyncExec(new Runnable() {
			public void run() {
				if (monitor != null) {
					// ask the project to perform a refresh to pick up the
					// newly generated classfiles - bug 30462
					refreshOutputDir( );
					monitor.worked(MONITOR_MAX);
					monitor.done();
					monitor = null;
				}
				//monitoringInProgress = false;
				// Attempting this here will lock the GUI up if there is another compile about to
				// be done.  So it has been moved to an equivalent call (showOutstandingProblems())
				// in the Builder.java file.
				//showMessages();
			}
		});
	}

	/**
	 * Classfiles have been generated in the projects output directory,
	 * but eclipse doesn't know about them yet... time to let eclipse know.
	 */
	private void refreshOutputDir( ) {
		try {
			IProject currProject = AspectJPlugin.getDefault().getCurrentProject();
			IJavaProject jProject = JavaCore.create(currProject);
			IResource res = jProject.getResource();
			res.refreshLocal( IResource.DEPTH_INFINITE, monitor );
			
		} catch (Exception ex) {
			Ajde.getDefault().getErrorHandler().handleError( "Unable to refresh project dir", ex);
		}		
	}

	/**
	 * Has the most recent compile finished?
	 */
	public boolean finished() {
		return !monitoringInProgress;
	}

	/********* CompileProgressMonitor interface implementation *************  END  */

	/**
	 * A compiler message has been emitted, which potentially includes the 
	 * fully qualifed path of a resource. Chop it down to just the project-relative
	 * portion.
	 */
	private String removePrefix(String msg) {
		String ret = msg;
		IProject p = AspectJPlugin.getDefault().getCurrentProject();
		String projectLocation = p.getLocation().toOSString();
		if (msg.indexOf(projectLocation) != -1) {
			ret = msg.substring(0,msg.indexOf(projectLocation)) + 
					msg.substring(msg.indexOf(projectLocation) + projectLocation.length());
		} else {
			projectLocation = projectLocation.replace('\\','/');
			if (msg.indexOf(projectLocation) != -1) {
				ret = msg.substring(0,msg.indexOf(projectLocation)) + 
						msg.substring(msg.indexOf(projectLocation) + projectLocation.length());
			}
		}
		
		// special cases...
		// this message is too long to be meaningful
		if ( ret.startsWith("might need to weave")) { ret = "weaving ..."; }
		// we always get this next one, and it seems to be nonsense
		if ( ret.startsWith("directory classpath entry does not exist: null")) { ret = ""; }
		return ret;
	}


	/********* TaskListManager interface implementation ************* START */


	/**
	 * Add a problem to the tasks list for the given file and line number
	 */
	public void addSourcelineTask(
		final String message,
		final ISourceLocation location,
		final IMessage.Kind kind
		) {

		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("CompilerMessage received ]" + message + "[");


		problems.add(new ProblemTracker(message, location, kind));
        // When will showMessages() get called if we are not 'finishing off a compilation' - the reason this
        // is important is that when AJDE is asked to build a model of a lst file, it calls this routine
        // with any errors it finds whilst parsing the file...  if no one calls showMessages then the
        // messages will be stuck in the problems List and not pushed out until later!
	}

	/**
	 * Add a problem to the tasks list associated with the current project
	 */
	public void addProjectTask(String message, IMessage.Kind kind) {
		problems.add(new ProblemTracker( message, null, kind ) );
	}

	/**
	 * Called from Ajde to clear all tasks in problem list
	 */
	public void clearTasks() {
		if (AspectJPlugin.DEBUG_COMPILER)
			System.err.println("clearTasks() called");
		problems = new ArrayList();
	}

	/********* TaskListManager interface implementation ************* END */

    /**
     * Callable from anywhere in the plugin, will put any unreported problems
     * onto the task bar.  This is currently used by the model builder for
     * build configuration files.  Ajde builds the model and reports errors through this 
     * class - BuildConfigurationEditor then asks this helper method to report
     * them.  We need to move this error reporting stuff out of here if it
     * is going to be used by more than just the compiler.
     */
    public static void showOutstandingProblems() {
    	if (problems.size()>0) {
    		new CompilerMonitor().showMessages();
    	}
    }

	private void showMessages() {
		
        // THIS MUST STAY IN A SEPARATE THREAD - This is because we need
        // to create and setup the marker in an atomic operation.  See
        // AMC or ASC.
		IWorkspaceRunnable r = new IWorkspaceRunnable() {
			public void run( IProgressMonitor monitor ) {
				Iterator problemIterator = problems.iterator();
				while ( problemIterator.hasNext() ) {
					ProblemTracker p = (ProblemTracker) problemIterator.next();
					IResource ir = null;
					if ( p.location != null ) {
						String loc = p.location.getSourceFile().getAbsolutePath();
						ir = AspectJPlugin.getDefault().getAjdtProjectProperties().
						       findResource(loc);
						if (ir == null) {
							// fix for declare warning/error bug which returns only file name
							// (unqualified)
							ir = tryToFindResource( p.location.getSourceFile().getPath() );
							//ir = AspectJPlugin.getDefault().getCurrentProject();
						}
						try {
							IMarker im = ir.createMarker(IMarker.PROBLEM);
							if ( p.location.getLine() > 0 ) {
								im.setAttribute(IMarker.LINE_NUMBER, new Integer(p.location.getLine()));
							}
							if ( p.location.getColumn() > 0 ) {
								// setting these causes the problem marker to appear in the
								// wrong place in the editor margin!!!
								//im.setAttribute(IMarker.CHAR_START , new Integer(p.location.getColumnNumber()));
								//im.setAttribute(IMarker.CHAR_END, new Integer( p.location.getColumnNumber() + 1 ));
							}
						
							if ( p.kind == IMessage.ERROR ) {
								im.setAttribute(IMarker.SEVERITY, new Integer(IMarker.SEVERITY_ERROR));
							} else if ( p.kind == IMessage.WARNING ) {
								im.setAttribute(IMarker.SEVERITY, new Integer(IMarker.SEVERITY_WARNING) );
							} else {
								im.setAttribute(IMarker.SEVERITY, new Integer(IMarker.SEVERITY_INFO) );
							}
		
			                // FIXME: Remove this horrid hack.
							// Hack the filename off the front and the line number off the end
							if (p.message.indexOf("\":")!=-1 && p.message.indexOf(", at line")!=-1) {
							  String hackedMessage = p.message.substring(p.message.indexOf("\":") + 2);
							  hackedMessage = hackedMessage.substring(0, hackedMessage.indexOf(", at line"));
							  im.setAttribute(IMarker.MESSAGE, hackedMessage);
							} else {
								im.setAttribute(IMarker.MESSAGE,p.message);
							}
						} catch (Exception e) {
							AspectJPlugin.getDefault().getErrorHandler().handleError( "Error creating marker", e );
						}
					}			
				}
				clearTasks( );												
			}
		};
		
		
		try {
			AspectJPlugin.getWorkspace().run( r, null );	
		} catch ( CoreException cEx ) {
			AspectJPlugin.getDefault().getErrorHandler().handleError( 
				"AJDT Error adding problem markers", cEx );			
		}
	}
	
	private IResource tryToFindResource( String fileName ) {		
		IResource ret = null;
		String toFind = fileName.replace('\\','/');
		IProject project = AspectJPlugin.getDefault().getCurrentProject();
		IJavaProject jProject = JavaCore.create(project);
		try {
			IClasspathEntry[] classpathEntries = jProject.getResolvedClasspath(false);
			for (int i = 0; i < classpathEntries.length; i++) {
				if (classpathEntries[i].getEntryKind() == IClasspathEntry.CPE_SOURCE) {
					IClasspathEntry sourceEntry = classpathEntries[i];
					IPath sourcePath = sourceEntry.getPath();
					// remove the first segment because the findMember call following
					// always adds it back in under the covers (doh!) and we end up
					// with two first segments otherwise!
					sourcePath = sourcePath.removeFirstSegments(1);
					IResource[] srcContainer = new IResource[] { project.findMember(sourcePath)};
					ret = findFile(srcContainer, toFind);
					if ( ret != null) break;					
				}
			}
		} catch (JavaModelException jmEx) {
			String message = AspectJPlugin.getResourceString("jmCoreException");
			Status status = new Status(Status.ERROR, AspectJPlugin.PLUGIN_ID, Status.OK, message, jmEx);
			Shell shell = AspectJPlugin.getDefault().getActiveWorkbenchWindow().getShell();
			org.eclipse.jface.dialogs.ErrorDialog.openError(shell, AspectJPlugin.getResourceString("ajErrorDialogTitle"), message, status);
		}

		if ( ret == null ) ret = project;
		return ret;	
	}

	private IResource findFile( IResource[] srcContainer, String name ) {
		IResource ret = null;
		try {
			for (int i = 0; i < srcContainer.length; i++) {
				IResource ir = srcContainer[i];
				if (ir.getFullPath().toString().endsWith(name)) {
					ret = ir;
					break;
				}
				if (ir instanceof IContainer) {
					ret = findFile( ((IContainer)ir).members(), name );
					if ( ret != null ) break;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return ret;	
	}

	/**
	 * Inner class used to ensure marker creation and attribute setting occur
	 * as an atomic unit (otherwise tick marks do not appear correctly in editor
	 * margin.
	 */
	class ProblemAdder implements IWorkspaceRunnable {
		private Iterator problemIterator;
		
		public ProblemAdder( Iterator it ) {
			this.problemIterator = it;	
		}
		
		public void run( IProgressMonitor monitor ) {
			System.out.println( "Adding problem markers" );
			while ( problemIterator.hasNext() ) {
				ProblemTracker p = (ProblemTracker) problemIterator.next();
				IResource ir = null;
				if ( p.location != null ) {
					String loc = p.location.getSourceFile().getAbsolutePath();
					ir = AspectJPlugin.getDefault().getAjdtProjectProperties().
					       findResource(loc);
					if (ir == null) {
						ir = AspectJPlugin.getDefault().getCurrentProject();
					}
					try {
						IMarker im = ir.createMarker(IMarker.PROBLEM);
						if ( p.location.getLine() > 0 ) {
							im.setAttribute(IMarker.LINE_NUMBER, p.location.getLine());
						}
						if ( p.location.getColumn() > 0 ) {
							//im.setAttribute(IMarker.CHAR_START , p.location.getColumnNumber());
							//im.setAttribute(IMarker.CHAR_END,...);
						}
					
						if ( p.kind == IMessage.ERROR ) {
							im.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
						} else if ( p.kind == IMessage.WARNING ) {
							im.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_WARNING );
						} else {
							im.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_INFO );
						}
	
		                // FIXME: Remove this horrid hack.
						// Hack the filename off the front and the line number off the end
						if (p.message.indexOf("\":")!=-1 && p.message.indexOf(", at line")!=-1) {
						  String hackedMessage = p.message.substring(p.message.indexOf("\":") + 2);
						  hackedMessage = hackedMessage.substring(0, hackedMessage.indexOf(", at line"));
						  im.setAttribute(IMarker.MESSAGE, hackedMessage);
						} else {
							im.setAttribute(IMarker.MESSAGE,p.message);
						}
					} catch (Exception e) {
						AspectJPlugin.getDefault().getErrorHandler().handleError( "Error creating marker", e );
					}
				}			
			}
			clearTasks( );						
		}	// end of run method
		
	}; // end of ProblemAdder class

	/**
	 * Inner class used to track problems found during compilation
	 * Values of -1 are used to indicate no line or column number available.
	 */
	class ProblemTracker {

		public ISourceLocation location;
		public String message;
		public IMessage.Kind kind;

		public ProblemTracker(String m, ISourceLocation l, IMessage.Kind k ) {
			location = l;
			message = m;
			kind = k;
		}
	};
}