/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
 
   AMC 08/12/2002  Changed getAspectjrtClasspath to get version info 
                   from Ajde instead of being hard-coded.
                   
   Geoff Longman 11/27/2002 Change getClasspath to retrieve entire classpath from
    						Project dependencies.

**********************************************************************/
package org.eclipse.ajdt.internal.ui.ajde;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import org.aspectj.ajde.Ajde;
import org.aspectj.ajde.ProjectPropertiesAdapter;
import org.eclipse.ajdt.internal.ui.AspectJPreferences;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.IPluginRegistry;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.PluginVersionIdentifier;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.swt.widgets.Shell;

/**
 * ProjectProperties is used to pass all the user, project and plugin settings to 
 * AJ Tools.
 */
public class ProjectProperties implements ProjectPropertiesAdapter {

	// when working from a development build of AspectJ (rather than a 
	// release), the version string comes back as "DEVELOPMENT." This
	// breaks the classpath creation, so this constant tells us a version
	// number to use instead.
	private static final String DEV_AJDE_LEVEL = "1.1.1";

	/**
	 * A map from java.io.File to the corresponding IResource in the project.
	 */
	private Hashtable fileToResourceHt = null;

	/**
	 * Computed classpath to aspectjrt.jar
	 */
	private String aspectjrtPath = null;

	/*
	 * @see ProjectPropertiesAdapter#getAjcWorkingDir()
	 */
	public String getAjcWorkingDir() {
		return AspectJPlugin.getWorkspace().getRoot().getLocation().toOSString();
	}

	/**
	 * Called to determine where the resultant class files should
	 * go when AJC is compiling code.  We grab the location from
	 * the current project and convert it to a string.
	 */
	public String getOutputPath() {
		try {
			IProject currProject = AspectJPlugin.getDefault().getCurrentProject();

			// Bug21998: Use the project location rather than the workspace location.
			String workspaceLoc = currProject.getLocation().removeLastSegments(1).toOSString();

			IJavaProject jProject = JavaCore.create(currProject);
			String relativePath = jProject.getOutputLocation().toOSString();
			return workspaceLoc + relativePath;
		} catch (JavaModelException jme) {
			Ajde.getDefault().getErrorHandler().handleError(AspectJPlugin.getResourceString("noOutputDir"), jme);
		}
		return null;
	}

	/**
	 * The name of the current project
	 */
	public String getProjectName() {
		return AspectJPlugin.getDefault().getCurrentProject().getName();
	}

	/*
	 * @see ProjectPropertiesAdapter#getClassToExecute()
	 */
	public String getClassToExecute() {
		System.err.println("*** AJ Plugin: ProjectProperties.getClassToExecute NOT IMPLEMENTED ***");
		return null;
	}

	/**
	 * The home directory of the current project
	 */
	public String getRootProjectDir() {
		IProject project = AspectJPlugin.getDefault().getCurrentProject();
		return project.getLocation().toOSString();
	}

	/**
	 * All the source files in the current project, as a List of
	 * java.io.Files.
	 */
	public List getProjectSourceFiles() {
		IProject activeProject = AspectJPlugin.getDefault().getCurrentProject();
		return getProjectSourceFiles(activeProject);
	}

	/**
	 * version to use when you know the project
	 */
	public List getProjectSourceFiles(IProject project) {

		List allProjectSourceFiles = new ArrayList();
		fileToResourceHt = new Hashtable();

		// bug 23045 - we don't want every file in the project, just those in
		// the active source directories	
		IJavaProject jProject = JavaCore.create(project);
		try {
			IClasspathEntry[] classpathEntries = jProject.getResolvedClasspath(false);
			for (int i = 0; i < classpathEntries.length; i++) {
				if (classpathEntries[i].getEntryKind() == IClasspathEntry.CPE_SOURCE) {
					IClasspathEntry sourceEntry = classpathEntries[i];
					IPath sourcePath = sourceEntry.getPath();
					// remove the first segment because the findMember call following
					// always adds it back in under the covers (doh!) and we end up
					// with two first segments otherwise!
					sourcePath = sourcePath.removeFirstSegments(1);
					IResource[] srcContainer = new IResource[] { project.findMember(sourcePath)};
					getAllFiles(srcContainer, allProjectSourceFiles, fileToResourceHt);
				}
			}
		} catch (JavaModelException jmEx) {
			String message = AspectJPlugin.getResourceString("jmCoreException");
			Status status = new Status(Status.ERROR, AspectJPlugin.PLUGIN_ID, Status.OK, message, jmEx);
			Shell shell = AspectJPlugin.getDefault().getActiveWorkbenchWindow().getShell();
			ErrorDialog.openError(shell, AspectJPlugin.getResourceString("ajErrorDialogTitle"), message, status);
		}

		return allProjectSourceFiles;
	}

	/*
	 * @see ProjectPropertiesAdapter#getProjectSourcePath()
	 */
	public String getProjectSourcePath() {
		IProject p = AspectJPlugin.getDefault().getCurrentProject();
		// todo - how to tell if the root of the project source tree??
		// is it a getPersistentProperty( )??
		return p.getLocation().toOSString();
	}

	/**
	 * get the classpath to use for compiling the current project.
	 */
	public String getClasspath() {
		StringBuffer classpath = new StringBuffer();

		// Related to bug21998: Use the project location rather than the workspace location, this allows
		// for projects who keep their source/binaries in a location other than the workspace directories.
		//String prependPath = AspectJPlugin.getWorkspace().getRoot().getLocation().toOSString();
		IProject currProject = AspectJPlugin.getDefault().getCurrentProject();

		classpath.append(getAspectjrtClasspath());
		classpath.append(File.pathSeparator);
		IProject project = AspectJPlugin.getDefault().getCurrentProject();

		Set projectClasspathSet = new HashSet();
		Set circularDependencyPreventer = new HashSet();

		circularDependencyPreventer.add(project);

		if (project != null) {
			try {

				getJavaProjectClasspath(currProject, projectClasspathSet, circularDependencyPreventer,false);

			} catch (CoreException corex) {
				String message = AspectJPlugin.getResourceString("jmCoreException");
				Status status = new Status(Status.ERROR, AspectJPlugin.PLUGIN_ID, Status.OK, message, corex);
				Shell shell = AspectJPlugin.getDefault().getActiveWorkbenchWindow().getShell();
				ErrorDialog.openError(shell, AspectJPlugin.getResourceString("ajErrorDialogTitle"), message, status);
			}
		}

		for (Iterator iter = projectClasspathSet.iterator(); iter.hasNext();) {
			classpath.append((String) iter.next());
			classpath.append(File.pathSeparator);
		}

		if (AspectJPlugin.isDebugging) {
			// It's *always* a classpath problem ;-). Better print it out.
			System.out.println("Using classpath: " + classpath);
		}
		return classpath.toString();
	}

	private void getJavaProjectClasspath(IProject project, Set projectClasspathSet, Set circularDependencyPreventer,
	  boolean processingDependantProjects)
		throws CoreException {
			
	    // Bug 30461 discussion: Rons bug about not supporting dependant projects correctly.
	    // The intent here is that the routine is gathering 'classpath elements' in order to build 
	    // a classpath that can be passed to AJC.  The interesting case is when you have a dependant
	    // project.  In this case recursion occurs - and this routine breaks down.  In the recursive case, we are
	    // processing the project upon which our *REAL* project depends.  And so we want to add components that this
	    // dependant project exports, not components that make up its own classpath.  To this end I have added the
	    // boolean flag to the parameter set for this method, which tells us whether we are in the recursive case.
	    // I then added code that (in the case of processing a dependant project) checks that the dependant
	    // project has each classpath entry marked as exported, before it is included in the classpath that
	    // is being built.
	    //
	    // However, it is also true that getResolvedClasspath() does not return the outputlocation of the project -
	    // this is perfectly reasonable.  So part (2) of this fix is to check if we are processing a dependant
	    // project (using the new flag) - if we are then we add the outputlocation of the project to the classpath
	    // we are building up.

		String projectLocalPrefix = File.separator + project.getName();

        // After this call prependPath will be something like "c:\eclipse\runtime-workspace"
		String prependPath = project.getLocation().removeLastSegments(1).toOSString();

		IJavaProject jProject = JavaCore.create(project);

        // After this call outputLocation will be something like "\MyProject\bin"
		String outputLocation = jProject.getOutputLocation().toOSString();
		
		// In the recursive case, the project we depend on should have its output directory
		// added to the classpath set.
		if (processingDependantProjects) projectClasspathSet.add(prependPath+outputLocation);

		IClasspathEntry[] classpathEntries = jProject.getResolvedClasspath(false);
		for (int i = 0; i < classpathEntries.length; i++) {

			String cpstring = classpathEntries[i].getPath().toOSString();
			
			// This says, for dependant project processing, check the classpath element is marked exported
			// before proceeding.
			if (!processingDependantProjects || classpathEntries[i].isExported()) {
			// what to do next depends on the entry kind
			switch (classpathEntries[i].getEntryKind()) {

				case IClasspathEntry.CPE_LIBRARY :
					// could be an external jar (fully qualified),
					// or a local jar (starts with our project name).
					// the latter needs qualifying for ajc						
					if (cpstring != null && cpstring.startsWith(projectLocalPrefix)) {
						cpstring = prependPath + cpstring;
					}
					break;

				case IClasspathEntry.CPE_PROJECT :

					String name = classpathEntries[i].getPath().lastSegment();
					IProject projectDependancy = AspectJPlugin.getWorkspace().getRoot().getProject(name);
					
					//					if classpathEntries[i].isExported()) 


					if (projectDependancy.exists()
						&& projectDependancy.isOpen()
						&& projectDependancy.hasNature(JavaCore.NATURE_ID)
						&& !circularDependencyPreventer.contains(projectDependancy)) {

						circularDependencyPreventer.add(projectDependancy);
						getJavaProjectClasspath(projectDependancy, projectClasspathSet, circularDependencyPreventer,true);

					}
					continue;

				case IClasspathEntry.CPE_SOURCE :
					// a source directory under my project, needs
					// qualification with the base location for the project.
					cpstring = prependPath + cpstring;
					break;

				case IClasspathEntry.CPE_VARIABLE :
				case IClasspathEntry.CPE_CONTAINER :
					// these two should already be qualified through use of
					// getResolvedClasspath	

				default :
					// do nothing
			}
			
			projectClasspathSet.add(cpstring);
			}

		}
	}

	/*
	 * @see ProjectPropertiesAdapter#getBootClasspath()
	 */
	public String getBootClasspath() {
		return null;
	}

	/*
	 * @see ProjectPropertiesAdapter#getExecutionArgs()
	 */
	public String getExecutionArgs() {
		String options = AspectJPreferences.getCompilerOptions();
		return options;
	}

	/*
	 * @see ProjectPropertiesAdapter#getVmArgs()
	 */
	public String getVmArgs() {
		return null;
	}

	// ----------- end of ProjectPropertiesAdapter interface methods ------------

	/**
	 * Called from builder before doing a build in order to 
	 * clear all problem markers
	 */
	public void clearMarkers() {
		IProject currProject = AspectJPlugin.getDefault().getCurrentProject();
		try {
			currProject.deleteMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
		} catch (Exception ex) {
		}
	}

	/**
	 * Return the IResource within the project that maps to the
	 * given File
	 */
	public IResource findResource(String fullPath) {
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IPath path = new Path(fullPath);
		return root.getFileForLocation(path);
	}

	/**
	 * Get the aspectjrt.jar classpath entry. This is usually in 
	 * plugins/org.aspectj.ajde_<VERSION>/aspectjrt.jar
	 */
	public String getAspectjrtClasspath() {
		
		if (aspectjrtPath == null) {
			StringBuffer cpath = new StringBuffer();
	
			IPluginRegistry reg = Platform.getPluginRegistry();

			String ajVersion = Ajde.getDefault().getVersion();
			if ( ajVersion.indexOf( "b" ) != -1 ) {
				// this is a beta version, replace beta with ".0" to
				// give plugin version
				ajVersion = ajVersion.substring(0,ajVersion.lastIndexOf("b")) + ".1";
			}
			if ( ajVersion.indexOf( "rc" ) != -1 ) {
				// this is a release candidate, replace rc with ".0" to
				// give plugin version
				ajVersion = ajVersion.substring(0,ajVersion.lastIndexOf("rc")) + ".1";
			}
			if ( ajVersion.equals( "DEVELOPMENT")) {
				// this breaks the classpath, so we have to use the
				// dev time constant that tells the level we're aiming for...
				ajVersion = DEV_AJDE_LEVEL;
			}
	
			int maj = 1;
			int min = 1;
			int svc = 1;
			try {
				StringTokenizer tok = new StringTokenizer(ajVersion,".");
				maj = Integer.parseInt(tok.nextToken());
				min = Integer.parseInt(tok.nextToken());
				svc = Integer.parseInt(tok.nextToken());
			} catch ( Exception ex ) {
				System.err.println( "Exception parsing AJDE version: " + ex );	
			}
	
	
			// first look for the version we really want...
			IPluginDescriptor ajdePluginDesc = 
				reg.getPluginDescriptor( "org.aspectj.ajde", new PluginVersionIdentifier(maj,min, svc));
	
			if (ajdePluginDesc == null) {
				// then try *any* version
				reg.getPluginDescriptor( "org.aspectj.ajde" );	
			}

			String pluginLoc = null;
			if ( reg != null ) {
				URL installLoc = ajdePluginDesc.getInstallURL();
				URL resolved = null;
				try {
					resolved = Platform.resolve(installLoc);
					pluginLoc = resolved.toExternalForm();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}			
			if ( pluginLoc != null ) {
				if ( pluginLoc.startsWith( "file:" )) {
					cpath.append(pluginLoc.substring( "file:".length() ));	
					cpath.append( "aspectjrt.jar" );
				}
			}
	
			// Verify that the file actually exists at the plugins location
			// derived above. If not then it might be because we are inside
			// a runtime workbench. Check under the workspace directory.
			if (new File(cpath.toString()).exists())	
			{
				// File does exist under the plugins directory 
				aspectjrtPath = cpath.toString();
			}
			else
			{
				// File does *not* exist under plugins. Try under workspace...
				IPath rootPath = AspectJPlugin.getWorkspace().getRoot().getLocation();
				IPath installPath = rootPath.removeLastSegments(1);
				cpath = new StringBuffer().append(installPath.toOSString());
				cpath.append(File.separator);
				cpath.append("workspace");
				cpath.append(File.separator);
				cpath.append("org.aspectj.ajde");
				cpath.append(File.separator);
				cpath.append("aspectjrt.jar");

				// Only set the aspectjrtPath if the jar file exists here.				
				if (new File(cpath.toString()).exists())
					aspectjrtPath = cpath.toString();
			}
		}
		return aspectjrtPath;	
	}

	/**
	 * Find all the ".java" and ".aj" files in the project.
	 */
	private void getAllFiles(IResource[] resource_list, List allProjectFiles, Hashtable fileToResourceHt) {
		try {
			for (int i = 0; i < resource_list.length; i++) {
				IResource ir = resource_list[i];
				if (ir.getName().endsWith(".java") || ir.getName().endsWith(".aj")) {
					allProjectFiles.add(new File(ir.getLocation().toOSString()));
					if (fileToResourceHt != null)
						fileToResourceHt.put(new File(ir.getLocation().toOSString()), ir);
				}
				if (ir instanceof IContainer) {
					getAllFiles(((IContainer) ir).members(), allProjectFiles, fileToResourceHt);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	/** New interface methods follow for build configuration management */

	/**
	 * Return a List containing strings, each string is the full path to 
	 * a build configuration file.
	 */
	public List getBuildConfigFiles() {
		List lstFiles_Strings = new ArrayList();
		List lstFiles_IResources = AspectJPlugin.getDefault().getListOfConfigFilesForCurrentProject();
		// Convert the IResource list to a list of strings for the full paths
		Iterator iter = lstFiles_IResources.iterator();
		IResource ir;
		while (iter.hasNext()) {

			ir = (IResource) iter.next();
			//System.err.println("AC_temp_debug: ProjectProperties.getBuildConfigFiles(): Path to a .lst is "+ir.getFullPath().toOSString());
			lstFiles_Strings.add(ir.getFullPath().toOSString());
		}
		return lstFiles_Strings;
	}

	public String getDefaultBuildConfigFile() {

		String defaultLstFile =
			(AspectJPlugin.getDefault().getCurrentProject().getLocation().toOSString()
				+ File.separator
				+ org.eclipse.ajdt.internal.builder.Builder.DEFAULT_CONFIG_FILE);
		//System.err.println("AC_temp_debug: ProjectProperties.getDefaultBuildConfigFile(): Returning "+defaultLstFile);

		return defaultLstFile;
		//return defaultLstFile + "TMP FIX DUE TO BUG IN AJDE";
	}

	public String getLastActiveBuildConfigFile() {

		String currentLstFile =
			AspectJPlugin.getBuildConfigurationFile(AspectJPlugin.getDefault().getCurrentProject());
		//System.err.println("AC_temp_debug: ProjectProperties.getLastActiveBuildConfigFile(): Returning "+currentLstFile);
		return currentLstFile;
	}
	
	// The following methods added for AspectJ 1.1
	//--------------------------------------------
	
	/**
	 * Get the set of input jar files for this compilation.
	 * Set members should be of type java.io.File.
	 * An empty set or null is acceptable for this option.
	 * From -injars.
	 */
	public Set getInJars() {
		return AspectJPlugin.getDefault().
		  getAjdtBuildOptionsAdapter().getInJars();
	}
	
	/**
	 * Get the output jar file for the compilation results.
	 * Return null to leave classfiles unjar'd in output directory
	 * From -outjar
	 */
	public String getOutJar() { 
		return AspectJPlugin.getDefault().
		  getAjdtBuildOptionsAdapter().getOutJar();
    } 
	
	/**
	 * Get a set of root source directories for the compilation.
	 * Set members should be of type java.io.File
	 * Returning null or an empty set disables the option.
	 * From -sourceroots
	 */
	public Set getSourceRoots() {
		return AspectJPlugin.getDefault().
		  getAjdtBuildOptionsAdapter().getSourceRoots();
	}
	
	/**
	 * Get the set of aspect jar files to be used for the compilation.
	 * Returning null or an empty set disables this option. Set members
	 * should be of type java.io.File.
	 * From -aspectpath
	 */
	public Set getAspectPath() { 
		return AspectJPlugin.getDefault().
		  getAjdtBuildOptionsAdapter().getAspectPath();
	}

	
}