/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer, Andy Clement, Tracy Gardner - initial version
...
**********************************************************************/
package org.eclipse.ajdt.internal.ui.editor;


import java.io.File;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.aspectj.asm.StructureNode;
import org.aspectj.bridge.ISourceLocation;
import org.eclipse.ajdt.internal.core.AJDTEventTrace;
import org.eclipse.ajdt.internal.core.AJDTUtils;
import org.eclipse.ajdt.internal.ui.AspectJPreferences;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.ajdt.ui.visualiser.StructureModelUtil;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.IWorkingCopyManager;
import org.eclipse.jdt.ui.text.JavaSourceViewerConfiguration;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.texteditor.IDocumentProvider;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;

/**
 * AspectJ Editor extends internal JDT editor in order to use our TextTools.
 * Better would be a clean interface in JDT to achieve the same effect.
 * We also override getAdapter( ) to plug-in the AspectJ-aware outline view.
 */
public class AspectJEditor extends CompilationUnitEditor {

	private AspectJContentOutlinePage contentOutlinePage;

    private boolean markersNeedUpdating = true;

	/**
	 * Constructor for AspectJEditor
	 */
	public AspectJEditor() {
		super();
				
		AspectJTextTools textTools= AspectJPlugin.getDefault().getAspectJTextTools();
		setSourceViewerConfiguration(new JavaSourceViewerConfiguration(textTools, this));
		
	}

	/**
	 * Get an adapter - if they want an outliner, give them ours!
	 * Other adapters we could consider providing:
	 * <ol>
	 * <li>ITextOperationTarget</li>
	 * <li>IFindReplaceTarget</li>
	 * </ol>
	 */
	public Object getAdapter( Class key ) {
		//System.err.println( "Asked for adapter: " + key.getName() );
		if ( key.equals( IContentOutlinePage.class ) ) {
			return getContentOutlinePage( key );			
		}		
		return super.getAdapter( key );
	}


	/**
	 *  Override of doSave to comment-out call to getStatusLineManager - 
	 * always returns null (why?) in our environment.
	 * Also ask the contentOutlinePage to update - not strictly required at the
	 * moment, but will be in the future.
	 */
	public void doSave(IProgressMonitor progressMonitor) {
		
		IDocumentProvider p= getDocumentProvider();
		if (p == null)
			return;
			
		if (p.isDeleted(getEditorInput())) {
			
			if (isSaveAsAllowed()) {
				
				/*
				 * 1GEUSSR: ITPUI:ALL - User should never loose changes made in the editors.
				 * Changed Behavior to make sure that if called inside a regular save (because
				 * of deletion of input element) there is a way to report back to the caller.
				 */
				 performSaveAs(progressMonitor);
			
			} else {
				
				/* 
				 * 1GF5YOX: ITPJUI:ALL - Save of delete file claims it's still there
				 * Missing resources.
				 */
				Shell shell= getSite().getShell();
				MessageDialog.openError(shell, "CompilationUnitEditor.error.saving.title1", "CompilationUnitEditor.error.saving.message1"); //$NON-NLS-1$ //$NON-NLS-2$
			}
			
		} else {	
			
			//getStatusLineManager().setErrorMessage(""); //$NON-NLS-1$
			
			IWorkingCopyManager manager= JavaPlugin.getDefault().getWorkingCopyManager();
			ICompilationUnit unit= manager.getWorkingCopy(getEditorInput());
			
			if (unit != null) {
				synchronized (unit) { 
					performSaveOperation(createSaveOperation(false), progressMonitor); 
				}
			} else 
				performSaveOperation(createSaveOperation(false), progressMonitor);
		}
		if (contentOutlinePage == null) {
			Object outlinePage = getContentOutlinePage( IContentOutlinePage.class );		
			if (outlinePage instanceof AspectJContentOutlinePage )
			  contentOutlinePage = (AspectJContentOutlinePage)outlinePage;	
		}
		
		// AMC - commented this out - it sometimes causes "Save Failed 'Null'"
		// problems and the update will never do anything since we don't have 
		// eager parsing and incremental compilation yet.
		
		//if (contentOutlinePage != null) 
		// contentOutlinePage.update( );
	}


	/**
	 * Get the content outline page - either the JDT one or our own
	 * depending on the preference setting.
	 */
	private Object getContentOutlinePage( Class key ) {
		Object outlinePage = null;

		if ( AspectJPreferences.isAspectJOutlineEnabled() ) {
			IEditorInput input = getEditorInput( );
			if ( input instanceof IFileEditorInput ) {
				IFile f = ((IFileEditorInput)input).getFile( );
				IProject p = f.getProject();
				try {
					if ( p.hasNature( AspectJPlugin.ID_NATURE ) ) {
						contentOutlinePage = new AspectJContentOutlinePage(
						  this,
						  f
						);	
						outlinePage = contentOutlinePage;
					} else { outlinePage = super.getAdapter( key ); }
				} catch ( CoreException cEx ) {
					outlinePage = super.getAdapter( key );
				}
			} else {
				outlinePage = super.getAdapter( key );		
			}
		} else {
			outlinePage = super.getAdapter( key );
		}

		return outlinePage;
	}



	public void doSetInput(IEditorInput input) throws CoreException {
		super.doSetInput(input);
		if (input instanceof IFileEditorInput) {
			IFileEditorInput fInput = (IFileEditorInput) input;
			AJDTEventTrace.editorOpened(fInput.getFile());
            // Ensure any advice markers are created since they are not persisted.
            updateAdviceMarkers(fInput); 
		}
	}



    /**
     * Shortcut method for updateAdviceMarkers() that takes a fileeditorinput
     */
    public void updateAdviceMarkers() {
    	updateAdviceMarkers((IFileEditorInput)getEditorInput());
    }


    
    /**
     * This method forces an update of the markers (i.e. they will all be deleted
     * and then readded) - this is called from the build updating code in
     * AspectJContentOutlinePage - to ensure that every editor has the right
     * markers after a compilation
     */
    public void forceUpdateOfAdviceMarkers() {
    	markersNeedUpdating = true;
    	updateAdviceMarkers((IFileEditorInput)getEditorInput());
    }


	
	/**
	 * Adds the advice markers for a file to the left hand gutter.
	 * It kicks off a thread that does a delete then adds all the new markers.
	 */
	public void updateAdviceMarkers(final IFileEditorInput fInput) {
		if (!markersNeedUpdating )
			return;
			
			 AspectJPlugin.getDefault( ).getDisplay().asyncExec( new Runnable( ) {
			  public void run( ) {
				try {
			      // Wipe all the current advice markers
 	              fInput.getFile().deleteMarkers("org.eclipse.ajdt.ui.advicemarker",
 	                true,IResource.DEPTH_INFINITE);	
	              	      
	              addNewMarkers(fInput);

				} catch (CoreException ce) {
                  AspectJPlugin.getDefault().getErrorHandler().
                    handleError("Advice marker updates failed", ce);	
				}
			}	
		} );	
	}	


	/**
	 * Adds advice markers to mark each line in the source that is affected by
	 * advice from an aspect.  It uses the StructureModelUtil code that the
	 * visualizer also uses - to determine what aspects are in effect on a
	 * specific source file.
	 * 
	 * @param fInput The file editor input resource against which the markers will be added.
	 */	
	public void addNewMarkers(IFileEditorInput fInput) {

		// Work out a pathname for the StructureModelUtil method.
		// Something like c:/eclipse/runtime-workspace/DemoProject/package1/source1.java

		// Get the path to the resource - e.g. "DemoProject\package1\source1.java"
		String fpath = fInput.getFile().getFullPath().toOSString();
		
		// Chop the project name off the front of that path. e.g. "package1\source1.java"
		fpath = fpath.substring(fpath.indexOf(File.separator) + 1);
		
		// Put the project root directory on the front of the resource path
		// e.g. "c:\eclipse\runtime-workspace\DemoProject"+"\"+"package1\source1.java"
		String path =
			AJDTUtils.getProjectRootDirectory(fInput.getFile().getProject())
				+ File.separator
				+ fpath.substring(fpath.indexOf(File.separator) + 1);

		// retrieve a map of line numbers to Vectors containing StructureNode objects
		// Ask for the detailed version of the map (by specifying 'true') which maps
		// line numbers to nodes representing advice (rather than just nodes representing
		// aspects).
		Map m = StructureModelUtil.getLinesToAspectMap(path,true);

		if (m != null) {

			// iterate through the line numbers in the map
			Set keys = m.keySet();
			Iterator i = keys.iterator();
			while (i.hasNext()) {
				Object o = i.next();
				final Integer linenumberInt = (Integer) o;
				
				// for that line, go through all the advice in effect
				Vector v = (Vector) m.get(o);
				for (int j = 0; j < v.size(); j++) {
					// sn will represent the advice in affect at the given line.
					final StructureNode sn = (StructureNode) v.get(j);
					final IResource ir     = (IResource) fInput.getFile();
					// Thread required to ensure marker created and set atomically 
					// (and so reflected correctly in the ruler).
					IWorkspaceRunnable r = new IWorkspaceRunnable() {
						public void run(IProgressMonitor monitor) {
							try {
								IMarker marker = ir.createMarker("org.eclipse.ajdt.ui.advicemarker");
								marker.setAttribute(IMarker.LINE_NUMBER,linenumberInt.intValue());

								ISourceLocation sl_sn = sn.getSourceLocation();
								
								marker.setAttribute(IMarker.MESSAGE,
									"Advice: "+sn.getName()+"  "+sn.getParent().getName()+ "(Lines "+sl_sn.getLine()+">"+sl_sn.getEndLine()+")"); 
								// was "Advice: "+ sn.getName()+ "."+ asn.getName());
										
								// Crude format is "FFFF:::NNNN:::NNNN:::NNNN"
								// Filename:::StartLine:::EndLine:::ColumnNumber
								
								// Grab the location of the pointcut
								ISourceLocation sLoc2 = sn.getSourceLocation(); // was asn
								
								marker.setAttribute(
									"sourceLocationOfAdvice",
									sLoc2.getSourceFile().getAbsolutePath()  + ":::"
						 			+ sLoc2.getLine()	   + ":::"
									+ sLoc2.getEndLine() + ":::"
									+ sLoc2.getColumn());
								/*
   								System.err.println(
									"Creating advicemarker at line="+ linenumberInt.intValue() +
									" advice="+ sn.getName() +
									"  sourcefilepath="	+ sLoc2.getSourceFile() +
								    " line="+ sLoc2.getLine());
								*/

							} catch (CoreException ce) {
								AspectJPlugin
									.getDefault()
									.getErrorHandler()
									.handleError(
									"Exception creating advice marker",
									ce);
							}
						}
					};

					// Kick off the thread to add the marker...
					try {
						AspectJPlugin.getWorkspace().run(r, null);
					} catch (CoreException cEx) {
						AspectJPlugin
							.getDefault()
							.getErrorHandler()
							.handleError(
							"AJDT Error adding advice markers",
							cEx);
					}
				}
			}
		}
		// Keep note that we are now up to date
		markersNeedUpdating = false;
	}


	
	public void dispose( ) {
		IEditorInput input = getEditorInput();
		if ( input instanceof IFileEditorInput ) {
			IFileEditorInput fInput = (IFileEditorInput) input;
			AJDTEventTrace.editorClosed( fInput.getFile() );
		}
		super.dispose();				
	}	
	
	
	
	/**
	 * @see org.eclipse.ui.IWorkbenchPart#setFocus()
	 */
	public void setFocus() {
		// AMC -added this method to catch user switching between multiple open files
		// in the editor - used to trigger a structured selection event, but now
		// triggers a TextSelection event. TextSelection event does not give AJP
		// enouhg info to determine project, so have to do from here instead.
		//System.out.println( "Focus given to: " + getEditorInput( ).getName() );
		IEditorInput input = getEditorInput();
		if ( input instanceof IFileEditorInput ) {
			IFileEditorInput fInput = (IFileEditorInput) input;
			AspectJPlugin.getDefault().setCurrentProject( fInput.getFile().getProject() );
            // Ensure any advice markers are created since they are not persisted - but
            // in this case (unlike when opening a new editor) they are likely to
            // be correct already.
	        updateAdviceMarkers(fInput);			
		}
		super.setFocus();
	}

}


