/**********************************************************************
Copyright (c) 2002 IBM Corporation and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
Contributors:
Adrian Colyer - <description of contribution>
...
**********************************************************************/
package org.eclipse.ajdt.internal.ui.wizards;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.aspectj.ajde.ui.AbstractIcon;
import org.aspectj.asm.ProgramElementNode;
import org.eclipse.ajdt.internal.core.resources.AJDTIcon;
import org.eclipse.ajdt.internal.core.resources.AspectJImages;
import org.eclipse.ajdt.ui.AspectJPlugin;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.IPluginRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.PluginVersionIdentifier;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.actions.WorkbenchRunnableAdapter;
import org.eclipse.jdt.internal.ui.util.ExceptionHandler;
import org.eclipse.jdt.internal.ui.wizards.NewElementWizard;
import org.eclipse.jdt.internal.ui.wizards.NewWizardMessages;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.actions.WorkspaceModifyDelegatingOperation;

/**
 * This wizard is used to create a new aspect. It bases from the 
 * *internal* JDT class creation wizard - ideally this would be 
 * exposed as a wizard we could legitimately extend.
 */
public class NewAspectWizard extends NewElementWizard {

	private NewAspectCreationWizardPage fPage;

	public NewAspectWizard() {
		super();
		AbstractIcon aspectIcon =
			AspectJImages.registry().getStructureIcon(ProgramElementNode.Kind.ADVICE);
		setDefaultPageImageDescriptor(((AJDTIcon) aspectIcon).getImageDescriptor());
		setDialogSettings(JavaPlugin.getDefault().getDialogSettings());
		setWindowTitle(
			AspectJPlugin.getResourceString("NewAspectCreationWizard.title"));
	}

	/*
	 * @see Wizard#createPages
	 */
	public void addPages() {
		super.addPages();
		IWorkspace workspace = JavaPlugin.getWorkspace();
		fPage = new NewAspectCreationWizardPage();
		addPage(fPage);
		fPage.init(getSelection());
	}

	/*
	 * Although this class is almost a duplicated of the wizard for creating new
	 * classes, the performFinish method *is* different.  Normally there are only
	 * two kinds of resource that can be created by the pages and underlying JDT
	 * infrastructure being reused in this wizard - either a class is being
	 * created or an interface.  It is the performFinish method that alters the
	 * class produced by the underlying wizard code such that it will say "aspect MyAspect"
	 * rather than "class MyAspect".
	 * 
	 * @see Wizard#performFinish
	 */
	public boolean performFinish() {
		// in eclipse 2.0.x perform finish has to call a version of 
		// finish page that takes an IRunnableWithProgress. In eclipse 2.1 and
		// later the processing is different and we should not interfere with
		// this method
		if (!isEclipse20x()) {
			return eclipse21PerformFinishImpl();
		} else {
			try{
				finishPage(fPage.getRunnable());
				makeAspect();
			} catch (Exception e) {
				System.err.println(e.getMessage());
			}
		}
		
		return true;
	}
	
	
	private boolean eclipse21PerformFinishImpl() {
		IWorkspaceRunnable op= new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException, OperationCanceledException {
				try {
					finishPage(monitor);
				} catch (InterruptedException e) {
					throw new OperationCanceledException(e.getMessage());
				}
			}
		};
		try {
			getContainer().run(false, true, new WorkbenchRunnableAdapter(op));
		} catch (InvocationTargetException e) {
			//handleFinishException(getShell(), e);
			// call this reflectively instead... we don't have all the types needed in eclipse 2.0
			try {
				Class superclass = getClass().getSuperclass();
				Method handleFinishException = superclass.getDeclaredMethod( "handleFinishException",
					new Class[]{ Shell.class, InvocationTargetException.class });
				handleFinishException.invoke(this,new Object[]{getShell(), e});
			} catch (Exception ex ) {
				System.err.println( "Reflective call to handleFinishException failed: " + ex);
			}
			return false;
		} catch  (InterruptedException e) {
			return false;
		}
		return true;
	}
	
	
	/**
	 * Declaration of this method needed for cross 2.0.x / 2.1 compatibility
	 * Under 2.1, do nothing (should never be called)
	 * Under 2.0, duplicate logic from super (can't call super, because is abstract
	 * in 2.1.
	 * @param r
	 * @return boolean
	 */
	protected boolean finishPage(IRunnableWithProgress runnable ) {
		if (!isEclipse20x()) return true;
		
		IRunnableWithProgress op= new WorkspaceModifyDelegatingOperation(runnable);
		try {
			getContainer().run(false, true, op);
		} catch (InvocationTargetException e) {
			Shell shell= getShell();
			String title= NewWizardMessages.getString("NewElementWizard.op_error.title"); //$NON-NLS-1$
			String message= NewWizardMessages.getString("NewElementWizard.op_error.message"); //$NON-NLS-1$
			ExceptionHandler.handle(e, shell, title, message);
			return false;
		} catch  (InterruptedException e) {
			return false;
		}
		return true;
	}
	
	/**
	 * This version needed for eclipse 2.1 or later, not called in
	 * eclipse 2.0.x
	 * @see org.eclipse.jdt.internal.ui.wizards.NewElementWizard#finishPage(IProgressMonitor)
	 */
	protected void finishPage(IProgressMonitor monitor) throws CoreException, InterruptedException {
		fPage.createType(monitor);
		makeAspect();
	}
	
	/**
	 * this logic refactored from performFinish so that we can create a
	 * wizard that across both 2.0.2 and 2.1 
	 */
	public void makeAspect() {
		ICompilationUnit cu = fPage.getCreatedType().getCompilationUnit();
		if (cu.isWorkingCopy()) {
			cu = (ICompilationUnit) cu.getOriginalElement();
		}
		
		try {
			IResource resource = cu.getUnderlyingResource();
		
			// Here is where the aspect wizard differs from the JDT class wizard.
			// Once the resource is obtained, the contents are altered such that
			// it says "aspect <abcdef>" rather than "class <abcdef>"
		
			// First cast the resource to a file
			IFile ifile = ((IFile) resource);
		
			try {
				// Grab a copy of the file into a byte array
				InputStream is = ifile.getContents();
				byte[] inputdata = new byte[is.available()];
				is.read(inputdata);
		
				// Create a string representation of the current file contents
				String currentContents = new String(inputdata);
				String newcontents = "";
				
				// It is difficult to identify what to replace in the contents of the source generated
				// by the underlying NewTypeWizardPage code.  The simple phrase searched for here is
				// "class <TypeName>" - so a type called "Lemon" is being created, this snippet of
				// code looks for "class Lemon" and replaces the word "class" with the word
				// "aspect".
				// This is better than just looking for the first occurrence of "class" - but it could
				// still fail - if, for example, the javadoc header created within the source file
				// contains the "class Lemon" statement!  The phrase search could be made more sophisticated
				// by grabbing the whole line with the "class Lemon" phrase on it and ensuring
				// there is no "//" "/*" or "*" ahead of the "class " phrase found -
				// if any of those 'comment' markers are found, try the next occurrence of class
		
				String phraseToLocate = new String("class "+fPage.getTypeName());
				int i = currentContents.indexOf(phraseToLocate);
				if (i != -1) {
					newcontents =
						new String(
							currentContents.substring(0, i) + "aspect " + currentContents.substring(i + 6)); // 6 = length of "class "
		
				} else {
					// Bugger!!!!
					System.err.println("class keyword not found in (" + currentContents + ")");
				}
		
				// Build an appropriate input stream to pass to setContents()
				ByteArrayInputStream bais = null;
				bais = new ByteArrayInputStream(newcontents.getBytes());
		
				// Alter the contents
				ifile.setContents(bais, false, false, null);
			} catch (Exception e) {
				AspectJPlugin.getDefault().getErrorHandler().handleError(
					AspectJPlugin.getResourceString(
						"NewAspectCreationWizard.exceptionChangingFileContents")
						+ ifile.getFullPath().toOSString(),
					e);
			}
			selectAndReveal(resource);
			openResource((IFile)resource);
		} catch (JavaModelException e) {
			// Could skip reporting this, its not the end of the world.
			AspectJPlugin.getDefault().getErrorHandler().handleError(
			AspectJPlugin.getResourceString("NewAspectCreationWizard.exceptionDuringFinish"),e);
		}
	}

	/**
	 * test whether this is eclipse 2.0.x, if not assume a later version)
	 * @return boolean
	 */
	private boolean isEclipse20x() {
		boolean itIs = false;
		IPluginRegistry registry = Platform.getPluginRegistry();
		IPluginDescriptor desc = 
			registry.getPluginDescriptor( ResourcesPlugin.PI_RESOURCES );
		PluginVersionIdentifier ident = desc.getVersionIdentifier();
		if ( ident.getMajorComponent() == 2  && ident.getMinorComponent() == 0) {
			itIs = true;		
		}
		return itIs;
	}
}