;; aspectj-support --- AspectJ extensions to JDE mode and Java mode

;; Author: Gregor Kiczales and William Griswold
;; Version: 0.7b8
;; Keywords: JDE, tools, AspectJ
;;
;; Copyright (c) Xerox Corporation 1998, 1999, 2000.  
;;
;; This file is part of GNU Emacs.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.
;;
;; Please send bug reports, etc. to support@aspectj.org.

;<--- edit this file with an editor window that is exactly 80 columns wide --->|
;
; aspectj-support.el
;

;;;
;;; Provides a minor mode for aspectj files that builds on either java-mode
;;; or jde-mode.
;;;

;;; ------------------------------
;;; Requirements

;;; The following must be installed for aspectj minor mode to work properly:
;;;
;;;(1) emacs 20.3.1 or higher (ftp://ftp.cs.washington.edu/pub/ntemacs or
;;;    http://www.gnu.org/software/emacs/).  Follow install instructions
;;;    carefully, like running addpm on Windows.  XEmacs is not currently
;;;    supported, although it may work under Unix; please let us know if
;;;    you get it to work.
;;;
;;;(2) aspectj 0.7 (http://aspectj.org).
;;;
;;;(3a)[Java major mode only, Optional] Beanshell (i.e., bsh.jar), available
;;;    at http://www.beanshell.org.  (Note that JDE provides its own copy of
;;;    beanshell.)  If you don't install beanshell, you will still get
;;;    keyword highlighting and indentation.
;;;
;;;(3b)[JDE major mode only] JDE 2.1.8 or higher (http://sunsite.auc.dk/jde/).
;;;    Follow the install instructions from the website.  JDE requires three
;;;    auxiliary packages, elib, bovine (semantic) and speedbar, which are
;;;    are available via the JDE home page.
;;;

;;; ------------------------------
;;; Usage/Installation

;;;
;;; Java mode installation
;;;
;;; NOTE: See also the example .emacs/_emacs init file "_emacs.java-mode"
;;;
;;; (1) The files in this package need to be in the load-path and required.
;;;     For example:
;;;
;;;       (setq load-path (cons "<emacs-pkgs-home>/ajdeForEmacs07b8" load-path))
;;;       (require 'aspectj-support)
;;;
;;; (2) [Optional] Put beanshell (<bsh-home>/bsh.jar) and
;;;     <aspectj-home>/lib/aspectjtools.jar on your java classpath via the
;;;     CLASSPATH environment variable.
;;;
;;; (3) [Helpful at least on Unix] Compile aspectj-support.el.  From emacs:
;;;     M-x byte-compile-file aspectj-support.el.  Compiling can avoid
;;;     stack overruns when interfacing with beanshell in some circumstances.
;;;

;;; JDE mode installation
;;;
;;; NOTE: See also the example .emacs/_emacs init file "_emacs.jde-mode"
;;;
;;; (1) Make sure aspectj mode, JDE, and related packages are on your load-path
;;;     and are required.
;;;
;;;    ;; I keep my emacs packages on D:/apps/emacs-pkgs
;;;    (setq load-path 
;;;          (append
;;;           '( 
;;;             "<emacs-pkgs-home>/ajdeForEmacs07b8"
;;;             "<emacs-pkgs-home>/jde-2.2.4/lisp"
;;;             "<emacs-pkgs-home>/speedbar-0.12"	; for JDE
;;;             "<emacs-pkgs-home>/semantic-1.2.1"	; for JDE/speedbar
;;;             "<emacs-pkgs-home>/elib-1.0"            ; for JDE
;;;            ) 
;;;           load-path))
;;;
;;;    (require 'jde)
;;;    (require 'aspectj-support)
;;;
;;; (2) Put beanshell (<jde-home>/java/lib/bsh.jar) and
;;;     <aspectj-home>/lib/aspectjtools.jar on your java classpath via the
;;;     CLASSPATH environment variable.
;;;
;;; (3) [Helpful at least on Unix] Compile beanshell.el.  From emacs:
;;;     M-x byte-compile-file beanshell.el.  Compiling beanshell can
;;;     avoid stack overruns in certain circumstances.
;;;
;;; (4) Customize JDE + aspectj mode's compile options by putting a version
;;;     of the following in your .emacs/_emacs file or in a JDE project file
;;;     "prj.el" in your project's hierarchy (see the JDE Project File Name
;;;     option for the latter).  Here are two examples:
;;;
;;;       ;;; A default version for simple projects, maybe good for
;;;       ;;; .emacs/_emacs file.
;;;       (custom-set-variables
;;;         '(jde-compiler "ajc")
;;;         ;; ajc requires all files to be named for a compile
;;;         '(jde-compile-file-specification "*.java")
;;;         )
;;;
;;;       ;;; These options are for the spacewar, in examples/spacewar.
;;;       (custom-set-variables
;;;         '(jde-compiler "ajc")
;;;         ;; ajc provides an ``argfile'' mechanism for specifying all files.
;;;         '(jde-compile-file-specification "-argfile demo.lst")
;;;         ;; *if* compiling packages, name root dir for package hierarchy
;;;         ;; to tell ajc where .class files should go.
;;;         '(jde-compile-option-directory "..")
;;;         )
;;;

;;; ------------------------------
;;; general header

(require 'cl)
(require 'easy-mmode)

;;;
;;; This file provides the package named ``aspectj-support''.  When compiled,
;;; if it contains macros, it should require itself to be loaded before the
;;; program is compiled, thus the trailing require expression.
;;;
(provide 'aspectj-support)
(require 'aspectj-support)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; This file consists of several subcomponents:
;;;
;;;    - mode-managing code: hooks, hook actions
;;;    - overlay management for storing points-to/by information
;;;    - jump-menu code for displaying points-to/by
;;;    - conversion of SymbolManager data to speedbar data
;;;    - interface to the SymbolManager API provided by ajtools
;;;    - miscellaneous
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; definition of aspectj-mode.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; mode-setup stuff
;;;
;;; Steps on opening a .java file
;;;
;;; 1. Open the file
;;;
;;; 2. Find the symbols file
;;;    * if error, just fail silently, it might not be aspectj code!
;;;
;;; 3. Check the version number of the table.  If out of date,
;;;    print a message and don't setup links
;;;
;;; 4. setup links.
;;;
;;; 5. Add menu options.
;;;

(easy-mmode-define-minor-mode aspectj-mode
"provides three services on top of Java mode or JDE mode:

1. Navigational support between aspect code and the code that it
   affects, via a jump menu (and in the speedbar for JDE users).

2. Source code annotation of aspect, advice, and introduction
   declarations, as well as the code they affect (default is off for JDE).

3. AspectJ-style compilation.  The Compile... command accessible from the
   Emacs Tools menu is customized through the project customization
   option Aspectj Tools Compile Command, customizable from the AspectJ menu.
   For the extension of JDE, setting the value of Jde Compile File
   Specification in JDE's project options (and changing the default
   compiler to ajc) enables ajc-compatible compiles and builds.

The first two are derived from ajc's last build of the AspectJ program.

There are three commands, two of which customize the mode:

  \\[aspectj-jump-menu]     aspectj-jump-menu\t  Display menu of advisers, advisees, & intro's
  M-x aspectj-setup-annotations\t  Add tagging on the text
  M-x aspectj-clear-annotations\t  Remove tagging from text

JAVA MODE:  As a minor mode of Java mode, the tagging appears in the form of
textual overlays on the Java source (optionally can be turned off), such as
[4 advisees] on an advice, or [3 advisers] on a method that is advised by
three different advice.

JDE MODE: As a minor mode of JDE mode, aspectj-mode enhances the speedbar to
show the location of aspect, advice, and introduction declarations.
The affects/affected-by relationships are shown in the speedbar rather than
embedding tags in the text (available as an option), and selecting the items
in the speedbar will perform the expected navigation. The speedbar symbols have
been extended for AspectJ as follows:

Indication            Meaning
----------            -------
 +  methodSignature   method has an advice that applies to it; double
                      double mouse-1 will display the relevant advice

 +  adviceSignature   advice declared by the containing aspect; double
                      mouse-1 will display affected methods.

 +  introductionSig   introduction declared by the containing class; double
                      mouse-1 will display affected methods or classes.

| | methodOrFieldSig  method or field has been introduced by another class;
                      double mouse-1 on text will navigate to introduction;
                      a + within the bars means that it has an advice
                      that applies to it."
			      nil
			      " AspectJ")

;;;
;;; Note that jde-mode is derived from java-mode.  Means that when jde-mode
;;; starts it runs the java-mode init, including hooks, before its own.
;;;

;;; True if the JDE sources (in particular jde.el) are loaded
(defun aspectj-jde-loaded ()
  (functionp 'jde-find-jde-data-directory))

;;;
;;; Basic switching has been set up so that XEmacs will not execute
;;;  the overlays code (XEmacs supports a similar-but-different mechanism
;;;  called "annotations").  However, forking bsh under ntemacs does not
;;;  work, so this is currently moot.  May work for XEmacs 21 under
;;;  Unix.  Describe-mode info would still need to be updated to reflect
;;;  omission of overlays functionality.
;;;
(defvar aspectj-xemacsp (string-match "XEmacs" emacs-version)
  "Non-nil if we are running in the XEmacs environment.")

;;; Local version of speedbar's variable (see speedbar.el)
(defvar speedbar-dynamic-tags-function-list)
(make-variable-buffer-local 'speedbar-dynamic-tags-function-list)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Customizations using defcustom.
;;;
;;; See mode management, below.
;;;

;; From custom web page for compatibility between versions of custom:
(eval-and-compile
  (condition-case ()
      (require 'custom)
    (error nil))
  (if (and (featurep 'custom) (fboundp 'custom-declare-variable))
      nil ;; We've got what we needed
    ;; We have the old custom-library, hack around it!
    (defmacro defgroup (&rest args)
      nil)
    (defmacro defface (var values doc &rest args)
      (` (progn
	   (defvar (, var) (quote (, var)))
	   ;; To make colors for your faces you need to set your .Xdefaults
	   ;; or set them up ahead of time in your .emacs file.
	   (make-face (, var))
	   )))
    (defmacro defcustom (var value doc &rest args)
      (` (defvar (, var) (, value) (, doc))))))

;;;
;;; Next several declarations provide for customization of AspectJ features.
;;;
(defgroup aspectj nil
  "AspectJ Extensions to JDE/Java Development Environment"
  :group 'tools
  :prefix "aspectj-")

(defcustom aspectj-jde-mode-show-annotations nil
  "*Show source code annotations in JDE mode.
Set this option on if you want annotations in the source code showing
the possible effects of aspects on program behavior.  Change of this
option will not be reflected in open buffers."
  :group 'aspectj
  :type 'boolean)

(defcustom aspectj-java-mode-show-annotations t
  "*Show source code annotations in Java mode.
Set this option off if you do not want annotations in the source code showing
the possible effects of aspects on program behavior.  Change of this
option will not be reflected in open buffers."
  :group 'aspectj
  :type 'boolean)

(defcustom aspectj-tools-compile-command "ajc *.java"
  "*Default compile-command accessed from the Emacs Tools menu."
  :group 'aspectj
  :type 'string)

;;; Default unbound
(defvar aspectj-buffer-show-annotations)
(make-variable-buffer-local 'aspectj-buffer-show-annotations)

(defun aspectj-show-annotations-p ()
  (if (boundp 'aspectj-buffer-show-annotations)
      aspectj-buffer-show-annotations
    (if (eq major-mode 'jde-mode)
	aspectj-jde-mode-show-annotations
      aspectj-java-mode-show-annotations)))

;;; Sets up the customization window.
(defun aspectj-show-customizations ()
  "Show the AspectJ Mode Customizations panel."
  (interactive)
  (customize-apropos "aspectj" 'groups))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mode management.
;;;
;;; This is complicated by the fact that aspectj-mode must work with both
;;; java-mode and jde-mode.  WARNING - there are little bits of code that
;;; have to check which mode we're in so that certain actions are dis/en-abled.
;;; For example, by default text annotations are *disabled* for jde-mode,
;;; since it has the benefits of speedbar.  This is also the sole reason
;;; for having the defcustoms above.  
;;; 
;;; Another complicating factor is XEmacs, which doesn't support overlays
;;; (hence neither annotations nor jump menus).
;;;
;;; A third problem is that turning on/off java-mode or jde-mode should
;;; cause reasonble things to happen for aspectj-mode.  Some of these
;;; things are hard to control, such as deleting menus and reverting
;;; global customizations that have been modified to accomodate aspectj.
;;;

(defun aspectj-mode-setup ()
;  (aspectj-setup-buffer-properties)
  (aspectj-setup-local-compile-command)
  (aspectj-setup-local-fetch-dynamic-menu-function)
  (when (not aspectj-xemacsp)
    (add-hook 'post-command-hook
	      'aspectj-reset-annotations-and-jumps-on-command nil t)
    (aspectj-setup-menu)  ; really one-time per buffer, since can't tear down
    (aspectj-reset-annotations-and-jumps))) ; WGG - not setup; trust me.

;;; Special function, should we ever figure out how to tear down menu
(defun aspectj-mode-clear ()
;  (aspectj-clear-buffer-properties)
  (aspectj-clear-local-compile-command)
  (aspectj-clear-local-fetch-dynamic-menu-function)
  (when (not aspectj-xemacsp)
    (aspectj-clear-annotations-and-jumps)))

(defun aspectj-mode-toggle ()
  ;; minor-mode flag is set before hooks are run
  (if aspectj-mode
      (aspectj-mode-setup)
    (aspectj-mode-clear)))

(add-hook 'java-mode-hook 'aspectj-mode)

;;; Because java-mode is on when aspectj-mode-setup is run
;;; (JDE mode runs java-mode's hooks, first)
(when (not aspectj-xemacsp)
  (add-hook 'jde-mode-hook '(lambda () (if (not (aspectj-show-annotations-p))
					   (aspectj-clear-annotations)))))

;;; Minor mode commands toggle, major mode commands don't
(add-hook 'aspectj-mode-hook 'aspectj-mode-toggle)

;;; WGG - this doesn't work, comments in easy-menu say that a menu
;;; disappears automatically when the keymap is changed. ????
;;; (local-unset-key [menu-bar]) will kill all local menus.
;;; (kill-all-local-variables) works...too well.
;(add-hook 'aspectj-mode-off-hook
;	  '(lambda () (local-unset-key [menu-bar AspectJ])))


(defvar aspectj-saved-compile-command "")
(make-variable-buffer-local 'aspectj-saved-compile-command)

;; Doesn't strictly toggle because it doesn't save aspectj command
(defun aspectj-setup-local-compile-command ()
  (make-local-variable 'compile-command)
;  (setq compile-command (default-value 'compile-command))
  (when (boundp 'compile-command)
    (setq aspectj-saved-compile-command compile-command))
  (setq compile-command aspectj-tools-compile-command))

(defun aspectj-clear-local-compile-command ()
  (setq compile-command aspectj-saved-compile-command))


;;;
;;; Using buffer-local variables, from aspectj-support and speedbar,
;;; toggle which function builds the menu underneath a file.
;;;
;;; NOTE: Perhaps by writing our own version of speedbar-insert-imenu-list,
;;; we can get rid of some more of our changes to speedbar.el.  WGG
;;;
(defun aspectj-setup-local-fetch-dynamic-menu-function ()
  (when (aspectj-jde-loaded)
    (setq speedbar-dynamic-tags-function-list
	  (cons '(aspectj-fetch-dynamic-menu . aspectj-insert-imenu-list)
		(default-value 'speedbar-dynamic-tags-function-list)))))

;;; Only clear if something to restore (protects against ``openness'')
(defun aspectj-clear-local-fetch-dynamic-menu-function ()
  (when (aspectj-jde-loaded)
	(setq speedbar-dynamic-tags-function-list
	      (default-value 'speedbar-dynamic-tags-function-list))))

;;; Oh the humanity.
(defadvice speedbar-trim-words-hierarchy (around aspectj-speedbar activate)
  "Makes sure the advised method doesn't happen in aspectj-mode, because it
screws up the menus."
  (when (not aspectj-mode)
    ad-do-it))


(defun aspectj-setup-menu ()
  (local-set-key [menu-bar AspectJ]
		 (cons "Aspectj" (make-sparse-keymap "Aspectj")))
  (local-set-key [menu-bar AspectJ customize]
		 '("Customize" . aspectj-show-customizations))
  (local-set-key [menu-bar AspectJ show]
		 '("Show annotations" . aspectj-menu-setup-annotations))
  (local-set-key [menu-bar AspectJ clear]
		 '("Clear annotations" . aspectj-menu-clear-annotations))
  (local-set-key [menu-bar AspectJ update]
		 '("Update annotations" . aspectj-reset-annotations-and-jumps))
  (local-set-key [menu-bar AspectJ jump-menu]
		 '("Jump menu" . aspectj-jump-menu)))

(defun aspectj-setup-annotations-and-jumps ()
  "Turn on both jumps and text annotations in the buffer."
  (interactive)
  (aspectj-clear-annotations-and-jumps)
  ;; WGG - for some reason I can't call aspectj-declarations here.
  (aspectj-add-overlays (aspectj-fetch-declarations (buffer-file-name))))

;;;
;;; Only for interactive purposes.  See mode hooks for internal handling.
;;;
(defun aspectj-menu-setup-annotations ()
  "Turn on text annotations in the buffer."
  (interactive)
  (setq aspectj-buffer-show-annotations t)
  (aspectj-setup-annotations-and-jumps))

(defun aspectj-menu-clear-annotations ()
  "Turn on text annotations in the buffer."
  (interactive)
  (setq aspectj-buffer-show-annotations nil)
  (aspectj-clear-annotations))

(defun aspectj-clear-annotations-and-jumps ()
  (aspectj-clear-annotations)
  (aspectj-clear-jumps))

(defun aspectj-reset-annotations-and-jumps ()
  "Recompute the text annotations and jumps of the buffer as appropriate."
  (interactive)
  (aspectj-setup-annotations-and-jumps)
  (when (not (aspectj-show-annotations-p)) (aspectj-clear-annotations))
  t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Overlay cache management.  Borrowed in part from JDE.
;;;

(defun aspectj-reset-annotations-and-jumps-on-command ()
  "Detects when a user activates a buffer and recomputes the declarations
information and resets the overlays if they are stale."
  (when (and aspectj-mode
	     (not aspectj-xemacsp)
	     (not (aspectj-declarations-cache-fresh
		   (buffer-file-name (current-buffer)))))
    (aspectj-reset-annotations-and-jumps)))

(defun aspectj-reset-annotations-and-jumps-on-command-hook (buffer msg)
  (aspectj-reset-annotations-and-jumps-on-command))

(defun aspectj-count-open-java-buffers ()
  "Returns non-nil if any java buffers are open."
  (count 
   ".java"
   (buffer-list)
   :test
   (lambda (file-type buffer)
     (let ((file-name (buffer-file-name buffer)))
       (if file-name
	   (string-match file-type file-name))))))
	 
(defun aspectj-remove-aspectj-hook ()
  "Removes `aspectj-reset-annotations-and-jumps-on-command' when
all Java source buffers have been closed."
  (unless (> (aspectj-count-open-java-buffers) 1)
    (remove-hook 'post-command-hook
		 'aspectj-reset-annotations-and-jumps-on-command)))

(add-hook 'kill-buffer-hook 'aspectj-remove-aspectj-hook)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Overlay handling.
;;;
;;; This code stores points-to and pointed-to-by information in overlays
;;; to permit contextual retrieval (using the current buffer point).
;;;

(defun aspectj-add-overlays (decls)
  "Add overlays to the current buffer.  Do nothing else"
  (when decls
    (while decls
      (let ((decl          (car decls)))
	(unless (aspectj-decl-is-introduced-p decl)
	  (let* ((startline     (aspectj-decl-begin-line decl))
		 (endline       (aspectj-decl-end-line decl))
		 (kind          (aspectj-decl-kind decl))
		 (points-to     (aspectj-decl-points-to decl))
		 (pointed-to-by (aspectj-decl-pointed-to-by decl))
		 
		 (startpoint (save-excursion 
			       (goto-line startline)
			      ;(beginning-of-line)
			       (point)))
		 (endpoint   (save-excursion
			       (goto-line endline)
			       (end-of-line)
			       (point))))
	    (aspectj-add-points-to-overlay     startpoint endpoint kind points-to)
	    (aspectj-add-pointed-to-by-overlay startpoint endpoint kind pointed-to-by)
	    (if (aspectj-decl-is-type-p decl)
		(aspectj-add-overlays (aspectj-decl-declarations decl)))))
	(setq decls (cdr decls))))))

(defun aspectj-add-points-to-overlay (start-char end-char kind points-to)
  (setq points-to (remove-if #'aspectj-decl-is-introduced-p points-to))
  (when points-to 
    (let* ((n (length points-to))
	   (button-text 
	     (format (cond;((eq kind 'pointcut)     "[%s targets%s]")
		           ((eq kind 'introduction) "[%s introduction%s]")
			   ((eq kind 'advice)       "[%s advisee%s]")
			   (t (error "Unrecognized kind of declaration: %S" kind)))
		     n (if (= n 1) "" "s"))))
      (aspectj-add-overlay start-char end-char points-to button-text))))

(defun aspectj-add-pointed-to-by-overlay (start-char end-char kind pointed-to-by)
  (setq pointed-to-by (remove-if #'aspectj-decl-is-introduced-p pointed-to-by))
  (when pointed-to-by
    (let* ((n (length pointed-to-by))
	   (button-text
	     (format (cond ((eq kind 'interface)   "[%s introducer%s]")
			   ((eq kind 'class)       "[%s introducer%s]")
			   ((eq kind 'field)       "[%s adviser%s]")
			   ((eq kind 'constructor) "[%s adviser%s]")
			   ((eq kind 'method)      "[%s adviser%s]")
			   (t (error "Unrecognized kind of declaration: %S" kind)))
		     n (if (= n 1) "" "s"))))
      (aspectj-add-overlay start-char end-char pointed-to-by button-text))))

(defun aspectj-add-overlay (start-char end-char jumps button-text)
  (let* ((eol (save-excursion (goto-char start-char) (end-of-line) (point)))
	 (n (length points-to))
	 (region-overlay (make-overlay start-char end-char))
	 (eol-overlay    (make-overlay start-char eol))
	 (nfill (max 1 (- (window-width) eol (length button-text))))
	 (overlay-text (concat (make-string nfill ?\ ) button-text)))
    (overlay-put eol-overlay 'after-string overlay-text)
    (overlay-put eol-overlay 'aspectj-annotation-p t)
    (overlay-put region-overlay 'aspectj-jumps jumps)))

(defun aspectj-clear-annotations ()
  "Clear the annotations from buffer."
  (interactive)
  (let ((overlays (aspectj-get-overlays t)))
    (aspectj-clear-overlays 
      (remove-if-not #'(lambda (o) (aspectj-annotation-p o)) overlays))))

(defun aspectj-clear-jumps ()
  (let ((overlays (aspectj-get-overlays t)))
    (aspectj-clear-overlays 
      (remove-if-not #'(lambda (o) (aspectj-jumps o))        overlays))))

(defun aspectj-clear-overlays (overlays)
  (while overlays
    (delete-overlay (car overlays))
    (setq overlays (cdr overlays)))
  (recenter))  ; WGG - was call to gud-refresh, which calls recenter
  
(defun aspectj-get-jumps-at ()
  (apply 'append
	 (remove nil
		 (mapcar 'aspectj-jumps (aspectj-get-overlays (point))))))

(defun aspectj-get-all-jumps ()
  (apply 'append
	 (remove nil
		 (mapcar 'aspectj-jumps (aspectj-get-overlays t)))))

(defun aspectj-get-annotations-at ()
  (remove-if-not 'aspectj-annotation-p  (aspectj-get-overlays (point))))

(defun aspectj-get-all-annotations ()
  (remove-if-not 'aspectj-annotation-p  (aspectj-get-overlays t)))

(defun aspectj-get-overlays (point-or-t)
  (if (eq point-or-t 't)
      (save-excursion (overlays-in (progn (beginning-of-buffer) (point))
				   (progn (end-of-buffer) (point))))
      (overlays-at point-or-t)))


(defun aspectj-annotation-p (overlay)
  (overlay-get overlay 'aspectj-annotation-p))

(defun aspectj-jumps (overlay)
  (overlay-get overlay 'aspectj-jumps))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Menu helper(s).
;;;

(defun aspectj-substring (string maxlen)
  (if (<= (length string) maxlen)
      string
    (concat (substring string 0 (- maxlen 2)) "..")))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Jump Menu functions.
;;;
;;; The Jump Menu displays the points-tos and pointed-to-bys relevant to the
;;; program point current when the Jump Menu is activated.  Activating items
;;; in the list "jumps" you to the selected item to see the nature of the
;;; points-to/by.
;;;

(when (not aspectj-xemacsp)
  (define-key ctl-x-map "\C-j" 'aspectj-jump-menu))

;;; Precondition: 
(defun aspectj-jump-menu ()
  "Make a menu of jumps to/from selected code."
  (interactive)
  (let ((buffer (aspectj-list-jumps-noselect)))
    (switch-to-buffer-other-window buffer)
    (let* ((start (progn (beginning-of-buffer) (point)))
	   (end   (progn (end-of-buffer) (point)))
	   (nlines (count-lines start end)))
      (enlarge-window (+ 2 (- nlines (window-height))))
      (goto-line 3)  ; moves point to first entry after header
      (recenter)     ; moves cursor to point and recenters
      (message "Commands: q to quit; n or p to move; SPC or x to jump; ? for help."))))

(defun aspectj-list-jumps-noselect ()
  "Create and return a buffer with a list ..."
  (let ((old-buffer (current-buffer))
	(standard-output standard-output)
	(jumps (aspectj-get-jumps-at)))
    (save-excursion
      (set-buffer (get-buffer-create "*AspectJ Jumps*"))
      (setq buffer-read-only nil)
      (erase-buffer)
      (setq standard-output (current-buffer))
      (princ "\
 Signature                                        Filename
 ---------                                        --------
")
      (while jumps
	(let* ((decl       (car jumps))
	       (filename   (aspectj-decl-filename decl))
	       (kind       (aspectj-decl-kind decl))
	       (signature  (aspectj-decl-signature decl))
	       (midline    50) ; leave 30 characters for relative filename
	       (compile-dir                 ; keep jump-menu entries short
		(if (and (boundp 'jde-compile-option-directory)
			 (not (string= jde-compile-option-directory "")))
		    jde-compile-option-directory
		  "..")) ; ".." means "parent directory"
	       line-start
	       line-end)
	  (indent-to 1)
	  (setq line-start (point))
	  (princ (aspectj-substring signature (- midline 4)))
	  (indent-to midline 2)
	  (princ (file-relative-name filename compile-dir))
	  (setq line-end (point))
	  (put-text-property line-start line-end 'jump decl)
	  (put-text-property line-start line-end 'mouse-face 'highlight)
	  (princ "\n")
	  (setq jumps (cdr jumps))))	
      (aspectj-jump-menu-mode)
      (current-buffer))))

(defvar aspectj-jump-menu-mode-map nil "")

(when (not aspectj-jump-menu-mode-map)
  (setq aspectj-jump-menu-mode-map (make-keymap))
  (suppress-keymap aspectj-jump-menu-mode-map t)
  (define-key aspectj-jump-menu-mode-map "q" 'aspectj-jump-menu-quit)
  (define-key aspectj-jump-menu-mode-map "x" 'aspectj-jump-menu-select)
  (define-key aspectj-jump-menu-mode-map " " 'aspectj-jump-menu-select)
  (define-key aspectj-jump-menu-mode-map "n" 'next-line)
  (define-key aspectj-jump-menu-mode-map "p" 'previous-line)
  (define-key aspectj-jump-menu-mode-map "?" 'describe-mode)
  (define-key aspectj-jump-menu-mode-map [mouse-2] 'aspectj-jump-menu-mouse-select)
)

;; AspectJ Jump mode is suitable only for specially formatted data.
(put 'aspectj-jump-menu-mode 'mode-class 'special)

(defun aspectj-jump-menu-mode ()
  "Major mode for viewing advice, introductions, or their targets that
   apply to selected code in source window.  All keystrokes are commands.
\\{aspectj-jump-menu-mode-map}"

  (kill-all-local-variables)
  (use-local-map aspectj-jump-menu-mode-map)
  (setq major-mode 'aspectj-jump-menu-mode)
  (setq mode-name "AspectJ Jump Menu")
  (setq truncate-lines t)
  (setq buffer-read-only t)
  (run-hooks 'aspectj-jump-menu-mode-hook))

(defun aspectj-jump-menu-quit ()
  "Quit the jump menu."
  (interactive)
  (switch-to-buffer (window-buffer (get-lru-window)))
  (bury-buffer (other-buffer))
  (delete-other-windows))

(defun aspectj-jump-menu-select ()
  "Visit entry under the cursor in the buffer menu."
  (interactive)
  (aspectj-jump-menu-select-internal (current-buffer)
				     (point)))

(defun aspectj-jump-menu-mouse-select (event)
  "Visit entry under mouse cursor in the buffer menu."
  (interactive "e")  
  (let ((window (posn-window (event-end event))))
    (select-window window)
    (aspectj-jump-menu-select-internal (window-buffer window)
				       (posn-point (event-end event)))))

(defun aspectj-jump-menu-select-internal (buffer char)
  (let* ((decl       (aspectj-jump-menu-lookup-jump buffer char))
	 (filename   (aspectj-decl-filename decl))
	 (start-line (aspectj-decl-begin-line decl))
	 (back-to    (get-lru-window)))
    (find-file filename)
    (goto-line start-line)
    (recenter 0)
    (kill-buffer buffer)
    (select-window back-to)
    (balance-windows)))

(defun aspectj-jump-menu-lookup-jump (buffer char)
  "Return jump (an internal structure) described by this line of buffer menu."
  (save-excursion 
    (switch-to-buffer buffer)
    (goto-char char)
    (beginning-of-line)
    (and (not (eobp)) (get-text-property (+ (point) 2) 'jump))))

;;;
;;; Start of testing script.  This just prints out what came from SymbolManager
;;; and what goes out to speedbar.
;;;
(defun aspectj-test ()
;  (setq debug-on-error t)
   (let* ((filename (buffer-file-name)))
     (aspectj-test-fetch-decls filename)))

(defun aspectj-test-fetch-decls (filename)
   (let* ((decls (aspectj-fetch-declarations filename))
	  (out-buffer (get-buffer-create "*scratch*")))
     (pp decls out-buffer)
     (pp (aspectj-fetch-dynamic-menu filename) out-buffer)
     nil))

;     (list 'raw-decls: decls
;	   'speedbar-menu: (aspectj-fetch-dynamic-menu filename))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Creates speedbar data from lisp-ified SymbolManager data.
;;;
;;; we need to produce what speedbar calls a "generic list"
;;; (... nil
;;;      ("foo" . 32)
;;;      ("foo" 32 <button> . ...)
;;;      ("foo" . ...)
;;;
;;; <-> spacewar/
;;;  [-] Ship.java
;;;   (-) Ship
;;;    (+) MyInnerClass
;;;        getX()
;;;     +  getY()
;;;     -  getZ()
;;;         Debug: static before
;;;
;;;    |+| paint(..)
;;;
;;; 
;;;    |-| paint(..)
;;;         introduction FooCut
;;;
;;;
;;;     +  introduction Game | ..
;;;         Game 
;;;         Player 
;;;         Bozo
;;;
;;;     + static before ...
;;;
;;;     - static before ...
;;;         Frotz: doit()
;;;
;;;   (+) GameCoordinator
;;;
;;; The "brackets" say what kind of thing is on this line, NOT what kind of 
;;; thing is underneath it.  The presence of +/- means you can expand the
;;; thing to get what's under it.
;;;

;;;
;;; WGG WARNING!  Horrific blob of essentially copied code below, as dictated
;;; by "extensible" speedbar interface.  Sigh....
;;;

;(setq debug-on-error t)
;(mapcar '(lambda (fn) (trace-function-background fn "*scratch*"))
;	'(speedbar-insert-imenu-list 
;          aspectj-insert-imenu-list
;	  speedbar-insert-generic-list
;	  aspectj-insert-generic-list
;	  speedbar-make-tag-line
; 	  aspectj-make-tag-line
; 	  speedbar-create-tag-hierarchy
; 	  speedbar-tag-expand
; 	  aspectj-tag-expand
; 	  speedbar-tag-find
; 	  aspectj-tag-find
; 	  speedbar-tag-face))

(defun aspectj-insert-imenu-list (indent lst)
  "At level INDENT, insert the imenu generated LST."
  (aspectj-insert-generic-list indent lst
				'aspectj-tag-expand
				'aspectj-tag-find))

;;; Elements of the list are one of the following, where ... is a 
;;; recursive list of the same kind.
;;;   nil
;;;   ("foo" . <int-or-mark>)
;;;   ("foo" <int-or-mark> button-type ...)  !!!  this one is new
;;;   ("foo" . ...)
;;; 

;;;
;;; Our modified version of speedbar-insert-generic-list.  (Mods marked.)
;;;
(defun aspectj-insert-generic-list (level lst expand-fun find-fun)
  "At LEVEL, insert a generic multi-level alist LST.
Associations with lists get {+} tags (to expand into more nodes) and
those with positions just get a > as the indicator.  {+} buttons will
have the function EXPAND-FUN and the token is the CDR list.  The token
name will have the function FIND-FUN and not token."
  ;; Remove imenu rescan button
  (if (string= (car (car lst)) "*Rescan*")
      (setq lst (cdr lst)))
  ;; Adjust the list.
  (setq lst (speedbar-create-tag-hierarchy lst))
  ;; insert the parts
  (while lst
    (cond ((null (car-safe lst)) nil)	;this would be a separator
	  ((or (numberp (cdr-safe (car-safe lst)))
	       (markerp (cdr-safe (car-safe lst))))
	   (aspectj-make-tag-line nil nil nil nil ;no expand button data
				   (car (car lst)) ;button name
				   find-fun        ;function
				   (cdr (car lst)) ;token is position
				   'speedbar-tag-face
				   (1+ level)))
	  ;;!!!vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
 	  ((and (not (null (cdr-safe (cdr-safe (car-safe lst)))))
 		(or (numberp (car-safe (cdr-safe (car-safe lst))))
 		    (markerp (car-safe (cdr-safe (car-safe lst))))))
	   (let ((button-name (car   (car lst)))
		 (mark        (cadr  (car lst)))
		 (button-type (caddr (car lst)))
		 (subs        (cdddr (car lst))))
	     (aspectj-make-tag-line button-type
				     (if subs ?+ ? )
				     (and subs expand-fun)
				     subs
				     button-name
				     find-fun
				     mark
				     'speedbar-tag-face
				     (1+ level))))
	  ;;!!!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	  ((listp (cdr-safe (car-safe lst)))
	   (aspectj-make-tag-line 'curly ?+ expand-fun (cdr (car lst))
				   (car (car lst)) ;button name
				   nil nil 'speedbar-tag-face
				   (1+ level)))
	  (t (speedbar-message "Ooops!")))
    (setq lst (cdr lst))))

;;;
;;; Only here to call other aspectj-... functions.
;;;
(defun aspectj-tag-expand (text token indent)
  "Expand a tag sublist.  Imenu will return sub-lists of specialized tag types.
Etags does not support this feature.  TEXT will be the button
string.  TOKEN will be the list, and INDENT is the current indentation
level."
  (cond ((string-match "+" text)	;we have to expand this file
	 (speedbar-change-expand-button-char ?-)
	 (speedbar-with-writable
	   (save-excursion
	     (end-of-line) (forward-char 1)
	     (aspectj-insert-generic-list indent token 'aspectj-tag-expand
					   'aspectj-tag-find))))
	((string-match "-" text)	;we have to contract this node
	 (speedbar-change-expand-button-char ?+)
	 (speedbar-delete-subblock indent))
	(t (error "Ooops...  not sure what to do")))
  (speedbar-center-buffer-smartly))

(defun aspectj-make-tag-line  (exp-button-type
			       exp-button-char exp-button-function
			       exp-button-data
			       tag-button tag-button-function tag-button-data
			       tag-button-face depth)
  "Create a tag line with EXP-BUTTON-TYPE for the small expansion button.
This is the button that expands or contracts a node (if applicable),
and EXP-BUTTON-CHAR the character in it (+, -, ?, etc).  EXP-BUTTON-FUNCTION
is the function to call if it's clicked on.  Button types are
'bracket, 'angle, 'curly, or nil.  EXP-BUTTON-DATA is extra data
attached to the text forming the expansion button.

Next, TAG-BUTTON is the text of the tag.  TAG-BUTTON-FUNCTION is the
function to call if clicked on, and TAG-BUTTON-DATA is the data to
attach to the text field (such a tag positioning, etc).
TAG-BUTTON-FACE is a face used for this type of tag.

Lastly, DEPTH shows the depth of expansion.

This function assumes that the cursor is in the speedbar window at the
position to insert a new item, and that the new item will end with a CR"
  (let ((start (point))
	(end (progn
	       (insert (int-to-string depth) ":")
	       (point)))
	(depthspacesize (* depth speedbar-indentation-width)))
    (put-text-property start end 'invisible t)
    (insert-char ?  depthspacesize nil)
    (put-text-property (- (point) depthspacesize) (point) 'invisible nil)
    (let* ((exp-button (cond ((eq exp-button-type 'bracket) "[%c]")
			     ((eq exp-button-type 'angle) "<%c>")
			     ((eq exp-button-type 'curly) "{%c}")
			     ((eq exp-button-type 'paren)        "(%c)") ;!!!
			     ((eq exp-button-type 'blank)        " %c ") ;!!!
			     ((eq exp-button-type 'vertical-bar) "|%c|") ;!!!
			     (t ">")))
	   (buttxt (format exp-button exp-button-char))
	   (start (point))
	   (end (progn (insert buttxt) (point)))
	   (bf (if exp-button-type 'speedbar-button-face nil))
	   (mf (if exp-button-function 'speedbar-highlight-face nil))
	   )
      (speedbar-make-button start end bf mf exp-button-function exp-button-data)
      (if speedbar-hide-button-brackets-flag
	  (progn
	    (put-text-property start (1+ start) 'invisible t)
	    (put-text-property end (1- end) 'invisible t)))
      )
    (insert-char ?  1 nil)
    (put-text-property (1- (point)) (point) 'invisible nil)
    (let ((start (point))
	  (end (progn (insert tag-button) (point))))
      (insert-char ?\n 1 nil)
      (put-text-property (1- (point)) (point) 'invisible nil)
      (speedbar-make-button start end tag-button-face
			    (if tag-button-function 'speedbar-highlight-face nil)
			    tag-button-function tag-button-data))
    ))

(defun aspectj-tag-find (text token indent)
  "For the tag TEXT in a file TOKEN, goto that position.
INDENT is the current indentation level."
  (if (consp text)                                           ;!!!
      (aspectj-tag-find (cdr text) (car text) indent)        ;!!! WGG - speedbar-tag-find OK?
  (let ((file (if (markerp token)                            ;!!!
		  (buffer-file-name (marker-buffer token))   ;!!!
		(speedbar-line-path indent))))
    (speedbar-find-file-in-frame file)
    (save-excursion (speedbar-stealthy-updates))
    ;; Reset the timer with a new timeout when cliking a file
    ;; in case the user was navigating directories, we can cancel
    ;; that other timer.
    (speedbar-set-timer speedbar-update-speed)
    (goto-char token)
    (run-hooks 'speedbar-visiting-tag-hook)
    (speedbar-maybee-jump-to-attached-frame)
    ))
  )

;;;
;;; The variable aspectj-buffers-declarations holds the most recently
;;; computed declarations, which could have been recomputed by the
;;; jump-menu and annotations stuff.  So these variables aid in caching
;;; the dynamic menu and determining its freshness.
;;;
(defvar aspectj-buffers-last-dynamic-menu nil)
(defvar aspectj-buffers-last-dynamic-menu-declarations nil)
  
;;;
;;; Returns an approximation of an "index-alist", a data structure defined
;;; by imenu.el.  It is not quite one of these beasts, though, due to
;;; extensions to support aj weirdness.
;;;
;;; Returns "t" on failure because that's what speedbar-fetch-dynamic-tags
;;; wants to see (see speedbar.el).
;;;
(defun aspectj-fetch-dynamic-menu (fn)
  (let ((declarations (aspectj-declarations fn)))
    (if declarations
	(if (eq declarations aspectj-buffers-last-dynamic-menu-declarations)
	    aspectj-buffers-last-dynamic-menu
	  (setq aspectj-buffers-last-dynamic-menu-declarations declarations)
	  (setq aspectj-buffers-last-dynamic-menu
		(aspectj-convert-decls 0 declarations t nil)))
      t)))

(defun aspectj-convert-decls (depth decls recurse parent-decl)
  (mapcar #'(lambda (decl)
	      (aspectj-convert-decl (+ depth 1) decl recurse parent-decl)) 
	  (aspectj-sort-decls decls)))

;;; Helper for below to keep speedbar entries from being too long.
;;; Threshold length of 5 avoids over-elision.
(defun aspectj-shorten (depth string fraction)
  (let ((maxlen
	 (max 5 (- (floor (* fraction (speedbar-frame-width))) depth 1))))
    (aspectj-substring string maxlen)))


(defun aspectj-convert-decl (depth decl recurse parent-decl)
  (let* ((fn             (aspectj-decl-filename       decl))
	 (begin-line     (aspectj-decl-begin-line     decl))
	 (kind           (aspectj-decl-kind           decl))
	 (sig            (aspectj-decl-signature      decl))
	 (declaring-type (aspectj-decl-declaring-type decl))
	 (pt             (aspectj-decl-points-to      decl))
	 (ptb            (aspectj-decl-pointed-to-by  decl))
	 (decls          (aspectj-decl-declarations   decl))

	 (button-type    (aspectj-decl-button-type decl))
	 (mark           (if (and (aspectj-unmapped fn begin-line) parent-decl)
			     (aspectj-point-or-mark  ; then mark from parent
			      (aspectj-decl-filename parent-decl)
			      (aspectj-decl-begin-line parent-decl))
			   (aspectj-point-or-mark fn begin-line)))
	 
	 (qualified-sig 
	  (if recurse
	      sig
	    (concat (aspectj-shorten depth declaring-type .5) ; so can see sig
		    ": "
		    sig))))

    ;;; WGG - want introduction's to show their decls and to whom they
    ;;;  introduce, but only at the actual declaration of introduction, not
    ;;;  not at each site it affects (hence recurse flag test).
    ;;; WGG - the recurse decision is subject to revision; programmer may
    ;;;  want to know what the decl of introduced object is; however, JDE
    ;;;  does not show this information for normal field/method decls.
    (if (and (eq kind 'introduction) recurse)
	(list qualified-sig mark button-type	      
	      (list* "Targets"
		     (aspectj-convert-introduction-targets pt))
	      (list* "Decls"
		     (append (aspectj-convert-decls (1+ depth) decls recurse decl))))
      (list* qualified-sig
	     mark
	     button-type
	     (and (and recurse (or pt ptb decls))
		  ;; WGG - sort only pt/ptb, as they have no predefined order
		  ;;  that programmer may expect items to be in, but decls do.
		  (append
		   (aspectj-converted-decls-sort
		    (aspectj-convert-decls (1+ depth) pt    nil nil))
		   (aspectj-converted-decls-sort
		    (aspectj-convert-decls (1+ depth) ptb   nil nil))
		   (aspectj-convert-decls (1+ depth) decls t decl)))))))

(defun aspectj-convert-introduction-targets (pts)
    (mapcar #'(lambda (pt)
		(let* ((name (aspectj-decl-signature pt))
		       (file (aspectj-decl-filename pt))
		       (line (aspectj-decl-begin-line pt)))
		  (list name (aspectj-point-or-mark file line) 'blank)))
	    pts))


;;;
;;; Lists of advice points-tos can be quite long, especially in debugging,
;;;  so sorting helps to get things under control.  WGG
;;; 
(defun aspectj-converted-decls-sort (cdecls)
 (sort cdecls #'(lambda (cdecl1 cdecl2) (string< (car cdecl1) (car cdecl2)))))



;; WGG - above used to filter for blanks *and* remove duplicates
;	    (remove-duplicates type-names :test 'equal))))
	

(defvar *aspectj-decls-order* '(interface 
				class
				field        ;introduced-field
				initializer  ;introduced-initializer
				constructor  ;introduced-constructor
				method       ;introduced-method
				pointcut
				introduction 
				advice))

(defun aspectj-sort-decls (decls)    
  (sort decls 
	#'(lambda (a b)
	    (member (aspectj-decl-kind b)
		    (cdr (member (aspectj-decl-kind a) 
				 *aspectj-decls-order*))))))

(defun aspectj-decl-button-type (decl)
  (let ((kind (aspectj-decl-kind decl)))
    (ecase kind 
      ((aspect interface class) 'paren)
      ((field initializer constructor method)
       (if (aspectj-decl-is-introduced-p decl) 'vertical-bar 'blank))
      ((pointcut introduction advice) 'blank))))
  

(defun aspectj-point-or-mark (fn line-number)
  (save-excursion
    (let ((other-buffer (find-file-noselect fn)))
      (unless (eq other-buffer (current-buffer)) (set-buffer other-buffer))
      (goto-line line-number)
      (if  (eq other-buffer (current-buffer))
	  (point-marker)
	(point)))))

;;;
;;; Convention is that if line-number is less than 1, it's bogus.
;;; See SymbolManager for details.
;;;
(defun aspectj-unmapped (fn line-number)  (< line-number 1))

(defun line->bol (line)
  (save-excursion (goto-line line)  (beginning-of-line) (point)))

(defun point->bol (point)  
  (save-excursion (goto-char point) (beginning-of-line) (point)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Here's where we bridge the gap between Lisp and the SymbolManager API
;;; the compiler provides.  We've got helper code, on the Java side, called
;;; StringBasedSymbolManager.
;;;
;;; The aspectj-fetch-... functions go get data from the compiler API.  The
;;; aspectj-decl-... functions are accessors over that data.  
;;;
;;; With a lot more work this could be made lazy.  Don't do it!
;;;
(defvar aspectj-classname "org.aspectj.tools.ide.StringBasedSymbolManager")

(defun aspectj-fetch-data-internal (stmt)
  (let ((data
	  (let ((max-lisp-eval-depth (* max-lisp-eval-depth 10)))
	    (read (aspectj-bsh-eval stmt)))))
    (if (not (listp data))
	(error (prin1-to-string data))
      data)))

;;; Assumes cadr is string with a colon in it.  A bit of a hack because it
;;;  presumes to know something of structure of string returning from
;;;  StringBasedSymbolManager.  (Can't look just for colon because colon is
;;;  also used in Windows file specifiers.
(defun aspectj-get-message (java-message)
  (let* ((long-message (cadr java-message))
	 (colon-posn (string-match "^.*Error:" long-message)))
    (if (null colon-posn)
	long-message
      (substring long-message (match-end 0)))))

(defun aspectj-process-result (result filename)
  (if (eq (car result) 'ERROR)	; WGG - call to function error broke speedbar
      (progn
;	(ding t)	; bell to alert user since message is gone quickly
	(message "AspectJ Mode Warning:%s (compile with ajc)"
		 (aspectj-get-message result) filename)
	nil)
    result))

;;; The time returned by file attributes are two 16-bit ints in hi-order
;;; low-order bit order.  So I'm shifting the high bits up here.
;;;
;;; The next several functions implement caching of the 
(defun aspectj-file-timestamp (fn) (nth 5 (file-attributes fn)))
(defconst aspectj-notexist-timestamp -1)

(defun aspectj-declaration-file (fn)
  (concat (file-name-sans-extension fn) ".ajsym"))

;;;
;;; Note that the cache is "fresh" if there was no file last time and no
;;; file this time.
;;;
(defun aspectj-declarations-cache-fresh (fn)
  (let ((declarations-file (aspectj-declaration-file fn)))
    (equal (if (file-exists-p declarations-file)
		    (aspectj-file-timestamp declarations-file)
		  aspectj-notexist-timestamp)
		aspectj-declarations-timestamp)))

(defvar aspectj-buffers-declarations nil)
(make-variable-buffer-local 'aspectj-buffers-declarations)

(defvar aspectj-declarations-timestamp aspectj-notexist-timestamp)
(make-variable-buffer-local 'aspectj-declarations-create-time)

;;; Cached version of aspectj-fetch-declarations
(defun aspectj-declarations (filename) 
  (if (aspectj-declarations-cache-fresh filename)
      aspectj-buffers-declarations
    (aspectj-fetch-declarations filename)))

(defun aspectj-fetch-declarations (filename) 
  (let* ((method-name "fetchDeclarations")
	 (stmt (concat aspectj-classname "." method-name
		       "(\"" filename "\");"))
	 (result (aspectj-fetch-data-internal stmt)))
    (aspectj-process-result result filename)))

(defadvice aspectj-fetch-declarations (around aspectj-decls activate)
  "Cache the declarations data."
  (let ((timestamp ; do first to get stale-est timestamp
	 (aspectj-file-timestamp (aspectj-declaration-file filename))))
    ad-do-it
    (setq aspectj-declarations-timestamp
	  (if ad-return-value timestamp aspectj-notexist-timestamp))
    (setq aspectj-buffers-declarations ad-return-value)))


(defun aspectj-fetch-source-line (filename line)
  (let* ((method-name "fetchSourceLine")
	 (stmt (concat aspectj-classname "." method-name
		       "(\"" filename "\", " line");"))
	 (result (aspectj-fetch-data-internal stmt)))
    (aspectj-process-result result filename)))


;; Helper for below.  Copied from dired to avoid pkg include. :(
(defun aspectj-string-replace-match (regexp string newtext
					  &optional literal global)
  "Replace first match of REGEXP in STRING with NEWTEXT.
If it does not match, nil is returned instead of the new string.
Optional arg LITERAL means to take NEWTEXT literally.
Optional arg GLOBAL means to replace all matches."
  (if global
      (let ((start 0))
	(while (string-match regexp string start)
	  (let ((from-end (- (length string) (match-end 0))))
	    (setq string (replace-match newtext t literal string))
	    (setq start (- (length string) from-end))))
	  string)
    (if (not (string-match regexp string 0))
	nil
      (replace-match newtext t literal string))))

;;;
;;; This order lines up with the code in StringBasedSymbolManager.  We use
;;; order rather than a plist to make loading the data faster.
;;;
(defun aspectj-decl-begin-line         (decl) (car (nth 0 decl)))
(defun aspectj-decl-begin-column       (decl) (cdr (nth 0 decl)))
(defun aspectj-decl-end-line           (decl) (car (nth 1 decl)))
(defun aspectj-decl-end-column         (decl) (cdr (nth 1 decl)))
(defun aspectj-decl-kind               (decl) (nth 2 decl))
(defun aspectj-decl-signature          (decl)
  (aspectj-string-replace-match "  " (nth 3 decl) " " t t)) ;; no extra blanks
(defun aspectj-decl-filename           (decl) (nth 4 decl))
(defun aspectj-decl-declaring-type     (decl) (nth 5 decl))
(defun aspectj-decl-points-to          (decl) (nth 6 decl)) ;; getTargets !!!
(defun aspectj-decl-pointed-to-by      (decl) (nth 7 decl)) ;; getPointedToBy
(defun aspectj-decl-declarations       (decl) (nth 8 decl))
(defun aspectj-decl-is-type-p          (decl) (nth 9 decl))
(defun aspectj-decl-is-introduced-p    (decl) (nth 10 decl))
(defun aspectj-decl-has-body-p         (decl) (nth 11 decl))
(defun aspectj-decl-has-signature-p    (decl) (nth 12 decl))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Highlighting behavior for AspectJ.
;;;
(require 'font-lock)
(require 'cc-mode)

;;; There are a bunch of new keywords in AspectJ that should be highlighted
;;; properly and might impact indentation.  We treat pointcut and aspect as
;;; method/type decls.  before/after/around are special keywords that appear
;;; in the method-name position as far as Java is concerned, so we just color.
;;;
;;; The major challenge here is that we're hanging stuff off of Java mode.
;;; There are plenty of hooks, but they are not exactly designed for
;;; language extension.  It also makes them hard to turn off while leaving
;;; Java mode on.  Basically buffer-local variables need to get macked
;;; for indentation, and maybe globals for highlighting.  The global stuff
;;; is deferred.
;;;
;;; All this would be so much easier if we could hook off minor modes.
;;; Perhaps we should consider deriving AspectJ mode from Java/JDE mode.
;;; The problem here is that we have to pick one, unlike with the minor mode.
;;;
;;; We are doing essentially nothing for indentation.
;;; There is a mechanism for controlling the indentation around ":", but it
;;; assumes a finite number of C uses (case, goto, etc.), so it would be
;;; a major hack to get the ":" after pointcut to behave.  So we'll do
;;; without (see progmode/cc-vars.el, part of cc-mode, in particular the
;;; c-hanging-colons-alist).  The real question in getting this to work is
;;; how would java mode misinterpret the pointcut declaration, so what
;;; syntactic class would it fall into?  Of course, if we tried to do this,
;;; it would screw up the colon formatting of the intended C entity as well.
;;; A quick check showed that java-mode does not have any settings on this
;;; var, so maybe we should be ignoring, too.
;;;

;;;
;;; Highlighting.  (Global and invasive to major modes).  All the magic
;;; for this for Java itself is in font-lock.el, not cc-mode.  That's
;;; where a lot of these patterns came from.
;;;
;;; We assign a color to introduction so it looks like class and pointcut,
;;; but doesn't highlight the following class spec (since using, not
;;; defining, in a way).  It's also hard to highlight more than a word.
;;;
;;; Many of the choices/omissions here are highly subjective (e.g., what
;;; to do about receptions, within, eachobject etc.  I haven't fully worked
;;; out the "levels" (1,2,3) of highlighting, either.
;;;
(defvar aspectj-keyword-spec
  (list	'("\\<\\(introduction\\)\\>" . font-lock-type-face)
        '("\\<\\(aspect\\|pointcut\\)\\>[ \t]*\\(\\sw+\\)?"
           (1 font-lock-type-face) (2 font-lock-function-name-face nil t))
	'("\\<\\(of\\)\\>" . font-lock-keyword-face)
	'("\\<\\(before\\|after\\|around\\)\\>" . font-lock-keyword-face)
	'("\\<\\(returning\\|throwing\\)\\>" . font-lock-keyword-face)
        '("\\<\\(returns\\)\\>[ \t]*\\(\\sw+\\)?"
           (1 font-lock-keyword-face) (2 font-lock-type-face t t))
	'("\\<\\(hasaspect\\)\\>" . font-lock-keyword-face)
	'("\\<\\(thisJoinPoint\\)\\>" . font-lock-keyword-face)
))

;;;
;;; Note we need to do this for both modes (actually, just the actual major
;;; mode being fired up, but that's hard to tell here and could even change).
;;;
(font-lock-add-keywords 'java-mode aspectj-keyword-spec)
(font-lock-add-keywords 'jde-mode  aspectj-keyword-spec)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Overrides JDE compile/build behaviors when JDE is present.
;;;

(when (aspectj-jde-loaded)
;;; WGG - if defined, overrides JDE's mechanism to find .java to compile.
;;;  Might also be put in group jde-compile-options.
(defcustom jde-compile-file-specification ""
  "*Specification of file(s) to compile as alternate to current buffer's or application class's file. 
This file specification supplants the standard mechanism by which JDE
determines which file to compile.  If this option's value is the empty string,
the standard mechanism applies.  If non-empty, it is appended directly to the compiler invocation's argument list."
  :group 'jde-project
  :type 'string)


;;; *Replaces* jde version to choose jde-compile-file-spec if defined.  New
;;; code is commented with "WGG".
(defadvice jde-make-compile-command (around aspectj-compile activate)
  "Constructs the java compile command as: jde-compiler + options + buffer file name or file-spec."
  (cond 
   ((string= jde-compile-file-specification "")
    ad-do-it)
   (t
    (let ((more-args (ad-get-arg 0)))
      (setq ad-return-value
	    (concat jde-compiler " " 
		    (jde-get-compile-options) 
		    (if (not (string= more-args ""))
			(concat " " more-args))
		    " "
		    (if (not (string= jde-compile-file-specification "")) ; WGG
			jde-compile-file-specification                    ; WGG
		      (file-name-nondirectory buffer-file-name))))))))


(defadvice jde-java-build (around aspectj-build activate)
  "Extends jde-java-build to use jde-compile-file-specification."
  (cond 
   ((string= jde-compile-file-specification "")
    ad-do-it)
   (t
    (setq ad-return-value (jde-compile)))))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Provides our own beanshell functionality for java-mode users.
;;;

(require 'comint)

;; WGG - from jde-run.el (modified)
(defcustom aspectj-java-vm "java"
  "*Specify Java interpreter for non-Windows platforms.  Applies only to
extension of java-mode."
  :group 'aspectj
  :type 'string)

;; WGG - from jde-run.el (modified)
(defcustom aspectj-java-vm-w "javaw"
  "*Specify Java interpreter for Windows platforms.  Applies only to
extension of java-mode.
On NTEmacs, the aspectj-mode must use the NTEmacs show window option in order
to run the Java interpreter. This in turn requires that aspectj-mode use
the javaw version of the JDK Java interpreter to prevent a console
window from appearing every time you run an application. Sound
confusing? It is, but it works."
  :group 'aspectj
  :type 'string)

(defcustom aspectj-bsh-startup-timeout 10
  "*Length of time in seconds aspectj minor mode waits for the Beanshell to
startup.  Increase the value of this variable if you get Lisp errors
on BeanShell startup on Unix."
  :group 'aspectj
  :type 'integer)

(defcustom aspectj-bsh-eval-timeout 20
  "*Length of time in seconds aspectj minor mode waits for the Beanshell to
evaluate a Java expression before giving up and signaling an error.  Applies
only to extension of java-mode."
  :group 'aspectj
  :type 'integer)

;; WGG - from jde-run.el (unmodified)
(defmacro aspectj-save-w32-show-window (&rest body)
  "Saves the value of the w32-start-process-show-window variable
before evaluating body and restores the value afterwards."
  `(if (and (eq system-type 'windows-nt)
	    (not aspectj-xemacsp))
       (if (boundp 'win32-start-process-show-window)
	   (let ((save win32-start-process-show-window))
	     (setq win32-start-process-show-window t)
	     ,@body
	     (setq win32-start-process-show-window save))
	 (let ((save w32-start-process-show-window))
	   (setq w32-start-process-show-window t)
	   ,@body
	   (setq w32-start-process-show-window save)))
     ,@body))


;; WGG - from beanshell.el, revision 1.16
(defun aspectj-bsh-internal ()
  (if (eq major-mode 'jde-mode)
      (bsh-internal)
    (let ((bsh-buffer-name "*bsh*"))
      (if (not (comint-check-proc bsh-buffer-name))
	  (let* ((bsh-buffer (get-buffer-create bsh-buffer-name))
		 (vm (if (eq system-type 'windows-nt)
			 aspectj-java-vm-w
		       aspectj-java-vm))
		 (vm-args
		  (list
		   "-classpath"
		   (getenv "CLASSPATH")
		   "bsh.Interpreter"))
		 (dir (if (buffer-file-name)
			  (file-name-directory (buffer-file-name))
			default-directory)))

	    (save-excursion
	      (set-buffer bsh-buffer)
	      (erase-buffer)

	      (cd dir)
	      (insert (concat "cd " dir "\n"))
	      (insert 
	       (concat vm " "
		       (mapconcat (lambda (x) x) vm-args " ") "\n\n"))

	      (comint-mode)
	      (setq comint-prompt-regexp "bsh % "))
	    (aspectj-save-w32-show-window
	     ;; (message "%s" (nth 1 vm-args))
	     (message "%s" "Starting the BeanShell. Please wait...")
	     (comint-exec bsh-buffer "bsh" vm nil vm-args))
	    (process-kill-without-query (get-buffer-process bsh-buffer)))))))


(defun aspectj-bsh-eval-filter (process result)
  (if (eq major-mode 'jde-mode)
      (bsh-eval-filter process result)
    (let ((end-of-result (string-match ".*bsh % " result)))
      ;; Check for case
      ;;   %bsh\n...eval output...%bsh\n
      ;; This can happen because the beanshell outputs two or more
      ;; prompts after evaluating some expressions.
      ;; Thanks to Stephane Nicolas.
      ;; (if (eq end-of-result 0)
      ;; (accept-process-output process 0 5))
      (if end-of-result
	  (setq bsh-tq-reply (concat bsh-tq-reply (substring result 0 end-of-result)))
	(setq bsh-tq-reply (concat bsh-tq-reply result))
	(accept-process-output process aspectj-bsh-eval-timeout 5)))))


;;; Keep track of whether bsh can be fired up.
(defvar aspectj-bsh-available nil)
(defvar aspectj-bsh-checked nil)

;;; Maybe a bit of a hack on the pattern, but accounting for a number of
;;; legal jar-file naming conventions.
(defun aspectj-bsh-availablep ()
  (if aspectj-bsh-checked
      aspectj-bsh-available
    (setq aspectj-bsh-checked t)
    (setq aspectj-bsh-available
      (string-match "[\\\\\\./;:-]bsh[\\.-]" (getenv "CLASSPATH")))))


(defun aspectj-bsh-eval (expr)
  "Uses the BeanShell Java interpreter to evaluate a Java statement.
If the interpreter is not already running, this function starts
the interpreter. This function returns any text output by the
Java interpreter's standard out or standard error pipes."
  (if (eq major-mode 'jde-mode)
      (bsh-eval expr)
    (if (not (aspectj-bsh-availablep))
	     (progn
	       (message "AspectJ Mode Warning: beanshell not found on CLASSPATH")
	       "nil")
      (let* ((bsh-process
	      (if (get-process "bsh")
		  (get-process "bsh")
		(let (proc)
		  (aspectj-bsh-internal)
		  (setq proc (get-process "bsh"))
		  (if (eq system-type 'windows-nt)
		      (accept-process-output proc)
		    (while (accept-process-output proc aspectj-bsh-startup-timeout 0)))
		  proc)))
	     (comint-filter (if bsh-process (process-filter bsh-process))))
	(when bsh-process
	  (setq bsh-tq-reply nil)
	  (set-process-filter bsh-process 'aspectj-bsh-eval-filter)
	  ;; (message "Evaluating: %s" expr)
	  (process-send-string bsh-process (concat expr "\n"))
	  (if (not (accept-process-output bsh-process aspectj-bsh-eval-timeout 100))
	      (error "No reply from BeanShell"))
	  (set-process-filter bsh-process comint-filter)
	  (if (string-match "// Error:" bsh-tq-reply)
	      (progn
		(message
		 "Beanshell expression evaluation error.\n  Expression: %s\n  Error: %s"
		 expr bsh-tq-reply)
		(error "Beanshell eval error. See messages buffer for details.")))
	  bsh-tq-reply)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Protect against bsh invocation race conditions.  For Unix/Linux systems.
;;;
;;; It may be that Windows has this problem, too.  I've seen it happen after
;;; Windows was running for a few weeks.  WGG - WARNING.
;;;

;;; Protects against recursive bsh-eval process thing from running out of
;;; control while waiting for data.  Fairly harmless.  Should probably 
;;; rewrite bsh-eval so that it isn't unnecessarily recursive.  WGG
(setq max-specpdl-size 10000) ; default is 600, adjust as needed :)

;;; Waits (a limited amount of time) for bsh to emit initial prompt
;;; (which it will eat).  Fixes an apparent race condition (discovered on
;;; Solaris) in Emacs starting a process and then recursively waiting for
;;; it's first output while the process slowly starts.  WGG
;;;

; ;; WGG - not currently used.  Beanshell was updated with a fix.
; (defun aspectj-bsh-prime ()
;   "Waits for initial prompt from bsh on Unix.  Prints the output it reads,
; which is often empty.  Borrowed from function bsh-eval in beanshell.el."
;   (when (not (memq system-type '(vax-vms windows-nt ms-dos))) ; from version.el
;     (let* ((bsh-process
; 	    (if (get-process "bsh")
; 		(get-process "bsh")
; 	      (progn
; 		(aspectj-bsh-internal)
; 		(accept-process-output (get-process "bsh"))
; 		(get-process "bsh")))) 
; 	   (comint-filter (process-filter bsh-process)))
;       (setq bsh-tq-reply nil)
;       (set-process-filter bsh-process 'aspectj-bsh-eval-filter)
;       (if (not (accept-process-output bsh-process aspectj-bsh-startup-timeout 5))
; 	  (message "Saw no initial prompt from Bean Shell"))
;       (set-process-filter bsh-process comint-filter)
;       (message "Initial Bean Shell output: %s" bsh-tq-reply))))

;; ---- end aspectj-mode.el
(provide 'aspectj-support)
