/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jesse Glick.
 */

package org.apache.tools.ant.module.api;

import java.util.*;
import java.util.Map; // override org.apache.tools.ant.Map

import org.apache.tools.ant.Task;

import org.apache.tools.ant.module.AntModule;

// [PENDING] Perhaps this class could be replaced by folder lookup in system file system,
// e.g. if you add in a layer /Ant/Tasks/footask[org-netbeans-modules-foo-FooTask].instance,
// then the Ant module will define <footask> for you.

/** Place where ad-hoc taskdef's and typedef's can be registered by other modules or code.
 */
public final class DefinitionRegistry {

    private DefinitionRegistry () {}

    private static final Map taskDefs = new HashMap (); // Map<String,Class>
    private static final Map typeDefs = new HashMap (); // Map<String,Class>

    /** Register a task or type definition.
     * @param name the name of the task or type
     * @param taskClazz the implementing class
     * @param task true for tasks, false for types
     * @throws ClassCastException if supposed to be a task and the class is not a subclass of Task
     */
    public static void register (String name, Class taskClazz, boolean task) throws ClassCastException {
        if (task && ! Task.class.isAssignableFrom (taskClazz)) {
            throw new ClassCastException ("Not a subclass of org.apache.tools.ant.Task: " + taskClazz.getName ()); // NOI18N
        }
        AntModule.err.log ("registering def: " + name + "=" + taskClazz.getName () + " (task=" + task + ")");
        Map defs = task ? taskDefs : typeDefs;
        synchronized (defs) {
            defs.put (name, taskClazz);
        }
        if (IntrospectedInfo.isDefaultsPrepared ()) {
            IntrospectedInfo.getDefaults ().register (name, taskClazz, task);
        } // else when prepared it will read defs itself
    }

    /** Unregister a task or type definition.
     * @param name the name to unregister
     * @param task if true, a task, else a type
     */
    public static void unregister (String name, boolean task) {
        AntModule.err.log ("unregistering def: " + name + " (task=" + task + ")");
        Map defs = task ? taskDefs : typeDefs;
        synchronized (defs) {
            defs.remove (name);
        }
    }

    /** Get a set of definitions.
     * @param task if true, get tasks, else get types
     * @return an immutable map of names to implementing classes
     */
    public static Map getDefs (boolean task) {
        Map defs = task ? taskDefs : typeDefs;
        synchronized (defs) {
            return new HashMap (defs);
        }
    }

}
