/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jesse Glick.
 */
 
package org.apache.tools.ant.module.nodes;

import java.beans.*;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.*;
import java.util.*;

import org.w3c.dom.*;

import org.apache.tools.ant.types.Reference;
import org.apache.tools.ant.types.EnumeratedAttribute;

import org.openide.*;
import org.openide.cookies.InstanceCookie;
import org.openide.execution.NbClassPath;
import org.openide.nodes.*;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.datatransfer.NewType;

import org.apache.tools.ant.module.AntModule;
import org.apache.tools.ant.module.AntSettings;
import org.apache.tools.ant.module.api.ElementCookie;
import org.apache.tools.ant.module.api.IntrospectedInfo;
import org.apache.tools.ant.module.xml.ElementSupport;

public class DataTypeNode extends ElementNode {
    
    // known attributes:
    private Map attrs; // Map<String,String>
    // known subelements:
    private Map subels; // Map<String,String>
    private Boolean supportsTextFlag;
    
    private String clazz;
    
    public DataTypeNode (Element el, String clazz) {
        this (el, computeChildren (el, clazz));
        //AntModule.err.log ("DataTypeNode: " + getDisplayName () + " clazz=" + clazz);
        this.clazz = clazz;
    }
    
    protected DataTypeNode (Element el, Children children) {
        super (el, children);
    }
    
    private static Children computeChildren (Element el, String clazz) {
        if (isEmpty (clazz)) {
            return Children.LEAF;
        } else {
            return new ElementChildren (el);
        }
    }
    
    protected static boolean isEmpty (String clazz) {
        return isEmpty (clazz, IntrospectedInfo.getDefaults ()) ||
               isEmpty (clazz, AntSettings.getDefault ().getCustomDefs ());
    }
    
    private static boolean isEmpty (String clazz, IntrospectedInfo info) {
        if (info.isKnown (clazz)) {
            return info.getElements (clazz).isEmpty ();
        } else {
            // Be safe:
            return false;
        }
    }
    
    protected DataTypeNode (Element el) {
        super (el);
    }
    
    protected ElementCookie createElementCookie () {
        //AntModule.err.log ("DataTypeNode.elementCookie: clazz=" + clazz);
        if (clazz != null) {
            return new ElementSupport.Instance (el, clazz);
        } else {
            return new ElementSupport (el);
        }
    }
    
    protected void initDisplay () {
        super.initDisplay ();
        setIconBase ("/org/apache/tools/ant/module/resources/DataTypeIcon");
    }
    
    protected Map getAttrs () {
        if (attrs != null) return attrs;
        synchronized (this) {
            if (attrs != null) return attrs;
            updateStuff ();
            return attrs;
        }
    }
    
    protected Map getSubels () {
        if (subels != null) return subels;
        synchronized (this) {
            if (subels != null) return subels;
            updateStuff ();
            return subels;
        }
    }
    
    protected boolean supportsText () {
        if (supportsTextFlag != null) return supportsTextFlag.booleanValue ();
        synchronized (this) {
            if (supportsTextFlag != null) return supportsTextFlag.booleanValue ();
            updateStuff ();
            return supportsTextFlag.booleanValue ();
        }
    }
    
    private void updateStuff () {
        attrs = new HashMap (); // Collections.EMPTY_MAP
        subels = new HashMap (); // Collections.EMPTY_MAP
        supportsTextFlag = Boolean.FALSE;
        String name = el.getNodeName ();
        //findStuff (attrs, subels, , name);
        //findStuff (attrs, subels, , name);
        InstanceCookie cookie = (InstanceCookie) getCookie (InstanceCookie.class);
        if (cookie == null) return;
        try {
            // [PENDING] probably there should really be a IntrospectionCookie
            // that would give the right info without trying to load classes...
            // ElementChildren could benefit as well.
            String clazz = cookie.instanceClass ().getName ();
            IntrospectedInfo info = IntrospectedInfo.getDefaults ();
            if (info.isKnown (clazz)) {
                attrs = info.getAttributes (clazz);
                subels = info.getElements (clazz);
                supportsTextFlag = new Boolean (info.supportsText (clazz));
            } else {
                info = AntSettings.getDefault ().getCustomDefs ();
                if (info.isKnown (clazz)) {
                    attrs = info.getAttributes (clazz);
                    subels = info.getElements (clazz);
                    supportsTextFlag = new Boolean (info.supportsText (clazz));
                } // else oh well, no info
            }
        } catch (IOException ioe) {
            // whatever
        } catch (ClassNotFoundException cnfe) {
            // whatever
        }
    }
    
    protected void addProperties (Sheet.Set props) {
        Iterator it = getAttrs ().entrySet ().iterator ();
        while (it.hasNext ()) {
            Map.Entry entry = (Map.Entry) it.next ();
            String attr = (String) entry.getKey ();
            if (el.getAttribute (attr).indexOf ("${") != -1) { // NOI18N
                // If it is already using variable substitution, we cannot provide a reasonable
                // proped for it, so just treat it as an unknown string property.
                props.put (new AntProperty (el, attr));
                continue;
            }
            String clazzname = (String) entry.getValue ();
            if (clazzname.equals ("org.apache.tools.ant.types.Reference")) { // NOI18N
                // Has a special proped.
                props.put (new ReferenceProperty (attr));
                continue;
            }
            if (clazzname.equals ("org.apache.tools.ant.types.Path")) { // NOI18N
                props.put (new PathProperty (attr));
                continue;
            }
            if (clazzname.equals ("java.lang.String") && attr.equals ("path") && // NOI18N
                (clazz.equals ("org.apache.tools.ant.types.Path") || // NOI18N
                 clazz.equals ("org.apache.tools.ant.types.Path$PathElement"))) { // NOI18N
                // These are also really Path-like things, though of type String rather than Path
                props.put (new PathProperty (attr));
                continue;
            }
            Class clazz;
            if (clazzname.equals ("boolean")) clazz = Boolean.TYPE; // NOI18N
            else if (clazzname.equals ("char")) clazz = Character.TYPE; // NOI18N
            else if (clazzname.equals ("byte")) clazz = Byte.TYPE; // NOI18N
            else if (clazzname.equals ("short")) clazz = Short.TYPE; // NOI18N
            else if (clazzname.equals ("int")) clazz = Integer.TYPE; // NOI18N
            else if (clazzname.equals ("long")) clazz = Long.TYPE; // NOI18N
            else if (clazzname.equals ("float")) clazz = Float.TYPE; // NOI18N
            else if (clazzname.equals ("double")) clazz = Double.TYPE; // NOI18N
            else {
                try {
                    clazz = TopManager.getDefault ().currentClassLoader ().loadClass (clazzname);
                } catch (ClassNotFoundException cnfe) {
                    // Fine, maybe a custom type or something.
                    clazz = String.class;
                }
            }
            if (EnumeratedAttribute.class.isAssignableFrom (clazz)) {
                // Special propeds.
                props.put (new EnumeratedProperty (attr, clazz));
                continue;
            }
            props.put (new DataProperty (attr, clazz));
        }
        org.openide.nodes.Node.Property idProp = new AntProperty (el, "id"); // NOI18N
        idProp.setDisplayName (NbBundle.getMessage (DataTypeNode.class, "PROP_data_id"));
        idProp.setShortDescription (NbBundle.getMessage (DataTypeNode.class, "HINT_data_id"));
        props.put (idProp);
        if (supportsText ()) props.put (new TextProperty ());
    }
    
    private class DataProperty extends PropertySupport.ReadWrite {
        private Constructor cons;
        public DataProperty (String attr, Class clazz) {
            super (attr, clazz, attr, attr);
            if (! clazz.isPrimitive ()) {
                try {
                    cons = clazz.getConstructor (new Class[] { String.class });
                } catch (NoSuchMethodException nsme) {
                    // E.g. Path has no such constructor.
                    //AntModule.err.notify (nsme);
                    cons = null;
                }
            }
        }
        public Object getValue () {
            Class clazz = getValueType ();
            String val = el.getAttribute (this.getName ());
            try {
                if (clazz == Boolean.TYPE) return new Boolean (val);
                else if (clazz == Character.TYPE) return new Character (val.length () > 0 ? val.charAt (0) : 0);
                else if (clazz == Byte.TYPE) return new Byte (val);
                else if (clazz == Short.TYPE) return new Short (val);
                else if (clazz == Integer.TYPE) return new Integer (val);
                else if (clazz == Long.TYPE) return new Long (val);
                else if (clazz == Float.TYPE) return new Float (val);
                else if (clazz == Double.TYPE) return new Double (val);
                else {
                    if (cons == null) return null;
                    try {
                        return cons.newInstance (new Object[] { val });
                    } catch (Exception e) { // all different kinds...
                        AntModule.err.notify (ErrorManager.INFORMATIONAL, e);
                        return null;
                    }
                }
            } catch (NumberFormatException nfe) {
                if (! val.equals ("")) { // NOI18N
                    AntModule.err.notify (ErrorManager.INFORMATIONAL, nfe);
                }
                return null;
            }
        }
        public void setValue (Object v) throws InvocationTargetException {
            try {
                String s = ((v == null) ? null : v.toString ());
                if (s == null || s.equals ("")) {
                    el.removeAttribute (this.getName ());
                } else {
                    //Class clazz = getValueType ();
                    el.setAttribute (this.getName (), s);
                }
            } catch (DOMException dome) {
                throw new InvocationTargetException (dome);
            }
        }
        public boolean supportsDefaultValue () {
            return true;
        }
        public void restoreDefaultValue () throws InvocationTargetException {
            setValue (null);
        }
    }
    
    private class EnumeratedProperty extends AntProperty {
        private Class clazz;
        public EnumeratedProperty (String attr, Class clazz) {
            super (el, attr);
            this.clazz = clazz;
        }
        public PropertyEditor getPropertyEditor () {
            return new EnumeratedEditor (clazz);
        }
    }
    
    private class EnumeratedEditor extends PropertyEditorSupport {
        String[] tags;
        public EnumeratedEditor (Class clazz) {
            try {
                EnumeratedAttribute instance = (EnumeratedAttribute) clazz.newInstance ();
                tags = instance.getValues ();
            } catch (Exception e) {
                AntModule.err.notify (ErrorManager.INFORMATIONAL, e);
                tags = new String[] { };
            }
        }
        public String getAsText () {
            return (String) getValue ();
        }
        public void setAsText (String v) {
            setValue (v);
        }
        public String[] getTags () {
            return tags;
        }
    }
    
    private class ReferenceProperty extends PropertySupport.ReadWrite {
        public ReferenceProperty (String attr) {
            super (attr, Reference.class,
                   attr, //NbBundle.getMessage (DataTypeNode.class, "PROP_data_refid"),
                   NbBundle.getMessage (DataTypeNode.class, "HINT_data_refid"));
        }
        public Object getValue () {
            return new Reference (el.getAttribute (this.getName ()));
        }
        public void setValue (Object o) throws IllegalArgumentException, InvocationTargetException {
            try {
                if (o == null) {
                    el.removeAttribute (this.getName ());
                } else {
                    if (! (o instanceof Reference)) throw new IllegalArgumentException ();
                    //AntModule.err.log ("ReferenceProperty.setValue: o=" + o + " id=" + ((Reference) o).getRefId ());
                    el.setAttribute (this.getName (), ((Reference) o).getRefId ());
                }
            } catch (DOMException dome) {
                throw new InvocationTargetException (dome);
            }
        }
        public boolean supportsDefaultValue () {
            return true;
        }
        public void restoreDefaultValue () throws IllegalArgumentException, InvocationTargetException {
            setValue (null);
        }
        public PropertyEditor getPropertyEditor () {
            return new ReferenceEditor ();
        }
    }
    
    private class ReferenceEditor extends PropertyEditorSupport {
        public String getAsText () {
            return ((Reference) getValue ()).getRefId ();
        }
        public void setAsText (String s) {
            //AntModule.err.log ("ReferenceEditor.setAsText: s=" + s);
            setValue (new Reference (s));
        }
        public String[] getTags () {
            Set tags = new TreeSet ();
            NodeList nl = el.getOwnerDocument ().getElementsByTagName ("*"); // NOI18N
            int len = nl.getLength ();
            for (int i = 0; i < len; i++) {
                Attr id = ((Element) nl.item (i)).getAttributeNode ("id"); // NOI18N
                if (id != null) {
                    String rid = id.getValue ();
                    if (rid != null && ! rid.equals ("")) { // NOI18N
                        tags.add (rid);
                    }
                }
            }
            //AntModule.err.log ("ReferenceEditor.tags=" + tags);
            return (String[]) tags.toArray (new String[tags.size ()]);
        }
    }
    
    private class PathProperty extends PropertySupport.ReadWrite {
        public PathProperty (String attr) {
            super (attr, NbClassPath.class, attr, attr);
        }
        public Object getValue () {
            String path = el.getAttribute (this.getName ());
            path = subst (path, File.separatorChar, '/');
            path = subst (path, File.separatorChar, '\\');
            path = subst (path, File.pathSeparatorChar, ':');
            path = subst (path, File.pathSeparatorChar, ';');
            return new NbClassPath (path);
        }
        private String subst (String path, char local, char generic) {
            if (local == generic) {
                return path;
            } else {
                return path.replace (generic, local);
            }
        }
        public void setValue (Object o) throws IllegalArgumentException, InvocationTargetException {
            try {
                if (o == null) {
                    el.removeAttribute (this.getName ());
                } else {
                    if (! (o instanceof NbClassPath)) throw new IllegalArgumentException ();
                    String path = ((NbClassPath) o).getClassPath ();
                    path = subst (path, '/', File.separatorChar);
                    path = subst (path, ':', File.pathSeparatorChar);
                    el.setAttribute (this.getName (), path);
                }
            } catch (DOMException dome) {
                throw new InvocationTargetException (dome);
            }
        }
        public boolean supportsDefaultValue () {
            return true;
        }
        public void restoreDefaultValue () throws IllegalArgumentException, InvocationTargetException {
            setValue (null);
        }
    }
    
    private class TextProperty extends PropertySupport.ReadWrite {
        public TextProperty () {
            super ("text", String.class, // NOI18N
                   NbBundle.getMessage (DataTypeNode.class, "PROP_data_text"),
                   NbBundle.getMessage (DataTypeNode.class, "HINT_data_text"));
        }
        public Object getValue () {
            // [PENDING] should the value be trimmed somehow?
            NodeList nl = el.getChildNodes ();
            StringBuffer buf = new StringBuffer ();
            for (int i = 0; i < nl.getLength (); i++) {
                if (nl.item (i) instanceof Text) {
                    // Note: can throw DOMException.
                    buf.append (((Text) nl.item (i)).getData ());
                }
            }
            return buf.toString ();
        }
        public void setValue (Object o) throws IllegalArgumentException, InvocationTargetException {
            if (! (o instanceof String)) throw new IllegalArgumentException ();
            NodeList nl = el.getChildNodes ();
            try {
                // Wipe out any existing text/CDATA nodes.
                // [PENDING] would be better to replace exactly one existing one
                // with a node of the same type, to minimize disruption.
                for (int i = 0; i < nl.getLength (); i++) {
                    if (nl.item (i) instanceof Text) {
                        el.removeChild (nl.item (i));
                    }
                }
                // Readd what we want.
                // Always creates CDATA, on the grounds that that is safest.
                // Don't try to indent--want this inline with no intervening whitespace.
                el.appendChild (el.getOwnerDocument ().createCDATASection ((String) o));
            } catch (DOMException dome) {
                throw new InvocationTargetException (dome);
            }
        }
    }
    
    protected boolean canPasteElement (Element el2) {
        return getSubels ().containsKey (el2.getNodeName ());
    }
    
    public NewType[] getNewTypes () {
        List names = new ArrayList (getSubels ().keySet ()); // List<String>
        Collections.sort (names);
        NewType[] types = new NewType[names.size ()];
        for (int i = 0; i < types.length; i++) {
            types[i] = new SubElementNewType ((String) names.get (i));
        }
        return types;
    }
    
    private class SubElementNewType extends NewType {
        private String name;
        public SubElementNewType (String name) {
            this.name = name;
        }
        public String getName () {
            return name;
        }
        public HelpCtx getHelpCtx () {
            return new HelpCtx ("org.apache.tools.ant.module.node-manip");
        }
        public void create () throws IOException {
            try {
                Element el2 = el.getOwnerDocument ().createElement (name);
                ElementNode.appendWithIndent (el, el2);
            } catch (DOMException dome) {
                IOException ioe = new IOException ();
                AntModule.err.annotate (ioe, dome);
                throw ioe;
            }
        }
    }
    
}
