/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jesse Glick.
 */
 
package org.apache.tools.ant.module.nodes;

import java.io.IOException;
import java.awt.datatransfer.*;
import java.beans.*;
import java.util.*;

import org.openide.*;
import org.openide.actions.*;
import org.openide.cookies.CompilerCookie;
import org.openide.cookies.ExecCookie;
import org.openide.cookies.InstanceCookie;
import org.openide.nodes.*;
import org.openide.util.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.*;

import org.w3c.dom.*;
import org.w3c.dom.events.*;

import org.apache.tools.ant.module.AntModule;
import org.apache.tools.ant.module.api.AntProjectCookie;
import org.apache.tools.ant.module.api.ElementCookie;
import org.apache.tools.ant.module.xml.ElementSupport;

public class ElementNode extends AbstractNode implements org.w3c.dom.events.EventListener, Runnable, NodeListener {
    
    protected Element el;
    private String error;
    private ElementCookie elCookie;
    private boolean attachedCookieListener = false; // #9952
    
    public ElementNode (Element el, String error) {
        this (el);
        if (error == null) throw new NullPointerException ();
        this.error = error;
        initDisplay (); // again, since error was null the first time
    }
    
    protected ElementNode (Element el) {
        this (el, new ElementChildren (el));
    }
    
    protected ElementNode (Element el, Children ch) {
        super (ch);
        this.el = el;
        initDisplay ();
        if (el instanceof EventTarget) {
            // [PENDING] WeakListener.eventDOM? However EventListener !< java.util.EventListener.
            ((EventTarget) el).addEventListener ("DOMSubtreeModified", this, false); // NOI18N
            // XXX just DOMSubtreeModified should be enough, but in current Xerces then
            // attr changes will not be fired!
            ((EventTarget) el).addEventListener ("DOMAttrModified", this, false); // NOI18N
            //((EventTarget) el).addEventListener ("DOMNodeInserted", this, false); // NOI18N
            //((EventTarget) el).addEventListener ("DOMNodeRemoved", this, false); // NOI18N
            //((EventTarget) el).addEventListener ("DOMCharacterDataModified", this, false); // NOI18N
        } else {
            AntModule.err.log (ErrorManager.WARNING, "Element was not an EventTarget: " + el); // NOI18N
        }
    }
    
    protected ElementCookie createElementCookie () {
        return new ElementSupport (el);
    }
    
    protected void initDisplay () {
        setNameSuper (semiUniqueName (el));
        setDisplayName (displayName (el));
        if (error != null) {
            AntModule.err.log ("ElementNode.initDisplay: error=" + error);
            setShortDescription (error);
        } else {
            setShortDescription (fullText (el));
        }
        // Hack to kill expand handles when feasible; doesn't work!
        /*
         if (getShortDescription ().indexOf ("...") == -1) { // NOI18N
             getChildren ().getNodes ();
         }
         */
        setIconBase ("/org/apache/tools/ant/module/resources/ElementIcon");
    }
    
    protected void setNameSuper (String n) {
        super.setName (n);
    }
    
    protected static String semiUniqueName (Element el) {
        Attr attr = el.getAttributeNode ("id"); // NOI18N
        if (attr != null) return attr.getValue ();
        attr = el.getAttributeNode ("name"); // NOI18N
        if (attr != null) return attr.getValue ();
        return el.getNodeName ();
    }
    
    protected static String displayName (Element el) {
        Attr attr = el.getAttributeNode ("name"); // NOI18N
        if (attr != null) {
            return NbBundle.getMessage (ElementNode.class, "LBL_node_with_name", el.getNodeName (), attr.getValue ());
        } else {
            return el.getNodeName ();
        }
    }
    
    protected static String fullText (Element el) {
        // NOI18N thruout...
        StringBuffer buf = new StringBuffer ();
        buf.append ('<');
        buf.append (el.getNodeName ());
        NamedNodeMap attrs = el.getAttributes ();
        int len = attrs.getLength ();
        for (int i = 0; i < len; i++) {
            Attr attr = (Attr) attrs.item (i);
            buf.append (' ');
            buf.append (attr.getName ());
            buf.append ('=');
            buf.append ('"');
            // Note: no attempt at escaping vals, etc.
            buf.append (attr.getValue ());
            buf.append ('"');
        }
        if (el.getChildNodes ().getLength () > len) {
            // Have some children other than attributes.
            buf.append (">...</"); // NOI18N
            buf.append (el.getNodeName ());
            buf.append ('>');
        } else {
            buf.append ("/>"); // NOI18N
        }
        return buf.toString ();
    }

    /** Inherit cookies from parent node.
     * Permits e.g. subnodes to be saved directly.
     */
    public org.openide.nodes.Node.Cookie getCookie (Class clazz) {
        org.openide.nodes.Node.Cookie supe = super.getCookie (clazz);
        if (supe != null) return supe;
        if (ElementCookie.class.isAssignableFrom (clazz) ||
            InstanceCookie.class.isAssignableFrom (clazz)) {
            // Do not propagate ElementCookie/InstanceCookie!
            if (elCookie == null) { // don't bother synchronizing, no harm if collision
                final ElementCookie cookie = createElementCookie ();
                elCookie = cookie;
                // Do not add in this thread; can go into cycle with CookieAction:
                // Note that we caches the result, so even if a zillion
                // NodeAction's ask for this in sequence, still we will have only one.
                RequestProcessor.postRequest (new Runnable () {
                    public void run () {
                        getCookieSet ().add (cookie);
                    }
                });
            }
            if (clazz.isInstance (elCookie)) {
                return elCookie;
            } else {
                return null;
            }
        }
        if (CompilerCookie.class.isAssignableFrom (clazz) ||
            ExecCookie.class.isAssignableFrom (clazz)) {
            // Never propagate these, they should be taken from the
            // relevant object only.
            return null;
        }
        org.openide.nodes.Node parent = getParentNode ();
        if (parent != null) {
            if (! attachedCookieListener) {
                attachedCookieListener = true;
                parent.addNodeListener (WeakListener.node (this, parent));
            }
            return parent.getCookie (clazz);
        }
        return null;
    }
    
    public HelpCtx getHelpCtx () {
        return new HelpCtx ("org.apache.tools.ant.module.node-manip");
    }
    
    protected Sheet createSheet () {
        Sheet sheet = super.createSheet ();
        Sheet.Set props = sheet.get (Sheet.PROPERTIES);
        if (props == null) {
            props = Sheet.createPropertiesSet ();
            sheet.put (props);
        }
        addProperties (props);
        return sheet;
    }
    
    protected void addProperties (Sheet.Set props) {
        // Basic implementation just adds existing ones.
        NamedNodeMap attrs = el.getAttributes ();
        int len = attrs.getLength ();
        for (int i = 0; i < len; i++) {
            props.put (new AntProperty (el, attrs.item (i).getNodeName ()));
        }
        props.put (new AntProperty (el, "id")); // NOI18N
        // [PENDING] text attributes?
    }
    
    protected SystemAction[] createActions () {
        return new SystemAction[] {
            SystemAction.get (OpenLocalExplorerAction.class),
            null,
            SystemAction.get (CutAction.class),
            SystemAction.get (CopyAction.class),
            SystemAction.get (PasteAction.class),
            null,
            SystemAction.get (DeleteAction.class),
            SystemAction.get (RenameAction.class),
            null,
            SystemAction.get (NewAction.class),
            null,
            SystemAction.get (ToolsAction.class),
            SystemAction.get (PropertiesAction.class),
        };
    }
    
    public SystemAction getDefaultAction () {
        return SystemAction.get (PropertiesAction.class);
    }
    
    public boolean canDestroy () {
        org.w3c.dom.Node parent = el.getParentNode ();
        if (parent == null) return false;
        short type = parent.getNodeType ();
        boolean ok = (type == org.w3c.dom.Node.ELEMENT_NODE);
        //AntModule.err.log ("ElementNode.canDestroy: type=" + type + " ENTITY_NODE=" + org.w3c.dom.Node.ENTITY_NODE + " ok=" + ok);
        return ok;
    }
    
    public void destroy () throws IOException {
        org.w3c.dom.Node parent = el.getParentNode ();
        if (parent == null) throw new IOException ("Node had no DOM parent"); // NOI18N
        try {
            parent.removeChild (el);
            AntModule.err.log ("ElementNode.destroy: el=" + el + " done");
        } catch (DOMException dome) {
            IOException ioe = new IOException ();
            AntModule.err.annotate (ioe, dome);
            throw ioe;
        }
        // XXX sanity check...
        NodeList nue = parent.getChildNodes ();
        for (int i = 0; i < nue.getLength (); i++) {
            if (nue.item (i) == el) {
                throw new IOException ("Node.removeChild did not work!"); // NOI18N
            }
        }
        AntProjectCookie cookie = (AntProjectCookie) getCookie (AntProjectCookie.class);
        if (cookie != null) {
            if (cookie.getDocument () != el.getOwnerDocument ()) {
                throw new IOException ("DOM mismatch"); // NOI18N
            }
        }
        org.openide.nodes.Node parent2 = getParentNode ();
        if (parent2 != null) {
            ElementCookie cookie2 = (ElementCookie) parent2.getCookie (ElementCookie.class);
            if (cookie2 != null) {
                if (cookie2.getElement () != parent) {
                    throw new IOException ("I am not my father's son!"); // NOI18N
                }
            }
        }
    }
    
    public boolean canRename () {
        return el.getAttributeNode ("id") != null || el.getAttributeNode ("name") != null; // NOI18N
    }
    
    public void setName (String nue) throws IllegalArgumentException {
        try {
            Attr attr = el.getAttributeNode ("id"); // NOI18N
            if (attr != null) {
                attr.setNodeValue (nue);
            } else {
                attr = el.getAttributeNode ("name"); // NOI18N
                if (attr != null) {
                    attr.setNodeValue (nue);
                } else {
                    throw new IllegalArgumentException ();
                }
            }
        } catch (DOMException dome) {
            IllegalArgumentException iae = new IllegalArgumentException ();
            AntModule.err.annotate (iae, dome);
            throw iae;
        }
    }
    
    public boolean canCopy () {
        return true;
    }
    
    public boolean canCut () {
        return true;
    }
    
    private static final int INDENT_STEP = 4;
    static void appendWithIndent (Element parent, org.w3c.dom.Node child) throws DOMException {
        org.w3c.dom.Node doc = parent;
        int depth = -1;
        while (! (doc instanceof Document)) {
            doc = doc.getParentNode ();
            depth++;
        }
        String ws1;
        if (parent.hasChildNodes ()) {
            ws1 = spaces (INDENT_STEP);
        } else {
            ws1 = "\n" + spaces ((depth + 1) * INDENT_STEP);
        }
        parent.appendChild (((Document) doc).createTextNode (ws1));
        parent.appendChild (child);
        parent.appendChild (((Document) doc).createTextNode ("\n" + spaces (depth * INDENT_STEP)));
    }
    private static String spaces (int size) {
        char[] chars = new char[size];
        for (int i = 0; i < size; i++) {
            chars[i] = ' ';
        }
        return new String (chars);
    }
    
    public static class ElementPaste extends PasteType {
        private boolean cut;
        private Element el;
        private Element el2;
        public ElementPaste (Element el, Element el2, boolean cut) {
            this.el = el;
            this.el2 = el2;
            this.cut = cut;
        }
        public HelpCtx getHelpCtx () {
            return HelpCtx.DEFAULT_HELP; // XXX
        }
        public String getDisplayName () {
            // XXX is this really appropriate?
            return fullText (el2);
        }
        public Transferable paste () throws IOException {
            try {
                org.w3c.dom.Node toadd;
                boolean xdoc = (el.getOwnerDocument () != el2.getOwnerDocument ()); // #10757
                if (xdoc) {
                    toadd = el.getOwnerDocument ().importNode (el2, true);
                } else if (cut) {
                    toadd = el2;
                } else {
                    toadd = el2.cloneNode (true);
                }
                AntModule.err.log ("ElementPaste: el=" + el + " el2=" + el2 + " toadd=" + toadd);
                if (cut && xdoc) {
                    el2.getParentNode ().removeChild (el2);
                }
                appendWithIndent (el, toadd);
            } catch (DOMException dome) {
                IOException ioe = new IOException ();
                AntModule.err.annotate (ioe, dome);
                throw ioe;
            }
            return cut ? ExTransferable.EMPTY : null;
        }
    }
    
    protected void createPasteTypes (Transferable t, List l) {
        ElementCookie cookie = (ElementCookie) NodeTransfer.cookie (t, NodeTransfer.COPY, ElementCookie.class);
        if (cookie != null && canPasteElement (cookie.getElement ())) {
            l.add (new ElementPaste (el, cookie.getElement (), false));
        }
        cookie = (ElementCookie) NodeTransfer.cookie (t, NodeTransfer.MOVE, ElementCookie.class);
        if (cookie != null && canPasteElement (cookie.getElement ())) {
            l.add (new ElementPaste (el, cookie.getElement (), true));
        }
    }
    
    protected boolean canPasteElement (Element el2) {
        return true;
    }
    
    public NewType[] getNewTypes () {
        // None by default
        return new NewType[] { };
    }
    
    public void handleEvent (Event evt) {
        if (isRelevant (evt)) {
            RequestProcessor.postRequest (this);
        }
    }
    
    public void run () {
        initDisplay ();
        // Crude but good enough:
        firePropertyChange (null, null, null);
    }
    
    protected boolean isRelevant (Event evt) {
        /*
        String type = evt.getType ();
        if (type.equals ("DOMAttrModified")) { // NOI18N
         */
            return evt.getTarget () == el;
            /*
        } else if (type.equals ("DOMNodeInserted") || type.equals ("DOMNodeRemoved")) { // NOI18N
            MutationEvent mev = (MutationEvent) evt;
            return mev.getRelatedNode () == el;
        } else {
            return false;
        }
             */
    }
    
    // NodeListener:
    
    public void propertyChange (PropertyChangeEvent ev) {
        if (org.openide.nodes.Node.PROP_COOKIE.equals (ev.getPropertyName ())) { // #9952
            fireCookieChange ();
        }
    }
    
    public void childrenAdded (NodeMemberEvent ev) {}
    public void nodeDestroyed (NodeEvent ev) {}
    public void childrenRemoved (NodeMemberEvent ev) {}
    public void childrenReordered (NodeReorderEvent ev) {}
    
}
