/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jesse Glick.
 */

package org.apache.tools.ant.module.run;

import java.io.IOException;
import java.util.*;

import org.openide.*;
import org.openide.compiler.*;
import org.openide.compiler.Compiler;
import org.openide.cookies.CompilerCookie;
import org.openide.loaders.DataObject;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

import org.apache.tools.ant.module.AntModule;
import org.apache.tools.ant.module.api.AntProjectCookie;

public class AntCompilerType extends CompilerType {

    private String compileTarget = ""; // NOI18N
    private String buildTarget = "clean, all"; // NOI18N
    private String cleanTarget = "clean"; // NOI18N
    private Properties properties = new Properties ();
    
    private static final long serialVersionUID = -1209809503974704891L;

    public String getCompileTarget () {
        return compileTarget;
    }

    public synchronized void setCompileTarget (String nue) {
        String old = compileTarget;
        compileTarget = nue;
        firePropertyChange ("compileTarget", old, nue); // NOI18N
    }

    public String getCleanTarget () {
        return cleanTarget;
    }

    public synchronized void setCleanTarget (String nue) {
        String old = cleanTarget;
        cleanTarget = nue;
        firePropertyChange ("cleanTarget", old, nue); // NOI18N
    }

    public String getBuildTarget () {
        return buildTarget;
    }

    public synchronized void setBuildTarget (String nue) {
        String old = buildTarget;
        buildTarget = nue;
        firePropertyChange ("buildTarget", old, nue); // NOI18N
    }
    
    public Properties getProperties () {
        return properties;
    }
    
    public synchronized void setProperties (Properties nue) {
        Properties old = properties;
        properties = nue;
        firePropertyChange ("properties", old, nue); // NOI18N
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx ("org.apache.tools.ant.module.adv-comp-exec");
    }

    public void prepareJob (CompilerJob job, Class type, DataObject obj) {
        try {
            prepareJob0 (job, type, obj);
        } catch (RuntimeException re) {
            job.add (new BogusCompiler (re));
            AntModule.err.notify (ErrorManager.INFORMATIONAL, re);
        }
    }

    protected void prepareJob0 (CompilerJob job, Class type, DataObject obj) throws IllegalArgumentException {
        AntProjectCookie cookie = (AntProjectCookie) obj.getCookie (AntProjectCookie.class);
        if (cookie == null) { // #10680
            IllegalArgumentException iae = new IllegalArgumentException ("No AntProjectCookie for " + obj.getPrimaryFile ()); // NOI18N
            AntModule.err.annotate (iae, NbBundle.getMessage
                                    (AntCompilerType.class, "EXC_ant_compiler_on_non_ant_script", // NOI18N
                                     getName (), obj.getNodeDelegate ().getDisplayName ()));
            throw iae;
        }
        reallyPrepareJob (job, type, cookie);
    }
    
    protected void reallyPrepareJob (CompilerJob job, Class type, AntProjectCookie cookie) {
        if (cookie == null) throw new NullPointerException ();
        String target;
        if (type == CompilerCookie.Compile.class) {
            target = compileTarget;
        } else if (type == CompilerCookie.Build.class) {
            target = buildTarget;
        } else if (type == CompilerCookie.Clean.class) {
            target = cleanTarget;
        } else {
            // Other types not supported.
            return;
        }
        if (target == null || target.equals ("")) { // NOI18N
            job.add (new AntCompiler (cookie, null, properties));
        } else {
            // Handle comma-separated list of targets:
            StringTokenizer tok = new StringTokenizer (target, ", "); // NOI18N
            Compiler last = null;
            while (tok.hasMoreTokens ()) {
                String actualTarget = tok.nextToken ();
                Compiler nue = new AntCompiler (cookie, actualTarget, properties);
                job.add (nue);
                // Run in strict order, e.g. "clean,all" creates two compiler levels // NOI18N
                // which has a similar effect to "ant clean all": // NOI18N
                if (last != null) nue.dependsOn (last);
                last = nue;
            }
        }
    }

    static final class BogusCompiler extends Compiler { // #11099
        Throwable t;
        public BogusCompiler (Throwable t) {
            this.t = t;
        }
        public Class compilerGroupClass () {
            return BogusCompilerGroup.class;
        }
        protected boolean isUpToDate () {
            return false;
        }
    }
    public static final class BogusCompilerGroup extends CompilerGroup {
        private List throwables = new LinkedList (); // List<Throwable>
        public void add (Compiler c) throws IllegalArgumentException {
            if (! (c instanceof BogusCompiler)) throw new IllegalArgumentException ();
            throwables.add (((BogusCompiler) c).t);
        }
        public boolean start () {
            Iterator it = throwables.iterator ();
            while (it.hasNext ()) {
                Throwable t = (Throwable) it.next ();
                ErrorManager.Annotation[] anns = AntModule.err.findAnnotations (t);
                boolean shown = false;
                if (anns != null) {
                    for (int i = 0; i < anns.length; i++) {
                        String lm = anns[i].getLocalizedMessage ();
                        if (lm != null) {
                            fireErrorEvent (new ErrorEvent (this, null, 0, 0, lm, null));
                            shown = true;
                        }
                    }
                }
                if (! shown) {
                    fireErrorEvent (new ErrorEvent (this, null, 0, 0, t.toString (), null));
                }
            }
            return false;
        }
    }

}
