/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jayme C. Edwards, Jesse Glick.
 */
 
package org.apache.tools.ant.module.run;

import java.io.*;
import java.util.*;

import org.openide.*;
import org.openide.cookies.EditorCookie;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem; // override java.io.FileSystem
import org.openide.loaders.*;
import org.openide.text.Line;
import org.openide.util.NbBundle;
import org.openide.windows.*;

/** OutputStream for wrapping writes to the IDE.
 * Handles hyperlinks and so on.
 */
public class OutputWriterOutputStream extends OutputStream {

    private OutputWriter writer;
    /** buffer which will be used for the next line */
    private StringBuffer buffer = new StringBuffer (1000);
    /** have we printed any lines yet? used to prevent initial blank line */
    private boolean hadFirst = false;
    private HyperlinkFactory factory = new HyperlinkFactory ();

    public OutputWriterOutputStream (OutputWriter writer) {
        this.writer = writer;
    }

    public void close() throws IOException {
        flush ();
        writer.close();
    }

    public void flush() throws IOException {
        flushLines ();
        if (buffer.length () > 0) {
            // No chance to hyperlink individual lines at this point:
            writer.print (buffer.toString ());
            buffer.setLength (0);
        }
        writer.flush();
    }

    public void write(byte[] b) throws IOException {
        write (b, 0, b.length);
    }

    public void write(byte[] b, int offset, int length) throws IOException {
        buffer.append (new String (b, offset, length));
        // Will usually contain at least one newline:
        flushLines ();
    }

    public void write(int b) throws IOException {
        buffer.append ((char) b);
        if ((char) b == '\n') {
            flushLines ();
        }
    }

    private void flushLines () throws IOException {
        // Not as efficient as it could be, surely, but keep it simple for now:
        //System.err.println("flushLines: buffer=" + buffer);
    MAIN:
        while (true) {
            int len = buffer.length ();
            for (int i = 0; i < len; i++) {
                if (buffer.charAt (i) == '\n') {
                    //System.err.println("flushing; i=" + i);
                    // For Windows:
                    int end = i;
                    if (end > 0 && buffer.charAt (end - 1) == '\r') {
                        end--;
                    }
                    flushLine (buffer.substring (0, end));
                    buffer.delete (0, i + 1);
                    continue MAIN;
                }
            }
            //System.err.println("not found");
            break MAIN;
        }
    }

    private void flushLine (String l) throws IOException {
        //System.err.println("flushing: " + l);
        if (! hadFirst) {
            hadFirst = true;
            // Do not print an initial blank line.
            if (l.trim ().length () == 0) {
                return;
            }
        }
        Hyperlink link = factory.create (l);
        if (link != null) {
            writer.println (link.getMessage (), link);
        } else {
            writer.println (l);
        }
    }

    /** Represents a linkable line (appears in red in Output Window...).
     * Its functionality could be replaced in the future by ErrorEvent's from the Compiler API
     * (though something like the HyperlinkFactory would have to remain).
     * This would also handle the proper error highlighting and so on.
     */
    private static final class Hyperlink implements OutputListener {
        private FileObject file; // file to jump to
        private int line1, col1, line2, col2; // line/col number to jump to, 0-based, -1 for unspecified
        private String message; // message it is saying, or null
        Hyperlink (FileObject file, int line1, int col1, int line2, int col2, String message) {
            this.file = file;
            this.line1 = line1;
            this.col1 = col1;
            this.line2 = line2;
            this.col2 = col2;
            this.message = message;
        }
        public String getMessage () {
            String fname = file.getPackageNameExt ('/', '.');
            String m = (message != null ? message : NbBundle.getMessage (OutputWriterOutputStream.class, "ERR_unknown"));
            if (line1 == -1) {
                return NbBundle.getMessage
                    (OutputWriterOutputStream.class, "MSG_err", fname, m); // NOI18N
            } else {
                if (col1 == -1) {
                    return NbBundle.getMessage
                        (OutputWriterOutputStream.class, "MSG_err_line", fname, m, // NOI18N
                         new Integer (line1 + 1));
                } else {
                    if (line2 == -1 || col2 == -1 || (line1 == line2 && col1 == col2)) {
                        return NbBundle.getMessage
                            (OutputWriterOutputStream.class, "MSG_err_line_col", // NOI18N
                             new Object[] { fname, m, new Integer (line1 + 1), new Integer (col1 + 1) });
                    } else {
                        if (line1 == line2) {
                            return NbBundle.getMessage
                                (OutputWriterOutputStream.class, "MSG_err_line_col_col", // NOI18N
                                 new Object[] { fname, m, new Integer (line1 + 1), new Integer (col1 + 1), new Integer (col2 + 1) });
                        } else {
                            return NbBundle.getMessage
                                (OutputWriterOutputStream.class, "MSG_err_line_col_line_col", // NOI18N
                                 new Object[] { fname, m, new Integer (line1 + 1), new Integer (col1 + 1), new Integer (line2 + 1), new Integer (col2 + 1) });
                        }
                    }
                }
            }
        }
        public void outputLineAction (OutputEvent ev) {
            if (message != null) {
                TopManager.getDefault ().setStatusText (message);
            }
            try {
                DataObject dob = DataObject.find (file);
                EditorCookie ed = (EditorCookie) dob.getCookie (EditorCookie.class);
                if (ed != null) {
                    if (line1 == -1) {
                        // OK, just open it.
                        ed.open ();
                    } else {
                        Line l = ed.getLineSet ().getOriginal (line1);
                        l.markError ();
                        if (col1 == -1) {
                            l.show (Line.SHOW_GOTO);
                        } else {
                            l.show (Line.SHOW_GOTO, col1);
                        }
                    }
                }
            } catch (DataObjectNotFoundException donfe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, donfe);
            } catch (IndexOutOfBoundsException iobe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, iobe);
            }
        }
        public void outputLineSelected (OutputEvent ev) {
            try {
                DataObject dob = DataObject.find (file);
                EditorCookie ed = (EditorCookie) dob.getCookie (EditorCookie.class);
                if (ed != null) {
                    if (line1 != -1) {
                        Line l = ed.getLineSet ().getOriginal (line1);
                        l.markError ();
                        if (col1 == -1) {
                            l.show (Line.SHOW_TRY_SHOW);
                        } else {
                            l.show (Line.SHOW_TRY_SHOW, col1);
                        }
                    }
                }
            } catch (DataObjectNotFoundException donfe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, donfe);
            } catch (IndexOutOfBoundsException iobe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, iobe);
            }
        }
        public void outputLineCleared (OutputEvent ev) {
            try {
                DataObject dob = DataObject.find (file);
                EditorCookie ed = (EditorCookie) dob.getCookie (EditorCookie.class);
                if (ed != null) {
                    if (line1 != -1) {
                        Line l = ed.getLineSet ().getOriginal (line1);
                        l.unmarkError ();
                    }
                }
            } catch (DataObjectNotFoundException donfe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, donfe);
            } catch (IndexOutOfBoundsException iobe) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, iobe);
            }
        }
        public String toString () {
            return "Hyperlink[" + file + ":" + line1 + ":" + col1 + ":" + line2 + ":" + col2 + "]"; // NOI18N
        }
    }

    private static final class HyperlinkFactory implements Comparator {
        /** used only in constructor */
        private Map fss0; // Map<String,FileSystem>
        /** list of filesystem prefixes, akin to a classpath, mapped to filesystems */
        private SortedMap fss; // SortedMap<String,FileSystem>
        public HyperlinkFactory () {
            fss0 = new HashMap ();
            fss = new TreeMap (this);
            Repository repo = TopManager.getDefault ().getRepository ();
            Enumeration e = repo.getFileSystems ();
            while (e.hasMoreElements ()) {
                FileSystem fs = (FileSystem) e.nextElement ();
                FileObject root = fs.getRoot ();
                File path = NbClassPath.toFile (root);
                if (path == null) {
                    continue;
                }
                String prefix = path.getAbsolutePath ();
                if (! prefix.endsWith (File.separator)) {
                    prefix += File.separator;
                }
                if (path != null) {
                    fss0.put (prefix, fs);
                    fss.put (prefix, fs);
                    //debug ("prefix=" + prefix + " fs=" + fs); // NOI18N
                }
            }
            //debug ("fss0=" + fss0); // NOI18N
            fss0 = null;
            //debug ("fss=" + fss); // NOI18N
        }
        /** returns a hyperlink for the line, or null if not apropos */
        public Hyperlink create (String line) {
            // Look through the list of filesystem prefixes in order...
            Iterator it = fss.entrySet ().iterator ();
            while (it.hasNext ()) {
                Map.Entry entry = (Map.Entry) it.next ();
                String prefix = (String) entry.getKey ();
                if (line.startsWith (prefix) ||
                    // #11513 - Jikes may use wrong sep on Win
                    line.replace ('/', File.separatorChar).startsWith (prefix)) {
                    int colon1 = line.indexOf (':', prefix.length ());
                    if (colon1 != -1) {
                        String filename = line.substring (prefix.length (), colon1)
                            .replace (File.separatorChar, '/');
                        FileSystem fs = (FileSystem) entry.getValue ();
                        FileObject fo = fs.findResource (filename);
                        if (fo != null) {
                            int line1 = -1, col1 = -1, line2 = -1, col2 = -1;
                            int start = colon1 + 1; // start of message
                            int colon2 = line.indexOf (':', colon1 + 1);
                            if (colon2 != -1) {
                                try {
                                    line1 = Integer.parseInt (line.substring (colon1 + 1, colon2).trim ()) - 1;
                                    start = colon2 + 1;
                                    int colon3 = line.indexOf (':', colon2 + 1);
                                    if (colon3 != -1) {
                                        col1 = Integer.parseInt (line.substring (colon2 + 1, colon3).trim ()) - 1;
                                        start = colon3 + 1;
                                        int colon4 = line.indexOf (':', colon3 + 1);
                                        if (colon4 != -1) {
                                            line2 = Integer.parseInt (line.substring (colon3 + 1, colon4).trim ()) - 1;
                                            start = colon4 + 1;
                                            int colon5 = line.indexOf (':', colon4 + 1);
                                            if (colon5 != -1) {
                                                col2 = Integer.parseInt (line.substring (colon4 + 1, colon5).trim ()) - 1;
                                                start = colon5 + 1;
                                            }
                                        }
                                    }
                                } catch (NumberFormatException nfe) {
                                    // Fine, rest is part of the message.
                                }
                            }
                            String message = line.substring (start).trim ();
                            if (message.length () == 0) {
                                message = null;
                            }
                            Hyperlink h = new Hyperlink (fo, line1, col1, line2, col2, message);
                            //debug ("Got " + h + " for: " + line); // NOI18N
                            return h;
                        } else {
                            //debug ("no fo for: " + line + " fs=" + fs.getDisplayName () + " prefix=" + prefix + " filename=" + filename); // NOI18N
                        }
                    } else {
                        //debug ("colon1 not found for: " + line); // NOI18N
                    }
                }
            }
            //debug ("nothing for: " + line); // NOI18N
            return null;
        }
        /*
      private static PrintWriter debugwriter = null;
      private static void debug (String s) {
          if (debugwriter == null) {
              try {
                  debugwriter = new PrintWriter (new FileWriter ("/space/nb/dev-dec-12/Development/testme/owos.log"));
              } catch (IOException ioe) {
                  TopManager.getDefault ().notifyException (ioe);
                  return;
              }
          }
          debugwriter.println("[owos] " + s);
          debugwriter.flush ();
      }
       */
        /** compare prefixes */
        public int compare (Object o1, Object o2) {
            FileSystem f1 = (FileSystem) fss0.get (o1);
            FileSystem f2 = (FileSystem) fss0.get (o2);
            boolean c1 = f1.getCapability ().capableOf (FileSystemCapability.COMPILE);
            boolean c2 = f2.getCapability ().capableOf (FileSystemCapability.COMPILE);
            //debug ("compare: o1=" + o1 + " o2=" + o2 + " f1=" + f1 + " f2=" + f2 + " c1=" + c1 + " c2=" + c2); // NOI18N
            if (c1 && !c2) {
                // Only first is for compilation, so it gets preference (earlier in list)
                //debug ("-1"); // NOI18N
                return -1;
            } else if (!c1 && c2) {
                // Only second, so it gets preference (earlier in list)
                //debug ("+1"); // NOI18N
                return 1;
            } else {
                // The same; compare length of prefixes. Longer prefixes
                // are more specific, thus preference (earlier in list).
                String p1 = (String) o1;
                String p2 = (String) o2;
                int comp = p2.length () - p1.length ();
                //debug ("comp=" + comp); // NOI18N
                if (comp != 0) return comp;
                return System.identityHashCode (f1) - System.identityHashCode (f2);
            }
        }
    }

}
