/*
 *                         Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the Ant module
 * The Initial Developer of the Original Code is Jayme C. Edwards.
 * Portions created by Jayme C. Edwards are Copyright (c) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Jesse Glick.
 */
 
package org.apache.tools.ant.module.xml;

import java.io.IOException;
import java.util.Map;

import org.w3c.dom.Element;

import org.openide.TopManager;
import org.openide.cookies.InstanceCookie;
import org.openide.nodes.Node;

import org.apache.tools.ant.module.AntSettings;
import org.apache.tools.ant.module.api.ElementCookie;
import org.apache.tools.ant.module.api.IntrospectedInfo;

public class ElementSupport implements ElementCookie {

    private final Element el;

    public ElementSupport (Element el) {
        this.el = el;
    }

    public Element getElement () {
        return el;
    }

    /** More sophisticated support that also implements InstanceCookie.
     * The associated class represents the real Ant data type being represented
     * by the XML. For example, <project> nodes would have Project.class;
     * then Target.class for targets; then e.g. Jar.class for a task;
     * then data types such as FileSet.class. All this information can be used
     * to help make the nodes easier to edit using IntrospectedInfo.
     */
    public static class Instance extends ElementSupport implements InstanceCookie {
        
        private Class clazz;
        private String clazzname;
        private ClassNotFoundException cnfe = null;
        private Map subels = null;
        
        /** Specify a class in advance.
         */
        public Instance (Element el, Class clazz) {
            super (el);
            this.clazz = clazz;
            this.clazzname = clazz.getName ();
        }

        /** Specify a class in advance by name.
         */
        public Instance (Element el, String clazzname) {
            super (el);
            this.clazzname = clazzname;
        }
        
        public String toString () {
            return "ElementSupport.Instance[clazzname=" + clazzname + ",subels=" + subels + "]"; // NOI18N
        }

        public String instanceName () {
            return this.getElement ().getNodeName ();
        }
        
        /** Convenience method for ElementChildren giving defined subelements.
         * Its advantage over moving this code into ElementChildren is that it will
         * still work even if the parent class cannot be loaded, since the class name
         * is known and is sufficient to look it up in the table.
         */
        public Map getSubelements () throws IOException, ClassNotFoundException {
            if (subels != null) return subels;
            synchronized (this) {
                if (subels != null) return subels;
                if (clazzname == null) instanceClass ();
                if (IntrospectedInfo.getDefaults ().isKnown (clazzname)) {
                    subels = IntrospectedInfo.getDefaults ().getElements (clazzname);
                } else if (AntSettings.getDefault ().getCustomDefs ().isKnown (clazzname)) {
                    subels = AntSettings.getDefault ().getCustomDefs ().getElements (clazzname);
                } else {
                    subels = null;
                }
                return subels;
            }
        }
        
        public Class instanceClass () throws IOException, ClassNotFoundException {
            if (clazz != null) return clazz;
            if (cnfe != null) throw cnfe;
            synchronized (this) {
                if (clazz != null) return clazz;
                if (cnfe != null) throw cnfe;
                if (clazzname == null) throw new IOException ("Must specify clazz or clazzname"); // NOI18N
                try {
                    return (clazz = TopManager.getDefault ().currentClassLoader ().loadClass (clazzname));
                } catch (ClassNotFoundException cnfe) {
                    this.cnfe = cnfe;
                    throw cnfe;
                }
            }
        }
        
        public Object instanceCreate () throws IOException, ClassNotFoundException {
            // In the future, could use IntrospectionHelper to traverse project build
            // tree and actually create a valid object!
            throw new IOException ("UNIMPLEMENTED"); // NOI18N
        }
        
    }
    
}
