/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport;

import java.awt.Image;
import java.beans.*;
import java.io.*;
import java.util.*;

import org.openide.ServiceType;
import org.openide.TopManager;
import org.openide.execution.ThreadExecutor;
import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem;
import org.openide.loaders.XMLDataObject;
import org.openide.modules.ModuleInstall;
import org.openide.util.RequestProcessor;
import org.openide.util.SharedClassObject;

public class APIModule extends ModuleInstall {

    private static final long serialVersionUID = -8817465769852478093L;

    private static final String PROP_INSTALL_COUNT = "installCount";
    
    private static final String LAYER_DTD = "-//NetBeans//DTD Filesystem 1.0//EN";
    private static final String JAVAHELP_DTD = "-//Sun Microsystems Inc.//DTD JavaHelp HelpSet Version 1.0//EN";

    public void installed () {
        restored ();
    }
    
    private static FileSystem createJarFS (File zipfile) throws Exception {
        JarFileSystem fs = new GlobalJarFileSystem ();
        fs.setJarFile (zipfile);
        fs.setHidden (true);
        FileSystemCapability capab = fs.getCapability ();
        if (capab instanceof FileSystemCapability.Bean) {
            FileSystemCapability.Bean bean = (FileSystemCapability.Bean) capab;
            bean.setCompile (false);
            bean.setExecute (false);
            bean.setDebug (false);
            bean.setDoc (true);
        } else {
            System.err.println ("Warning: JarFileSystem had strange capability: " + capab);
        }
        return fs;
    }

    public void restored () {
        System.setProperty ("netbeans.module.test", "true");
        // Mount docs, or remount if project was discarded:
        Integer count = (Integer) getProperty (PROP_INSTALL_COUNT);
        int icount = count == null ? 1 : count.intValue () + 1;
        //if (Boolean.getBoolean ("netbeans.debug.exceptions"))
        //    System.err.println ("APIModule: icount=" + icount);
        putProperty (PROP_INSTALL_COUNT, new Integer (icount));
        // 1: first install (project is discarded anyway)
        // 2: first restore as actual user
        // 3: next restore (project settings incl. Repository loaded)
        if (icount <= 2) {
            final File f = findAPIDocs ();
            if (f != null) {
                try {
                    final FileSystem fs = createJarFS (f);
                    // Mount docs in Documentation Repository:
                    final Repository repo = TopManager.getDefault ().getRepository ();
                    if (repo.findFileSystem (fs.getSystemName ()) == null) {
			repo.addFileSystem (fs);
                    } else {
                        System.err.println ("Note: OpenAPIs.zip was already present in Filesystems.");
		    }
                } catch (Exception e) {
                    if (Boolean.getBoolean ("netbeans.debug.exceptions"))
                        e.printStackTrace ();
                }
            } else {
                System.err.println("Note: OpenAPIs.zip not found to add to Javadoc, ignoring...");
            }
            /* Replaced with XML layer:
            // Now mount openide*.jar so that people can actually compile...
            mountLib ("openide.jar");
            mountLib ("openide-fs.jar");
            mountLib ("openide-nodes.jar");
            mountLib ("openide-util.jar");
            // [PENDING] also mount regexp stuff? XML?
             */
        }
        XMLDataObject.Info info = new XMLDataObject.Info ();
        //info.setIconBase ("/org/netbeans/modules/apisupport/resources/LayerDataNode");
        info.addProcessorClass (LayerDataNode.class);
        XMLDataObject.registerInfo (LAYER_DTD, info);
        // catalog entry already registered by XMLFileSystem
        info = new XMLDataObject.Info ();
        info.setIconBase ("/org/netbeans/modules/apisupport/resources/JavaHelpDataIcon"); // NOI18N
        info.addProcessorClass (JavaHelpProcessor.class);
        XMLDataObject.registerInfo (JAVAHELP_DTD, info);
        // catalog entry registered by JavaHelpDataLoader
    }
    
    /*
    private void mountLib (String libname) {
        String nbhome = System.getProperty ("netbeans.home");
        if (nbhome == null) {
            System.err.println("Note: -Dnetbeans.home not set");
            return;
        }
        File lib = new File (new File (nbhome, "lib"), libname);
        if (! lib.isFile ()) {
            System.err.println("Library not found to mount: " + lib);
            return;
        }
        try {
            JarFileSystem fs = new GlobalJarFileSystem ();
            fs.setJarFile (lib);
            fs.setHidden (true);
            Repository repo = TopManager.getDefault ().getRepository ();
            if (repo.findFileSystem (fs.getSystemName ()) == null) {
                repo.addFileSystem (fs);
            } else {
                System.err.println ("Note: " + libname + " was already present in Filesystems.");
            }
        } catch (Exception e) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions"))
                e.printStackTrace ();
        }
    }
     */

    public void uninstalled () {
        // Unmount docs (AutoUpdate should handle actually removing the file):
        Repository repo = TopManager.getDefault ().getRepository ();
        Enumeration e = repo.fileSystems ();
        while (e.hasMoreElements ()) {
            Object o = e.nextElement ();
            //System.err.println("repo elt: " + o);
            if (o instanceof GlobalJarFileSystem) {
                repo.removeFileSystem ((FileSystem) o);
            }
        }
        XMLDataObject.registerInfo (LAYER_DTD, null);
        XMLDataObject.registerInfo (JAVAHELP_DTD, null);
    }

    static File findAPIDocs () {
        try {
            String suffix = "docs" + File.separator + "OpenAPIs.zip";
            String user = System.getProperty ("netbeans.user");
            if (user != null) {
                File f = new File (user, suffix);
                if (f.exists ()) return f.getCanonicalFile ();
            }
            String home = System.getProperty ("netbeans.home");
            if (home != null) {
                File f = new File (home, suffix);
                if (f.exists ()) return f.getCanonicalFile ();
            }
        } catch (IOException ioe) {
            TopManager.getDefault ().notifyException (ioe);
        }
        return null;
    }

    public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal (in);
        putProperty (PROP_INSTALL_COUNT, in.readObject ());
    }

    public void writeExternal (ObjectOutput out) throws IOException {
        super.writeExternal (out);
        out.writeObject (getProperty (PROP_INSTALL_COUNT));
    }

    // Static attribute-getting methods called from mf-layer.xml:

    public static ServiceType.Handle lookupThreadExecutor () {
        // Core no longer provides ThreadExecutor, rather Java provides JavaThreadExecutor,
        // which cannot be easily looked up -> heck with it.
        return new ServiceType.Handle (new ThreadExecutor ());
    }
    public static ServiceType.Handle lookupSystemActionTester () {
        return lookupService (SystemActionTester.class);
    }
    public static ServiceType.Handle lookupPresenterTester () {
        return lookupService (PresenterTester.class);
    }
    public static ServiceType.Handle lookupCompilerTypeTester () {
        return lookupService (CompilerTypeTester.class);
    }
    public static ServiceType.Handle lookupBeanTester () {
        return lookupService (BeanTester.class);
    }
    public static ServiceType.Handle lookupExecutorTester () {
        return lookupService (ExecutorTester.class);
    }

    private static ServiceType.Handle lookupService (Class clazz) {
        ServiceType service = TopManager.getDefault ().getServices ().find (clazz);
        if (service == null) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) {
                System.err.println ("Warning: services registry did not contain " + clazz.getName ()); // NOI18N
            }
            return null;
        } else {
            return new ServiceType.Handle (service);
        }
    }

    /** Exists only for the sake of its bean info. */
    public static final class GlobalJarFileSystem extends JarFileSystem {
        private static final long serialVersionUID = -2165158869303960139L;
    }
    /** Marks it as global (not project-specific). */
    public static final class GlobalJarFileSystemBeanInfo extends SimpleBeanInfo {
        public BeanDescriptor getBeanDescriptor () {
            BeanDescriptor bd = new BeanDescriptor (GlobalJarFileSystem.class);
            bd.setValue ("global", Boolean.TRUE); // NOI18N
            return bd;
        }
        public BeanInfo[] getAdditionalBeanInfo () {
            try {
                return new BeanInfo[] { Introspector.getBeanInfo (JarFileSystem.class) };
            } catch (IntrospectionException ie) {
                TopManager.getDefault ().notifyException (ie);
                return null;
            }
        }
        public Image getIcon (int kind) {
            try {
                return Introspector.getBeanInfo (JarFileSystem.class).getIcon (kind);
            } catch (IntrospectionException ie) {
                TopManager.getDefault ().notifyException (ie);
                return null;
            }
        }
    }

}
