/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport;

import java.io.IOException;
import java.lang.reflect.*;
import java.text.Format;
import java.util.*;

import org.openide.*;
import org.openide.actions.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.util.MapFormat;
import org.openide.util.SharedClassObject;
import org.openide.util.actions.SystemAction;

// [PENDING] this whole class would probably fare better as an XML processor
// plus one custom template iterator directly assigned to the docs templates...

public class JavaHelpDataLoader extends UniFileLoader {
    
    private static boolean haveXmlModule = false;
    static {
        try {
            Class.forName ("org.netbeans.modules.xml.XMLDataLoader",
                           true,
                           TopManager.getDefault ().systemClassLoader ());
            haveXmlModule = true;
        } catch (ClassNotFoundException cnfe) {
            // OK.
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    private static final long serialVersionUID =-7728960313890246535L;
    public JavaHelpDataLoader() {
        // Just has to be superclass, does not need to be exact:
        this (XMLDataObject.class);
    }
    /*
    private static Class getDataObjectClass () {
        if (haveXmlModule) {
            try {
                return Class.forName ("org.netbeans.modules.xml.XMLDataObject",
                                      true,
                                      TopManager.getDefault ().systemClassLoader ());
            } catch (ThreadDeath td) {
                throw td;
            } catch (Throwable t) {
                TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, t);
                // fallthrough
            }
        }
        return XMLDataObject.class; // org.openide.loaders
    }
     */

    public JavaHelpDataLoader(Class recognizedObject) {
        super (recognizedObject);
    }

    // List of XML file types in use. Each is extension, then public ID, then resource path to DTD.
    private static final String[][] types = {
        { "hs",  "-//Sun Microsystems Inc.//DTD JavaHelp HelpSet Version 1.0//EN", "org/netbeans/modules/apisupport/resources/helpset_1_0.dtd" },
        { "jhm", "-//Sun Microsystems Inc.//DTD JavaHelp Map Version 1.0//EN",     "org/netbeans/modules/apisupport/resources/map_1_0.dtd" },
        { "toc", "-//Sun Microsystems Inc.//DTD JavaHelp TOC Version 1.0//EN",     "org/netbeans/modules/apisupport/resources/toc_1_0.dtd" },
        { "idx", "-//Sun Microsystems Inc.//DTD JavaHelp Index Version 1.0//EN",   "org/netbeans/modules/apisupport/resources/index_1_0.dtd" }
    };
    
    protected String defaultDisplayName () {
        return "JavaHelp Files";
    }

    protected void initialize () {

        super.initialize ();

        ExtensionList extensions = new ExtensionList ();
        for (int i = 0; i < types.length; i++)
            extensions.addExtension (types[i][0]);
        setExtensions (extensions);

        ClassLoader loader = JavaHelpDataLoader.class.getClassLoader ();
        for (int i = 0; i < types.length; i++)
            // org.openide.loaders is where the API is anyway:
            XMLDataObject.registerCatalogEntry (types[i][1], types[i][2], loader);
        
    }
    
    protected SystemAction[] defaultActions () {

        // Optionally:
        if (haveXmlModule) {
            try {
                Class clazz = Class.forName ("org.netbeans.modules.xml.XMLDataLoader",
                                             true,
                                             TopManager.getDefault ().systemClassLoader ());
                DataLoader loader = (DataLoader) SharedClassObject.findObject (clazz, true);
                return loader.getActions ();
            } catch (ThreadDeath td) {
                throw td;
            } catch (Throwable t) {
                TopManager.getDefault ().notifyException (t);
                // fine, forget it
            }
        }

        return new SystemAction[] {
                       SystemAction.get(OpenAction.class),
                       SystemAction.get(FileSystemAction.class),
                       null,
                       SystemAction.get (ExecuteAction.class),
                       null,
                       SystemAction.get(CutAction.class),
                       SystemAction.get(CopyAction.class),
                       SystemAction.get(PasteAction.class),
                       null,
                       SystemAction.get(DeleteAction.class),
                       SystemAction.get(RenameAction.class),
                       null,
                       SystemAction.get(SaveAsTemplateAction.class),
                       null,
                       SystemAction.get(ToolsAction.class),
                       SystemAction.get(PropertiesAction.class)
                   };

    }

    protected MultiDataObject createMultiObject (FileObject primaryFile)
    throws DataObjectExistsException, IOException {
        if (haveXmlModule) {
            try {
                Class clazz = Class.forName ("org.netbeans.modules.xml.XMLDataObject",
                                             true,
                                             TopManager.getDefault ().systemClassLoader ());
                Constructor c = clazz.getConstructor (new Class[] { FileObject.class, MultiFileLoader.class });
                return (MultiDataObject) c.newInstance (new Object[] { primaryFile, this });
            } catch (InvocationTargetException ite) {
                Throwable itet = ite.getTargetException ();
                if (itet instanceof DataObjectExistsException)
                    throw (DataObjectExistsException) itet;
                else if (itet instanceof IOException)
                    throw (IOException) itet;
                else
                    TopManager.getDefault ().notifyException (itet);
                    // fallthrough
            } catch (ThreadDeath td) {
                throw td;
            } catch (Throwable t) {
                TopManager.getDefault ().notifyException (t);
                // fallthrough
            }
        }
        return new XMLDataObject (primaryFile, this); // org.openide.loaders
    }

    protected MultiDataObject.Entry createPrimaryEntry (MultiDataObject mdo, FileObject fo) {
        return new JavaHelpEntry (mdo, fo);
    }

    static class JavaHelpEntry extends FileEntry.Format {

        private static final long serialVersionUID =2248161434771748311L;
        public JavaHelpEntry (MultiDataObject mdo, FileObject fo) {
            super (mdo, fo);
        }

        protected java.text.Format createFormat (FileObject target, String name, String ext) {
            String homeID = target.isRoot () ? "HOMEID" : target.getPackageName ('.') + ".HOMEID";
            Map m = new HashMap ();
            if ("hs".equals (ext)) {
                m.put ("LIKELY_HOME_ID", homeID);
                String base;
                if (name.endsWith ("HelpSet"))
                    base = name.substring (0, name.length () - 7) + "HelpMap";
                else
                    base = "Map";
                m.put ("LIKELY_MAP_URL_QUOTED", "\"" + base + ".jhm\"");
                if (name.endsWith ("HelpSet"))
                    base = name.substring (0, name.length () - 7) + "HelpContents";
                else
                    base = "Contents";
                m.put ("LIKELY_TOC_URL", base + ".toc");
                if (name.endsWith ("HelpSet"))
                    base = name.substring (0, name.length () - 7) + "HelpIndex";
                else
                    base = "Index";
                m.put ("LIKELY_INDEX_URL", base + ".idx");
            } else if ("jhm".equals (ext)) {
                m.put ("LIKELY_HOME_ID_QUOTED", "\"" + homeID + "\"");
                String base;
                if (name.endsWith ("HelpMap"))
                    base = name.substring (0, name.length () - 7) + "HelpPage";
                else
                    base = "index";
                m.put ("LIKELY_HTML_URL_QUOTED", "\"" + base + ".html\"");
            }
            MapFormat format = new MapFormat (m);
            format.setLeftBrace ("__");
            format.setRightBrace ("__");
            return format;
        }

    }

}
