/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.*;
import java.io.PrintWriter;
import java.net.URL;
import java.util.*;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.openide.TopManager;
import org.openide.actions.OpenAction;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.*;
import org.openide.util.actions.SystemAction;

public class LayerDataNode extends DataNode {
    
    private boolean parseError = false;

    public LayerDataNode (XMLDataObject obj) {
        super (obj, new LayerChildren (obj));
        ((LayerChildren) getChildren ()).setParentNode (this);
        setIconBase ("/org/netbeans/modules/apisupport/resources/LayerDataIcon");
    }

    public SystemAction getDefaultAction () {
	return SystemAction.get (OpenAction.class);
    }
    
    public HelpCtx getHelpCtx () {
        return new HelpCtx ("org.netbeans.modules.apisupport.layers");
    }
    
    private void setParseError (boolean pe) {
        boolean old = parseError;
        parseError = pe;
        if (! old && pe) {
            setIconBase ("/org/netbeans/modules/apisupport/resources/LayerDataIconError");
        } else if (old && ! pe) {
            setIconBase ("/org/netbeans/modules/apisupport/resources/LayerDataIcon");
        }
    }
    
    private static PrintWriter pw = null;
    private static synchronized PrintWriter getErr () {
        if (pw == null) {
            pw = TopManager.getDefault ().getIO ("XML Layers", true).getErr ();
        }
        return pw;
    }

    // key type: FileSystem
    private static class LayerChildren extends Children.Array implements FileChangeListener, PropertyChangeListener {
        
        private DataObject obj;
        private FileObject xml;
        private LayerDataNode parent;
        private Node subnode;
        private AnnotatingFileSystem fs;
        
        public LayerChildren (XMLDataObject obj) {
            this.obj = obj;
            xml = obj.getPrimaryFile ();
        }
        
        void setParentNode (LayerDataNode node) {
            parent = node;
        }
        
        protected void addNotify () {
            if (fs == null) {
                fs = new AnnotatingFileSystem ();
                try {
                    fs.setURL (xml.getURL ());
                    parent.setParseError (false);
                } catch (Exception e) {
                    handleErr (e);
                }
                obj.addPropertyChangeListener (WeakListener.propertyChange (this, obj));
                xml.addFileChangeListener (WeakListener.fileChange (this, xml));
            }
            if (subnode == null) {
                try {
                    subnode = DataObject.find (fs.getRoot ()).getNodeDelegate ();
                } catch (DataObjectNotFoundException donfe) {
                    TopManager.getDefault ().notifyException (donfe);
                    return;
                }
            }
            add (new Node[] { subnode });
        }
        
        protected void removeNotify () {
            if (subnode != null) {
                remove (new Node[] { subnode });
                subnode = null;
            }
        }
        
        // FileChangeListener:
        
        public void fileChanged (FileEvent fe) {
            try {
                fs.setURL (xml.getURL ());
                parent.setParseError (false);
            } catch (Exception e) {
                handleErr (e);
            }
        }

        public void fileFolderCreated (FileEvent fe) {}
        public void fileDataCreated (FileEvent fe) {}
        public void fileDeleted (FileEvent fe) {}
        public void fileRenamed (FileRenameEvent fe) {}
        public void fileAttributeChanged (FileAttributeEvent fe) {}
        
        public void propertyChange (PropertyChangeEvent ev) {
            if (DataObject.PROP_PRIMARY_FILE.equals (ev.getPropertyName ())) { // #9253
                xml = obj.getPrimaryFile ();
                xml.addFileChangeListener (WeakListener.fileChange (this, xml));
                fileChanged (null);
            }
        }
        
        private void handleErr (Exception e) {
            parent.setParseError (true);
            if (e instanceof SAXParseException) {
                SAXParseException saxpe = (SAXParseException) e;
                getErr ().println ("XML parse error [" + saxpe.getLineNumber () + ":" + saxpe.getColumnNumber () + "]: " + saxpe.getMessage ());
            } else if (e instanceof SAXException) {
                getErr ().println ("XML parse error: " + e.getMessage ());
            } else {
                e.printStackTrace (getErr ());
            }
        }
        
    }

    private static class AnnotatingFileSystem extends MultiFileSystem implements FileSystem.Status {
        
        public AnnotatingFileSystem () {
            // Default constructor should suffice; this works around a bug
            // in MultiFileSystem:
            super (new FileSystem[] { });
        }
        
        void setURL (URL u) throws Exception {
            try {
                setDelegates (new FileSystem[] { new XMLFileSystem (u) });
            } catch (Exception e) {
                setDelegates (new FileSystem[] { });
                throw e;
            }
            fireFileStatusChanged (new FileStatusEvent (this, true, true));
        }
        
        public FileSystem.Status getStatus () {
            return this;
        }
        
        // Largely copied from org.netbeans.core.projects.SystemFileSystem:

        public String annotateName (String base, Set files) {
            Iterator it = files.iterator ();
            while (it.hasNext ()) {
                FileObject fo = (FileObject) it.next ();
                String bundleName = (String) fo.getAttribute ("SystemFileSystem.localizingBundle");
                if (bundleName != null) {
                    try {
                        ResourceBundle b = NbBundle.getBundle (bundleName, Locale.getDefault (), TopManager.getDefault ().currentClassLoader ());
                        try {
                            return b.getString (fo.getPackageNameExt ('/', '.'));
                        } catch (MissingResourceException ex) {
                            // ignore--normal
                        }
                    } catch (MissingResourceException ex) {
                        getErr ().println ("Note: " + bundleName + " does not yet exist to localize file " + fo);
                        // ignore
                    }
                }
            }
            if (files.size () == 1 && ((FileObject) files.iterator ().next ()).isRoot ()) {
                return "<Root Folder>";
            }
            return base;
        }
        
        public Image annotateIcon (Image base, int type, Set files) {
            String attr;
            if (type == BeanInfo.ICON_COLOR_16x16) {
                attr = "SystemFileSystem.icon";
            } else if (type == BeanInfo.ICON_COLOR_32x32) {
                attr = "SystemFileSystem.icon32";
            } else {
                return base;
            }
            Iterator it = files.iterator ();
            while (it.hasNext ()) {
                FileObject fo = (FileObject) it.next ();
                Object value = fo.getAttribute (attr);
                if (value != null) {
                    try {
                        return Toolkit.getDefaultToolkit ().getImage ((URL) value);
                    } catch (Exception e) {
                        e.printStackTrace (getErr ());
                    }
                }
            }
            return base;
        }
        
    }
    
}
