/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport;

import java.lang.reflect.Modifier;
import java.io.IOException;
import java.net.URL;
import java.util.*;
import javax.swing.SwingUtilities;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.cookies.SourceCookie;
import org.openide.cookies.InstanceCookie;
import org.openide.execution.NbfsURLConnection;
import org.openide.filesystems.*;
import org.openide.nodes.Node;
import org.openide.src.*;
import org.openide.util.*;
import org.openide.util.actions.*;
import org.openide.windows.TopComponent;

public class ShowAPIJavadocAction extends CookieAction {

    private static final long serialVersionUID =-2438079845105114834L;

    private static final ErrorManager err = TopManager.getDefault ().getErrorManager ().
        getInstance ("org.netbeans.modules.apisupport.ShowAPIJavadocAction");

    public String getName () {
        if (lastCalced == null) {
            return "Show API Javadoc";
        } else {
            String name = lastCalced;
            int idx = name.lastIndexOf ((int) '.');
            if (idx != -1) name = name.substring (idx + 1);
            return "Javadoc: " + name.replace ('$', '.');
        }
    }

    protected String iconResource () {
        return "/org/netbeans/modules/apisupport/resources/ShowAPIJavadocIcon.gif";
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx ("org.netbeans.modules.apisupport.utils");
    }

    protected Class[] cookieClasses () {
        return new Class[] { SourceCookie.class, ClassElement.class, MemberElement.class, InstanceCookie.class };
    }

    protected int mode () {
        return MODE_EXACTLY_ONE;
    }

    private static int sequence = 0;
    private static Node lastCalcedNode = null;
    private static String lastCalced = null;
    protected boolean enable (Node[] nodes) {
        if (! super.enable (nodes)) return false;
        final int currentSequence = ++sequence;
        err.log ("sequence=" + sequence);
        final Node n = nodes[0];
        RequestProcessor.postRequest (new Runnable () {
            public void run () {
                if (sequence > currentSequence) {
                    err.log ("Preempted before even doing check");
                    return;
                }
                String classname = findClass (n);
                if (sequence > currentSequence) {
                    err.log ("Ran out after checking");
                    return;
                }
                err.log ("setting enabled; currentSequence=" + currentSequence);
                setEnabled (classname != null);
                lastCalcedNode = n;
                lastCalced = classname;
                firePropertyChange ("name", null, null);
            }
        }, 2500);
        return false;
    }
    // 1.2 Javac bug:
    protected void firePropertyChange (String name, Object old, Object nue) {
        super.firePropertyChange (name, old, nue);
    }

    protected void performAction (Node[] nodes) {
        if (! super.enable (nodes)) {
            err.log (ErrorManager.WARNING, "performed when not enabled");
            return;
        }
        Node n = nodes[0];
        String classname;
        if (n == lastCalcedNode) {
            err.log ("using last-calculated");
            classname = lastCalced;
        } else {
            err.log ("recalcing");
            classname = findClass (n);
        }
        if (classname == null) {
            err.log (ErrorManager.WARNING, "enabled, but now cannot find it");
            return;
        }
        String resource = classname.replace ('.', '/').replace ('$', '.') + ".html";
        FileObject fo = FileSystemCapability.DOC.findResource (resource);
        if (fo == null) fo = FileSystemCapability.DOC.findResource ("api/" + resource);
        if (fo == null) {
            err.log (ErrorManager.USER, "Could not find docs for " + resource +
		   " in Documentation Repository!");
            return;
        }
        try {
            URL url = NbfsURLConnection.encodeFileObject (fo);
            TopManager.getDefault ().showUrl (url);
        } catch (FileStateInvalidException fsie) {
            err.notify (ErrorManager.INFORMATIONAL, fsie);
        }
    }

    private static String findClass (Node n) {
        ClassElement clazz = (ClassElement) n.getCookie (ClassElement.class);
        if (clazz == null) {
            err.log ("no ClassElement, trying MemberElement...");
            MemberElement mem = (MemberElement) n.getCookie (MemberElement.class);
            if (mem != null) {
                err.log ("got MemberElement");
                clazz = mem.getDeclaringClass ();
            }
        }
        if (clazz == null) {
            err.log ("neither, trying SourceCookie...");
            SourceCookie source = (SourceCookie) n.getCookie (SourceCookie.class);
            if (source == null) {
                err.log ("No SourceCookie, trying InstanceCookie");
                InstanceCookie inst = (InstanceCookie) n.getCookie (InstanceCookie.class);
                if (inst == null) {
                    err.log ("No InstanceCookie");
                    return null;
                }
                try {
                    clazz = ClassElement.forClass (inst.instanceClass ());
                    if (clazz == null) {
                        err.log ("Instance class not found");
                        return null;
                    }
                } catch (ClassNotFoundException cnfe) {
                    err.notify (ErrorManager.INFORMATIONAL, cnfe);
                    return null;
                } catch (IOException ioe) {
                    err.notify (ErrorManager.INFORMATIONAL, ioe);
                    return null;
                }
            } else {
                SourceElement element = source.getSource ();
                switch (element.getStatus ()) {
                case SourceElement.STATUS_ERROR:
                    // fallthrough
                case SourceElement.STATUS_PARTIAL:
                    err.log ("bad parse, ignoring");
                    return null;
                case SourceElement.STATUS_NOT:
                    /*
                    err.log ("partial parse, will finish and then try again", false);
                    element.prepare ().addTaskListener (new TaskListener () {
			    public void taskFinished (Task ignore) {
				SwingUtilities.invokeLater (new Runnable () {
					public void run () {
					    ShowAPIJavadocAction action = (ShowAPIJavadocAction) SystemAction.get (ShowAPIJavadocAction.class);
					    action.setEnabled (action.enable (TopComponent.getRegistry ().getActivatedNodes ()));
					}
				    });
			    }
			});
                    return null;
                    */
                    // fallthrough
                case SourceElement.STATUS_OK:
                    err.log ("full parse (or delayed)");
                    if (element.getClasses ().length == 0) {
                        err.log ("\tno classes found");
                        return null;
                    }
                    // [PENDING] ought to check specifically for main public one
                    clazz = element.getClasses ()[0];
                    break;
                default:
                    throw new InternalError ("illegal parsing state for SourceElement: " + element.getStatus ());
                }
            }
        }
        Type type = testCache (Type.createClass (clazz.getName ()));
        if (type == null) {
            err.log ("no found type");
            return null;
        }
        ClassElement clazz2 = ClassElement.forName (type.getFullString ());
        if (clazz2 == null) {
            err.log (ErrorManager.WARNING, "could not find clazz for " + type.getFullString ());
            return null;
        }
        SourceElement source = clazz2.getSource ();
        if (source == null) {
            err.log (ErrorManager.WARNING, "clazz " + clazz2.getName () + " had no source");
            return null;
        }
        String pkg = source.getPackage ().getFullName ();
        if (pkg == null) {
            err.log (ErrorManager.WARNING, "pkg==null");
            return null;
        }
        if (! pkg.equals ("")) pkg += '.';
        String classname = clazz2.getName ().getName ();
        while ((clazz2 = clazz2.getDeclaringClass ()) != null) {
            classname = clazz2.getName ().getName () + '$' + classname;
        }
        return pkg + classname;
    }

    // [PENDING] Using a weak map would seem to make sense, but for some reason it
    // seems that storing somekey -> null in the weak map "sticks" temporarily, but
    // is erased a few seconds later...even while regular keys stay around as expected.
    // This does not seem to have anything to do with garbage collection. Using
    // a regular hash map all works well, so this (hopefully minor) memory leak is
    // reasonable given the likely performance benefit.
    private static Map cache = Collections.synchronizedMap (new /*Weak*/HashMap ());
    private static Type testCache (Type type) {
        if (cache.containsKey (type)) {
            Type hit = (Type) cache.get (type);
            err.log ("Cache hit: " + type + " -> " + hit);
            return hit;
        }
        err.log ("Cache miss: " + type);
        cache.put (type, null);
        Type toRet = test (type);
        cache.put (type, toRet);
        err.log ("Cache add: " + type + " -> " + toRet);
        err.log ("\tfound: " + cache.containsKey (type) + " " + cache.get (type));
        return toRet;
    }

    private static Type test (Type type) {
        err.log ("testing: " + type);
        if (type == null) {
            err.log ("\twas null anyway");
            return null;
        }
        if (type.isArray ()) {
            err.log ("\ttrying element type");
            return testCache (type.getElementType ());
        }
        if (type.isPrimitive ()) {
            err.log ("\twas primitive");
            return null;
        }
        ClassElement clazz = ClassElement.forName (type.getFullString ());
        if (clazz == null) {
            // Due to a parser bug, this will actually happen sometimes,
            // though it shouldn't: method-local inner classes get here.
            err.log ("No such clazz");
            return null;
        }
        err.log ("Got clazz: " + clazz.getName ());
        if ((Modifier.isPublic (clazz.getModifiers ()) || Modifier.isProtected (clazz.getModifiers ()))
                && clazz.getName ().getFullName ().startsWith ("org.openide.")) {
            err.log ("\tsucceeded");
            return Type.createClass (clazz.getName ());
        } else {
            Identifier superid = clazz.getSuperclass ();
            err.log ("\ttrying on superclass " + superid);
            Type supe = superid == null ? null : testCache (Type.createClass (superid));
            if (supe != null) {
                err.log ("\tsucceeded with superclass");
                return supe;
            } else {
                err.log ("\ttrying with interfaces");
                Identifier[] xfaces = clazz.getInterfaces ();
                for (int i = 0; i < xfaces.length; i++) {
                    Type xface = testCache (Type.createClass (xfaces[i]));
                    if (xface != null) return xface;
                }
                err.log ("\tno luck with interfaces");
                return null;
            }
        }
    }

}
