/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport.lite;

import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Collection;
import java.util.Iterator;
import java.util.ResourceBundle;
import javax.swing.event.EventListenerList;

import org.openide.DialogDescriptor;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.WizardDescriptor;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

/** Implements the behavior of new module wizard
 *
 * @author Ales Kemr
 */
class Wizard extends Object implements WizardDescriptor.Iterator {

    private static final ResourceBundle bundle = NbBundle.getBundle( Wizard.class );

    /** Panels of the wizard */
    WizardPanel[] panels = new WizardPanel[] {
                                           new IntroWizardPanel(),
                                           new ModuleWizardPanel(),
                                           new SelectionWizardPanel(),
                                           new LoaderWizardPanel(),
                                           new MenuWizardPanel(),
                                           new ManifestWizardPanel()
                                     };

    /** Current panel */
    private int current = 0;

    private boolean canceled = false;

    /** The wizard descriptor */
    private WizardDescriptor wizardDescriptor;

    /** The dialog */
    Dialog dialog;

    /** Creates the wizard */
    Wizard() {


        // Create the wizard

        PropertyChangeListener listener = new PropertyChangeListener() {
                                              public void propertyChange(PropertyChangeEvent event) {
                                                  if (event.getPropertyName().equals(DialogDescriptor.PROP_VALUE)) {
                                                      Object option = event.getNewValue();

                                                      if (option == WizardDescriptor.FINISH_OPTION ||
                                                              option == NotifyDescriptor.CANCEL_OPTION ||
                                                              option == NotifyDescriptor.CLOSED_OPTION ) {
                                                          //panels[ current ].end( true );
                                                          if ( option != WizardDescriptor.FINISH_OPTION )
                                                              canceled = true;
                                                          getCurrent().end( true );
                                                          dialog.setVisible(false);
                                                          dialog.dispose();
                                                      }
                                                  }
                                              }
                                          };


        wizardDescriptor = new WizardDescriptor( this, new Object() );
        //wizardDescriptor.setModal( false );
        wizardDescriptor.setTitle(bundle.getString( "CTL_Wizard"));
        wizardDescriptor.setTitleFormat (new java.text.MessageFormat (" {0}"));
        wizardDescriptor.setAdditionalOptions (new Object[] { });
        wizardDescriptor.addPropertyChangeListener(listener);
        wizardDescriptor.setOptions (new Object[] {
                                         WizardDescriptor.PREVIOUS_OPTION,
                                         WizardDescriptor.NEXT_OPTION,
                                         WizardDescriptor.FINISH_OPTION,
                                         NotifyDescriptor.CANCEL_OPTION } );

    }

    /** Runs the wizard */
    void go() {

        dialog = TopManager.getDefault().createDialog( wizardDescriptor );

        canceled = false;
        
        dialog.addWindowListener(new java.awt.event.WindowAdapter() {
                public void windowClosing(java.awt.event.WindowEvent evt) {
                    cancel();
                }
            }
            );
            
        dialog.show();

        if ( !canceled ) {
            for (int i=0; i< panels.length; i++) {
                WizardData data = panels[i].getWizardData();
                if ( data != null && panels[i].isActive() ) {
                    data.run();                    
                }
            }

            SharedData.getShared().manifest();
            SharedData.getShared().generateBundle();
        }

    }


    void cancel() {
        canceled = true;        
        getCurrent().end( true );
    }

    boolean isCanceled() {
        return canceled;
    }

    // Implementation of Iterator --------------------------------------------------------

    public String name() {
        return getCurrent().getName();
    }

    public WizardDescriptor.Panel current() {

        return getCurrent();
    }

    private WizardPanel getCurrent() {

        return panels[ current ];
    }

    public boolean hasNext() {

        return current < panels.length - 1;
    }

    public boolean hasPrevious() {

        return current > 0;
    }

    public void nextPanel() {

        getCurrent().end( true );

        current ++;
        while (!getCurrent().isActive() && hasNext())
            current ++;

        getCurrent().start( true );

    }

    public void previousPanel() {

        getCurrent().end( false );

        current --;
        while (!getCurrent().isActive() && hasPrevious())
            current --;

        getCurrent().start( false );
    }        

    public synchronized void addChangeListener(javax.swing.event.ChangeListener listener) {}
    public synchronized void removeChangeListener(javax.swing.event.ChangeListener listener) {}

    // Inner classes ----------------------------------------------------------------------

    static interface Validator {

        // Called from component when the next button should be enabled or disbled
        public void setValid( boolean valid );

    }

    abstract class WizardPanel implements WizardDescriptor.Panel, Validator {

        protected Dimension WIZARD_SIZE = new Dimension( 550, 400 );

        protected boolean valid = true;

        /** Utility field used by event firing mechanism. */
        private EventListenerList listenerList = new EventListenerList();

        void start( boolean forward ) {}

        void end( boolean forward ) { }

        /** Called to get offset of the new panel */
        int nextPanelOffset() {
            return 1;
        }

        protected WizardData wizardData;

        private boolean active = true;

        protected WizardData getWizardData() {
            return wizardData;
        }

        protected void setWizardData( WizardData wizardData ) {
            this.wizardData = wizardData;
        }

        abstract String getName();

        // Implementation of Validator

        public void setValid(boolean valid) {
            this.valid = valid;
            fireChangeListenerStateChanged( this );
        }

        // Implementation of WizardDescriptor.Panel

        public boolean isValid() {
            return valid;
        }

        public HelpCtx getHelp() {
            return new HelpCtx ( Wizard.class );
        }

        public abstract Component getComponent();

        public void readSettings( Object settings ) {}

        public void storeSettings( Object settings ) {}

        /** Registers ChangeListener to receive events.
         *@param listener The listener to register.
         */
        public synchronized void addChangeListener(javax.swing.event.ChangeListener listener) {
            listenerList.add (javax.swing.event.ChangeListener.class, listener);
        }

        /** Removes ChangeListener from the list of listeners.
         *@param listener The listener to remove.
         */
        public synchronized void removeChangeListener(javax.swing.event.ChangeListener listener) {
            listenerList.remove (javax.swing.event.ChangeListener.class, listener);
        }

        /** Notifies all registered listeners about the event.
         *
         *@param param1 Parameter #1 of the <CODE>ChangeEvent<CODE> constructor.
         */
        protected void fireChangeListenerStateChanged(java.lang.Object param1) {
            javax.swing.event.ChangeEvent e = null;
            Object[] listeners = listenerList.getListenerList ();
            for (int i = listeners.length-2; i>=0; i-=2) {
                if (listeners[i]==javax.swing.event.ChangeListener.class) {
                    if (e == null)
                        e = new javax.swing.event.ChangeEvent (param1);
                    ((javax.swing.event.ChangeListener)listeners[i+1]).stateChanged (e);
                }
            }
        }

        public boolean isActive() {
            return active;
        }

        public void setActive(boolean active) {
            this.active = active;
        }

    }

    class IntroWizardPanel extends WizardPanel {

        IntroPanel panel;

        IntroWizardPanel () {
            panel = new IntroPanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void end( boolean forward ) {
        }

    }

    class ModuleWizardPanel extends WizardPanel {

        ModulePanel panel;
        ModuleData data = new ModuleData();

        ModuleWizardPanel () {
            panel = new ModulePanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
            setWizardData( data );
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void start( boolean forward ) {
            panel.addListeners();
        }
        
        void end( boolean forward ) {
            data.setModuleName(panel.getModuleName());
            data.setTargetFS(panel.getTargetFS());
            data.setPackageName(panel.getPackageName());
            panel.removeListeners();
        }

    }

    class SelectionWizardPanel extends WizardPanel {

        SelectionPanel panel;

        SelectionWizardPanel () {
            panel = new SelectionPanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void end( boolean forward ) {
            panels[3].setActive( panel.isDataLoader() );
            panels[4].setActive( panel.isMenu() );
        }

    }

    class LoaderWizardPanel extends WizardPanel {

        LoaderPanel panel;
        LoaderData data = new LoaderData();

        LoaderWizardPanel () {
            panel = new LoaderPanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
            setWizardData( data );
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void end( boolean forward ) {
            data.setLoaderName( SharedData.getShared().getModuleName() + " Data Loader" );
            data.setLoaderCode( SharedData.CLASS_PREFIX );
            data.setExtensions( panel.getExtensions() );
        }

    }

    class MenuWizardPanel extends WizardPanel {

        MenuPanel panel;
        MenuData data = new MenuData();

        MenuWizardPanel () {
            panel = new MenuPanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
            setWizardData( data );
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void start( boolean forward ) {
            panel.addListeners();
        }
        
        void end( boolean forward ) {
            data.setMenuName(panel.getMenuName());
            data.setMenuCode( SharedData.CLASS_PREFIX );
            data.setMenuHierarchy(panel.getMenuHierarchy());
            data.setToolbar( panel.isToolbar() );
            data.setToolHierarchy( panel.getToolHierarchy());
            data.setOtherMenuClass( panel.getOtherMenuClass());
            data.setBeforeOtherMenu( panel.isBeforeOtherMenu());
            data.setSeparatorBefore( panel.isSeparatorBefore());
            data.setSeparatorAfter( panel.isSeparatorAfter());
            data.setBeforeMenuClass( panel.getBeforeMenuClass());
            data.setAfterMenuClass( panel.getAfterMenuClass());
            data.setIconFile( panel.getIconFile());
            data.setShortcut( panel.getShortcut());
            panel.removeListeners();
        }

    }

    class ManifestWizardPanel extends WizardPanel {

        ManifestPanel panel;

        ManifestWizardPanel () {
            panel = new ManifestPanel( this ) {
                             public Dimension getPreferredSize() {
                                 return WIZARD_SIZE;
                             }
                         };
        }

        String getName() {
            return "";
        }

        public Component getComponent() {
            return panel;
        }

        void end( boolean forward ) {
            SharedData.getShared().setVersion(panel.getVersion());
            SharedData.getShared().setModuleDesc(panel.getModuleDesc());
            SharedData.getShared().setCategory(panel.getCategory());
        }

    }

}
