/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport.wizard;

import java.io.IOException;
import java.util.*;
import java.util.jar.Manifest;

import org.w3c.dom.Document;

import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.src.ClassElement;

/**
 * @author Jesse Glick
 */
public final class APIWizardData implements Cloneable {
    
    /** Property giving folder where new objects should by default be stored.
     * Of type org.openide.loaders.DataFolder;
     */
    public static final String PROP_MAIN_FOLDER = "main-folder"; // NOI18N
    
    private Manifest manifest = null;
    private Document layer = null;
    private ClassElement installer = null;
    private final Map misc = new HashMap (); // Map<Object,Object>
    // Accessed from APIIterator:
    final SortedSet tasks = new TreeSet (); // SortedSet<ExtraTask>
    final Set instantiations = new HashSet (); // Set<Instantiate>
    
    public APIWizardData () {
    }
    
    public Object clone () throws CloneNotSupportedException {
        APIWizardData nue = new APIWizardData ();
        if (manifest != null) nue.manifest = new Manifest (manifest);
        if (layer != null) nue.layer = (Document) layer.cloneNode (true);
        if (installer != null) nue.installer = (ClassElement) installer.clone ();
        nue.tasks.addAll (tasks);
        nue.instantiations.addAll (instantiations);
        nue.misc.putAll (misc);
        return nue;
    }
    
    public String toString () {
        Set stuff = new TreeSet ();
        if (manifest != null) stuff.add ("manifest"); // NOI18N
        if (layer != null) stuff.add ("layer"); // NOI18N
        if (installer != null) stuff.add ("installer"); // NOI18N
        return "APIWizardData[contents=" + stuff + ",misc=" + misc + ",tasks=" + tasks + ",instantiations=" + instantiations + "]"; // NOI18N
    }

    public Manifest getManifest () {
        return manifest;
    }
    
    public void setManifest (Manifest manifest) {
        this.manifest = manifest;
    }
    
    public Document getLayer () {
        return layer;
    }
    
    public void setLayer (Document layer) {
        this.layer = layer;
    }
    
    public ClassElement getInstaller () {
        return installer;
    }
    
    public void setInstaller (ClassElement installer) {
        this.installer = installer;
    }
    
    public void addTask (ExtraTask task) {
        tasks.add (task);
    }
    
    public void addInstantiation (Instantiate inst) {
        instantiations.add (inst);
    }
    
    public Map getProperties () {
        return misc;
    }
    
    public static abstract class ExtraTask implements Comparable {
        
        private final int priority;
        // Accessed from APIIterator:
        final Set objs = new HashSet (); // Set<DataObject>
        
        /** Make an extra task to run at wizard finish time.
         * Lower-priority-value tasks are run first.
         * Use priorities above zero normally; below zero for
         * updating manifest/layer/installer.
         * @param priority the order to run in
         */
        protected ExtraTask (int priority) {
            this.priority = priority;
        }
        
        protected abstract void run () throws IOException;
        
        protected final void created (DataObject obj) {
            objs.add (obj);
        }
        
        public int compareTo (Object o) {
            if (! (o instanceof ExtraTask)) {
                return -1;
            }
            return priority - ((ExtraTask) o).priority;
        }
        
    }
    
    public static final class Instantiate {
        
        private final DataObject template;
        private final DataFolder folder;
        private final String name;
        private final Map tokens = new HashMap (); // Map<String,String>
        // accessed from APIIterator:
        DataObject result = null;
        
        public Instantiate (DataObject template, DataFolder folder, String name) {
            this.template = template;
            this.folder = folder;
            this.name = name;
        }
        
        public void addToken (String name, String value) {
            tokens.put (name, value);
        }
        
        public DataObject getTemplate () {
            return template;
        }
        
        public DataFolder getFolder () {
            return folder;
        }
        
        public String getName () {
            return name;
        }
        
        public Map getTokens () {
            return new HashMap (tokens);
        }
        
        public DataObject getResult () {
            if (result == null) throw new IllegalStateException (toString ());
            return result;
        }
        
        public String toString () {
            return "Instantiate@" + System.identityHashCode (this) + "[template=" + template.getPrimaryFile () + ",folder=" + folder.getPrimaryFile () + ",name=" + name + ",tokens=" + tokens + ",result=" + result + "]"; // NOI18N
        }
        
        public boolean equals (Object o) {
            if (! (o instanceof Instantiate)) return false;
            Instantiate i = (Instantiate) o;
            return template == i.template && folder == i.folder && name.equals (i.name) && tokens.equals (i.tokens);
        }
        
        public int hashCode () {
            return 884743 ^ template.hashCode () ^ folder.hashCode () ^ name.hashCode () ^ tokens.hashCode ();
        }
        
    }
    
}
