/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport.wizard;

import java.awt.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

import org.openide.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.util.*;

/**
 * @author Jesse Glick
 */
public abstract class PolymorphicIterator implements TemplateWizard.Iterator, ChangeListener, TaskListener {
    
    private static final String TEMPLATE_PREFIX = "Templates/"; // NOI18N
    private static final String WIZARD_PREFIX = "Wizards/"; // NOI18N
    
    private WizardDescriptor.Iterator realWizard = null;
    private FolderInstance realWizardInstance = null;
    private final Set listeners = new HashSet (); // Set<ChangeListener>

    protected PolymorphicIterator () {
    }

    public synchronized void initialize (TemplateWizard templateWizard) {
        DataObject template = templateWizard.getTemplate ();
        String id = template.getPrimaryFile ().getPackageNameExt ('/', '_');
        if (! id.startsWith (TEMPLATE_PREFIX)) throw new IllegalStateException (); // NOI18N
        String foldername = WIZARD_PREFIX + id.substring (TEMPLATE_PREFIX.length ());
        FileObject folderfile = TopManager.getDefault ().getRepository ().getDefaultFileSystem ().findResource (foldername);
        // Will be set later:
        setRealWizard (null);
        if (folderfile != null && folderfile.isFolder ()) {
            DataFolder folder = DataFolder.findFolder (folderfile);
            realWizardInstance = new WizardFolder (folder);
            // Ask to be notified when it is finished:
            realWizardInstance.addTaskListener (this);
            // Scan it in the background:
            RequestProcessor.postRequest (realWizardInstance);
        } else {
            System.err.println("No such wizard folder: " + foldername);
        }
    }
    
    public void uninitialize (TemplateWizard templateWizard) {
        setRealWizard (null);
    }
    
    protected synchronized void setRealWizard (WizardDescriptor.Iterator wizard) {
        if (realWizard == wizard) return;
        if (realWizard != null) realWizard.removeChangeListener (this);
        realWizard = wizard;
        if (realWizard != null) realWizard.addChangeListener (this);
        // Probably having a real wizard means something has changed:
        fireChangeEvent ();
    }
    
    // get the underlying wizard; may return null in case of problems; may block
    protected WizardDescriptor.Iterator getRealWizard () {
        if (realWizard != null) return realWizard;
        synchronized (this) {
            if (realWizard != null) return realWizard;
            if (realWizardInstance == null) return null;
            // Else we are looking it up.
            realWizardInstance.instanceFinished ();
            try {
                setRealWizard ((WizardDescriptor.Iterator) realWizardInstance.instanceCreate ());
            } catch (Exception e) { // IOException, ClassNotFoundException, ClassCastException, ...
                e.printStackTrace();
            } finally {
                // Done with it. No need to try to listen to further changes etc.
                // If it worked, great, if not then skip it.
                realWizardInstance = null;
            }
            return realWizard;
        }
    }
    
    public final void addChangeListener (ChangeListener changeListener) {
        synchronized (listeners) {
            listeners.add (changeListener);
        }
    }
    
    public final void removeChangeListener (ChangeListener changeListener) {
        synchronized (listeners) {
            listeners.remove (changeListener);
        }
    }
    
    protected final void fireChangeEvent () {
        // [PENDING] fire asynch?
        ChangeEvent ev = new ChangeEvent (this);
        Set l2; // Set<ChangeListener>
        synchronized (listeners) {
            l2 = new HashSet (listeners);
        }
        Iterator it = l2.iterator ();
        while (it.hasNext ()) {
            ((ChangeListener) it.next ()).stateChanged (ev);
        }
    }
    
    public abstract Set instantiate (TemplateWizard wiz) throws IOException;
    
    public WizardDescriptor.Panel current () {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            return it.current ();
        } else {
            return new DummyPanel ();
        }
    }
    
    public String name () {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            return it.name ();
        } else {
            return "Loading...";
        }
    }
    
    public boolean hasNext () {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            return it.hasNext ();
        } else {
            return false;
        }
    }
    
    public boolean hasPrevious () {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            return it.hasPrevious ();
        } else {
            return false;
        }
    }
    
    public void nextPanel () throws NoSuchElementException {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            it.nextPanel ();
        } else {
            throw new NoSuchElementException ();
        }
    }
    
    public void previousPanel () throws NoSuchElementException {
        WizardDescriptor.Iterator it = getRealWizard ();
        if (it != null) {
            it.previousPanel ();
        } else {
            throw new NoSuchElementException ();
        }
    }
    
    // Fired from real wizard, propagate it:
    public void stateChanged (ChangeEvent ev) {
        fireChangeEvent ();
    }
    
    // Fired when folder has finished scanning.
    public void taskFinished (Task task) {
        // Probably not really needed, initialization of the wizard involves
        // calling current() which forces getRealWizard to finish anyway, but:
        fireChangeEvent ();
    }
    
    /** Temporary panel until real wizard finishes loading.
     */
    private static class DummyPanel implements WizardDescriptor.Panel {
        
        public Component getComponent () {
            JPanel panel = new JPanel ();
            panel.setLayout (new BorderLayout ());
            panel.add (new JLabel ("Loading wizard...", SwingConstants.CENTER), BorderLayout.CENTER);
            return panel;
        }
        
        public HelpCtx getHelp () {
            return HelpCtx.DEFAULT_HELP;
        }
        
        public boolean isValid () {
            return false;
        }
        
        public void readSettings (Object settings) {}
        public void addChangeListener (ChangeListener l) {}
        public void removeChangeListener (ChangeListener l) {}
        public void storeSettings (Object settings) {}
        
    }
    
}
