/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.apisupport.wizard.panels;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;
import javax.swing.*;
import javax.swing.event.*;

import org.openide.*;
import org.openide.cookies.*;
import org.openide.explorer.*;
import org.openide.explorer.view.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.modules.*;
import org.openide.nodes.*;

import org.netbeans.modules.apisupport.ManifestProvider;
import org.netbeans.modules.apisupport.wizard.*;

/** Panel permitting user to select a manifest for his module.
 * Some use scenarios:
 * 1. This is the first time this panel has ever been used.
 *    APIIterator.manifestDataObject (stored manifest) is null.
 *    Panel chooses a standard manifest template by name; user
 *    selects a target folder and filename for it; panel creates
 *    a manifest folder in that position, with entries acc. to later panels.
 * 2. The wizard is rerun; stored manifest exists. Panel shows the
 *    position of the stored manifest, which the user approves. Panel
 *    later updates manifest with new entries from other panels.
 * 3. The wizard is rerun; the user chooses to make a new module (manifest),
 *    the rest much like 1.
 * 4. The wizard is rerun; the user chooses a different existing manifest;
 *    the rest much like 2.
 * This panel should generally be included as the first panel of any wizard
 * that might need to access manifests (most of them).
 * @author Jesse Glick
 */
public class ManifestPanel extends APIPanel {
    
    // will be non-null if there was already a manifest on disk:
    private DataObject manifestFile;
    // never null:
    private Manifest manifest;
    // non-null if we will create a new manifest:
    private APIWizardData.Instantiate inst;
    // may or may not be null (this is not the dest for the manifest, but for regular sources):
    private DataFolder destination;
    // null initially:
    private Component component;
    // non-null if there is a template we can use:
    private DataObject template;
    // true if creating a new manifest, false if using an existing one:
    private boolean makingNew;

    public ManifestPanel () {
    }
    
    protected String manifestTemplate () {
        return "Templates/API_Support/Modules_API/Manifest.mf"; // NOI18N
    }

    public Component getComponent () {
        if (component == null) {
            component = new Pane ();
        }
        return component;
    }
    
    public boolean isValid () {
        return makingNew ? inst != null : manifestFile != null;
    }
    
    protected void readSettings (APIWizardData data) {
        manifest = data.getManifest (); // probably will be null
        manifestFile = APIIterator.getManifestDataObject ();
        if (manifestFile != null) {
            ManifestProvider p = (ManifestProvider) manifestFile.getCookie (ManifestProvider.class);
            if (p != null) {
                try {
                    manifest = new Manifest (p.getManifest ());
                } catch (IOException ioe) {
                    TopManager.getDefault ().notifyException (ioe);
                }
            }
        }
        destination = (DataFolder) data.getProperties ().get (APIWizardData.PROP_MAIN_FOLDER);
    }
    
    protected void storeSettings (APIWizardData data) {
        data.setManifest (manifest);
        if (inst != null) {
            data.addInstantiation (inst);
        }
        data.addTask (new MyTask ());
    }
    
    private class MyTask extends APIWizardData.ExtraTask {
        
        public MyTask () {
            super (-3);
        }
        
        protected void run () throws IOException {
            DataObject file;
            if (inst != null) {
                file = inst.getResult ();
            } else {
                file = manifestFile;
            }
            if (file == null) throw new IllegalStateException ();
            ManifestProvider p = (ManifestProvider) file.getCookie (ManifestProvider.class);
            if (p == null) throw new IOException ();
            APIIterator.setManifestDataObject (file);
            p.setManifest (manifest);
            SaveCookie save = (SaveCookie) file.getCookie (SaveCookie.class);
            if (save != null) {
                save.save ();
            }
        }
        
    }
    
    private class Pane extends JPanel implements ManifestProvider, PropertyChangeListener, KeyListener {
        
        private final Set listeners = new HashSet (); // Set<ChangeListener>
        
        public Pane () {
            initComponents ();
            // XXX crude temporary measure to make whole dialog visible on screen:
            detailsSheet.setPreferredSize (new Dimension (500, 200));
            newFolderTree.setPreferredSize (new Dimension (250, 200));
            existingManifest.setPreferredSize (new Dimension (250, 200));
            
            FileObject fo = TopManager.getDefault ().getRepository ().getDefaultFileSystem ().findResource (manifestTemplate ());
            if (fo != null) {
                try {
                    DataObject tryTemplate = DataObject.find (fo);
                    ManifestProvider p = (ManifestProvider) tryTemplate.getCookie (ManifestProvider.class);
                    if (p != null) {
                        template = tryTemplate;
                        if (manifest == null) {
                            // Use a copy of the template's manifest:
                            manifest = new Manifest (p.getManifest ());
                        }
                    }
                } catch (IOException e) {
                    TopManager.getDefault ().notifyException (e);
                }
            }
            if (manifest == null) {
                // Should not normally occur:
                manifest = new Manifest ();
                manifest.getMainAttributes ().put (ModuleDescription.TAG_MAGIC, "some.code.name"); // NOI18N
                manifest.getMainAttributes ().put (ModuleDescription.TAG_NAME, "Module Name");
                manifest.getMainAttributes ().put (ModuleDescription.TAG_SPEC_VERSION, "0.1"); // NOI18N
            }
            makingNew = (manifestFile == null);
            if (makingNew && destination != null && ! destination.getPrimaryFile ().isRoot ()) {
                manifest.getMainAttributes ().put (ModuleDescription.TAG_MAGIC, destination.getPrimaryFile ().getPackageName ('.'));
            }
            DataFilter existingFilter = new DataFilter () {
                public boolean acceptDataObject (DataObject obj) {
                    return obj.getCookie (DataFolder.class) != null ||
                           obj.getCookie (ManifestProvider.class) != null;
                }
            };
            Node existingRepo = TopManager.getDefault ().getPlaces ().nodes ().repository (existingFilter);
            existingPanel.getExplorerManager ().setRootContext (existingRepo);
            if (manifestFile != null) {
                FileObject mfPrim = manifestFile.getPrimaryFile ();
                Node[] fss = existingRepo.getChildren ().getNodes ();
                boolean foundIt = false;
                for (int i = 0; i < fss.length; i++) {
                    DataObject obj = (DataObject) fss[i].getCookie (DataObject.class);
                    if (obj != null) {
                        FileObject prim = obj.getPrimaryFile ();
                        try {
                            if (prim.isRoot () && prim.getFileSystem () == mfPrim.getFileSystem ()) {
                                Node manifestNode = NodeOp.findPath (fss[i], new StringTokenizer (mfPrim.getPackageName ('/'), "/")); // NOI18N
                                existingPanel.getExplorerManager ().setSelectedNodes (new Node[] { manifestNode });
                                break;
                            }
                        } catch (Exception e) { // FileStateInvalid, NodeNotFound, PropertyVeto, ...
                            TopManager.getDefault ().notifyException (e);
                        }
                    }
                }
                if (! foundIt) {
                    System.err.println("did not find proper node");//XXX
                }
            }
            existingPanel.getExplorerManager ().addPropertyChangeListener (this);
            DataFilter folderFilter = new DataFilter () {
                public boolean acceptDataObject (DataObject obj) {
                    return obj.getCookie (DataFolder.class) != null;
                }
            };
            Node newRepo = TopManager.getDefault ().getPlaces ().nodes ().repository (folderFilter);
            newPanel.getExplorerManager ().setRootContext (newRepo);
            newPanel.getExplorerManager ().addPropertyChangeListener (this);
            existingRadio.setSelected (! makingNew);
            newRadio.setSelected (makingNew);
            // XXX why doesn't just PropertyChangeListener work?
            // even this does not really work right...
            newManifestName.addKeyListener (this);
            updateStateNoFire ();
        }
        
        private void updateStateNoFire () {
            // XXX when !makingNew, if a new legitimate node is selected, should refresh
            // manifest with the manifest from that node; otherwise it should permit the
            // user to make changes to the manifest, which it should keep
            System.err.println("updateStateNoFire");//XXX
            existingPanel.setEnabled (manifestFile != null);
            newPanel.setEnabled (template != null);
            detailsPanel.setEnabled (manifest != null);
            Node[] newFolder = newPanel.getExplorerManager ().getSelectedNodes ();
            if (newFolder.length == 1 && template != null && newManifestName.getText ().length () > 0) {
                APIWizardData.Instantiate inst2 = new APIWizardData.Instantiate (template, (DataFolder) newFolder[0].getCookie (DataFolder.class), newManifestName.getText ());
                if (! inst2.equals (inst)) {
                    inst = inst2;
                } // else avoid changing it, for we might get a dummy one with no result
            } else {
                inst = null;
            }
            Node[] manifestLocation = existingPanel.getExplorerManager ().getSelectedNodes ();
            manifestFile = null;
            if (manifestLocation.length == 1) {
                ManifestProvider p = (ManifestProvider) manifestLocation[0].getCookie (ManifestProvider.class);
                if (p != null) {
                    manifestFile = (DataObject) manifestLocation[0].getCookie (DataObject.class);
                }
            }
            Node manifestDetail = new AbstractNode (Children.LEAF) {
                protected Sheet createSheet () {
                    Sheet sheet = new Sheet ();
                    ManifestProvider.Util.addToSheet (sheet, Pane.this);
                    return sheet;
                }
            };
            detailsPanel.getExplorerManager ().setRootContext (manifestDetail);
            try {
                detailsPanel.getExplorerManager ().setSelectedNodes (new Node[] { manifestDetail });
            } catch (PropertyVetoException pve) {
                TopManager.getDefault ().notifyException (pve);
            }
        }
        
        private void updateState () {
            System.err.println("updateState");//XXX
            updateStateNoFire ();
            fireManifestProviderChangeEvent ();
            ManifestPanel.this.fireChangeEvent (); // not always useful, but just to be safe
        }
        
        private void fireManifestProviderChangeEvent () {
            ChangeEvent ev = new ChangeEvent (this);
            Iterator it = listeners.iterator ();
            while (it.hasNext ()) {
                ((ChangeListener) it.next ()).stateChanged (ev);
            }
        }
        
        private void initComponents () {//GEN-BEGIN:initComponents
            existingPanel = new org.openide.explorer.ExplorerPanel ();
            existingManifest = new org.openide.explorer.view.BeanTreeView ();
            newPanel = new org.openide.explorer.ExplorerPanel ();
            newFolderTree = new org.openide.explorer.view.BeanTreeView ();
            newManifestName = new javax.swing.JTextField ();
            newManifestLabel = new javax.swing.JLabel ();
            detailsPanel = new org.openide.explorer.ExplorerPanel ();
            detailsSheet = new org.openide.explorer.propertysheet.PropertySheetView ();
            existingRadio = new javax.swing.JRadioButton ();
            newRadio = new javax.swing.JRadioButton ();
            setLayout (new java.awt.GridBagLayout ());
            java.awt.GridBagConstraints gridBagConstraints1;
            
            existingPanel.setLayout (new java.awt.GridBagLayout ());
            java.awt.GridBagConstraints gridBagConstraints4;
            existingPanel.setBorder (new javax.swing.border.TitledBorder ("Existing Manifest"));
            
            gridBagConstraints4 = new java.awt.GridBagConstraints ();
            gridBagConstraints4.fill = java.awt.GridBagConstraints.BOTH;
            existingPanel.add (existingManifest, gridBagConstraints4);
            
            
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 1;
            gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
            add (existingPanel, gridBagConstraints1);
            
            
            newPanel.setLayout (new java.awt.GridBagLayout ());
            java.awt.GridBagConstraints gridBagConstraints5;
            newPanel.setBorder (new javax.swing.border.TitledBorder ("New Manifest"));
            
            gridBagConstraints5 = new java.awt.GridBagConstraints ();
            gridBagConstraints5.gridx = 0;
            gridBagConstraints5.gridy = 0;
            gridBagConstraints5.gridwidth = 2;
            gridBagConstraints5.fill = java.awt.GridBagConstraints.BOTH;
            newPanel.add (newFolderTree, gridBagConstraints5);
            
            
            gridBagConstraints5 = new java.awt.GridBagConstraints ();
            gridBagConstraints5.gridx = 1;
            gridBagConstraints5.gridy = 1;
            gridBagConstraints5.fill = java.awt.GridBagConstraints.HORIZONTAL;
            gridBagConstraints5.anchor = java.awt.GridBagConstraints.WEST;
            newPanel.add (newManifestName, gridBagConstraints5);
            
            
            newManifestLabel.setText ("Name:");
            gridBagConstraints5 = new java.awt.GridBagConstraints ();
            gridBagConstraints5.gridx = 0;
            gridBagConstraints5.gridy = 1;
            gridBagConstraints5.anchor = java.awt.GridBagConstraints.EAST;
            newPanel.add (newManifestLabel, gridBagConstraints5);
            
            
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 1;
            gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
            add (newPanel, gridBagConstraints1);
            
            
            detailsPanel.setLayout (new java.awt.GridBagLayout ());
            java.awt.GridBagConstraints gridBagConstraints7;
            detailsPanel.setBorder (new javax.swing.border.TitledBorder ("Manifest Details"));
            
            gridBagConstraints7 = new java.awt.GridBagConstraints ();
            gridBagConstraints7.fill = java.awt.GridBagConstraints.BOTH;
            detailsPanel.add (detailsSheet, gridBagConstraints7);
            
            
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 2;
            gridBagConstraints1.gridwidth = 2;
            gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
            add (detailsPanel, gridBagConstraints1);
            
            
            existingRadio.setText ("Use existing manifest");
            existingRadio.addActionListener (new java.awt.event.ActionListener () {
                public void actionPerformed (java.awt.event.ActionEvent evt) {
                    existingRadioActionPerformed (evt);
                }
            }
            );
            
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 0;
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            add (existingRadio, gridBagConstraints1);
            
            
            newRadio.setText ("Create new manifest");
            newRadio.addActionListener (new java.awt.event.ActionListener () {
                public void actionPerformed (java.awt.event.ActionEvent evt) {
                    newRadioActionPerformed (evt);
                }
            }
            );
            
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 0;
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            add (newRadio, gridBagConstraints1);
            
        }//GEN-END:initComponents

        private void newRadioActionPerformed (java.awt.event.ActionEvent evt) {//GEN-FIRST:event_newRadioActionPerformed
            makingNew = true;
            updateState ();
    }//GEN-LAST:event_newRadioActionPerformed

        private void existingRadioActionPerformed (java.awt.event.ActionEvent evt) {//GEN-FIRST:event_existingRadioActionPerformed
            makingNew = false;
            updateState ();
    }//GEN-LAST:event_existingRadioActionPerformed

        public Manifest getManifest () /*throws IOException*/ {
            return manifest;
        }    

        public void setManifest (Manifest m) /*throws IOException*/ {
            manifest = m;
            updateStateNoFire ();
            fireManifestProviderChangeEvent ();
        }

        public void addFiles (Set files) /*throws IOException*/ {
            // do nothing
        }

        public void removeFiles (Set files) /*throws IOException*/ {
            // do nothing
        }

        public Set getFiles () /*throws IOException*/ {
            return Collections.EMPTY_SET;
        }

        public void addChangeListener (ChangeListener list) {
            listeners.add (list);
        }

        public void removeChangeListener (ChangeListener list) {
            listeners.remove (list);
        }

        public boolean isValid () {
            return true;
        }

        public Exception getParseException () {
            return null;
        }

        public File getManifestAsFile () {
            return null;
        }

        public File getModuleJar () {
            return null;
        }
    
        public void propertyChange (PropertyChangeEvent ev) {
            updateState ();
        }
        
        public void keyReleased (KeyEvent ev) {
        }
        
        public void keyPressed (KeyEvent ev) {
        }
        
        public void keyTyped (KeyEvent ev) {
            updateState ();
        }
        
        // Variables declaration - do not modify//GEN-BEGIN:variables
        private org.openide.explorer.ExplorerPanel existingPanel;
        private org.openide.explorer.view.BeanTreeView existingManifest;
        private org.openide.explorer.ExplorerPanel newPanel;
        private org.openide.explorer.view.BeanTreeView newFolderTree;
        private javax.swing.JTextField newManifestName;
        private javax.swing.JLabel newManifestLabel;
        private org.openide.explorer.ExplorerPanel detailsPanel;
        private org.openide.explorer.propertysheet.PropertySheetView detailsSheet;
        private javax.swing.JRadioButton existingRadio;
        private javax.swing.JRadioButton newRadio;
        // End of variables declaration//GEN-END:variables

    }
    
}
