/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.applet;

import java.io.IOException;
import java.io.File;
import java.text.MessageFormat;

import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.cookies.SourceCookie;
import org.openide.loaders.DataObject;
import org.openide.execution.NbProcessDescriptor;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.HttpServer;
import org.openide.execution.ProcessExecutor;
import org.openide.execution.ExecInfo;
import org.openide.execution.ExecutorTask;

/** Executes a class externally (in a separate process). Provides
* basic implementation that allows to specify the process to 
* execute, its parameters and also to substitute the content of repositorypath,
* classpath, bootclasspath and librarypath. This is done by inner class Format.
* <P>
* The behaviour described here can be overriden by subclasses to use different
* format (extend the set of recognized tags), execute the 
* process with additional environment properties, etc.
*
* @author Petr Jiricka, Ales Novak, Jaroslav Tulach
*/
public class AppletExecutor extends ProcessExecutor {

    private String url;

    /** default descriptor to use */
    private static NbProcessDescriptor DEFAULT_APPLET_DESCRIPTOR;

    static final long serialVersionUID =5682139532418769413L;
    /** Create a new executor.
    * The default Java launcher associated with this VM's installation will be used,
    * and the user repository entries will be used for the class path.
    */
    public AppletExecutor() {
        super();
        // setExternalExecutor(DEFAULT_APPLET_DESCRIPTOR);
    }


    /* Default human-presentable name of the executor.
    * In the default implementation, just the class name.
    * @return initial value of the human-presentable name
    */
    public String displayName() {
        return NbBundle.getBundle(AppletExecutor.class).getString("CTL_Exec_Name");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (AppletExecutor.class);
    }

    /** Called to create the java.lang.Process for given exec info.
    * Current implementation scans creates new Format with provided
    * exec info and asks the current executor to start with that
    * format.
    * <P>
    * Subclasses can override this to achive the right behaviour, add
    * system properties, own format, etc.
    * 
    * @param info exec info 
    * @return the executed process
    * @exception IOException if the action fails
    */
    protected Process createProcess (DataObject obj) throws IOException {
        if (url == null)
            throw new IOException(NbBundle.getBundle(AppletExecutor.class).getString("EXC_NoHtmlPage"));
        ExecInfo info = new ExecInfo(obj.getPrimaryFile().getPackageName ('.'));
        return getExternalExecutor ().exec(new AppletFormat(info, url));
    }


    /* Executes given class by creating new process in underlying operating system.
    * @param ctx used to write to the Output Window
    * @param info information about the class to be executed
    */
    public ExecutorTask execute(DataObject appletOrPage) throws IOException {
        url = null;
        try {
            if (appletOrPage.getCookie(SourceCookie.class) != null) {
                // this is a Java file or class, generate the HTML page for it
                url = AppletSupport.generateHtmlFileURL(appletOrPage.getPrimaryFile()).toString();
            }
            else {
                // this is probably a HTML file, just show that
                url = HttpServer.getResourceURL(appletOrPage.getPrimaryFile().getPackageNameExt('/','.')).toString();
            }
        } catch (HttpServerNotFoundException ex) {
            AppletSupport.reportNoHttpServer();
        }
        return super.execute(appletOrPage);
    }

    /** Get the current external execution command.
 * @return the settings for the current external executor
 */
    public NbProcessDescriptor getExternalExecutor () {
        if (externalExecutor == null) {
            synchronized (AppletExecutor.class) {
                if (externalExecutor == null) 
                    externalExecutor = new NbProcessDescriptor (
                        // /usr/local/bin/appletviewer
                        "{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}..{" + Format.TAG_SEPARATOR + // NOI18N
                        "}bin{" + Format.TAG_SEPARATOR + "}appletviewer", // NOI18N
                        // {URL}
                        " {" + AppletFormat.TAG_URL + "}", // NOI18N
                        NbBundle.getBundle(AppletExecutor.class).getString("MSG_AppletExecutorHint")
                    );
            }
        }
        return externalExecutor;
    }
    
    /** Default format that can format tags related to execution. Currently this is only the URL.
    */
    public static class AppletFormat extends Format {
        /** Tag replaced with the URL */
        public static final String TAG_URL = "URL"; // NOI18N

        static final long serialVersionUID =4315554797414856261L;
        /** @param info exec info about class to execute
        * @param classPath to substitute instead of CLASSPATH
        * @param bootClassPath boot class path
        * @param repository repository path
        * @param library library path
        */
        public AppletFormat (ExecInfo info, String url) {
            super(info);
            java.util.Map map = getMap ();

            map.put (TAG_URL, url);
        }

    }

}
