/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.updater;

import java.io.*;
import java.util.Enumeration;
import java.util.jar.*;
import java.util.ResourceBundle;
import java.util.Properties;

import org.openide.util.NbBundle;

/** Class used by autoupdate module for the work with module files and
 * for installing / uninstalling modules
 *
 * @author  Petr Hrebejk, Ales Kemr
 * @version 
 */
public class ModuleUpdater extends Thread {

    ResourceBundle bundle = java.util.ResourceBundle.getBundle("org/netbeans/updater/Bundle"); // NOI18N

    /** Platform dependent file name separator */
    private static final String FILE_SEPARATOR = System.getProperty ("file.separator");

    /** Relative name of update directory */
    private static final String UPDATE_DIR = "update"; // NOI18N

    /** Relative name of directory where the .NBM files are downloaded */
    private static final String DOWNLOAD_DIR =UPDATE_DIR + FILE_SEPARATOR + "download"; // NOI18N

    /** Relative name of backup directory */
    private static final String BACKUP_DIR = UPDATE_DIR + FILE_SEPARATOR + "backup"; // NOI18N

    /** Relative name of netbeans/lib directory */
    private static final String NB_LIB_DIR = "lib"; // NOI18N

    /** Name of extension directory */
    private static final String EXT_DIR = "ext"; // NOI18N

    /** Relative name of java/lib/ext directory */
    private static final String JAVA_LIB_EXT_DIR = "lib" +  FILE_SEPARATOR + EXT_DIR; // NOI18N

    /** The name of zip entry containing netbeans files */
    public static final String UPDATE_NETBEANS_DIR = "netbeans"; // NOI18N

    /** The name of zip entry containing java_extension files */
    public static final String UPDATE_JAVA_EXT_DIR = "java_ext"; // NOI18N

    /** The name of zip entry containing files for external installer */
    public static final String UPDATE_MAIN_DIR = "main"; // NOI18N
    
    /** Absolut name of directory where are external installers*/
    private static final String MAIN_DIR = getNetbeansDirectory().getPath() + FILE_SEPARATOR 
        + "update" + FILE_SEPARATOR + UPDATE_MAIN_DIR; // NOI18N
    
    /** Name of external installer parameters file*/
    private static final String JVM_PARAMS_FILE = "main.properties";

    /** Extension of the distribution files */
    private static final String NBM_EXTENSION = "nbm"; // NOI18N

    /** The name of the log file */
    public static final String LOG_FILE_NAME = "update.log"; // NOI18N
    
    private static File downloadDirectory = null;
    private static File backupDirectory = null;
    private static File netbeansDirectory = null;
    private static File nbLibDirectory = null;
    private static File javaLibExtDirectory = null;
    private static File updateDirectory = null;
    private static File mainDirectory = new File (MAIN_DIR);

    /** All distribution files in the download directory */
    private File[] nbmFiles = null;

    /** Should the thread stop */
    private volatile boolean stop = false;

    private volatile boolean suspend = false;

    /** Total length of unpacked files */
    private long totalLength;

    /** Creates new ModuleUpdater */
    public void run() {


        getDownloadDirectory();

        checkStop();

        if ( downloadDirectory == null ) {
            endRun();
        }

        checkStop();
        nbmFiles = getNBMFiles();

        if ( nbmFiles == null || nbmFiles.length == 0 ) {
            endRun();
        }

        checkStop();

        totalLength();

        checkStop();

        unpack();

        UpdaterFrame.unpackingFinished();
    }

    /** ends the run of update */
    void endRun() {
        stop = true;
    }

    /** checks wheter ends the run of update */
    private void checkStop() {

        if ( suspend )
            while ( suspend );

        if ( stop ) {
            System.exit( 0 );
        }
    }

    /** Creates array of all .NBM files in the update/download directory
    */
    static File[] getNBMFiles() {


        File dirList[] = downloadDirectory.listFiles( new FilenameFilter() {
                             public boolean accept( File dir, String name ) {
                                 return name.endsWith( NBM_EXTENSION );
                             }
                         });

        return dirList;
    }


    /** Determines size of unpacked modules */
    private void totalLength() {
        totalLength = 0L;

        UpdaterFrame.setLabel( bundle.getString( "CTL_PreparingUnpack" ) );
        UpdaterFrame.setProgressRange( 0, nbmFiles.length );

        for( int i = 0; i < nbmFiles.length; i++ ) {

            JarFile jarFile = null;

            try {
                UpdaterFrame.setProgressValue( i + 1 );

                jarFile = new JarFile( nbmFiles[i] );
                Enumeration entries = jarFile.entries();
                while( entries.hasMoreElements() ) {
                    JarEntry entry = (JarEntry) entries.nextElement();

                    checkStop();

                    if ( ( entry.getName().startsWith( UPDATE_NETBEANS_DIR ) ||
                            entry.getName().startsWith( ModuleUpdater.UPDATE_JAVA_EXT_DIR ) ||
                            entry.getName().startsWith( UPDATE_MAIN_DIR) ) &&
                            !entry.isDirectory() ) {
                        totalLength += entry.getSize();
                    }
                }
            }
            catch ( java.io.IOException e ) {
                // Ignore non readable files
            }
            finally {
                try {
                    if ( jarFile != null )
                        jarFile.close();
                }
                catch ( java.io.IOException e ) {
                    // We can't close the file do nothing
                    // System.out.println( "Cant close : " + e ); // NOI18N
                }
            }
        }
    }


    /** Unpack the distribution files into update directory */

    void unpack ()  {

        long bytesRead = 0L;
        boolean hasMainClass;

        // System.out.println("Total lengtg " + totalLength ); // NOI18N

        UpdaterFrame.setLabel( "" ); // NOI18N
        UpdaterFrame.setProgressRange( 0, totalLength );
        
        LogWriter log = new LogWriter();
        log.println("Updater: "); // NOI18N                    
        log.println("NBM files installed: "); // NOI18N

        for( int i = 0; i < nbmFiles.length; i++ ) {
            
            hasMainClass = false;
            UpdaterFrame.setLabel( bundle.getString("CTL_UnpackingFile") + "  " + nbmFiles[i].getName() ); //NOI18N
            UpdaterFrame.setProgressValue( bytesRead );
            log.println("> " + nbmFiles[i].getName()); // NOI18N 
            JarFile jarFile = null;

            try {
                jarFile = new JarFile( nbmFiles[i] );
                Enumeration entries = jarFile.entries();
                while( entries.hasMoreElements() ) {
                    JarEntry entry = (JarEntry) entries.nextElement();

                    checkStop();

                    if ( entry.getName().startsWith( UPDATE_NETBEANS_DIR ) ) {
                        // Copy files into netbeans directory
                        if ( entry.isDirectory() ) {
                            File newDir = new File( getNetbeansDirectory(), entry.getName().substring( UPDATE_NETBEANS_DIR.length() ) );
                            if ( !newDir.isDirectory() )
                                newDir.mkdirs();
                            File newBckDir = new File( getBackupDirectory(), entry.getName() );
                            if ( !newBckDir.isDirectory() )
                                newBckDir.mkdirs();
                        }
                        else {
                            File destFile = new File( getNetbeansDirectory(), entry.getName().substring( UPDATE_NETBEANS_DIR.length() ) );
                            if ( destFile.exists() ) {
                                File bckFile = new File( getBackupDirectory(), entry.getName() );
				bckFile.getParentFile ().mkdirs ();
                                // System.out.println("Backing up" ); // NOI18N
                                copyStreams( new FileInputStream( destFile ), new FileOutputStream( bckFile ) );
                            } else {
				destFile.getParentFile ().mkdirs ();
			    }
                            copyStreams( jarFile.getInputStream( entry ), new FileOutputStream( destFile ) );
                            bytesRead += entry.getSize();
                            UpdaterFrame.setProgressValue( bytesRead );

                        }
                    }
                    else if ( entry.getName().startsWith( ModuleUpdater.UPDATE_JAVA_EXT_DIR ) &&
                              !entry.isDirectory() ) {
                        // Copy files into java/lib/ext directory
                        File destFile = new File( getJavaLibExtDirectory(), entry.getName().substring( UPDATE_JAVA_EXT_DIR.length() ) );
                        if ( destFile.exists() ) {
                            File bckFile = new File( getBackupDirectory(), ModuleUpdater.UPDATE_JAVA_EXT_DIR + FILE_SEPARATOR + entry.getName().substring( UPDATE_NETBEANS_DIR.length() ) );
			    bckFile.getParentFile ().mkdirs ();
                            copyStreams( new FileInputStream( destFile ), new FileOutputStream( bckFile ) );
			} else {
			    destFile.getParentFile ().mkdirs ();
                        }
                        copyStreams( jarFile.getInputStream( entry ), new FileOutputStream( destFile ) );
                        bytesRead += entry.getSize();
                        UpdaterFrame.setProgressValue( bytesRead );
                    }
                    else if ( entry.getName().startsWith( UPDATE_MAIN_DIR )&&
                              !entry.isDirectory() ) {
                        // run main                  
                        File destFile = new File( mainDirectory, 
                            entry.getName().substring(UPDATE_MAIN_DIR.length() + 1) );
                        destFile.getParentFile ().mkdirs ();
                        hasMainClass = true;
                        copyStreams( jarFile.getInputStream( entry ), new FileOutputStream( destFile ) );
                        bytesRead += entry.getSize();
                        UpdaterFrame.setProgressValue( bytesRead );
                    }
                }
                if ( hasMainClass ) {                    
                    MainConfig mconfig = new MainConfig(MAIN_DIR + FILE_SEPARATOR + JVM_PARAMS_FILE);
                    if (mconfig.isValid()) {
                        String java_path = System.getProperty ("java.home") + FILE_SEPARATOR
                            + "bin"  + FILE_SEPARATOR;
                        String torun = java_path + "java -cp " + MAIN_DIR + mconfig.getCommand();                                
                        startCommand(torun);
                        File[] mainfiles=mainDirectory.listFiles();
                        for(int j = 0;j < mainfiles.length;j++)
                            mainfiles[j].delete();                        
                    }
                }
            }
            catch ( java.io.IOException e ) {
                // Ignore non readable files
		if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
		    e.printStackTrace ();
            }
            finally {
                try {
                    if ( jarFile != null )
                        jarFile.close();
                }
                catch ( java.io.IOException e ) {
                    // We can't close the file do nothing
                    // System.out.println("Can't close : " + e ); // NOI18N
                }
                //System.out.println("Dleting :" + nbmFiles[i].getName() + ":" + nbmFiles[i].delete() ); // NOI18N

                nbmFiles[i].delete();

                //nbmFiles[i].deleteOnExit();

            }
        }
        log.println("Finished."); // NOI18N
        log.close();
    }

    /** The directory where to download the distribution files of modules */
    public static File getUpdateDirectory() {
        if ( updateDirectory == null ) {

            if ( isMultiuser() )
                updateDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + UPDATE_DIR );
            else
                updateDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + UPDATE_DIR );

            if ( !updateDirectory.isDirectory() )
                updateDirectory.mkdirs();
        }

        return updateDirectory;
    }

    /** The directory where to download the distribution files of modules */
    public static File getDownloadDirectory() {
        if ( downloadDirectory == null ) {

            if ( isMultiuser() )
                downloadDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + DOWNLOAD_DIR );
            else
                downloadDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + DOWNLOAD_DIR );

            if ( !downloadDirectory.isDirectory() )
                downloadDirectory.mkdirs();
        }
        return downloadDirectory;
    }

    private void startCommand(String torun) {
        Runtime runtime=Runtime.getRuntime();
        Process proces;            
        try {
            proces=runtime.exec(torun);
            final Process proc2 = proces;
            new Thread() {
                public void run() {
                    try {
                        InputStreamReader stream= new InputStreamReader (proc2.getErrorStream());
                        BufferedReader reader= new BufferedReader(stream);
                        String vystup;
                        do {
                            vystup = reader.readLine();
                            if (vystup!=null)
                                System.out.println(vystup);
                        } while (vystup != null);
                    } catch (Exception e) {
                        e.printStackTrace();
                  }
                }
            }.start();
            int x=proces.waitFor();
        }
        catch (Exception e){
          e.printStackTrace();
        }
    }
    
    /** The directory where to backup old versions of modules */
    public static File getBackupDirectory() {
        if ( backupDirectory == null ) {

            if ( isMultiuser() )
                backupDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + BACKUP_DIR );
            else
                backupDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + BACKUP_DIR );

            if ( !backupDirectory.isDirectory() )
                backupDirectory.mkdirs();
        }

        return backupDirectory;
    }


    /** Gets the netbeans directory */
    public static File getNetbeansDirectory() {
        if ( netbeansDirectory == null ) {
            if ( isMultiuser() )
                netbeansDirectory = new File (System.getProperty ("netbeans.user") );
            else
                netbeansDirectory = new File (System.getProperty ("netbeans.home") );
        }
        return netbeansDirectory;
    }

    /** The directory of libraries that are added to CLASSPATH on startup */

    public static File getNbLibDirectory() {
        if ( nbLibDirectory == null ) {

            if ( isMultiuser() )
                nbLibDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + NB_LIB_DIR );
            else
                nbLibDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + NB_LIB_DIR );
        }

        File nbLibExt = new File( nbLibDirectory, EXT_DIR );
        if ( !nbLibExt.isDirectory() )
            nbLibExt.mkdirs();

        return nbLibDirectory;
    }


    /** The directory lib/ext directory of JDK */
    public static File getJavaLibExtDirectory() {
        if ( javaLibExtDirectory == null ) {
            javaLibExtDirectory = new File (System.getProperty ("java.home") + FILE_SEPARATOR + JAVA_LIB_EXT_DIR );
        }

        if ( canWrite( javaLibExtDirectory, true ) )
            return javaLibExtDirectory;
        else
            return getNbLibDirectory();
    }

    private static boolean isMultiuser() {

        if ( System.getProperty ("netbeans.user") == null ||
                System.getProperty ("netbeans.home").equals( System.getProperty ("netbeans.user") ) )
            return false;
        else
            return true;
    }


    /*
    private static Manifest getManifest( File file ) throws java.io.IOException {
      JarFile jar = new JarFile( file );
      Manifest man = jar.getManifest();
      jar.close();
      return man; 
}
    */



    /** Copies two streams */
    private void copyStreams( InputStream src, OutputStream dest ) throws java.io.IOException {

        BufferedInputStream bsrc = new BufferedInputStream( src );
        BufferedOutputStream bdest = new BufferedOutputStream( dest );

        int count = 0;

        int c;

        try {
            while( ( c = bsrc.read() ) != -1 ) {
                bdest.write( c );
                count++;
                if ( count > 8500 ) {
                    count = 0;
                    checkStop();
                }
            }
        }
        finally {
            bsrc.close();
            bdest.close();
        }
    }

    /** Test whether the user has rights to write into directory */

    private static boolean canWrite( File dir, boolean create ) {
        if ( !dir.exists() && create )

            dir.mkdirs();

        if ( !dir.isDirectory() || !dir.canWrite() )
            return false;



        File tmp = null;

        try {
            tmp = File.createTempFile( "test", "access", dir ); // NOI18N
        }
        catch ( java.io.IOException e ) {
            return false;
        }

        if ( tmp == null )
            return false;

        boolean cw = tmp.canWrite();
        if (cw)
            tmp.delete();

        return cw;
    }

    /** read jvm parameters from jvm parameters file */
    class MainConfig extends Object {
        
        /** The names of properties from jvm parameters file */
        private final String PAR_MAIN = "mainClass";               // NOI18N
        private final String PAR_RELCP = "relativeClassPath";      // NOI18N
        private final String PAR_JVMPAR = "jvm.parameters";        // NOI18N
        private final String PAR_MAINARGS = "mainClass.arguments"; // NOI18N
        
        /** The names of variables allow to use in jvm parameters file */
        private final String VAR_IDE_HOME = "%IDE_HOME%";          // NOI18N
        private final String VAR_IDE_USER = "%IDE_USER%";          // NOI18N
        private final String VAR_FILE_SEPARATOR = "%FS%";          // NOI18N        
        private final String VAR_JAVA_HOME = "%JAVA_HOME%";        // NOI18N        
        private final String VAR_ICON_ROOT = "%ICON_ROOT%";        // NOI18N        
        private final String ICON_ROOT = "Forte for Java"; // NOI18N        
    
        /** joined all parameters of jvm java command */
        private String parameters = "";
        
        /** is jvm parameters file in valid stucture */
        private boolean valid = false;
        
        public MainConfig(String spath) {
            valid = readParms(spath);
        }
        
        /** returns all parameters needed by jvm java command */
        public String getCommand() {
            return parameters;
        }
        
        /** is jvm parameters file in valid stucture */
        public boolean isValid() {
            return valid;
        }
        
        /** read jvm parameters from jvm parameters file */
        private boolean readParms(String spath) {
            Properties details = new Properties();
            FileInputStream fis = null;
            try {
                details.load(fis = new FileInputStream(spath)); // NOI18N
            } catch (IOException e) {            
                return false;
            } finally {
                if (fis != null) try { fis.close(); } catch (IOException e) { /* ignore */ };
            }
            
            String mainclass;
            String classpath;
            String jvmparms;
            String mainargs;        
        
            classpath = details.getProperty(PAR_RELCP,null);
            if (classpath != null)
                parameters = ";" + classpath;
            
            jvmparms = details.getProperty(PAR_JVMPAR,null);
            if (jvmparms != null)
                parameters = parameters + " " + jvmparms;
            
            mainclass = details.getProperty(PAR_MAIN,null);
            if (mainclass == null)
                return false;
            else
                parameters = parameters + " " + mainclass;
            
            mainargs = details.getProperty(PAR_MAINARGS,null);
            if (mainargs != null)
                parameters = parameters + " " + mainargs;

            parameters = replaceAll(parameters,VAR_IDE_HOME,
                new File (System.getProperty ("netbeans.home") ).getPath());                
            parameters = replaceAll(parameters,VAR_IDE_USER,
                ModuleUpdater.this.getNetbeansDirectory().getPath());            
            parameters = replaceAll(parameters,VAR_FILE_SEPARATOR,
                ModuleUpdater.this.FILE_SEPARATOR);            
            parameters = replaceAll(parameters,VAR_JAVA_HOME,
                System.getProperty ("java.home"));
            parameters = replaceAll(parameters,VAR_ICON_ROOT,
                getIconRoot());            
            return true;            
        }
        
        private String getIconRoot() {
            Properties prop = new Properties();
            FileInputStream fis = null;
            String ret;
            try {
                File propf = new File (System.getProperty ("netbeans.home") 
                        + FILE_SEPARATOR + "system" + FILE_SEPARATOR + "product.properties");                
                prop.load(fis = new FileInputStream(propf)); // NOI18N
                ret = prop.getProperty("start.menu.icon.root");
                if ( ret == null )
                    ret = getDefaultIconRoot();
            } catch (IOException e) {            
                ret = getDefaultIconRoot();
            } finally {
                if (fis != null) try { fis.close(); } catch (IOException e) { /* ignore */ };
            }
            return ('"' + ret + '"');
        }
        
        private String getDefaultIconRoot() {
            BufferedReader in = null;
            String ret = ICON_ROOT;
            String ext = "";
            String s;
            try {
                File brandf = new File (System.getProperty ("netbeans.home") 
                        + FILE_SEPARATOR + "lib" + FILE_SEPARATOR + "branding");                
                in = new BufferedReader(new FileReader(brandf));
                if (in.ready()) {                    
                    s = in.readLine();
                    if ( s.indexOf("_ce") > 0 )
                        ext = " CE";
                    else if ( s.indexOf("_ie") > 0 )
                        ext = " IE";
                    else if ( s.indexOf("_ee") > 0 )
                        ext = " EE";
                }
            } catch (IOException e) {                            
            } finally {
                if (in != null) try { in.close(); } catch (IOException e) { /* ignore */ };
            }
            return ret + ext;
        }
        
        /** replace all occurences of String what by String repl in the String sin */
        private String replaceAll(String sin, String what, String repl) {
            StringBuffer sb = new StringBuffer(sin);
            int i = sb.toString().indexOf(what);
            int len = what.length();
            while ( i > -1 ) {
                sb.replace(i,i + len,repl);
                i = sb.toString().indexOf(what,i+1);                
            }

            return sb.toString();
        }
    }
    
    static class LogWriter extends Object {
            
        private FileWriter log = null;

        public LogWriter () {
            try {
                log = getLogWriter();
            } catch (IOException e) {
            }
        }

        private FileWriter getLogWriter() throws IOException {                
            File logFile = new File(getUpdateDirectory() + FILE_SEPARATOR + LOG_FILE_NAME);
            if ((logFile.exists() && !logFile.canWrite()) || logFile.isDirectory()) {
                throw new IOException("Cannot write to " + LOG_FILE_NAME);  // NOI18N
            }        
            return new FileWriter(logFile.getAbsolutePath(),true);
        }

        void println(String line) {
            if (log != null) {
                try {
                    log.write(line);
                    log.write('\n');
                } catch (IOException e) {
                }
            }                
        }

        void close() {
            if (log != null) {
                try {
                    log.flush();
                    log.close();
                } catch (IOException e) {
                }
            }
        }

    }

    /*
    void cancel() {
      suspend = true;
      
      StopWarning stopWarning = new StopWarning();
      stopWarning.show();
     
      stop = stopWarning.isStop();
      suspend = false; 
}
    */

}