/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.autoupdate;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ResourceBundle;

import org.openide.TopManager;
import org.openide.modules.ModuleDescription;
import org.openide.util.NbBundle;

import org.netbeans.core.UpdateSupport;


/** This singleton class is a skeleton for running autoupdate, it also
 * contains all communication with core of IDE implementation. This
 * communication is handled by org.netbeans.core.UpdateSupport
 * on the side of IDE Core implementation.
 *
 * @author  Petr Hrebejk
 */
public class Autoupdater extends Object {

    /** Resource bundle */
    static ResourceBundle bundle = NbBundle.getBundle(Autoupdater.class);

    /** Is the autoupdate running ? */
    private static boolean isRunning = false;

    /** Getter for static property isRunning */
    static boolean isRunning() {
        return isRunning;
    }

    /** Setter for static property isRunning */
    static void setRunning( boolean isRunning ) {
        Autoupdater.isRunning = isRunning;
    }

    /** Restarts IDE in order to run Updater */
    static void restart() {
        UpdateSupport.restart();
    }

    /** Installs autoupdate checker */
    static void installUpdateChecker( UpdateSupport.UpdateChecker updateChecker ) {
        UpdateSupport.installUpdateChecker( updateChecker );
    }

    /** Gets proxy usage */
    static boolean isUseProxy() {
        return UpdateSupport.isUseProxy();
    }

    /** Gets Proxy Host */
    static String getProxyHost() {
        return UpdateSupport.getProxyHost();
    }

    /** Gets Proxy Port */
    static String getProxyPort() {
        return UpdateSupport.getProxyPort();
    }

    /** Sets the whole proxy configuration */
    static void setProxyConfiguration( boolean use, String host, String port ) {
        UpdateSupport.setProxyConfiguration( use, host, port );
    }

    /** Singleton support class for getting directories */
    static class Support extends Object {

        /** Platform dependent file name separator */
        private static final String FILE_SEPARATOR = System.getProperty ("file.separator");
        
        /** Relative name of update directory */
        private static final String UPDATE_DIR = "update"; // NOI18N
        
        /** Relative name of directory where the .NBM files are downloaded */
        private static final String DOWNLOAD_DIR =UPDATE_DIR + FILE_SEPARATOR + "download"; // NOI18N
        
        /** Relative name of directory where the patch files are stored */
        private static final String PATCH_DIR = "lib" + FILE_SEPARATOR + "patches"; // NOI18N

        /** Relative name of key store file */
        private static final String KS_FILE = "lib" + FILE_SEPARATOR + "ide.ks"; // NOI18N        
        
        /** The name of the log file */
        public static final String LOG_FILE_NAME = "update.log"; // NOI18N
        
        /** File representing the download directory */
        private static File downloadDirectory = null;

        /** File representing the update directory */
        private static File updateDirectory = null;

        /** File representing the central patch directory */
        private static File centralPatchDirectory = null;

        /** File representing the users patch directory */
        private static File userPatchDirectory = null;

        /** The file with cenral KeySotre */
        private static File centralKSFile = null;

        /** The file with users KeySotre */
        private static File userKSFile = null;

        /** Disable instantiation */
        private Support() {}

        /** Retruns array of module descriptions of installed modules */
        public static ModuleDescription[] getModuleDescriptions() {
            return UpdateSupport.getModuleDescriptions();
        }        

        /** The directory where to download the distribution files of modules */
        public static File getDownloadDirectory() {
            if ( downloadDirectory == null ) {

                if ( isMultiuser() )
                    downloadDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + DOWNLOAD_DIR );
                else
                    downloadDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + DOWNLOAD_DIR );

                if ( !downloadDirectory.isDirectory() )
                    downloadDirectory.mkdirs();
            }

            return downloadDirectory;
        }
        
        /** The directory where to download the distribution files of modules */
        public static File getUpdateDirectory() {
            if ( updateDirectory == null ) {

                if ( isMultiuser() )
                    updateDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + UPDATE_DIR );
                else
                    updateDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + UPDATE_DIR );

                if ( !updateDirectory.isDirectory() )
                    updateDirectory.mkdirs();
            }

            return updateDirectory;
        }

        /** Decides about running multiuser */
        private static boolean isMultiuser() {

            if ( System.getProperty ("netbeans.user") == null ||
                    System.getProperty ("netbeans.home").equals( System.getProperty ("netbeans.user") ) )
                return false;
            else
                return true;
        }

        /** Gets the central directory of patches */
        public static File getCentralPatchDirectory() {
            if ( centralPatchDirectory == null ) {
                centralPatchDirectory = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + PATCH_DIR );
            }
            return centralPatchDirectory;
        }


        /** Gets the users directory of patches */
        public static File getUserPatchDirectory() {
            if ( userPatchDirectory == null ) {
                userPatchDirectory = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + PATCH_DIR );
            }
            return userPatchDirectory;
        }


        /** Gets the central keystore */
        static File getCentralKSFile() {
            if ( centralKSFile == null ) {
                centralKSFile = new File (System.getProperty ("netbeans.home") + FILE_SEPARATOR + KS_FILE );
            }

            return centralKSFile;
        }

        /** Gets the users keystore */
        static File getUserKSFile() {
            if ( userKSFile == null ) {
                userKSFile = new File (System.getProperty ("netbeans.user") + FILE_SEPARATOR + KS_FILE );
            }

            return  userKSFile;
        }

        static class LogWriter extends Object {
            
            private FileWriter log = null;
            
            public LogWriter () {
                try {
                    log = getLogWriter();
                } catch (IOException e) {
                }
            }
         
            private FileWriter getLogWriter() throws IOException {                
                File logFile = new File(getUpdateDirectory() + FILE_SEPARATOR + LOG_FILE_NAME);
                if ((logFile.exists() && !logFile.canWrite()) || logFile.isDirectory()) {
                    throw new IOException("Cannot write to " + LOG_FILE_NAME);  // NOI18N
                }        
                return new FileWriter(logFile.getAbsolutePath(),true);
            }
            
            void println(String line) {
                if (log != null) {
                    try {
                        log.write(line);
                        log.write('\n');
                    } catch (IOException e) {
                    }
                }                
            }
            
            void close() {
                if (log != null) {
                    try {
                        log.flush();
                        log.close();
                    } catch (IOException e) {
                    }
                }
            }
            
        }
        
        /** Test whether the user has rights to write into directory */
        /*
        private static boolean canWrite( File dir, boolean create ) {
          if ( !dir.exists() && create ) 

            dir.mkdirs();

          if ( !dir.isDirectory() || !dir.canWrite() )
            return false;



          File tmp = null;

          try {
            tmp = File.createTempFile( "test", "access", dir ); // NOI18N
          }
          catch ( java.io.IOException e ) {
            return false;
          }

          if ( tmp == null )
            return false;

          boolean cw = tmp.canWrite();
          if (cw)
            tmp.delete();

          return cw;
    }
        */        

    }


}

/*
 * Log
 *  19   Gandalf   1.18        3/20/00  Milan Kubec     new version of 
 *       autoupdate 1.0
 *  18   Gandalf   1.17        3/8/00   Petr Hrebejk    URL version raised
 *  17   Gandalf   1.16        2/23/00  Petr Hrebejk    Notifications added into
 *       autoupdate
 *  16   Gandalf   1.15        1/13/00  Petr Hrebejk    Multiuser bugfix
 *  15   Gandalf   1.14        1/12/00  Petr Hrebejk    i18n
 *  14   Gandalf   1.13        1/12/00  Petr Hrebejk    Regnum can be changed at
 *       any time
 *  13   Gandalf   1.12        1/10/00  Petr Hrebejk    Version for finding XML 
 *       raised
 *  12   Gandalf   1.11        1/9/00   Petr Hrebejk    Support for i18n added
 *  11   Gandalf   1.10        1/9/00   Petr Hrebejk    Proxy Config and 
 *       Registration number added
 *  10   Gandalf   1.9         12/16/99 Petr Hrebejk    Sign checking added
 *  9    Gandalf   1.8         12/1/99  Petr Hrebejk    Checkin signatures of 
 *       NBM files & automatic autoupdate check added
 *  8    Gandalf   1.7         11/9/99  Petr Hrebejk    Better selection of nbm 
 *       files
 *  7    Gandalf   1.6         11/1/99  Petr Hrebejk    Remove of 
 *       org.netbeans.core.ModuleUpdater fixed
 *  6    Gandalf   1.5         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun
 *       Microsystems Copyright in File Comment
 *  5    Gandalf   1.4         10/11/99 Petr Hrebejk    Last minute fixes
 *  4    Gandalf   1.3         10/11/99 Petr Hrebejk    Version before Beta 5
 *  3    Gandalf   1.2         10/10/99 Petr Hrebejk    AutoUpdate made to 
 *       wizard
 *  2    Gandalf   1.1         10/8/99  Petr Hrebejk    Next development version
 *  1    Gandalf   1.0         10/7/99  Petr Hrebejk    
 * $
 */
