/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.autoupdate;

import java.net.URL;

import org.openide.TopManager;
import org.openide.util.NbBundle;

/** Options autoupdate module
*
* @author Ales Kemr
*/
public class XMLAutoupdateType extends AutoupdateType {

    final static String PROP_URL = "url";
    
    /** Holds value of property URL. */
    private URL URL;
    
    /** Holds value of property defaultURL. */
    protected String defaultURL;
    
    /** System property holding autoupdate version */
    private static final String UPDATE_VERSION_PROP = "netbeans.autoupdate.version"; // NOI18N

    /** Current version of autoupdate */
    private static final String UPDATE_VERSION = "1.5"; // NOI18N

    /** Important system properies */
    private static final String SYSPROP_COUNTRY = "netbeans.autoupdate.country"; // NOI18N
    private static final String SYSPROP_LANGUAGE = "netbeans.autoupdate.language"; // NOI18N
    private static final String SYSPROP_VARIANT = "netbeans.autoupdate.variant"; // NOI18N

    /** serialVersionUID */
    static final long serialVersionUID = 362844553432169452L;
    
    /** Creates new XMLAutoupdateType */
    public XMLAutoupdateType() {
        try {
            URL = new URL( getDefaultURL() );
        }
        catch (java.net.MalformedURLException e) {
            URL = null;
        }
    }
    
    /** @return human presentable name */
    public String displayName() {
        return NbBundle.getBundle( Settings.class).getString("CTL_XMLAutoupdateType_Name");
    }
    
    /** Getter for property URL.
     * @return Value of property URL.
    */
    public URL getURL() {
        return URL;
    }
    
    /** Setter for property URL.
     * @param URL New value of property URL.
    */
    public void setURL(URL URL) {
        URL old = this.URL;
        this.URL = URL;
        firePropertyChange( PROP_URL, old, URL );
    }
    
    public Updates connectForUpdates() {
        return new XMLUpdates( modifyURL( URL ) );
    }
    
    /** Gets the update URL */
    URL modifyURL (URL original) {
        
        URL updateURL = null;
        
        if ( System.getProperty( UPDATE_VERSION_PROP ) == null ) {
            System.setProperty( UPDATE_VERSION_PROP, UPDATE_VERSION );
        }

        try {
            updateURL = new URL( replace( original.toString() ) );
        }
        catch (java.net.MalformedURLException e) {
            TopManager.getDefault ().notifyException (e);
        };

        //System.out.println( updateURL );

        return updateURL;
        
    }
    
    /** Utility method for replacing {$xxx} with value of system property xxx,
     * or with another special value - see getReplacement
    */

    protected String replace( String string ) {

        // First of all set our system properties
        setSystemProperties();

        if ( string == null )
            return null;

        StringBuffer sb = new StringBuffer();

        int index, prevIndex;
        index = prevIndex = 0;
        while( ( index = string.indexOf( "{", index )) != -1 && index < string.length() - 1) { // NOI18N

            if ( string.charAt( index + 1 ) == '{' || string.charAt( index + 1 ) != '$'  ) {
                ++index;
                continue;
            }

            sb.append( string.substring( prevIndex, index ) );
            int endBracketIndex = string.indexOf( "}", index ); // NOI18N
            if ( endBracketIndex != -1 ) {
                String whatToReplace = string.substring( index + 2, endBracketIndex );
                sb.append( getReplacement(whatToReplace) );
            }
            prevIndex = endBracketIndex == -1 ? index + 2 : endBracketIndex + 1;
            ++index;
        }

        if ( prevIndex < string.length() - 1 )
            sb.append( string.substring( prevIndex ) );

        return sb.toString();
    }

    /** What to replace by what in URL string         
    */
    protected String getReplacement(String whatToReplace) {        
        return System.getProperty( whatToReplace, ""  );
    }

    private static void setSystemProperties() {
            
        if ( System.getProperty( SYSPROP_COUNTRY, null ) == null ) {
            System.setProperty( SYSPROP_COUNTRY, java.util.Locale.getDefault().getCountry() );
        }
        if ( System.getProperty( SYSPROP_LANGUAGE, null ) == null ) {
            System.setProperty( SYSPROP_LANGUAGE, java.util.Locale.getDefault().getLanguage() );
        }
        if ( System.getProperty( SYSPROP_VARIANT, null ) == null ) {
            System.setProperty( SYSPROP_VARIANT, java.util.Locale.getDefault().getVariant() );
        }
    }
    
    protected String getDefaultURL() {
        if ( defaultURL == null )
            defaultURL = NbBundle.getBundle( XMLAutoupdateType.class).getString( "URL_Default_N" );
        return defaultURL;        
    }
    
    private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException {
        if ( URL.toString() == getDefaultURL() )
            out.writeObject( null );        
        else
            out.writeObject( URL );        
    }

    private void readObject(java.io.ObjectInputStream in) throws java.io.IOException,
            ClassNotFoundException {
        URL u = (URL)in.readObject();
        if ( u == null || u.toString().startsWith("http://www.netbeans.org/updates/31_" ))  // NOI18N
            setURL(new URL( getDefaultURL() ));
        else
            setURL(u);        
    }
    
}
