/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.beans.PropertyVetoException;
import java.io.*;
import java.util.Enumeration;

import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.loaders.ExtensionList;
import org.openide.util.RequestProcessor;

/** Enhanced version of local file system that also takes care about
* backuping of saving files.
*
* @author  Jaroslav Tulach
*/
public final class ExLocalFileSystem extends LocalFileSystem implements Runnable {
    /** serial UID */
    private static final long serialVersionUID = -6117993653210115798L;

    /** default extensions to use for the file system */
    private static final ExtensionList DEFAULT_EXTENSIONS = new ExtensionList ();
    static {
        DEFAULT_EXTENSIONS.addExtension("java"); // NOI18N
    }

    /** extension for backuped files */
    private static final String BACKUP_EXT = System.getProperty
("org.netbeans.core.ExLocalFileSystem.BACKUP_EXT", "~");// NOI18N


    /** for which extensions create backup? */
    private ExtensionList backupExtensions;
    
    /** regexp of ignorable children */
    private String ignoredFiles = ""; // NOI18N
    
    /** regexp matcher for ignoredFiles, null if not needed */
    private transient RE ignoredRE = null;

    /** relative name to netbeans.user */
    private String relativeName;

    /** Before modification is done backup, if necessary */    
    public OutputStream outputStream (String name) throws IOException {
        backupIfNecessary (name);
        return super.outputStream (name);
    }
    
    /**
     * Before delete is done backup, if necessary
     *
     * @param name name of file
     * @exception IOException if the file could not be deleted
     */
    protected void delete(String name) throws IOException {
        backupIfNecessary (name);        
        super.delete(name);
    }
    
    /** Backups the file if necessary */    
    private void backupIfNecessary (String name)  throws IOException  {
        ExtensionList b = (backupExtensions == null) ? DEFAULT_EXTENSIONS : backupExtensions;
                     
        if ( b.isRegistered (name) ) {                
            InputStream is = null;
            try {
                is = super.inputStream(name);                
                OutputStream os = super.outputStream (name + BACKUP_EXT);
                try {
                    FileUtil.copy (is, os);
                } finally {
                    os.close ();
                }
            } catch (FileNotFoundException fex) {
                 /*not important: File was probably moved or removed*/ //I18N
            } finally {
                if (is != null) is.close ();
            }
        }        
    }

    /** Filters the backuped files.
    */
    public String[] children (String name) {
        String[] arr = super.children (name);

        if (arr == null) {
            return null;
        }

        int j = 0;
        for (int i = 0; i < arr.length; i++) {
            if (arr[i].endsWith (BACKUP_EXT) || (ignoredRE != null && ignoredRE.match (arr[i]))) {
                arr[i] = null;
            }
        }

        return arr;
    }

    /** Set the number of milliseconds between automatic
    * refreshes of the directory structure.
    *
    * @param ms number of milliseconds between two refreshes; if <code><= 0</code> then refreshing is disabled
    */    
    public  void setRefreshTime2 (int ms) {
        super.setRefreshTime (ms);
    }

    /** Get the number of milliseconds between automatic
    * refreshes of the directory structure.
    * By default, automatic refreshing is disabled.
    * @return the number of milliseconds, or <code>0</code> if refreshing is disabled
    */        
    public  int getRefreshTime2 () {
        return super.getRefreshTime ();
    }
    
    
    //
    // properties
    //

    /** Setter disabling/enabling backup.
    */
    public void setBackupExtensions (ExtensionList l) {
        backupExtensions = l;
        firePropertyChange ("backupExtensions", null, null); // NOI18N
    }

    /** @return true if backup of files is enabled.
    */
    public ExtensionList getBackupExtensions () {
        if (backupExtensions == null) {
            return (ExtensionList)DEFAULT_EXTENSIONS.clone ();
        }
        return backupExtensions;
    }
    
    public String getIgnoredFiles () {
        return ignoredFiles;
    }
    
    public synchronized void setIgnoredFiles (String nue) throws IllegalArgumentException {
        if (! nue.equals (ignoredFiles)) {
            if (nue.length () > 0) {
                try {
                    ignoredRE = new RE (nue);
                } catch (RESyntaxException rese) {
                    IllegalArgumentException iae = new IllegalArgumentException ();
                    TopManager.getDefault ().getErrorManager ().annotate (iae, rese);
                    throw iae;
                }
            } else {
                ignoredRE = null;
            }
            ignoredFiles = nue;
            firePropertyChange ("ignoredFiles", null, null); // NOI18N
            RequestProcessor.postRequest (this);
        }
    }
    
    public void run () {
        Enumeration e = existingFileObjects (getRoot ());
        while (e.hasMoreElements ()) {
            FileObject fo = (FileObject) e.nextElement ();
            if (fo.isFolder ()) {
                fo.refresh (true);
            }
        }
    }


    private boolean relativeFlag = false;

    /** Relative file name to netbeans.user
    *
    * @param relativeName the relative file name
    */
    public void setRelativeDirectory (String relativeName)
    throws java.beans.PropertyVetoException, IOException {

        String oldRelative = this.relativeName;
        try {
            this.relativeName = relativeName;
            relativeFlag = true;

            String file =
                System.getProperty ("netbeans.user") +
                File.separatorChar +
                relativeName;
            File f = new File (file);

            try {
                f = f.getCanonicalFile();
            } catch (Exception ex) {
                // not important
            }

            setRootDirectory(f);
        } catch (PropertyVetoException ex) {
            this.relativeName = oldRelative;
            throw ex;
        } catch (IOException ex) {
            this.relativeName = oldRelative;
            throw ex;
        } finally {
            relativeFlag = false;
        }
    }

    /** Getter for relative name to netbeans.user or null
    * @return the name or null if not set
    */
    public String getRelativeDirectory () {
        return relativeName;
    }

    /** Compute the system name of this file system for a given root directory.
    * <P>
    * The default implementation simply returns the filename separated by slashes.
    * @see FileSystem#setSystemName
    * @param rootFile root directory for the filesystem
    * @return system name for the filesystem
    */
    protected String computeSystemName (File rootFile) {
        if (relativeName != null) {
            return "{netbeans.user}" + '/' + relativeName; // NOI18N
        } else return super.computeSystemName(rootFile);
    }

    /** Set the root directory of the file system.
    * @param r file to set root to
    * @exception PropertyVetoException if the value if vetoed by someone else (usually
    *    by the {@link org.openide.filesystems.Repository Repository})
    * @exception IOException if the root does not exists or some other error occured
    */
    public synchronized void setRootDirectory (File r) throws PropertyVetoException, IOException {
        String oldRelative = relativeName;
        if (!relativeFlag) relativeName = null;
        try {
            super.setRootDirectory(r);
        } catch (PropertyVetoException ex) {
            relativeName = oldRelative;
            throw ex;
        } catch (IOException ex) {
            relativeName = oldRelative;
            throw ex;
        }
    }

    /** Read object refreshes the relative name.
    */
    private void readObject (ObjectInputStream ois)
    throws IOException, ClassNotFoundException {
        ois.defaultReadObject ();
        if (ignoredFiles == null) {
            ignoredFiles = ""; // NOI18N
        } else if (ignoredFiles.length () > 0) {
            try {
                ignoredRE = new RE (ignoredFiles);
            } catch (RESyntaxException rese) {
                TopManager.getDefault ().notifyException (rese);
            }
        }
        if (relativeName != null) {
            try {
                setRelativeDirectory (relativeName);
            } catch (java.beans.PropertyVetoException ex) {
            }
        }
    }
    
    
}
