/*
 * FilterList.java
 *
 * Created on 27. duben 2000, 11:12
 */
 
package org.netbeans.core;

import java.util.*;

/** Implementation of list interface that delegates to
* different list and filters its content.
* @author Jaroslav Tulach
*/
abstract class FilterList extends AbstractList 
implements java.beans.PropertyChangeListener {
    /** list to delegate to */
    private List delegate;
    /** array to map the public index to the index in the delate list */
    private int[] map;

    /** Creates new FilterList */
    public FilterList(List delegate) {
      this.delegate = delegate;
    }

    /** Method to filter the content of the delegate.
    * @param obj the object to check
    * @return true if the object should be included in this list
    */
    protected abstract boolean check (Object obj);

    /** When property change arrives, clears mapping.
    */
    public void propertyChange (java.beans.PropertyChangeEvent ev) {
        map = null;
    }

    /** Adds the object at given possition.
    */
    public void add(int p1, Object p2) {
      if (map ().length == p1) {
        // adds at the end
        delegate.add (p2);
      } else {
        // adds at given position
        delegate.add (map ()[p1], p2);
      }
      map = null;
    }

    /** Adds the collection at given possition. Faster implementation than the
    * one provided by abstract list.
    */
    public boolean  addAll (int p1, Collection p2) {
      if (p1 == map ().length) {
        // simply add at the end
        return addAll (p2);
      }
      try {
        return delegate.addAll (map ()[p1], p2);
      } finally {
        // clean the mapping
        map = null;
      }
    }

    /** Adds the collection at given possition. Faster implementation than the
    * one provided by abstract list.
    */
    public boolean  addAll (Collection p2) {
      try {
        return delegate.addAll (p2);
      } finally {
        // clean the mapping
        map = null;
      }
    }

    /** Finds the right possition in the original list and 
    * sets the object on it.
    * @param p1 index into the list
    * @param p2 the object to place at the position
    */
    public Object set(int p1, Object p2) {
      return delegate.set (map ()[p1], p2);
    }

    /** The size of the array.
    */
    public int size() {
      return map ().length;
    }

    /** Removes the object at specified index.
    */
    public Object remove(int p1) {
      try {
        return delegate.remove (map ()[p1]);
      } finally {
        map = null;
      }
    }

    /** Getter for the object of specified index.
    */
    public Object get(int p1) {
      return delegate.get (map ()[p1]);
    }

    //
    // mapping info
    //

    /** Returns the map array.
    */
    private int[] map () {
      int[] localMap = map;  
      if (localMap != null) return localMap;

      int index = 0;
      List mapList = new LinkedList ();
      Iterator it = delegate.iterator();
      while (it.hasNext()) {
        Object obj = it.next ();
        if (check (obj)) {
          // include this object in this list
          mapList.add (new Integer (index));
        }
        index++;
      }

      // copy mapList to array
      int[] arr = new int[mapList.size ()];

      it = mapList.iterator();
      index = 0;
      while (it.hasNext()) {
        arr[index++] = ((Integer)it.next ()).intValue ();
      }

      map = arr;
      return arr;
    }


/*
  public static void main (String[] args) {
    List arr = Arrays.asList(args);
    List ll = new LinkedList (arr);

    FilterList l = new FilterList (ll) {
      protected boolean check (Object o) {
        return ((String)o).length() > 3;
      }
    };


    Collections.sort(l);
    l.add (0, "Kukuc");
    l.remove (1);

    System.out.println("args: " + arr);
    System.out.println("4444: " + l);
    System.out.println("list: " + ll);
  }
*/
}