/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.io.*;
import java.lang.ref.*;
import java.util.*;
import java.net.*;
import javax.swing.event.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.execution.NbfsStreamHandlerFactory;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

/** Reference implementation of the helpfiles of JavaHelp.
* Also holds all installed helpIDs and can display a 
* help page when such HelpCtx should be opened.
*
* @author Jaroslav Tulach
*/
public abstract class Help extends Object {
    private Help () {}
    /** instance of the Help */
    private static Impl help;

    /** Finds default instance of the help.
    */
    public static Impl getDefault () {
        if (help == null) {

            try {
                Class.forName ("javax.help.HelpSet"); // NOI18N
                try {
                    // Avoid making any static references to javax.help.*
                    // (just new JavaHelp() should work, but this is a little safer):
                    help = (Impl) Class.forName ("org.netbeans.core.JavaHelp").newInstance (); // NOI18N
                } catch (Exception e) {
                    // These would actually be unexpected.
                    TopManager.getDefault ().notifyException (e);
                    help = new XMLImpl ();
                }
            } catch (ClassNotFoundException cnfe) {
                // Fine, JavaHelp not installed.
                help = new XMLImpl ();
            }

        }
        return help;
    }

    /** Methods delegated to different implementations.
    */
    public static interface Impl {
        /** Adds help set URL into the system */
        public void addHelpSet (URL resource, String moduleCodeName, String moduleDisplayName, ClassLoader cl);

        /** Removes help set URL from the system */
        public void removeHelpSet (URL resource);

        /** Shows help.
        */
        public void showHelp (HelpCtx ctx);

        /** Get a map from module code names to home IDs. */
        public Map getHomesByCode ();

        /** Get a map from module display names to home IDs. */
        public Map getHomesByDisplay ();

        /** Display name for the master index, or <code>null</code> if none. */
        public String getMasterDisplayName ();

        /** Help ID for the master index, or <code>null</code> if none. */
        public String getMasterID (boolean clever);

        /** Test whether a given homeID is particularly distinguished, so should be displayed first.
        * @param id the home ID
        * @return true if it should be considered distinguished
        */
        public boolean isDistinguished (String homeID);

        /** Add a change listener for when help sets change.
         * @param l the listener to add
         */
        public void addChangeListener (ChangeListener l);

        /** Remove a change listener.
         * @param l the listener to remove
         */
        public void removeChangeListener (ChangeListener l);
    }

    // --- DOCUMENTATION CLASSLOADER AND URLS ---

    // [PENDING] probably createDocsURL should be part of Impl...but in fact XMLImpl and JavaHelp
    // would implement it exactly the same way, so don't bother for now.

    /** Create a URL to load documentation with.
     * Pass in a resource path to a helpset in the base locale,
     * and get a URL capable of loading that helpset (be it in some module JAR
     * or in the docs/ directory) in the correct locale.
     * Note that a valid resource will always produce some URL, which may or may
     * not actually be able to load the helpset; try connecting to tell for sure.
     * @param resource a classloader-style resource path to a helpset in the base locale
     * @return a URL capable of loading such documentation
     */
    public static URL createDocsURL (String resource) {
        initStreamHandler ();
        try {
            return new URL ("nbdocs:/" + resource); // NOI18N
        } catch (MalformedURLException mfue) {
            InternalError ie = new InternalError ("Could not create nbdocs: URL"); // NOI18N
            TopManager.getDefault ().getErrorManager ().annotate (ie, mfue);
            throw ie;
        }
    }

    private static boolean streamHandlerInited = false;

    private static void initStreamHandler () {
        if (! streamHandlerInited) {
            synchronized (Help.class) {
                if (! streamHandlerInited) {
                    NbfsStreamHandlerFactory.getDefault ().register ("nbdocs", new NbDocsStreamHandler ());
                    streamHandlerInited = true;
                }
            }
        }
    }

    /** Handler & connection cribbed from NbResourceStreamHandler. */
    private static final class NbDocsStreamHandler extends URLStreamHandler {

        public URLConnection openConnection (URL u) throws IOException {
            if (u.getProtocol ().equals ("nbdocs")) {
                return new NbDocsURLConnection (u);
            } else {
                throw new IOException ("mismatched protocol"); // NOI18N
            }
        }

    }

    private static final class NbDocsURLConnection extends URLConnection {

	private URLConnection real = null;
        private IOException exception = null;

        public NbDocsURLConnection (URL u) {
            super (u);
        }

	public synchronized void connect () throws IOException {
            if (exception != null) {
                IOException e = exception;
                exception = null;
                throw e;
            }
	    if (! connected) {
		String resource = url.getFile ();
		if (resource.startsWith ("/")) resource = resource.substring (1); //NOI18N
                // [PENDING] could probably be simplified (no need for a special ClassLoader at all)
                // if only NbBundle had a static method to enumerate locale suffixes...then could check
                // docsURLs directly, followed by nbresloc: URL or something.
		URL target;
                String ext, basename;
                int index = resource.lastIndexOf ('.');
                if (index != -1) {
                    ext = resource.substring (index + 1);
                    basename = resource.substring (0, index).replace ('/', '.');
                } else {
                    ext = null;
                    basename = resource.replace ('/', '.');
                }
                try {
                    target = NbBundle.getLocalizedFile (basename, ext, Locale.getDefault (), getDocsLoader ());
                } catch (MissingResourceException mre) {
                    throw (IOException) TopManager.getDefault ().getErrorManager ().annotate
                        (new IOException (NbBundle.getMessage (Help.class, "EXC_nbdocs_cannot_connect", url)), mre);
                }
		real = target.openConnection ();
                real.connect ();
		connected = true;
	    }
	}
        
        private void tryToConnect () {
            if (connected || exception != null) return;
            try {
                connect ();
            } catch (IOException ioe) {
                exception = ioe;
            }
        }

	public String getHeaderField (int n) {
            tryToConnect ();
	    if (connected)
		return real.getHeaderField (n);
	    else
		return null;
	}

	public String getHeaderFieldKey (int n) {
            tryToConnect ();
	    if (connected)
		return real.getHeaderFieldKey (n);
	    else
		return null;
	}

	public String getHeaderField (String key) {
            tryToConnect ();
	    if (connected)
		return real.getHeaderField (key);
	    else
		return null;
	}

	public InputStream getInputStream () throws IOException {
            connect ();
            return real.getInputStream ();
	}

	public OutputStream getOutputStream () throws IOException {
            connect ();
            return real.getOutputStream ();
	}
        
        public String getContentType () {
            tryToConnect ();
            if (connected)
                return real.getContentType ();
            else
                return "application/octet-stream"; // NOI18N
        }
        
        public int getContentLength () {
            tryToConnect ();
            if (connected)
                return real.getContentLength ();
            else
                return 0;
        }

    }

    private static Reference docsLoader = null; // Reference<ClassLoader>

    // Package-private only for debugging:
    static ClassLoader getDocsLoader () {
        Reference r = docsLoader;
        ClassLoader l;
        if (r != null) {
            l = (ClassLoader) r.get ();
        } else {
            l = null;
        }
        if (l == null) {
            l = new URLClassLoader (getDocsURLs (), TopManager.getDefault ().systemClassLoader ());
            docsLoader = new WeakReference (l);
        }
        return l;
    }

    // Called from ModuleInstaller when systemClassLoader changes:
    static void resetDocsLoader () {
        docsLoader = null;
    }

    private static URL[] docsURLs = null;
    private static URL[] getDocsURLs () {
        if (docsURLs == null) {
            List urls = new ArrayList (2);
            String nbhome = System.getProperty ("netbeans.home");
            if (nbhome != null) {
                try {
                    // Java bug: if you do not canonicalize, it will make an absolute path,
                    // however file:/d:/nbdir/./docs/ is not treated as a valid URL for the
                    // URLClassLoader, for some reason.
                    URL url = new File (nbhome, "docs").getCanonicalFile ().toURL (); // NOI18N
                    if (! url.toString ().endsWith ("/")) // NOI18N
                        url = new URL (url.toString () + "/"); // NOI18N
                    urls.add (url);
                } catch (Exception e) {
                    TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, e);
                }
            }
            String nbuser = System.getProperty ("netbeans.user");
            if (nbuser != null && ! nbuser.equals (nbhome)) {
                try {
                    URL url = new File (nbuser, "docs").getCanonicalFile ().toURL (); // NOI18N
                    if (! url.toString ().endsWith ("/")) // NOI18N
                        url = new URL (url.toString () + "/"); // NOI18N
                    urls.add (url);
                } catch (Exception e) {
                    TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, e);
                }
            }
            Collections.reverse (urls); // nbuser docs should take precedence
            docsURLs = (URL[]) urls.toArray (new URL[urls.size ()]);
        }
        return docsURLs;
    }

    // --- END DOCUMENTATION CLASSLOADER AND URLS ---

    /** Simple implementation that takes only important tags from the HelpSet file.
    * Serves for two purposes:
    * <OL>
    *   <LI>Implementation of Help.Impl
    *   <LI>DocumentHandler callback for XML
    * </OL>
    */
    private static final class XMLImpl extends HandlerBase implements Impl {
        /** tags in map file */
        private static final String MAP_ID = "mapID"; // NOI18N
        private static final String TARGET = "target"; // NOI18N
        private static final String URL = "url"; // NOI18N

        /** important tags in help set file */
        private static final String MAPS = "maps"; // NOI18N
        private static final String HOME_ID = "homeID"; // NOI18N
        private static final String MAPREF = "mapref"; // NOI18N
        private static final String LOCATION = "location"; // NOI18N

        // [PENDING] these should not be static--this class should be
        // untangled so they do not have to be

        /** map of URL of help set to map of string ids to string URLs.
        * Type (URL, Map (String, String)).
        */
        private static Map help = new HashMap (7);

        /** map of helpset URLs to code names */
        private static Map moduleCodeNames = new HashMap (); // Map<URL, String>

        /** map of helpset URLs to display names */
        private static Map moduleDisplayNames = new HashMap (); // Map<URL, String>

        /** map of code names to home IDs */
        private static Map homesByCode = new HashMap (); // Map<String, String>

        /** map of display names to home IDs */
        private static Map homesByDisplay = new HashMap (); // Map<String, String>

        public synchronized Map getHomesByCode () {
            return new HashMap (homesByCode);
        }

        public synchronized Map getHomesByDisplay () {
            return new HashMap (homesByDisplay);
        }

        /** buffer to contain the homeID or null if we are not reading homeID */
        private StringBuffer homeID;

        /** base URL for parsing file */
        private URL base;

        private Parser parser;

        /** map to currently work with (String, String) */
        private Map current;

        /** module info to currently work with */
        private String moduleCodeName, moduleDisplayName;

        /** Creates new XML parser/saver for HelpSet file.
        * This constructor is for the 
        */
        public XMLImpl () {
        }

        /** Creates new XML parser/saver for HelpSet file.
        */
        private XMLImpl (URL base, Map current, String moduleCodeName, String moduleDisplayName) {
            this.base = base;
            this.current = current;
            this.moduleCodeName = moduleCodeName;
            this.moduleDisplayName = moduleDisplayName;

            parser = org.openide.loaders.XMLDataObject.createParser ();
            parser.setDocumentHandler (this);
        }

        // [PENDING] synchronized??
        /** Adds help set URL into the system */
        public void addHelpSet (URL resource, String moduleCodeName, String moduleDisplayName, ClassLoader ignore) {
            //System.err.println ("Adding help set: " + resource + " " + moduleCodeName + " " + moduleDisplayName);
            try {
                HashMap map = new HashMap ();
                XMLImpl xml = new XMLImpl (resource, map, moduleCodeName, moduleDisplayName);
                xml.parse ();
                // store the map
                help.put (resource, map);
                moduleCodeNames.put (resource, moduleCodeName);
                moduleDisplayNames.put (resource, moduleDisplayName);
            } catch (SAXException e) {
                TopManager.getDefault ().notifyException (e);
            } catch (IOException e) {
                TopManager.getDefault ().notifyException (e);
            }
            fireChangeEvent ();
        }

        /** Removes help set URL from the system */
        public synchronized void removeHelpSet (URL resource) {
            help.remove (resource);
            homesByCode.remove (moduleCodeNames.remove (resource));
            homesByDisplay.remove (moduleDisplayNames.remove (resource));
            fireChangeEvent ();
        }

        private final Set listeners = new HashSet (1); // Set<ChangeListener>
        public void addChangeListener (ChangeListener l) {
            synchronized (listeners) {
                listeners.add (l);
            }
        }
        public void removeChangeListener (ChangeListener l) {
            synchronized (listeners) {
                listeners.remove (l);
            }
        }
        protected void fireChangeEvent () {
            Iterator it;
            synchronized (listeners) {
                it = new HashSet (listeners).iterator ();
            }
            ChangeEvent ev = new ChangeEvent (this);
            while (it.hasNext ()) {
                ((ChangeListener) it.next ()).stateChanged (ev);
            }
        }

        /** Shows help.
        */
        public synchronized void showHelp (HelpCtx ctx) {
            URL url = ctx.getHelp ();
            String s = ctx.getHelpID ();

            if (url == null) {
                // iterator over maps
                String x = null;
                Iterator it = help.values ().iterator ();
                while (x == null && it.hasNext ()) {
                    Map m = (Map)it.next ();
                    x = (String)m.get (s);
                }
                if (x != null) {
                    try {
                        url = new URL (x);
                    } catch (MalformedURLException e) {
                        // go on, but it should not happen
                    }
                }
            }

            if (url != null) {
                TopManager.getDefault ().showUrl (url);
            } else {
                TopManager.getDefault ().notify (new NotifyDescriptor.Message (Main.getString ("EXC_HelpIDNotFound", s)));
            }
        }

        /** Registration of the id */
        private void registerID (String id, URL url) {
            // saving the URL as strings could save some memory
            current.put (id.intern (), url.toExternalForm ().intern ());
        }

        /** If we have homeID buffer non-null we add the text to it.
        * Because it composes the homeID.
        */
        public void characters(char[] ch, int start, int length) throws SAXException {
            if (homeID != null) {
                homeID.append (ch, start, length);
            }
        }

        /** Accepts module item */
        public void startElement (String name, AttributeList attr) throws SAXException {
            if (MAP_ID.equals (name)) {
                String target = attr.getValue (TARGET);
                String url = attr.getValue (URL);
                if (target != null && url != null) {
                    try {
                        registerID (target, new URL (base, url));
                    } catch (MalformedURLException e) {
                        throw new SAXException ("Malformed help URL: " + url + " for target: " + target); // NOI18N
                    }
                }
                return;
            }
            if (HOME_ID.equals (name)) {
                homeID = new StringBuffer ();
                return;
            }
            if (MAPREF.equals (name)) {
                processMap (attr.getValue (LOCATION));
                return;
            }
        }

        /** Accepts module item */
        public void endElement (String name) {
            if (HOME_ID.equals (name)) {
                String id = homeID.toString ();
                //System.err.println ("Found home ID: " + id + " " + moduleCodeName + " " + moduleDisplayName);
                homesByCode.put (moduleCodeName, id);
                homesByDisplay.put (moduleDisplayName, id);
                homeID = null;
            }
        }

        /** Processes map file.
        */
        private void processMap (String url) throws SAXException {
            try {
                URL map = new URL (base, url);
                XMLImpl impl = new XMLImpl (map, current, moduleCodeName, moduleDisplayName);
                impl.parse ();
            } catch (IOException e) {
                throw new SAXException (e.getMessage ());
            }
        }

        /** Parses the HelpSet file.
        */
        public void parse () throws IOException, SAXException {
            parser.parse (base.toExternalForm ());
        }

        /** Display name for the master index, or <code>null</code> if none. */
        public String getMasterDisplayName() {
            return null;
        }
        /** Help ID for the master index, or <code>null</code> if none. */
        public String getMasterID(boolean clever) {
            return null;
        }
        /** Test whether a given homeID is particularly distinguished, so should be displayed first.
         * @param id the home ID
         * @return true if it should be considered distinguished
         */
        public boolean isDistinguished(String homeID) {
            return false;
        }
    }


}
