/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.*;
import java.beans.*;
import java.io.*;
import java.lang.ref.*;
import java.net.URL;
import java.util.*;
import java.util.List; // override java.awt.List
import java.util.Map; // override javax.help.Map
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.ComponentUI;

import org.openide.*;
import org.openide.awt.HtmlBrowser;
//import org.openide.text.CloneableEditorSupport;
import org.openide.util.*;
//import org.openide.util.io.SafeException;
import org.openide.windows.CloneableTopComponent;

import javax.help.*;
import javax.help.event.*;
import javax.help.plaf.HelpContentViewerUI;
import javax.help.plaf.basic.BasicHelpUI;

// [PENDING] should event dispatch thread be used thruout?

/** Help implementation using the JavaHelp 1.0 system.
* Should only be loaded if JavaHelp is installed in the class path.
* @author Jesse Glick
*/
class JavaHelp implements Help.Impl, ChangeListener {

    /** The default instance of the JavaHelp implementation; there
     *should not be more than one.
     */
    //private static JavaHelp DEFAULT =null;

    /** Make a JavaHelp implementation of the Help.Impl interface.
     *Or, use {@link #getDefaultJavaHelp}.
     */
    public JavaHelp() {
        //DEFAULT = this;
        NbPresenter.addChangeListener (this);
        err.log ("JavaHelp created");
    }

    /** Get the default instance of the JavaHelp implementation.
     * @return the default instance
     * /
    static JavaHelp getDefaultJavaHelp() {
        if (DEFAULT == null)
            return new JavaHelp ();
        else
            return DEFAULT;
    }*/

    /** logging for this class */
    private ErrorManager err = TopManager.getDefault ().getErrorManager ().getInstance ("org.netbeans.core.JavaHelp"); // NOI18N

    /** Base resource string to the dummy master help set.
     *Just contains a title and empty standard content navigators
     *permitting other help sets to be merged into it.
     */
    private static final String MASTER_RESOURCE ="org.netbeans.core.resources.masterHelpSet"; // NOI18N
    /** As MASTER_RESOURCE, but now the file extension. */
    private static final String MASTER_RESOURCE_EXT = "xml"; // NOI18N
    /** A helpID present only in the master help set;
     *however, when displayed by {@link #showHelp} as the helpID in a context,
     *the master help set (with merged-in children) will be shown instead,
     *with no change made to the content pane.
     *Also, this is the help ID mapped to the "default" page in the master help viewer.
     */
    private static final String MASTER_ID ="org.netbeans.core.JavaHelp.MASTER_ID"; // NOI18N

    /** Internal name of the window manager {@link Mode} to be used
     *by default for {@link HelpComponent}s.
     */
    //private static final String JAVAHELP_MODE ="JavaHelp"; // window manager mode // NOI18N

    /** Map from help set URLs to the help sets themselves.
     *Type <CODE>Map&lt;URL,HelpSet&rt;</CODE>.
     */
    private Map setsByUrl = new HashMap (); // Map<URL,HelpSet>
    /** Map from module code name to the home ID for that module.
     *Of type <CODE>Map&lt;String,String&gt;</CODE>.
     */
    private Map homesByCode = new HashMap (); // Map<String,String>
    /** Map from display names of modules (or the displayed name under Features,
     *often help set title) to the home IDs. Note that the display names
     *are guaranteed to be unique, using suffixes if needed.
     *Of type <CODE>Map&lt;String,String&gt;</CODE>.
     */
    private Map homesByDisplay =new HashMap (); // Map<String,String>
    /** Map from help set URLs to module code names.
     *Of type <CODE>Map&lt;URL,String&gt;</CODE>.
     */
    private Map codesByUrl =new HashMap (); // Map<URL,String>
    /** Map from help set URLs to module display names (or help set titles).
     *Of type <CODE>Map&lt;URL,String&gt;</CODE>.
     */
    private Map displaysByUrl =new HashMap (); // Map<URL,String>
    /** Set of home IDs which should be prominently displayed.
    * Of type <code>Set&lt;String&rt;</code>.
    */
    private Set distinguishedHomeIDs = new HashSet (); // Set<String>
    /** The master help set.
     */
    private HelpSet master =null;
    /** All distinguished help sets. */
    private Set distinguishedHelpSets = new HashSet (); // Set<HelpSet>
    /** All undistinguished help sets. */
    private Set undistinguishedHelpSets = new HashSet (); // Set<HelpSet>

    /** map from help sets to (soft refs to) components showing them */
    private Map availableJHelps = new HashMap (); // Map<HelpSet,Reference<JHelp>>
    /** viewer (may be invisible) showing help normally; null until first used; if invisible, is empty */
    private JFrame frameViewer = null;
    /** viewer showing help parented to current modal dialog; initially null */
    private JDialog dialogViewer = null;
    /** the last help set shown in the frame viewer, may be null */
    private HelpSet lastFramedHelpSet = null;
    /** the last help set shown in a modal dialog, may be null */
    private HelpSet lastDialogHelpSet = null;

    /** Get the master help set that others will be merged into.
     * @return the master help set
     */
    private HelpSet getMaster() {
        if (master == null) {
            synchronized (this) {
                if (master == null) {
                    ClassLoader loader = JavaHelp.class.getClassLoader ();
                    try {
                        master = new HelpSet (loader, NbBundle.getLocalizedFile
                                              (MASTER_RESOURCE, MASTER_RESOURCE_EXT, Locale.getDefault (), loader));
                    } catch (HelpSetException hse) {
                        err.notify (hse);
                        master = new HelpSet ();
                    }
                }
            }
        }
        return master;
    }

    /** Add a help set to the system from a module.
     * @param resource The URL to the help set (<CODE>.hs</CODE>) file.
     * @param moduleCodeName The module code name (mostly used to look up context
     *help from {@link ModuleNode.Item}).
     * @param moduleDisplayName The module's display name, used as a backup display
     *name for the help in case the <CODE>&lt;TITLE&rt;</CODE>
     *tag is missing.
     * @param cl A classloader which may be used to load classes for navigators etc.
     */
    public synchronized void addHelpSet(URL resource, String moduleCodeName, String moduleDisplayName, ClassLoader cl) {
        try {
            HelpSet sub = new HelpSet (cl, resource);
            String displayName = sub.getTitle ();
            boolean distinguished = displayName != null && displayName.endsWith ("*"); // NOI18N
            if (distinguished) {
                displayName = displayName.substring (0, displayName.length () - 1);
                sub.setTitle (displayName);
                err.log ("Found distinguished help set: " + displayName);
            }
            if (displayName == null || displayName.equals ("")) displayName = moduleDisplayName; // NOI18N
            if (homesByDisplay.containsKey (displayName)) {
                int i = 2;
                String test;
                while (homesByDisplay.containsKey (test = Main.getString ("LBL_HelpSet_Duplicate", displayName, String.valueOf (i)))) i++;
                displayName = test;
            }
            javax.help.Map.ID homeID = sub.getHomeID ();
            if (homeID != null) {
                homesByCode.put (moduleCodeName, homeID.id);
                homesByDisplay.put (displayName, homeID.id);
                if (distinguished) distinguishedHomeIDs.add (homeID.id);
            }
            setsByUrl.put (resource, sub);
            codesByUrl.put (resource, moduleCodeName);
            displaysByUrl.put (resource, displayName);
            HelpSet master = getMaster ();
            if (master == null) return; // shouldn't happen
            if (distinguished) {
                distinguishedHelpSets.add (sub);
                // Make sure it comes before any undistinguished help sets in merged navs.
                // [PENDING] in principle this could slow things down...does it?
                Iterator it = undistinguishedHelpSets.iterator ();
                while (it.hasNext ()) {
                    master.remove ((HelpSet) it.next ());
                }
            } else {
                undistinguishedHelpSets.add (sub);
            }
            master.add (sub);
            if (distinguished) {
                // Now add them back (at the end, hopefully).
                Iterator it = undistinguishedHelpSets.iterator ();
                while (it.hasNext ()) {
                    master.add ((HelpSet) it.next ());
                }
            }
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            // Only declared to throw HelpSetException but in fact all sorts of
            // nastiness seems possible with a sufficiently ugly help set--
            // e.g. InvocationTargetException, etc.
            err.annotate (t, Main.getString ("EXC_cannot_add_help_set", resource.toString (), moduleDisplayName));
            err.notify (t);
        }
        fireChangeEvent ();
    }

    /** Remove a help set from the system (i.e. when the module is uninstalled).
     * @param resource The URL to the help set to be removed.
     */
    public synchronized void removeHelpSet(URL resource) {
        String code = (String) codesByUrl.remove (resource);
        if (code != null) homesByCode.remove (code);
        String display = (String) displaysByUrl.remove (resource);
        if (display != null) {
            String home = (String) homesByDisplay.remove (display);
            if (home != null) distinguishedHomeIDs.remove (home);
        }
        HelpSet hs = (HelpSet) setsByUrl.remove (resource);
        if (hs != null) {
            getMaster ().remove (hs);
            undistinguishedHelpSets.remove (hs);
            distinguishedHelpSets.remove (hs);
            availableJHelps.remove (hs);
            if (lastFramedHelpSet == hs) {
                lastFramedHelpSet = null;
            }
            if (lastDialogHelpSet == hs) {
                lastDialogHelpSet = null;
            }
        }
        fireChangeEvent ();
    }

    private final Set listeners = new HashSet (1); // Set<ChangeListener>
    public void addChangeListener (ChangeListener l) {
        synchronized (listeners) {
            listeners.add (l);
        }
    }
    public void removeChangeListener (ChangeListener l) {
        synchronized (listeners) {
            listeners.remove (l);
        }
    }
    protected void fireChangeEvent () {
        Iterator it;
        synchronized (listeners) {
            it = new HashSet (listeners).iterator ();
        }
        ChangeEvent ev = new ChangeEvent (this);
        while (it.hasNext ()) {
            ((ChangeListener) it.next ()).stateChanged (ev);
        }
    }

    /** Get the master index name.
    * @return a display name
    */
    public String getMasterDisplayName () {
        return getMaster ().getTitle ();
    }

    /** Get the master ID.
    * Will instead return nothing if either:
    * <ol>
    * <li>There are no help sets installed, in which case the whole
    * thing would not even be useful to display.
    * <li>There is exactly one help set installed, and it specifies
    * a home ID, meaning that all possible help information is already
    * displayed in its own frame--provided clever mode is on!
    * </ol>
    * @param clever see above
    * @return the dummy help ID, or <code>null</code>
    */
    public String getMasterID (boolean clever) {
        if (setsByUrl.size () == 0 ||
                (clever && setsByUrl.size () == 1 && homesByCode.size () == 1))
            return null;
        else
            return MASTER_ID;
    }

    /** Test whether a helpset should be displayed prominently.
    * Currently this is only true if the title of the help set
    * as originally found in XML ended with an asterisk, as a
    * special hack marker.
    * @param homeID the home ID to test
    * @return whether the helpset should be prominent
    */
    public boolean isDistinguished (String homeID) {
        return distinguishedHomeIDs.contains (homeID);
    }

    /** Show some help.
     *This is the basic call which should be used externally
     *and is the result of {@link TopManager#showHelp}.
     *Handles null contexts, missing or null help IDs, and null URLs.
     *If possible, looks for an existing help window in the proper mode
     *on this workspace that is showing the same help set as this context
     *help object requests. If there is any problem, shows the master set
     *instead, or it may also create a new help window.
     *All display is done using {@link HelpComponent}.
     *Works correctly if invoked while a dialog is open--creates a new modal
     *dialog with the help.
     * @param ctx the help context to display
     */
    public synchronized void showHelp (HelpCtx ctx) {
        if (! SwingUtilities.isEventDispatchThread ()) {
            err.log ("showHelp later...");
            final HelpCtx ctx2 = ctx;
            SwingUtilities.invokeLater (new Runnable () {
                    public void run () {
                        showHelp (ctx2);
                    }
                });
            return;
        }
        err.log ("showing help: " + ctx);
        if (ctx == null) ctx = HelpCtx.DEFAULT_HELP;
        HelpSet hs = findHelpSetForID (ctx.getHelpID (), true);
        JHelp jh = createJHelp (hs);
        NbPresenter dlg = NbPresenter.currentModalDialog;
        if (dlg == null) {
            err.log ("showing as non-dialog");
            boolean nue = false;
            if (frameViewer == null) {
                nue = true;
                err.log ("creating new frame viewer");
                frameViewer = new JFrame ();
                frameViewer.setIconImage (Toolkit.getDefaultToolkit ().getImage
                                          (JavaHelp.class.getClassLoader ().getResource
                                           ("org/netbeans/core/resources/actions/moduleHelp.gif"))); // NOI18N
            }
            if (lastFramedHelpSet != hs) {
                err.log ("changing displayed help set");
                frameViewer.setTitle (hs.getTitle ());
                frameViewer.getContentPane ().removeAll ();
                if (dialogViewer != null) {
                    err.log ("cleaning up old dialog viewer");
                    dialogViewer.setVisible (false);
                    dialogViewer.getContentPane ().removeAll ();
                }
                frameViewer.getContentPane ().add (jh, BorderLayout.CENTER);
                frameViewer.pack ();
                frameViewer.repaint ();
                lastFramedHelpSet = hs;
            }
            if (nue) {
                Dimension screenSize = Toolkit.getDefaultToolkit ().getScreenSize ();
                Dimension frameSize = frameViewer.getSize ();
                frameViewer.setLocation ((screenSize.width - frameSize.width) / 2,
                                         (screenSize.height - frameSize.height) / 2);
            } else {
                frameViewer.setState (Frame.NORMAL);
            }
            frameViewer.setVisible (true);
            frameViewer.requestFocus ();
        } else {
            err.log ("showing as dialog");
            if (dialogViewer != null && ! dialogViewer.isVisible ()) {
                err.log ("destroying old dialog viewer and creating new");
                dialogViewer.getContentPane ().removeAll ();
                dialogViewer.dispose ();
                dialogViewer = new JDialog (dlg, false);
                lastDialogHelpSet = null;
            }
            if (dialogViewer == null) {
                err.log ("creating new dialog viewer");
                dialogViewer = new JDialog (dlg, false);
                if (frameViewer != null && frameViewer.isVisible ()) {
                    err.log ("hiding frame viewer");
                    dialogViewer.setBounds (frameViewer.getBounds ());
                    frameViewer.setVisible (false);
                    frameViewer.getContentPane ().removeAll ();
                } else {
                    lastFramedHelpSet = null;
                }
            }
            if (lastDialogHelpSet != hs) {
                err.log ("changing displayed help set");
                dialogViewer.setTitle (hs.getTitle ());
                dialogViewer.getContentPane ().removeAll ();
                dialogViewer.getContentPane ().add (jh, BorderLayout.CENTER);
                dialogViewer.pack ();
                dialogViewer.repaint ();
                lastDialogHelpSet = hs;
            }
            rearrange (dlg);
            //dialogViewer.setVisible (true);
            dialogViewer.show ();
            //dialogViewer.requestFocus ();
        }
        displayInJHelp (jh, ctx.getHelpID (), ctx.getHelp ());
    }

    /** Called (asynch in the event thread) after a new modal dialog is shown, or hidden.
     * @param ev ignored
     */
    public void stateChanged (ChangeEvent ev) {
        if (! SwingUtilities.isEventDispatchThread ()) throw new IllegalStateException ();
        NbPresenter dlg = NbPresenter.currentModalDialog;
        if (dlg == null) {
            err.log ("modal dialog was hidden");
            if (lastFramedHelpSet != null && frameViewer != null && ! frameViewer.isVisible ()) {
                err.log ("reparenting to frame viewer");
                frameViewer.getContentPane ().removeAll ();
                frameViewer.getContentPane ().add (createJHelp (lastFramedHelpSet), BorderLayout.CENTER);
                //frameViewer.pack ();
                frameViewer.setState (Frame.NORMAL);
                frameViewer.setVisible (true);
                frameViewer.repaint ();
            } else if (lastFramedHelpSet == null) {
                err.log ("no existing framed help set");
            } else if (frameViewer == null) {
                err.log ("no existing frame viewer");
            } else {
                err.log ("frame viewer was already visible");
            }
            dialogViewer = null;
            lastDialogHelpSet = null;
        } else {
            err.log ("modal dialog was shown");
            if (frameViewer != null && frameViewer.isVisible ()) {
                err.log ("frame viewer was visible");
                frameViewer.setVisible (false);
                frameViewer.getContentPane ().removeAll ();
                if (dialogViewer == null || ! dialogViewer.isVisible ()) {
                    err.log ("dialog viewer did not exist or was invisible");
                    dialogViewer = new JDialog (dlg, false);
                    dialogViewer.setTitle (lastFramedHelpSet.getTitle ());
                    dialogViewer.getContentPane ().add (createJHelp (lastFramedHelpSet), BorderLayout.CENTER);
                    dialogViewer.setBounds (frameViewer.getBounds ());
                    //dialogViewer.pack ();
                    rearrange (dlg);
                    dialogViewer.setVisible (true);
                    lastDialogHelpSet = lastFramedHelpSet;
                }
            } else {
                err.log ("frame viewer was not visible");
                lastFramedHelpSet = null;
                if (frameViewer != null) {
                    err.log ("cleaning up invisible frame viewer");
                    frameViewer.getContentPane ().removeAll ();
                }
            }
        }
    }

    /** If needed, visually rearrange dialogViewer and dlg on screen.
     * If they overlap, try to make them not overlap.
     * @param dlg the visible modal dialog
     */
    private void rearrange (NbPresenter dlg) {
        Rectangle r1 = dlg.getBounds ();
        Rectangle r2 = dialogViewer.getBounds ();
        if (r1.intersects (r2)) {
            err.log ("modal dialog and dialog viewer overlap");
            Dimension s = Toolkit.getDefaultToolkit ().getScreenSize ();
            if (r2.width <= r1.x) {
                err.log ("fit help on left");
                r2.x = (r1.x - r2.width) / 2;
            } else if (r2.height <= r1.y) {
                err.log ("fit help on top");
                r2.y = (r1.y - r2.height) / 2;
            } else if (r2.width <= s.width - r1.x - r1.width) {
                err.log ("fit help on right");
                r2.x = (r1.x + r1.width + s.width - r2.width) / 2;
            } else if (r2.height <= s.height - r1.y - r1.height) {
                err.log ("fit help on bottom");
                r2.y = (r1.y + r1.height + s.height - r2.height) / 2;
            } else if (r1.width + r2.width <= s.width) {
                err.log ("fit both horizontally");
                int extra = s.width - r1.width - r2.width;
                int gap = extra / 3;
                r2.x = gap;
                r1.x = s.width - gap - r1.width;
            } else if (r1.height + r2.height <= s.height) {
                err.log ("fit both vertically");
                int extra = s.height - r1.height - r2.height;
                int gap = extra / 3;
                r2.y = gap;
                r1.y = s.height - gap - r1.height;
            } else {
                err.log ("squeeze help vertically");
                r1.y = 0;
                r2.y = r1.height;
                r2.height = s.height - r2.y;
            }
            dlg.setBounds (r1);
            dialogViewer.setBounds (r2);
        }
    }

    /** Find the proper help set for an ID.
    * @param id the ID to look up (may be null)
    * @param preferMaster if true, the master is always used unless the id is a home ID
    * @return the proper help set (master if not otherwise found)
    */
    private HelpSet findHelpSetForID (String id, boolean preferMaster) {
        if (id != null && (! preferMaster || homesByCode.values ().contains (id))) {
            Iterator it = setsByUrl.values ().iterator ();
            while (it.hasNext ()) {
                HelpSet hs = (HelpSet) it.next ();
                if (hs.getCombinedMap ().isValidID (id, hs))
                    return hs;
            }
            warnBadID (id);
        }
        return getMaster ();
    }

    /** Get home IDs by code name.
     * @see #homesByCode
     * @return a map
     */
    public Map getHomesByCode() {
        return homesByCode;
    }

    /** Get the home IDs by display name.
     * @see #homesByDisplay
     * @return a map
     */
    public Map getHomesByDisplay() {
        return homesByDisplay;
    }

    /** Warn that an ID was not found in any help set.
    * @param id the help ID
    */
    private static void warnBadID (String id) {
        // PLEASE DO NOT COMMENT OUT...localized warning
        System.err.println (Main.getString ("MSG_jh_id_not_found", id));
    }

    /** Display something in a JHelp.
     *Handles {@link #MASTER_ID}, as well as help IDs
     *that were not found in any help set, various exceptions, etc.
     * @
     * @param helpID a help ID string to display, may be <CODE>null</CODE>
     * @param url a URL to display, may be <CODE>null</CODE>; lower priority than the help ID
     */
    private synchronized void displayInJHelp (JHelp jh, String helpID, URL url) {
        err.log ("displayInJHelp: " + helpID + " " + url);
        try {
            if (helpID != null && ! helpID.equals (MASTER_ID)) {
                HelpSet hs = jh.getModel ().getHelpSet ();
                if (hs.getCombinedMap ().isValidID (helpID, hs)) {
                    jh.setCurrentID (helpID);
                } else {
                    warnBadID (helpID);
                }
            } else if (url != null) {
                jh.setCurrentURL (url);
            }
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Create &amp; return a JHelp with the supplied help set.
     * In the case of the master help, will show the home page for
     * the distinguished help set if there is exactly one such,
     * or in the case of exactly one home page, will show that.
     * Caches the result and the result may be a reused JHelp.
    * @return the new JHelp
    */
    private JHelp createJHelp (HelpSet hs) {
        JHelp jh;
        synchronized (availableJHelps) {
            Reference r = (Reference) availableJHelps.get (hs);
            if (r != null) {
                jh = (JHelp) r.get ();
                if (jh != null) {
                    return jh;
                }
            }
        }
        try {
            jh = new JHelp (hs);
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
            return new JHelp ();
        }
        synchronized (availableJHelps) {
            availableJHelps.put (hs, new SoftReference (jh));
        }
        try {
            if (hs == getMaster ()) {
                if (distinguishedHomeIDs.size () == 1) {
                    jh.setCurrentID ((String) distinguishedHomeIDs.iterator ().next ());
                } else if (homesByCode.size () == 1) {
                    jh.setCurrentID ((String) homesByCode.values ().iterator ().next ());
                } else {
                    jh.setCurrentID (MASTER_ID);
                }
            } else {
                javax.help.Map.ID home = hs.getHomeID ();
                if (home != null) {
                    jh.setCurrentID (home);
                }
            }
        } catch (Exception e) {
            err.notify (ErrorManager.INFORMATIONAL, e);
        }
        return jh;
    }

    // --- BACKWARDS COMPATIBILITY of serialized window system ---

    public static class HelpComponent extends CloneableTopComponent {

        private static final long serialVersionUID =882544693090803112L;

        public HelpComponent () {
        }

        public Object readResolve () throws ObjectStreamException {
            //System.err.println("HelpComponent.readResolve; discard=" + discard);
            return null;
        }

    }

    // --- OPTIONAL SUPPORT for using alternate HTML browser with -J-Dorg.netbeans.core.JavaHelp.NOSWINGHTML=true

    /** The current UIDefaults object, whose property changes will be listened to.
     */
    private static UIDefaults curr;
    /** A listener to changes in the UI which attempts to keep the custom UIs
     *installed over the JavaHelp defaults.
     */
    private static PropertyChangeListener uiListener;
    static {
        if (Boolean.getBoolean ("org.netbeans.core.JavaHelp.NOSWINGHTML")) { // NOI18N
            System.err.println("JavaHelp: NOT using default Swing HTML rendering (using HTML browser module)"); // NOI18N
            curr = UIManager.getLookAndFeelDefaults ();
            uiListener =new PropertyChangeListener () {
                public void propertyChange (PropertyChangeEvent ev) {
                    //System.err.println("uiListener.pC");
                    if (! BrowserContentViewerUI.class.getName ().equals (curr.get ("HelpContentViewerUI")) || // NOI18N
                            ! ToolbarlessHelpUI.class.getName ().equals (curr.get ("HelpUI"))) { // NOI18N
                        //System.err.println("uiListener: out of date");
                        //System.err.println("HelpContentViewerUI=" + curr.get ("HelpContentViewerUI"));
                        //System.err.println("HelpUI=" + curr.get ("HelpUI"));
                        curr.removePropertyChangeListener (uiListener);
                        curr.putDefaults (new Object[] {
                                              "HelpContentViewerUI", BrowserContentViewerUI.class.getName (), // NOI18N
                                              "HelpUI", ToolbarlessHelpUI.class.getName () // NOI18N
                                          });
                        curr.addPropertyChangeListener (uiListener);
                    }
                }
            };
            curr.addPropertyChangeListener (uiListener);
            uiListener.propertyChange (null);
            UIManager.addPropertyChangeListener (new PropertyChangeListener () {
                                                     public void propertyChange (PropertyChangeEvent ev) {
                                                         //System.err.println("UIManager PCE: " + ev.getPropertyName ());
                                                         if ("lookAndFeel".equals (ev.getPropertyName ())) {
                                                             curr.removePropertyChangeListener (uiListener);
                                                             curr = UIManager.getLookAndFeelDefaults ();
                                                             curr.addPropertyChangeListener (uiListener);
                                                             uiListener.propertyChange (null);
                                                         }
                                                     }
                                                 });
        }
    }

    /** Hack to take the toolbar out of the JHelp window. It is big and ugly
    * and the web browser already has a toolbar which contains back and
    * forward buttons which also work better (e.g. with frames).
    */
    public static class ToolbarlessHelpUI extends BasicHelpUI {

        /** Create the proper UI.
         * @param comp the JHelp component to render
         * @return the desired UI
         */
        public static ComponentUI createUI(JComponent comp) {
            return new ToolbarlessHelpUI ((JHelp) comp);
        }

        static final long serialVersionUID =-4593005151581696837L;
        /** Create a new UI.
         * @param jHelp the component to render
         */
        public ToolbarlessHelpUI(JHelp jHelp) {
            super (jHelp);
        }

        /** Installs the UI.
         *Just tries to remove the JToolBar.
         * @param c the component to render
         */
        public void installUI(JComponent c) {
            super.installUI (c);
            Component[] components = c.getComponents ();
            for (int i = 0; i < components.length; i++) {
                if (components[i] instanceof JToolBar) {
                    c.remove (components[i]);
                    return;
                }
            }
            // PLEASE DO NOT COMMENT OUT
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                System.err.println ("Warning: could not remove toolbar from JHelp UI");
        }

    }

    /** Alternate UI for the content browser.
    * Instead of using the slow and buggy Swing Text HTML support,
    * uses {@link #HtmlBrowser} embedded browser.
    */
    public static class BrowserContentViewerUI extends HelpContentViewerUI implements Serializable, PropertyChangeListener, HelpModelListener {

        /** Create UI for a component.
         * @param comp the component to render
         * @return the UI
         */
        public static ComponentUI createUI(JComponent comp) {
            return new BrowserContentViewerUI ();
        }

        /** The actual browser component.
         */
        private HtmlBrowser browser;
        /** The content viewer being rendered.
         */
        private JHelpContentViewer viewer;

        static final long serialVersionUID =2565972402655318909L;
        /** Create a new UI.
         */
        public BrowserContentViewerUI() {
            //System.err.println("creating UI...");
        }

        /** Install UI for the component.
         * @param c the content navigator to render
         */
        public void installUI(JComponent c) {
            //System.err.println("installing UI to " + c);
            viewer = (JHelpContentViewer) c;
            viewer.addPropertyChangeListener (this);
            TextHelpModel model = viewer.getModel ();
            // [PENDING] should viewer.addHelpModelListener (this) be used instead??
            if (model != null) model.addHelpModelListener (this);
            // No TextHelpModelListener--we do not support highlighting.
            browser = new HtmlBrowser ();
            browser.setEnableHome (false);
            browser.setEnableLocation (false);
            browser.addPropertyChangeListener (this);
            viewer.setLayout (new BorderLayout ());
            viewer.add (browser, BorderLayout.CENTER);
        }

        /** Uninstall UI from the component.
         * @param c the content navigator
         */
        public void uninstallUI(JComponent c) {
            //System.err.println("uninstalling UI from " + c);
            JHelpContentViewer jhcv = (JHelpContentViewer) c;
            viewer.removePropertyChangeListener (this);
            TextHelpModel model = viewer.getModel ();
            if (model != null) model.removeHelpModelListener (this);
            if (browser != null) browser.removePropertyChangeListener (this);
            browser = null;
            viewer.setLayout (null);
            viewer.removeAll ();
            viewer = null;
        }

        /** Handle changes in browser.title and viewer.helpModel.
         * @param ev the event
         */
        public void propertyChange(PropertyChangeEvent ev) {
            //System.err.println("UI prop chg: [" + ev.getPropertyName () + "] " + ev.getOldValue () + " -> " + ev.getNewValue ());
            if (ev.getSource () == browser) {
                // Really a PROP_TITLE change, but what the heck.
                if (browser == null) return;
                URL url = browser.getDocumentURL ();
                if (url == null) return; // Probably should never happen.
                TextHelpModel model = viewer.getModel ();
                if (model != null) {
                    //System.err.println("Browser url set to " + url);
                    try {
                        model.removeHelpModelListener (this);
                        // model.setDocumentTitle (???);
                        javax.help.Map.ID id = model.getHelpSet ().getCombinedMap ().getIDFromURL (url);
                        if (id != null) {
                            //System.err.println("(setting by id)");
                            try {
                                model.setCurrentID (id);
                            } catch (InvalidHelpSetContextException ihsce) {
                                // Should probably not happen:
                                TopManager.getDefault ().notifyException (ihsce);
                                model.setCurrentURL (url);
                            }
                        } else {
                            //System.err.println("(setting by URL)");
                            model.setCurrentURL(url);
                        }
                    } finally {
                        model.addHelpModelListener (this);
                    }
                }
            } else {
                // Viewer.
                if ("helpModel".equals (ev.getPropertyName ())) {
                    //System.err.println("Changing help model...");
                    HelpModel old = (HelpModel) ev.getOldValue ();
                    if (old != null) old.removeHelpModelListener (this);
                    HelpModel nue = (HelpModel) ev.getNewValue ();
                    if (nue != null) nue.addHelpModelListener (this);
                }
            }
        }

        /** Handle changes in the help model's ID.
         * @param ev the event
         */
        public void idChanged(HelpModelEvent ev) {
            //System.err.println("UI: ID changed to " + ev.getID () + " " + ev.getURL ());
            if (browser != null) {
                try {
                    browser.removePropertyChangeListener (this);
                    browser.setURL (ev.getURL ());
                } finally {
                    browser.addPropertyChangeListener (this);
                }
            }
        }

    }

}
