/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.beans.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.ResourceBundle;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import org.openide.*;
import org.openide.execution.*;
import org.openide.filesystems.*;
import org.openide.modules.*;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.io.*;

/** This class implementats the main functions which are needed for
* installing of new module into the system and storage of them.
*
* @author Petr Hamernik, Jaroslav Tulach
*/
public final class ModuleInstaller extends Object {
    /** synch lock for moduleClassLoader; --> INITIALIZE BEFORE err <-- */
    private static final Object moduleClassLoaderLock = new Object ();

    /** logging for module subsystem */
    static final ErrorManager err = TopManager.getDefault ().getErrorManager ().getInstance ("org.netbeans.core.modules"); // NOI18N

    /** Tes ResourceBundle */
    private static final ResourceBundle bundle = NbBundle.getBundle( ModuleInstaller.class );

    /** name of file to contain stored externalizable modules */
    private static final String EXTERNALIZED_MODULES = "externalizedModules.ser"; // NOI18N

    /** Instance of ModuleInstaller support */
    static ModuleInstallerSupport miSupport = null;

    /** Boolean property to remember if some modules were updated or disabled
     * to write new installedModules.xml during installing modules 
     */
    private static boolean writeRegistry = true;

    /** The only instance of the installer */
    private static final PropertyChangeSupport INSTANCE = new PropertyChangeSupport (new ModuleInstaller ());

    /** The current system class loader which can load all modules. */
    private static ClassLoader moduleClassLoader = null;
    
    /** Private constructor */
    private ModuleInstaller () {
    }


    // PUBLIC METHODS ------------------------------------------------------------

    /** Get the master system class loader loading all installed non-test modules. */
    public static ClassLoader getModuleClassLoader () {
        synchronized (moduleClassLoaderLock) {
            if (moduleClassLoader == null) {
                moduleClassLoader = recomputeModuleClassLoader ();
                //System.err.println ("ModuleInstaller.moduleClassLoader=" + moduleClassLoader);
                //ModuleItem.maybeDumpLoaderURLs (moduleClassLoader);
            }
            return moduleClassLoader;
        }
    }
    
    /** Adds listener on the nodes.
    */
    public static void addPropertyChangeListener (PropertyChangeListener l) {
        INSTANCE.addPropertyChangeListener (l);
    }

    /** Removes listener on the nodes.
    */
    public static void removePropertyChangeListener (PropertyChangeListener l) {
        INSTANCE.removePropertyChangeListener (l);
    }

    /** Getter for array of modules of given kind
    * @return array of modules
    */
    public static ModuleItem[] getModuleItems ( int kind ) {
        Collection modules = miSupport.get( kind );
        return (ModuleItem[])modules.toArray (new ModuleItem[ modules.size ()]);
    }

    /** Retruns array of module descriptions of installed modules */
    public static ModuleDescription[] getModuleDescriptions( int kind ) {
        Collection modules = miSupport.get( kind );
        ModuleItem[] items = (ModuleItem[])modules.toArray (new ModuleItem[ modules.size ()]);
        ModuleDescription[] result = new ModuleDescription[ items.length ];

        for( int i = 0; i < items.length; i++ ) {
            result[i] = items[i].getDescription();
        }
        return result;
    }
    
    

    /** Returns module item for name */
    static ModuleItem getByName( String codeNameBase ) {
        return miSupport.get( codeNameBase );
    }


    /** Initializes the modules loads the previously installed modules
     */
    public synchronized static void initialize(
        File centralModuleDirectory,
        File userModuleDirectory
    ) {
        // first of all initialize all XML stuff
        org.netbeans.core.xml.XML.init ();
        
        // System.out.println("LOADING MODULES ------------------------------- " ); //NOI18N
        org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
        long time = System.currentTimeMillis();

        miSupport = new ModuleInstallerSupport ();
        miSupport.initialize(
                        centralModuleDirectory,
                        userModuleDirectory
                    );  // Reads the miSupport
        pm.notifyTime("ModuleInstallerSupport.initialize", System.currentTimeMillis() - time);
        // System.out.println( miSupport );

        Collection m2i = miSupport.get( miSupport.ENABLED_MODULE );

        // #9806 Check if some loaded modules (now) depend on uninstalled
        // autoload modules. If so, load those autoloads first, then
        // proceed as before. They will be "installed" later anyway.
        // Note this means a module could be restored, then later installed.
        // Weird, but rare and hopefully harmless.
        Collection preloads = new HashSet (); // Collection<ModuleItem>
        Collection possiblePreloads = miSupport.get (miSupport.AUTOLOAD_USER_MODULE | miSupport.AUTOLOAD_CENTRAL_MODULE | miSupport.DISABLED_MODULE);
        checkForPreloads (m2i, possiblePreloads, preloads);
        ModuleInstaller.err.log ("Preloads: " + preloads);
        m2i.addAll (preloads);

        // Sort enabled modules
        Collection orderedModules = null;
        try {
            orderedModules = resolveOrdering( m2i/*, orderedPreloads*/ );
        }
        catch ( IllegalModuleException e ) {
            TopManager.getDefault().notifyException( e );
        }
        if ( orderedModules == null ) {
            orderedModules = new ArrayList();
        }

        // Look wther some modules have to be disabled
        Collection dm = miSupport.get( miSupport.ENABLED_MODULE );
        // ignore preloads here
        dm.removeAll( orderedModules );

        if ( !dm.isEmpty() ) {
            NotifyDescriptor nd = new NotifyDescriptor.Message(
                                      bundle.getString ("MSG_Some_Missed") + "\n" +
                                      getMissed( new ArrayList(), m2i ) );
            notifyLater (nd);
        }

        // Print info
        if ( orderedModules.size() > 0 )
            System.err.println (Main.getString ( "INFO_LoadedModules",
                                                 moduleItemsDescr (m2i)));


        // Add disabled modules to ordered modules
        orderedModules.addAll( dm );
        
        NbTopManager.get ().modulesClassPathInitialized ();

        // Install default sections of all modules
        {
            Iterator it = orderedModules.iterator ();
            while (it.hasNext ()) {
                ModuleItem mi = (ModuleItem)it.next ();
                TopManager.getDefault ().setStatusText (Main.getString ( "MSG_Install_Section",
                                                        mi.getDescription ().getName ()));
                if ( dm.contains( mi ) ) {
                    writeRegistry = true;
                    mi.setEnabledNoNotify( false );
                }
                else {
                    // Needed to ensure preloads are accessible in the loader too:
                    miSupport.enableLoader (mi);
                    mi.restoreLayer ();
                    mi.restoreSection ();
                    mi.restoreDefault ();
                }
            }
        }

        // Make an attempt to put even autoloaded modules into the
        // loader, now, though they will not be really installed till
        // autoLoadModules() is called, and no attempt is made to
        // enable disabled dependencies etc. The reason is that
        // loading a project can, under obscure cirsumstances, require
        // autoload modules to be available: e.g. if a module is split
        // in half, and old settings are loaded which refer to classes
        // now in an autoloaded (not restored!) half.
        try {
            Collection autoloads = resolveOrdering (miSupport.get (miSupport.AUTOLOAD_CENTRAL_MODULE | miSupport.AUTOLOAD_USER_MODULE),
                                                    miSupport.get (miSupport.ENABLED_MODULE));
            Iterator it = autoloads.iterator ();
            while (it.hasNext ()) {
                miSupport.enableLoader ((ModuleItem) it.next ());
            }
        } catch (IllegalModuleException ime) {
            // Don't really worry about it here, it will be noticed later anyway.
            err.notify (ErrorManager.INFORMATIONAL, ime);
        }

        // Tell loader pool node the installation is finished
        // No longer necessary: LoaderPoolNode.finishInstallation();

        // Try to deexternalize modules
        try {
            time = System.currentTimeMillis();
            deExternalizeModules();
            pm.notifyTime("deExternalizeModules", System.currentTimeMillis() - time);
        }
        catch ( IOException e ) { 
            TopManager.getDefault().notifyException( e );
        }

        // Restored or Updated code of all modules
        {
            Iterator it = orderedModules.iterator ();
            while (it.hasNext ()) {
                ModuleItem mi = (ModuleItem)it.next ();

                if (mi.isEnabled ()) {

                    TopManager.getDefault ().setStatusText (
                        Main.getString (  "MSG_Install_Module",  mi.getDescription ().getName ()));

                    if ( mi.isUpdated() ) {
                        mi.updatedCode();
                        writeRegistry = true;
                    }
                    else {
                        mi.restoreCode ();
                    }
                }

                mi.verifyClassLoader ();
            }
        }

        // Fire module change
        INSTANCE.firePropertyChange (null, null, null);
    }

    /** See if some extra modules present in possible need to be
     * installed before things in target can be restored. If so, put
     * them in preloads. Recursive, i.e. if A in target depends on B
     * in possible depends on C in possible, then both B and C will be
     * copied from possible to preloads, so that A is all ready. Cf.
     * #9806. All collections are of ModuleItem.
     */
    private static void checkForPreloads (Collection target, Collection possible, Collection preloads) {
        // Track the possible ones according to code name (incl. base!):
        Map possibleByName = new HashMap (); // Map<String, ModuleItem>
        Iterator it = possible.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem) it.next ();
            possibleByName.put (mi.getDescription ().getCodeName (), mi);
        }
        it = target.iterator ();
        while (it.hasNext ()) {
            checkForPreloads ((ModuleItem) it.next (), possibleByName, preloads);
        }
        /*
        if (err.isLoggable (ErrorManager.UNKNOWN)) {
            err.log ("Check for preloads: restored modules=" + target + " possible preloads=" + possible + " actual preloads=" + preloads);
        }
        */
    }
    private static void checkForPreloads (ModuleItem target, Map possibleByName, Collection preloads) {
        ModuleDescription.Dependency[] deps = target.getDescription ().getDependencies ();
        for (int i = 0; i < deps.length; i++) {
            if (deps[i].getType () == ModuleDescription.Dependency.TYPE_MODULE) {
                ModuleItem required = (ModuleItem) possibleByName.get (deps[i].getName ());
                if (required != null && ! preloads.contains (required)) {
                    preloads.add (required);
                    checkForPreloads (required, possibleByName, preloads);
                }
            }
        }
    }

    /** Autoloads all modules into the system. This is called from Main and
    * installs new modules. In multiuser environment only the modules in USER directory
    * are installed. The modules from central directory are already restored
    */
    public static synchronized void autoLoadModules () {
        autoLoadModules( miSupport.AUTOLOAD_USER_MODULE | miSupport.AUTOLOAD_CENTRAL_MODULE );

        TopManager.getDefault().setStatusText( "" ); // NOI18N
        INSTANCE.firePropertyChange (null, null, null);

    }

    private static synchronized void autoLoadModules ( int kind ) {

        // resort new modules
        Collection m2iCentral = miSupport.get (miSupport.AUTOLOAD_CENTRAL_MODULE);
        Collection m2iUser = miSupport.get (miSupport.AUTOLOAD_USER_MODULE);
        Collection m2i = new ArrayList ();
        if ((kind & miSupport.AUTOLOAD_CENTRAL_MODULE) != 0) {
            m2i.addAll (m2iCentral);
        }
        if ((kind & miSupport.AUTOLOAD_USER_MODULE) != 0) {
            m2i.addAll (m2iUser);
        }
        if (((kind & (miSupport.AUTOLOAD_USER_MODULE | miSupport.AUTOLOAD_CENTRAL_MODULE)) == 0) ||
            ((kind & (~ miSupport.AUTOLOAD_USER_MODULE) & (~ miSupport.AUTOLOAD_CENTRAL_MODULE)) != 0)) {
            throw new IllegalArgumentException ("ModuleInstaller.autoLoadModules: kind=" + kind); // NOI18N
        }
        Collection newModules = null;

        try {
            newModules = resolveOrdering ( m2i, miSupport.get( miSupport.ENABLED_MODULE | miSupport.DISABLED_MODULE ) );
        } catch (IllegalModuleException e) {
            TopManager.getDefault ().notifyException (e);
        }

        Collection dm = miSupport.get( kind );
        dm.removeAll( newModules );

        if ( !dm.isEmpty() ) {
            NotifyDescriptor nd = new NotifyDescriptor.Message(
                                      bundle.getString ("MSG_Some_New_Missed") + "\n" +
                                      getMissed( miSupport.get( miSupport.ENABLED_MODULE ), m2i ) );
            notifyLater (nd);
        }


        if ( newModules == null )
            newModules = new ArrayList();


        Collection modules4Check = miSupport.get( kind );
        modules4Check.removeAll( dm );
        Collection enModules = miSupport.checkDependenciesOnDisabled ( modules4Check );
        if ( enModules != null && enModules.size() > 0) {
            StringBuffer sb = new StringBuffer();
            sb.append( bundle.getString( "MSG_EnableToInstall") );
            Iterator it = enModules.iterator();
            while( it.hasNext() ) {
                sb.append ( ((ModuleItem)it.next() ).getDescription().getName() ).append( "\n" ); // NOI18N
            }
            NotifyDescriptor nd = new NotifyDescriptor.Message( sb.toString() );
            notifyLater (nd);

            it = enModules.iterator();
            while( it.hasNext() ) {
                ((ModuleItem)it.next() ).setEnabled( true );
            }
        }

        // Print information about installed modules ...
        if ( newModules.size () > 0)
            System.err.println (Main.getString ("INFO_InstalledModules", moduleItemsDescr ( newModules )));


        // Add disabled modules to ordered modules
        newModules.addAll( dm );


        try {
            {
                Iterator it = newModules.iterator ();
                while (it.hasNext ()) {
                    ModuleItem mi = (ModuleItem)it.next ();

                    TopManager.getDefault ().setStatusText (Main.getString ("MSG_Install_Section", mi.getDescription ().getName ()));

                    // add the module to list of installed modules
                    miSupport.remove( mi, miSupport.getKind( mi ) );

                    if ( dm.contains( mi ) ) {
                        mi.setEnabledNoNotify( false );
                        miSupport.add (mi, miSupport.DISABLED_MODULE);
                        continue;
                    }

                    miSupport.add ( mi, miSupport.ENABLED_MODULE );

                    mi.restoreLayer ();

                    if ( miSupport.isMultiuser() && m2iCentral.contains (mi) )
                        mi.restoreSection();
                    else
                        mi.installSection();

                    mi.restoreDefault();
                }
            }

            // Install code of all modules ...
            {
                Iterator it = newModules.iterator ();
                while (it.hasNext ()) {
                    ModuleItem mi = (ModuleItem)it.next ();


                    TopManager.getDefault ().setStatusText (Main.getString ("MSG_Install_Module", mi.getDescription ().getName ()));

                    //System.err.println ("Maybe install/restore: " + mi + " enabled=" + mi.isEnabled () + " multiuser=" + miSupport.isMultiuser () + " kind=" + miSupport.getKind (mi));
                    if ( mi.isEnabled() ) {
                        if ( miSupport.isMultiuser() && m2iCentral.contains (mi) )
                            mi.restoreCode();
                        else
                            mi.installCode();
                    }

                    mi.verifyClassLoader ();
                }
            }

            if ( !newModules.isEmpty() ) {

                TopManager.getDefault().setStatusText( Main.getString (
                                                           "MSG_Install_Complete", // NOI18N
                                                           new Integer( newModules.size ()),
                                                           ((ModuleItem)newModules.iterator().next()).getDescription().getName()
                                                       ));

                INSTANCE.firePropertyChange (null, null, null);
            }
        }
        finally {
        }

        if ( !newModules.isEmpty () || writeRegistry ) {
            miSupport.writeRegistry();
        }

    }

    /** Instructs all modules to exit.
    * @return false if the exit should not continue
    */ 
    public static boolean exit () {

        Collection enabledModules = miSupport.get( miSupport.ENABLED_MODULE /*|
                                                   miSupport.TEST_MODULE*/ );

        Iterator it = enabledModules.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem)it.next ();
            if (!mi.closing ()) {
                return false;
            }
        }
        
        try {
            externalizeModules ();
        } catch (IOException ex) {
            TopManager.getDefault ().notifyException (ex);
        }

        // No of the modules voted against exit notify it to modules
        it = enabledModules.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem)it.next ();
            mi.closeCode ();
        }

        return true;
    }

    /** Installs module from file */
    public synchronized static ModuleItem installFromFile (File file, boolean reloadable) throws IOException {

        int base = ModuleItem.BASE_JAR;

        if ( file.getParentFile().equals( miSupport.getCentralModuleDirectory() ) ) {
            base = ModuleItem.BASE_CENTRAL;
        }
        else if ( file.getParentFile().equals( miSupport.getUserModuleDirectory() ) ) {
            base = ModuleItem.BASE_USER;
        }

        ModuleItem mi = null;

        if ( base == ModuleItem.BASE_JAR )
            mi = new ModuleItem( base, file.getPath(), true, reloadable );
        else
            mi = new ModuleItem( base, file.getName(), true, reloadable );

        ModuleItem exMi = miSupport.get( mi.getDescription().getCodeNameBase() );

        if ( exMi != null ) {
            switch ( miSupport.getKind( exMi ) ) {
            case ModuleInstallerSupport.ENABLED_MODULE:
            case ModuleInstallerSupport.DISABLED_MODULE:
                NotifyDescriptor nd = new NotifyDescriptor.Message(
                                          bundle.getString( "MSG_Already_Installed" ) );
                notifyLater (nd);
                return null;
            }

        }

        if (resolveOrdering ( Collections.nCopies (1, mi),
                              miSupport.ENABLED_MODULE | miSupport.DISABLED_MODULE ).isEmpty () ) {

            NotifyDescriptor.Message nd = new NotifyDescriptor.Message(
                                              bundle.getString( "MSG_Module_Missed" ) + "\n" +
                                              getMissed( miSupport.get( miSupport.ENABLED_MODULE | miSupport.DISABLED_MODULE ), Collections.nCopies (1, mi) ) );
            notifyLater (nd);
            return null;
            //throw new IOException ();
        }

        Collection enModules = checkDependenciesOnDisabled ( mi.getDescription() );

        if ( enModules != null ) {
            Iterator it = enModules.iterator();
            while( it.hasNext() ) {
                ((ModuleItem)it.next() ).setEnabled( true );
            }
        }
        else {
            return null;
        }

        doAddModule( mi );
        miSupport.add( mi, miSupport.ENABLED_MODULE );

        INSTANCE.firePropertyChange (null, null, null);
        miSupport.writeRegistry();

        return mi;
    }

    /** Called publically from apisupport, scripting, and so on to install a test module.
     * Will perform internal logic to install, reinstall, move base, etc.
     * @param jarfile a JAR to install in test mode
     * @throws IOException if there is any problem
     */
    public synchronized static void deployTestModule (File jarfile) throws IOException {
        jarfile = jarfile.getAbsoluteFile ();
        err.log ("Deploy test module: " + jarfile);
        Collection all = miSupport.get (miSupport.ENABLED_MODULE | miSupport.DISABLED_MODULE | miSupport.AUTOLOAD_CENTRAL_MODULE | miSupport.AUTOLOAD_USER_MODULE);
        Iterator it = all.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem) it.next ();
            String miName = mi.getDescription ().getCodeNameBase ();
            if (jarfile.equals (mi.getJar ())) {
                err.log ("Redeploying existing module " + miName);
                // Found existing one. Make sure it is reloadable, and install or reload it.
                mi.setReloadable (true);
                // If we are dealing with A, and B depends on A, and C on B...
                Collection dependents = getDependentModules (mi);
                List revDependents = new ArrayList (dependents);
                Collections.reverse (revDependents);
                it = revDependents.iterator ();
                // ...uninstall C, then B, ...
                while (it.hasNext ()) {
                    ((ModuleItem) it.next ()).setEnabled (false);
                }
                // ... then A...
                mi.setEnabled (false); // will do nothing if already disabled
                // ... then install A...
                mi.setEnabled (true);// now load it
                it = dependents.iterator ();
                // ...then B, then C
                while (it.hasNext ()) {
                    ((ModuleItem) it.next ()).setEnabled (true);
                }
                return;
            }
        }
        // It could be that the same module is already installed elsewhere. In this case we
        // should kill the old ModuleItem and replace it with this one. First we have to check
        // for a manifest so we can compare code name bases.
        String myName; // codeNameBase, or null
        try {
            JarFile jar = new JarFile (jarfile);
            try {
                String codeName = jar.getManifest ().getMainAttributes ().getValue (ModuleDescription.TAG_MAGIC);
                if (codeName != null) {
                    int idx = codeName.lastIndexOf ('/');
                    if (idx != -1) {
                        myName = codeName.substring (0, idx);
                    } else {
                        myName = codeName;
                    }
                } else {
                    throw new IOException ("JAR " + jarfile + " had no tag " + ModuleDescription.TAG_MAGIC + " in its manifest");
                }
            } finally {
                jar.close ();
            }
        } catch (IOException ioe) {
            // Screw it.
            err.notify (ioe);
            myName = null;
        }
        err.log ("Found this code name base in the JAR: " + myName);
        Collection dependents = Collections.EMPTY_SET;
        if (myName != null) {
            // Now check existing modules.
            it = all.iterator ();
            while (it.hasNext ()) {
                ModuleItem mi = (ModuleItem) it.next ();
                String miName = mi.getDescription ().getCodeNameBase ();
                if (miName.equals (myName)) {
                    err.log ("Found existing module with different JAR: " + mi.getJar ());
                    // Kill it.
                    dependents = getDependentModules (mi);
                    List revDependents = new ArrayList (dependents);
                    Collections.reverse (revDependents);
                    it = revDependents.iterator ();
                    while (it.hasNext ()) {
                        ((ModuleItem) it.next ()).setEnabled (false);
                    }
                    mi.setEnabled (false);
                    miSupport.remove (mi, miSupport.getKind (mi));
                    INSTANCE.firePropertyChange (null, null, null);
                    miSupport.writeRegistry ();
                    break;
                }
            }
        }
        // OK, make a new module for it.
        err.log ("Deploying brand-new module...");
        installFromFile (jarfile, true);
        // Reenable dependents (only nonempty if was existing w/ different JAR):
        it = dependents.iterator ();
        while (it.hasNext ()) {
            ((ModuleItem) it.next ()).setEnabled (true);
        }
    }


    static String getMissed( int kind, Collection installed ) {
        return getMissed( miSupport.get( kind ), installed );
    }

    private static String getMissed( Collection restored, Collection installed ) {
        // List of messages for modules which are *not* to be installed.
        StringBuffer sb = new StringBuffer();
        List missed = new ArrayList (); // List<String>
        // Set of modules which will actually be installed.
        Set actual = new HashSet (); // Set<ModuleDescription>
        actual.addAll (installed);
        int misscount;              // this time around
        do {
            misscount = 0;
            Iterator it = actual.iterator ();
            while (it.hasNext ()) {
                ModuleDescription test = ((ModuleItem) it.next ()).getDescription();
                Set whatCanIStillUse = new HashSet (); // Set<ModuleDescription>

                //whatCanIStillUse.addAll (restored);
                Iterator it2 = restored.iterator();
                while (it2.hasNext ())
                    whatCanIStillUse.add( ((ModuleItem) it2.next ()).getDescription() );
                // whatCanIStillUse.addAll (actual);
                it2 = actual.iterator();
                while (it2.hasNext ())
                    whatCanIStillUse.add( ((ModuleItem) it2.next ()).getDescription() );

                try {
                    String miss = test.reasonWhyUnsatisfied ((ModuleDescription[]) whatCanIStillUse.toArray (new ModuleDescription[0]));
                    if (miss != null) {
                        misscount++;
                        it.remove ();
                        missed.add (miss);
                        sb.append( miss ).append( "\n" ); // NOI18N
                    }
                }
                catch ( org.openide.modules.IllegalModuleException e ) {}
            }
        } while (misscount > 0);

        if (missed.size () > 0) {
            return sb.toString();
            /*
            System.out.println( missed );
            return java.text.MessageFormat.format (bundle.getString ("MSG_Some_Missed") + "\n", 
              new Object[] { (String[]) missed.toArray (new String[missed.size()] ) } );
            */
        }

        return null;
    }

    /** Deletes a module with given by ModuleItem parameter.
    * @return true if the module was removed, false if it was not installed
    */
    public synchronized static boolean deleteModule( ModuleItem item )  {

        if (item != null ) {

            int kind = miSupport.getKind( item );

            if ( kind != miSupport.ENABLED_MODULE &&
                    kind != miSupport.DISABLED_MODULE )
                return false;


            if ( item.canDestroy() ) {
                doRemoveModule (item);
                miSupport.remove( item, kind );
                INSTANCE.firePropertyChange (null, null, null);
                miSupport.writeRegistry();
            }
            else {
                item.setEnabled( false );
            }

            return true;
        } else {
            return false;
        }
    }

    /** Changes the enabled status
    */
    static synchronized void changeEnabled (ModuleItem item) {
        try {
            if (item.isEnabled ()) {
                try {
                    item.reload ();
                } catch (Exception e) {
                    TopManager.getDefault ().notifyException (e);
                    return;
                }
                doAddModule (item);
            } else {
                doRemoveModule (item);
            }
        } finally {
            miSupport.writeRegistry();
        }
    }

    /** Checks for dependencies on disabled modules. If the module depends on
    * some disabled modules then the method returns collection of modules which 
    * should be enabled. If there are some other unsatisfied dependencies it returns 
    * null. If there are no problems it returns empty Collection.
    */
    static Collection checkDependenciesOnDisabled( ModuleDescription md ) {
        return askAndOrderToEnable( miSupport.checkDependenciesOnDisabled( md ) );
    }

    static Collection checkDependenciesOnDisabled( Collection moduleItems ) {
        return askAndOrderToEnable( miSupport.checkDependenciesOnDisabled( moduleItems ) );
    }

    /** Produce a list of JAR files including all installed modules,
     * their extensions, and enabled locale variants of both.
     * Will be returned in a classpath-like order.
     * Intended for use by the execution engine.
     * @return <code>List&lt;File&gt;</code> of module-related JARs/ZIPs
     */
    public static List getModuleJars () {
        Iterator modules = miSupport.get (miSupport.ENABLED_MODULE | miSupport.AUTOLOAD_CENTRAL_MODULE | miSupport.AUTOLOAD_USER_MODULE).iterator ();
        List l = new ArrayList (); // List<File>
        while (modules.hasNext ()) {
            // XXX remove duplicates
            l.addAll (((ModuleItem) modules.next ()).getAllJars ());
        }
        /*
        if (err.isLoggable (ErrorManager.UNKNOWN)) {
            err.log ("moduleJars=" + l);
        }
        */
        return l;
    }

    // PRIVATE METHODS -----------------------------------------------------------

    private static ClassLoader recomputeModuleClassLoader () {
        if (err != null) {
            err.log ("ModuleInstaller.recomputeModuleClassLoader");
        } // else maybe we have a custom ErrorManager which is still
          // being initialized here: ModuleInstaller.err is being loaded,
          // hence systemClassLoader asked to load it
        ClassLoader[] parents;
        if (miSupport != null) {
            Collection enabled = miSupport.get (miSupport.ENABLED_MODULE);
            List parentsl = new ArrayList (Math.min (1, enabled.size ())); // List<ClassLoader>
            Iterator it = enabled.iterator ();
            while (it.hasNext ()) {
                ModuleItem item = (ModuleItem) it.next ();
                if (item.getBase () == ModuleItem.BASE_LOADER) continue;
                item.ensureClassLoader ();
                parentsl.add (item.getDescription ().getClassLoader ());
            }
            parents = (ClassLoader[]) parentsl.toArray (new ClassLoader[parentsl.size ()]);
        } else {
            parents = new ClassLoader[] { };
        }
        return new ModuleClassLoader (parents);
    }
    private static final class ModuleClassLoader extends MultiURLClassLoader {
        public ModuleClassLoader (ClassLoader[] parents) {
            super (new URL[] { }, parents);
        }
        protected void finalize () {
            err.log ("Collected moduleClassLoader");
        }
    }

    // Called from ModuleInstallerSupport:
    static void appendToModuleClassLoader (ModuleItem item) {
        item.ensureClassLoader ();
        ((MultiURLClassLoader) getModuleClassLoader ()).append (item.getDescription ().getClassLoader ());
        err.log ("ModuleInstaller.appendToModuleClassLoader: " + item.getDescription ().getCodeNameBase ());
    }
    static void invalidateModuleClassLoader () {
        if (moduleClassLoader == null) return;
        // Probably has no effect but just in case:
        ((MultiURLClassLoader) moduleClassLoader).destroy ();
        moduleClassLoader = null;
        err.log ("ModuleInstaller.invalidateModuleClassLoader");
        ClassLoaderSupport.resetLoader ();
        Help.resetDocsLoader ();
    }

    /** Get a brief localized string describing a ModuleItem.
    */ 
    private static String moduleItemDescr (ModuleItem item) {

        String spec = item.getDescription ().getSpecVersion ();
        String impl = item.getDescription ().getImplVersion ();

        java.text.MessageFormat mf = new java.text.MessageFormat(Main.getString ("INFO_ModuleSummary"));


        return mf.format (new Object[] {
                              item.getDescription ().getCodeName (),
                              spec == null ? Main.getString ("INFO_NullSpec") : spec,
                              impl == null ? Main.getString ("INFO_NullImpl") : impl
                          });

    }

    /** Get a brief localized string describing a list of ModuleItem's. */

    private static String moduleItemsDescr (Collection items) {

        StringBuffer buf = new StringBuffer ();
        Iterator it = items.iterator ();
        boolean first = true;
        while (it.hasNext ()) {
            if (first)
                first = false;
            else
                buf.append (Main.getString ("INFO_ModuleSummarySeparator"));
            buf.append (moduleItemDescr ((ModuleItem) it.next ()));
        }
        return buf.toString ();

    }

    /** @exception if something fails
    */
    private static void doAddModule (ModuleItem item) {

        appendToModuleClassLoader (item);

        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Install_Section", item.getDescription ().getName ()));
        item.restoreLayer ();
        item.installSection ();
        item.restoreDefault ();
        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Install_Module", item.getDescription ().getName ()));
        item.installCode ();
        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Install_Complete", new Integer (1), item.getDescription ().getName ()));
        item.verifyClassLoader ();

        System.err.println (Main.getString ("INFO_ModuleInstalled", moduleItemDescr (item)));

    }

    /** Removes a module.
    */

    private static void doRemoveModule (ModuleItem item) {
        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Uninstall_Module", item.getDescription ().getName ()));
        item.uninstallCode ();
        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Uninstall_Section", item.getDescription ().getName ()));
        item.uninstallSection ();
        item.unrestoreLayer ();
        item.unrestoreDefault ();

        invalidateModuleClassLoader ();

        // Release its classloader so it could be reloaded:
        item.release ();

        TopManager.getDefault ().setStatusText (Main.getString ("MSG_Uninstall_Complete", new Integer (1), item.getDescription ().getName ()));

        System.err.println (Main.getString ("INFO_ModuleUninstalled", moduleItemDescr (item)));
    }

    /** Externalize modules.
    */
    private static void externalizeModules () throws IOException {
        IOException master = null;
        try {
            ObjectOutputStream os = new ObjectOutputStream
                (new BufferedOutputStream
                    (new FileOutputStream
                        (new File ( miSupport.getUserModuleDirectory(), EXTERNALIZED_MODULES))));
            try {

                Iterator it = miSupport.get( miSupport.ENABLED_MODULE ).iterator();

                while (it.hasNext ()) {
                    ModuleItem mi = (ModuleItem)it.next ();

                    try {
                        NbObjectOutputStream.writeSafely (os, mi.getDescription ().getModule ());
                    } catch (SafeException se) {
                        if (master == null) master = new IOException ();
                        TopManager.getDefault ().getErrorManager ().annotate (master, se);
                    }
                }
                NbObjectOutputStream.writeSafely (os, null);

                try {
                    NbObjectOutputStream.writeSafely (os, TopManager.getDefault ().getLoaderPool ());
                } catch (SafeException se) {
                    if (master == null) master = new IOException ();
                    TopManager.getDefault ().getErrorManager ().annotate (master, se);
                }

            } finally {
                os.close ();
            }
        } catch (IOException ioe) {
            throw ioe;
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            if (master == null) master = new IOException ();
            TopManager.getDefault ().getErrorManager ().annotate (master, t);
        }
        if (master != null) throw master;
    }

    /** Deexternalize modules.
    */
    private static void deExternalizeModules () throws IOException {

        File f = new File ( miSupport.getUserModuleDirectory(), EXTERNALIZED_MODULES);
        if (!f.exists ()) {
            f = new File ( miSupport.getCentralModuleDirectory(), EXTERNALIZED_MODULES);
            if (!f.exists ()) {
                return;
            }
        }

        ObjectInputStream is = new ObjectInputStream (
                                   new BufferedInputStream (new FileInputStream (f))
                               );

        while (true) {
            try {
                if (NbObjectInputStream.readSafely (is) == null) {
                    break;
                }
            } catch (SafeException ex) {
                err.notify (ErrorManager.INFORMATIONAL, ex);
            }
        }

        try {
            // TopManager.getDefault ().getLoaderPool ()
            NbObjectInputStream.readSafely (is);
        } catch (EOFException ex) {
            // this object has not been saved before => ignore
        } catch (SafeException ex) {
            err.notify (ErrorManager.INFORMATIONAL, ex);
        }

        is.close ();
    }


    static Collection resolveOrdering( Collection in, int kind ) {
        return resolveOrdering( in, kind, false );
    }

    static Collection resolveOrdering( Collection in, int kind, boolean containsIn ) {

        Collection installed = miSupport.get( kind );

        if ( containsIn ) {
            installed.removeAll( in );
        }

        try {
            Collection result = resolveOrdering( in, installed );
            return result;
        }
        catch ( IllegalModuleException e ) {
            TopManager.getDefault().notifyException( e );
            return new ArrayList();
        }
    }


    /** Reordres module to folow the dependencies.
    *
    * @param in collection of ModuleItems to install
    * @return list of module items
    */
    private static List resolveOrdering (Collection in) throws IllegalModuleException {
        return resolveOrdering (in, new ArrayList() );
    }


    /** Reordres module to folow the dependencies.
    *
    * @param in collection of ModuleItems to install
    * @param current current modules
    * @return list of module items
    * @exception IllegalModuleException if the modules are not ok.
    */
    private static List resolveOrdering (Collection in, Collection current)
    throws IllegalModuleException {
        HashMap map = new HashMap (in.size ()); // Map (ModuleDescription, ModuleItem)
        Iterator it = in.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem)it.next ();
            map.put (mi.getDescription (), mi);
        }

        List descr;

        if (current == null) {
            descr = ModuleDescription.resolveOrdering (map.keySet ());
        } else {
            it = current.iterator ();
            HashSet s = new HashSet (); // set of descriptions
            while (it.hasNext ()) {
                ModuleItem mi = (ModuleItem)it.next ();
                s.add (mi.getDescription ());
            }
            it = map.entrySet ().iterator ();
            //ModuleDescription[] others = (ModuleDescription[]) s.toArray (new ModuleDescription[s.size ()]);
        PREPARE:
            while (it.hasNext ()) {
                Map.Entry entry = (Map.Entry) it.next ();
                // Alternate "solution": workaround in ModuleItem to just not throw NPE.
                /*
                ModuleDescription desc = (ModuleDescription) entry.getKey ();
                // We want to make sure the classloader is there, since otherwise
                // package dep checks will give an NPE. However, this call assumes
                // that all required modules exist (possibly disabled, I guess), so
                // if they don't it will itself NPE. Thus, in case this mi to be installed
                // does not meet its module deps, we should remove it now.
                ModuleDescription.Dependency[] deps = desc.getDependencies ();
                for (int i = 0; i < deps.length; i++) {
                    if (deps[i].getType () == ModuleDescription.Dependency.TYPE_MODULE) {
                        if (deps[i].checkForMiss (others) != null) {
                            // XXX this might happen even if the dep is on a module in the in-list!
                            // That would prevent a block of modules from being installed at once.
                            it.remove ();
                            continue PREPARE;
                        }
                    }
                }
                // Well module deps at least are OK. Continue with the others.
                */
                ModuleItem mi = (ModuleItem) entry.getValue ();
                mi.ensureClassLoader ();
            }
            descr = ModuleDescription.resolveOrderingForRealInstall (s, map.keySet ());
        }

        List out = new ArrayList (in.size ()); // List (ModuleItem)

        it = descr.iterator ();
        while (it.hasNext ()) {
            out.add (map.get (it.next ()));
        }
        return out;
    }


    private static Collection askAndOrderToEnable( Collection modules ) {

        if ( modules == null ) {
            return null;
        }

        if ( modules.size() == 0 )
            return modules;

        StringBuffer sb = new StringBuffer( 200 );

        sb.append( bundle.getString( "MSG_EnableOthers" ) ).append( "\n" );

        Iterator it = modules.iterator();
        while( it.hasNext() ) {
            sb.append( ((ModuleItem)it.next()).getDescription().getName() );
            sb.append( "\n" ); // NOI18N
        }
        NotifyDescriptor nd = new NotifyDescriptor.Confirmation( sb.toString(), NotifyDescriptor.OK_CANCEL_OPTION );
        TopManager.getDefault().notify( nd );
        if ( nd.getValue() == DialogDescriptor.CANCEL_OPTION )
            return null;
        else {
            Collection orderedModules = null;

            // Order the modules to enable
            try {
                orderedModules = resolveOrdering(
                                     modules,
                                     miSupport.get( ModuleInstallerSupport.ENABLED_MODULE ) );
                return orderedModules;
            }
            catch ( IllegalModuleException e ) {
                TopManager.getDefault().notifyException( e );
                return null;
            }
        }
    }

    /** Utility method gets all modules dependent on the module.
     * Returned in a useful order, e.g. if A is requested, and B and C depend on A,
     * and D depends on B and C, will return [B, C, D]. To disable, disable the dependencies
     * in reverse.
     */
    static List getDependentModules( ModuleItem item ) {
        return miSupport.getDependentModules( item.getDescription(), miSupport.ENABLED_MODULE );
    }

    private static void notifyLater (final NotifyDescriptor nd) {
        javax.swing.SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    TopManager.getDefault ().notify (nd);
                }
            });
    }

}
