/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.util.*;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import org.openide.*;
import org.openide.modules.*;
import org.openide.util.Utilities;

/** ModuleInstallerSupport provides supporting methods for ModuleInstaller. It
 * cares about writing/reading modules to/from XML. Maintains list of modules etc.
 * check dependencies
 */
class ModuleInstallerSupport extends Object implements Comparator {

    /** Extensions for modules in the folders */
    private static final String JAR_EXT = ".jar"; // NOI18N

    /** Installed enabled modules */
    static final int ENABLED_MODULE = 1;
    /** Installed disabled modules */
    static final int DISABLED_MODULE = 2;
    /** New modules added in autoload directory since last start */
    static final int AUTOLOAD_CENTRAL_MODULE = 4;
    /** New modules found in users autoload directory */
    static final int AUTOLOAD_USER_MODULE = 8;
    /** Patches - not used yet */
    //static final int PATCH = 16;
    /** modules which should be enabled, and will be, just wait a moment */
    private static final int RESTORED_MODULE = 32;

    /** Name of file to contain stored modules */
    private static final String INSTALLED_MODULES = "installedModules.xml"; // NOI18N

    /** Central directory for modules */
    private static File centralModuleDirectory = null;

    /** User directory for modules */
    private static File userModuleDirectory = null;

    /** Map between module names and ModuleItems */
    private HashMap nameModuleMap = new HashMap( 77 );

    /** Enabled and disabled - installed (& restored) modules */
    private LinkedList installedModules = new LinkedList();

    /** restored modules only (not yet set up for classloader) */
    private final List restoredModules = new LinkedList ();

    /** Modules found in autoload director(ies) */
    private LinkedList autoloadCentralModules = new LinkedList();

    /** Modules found in local autoload directory */
    private LinkedList autoloadUserModules = new LinkedList();

    /** Ghost modules that are listed as enabled in XML but whose JAR cannot be found.
     * If we find another JAR matching its supposed code name, we use it instead,
     * otherwise it will just be dropped. #9806
     */
    private final Map ghosts = new HashMap (1); // Map<String,AttributeList>

    /** Creates new ModuleRegistry
    */
    ModuleInstallerSupport() {
    }

    /** Initializes the modules loads the previously installed modules
     */
    public void initialize(
        File centralModuleDirectory,
        File userModuleDirectory
    ) {
        ModuleInstallerSupport.centralModuleDirectory = centralModuleDirectory;
        ModuleInstallerSupport.userModuleDirectory = userModuleDirectory;
        
        // initialize modules from classpath
        initClassPathModules (getClass ().getClassLoader ()); 
        
        if (userModuleDirectory == null) return;
        

        ModuleInstaller.err.log ("ModuleInstallerSupport initialize");

        try {
            boolean read = readRegistry();
            ModuleInstaller.err.log ("ModuleInstallerSupport initialize: registry read=" + read);
        }
        catch ( java.io.IOException e ) {
            // Do nothing - leave the registry empty
            ModuleInstaller.err.log ("ModuleInstallerSupport initialize: registry unreadable...");
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.WARNING, e);
        }

        // System.out.println("Looking for new modules" );


        findNewModules();
    }

    /** Gets the central directory of modules
    */
    static File getCentralModuleDirectory() {
        return centralModuleDirectory;
    }

    /** Gets the user modules directory
    */
    static File getUserModuleDirectory() {
        return userModuleDirectory;
    }

    /** Gets all installed parts of given kind
    */
    Collection get( int kind ) {
        ArrayList result = new ArrayList();

        // Standard modules
        Iterator it = installedModules.iterator();
        while( it.hasNext() ) {
            ModuleItem mi = (ModuleItem) it.next();

            if ( ( kind & ENABLED_MODULE ) > 0 && mi.isEnabled() ) {
                result.add( mi );
            }
            else if ( ( kind & DISABLED_MODULE ) > 0 && !mi.isEnabled() ) {
                result.add( mi );
            }

        }

        // Autoload Modules
        if ( ( kind & AUTOLOAD_CENTRAL_MODULE ) > 0 ) {
            result.addAll( autoloadCentralModules );
        }

        // AutloadLocal Modules
        if ( ( kind & AUTOLOAD_USER_MODULE ) > 0 ) {
            result.addAll( autoloadUserModules );
        }

        return result;
    }

    /** Adds module into right into specified collection */
    void add( ModuleItem mi, int kind ) {
        boolean enableIt = false;
        switch ( kind ) {
        case ENABLED_MODULE:
            enableIt = true;
            // FALLTHRU
        case RESTORED_MODULE:
            // FALLTHRU
        case DISABLED_MODULE:
            installedModules.add( mi );
            break;
        case AUTOLOAD_CENTRAL_MODULE:
            autoloadCentralModules.add( mi );
            break;
        case AUTOLOAD_USER_MODULE:
            autoloadUserModules.add( mi );
            break;
        default:
            return;
        }

        nameModuleMap.put( mi.getDescription().getCodeNameBase(), mi );
        if (enableIt) {
            enableLoader (mi);
        }
    }

    /** Just fix up its classloader. */
    void enableLoader (ModuleItem mi) {
        //ModuleInstaller.err.log ("ModuleInstallerSupport adding to loader: " + mi.getDescription ().getCodeNameBase ());
        if (mi.getBase () != ModuleItem.BASE_LOADER) {
            ModuleInstaller.appendToModuleClassLoader (mi);
        }
    }


    /** Removes module from given collection */
    void remove( ModuleItem mi, int kind ) {
        switch ( kind ) {
        case ENABLED_MODULE:
            ModuleInstaller.invalidateModuleClassLoader ();
            // FALLTHRU
        case DISABLED_MODULE:
            installedModules.remove( mi );
            break;
        case AUTOLOAD_CENTRAL_MODULE:
            autoloadCentralModules.remove( mi );
            break;
        case AUTOLOAD_USER_MODULE:
            autoloadUserModules.remove( mi );
            break;
        default:
            return;
        }

        nameModuleMap.remove( mi.getDescription().getCodeNameBase() );
    }

    /** Gets the kind of given module */
    int getKind( ModuleItem mi ) {

        // Standard modules
        if ( installedModules.contains( mi ) ) {
            if ( mi.isEnabled() )
                return ENABLED_MODULE;
            else
                return DISABLED_MODULE;
        }

        // Autoload Central Modules
        if ( autoloadCentralModules.contains( mi ) ) {
            return AUTOLOAD_CENTRAL_MODULE;
        }

        // Autload Local Modules
        if ( autoloadUserModules.contains( mi ) ) {
            return AUTOLOAD_USER_MODULE;
        }

        return -1;
    }

    /** Gets installed part identified by it's key
    */
    ModuleItem get( String key ) {
        return (ModuleItem)nameModuleMap.get( key );
    }

    /** Writes the whole registry to XML file
    */
    void writeRegistry() {

        OutputStream os = null;

        try {
            os = new BufferedOutputStream (
                     new FileOutputStream ( new File ( getUserModuleDirectory(), INSTALLED_MODULES )));
            new XML().write ( os );
        }
        catch ( IOException e ) {
            TopManager.getDefault().notifyException( e );
        }
        finally {
            if ( os != null )
                try {
                    os.close ();
                }
                catch( IOException e ) {
                    ModuleInstaller.err.notify (ErrorManager.INFORMATIONAL, e);
                }
        }

        ModuleInstaller.err.log ("Wrote module registry");
    }
    
    /** Recognizes all modules accessible from given class loader.
     * @param classloader the class loader to initilize loaders from
     */
    private void initClassPathModules (ClassLoader classloader) {
        try {
            java.util.Enumeration en = classloader.getResources ("META-INF/MANIFEST.MF"); // NOI18N

            while (en.hasMoreElements ()) {
                URL u = (URL)en.nextElement();

                Manifest m = new Manifest (u.openStream());
                Attributes attr = m.getMainAttributes ();
                if (attr.getValue (ModuleDescription.TAG_MAGIC) != null) {
                    if (attr.getValue ("OpenIDE-Archive-Locale") != null || // NOI18N
                        attr.getValue ("OpenIDE-Archive-Branding") != null) { // NOI18N
                        // Localized variant of some module, please ignore it...
                        continue;
                    }
                    // looks like NetBeans module
                    try {
                        ModuleItem item = new ModuleItem (u.toExternalForm (), m, classloader);
                        add (item, ModuleInstallerSupport.ENABLED_MODULE);
                    } catch (IllegalModuleException ex) {
                        ModuleInstaller.err.notify (ex);
                    }
                }
            }
        } catch (java.io.IOException ex) {
            ModuleInstaller.err.notify (ex);
        }
    }

    /** Reads the whole registry. The read of registry is not restricted to
    * to reading of XML file, it also checks the module directory for newly
    * installed modules (Autoload feature)
    * @return true if actually read
    * @throws IOException if installedModules.xml was there but could not be read
    */ 
    boolean readRegistry() throws java.io.IOException {

        File f = new File ( getUserModuleDirectory(), INSTALLED_MODULES);
        if (! f.exists ()) {
            f = new File (getCentralModuleDirectory (), INSTALLED_MODULES);
            if (! f.exists ()) {
                return false; // it is OK
            }
        }
        InputStream is = new BufferedInputStream (new FileInputStream (f));
        try {
            new XML().read( /*f.toURL (), */is );
        }
        finally {
            is.close ();
        }
        return true;
    }

    /** Test's wether IDE is runs under multiuser installation */
    static boolean isMultiuser() {
        return !Utilities.compareObjects (getUserModuleDirectory(), getCentralModuleDirectory());
    }

    /** Loadsthe new modules from modules directory.
    */
    void findNewModules() {

        if ( isMultiuser() ) {
            // In multiuser install first scan users directory
            // System.out.println("Looking in USER" );

            findModulesInFolder( getUserModuleDirectory(), AUTOLOAD_USER_MODULE );
        }

        // System.out.println("LOOKING iN CENTRAL" );

        findModulesInFolder( getCentralModuleDirectory(), AUTOLOAD_CENTRAL_MODULE );
    }

    /** Finds all jars in given folder and loads them into collection specified by
    * moduleType
    * @param folder Folder to search
    * @param moduleType Type of the module ( AUTOLOAD_CENTRAL_MODULE || AUTOLOAD_USER_MODULE )
    */
    private void findModulesInFolder( File folder, int moduleType ) {

        // System.out.println("Looking in : " + folder ); // NOI18N

        if ( moduleType != AUTOLOAD_CENTRAL_MODULE && moduleType != AUTOLOAD_USER_MODULE ) {
            throw new InternalError();
        }

        // Find all jarfiles in the folder

        final String[] list = folder.list ( new FilenameFilter() {
                                                public boolean accept( File dir, String name ) {
                                                    return name.toLowerCase().endsWith( JAR_EXT ) &&
                                                        ! name.toLowerCase ().endsWith ("-test.jar"); // NOI18N
                                                }
                                            } );

        if ( list == null || list.length == 0 )
            return;



        // Go through the list and add ModuleItems into the collection
        for ( int i = 0; i < list.length; i++ ) {

            try {
                ModuleItem mi = new ModuleItem (
                                    moduleType == AUTOLOAD_CENTRAL_MODULE ? ModuleItem.BASE_CENTRAL : ModuleItem.BASE_USER,
                                    list[i],
                                    true,
                                    false);

                String cnb = mi.getDescription ().getCodeNameBase ();
                ModuleItem existing = get (cnb);
                AttributeList ghost = (AttributeList) ghosts.get (cnb);

                if ( existing != null ) { // The module already exists
                    int exRelease = existing.getDescription ().getCodeNameRelease ();
                    int miRelease = mi.getDescription ().getCodeNameRelease ();
                    String exSpec = existing.getDescription ().getSpecVersion ();
                    String miSpec = mi.getDescription ().getSpecVersion ();
                    ModuleInstaller.err.log ("Module already exists " + cnb + ": existing.base=" + existing.getBase() + " existing.release=" + exRelease + " existing.spec=" + exSpec + " mi.base="  + mi.getBase() + " mi.release=" + miRelease + " mi.spec=" + miSpec);
                    if ((exRelease == miRelease) && Utilities.compareObjects (exSpec, miSpec)) {
                        ModuleInstaller.err.log ("Duplicate of existing module, ignoring...");
                    } else if (exRelease > miRelease) {
                        ModuleInstaller.err.log ("Way older than existing module, ignoring...");
                    } else if (exSpec != null && (miSpec == null || ModuleDescription.compatibleWith (miSpec, exSpec))) {
                        ModuleInstaller.err.log ("Older than existing module, ignoring...");
                    } else {
                        ModuleInstaller.err.log ("Newer, will update...");
                        mi.setOldRelease (existing.getOldRelease ());
                        mi.setOldSpecVersion (existing.getOldSpecVersion ());
                        if (! mi.isUpdated ()) ModuleInstaller.err.log (ErrorManager.WARNING, "WARNING - module will not really be updated");
                        int kind = getKind (existing);
                        remove (existing, kind);
                        add (mi, kind);
                    }
                } else if (ghost != null) { // #9806
                    int exRelease;
                    try {
                        exRelease = Integer.parseInt (ghost.getValue (ModuleTags.RELEASE));
                    } catch (Exception e) {
                        ModuleInstaller.err.notify (e);
                        exRelease = -1;
                    }
                    int miRelease = mi.getDescription ().getCodeNameRelease ();
                    String exSpec = ghost.getValue (ModuleTags.SPECVERSION);
                    if (exSpec != null && exSpec.length () == 0) exSpec = null;
                    String miSpec = mi.getDescription ().getSpecVersion ();
                    ModuleInstaller.err.log ("Module found as ghost " + cnb + ": existing.release=" + exRelease + " existing.spec=" + exSpec + " mi.base="  + mi.getBase() + " mi.release=" + miRelease + " mi.spec=" + miSpec);
                    mi.setOldRelease (exRelease);
                    mi.setOldSpecVersion (exSpec);
                    String reloadable = ghost.getValue (ModuleTags.RELOADABLE);
                    mi.setReloadable (reloadable != null && Boolean.valueOf (reloadable).booleanValue ());
                    add (mi, RESTORED_MODULE);
                } else { // If the module does not exist, add it into new modules collection
                    add( mi, moduleType );
                }
            }
            catch (IOException e) {
                // wrong module
                TopManager.getDefault ().notifyException (e);
            }
        }
    }

    /** Returns all modules of given kind depending on module
    */
    List getDependentModules( ModuleDescription md, int kind ) {
        List result = new ArrayList();
        getDependentModules( md, get (kind), result );
        try {
            return Utilities.partialSort (result, this, false);
        } catch (Utilities.UnorderableException uue) {
            // Should never happen.
            ModuleInstaller.err.notify (uue);
            return Collections.EMPTY_LIST;
        }
    }

    /** Private/recursive method for retrieving dependencies
    */
    private void getDependentModules( ModuleDescription md, Collection modules, List result ) {
        Iterator it = modules.iterator();
        while ( it.hasNext() ) {
            //ModuleDescription imd = ims[j].getDescription();
            ModuleItem mi = (ModuleItem)it.next();
            ModuleDescription imd = mi.getDescription();

            if ( result.contains( mi ) || md == imd ||
                 md.getCodeNameBase().equals( imd.getCodeNameBase() ) ) {
                continue;
            }

            ModuleDescription.Dependency[] deps = imd.getDependencies();

            for( int k = 0; k < deps.length; k++ ) {
                if ( deps[k].getName().equals ( md.getCodeName ()) ) {
                    if ( !result.contains( mi ) ) {
                        result.add( mi );
                        getDependentModules (imd, modules, result);
                    }
                }
            }
        }
    }

    public int compare (Object o1, Object o2) {
        ModuleDescription m1 = ((ModuleItem) o1).getDescription ();
        ModuleDescription m2 = ((ModuleItem) o2).getDescription ();
        if (m1 == m2) return 0;
        ModuleDescription.Dependency[] deps = m1.getDependencies ();
        for (int i = 0; i < deps.length; i++) {
            if (deps[i].getName ().equals (m2.getCodeName ())) {
                return 1;
            }
        }
        deps = m2.getDependencies ();
        for (int i = 0; i < deps.length; i++) {
            if (deps[i].getName ().equals (m1.getCodeName ())) {
                return -1;
            }
        }
        return 0;
    }

    /** Checks for dependencies on disabled modules. If the module depends on
    * some disabled modules then the method returns collection of modules which 
    * should be enabled. If there are some other unsatisfied dependencies or
    * no problems it returns null.
    */
    Collection checkDependenciesOnDisabled( ModuleDescription md ) {
        Collection result = new ArrayList( 10 );
        return checkDependenciesOnDisabled( md, result ) ? result : null;
    }

    Collection checkDependenciesOnDisabled( Collection moduleItems ) {
        Collection result = new ArrayList( 10 );

        Iterator it = moduleItems.iterator();
        while( it.hasNext() ) {
            ModuleDescription md = ((ModuleItem)it.next()).getDescription();
            if ( !checkDependenciesOnDisabled( md, result ) )
                return null;
        }

        return result;
    }

    private boolean checkDependenciesOnDisabled( ModuleDescription md, Collection result ) {
        ModuleDescription[] otherModules = ModuleInstaller.getModuleDescriptions(
                                               ENABLED_MODULE | DISABLED_MODULE );
        ModuleDescription.Dependency[] deps = md.getDependencies();

        for ( int i = 0; i < deps.length; i++ )  {

            String miss = null;

            try {
                miss = deps[i].checkForMiss (otherModules);
            }
            catch ( IllegalModuleException e ) {
                return false;
            }
            if ( miss != null) {
                return false; // Some dependencies are not satisfied
            }
            else if ( deps[i].getType() != ModuleDescription.Dependency.TYPE_MODULE ) {
                continue;
            }
            else {
                // Dependency is satisfied lets look if the module is enabled
                String moduleName = deps[i].getName();
                int index = moduleName.lastIndexOf( '/' );

                if ( index == -1 ) {
                    index = moduleName.length();
                }

                ModuleItem dm = get( moduleName.substring( 0, index ) );

                if ( getKind( dm ) == DISABLED_MODULE && !result.contains( dm ) ) {
                    result.add( dm );
                    if ( !checkDependenciesOnDisabled( dm.getDescription(), result ) )
                        return false;
                }
            }
        }
        return true;
    }


    // InnerClasses -----------------------------------------------------------


    /** XML writes and reads all modules to / from XML.
    */
    private class XML extends HandlerBase {

        /** base URL */
        //private URL base;

        /** Write the document to stream.
        * @param os output stream
        * @param list of ModuleItems
        * @exception IOException if I/O error occured
        */
        void write( OutputStream os ) throws IOException {

            Collection list = get( ENABLED_MODULE | DISABLED_MODULE );

            PrintStream p = new PrintStream( os );

            p.print( ModuleTags.XML_HEADER );
            p.print( ModuleTags.NEW_LINE + ModuleTags.NEW_LINE );
            p.print( ModuleTags.MODULES_HEADER );
            p.print( ModuleTags.NEW_LINE );

            Iterator it = list.iterator ();
            while (it.hasNext ()) {
                ModuleItem mi = (ModuleItem)it.next ();

                p.print( mi.toXML() );
                p.print( ModuleTags.NEW_LINE );
            }

            p.print( ModuleTags.MODULES_FOOTER );
            p.print( ModuleTags.NEW_LINE );

            p.close ();
        }

        /** Parse the input stream.
        * @param url the base URL to referer all relative to
        * @return list of ModuleItems
        * @exception IOException if I/O error occured
        */
        void read(/* URL url, */InputStream is ) throws IOException {
            //base = url;

            Parser p = org.openide.loaders.XMLDataObject.createParser ();

            p.setDocumentHandler ( this );
            try {
                p.parse (new InputSource (is));
            } catch (SAXException e) {
                IOException ioe = new IOException ();
                TopManager.getDefault ().getErrorManager ().annotate (ioe, e);
                throw ioe;
            }
            return;
        }

        /** Creates new XML parser/saver for given list of ModuleItems.
         */
        XML () {
        }

        /** Creates the right type of module item and adds it to list of modules.
         */
        public void startElement (String name, AttributeList attr) {
            try {
                
                if ( name.equals( ModuleTags.MODULE )  ) {
                    ModuleItem mi = ModuleItem.fromXML( attr/*, base*/ );
                    if ( mi.isEnabled() )
                        add( mi, ModuleInstallerSupport.RESTORED_MODULE );
                    else
                        add( mi, ModuleInstallerSupport.DISABLED_MODULE );
                }

            } catch (FileNotFoundException fnfe) {
                // OK, some module disappeared on us!
                // If it was already disabled, so just ignore it.
                // If it was enabled, remember its basic characteristics
                // so that if an "autoload" module shows up as a match, we
                // silently slip in the autoloaded JAR in its place.
                // See #9806: e.g. projectx.jar -> projects.jar.
                String enabled = attr.getValue (ModuleTags.ENABLED);
                if (enabled == null || Boolean.valueOf (enabled).booleanValue ()) {
                    ModuleInstaller.err.log ("Tracking missing enabled module as ghost: " + fnfe);
                    String codenamebase = attr.getValue (ModuleTags.CODENAMEBASE);
                    ghosts.put (codenamebase, new AttributeListImpl (attr));
                } else {
                    ModuleInstaller.err.log ("Ignoring missing disabled module: " + fnfe);
                }
            } catch (IOException ioe) {
                TopManager.getDefault ().notifyException (ioe);
            }
        }

    }

    public String toString() {
        StringBuffer sb = new StringBuffer( 500 );


        sb.append("ENABLED --------------------------\n" ); // NOI18N
        sb.append( modulesToString( get( ModuleInstallerSupport.ENABLED_MODULE ) ) );
        sb.append( "\n" ); // NOI18N
        sb.append("DISABLED --------------------------\n" ); // NOI18N
        sb.append( modulesToString( get( ModuleInstallerSupport.DISABLED_MODULE ) ) );
        sb.append( "\n" ); // NOI18N
        sb.append("CENTRAL --------------------------\n" ); // NOI18N
        sb.append( modulesToString( get( ModuleInstallerSupport.AUTOLOAD_CENTRAL_MODULE ) ) );
        sb.append( "\n" ); // NOI18N
        sb.append("USER --------------------------\n" ); // NOI18N
        sb.append( modulesToString( get( ModuleInstallerSupport.AUTOLOAD_USER_MODULE ) ) );
        sb.append( "\n" ); // NOI18N

        return sb.toString();
    }

    String modulesToString( Collection modules ) {
        StringBuffer sb = new StringBuffer( 100 );
        Iterator it = modules.iterator();
        while( it.hasNext() ) {
            ModuleItem mi = (ModuleItem)it.next();
            sb.append( " " + mi.getDescription().getName() + mi.getJar () ); // NOI18N
            sb.append( "\n" ); // NOI18N
        }

        return sb.toString();
    }

}
