/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.io.*;
import java.security.PermissionCollection;
import java.security.CodeSource;
import java.security.Permissions;
import java.security.AllPermission;
import java.util.*;
import java.util.jar.*;
import javax.swing.SwingUtilities;

import org.xml.sax.AttributeList;

import org.openide.modules.ManifestSection;
import org.openide.modules.ModuleDescription;
import org.openide.modules.IllegalModuleException;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.MultiFileSystem;
import org.openide.filesystems.XMLFileSystem;
import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.NotifyDescriptor;
import org.openide.util.NbBundle;

import org.netbeans.core.projects.ModuleLayeredFileSystem;

/**
* Item representing one module.
*
* @author Jaroslav Tulach, Petr Hrebejk, Jesse Glick
*/
final class ModuleItem extends Object {

    /** Constants for module bases */
    static final int BASE_USER = 1;
    static final int BASE_CENTRAL = 2;
    static final int BASE_JAR = 3;
    static final int BASE_LOADER = 4;

    /** is enabled or not */
    private boolean enabled;

    /** can be reloaded easily */
    private boolean reloadable = false;

    /** Old release number of the module */
    private transient int oldRelease = -1;

    /** Old specification version of the module */
    private transient String oldSpecVersion = null;

    /** Description of the module */
    private transient ModuleDescription descr;

    /** The module was deleted from autoload folder */
    private transient boolean deleted = false;

    /** Base of the module local|cenral|file */
    private int base = BASE_USER;

    /** The jar file the module is loaded from - used for modules
    * with base == BASE_CENTRAL || BASE_USER || BASE_JAR )
    */
    private final File moduleFile;
    /** the actual JAR used; normally null unless reloadable==true */
    private File physicalModuleFile = null;

    /** Used as a flag to tell if this module was really successfully released.
     * Currently does not work, so if it cannot be made to work, delete it.
     * (Someone seems to be holding a strong reference to the classloader--who?!)
     */
    private transient boolean released;
    /** Count which release() call is really being checked. */
    private transient int releaseCount = 0;

    /** Map from extension JARs to sets of JAR that load them via Class-Path.
     * Used only for debugging purposes, so that a warning is printed if two
     * different modules try to load the same extension (which would cause them
     * to both load their own private copy, which may not be intended).
     */
    private static final Map extensionOwners = new HashMap (); // Map<File,Set<File>>

    /** Set of locale-variants JARs for this module.
     * Added explicitly to classloader, and can be used by execution engine.
     */
    private final Set localeVariants = new HashSet (); // Set<File>
    /** Set of extension JARs that this module loads via Class-Path.
     * Can be used e.g. by execution engine. (#9617)
     */
    private final Set plainExtensions = new HashSet (); // Set<File>
    /** Set of localized extension JARs derived from plainExtensions.
     * Used to add these to the classloader. (#9348)
     * Can be used e.g. by execution engine.
     */
    private final Set localeExtensions = new HashSet (); // Set<File>

    /** Creates new module for Jar file.
    * @param base Base of the module BASE_USER || BASE_CENTRAL || BASE_JAR
    * @param name path to the module jar
    * @param enabled enabled?
    * @exception IllegalModuleException if the jar is not module
    * @exception IOException if an error occures during reading
    */
    public ModuleItem (int base, String location, boolean enabled, boolean reloadable)
    throws IllegalModuleException, FileNotFoundException,
        MalformedURLException, IOException {

        this.enabled = enabled;
        this.reloadable = reloadable;
        this.base = base;

        switch( base ) {
        case BASE_USER:
            moduleFile = new File( ModuleInstallerSupport.getUserModuleDirectory(), location );
            break;
        case BASE_CENTRAL:
            moduleFile = new File( ModuleInstallerSupport.getCentralModuleDirectory(), location );
            break;
        case BASE_JAR:
            moduleFile = new File( location );
            break;
        default:
            throw new IllegalArgumentException ( "Unknown module base " ); // NOI18N
        }

        if ( ! moduleFile.canRead() ) {
            throw new FileNotFoundException( "Can't find module jarfile : "  + moduleFile.toString() ); // NOI18N
        }

        this.descr = new ModuleDescription (moduleFile.getName (), loadManifest (), null);
    }
    
    /** Constructor for module item provided by existing class loader.
     * @param manifest created manifest
     * @param cl class loader
     * @exception IllegalModuleException if the jar is not module
     */
    public ModuleItem (String name, Manifest manifest, ClassLoader cl) throws IllegalModuleException {
        this.enabled = true;
        this.reloadable = false;
        this.base = BASE_LOADER;
        this.moduleFile = null;

        this.descr = new ModuleDescription (name, manifest, cl);
    }

    private Manifest loadManifest () throws IOException {
        JarFile jarFile;
        if (reloadable) {
            ensurePhysicalModuleFile ();
            jarFile = new JarFile (physicalModuleFile);
        } else {
            jarFile = new JarFile (moduleFile);
        }
        localeVariants.clear ();
        localeVariants.addAll (findVariantsOf (moduleFile));
        try {
            Manifest m = jarFile.getManifest ();
            plainExtensions.clear ();
            localeExtensions.clear ();
            String classPath = m.getMainAttributes ().getValue (Attributes.Name.CLASS_PATH);
            if (classPath != null) {
                StringTokenizer tok = new StringTokenizer (classPath);
                while (tok.hasMoreTokens ()) {
                    String ext = tok.nextToken ();
                    File extfile = new File (moduleFile.getParentFile (), ext.replace ('/', File.separatorChar));
                    if (! extfile.exists ()) {
                        // Ignore unloadable extensions.
                        continue;
                    }
                    synchronized (extensionOwners) {
                        Set owners = (Set) extensionOwners.get (extfile);
                        if (owners == null) {
                            owners = new HashSet (1);
                            owners.add (moduleFile);
                            extensionOwners.put (extfile, owners);
                        } else if (! owners.contains (moduleFile)) {
                            owners.add (moduleFile);
                            System.err.println ("Warning: the extension " + extfile + " might be multiply loaded by " + owners); // NOI18N
                        } // else already know about it (OK or warned)
                    }
                    plainExtensions.add (extfile);
                    localeExtensions.addAll (findVariantsOf (extfile));
                }
            }
            return m;
        } finally {
            jarFile.close();
        }
    }

    /** find existing locale variants of f */
    private static Set/*<File>*/ findVariantsOf (File f) {
        if (! f.isFile ()) {
            return Collections.EMPTY_SET;
        }
        File dir = new File (f.getParentFile (), "locale"); // NOI18N
        if (! dir.isDirectory ()) {
            return Collections.EMPTY_SET;
        }
        Set s = new HashSet (1); // Set<File>
        String nameExt = f.getName ();
        int idx = nameExt.lastIndexOf ('.'); // NOI18N
        String name, ext;
        if (idx != -1) {
            name = nameExt.substring (0, idx);
            ext = nameExt.substring (idx);
        } else {
            name = nameExt;
            ext = ""; // NOI18N
        }
        Iterator it = NbBundle.getLocalizingSuffixes ();
        while (it.hasNext ()) {
            String suffix = (String) it.next ();
            File v = new File (dir, name + suffix + ext);
            if (v.isFile ()) {
                s.add (v);
            }
        }
        return s;
    }

    private void ensurePhysicalModuleFile () throws IOException {
        if (physicalModuleFile == null && reloadable) {
            physicalModuleFile = makeTempJar (moduleFile);
        }
    }

    /** PermissionCollection with an instance of AllPermission. */
    private static PermissionCollection modulePermissions;
    /** @return initialized @see #modulePermission */
    private static PermissionCollection getAllPermission() {
        if (modulePermissions == null) {
            synchronized (ModuleItem.class) {
                if (modulePermissions == null) {
                    modulePermissions = new Permissions();
                    modulePermissions.add(new AllPermission());
                    modulePermissions.setReadOnly();
                }
            }
        }
        return modulePermissions;
    }
    /** Class loader to load a single module.
     * Auto-localizing, multi-parented, permission-granting, the works.
     */
    private class OneModuleClassLoader extends MultiURLClassLoader {
        private int rc;
        public final URL[] classpath;
        /** Create a new loader for a module.
         * @param urls the URLs to explicitly add; includes the module itself, its locale variants, and variants of extensions
         * @param classpath the URLs formed by Class-Path which URLClassLoader will add itself
         * @param parents a list of parent classloaders (from other modules)
         */
        public OneModuleClassLoader (URL[] urls, URL[] classpath, ClassLoader[] parents) {
            super (urls, parents);
            this.classpath = classpath;
            rc = releaseCount++;
        }
        /** Inherited.
         * @param cs is ignored
         * @return PermissionCollection with an AllPermission instance
         */
        protected PermissionCollection getPermissions(CodeSource cs) {
            return getAllPermission();
        }
        /** look for JNI libraries also in modules/bin/ */
        protected String findLibrary (String libname) {
            String mapped = System.mapLibraryName (libname);
            String nbuser = System.getProperty ("netbeans.user");
            if (nbuser != null) {
                File lib = new File (new File (new File (nbuser, "modules"), "bin"), mapped); // NOI18N
                if (lib.isFile ()) {
                    return lib.getAbsolutePath ();
                }
            }
            String nbhome = System.getProperty ("netbeans.home");
            if (nbhome != null && ! nbhome.equals (nbuser)) {
                File lib = new File (new File (new File (nbhome, "modules"), "bin"), mapped); // NOI18N
                if (lib.isFile ()) {
                    return lib.getAbsolutePath ();
                }
            }
            return null;
        }
        public String toString () {
            return super.toString () + "[" + descr.getCodeNameBase () + "#" + rc + "]"; // NOI18N
        }
        protected void finalize () {
            ModuleInstaller.err.log ("Finalize for " + this + ": rc=" + rc + " releaseCount=" + releaseCount + " released=" + released);
            if (rc == releaseCount) {
                // Hurrah! release() worked.
                released = true;
            } else {
                ModuleInstaller.err.log ("Now resources for " + descr.getName () + " have been released.");
            }
        }
    }

    private static File makeTempJar (File moduleFile) throws IOException {
        // [PENDING] JDK 1.3 provides ZipFile(File,OPEN_READ|OPEN_DELETE).
        // Would this be useful here?
        String prefix = moduleFile.getName ();
        if (prefix.endsWith (".jar") || prefix.endsWith (".JAR")) { // NOI18N
            prefix = prefix.substring (0, prefix.length () - 4);
        }
        if (prefix.length () < 3) prefix += '.';
        if (prefix.length () < 3) prefix += '.';
        if (prefix.length () < 3) prefix += '.';
        File physicalModuleFile = File.createTempFile (prefix, "-test.jar", moduleFile.getParentFile ()); // NOI18N
        physicalModuleFile.deleteOnExit ();
        InputStream is = new FileInputStream (moduleFile);
        try {
            OutputStream os = new FileOutputStream (physicalModuleFile);
            try {
                byte[] buf = new byte[4096];
                int i;
                while ((i = is.read (buf)) != -1) {
                    os.write (buf, 0, i);
                }
            } finally {
                os.close ();
            }
        } finally {
            is.close ();
        }
        ModuleInstaller.err.log ("Made " + physicalModuleFile);
        return physicalModuleFile;
    }

    /** Class loader to use when loading this module.
     * @return the class loader
     */
    protected ClassLoader createClassLoader () {
        // Find classloaders for dependent modules and parent to them.
        ModuleDescription.Dependency[] dependencies = descr.getDependencies ();
        List loaders = new ArrayList (Math.max (1, dependencies.length));
        for (int i = 0; i < dependencies.length; i++) {
            if (dependencies[i].getType () != ModuleDescription.Dependency.TYPE_MODULE) {
                continue;
            }
            String codeNameBase = dependencies[i].getName ();
            int slash = codeNameBase.lastIndexOf ('/');
            if (slash != -1) {
                codeNameBase = codeNameBase.substring (0, slash);
            }
            ModuleItem other = ModuleInstaller.getByName (codeNameBase);
            if (other == null) {
                // This should not normally happen. But if you e.g. try to install a new module A
                // depends on B, where B is nowhere to be found, ModuleInstaller will call ensureClassLoader
                // on A before checking its dependencies (to make sure package deps work). Ideally
                // it would check all module deps first, then ensure classloaders, then check all package
                // deps. Since that would require a lot of rewriting of ModuleInstaller, for now just
                // work around the situation by not giving a NPE here.
                ModuleInstaller.err.log ("ModuleItem " + descr.getCodeNameBase () + ": no dep " + codeNameBase + " to base classloader on; ignoring");
                continue;
            }
            other.ensureClassLoader ();
            loaders.add (other.descr.getClassLoader ());
        }
        // [PENDING] provide some kind of mechanism for modules/patches/ loading (#9273)
        // (must be more complicated, since need separate patch dir per-module, perhaps)
        List urls = new ArrayList (1); // List<URL>
        List classpath = new ArrayList (1); // List<URL>
        try {
            if (reloadable) {
                ensurePhysicalModuleFile ();
                urls.add (physicalModuleFile.toURL ());
            } else {
                urls.add (moduleFile.toURL ());
            }
            // URLClassLoader would not otherwise find these, so:
            Iterator it = localeVariants.iterator ();
            while (it.hasNext ()) {
                urls.add (((File) it.next ()).toURL ());
            }
            it = localeExtensions.iterator ();
            while (it.hasNext ()) {
                urls.add (((File) it.next ()).toURL ());
            }
            it = plainExtensions.iterator ();
            while (it.hasNext ()) {
                classpath.add (((File) it.next ()).toURL ());
            }
        } catch (IOException ioe) {
            TopManager.getDefault ().notifyException (ioe);
        }
        return new OneModuleClassLoader ((URL[]) urls.toArray (new URL[urls.size ()]),
                                         (URL[]) classpath.toArray (new URL[classpath.size ()]),
                                         (ClassLoader[]) loaders.toArray (new ClassLoader[loaders.size ()]));
    }

    void ensureClassLoader () {
        if (descr.getClassLoader () == null) {
            descr.setClassLoader (createClassLoader ());
        }
    }

    /** Try to release its classloader and associated objects completely.
     * Hopefully after this method is called there will be no operating-system-
     * level locks held on the JAR file.
     * See #4405789.
     */
    void release () {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();
        ClassLoader l = descr.getClassLoader ();
        if (l != null && (l instanceof MultiURLClassLoader)) {
            // Remove references to it from other places, i.e. ModuleClassLoader:
            ((MultiURLClassLoader) l).destroy ();
        }
        l = null;
        descr.setClassLoader (null); // clear caches of ModuleInstaller, sections, ...
        ModuleInstaller.err.log ("ModuleItem.release on " + descr.getName () + ": releaseCount=" + releaseCount + " released=" + released);
        released = false;
        System.gc (); // hope OneModuleClassLoader.finalize() is called...
        /* [PENDING] for some reason does not actually work anyway:
        if (! reloadable && ! released) {
            // ResourceBundle stupidly holds onto classloaders with soft references.
            // See #4405807.
            // That means that the loader may still be softly reachable at this point.
            // Try harder to force the GC to kill the cache key.
            ModuleInstaller.err.log ("Trying hard to release " + moduleFile + " ...");
            List crud = new ArrayList (1000);
            try {
                while (true) {
                    crud.add (new byte[1024 * 1024]);
                }
            } catch (OutOfMemoryError oome) {
                // Thanks!
            }
            crud.clear ();
            crud = null;
            System.gc ();
        }
        */
        if (! released) {
            ModuleInstaller.err.log (ErrorManager.WARNING, "WARNING: not all resources associated with module " + moduleFile + " were successfully released.");
            released = true;
        } else {
            ModuleInstaller.err.log ("All resources associated with module " + moduleFile + " were successfully released.");
        }
        if (physicalModuleFile != null) {
            if (! physicalModuleFile.delete ()) {
                ModuleInstaller.err.log (ErrorManager.WARNING, "WARNING: temporary JAR " + physicalModuleFile + " not currently deletable.");
            } else {
                ModuleInstaller.err.log ("deleted: " + physicalModuleFile);
            }
            physicalModuleFile = null;
        } else {
            ModuleInstaller.err.log ("no physicalModuleFile to delete for " + descr.getCodeNameBase ());
        }
    }

    /** Try to reload it from the JAR. */
    void reload () throws IOException, IllegalModuleException {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();
        descr.setClassLoader (null);
        String codeName = descr.getCodeName ();
        descr = new ModuleDescription (moduleFile.getName (), loadManifest (), null);
        if (! codeName.equals (descr.getCodeName ())) {
            // [PENDING] I18N
            throw new IllegalModuleException ("Attempt to reload module with different code name! " + codeName + " -> " + descr.getCodeName ());
        }
        ModuleInstaller.err.log ("ModuleItem.reload: " + descr.getName());
    }

    /** @param en true if enabled
    */
    public void setEnabled (boolean en) {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();
        if (enabled == en) return;

        enabled = en;
        ModuleInstaller.changeEnabled (this);
    }

    void setEnabledNoNotify( boolean en ) {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();
        enabled = en;
    }


    /** @return true if enabled
    */
    public boolean isEnabled () {
        return enabled;
    }

    public boolean isReloadable () {
        return reloadable;
    }

    public void setReloadable (boolean r) {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();
        reloadable = r;
    }

    /** @return Representation of ModuleItem location in XML file */
    private String getLocationForXML () {
        if (base == BASE_LOADER) throw new IllegalArgumentException ();

        switch ( base ) {
        case BASE_CENTRAL:
        case BASE_USER:
            return moduleFile.getName();
        case BASE_JAR:
            return moduleFile.getPath();
        }
        throw new IllegalStateException ();
    }

    public File getJar () {
        return moduleFile;
    }

    public String toString () {
        return "ModuleItem[jar=" + moduleFile + ",name=" + descr.getCodeName () + ",spec=" + descr.getSpecVersion () + "]"; // NOI18N
    }

    /** Get all JARs loaded by this module.
     * Includes the module itself, any locale variants of the module,
     * any extensions specified with Class-Path, any locale variants
     * of those extensions.
     * If patches are implemented (#9273), the list will be in classpath order (patches first).
     * Currently the temp JAR is provided in the case of test modules, to prevent
     * sporadic ZIP file exceptions when background threads (like Java parsing) tries
     * to open libraries found in the library path.
     * JARs already present in the classpath are <em>not</em> listed.
     * @return a <code>List&lt;File&gt;</code> of JARs
     */
    public List getAllJars () {
        List l = new ArrayList (); // List<File>
        if (base != BASE_LOADER) {
            l.add (reloadable ? physicalModuleFile : moduleFile);
        }
        l.addAll (plainExtensions);
        l.addAll (localeVariants);
        l.addAll (localeExtensions);
        return l;
    }

    /** @return Release of the module from previous start of the Ide or -1 */
    public int getOldRelease() {
        return oldRelease;
    }

    /** Old release is set only by ModuleInstaller */
    void setOldRelease( int oldRelease ) {
        this.oldRelease = oldRelease;
    }

    /** @return Specification version of the module from previous start
     * of the IDE or <CODE>null</CODE> */
    public String getOldSpecVersion() {
        return oldSpecVersion;
    }

    /** Old release is set only by ModuleInstaller */
    void setOldSpecVersion( String oldSpecVersion ) {
        this.oldSpecVersion = oldSpecVersion;
    }

    /** Test if the module was updated after last shutdown of IDE
    */
    boolean isUpdated() {
        if ( descr.getCodeNameRelease() > oldRelease ) {
            return true;
        }

        try {
            if ( !ModuleDescription.compatibleWith( descr.getSpecVersion(), oldSpecVersion ) ) {
                return true;
            }
        }
        catch ( org.openide.modules.IllegalModuleException e ) {
            TopManager.getDefault().notifyException( e );
        }

        return false;
    }

    /** Getter for the module base */
    public int getBase() {
        return base;
    }

    /** Getter for description */
    public ModuleDescription getDescription() {
        return descr;
    }

    /** Is the deletion of this item allowed */
    public boolean canDestroy() {
        return base != BASE_CENTRAL && base != BASE_USER && base != BASE_LOADER;
    }
    
    /** Is it allowed to disable this module?
     */
    public boolean canDisable () {
        return base != BASE_LOADER;
    }

    /** Get the actual URL to point to this module's help set.
     * If bundled in the JAR, then nbresloc: will be used (permitting crosslinking).
     * Currently systemClassLoader actually provides the resources, not the module classloader.
    * @return the URL to the help set file, or null if no help set was specified, or if it could not be found
    */
    private URL getRealHelpSet () {
        ensureClassLoader ();
        String resource = descr.getDescriptionResource ();
        if (resource == null) return null;
        URL u = Help.createDocsURL (resource);
        try {
            // Sanity-check it.
            u.openConnection ().connect ();
            return u;
        } catch (IOException ioe) {
            warnAboutMissingHelp (ioe);
            return null;
        }
    }

    private void warnAboutMissingHelp (final Exception e) {
        TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, e);
        if (ModuleInstaller.err.isLoggable (ErrorManager.INFORMATIONAL)) {
            ClassLoader loader = Help.getDocsLoader ();
            ModuleInstaller.err.log ("docsLoader=" + loader);
            maybeDumpLoaderURLs (loader);
        }
        String name = descr.getName ();
        if (name == null) name = descr.getCodeName ();
        final String _name = name;
        SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    TopManager.getDefault ().notify
                        (new NotifyDescriptor.Message
                            (NbBundle.getMessage
                             (ModuleItem.class, "MSG_no_help_set_found_skipping", _name, e.getLocalizedMessage ()),
                             NotifyDescriptor.WARNING_MESSAGE));
                }
            });
    }

    /** Initializes defaults.
    */
    public void restoreDefault () {
        try {
            URL helpSet = getRealHelpSet ();
            if (helpSet != null) {
                Help.getDefault ().addHelpSet (helpSet, descr.getCodeName (), descr.getName (), descr.getClassLoader ());
            }
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Deinitializes them.
    */
    public void unrestoreDefault () {
        ensureClassLoader ();
        try {
            URL helpSet = getRealHelpSet ();
            if (helpSet != null) {
                Help.getDefault ().removeHelpSet (helpSet);
            }
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Called when the module is installed for the first time
    */
    public void installSection () {
        ensureClassLoader ();
        ModuleActions.attachTo (this);
        try {
            descr.forEachSection (INSTALL);
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        } finally {
            ModuleActions.attachTo (null);
        }
    }

    /** Called when the module is restored on startup
    */
    public void restoreSection () {
        ensureClassLoader ();
        ModuleActions.attachTo (this);
        try {
            restoreSection (descr);
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        } finally {
            ModuleActions.attachTo (null);
        }
    }

    private static void restoreSection (ModuleDescription descr) throws Exception {
        descr.forEachSection (RESTORE);
    }

    /** Called when the module should be uninstalled.
    */
    public void uninstallSection () {
        ensureClassLoader ();
        ModuleActions.attachTo (this);
        try {
            descr.forEachSection (REMOVE);
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        } finally {
            ModuleActions.attachTo (null);
        }
    }

    /** Called when the module is installed for the first time
    */
    public void installCode () {
        ModuleInstaller.err.log ("Install module: " + getDescription ().getCodeNameBase ());
        ensureClassLoader ();
        try {
            descr.getModule ().installed ();
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Called when the module is restored on startup
    */
    public void restoreCode () {
        ModuleInstaller.err.log ("Restore module: " + getDescription ().getCodeNameBase ());
        ensureClassLoader ();
        try {
            descr.getModule ().restored ();
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Called when the module should be uninstalled.
    */
    public void uninstallCode () {
        ModuleInstaller.err.log ("Uninstall module: " + getDescription ().getCodeNameBase ());
        ensureClassLoader ();
        try {
            descr.getModule ().uninstalled ();
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** Called when the module was updated to higher version.
    */ 
    public void updatedCode() {
        ModuleInstaller.err.log ("Updating module " + getDescription ().getCodeNameBase () + // NOI18N
                                 " from /" + oldRelease + " " + oldSpecVersion + // NOI18N
                                 " to /" + getDescription ().getCodeNameRelease () + " " + getDescription ().getSpecVersion ()); // NOI18N
        ensureClassLoader ();
        try {
            descr.getModule ().updated( oldRelease, oldSpecVersion );
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }
    
    /** Called when module is installed or restored, to add its layer. */
    public void restoreLayer () {
        ensureClassLoader ();
        try {
            restoreLayer (descr, getHostingFileSystem ());
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    private static void restoreLayer (ModuleDescription descr, ModuleLayeredFileSystem hostingFileSystem) throws Exception {
            String s = descr.getLayerResource ();
            if (s != null) {
                // Actually add a sequence of layers, in locale order.
                String base, ext;
                int idx = s.lastIndexOf ('.');
                if (idx == -1) {
                    base = s;
                    ext = "";
                } else {
                    base = s.substring (0, idx);
                    ext = s.substring (idx);
                }
                ClassLoader cl = descr.getClassLoader ();
                List fss = new ArrayList ();
                Iterator it = NbBundle.getLocalizingSuffixes ();
                while (it.hasNext ()) {
                    String suffix = (String) it.next ();
                    String resource = base + suffix + ext;
                    URL u = cl.getResource (resource);
                    if (u != null) {
                        fss.add (new XMLFileSystem (u));
                    }
                }
                FileSystem layer;
                if (fss.size () == 0) {
                    throw new MissingResourceException ("Module layer not found", null, s); // NOI18N
                } else if (fss.size () == 1) {
                    // Optimize for the common case:
                    layer = (FileSystem) fss.get (0);
                } else {
                    final FileSystem[] systems = (FileSystem[]) fss.toArray (new FileSystem[fss.size ()]);
                    class LocaleMergedFileSystem extends MultiFileSystem {
                        public LocaleMergedFileSystem () {
                            super (systems);
                            // Ensure that layers from one module can mask files
                            // from other modules. E.g. projects module wants to mask
                            // some menu items supplied by the core.
                            setPropagateMasks (true);
                        }
                    }
                    // Merge them all together and add that:
                    layer = new LocaleMergedFileSystem ();
                }
                hostingFileSystem.addModuleLayer (descr.getCodeName (), layer);
            }
    }

    /** Called when module is uninstalled, to remove its layer. */
    public void unrestoreLayer () {
        try {
            getHostingFileSystem ().removeModuleLayer (descr.getCodeName ());
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    private ModuleLayeredFileSystem getHostingFileSystem () {
        switch (base) {
        case BASE_CENTRAL:
        case BASE_LOADER:
            return ModuleLayeredFileSystem.getInstallationModuleLayer ();
        case BASE_USER:
        case BASE_JAR:
            // also covers test modules, should be in BASE_USER by default:
            return ModuleLayeredFileSystem.getUserModuleLayer ();
        default:
            throw new InternalError ();
        }
    }

    /** Called when the IDE is closing
    * @return false if the IDE should not close
    */
    public boolean closing () {
        ensureClassLoader ();
        try {
            return descr.getModule ().closing ();
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
            // if there is an error, assume the module is completely bad
            return true;
        }
    }

    /** Called when the IDE is finallty closing. All modules accepted the close
    * of the IDE
    */
    public void closeCode() {
        ensureClassLoader ();
        try {
            descr.getModule ().close();
        } catch (ThreadDeath t) {
            throw t;
        } catch (Throwable t) {
            TopManager.getDefault ().notifyException (t);
        }
    }

    /** May be called after a batch of things has been done to a module.
     * (For example, restoring sections, default, and code.)
     * Will check to see if any of the classes loaded by the module
     * were in fact not loaded with the intended classloader, and
     * if so warn the user.
     */
    public void verifyClassLoader () {
        if (System.getProperty("netbeans.moduleitem.dontverifyclassloader") != null)
            return;
        
        Set s = descr.getBadClasses ();
        if (s.isEmpty ()) return;
        String name = descr.getName ();
        if (name == null) name = descr.getCodeName ();
        TopManager.getDefault ().notify
            (new NotifyDescriptor.Message
                (NbBundle.getMessage (ModuleItem.class, "MSG_warn_incorrect_classloaders_found", name),
                 NotifyDescriptor.WARNING_MESSAGE));
        System.err.println ("[WARNING] Bad classloader found in " + name + " for classes:"); // NOI18N
        Iterator it = s.iterator ();
        while (it.hasNext ()) {
            Class clazz = (Class) it.next ();
            ClassLoader loader = clazz.getClassLoader ();
            System.err.println ("\t" + clazz.getName () + " from loader " + loader); // NOI18N
            maybeDumpLoaderURLs (loader);
        }
        ClassLoader loader = descr.getClassLoader ();
        System.err.println ("\tExpected class loader: " + loader); // NOI18N
        maybeDumpLoaderURLs (loader);
    }

    /** If supplied loader is a URLClassLoader, dump its URLs to syserr. */
    static void maybeDumpLoaderURLs (ClassLoader loader) {
        if (loader != null) {
            // Recurse since URLClassLoader only lists URLs it adds:
            if (loader instanceof MultiURLClassLoader) {
                // Handled specially.
                ((MultiURLClassLoader) loader).dumpURLs ();
            } else {
                maybeDumpLoaderURLs (loader.getParent ());
            }
            if (loader instanceof URLClassLoader) {
                URLClassLoader urlloader = (URLClassLoader) loader;
                URL[] urls = urlloader.getURLs ();
                for (int i = 0; i < urls.length; i++) {
                    System.err.println ("\t\t[" + urlloader + "] " + urls[i]); // NOI18N
                }
                if (loader instanceof OneModuleClassLoader) {
                    // We also know what the Class-Path entries are in this case, so show them.
                    urls = ((OneModuleClassLoader) loader).classpath;
                    for (int i = 0; i < urls.length; i++) {
                        System.err.println ("\t\t[" + urlloader + "] " + urls[i]); // NOI18N
                    }
                }
            } else {
                System.err.println ("\t\t[" + loader + "] not a URLClassLoader"); // NOI18N
            }
        }
    }


    /** Mehod called by ModuleInstaller to get the XML format of ModuleItem
     */
    String toXML() {
        if (base == BASE_LOADER) {
            // module is not persistent was added to class path
            return "";
        }

        StringBuffer moduleBuffer = new StringBuffer( 150 );

        /*
        if ( isDeleted() && !( base == BASE_CENTRAL || base == BASE_USER ) )
          return moduleBuffer.toString();
        */

        // Module header
        moduleBuffer.append( ModuleTags.TAB ).append("<" ).append( ModuleTags.MODULE ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // CodeBaseName
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.CODENAMEBASE ).append( "=\"" ); // NOI18N
        moduleBuffer.append( getDescription().getCodeNameBase() ).append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // Version
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.RELEASE ).append( "=\"" ); // NOI18N
        moduleBuffer.append( getDescription().getCodeNameRelease() ).append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.TAB ).append( ModuleTags.SPECVERSION ).append( "=\"" ); // NOI18N
        moduleBuffer.append( getDescription().getSpecVersion() ).append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // Base
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.BASE ).append( "=\"" ); // NOI18N
        switch ( base ) {
        case BASE_USER:
            moduleBuffer.append( ModuleTags.BASE_USER );
            break;
        case BASE_CENTRAL:
            moduleBuffer.append( ModuleTags.BASE_CENTRAL );
            break;
        case BASE_JAR:
            moduleBuffer.append( ModuleTags.BASE_JAR );
            break;
        }
        moduleBuffer.append( "\"" ).append( ModuleTags.NEW_LINE ); // NOI18N

        // URL
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.URL ).append( "=\"" ); // NOI18N
        moduleBuffer.append( getLocationForXML() );
        moduleBuffer.append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // Enabled
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.ENABLED ).append( "=\"" ); // NOI18N
        moduleBuffer.append( new Boolean (isEnabled ()) ).append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // Reloadable
        moduleBuffer.append( ModuleTags.TABx2 ).append( ModuleTags.RELOADABLE ).append( "=\"" ); // NOI18N
        moduleBuffer.append( new Boolean (reloadable) ).append( "\"" ); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        // Module footer
        moduleBuffer.append( "/>"); // NOI18N
        moduleBuffer.append( ModuleTags.NEW_LINE );

        return moduleBuffer.toString();
    }


    /** Restores the item from XML. Creates new instance of ModuleItem
     * represented by the attribute list. 
     * @return new ModuleItem
     * @throws IOException if there was some problem loading it
     * @throws FileNotFoundException if the module JAR no longer exists
     */
    static ModuleItem fromXML( AttributeList attr/*, URL base */) throws IOException {

        // Read main flags

        String url = attr.getValue( ModuleTags.URL );
        String enabled = attr.getValue( ModuleTags.ENABLED );
        // String deleted = attr.getValue( ModuleTags.DELETED );
        String reloadable = attr.getValue (ModuleTags.RELOADABLE);
        String baseAttr = attr.getValue( ModuleTags.BASE );

        // Read old version of module

        int relase;
        try {
            relase = Integer.parseInt( attr.getValue( ModuleTags.RELEASE ) );
        }
        catch ( NumberFormatException e ) {
            relase = -1;
        }

        String specVersion = attr.getValue( ModuleTags.SPECVERSION );
        
        int baseCode;
                if (  ModuleTags.BASE_USER.equals( baseAttr ) ) {
                    baseCode = BASE_USER;
                }
                else if ( ModuleTags.BASE_CENTRAL.equals( baseAttr ) ) {
                    baseCode = BASE_CENTRAL;
                }
                else if ( ModuleTags.BASE_JAR.equals( baseAttr ) ) {
                    baseCode = BASE_JAR;
                } else {
                    throw new IOException ("invalid module base"); // NOI18N
                }

                ModuleItem mi = new ModuleItem( baseCode, url,
                                                enabled == null || Boolean.valueOf (enabled).booleanValue (),
                                                reloadable != null && Boolean.valueOf (reloadable).booleanValue ());

                mi.setOldRelease( relase );
                mi.setOldSpecVersion( specVersion );
                return mi;
    }


    /** Installs everything.
    */
    private static class InstallIterator extends Object implements ManifestSection.Iterator {

        /** Process action section.
         */
        public void processAction (ManifestSection.ActionSection as) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            ModuleActions.add (as);
            pm.notifyTime("processAction ".concat(as.getAction().getClass().toString()), System.currentTimeMillis() - time);
        }

        /** Processes the option section.
         */
        public void processOption (ManifestSection.OptionSection os) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            NbControlPanel pool = NbControlPanel.getDefault ();
            pool.add (os.getOption ());
            pm.notifyTime("processOption ".concat(os.getOption().getClass().toString()), System.currentTimeMillis() - time);
        }

        /** Processes the loader section.
        */
        public void processLoader (ManifestSection.LoaderSection ls) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            LoaderPoolNode.add (ls);
            pm.notifyTime("processLoader ".concat(ls.getLoader().getClass().toString()), System.currentTimeMillis() - time);
        }


        /** Processes debugger section
        */
        public void processDebugger (ManifestSection.DebuggerSection ds) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            NbTopManager.get ().setDebugger (ds.getDebugger ());
            pm.notifyTime("processDebugger ".concat(ds.getDebugger().toString()), System.currentTimeMillis() - time);
        }

        /** Processes executor section
        */
        public void processService (ManifestSection.ServiceSection es) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            Services.addService (es);
            NbTopManager.get ().register (es);
            pm.notifyTime("processService ".concat(es.getServiceType().getClass().toString()), System.currentTimeMillis() - time);
        }

        /** Processes filesystem section
        */
        public void processFileSystem (ManifestSection.FileSystemSection fs) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            NbTopManager.get ().register (fs);
            pm.notifyTime("processFileSystem ".concat(fs.getName()), System.currentTimeMillis() - time);
        }

        /** Processes environment node section.
        * @exception Exception if there is an error and the section should be removed
        */
        public void processNode (ManifestSection.NodeSection es) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            NbTopManager.get ().register (es);
            pm.notifyTime("processNode ", System.currentTimeMillis() - time);
        }

        /** Processes clipboard convertor section.
        * @exception Exception if there is an error and the section should be removed
        */
        public void processClipboardConvertor (ManifestSection.ClipboardConvertorSection es) throws InstantiationException {
            org.netbeans.core.perftool.PerformanceMeter pm = org.netbeans.core.perftool.PerformanceMeterFactory.getPerformanceMeter();
            long time = System.currentTimeMillis();
            CoronaClipboard.addConvertor (es.getConvertor ());
            pm.notifyTime("processClipboardConvertor " + es.getConvertor().getClass(), System.currentTimeMillis() - time);
        }

    }

    /** Installs new modules. */
    private static final ManifestSection.Iterator INSTALL = new InstallIterator ();

    /** Restores existing modules.
    * Not private because ModuleInstaller uses it for the IDE manifest.
    */
    static final ManifestSection.Iterator RESTORE = INSTALL;

    /** Deinstalator of sections.
    */
    private static final ManifestSection.Iterator REMOVE = new ManifestSection.Iterator () {
                /** Process action section.
                */
                public void processAction (ManifestSection.ActionSection as) throws InstantiationException {
                    ModuleActions.remove (as);
                }

                /** Processes the option section.
                */
                public void processOption (ManifestSection.OptionSection os) throws InstantiationException {
                    NbControlPanel pool = NbControlPanel.getDefault ();
                    pool.remove (os.getOption ());
                }

                /** Processes the loader section.
                */
                public void processLoader (ManifestSection.LoaderSection ls) throws InstantiationException {
                    LoaderPoolNode.remove (ls.getLoader ());
                }


                /** Processes debugger section
                */
                public void processDebugger (ManifestSection.DebuggerSection ds) throws InstantiationException {
                    NbTopManager.get ().setDebugger (null);
                }

                /** Processes executor section
                */
                public void processService (ManifestSection.ServiceSection es) throws InstantiationException {
                    Services.removeService (es);
                    NbTopManager.get ().unregister(es);
                }

                /** Processes filesystem section
                */
                public void processFileSystem (ManifestSection.FileSystemSection fs) throws InstantiationException {
                    NbTopManager.get ().unregister (fs);
                }

                /** Processes environment node section.
                * @exception Exception if there is an error and the section should be removed
                */
                public void processNode (ManifestSection.NodeSection es) throws InstantiationException {
                    NbTopManager.get ().unregister (es);
                    /*
                    if (ManifestSection.NodeSection.TYPE_ROOTS.equalsIgnoreCase (es.getType ())) {
                        NbPlaces.removeRoot (es.getNode ());
                    } else if (ManifestSection.NodeSection.TYPE_SESSION.equalsIgnoreCase (es.getType ())) {
                        NbPlaces.removeSession (es.getNode ());
                    } else {
                        EnvironmentNode.removeNode (es.getNode ());
                    }
                     */
                }

                /** Processes clipboard convertor section.
                * @exception Exception if there is an error and the section should be removed
                */
                public void processClipboardConvertor (ManifestSection.ClipboardConvertorSection es) throws InstantiationException {
                    CoronaClipboard.removeConvertor (es.getConvertor ());
                }

            };
}
