/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.beans.*;
import java.io.*;
import java.net.URL;
import java.util.*;
import javax.swing.event.*;
import javax.swing.JFileChooser;
import javax.swing.AbstractAction;

import org.openide.*;
import org.openide.actions.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.modules.*;
import org.openide.nodes.*;
import org.openide.options.*;
import org.openide.util.*;
import org.openide.util.actions.*;
import org.openide.util.datatransfer.*;
import org.openide.util.Lookup;

/** Node representing modules */
public class ModuleNode extends AbstractNode {

    private static final ResourceBundle bundle = NbBundle.getBundle( ModuleNode.class );

    /** Icon bases */
    private static final String MODULE_ITEM_ICON_BASE="/org/netbeans/core/resources/moduleItem"; // NOI18N
    private static final String MODULE_ITEM_DISABLED_BASE="/org/netbeans/core/resources/moduleItemDisabled"; // NOI18N
    private static final String MODULE_TEST_ITEM_ICON_BASE="/org/netbeans/core/resources/testModuleItem"; // NOI18N
    private static final String MODULES_ICON_BASE="/org/netbeans/core/resources/modules"; // NOI18N

    /** Last directory used by the install new module file chooser. */
    private static File lastChosenDir = null;


    /** New types */
    private static NewType[] newTypes = null;

    /** map from (Item, Boolean) of modules we should enable/disable
    */
    private static Map changeState = new HashMap (11);
    
    /** task to run when changeState should be processed */
    private static RequestProcessor.Task changeStateRequest = RequestProcessor.createRequest (
        new Update ()
    );
    
    /** whether modules should be enabled/disabled immediately or after signal performLateUpdate() */
    private static boolean lateUpdate = false;
    
    /** modules which should be disabled later */
    private static LinkedList toDisableLater = new LinkedList();
    
    /** modules which should be disabled later */
    private static LinkedList toEnableLater = new LinkedList();
    
    /** listener on changes in modules state */
    private static ModuleSelectionPanel msp;

    /** Creates a new ModulesNode */
    public ModuleNode() {
        super(new Modules ());
        setName(Main.getString("CTL_Modules_name"));
        setShortDescription (Main.getString ("CTL_Modules_hint"));
        setIconBase (MODULES_ICON_BASE);
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (ModuleNode.class);
    }

    public Node cloneNode () {
        return new ModuleNode ();
    }

    protected SystemAction[] createActions () {

        return new SystemAction[] {
                   SystemAction.get (NewAction.class),
                   null,
                   SystemAction.get (ToolsAction.class),
                   SystemAction.get (PropertiesAction.class),
               };

    }

    /** Add a sorting property. */
    protected Sheet createSheet () {
        Sheet.Set set = new Sheet.Set ();
        set.setName ("sorting"); // NOI18N
        set.setDisplayName (bundle.getString ("LBL_ModuleNode_sheet_sorting"));
        set.put (((Modules) getChildren ()).createSortingProperty ());
        Sheet sheet = new Sheet ();
        sheet.put (set);
        return sheet;
    }

    public NewType[] getNewTypes () {

        if ( newTypes == null ) {
            int size = Boolean.getBoolean ("netbeans.module.test") ? 2 : 1;
            
            Lookup lkp = Lookup.getDefault();
            Lookup.Result result = lkp.lookup( new Lookup.Template( ModuleNewType.class ) );            
            
            newTypes = new NewType[size + result.allInstances().size()];
            newTypes[0] = new ModuleNewType () {
                          public String getName () {
                              return Main.getString ("CTL_NewModuleByFile");
                          }

                          public void create () throws IOException {
                              addFile (false);
                          }
                      };            
            if ( Boolean.getBoolean ("netbeans.module.test")) { // NOI18N
                 newTypes[1] = new ModuleNewType () {
                              public String getName () {
                                  return Main.getString( "CTL_NewTestModule");
                              }

                              public void create () throws IOException {
                                  addFile (true);
                              }
                          };            
            }
                                      
            Iterator instances = result.allInstances().iterator();
            while (instances.hasNext()) {
                newTypes[size] = (NewType) instances.next();
                size++;
            }             

        }

        return newTypes;


    }

    /** Allows to add new module by specifying its file.
    */
    void addFile (final boolean reloadable) throws IOException {
        final JFileChooser chooser = new JFileChooser ();
        if (lastChosenDir != null) chooser.setCurrentDirectory (lastChosenDir);
        chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        chooser.setApproveButtonText (Main.getString ("CTL_ModuleSelect"));
        chooser.setApproveButtonToolTipText (Main.getString ("CTL_ModuleSelectToolTip"));
        chooser.setFileFilter (new javax.swing.filechooser.FileFilter () {
                                   public String getDescription () {
                                       return Main.getString ("CTL_ModuleSelectFilter");
                                   }
                                   public boolean accept (File f) {
                                       return f.isDirectory () || f.getName ().endsWith (".jar"); // NOI18N
                                   }
                               });

        int result = chooser.showOpenDialog (TopManager.getDefault ().getWindowManager ().getMainWindow ());
        lastChosenDir = chooser.getCurrentDirectory ();
        if (result == JFileChooser.APPROVE_OPTION) {
            // install the file
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    try {
                        ModuleInstaller.installFromFile ( chooser.getSelectedFile(), reloadable );
                    } catch (IOException ex) {
                        TopManager.getDefault ().getErrorManager().notify (ex);
                    }
                }
            });
        }

    }

    /** Class representing node of one standard module
     */
    static class Item extends AbstractNode  {

        protected ModuleItem item;

        public Item (ModuleItem item) {
            super (Children.LEAF);
            this.item = item;

            ModuleDescription moduleDesc = item.getDescription ();
            // [PENDING] somehow reset these after a reload (e.g. override getName() etc.?)
            setName (moduleDesc.getCodeName ());
            setDisplayName (moduleDesc.getName ());
            String tooltip = moduleDesc.getShortDescription ();
            if (tooltip != null) {
                setShortDescription (tooltip);
            }
            setIconBase ( item.isReloadable () ? MODULE_TEST_ITEM_ICON_BASE : ( item.isEnabled() ? MODULE_ITEM_ICON_BASE : MODULE_ITEM_DISABLED_BASE ) );
            setDefaultAction (SystemAction.get (PropertiesAction.class));
        }

        public HelpCtx getHelpCtx () {
            String id = (String) Help.getDefault ().getHomesByCode ().get (item.getDescription ().getCodeName ());
            if (id != null && item.isEnabled ())
                return new HelpCtx (id);
            else
                return new HelpCtx (Item.class);
        }

        protected SystemAction[] createActions () {
            return new SystemAction[] {
                SystemAction.get (DeleteAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class),
            };
        }

        public void destroy () {

            List depModules = ModuleInstaller.getDependentModules( item );
            Collections.reverse (depModules);

            if ( depModules.size() > 0 ) {
                StringBuffer sb = new StringBuffer( 200 );

                sb.append( bundle.getString( "MSG_UninstallOthers" ) ).append( "\n" );

                Iterator it = depModules.iterator();
                while( it.hasNext() ) {
                    sb.append( ((ModuleItem)it.next()).getDescription().getName() );
                    sb.append( "\n" ); // NOI18N
                }

                NotifyDescriptor nd = new NotifyDescriptor.Confirmation( sb.toString(), NotifyDescriptor.OK_CANCEL_OPTION );
                TopManager.getDefault().notify( nd );
                if ( nd.getValue() == DialogDescriptor.CANCEL_OPTION )
                    return;
                else {
                    it = depModules.iterator();
                    while( it.hasNext() ) {
                        ModuleInstaller.deleteModule( (ModuleItem)it.next() );
                    }
                }
            }

            // [PENDING] should this rather be moduleItem(this, Boolean.FALSE)?
            // If so, how to handle the uninstallation of dependencies, above?
            // Probably the code block above should be removed, first. Then Update
            // should actually have three actions--enable, disable, and delete.
            // Delete would be handled much like disable, except of course the final
            // step is to call ModuleInstall.deleteModule(item) rather than item.setEnabled(false).
            // Everything else could be about the same.
            ModuleInstaller.deleteModule( item );
        }

        public boolean canDestroy () {
            return item.canDestroy();
        }

        /** Creates properties.
        */
        protected Sheet createSheet () {
            Sheet s = Sheet.createDefault ();
            Sheet.Set ss = s.get (Sheet.PROPERTIES);

            Sheet.Set sse = Sheet.createExpertSet ();
            s.put(sse);

            PropertySupport p;

            p = new PropertySupport.ReadOnly (
                    "name", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_name"),
                    Main.getString ("HINT_modules_name")
                ) {
                    public Object getValue () {
                        return item.getDescription ().getName ();
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadOnly (
                    "shortDescription", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_shortDescription"),
                    Main.getString ("HINT_modules_shortDescription")
                ) {
                    public Object getValue () {
                        String v = item.getDescription ().getShortDescription ();
                        if (v == null) v = ""; // NOI18N
                        return v;
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadOnly (
                    "longDescription", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_longDescription"),
                    Main.getString ("HINT_modules_longDescription")
                ) {
                    public Object getValue () {
                        String v = item.getDescription ().getLongDescription ();
                        if (v == null) v = ""; // NOI18N
                        return v;
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadOnly (
                    "category", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_category"),
                    Main.getString ("HINT_modules_category")
                ) {
                    public Object getValue () {
                        String v = item.getDescription ().getDisplayCategory ();
                        if (v == null) v = ""; // NOI18N
                        return v;
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadWrite (
                    "enabled", // NOI18N
                    Boolean.TYPE,
                    Main.getString ("PROP_modules_enabled"),
                    Main.getString ("HINT_modules_enabled")
                ) {
                    public Object getValue () {
                        return new Boolean (item.isEnabled ());
                    }

                    public void setValue (Object o) {
                        moduleItem (Item.this, (Boolean)o);
                    }
                    
                    public boolean canWrite () {
                        return item.canDisable ();
                    }
                };
            ss.put (p);

            if (Boolean.getBoolean ("netbeans.module.test")) {
                p = new PropertySupport.ReadWrite (
                        "reloadable", // NOI18N
                        Boolean.TYPE,
                        Main.getString ("PROP_modules_reloadable"),
                        Main.getString ("HINT_modules_reloadable")
                    ) {
                        public Object getValue () {
                            return new Boolean (item.isReloadable ());
                        }
                        public void setValue (Object o) {
                            item.setReloadable (((Boolean) o).booleanValue ());
                        }
                        public boolean canWrite () {
                            // Need to be able to write to the containing directory to use this feature.
                            if (item.getJar () == null) return false;
                            return item.isReloadable () || item.getJar ().getParentFile ().canWrite ();
                        }
                    };
                ss.put (p);
            }

            p = new PropertySupport.ReadOnly (
                    "specVersion", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_specversion"),
                    Main.getString ("HINT_modules_specversion")
                ) {
                    public Object getValue () {
                        return item.getDescription ().getSpecVersion ();
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadOnly (
                    "implVersion", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_implversion"),
                    Main.getString ("HINT_modules_implversion")
                ) {
                    public Object getValue () {
                        return item.getDescription ().getImplVersion ();
                    }

                };
            ss.put (p);

            p = new PropertySupport.ReadOnly (
                    "codeName", // NOI18N
                    String.class,
                    Main.getString ("PROP_modules_codename"),
                    Main.getString ("HINT_modules_codename")
                ) {
                    public Object getValue () {
                        return item.getDescription ().getCodeName ();
                    }

                };
            sse.put (p);

            p = new PropertySupport.ReadOnly (
                    "jar", // NOI18N
                    java.io.File.class,
                    Main.getString ("PROP_modules_jar"),
                    Main.getString ("HINT_modules_jar")
                ) {
                    public Object getValue () {
                        return item.getJar ();
                    }

                };
            if (item.getJar () != null) sse.put (p);

            return s;
        }

        public void refresh () {
            firePropertyChange (Node.PROP_PROPERTY_SETS, null, null);
        }        

    }
    
    
    /** Enables/disables module item.
    * @param item the item to change
    * @param state the state to change
    */
    private static void moduleItem (Item item, Boolean state) {
        synchronized (changeState) {
            changeState.put (item, state);
            changeStateRequest.schedule (200);
        }
        if (msp != null) {
            msp.modulesUpdateScheduled();
        }
    }
    
    /** Performes actuall update of modules with ModuleItem.setEnabled()
     */
    static void performLateUpdate() {
        lateUpdate = false;
        changeStateRequest.schedule (200);
    }
    
    /** All calls of moduleItem() will update modules with ModuleItem.setEnabledNoNotify()
     */
    static void setUpdateLater() {
        lateUpdate = true;
    }

    /** Children that contains modules installed it has to
     * dissingushg between modules and test modules.
     * Keys are of type ModuleItem.
     */
    private static class Modules extends Children.Keys implements PropertyChangeListener {

        IDESettings settings;

        /** Refreshed list of nodes acc. to current sorting and contents. */
        private void refreshKeys () {
            ModuleItem[] items = ModuleInstaller.getModuleItems (
                                     ModuleInstallerSupport.ENABLED_MODULE |
                                     ModuleInstallerSupport.DISABLED_MODULE );
            if (settings.getModulesSortMode () == IDESettings.MODULES_SORT_CATEGORY) {
                // Special keys: category names. Also ModuleItem otherwise.
                SortedSet categories = new TreeSet ();
                List uncategorized = new ArrayList ();
                for (int i = 0; i < items.length; i++) {
                    ModuleItem mi = items[i];
                    String category = mi.getDescription ().getDisplayCategory ();
                    if (category != null) {
                        categories.add (category);
                    } else {
                        uncategorized.add (mi);
                    }
                }
                Collections.sort (uncategorized, new Comparator () {
                        public int compare (Object o1, Object o2) {
                            ModuleItem m1 = (ModuleItem) o1;
                            ModuleItem m2 = (ModuleItem) o2;
                            return m1.getDescription ().getName ().compareTo (m2.getDescription ().getName ());
                        }
                    });
                uncategorized.addAll (categories);
                setKeys (uncategorized);
                return;
            } else if (settings.getModulesSortMode () != IDESettings.MODULES_SORT_UNSORTED) {
                items = (ModuleItem[]) items.clone ();
                Arrays.sort (items, new Comparator () {
                        public int compare (Object o1, Object o2) {
                            ModuleItem m1 = (ModuleItem) o1;
                            ModuleItem m2 = (ModuleItem) o2;
                            switch (settings.getModulesSortMode ()) {
                            case IDESettings.MODULES_SORT_CODENAME:
                                return m1.getDescription ().getCodeName ().compareTo (m2.getDescription ().getCodeName ());
                            case IDESettings.MODULES_SORT_ENABLED:
                                if (m1.isEnabled () && ! m2.isEnabled ()) return -1;
                                if (! m1.isEnabled () && m2.isEnabled ()) return 1;
                                // fallthrough
                            case IDESettings.MODULES_SORT_DISPLAYNAME:
                                return m1.getDescription ().getName ().compareTo (m2.getDescription ().getName ());
                            case IDESettings.MODULES_SORT_URL:
                                File j1 = m1.getJar ();
                                File j2 = m2.getJar ();
                                if (j1 != null) {
                                    if (j2 != null) {
                                        return j1.compareTo (j2);
                                    } else {
                                        return 1;
                                    }
                                } else {
                                    if (j2 != null) {
                                        return -1;
                                    } else {
                                        return 0;
                                    }
                                }
                            default:
                                return 0;
                            }
                        }
                    });
            }
            setKeys (items);
        }

        Node.Property createSortingProperty () {
            settings = (IDESettings) SharedClassObject.findObject (IDESettings.class, true);
            return new PropertySupport.ReadWrite ("sorted", Integer.TYPE, // NOI18N
                                                  bundle.getString ("PROP_ModuleNode_sorted"),
                                                  bundle.getString ("HINT_ModuleNode_sorted")) {
                       public Object getValue () {
                           return new Integer (settings.getModulesSortMode ());
                       }
                       public void setValue (Object o) {
                           settings.setModulesSortMode (((Integer) o).intValue ());
                       }
                       public boolean supportsDefaultValue () {
                           return true;
                       }
                       public void restoreDefaultValue () {
                           setValue (new Integer (IDESettings.MODULES_SORT_CATEGORY));
                       }
                       public PropertyEditor getPropertyEditor () {
                           return new PropertyEditorSupport () {
                                      private final String[] tags = new String[] {
                                                                        bundle.getString ("LBL_ModuleNode_SORT_UNSORTED"),
                                                                        bundle.getString ("LBL_ModuleNode_SORT_DISPLAYNAME"),
                                                                        bundle.getString ("LBL_ModuleNode_SORT_CODENAME"),
                                                                        bundle.getString ("LBL_ModuleNode_SORT_ENABLED"),
                                                                        bundle.getString ("LBL_ModuleNode_SORT_URL"),
                                                                        bundle.getString ("LBL_ModuleNode_SORT_CATEGORY"),
                                                                    };
                                      public String[] getTags () {
                                          return tags;
                                      }
                                      public String getAsText () {
                                          return tags[((Integer) this.getValue ()).intValue ()];
                                      }
                                      public void setAsText (String text) {
                                          for (int i = 0; i < tags.length; i++) {
                                              if (tags[i].equals (text)) {
                                                  this.setValue (new Integer (i));
                                                  return;
                                              }
                                          }
                                          throw new IllegalArgumentException ();
                                      }
                                  };
                       }
                   };
        }

        /** Initializes content */
        public void addNotify () {
            settings = (IDESettings) SharedClassObject.findObject (IDESettings.class, true);
            ModuleInstaller.addPropertyChangeListener (this);
            settings.addPropertyChangeListener (this);
            refreshKeys ();
        }

        /** Releases listener. */
        public void removeNotify () {
            settings.removePropertyChangeListener (this);
            ModuleInstaller.removePropertyChangeListener (this);
        }

        /** Reacts to changes */
        public void propertyChange (PropertyChangeEvent ev) {
            refreshKeys ();
        }

        /** Generates node for the ModuleItem key */
        protected Node[] createNodes (Object key) {
            if (key instanceof String) {
                return new Node[] { new CategoryNode ((String) key) };
            } else {
                return new Node[] { new Item ((ModuleItem)key) };
            }
        }
    }

    private static final class CategoryNode extends AbstractNode 
                                            implements PropertyChangeListener {

        public CategoryNode (String category) {
            super (new CategoryChildren (category));
            setName (category); // displayName should be the same
            setIconBase (MODULES_ICON_BASE);
        }

        public boolean canDestroy () {
            return false;
        }

        public boolean canRename () {
            return false;
        }

        public boolean canCut () {
            return false;
        }

        /** Creates property.
        */
        protected Sheet createSheet () {
            Sheet s = Sheet.createDefault ();
            Sheet.Set ss = s.get (Sheet.PROPERTIES);

            PropertySupport p;
            // value of this property is 'sum' of children node 'enabled' properties
            p = new PropertySupport.ReadWrite (
                    "enabled", // NOI18N
                    Boolean.TYPE,
                    Main.getString ("PROP_modules_enabled"),
                    Main.getString ("HINT_modules_enabled")
                ) {
                    public Object getValue () {
                        Enumeration subnodes = getChildren().nodes();
                        if (subnodes.hasMoreElements()) {
                            boolean enabled = ((Item)subnodes.nextElement()).item.isEnabled();
                            while (subnodes.hasMoreElements()) {
                                if (((Item)subnodes.nextElement()).item.isEnabled() != enabled)
                                    return null;
                            }
                            return new Boolean (enabled);
                        }
                        return null;
                    }

                    public void setValue (Object o) {
                        if (o != null) {
                            Enumeration subnodes = getChildren().nodes();
                            while (subnodes.hasMoreElements()) {
                                Item i;
                                if ((i = (Item)subnodes.nextElement()).item.canDisable())
                                    moduleItem(i, (Boolean)o);
                            }
                        }
                    }
                    
                    public boolean canWrite () {
                        Enumeration subnodes = getChildren().nodes();
                        while (subnodes.hasMoreElements()) {
                            if (((Item)subnodes.nextElement()).item.canDisable())
                                return true;
                        }
                        return false;
                    }
                    
                    public PropertyEditor getPropertyEditor() {
                        return new BooleanPropertyEditor();
                    }
                };
            ss.put (p);
            return s;
        }

        /** Propagate children property change */
        public void propertyChange (PropertyChangeEvent ev) {
            if (Node.PROP_PROPERTY_SETS.equals(ev.getPropertyName()))
                firePropertyChange(Node.PROP_PROPERTY_SETS, null, null);
        }

        private static final class CategoryChildren extends Children.Keys implements PropertyChangeListener {

            private final String category;

            public CategoryChildren (String category) {
                this.category = category;
            }

            protected void addNotify () {
                refreshKeys ();
                ModuleInstaller.addPropertyChangeListener (this);
            }

            protected void removeNotify () {
                ModuleInstaller.removePropertyChangeListener (this);
                setKeys (Collections.EMPTY_SET);
            }

            private void refreshKeys () {
                ModuleItem[] items = ModuleInstaller.getModuleItems (ModuleInstallerSupport.ENABLED_MODULE | ModuleInstallerSupport.DISABLED_MODULE);
                List keys = new ArrayList ();
                for (int i = 0; i < items.length; i++) {
                    if (category.equals (items[i].getDescription ().getDisplayCategory ())) {
                        keys.add (items[i]);
                    }
                }
                Collections.sort (keys, new Comparator () {
                        public int compare (Object o1, Object o2) {
                            ModuleItem m1 = (ModuleItem) o1;
                            ModuleItem m2 = (ModuleItem) o2;
                            return m1.getDescription ().getName ().compareTo (m2.getDescription ().getName ());
                        }
                    });
                setKeys (keys);
            }

            protected Node[] createNodes (Object key) {
                Item item = new Item ((ModuleItem) key);
                item.addPropertyChangeListener((CategoryNode)getNode());
                return new Node[] { item };
            }

            protected void destroyNodes (Node[] arr) {
                for (int i = 0; i < arr.length; i++) {
                    arr[i].removePropertyChangeListener((CategoryNode)getNode());
                }
                super.destroyNodes(arr);
            }
            
            public void propertyChange (PropertyChangeEvent ev) {
                refreshKeys ();
            }

        }

    }

    /** Action that refreshes the modules in the modules directory.
    * /
    public static final class AutoLoadModulesAction extends CallableSystemAction {

        static final long serialVersionUID =-6598195161786990764L;

        public String getName () {
            return Main.getString ("CTL_AutoLoadModules");
        }

        public HelpCtx getHelpCtx () {
            return new HelpCtx (AutoLoadModulesAction.class);
        }

        public void performAction () {
            ModuleInstaller.autoLoadModules ();
        }
    }*/

    /** Runnable that prepares list of modules to update and 
    * if lateUpdate is false then also updates modules.
    */
    private static final class Update implements Runnable {
        /** Invoked when we should check state of modules.
        */
        public void run () {
            List toEnable = new LinkedList ();
            List toDisable = new LinkedList ();
            Set nodesToRefresh = new HashSet ();

            // fill the toDisable and toEnable sets
            synchronized (changeState) {

                Iterator it = changeState.entrySet ().iterator ();
                while (it.hasNext ()) {
                    Map.Entry en = (Map.Entry)it.next ();
                    
                    Item node = (Item)en.getKey ();
                    
                    nodesToRefresh.add (node);
                    
                    ModuleItem item = node.item;

                    if (Boolean.TRUE.equals (en.getValue ())) {
                        // module to enable
                        if (!toEnable.contains (en.getKey ())) {
                            toEnable.add (item);
                        }
                    } else {
                        if (!toDisable.contains (en.getKey ())) {
                            toDisable.add (item);
                        }
                    }
                }

                changeState.clear();
            }

            synchronized(toEnableLater) {
                // process modules which should be disabled
                {
                    Iterator it = new LinkedList (toDisable).iterator();
                    while (it.hasNext ()) {
                        ModuleItem item = (ModuleItem)it.next ();

                        // Ask about other modules to disable
                        List depModules = ModuleInstaller.getDependentModules (item);
                        Collections.reverse (depModules);
                        // [PENDING] probably if multiple modules are disabled at once,
                        // the order could still be incorrect--TBD. See #11145.

                        // remove all other modules that will be disabled anyway
                        depModules.removeAll(toDisable);

                        if ( depModules.size() > 0 ) {

                            StringBuffer sb = new StringBuffer( 200 );

                            sb.append( bundle.getString( "MSG_DisableOthers" ) ).append( "\n" );

                            Iterator it2 = depModules.iterator();
                            while( it2.hasNext() ) {
                                sb.append( ((ModuleItem)it2.next()).getDescription().getName() );
                                sb.append( "\n" ); // NOI18N
                            }

                            NotifyDescriptor nd = new NotifyDescriptor.Confirmation( sb.toString(), NotifyDescriptor.OK_CANCEL_OPTION );
                            TopManager.getDefault().notify( nd );
                            if ( nd.getValue() == DialogDescriptor.CANCEL_OPTION ) {
                                // removes the item from collection, because it should not be removed
                                toDisable.remove (item);
                            } else {
                                // add those modules, to the colleciton
                                toDisable.addAll (0, depModules);
                            }
                        }
                    }
                }

                // process modules which should be enabled
                {
                    if (!toEnable.isEmpty ()) {
                        Collection c = ModuleInstaller.resolveOrdering (
                            toEnable, 
                            ModuleInstallerSupport.ENABLED_MODULE | ModuleInstallerSupport.DISABLED_MODULE,
                            true
                        );

                        if (c.isEmpty ()) {
                            NotifyDescriptor.Message nd = new NotifyDescriptor.Message(
                              bundle.getString( "MSG_Module_Missed_Enabled" ) + "\n" +
                              ModuleInstaller.getMissed(
                                  ModuleInstallerSupport.ENABLED_MODULE | ModuleInstallerSupport.DISABLED_MODULE,
                                  toEnable 
                              ) 
                            );
                            TopManager.getDefault().notify( nd );
                            toEnable.clear ();
                        } else {
                            // remember order of objects in which they should be enabled
                            toEnable = new LinkedList (c);
                        }
                    }
                }

                // silently disable modules
                {
                    Iterator it = new LinkedList(toDisable).iterator ();
                    while (it.hasNext ()) {
                        ModuleItem item = (ModuleItem)it.next ();
                        if (item.isEnabled())
                            item.setEnabledNoNotify (false);
                        else
                            toDisable.remove(item);
                    }
                }

                // list which contains modules to enable later in right order
                LinkedList orderedToEnable = new LinkedList();

                // silently enable modules
                {
                    Iterator it = toEnable.iterator ();
                    while (it.hasNext ()) {
                        ModuleItem item = (ModuleItem)it.next ();
                        // Test if some other modules should be enabled

                        Collection orderedModules = ModuleInstaller.checkDependenciesOnDisabled ( 
                            item.getDescription() 
                        );

                        if ( orderedModules != null ) {
                            Iterator inner = orderedModules.iterator();
                            while (inner.hasNext()) {
                                ModuleItem mi = (ModuleItem)inner.next ();
                                if (!item.isEnabled()) {
                                    mi.setEnabledNoNotify (true);
                                    orderedToEnable.add(mi);
                                }
                            }
                        } else {
                            break;
                        }

                        if (!item.isEnabled()) {
                            item.setEnabledNoNotify (true);
                            orderedToEnable.add(item);
                        }
                    }
                }

                // toDisableLater = (toDisableLater - orderToEnable) + (toDisable - toEnableLater)
                // toEnableLater = (toEnableLater - toDisable) + (orderedToEnable - toDisableLater)
                LinkedList actuallyToEnable = new LinkedList(orderedToEnable);
                actuallyToEnable.removeAll(toDisableLater);
                toDisableLater.removeAll(orderedToEnable);

                LinkedList actuallyToDisable = new LinkedList(toDisable);
                actuallyToDisable.removeAll(toEnableLater);
                toEnableLater.removeAll(toDisable);

                toDisableLater.addAll(actuallyToDisable);
                toEnableLater.addAll(actuallyToEnable);
                
                // This should partialy fixes order of disabled modules. Modules are
                // disabled in same order in which they are held by ModuleInstaller.
                // This may not work if MI modules order is not correct.
                List reverseList = new ArrayList(Arrays.asList(ModuleInstaller.getModuleItems (ModuleInstallerSupport.ENABLED_MODULE | ModuleInstallerSupport.DISABLED_MODULE)));
                List reverseListWithoutToDisable = new ArrayList(reverseList);
                reverseListWithoutToDisable.removeAll(toDisableLater);
                Collections.reverse(reverseList);
                reverseList.removeAll(reverseListWithoutToDisable);
                
                toDisableLater.clear();
                toDisableLater.addAll(reverseList);

                if (!lateUpdate) {
                    // firstly get modules to right state
                    undoUpdate();
                    // actually make changes
                    synchronized (toEnableLater) {
                        // disable modules
                        Iterator it = toDisableLater.iterator ();
                        while (it.hasNext ()) {
                            ModuleItem mi = (ModuleItem)it.next ();
                            mi.setEnabled (false);
                            addNodeToRefresh(nodesToRefresh, mi);
                        }

                        // enable modules
                        it = toEnableLater.iterator();
                        while (it.hasNext()) {
                            ModuleItem mi = (ModuleItem)it.next ();
                            mi.setEnabled (true);
                            addNodeToRefresh(nodesToRefresh, mi);
                        }
                        toEnableLater.clear();
                        toDisableLater.clear();
                    }
                } else {
                    // notify module selection panel
                    if (msp != null) {
                        msp.modulesUpdateFinished(new LinkedList(toDisableLater));
                    }
                }
            }

            {
                // refresh all nodes
                Iterator it = nodesToRefresh.iterator ();
                while (it.hasNext ()) {
                    Item i = (Item)it.next ();
                    i.setIconBase(i.item.isEnabled() ? MODULE_ITEM_ICON_BASE : MODULE_ITEM_DISABLED_BASE);
                    i.refresh ();
                }
            }
        }

        private void addNodeToRefresh (Set nodesToRefresh, ModuleItem moduleItem) {
            int index = 0;
            int count;
            Item item = null;
            ModuleNode moduleNode = findModuleNode((Node)(nodesToRefresh != null && nodesToRefresh.size() > 0 ? nodesToRefresh.toArray()[0] : null));
            Children children = moduleNode != null ? moduleNode.getChildren() : null;
            Node[] nodes = children != null ? children.getNodes() : null;
            count = nodes != null ? nodes.length : 0;
            while(index < count && (item = getModuleNodeByModuleItem(nodes[index], moduleItem)) == null)
                index++;
            if(item != null)
                nodesToRefresh.add(item);
        }

        private ModuleNode findModuleNode (Node node) {
            if(node == null)
                return null;
            if(node instanceof ModuleNode)
                return (ModuleNode)node;
            return findModuleNode(node.getParentNode());
        }

        private Item getModuleNodeByModuleItem (Node rootNode, ModuleItem moduleItem) {
            int count;
            Item item;
            Node[] nodes;
            if(rootNode instanceof Item && (item = (Item)rootNode).item == moduleItem)
                return item;
            nodes = rootNode.getChildren().getNodes();
            count = nodes.length;
            for(int i = 0; i < count; i++)
                if((item = getModuleNodeByModuleItem(nodes[i], moduleItem)) != null)
                    return item;
            return null;
        }
    }

    /** ModuleSelectionPanel have to listen on modules update, so that it can update
     * advanced part of the wizard
     */
    static void setModuleUpdateListener(ModuleSelectionPanel m) {
        msp = m;
    }
    
    /** ModuleSelectionPanel asks whether change of modules state is scheduled 
     */
    static boolean isUpdateScheduled() {
        synchronized (changeState) {
            return !changeState.isEmpty();
        }
    }
    
    /** Reverse back temporary change of modules state and clear lists so module state 
     * changes will not done later.
     */
    static void clearUpdateLater() {
        synchronized (toEnableLater) {
            undoUpdate();
            toEnableLater.clear();
            toDisableLater.clear();
            lateUpdate = false;
        }
    }
    
    /** This allows to preset list of ModuleItems, which should be enabled/disabled without
     * knowing appropriate module nodes.
     */
    static void updateLater(List enable, List disable) {
        synchronized (toEnableLater) {
            undoUpdate();
            
            // enable silently
            List yetEnabled = Arrays.asList(
                                        ModuleInstaller.getModuleItems(
                                            ModuleInstallerSupport.ENABLED_MODULE));
            enable.removeAll(yetEnabled);

            Collection c = ModuleInstaller.resolveOrdering (
                enable, 
                ModuleInstallerSupport.DISABLED_MODULE | ModuleInstallerSupport.ENABLED_MODULE,
                true
            );

            Iterator it = c.iterator ();
            while (it.hasNext ()) {
                ModuleItem item = (ModuleItem)it.next ();
                item.setEnabledNoNotify (true);
            }
            
            toEnableLater.clear();
            toEnableLater.addAll(c);

            // disable silently
            List yetDisabled = Arrays.asList(
                                        ModuleInstaller.getModuleItems(
                                            ModuleInstallerSupport.DISABLED_MODULE));
            disable.removeAll(yetDisabled);

            it = disable.iterator ();
            while (it.hasNext ()) {
                ModuleItem item = (ModuleItem)it.next ();
                item.setEnabledNoNotify (false);
            }
            
            toDisableLater.clear();
            toDisableLater.addAll(disable);
        }
    }
    
    /* Used by ModuleConfigPanel to get prepared modules to enabled
     */
    static List getToEnableLater() {
        synchronized (toEnableLater) {
            return new LinkedList(toEnableLater);
        }
    }
    
    /* Used by ModuleConfigPanel to get prepared modules to disable
     */
    static List getToDisableLater() {
        synchronized (toEnableLater) {
            return new LinkedList(toDisableLater);
        }
    }

    /** Reverse back temporary change of modules state, should be called under toEnableLater lock
     */
    private static void undoUpdate() {
        Iterator it = toEnableLater.iterator ();
        while (it.hasNext ()) {
            ModuleItem mi = (ModuleItem)it.next ();
            mi.setEnabledNoNotify (false);
        }

        // enable modules
        it = toDisableLater.iterator();
        while (it.hasNext()) {
            ModuleItem mi = (ModuleItem)it.next ();
            mi.setEnabledNoNotify (true);
        }
    }
    
    /** Property editor for boolean property. Returns null if its value is uknown. Is used
     * instead of standard boolean editor because PropertyDisplayer displays null value as '[boolean]'
     */
    public static class BooleanPropertyEditor extends PropertyEditorSupport {

        public BooleanPropertyEditor() {
        }

        public String getAsText() {
            if (getValue() == null)
                return null;
            return ((Boolean)getValue()).booleanValue() ? "True" : "False";
        }

        public void setAsText(String text) throws java.lang.IllegalArgumentException {
            if (text == null)
                setValue(null);
            else if ("true".equalsIgnoreCase(text))
                setValue(Boolean.TRUE);
            else if ("false".equalsIgnoreCase(text))
                setValue(Boolean.FALSE);
            else
                throw new java.lang.IllegalArgumentException(text);
        }

        public String[] getTags() {
            return new String[] {"True", "False"};
        }
    }
    
    public static abstract class ModuleNewType extends NewType {
    }

}
