/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*; // Collections methods

import org.openide.modules.ModuleDescription;
import org.openide.util.WeakSet;
import org.openide.util.enum.RemoveDuplicatesEnumeration;
import org.openide.util.enum.SequenceEnumeration;

/** Classloader which behaves as if it had multiple parents.
 * That is, all parents are searched for classes and resources.
 *
 * <p>As a speed optimization, each classloader (this one and its parents) receives
 * a "domain" consisting of some set of packages. Once a class or resource has been
 * successfully found in some classloader (this one's own URLs, or one of its parents)
 * then only that classloader will subsequently be asked to look for other classes or
 * resources in the same package. This acts as a shortcut permitting searches to be
 * faster. If one classloader attempts to "patch" classes or resources occurring in
 * another classloader in the same package, that attempt will result in failure to
 * find/resolve something else, which may be just as well. Also note that this classloader
 * will refuse to load anything from the default package, again probably just as well.
 *
 * <p>As a further speed optimization, all recursive parent classloaders are precomputed
 * and uniquified and indirect ancestors are directly asked to load classes/resources,
 * without wasting time going through intermediate classloader, whenever this is possible.
 *
 * @author Jesse Glick
 */
class MultiURLClassLoader extends URLClassLoader implements ModuleDescription.PackageAccessibleClassLoader {

    private static final Set instances = new WeakSet (); // Set<MultiURLClassLoader>
    
    private static final ClassLoader systemClassLoader = ClassLoader.getSystemClassLoader ();
    static {
        if (systemClassLoader == null) {
            // Should still work, but definitely not a good sign:
            System.err.println ("Warning: appear to be running in oldjava! ClassLoader.systemClassLoader -> null"); // NOI18N
        }
    }

    private ClassLoader[] parents;

    /** Create a multi-parented classloader (which can also have its own URLs added).
     * Note that if you pass an empty list of parents, it will behave almost like a plain URLClassLoader
     * with the default parent (system classloader), other than the special domain behavior.
     * @param urls initial URLs specific to this classloader
     * @param parents list of parent classloaders (maybe none). It is best if they implement
     *        {@link ModuleDescription.PackageAccessibleClassLoader}, or else package-related
     *        methods will not fully work.
     * @throws IllegalArgumentException if there are any nulls or duplicate parent loaders or cycles
     */
    public MultiURLClassLoader (URL[] urls, ClassLoader[] parents) throws IllegalArgumentException {
        super (urls);
        synchronized (instances) {
            instances.add (this);
        }
        if (parents.length == 0 && systemClassLoader != null) {
            parents = new ClassLoader[] { systemClassLoader };
        }
        Set check = new HashSet (Arrays.asList (parents)); // Set<ClassLoader>
        if (check.size () < parents.length) throw new IllegalArgumentException ("duplicate parents"); // NOI18N
        if (check.contains (null)) throw new IllegalArgumentException ("null parent"); // NOI18N
        this.parents = coalesceParents (parents);
    }

    /** Add a new parent dynamically.
     * @param parent the new parent to add (appended to list)
     * @throws IllegalArgumentException in case of a null or cyclic parent (duplicate OK)
     */
    public void append (ClassLoader parent) throws IllegalArgumentException {
        if (parent == null) throw new IllegalArgumentException ("null parent"); // NOI18N
        parents = coalesceAppend (parents, parent);
    }

    /** Try to destroy this classloader.
     * It will be removed from the parent lists of any existing multiclassloaders.
     */
    public void destroy () {
        synchronized (instances) {
            Iterator it = instances.iterator ();
            while (it.hasNext ()) {
                MultiURLClassLoader l = (MultiURLClassLoader) it.next ();
                synchronized (l) {
                    ClassLoader[] p = l.parents;
                    int i;
                    for (i = 0; i < p.length; i++) {
                        if (p[i] == this) {
                            break;
                        }
                    }
                    if (i < p.length) {
                        ClassLoader[] p2 = new ClassLoader[p.length - 1];
                        if (i > 0) System.arraycopy (p, 0, p2, 0, i);
                        if (i < p.length - 1) System.arraycopy (p, i + 1, p2, i, p.length - i - 1);
                        l.parents = p2;
                    }
                    Collection c = l.domainsBySlashPackage.values ();
                    while (c.remove (this)) ;
                    c = l.domainsByDotPackage.values ();
                    while (c.remove (this)) ;
                }
            }
        }
    }

    // Caches which tell it which packages are "owned" by which classloaders.
    // Both the maps are kept in synch; they only differ in whether dot-separated
    // packages names are used for keys (good for classes) or slash-separated (good
    // for resources). Each value is either a parent classloader, or this.
    private final Map domainsBySlashPackage = new HashMap (); // Map<String,ClassLoader>
    private final Map domainsByDotPackage = new HashMap (); // Map<String,ClassLoader>

    // XXX fix later!
    // Currently ICE browser JAR includes a copy of many XML classes, including
    // two classes in org.w3c.dom.events which are not in Xerces for some reason.
    // This is a temporary workaround to permit these two classes to be loaded
    // by the ICE browser even though there is package skew going on.
    private static final boolean hacky (final String pkg) {
        return pkg.equals ("org.w3c.dom.events") || // NOI18N
               pkg.equals ("org/w3c/dom/events"); // NOI18N
    }

    protected Class loadClass (final String name, final boolean resolve) throws ClassNotFoundException {
        return loadClass (name, resolve, false);
    }
    // Optimized version we can call recursively
    // which will avoid checking parents unnecessarily:
    private synchronized final Class loadClass (final String name, final boolean resolve, final boolean fast) throws ClassNotFoundException {
        if (parents.length == 0) {
            // Should never happen
            return super.loadClass (name, resolve);
        }
        // First check if it isn't already loaded.
        Class clazz = findLoadedClass (name);
        if (clazz != null) {
            if (resolve) resolveClass (clazz);
            return clazz;
        }
        if (fast) {
            // Child told me that I specifically should be able to load it, if anyone can.
            // If this fails, throws CNFE, end of story.
            clazz = findClass (name);
            // Got it.
            final int dotIdx = name.lastIndexOf ('.');
            if (dotIdx == -1) {
                // Sorry.
                throw new ClassNotFoundException ("Will not load classes from the default package"); // NOI18N
            }
            final String pkg = name.substring (0, dotIdx);
            domainsByDotPackage.put (pkg, this);
            domainsBySlashPackage.put (pkg.replace ('.', '/'), this);
            if (resolve) resolveClass (clazz);
            return clazz;
        } else {
            // Normal mode--check parents as well as this.
            // First MURLCL consulted. Do a (hopefully) fast domain check.
            final int dotIdx = name.lastIndexOf ('.');
            if (dotIdx == -1) {
                // Sorry.
                throw new ClassNotFoundException ("Will not load classes from the default package"); // NOI18N
            }
            final String pkg = name.substring (0, dotIdx);
            final ClassLoader owner = hacky (pkg) ? null : (ClassLoader) domainsByDotPackage.get (pkg);
            if (owner == this) {
                // May throw CNFE:
                clazz = findClass (name);
                if (resolve) resolveClass (clazz);
                return clazz;
            } else if (owner != null) {
                // May throw CNFE:
                if (owner instanceof MultiURLClassLoader) {
                    // We know that MURLCL really owns it, so it should not check its parents:
                    clazz = ((MultiURLClassLoader) owner).loadClass (name, false, true);
                } else {
                    // No way to tell an arbitrary ClassLoader not to check parents, oh well.
                    // However if a parent finds it, next time we will know...
                    clazz = owner.loadClass (name);
                }
                if (resolve) resolveClass (clazz);
                return clazz;
            } else {
                // Virgin package? OK, check parents in turn.
                for (int i = 0; i < parents.length; i++) {
                    try {
                        final ClassLoader owner2; // who got it?
                        if (parents[i] instanceof MultiURLClassLoader) {
                            // Again, don't check *its* parent! We already have its parent
                            // in our *own* parent list, and before it too.
                            clazz = ((MultiURLClassLoader) parents[i]).loadClass (name, false, true);
                            owner2 = parents[i];
                        } else {
                            clazz = parents[i].loadClass (name);
                            // Not really sure if this classloader got it, or some parent. Let's check.
                            owner2 = clazz.getClassLoader ();
                        }
                        // Got it.
                        domainsByDotPackage.put (pkg, owner2);
                        domainsBySlashPackage.put (pkg.replace ('.', '/'), owner2);
                        if (resolve) resolveClass (clazz);
                        return clazz;
                    } catch (ClassNotFoundException cnfe2) {
                        // OK.
                    }
                }
                // That didn't work--we have to load it ourselves.
                // If this throws CNFE, it will be thrown from the method.
                clazz = findClass (name);
                // Got it.
                domainsByDotPackage.put (pkg, this);
                domainsBySlashPackage.put (pkg.replace ('.', '/'), this);
                if (resolve) resolveClass (clazz);
                return clazz;
            }
        }
    }
    
    // Logic essentially the same as above.
    public URL getResource (final String name) {
        return getResource (name, false);
    }
    private synchronized final URL getResource (final String name, final boolean fast) {
        if (parents.length == 0) {
            // Should never happen
            return super.getResource (name);
        }
        if (fast) {
            // Don't check parent(s)...
            final URL u = findResource (name);
            if (u != null) {
                // Got it.
                final int slashIdx = name.lastIndexOf ('/');
                if (slashIdx == -1) {
                    // No dice.
                    return null;
                }
                final String pkg = name.substring (0, slashIdx);
                domainsByDotPackage.put (pkg.replace ('/', '.'), this);
                domainsBySlashPackage.put (pkg, this);
            }
            return u;
        } else {
            // Fine, entry to getResource, check parents if necessary.
            final int slashIdx = name.lastIndexOf ('/');
            if (slashIdx == -1) {
                // No dice.
                return null;
            }
            final String pkg = name.substring (0, slashIdx);
            final ClassLoader owner = hacky (pkg) ? null : (ClassLoader) domainsBySlashPackage.get (pkg);
            if (owner == this) {
                return findResource (name);
            } else if (owner != null) {
                if (owner instanceof MultiURLClassLoader) {
                    return ((MultiURLClassLoader) owner).getResource (name, true);
                } else {
                    return owner.getResource (name);
                }
            } else {
                // Virgin package.
                for (int i = 0; i < parents.length; i++) {
                    final URL u;
                    if (parents[i] instanceof MultiURLClassLoader) {
                        u = ((MultiURLClassLoader) parents[i]).getResource (name, true);
                    } else {
                        u = parents[i].getResource (name);
                    }
                    if (u != null) {
                        // Got it. In fact if the parent was not a MURLCL it might be an ancestor of this parent that
                        // got it, rather than directly, but since there is no way to extract the originating CL from
                        // the URL we just treat the direct parent as the owner. It will always search its own parents
                        // later anyway, so the search will work, just a trifle slower.
                        domainsByDotPackage.put (pkg.replace ('/', '.'), parents[i]);
                        domainsBySlashPackage.put (pkg, parents[i]);
                        return u;
                    }
                }
                final URL u = findResource (name);
                if (u != null) {
                    // Got it.
                    domainsByDotPackage.put (pkg.replace ('/', '.'), this);
                    domainsBySlashPackage.put (pkg, this);
                }
                return u;
            }
        }
    }
    
    public synchronized Enumeration findResources (String name) throws IOException {
        if (parents.length == 0) {
            return super.findResources (name);
        }
        // Somewhat weirder due to illogical design of related methods in Java 2
        // platform but still feasible.
        // (Why is it public in URLClassLoader?)
        // (Who would throw IOException?)
        // (Why is getResources final?)
        // (Does anyone ever really call this method anyway?)
        // Don't bother optimizing this call by domains--it is practically unused.
        Enumeration e = super.findResources (name);
        for (int i = parents.length - 1; i >= 0; i--) {
            e = new SequenceEnumeration (parents[i].getResources (name), e);
        }
        return new RemoveDuplicatesEnumeration (e);
    }
    
    // Now it gets a little stickier.
    // The Java API is not very friendly.
    private final Map packageCache = new HashMap (); // Map<String,Package>

    protected synchronized Package getPackage (String name) {
        if (parents.length == 0) {
            return super.getPackage (name);
        }
        synchronized (packageCache) {
            Package p = (Package) packageCache.get (name);
            if (p == null) {
                for (int i = 0; i < parents.length; i++) {
                    if (parents[i] instanceof ModuleDescription.PackageAccessibleClassLoader) {
                        p = ((ModuleDescription.PackageAccessibleClassLoader) parents[i]).getPackageAccessibly (name);
                        if (p != null) {
                            packageCache.put (name, p);
                            break;
                        }
                    }
                }
                if (p == null) {
                    p = super.getPackage (name);
                    if (p != null) {
                        packageCache.put (name, p);
                    }
                }
            }
            return p;
        }
    }
    
    protected synchronized Package[] getPackages () {
        if (parents.length == 0) {
            return super.getPackages ();
        }
        Map all = new HashMap (); // Map<String,Package>
        addPackages (all, super.getPackages ());
        for (int i = 0; i < parents.length; i++) {
            if (parents[i] instanceof ModuleDescription.PackageAccessibleClassLoader) {
                addPackages (all, ((ModuleDescription.PackageAccessibleClassLoader) parents[i]).getPackagesAccessibly ());
            }
        }
        synchronized (packageCache) {
            Iterator it = all.entrySet ().iterator ();
            while (it.hasNext ()) {
                Map.Entry entry = (Map.Entry) it.next ();
                Object name = entry.getKey ();
                if (! packageCache.containsKey (name)) {
                    packageCache.put (name, entry.getValue ());
                }
            }
        }
        return (Package[]) all.values ().toArray (new Package[all.size ()]);
    }
    
    private void addPackages (Map all, Package[] pkgs) {
        // Would be easier if Package.equals() was just defined sensibly...
        for (int i = 0; i < pkgs.length; i++) {
            all.put (pkgs[i].getName (), pkgs[i]);
        }
    }
    
    public Package getPackageAccessibly (String name) {
        return getPackage (name);
    }
    
    public Package[] getPackagesAccessibly () {
        return getPackages ();
    }

    /** Coalesce parent classloaders into an optimized set.
     * This means that all parents of the specified classloaders
     * are also added recursively, removing duplicates along the way.
     * Search order should be preserved (parents before children, stable w.r.t. inputs).
     * @param loaders list of suggested parents (no nulls or duplicates permitted)
     * @return optimized list of parents (no nulls or duplicates)
     * @throws IllegalArgumentException if there are cycles
     */
    private ClassLoader[] coalesceParents (ClassLoader[] loaders) throws IllegalArgumentException {
        int likelySize = loaders.length * 3 + 1;
        Set resultingUnique = new HashSet (likelySize); // Set<ClassLoader>
        List resulting = new ArrayList (likelySize); // List<ClassLoader>
        for (int i = 0; i < loaders.length; i++) {
            addRec (resultingUnique, resulting, loaders[i]);
        }
        return (ClassLoader[]) resulting.toArray (new ClassLoader[resulting.size ()]);
    }

    /** Coalesce a new loader into the existing ones.
     */
    private ClassLoader[] coalesceAppend (ClassLoader[] existing, ClassLoader appended) throws IllegalArgumentException {
        int likelySize = existing.length + 3;
        Set resultingUnique = new HashSet (likelySize);
        List existingL = Arrays.asList (existing);
        resultingUnique.addAll (existingL);
        if (resultingUnique.contains (appended)) {
            // No change required.
            return existing;
        }
        List resulting = new ArrayList (likelySize);
        resulting.addAll (existingL);
        addRec (resultingUnique, resulting, appended);
        return (ClassLoader[]) resulting.toArray (new ClassLoader[resulting.size ()]);
    }

    private void addRec (Set resultingUnique, List resulting, ClassLoader loader) throws IllegalArgumentException {
        if (loader == this) throw new IllegalArgumentException ("cycle in parents"); // NOI18N
        if (resultingUnique.contains (loader)) return;
        if (loader instanceof MultiURLClassLoader) {
            ClassLoader[] parents = ((MultiURLClassLoader) loader).parents;
            for (int i = 0; i < parents.length; i++) {
                addRec (resultingUnique, resulting, parents[i]);
            }
        } else {
            ClassLoader parent = loader.getParent ();
            if (parent != null) {
                addRec (resultingUnique, resulting, parent);
            }
        }
        resultingUnique.add (loader);
        resulting.add (loader);
    }

    // Debugging method called from ModuleItem:
    synchronized void dumpURLs () {
        // Not recursive since we already have the full search path precomputed.
        // Just check direct URLs.
        for (int i = 0; i < parents.length; i++) {
            if (parents[i] instanceof URLClassLoader) {
                URLClassLoader urlloader = (URLClassLoader) parents[i];
                URL[] urls = urlloader.getURLs ();
                for (int j = 0; j < urls.length; j++) {
                    System.err.println ("\t\t[" + urlloader + "] " + urls[j]); // NOI18N
                }
            } else {
                System.err.println ("\t\t[" + parents[i] + "] not a URLClassLoader"); // NOI18N
            }
        }
    }
    
}
