/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.*;
import java.awt.datatransfer.Clipboard;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.lang.ref.SoftReference;
import java.net.URL;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;

import org.openide.*;
import org.openide.explorer.*;
import org.openide.explorer.propertysheet.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.options.ControlPanel;
import org.openide.util.*;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.ExClipboard;
import org.openide.util.io.*;
import org.openide.windows.*;

import org.netbeans.core.actions.*;
import org.netbeans.core.windows.ModeImpl;
import org.netbeans.core.windows.WellKnownModeNames;
import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.WorkspaceImpl;

/** Class that provides operations on nodes. Any part of system can
* ask for opening a customizer or explorer on any node. These actions
* are accessible thru this class.
*/
public class NbNodeOperation extends TopManager.NodeOperation {

    /**
     * Name of a property that can be passed in a Node instance. The value
     * of the property must be String and can be an alternative to displayName.
     */
    private static final String PROP_LONGER_DISPLAY_NAME = "longerDisplayName"; // NOI18N
    
    /** Shows an explorer on the given root Node.
    * @param n the Node that will be the rootContext of the explored hierarchy
    */
    public void explore (final Node n) {
        Mutex.EVENT.readAccess (new Runnable () {
                public void run () {
                    NbMainExplorer.ExplorerTab et = new NbMainExplorer.ExplorerTab ();
                    et.getExplorerManager ().setRootContext (n);
                    // create special mode for explorer and name it properly, then opet it
                    WindowManagerImpl wm =
                        (WindowManagerImpl)TopManager.getDefault().getWindowManager();
                    WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
                    String modeName = wm.findUnusedModeName(et.getName(), curWorkspace);
                    curWorkspace.createMode
                        (modeName, NbBundle.getBundle(NbNodeOperation.class).getString("CTL_ExplorerTitle"),
                         null, ModeImpl.SINGLE, true)
                        .dockInto(et);
                    et.open ();
                }
            });
    }

    /** Tries to open customization for specified node. The dialog is
    * open in modal mode and the function returns after successful
    * customization.
    *
    * @param n the node to customize
    * @return <CODE>true</CODE> if the node has customizer,
    * <CODE>false</CODE> if not
    */
    public boolean customize (Node n) {
        final Component customizer = n.getCustomizer ();
        if (customizer == null) return false;
        return ((Boolean) Mutex.EVENT.readAccess (new Mutex.Action () {
                public Object run () {
                    if (customizer instanceof NbPresenter) { // #9466
                        ((NbPresenter) customizer).pack ();
                        ((NbPresenter) customizer).show ();
                        return Boolean.TRUE;
                    }
                    if (customizer instanceof Window) {
                        ((Window) customizer).pack ();
                        customizer.setVisible (true);
                        return Boolean.TRUE;
                    }

                    DialogDescriptor dd = new DialogDescriptor
                        (customizer, 
                         NbBundle.getMessage (ModuleFSSection.class, "CTL_Customizer_dialog_title"));
                    dd.setOptions(new Object[] { DialogDescriptor.CLOSED_OPTION });

                    Dialog dialog = TopManager.getDefault ().createDialog(dd);

                    dialog.pack();
                    dialog.show();
                    return Boolean.TRUE;
                }
            })).booleanValue ();
    }

    /** Opens a modal propertySheet on given Node
    * @param n the node to show properties for
    */
    public void showProperties (Node n) {
        Sheet s = new Sheet ();
        s.setNodes (new Node[] { n });
        openProperties(s);
    }

    /** Opens a modal propertySheet on given set of Nodes
    * @param n the array of nodes to show properties for
    */
    public void showProperties (Node[] nodes) {
        Sheet s = new Sheet ();
        s.setNodes (nodes);
        openProperties(s);
    }

    /** Opens explorer for specified root in modal mode. The set
    * of selected components is returned as a result. The acceptor
    * should be asked each time selected nodes changes to accept or
    * reject the current result. This should affect for example the
    * <EM>OK</EM> button.
    *
    * @param title is a title that will be displayed as a title of the window
    * @param root the root to explore
    * @param acceptor the class that is asked for accepting or rejecting
    *    current selection
    * @param top is a component that will be displayed on the top
    * @return array of selected (and accepted) nodes
    *
    * @exception UserCancelException selection interrupted by user
    */
    public Node[] select (String title, String rootTitle, Node root, NodeAcceptor acceptor, Component top)
    throws UserCancelException {
        final FileSelector selector = new FileSelector (title, rootTitle, root, acceptor, top);
        Mutex.EVENT.readAccess (new Mutex.Action () {
                public Object run () {
                    selector.show();
                    return null;
                }
            });
        if (selector.cancelFlag)
            throw new UserCancelException ();
        return selector.getNodes ();
    }

    /** Helper method, opens properties top component in single mode
    * and requests a focus for it */
    private void openProperties (final TopComponent tc) {
        Mutex.EVENT.readAccess (new Runnable () {
                public void run () {
                    WindowManagerImpl wm =
                        (WindowManagerImpl)TopManager.getDefault().getWindowManager();
                    WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
                    String modeName = wm.findUnusedModeName(tc.getName(), curWorkspace);
                    curWorkspace.createMode(
                                            modeName, modeName, null, ModeImpl.SINGLE, true
                                            ).dockInto(tc);
                    tc.open();
                    tc.requestFocus();
                }
            });
    }

    /** Default view for properties.
    */
    public static class Sheet extends TopComponent {
        /** generated Serialized Version UID */
        static final long serialVersionUID = 7807519514644165460L;

        /** shared sheet */
        private static Sheet sharedSheet;
        /** The SheetFrame's icon SoftReference (Image) */
        private static SoftReference sheetIcon;
        /** listener to the property changes */
        transient private Listener listener;
        /** listener to the node changes, especially their destruction */
        transient private SheetNodesListener snListener;
        /** Should property sheet listen to the global changes ? */
        boolean global;
        /** the property sheet that is used to display nodes */
        private PropertySheet propertySheet;
        /** the nodes that are displayed in the property sheet */
        private Node[] nodes;

        /** Constructor for new sheet.
        * The sheet does not listen to global changes */
        public Sheet () {
            this (false);
        }

        /** @param global should the content change when global properties changes?
        */
        public Sheet (boolean global) {
            this.global = global;
            this.propertySheet = new PropertySheet ();

            setLayout (new java.awt.BorderLayout ());
            add ("Center", propertySheet); // NOI18N

            initIcon ();

            // name listener and node listener
            listener = new Listener ();
            if (!global) {
                snListener = new SheetNodesListener();
            }

            updateGlobalListening(true);
        }

        /** Provides the shared instance of SheetFrame */
        public static Sheet getDefault () {
            if (sharedSheet == null)
                sharedSheet = new Sheet (true);
            return sharedSheet;
        }

        public HelpCtx getHelpCtx () {
            return ExplorerPanel.getHelpCtx (nodes, new HelpCtx (Sheet.class));
        }
        
        /** always open global property sheet in its special mode */
        public void open (Workspace workspace) {
            if (global) {
                Workspace realWorkspace = (workspace == null)
                                          ? TopManager.getDefault().getWindowManager().getCurrentWorkspace()
                                          : workspace;
                Mode tcMode = realWorkspace.findMode(this);
                if (tcMode == null) {
                    // dock into our mode if not docked yet
                    Mode mode = realWorkspace.findMode(WellKnownModeNames.PROPERTIES);
                    if (mode == null) {
                        mode = realWorkspace.createMode(
                            WellKnownModeNames.PROPERTIES,
                            NbBundle.getBundle(WindowManagerImpl.class).getString("CTL_PropertiesWindow"),
                            null
                        );
                    }
                    mode.dockInto(this);
                }
            }
            // behave like superclass
            super.open(workspace);
        }
        

        /** Provides the icon to be shown as this Window's icon.
        * @return the window's icon
        */
        private void initIcon () {
            Image icon = sheetIcon == null ? null : (Image)sheetIcon.get ();
            if (icon == null) {
                icon = java.awt.Toolkit.getDefaultToolkit ().getImage (
                           getClass ().getResource (
                               org.openide.util.Utilities.isLargeFrameIcons() ?
                               "/org/netbeans/core/resources/frames/properties32.gif" : // NOI18N
                               "/org/netbeans/core/resources/frames/properties.gif" // NOI18N
                           )
                       );
                sheetIcon = new SoftReference (icon);
            }
            setIcon (icon);
        }

        /** Changes name of the component to reflect currently displayed nodes.
        * Called when set of displayed nodes has changed.
        */
        protected void updateTitle () {
            // different naming for global and local sheets
            Mode ourMode = TopManager.getDefault().getWindowManager().
                           getCurrentWorkspace().findMode(this);
            String nodeTitle =  null;
            if (nodes.length == 0) {
                nodeTitle = "";  // NOI18N
            } else {
                nodeTitle = nodes[0].getDisplayName();
                Object alternativeDisplayName = nodes[0].getValue(PROP_LONGER_DISPLAY_NAME);
                if (alternativeDisplayName instanceof String) {
                    nodeTitle = (String)alternativeDisplayName;
                }
            }
            Object[] titleParams = new Object[] {
                new Integer (nodes.length),
                nodeTitle
            };
            // different naming if docked in properties mode
            if ((ourMode != null) && 
                (WellKnownModeNames.PROPERTIES.equals(ourMode.getName()))) {
                setName(
                    java.text.MessageFormat.format(
                        NbBundle.getBundle(NbNodeOperation.class).getString("CTL_FMT_GlobalProperties"),
                        titleParams
                    )
                );
            } else {
                setName(
                    java.text.MessageFormat.format(
                        NbBundle.getBundle(NbNodeOperation.class).getString("CTL_FMT_LocalProperties"),
                        titleParams
                    )
                );
            }
        }

        /** Nodes to display.
        */
        public void setNodes (Node[] nodes) {
            setNodesWithoutReattaching(nodes);
            // re-attach to listen to new nodes
            if (!global) {
                snListener.detach();
                snListener.attach(nodes);
            }
        }

        /** Helper method, called from SheetNodesListener inner class */
        private void setNodesWithoutReattaching (Node[] nodes) {
            this.nodes = nodes;
            propertySheet.setNodes(nodes);
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    updateTitle();
                }
            });
        }

        public Dimension getPreferredSize () {
            return new Dimension (250, 400);
        }

        /** Serialize this property sheet */
        public void writeExternal (ObjectOutput out)
        throws IOException {
            super.writeExternal(out);
            
            if (global) {
                // write dummy array
                out.writeObject (null);
            } else {
                Node.Handle[] arr = NodeOp.toHandles (nodes);
                out.writeObject(arr);
            }
            
            out.writeBoolean(global);
        }

        /** Deserialize this property sheet. */
        public void readExternal (ObjectInput in)
        throws IOException, ClassNotFoundException {
            try {
                super.readExternal(in);
            } catch (SafeException se) {
                // ignore--we really do not care about the explorer manager that much
                //System.err.println("ignoring a SafeException: " + se.getLocalizedMessage ());
            }
            Object obj = in.readObject ();

            if (obj instanceof NbMarshalledObject || obj instanceof ExplorerManager) {
                // old version read the Boolean
                global = ((Boolean)in.readObject()).booleanValue();
            } else {
                Node[] nodes;
              
                if (obj == null) {
                    // handles can also be null for global 
                    // property sheet
                    nodes = TopComponent.getRegistry().getActivatedNodes();
                } else {
                    // new version, first read the nodes and then the global boolean
                    Node.Handle[] arr = (Node.Handle[])obj;

                    try {
                        nodes = NodeOp.fromHandles (arr);
                    } catch (IOException ex) {
                        ErrorManager errMgr = TopManager.getDefault ().getErrorManager ();
                        errMgr.annotate (ex, NbBundle.getBundle(NbNodeOperation.class).getString("EXC_CannotLoadNodes"));
                        errMgr.notify (ErrorManager.INFORMATIONAL, ex);
                        nodes = new Node[0];
                    }
                }

                global = in.readBoolean ();

                setNodes (nodes);
            }

            /*
                  if (obj instanceof Boolean) {
                    global = (Boolean)in.readObject ()

                  global = ((Boolean)in.readObject()).booleanValue();
            /*
                  // start global listening if needed, but wait until
                  // deserialization is done (ExplorerManager is uses
                  // post-deserialization validating too, so we are forced
                  // to use it)
                  ((ObjectInputStream)in).registerValidation(
                    new ObjectInputValidation () {
                      public void validateObject () {
                        updateGlobalListening(false);
                      }
                    }, 0
                  );
            */
            // JST: I guess we are not and moreover the type casting is really ugly
            //      updateGlobalListening (global);
        }

        /** Resolve to singleton instance, if needed. */
        public Object readResolve ()
        throws ObjectStreamException {
            if (global) {
                return getDefault();
            } else {
                if ((nodes == null) || (nodes.length <= 0)) {
                    return null;
                }
            }
            return this;
        }

        /** Helper, listener variable must be initialized before
        * calling this */
        private void updateGlobalListening (boolean activateNow) {
            if (global) {
                TopComponent.getRegistry().addPropertyChangeListener(
                    WeakListener.propertyChange(listener, TopComponent.getRegistry ())
                );
                if (activateNow) {
                    listener.activate ();
                }
            }
        }

        /** Change listener to changes in selected nodes. And also
        * nodes listener to listen to global changes of the nodes.
        */
        private class Listener extends Object
            implements PropertyChangeListener {
            public void propertyChange (PropertyChangeEvent ev) {
                if ((ev.getPropertyName().equals(TopComponent.Registry.PROP_ACTIVATED)) && 
                    (ev.getNewValue() == Sheet.this)) {
                    return; // we do not want to call setNodes if we are
                            // the activated window
                }
                activate ();
            }

            public void activate () {
                Node[] arr = TopComponent.getRegistry ().getActivatedNodes();
                setNodes (arr);
            }

        }
        
        /** Change listener to changes in selected nodes. And also
        * nodes listener to listen to global changes of the nodes.
        */
        private class SheetNodesListener extends Object implements NodeListener {
            
            /* maps nodes to their listeners (Node, WeakListener) */
            HashMap listenerMap;
            
            public void propertyChange (PropertyChangeEvent ev) {
            }
            
            public void childrenAdded(NodeMemberEvent ev) {
            }
            
            public void childrenRemoved(NodeMemberEvent ev) {
            }
            
            public void childrenReordered(NodeReorderEvent ev) {
            }

            /** Fired when the node is deleted.
             * @param ev event describing the node
             */
            public void nodeDestroyed(NodeEvent ev) {
                Node destroyedNode = ev.getNode();
                NodeListener listener = (NodeListener)listenerMap.get(destroyedNode);
                // stop to listen to destroyed node
                destroyedNode.removeNodeListener(listener);
                listenerMap.remove(destroyedNode);
                // close top component (our outer class) if last node was destroyed
                if (listenerMap.isEmpty()) {
                    close();
                } else {
                    setNodesWithoutReattaching(
                        (Node[])(listenerMap.keySet().toArray(new Node[listenerMap.size()]))
                    );
                }
            }
            
            public void attach (Node[] nodes) {
                listenerMap = new HashMap(nodes.length * 2);
                NodeListener curListener = null;
                // start to listen to all given nodes and map nodes to
                // their listeners
                for (int i = 0; i < nodes.length; i++) {
                    curListener = WeakListener.node(this, nodes[i]);
                    listenerMap.put(nodes[i], curListener);
                    nodes[i].addNodeListener(curListener);
                };
            }
            
            public void detach () {
                if (listenerMap == null) {
                    return;
                }
                // stop to listen to all nodes
                Node curNode = null;
                NodeListener curListener = null;
                Map.Entry curEntry = null;
                for (Iterator iter = listenerMap.entrySet().iterator(); iter.hasNext(); ) {
                    curEntry = (Map.Entry)iter.next();
                    ((Node)curEntry.getKey()).removeNodeListener((NodeListener)curEntry.getValue());
                }
                // destroy the map
                listenerMap = null;
            }
            
        } // end of SheetNodesListener

    }


}
