/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.MessageFormat;

import org.openide.*;
import org.openide.cookies.*;
import org.openide.loaders.*;
import org.openide.filesystems.*;
import org.openide.util.NbBundle;
import org.openide.util.io.*;
import org.openide.nodes.*;

import org.netbeans.core.windows.WindowManagerImpl;

/** Holder for current project. Also provides the default one.
*
* @author Jaroslav Tulach
*/
class NbProjectOperation extends Object {
    /** current project */
    private static ProjectCookie project;

    /** node to be used to mark that the project does not have desktop */
    private static final Node NO_DESKTOP = new DesktopNode ();

    /** desktop node for current project */
    private static Node projectDesktop = NO_DESKTOP;

    // ----------------------------------------------------------------------------------------
    // Default simple ProjectCookie persistence

    /** Save the basic project.
    */
    static void saveBasicProject () throws IOException {
        project.projectSave ();
    }

    // ----------------------------------------------------------------------------------------
    // Static methods for manipulation with projects

    static void storeLastProject () {
        //    if (new IDESettings ().getConfirmSaveOnExit ()) {
        //      if (TopManager.notify (...)
        //    }
        try {
            project.projectSave ();
            NbProjectOperation.saveLastProjectUsed ();
        } catch (SafeException exc) {
            TopManager.getDefault ().getErrorManager ().notify (
                ErrorManager.INFORMATIONAL, exc.getException()
            );
        } catch (IOException exc) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, exc);
        }
    }


    static void saveLastProjectUsed () throws IOException {
        FileObject nameFile = getLastProjectFile (true);
        FileLock lock = nameFile.lock ();
        try {
            ObjectOutputStream oos = new NbObjectOutputStream (nameFile.getOutputStream (lock));
            try {
                oos.writeObject (project);
            } finally {
                oos.close ();
            }
        } finally {
            lock.releaseLock ();
        }
    }

    /** Getter for current project.
    * @return the project
    */
    public static ProjectCookie getProject () {
        return project;
    }

    /** Getter for the project desktop node.
    */
    public static Node getProjectDesktop () {
        return projectDesktop;
    }

    /** Test if the project has own desktop or not */
    public static boolean hasProjectDesktop () {
        return projectDesktop != NO_DESKTOP;
    }

    /** Setter for changing the project.
    * @param p the new project to use
    * @exception IOException if the previous project cannot be closed or new opened
    */
    public static void setProject (final ProjectCookie p) throws IOException {
        if (project == p) {
            // no change in project
            return;
        }

        if (project != null) {
            project.projectSave();
        }
        
        Node oldDesktop = projectDesktop;
        projectDesktop = p.projectDesktop ();
        if (projectDesktop == null) {
            projectDesktop = NO_DESKTOP;
        }

        try {
            // open the project
            p.projectOpen ();
            if (project != null) project.projectClose();
        }
        catch (IOException e) {
            projectDesktop = oldDesktop;
            throw e;
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            // Random bug in project implementation -> deal with it.
            projectDesktop = oldDesktop;
            IOException ioe = new IOException ();
            TopManager.getDefault ().getErrorManager ().annotate (ioe, t);
            throw ioe;
        }

        project = p;

        if (!(project instanceof NbProject)) {
            saveLastProjectUsed ();
        }

        NbTopManager.get ().firePropertyChange (NbTopManager.PROP_PLACES, null, null);
    }

    /** Opens last used project or creates a default one. */
    public static synchronized void openOrCreateProject () throws IOException {
        ProjectCookie project = null;
        // the file, but do not create new one
        try {
            FileObject nameFile = getLastProjectFile (false);
            if (nameFile != null) {
                ObjectInputStream ois = new NbObjectInputStream (nameFile.getInputStream ());
                try {
                    project = (ProjectCookie)ois.readObject ();
                } finally {
                    ois.close ();
                }
            }
        } catch (SafeException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex.getException ());
        } catch (IOException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        } catch (ClassNotFoundException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        } catch (Exception ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        }

        NbProjectOperation.project = setOpeningProject(project);

    }
    
    /** Try to set up new project. If it fails then NbProject is used as default.
     * @param project project to set up. If <code>null</code> then NbProject
     * is used as default.
     * @return set project.
     */
    private static ProjectCookie setOpeningProject(ProjectCookie project) {
        try {
            if (project == null) {
                project = createDefaultProject();
            }
            setProject (project);
            return project;
        } catch (Exception exc) {
            if (project instanceof NbProject) {
                TopManager.getDefault ().getErrorManager ().notify (exc);
            } else {
                TopManager.getDefault ().getErrorManager ()
                .log (NbBundle.getBundle(NbProjectOperation.class)
                .getString("EXC_CorruptedProject"));
                return setOpeningProject(null);
            }
        }
        return createDefaultProject();
    }

    /** Creates and returns newly created default project */
    private static ProjectCookie createDefaultProject () {
        //System.out.println(NbBundle.getBundle(NbProjectOperation.class).
        //                   getString("MSG_DefaultProject"));
        // create empty new
        ProjectCookie project = new NbProject();
        WindowManagerImpl.createFromScratch();
        return project;
    }


    /** name and extension for basic serialized project file */
    private static final String LAST_PROJECT_NAME = "project"; // NOI18N
    private static final String LAST_PROJECT_EXT = "last"; // NOI18N

    /** Returns file object (file) where the name of last opened project is stored.
    * @param create true if the file should be created
    * @return the file object
    */
    private static FileObject getLastProjectFile (boolean create) throws java.io.IOException {
        org.openide.filesystems.FileSystem def =
            TopManager.getDefault().getRepository().getDefaultFileSystem ();

        FileObject fo = def.find ("", LAST_PROJECT_NAME, LAST_PROJECT_EXT); // NOI18N
        if (fo == null && create) {
            fo = def.getRoot ().createData (LAST_PROJECT_NAME, LAST_PROJECT_EXT);
        }
        return fo;

    }

}
