/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.ArrayList;

import javax.swing.*;

import org.openide.windows.*;
import org.openide.*;
import org.openide.util.enum.*;
import org.openide.util.NbBundle;

import org.netbeans.core.output.OutPane;
/**
 * Notifies exceptions.
 *
 * @author  Jaroslav Tulach
 */

final class NotifyException extends JScrollPane implements ActionListener {
    static final long serialVersionUID =3680397500573480127L;


    /** the instance */
    private static NotifyException INSTANCE = null;

    /** max text of exception */
    private static final int MAXIMUM_TEXT_WIDTH = 40;
    /** preferred width of this component */
    private static final int SIZE_PREFRERRED_WIDTH=550;
    /** preferred height of this component */
    private static final int SIZE_PREFRERRED_HEIGHT=250;

    /** enuration of NbExceptionManager.Exc to notify */
    private ArrayListPos exceptions;
    /** current exception */
    private NbErrorManager.Exc current;

    /** dialog descriptor */
    private DialogDescriptor descriptor;
    /** dialog that displayes the exceptions */
    private java.awt.Dialog dialog;
    /** button to show next exceptions */
    private JButton next;
    /** button to show previous exceptions */
    private JButton previous;
    /** details button */
    private JButton details;
    /** details window */
    private OutPane output;


    /** boolean to show/hide details */
    private boolean showDetails;

    /** Constructor.
    */
    private NotifyException () {
        super ();

        setPreferredSize(new Dimension(SIZE_PREFRERRED_WIDTH,SIZE_PREFRERRED_HEIGHT));
        exceptions=new ArrayListPos();

        next = new JButton (org.openide.util.NbBundle.getBundle(NotifyException.class).getString("CTL_NextException"));
        previous = new JButton (org.openide.util.NbBundle.getBundle(NotifyException.class).getString("CTL_PreviousException"));
        details = new JButton ();

        output = new OutPane ();
        setViewportView (output);


        descriptor = new DialogDescriptor ("", ""); // NOI18N

        descriptor.setModal (false);
        
        descriptor.setMessageType (DialogDescriptor.ERROR_MESSAGE);
        descriptor.setOptions (new Object[] {
                                   DialogDescriptor.OK_OPTION,
                                   previous,
                                   next
                               });
        descriptor.setAdditionalOptions (new Object[] {
                                             details
                                         });
        descriptor.setClosingOptions (new Object[0]);
        descriptor.setButtonListener (this);

        dialog = TopManager.getDefault ().createDialog (descriptor);
    }



    /** Adds new exception into the queue.
    */
    public static void notify (
        final NbErrorManager.Exc t
    ) {
        SwingUtilities.invokeLater (new Runnable () {
            public void run() {
                String glm = t.getLocalizedMessage();
                int gs = t.getSeverity();
                
                if (gs == ErrorManager.WARNING  &&  glm != null  &&  !glm.equals(""))
                    TopManager.getDefault().notify(
                        new NotifyDescriptor.Message(glm, NotifyDescriptor.WARNING_MESSAGE));
                else if (gs == ErrorManager.USER  &&  glm != null  &&  !glm.equals(""))
                    TopManager.getDefault().notify(
                        new NotifyDescriptor.Message(glm, NotifyDescriptor.INFORMATION_MESSAGE));
                else if (gs == ErrorManager.ERROR  &&  glm != null  &&  !glm.equals(""))
                    TopManager.getDefault().notify(
                        new NotifyDescriptor.Message(glm, NotifyDescriptor.ERROR_MESSAGE));
                else {
                    if (INSTANCE == null)
                        INSTANCE = new NotifyException();
                    INSTANCE.updateState(t);
                }
            }
        });
    }



    /** updates the state of the dialog. called only in AWT thread.
    */
    private void updateState (NbErrorManager.Exc t) {
        exceptions.add(t);
        if (!exceptions.existsNextElement()) {
            // the dialog is not shown
            current = t;
            update ();
            dialog.show ();
        } else {
            // add the exception to the queue
            //Fixed bug #9435, call of setVisible(true) replaced by call of show()
            //It did not work on Linux when JDialog is reused.
            dialog.show ();
            //if (!dialog.isVisible()) dialog.setVisible(true);
            next.setVisible (true);
        }
    }

    /** Updates the visual state of the dialog.
    */
    private void update () {
        // JST: this can be improved in future...
        String lm = current.getLocalizedMessage ();
        String nm = current.getMessage ();
        boolean isLocalized = current.isLocalized();

        next.setVisible (exceptions.existsNextElement());
        previous.setVisible (exceptions.existsPreviousElement());

        details.setText (
            showDetails
            ?
            org.openide.util.NbBundle.getBundle(NotifyException.class).getString("CTL_Exception_Hide_Details")
            :
            org.openide.util.NbBundle.getBundle(NotifyException.class).getString("CTL_Exception_Show_Details")
        );


        //    setText (current.getLocalizedMessage ());
        String title = org.openide.util.NbBundle.getBundle(NotifyException.class).getString("CTL_Title_Exception");

        if (showDetails) {
            descriptor.setMessage (createDetails ());
        } else {
            if (isLocalized) {
                String msg = current.getLocalizedMessage ();
                if (msg == null || "".equals(msg)) {  // NOI18N
                    msg = org.openide.util.Utilities.wrapString (
                              msg, MAXIMUM_TEXT_WIDTH, java.text.BreakIterator.getCharacterInstance(), false
                          );
                }
                descriptor.setMessage (msg);
            } else {
                // emphasize user-non-friendly exceptions
                //      if (this.getMessage() == null || "".equals(this.getMessage())) { // NOI18N
                descriptor.setMessage (
                    java.text.MessageFormat.format(
                        NbBundle.getBundle (NotifyException.class).getString("NTF_ExceptionalException"),
                        new Object[] {
                            current.getClassName (),
                            System.getProperty("netbeans.user").replace('\\','/')
                        }
                    )
                );

                title = NbBundle.getBundle (NotifyException.class).getString(
                            "NTF_ExceptionalExceptionTitle" // NOI18N
                        );
            }
        }

        descriptor.setTitle (title);
    }

    /** Create details.
    */
    private JComponent createDetails () {
        try {
            output.getOut ().reset ();
            current.printStackTrace (output.getOut ());
            output.setSelectedIndex (0);
            output.requestFocus ();
        } catch (java.io.IOException ex) {
        }

        return this;
    }

    //
    // Handlers
    //

    public void actionPerformed(final java.awt.event.ActionEvent ev) {
        if (ev.getSource () == next && exceptions.setNextElement() || ev.getSource () == previous && exceptions.setPreviousElement()) {
            current = (NbErrorManager.Exc)exceptions.get();
            update ();
            return;
        }

        if (ev.getSource () == details) {
            showDetails = !showDetails;
            update ();
            return;
        }

        if (ev.getSource () == DialogDescriptor.OK_OPTION) {
            exceptions.removeAll();
            //Fixed bug #9435, call of setVisible(false) replaced by call of dispose()
            //It did not work on Linux when JDialog is reused.
            //dialog.setVisible (false);
            dialog.dispose();
            return;
        }
    }

    //
    // Listeners
    //

    /**
     * Shows or hides this component depending on the value of parameter.
     * @param show if true, shows this component; otherwise, hides it.
     *
    public void setVisible(boolean visible) {
        if (visible == true) {
            // print details to log file
            if (descriptor != null) {
                PrintStream ps = TopLogging.getLogOutputStream ();
                Object detail = descriptor.getDetail();
                if (detail instanceof Throwable) {
                    if (detail instanceof InvocationTargetException) {
                        ps.println ("InvocationTargetException:");
                        ((InvocationTargetException)detail).getTargetException ().printStackTrace (ps);
                    } else {
                        ((Throwable)detail).printStackTrace(ps);
                    }
                } else {
                    ps.println(detail);
                }
            }
        }
        super.setVisible(visible);
}
    */

    protected static class ArrayListPos extends ArrayList {
        protected int position;

        protected ArrayListPos () {
            super();
            position=0;
        }

        protected boolean existsElement () {
            return size()>0;
        }

        protected boolean existsNextElement () {
            return position+1<size();
        }

        protected boolean existsPreviousElement () {
            return position>0&&size()>0;
        }

        protected boolean setNextElement () {
            if(!existsNextElement())
                return false;
            position++;
            return true;
        }

        protected boolean setPreviousElement () {
            if(!existsPreviousElement())
                return false;
            position--;
            return true;
        }

        protected Object get () {
            return existsElement()?get(position):null;
        }

        protected void removeAll () {
            clear();
            position=0;
        }
    }
}
