/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.datatransfer.Transferable;
import java.awt.Image;
import java.beans.*;
import java.util.*;
import java.io.IOException;
import java.text.MessageFormat;

import org.openide.nodes.*;
import org.openide.*;
import org.openide.cookies.InstanceCookie;
import org.openide.loaders.InstanceSupport;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.WeakListener;
import org.openide.util.actions.CookieAction;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.NewType;
import org.openide.util.datatransfer.PasteType;
import org.openide.util.io.NbMarshalledObject;

import org.netbeans.core.Services;

/** Node that represents the executors.
*
* @author Jaroslav Tulach
*/
public final class ServicesNode extends AbstractNode {
    /** default icons for SubLevel (Debugger Types, Executor Types, etc.) */
    private static final String ICON_SUB_LEVEL = "org/netbeans/core/resources/subLevel"; // NOI18N
    
    /** default instance of this node */
    private static Node INSTANCE;

//    private static final String ICON_TYPE_LEVEL = "org/netbeans/core/resources/typeLevel"; // NOI18N

    /** Creates new ServicesNode */
    private ServicesNode(Children ch) {
        super (ch);
        getCookieSet ().add (new InstanceSupport.Instance (
                                 Services.getDefault ()
                             ));
    }
    
    public HelpCtx getHelpCtx () {
        return new HelpCtx (ServicesNode.class);
    }

    /** Creates new service node.
    * @return the service node.
    */
    public static Node getDefault () {
        if (INSTANCE == null) {
            
            ServicesChildren sc = new ServicesChildren (
                Services.PROP_KINDS, ServicesChildren.TYPE, Services.kinds
            );
            Services.getDefault().addWeakListener (sc);

            INSTANCE = new ServicesNode (sc);
        }
        return INSTANCE;
    }
            
    /** Create default node for one class.
    */
    static Node createTypeNode (final Class clazz) throws InstantiationException {
        FilterList fl = new FilterList (Services.current) {
            protected boolean check (Object o) {
                return clazz.isInstance(o);
            }
        };

        Services.getDefault().addWeakListener (fl);
        
        ServicesChildren sc = new ServicesChildren (
            Services.PROP_SERVICE_TYPES, ServicesChildren.INSTANCE, fl);
        Services.getDefault().addWeakListener (sc);
      
        return new Type (clazz, sc);        
    }

    /** Create default node for one instance.
    */
    static Node createInstanceNode (ServiceType serv) throws IntrospectionException {
        return new Instance (serv);
    }

    /** Node that displayes informations about one class.
    */
    private static class Type extends AbstractNode {
        protected Class clazz;
        protected BeanInfo bi;

        public Type (Class clazz, ServicesChildren ch) throws InstantiationException {
            super (ch);

            this.clazz = clazz;

            // icon for types that do not define own
            setIconBase (ICON_SUB_LEVEL);

            getCookieSet ().add (ch.getIndex());

            getCookieSet ().add (getDummyInstance());
            
            try {
                bi = Introspector.getBeanInfo (clazz);
            } catch (IntrospectionException ex) {
                InstantiationException newEx = new InstantiationException (ex.getMessage ());
                TopManager.getDefault().getErrorManager ().copyAnnotation (newEx, ex);
                throw newEx;
            }

            setName (clazz.getName ());
            setDisplayName (bi.getBeanDescriptor ().getDisplayName ());
            setShortDescription (bi.getBeanDescriptor ().getShortDescription ());
        }

        /** Finds help associated with the variety class acc. to the BeanDescriptor. */
        public HelpCtx getHelpCtx () {
            return InstanceSupport.findHelp(getDummyInstance());
        }
        
        public Image getIcon (int type) {
            Image i = bi.getIcon (type);
            if (i == null) {
                i = super.getIcon (type);
            }
            return i;
        }

        public Image getOpenedIcon (int type) {
            return getIcon (type);
        }

        public boolean canCopy () { return false; }
        public boolean canCut () { return false; }
        public boolean canRename () { return false; }
        public boolean canDestroy () { return false; }


         /** Actions.
         */
         protected SystemAction[] createActions () {
             return new SystemAction[] {
                        SystemAction.get (org.openide.actions.ReorderAction.class),
                        null,
                        SystemAction.get (org.openide.actions.PasteAction.class),
                        null,
                        SystemAction.get (org.openide.actions.NewAction.class),
                        null,
                        SystemAction.get (org.openide.actions.ToolsAction.class),
                        SystemAction.get (org.openide.actions.PropertiesAction.class)
                    };
         }
 
        /** New types.
        */
        public NewType[] getNewTypes () {
            return Services.createNewTypes (clazz);
        }

        /** Paste types.
        */
        protected void createPasteTypes (Transferable t, List l) {
            super.createPasteTypes (t, l);

            final InstanceCookie ic = (InstanceCookie)NodeTransfer.cookie (
                                          t, NodeTransfer.COPY, InstanceCookie.class
                                      );
            try {
                if (ic != null && clazz.isAssignableFrom(ic.instanceClass ())) {
                    l.add (new PasteType () {
                               public String getName () {
                                   return Main.getString ("LAB_PasteExecutor", getDisplayName ());
                               }

                               public Transferable paste () throws java.io.IOException {
                                   try {
                                       ServiceType s = (ServiceType)ic.instanceCreate ();

                                       Services.getDefault ().addServiceType (s);
                                   } catch (Exception ex) {
                                       IOException io = new IOException (
                                         ex.getMessage ()
                                       );
                                       TopManager.getDefault ().getErrorManager ().copyAnnotation (io, ex);
                                       throw io;
                                   }
                                   return null;
                               }
                           });
                }
            } catch (ClassNotFoundException ex) {
                // consume exceptions
            } catch (java.io.IOException ex) {
                // consume exceptions
            }
        }
        
        private InstanceCookie getDummyInstance() {
            return new InstanceCookie () {
                public String instanceName () {
                    return clazz.getName ();
                }
                public Class instanceClass () {
                    return clazz;
                }
                public Object instanceCreate () {
                    return null;
                }
            };
        }
        
    }

    /** Node that represents one service type.
    */
    private static final class Instance extends BeanNode {
        public Instance (final ServiceType s) throws IntrospectionException {
            super (s);
        }

        ServiceType getService () {
            return (ServiceType) getBean ();
        }

        /** @return false
        */
        public boolean canCopy () {
            return true;
        }

        public boolean canCut () {
            return true;
        }

        public boolean canRename () {
            return true;
        }

        public boolean canDestroy () {
            return true;
        }

        public Transferable clipboardCut () throws java.io.IOException {
            destroy ();
            return clipboardCopy ();
        }

        /** Destroys the node.
        */
        public void destroy () {
            Services.getDefault ().removeServiceType (getService ());
        }

        /** Actions.
        */
        protected SystemAction[] createActions () {
            return new SystemAction[] {
                       SystemAction.get (org.openide.actions.MoveUpAction.class),
                       SystemAction.get (org.openide.actions.MoveDownAction.class),
                       null,
                       SystemAction.get (org.openide.actions.CustomizeBeanAction.class),
                       null,
                       SystemAction.get (org.openide.actions.CopyAction.class),
                       SystemAction.get (org.openide.actions.CutAction.class),
                       null,
                       SystemAction.get (org.openide.actions.DeleteAction.class),
                       null,
                       SystemAction.get (org.openide.actions.RenameAction.class),
                       null,
                       SystemAction.get (org.openide.actions.ToolsAction.class),
                       SystemAction.get (org.openide.actions.PropertiesAction.class)
                   };
        }

    }

}

/*
* Log
*  11   Gandalf   1.10        1/13/00  Jaroslav Tulach I18N
*  10   Gandalf   1.9         12/20/99 Jesse Glick     No more "default 
*       services", all are freely reorderable.
*  9    Gandalf   1.8         11/26/99 Patrik Knakal   
*  8    Gandalf   1.7         11/5/99  Jesse Glick     Context help jumbo patch.
*  7    Gandalf   1.6         11/5/99  Jaroslav Tulach WeakListener has now 
*       registration methods.
*  6    Gandalf   1.5         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  5    Gandalf   1.4         10/4/99  Jesse Glick     Make Default action on 
*       service types.
*  4    Gandalf   1.3         10/1/99  Jesse Glick     Cleanup of service type 
*       name presentation.
*  3    Gandalf   1.2         9/23/99  Jaroslav Tulach #3976
*  2    Gandalf   1.1         9/17/99  Jaroslav Tulach Reorder of nodes works.
*  1    Gandalf   1.0         9/10/99  Jaroslav Tulach 
* $
*/
