/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.net.URL;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;
import java.lang.ref.WeakReference;
import java.lang.ref.Reference;

import org.openide.TopManager;
import org.openide.util.Utilities;
import org.openide.util.NbBundle;

import org.netbeans.core.ui.ProductInformationPanel;

/** A class that encapsulates all the splash screen things.
*
* @author Ian Formanek, David Peroutka
*/
public final class Splash  implements SwingConstants {

    /** The splash image */
    static Reference splashRef;

    /** The about image */
    static Reference aboutRef;

    private static String getMainWindowTitle () {
        String buildNumber = System.getProperty ("netbeans.buildnumber"); // NOI18N
        return MessageFormat.format (NbBundle.getBundle ("org.netbeans.core.windows.Bundle", Locale.getDefault (), Splash.class.getClassLoader ()).getString ("CTL_MainWindow_Title"),
                                               new Object[] { buildNumber });
    }

    public static void showSplashDialog () {
        SplashDialog splashDialog = new SplashDialog ();
        splashDialog.print (getMainWindowTitle ());
        splashDialog.show ();
    }

    public static SplashOutput showSplash () {
        Window splashWindow;
        if (Utilities.isWindows () ||
            (Utilities.getOperatingSystem () == Utilities.OS_SOLARIS) ||
            (Utilities.getOperatingSystem () == Utilities.OS_OS2)
            || (Utilities.getOperatingSystem () == Utilities.OS_LINUX)
            ) {
            // only some systems supports non-frame windows
            splashWindow = new SplashWindow();
        } else {
            splashWindow = (Window)new SplashFrame();
        }

        //splashOutput = (SplashOutput)splashWindow;
        // show splash
        splashWindow.show ();
        splashWindow.toFront ();
        return (SplashOutput) splashWindow;
    }

    public static void hideSplash (SplashOutput xsplashWindow) {
        final Window splashWindow = (Window) xsplashWindow;
	javax.swing.SwingUtilities.invokeLater (new Runnable () {
		public void run () {
		    splashWindow.setVisible (false);
		    splashWindow.dispose ();
		}
	    });
    }

    static interface SplashOutput {
        public void print (String s);
    }

    /**
     * Standard way how to place the window to the center of the screen.
     */
    public static final void center(Window c) {
        c.pack();
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension dialogSize = c.getSize();
        c.setLocation((screenSize.width-dialogSize.width)/2,(screenSize.height-dialogSize.height)/2);
    }

    /** @return splash image */
    static Image getSplash() {
        Image ret;
        if ((splashRef == null) ||
                ((ret = (Image) splashRef.get()) == null)) {
            ret = loadSplash();
            splashRef = new WeakReference(ret);
        }
        return ret;
    }

    /** @return about image */
    static Image getAbout() {
        Image ret;
        if ((aboutRef == null) ||
                ((ret = (Image) aboutRef.get()) == null)) {
            ret = loadAbout();
            aboutRef = new WeakReference(ret);
        }
        return ret;
    }

    /** Loads a splash image from its source */
    private static Image loadSplash() {
        URL u = NbBundle.getLocalizedFile
            ("org.netbeans.core.resources.splash", // NOI18N
             "gif", // NOI18N
             Locale.getDefault(),
             Splash.class.getClassLoader()
             );
        return Toolkit.getDefaultToolkit().getImage(u);
    }

    /** Loads an about image from its source */
    private static Image loadAbout() {
        try {
            URL u = NbBundle.getLocalizedFile(
                        "org.netbeans.core.resources.about", // NOI18N
                        "gif", // NOI18N
                        Locale.getDefault(),
                        Splash.class.getClassLoader()
                    );

            return Toolkit.getDefaultToolkit().getImage(u);
        } catch (java.util.MissingResourceException exception) {
            return loadSplash();
        }
    }

    /**
     * This class implements double-buffered splash screen component.
     */
    static class SplashComponent extends JComponent {
        /** generated Serialized Version UID */
        static final long serialVersionUID = -1162806313274828742L;

        private final Font font = new Font("SansSerif", Font.PLAIN, 12); // NOI18N
        private FontMetrics fm;
        static Rectangle view;
        static Color color_text;

        Image image;
        private Rectangle dirty = new Rectangle();
        private String text;
        private Rectangle rect;

        /**
         * Creates a new splash screen component.
         */
        public SplashComponent() {
            image = new ImageIcon(getSplash()).getImage(); // load!
            
            ResourceBundle bundle = NbBundle.getBundle(
                "org.netbeans.core.Bundle", // NOI18N
                Locale.getDefault(),
                Splash.class.getClassLoader());
            try {
                StringTokenizer st = new StringTokenizer(
                    bundle.getString("SplashRunningTextBounds"), " ,"); // NOI18N
                view = new Rectangle(Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()));
                Integer rgb = Integer.decode(
                    bundle.getString("SplashRunningTextColor"));
                color_text = new Color(rgb.intValue());
            }
            catch (Throwable ex) {
                view = new Rectangle(9, 52, 380, 73);
                color_text = Color.white;
            }
        }

        /**
         * Defines the single line of text this component will display.
         */
        public void setText(String text) {
            this.text = text;
            if (text == null) {
                repaint(dirty);
                return;
            }
            // fm - first paint is called when the window is exposed - before setText
            if (fm == null) {
                return;
            }

            rect = new Rectangle();
            SwingUtilities.layoutCompoundLabel(fm, text, null,
                                               BOTTOM, LEFT, BOTTOM, LEFT,
                                               this.view, new Rectangle(), rect, 0);
            dirty = dirty.union(rect);
            // update screen (assume repaint manager optimizes unions;)
            repaint(dirty);
            this.dirty = rect;
        }

        /**
         * Override update to *not* erase the background before painting.
         */
        public void update(Graphics g) {
            paint(g);
        }

        /**
         * Renders this component to the given graphics.
         */
        public void paint(Graphics graphics) {
            /* JComponent paint is *always* double buffered
               The clip is also always set
            */
            // draw background
            graphics.setColor(color_text);
            graphics.setFont(font);
            graphics.drawImage(image, 0, 0, null);

            if (fm == null && text == null) {
                fm = graphics.getFontMetrics();
                return;
            }
            // draw text
            if (text != null) {
                if (fm == null)
                    fm = graphics.getFontMetrics();
                if (rect == null)
                    rect = new Rectangle();
                
                SwingUtilities.layoutCompoundLabel(fm, text, null,
                                                       BOTTOM, LEFT, BOTTOM, LEFT,
                                                       this.view, new Rectangle(), rect, 0);
                graphics.drawString(text, rect.x, rect.y + fm.getAscent());
            }
        }

        public Dimension getPreferredSize() {
            return new Dimension(image.getWidth (null), image.getHeight (null));
        }

        public boolean isOpaque () {
            return true;
        }
    }

    static class AboutComponent extends SplashComponent {
        
        public AboutComponent () {
            Integer rgb;
            ResourceBundle bundle = NbBundle.getBundle(
                "org.netbeans.core.Bundle", // NOI18N
                Locale.getDefault(),
                Splash.class.getClassLoader());
            Color colortext;
            StringTokenizer st;
            image = new ImageIcon(getAbout()).getImage(); // load!
            try {
                st = new StringTokenizer(
                    bundle.getString("AboutTextBounds"), " ,"); // NOI18N
                view = new Rectangle(Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()),
                                     Integer.parseInt(st.nextToken()));
                rgb = Integer.decode(
                    bundle.getString("AboutTextColor"));
                color_text = new Color(rgb.intValue());
            } catch(Throwable exc) {
                view = new Rectangle(78, 60, 362, 20);
                color_text = Color.white;
            }
        }
    }

    static class SplashWindow extends Window implements SplashOutput {
        /** generated Serialized Version UID */
        static final long serialVersionUID = 4838519880359397841L;

        private final SplashComponent splashComponent = new SplashComponent();

        /** Creates a new SplashWindow */
        public SplashWindow () {
            super(new Frame());
            // add splash component
            setLayout (new java.awt.BorderLayout ());
            add(splashComponent, java.awt.BorderLayout.CENTER);
            Splash.center(this);
        }

        public java.awt.Dimension getPreferredSize () {
            int width = 520;
            int height = 316;
            try {
                width = Integer.parseInt(NbBundle.getBundle("org.netbeans.core.Bundle",Locale.getDefault(),Splash.class.getClassLoader()).getString ("SPLASH_WIDTH"));
                height = Integer.parseInt(NbBundle.getBundle("org.netbeans.core.Bundle",Locale.getDefault(),Splash.class.getClassLoader()).getString ("SPLASH_HEIGHT"));
            } catch (Exception ex) {
                // silently ignore ???
            }
            return new java.awt.Dimension (width, height);
        }

        /**
         * Prints the given progress message on the splash screen.
         * @param x specifies a string that is to be displayed
         */
        public void print(String x) {
            splashComponent.setText(x);
        }
    }

    static class SplashFrame extends JFrame implements SplashOutput {
        /** generated Serialized Version UID */
        static final long serialVersionUID = 3200319077899134000L;

        private final SplashComponent splashComponent = new SplashComponent();

        /** Creates a new SplashFrame */
        public SplashFrame () {
            super (getMainWindowTitle ());
            setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
            setCursor(java.awt.Cursor.getPredefinedCursor(java.awt.Cursor.WAIT_CURSOR));
            // add splash component
            getContentPane().add(splashComponent);
            Splash.center(this);
        }

        /**
         * Prints the given progress message on the splash screen.
         * @param x specifies a string that is to be displayed
         */
        public void print(String x) {
            splashComponent.setText(x);
        }
    }

    static class SplashDialog extends JDialog implements SplashOutput {
        /** generated Serialized Version UID */
        static final long serialVersionUID = 5185644855500178404L;

        private final SplashComponent splashComponent = new AboutComponent();

        /** Creates a new SplashDialog */
        public SplashDialog () {
            super (TopManager.getDefault ().getWindowManager ().getMainWindow (), true);
            JPanel splashPanel = new JPanel();
            JTabbedPane tabbedPane = new JTabbedPane();
            setTitle (NbBundle.getBundle("org.netbeans.core.Bundle",Locale.getDefault(),Splash.class.getClassLoader()).
                      getString ("CTL_About_Title"));
            setResizable (false);
            // add splash component
            splashPanel.add(splashComponent, "Center");
            tabbedPane.addTab(NbBundle.getBundle(Splash.class).getString("CTL_About_Title"), splashPanel);
            tabbedPane.addTab(NbBundle.getBundle(Splash.class).getString("CTL_About_Detail"), new ProductInformationPanel());
            getContentPane().add(tabbedPane, "Center");  // NOI18N
            addKeyListener (new java.awt.event.KeyAdapter () {
                                public void keyPressed (java.awt.event.KeyEvent evt) {
                                    setVisible (false);
                                    dispose();
                                }
                            }
                           );
            Splash.center(this);
        }

        /**
         * Prints the given progress message on the splash screen.
         * @param x specifies a string that is to be displayed
         */
        public void print(String x) {
            splashComponent.setText(x);
        }
    }
}
