/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core;

import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.Locale;

import org.openide.util.NbBundle;

/** A class that provides logging facility for the IDE - once instantiated,
* it redirects the System.err into a log_file.
*
* @author   Ian Formanek, Ales Novak
*/
public class TopLogging extends Object {
    /** The name of the log file */
    public static final String LOG_FILE_NAME = "ide.log"; // NOI18N

    private static RandomAccessFile fileOutput;
    private static boolean disabledConsole = true;

    private PrintStream logPrintStream;
    private PrintStream consoleErrStream;
    private OutputStream demultiplex;
    
    /** Maximal size of log file.*/
    private static final long LOG_MAX_SIZE = 
    Long.getLong("org.netbeans.core.TopLogging.LOG_MAX_SIZE",1000000L).longValue();
    /** Number of old log files that are maintained.*/    
    private static final int LOG_COUNT = 
    Integer.getInteger("org.netbeans.core.TopLogging.LOG_COUNT",3).intValue();    
        

    /** Creates a new TopLogging - redirects the System.err to a log file.
    * @param logDir A directory for the log file
    */
    public TopLogging (String logDir) throws IOException  {
        File logFileDir = new File (logDir);
        if (! logFileDir.exists () && ! logFileDir.mkdirs ()) { // #8865
            throw new IOException ("Cannot make directory to contain log file"); // NOI18N
        }
        File logFile = createLogFile (logFileDir, LOG_FILE_NAME);
        if ((logFile.exists() && !logFile.canWrite()) || logFile.isDirectory()) {
            throw new IOException ("Cannot write to file"); // NOI18N
        }
        // output to file
        fileOutput = new RandomAccessFile(logFile, "rw"); //only write // NOI18N
        fileOutput.seek(fileOutput.length());

        // output to console
        consoleErrStream = System.err;

        DateFormat df = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL);
        java.util.Date date = new java.util.Date();

        logPrintStream = new PrintStream(demultiplex = new StreamDemultiplexor());
        disabledConsole = true;
        logPrintStream.println("-------------------------------------------------------------------------------"); // NOI18N
        logPrintStream.println(">Log Session: "+df.format (date)); // NOI18N
        logPrintStream.println(">System Info: "); // NOI18N
        printSystemInfo(logPrintStream);
        logPrintStream.println("-------------------------------------------------------------------------------"); // NOI18N
        disabledConsole = false;
        consoleErrStream = System.err;  // reacquire needed - cause of System.err has changed!!! see top.exec.ExecEng class
        System.setErr(logPrintStream);
    }

    /** This method limits size of log files. There is kept: actual log file   
     *  and old log files. This method prevents from growing log file infinitely.*/
    private static File createLogFile (File parent, String chld) {
        long firstModified = 0;
        File renameTo = null;
        File retFile = new File (parent, chld);
        
        if (!retFile.exists() || retFile.length() < LOG_MAX_SIZE)
            return retFile;
        
        for (int i = 1; i < LOG_COUNT;i++) {
            String logName = chld + "_"+i; // NOI18N
            File logFile = new File (parent, logName);
            
            if (!logFile.exists()) {
                renameTo = logFile;               
                break;
            }
            
            long logModif = logFile.lastModified();
            if ((firstModified == 0 || logModif < firstModified) &&  logModif > 0) {
                firstModified = logModif;
                renameTo = logFile;
            }            
        }

        if (renameTo != null) {
            if (renameTo.exists()) renameTo.delete();
            retFile.renameTo(renameTo);
        }
        
        return retFile;
    }
    
    /**
     * @return name of JIT (if installed). If Symantect JIT is available, also its version.
     **/
    private static String extractJITName() {
        // get as much information about JIT as possible
        String jit = System.getProperty ("java.compiler"); // is not supported e.g. on HP 1.1.5
        if ((jit == null) || ("".equals (jit))) // NOI18N
            return null;

        // specific for Symantec JIT (on Windows only)!
        if (jit.equals("symcjit")) { // NOI18N
            // allows to extract version info (variable JAVA_COMPCMD must be set FORCE_SIGNON,
            // any java app run and first line of result contains version info
            // (like: Symantec Java! JustInTime Compiler Version 3.00.039(x) for JDK 1.1.x)
            String fileSeparator = System.getProperty ("file.separator");
            String javaRoot = System.getProperty ("java.home") + fileSeparator;
            String cmd = javaRoot + "bin" + fileSeparator + "java nonexistent"; // NOI18N
            // run fake Java to get version info
            String[] envp = new String[1];
            envp[0] = "JAVA_COMPCMD=FORCE_SIGNON";  // magic env. setting, every Java app wil print JIT version first // NOI18N

            try {
                Process p = Runtime.getRuntime().exec(cmd, envp);
                BufferedReader brd = new BufferedReader(new InputStreamReader(p.getErrorStream()));
                p.waitFor();  // the output is short so this will not block (this is JDK bug)
                String firstLine = brd.readLine();
                brd.close();
                if (firstLine != null) {
                    // usual output is too long, try to shorten it if can
                    String stdStart = "Symantec Java! JustInTime Compiler Version "; // NOI18N
                    if (firstLine.startsWith(stdStart)) {
                        jit = "Symantec, version " + firstLine.substring(stdStart.length()); // NOI18N
                    } else {
                        jit = firstLine; // unknown message structure, let it be
                    }
                }
            } catch (IOException e) { // version can't be read
            } catch (InterruptedException e) { // dtto
            }
        }
        return jit;
    }

    public static void printSystemInfo(PrintStream ps) {
        final java.util.ResourceBundle topBundle = NbBundle.getBundle(TopLogging.class);
        String buildNumber = System.getProperty ("netbeans.buildnumber"); // NOI18N
        String currentVersion = new MessageFormat (org.openide.util.NbBundle.getBundle("org.netbeans.core.Bundle",java.util.Locale.getDefault(),TopLogging.class.getClassLoader()).getString ("currentVersion")).format (new Object[] { buildNumber });
        ps.println("  Product Version       = " + currentVersion); // NOI18N
        ps.println("  IDE Versioning        = " + System.getProperty ("org.openide.major.version") +
                   " spec=" + System.getProperty ("org.openide.specification.version") +
                   " impl=" + System.getProperty ("org.openide.version"));
        ps.println("  Operating System      = " + System.getProperty("os.name", "unknown")
                   + " version " + System.getProperty("os.version", "unknown")
                   + " running on " +  System.getProperty("os.arch", "unknown"));
        ps.println("  Java; VM; Vendor      = " + System.getProperty("java.version", "unknown") + "; " +
                   System.getProperty("java.vm.name", "unknown") + " " + System.getProperty("java.vm.version", "") + "; " +
                   System.getProperty("java.vendor", "unknown"));
        //ps.println("  Java Vendor URL          = " + System.getProperty("java.vendor.url", "unknown"));
        ps.println("  Java Home             = " + System.getProperty("java.home", "unknown"));
        //ps.println("  Java Class Version       = " + System.getProperty("java.class.version", "unknown"));
        ps.print  ("  System Locale         = " + Locale.getDefault()); // NOI18N
        String branding = NbBundle.getBranding ();
        if (branding == null)
            ps.println ();
        else
            ps.println (" (" + branding + ")"); // NOI18N
        String jit = extractJITName ();
        if (jit != null)
            ps.println("  JIT                   = " + jit); // NOI18N
        ps.println("  Home Dir; Current Dir = " + System.getProperty("user.home", "unknown") + "; " +
                   System.getProperty("user.dir", "unknown"));
        ps.println("  IDE Install; User Dir = " + Main.getHomeDir () + "; " + // NOI18N
                   Main.getUserDir ()); // NOI18N
        //ps.println("  System Directory         = " + Main.getSystemDir ()); // NOI18N
        ps.println("  CLASSPATH             = " + System.getProperty("java.class.path", "unknown")); // NOI18N
    }

    public void finalize() throws Throwable {
        demultiplex.flush();
        demultiplex.close();
    }

    static PrintStream getLogOutputStream() {
        return new PrintStream(new BufferedOutputStream(new LogOutputStream()));
    }

    class StreamDemultiplexor extends OutputStream {
        public void write(int b) throws IOException {
            fileOutput.write(b);
            if (! disabledConsole) consoleErrStream.write(b);
        }

        public void write(byte b[]) throws IOException {
            fileOutput.write(b);
            if (! disabledConsole) consoleErrStream.write(b);
        }

        public void write(byte b[],
                          int off,
                          int len)
        throws IOException {
            fileOutput.write(b, off, len);
            if (! disabledConsole) consoleErrStream.write(b, off, len);
        }

        public void flush() throws IOException {
            fileOutput.getFD().sync();
            consoleErrStream.flush();
        }

        public void close() throws IOException {
            fileOutput.close();
            consoleErrStream.close();
        }
    }

    private static class LogOutputStream extends OutputStream {
        public void write(int b) throws IOException {
            if (TopLogging.fileOutput != null)
                TopLogging.fileOutput.write(b);
        }

        public void write(byte b[]) throws IOException {
            if (TopLogging.fileOutput != null)
                TopLogging.fileOutput.write(b);
        }

        public void write(byte b[],
                          int off,
                          int len)
        throws IOException {
            if (TopLogging.fileOutput != null)
                TopLogging.fileOutput.write(b, off, len);
        }

        public void flush() throws IOException {
            if (TopLogging.fileOutput != null) {
                // TopLogging.fileOutput.getFD().sync();
            }
        }

        public void close() throws IOException {
            if (TopLogging.fileOutput != null)
                TopLogging.fileOutput.close();
        }
    }
}
