/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.*;

import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.awt.JMenuPlus;
import org.openide.awt.Actions;
import org.openide.windows.*;
import org.openide.util.actions.*;

import org.netbeans.core.windows.*;
import org.netbeans.core.windows.frames.*;
import org.netbeans.core.windows.util.WindowUtils;

/** Attaches frame to some side of desktop pane.
 * Meaningful only in MDI, disabled in SDI.
 *
 * @author Jano Rojcek, Dafe Simonek
 */
public final class AttachFrameAction extends CallableSystemAction implements Presenter.Menu {
    
    /** asociation with window manager implementation */
    private static WindowManagerImpl wm =
        (WindowManagerImpl)TopManager.getDefault().getWindowManager();

    /* Constructs a new action */
    public AttachFrameAction() {
    }

    /* Returns resource string pointing to actions' icon.
    * @return resource string of actions' icon
    */
    protected String iconResource () {
        return "/org/netbeans/core/resources/actions/attachFrame.gif"; // NOI18N
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx (AttachFrameAction.class);
    }

    /* @return Returns localized name of this action */
    public String getName() {
        return getBundleString("CTL_AttachFrameAction"); //NOI18N
    }

    /* Returns a submneu that will present this action in a Menu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getMenuPresenter () {
        JMenu mainItem = new JMenuPlus();
        Actions.connect(mainItem, this, false);
//        HelpCtx.setHelpIDString (mainItem, ConvertAction.class.getName ());
        mainItem.addMenuListener(getDesktopTypeListener());
        return mainItem;
    }
    
    /** This action is active only in MDI mode */
    public boolean isEnabled () {
        return WindowUtils.isMdi() && 
               (WindowUtils.getActiveMode() != null); 
    }

    /* Returns a submneu that will present this action in a PopupMenu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getPopupPresenter() {
        JMenu mainItem = new JMenuPlus();
        Actions.connect(mainItem, this, true);
//        HelpCtx.setHelpIDString (mainItem, ConvertAction.class.getName ());
        mainItem.addMenuListener(getDesktopTypeListener());
        return mainItem;
    }

    /* Performs nothing
    */
    public void performAction () {
        // all functionality is accomplished by menu listeners
    }
    
    /** @return returns string from bundle for given string pattern */
    private static final String getBundleString (String bundleStr) {
        return NbBundle.getBundle(AttachFrameAction.class).getString(bundleStr);
    }

    /** Lazy init of mapping between constraints and names for menu items
     * @return constraint-menu name mapping
     */
    private static final Map getConstrMap () {
        HashMap constrMap = new HashMap(8);
        constrMap.put(PerimeterLayout.NORTH, "CTL_Split_Top");
        constrMap.put(PerimeterLayout.SOUTH, "CTL_Split_Bottom");
        constrMap.put(PerimeterLayout.WEST, "CTL_Split_Left");
        constrMap.put(PerimeterLayout.EAST, "CTL_Split_Right");
        return constrMap;
    }
    
    private static final DesktopTypeListener getDesktopTypeListener () {
        return new DesktopTypeListener();
    }

    // innerclasses .......................................................
    
    /** Listens to main menu item, constructs submenu when main item selected.
     * Submenu contains choices for converting into constrained (desktop) frame,
     * choices that may not be performed at this time are greyed out
     */
    private static final class DesktopTypeListener implements MenuListener, ActionListener, Runnable {

        /** action command to run */
        private ActionEvent evt;
            
        public void menuCanceled(MenuEvent evt) {
        }
        
        /** Constructs submenu from mask.
         */
        public void menuSelected(MenuEvent evt) {
            WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
            JMenu menu = (JMenu)evt.getSource();
            JMenuItem mi = null;
            Map.Entry curEntry = null;
            for (Iterator iter = getConstrMap().entrySet().iterator(); iter.hasNext(); ) {
                curEntry = (Map.Entry)iter.next();
                mi = new JMenuItem(getBundleString((String)curEntry.getValue()));
                mi.setActionCommand((String)curEntry.getKey());
                mi.addActionListener(this);
                mi.setEnabled(
                    null == WindowUtils.findConstrainedMode(
                        curWorkspace, curEntry.getKey(), true
                    )
                );
                menu.add(mi);
            }
        }
        
        public void menuDeselected(MenuEvent evt) {
            ((JMenu)evt.getSource()).removeAll();
        }
        
        /** Performs converting of mode into desktop frame type
         */
        public synchronized void actionPerformed(ActionEvent evt) {
            this.evt = evt;
            if (SwingUtilities.isEventDispatchThread()) {
                run();
            } else {
                SwingUtilities.invokeLater(this);
            }
        }
        
        /** Implementation of runnable interface, runs the action command */
        public synchronized void run() {
            if (evt != null) {
                WindowUtils.changeActiveMode(
                    evt.getActionCommand(),
                    WindowTypesManager.DESKTOP_FRAME
                );
                evt = null;
            }
        }
            
    } // end of DesktopTypeListener inner class

}
