/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import java.io.Serializable;
import java.awt.MenuShortcut;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.MenuListener;
import java.util.List;
import java.util.Set;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.event.*;
import javax.swing.JMenuItem;
import javax.swing.JMenu;

import org.netbeans.core.windows.ModeImpl;
import org.netbeans.core.windows.WorkspaceImpl;
import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.windows.frames.TopComponentContainer;

import org.openide.TopManager;
import org.openide.windows.Mode;
import org.openide.windows.TopComponent;
import org.openide.windows.Workspace;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.*;
import org.openide.awt.*;

/** Dock the top component to the mode.
*
* @author jrojcek
*/
public final class DockingAction extends CallableSystemAction
    implements Presenter.Menu {
    
    /* constant special items in docking menu */
    static final String SINGLE = "CTL_NewSingle"; //NOI18N
    static final String MULTI_TAB = "CTL_NewMultiTab"; //NOI18N
    static final String SPLIT = "CTL_NewSplit"; //NOI18N
    
    /** asociation with window manager implementation */
    static WindowManagerImpl wm =
        (WindowManagerImpl)TopManager.getDefault().getWindowManager();

    /* Constructs a new docking action */
    static final long serialVersionUID =8679037289261745865L;
    public DockingAction() {
    }

    /* Returns resource string pointing to actions' icon.
    * @return resource string of actions' icon
    */
    protected String iconResource () {
        return "/org/netbeans/core/resources/actions/dock.gif"; // NOI18N
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx (DockingAction.class);
    }

    /* @return Returns localized name of this action */
    public String getName() {
        return NbBundle.getBundle(DockingAction.class).getString("Docking"); //NOI18N
    }

    /* Returns a submneu that will present this action in a Menu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getMenuPresenter () {
      JMenu mainItem = new JMenuPlus();
        Actions.setMenuText(mainItem, getName(), true); 
        mainItem.setIcon (SystemAction.get (DockingAction.class).getIcon ());
        HelpCtx.setHelpIDString (mainItem, DockingAction.class.getName ());
        mainItem.addMenuListener(new MainItemListener());
        return mainItem;
    }

    /* Returns a submneu that will present this action in a PopupMenu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getPopupPresenter() {
      JMenu mainItem = new JMenuPlus();
        Actions.setMenuText(mainItem, getName(), true); 
        HelpCtx.setHelpIDString (mainItem, DockingAction.class.getName ());
        mainItem.addMenuListener(new MainItemListener());
        return mainItem;
    }

    /* Actually performs the "docking" action by setting the mode
    * of how component will be displayed
    */
    public void performAction () {
        // all functionality is accomplished by menu listeners
    }

    /** @return returns string from bundle for given string pattern */
    private static final String getBundleString (String bundleStr) {
        return NbBundle.getBundle(DockingAction.class).getString(bundleStr);
    }

    // innerclasses .......................................................
    
    /** Listens to selecting of main item and expands it to the
     * submenu of exiting and new modes
     */
    private static final class MainItemListener implements MenuListener {

        /** Source of the events */
        private JMenu menu;
        /** listener for new mode creation action */
        private NewModeActionListener newModeActionListener;
        /** Counter of the menu items (because menu.getComponent() returns always zero)*/
        private int count = 0;
        
        public void menuCanceled (MenuEvent e) {
        }

        public void menuDeselected (MenuEvent e) {
            JMenu menu = (JMenu)e.getSource();
            menu.removeAll();
        }

        public void menuSelected (MenuEvent e) {
            count = 0;
            TopComponent comp = TopComponent.getRegistry ().getActivated ();
            if (comp == null) 
                return;
            this.menu = (JMenu)e.getSource();
            Workspace curWorkspace =
                TopManager.getDefault().getWindowManager().getCurrentWorkspace();
            Set modes = curWorkspace.getModes();
            int count = modes.size();
            // obtain mode active top component is in
            Mode compMode = curWorkspace.findMode(comp);
            // create mode list
            ArrayList userDefined = new ArrayList(count);
            Mode curMode;
            for (Iterator iter = modes.iterator(); iter.hasNext(); ) {
                curMode = (Mode)iter.next();
                if (curMode.canDock(comp) && ((!curMode.equals(compMode)) ||
                    // enable split TCC even if TC is inside this TCC
                    (((ModeImpl)compMode).getContainerType().equals(
                    WindowTypesManager.SPLIT_CONTAINER)))) {
                    if (((ModeImpl)curMode).isUserDefined()) {
                        // exclude user defined orphans
                        if (!((ModeImpl)curMode).isOrphan())
                            userDefined.add(curMode);
                    } else {
                        menu.add(createMenuItem(curMode));
                    }
                }
            }
            if (userDefined.size() > 0)
                menu.addSeparator();
            for (Iterator iter = userDefined.iterator(); iter.hasNext(); ) {
                curMode = (Mode)iter.next();
                menu.add(createMenuItem(curMode));
            }
            menu.addSeparator();
            if (!((ModeImpl)compMode).isSingle()) {
                // allow docking to single only if in multi
                menu.add(createNewModeMenuItem(SINGLE));
            }
            // new multi mode
            menu.add(createNewModeMenuItem(MULTI_TAB));
            // new split mode
            menu.add(createNewModeMenuItem(SPLIT));
        }
        
        private String getLabel(Mode mode) {
            // create menu display name in format: 'order_number name_of_mode'
            // where order number works as mnemonic
            StringBuffer buffer = new StringBuffer("&");
            //buffer.append(menu.getComponentCount() + 1);
            buffer.append(++count);
            buffer.append(' ');
            buffer.append(mode.getDisplayName());
            return buffer.toString();
        }

        /**Creates menu item for existing mode. If mode has split TCC then use menu
         * instead of menu item. Adds listener to menu item.
         */
        private JMenuItem createMenuItem(Mode mode) {
            if (((ModeImpl)mode).getContainerType().equals(
                WindowTypesManager.SPLIT_CONTAINER)) {
                JMenu curMenu = new JMenuPlus();
                Actions.setMenuText(curMenu, getLabel(mode), false); 
                curMenu.addMenuListener(new ModeItemListener(mode));
                return curMenu;
            }
            else {
                JMenuItem curMenuItem = new JMenuItem();
                Actions.setMenuText(curMenuItem, getLabel(mode), true); 
                curMenuItem.addActionListener(new ModeItemListener(mode));
                return curMenuItem;
            }
        }
        
        /**Creates menu item for new (not yet created) mode. Adds listener to it.
         *
         */
        private JMenuItem createNewModeMenuItem(String type) {
            JMenuItem curMenuItem;
            curMenuItem = new JMenuItem();
            Actions.setMenuText(curMenuItem, NbBundle
                .getBundle(DockingAction.class).getString(type), false); 
            curMenuItem.setActionCommand(type);
            curMenuItem.addActionListener(newModeActionListener());
            return curMenuItem;
        }

        /** accessor for new modes menu item listener */
        private NewModeActionListener newModeActionListener () {
            if (newModeActionListener == null) {
                newModeActionListener = new NewModeActionListener();
            }
            return newModeActionListener;
        }

    } // end of MainItemListener inner class

    
    /** Listens to menu choice for docking into new modes 
     */
    private static final class NewModeActionListener implements ActionListener {

        public void actionPerformed(ActionEvent evt) {
            // special items
            WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
            TopComponent comp = TopComponent.getRegistry ().getActivated ();
            // dock into specially created modes
            Mode mode = null;
            String modeName = null;
            String type = evt.getActionCommand();
            if (type.equals(SINGLE)) {
                modeName = wm.findUnusedModeName(comp.getName(),
                                                 curWorkspace);
                mode = curWorkspace.createMode(
                           modeName, modeName, null,
                           true
                       );
                ((ModeImpl)mode).setContainerType(WindowTypesManager.MULTI_TABBED_CONTAINER);
            }
            else if (type.equals(MULTI_TAB)) { 
                modeName = wm.findUnusedModeName(getBundleString("CTL_NewMultiTabName"),
                                                 curWorkspace, true);
                mode = curWorkspace.createMode(
                           modeName, modeName, null, // PENDING - what to use ???
                           true
                       );
                ((ModeImpl)mode).setContainerType(WindowTypesManager.MULTI_TABBED_CONTAINER);
            }
            else if (type.equals(SPLIT)) {
                modeName = wm.findUnusedModeName(getBundleString("CTL_NewSplitName"),
                                                 curWorkspace, true);
                mode = curWorkspace.createMode(
                           modeName, modeName, null, // PENDING - what to use ???
                           true
                       );
                ((ModeImpl)mode).setContainerType(WindowTypesManager.SPLIT_CONTAINER);
            }
            mode.dockInto(comp);
        }        

    }

    /** Listens to menu choice for docking into existing split or multitabbed modes 
     */
    private static final class ModeItemListener implements
        MenuListener, ActionListener {
            
        private Mode mode;
        
        public ModeItemListener(Mode mode) {
            this.mode = mode;
        }
        
        public void menuCanceled(MenuEvent evt) {
        }
        
        public void menuSelected(MenuEvent evt) {
            JMenu menu = (JMenu)evt.getSource();
            JMenuItem mi = new JMenuItem(getBundleString("CTL_Split_Center"));
            mi.setActionCommand(PerimeterLayout.CENTER);
            mi.addActionListener(this);
            menu.add(mi);
            mi = new JMenuItem(getBundleString("CTL_Split_Right"));
            mi.setActionCommand(PerimeterLayout.EAST);
            mi.addActionListener(this);
            menu.add(mi);
            mi = new JMenuItem(getBundleString("CTL_Split_Left"));
            mi.setActionCommand(PerimeterLayout.WEST);
            mi.addActionListener(this);
            menu.add(mi);
            mi = new JMenuItem(getBundleString("CTL_Split_Top"));
            mi.setActionCommand(PerimeterLayout.NORTH);
            mi.addActionListener(this);
            menu.add(mi);
            mi = new JMenuItem(getBundleString("CTL_Split_Bottom"));
            mi.setActionCommand(PerimeterLayout.SOUTH);
            mi.addActionListener(this);
            menu.add(mi);
        }
        
        public void menuDeselected(MenuEvent evt) {
            ((JMenu)evt.getSource()).removeAll();
        }
        
        public void actionPerformed(ActionEvent evt) {
            TopComponent comp = TopComponent.getRegistry ().getActivated ();
            if (evt.getActionCommand() == null)
                mode.dockInto(comp);
            else {
                // try to remove TC from container
                TopComponentContainer tcc = ((ModeImpl)mode).getContainerInstance();
                if (tcc != null)
                    tcc.removeTopComponent(comp);
                ((ModeImpl)mode).dockInto(comp, evt.getActionCommand());
            }
            comp.requestFocus();
        }
    }

}
