/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.*;

import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.awt.JMenuPlus;
import org.openide.awt.Actions;
import org.openide.windows.*;
import org.openide.util.actions.*;

import org.netbeans.core.windows.*;
import org.netbeans.core.windows.frames.*;
import org.netbeans.core.windows.util.WindowUtils;

/** Action can either pull the frame out of the IDE desktop or to put it
 * back from standalone frame to frame inside IDE desktop.
 * Meaningful only in MDI, disabled in SDI.
 *
 * @author Dafe Simonek
 */
public final class FrameResidenceAction extends CallableSystemAction implements Presenter.Menu {
    
    private static final String FRAME_INSIDE = "inside"; // NOI18N
    private static final String FRAME_OUTSIDE = "outside"; // NOI18N
    
    /* Constructs a new action */
    public FrameResidenceAction () {
    }

    /* Returns resource string pointing to actions' icon.
    * @return resource string of actions' icon
    */
    protected String iconResource () {
        return "/org/netbeans/core/resources/actions/frameResides.gif"; // NOI18N
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx (AttachFrameAction.class);
    }

    /* @return Returns localized name of this action */
    public String getName() {
        return getBundleString("CTL_FrameResidenceAction"); //NOI18N
    }

    /* Returns a submneu that will present this action in a Menu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getMenuPresenter () {
        return createMenuPresenter(false);
    }
    
    /* Returns a submneu that will present this action in a PopupMenu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getPopupPresenter() {
        return createMenuPresenter(true);
    }
    
    /** This action is active only in MDI mode for desktop-type frames */
    public boolean isEnabled () {
        return WindowUtils.isMdi() && (WindowUtils.getActiveMode() != null);
    }

    /* Detaches frame containing active top component and makes it floating
    */
    public void performAction () {
        WindowUtils.changeActiveMode(null, WindowTypesManager.INTERNAL_FRAME);
    }
    
    private JMenuItem createMenuPresenter (boolean popup) {
        JMenu mainItem = new JMenuPlus();
        Actions.connect(mainItem, this, popup);
//        HelpCtx.setHelpIDString (mainItem, DetachFrameAction.class.getName ());
        mainItem.addMenuListener(new ResidenceTypeListener());
        return mainItem;
    }
    
    /** @return returns string from bundle for given string pattern */
    private static final String getBundleString (String bundleStr) {
        return NbBundle.getBundle(DetachFrameAction.class).getString(bundleStr);
    }
    
    // innerclasses .......................................................
    
    /** Listens to main menu item, constructs submenu when main item selected.
     * Submenu contains radio-style choices for changing residence of
     * active frame, either inside IDE desktop or outside.
     */
    private static final class ResidenceTypeListener 
                         implements MenuListener, ActionListener, Runnable {
            
        /** action command to run */
        private ActionEvent evt;
            
        public void menuCanceled(MenuEvent evt) {
        }
        
        /** Constructs submenu from mask.
         */
        public void menuSelected(MenuEvent evt) {
            String activeFrameType = WindowUtils.getActiveMode().getFrameType();
            boolean isInside = 
                WindowTypesManager.DESKTOP_FRAME.equals(activeFrameType) ||
                WindowTypesManager.INTERNAL_FRAME.equals(activeFrameType);
            JMenu menu = (JMenu)evt.getSource();
            JRadioButtonMenuItem mi = new JRadioButtonMenuItem(getBundleString("CTL_FrameInside"));
            mi.setActionCommand(FRAME_INSIDE);
            mi.addActionListener(this);
            mi.setSelected(isInside);
            menu.add(mi);
            mi = new JRadioButtonMenuItem(getBundleString("CTL_FrameOutside"));
            mi.setActionCommand(FRAME_OUTSIDE);
            mi.addActionListener(this);
            mi.setSelected(!isInside);
            menu.add(mi);
        }
        
        public void menuDeselected(MenuEvent evt) {
            ((JMenu)evt.getSource()).removeAll();
        }
        
        /** Performs converting of mode into desired frame type
         */
        public synchronized void actionPerformed(ActionEvent evt) {
            this.evt = evt;
            if (SwingUtilities.isEventDispatchThread()) {
                run();
            } else {
                SwingUtilities.invokeLater(this);
            }
        }

        /** Implementation of runnable interface, runs the action command */
        public synchronized void run () {
            if (evt == null) {
                return;
            }
            String resultType = null;
            ModeImpl activeMode = WindowUtils.getActiveMode();
            if (FRAME_INSIDE.equals(evt.getActionCommand())) {
                Object constr = activeMode.getConstraints();
                resultType = ((constr == null) || PerimeterLayout.CENTER.equals(constr))
                                ? WindowTypesManager.INTERNAL_FRAME
                                : WindowTypesManager.DESKTOP_FRAME;
            } else {
                resultType = WindowTypesManager.TOP_FRAME;
            }
            activeMode.setFrameType(resultType);
            evt = null;
        }
        
    } // end of DesktopTypeListener inner class

    
}
