/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import java.awt.event.ActionEvent;
import java.util.*;
import javax.swing.JMenuItem;
import javax.swing.event.*;

import org.openide.TopManager;
import org.openide.awt.Actions;
import org.openide.util.*;
import org.openide.util.actions.*;

import org.netbeans.core.Help;

/** Shows a list of home help pages for modules.
*
* @author Jesse Glick
*/
public class ModuleHelpAction extends SystemAction implements Presenter.Menu, Presenter.Popup, Actions.SubMenuModel {
    /** Implementation of ActSubMenuInt */
    //private ActSubMenuModel model;

    private static final long serialVersionUID =2022674936562918639L;

    /** Do nothing.
    */
    public void actionPerformed (ActionEvent e) {
    }

    /** URL to this action.
    * @return URL to the action icon
    */
    public String iconResource () {
        return "/org/netbeans/core/resources/actions/moduleHelp.gif"; // NOI18N
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx (ModuleHelpAction.class);
    }

    public String getName() {
        return NbBundle.getBundle (ModuleHelpAction.class).getString("ModuleHelp");
    }

    protected boolean showingMaster () {
        return true;
    }

    /** Alternate action displaying only the master index (no submenu).
     */
    public static class MasterOnly extends CallableSystemAction implements ChangeListener {

        private static String masterID;

        protected void initialize () {
            super.initialize ();
            refresh ();
            Help.getDefault ().addChangeListener (WeakListener.change (this, Help.getDefault ()));
        }

        private void refresh () {
            masterID = Help.getDefault ().getMasterID (false);
            setEnabled (masterID != null);
        }

        public void stateChanged (ChangeEvent ev) {
            refresh ();
        }

        public String getName () {
            return NbBundle.getBundle (ModuleHelpAction.class).getString("ModuleHelp.MasterOnly");
        }

        public String iconResource () {
            return "/org/netbeans/core/resources/actions/moduleHelp.gif"; // NOI18N
        }

        public HelpCtx getHelpCtx () {
            return new HelpCtx (ModuleHelpAction.class);
        }

        public void performAction () {
            TopManager.getDefault ().showHelp (new HelpCtx (masterID));
        }

    }

    /** Alternate action displaying only the home IDs (no master index).
     */
    public static class ShortcutsOnly extends ModuleHelpAction {

        protected boolean showingMaster () {
            return false;
        }

        public String getName () {
            return NbBundle.getBundle (ModuleHelpAction.class).getString("ModuleHelp.ShortcutsOnly");
        }

        /** Want to disable when there is only one helpset, since the master should be used instead.
         */
        public int getCount () {
            int i = super.getCount ();
            return (i == 1 ? 0 : i);
        }

    }

    /* Returns a submenu that will presents this action in menu bar
    * @return the JMenuItem representation for this action (which is submenu)
    */
    public JMenuItem getMenuPresenter() {
        return new SpecialSubMenu(false);
    }

    /* Returns a submneu that will present this action in a PopupMenu.
    * @return the JMenuItem representation for this action
    */
    public JMenuItem getPopupPresenter() {
        return new SpecialSubMenu(true);
    }

    /** Convenience method for obtaining submenu model * /
    ActSubMenuModel getModel () {
        if (model == null)
            model = new ActSubMenuModel();
        return model;
    } */

    /** SubMenu of this action - notifies data model about
    * ancestor changes (addNotify, removeNotify)
    */
    private final class SpecialSubMenu extends Actions.SubMenu {

        static final long serialVersionUID =-4734612283802319986L;
        SpecialSubMenu (boolean popup) {
            super(ModuleHelpAction.this, ModuleHelpAction.this/*.getModel ()*/, popup);
        }

        /** Notifies data model in adition to normal behaviour */
        public void addNotify () {
            ModuleHelpAction.this/*.getModel ()*/.addNotify();
            super.addNotify();
        }

        /** Notifies data model in adition to normal behaviour */
        public void removeNotify () {
            ModuleHelpAction.this/*.getModel ()*/.removeNotify();
            super.removeNotify();
        }

    } // end of SpecialSubMenu

    /** Implementation of SubMenuModel.
    */
    //private final class ActSubMenuModel implements Actions.SubMenuModel {

        private List items; // List<{String homeID, String display}>

        public int getCount() {
            //System.err.println ("getCount");
            return items.size ();
        }

        public String getLabel (int index) {
            //System.err.println ("getLabel " + index);
            return ((String[]) items.get (index))[1];
        }

        public HelpCtx getHelpCtx (int index) {
            String homeID = ((String[]) items.get (index))[0];
            return homeID != null ? new HelpCtx (homeID) : ModuleHelpAction.this.getHelpCtx ();
        }

        public void performActionAt (int index) {
            //System.err.println ("performActionAt " + index);
            String homeID = ((String[]) items.get (index))[0];
            if (homeID != null)
                TopManager.getDefault ().showHelp (new HelpCtx (homeID));
        }

        public void addChangeListener (ChangeListener l) {
        }

        public void removeChangeListener (ChangeListener l) {
        }

        public void addNotify () {
            super.addNotify ();
            //System.err.println ("addNotify called");
            Help.Impl help = Help.getDefault ();
            Map homesByDisplay = help.getHomesByDisplay ();
            List dists = new ArrayList ();
            List regs = new ArrayList ();
            Iterator it = homesByDisplay.entrySet ().iterator ();
            while (it.hasNext ()) {
                Map.Entry entry = (Map.Entry) it.next ();
                String display = (String) entry.getKey ();
                String home = (String) entry.getValue ();
                boolean dist = help.isDistinguished (home);
                (dist ? dists : regs).add (new String[] { home, display });
                //System.err.println("Help set: display=" + display + "; home=" + home + "; dist=" + dist);
            }
            Comparator c = new Comparator () {
                               public int compare (Object o1, Object o2) {
                                   return ((String[]) o1)[1].compareTo (((String[]) o2)[1]);
                               }
                           };
            Collections.sort (dists, c);
            Collections.sort (regs, c);
            items = new ArrayList (dists);
            if (dists.size () > 0 && regs.size () > 0)
                items.add (new String[] { null, null });
            items.addAll (regs);
            /*
            for (int i = 0; i < displayNames.size (); i++)
              System.err.println ("Pair: " + displayNames.get (i) + " " + homeIDs.get (i));
            */
            if (showingMaster ()) {
                String id = help.getMasterID (true);
                if (id != null) {
                    // Add a Master Index item if possible. Might be the only one,
                    // if there are helpsets but no real home IDs.
                    if (! items.isEmpty ())
                        items.add (new String[] { null, null });
                    String masterName = help.getMasterDisplayName ();
                    if (masterName == null)
                        masterName = ModuleHelpAction.this.getName ();
                    items.add (new String[] { id, masterName });
                }
            }
        }

        public void removeNotify () {
            super.removeNotify ();
            /* Not safe to implement--called at inappropriate times.
            //System.err.println ("removeNotify called");
            displayNames = null;
            homeIDs = null;
            */
        }

    //} // end of ActSubMenuModel

}
