/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import java.awt.event.*;
import java.io.IOException;
import java.util.*;

import javax.swing.*;

import org.openide.*;
import org.openide.explorer.propertysheet.PropertyPanel;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.*;
import org.openide.util.actions.*;

/** Presents a dialog allowing the user to mount a new file system
 * (of one of the known types).
 * @author David Strupl
 */
public class NewFileSystemAction extends CallableSystemAction {

    public NewFileSystemAction() {
        setEnabled(true);
    }
    
    /* Human presentable name of the action. This should be
    * presented as an item in a menu.
    * @return the name of the action
    */
    public String getName() {
        return getString("NewFileSystem");
    }

    /* Help context where to find more about the action.
    * @return the help context for this action
    */
    public HelpCtx getHelpCtx() {
        return new HelpCtx (NewFileSystemAction.class);
    }

    /* Resource name for the icon.
    * @return resource name
    */
    protected String iconResource () {
        return "/org/netbeans/core/resources/actions/addDirectory.gif"; // NOI18N
    }
    
    /** Property for editing in a PropertyPanel. */
    private FileSystem fs;
    /** setter - Property for editing in a PropertyPanel. */
    public void setFileSystem(FileSystem f) {
        fs = f;
    }
    /** getter - Property for editing in a PropertyPanel. */
    public FileSystem getFileSystem() {
        return fs;
    }
    /** This method is called by one of the "invokers" as a result of
     * some user's action that should lead to actual "performing" of the action.
     * This default implementation calls the assigned actionPerformer if it
     * is not null otherwise the action is ignored.
     */
    public void performAction () {
        fs = null;
        PropertyPanel panel = new PropertyPanel(this, "fileSystem", PropertyPanel.PREF_CUSTOM_EDITOR); // NOI18N
        final Object [] options = new Object [] {
            DialogDescriptor.OK_OPTION,
            DialogDescriptor.CANCEL_OPTION,
        };
        DialogDescriptor descriptor = new DialogDescriptor (
            panel,
            getString("CTL_Mount_Dialog_Title"),
            true,
            options,
            options[0],
            DialogDescriptor.BOTTOM_ALIGN,
            new HelpCtx(NewFileSystemAction.class), // HelpCtx
            null // Action Listener
        );
        TopManager.getDefault().createDialog(descriptor).show();
        if (descriptor.getValue() == DialogDescriptor.OK_OPTION) {
            // first update the panel's value in case it was edited by 
            // EnhancedCustomPropertyEditor
            panel.updateValue();
            // get the value
            Object value = null;
            try {
                value = panel.getModel().getValue();
            } catch (java.lang.reflect.InvocationTargetException x) {
                TopManager.getDefault().notifyException(x);
            }
            // the value should be FileSystem
            if ((value != null) && (value instanceof FileSystem)) {
                Repository r = TopManager.getDefault().getRepository();
                FileSystem filesystem;
                
                if ((filesystem = r.findFileSystem(((FileSystem)value).getSystemName ())) == null) {
                    r.addFileSystem((FileSystem)value);                        
                }
                else {
                    if(filesystem.isHidden()) {
                        NotifyDescriptor description = new NotifyDescriptor.Confirmation(
                                        getString("MSG_LocalFSAlreadyMountedHidden"),
                                        getString("CTL_LocalFSAlreadyMountedHidden_Title"),
                                        NotifyDescriptor.YES_NO_CANCEL_OPTION
                                    );
                        if(NotifyDescriptor.YES_OPTION.equals(TopManager.getDefault().notify(description))) {
                            filesystem.setHidden(false);
                        }
                    } else {
                        TopManager.getDefault().notify(
                                new NotifyDescriptor.Message(getString("MSG_LocalFSAlreadyMounted"),
                                                             NotifyDescriptor.ERROR_MESSAGE));
                    }
                }
            } else
                TopManager.getDefault().notify(
                            new NotifyDescriptor.Message(getString("MSG_LocalFSNoMounted"),
                                                         NotifyDescriptor.ERROR_MESSAGE));
        }
    }

    private static String getString(String key) {
        return org.openide.util.NbBundle.getBundle(NewFileSystemAction.class).getString(key);
    }
}
