/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.actions;

import java.io.IOException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import org.openide.DialogDescriptor;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.cookies.ArgumentsCookie;
import org.openide.execution.ExecInfo;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.CookieAction;
import org.openide.nodes.Node;

/** SetArguments action. Is enabled if the activated node implements ArgumentsCookie.
*
* @see ArgumentsCookie
*
* This class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Ian Formanek
*/
public final class SetArgumentsAction extends CookieAction {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 6015899396428466799L;

    /** Manages enable / disable logic of this action */
    protected boolean enable (Node[] activatedNodes) {
        if (!super.enable (activatedNodes)) return false;

        // This action is enabled if the ArgumentsCookie returns true from getArgumentsSupported ()
        ArgumentsCookie rightCookie =
            (ArgumentsCookie)(activatedNodes[0].getCookie(ArgumentsCookie.class));
        return rightCookie != null;
    }

    /** Actually performs SetArguments action */
    protected void performAction (final Node[] activatedNodes) {
        ArgumentsCookie rightCookie =
            (ArgumentsCookie)(activatedNodes[0].getCookie(ArgumentsCookie.class));
        if (rightCookie != null)
            displayArgumentsDialog (rightCookie);
    }

    protected int mode () {
        return CookieAction.MODE_EXACTLY_ONE;
    }

    protected Class[] cookieClasses () {
        return new Class[] { ArgumentsCookie.class };
    }

    /** Human presentable name of the action. This should be
    * presented as an item in a menu.
    * @return the name of the action
    */
    public String getName() {
        return NbBundle.getBundle(SetArgumentsAction.class).getString("SetArguments");
    }

    /** Help context where to find more about the acion.
    * @return the help context for this action
    */
    public HelpCtx getHelpCtx() {
        return new HelpCtx (SetArgumentsAction.class);
    }

    private static void displayArgumentsDialog (ArgumentsCookie cookie) {
        String[] args = cookie.getArguments ();
        StringBuffer sb = new StringBuffer ();
        for (int i = 0; i < args.length; i++) {
            sb.append (args[i]);
            if (i != args.length - 1) sb.append (" "); // NOI18N
        }

        ArgumentsPanel ap = new ArgumentsPanel (sb.toString ());
        HelpCtx.setHelpIDString(ap, "set.arguments.window");
        DialogDescriptor dd = new DialogDescriptor (ap, NbBundle.getBundle(SetArgumentsAction.class).getString("CTL_SetArgumentsTitle"));
        dd.setHelpCtx(new HelpCtx("set.arguments.window"));
        TopManager.getDefault ().createDialog (dd).show ();

        if (dd.getValue () == DialogDescriptor.OK_OPTION) {
            StringTokenizer st = new StringTokenizer (ap.getArguments ());
            args = new String[st.countTokens ()];
            int i = 0;
            while (st.hasMoreTokens ()) {
                args[i++] = st.nextToken ();
            }
            try {
                cookie.setArguments (args);
            } catch (IOException ex) {
                TopManager.getDefault().notify(
                    new NotifyDescriptor.Message(
                        NbBundle.getBundle(SetArgumentsAction.class).getString("MSG_ArgumentsCannotBeSet"),
                        NotifyDescriptor.ERROR_MESSAGE
                    )
                );
            }
        }
    }

    static class ArgumentsPanel extends javax.swing.JPanel {

        private javax.swing.JLabel jLabel1;
        private javax.swing.JTextField argField;

        /** Initializes the Form */
        public ArgumentsPanel (String args) {
            ResourceBundle bundle;

            setLayout (new java.awt.GridBagLayout ());
            java.awt.GridBagConstraints gridBagConstraints1;

            jLabel1 = new javax.swing.JLabel ();
            jLabel1.setText ((bundle = NbBundle.getBundle(SetArgumentsAction.class)).getString("CTL_SetArguments"));
            jLabel1.setDisplayedMnemonic(bundle.getString("CTL_SetArguments_Mnemonic").charAt(0));
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            add (jLabel1, gridBagConstraints1);

            argField = new javax.swing.JTextField ();
            jLabel1.setLabelFor(argField);
            gridBagConstraints1 = new java.awt.GridBagConstraints ();
            gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
            gridBagConstraints1.weightx = 1.0;
            gridBagConstraints1.insets = new java.awt.Insets (0, 12, 0, 0);
            add (argField, gridBagConstraints1);

            argField.setText (args);
            argField.unregisterKeyboardAction(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_ENTER, 0));

            setBorder (new javax.swing.border.EmptyBorder (new java.awt.Insets(12, 12, 0, 11)));
            argField.requestFocus ();
        }

        public java.awt.Dimension getPreferredSize () {
            return new java.awt.Dimension (300, 50);
        }

        String getArguments () {
            return argField.getText ();
        }
        
        
    }
}
