/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.execution;

import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.util.ResourceBundle;

import org.openide.options.SystemOption;
import org.openide.actions.ExecuteAction;
import org.openide.util.HelpCtx;

/** Settings for Execution
*
* @author Ales Novak
*/
public class ExecutionSettings extends SystemOption {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 4261950851983665892L;

    /** Property name of the workspace property */
    public static final String PROP_WORKSPACE = "workspace"; // NOI18N
    /** Property name of the reuse property */
    public static final String PROP_TAB_HANDLER = "handler"; // NOI18N
    /** Property name of the runCompilation property */
    public static final String PROP_RUN_COMPILATION = "runCompilation"; // NOI18N
    
    // old props
    private static final String PROP_REUSE = "reuse";
    private static final String PROP_CLEAR = "clear";
    
    static final int NONE = 0;
    static final int REUSE = 1;
    static final int REUSE_AND_CLEAR = 2;

    private transient int tabHandler = REUSE_AND_CLEAR;
    
    /** @return the display name of Execution Settings */
    public String displayName () {
        return ProcessNode.getBundle().getString("CTL_Execution_option");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (ExecutionSettings.class);
    }

    // ------------------------------------------
    // property access methods

    public void setWorkspace (String workspace) {
        String oldValue = getWorkspace ();
        if (workspace.equals (oldValue)) return;
        ExecuteAction.setWorkspace (workspace);
        // fire the PropertyChange
        firePropertyChange (PROP_WORKSPACE, new Boolean (oldValue), new Boolean (workspace));
    }
    public String getWorkspace () {
        return ExecuteAction.getWorkspace ();
    }
    
    /** Getter for tab handler */
    public int getTabHandler() {
        return tabHandler;
    }
    
    /** Setter for tab handler */
    public void setTabHandler(int nval) {
        int old = tabHandler;
        tabHandler = nval;
        if (old != tabHandler) {
            firePropertyChange(PROP_TAB_HANDLER, new Integer(old), new Integer(nval));
        }
    }
    
    /** setter for runCompilation property */
    public void setRunCompilation(boolean value) {
        boolean oldValue = getRunCompilation();
        ExecuteAction.setRunCompilation(value);
        if (oldValue != value)
            firePropertyChange(PROP_RUN_COMPILATION, new Boolean(oldValue), new Boolean(value));
    }
    /** getter for runCompilation property */
    public boolean getRunCompilation() {
        return ExecuteAction.getRunCompilation();
    }
    
    
    private static final String version11 = "version 1.1";
    
    public void writeExternal(ObjectOutput oo) throws IOException {
        oo.writeObject(version11);
        oo.writeUTF(getWorkspace());
        oo.writeBoolean(getRunCompilation());
        oo.writeInt(getTabHandler());
    }
    
    public void readExternal(ObjectInput oi) throws IOException, ClassNotFoundException {
        String read = (String) ((java.io.ObjectInputStream) oi).readObject();
        
        if (version11.equals(read)) {
            read11(oi);
        } else {
            read10(oi, read);
        }
    }
    
    void read10(ObjectInput oi, String read) throws IOException, ClassNotFoundException {
        Object val;
        Object preread;
        Object useMethod;
        boolean inClear = false;
        do {
            val = oi.readObject();
            
            if (read.equals(PROP_WORKSPACE)) {
                setWorkspace((String) val);
            } else if (read.equals(PROP_RUN_COMPILATION)) {
                setRunCompilation(((Boolean) val).booleanValue());
            } else if (read.equals(PROP_REUSE)) {
                if (! inClear) {
                    tabHandler = REUSE;
                }
            } else if (read.equals(PROP_CLEAR)) {
                inClear = true;
                tabHandler = REUSE_AND_CLEAR;
            }
             
            useMethod = oi.readObject();
            if (useMethod instanceof String) {
                read = (String) useMethod;
            } else if (useMethod == null) {
                break;
            } else {
                read = (String) oi.readObject();
                if (read == null) {
                    break;
                }
            }
        } while (true);
    }
    
    void read11(ObjectInput oi) throws IOException, ClassNotFoundException {
        setWorkspace(oi.readUTF());
        setRunCompilation(oi.readBoolean());
        setTabHandler(oi.readInt());
    }
}

/*
 * Log
 *  8    Gandalf   1.7         1/12/00  Ales Novak      i18n
 *  7    Gandalf   1.6         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun
 *       Microsystems Copyright in File Comment
 *  6    Gandalf   1.5         7/2/99   Jesse Glick     More help IDs.
 *  5    Gandalf   1.4         6/8/99   Ian Formanek    ---- Package Change To 
 *       org.openide ----
 *  4    Gandalf   1.3         5/14/99  Ales Novak      bugfix for #1667 #1598 
 *       #1625
 *  3    Gandalf   1.2         2/12/99  Ian Formanek    Reflected renaming 
 *       Desktop -> Workspace
 *  2    Gandalf   1.1         1/21/99  Ales Novak      
 *  1    Gandalf   1.0         1/5/99   Ian Formanek    
 * $
 */
