/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.execution;

import java.io.FileDescriptor;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.URL;
import java.net.MalformedURLException;
import java.security.Permission;
import java.security.AllPermission;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;
import java.security.AccessController;
import java.security.AccessControlContext;
import java.util.*;
import java.lang.reflect.Field;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.execution.NbClassLoader;

//import ice.iblite.BrowserClassLoader;

/** See java.lang.SecurityManager for more details.
*
* @author Ales Novak
*/
public class TopSecurityManager extends SecurityManager {

    /** thread group of executed classes */
    private ThreadGroup base;
    
    private Permission allPermission;
    
    /* JVMPI sometimes deadlocks sync getForeignClassLoader
        and Class.forName
    */
    private static final Class classLoaderClass;
    private static final Class nbClassLoaderClass;
    private static final Class URLClass;

    static {
        try {
            classLoaderClass = Class.forName("java.lang.ClassLoader");
            nbClassLoaderClass = Class.forName("org.openide.execution.NbClassLoader");
            URLClass = Class.forName("java.net.URL");
            getAccessControllerClass();
        } catch (ClassNotFoundException e) {
            throw new ExceptionInInitializerError();
        }
    }

    /**
    * constructs new TopSecurityManager
    */
    public TopSecurityManager () {
        base = ExecutionEngine.base;
        allPermission = new AllPermission();
    }

    public void checkExit(int status) throws SecurityException {
        if (! check) {
            return;
        }
        
        PrivilegedCheck.checkExit(status, this);
    }


    final void checkExitImpl(int status, AccessControlContext acc) throws SecurityException {             
        ThreadGroup g = Thread.currentThread().getThreadGroup ();
        ThreadGroup old = null;
        while (g != null && g != base) {
            old = g;
            g = g.getParent ();
        }

        IOPermissionCollection iopc;
        iopc = AccController.getIOPermissionCollection(acc);

        /*if (g != null) { // g == base
            if (old instanceof TaskThreadGroup) {
                if (iopc != null) {
                    ExecutionEngine.getTaskIOs().free(old, iopc.getIO()); // closes output
                } // else loaded by App classloader
                ExecutionEngine.closeGroup(old);
                stopTaskThreadGroup((TaskThreadGroup) old);
            }
        } else {*/
        if ((iopc != null) &&
                (iopc.grp != null)) {
            ExecutionEngine.getTaskIOs().free(iopc.grp, iopc.getIO()); // closes output
            ExecutionEngine.closeGroup(iopc.grp);
            stopTaskThreadGroup(iopc.grp);
            throw new ExitSecurityException();
        }

        if (isNbClassLoader()) {
            throw new SecurityException();
        }
        //}
        super.checkExit(status);
    }

    public boolean checkTopLevelWindow(Object window) {
        IOPermissionCollection iopc = AccController.getIOPermissionCollection();
        if (iopc != null && iopc.grp != null && (window instanceof java.awt.Window)) {
            ExecutionEngine.putWindow((java.awt.Window) window, iopc.grp);
        }
        return super.checkTopLevelWindow(window);
    }

    /** Hack against permissions of Launcher$AppLoader. */
    public void checkPackageAccess(String pckg) {
        if (pckg == null) return;
        if (pckg.startsWith("sun.")) { // NOI18N
            if (inClazz("sun.misc.Launcher") || inClazz("java.lang.Class")) { // NOI18N
                return;
            }
        }
        super.checkPackageAccess(pckg);
    }

    /** Performance - all props accessible */
    public final void checkPropertyAccess(String x) {
        return;
    }

    /* ----------------- private methods ------------- */

    private boolean inClazz(String s) {
        Class[] classes = getClassContext();
        int i = 0;
        for (; (i < classes.length) && (classes[i] == TopSecurityManager.class); i++);
        if (i == classes.length) {
            return false;
        }
        return classes[i].getName().startsWith(s);
    }

    private void stopTaskThreadGroup(TaskThreadGroup old) {
        synchronized(old) {
            int count = old.activeCount();
            int icurrent = -1;
            Thread current = Thread.currentThread();
            Thread[] thrs = new Thread[count];
            old.enumerate(thrs, true);
            for (int i = 0; i < thrs.length; i++) {
                if (thrs[i] == null) break;
                if (thrs[i] == current) icurrent = i;
                else thrs[i].stop();
            }
            if (icurrent != -1) thrs[icurrent].stop();
            //throw new ExitSecurityException();
        }
    }
    
    /**
     * The method is empty. This is not "secure", but on the other hand,
     * it reduces performance penalty of startup about 10%
     */
    public void checkRead(String file) {
    }
    
    public void checkRead(FileDescriptor fd) {
    }

    public void checkWrite(FileDescriptor fd) {
    }

    /** The method has awful performance in super class */
    public void checkDelete(String file) {
        try {
            checkPermission(allPermission);
            return;
        } catch (SecurityException e) {
            super.checkDelete(file);
        }
    }
    /** The method has awful performance in super class */
    public void checkWrite(String file) {
        try {
            checkPermission(allPermission);
            return;
        } catch (SecurityException e) {
            super.checkWrite(file);
        }
    }
    
    /** Checks connect */
    public void checkConnect(String host, int port) {
        if (! check) {
            return;
        }
        
        try {
            checkPermission(allPermission);
            return;
        } catch (SecurityException e) {
        }
        
        try {
            super.checkConnect(host, port);
            return;
        } catch (SecurityException e) {
        }
        
        PrivilegedCheck.checkConnect(host, port, this);
    }
     
    final void checkConnectImpl(String host, int port) {
        Class insecure = getInsecureClass();
        if (insecure != null) {  
            URL ctx = getClassURL(insecure);
            if (ctx == URL_UNKNOWN) {
                throw new SecurityException();
            }
            if (ctx != null) {
                try {
                    String fromHost = ctx.getHost();
                    InetAddress ia2 = InetAddress.getByName(host);
                    InetAddress ia3 = InetAddress.getByName(fromHost);
                    if (ia2.equals(ia3)) {
                        return;
                    }
                } catch (UnknownHostException e) { // ignore
                    TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, e);
                }
            }
            throw new SecurityException();
        }
    }

    public void checkConnect(String s, int port, Object context) {
        checkConnect(s, port);
    }

    public void checkPermission(Permission perm) {
        checkSetSecurityManager(perm);
        if (! check) {
            return;
        }
        super.checkPermission(perm);
    }

    public void checkPermission(Permission perm, Object context) {
        checkSetSecurityManager(perm);
        if (! check) {
            return;
        }
        super.checkPermission(perm, context);
    }
    
    /** Prohibits to set another SecurityManager */
    private static void checkSetSecurityManager(Permission perm) {
        if (perm instanceof RuntimePermission) {
            if ("setSecurityManager".equals(perm.getName())) { // NOI18N - hardcoded in java.lang
                throw new SecurityException();
            }
        }
    }

    public void checkMemberAccess(Class clazz, int which) {
        if ((which == java.lang.reflect.Member.PUBLIC) ||
                javax.swing.text.JTextComponent.class.isAssignableFrom(clazz)) {
            return;
        } else {
            super.checkMemberAccess(clazz, which);
        }
    }

    /** @return true iff an instance of the NbClassLoader class is on the stack
    */
    protected boolean isNbClassLoader() {
        Class[] ctx = getClassContext();
        ClassLoader cloader;

        for (int i = 0; i < ctx.length; i++) {
            if ((ctx[i].getClassLoader() instanceof NbClassLoader) &&
                (ctx[i].getProtectionDomain().getCodeSource() != null)) {
                return true;
            }
        }
        return false;
    }

    private Class getInsecureClass() {

        Class[] ctx = getClassContext();
        ClassLoader cloader;
        boolean firstACClass = false;

LOOP:   for (int i = 0; i < ctx.length; i++) {

            if (ctx[i] == getAccessControllerClass()) {
                // privileged action is on the stack before an untrusted class loader
                // #3950
                if (firstACClass) {
                    return null;
                } else {
                    firstACClass = true;
                    continue LOOP;
                }
            } else if ((cloader = ctx[i].getClassLoader()) != null) {
                final Class hisClass = cloader.getClass();

                if (isSecureClass(ctx[i])) {
                    if (classLoaderClass.isAssignableFrom(ctx[i])) {
                        return null;
                    } else {
                        // OK process next one
                        continue LOOP;
                    }
                }

                return ctx[i];
            } else if (classLoaderClass.isAssignableFrom(ctx[i])) { // cloader == null
                return null; // foreign classloader wants to do work...
            }
        }

        return null;
    }

    /** Checks if the class is loaded through the nbfs URL */
    static boolean isSecureClass(final Class clazz) {
        URL source = getClassURL(clazz);
        if (source == URL_UNKNOWN) {
            return false;
        }
        if (source != null) {
            return isSecureProtocol(source.getProtocol());
        } else {
            return true;
        }
    }
    
    static URL URL_UNKNOWN;
    
    static {
        try {
            URL_UNKNOWN = new URL("http", "UNKNOWN", -1, "UNKNOWN");
        } catch (MalformedURLException e) {
            Error err = new ExceptionInInitializerError();
            TopManager.getDefault().getErrorManager().annotate(err, e);
            throw err;
        }
    }
            
    
    /** @return a protocol through which was the class loaded (file://...) or null
    */
    static URL getClassURL(Class clazz) {
        java.security.CodeSource cs = clazz.getProtectionDomain().getCodeSource();                                                     
        if (cs != null) {
            URL url = cs.getLocation();
            if (url != null) {
                return url;
            } else {
                // [PENDING] remove this else if iceblit is not used
                ClassLoader him = clazz.getClassLoader();
                Class hisClass = him.getClass();
                if (hisClass.getName().indexOf("ice.iblite.BrowserClassLoader") >= 0) { // NOI18N
                    try {
                        Field fld = getUrlField(hisClass);
                        if (fld == null) {
                            return null;
                        } else {
                            return (URL) fld.get(him);
                        }
                    } catch (Exception e) {
                        TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, e);
                    }
                }
                return URL_UNKNOWN;
            }
        } else { // PROXY CLASS?
            return null;
        }
    }

    static Field getUrlField(Class clazz) {
        if (urlField == null) {
            try {
                Field[] fds = clazz.getDeclaredFields();
                for (int i = 0; i < fds.length; i++) {
                    if (fds[i].getType() == URLClass) {
                        fds[i].setAccessible(true);
                        urlField = fds[i];
                        break;
                    }
                }
            } catch (Exception e) {
                TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, e);
            }
        }
        return urlField;
    }

    private static Field urlField;

    /** @return Boolean.TRUE iff the string is a safe protocol (file, nbfs, ...) */
    static boolean isSecureProtocol(String protocol) {
        if (protocol.equals("http") || // NOI18N
            protocol.equals("ftp") || // NOI18N
            protocol.equals("rmi")) { // NOI18N
            return false;
        } else {
            return true;
        }
    }

    private static Class accessControllerClass;

    /** @return java.security.AccessController.class */
    static Class getAccessControllerClass() {
        if (accessControllerClass == null) {
            accessControllerClass = forName("java.security.AccessController"); // NOI18N
        }
        return accessControllerClass;
    }

    private static Class forName(String fn) {
        try {
            return Class.forName(fn);
        } catch (ClassNotFoundException ex) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, ex);
            return null;
        }
    }

    private static final class PrivilegedCheck implements PrivilegedExceptionAction {
        int action;
        TopSecurityManager tsm;
        
        // exit
        int status;
        AccessControlContext acc;

        // connect
        String host;
        int port;
        
        
        public PrivilegedCheck(int action, TopSecurityManager tsm) {
            this.action = action;
            this.tsm = tsm;
            
            if (action == 0) {
                acc = AccessController.getContext();
            }
        }
        
        public Object run() throws Exception {
            switch (action) {
                case 0 : 
                    tsm.checkExitImpl(status, acc);
                    break;
                case 1 :
                    tsm.checkConnectImpl(host, port);
                    break;
                default :
            }
            return null;
        }
        
        static void checkExit(int status, TopSecurityManager tsm) {
            PrivilegedCheck pea = new PrivilegedCheck(0, tsm);
            pea.status = status;
            check(pea);
        }
        
        static void checkConnect(String host, int port, TopSecurityManager tsm) {
            PrivilegedCheck pea = new PrivilegedCheck(1, tsm);
            pea.host = host;
            pea.port = port;
            check(pea);
        }
        
        private static void check(PrivilegedCheck action) {
            try {
                AccessController.doPrivileged(action);
            } catch (PrivilegedActionException e) {
                Exception orig = e.getException();
                if (orig instanceof RuntimeException) {
                    throw ((RuntimeException) orig);
                }

                TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, orig);
            }
        }
    }
    
    
    private static boolean check;
    static {
        if (Boolean.getBoolean("netbeans.security.nocheck")) { // NOI18N
            check = false;
        } else {
            check = true;
        }
    }
}

/*
 * Log
 *  29   Jaga      1.26.1.1    4/10/00  Ales Novak      deadlock
 *  28   Jaga      1.26.1.0    3/7/00   Ales Novak      made faster
 *  27   src-jtulach1.26        1/20/00  Petr Hamernik   rolled back
 *  26   src-jtulach1.25        1/19/00  Petr Nejedly    Commented out debug 
 *       messages
 *  25   src-jtulach1.24        1/13/00  Jaroslav Tulach I18N
 *  24   src-jtulach1.23        1/12/00  Ales Novak      i18n
 *  23   src-jtulach1.22        12/20/99 Ales Novak      checkRead/Write/Delete 
 *       is back - FilePermissions are skipped
 *  22   src-jtulach1.21        12/14/99 Ales Novak      simplified security
 *  21   src-jtulach1.20        11/5/99  Ales Novak      secure protocols
 *  20   src-jtulach1.19        11/1/99  Ales Novak      cloudscape classloader 
 *       added into a list of trusted classloaders
 *  19   src-jtulach1.18        10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun
 *       Microsystems Copyright in File Comment
 *  18   src-jtulach1.17        10/8/99  Ales Novak      security checks
 *  17   src-jtulach1.16        10/6/99  Ales Novak      dumpStack removed
 *  16   src-jtulach1.15        10/5/99  Ales Novak      #4166
 *  15   src-jtulach1.14        10/1/99  Martin Ryzl     "RMIClassLoader" added 
 *       to secure classloaders
 *  14   src-jtulach1.13        9/10/99  Jaroslav Tulach Does not use any weak 
 *       maps.
 *  13   src-jtulach1.12        9/2/99   Jaroslav Tulach ClassLoaderSupport is 
 *       trusted too.
 *  12   src-jtulach1.11        9/1/99   Jaroslav Tulach Treats jars in extension
 *       directory as trusted.
 *  11   src-jtulach1.10        8/31/99  Jaroslav Tulach Works not only for 
 *       Launcher$AppLoader, but also for Launcher$ExtLoader
 *  10   src-jtulach1.9         8/31/99  Jaroslav Tulach checkPermisson only for 
 *       foreign loaders
 *  9    src-jtulach1.8         8/30/99  Ales Novak      applet security
 *  8    src-jtulach1.7         8/6/99   Ales Novak      NullPointerExc
 *  7    src-jtulach1.6         7/28/99  Ales Novak      new window system/#1409
 *  6    src-jtulach1.5         7/28/99  Ales Novak      bugfix #1409
 *  5    src-jtulach1.4         7/9/99   Ales Novak      #2413 fix
 *  4    src-jtulach1.3         6/3/99   Ales Novak      security checks disabled
 *  3    src-jtulach1.2         4/10/99  Ales Novak      
 *  2    src-jtulach1.1         4/8/99   Ales Novak      
 *  1    src-jtulach1.0         3/31/99  Ales Novak      
 * $
 */
