/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.lookup;

import java.util.*;

import org.openide.cookies.InstanceCookie;
import org.openide.loaders.*;
import org.openide.util.Lookup;
import org.openide.ErrorManager;
import org.openide.TopManager;

/** Implements a lookup, that scans a content of a folder for its
 * data objects and asks them for instance cookie, the created objects
 * are then used to for the content of the lookup.
 *
 * @author  Jaroslav Tulach
 */
public class FolderLookup extends FolderInstance {
    
    /** lookup to delegate to */
    private NbLookup pairs;
    /** list of lookups to delegate to */
    private ProxyLookup proxies;
    
    /** Constructor
     * @param df the folder containing all 
     */
    public FolderLookup (String folderName) {
        this (createFolder (folderName));
    }
    
    /** Constructor.
     * @param df folder
     */
    private FolderLookup (DataFolder df) {
        super (df);
    }

    /* 
    * @return FolderLookup class.
    */
    public Class instanceClass () {
        return ProxyLookup.class;
    }
    
    /** Getter for the lookup, that should be used.
     */
    public Lookup getLookup () {
        if (proxies == null) {
            synchronized (this) {
                pairs = new Lkp ();
                proxies = new ProxyLookup (new Lookup[] { pairs });
            }
            recreate ();
        }
                
        return proxies;
    }
    
    /** Create the folder.
     * @param folderName name to use
     */
    private static DataFolder createFolder (String folderName) {
        try {
            DataFolder rootFolder = DataFolder.findFolder (
                org.openide.TopManager.getDefault ().getRepository ().getDefaultFileSystem ().getRoot ()
            );
            DataFolder df = DataFolder.create (rootFolder, folderName);
            return df;
        } catch (java.io.IOException ex) {
            ex.printStackTrace();
            throw new IllegalStateException ("Cannot initialize folder: " + folderName); // NOI18N
        }
    }
        
    /** Updates the content of the lookup.
     *
     *
     * @param cookies updated array of instance cookies for the folder
     * @return object to represent these cookies
     *
     * @exception IOException an I/O error occured
     * @exception ClassNotFoundException a class has not been found
     */
    protected Object createInstance(InstanceCookie[] cookies) 
    throws java.io.IOException, ClassNotFoundException {
        LinkedList ll = new LinkedList ();
        LinkedList del = new LinkedList ();

        for (int i = 0; i < cookies.length; i++) {
            InstanceCookie ic = cookies[i];
            try {
                if (Lookup.class.isAssignableFrom (ic.instanceClass ())) {
                    // ok, delegate to this lookup
                    del.add (ic.instanceCreate ());
                    continue;
                }
            } catch (java.io.IOException ex) {
                exception (ex);
            } catch (ClassNotFoundException ex) {
                exception (ex);
            }

            ll.add (new ICItem (ic));
        }
        
        // initializes all lookups
        Lookup result = getLookup ();

        pairs.setPairs (ll);
        del.addFirst (pairs);
        Lookup[] arr = (Lookup[])del.toArray (new Lookup[0]);
        proxies.setLookups (arr);

        return result;
    }
    
    /** Folder is recognized as another lookup.
     */
    protected InstanceCookie acceptFolder (DataFolder df) {
        return new FolderLookup (df);
    }
    
    /** Item that delegates to InstanceCookie.
     */
    private static final class ICItem extends NbLookup.Item {
        private InstanceCookie ic;
        
        public ICItem (InstanceCookie ic) {
            this.ic = ic;
        }
        
        
        /** The class of the result item.
         * @return the class of the item
         */
        public Class getType() {
            try {
                return ic.instanceClass ();
            } catch (ClassNotFoundException ex) {
                exception (ex);
            } catch (java.io.IOException ex) {
                exception (ex);
            }
            return Object.class;
        }
        
        /** The class of the result item.
         * @return the instance of the object.
         */
        public Object getInstance() {
            try {
                return ic.instanceCreate();
            } catch (ClassNotFoundException ex) {
                exception (ex);
            } catch (java.io.IOException ex) {
                exception (ex);
            }
            return null;
        }
        
    } // end of ICItem
    
    
    /** Notifies the exception
     */
    private static void exception (Exception ex) {
        org.openide.TopManager.getDefault ().getErrorManager ().notify (
            org.openide.ErrorManager.INFORMATIONAL, 
            ex
        );
    }
    
    /** list of not safe to wait classes.
     */
    private static final HashSet dangerousClasses = new HashSet (37);
    static {
        dangerousClasses.add (org.openide.filesystems.MIMEResolver.class);
    }
    
    /** Some classes are not safe to be waited for. Here is the list.
     */
    private static boolean isSafeToWaitFor (Class c) {
        return !dangerousClasses.contains (c);
    }
    
    /** Lookup used for access to folder instances.
     */
    private final class Lkp extends NbLookup {
        
        /** Waits before the processing of changes is finished.
         */
        protected void beforeLookup (Class clazz) {
            if (isSafeToWaitFor (clazz)) {
                instanceFinished ();
            }
        }
    }
}
