/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.lookup;


import java.lang.ref.*;
import java.util.*;

import org.openide.util.Lookup;
import org.openide.util.LookupListener;
import org.openide.util.LookupEvent;
import org.openide.util.WeakSet;

/** Implementation of the lookup from OpenAPIs that is based on the
 * introduction of Item. This class should provide the default way
 * of how to store (Class, Object) pairs in the lookups. It offers 
 * protected methods for subclasses to register the pairs.
 *
 * @author  Jaroslav Tulach
 */
public class NbLookup extends Lookup {
    /** lock for initialization of the map */
    private Object mapLock = new Object ();
    /** map (Class, List(Item)) */
    private Map map;
    /** set (Class, List (Reference (Result)) of all listeners that are waiting in 
     * changes in class Class
     */
    private Map reg;
    
    
    /** Getter for the map object. If not initilized calls initialize first.
     */
    private Map getMap () {
        if (map != null) {
            return map;
        }

        synchronized (mapLock) {
            if (map == null) {
                // ok, I am the thread that is going to initialize the map 
                map = new HashMap ();
                initialize ();
            }
        }
        
        return map;
    }
    
    /** Method for subclasses to initialize them selves.
     */
    protected void initialize () {
    }
    
    

    /** The method to add instance to the lookup with.
     * @param pair class/instance pair
     */
    protected final void addPair (Item pair) {
        Class c = pair.getType ();
        HashSet toNotify = new HashSet ();
        
        Map m = getMap ();
        
        // manipulation with map must be synchronized on this
        synchronized (this) {
            recursiveAdd (m, c, pair);
            
            collectListeners (toNotify, c);
        }
        
        notifyListeners (toNotify);
    }
    
    /** Remove instance.
     * @param pair class/instance pair
     */
    protected final void removePair (Item pair) {
        Class c = pair.getType ();
        HashSet toNotify = new HashSet ();
        
        synchronized (this) {
            if (map == null) {
                // nothing needs to be done
                return;
            }
            
            recursiveRemove (map, c, pair);
            
            collectListeners (toNotify, c);
        }
        
        notifyListeners (toNotify);
    }
    
    /** Changes all pairs in the lookup to new values.
     * @param collection the collection of (Item) objects
     */
    protected final void setPairs (Collection collection) {
        HashSet toNotify = new HashSet ();

        Map m = getMap ();
        
        // manipulation with map must be synchronized on this
        synchronized (this) {
            // collect all listeners of objects to remove
            Iterator it = m.keySet ().iterator ();
            while (it.hasNext ()) {
                Class c = (Class)it.next ();
                collectListeners (toNotify, c);
            }
            
            // clear the current map, because we will have new pairs
            m.clear ();
            
            
            it = collection.iterator ();
            while (it.hasNext ()) {
                Item pair = (Item)it.next ();
                Class c = pair.getType ();
                
                recursiveAdd (m, c, pair);

                collectListeners (toNotify, c);
            }
        }
        
        notifyListeners (toNotify);
    }
                
    
    /** Lookups an object of given interface. This is the simplest method
     * for the lookuping, if more registered objects implement the given
     * class any of them can be returned. 
     *
     * @param clazz class of the object we are searching for
     * @return the object implementing given class or null if no such 
     *    has been found
     */
    public Object lookup (Class clazz) {
        List l = (List)getMap ().get (clazz);
        
        // manipulation with map must be synchronized
        synchronized (this) {
            if (l == null || l.isEmpty ()) {
                return null;
            } else {
                Item i = (Item)l.get (0);

                return i.getInstance ();
            }
        }
    }
    
    /** The general lookup method.
     * @param template the template describing the services we are looking for
     * @return object containing the results
     */
    public synchronized final NbLookup.Result lookup (NbLookup.Template template) {
        R result = new R  (template);
        
        if (reg == null) {
            reg = new HashMap (11);
        }
        
        List ws = (List)reg.get (template.getType ());
        if (ws == null) {
            ws = new LinkedList ();
            reg.put (template.getType (), ws);
        }
        ws.add (new WeakReference (result));
        
        return result;
    }
    
    
    /** Delegate method to implement the the lookup method from the original
     * lookup.
     */
    public final Lookup.Result lookup (Lookup.Template template) {
        return lookup (new NbLookup.Template (template));
    }

    /** Notifies all listeners that are interested in changes in this class.
     * Should be called from synchronized places.
     * @param allAffectedResults adds Results into this set
     * @param c the class that has changed
     */
    private void collectListeners (HashSet allAffectedResults, Class c) {
        if (reg == null) {
            return;
        }
    
        while (c != null) {
            List l = (List)reg.get (c);
            if (l != null && !l.isEmpty ()) {
                Iterator it = l.iterator ();
                while (it.hasNext ()) {
                    Reference ref = (Reference)it.next ();
                    R result = (R)ref.get ();
                    if (result == null) {
                        // the result has been garbage collected
                        it.remove ();
                    } else {
                        allAffectedResults.add (result);
                    }
                }
            }

            c = c.getSuperclass ();
        }
    }   
        
    
    /** Notify change to all Results in the set.
     */
    private static void notifyListeners (HashSet allAffectedResults) {
        if (allAffectedResults.isEmpty ()) {
            return;
        }
        
        Iterator it = allAffectedResults.iterator ();
        while (it.hasNext ()) {
            R result = (R)it.next ();
            result.fireStateChanged ();
        }
    }
    
    /** Fills the map with all super classes and classes as keys of 
     * item any object
     */
    private static void recursiveAdd (Map map, Class c, Object item) {
        if (c == null) {
            return;
        }
        
        
        List list = (List)map.get (c);
        if (list != null) {
            // add the item to this class
            list.add (item);
        } else {
            list = new LinkedList ();
            list.add (item);
            map.put (c, list);
        }
        
        Class[] arr = c.getInterfaces ();
        for (int i = 0; i < arr.length; i++) {
            recursiveAdd (map, arr[i], item);
        }
        recursiveAdd (map, c.getSuperclass (), item);
    }

    /** Removes all instances of item from the map for c and all its superclasses and
     * superinterfaces.
     *
     * @param map map of (Class, List(Object))
     * @param c class 
     * @param item item
     */
    private static void recursiveRemove (Map map, Class c, Object item) {
        if (c == null) {
            return;
        }
        
        List list = (List)map.get (c);
        if (list != null) {
            // add the item to this class
            list.remove (item);
        }
        
        Class[] arr = c.getInterfaces ();
        for (int i = 0; i < arr.length; i++) {
            recursiveRemove (map, arr[i], item);
        }
        recursiveRemove (map, c.getSuperclass (), item);
    }
    
    /** Template to define pattern of object to look for. 
     * <P>
     * I expect this class to grow in the future, but for now, it is 
     * enough to start with something simple.
     */
    public static final class Template extends Object {
        /** type of the service */
        private Class type;
        
        /** General template to find all possible instances.
         */
        public Template () {
        }
        
        /** Copy constructor to bridge with the Lookup.Template class
         */
        public Template (Lookup.Template t) {
            type = t.getType ();
        }
        
        /** Constructor to create new template.
         * @param serviceClass the class of service we are looking for
         */
        public Template (Class type) {
            this.type = type;
        }
        
        /** @return the class we are looking for */
        public Class getType () {
            return type;
        }
    }
    
    /** Result of a lookup request. Allows access to single object
     * that was found (not too useful) and also to all objects found
     * (more useful).
     */
    public static abstract class Result extends Lookup.Result {
        /** Registers a listener that is invoked when there is a possible 
         * change in this result.
         *
         * @param l listener to invoke when there is an change
         */
        public abstract void addLookupListener (LookupListener l);

        /** Unregisters a listener previously added by addLookupListener.
         * @param l the listener
         */
        public abstract void removeLookupListener (LookupListener l);
        
        /** Access to all instances in the result.
         * @return collection of all instances
         */
        public abstract java.util.Collection allInstances ();
        
        /** Classes of all results. Set of the most concreate classes
         * that are registered in the system.
         * @return set of Class objects
         */
        public abstract java.util.Set allClasses ();
        
        /** All registered items. The collection of all pairs of 
         * instances and their classes.
         * @return collection of Lookup.Item
         */
        public abstract java.util.Collection allItems ();
    }
    
    /** A single item in the result. An object to bind together the 
     * instance and its class.
     * <P>
     * In future it could also get access to additional attributes.
     */
    public static abstract class Item extends Object {
        /** The class of the result item.
         * @return the instance of the object.
         */
        public abstract Object getInstance ();
        
        /** The class of the result item.
         * @return the class of the item
         */
        public abstract Class getType ();
    }

    
    /** Result based on one instance returned.
     */
    private final class R extends Result {
        private Template template;
        
        /** temporary caches */
        private Set classesCache;
        private Collection instancesCache;
        private Collection itemsCache;
        
        /** listeners on the results */
        private ArrayList listeners;
        
        private R (Template template) {
            this.template = template;
        }
        
        /** Delete all cached values, the template changed.
         */
        public void fireStateChanged () {
            classesCache = null;
            instancesCache = null;
            itemsCache = null;
            
            LookupListener[] arr;
            synchronized (this) {
                if (listeners == null) return;
                
                arr = (LookupListener[])listeners.toArray (
                    new LookupListener[listeners.size ()]
                );
            }
            
            LookupEvent ev = new LookupEvent (this);
            for (int i = 0; i < arr.length; i++) {
                arr[i].resultChanged (ev);
            }
        }
        
        /** Ok, register listeners to all classes and super classes.
         */
        public synchronized void addLookupListener (LookupListener l) {
            if (listeners == null) {
                listeners = new ArrayList ();
            }
            listeners.add (l);
        }    
        
        /** Ok, register listeners to all classes and super classes.
         */
        public synchronized void removeLookupListener (LookupListener l) {
            listeners.remove (l);
        }    
        

        public Collection allInstances () {
            Collection s = instancesCache;
            if (s != null) {
                return s;
            }
            
            s = new LinkedList ();
            Iterator it = allItems ().iterator ();
            while (it.hasNext ()) {
                Item item = (Item)it.next ();
                Object obj = item.getInstance ();
                if (obj != null) {
                    s.add (obj);
                }
            }
            instancesCache = s;
            
            return s;
        }
        
        public Set allClasses () {
            Set s = classesCache;
            if (s != null) {
                return s;
            }
            
            s = new HashSet ();
            Iterator it = allItems ().iterator ();
            while (it.hasNext ()) {
                Item item = (Item)it.next ();
                Class clazz = item.getType ();
                if (clazz != null) {
                    s.add (clazz);
                }
            }
            classesCache = s;
            
            return s;
        }

        /** Items are stored directly in the allItems.
         */
        public Collection allItems () {
            if (itemsCache != null) {
                return itemsCache;
            }
            
            List l = (List)getMap ().get (template.getType ());
            
            if (l != null) {
                // manipulation with map must be synchronized on this
                synchronized (NbLookup.this) {
                    return itemsCache = new LinkedList (l);
                }
            } else {
                return Collections.EMPTY_SET;
            }
        }
        
    }
    
}