/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.lookup;

import java.util.*;

import org.openide.util.Lookup;
import org.openide.util.LookupListener;
import org.openide.util.LookupEvent;
import org.openide.util.WeakSet;

/** Implementation of lookup that can delegate to provided to others 
 * provided.
 *
 * @author  Jaroslav Tulach
 */
public final class ProxyLookup extends Lookup {
    /** lookups to delegate to */
    private Lookup[] lookups;
    /** list of all currently active results */
    private WeakSet results;
    
    /** Constructor.
     * @param lookups delegator's lookups
     */
    public ProxyLookup (Lookup[] lookups) {
        this.lookups = lookups;
    }
    
    /** Changes the lookups.
     */
    public void setLookups (Lookup[] lookups) {
        R[] arr;
        HashSet newL;
        HashSet current;
        Lookup[] old;
        
        synchronized (this) {
            current = new HashSet (Arrays.asList (this.lookups));
            newL = new HashSet (Arrays.asList (lookups));

            old = this.lookups;
            this.lookups = lookups;
            
            if (results == null || results.isEmpty ()) {
                // no affected results => exit
                return;
            }
	    arr = (R[])results.toArray( new R[0] );
        }
        
        HashSet removed = new HashSet (current);
        removed.removeAll (newL); // current contains just those lookups that have disappeared
        newL.removeAll (current); // really new lookups
        
        if (removed.isEmpty () && newL.isEmpty ()) {
            // no need to notify changes
            return;
        }
                                    
        for (int i = 0; i < arr.length; i++) {
            arr[i].lookupChange (newL, removed, old, lookups);
        }
    }
    
    /** Lookups an object of given interface. This is the simplest method
     * for the lookuping, if more registered objects implement the given
     * class any of them can be returned. 
     *
     * @param clazz class of the object we are searching for
     * @return the object implementing given class or null if no such 
     *    has been found
     */
    public Object lookup (Class clazz) {
        Lookup[] lookups = this.lookups;
        
        for (int i = 0; i < lookups.length; i++) {
            Object o = lookups[i].lookup (clazz);
            if (o != null) {
                return o;
            }
        }
        
        return null;
    }
    
    /** The general lookup method.
     * @param template the template describing the services we are looking for
     * @return object containing the results
     */
    public synchronized Result lookup (Lookup.Template template) {
        R r = new R (template);
        
        if (results == null) {
            results = new WeakSet ();
        }
        results.add (r); 
        return r;
    }
    
    /** Result of a lookup request. Allows access to single object
     * that was found (not too useful) and also to all objects found
     * (more useful).
     */
    private final class R extends Lookup.Result implements LookupListener {
        /** list of listeners added */
        private ArrayList listeners;
        
        /** template for this result */
        private Lookup.Template template;
        /** all results */
        private Lookup.Result[] results;
        /** collection of Objects */
        private Collection instances;
        /** set of Class */
        //private Set classes;
        /** collection of Item */
        //private Collection items;
        
        
        /** Constructor.
         */
        public R (Lookup.Template t) {
            template = t;
        }

        /** initializes the results 
         */
        private Result[] initResults () {
            if (results != null) return results;
            
            Result[] arr = new Result[lookups.length];
            for (int i = 0; i < arr.length; i++) {
                arr[i] = lookups[i].lookup (template);
                arr[i].addLookupListener (this);
            }
            
            results = arr;
            return arr;
        }
        
        /** Called when there is a change in the list of proxied lookups.
         * @param added set of added lookups
         * @param remove set of removed lookups
         * @param current array of current lookups
         */
        protected synchronized void lookupChange (
            Set added, Set removed, Lookup[] old, Lookup[] current
        ) {
            if (results == null) {
                // not computed yet, do not need to do anything
                return;
            }
            
            // map (Lookup, Lookup.Result)
            HashMap map = new HashMap (old.length * 2);
            
            
            for (int i = 0; i < old.length; i++) {
                if (removed.contains (old[i])) {
                    // removed lookup
                    results[i].removeLookupListener (this);
                } else {
                    // remember the association
                    map.put (old[i], results[i]);
                }
            }

            Lookup.Result[] arr = new Lookup.Result[current.length];
            for (int i = 0; i < current.length; i++) {
                if (added.contains (current[i])) {
                    // new lookup
                    arr[i] = current[i].lookup (template);
                    arr[i].addLookupListener (this);
                } else {
                    // old lookup
                    arr[i] = (Lookup.Result)map.get (current[i]);
                    if (arr[i] == null) {
                        // assert
                        throw new IllegalStateException ();
                    }
                }
            }
            
            // remember the new results
            results = arr;
            
            resultChanged (null);
        }
        
        /** Just delegates.
         */
        public synchronized void addLookupListener (LookupListener l) {
            if (listeners == null) {
                listeners = new ArrayList ();
            }
            listeners.add (l);
        }
            
        /** Just delegates.
         */
        public synchronized void removeLookupListener (LookupListener l) {
            if (listeners != null) {
                listeners.remove (l);
            }
        }
        
        /** Access to all instances in the result.
         * @return collection of all instances
         */
        public java.util.Collection allInstances () {
            java.util.Collection res = instances;
            if (res != null) return res;
            
            // results to use
            Lookup.Result[] arr;
            
            synchronized (this) {
                res = instances;
                if (res != null) return res;

                arr = initResults ();
            }
                
            ArrayList ll = new ArrayList (arr.length * 2);
            for (int i = 0; i < arr.length; i++) {
                ll.addAll (arr[i].allInstances ());
            }

            synchronized (this) {
                if (arr == results) {
                    // updates the results, if the results have not been
                    // changed during the computation of allInstances
                    instances = ll;
                }
            }
            
            return ll;
        }
        
        /** When the result changes, fire the event.
         */
        public void resultChanged (LookupEvent ev) {
            // clear cached instances
            instances = null;
            
            if (listeners == null) return;
            
            ev = new LookupEvent (this);
            
            LookupListener[] arr;
            synchronized (this) {
                arr = (LookupListener[])listeners.toArray (
                    new LookupListener [
                        listeners.size ()
                    ]
                );
            }
            
            for (int i = 0; i < arr.length; i++) {
                arr[i].resultChanged (ev);
            }
        }
                    
        
        /** Classes of all results. Set of the most concreate classes
         * that are registered in the system.
         * @return set of Class objects
         *
        public java.util.Set allClasses () {
            if (classes != null) return classes;
            
            synchronized (this) {
                if (classes != null) return classes;
                
                initResults ();
                
                HashSet ll = new HashSet ();
                for (int i = 0; i < results.length; i++) {
                    ll.addAll (results[i].allClasses ());
                }
                
                classes = ll;
            }
            
            return classes;
        }
         */
        
        /** All registered items. The collection of all pairs of 
         * ii and their classes.
         * @return collection of Lookup.Item
         *
        public java.util.Collection allItems () {
            if (items != null) return items;
            
            synchronized (this) {
                if (items != null) return items;
                
                initResults ();
                
                LinkedList ll = new LinkedList ();
                for (int i = 0; i < results.length; i++) {
                    ll.addAll (results[i].allItems ());
                }
                
                items = ll;
            }
            
            return items;
        }
        */
    }
}