/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.output;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Event;
import java.awt.EventQueue;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Point;
import java.awt.event.*;
import java.awt.datatransfer.StringSelection;
import java.io.IOException;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.util.*;
import java.text.MessageFormat;
import javax.swing.*;
import javax.swing.Timer;
import javax.swing.event.*;

import org.openide.TopManager;
import org.openide.awt.JPopupMenuPlus;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.cookies.EditorCookie;
import org.openide.text.Line;
import org.openide.windows.*;
import org.openide.actions.CopyAction;
import org.netbeans.core.actions.NextOutJumpAction;
import org.netbeans.core.actions.PreviousOutJumpAction;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallbackSystemAction;
import org.openide.util.io.NullOutputStream;

/** This class represents one of output panes in one OutputTab (It can be
* either stdout either errout). It extends JComponent. Inside this component
* is inserted org.openide.text.view.ViewManager.
*
* @author Petr Hamernik, Jaroslav Tulach
*/
public final class OutPane extends JList
            implements MouseListener, ListSelectionListener, KeyListener,
    ActionPerformer, FocusListener, ActionListener {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -633812069000420549L;
    /** Information channel for this pane */
    PaneWriter writer;
    /** model for the list */
    private PaneWriter model;

    /** My parent output tab */
    OutputTab tab;

    /** Boolean flag - First jump line was set or not */
    boolean jumpLineSet;

    /** Copy action */
    private static CopyAction copyAction = (CopyAction)CopyAction.get (CopyAction.class);

    /** Private instance of Next jump action */
    private static NextOutJumpAction nextAction = (NextOutJumpAction)NextOutJumpAction.get (NextOutJumpAction.class);

    /** Private instance of Previous jump action */
    private static PreviousOutJumpAction previousAction = (PreviousOutJumpAction)PreviousOutJumpAction.get (PreviousOutJumpAction.class);

    /** output settings */
    private static OutputSettings outputSettings = (OutputSettings)OutputSettings.findObject (OutputSettings.class, true);

    /** Performer for jump actions */
    private JumpActionPerformer jumpPerformer = new JumpActionPerformer();

    /** Mapping from the line numbers to listeners (Integer, Listener) */
    private TreeMap listeners = new TreeMap ();
    /** Map of lineNO:ide.text.Line */
    private Map int2Line;
    /** Message format for exception parsing */
    private static MessageFormat formatOfException;
    /** PopupMenu */
    JPopupMenu jPopup;
    private JMenuItem copyItem;
    private JMenuItem clearItem;
    /** Timer for ensureIndexIsVisible */
    javax.swing.Timer fireTimer;

    /** Creates pane without association to a tab.
    */
    public OutPane() {
        this (null);
    }

    /** Creates new OutPane in the specific OutputTab */
    public OutPane(OutputTab tab) {
        this.tab = tab;
        int2Line = new HashMap(29);
        this.writer = this.model = new PaneWriter();

        setModel (model);
        setCellRenderer (jumpPerformer);
        setBackground (outputSettings.getBaseBackground ());
        setSelectionMode (ListSelectionModel.SINGLE_SELECTION);

        jumpLineSet = false;

        addMouseListener(this);
        addListSelectionListener (this);
        addKeyListener(this);
        addFocusListener(this);
        jPopup = new JPopupMenuPlus();

        // add copy
        copyItem = new JMenuItem(org.openide.util.NbBundle.getBundle (OutPane.class).getString ("CTL_ClipboardCopy"));
        copyItem.addActionListener(this);
        jPopup.add(copyItem);
        jPopup.addSeparator();
        // add clear
        clearItem = new JMenuItem(org.openide.util.NbBundle.getBundle (OutPane.class).getString ("CTL_Clear"));
        clearItem.addActionListener(this);
        jPopup.add(clearItem);

        // add(jPopup);
    }

    /** The writer for this pane.
    */
    public OutputWriter getOut () {
        return writer;
    }

    private OutputListener getListenerForPosition(int pos) {
        return (OutputListener)listeners.get (new Integer (pos));
    }

    /** Checks (sets/unsets) performer for CopyAction */
    final void checkCopyAction() {
        ActionPerformer aperf = (getSelectedIndex () != -1 ? this : null);
        copyAction.setActionPerformer(aperf);
    }

    void checkNextPrevActions() {
        //    if (tab.amISelected(this)) {
        nextAction.setActionPerformer((nextJump() != -1) ? jumpPerformer : null);
        previousAction.setActionPerformer((previousJump() != -1) ? jumpPerformer : null);
        checkCopyAction();
        //    }
        // Those comments are part of a bugfix - checkNextPrevAction doesn't work
        // when you are pressing alt-F8/alt-F7 in the editor
    }

    private int nextJump() {
        Integer i = new Integer (getSelectedIndex () + 1);

        synchronized (listeners) {
            SortedMap sm = listeners.tailMap (i);
            if (sm.isEmpty ()) return -1;
            i = (Integer)sm.firstKey ();
        }
        return i.intValue ();
    }

    private int previousJump() {
        Integer i = new Integer (getSelectedIndex () - 1);

        synchronized (listeners) {
            SortedMap sm = listeners.headMap (i);
            if (sm.isEmpty ()) {
                return -1;
            }
            i = (Integer)sm.lastKey ();
        }
        return i.intValue ();
    }

    /** This method is called whenever jump line is written. It tests if this line
    * is first and if it is sets the cursor to this position
    */
    private void stopAtFirstJumpLine(int index) {
        if (!jumpLineSet) {
            setSelectedIndex (index);
            jumpLineSet = true;
        }
    }

    /** Clears whole lines table */
    void clearLineTable() {
        synchronized (listeners) {
            Iterator it = listeners.entrySet ().iterator ();
            while (it.hasNext ()) {
                Map.Entry e = (Map.Entry)it.next ();
                int pos = ((Integer)e.getKey ()).intValue ();
                OutputListener listener = (OutputListener) e.getValue ();
                listener.outputLineCleared(new OutputEventImpl(tab, pos));
            }
            listeners.clear ();
            jumpLineSet = false;
            checkNextPrevActions();
        }
    }

    /** My implementation of OutputEvent. It has lazy initialized line text. */
    private final class OutputEventImpl extends OutputEvent {
        private int index;

        static final long serialVersionUID =-437312909583471519L;

        public OutputEventImpl(InputOutput src, int index) {
            super(src);
            // #7363
            this.index = index;
        }

        /** Returns text on the line.
        * @return the text on the line
        */
        public String getLine () {
            return model.getElementAt(index).toString ();
        }
    }

    private static final String EMPTY = " "; // NOI18N

    /** Writer, which can insert text into the output document.
    * @see java.io.Writer
    */
    final class PaneWriter extends OutputWriter implements ListModel, Runnable {

        /** Array with all lines (String) except current line */
        ArrayList lines;

        /** Store current line */
        private StringBuffer currentLine;
        private String currentLineStr;

        /** the buffer with text to add */
        private StringBuffer sb;

        /** Last printed character was '\r' */
        private boolean lastR = false;

        /** only one listener for the model */
        private ListDataListener dataListener;
        /** up to origSize JList knows about lines
         * To ensure that JList knows about lines that 
         * you have to fire events...
         */
        int origSize;
        
        public PaneWriter() {
            super(new OutputStreamWriter(new NullOutputStream()));
            this.out = this;
            this.lock = this;

            lines = new ArrayList();
            currentLine = new StringBuffer();
            currentLineStr = EMPTY;
            sb = new StringBuffer();
            origSize = 0;
        }

        void cleaned () {
            redraw();
        }

        /** Draws text from buffer to editor.
        * Must be called from synchronized methods.
        */
        private synchronized void redraw() {

            String bufferStr = sb.toString();
            StringTokenizer tok = new StringTokenizer(bufferStr, "\n\r\t", true); // NOI18N

            String tab = null;

            while (tok.hasMoreTokens()) {
                String token = tok.nextToken();
                if (token.equals("\n") || token.equals("\r")) { // NOI18N
                    boolean r = token.equals("\r"); // NOI18N
                    if (r || !lastR) {
                        String addLine = (currentLine.length() == 0) ? EMPTY : currentLine.toString();
                        lines.add(addLine);
                        currentLine.setLength(0);
                        currentLineStr = EMPTY;
                    }
                    lastR = r;
                }
                else {
                    if (token.equals("\t")) { // NOI18N
                        if (tab == null)
                            tab = getTab();
                        token = tab;
                    }
                    currentLine.append(token);
                    currentLineStr = null;
                }
            }

            sb.setLength(0);
            restartFire();
        }

        long lastPaintedTime = 0L;
        int lastPaintedIdx = 0;

        void restartFire() {
            Timer timer = getFireTimer();
            long thisTime = System.currentTimeMillis();
            if (lastPaintedTime == 0) {
                lastPaintedTime = System.currentTimeMillis();
            } else {
                int lineDelta = (getUncommittedSize() - lastPaintedIdx);
                if (lineDelta > 1) {
                    long delta = (thisTime - lastPaintedTime) / 1000;
                    if (delta > 0) {
                        // flat the profile
                        //long ratio = (lineDelta + 10) / (delta + 10);
                        //if (ratio < 15) {
                            timer.stop();
                            fireChanges();
                            return;
                        //}
                    }
                }
            }
            timer.restart();
        }
        
        void fireChanges() {
            
            if (EventQueue.isDispatchThread()) {
                getFireTimer().stop();
                lastPaintedTime = System.currentTimeMillis();
                int currentSize = getUncommittedSize();
                int tmp = origSize;
                origSize = currentSize;
                
                if (currentSize != tmp) {
                    fireIntervalAdded(tmp, currentSize);
                }  else {
                    fireContentsChanged(tmp, currentSize);
                }

                if ((currentSize > 0) && (getSelectedIndex() < 0)) {
                    ensureIndexIsVisible(currentSize - 1);
                }
                
                lastPaintedTime = System.currentTimeMillis();
                lastPaintedIdx = origSize;
            } else {
                EventQueue.invokeLater(this);
            }
        }
        
        public void run() {
            fireChanges();
        }


        public void write(char cbuf[], int off, int len) {
            OutputTab local = tab;
            if (local != null && local.isClosed()) {
                local.rebindTab();
            }
            
            synchronized (this) {
                sb.append(cbuf, off, len);
                redraw();
            }
        }

        public void write(int c) {
            OutputTab local = tab;
            if (local != null && local.isClosed()) {
                local.rebindTab();
            }
            
            synchronized (this) {
                sb.append((char) c);
                redraw();
            }
        }

        public void write(String str, int off, int len) {
            OutputTab local = tab;
            if (local != null && local.isClosed()) {
                local.rebindTab();
            }

            synchronized (this) {
                sb.append(str.substring(off, len));
                redraw();
            }
        }

        /** Prints the given string and register OutputListener on this line */
        public void println(String s, OutputListener l) throws IOException {
            OutputTab local = tab;
            if (local != null && local.isClosed()) {
                local.rebindTab();
            }

            synchronized (this) {
                if (sb.length() > 0) {
                    sb.append("\n"); // NOI18N
                }

                // #7363
                int indx = lines.size();

                sb.append(s);
                sb.append("\n"); // NOI18N

                boolean empty = listeners.isEmpty();

                listeners.put(new Integer(indx), l);

                // flush the new line
                redraw();

                if (empty) {
                    // stop at this position
                    setSelectedIndex(indx);
                    // fires info about the change
                    valueChanged (null);
                }

                checkNextPrevActions();
            }
        }

        public void flush() {
            if (tab != null && tab.isClosed()) {
                return;
            }
            redraw();
        }

        public void close() {
            reset();
        }

        public synchronized void reset() {
            clearLineTable();
            int2Line.clear();

            final int size = getUncommittedSize();
            lines.clear();
            sb.setLength(0);
            currentLine.setLength(0);
            currentLineStr = EMPTY;
            if (size >= 0) {
                if (EventQueue.isDispatchThread()) {
                    fireIntervalRemoved (0, size);
                } else {
                    Runnable run = new Runnable() {
                        public void run() {
                            fireIntervalRemoved (0, size);
                        }
                    };
                    EventQueue.invokeLater(run);
                }
            }
            origSize = 0;
            lastPaintedTime = 0;
            lastPaintedIdx = 0;
        }

        //
        // ListDataListener methods
        //

        public Object getElementAt (int i) {
            if (i < lines.size())
                return lines.get(i);
            else if ((i == lines.size()) && (currentLine.length() > 0)) {
                if (currentLineStr == null)
                    currentLineStr = currentLine.toString();
                return currentLineStr;
            }
            else {
                return EMPTY;
            }
        }

        public int getSize() {
            // return number of commited lines
            return origSize;
        }
        
        int getUncommittedSize() {
            return lines.size() + ((currentLine.length() > 0) ? 1 : 0);
        }

        /**
         * Add a listener to the list that's notified each time a change
         * to the data model occurs.
         * @param l the ListDataListener
         */
        public void addListDataListener(ListDataListener l) {
            if (dataListener != null) {
                // only one listener supported
                throw new InternalError ();
            }
            dataListener = l;
        }

        /**
         * Remove a listener from the list that's notified each time a
         * change to the data model occurs.
         * @param l the ListDataListener
         */
        public void removeListDataListener(ListDataListener l) {
            dataListener = null;
        }


        protected void fireContentsChanged(int index0, int index1) {
            if (dataListener != null) dataListener.contentsChanged(
                    new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, index0, index1)
                );
        }


        protected void fireIntervalAdded(int index0, int index1) {
            if (dataListener != null) dataListener.intervalAdded (
                    new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED, index0, index1)
                );
        }


        protected void fireIntervalRemoved(int index0, int index1) {
            if (dataListener != null) dataListener.intervalRemoved (
                    new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED, index0, index1)
                );
        }
    }

    final class JumpActionPerformer extends JLabel implements ActionPerformer, ListCellRenderer  {
        {
            // initializer
            this.setOpaque (true);
        }

        static final long serialVersionUID =4405590143900383138L;

        /** getPreferredSize is about 27% faster than default impl
        * if we cache FontMetrics 
        */
        private FontMetrics fm;
        private String txt = "";

        /** Performer for actions */
        public void performAction(SystemAction action) {
            int jump = -1;
            if (action instanceof NextOutJumpAction) {
                jump = nextJump();
            }
            if (action instanceof PreviousOutJumpAction) {
                jump = previousJump();
            }
            if (jump != -1) {
                setSelectedIndex (jump);
                if (action instanceof NextOutJumpAction) {
                    ensureIndexIsVisible (jump + 2);
                }
                ensureIndexIsVisible (jump);
                invokeJumpListener(jump);
            }
            checkNextPrevActions();
        }
        
        private Dimension recycle;

        public Dimension getPreferredSize() {
            checkFont();
            String txt = getText();
            if (txt == null) {
                txt = OutPane.this.EMPTY;
            }
            
            if (recycle == null) {
                recycle = new Dimension(fm.stringWidth(txt), fm.getHeight());
                return recycle;
            }
            
            recycle.width = fm.stringWidth(txt);
            recycle.height = fm.getHeight();
            return recycle;
        }

        public String getText() {
            return txt;
        }

        public void setText(String text) {
            String old = txt;
            txt = text;
            //this.firePropertyChange("text", old, text);
        }

        /** Component for rendering the cell.
        */
        public java.awt.Component getListCellRendererComponent (JList list, Object value,
                int index, boolean isSelected,
                boolean cellHasFocus) {

            String newVal = (value instanceof String) ? (String) value : ((StringBuffer)value).toString ();

            checkFont();
            setText(newVal);
            Integer bigIdx = new Integer(index);

            boolean isJump = listeners.containsKey (bigIdx) ||
                             parseException(bigIdx, false);
            if (index != list.getSelectedIndex()) {
                this.setBackground (outputSettings.getBaseBackground ());
                this.setForeground (outputSettings.getBaseForeground ());
            } else {
                if (isJump) {
                    this.setBackground (outputSettings.getJumpCursorBackground ());
                    this.setForeground (outputSettings.getJumpCursorForeground ());
                } else {
                    this.setBackground (outputSettings.getCursorBackground ());
                    this.setForeground (outputSettings.getCursorForeground ());
                }
            }

            return this;
        }

        void checkFont() {
            Font font = this.getFont();
            if (font != null &&
                fm != null &&
                font.isPlain() &&
                (font.getSize() == outputSettings.getFontSize())) {
                return;
            } else {
                Font nf = new Font("monospaced", java.awt.Font.PLAIN, outputSettings.getFontSize()); // NOI18N
                fm = java.awt.Toolkit.getDefaultToolkit().getFontMetrics(nf);
                this.setFont(nf);
            }
        }
    }


    /** @return true if there were an listner */
    private boolean invokeJumpListener(int index) {
        OutputListener listener = getListenerForPosition(index);
        if (listener != null) {
            listener.outputLineAction(new OutputEventImpl(tab, index));
            return true;
        } else {
            return false;
        }
    }

    // ListSelectionListener method

    /** previous selected index */
    private int previousIndex = -1;


    public void valueChanged (ListSelectionEvent ev) {
        int indx = getSelectedIndex ();

        if (indx == previousIndex) return;

        previousIndex = indx;

        OutputListener listener = getListenerForPosition(indx);
        if (listener != null) {
            listener.outputLineSelected(new OutputEventImpl(tab, indx));
        }

        checkNextPrevActions();
    }

    // mouse listener methods
    public void mouseClicked (MouseEvent evt) {
        if ((evt.getClickCount() == 2) && ((evt.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)) {
            int loc = locationToIndex(evt.getPoint());
            if (! invokeJumpListener(loc)) {
                parseException(new Integer(loc), true);
            }
        }
    }

    public void mousePressed(MouseEvent e) {
        if (isPopupTrigger(e)) {
            showPopup(e.getX(), e.getY());
        }
    }


    /** @return true if this event is popup menu trigger... */
    private static boolean isPopupTrigger(MouseEvent evt) {
        return (evt.getModifiers() & (MouseEvent.BUTTON2_MASK | MouseEvent.BUTTON3_MASK)) != 0;
    }

    /** Shows "copy" popup menu */ // NOI18N
    void showPopup(int xpos, int ypos) {
        jPopup.show(this, xpos, ypos);
    }
    
    // focus listener
    CallbackSystemAction csa;
    int lastidx = -1;
    
    
    public void focusGained(FocusEvent ev) {
        checkCopyAction();
        if (csa == null) {
            try {
                Class popup = Class.forName("org.openide.actions.PopupAction");
                csa = (CallbackSystemAction) CallbackSystemAction.get(popup);
            } catch (ClassNotFoundException e) {
                Error err = new NoClassDefFoundError();
                TopManager.getDefault().getErrorManager().annotate(err, e);
                throw err;
            }
        }
        if (lastidx >= 0) {
            setSelectedIndex(lastidx);
        }
        csa.setActionPerformer(this);
    }
    
    public void focusLost(FocusEvent ev) {
        if (csa != null) {
            csa.setActionPerformer(null);
        }
        int i = getSelectedIndex();
        lastidx = i;
        // do not do copyAction.setActionPerformer(null);
        // doesn't work with main window
    }

    /** Parses exception or tries to find
    * @param openEditor <tt>true</tt> if an editor should be open
    * @return <tt>true</tt> if the line is recognized as part of exception dump
    */
    boolean parseException(Integer loc, boolean openEditor) {

        int smallLoc = loc.intValue();
        if (smallLoc < 0) {
            return false;
        }

        String s = ((String) model.getElementAt(smallLoc)).trim();

        try {
            Object aLine = int2Line.get(loc);
            
            if (aLine == int2Line) { 
                return false;
            }
            
            Line l;
            if (isExceptionHeuristic(s)) {
                if (aLine == null) {
                    int idx = s.indexOf("at ");
                    if (idx < 0) {
                        int2Line.put(loc, int2Line);
                        return false;
                    }
                    
                    idx += 3;
                    s = s.substring(idx);
                    Object[] o = getExceptionFormat().parse(s);

                    String all = o[0].toString();
                    String file = o[1].toString();
                    String ext = o[2].toString();
                    int line = Integer.parseInt(o[3].toString());

                    int i = all.indexOf(file);
                    String path = all.substring(0, i + file.length()).replace('.', '/') + ".java"; // NOI18N

                    FileObject fo = TopManager.getDefault().getRepository().findResource(path);
                    if (fo == null) {
                        int2Line.put(loc, int2Line);
                        return false;
                    }

                    DataObject data = DataObject.find(fo);
                    EditorCookie cookie = (EditorCookie) data.getCookie(EditorCookie.class);
                    if (cookie == null) {
                        int2Line.put(loc, int2Line);
                        return false;
                    }
                    l = cookie.getLineSet().getOriginal (line - 1);
                } else {
                    l = (Line) aLine;
                }

                if (openEditor) {
                    l.show(Line.SHOW_GOTO, 0);
                }
                return true;
            }
            // ignore all
        } catch (java.text.ParseException e) { // MsgForm.parse
        } catch (NumberFormatException e) { // Integer.parseInt
        } catch (DataObjectNotFoundException e) { // DO.find
        } catch (IndexOutOfBoundsException e) { // getLine
        }

        int2Line.put(loc, int2Line);
        return false;
    }

    static boolean isExceptionHeuristic(String s) {
        s = s.trim();
        return (startsOK(s) && s.charAt(s.length() - 1) == ')'); // NOI18N
    }
    
    private static boolean startsOK(String s) {
        return (s.startsWith("at ") || s.startsWith("[catch] at "));
    }

    /** Getter for a MessageFormat that describes a line of exception dump. */
    private static MessageFormat getExceptionFormat() {
        if (formatOfException == null) {
            formatOfException = new MessageFormat("{0}({1}.{2}:{3})");
        }
        return formatOfException;
    }

    /** Performer for copy action */
    public void performAction(SystemAction action) {
        if (action instanceof CopyAction) {
            doCopy();
        } else if (action instanceof org.openide.actions.PopupAction) {
            int i = getSelectedIndex();
            if (i < 0) {
                return;
            }
            Point p = indexToLocation(i);
            if (p == null) {
                return;
            }
            showPopup(p.x, p.y);
        }
    }
    
    /** Performer for registerKeyboardAction */
    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == copyItem) {
            doCopy();
        } else if (e.getSource() == clearItem) {
            doClear();
        } else {
            getFireTimer().stop();
            model.fireChanges();
        }
    }

    /** Inserts selected indexes into the clipboard */
    private void doCopy() {
        final Object[] o = model.lines.toArray();

        StringBuffer buff = new StringBuffer(o.length * 25);

        if (o.length >= 1) {
            buff.append(o[0].toString());

            for (int i = 1; i < o.length; i++) {
                buff.append('\n').append(o[i].toString()); // NOI18N
            }
            
            if (model.currentLine.length() > 0) {
                buff.append('\n').append(model.currentLine.toString());
            }
        } else {
            buff.append(model.currentLine.toString());
        }

        StringSelection ss = new StringSelection(buff.toString());
        TopManager.getDefault().getClipboard ().setContents (ss, ss);
    }

    private void doClear() {
        model.close();
        model.cleaned ();
    }

    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

    // key listener waits for enter

    public void keyTyped (KeyEvent e) {
    }

    public void keyPressed (KeyEvent e) {
        if (e.getKeyCode () == KeyEvent.VK_ENTER) {
            e.consume ();

            int indx = getSelectedIndex();

            if (indx < 0) {
                return;
            }

            if (! invokeJumpListener(indx)) {
                parseException(new Integer(indx), true);
            }
        }
    }

    public void keyReleased (KeyEvent e) {
    }

    static String getTab() {
        int tabSize = outputSettings.getTabSize();
        StringBuffer buf = new StringBuffer(tabSize);
        for (int j = 0; j < tabSize; j++)
            buf.append(" "); // NOI18N
        return buf.toString();
    }

    Timer getFireTimer() {
        if (fireTimer == null) {
            fireTimer = new Timer(80, this);
            fireTimer.setRepeats(false);
            fireTimer.setCoalesce(true);
        }
        return fireTimer;
    }
}
