/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.perftool;

import java.io.*;
import java.util.*;
import java.text.DateFormat;
import org.apache.regexp.*;

/** Parses ide.log, searches for sessions, and acquires
 * data about the specific session
 *
 */
final class Parser {

    private static final String nline = System.getProperty("line.separator");
    private static final String SESSION_START_PATTERN = "^([\\-]+)" + nline + "^(>Log Session: .*)";
    private static final String SESSION_END_PATTERN = "^([\\-]+)".concat(nline);
    // [PENDING] needs to escape metacharacters in LOG_PREFIX:
    private static final String ENTRY_PATTERN = "^" + PerformanceMeterImpl.LOG_PREFIX + "([^0-9]+)([0-9]+)\\.([0-9]+) s" ;
    
    private static RE logstart;
    private static RE logend;
    private static RE entries;
    
    static RE getLogStart() throws RESyntaxException {
        if (logstart == null) {
            RECompiler rec = new RECompiler();
            REProgram rep = rec.compile(SESSION_START_PATTERN);
            logstart = new RE(rep, RE.MATCH_MULTILINE);
        }
        return logstart;
    }
    
    static RE getLogEnd() throws RESyntaxException {
        if (logend == null) {
            RECompiler rec = new RECompiler();
            REProgram rep = rec.compile(SESSION_END_PATTERN);
            logend = new RE(rep, RE.MATCH_MULTILINE);
        }
        return logend;
    }
    
    static RE getEntries() throws RESyntaxException {
        if (entries == null) {
            RECompiler rec = new RECompiler();
            REProgram rep = rec.compile(ENTRY_PATTERN);
            entries = new RE(rep, RE.MATCH_MULTILINE);
        }
        return entries;
    }
    
    /** Constructs new Parser */
    private Parser() {
    }
    
    public static SessionInfo[] getData(File f) throws IOException {
        try {
            Reader reader = new BufferedReader(new InputStreamReader(new FileInputStream(f)));
            ReaderCharacterIterator iter = new ReaderCharacterIterator(reader);

            RE logstart = getLogStart();
            RE logend = getLogEnd();
            
            List sessions = new ArrayList(20);

            int idx = 0; // index of last processed char
            for (;;) {
                if (logstart.match(iter, idx)) {
                    int start = logstart.getParenStart(0);
                    parseEntries(iter, idx, start, sessions);
                    if (logend.match(iter, logstart.getParenEnd(0))) {
                        int end = logend.getParenEnd(0);
                        SessionInfo info = new SessionInfo(trimString(iter.substring(start, end)));
                        sessions.add(info);
                        idx = end;
                        continue;
                    }
                }
                break;
            }

            reader.close();
            return (SessionInfo[]) sessions.toArray(new SessionInfo[sessions.size()]);
        } catch (RESyntaxException e) {
            e.printStackTrace();
        }
        
        return null;
    }
    
    private static void parseEntries(CharacterIterator iter, int startpos, int endpos, List sessions) throws IOException, RESyntaxException {
        int size = sessions.size();
        if (size == 0) {
            return;
        }
        
        SessionInfo session = (SessionInfo) sessions.get(size - 1);
        RE entries = getEntries();
        
        while ((endpos > startpos) && entries.match(iter, startpos)) {
            if (entries.getParenStart(0) > endpos) {
                startpos = entries.getParenStart(0);
                continue;
            }
            
            String entry = entries.getParen(0);
            double timeVal = strings2Double(entries.getParen(2), entries.getParen(3));
            Field f = new Field(session, entries.getParen(1), timeVal);
            session.addField(f);
            startpos = entries.getParenEnd(0);
        }
        
        if (session.size() == 0) {
            sessions.remove(size - 1);
        }
    }
    
    private static double strings2Double(String sx, String sy) {
        try {
            double ret = Double.parseDouble(sx);
            ret += Double.parseDouble(sy) / Util.tenpower(sy.length());
            return ret;
        } catch (NumberFormatException e) {
            e.printStackTrace();
        }
        
        return -1d;
    }

    private static String trimString(String s) {
        int idx = 0;
        char c;
        final int slen = s.length();

        if (slen == 0) {
            return s;
        }

        do {
            c = s.charAt(idx++);
        } while ((c == '\n' || c == '\r') &&
                 (idx < slen));

        s = s.substring(--idx);
        idx = s.length() - 1;

        if (idx < 0) {
            return s;
        }

        do {
            c = s.charAt(idx--);
        } while ((c == '\n' || c == '\r') &&
                 (idx >= 0));

        return s.substring(0, idx + 2);
    }
    
    /** Basic Info about the session
     * whole desc
     * build number (-1 if a bug)
     * Date (null if a problem)
     * Fields related to the session
     */
    public static class SessionInfo {
        private String desc;
        private Map fields;
        private int build;
        private Date date;
        
        SessionInfo(String desc) {
            this.desc = desc;
            this.fields = new HashMap(80);
            this.build = getBuildNumber(desc);
            Locale loc = getLocale(desc);
            this.date = getDate(desc, loc);
        }
        
        public String getDescription() {
            return desc;
        }
        
        void addField(Field f) {
            fields.put(f.getName(), f);
        }
        
        public Field getField(String name) {
            return (Field) fields.get(name);
        }
        
        public int getBuildNumber() {
            return build;
        }
        
        public Date getDate() {
            return date;
        }
        
        int size() {
            return fields.size();
        }
        
        private static int getBuildNumber(String desc) {
            try {
                int idx = desc.indexOf("(Build ");
                int paren = desc.indexOf(')', idx);
                String build = desc.substring(idx + 7, paren);
                if (build.equals("@BUILD_NUMBER_SUBST@")) {
                    return -1;
                }
                return Integer.parseInt(build);
            } catch (NumberFormatException e) {
                e.printStackTrace();
                return -1;
            }
        }
        
        private static Locale getLocale(String desc) {
            int idx = desc.indexOf("System Locale");
            idx = desc.indexOf("= ", idx);
            idx += 2;
            String[] locale = new String[] { null, "", "" };
            int ptr = 0;
            StringBuffer sb = new StringBuffer(3);
            char c = desc.charAt(idx);
            while (Character.isLetter(c)) {
                if (c == ':') {
                    locale[ptr++] = sb.toString();
                    sb.setLength(0);
                    continue;
                }
                sb.append(c);
                c = desc.charAt(++idx);
            }
            locale[ptr] = sb.toString();
            return new Locale(locale[0], locale[1], locale[2]);
        }
        
        private static Date getDate(String desc, Locale locale) {
            int idx = desc.indexOf(">Log Session: ");
            int end = desc.indexOf('>', idx + 14);
            String date = trimString(desc.substring(idx + 14, end));
            DateFormat df = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL, locale);
            Date dt = df.parse(date, new java.text.ParsePosition(0));
            return dt;
        }
    }
    
    /** Represents one line of Performance journal in the log */
    public static class Field {
        private String name;
        private double time;
        private SessionInfo session;
        
        Field(SessionInfo session, String name, double time) {
            this.name = name;
            this.time = time;
            this.session = session;
        }
        
        public String getName() {
            return name;
        }
        
        public double getTime() {
            return time;
        }
        
        public SessionInfo getSessionInfo() {
            return session;
        }
    }
}
