/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.projects;

import java.beans.*;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.*;

import org.openide.TopManager;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.MultiFileSystem;

/** Layered file system serving itself as either the user or installation layer.
 * Holds one layer of a writable system directory, and some number
 * of module layers.
 * @author Jesse Glick
 */
public class ModuleLayeredFileSystem extends MultiFileSystem {
    /** serial version UID */
    static final long serialVersionUID = 782910986724201983L;

    private final FileSystem writableLayer;
    private final List moduleLayers = new LinkedList (); // List<FileSystem>
    private final Map layersByModule = new HashMap (); // Map<String,FileSystem>

    /** Create layered filesystem based on a supplied writable layer.
     * @param writableLayer the writable layer to use, typically a LocalFileSystem
     */
    ModuleLayeredFileSystem (FileSystem writableLayer) {
        super (new FileSystem[] { writableLayer });
        this.writableLayer = writableLayer;
        // Wish to permit e.g. a user-installed module to mask files from a
        // root-installed module, so propagate masks up this high.
        // SystemFileSystem leaves this off, so that the final file system
        // will not show them if there are some left over.
        setPropagateMasks (true);
    }

    /** Get all layers.
     * @return all filesystems making layers
     */
    public/*but just for debugging*/ final synchronized FileSystem[] getLayers () {
        return getDelegates ();
    }

    /** Get all module-installed layers.
     * @return all of the layers added by modules
     */
    final synchronized FileSystem[] getModuleLayers () {
        return (FileSystem[]) moduleLayers.toArray (new FileSystem[moduleLayers.size ()]);
    }

    /** Get the writable layer.
     * @return the writable layer
     */
    final FileSystem getWritableLayer () {
        return writableLayer;
    }
   
    /** Creates the system file system.
     */
    public static FileSystem create (File x, File y) 
    throws java.beans.PropertyVetoException, java.io.IOException {
        return SystemFileSystem.create (x, y);
    }
    
    /** Get the installation layer.
     * You can take advantage of the specialized return type
     * if working within the core.
     */
    public static ModuleLayeredFileSystem getInstallationModuleLayer () {
        SystemFileSystem sfs = (SystemFileSystem)
            TopManager.getDefault ().getRepository ().getDefaultFileSystem();
        ModuleLayeredFileSystem home = sfs.getInstallationLayer ();
        if (home != null)
            return home;
        else
            return sfs.getUserLayer ();
    }    
    
    /** Get the user layer.
     * You can take advantage of the specialized return type
     * if working within the core.
     */
    public static ModuleLayeredFileSystem getUserModuleLayer () {
        SystemFileSystem sfs = (SystemFileSystem)
            TopManager.getDefault ().getRepository ().getDefaultFileSystem();
        return sfs.getUserLayer ();
    }
    

    /** Create a layer filesystem for a particular module.
     * @param codeName code name of the module; probably informational only
     * @param fs the filesystem to add
     * @throws IOException if the layer cannot be created
     */
    public synchronized void addModuleLayer (String codeName, FileSystem fs) throws IOException {
        moduleLayers.add (0, fs);
        layersByModule.put (codeName, fs);
        reSetDelegates ();
    }

    /** Remove an existing filesystem layer.
     * If this module had no such layer anyway, does nothing.
     * @param codeName module to remove layer for
     * @throws IOException currently doesn't
     */
    public synchronized void removeModuleLayer (String codeName) throws IOException {
        FileSystem fs = (FileSystem) layersByModule.get (codeName);
        if (fs == null) return;
        moduleLayers.remove (fs);
        layersByModule.remove (codeName);
        reSetDelegates ();
    }

    /** Update list of delegates based on changes to moduleLayers. */
    private void reSetDelegates () {
        FileSystem[] fss = new FileSystem[moduleLayers.size () + 1];
        fss[0] = writableLayer;
        for (int i = 0; i < moduleLayers.size (); i++)
            fss[i + 1] = (FileSystem) moduleLayers.get (i);
        setDelegates (fss);
        firePropertyChange ("layers", null, null);
    }

}
