/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.projects;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.*;
import java.io.*;
import java.net.URL;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.JMenuItem;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.nodes.*;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.WeakListener;

import org.netbeans.core.Main;

/*
import org.openide.util.HelpCtx;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.NodeAction;
import org.openide.nodes.Node;
import org.openide.awt.JInlineMenu;
import org.openide.windows.TopComponent;
import org.openide.loaders.DataObject;
*/

/** The system FileSystem - represents system files under $NETBEANS_HOME/system.
*
* @author Jan Jancura, Ian Formanek, Petr Hamernik
*/
public final class SystemFileSystem extends MultiFileSystem implements FileSystem.Status {
    // Must be public for BeanInfo to work: #11186.

    /** generated Serialized Version UID */
    static final long serialVersionUID = -7761052280240991668L;


    /** Resource for all localized strings in jar file system. */
    //org.openide.util.NbBundle.getBundle(SystemFileSystem.class);

    /** system name of this filesystem */
    private static final String SYSTEM_NAME = "SystemFileSystem"; // NOI18N

    /** name of file attribute with localizing bundle */
    private static final String ATTR_BUNDLE = "SystemFileSystem.localizingBundle"; // NOI18N

    /** name of file attribute with URL to 16x16 color icon */
    private static final String ATTR_ICON_16 = "SystemFileSystem.icon"; // NOI18N
    /** name of file attribute with URL to 32x32 color icon */
    private static final String ATTR_ICON_32 = "SystemFileSystem.icon32"; // NOI18N

    /** user fs */
    private ModuleLayeredFileSystem user;
    /** home fs */
    private ModuleLayeredFileSystem home;

    /** message to format file in netbeans.home */
    private static MessageFormat homeFormat;
    /** message to format file in netbeans.user */
    private static MessageFormat userFormat;

    /** @param fss list of file systems to delegate to
    */
    private SystemFileSystem (ModuleLayeredFileSystem[] fss) throws PropertyVetoException {
        super (fss);
        user = fss[0];
        home = fss.length > 1 ? fss[1] : null;

        setSystemName (SYSTEM_NAME);
        setHidden (true);
    }


    /** Name of the system */
    public String getDisplayName () {
        return NbBundle.getMessage(SystemFileSystem.class, "CTL_SystemFileSystem");
    }
    
    /** Getter for the instalation layer filesystem.
     * May be null if there is none such.
    */
    public ModuleLayeredFileSystem getInstallationLayer () {
        return home;
    }
    
    /** Getter for the user layer filesystem.
    */
    public ModuleLayeredFileSystem getUserLayer () {
        return user;
    }
    
    /** Changes layers to provided values.
     * @param arr the new layers
     * @throws IllegalArgumentException if there is an overlap
     */
    public final void setLayers (FileSystem[] arr) throws IllegalArgumentException {
        Set s = new HashSet ();
        for (int i = 0; i < arr.length; i++)
            if (s.contains (arr[i]))
                throw new IllegalArgumentException ("Overlap in filesystem layers");
            else
                s.add (arr[i]);
        setDelegates (arr);
        firePropertyChange ("layers", null, null);
    }
    
    /** Getter for the array of filesystems that are currently used 
    * in the IDE.
    *
    * @return array of filesystems
    */
    public FileSystem[] getLayers() {
        return getDelegates ();
    }

    protected FileSystem createWritableOn (String name) throws IOException {
        FileObject fo = findResource (name);
        if (fo == null) {
            // It does not exist yet. Let super decide (probably this means
            // it will write to the first one).
            return super.createWritableOn (name);
        } else {
            // See where it came from.
            FileSystem fs = findSystem (fo);
            FileSystem[] fss = getDelegates ();
            int idx = -1;
            for (int i = 0; i < fss.length; i++) {
                if (fs == fss[i]) {
                    idx = i;
                    break;
                }
            }
            if (idx == -1) {
                TopManager.getDefault ().getErrorManager ().notify
                    (ErrorManager.WARNING, new IllegalStateException
                        ("File object " + name + " found on a file system not within our delegates"));
                return super.createWritableOn (name);
            }
            // Starting here, look for the lowest writable layer.
            for (int i = idx; i >= 0; i--) {
                if (! fss[i].isReadOnly ())
                    return fss[i];
            }
            throw new IllegalStateException ("No writable filesystems in our delegates!");
        }
    }

    /** This filesystem cannot be removed from pool, it is persistent.
    */
    public boolean isPersistent () {
        return true;
    }

    public FileSystem.Status getStatus () {
        return this;
    }

    /** Annotate name
    */
    public String annotateName (String s, Set set) {

        // Look for a localized file name.
        // [PENDING] this will mask [Local] annotations, which sometimes will be annoying.
        // Perhaps there should be a file system action which indicates with a readonly checkbox
        // whether the given file is local or not.
        // Note: all files in the set are checked. But please only place the attribute
        // on the primary file, and use this primary file name as the bundle key.
        Iterator it = set.iterator ();
        while (it.hasNext ()) {
            // annotate a name
            FileObject fo = (FileObject) it.next ();

            String bundleName = (String)fo.getAttribute (ATTR_BUNDLE); // NOI18N
            if (bundleName != null) {
                try {
                    bundleName = org.openide.util.Utilities.translate(bundleName);
                    ResourceBundle b = NbBundle.getBundle (bundleName, Locale.getDefault (), TopManager.getDefault ().systemClassLoader ());
                    try {
                        return b.getString (fo.getPackageNameExt ('/', '.'));
                    } catch (MissingResourceException ex) {
                        // ignore--normal
                    }
                } catch (MissingResourceException ex) {
                    // [PENDING] merge improvements from old org.netbeans.core.SystemFileSystem
                    if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                        ex.printStackTrace ();
                    // ignore
                }
            }
        }

        return annotateNameNoLocalization (s, set);
    }

    /** Annotate name but do not consider using localized name. */
    private String annotateNameNoLocalization (String s, Set set) {

        if (home == null || user == null) {
            // no annotation if not running as multiuser
            return s;
        }

        Iterator it = set.iterator ();
        int cnt = 0;
        while (it.hasNext ()) {
            FileObject fo = (FileObject)it.next ();
            if (!fo.isRoot ()) {
                cnt++;
            }
            if (findSystem (fo) == home) {
                return getHomeFormat ().format (new Object[] { s });
            }
        }

        if (cnt == 0) {
            // only roots
            return s;
        }

        return getUserFormat ().format (new Object[] { s });
    }

    /** Annotate icon
    */
    public Image annotateIcon (Image im, int type, Set s) {
        String attr;
        if (type == BeanInfo.ICON_COLOR_16x16) {
            attr = ATTR_ICON_16;
        } else if (type == BeanInfo.ICON_COLOR_32x32) {
            attr = ATTR_ICON_32;
        } else {
            // mono icons not supported
            return im;
        }
        Iterator it = s.iterator ();
        while (it.hasNext ()) {
            FileObject fo = (FileObject) it.next ();
            Object value = fo.getAttribute (attr);
            if (value != null) {
                if (value instanceof URL) {
                    return Toolkit.getDefaultToolkit ().getImage ((URL) value);
                } else {
                    if (Boolean.getBoolean ("netbeans.debug.exceptions")) {
                        System.err.println ("Attribute " + attr + " on " + fo + " expected to be a URL; was: " + value); // NOI18N
                    }
                }
            }
        }
        return im;
    }

    /** Add an action to show the unlocalized name, if applicable. */
    /*
    public SystemAction[] getActions () {
      System.err.println("SFS.getActions");
      SystemAction[] orig = super.getActions ();
      SystemAction mine = SystemAction.get (ShowUnlocalizedNameAction.class);
      if (orig == null || orig.length == 0) {
        return new SystemAction[] { mine };
      } else {
        SystemAction[] nue = new SystemAction[orig.length + 1];
        System.arraycopy (orig, 0, nue, 0, orig.length);
        nue[orig.length] = mine;
        return nue;
      }
}

    public static class ShowUnlocalizedNameAction extends NodeAction {
      
      public String getName () {
        System.err.println("SUNA.getName");
        return ""; // NOI18N
      }
      
      protected boolean enable (Node[] nodes) {
        System.err.println("SUNA.enable");
        // [PENDING] shows an extra separator even on unloc. files
        return true;
      }
      
      protected void performAction (Node[] nodes) {
        System.err.println("SUNA.performAction");
        // ignore
      }
      
      public HelpCtx getHelpCtx () {
        return new HelpCtx (ShowUnlocalizedNameAction.class);
      }
      
      public JMenuItem getPopupPresenter () {
        System.err.println("SUNA.getPP");
        JInlineMenu menu = new JInlineMenu ();
        Node[] nodes = TopComponent.getRegistry ().getActivatedNodes ();
        System.err.println("#1");
        if (nodes.length != 1) return menu;
        DataObject dob = (DataObject) nodes[0].getCookie (DataObject.class);
        System.err.println("#2");
        if (dob == null) return menu;
        FileObject fo = dob.getPrimaryFile ();
        FileSystem fs;
        try {
          System.err.println("#3");
          if (! ((fs = fo.getFileSystem ()) instanceof SystemFileSystem)) return menu;
        } catch (FileStateInvalidException fsie) {
          if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
            fsie.printStackTrace ();
          return menu;
        }
        System.err.println("#4");
        if (fo.getAttribute (ATTR_BUNDLE) == null) return menu;
        String unlocName = ((SystemFileSystem) fs).annotateNameNoLocalization (fo.getName (), Collections.singleton (fo));
        JMenuItem item = new JMenuItem ();
        // [PENDING] it does not actually display although item is set correctly! why?
        //item.setEnabled (false);
        item.setText (MessageFormat.format (NbBundle.getBundle (SystemFileSystem.class).getString ("LBL_unlocalized_name"),
                                            new Object[] { unlocName }));
        System.err.println("item.text=" + item.getText ());
        menu.setMenuItems (new JMenuItem[] { item });
        System.err.println("#5");
        return menu;
      }
      
}
    */

    /** Initializes and creates new repository. This repository's system fs is
    * based on the content of ${HOME_DIR}/system and ${USER_DIR}/system directories
    *
    * @param userDir directory where user can write 
    * @param homeDir directory where netbeans has been installed, user need not have write access
    * @return repository
    * @exception PropertyVetoException if something fails
    */
    public static SystemFileSystem create (File userDir, File homeDir)
    throws java.beans.PropertyVetoException, IOException {
        FileSystem user;
        LocalFileSystem home;

        if (userDir != null) {
            // only one file system
            LocalFileSystem l = new LocalFileSystem ();
            l.setRootDirectory (userDir);
            user = l;
        } else {
            // use some replacement
            user = new XMLFileSystem ();
        }

        if (org.openide.util.Utilities.compareObjects (userDir, homeDir)) {
            home = null;
        } else {
            // home directory is different
            home = new LocalFileSystem ();
            home.setRootDirectory (homeDir);
            home.setReadOnly (true);                        
        }

        ModuleLayeredFileSystem[] arr = new ModuleLayeredFileSystem[home == null ? 1 : 2];
        arr[0] = new ModuleLayeredFileSystem (user);
        if (home != null) {
            arr[1] = new ModuleLayeredFileSystem (home);
        }

        return new SystemFileSystem (arr);
    }

    /** The instance of this file system.
    */
    //  public static SystemFileSystem getDefault () {
    //    return (SystemFileSystem)NbTopManager.getDefaultRepository ().getDefaultFileSystem ();
    //  }

    /** Getter for message.
    */
    private static MessageFormat getUserFormat () {
        if (userFormat == null) {
            userFormat = new MessageFormat (NbBundle.getMessage(SystemFileSystem.class, "CTL_UserFile"));
        }
        return userFormat;
    }

    /** Getter for message.
    */
    private static MessageFormat getHomeFormat () {
        if (homeFormat == null) {
            homeFormat = new MessageFormat (NbBundle.getMessage(SystemFileSystem.class, "CTL_HomeFile"));
        }
        return homeFormat;
    }

    /** Notification that a file has migrated from one file system
    * to another. Usually when somebody writes to file on readonly file
    * system and the file has to be copied to write one. 
    * <P>
    * This method allows subclasses to fire for example FileSystem.PROP_STATUS
    * change to notify that annotation of this file should change.
    *
    * @param fo file object that change its actual file system
    */
    protected void notifyMigration (FileObject fo) {
        fireFileStatusChanged (new FileStatusEvent (this, fo, false, true));
    }

    // --- SAFETY ---
    private Object writeReplace () throws ObjectStreamException {
        new NotSerializableException ("WARNING - SystemFileSystem is not designed to be serialized").printStackTrace (); // NOI18N
        return new SingletonSerializer ();
    }
    private static final class SingletonSerializer extends Object implements Serializable {
        private static final long serialVersionUID = 6436781994611L;
        private Object readResolve () throws ObjectStreamException {
            return TopManager.getDefault ().getRepository ().getDefaultFileSystem ();
        }
    }
    // --- SAFETY ---

    /** Node showing layered structure of system file system graphically.
     * Useful for debugging and for finding where settings came from.
     * Presentation is not very refined yet but can serve as a start.
     */
    public static class StructureNode extends AbstractNode {

        public StructureNode () {
            this (TopManager.getDefault ().getRepository ().getDefaultFileSystem ());
        }

        public StructureNode (FileSystem fs) {
            super (new StructureChildren (fs));
            setName (fs.getSystemName ());
            setDisplayName (fs.getDisplayName ());
        }

        private static class StructureChildren extends Children.Keys implements PropertyChangeListener {

            private FileSystem fs;

            public StructureChildren (FileSystem fs) {
                this.fs = fs;
                if ((fs instanceof SystemFileSystem) || (fs instanceof ModuleLayeredFileSystem))
                    fs.addPropertyChangeListener (WeakListener.propertyChange (this, fs));
            }

            protected void addNotify () {
                List k = new ArrayList ();
                k.add (fs.getRoot ());
                if (fs instanceof SystemFileSystem)
                    k.addAll (Arrays.asList (((SystemFileSystem) fs).getLayers ()));
                else if (fs instanceof ModuleLayeredFileSystem)
                    k.addAll (Arrays.asList (((ModuleLayeredFileSystem) fs).getLayers ()));
                setKeys (k);
            }

            protected void removeNotify () {
                setKeys (Collections.EMPTY_SET);
            }

            protected Node[] createNodes (Object k) {
                if (k instanceof FileObject) {
                    try {
                        return new Node[] { new RootNode (fs, DataObject.find ((FileObject) k).getNodeDelegate ()) };
                    } catch (DataObjectNotFoundException donfe) {
                        TopManager.getDefault ().notifyException (donfe);
                        return new Node[] { };
                    }
                } else {
                    return new Node[] { new StructureNode ((FileSystem) k) };
                }
            }

            public void propertyChange (PropertyChangeEvent pce) {
                if ("layers".equals (pce.getPropertyName ())) {
                    RequestProcessor.postRequest (new Runnable () {
                            public void run () {
                                addNotify ();
                            }
                        });
                }
            }

            private static class RootNode extends FilterNode {

                public RootNode (FileSystem fs, Node orig) {
                    super (orig);
                    disableDelegation (DELEGATE_GET_NAME | DELEGATE_SET_NAME |
                                       DELEGATE_GET_DISPLAY_NAME | DELEGATE_SET_DISPLAY_NAME);
                    setName (fs.getSystemName ());
                    setDisplayName ("Root of " + fs.getDisplayName ());
                }

            }

        }
    
    }

}
