/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import javax.swing.Icon;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JComponent;
import javax.swing.JRadioButton;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.TabbedPaneUI;

import org.openide.awt.JPopupMenuPlus;

/** User interface for javax.swing.JTabbedPane component with only one line of tabs.
 *  If JTabbedPane includes more tabs then can be displayed in one line, arrows and list icon are displayed.
 *  Left-click to the arrow shifts displays tabs.
 *  Left-Click to the list icon displays the list of tabs.
 *  @author  Rostislav Levy
 */

public class ArrowsTabbedPaneUI extends TabbedPaneUI {

//-----------------------------------------------------------------------------
// Constants

    /** Type of probed size - Minimum size
     */
    private static final int SIZE_MINIMUM = 1;
    /** Type of probed size - Maximum size
     */
    private static final int SIZE_MAXIMUM = 2;
    /** Type of probed size - Preferred size
     */
    private static final int SIZE_PREFERRED = 3;

    /** Horizontal space
     */
    private static final int TAB_SPACEX = 2;
    /** Vertical space
     */
    private static final int TAB_SPACEY = 2;
    /** Helpful property
     */
    private static final int TAB_2SPACEX = 2 * TAB_SPACEX;
    /** Helpful property
     */
    private static final int TAB_2SPACEY = 2 * TAB_SPACEY;
    /** Space between tabs and including components
     */
    private static final int SPACE_TAB = 5;
    /** With of arrow
     */
    private static final int SPACE_ARROW_WIDTH = 10;
    /** Height of arrow
     */
    private static final int SPACE_ARROW_HEIGHT = 10;
    /** Space between arrows
     */
    private static final int SPACE_ARROWS = 2;
    /** Width of the left border of tab
     */
    private static final int SPACE_TAB_BORDER_LEFT = 3;
    /** Width of the right border of tab
     */
    private static final int SPACE_TAB_BORDER_RIGHT = 3;
    /** Width of the top border of tab
     */
    private static final int SPACE_TAB_BORDER_TOP = 2;
    /** Width of the bottom border of tab
     */
    private static final int SPACE_TAB_BORDER_BOTTOM = 2;
    /** Space between tab's text and tab's icon
     */
    private static final int SPACE_TAB_TEXTICON = 3;
    /** Helpful constant
     */
    private static final int SPACE_TAB_BORDER_LEFT_RIGHT = SPACE_TAB_BORDER_LEFT + SPACE_TAB_BORDER_RIGHT;
    /** Helpful constant
     */
    private static final int SPACE_TAB_BORDER_TOP_BOTTOM = SPACE_TAB_BORDER_TOP + SPACE_TAB_BORDER_BOTTOM;
    /** Helpful constant
     */
    private static final int SPACE_TAB_BORDER_LEFT_TEXTICON = SPACE_TAB_BORDER_LEFT + SPACE_TAB_TEXTICON;

    /** Color of left and upper border
     */
    private static final Color COLOR_BORDER_LEFTTOP = Color.white;
    /** Color of right and lower border
     */
    private static final Color COLOR_BORDER_RIGHTBOTTOM = new Color(102, 102, 102);
    /** Border's color of nonactual tab
     */
    private static final Color COLOR_BACKGROUND_TABNOACTUAL = new Color(153, 153, 153);
    /** Color of focus
     */
    private static final Color COLOR_BORDER_FOCUS = new Color(102, 102, 153);
    /** Color of focus
     */
    private static final Color COLOR_ARROW = Color.black;

//-----------------------------------------------------------------------------
// Attributes

    /** Index of the first displayed tab
     */
    private int tabstart;
    /** Popup menu includes list of tabs
     */
    private JPopupMenuPlus popupmenu;
    /** Displayed instance of JTabbedPane component
     */
    private JTabbedPane tabbedpane;
    /** Previous layout managet of JTabbedPane component
     */
    private LayoutManager layoutmanager;
    /** Bounds of tabs
     */
    private Rectangle tabbounds[];
    /** Bound of previous arrow
     */
    private Rectangle boundsarrowprevious;
    /** Bound of next arrow
     */
    private Rectangle boundsarrownext;
    /** Bound of list icon
     */
    private Rectangle boundsarrowlist;
    /** Added mouse adapter to the JTabbedPane
     */
    private TabbedPaneMouseAdapter mouseadapter;

//-----------------------------------------------------------------------------
// Constructors

    /** Default constructor
     */
    public ArrowsTabbedPaneUI() {
        tabstart = 0;
        popupmenu = null;
        tabbedpane = null;
        layoutmanager = null;
        tabbounds = null;
        boundsarrowprevious = null;
        boundsarrownext = null;
        boundsarrowlist = null;
        mouseadapter = new TabbedPaneMouseAdapter();
    }

//-----------------------------------------------------------------------------
// Methods - TabbedPaneUI APIs - required

    /** Establishment of tab on a position
     *  tabbedpane - JTabbedPane component
     *  x          - horizontal position
     *  y          - vertical position
     *  return     - tab on the location
     */
    public int tabForCoordinate (JTabbedPane tabbedpane, int x, int y) {
        int count = tabbounds != null ? tabbounds.length : 0;
        for(int i = 0; i < count; i++)
            if(tabbounds[i] != null && tabbounds[i].contains(x, y))
                return i;
        return -1;
    }

    /** Establishment of a tab location
     *  tabbedpane - JTabbedPane component
     *  index      - index of tab in tabbedpane
     *  return     - bounds of tab
     */
    public Rectangle getTabBounds (JTabbedPane tabbedpane, int index) {
        if(tabbounds != null && index >= 0 && index < tabbounds.length)
            return tabbounds[index];
        return null;
    }

    /** Establishment of a tab lines
     *  tabbedpane - JTabbedPane component
     *  return     - 0, because tabs are not displayed
     */
    public int getTabRunCount (JTabbedPane tabbedpane) {
        return 1;
    }

//-----------------------------------------------------------------------------
// Methods - TabbedPaneUI APIs - free

    /** Creating a new instance of ArrowsTabbedPaneUI
     *  component - JTabbedPane component
     *  return    - new instance of ArrowsTabbedPaneUI
     */
    public static ComponentUI createUI (JComponent component) {
        return new ArrowsTabbedPaneUI();
    }

    /** Install PopupTabbedPaneUI attributes to the component
     *  component - JTabbedPane component where attributes are installed to
     */
    public void installUI (JComponent component) {
        tabbedpane = (JTabbedPane)component;
        layoutmanager = tabbedpane.getLayout();
        tabbedpane.setLayout(createLayoutManager());
        installListeners();
    }

    /** Uninstall PopupTabbedPaneUI attributes from the component
     *  component - JTabbedPane component where attributes are uninstalled from
     */
    public void uninstallUI (JComponent component) {
        tabbedpane.setLayout(layoutmanager);
        uninstallListeners();
        tabbedpane = null;
    }

    /** Paint JTabbedPane component
     *  graphics  - Graphics where JTabbedPane is painted to
     *  component - painted JTabbedPane component
     */
    public void paint (Graphics graphics, JComponent component) {
        paintTabs(graphics);
        paintBorder(graphics);
    }

//-----------------------------------------------------------------------------
// Methods

    /** Set the first displayed tab
     *  tabstart - index of the first displayed tab
     */
    private void setTabStart (int tabstart) {
        if(tabstart >= 0 && tabstart < tabbedpane.getTabCount()) {
            this.tabstart = tabstart;
            tabbedpane.setSelectedIndex(tabstart);
            tabbedpane.revalidate();
            tabbedpane.repaint();
        }
    }

    /** Create ArrowsTabbedPaneUI layout manager
     *  return - TabbedPaneLayout
     */
    private LayoutManager createLayoutManager () {
        return new TabbedPaneLayout();
    }

    /** Install listeners
     */
    private void installListeners () {
        tabbedpane.addMouseListener(mouseadapter);
    }

    /** Uninstall listeners
     */
    private void uninstallListeners () {
        tabbedpane.removeMouseListener(mouseadapter);
    }

    /** Paint tabs
     *  graphics - Graphics where tabs are painted to
     */
    private void paintTabs (Graphics graphics) {
        boolean display = true;
        int count = tabbedpane.getTabCount();
        int actual = tabbedpane.getSelectedIndex();
        int index = tabstart;
        FontMetrics fontmetrics = graphics.getFontMetrics();
        Dimension maxdimension = calculateTabsDimension(fontmetrics);
        Rectangle rectangle = new Rectangle(0, 0, tabbedpane.getWidth(), tabbedpane.getHeight());
        Rectangle tabrectangle = new Rectangle(TAB_SPACEX, TAB_SPACEY, 0, 0);
        tabbounds = new Rectangle[count];
        for(int i = 0; i < count; i++)
            if(i < tabstart)
                tabbounds[i] = null;
            else
                if(display && (display = calculateTabBounds(rectangle, i, maxdimension, fontmetrics, tabrectangle)) == true)
                    tabbounds[i] = new Rectangle(tabrectangle);
                else
                    tabbounds[i] = null;
        calculateTabsCorrection(rectangle, maxdimension, tabstart > 0, !display);
        while(index < count && tabbounds[index] != null) {
            paintTab(graphics, index, index == actual, tabbounds[index], fontmetrics);
            index++;
        }
        paintArrows(graphics, rectangle, maxdimension, tabstart > 0, !display);
    }

    /** Paint border around JTabbedPane component
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorder (Graphics graphics) {
        if(tabbounds != null && tabstart >= 0 && tabstart < tabbounds.length && tabbounds[tabstart] != null)
            switch(tabbedpane.getTabPlacement()) {
                case SwingConstants.BOTTOM:
                    paintBorderBottom(graphics);
                    break;
                case SwingConstants.LEFT:
                    paintBorderLeft(graphics);
                    break;
                case SwingConstants.RIGHT:
                    paintBorderRight(graphics);
                    break;
                default:
                    paintBorderTop(graphics);
            }
    }

    /** Paint border around JTabbedPane component with top placement
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorderTop (Graphics graphics) {
        int xstart = 0;
        int xend = tabbedpane.getWidth() - 1;
        int ystart = tabbounds[tabstart].y + tabbounds[tabstart].height;
        int yend = tabbedpane.getHeight() - 1;
        int xtabstart = tabbounds[tabstart].x;
        int xtabend = calculateTabBoundsHorizontalEnd();
        Color color = graphics.getColor();
        graphics.setColor(COLOR_BORDER_LEFTTOP);
        graphics.drawLine(xstart, ystart, xtabstart, ystart);
        graphics.drawLine(xtabend, ystart, xend, ystart);
        graphics.drawLine(xstart, ystart, xstart, yend);
        graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
        graphics.drawLine(xstart, yend, xend, yend);
        graphics.drawLine(xend, ystart, xend, yend);
        graphics.setColor(color);
    }

    /** Paint border around JTabbedPane component with bottom placement
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorderBottom (Graphics graphics) {
        int xstart = 0;
        int xend = tabbedpane.getWidth() - 1;
        int ystart = 0;
        int yend = tabbounds[tabstart].y;
        int xtabstart = tabbounds[tabstart].x;
        int xtabend = calculateTabBoundsHorizontalEnd();
        Color color = graphics.getColor();
        graphics.setColor(COLOR_BORDER_LEFTTOP);
        graphics.drawLine(xstart, ystart, xend, ystart);
        graphics.drawLine(xstart, ystart, xstart, yend);
        graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
        graphics.drawLine(xstart, yend, xtabstart, yend);
        graphics.drawLine(xtabend, yend, xend, yend);
        graphics.drawLine(xend, ystart, xend, yend);
        graphics.setColor(color);
    }

    /** Paint border around JTabbedPane component with left placement
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorderLeft (Graphics graphics) {
        int xstart = tabbounds[tabstart].x + tabbounds[tabstart].width;
        int xend = tabbedpane.getWidth() - 1;
        int ystart = 0;
        int yend = tabbedpane.getHeight() - 1;
        int ytabstart = tabbounds[tabstart].y;
        int ytabend = calculateTabBoundsVerticalEnd();
        Color color = graphics.getColor();
        graphics.setColor(COLOR_BORDER_LEFTTOP);
        graphics.drawLine(xstart, ystart, xend, ystart);
        graphics.drawLine(xstart, ystart, xstart, ytabstart);
        graphics.drawLine(xstart, ytabend, xstart, yend);
        graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
        graphics.drawLine(xstart, yend, xend, yend);
        graphics.drawLine(xend, ystart, xend, yend);
        graphics.setColor(color);
    }

    /** Paint border around JTabbedPane component with right placement
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorderRight (Graphics graphics) {
        int xstart = 0;
        int xend = tabbounds[tabstart].x;
        int ystart = 0;
        int yend = tabbedpane.getHeight() - 1;
        int ytabstart = tabbounds[tabstart].y;
        int ytabend = calculateTabBoundsVerticalEnd();
        Color color = graphics.getColor();
        graphics.setColor(COLOR_BORDER_LEFTTOP);
        graphics.drawLine(xstart, ystart, xend, ystart);
        graphics.drawLine(xstart, ystart, xstart, yend);
        graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
        graphics.drawLine(xstart, yend, xend, yend);
        graphics.drawLine(xend, ystart, xend, ytabstart);
        graphics.drawLine(xend, ytabend, xend, yend);
        graphics.setColor(color);
    }

    /** Establishment the end horizontal position of painted tabs
     *  return - the end horizontal position of last painted tab
     */
    private int calculateTabBoundsHorizontalEnd () {
        int i = tabstart;
        int count = tabbedpane.getTabCount();
        while(i < count && tabbounds[i] != null)
            i++;
        if(i == 0 || tabbounds == null || tabbounds[i - 1] == null)
            return 0;
        return tabbounds[i - 1].x + tabbounds[i - 1].width;
    }

    /** Establishment the end vertical position of painted tabs
     *  return - the end vertical position of last painted tab
     */
    private int calculateTabBoundsVerticalEnd () {
        int i = tabstart;
        int count = tabbedpane.getTabCount();
        while(i < count && tabbounds[i] != null)
            i++;
        if(i == 0 || tabbounds == null || tabbounds[i - 1] == null)
            return 0;
        return tabbounds[i - 1].y + tabbounds[i - 1].height;
    }

    /** Paint tab
     *  graphics     - Graphics where tab is painted to
     *  tabtext      - the title of the tab
     *  istabactual  - true, if tab is actual, otherwise false
     *  tabrectangle - bounds where tab is painted to
     *  fontmetrics  - FontMetrics of actual font
     */
    private void paintTab (Graphics graphics, int index, boolean istabactual, Rectangle tabrectangle, FontMetrics fontmetrics) {
        int iconendx;
        String text = tabbedpane.getTitleAt(index);
        Color color;
        if(!istabactual) {
            color = graphics.getColor();
            graphics.setColor(COLOR_BACKGROUND_TABNOACTUAL);
            graphics.fillRect(tabrectangle.x, tabrectangle.y, tabrectangle.width, tabrectangle.height);
            graphics.setColor(color);
        }
        paintTabBorder(graphics, istabactual, tabrectangle);
        iconendx = paintTabIcon(graphics, index, tabrectangle);
        graphics.drawString(text, iconendx + calculateCenter(tabrectangle.width + tabrectangle.x - iconendx, fontmetrics.stringWidth(text)) - 1, tabrectangle.y + fontmetrics.getHeight() + calculateCenter(tabrectangle.height, fontmetrics.getHeight() + fontmetrics.getDescent()) - fontmetrics.getDescent());
    }

    /** Paint border around tab
     *  graphics     - Graphics where border is painted to
     *  istabactual  - true, if tab is actual, otherwise false
     *  tabrectangle - bounds where border is painted to
     */
    private void paintTabBorder (Graphics graphics, boolean istabactual, Rectangle tabrectangle) {
        int placement = tabbedpane.getTabPlacement();
        paintTabBorderLineTop(graphics, istabactual, placement != SwingConstants.BOTTOM, tabrectangle);
        paintTabBorderLineBottom(graphics, istabactual, placement != SwingConstants.TOP, tabrectangle);
        paintTabBorderLineLeft(graphics, istabactual, placement != SwingConstants.RIGHT, tabrectangle);
        paintTabBorderLineRight(graphics, istabactual, placement != SwingConstants.LEFT, tabrectangle);
    }

    /** Paint top tab border's line
     *  graphics     - Graphics where border is painted to
     *  istabactual  - true, if tab is actual, otherwise false
     *  islineborder - false, if line is between tab and component, otherwise true
     *  tabrectangle - bounds where border is painted to
     */
    private void paintTabBorderLineTop (Graphics graphics, boolean istabactual, boolean islineborder, Rectangle tabrectangle) {
        int x1 = tabrectangle.x;
        int x2 = x1 + tabrectangle.width;
        int y = tabrectangle.y;
        Color color = graphics.getColor();
        Color colors[];
        if(!istabactual || islineborder) {
            if(istabactual)
                colors = new Color[] {COLOR_BORDER_RIGHTBOTTOM, COLOR_BORDER_LEFTTOP};
            else
                colors = new Color[] {COLOR_BORDER_RIGHTBOTTOM};
            for(int i = 0; i < colors.length; i++) {
                graphics.setColor(colors[i]);
                graphics.drawLine(x1 + i, y + i, x2 - i - 1, y + i);
            }
            graphics.setColor(color);
        }
    }

    /** Paint bottom tab border's line
     *  graphics     - Graphics where border is painted to
     *  istabactual  - true, if tab is actual, otherwise false
     *  islineborder - false, if line is between tab and component, otherwise true
     *  tabrectangle - bounds where border is painted to
     */
    private void paintTabBorderLineBottom (Graphics graphics, boolean istabactual, boolean islineborder, Rectangle tabrectangle) {
        Color color = graphics.getColor();
        if(!istabactual || islineborder) {
            graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
            graphics.drawLine(tabrectangle.x, tabrectangle.y + tabrectangle.height - 1, tabrectangle.x + tabrectangle.width - 1, tabrectangle.y + tabrectangle.height - 1);
            graphics.setColor(color);
        }
    }

    /** Paint left tab border's line
     *  graphics     - Graphics where border is painted to
     *  istabactual  - true, if tab is actual, otherwise false
     *  islineborder - false, if line is between tab and component, otherwise true
     *  tabrectangle - bounds where border is painted to
     */
    private void paintTabBorderLineLeft (Graphics graphics, boolean istabactual, boolean islineborder, Rectangle tabrectangle) {
        int x = tabrectangle.x;
        int y1 = tabrectangle.y;
        int y2 = y1 +  tabrectangle.height;
        Color color = graphics.getColor();
        Color colors[];
        if(!istabactual || islineborder) {
            if(istabactual)
                colors = new Color[] {COLOR_BORDER_RIGHTBOTTOM, COLOR_BORDER_LEFTTOP};
            else
                colors = new Color[] {COLOR_BORDER_RIGHTBOTTOM};
            for(int i = 0; i < colors.length; i++) {
                graphics.setColor(colors[i]);
                graphics.drawLine(x + i, y1 + i, x + i, y2 - i - 1);
            }
            graphics.setColor(color);
        }
    }

    /** Paint right tab border's line
     *  graphics     - Graphics where border is painted to
     *  istabactual  - true, if tab is actual, otherwise false
     *  islineborder - false, if line is between tab and component, otherwise true
     *  tabrectangle - bounds where border is painted to
     */
    private void paintTabBorderLineRight (Graphics graphics, boolean istabactual, boolean islineborder, Rectangle tabrectangle) {
        Color color = graphics.getColor();
        if(!istabactual || islineborder) {
            graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
            graphics.drawLine(tabrectangle.x + tabrectangle.width - 1, tabrectangle.y, tabrectangle.x + tabrectangle.width - 1, tabrectangle.y + tabrectangle.height - 1);
            graphics.setColor(color);
        }
    }

    /** Paint tab icon
     *  graphics - Graphics where icon is painted to
     *  tabrectangle - bounds where border is painted to
     */
    private int paintTabIcon (Graphics graphics, int index, Rectangle tabrectangle) {
        Icon icon = tabbedpane.getIconAt(index);
        if(icon == null)
            return tabrectangle.x;
        icon.paintIcon(tabbedpane, graphics, tabrectangle.x + SPACE_TAB_BORDER_LEFT, tabrectangle.y + SPACE_TAB_BORDER_TOP);
        return tabrectangle.x + icon.getIconWidth() + SPACE_TAB_BORDER_LEFT_TEXTICON;
    }

    /** Paint arrows to shift displayed tabs
     *  graphics      - Graphics where border is painted to
     *  rectangle     - 
     *  maxdimension  - 
     *  arrowprevious - true, if previous arrow might be displayed, otherwise false
     *  arrownext     - true, if next arrow might be displayed, otherwise false
     */
    private void paintArrows (Graphics graphics, Rectangle rectangle, Dimension maxdimension, boolean arrowprevious, boolean arrownext) {
        boolean ishorizontal;
        Point point1 = new Point(0, 0);
        Point point2 = new Point(0, 0);
        Point point3 = new Point(0, 0);
        Color color = graphics.getColor();
        ishorizontal = calculateArrowsPoints(rectangle, maxdimension, point1, point2, point3);
        graphics.setColor(COLOR_ARROW);
        paintArrowPrevious(graphics, arrowprevious && arrownext ? point1 : arrowprevious ? point2 : null, ishorizontal);
        paintArrowNext(graphics, arrownext ? point3 : null, ishorizontal);
        paintArrowList(graphics, arrownext ? point2 : arrowprevious ? point3 : null);
        graphics.setColor(color);
    }

    private void paintArrowPrevious (Graphics graphics, Point point, boolean ishorizontal) {
        if(point == null)
            boundsarrowprevious = null;
        else {
            boundsarrowprevious = new Rectangle(point.x, point.y, SPACE_ARROW_WIDTH, SPACE_ARROW_HEIGHT);
            if(ishorizontal)
                graphics.fillPolygon(new int[]{point.x + SPACE_ARROW_WIDTH, point.x + SPACE_ARROW_WIDTH, point.x}, new int[]{point.y, point.y + SPACE_ARROW_HEIGHT, point.y + ((int)(SPACE_ARROW_HEIGHT + .5) / 2)}, 3);
            else
                graphics.fillPolygon(new int[]{point.x + ((int)(SPACE_ARROW_WIDTH + .5) / 2), point.x + SPACE_ARROW_WIDTH, point.x}, new int[]{point.y, point.y + SPACE_ARROW_HEIGHT, point.y + SPACE_ARROW_HEIGHT}, 3);
        }
    }

    private void paintArrowNext (Graphics graphics, Point point, boolean ishorizontal) {
        if(point == null)
            boundsarrownext = null;
        else {
            boundsarrownext = new Rectangle(point.x, point.y, SPACE_ARROW_WIDTH, SPACE_ARROW_HEIGHT);
            if(ishorizontal)
                graphics.fillPolygon(new int[]{point.x, point.x + SPACE_ARROW_WIDTH, point.x}, new int[]{point.y, point.y + ((int)(SPACE_ARROW_HEIGHT + .5) / 2), point.y + SPACE_ARROW_HEIGHT}, 3);
            else
                graphics.fillPolygon(new int[]{point.x, point.x + SPACE_ARROW_WIDTH, point.x + ((int)(SPACE_ARROW_WIDTH + .5) / 2)}, new int[]{point.y, point.y, point.y + SPACE_ARROW_HEIGHT}, 3);
        }
    }

    private void paintArrowList (Graphics graphics, Point point) {
        if(point == null)
            boundsarrowlist = null;
        else {
            boundsarrowlist = new Rectangle(point.x, point.y, SPACE_ARROW_WIDTH, SPACE_ARROW_HEIGHT);
            graphics.fillRect(point.x, point.y, SPACE_ARROW_WIDTH, SPACE_ARROW_HEIGHT);
        }
    }

    private boolean calculateArrowsPoints (Rectangle rectangle, Dimension maxdimension, Point point1, Point point2, Point point3) {
        switch(tabbedpane.getTabPlacement()) {
            case SwingConstants.BOTTOM:
                return calculateArrowsPointsBottom(rectangle, maxdimension, point1, point2, point3);
            case SwingConstants.LEFT:
                return calculateArrowsPointsLeft(rectangle, maxdimension, point1, point2, point3);
            case SwingConstants.RIGHT:
                return calculateArrowsPointsRight(rectangle, maxdimension, point1, point2, point3);
            default:
                return calculateArrowsPointsTop(rectangle, maxdimension, point1, point2, point3);
        }
    }

    private boolean calculateArrowsPointsTop (Rectangle rectangle, Dimension maxdimension, Point point1, Point point2, Point point3) {
        point1.x = rectangle.x + rectangle.width - 3 * SPACE_ARROW_WIDTH - 2 * SPACE_ARROWS;
        point1.y = rectangle.y + calculateCenter(maxdimension.height, SPACE_ARROW_HEIGHT);
        point2.x = point1.x + SPACE_ARROW_WIDTH + SPACE_ARROWS;
        point2.y = point1.y;
        point3.x = point2.x + SPACE_ARROW_WIDTH + SPACE_ARROWS;
        point3.y = point1.y;
        return true;
    }

    private boolean calculateArrowsPointsBottom (Rectangle rectangle, Dimension maxdimension, Point point1, Point point2, Point point3) {
        point1.x = rectangle.x + rectangle.width - 3 * SPACE_ARROW_WIDTH - 2 * SPACE_ARROWS;
        point1.y = rectangle.y + rectangle.height - maxdimension.height + calculateCenter(maxdimension.height, SPACE_ARROW_HEIGHT);
        point2.x = point1.x + SPACE_ARROW_WIDTH + SPACE_ARROWS;
        point2.y = point1.y;
        point3.x = point2.x + SPACE_ARROW_WIDTH + SPACE_ARROWS;
        point3.y = point1.y;
        return true;
    }

    private boolean calculateArrowsPointsLeft (Rectangle rectangle, Dimension maxdimension, Point point1, Point point2, Point point3) {
        point1.x = rectangle.x + calculateCenter(maxdimension.width, SPACE_ARROW_WIDTH);
        point1.y = rectangle.y + rectangle.height - 3 * SPACE_ARROW_HEIGHT - 2 * SPACE_ARROWS;
        point2.x = point1.x;
        point2.y = point1.y + SPACE_ARROW_HEIGHT + SPACE_ARROWS;
        point3.x = point1.x;
        point3.y = point2.y + SPACE_ARROW_HEIGHT + SPACE_ARROWS;
        return false;
    }

    private boolean calculateArrowsPointsRight (Rectangle rectangle, Dimension maxdimension, Point point1, Point point2, Point point3) {
        point1.x = rectangle.x + rectangle.width - maxdimension.width + calculateCenter(maxdimension.width, SPACE_ARROW_WIDTH);
        point1.y = rectangle.y + rectangle.height - 3 * SPACE_ARROW_HEIGHT - 2 * SPACE_ARROWS;
        point2.x = point1.x;
        point2.y = point1.y + SPACE_ARROW_HEIGHT + SPACE_ARROWS;
        point3.x = point1.x;
        point3.y = point2.y + SPACE_ARROW_HEIGHT + SPACE_ARROWS;
        return false;
    }

    private Dimension calculateTabsDimension (FontMetrics fontmetrics) {
        int count = tabbedpane.getTabCount();
        int temp;
        Dimension dimension;
        if(fontmetrics == null)
            return new Dimension(0, 0);
        dimension = new Dimension(0, 0);
        for(int i = 0; i < count; i++) {
            if((temp = calculateTabWidth(i, fontmetrics)) > dimension.width)
                dimension.width = temp;
            if((temp = calculateTabHeight(i, fontmetrics)) > dimension.height)
                dimension.height = temp;
        }
        return dimension;
    }

// SPUR - nechodi
    private void calculateTabsCorrection (Rectangle rectangle, Dimension maxdimension, boolean arrowleft, boolean arrowright) {
        int count = tabbounds.length;
        int tabend = tabstart;
        Point point1 = new Point(0, 0);
        Point point2 = new Point(0, 0);
        Point point3 = new Point(0, 0);
        if(arrowleft || arrowright) {
            while(tabend < count && tabbounds[tabend] != null)
                tabend++;
            if(tabend == count || tabbounds[tabend] == null)
                tabend--;
            if(tabend >= tabstart && tabbounds[tabend] != null) {
                calculateArrowsPoints(rectangle, maxdimension, point1, point2, point3);
                switch(tabbedpane.getTabPlacement()) {
                    case SwingConstants.LEFT:
                    case SwingConstants.RIGHT:
                        calculateTabsCorrectionVertical(tabend, arrowleft && arrowright ? point1.x : point2.x);
                        break;
                    default:
                        calculateTabsCorrectionHorizontal(tabend, arrowleft && arrowright ? point1.y : point2.y);
                }
            }
        }
    }

    private void calculateTabsCorrectionHorizontal (int tabend, int maxx) {
//        tabbounds[tabend].width = maxx - tabbounds[tabend].x - 1;
    }

    private void calculateTabsCorrectionVertical (int tabend, int maxy) {
        
    }

    private int calculateTabWidth (int index,FontMetrics fontmetrics) {
        Icon icon = tabbedpane.getIconAt(index);
        return fontmetrics.stringWidth(tabbedpane.getTitleAt(index)) + SPACE_TAB_BORDER_LEFT_RIGHT + (icon == null ? 0 : icon.getIconWidth() + SPACE_TAB_TEXTICON);
    }

    private int calculateTabHeight (int index,FontMetrics fontmetrics) {
        int fontheight = fontmetrics.getHeight();
        int iconheight;
        Icon icon = tabbedpane.getIconAt(index);
        return (icon == null || fontheight > (iconheight = icon.getIconHeight()) ? fontheight : iconheight) + SPACE_TAB_BORDER_TOP_BOTTOM;
    }

    private boolean calculateTabBounds (Rectangle rectangle, int index, Dimension maxdimension, FontMetrics fontmetrics, Rectangle tabrectangle) {
        switch(tabbedpane.getTabPlacement()) {
            case SwingConstants.BOTTOM:
                return calculateTabBoundsBottom(rectangle, maxdimension, index, fontmetrics, tabrectangle);
            case SwingConstants.LEFT:
                return calculateTabBoundsLeft(rectangle, maxdimension, index, fontmetrics, tabrectangle);
            case SwingConstants.RIGHT:
                return calculateTabBoundsRight(rectangle, maxdimension, index, fontmetrics, tabrectangle);
        }
        return calculateTabBoundsTop(rectangle, maxdimension, index, fontmetrics, tabrectangle);
    }

    private boolean calculateTabBoundsTop (Rectangle rectangle, Dimension maxdimension, int index, FontMetrics fontmetrics, Rectangle tabrectangle) {
        tabrectangle.x = tabrectangle.x + tabrectangle.width;
        tabrectangle.width = calculateTabWidth(index, fontmetrics) - 2;
        tabrectangle.y = rectangle.y;
        tabrectangle.height = maxdimension.height + SPACE_TAB_BORDER_TOP_BOTTOM - 2;
        return canDisplayHorizontal(index, rectangle, tabrectangle);
    }

    private boolean calculateTabBoundsBottom (Rectangle rectangle, Dimension maxdimension, int index, FontMetrics fontmetrics, Rectangle tabrectangle) {
        tabrectangle.x = tabrectangle.x + tabrectangle.width;
        tabrectangle.width = calculateTabWidth(index, fontmetrics) - 2;
        tabrectangle.height = maxdimension.height + SPACE_TAB_BORDER_BOTTOM - 2;
        tabrectangle.y = rectangle.y + rectangle.height - tabrectangle.height;
        return canDisplayHorizontal(index, rectangle, tabrectangle);
    }

    private boolean calculateTabBoundsLeft (Rectangle rectangle, Dimension maxdimension, int index, FontMetrics fontmetrics, Rectangle tabrectangle) {
        tabrectangle.x = rectangle.x;
        tabrectangle.width = maxdimension.width + SPACE_TAB_BORDER_LEFT_RIGHT - 2;
        tabrectangle.y = tabrectangle.y + tabrectangle.height;
        tabrectangle.height = calculateTabHeight(index, fontmetrics) - 2;
        return canDisplayVertical(index, rectangle, tabrectangle);
    }

    private boolean calculateTabBoundsRight (Rectangle rectangle, Dimension maxdimension, int index, FontMetrics fontmetrics, Rectangle tabrectangle) {
        tabrectangle.width = maxdimension.width + SPACE_TAB_BORDER_LEFT_RIGHT - 2;
        tabrectangle.x = rectangle.x + rectangle.width - tabrectangle.width;
        tabrectangle.y = tabrectangle.y + tabrectangle.height;
        tabrectangle.height = calculateTabHeight(index, fontmetrics) - 2;
        return canDisplayVertical(index, rectangle, tabrectangle);
    }

    private boolean canDisplayHorizontal (int index, Rectangle rectangle, Rectangle tabrectangle) {
        int count = tabbedpane.getTabCount();
        int space = 0;
        if(tabstart > 0 && index + 1 < count)
            space = 3 * SPACE_ARROW_WIDTH + 2 * SPACE_ARROWS;
        else
            if(tabstart > 0 || index + 1 < count)
                space = 2 * SPACE_ARROW_WIDTH + SPACE_ARROWS;
        return tabrectangle.x + tabrectangle.width + space < rectangle.x + rectangle.width;
    }

    private boolean canDisplayVertical (int index, Rectangle rectangle, Rectangle tabrectangle) {
        int count = tabbedpane.getTabCount();
        int space = 0;
        if(tabstart > 0 && index + 1 < count)
            space = 3 * SPACE_ARROW_WIDTH + 2 * SPACE_ARROWS;
        else
            if(tabstart > 0 || index + 1 < count)
                space = 2 * SPACE_ARROW_WIDTH + SPACE_ARROWS;
        return tabrectangle.y + tabrectangle.height + space < rectangle.y + rectangle.height;
    }

    private int getComponentWidth (Component component, int sizetype) {
        switch(sizetype) {
            case SIZE_MINIMUM:
                return component.getMinimumSize().width;
            case SIZE_MAXIMUM:
                return component.getMaximumSize().width;
            case SIZE_PREFERRED:
                return component.getPreferredSize().width;
        }
        return 0;
    }

    private int getComponentHeight (Component component, int sizetype) {
        switch(sizetype) {
            case SIZE_MINIMUM:
                return component.getMinimumSize().height;
            case SIZE_MAXIMUM:
                return component.getMaximumSize().height;
            case SIZE_PREFERRED:
                return component.getPreferredSize().height;
        }
        return 0;
    }

    private Dimension calculateSize (int sizetype) {
// SPUR - upravit aby kontrolovalo i taby
        int count = tabbedpane.getComponentCount();
        int temp;
        Component component;
        Dimension dimensiontab = new Dimension(0, 0);
        Insets insets = tabbedpane.getInsets();
        for(int i = 0; i < count; i++) {
            if((temp = getComponentWidth(component = tabbedpane.getComponent(i), sizetype)) > dimensiontab.width)
                dimensiontab.width = temp;
            if((temp = getComponentHeight(component, sizetype)) > dimensiontab.height)
                dimensiontab.height = temp;
        }
        return dimensiontab;
    }

    private int calculateCenter (int length, int in) {
        if(in >= length)
            return 0;
        return (length - in) / 2;
    }

    private void displayTabList (int x, int y) {
        int count = tabbedpane.getTabCount();
        int actual = tabbedpane.getSelectedIndex();
        JRadioButton menuitem;
        if(popupmenu == null)
            popupmenu = new JPopupMenuPlus();
        popupmenu.removeAll();
        for(int i = 0; i < count; i++) {
            final int index = i;
            menuitem = new JRadioButton(tabbedpane.getTitleAt(i));
            menuitem.addActionListener(new ActionListener() {
                public void actionPerformed (ActionEvent event) {
                    popupmenu.setVisible(false);
                    setTabStart(index);
                }
            });
            if(i == actual)
                menuitem.setSelected(true);
            popupmenu.add(menuitem);
        }
        try {
            popupmenu.show(tabbedpane, x, y);
        } catch(Exception exception) {
            popupmenu.setLocation(x, y);
            popupmenu.setVisible(true);
        }
    }

//-----------------------------------------------------------------------------
// Class TabbedPaneLayout - Layout Manager

    private class TabbedPaneLayout implements LayoutManager {

        private TabbedPaneLayout () {
        }

        public void addLayoutComponent (String string, Component component) {
        }

        public void removeLayoutComponent (Component component) {
        }

        public Dimension minimumLayoutSize (Container container) {
            return calculateSize(SIZE_MINIMUM);
        }

        public Dimension maximumLayoutSize (Container container) {
            return calculateSize(SIZE_MAXIMUM);
        }

        public Dimension preferredLayoutSize (Container container) {
            return calculateSize(SIZE_PREFERRED);
        }

        public void layoutContainer (Container container) {
            int count;
            Component component;
            Dimension maxdimension;
            Insets insetstabbedpane;
            Rectangle rectangle;
            Rectangle boundstab = new Rectangle(0, 0, 0, 0);
            if(tabbedpane != null) {
                count = tabbedpane.getTabCount();
                maxdimension = calculateTabsDimension(tabbedpane.getGraphics() != null ? tabbedpane.getGraphics().getFontMetrics() : null);
                if((insetstabbedpane = tabbedpane.getInsets()) == null)
                    insetstabbedpane = new Insets(0, 0, 0, 0);
                rectangle = new Rectangle(TAB_SPACEX + insetstabbedpane.left, TAB_SPACEY + insetstabbedpane.top, tabbedpane.getWidth() - TAB_2SPACEX - insetstabbedpane.left - insetstabbedpane.right, tabbedpane.getHeight() - TAB_2SPACEY - insetstabbedpane.top - insetstabbedpane.bottom);
                for(int i = 0; i < count; i++)
                    tabbedpane.getComponent(i).setVisible(false);
                if((component = tabbedpane.getSelectedComponent()) != null) {
                    switch(tabbedpane.getTabPlacement()) {
                        case SwingConstants.BOTTOM:
                            calculateBoundsBottom(rectangle, maxdimension, boundstab);
                            break;
                        case SwingConstants.LEFT:
                            calculateBoundsLeft(rectangle, maxdimension, boundstab);
                            break;
                        case SwingConstants.RIGHT:
                            calculateBoundsRight(rectangle, maxdimension, boundstab);
                            break;
                        default:
                            calculateBoundsTop(rectangle, maxdimension, boundstab);
                    }
                    component.setVisible(true);
                    component.setBounds(boundstab);
                }
            }
        }

        private void calculateBoundsTop (Rectangle rectangle, Dimension maxdimension, Rectangle boundstab) {
            boundstab.x = rectangle.x;
            boundstab.width = rectangle.width;
            boundstab.y = rectangle.y + maxdimension.height + SPACE_TAB;
            boundstab.height = rectangle.height - boundstab.y;
        }

        private void calculateBoundsBottom (Rectangle rectangle, Dimension maxdimension, Rectangle boundstab) {
            boundstab.x = rectangle.x;
            boundstab.width = rectangle.width;
            boundstab.y = rectangle.y;
            boundstab.height = rectangle.height - boundstab.y - maxdimension.height - SPACE_TAB;
        }

        private void calculateBoundsLeft (Rectangle rectangle, Dimension maxdimension, Rectangle boundstab) {
            boundstab.x = rectangle.x + maxdimension.width + SPACE_TAB;
            boundstab.width = rectangle.width - boundstab.x;
            boundstab.y = rectangle.y;
            boundstab.height = rectangle.height;
        }

        private void calculateBoundsRight (Rectangle rectangle, Dimension maxdimension, Rectangle boundstab) {
            boundstab.x = rectangle.x;
            boundstab.width = rectangle.width - boundstab.x - maxdimension.width - SPACE_TAB;
            boundstab.y = rectangle.y;
            boundstab.height = rectangle.height;
        }
    }

//-----------------------------------------------------------------------------
// Class TabbedPaneMouseAdapter - Mouse Listener

    private class TabbedPaneMouseAdapter extends MouseAdapter {

        public void mousePressed (MouseEvent event) {
            boolean dorepaint = false;
            int x = event.getX();
            int y = event.getY();
            int index;
            if((event.getModifiers() & MouseEvent.BUTTON1_MASK) > 0 && boundsarrowprevious != null && boundsarrowprevious.contains(x, y)) {
                tabstart--;
                dorepaint = true;
            } else {
                if((event.getModifiers() & MouseEvent.BUTTON1_MASK) > 0 && boundsarrownext != null && boundsarrownext.contains(x, y)) {
                    tabstart++;
                    dorepaint = true;
                } else {
                    if((event.getModifiers() & MouseEvent.BUTTON1_MASK) > 0 && boundsarrowlist != null && boundsarrowlist.contains(x, y)) {
                        displayTabList(x, y);
                    } else {
                        if((index = tabForCoordinate(tabbedpane, x, y)) >= 0) {
                            tabbedpane.setSelectedIndex(index);
                            dorepaint = true;
                        } else {
                        }
                    }
                }
            }
            if(dorepaint) {
                tabbedpane.revalidate();
                tabbedpane.repaint();
            }
        }
    }
}
