/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JComponent;
import javax.swing.JRadioButton;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.TabbedPaneUI;

import org.openide.awt.JPopupMenuPlus;

/** User interface for javax.swing.JTabbedPane component including special line.
 *  Left-click to the line displays a popup menu includes tabs and susceptible change an actual tab.
 *  @author  Rostislav Levy
 *  @version 0.9
 */

public class PopupTabbedPaneUI extends TabbedPaneUI {

//-----------------------------------------------------------------------------
// Constants

    /** Type of probed size - Minimum size
     */
    private static final int SIZE_MINIMUM = 1;
    /** Type of probed size - Maximum size
     */
    private static final int SIZE_MAXIMUM = 2;
    /** Type of probed size - Preferred size
     */
    private static final int SIZE_PREFERRED = 3;

    /** Horizontal space
     */
    private static final int TAB_SPACEX = 5;
    /** Vertical space
     */
    private static final int TAB_SPACEY = 5;
    /** Width of special line
     */
    private static final int TAB_WIDTH = 3;
    /** Height of special line
     */
    private static final int TAB_HEIGHT = 3;
    /** Helpful property
     */
    private static final int TAB_SPACEX_WIDTH = TAB_SPACEX + TAB_WIDTH;
    /** Helpful property
     */
    private static final int TAB_2SPACEX = 2 * TAB_SPACEX;
    /** Helpful property
     */
    private static final int TAB_2SPACEX_WIDTH = TAB_SPACEX_WIDTH + TAB_SPACEX;
    /** Helpful property
     */
    private static final int TAB_3SPACEX_WIDTH = TAB_2SPACEX_WIDTH + TAB_SPACEX;
    /** Helpful property
     */
    private static final int TAB_SPACEY_HEIGHT = TAB_SPACEY + TAB_HEIGHT;
    /** Helpful property
     */
    private static final int TAB_2SPACEY = 2 * TAB_SPACEY;
    /** Helpful property
     */
    private static final int TAB_2SPACEY_HEIGHT = TAB_SPACEY_HEIGHT + TAB_SPACEY;
    /** Helpful property
     */
    private static final int TAB_3SPACEY_HEIGHT = TAB_2SPACEY_HEIGHT + TAB_SPACEY;

    /** Color of special line
     */
    private static final Color COLOR_LINE = Color.black;
    /** Color of left and upper border
     */
    private static final Color COLOR_BORDER_LEFTTOP = Color.white;
    /** Color of right and lower border
     */
    private static final Color COLOR_BORDER_RIGHTBOTTOM = new Color(102, 102, 102);

//-----------------------------------------------------------------------------
// Attributes

    /** Displayed JTabbedPane component
     */
    private JTabbedPane tabbedpane;
    /** Displaying JPopupMenu includes list of tabs
     */
    private JPopupMenuPlus popupmenu;
    /** Previous JTabbedPane's layout managet
     */
    private LayoutManager layoutmanager;
    /** Location of special line
     */
    private Rectangle tabbounds;
    /** New JTabbedPane's Mouse Listener
     */
    private TabbedPaneMouseAdapter mouselistener;

//-----------------------------------------------------------------------------
// Constructors

    /** Default constructor
     */
    public PopupTabbedPaneUI() {
        tabbedpane = null;
        popupmenu = null;
        layoutmanager = null;
        tabbounds = new Rectangle(0, 0, 0, 0);
        mouselistener = new TabbedPaneMouseAdapter();
    }

//-----------------------------------------------------------------------------
// Methods - TabbedPaneUI APIs - required

    /** Establishment of tab on a position
     *  tabbedpane - JTabbedPane component
     *  x          - horizontal position
     *  y          - vertical position
     *  return     - selected tab, because tabs are not displayed
     */
    public int tabForCoordinate (JTabbedPane tabbedpane, int x, int y) {
        return this.tabbedpane.getSelectedIndex();
    }

    /** Establishment of a tab location
     *  tabbedpane - JTabbedPane component
     *  index      - index of tab in tabbedpane
     *  return     - null, because tabs are not displayed
     */
    public Rectangle getTabBounds (JTabbedPane tabbedpane, int index) {
        return null;
    }

    /** Establishment of a tab lines
     *  tabbedpane - JTabbedPane component
     *  return     - 0, because tabs are not displayed
     */
    public int getTabRunCount (JTabbedPane tabbedpane) {
        return 0;
    }

//-----------------------------------------------------------------------------
// Methods - TabbedPaneUI APIs - free

    /** Creating a new instance of PopupTabbedPaneUI
     *  component - JTabbedPane component
     *  return    - new instance of PopupTabbedPaneUI
     */
    public static ComponentUI createUI (JComponent component) {
        return new PopupTabbedPaneUI();
    }

    /** Install PopupTabbedPaneUI attributes to the component
     *  component - JTabbedPane component where attributes are installed to
     */
    public void installUI (JComponent component) {
        tabbedpane = (JTabbedPane)component;
        layoutmanager = tabbedpane.getLayout();
        tabbedpane.setLayout(createLayoutManager());
        installListeners();
    }

    /** Uninstall PopupTabbedPaneUI attributes from the component
     *  component - JTabbedPane component where attributes are uninstalled from
     */
    public void uninstallUI (JComponent component) {
        tabbedpane.setLayout(layoutmanager);
        uninstallListeners();
        tabbedpane = null;
    }

    /** Paint JTabbedPane component
     *  graphics  - Graphics where JTabbedPane is painted to
     *  component - painted JTabbedPane component
     */
    public void paint (Graphics graphics, JComponent component) {
        paintTabs(graphics);
        paintBorder(graphics);
    }

//-----------------------------------------------------------------------------
// Methods

    /** Create PopupTabbedPaneUI layout manager
     *  return - TabbedPaneLayout
     */
    private LayoutManager createLayoutManager () {
        return new TabbedPaneLayout();
    }

    /** Install listeners
     */
    private void installListeners () {
        tabbedpane.addMouseListener(mouselistener);
    }

    /** Uninstall listeners
     */
    private void uninstallListeners () {
        tabbedpane.removeMouseListener(mouselistener);
    }

    /** Paint a special line
     *  graphics - Graphics where special line is painted to
     */
    private void paintTabs (Graphics graphics) {
        int placement = tabbedpane.getTabPlacement();
        int width = tabbedpane.getWidth();
        int height = tabbedpane.getHeight();
        Color color=graphics.getColor();
        tabbounds.setBounds(
            placement == SwingConstants.RIGHT ? width - TAB_SPACEX_WIDTH : TAB_SPACEX,
            placement == SwingConstants.BOTTOM ? height - TAB_SPACEY_HEIGHT : TAB_SPACEY,
            placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? width - TAB_2SPACEX : TAB_WIDTH,
            placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? TAB_HEIGHT : height - TAB_2SPACEY);
        graphics.setColor(COLOR_LINE);
        graphics.fillRect(tabbounds.x ,tabbounds.y, tabbounds.width, tabbounds.height);
        graphics.setColor(color);
    }

    /** Paint border around JTabbedPane component
     *  graphics - Graphics where JTabbedPane is painted to
     */
    private void paintBorder (Graphics graphics) {
        int width = tabbedpane.getWidth() - 1;
        int height = tabbedpane.getHeight() - 1;
        Color color=graphics.getColor();
        graphics.setColor(COLOR_BORDER_LEFTTOP);
        graphics.drawLine(0, 0, width, 0);
        graphics.drawLine(0, 0, 0, height);
        graphics.setColor(COLOR_BORDER_RIGHTBOTTOM);
        graphics.drawLine(0, height, width, height);
        graphics.drawLine(width, 0, width, height);
        graphics.setColor(color);
    }

    /** Establishment of JTabbedPane component size
     *  sizetype - type of establising size (constants SIZE_...)
     *  return   - size of JTabbedPane component
     */
    private Dimension calculateSize (int sizetype) {
        int count = tabbedpane.getTabCount();
        int placement = tabbedpane.getTabPlacement();
        Dimension dimension = new Dimension(0, 0);
        Dimension dimensioncomponent;
        for(int i = 0; i < count; i++) {
            dimensioncomponent = getComponentSize(tabbedpane.getComponent(i), sizetype);
            if(dimensioncomponent.width > dimension.width)
                dimension.width = dimensioncomponent.width;
            if(dimensioncomponent.height > dimension.height)
                dimension.height = dimensioncomponent.height;
        }
        dimension.width += (placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? TAB_2SPACEX : TAB_3SPACEX_WIDTH);
        dimension.height += (placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? TAB_3SPACEY_HEIGHT : TAB_2SPACEY);
        return dimension;
    }

    /** Establisment of component size
     *  component - establishing component
     *  sizetype  - type of establising size (constants SIZE_...)
     *  return    - size of component
     */
    private Dimension getComponentSize (Component component, int sizetype) {
        switch(sizetype) {
            case SIZE_MINIMUM:
                return component.getMinimumSize();
            case SIZE_MAXIMUM:
                return component.getMaximumSize();
        }
        return component.getPreferredSize();
    }

//-----------------------------------------------------------------------------
// Class TabbedPaneLayout - Layout Manager

    /** Layout manager for PopuTabbedPaneUI
     */
    private class TabbedPaneLayout implements LayoutManager {

        /** Default constructor
         */
        private TabbedPaneLayout () {
        }

        /** Add component to the layout manager
         *  string    - constraint
         *  component - adding component
         */
        public void addLayoutComponent (String string, Component component) {
        }

        /** Remove component from the layout manager
         *  component - removing component
         */
        public void removeLayoutComponent (Component component) {
        }

        /** Establishment of minimum size of component
         *  container - established component
         *  return    - minimum size of the component
         */
        public Dimension minimumLayoutSize (Container container) {
            return calculateSize(SIZE_MINIMUM);
        }

        /** Establishment of maximum size of component
         *  container - established component
         *  return    - maximum size of the component
         */
        public Dimension maximumLayoutSize (Container container) {
            return calculateSize(SIZE_MAXIMUM);
        }

        /** Establishment of preferred size of component
         *  container - established component
         *  return    - preferred size of the component
         */
        public Dimension preferredLayoutSize (Container container) {
            return calculateSize(SIZE_PREFERRED);
        }

        /** Layout a component
         *  container - layouted component
         */
        public void layoutContainer (Container container) {
            int count = tabbedpane.getTabCount();
            int placement = tabbedpane.getTabPlacement();
            int width = tabbedpane.getWidth();
            int height = tabbedpane.getHeight();
            Component component;
            Rectangle bounds = new Rectangle();
            for(int i = 0; i < count; i++)
                tabbedpane.getComponent(i).setVisible(false);
            if((component = tabbedpane.getSelectedComponent()) != null) {
                component.setBounds(
                    placement == SwingConstants.LEFT ? TAB_2SPACEX_WIDTH : TAB_SPACEX,
                    placement == SwingConstants.TOP ? TAB_2SPACEY_HEIGHT : TAB_SPACEY,
                    width - (placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? TAB_2SPACEX : TAB_3SPACEX_WIDTH),
                    height - (placement == SwingConstants.TOP || placement == SwingConstants.BOTTOM ? TAB_3SPACEY_HEIGHT : TAB_2SPACEY));
                component.setVisible(true);
            }
        }
    }

//-----------------------------------------------------------------------------
// Class TabbedPaneMouseAdapter - Mouse Listener

    /** Mouse adapter for PopupTabbedPaneUI
     */
    private class TabbedPaneMouseAdapter extends MouseAdapter {

        /** Mouse clicked
         *  event - mouse event
         */
        public void mouseClicked (MouseEvent event) {
            int x;
            int y;
            if((event.getModifiers() & MouseEvent.BUTTON1_MASK) > 0 && tabbounds.contains(x = event.getX(), y = event.getY()))
                displayTabList(x, y);
        }

        /** Display a popup menu includes list of tabs
         *  x - horizontal location of popup menu
         *  y - vertical location of popup menu
         */
        private void displayTabList (int x,int y) {
            int count = tabbedpane.getTabCount();
            int actual = tabbedpane.getSelectedIndex();
            int lengthtabbedpane;
            int lengthpopup;
            JRadioButton menuitem;
            if(popupmenu == null)
                popupmenu = new JPopupMenuPlus();
            popupmenu.removeAll();
            for(int i = 0; i < count; i++) {
                final int index = i;
                menuitem = new JRadioButton(tabbedpane.getTitleAt(i));
                menuitem.addActionListener(new ActionListener() {
                    public void actionPerformed (ActionEvent event) {
                        popupmenu.setVisible(false);
                        tabbedpane.setSelectedIndex(index);
                        tabbedpane.revalidate();
                        tabbedpane.repaint();
                    }
                });
                if(i == actual)
                    menuitem.setSelected(true);
                popupmenu.add(menuitem);
            }
            try {
                popupmenu.show(tabbedpane, x, y);
            } catch(Exception exception) {
                popupmenu.setLocation(x, y);
                popupmenu.setVisible(true);
            }
        }
    }
}
