/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.ui;

import javax.swing.event.*;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.table.TableColumn;
import javax.swing.tree.TreePath;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.*;

import org.openide.explorer.view.BeanTreeView;
import org.openide.explorer.view.Visualizer;
import org.openide.explorer.ExplorerManager;
import org.openide.util.WeakListener;
import org.openide.nodes.*;

/** View compounded from - tree view, table view. Allows to view tree of nodes on the left
 * and its properties in table on the right.
 *
 * @author  jrojcek
 * @version 
 */
public class TreeTableView extends BeanTreeView {
        
    /** flag that allows group more requests into one */
    private boolean tableChanging = false;
    /** manager is used only for synchronization of header name */
    private ExplorerManager manager;
    /** listener on explorer manager */
    private PropertyChangeListener wlpc;
    /** table view that is controlled by tree view */
    private TableView.ControlledTableView controlledTableView;
    /** listener on changes in tree and its model */
    private Listener listener;

    /** Create TreeTableView with default NodeTableModel
     */
    public TreeTableView() {
        this(null);
    }

    /** Creates TreeTableView with provided NodeTableModel.
     * @param ntm node table model
     */
    public TreeTableView(NodeTableModel ntm) {
        tree.putClientProperty("JTree.lineStyle", "None"); // NOI18N
        tree.setUI(new TableLikeTreeUI());
        // do not use scroll bars, this scroll pane is dummy, only for border painting
        setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        // remove tree from viewportview
        setViewportView(null);

        // insert tree into new scrollpane
        JScrollPane treeView = new JScrollPane(tree);
        treeView.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        treeView.setBorder(null);
        
        // create table view controlled by new scrollpane
        controlledTableView = ntm == null 
                              ? new TableView.ControlledTableView(treeView)
                              : new TableView.ControlledTableView(treeView, ntm);
        // hack to avoid moving of compound view
        setLayout(new TableView.MyScrollPaneLayout());
        // add compound scrollpane (JPanel) into this scrollpane
        add(TableView.MyScrollPaneLayout.COMPOUND_SCROLLPANE,
            controlledTableView.compoundScrollPane());

        listener = new Listener();
        changeTableModel();
    }
    /** Set columns.
     * @param nodes columns
     */
    public void setProperties(Node.Property[] props) {
        controlledTableView.setProperties(props);
    }

    /** Sets resize mode of table.
     * 
     * @param mode - One of 5 legal values: <pre>JTable.AUTO_RESIZE_OFF,
     *                                           JTable.AUTO_RESIZE_NEXT_COLUMN,
     *                                           JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS, 
     *                                           JTable.AUTO_RESIZE_LAST_COLUMN, 
     *                                           JTable.AUTO_RESIZE_ALL_COLUMNS</pre>
     */    
    public void setAutoResizeMode(int mode) {
        controlledTableView.setAutoResizeMode(mode);
    }

    /** Allows set parameters of column.
     * @param i column index
     * @return table column avaliable to set sizes of columns, ...
     */    
    public TableColumn getTableColumn(int columnIndex) {
        return controlledTableView.getTableColumn(columnIndex);
    }
    
    /** Set preferred size of tree view
     * @param dim
     */    
    public void setTreeViewPreferredSize(Dimension dim) {
        controlledTableView.setControllingViewPreferredSize(dim);
    }
    
    /** Set preferred size of table view
     * @param dim preferred size
     */    
    public void setTableViewPreferredSize(Dimension dim) {
        controlledTableView.setPreferredSize(dim);
    }

    public void addNotify() {
        super.addNotify();

        ExplorerManager newManager = ExplorerManager.find (this);
        if (newManager != manager) {
            if (manager != null) {
                manager.removePropertyChangeListener (wlpc);
            }
            manager = newManager;
            manager.addPropertyChangeListener(wlpc = WeakListener.propertyChange (listener, manager));
            controlledTableView.setHeaderText(manager.getRootContext().getDisplayName());
        }
        tree.addTreeExpansionListener(listener);
        tree.getModel().addTreeModelListener(listener);
        changeTableModel();
    }

    public void removeNotify() {
        super.removeNotify();
        tree.removeTreeExpansionListener(listener);
        tree.getModel().removeTreeModelListener(listener);
        // clear node listeners
        controlledTableView.setNodes(new Node[] {});
    }

    /** Chnage table data in awt thread */
    private void delayedFireTableDataChanged() {
        if (tableChanging)
            return;
        tableChanging = true;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                if (tree.getRowBounds(0) != null)
                    controlledTableView.setRowHeight(tree.getRowBounds(0).height);
                changeTableModel();
                tableChanging = false;
            }
        });
    }

    /** Change table model. Sets rows (nodes) of table mmodel */
    private void changeTableModel() {
        Node[] nodes = new Node[tree.getRowCount()];
        for (int i = 0; i < tree.getRowCount(); i++) {
            nodes[i] = Visualizer.findNode(tree.getPathForRow(i).getLastPathComponent());
        }
        controlledTableView.setNodes(nodes);
    }

    /** Listenes on changes in tree and its model.
     */
    private class Listener implements TreeExpansionListener, 
                                      TreeModelListener,
                                      PropertyChangeListener {

        public void treeExpanded(TreeExpansionEvent e) {
            changeTableModel();
        }

        public void treeCollapsed(TreeExpansionEvent e) {
            changeTableModel();
        }

        public void treeNodesChanged(TreeModelEvent e) {
            delayedFireTableDataChanged();
        }

        public void treeNodesInserted(TreeModelEvent e) {
            delayedFireTableDataChanged();
        }

        public void treeNodesRemoved(TreeModelEvent e) {
            delayedFireTableDataChanged();
        }

        public void treeStructureChanged(TreeModelEvent e) {
            delayedFireTableDataChanged();
        }

        /** Updates header name.
         * @param evt event
         */
        public void propertyChange(PropertyChangeEvent evt) {
            if (ExplorerManager.PROP_ROOT_CONTEXT.equals(evt.getPropertyName()))
                controlledTableView.setHeaderText(manager.getExploredContext().getDisplayName());
        }
    }
    
    /** Overriden to paint row separator lines
     */
    private class TableLikeTreeUI extends javax.swing.plaf.metal.MetalTreeUI {

        protected void paintRow(Graphics g, Rectangle clipBounds,
                                Insets insets, Rectangle bounds, TreePath path,
                                int row, boolean isExpanded,
                                boolean hasBeenExpanded, boolean isLeaf) {
            super.paintRow(g, clipBounds, insets, bounds, path, row, isExpanded,
                           hasBeenExpanded, isLeaf);
            Graphics cg = g.create(0, bounds.y, super.tree.getSize().width, bounds.height);
            try {
                cg.setColor(SystemColor.controlDkShadow);
                cg.drawLine(0, bounds.height - 1, super.tree.getSize().width - 1, bounds.height - 1);
            }
            finally {
                cg.dispose();
            }
        }
    }
/*    
    public static void main (String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                Node n = new org.netbeans.core.ModuleNode();
                //org.openide.TopManager.getDefault ().getPlaces ().nodes ().repository ();
                TreeTableView ttv = new TreeTableView();
//                ttv.setAutoResizeMode(javax.swing.JTable.AUTO_RESIZE_OFF);
                ttv.setProperties(
//                    n.getChildren().getNodes()[0].getPropertySets()[0].getProperties());
                    new Node.Property[]{
                        new PropertySupport.ReadWrite (
                            "enabled", // NOI18N
                            Boolean.TYPE,
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "PROP_modules_enabled"),
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "HINT_modules_enabled")
                        ) {
                            public Object getValue () {
                                return null;
                            }

                            public void setValue (Object o) {
                            }
                        },
                        new PropertySupport.ReadOnly (
                            "specVersion", // NOI18N
                            String.class,
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "PROP_modules_specversion"),
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "HINT_modules_specversion")
                        ) {
                            public Object getValue () {
                                return null;
                            }

                        }
                    }
                );

                org.openide.explorer.ExplorerManager em = new org.openide.explorer.ExplorerManager();
                em.setRootContext(n);
                org.openide.explorer.ExplorerPanel ep = new org.openide.explorer.ExplorerPanel (em);
                ep.setLayout (new java.awt.BorderLayout ());
                ep.add ("Center", ttv);
                ep.open ();
            }
        });
    }
*/

}
