/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.upgrade;

import java.io.*;
import java.util.*;
import java.beans.PropertyVetoException;
import org.openide.util.*;
import java.util.jar.*;

import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem;
import org.openide.execution.NbClassPath;

import org.openide.util.NbBundle;

/** Basic copy utilities.
 *
 * @author Petr Zajac, 
 */
public abstract class CopyUtil extends Object {
    /* time deviation in ms
     */
    private static final int timeDev = 100000;

    private int items;
    private int maxItems;

    /** Previous names of the IDE.
     */
    static String versions[];

    static String ffj20IE = "Forte(tm) for Java(tm), release 2.0, Internet Edition";
    
    /** Bundles with files that should not be copied.
     */
    private static String nonCpFiles[];

    static {
        if (isForte()) {
            versions = new String [] {
                "NetBeans IDE v3.0",
                "Forte(tm) for Java(tm), release 2.0, Community Edition",
                ffj20IE,
                "NetBeans IDE 3.1", 
                "NetBeans IDE, Development Version"
            };
            nonCpFiles =  new String [] {
                "netbeans30.cfg",
                "forte4j20CE.cfg",
                "forte4j20IE.cfg",
                "netbeans31.cfg",
                "nbdev.cfg"
            };
        } else {
            // import from FFJ 2.0 IE -> Nb isn't allowed
            versions = new String [] {
                "NetBeans IDE v3.0",
                "Forte(tm) for Java(tm), release 2.0, Community Edition",
                "NetBeans IDE 3.1", 
                "NetBeans IDE, Development Version"
            };
            nonCpFiles =  new String [] {
                "netbeans30.cfg",
                "forte4j20CE.cfg",
                "netbeans31.cfg",
                "nbdev.cfg"
            };
        }
    }
    
    /** Subclasses should provide reasonable handling of messages.
     * @param msg message to print
     */
    protected abstract void message (String msg);

    /** Subclasses should provide reasonable handling of progress status.
     * @param current number of items already processed
     * @param max total number of items being processed
     */
    protected abstract void progress (int current, int max);

    /** Updates the IDE.
     * @param sourceDir original instalation of the IDE
     * @param targetSystem target system to copy files to
     * @param backupSystem filesystem to do backupSystemFo to (or null)
     * @exception IOException if the copying fails
     */
    protected final void updateIde (
        File sourceDir, FileSystem targetSystem, FileSystem backupSystem
    ) throws Exception {
        
        int version = getIdeVersion (sourceDir);
        updateIde_internal(version, sourceDir, targetSystem, backupSystem, -1);
    }

    protected final void updateIde (
        int version, File sourceDir, FileSystem targetSystem, FileSystem backupSystem
    ) throws Exception {
        
        updateIde_internal(version, sourceDir, targetSystem, backupSystem, -timeDev);
    }

    private final void updateIde_internal(
        int version, File sourceDir, FileSystem targetSystem, FileSystem backupSystem, long sourceBaseTime
    ) throws Exception {
        
        if (version < 0 || version >= versions.length) {
            message (getString("MSG_BAD_IDE"));
            for (int i = 0 ; i < versions.length ; i++ ) {
                message (versions[i]);
            }
            throw new Exception("Invalid IDE version"); // NOI18N
        }

        message (getString("MSG_UPDATE_FROM", versions[version] ));
        
        // create old/system FileSystem
        //
        if (sourceDir.exists() == false || sourceDir.isDirectory() == false ) {
            message (getString("MSG_NO_OLD_HOME") );
            throw new Exception("Old IDE home directory not found."); // NOI18N
        }
        File systemDirFile = new File (sourceDir, "system"); // NOI18N
        if (systemDirFile.exists() == false || systemDirFile.isDirectory() == false ) {
            message (getString("MSG_NO_OLD_SYSTEM"));
            throw new Exception("Old IDE system directory not found."); // NOI18N
        }

        if (backupSystem != null) {
            copyAttributes(targetSystem.getRoot (), backupSystem.getRoot ());
            recursiveCopy(targetSystem.getRoot (), backupSystem.getRoot ());
        }
        
        try {
            long tm = -1 == sourceBaseTime ? getLastModified(sourceDir) : sourceBaseTime;
            update(version, systemDirFile, targetSystem, tm);
        }
        catch (Exception e) {
            if (backupSystem != null) {
                message(getString("MSG_recovery_started"));
                deleteAll(targetSystem.getRoot());
                copyAttributes(backupSystem.getRoot (), targetSystem.getRoot ());
                recursiveCopy(backupSystem.getRoot (), targetSystem.getRoot ());
                message(getString("MSG_recovery_finished"));
            }
            throw e;
        }
    }
    
    private void update(
        int version, File systemDirFile, FileSystem targetSystem, long sourceBaseTime
    ) throws IOException {
        LocalFileSystem sourceSystem = new LocalFileSystem();
        try {
            sourceSystem.setRootDirectory(systemDirFile);
        } catch (PropertyVetoException ex) {
            ex.printStackTrace();
        }
        
        items = 0;
        maxItems = 0;

        copyAttributes(sourceSystem.getRoot (),targetSystem.getRoot ());
        recursiveCopyWithFilter(
            sourceSystem.getRoot (),
            targetSystem.getRoot (),
            originalFiles (version),
            sourceBaseTime
        );

        // tell the filesystem to clean itself
        sourceSystem.removeNotify ();
    }
    
    /** copies recursively directory, skips files existing in target location
     *  @param source source directory
     *  @param dest destination directory
     */
    private void recursiveCopy (FileObject sourceFolder, FileObject destFolder) throws IOException {
        FileObject childrens []  = sourceFolder.getChildren();
        for (int i = 0 ; i < childrens.length ; i++ ) {
            final FileObject subSourceFo = childrens[i];
            FileObject subTargetFo = null;
            
            if (subSourceFo.isFolder()) {
                subTargetFo =  destFolder.getFileObject(subSourceFo.getName());
                if (subTargetFo == null) {
                    subTargetFo = destFolder.createFolder(subSourceFo.getName());
                    
                }
                copyAttributes(subSourceFo,subTargetFo);
                recursiveCopy(subSourceFo,subTargetFo);
            } else {
                subTargetFo =  destFolder.getFileObject(subSourceFo.getNameExt());
                if (subTargetFo == null) {
                    subTargetFo = FileUtil.copyFile(subSourceFo, destFolder, subSourceFo.getName(), subSourceFo.getExt());
                }
                copyAttributes(subSourceFo,subTargetFo);
            }
        }
    }
    
    /** Copies recursively dircectory. Files are copied when when basicTime + timeDev < time of file.
     *  @param source source directory
     *  @param #dest destination dirctory
     */
    private void recursiveCopyWithFilter (
        FileObject source, FileObject dest, String[] noCopy, long basicTime
    ) throws IOException {
        FileObject childrens []  = source.getChildren();
        if (source.isFolder() == false ) {
            message (getString("MSG_IS_NOT_FOLDER", source.getName()));
        }

        // adjust max number of items
        maxItems += childrens.length;

        for (int i = 0 ; i < childrens.length ; i++ ) {
            // report progress
            items++;
            progress(items, maxItems);

            final FileObject subSourceFo = childrens[i];
            
            FileObject subTargetFo = null;
            if (subSourceFo.isFolder()) {
                subTargetFo =  dest.getFileObject(subSourceFo.getName());
                if (subTargetFo == null) {
                    subTargetFo = dest.createFolder(subSourceFo.getName());
                    
                }
                copyAttributes(subSourceFo,subTargetFo);
                
                recursiveCopyWithFilter(subSourceFo,subTargetFo, noCopy, basicTime);
            } else {
                if (canCopy(subSourceFo, noCopy, basicTime)) {
                    subTargetFo = dest.getFileObject(subSourceFo.getName(),subSourceFo.getExt());
                    if (subTargetFo != null) {
                        final FileLock lock = subTargetFo.lock();
                        subTargetFo.delete(lock);
                        lock.releaseLock();
                    }
                    subTargetFo = FileUtil.copyFile(subSourceFo,dest,subSourceFo.getName());
                    copyAttributes(subSourceFo,subTargetFo);
//                    message (getString("MSG_WRITE",subSourceFo.getPackageNameExt('/', '.')));
                }
            }
        }
    }
    
    private static void copyAttributes (FileObject source, FileObject dest) throws IOException {
        Enumeration attrKeys = source.getAttributes();
        while (attrKeys.hasMoreElements()) {
            String key = (String) attrKeys.nextElement();
            Object value = source.getAttribute(key);
            if (value != null) {
                dest.setAttribute(key, value);
            }
        }
    }
    /** test if file can be copied
     */
    private boolean canCopy (FileObject fo, String[] nonCopiedFiles, long basicTime) throws IOException {
        String fileName =  fo.getPackageNameExt('/', '.');
        long time =  fo.lastModified().getTime();
        
        if (( Arrays.binarySearch(nonCopiedFiles,fileName) >= 0 ||
        basicTime + timeDev > time) && fo.isFolder() == false ) {
            return false ;
        } else {
            return true;
        }
    }
    
    private long getLastModified (File sourceDirFile) {
        long basicTime = sourceDirFile.lastModified();
        
        long lastModified = 0;
        File f = new File (sourceDirFile,getString("CTL_README"));
        if (f.exists()) {
            lastModified  = f.lastModified();
            if (lastModified < basicTime) {
                basicTime = lastModified;
            }
        } else {
            message (getString("MSG_NOT_FOUNDED", f.getPath()));
        }
        
        f = new File(new File (sourceDirFile,"beans"),"TimerBean.jar"); //NOI18N
        if (f.exists ()) {
            lastModified  = f.lastModified();
            
            if (lastModified < basicTime) {
                basicTime = lastModified;
            }
        } else {
            message ( getString("MSG_NOT_FOUNDED",f.getPath()));
        }
        return lastModified;
    }
    
    /** Creates sorted array of file names that should not be copied.
     * @param version the version of the IDE
     * @return array of the file names
     */
    private String[] originalFiles (int version) throws IOException {
        BufferedReader reader = new BufferedReader (
            new InputStreamReader (FileCopy.class.getResourceAsStream(nonCpFiles[version]))
        ); //NOI18N
        String line = null;
        Vector fileNames = new Vector ();
        while ((line = reader.readLine ()) != null ) {
            if (line.startsWith("./")) { //NOI18N
                line = line.substring(2);
            }
            fileNames.add (line);
        }
        String files[] = new String [fileNames.size ()];
        fileNames.copyInto(files);
        Arrays.sort (files);
        return files;
    }
    
    private void deleteAll(FileObject root) { //throws IOException
        FileObject children[] = root.getChildren();
        
        for(int i = 0; i < children.length; i++) {
            FileObject fo = children[i];
            FileLock lock = null;
          
            if (fo.getNameExt().equals("ide.log"))
                continue;
            
            if (fo.isFolder())
                deleteAll(fo);
            
            try {
                lock = fo.lock();
                fo.delete(lock);
            }
            catch (IOException e) {
                System.out.println("FileObject " + fo.getPackageNameExt('/', '.') + " couldn't be deleted.");
            }
            finally {
                if (null != lock)
                    lock.releaseLock();
            }
        }
    }
    
    static int getIdeVersion (File home) {
        String verDescription = getIdeVersionDescription(home);
        
        if (verDescription != null ) {
            return getIdeVersion(verDescription);
        }
        return -1;
    }

    static String getIdeVersionDescription (File home) {
        File libDir = new File (home,"lib"); //NOI18N
        File coreJar = new File (libDir ,"core.jar"); //NOI18N
        if (!coreJar.exists()) {
            coreJar = new File (libDir ,"developer.jar"); //NOI18N
        }
        try {
            JarFile jar = new JarFile(coreJar);
            JarEntry entry = jar.getJarEntry("org/netbeans/core/Bundle.properties"); //NOI18N
            BufferedReader reader = new BufferedReader(new InputStreamReader(jar.getInputStream(entry)));
            String line ;
            while ((line = reader.readLine()) != null) {
                StringTokenizer tokenizer  = new StringTokenizer(line,"="); //NOI18N
                try {
                    String token1 = tokenizer.nextToken();
                    if (token1.equals ("currentVersion")) { //NOI18N
                        return tokenizer.nextToken();
                    }
                } catch (NoSuchElementException e) {};
            }
        } catch (IOException e) {}
        return null;
    }
    
    static int getIdeVersion (String versionDescription) {
        for (int i = 0 ; i < versions.length ; i++ ) {
            if (versionDescription.startsWith(versions[i])) {
                return i;
            }
        }
        return -1;
    }

    static boolean canUpgradeFrom(String versionDescription) {
        return !ffj20IE.equals(versionDescription) || isForte();
    }
    
    private static boolean isForte() {
        String ver = NbBundle.getMessage(org.netbeans.core.NbTopManager.class, "currentVersion");
        
        if (null == ver)
            return true;
        
        return ver.startsWith("Forte");
    }
    
    static String getString (String key) {
        return NbBundle.getMessage (CopyUtil.class, key);
    }
    static String getString (String key,String param) {
        return NbBundle.getMessage(CopyUtil.class,key,param);
    }
}
