/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.upgrade;

import java.util.*;
import java.text.MessageFormat;
import java.beans.*;
//import javax.swing.event.ChangeListener;
//import javax.swing.event.ChangeEvent;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.WizardDescriptor;
import org.openide.filesystems.*;
import org.openide.util.NbBundle;
import org.netbeans.core.ui.*;

/** Upgrade wizard displays only one panel allowing user to upgrade his
 * settings from previous product version.
 *
 * @author  vstejskal
 * @version 
 */
public final class UpgradeWizard implements PropertyChangeListener { // extends WizardFolder 

    /** See org.openide.WizardDescriptor.PROP_AUTO_WIZARD_STYLE
     */
    private static final String PROP_AUTO_WIZARD_STYLE = "WizardPanel_autoWizardStyle"; // NOI18N
    /** See org.openide.WizardDescriptor.PROP_CONTENT_DISPLAYED
     */
    private static final String PROP_CONTENT_DISPLAYED = "WizardPanel_contentDisplayed"; // NOI18N
    /** See org.openide.WizardDescriptor.PROP_CONTENT_NUMBERED
     */
    private static final String PROP_CONTENT_NUMBERED = "WizardPanel_contentNumbered"; // NOI18N
    /** See org.openide.WizardDescriptor.PROP_CONTENT_SELECTED_INDEX
     */
    private static final String PROP_CONTENT_SELECTED_INDEX = "WizardPanel_contentSelectedIndex"; // NOI18N
    /** See org.openide.WizardDescriptor.PROP_CONTENT_DATA
     */
    private static final String PROP_CONTENT_DATA = "WizardPanel_contentData"; // NOI18N

    static final String PROP_CONTENT_PANEL_SELECTED = "WizardPanel_selected"; // NOI18N
    
    private UpgradeWizard() {
    }
    
    private FinalNoBackIterator     it = null;
    private WizardDescriptor        wd = null;
    private boolean                 cancelOnClose = true;
    private UpgradeWizardSettings   uwSettings = null;
    
    /** Depending on the presence of upgrade.properties file run wizard in 
     * silent/interactive mode.
     * @return true when user canceled import procedure, false otherwise (it
     * includes both situations, when procedure was skipped or proceeded and finished).
     */
    private boolean show() {
        uwSettings = UpgradeWizardSettings.getInstance();
        boolean result;
        
        if (null == uwSettings.getIDEVersion() || null == uwSettings.getIDELocation()) {
            result = showInteractive();
        } else {
            if (CopyUtil.canUpgradeFrom(uwSettings.getIDEVersion())) {
                result = showAutomatic();
            } else {
                result = false; //silently skip
            }
        }
        return result;
    }

    private boolean showInteractive() {
        WizardDescriptor.Panel panels[] = new WizardDescriptor.Panel[] {
            new PreUpgradePanel(),
            new UpgradePanel(), 
            new TransferSettingsPanel()
        };
        
        Object options[] = new Object [] {
            WizardDescriptor.PREVIOUS_OPTION,
            WizardDescriptor.NEXT_OPTION,
            WizardDescriptor.FINISH_OPTION,
            WizardDescriptor.CANCEL_OPTION
        };
        
        return show_internal(options, panels);
    }
    
    private boolean showAutomatic() {
        WizardDescriptor.Panel panels[] = new WizardDescriptor.Panel[] {
            new TransferSettingsPanel()
        };
        
        Object options[] = new Object [] {
            WizardDescriptor.PREVIOUS_OPTION,
            WizardDescriptor.NEXT_OPTION,
            WizardDescriptor.FINISH_OPTION,
        };
        
        return show_internal(options, panels);
    }

    private boolean show_internal(Object options[], WizardDescriptor.Panel panels[]) {
        TransferSettingsPanel transPanel = (TransferSettingsPanel)panels[panels.length - 1];
        boolean hidden = 1 == panels.length; // if there is just transfer panel run in silent mode
        
        it = new FinalNoBackIterator(panels);
        wd = new WizardDescriptor(
            it, 
            uwSettings);
        
        it.addPropertyChangeListener(this);
        it.addPropertyChangeListener(transPanel);
        
        wd.putProperty(PROP_AUTO_WIZARD_STYLE, new Boolean(true));
        wd.putProperty(PROP_CONTENT_DISPLAYED, new Boolean(true));
        wd.putProperty(PROP_CONTENT_NUMBERED, new Boolean(true));
        wd.setTitleFormat(new MessageFormat("{0}")); // NOI18N
        wd.setTitle(NbBundle.getMessage(UpgradeWizard.class, "LBL_UpgradeWizard_Title")); // NOI18N
        wd.putProperty(PROP_CONTENT_DATA, it.getContentData());
        wd.putProperty(PROP_CONTENT_SELECTED_INDEX, new Integer(it.getContentSelectedIndex()));
        wd.setOptions(options);
        wd.setClosingOptions(new Object [] { 
            WizardDescriptor.FINISH_OPTION,
            WizardDescriptor.CANCEL_OPTION });

        java.awt.Dialog d = TopManager.getDefault().createDialog(wd);
        if (hidden) {
            // kick-off upgrade procedure
            it.hasNext();
            
            try {
                // wait for the import to finish
                while (TransferSettingsPanel.STATUS_NONE == transPanel.getStatus() ||
                        TransferSettingsPanel.STATUS_RUNNING == transPanel.getStatus())
                {
                    Thread.currentThread().sleep(500);
                }

                hidden = TransferSettingsPanel.STATUS_FINISHED == transPanel.getStatus();
            }
            catch (InterruptedException e) {
                //something went wrong, show dialog and let user know
                hidden = false;
            }
        }
        
        if (!hidden)
            d.show();
        
        d.dispose();
        
        it.removePropertyChangeListener(this);
        it.removePropertyChangeListener(transPanel);
        
        return  wd.getValue() == wd.CANCEL_OPTION || 
                (cancelOnClose && wd.getValue() == wd.CLOSED_OPTION);
    }

    /** Listens on content property changes in Compound iterator. And updates Wizard
     * descriptor properties.
     */
    public void propertyChange(PropertyChangeEvent ev) {
        if (PROP_CONTENT_SELECTED_INDEX.equals(ev.getPropertyName()))
            wd.putProperty(PROP_CONTENT_SELECTED_INDEX,
                new Integer(it.getContentSelectedIndex()));
        else if (PROP_CONTENT_DATA.equals(ev.getPropertyName()))
            wd.putProperty(PROP_CONTENT_DATA, it.getContentData());
        else if (PROP_CONTENT_PANEL_SELECTED.equals(ev.getPropertyName())) {
            if (ev.getNewValue() instanceof TransferSettingsPanel) {
                cancelOnClose = false;
                wd.setOptions(new Object [] { 
                    WizardDescriptor.PREVIOUS_OPTION,
                    WizardDescriptor.NEXT_OPTION,
                    WizardDescriptor.FINISH_OPTION});
            }
            else {
                cancelOnClose = true;
                wd.setOptions(new Object [] { 
                    WizardDescriptor.PREVIOUS_OPTION,
                    WizardDescriptor.NEXT_OPTION,
                    WizardDescriptor.FINISH_OPTION,
                    WizardDescriptor.CANCEL_OPTION });
            }
        }
    }

    /** Shows the upgrade wizard if there is no folder under the Projects
     * subfolder of system default filesystem. This condition is met during
     * the first start only.
     * 
     * @return true when and only when the wizard was displayed and user has canceled it.
     */
    public static boolean showWizard() {
        boolean status = false;
        LocalFileSystem fs = new LocalFileSystem();
        
        try {
            fs.setRootDirectory(new java.io.File(System.getProperty("netbeans.user", "")));
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        FileObject foProjectLast = fs.findResource("system/project.last"); // NOI18N

        if (null == foProjectLast) {
            // no project was saved as yet (during the first start only)
            // show Upgrade Wizard
            UpgradeWizard f = new UpgradeWizard();
            status = f.show();
        }
        return status;
    }

    public static void main(String args[]) {
        UpgradeWizard uw = new UpgradeWizard();
        uw.show();
        
        System.out.println("exit");
    }

    /** This iterator prevents user from pressing Back button, when it
     * displays last panel.
     */
    private class FinalNoBackIterator extends WizardDescriptor.ArrayIterator {
        protected WizardDescriptor.Panel [] panels;
        protected LinkedList listeners = new LinkedList();
        private boolean first = true;
        
        public FinalNoBackIterator(WizardDescriptor.Panel [] list) {
            super(list);
            panels = list;
        }

        public boolean hasPrevious() {
            checkFirst();
            int index = getContentSelectedIndex();
            boolean hasPrev = super.hasPrevious();
            
            if (hasPrev)
                hasPrev = index < panels.length - 1;
            
            if (hasPrev && panels[index] instanceof ExtendedPanel) {
                hasPrev = ((ExtendedPanel)panels[index]).hasPrevious();
            }
            
            return hasPrev;
        }

        public boolean hasNext() {
            checkFirst();
            int index = getContentSelectedIndex();
            boolean hasNext = super.hasNext();
            
            if (hasNext && panels[index] instanceof ExtendedPanel) {
                hasNext = ((ExtendedPanel)panels[index]).hasNext();
            }
            
            return hasNext;
        }
        
        public void previousPanel() {
            super.previousPanel();
            firePanelChangedEvent();
        }

        public void nextPanel() {
            super.nextPanel();
            firePanelChangedEvent();
        }
        
        public String[] getContentData() {
            String[] cData = new String[panels.length];
            for (int i = 0; i < panels.length; i++) {
                cData[i] = panels[i].getComponent().getName();
            }
            return cData;
        }

        public int getContentSelectedIndex() {
            for (int i = 0; i < panels.length; i++) {
                if (current() == panels[i])
                    return i;
            }
            return -1;
        }
        
        public void addPropertyChangeListener(PropertyChangeListener l) {
            if (!listeners.contains(l))
                listeners.add(l);
        }
        
        public void removePropertyChangeListener(PropertyChangeListener l) {
            listeners.remove(l);
        }

        private void checkFirst() {
            if (first) {
                first = false;
                
                firePanelChangedEvent();
            }
        }
        
        private void firePanelChangedEvent() {
            Iterator i = listeners.iterator();
            while(i.hasNext()) {
                PropertyChangeListener l = (PropertyChangeListener)i.next();
                
                l.propertyChange(new PropertyChangeEvent(
                    this,
                    PROP_CONTENT_SELECTED_INDEX,
                    null,
                    new Integer (getContentSelectedIndex())));
                
                l.propertyChange(new PropertyChangeEvent(
                    this,
                    PROP_CONTENT_PANEL_SELECTED,
                    null,
                    current()));
            }
        }
    }
    
    public interface ExtendedPanel {
        public boolean hasNext();
        public boolean hasPrevious();
    }
}
