/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.awt.Rectangle;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;

import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;


/** Allows to convert bounds of mode when changing mode frame type.
 *
 * @author  Dafe Simonek
 */
final class BoundsConvertor extends Object {
    
    private static BoundsConvertor instance;
    
    /** Array of possible conversions */
    private ArrayList convertors;
    /** helper, holds previous frame type when bounds conversion is pending */
    private String previousFrameType;
    
    /** Singleton, no need to instantiate externally */ 
    private BoundsConvertor () {
    }

    /** Instance accessor */
    public static final BoundsConvertor getInstance () {
        if (instance == null) {
            instance = new BoundsConvertor();
        }
        return instance;
    }

    public Rectangle convertBounds (ModeImpl mode, String oldFrameType, String newFrameType) {
        if (convertors == null) {
            convertors = createConversions();
        }
        Conversion curConv = null;
        for (Iterator iter = convertors.iterator(); iter.hasNext(); ) {
            curConv = (Conversion)iter.next();
            if (curConv.accept(mode, oldFrameType, newFrameType)) {
                return curConv.convertBounds(mode, oldFrameType, newFrameType);
            }
        }
        // no conversion
        return mode.getBounds();
    }
    
    public boolean updateBounds (ModeImpl mode) {
        return updateBounds(mode, previousFrameType, mode.getFrameType());
    }
    
    /** Updates bounds if needed. Bounds update is needed if bounds conversion
     * is pending due to MDI/SDI mode change.
     * @return true if bounds were updated, false otherwise
     */
    public boolean updateBounds (ModeImpl mode, String oldFrameType, String newFrameType) {
        if (oldFrameType == null) {
            // no request pending, bounds are ok
            return false;
        }
        if (!WindowManagerImpl.getDefault().deferredPerformer().canImmediatelly()) {
            // can't update bounds now, we must wait for consistent state of IDE
            if (previousFrameType == null) {
                previousFrameType = oldFrameType;
            }
            return false;
        }
        if (mode.getBounds() == null) {
            // mode bounds not initialized yet, no conversion
            return false;
        }
        mode.setBounds(convertBounds(mode, oldFrameType, newFrameType));
        previousFrameType = null;
        return true;
    }
    
    
    private ArrayList createConversions () {
        ArrayList convArray = new ArrayList(3);
        convArray.add(new MdiSdi());
        convArray.add(new Desktop2Internal());
        return convArray;
    }
    
    /** Specifies protocol for conversions, conversion knows its type 
     * and accepts only proper frame types combinations which it can handle
     */
    private interface Conversion {
        
        /** @return true if this conversion can convert coordinates for given
         * specified frame types, false otherwise
         */ 
        public boolean accept (ModeImpl mode, String oldFrameType, String newFrameType);
        
        /** Performs conversion task */
        public Rectangle convertBounds (ModeImpl mode, String oldFrameType, String newFrameType);
        
    } // end of Conversion interface

    /** Can convert between MDI and SDI frame types (both directions) */
    private static final class MdiSdi implements Conversion {
        
        public boolean accept (ModeImpl mode, String oldFrameStr, String newFrameStr) {
            boolean isOldMdi = (WindowTypesManager.INTERNAL_FRAME.equals(oldFrameStr)) ||
                               (WindowTypesManager.DESKTOP_FRAME.equals(oldFrameStr));
            boolean isNewMdi = (WindowTypesManager.INTERNAL_FRAME.equals(newFrameStr)) ||
                              (WindowTypesManager.DESKTOP_FRAME.equals(newFrameStr));
            return isOldMdi != isNewMdi;
        }
        
        public Rectangle convertBounds (ModeImpl mode, String oldFrameStr, String newFrameStr) {
            boolean isOldMdi = (WindowTypesManager.INTERNAL_FRAME.equals(oldFrameStr)) ||
                               (WindowTypesManager.DESKTOP_FRAME.equals(oldFrameStr));
            boolean isNewMdi = (WindowTypesManager.INTERNAL_FRAME.equals(newFrameStr)) ||
                              (WindowTypesManager.DESKTOP_FRAME.equals(newFrameStr));
            Rectangle bounds = mode.getBounds();
            Point newLeftTop = bounds.getLocation();
            Object constraints = mode.getConstraints();
            Point offset = null;
            Rectangle result = null;
            if ((constraints == null) || (PerimeterLayout.CENTER.equals(constraints))) {
                // floating modes. move them relatively to inner desktop
                offset = ((WorkspaceImpl)mode.getWorkspace()).getDesktopScreenLocation();
            } else {
                // modes attached to sides, move them relatively to whole desktop
                offset = WindowManagerImpl.getDefault().mainWindow().getDesktopScreenLocation();
            }
            if (isOldMdi && !isNewMdi) {
                newLeftTop.translate(offset.x, offset.y);
            } else {
                newLeftTop.translate(-offset.x, -offset.y);
            }
            result = new Rectangle(newLeftTop, bounds.getSize());
            return result;
        }
        
    } // end of MdiSdi inner class
    
    /** Can convert desktop frame type to internal frame type */
    private static final class Desktop2Internal implements Conversion {
        
        public boolean accept (ModeImpl mode, String oldFrameStr, String newFrameStr) {
            return WindowTypesManager.DESKTOP_FRAME.equals(oldFrameStr) &&
                   WindowTypesManager.INTERNAL_FRAME.equals(newFrameStr);
        }
        
        public Rectangle convertBounds (ModeImpl mode, String oldFrameStr, String newFrameStr) {
            Rectangle bounds = mode.getBounds();
            Point leftTop = bounds.getLocation();
            Point innerDesktopLoc = ((WorkspaceImpl)mode.getWorkspace()).getDesktopScreenLocation();
            Point desktopLoc = WindowManagerImpl.getDefault().mainWindow().getDesktopScreenLocation();
            Rectangle result = new Rectangle(
                leftTop.x - (innerDesktopLoc.x - desktopLoc.x),
                leftTop.y - (innerDesktopLoc.y - desktopLoc.y),
                bounds.getSize().width - 10,
                bounds.getSize().height - 10
            );
            return result;
            
        }
        
    } // end of Desktop2Internal inner class
    
}
