/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.awt.*;
import java.util.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeSupport;
import java.net.URL;
import java.text.MessageFormat;

import javax.swing.SwingUtilities;

import org.openide.windows.*;
import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.text.CloneableEditorSupport;
import org.openide.util.actions.SystemAction;

import org.netbeans.core.output.OutputTab;
import org.netbeans.core.actions.ExecutionViewAction;
import org.netbeans.core.actions.OpenExplorerAction;
import org.netbeans.core.NbMainExplorer;
import org.netbeans.core.NbNodeOperation;
import org.netbeans.core.windows.ModeImpl;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.openide.nodes.Node;

import org.netbeans.core.windows.util.WindowUtils;

/** Creates default workspace pool.
* Creates default workspaces and default positions of windows
*
* @author Dafe Simonek
*/
final class DefaultCreator extends WindowAdapter implements WellKnownModeNames, WellKnownWorkspaceNames {

    /** Window manager to initialize and fill */
    private WindowManagerImpl wm;
    /** Default workspaces */
    private WorkspaceImpl editing;
    private WorkspaceImpl browsing;
    private WorkspaceImpl running;
    private WorkspaceImpl debugging;
    private WorkspaceImpl visual;
    private WorkspaceImpl[] workspaces;
    
    /** functionality enable / disable flag */
    private boolean enabled;

    private static HashMap defaultModeDefs;
    
    static {
        ModeImpl.ModeDefinition md;
        defaultModeDefs = new HashMap();
        
        md = new ModeImpl.ModeDefinition();
        md.name = EXPLORER;
        md.displayName = NbBundle.getBundle(DefaultCreator.class).getString("CTL_ExplorerWindow");
        md.iconURL = getIconURL("nbresboot:/org/netbeans/core/resources/frames/explorer.gif"); // NOI18N
        md.containerType = WindowTypesManager.SPLIT_CONTAINER;
        md.constraints = PerimeterLayout.WEST;
        defaultModeDefs.put(EXPLORER, md);

        md = new ModeImpl.ModeDefinition();
        md.name = OUTPUT;
        md.displayName = NbBundle.getBundle(DefaultCreator.class).getString("CTL_OutputWindow");
        md.iconURL = getIconURL("nbresboot:" + OutputTab.ICON_RESOURCE);
        md.containerType = WindowTypesManager.SPLIT_CONTAINER;
        md.constraints = PerimeterLayout.SOUTH;
        defaultModeDefs.put(OUTPUT, md);

        md = new ModeImpl.ModeDefinition();
        md.name = EXECUTION;
        md.displayName = NbBundle.getBundle(DefaultCreator.class).getString("CTL_ExecutionWindow");
        md.iconURL = getIconURL("nbresboot:/org/netbeans/core/resources/frames/execution.gif"); // NOI18N
        md.containerType = WindowTypesManager.MULTI_TABBED_CONTAINER;
        md.constraints = null;
        defaultModeDefs.put(EXECUTION, md);

        md = new ModeImpl.ModeDefinition();
        md.name = PROPERTIES;
        md.displayName = NbBundle.getBundle(DefaultCreator.class).getString("CTL_PropertiesWindow");
        md.iconURL = getIconURL("nbresboot:/org/netbeans/core/resources/frames/properties.gif"); // NOI18N
        md.containerType = WindowTypesManager.MULTI_TABBED_CONTAINER;
        md.constraints = null;
        defaultModeDefs.put(PROPERTIES, md);

        md = new ModeImpl.ModeDefinition();
        md.name = EDITOR;
        md.displayName = NbBundle.getBundle(DefaultCreator.class).getString("CTL_EditorWindow");
        md.iconURL = getIconURL("nbresboot:/org/netbeans/core/resources/frames/editor.gif"); // NOI18N
        md.containerType = WindowTypesManager.SPLIT_CONTAINER;
        md.constraints = null;
        defaultModeDefs.put(EDITOR, md);
    }

    DefaultCreator(WindowManagerImpl wm) {
        this.wm = wm;
        this.enabled = true;
    }

    /** Starts creation of default window system */
    void start () {
        initWorkspaces();
        wm.setCreated(true);
        // listen to main window (and set initial positions
        // only after main window is opened)
        WindowManagerImpl.mainWindow().addWindowListener(this);
    }
    
    /** Sets flag which will stop further creation of default content
     * of window system */ 
    void setEnabled (boolean enabled) {
        this.enabled = enabled;
    }

    /** Initializes default workspace */
    void initWorkspaces () {
        // Initialize the default workspace pool
        ResourceBundle bundle = NbBundle.getBundle(DefaultCreator.class);
        editing = new WorkspaceImpl(EDITING, "CTL_Workspace_Editing", true);
        browsing = new WorkspaceImpl(BROWSING, "CTL_Workspace_Browsing", true);
        running = new WorkspaceImpl(RUNNING, "CTL_Workspace_Running", true);
        debugging = new WorkspaceImpl(DEBUGGING, "CTL_Workspace_Debugging", true);
        ((WorkspaceImpl)debugging).setToolbarConfigName(DEBUGGING);
        visual = new WorkspaceImpl(VISUAL, "CTL_Workspace_Visual", true);
        // modify the workspace pool
        workspaces = new WorkspaceImpl[] {
                         editing, visual, browsing, running, debugging
                     };
        wm.setWorkspaces(new Workspace[0]);
        wm.setWorkspaces(workspaces);
        wm.setCurrentWorkspace(workspaces[0]);
    }

    /** Set initial positions after main window is opened */
    public void windowOpened (WindowEvent e) {
        // notify me no more
        WindowManagerImpl.mainWindow().removeWindowListener(this);
        if (!enabled) {
            return;
        }
        // set important positioning variables
        Window mainWindow = WindowManagerImpl.mainWindow();
        // cascade from different point on visual workspace in SDI mode
        if (!WindowUtils.isMdi()) {
            ((WorkspaceImpl)visual).setCascadingOrigin(new Point(0, 50));
        }
        // set default positions for frames
        placeFrames();
    }

    public void placeFrames () {
        // set default window positions (main window is already visible)
        FramePlacer placer = null;
        if (WindowUtils.isMdi()) {
            placer = new MDIPlacer();
        } else {
            placer = new SDIPlacer();
        }
        placer.placeEditing();
        placer.placeRunning();
        placer.placeVisual();
        placer.placeBrowsing();
        placer.placeDebugging();
        // set current workspace
        editing.activate();
        // open neede components
        openComponents();
        
        //Select FileSystem tab in Explorer
        //Temporary solution for bugfix #9352. There is currently
        //no way how to select given tab other than focused in split container.
        //It requires better solution.
        TopComponent toBeActivated;
        Node[] rootsArray = (Node[])NbMainExplorer.getRoots().toArray(new Node[0]);
        
        toBeActivated = NbMainExplorer.getExplorer().getRootPanel(rootsArray[0]);
        
        ModeImpl mode = (ModeImpl)editing.findMode(toBeActivated);
        if (mode != null) {
            mode.requestFocus(toBeActivated);
        }
        //End of bugfix #9352
    }

    /** Creates mode on given workspace using default mode attributes. If the mode
     * definition doesn't exist, nothing is created and null is returned.
     *
     * @param wsp workspace, in which mode will be created
     * @param name name of the mode, ModeDefinition will be found and mode will be instantiated
     * @return ModeImpl instance with attributes preset to default values or null
     * in case there is no definition for mode with given name
     */
    public static ModeImpl createMode(Workspace wsp, String name) {
        WindowManagerImpl wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        ModeImpl.ModeDefinition mdef;
        ModeImpl m;
        
        if (null == (mdef = getModeDefinition(name)))
            return null;
        
        m = (ModeImpl)wsp.createMode(name, mdef.displayName, mdef.iconURL);
        m.setContainerType(mdef.containerType);
        m.setConstraints(mdef.constraints);
        m.setFrameType(wm.findFrameType(m));
        return m;
    }
    
    public static ModeImpl.ModeDefinition getModeDefinition(Mode mode) {
        return getModeDefinition(mode.getName());
    }

    /** Finds mode definition for given name of mode.
     * @return mode definition or null if there is no definition for given name of mode
     */
    public static ModeImpl.ModeDefinition getModeDefinition(String modeName) {
        return (ModeImpl.ModeDefinition)defaultModeDefs.get(modeName);
    }

    private void openComponents () {
        // open main explorer and global property sheet
        // on current (editing) workspace        
        NbMainExplorer.getExplorer().openRoots(editing);        
        NbNodeOperation.Sheet.getDefault().open(editing);
        // open output on running and debugging workspaces
        TopComponent outputTc = OutputTab.getStdOutputTab();
        outputTc.open(running);
        outputTc.open(debugging);
        // open explorer also on visual workspace
        NbMainExplorer.getExplorer().openRoots(visual);
        // property sheet on browsing
        NbNodeOperation.Sheet.getDefault().open(browsing);
        // execution on running
        ExecutionViewAction.getExecutionView().open(running);
    }

 
    /** Returns URL for given resource string or null if URL cannot
     * be constructed (is ill formattted, for example) */
    private static final URL getIconURL (String url) {
        URL result = null;
        try {
            result = new URL(url); // NOI18N
        } catch (java.net.MalformedURLException exc) {
            // log warning to the log file and return null
            String pattern = NbBundle.getBundle(DefaultCreator.class).getString("LOG_NoIcon");
            TopManager.getDefault().getErrorManager().log(
                ErrorManager.WARNING,
                MessageFormat.format(pattern, new Object[] { url })
            );
        }
        return result;
    }
    
    private ModeImpl createExplorer (Workspace workspace) {
        // create appropriate modes and set the bounds
        TopComponent explorerTc = NbMainExplorer.getExplorer();
        ModeImpl explorerMode = createMode(workspace, EXPLORER);
        explorerMode.dockInto(explorerTc);
        return explorerMode;
    }

    private ModeImpl createOutput (Workspace workspace) {
        // create appropriate modes and set the bounds
        TopComponent outputTc = OutputTab.getStdOutputTab();
        ModeImpl curMode = createMode(workspace, OUTPUT);
        curMode.addIgnoredNameClass(outputTc.getClass());
        curMode.dockInto(outputTc);
        return curMode;
    }

    private ModeImpl createExecution (Workspace workspace) {
        // create appropriate modes and set the bounds
        TopComponent executionTc = ExecutionViewAction.getExecutionView();
        ModeImpl curMode = createMode(workspace, EXECUTION);
        curMode.addIgnoredNameClass(executionTc.getClass());
        curMode.dockInto(executionTc);
        // attached to right side on Running workspace
        if (workspace.equals(running)) {
            curMode.setConstraints(PerimeterLayout.EAST);
        }
        return curMode;
    }

    private ModeImpl createPropertySheet (Workspace workspace) {
        TopComponent propertiesTc = NbNodeOperation.Sheet.getDefault();
        ModeImpl mode = createMode(workspace, PROPERTIES);
        mode.dockInto(propertiesTc);
        return mode;
    }

    private ModeImpl createEditor (Workspace workspace) {
        ModeImpl curMode = createMode(workspace, EDITOR);
        return curMode;
    }
    
    /** Basic frames placement that is the same both for SDI and MDI modes.   
     */
    private abstract class FramePlacer extends Object {
        
        /** Two dimensional mapping which holds all default bounds of modes
         * on all workspaces.
         * <Workspace, <String (mode_name), Rectangle (mode bounds)>> */
        private Map allBounds;

        /** Creates and places modes for editing workspace. */
        protected void placeEditing () {
            installBasicModes(editing);
        }
        
        /** Creates and places modes for GUI editing workspace. */
        protected void placeVisual () {
            installBasicModes(visual);
            setModeBounds(createPropertySheet(visual));
        }
        
        /** Creates and places modes for browsing workspace. */
        protected void placeBrowsing () {
            installBasicModes(browsing);
            setModeBounds(createPropertySheet(browsing));
        }

        /** Creates and places modes for running workspace. */
        protected void placeRunning () {
            installBasicModes(running);
        }
        
        /** Creates and places modes for debugging workspace. */
        protected void placeDebugging () {
            installBasicModes(debugging);
        }
        
        private void installBasicModes (Workspace workspace) {
            setModeBounds(createExplorer(workspace));
            setModeBounds(createOutput(workspace));
            setModeBounds(createExecution(workspace));
            setModeBounds(createEditor(workspace));
        }

        /** Sets bounds for given mode by searching bounds mappings */
        protected void setModeBounds(ModeImpl mode) {
            if (allBounds == null) {
                buildBounds();
            }
            mode.setBounds(
                (Rectangle)((Map)allBounds.get(mode.getWorkspace())).
                            get(mode.getName())
            );
        }
        
        private void buildBounds () {
            allBounds = new HashMap(10);
            allBounds.put(editing, editingBounds());
            allBounds.put(running, runningBounds());
            allBounds.put(visual, visualBounds());
            allBounds.put(browsing, browsingBounds());
            allBounds.put(debugging, debuggingBounds());
        }
        
        /** To be overriden by sublasses, which will specify bounds
         * either for SDI or MDI enviromnemt */
        protected Map editingBounds () {
            Rectangle workingSpace = editing.getBounds();
            int realHeight = workingSpace.height * 2 / 3;
            Map modesBounds = new HashMap(15);
            // explorer
            int explorerHeight = (realHeight > NbMainExplorer.MIN_HEIGHT) 
                                 ? realHeight : workingSpace.height;
            modesBounds.put(EXPLORER, new Rectangle(
                workingSpace.x, workingSpace.y,
                NbMainExplorer.DEFAULT_WIDTH, explorerHeight
            ));
            // output    
            modesBounds.put(OUTPUT, new Rectangle(
                workingSpace.x + NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.y + workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT,
                workingSpace.x + workingSpace.width - NbMainExplorer.DEFAULT_WIDTH,
                OutputTab.DEFAULT_WINDOW_HEIGHT
            ));
            return modesBounds;
        }
        
        protected Map runningBounds () {
            Rectangle workingSpace = running.getBounds();
            Map modesBounds = new HashMap(15);
            // explorer
            int outputHeight = workingSpace.height * 4 / 9;
            modesBounds.put(EXPLORER, new Rectangle(
                workingSpace.x, workingSpace.y,
                NbMainExplorer.DEFAULT_WIDTH, workingSpace.height - outputHeight
            ));
            // output
            modesBounds.put(OUTPUT, new Rectangle(
                workingSpace.x + NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.y + workingSpace.height - outputHeight,
                workingSpace.x + workingSpace.width - NbMainExplorer.DEFAULT_WIDTH,
                outputHeight
            ));
            return modesBounds;
        }
        
        protected Map visualBounds () {
            Rectangle workingSpace = visual.getBounds();
            Map modesBounds = new HashMap(15);
            // explorer
            modesBounds.put(EXPLORER, new Rectangle(
                workingSpace.x, workingSpace.y, workingSpace.width * 3 / 10, workingSpace.height / 2
            ));
            // output
            modesBounds.put(OUTPUT, new Rectangle(
                workingSpace.x + (workingSpace.width / 2),
                workingSpace.y + workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT,
                workingSpace.width / 2,
                OutputTab.DEFAULT_WINDOW_HEIGHT
            ));
            return modesBounds;
        }
        
        protected Map browsingBounds () {
            Rectangle workingSpace = browsing.getBounds();
            int realHeight = workingSpace.height * 2 / 3;
            Map modesBounds = new HashMap(15);
            // explorer
            modesBounds.put(EXPLORER, new Rectangle(
                workingSpace.x, workingSpace.y, NbMainExplorer.DEFAULT_WIDTH,
                (realHeight > NbMainExplorer.MIN_HEIGHT) ? realHeight : workingSpace.height
            ));
            // output    
            modesBounds.put(OUTPUT, new Rectangle(
                workingSpace.x + NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.y + workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT,
                workingSpace.x + workingSpace.width - NbMainExplorer.DEFAULT_WIDTH,
                OutputTab.DEFAULT_WINDOW_HEIGHT
            ));
            return modesBounds;
        }
        
        protected Map debuggingBounds () {
            Rectangle workingSpace = debugging.getBounds();
            int realHeight = workingSpace.height * 2 / 3;
            Map modesBounds = new HashMap(15);
            // explorer
            modesBounds.put(EXPLORER, new Rectangle(
                workingSpace.x, workingSpace.y, NbMainExplorer.DEFAULT_WIDTH,
                (realHeight > NbMainExplorer.MIN_HEIGHT) ? realHeight : workingSpace.height
            ));
            // output    
            modesBounds.put(OUTPUT, new Rectangle(
                workingSpace.x + NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.y + realHeight,
                workingSpace.x + workingSpace.width - NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.height - realHeight
            ));
            return modesBounds;
        }
        
    } // end of FramePlacer
    
    /** Specifies default bounds for all modes in SDI environment. */
    private final class SDIPlacer extends FramePlacer {
        
        /** Creates and places modes for editing workspace. */
        protected void placeEditing () {
            super.placeEditing();
            setModeBounds(createPropertySheet(editing));
        }
        
        /** Creates and places modes for running workspace. */
        protected void placeRunning () {
            super.placeRunning();
            setModeBounds(createPropertySheet(running));
        }
        
        /** Creates and places modes for debugging workspace. */
        protected void placeDebugging () {
            super.placeDebugging();
            setModeBounds(createPropertySheet(debugging));
        }
        
        protected Map editingBounds () {
            Map modesBounds = super.editingBounds();
            Rectangle workingSpace = editing.getBounds();
            // editor
            int editorHeight = workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT;
            if (editorHeight < 200) {
                editorHeight = workingSpace.height;
            }
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x + NbMainExplorer.DEFAULT_WIDTH, workingSpace.y,
                workingSpace.width - (workingSpace.x + NbMainExplorer.DEFAULT_WIDTH),
                editorHeight
            ));
            // property sheet
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x, workingSpace.y + workingSpace.height * 2 / 3,
                NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.height - workingSpace.height * 2 / 3
            ));
            // execution
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width, workingSpace.y,
                prefSize.width, prefSize.height
            ));
            return modesBounds;
        }
        
        protected Map visualBounds() {
            Map modesBounds = super.visualBounds();
            Rectangle workingSpace = visual.getBounds();
            // editor
            int realHeight = workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT;
            if (realHeight < 200) {
                realHeight = workingSpace.height;
            }
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x + (workingSpace.width / 2), workingSpace.y,
                workingSpace.width / 2, realHeight
            ));
            // property sheet
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x + (workingSpace.width * 3 / 10), workingSpace.y,
                workingSpace.width * 2 / 10, workingSpace.height / 2
            ));
            return modesBounds;
        }

        protected Map runningBounds() {
            Map modesBounds = super.runningBounds();
            Rectangle workingSpace = running.getBounds();
            int outputHeight = workingSpace.height * 4 / 9;
            // editor
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y, 
                workingSpace.x + workingSpace.width - prefSize.width,
                workingSpace.height - outputHeight
            ));
            // execution
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width, workingSpace.y,
                prefSize.width, prefSize.height
            ));
            // property sheet
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x, workingSpace.y + workingSpace.height - outputHeight,
                NbMainExplorer.DEFAULT_WIDTH,
                outputHeight
            ));
            return modesBounds;
        }
        
        protected Map browsingBounds() {
            Map modesBounds = super.browsingBounds();
            Rectangle workingSpace = browsing.getBounds();
            // editor
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y + (workingSpace.height / 2),
                workingSpace.width, workingSpace.height / 2
            ));
            // property sheet
            int widthOfProps = workingSpace.width / 4;
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x + (workingSpace.width - widthOfProps), workingSpace.y,
                widthOfProps, workingSpace.height / 2
            ));
            // execution
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width, workingSpace.y,
                prefSize.width, prefSize.height
            ));
            return modesBounds;
        }
            
        protected Map debuggingBounds() {
            Map modesBounds = super.debuggingBounds();
            Rectangle workingSpace = debugging.getBounds();
            int commonHeight = workingSpace.height * 2 / 3;
            // editor
               modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y,
                workingSpace.width / 2, commonHeight
            ));
            // execution
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x, workingSpace.y + workingSpace.height * 2 / 3,
                NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.height - workingSpace.height * 2 / 3
            ));
            // property sheet
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x, workingSpace.y + workingSpace.height * 2 / 3,
                NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.height - workingSpace.height * 2 / 3
            ));
            return modesBounds;
        }
        
    } // end of SDIPlacer

    /** Specifies default bounds for all modes known for core in MDI environment. */
    private final class MDIPlacer extends FramePlacer {

        /** size (thick) of splitter */
        private int gap;
        
        MDIPlacer () {
            super();
            gap = ((PerimeterLayout)editing.desktopPane().getLayout()).getGap();
        }
        
        /** Creates and places modes for editing workspace. */
        protected void placeEditing () {
            super.placeEditing();
            ModeImpl explorerMode = (ModeImpl)editing.findMode(EXPLORER);
            explorerMode.dockInto(
                NbNodeOperation.Sheet.getDefault(), PerimeterLayout.SOUTH
            );
        }
        
        /** Creates and places modes for editing workspace. */
        protected void placeVisual () {
            super.placeVisual();
            ModeImpl explorerMode = (ModeImpl)visual.findMode(EXPLORER);
            explorerMode.dockInto(
                NbNodeOperation.Sheet.getDefault(), PerimeterLayout.SOUTH
            );
        }
        
        /** Creates and places modes for running workspace. */
        protected void placeRunning () {
            super.placeRunning();
            ModeImpl explorerMode = (ModeImpl)running.findMode(EXPLORER);
            explorerMode.dockInto(
                NbNodeOperation.Sheet.getDefault(), PerimeterLayout.SOUTH
            );
        }
        
        /** Creates and places modes for debugging workspace. */
        protected void placeDebugging () {
            super.placeDebugging();
        }
        
        protected Map editingBounds() {
            Map modesBounds = super.editingBounds();
            Rectangle workingSpace = editing.getBounds();
            // editor
            int editorHeight = workingSpace.height - 
                               (OutputTab.DEFAULT_WINDOW_HEIGHT + gap);
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y,
                workingSpace.width - 
                    (workingSpace.x + NbMainExplorer.DEFAULT_WIDTH + gap),
                editorHeight
            ));
            // property sheet
            Dimension prefSize = new Dimension(
                NbMainExplorer.DEFAULT_WIDTH,
                workingSpace.height - (workingSpace.height * 2 / 3)
            );
            NbNodeOperation.Sheet.getDefault().setSize(prefSize);
            // execution
            Dimension execSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - 
                    (execSize.width + NbMainExplorer.DEFAULT_WIDTH + gap),
                workingSpace.y,
                execSize.width, execSize.height
            ));
            return modesBounds;
        }
        
        protected Map visualBounds() {
            Map modesBounds = super.visualBounds();
            Rectangle workingSpace = visual.getBounds();
            int gap = ((PerimeterLayout)visual.desktopPane().getLayout()).getGap();
            int desktopWidth = workingSpace.width - (workingSpace.width * 3 / 10 + gap);
            // editor
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y + workingSpace.height / 2,
                desktopWidth,
                workingSpace.height / 2
            ));
            return modesBounds;
        }

        protected Map runningBounds() {
            Map modesBounds = super.runningBounds();
            Rectangle workingSpace = running.getBounds();
            int outputHeight = workingSpace.height * 4 / 9;
            // editor
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            int editorHeight = workingSpace.height - OutputTab.DEFAULT_WINDOW_HEIGHT;
            if (editorHeight < 200) {
                editorHeight = workingSpace.height;
            }
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y, 
                workingSpace.x + workingSpace.width - 
                    (prefSize.width + 2*gap + NbMainExplorer.DEFAULT_WIDTH),
                workingSpace.height - (outputHeight + gap)
            ));
            // execution
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width,
                workingSpace.y, prefSize.width, prefSize.height
            ));
            return modesBounds;
        }
        
        protected Map browsingBounds() {
            Map modesBounds = super.browsingBounds();
            Rectangle workingSpace = browsing.getBounds();
            // editor
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y + (workingSpace.height / 2),
                workingSpace.width, workingSpace.height / 2
            ));
            // property sheet
            int widthOfProps = workingSpace.width / 4;
            modesBounds.put(PROPERTIES, new Rectangle(
                workingSpace.x + (workingSpace.width - widthOfProps), workingSpace.y,
                widthOfProps, workingSpace.height / 2
            ));
            // execution
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width, workingSpace.y,
                prefSize.width, prefSize.height
            ));
            return modesBounds;
        }
            
        protected Map debuggingBounds() {
            Map modesBounds = super.debuggingBounds();
            Rectangle workingSpace = debugging.getBounds();
            // editor
            int basicHeight = workingSpace.height * 2 / 3 - gap;
            modesBounds.put(CloneableEditorSupport.EDITOR_MODE, new Rectangle(
                workingSpace.x, workingSpace.y,
                workingSpace.width / 2, basicHeight
            ));
            // execution
            Dimension prefSize = ExecutionViewAction.getExecutionView().getPreferredSize();
            modesBounds.put(EXECUTION, new Rectangle(
                workingSpace.x + workingSpace.width - prefSize.width,
                workingSpace.y + basicHeight - prefSize.height,
                prefSize.width, prefSize.height
            ));
            return modesBounds;
        }
        
    } // end of MDIPlacer
    
    
}
