/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.ArrayList;
import java.util.Iterator;

import org.openide.windows.TopComponent;
import org.openide.windows.Workspace;
import org.openide.TopManager;

/** Support for deferred performing of various actions which
* need window system in consistent state.
* Clients will implement inner interface DeferredCommand
* and use putRequest method to achieve desired functionality.<p>
*
* Design follows principles of well known "Command" design pattern.
*
* @author  Dafe Simonek
*/
public class DeferredPerformer extends Object implements StateManager.StateListener {

    /** not initialized yet (before first request) */
    static final int NOT_INITIALIZED = 0;
    /** ready for immediate open */
    static final int READY = 1;
    /** immediate open is not possible now, must be deferred */
    static final int DEFERRED = 2;
    /** holds simplified state of window manager
    * either READY or DEFERRED (unitialized at startup) */
    int state;
    /** Array of DeferredEntry entries containing top components
    * whose open is deferred */
    ArrayList requests;
    /** Asociation with window manager's state machine support */
    StateManager stateManager;

    /** Consider it private, should be called only from
    * window manager implementation.
    */
    DeferredPerformer() {
        state = NOT_INITIALIZED;
    }

    /** Puts a command to the deferred performer.
    * If window manager is in consistent state, specified command
    * will be executed immediately, otherwise it will be deferred
    * until window manager comes to consistent state. */
    public void putRequest (DeferredCommand dc, Object context) {
        putRequest(dc, new DeferredContext(context, false));
    }
    
    public void putRequest (DeferredCommand dc, DeferredContext context) {
        if (context == null) {
            context = new DeferredContext(null, false);
        }
        // synchronize to prevent from concurrent state & requests queue accesses
        synchronized (this) {
            if (state == NOT_INITIALIZED) {
                initialize();
            }
            if (state == DEFERRED) {
                // delay the execution (add to the queue)
                DeferredEntry entry = new DeferredEntry();
                entry.command = dc;
                entry.context = context;
                requests.add(entry);
                return;
            }
        }
        // execute immediately (and without sync)
        processCommand(dc, context);
    }
    
    /** Removes request from the queue, it means that request will not be
     * performed by this class.
     * @param request Request to remove from queue
     * @return true if request was found and removed, false otherwise
     */
    public boolean removeRequest (DeferredCommand command) {
        DeferredEntry curEntry = null;
        synchronized (this) {
            for (Iterator iter = requests.iterator(); iter.hasNext(); ) {
                curEntry = (DeferredEntry)iter.next();
                if (curEntry.command.equals(command)) {
                    requests.remove(curEntry);
                    return true;
                }
            }
        }
        return false;
    }

    /** @return true if window manager is in the state that allows
    * immediate opening of top components and showing workspaces */
    public boolean canImmediatelly () {
        return state == READY;
    }
    
    /** Initializes listening to window manager states
    * and update inner state */
    private void initialize () {
        requests = new ArrayList(40);
        stateManager = ((WindowManagerImpl)TopManager.getDefault().
                        getWindowManager()).stateManager();
        updateState();
        stateManager.addStateListener(this);
    }

    public void stateChanged (int newState) {
        DeferredEntry[] entries = null;
        synchronized (this) {
            int oldState = state;
            updateState();
            // execute all requests
            // if we changing back to READY state
            if ((state != oldState) && (state == READY)) {
                entries = (DeferredEntry[])requests.toArray(new DeferredEntry[0]);
                requests.clear();
            }
        }
        // process commands without sync
        if (entries != null) {        
            for (int i = 0; i < entries.length; i++) {
                processCommand(entries[i].command, entries[i].context);
            }
        }
    }

    /** Helper method, updates inner state according to
    * current window manager state */
    private void updateState () {
        int wmState = stateManager.getState();
        if ((((wmState & StateManager.VISIBLE) != 0) || 
              ((wmState & StateManager.NOT_PRESENT) != 0)) &&
                ((wmState & (StateManager.READY | StateManager.SWITCHING)) != 0)) {
            state = READY;
        } else {
            state = DEFERRED;
        }
    }

    /** Processed command, using event thread if requested in context */    
    private void processCommand (final DeferredCommand dc, final DeferredContext context) {
        // execute immediately
        // (use event thread if requested)
        if (context.isRunInEventThread() && !java.awt.EventQueue.isDispatchThread()) {
            javax.swing.SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    dc.performCommand(context);
                }
            });
        } else {
            dc.performCommand(context);
        }    
    }

    /** Interface for commands to be performed using DeferredPerformer.
    * Clients that want to perform some action which relies on consistent
    * state of window manager will need to implement this interface.
    */
    public interface DeferredCommand {

        /** Performs the command */
        public void performCommand (DeferredContext context);

    } // end of DeferredCommand inner interface

    /** Context for deferred commands */    
    public static class DeferredContext extends Object {

        /** Holds value of property data. */
        private Object data;
        /** Holds value of property runInEventThread. */
        private boolean runInEventThread;
        
        public DeferredContext (Object data, boolean runInEventThread) {
            this.data = data;
            this.runInEventThread = runInEventThread;
        }
        
        /** Getter for property data.
        * @return Value of property data.
        */
        public Object getData() {
            return data;
        }
        /** Setter for property data.
        * @param data New value of property data.
        */
        public void setData(Object data) {
            this.data = data;
        }
        /** Getter for property runInEventThread.
        * @return Value of property runInEventThread.
        */
        public boolean isRunInEventThread() {
            return runInEventThread;
        }
        /** Setter for property runInEventThread.
        * @param runInEventThread New value of property runInEventThread.
        */
        public void setRunInEventThread(boolean runInEventThread) {
            this.runInEventThread = runInEventThread;
        }
        
    } // end of DeferredContext inner class

    /** Stupid data structure, serves as item in deferred list */
    private static final class DeferredEntry {
        DeferredCommand command;
        DeferredContext context;
    }


}

/*
* Log
*  2    Gandalf   1.1         11/30/99 David Simonek   neccessary changes needed
*       to change main explorer to new UI style  (tabs are full top components 
*       now, visual workspace added, layout of editing workspace chnaged a bit)
*  1    Gandalf   1.0         11/6/99  David Simonek   
* $ 
*/ 
