/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.ImageIcon;
import javax.swing.*;

import org.openide.TopManager;
import org.openide.awt.MenuBar;
import org.openide.filesystems.*;
import org.openide.loaders.DataFolder;
import org.openide.loaders.FolderInstance;
import org.openide.nodes.*;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.SharedClassObject;
import org.openide.util.actions.*;
import org.openide.windows.Workspace;
import org.openide.windows.TopComponent;
import org.openide.awt.ToolbarPool;

import org.netbeans.core.actions.SystemExit;
import org.netbeans.core.windows.frames.DesktopPane;
import org.netbeans.core.DesktopNode;
import org.netbeans.core.windows.toolbars.ToolbarConfiguration;


/** The MainWindow of IDE. Holds toolbars, main menu and also entire desktop
 * if in MDI user interface. Singleton.
 * This class is final only for performance reasons, can be unfinaled
 * if desired.
 *
 * @author Ian Formanek, Petr Hamernik
 */
public final class MainWindow extends JFrame {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -1160791973145645501L;

    /** The only instance of main window in the system */
    private static MainWindow instance;
    /** window manager association */
    private static WindowManagerImpl wm;
    
    /** help context for main window */
    private HelpCtx mainWindowHelp;
    /** The MainWindow's menu */
    private MenuBar menuBar;
    /** The MainWindow's toolbar */
    private JPanel toolbars;
    /** Current project desktop node. */
    transient private Node projectNode = null;
    /** Current listener to changes in project name (other than project itself changing). */
    transient private NodeListener projectNameListener = null;    
    /** Desktop panel */
    private DesktopPane desktopPane;
    /** Helper flag - true if main window is fully constructed and initialized */
    private boolean initialized;
    /** Helper flag - true if desktop is currently visible */
    private boolean desktopVisible;

    /** Singleton. Allows to construct main window only from this class.
     */
    private MainWindow () {
        super();
    }
    
    /** Getter for only instance of main window in the system */
    static MainWindow getDefault () {
        if (instance == null) {
            instance = new MainWindow();
            wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        }
        return instance;
    }

    /**
     * Initializes menu and toolbar systems.
     */
    public void initialize () {
        if (Boolean.getBoolean ("netbeans.windows.small_main_window")) {
            menuBar = new StatusMenuBar ();
        } else {
            menuBar = new MenuBar(null);
        }
        // initialize frame
        javax.swing.ImageIcon ideIcon = new javax.swing.ImageIcon (
            Toolkit.getDefaultToolkit ().getImage (
                NbBundle.getLocalizedFile("org.netbeans.core.resources.frames.ide" + (org.openide.util.Utilities.isLargeFrameIcons() ? "32" : ""), // NOI18N
                                          "gif", // NOI18N
                                          Locale.getDefault(),
                                          MainWindow.class.getClassLoader()
            )));
        setIconImage (ideIcon.getImage ());
        projectNameListener = new NodeAdapter () {
                public void propertyChange (PropertyChangeEvent ev2) {
                    if (Node.PROP_DISPLAY_NAME.equals (ev2.getPropertyName ())) {
                        updateTitle ();
                    }
                }
            };
        TopManager.getDefault ().addPropertyChangeListener (new PropertyChangeListener () {
                public void propertyChange (PropertyChangeEvent ev) {
                    if (TopManager.PROP_PLACES.equals (ev.getPropertyName ())) {
                        if (projectNode != null) {
                            projectNode.removeNodeListener (projectNameListener);
                        }
                        projectNode = TopManager.getDefault ().getPlaces ().nodes ().projectDesktop ();
                        if (projectNode != null) {
                            projectNode.addNodeListener (projectNameListener);
                        }
                        updateTitle ();
                    }
                }
            });
        projectNode = TopManager.getDefault ().getPlaces ().nodes ().projectDesktop ();
        if (projectNode != null) {
            projectNode.addNodeListener (projectNameListener);
        }
        updateTitle ();
        setDefaultCloseOperation (DO_NOTHING_ON_CLOSE);

        addWindowListener (new WindowAdapter() {
                               public void windowClosing(WindowEvent evt) {
                                    ((SystemExit) SharedClassObject.findObject (
                                        SystemExit.class, true)
                                    ).performAction();
                               }

                               public void windowActivated (WindowEvent evt) {
                                   if (mainWindowHelp == null) {
                                       mainWindowHelp = new HelpCtx(MainWindow.class);
                                       HelpCtx.setHelpIDString (getRootPane (), mainWindowHelp.getHelpID ());
                                   }
                               }

                           }
                          );
    }
    
    
    /** Shows the main window.
     */
    public void showWindow () {
        if (menuBar == null) {
            initialize();
        }
        menuBar.waitFinished();
        setJMenuBar(menuBar);
        
        if (! Boolean.getBoolean ("netbeans.windows.small_main_window")) {
            // Register DTDs and so on. See Javadoc for this method for details.
            ToolbarConfiguration.init ();
            // Normal appearance.
            ToolbarPool tp = ToolbarPool.getDefault();
            tp.waitFinished ();
            tp.setConfiguration ("Standard"); // NOI18N

            toolbars = new JPanel();
            toolbars.setLayout(new BorderLayout());
            toolbars.add(tp, BorderLayout.CENTER);
            JPanel southPanel = new JPanel();
            southPanel.setLayout(new BorderLayout());
            WorkspaceSwitcher ws = new WorkspaceSwitcher(wm);
            southPanel.add(ws.getComponent(), BorderLayout.WEST);
            Component status = StatusLine.createLabel();
            southPanel.add(status, BorderLayout.CENTER);
            toolbars.add(southPanel, BorderLayout.SOUTH);
            getContentPane().add(toolbars, BorderLayout.NORTH);
        }

        if (!wm.isMainPositioned()) {
            pack();
        }
        show();
        getRootPane().requestDefaultFocus();
        initialized = true;
    }
    
    /** Updates the MainWindow's title */
    void updateTitle () {
        SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    String buildNumber = System.getProperty ("netbeans.buildnumber"); // NOI18N
                    if (projectNode == null || (projectNode instanceof DesktopNode)) {
                        setTitle (NbBundle.getMessage (MainWindow.class, "CTL_MainWindow_Title", buildNumber));
                    } else {
                        setTitle (NbBundle.getMessage (MainWindow.class, "CTL_MainWindow_Title_Project", buildNumber, projectNode.getDisplayName ()));
                    }
                }
            });
    }

    /** @return Screen location of left top corner of desktop component
     * (center component of content pane). Useful only for MDI.
     */
    Point getDesktopScreenLocation () {
        Point loc = new Point(0, 0);
        if (toolbars != null) {
            loc.y += toolbars.getSize().height;
        }
        SwingUtilities.convertPointToScreen(loc, getContentPane());
        return loc;
    }
    
    /** @return Size of desktop component - the component in the center of
     * content pane, where all frames live in MDI. Useful only for MDI mode,
     * in SDI size is zero.
     */
    Dimension getDesktopSize () {
        Dimension result = getContentPane().getSize();
        if (toolbars != null) {
            result = new Dimension(
                result.width, result.height - toolbars.getSize().height
            );
        }
        return result;
    }

    /** Takes care of the MainWindow size
     * the height is always equal to preferred height for SDI, anything
     * between min and max for MDI, width can be anything between min and max
     * for both MDI and SDI.
    */
    public void validate() {
        super.validate ();
        Dimension curSize = getSize();
        Dimension minSize = getMinimumSize();
        Dimension maxSize = getMaximumSize();
        Dimension newSize = new Dimension(
            Math.min(maxSize.width, Math.max(curSize.width, minSize.width)),
            Math.min(maxSize.height, Math.max(curSize.height, minSize.height))
        );
        if (!newSize.equals(curSize)) {
            setSize(newSize);
            validate();
        }
    }

    /** Returns preferredSize as the preferred height and the width of the screen */
    public Dimension getPreferredSize() {
        if (wm.getUIMode() == UIModeManager.MDI_MODE) {
            return org.openide.util.Utilities.getScreenSize();
        } else {
            return new Dimension(
                org.openide.util.Utilities.getScreenSize().width,
                super.getPreferredSize().height
            );
        }
    }

    /** @return Minimum reasonable size of main window, estimated so that
     * standard top level menu is still visible */
    public Dimension getMinimumSize() {
        if (wm.getUIMode() == UIModeManager.MDI_MODE) {
            return new Dimension(320, 320);
        } else {
            return new Dimension(320, super.getPreferredSize().height);
        }
    }

    /** @return Maximum possible size of IDE's main window. In SDI mode, main
     * window don't allow its height to be other than preferred size, in MDI
     * it's possible to resize main window up to current screen size */
    public Dimension getMaximumSize() {
        if (wm.getUIMode() == UIModeManager.MDI_MODE) {
            return org.netbeans.core.windows.util.WindowUtils.getMaximizedSize(this);
        } else {
            return new Dimension(
                org.netbeans.core.windows.util.WindowUtils.getMaximizedSize(this).width,
                super.getPreferredSize().height
            );
        }
    }

    /** Disables desktop pane auto-refocus activity in special SDI x MDI mixed
     * mode, where this behaviour is not desirable */ 
    protected void processWindowEvent (WindowEvent evt) {
        DesktopPane dp = null;
        // condition says - MDI mode while last activated is SDI frame (mixed mode)
        if (wm.getUIMode() == UIModeManager.MDI_MODE) {
            TopComponent activatedTc = TopComponent.getRegistry().getActivated();
            if ((activatedTc != null) && 
                !this.equals(SwingUtilities.windowForComponent(activatedTc))) {
                dp = ((WorkspaceImpl)wm.getCurrentWorkspace()).desktopPane();
                dp.setAutoFocusEnabled(false);
            }
        }
        // do parent's job
        super.processWindowEvent(evt);
        // reset flag again
        if (dp != null) {
            dp.setAutoFocusEnabled(true);
        }
    }
    
    /** Includes the status line directly at the end of the menu bar. */
    private static class StatusMenuBar extends MenuBar {
        private boolean hasStatus = false;
        public StatusMenuBar () {
            super (null);
        }
        public void removeAll () {
            super.removeAll ();
            hasStatus = false;
        }
        public void validate () {
            if (! hasStatus) {
                hasStatus = true;
                add (new StatusLine ());
            }
            super.validate ();
        }
    } // end of StatusMenuBar

}
