/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.*;
import java.awt.Rectangle;
import java.awt.Image;
import java.awt.Frame;
import java.awt.Window;
import java.awt.Container;
import java.awt.Toolkit;
import java.awt.Point;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectStreamException;
import java.io.ObjectStreamField;
import java.io.Serializable;
import java.lang.ref.WeakReference;
import java.net.URL;
import java.text.MessageFormat;

import javax.swing.SwingUtilities;

import org.openide.windows.*;
import org.openide.util.Mutex;
import org.openide.util.io.SafeException;
import org.openide.util.NbBundle;
import org.openide.TopManager;
import org.openide.ErrorManager;

import org.netbeans.core.windows.util.*;
import org.netbeans.core.windows.frames.TopComponentContainer;
import org.netbeans.core.windows.frames.MultiTabbedContainerImpl;
import org.netbeans.core.windows.frames.FrameType;
import org.netbeans.core.windows.frames.FrameTypeEvent;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.FrameTypeListener;
import org.netbeans.core.windows.frames.DesktopPane;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.NbPresenter;

/** This class is an implementation of Mode interface.
* It's responsible for managing top component container.
* This implementation supports switching of various types of
* top component containers.
*
* @author Dafe Simonek
*/
public final class ModeImpl implements Mode, FrameTypeListener, ComponentListener,
                                        DeferredPerformer.DeferredCommand {

    /** serial version UID */
    static final long serialVersionUID = 2721445375634234372L;

    /** @deprecated Has no function now */
    public static final int SINGLE = 1;
    /** @deprecated Has no function now */
    public static final int MULTI_TAB = 2;
    /** @deprecated Has no function now */
    public static final int SIDE_BY_SIDE = 4;
    /** @deprecated Has no function now */
    public static final int INTERNAL_FRAME = 8;
    
    /** Top components changes property name */
    public static final String PROP_TOP_COMPONENTS = "topComponents"; // NOI18N
    public static final String PROP_FRAME_INSTANCE = "frameInstance"; // NOI18N
    public static final String PROP_CONTAINER_INSTANCE = "containerInstance"; // NOI18N
    public static final String PROP_CONSTRAINTS = "containerInstance"; // NOI18N

    /** Default name for the components with no name */
    private static String untitledComponent;
    /** format for the title of the default workspace name */
    private static MessageFormat windowTitle;
    
    /** Icon of the mode */
    private URL icon;
    /** icon as an image, created from URL */
    private Image iconImage;
    /** default icon for frames */
    private static Image defaultIcon;
    /** Programmatic unique name of the mode */
    private String name;
    /** Human presentable name of the mode */
    private String displayName;
    /** The bounds of the mode - should be kept synchronized with
    * top component container when container exists and visible */ 
    private Rectangle bounds;
    /** true if this mode was created by user, not system */
    private boolean userDefined;
    /** The workspace which this mode belongs to */
    private Workspace workspace;
    /** Name of workspace which this mode belongs to.
    * Used only during deserialization when delayed validation is needed */
    private String workspaceName;
    /** Composited container for top components. */
    private TopComponentContainer tcc;
    /** Composited frame type which represents this mode as frame of some type.
     * Can be null if frame was not used yet */
    private FrameType frame;
    /** String representation of asociated container type */
    private String containerType;
    /** String representation of asociated container type */
    private String frameType;
    /** State of associated frame type as it was deserialized or the last state of
     * associated frame, which  was destroyed. It is used in setVisible, when mode
     * is shown first time. Value of frameTypeState is also serialized, when mode
     * doesn't have frame associated.
     */
    private int frameTypeState = -1;
    
    /** Layout constraints, null by default */
    private Object constraints;
    
    /** List of top component classes which names should be ignored in certain
    * circumstances */
    private List ignoredTcList;
    /** Weak map which holds weak references to the top components which
    * are docked in this mode but closed, together with their mode constraints */
    private WeakHashMap closedComponents;
    /** Asociation with window manager implementation */
    private WindowManagerImpl wm;
    /** asociated property change support for firing property changes */
    private PropertyChangeSupport changeSupport;
    /** helper variable, true when asociated top component container
    * is showing on the screen, false otherwise */
    private boolean showing;
    /** helper variable, true when there is some pending focus request */
    private boolean deferredFocusRequest;
    /** helper variable, holds top component as context of pending
    * focus request */
    private TopComponent compToReceiveFocus;
    /** Listener to the changes of the name of top component */
    private NameListener nameListener;
    /** Helper variable which holds  */
    private String oldDisplayName;

    /** manager of versioned serialization */
    private static VersionSerializator serializationManager;

    /** Construct new mode with given properties
     * @deprecated Use constructor without containerType parameter
     */
    public ModeImpl (String name, String displayName, URL icon,
                     int containerType, boolean userDefined,
                     Workspace workspace) {
        this(name, displayName, icon, userDefined, workspace);
    }
    
    /** Construct new mode with given properties
     */
    public ModeImpl (String name, String displayName, URL icon,
                     boolean userDefined, Workspace workspace) {
        this.name = name;
        this.displayName = displayName;
        this.icon = icon;
        this.workspace = workspace;
        this.userDefined = userDefined;
        initialize();
    }
    

    /** Creates new mode as a shallow copy of original mode on
    * given workspace. New mode will have the same characterists,
    * (displayname, icon, container type, bounds).
    * Contained components are taken from original.
    * The name can differ (name uniquennes is ensured automatically).
    */
    public ModeImpl (Workspace workspace, ModeImpl original) {
        this(original.name, original.displayName, original.icon,
             original.userDefined, workspace);
        // set unigue name
        if (workspace.findMode(name) != null) {
            int i = 1;
            while (workspace.findMode(name = original.name + "_" + i++) != null); // NOI18N
        }
        // bounds
        this.bounds = (original.bounds == null) ? null
                      : new Rectangle(original.bounds);
        this.containerType = original.containerType;
        this.frameType = original.frameType;
        this.constraints = original.constraints;

        //Add mode to workspace before top components are opened.
        ((WorkspaceImpl)workspace).addMode((Mode)this);
        
        TopComponent[] tcs = original.getTopComponents();
        Object constr;
        for (int i = 0; i < tcs.length; i++) {
            constr = original.tcc.getConstraints(tcs[i]);
            if (canDock(tcs[i])) {
                if (this.dockInto(tcs[i], constr)) {
                    tcs[i].open(workspace);
                }
            }
        }        
    }
    
    /** Initialization, called from constructors and deserialization */
    private void initialize () {
        this.wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        changeSupport = new PropertyChangeSupport(this);
    }

    /** Get the programmatic name of the mode.
     * This name should be unique, as it is used to find modes etc.
     * @return programmatic name of the mode
     */
    public String getName () {
        return name;
    }

    /* @return true if given top component can be docked to this mode,
    * false otherwise */
    public boolean canDock (TopComponent tc) {
        return (tcc == null) ? true : tcc.canAdd( new TopComponent[] {tc} ); 
    }

    /** Attaches a component to a mode for this workspace.
    * If the component is in different mode on this desktop, it is 
    * removed from the original and moved to this one.
    *
    * @param tc top component to dock into this mode
    * @return true if top component was succesfully docked to this
    * mode, false otherwise */
    public boolean dockInto (final TopComponent tc) {
        return dockInto(tc, null);
    }
    
    /** Attaches a component to a mode for this workspace.
    * If the component is in different mode on this desktop, it is 
    * removed from the original and moved to this one.
    *
    * @param tc top component to dock into this mode
    * @param constr constraints used in TopComponentContainer
    * @return true if top component was succesfully docked to this
    * mode, false otherwise */
    public boolean dockInto (final TopComponent tc, final Object constraints) {
        if (!canDock(tc))
            return false;
        // synchronize with window manager state and also run in AWT queue
        // because of 'dangerous' swing operations with AWT Event queue
        wm.deferredPerformer().putRequest(
            new DeferredPerformer.DeferredCommand () {
                public void performCommand (DeferredPerformer.DeferredContext dc) {
                    ModeImpl mi = (ModeImpl)workspace.findMode(tc);
                    // find out if component is opened
                    boolean opened = wm.findManager(tc).isOpened(workspace);
                    if ((mi != null) && (!ModeImpl.this.equals(mi))) {
                        mi.release(tc);
                    }
                    if (opened) {
                        // compute mode bounds if not known yet
                        if (bounds == null) {
                            ((WorkspaceImpl)workspace).placeMode(ModeImpl.this, tc);
                        }
                        // remove top component in a case it was previously
                        // docked in this mode in closed state
                        Object savedConstraints = removeClosedComponent(tc);
                        // adds component to the container
                        addToContainer(tc, constraints == null ? savedConstraints : constraints);
                        // show container if we are on current workspace only
                        if (workspace.equals(wm.getCurrentWorkspace())) {
                            setVisible(true);
                        }
                    } else {
                        // component not opened
                        addClosedComponent(tc, constraints);
                    }
                    // notify
                    changeSupport.firePropertyChange(PROP_TOP_COMPONENTS, null, null);
                }
            },
            new DeferredPerformer.DeferredContext(null, true)
        );
        return true;
    }

    /** Releases given top component from current association
    * with this mode. 
    */
    private void release (TopComponent tc) {
        // find out if component is opened
        if (wm.findManager(tc).isOpened(workspace)) {
            int left = tcc.removeTopComponent(tc);
            if (left <= 0) {
                destroyFrame();
            } else {
                updateNameListener();
            }
        } else {
            removeClosedComponent(tc);
        }
        // notify
        changeSupport.firePropertyChange(PROP_TOP_COMPONENTS, null, null);
    }

    /** Closes given top component.
    * Closing here means removing from top component container and
    * adding to the set of closed components docked in this mode.
     * This method is thead safe and window manager safe.
    */
    public void close (final TopComponent tc) {
        // synchronize with window manager state and also run in AWT queue
        // because of 'dangerous' swing operations
        wm.deferredPerformer().putRequest(
            new DeferredPerformer.DeferredCommand () {
                public void performCommand (DeferredPerformer.DeferredContext dc) {
                    // check validity of request, request may become
                    // invalid (container already destroyed) due to delayed processing
                    if (tcc == null) {
                        return;
                    }
                    if (tcc.containsTopComponent(tc)) {
                        Object constraints = tcc.getConstraints(tc);
                        int left = tcc.removeTopComponent(tc);
                        if (left <= 0) {
                            destroyFrame();
                        } else {
                            updateNameListener();
                        }
                        addClosedComponent(tc, constraints);
                    }
                    // notify
                    changeSupport.firePropertyChange(PROP_TOP_COMPONENTS, null, null);
                }
            },
            new DeferredPerformer.DeferredContext(null, true)
        );
    }

    /** Closes this mode - it means, that opened top components
    * are closed and closed components docked to this mode are
    * removed 
    * @return true if mode was succesfully cleared and closed,
    * false if some top component refused to close
    */
    public boolean close () {
        // clear closed components
        if (closedComponents != null)
            closedComponents.clear();
        if (tcc == null) {
            return true;
        }
        // try to close all opened top components
        // selected top component will be closed at last
        // to prevent from focus transfering between components during closing
        TopComponent[] tcs = tcc.getTopComponents();
        TopComponent selected = tcc.getSelectedTopComponent();
        boolean result = true;
        for (int i = 0; i < tcs.length; i++) {
            if ((!tcs[i].equals(selected)) && (!tcs[i].close(workspace))) {
                result = false;
            }
        }
        // close selected top component, if possible
        if ((selected != null) && (!selected.close(workspace))) {
            result = false;
        }
        return result;
    }

    /** Sets the bounds of the mode.
     * @param s the bounds for the mode 
     */
    public void setBounds (Rectangle rect) {
        if ((bounds != null) && (bounds.equals(rect))) {
            return;
        }
        Rectangle old = bounds;
        bounds = rect;
        // notify frame if possible
        if (frame != null) {
            frame.getRootPaneContainer().setBounds(bounds);
        }
        // notify others interested
        changeSupport.firePropertyChange(PROP_BOUNDS, old, bounds);
    }

    /** Getter for current bounds of the mode.
     * @return the bounds of the mode
     */
    public Rectangle getBounds () {
        return bounds;
    }

    /** Getter for asociated workspace.
     * @return The workspace instance to which is this mode asociated.
     */
    public Workspace getWorkspace () {
        return workspace;
    }

    /** @return array of top components which are currently
     * docked in this mode. May return empty array if no top component
     * is docked in this mode.
     */
    public TopComponent[] getTopComponents () {
        TopComponent[] containerTcs =
            (tcc == null) ? new TopComponent[0] : tcc.getTopComponents();
        if (closedComponents == null) {
            return containerTcs;
        }
        // merge container and closed tcs into one array
        TopComponent[] result = null;
        synchronized (closedComponents) {
            // implemetation of WeakHashMap should assure that garbage collected
            // keys (top components) were removed
            Set closed = closedComponents.keySet();
            // merge!
            result = new TopComponent[containerTcs.length + closed.size()];
            System.arraycopy(containerTcs, 0, result, 0, containerTcs.length);
            System.arraycopy(closed.toArray(new TopComponent[closed.size()]), 0,
                             result, containerTcs.length, closed.size());
        }
        return result;
    }

    /** @return an array of opened top components currently
    * docked in this mode.
    */
    public TopComponent[] getOpenedTopComponents () {
        return (tcc == null) ? new TopComponent[0] : tcc.getTopComponents();
    }

    /** The component requests focus. Request can be delayed
    * if top component container is not in consistent state.
    */
    public void requestFocus (final TopComponent comp) {
        if (SwingUtilities.isEventDispatchThread()) {
            doRequestFocus(comp);
        } else {
            SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    doRequestFocus(comp);
                }
            });
        }
    }

    /** Requests focus for whole mode. Request can be delayed
    * if top component container is not in consistent state.
    */
    public void requestFocus () {
        if (SwingUtilities.isEventDispatchThread()) {
            doRequestFocus();
        } else {
            SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    doRequestFocus();
                }
            });
        }
    }

    /** requests focus for given top component */
    private void doRequestFocus (TopComponent comp) {
        if (!showing) {
            deferredFocusRequest = true;
            compToReceiveFocus = comp;
        } else {
            // due to possibility of 'too much' delayed requests
            // we must perform additional check 
            if ((tcc != null) && tcc.containsTopComponent(comp)) {
                requestFocus();
                tcc.requestFocus(comp);
            }
        }
    }
    
    /** requests focus for this mode */
    private void doRequestFocus () {
        if (!showing) {
            deferredFocusRequest = true;
        } else {
            // due to possibility of 'too much' delayed requests
            // we must perform additional check 
            if (frame != null) {
                try {
                    frame.setSelected(true);
                } catch (PropertyVetoException exc) {
                    // not sure what I'm supposed to do ?
                }
            }
        }
    }

    /** @return asociated implementation of TopComponentContainer.
    * Can return null if no container is asociated at present.
    * (it means that no opened top component is docked in this mode) */
    public TopComponentContainer getContainerInstance () {
        return tcc;
    }

    /** @return asociated implementation of FrameType.
    * Can return null if no container is asociated at present.
    * (it means that no opened top component is docked in this mode) */
    public FrameType getFrameInstance () {
        return frame;
    }
    
    
    /** @return Human presentable name of this mode implementation */
    public String getDisplayName () {
        return displayName;
    }

    /** Sets new display name of this mode */
    public void setDisplayName (String s) {
        if (s == displayName) {
            return;
        }
        String old = displayName;
        displayName = s;
        changeSupport.firePropertyChange(PROP_DISPLAY_NAME, old, displayName);
    }

    /** @return icon of this mode */
    public Image getIcon () {
        if ((iconImage == null) && (icon != null)) {
            try {
                iconImage = Toolkit.getDefaultToolkit().getImage(icon);
            } catch (Exception exc) {
                // sometimes null pointer exceptions are thrown for no good reason
                // we return null and log exception for further processing
                String pattern = NbBundle.getBundle(ModeImpl.class).getString("FMT_NoIconWarning");
                TopManager.getDefault().getErrorManager().annotate(
                    exc,
                    MessageFormat.format(pattern, new Object[] { icon })
                );
                TopManager.getDefault().getErrorManager().notify(
                    ErrorManager.INFORMATIONAL,
                    exc
                );
                    
            }
        }
        return iconImage;
    }

    public URL getIconURL () {
        return icon;
    }
    
    /** Updates UI of asociated top component container,
    * if possible */
    public void updateUI () {
        if (frame != null)
            frame.updateUI();
    }

    /** Sets the state of asociated top component container
    */
    public void setState (int state) {
        if (frame != null) {
            try {
                frame.setState(state);
            }
            catch (PropertyVetoException pve) {
                pve.printStackTrace();
            }
        }
    }
    
    /** Shows or hides asociated top component container,
    * if possible (if container exists)
    */
    public void setVisible (boolean state) {
        if (frame == null) {
            return;
        }
        // check and convert bounds, if needed
        BoundsConvertor.getInstance().updateBounds(this);
        Container frameContainer = frame.getRootPaneContainer();
        if ((state != frameContainer.isVisible()) && (!state || shouldShow())) {
            // handle MDI frame types
            if ((WindowTypesManager.DESKTOP_FRAME.equals(frameType)) || 
                (WindowTypesManager.INTERNAL_FRAME.equals(frameType))) {
                WorkspaceImpl w = (WorkspaceImpl)workspace;
                if (state) {
                    // ensure that desktop pane is attached to comp hierarchy
                    if (w.desktopPane().getParent() == null) {
                        WindowManagerImpl.mainWindow().getContentPane().add(w.desktopPane());
                    }
                    // add or remove frame to/from desktop
                    if (WindowTypesManager.DESKTOP_FRAME.equals(frameType)) {
                        w.desktopPane().addFrameType(frame, constraints);
                    } else {
                        w.desktopPane().addFrameType(frame);
                    }
                } else {
                    w.desktopPane().removeFrameType(frame);
                }
            }
            frameContainer.setVisible(state);
            //Bugfix #11153 before setAttached was called from DefaultContainerImpl
            //componentShown, componentHidden
            tcc.setAttached(state);
            
            // when mode is displayed first time, set its deserialized state
            // bugfix #10455, #10533
            if (state && -1 != frameTypeState) {
                final int fts = frameTypeState;
                frameTypeState = -1;
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        setState(fts);
                    }
                });
            }
        }
    }
    
    /** @return True if frame that represents this mode is visible, false
     * otherwise
     */
    public boolean isVisible () {
        return (frame == null) ? false : frame.getRootPaneContainer().isVisible();
    }

    /** @return true if it is ok to show frame
    * false otherwise (all top components are closed)
    * Called from setVisible.
    */
    private boolean shouldShow () {
        TopComponent[] tcs = tcc.getTopComponents();
        for (int i = 0; i < tcs.length; i++) {
            if (WindowManagerImpl.findManager(tcs[i]).isOpened()) {
                return true;
            }
        }
        return false;
    }

    /** @return true if no component is docked to this mode
    * or all components docked to this mode are closed */
    public boolean isOrphan () {
        return (frame == null) || (tcc == null) || (tcc.getTopComponents().length <= 0);
    }

    /** @return true if this mode is currently in the state where
    * it contains exactly one OPENED top component.
    * (it can contains any number of closed components but this method
    * will still return true if there is exatly one opened)
    */
    public boolean isSingle () {
        return (frame != null) && (tcc != null) && (tcc.getTopComponents().length == 1);
    }

    public boolean isSingleType () {
        return isSingle() && !tcc.canAdd(null);
    }
    
    /** @return true if mode is user defined, false otherwise
    * (defined programmatically) */
    public boolean isUserDefined () {
        return userDefined;
    }

    /** Fills this mode with top components contained in given source
    * mode.
    */
    void fillTopComponents (ModeImpl source) {
        // copy references to closed top components via dockInto calls
        if (source.closedComponents != null) {
            Map.Entry[] closedEntries = null;
            synchronized (source.closedComponents) {
                closedEntries = (Map.Entry[])source.closedComponents.entrySet().
                                toArray(new Map.Entry[source.closedComponents.size()]);
            }
            Map.Entry curEntry = null;
            for (int i = 0; i < closedEntries.length; i++) {
                dockInto((TopComponent)closedEntries[i].getKey(), closedEntries[i].getValue());
            }
        }
        // copy references to opened top components (and reopen them)
        TopComponent[] tcs = source.getOpenedTopComponents();
        for (int i = 0; i < tcs.length; i++) {
            dockInto(tcs[i], source.tcc.getConstraints(tcs[i]));
            tcs[i].open(workspace);
        }
    }

    /**** Implementation of frame type and component listeners
     * we listen to the events in frame and react properly
     */

    /** When frame is being deactivated  */
    public void frameDeactivated(FrameTypeEvent fe) {
    }
    
    /** When frame was closed  */
    public void frameClosed(FrameTypeEvent fe) {
    }
    
    /** When frame was brought back from icon to normal state  */
    public void frameDeiconified(FrameTypeEvent fe) {
    }
    
    /** When frame was brought back from maximize to normal state  */
    public void frameNormalized(FrameTypeEvent fe) {
    }
    
    /** When frame was opened - showed  */
    public void frameOpened(FrameTypeEvent fe) {
    }
    
    /** When frame was iconified  */
    public void frameIconified(FrameTypeEvent fe) {
    }
    
    /** When user tried to invoke frame closing
     * So try to close all contained top components, if they agree.
     */
    public void frameClosing(FrameTypeEvent fe) {
        // check if we are in consistent state
        if (!isInClosableState()) {
            return;
        }
        TopComponent[] tcs = tcc.getTopComponents();
        TopComponent selected = tcc.getSelectedTopComponent();
        boolean shouldClose = true;
        for (int i = 0; i < tcs.length; i++) {
            if ((!tcs[i].equals(selected)) && (!tcs[i].close())) {
                shouldClose = false;
                break;
            }
        }
        // close selected top component, if possible
        if ((shouldClose) && (selected != null) && (!selected.close())) {
            shouldClose = false;
        }
        // close frame if nothing left
        if (shouldClose) {
            destroyFrame();
        }   
    }
    
    /** When frame was made active - receives focus  */
    public void frameActivated(FrameTypeEvent fe) {
    }
    
    /** When frame was made active - receives focus  */
    public void frameMaximized(FrameTypeEvent fe) {
    }
    
    /** When container was resized, its size was changed */
    public void componentResized (ComponentEvent ce) {
        setBounds(ce.getComponent().getBounds());
    }

    /** When container was moved, its position was changed */
    public void componentMoved (ComponentEvent ce) {
        setBounds(ce.getComponent().getBounds());
    }

    /** Called when container was shown. */
    public void componentShown (ComponentEvent ce) {
        showing = true;
        if (deferredFocusRequest) {
            deferredFocusRequest = false;
            if (compToReceiveFocus != null) {
                requestFocus(compToReceiveFocus);
                compToReceiveFocus = null;
            } else {
                requestFocus();
            }
        }
    }

    /** Called when container was hidden. */
    public void componentHidden (ComponentEvent ce) {
        showing = false;
    }

    /** Sets new container type and transfers content
    * of current container to the new one
    * @return true if new container was switched succesfully
    */
    public void setContainerType (String containerType) {
        if (containerType.equals(this.containerType)) {
            return;
        }
        this.containerType = containerType;
        if (tcc != null) {
            TopComponentContainer old = tcc;
            tcc = createContainer(containerType);
            changeContainer(old, tcc);
        }
    }

    /** @return unique string representation of current container type */
    public String getContainerType () {
        if (containerType == null) {
            containerType = WindowTypesManager.getDefaultContainer();
        }
        return containerType;
    }
    
    /** Sets new frame type and associates new frame with existing container.
     * Also updates visual state in case that mode is showing on the screen
     * when executing this method.
     */
    public void setFrameType (String frameType) {
        if (frameType.equals(this.frameType)) {
            return;
        }
        String oldFrameType = this.frameType;
        this.frameType = frameType;
        // check and convert bounds
        BoundsConvertor.getInstance().updateBounds(this, oldFrameType, frameType);
        // update visual state if needed
        if (frame != null) {
            // hide first if now showing 
            boolean wasVisible = frame.getRootPaneContainer().isVisible();
            if (wasVisible) {
                setVisible(false);
            }
            // create new frame, copy icon
            FrameType old = frame;
            frame = createFrame(frameType);
            Image ourIcon = old.getIconImage();
            if (ourIcon != null) {
                frame.setIconImage(ourIcon);
            }
            changeFrame(old, frame);
            // throws away old frame
            old.dispose();
            // restore visibility
            if (wasVisible) {
                setVisible(true);
            }
            changeSupport.firePropertyChange(PROP_FRAME_INSTANCE, old, frame);
        }
    }
    
    /** @return unique string representation of current frame type */
    public String getFrameType () {
        if (frameType == null) {
            frameType = WindowTypesManager.getDefaultFrame();
        }
        return frameType;
    }

    
    /** Sets constraints for mode placement in its parent. Null constraints
     * means no constraints. Null constraints are default.
     * Also directly update visual state, if mode is visible when entering
     * this method */
    public void setConstraints (Object constraints) {
        if (this.constraints == constraints) {
            return;
        }
        // check validity of request - multiple modes with the same side
        // constraints on one workspace are not allowed
        if ((constraints != null) && !PerimeterLayout.CENTER.equals(constraints)) {
            ModeImpl foundMode = WindowUtils.findConstrainedMode(workspace, constraints);
            if (foundMode != null) {
                throw new IllegalArgumentException(
                    "Cannot attach " + getDisplayName() + " to " + constraints + // NOI18N
                    " side. " + foundMode.getDisplayName() + // NOI18N
                    " is already attached there."); // NOI18N
            }
        }
        Object oldValue = this.constraints;
        if ((frame != null) && frame.getRootPaneContainer().isVisible()) {
            setVisible(false);
            this.constraints = constraints;
            setVisible(true);
        } else {
            this.constraints = constraints;
        }
        changeSupport.firePropertyChange(PROP_CONSTRAINTS, oldValue, this.constraints);
    }

    /** @return Current constraints of this mode, null by default */
    public Object getConstraints () {
        return constraints;
    }

    /** Add listener to the property changes */
    public void addPropertyChangeListener (PropertyChangeListener pchl) {
        changeSupport.addPropertyChangeListener(pchl);
    }

    /** Remove listener to the property changes */
    public void removePropertyChangeListener (PropertyChangeListener pchl) {
        changeSupport.removePropertyChangeListener(pchl);
    }

    /** Adds given class to the list of classes which name is ignored
    * when top component container enters some special state 
    * (usually when there is only one top component docked in it)
    * @param tcClass Class of top component.
    * @return true if list of classes was changed as a result of this call
    */
    public boolean addIgnoredNameClass (Class tcClass) {
        if (ignoredTcList == null) {
            ignoredTcList = new ArrayList(5);
        }
        if (!TopComponent.class.isAssignableFrom(tcClass)) {
            throw new IllegalArgumentException("tcClass parameter is not a subclass of TopComponent class"); // NOI18N
        }
        return ignoredTcList.add(tcClass);
    }

    /** Removes given class from the list. */
    public boolean removeIgnoredNameClass (Class tcClass) {
        if (ignoredTcList == null) {
            return false;
        }
        return ignoredTcList.remove(tcClass);
    }

    /** Adds top component to the set of closed tcs docked
    * in this mode */ 
    void addClosedComponent (TopComponent tc, Object constraints) {
        if (closedComponents == null) {
            closedComponents = new WeakHashMap(10);
        }
        synchronized (closedComponents) {
            closedComponents.put(tc, constraints);
        }
    }

    /** Removes top component from the set of closed tcs docked
    * in this mode */ 
    Object removeClosedComponent (TopComponent tc) {
        if (closedComponents == null) {
            return null;
        }
        synchronized (closedComponents) {
            Object result = closedComponents.get(tc);
            closedComponents.remove(tc);
            return result;
        }
    }
    
    
    /* Helper method, adds specified top component to the container.
     * Creates frame and container, if needed. */
    private void addToContainer (TopComponent tc, Object constr) {
        // assure frame is created and in normal state
        if (frame == null) {
            frame = createFrame(getFrameType());
            changeFrame(null, frame);
            Image frameIcon = obtainFrameIcon(tc);
            if (frameIcon != null) {
                frame.setIconImage(frameIcon);
            }
            changeSupport.firePropertyChange(PROP_FRAME_INSTANCE, null, frame);
        }
        if (frame.getState() == FrameType.ICONIFIED) {
            try {
                frame.setState(FrameType.NORMAL);
            }
            catch (PropertyVetoException pve) {
                pve.printStackTrace();
            }
        }
        // assure container is created and perform adding  
        if (tcc == null) {
            tcc = createContainer(getContainerType());
            changeContainer(null, tcc);
        }
        // add only if not re-opened multiple times
        if (constr == null ? tcc.addTopComponent(tc) : tcc.addTopComponent(tc, constr)) {
            updateNameListener();
        }
    }

    /** Returns proper icon for this mode's frame or null if everything failed */ 
    private Image obtainFrameIcon (TopComponent firstTc) {
        Image modeIcon = getIcon();
        if (modeIcon != null) {
            return modeIcon;
        } else {
            Image tcIcon = firstTc.getIcon();
            if (tcIcon != null) {
                return tcIcon;
            }
        }
        return defaultIcon();
    }
    
    /** Returns default icon for frames */
    private static Image defaultIcon () {
        if (defaultIcon == null) {
            try {
                defaultIcon = Toolkit.getDefaultToolkit().createImage(new java.net.URL(
                    "nbresboot:/org/netbeans/core/resources/frames/default.gif" // NOI18N
                ));
            } catch (java.net.MalformedURLException exc) {
                // should never happen, so we just catch it and default icon
                // will remain null - but log the warning so that we know what is
                // going on
                TopManager.getDefault().getErrorManager().notify(
                    ErrorManager.WARNING,
                    exc
                );
            }
        }
        return defaultIcon;
    }
    
    /** @return Newly created top component container of specified
     * type (class name). 
     */
    private TopComponentContainer createContainer (String className) {
        return WindowTypesManager.createContainer(className);
    }
    
    /** @return Newly created frame type of specified type (class name) */
    private FrameType createFrame (String className) {
        FrameType result = WindowTypesManager.createFrame(className);
        // set properties that we expect
        result.setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        Container frame = result.getRootPaneContainer();
        frame.setVisible(false);
        if (bounds != null) {
            frame.setBounds(bounds);
        }
        return result;
    }
    
  
    /** Destroys asociated frame */
    private void destroyFrame () {
        if (frame == null) {
            return;
        }
        updateNameListener ();
        changeFrame(frame, null);
        
        //Fixed bug #10580 Call of frame.dispose is replanned to AWT thread
        //It is necessary because mode deserialization is done in RequestProcessor
        //when new Project is opened.
        if (SwingUtilities.isEventDispatchThread()) {
            frame.dispose();
        } else {
            final FrameType localFrame = frame;
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    localFrame.dispose();
                }
            });
        }
        //End of bugfix #10580

        // remember state of frame being destroyed, it will be used 
        // when frame will be displayed again, but change minimized
        // frames to normal (bugfix #11202)
        frameTypeState = frame.getState();
        if (FrameType.ICONIFIED == frameTypeState)
            frameTypeState = FrameType.NORMAL;
        
        FrameType old = frame;
        frame = null;
        changeSupport.firePropertyChange(PROP_FRAME_INSTANCE, old, null);
    }
    
    /** Destroys asociated container */
    private void destroyContainer () {
        if (tcc == null) {
            return;
        }
        changeContainer(tcc, null);
        TopComponentContainer old = tcc;
        tcc = null;
        changeSupport.firePropertyChange(PROP_CONTAINER_INSTANCE, old, null);
    }
    
    /** Updates title of multi frame. Title is constructed from
     * the multi frame name and the name of selected view. Returns updated
     * title string value.
     */
    String frameTitle () {
        if (windowTitle == null) {
            ResourceBundle bundle = NbBundle.getBundle(ModeImpl.class);
            windowTitle = new MessageFormat(bundle.getString("CTL_MultiTabTitle"));
            if (untitledComponent == null)
                untitledComponent = bundle.getString("CTL_UntitledComponent");
        }
        TopComponent current = getContainerInstance().getSelectedTopComponent();
        String selectedName = (current == null) ? "?" : current.getName(); // NOI18N
        if (selectedName == null)
            selectedName = untitledComponent;
        // solve really-single state (use name of the only TC)
        if (isSingleType()) {
            return selectedName;
        }
        // solve special states in which name of selected component should be hidden
        if ((current != null) && shouldIgnoreName(current)) {
            return getDisplayName();
        }
        // other cases (normally formatted title)
        return windowTitle.format(new Object[] { getDisplayName(), selectedName });
    }

    /** Refreshes mode's title in title bar */
    public void updateTitle () {
        if (frame != null) {
            frame.setTitle(frameTitle());
        }
    }
        
    /** attaches new frame to this mode, detaches old frame. */
    private void changeFrame (FrameType oldFrame, FrameType newFrame) {
        if (oldFrame != null) {
            oldFrame.removeFrameTypeListener(this);
            oldFrame.getRootPaneContainer().removeComponentListener(this);
        }
        if (newFrame != null) {
            newFrame.addFrameTypeListener(this);
            newFrame.getRootPaneContainer().addComponentListener(this);
            TopComponentContainer container = getContainerInstance();
            if (container != null) {
                newFrame.setContentPane(container.getContentPane());
                frame.setTitle(frameTitle());
            }
        }
    }

    /** attaches new container to this mode, detaches old container. */
    private void changeContainer (TopComponentContainer oldContainer, TopComponentContainer newContainer) {
        if (oldContainer != null) {
            oldContainer.attachToMode(null);
        }
        if (newContainer != null) {
            newContainer.attachToMode(this);
            FrameType frame = getFrameInstance();
            if (frame != null) {
                frame.setContentPane(newContainer.getContentPane());
                frame.setTitle(frameTitle());
            }
        }
    }
    
    /** Decides whether to ignore the name of given top component or not.
    * @param tc top component to decide on
    * @return true TC containers should now ignore the name of currently
    * selected top component, false otherwise.
    * Should be called only if tcc is not null.
    */
    public boolean shouldIgnoreName (TopComponent tc) {
        if (!isSingle()) {
            return false;
        }
        return getDisplayName().equalsIgnoreCase(tc.getName()) ||
               ((ignoredTcList != null) && ignoredTcList.contains(tc.getClass()));
    }

    /** Asigns or removes listener to the name of the top component.
    * If mode is single, its display name should be the same as
    * the name of contained component. */
    private void updateNameListener () {
        if (isSingle()) {
            if (nameListener == null) {
                nameListener = new NameListener();
            }
            nameListener.activate();
        } else {
            if (nameListener != null) {
                nameListener.passivate();
            }
        }
        // the style of mode display name is changing,
        // so notify about the change
        changeSupport.firePropertyChange(PROP_DISPLAY_NAME, null, getDisplayName());
    }

    /** Implementation of DeferredPerformer.DeferredCommand interface.
    * Actually opens managed top component. */
    public void performCommand (DeferredPerformer.DeferredContext context) {
        DeferredOpenContext openContext = (DeferredOpenContext)(context.getData());
        openContext.tc.open(openContext.workspace);
    }

    /** @return true if it is in 'closable' state, it means that no modal
    * windows are opened at that time
    * (it's an ugly HACK, needed on unixes to prevent top component container
    * from closing during the time when modal dialog is shown)
    */
    private boolean isInClosableState () {
        return NbPresenter.currentModalDialog == null;
    }

    /** Holds data context for delayed opening */
    private static final class DeferredOpenContext {
        TopComponent tc;
        Workspace workspace;
    }
    
    /** Handles mode's display name refresh based upon top components' name
     * changes */
    private final class NameListener implements PropertyChangeListener {

        WindowManagerImpl.TopComponentManager tcm;

        void activate () {
            if (tcm == null) {
                tcm = wm.findManager(tcc.getTopComponents()[0]);
                tcm.addPropertyChangeListener(this);
            }
        }

        void passivate () {
            if (tcm != null) {
                tcm.removePropertyChangeListener(this);
                tcm = null;
            }
        }

        public void propertyChange (PropertyChangeEvent evt) {
            String propName = evt.getPropertyName();
            if (WindowManagerImpl.TopComponentManager.PROP_NAME.equals(propName)) {
                // name of top component changed, update title 
                // and fire display name change
                if (frame != null) {
                    frame.setTitle(frameTitle());
                }
                changeSupport.firePropertyChange(
                    PROP_DISPLAY_NAME,
                    evt.getOldValue(),
                    evt.getNewValue()
                );
            }
        }

    } // end of NameListener inner class

    public static class ModeDefinition {
        String  name;
        String  displayName;
        URL     iconURL;
        String  containerType;
        Object  constraints;
    }
    
/************** Serialization ********/    
    
    /** Let instance of SerializationReplacer to deal with persistence of the
     * Mode */
    private Object writeReplace () throws ObjectStreamException {
        return new SerializationReplacer(this);
    }

    /** Called when first phase of WS deserialization is done.
    * Validates asociation with its workspace.
    */
    void validateSelf (Workspace workspace) {
        this.workspace = workspace;
    }

    /** Called when first phase of WS deserialization is done.
    * Validates its top component container, if possible
    */
    void validateData () {
        // validate container, if possible
        if (tcc != null) {
            tcc.validateData();

            // swap tcc if needed: multitabbed -> split after deserialization
            if (tcc instanceof MultiTabbedContainerImpl &&
                null != containerType &&
                containerType.equals(WindowTypesManager.SPLIT_CONTAINER)) {
                TopComponentContainer newTcc = createContainer(containerType);
                TopComponent tcs[] = tcc.getTopComponents();
                
                newTcc.attachToMode(this);
                for(int i = 0; i < tcs.length; i++) {
                    newTcc.addTopComponent(tcs[i], PerimeterLayout.CENTER);
                }
                tcc = newTcc;
            }
            
            if (tcc.getTopComponents().length > 0) {
                //Store current top selected component
                final TopComponent tc = tcc.getSelectedTopComponent();
                // put a request for reopening all top components in
                // the container (actual opening is delayed and performed
                // after deserializaton)
                TopComponent[] tcs = tcc.getTopComponents();
                DeferredOpenContext openContext = null;
                for (int i = 0; i < tcs.length; i++) {
                    openContext = new DeferredOpenContext();
                    openContext.tc = tcs[i];
                    openContext.workspace = workspace;
                    WindowManagerImpl.deferredPerformer().putRequest(
                        this, new DeferredPerformer.DeferredContext(openContext, true)
                    );
                }
                //Bugfix 9579, 21 Feb 2001 by Marek Slama
                //Method requestFocus is called when ALL top components are opened
                //and added to container.
                //Every addComponent calls setSelectedTopComponent so in the end we must
                //make sure that correct top component is selected in container.
                if (tc != null) {
                    wm.deferredPerformer().putRequest(
                        new DeferredPerformer.DeferredCommand () {
                            public void performCommand (DeferredPerformer.DeferredContext dc) {
                                synchronized (ModeImpl.this) {
                                    // check validity of request, request may become
                                    // invalid (container already destroyed) due to delayed processing
                                    if (tcc == null) {
                                        return;
                                    }
                                    requestFocus(tc);
                                }
                            }
                        },
                        new DeferredPerformer.DeferredContext(null, true)
                    );
                }
                //End
            } else {
                tcc = null;
            }
        }
        // validate frame
        if (frame != null) {
            frame.addFrameTypeListener(this);
            frame.getRootPaneContainer().addComponentListener(this);
        }
        // fire changes
        changeSupport.firePropertyChange(PROP_CONTAINER_INSTANCE, null, tcc);
        changeSupport.firePropertyChange(PROP_FRAME_INSTANCE, null, frame);
        changeSupport.firePropertyChange(PROP_TOP_COMPONENTS, null, null);
    }

    /** Class that acts as serialization manager for mode implementation */
    private static final class SerializationReplacer implements Serializable {
        /** Unique ID */
        static final long serialVersionUID =1230255666898346575L;
        /** Description of serializable fields for mode */
        private static final String NAME = "name";
        private static final String DISPLAY_NAME = "displayName";
        private static final String BOUNDS = "bounds";
        private static final String ICON_URL = "iconURL";
        private static final String USER_DEFINED = "userDefined";
        private static final String WORKSPACE_NAME = "workspaceName";
        private static final String TC_CONTAINER = "tcContainer";
        private static final String FRAME_TYPE = "frameType";
        private static final String FRAME_TYPE_STATE = "frameTypeState";
        private static final String CONTAINER_TYPE = "containerType";
        private static final String CONSTRAINTS = "constraints";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(NAME, String.class),
            new ObjectStreamField(DISPLAY_NAME, String.class),
            new ObjectStreamField(BOUNDS, Rectangle.class),
            new ObjectStreamField(ICON_URL, URL.class),
            new ObjectStreamField(USER_DEFINED, Boolean.class),
            new ObjectStreamField(WORKSPACE_NAME, String.class),
            new ObjectStreamField(TC_CONTAINER, TopComponentContainer.class),
            new ObjectStreamField(FRAME_TYPE, String.class),
            new ObjectStreamField(FRAME_TYPE_STATE, Integer.class),
            new ObjectStreamField(CONTAINER_TYPE, String.class),
            new ObjectStreamField(CONSTRAINTS, Object.class)
        };

        /** backward compatibility, we must recognize old container's name */
        private static final String OLD_SPLIT_CONTAINER = "org.netbeans.core.windows.frames.SplittedContainerImpl";
        
        /** Asociation with container which data needs to be serialized */
        transient ModeImpl source;
        
        SerializationReplacer (ModeImpl source) {
            this.source = source;
        }

        /** Serialization of all workspaces */
        private void writeObject (ObjectOutputStream oos)
                     throws IOException {
            FrameType f;
            
            // write fields
            ObjectOutputStream.PutField pf = oos.putFields();
            pf.put(NAME, source.getName());
            pf.put(DISPLAY_NAME, source.getDisplayName());
            pf.put(BOUNDS, source.getBounds());
            pf.put(ICON_URL, source.getIconURL());
            pf.put(USER_DEFINED, new Boolean(source.isUserDefined()));
            pf.put(WORKSPACE_NAME, source.workspace.getName());
            pf.put(TC_CONTAINER, source.tcc);
            pf.put(FRAME_TYPE, source.getFrameType());
            
            if (null != (f = source.getFrameInstance()))
                pf.put(FRAME_TYPE_STATE, new Integer(f.getState()));
            else if (-1 != source.frameTypeState)
                pf.put(FRAME_TYPE_STATE, new Integer(source.frameTypeState));
            else
                pf.put(FRAME_TYPE_STATE, new Integer(FrameType.NORMAL));
                
            pf.put(CONTAINER_TYPE, source.getContainerType());
            pf.put(CONSTRAINTS, source.getConstraints());
            oos.writeFields();
        }

        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField gf = ois.readFields();
            String name = (String)gf.get(NAME, null);
            String workspaceName = (String)gf.get(WORKSPACE_NAME, null);
            Workspace workspace = TopManager.getDefault().getWindowManager().
                                  findWorkspace(workspaceName);
            source = (workspace == null) ? null : (ModeImpl)workspace.findMode(name);
            
            if (source == null) {
                // mode don't exist, create new one
                source = new ModeImpl(
                    name, (String)gf.get(DISPLAY_NAME, null),
                    (URL)gf.get(ICON_URL, null),
                    ((Boolean)gf.get(USER_DEFINED, null)).booleanValue(),
                    workspace
                );
            } else {
                // old mode content cleanup
                source.close();
                source.destroyFrame();
                source.destroyContainer();
                source.setDisplayName((String)gf.get(DISPLAY_NAME, null));
            }
            source.setBounds((Rectangle)gf.get(BOUNDS, null));
            source.frameType = (String)gf.get(FRAME_TYPE, null);
            
            source.containerType = (String)gf.get(CONTAINER_TYPE, null);
            if (OLD_SPLIT_CONTAINER.equals(source.containerType)) {
                source.containerType = WindowTypesManager.SPLIT_CONTAINER;
            }
            source.constraints = (Object)gf.get(CONSTRAINTS, null);
            source.tcc = (TopComponentContainer)gf.get(TC_CONTAINER, null);
            // assign workspace if needed for later validation
            if (workspace == null) {
                source.workspaceName = workspaceName;
            }

            if (null != source.frameType &&
                WindowUtils.hasObjectStreamField(gf, FRAME_TYPE_STATE)) {
                // current format - read the state of frame: iconified, normal, maximized
                source.frameTypeState = 
                    ((Integer)gf.get(FRAME_TYPE_STATE, new Integer(FrameType.NORMAL))).intValue();
            }
        }
    
        /** Resolves deserialized SerializationReplacer to the singleton
        * instance of WindowManagerImpl */
        private Object readResolve () throws ObjectStreamException {
            return source;                            
        }

    } // end of SerializationReplacer
    
/******** Old serialization, must be here for compatibility with 3.1 and older versions */    
    
    /** Accessor to the versioned serialization manager */
    private VersionSerializator serializationManager () {
        if (serializationManager == null) {
            serializationManager = createSerializationManager();
        }
        return serializationManager;
    }

    /** Creates new serialization manager filled with our versions */
    private static VersionSerializator createSerializationManager () {
        VersionSerializator result = new VersionSerializator();
        result.putVersion(new Version1());
        result.putVersion(new Version2());
        return result;
    }
    
    /** Basic version of persistence for mode implementation.
    * Method assignData(modeImpl) must be called prior to serialization */
    private static class Version1
        implements DefaultReplacer.ResVersionable {

        /* identification string */
        public static final String NAME = "Version_1.0"; // NOI18N

        /** variables of persistent state of the mode implementation */
        String name;
        String displayName;
        Rectangle bounds;
        URL icon;
        int containerType;
        boolean userDefined;
        String workspaceName;
        TopComponentContainer tcc;

        /** asociation with mode implementation, used when writing */
        ModeImpl mode;

        /** Identification of the version */
        public String getName () {
            return NAME;
        }

        /** Assigns data to be written. Must be called before writing */
        public void assignData (ModeImpl mode) {
            this.mode = mode;
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            // read mode fields
            name = (String)in.readObject();
            displayName = (String)in.readObject();
            bounds = (Rectangle)in.readObject();
            icon = (URL)in.readObject();
            containerType = ((Integer)in.readObject()).intValue();
            userDefined = ((Boolean)in.readObject()).booleanValue();
            workspaceName = (String)in.readObject();
            tcc = (TopComponentContainer)in.readObject();
        }

        /** Should be never called */
        public void writeData (ObjectOutput out) throws IOException {
            throw new InternalError();
        }

        public Object resolveData ()
        throws ObjectStreamException {
            Workspace workspace =
                TopManager.getDefault().getWindowManager().findWorkspace(workspaceName);
            ModeImpl result =
                (workspace == null) ? null : (ModeImpl)workspace.findMode(name);

            if (result == null && null != workspace) {
                result = DefaultCreator.createMode(workspace, name);
            }
                
            if (result == null) {
                // mode don't exist, create new one and fill it
                result = new ModeImpl(name, displayName, icon, 
                                        userDefined, workspace);
                // assign workspace if needed for later validation
                if (workspace == null) {
                    result.workspaceName = workspaceName;
                }
            }

            if (null != result) {
                result.bounds = bounds;
                result.tcc = tcc;
            }
            return result;
        }

    } // end of Version1 inner class

    /** Added list of ignored top component classes */
    private static class Version2 extends Version1 {

        /* identification string */
        public static final String NAME = "Version_2.0"; // NOI18N

        /** variables of persistent state of the mode implementation */
       List ignoredTcList;

        /** Identification of the version */
        public String getName () {
            return NAME;
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            // read version 1
            super.readData(in);
            // read list
            // things are messy here, because we must handle different
            // protocols. either there will be list of Class objects or
            // the list of Strings
            ArrayList readList = (ArrayList)in.readObject();
            if ((readList == null) || (readList.size() <= 0) ||
                (readList.get(0) instanceof Class)) {
                // old protocol
                ignoredTcList = readList;
            } else {
                // new protocol, convert strings back to Classes
                ignoredTcList = new ArrayList(readList.size());
                String curName = null;
                for (Iterator iter = readList.iterator(); iter.hasNext(); ) {
                    curName = (String)iter.next();
                    ignoredTcList.add(
                        Class.forName(curName, false, TopManager.getDefault().systemClassLoader())
                    );
                }
            }
        }

        /** write the data of the version to given output */
        public void writeData (ObjectOutput out)
        throws IOException {
            // write version 1
            super.writeData(out);
            // write list of names of ignored classes. do not save classes
            // because it can cause versioning problems
            ArrayList ignoredNames = null;
            if (mode.ignoredTcList != null) {
                ignoredNames = new ArrayList(mode.ignoredTcList.size());
                String curName = null;
                for (Iterator iter = mode.ignoredTcList.iterator(); iter.hasNext(); ) {
                    curName = ((Class)iter.next()).getName();
                    ignoredNames.add(curName);
                }
            }
            out.writeObject(ignoredNames);
        }

        public Object resolveData ()
        throws ObjectStreamException {
            ModeImpl result = (ModeImpl)super.resolveData();
            // add ignored list to the deserialization result, if possible
            if (result.ignoredTcList == null) {
                result.ignoredTcList = ignoredTcList;
            } else {
                if (ignoredTcList != null) {
                    result.ignoredTcList.addAll(ignoredTcList);
                }
            }
            return result;
        }

    } // end of Version2 inner class


    /** Implementation of persistent access to our version serializator */
    private static final class VSAccess implements DefaultReplacer.Access {
        /** version serializator, used only during writing */
        transient VersionSerializator vs;

        /** serialVersionUID */
        private static final long serialVersionUID = -7577235918945664917L;

        public VSAccess (VersionSerializator vs) {
            this.vs = vs;
        }

        public VersionSerializator getVersionSerializator () {
            return (vs == null) ? createSerializationManager() : vs;
        }

    } // end of VSAccess inner class

}
