/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.HashMap;
import java.util.Map;
import java.util.ResourceBundle;

import org.openide.util.NbBundle;
import org.openide.windows.Workspace;

import org.netbeans.core.windows.frames.MultiTabbedContainerImpl;
import org.netbeans.core.windows.frames.SplitContainerImpl;
import org.netbeans.core.windows.frames.TopComponentContainer;
import org.netbeans.core.windows.nodes.TabbedContainerUIPropertyEditor;

/** Performs setting of the UI modes (BasicTabbedPaneUI, ArrowsTabbedPaneUI,
 *  PopupTabbedPaneUI) of containers include JTabbedPane
 *  (MultiTabbedContainerImpl, SplitContainerImpl).
 *  encapsulates logic to convert between them.
 *
 *  @author Rostislav Levy
 */
public class TabbedContainerUIManager extends Object {

    /** javax.swing.plaf.basic.BasicTabbedPaneUI mode
     */
    public static final int MODE_BASIC = 1;
    /** Using org.netbeans.core.ui.ArrowsTabbedPaneUI mode
     */
    public static final int MODE_ARROWS = 2;
    /** Using org.netbeans.core.ui.PopupTabbedPaneUI mode
     */
    public static final int MODE_POPUP = 3;
    /** Default TabbedPaneUI
     */
    public static final int MODE_DEFAULT = MODE_BASIC;
    /** Minimum value of TabbedPaneUI mode
     */
    public static final int MODE_MIN = MODE_BASIC;
    /** Maximum value of TabbedPaneUI mode
     */
    public static final int MODE_MAX = MODE_POPUP;

    /** TabbedPaneUI mode (constants MODE_...)
     */
    private int tabbedContainerUI;
    /** List of TabbedPaneUI modes (mode name, mode value)
     */
    private Map names2Modes;
    /** Bundle with texts
     */
    private static ResourceBundle bundle;

    /** Default constructor
     */
    public TabbedContainerUIManager() {
        tabbedContainerUI = MODE_BASIC;
        names2Modes = null;
    }

    /** Sets TabbedPaneUI mode
     *  value - TabbedPaneUI mode (constants MODE_...)
     */
    public void setTabbedContainerUI (int value) {
        int countworkspace;
        int countmode;
        Object modes[];
        TopComponentContainer topcomponentcontainer;
        Workspace workspaces[];
        if (value >= MODE_MIN && value <= MODE_MAX) {
            if (tabbedContainerUI != value) {
                tabbedContainerUI = value;
                workspaces = WindowManagerImpl.getDefault().getWorkspaces();
                countworkspace = workspaces.length;
                for (int i = 0; i < countworkspace; i++) {
                    modes = workspaces[i].getModes().toArray();
                    countmode = modes.length;
                    for (int j = 0; j < countmode; j++) {
                        if ((topcomponentcontainer = ((ModeImpl)modes[j]).getContainerInstance()) instanceof MultiTabbedContainerImpl) {
                            ((MultiTabbedContainerImpl)topcomponentcontainer).changeUI();
                        } else {
                            if (topcomponentcontainer instanceof SplitContainerImpl)
                                ((SplitContainerImpl)topcomponentcontainer).changeUI();
                        }
                    }
                }
            }
        }
    }

    /** Reads TabbedPaneUI mode
     *  return - TabbedPaneUI mode (constants MODE_...)
     */
    public int getTabbedContainerUI () {
        return tabbedContainerUI;
    }

    /** Reads the list of TabbedPaneUI modes
     *  return - list of TabbedPaneUI modes
     */
    public Map getModeNames () {
        if (names2Modes == null) {
            names2Modes = new HashMap(10);
            names2Modes.put(bundle().getString("CTL_BasicUI"), new Integer(MODE_BASIC));
            names2Modes.put(bundle().getString("CTL_ArrowsUI"), new Integer(MODE_ARROWS));
            names2Modes.put(bundle().getString("CTL_PopupUI"), new Integer(MODE_POPUP));
        }
        return names2Modes;
    }

    /** Reads resource bundle
     *  return - resource bundle
     */
    private static final ResourceBundle bundle () {
        if (bundle == null) {
            bundle = NbBundle.getBundle(TabbedContainerUIManager.class);
        }
        return bundle;
    }
}
