/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import org.openide.TopManager;
import org.openide.windows.Workspace;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;

/** Performs setting of the UI mode types of the IDE (SDI, MDI) and
 * encapsulates logic to convert between them.
 *
 * @author Dafe Simonek
*/
public final class UIModeManager extends Object {

    /** Constant for traditional SDI interface of the IDE */
    public static final int SDI_MODE = 1;
    /** Constant for mupliple document interface mode of the IDE */
    public static final int MDI_MODE = 2;
    /** Constant for SDI with dialogs instead of frames */
    public static final int DIALOG_SDI_MODE = 4;
    
    /** current mode of UI of the IDE */
    private int uiMode;
    /** mappings between UI modes and frame types, <Integer, String> */
    private Map uiModes2Frames;
    /** <integer, string> mapping which maps display names to integer mode values */
    private Map names2Modes;
    
    public UIModeManager () {
        uiMode = SDI_MODE;
   }

    /** Sets new UI mode for whole IDE. This method will take all windows and
     * convert them to the new UI mode, so this can be expensive operation,
     * including repaint of every window in the system */
    public void setUIMode (int uiMode) {
        setUIMode(uiMode, true, true);
    }
    
    /** Sets new UI mode for whole IDE.
     * @param update True when all windows should be updated, false otherwise.
     */
    public void setUIMode (int uiMode, boolean update) {
        setUIMode(uiMode, update, true);
    }
        
    /** Sets new UI mode for whole IDE.
     * @param update True when all windows should be updated, false otherwise.
     */
    public void setUIMode (int uiMode, boolean update, boolean alterTypes) {
        if (this.uiMode == uiMode) {
            return;
        }
        this.uiMode = uiMode;
        WindowTypesManager.setDefaultFrame((String)uiModes2Frames().get(new Integer(uiMode)));
        if (update) {
            updateAll(alterTypes);
        } else if (alterTypes) {
            changeModeFrameTypes();
        }
    }
    
    /** @return Returns current UI mode IDE is in */
    public int getUIMode () {
        return uiMode;
    }
    
    /** Updates all windows (modes) in the system and convert them to the
     * currently set UI mode, so this can be expensive operation,
     * including repaint of every window in the system */
    public void updateAll () {
        updateAll(true);
    }
    
    /** Go through all workspaces and all modes and update their frametype
     * according to the ui mode IDE is currently in */
    public void changeModeFrameTypes () {
        WindowManagerImpl wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        Workspace[] workspaces = wm.getWorkspaces();
        if (workspaces == null) {
            // not initialized yet
            return;
        }
        Set modes = null;
        ModeImpl curMode = null;
        for (int i = 0; i < workspaces.length; i++) {
            modes = workspaces[i].getModes();
            for (Iterator iter = modes.iterator(); iter.hasNext(); ) {
                curMode = (ModeImpl)iter.next();
                curMode.setFrameType(findFrameType(curMode));
            }
        }
    }

    /** @return Map of <String, Integer>, where Integer identifies UI mode
     * and String is display name of this mode */
    public final Map getModeNames () {
        if (names2Modes == null) {
            names2Modes = new HashMap(10);
            names2Modes.put(
                NbBundle.getBundle(UIModeManager.class).getString("CTL_SdiMode"),
                new Integer(SDI_MODE)
            );
            names2Modes.put(
                NbBundle.getBundle(UIModeManager.class).getString("CTL_MdiMode"),
                new Integer(MDI_MODE)
            );
        }
        return names2Modes;
    }


    /** Utility static method. Finds proper frame type for given mode
     * which suits current ui mode */
    String findFrameType (ModeImpl mode) {
        String result = (String)uiModes2Frames().get(new Integer(uiMode));
        if (uiMode == MDI_MODE) {
            Object constraints = mode.getConstraints();
            if ((constraints != null) && (!PerimeterLayout.CENTER.equals(constraints))) {
                result = WindowTypesManager.DESKTOP_FRAME;
            }
        }
        return result;
    }
    
    private void updateAll (boolean alterTypes) {
        WindowManagerImpl wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
        if (curWorkspace == null) {
            // not initialized yet
            return;
        }
        curWorkspace.setVisible(false);
        if (alterTypes) {
            changeModeFrameTypes();
        }
        wm.mainWindow().pack();
        curWorkspace.setVisible(true);
    }

    /** Accessor for ui modes to frame types mapping */
    private Map uiModes2Frames () {
        if (uiModes2Frames == null) {
            uiModes2Frames = new HashMap(10);
            uiModes2Frames.put(new Integer(SDI_MODE), WindowTypesManager.TOP_FRAME);
            uiModes2Frames.put(new Integer(MDI_MODE), WindowTypesManager.INTERNAL_FRAME);
            //uiModes2Frames.put(new Integer(DIALOG_SDI_MODE), WindowTypesManager.DIALOG_FRAME);
        }
        return uiModes2Frames;
    }
    
}