/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.awt.*;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowEvent;
import java.text.MessageFormat;
import java.util.*;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.ObjectInputValidation;
import java.io.IOException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Arrays;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.SwingUtilities;
import javax.swing.RepaintManager;

import org.openide.TopManager;
import org.openide.windows.*;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.Mutex;
import org.openide.util.io.NbMarshalledObject;
import org.openide.NotifyDescriptor;

import org.netbeans.core.NbTopManager;
import org.netbeans.core.NbPresenter;

import org.netbeans.core.windows.WeakHash.Value;
import org.netbeans.core.windows.util.*;
import org.netbeans.core.windows.frames.PinRepaintRM;
import org.netbeans.core.windows.frames.FrameType;
import org.netbeans.core.windows.frames.TopComponentContainer;

/** Implementation of manager of windows in the IDE.
* Handles work with workspaces, serialization of all window system,
* allows to listen to workspace events.
* This class is final only for performance reasons. Can be freely
* unfinaled if desired.
*
* @author Jaroslav Tulach, Dafe Simonek
*/
public final class WindowManagerImpl extends WindowManager {
  /** The Corona IDE's main window */
  static MainWindow mainWindow;
  /** The only instance of the window manager implementation
  * in the system */
  static WindowManagerImpl defaultInstance;
  /** State machine support */
  static StateManager stateManager;
  /** Support for deferred opening of top components */
  static DeferredPerformer deferredPerformer;
  
  /** Manages transitions between UI IDE modes */
  private UIModeManager uiModeManager;

  /** Manages displaying modes of ContainerUI */
  private TabbedContainerUIManager tabbedContainerUIManager;

  /** array of workspaces */
  private Workspace[] workspaces;
  /** Current active workspace */
  private Workspace current;
  /** properties support */
  private PropertyChangeSupport changeSupport;
  /** the set of listeners which listen to the changes of
  * top component set in system */
  private HashSet tcListeners;

  /** The history of activated components mapped to their classes
  * (mapping class - top component) */
  private static Hashtable lastActivatedComponents;
  /** initialized already? */
  private boolean isCreated;
  /** map of workspace::Value (activated nodes) */
  private WeakHash workspace2Nodes;
  /** helper flag */
  private boolean serializationInProgress = false;
  /** true if main window was already positioned 
  * during deserialization */ 
  private boolean mainPositioned = false;
  /** Helper temporary variable, holds TC manager currently
  * being validated, used in createTopComponentManager() method */
  private TopComponentManager validatedManager;
  /** 'father' of all exceptions thrown during serialization / deserialization */
  private Exception mainEx;
  /** map of names of badly persistenced top components,
   * serves as additional annotation of main exception */
  private Map failedCompsMap;
  /** Screen size in at the time of window system last serialization */
  private Dimension oldScreenSize;
  /** Main window bounds in at the time of window system last serialization */
  private Rectangle oldMainWindowBounds;
  
  /** Standard creator for proper initialixation of the window system */
  private static DefaultCreator dc;
  
  static final long serialVersionUID =680725949680433701L;

  /** default base name for noname modes */
  private static final String DEFAULT_NAME = "untitled_mode"; // NOI18

  /** Default constructor. Don't use directly, use getDefault()
  * instead.
  */
  private WindowManagerImpl () {
    initialize();
  }

  /** Needed initialization, called from constrcutor and
  * during deserialization */
  private void initialize () {
    initToolbars();  
    changeSupport = new PropertyChangeSupport(this);
    workspace2Nodes = new WeakHash();
    RepaintManager rm = RepaintManager.currentManager(null);
    RepaintManager.setCurrentManager(new PinRepaintRM(rm));    
  }
  
  
  private void initToolbars() {
      
    /** toolbar dtd location */
    final String TOOLBAR_DTD               =
        "org/netbeans/core/windows/toolbars/toolbar.dtd"; // NOI18N
    /** toolbar dtd public id */
    final String TOOLBAR_DTD_PUBLIC_ID     =
        "-//Forte for Java//DTD toolbar//EN"; // NOI18N
    /** toolbar prcessor class */
    final Class  TOOLBAR_PROCESSOR_CLASS   =
        org.netbeans.core.windows.toolbars.ToolbarProcessor.class;
    /** toolbar icon base */
    final String TOOLBAR_ICON_BASE         =
        "/org/netbeans/core/windows/toolbars/xmlToolbars"; // NOI18N
    
    org.openide.loaders.XMLDataObject.Info xmlinfo = new org.openide.loaders.XMLDataObject.Info();
    xmlinfo.addProcessorClass (TOOLBAR_PROCESSOR_CLASS);
    xmlinfo.setIconBase (TOOLBAR_ICON_BASE);
    
    /* Register public toolbar dtd id with toolbar configuration xml info */
    org.openide.loaders.XMLDataObject.registerInfo (TOOLBAR_DTD_PUBLIC_ID, xmlinfo);
      
  }
  
  public synchronized static WindowManagerImpl getDefault() {
    if (defaultInstance == null) {
      defaultInstance = new WindowManagerImpl();
      // initialize window manager with default settings
      // we are not doing this in constructor to prevent from cycling
      if ("mdi".equalsIgnoreCase(System.getProperty("netbeans.windows"))) {
          defaultInstance.uiModeManager().setUIMode(UIModeManager.MDI_MODE, false);
      } else {
          defaultInstance.uiModeManager().setUIMode(UIModeManager.SDI_MODE, false);
      }
      if (NbTopManager.get ().isInteractive (NbTopManager.IL_WORKSPACES)) {
        dc = new DefaultCreator(defaultInstance);
        dc.start();
      }
    }
    return defaultInstance;
  }
  
  /** @return The only instance of window manager's state support
  * in the system */
  public static StateManager stateManager () {
    if (stateManager == null) {
      stateManager = new StateManager();
    }
    return stateManager;  
  }
  
  /** @return The only instance of support for deferred opening.
  */
  public static DeferredPerformer deferredPerformer () {
    if (deferredPerformer == null) {
      deferredPerformer = new DeferredPerformer();
    }
    return deferredPerformer;
  }
  
 
    /** Helper method - create workspaces from scratch by
    * adding standard workspaces and setting initial positions */
    public static void createFromScratch () {
        if ("full".equals (System.getProperty ("netbeans.full.hack"))) {
            return;
        }
        if (defaultInstance == null) {
            getDefault();
        } else if (dc != null) { // non-visual mode if dc is null
            // remove and close old workspaces and all their content
            Workspace[] ws = getDefault().getWorkspaces();
            for(int i = 0; i < ws.length; i++) {
                ws[i].remove();
            }
            // initialize default WS layout
            dc.setEnabled(true);
            dc.start();
        }
    }
  
  /** Provides access to the MainWindow of the IDE.
  * This should ONLY be used for:
  * <UL>
  *   <LI>Using the MainWindow as the parent for dialogs</LI>
  *   <LI>Using the MainWindow's positions for positioning of windows
  *       that need to be prepositioned </LI>
  * </UL>
  * @return the MainWindow of the IDE.
  */
  public java.awt.Frame getMainWindow () {
    return mainWindow ();
  }

  /** Called after a current LookAndFeel change to update the IDE's UI
  * Should call updateUI on all opened windows */
  public void updateUI () {
    // update main window first
    SwingUtilities.updateComponentTreeUI(mainWindow);
    mainWindow.pack();
    // update all other opened windows on workspaces
    Workspace[] wArray = getWorkspaces();
    for (int i = 0; i < wArray.length; i++) {
      for (Iterator iter = wArray[i].getModes().iterator(); iter.hasNext(); ) {
        ((ModeImpl)iter.next()).updateUI();
      }
    }
  }

  /** Returns newly created and properly 
  * initialized top component registry.
  * 
  * @return the registry 
  */
  protected TopComponent.Registry componentRegistry () {
    RegistryImpl result = new RegistryImpl();
    // force registry to listen to top components' events
    addTopComponentListener(result);
    return result;
  }
  
  /** Creates new workspace with given name.
  * @param name the name of the workspace
  * @return new workspace
  */
  public Workspace createWorkspace (String name, String displayName) {
    return new WorkspaceImpl(name, displayName);
  }
  
  /** Finds workspace given its name.
  * @param name the name of workspace to find
  * @return workspace or null if not found
  */
  public Workspace findWorkspace (String name) {
    if (workspaces == null)
      return null;
    for (int i = 0; i < workspaces.length; i++) {
      if (name.equals(workspaces[i].getName()))
        return workspaces[i];
    }
    return null;
  }
  
    /** List of all currenty available workspaces.
    */
    public Workspace[] getWorkspaces () {
        return workspaces;
    }

    /** Sets new workspaces.
    * @param workspaces array of new workspaces
    */
    public void setWorkspaces (Workspace[] workspaces) {
        if (Arrays.equals(this.workspaces, workspaces))
            return;
        Workspace[] old = this.workspaces;
        this.workspaces = workspaces;
        changeSupport.firePropertyChange(
            PROP_WORKSPACES, old, this.workspaces
        );
    }
  
    /** @return Current workspace.
    * Can be changed by calling Workspace.activate ()
    */
    public Workspace getCurrentWorkspace () {
        if (current == null) {
            // create default empty workspace if setCurrentWorkspace wasn't
            // called yet
            current = new WorkspaceImpl();
        }
        return current;
    }

  /** Sets current workspace.
  * @return true if succeded, false otherwise (when workspace is not
  * known to this window manager...)
  */
  public boolean setCurrentWorkspace (Workspace workspace) {
    if ((current != null) && (current.equals(workspace))) {
      return true;
    }
    // check if present
    boolean found = false;
    for (int i = 0; i < workspaces.length; i++) {
      if (workspaces[i].equals(workspace)) {
        found = true;
        break;
      }
    }
    if (!found) 
      return false;
    // perform the change
    StateManager stateMan = stateManager();
    if ((stateMan.getState() & StateManager.READY) != 0) {
      stateMan.setMainState(StateManager.SWITCHING);
    }
    // close all menus before workspace switch
    // (opened popup menus caused some serious problems)
    javax.swing.MenuSelectionManager.defaultManager().clearSelectedPath();
    Workspace old = current;
    current = workspace;
    if (old != null) {
      workspace2Nodes.put(old, getSelectedNodes());
      ((WorkspaceImpl)old).setVisible(false);
    }
    ((WorkspaceImpl)current).setVisible(true);
    // notify others
    changeSupport.firePropertyChange(
      PROP_CURRENT_WORKSPACE, old, current
    );
    setSelectedNodes(workspace2Nodes.get(current));
    if ((stateMan.getState() & StateManager.SWITCHING) != 0) {
      stateMan.setMainState(StateManager.READY);
    }
    return true;
  }


  /** switches to next workspace as current */
  public synchronized void nextWorkspace() {
    Workspace current = getCurrentWorkspace();
    int len = workspaces.length - 1;
    for (int i = len; i >= 0; i--) {
      if (workspaces[i] == current) {
        if (i == len) i = 0; // cycle it
        else ++i;
        // i will be current
        current = (Workspace) workspaces[i];
        current.activate();
        return;
      }
    }
  }

  /** switche to previous workspace */
  public synchronized void previousWorkspace() {
    Workspace current = getCurrentWorkspace();
    int len = workspaces.length - 1;
    for (int i = len; i >= 0; i--) {
      if (workspaces[i] == current) {
        if (i == 0) i = len; // cycle it
        else --i;
        // i will be current
        current = (Workspace) workspaces[i];
        current.activate();
        return;
      }
    }
  }

    /** Adds new explorer manager that will rule the selection of current
    * nodes until the runnable is running.
    *
    * @param run runnable to execute (till it is running the explorer manager is in progress)
    * @param em explorer manager 
    */
    public void attachExplorer (Runnable run, org.openide.explorer.ExplorerManager em) {
        ((RegistryImpl)getRegistry ()).attachExplorer (run, em);
    }

    /** @return selected nodes at 0 and activated at 1 */
    static Value getSelectedNodes() {
        Node[] current = getDefault().getRegistry().getCurrentNodes();
        Node[] activated = getDefault().getRegistry().getActivatedNodes();
        TopComponent tc = getDefault().getRegistry().getActivated();
        return new Value(current, activated, tc);
    }

    void setSelectedNodes(Value active) {
        Node[] activated = (active == null ? new Node[0] : active.getActivatedNodes());
        Node[] current = (active == null ? new Node[0] : active.getCurrentNodes());
        Node[][] nodes = new Node[][] { current, activated };
        RegistryImpl rimpl = (RegistryImpl) getRegistry();
        SelectedNodesChangedEvent ev = new SelectedNodesChangedEvent(nodes, null, activated);
        rimpl.selectedNodesChanged(ev);
        ev = new SelectedNodesChangedEvent(nodes, null, current);
        rimpl.selectedNodesChanged(ev);

        if (active != null) {
//      activateComponent(active.activatedTC);
      /*
      TopComponentChangedEvent tcev = new TopComponentChangedEvent(nodes,
                                                                 active.activatedTC,
                                                                 null,
                                                                 1
                                                                );
                                                                rimpl.topComponentActivated(tcev);
                                                                */
    }
  }
  
  //
  // You can add implementation to this class (+firePropertyChange), or implement it in subclass
  // Do as you want.
  //

  /** Attaches listener for changes in workspaces
  */
  public void addPropertyChangeListener (PropertyChangeListener l) {
    changeSupport.addPropertyChangeListener(l);
  }
  
  /** Removes listener.
  */
  public void removePropertyChangeListener (PropertyChangeListener l) {
    changeSupport.removePropertyChangeListener(l);
  }
  
  /** Adds top component listener for listening to the changes of
  * the set of top components in the system */
  public synchronized void addTopComponentListener (TopComponentListener tcl) {
    if (tcListeners == null)
      tcListeners = new HashSet(5);
    tcListeners.add(tcl);
  }
  
  /** Removes top component listener */
  public synchronized void removeTopComponentListener (TopComponentListener tcl) {
    if (tcListeners == null)
      return;
    tcListeners.remove(tcl);
  }
  
  /** @return True if workspace pool was created from scratch or
  * deserialized already, false if not initialized yet.
  */
  public boolean isCreated () {
    return isCreated;
  }
  
  /** Sets created flag. Accessible only for class in this
  * package */
  void setCreated (boolean isCreated) {
    this.isCreated = isCreated;
  }

  /** @return true if main window was already positioned */
  public boolean isMainPositioned () {
    return mainPositioned;
  }

  /** Helper method.
  * @return set of listeners, which is prepared for firing.
  * Can return null if no listeners are attached */
  Set getTcListenersForFiring () {
    if (tcListeners == null)
      return null;
    Set cloned = null;
    synchronized (this) {
      cloned = (Set)tcListeners.clone();
    }
    return cloned;
  }

    public static TopComponentContainer findContainer(String mode) {
        WindowManagerImpl wm = getDefault();
        Workspace aWsp;
        TopComponentContainer c = null;
        
        // find selected component of Source Editor window in selected workspace
        aWsp = wm.getCurrentWorkspace();
        if (null != (c = findContainer(mode, aWsp)))
            return c;
        
        // find selected component of Source Editor window in other workspaces
        Workspace allWsp[] = wm.getWorkspaces();
        for(int i = 0; i < allWsp.length; i++) {
            if (allWsp[i] == aWsp)
                continue;

            if (null != (c = findContainer(mode, allWsp[i])))
                break;
        }
        
        return c;
    }

    /** Utility method, finds container associated with given mode in workspace.
     * This method is supposed to be used for searching components docked in
     * specified mode, thus empty containers are treated as non-existing and are not returned.
     */
    public static TopComponentContainer findContainer(String mode, Workspace wsp) {
        ModeImpl m;
        TopComponentContainer c = null;
        
        m = (ModeImpl)wsp.findMode(mode);
        if (null != m) {
            c = m.getContainerInstance();
            if (null != c && 0 == c.getTopComponents().length)
                c = null;
        }
        
        return c;
    }
    
    /** Utility method, finds unused name of the mode on given workspace
    * based on given string.
    * @param base Base name of the mode. Can be null - in this cas
    * some default string constant will be used
    * @return string representing mode name which is not used yet
    * on current workspace */
    public static String findUnusedModeName (String base, Workspace workspace) {
        return findUnusedModeName(base, workspace, false);
    }
  
    public static String findUnusedModeName (String base, Workspace workspace, boolean forceNumber) {
        // be prepared when base is null
        if (base == null) {
            base = DEFAULT_NAME;
        }
        if (!forceNumber && (workspace.findMode(base) == null)) {
            return base;
        }
        // add numbers to the name
        String result = null;
        int modeNumber = 1;
        while (workspace.findMode(result = base + " " + modeNumber) != null) { // NOI18N
            modeNumber++;
        }
        return result;
    }

  /** Activate given top component and register component into history of
   * activated top components.
  */
  public void activateComponent (TopComponent tc) {
    if (tc != null) {
      // notify workspace
      ((WorkspaceImpl)getCurrentWorkspace()).componentActivated(tc);
    }
    // fire info that activated component changed
    Set listeners = getTcListenersForFiring();
    if (listeners != null) {
      for (Iterator iter = listeners.iterator(); iter.hasNext(); ) {
        ((TopComponentListener)iter.next()).topComponentActivated(
          new TopComponentChangedEvent(this, tc, getCurrentWorkspace(), 
            TopComponentChangedEvent.ACTIVATED));
      }
    }
    super.activateComponent(tc);
  }

    /** Annotate persistence exception. Exception is added to the exception
     * list, which is displayed at once when whole persistence process 
     * (either serialization or deserialization) is about to finish.
     */
    void annotatePersistenceError (Exception exc, String tcName) {
        if (mainEx == null) {
            // mainEx is a marker which serves as 'father' of all our
            // persistence exceptions
            mainEx = new Exception();
        }
        TopManager.getDefault().getErrorManager().annotate(mainEx, exc);
        if (failedCompsMap == null) {
            failedCompsMap = new HashMap();
        }
        failedCompsMap.put(tcName, tcName);
    }
    
    /** Checks for some persistence errors and notifies the user if some
     * persistence errors occured. Shouild be called after serialization
     * and deserialization of window manager.
     */
    void checkPersistenceErrors (boolean reading) {
        if (mainEx == null) {
            // everything ok, nothing to notify...
            return;
        }
        // build buffer containing names of all failed top components
        StringBuffer buffer = new StringBuffer();
        Iterator iter = failedCompsMap.values().iterator();
        buffer.append((String)iter.next());
        for (; iter.hasNext(); ) {
            buffer.append(", " + (String)iter.next());
        }
        // create message
        String message = null;
        if (reading) {
            message = MessageFormat.format(
                NbBundle.getBundle(WindowManagerImpl.class).getString("FMT_TCReadError"),
                new Object[] { buffer.toString() }
            );
        } else {
            message = MessageFormat.format(
                NbBundle.getBundle(WindowManagerImpl.class).getString("FMT_TCWriteError"),
                new Object[] { buffer.toString() }
            );
        }
        TopManager.getDefault().getErrorManager().annotate(mainEx, message);
        TopManager.getDefault().getErrorManager().notify(mainEx);
        // clear for futher use
        mainEx = null;
        failedCompsMap = null;
    }
    
 
  /** Creates a component manager for given top component.
  * @param c the component
  * @return the manager that handles opening, closing and
  * selecting a component
  */
  protected synchronized org.openide.windows.WindowManager.Component createTopComponentManager (TopComponent c) {
    // create new manager or returng manager being validated
    org.openide.windows.WindowManager.Component result = 
      (validatedManager == null) ? new TopComponentManager(c) : validatedManager;
    // clear after each usage
    validatedManager = null;
    return result;
  }

  /** Helper; used from TopComponentManager to identify itself
  * during deserialization to achieve asociation with its top component */
  private void setValidatedManager (TopComponentManager tcm) {
    validatedManager = tcm;
  }
  
  private TopComponentManager getValidatedManager () {
    return validatedManager;
  }
  
  /** Finds top component manager for given top component.
  * It's here just to rovide access for classes in this package.
  */
  public static TopComponentManager findManager (TopComponent tc) {
    return (TopComponentManager) /*WindowManager. not compilable by JIKES */findComponentManager(tc);
  }

  /** Convenience static method for simpler obtaining mainWindow
  * reference
  */
  public static final MainWindow mainWindow () {
    if (mainWindow == null) {
      synchronized (WindowManagerImpl.class) {
        if (mainWindow == null) {
          mainWindow = MainWindow.getDefault();
          mainWindow.addWindowListener(new IconifyManager());
        }
      }
    }
    return mainWindow;
  }
  
    /** Sets UI mode for whole IDE. Possible IDE UI modes are defined as
    * constants in 
    */
    public void setUIMode (int uiMode) {
        uiModeManager().setUIMode(uiMode);
    }

    /** @return Returns current ui mode IDE is in */
    public int getUIMode () {
        return uiModeManager().getUIMode();
    }

    /** Utility method. Finds proper frame type for given mode, which suits
     * current ui mode */
    public String findFrameType (ModeImpl mode) {
        return uiModeManager().findFrameType(mode);
    }

    /** Accessor for ui mode mnanager */
    public final UIModeManager uiModeManager () {
        if (uiModeManager == null) {
            uiModeManager = new UIModeManager();
        }
        return uiModeManager;
    }

    /** Sets a UI type for containers with JTabbedPane
     *  tabbedContainerUI - constant of TabbedPaneUI (TabbedContainerUIManager.MODE_...)
     */
    public void setTabbedContainerUI (int tabbedContainerUI) {
        tabbedContainerUIManager().setTabbedContainerUI(tabbedContainerUI);
    }

    /** Reads a UI type for containers with JTabbedPane
     *  return - constant of TabbedPaneUI (TabbedContainerUIManager.MODE_...)
     */
    public int getTabbedContainerUI () {
        return tabbedContainerUIManager().getTabbedContainerUI();
    }

    /** Reads a TabbedContainedUIManager
     *  return - TabbedContainedUIManager
     */
    public final TabbedContainerUIManager tabbedContainerUIManager () {
        if (tabbedContainerUIManager == null) {
            tabbedContainerUIManager = new TabbedContainerUIManager();
        }
        return tabbedContainerUIManager;
    }

    /** @return Screen size of last deserialized window system or null if
     * information is not available (no window system info was deserialized
     * in this IDE session */
    public Dimension getOldScreenSize () {
        return oldScreenSize;
    }
    
    /** @return Bounds of last deserialized main window or null if
     * information is not available (no window system info was deserialized
     * in this IDE session */
    public Rectangle getOldMainWindowBounds () {
        return oldMainWindowBounds;
    }
  
  /** Delegates serialization manager instance to be serialized 
  * instead of window manager impl */
  private Object writeReplace () 
                 throws ObjectStreamException {
    return new SerializationReplacer();
  }
  
    /** Instance of this class is serialized instead of WindowManagerImpl.
    * It saves all needed information and deserializesback  to the signleton
    * instance of WindowManagerImpl */
    private static final class SerializationReplacer implements Serializable {
        static final long serialVersionUID =-8212722893309295268L;
        /** Description of persistent fields */
        private static final String WORKSPACES = "workspaces";
        private static final String CURRENT_WORKSPACE = "currentWorkspace";
        private static final String MAIN_WINDOW_BOUNDS = "mainWindowBounds";
        private static final String UI_MODE = "uiMode";
        private static final String SCREEN_SIZE = "screenSize";
        private static final String TABBED_CONTAINER_UI = "tabbedContainerUI";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(WORKSPACES, java.util.List.class),
            new ObjectStreamField(CURRENT_WORKSPACE, WorkspaceImpl.class),
            new ObjectStreamField(MAIN_WINDOW_BOUNDS, Rectangle.class),
            new ObjectStreamField(UI_MODE, Integer.class),
            new ObjectStreamField(SCREEN_SIZE, Dimension.class),
            new ObjectStreamField(TABBED_CONTAINER_UI, Integer.class),
        };
        
        // deserialized workspaces
        private Workspace[] workspaces;
        // deserialized current workspaces
        private Workspace current;
        // deserialized size of screen
        private Dimension oldScreenSize;
        // deserialized ui mode of the IDE
        private int uiMode;
        // deserialized ContainerUI mode of the IDE
        private int tabbedContainerUI;
        // deserialized ui mode of the IDE
        private Rectangle mainWindowBounds;
        // flag for recognizing first startup 
        private boolean isStartup = true;
    
        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            // obtain WindowManagerImpl whose serialization we manage
            WindowManagerImpl wm = WindowManagerImpl.getDefault();
            StateManager stateMan = wm.stateManager();
            stateMan.setMainState(StateManager.DESERIALIZING);
            try {
                ObjectInputStream.GetField gf = ois.readFields();
                //Bugfix #10993, workaround for JDK 1.2.2
                //if (gf.getObjectStreamClass().getField(MAIN_WINDOW_BOUNDS) != null) {
                if (WindowUtils.hasObjectStreamField(gf, MAIN_WINDOW_BOUNDS)) {
                    // current data format
                    java.util.List workspaceList = (java.util.List)gf.get(WORKSPACES, null);
                    workspaces = (Workspace[])workspaceList.toArray(new Workspace[workspaceList.size()]);
                    current = (Workspace)gf.get(CURRENT_WORKSPACE, null);
                    oldScreenSize = (Dimension)gf.get(SCREEN_SIZE, null);
                    uiMode = ((Integer)gf.get(UI_MODE, null)).intValue();
                    mainWindowBounds = (Rectangle)gf.get(MAIN_WINDOW_BOUNDS, null);
                } else {
                    // old data format
                    // read workspaces - first phase
                    int count = ois.readInt();
                    workspaces = new Workspace[count];
                    for (int i = 0; i < count; i++) {
                        workspaces[i] = (Workspace)ois.readObject();
                    }
                    // read current workspace
                    current = (Workspace)ois.readObject();
                    // bounds of main window
                    mainWindowBounds = (Rectangle)ois.readObject();
                    uiMode = UIModeManager.SDI_MODE;
                }
                if (gf.getObjectStreamClass().getField(TABBED_CONTAINER_UI) != null) {
                    tabbedContainerUI = ((Integer)gf.get(TABBED_CONTAINER_UI, null)).intValue();
                } else {
                    tabbedContainerUI = TabbedContainerUIManager.MODE_DEFAULT;
                }
                // deserialization seems to be succesful, so disable default
                // window system content creation and positioning
                wm.dc.setEnabled(false);
                // testing
                //throw new IOException();
            } finally {
                isStartup = false;
                stateMan.setMainState(StateManager.READY);
            }
        }

        /** Resolves deserialized SerializationReplacer to the singleton
        * instance of WindowManagerImpl */
        private synchronized Object readResolve () 
                                    throws ObjectStreamException {
          // obtain WindowManagerImpl whose serialization we manage
          WindowManagerImpl wm = WindowManagerImpl.getDefault();
          StateManager stateMan = wm.stateManager();
          stateMan.setMainState(StateManager.DESERIALIZING);
          try {
            initGlobalProperties();
            // validating workspaces and modes
            for (int i = 0; i < workspaces.length; i++) {
              ((WorkspaceImpl)workspaces[i]).validateSelf();
            }
            // set new workspaces
            wm.setWorkspaces(workspaces);
            // validate top components
            if(wm.getUIMode() != uiMode) {
                wm.uiModeManager.changeModeFrameTypes();
            }
            for (int i = 0; i < workspaces.length; i++) {
              ((WorkspaceImpl)workspaces[i]).validateData();
            }
            wm.setCurrentWorkspace(current);
          } finally {
            stateMan.setMainState(StateManager.READY);
            wm.checkPersistenceErrors(true);
          }
          // reactivate 
          return wm;               
        }

        /** Init global properties of window system like main window bounds,
         * UI mode */
        private void initGlobalProperties () {
            WindowManagerImpl wm = WindowManagerImpl.getDefault();
            wm.mainPositioned = true;
            wm.oldScreenSize = oldScreenSize;
            wm.oldMainWindowBounds = new Rectangle(mainWindowBounds);
            wm.setTabbedContainerUI(tabbedContainerUI);
            if(wm.uiModeManager().getUIMode() == uiMode) {
                if (oldScreenSize != null) {
                    WindowUtils.convertRectangle(
                        mainWindowBounds, oldScreenSize,
                        Toolkit.getDefaultToolkit().getScreenSize()
                    );
                }
                wm.mainWindow().setBounds(mainWindowBounds);
            }
        }

        /** Serialization of all workspaces */
        private void writeObject (ObjectOutputStream oos)
                 throws IOException {
            // obtain WindowManagerImpl whose serialization we manage
            WindowManagerImpl wm = WindowManagerImpl.getDefault();
            StateManager stateMan = wm.stateManager();
            stateMan.setMainState(StateManager.SERIALIZING);
            try {
                ObjectOutputStream.PutField pf = oos.putFields();
                pf.put(WORKSPACES, Arrays.asList(wm.workspaces));
                pf.put(CURRENT_WORKSPACE, wm.current);
                pf.put(MAIN_WINDOW_BOUNDS, wm.mainWindow().getBounds());
                pf.put(SCREEN_SIZE, Toolkit.getDefaultToolkit().getScreenSize());
                pf.put(UI_MODE, new Integer(wm.getUIMode()));
                pf.put(TABBED_CONTAINER_UI, new Integer(wm.getTabbedContainerUI()));
                oos.writeFields();
            } finally {
                stateMan.setMainState(StateManager.READY);
                wm.checkPersistenceErrors(false);
            }
        }
    
  } // end of inner class SerializationReplacer
  
    /** Manager for iconifying/deiconifying main window */
    static final class IconifyManager extends WindowAdapter implements Runnable {
        /** The workspace to be shown on deiconification */
        private WorkspaceImpl workspace;
    
        public void windowIconified (WindowEvent evt) {
            if (getDefault().getUIMode() != UIModeManager.MDI_MODE) {
                // in SDI, post request to iconify all other windows
                // the iconify code is called later, when all Mode windows would
                // have to be iconified if this actions was trigged by some
                // XWindow manager (when switching workspace)
                //
                // If the inconification was requested by minimalization of just
                // main window the Frames of all other windows will remain in state
                // Frame.NORMAL, if the inconification is invoked by switching 
                // virtual desktops, then all other windows will already (after 200ms)
                // be in Frame.ICONIFIED state.
                RequestProcessor.postRequest (this, 200);
            }
        }

        public void run () {
            if (SwingUtilities.isEventDispatchThread()) {
                // test iconify
                iconify ();
            } else {
                // call this code later in event thread
                SwingUtilities.invokeLater (this);
            }
        }

        public void windowDeiconified (WindowEvent evt) {
            if (workspace != null) {
                workspace.setVisible(true);
                // make gc' able
                workspace = null;
            }
        }

        /** Iconify current workspace content */
        private void iconify () {
            workspace = (WorkspaceImpl)TopManager.getDefault().
                      getWindowManager().getCurrentWorkspace();
            Iterator it = workspace.getModes ().iterator ();
            while (it.hasNext ()) {
                FrameType curFrame = ((ModeImpl)it.next()).getFrameInstance();
                if ((curFrame != null) && (curFrame.getState() == FrameType.NORMAL)) {
                    workspace.setVisible (false);
                    return;    
                }
            }
        }
        
    } // end of IconifyManager  
  
  
    /** The manager that handles operations on a top component.
    * It is always attached to a TopComponent via one-to-one
    * relationship.
    *
    * @author Dafe Simonek
    */
    public static final class TopComponentManager extends ComponentAdapter 
                                                implements WindowManager.Component, DeferredPerformer.DeferredCommand {
    /** The constants for properties of managed top component */                                           
    public static final String PROP_ACTIVATED_NODES = "activatedNodes"; // NOI18N
    public static final String PROP_NAME = "name"; // NOI18N
    public static final String PROP_ICON = "icon"; // NOI18N

    /** Set of workspaces where top component is opened. */
    transient HashSet whereOpened;
    /** Agregation of the top component which we're trying to manage */
    TopComponent component;
    /** top component we manage in hlaf-way deserialized form,
    * used only during deserialization */
    private NbMarshalledObject marshalledComponent;
    /** helper variable, holds top component's name, used when
    * reporting failure of deserialization of the top copmponent */
    String componentName;

    /** Icon of the component we manage */
    transient Image icon;
    /** Activated nodes of the component we manage */
    transient Node[] nodes;
    /** Nodes which will be set as activated when the component becomes non-null */
    transient Node[] tempNodes;
    /** asociation with the window manager */
    transient WindowManagerImpl wm;
    /** Support for property changes */
    transient PropertyChangeSupport changeSupport;
    /** asociation with support for deferred opening */
    transient DeferredPerformer deferredPerformer;
    /** helper flag, holds current state of validation during
    * deserialization */
    private transient int innerState = READY;
    /** constant for all-right state */
    private static final int READY = 1;
    /** constant for state when validation was not performed yet */
    private static final int INVALID = 2;
    /** constant for state when deserialization of tc completely failed */
    private static final int FAILED = 4;

    /** manager of versioned serialization */
    private static VersionSerializator serializationManager;

    static final long serialVersionUID =5669852754182098300L;
    /** Constructs new top component manager. Allow only classes in package
    * to construct us
    * @param component Managed top component.
    */
    TopComponentManager (TopComponent component) {
      this.component = component;
      whereOpened = new HashSet(5);
      initialize();
    }

    /** Initialization of this manager, called also when deserializaing */
    private void initialize () {
      this.wm = 
        (WindowManagerImpl)TopManager.getDefault().getWindowManager();
      changeSupport = new PropertyChangeSupport(this);
      deferredPerformer = deferredPerformer();
    }

    /** Opens a component on current workspace. If main window is
    * not visible, open action is delayed until it's open.
    */
    public void open () {
      open(null);
    }

    /** Opens a component on given workspace. If window system is
    * in inconsistent state (main window is not visible, serializing etc..)
    8 open action is delayed.
    * @workspace the workspace where to open managed top component
    */
    public void open (Workspace workspace) {
        if (deferredPerformer.canImmediatelly()) {
            // immediate open (replan to event thread if needed)
            final Workspace realWorkspace = 
                (workspace == null) ? wm.getCurrentWorkspace() : workspace;
            if (java.awt.EventQueue.isDispatchThread()) {
                doOpen(realWorkspace);
            } else {
                javax.swing.SwingUtilities.invokeLater(new Runnable () {
                    public void run () {
                        doOpen(realWorkspace);
                    }
                });
            }
      } else {
        // deferred open
        deferredPerformer.putRequest(this, workspace);
      }
    }

    /** Implementation of DeferredPerformer.DeferredCommand interface.
    * Actually opens managed top component. */
    public void performCommand (DeferredPerformer.DeferredContext context) {
        component.open((Workspace)(context.getData()));
    }

    /** Opens a component on given workspace.
    * If given workspace differs from current, component will
    * be visible only after switching to given workspace.
    * @workspace the workspace where to open managed top component
    */
    private void doOpen (Workspace workspace) {
        if (!whereOpened.add(workspace)) {
            return;
        };
        Mode mode = workspace.findMode(component);
        if (mode == null) {
            // create new mode for given tc
            // important bugfix - create properly named mode even 
            // for top components with null name
            String modeName = wm.findUnusedModeName(component.getName(), workspace);
            mode = ((WorkspaceImpl)workspace).createMode(
                modeName, modeName, null, true
            );
        }
        mode.dockInto(component);
        // inform top component's subclasses
        if (whereOpened.size() == 1) {
            wm.componentOpenNotify(component);
        }
        // then let others know that top component was opened...
        Set listeners = wm.getTcListenersForFiring();
        if (listeners != null) {
            for (Iterator iter = listeners.iterator(); iter.hasNext(); ) {
                ((TopComponentListener)iter.next()).topComponentOpened(
                    new TopComponentChangedEvent(this, component, workspace,
                                                 TopComponentChangedEvent.OPENED)
                );
            }
        }
    }

    /** @return true if managed top component is currently opened
    * on some workspace, false otherwise */
    public boolean isOpened () {
        return whereOpened.size() > 0;
    }

    /** @return true if managed top component is currently opened
    * on given workspace, false otherwise */
    public boolean isOpened (Workspace workspace) {
        return whereOpened.contains(workspace);
    }

    /** Closes the component on given workspace. 
    * @param workspace the workspace where managed top component
    * should be closed.
    */
    public void close (Workspace workspace) {
      switch (component.getCloseOperation()) {
        case TopComponent.CLOSE_LAST:
          if (isOpened(workspace))
            doClose(workspace);
          break;
        case TopComponent.CLOSE_EACH:
          // special mode for editor etc...
          // close on all workspaces
          Workspace[] workspaces = wm.getWorkspaces();
          for (int i = 0; i < workspaces.length; i++) {
            if (isOpened(workspaces[i]))
              doClose(workspaces[i]);
          }
          break;
      }
    }

    /** The component requests focus. Works only on opened component.
    * When the request comes too early, implemetation waits
    * till component is shown.
    */
    public void requestFocus () {
        //09 Feb 2001 by Marek Slama
        //Bug fix #9445
        //Call of requestFocus replanned to AWT to make sure that
        //it is performed AFTER open.
        DoRequestFocus doRequestFocus = new DoRequestFocus();
        if (java.awt.EventQueue.isDispatchThread()) {
            doRequestFocus.run();
        } else {
            javax.swing.SwingUtilities.invokeLater(doRequestFocus);
        }
    }
        
    /** @return the mode which belongs to managed top component on
    * CURRENT workspace.
    */
    Mode getMode () {
        return wm.getCurrentWorkspace().findMode(component);
    }

    /** Actually perform the work of closing ther managed component on
    * given workspace, without checking. */
    void doClose (Workspace workspace) {
        ModeImpl mode = (ModeImpl)workspace.findMode(component);
        whereOpened.remove(workspace);
        if (mode != null) {
            mode.close(component);
        }
        // inform subclasses
        if (!isOpened()) {
            wm.componentCloseNotify(component);
        }
        // let others know that top component was closed...
        Set listeners = wm.getTcListenersForFiring();
        if (listeners != null) {
            for (Iterator iter = listeners.iterator(); iter.hasNext(); ) {
                ((TopComponentListener)iter.next()).topComponentClosed(
                    new TopComponentChangedEvent(this, component, workspace,
                                                 TopComponentChangedEvent.CLOSED)
                );
            }
        }
    }

    /** Getter for set of activated nodes
    * @return activated nodes for the component
    */
    public Node[] getActivatedNodes () {
        return nodes;
    }

    /** Setter for set of activated nodes for the component
    * @param nodes activated nodes for this component
    */
    public void setActivatedNodes (Node[] nodes) {
      if (component == null) {
        // store the nodes to a temporary variable and set them properly
        // when a non-null value of the component is set
        tempNodes = nodes;
        return;
      }
      if (Arrays.equals(this.nodes, nodes))
        return;
      Node[] old = this.nodes;
      this.nodes = nodes;
      // notify all that are interested...
      changeSupport.firePropertyChange(PROP_ACTIVATED_NODES, old, nodes);
      Set listeners = wm.getTcListenersForFiring();
      if (listeners != null) {
        for (Iterator iter = listeners.iterator(); iter.hasNext(); ) {
          ((TopComponentListener)iter.next()).selectedNodesChanged(
            new SelectedNodesChangedEvent(this, component, nodes));
        }
      }
    }

    /** Notify about name change.
    */
    public void nameChanged () {
      // component can be null for a while during deserialization
      if (component != null) {
        changeSupport.firePropertyChange(PROP_NAME, null, component.getName());
      }
    }

    /** Sets the icon of the top component which will be used for
    * component representaion on the screen.
    * @param icon New components' icon.
    */
    public void setIcon (final Image icon) {
      if (((this.icon != null) && this.icon.equals(icon)) ||
          ((this.icon == null) && (icon == null)))
        return;
      Image old = this.icon;
      this.icon = icon;
      changeSupport.firePropertyChange(PROP_ICON, old, this.icon);
    }

    public Image getIcon () {
      return icon;
    }

    /** Getter for managed component 
    * @return managed component
    */
    public TopComponent getComponent () {
      return component;
    }

    /** @return the set of workspaces where managed component is open */
    public Set whereOpened () {
      return whereOpened;
    }

    /** Adds listener for listening to top component property changes.
    */
    public void addPropertyChangeListener (PropertyChangeListener pchl) {
      changeSupport.addPropertyChangeListener(pchl);
    }

    /** Removes property change listener.
    */
    public void removePropertyChangeListener (PropertyChangeListener pchl) {
      changeSupport.removePropertyChangeListener(pchl);
    }

    /** Called when managed top component becomes
    * visible - just for the requestFocus, when
    * it comes too early.
    */
    public void componentShown (ComponentEvent ev) {
      // notify me no more
      component.removeComponentListener(this);
      // now request a focus
      SwingUtilities.invokeLater(new Runnable() {
        public void run () {
          component.requestFocus();
        }
      });
    }

    /** Serialization */
    private Object writeReplace ()
                 throws ObjectStreamException {
      NbMarshalledObject marshalledTc = null;           
      try {
        marshalledTc = new NbMarshalledObject(component);
        // testing
        // throw new Exception();
      } catch (Exception exc) {
        wm.annotatePersistenceError(exc, component.getName());
        // write null as a marker that there was serialization problem
        return null;
      }
      return new SerializationReplacer(componentName, marshalledTc);
    }             

    /** Called when first phase of WS deserialization is done.
    * Deserialization of managed TC is finished here.
    */
    public boolean validateData () {
      // deserialize only once, then only return succes or failure
      // status
      if (innerState != INVALID) {
        return innerState == READY;
      }
      // component written badly, consider it a failure
      if (marshalledComponent == null) {
        innerState = FAILED;
        return false;
      }
      // Actually deserializes top component.
      // Uses ugly hack - redundant TopComponent<init>.getManager() call 
      // to enforce wm.createTopComponentManager() to be called,
      // in which we can re-asociate deserialized top component
      // with this manager. 
      // This is UGLY implementation, based on the fact that there's
      // no way to set new manager for top component and we want to
      // keep the Open API unchanged if possible
      try {
        wm.setValidatedManager(this);
        component = (TopComponent)marshalledComponent.get();
        // component can resolve itself to null,
        // be prepared to such situation
        /*if (component != null) {
          component.getActivatedNodes();
        }*/
        innerState = (component == null) ? FAILED : READY;
        if ((component != null) && (tempNodes != null)) {
          setActivatedNodes(tempNodes);
          tempNodes = null;
        }
      } catch (Exception exc) {
        // do catch all exceptions, because exceptions
        // during deserialization of one top component
        // shoudn't break whole window system
        innerState = FAILED;
        wm.annotatePersistenceError(exc, componentName);
        return false;
      } finally {
        // ensure that manager is not set anymore
        wm.setValidatedManager(null);
      }
      return innerState == READY;
    }

    /** Utility method, creates, fills and returns instance of top component
     * manager */
    private static TopComponentManager createTCM (String componentName, 
                                                  NbMarshalledObject marshalledComponent) {
        TopComponentManager result = new TopComponentManager(null);
        result.componentName = componentName;
        result.marshalledComponent = marshalledComponent;
        result.innerState = INVALID;
        return result;
    }                                                      

    /** Used to do requestFocus in AWT thread. Necessary when open is done
     * in AWT thread too to make sure correct order first open, then requestFocus.
     * Called from TopComponentManager.requestFocus
     */
    private class DoRequestFocus extends Object implements Runnable {        
        public void run () {
            if (!isOpened()) {
                return;
            }
            if (component.isDisplayable()) {
                Mutex.EVENT.readAccess(new Runnable() {
                    public void run () {
                        // fulfill only if the request is still valid,
                        // be aware not to steal focus from modal dialog
                        ModeImpl modeImpl = (ModeImpl)getMode();
                        if ((modeImpl != null) && (component != null) &&
                            (NbPresenter.currentModalDialog == null)) {
                            modeImpl.requestFocus(component);
                        }
                    }
                });
            } else { 
                // component not yet shown, we must wait
                // until the component is shown
                component.addComponentListener(TopComponentManager.this);
            }
        }
    }
    
    /** Class that acts as serialization manager for component manager implementation */
    private static final class SerializationReplacer implements Serializable {
        /** Unique ID */
        static final long serialVersionUID =-9009124438910567345L;
        /** Description of serializable fields */
        private static final String NAME = "name";
        private static final String MARSHALLED_COMPONENT = "marshalledComponent";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(NAME, String.class),
            new ObjectStreamField(MARSHALLED_COMPONENT, NbMarshalledObject.class),
        };

        private String componentName;
        private NbMarshalledObject marshalledComponent;

        SerializationReplacer (String componentName, NbMarshalledObject marshalledComponent){
            this.componentName = componentName;
            this.marshalledComponent = marshalledComponent;
        }

        /** Serialization of one top component, driven by its manager */
        private void writeObject (ObjectOutputStream oos)
                     throws IOException {
            // write fields
            ObjectOutputStream.PutField pf = oos.putFields();
            pf.put(NAME, componentName);
            pf.put(MARSHALLED_COMPONENT, marshalledComponent);
            oos.writeFields();
        }

        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField gf = ois.readFields();
            componentName = (String)gf.get(NAME, null);
            marshalledComponent = (NbMarshalledObject)gf.get(MARSHALLED_COMPONENT, null);
        }

        /** Resolves deserialized SerializationReplacer to the singleton
        * instance of WindowManagerImpl */
        private synchronized Object readResolve () 
                                    throws ObjectStreamException {
            return createTCM(componentName, marshalledComponent);
        }

    } // end of SerializationReplacer

    
/******** Old serialization, must be here for compatibility with 3.1 and older versions */    
    
    /** Accessor to the versioned serialization manager */
    private VersionSerializator serializationManager () {
      if (serializationManager == null) {
        serializationManager = createSerializationManager();
      }
      return serializationManager;
    }
    
    /** Creates new serialization manager filled with our versions */
    private static VersionSerializator createSerializationManager () {
      VersionSerializator result = new VersionSerializator();
      result.putVersion(new Version1());
      return result;
    }

    /** Basic version of persistence for mode implementation.
    * Method assignData(modeImpl) must be called prior to serialization */
    private static final class Version1 
                         implements DefaultReplacer.ResVersionable {

      /* identification string */
      public static final String NAME = "Version_1.0"; // NOI18N

      /** variables of persistent state of the tc manager */
      String componentName;
      NbMarshalledObject marshalledComponent;
      /** asociation with outerclass, used when writing */
      TopComponentManager tcm;
      /** set of workspaces where managed tc is opened */
      transient HashSet whereOpened;

      /** Identification of the version */
      public String getName () {
        return "Version_1.0"; // NOI18N
      }

      /** read the data of the version from given input */
      public void readData (ObjectInput in) 
                  throws IOException, ClassNotFoundException {
        // read the fields
        componentName = (String)in.readObject();
        marshalledComponent = (NbMarshalledObject)in.readObject();
      }              

      /** write the data of the version to given output */
      public void writeData (ObjectOutput out)
                  throws IOException {
        // should be never called              
        throw new InternalError();          
      }

      /** Resolves this object to top component manager instance again */
      public Object resolveData ()
                    throws ObjectStreamException {
        return createTCM(componentName, marshalledComponent);
      }

    } // end of Version1 inner class

    /** Implementation of persistent access to our version serializator */
    private static final class VSAccess implements DefaultReplacer.Access {

        /** serialVersionUID */
        private static final long serialVersionUID = -6484558550904999459L;

        /** version serializator, used only during writing */
        transient VersionSerializator vs;

        public VSAccess (VersionSerializator vs) {
            this.vs = vs;
        }

        public VersionSerializator getVersionSerializator () {
            return (vs == null) ? createSerializationManager() : vs;
        }

    } // end of VSAccess inner class
   
  } // end of TopComponentManager inner class

}
