/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.Component;
import java.awt.Container;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.SwingUtilities;

import org.openide.windows.Mode;
import org.openide.windows.TopComponent;

import org.netbeans.core.output.OutputTab;
import org.netbeans.core.windows.util.WindowUtils;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.windows.frames.DesktopPane;

/** Takes care of frame positiong issues of asociated workspace. Corrects bounds,
 * transforms frames bounds between MDI and SDI environment etc.
 * This class plays a "part" role in the part-whole relationship with WorkspaceImpl.
 *
 * @author  Dafe Simonek
 */
class WorkspaceBoundsManager extends Object {
    
    /** Default height and width of modes */
    public static final int DEFAULT_MODE_WIDTH = 500;
    public static final int DEFAULT_MODE_HEIGHT = 350;
    
    /** helper - last bounds of main window */
    private static Rectangle lastMainBounds;
    
    /** agregation with WorkspaceImpl */
    private WorkspaceImpl workspace;
    /** window manager asociation */
    private WindowManagerImpl wm;
    /** default size of one cascading step */
    private Dimension cascadeStep;
    /** holds number of cascade steps */
    private int cascadeStepsCount;
    /** Current cascade point - to which cascaded windows are placed */
    private Point cascadePoint;
    /** Origin of the cascading on this workspace */
    private Point cascadingOrigin;
    
    /** Creates new WorkspacePositioningManager, aggtegated with workspace. */
    WorkspaceBoundsManager (WorkspaceImpl workspace) {
        this.workspace = workspace;
        wm = WindowManagerImpl.getDefault();
    }
    
    /** Initializes itself with values from other bounds manager */ 
    public void copyFrom (WorkspaceBoundsManager other) {
        cascadeStep = new Dimension(other.getCascadeStep());
        cascadeStepsCount = other.getCascadeStepsCount();
        Point helper = other.getCascadePoint();
        cascadePoint = (helper == null) ? null : new Point(helper);
    }

    /** Getter for property cascadingOrigin.
     * @return Value of property cascadingOrigin.
     */
    public Point getCascadingOrigin () {
        if (cascadingOrigin == null) {
            cascadingOrigin = new Point(0, 0);
        }
        return cascadingOrigin;
    }

    /** Setter for property cascadingOrigin. Also properly recomputes
     * current cascading point.
     * @param cascadingOrigin New value of property cascadingOrigin.
     */
    public void setCascadingOrigin (Point cascadingOrigin) {
        this.cascadingOrigin = cascadingOrigin;
        recomputeCascadePoint();
    }
    
    /** Getter for property cascadeStep.
     * @return Value of property cascadeStep.
     */
    public Dimension getCascadeStep () {
        if (cascadeStep == null) {
            cascadeStep = new Dimension(20, 20);
        }
        return cascadeStep;
    }    

    /** Setter for property cascadeStep.
     * @param cascadeStep New value of property cascadeStep.
     */
    public void setCascadeStep (Dimension cascadeStep) {
        this.cascadeStep = cascadeStep;
    }
    
    /** Getter for property cascadeStepsCount.
     * @return Value of property cascadeStepsCount.
     */
    public int getCascadeStepsCount () {
        return cascadeStepsCount;
    }    
    
    /** Setter for property cascadeStepsCount.
     * @param cascadeStepsCount New value of property cascadeStepsCount.
     */
    public void setCascadeStepsCount (int cascadeStepsCount) {
        this.cascadeStepsCount = cascadeStepsCount;
    }    
    
    /** @return Current cascade point of this workspace. Cascade point
    * is a point where mode screen representation (window, internal
    * frame or whatever) will be placed when added to this workspace.
    */
    public Point getCascadePoint () {
        Window mainWindow = wm.getMainWindow();
        Rectangle mainBounds = mainWindow.getBounds();
        // check variables
        if (lastMainBounds == null)
            lastMainBounds = mainBounds;
        if (cascadeStepsCount <= 0)
            cascadePoint = null;
        // recompute cascade point if needed
        if ((cascadePoint == null) || (!lastMainBounds.equals(mainBounds)) ||
                (((mainBounds.width == 0) && (mainBounds.height == 0)))) {
            recomputeCascadePoint();
            lastMainBounds = mainBounds;
        }
        return cascadePoint;
    }

    /** Non-public, only for WorkspaceImpl tobe able to set cascadePoint
     * during deserialization
     */
    void setCascadePoint (Point cascadePoint) {
        this.cascadePoint = cascadePoint;
    }
    
    /** Corrects the bounds of given list of modes, if the screen size
     * is different than it was at the time of serialization 
     */
    public void correctBounds () {
        Dimension currentScreen = Toolkit.getDefaultToolkit().getScreenSize ();
        Dimension oldScreen = wm.getOldScreenSize();
        // do nothing if the same
        if (oldScreen == null || currentScreen.equals(oldScreen)) {
            return;
        }
        List modes = workspace.getModeArray();
        // correct the bounds of modes
        boolean isMdi = wm.getUIMode() == UIModeManager.MDI_MODE;
        ModeImpl curMode = null;
        for (Iterator iter = modes.iterator(); iter.hasNext(); ) {
            curMode = (ModeImpl)iter.next();
            Rectangle modeRect = curMode.getBounds();
            if (modeRect != null) {
                Rectangle newBounds = new Rectangle(modeRect);
                WindowUtils.convertRectangle(newBounds, oldScreen, currentScreen);
                if (!isMdi) {
                    removeMainWindowOverlapping(newBounds);
                }
                curMode.setBounds(newBounds);
            }
        }
    }
    
    /** Returns bounds for working space of the IDE. called from
    * getBounds() method. */
    public Rectangle getWorkingSpaceBounds () {
        MainWindow mainWindow = WindowManagerImpl.mainWindow();
        Rectangle mainBounds = mainWindow.getBounds();
        if (UIModeManager.MDI_MODE == WindowManagerImpl.getDefault().getUIMode()) {
            // MDI case
            Dimension desktopSize = mainWindow.getDesktopSize();
            // use SDI computation if MDI desktop pane size is not initialized  
            if ((desktopSize.width != 0) && (desktopSize.height != 0)) {
                return new Rectangle(0, 0, desktopSize.width, desktopSize.height);
            }
        }
        // SDI case
        Rectangle workingSpace = null;
        int bottomSpace = 0;
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        if (org.openide.util.Utilities.isWindows()) {
            bottomSpace = OutputTab.TYPICAL_WINDOWS_TASKBAR_HEIGHT;
        }
        if (mainBounds.y < (screenSize.height / 2)) {
            workingSpace = new Rectangle(
                               0, mainBounds.y + mainBounds.height,
                               screenSize.width,
                               screenSize.height - bottomSpace - (mainBounds.y + mainBounds.height)
                           );
        } else {
            workingSpace = new Rectangle(
                               0, 0, screenSize.width, mainBounds.y
                           );
        }
        return workingSpace;
    }


    /** Compute the bounds for the mode according to the current
    * cascading point and preferred size of given top component.
    * Update (move) cascading point.
    * This methoid is used when no mode bounds are specified explixitly.
    */
    public void placeMode (Mode mode, TopComponent tc) {
        // check
        if (mode.getBounds() != null)
            return;
        // compute
        Rectangle modeBounds = new Rectangle();
        // start in cascading point
        Point cascadePoint = getCascadePoint();
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        modeBounds.x = cascadePoint.x;
        modeBounds.y = cascadePoint.y;
        // set size like top component wishes, if possible
        Dimension d = tc.getPreferredSize();
        if ((d == null) || ((d.width == 0) && (d.height == 0))) {
            // set default size
            modeBounds.width =
                Math.min(DEFAULT_MODE_WIDTH, screenSize.width - modeBounds.x);
            modeBounds.height =
                Math.min(DEFAULT_MODE_HEIGHT, screenSize.height - modeBounds.y);
        } else {
            // the bounds must not exceed screen bounds
            modeBounds.width = Math.min(d.width, screenSize.width - modeBounds.x);
            modeBounds.height = Math.min(d.height, screenSize.height - modeBounds.y);
        }
        // update cascading point
        Dimension cascadeStep = getCascadeStep();
        cascadePoint.x += cascadeStep.height;
        cascadePoint.y += cascadeStep.width;
        cascadeStepsCount--;
        // finally place the mode
        mode.setBounds(modeBounds);
    }

    public Point getDesktopScreenLocation () {
        Point loc = wm.mainWindow().getDesktopScreenLocation();
        ModeImpl westMode = WindowUtils.findConstrainedMode(workspace, PerimeterLayout.WEST, true);
        int draggerWidth = ((PerimeterLayout)workspace.desktopPane().getLayout()).getGap();
        if (westMode != null) {
            loc.x += westMode.getBounds().width + draggerWidth;
        }
        ModeImpl northMode = WindowUtils.findConstrainedMode(workspace, PerimeterLayout.NORTH, true);
        if (northMode != null) {
            loc.y += northMode.getBounds().height + draggerWidth;
        }
        return loc;
    }
    
    /** Helper utility method, corrects rectangle so that it doesn't overlap
     * with main window. Useful only in SDI mode and for situation when
     * SDI main window is near the top of the screen (which is default) */
    private void removeMainWindowOverlapping (Rectangle rect) {
        Rectangle mainBounds = wm.getOldMainWindowBounds();
        if ((rect.y > mainBounds.y) && (rect.y < (mainBounds.y + mainBounds.height))) {
            rect.y = mainBounds.y + mainBounds.height;
        }
    }
    

    /** Recomputes location of the cascading point. Called when
    * no cascade point is computed yet or when dimensions of 
    * IDE working space (either in SDI or MDI) has changed. */
    private void recomputeCascadePoint () {
        // obtain IDE working space
        Rectangle workingSpace = getWorkingSpaceBounds();
        Point cascadingOrigin = getCascadingOrigin();
        // compute new cascading point
        cascadePoint = new Point(
                           workingSpace.x + (cascadingOrigin.x * workingSpace.width / 100),
                           workingSpace.y + (cascadingOrigin.y * workingSpace.height / 100)
                       );
        // now compute how deep cascade could be
        int xSteps, ySteps;
        Dimension cascadeStep = getCascadeStep();
        if (cascadeStep.width < 0) {
            xSteps = Math.abs(
                         (cascadingOrigin.x * workingSpace.width / 100) / cascadeStep.width
                     );
        } else {
            xSteps = Math.abs(
                         ((100 - cascadingOrigin.x) * workingSpace.width / 100) / cascadeStep.width
                     );
        }
        if (cascadeStep.height < 0) {
            ySteps = Math.abs(
                         (cascadingOrigin.y * workingSpace.height / 100) / cascadeStep.height
                     );
        } else {
            ySteps = Math.abs(
                         ((100 - cascadingOrigin.y) * workingSpace.height / 100) / cascadeStep.height
                     );
        }
        cascadeStepsCount = Math.min(xSteps, ySteps) / 2;
    }

    
}
