/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.Collections;
import java.util.Collection;
import java.util.ResourceBundle;
import java.awt.Image;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.Container;
import java.awt.Component;
import java.text.MessageFormat;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectStreamException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.IOException;
import java.io.ObjectStreamField;
import java.io.Serializable;
import java.net.URL;
import javax.swing.SwingUtilities;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import org.openide.windows.*;
import org.openide.util.NbBundle;
import org.openide.TopManager;
import org.openide.awt.Actions;
import org.openide.awt.ToolbarPool;
import org.openide.util.Mutex;

import org.netbeans.core.output.OutputTab;
import org.netbeans.core.windows.util.*;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.windows.frames.DesktopPane;


/** Implementation of the workspace interface.
*
* @author Dafe Simonek
*/
public final class WorkspaceImpl implements Workspace {

    /** Name of property for the list of elements on this workspace. */
    public static final String PROP_MODES = "modes"; // NOI18N
    /** Name of property for the name of this workspace. */
    public static final String PROP_NAME = "name"; // NOI18N
    /** count number of workspaces */
    static int numberOfWorkspaces = 0;
    /** Information message shown during loading */
    static MessageFormat loadingMessage;
    /** a pool of toolbars */
    static ToolbarPool toolbarPool;

    /** programmatic name of the workspace */
    private String name;
    /** display name of the workspace. actually, it serves two purposes.
    * if fromBundle helper variable is true, then displayName contains name
    * of the bundle string to take display name from. */
    private String displayName;
    /** toolbar configuration name */
    private String toolbarConfigName = "Standard"; // NOI18N
    /** A list of all modes on this workspace */
    private ArrayList modes;
    /** Currently activated mode */
    private Mode current;
    /** Helper flag variable, true if display name is taken from the bundle */
    private boolean fromBundle;

    /** Helper variable - holds information needed
    * to renew activated top component after deserialization. */
    private WindowManagerImpl.TopComponentManager savedManager;

    /** asociation with window manager implementation */
    private WindowManagerImpl wm;
    /** property change support */
    private PropertyChangeSupport changeSupport;
    /** default workspace name */
    private static MessageFormat defaultWorkspaceName;
    /** manager of versioned serialization */
    private VersionSerializator serializationManager;
    /** Agregated visibility manager */
    private WorkspaceVisibilityManager visManager;
    /** Agregated positioning manager */
    private WorkspaceBoundsManager boundsManager;

    static final long serialVersionUID =-1865300855714270375L;

    /** Creates new workspace implementation with generated default name
    * and display name */
    public WorkspaceImpl () {
        this((String)null);
    }

    /** Creates new workspace implementation with given name.
    * Given parameter is used both for name and display name.
    */
    public WorkspaceImpl (String name) {
        this(name, name, false);
    }

    /** Creates new workspace with given programmatic name and
    * given display name. */
    public WorkspaceImpl (String name, String displayName) {
        this(name, displayName, false);
    }

    /** Creates new workspace with given programmatic name and
    * given display name.
    * @param fromBundle true if displayName parameter contains bundle string
    * rather than real display name */
    public WorkspaceImpl (String name, String displayName, boolean fromBundle) {
        this.name = name;
        this.fromBundle = fromBundle;
        if (name == null) {
            // set default name
            Object[] arr = { new Integer (++numberOfWorkspaces) };
            if (defaultWorkspaceName == null) {
                ResourceBundle bundle = NbBundle.getBundle (WorkspaceImpl.class);
                defaultWorkspaceName =
                    new MessageFormat (bundle.getString ("WorkspaceNumber"));
            }
            this.name = defaultWorkspaceName.format (arr);
        }
        // display name
        this.displayName = (displayName == null) ? this.name : displayName;
        modes = new ArrayList(10);
        this.toolbarConfigName = name;
        initialize();
    }

    /** Copy constructor, creates new workspace as a copy of given workspace.
    * Created workspace will have a list of modes with the same names,
    * bounds etc as the original, and the same top components will
    * be opened on the clone as they are in the original.
    * The name of cloned workspace will be unique, based on the
    * name of the original.
    */
    public WorkspaceImpl (WorkspaceImpl original) {
        // usual initialization
        modes = new ArrayList(10);
        initialize();
        // create a name
        int i = 1;
        while (wm.findWorkspace(name = original.name + "_" + i++) != null); // NOI18N
        // display name
        displayName = original.getDisplayName() + (i - 1);
        // copy content
        boundsManager().copyFrom(original.boundsManager());
        this.toolbarConfigName = new String(original.toolbarConfigName);
        // copy modes
        ModeImpl curMode = null;
        ModeImpl newMode = null;
        for (Iterator iter = original.modes.iterator(); iter.hasNext(); ) {
            curMode = (ModeImpl)iter.next();
            newMode = new ModeImpl(this, curMode);
            if (curMode.equals(original.current)) {
                this.current = newMode;
            }
            modes.add(newMode);
            // copy reference to the contained tcs
            newMode.fillTopComponents(curMode);
        }
    }
    
    private void initialize () {
        changeSupport = new PropertyChangeSupport(this);
        wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
    }

    /** Get the programmatic unique name of this workspace.
     * @return the programmatic name
     */
    public String getName () {
        return name;
    }

    /** Get human-presentable name of the workspace.
    * May be used, e.g., for display on the workspaces tab on the main window.
    * @return the diplay name of the workspace
    */
    public String getDisplayName () {
        String result;
        if (fromBundle) {
            result = NbBundle.getBundle(WorkspaceImpl.class).getString(displayName);
        } else {
            result = displayName;
        }
        return result;
    }

    /** Set human-presentable name of the workspace.
    * @param s the new display name
    */
    public void setDisplayName (String s) {
        //Remove shortcut due to detection of change
        String localDisplay = Actions.cutAmpersand(getDisplayName());
        String inNoShortcut = Actions.cutAmpersand(s);
        if (((localDisplay != null) && localDisplay.equals(inNoShortcut)) ||
                (localDisplay == null) && (inNoShortcut == null)) {
            // no real change
            return;
        }
        // disable fromBundle feature, it user's choice now
        fromBundle = false;
        String old = displayName;
        displayName = s;
        changeSupport.firePropertyChange (PROP_DISPLAY_NAME, old, displayName);
    }

    /** Array of all modes on this workspace.
     */
    public Set getModes () {
        return Collections.unmodifiableSet(new HashSet(modes));
    }
    
    List getModeArray () {
        List result = null;
        synchronized (modes) {
            result = new ArrayList(modes);
        }
        return result;
    }

    /** Get bounds of the workspace.
    * Delegates to static getWorkingSpaceBounds().
    */
    public Rectangle getBounds () {
        return boundsManager().getWorkingSpaceBounds();
    }

    /** Activates this workspace to be current one.
     * This leads to change of current workspace of the WindowManager.
     */
    public void activate () {
        if (java.awt.EventQueue.isDispatchThread()) {
            wm.setCurrentWorkspace(this);
        } else {
            java.awt.EventQueue.invokeLater(new Runnable() {
                public void run () {
                    wm.setCurrentWorkspace(WorkspaceImpl.this);
                }
            });
        }
    }

    /** Create a new mode.
     * @param name a unique programmatic name of the mode 
     * @param displayName a human presentable (probably localized) name
     *                    of the mode (may be used by
     *{@link org.openide.actions.DockingAction}, e.g.)
     * @param icon an icon to use for the mode (e.g. on a tab or window corner);
     *             may be <code>null</code>
     * @return the new mode
     */
    public Mode createMode (String name, String displayName, URL icon) {
        return createMode(name, displayName, icon, false);
    }

    /* @deprecated Use createMode without containerType parameter. */
    public Mode createMode (String name, String displayName, URL icon,
                            int containerType, boolean userDefined) {
        return createMode(name, displayName, icon, userDefined);
    }
    
    /** Create new mode.
     * @param name a unique programmatic name of the mode 
     * @param displayName a human presentable (probably localized) name
     *                    of the mode (may be used by
     *{@link org.openide.actions.DockingAction}, e.g.)
     * @param icon an icon to use for the mode (e.g. on a tab or window corner);
     *             may be <code>null</code>
     * @return the new mode
    * @param containerType type specifies type of frame representing
    *        the mode
    * @param userDefined true if mode is creted by user
    */
    public Mode createMode (String name, String displayName, URL icon,
                            boolean userDefined) {
        // throw exception when method is used wrongly
        if (name == null) {
            throw new IllegalArgumentException("Cannot create mode with null name!"); // NOI18N
        }
        Mode newMode = findMode(name);
        if (newMode == null) {
            newMode = new ModeImpl(name, displayName, icon, userDefined, this);
            addMode(newMode);
        }
        return newMode;
    }
    
    /** Search all modes on this workspace by name.
     * @param name the name of the mode to search for
     * @return the mode with that name, or <code>null</code> if no such mode
     *         can be found
     */
    public Mode findMode (String name) {
        if (name == null) {
            return null;
        }
        Mode[] modeArray = null;
        synchronized (modes) {
            modeArray = (Mode[])modes.toArray(new Mode[0]);
        }
        for (int i = 0; i < modeArray.length; i++) {
            if (name.equals(modeArray[i].getName()))
                return modeArray[i];
        }
        return null;
    }

    /** Finds mode the component is in on this workspace.
     *
     * @param c component to find mode for
     * @return the mode or null if the component is not visible on this workspace
     */
    public Mode findMode (TopComponent c) {
        Mode[] modeArray = null;
        synchronized (modes) {
            modeArray = (Mode[])modes.toArray(new Mode[0]);
        }
        for (int i = 0; i < modeArray.length; i++) {
            TopComponent[] tcs = modeArray[i].getTopComponents();
            for (int y = 0; y < tcs.length; y++) {
                if (c.equals(tcs[y])) {
                    return modeArray[i];
                }
            }
        }
        return null;
    }

    /** @return true if given top component is opened on this workspace,
    * false otherwise */
    public boolean isOpened (TopComponent tc) {
        return wm.findManager(tc).isOpened(this);
    }

    /** Clears this workspace and removee this workspace from
    * window manager, if possible */
    public void remove () {
        ModeImpl[] modeArray = null;
        synchronized (modes) {
            modeArray = (ModeImpl[])modes.toArray(new ModeImpl[0]);
        }
        // clear workspace
        if (modeArray.length > 0) {
            // iterate through modes and try to close them
            for (int i = 0; i < modeArray.length; i++) {
                modeArray[i].close();
            }
            synchronized (modes) {
                modes.clear();
            }
            // fire changes
            changeSupport.firePropertyChange(PROP_MODES, null, null);
        }
        // remove this workspace
        Workspace[] wss = wm.getWorkspaces();
        int i = 0;
        for (i = 0; i < wss.length; i++) {
            if (this.equals(wss[i])) {
                break;
            }
        }
        if (i < wss.length) {
            Workspace[] nwss = new Workspace[wss.length-1];
            System.arraycopy(wss, 0, nwss, 0, i);
            System.arraycopy(wss, i + 1, nwss, i, wss.length - (i + 1));
            wm.setWorkspaces(nwss);
        }
    }

    /** Add a property change listener.
     * @param list the listener to add
     */
    public void addPropertyChangeListener (PropertyChangeListener l) {
        changeSupport.addPropertyChangeListener(l);
    }

    /** Remove a property change listener.
     * @param list the listener to remove
     */
    public void removePropertyChangeListener (PropertyChangeListener l) {
        changeSupport.removePropertyChangeListener(l);
    }

    /** Shows or hides all content of this workspace. */
    public void setVisible (boolean visible) {
        visManager().setVisible(visible);
    }

    /** @return true when workspace is visible, false otherwise */
    public boolean isVisible () {
        return visManager().isVisible();
    }
   
    /**
     * @return toolbar configuration name
     */
    public String getToolbarConfigName () {
        return toolbarConfigName;
    }

    /**
     * Sets toolbar configuration for workspace
     * @param name of toolbar configuration for this workspace
     */
    public void setToolbarConfigName (String name) {
        toolbarConfigName = name;
        if (this.equals(wm.getCurrentWorkspace()))
            activateConfiguration();
    }

    /**
     * Prepare toolbar configuration
     */
    void activateConfiguration () {
        if (toolbarPool == null)
            toolbarPool = ToolbarPool.getDefault();
        toolbarPool.setConfiguration (toolbarConfigName);
    }

    /** Fire mode change after changes was finished. */
    public void fireModeChange () {
        changeSupport.firePropertyChange(PROP_MODES, null, null);
    }
    
    /** Adds given mode to this workspace. */
    public void addMode (Mode mode) {
        if (mode == null) {
            throw new IllegalArgumentException("Cannot add null mode to the workspace"); // NOI18N
        }
        synchronized (modes) {
            if (modes.contains(mode)) {
                return;
            }
            modes.add(mode);
        }
        changeSupport.firePropertyChange(PROP_MODES, null, null);
    }

    /** Removes given mode from this workspace. */
    public void removeMode (Mode mode) {
        boolean removed = false;
        synchronized (modes) {
            removed = modes.remove(mode);
        }
        if (removed) {
            changeSupport.firePropertyChange(PROP_MODES, null, null);
        }
    }

    /** Sets new cascading origin for this workspace.
    * Cascading origin is specified using percentage and 
    * relative to left upper corner of IDE desktop
    * (main window's left bottom corner for SDI,
    * left bottom corner of main toolbar and workspace panel for MDI.<br>
    * So - setCascadingOrigin(0, 0) means upper left corner,
    * setCascadingOrigin(100, 100) means bottom right
    * corner of "IDE working space"
    */
    public void setCascadingOrigin (Point cascadingOrigin) {
        boundsManager().setCascadingOrigin(cascadingOrigin);
    }

    /* @return Current cascading origin for this workspace */
    public Point getCascadingOrigin () {
        return boundsManager.getCascadingOrigin();
    }
    
    /** Compute the bounds for the mode according to the current
    * cascading point and preferred size of given top component.
    * Update (move) cascading point.
    * This methoid is used when no mode bounds are specified explixitly.
    */
    void placeMode (Mode mode, TopComponent tc) {
        boundsManager().placeMode(mode, tc);
    }

    /** Utility method which activates top component that was active
    * before serialization. */
    void reactivateSavedTc () {
        if (savedManager == null) {
            return;
        }
        final TopComponent toBeActivated = savedManager.getComponent();
        savedManager = null;
        // component can be resolved to null after bad serialization,
        // be prepared to such situation
        if (toBeActivated != null) {
            // delay focus transferring and pass it to the AWT Event queue
            SwingUtilities.invokeLater(new Runnable() {
                                           public void run () {
                                               toBeActivated.requestFocus();
                                           }
                                       });
        }
    }
    
    /** Called when some top component is activated, puts mode containing given
     * top component to the first slot in modes array. Modes array then keeps
     * order properly. Order is needed to show cascade of overlapping frames
     * correctly anytime */ 
    void componentActivated (TopComponent activeTc) {
        if (visManager().isProcessing()) {
            return;
        }
        Mode activeMode = findMode(activeTc);
        if (activeMode != null) {
            synchronized (modes) {
                modes.remove(activeMode);
                modes.add(0, activeMode);
            }
        }
    }

    /** Getter for the formatted information message during loading */
    static MessageFormat getLoadingMessage () {
        if (loadingMessage == null) {
            loadingMessage =
                new MessageFormat(NbBundle.getBundle(Workspace.class).
                                  getString("FMT_LoadingWorkspace"));
        }
        return loadingMessage;
    }

    /** Utility method, removes modes which are user defined and
    * orphan from the given set of the modes */
    void excludeUserOrphans (ArrayList modes) {
        ModeImpl curMode = null;
        ModeImpl[] modeArray = (ModeImpl[])modes.toArray(new ModeImpl[0]);
        for (int i = 0; i < modeArray.length; i++) {
            if (modeArray[i].isUserDefined() && modeArray[i].isOrphan()) {
                modes.remove(modeArray[i]);
            }
        }
    }
    
    /** Safe accessor for desktop pane instance */
    public DesktopPane desktopPane () {
        return visManager().desktopPane();
    }
    
    Point getDesktopScreenLocation () {
        return boundsManager().getDesktopScreenLocation();
    }

    /** Safe accessor for aggregated visibility manager */
    private WorkspaceVisibilityManager visManager () {
        if (visManager == null) {
            visManager = new WorkspaceVisibilityManager(this);
        }
        return visManager;
    }
    
    /** Safe accessor for aggregated visibility manager */
    private WorkspaceBoundsManager boundsManager () {
        if (boundsManager == null) {
            boundsManager = new WorkspaceBoundsManager(this);
        }
        return boundsManager;
    }

    /** Let instance of properly parametrized SerializationReplacer to keep
    * persistent state of this workspace
    */
    private Object writeReplace ()
    throws ObjectStreamException {
        // exclude unwanted modes (user defined orphans)
        excludeUserOrphans(modes);
        return new SerializationReplacer(this);
    }

    /** Called when first phase of WS deserialization is done.
    * Ask all contained modes for self validation.
    */
    void validateSelf () {
        // turn to array to prevent concurrent modifications
        ModeImpl[] modeArray =
            (ModeImpl[])modes.toArray(new ModeImpl[modes.size()]);
        // validate modes themself
        for (int i = 0; i < modeArray.length; i++) {
            modeArray[i].validateSelf(this);
        }
    }

    /** Called after validateSelf() as second validation step.
    * Ask all contained modes data validation. */
    void validateData () {
        boundsManager().correctBounds();
        visManager().validateModes();
        // notify about modes change
        changeSupport.firePropertyChange(PROP_MODES, null, null);
    }

    /** Class that acts as serialization manager for workspace implementation */
    private static final class SerializationReplacer implements Serializable {
        /** Unique ID */
        static final long serialVersionUID =-1110255063830989275L;
        /** Description of serialized fields */
        private static final String NAME = "name";
        private static final String DISPLAY_NAME = "displayName";
        private static final String CASCADE_STEP = "cascadeStep";
        private static final String CASCADE_STEPS_COUNT = "cascadeStepsCount";
        private static final String CASCADE_POINT = "cascadePoint";
        private static final String CASCADING_ORIGIN = "cascadingOrigin";
        private static final String TOOLBAR_CONFIG_NAME = "toolbarConfigName";
        private static final String MODES = "modes";
        private static final String CURRENT_MODE = "current";
        private static final String ACTIVE_TCM = "activeTCM";
        private static final String FROM_BUNDLE = "fromBundle";
        private static final String DOMINANCE_LIST = "domList";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(NAME, String.class),
            new ObjectStreamField(DISPLAY_NAME, String.class),
            new ObjectStreamField(CASCADE_STEP, Dimension.class),
            new ObjectStreamField(CASCADE_STEPS_COUNT, Integer.class),
            new ObjectStreamField(CASCADE_POINT, Point.class),
            new ObjectStreamField(CASCADING_ORIGIN, Point.class),
            new ObjectStreamField(TOOLBAR_CONFIG_NAME, String.class),
            new ObjectStreamField(MODES, java.util.ArrayList.class),
            new ObjectStreamField(CURRENT_MODE, Mode.class),
            new ObjectStreamField(ACTIVE_TCM, WindowManagerImpl.TopComponentManager.class),
            new ObjectStreamField(FROM_BUNDLE, Boolean.class),
            new ObjectStreamField(DOMINANCE_LIST, ArrayList.class)
        };
        
        /** Asociation with container which data needs to be serialized */
        transient WorkspaceImpl source;
        
        SerializationReplacer (WorkspaceImpl source) {
            this.source = source;
        }

        /** Serialization of all workspaces */
        private void writeObject (ObjectOutputStream oos)
                     throws IOException {
            // write fields
            ObjectOutputStream.PutField pf = oos.putFields();
            WorkspaceBoundsManager boundsManager = source.boundsManager();
            pf.put(NAME, source.name);
            pf.put(DISPLAY_NAME, source.displayName);
            pf.put(CASCADE_STEP, boundsManager.getCascadeStep());
            pf.put(CASCADE_STEPS_COUNT, new Integer(boundsManager.getCascadeStepsCount()));
            pf.put(CASCADE_POINT, boundsManager.getCascadePoint());
            pf.put(TOOLBAR_CONFIG_NAME, source.toolbarConfigName);
            pf.put(MODES, source.modes);
            pf.put(CURRENT_MODE, source.current);
            pf.put(CASCADING_ORIGIN, boundsManager.getCascadingOrigin());
            pf.put(FROM_BUNDLE, new Boolean(source.fromBundle));
            // save active top component
            TopComponent activeTc = source.visManager().getSavedActive();
            if (activeTc == null) {
                activeTc = TopComponent.getRegistry().getActivated();
            }
            if ((activeTc != null) && (activeTc.isOpened(source))) {
                pf.put(ACTIVE_TCM, WindowManagerImpl.findManager(activeTc));
            } else {
                pf.put(ACTIVE_TCM, null);
            }
            List domList = source.visManager().getDominanceList();
            ArrayList arDomList = null;
            if (domList != null) {
                arDomList = new ArrayList(domList.size());
                arDomList.addAll(domList);
            }
            pf.put(DOMINANCE_LIST, arDomList); 
            oos.writeFields();
        }

        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField gf = ois.readFields();
            String name = (String)gf.get(NAME, null);
            source = (WorkspaceImpl)TopManager.getDefault().getWindowManager().findWorkspace(name);
            if (source == null) {
                // workspace don't exist, create new one
                source = new WorkspaceImpl(name);
            }
            source.setDisplayName((String)gf.get(DISPLAY_NAME, null));
            WorkspaceBoundsManager boundsManager = source.boundsManager();
            boundsManager.setCascadeStep((Dimension)gf.get(CASCADE_STEP, null));
            boundsManager.setCascadeStepsCount(((Integer)gf.get(CASCADE_STEPS_COUNT, null)).intValue());
            boundsManager.setCascadePoint((Point)gf.get(CASCADE_POINT, null));
            source.setToolbarConfigName((String)gf.get(TOOLBAR_CONFIG_NAME, null));
            source.current = (Mode)gf.get(CURRENT_MODE, null);
            boundsManager.setCascadingOrigin((Point)gf.get(CASCADING_ORIGIN, null));
            source.fromBundle = ((Boolean)gf.get(FROM_BUNDLE, null)).booleanValue();
            source.savedManager = (WindowManagerImpl.TopComponentManager)gf.get(ACTIVE_TCM, null);
            // beware, we must deal both with older and newer data - older
            // data type was HashSet, newer is ArrayList. So we can cast
            // to collection and we can cover both cases
            Collection newModes = (Collection)gf.get(MODES, null);
            ArrayList modes = source.modes;
            synchronized (modes) {
                modes.clear();
                modes.addAll(newModes);
            }
            // older data didn't have DOMINANCE_LIST field, we must take care            
            //Bugfix #10993, workaround for JDK 1.2.2
            //if (gf.getObjectStreamClass().getField(DOMINANCE_LIST) != null) {
            if (WindowUtils.hasObjectStreamField(gf, DOMINANCE_LIST)) {
                source.visManager().setDominanceList((List)gf.get(DOMINANCE_LIST, null));
            }
        }
    
        /** Resolves deserialized SerializationReplacer to the singleton
        * instance of WindowManagerImpl */
        private Object readResolve () throws ObjectStreamException {
            return source;                            
        }

    } // end of SerializationReplacer
    
    
/******** Old serialization, must be here for compatibility with 3.1 and older versions */    

    /** Accessor to the versioned serialization manager */
    private VersionSerializator serializationManager () {
        if (serializationManager == null) {
            serializationManager = createSerializationManager();
        }
        return serializationManager;
    }

    /** Creates new serialization manager filled with our versions */
    private static VersionSerializator createSerializationManager () {
        VersionSerializator result = new VersionSerializator();
        result.putVersion(new Version1());
        result.putVersion(new Version2());
        result.putVersion(new Version3());
        return result;
    }
    
    /** Basic version of persistence for workspace implementation.
    * Method assignData(workspace) must be call before writing */
    private static class Version1
        implements DefaultReplacer.ResVersionable {

        /* identification string */
        public static final String NAME = "Version_1.0"; // NOI18N

        /** variables which form persistent state of the workspace */
        String name;
        String displayName;
        Dimension cascadeStep;
        int cascadeStepsCount;
        Point cascadePoint;
        String toolbarConfigName;
        HashSet modes;
        Mode current;
        /** asociation with workspace to be written, used during writing */
        WorkspaceImpl workspace;

        /** Identification of the version */
        public String getName () {
            return NAME;
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            // read fields
            name = (String)in.readObject();
            displayName = (String)in.readObject();
            cascadeStep = (Dimension)in.readObject();
            cascadeStepsCount = ((Integer)in.readObject()).intValue();
            cascadePoint = (Point)in.readObject();
            toolbarConfigName = (String)in.readObject();
            modes = (HashSet)in.readObject();
            current = (Mode)in.readObject();
        }

        /** write the data of the version to given output. Should be never
         * called, here only for backward compatibility reasons */
        public void writeData (ObjectOutput out)
        throws IOException {
            throw new InternalError();
        }

        /** First tries to resolve to existing workspace with the name
        * equal to deserialized name. If no such workspace can be found,
        * returns properly initialized newly created instance of workspace
        * implementation */
        public Object resolveData ()
        throws ObjectStreamException {
            WorkspaceImpl result = (WorkspaceImpl)
                                   TopManager.getDefault().getWindowManager().findWorkspace(name);
            if (result == null) {
                // workspace don't exist, create new one and fill it
                result = new WorkspaceImpl(name);
                result.displayName = displayName;
                WorkspaceBoundsManager boundsManager = result.boundsManager();
                boundsManager.setCascadeStep(cascadeStep);
                boundsManager.setCascadeStepsCount(cascadeStepsCount);
                boundsManager.setCascadePoint(cascadePoint);
                result.toolbarConfigName = toolbarConfigName;
                result.modes = new ArrayList(modes);
                result.current = current;
            } else {
                // try to add modes that don't exist yet in the workspace,
                // but leave other attributes untouched
                Mode curMode = null;
                for (Iterator iter = modes.iterator(); iter.hasNext(); ) {
                    curMode = (Mode)iter.next();
                    if (result.findMode(curMode.getName()) == null) {
                        result.addMode(curMode);
                    }
                }
            }
            return result;
        }

    } // end of Version1 inner class

    /** Enhanced serialization protocol - adds cascading origin,
    * identification of last activated top component */
    private static class Version2 extends Version1 {
        /* identification string */
        public static final String NAME2 = "Version_2.0"; // NOI18N

        Point cascadingOrigin;
        WindowManagerImpl.TopComponentManager savedManager;

        /** Identification of the version */
        public String getName () {
            return NAME2;
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            super.readData(in);
            cascadingOrigin = (Point)in.readObject();
            savedManager = (WindowManagerImpl.TopComponentManager)in.readObject();
        }

        /** write the data of the version to given output */
        public void writeData (ObjectOutput out)
        throws IOException {
            throw new InternalError();
        }

        public Object resolveData ()
        throws ObjectStreamException {
            WorkspaceImpl result = (WorkspaceImpl)super.resolveData();
            WorkspaceImpl existing = (WorkspaceImpl)
                                     TopManager.getDefault().getWindowManager().findWorkspace(this.name);
            if (existing == null) {
                // workspace didn't exist, fill it
                result.boundsManager().setCascadingOrigin(cascadingOrigin);
                result.savedManager = savedManager;
            }
            return result;
        }

    } // end of Version2 inner class

    /** Enhanced serialization protocol - adds fromBundle flag */
    private static class Version3 extends Version2 {
        /* identification string */
        public static final String NAME3 = "Version_3.0"; // NOI18N

        boolean fromBundle;

        /** Identification of the version */
        public String getName () {
            return NAME3;
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            super.readData(in);
            fromBundle = ((Boolean)in.readObject()).booleanValue();
        }

        /** write the data of the version to given output */
        public void writeData (ObjectOutput out)
        throws IOException {
            throw new InternalError();
        }

        public Object resolveData ()
        throws ObjectStreamException {
            WorkspaceImpl result = (WorkspaceImpl)super.resolveData();
            WorkspaceImpl existing = (WorkspaceImpl)
                                     TopManager.getDefault().getWindowManager().findWorkspace(this.name);
            if (existing == null) {
                // workspace didn't exist, fill it
                result.fromBundle = fromBundle;
            }
            return result;
        }

    } // end of Version3 inner class

    /** Implementation of persistent access to our version serializator */
    private static final class VSAccess implements DefaultReplacer.Access {
        /** serialVersionUID */
        private static final long serialVersionUID = 2367034636261092266L;

        /** version serializator, used only during writing */
        transient VersionSerializator vs;

        public VSAccess (VersionSerializator vs) {
            this.vs = vs;
        }

        public VersionSerializator getVersionSerializator () {
            return (vs == null) ? createSerializationManager() : vs;
        }

    } // end of VSAccess inner class

}

