/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows;

import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.awt.Container;

import java.beans.PropertyVetoException;

import javax.swing.SwingUtilities;

import org.openide.windows.TopComponent;

import org.netbeans.core.windows.frames.DesktopPane;
import org.netbeans.core.windows.frames.PerimeterLayout;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.FrameType;

/** Utility class which works as visibility manager for workspace.
 * This class is agregated with WorkspaceImpl (in fact part of WorkspaceImpl,
 * just isolated to separate outer class for better maintainability. 
 *
 * @author  Dafe Simonek
 */
class WorkspaceVisibilityManager extends Object 
                                 implements Runnable, 
                                            DeferredPerformer.DeferredCommand {
                                            
    /** workspace visibility flag */
    private boolean visible = false;
    /** agregation with WorkspaceImpl */
    private WorkspaceImpl workspace;
    /** window manager asociation */
    private WindowManagerImpl wm = WindowManagerImpl.getDefault();
    /** helper flag, true during process of displaying or hiding of modes */
    private boolean isProcessing = false;
    /** dominance info of desktop pane. Used only in MDI mode */
    private List dominanceList;
    /** holds previously active top component when
    * workspace was visible last time */
    private TopComponent savedActive;
    /** reference to pending request to show workspace */
    private DeferredPerformer.DeferredCommand pendingRequest;
    /** Asociated MDI desktop, (can be null if in SDI mode) */
    private DesktopPane desktopPane;
    

    /** Creates new WorkspaceVisibilityManager */
    WorkspaceVisibilityManager (WorkspaceImpl workspace) {
        this.workspace = workspace;
    }

    /** Sets visibility state for asociated workspace */
    public void setVisible (boolean visible) {
        if (pendingRequest != null) {
            if (visible) {
                // coalesce this request if already pending 
                return;
            } else {
                // cancel pending request, because this new request is opposite
                boolean cancelled = wm.deferredPerformer().removeRequest(pendingRequest);
                pendingRequest = null;
            }
        }
        if (visible && (!wm.deferredPerformer().canImmediatelly())) {
            // show workspace later
            pendingRequest = this;
            wm.deferredPerformer().putRequest(
                this, new DeferredPerformer.DeferredContext(null, true)
            );
            return;
        } else {
            // show immediatelly
            doSetVisible(visible);
        }
    }
    
    /** Implementation of DeferredPerformer.DeferredCommand interface.
     * Sets this workspace to be visible.
     */
    public void performCommand (DeferredPerformer.DeferredContext context) {
        pendingRequest = null;
        doSetVisible(true);
    }

    /** Implementation of runnable interface. Requests focus for previously
     * active top component. Called from refreshProperties */
    public void run () {
        if (savedActive != null) {
            savedActive.requestFocus();
            savedActive = null;
        }
    }
    
    /** @return currently active mode or null. Active mode is
     * the mode on current workspace which contains currently activated
     * top component.
     */
    private ModeImpl getActiveMode () {
        TopComponent comp = TopComponent.getRegistry().getActivated();
        if (comp == null) {
            return null;
        }
        return (ModeImpl) workspace.findMode(comp);
    }
    
    /** Real implementation of setting workspace visibility */
    private void doSetVisible (boolean visible) {
        if (this.visible == visible) {
            return;
        }
        isProcessing = true;
        this.visible = visible;
        // save active for later reactivation if hiding
        if (!visible) {
            //Bugfix #9932 explicitly deselect frame of active mode
            //when workspace is to be hidden.
            //It was not deselected before at all => frame was not
            //selected when workspace was shown again => caused
            //bug #9932 - actions was disabled in context menu
            //for nodes in explorer.
            ModeImpl mode = getActiveMode();
            if (mode != null) {
                FrameType frame = mode.getFrameInstance();
                if (frame != null) {
                    try {
                        frame.setSelected(false);
                    } catch(PropertyVetoException e){}
                }
            }
            //End of bugfix #9932
            savedActive = TopComponent.getRegistry().getActivated();
            if ((savedActive != null) && (workspace.findMode(savedActive) == null))
                savedActive = null;
            wm.activateComponent(null);
        }
        boolean isMdi = UIModeManager.MDI_MODE == wm.getUIMode();
        Container contentPane = WindowManagerImpl.mainWindow().getContentPane();
        // show desktop if in MDI mode
        if (isMdi && visible) {
            //Fixed bug #9579
            //Added check if desktopPane() has already parent
            //because it could be already added to contentPane in
            //ModeImpl.setVisible
            //Adding again causes removing from previous parent 
            //(new and previous parnet are the same) and removing
            //all events for this source from event queue -> method 
            //componentShown of ModeImpl is not called -> wrong tab (Runtime)
            //was selected in Explorer after IDE restart.
            if (desktopPane().getParent() == null) {
                contentPane.add(desktopPane());
            }
        }
        // show or hide modes
        processModes(visible, isMdi);
        // hide desktop if needed
        if ((isMdi && !visible) || (!isMdi && (desktopPane() != null))) {
            contentPane.remove(desktopPane());
        }
        contentPane.invalidate();
        contentPane.validate();
        contentPane.repaint();
        
        // update toolbar configuration, active component etc
        refreshProperties(visible, isMdi);
        isProcessing = false;
    }

    /** @return current visibility state of asociated workspace */
    public boolean isVisible () {
        return visible;
    }

    /** @return true when busy processing visibility operations,
     * false otherwise */
    public boolean isProcessing () {
        return isProcessing;
    }
    
    /** @return top component previously active (during last visible session)
     * or null if no such component is available */
    public TopComponent getSavedActive () {
        return savedActive;
    }
    
    /** Assigns dominance list which will be used for mode order
     * during displaying process. Method don't update current state if
     * workspace is visible. Use only for re-setting of saved dominance.
     * Usable only in MDI
     */
    public void setDominanceList (List dominanceList) {
        this.dominanceList = dominanceList;
    }

    /** @return perimeter pane dominance list that affects displaying order.
     * Usable only in MDI
     */
    public List getDominanceList () {
        List result = dominanceList;
        if ((result == null) && visible && 
            (UIModeManager.MDI_MODE == wm.getUIMode())) {
            result = ((PerimeterLayout)desktopPane().
                     getLayout()).getDominance();
        }
        return result;
    }
    
    /** Performs validation of workspace's modes in right order */
    public void validateModes () {
        List modes = workspace.getModeArray();
        ArrayList cascadedModes = new ArrayList(modes.size());
        ArrayList nonCascadedModes = new ArrayList(modes.size());
        divideModes(modes, cascadedModes, nonCascadedModes);
        if (!nonCascadedModes.isEmpty()) {
            if (dominanceList != null) {
                prioritize(nonCascadedModes, dominanceList);
                dominanceList = null;
            }
            processModesValidation(nonCascadedModes, false);
        }
        processModesValidation(cascadedModes, true);
    }

    /** Safe accessor for desktop pane instance */
    DesktopPane desktopPane () {
        if (desktopPane == null) {
            desktopPane = new DesktopPane();
        }
        return desktopPane;
    }
    
    /** Takes care of transferring visibility status to all modes of asociated
     * workspace, respects saved perimeter layout dominance, if any */
    private void processModes (boolean visible, boolean isMdi) {
        List modes = workspace.getModeArray();
        if (visible) {
            ArrayList cascadedModes = new ArrayList(modes.size());
            ArrayList nonCascadedModes = new ArrayList(modes.size());
            divideModes(modes, cascadedModes, nonCascadedModes);
            if (!nonCascadedModes.isEmpty()) {
                if (dominanceList != null) {
                    prioritize(nonCascadedModes, dominanceList);
                    dominanceList = null;
                }
                setModesVisibility(nonCascadedModes, visible, false);
            }
            setModesVisibility(cascadedModes, visible, true);
        } else {
            if (isMdi) {
                dominanceList = ((PerimeterLayout)desktopPane().
                                 getLayout()).getDominance();
            }
            setModesVisibility(modes, visible, true);
        }
    }
    
    /** Divides given mode list to list of modes which are represented by
     * floating (possible overlapped, cascaded) windows and list of others
     */
    private void divideModes (List sourceModes, List cascadedModes, List nonCascadedModes) {
        ModeImpl curMode = null;
        for (Iterator iter = sourceModes.iterator(); iter.hasNext(); ) {
            curMode = (ModeImpl)iter.next();
            if (WindowTypesManager.DESKTOP_FRAME.equals(curMode.getFrameType())) {
                nonCascadedModes.add(curMode);
            } else {
                cascadedModes.add(curMode);
            }
        }
    }
    
    /** Shuffles given mode list to reflect given dominance order */
    private void prioritize (List modes, List dominance) {
        String[] dominanceArray = (String[])dominance.toArray(new String[dominance.size()]);
        ModeImpl curMode = null;
        ModeImpl[] modesArray = null;
        for (int i = 0; i < dominanceArray.length; i++) {
            modesArray = (ModeImpl[])modes.toArray(new ModeImpl[modes.size()]);
            for (int y = 0; y < modesArray.length; y++) {
                curMode = modesArray[y];
                if (curMode.getConstraints().equals(dominanceArray[i])) {
                    // move processed mode to proper position
                    modes.remove(curMode);
                    modes.add(i, curMode);
                }
            }
        }
    }

    /** Sets specified visibility to modes in specified order. */
    private void setModesVisibility (List modes, boolean visible, boolean reverse) {
        ModeImpl[] modesArray = (ModeImpl[])modes.toArray(new ModeImpl[modes.size()]);
        // we don't use Collections.reverse, because this will be faster
        if (reverse) {
            for (int i = modesArray.length - 1; i >= 0; i--) {
                modesArray[i].setVisible(visible);
            }
        } else {
            for (int i = 0; i < modesArray.length; i++) {
                modesArray[i].setVisible(visible);
            }
        }
    }
    
    /** Sets specified visibility to modes in specified order. */
    private void processModesValidation (List modes, boolean reverse) {
        ModeImpl[] modesArray = (ModeImpl[])modes.toArray(new ModeImpl[modes.size()]);
        // we don't use Collections.reverse, because this will be faster
        if (reverse) {
            for (int i = modesArray.length - 1; i >= 0; i--) {
                modesArray[i].validateData();
            }
        } else {
            for (int i = 0; i < modesArray.length; i++) {
                modesArray[i].validateData();
            }
        }
        
    }
    
     /** Activates toolbar config and saved active component. Helper method,
     * called from setVisible
     */
    private void refreshProperties (boolean visible, boolean isMdi) {
        if (visible) {
            workspace.activateConfiguration();
            if (savedActive != null) {
                // request focus later
                SwingUtilities.invokeLater(this);
            } else {
                // activation after deserialization
                workspace.reactivateSavedTc();
            }
        }
    }
    
}
