/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.awt.Container;
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Image;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ComponentListener;
import java.util.Collection;
import java.util.Iterator;
import java.util.EventObject;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.ObjectStreamField;
import java.io.Serializable;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

import org.openide.TopManager;
import org.openide.windows.*;
import org.openide.actions.*;
import org.openide.nodes.Node;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.CallableSystemAction;
import org.openide.awt.JPopupMenuPlus;
import org.openide.util.WeakListener;

import org.netbeans.core.windows.*;
import org.netbeans.core.actions.DockingAction;
import org.netbeans.core.NbNodeOperation;
import org.netbeans.core.windows.WindowManagerImpl.TopComponentManager;


/** Common abstract superclass for top component containers.
 * Implements basic functionality which is expected to be common for vast
 * majority of top component containers.
 * Various top component container implementations are expected to extend this
 * default implementation, add concrete visual representation.
 *
 * @author Dafe Simonek
*/
abstract class DefaultContainerImpl extends JPanel 
                                    implements TopComponentContainer,
                                                ActionPerformer,
                                                PropertyChangeListener,
                                                FrameTypeListener,
                                                ComponentListener {
    
    /** List of all contained top components */
    private ArrayList topComps;
    /** Helper variable, is true if this container is activated (has focus). */
    private boolean hasFocus = false;
    /** Currently selected top component in this container */
    private TopComponent selected;
    
    /** Clone action */
    protected static CloneViewAction clone =
        (CloneViewAction)SystemAction.get(CloneViewAction.class);
    /** Undock action */
    protected static UndockAction undock =
        (UndockAction)SystemAction.get(UndockAction.class);
    /** Nexttab action */
    protected static NextTabAction nextTab =
        (NextTabAction)SystemAction.get(NextTabAction.class);
    /** Previoustab action */
    protected static PreviousTabAction prevTab =
        (PreviousTabAction)SystemAction.get(PreviousTabAction.class);
    /** Close view action */
    protected static CloseViewAction closeView =
        (CloseViewAction)SystemAction.get(CloseViewAction.class);
    /** Docking action */
    protected static DockingAction docking =
        (DockingAction)SystemAction.get(DockingAction.class);

    /** Asociation with window manager implementation */
    protected WindowManagerImpl wm;
    /** Asociation with the mode which we represent */
    protected ModeImpl mode;
    /** listener to the mode which owns this container */
    private PropertyChangeListener weakModeL;
    /** listener to the container frame itself */
    private FrameTypeListener weakFrameL;
    /** listener to the container frame itself */
    private ComponentListener weakComponentL;
    /** atttached flag - true if top components are attached to AWT hierarchy, false otherwise */
    private boolean attached;
    /** support for property changes event source */
    private PropertyChangeSupport propChangeSupport;
    /** holds deserialized data between deserialization phases */
    private SerializedData serData;
    
    /** holds value of delay for reactivating selected component */
    private static int waitBeforeActivateComponent = -1; // msec
    /** selected component reactivation task, non-null when running */
    private static Runnable pendingNodeActivator = null;    
    
    /** Default constructor */
    protected DefaultContainerImpl () {
        this(null);
    }

    /** Constructor for use after deserialization. Asociates given data
     * with this container */
    protected DefaultContainerImpl (SerializedData serData) {
        super();
        wm = (WindowManagerImpl)TopManager.getDefault().getWindowManager();
        this.serData = serData;
    }

    /** Attaches this container to the given mode implementation and detaches
     * from old mode, if possible */
    public void attachToMode (ModeImpl mode) {
        ModeImpl oldMode = this.mode;
        ModeImpl newMode = mode;
        changeModeDetachOld(oldMode);
        this.mode = newMode;
        changeModeAttachNew(newMode);
        changeFrame(
            (oldMode == null) ? null : oldMode.getFrameInstance(),
            (newMode == null) ? null : newMode.getFrameInstance()
        );
    }
    
    //Replaced by changeModeDetachOld and changeModeAttachNew
    /** Detaches from old mode and attaches to new mode. It is now split
        to changeModeDetachOld and changeModeAttachNew to make sure that
        property change event for old mode is not processed after new mode
        is set. Property change event can come from non AWT thread. */
    /*private void changeMode (ModeImpl oldMode, ModeImpl newMode) {
        if (oldMode != null) {
            oldMode.removePropertyChangeListener(weakModeL);
            //07 Feb 2001 by Marek Slama
            //Stop listening to top components in old mode
            //Bug fix 9433
            ArrayList topComps = topComponents();
            TopComponent curTc;
            
            for (Iterator iter = topComps.iterator(); iter.hasNext(); ) {
                curTc = (TopComponent)iter.next();
                wm.findManager(curTc).removePropertyChangeListener(this);
            }
            //End
            
            weakModeL = null;
        }
        if (newMode != null) {
            weakModeL = WeakListener.propertyChange(this, newMode);
            mode.addPropertyChangeListener(weakModeL);
        }
        changeFrame(
            (oldMode == null) ? null : oldMode.getFrameInstance(),
            (newMode == null) ? null : newMode.getFrameInstance()
        );
            
    }*/

    /** Detaches from old mode, together with changeModeAttachNew replacement 
        of changeMode. changeFrame must called separately. */
    private void changeModeDetachOld (ModeImpl oldMode) {
        if (oldMode != null) {
            oldMode.removePropertyChangeListener(weakModeL);
            //07 Feb 2001 by Marek Slama
            //Stop listening to top components in old mode
            //Bug fix 9433
            ArrayList topComps = topComponents();
            TopComponent curTc;
            
            for (Iterator iter = topComps.iterator(); iter.hasNext(); ) {
                curTc = (TopComponent)iter.next();
                wm.findManager(curTc).removePropertyChangeListener(this);
            }
            //End
            
            weakModeL = null;
        }
    }

    /** Attaches to new mode, together with changeModeDetachOld replacement 
        of changeMode. changeFrame must called separately. */
    private void changeModeAttachNew (ModeImpl newMode) {
        if (newMode != null) {
            weakModeL = WeakListener.propertyChange(this, newMode);
            newMode.addPropertyChangeListener(weakModeL);
        }
    }

    /** Detaches from old frame and attaches to new frame */
    private void changeFrame (FrameType oldFrame, FrameType newFrame) {
        if (oldFrame != null) {
            oldFrame.removeFrameTypeListener(weakFrameL);
            oldFrame.getRootPaneContainer().removeComponentListener(weakComponentL);
            weakFrameL = null;
            weakComponentL = null;
        }
        if (newFrame != null) {
/* Commented out for now - weak listeners should be used when merged back
 * to main trunk. Now for compatibility reasons we temporarily use normal
 * listeners.
            weakFrameL = (FrameTypeListener)WeakListener.create(
                            FrameTypeListener.class, this, newFrame
                         );
            weakComponentL = (ComponentListener)WeakListener.create(
                            ComponentListener.class, this, newFrame
                         );
 */
            weakFrameL = this;
            weakComponentL = this;
            newFrame.addFrameTypeListener(weakFrameL);
            newFrame.getRootPaneContainer().addComponentListener(weakComponentL);
        }
    }
    
    /** @return an array containing all top components which are currently
    * managed by this container */
    public TopComponent[] getTopComponents () {
        ArrayList comps = topComponents();
        return (TopComponent[])(comps.toArray(new TopComponent[comps.size()]));
    }

    /** Adds given top component to this container.
    * Beware, the code must be written to ensure that top component
    * will receive addNotify() before componentActivated(). 
    *
    * @param tc top component to add
    * @return false if top component already exixts in container or 
    * cannot be added due to container-dependent restrictions,
    * true otherwise
    */
    public boolean addTopComponent (TopComponent tc) {
        return addTopComponent(tc, null);
    }

    /** Adds given top component to this container with specified constraints.
    * Beware, the code must be written to ensure that top component
    * will receive addNotify() before componentActivated(). 
    *
    * @param tc top component to add
    * @param constr constraints
    * @return false if top component already exixts in container or 
    * cannot be added due to container-dependent restrictions,
    * true otherwise
    */
    public boolean addTopComponent (TopComponent tc, Object constr) {
        // check if we can add or not
        if (!canAdd(new TopComponent[] {tc} )) {
            return false;
        }
        // select only if top component is already contained
        if (topComponents().contains(tc)) {
            setSelectedTopComponent(tc);
            return false;
        }
        topComponents().add(tc);
        if (constr == null)
            componentAdded(tc);
        else
            componentAdded(tc, constr);
        if (isAttached()) {
            doActivateComponent(tc);
        }
        // start to listen to top component's property changes
        wm.findManager(tc).addPropertyChangeListener(this);
        return true;
    }

    /** Removes specified top component from this container.
    * Beware! The code in this method must be written to ensure
    * that each top component will receive componentDeactivated() before
    * removeNotify().
    *
    * @param tc top component to remove
    * @return the count of remaining contained components
    */
    public final int removeTopComponent (TopComponent tc) {
        ArrayList tcs = topComponents();
        if (!tcs.contains(tc)) {
            return tcs.size();
        }
        // stop to listen to the component
        wm.findManager(tc).removePropertyChangeListener(this);
        topComponents().remove(tc);
        int count = tcs.size();
        // update selected if needed
        boolean selectedChanged = false;
        if (tc.equals(selected)) {
            selected = (count == 0) ? null : (TopComponent)tcs.get(0);
            selectedChanged = true;
        }
        // deactivate if needed
        if (tc.equals(TopComponent.getRegistry().getActivated())) {
            doActivateComponent(null);
        }
        // let subclasses do visual update job
        componentRemoved(tc);
        if (isAttached() && (selectedChanged)) {
            reactivateSelectedComponent();
        }
        return count;
    }

    /** @return true if container contains specified top component,
    * false otherwise
    */
    public boolean containsTopComponent (TopComponent tc) {
        return topComponents().contains(tc);
    }

    /** @return Currently selected top component or null if no top
    * component is selected at a time */ 
    public TopComponent getSelectedTopComponent () {
        return selected;
    }

    /** The component requests focus. Move it to be visible.
    * When in iconified state, deiconify first.
    */
    public void requestFocus (TopComponent tc) {
        setSelectedTopComponent(tc);
    }
    
    /** Default container itself plays a role of content pane */
    public Container getContentPane () {
        return this;
    }
    
    /** Reactions to the property changes of contained top components
    * and property changes of the mode.
    */
    public void propertyChange (PropertyChangeEvent ev) {
        final String propName = ev.getPropertyName();
        final Object source = ev.getSource();
        
        DoPropertyChange doPropChange = new DoPropertyChange
        (propName, source, ev);
        
        if (EventQueue.isDispatchThread()) {
            //We are in AWT thread, update GUI directly
            doPropChange.run();
        } else {
            //We are NOT in AWT thread, update GUI later in AWT thread
            EventQueue.invokeLater(doPropChange);
        }
    }
    
    /** Implementation of the ActionPerformer interface
    * Performs differently depending on given action.
    */
    public void performAction (final SystemAction action) {
        if (SwingUtilities.isEventDispatchThread()) {
            doPerformAction(action);
        } else {
            SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    doPerformAction(action);
                }
            });
        }
    }

    /** performs proper action */
    private void doPerformAction (SystemAction action) {
        TopComponent selected = getSelectedTopComponent();
        if (selected == null) {
            return;
        }
        if (action instanceof CloseViewAction) {
            // close view action performing
            selected.close();
            actionPerformed(action, true);
        } else if (action instanceof CloneViewAction || 
                 action instanceof UndockAction) {
            // prepare new single mode
            WorkspaceImpl curWorkspace = (WorkspaceImpl)wm.getCurrentWorkspace();
            String modeName = wm.findUnusedModeName(selected.getName(), curWorkspace);
            Mode singleMode = curWorkspace.createMode(
                                  modeName, modeName, null, true
                              );
            if (action instanceof CloneViewAction) {
                // clone the view and dock into new single mode
                TopComponent newComp =
                    (TopComponent)((TopComponent.Cloneable)selected).cloneComponent();
                newComp.setIcon(selected.getIcon());
                singleMode.dockInto(newComp);
                newComp.open();
            } else {
                // undock to the single mode
                singleMode.dockInto(selected);
            }
            actionPerformed(action, true);
        } else {
            actionPerformed(action, false);
        }
    }
    
    /** Method for subclasses, returns true if container is in
     * attached state.
     */
    protected boolean isAttached () {
        return attached;
    }
    
    public void addNotify () {
        super.addNotify();
        //Bugfix #10619, commented redundant call of setAttached
        //setAttached(true);
    }
    
    public void removeNotify () {
        super.removeNotify();
        //Bugfix #10619, commented redundant call of setAttached
        //setAttached(false);
    }
    
    /** Sets visibility status for container
    * @param visible true if container should be made visible,
    * false otherwise
    */
    public synchronized void setAttached (boolean attached) {
        // do nothing if desired state is already reached 
        if (this.attached == attached) {
            return;
        }
        
        this.attached = attached;
        //PinRepaintRM rman = (PinRepaintRM) javax.swing.RepaintManager.currentManager(null);
        try {
            //rman.setRegisteringAllowed(false);
            if (attached) {
                attachComponents();
                mode.updateTitle();
            } else {
                detachComponents();
            }
        } finally {
            //rman.setRegisteringAllowed(true);
        }
    }
    
    /** Adds specified listener to property changes of this container */
    public void addPropertyChangeListener (PropertyChangeListener listener) {
        if (propChangeSupport == null) {
            propChangeSupport = new PropertyChangeSupport(this);
        }
        propChangeSupport.addPropertyChangeListener(listener);
    }
    
    /** Removes specified listener for property changes of this container */
    public void removePropertyChangeListener (PropertyChangeListener listener) {
        if (propChangeSupport != null) {
            propChangeSupport.removePropertyChangeListener(listener);
        }
    }

    /** Fires property change. Expected to be called by subclasses to notify
     * about changes */
    protected void firePropertyChange (String propertyName, Object oldValue, Object newValue) {
        if (propChangeSupport != null) {
            propChangeSupport.firePropertyChange(propertyName, oldValue, newValue);
        }
    }
    
    /** Fires property change. Expected to be called by subclasses to notify
     * about changes */
    protected void firePropertyChange (PropertyChangeEvent evt) {
        if (propChangeSupport != null) {
            propChangeSupport.firePropertyChange(evt);
        }
    }

    /** Safe getter for top components list data */
    protected ArrayList topComponents () {
        if (topComps == null) {
            topComps = new ArrayList(10);
        }
        return topComps;
    }
    
    /** Activates component and updates selected nodes.
    * If param is null, clears both active top component and
    * selected nodes.
    */
    protected void doActivateComponent (TopComponent tc) {
        //Fixed bug #9932
        //Do not call wm.activateComponent(tc) if container does not 
        //have focus.
        if (hasFocus) {
            wm.activateComponent(tc);
        }
        
        // selected nodes
        // don't change sel nodes if activated top component
        // is property sheet
        if (!(tc instanceof NbNodeOperation.Sheet)) {
            RegistryImpl rimpl = (RegistryImpl) TopComponent.getRegistry();
            Node[] nodes = (tc == null) ? new Node[0] : tc.getActivatedNodes();
            rimpl.selectedNodesChanged(new SelectedNodesChangedEvent(this, tc, nodes));
        }
    }
    
    /** Register as performer for actions when get focus. */
    void registerPerformers () {
        // next and previous actions
        updateNextAndPreviousState();
        if (topComponents().size() <= 0) {
            return;
        }
        // close action
        closeView.setActionPerformer(this);
        // undock action
        TopComponent selected = getSelectedTopComponent();
        if (selected != null) {
            // obtain mode active top component is in
            Workspace curWorkspace = wm.getCurrentWorkspace();
            Mode compMode = curWorkspace.findMode(selected);
            // activate undock action if component is in multi
            if ((compMode != null) && (!((ModeImpl)compMode).isSingle()))
                undock.setActionPerformer(this);
        }
        // clone action
        if (selected instanceof TopComponent.Cloneable)
            clone.setActionPerformer(this);
    }

    /** Enable or disable next and previous actions depending on
    * current inner state */
    void updateNextAndPreviousState () {
        if (topComponents().size() > 1) {
            nextTab.setActionPerformer(this);
            prevTab.setActionPerformer(this);
        } else {
            nextTab.setActionPerformer(null);
            prevTab.setActionPerformer(null);
        }
    }
    
    /** Setter for selected top component - only for subclasses. */
    protected final void setSelectedTopComponent (TopComponent tc) {
        if (selected == tc) {
            return;
        }
        selected = tc;
        if ((tc != null) && !containsTopComponent(tc)) {
            throw new IllegalArgumentException(
                "Cannot set " + tc.getName() + " as selected. " +
                "It's not present in the container."
            );
        }
        if (isAttached()) {
            mode.updateTitle();
            componentSelected(selected);
        }
    }


    /** Creates and returns popup menu for given top component,
     * filled with standard action presenters for window actions */
    protected static JPopupMenu createPopupMenu (TopComponent tc) {
        JPopupMenu popup = new JPopupMenuPlus();
        // constructs pop-up menu from actions of selected component
        SystemAction[] compActions = tc.getSystemActions();
        for (int i = 0; i < compActions.length; i++) {
            if (compActions[i] == null)
                popup.addSeparator();
            else if (compActions[i] instanceof CallableSystemAction)
                popup.add(((CallableSystemAction)compActions[i]).
                          getPopupPresenter());
            //add FileSystemAction to pop-up menu
            else if (compActions[i] instanceof FileSystemAction)
                popup.add(((FileSystemAction)compActions[i]).
                          getPopupPresenter());
        }
        if (compActions.length != 0) {
            popup.addSeparator();
        }
        // and add our docking action
        popup.add(docking.getPopupPresenter());
        return popup;
    }

    /** Shows given popup on given coordinations and takes care about the
     * situation when menu can exceed screen limits */
    protected static void showPopupMenu (JPopupMenu popup, Point p, Component comp) {
        SwingUtilities.convertPointToScreen (p, comp);
        Dimension popupSize = popup.getPreferredSize ();
        Dimension screenSize = Toolkit.getDefaultToolkit ().getScreenSize ();
        int yCorrection = 0;
        if (org.openide.util.Utilities.isWindows ()) {
            yCorrection = org.netbeans.core.output.OutputTab.TYPICAL_WINDOWS_TASKBAR_HEIGHT;
        }
        if (p.x + popupSize.width > screenSize.width) {
            p.x = screenSize.width - popupSize.width;
        }
        if (p.y + popupSize.height > screenSize.height - yCorrection) {
            p.y = screenSize.height - popupSize.height - yCorrection;
        }
        SwingUtilities.convertPointFromScreen (p, comp);
        popup.show(comp, p.x, p.y);
    }

    protected boolean isActivated () {
        return hasFocus;
    }
    
    /** Reactivate currently selected top component, if possible */
    protected void reactivateSelectedComponent () {
        if (isActivated()) {
            TopComponent selected = getSelectedTopComponent();
            doActivateComponent(selected);
            // bugfix #5467
            registerPerformers();
            if (selected != null) {
                selected.requestFocus();
            }
        }
    }

    /** @return true if new top component can be added, false otherwise */
    public abstract boolean canAdd (TopComponent[] tc);
    
    /** @return Constraints in which given top components lives in this
     * container. Null is valid and means that no constraints for specified
     * top component exist  */
    public abstract Object getConstraints(TopComponent tc);
    
    protected abstract void attachComponents ();
    
    protected abstract void detachComponents ();

    /** Updates visual state of container after given top component has beeen
     * added to this container.
     * Beware, the code must be written to ensure that top component
     * will receive addNotify() before componentActivated(). 
     * @param tc top component tha was added
    */
    protected abstract void componentAdded (TopComponent tc);

    /** Updates visual state of container after given top component has beeen
     * added to this container with constraints.
     * Beware, the code must be written to ensure that top component
     * will receive addNotify() before componentActivated(). 
     * @param tc top component tha was added
     * @param constr constraints
    */
    protected abstract void componentAdded (TopComponent tc, Object constr);
    
    protected abstract void componentRemoved (TopComponent tc);
    
    protected abstract void componentSelected (TopComponent tc);
    
    /** Called when the name of some component in this multi frame has changed */
    protected abstract void componentNameChanged (TopComponent tc);
    
    /** Called when icon of some component in this multi frame has changed */
    protected abstract void componentIconChanged (TopComponent component, Image icon);
    
    /** Called when some of standard windows actions like close, clone, next tab
     * was invoked. Boolean flag is true when action was already processed
     * by this base abstract class.
     */
    protected abstract void actionPerformed (SystemAction action, boolean processed);
    
    
/******* implementation of listeners **********/    

    public void componentShown(java.awt.event.ComponentEvent p1) {
        //Bugfix #11153, setAttached now called from ModeImpl.setVisible
        //setAttached(true);
    }
    
    public void componentResized(java.awt.event.ComponentEvent p1) {
    }
    
    public void componentHidden(java.awt.event.ComponentEvent p1) {
        //Bugfix #11153, setAttached now called from ModeImpl.setVisible
        //setAttached(false);
    }
    
    public void componentMoved(java.awt.event.ComponentEvent p1) {
    }
     
    /** When frame is being deactivated  */
    public void frameDeactivated(FrameTypeEvent fe) {
        hasFocus = false;
    }    
    
    /** When frame was closed  */
    public void frameClosed(FrameTypeEvent fe) {
    }
    
    /** When frame was brought back from icon to normal state  */
    public void frameDeiconified(FrameTypeEvent fe) {
    }
    
    /** When frame was brought back from maximize to normal state  */
    public void frameNormalized(FrameTypeEvent fe) {
    }
    
    /** When frame was opened - showed  */
    public void frameOpened(FrameTypeEvent fe) {
    }
    
    /** When frame was iconified  */
    public void frameIconified(FrameTypeEvent fe) {
    }
    
    /** When user tried to invoke frame closing  */
    public void frameClosing(FrameTypeEvent fe) {
    }
    
    /** When frame was made active - receives focus  */
    public void frameActivated(FrameTypeEvent fe) {
        hasFocus = true;
        reactivateWithDelay();
    }
    
    /** When frame was made active - receives focus  */
    public void frameMaximized(FrameTypeEvent fe) {
    }

    /** Run reactivateSelectedComponent with some delay, if activation
     * delay property is set */
    protected final void reactivateWithDelay () {
        if (waitBeforeActivateComponent < 0) {
            // try to read delay property, default is 100 to solve some issues
            // with SplitContainerImpl - PENDING, we should change it
            // to be more robust 
            String str = System.getProperty("netbeans.frames.activate.delay", "100");
            try {
                waitBeforeActivateComponent = Integer.parseInt(str);
                if (waitBeforeActivateComponent < 0)
                    waitBeforeActivateComponent = 0;
            }
            catch (NumberFormatException nfe) {
                waitBeforeActivateComponent = 0;
                System.err.println("** BAD NUMBER FORMAT: netbeans.frames.activate.delay = " + str); // XXX
            }
        }
        if (waitBeforeActivateComponent > 0) {
            // construct delayed reactivation call
            pendingNodeActivator = new Runnable() {
                public void run() {
                    if (isActivated()) {
                        reactivateSelectedComponent();
                    }
                    // request processsed, do cleanup
                    pendingNodeActivator = null;
                    System.getProperties().remove("hack.pendingNodeActivator");
                }
            };
            // mark pending delayed call and post request
            System.getProperties().put("hack.pendingNodeActivator", pendingNodeActivator);
            org.openide.util.RequestProcessor.postRequest(new Runnable() {
                public void run() {
                    // must run in AWT
                    SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            if (pendingNodeActivator != null) {
                                pendingNodeActivator.run();
                            }
                        }
                    });
                }
            }, waitBeforeActivateComponent, Thread.MAX_PRIORITY);
        } else {
            // delayed activation not desired, run now
            reactivateSelectedComponent();
        }
    }

    /** force running of pending activation request */
    static void runPendingNodeActivator () {
        if (pendingNodeActivator != null) {
            pendingNodeActivator.run();
        }
    }

    /** Used to do UI change in AWT thread, when propertyChange is not called
     * from AWT thread.
     * Called from DefaultContainerImpl.propertyChange
     */
    private final class DoPropertyChange extends Object implements Runnable {
        private String propName;
        private Object source;
        private PropertyChangeEvent ev;
        
        DoPropertyChange (String propName, Object source, PropertyChangeEvent ev) {
            this.propName = propName;
            this.source = source;
            this.ev = ev;
        }
        
        public void run () {
            // handle tcmanager's changes
            if (source instanceof WindowManagerImpl.TopComponentManager) {
                // notification from top component manager
                TopComponent tc = ((WindowManagerImpl.TopComponentManager)source).getComponent();
                if (WindowManagerImpl.TopComponentManager.PROP_NAME.equals(propName)) {
                    if (isAttached()) {
                        componentNameChanged(tc);
                        //Added to update window title
                        mode.updateTitle();
                    }
                } else if (WindowManagerImpl.TopComponentManager.PROP_ICON.equals(propName)) {
                    if (isAttached()) {
                        componentIconChanged(tc, (Image)ev.getNewValue());
                    }
                }
                // forward event to our listeners too
                firePropertyChange(ev);
                return;
            }
            // handle mode changes
            if (source instanceof ModeImpl) {
                if (ModeImpl.PROP_FRAME_INSTANCE.equals(propName)) {
                    changeFrame((FrameType)ev.getOldValue(), (FrameType)ev.getNewValue());
                }
            }
        }
    }
/************** serialization - deserialization mechanisms ***********/    
    
    /** Asociates given data with this container. Data are used during
     * validation process. Must be called prior to validateData method.
     */
    protected final void attachData (SerializedData serData) {
        this.serData = serData;
    }

    /** Called when first phase of WS deserialization is done.
     * Finish deserialization of all contained top components
     * and attach listeners properly.
     * Caution: This method should be called only during deserialization
     * (or when container was not displayed yet)
    */
    public void validateData () {
        validateData(serData);
    }

    /** Should be called only from MultiTabbedContainerImpl externally, to deal
     * with old serialization protocol.
     */
    private void validateData (SerializedData serData) {
        // restore connection with our mode
        if (serData == null) {
            //Do not perform data validation when serData is null
            return;
        }
                    
        ModeImpl ourMode = (ModeImpl)TopManager.getDefault().getWindowManager().
                            findWorkspace(serData.workspaceName).findMode(serData.modeName);
        if (ourMode == null) {
            throw new InternalError("Mode " + serData.modeName + " cannot be found!"); //NOI18N
        }
        // listen to our mode
        attachToMode(ourMode);
        // obtain top component instances from
        // instances of top component managers
        topComps = new ArrayList(serData.tcManagers.size());
        WindowManagerImpl.TopComponentManager cur = null;
        TopComponent current = null;
        for (Iterator iter = serData.tcManagers.iterator(); iter.hasNext(); ) {
            cur = (WindowManagerImpl.TopComponentManager)iter.next();
            /*if (cur == null) {
                System.out.println("THE PROBLEM....manager is null, mode "
                                    + ourMode.getDisplayName() + " workspace "
                                    + TopManager.getDefault().getWindowManager().
                                      findWorkspace(serData.workspaceName).getDisplayName()
                                    );
            }*/
            if ((cur != null) && cur.validateData()) {
                topComps.add(cur.getComponent());
                // set current if possible
                if (cur.equals(serData.currentTcm)) {
                    current = cur.getComponent();
                }
            }
        }
        // update current if needed
        int count = topComps.size();
        if ((current == null) && (count > 0)) {
            current = (TopComponent)topComps.get(0);
        }
        setSelectedTopComponent(current);
        // listen to top components
        TopComponent curTc = null;
        for (Iterator iter = topComps.iterator(); iter.hasNext(); ) {
            curTc = (TopComponent)iter.next();
            wm.findManager(curTc).addPropertyChangeListener(this);
        }
        // clear serialized data
        serData = null;
    }
    
   
    /** Data class for storing persistent information about DefaultContainerImpl.
     * Remember that deserialization is multiple-phase, information from this
     * class are processed in validateData() methods and only after then
     * container instance is fully deserialized and initialized
     */
    protected static final class SerializedData implements Serializable {
        /** Unique ID */
        static final long serialVersionUID =2108673887921434234L;
        /** Description of serialized fields */
        public static final String MODE_NAME = "modeName";
        public static final String WORKSPACE_NAME = "workspaceName";
        public static final String MANAGERS_LIST = "tcManagers";
        public static final String CURRENT_MANAGER = "currentTcm";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(MODE_NAME, String.class),
            new ObjectStreamField(WORKSPACE_NAME, String.class),
            new ObjectStreamField(MANAGERS_LIST, ArrayList.class),
            new ObjectStreamField(CURRENT_MANAGER, WindowManagerImpl.TopComponentManager.class)
        };
        
        /** name of the containing mode */
        String modeName;
        /** name of the containing workspace */
        String workspaceName;
        /** array of managers of top components */
        ArrayList tcManagers;
        /** manager of last selected top component (can be null) */
        WindowManagerImpl.TopComponentManager currentTcm;
        /** Source for data that are to be serialized */
        private DefaultContainerImpl source;
        
        /** Default constructor, used when writing data, asociates
         * with given container */
        protected SerializedData (DefaultContainerImpl source) {
            this(source, null, null, null, null);
        }

        /** Used when serialized data are obtained externally, not through
         * this class itself - for example from older deserialization format */
        protected SerializedData (DefaultContainerImpl source, String modeName,
                                  String workspaceName, ArrayList tcManagers,
                                  WindowManagerImpl.TopComponentManager currentTcm) {
            this.source = source;
            this.modeName = modeName;
            this.workspaceName = workspaceName;
            this.tcManagers = tcManagers;
            this.currentTcm = currentTcm;
        }
        
        /** Serialization of all workspaces */
        private void writeObject (ObjectOutputStream oos)
                     throws IOException {
            ObjectOutputStream.PutField pf = oos.putFields();
            // write workspace and mode name
            pf.put(WORKSPACE_NAME, source.mode.getWorkspace().getName());
            pf.put(MODE_NAME, source.mode.getName());
            // create and write list of tc managers
            TopComponent[] tcs = source.getTopComponents();
            ArrayList tcManagers = new ArrayList(source.topComps.size());
            for (int i = 0; i < tcs.length; i++) {
                /*if (WindowManagerImpl.findManager(tcs[i]) == null) {
                    System.out.println("MANAGER IS NULL FOR TC " + tcs[i].getName());
                    System.out.println("Mode " + source.mode.getDisplayName() +
                        " workspace " + source.mode.getWorkspace().getDisplayName()
                    );
                }*/
                tcManagers.add(WindowManagerImpl.findManager(tcs[i]));
            }
            pf.put(MANAGERS_LIST, tcManagers);
            // write manager of selected tc
            WindowManagerImpl.TopComponentManager selectedTcm = 
                (source.selected == null) ? null : WindowManagerImpl.findManager(source.selected);
            pf.put(CURRENT_MANAGER, selectedTcm);
            oos.writeFields();
        }

        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField gf = ois.readFields();
            modeName = (String)gf.get(MODE_NAME, null);
            workspaceName = (String)gf.get(WORKSPACE_NAME, null);
            tcManagers = (ArrayList)gf.get(MANAGERS_LIST, null);
            currentTcm = (WindowManagerImpl.TopComponentManager)gf.get(CURRENT_MANAGER, null);
        }
        
    } // end of SerializedData inner class
    
}

