/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.util.*;
import java.beans.PropertyVetoException;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

/**
 * This class is a specialization of the PerimeterPane that sets the 
 * PerimeterLayout.CENTER component to a JDesktopPane for implementation
 * of an MDI desktop panel with desktop frames.
 * 
 * @see PerimeterPane
 *
 * @author  Hippy
 */
public class DesktopPane extends PerimeterPane implements SwingConstants {
    // data members
        
    // the JDesktopPane for InternalFrames
    private JDesktopPane desktopPane = new JDesktopPane();
        
    // component listener for each internal frame for sizing viewport of
    // desktop
    private DesktopPane.DesktopScrollbarListener desktopScrollbarListener = 
        new DesktopPane.DesktopScrollbarListener();
    
    // component listener for each internal frame for sizing viewport of
    // desktop
    private DesktopPane.MinMaxListener minMaxListener = 
        new DesktopPane.MinMaxListener();
    
    // activation listener for all FrameTypes
    private DesktopPane.DeactivationListener deactivationListener = 
        new DesktopPane.DeactivationListener();    
    
    // scroll pane for the desktop (adding desktop, of course)
    private JScrollPane desktopScrollPane = 
        new JScrollPane(desktopPane);
    
    // maximized mode on/off
    private boolean maxMode;
    
    // <JInternalFrame, Boolean> mapping which holds info about iconified frames
    private WeakHashMap frames2States;

    // holds reference to currently containing window 
    // needed for proper automatic refocus behaviour
    // PENDING - WeakReference should be used
    private Window topLevelWindow;
    
    /** 
     * Creates new DesktopPane.
     */
    public DesktopPane() {
        super();
        
        setBorder(BorderFactory.createEmptyBorder());
        
        // set up the desktopPane
        
        //Make dragging faster
        desktopPane.putClientProperty("JDesktopPane.dragMode", "outline");
                
        // add complete our scrollbar listener for making sure scrollbars are visible
        // ONLY when an internal frame is outside the viewport view        
        desktopPane.addContainerListener(desktopScrollbarListener);
                
        // set up the desktopPane's scroll pane
        desktopScrollPane.getVerticalScrollBar().addAdjustmentListener(desktopScrollbarListener);
        desktopScrollPane.getHorizontalScrollBar().addAdjustmentListener(desktopScrollbarListener);
                       
        // add the desktopScrollPane to its proper place
        super.addImpl(desktopScrollPane, PerimeterLayout.CENTER, 0);
        
            
        // reattach Ctrl-F4 action on desktop pane to our control 
        desktopPane.registerKeyboardAction(
            new CloseViewActionRunner(),
            KeyStroke.getKeyStroke(KeyEvent.VK_F4, KeyEvent.CTRL_MASK),
            WHEN_ANCESTOR_OF_FOCUSED_COMPONENT
        );
    }
    
    /** add a hierarchy listener to add the deactivation listener to the top level
    * container */
    public void addNotify () {
        super.addNotify();
        Window curTopLevel = SwingUtilities.windowForComponent(this);
        if (curTopLevel != topLevelWindow) {
            curTopLevel.addWindowListener(deactivationListener);
            if (topLevelWindow != null) {
                topLevelWindow.removeWindowListener(deactivationListener);
            }
            topLevelWindow = curTopLevel;
        }
    }
    
    public void removeNotify () {
        super.removeNotify();
        if (topLevelWindow != null) {
            topLevelWindow.removeWindowListener(deactivationListener);
            topLevelWindow = null;
        }
    }
    
    /**
     * Overrides <code>Component#addImpl</code> to ensure the center 
     * <code>JDesktopPane</code> cannot be replaced without throwing an exception.
     * 
     */
    protected void addImpl(Component comp, Object constraints, int index) {
        if (constraints == PerimeterLayout.CENTER) {
            throw new IllegalArgumentException("Cannot replace desktop.");
        }        
        super.addImpl(comp, constraints, index + 1);
    }
        
    /** 
     *  Adds the specified FrameType to the central JDesktopPane.
     *
     * @param frame   The frame to be added.
     * 
     */
    public void addFrameType(FrameType frame) {
        // add the component listener for scrollbars
        Container cont = frame.getRootPaneContainer();
        cont.addComponentListener(desktopScrollbarListener);
        
        // add frame for min/max listener
        frame.addFrameTypeListener(minMaxListener);
        if (frame.getIcon() != null) {
            frame.getIcon().addComponentListener(desktopScrollbarListener);
        }
        
        // add deactivation frame listener
        frame.addFrameTypeListener(deactivationListener);
        
        // add the frame
        desktopPane.add(cont);
    }
        
    /** 
     *  Adds the specified FrameType with the specified constraints, in order to support
     *  desktop frames addition.
     * 
     * @param frame   The frame to be added.
     * @param constraints   The constraints associated with the added frame.
     * 
     * @see PerimeterLayout#CENTER
     * @see PerimeterLayout#NORTH
     * @see PerimeterLayout#SOUTH
     * @see PerimeterLayout#EAST
     * @see PerimeterLayout#WEST
     * 
     */
    public void addFrameType(FrameType frame, Object constraints) {
        if (constraints == PerimeterLayout.CENTER) {
            addFrameType(frame);
        }
        else {        
            // add deactivation frame listener
            frame.addFrameTypeListener(deactivationListener);
            
            // add the goofy frame
            add(frame.getRootPaneContainer(), constraints);
        }
    }
    
    /** 
     * Removes the specified  FrameType from the DesktopPane.
     * 
     * @param frame   The frame to be removed.
     * 
     */
    public void removeFrameType(FrameType frame) {
        Container cont = frame.getRootPaneContainer();
        desktopPane.remove(cont);
        if (frame.getIcon() != null) {
            frame.getIcon().removeComponentListener(desktopScrollbarListener);
        }
        cont.removeComponentListener(desktopScrollbarListener);
        frame.removeFrameTypeListener(minMaxListener);
        frame.removeFrameTypeListener(deactivationListener);
    }

    /** 
     * Remove the component from the DesktopPane at the specified index.
     * 
     * @param index   The index of the component to be removed.
     * 
     */
    public void remove(int index) {
        Component comp = getComponent(index);
        if (comp != desktopScrollPane) {
            super.remove(index);
        }
    }

    /** 
     * Remove the component from the DesktopPane.
     * 
     * @param comp   The component to be removed.
     * 
     */
    public void remove(Component comp) {
        if (comp != desktopScrollPane) {
            super.remove(comp);
        }
    }

    /** 
     * Remove all the components from the DesktopPane.
     * 
     */
    public void removeAll() {
        // remove all from the desktop
        desktopPane.removeAll();
        
        // remove all from the the DesktopPane (DesktopScrollPane can't be 
        // removed() due to remove override above...
        super.removeAll();
    }

    /** 
     * Convienience method to the frame to the front of the hierarchy.
     * 
     * @param frame   The frame to be moved tot the front.
     * 
     */
    public void moveToFront(FrameType frame) {
        Container cont = frame.getRootPaneContainer();
        if (desktopPane.isAncestorOf(frame.getRootPaneContainer())) {
            desktopPane.moveToFront(frame.getRootPaneContainer());
            try {
                ((JInternalFrame)frame).setSelected(true);
            }
            catch(PropertyVetoException pve) {
                pve.printStackTrace();
            }
        }
        else {
            super.moveToFront(cont);
        }
        frame.getRootPaneContainer().requestFocus();        
    }

    public void setAutoFocusEnabled (boolean autoFocus) {
        deactivationListener.setAutoFocusEnabled(autoFocus);
    }
    
    public boolean isAutoFocusEnabled () {
        return deactivationListener.isAutoFocusEnabled();
    }
    
    /** Enables or disables maximum mode for desktop pane. When in maximum mode,
     * all contained frames are maximized */
    private void setMaxMode (boolean maxMode) {
        if (this.maxMode == maxMode) {
            return;
        }
        this.maxMode = maxMode;
        // don't allow reactions to event firings
        desktopScrollbarListener.passivate();
        
        JInternalFrame[] frames = desktopPane.getAllFrames();
        if (maxMode) {
            frames2States = new WeakHashMap(frames.length * 2);
        }
        for (int i = 0; i < frames.length; i++) {
            setMaxModeForFrame(frames[i], maxMode);
        }
        if (!maxMode) {
            // make gc' able
            frames2States = null;
        }
        desktopScrollbarListener.activate();
    }
    
    private boolean isMaxMode () {
        return maxMode;
    }
    
    /** Helper method, puts given frame into max mode or changes it back,
     * depending on given flag. Map frames2States must not be null */
    private void setMaxModeForFrame (JInternalFrame frame, boolean enable) {
        if (enable) {
            // switch to max mode
            Rectangle visibleRect = desktopScrollPane.getViewport().getViewRect();
            try {
                if (frame.isIcon()) {
                    frames2States.put(frame, Boolean.TRUE);
                    //frame.setIcon(false);
                }
                frame.setMaximum(true);
                frame.setBounds(visibleRect);
            }
            catch(PropertyVetoException pve) {
                // PENDING - don't know what I'm supposed to do..log it at least
                pve.printStackTrace();
            }
        } else {
            // restore regular mode
            try {
                frame.setMaximum(false);
                if (frames2States.get(frame) != null) {
                    frame.setIcon(true);
                }
            }
            catch(PropertyVetoException pve) {
                // PENDING - don't know what I'm supposed to do..log it at least
                pve.printStackTrace();
            }
        }
    }
    
    /** Dumb helper class which only task is to simulate running of 
     * close view action like if it is invoked from kayboard shortcut.
     * Needed as replacement of standard swing Ctrl-F4 behaviour
     */
    private static final class CloseViewActionRunner implements ActionListener {
        public void actionPerformed(ActionEvent evt) {
            KeyStroke ctrlF4 = KeyStroke.getKeyStroke(KeyEvent.VK_F4, KeyEvent.CTRL_MASK);
            ActionEvent aev = new ActionEvent(
                evt.getSource(), ActionEvent.ACTION_PERFORMED,
                org.openide.util.Utilities.keyToString (ctrlF4)
            );
            NbFocusManager.processKeyStroke(ctrlF4, aev);
        }
    } // end of CloseViewActionRunner
    
    
    private static class DeactivationListener extends FrameTypeAdapter 
                                              implements WindowListener {
        
        // keep track of last activated frame type
        private FrameType activated;
        // flag that enables or disables reactions on window events
        private boolean autoFocusEnabled = true;
        
        public void setAutoFocusEnabled (boolean autoFocusEnabled) {
            if (this.autoFocusEnabled == autoFocusEnabled) {
                return;
            }
            this.autoFocusEnabled = autoFocusEnabled;
        }
        
        public boolean isAutoFocusEnabled () {
            return autoFocusEnabled;
        }
        
        // when frame or top level parent is deactivated
        public void frameActivated(final FrameTypeEvent fte) {
            FrameType frame = (FrameType)fte.getSource();
            if (activated != null && frame != activated) {
                try {
                    activated.setSelected(false);
                }
                catch(PropertyVetoException pve) {
                    pve.printStackTrace();
                }
            }
            activated = frame;
        }
        
        public void windowDeactivated(final java.awt.event.WindowEvent p1) {
            if (autoFocusEnabled && (activated != null)) {
                try {
                    activated.setSelected(false);
                }
                catch(PropertyVetoException pve) {
                    pve.printStackTrace();
                }
            }
        }
        
        public void windowActivated(final java.awt.event.WindowEvent p1) {
            if (autoFocusEnabled && (activated != null)) {
                try {
                    activated.setSelected(true);
                }
                catch(PropertyVetoException pve) {
                    pve.printStackTrace();
                }
            }
        }
        
        public void windowClosed(final java.awt.event.WindowEvent p1) {
        }
        
        public void windowDeiconified(final java.awt.event.WindowEvent p1) {
        }
        
        public void windowOpened(final java.awt.event.WindowEvent p1) {
        }
        
        public void windowIconified(final java.awt.event.WindowEvent p1) {
        }
        
        public void windowClosing(final java.awt.event.WindowEvent p1) {
        }
    }
    
    // for setting the scrollpane viewport size
    private class DesktopScrollbarListener
        implements ComponentListener, ContainerListener, AdjustmentListener {
            
        // flag tha aloows to refuse recursive updating of desktop pane
        private boolean nowUpdating = false;

        public void componentShown(final java.awt.event.ComponentEvent p1) {
            checkDesktopSize();
        }
        
        public void componentResized(final java.awt.event.ComponentEvent p1) {
            checkDesktopSize();
        }
        
        public void componentHidden(final java.awt.event.ComponentEvent p1) {
            checkDesktopSize();
        }
        
        public void componentMoved(final java.awt.event.ComponentEvent ce) {
            checkDesktopSize();
        }
        
        public void componentRemoved(final java.awt.event.ContainerEvent p1) {
            checkDesktopSize();
        }
        
        public void componentAdded(final java.awt.event.ContainerEvent ce) {
            if (isMaxMode()) {
                if (ce.getChild() instanceof FrameType) {
                    try {
                        ((FrameType)ce.getChild()).setState(FrameType.MAXIMIZED);
                    }
                    catch(PropertyVetoException pve) {
                        pve.printStackTrace();
                    }
                }
            }
            checkDesktopSize();
        }
        
        public void adjustmentValueChanged(AdjustmentEvent evt) {
            // we are adjusting
            checkDesktopSize();
        }        

        // sets the size of the desktopPane, and adjusts children and viewport
        // accordingly
        private void checkDesktopSize () {
            // refuse cyclic calls
            if (nowUpdating) {
                return;
            }
            nowUpdating = true;
            // get an easier (and still proper) way of referring to goofy 
            // JDesktopPane and JScrollPane
            JDesktopPane desktopPane = DesktopPane.this.desktopPane;
            JScrollPane scrollPane = DesktopPane.this.desktopScrollPane;
            
            // get the viewRect
            Rectangle minRect = scrollPane.getViewport().getViewRect();
            // save old view pos
            Point oldViewPos = scrollPane.getViewport().getViewPosition();
            
            // get the correct size of the desktoppane based on frame/icon pos
            JInternalFrame[] frames = desktopPane.getAllFrames();
            if (!isMaxMode()) {            
                for (int i = 0; i < frames.length; i++) {
                    Rectangle compRect = frames[i].isIcon() ?
                        frames[i].getDesktopIcon().getBounds() :
                        frames[i].getBounds();
                    minRect.add(compRect);
                }

                /****** dafe - commented out, caused problems during scrolls
                 * to maximum value with both scrollbars enabled */
                // resize the desktop and position view correctly
                // ensure that scroll bars are accounted for
                /*minRect.height += 
                    scrollPane.getHorizontalScrollBar().isVisible() ?
                    scrollPane.getHorizontalScrollBar().getSize().height : 0;
                minRect.width += 
                    scrollPane.getVerticalScrollBar().isVisible() ?
                    scrollPane.getVerticalScrollBar().getSize().width : 0;*/

                // check if size or positioning are off now, do something about it
                boolean sizeChanged = !minRect.getSize().equals(desktopPane.getSize());
                boolean viewPosChanged = (minRect.x != 0 || minRect.y != 0) &&
                    (oldViewPos.x != - minRect.x && oldViewPos.y != - minRect.y);
                if (sizeChanged || viewPosChanged) {
                    resizeDesktop(minRect, oldViewPos);
                }
            }
            else {            
                // ensure maximized frame are set to proper size
                for (int i = 0; i < frames.length; i++) {
                    frames[i].setBounds(scrollPane.getViewport().getViewRect());
                }
                resizeDesktop(minRect, oldViewPos);
            }
            // reset flag that we are updating, but don't let waiting jobs in
            // AWT to cause resizing again
            SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    nowUpdating = false;
                }
            });
        }
            
        // sets the size of the desktopPane, adjust children and viewport
        // positions accordingly
        private void resizeDesktop(Rectangle newDRect, Point oldViewPos) {
            // get an easier way of referring to goofy JDesktopPane, 
            // JScrollPane, and JViewport
            JDesktopPane desktopPane = DesktopPane.this.desktopPane;
            JScrollPane scrollPane = DesktopPane.this.desktopScrollPane;
            
            passivate();
            
            JViewport viewport = scrollPane.getViewport();
            
            // set the scroll pane invisible for cleaner drawing
            scrollPane.setVisible(false);
            
            // move all the internal frames for proper placement
            JInternalFrame[] frames = desktopPane.getAllFrames();
            JInternalFrame curFrame = null;
            Component curComp = null;
            boolean isIcon = false;
            for (int i = 0; i < frames.length; i++) {
                curFrame = frames[i];
                isIcon = curFrame.isIcon();
                curComp = isIcon ? 
                    (Component)curFrame.getDesktopIcon() :
                    (Component)curFrame;
                Point loc = curComp.getLocation();
                loc.x -= newDRect.x; 
                loc.y -= newDRect.y;
                // workaround for BasicInternalFrameUI trying to position icon
                // bugfix #9166
                if (isIcon) {
                    loc.y -= newDRect.height - desktopPane.getSize().height;
                }
                curComp.setLocation(loc);
            }
            
            // set the desktop size
            desktopPane.setPreferredSize(newDRect.getSize());
            desktopPane.setSize(newDRect.getSize());
            desktopPane.revalidate();

            // set view's position
            if ((newDRect.x != 0) || (newDRect.y != 0)) {
                oldViewPos.translate(-newDRect.x, -newDRect.y);
                viewport.setViewPosition(oldViewPos);
            }
            
            determineScrollBarPolicy(newDRect);                     
            scrollPane.doLayout();
            scrollPane.revalidate();
    
            scrollPane.setVisible(true);
            
            activate();
        }
        
        private void determineScrollBarPolicy(Rectangle newDRect) {
            // get an easier (and still proper) way of referring to goofy 
            // JScrollPane
            JScrollPane scrollPane = DesktopPane.this.desktopScrollPane;   
            
            // get the size of the desktop pane and the scrollpane         
            Dimension sSize = scrollPane.getSize();
            
            // determine the scrollbar policies of the scroll pane
            int vPolicy = newDRect.height > sSize.height ?
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS : 
                JScrollPane.VERTICAL_SCROLLBAR_NEVER;
            boolean vEnabled = 
                    vPolicy == JScrollPane.VERTICAL_SCROLLBAR_ALWAYS ?
                    true : false;
            
            int hPolicy = newDRect.width > sSize.width ?
                JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS : 
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
            boolean hEnabled = 
                    hPolicy == JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS ?
                    true : false;
            
            // set the scrollbar policies of the scroll pane
            scrollPane.setVerticalScrollBarPolicy(vPolicy);
            scrollPane.setHorizontalScrollBarPolicy(hPolicy);
            scrollPane.getVerticalScrollBar().setEnabled(vEnabled);
            scrollPane.getHorizontalScrollBar().setEnabled(hEnabled);
        }
        
        private void passivate () {
            JDesktopPane desktopPane = DesktopPane.this.desktopPane;
            JScrollPane scrollPane = DesktopPane.this.desktopScrollPane;
    
            desktopPane.removeContainerListener(this);
            scrollPane.getVerticalScrollBar().removeAdjustmentListener(this);
            scrollPane.getHorizontalScrollBar().removeAdjustmentListener(this);   
            
            JInternalFrame[] frames = desktopPane.getAllFrames();
            for (int i = 0; i < frames.length; i++) {
                frames[i].getDesktopIcon().removeComponentListener(this);
                frames[i].removeComponentListener(this);
            }
        }
        
        private void activate () {
            JDesktopPane desktopPane = DesktopPane.this.desktopPane;
            JScrollPane scrollPane = DesktopPane.this.desktopScrollPane;
            
            desktopPane.addContainerListener(this);
            scrollPane.getVerticalScrollBar().addAdjustmentListener(this);
            scrollPane.getHorizontalScrollBar().addAdjustmentListener(this);
            
            JInternalFrame[] frames = desktopPane.getAllFrames();
            for (int i = 0; i < frames.length; i++) {
                frames[i].getDesktopIcon().addComponentListener(this);
                frames[i].addComponentListener(this);
            }
        }
        
    }
    
    private class MinMaxListener extends FrameTypeAdapter {
        // are the internal frames maximized?
        private boolean maximized = false;
        
        public void frameClosed(FrameTypeEvent fe) {
            if (DesktopPane.this.desktopPane.getAllFrames().length == 0) {
                setMaxMode(false);
            }
        }
                
        /** When frame was iconified  */
        public void frameIconified(FrameTypeEvent fe) {
            InternalFrameTypeImpl frame = (InternalFrameTypeImpl)fe.getSource();
            setMaxMode(false);
            
            Rectangle minRect = desktopScrollPane.getViewport().getViewRect();
            
            Component icon = frame.getDesktopIcon();
            Rectangle iconRect = icon.getBounds();
            // lay icon down and from left
            iconRect.setLocation(minRect.x, 
                                 minRect.y + minRect.height - iconRect.height);
            // move icon if place is already occupied by another icon                     
            Component comp = null;
            do {
                comp = desktopPane.getComponentAt(iconRect.getLocation());
                // take only icons except me
                comp = (comp instanceof RootPaneContainer || comp == icon ||
                        comp == desktopPane) ? null : comp;
                if (comp != null) {
                    iconRect.setLocation(iconRect.x + comp.getBounds().width, 
                                         iconRect.y);
                }
            } while (comp != null);
        }
                
        /** When frame was maximized  */
        public void frameMaximized(FrameTypeEvent fe) {
            if (!isMaxMode()) {
                setMaxMode(true);
                selectFrame((FrameType)fe.getSource());
            }
        }
                
        /** When frame was normalized back to regular state */
        public void frameNormalized(FrameTypeEvent fe) {
            if (isMaxMode()) {
                setMaxMode(false);
                selectFrame((FrameType)fe.getSource());
            }
        }
        
        private void selectFrame(FrameType frame) {
            moveToFront(frame);
            frame.getRootPaneContainer().requestFocus();
        }
        
    } // and of MinMaxListener
    
}
