/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.Component;
import java.awt.Container;
import java.awt.Image;
import java.io.Serializable;

import java.beans.PropertyVetoException;

import javax.swing.RootPaneContainer;
import javax.swing.WindowConstants;

/** 
 * Interface for frame types. It's associated with exactly one
 * mode. Implementors of this interface should manage the task of containing
 * multiple top components, switching between them etc.
 * Implementors should also listen to the property changes of the contained
 * top components and the mode which they represent.
 *
 * @author Dafe Simonek, Evan 'Hippy' Slatis
 * 
 */
public interface FrameType extends RootPaneContainer, WindowConstants, Serializable {
    
    /** 
     * Represents NORMAL state a frame can be in. 
     */
    public static final int NORMAL = 0;

    /** 
     * Represents ICONIFIED state a frame can be in. 
     */
    public static final int ICONIFIED = 1;
    
    /** 
     * Represents MAXIMIZED state a frame can be in. 
     */
    public static final int MAXIMIZED = 2;

    /** 
     * Represents INTERNAL frame type, one that floats inside a parent frame. 
     */
    public static final int INTERNAL = 0;
    
    /** 
     * Represents DESKTOP frame type, one that docks to the side of a container. 
     */
    public static final int DESKTOP = 1;
    
    /** 
     * Represents TOP_LEVEL frame type, one that floats on the OS desktop. 
     */
    public static final int TOP_LEVEL = 2;
    
    /** 
     * Represents DIALOG frame type, a specialized, one time use, TOP_LEVEL frame. 
     */
    public static final int DIALOG = 4;
    
    /** 
     * Represents OTHER frame type, a sick and twisted creation that no one can figure
     * out, like Hippy. 
     */
    public static final int OTHER = 8;
    
    /** 
     * Method for retrieving the Container that is the frame (for example, the
     * JInternalFrame of an INTERNAL frame type).
     *
     * @return    Returns the container implementing the FrameType.
     */
    public Container getRootPaneContainer();
    
    /** 
     * Sets the state of container - normal, iconified, maximized.
     * 
     *  @param state new state of the comtainer (constants from FrameType.xxx)
     * 
     */
    public void setState (int state) throws PropertyVetoException;

    /** 
     * Retrieve the state of the FrameType.
     * 
     * @return current state (normal x iconified) of the container.
     *
     */
    public int getState ();
    
    /** 
     * Set whether the frame is selected or not.
     *
     * @param selected  True if the FrameType is selected.
     * 
     * @throws PropertyVetoException  If the FrameType annot be selected.
     * 
     */
    public void setSelected(boolean selected) throws PropertyVetoException;
    
    /** 
     * Gets the default close operation, defined by WindowConstants.
     *
     * @returns The default close operation in WindowConstants.
     * 
     */
    public int getDefaultCloseOperation();
    
    /** 
     * Sets the default close operation, defined by WindowConstants.
     *
     * @param operation   The default close operation in WindowConstants.
     * 
     */
    public void setDefaultCloseOperation(int operation);
    
    /** 
     * Request for UI update of look anf feel.     * 
     */
    public void updateUI ();
    
    /** 
     * Closes the container and frees system resources.
     */
    public void dispose ();

    /** 
     * Gets the frame's title.
     *
     * @returns The frames title.
     * 
     */
    public String getTitle();

    /** 
     * Sets the frame's title.
     *
     * @param title   The frames title.
     * 
     */
    public void setTitle(String title);
    
    /** 
     * Gets the frame's icon image, null if no icon available. This can mean 
     * either the desktop, frame, both, or neither depending on 
     * implementation.
     *
     * @returns The frame's icon's image.
     * 
     */
    public Image getIconImage();
    
    /** 
     * Sets the frame's icon image.  This can mean either the desktop,
     * frame, both, or neither depending on implementation.
     *
     * @param icon The frame's icon's image.
     * 
     */
    public void setIconImage(Image icon);
    
    /** 
     * Gets the frame's desktop icon, if it exists.
     *
     * @returns The frame's icon component.
     * 
     */
    public Component getIcon();

    /** 
     * Get the type of frame. 
     *
     * @return  The type of the FrameType, such as FrameType#Internal.
     */
    public int getType();
    
    /** 
     * Adds given frame listener for listening to the frame events.
     *
     * @param fl   The listener to be added.
     * 
     */
    public void addFrameTypeListener(FrameTypeListener fl);
    
    /** 
     * Removes given frame listener.
     *
     * @param fl   The listener to be removed.
     * 
     */
    public void removeFrameTypeListener(FrameTypeListener fl);

}
