/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;


import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.LayoutManager;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.*;
import java.beans.PropertyVetoException;
import java.util.HashMap;
import java.util.Collection;
import java.util.Iterator;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JRootPane;
import javax.swing.event.InternalFrameEvent;

import javax.swing.plaf.InternalFrameUI;
import javax.swing.plaf.basic.BasicInternalFrameUI;

/**
 * This is the FrameType implementation of an internal frame, one that resides
 * inside a desktop.
 * 
 * @see FrameType
 *
 * @author  Hippy, Dafe
 * 
 */
public class InternalFrameTypeImpl extends JInternalFrame implements FrameType {
    
    // Keep track of frame state.
    private int state = FrameType.NORMAL;
    
    // FrameType event
    private FrameTypeEvent frameEvent = null;
    
    // keep track of FrameListeners...
    HashMap flMap = new HashMap();
        
    /** 
     * Creates a non-resizable, non-closable, non-maximizable, non-iconifiable 
     * JInternalFrame with no title.
     */
    public InternalFrameTypeImpl() {
        this("");
    }
    
    /** 
     * Creates a non-resizable, non-closable, non-maximizable, non-iconifiable 
     * JInternalFrame with the specified title. 
     * 
     * @param title  The title of the frame.
     * 
     */
    public InternalFrameTypeImpl(String title) {
        this(title, false);
    }
    
    /** 
     * Creates a non-closable, non-maximizable, non-iconifiable JInternalFrame 
     * with the specified title and with resizability specified.
     * 
     * @param title  The title of the frame.
     * @param resizable  True if the frame is resizable.
     * 
     */
    public InternalFrameTypeImpl(String title, boolean resizable) {
        this(title, resizable, false);
    }
           
    /** 
     * Creates a non-maximizable, non-iconifiable JInternalFrame with the 
     * pecified title and with resizability and closability specified.
     * 
     * @param title  The title of the frame.
     * @param resizable  True if the frame is resizable.
     * @param closable   True if the frame is closable.
     *
     */
    public InternalFrameTypeImpl(String title,boolean resizable, boolean closable) {
        this(title, resizable, closable, false);
    }
           
    /** 
     * Creates a non-iconifiable JInternalFrame with the specified title and 
     * with resizability, closability, and maximizability specified.
     * 
     * @param title  The title of the frame.
     * @param resizable  True if the frame is resizable.
     * @param closable   True if the frame is closable.
     * @param maximizable   True if the frame is maximizable.
     *
     */
    public InternalFrameTypeImpl(String title,boolean resizable,boolean closable,boolean maximizable) {
        this(title, resizable, closable, maximizable, false);
    }
           
    /**
     * Creates a JInternalFrame with the specified title and with 
     * resizability, closability, maximizability, and iconifiability specified. 
     * 
     * @param title  The title of the frame.
     * @param resizable  True if the frame is resizable.
     * @param closable   True if the frame is closable.
     * @param iconifiable   True if the frame is iconifiable.
     * @param
     * 
     */
    public InternalFrameTypeImpl(String title,boolean resizable,boolean closable,boolean maximizable,boolean iconifiable) {
        super(title, resizable, closable, maximizable, iconifiable);
        
        initialize();
    }

    /** Allows subclasses to override initialization */
    protected void initialize () {
        // PENDING - should be deleted
        //replaceLayout();
        getRootPane().getGlassPane().setVisible(true);
        revalidate();
    }
    
    // PENDING - should be deleted when new approach will be approved 
    /*private void replaceLayout () {
        // now replace the layout manager (God help us...)
        setRootPaneCheckingEnabled(false);
        setLayout(new MDIFrameTypeLayoutManager());
        setRootPaneCheckingEnabled(true);
    }*/
    
    /** 
     * Get the Container containing the RootPane.
     *
     * @return The container actually representing FrameType. Normallly it is
     * instance of JInternalFrame, but if iconified, it's desktop icon instance.
     * 
     */
    public Container getRootPaneContainer() {
        Container result = (getState() == FrameType.ICONIFIED)
                    ? (Container)getDesktopIcon()
                    : (Container)this;
        return result;
    }
    
    /** 
     * Sets the state of container - normal, iconified, maximized.
     * 
     * @param state   New state of the comtainer (constants from FRAME.xxx).
     * 
     */
    public void setState(int state) throws PropertyVetoException {
        switch (state) {
            case FrameType.ICONIFIED:
                super.setIcon(true);
                break;
            case FrameType.NORMAL:
                if (isIcon()) {
                    super.setIcon(false);
                } else if (isMaximum()) {
                    super.setMaximum(false);
                }
                break;
            case FrameType.MAXIMIZED:
                super.setMaximum(true);
                break;
            default:
                throw new IllegalArgumentException("Illegal state");
        }
        this.state = state;
    }
    
    /** Overrides superclass version to keep our state synchronized with
     * superclass' iconification state */
    public void setIcon (boolean b) throws PropertyVetoException {
        super.setIcon(b);
        this.state = b ? FrameType.ICONIFIED : FrameType.NORMAL;
    }
    
    /** Overriden to ensure proper firing of FrameTypeEvents events and
     * to keep our state property in sync.
     * 
     * @param b  True if maximized.
     */
    public void setMaximum (boolean b) throws PropertyVetoException {
        super.setMaximum(b);
        this.state = b ? FrameType.MAXIMIZED : FrameType.NORMAL;
        Collection ifas = flMap.values();
        Iterator ifaIter = ifas.iterator();
        
        while (ifaIter.hasNext()) {
            InternalFrameListenerAdapter ifa = 
                (InternalFrameListenerAdapter)ifaIter.next();
            if (true == b) {
                ifa.internalFrameMaximized();
            }
            else {
                ifa.internalFrameNormalized();
            }
        }
    }
    
    /** 
     * 
     * Return the current state of the frame.
     * 
     * @return   Current state (normal, iconified, or maximized) of the container.
     * 
     */
    public int getState() {
        return state;
    }
    
    /** 
     *
     * Add a frame listener.
     *
     * @param fl  The frame listener to add.
     * 
     */
    public synchronized void addFrameTypeListener(FrameTypeListener fl) {
        InternalFrameListenerAdapter ifa = new InternalFrameListenerAdapter(this, fl);
        addInternalFrameListener(ifa);
        flMap.put(fl, ifa);
    }
    
    /**
     * 
     * Remove a frame listener.
     *
     * @param fl  The frame listener to remove.
     *
     */
    public synchronized void removeFrameTypeListener(FrameTypeListener fl) {
        InternalFrameListenerAdapter ifa = 
            (InternalFrameListenerAdapter)flMap.get(fl);
        if (ifa != null) {
            removeInternalFrameListener(ifa);
        }
    }
    
    /** 
     * 
     * Get the type of frame.
     *
     * @return FrameType#INTERNAL
     * 
     */
    public int getType() {
        return FrameType.INTERNAL;
    }
    
    /** 
     *
     * Gets the frame's icon image, null if no icon available.
     * 
     * @return  The image of the icon.
     *
     */
    public Image getIconImage() {
        if (getFrameIcon() instanceof ImageIcon) {
            return ((ImageIcon)getFrameIcon()).getImage();
        }
        return null;
    }
    
    /** 
     * Set's the frame's icon image, if available.
     *
     * @param icon  The image of the icon.
     */
    public void setIconImage(Image icon) {
        setFrameIcon(new ImageIcon(icon));
    }
        
    /**
     *
     * Get's the frame's desktop icon.
     *
     * @return The component representing the minimized desktop icon.
     * 
     */
    public Component getIcon() {
        return getDesktopIcon();
    }    
    
    /** Resize title bar to smaller height, which is better acceptable default
     */
    public void setUI(InternalFrameUI ui) {
        super.setUI(ui);
        // PENDING - should be available as option
        setCaptionHeight(20);
    }
    
    /** Sets new height of the title bar of this frame.
     * @return true if new height was set succesfuly, false if height
     * cannot be changed
     */
    public boolean setCaptionHeight (int height) {
        // works only for UIs derived from BasicInternalFrameUI
        InternalFrameUI ui = getUI();
        if (!(ui instanceof BasicInternalFrameUI)) {
            return false;
        }
        // title bar needn't be here, we have to be prepared
        JComponent north = ((BasicInternalFrameUI)ui).getNorthPane();
        if (north == null) {
            return false;
        }
        Dimension oldPref = north.getPreferredSize();
        oldPref.height = height;
        north.setPreferredSize(oldPref);
        return true;
    }
    
    /* not needed now, will be deleted when tehre wil be no problems with new
     approach 

    // for resizing the title bar to something a bit more managemable
    private class MDIFrameTypeLayoutManager extends Object implements LayoutManager {

        private static final int TITLE_BAR_SIZE = 19;

        // Creates new ChildFrameTypeLayoutManager
        public MDIFrameTypeLayoutManager() {
        }

        public java.awt.Dimension preferredLayoutSize(final Container cont) {
            return cont.getSize();
        }

        public void removeLayoutComponent(final Component comp) {
        }
        
        public java.awt.Dimension minimumLayoutSize(final Container cont) {
            //return cont.getMinimumSize();
            return new Dimension(20, 20);
        }

        public void addLayoutComponent(String str,final Component comp) {
        }

        public void layoutContainer(final Container cont) {
            Component[] comps = cont.getComponents();
            JRootPane rootPane = (JRootPane) ((comps[0] instanceof JRootPane) ?
                comps[0] : comps[1]);
            JComponent comp = (JComponent) (!(comps[0] instanceof JRootPane) ?
                comps[0] : comps[1]);

            // get the cont size - insets
            Rectangle contRect = cont.getBounds();
            Insets insets = cont.getInsets();
            contRect.x = insets.left;
            contRect.width -= insets.left + insets.right;
            contRect.y = insets.top;
            contRect.height -= insets.top + insets.bottom;

            Rectangle compRect = comp.getBounds();
            compRect.width = contRect.width;
            compRect.height = TITLE_BAR_SIZE;
            compRect.setLocation(contRect.getLocation());
            comp.setBounds(compRect);
            comp.setPreferredSize(comp.getSize());

            Rectangle rootRect = rootPane.getBounds();
            rootRect.width = contRect.width;
            rootRect.height = contRect.height - TITLE_BAR_SIZE;
            rootRect.setLocation(contRect.x, contRect.y + TITLE_BAR_SIZE);
            rootPane.setBounds(rootRect);
        }
    } */
}
