/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.*;
import java.awt.event.*;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.*;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.*;
import javax.swing.text.Keymap;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.TabbedPaneUI;
import javax.swing.SwingConstants;
import javax.swing.plaf.basic.BasicTabbedPaneUI;
import javax.swing.FocusManager;

import org.openide.loaders.DataObject;
import org.openide.awt.UndoRedo;
import org.openide.awt.MouseUtils;
import org.openide.TopManager;
import org.openide.actions.*;
import org.openide.windows.*;
import org.openide.util.NbBundle;
import org.openide.util.WeakListener;
import org.openide.util.Task;
import org.openide.util.HelpCtx;
import org.openide.util.RequestProcessor;
import org.openide.util.actions.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.io.SafeException;
import org.openide.nodes.Node;
import org.openide.nodes.NodeListener;
import org.openide.nodes.NodeAdapter;

import org.netbeans.core.actions.DockingAction;
import org.netbeans.core.windows.util.*;
import org.netbeans.core.windows.*;
import org.netbeans.core.ui.ArrowsTabbedPaneUI;
import org.netbeans.core.ui.PopupTabbedPaneUI;

/** Implementation of multi-tabbed top component container.
* The window should be invisible if there is no component in it.
*/
public final class MultiTabbedContainerImpl extends DefaultContainerImpl implements ChangeListener {

    /** Time delta between two successive calls to name change method */
    static final int NAME_CHANGE_DELTA = 1000;
    /* Task for renaming */
    //private RequestProcessor.Task nameTask;
    /* Rename performer */
    //private NameChanger nameChanger;
    /* The time of last call to name change method */
    //private long nameChangeTime = -1;
    /** JTabbedPane for this frame */
    private JTabbedPane tabc;
    /** helper variable - stores selected tab in tabbed pane */
    private TopComponent wasSelected;
    /** listener to the mode which owns this container */
    private PropertyChangeListener weakModeL;
    /** Tab placement of the tabs */
    private int tabPlacement;
    /** Maximal allowed count of contained top components */
    private int maxCount;

    /** Internal list of deserialized tc managers. This list is used
    * in validateData() method, where list of deserialized top components
    * is finnaly obtained */
    private ArrayList tcManagers;
    /** Holds the manager of selected top component,
    * used only during deserialization */
    private WindowManagerImpl.TopComponentManager currentTcm;
    /** Holds mode and workspace name during deserialization.
    * Used to restore association with the mode after deserialization */
    private String modeName;
    private String workspaceName;

    /** Internal status in which currently this container resides */
    private int innerStatus;
    /** inner status constants */
    private static final int EMPTY = 0;
    private static final int SINGLE = 1;
    private static final int MULTI = 2;
    
    /** manager of versioned serialization - for old deserialization only*/
    private static VersionSerializator serializationManager;

    /** Default TabbedPaneUI of JTabbedPane component
     */
    private BasicTabbedPaneUI basictabbedpaneui;

    /** Creates new multiFrame */
    public MultiTabbedContainerImpl () {
        this(null);
    }
    
    /** Constructor for use during deserialization, asociates data
     * with container */
    private MultiTabbedContainerImpl (DefaultContainerImpl.SerializedData serData) {
        super(serData);
        innerStatus = EMPTY;
        maxCount = 0;
        initGui();
        initListeners();
        NbFocusManager.install ();
    }

    /** initializes the gui of this container */
    private void initGui () {
        tabc = new JTabbedPane();
        basictabbedpaneui = (BasicTabbedPaneUI)tabc.getUI();
        changeUI();
	if (! Boolean.getBoolean ("netbeans.windows.tabs_on_top"))
	    tabc.setTabPlacement (SwingConstants.BOTTOM);
        setLayout(new BorderLayout());
    }

    /* performs needed initialization */
    private void initListeners () {
        // listen to various things happening in tabbed pane
        tabc.addChangeListener(this);
        tabc.addMouseListener(new PopupMouseImpl());
    }

    protected void attachComponents () {
        // attach our top components to the AWT's hierarchy
        if (innerStatus == MULTI) {
            // add all contained top components
            final TopComponent[] curComps = getTopComponents();
            Image icon = null;
            TopComponent curComp = null;
            for (int i = 0; i < curComps.length; i++) {
                curComp = curComps[i];
                icon = curComp.getIcon();
                tabc.addTab(
                    curComp.getName(),
                    (icon == null) ? null : new ImageIcon(icon),
                    curComp, curComp.getToolTipText()
                );
            }
            if (wasSelected != null) {
                // restore selected tab and ensure that it is visible
                final TopComponent toBeSelected = wasSelected;
                // select the component a little bit later,
                // when tabs are all added and initialized
                SwingUtilities.invokeLater(new Runnable() {
                    public void run () {
                        TopComponent localToBeSelected = toBeSelected;
                        // beware, previously selected component may become
                        // invalid, select last one then
                        if (!containsTopComponent(localToBeSelected)) {
                            localToBeSelected = curComps[curComps.length - 1];
                        }
                        if (!localToBeSelected.isVisible()) {
                            localToBeSelected.setVisible(true);
                        }
                        setSelectedTopComponent(localToBeSelected);
                    }
                });
            }
            if (getComponentCount() <= 0) {
                add(tabc);
            }
            tabc.validate();
        } else if (innerStatus == SINGLE) {
            //if (getComponentCount() <= 0) {
                TopComponent selected = getSelectedTopComponent();
                add(selected);
                // ensure that component is visible
                if (!selected.isVisible()) {
                    selected.setVisible(true);
                }
                invalidate();
                validate();
            //}
        }
    }
    
    protected void detachComponents () {
        // free our top components from AWT's component hierarchy
        // to be available for other workspaces
        if (innerStatus == MULTI) {
            wasSelected = (TopComponent)tabc.getSelectedComponent();
            tabc.removeAll();
        } else if (innerStatus == SINGLE) {
            remove(getSelectedTopComponent());
        }
    }
    
    
    /** Updates visual state of container after given top component has been
     * added to this container. Ignores constraints by calling componentAdded(TC).
     * @param tc top component that was added
     * @param constr constraints of TC
    */
    public void componentAdded (TopComponent tc, Object constr) {
        componentAdded(tc);
    }

    /** Updates visual state of container after given top component has been
     * added to this container.
     * @param tc top component tha was added
    */
    public void componentAdded (TopComponent tc) {
        Image icon = tc.getIcon();
        TopComponent current = getSelectedTopComponent();
        // perform adding
        synchronized (this) {
            switch (innerStatus) {
                // first component added -> act as single
            case EMPTY:
                innerStatus = SINGLE;
                setSelectedTopComponent(tc);
                if (isAttached()) {
                    add(tc);
                }
                break;
                // second added -> turn to multi
            case SINGLE:
                innerStatus = MULTI;
                // deactivate first, if needed
                if (current.equals(TopComponent.getRegistry().getActivated())) {
                    doActivateComponent(null);
                }
                removeAll();
                add(tabc, BorderLayout.CENTER);
                // add first top component
                Image currIcon = current.getIcon();
                if (isAttached()) {
                    tabc.addTab(current.getName(),
                                currIcon == null ? null : new ImageIcon(currIcon),
                                current, current.getToolTipText());
                }
                // BEWARE, WITHOUT BREAK!
            case MULTI:
                if (isAttached()) {
                    tabc.addTab(tc.getName(),
                                icon == null ? null : new ImageIcon(icon),
                                tc, tc.getToolTipText());
                }
                break;
            }
            refreshAfterAdding(tc);
        }
        // start to listen to top component's property changes
        //Commented, not necessary superclass DefaultContainerImpl
        //is already listening to component.
        //wm.findManager(tc).addPropertyChangeListener(this);
    }
    
    /** Updates visual state after top component was removed from this container.
    * @param tc top component to remove
    */
    public void componentRemoved (TopComponent tc) {
        int count = -1;
        synchronized (this) {
            count = topComponents().size();
            switch (innerStatus) {
                // the only component removed -> turn into EMPTY
            case SINGLE:
                innerStatus = EMPTY;
                doActivateComponent(null);
                remove(tc);
                break;
            case MULTI:
                if (count == 1) {
                    // only one component left -> turn to single
                    innerStatus = SINGLE;
                    //Bugfix #10265 eliminate call of stateChanged
                    tabc.removeChangeListener(this);
                    tabc.removeAll();
                    tabc.addChangeListener(this);
                    remove(tabc);
                    if (isAttached()) {
                        final TopComponent singleTc = (TopComponent)topComponents().get(0);
                        // hack - workaround to JTabbedPaneUI bug, which hides
                        // our top component, so we have to reset it again 
                        SwingUtilities.invokeLater(new Runnable() {
                            public void run () {
                                singleTc.setVisible(true);
                                //09 Feb 2001 by Marek Slama
                                //Bugfix #9521, update window title when last but one
                                //top component is removed.
                                mode.updateTitle();
                                //End
                            }
                        });
                        add(singleTc);
                        setSelectedTopComponent(singleTc);
                        reactivateSelectedComponent();
                    }
                } else {
                    int index = tabc.indexOfComponent(tc);
                    if (index != -1) {
                        //Bugfix #10265 eliminate call of stateChanged
                        tabc.removeChangeListener(this);
                        tabc.removeTabAt(index);
                        tabc.addChangeListener(this);
                        // synchronize current variable with tab selection
                        setSelectedTopComponent((TopComponent)tabc.getSelectedComponent());
                        reactivateSelectedComponent();
                    }
                    // invoke relayout of multi tab
                    tabc.invalidate ();
                    tabc.validate ();
                    tabc.repaint();
                }
                break;
            }
        }
    }
    
    

    /** @return True if count of contained top components is
    * smaller than maxCount property */
    public boolean canAdd (TopComponent[] tc) {
        int newSize = topComponents().size() + ((tc == null) ? 0 : tc.length);
        return (maxCount <= 0) || (newSize <= maxCount);
    }

    /** Constraints not supported in this container implementation */
    public Object getConstraints(TopComponent tc) {
        return null;
    }

    /** Sets new maximal count of top components that can be
    * contained in this container */
    public void setMaxCount (int maxCount) {
        this.maxCount = maxCount;
    }

    /** @return Maximal allowed count of contained top components */
    public int getMaxCount () {
        return maxCount;
    }

    /** Called when the name of some component in this multi frame has changed */
    protected void componentNameChanged (TopComponent component) {
        if (!containsTopComponent(component))
            return;  // not added yet
        String newName = component.getName();
        if (innerStatus == MULTI) {
           // update tabbed pane
           int compIndex = tabc.indexOfComponent(component);
           if (compIndex < 0)
               return;  // removed or somewhat..
           tabc.setTitleAt(compIndex, newName);
           tabc.invalidate();
           tabc.validate();
           tabc.repaint();
        }
       // notify name change listeners
       fireNameChange(newName);
       
       /////////////////////////////////////////////////////////////////////
       //Commented 05 Feb 2001 by Marek Slama
       //Delaying is not used anymore.
       //Replanning to AWT thread is done
       //in DefaultContainerImpl.propertyChange.
       
        /*long now = System.currentTimeMillis();
        if ((nameChangeTime < 0) ||
                ((now - nameChangeTime) > NAME_CHANGE_DELTA)) {
            new NameChanger(this, component).run();
        } else {
            // run the request for the change of the name of tab
            if ((nameTask == null) || (nameChanger.component != component)) {
                nameTask = RequestProcessor.createRequest(
                               nameChanger = new NameChanger(this, component));
            }
            nameTask.schedule(1500);
        }
        nameChangeTime = now;*/
       //End of commented code
    }

    /** Called when icon of some component in this multi frame has changed */
    protected void componentIconChanged (TopComponent component, Image icon) {
        if (!containsTopComponent(component))
            return;  // not added yet
        if (innerStatus == MULTI) {
            int compIndex = tabc.indexOfComponent(component);
            if (compIndex < 0)
                return;
            tabc.setIconAt(compIndex, new ImageIcon(icon));
            if (isAttached()) {
                tabc.revalidate();
                tabc.repaint();
            }
        }
    }
    
    /** Called when some of standard windows actions like close, clone, next tab
     * was invoked. Boolean flag is true when action was already processed
     * by this base abstract class.
     */
    protected void actionPerformed (SystemAction action, boolean processed) {
        if (processed) {
            return;
        }
        if (action instanceof NextTabAction || action instanceof PreviousTabAction) {
            // select previous or next component
            int count = tabc.getTabCount();
            int index = tabc.getSelectedIndex();
            if ((count <= 1) || (index == -1)) {
                return;
            }
            int delta = (action instanceof NextTabAction) ? 1 : -1;
            int newIndex = (index + delta + count) % count;
            tabc.setSelectedIndex(newIndex);
        }
    }
    

    /** Post-initialization after adding - called from addTopComponent method */
    protected void refreshAfterAdding (TopComponent tc) {
        // workaround for JTabbedPane bug, make visible if needed
        if (isAttached() && innerStatus == MULTI) {
            componentSelected(tc);
            if (!tc.isVisible())
                tc.setVisible(true);
            tabc.invalidate();
            tabc.validate();
            tabc.repaint();
        }
    }

    /** Selects specified top component in tabbed pane */
    protected void componentSelected (TopComponent tc) {
        if ((innerStatus == MULTI) && (tabc.indexOfComponent(tc) >= 0)) {
            tabc.setSelectedComponent(tc);
        }
    }
    
    /** @return context help for the frame. */
    public HelpCtx getHelpCtx () {
        return new HelpCtx(MultiTabbedContainerImpl.class);
    }

    /** Implementation of the ChangeListener interface
    * Called when some tab is the tabbed pane is activated.
    */
    public void stateChanged (ChangeEvent e) {
        if ((!isVisible()) || (!isShowing()) || (tabc.getTabCount() <= 0) ||
                (innerStatus != MULTI))
            return;
        int index = tabc.getSelectedIndex();
        TopComponent current = (index < 0) ? null : (TopComponent)tabc.getSelectedComponent();
        setSelectedTopComponent(current);
        reactivateSelectedComponent();
    }

    /** Sets the placement of the tabs.
    * @param placement SwingConstant.XXX constant representing new
    * placement of the tabs (TOP, LEFT...)
    */
    public void setTabPlacement (int tabPlacement) {
        if (this.tabPlacement == tabPlacement)
            return;
        this.tabPlacement = tabPlacement;
        tabc.setTabPlacement(tabPlacement);
        // PENDING - need to repaint?
    }

    /** @return current placement of the tab */
    public int getTabPlacement () {
        return tabPlacement;
    }

    /** Called when first phase of WS deserialization is done.
    * Finish deserialization process by doing proper initialization
     * Caution: This method should be called only during deserialization
     * (or when container was not displayed yet) 
    */
    public void validateData () {
        // validate superclass too (data was asociated with superclass using
        // attachData(data) call)
        super.validateData();
        // update inner status
        int count = topComponents().size();
        if (count <= 0) {
            innerStatus = EMPTY;
        } else if (count == 1) {
            innerStatus = SINGLE;
        } else {
            innerStatus = MULTI;
        }
        // selected field synchronization
        wasSelected = getSelectedTopComponent();
    }

    /** Change TabbedPaneUI for JTabbedPane component
     */
    public void changeUI () {
        switch(WindowManagerImpl.getDefault().tabbedContainerUIManager().getTabbedContainerUI()) {
            case TabbedContainerUIManager.MODE_BASIC:
                tabc.setUI(basictabbedpaneui/*new BasicTabbedPaneUI()*/);
                break;
            case TabbedContainerUIManager.MODE_ARROWS:
                tabc.setUI(new ArrowsTabbedPaneUI());
                break;
            case TabbedContainerUIManager.MODE_POPUP:
                tabc.setUI(new PopupTabbedPaneUI());
                break;
        }
    }

    /** Only accessor for inner class */
    private void fireNameChange (final String newName) {
        firePropertyChange("name", null, newName); // NOI18N
    }

    /** Popup menu reaction implementation */
    final class PopupMouseImpl extends MouseUtils.PopupMouseAdapter {
        /** Called when the seqeunce of mouse events should lead to actual
        * showing of the popup menu. */
        protected void showPopup (MouseEvent e) {
            TabbedPaneUI tabUI = tabc.getUI();
            int clickTab = tabUI.tabForCoordinate(tabc, e.getX(), e.getY());
            TopComponent selected = getSelectedTopComponent();
            if ((selected == null) ||
                    (!selected.equals(TopComponent.getRegistry().getActivated()))) {
                return;
            }
            if (clickTab != -1) {
                //Click is on valid tab, not on empty area in tab
                showPopupMenu(createPopupMenu(selected), e.getPoint(), tabc);
            }
        }
    } // end of PopupMouseImpl

    /////////////////////////////////////////////////////////////////////
    // Commented 05 Feb 2001 by Marek Slama. Not used anymore.
    // Code to update GUI is replanned to AWt thread 
    // in DefaultContainerImpl.propertyChange. It covers also subclasses.
    
    /* Takes care of lazy renaming of tabs in tabbed panel.
    * Waits until the name stays the same for some period of time,
    * and only then calls rename of the tab.
    */
    /* private static final class NameChanger extends Object implements Runnable { */
        /** Outer class */
        //private MultiTabbedContainerImpl outer;
        /** The component whose name has changed */
        /*TopComponent component;

        NameChanger (MultiTabbedContainerImpl outer, TopComponent component) {
            this.outer = outer;
            this.component = component;
        }

        public synchronized void run () {
            // do name change always in AWT thread
            EventQueue.invokeLater(new Runnable() {
                                       public void run () {
                                           String newName = component.getName();
                                           if (outer.innerStatus == MULTI) {
                                               // update tabbed pane
                                               JTabbedPane tabc = outer.tabc;
                                               int compIndex = tabc.indexOfComponent(component);
                                               if (compIndex < 0)
                                                   return;  // removed or somewhat..
                                               tabc.setTitleAt(compIndex, newName);
                                               tabc.invalidate();
                                               tabc.validate();
                                               tabc.repaint();
                                           }
                                           // notify name change listeners
                                           outer.fireNameChange(newName);
                                       }
                                   });
        }

    } // end of NameChanger inner class */
   //End of commented code

/******************** Serialization management ***************/    
    
    /** Nominates replacer for serialization.
    */
    private Object writeReplace () throws ObjectStreamException {
        return new SerializationReplacer(this);
    }
    
    /** Class that acts as serialization manager for multi tabbed container. */
    private static final class SerializationReplacer implements Serializable {
        /** Unique ID */
        static final long serialVersionUID =2202987849834763838L;
        /** Asociation with container which data needs to be serialized */
        private MultiTabbedContainerImpl source;
        
        public static final String DEFAULT_CONTAINER = "defaultContainer";
        private static final ObjectStreamField[] serialPersistentFields = {
            new ObjectStreamField(DEFAULT_CONTAINER, DefaultContainerImpl.class),
        };
        
        SerializationReplacer (MultiTabbedContainerImpl source) {
            this.source = source;
        }

        /** Serialization of all workspaces */
        private void writeObject (ObjectOutputStream oos)
                     throws IOException {
            ObjectOutputStream.PutField pf = oos.putFields();
            // write workspace and mode name
            pf.put(DEFAULT_CONTAINER, new DefaultContainerImpl.SerializedData(source));
            oos.writeFields();
        }

        /** Deserialization of the workspace */
        private void readObject (ObjectInputStream ois)
                     throws IOException, ClassNotFoundException {
            source = new MultiTabbedContainerImpl();
            ObjectInputStream.GetField gf = ois.readFields();
            source.attachData((DefaultContainerImpl.SerializedData)gf.get(DEFAULT_CONTAINER, null));
        }

        /** Resolves deserialized SerializationReplacer to the singleton
        * instance of WindowManagerImpl */
        private synchronized Object readResolve () 
                                    throws ObjectStreamException {
            return source;                            
        }

    } // end of SerializationReplacer
    
    
/******** Old serialization, must be here for compatibility with 3.1 and older versions */    
    
    /** Accessed from MultiTabContainer$VSAccess, which is remainder from
     * old serialization mechanism */
    public static DefaultReplacer.ResVersionable getOldSerializator () {
        return new Version1();
    }
    
    /** Basic version of persistence for mode implementation.
     * Method assignData(modeImpl) must be called prior to serialization.
     */
    private static final class Version1
        implements DefaultReplacer.ResVersionable {

        /* identification string */
        public static final String NAME = "Version_1.0"; // NOI18N

        /** variables of persistent state of the mode implementation */
        int innerStatus;
        String workspaceName;
        String modeName;
        Rectangle bounds;
        ArrayList tcManagers;
        WindowManagerImpl.TopComponentManager currentTcm;

        /** Identification of the version */
        public String getName () {
            return "Version_1.0"; // NOI18N
        }

        /** read the data of the version from given input */
        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            // read the fields
            innerStatus = ((Integer)in.readObject()).intValue();
            workspaceName = (String)in.readObject();
            modeName = (String)in.readObject();
            bounds = (Rectangle)in.readObject();
            // read managers of top components (but leave top components
            // in byte streams, not fully deserialized).
            // Deserialization of tcs will be finished
            // in validateData() of MultiTabContainer
            tcManagers = (ArrayList)in.readObject();
            // remove tcManagers deserialized to null, because
            // it signalizes that there was a problem during
            // serialization of top component
            WindowManagerImpl.TopComponentManager[] tcmArray =
                (WindowManagerImpl.TopComponentManager[])
                tcManagers.toArray(new WindowManagerImpl.TopComponentManager[tcManagers.size()]);
            for (int i = tcmArray.length - 1; i >= 0; i--) {
                if (tcmArray[i] == null) {
                    tcManagers.remove(i);
                }
            }
            // obtain manager of selected tc
            Object temp = in.readObject();
            currentTcm = (WindowManagerImpl.TopComponentManager)temp;
        }

        /** write the data of the version to given output */
        public void writeData (ObjectOutput out)
        throws IOException {
            // never should be called again
            throw new InternalError();
        }

        public Object resolveData ()
        throws ObjectStreamException {
            MultiTabbedContainerImpl result = new MultiTabbedContainerImpl();
            result.innerStatus = innerStatus;
                        
            DefaultContainerImpl.SerializedData oldSerData = 
                new DefaultContainerImpl.SerializedData(
                    result, modeName, workspaceName, tcManagers, currentTcm
                );
            result.attachData(oldSerData);
            return result;
        }

    } // end of Version1 inner class

}
