/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.Component;

import javax.swing.FocusManager;
import javax.swing.KeyStroke;
import javax.swing.text.Keymap;
import javax.swing.JFrame;
import javax.swing.Action;
import javax.swing.SwingUtilities;
import javax.swing.MenuSelectionManager;
import javax.swing.MenuElement;

import org.openide.TopManager;
import org.openide.util.Utilities;

import org.netbeans.core.windows.WindowManagerImpl;

/** Our focus manager to catch keys.
 * @author Jaroslav Tulach
*/
final class NbFocusManager extends FocusManager {
    /** manager to deletage operations to */
    private FocusManager delegate;

    /** @param d delegate focus manager
    */
    private NbFocusManager(FocusManager d) {
        delegate = d;
    }

    /** Checks whether the calling thread uses this focus manager.
    * If no, installs new one.
    */
    public static void install () {
        FocusManager fm = getCurrentManager ();
        if (fm == null || fm.getClass () != NbFocusManager.class) {
            setCurrentManager (new NbFocusManager (fm));
        }
    }

    /** Processes the key */
    public void processKeyEvent(Component focusedComponent, final KeyEvent anEvent) {
        // force run of any pending activation requests
        DefaultContainerImpl.runPendingNodeActivator();
        if (isTransmodalAction (KeyStroke.getKeyStrokeForEvent (anEvent))) {
            //System.err.println ("Got a transmodal action...");
        } else {
            // Bug #500 fixed here. and #2482 too.
            Component c = focusedComponent;
            Component mw = TopManager.getDefault ().getWindowManager ().getMainWindow ();

            java.awt.Window w = SwingUtilities.windowForComponent (c);
            if (w instanceof java.awt.Dialog) {
                java.awt.Dialog d = (java.awt.Dialog)w;
                if (d.isModal ()) {
                    return;
                }
            }
        }
        /*
                while (true) {
                  System.err.println ("Considering: " + c);
                  if ((c instanceof TopComponent) || (c == mw)) {
                    //System.err.println ("Fine, MW or TC.");
                    break;
                  } else if (c instanceof java.awt.Dialog) {
                    //System.err.println ("Some window, maybe dialog; skip.");
                    return;
                  } else if (c instanceof java.awt.Window) {
                    // other top window, proceed bugfix of #2482
                    break;
                  } else {
                    c = c.getParent ();
                  }
                }
              }
        */
        // delegates to the original focus manager
        delegate.processKeyEvent (focusedComponent, anEvent);
        //System.err.println ("Was consumed by delegate: " + anEvent.isConsumed ());
        if (!anEvent.isConsumed ()) {
            process (anEvent, focusedComponent);
        }
    }

    /** Delegates to old manager */
    public void focusNextComponent(Component aComponent) {
        delegate.focusNextComponent (aComponent);
    }

    /** Delegates to old manager */
    public void focusPreviousComponent(Component aComponent) {
        delegate.focusPreviousComponent (aComponent);
    }

    /** HashSet of posted events. Ensures that no event will be posted
    * twice. (KeyEvent, KeyEvent)
    */
    private static java.util.Set posted = new java.util.HashSet ();

    /** Processes the event by posting it to shortcut processor.
    * @param ev the event to process
    */
    static void process (final KeyEvent ev, final Component comp) {
        // test if ev is not posted
        if (posted.contains (ev)) return;

        // register that this event has been posted
        posted.add (ev);

        // if there is opened menu in the main window, then dispatch the
        // messages to the menu first
        {
            JFrame main = WindowManagerImpl.mainWindow();
            MenuSelectionManager selManager = MenuSelectionManager.defaultManager();
            MenuElement[] elems = selManager.getSelectedPath();
            if ((elems != null) && (elems.length > 0)) {
                for (int i = elems.length; --i >= 0 && !ev.isConsumed(); ) {
                    Component compo = elems[i].getComponent();
                    compo.dispatchEvent(ev);
                }
                //main.dispatchEvent(ev);
                ev.consume();

                // if there is a menu in main window => never go thru the default processing
                return;
            }
        }

        if (ev.getID () != KeyEvent.KEY_PRESSED) {
            posted.remove(ev);
            return;
        }

        // if the event is not consumed, then plans code
        // that will be run after the event is processed by component
        // an will try to find the right shortcut
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                //System.err.println ("Running later; was consumed: " + ev.isConsumed ());
                if (!ev.isConsumed ()) {
                    // process only if the event is not consumed
                    processEvent(ev, comp);
                }
                posted.remove (ev);
            }
        });
    }

    /** Directly processes given key event, event should not be already consumed */
    static void processEvent (KeyEvent ev, Component comp) {
        KeyStroke key = KeyStroke.getKeyStrokeForEvent (ev);

        Component source = comp; // first guess
        // For menu items, the first guess does not work well--really we want
        // to examine the currently open menu. Note special treatment of submenus
        // which are open, but no items selected--ignore, and use the JMenu instead.
        javax.swing.MenuElement[] menuitems =
            javax.swing.MenuSelectionManager.defaultManager ().getSelectedPath ();
        if (menuitems != null) {
            for (int idx = menuitems.length - 1; idx >= 0; idx--) {
                Component menuitem = menuitems[idx].getComponent ();
                if (! (menuitem instanceof javax.swing.JPopupMenu) &&
                        (menuitem != null)) {
                    source = menuitem;
                    break;
                }
            }
        }
        // Provide a reasonably useful action event that identifies what was focused
        // when the key was pressed, as well as what keystroke ran the action.
        ActionEvent aev = new ActionEvent(
            source, ActionEvent.ACTION_PERFORMED, Utilities.keyToString (key)
        );

        if (processKeyStroke (key, aev)) {
            ev.consume ();
        } else {
            java.awt.Window w = SwingUtilities.windowForComponent(ev.getComponent ());
            JFrame mw = WindowManagerImpl.mainWindow();
            if (mw != w) {
                mw.getJMenuBar().dispatchEvent (ev);
/*                            
                if (ev.isConsumed()) {
                    MenuElement[] arr = MenuSelectionManager.defaultManager().getSelectedPath();
                    if (arr != null && arr.length > 0) {
                        Component c = arr[arr.length - 1].getComponent ();
                        c.requestFocus();
                    }
                }
*/                            
            }
        }
    }

    /** base context */
    private static Keymap root = TopManager.getDefault ().getGlobalKeymap ();

    /**
    * Process keyStroke action. Current context will be set to return value
    * of distributeKey(key) method.
    * @param key Key to be processed.
    * @param ev an event to send
    * @return true if the key has been processed and false otherwise
    */
    static boolean processKeyStroke(KeyStroke key, ActionEvent ev) {
        Action a = root.getAction (key);
        //System.err.println ("Running in SM");
        if (a != null) {
            //System.err.println ("Performing action with event: " + ev);
            //System.err.println ("\tsource: " + ev.getSource ());
            //System.err.println ("\tID: " + ev.getID ());
            //System.err.println ("\tcommand: " + ev.getActionCommand ());
            //a.actionPerformed (ev);
            TopManager.getDefault().getActionManager().invokeAction(a, ev);
            return true;
        } else {
            return false;
        }
    }

    /** Checks to see if a given keystroke is bound to an action
    * which should function on all focused components.
    * This includes the Main Window, dialogs, popup menus, etc.
    * Otherwise only the Main Window and TopComponents will receive the keystroke.
    * By default, off, unless the action has a property named <code>OpenIDE-Transmodal-Action</code>
    * which is set to {@link Boolean#TRUE}.
    * @param key the keystroke to check
    * @return <code>true</code> if transmodal; <code>false</code> if a normal action, or
    * the key is not bound to anything in the global keymap
    */
    static boolean isTransmodalAction (KeyStroke key) {
        Action a = root.getAction (key);
        if (a == null) return false;
        Object val = a.getValue ("OpenIDE-Transmodal-Action"); // NOI18N
        return val != null && val.equals (Boolean.TRUE);
    }

} // end of IDEFocusManager
