/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * 
 * This class is a lyout manager that arranges four components around a central component,
 * much in thge same way as a BorderLayout, except side components can now dominate a 
 * complete side.
 *
 * @author  Hippy
 */
public class PerimeterLayout implements LayoutManager {
    /** 
     * Represents the the NORTH, or top, side.
     */
    public static final String NORTH = "NORTH";

    /** 
     * Represents the the SOUTH, or bottom, side.
     */
    public static final String SOUTH="SOUTH";

    /** 
     * Represents the the EAST, or right, side.
     */
    public static final String EAST = "EAST";

    /** 
     * Represents the the WEST, or left, side.
     */
    public static final String WEST="WEST";

    /** 
     * Represents the the CENTER.
     */
    public static final String  CENTER = "CENTER";
    
    /** Constant for domimance policy - first added component is dominant */
    public static final int FIRST_DOMINANT = 0;
    
    /** Constant for domimance policy - last added component is dominant */
    public static final int LAST_DOMINANT = 1;

    // side indexes
    private static final int NORTHINDEX = 0;
    private static final int SOUTHINDEX = 1;
    private static final int EASTINDEX = 2;
    private static final int WESTINDEX = 3;
    private static final int CENTERINDEX = 4;

    private static final ArrayList sideNames = new ArrayList(5);
    static {
        sideNames.add(NORTHINDEX, NORTH);
        sideNames.add(SOUTHINDEX, SOUTH);
        sideNames.add(EASTINDEX, EAST);
        sideNames.add(WESTINDEX, WEST);
        sideNames.add(CENTERINDEX, CENTER);
    }

    // gap between components
    private int gap = 4;

    // miniumim size of center
    private int minBreadth = 50;

    // miniumim size of center
    private Rectangle centerRect = null;

    // array of components and dominance
    private ArrayList sides = new ArrayList(5);
    {
        sides.add(null);
        sides.add(null);
        sides.add(null);
        sides.add(null);
        sides.add(null);
    }
    private ArrayList dominance = new ArrayList(4);
    
    /** current dominance policy when adding layout components */ 
    private int dominancePolicy = FIRST_DOMINANT;

    /** 
     * Creates new PerimeterLayout.
     */
    public PerimeterLayout() {
    }

    /** 
     * 
     * Creates new PerimeterLayout with specified gap between components. 
     *
     * @param gap  Specifies the gap between components.
     *
     */
    public PerimeterLayout(int gap) {
        this();
        // set the gap
        this.gap = gap;
    }

    /** 
     * 
     * Adds component to lay out.
     *
     * @param sideName  Name of the component
     * @param comp  The component to add.
     *
     */
    public void addLayoutComponent(String sideName, final Component comp) {
        // find out which side is requested and confirm its existence
        int side = sideNames.indexOf(sideName);
        if (side != -1) {
            // set the component to its proper side
            sides.set(side, comp);

            // CENTERINDEX component always bottom of dominance
            if (side != CENTERINDEX) {
                // add comp at dominance end if not already present
                if (!dominance.contains(comp)) {
                    switch (dominancePolicy) {
                        case LAST_DOMINANT:
                            dominance.add(0, comp);
                            break;
                        case FIRST_DOMINANT:
                            dominance.add(comp);
                            break;
                    }
                }
            }

            // set the basic size from preferred if current size is not inited
            Dimension curSize = comp.getSize();
            if ((curSize.width <= 0) || (curSize.height <= 0)) {
                comp.setSize(comp.getPreferredSize());
            }
        }
        else {
            // added to nonexistant side
            throw new IllegalArgumentException("invalid constraint: " + sideName + " " + side); // NOI18N
        }
    }

    /** 
     *
     * Remove a component.
     * 
     * @param comp The compontent to remove.
     * 
     */
    public void removeLayoutComponent(final Component comp) {
        // find the component
        int side = sides.indexOf(comp);
        if (side != -1) {
            // if a side, nullify its existence
            sides.set(side, null);
            dominance.remove(comp);
        }
    }

    /** 
     *
     * Returns the containers current size.
     * 
     * @param cont The container in question.
     *
     * @return The containers curretn size.
     * 
     */
    public Dimension preferredLayoutSize(final Container cont) {
        // return the current size of the container
        return cont.getSize();
    }

    /** 
     *
     * Returns the containers minumum size.
     * 
     * @param cont The container in question.
     *
     * @return The minimum size of the container.
     * 
     */
    public Dimension minimumLayoutSize(final Container container) {
        // return the container's minimum size.
        Dimension dimMin;
        synchronized(container.getTreeLock()) {
            Dimension dimension = new Dimension(0, 0);
            Component component = null;
            if((component = getChild(CENTERINDEX)) != null) {
                 Dimension dimCenter = component.getMinimumSize();
                 dimension.width += dimCenter.width;
                 dimension.height += dimCenter.height;
            }
            if((component = getChild(EASTINDEX)) != null) {
                 Dimension dimEast = component.getMinimumSize();
                 dimension.width += dimEast.width + gap;
                 dimension.height = Math.max(dimEast.height, dimension.height);
            }
            if((component = getChild(WESTINDEX)) != null) {
                 Dimension dimWest = component.getMinimumSize();
                 dimension.width += dimWest.width + gap;
                 dimension.height = Math.max(dimWest.height, dimension.height);
            }
            if((component = getChild(NORTHINDEX)) != null) {
                 Dimension dimNorth = component.getMinimumSize();
                 dimension.width = Math.max(dimNorth.width, dimension.width);
                 dimension.height += dimNorth.height + gap;
            }
            if((component = getChild(SOUTHINDEX)) != null) {
                 Dimension dimSouth = component.getMinimumSize();
                 dimension.width = Math.max(dimSouth.width, dimension.width);
                 dimension.height += dimSouth.height + gap;
            }
            Insets insets = container.getInsets();
            dimension.width += insets.left + insets.right;
            dimension.height += insets.top + insets.bottom;
            dimMin = dimension;
        }
        return dimMin;
    }

    private Component getChild(int index) {
        return (Component)sides.get(index);
    }    

    /** 
     *
     * Lays out the container and the center frames.
     * 
     * @param cont The container in question.
     *
     */
    public void layoutContainer(final Container cont) {
        // get the container rect and adjust for insets
        centerRect = cont.getBounds();
        centerRect.setLocation(0,0);
        Insets insets = cont.getInsets();
        centerRect.x -= insets.top;
        centerRect.y -= insets.left;
        centerRect.width -= insets.left + insets.right;
        centerRect.height -= insets.top + insets.bottom;

        // save origianl for container rect
        Rectangle contRect = (Rectangle)centerRect.clone();

        // find the last component to size, starting with the center one
        for (int i = 0; i < dominance.size(); i++) {
            int side = sides.indexOf(dominance.get(i));
            setPeriCompBounds(centerRect, contRect, side);
        }

        // now size the center rect
        Component centerComp = (Component)sides.get(CENTERINDEX);
        if (centerComp != null) {
            centerComp.setBounds(centerRect);
        }
        else if (!dominance.isEmpty()) {
            centerComp = (Component)dominance.get(dominance.size() - 1);
            centerComp.setBounds(centerComp.getBounds().union(centerRect));
        }
    }
    
    // Calculate the size of the north component
    private void setPeriCompBounds(Rectangle remaining, 
                                   Rectangle container,
                                   int side) {
        // are we bottom right, or vertical?
        boolean vertical = (side == WESTINDEX) || (side == EASTINDEX);
        boolean btmRght = (side == SOUTHINDEX) || (side == EASTINDEX);
        
        // get component and its bounds
        Component comp = (Component)sides.get(side);
        Rectangle compRect = comp.getBounds();
        
        if (vertical) {
            // rotate the rectangles in order to ensure proper sizing
            flip(compRect);
            flip(remaining);
            flip(container);
        }

        // width is always the remaining
        compRect.x = remaining.x;
        compRect.width = remaining.width;

        // calculate max height to allow for EACH components minimum: must have
        // at least minBreadth + gap
        int max = remaining.height;
        int min = minBreadth + gap;
        max -= (dominance.size() > 1) ? min : 0;
        max -= (sides.get(CENTERINDEX) == null) ? 0 : min;

        compRect.height = (compRect.height > max) ? max : compRect.height;
        compRect.height = (compRect.height < min) ? min : compRect.height;
        
        // the y coord is ALWAYS 0
        compRect.y = (btmRght == true) ? container.height - compRect.height : 0;

        // remove the comp's height from the remaining
        remaining.y = (btmRght == true) ? remaining.y : compRect.height + gap;
        remaining.height -= compRect.height + gap;
        
        if (vertical) {
            // rotate the rectangles in order to ensure proper placement
            flip(compRect);
            flip(remaining);
            flip(container);
        }

        // place the component
        comp.setBounds(compRect);
    }
    
    private void flip(Rectangle rect) {
        rect.setSize(rect.height, rect.width);
        rect.setLocation(rect.y, rect.x);
    }
    
    /** 
     *
     * Retrieves the gap between components in pixels.
     * 
     * @return The current gap the layout manager sets between components.
     * 
     */
    public int getGap() {
        return gap;
    }

    /** 
     *
     * Sets the gap between components in pixels.
     * 
     * @param g The gap the layout manager should set between components.
     * 
     */
    public void setGap(int g) {
        gap = g;
    }

    /** 
     *
     * Retrieves the minumum breadth a component can have.
     * 
     * @returns The minumum breadth the layout manager should allow a components.
     * 
     */
    public int getMinimumBreadth() {
        return minBreadth;
    }

    /** 
     *
     * Sets the minumum breadth a component can have.
     * 
     * @param d The minumum breadth the layout manager should allow a components.
     * 
     */
    public void setMinimumBreadth(int d) {
        minBreadth = d;
    }

    /** 
     *
     * Move the specified component to the front of the hierarchy, meaning that it
     * will cover the total side it resides upon.
     * 
     * @param comp The component in question.
     * 
     */
    public void setDominant(Component comp) {
        if (dominance.contains(comp)) {
            dominance.add(0, dominance.remove(dominance.indexOf(comp)));
        }
    }

    /** 
     *
     * Move the specified component with name sidename to the front of the hierarchy, 
     * meaning that it will cover the total side it resides upon.
     * 
     * @param sideName The name of the component in question.
     * 
     */
    public void setDominant(String sideName) {
        int side = sideNames.indexOf(sideName);
        if (side > -1 && side < CENTERINDEX) {
            Component comp = (Component)sides.get(side);
            if (comp != null) {
                dominance.remove(comp);
                dominance.add(0, comp);
            }
        }
        else {
            throw new IllegalArgumentException("unknown constraint: " + side);
        }
    }

    /** 
     *
     * Get the name of the dominant component.
     * 
     * @return The name of the dominant component.
     * 
     */
    public String getDominant() {
        int side = 0;
        Component dominant = (Component)dominance.get(0);
        for (side = 0; side < sides.size(); side++) {
            if (sides.get(side) == dominant) {
                return (String)sideNames.get(side);
            }
        }
        return CENTER;
    }

    /** Sets dominance policy that will be used for added layout components
     * @param dominancePolicy dominance policy - either FIRST_DOMINANT or LAST_DOMINANT 
     * Default value is FIRST_DOMINANT
     */
    public void setDominancePolicy (int dominancePolicy) {
        if ((dominancePolicy != FIRST_DOMINANT) && (dominancePolicy != LAST_DOMINANT)) {
            throw new IllegalArgumentException("invalid dominance policy " + dominancePolicy); // NOI18N
        }
        this.dominancePolicy = dominancePolicy;
    }
    
    /** @return Current dominance policy - either FIRST_DOMINANT or LAST_DOMINANT.
     * Default value is FIRST_DOMINANT
     */
    public int getDominancePolicy () {
        return dominancePolicy;
    }
    
    /** @return Dominance list of side names in current component layout */ 
    public java.util.List getDominance () {
        ArrayList result = new ArrayList(dominance.size());
        synchronized (this) {
            int curIndex = -1;
            for (Iterator iter = dominance.iterator(); iter.hasNext(); ) {
                curIndex = sides.indexOf(iter.next());
                result.add(sideNames.get(curIndex));
            }
        }
        return result;
    }
    
}
