/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.util.*;
import java.beans.PropertyVetoException;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

/**
 * A lightweight JComponent that uses a PerimeterLayout object as its layout manager. 
 *
 * @author  Hippy
 *
 */
public class PerimeterPane extends JComponent {
    // data members
        
    // the panels layout
    private PerimeterLayout layout = new PerimeterLayout();
        
    // keep track of components
    private HashMap components = new HashMap(5);
    
    
    private PerimeterGapListener pgl = new PerimeterGapListener();

    /** 
     * Creates new PerimeterPane.
     */
    public PerimeterPane() {
        // set the layout
        setLayout(layout);
        setBorder(BorderFactory.createEmptyBorder());
        
        PerimeterGapListener pgl = new PerimeterGapListener();
        addMouseMotionListener(pgl);
        addMouseListener(pgl);
    }
    
    /**
     * Overridden to keep better track of components and their sides.
     */
    protected void addImpl(Component comp, Object constraints, int index) {
        super.addImpl(comp, constraints, index);
        components.put(constraints, comp);
    }
    
    /**
     * Overridden to ensure that only a PerimeterLayout manager is used.  An AWTError will
     * thrown otherwise.
     */
    public void setLayout(LayoutManager mgr) {
        if (!(mgr instanceof PerimeterLayout)) {
            throw new AWTError("PerimeterPanel can only use PerimeterLayout");
        }
        else {
            super.setLayout(mgr);
        }
    }
    
    /**
     * 
     * Retrieve the component at the named side.
     *
     * @param side  The side in question.
     *
     * @return The component at the specified side, or null if no component.
     */
    public Component getComponentAt(String side) {
        return (Component)components.get(side);
    }
    
    /** 
     *
     * Convienience method to the component to the front of the hierarchy. 
     *
     * @param comp The component to move to the front.
     * 
     */
    public void moveToFront(Component comp) {
        layout.setDominant(comp);
        comp.requestFocus();        
    }

    /** 
     *
     * Convienience method to the component to the front of the hierarchy. 
     *
     * @param sideName The name of the component to move to the front.
     * 
     */
    public void moveToFront(String sideName) {
        Component comp = (Component)components.get(sideName);
        if (comp != null) {
            comp.requestFocus();
        }
        layout.setDominant(sideName);
    }

    /** Convenience method that delegates to PerimeterLayout.
     * PerimeterLayout is the only layout allowed for this pane.
     */
    public void setDominant(Component comp) {
        layout.setDominant (comp);
        revalidate();
    }
    
    /** Convenience method that delegates to PerimeterLayout.
     * PerimeterLayout is the only layout allowed for this pane.
     */
    public void setDominant(String sideName) {
        layout.setDominant (sideName);
        revalidate();
    }
    
    // for resizing 
    private class PerimeterGapListener 
        extends MouseAdapter implements MouseMotionListener {
        // minimum mouse must change to move
        private static final int CHANGE = 3;
        
        // sides
        private static final int NORTHSIDE = 0;
        private static final int SOUTHSIDE = 1;
        private static final int EASTSIDE = 2;
        private static final int WESTSIDE = 3;
        private static final int MIDDLE = 4;

        // the center, north, south, east, and west outcodes
        private int[] outcodes = new int[5];
        
        // the enclosing perimeter pane
        private PerimeterPane pPane = PerimeterPane.this;

        // keeps track if were in the middle of resizing
        private boolean resizing = false;
        private boolean dragging = false;

        // keep track of which directions
        private Point mouseLocation;

        // save the old cursor...
        int oldCursorType = pPane.getCursor().getType();
        boolean sizingCursorSet = false;
        
        public PerimeterGapListener() {
        }

        public void mousePressed(final MouseEvent me) {
            if ((me.getModifiers() & me.BUTTON1_MASK) != 0) {
                // get original position
                mouseLocation = me.getPoint();
                // resizing if clicked directly on JPanel
                resizing = (pPane.getComponentAt(mouseLocation) == pPane);
            }
        }
        
        public void mouseReleased(final MouseEvent me) {
            // resizing done
            resizing = false;
            dragging = false;
            pPane.setCursor(Cursor.getPredefinedCursor(oldCursorType));
        }

        public void mouseDragged(final MouseEvent me) {
            if (resizing == true) {
                dragging = true;
                // get new position
                Point newLoc = me.getPoint();
                // only move if greater than CHANGE pixels than the last position
                if (newLoc.distance(mouseLocation) > CHANGE) {
                    Component comp = null;
                    Rectangle rect = null;
                    HashMap comps = pPane.components;
                    Rectangle pRect = pPane.getBounds();
                    // resize appropriate comps if mouse pressed in proper pos
                    // e.g. for the SOUTH component, we check that we are south
                    // of the desktop/NORTH component and north of the SOUTH 
                    // component
                    int cType = pPane.getCursor().getType();
                    if (outcodes[NORTHSIDE] == Rectangle.OUT_BOTTOM) {
                        // resize the north panel
                        comp = (Component)comps.get(PerimeterLayout.NORTH);
                        rect = comp.getBounds();
                        rect.height = newLoc.y;
                        comp.setBounds(rect);
                    }
                    if (outcodes[SOUTHSIDE] == Rectangle.OUT_TOP) {
                        // resize the south panel
                        comp = (Component)comps.get(PerimeterLayout.SOUTH);
                        rect = comp.getBounds();
                        rect.height = pRect.height - newLoc.y - layout.getGap();
                        comp.setBounds(rect);
                    }
                    if (outcodes[EASTSIDE] == Rectangle.OUT_LEFT) {
                        // resize the east panel
                        comp = (Component)comps.get(PerimeterLayout.EAST);
                        rect = comp.getBounds();
                        rect.width = pRect.width - newLoc.x - layout.getGap();
                        rect.x = newLoc.x;
                        comp.setBounds(rect);
                    }
                    if (outcodes[WESTSIDE] == Rectangle.OUT_RIGHT) {
                        // resize the west panel
                        comp = (Component)comps.get(PerimeterLayout.WEST);
                        rect = comp.getBounds();
                        rect.width = newLoc.x;
                        comp.setBounds(rect);
                    }
                    
                    // update the postition
                    mouseLocation = newLoc;
                    pPane.revalidate();
                }
            }
        }
        
        /** Required for MouseMotionListener. */
        public void mouseMoved(final java.awt.event.MouseEvent me) {
            Component comp = pPane.getComponentAt(me.getPoint());
            
            // update the outcodes for each perimeter component, and thus the
            // center, empty or not
            updateOutcodes(me.getPoint());
            
            int newCursorType = getCursorType(outcodes[MIDDLE]);
            if (comp == pPane) {
                if (sizingCursorSet == false) {
                    // save opld cursor
                    oldCursorType = pPane.getCursor().getType();
                    // resizing is true
                    sizingCursorSet = true;
                }

                // set the cursor to resizing type
                pPane.setCursor(Cursor.getPredefinedCursor(newCursorType));
            }
            else if (sizingCursorSet == true) {
                pPane.setCursor(Cursor.getPredefinedCursor(oldCursorType));
                sizingCursorSet = false;
            }
        }

        /** Restores old cursor upon exit from component if not
         * during resizing */ 
        public void mouseExited (MouseEvent evt) {
            if (!dragging) {
                pPane.setCursor(Cursor.getPredefinedCursor(oldCursorType));
            }
        }
        
        private void updateOutcodes(Point newLoc) {            
            Component comp = 
                (Component)pPane.components.get(PerimeterLayout.NORTH);
            outcodes[NORTHSIDE] = (comp != null) ? 
                getOutcode(comp, newLoc) & Rectangle.OUT_BOTTOM : 0;
            
            comp = (Component)pPane.components.get(PerimeterLayout.SOUTH);
            outcodes[SOUTHSIDE] = (comp != null) ? 
                getOutcode(comp, newLoc) & Rectangle.OUT_TOP : 0;
            
            comp = 
                (Component)pPane.components.get(PerimeterLayout.EAST);
            outcodes[EASTSIDE] = (comp != null) ?
                getOutcode(comp, newLoc) & Rectangle.OUT_LEFT : 0;
            
            comp = 
                (Component)pPane.components.get(PerimeterLayout.WEST);
            outcodes[WESTSIDE] = (comp != null) ? 
                getOutcode(comp, newLoc) & Rectangle.OUT_RIGHT : 0;
            
            // set the middle outcode, which is the between one based on 
            // what we just calculated above
            outcodes[MIDDLE] = outcodes[NORTHSIDE] |
                               outcodes[SOUTHSIDE] |
                               outcodes[EASTSIDE] |
                               outcodes[WESTSIDE];
        }
        
        private int getOutcode(Component comp, Point newLoc) {
            // get the rect and grow it to the surrounding border
            Rectangle compRect = comp.getBounds();
            int gap = PerimeterPane.this.layout.getGap();
            compRect.setLocation(compRect.x - gap, compRect.y - gap);
            compRect.grow(gap * 2, gap * 2);
            
            // return an outcode if the mouse is within the gaps between comp's
            return (compRect.contains(newLoc) ? comp.getBounds().outcode(newLoc) : 0);
        }

        // note that the center component is not considered when choosing 
        // cursors
        private int getCursorType(int outcode) {
            switch (outcode) {
                case Rectangle.OUT_TOP:
                    return Cursor.S_RESIZE_CURSOR;
                case Rectangle.OUT_BOTTOM:
                    return Cursor.N_RESIZE_CURSOR;
                case Rectangle.OUT_LEFT:
                    return Cursor.E_RESIZE_CURSOR;
                case Rectangle.OUT_RIGHT:
                    return Cursor.W_RESIZE_CURSOR;
                case Rectangle.OUT_TOP | Rectangle.OUT_LEFT:
                    return Cursor.SE_RESIZE_CURSOR;
                case Rectangle.OUT_BOTTOM | Rectangle.OUT_RIGHT:
                    return Cursor.NW_RESIZE_CURSOR;
                case Rectangle.OUT_TOP | Rectangle.OUT_RIGHT:
                    return Cursor.SW_RESIZE_CURSOR;
                case Rectangle.OUT_BOTTOM | Rectangle.OUT_LEFT:
                    return Cursor.NE_RESIZE_CURSOR;
                case Rectangle.OUT_TOP | Rectangle.OUT_LEFT | Rectangle.OUT_RIGHT:
                case Rectangle.OUT_BOTTOM | Rectangle.OUT_LEFT | Rectangle.OUT_RIGHT:
                case Rectangle.OUT_TOP | Rectangle.OUT_BOTTOM | Rectangle.OUT_LEFT:
                case Rectangle.OUT_TOP | Rectangle.OUT_BOTTOM | Rectangle.OUT_RIGHT:
                case Rectangle.OUT_TOP | Rectangle.OUT_BOTTOM | 
                     Rectangle.OUT_LEFT | Rectangle.OUT_RIGHT:
                    return Cursor.MOVE_CURSOR;
                default:
                    return pPane.getCursor().getType();
            }
        }
    }
}
