/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.ImageIcon;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.plaf.TabbedPaneUI;
import javax.swing.plaf.basic.BasicTabbedPaneUI;

import org.openide.util.actions.SystemAction;
import org.openide.actions.NextTabAction;
import org.openide.actions.PreviousTabAction;
import org.openide.windows.TopComponent;
import org.openide.awt.MouseUtils;

import org.netbeans.core.ui.ArrowsTabbedPaneUI;
import org.netbeans.core.ui.PopupTabbedPaneUI;
import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.TabbedContainerUIManager;
import org.netbeans.core.windows.util.WindowUtils;

/**Split implementation of TopComponentContainer. It allows to add top
 * components to five positions - EAST, WEST, NORTH, SOUTH and CENTER. Every position
 * uses tabbed approach for managing top components.
 *
 * @author  Jano Rojcek
 */
public class SplitContainerImpl extends DefaultContainerImpl implements ChangeListener {
    
    /** helper variable - stores selected TC */
    private TopComponent wasSelected;
    
    /** position names from PerimeterLayout */
    private static final ArrayList positionNames = new ArrayList(5);
    static {
        positionNames.add(PerimeterLayout.NORTH);
        positionNames.add(PerimeterLayout.SOUTH);
        positionNames.add(PerimeterLayout.EAST);
        positionNames.add(PerimeterLayout.WEST);
        positionNames.add(PerimeterLayout.CENTER);
    }
    
    /** constraints for TCs : (TC, constr) */
    private HashMap constraints = new HashMap(15);
    
    /** Tabbed containers */
    private JTabbedPane[] tabbedContainer = new JTabbedPane[positionNames.size()];
    private BasicTabbedPaneUI[] basicTabbedPaneUI = new BasicTabbedPaneUI[positionNames.size()];
    private Dimension[] prefTabbedSize = new Dimension[positionNames.size()];
    private int[] selectedTabs = new int[positionNames.size()];
    
    /** Number of opened tabbed containers */
    private int counter = 0;
    
    /** Content pane */
    private PerimeterPane contentPane = new PerimeterPane();
    
    /** Popup menu invoker */
    private PopupMouseImpl popup = new PopupMouseImpl();
    
    /** Activates TCs */
    private FocusListener tabbedFocusListener = new TabbedContainerFocusL();
    private FocusListener tcFocusListener = new TopComponentFocusL();
    private ContainerListener containerListener = new ContainerL();
    
    /** deserialized constraints values (converted to regular constraints
     * during validation (validate method) */
    private ArrayList serConstraints;
    /** deserialized indexes of selected tabbed pane tabs */
    private ArrayList serSelectedTabs;
    
    /** Creates new SplitContainerImpl */
    public SplitContainerImpl() {
        super();
        for (int i = 0; i < positionNames.size(); i++)
            selectedTabs[i] = -1;
        setLayout(new BorderLayout());
        add(contentPane, BorderLayout.CENTER);
        NbFocusManager.install();
    }
    
    // -------------------- DefaultContainerImpl abstract methods ---------------------------
    
    
    /** Called when some of standard windows actions like close, clone, next tab
     * was invoked. Boolean flag is true when action was already processed
     * by this base abstract class.
     */
    protected void actionPerformed (SystemAction action, boolean processed) {
        if (processed) {
            return;
        }
        if (action instanceof NextTabAction || action instanceof PreviousTabAction) {
            // select previous or next component
            TopComponent selected = getSelectedTopComponent();
            JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(selected))];
            int count = tab.getTabCount();
            int index = tab.getSelectedIndex();
            if ((count <= 1) || (index == -1)) {
                return;
            }
            int delta = (action instanceof NextTabAction) ? 1 : -1;
            int newIndex = (index + delta + count) % count;
            tab.setSelectedIndex(newIndex);
        }
    }

    /** Enable or disable next and previous actions depending on
    * current inner state */
    void updateNextAndPreviousState () {
        TopComponent selected = getSelectedTopComponent();
        if (selected == null)
            return;
        JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(selected))];
        if (tab != null && tab.getComponentCount() > 1) {
            nextTab.setActionPerformer(this);
            prevTab.setActionPerformer(this);
        } else {
            nextTab.setActionPerformer(null);
            prevTab.setActionPerformer(null);
        }
    }
    
    /** @return true if new top component can be added, false otherwise  */
    public boolean canAdd(TopComponent[] tc) {
        return true;
    }
    
    /** @return Constraints of given top component in perimeter pane or null
     * if no constraints exist */
    public Object getConstraints(TopComponent tc) {
        return constraints.get(tc);
    }
    
    /** Attach TCs to AWT container hierarchy */
    protected void attachComponents() {
        // if there is only one TC in container then insert it directly into contentPane
        if (constraints.size() == 1) {
            TopComponent selected = getSelectedTopComponent();
            contentPane.add(selected, constraints.get(selected));
            // ensure that component is visible
            if (!selected.isVisible()) {
                selected.setVisible(true);
            }
            invalidate();
            validate();
            return;
        }        
        // otherwise use tabbed panes
        TopComponent[] curComps = getTopComponents();
        Image icon = null;
        for (int i = 0; i < curComps.length; i++) {
            hangUpFocusListener(curComps[i]);
            icon = curComps[i].getIcon();
            
            int ind = positionNames.indexOf(constraints.get(curComps[i]));
            //Fixed bug #10551 26 Mar 2001 by Marek Slama
            //Added opening of tabbed container when it is null. It happens after 
            //first user start when SplitContainer is Copy&Paste to another 
            //workspace where it was not present before. In such case tabbed container 
            //was not initialized.
            if (tabbedContainer[ind] == null) {
                openTabbedContainer(ind, curComps[i]);
            }
            //End of bugfix #10551
            
            tabbedContainer[ind].addTab(
                curComps[i].getName(),
                (icon == null) ? null : new ImageIcon(icon),
                curComps[i], curComps[i].getToolTipText()
            );
            refreshAfterAdding(curComps[i]);
        }
        // initialize selected tabs of tabbed panes
        for (int i = 0; i < positionNames.size(); i++) {
            if ((tabbedContainer[i] != null) && (selectedTabs[i] > -1) 
                && (selectedTabs[i] < tabbedContainer[i].getComponentCount())) {
                tabbedContainer[i].setSelectedIndex(selectedTabs[i]);
            }
        }

        //Bugfix #9352 20 Feb 2001 by Marek Slama
        //Added synchronization of selected top component with selected tab
        //when wasSelected is null
        //Synchronize selected top component with selected tab
        //Bugfix #10986 added check if wasSelected component is still present in container
        if ((wasSelected != null) && containsTopComponent(wasSelected)) {
            if (!wasSelected.isVisible()) {
                wasSelected.setVisible(true);
            }
            setSelectedTopComponent(wasSelected);
        } else {
            if (curComps.length > 0) {
                if (!curComps[curComps.length - 1].isVisible()) {
                    curComps[curComps.length - 1].setVisible(true);
                }
                setSelectedTopComponent(curComps[curComps.length - 1]);
            }
        }
        //End of bugfix #9352
    }
    
    /** Detach TCs from AWT container hierarchy */
    protected void detachComponents() {
        // if there is only one TC in container then remove it from contentPane
        if (constraints.size() == 1) {
            contentPane.remove(getSelectedTopComponent());
            return;
        }
        TopComponent[] curComps = getTopComponents();
        for (int i = 0; i < curComps.length; i++) {
            hangDownFocusListener(curComps[i]);
        }
        // otherwise remove TC's from tabbed panes and remember their selected indexes
        wasSelected = getSelectedTopComponent();
        for (int i = 0; i < positionNames.size(); i++) {
            if (tabbedContainer[i] != null) {
                selectedTabs[i] = tabbedContainer[i].getSelectedIndex();
                tabbedContainer[i].removeAll();
            } else {
                selectedTabs[i] = -1;
            }
        }
    }
        
    /** Updates visual state of container after given top component has beeen
     * added to this container.
     * Beware, the code must be written to ensure that top component
     * will receive addNotify() before componentActivated().
     * @param tc top component that was added
     */
    protected void componentAdded(TopComponent tc) {
        componentAdded(tc, PerimeterLayout.CENTER);
    }
    
    /** Updates visual state of container after given top component has beeen
     * added to this container.
     * Beware, the code must be written to ensure that top component
     * will receive addNotify() before componentActivated().
     * @param tc top component that was added
     */
    protected void componentAdded(TopComponent tc, Object constr) {
        if (positionNames.indexOf(constr) == -1)
            throw new IllegalArgumentException("invalid constraint: " + constr); //NOI18N
        synchronized (this) {
            constraints.put(tc, constr);
            // if this is first TC in container then insert it into contentPane
            if (constraints.size() == 1) {
                setSelectedTopComponent(tc);
                if (isAttached())
                    contentPane.add(tc, constraints.get(tc));
                return;
            }
            // if this is second TC then firstly remove previous from contentPane
            // and add it into tabbed pane
            if (constraints.size() == 2) {
                //Fixed bug #9885 15 Mar 2001 by Marek Slama
                //Sometimes it can happen that getSelectedTopComponent()
                //returns component which was not added to container yet
                //so we use first component here.
                //TopComponent current = getSelectedTopComponent();
                //In addition call of contentPane.remove(current)
                //was added to properly clean PerimeterLayout
                TopComponent[] tcs = getTopComponents();                
                TopComponent current = tcs[0];
                int i = positionNames.indexOf(constraints.get(current));
                
                contentPane.remove(current);
                
                openTabbedContainer(i, current);
                if (isAttached()) {
                    hangUpFocusListener(current);
                    Image icon = current.getIcon();
                    tabbedContainer[i].addTab(current.getName(),
                        icon == null ? null : new ImageIcon(icon),
                        current, current.getToolTipText());
                    refreshAfterAdding(current);
                }
            }
            // remove TC from tabbed pane
            int i = positionNames.indexOf(constraints.get(tc));
            if (tabbedContainer[i] == null) {
                openTabbedContainer(i, tc);
            }

            if (isAttached()) {
                hangUpFocusListener(tc);
                Image icon = tc.getIcon();
                tabbedContainer[i].addTab(tc.getName(),
                    icon == null ? null : new ImageIcon(icon),
                    tc, tc.getToolTipText());
                refreshAfterAdding(tc);
            }
        }
        // start to listen to top component's property changes
        //        wm.findManager(tc).addPropertyChangeListener(this);
    }
    
    protected void componentRemoved(TopComponent tc) {
        synchronized (this) {
            // if only one TC is in container then clear contentPane
            if (constraints.size() == 1) {
                constraints.remove(tc);
                doActivateComponent(null);
                contentPane.remove(tc);
                return;
            }
            // otherwise firstly remove supplied TC
            int i = positionNames.indexOf(constraints.get(tc));            
            JTabbedPane tab = tabbedContainer[i];
            if (tab != null) {
                //Bugfix #11990 check if component is in tab before it is removed
                //needed in jdk 1.4
                if (tab.indexOfComponent(tc) != -1) {
                    //Bugfix #10265 eliminate call of stateChanged
                    tab.removeChangeListener(this);
                    tab.remove(tc);
                    tab.addChangeListener(this);
                }
            }
            constraints.remove(tc);
            hangDownFocusListener(tc);
            //Fixed bug #9885 15 Mar 2001 by Marek Slama
            //It is necessary to call closeTabbedContainer(i) here
            //to correctly clean contentPane.
            //Check for (tab != null) must be added because when workspace
            //is closed closeTabbedContainer(i) is called when first component
            //is removed -- bug #9506 it is not possible to test when last
            //component is removed when workspace is not displayed.
            //
            //Fixed bug #9506 02 Mar 2001 by Marek Slama
            //When workspace is deleted from popup menu in workspace tab
            //workspace is not active - in such case tab.getTabCount() returns zero
            //and closeTabbedContainer() is called even it should not be called.
            //It causes exception when next component is removed because tab will 
            //be already null.
            //closeTabbedContainer() is called correctly later when last but one 
            //component is removed.
            //Following code commented
            if ((tab != null) && (tab.getTabCount() == 0)) {
                closeTabbedContainer(i);
            }
            //End of commented code end of bug fix #9506
            // if only one TC left then remove its tabbed container
            // and insert it directly into contentPane
            if (constraints.size() == 1) {
                TopComponent singleTc = (TopComponent)topComponents().get(0);
                i = positionNames.indexOf(constraints.get(singleTc));
                tab = tabbedContainer[i];
                if (tab != null) {
                    //Bugfix #10265 eliminate call of stateChanged    
                    tab.removeChangeListener(this);
                    //Bugfix #11668 remove was called incorrectly with tc instead of singleTc
                    tab.remove(singleTc);
                    tab.addChangeListener(this);
                    closeTabbedContainer(i);
                }
                hangDownFocusListener(singleTc);
                if (isAttached()) {
                    singleTc.setVisible(true);
                    contentPane.add(singleTc, constraints.get(singleTc));
                    //Bugfix #11182 added to update window title when last but one
                    //component is removed
                    mode.updateTitle();
                }
                setSelectedTopComponent(singleTc);
                reactivateSelectedComponent();
            } else if ((tab != null) && (tab.getSelectedComponent() != null)) {
                // synchronize current variable with tab selection
                setSelectedTopComponent((TopComponent)tab.getSelectedComponent());
                reactivateSelectedComponent();
                tab.invalidate ();
                tab.validate ();
                tab.repaint();
            }
        }
    }
    
    /** Called when icon of some component in this multi frame has changed  */
    protected void componentIconChanged(TopComponent tc, Image icon) {
        if (!containsTopComponent(tc))
            return;  // not added yet
        if (constraints.size() > 1) {
            JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(tc))];
            int index = tab.indexOfComponent(tc);
            if (index < 0)
                return;
            tab.setIconAt(index, new ImageIcon(icon));
            if (isAttached()) {
                tab.revalidate();
                tab.repaint();
            }
        }
    }
    
    /** Called when the name of some component in this multi frame has changed  */
    protected void componentNameChanged(TopComponent tc) {
        if (!containsTopComponent(tc))
            return;
        String newName = tc.getName();
        if (constraints.size() > 1) {
            JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(tc))];
            int index = tab.indexOfComponent(tc);
            if (index < 0)
                return;
            tab.setTitleAt(index, newName);
        }
        firePropertyChange("name", null, newName); // NOI18N
    }
    
    protected void componentSelected(TopComponent tc) {
        if (tc == null) {
            return;
        }
        if (constraints.size() > 1) {
            JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(tc))];
            if (tab.indexOfComponent(tc) >= 0) {
                tab.setSelectedComponent(tc);
            }
        }
    }
    
    /** Implementation of the ChangeListener interface
     * Called when some tab in the tabbed pane is activated.
     */
    public void stateChanged (ChangeEvent e) {
        JTabbedPane tab = (JTabbedPane)e.getSource();
        if ((!isVisible()) || (!isShowing()) || (tab.getTabCount() <= 0)) {
            return;
        }
        int index = tab.getSelectedIndex();
        TopComponent current = (index < 0) ? null : (TopComponent)tab.getComponentAt(index);
        setSelectedTopComponent(current);
        reactivateSelectedComponent();
    }
    
    /** Post-initialization after adding - called from addTopComponent method */
    protected void refreshAfterAdding (TopComponent tc) {
        JTabbedPane tab = tabbedContainer[positionNames.indexOf(constraints.get(tc))];
        // workaround for JTabbedPane bug, make visible if needed
        componentSelected(tc);
        if (!tc.isVisible())
            tc.setVisible(true);
        tab.invalidate();
        tab.validate();
        tab.repaint();
    }
    
    /** Called when first phase of WS deserialization is done.
     * Finish deserialization process by doing proper initialization
     * Caution: This method should be called only during deserialization
     * (or when container was not displayed yet)
     */
    public void validateData () {
        // validate superclass too (data was asociated with superclass using
        // attachData(data) call)
        super.validateData();
        
        //Bug fix 9579
        //Keep top selected component, because method componentAdded
        //reset it. It is set correctly in super.validateData().
        TopComponent tc = getSelectedTopComponent();
        //End
        // initialize constraints and tabbed panes
        if (serConstraints != null) {
            final Iterator constraintsIter = serConstraints.iterator();
            for (Iterator iter = topComponents().iterator();
                iter.hasNext() && constraintsIter.hasNext(); ) {
                componentAdded((TopComponent)iter.next(), constraintsIter.next());
            }
            serConstraints = null;
        }
        // initialize selected tabs in tabbed panes
        if (serSelectedTabs != null) {
            final Iterator selectedTabIterator = serSelectedTabs.iterator();
            for (int i = 0; i < positionNames.size() && selectedTabIterator.hasNext(); i++) {
                selectedTabs[i] = ((Integer)selectedTabIterator.next()).intValue();
            }
            serConstraints = null;
        } else {
            for (int i = 0; i < positionNames.size(); i++) {
                selectedTabs[i] = -1;
            }
        }
        //Bug fix 9579
        //Set top selected component to previous one
        setSelectedTopComponent(tc);
        //End
        // selected field synchronization
        wasSelected = getSelectedTopComponent();
    }

    // ----------- End of DefaultContainerImpl abstract methods ----------------------------

    /** Change TabbesPaneUI for all JTabbedPanes
     */
    public void changeUI () {
        int count = tabbedContainer.length;
        for (int i = 0; i < count; i++)
            changeUI(i);
    }

    /** Change TabbesPaneUI for specific JTabbedPane
     *  index - index of JTabbedPane
     */
    public void changeUI (int index) {
        if (tabbedContainer[index] != null) {
            switch(WindowManagerImpl.getDefault().tabbedContainerUIManager().getTabbedContainerUI()) {
                case TabbedContainerUIManager.MODE_BASIC:
                    tabbedContainer[index].setUI(basicTabbedPaneUI[index] != null ? basicTabbedPaneUI[index] : new BasicTabbedPaneUI());
                    break;
                case TabbedContainerUIManager.MODE_ARROWS:
                    tabbedContainer[index].setUI(new ArrowsTabbedPaneUI());
                    break;
                case TabbedContainerUIManager.MODE_POPUP:
                    tabbedContainer[index].setUI(new PopupTabbedPaneUI());
                    break;
            }
        }
    }

    /**Creates tabbed container on supplied position
     * @param i position
     * @param firstTc first top component being added to container, useful
     * for computing preferred size if no pref size is saved
     */
    private void openTabbedContainer(int i, TopComponent firstTc) {
        tabbedContainer[i] = new JTabbedPane();
        basicTabbedPaneUI[i] = tabbedContainer[i].getUI() instanceof BasicTabbedPaneUI ? (BasicTabbedPaneUI)tabbedContainer[i].getUI() : null;
        changeUI(i);
        tabbedContainer[i].setTabPlacement(JTabbedPane.BOTTOM);
        if (prefTabbedSize[i] == null) {
            // resize tabbed container according to size of
            // first top component that will be added to the tabbed pane
            Dimension prefSize = firstTc.getSize();
            Insets insets = tabbedContainer[i].getInsets();
            prefSize.width += insets.left + insets.right;
            prefSize.height += insets.top + insets.bottom;
            prefTabbedSize[i] = prefSize;
        }
        tabbedContainer[i].setPreferredSize(prefTabbedSize[i]);
        
        tabbedContainer[i].addChangeListener(this);
        tabbedContainer[i].addMouseListener(popup);
        tabbedContainer[i].addFocusListener(tabbedFocusListener);
        // if there is only one tabbed container then do not set border
        if (counter == 1) {
            for (int j = 0; j < positionNames.size(); j++) {
                if (tabbedContainer[j] != null) 
                    tabbedContainer[j].setBorder(BorderFactory.createEtchedBorder(Color.white, Color.darkGray));
            }
        } else if (counter > 1) {
                tabbedContainer[i].setBorder(BorderFactory.createEtchedBorder(Color.white, Color.darkGray));
        }
        counter++;
        contentPane.add(tabbedContainer[i], positionNames.get(i));
    }

    /**Deletes tabbed container on supplied position
     * @param i position
     */
    private void closeTabbedContainer(int i) {
        tabbedContainer[i].removeChangeListener(this);
        tabbedContainer[i].removeMouseListener(popup);
        tabbedContainer[i].removeFocusListener(tabbedFocusListener);
        prefTabbedSize[i] = tabbedContainer[i].getSize();
        contentPane.remove(tabbedContainer[i]);
        tabbedContainer[i] = null;
        counter--;
        // try to select some TC and remember size of other containers
        for (int j = 0; j < positionNames.size(); j++) {
            if (tabbedContainer[j] != null) {
                tabbedContainer[j].setPreferredSize(tabbedContainer[j].getSize());
                setSelectedTopComponent(
                    (TopComponent)tabbedContainer[j].getSelectedComponent()
                );
                // if there is only one tabbed pane yet then clear its border
                if (counter == 1) tabbedContainer[j].setBorder(null);
            }
        }
    }

    /** Hangs up focus listener on TC's component hierarchy */
    private void hangUpFocusListener(Component comp) {
        comp.addFocusListener(tcFocusListener);
        if (comp instanceof Container) {
            ((Container)comp).addContainerListener(containerListener);
            Component[] comps = ((Container)comp).getComponents();
            for (int i = 0; i < comps.length; i++) {
                hangUpFocusListener(comps[i]);
            }
        }
    }
    
    /** Hangs down focus listener from TC's component hierarchy */
    private void hangDownFocusListener(Component comp) {
        comp.removeFocusListener(tcFocusListener);
        if (comp instanceof Container) {
            ((Container)comp).removeContainerListener(containerListener);
            Component[] comps = ((Container)comp).getComponents();
            for (int i = 0; i < comps.length; i++) {
                hangDownFocusListener(comps[i]);
            }
        }
    }
    
    /** Listener on focus changes in tabbed container. Activates selected TC. */
    private class TabbedContainerFocusL extends FocusAdapter {
        /** Activates focused TC.
         * @param ev focus event from Tabbed container, TC or one of its subcomponents
         */
        public void focusGained(FocusEvent ev) {
            TopComponent tc = (TopComponent)((JTabbedPane)ev.getComponent()).getSelectedComponent();
            if ((tc != null) && (tc != getSelectedTopComponent())) {
                setSelectedTopComponent(tc);
                reactivateWithDelay();
            }
        }
    }
    
    /** Listener on focus changes in TC's hierarchy. Activates focused TC. */
    private class TopComponentFocusL extends FocusAdapter {
        /** Activates focused TC.
         * @param ev focus event from Tabbed container, TC or one of its subcomponents
         */
        public void focusGained(FocusEvent ev) {
            // if source was TC or one of its subcomponents find out which one
            TopComponent[] tcs = getTopComponents();
            for (int i = 0; i < tcs.length; i++) {
                if (SwingUtilities.isDescendingFrom(ev.getComponent(), tcs[i])) {
                    if (tcs[i] != getSelectedTopComponent()) {
                        //Bugfix #11604 for JDK 1.4, disable change listener on tab to avoid tab cycling
                        int j = positionNames.indexOf(constraints.get(tcs[i]));
                        JTabbedPane tab = tabbedContainer[j];
                        if (tab != null) {
                            tab.removeChangeListener(SplitContainerImpl.this);
                        }
                        setSelectedTopComponent(tcs[i]);
                        reactivateWithDelay();
                        if (tab != null) {
                            tab.addChangeListener(SplitContainerImpl.this);
                        }
                    }
                    break;
                }
            }
        }
    }
    
    /** Listens on changes in TCs and adds/removes focus listeners */
    private class ContainerL implements ContainerListener {
        /** When component is added to TC's hierarchy then hang up focus listener on it */
        public void componentAdded(ContainerEvent ev) {
            hangUpFocusListener(ev.getChild());
        }

        /** When component is removed from TC's hierarchy then hang down focus listener from it */
        public void componentRemoved(ContainerEvent ev) {
            hangDownFocusListener(ev.getChild());
        }
    }
    
    /** Popup menu reaction implementation */
    final private class PopupMouseImpl extends MouseUtils.PopupMouseAdapter {
        /** Called when the seqeunce of mouse events should lead to actual
         * showing of the popup menu. */
        protected void showPopup (MouseEvent e) {
            //TabbedPaneUI tabUI = tabc.getUI();
            //int clickTab = tabUI.tabForCoordinate(tabc, e.getX(), e.getY());
            //TopComponent selected = getSelectedTopComponent();
            JTabbedPane tab = (JTabbedPane)e.getSource();

            //Find on which tab mouse click is.
            TabbedPaneUI tabUI = tab.getUI();
            int clickTab = tabUI.tabForCoordinate(tab, e.getX(), e.getY());

            TopComponent selected = (TopComponent)tab.getSelectedComponent();
            setSelectedTopComponent(selected);
            reactivateSelectedComponent();
            //System.err.println("Trying to show popup for : " + selected.getName());
            //System.err.println("Activated : " + TopComponent.getRegistry().getActivated().getName());
            if ((selected == null) ||
            (!selected.equals(TopComponent.getRegistry().getActivated()))) {
                return;
            }
            //Bugfix #8581, popup menu is now opened only when user clicks on tab
            //not in empty area.
            if (clickTab != -1) {
                showPopupMenu(createPopupMenu(selected), e.getPoint(), tab);
            }
        }        
    } // end of PopupMouseImpl
    
/******************** Serialization management ***************/
    
        public static final String DEFAULT_CONTAINER = "defaultContainer";
        public static final String CONSTRAINTS_LIST = "constraintsArray";
        public static final String TABBED_CONTAINERS_SIZE = "tabbedSizesArray";
        public static final String INDEX_OF_SELECTED_TAB = "selectedTabIndexes";

    /** Nominates replacer for serialization.
     */
    private Object writeReplace () throws ObjectStreamException {
        return new SplittedContainerImpl$SerializationReplacer(this);
    }
    
        /** Serialization of properties of split container - default
         * container plus array of side constraints. */
        void writeData (ObjectOutputStream oos)
        throws IOException {
            ObjectOutputStream.PutField pf = oos.putFields();
            pf.put(DEFAULT_CONTAINER, new DefaultContainerImpl.SerializedData(this));
            ArrayList tcList = topComponents();
            ArrayList constraintsList = new ArrayList(tcList.size());
            for (Iterator iter = tcList.iterator(); iter.hasNext(); ) {
                constraintsList.add(constraints.get(iter.next()));
            }
            pf.put(CONSTRAINTS_LIST, constraintsList);
            ArrayList tabbedContSize = new ArrayList(prefTabbedSize.length);
            for (int i = 0; i < prefTabbedSize.length; i++) {
                tabbedContSize.add(tabbedContainer[i] == null 
                    ? prefTabbedSize[i]
                    : tabbedContainer[i].getSize()
                );
            }
            pf.put(TABBED_CONTAINERS_SIZE, tabbedContSize);
            ArrayList selectedTabIndex = new ArrayList(positionNames.size());
            for (int i = 0; i < positionNames.size(); i++) {
                selectedTabIndex.add(tabbedContainer[i] == null 
                    ? new Integer(-1)
                    : new Integer(tabbedContainer[i].getSelectedIndex()));
            }
            pf.put(INDEX_OF_SELECTED_TAB, selectedTabIndex);
            oos.writeFields();
        }
        
        /** Deserialization of the workspace */
        void readData (ObjectInputStream ois)
        throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField gf = ois.readFields();
            attachData((DefaultContainerImpl.SerializedData)gf.get(DEFAULT_CONTAINER, null));
            serConstraints = (ArrayList)gf.get(CONSTRAINTS_LIST, null);
            //Bugfix #10993, workaround for JDK 1.2.2
            //if (gf.getObjectStreamClass().getField(TABBED_CONTAINERS_SIZE) != null) {
            if (WindowUtils.hasObjectStreamField(gf, TABBED_CONTAINERS_SIZE)) {
                ArrayList tabbedContSize = (ArrayList)gf.get(TABBED_CONTAINERS_SIZE, null);
                if (tabbedContSize != null)
                    prefTabbedSize = (Dimension[])tabbedContSize.toArray(new Dimension[5]);
            }            
            //Bugfix #10993, workaround for JDK 1.2.2
            //if (gf.getObjectStreamClass().getField(INDEX_OF_SELECTED_TAB) != null) {
            if (WindowUtils.hasObjectStreamField(gf, INDEX_OF_SELECTED_TAB)) {
                serSelectedTabs = (ArrayList)gf.get(INDEX_OF_SELECTED_TAB, null);
            }
        }
        
}
