/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.util.ArrayList;
import java.io.Serializable;
import java.awt.Container;
import java.beans.PropertyChangeListener;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JPanel;

import org.openide.TopManager;
import org.openide.windows.*;

import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.ModeImpl;

/** Interface for top component containers.
 * It's associated (contained) in mode and defines style in which set of
 * contained top components is visualized and represented. Concrete implementors 
 * should manage the task of visualizing contained components, switching
 * between them etc.
 * Implementors should also be aware of property changes of the mode which style
 * they represent and keep themself in sync.
 *
 * @author Dafe Simonek
*/
public interface TopComponentContainer extends Serializable {
    
    public static final String PROP_NAME = WindowManagerImpl.TopComponentManager.PROP_NAME; // NOI18N
    public static final String PROP_ICON = WindowManagerImpl.TopComponentManager.PROP_ICON; // NOI18N
    public static final String PROP_SELECTED = "selected"; // NOI18N

    /** @return an array containing all top components which are currently
    * managed by this container */
    public TopComponent[] getTopComponents ();

    /** Adds given top component to this container.
    * Beware, the code must be written to ensure that top component
    * will receive addNotify() before componentActivated(). 
    *
    * @param tc top component to add
    * @return false if top component already exixts in container or 
    * cannot be added due to container-dependent restrictions,
    * true otherwise
    */
    public boolean addTopComponent (TopComponent tc);

    /** Adds given top component to this container with specified constraints.
    * Beware, the code must be written to ensure that top component
    * will receive addNotify() before componentActivated(). 
    *
    * @param tc top component to add
    * @param constr constraints
    * @return false if top component already exixts in container or 
    * cannot be added due to container-dependent restrictions,
    * true otherwise
    */
    public boolean addTopComponent (TopComponent tc, Object constr);

    /** @return true if array of top components can be added, false otherwise
     * Given array can be empty or null - in this case it means that client
     * is just asking if container will be able to hold any other component
     * or not.
     */
    public abstract boolean canAdd (TopComponent[] tc);

    /** Removes specified top component from this container.
    * Beware! The code in this method must be written to ensure
    * that each top component will receive componentDeactivated() before
    * removeNotify().
    *
    * @param tc top component to remove
    * @return the count of remaining contained components
    */
    public abstract int removeTopComponent (TopComponent tc);

    /** @return true if container contains specified top component,
    * false otherwise
    */
    public boolean containsTopComponent (TopComponent tc);
    
    /** @return Constraints in which given top components lives in this
     * container. Null is valid and means that no constraints for specified
     * top component exist */
    public Object getConstraints (TopComponent tc);

    /** @return Currently selected top component or null if no top
    * component is selected at a time */ 
    public TopComponent getSelectedTopComponent ();

    /** Requests a focus for given top component. Do nothing if component
     * is not present in the container */
    public void requestFocus (TopComponent tc);
    
    /** Called when first phase of WS deserialization is done.
    * TC container will usually finish the deserialization of all
    * contained top components here.
    */
    public void validateData ();
    
    public Container getContentPane ();
    
    /** Sets visibility status for container
    * @param visible true if container should be made visible,
    * false otherwise
    */
    public void setAttached (boolean attached);
        
    /** Adds specified listener to property changes of this container */
    public void addPropertyChangeListener (PropertyChangeListener listener);
    
    /** Removes specified listener for property changes of this container */
    public void removePropertyChangeListener (PropertyChangeListener listener);

    /** Attaches this container to the given mode implementation and detaches
     * from old mode, if possible */
    public void attachToMode (org.netbeans.core.windows.ModeImpl mode);

}

/*
* Log
*  7    Gandalf   1.6         12/17/99 David Simonek   #1913, #2970
*  6    Gandalf   1.5         11/30/99 David Simonek   neccessary changes needed
*       to change main explorer to new UI style  (tabs are full top components 
*       now, visual workspace added, layout of editing workspace chnaged a bit)
*  5    Gandalf   1.4         11/3/99  David Simonek   completely rewritten 
*       serialization of windowing system...
*  4    Gandalf   1.3         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  3    Gandalf   1.2         9/13/99  David Simonek   request focus for the 
*       mode added
*  2    Gandalf   1.1         8/14/99  David Simonek   bugfixes, #3347, #3274 
*       etc.
*  1    Gandalf   1.0         7/11/99  David Simonek   
* $
*/
