/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.awt.event.WindowListener;
import java.util.HashMap;
import java.awt.Image;
import java.awt.GraphicsConfiguration;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;
import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;

import javax.swing.JFrame;
import javax.swing.WindowConstants;
import javax.swing.SwingUtilities;

import org.netbeans.core.windows.ModeImpl;


/** Implementation of SDI frame type based on JFrame swing class.
 *
 * @author  Dafe Simonek
 */
public class TopFrameTypeImpl extends JFrame implements FrameType, WindowListener {

    /** Map of <FrameTypeListener, WindowContainerAdapter>
     * Mapping between frame type listeners and WindowContainerAdapters.*/
    private HashMap flMap;
    /** Support for common operations */
    private FrameTypeSupport support;
    // Keep track of frame state.
    private int state = FrameType.NORMAL;

    
    /** Creates new TopFrameTypeImpl.  Constructs a new frame that is
        initially invisible. */ 
    public TopFrameTypeImpl() {
        this("");
    }

    /** Creates a new, initially invisible TopFrameTypeImpl with the specified 
        title. */
    public TopFrameTypeImpl(String title) {
        super(title);
        support = new FrameTypeSupport(this);
        addWindowListener(this);
    }

// commented out as we still should be jdk 1.2 compliant    
    /** Creates a TopFrameTypeImpl in the specified GraphicsConfiguration of 
        a screen device and a blank title. */
    /*public TopFrameTypeImpl(GraphicsConfiguration gc) {
        this("", gc);
    }

    /** Creates a TopFrameTypeImpl with the specified title and the specified 
        GraphicsConfiguration of a screen device. */
    /*public TopFrameTypeImpl(String title,GraphicsConfiguration gc) {
        super(title, gc);
        support = new FrameTypeSupport(this);
    }*/
    
    /** Get the Container containing the RootPane. */
    public Container getRootPaneContainer() {
        return this;
    }

    public void addNotify () {
        super.addNotify();
        support.addNotify();
    }
    
    public Component getIcon() {
        return null;
    }
    
    public void requestFocus () {
        if (getState() == Frame.ICONIFIED) {
            setState(Frame.NORMAL);
        }
        toFront();
        super.requestFocus();
        Component contentPane = getContentPane();
        if (contentPane != null) {
            contentPane.requestFocus();
        }
    }

    /** Request for UI update (when changing L&F) */
    public void updateUI () {
        SwingUtilities.updateComponentTreeUI(this);
    }

    /** Adds given container listener for listening to the container
    * events */
    public synchronized void addFrameTypeListener (FrameTypeListener fl) {
        if (flMap == null)
            flMap = new HashMap(5);
        WindowFrameListenerAdapter wfa = new WindowFrameListenerAdapter(this, fl);
        addWindowListener(wfa);
        flMap.put(fl, wfa);
    }

    /** Removes given container listener */
    public synchronized void removeFrameTypeListener (FrameTypeListener fl) {
        if (flMap == null)
            return;
        WindowFrameListenerAdapter wfa = (WindowFrameListenerAdapter)flMap.get(fl);
        if (wfa != null) {
            removeWindowListener(wfa);
        }
        flMap.remove(fl);
    }

    /** Get the type of frame. */
    public int getType () {
        return FrameType.TOP_LEVEL;
    }
    
    public void setSelected(boolean selected) throws PropertyVetoException {
        if (true == selected) {
            if (SwingUtilities.findFocusOwner(this) == null) {
                requestFocus();
            }
        } else {
            transferFocus();
        }
    }
 
    public void setState(int state) {
        super.setState(state);
        this.state = state;
    }

    public int getState() {
        return state;
    }
    
    public void windowOpened(java.awt.event.WindowEvent windowEvent) {
        state = getState();
    }
    
    public void windowClosed(java.awt.event.WindowEvent windowEvent) {
    }
    
    public void windowDeiconified(java.awt.event.WindowEvent windowEvent) {
        state = org.netbeans.core.windows.frames.FrameType.NORMAL;
    }
    
    public void windowClosing(java.awt.event.WindowEvent windowEvent) {
    }
    
    public void windowDeactivated(java.awt.event.WindowEvent windowEvent) {
    }
    
    public void windowActivated(java.awt.event.WindowEvent windowEvent) {
    }
    
    public void windowIconified(java.awt.event.WindowEvent windowEvent) {
        state = org.netbeans.core.windows.frames.FrameType.ICONIFIED;
    }
}
