/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.frames;

import java.util.Map;
import java.util.HashMap;
import java.lang.reflect.Constructor;

import org.openide.windows.TopComponent;
import org.openide.util.NbBundle;
import org.openide.TopManager;
import org.openide.ErrorManager;

/** Utility class, with class-only content. Its responsibility is to act as
 * a factory for FrameTypes and TopComponentContainers and provides access
 * to all existing frame and container types in the window system.
 *
 * @author Dafe Simonek
*/
public final class WindowTypesManager {

    /** constants identifying frame types */ 
    public static final String TOP_FRAME = "org.netbeans.core.windows.frames.TopFrameTypeImpl"; // NOI18N
    public static final String INTERNAL_FRAME = "org.netbeans.core.windows.frames.InternalFrameTypeImpl"; // NOI18N
    public static final String DESKTOP_FRAME = "org.netbeans.core.windows.frames.DesktopFrameTypeImpl"; // NOI18N
    // TBD - dialog frames not implemented yet, just constant is prepared 
    public static final String DIALOG_FRAME = "org.netbeans.core.windows.frames.DialogFrameTypeImpl"; // NOI18N
    
    /** constants identifying frame types */ 
    public static final String MULTI_TABBED_CONTAINER = "org.netbeans.core.windows.frames.MultiTabbedContainerImpl"; // NOI18N
    public static final String SPLIT_CONTAINER = "org.netbeans.core.windows.frames.SplitContainerImpl"; // NOI18N
    
    /* Map holding <String, String>, which means <containerClassName, containerDisplayName> */
    private static Map containerMap;
    
    /* Map holding <String, String>, which means <frameClassName, frameDisplayName> */
    private static Map frameMap;

    /** default frame and container types */
    private static String defaultFrameType = TOP_FRAME;
    private static String defaultContainerType = MULTI_TABBED_CONTAINER;
    
    private static final Class[] INTERNAL_CONSTRUCTOR_TYPES =
        new Class[] { String.class, Boolean.TYPE, Boolean.TYPE, Boolean.TYPE, Boolean.TYPE };
        
    private static final Object[] INTERNAL_CONSTRUCTOR_VALUES = 
        new Object[] { "", Boolean.TRUE, Boolean.TRUE, Boolean.TRUE, Boolean.TRUE };
    
    /** Creates and returns instance of container of top components,
     * specified by given class name. Throws IllegalArgumentException
     * if specified className is not known to this manager. */ 
    public static TopComponentContainer createContainer (String className) {
        return (TopComponentContainer)createInstance(className, containerMap());
    }

    /** Creates and returns instance of frame, specified
     * by given class name. Throws IllegalArgumentException if specified
     * className is not known to this manager. */ 
    public static FrameType createFrame (String className) {
        if (INTERNAL_FRAME.equals(className)) {
            return (FrameType)createInstance(className, frameMap(),
                INTERNAL_CONSTRUCTOR_TYPES, INTERNAL_CONSTRUCTOR_VALUES);
        }
        return (FrameType)createInstance(className, frameMap());
    }

    /** Returns map of existing containers.
     * Map holds entries of <String, String>, and semantically it is
     * <containerClassName, containerDisplayName>
     */
    public static Map getContainerMap () {
        return containerMap();
    }
    
    /** Returns map of existing frames.
     * Map holds entries of <String, String>, and semantically it is
     * <frameClassName, frameDisplayName>
     */
    public static Map getFrameMap () {
        return frameMap();
    }
    
    public static String getDefaultFrame () {
        return defaultFrameType;
    }
    
    public static void setDefaultFrame (String frameType) {
        if (!frameMap().containsKey(frameType)) {
            throw new IllegalArgumentException(
                NbBundle.getBundle(WindowTypesManager.class).getString("EXC_UnknownType")
                + frameType
            );
        }
        defaultFrameType = frameType;
    }
    
    public static String getDefaultContainer () {
        return defaultContainerType;
    }

    
    public static void setDefaulContainer (String containerType) {
        if (!containerMap().containsKey(containerType)) {
            throw new IllegalArgumentException(
                NbBundle.getBundle(WindowTypesManager.class).getString("EXC_UnknownType")
                + containerType
            );
        }
        defaultContainerType = containerType;
    }

    /** Safe getter for container map, actually creates it */
    private static Map containerMap () {
        if (containerMap == null) {
            containerMap = new HashMap(5);
            containerMap.put(
                MULTI_TABBED_CONTAINER,
                NbBundle.getBundle(WindowTypesManager.class).getString("CTL_TabbedContainerDisplayName") // NOI18N
            );
            containerMap.put(
                SPLIT_CONTAINER,
                NbBundle.getBundle(WindowTypesManager.class).getString("CTL_SplitContainerDisplayName") // NOI18N
            );
            // PENDING - add other entries
        }
        return containerMap;
    }

    /** Safe getter for frame map, actually creates it */
    private static Map frameMap () {
        if (frameMap == null) {
            frameMap = new HashMap(5);
            frameMap.put(
                TOP_FRAME,
                NbBundle.getBundle(WindowTypesManager.class).getString("CTL_TopFrameDisplayName") // NOI18N
            );
            frameMap.put(
                INTERNAL_FRAME,
                NbBundle.getBundle(WindowTypesManager.class).getString("CTL_InternalFrameDisplayName") // NOI18N
            );
            frameMap.put(
                DESKTOP_FRAME,
                NbBundle.getBundle(WindowTypesManager.class).getString("CTL_DesktopFrameDisplayName") // NOI18N
            );
            // PENDING - add other entries
        }
        return frameMap;
    }

    private static Object createInstance (String className, Map classesMap) {
        return createInstance(className, classesMap, null, null);
    }
    
    /** Actually creates instance from given class name, assuring that class
     * name is contained in specified map */
    private static Object createInstance (String className, Map classesMap, Class[] paramTypes, Object[] paramValues) {
        if (!classesMap.containsKey(className)) {
            throw new IllegalArgumentException(
                NbBundle.getBundle(WindowTypesManager.class).getString("EXC_UnknownType")
                + className
            );
        }
        Object instance = null;
        try {
            Class clazz = Class.forName(className);
            if ((paramTypes == null) || (paramValues == null)) {
                // use default constructor
                return clazz.newInstance();
            } else {
                // use constructor of given type
                Constructor constructor = clazz.getConstructor(paramTypes);
                return constructor.newInstance(paramValues);
            }
        } catch (Exception exc) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.ERROR, exc);
        }
        return instance;
    }
    
    
}

