/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.nodes;

import java.awt.Image;
import java.awt.datatransfer.Transferable;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyEditor;
import java.io.IOException;
import java.util.*;

import org.openide.util.datatransfer.PasteType;
import org.openide.util.datatransfer.ExTransferable;
import org.openide.util.datatransfer.NewType;
import org.openide.TopManager;
import org.openide.windows.Workspace;
import org.openide.windows.WindowManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.WeakListener;
import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;

import org.openide.actions.*;
import org.openide.cookies.InstanceCookie;
import org.openide.util.actions.SystemAction;
import org.openide.nodes.*;

import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.WorkspaceImpl;
import org.netbeans.core.windows.UIModeManager;
import org.netbeans.core.windows.TabbedContainerUIManager;

/** Node representing all workspaces.
*
* @author Ales Novak, Jan Jancura, Dafe Simonek
*/
public final class WorkspacePoolContext extends AbstractNode {
    /** default instance */
    private static WorkspacePoolContext defaultContext;

    /** Constructs new WorkspacePoolContext object
    */
    private WorkspacePoolContext () {
        // uses workspaces as children
        super (new Workspaces ());
        setName (NbBundle.getBundle (WorkspacePoolContext.class).getString ("Corona_workspaces"));
        setShortDescription (NbBundle.getBundle (WorkspacePoolContext.class).getString ("Corona_workspaces_HINT"));
        setIconBase ("/org/netbeans/core/resources/workspaces"); // NOI18N
    }

    /** Getter for the workspace pool node in the context.
    */
    public static WorkspacePoolContext getDefault () {
        if (defaultContext ==  null) {
            synchronized (WorkspacePoolContext.class) {
                if (defaultContext ==  null) {
                    defaultContext = new WorkspacePoolContext ();
                }
            }
        }
        return defaultContext;
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (WorkspacePoolContext.class);
    }

    //************** paste of workspace *******************

    /** Checks for instances of workspace elements.
    */
    protected void createPasteTypes (Transferable t, List s) {
        super.createPasteTypes (t, s);
        Workspace e = element((InstanceCookie)NodeTransfer.cookie (t, NodeTransfer.COPY, InstanceCookie.class));
        if (e != null) {
            // copy flavor
            s.add (new WorkspacePasteType (e, null));
            return;
        }

        // node to be cut
        Node n = NodeTransfer.node(t, NodeTransfer.CLIPBOARD_CUT | NodeTransfer.DND_MOVE);
        if (n != null && n.canDestroy ()) {
            e = element ((InstanceCookie)n.getCookie (InstanceCookie.class));
            s.add (new WorkspacePasteType (e, n));
            return;
        }
    }

    /** Takes workspace from the instance cookie or not.
    * @param cookie instance cookie
    * @return the workspace
    */
    private Workspace element (InstanceCookie cookie) {
        try {
            if (cookie != null && Workspace.class.isAssignableFrom (cookie.instanceClass ())) {
                return (Workspace)cookie.instanceCreate ();
            }
        } catch (Exception ex) {
            if (System.getProperty("netbeans.debug.exceptions") != null) ex.printStackTrace();
        }
        return null;
    }


    /** Paste type for transfering workspaces.
    */
    private class WorkspacePasteType extends PasteType {
        /** the workspace */
        private Workspace workspace;
        /** the node to destroy or null */
        private Node node;

        /** Constructs new WorkspacePasteType for the specific
        * type of operation paste.
        */
        public WorkspacePasteType(Workspace workspace, Node node) {
            this.workspace = workspace;
            this.node = node;
        }

        /* @return Human presentable name of this paste type. */
        public String getName() {
            return NbBundle.getBundle (WorkspacePoolContext.class).getString ("PASTE_Workspace");
        }

        /* @return help */
        public HelpCtx getHelpCtx() {
            return new HelpCtx (WorkspacePasteType.class);
        }

        /** Performs the paste action.
        * @return Transferable which should be inserted into the clipboard after
        *         paste action. It can be null, which means that clipboard content
        *         should stay the same.
        */
        public Transferable paste() throws IOException {
            addWorkspace(workspace);
            if (node != null) {
                node.destroy ();
            }
            return null;
        }
    }

    private static void addWorkspace(Workspace workspace) {
        Workspace[] wss = TopManager.getDefault().getWindowManager().getWorkspaces();
        Workspace[] nwss = new Workspace[wss.length + 1];
        System.arraycopy(wss, 0, nwss, 0, wss.length);
        nwss[wss.length] = workspace;
        ((WindowManagerImpl) TopManager.getDefault().getWindowManager()).setWorkspaces(nwss);
    }

    /**
    * @return an array of NewType that can be created
    */
    public NewType[] getNewTypes() {
        return new NewType[] {new WorkspaceNewType(this)};
    }

    void newChildWorkspace(String name) {
        Workspace dtop = new WorkspaceImpl(name);
        addWorkspace(dtop);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] createActions () {
        return new SystemAction []{
                   SystemAction.get (PasteAction.class),
                   null,
                   SystemAction.get (NewAction.class),
                   null,
                   SystemAction.get (ToolsAction.class),
                   SystemAction.get (PropertiesAction.class),
               };
    }
    
    /** Creates properties for this node */
//    protected Sheet createSheet () {
//        Sheet s = Sheet.createDefault ();
//        Sheet.Set ss = s.get (Sheet.PROPERTIES);
//        ss.put (new Node.Property[] {
//                    new UIModeProperty(),
//                    new TabbedContainerUIProperty()
//                });
//        return s;
//    }

    /* Specialized extension of Children.Keys which manages workspace list */
    private final static class Workspaces extends Children.Keys
        implements PropertyChangeListener {
        /** Holds listener to the workspace pool content */
        PropertyChangeListener propL;

        public void propertyChange (PropertyChangeEvent che) {
            if (che.getPropertyName().equals(WindowManager.PROP_WORKSPACES)) {
                Workspace[] newDesks = (Workspace[]) che.getNewValue();
                setKeys (newDesks);
            }
        }

        /** Initialize the keys from workspace pool */
        protected void addNotify () {
            WindowManager wp = TopManager.getDefault().getWindowManager();
            setKeys (wp.getWorkspaces());
            if (propL == null) {
                propL = WeakListener.propertyChange(this, wp);
            }
            wp.addPropertyChangeListener(propL);
        }

        /** Clear the keys */
        protected void removeNotify () {
            TopManager.getDefault().getWindowManager().
            removePropertyChangeListener(propL);
            setKeys(new Object[0]);
        }

        /** Creates a node for the workspace. */
        protected Node[] createNodes (Object o) {
            return new Node[] { new WorkspaceContext ((Workspace)o) };
        }
    } // end of Workspaces

    
    /** Property for changing the type of user interface of window system.
     * (values: SDI - Single Document Interface, MDI - Multiple Document Interface)
     *
     * @author  Dafe Simonek
     */
    public static class UIModeProperty extends PropertySupport.ReadWrite {

        /** ResourceBundle to get localized strings. */
        private static ResourceBundle bundle;
        /** Mode manager */
        private UIModeManager modeManager;

        /** Creates new UIModeProperty */
        public UIModeProperty() {
            super ("uiMode", Integer.class, // NOI18N
                    bundle().getString ("PROP_UI_Mode"),
                    bundle().getString ("HINT_UI_Mode")
                  );
            modeManager = WindowManagerImpl.getDefault().uiModeManager();
        }

        /** Get the value.
         * @return the value of the property
         */
        public Object getValue () {
            return new Integer(modeManager.getUIMode());
        }

        /** Set the value.
         * @param val the new value of the property
         */
        public void setValue (Object n) {
            if (n instanceof Integer) {
                modeManager.setUIMode(((Integer)n).intValue());
            } else {
                throw new IllegalArgumentException();
            }
        }

        /** Get a property editor for this property.
         * @return Get a property editor for this property.
         */
        public PropertyEditor getPropertyEditor () {
            return new UIModePropertyEditor();
        }

        private static final ResourceBundle bundle () {
            if (bundle == null) {
                bundle = NbBundle.getBundle(UIModeProperty.class);
            }
            return bundle;
        }
        
    } // end of UIModeProperty inner class

    /** Property for change TabbedPaneUI for containers including JTabbedPane
     */
    public static class TabbedContainerUIProperty extends PropertySupport.ReadWrite {

        /** Property label's bundle
         */
        private static ResourceBundle bundle;
        /** Manager of actual TabbedPaneUI
         */
        private TabbedContainerUIManager tabbedContainerUIManager;

        /** Constructor
         */
        public TabbedContainerUIProperty () {
            super ("TabbedContainerUI", Integer.class,           // NOI18N
                bundle().getString("PROP_TabbedContainerUI"),
                bundle().getString("HINT_TabbedContainerUI"));
                tabbedContainerUIManager = WindowManagerImpl.getDefault().tabbedContainerUIManager();
        }

        /** Reads an actual TabbedPaneUI value
         *  return - constant of TabbedPaneUI (TabbedContainerUIManager.MODE_...)
         */
        public Object getValue () {
            return new Integer(tabbedContainerUIManager.getTabbedContainerUI());
        }

        /** Sets an actual TabbedPaneUI value
         *  value - constant of TabbedPaneUI (TabbedContainerUIManager.MODE_...)
         */
        public void setValue (Object value) {
            if (value instanceof Integer) {
                    tabbedContainerUIManager.setTabbedContainerUI(((Integer)value).intValue());
            } else {
                throw new IllegalArgumentException();
            }
        }

        /** Reads a property editor with list of TabbedPaneUI values 
         */
        public PropertyEditor getPropertyEditor () {
            return new TabbedContainerUIPropertyEditor();
        }

        /** Reads a bundle
         */
        private static final ResourceBundle bundle () {
            if (bundle == null) {
                bundle = NbBundle.getBundle(TabbedContainerUIProperty.class);
            }
            return bundle;
        }
    }
}
