/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.toolbars;

import java.beans.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.awt.Window;
import java.awt.Container;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.*;
import javax.swing.*;

import org.openide.util.NbBundle;
import org.openide.NotifyDescriptor;
import org.openide.DialogDescriptor;
import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileLock;
import org.openide.awt.JPopupMenuPlus;
import org.openide.awt.Toolbar;
import org.openide.awt.ToolbarPool;
import org.openide.loaders.XMLDataObject;
import org.openide.util.SharedClassObject;

import org.netbeans.core.actions.DisplayToolbarsAction;
import org.netbeans.core.windows.WorkspaceImpl;

import org.w3c.dom.*;

/** Toolbar configuration.
 * It can load configuration from DOM Document, store configuration int XML file. 
 * Toolbar configuration contains list of all correct toolbars (toolbars which are
 * represented int ToolbarPool too), waiting toolbars (toolbars which was described
 * [it's position, visibility] but there is representation int ToolbarPool).
 * There is list of rows (ToolbarRow) and map of invisible toolbars.
 *
 * @author Libor Kramolis
 */
public final class ToolbarConfiguration extends Object 
implements ToolbarPool.Configuration, PropertyChangeListener {
    /** toolbar dtd location */
    protected static final String TOOLBAR_DTD               =
        "org/netbeans/core/windows/toolbars/toolbar.dtd"; // NOI18N
    /** location outside the IDE */
    protected static final String TOOLBAR_DTD_WEB           =
        "http://www.netbeans.org/dtds/toolbar.dtd"; // NOI18N
    /** toolbar dtd public id */
    protected static final String TOOLBAR_DTD_PUBLIC_ID     =
        "-//NetBeans IDE//DTD toolbar//EN"; // NOI18N
    /** toolbar prcessor class */
    protected static final Class  TOOLBAR_PROCESSOR_CLASS   =
        org.netbeans.core.windows.toolbars.ToolbarProcessor.class;
    /** toolbar icon base */
    protected static final String TOOLBAR_ICON_BASE         =
        "/org/netbeans/core/windows/toolbars/xmlToolbars"; // NOI18N
    
    /** error manager */
    private static ErrorManager ERR = TopManager.getDefault ().getErrorManager ().getInstance ("org.netbeans.core.windows.toolbars");

    /** last time the document has been reloaded */
    private volatile long lastReload;
    
    /** xml extension */
    protected static final String EXT_XML                   = "xml"; // NOI18N
//      /** xmlinfo extension */
//      protected static final String EXT_XMLINFO               = "xmlinfo"; // NOI18N

    /** Static initialization needed before toolbar XML files will be recognized.
     * [PENDING] Better would be to fix XMLDataObject so that it changes cookies etc.
     * when a new catalog entry is added that matches an existing object, but there
     * are probably a number of related bugfixes needed to make that work.
     * This is simpler.
     */
    public static void init () {
	/* Register public toolbar dtd id with toolbar configuration xml info */
  	XMLDataObject.registerInfo (TOOLBAR_DTD_PUBLIC_ID, getXMLInfo());
    }

    /** xml element for configuration (root element) */
    protected static final String TAG_CONFIG                = "Configuration"; // NOI18N
    /** xml element for row */
    protected static final String TAG_ROW                   = "Row"; // NOI18N
    /** xml element for toolbar */
    protected static final String TAG_TOOLBAR               = "Toolbar"; // NOI18N
    /** xml attribute for toolbar name */
    protected static final String ATT_TOOLBAR_NAME          = "name"; // NOI18N
    /** xml attribute for toolbar position */
    protected static final String ATT_TOOLBAR_POSITION      = "position"; // NOI18N
    /** xml attribute for toolbar visible */
    protected static final String ATT_TOOLBAR_VISIBLE       = "visible"; // NOI18N

    /** standard panel for all configurations */
    private static ToolbarPanel  toolbarPanel;
    /** toolbar layout manager for this configuration */
    private        ToolbarLayout toolbarLayout;
    /** toolbar drag and drop listener */
    private   ToolbarDnDListener toolbarListener;

    /** configuration xml info */
    private static XMLDataObject.Info xmlinfo;

    /** All toolbars which are represented in ToolbarPool too. */
    private WeakHashMap allToolbars;
    /** List of visible toolbar rows. */
    private Vector      toolbarRows;
    /** All invisible toolbars (visibility==false || tb.isCorrect==false). */
    private HashMap     invisibleToolbars;
    /** Toolbars which was described in DOM Document,
	but which aren't represented in ToolbarPool.
	For exapmle ComponentPalette and first start of IDE. */
    private WeakHashMap waitingToolbars;
    /** Name of configuration. */
    private String      configName;
    /** Cached preferred width. */
    private int         prefWidth;

   // private static final ResourceBundle bundle = NbBundle.getBundle (ToolbarConfiguration.class);

    /** Creates new empty toolbar configuration for specific name.
     * @param name new configuration name
     */
    public ToolbarConfiguration (String name) {
        configName = name;
        initInstance ();
    }

    /** Creates new toolbar configuration for specific name and from specific DOM Document.
     * @param name new configuration name
     * @param doc DOM Document represented toolbar configuration
     */
    public ToolbarConfiguration (String name, Document doc) {
        this (name);
        parseDocument (doc);
    }
    
    /** Clean all the configuration parameters.
     */
    private void initInstance () {
        allToolbars = new WeakHashMap();
        waitingToolbars = new WeakHashMap();
        toolbarRows = new Vector();
        invisibleToolbars = new HashMap();
        toolbarListener = new ToolbarDnDListener (this);
    }
    
    /** @return returns string from bundle for given string pattern */
    static final String getBundleString (String bundleStr) {
        return NbBundle.getBundle(ToolbarConfiguration.class).getString(bundleStr);
    }

    /** Add toolbar to list of all toolbars.
     * If specified toolbar constraints represents visible component 
     * it is added to specified toolbar row.
     * Othewise toolbar constraints is added to invisible toolbars.
     *
     * @param row toolbar row of new toolbar is part
     * @param tc added toolbar represented by ToolbarConstraints
     */
    void addToolbar (ToolbarRow row, ToolbarConstraints tc) {
        if (tc == null)
            return;

        if (tc.isVisible())
            row.addToolbar (tc);
        else {
            int rI;
            if (row == null)
                rI = toolbarRows.size();
            else
                rI = toolbarRows.indexOf (row);
            invisibleToolbars.put (tc, new Integer (rI));
        }
        allToolbars.put (tc.getName(), tc);
    }

    /** Remove toolbar from list of all toolbars.
     * This could mean that toolbar is represented only in DOM document.
     *
     * @param name name of removed toolbar
     */
    ToolbarConstraints removeToolbar (String name) {
        ToolbarConstraints tc = (ToolbarConstraints)allToolbars.remove (name);
        if (tc.destroy())
            checkToolbarRows();
        return tc;
    }

    /** Add toolbar row as last row.
     * @param row added toolbar row
     */
    void addRow (ToolbarRow row) {
        addRow (row, toolbarRows.size());
    }

    /** Add toolbar row to specific index.
     * @param row added toolbar row
     * @param index specified index of toolbar position
     */
    void addRow (ToolbarRow row, int index) {
	/* It is important to recompute row neighbournhood. */
        ToolbarRow prev = null;
        ToolbarRow next = null;
        try {
            prev = (ToolbarRow)toolbarRows.elementAt (index - 1);
        } catch (ArrayIndexOutOfBoundsException e) { }
        try {
            next = (ToolbarRow)toolbarRows.elementAt (index);
        } catch (ArrayIndexOutOfBoundsException e) { }

        if (prev != null)
            prev.setNextRow (row);
        row.setPrevRow (prev);
        row.setNextRow (next);
        if (next != null)
            next.setPrevRow (row);

        toolbarRows.insertElementAt (row, index);
        updateBounds (row);
    }

    /** Remove toolbar row from list of all rows.
     * @param row removed toolbar row
     */
    void removeRow (ToolbarRow row) {
	/* It is important to recompute row neighbournhood. */
        ToolbarRow prev = row.getPrevRow();
        ToolbarRow next = row.getNextRow();
        if (prev != null) {
            prev.setNextRow (next);
        }
        if (next != null) {
            next.setPrevRow (prev);
        }

        toolbarRows.removeElement (row);
        updateBounds (next);
        revalidateWindow();
    }

    /** Update toolbar row cached bounds.
     * @param row updated row
     */
    void updateBounds (ToolbarRow row) {
        while (row != null) {
            row.updateBounds();
            row = row.getNextRow();
        }
    }
    
    private static final ToolbarPool toolbarPool () {
        return ToolbarPool.getDefault ();
    }

    /** Revalidates toolbar pool window.
     * It is important for change height when number of rows is changed.
     */
    void revalidateWindow () {
        // replan to AWT thread, if needed (because of direct swing calls)
        if (SwingUtilities.isEventDispatchThread()) {
            doRevalidateWindow();
        } else {
            SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                    doRevalidateWindow();
                }
            });
        }
    }
    
    /** Performs revalidating work */
    private void doRevalidateWindow () {
        toolbarPanel().revalidate();
        java.awt.Window w = SwingUtilities.windowForComponent (toolbarPool ());
        if (w != null) {
            w.validate ();
        }
    }

    /** 
     * @param row specified toolbar row
     * @return index of toolbar row
     */
    int rowIndex (ToolbarRow row) {
        return toolbarRows.indexOf (row);
    }

    /** Updates cached preferred width of toolbar configuration.
     */
    void updatePrefWidth () {
        Iterator it = toolbarRows.iterator();
        prefWidth = 0;
        int tryPrefWidth;
        while (it.hasNext()) {
            prefWidth = Math.max (prefWidth, ((ToolbarRow)it.next()).getPrefWidth());
        }
    }

    /**
     * @return configuration preferred width
     */
    int getPrefWidth () {
        return prefWidth;
    }

    /**
     * @return configuration preferred height.
     * If there is no row, preferred height is 25% of BASIC_HEIGHT.
     */
    int getPrefHeight () {
        double rowCount = getRowCount();
        if (rowCount == 0)
            rowCount = 0.25;
        return (ToolbarLayout.VGAP + (int)((ToolbarLayout.VGAP + Toolbar.BASIC_HEIGHT) * rowCount));
    }

    /** Checks toolbar rows. If there is some empty row it is removed.
     */
    void checkToolbarRows () {
        Object[] rows = toolbarRows.toArray();
        ToolbarRow row;

        for (int i = rows.length - 1; i >= 0; i--) {
            row = (ToolbarRow)rows[i];
            if (row.isEmpty())
                removeRow (row);
        }
    }

    /**
     * @return number of rows.
     */
    int getRowCount () {
        return toolbarRows.size();
    }

    /**
     * @param name toolbar constraints name
     * @return toolbar constraints of specified name
     */
    ToolbarConstraints getToolbarConstraints (String name) {
        return (ToolbarConstraints)allToolbars.get (name);
    }

    /** Checks toolbars constraints if there is some of specific name.
     * If isn't then is created new toolbar constraints. Othewise is old
     * toolbar constraints confronted with new values (position, visibility).
     * @param name name of checked toolbar
     * @param position position of toolbar
     * @param visible visibility of toolbar
     * @return toolbar constraints for specifed toolbar name
     */
    ToolbarConstraints checkToolbarConstraints (String name, Integer position, Boolean visible) {
        ToolbarConstraints tc = (ToolbarConstraints)allToolbars.get (name);
        if (tc == null)
            tc = new ToolbarConstraints (this, name, position, visible);
        else
            tc.checkNextPosition (position, visible);
        return tc;
    }

    /** Checks whole toolbar configuration.
     * It confronts list of all toolbars and waiting toolbars
     * with toolbars represented by ToolbarPool.
     *
     * @return true if there is some change and is important another check.
     */
    boolean checkConfigurationOver () {
        boolean change = false;
        String name;
        Object[] waNas = waitingToolbars.keySet().toArray();
        Object[] names = allToolbars.keySet().toArray();

	/* Checks ToolbarPool with waiting list. */
        for (int i = 0; i < waNas.length; i++) {
            name = (String)waNas[i];
            if (toolbarPool ().findToolbar (name) != null) {  /* If there is new toolbar in the pool
							      which was sometimes described ... */
                ToolbarConstraints tc = (ToolbarConstraints)waitingToolbars.remove (name);
		                                           /* ... it's removed from waiting ... */
                allToolbars.put (name, tc);                /* ... so it's added to correct toolbars ... */
                addInvisible (tc);                         /* ... and added to visible toolbars. */
                change = true;
            }
        }

	/* Checks ToolbarPool with list of all toolbars ... reverse process than previous for. */
        for (int i = 0; i < names.length; i++) {
            name = (String)names[i];
            if (toolbarPool ().findToolbar (name) == null) {  /* If there is toolbar which is not represented int pool ... */
                ToolbarConstraints tc = removeToolbar (name);  /* ... so let's remove toolbar from all toolbars ... */
                waitingToolbars.put (name, tc);                /* ... and add to waiting list. */
                invisibleToolbars.put (tc, new Integer (tc.rowIndex()));
                change = true;
            }
        }
        return change;
    }

    /** Removes toolbar from visible toolbars.
     * @param tc specified toolbar
     */
    void removeVisible (ToolbarConstraints tc) {
        invisibleToolbars.put (tc, new Integer (tc.rowIndex()));
        if (tc.destroy())
            checkToolbarRows();
        tc.setVisible (false);

        //reflectChanges();
    }

    /** Adds toolbar from list of invisible to visible toolbars.
     * @param tc specified toolbar
     */
    void addInvisible (ToolbarConstraints tc) {
        int rC = toolbarRows.size();
        int pos = ((Integer)invisibleToolbars.remove (tc)).intValue();
        tc.setVisible (true);
        for (int i = pos; i < pos + tc.getRowCount(); i++) {
            getRow (i).addToolbar (tc, tc.getPosition());
        }

        if (rC != toolbarRows.size())
            revalidateWindow();
        //reflectChanges();
    }

    /**
     * @param rI index of required row
     * @return toolbar row of specified index.
     * If rI is out of bounds then new row is created.
     */
    ToolbarRow getRow (int rI) {
        ToolbarRow row;
        int s = toolbarRows.size();
        if (rI < 0) {
            row = new ToolbarRow (this);
            addRow (row, 0);
        } else if (rI >= s) {
            row = new ToolbarRow (this);
            addRow (row);
        } else {
            row = (ToolbarRow)toolbarRows.elementAt (rI);
        }
        return row;
    }

    /**
     * @return toolbar row at last row position.
     */
    ToolbarRow createLastRow () {
        return getRow (toolbarRows.size());
    }

    /** Reactivate toolbar panel.
     * All components are removed and again added using ToolbarPool's list of correct toolbars.
     *
     * @param someBarRemoved if some toolbar was previously removed and is important to reflect changes
     * @param writeAtAll if false the content of disk will not be updated at all
     */
    void reactivatePanel (boolean someBarRemoved, boolean writeAtAll) {
        toolbarPanel().removeAll();
        prefWidth = 0;

        Toolbar tbs[] = toolbarPool ().getToolbars();
        Toolbar tb;
        ToolbarConstraints tc;
        String name;
        ToolbarRow newRow = null;
        boolean someBarAdded = false;

        for (int i = 0; i < tbs.length; i++) {
            tb = tbs[i];
            name = tb.getName();
            tc = (ToolbarConstraints)allToolbars.get (name);
            if (tc == null) {                          /* If there is no toolbar constraints description defined yet ... */
                if (newRow == null)
                    newRow = createLastRow();
                tc = new ToolbarConstraints (this, name, null, Boolean.TRUE);  /* ... there is created a new constraints. */
                addToolbar (newRow, tc);
                someBarAdded = true;
            }
            toolbarPanel().add (tb, tc);
        }

        revalidateWindow();

        if (writeAtAll) {
            if (someBarRemoved || someBarAdded)
            {
                //Do not reflect non user change
                //reflectChanges();
            }
        }
    }

    /**
     * @return true if if important reactivate component.
     */
    boolean isImportantActivateComponent () {
        Object[] names = allToolbars.keySet().toArray();
        Toolbar[] toolbars = toolbarPool ().getToolbars();

	/* Is number of toolbars int local list and toolbar pool list different? */
        if (names.length != toolbars.length)
            return true;

	/* Is name of current configuration differrent of last toolbar pool configuration? */
        if (! configName.equals (toolbarPool ().getConfiguration()))
            return true;

        return false;
    }

    /** Reflects configuration changes ... write it to document.
     */
    void reflectChanges () {
        try {
            writeDocument();
        } catch (IOException e) { /* ??? */ }
    }

    /////////////////////////////////
    // from ToolbarPool.Configuration

    /** Activates the configuration and returns right
     * component that can display the configuration.
     * @return representation component
     */
    public Component activate () {
        return activate (isImportantActivateComponent (), true);
    }
        
        
    /** Activate.
     * @param isImportant is the change of structure important
     * @param writeAtAll write changes to disk or not?
     */
    private Component activate (boolean isImportant, boolean writeAtAll) {
        toolbarPool().setToolbarsListener (toolbarListener);

        boolean someBarRemoved = checkConfigurationOver();

        if (isImportant || someBarRemoved) {
            if (toolbarLayout == null) {
                toolbarLayout = new ToolbarLayout (this);
            }
            toolbarPanel().setLayout (toolbarLayout);
            reactivatePanel (someBarRemoved, writeAtAll);
        }

        return toolbarPanel();
    }

    /** Name of the configuration.
     * @return the name
     */
    public String getName () {
        return configName;
    }


    /** Popup menu that should be displayed when the users presses
     * right mouse button on the panel. This menu can contain
     * contains list of possible configurations, additional actions, etc.
     *
     * @return popup menu to be displayed
     */
    public JPopupMenu getContextMenu () {
        JPopupMenu menu = new JPopupMenuPlus ();

        // generate list of available toolbars
        Iterator it = Arrays.asList (ToolbarPool.getDefault ().getToolbars ()).iterator ();
        while (it.hasNext()) {
            final Toolbar tb = (Toolbar)it.next();
            final ToolbarConstraints tc = 
                (ToolbarConstraints)allToolbars.get (tb.getName());
/*            
            if (tc == null) {
                System.err.println("ToolbarConfiguration.java - error"); // NOI18N
                System.err.println("name: " + name); // NOI18N
                System.err.println(" all: " + allToolbars); // NOI18N
                continue;
            }
*/
            
            JCheckBoxMenuItem mi = new JCheckBoxMenuItem (
                tb.getDisplayName(), tc.isVisible()
            );
            mi.addActionListener (new ActionListener () {
                                      public void actionPerformed (ActionEvent ae) {
                                          boolean wasVisible = tc.isVisible();
                                          if (wasVisible) {
                                              removeVisible (tc);
                                              tb.setVisible (false);                                              
                                          } else {
                                              addInvisible (tc);
                                              tb.setVisible (true);
                                          }
                                          reflectChanges();
                                      }
                                  });
            menu.add (mi);
        }
        menu.add (new JPopupMenu.Separator());
        // generate list of available toolbar panels
        it = Arrays.asList (ToolbarPool.getDefault ().getConfigurations ()).iterator ();
        ButtonGroup bg = new ButtonGroup ();
        String current = ToolbarPool.getDefault ().getConfiguration ();
        while (it.hasNext()) {
            final String name = (String)it.next ();
            JRadioButtonMenuItem mi = new JRadioButtonMenuItem (name, (name.compareTo (current) == 0));
            mi.addActionListener (new ActionListener () {
                                      public void actionPerformed (ActionEvent e) {
                                          WorkspaceImpl curWs = (WorkspaceImpl)TopManager.getDefault().
                                                                getWindowManager().getCurrentWorkspace();
                                          curWs.setToolbarConfigName (name);
                                      }
                                  });
            bg.add (mi);
            menu.add (mi);
        }
        menu.add (new JPopupMenu.Separator());
        JMenuItem menuItem = new JMenuItem(((DisplayToolbarsAction)SharedClassObject.findObject(DisplayToolbarsAction.class, true)).getName());
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed (ActionEvent event) {
                ((DisplayToolbarsAction)SharedClassObject.findObject(DisplayToolbarsAction.class, true)).performAction();
            }
        });
        menu.add(menuItem);
        JMenuItem mi = new JMenuItem (getBundleString("PROP_saveAs"));
        mi.addActionListener (new ActionListener () {
                                  public void actionPerformed (ActionEvent e) {
                                      NotifyDescriptor.InputLine il = new NotifyDescriptor.InputLine
                                                                      (getBundleString("PROP_saveLabel"),
                                                                       getBundleString("PROP_saveDialog"));
                                      il.setInputText (getBundleString("PROP_saveName"));

                                      Object ok = TopManager.getDefault ().notify (il);
                                      if (ok == NotifyDescriptor.OK_OPTION) {
                                          String s = il.getInputText();
                                          if (s.length() != 0) {
                                              try {
                                                  String newName = il.getInputText();
                                                  if (tryWriteDocument (newName)) {
                                                      WorkspaceImpl curWs = (WorkspaceImpl)TopManager.getDefault().
                                                                            getWindowManager().getCurrentWorkspace();
                                                      curWs.setToolbarConfigName (newName);
                                                  }
                                              } catch (IOException ioe) {
                                                  TopManager.getDefault ().notifyException (ioe);
                                              }
                                          }
                                      }
                                  }
                              });
        menu.add (mi);

        return menu;
    } // getContextMenu



    ////// IO operations

    /** Parse toolbar configuration from DOM Document.
     * @param doc configuration described by DOM Document.
     */
    void parseDocument (Document doc) {
        Element rE = doc.getDocumentElement();
        if (!TAG_CONFIG.equals (rE.getTagName()))
            return;
        readConfiguration (rE);
        checkToolbarRows();
    }

    //// reading

    /** Parse toolbar configuration from root DOM Element.
     * @param doc root DOM Element.
     */
    void readConfiguration (Element conf) {
        NodeList nL = conf.getElementsByTagName (TAG_ROW);
        toolbarRows = new Vector();
        for (int i = 0; i < nL.getLength(); i++) {
            Node node = nL.item (i);
            readRow ((Element)node);
        }
        
        lastReload = System.currentTimeMillis();
        ERR.log ("readConfiguration: " + configName + " time: " + lastReload); // NOI18N
    }

    /** Parse toolbar row from DOM Element.
     * @param row row in DOM Element.
     */
    void readRow (Element row) {
        ToolbarRow tbRow = new ToolbarRow (this);
        addRow (tbRow);

        NodeList nL = row.getElementsByTagName (TAG_TOOLBAR);
        for (int i = 0; i < nL.getLength(); i++) {
            Node node = nL.item (i);
            addToolbar (tbRow, readToolbar ((Element)node));
        }
    }

    /** Parse toolbar constraints from DOM Element.
     * @param row toolbar in DOM Element.
     */
    ToolbarConstraints readToolbar (Element tb) {
        String  name;
        String  posStr;
        String  visStr;
        Integer pos;
        Boolean vis;

        name = tb.getAttribute (ATT_TOOLBAR_NAME);
        if (name.length() == 0)
            return null;

        posStr = tb.getAttribute (ATT_TOOLBAR_POSITION);
        if (posStr.length() == 0)
            pos = null;
        else
            pos = new Integer (posStr);

        visStr = tb.getAttribute (ATT_TOOLBAR_VISIBLE);
        if (visStr.length() == 0)
            vis = Boolean.TRUE;
        else
            vis = new Boolean (visStr);

        return checkToolbarConstraints (name, pos, vis);
    }

    //// writting

    /** Try write document to file. It is asked for replacing if there is configuration of same name.
     * @param cn configuration file name
     * @return false if don't want replace
     */
    boolean tryWriteDocument (String cn) throws IOException {
        final FileObject tbFO = TopManager.getDefault().getPlaces().folders().toolbars().getPrimaryFile();
        final FileSystem tbFS = tbFO.getFileSystem();

        FileObject newFO = tbFS.find (tbFO.getName(), cn, EXT_XML);
        if (newFO != null) {
            NotifyDescriptor replaceD = new NotifyDescriptor.Confirmation
                                        (MessageFormat.format (getBundleString("MSG_replaceConfiguration"),
                                                               new String [] { cn }),
                                         NotifyDescriptor.OK_CANCEL_OPTION, NotifyDescriptor.WARNING_MESSAGE);
            TopManager.getDefault().notify (replaceD);
            if (replaceD.getValue() != DialogDescriptor.OK_OPTION) {
                return false;
            }
        }
        writeDocument (cn);
        return true;
    }

    /** Write actual toolbar configuration. */
    public void writeDocument () throws IOException {
        writeDocument (configName);
    }

    /** Write toolbar configuration for specified file name to xml.
     * @param cn configuration file name
     */
    private void writeDocument (final String cn) throws IOException {
        ERR.log ("writeDocument: " + cn); // NOI18N
        WritableToolbarConfiguration wtc = new WritableToolbarConfiguration (toolbarRows, invisibleToolbars);
        final StringBuffer sb = new StringBuffer ("<?xml version=\"1.0\"?>\n\n"); // NOI18N
        sb.append ("<!DOCTYPE ").append (TAG_CONFIG).append (" PUBLIC \""). // NOI18N
        append (TOOLBAR_DTD_PUBLIC_ID).append ("\" \"" + TOOLBAR_DTD_WEB + "\">\n\n").append (wtc.toString()); // NOI18N

        final FileObject tbFO = TopManager.getDefault().getPlaces().folders().toolbars().getPrimaryFile();
        final FileSystem tbFS = tbFO.getFileSystem();

        tbFS.runAtomicAction (new FileSystem.AtomicAction () {
		public void run () throws IOException {
		    FileLock lock = null;
		    OutputStream os = null;
		    FileObject xmlFO = tbFS.find (tbFO.getName(), cn, EXT_XML);
		    if (xmlFO == null)
			xmlFO = tbFO.createData (cn, EXT_XML);
		    try {
			lock = xmlFO.lock ();
			os = xmlFO.getOutputStream (lock);
			
			PrintWriter writer = new PrintWriter (os);
			writer.print (sb.toString());
			writer.close();
		    } finally {
                        lastReload = System.currentTimeMillis ();
                        ERR.log ("Setting last reload: " + lastReload); // NOI18N
                        
			if (os != null)
			    os.close ();
			if (lock != null)
			    lock.releaseLock ();
		    }
		}
	    });
        ERR.log ("writeDocument finished"); // NOI18N
    }

    /** lazy init of toolbar panel */
    private static final ToolbarPanel toolbarPanel () {
        if (toolbarPanel == null) {
            toolbarPanel = new ToolbarPanel();
        }
        return toolbarPanel;
    }

    /** @return xmlinfo for toolbar configuration. */
    static XMLDataObject.Info getXMLInfo () {
        if (xmlinfo == null) {
            xmlinfo = new XMLDataObject.Info();
            xmlinfo.addProcessorClass (TOOLBAR_PROCESSOR_CLASS);
            xmlinfo.setIconBase (TOOLBAR_ICON_BASE);
        }
        return xmlinfo;
    }

    /** Listens on changes in the document.
     */
    public void propertyChange(PropertyChangeEvent ev) {
        if (!XMLDataObject.PROP_DOCUMENT.equals(ev.getPropertyName ())) {
            // interested only in PROP_DOCUMENT properties
            return;
        }
        
        final XMLDataObject obj = (XMLDataObject)ev.getSource ();
        long mod = obj.getPrimaryFile ().lastModified().getTime ();
        ERR.log ("Checking modified: " + lastReload); // NOI18N
        //Bugfix #10196, this condition commented to make sure that all changes
        //will be applied.
        /*if (lastReload >= mod) {
            // not changed since last refresh
            return;
        }*/
        
        SwingUtilities.invokeLater(new Runnable() {
            public void run () {
                try {
                    Document doc = obj.getDocument ();
                    ERR.log ("Reparsing file: " + obj.getPrimaryFile ()); // NOI18N

                    initInstance ();
                    parseDocument (doc);
                    checkConfigurationOver ();


                    if (configName.equals (toolbarPool ().getConfiguration())) {
                        ERR.log ("Activating the configuration"); // NOI18N
                        // 1st argument is true because the change is important
                        // 2nd argument is false, because it should prevent the system
                        // to write anything do
                        activate (true, false);
                    }

                } catch (IOException ex) {
                    TopManager.getDefault ().getErrorManager().notify (
                        ErrorManager.INFORMATIONAL, ex
                    );
                } catch (org.xml.sax.SAXException ex) {
                    TopManager.getDefault ().getErrorManager().notify (
                        ErrorManager.INFORMATIONAL, ex
                    );
                }
            }
        });
    }    

    // class WritableToolbarConfiguration
    class WritableToolbarConfiguration {
	/** List of rows. */
        Vector rows;

	/** Create new WritableToolbarConfiguration.
	 * @param rs list of rows
	 * @param iv map of invisible toolbars
	 */
        public WritableToolbarConfiguration (Vector rs, Map iv) {
            initRows (rs);
            initInvisible (iv);
            removeEmptyRows();
        }

        /** Init list of writable rows.
	 * @param rs list of rows
	 */
        void initRows (Vector rs) {
            rows = new Vector();

            Iterator it = rs.iterator();
            while (it.hasNext()) {
                rows.addElement (new ToolbarRow.WritableToolbarRow ((ToolbarRow)it.next()));
            }
        }

	/** Init invisible toolbars in toolbar rows.
	 * @param iv map of invisible toolbars
	 */
        void initInvisible (Map iv) {
            Iterator it = iv.keySet().iterator();
            ToolbarConstraints tc;
            int row;
            while (it.hasNext()) {
                tc = (ToolbarConstraints)it.next();
                row = ((Integer)iv.get (tc)).intValue();
                for (int i = row; i < row + tc.getRowCount(); i++) {
                    getRow (i).addToolbar (tc);
                }
            }
        }

	/** Removes empty rows. */
        void removeEmptyRows () {
            ToolbarRow.WritableToolbarRow row;
            for (int i = rows.size() - 1; i >= 0; i--) {
                row = (ToolbarRow.WritableToolbarRow)rows.elementAt (i);
                if (row.isEmpty())
                    rows.removeElement (row);
            }
        }

	/**
	 * @param r row index
	 * @return WritableToolbarRow for specified row index
	 */
        ToolbarRow.WritableToolbarRow getRow (int r) {
            try {
                return (ToolbarRow.WritableToolbarRow)rows.elementAt (r);
            } catch (ArrayIndexOutOfBoundsException e) {
                rows.addElement (new ToolbarRow.WritableToolbarRow ());
                return getRow (r);
            }
        }

        /** @return ToolbarConfiguration in xml format. */
        public String toString () {
            StringBuffer sb = new StringBuffer();

            sb.append ("<").append (TAG_CONFIG).append (">\n"); // NOI18N
            Iterator it = rows.iterator();
            while (it.hasNext()) {
                sb.append (it.next().toString());
            }
            sb.append ("</").append (TAG_CONFIG).append (">\n"); // NOI18N

            return sb.toString();
        }
    } // end of class WritableToolbarConfiguration
} // end of class Configuration
