/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.core.windows.util;

import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.event.MouseEvent;
import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Toolkit;
import java.awt.Insets;
import java.util.Set;
import java.util.Iterator;
import java.io.ObjectInputStream;
import java.io.ObjectStreamField;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JRootPane;
import javax.swing.SwingUtilities;

import org.openide.windows.Workspace;
import org.openide.windows.TopComponent;
import org.openide.TopManager;

import org.netbeans.core.windows.ModeImpl;
import org.netbeans.core.windows.WorkspaceImpl;
import org.netbeans.core.windows.WindowManagerImpl;
import org.netbeans.core.windows.UIModeManager;
import org.netbeans.core.windows.frames.WindowTypesManager;
import org.netbeans.core.windows.frames.PerimeterLayout;

/** This utility class serves as storage for common useful utility
 * methods tied to window system.
 *
 * @author Dafe Simonek
 */
public final class WindowUtils extends Object {
    
    /** No need to instantiate */
    private WindowUtils () {
    }
    
    /** Simple transformation to scale given rectangle between dimensions */
    public static void convertRectangle (Rectangle rect, Dimension oldDimension,
                                         Dimension newDimension) {
        if (oldDimension.equals(newDimension)) {                                 
            return;
        }
        double dx = ((double)newDimension.width) / ((double)oldDimension.width);
        double dy = ((double)newDimension.height) / ((double)oldDimension.height);
        rect.x = (int)(rect.x * dx);
        rect.y = (int)(rect.y * dy);
        rect.width = (int)(rect.width * dx);
        rect.height = (int)(rect.height * dy);
    }

    /** @return true if IDE currently in MDI mode, false otherwise */
    public static final boolean isMdi () {
        return WindowManagerImpl.getDefault().getUIMode() == UIModeManager.MDI_MODE;
    }

    /** Finds mode on specified workspace, which has specified constraints.
     * @param workspace Workspace to find mode in.
     * @param constraints Constraints of mode we are searching for
     * @return first mode which meets criterions or null if no such mode exist.
     * Both visible and invisible modes are searched for.
     */
    public static ModeImpl findConstrainedMode (Workspace workspace,
                                                Object constraints) {
        return findConstrainedMode(workspace, constraints, false);
    }
    
    /** Finds mode on specified workspace, which has specified constraints.
     * @param workspace Workspace to find mode in.
     * @param constraints Constraints of mode we are searching for
     * @param onlyVisible when true, only visible modes are searched for
     * @return first mode which meets criterions or null if no such mode exist
     */
    public static ModeImpl findConstrainedMode (Workspace workspace,
                                                Object constraints,
                                                boolean onlyVisible) {
        Set modes = workspace.getModes();
        ModeImpl curMode = null;
        Object curConstraints = null;
        for (Iterator iter = modes.iterator(); iter.hasNext(); ) {
            curMode = (ModeImpl)iter.next();
            if (onlyVisible && curMode.getOpenedTopComponents().length <= 0) {
                // ignore modes without opened components
                // if onlyVisible flag is set
                continue;
            }
            curConstraints = curMode.getConstraints();
            if (constraints == null) {
                if (curConstraints == null) {
                    return curMode;
                }
            } else if (constraints.equals(curConstraints)) {
                return curMode;
            }
        }
        return null;
    }
    
    /** Utility method, changes mode constraints, but avoids constraints
     * conflicts by deleting constraint of mode which occupied the constraint
     * given mode wants to have */
    public static void changeModeConstraints (ModeImpl mode, Object constraints) {
        if ((constraints != null) && !PerimeterLayout.CENTER.equals(constraints)) {
            ModeImpl conflictingMode = findConstrainedMode(mode.getWorkspace(), constraints);
            if (conflictingMode != null) {
                conflictingMode.setConstraints(null);
                conflictingMode.setFrameType(WindowTypesManager.getDefaultFrame());
            }
        }
        mode.setConstraints(constraints);
    }
    
    /** @return currently active mode or null. Active mode is
     * the mode on current workspace which contains currently activated
     * top component.
     */
    public static ModeImpl getActiveMode () {
        TopComponent comp = TopComponent.getRegistry().getActivated();
        if (comp == null) {
            return null;
        }
        WorkspaceImpl curWorkspace = (WorkspaceImpl)TopManager.getDefault().
                                     getWindowManager().getCurrentWorkspace();
        return (ModeImpl)curWorkspace.findMode(comp);
    }
    
    /** Sets new constraints and frame type for active mode. Active mode is
     * the mode on current workspace which contains currently activated
     * top component.
     * @return true if change was performed well, false if change cannot be
     * made, because no top component is active at the time
     */
    public static boolean changeActiveMode (Object constraints, String frameType) {
        ModeImpl activeMode = getActiveMode();
        if (activeMode == null) {
            // cannot perform
            return false;
        }
        boolean wasVisible = activeMode.isVisible();
        if (wasVisible) {
            activeMode.setVisible(false);
        }
        WindowUtils.changeModeConstraints(activeMode, constraints);
        activeMode.setFrameType(frameType);
        if (wasVisible) {
            activeMode.setVisible(true);
            WindowManagerImpl.mainWindow().getContentPane().validate();
        }
        return true;
    }

    /** @return Creates proxy for given component so that it's possible to
     * control all events that goes to given component. Useful for managing
     * existing components which is not possible to subclass for some reason.
     */
    public static JRootPane createComponentProxy (JComponent comp, AWTEventProcessor handler) {
        JRootPane rootPane = new JRootPane();
        rootPane.getContentPane().add(comp);
        JPanel glassPane = new GlassProxy(rootPane, comp, handler);
        rootPane.setGlassPane(glassPane);
        glassPane.setVisible(true);
        glassPane.setOpaque(false);
        return rootPane;
    }

    public static Dimension getMaximizedSize(java.awt.Container c) {
//        Dimension screen = org.openide.util.Utilities.getScreenSize();
        Dimension screen = java.awt.Toolkit.getDefaultToolkit().getScreenSize();
        Dimension max = null;
        if (org.openide.util.Utilities.isWindows()) {
            Insets borders = c.getInsets();
            max = new Dimension(
                screen.width + borders.left + borders.right,
                screen.height + 2 * borders.bottom              // top border size includes window caption
            );
        }
        else
            max = screen;
        
        return max;
    }
    
    /** @return true if gf contains field with given fieldName */
    //Bugfix #10993, in JDK 1.2.2 method ObjectStreamClass.getField() returns always null
    //whereas getFields() returns correct array so we must use getFields() instead of
    //getField() to check presence of field with given name.
    public static boolean hasObjectStreamField (ObjectInputStream.GetField gf, String fieldName) {
        ObjectStreamField[] osf = gf.getObjectStreamClass().getFields();
        for (int i = 0; i < osf.length; i++) {
            if (fieldName.equals(osf[i].getName())) {
                return true;
            }
        }
        return false;
    }
    
    /** Interface for processing awt events. */
    public interface AWTEventProcessor {
        
        /** Processes given event */
        public boolean handleEvent (AWTEvent evt);
        
    } // end of AWTEventProcessor

    /** Proxy component that is used to trap incoming events and pass it to
     * asociated event procesor */
    private static final class GlassProxy extends JPanel {
        
        /** root pane containing this proxy in a role of glass pane */
        private JRootPane rootPane;
        /** original component we are playing proxy for */
        private JComponent original;
        /** event handler for processing events before they go to original */
        private AWTEventProcessor handler;
        
        public GlassProxy (JRootPane rootPane, JComponent original,
                           AWTEventProcessor handler) {
            this.rootPane = rootPane;
            this.original = original;
            this.handler = handler;
            enableEvents(0xffffffffffffffffL);
        }
        
        /** Override event processing so that first goes associated
         * custom handler, then original dispatching */
        protected void processEvent(AWTEvent evt) {
            if (handler.handleEvent(evt)) {
                return;
            }
            // pass all events to the original component if handler
            // didn't consume event
            if (evt instanceof MouseEvent) {
                rootPane.dispatchEvent(evt);
                MouseEvent me = (MouseEvent)evt;
                Component c = original.getComponentAt(me.getPoint());
                if (c != null) {
                    c.dispatchEvent(
                        SwingUtilities.convertMouseEvent(this, me, c)
                    );
                }
            } else {
                original.dispatchEvent(evt);
            }
        }
        
    } // end of GlassProxy
}

