/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ResourceBundle;
import java.util.ArrayList;
import java.util.Arrays;
import javax.swing.JComponent;

import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.text.Line;
import org.openide.cookies.LineCookie;
import org.openide.cookies.DebuggerCookie;
import org.openide.debugger.Breakpoint;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.Watch;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.text.Line;
import org.openide.src.ConstructorElement;

import org.netbeans.modules.debugger.*;


/**
* Standart implementation of breakpoint interface.
* @see org.openide.debugger.Breakpoint
*
* @author   Jan Jancura
* @version  0.23, May 26, 1998
*/
public abstract class CoreBreakpoint extends Breakpoint implements java.io.Serializable {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 3686204002781932191L;

    /** Property name constant */
    public static final String      PROP_EVENT = "event"; // NOI18N
    /** Property name constant */
    public static final String      PROP_ACTIONS = "actions"; // NOI18N
    /** Property name constant */
    public static final String      PROP_HIDDEN = "hidden"; // NOI18N


    private transient PropertyChangeSupport   pcs;
    protected transient Debugger              debugger;
    private Action[]                          actions;
    protected Event                           event;
            
    
    // init ....................................................................
    
    public CoreBreakpoint () {
        init ();
    }
    
    protected void init () {
        pcs = new PropertyChangeSupport (this);
        if (debugger == null) 
            try {
                debugger = TopManager.getDefault ().getDebugger ();
            } catch (DebuggerNotFoundException e) {
            }
    }

    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        init ();
    }
    
    
    // main methods ............................................................
    
    /**
    * Returns current breakpoint event for given instance of debugger.
    */
    public Event getEvent (Debugger debugger) {
        if (debugger == this.debugger)
            return event;
        return null;
    }

    /**
     */
    protected Event createEvent (Debugger debugger, String type) {
        try {
            DebuggerImpl impl = null;
            if (debugger == TopManager.getDefault ().getDebugger ())
                impl = Register.getDebuggerCoreImpl ();
            else
                impl = Register.getDebuggerImpl (debugger);
            Event es[] = ((EventsProducer) impl).getEvents ();
            int i, k = es.length;
            for (i = 0; i < k; i++)
                if (es [i].getTypeName ().equals (type)) {
                    try {
                        Event e = (Event) es [i].getClass ().newInstance ();
                        e.breakpoint = this;
                        e.debugger = debugger;
                        return e;
                    } catch (Exception ex) {
                        ex.printStackTrace ();
                    }
                }
        } catch (DebuggerNotFoundException e) {
        }
        return null;
    }

    /**
     */
    public static Event createEvent (String type) {
        DebuggerImpl impl = Register.getDebuggerCoreImpl ();
        Event es[] = ((EventsProducer) impl).getEvents ();
        int i, k = es.length;
        for (i = 0; i < k; i++)
            if (es [i].getTypeName ().equals (type)) {
                try {
                    Event e = (Event) es [i].getClass ().newInstance ();
                    return e;
                } catch (Exception ex) {
                }
            }
        return null;
    }
    
    /**
    * Returns current breakpoint event for given instance of debugger.
    */
    public Event getEvent () {
        try {
            Debugger d = ((CoreDebugger) TopManager.getDefault ().getDebugger ()).
                getCurrentDebugger ();
            if (!(d instanceof BreakpointProducer)) return event;
            if (d != null) return getEvent (d);
            return event;
        } catch (DebuggerNotFoundException e) {
        }
        return null;
    }
    
    /**
    * Sets current breakpoint event.
    */
    public void setEvent (Event e) {
        Event old = event;
        e.breakpoint = this;
        event = e;
        firePropertyChange (PROP_EVENT, old, event);
        setActions (event.getActions ());
    }

    /**
    * Returns actions availabled for this breakpoint.
    */
    public CoreBreakpoint.Action[] getActions () {
        return actions;
    }
    
    /**
    * Returns actions availabled for this breakpoint.
    */
    public void setActions (CoreBreakpoint.Action[] act) {
        Action[] old = actions;
        actions = act;
        int i, k = actions.length;
        for (i = 0; i < k; i ++)
            actions [i].breakpoint = this;
        firePropertyChange (PROP_ACTIONS, old, actions);
    }

    public abstract void setHidden (boolean hidden);
    
    
    // other methods ............................................................
    
    /**
    * Performs given event for all actions.
    */
    protected void perform (Event e) {
        int i, k = actions.length;
        for (i = 0; i < k; i ++) 
              actions [i].perform (e);
    }
    
    /**
    * Performs given event for all actions.
    */
    protected abstract void setValid (boolean b);
        
    
    /**
    * Adds listener on the property changing.
    */
    public synchronized void addPropertyChangeListener (PropertyChangeListener listener) {
        pcs.addPropertyChangeListener (listener);
    }

    /**
    * Removes listener on the property changing.
    */
    public synchronized void removePropertyChangeListener (PropertyChangeListener listener){
        pcs.removePropertyChangeListener (listener);
    }

    /**
    * Fires property change of some event property (Needs re-setBreakpoint).
    */
    protected void fireEventPropertyChange (Event e, String s, Object o, Object n) {
        firePropertyChange (s, o, n);
    }
    
    /**
    * Fires property change.
    */
    protected void firePropertyChange (String s, Object o, Object n) {
        pcs.firePropertyChange (s, o, n);
    }

    
    // innerclasses .................................................................................

    /**
    * Breakpoint event implementation represents type of breakpoint (like breakpoint on class, 
    * method, exception). 
    */
    public abstract static class Event implements java.io.Serializable {
        static final long serialVersionUID =-4232563710736961248L;

        /** Proper instance of breakpoint. */
        private CoreBreakpoint         breakpoint;
        /** Proper instance of debugger. */
        private Debugger               debugger;

        
/*        protected Event (Debugger debugger, CoreBreakpoint) {
            this.debugger = debugger;
        }*/
        
        private void readObject (java.io.ObjectInputStream in)
        throws java.io.IOException, ClassNotFoundException {
            in.defaultReadObject ();
            try {
                debugger = (Debugger) TopManager.getDefault ().getDebugger ();
            } catch (DebuggerNotFoundException e) {
                throw new java.io.IOException ();
            }
        }

        // main methods ........................................................

        /**
        * Sets breakpoint.
        */
        public abstract boolean set ();

        /**
        * Clears breakpoint.
        */
        public abstract void remove ();

        /**
        * Returns name of type of this event.
        */
        public abstract String getTypeName ();

        /**
        * Returns display name of this event.
        */
        public abstract String getTypeDisplayName ();

        /**
        * Returns display name of this instance of event. It will be used
        * as the name of the breakpoint.
        */
        public abstract String getDisplayName ();

        /**
        * Returns name of icon.
        */
        public String getIconBase () {
            return null;
        }

        /**
        * Returns lines to highlite in the editor.
        */
        public abstract Line[] getLines ();

        /**
        * Returns specific properties of this event.
        */
        public abstract Node.Property[] getProperties();
        
        /**
        * Returns customizer visual component.
        */
        public abstract JComponent getCustomizer ();

        /**
        * Returns actions available specially for this version of event.
        */
        public CoreBreakpoint.Action[] getActions () {
            return new CoreBreakpoint.Action [0];
        }
        
        
        // helper methods ......................................................
        
        /**
        * Returns breakpoint for this event.
        */
        public CoreBreakpoint getBreakpoint () {
            return breakpoint;
        }

        /**
        * Returns debugger instance for this event.
        */
        public Debugger getDebugger () {
            return debugger;
        }
        
        protected void setValid (boolean b) {
            getBreakpoint ().setValid (b);
        }
        
        protected void perform () {
            getBreakpoint ().perform (this);
        }

        /**
        * Fires change of property.
        */
        protected void firePropertyChange (String s, Object o, Object n) {
            if (breakpoint != null)
                breakpoint.fireEventPropertyChange (this, s, o, n);
        }
    }

    /**
    * Breakpoint action implementation represents actions which will be performed on some 
    * breakpoint event (like stop debugger, show some message, show stack trace). 
    */
    public abstract static class Action implements java.io.Serializable {
        static final long serialVersionUID = 5993327662505228239L;

        /** Proper instance of debugger. */
        private CoreBreakpoint breakpoint;

        
        /*protected Action (CoreBreakpoint breakpoint) {
            this.breakpoint = breakpoint;
        }*/
        
        
        // main methods ........................................................
        
        /**
        * This method is called for each action when some breakpoint event 
        * is riched.
        */
        protected abstract void perform (Event e);

        /**
        * Returns specific properties of this event.
        */
        public abstract Node.Property[] getProperties ();

        /**
        * Returns customizer visuall component or null, if Action is not 
        * customizable.
        */
        public abstract JComponent getCustomizer ();


        // helper methods ......................................................

        /**
        * Returns breakpoint for this event.
        */
        public CoreBreakpoint getBreakpoint () {
            return breakpoint;
        }

        /**
        * Fires change of property.
        */
        protected void firePropertyChange (String s, Object o, Object n) {
            breakpoint.firePropertyChange (s, o, n);
        }
    }
}
