/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger;

import org.openide.debugger.Debugger;

import java.util.*;


/**
 * This class represents one type of debugger - Debugger Implementation.
 *
 * @author Jan Jancura
 */
public class Register {
    
    /** Links to default instance of DebuggerImpl. */
    private static DebuggerImpl defaultImpl;
    /** Set of curently installed DebuggerImpls. */
    private static LinkedList debuggerImpls = new LinkedList ();
    /** . */
    private static HashMap listeners = new HashMap ();

    
    /**
     * Returns default DebuggerImpl instance.
     *
     * @return default DebuggerImpl instance
     */
    public static DebuggerImpl getDebuggerCoreImpl () {
        if ((defaultImpl == null) &&
            (debuggerImpls.size () == 1)
        ) return (DebuggerImpl) debuggerImpls.get (0);
        return defaultImpl;
    }
    
    /**
     * Sets default DebuggerImpl instance. It can be set only once.
     * Throws SecurityException on the other hand. This method
     * should call DebuggerCore module implementation only.
     *
     * @throws SecurityException if default DebuggerImpl instance is already set
     */
    public static void setDebuggerCoreImpl (DebuggerImpl impl) {
        if (defaultImpl != null)
            throw new SecurityException ("Second default DebuggerImpl can not be set!");
        if (!(impl instanceof DebuggerCoreImpl))
            throw new SecurityException ("DebuggerImpl does not implement DebuggerCoreImpl!");
        defaultImpl = impl;
    }
    

    /**
     * Returns debugger impl for given debugger instance.
     */
    public static DebuggerImpl getDebuggerImpl (Debugger debugger) {
        if (defaultImpl != null) 
            return ((DebuggerCoreImpl) defaultImpl).getDebuggerImpl (debugger);
        if (debuggerImpls.size () == 1)
            return (DebuggerImpl) debuggerImpls.get (0);
        return null;
    }

    /**
     * Returns current debugger. Returns null if there is only one debugger
     * installed.
     */
/*    public static Debugger getCurrentDebugger () {
        if (defaultImpl != null) 
            return ((DebuggerCoreImpl) defaultImpl).getCurrentDebugger ();
        return null;
    }
*/    
    /**
     * Registers a new Debugger Implementation. 
     * Should be called from ModuleInstall.installed () and restore () 
     * for each Debugger Implementation.
     */
    public static void registerDebuggerImpl (DebuggerImpl debuggerImpl) {
        debuggerImpls.add (debuggerImpl);
        Iterator i = listeners.keySet().iterator();
        while (i.hasNext())
            ((RegisterListener)i.next()).debuggerImplAdded (debuggerImpl);
    }
    
    /**
     * Unregisters Debugger Implementation.
     * Should be called from ModuleInstall.uninstalled () for each Debugger 
     * Implementation.
     */
    public static void unregisterDebuggerImpl (DebuggerImpl debuggerImpl) {
        debuggerImpls.remove (debuggerImpl);
        Iterator i = listeners.keySet().iterator();
        while (i.hasNext())
            ((RegisterListener)i.next()).debuggerImplRemoved (debuggerImpl);
    }
    
    public static void addRegisterListener (RegisterListener rl) {
        listeners.put (rl, rl);
    }
    
    public static void removeRegisterListener (RegisterListener rl) {
        listeners.remove (rl);
    }
    
    /**
     * Listener for debugger registry.
     */
    public abstract static class RegisterListener {
        public abstract void debuggerImplAdded (DebuggerImpl debuggerImpl);
        public abstract void debuggerImplRemoved (DebuggerImpl debuggerImpl);
    }
}

