/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.delegator;

import java.util.Arrays;
import java.util.ArrayList;

import org.openide.TopManager;
import org.openide.text.Line;
import org.openide.debugger.*;
import org.openide.nodes.Node.Property;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.BreakpointSupport;


/**
* Core breakpoint delegates all functionality on inner instance of breakpint.
* @see org.openide.debugger.Breakpoint
*
* @author   Jan Jancura
*/
public class DelegatingBreakpoint extends BreakpointSupport {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 2246204002781932191L;
    
    
    // private variables .....................................................

    private transient Event         delEvent;
//    private boolean                 doNotFire = false;


    // init ....................................................................

    /**
    * Non public constructor called from the Debugger only.
    * User must create breakpoint from Debugger.getNewBreakpoint () method.
    */
    DelegatingBreakpoint (boolean hidden) {
        super (hidden);
    }

    protected void init () {
        delEvent = null;
        super.init ();
    }


    // CoreBreakpoint ..........................................................

    /**
    * Returns current breakpoint event for given instance of debugger (used
    * in multisession debugger implementation).
    */
    public Event getEvent (Debugger d) {
        if (d == debugger)
            return event;
        if ((delEvent != null) && (d == delEvent.getDebugger ()))
            return delEvent;
        return null;
    }

    
    // BreakpointSupport .......................................................
    
    /**
    * Sets or updates breakpoint. Calls set method of currently selected event for all debuggers, sets
    * value of valid property and marks lines in the editor.
    */
    protected void setBreakpoint () {
        if (!isEnabled ()) return;
        boolean valid = false;
        if (event != null) {
            // 1) set / remove
            if (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING) {
                if (delEvent != null)
                    delEvent.remove ();
                delEvent = null;
            } else {
                try {
                    // 1.1) init event
                    if (delEvent == null) {
                        Debugger d = ((DelegatingDebugger) TopManager.
                            getDefault ().getDebugger ()).getCurrentDebugger ();
                        if (d != null) {
                            Event e = createEvent (d, event.getTypeName ());
                            if (e != null) {
                                synchronizeProperties (e);
                                delEvent = e;
                            }
                        }
                    }
                    // 1.2) set event
                    if (delEvent != null)
                        valid = delEvent.set ();
                } catch (DebuggerException e) {
                }
            }
            // 2) remark
            Line[] l = null;
            if (delEvent != null)
                l = delEvent.getLines ();
            else
                l = event.getLines ();
            if (l != null)
                remark (new ArrayList (Arrays.asList (l)));
            else
                remark (null);
        }
        // 3) setValid
        setValid (valid);
    }

    /**
    * Removes breakpoint from this position.
    * Clears Line and breakpoint from RemoteDebugger.
    */
    protected void clearBreakpoint () {
        remark (false);
        lines = null;
        if (delEvent != null)
            delEvent.remove ();
        setValid (false);
    }

    /**
    * Fires property change of some event property (Needs re-setBreakpoint).
    */
    protected void fireEventPropertyChange (Event e, String s, Object o, Object n) {
        if (e != event) return;
        firePropertyChange (null, o, n);
        if (isEnabled ()) setBreakpoint ();
    }
    

    // helper methods ..........................................................

    /**
    * Synchronizes properties of all events of all debuggers.
    */
    protected void synchronizeProperties () {
        if (delEvent != null)
            synchronizeProperties (delEvent);
    }

    /**
    * Synchronizes properties with given event. [event => e}
    */
    protected void synchronizeProperties (Event e) {
        if (event == null) return;
//        doNotFire = true;
        Property[] s = event.getProperties ();
        Property[] d = e.getProperties ();
        int i,k = Math.min (d.length, s.length);
        for (i = 0; i < k; i++)
            try {
                d [i].setValue (s [i].getValue ());
            } catch (Exception ex) {  
            }
//        doNotFire = false;
    }

    /**
    * Creates breakpoint event for given debugger instance, adds it to 
    * events and calls set () for it.
    */
    void setDelegatingDebugger (Debugger debugger) {
        if (event == null) return;
        Event e = createEvent (debugger, event.getTypeName ());
        if (e == null) return;
        synchronizeProperties (e);
        delEvent = e;
        setBreakpoint ();
    }

    /**
    * Removes debugger event for given debugger instance.
    */
    void removeDelegatingDebugger () {
        if (delEvent != null)
            delEvent.remove ();
        delEvent = null;
        setBreakpoint ();
    }
}
