/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.delegator;

import java.beans.*;
import java.awt.Component;
import java.awt.Window;
import java.awt.Insets;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.lang.reflect.Constructor;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;
import javax.swing.border.EmptyBorder;

import org.openide.TopManager;
import org.openide.DialogDescriptor;
import org.openide.NotifyDescriptor;
import org.openide.actions.StartDebuggerAction;
import org.openide.debugger.*;
import org.openide.text.Line;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.util.actions.*;
import org.openide.windows.InputOutput;
import org.openide.windows.WindowManager;
import org.openide.windows.Workspace;
import org.openide.actions.StartDebuggerAction;
import org.openide.src.ClassElement;
import org.openide.src.ConstructorElement;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.util.Validator;

/**
 * NetBeans implementation of debugger. This class implements 
 * {@link org.openide.debugger.Debugger} class and acts the only one
 * debugger implementation for OpenIDE. It redirects all debugger calls 
 * to one of registered Debugger Implementations. 
 *
 * @author Jan Jancura, Daniel Prusa
 */
public class DelegatingDebugger extends DebuggerSupport 
implements ValidatorHolder, ThreadsRoot, BreakpointProducer, CoreDebugger {


    // static ..................................................................

    static final long                         serialVersionUID =
                                                7558733005739651906L;
    /** InputOutput of old session. */
    private static InputOutput                ioToClose;
    /** bundle to obtain text information from */
    private static transient ResourceBundle   bundle;
    /** Helper class for updating values of variales */
    private transient Validator               validator;
    
    
    /*
     * Returns bundle for this package.
     *
     * @return bundle for this package
     */
    static String getLocString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (DelegatingDebugger.class);
        return bundle.getString (s);
    }
    

    // variables .................................................................

    /** Currently running debugger session. */
    private transient Debugger                   debugger;
    /** Listens on inner istance of debugger. */
    private transient SessionListener            sessionListener = new SessionListener ();
    /** Delegating root of thread groups. */
    protected transient DelegatingThreadProducer threadProducer = new DelegatingThreadProducer ();


    // init .......................................................................

    public DelegatingDebugger () {
        if (Register.getDebuggerCoreImpl () == null)
            Register.setDebuggerCoreImpl (new DelegatingDebuggerImpl ());
        
        setValidator (new Validator ());
        
        addPropertyChangeListener (new StateListener ());
    }
    
    
    // Debugger implementation .................................................................

    /** Starts the debugger. The method takes information from the provided
    * info (containing the class to start and
    * arguments to pass it and name of class to stop debugging in) and starts
    * new debugging session.
    *
    * @param info debugger info about class to start
    * @exception DebuggerException if an error occures during the start of the debugger
    */
    public void startDebugger (DebuggerInfo info) throws DebuggerException {
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).changeEnabled (true);

        // close old tab in Output window
        if (ioToClose != null)
            ioToClose.closeInputOutput ();

        // start current session
        if (info != null)
            startSession (info);

        // ?!?! why twice Dan?!?!
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).changeEnabled (true);
    }

    /**
    * Finishes debugger.
    */
    public void finishDebugger () throws DebuggerException {
        try {
            if (debugger != null) 
                debugger.finishDebugger ();
            super.finishDebugger ();
            //      setConnectionState (STATE_NOT_RUNNING);
        } catch (DebuggerException ex) {
            TopManager.getDefault ().notify (
                new NotifyDescriptor.Exception (
                    ex.getTargetException () == null ? ex : ex.getTargetException (),
                    getLocString ("EXC_Debugger") + ": " + ex.getMessage () // NOI18N
                )
            );
        }
    }

    /**
    * Trace into.
    */
    public void traceInto () throws DebuggerException {
        Debugger debugger = getCurrentDebugger ();
        if (debugger == null) return;
        setLastAction (ACTION_TRACE_INTO);
        debugger.traceInto ();
    }

    /**
    * Trace over.
    */
    public void traceOver () throws DebuggerException {
        Debugger debugger = getCurrentDebugger ();
        if (debugger == null) return;
        setLastAction (ACTION_TRACE_OVER);
        debugger.traceOver ();
    }

    /**
    * Go.
    */
    public void go () throws DebuggerException {
        Debugger debugger = getCurrentDebugger ();
        if (debugger == null) return;
        setLastAction (ACTION_GO);
        debugger.go ();
    }


    /**
    * Step out.
    */
    public void stepOut () throws DebuggerException {
        Debugger debugger = getCurrentDebugger ();
        if (debugger == null) return;
        setLastAction (ACTION_STEP_OUT);
        debugger.stepOut ();
    }


    // BREAKPOINTS .......................................................

    /** Creates new breakpoint.
    *
    * @return new line breakpoint
    */
    public CoreBreakpoint createBreakpoint (boolean hidden) {
        DelegatingBreakpoint b = new DelegatingBreakpoint (hidden);
        addBreakpoint (b);
        return b;
    }
    
    // WATCHES ..............................................................

    /** Creates new uninitialized watch. The watch is visible (not hidden).
    *
    * @return new uninitialized watch
    */
    public Watch createWatch () {
        DelegatingWatch w = new DelegatingWatch (false);
        addWatch (w);
        return w;
    }

    /** Creates a watch its expression is set to initial value. Also
    * allows to create a watch not presented to the user, for example
    * for internal usage in the editor to obtain values of variables
    * under the mouse pointer.
    *
    * @param expr expresion to watch for
    * @param hidden allows to specify whether the watch should be presented
    *   to the user or not (be only of internal usage of the IDE).
    * @return new watch
    */
    public Watch createWatch (String expr, boolean hidden) {
        DelegatingWatch w = new DelegatingWatch (hidden);
        w.setVariableName (expr);
        addWatch (w);
        return w;
    }


    // AbstractDebugger implementation ................................................................


    // properties .......................


    /**
    * Getter for last action property.
    *
    * @return one of XXX_ACTION constants
    */
    public int getLastAction () {
        if (getCurrentDebugger () == null) return ACTION_START;
        if (!(getCurrentDebugger () instanceof AbstractDebugger)) return ACTION_START;
        return ((AbstractDebugger) getCurrentDebugger ()).getLastAction ();
    }
    
    /** Get the current line of debugger.
    *
    * @return current line
    */
    public Line getCurrentLine () {
        if (getCurrentDebugger () == null) return null;
        return getCurrentDebugger ().getCurrentLine ();
    }
    
    /**
    * Sets debugger state.
    */
//    public void setSuspended (boolean suspended) {
//        if (debugger instanceof AbstractDebugger) 
//            ((AbstractDebugger) debugger).setSuspended (suspended);
//    }

    /**
    * @return name of proces for given DebuggerInfo
    */
    public String getProcessName (DebuggerInfo info) {
        return "???"; // NOI18N nobody calls this method
    }

    /**
    * @return name of location for given DebuggerInfo
    */
    public String getLocationName (DebuggerInfo info) {
        return "???"; // NOI18N nobody calls this method
    }


    // ThreadRoot interface ....................................................

    /**
    * Returns root of all threads
    */
    public ThreadsProducer getThreadsRoot () {
        return threadProducer;
    }

    /**
    * Returns current thread or null
    */
    public AbstractThread getCurrentThread () {
        if (getCurrentDebugger () == null) return null;
        if (!(getCurrentDebugger () instanceof ThreadsRoot)) return null;
        return ((ThreadsRoot) getCurrentDebugger ()).getCurrentThread ();
    }

    /**
    * Sets current thread. If thread is null, unsets curent thread
    */
    public void setCurrentThread (AbstractThread thread) {
        thread.setCurrent (true);
    }

    /**
    * Returns validator.
    */
    public InputOutput getIO () {
        //return io.getIO ();
        return null;
    }
    
    /**
    * Prints given text to the output.
    */
    public void println (String text) {
        
    }
    
    
    // ValidatorCarrier interface .............................................. 
    
    /**
    * Returns validator.
    */
    public Validator getValidator () {
        return validator;
    }
    
    /**
    * Sets validator.
    */
    public void setValidator (Validator validator) {
        this.validator = validator;
    }
    
    
    // other methods ...........................................................

    /**
    * @return delegating debugger instance.
    */
    public Debugger getCurrentDebugger () {
        return debugger;
    }

    /**
     * Returns currently running debugger or null if debugger is not running.
     */
    public Debugger[] getDebuggers () {
        if (getCurrentDebugger () != null)
            return new Debugger [] {getCurrentDebugger ()};
        return new Debugger [0];
    }

    /**
    * Creates a new debugger session for given debugger type.
    */
    public void startSession (DebuggerInfo info) throws DebuggerException {
        if (getState () == DEBUGGER_NOT_RUNNING)
            setState (DEBUGGER_STARTING);
        
        debugger = createDebugger (info);
        firePropertyChange (PROP_CURRENT_DEBUGGER, null, debugger);
        firePropertyChange (PROP_DEBUGGERS, new Debugger [0], getDebuggers ());
        
        debugger.addPropertyChangeListener (sessionListener);
        debugger.startDebugger (info);
        if (debugger.getState () == DEBUGGER_NOT_RUNNING) {
            debugger.removePropertyChangeListener (sessionListener);
            Debugger o = debugger;
            Debugger[] oo = getDebuggers ();
            debugger = null;
            firePropertyChange (PROP_CURRENT_DEBUGGER, o, null);
            firePropertyChange (PROP_DEBUGGERS, oo, getDebuggers ());
            return;
        }
        if (debugger instanceof ThreadsRoot) 
            threadProducer.setRemoteThreadGroup (
                ((ThreadsRoot) debugger).getThreadsRoot ()
            );
    }
    
    /**
     * Creates new debugger instance.
     */
    protected Debugger createDebugger (DebuggerInfo info) {
        // get debugger instance
        Debugger debugger = DelegatingDebuggerImpl.getDelegatingImpl ().createDebugger (
            info instanceof SessionDebuggerInfo ? 
                ((SessionDebuggerInfo) info).getDebuggerImpl () :
                null
        );
        if (debugger instanceof ValidatorHolder)
            ((ValidatorHolder) debugger).setValidator (getValidator ());
        return debugger;
    }
    

    // private helper methods ....................................................

    /**
    * Called when some debugger instance is running first time.
    */
    private void debuggerRunning () {
        //    updateWatches ()
/*        CallbackSystemAction a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
        a.setActionPerformer (new ActionPerformer () {
            public void performAction (SystemAction a2) {
            setSuspended (true);
            }
            });
        a = (CallbackSystemAction) SystemAction.get (ResumeDebuggerAction.class);
        a.setActionPerformer (new ActionPerformer () {
              public void performAction (SystemAction a2) {
                  setSuspended (false);
              }
          });*/
        CallableSystemAction b = (CallableSystemAction) SystemAction.get (ConnectAction.class);
        b.setEnabled (false);
        
        if (debugger instanceof AbstractDebugger) {
            Breakpoint[] br = getBreakpoints ();
            int i, k = br.length;
            for (i = 0; i < k; i++)
                if (br [i] instanceof DelegatingBreakpoint)
                    ((DelegatingBreakpoint) br [i]).setDelegatingDebugger (
                        (AbstractDebugger) debugger
                    );
        }
    }

    /**
    * Called when last debugger session is finished.
    */
    private void debuggerNotRunning () {
        //    setBreakpoints ();
        CallbackSystemAction a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
        a.setActionPerformer (null);
        a = (CallbackSystemAction) SystemAction.get (ResumeDebuggerAction.class);
        a.setActionPerformer (null);
        CallableSystemAction b = (CallableSystemAction) SystemAction.get (ConnectAction.class);
        b.setEnabled (DelegatingDebuggerImpl.getDelegatingImpl ().
            supportsConnectAction ());

        try {
            super.finishDebugger (); 
        } catch (DebuggerException e) {
            TopManager.getDefault ().notify (
                new NotifyDescriptor.Exception(
                    e.getTargetException () == null ? e : e.getTargetException (),
                    getLocString ("EXC_Debugger") + ": " + e.getMessage ())
            );
        }
        TopManager.getDefault ().setStatusText (
            getLocString ("CTL_Debugger_end")
        );

        Breakpoint[] br = getBreakpoints ();
        int i, k = br.length;
        for (i = 0; i < k; i++)
            ((DelegatingBreakpoint) br [i]).removeDelegatingDebugger ();

/*        if (debugger instanceof SharedIO) {
            ioToClose = ((SharedIO) debugger).getIO ();
            ((AbstractDebugger) debugger).println (
                getLocString ("CTL_Debugger_session_end")
            );
        }*/

        debugger.removePropertyChangeListener (sessionListener);
        Debugger o = debugger;
        Debugger[] oo = getDebuggers ();
        debugger = null;
        firePropertyChange (PROP_CURRENT_DEBUGGER, o, null);
        firePropertyChange (PROP_DEBUGGERS, oo, getDebuggers ());
    }
    
    /** Stores set of breakpoints for each "generic" debugger instance. */
    HashMap gDebuggerToBreakpoints = new HashMap ();
    
    /**
     * Synchronizes breakpoint set betweens given debugger and this one.
     */
    void refreshBreakpoints (Debugger debugger) {
        HashMap sbh = new HashMap ();
        // adds new breakpoints from D to DD
        Breakpoint[] sb = debugger.getBreakpoints ();
        HashSet breakpoints = new HashSet (Arrays.asList (getBreakpoints ()));
        int i, k = sb.length;
        for (i = 0; i < k; i++) {
            if (!breakpoints.contains (sb [i])) {
                addBreakpoint (sb [i]);
            }
            sbh.put (sb [i], sb [i]);
        }
        // removes old breakpoint from DD
        Breakpoint[] bs = (Breakpoint[]) gDebuggerToBreakpoints.get (debugger);
        if (bs != null) {
            k = bs.length;
            for (i = 0; i < k; i++) {
                if (!sbh.containsKey (bs [i])) {
                    removeBreakpoint (bs [i]);
                }
            }
        }
        gDebuggerToBreakpoints.put (debugger, sb);
    }
    
    /** Stores set of watches for each "generic" debugger instance. */
    HashMap gDebuggerToWatches = new HashMap ();
    
    /**
     * Synchronizes watches set betweens given debugger and this one.
     */
    void refreshWatches (Debugger debugger) {
        HashMap newW = new HashMap ();
        // adds new breakpoints from D to DD
        Watch[] nw = debugger.getWatches ();
        HashSet watches = new HashSet (Arrays.asList (getWatches ()));
        int i, k = nw.length;
        for (i = 0; i < k; i++) {
            if (!watches.contains (nw [i])) {
                addWatch (nw [i]);
            }
            newW.put (nw [i], nw [i]);
        }
        // removes old breakpoint from DD
        Watch[] ow = (Watch[]) gDebuggerToWatches.get (debugger);
        if (ow != null) {
            k = ow.length;
            for (i = 0; i < k; i++) {
                if (!newW.containsKey (ow [i])) {
                    removeWatch (ow [i]);
                }
            }
        }
        gDebuggerToWatches.put (debugger, nw);
    }
    
    /**
    * Fires property change.
    */
    protected void firePropertyChange (String name, Object o, Object n) {
        super.firePropertyChange (name, o, n);
    }

    
    // innerclasses .............................................................

    /**
    * Listens on each session for current thread changes and debugger state.
    */  
    private class SessionListener implements PropertyChangeListener {
        public void propertyChange (PropertyChangeEvent e) {
            int os = getState ();
            int ns = debugger.getState ();

            if (e.getPropertyName ().equals (PROP_STATE)) {
                //S ystem.out.println ("Core debugger.state: " + os + " : " + ns);

                if ((os == DEBUGGER_NOT_RUNNING) || (os == ns)) return;
                setState (ns);
                if ((os == DEBUGGER_STARTING) && (ns == DEBUGGER_RUNNING))
                    debuggerRunning ();
                if (ns == DEBUGGER_NOT_RUNNING)
                    debuggerNotRunning ();

            } else
            if (e.getPropertyName ().equals (PROP_CURRENT_LINE)) {
                firePropertyChange (PROP_CURRENT_LINE, e.getOldValue (), e.getNewValue ());
            } else
            if (e.getPropertyName ().equals (ThreadsRoot.PROP_CURRENT_THREAD)) {
                firePropertyChange (ThreadsRoot.PROP_CURRENT_THREAD, e.getOldValue (), e.getNewValue ());
            } else
            if (e.getPropertyName ().equals (PROP_LAST_ACTION)) {
                firePropertyChange (PROP_LAST_ACTION, e.getOldValue (), e.getNewValue ());
            }
        }
    }
    
    private class StateListener implements PropertyChangeListener {
        public void propertyChange (PropertyChangeEvent e) {
            if (!e.getPropertyName ().equals (Debugger.PROP_STATE)) return;
            CallbackSystemAction a;
            switch (getState ()) {
                case Debugger.DEBUGGER_STARTING:
                    break;
                case Debugger.DEBUGGER_RUNNING:
                    a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
                    Debugger deb = getCurrentDebugger ();
                    if (deb instanceof SuspendResumeSupport)
                        a.setActionPerformer (new ActionPerformer () {
                            public void performAction (SystemAction action) {
                                Debugger d = getCurrentDebugger ();
                                if (d instanceof SuspendResumeSupport)
                                    ((SuspendResumeSupport) d).setSuspended (true);
                            }
                        });
                    break;
                case Debugger.DEBUGGER_NOT_RUNNING:
                case Debugger.DEBUGGER_STOPPED:
                    a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
                    a.setActionPerformer (null);
                    break;
                }
        }
    }
    
}
