/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.delegator;

import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import javax.swing.JComponent;

import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerNotFoundException;
import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.ConnectAction;
import org.netbeans.modules.debugger.support.actions.AddBreakpointAction;


/**
 * This class represents one type of debugger - Debugger Implementation.
 *
 * @author Jan Jancura
 */
public class DelegatingDebuggerImpl extends DebuggerImpl implements EventsProducer,
 DebuggerCoreImpl, ConnectSupport {

    /** Common set of breakpoint events. */
    private static CoreBreakpoint.Event[]   breakpointEvents =
                                                new CoreBreakpoint.Event [0];
    /** Array of all registered Debugger Implementations. */
    private ArrayList                       debuggerImpls = new ArrayList ();
    /** is true if some registered debugger supports connect action */
    private boolean                         supportsConnect = false;
    /** is true if some registered debugger supports expressions */
    private boolean                         supportsExpressions = false;
    /** stores DebugerImpl for each debugger returned by createDebugger (DebuggerImpl) */
    private WeakHashMap                     debuggerToImpl = new WeakHashMap ();
    
    
    // DebuggerImpl methods ....................................................
    
    {
        Register.addRegisterListener (new Register.RegisterListener () {
            public void debuggerImplAdded (DebuggerImpl di) {
                registerDebuggerImpl (di);
            }
            public void debuggerImplRemoved (DebuggerImpl di) {
                unregisterDebuggerImpl (di);
            }
        });
    }
    
    /**
     * Returns displayable name of debugger.
     *
     * @return Connect Panel for this version of debugger.
     */
    public String getDisplayName () {
        return DelegatingDebugger.getLocString ("CTL_Debugger_version");
    }

    /**
     * Returns priority, which serves to determine the default debugger.
     *
     * @return Connect Panel for this version of debugger.
     */
    public int getVersionPriority () {
        return 0; // should not be called
    }

    /**
     * Returns true if current Debugger Implementation supports ConnectAction.
     *
     * @return Connect Panel for this version of debugger.
     */
    public boolean supportsConnectAction () {
        return supportsConnect;
    }

    /**
     * Returns panel for ConectAction dialog or null.
     *
     * @return panel for ConectAction dialog or null.
     */
    public JComponent getConnectPanel () {
        if (!supportsConnect) return null;
        return new ConnectorPanel (getDebuggerImpls ());
    }

    /**
     * True, if debugger supports evaluation of expressions.
     */
    public boolean supportsExpressions () {
        return supportsExpressions;
    }

    /**
     * Each Debugger Implementation can define its own set of breakpoint
     * types (= Breakpoint Events). Set of Breakpoint Events should be 
     * static. This property can be set invoking {@link #setBreakpointEvents} method.
     * Default implementation returns empty field.
     *
     * @return set of Breakpoint Events of this Debugger Implementation
     */
    public CoreBreakpoint.Event[] getEvents () {
        return breakpointEvents;
    }

    /**
     * Returns breakpoint event with given name.
     * 
     * @return breakpoint event with given name
     * @see #getBreakpointEvents
     */
/*    public CoreBreakpoint.Event getEvent (String name) {
        CoreBreakpoint.Event[] e = getEvents ();
        int i, k = e.length;
        try {
            for (i = 0; i < k; i++)
                if (e [i].getTypeName ().equals (name))
                    return (CoreBreakpoint.Event) e [i].getClass ().newInstance ();
        } catch (Exception ex) {
        }         
        return null;
    }*/

    /**
     * Returns actions available for this version of debugger.
     */
    public CoreBreakpoint.Action[] getActions () {
        return new CoreBreakpoint.Action[0];
    }
    
    /**
     * Returns a new instance of Debugger.
     */
    public Debugger createDebugger () {
        if (debuggerImpls.size () < 1) return null;
        DebuggerImpl di = (DebuggerImpl) debuggerImpls.get (0);
        return createDebugger (di);
    }
    
    /**
     * Returns DebuggerImpl instance for given Debugger instance.
     *
     * @return DebuggerImpl instance for given Debugger instance
     */
    public DebuggerImpl getDebuggerImpl (Debugger debugger) {
        return (DebuggerImpl) debuggerToImpl.get (debugger);
    }

    /**
     * Returns current debugger.
     */
    public Debugger getCurrentDebugger () {
        try {
            DelegatingDebugger d = (DelegatingDebugger) TopManager.getDefault ().getDebugger ();
            return d.getCurrentDebugger ();
        } catch (DebuggerNotFoundException e) {
        }
        return null;
    }
    
    
    // other methods ...........................................................
    
    
    /**
     * Returns a new instance of Debugger.
     */
    public Debugger createDebugger (DebuggerImpl di) {
        if (di == null) return createDebugger ();
        Debugger d = (Debugger) gDebuggerInfoToDebugger.get (di);
        if (d != null) return d;
        d = di.createDebugger ();
        debuggerToImpl.put (d, di);
        return d;
    }
    
    public static DelegatingDebuggerImpl getDelegatingImpl () {
        return (DelegatingDebuggerImpl) Register.getDebuggerCoreImpl ();
    }
    
    public List getDebuggerImpls () {
        return Collections.unmodifiableList (debuggerImpls);
    }
    
    /**
     * Registers a new Debugger Implementation. 
     * Should be called from ModuleInstall.restore () for each Debugger 
     * Implementations.
     */
    private void registerDebuggerImpl (DebuggerImpl debuggerImpl) {
        debuggerImpls.add (debuggerImpl);
        if (debuggerImpl instanceof EventsProducer) 
            updateEventSet ();
        else
            registerGenericDebugger (debuggerImpl);
        updateConnectState ();
    }

    /**
    * Unregisters Debugger Implementation.
    */
    private void unregisterDebuggerImpl (DebuggerImpl debuggerImpl) {
        debuggerImpls.remove (debuggerImpl);
        if (debuggerImpl instanceof EventsProducer) 
            updateEventSet ();
        else
            unregisterGenericDebugger (debuggerImpl);
        updateConnectState ();
    }
    
    /**
     * Updates state of supportsConnect&supportsExpressions properties,
     * and ConnectAction.
     */
    private void updateConnectState () {
        supportsConnect = false;
        int i, k = debuggerImpls.size ();
        for (i = 0; i < k; i++) {
            DebuggerImpl impl = (DebuggerImpl) debuggerImpls.get (i);
            if (impl instanceof ConnectSupport)
                supportsConnect = true;
        }
        try {
            Debugger d = TopManager.getDefault().getDebugger();
            ConnectAction ca = (ConnectAction) ConnectAction.get (ConnectAction.class);
            ca.setEnabled (supportsConnect && (d.getState() == Debugger.DEBUGGER_NOT_RUNNING));
        } catch (DebuggerNotFoundException e) {
        }
    }
    
    private void updateEventSet () {
        HashMap breakpointEventsRegister = new HashMap ();
        int i;
        for (i = debuggerImpls.size () - 1; i >= 0; i--) {
            DebuggerImpl impl = (DebuggerImpl) debuggerImpls.get (i);
            if (!(impl instanceof EventsProducer))
                continue;
            CoreBreakpoint.Event[] br = ((EventsProducer) impl).
                getEvents ();
            int j, jj = br.length;
            for (j = 0; j < jj; j++) {
                CoreBreakpoint.Event old = (CoreBreakpoint.Event) 
                    breakpointEventsRegister.get (br [j].getTypeName ());
                if ( (old == null) ||
                     (old.getProperties ().length < br [j].getProperties ().length )
                )
                    breakpointEventsRegister.put (br [j].getTypeName (), br [j]);
            }
        }
        int k = breakpointEventsRegister.size ();
        breakpointEvents = new CoreBreakpoint.Event [k];
        Iterator it = breakpointEventsRegister.values ().iterator ();
        for (i = 0; i < k; i++)
            breakpointEvents [i] = (CoreBreakpoint.Event) it.next ();
        
        AddBreakpointAction aba = (AddBreakpointAction) AddBreakpointAction.get (AddBreakpointAction.class);
        aba.setEnabled (k > 0);
    }
    
    HashMap gDebuggerInfoToDebugger = new HashMap ();
    PropertyChangeListener breakpointListener;
    
    private void registerGenericDebugger (DebuggerImpl impl) {
        Debugger debugger = impl.createDebugger ();
        gDebuggerInfoToDebugger.put (impl, debugger);
        if (breakpointListener == null)
            breakpointListener = new BreakpointListener (debugger);
        debugger.addPropertyChangeListener (breakpointListener);
    }
    
    private void unregisterGenericDebugger (DebuggerImpl impl) {
        Debugger debugger = (Debugger) gDebuggerInfoToDebugger.remove (impl);
        debugger.removePropertyChangeListener (breakpointListener);
    }
    
    
    class BreakpointListener implements PropertyChangeListener {
        
        private Debugger debugger;
        
        BreakpointListener (Debugger debugger) {
            this.debugger = debugger;
        }
        
        public void propertyChange (PropertyChangeEvent e) {
            if ( (e.getPropertyName () != null) &&
                 (e.getPropertyName ().equals (Debugger.PROP_BREAKPOINTS))
            ) {
                try {
                    DelegatingDebugger dd = (DelegatingDebugger) TopManager.
                        getDefault ().getDebugger ();
                    dd.refreshBreakpoints (debugger);
                } catch (DebuggerNotFoundException ex) {
                }
                return;
            }
            if ( (e.getPropertyName () != null) &&
                 (e.getPropertyName ().equals (Debugger.PROP_WATCHES))
            ) {
                try {
                    DelegatingDebugger dd = (DelegatingDebugger) TopManager.
                        getDefault ().getDebugger ();
                    dd.refreshWatches (debugger);
                } catch (DebuggerNotFoundException ex) {
                }
                return;
            }
        }
    }
}

