/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.delegator;

import java.beans.*;

import org.openide.TopManager;
import org.openide.util.NbBundle;
import org.openide.debugger.Watch;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.DebuggerException;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.util.Validator;

import org.netbeans.modules.debugger.support.nodes.WatchDelegate;


/**
* Core watch delegates all functionality on inner instance of watch.
*
* @author Jan Jancura
*/
public class DelegatingWatch extends AbstractWatch implements VariablesProducer,
WatchDelegate, Validator.Object {

    static final long serialVersionUID = -8584610162523086115L;

    // variables ...............................................................

    /** Delegating instance of watch. */
    private transient Watch                   watch;
    private transient Debugger                oldDebugger;
    private String                            expression;
    private String                            errorMessage;
    private transient PropertyChangeSupport   pcs;
    private boolean                           isHidden = false;
    private Debugger                          coreDebugger;


    // init ....................................................................

    /** Creates new DelegatingWatch */
    public DelegatingWatch (boolean hidden) {
        this.isHidden = hidden;
        init ();
    }

    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        init ();
    }

    protected void init () {
        pcs = new PropertyChangeSupport (this);
        try {
            coreDebugger = TopManager.getDefault ().getDebugger ();
        } catch (DebuggerNotFoundException e) {
            e.printStackTrace (); // DelegatingWatch does not work in this case!
            return;
        }
        ((ValidatorHolder) coreDebugger).getValidator ().add (this);
    }
    

    // Watch implementation ....................................................

    public void validate () {
        Debugger newDebugger = ((CoreDebugger) coreDebugger).
            getCurrentDebugger ();

        if (coreDebugger.getState () == Debugger.DEBUGGER_NOT_RUNNING) {
            // debugger stopped
            setInnerWatch (null);
            oldDebugger = null;
            setError (NbBundle.getBundle (DelegatingWatch.class).
                getString ("EXC_No_session"));
            firePropertyChange (null, null, null);
            return;
        }
        
        if ( (newDebugger == null) && 
             (coreDebugger.getState () != Debugger.DEBUGGER_STOPPED)
        ) {
            // debugger running & no current session
            setInnerWatch (null);
            setError (NbBundle.getBundle (DelegatingWatch.class).
                getString ("CTL_No_context"));
            firePropertyChange (null, null, null);
            return;
        }
        
        errorMessage = null; 
        if ( (oldDebugger == null) && (newDebugger == null)) {
            setInnerWatch (null);
            firePropertyChange (null, null, null);
            return;
        }
        
        if ( (oldDebugger != null) && 
             (newDebugger != null) &&
             oldDebugger.equals (newDebugger)
        ) {
            if (watch != null) {
                if (watch instanceof AbstractWatch)
                    ((AbstractWatch) watch).validate (); //PENDING we should listen on it!!!
            } else 
                setInnerWatch (newDebugger.createWatch (expression, true));
        } else {
            if (newDebugger == null)
                setInnerWatch (null);
            else
                setInnerWatch (newDebugger.createWatch (expression, true));
        }
        oldDebugger = newDebugger;
        firePropertyChange (null, null, null);
    }

    /**
    * @return false, watch cannot be removed from validator when debugger is finished
    */
    public boolean canRemove () {
        return false;
    }

    /**
    * Returns the name of this variable.
    *
    * @return the name of this variable.
    */
    public String getVariableName () {
        return expression;
    }

    /**
    * Returns string representation of type of this variable.
    *
    * @return string representation of type of this variable.
    */
    public String getType () {
        if (watch != null) return watch.getType ();
        return ""; // NOI18N
    }

    /**
     * Getter for textual representation of the value. It converts
     * the value to a string representation. So if the watch represents
     * null reference, the returned string will be for example "null".
     * That is why null can be returned when the watch is not valid
     *
     * @return the value of this watch or null if the watch is not in the scope
     */
    public String getAsText () {
        if (watch != null) return watch.getAsText ();
        return ""; // NOI18N
    }

    /**
    * Setter that allows to change value of the watched variable.
    *
    * @param value text representation of the value
    * @exception DebuggerException if the value cannot be changed or the
    *    string does not represent valid value
    */
    public void setAsText (String value) throws DebuggerException {
        if (watch != null) watch.setAsText (value);
        firePropertyChange (null, null, null);
    }

    /**
    * Returns human-readable text info about currently returned variables 
    * or null. This text should be displayed with list of variables. 
    *
    * @return human-readable text info about currently returned variables 
    * or null
    */
    public String getInfo () {
        if ( (watch != null) &&
             (watch instanceof VariablesProducer)
        ) return ((VariablesProducer) watch).getInfo ();
        return null;
    }
    
    /**
    * If this AbstractVariable object represents instance of some class or array this method
    * returns variables (static and non-static) of this object.
    *
    * @return variables (static and non-static) of this object.
    */
    public AbstractVariable[] getVariables () {
        if ( (watch != null) &&
             (watch instanceof VariablesProducer)
        ) return ((VariablesProducer) watch).getVariables ();
        return new AbstractVariable [0];
    }
    
    
    /**
    */
    public AbstractVariable[] getFilteredVariables () {
        if ( (watch != null) &&
             (watch instanceof VariablesProducer)
        ) return ((VariablesProducer) watch).getFilteredVariables ();
        return new AbstractVariable [0];
    }

    
    /**
     * If this AbstractVariable object represents instance of some array this method
     * returns (a part of) its items. This method does not check if the given interval
     * is inside the array!
     *
     * @param startindex index of the first item to be returned
     * @param numberofitems number of items to be returned
     * @return array of the items of the array
     */
/*    public AbstractVariable[] getFields (int startindex, int numberofitems) {
        if ( (watch != null) &&
             (watch instanceof VariableProducer)
        ) return ((VariableProducer) watch).getFields (startindex, numberofitems);
        return new AbstractVariable [0];
    }*/
    
    /**
     * If this AbstractVariable object represents instance of some array this method
     * returns number of its items.
     *
     * @return number of items of the array
     */
   /* public int getFieldsNumber () {
        if ( (watch != null) &&
             (watch instanceof VariableProducer)
        ) return ((VariableProducer) watch).getFieldsNumber ();
        return 0;
    }*/

    /**
    * Returns true if this variable hasn't any fields.
    *
    * @return True if this variable hasn't any fields.
    */
    public boolean isLeaf() {
        return false;
    }

    /**
    * Returns error message if watch cannot be resolved or null.
    *
    * @return AbstractVariable object for this class.
    */
    public String getErrorMessage () {
        if (watch == null)
            return errorMessage;
        if (watch instanceof AbstractWatch) 
            return ((AbstractWatch) watch).getErrorMessage ();
        return null;
    }

    /** Set the variable name to watch.
     *
     * @param name string name of the variable to watch
     */
    public void setVariableName (String name) {
        expression = name;
        if (watch != null) watch.setVariableName (name);
        validate ();
    }

    /**
    * Sets error message for this watch.
    */
    protected void setError (String description) {
        errorMessage = description;
    }

    /** Remove the watch from the list of all watches in the system.
     */
    public void remove () {
        try {
            ((DelegatingDebugger) TopManager.getDefault ().getDebugger ()).
                removeWatch (this);
        } catch (DebuggerNotFoundException e) {
            return;
        }
    }

    /** Getter to test if the property is "hidden". If a property is hidden it
    * is not presented in the list of all properties. Such a property can be used
    * for private usage of the IDE, not displaying anything to user.
    * <P>
    * To create hidden watch, use <CODE>Debugger.createWatch ("name", true)</CODE> method.
    *
    * @return true if the watch is hidden, false otherwise
    */
    public boolean isHidden () {
        return isHidden;
    }

    public synchronized void addPropertyChangeListener (PropertyChangeListener listener) {
        pcs.addPropertyChangeListener (listener);
    }

    public synchronized void removePropertyChangeListener (PropertyChangeListener listener) {
        pcs.removePropertyChangeListener (listener);
    }

    
    // other methods ...........................................................

    public Watch getInnerWatch () {
        return watch;
    }

    public void setInnerWatch (Watch w) {
        if (w == watch) return;
        Watch o = watch;
        if (watch != null) watch.remove ();
        watch = w;
        firePropertyChange (PROP_INNER_WATCH, o, watch);
    }
    
    protected void firePropertyChange (String s, Object o, Object n) {
        pcs.firePropertyChange (s, o, n);
    }
}

