/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.multisession;


import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.Reader;
import java.io.Writer;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.ArrayList;
import java.util.StringTokenizer;

import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.execution.ProcessExecutor;
import org.openide.execution.NbProcessDescriptor;
import org.openide.execution.NbClassPath;
import org.openide.execution.ExecInfo;
import org.openide.util.HelpCtx;
import org.openide.util.MapFormat;
import org.openide.util.NbBundle;
import org.openide.util.Utilities;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;

import org.netbeans.modules.debugger.support.util.Utils;

/** Executes a class externally (in a separate process). Provides
* basic implementation that allows to specify the process to 
* execute, its parameters and also to substitute the content of repositorypath,
* classpath, bootclasspath and librarypath. This is done by inner class Format.
* <P>
* The behaviour described here can be overriden by subclasses to use different
* format (extend the set of recognized tags), execute the 
* process with additional environment properties, etc.
*
* @author Jan Jancura
*/
public class DebuggerExecutor extends ProcessExecutor {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 1440216775212461457L;

    /** Property name of the debuggerExecutorType property */
    public static String PROP_DEBUGGER_EXECUTOR_TYPE = "debuggerExecutorType";
    
    /** Property name of the classic switch property */
    public static String PROP_CLASSIC_SWITCH = "classicSwitch";
    
    /** default descriptor to use */
    static final NbProcessDescriptor [] DEFAULT;
    
    static String [] description;
    
    private static boolean classicPreferred;
    
    static {
        classicPreferred = Utils.classicPreferred (System.getProperty ("jdk.home") + // NOI18N
            File.separatorChar + "bin" + File.separatorChar + "java"); // NOI18N
        description = new String [2];
        description [0] = NbBundle.getBundle (DebuggerExecutor.class).getString ("CTL_JPDADebuggerExecutorName");
        description [1] = NbBundle.getBundle (DebuggerExecutor.class).getString ("CTL_JDK11DebuggerExecutorName");
        
        DEFAULT = new NbProcessDescriptor [2];
        DEFAULT [1] = new NbProcessDescriptor (
                      ( // /usr/local/bin/java
                          "{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}bin{" + Format.TAG_SEPARATOR + "}java"
                      ), ( //-Xdebug -Djava.compiler=NONE -Xbootclasspath:{BOOTCLASSPATH} -cp {REPOSITORY}:{CLASSPATH} {CLASSNAME} {ARGUMENTS}
                          "{" + Format.TAG_CLASSIC_SWITCH + "}" +
                          "-Xdebug -Djava.compiler=NONE -Xbootclasspath:{" + Format.TAG_BOOTCLASSPATH + "} " +
                          "-cp {" + Format.TAG_REPOSITORY + "} " +
                          /*"{" + Format.TAG_PATHSEPARATOR + "}" + 
                          "{" + Format.TAG_CLASSPATH + "}" +
                          "{" + Format.TAG_PATHSEPARATOR + "}" + 
                          "{" + Format.TAG_LIBRARY + "} " +*/
                          "{" + Format.TAG_CLASSNAME + "} " +
                          "{" + Format.TAG_ARGUMENTS + '}'
                      ),
                      new ProcessExecutor ().getExternalExecutor ().getInfo ()
                  );
        DEFAULT [0] = new NbProcessDescriptor (
                      ( // /usr/local/bin/java
                          "{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}bin{" + Format.TAG_SEPARATOR + "}java"
                      ), ( //-Xdebug -Xnoagent -Xrunjdwp:transport=dt_socket,suspend=n -Djava.compiler=NONE -Xbootclasspath:{BOOTCLASSPATH} -cp {REPOSITORY}:{CLASSPATH} {CLASSNAME} {ARGUMENTS}
                          "{" + Format.TAG_CLASSIC_SWITCH + "}" +
                          "-Xdebug -Xnoagent -Xrunjdwp:transport=dt_socket,server=y,suspend=n -Djava.compiler=NONE -Xbootclasspath:{" 
                          + Format.TAG_BOOTCLASSPATH + "} " +
                          "-cp {" + Format.TAG_REPOSITORY + "} " +
                          /*"{" + Format.TAG_PATHSEPARATOR + "}" + "{" + Format.TAG_CLASSPATH + "}" +
                          "{" + Format.TAG_PATHSEPARATOR + "}" + "{" + Format.TAG_LIBRARY + "} " +*/
                          "{" + Format.TAG_CLASSNAME + "} " +
                          "{" + Format.TAG_ARGUMENTS + '}'
                      ),
                      new ProcessExecutor ().getExternalExecutor ().getInfo ()
                  );   
    }

    // variables ......................................................................................
    
    private NbProcessDescriptor [] debuggerProcess;
    
    private Integer selectedType;
    
    // classic switch value
    private boolean classic;
    // indicates if classic switch is set manually
    private boolean classicSet = false;
    
    // init ...................................................................................

    /** Read the object.
    */
    private void readObject (java.io.ObjectInputStream oos)
    throws java.io.IOException, ClassNotFoundException {
        oos.defaultReadObject ();
        if ((debuggerProcess == null) || (selectedType == null))
            init ();
    }
    
    private void init () {
        selectedType = new Integer (0);
        classicSet = false;
            
        debuggerProcess = new NbProcessDescriptor [DEFAULT.length];
        for (int x = 0; x < DEFAULT.length; x++)
            debuggerProcess [x] = DEFAULT [x];

        String fileSeparator = System.getProperty ("file.separator");
        String pathSeparator = System.getProperty ("path.separator");
        String javaRoot = System.getProperty ("java.home") + fileSeparator;
        String javaRoot1 = System.getProperty ("jdk.home") + fileSeparator;
        setBootClassPath (new NbClassPath (
                              javaRoot1 + "lib" + fileSeparator + "tools.jar" + pathSeparator +
                              getBootClassPath ().getClassPath ()
                          ));
        setName (
            NbBundle.getBundle (DebuggerExecutor.class).getString ("CTL_DebuggerExecutorDisplayName")
        );
        
        setExternalExecutor (debuggerProcess [selectedType.intValue ()]);
    }
    
    /** Create a new debugger executor.
    * The default Java launcher associated with this VM's installation will be used,
    * and the user repository entries will be used for the class path.
    */
    public DebuggerExecutor () {
        init ();
    }

    public Integer getDebuggerExecutorType () {
        debuggerProcess [selectedType.intValue ()] = getExternalExecutor ();
        return selectedType;
    }
    
    public void setDebuggerExecutorType (Integer sel) {
        debuggerProcess [selectedType.intValue ()] = getExternalExecutor ();
        if (sel.intValue () > DEFAULT.length)
            selectedType = new Integer (0);
        else
            selectedType = sel;
        setExternalExecutor (debuggerProcess [selectedType.intValue ()]);
    }

    public boolean isClassic () {
        if (classicSet)
            return classic;
        return classicPreferred;
    }
    
    public void setClassic (boolean classic) {
        classicSet = true;
        this.classic = classic;
    }
    
    // createProcess method overiden - DebuggerExecutor.Format should be passed instead of ProcessExecutor.Format
    protected Process createProcess (ExecInfo info) throws IOException {
        return getExternalExecutor ().exec (new Format (
                                                info,
                                                getClassPath (),
                                                getBootClassPath (),
                                                getRepositoryPath (),
                                                getLibraryPath (),
                                                isClassic ()
                                            ), getEnvironmentVariables (), getWorkingDirectory ());
    }
    
    public HelpCtx getHelpCtx () {
        return new HelpCtx (DebuggerExecutor.class);
    }
    
    // inner classes ..................................................................................
    
    public static class Format extends ProcessExecutor.Format {
        /** Tag replaced with ProcessExecutors.getClassPath () */
        public static final String TAG_CLASSIC_SWITCH = "classic"; // NOI18N
        
        static final long serialVersionUID =1070082399063112986L;
        
        public Format (
            ExecInfo info,
            NbClassPath classPath,
            NbClassPath bootClassPath,
            NbClassPath repository,
            NbClassPath library,
            boolean isClassic
        ) {
            super (info, classPath, bootClassPath, repository, library);
            java.util.Map map = getMap ();
            map.put (TAG_CLASSIC_SWITCH, isClassic ? "-classic " : ""); // NOI18N
        }
        
    }
}

/*
* <<Log>>
*  4    Gandalf   1.3         12/10/99 Jan Jancura     Hints
*  3    Gandalf   1.2         10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  2    Gandalf   1.1         8/2/99   Jan Jancura     
*  1    Gandalf   1.0         7/23/99  Jan Jancura     
* $
*/
