/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.multisession;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.Constructor;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;

import org.openide.TopManager;
import org.openide.DialogDescriptor;
import org.openide.NotifyDescriptor;
import org.openide.actions.StartDebuggerAction;
import org.openide.debugger.*;
import org.openide.text.Line;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallbackSystemAction;
import org.openide.util.actions.CallableSystemAction;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.windows.InputOutput;
import org.openide.src.ConstructorElement;
import org.openide.src.ClassElement;

import org.netbeans.modules.debugger.delegator.*;
import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.DebuggerSettings;

/**
* Enterprise debugger enables running more session of debugger in one time.
*
* @author   Jan Jancura
*/
public class EnterpriseDebugger extends DelegatingDebugger {

    // static ..................................................................
    
    /** generated Serialized Version UID */
    static final long                       serialVersionUID = 1969501498004072694L;

    /** Name of property currentDebugger. Fired when surrent session
        is changed. */
    public static final String              PROP_CURRENT_SESSION =
        "currentSession";
    /** Constant for runMultisession property. */
    public static final int                 RUN_SINGLE = 0;
    /** Constant for runMultisession property. */
    public static final int                 RUN_MULTI = 1;
    /** Constant for runMultisession property. */
    public static final int                 RUN_CANCEL = 2;
    
    /** bundle to obtain text information from */
    private static ResourceBundle           bundle;
    
    /**
    * @return localized string.
    */
    static String getString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (EnterpriseDebugger.class);
        return bundle.getString (s);
    }
    
    
    // variables ...............................................................
    
    /** Current debugger session. */
    private transient Session               currentSession;
    /** Currently running debuggers. */
    private transient ArrayList             debuggers = new ArrayList ();
    /** Stores all currently running (not starting) debuggers.*/
    private transient HashMap               runningDebuggers = new HashMap ();
    private transient Vector                sessionsListener = new Vector (10,20);
    /** Listens on inner istance of debugger. */
    private transient SessionListener       sessionListener = new SessionListener ();
    /** Array of all closed InputOutput. */
    private static ArrayList                ioToClose = new ArrayList ();
    /** Delegating root of thread groups. */
    private transient DelegatingThreadProducer threadProducer = new DelegatingThreadProducer ();
    /** If true the Finish sessions dialog will be showen on debugger finish. */
    protected boolean                       showFinishDialog = true; // OBSOLETE

    private Debugger                        currentDebugger; // PENDING

    // init ....................................................................

    public EnterpriseDebugger () {
    }

    /**
    * Deserializes debugger.
    */
    protected void setDebugger (AbstractDebugger debugger) {
        super.setDebugger (debugger);
        /* 
         All breakpoints should be instances of SessionBreakpoint.
         If there are some DelegatingBreakpoints (it is possible after CE -> IE update),
         remove them and create new corresponding SessionBreakpoints.
        */
        Breakpoint[] bs = getBreakpoints ();
        for (int i = 0; i < bs.length; i++) {
            if (bs [i] instanceof DelegatingBreakpoint) {
                bs [i].remove ();
                ((CoreBreakpoint) createBreakpoint (false)).setEvent (
                    ((DelegatingBreakpoint) bs [i]).getEvent (this)
                );
            }
        }
    }

    // Debugger implementation .................................................

    /** Starts the debugger. The method takes information from the provided
    * info (containing the class to start and
    * arguments to pass it and name of class to stop debugging in) and starts
    * new debugging session.
    *
    * @param info debugger info about class to start
    * @exception DebuggerException if an error occures during the start of the debugger
    */
    public void startDebugger (DebuggerInfo info) throws DebuggerException {
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).setMultisession (true);
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).changeEnabled (true);

        // open Start Session Dialog
        int action = RUN_MULTI;
        if (info != null) {
            if (debuggers.size () > 0)
                action = showStartSessionDialog ();
        }

        if (action == RUN_CANCEL) {
            Debugger d = getCurrentDebugger ();
            if (d == null)
                d = getDebuggers () [0];
            setState (DEBUGGER_STARTING);
            setState (d.getState ());
            return;
        }
        if (action == RUN_SINGLE) {
            // start sleeping sessions
            int i = debuggers.size () - 1;
            for (; i >= 0; i--) {
                Session s = (Session) debuggers.get (i);
                s.finishIn ();
            }
        } else {
            // start sleeping sessions
            int i = debuggers.size () - 1;
            for (; i >= 0; i--) {
                Session s = (Session) debuggers.get (i);
                s.start ();
            }
        }

        // close old tabs in Output window
        int i, k = ioToClose.size ();
        for (i = 0; i < k; i++)
            ((InputOutput) ioToClose.get (i)).closeInputOutput ();
        ioToClose = new ArrayList ();

        // start current session
        if (info != null)
            startSession (info);

        // ?!?! why twice Dan?!?!
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).setMultisession (true);
        ((StartDebuggerAction) SystemAction.get (StartDebuggerAction.class)).changeEnabled (true);
    }

    /**
    * Finishes debugger.
    */
    public void finishDebugger () throws DebuggerException {
        finishDebuggerIn ();
    }

    // BREAKPOINTS .......................................................

    /** Creates new breakpoint.
    *
    * @return new line breakpoint
    */
    public CoreBreakpoint createBreakpoint (boolean hidden) {
        //S ystem.out.println ("JavaD.createBreakpoint "); // NOI18N
        SessionBreakpoint b = new SessionBreakpoint (hidden);
        addBreakpoint (b);
        return b;
    }
    
    // AbstractDebugger implementation .........................................

    // properties .......................

    /**
     * Creates new debugger instance.
     */
    protected Debugger createDebugger (DebuggerInfo info) {
        // get debugger instance
        Debugger debugger = super.createDebugger (info);
//        if (debugger instanceof AbstractDebugger)
//            ((AbstractDebugger) debugger).setGeneralIOName (false);
        return debugger;
    }

    /**
    * Sets debugger state.
    */
    public void setSuspended (boolean suspended) {
/*        int i = debuggers.size () - 1;
        for (; i >= 0; i--)
            ((Session) debuggers.get (i)).getDebugger ().setSuspended (suspended);*/
    }

    /**
    * Returns version of this debugger.
    */
    public String getVersion () {
        return NbBundle.getBundle (EnterpriseDebugger.class).
            getString ("CTL_Debugger_version");
    }

    /**
    * Returns priority, which serves to determine the default debugger.
    */
    
    public int getVersionPriority () {
        return 0;
    }

    /**
    * @return name of proces for given DebuggerInfo
    */
    public String getProcessName (DebuggerInfo info) {
        return "???"; // NOI18N nobody calls this method
    }

    /**
    * @return name of location for given DebuggerInfo
    */
    public String getLocationName (DebuggerInfo info) {
        return "???"; // NOI18N nobody calls this method
    }


    // DelegatingDebugger method ...............................................

    /**
    * @return delegating debugger instance.
    */
    public Debugger getInnerDebugger () {
        return getCurrentDebugger ();
    }


    // other methods ...........................................................

    // properties .....................


    /**
    * Returns debugger for curent session.
    */
    public Debugger getCurrentDebugger () {
        if (currentSession == null) 
            return currentDebugger; // PENDING
            // return null;
        return currentSession.getDebugger ();
    }

    /**
    * @return current session.
    */
    public Session getCurrentSession () {
        return currentSession;
    }

    /**
    * Sets given session current.
    */
    public void setCurrentSession (Session s) {
        if (s == null) {
            Session old = currentSession;
            currentSession = null;
            firePropertyChange (PROP_CURRENT_SESSION, old, currentSession);
            return;
        }
        if (s.getCurrentThread () != null)
            s.getCurrentThread ().setCurrent (true);
        else
            setCurrentDebugger (s.getDebugger ());
    }


    // sessions .........................

    /**
    * Returns all currently availabed sessions.
    */
    public Session[] getSessions () {
        Session[] s = new Session [debuggers.size ()];
        return (Session[]) debuggers.toArray (s);
    }

    /**
     * Returns currently running debugger or null if debugger is not running.
     */
    public Debugger[] getDebuggers () {
        ArrayList ss = (ArrayList) debuggers.clone ();
        int i, k = ss.size ();
        Debugger[] ds = new Debugger [k];
        for (i = 0; i < k; i++)
            ds [i] = ((Session) ss.get (i)).getDebugger ();
        return ds;
    }
    
    /**
    * Returns session for gien debugger instance.
    */
    public Session findSession (Debugger debugger) {
        int i, k = debuggers.size ();
        for (i = 0; i < k; i++)
            if (((Session)debuggers.get (i)).getDebugger ().equals (debugger))
                return (Session)debuggers.get (i);
        return null;
    }

    // other .........................

    /**
    * Creates a new debugger session for given debugger type.
    */
    public void startSession (DebuggerInfo info) throws DebuggerException {
        if (getState () == DEBUGGER_NOT_RUNNING)
            setState (DEBUGGER_STARTING);

        // get debugger instance
        Debugger debugger = createDebugger (info);

        debugger.addPropertyChangeListener (sessionListener);
        currentDebugger = debugger; // PENDING
        debugger.startDebugger (info);
        currentDebugger = null; // PENDING
        if (debugger.getState () == DEBUGGER_NOT_RUNNING) {
            debugger.removePropertyChangeListener (sessionListener);
            return;
        }
        if (debugger instanceof ThreadsRoot) 
            threadProducer.setRemoteThreadGroup (
                ((ThreadsRoot) debugger).getThreadsRoot ()
            );
        if (info instanceof AbstractDebuggerInfo)
            addSession (new Session (
                ((AbstractDebuggerInfo) info).getProcessName (),
                ((AbstractDebuggerInfo) info).getLocationName (),
                debugger,
                info
            ));
        else
            addSession (new Session (
                "process",
                "localhost",
                debugger,
                info
            ));
    }


    // private helper methods ....................................................

    private int showStartSessionDialog () {
        final DebuggerSettings debuggerSettings = (DebuggerSettings) 
            DebuggerSettings.findObject (DebuggerSettings.class);
        if (!debuggerSettings.getShowRunMultisessionDialog ()) {
            return debuggerSettings.getRunMultisession () ? RUN_MULTI : RUN_SINGLE;
        }
        String msg = null;
        String title = getString ("TITLE_Start_session_dialog");
        if (debuggers.size () > 1)
            msg = getString ("MSG_Start_session_dialog");
        else {
            String s = getSessions () [0].getSessionName ();
            msg = new MessageFormat (
                getString ("MSG_Start_second_session_dialog")
            ).format (new Object[] {s});
        } 
        
        JPanel p = new JPanel ();
        p.setLayout (new BorderLayout (0, 8));
            JPanel p1 = new JPanel ();
            StringTokenizer st = new StringTokenizer (msg, "\n");
            int i, k = st.countTokens ();
            p1.setLayout (new GridLayout (k, 1));
            for (i = 0; i < k; i++)
                p1.add (new JLabel (st.nextToken ()));
        p.add (p1, BorderLayout.CENTER);
            JCheckBox chb = new JCheckBox ();
            chb.setText (getString ("CTL_DNSTDNT"));//"Do not show this dialog next time."
            chb.setMnemonic (getString ("CTL_DNSTDNT_Mnemonic").charAt (0));
            p.add (chb, BorderLayout.SOUTH);

        NotifyDescriptor.Confirmation descriptor = new NotifyDescriptor.Confirmation (
            p,
            title
        );
        descriptor.setOptions (new Object[] {
            getString ("CTL_FinishAndStart"),
            getString ("CTL_StartNewSession"),
            NotifyDescriptor.CANCEL_OPTION
        });
        Object o = TopManager.getDefault ().notify (descriptor);
        debuggerSettings.setShowRunMultisessionDialog (!chb.isSelected ());
        if (o.equals (getString ("CTL_FinishAndStart"))) {
            debuggerSettings.setRunMultisession (false);
            return RUN_SINGLE;
        }
        if (o.equals (getString ("CTL_StartNewSession"))) {
            debuggerSettings.setRunMultisession (true);
            return RUN_MULTI;
        }
        return RUN_CANCEL;
    }
    
    /**
    * Finishes debugger.
    */
    boolean finishDebuggerIn () throws DebuggerException {
        //S ystem.out.println("EnterpriseDebugger.finishDebugger " );
        if (debuggers.size () > 0) {

            // show dialog "Finish debugging sessions" // NOI18N
            DebuggerSettings debuggerSettings = (DebuggerSettings) 
                DebuggerSettings.findObject (DebuggerSettings.class);
            if (debuggerSettings.getShowFinishDialog ())
                if (!showFinishDialog ()) return false; // Cancel

            // finish all non persistent sessions
            int i = debuggers.size () - 1;
            for (; i >= 0; i--) {
                Session s = (Session) debuggers.get (i);
                s.finish ();
            }
        }
        debuggerNotRunning ();
        return true;
    }
    
    private boolean showFinishDialog () {
        final Session[] sessions = (Session[]) debuggers.toArray (new Session [debuggers.size ()]);
        final Dialog[] d = new Dialog [1];
        final FinishActionPanel panel = new FinishActionPanel ();
        panel.setSessions (sessions);
        final DebuggerSettings debuggerSettings = (DebuggerSettings) 
            DebuggerSettings.findObject (DebuggerSettings.class);
        panel.setShowFinishDialog (debuggerSettings.getShowFinishDialog ());
        final boolean[] ok = new boolean [1];
        ok [0] = false;
        DialogDescriptor descriptor = new DialogDescriptor (
            panel,
            getString ("CTL_Finish_debugging_dialog"),
            true,
            new ActionListener () {
                public void actionPerformed (ActionEvent e) {
                    if (DialogDescriptor.OK_OPTION.equals (e.getSource ())) {
                        boolean[] state = panel.getState ();
                        int j, jj = state.length;
                        for (j = 0; j < jj; j ++)
                        sessions [j].setPersistent (state [j]);
                        ok [0] = true;
                        debuggerSettings.setShowFinishDialog (
                            panel.getShowFinishDialog ()
                        );
                    }
                    d [0].dispose ();
                }
            }
        );
        descriptor.setHelpCtx (new HelpCtx (DelegatingDebugger.class.getName () + ".dialog")); // NOI18N
        d [0] = TopManager.getDefault ().createDialog (descriptor);
        d [0].setSize (300, 300);
        d [0].show ();
        return ok [0];
    }

    /**
    * Sets given debugger current.
    */
    void setCurrentDebugger (Debugger debugger) {

        // change current session
        if ( (debugger != getCurrentDebugger ()) &&
                (getCurrentDebugger () != null)
           ) {
            // deselect old current thread
            Debugger d = getCurrentDebugger ();
            AbstractThread tt = null;
            if (d instanceof ThreadsRoot)
                tt = ((ThreadsRoot) d).getCurrentThread ();
            if (tt != null)
                tt.setCurrent (false);
        }

        // set new one
        Session old = currentSession;
        currentSession = findSession (debugger);
        firePropertyChange (PROP_CURRENT_SESSION, old, currentSession);

        setState (debugger.getState ());
    }

    /**
    * Instantiates debugger of given type.
    */
/*    protected Debugger createDebugger (Class debuggerClass) {
        try {
            Constructor c = debuggerClass.getConstructor (
                new Class [] {
                    Boolean.TYPE,
                    Validator.class
                }
            );
            AbstractDebugger debugger =  (AbstractDebugger) c.newInstance (
                new Object[] {
                    new Boolean (true),
                    getValidator ()
                }
            );
            return debugger;
        } catch (Exception e) {
            return null;
        } catch (Error e) {
            return null;
        }
    }*/

    /**
    * Runns debugger when some hidden session stops on breakpoint.
    *
    * @return true if debugger gets out from bed...
    */
    boolean wakeUpDebugger (AbstractDebugger d) {
        // wake up from hidden state
        Session s = findSession (d);
        if (TopManager.getDefault ().notify (
                new NotifyDescriptor.Confirmation (
                    new MessageFormat (
                        getString ("CTL_Hidden_debugger_on_breakpoint")
                    ).format (new Object [] {
                        s.getSessionName ()
                    }),
                    NotifyDescriptor.YES_NO_OPTION
                )
            ) == NotifyDescriptor.YES_OPTION
        ) {
            EnterpriseModule.changeWorkspace ();
            setState (DEBUGGER_STARTING);
            setState (DEBUGGER_RUNNING);
            try {
                startDebugger (null);
            } catch (DebuggerException ex) {
            }
            return true;
        }
        return false;
    }

    /**
    * Adds new session to the list "debuggers" and fires it.
    */
    private void addSession (Session s) {
        debuggers.add (s);
        setCurrentSession (s);
        fireSessionCreated (s);
    }

    /**
    * Removes session from the list "debuggers" and fires it.
    */
    void removeSession (Session s) {
        if (s.getDebugger () != null)
            s.getDebugger ().removePropertyChangeListener (sessionListener);
//        s.getDebugger ().println (getString ("CTL_Debugger_session_end"), ERR_OUT);
        debuggers.remove (s);
        fireSessionDeath (s);
    }

    /**
    * Called when some debugger instance is running first time.
    */
    protected void firstDebuggerRunning (AbstractDebugger debugger) {
        //    updateWatches ()
        /*
        CallbackSystemAction a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
        a.setActionPerformer (new ActionPerformer () {
            public void performAction (SystemAction a2) {
                setSuspended (true);
            }
        });
        a = (CallbackSystemAction) SystemAction.get (ResumeDebuggerAction.class);
        a.setActionPerformer (new ActionPerformer () {
            public void performAction (SystemAction a3) {
                setSuspended (false);
            }
        });
         */
    }

    /**
    * Called when some debugger instance is running first time.
    */
    protected void debugerRunning (AbstractDebugger debugger) {
        addEvents (debugger);
    }

    /**
    * Called when last debugger session is finished.
    */
    protected void debuggerNotRunning () {
        //S ystem.out.println("EnterpriseDebugger.debuggerNotRunning " );

        CallbackSystemAction a = (CallbackSystemAction) SystemAction.get (SuspendDebuggerAction.class);
        a.setActionPerformer (null);
        a = (CallbackSystemAction) SystemAction.get (ResumeDebuggerAction.class);
        a.setActionPerformer (null);

        try {
            super.finishDebugger ();
        } catch (DebuggerException e) {
            TopManager.getDefault ().notify (
                new NotifyDescriptor.Exception(
                    e.getTargetException () == null ? e : e.getTargetException (),
                    getString ("EXC_Debugger") + ": " + e.getMessage ())
            );
        }
        runningDebuggers = new HashMap ();
        TopManager.getDefault ().setStatusText (getString ("CTL_Debugger_end"));
    }

    /**
    * Adds breakpoints to the new instance of debugger, when a new session 
    * is started.
    */
    private void addEvents (AbstractDebugger debugger) {  //S ystem.out.println("DelegatingDebugger.addEvents " + debugger); // NOI18N
        Breakpoint[] b = getBreakpoints ();
        int i, k = b.length;  //S ystem.out.println("DelegatingDebugger.addEvents stand. " + k); // NOI18N
        for (i = 0; i < k; i++)
            ((SessionBreakpoint) b [i]).addDebugger (debugger);
    }

    /**
    * Removes breakpoints from given instance of debugger.
    */
    private void removeEvents (AbstractDebugger debugger) {
        Breakpoint[] b = getBreakpoints ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            ((SessionBreakpoint) b [i]).removeDebugger (debugger);
    }


    // SessionListener support ...................................................

    /**
    * This listener notifikates about cahnges of breakpoints, watches and threads
    *
    * @param l listener object
    */
    public void addSessionsListener (SessionsListener l) {
        sessionsListener.addElement (l);
    }

    /**
    * Removes debugger listener
    *
    * @param l listener object
    */
    public void removeSessionsListener (SessionsListener l) {
        sessionsListener.removeElement (l);
    }

    /**
    * Notificates about creating a session
    *
    * @param threa
    */
    protected void fireSessionCreated (final Session s) {
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                int i, k = sessionsListener.size ();
                for (i = 0; i < k; i++)
                    ((SessionsListener)sessionsListener.elementAt (i)).sessionCreated (s);
            }
        });
    }

    /**
    * Notificates about session death
    *
    * @param thread
    */
    protected void fireSessionDeath (final Session s) {
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                int i, k = sessionsListener.size ();
                for (i = 0; i < k; i++)
                    ((SessionsListener)sessionsListener.elementAt (i)).sessionDeath (s);
            }
        });
    }
    
    /**
    * Fires property change.
    */
    protected void firePropertyChange (String name, Object o, Object n) {
        super.firePropertyChange (name, o, n);
    }

    // innerclasses .............................................................

    /**
    * Listens on each session for current thread changes and debugger state.
    */  
    private class SessionListener implements PropertyChangeListener {
        public void propertyChange (PropertyChangeEvent e) {
            AbstractDebugger d = (AbstractDebugger) e.getSource ();
            int os = getState ();
            int ns = d.getState ();

            if (e.getPropertyName ().equals (PROP_STATE)) {
                //S ystem.out.println ("Enterprise debugger.Session listener: " + d + " state: " + d.getState ());

                if (os == DEBUGGER_NOT_RUNNING) return;

                if ( (ns != DEBUGGER_NOT_RUNNING) &&
                        ( (d == getCurrentDebugger ()) ||
                          (debuggers.size () < 2)
                        )
                   ) {
                    // current session state changed
                    if (os != ns) {
                        setState (ns);
                        if ((os == DEBUGGER_STARTING) && (ns == DEBUGGER_RUNNING)) {
                            firstDebuggerRunning (d);
                        }
                        //            if (ns == DEBUGGER_STOPPED)
                        //              updateWatches ();
                    }
                }

                if (ns == DEBUGGER_RUNNING) {
                    if (!runningDebuggers.containsKey (d)) {
                        // add a new debugger
                        debugerRunning (d);
                        runningDebuggers.put (d, d);
                    }
                }

                if (ns == DEBUGGER_NOT_RUNNING) {
                    // debugger session finish
                    runningDebuggers.remove (d);
                    Session s = findSession (d);
                    //S ystem.out.println("Enterprise debugger.DebuggerNotRunning " + s);
                    removeEvents (d);

                    if ((s != null) && !s.isPersistent ()) {// s == null while debugger starts
//                        InputOutput io = s.getInputOutput ();
//                        if (io != null) ioToClose.add (io);
                        removeSession (s);
//                        if (debuggers.size () > 0)
//                            println (getString ("CTL_Debugger_session_end"), STL_OUT);
                    }
                    if (d == getCurrentDebugger ())
                        setCurrentSession (null);
                    if (debuggers.size () == 0)
                        debuggerNotRunning ();
                }

            } else
            if (e.getPropertyName ().equals (PROP_CURRENT_THREAD)) {

                // some thread from non current debugger sets uncurrent
                if ( (d != getCurrentDebugger ()) &&
                        (((ThreadsRoot) d).getCurrentThread () == null)
                   ) return;

                // firstly - when I sleep, I must wake up!!!
                if ( (os == DEBUGGER_NOT_RUNNING) &&
                        !wakeUpDebugger (d)
                   ) return;

                if (d != getCurrentDebugger ())
                    setCurrentDebugger (d);
                firePropertyChange (ThreadsRoot.PROP_CURRENT_THREAD, e.getOldValue (), e.getNewValue ());
            } else
            if (e.getPropertyName ().equals (PROP_CURRENT_LINE)) {
                if (d != getCurrentDebugger ()) return;
                firePropertyChange (PROP_CURRENT_LINE, e.getOldValue (), e.getNewValue ());
            } else
            if (e.getPropertyName ().equals (PROP_LAST_ACTION)) {
                if (d != getCurrentDebugger ()) return;
                firePropertyChange (PROP_LAST_ACTION, e.getOldValue (), e.getNewValue ());
            }
        }
    }
}
