/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.multisession;

import java.util.Arrays;
import java.util.ArrayList;

import org.openide.TopManager;
import org.openide.debugger.*;
import org.openide.text.Line;
import org.openide.nodes.Node.Property;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.BreakpointSupport;

/**
* Core breakpoint delegates all functionality on inner instance of breakpint.
* @see org.openide.debugger.Breakpoint
*
* @author   Jan Jancura
*/
public class SessionBreakpoint extends BreakpointSupport {
    
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 2246204002781932191L;


    // private variables .....................................................

    private transient ArrayList events = null;
    private boolean doNotFire = false;


    // init ....................................................................

    /**
    * Non public constructor called from the AbstractDebugger only.
    * User must create breakpoint from Debugger.getNewBreakpoint () method.
    */
    SessionBreakpoint (boolean hidden) {
        super (hidden);
    }

    protected void init () {
        super.init ();
    }


    // CoreBreakpoint ..........................................................

    /**
    * Returns current breakpoint event for given instance of debugger (used
    * in multisession debugger implementation).
    */
    public Event getEvent (Debugger d) {
        if (d == debugger)
            return event;
        if (events == null) return null;
        synchronized (events) {
            int i, k = events.size ();
            for (i = 0; i < k; i++) {
                Event e = (Event) events.get (i);
                if (d == e.getDebugger ())
                    return e;
            }
        }
        return null;
    }

    
    // BreakpointSupport .......................................................

    /**
    * Sets or updates breakpoint. Calls set method of currently selected event for all debuggers, sets
    * value of valid property and marks lines in the editor.
    */
    protected void setBreakpoint () {
        if (!isEnabled ()) return;
        boolean valid = false;
        if (event != null) {
            // 1) set / remove
            if (debugger.getState () == AbstractDebugger.DEBUGGER_NOT_RUNNING) {
                disableEvents ();
                events = null;
            } else {
                try {
                    // 1.1) init event
                    if (events == null) {
                        Debugger[] ds = ((EnterpriseDebugger) TopManager.
                            getDefault ().getDebugger ()).getDebuggers ();
                        int i, k = ds.length;
                        for (i = 0; i < k; i++)
                            addEventForDebugger (ds [i]);
                    }
                    // 1.2) set event
                    if (events != null) {
                        ArrayList a = (ArrayList) events.clone ();
                        int i, k = a.size ();
                        for (i = 0; i < k; i++)
                            if (((Event) a.get (i)).set ())
                                valid = true;
                    }
                } catch (DebuggerException e) {
                }
            }
            // 2) remark
            Line[] l = event.getLines ();
            if (l != null)
                remark (new ArrayList (Arrays.asList (l)));
            else
                remark (null);
        }
        // 3) setValid
        setValid (valid);
    }

    /**
    * Removes breakpoint from this position.
    * Clears Line and breakpoint from RemoteDebugger.
    */
    protected void clearBreakpoint () {
        remark (false);
        lines = null;
        disableEvents ();
        setValid (false);
    }

    /**
    * Fires property change of some event property (Needs re-setBreakpoint).
    */
    protected void fireEventPropertyChange (Event e, String s, Object o, Object n) {
        if (e != event) return;
        synchronizeProperties();
        firePropertyChange (null, o, n);
        if (isEnabled ()) setBreakpoint ();
    }

    // other methods ................................................................................

    /**
    * Synchronizes properties of all events of all debuggers.
    */
    private void synchronizeProperties () {
        if (events == null) return;
        ArrayList a = (ArrayList) events.clone ();
        int i, k = a.size ();
        for (i = 0; i < k; i++)
            synchronizeProperties ((Event) a.get (i));
    }

    /**
    * Synchronizes properties with given event. [event => e}
    */
    private void synchronizeProperties (Event e) {
        if (event == null) return;
//        doNotFire = true;
        Property[] s = event.getProperties ();
        Property[] d = e.getProperties (); 
        int i,k = Math.min (d.length, s.length);
        for (i = 0; i < k; i++)
            try {
                d [i].setValue (s [i].getValue ());
            } catch (Exception ex) {
            }
//        doNotFire = false;
    }
    
    /**
    * Calls remove () for each event.
    */
    private void disableEvents () {
        if (events == null) return;
        synchronized (events) {
            int i, k = events.size ();
            for (i = 0; i < k; i++) {
                ((Event) events.get (i)).remove ();
            }
        }
    }

    private Event addEventForDebugger (Debugger debugger) {
        if (event == null) return null;
        Event e = createEvent (debugger, event.getTypeName ());
        if (e == null) return null;
        synchronizeProperties (e);
        if (events == null) 
            events = new ArrayList ();
        synchronized (events) {
            events.add (e);
        }
        return e;
    }

    void addDebugger (Debugger debugger) {
        Event e = addEventForDebugger (debugger);
        if (e == null) return;
        if (isEnabled ())
            if (e.set ()) setValid (true);
    }

    void removeDebugger (AbstractDebugger debugger) {
        if (events == null) return;
        ArrayList a = (ArrayList) events.clone ();
        int i, k = a.size ();
        Event ee;
        for (i = 0; i < k; i++)
            if ((ee = (Event) a.get (i)).getDebugger () == debugger) {
                ee.remove ();
                synchronized (events) {
                    events.remove (ee);
                }
                return;
            }
        setBreakpoint ();
    }

    
    
    
    
    
    
    
    
    
    
    /*
    
    protected Event initEvent (Class cl) {
        Event old = event;
        event = getEvent (cl);
        event = event.get (this);
        removeEvents ();
        initActions ();

        // if event moved form class => line, we let property className
        if ( (old != null) &&
                (old instanceof ClassBreakpointEvent) &&
                (event instanceof ClassBreakpointEvent)
           ) ((ClassBreakpointEvent) event).setClassName (
                   ((ClassBreakpointEvent) old).getClassName ()
               );

        // add events for each started debugger
        Session[] s = ((EnterpriseDebugger) debugger).getSessions ();
        int i, k = s.length;
        for (i = 0; i < k; i++)
            addEvent (s [i].getDebugger ());
        return event;
    }
    
     */
    
}
