/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.multisession.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyEditor;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Arrays;

import org.openide.TopManager;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.explorer.propertysheet.editors.ChoicePropertyEditor;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;
import org.openide.util.WeakListener;
import org.openide.util.actions.SystemAction;
import org.openide.windows.InputOutput;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.nodes.ThreadGroupNode;
import org.netbeans.modules.debugger.support.nodes.ThreadNode;
import org.netbeans.modules.debugger.multisession.Session;
import org.netbeans.modules.debugger.multisession.EnterpriseDebugger;
import org.netbeans.modules.debugger.multisession.actions.*;


/**
* This class representates session as a Node.
*
* @author   Jan Jancura
*/
public final class SessionNode extends AbstractNode implements SessionCookie,
    SwitchOnCookie {
    /** generated Serialized Version UID */
    static final long                   serialVersionUID = -4736310787522307464L;


    // static ...........................................................................

    /** Property name */
    public static final String          PROP_SESSION_NAME        = "sessionName";
    /** Property name */
    public static final String          PROP_LOCATION_NAME       = "locationName";
    /** Property name */
    public static final String          PROP_PERSISTENT          = "persistent";
    /** Property name */
    public static final String          PROP_CONNECTED           = "connected";
    /** Property name */
    public static final String          PROP_STATE               = "state";
    /** Property name */
    public static final String          PROP_ACTION_ON_FINISH    = "actionOnFinish";

    private static final String         ICON_BASE =
        "/org/netbeans/modules/debugger/multisession/resources/session";

    private static final String         ICON_BASE_CURRENT =
        "/org/netbeans/modules/debugger/multisession/resources/sessionCurrent";

    /** System actions of this node */
    private static SystemAction[]     staticActions;


    // variables ...........................................................................

    /** Represented session. */
    private Session                     session;
    private SessionNodeListener         snl;


    // init ...........................................................................

    /**
    * Creates empty SessionNode.
    */
    SessionNode (Session s) {
        super (Children.LEAF);
        this.session = s;
        initialize ();
    }

    private void initialize () {
        updateName ();
        createProperties();
        getCookieSet ().add (this);
        snl = new SessionNodeListener ();
        session.addPropertyChangeListener (
            WeakListener.propertyChange (snl, session)
        );
        session.getDebugger ().addPropertyChangeListener (
            WeakListener.propertyChange (snl, session.getDebugger ())
        );
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerSessionNode");        //NOI18N
    }


    // Node implementation ...........................................................................

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null) {
            SwitchOnAction soa = (SwitchOnAction) SystemAction.get (
                                     SwitchOnAction.class
                                 );
            soa.setName (NbBundle.getBundle (SessionNode.class).
                         getString ("CTL_Switch_to_session"));
            staticActions = new SystemAction[] {
                soa,
                null,
                SystemAction.get (FinishSessionAction.class),
                //        SystemAction.get (RestartSessionAction.class),
                //        SystemAction.get (ConnectDisconnectAction.class),
                SystemAction.get (DestroySessionAction.class),
                null,
//                SystemAction.get (ShowOutputAction.class),
//                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class)
            };
        }
        return staticActions;
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (SwitchOnAction.class);
    }

    /** Creates property sets */
    private void createProperties () {
        final ResourceBundle bundle = NbBundle.getBundle (SessionNode.class);
        // default sheet with "properties" property set
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        ps.put (new PropertySupport.ReadOnly (
            PROP_SESSION_NAME,
            String.class,
            bundle.getString ("PROP_session_name"),
            bundle.getString ("HINT_session_name")
        ) {
            public Object getValue () throws InvocationTargetException {
                return session.getSessionName ();
            }
        });
        ps.put (new PropertySupport.ReadOnly (
            PROP_LOCATION_NAME,
            String.class,
            bundle.getString ("PROP_location_name"),
            bundle.getString ("HINT_location_name")
        ) {
            public Object getValue () throws InvocationTargetException {
                return session.getLocationName ();
            }
        });
        ps.put (new PropertySupport.ReadOnly (
            PROP_STATE,
            String.class,
            bundle.getString ("PROP_state"),
            bundle.getString ("HINT_state")
        ) {
            public Object getValue () throws InvocationTargetException {
                return getConnectionState ();
            }
        });
        ps.put (new PropertySupport.ReadWrite (
            PROP_PERSISTENT,
            Boolean.TYPE,
            bundle.getString ("PROP_persistent"),
            bundle.getString ("HINT_persistent")
        ) {
            public Object getValue () throws InvocationTargetException {
                return new Boolean (session.isPersistent ());
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    session.setPersistent (((Boolean)val).booleanValue ());
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
            }
        });
        ps.put (new PropertySupport.ReadWrite (
            PROP_ACTION_ON_FINISH,
            Integer.TYPE,
            bundle.getString ("PROP_action_on_finish"),
            bundle.getString ("HINT_action_on_finish")
        ) {
            public Object getValue () throws InvocationTargetException {
                return new Integer (session.getActionOnFinish ());
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    session.setActionOnFinish (((Integer)val).intValue ());
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
            }
            public PropertyEditor getPropertyEditor () {
                return new ChoicePropertyEditor (
                   new int[] {
                       Session.ACTION_FINISH,
                       //            Session.ACTION_DISCONNECT,
                       Session.ACTION_HIDE
                   },
                   new String[] {
                       bundle.getString ("CTL_Action_finish"),
                       //            bundle.getString ("CTL_Action_disconnect"),
                       bundle.getString ("CTL_Action_hide"),
                   }
               );
            }
        });
/*        ps.put (new PropertySupport.ReadWrite (
            PROP_CONNECTED,
            Boolean.TYPE,
            bundle.getString ("PROP_connected"),
            bundle.getString ("HINT_connected")
        ) {
            public Object getValue () throws InvocationTargetException {
                return new Boolean (session.isConnected ());
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    session.setConnected (((Boolean) val).booleanValue ());
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
            }
        });*/
        // and set new sheet
        setSheet (sheet);
    }


    // SessionCookie .............................................................

    /**
    * Connects to / disconnects from this session.
    */
    public void setConnected (boolean connected) {
        session.setConnected (connected);
    }

    /**
    * Returns state of session.
    */
    public boolean isConnected () {
        return session.isConnected ();
    }

    /**
    * Finish this session.
    */
    public void finish () {
        session.finish ();
    }

    /**
    * Show output window for this session.
    */
//    public void showOutput () {
//        InputOutput io = session.getInputOutput ();
//        io.select ();
//    }

    /**
    * Restarts this session.
    */
    public void restart () {
        destroy ();
        final PropertyChangeListener pcl = new PropertyChangeListener () {
            public void propertyChange (PropertyChangeEvent e) {
                if (session.getConnectionState () == Session.STATE_NOT_RUNNING) {
                    session.start ();
                }
            }
        };
    }

    /**
    * Destroyes curent session.
    */
    public void destroy () {
        session.setPersistent (false);
        session.finish ();
    }


    // SwitchCookie ..............................................................

    /**
    * Returns true if this node can be set current.
    */
    public boolean canSetCurrent () {
        return true;
    }

    /**
    * Sets state of thread.
    */
    public void setCurrent () {
        try {
            ((EnterpriseDebugger) TopManager.getDefault ().getDebugger ()).
            setCurrentSession (session);
        } catch (DebuggerNotFoundException e) {
        }
    }


    // private methods ...........................................................

    /**
    * Sets display name and fires property change.
    */
    void parameterChanged (PropertyChangeEvent e) {
        firePropertyChange (e.getPropertyName (), e.getOldValue (), e.getNewValue ());
    }

    private void updateName () {
        setIconBase (session.isCurrent () ? ICON_BASE_CURRENT : ICON_BASE);
        String s = session.getConnectionState ();
        if (s == Session.STATE_CONNECTED) {
            ResourceBundle bundle = NbBundle.getBundle (SessionNode.class);
            switch (session.getDebugger ().getState ()) {
            case Debugger.DEBUGGER_STARTING:
                s = bundle.getString ("CTL_Starting");
                break;
            case Debugger.DEBUGGER_RUNNING:
                s = bundle.getString ("CTL_Running");
                break;
            case Debugger.DEBUGGER_STOPPED:
                if ( ((AbstractDebugger) session.getDebugger ()).getLastAction () ==
                        AbstractDebugger.ACTION_BREAKPOINT_HIT
                   ) s = bundle.getString ("CTL_Stopped_on_breakpoint");
                else s = bundle.getString ("CTL_Stopped");
            }
        } else
            s = getConnectionState ();
        String n = session.getSessionName () + " (" + s + ")";
        setDisplayName (n);
        setName (n);
    }

    private String getConnectionState () {
        String s = session.getConnectionState ();
        ResourceBundle bundle = NbBundle.getBundle (SessionNode.class);
        if (s == Session.STATE_CONNECTED)
            return bundle.getString ("CTL_Connected");
        else
            if (s == Session.STATE_DISCONNECTED)
                return bundle.getString ("CTL_Disconnected");
            else
                if (s == Session.STATE_NOT_RUNNING)
                    return bundle.getString ("CTL_Not_running");
        return bundle.getString ("CTL_Hidden");
    }


    // innerclasses ..............................................................

    private class SessionNodeListener implements PropertyChangeListener {
        public void propertyChange (PropertyChangeEvent e) {
            if ( (e.getPropertyName () == Session.PROP_CONNECTION_STATE) ||
                 (e.getPropertyName () == Session.PROP_CURRENT) ||
                 (e.getPropertyName () == Debugger.PROP_STATE)
             ) updateName ();
            if (e.getSource () instanceof Session)
                parameterChanged (e);
        }
    }
}
