/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.multisession.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;

import org.openide.TopManager;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.multisession.Session;
import org.netbeans.modules.debugger.multisession.SessionsListener;
import org.netbeans.modules.debugger.multisession.EnterpriseModule;
import org.netbeans.modules.debugger.multisession.EnterpriseDebugger;


/**
* This class representates debugger sessions as a Node.
*
* @author Jan Jancura
*/
public final class SessionThreadRootNode extends AbstractNode {
    /** generated Serialized Version UID */
    static final long                   serialVersionUID = -4736310787523747464L;


    // static variables ...........................................................................

    private static final String         ICON_BASE =
        "/org/netbeans/modules/debugger/multisession/resources/sessions";
    private static final String         THREAD_GROUP_ICON_BASE =
        "/org/netbeans/core/resources/threads";

    
    // init .......................................................................................

    /**
    * Creates empty SessionRootNode.
    */
    public SessionThreadRootNode() {
        super (new SessionRootChildren ());
        initialize ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerThreadsRootNode");        //NOI18N
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new SessionsHandle ();
    }

    private void initialize () {
        String s = NbBundle.getBundle (SessionRootNode.class).getString ("CTL_Thread_group");
        setDisplayName (s);
        setName (s);
        setIconBase (THREAD_GROUP_ICON_BASE);
    }


    // Node implementation ...........................................................................


    // inner classes .......................................................

    /**
    * Shows SessionNode-s.
    */
    private static final class SessionRootChildren extends Children.Keys implements SessionsListener,
        PropertyChangeListener {

        /** Where to add variables. */
        private EnterpriseDebugger          debugger;

        SessionRootChildren () {
            try {
                debugger = (EnterpriseDebugger) TopManager.getDefault ().getDebugger ();
            } catch (DebuggerNotFoundException e) {
            }
        }

        /** Create a node for one data object.
        * @param key DataObject
        */
        protected Node[] createNodes (Object key) {
            if (((Session) key).getConnectionState () == Session.STATE_CONNECTED)
                return new Node [] {
                    new SessionThreadNode ((Session) key)
                };
            return new Node [] {
            };
        }

        /** Initializes the children.
        */
        protected void addNotify () {
            Session[] s = debugger.getSessions ();
            int i, k = s.length;
            for (i = 0; i < k; i++)
                s [i].addPropertyChangeListener (this);
            update ();
            debugger.addSessionsListener (this);
        }

        /** Deinitializes the children.
        */
        protected void removeNotify () {
        }

        /**
        * New session created.
        */
        public void sessionCreated (Session s) {
            update ();
            s.addPropertyChangeListener (this);
        }

        /**
        * Session death.
        */
        public void sessionDeath (Session s) {
            s.removePropertyChangeListener (this);
            update ();
        }

        public void propertyChange (PropertyChangeEvent e) {
            if ((e.getPropertyName () == null) || (!e.getPropertyName ().equals (Session.PROP_CONNECTION_STATE)))
                return;
            update ();
        }

        private void update () {
            //S ystem.out.println("SessionsContext.update " + debugger.getSessions ().length);
            Session[] s = debugger.getSessions ();
            int i, k = s.length;
            LinkedList ll = new LinkedList ();
            for (i = 0; i < k; i++)
                if (s [i].getConnectionState () == Session.STATE_CONNECTED
                ) ll.add (s [i]);
            setKeys (ll);
        }
    } // end of ThreadGroupContextChildren inner class

    /**
    * Serializable node reference. 
    */
    private static class SessionsHandle implements Node.Handle {
        static final long serialVersionUID = -4935665488987259653L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return EnterpriseModule.getSessionThreadsRootNode ();
        }
    }
}
