/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ResourceBundle;
import java.util.ArrayList;
import java.util.Arrays;
import javax.swing.JComponent;

import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.text.Line;
import org.openide.cookies.LineCookie;
import org.openide.cookies.DebuggerCookie;
import org.openide.debugger.Breakpoint;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.Watch;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.text.Line;
import org.openide.src.ConstructorElement;

import org.netbeans.modules.debugger.*;

import org.netbeans.modules.debugger.support.java.MethodBreakpointEvent;//PENDING

/**
* Standart implementation of breakpoint interface.
* @see org.openide.debugger.Breakpoint
*
* @author   Jan Jancura
* @version  0.23, May 26, 1998
*/
public class BreakpointSupport extends CoreBreakpoint {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 3686204002781932191L;

    /** bundle to obtain text information from */
    private static ResourceBundle   bundle = org.openide.util.NbBundle.getBundle
            (BreakpointSupport.class);


    // private variables .....................................................

    private boolean                   enabled = true;
    private boolean                   valid = false;
    private boolean                   hidden = false;
    protected ArrayList               lines = null;
    protected boolean                 linesVisible = true;


    // init ....................................................................

    /**
    * Non public constructor called from the Debugger only.
    * User must create breakpoint from Debugger.getNewBreakpoint () method.
    */
    protected BreakpointSupport (boolean hidden) {
        this.hidden = hidden;
    }

    protected void init () {
        super.init ();
        if (enabled) setBreakpoint ();
    }


    // interface Breakpoint .........................................................

    /** 
    * Destroys the breakpoint. Removes it from the list of all breakpoints in the system.
    */
    public void remove () {
        //S ystem.out.println("CoreBreakpoint.remove "); // NOI18N
        clearBreakpoint ();
        if (debugger instanceof DebuggerSupport)
            ((DebuggerSupport) debugger).removeBreakpoint (this);
    }

    /**
    * Returns the line this breakpoint is on.
    *
    * @return the line this breakpoint is on.
    */
    public Line getLine () {
        if (event instanceof LineBreakpointEvent)
            return ((LineBreakpointEvent) event).getLine ();
        if ((lines == null) || (lines.size () < 1)) return null;
        return (Line) lines.get (0);
    }

    /**
    * Returns the class name this breakpoint is in.
    *
    * @return the class name this breakpoint is in.
    */
    public String getClassName () { // OBSOLETE
        if (event instanceof MethodBreakpointEvent)
            return ((MethodBreakpointEvent) event).getClassName ();
        if (event instanceof LineBreakpointEvent)
            return ((LineBreakpointEvent) event).getSourceName ();
        return null;
    }

    /**
    * Getter for the method the breakpoint is assigned to.
    *
    * @return the method element or null
    */
    public ConstructorElement getMethod () { // OBSOLETE
        return null; //[PENDING]
    }

    /**
    * Getter method for enabled property.
    *
    * @return state of enabled property.
    */
    public boolean isEnabled () {
        return enabled;
    }

    /**
    * Setter method for enabled property.
    *
    * @param state of enabled property.
    */
    public void setEnabled (boolean enabled) {
        if (enabled == this.enabled) return;
        boolean old = this.enabled;
        this.enabled = enabled;
        if (enabled) setBreakpoint ();
        else clearBreakpoint ();
        firePropertyChange (PROP_ENABLED, new Boolean (old), new Boolean (enabled));
    }

    /**
    * Getter method for valid property.
    *
    * @return state of valid property.
    */
    public boolean isValid () {
        return valid;
    }

    public boolean isHidden () {
        return hidden;
    }


    // CoreBreakpoint impl .........................................................
    
    /**
    * Sets current breakpoint event.
    */
    public void setEvent (Event e) {
        clearBreakpoint ();
        super.setEvent (e);
        setBreakpoint ();
    }

    public void setHidden (boolean hidden) {
        if (this.hidden == hidden) return;
        this.hidden = hidden;
        firePropertyChange (PROP_HIDDEN, new Boolean (!hidden), new Boolean (hidden));
        remark ((!hidden) && enabled);
    }
    
    /**
    * Sets property valide and fires it.
    *
    * @param valid value of this property to be set.
    */
    protected void setValid (boolean valid) {
        if (valid == this.valid) return;
        boolean old = this.valid;
        this.valid = valid;
        firePropertyChange (PROP_VALID, new Boolean (old), new Boolean (valid));
    }
    
    /**
    * Fires property change of some event property (Needs re-setBreakpoint).
    */
    protected void fireEventPropertyChange (Event e, String s, Object o, Object n) {
        super.fireEventPropertyChange (e, s, o, n);
        if (enabled) setBreakpoint ();
    }

    
    // helper methods ..........................................................
    
    /**
    * Sets or updates breakpoint. Calls set method of currently selected event, sets
    * value of valid property and marks lines in the editor.
    */
    protected void setBreakpoint () {
        if (!enabled) return;
        boolean valid = false;
        if (event != null) {
            // 1) set / remove
            if (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING)
                event.remove ();
            else
                valid = event.set ();
            // 2) remark
            Line[] l = event.getLines ();
            if (l != null)
                remark (new ArrayList (Arrays.asList (l)));
            else
                remark (null);
        }
        // 3) setValid
        setValid (valid);
    }

    /**
    * Removes breakpoint from this position.
    * Clears Line and breakpoint from RemoteDebugger.
    */
    protected void clearBreakpoint () {
        remark (false);
        lines = null;
        if (event != null)
            event.remove ();
        setValid (false);
    }

    /**
    * Marks or unmarks all important lines (line variable).
    */
    protected void remark (boolean b) {
        linesVisible = b;
        if (lines == null) return;
        int i, k = lines.size ();
        for (i = 0; i < k; i++)
            ((Line) lines.get (i)).setBreakpoint (b);
        if (!b) remarkAll ();
    }

    /**
    * Remarks lines to new positions (checks hidden property, but not enabled).
    */
    protected void remark (ArrayList n) {
        int i, k = (n == null) ? 0 : n.size ();
        if (!isHidden ())
            for (i = 0; i < k; i++)  // mark new lines
                if ((lines == null) || !lines.remove (n.get (i)))
                    ((Line) n.get (i)).setBreakpoint (true);
        k = (lines == null) ? 0 : lines.size ();
        for (i = 0; i < k; i++)  // unmark old lines
            ((Line) lines.get (i)).setBreakpoint (false);
        lines = n;
        remarkAll ();
    }
    
    protected void remarkAll () {
        try {
            Breakpoint[] bs = TopManager.getDefault ().getDebugger ().getBreakpoints ();
            int i, k = bs.length;
            for (i = 0; i < k; i++)
                if ( (bs [i] instanceof BreakpointSupport) &&
                     ((BreakpointSupport) bs [i]).linesVisible
                ) ((BreakpointSupport) bs [i]).remark (true);
        } catch (DebuggerNotFoundException e) {
        }
    }

    /**
    * Returns action of given type.
    */
    public CoreBreakpoint.Action getAction (Class actionType) {
        CoreBreakpoint.Action[] actions = getActions ();
        int i, k = actions.length;
        for (i = 0; i < k; i++)
            if (actions [i].getClass ().isAssignableFrom (actionType))
                return actions [i];
        return null;
    }
    

    // helper methods ..........................................................
    
    /**
    * Sets line on this breakpoint.
    */
    public void setLine (Line line) {
        if (!(event instanceof LineBreakpointEvent)) {
            Event e = createEvent (debugger, LineBreakpointEvent.TYPE_NAME);
            ((LineBreakpointEvent) e).setLine (line);
            setEvent (e);
            return;
        }
        ((LineBreakpointEvent) event).setLine (line);
    }

    /**
    * Sets class name of this breakpoint.
    *
    * @param name name of class of this breakpoint.
    */
    public void setClassName (String name) {
        name = name.trim ();
        if (!((event instanceof MethodBreakpointEvent) ||
              (event instanceof LineBreakpointEvent))
        ) {
            setEvent (createEvent (debugger, MethodBreakpointEvent.TYPE_NAME));
        }
        if (event instanceof MethodBreakpointEvent)
            ((MethodBreakpointEvent) event).setClassName (name);
        if (event instanceof LineBreakpointEvent)
            ((LineBreakpointEvent) event).setSourceName (name);
    }

    /**
    * Sets breakpoint on the method with name specified.
    *
    * @param name name of method to set breakpoint on.
    */
    public void setMethodName (String name) {
        name = name.trim ();
        if (!(event instanceof MethodBreakpointEvent)) {
            setEvent (createEvent (debugger, MethodBreakpointEvent.TYPE_NAME));
        }
        ((MethodBreakpointEvent) event).setMethodName (name);
    }

    /**
    * Returns the method name this breakpoint is in.
    *
    * @return the class name this breakpoint is in.
    */
    public String getMethodName () {
        if (event instanceof MethodBreakpointEvent)
            return ((MethodBreakpointEvent) event).getMethodName ();
        return null;
    }
}
