/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.awt.Toolkit;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;
import java.net.URL;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.text.Keymap;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.nodes.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.java.nodes.*;

import org.openide.actions.StartDebuggerAction;
import org.openide.debugger.DebuggerException;
import org.openide.loaders.*;
import org.openide.windows.Workspace;
import org.openide.windows.Mode;
import org.openide.windows.WindowManager;
import org.openide.windows.TopComponent;
import org.openide.util.actions.SystemAction;
import org.openide.util.Utilities;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.nodes.Node;
import org.openide.TopManager;

/**
* Module installation class for HtmlModule
*
* @author Jan Jancura
*/
public class DebuggerModule extends org.openide.modules.ModuleInstall {

    static final long serialVersionUID = 2700701939804772882L;


    // static ..............................................................................

    /** name of debugger mode */
    public static final String            MODE_NAME = "debugger"; // NOI18N
    /** name of debugger workspace */
    public static final String            WORKSPACE_NAME = ((StartDebuggerAction) StartDebuggerAction.get
            (StartDebuggerAction.class)).getWorkspace ();

    private static Keymap                 map;
    /** Node representing debugger breakpoints. */
    private static Node                   callStackRootNode;
    /** Node representing debugger breakpoints. */
    private static Node                   breakpointsRootNode;
    /** Node representing debugger threads. */
    private static Node                   threadGroupsRootNode;
    /** Node representing debugger watches. */
    private static Node                   watchesRootNode;
    /** Node representing debugger variables in current context. */
    private static Node                   variablesRootNode;
    /** Node representing loaded classes in current context. */
    private static Node                   classesRootNode;
    /** Debugger view workspace */
    private static Workspace              workspace;
    /** Stores all views of debugger window. */
    private static LinkedList             views = new LinkedList ();
    /** Stores all nodes representing debugger in the explorer. */
    private static LinkedList             nodes = new LinkedList ();
    /** Stores all nodes classes representing debugger in the explorer. */
    private static HashMap                nodesClasses = new HashMap ();
    /** Number of starts counter. */
    private static int                    numberOfStarts = 0;
    protected static DebuggerModule       defaultModule;
    /** TopComponents to close when a main window is opened [PATCH]. */
    private static ArrayList              viewsToClose = new ArrayList ();
    /** If true method installWorkspace () will be called when a main window is
    * opened [PATCH]. 
    */
    private static boolean                patchInitialized = false;

    private static boolean                installWorkspaces = false;
    

    // init ..................................................................................

    public DebuggerModule () {
    }

    public void readExternal (final java.io.ObjectInput objectInput)
    throws java.io.IOException, java.lang.ClassNotFoundException {
        super.readExternal (objectInput);
        numberOfStarts = objectInput.readInt ();
    }

    public void writeExternal (final java.io.ObjectOutput objectOutput)
    throws java.io.IOException {
        super.writeExternal (objectOutput);
        objectOutput.writeInt (numberOfStarts);
    }


    // ModuleInstall implementation ........................................................

    /** Module installed for the first time. */
    public void installed () {     
        registerNodes ();
        registerRootNodeClasses ();
        java.awt.Window mainWindow = TopManager.getDefault ().
                                     getWindowManager ().getMainWindow ();
        installWorkspaces = true;
        if (mainWindow.isShowing ()) {
            if (defaultModule == null)
                defaultModule = this;
            installWorkspace ();
        } else {
            if (defaultModule == null) {
                defaultModule = this;
                mainWindow.addWindowListener (new Validator ()); // not "that" Validator ;-)
            } else {
                if (defaultModule.getClass ().isAssignableFrom (getClass ()))
                    defaultModule = this;
            }
        }
    }

    /**
    * Called when the module is loaded and the version is higher than
    * by the previous load
    * The default implementation calls restored().
    * @release The major release number of the <B>old</B> module code name or -1 if not specified.
    * @specVersion The specification version of the this <B>old</B> module.
    */
    public void updated ( int release, String specVersion ) {
        installed ();
    }

    /** Module installed again. */
    public void restored () {
        registerNodes ();
        registerRootNodeClasses ();
        if (defaultModule == null) {
            defaultModule = this;
            java.awt.Window mainWindow = TopManager.getDefault ().
                                         getWindowManager ().getMainWindow ();
            mainWindow.addWindowListener (new Validator ()); // not "that" Validator ;-)
        } else {
            if (defaultModule.getClass ().isAssignableFrom (getClass ()))
                defaultModule = this;
        }
        numberOfStarts ++;
    }

    /** Module was uninstalled. */
    public void uninstalled () {
        uninstallActions ();
        uninstallWorkspace ();
        map = null;
        callStackRootNode = null;
        breakpointsRootNode = null;
        threadGroupsRootNode = null;
        watchesRootNode = null;
        workspace = null;
        views = new LinkedList ();
        nodes = new LinkedList ();
    }

    /** Module is being closed. */
    public boolean closing () {
        try {
            TopManager.getDefault ().getDebugger ().finishDebugger ();
        } catch (Throwable e) {
        }
        return true; // agree to close
    }


    // main public methods ...................................................................

    /**
    * Adds a new debugger view.
    *
    * @param replace If true, previous view of the same type is replaced, otherwise not.
    */
    public static void addView (TopComponent view, boolean replace) {
        TopComponent dview;
        if (view instanceof DebuggerView)
            dview = getViewFor (
                        ((DebuggerView) view).getExplorerManager ().getRootContext ()
                    );
        else
            dview = getView (view.getClass ());

        if (dview == null) {
            views.add (view);
        } else
            if (replace) {
                int in = views.indexOf (dview);
                closeView (dview);
                views.add (in, view);
            }
    }

    /**
    * Closes given debugger view.
    */
    public static void closeView (final TopComponent view) {
        if (patchInitialized) {
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    SwingUtilities.invokeLater (new Runnable () {
                        public void run () {
                            if (view.isOpened ()) {
                                view.setCloseOperation (DebuggerView.CLOSE_EACH);
                                view.close (null);
                            }
                        }
                    });
                }
            }, 1000);
        } else
            viewsToClose.add (view);
        views.remove (view);
    }

    /**
    * Returns debugger view of given type.
    */
    public static DebuggerView getViewFor (Node rn) {
        Iterator it = views.iterator ();
        while (it.hasNext ()) {
            TopComponent viewObj = (TopComponent) it.next ();
            if (!(viewObj instanceof DebuggerView)) continue;
            DebuggerView view = (DebuggerView) viewObj;
            if (view.getExplorerManager ().getRootContext ().equals (rn))
                return view;
        }
        return null;
    }

    /**
    * Returns debugger view of given type.
    */
    public static TopComponent getView (Class cls) {
        Iterator it = views.iterator ();
        while (it.hasNext ()) {
            TopComponent view = (TopComponent) it.next ();
            if (view.getClass ().equals (cls))
                return view;
        }
        return null;
    }

    /**
    * Returns node of given type.
    */
    public static Node getNode (Class cls) {
        Iterator it = nodes.iterator ();
        while (it.hasNext ()) {
            Object node = it.next ();
            if (node.getClass ().equals (cls))
                return (Node) node;
        }
        return null;
    }

    /**
    * Returns all debugger views of debugger vindow.
    */
    public static Collection getViews () {
        return Collections.unmodifiableList (views);
    }

    /**
    * Adds custom node to the debugger root node in the explorer
    * hierarchy.
    */  
    public static void addNode (Node node) {
        nodes.add (node);
        if (node != null)
            registerRootNodeClass (node.getClass ());
        /*
        Node e = TopManager.getDefault ().getPlaces ().nodes ().environment ();
        Node[] ee = e.getChildren ().getNodes ();
        int i, k = ee.length;
        for (i = 0; i < k; i++)
            if (ee [i] instanceof DebuggerNode)
                break;
        if (i == k) return;
        ee [i].getChildren ().add (new Node[] {
                                       node
                                          });
        */
    }
    
    /**
    * Registers a debugger root node class.
    */
    public static void registerRootNodeClass (Class clazz) {
        if (! isRegisteredRootNodeClass (clazz))
            nodesClasses.put (clazz, clazz);
    }

    /**
    * Tests is debugger root node class is registered or not.
    */
    public static boolean isRegisteredRootNodeClass (Class clazz) {
        return nodesClasses.get (clazz) != null;
    }
    
    /**
    * Returns all debugger views of debugger vindow.
    */
    public static Collection getNodes () {
        if (defaultModule != null)
            defaultModule.installNodes ();
        return Collections.unmodifiableList (nodes);
    }

    /**
    * Return Node representing debugger breakpoints.
    */  
    public static Node getCallStackRootNode () {
        if (callStackRootNode == null) {
            callStackRootNode = new CallStackRootNode ();
            addNode (callStackRootNode);
        }
        return callStackRootNode;
    }

    /**
    * Return Node representing debugger breakpoints.
    */  
    public static Node getBreakpointsRootNode () {
        if (breakpointsRootNode == null) {
            breakpointsRootNode = new BreakpointsRootNode ();
            addNode (breakpointsRootNode);
        }
        return breakpointsRootNode;
    }

    /**
    * Return Node representing debugger threads.
    */  
    public static Node getThreadGroupsRootNode () {
        if (threadGroupsRootNode == null) {
            // obtain thread group
            try {
                ThreadsProducer threadProducer = ((ThreadsRoot) TopManager.
                    getDefault ().getDebugger ()).getThreadsRoot ();
                threadGroupsRootNode = NodesRegistry.getNode (
                    threadProducer, 
                    new Object[] {threadProducer, Boolean.TRUE}
                );
                //threadGroupsRootNode = new ThreadGroupNode (threadProducer, true);
                addNode (threadGroupsRootNode);
            } catch (DebuggerException e) {
                throw new InternalError ("Debugger not initialized"); // NOI18N
            }
        }
        return threadGroupsRootNode;
    }

    /**
    * Return Node representing debugger watches.
    */  
    public static Node getWatchesRootNode () {
        if (watchesRootNode == null) {
            watchesRootNode = new WatchesRootNode ();
            addNode (watchesRootNode);
        }
        return watchesRootNode;
    }

    /**
    * Return Node representing current context variables.
    */  
    public static Node getVariablesRootNode () {
        if (variablesRootNode == null) {
            variablesRootNode = new JavaVariablesRootNode ();
            addNode (variablesRootNode);
        }
        return variablesRootNode;
    }
    
    /**
    * Return Node representing loaded classes in current context.
    */  
    public static Node getClassesRootNode () {
        if (classesRootNode == null) {
            classesRootNode = new ClassesRootNode ();
            addNode (classesRootNode);
        }
        return classesRootNode;
    }
    
    /**
    * Returns Debugger workspace.
    */
    public static Workspace getWorkspace () {
        WindowManager wm = TopManager.getDefault ().getWindowManager ();
        workspace = wm.findWorkspace (WORKSPACE_NAME);
        if (workspace == null)
            workspace = wm.createWorkspace (WORKSPACE_NAME);
        return workspace;
    }

    /**
    * Switches to debugger workspace.
    */
    public static void changeWorkspace () {
        WindowManager dp = TopManager.getDefault ().getWindowManager ();
        Workspace d = dp.findWorkspace (org.openide.actions.StartDebuggerAction.getWorkspace());
        if (d != null) d.activate ();
    }

    /**
    * Installation of debugger actions.
    */
    protected void installActions () {
    }

    /**
    * Installation of debugger actions.
    */
    protected void registerNodes () {
        NodesRegistry.register (CoreBreakpoint.class, BreakpointNode.class);
        NodesRegistry.register (AbstractWatch.class, WatchNode.class);
        NodesRegistry.register (AbstractThread.class, ThreadNode.class);
        NodesRegistry.register (Location.class, LocationNode.class);
        NodesRegistry.register (AbstractVariable.class, VariableNode.class);
        NodesRegistry.register (ThreadsProducer.class, ThreadGroupNode.class);
        //NodesRegistry.register (new Class[] {AbstractThread.class, ThreadsProducer.class}, ThreadGroupNode.class);
        // java nodes
        NodesRegistry.register (JavaWatch.class, JavaWatchNode.class);
        NodesRegistry.register (JavaThread.class, JavaThreadNode.class);
        NodesRegistry.register (CallStackFrame.class, CallStackLineNode.class);
        NodesRegistry.register (JavaVariable.class, JavaVariableNode.class);
        NodesRegistry.register (new Class[] {JavaThreadGroup.class, ThreadsProducer.class}, JavaThreadGroupNode.class);
    }

    /**
    * Remove of debugger actions.
    */
    protected void uninstallActions () {
    }

    /**
    * Install all workspaces.
    */
    public static void installWorkspaces () {
        installWorkspaces = true;
    }

    /**
    * Installation of debugger workspace and open DebuggerWindow.
    */
    protected void installWorkspace () {
        WindowManager wm = TopManager.getDefault ().getWindowManager ();
        Workspace workspace = getWorkspace ();
        defaultModule.installViews ();

        // Opens DebuggerWindow for BuildMakera... ;-)
        ((DebuggerViewAction) SystemAction.get (DebuggerViewAction.class)).
        showDebuggerView (workspace);
    }

    /**
    * Installation of debugger views.
    */
    protected void installViews () {
        addView (new DebuggerView (true, getBreakpointsRootNode ()), false);
        addView (new DebuggerView (true, getThreadGroupsRootNode ()), false);
        addView (new DebuggerView (true, getCallStackRootNode ()), false);
        addView (new DebuggerView (true, getWatchesRootNode ()), false);
        addView (new DebuggerView (true, getVariablesRootNode ()), false);
        addView (new DebuggerView (true, getClassesRootNode ()), false);
    }

    /**
    * Installation of debugger nodes.
    */
    protected void installNodes () {
        getBreakpointsRootNode ();
        getThreadGroupsRootNode ();
        getCallStackRootNode ();
        getWatchesRootNode ();
        getVariablesRootNode ();
        getClassesRootNode ();
    }

    protected void registerRootNodeClasses () {
        registerRootNodeClass (BreakpointsRootNode.class);
        registerRootNodeClass (ThreadGroupNode.class);
        registerRootNodeClass (CallStackRootNode.class);
        registerRootNodeClass (WatchesRootNode.class);
        registerRootNodeClass (VariablesRootNode.class);
        registerRootNodeClass (ClassesRootNode.class);
    }
    
    /**
    * Remove debgger workspace.
    */
    protected void uninstallWorkspace () {
        int i, k = views.size ();
        for (i = 0; i < k; i++) {
            TopComponent t = (TopComponent) views.get (i);
            if (t.isOpened ()) {
                t.setCloseOperation (TopComponent.CLOSE_EACH);
                t.close (null);
            }
        }
    }


    // support methods .......................................................................

    /** Assigns a key to an action
    * @param key key name
    * @param action name of the action
    */
    protected static void assign (String key, String action) throws ClassNotFoundException {
        KeyStroke str = Utilities.stringToKey (key);
        if (str == null) {
            System.err.println ("Not a valid key: " + key);
            // go on
            return;
        }

        Class actionClass = Class.forName (action);

        // create instance of the action
        SystemAction a = SystemAction.get (actionClass);

        map.addActionForKeyStroke (str, a);
    }

    /**
    * Creates debugger mode on given workspace if debugger mode
    * doesn't exist on given worksspace 
    */
    public static Mode createMode (Workspace workspace, TopComponent dv) {
        URL iconURL = null;
        try {
            iconURL = new URL ("nbresboot:/org/netbeans/core/resources/debuggerView.gif"); // NOI18N
        } catch (java.net.MalformedURLException e) {
        }
        Mode m = workspace.createMode (
                     MODE_NAME,
                     NbBundle.getBundle (DebuggerModule.class).
                     getString ("CTL_Debugger_view"),
                     iconURL
                 );
        // compute the bounds for debugger
        Rectangle workspaceBounds = TopManager.getDefault ().
            getWindowManager ().getCurrentWorkspace ().getBounds();
        Dimension prefSize = dv.getPreferredSize ();
        Rectangle bounds = new Rectangle (
            workspaceBounds.x + workspaceBounds.width - prefSize.width,
            workspaceBounds.y, prefSize.width, prefSize.height
        );
        m.setBounds (bounds);
        return m;
    }


    // innerclasses ..............................................................

    class Validator extends WindowAdapter {
        public void windowOpened (java.awt.event.WindowEvent ev) {
            // notify me no more
            java.awt.Window mainWindow = TopManager.getDefault ().
                                         getWindowManager ().getMainWindow ();
            mainWindow.removeWindowListener (this);

            SwingUtilities.invokeLater (new Runnable () {
                public void run () {

                    if ( ((numberOfStarts == 1) && (views.size () < 1)) ||
                            installWorkspaces
                       ) {
                        installWorkspace ();
                    } else
                        defaultModule.installViews ();

                    // close waiting views [PATCH]
                    int i, k = viewsToClose.size ();
                    for (i = 0; i < k; i++) {
                        TopComponent tc = (TopComponent) viewsToClose.get (i);
                        if (tc.isOpened ()) {
                            tc.setCloseOperation (DebuggerView.CLOSE_EACH);
                            tc.close (null);
                        }
                    } //for
                    viewsToClose = new ArrayList ();
                    patchInitialized = true;
                } //run
            });
        }
    }
}

