/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import org.openide.options.SystemOption;
import org.openide.options.ContextSystemOption;
import org.openide.actions.StartDebuggerAction;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.execution.NbClassPath;
import org.openide.TopManager;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;

/**
* Settings for java debugger.
*
* @author Jan Jancura, Ian Formanek
* @version 0.11, May 16, 1998
*/
public class DebuggerSettings extends SystemOption {

    // static .....................................................................................

    /** generated Serialized Version UID */
    static final long serialVersionUID = 833203088075073629L;

    /** actionOnTraceInto property name */
    public static final String PROP_ACTION_ON_TRACE_INTO = "actionOnTraceInto"; // NOI18N
    /** actionOnTraceIntoSet property name */
    public static final String PROP_ACTION_ON_TRACE_INTO_SET = "actionOnTraceIntoSet"; // NOI18N
    /** desktop property name */
    public static final String PROP_DESKTOP = "desktop"; // NOI18N
    /** remoteDebugger property name */
    public static final String PROP_REMOTE_DEBUGGER = "remoteDebugger"; // NOI18N
    /** Property name of the runCompilation property */
    public static final String PROP_RUN_COMPILATION = "runCompilation"; // NOI18N
    /** Property name of the showFinishDialog property */
    public static final String PROP_SHOW_FINISH_DIALOG = "showFinishDialog"; // NOI18N
    /** Property name of the showRunMultisessionDialog property */
    public static final String PROP_SHOW_RUN_MULTISESSION_DIALOG = "showRunMultisessionDialog"; // NOI18N
    /** Property name of the runMultisession property */
    public static final String PROP_RUN_MULTISESSION = "runMultisession"; // NOI18N

    /** Constant for actionOnTraceInto property. */
    public static int ACTION_ON_TI_STOP;
    /** Constant for actionOnTraceInto property. */
    public static int ACTION_ON_TI_TRACE_OUT;

    private static int[] actionTIConstants;
    private static String[] actionTIComments;

    static {
        ACTION_ON_TI_STOP = putActionTI (
                                NbBundle.getBundle (DebuggerSettings.class).
                                getString ("CTL_Action_on_TI_stop")
                            );
        ACTION_ON_TI_TRACE_OUT = putActionTI (
                                     NbBundle.getBundle (DebuggerSettings.class).
                                     getString ("CTL_Action_on_TI_step_out")
                                 );
    }


    // SystemOption implementation ..................................................................

    /**
    * Returns name of this setings.
    */
    public String displayName () {
        return NbBundle.getBundle (DebuggerSettings.class).getString ("CTL_Debugger_option");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (DebuggerSettings.class);
    }


    // SystemOption implementation ..................................................................

    {
        setActionOnTraceInto (ACTION_ON_TI_TRACE_OUT);
    }

    static int[] getActionTIConstants () {
        return actionTIConstants; // NOI18N
    }

    static String[] getActionTIComments () {
        return actionTIComments; // NOI18N
    }

    public static String getActionTI (int actionTIConstant) {
        if (actionTIConstants == null) actionTIConstants = new int [] {};
        if (actionTIComments == null) actionTIComments = new String [] {};

        int i, k = actionTIConstants.length;
        for (i = 0; i < k; i++)
            if (actionTIConstants [i] == actionTIConstant)
                return actionTIComments [i];
        return null;
    }

    public static int putActionTI (String actionTIComment) {
        if (actionTIConstants == null) actionTIConstants = new int [] {};
        if (actionTIComments == null) actionTIComments = new String [] {};
        int actionTIConstant = actionTIConstants.length == 0 ?
                               1 :
                               actionTIConstants [actionTIConstants.length - 1] + 1;

        int[] ncon = new int [actionTIConstants.length + 1];
        String[] ncomm = new String [actionTIComments.length + 1];
        System.arraycopy (actionTIConstants, 0, ncon, 0, actionTIConstants.length);
        ncon [actionTIConstants.length] = actionTIConstant;
        System.arraycopy (actionTIComments, 0, ncomm, 0, actionTIComments.length);
        ncomm [actionTIComments.length] = actionTIComment;

        actionTIConstants = ncon; // NOI18N
        actionTIComments = ncomm; // NOI18N
        return actionTIConstant;
    }

    public static void removeActionTI (int actionTIConstant) {
        int i, k = actionTIConstants.length;
        for (i = 0; i < k; i++)
            if (actionTIConstants [i] == actionTIConstant)
                break;
        if (i == k) return;

        int[] ncon = new int [actionTIConstants.length - 1];
        String[] ncomm = new String [actionTIComments.length - 1];
        System.arraycopy (actionTIConstants, 0, ncon, 0, i);
        System.arraycopy (actionTIConstants, i + 1, ncon, i, actionTIConstants.length - i - 1);
        System.arraycopy (actionTIComments, 0, ncomm, 0, i);
        System.arraycopy (actionTIComments, i + 1, ncomm, 0, actionTIComments.length - i - 1);

        actionTIConstants = ncon; // NOI18N
        actionTIComments = ncomm; // NOI18N
    }


    // properties .................................................................................

    /**
    * Getter for remoteDebugger property. 
    */
    public DebuggerSupport getRemoteDebugger () throws DebuggerNotFoundException {
        return null;
    }

    /**
    * Setter for remoteDebugger property. 
    */
    public void setRemoteDebugger (DebuggerSupport javaDebugger) throws DebuggerNotFoundException {
        if (javaDebugger != null) {
            DebuggerSupport oldValue = (DebuggerSupport) TopManager.getDefault ().getDebugger ();
            oldValue.setDebugger (javaDebugger);
            // fire the PropertyChange
            firePropertyChange (PROP_REMOTE_DEBUGGER, oldValue, javaDebugger);
        }
    }

    /**
    * Getter for runCompilation property. 
    */
    public boolean getRunCompilation () {
        return StartDebuggerAction.getRunCompilation ();
    }

    /**
    * Setter for runCompilation property. 
    */
    public void setRunCompilation(boolean value) {
        boolean oldValue = getRunCompilation ();
        StartDebuggerAction.setRunCompilation (value);
        if (oldValue != value)
            firePropertyChange (PROP_RUN_COMPILATION, new Boolean (oldValue), new Boolean (value));
    }

    /**
    * Getter for desktop property. 
    */
    public String getDesktop () {
        return StartDebuggerAction.getWorkspace ();
    }

    /**
    * Setter for desktop property. 
    */
    public void setDesktop (String desktop) {
        String oldValue = getDesktop ();
        if (desktop.equals (oldValue)) return;
        StartDebuggerAction.setWorkspace(desktop);
        // fire the PropertyChange
        firePropertyChange (PROP_DESKTOP, new Boolean (oldValue), new Boolean (desktop));
    }

    /**
    * Getter for actionOnTraceInto property.
    */
    public int getActionOnTraceInto () {
        return ((Integer) getProperty (PROP_ACTION_ON_TRACE_INTO)).intValue ();
    }

    /**
    * Setter for actionOnTraceInto property.
    */
    public void setActionOnTraceInto (int actionOnTraceInto) {
        putProperty (PROP_ACTION_ON_TRACE_INTO, new Integer (actionOnTraceInto), true);
    }

    /**
    * Getter for actionOnTraceIntoSet property.
    */
    public boolean isActionOnTraceIntoSet () {
        Boolean b = (Boolean) getProperty (PROP_ACTION_ON_TRACE_INTO_SET);
        if (b == null) return true; // default value
        return b.booleanValue ();
    }

    /**
    * Setter for actionOnTraceIntoSet property.
    */
    public void setActionOnTraceIntoSet (boolean actionOnTraceIntoSet) {
        putProperty (PROP_ACTION_ON_TRACE_INTO_SET, new Boolean (actionOnTraceIntoSet), true);
    }

    /**
    * Getter for showFinishDialog property.
    */
    public boolean getShowFinishDialog () {
        Boolean b = (Boolean) getProperty (PROP_SHOW_FINISH_DIALOG);
        if (b == null) return true; // default value
        return b.booleanValue ();
    }

    /**
    * Setter for showFinishDialog property.
    */
    public void setShowFinishDialog (boolean showFinishDialog) {
        putProperty (PROP_SHOW_FINISH_DIALOG, new Boolean (showFinishDialog), true);
    }

    /**
    * Getter for showRunMultisessionDialog property.
    */
    public boolean getShowRunMultisessionDialog () {
        Boolean b = (Boolean) getProperty (PROP_SHOW_RUN_MULTISESSION_DIALOG);
        if (b == null) return true; // default value
        return b.booleanValue ();
    }

    /**
    * Setter for showRunMultisessionDialog property.
    */
    public void setShowRunMultisessionDialog (boolean showRunMultisessionDialog) {
        putProperty (PROP_SHOW_RUN_MULTISESSION_DIALOG, new Boolean (showRunMultisessionDialog), true);
    }

    /**
    * Getter for runMultisession property.
    */
    public boolean getRunMultisession () {
        Boolean i = (Boolean) getProperty (PROP_RUN_MULTISESSION);
        if (i == null) return true;
        return i.booleanValue ();
    }

    /**
    * Setter for runMultisession property.
    */
    public void setRunMultisession (boolean runMultisession) {
        putProperty (PROP_RUN_MULTISESSION, new Boolean (runMultisession), true);
    }
}
