/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.beans.*;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.JComponent;

import org.openide.loaders.DataObject;
import org.openide.text.Line;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.cookies.LineCookie;
import org.openide.debugger.*;
import org.openide.src.ConstructorElement;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.EnvironmentNotSupportedException;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.execution.NbProcessDescriptor;
import org.openide.util.NbBundle;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.CallbackSystemAction;
import org.openide.util.actions.SystemAction;
import org.openide.src.ClassElement;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.util.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.*;

import javax.swing.SwingUtilities;

/**
* Main Corona debugger class
*
* @author   Jan Jancura
*/
public abstract class DebuggerSupport extends AbstractDebugger {

    private Vector                            breakpoints = new Vector (); //[SERIALIZED]
    // watches
    private Vector                            watches = new Vector ();       //[SERIALIZED]
    
    private transient int                     lastAction = ACTION_START;

    private transient PropertyChangeSupport   pcs;
    // DebuggerListener support */
    private transient Vector                  listener = new Vector ();
    /** Current line support */
    private transient Line                    currentLine = null;
    // properties
    private transient int                     debuggerState = DEBUGGER_NOT_RUNNING;

    
    {
        pcs = new PropertyChangeSupport (this);
    }
    
    

    /**
    * Deserializes debugger.
    */
    protected void setDebugger (AbstractDebugger debugger) {
        // remove all debugger objects.
        removeAllBreakpoints ();
        removeAllWatches ();

        // add a new one
        Breakpoint[] bs = debugger.getBreakpoints ();
        int i, k = bs.length;
        for (i = 0; i < k; i++) {
            if (bs [i].isHidden ()) 
                continue;
            breakpoints.addElement (bs [i]);
            fireBreakpointCreated (bs [i]);
        }

        Watch[] ws = debugger.getWatches ();
        k = ws.length;
        for (i = 0; i < k; i++) {
            if (ws [i].isHidden ()) 
                continue;
            watches.addElement (ws [i]);
            fireWatchCreated (ws [i]);
        }
//        lastAction = ACTION_START;
    }
    
    /**
    * Opens output window & changes debugger state to STARTING
    *
    * @param info debugger info about class to start
    * @exception DebuggerException if an error occures during the start of 
    * the debugger
    */
    public void startDebugger (DebuggerInfo info)
    throws DebuggerException {
        setState (DEBUGGER_STARTING);
        setLastAction (ACTION_START);
    }

    /**
    * Destroyes threads, breakpoints, watches and current line. Debugger
    * state is changed to NOT_RUNNING.
    */
    public void finishDebugger () throws DebuggerException {
        setState (DEBUGGER_NOT_RUNNING);
        refreshBreakpoints ();
        setCurrentLine (null);
    }
    
    /**
    * Getter for last action property.
    *
    * @return one of XXX_ACTION constants
    */
    public int getLastAction () {
        return lastAction;
    }

    /**
    * Sets last action property value.
    */
    protected void setLastAction (int lastAction) {
        int old = this.lastAction;
        this.lastAction = lastAction;
        firePropertyChange (
            PROP_LAST_ACTION,
            new Integer (old),
            new Integer (lastAction)
        );
    }

    /**
    * Setter method for debugger state property.
    *
    * @param newState
    */
    public void setState (final int newState) {
        if (newState == debuggerState) return;
        Object old = new Integer (debuggerState);
        debuggerState = newState;
        firePropertyChange (PROP_STATE, old, new Integer (debuggerState));
    }

    /**
    * Getter for state of debugger.
    *
    * @return one of DEBUGGER_XXX constants
    */
    public int getState () {
        return debuggerState;
    }

    /** Get the current line of debugger.
    *
    * @return current line
    */
    public Line getCurrentLine () {
        return currentLine;
    }

    /**
    * Unmarks current line.
    */
    public void setCurrentLine (Line l) {
        Line old = currentLine;
        if (old != null) old.unmarkCurrentLine ();
        currentLine = l;
        if (currentLine != null) {
            Utils.showInEditor (currentLine);
            currentLine.markCurrentLine ();
        }
        firePropertyChange (PROP_CURRENT_LINE, old, currentLine);
    }

    
    // BREAKPOINTS .......................................................

    /** Creates new breakpoint.
    *
    * @return new line breakpoint
    */
    public CoreBreakpoint createBreakpoint (boolean hidden) {
        return new BreakpointSupport (hidden);
    }

    /** Creates new breakpoint that is assigned to specific line.
    * The line is represented by the Line object that changes its
    * position as the text is modified, etc.
    *
    * @param l line to create breakpoint for
    * @return new line breakpoint
    */
    public Breakpoint createBreakpoint (Line l) {
        BreakpointSupport b = (BreakpointSupport) createBreakpoint (false);
        b.setLine (l);
        return b;
    }

    /** Creates new breakpoint that is assigned to specific line.
    * Allows creation of a hidden breakpoint
    *
    * @param l line to create breakpoint for
    * @param hidden <code>true</code> if the breakpoint should be hidden from the user
    * @return new line breakpoint
    */
    public Breakpoint createBreakpoint (Line l, boolean hidden) {
        BreakpointSupport b = (BreakpointSupport) createBreakpoint (hidden);
        b.setLine (l);
        return b;
    }

    /** Create a new breakpoint assigned to a method (or constructor).
    * The method is represented by a method (or constructor) source element that
    * must have a declaring class.
    *
    * @param method method or constructor with {@link org.openide.src.MemberElement#getDeclaringClass valid} declaring class
    * @return the new breakpoint
    * @exception IllegalArgumentException if the method does not have a declaring class
    */
    public Breakpoint createBreakpoint (ConstructorElement method) {
        BreakpointSupport b = (BreakpointSupport) createBreakpoint (false);
        ClassElement clazz = method.getDeclaringClass ();
        if (clazz != null)
            b.setClassName (clazz.getName ().getFullName ());
        b.setMethodName (method.getName ().getName ());
        return b;
    }

    /** Create a new breakpoint assigned to a method (or constructor).
    * The method is represented by a method (or constructor) source element that
    * must have a declaring class.
    * Allows creation of a hidden breakpoint.
    *
    * @param method method or constructor with {@link org.openide.src.MemberElement#getDeclaringClass valid} declaring class
    * @param hidden <code>true</code> if the breakpoint should be hidden from the user
    * @return the new breakpoint
    * @exception IllegalArgumentException if the method does not have a declaring class
    */
    public Breakpoint createBreakpoint (ConstructorElement method, boolean hidden) {
        BreakpointSupport b = (BreakpointSupport) createBreakpoint (hidden);
        ClassElement clazz = method.getDeclaringClass ();
        if (clazz != null)
            b.setClassName (clazz.getName ().getFullName ());
        b.setMethodName (method.getName ().getName ());
        return b;
    }
     
    protected void addBreakpoint (Breakpoint b) {
        breakpoints.addElement (b);
        fireBreakpointCreated (b);
    }

    /** Find an enabled breakpoint assigned to a specific line.
    * The line is represented by a line object that can change its
    * position as the text is modified.
    *
    * @param l line to find breakpoint at
    * @return the breakpoint or null
    */
    public Breakpoint findBreakpoint (Line l) {
        CoreBreakpoint[] bs = Find.breakpoint (this, l, true, true, true, true);
        if (bs.length > 0) return bs [0];
        return null;
    }

    /** Find an enabled breakpoint assigned to a method.
    * The method is represented by a method source element that
    * must have a declaring class.
    *
    * @param method method with {@link ConstructorElement#getDeclaringClass valid} declaring class
    * @return the breakpoint or null
    * @exception IllegalArgumentException if the method does not have a declaring class
    */
    public Breakpoint findBreakpoint (ConstructorElement method) {
        CoreBreakpoint[] bs = Find.breakpoint (this, method, true, true, true, true);
        if (bs.length > 0) return bs [0];
        return null;
    }

    /** Getter for all breakpoints in the system.
    *
    * @return array of all breakpoints.
    */
    public Breakpoint[] getBreakpoints () {
        CoreBreakpoint[] b;
        if (breakpoints == null) return new Breakpoint [0]; // PATCH for deser.
                                                            // from core/1
        synchronized (breakpoints) {
            b = new CoreBreakpoint [breakpoints.size ()];
            breakpoints.copyInto (b);
        }
        return b;
    }

    /**
    * Removes all non hidden breakpoints.
    */
    public void removeAllBreakpoints () {
        Vector v = (Vector) breakpoints.clone ();
        int i, k = v.size ();
        for (i = k - 1; i >= 0; i--)
            ((Breakpoint) v.elementAt (i)).remove ();
    }

    /**
    * Removes breakpoint.
    *
    * @param b breakpoint to be removed
    */
    protected void removeBreakpoint (Breakpoint b) {
        //    ((CoreBreakpoint)b).clearBreakpoint ();
        breakpoints.removeElement (b);
        fireBreakpointRemoved (b);
    }

    /**
    * Tries to set or remove breakpoints. Is called when debugger starts and finishs.
    */
    protected void refreshBreakpoints () {
        Vector v = (Vector) breakpoints.clone ();
        int i, k = v.size ();
        for (i = 0; i < k; i++)
            if (v.elementAt (i) instanceof BreakpointSupport)
                ((BreakpointSupport) v.elementAt (i)).setBreakpoint ();
    }
    
    
    // WATCHES ..............................................................

    /**
    * Returns array of all watches.
    *
    * @return array of all watches.
    */
    public Watch[] getWatches () {
        Watch[] w;
        if (watches == null) return new Watch [0]; // PATCH for deser.
                                                   // from core/1
        synchronized (watches) {
            w = new Watch [watches.size ()];
            watches.copyInto (w);
        }
        return w;
    }

    /**
    * Removes all watches.
    */
    public void removeAllWatches () {
        Vector v = (Vector) watches.clone ();
        int i, k = v.size ();
        for (i = k - 1; i >= 0; i--)
            removeWatch ((Watch) v.elementAt (i));
    }

    /**
    * Removes watch.
    *
    * @param b watch to be removed
    */
    protected void addWatch (Watch w) {
        watches.addElement (w);
        fireWatchCreated (w);
    }

    /**
    * Removes watch.
    *
    * @param b watch to be removed
    */
    public void removeWatch (Watch w) {
        watches.removeElement (w);
        fireWatchRemoved (w);
    }
    
    
    // EVENTS ..................................................................
    

    /**
    * Adds property change listener.
    *
    * @param l new listener.
    */
    public void addPropertyChangeListener (PropertyChangeListener l) {
        pcs.addPropertyChangeListener (l);
    }

    /**
    * Removes property change listener.
    *
    * @param l removed listener.
    */
    public void removePropertyChangeListener (PropertyChangeListener l) {
        pcs.removePropertyChangeListener (l);
    }
    
    /**
    * Fires property change.
    */
    protected void firePropertyChange (String name, Object o, Object n) {
        pcs.firePropertyChange (name, o, n);
    }

    
    // DebuggerListener support ................................................

    /**
    * This listener notificates about cahnges of breakpoints, watches and threads.
    *
    * @param l listener object.
    */
    public void addDebuggerListener (DebuggerListener l) {
        listener.addElement (l);
    }

    /**
    * Removes debugger listener.
    *
    * @param l listener object.
    */
    public void removeDebuggerListener (DebuggerListener l) {
        listener.removeElement (l);
    }

    /**
    * Notificates about creating a breakpoint.
    *
    * @param breakpoint
    */
    private void fireBreakpointCreated (final Breakpoint breakpoint) {
        Vector l = (Vector) listener.clone ();
        int i, k = l.size ();
        for (i = 0; i < k; i++)
            ((DebuggerListener)l.elementAt (i)).breakpointAdded (breakpoint);
        pcs.firePropertyChange (PROP_BREAKPOINTS, null, null);
    }

    /**
    * Notificates about removing a breakpoint.
    *
    * @param breakpoint
    */
    private void fireBreakpointRemoved (final Breakpoint breakpoint) {
        Vector l = (Vector) listener.clone ();
        int i, k = l.size ();
        for (i = 0; i < k; i++)
            ((DebuggerListener) l.elementAt (i)).breakpointRemoved (breakpoint);
        pcs.firePropertyChange (PROP_BREAKPOINTS, null, null);
    }

    /**
    * Notificates about creating a AbstractWatch.
    *
    * @param AbstractWatch
    */
    private void fireWatchCreated (final Watch watch) {
        Vector l = (Vector) listener.clone ();
        int i, k = l.size ();
        for (i = 0; i < k; i++)
            ((DebuggerListener) l.elementAt (i)).watchAdded (watch);
        pcs.firePropertyChange (PROP_WATCHES, null, null);
    }

    /**
    * Notificates about removing a AbstractWatch.
    *
    * @param AbstractWatch
    */
    private void fireWatchRemoved (final Watch watch) {
        Vector l = (Vector) listener.clone ();
        int i, k = l.size ();
        for (i = 0; i < k; i++)
            ((DebuggerListener)l.elementAt (i)).watchRemoved (watch);
        pcs.firePropertyChange (PROP_WATCHES, null, null);
    }
}

