/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.io.*;
import java.util.*;

import javax.swing.SwingUtilities;

import org.openide.TopManager;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;

public class IOManager {

    /** Debugger output constant. */
    public static final int                   DEBUGGER_OUT = 1;
    /** Process output constant. */
    public static final int                   PROCESS_OUT = 2;
    /** Status line output constant. */
    public static final int                   STATUS_OUT = 4;
    /** All outputs constant. */
    public static final int                   ALL_OUT = DEBUGGER_OUT + 
                                                PROCESS_OUT + STATUS_OUT;

    /** Standart process output constant. */
    public static final int                   STD_OUT = 1;
    /** Error process output constant. */
    public static final int                   ERR_OUT = 2;

    private static Hashtable                  closedOutputs = new Hashtable ();
    
    protected transient InputOutput           io = null;
    private transient String                  name;

    /** Output window */
    private transient OutputWriter            output;
    private transient OutputWriter            error;

    /** output writer Thread */
    private transient Thread                  owThread = null;

    
    public IOManager (String name, boolean newIO, boolean processOutSeparated) {
        this.name = name;
        removeClosedOutputs (name);
        if (newIO)
            io = TopManager.getDefault ().getIO (name);
        else {
            io = TopManager.getDefault ().getIO (name, false);
            if (io == null)
                io = TopManager.getDefault ().getIO (name, true);
        }
        if (processOutSeparated)
            io.setErrSeparated (true);
        io.setOutputVisible (true);
        io.setErrVisible (true);
        io.setInputVisible (false);
        output = io.getOut ();
        error = io.getErr ();
        io.setFocusTaken (true);
        io.select ();
        try {
            output.reset ();
            error.reset ();
        } catch (java.io.IOException e) {
        }
    }
    
    /**
     * Returns used InputOutput instance.
     */
    public InputOutput getIO () {
        return io;
    }
    
    /**
    * 
    */
    public void select () {
        io.select ();
    }
    
    /**
    * Prints given text to the output.
    */
    public void print (final String text, final int where) {
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                if ((where & PROCESS_OUT) != 0) output.print (text);
                if ((where & DEBUGGER_OUT) != 0) error.print (text);
                if ((where & STATUS_OUT) != 0) TopManager.getDefault ().setStatusText (text);
            }
        });
    }


    /**
    * Prints given text to the output.
    */
    public void println (final String text) {
        println (text, DEBUGGER_OUT);
    }
    
    /**
    * Prints given text to the output.
    */
    public void println (final String text, final int where) {
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                if ((where & PROCESS_OUT) != 0) output.println (text);
                if ((where & DEBUGGER_OUT) != 0) error.println (text);
                if ((where & STATUS_OUT) != 0) TopManager.getDefault ().setStatusText (text);
            }
        });
    }

    /**
    * Shows output and error from this proces in output window.
    */
    public void showOutput (final Process process, int what, final int where) {
        if (process == null) throw new NullPointerException ();
        Thread t = null;
        if ((what | STD_OUT) != 0) {
            t = new Thread (new Runnable () {
                public void run () {
                    BufferedReader input = new BufferedReader (
                        new InputStreamReader (process.getInputStream ())
                    );
                    String s;
                    try {
                        while ((s = input.readLine ()) != null)
                            println (s, where);
                    } catch (IOException e) {
                    }
                }
            }, "Debugger input reader thread"); // NOI18N
            t.setPriority (Thread.MIN_PRIORITY);
            t.start ();
        }

        if ((what | ERR_OUT) != 0) {
            t = new Thread (new Runnable () {
                public void run () {
                    BufferedReader error = new BufferedReader (
                        new InputStreamReader (process.getErrorStream ())
                    );
                    String d;
                    try {
                        while ((d = error.readLine ()) != null)
                            println (d, where);
                    } catch (IOException e) {
                    }
                }
            }, "Debugger error reader thread"); // NOI18N
            t.setPriority (Thread.MIN_PRIORITY);
            t.start ();
        }
    }

    public void connectInput (final Process process) {
        if (process == null) throw new NullPointerException ();
        if (io == null) return;
        io.setInputVisible (true);
        io.flushReader ();
        final Reader input = io.getIn ();
        owThread = new Thread (
            new Runnable () {
                public void run () {
                    OutputStream out = process.getOutputStream ();
                    int b;
                    try {
                        while ((b = input.read ()) != -1) {
                            out.write (b);
                            out.flush ();
                        }
                    } catch (IOException e) {
                    }
                }
            },  
            "Debugger output writer thread" // NOI18N
        );
        owThread.setPriority (Thread.MIN_PRIORITY);
        owThread.start ();
    }

    /**
     * Stops communication between InputOutput and process.
     */
    public void stop () {
        addClosedOutput (name, io);
        if ((owThread != null) && (owThread.isAlive ()))
            try {
                owThread.stop ();
                owThread = null;
            }
            catch (SecurityException e) {
            }
    }
    
    private static synchronized void addClosedOutput (String name, InputOutput io) {
        List list = (List) closedOutputs.get (name);
        if (list == null) {
            list = new LinkedList ();
            closedOutputs.put (name, list);
        }
        list.add (io);
    }
    
    private static synchronized void removeClosedOutputs (String name) {
        List list = (List) closedOutputs.get (name);
        if (list != null) {
            Iterator iter = list.iterator ();
            while (iter.hasNext ())
                ((InputOutput) iter.next ()).closeInputOutput ();
            closedOutputs.remove (name);
        }
    }
    
}
