/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.util.ResourceBundle;
import java.text.MessageFormat;
import java.io.ObjectOutputStream;
import java.io.IOException;
import javax.swing.JComponent;

import org.openide.debugger.Debugger;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;
import org.openide.cookies.SourceCookie;
import org.openide.cookies.EditorCookie;
import org.openide.text.NbDocument;

import org.openide.src.*;

import org.netbeans.modules.debugger.CoreBreakpoint;
import org.netbeans.modules.debugger.support.util.Utils;
import org.netbeans.modules.debugger.support.actions.AddBreakpointAction;

/**
* Abstract implementation of breakpoint event on line.
*
* @author   Jan Jancura
*/
public abstract class LineBreakpointEvent extends CoreBreakpoint.Event {

    static final long             serialVersionUID = 5611559342537392476L;
    
    /** Property name constant */
    public static final String    PROP_SOURCE_NAME = "sourceName"; // NOI18N
    /** Property name constant */
    public static final String    PROP_LINE_NUMBER = "lineNumber"; // NOI18N

    public static final String    TYPE_NAME = "Line";

    
    // variables .....................................................................................................

    private int                   lineNumber;
    protected Line                line = null;
    protected String              sourceName;

    {
        lineNumber = ((AddBreakpointAction) SystemAction.get
            (AddBreakpointAction.class)).getCurrentLineNumber ();
        sourceName = ((AddBreakpointAction) SystemAction.get
           (AddBreakpointAction.class)).getCurrentClassName ();
        line = Utils.getCurrentLine (sourceName, lineNumber);
    }


    // init .....................................................................................................
    
    private void writeObject (ObjectOutputStream oos) throws IOException {
        if (line != null) lineNumber = line.getLineNumber () + 1;
        oos.defaultWriteObject ();
    }

    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
    }


    // Event implementation ....................................................................................

    /**
    * Returns specific properties of this event.
    */
    public Node.Property[] getProperties () {
        ResourceBundle bundle = NbBundle.getBundle (LineBreakpointEvent.class);
        return new Node.Property[] {
           new PropertySupport.ReadWrite (
               PROP_SOURCE_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_source_name"),
               bundle.getString ("HINT_breakpoint_source_name")
           ) {
               public Object getValue () {
                   return getSourceName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setSourceName (((String)val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_LINE_NUMBER,
               Integer.TYPE,
               bundle.getString ("PROP_breakpoint_line_number"),
               bundle.getString ("HINT_breakpoint_line_number")
           ) {
               public Object getValue () {
                   return new Integer (getLineNumber ());
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setLineNumber (((Integer)val).intValue ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           }
       };
    }

    /**
    * Returns name of type of this event.
    */
    public String getTypeName () {
        return TYPE_NAME;
    }

    /**
    * Returns display name of this event.
    */
    public String getTypeDisplayName () {
        return NbBundle.getBundle (LineBreakpointEvent.class).getString ("CTL_Line_event_type_name");
    }

    /**
    * Returns display name of this instance of event. It will be used
    * as the name of the breakpoint.
    */
    public String getDisplayName () {
        // 1.1 changes ...
        if (line != null)
            return line.getDisplayName ();
        // 1.1 changes end ...
        return new MessageFormat (
                   NbBundle.getBundle (LineBreakpointEvent.class).getString ("CTL_Line_event_name")
               ).format (new Object[] {getSourceName (), new Integer (getLineNumber ())});
    }

    /**
    * Returns lines to highlite in the editor.
    */
    public Line[] getLines () {
        if (line == null) return null;
        return new Line [] {line};
    }

    /**
    * Returns customizer visual component.
    */
    public JComponent getCustomizer () {
        return new LineBreakpointPanel (this);
    }


    // properties ........................................................................................

    /**
    * Set name of class to stop on.
    */
    public void setSourceName (String cn) {
        if ( (cn == sourceName) ||
             ((cn != null) && (sourceName != null) && sourceName.equals (cn))
        ) return;
        Object old = sourceName;
        sourceName = cn;
        int ln = getLineNumber ();
        line = null;
        if ( (sourceName != null) &&
                (sourceName.length () > 0) &&
                (ln > 0)
           ) if ((line = Utils.getLine (sourceName, ln)) != null
                    ) lineNumber = -1;
            else lineNumber = ln;
        firePropertyChange (PROP_SOURCE_NAME, old, sourceName);
    }

    /**
    * Get name of class to stop on.
    */
    public String getSourceName () {
        return sourceName;
    }

    /**
    * Set name of class to stop on.
    */
    public void setsourceName (String cn) {
        if ( (cn == sourceName) ||
             ((cn != null) && (sourceName != null) && sourceName.equals (cn))
        ) return;
        Object old = sourceName;
        sourceName = cn;
        firePropertyChange (PROP_SOURCE_NAME, old, sourceName);
    }
    
    /**
    * Get number of line to stop on.
    */
    public int getLineNumber() {
        if (line != null) {
            int num = line.getLineNumber () + 1;
            return num;
        }
        return lineNumber;
    }

    /**
    * Set number of line to stop on.
    */
    public void setLineNumber (int ln) {
        if (ln == lineNumber) return;
        int old = lineNumber;
        if (ln > 0) {
            line = Utils.getLine (sourceName, ln);
            lineNumber = ln;
        }
        else {
            line = null;
            lineNumber = -1;
        }
        firePropertyChange (
            PROP_LINE_NUMBER,
            new Integer (old),
            new Integer (getLineNumber ())
        );
    }

    /**
    * Get line to stop on.
    */
    public Line getLine () {
        return line;
    }

    /**
    * Set line to stop on.
    */
    public void setLine (Line l) {
        if (line == l) return;
        Object old = line;
        line = l;
        if (line != null) {
            SourceCookie.Editor sourceCookie = (SourceCookie.Editor) line.getDataObject ().getCookie (SourceCookie.Editor.class);
            EditorCookie editorCookie = (EditorCookie) line.getDataObject ().getCookie (EditorCookie.class);
            sourceName = null;
            if ((sourceCookie != null) && (editorCookie != null)) {
                try {
                    int offset = NbDocument.findLineOffset (editorCookie.openDocument (), l.getLineNumber());
                    sourceName = Utils.getClassNameForElement (sourceCookie.findElement (offset));
                } catch (java.io.IOException e) {
                }
            }
            if ((sourceName == null) || (sourceName.length () < 1))
                sourceName = line.getDataObject ().getPrimaryFile ().getPackageName ('.');
        }
        firePropertyChange (PROP_LINE_NUMBER, old, line);
    }
}
