/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support;

import java.io.File;

import org.openide.debugger.DebuggerInfo;
import org.openide.execution.NbProcessDescriptor;

import org.netbeans.modules.debugger.AbstractDebuggerInfo;


/**
* Contains information about a class to debug.
* Consists of these pieces of information:
* <UL>
* <LI>the class to run
* <LI>parameters for its main method
* <LI>a class name to stop execution in, if desired
* </UL>
*
* @author Jan Jancura
*/
public class ProcessDebuggerInfo extends AbstractDebuggerInfo
    implements java.io.Serializable {

    private NbProcessDescriptor         processDescriptor;
    private String                      classPath;
    private String                      bootClassPath;
    private String                      repositoryPath;
    private String                      libraryPath;
    private File                        workingDirectory;
    private boolean                     classic;
    private String                      processName;

    
    /**
    * Construct a new <code>DebuggerInfo</code> with the class to run, parameters, and a class to stop at.
    *
    * @param className name of debugged class
    * @param argv command-line arguments used for debugging this class; may be empty but not <code>null</code>
    * @param stopClassName name of class to stop in (may be <code>null</code>)
    */
    public ProcessDebuggerInfo (
        String className,
        String[] argv,
        String stopClassName,
        NbProcessDescriptor processDescriptor,
        String classPath,
        String bootClassPath,
        String repositoryPath,
        String libraryPath,
        File workingDirectory,
        boolean classic,
        String processName
    ) {
        super (className, argv, stopClassName);
        this.processDescriptor = processDescriptor;
        this.classPath = classPath;
        this.bootClassPath = bootClassPath;
        this.repositoryPath = repositoryPath;
        this.libraryPath = libraryPath;
        this.workingDirectory = workingDirectory;
        this.classic = classic;
        this.processName = processName;
    }

    /**
     * @deprecated
     */
    public ProcessDebuggerInfo (
        String className,
        String[] argv,
        String stopClassName,
        NbProcessDescriptor processDescriptor,
        String classPath,
        String bootClassPath,
        String repositoryPath,
        String libraryPath,
        boolean classic
    ) {
        super (className, argv, stopClassName);
        this.processDescriptor = processDescriptor;
        this.classPath = classPath;
        this.bootClassPath = bootClassPath;
        this.repositoryPath = repositoryPath;
        this.libraryPath = libraryPath;
        this.workingDirectory = null;
        this.classic = classic;
    }
    
    /**
    * Getter for debuggerProcess property. 
    */
    public NbProcessDescriptor getDebuggerProcess () {
        return processDescriptor;
    }

    /** Get the classpath or <code>null</code>, if default one must be used.
    *
    * @return classpath or <code>null</code>
    */
    public String getClassPath () {
        return classPath;
    }

    /** Get the boot classpath or <code>null</code>, if default one must be used.
    *
    * @return boot classpath or <code>null</code>
    */
    public String getBootClassPath () {
        return bootClassPath;
    }

    /** Get the repository path or <code>null</code>, if default one must be used.
    *
    * @return repository path  or <code>null</code>
    */
    public String getRepositoryPath () {
        return repositoryPath;
    }

    /** Get the library path or <code>null</code>, if default one must be used.
    *
    * @return library path  or <code>null</code>
    */
    public String getLibraryPath () {
        return libraryPath;
    }

    /** Get the working directory or <code>null</code>.
    *
    * @return working directory  or <code>null</code>
    */
    public File getWorkingDirectory () {
        return workingDirectory;
    }
    
    /**
    * Getter method for classic property.
    */
    public boolean isClassic () {
        return classic;
    }
    
    /**
    * Get the user-visible display name of debugged process.
    *
    * @return display name of debugged process
    */
    public String getProcessName () {
        return processName;
    }
    
    public String getLocationName () {
        return "localhost";
    }
}
