/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java;

import java.beans.*;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.JComponent;

import org.openide.loaders.DataObject;
import org.openide.text.Line;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.cookies.LineCookie;
import org.openide.debugger.*;
import org.openide.src.ConstructorElement;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.EnvironmentNotSupportedException;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.execution.NbProcessDescriptor;
import org.openide.util.NbBundle;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.CallbackSystemAction;
import org.openide.util.actions.SystemAction;
import org.openide.src.ClassElement;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.util.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.*;

import javax.swing.SwingUtilities;

/**
* Main Corona debugger class
*
* @author   Jan Jancura
*/
public abstract class JavaDebugger extends DebuggerSupport 
implements ThreadsRoot, SharedIO, ValidatorHolder, SuspendResumeSupport {


    // static ........................................................................

    static final long                         serialVersionUID = 2791304615739651906L;

    /** bundle to obtain text information from */
    private static ResourceBundle             bundle;

    /*
     * Returns bundle for this package.
     *
     * @return bundle for this package
     */
    static String getString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JavaDebugger.class);
        return bundle.getString (s);
    }
    

    // variables .................................................................

    /** = true if debugger IO name should be "Debugger" and 
      * = false if it should be name of process */
    private transient boolean                 generalIOName = true;
    /** Helper class for updating values of variales */
    private transient Validator               validator = new Validator ();


    boolean                                   showMessages = false;    //[SERIALIZED]
    boolean                                   followedByEditor = true; //[SERIALIZED]
    
    // IO support ...
    private IOManager                         io;
    private String                            sharedIOName;


    // Debugger implementation .................................................

    /**
    * Opens output window & changes debugger state to STARTING
    *
    * @param info debugger info about class to start
    * @exception DebuggerException if an error occures during the start of 
    * the debugger
    */
    public void startDebugger (DebuggerInfo info)
    throws DebuggerException {
        // open tab in Output Window
        if (sharedIOName != null) 
            io = new IOManager (sharedIOName, false, true);
        else
            io = new IOManager (getProcessName (info), true, true);
        sharedIOName = null;
        io.select ();
        io.println (getString ("CTL_Debugger_starting"), io.STATUS_OUT);
        
        setState (DEBUGGER_STARTING);
        setLastAction (ACTION_START);
        ((CreateVariableAction) SystemAction.get (CreateVariableAction.class)).
        changeEnabled (true);
    }

    /**
    * Destroyes threads, breakpoints, watches and current line. Debugger
    * state is changed to NOT_RUNNING.
    */
    public synchronized void finishDebugger () throws DebuggerException {
        if (getState () == DEBUGGER_NOT_RUNNING) return;

        io.stop ();
        JavaThreadGroup ttg = (JavaThreadGroup) getThreadsRoot ();
        if (ttg != null) ttg.removeAll ();
        setState (DEBUGGER_NOT_RUNNING);
        ((CreateVariableAction) SystemAction.get (CreateVariableAction.class)).changeEnabled (false);
        refreshBreakpoints ();
        getValidator ().clear ();
        updateWatches ();
        setCurrentLine (null);
    }

    /**
    * Trace into support.
    */
    public void traceInto () throws DebuggerException {
        io.println (getString ("CTL_Debugger_running"), io.STATUS_OUT);
        setState (DEBUGGER_RUNNING);
        setLastAction (ACTION_TRACE_INTO);
        setCurrentLine (null);
    }

    /**
    * Trace over support.
    */
    public void traceOver () throws DebuggerException {
        io.println (getString ("CTL_Debugger_running"), io.STATUS_OUT);
        setState (DEBUGGER_RUNNING);
        setLastAction (ACTION_TRACE_OVER);
        setCurrentLine (null);
    }

    /**
    * Go support.
    */
    public void go () throws DebuggerException {
        io.println (getString ("CTL_Debugger_running"), io.STATUS_OUT);
        setState (DEBUGGER_RUNNING);
        setLastAction (ACTION_GO);
        setCurrentLine (null);
    }

    /**
    * Step out support.
    */
    public void stepOut () throws DebuggerException {
        io.println (getString ("CTL_Debugger_running"), io.STATUS_OUT);
        setState (DEBUGGER_RUNNING);
        setLastAction (ACTION_STEP_OUT);
        setCurrentLine (null);
    }
    
    
    // AbstractDebugger implementation .........................................

    /**
    * Prints given text to the output.
    */
    public void println (String text) {
        io.println (text);
    }

    
    // SharedIO impl ...........................................................


    /**
    * Returns.
    */
    public InputOutput getIO () {
        if (io != null)
            return io.getIO ();
        return null;
    }

    /**
    * Sets name of Output window to be shared.
    */
    public void setSharedIOName (String name) {
        sharedIOName = name;
    }
    
    
    // ValidatorCarrier interface .............................................. 
    
    /**
    * Returns validator.
    */
    public Validator getValidator () {
        return validator;
    }

    /**
    * Sets validator.
    */
    public void setValidator (Validator validator) {
        this.validator = validator;
    }

    
    // ThreadRoot implementation ...............................................

    /**
    * Returns current thread or null.
    */
    public abstract AbstractThread getCurrentThread ();

    /**
    * Sets current thread. If thread is null, unsets curent thread.
    */
    public abstract void setCurrentThread (AbstractThread thread);
    
    
    // other methods ...........................................................

    /**
    * Sets debugger state.
    */
    public void setSuspended (boolean suspended) {
        ((JavaThreadGroup) getThreadsRoot ()).setSuspended (suspended);
        if (suspended)
            setState (DEBUGGER_STOPPED);
        else
            setState (DEBUGGER_RUNNING);
    }

    /**
    * @return name of proces for given DebuggerInfo.
    */
    public String getProcessName (DebuggerInfo info) {
        if (info instanceof ProcessDebuggerInfo) {
            String s = ((ProcessDebuggerInfo) info).getProcessName ();
            if (s != null) return s;
        }
        return (info.getStopClassName () != null) ? info.getStopClassName () :
                   info.getClassName ();
    }

    /**
    * @return name of location for given DebuggerInfo.
    */
    public abstract String getLocationName (DebuggerInfo info);


    // watches .................................


    /**
    * Update values of watches.
    */
    public void updateWatches () {
        getValidator ().validate ();
        //    int i, k = watch.size ();
        //    for (i = 0; i < k; i++) ((AbstractWatch)watch.elementAt (i)).validate ();
    }

    /**
    * Disconnects from Virtual Machine.
    */
    public void disconnect () throws DebuggerException {
    }

    /**
    * Reconnects to disconnected Virtual Machine.
    */
    public void reconnect () throws DebuggerException {
    }

    /**
    * Unmarks current line.
    */
    public boolean canBeCurrent (Line l, boolean onBreakpoint) {
        return l.canBeMarkedCurrent (
            onBreakpoint ? ACTION_BREAKPOINT_HIT : getLastAction (), 
            getCurrentLine ()
        );
    }

    
    // helper methods .................................

    /**
    * Updates state of debugger.
    *
    * @param newState
    */
    /*  public void updateDebuggerState () {
        try {
          AbstractThread currentThread = getCurrentThread ();
          if ((currentThread != null) && currentThread.isSuspended ())
            setState (DEBUGGER_STOPPED);
          else 
            setState (DEBUGGER_RUNNING);
        } catch (Throwable e) {
          if (e instanceof ThreadDeath) throw (ThreadDeath)e;
        }
      }*/
    
    public IOManager getIOManager () {
        return io;
    }
}

