/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java;

import java.beans.*;
import java.util.*;

import javax.swing.SwingUtilities;

import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerException;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.util.Utils;



public abstract class JavaThread extends AbstractThread 
implements Location, VariablesProducer, CallStackProducer {

    public static final String           PROP_ADITIONAL_THREAD_INFO =
        "aditionalThreadInfo";
    public static final String           PROP_CALL_STACK = "callStack";

    
    // variables ...............................................................

    protected JavaDebugger               debugger;
    private boolean                      current = false;
    private PropertyChangeSupport        pcs;
    /** Points to parent thread group or contains null. */
    private JavaThreadGroup              parentThreadGroup;
    private Node                         aditionalThreadInfo;
//    private Vector                       callStackFilters = new Vector ();

    
    // init ....................................................................

    protected JavaThread (
        JavaDebugger debugger, 
        JavaThreadGroup parentThreadGroup
    ) {
        this.parentThreadGroup = parentThreadGroup;
        this.debugger = debugger;
        pcs = new PropertyChangeSupport (this);
    }


    // AbstractThread methods ..................................................

    /**
    * Returns true if this thread is current.
    */
    public boolean isCurrent () {
        return current;
    }

    /**
    * Sets this thread current thread.
    * Calls debugger.setCurrentThread (this) and
    * marks current line in editor or show cursor on some stack frame
    * above, if current class is sourceless.
    *
    * @return true if the current thread switching is successfull.
    */
    public void setCurrent (boolean current) {
        this.current = current;
        AbstractThread tt = debugger.getCurrentThread ();
        if (!current) {
            // set non-current
            if ( (tt != null) &&
                    tt.equals (this)
               ) debugger.setCurrentThread (null);
            if (getParentThreadGroup () != null)
                getParentThreadGroup ().setCurrent (false);
            SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    pcs.firePropertyChange (null, null, null);
                }
            });
            debugger.setCurrentLine (null);
            return;
        }

        // make old current thread non-current
        if ((tt != null) && (tt != this))
            tt.setCurrent (false);

        debugger.setState (
            isSuspended () ? Debugger.DEBUGGER_STOPPED : 
                Debugger.DEBUGGER_RUNNING
        );
        
        // change thread state (icon..)
        debugger.setCurrentThread (this);
        if (getParentThreadGroup () != null)
            getParentThreadGroup ().setCurrent (true);
        SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    pcs.firePropertyChange (AbstractThread.PROP_CURRENT, null, null);
                }
            });

        debugger.updateWatches ();
        // mark current line in editor
        debugger.setCurrentLine (null);
        int ln = getLineNumber ();
        if (ln >=0) {
            Line l = getLine ();
            if (l != null) {
                debugger.setCurrentLine (l);
                return;
            }
        }

        // set cursor otherwise
        Location[] stack = getCallStack ();
        int i, k = stack.length;
        for (i = 1; i < k; i++) {
            if ( Utils.showInEditor (
                        stack [i].getLine ()
                    ) != null
               ) return;
        }
        return;
    }

    /**
    * If this thread is suspended returns line object representing position
    * in the editor where this thread is stopped. If thread has 
    * no stack returns null.
    *
    * @throws DebuggerException if informations about source are 
    *   not included or some other error occurres.
    * @return line object representing position where this thread is stopped.
    */
    public Line getLine () {
        String sn = getSourceName ();
        if (sn != null)
            return Utils.getLineForSource (
                getClassName (),
                getSourceName (),
                getLineNumber ()
            );
        return Utils.getLine (
            getClassName (),
            getLineNumber ()
        );
    }

    /**
    * Adds property change listener for this thread. All property changes are fired
    * together in one PropertyChangeEvent with name of property == null.
    *
    * @param l listener to be added.
    */
    public void addPropertyChangeListener (PropertyChangeListener l) {
        pcs.addPropertyChangeListener (l);
    }

    /**
    * Removes property change listener from this thread.
    *
    * @param l listener to be removed.
    */
    public void removePropertyChangeListener (PropertyChangeListener l) {
        pcs.removePropertyChangeListener (l);
    }

    
    // main methods ............................................................
    
    /**
    * Returns true if this thread is suspended.
    *
    * @throw DebuggerException if some problem occurs.
    * @return true if this thread is suspended.
    */
    public abstract boolean isSuspended ();

    /**
    * Setter method for the suspend property.
    *
    * @throw DebuggerException if some problem occurs.
    * @param suspend true if this thread might be suspend.
    */
    public void setSuspended (final boolean suspend) throws DebuggerException {
        if (isCurrent ()) {
            if (suspend) {
                setCurrent (true);
                debugger.updateWatches ();
                debugger.setState (debugger.DEBUGGER_STOPPED);
            } else {
                debugger.setCurrentLine (null);
                debugger.setState (debugger.DEBUGGER_RUNNING);
            }
        }
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                firePropertyChange (PROP_CALL_STACK, null, null);
            }
        });
    }
    
    /**
    * Returns parent thread group.
    *
    * @return parent thread group.
    */
    public JavaThreadGroup getParentThreadGroup () {
        return parentThreadGroup;
    }

    /**
    * If this thread is suspended returns line number where 
    * this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return line number where this thread is stopped.
    */
    public abstract int getLineNumber ();

    /**
    * Returns string representation of the current state of this 
    * thread (depends on debugger implementation).
    *
    * @throw DebuggerException if some problem occurs.
    * @return string representation of the current state of this thread.
    */
    public abstract String getState ();

    /**
    * If this thread is suspended returns class name 
    * where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return class name where this thread is stopped.
    */
    public abstract String getClassName ();

    /**
    * If this thread is suspended returns method 
    * name where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return method name where this thread is stopped.
    */
    public abstract String getMethod ();

    /**
    * Returns current stack depth.
    *
    * @throw DebuggerException if some problem occurs.
    * @return current stack depth.
    */
    public abstract int getStackDepth ();

    /**
    * Stops this thread.
    *
    * @throw DebuggerException if some problem occurs.
    */
    public abstract void stop () throws DebuggerException;

    /**
    * If this thread is suspended returns current call stack.
    *
    * @return current call stack.
    */
    public Location[] getFilteredCallStack () {
        return getCallStack ();
    }

/*    public void addCallStackFilter (CallStackFilter csf) {
        callStackFilters.addElement (csf);
    }

    public boolean removeCallStackFilter (CallStackFilter csf) {
        return callStackFilters.removeElement (csf);
    }*/

    /**
    * If this thread is suspended returns static and instance variables in current context.
    *
    * @return static and instance variables in current context.
    */
    public abstract AbstractVariable[] getContextVariables ();
    
    /**
    * Returns name of soucrce file of this frame or null if source is unknown.
    *
    * @return name of soucrce file of this frame or null if source is unknown
    */
    public abstract String getSourceName ();

    public void setAditionalThreadInfo (Node info) {
        //S ystem.out.println("JavaThread.addAditionalThreadInfo " + info  + " : "  + aditionalThreadInfo);
        if (aditionalThreadInfo == info) return;
        Node old = aditionalThreadInfo;
        aditionalThreadInfo = info;
        firePropertyChange (
            PROP_ADITIONAL_THREAD_INFO,
            old,
            aditionalThreadInfo
        );
    }

    public Node getAditionalThreadInfo () {
        if (aditionalThreadInfo == null) return null;
        return aditionalThreadInfo.cloneNode ();
    }

    protected void firePropertyChange (String s, Object o, Object n) {
        pcs.firePropertyChange (s, o, n);
    }

    /**
    * Returns true if it has no subelements.
    *
    * @return true if it has no subelements.
    */
    public boolean isLeaf () {
        return false;
    }
    
    /**
    */
    public AbstractVariable[] getFilteredVariables () {
        return getVariables ();
    }
    
    
    public String getLocationName () {
        CallStackFrame[] ls = (CallStackFrame[]) getCallStack ();
        if (ls.length < 1) return "";
        return ls [0].getClassName () + '.' +
               ls [0].getMethodName () + " : " + // NOI18N
               ls [0].getLineNumber ();
    }
    
    public String toString () {
        try {
            return "The Thread: " + getName () + " (" + super.toString () + ")"; // NOI18N
        } catch (Exception e) {
            return super.toString ();
        }
    }
}
