/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java;

import java.util.*;
import java.beans.*;

import javax.swing.SwingUtilities;

import org.openide.debugger.DebuggerException;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;


public abstract class JavaThreadGroup extends AbstractThread 
implements ThreadsProducer {

    /** Current property name. */
    public static final String      PROP_CURRENT = "current"; // NOI18N

    // variables .................................................................

    /** Current property value. */
    private boolean                 current = false;
    /** JavaThreadGroup-s are stored here.
        used in: getThreadGroups () */
    private Vector                  threadGroups = new Vector ();
    /** AbstractThread-s are stored here.
        used in: getThreads () */
    private Vector                  threads = new Vector ();
    /** ThreadGroupListener-s */
    private PropertyChangeSupport   pcs = new PropertyChangeSupport (this);
    /** Points to parent thread group or contains null. */
    private JavaThreadGroup         parentThreadGroup;


    // init ....................................................................

    /**
    * Creates ThreadGroup for some remote thread group.
    */
    protected JavaThreadGroup (JavaThreadGroup parentThreadGroup) {
        this.parentThreadGroup = parentThreadGroup;
    }


    // AbstractThread methods ..................................................

    /**
    * Getter for the current property.
    *
    * @return true if this thread group contains current thread or current 
    *  thread group.
    */
    public boolean isCurrent () {
        return current;
    }

    /**
    * Sets property current. Sets this property for parent thread group too.
    */
    public void setCurrent (boolean current) {
        if (this.current == current) return;
        this.current = current;
        firePropertyChange (PROP_CURRENT, new Boolean (!current), new Boolean (current));
        JavaThreadGroup ttg = getParentThreadGroup ();
        if (ttg == null) return;
        ttg.setCurrent (current);
    }

    /**
    * Adds property change listener for this thread. All property changes are fired
    * together in one PropertyChangeEvent with name of property == null.
    *
    * @param l listener to be added.
    */
    public void addPropertyChangeListener (PropertyChangeListener l) {
        pcs.addPropertyChangeListener (l);
    }

    /**
    * Removes property change listener from this thread.
    *
    * @param l listener to be removed.
    */
    public void removePropertyChangeListener (PropertyChangeListener l) {
        pcs.removePropertyChangeListener (l);
    }

    
    // main methods ............................................................
    
    /**
    * Returns array of thread in this thread group.
    *
    * @return array of thread in this thread group.
    */
    public JavaThread[] getJavaThreads () {
        return (JavaThread[]) threads.toArray (
            new JavaThread [threads.size ()]
        );
    }

    /**
    * Returns array of thread groups in this thread group.
    *
    * @return array of thread groups in this thread group.
    */
    public JavaThreadGroup[] getThreadGroups () {
        return (JavaThreadGroup[]) threadGroups.toArray (
            new JavaThreadGroup [threadGroups.size ()]
        );
    }

    /**
    * Returns parent thread group.
    *
    * @return parent thread group.
    */
    public JavaThreadGroup getParentThreadGroup () {
        return parentThreadGroup;
    }

    /**
    * Suspends / resumes all threads in this thread group recursivelly.
    */
    public void setSuspended (boolean suspended) {
        JavaThread [] threads = (JavaThread[]) getJavaThreads ();
        int i, k = threads.length;
        for (i = 0; i < k; i++)
            try {
                threads [i].setSuspended (suspended);
            } catch (DebuggerException e ) {
            }
        JavaThreadGroup[] groups = (JavaThreadGroup[]) getThreadGroups ();
        k = groups.length;
        for (i = 0; i < k; i++)
            groups [i].setSuspended (suspended);
    }


    // helper methods .....................................................................

    /**
    * Adds given thread group to this threadgroup.
    */
    protected void addThreadGroup (JavaThreadGroup threadGroup) {
        threadGroups.add (threadGroup);
        threadProducer.addThread (threadGroup);
    }

    /**
    * Removes given thread group from this thread group.
    */
    protected void removeThreadGroup (JavaThreadGroup threadGroup) {
        threadGroups.remove (threadGroup);
        threadProducer.removeThread (threadGroup);
    }

    /**
    * Adds given thread to this threadgroup.
    */
    protected void addThread (JavaThread thread) {
        threads.add (thread);
        threadProducer.addThread (thread);
    }

    /**
    * Removes given thread from this thread group.
    */
    protected void removeThread (JavaThread thread) {
        threads.remove (thread);
        threadProducer.removeThread (thread);
    }

    /**
    * Removes all threads & thread groups.
    */
    protected void removeAll () {
        int i = threadGroups.size () - 1;
        for (; i >= 0; i--)
            removeThreadGroup ((JavaThreadGroup) threadGroups.get (i));
        i = threads.size () - 1;
        for (; i >= 0; i--)
            removeThread ((JavaThread) threads.get (i));
    }

    protected void firePropertyChange (String name, Object oldValue, Object newValue) {
        pcs.firePropertyChange (name, oldValue, newValue);
    }

    
    // ThreadProducer ..........................................................
    
    private ThreadPAdapter threadProducer = new ThreadPAdapter ();
    
    /**
    * Returns array of thread in this thread group.
    *
    * @return array of thread in this thread group.
    */
    public AbstractThread[] getThreads () {
        return threadProducer.getThreads ();
    }

    /**
    * Adds thread group listener for this thread group.
    *
    * @param l listener to be added.
    */
    public void addThreadListener (ThreadListener l) {
        threadProducer.addThreadListener (l);
    }

    /**
    * Removes thread group listener from this thread group.
    *
    * @param l listener to be removed.
    */
    public void removeThreadListener (ThreadListener l) {
        threadProducer.removeThreadListener (l);
    }
}
