/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is Forte for Java, Community Edition. The Initial
 * Developer of the Original Code is Sun Microsystems, Inc. Portions
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.modules.debugger.support.java.nodes;

import java.util.*;

import org.openide.src.nodes.*;
import org.openide.src.*;
import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.debugger.DebuggerException;
import org.openide.util.actions.SystemAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.nodes.VariableNode;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* The class is final only for performance reasons.
* Can be unfinaled if desired.
*
* @author   dprusa
*/
public class ClassNode extends ClassElementNode {


    // static variables ..........................................................

    /** generated Serialized Version UID */
    static final long                   serialVersionUID = -5345233212410231L;

    private static ResourceBundle       bundle;
    
    // variables .................................................................
    
    public JavaClass                    clazz;
    
    private boolean                     prepared;
    public Hashtable                    innerClasses = null;
    private ClassElement                element = null;

    // init ......................................................................

    public ClassNode (JavaClass clazz, ClassElement element, LoadedClassFactory factory) {
        super (element, factory.createChildren (clazz, element), false);
        this.element = element;
        prepared = clazz.isPrepared ();
        String name = clazz.getName ();
        name = name.substring (name.lastIndexOf ('.') + 1);
        String s = name.substring (name.lastIndexOf ('$') + 1);
        if ((s.length () > 0) && (Character.isJavaIdentifierStart (s.charAt (0))))
            name = s;
        setDisplayName (name);
        this.clazz = clazz;
    }
    
    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerClassNode");        //NOI18N
    }

    protected Sheet createSheet () {
        Sheet sheet = Sheet.createDefault();
        Sheet.Set ps = sheet.get(Sheet.PROPERTIES);
        ps.put (new PropertySupport.ReadOnly (
                    "classLoaderTypeName", // NOI18N
                    String.class,
                    getLocalizedString ("PROP_class_loader_type"),
                    getLocalizedString ("HINT_class_loader_type")
                ) {
                    public Object getValue () {
                        String loaderType = clazz.classLoaderTypeName ();
                        if (loaderType == null)
                            return getLocalizedString ("CTL_bootstrap_class_loader");
                        return loaderType;
                    }
                });
        ps.put(createModifiersProperty(false));
        ps.put(createNameProperty(false));
        ps.put(createSuperclassProperty(false));
        ps.put(createInterfacesProperty(false));
        return sheet;
    }
    
    public boolean hasCustomizer () {
        return false;
    }

    public void update () {
        if (!prepared && clazz.isPrepared ()) {
            prepared = true;
            clazz.updateClassElement (element);
        } 
        else
            clazz.updateInnerClasses (element);
        Node [] children = (Node []) Children.MUTEX.writeAccess (new Mutex.Action () {
            public Object run () {
                return getChildren ().getNodes ();
            }
        });
        for (int x = 0; x < children.length; x++)
            if (children [x] instanceof ClassNode)
                ((ClassNode) children [x]).update ();
    }
    
    // helper methods ............................................................

    /*
    public void updateClassElement (ClassElement element) {
        
        boolean debug = clazz.getClassName ().equals ("Collections");
        if (debug) {
            System.out.println ("updateClassElement");
        }
        
        try {
            innerClasses = new Hashtable ();
            JavaClass [] inner = clazz.getInnerClasses ();
            ClassElement [] classes = new ClassElement [inner.length];
            for (int x = 0; x < inner.length; x++) {
                if ((classes [x] = inner [x].toClassElement ()) == null) {
                    
                    System.out.println ("JavaClass.toClassElement (), NULL");
                    
                    classes [x] = new ClassElement ();
                }
                innerClasses.put (inner [x].getClassName (), inner [x]);
            }
            
            element.setClasses (classes);
            if (clazz.isPrepared ()) {
                element.setConstructors (clazz.getConstructors ());
                element.setMethods (clazz.getMethods ());
                element.setInitializers (clazz.getInitializers ());
                element.setFields (clazz.getFields ());
            }
            element.setModifiers (clazz.modifiers ());
        } catch (SourceException e) {
            System.out.println (e);
            e.printStackTrace ();
        }
    }
    
    public void updateInnerClasses (ClassElement element) {
        
        boolean debug = clazz.getClassName ().equals ("Collections");
        if (debug) {
            System.out.println ("updateInnerClasses");
        }
        
        try {
            Hashtable temp = new Hashtable ();
            ClassElement [] last = element.getClasses ();
            JavaClass [] inner = clazz.getInnerClasses ();
            
            if (debug) System.out.println ("new inner:");
            
            for (int x = 0; x < inner.length; x++) {
                ClassElement el = inner [x].toClassElement ();
                if (innerClasses.get (inner [x].getClassName ()) == null) {
                    innerClasses.put (inner [x].getClassName (), inner [x]);
                    
                    element.addClass (el);
                }
                temp.put (inner [x].getClassName (), inner [x]);
                
                if (debug) System.out.println (el.getName ().getName ());
                
            }
            
            if (debug) System.out.println ("last:");
            
            for (int x = 0; x < last.length; x++) {
                
                if (debug) System.out.println (last [x].getName ().getName ());
                
                if (temp.get (last [x].getName ().getName ()) == null)
                    element.removeClass (last [x]);
            }
            innerClasses = temp;
        } catch (SourceException e) {
            System.out.println (e);
            e.printStackTrace ();
        }
    }
 
    public JavaClass findInnerClass (ClassElement el) {
        System.out.println ("find inner class: " + el.getName ().getName ());
        boolean debug = clazz.getClassName ().equals ("Collections");
        if (debug) {
            System.out.println ("findInnerClass");
        }
        return (JavaClass) innerClasses.get (el.getName ().getName ());
    }
    
     */
    
    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (ClassNode.class);
        return bundle.getString (s);
    }
    
}
