/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is Forte for Java, Community Edition. The Initial
 * Developer of the Original Code is Sun Microsystems, Inc. Portions
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.modules.debugger.support.java.nodes;

import java.util.*;
import java.beans.*;

import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.debugger.*;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.src.ClassElement;
import org.openide.src.nodes.ClassElementNode;
import org.openide.src.nodes.ClassChildren;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.util.RequestProcessor;


/**
* This class represents all loaded classes.
* This class is final only for performance reasons.
* Can be unfinaled if desired.
*
* @author   Daniel Prusa
*/
public final class ClassesRootNode extends PackageNode implements Validator.Object {


    // static variables ..........................................................

    /** generated Serialized Version UID */
    static final long                  serialVersionUID = -5432332119936122231L;

    private static String              ICON_BASE =
        "/org/netbeans/modules/debugger/resources/loadedClasses"; // NOI18N

    private static ResourceBundle      bundle;
    
    private static PackageContent      content = new PackageContent (""); //NOI18N

    // variables .................................................................
    
    private Debugger                   debugger;
    
    // init ......................................................................

    /**
    * Creates empty VariablesContext.
    */
    public ClassesRootNode () {
        super (null);
    }

    protected void init () {
        try {
            debugger = TopManager.getDefault ().getDebugger ();
            Validator validator = ((ValidatorHolder) debugger).getValidator ();
            validator.add (this);
            debugger.addPropertyChangeListener (new StateListener ());
        } catch (DebuggerException e) {
        }
        String s = getLocalizedString ("CTL_Classes_root");
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
        changeChildren ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerClassesRootNode");        //NOI18N
    }


    // helper methods ............................................................

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new ClassesHandle ();
    }
    
    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (ClassesRootNode.class);
        return bundle.getString (s);
    }

    void changeChildren () {
        List list = null;
        try {
            Debugger debugger = ((CoreDebugger) TopManager.getDefault ().getDebugger ()).
                getCurrentDebugger ();
            if (debugger == null) 
                list = new ArrayList ();
            else {
                if (debugger instanceof LoadedClassesProducer)
                    list = ((LoadedClassesProducer) debugger).getLoadedClasses ();
                else
                    list = new ArrayList ();
            }
            Iterator iter = list.iterator ();
            while (iter.hasNext())
                content.add ((JavaClass) iter.next ());
            content.reorganize ();
            ((PackageChildren)getChildren ()).setContent (content);
        } catch (DebuggerNotFoundException e) {
        }
    }
    
    // interface Validator.Object ................................................

    public void validate () {
        int state = debugger.getState ();
        if ((state == Debugger.DEBUGGER_STOPPED) || (state == Debugger.DEBUGGER_NOT_RUNNING)) {
            try {
                changeChildren ();
            } catch (Exception e) {
                e.printStackTrace ();
            }
        }
    }

    public boolean canRemove () {
        return false;
    }
    
    /**
    * Serializable node reference. 
    */
    private static class ClassesHandle implements Node.Handle {
        static final long serialVersionUID =-3418262400286559650L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return DebuggerModule.getClassesRootNode ();
        }
    }
 
    private class StateListener implements PropertyChangeListener {
                
        private RefreshThread thread = null;
        
        public void propertyChange (PropertyChangeEvent e) {
            if (!e.getPropertyName ().equals (Debugger.PROP_STATE)) return;
            switch (debugger.getState ()) {
                case Debugger.DEBUGGER_NOT_RUNNING:
                    synchronized (this) {
                        if (thread != null) {
                            thread.finish ();
                            thread = null;
                        }
                    }
                    break;   
                /* case Debugger.DEBUGGER_STARTING:
                    break; */
                case Debugger.DEBUGGER_RUNNING:
                    synchronized (this) {
                        if (thread != null)
                            thread.finish ();
                        thread = new RefreshThread ("Debugger Classes Refresher");
                        thread.start ();
                    }
                    break;   
                /* case Debugger.DEBUGGER_STOPPED:
                    break; */
            }
        }
    }
    
    private class RefreshThread extends Thread {
        
        private boolean finished = false;

        public RefreshThread (String name) {
            super (name);
        }
        
        public void finish () {
            finished = true;
        }
        
        public void run () {
            while (!finished) {
                try {
                    sleep (6000);
                } catch (InterruptedException e) {
                    finished = true;
                }
                RequestProcessor.postRequest (new Runnable () {
                    public void run () {
                        if ((debugger.getState () != Debugger.DEBUGGER_NOT_RUNNING) &&
                            (debugger.getState () != Debugger.DEBUGGER_STOPPED)) {
                            try {
                                changeChildren ();
                            } catch (Exception e) {
                                e.printStackTrace ();
                            }
                        }
                    }
                });
            } // while
        } // run ()
        
    }
}
