/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java.nodes;

import org.openide.TopManager;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.DebuggerException;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;

import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.JavaThreadGroup;
import org.netbeans.modules.debugger.support.nodes.ThreadGroupNode;
import org.netbeans.modules.debugger.support.actions.SuspendCookie;
import org.netbeans.modules.debugger.support.actions.SuspendAction;
import org.netbeans.modules.debugger.support.actions.ResumeAction;

/**
* This class represents threadGroup as a Node.
* This class is final only for performance reasons,
* can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class JavaThreadGroupNode extends ThreadGroupNode implements SuspendCookie {
    /** generated Serialized Version UID */
    static final long                   serialVersionUID = -4731990787523747464L;

    /** System actions of this node */
    private static SystemAction[]       staticActions;
    
    private JavaThreadGroup threadGroup;
    
    // init ......................................................................

    /**
    * Creates thread group context with given thread group
    */
    public JavaThreadGroupNode (JavaThreadGroup threadGroup) {
        super (threadGroup);
        this.threadGroup = threadGroup;
        getCookieSet ().add (this);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerThreadGroupNode");        //NOI18N
    }

    
    // Node implementation .......................................................

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                SystemAction.get (SuspendAction.class),
                SystemAction.get (ResumeAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class)
            };
        return staticActions;
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new ThreadsHandle ();
    }

    // SuspendCookie .............................................................

    /**
    * Returns state of thread.
    */
    public boolean isSuspended () {
        // throw new InternalError ();
        return false;
    }

    /**
    * Sets state of thread.
    */
    public void setSuspended (boolean suspended) {
        threadGroup.setSuspended (suspended);
    }

    /**
    * Serializable node reference. 
    */
    private static class ThreadsHandle implements Node.Handle {
        static final long serialVersionUID = -4518262935887212235L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            Node node = DebuggerModule.getNode (JavaThreadGroupNode.class);
            if (node == null)
                return new AbstractNode (Children.LEAF);
            return node;
        }
    }
}