/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.lang.ref.WeakReference;

import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.Debugger;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.text.Line;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallableSystemAction;

import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.JavaThread;
import org.netbeans.modules.debugger.support.nodes.CallStackNode;
import org.netbeans.modules.debugger.support.nodes.LocalsRootNode;
import org.netbeans.modules.debugger.support.actions.*;

import org.netbeans.modules.debugger.support.java.MonitorInfoProducer;

/**
* This class representates threads and threadGroups as a Node.
* This class is final only for performance reasons, can be happily
* unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class JavaThreadNode extends AbstractNode 
    implements SuspendCookie, SwitchOnCookie, GoToSourceCookie {

    /** generated Serialized Version UID */
    static final long serialVersionUID = 2202990736781010124L;


    // static ....................................................................

    /** Property names */
    public static final String        PROP_THREAD_NAME         = "threadName"; // NOI18N
    public static final String        PROP_THREAD_STATE        = "threadState"; // NOI18N
    public static final String        PROP_THREAD_CLASS        = "threadClass"; // NOI18N
    public static final String        PROP_THREAD_METHOD       = "threadMethod"; // NOI18N
    public static final String        PROP_THREAD_LINE         = "threadLine"; // NOI18N
    public static final String        PROP_THREAD_STACK_DEPTH  = "threadStackDepth"; // NOI18N
    public static final String        PROP_THREAD_FRAME_INDEX  = "threadFrameIndex"; // NOI18N
    public static final String        PROP_THREAD_SUSPENDED    = "threadSuspended"; // NOI18N

    public static final int           STATE_MAIN               = 1;
    public static final int           STATE_STOPPED            = 2;

    protected static String           ICON_CURRENT             =
        "/org/netbeans/modules/debugger/resources/threadCurrent"; // NOI18N
    protected static String           ICON_SUSPENDED           =
        "/org/netbeans/modules/debugger/resources/threadSuspended"; // NOI18N
    protected static String           ICON_RUNNING             =
        "/org/netbeans/modules/debugger/resources/threadRunning"; // NOI18N

    /** Icons for threads. */
    protected static Image            threadFolderIcon;
    protected static Image            threadFolderIcon32;
    protected static Image            threadCurrentIcon;
    protected static Image            threadCurrentIcon32;
    protected static Image            threadRunningIcon;
    protected static Image            threadRunningIcon32;

    /** System actions of this node */
    private static SystemAction[]     staticActions;
    private static ResourceBundle     bundle;


    // variables .................................................................

    private Debugger                  debugger;
    private JavaThread                thread;
    private ThreadListener            threadListener;
    private Node[]                    children = null;
    private CallStackNode             callStack;
    private LocalsRootNode            locals;
    private ContendedMonitorNode      contendedMonitor = null;
    private OwnedMonitorsNode         ownedMonitors = null;
    private boolean                   cmnPresent = false;
    private boolean                   omnPresent = false;
    private int                       state = 4;
    private boolean                   oldSuspended = false;
    private String                    dispNamePrefix;

    // init ......................................................................

    /**
    * Creates ThreadNode.
    */
    public JavaThreadNode (JavaThread thread) {
        this (thread, null);
    }

    /**
    * Creates ThreadNode having in its display name given prefix.
    */
    JavaThreadNode (JavaThread thread, String dispNamePrefix) {
        super (new Children.Array());
        this.thread = thread;
        this.dispNamePrefix = dispNamePrefix;
        initialize();
    }
    
    /**
    * Remove listeners.
    */
    protected void finalize () {
        debugger.removePropertyChangeListener (threadListener);
        thread.removePropertyChangeListener (threadListener);
    }

    /** Performs basic initialization of children, listeners, etc. */
    private void initialize () {
        initializeChildren ();
        try {
            debugger = TopManager.getDefault ().getDebugger ();
        } catch (DebuggerException e) {
        }
        setName ();
        createProperties ();
        getCookieSet ().add (this);
        threadListener = new ThreadListener (this);
        thread.addPropertyChangeListener (threadListener);
    }

    /** Initializes children (subnodes) of this node */
    private void initializeChildren () {
        Children myChildren = getChildren ();
        myChildren.add (
            new Node[] {
                callStack = new CallStackNode (thread),
                locals = new LocalsRootNode (thread)
            }
        );
        if (thread instanceof MonitorInfoProducer) {
            contendedMonitor = new ContendedMonitorNode (thread);
            ownedMonitors = new OwnedMonitorsNode (thread);
            if (((MonitorInfoProducer) thread).contendsMonitor ()) {
                cmnPresent = true;
                myChildren.add (new Node[] {contendedMonitor});
            }
            if (((MonitorInfoProducer) thread).ownsMonitors ()) {
                myChildren.add (new Node[] {ownedMonitors});
                omnPresent = true;
            }
        }
        /*
        if (thread instanceof MonitorInfoProducer) {
            contendedMonitor = new ContendedMonitorNode (variableHome, thread);
            ownedMonitors = new OwnedMonitorsNode (variableHome, thread);
            if (((MonitorInfoProducer) thread).contendsMonitor ())
                myChildren.add (new Node[] {contendedMonitor});
            if (((MonitorInfoProducer) thread).ownsMonitors ())
                myChildren.add (new Node[] {ownedMonitors});
        }
         */
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerThreadNode");        //NOI18N
    }


    // Node implementation .......................................................

    /** Method that prepares properties. Called from initialize.
    */
    private void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault();
        Sheet.Set props = sheet.get(Sheet.PROPERTIES);
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_NAME,
                       String.class,
                       getLocalizedString ("PROP_thread_name"),
                       getLocalizedString ("HINT_thread_name")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return thread.getName ();
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_STATE,
                       String.class,
                       getLocalizedString ("PROP_thread_state"),
                       getLocalizedString ("HINT_thread_state")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return thread.getState ();
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_CLASS,
                       String.class,
                       getLocalizedString ("PROP_thread_class"),
                       getLocalizedString ("HINT_thread_class")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return thread.getClassName ();
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_METHOD,
                       String.class,
                       getLocalizedString ("PROP_thread_method"),
                       getLocalizedString ("HINT_thread_method")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return thread.getMethod ();
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_LINE,
                       Integer.TYPE,
                       getLocalizedString ("PROP_thread_line"),
                       getLocalizedString ("HINT_thread_line")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return new Integer (thread.getLineNumber ());
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadOnly (
                       PROP_THREAD_STACK_DEPTH,
                       Integer.TYPE,
                       getLocalizedString ("PROP_thread_stack_depth"),
                       getLocalizedString ("HINT_thread_stack_depth")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return new Integer (thread.getStackDepth ());
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                   });
        props.put (new PropertySupport.ReadWrite (
                       PROP_THREAD_SUSPENDED,
                       Boolean.TYPE,
                       getLocalizedString ("PROP_thread_suspended"),
                       getLocalizedString ("HINT_thread_suspended")
                   ) {
                       public Object getValue () throws InvocationTargetException {
                           try {
                               return new Boolean (thread.isSuspended ());
                           } catch (java.lang.Exception e) {
                               throw new InvocationTargetException (e);
                           }
                       }
                       public void setValue (Object val) throws
                           InvocationTargetException {
                           if (!(val instanceof Boolean))
                               throw new IllegalArgumentException ();
                           setSuspended (((Boolean)val).booleanValue ());
                       }
                   });
        // and set new sheet
        setSheet (sheet);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                                SystemAction.get (SwitchOnAction.class),
                                SystemAction.get (GoToSourceAction.class),
                                SystemAction.get (SuspendResumeAction.class),
                                null,
                                SystemAction.get (ToolsAction.class),
                                SystemAction.get (PropertiesAction.class)
                            };
        return staticActions;
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (SwitchOnAction.class);
    }

    public void destroy () throws java.io.IOException {
        super.destroy ();
        debugger.removePropertyChangeListener (threadListener);
        thread.removePropertyChangeListener (threadListener);
    }


    // SwitchOnCookie ........................................................................................

    /**
    * Sets this thread to be suspended.
    */
    public boolean canSetCurrent () {
        return (!thread.isCurrent ()) && isSuspended ();
    }

    /**
    * Sets this thread to be current.
    */
    public void setCurrent () {
        thread.setCurrent (true);
    }


    // SuspendCookie .............................................................

    /**
    * Sets this thread to be suspended.
    */
    public void setSuspended (boolean b) {
        try {
            thread.setSuspended (b);
            changeChildren ();
        } catch (DebuggerException e) {
            TopManager.getDefault ().notifyException (e.getTargetException ());
        }
    }

    /**
    * Sets this thread to be suspended.
    */
    public boolean isSuspended () {
        return thread.isSuspended ();
    }


    // GoToSourceCookie implementation ...........................................

    /**
    * Goes to source.
    */
    public void goToSource () {
        Line l = thread.getLine ();
        if (l != null) l.show (Line.SHOW_GOTO);
    }

    /**
    * Returns true if source is availabled.
    */
    public boolean canGoToSource () {
        Line l = thread.getLine ();
        return l != null;
    }


    // helper methods ............................................................

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JavaThreadNode.class);
        return bundle.getString (s);
    }

    /**
    * Sets curent icon (running thread, current thread, suspended thread).
    */
    void refreshIcon () {
        if ((state & STATE_MAIN) != 0) {
            setIconBase (ICON_CURRENT); // NOI18N
        } else
            if ((state & STATE_STOPPED) != 0) {
                setIconBase (ICON_SUSPENDED); // NOI18N
            } else {
                setIconBase (ICON_RUNNING); // NOI18N
            }
    }

    /** equals if thread equals o.thread
    * @param o
    */
    public final boolean equals (Object o) {
        return o instanceof JavaThreadNode && ((JavaThreadNode) o).thread.equals (thread);
    }

    /**
    * @return hashCode of thread
    */
    public final int hashCode () {
        return thread.hashCode ();
    }

    /**
    * Updates name & icon, and fires change.
    */
    void changeProperties () {
        setName ();
        firePropertyChange (null, null, null);
    }

    /**
    * Updates name & icon.
    */
    void setName () {
        // read thread state
        int newS = 0;
        if (thread.isCurrent ())
            newS = STATE_MAIN;
        try {
            if (thread.isSuspended ()) newS |= STATE_STOPPED;
        } catch (java.lang.Exception e) {
        }

        // change icon
        if (state != newS) {
            state = newS;
            refreshIcon ();
        }

        // update name
        String s = thread.getName () + " (" + thread.getState () + ")"; // NOI18N
        if (dispNamePrefix != null)
            s = dispNamePrefix.concat (" ").concat (s); // NOI18N
        setDisplayName (s);
        setName (s);
    }

    /**
    * Refresh locales, callstack & monitors.
    */
    void changeChildren () {
        Children myChildren = getChildren ();
        if (thread instanceof MonitorInfoProducer) {
            if (((MonitorInfoProducer) thread).contendsMonitor ()) {
                contendedMonitor.update ();
                if (!cmnPresent)
                    myChildren.add (new Node[] {contendedMonitor});
                cmnPresent = true;
            } else {
                if (cmnPresent)
                    myChildren.remove (new Node[] {contendedMonitor});
                cmnPresent = false;
            }
            if (((MonitorInfoProducer) thread).ownsMonitors ()) {
                ownedMonitors.update ();
                if (!omnPresent)
                    myChildren.add (new Node[] {ownedMonitors});
                omnPresent = true;
            } else {
                if (omnPresent)
                    myChildren.remove (new Node[] {ownedMonitors});
                omnPresent = false;
            }
        }
    }

    JavaThread getDebuggerThread () {
        return thread;
    }


    // innerclasses ..............................................................

    /**
    * Listens on thread and debugger.
    */
    private static class ThreadListener implements PropertyChangeListener {

        private WeakReference         node;
        private Debugger              debugger;
        private JavaThread            thread;

        ThreadListener (JavaThreadNode n) {
            node = new WeakReference (n);
            debugger = n.debugger;
            thread = n.thread;
        }

        public void propertyChange (PropertyChangeEvent e) {
            JavaThreadNode tn = (JavaThreadNode) node.get ();
            if (tn == null) {
                debugger.removePropertyChangeListener (this);
                thread.removePropertyChangeListener (this);
                return;
            }
            tn.changeProperties ();
            String name = e.getPropertyName ();
            if ((name != null) && name.equals (MonitorInfoProducer.PROP_MONITORS))
                tn.changeChildren ();
        }
    }
}

