/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.IOException;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.Arrays;

import org.openide.actions.DeleteAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.Debugger;
import org.openide.debugger.Watch;
import org.openide.debugger.DebuggerException;
import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.nodes.BeanNode;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.nodes.Sheet;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.support.DebuggerModule;
import org.netbeans.modules.debugger.AbstractVariable;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.nodes.*;
import org.netbeans.modules.debugger.support.java.JavaVariable;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* This class represents variable as a Node.
*
* @author   Jan Jancura
*/
public class JavaVariableNode extends AbstractNode implements GetVariableCookie, CreateVariableCookie {

    /** generated Serialized Version UID */
    static final long                       serialVersionUID = -4167730413290269518L;

    protected static String                 ICON_BASE =
        "/org/netbeans/core/resources/variable"; // NOI18N
    protected static String                 ICON_STATIC_VARIABLE =
        "/org/netbeans/modules/debugger/resources/staticVariable"; // NOI18N

    public static final String              PROP_MODIFIERS = "modifiers"; // NOI18N
    public static final String              PROP_INNER_TYPE = "innerType"; // NOI18N
    public static final String              PROP_BASE_INDEX = "baseIndex"; // NOI18N
    public static final String              PROP_DISPLAYED_LENGTH = "displayedLength"; // NOI18N
    private static ResourceBundle           bundle;


    // variables .................................................................

    private boolean                         isRoot = false;

    protected JavaVariable                  variable = null;

    protected transient Debugger            debugger;

    private PropertyChangeListener          pcl;

    /** First index of an array to be displayed. */
    private int                             baseIndex = 0;

    /** Number of field of an array to be displayed. */
    private int                             displayedLength = 100;

    private boolean                         isArray;
    
    // init ......................................................................

    /**
    * Creates empty Variable context.
    */
    public JavaVariableNode (JavaVariable variable, boolean isRoot) {
        super (
            variable.isLeaf () ?
            Children.LEAF :
            new VariableContextChildren ()
        );
        this.isRoot = isRoot;
        this.variable = variable;
        isArray = variable.isArray ();
        init ();
    }

    /**
    * Creates empty Variable context.
    */
    public JavaVariableNode (JavaVariable variable) {
        this (variable, false);
    }

    protected void init () {
        variable.addPropertyChangeListener (pcl = new PropertyChangeListener () {
                                                public void propertyChange (PropertyChangeEvent e) {
                                                    parameterChanged (e);
                                                }
                                            });
        setDisplayName (getName ());
        setName (getName ());
        if ( (variable.getModifiers () != null) &&
             (variable.getModifiers ().indexOf ("static") > -1)
        )   setIconBase (ICON_STATIC_VARIABLE);
        else
            setIconBase (ICON_BASE);
        // obtain debugger reference
        try {
            debugger = TopManager.getDefault ().getDebugger ();
        } catch (DebuggerException exc) {
            TopManager.getDefault ().notifyException (exc);
        }
        createProperties ();
        getCookieSet ().add (this);
        changeChildren ();
    }

    /** deserializes object */
    private void readObject (java.io.ObjectInputStream obis)
    throws java.io.IOException, ClassNotFoundException,
        java.io.NotActiveException {
        obis.defaultReadObject ();
        init ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerVariableNode");        //NOI18N
    }


    // implementation of Node ....................................................

    /** Creates properties for this node */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        ps.put (new PropertySupport.ReadOnly (
                    Watch.PROP_VARIABLE_NAME,
                    String.class,
                    getLocalizedString ("PROP_watch_name"),
                    getLocalizedString ("HINT_watch_name")
                ) {
                    public Object getValue () {
                        return variable.getVariableName ();
                    }
                });
        createCommonProperties (bundle, ps);
        // and set new sheet
        setSheet (sheet);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (variable.isFixedVariable ())
            return new SystemAction[] {
                       //        CreateVariableModificationBreakpointAction.get (CreateVariableModificationBreakpointAction.class),
                       //        CreateVariableAccessBreakpointAction.get (CreateVariableAccessBreakpointAction.class),
                       //        null,
                       SystemAction.get (DeleteAction.class),
                       null,
                       SystemAction.get (ToolsAction.class),
                       SystemAction.get (PropertiesAction.class),
                   };
        else
            // if (variableHome != null)
                return new SystemAction[] {
                           SystemAction.get (CreateVariableAction.class),
                           null,
                           SystemAction.get (ToolsAction.class),
                           SystemAction.get (PropertiesAction.class),
                       };
            /*
            else
                return new SystemAction[] {
                           SystemAction.get (ToolsAction.class),
                           SystemAction.get (PropertiesAction.class),
                       };
             */
    }

    /** Helper method, creates properties common to variable
    * and watch contexts. */
    void createCommonProperties (
        final ResourceBundle bundle,
        final Sheet.Set ps
    ) {
        ps.put (new PropertySupport.ReadOnly (
                    Watch.PROP_TYPE,
                    String.class,
                    getLocalizedString ("PROP_watch_type"),
                    getLocalizedString ("HINT_watch_type")
                ) {
                    public Object getValue () {
                        return variable.getType ();
                    }
                });
        ps.put (new PropertySupport.ReadOnly (
                    PROP_INNER_TYPE,
                    String.class,
                    getLocalizedString ("PROP_watch_inner_type"),
                    getLocalizedString ("HINT_watch_inner_type")
                ) {
                    public Object getValue () {
                        return variable.getInnerType ();
                    }
                });
        ps.put (new PropertySupport.ReadOnly (
                    PROP_MODIFIERS,
                    String.class,
                    getLocalizedString ("PROP_watch_modifiers"),
                    getLocalizedString ("HINT_watch_modifiers")
                ) {
                    public Object getValue () {
                        return variable.getModifiers ();
                    }
                });
        ps.put (new PropertySupport.ReadWrite (
                    Watch.PROP_AS_TEXT,
                    String.class,
                    getLocalizedString ("PROP_watch_value"),
                    getLocalizedString ("HINT_watch_value")
                ) {
                    public Object getValue () {
                        return variable.getAsText ();
                    }
                    public void setValue (Object val) throws IllegalArgumentException {
                        try {
                            variable.setAsText ((String)val);
                        } catch (ClassCastException e) {
                            throw new IllegalArgumentException ();
                        }
                        catch (DebuggerException e) {}
                    }
                });
        if (isArray) {
            ps.put (new PropertySupport.ReadWrite (
                        PROP_BASE_INDEX,
                        Integer.TYPE,
                        getLocalizedString ("PROP_base_index"),
                        getLocalizedString ("HINT_base_index")
                    ) {
                        public Object getValue () {
                            return new Integer (baseIndex);
                        }
                        public void setValue (Object val) throws IllegalArgumentException {
                            try {
                                int x = ((Integer)val).intValue ();
                                if (x < 0) throw new ClassCastException ();
                                baseIndex = x;
                                if (isArray)
                                    changeChildren ();
                                // firePropertyChange (PROP_BASE_INDEX, null, null);
                            }
                            catch (ClassCastException e) {
                                new IllegalArgumentException ();
                            }
                        }
                    });
            ps.put (new PropertySupport.ReadWrite (
                        PROP_DISPLAYED_LENGTH,
                        Integer.TYPE,
                        getLocalizedString ("PROP_displayed_length"),
                        getLocalizedString ("HINT_displayed_length")
                    ) {
                        public Object getValue () {
                            return new Integer (displayedLength);
                        }
                        public void setValue (Object val) throws IllegalArgumentException {
                            try {
                                int x = ((Integer)val).intValue ();
                                if (x < 0) throw new ClassCastException ();
                                displayedLength = x;
                                if (isArray)
                                    changeChildren ();
                                // firePropertyChange (PROP_DISPLAYED_LENGTH, null, null);
                            }
                            catch (ClassCastException e) {
                                new IllegalArgumentException ();
                            }
                        }
                    });
        }
    }

    /**
    * Variable can be removed.
    *
    * @return <CODE>true</CODE>
    */
    public boolean canDestroy () {
        return variable.isFixedVariable ();
    }

    /**
    * Removes the Variable from its parent and deletes it.
    */
    public void destroy () throws IOException {
        if (variable == null) {
            super.destroy ();
            return;
        }
        variable.removePropertyChangeListener (pcl);
        variable = null;
        if (getChildren () == Children.LEAF) {
            super.destroy ();
            return;
        }
        VariableContextChildren myChildren = (VariableContextChildren)
                                             getChildren ();
        if (!myChildren.initialized) {
            super.destroy ();
            return;
        }
        Node[] n = myChildren.getNodes ();
        int i, k = n.length;
        for (i = 0; i < k; i++)
            ((JavaVariableNode) n [i]).destroy ();
        super.destroy ();
    }


    // GetVariableCookie implementation ..........................................

    /**
    * Returns variable instance.
    */
    public AbstractVariable getVariable () {
        return variable;
    }


    // other methods .............................................................

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JavaVariableNode.class);
        return bundle.getString (s);
    }

    /**
    * Returns human presentable name of this variable containing
    * informations about value or valide.
    *
    * @return human presentable name of this variable.
    */
    public String getName () {
        String name = variable.getVariableName ();
        if ((name == null) || (debugger == null)) return "???"; // NOI18N

        if (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING)
            return name + " " + getLocalizedString ("CTL_NoSession");
        String type = variable.getType ();
        String innerType = variable.getInnerType ();
        String value = variable.getAsText ();
        if (type == null)
            if (value == null)
                if (variable.isObject ()) return name + " = null"; // NOI18N
                else return name + " " + getLocalizedString ("CTL_NotInitialized");
            else return name + " = >" + value + "<"; // NOI18N
        if (variable.isObject ()) {
            if (type.equals (innerType))
                return name + " = (" + innerType + ") " + value; // NOI18N
            else
                return name + " = (" + type + ") (" + innerType + ") " + value; // NOI18N
        }
        else
            return name + " = (" + type + ") " + value; // NOI18N
    }

    // CreateVariableCookie implementation ............................................
    
    public void createVariable () {
        WatchesRootNode n = (WatchesRootNode) DebuggerModule.getNode (WatchesRootNode.class);
        if (n != null)
            n.createVariable (variable.createFixedVariable ());
    }
    
    public boolean canCreateVariable () {
        return true;
    }

    /**
    * Sets display name and fires property change.
    */
    void parameterChanged (PropertyChangeEvent e) {
        if (variable.isArray () != isArray) {
            isArray = !isArray;
            createProperties ();
        }
        String s = getName ();
        setDisplayName (s);
        setName (s);
        changeChildren ();
        firePropertyChange (
            e.getPropertyName (),
            e.getOldValue (),
            e.getNewValue ()
        );
    }

    /**
    * Sets current variable fields.
    */
    void changeChildren () {
        // set fields of this variable as keys into our children
        if (getChildren() == Children.LEAF) {
            Node n = getParentNode ();
            if (n == null)
                return;
            JavaVariable v = (JavaVariable) getVariable ();
            if (!v.isLeaf ()) {
                LeafRefresher parentChildren =
                    (LeafRefresher) n.getChildren ();
                parentChildren.refreshMyKey (v);
            }
            return;
        }
        VariableContextChildren myChildren =
            (VariableContextChildren) getChildren ();
        if (!myChildren.initialized) return;
        
        AbstractVariable [] fields;
        if (isArray) {
            // we do not need to sort arrays
            int length = Math.min (displayedLength, variable.getVariablesNumber() - baseIndex);
            fields = variable.getVariables (baseIndex, length);
        }
        else
        {
            fields = variable.getFilteredVariables ();
            Arrays.sort (fields, Utils.variablesComparator);
        }
        
        myChildren.setMyKeys (fields);
    }

    // innerclasses ..............................................................

    /** Special subnodes (children) for variable node */
    private static final class VariableContextChildren extends Children.Keys implements LeafRefresher {

        private boolean                         initialized = false;

        /** Initializes children.
        */
        protected void addNotify () {
            initialized = true;
            ((JavaVariableNode) getNode ()).changeChildren ();
        }

        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (Object key) {
            return new Node[] {
                       new JavaVariableNode ((JavaVariable)key)
                   };
        }

        /** Accessor for VariableNode outer class */
        private void setMyKeys (Object[] keys) {
            setKeys(keys);
        }

        // interface LeafRefresher

        /** Accessor for VariableNode class */
        public void refreshMyKey (Object key) {
            refreshKey (key);
        }

    } // end of VariableContextChildren inner class
}
