/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.IOException;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import javax.swing.JPopupMenu;

import org.openide.actions.DeleteAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.nodes.BeanNode;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.support.DebuggerModule;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.nodes.*;
import org.netbeans.modules.debugger.AbstractWatch;
import org.netbeans.modules.debugger.support.java.*;


/**
* This class represents watch as a Node.
* This class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class JavaWatchNode extends JavaVariableNode {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 7100312323961302436L;

    protected static String         ICON_BASE =
        "/org/netbeans/core/resources/watch"; // NOI18N
    private static ResourceBundle   bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JavaWatchNode.class);
        return bundle.getString (s);
    }

    
    // init ......................................................................

    /**
    * Creates empty Watch context.
    */
    public JavaWatchNode (final JavaWatch watch) {
        super (watch, true);
        setIconBase (ICON_BASE);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerWatchNode");        //NOI18N
    }


    // implementation of Node ....................................................

    /**
    * Variable can be removed.
    *
    * @return <CODE>true</CODE>
    */
    public boolean canDestroy () {
        return true;
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        // if (variableHome != null)
            return new SystemAction[] {
                SystemAction.get (CreateVariableAction.class),
                null,
                //      CreateVariableModificationBreakpointAction.get (CreateVariableModificationBreakpointAction.class),
                //      CreateVariableAccessBreakpointAction.get (CreateVariableAccessBreakpointAction.class),
                //      null,
                SystemAction.get (DeleteAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class),
            };
        /*
        else
            return new SystemAction[] {
                       //      CreateVariableModificationBreakpointAction.get (CreateVariableModificationBreakpointAction.class),
                       //      CreateVariableAccessBreakpointAction.get (CreateVariableAccessBreakpointAction.class),
                       //      null,
                       SystemAction.get (DeleteAction.class),
                       null,
                       SystemAction.get (ToolsAction.class),
                       SystemAction.get (PropertiesAction.class),
                   };
         */
    }

    /**
    * Removes the Variable from its parent and deletes it.
    */
    public void destroy () throws IOException {
        if (variable == null) return;
        getWatch ().remove ();
        super.destroy ();
    }

    /** Creates properties for watch context
    * (remember that watch extends variable context,
    * so we must pay an attention to its properties too)
    * Overrides createProperties from Variable context.
    */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        // and add rw property of the same name
        ps.put(new PropertySupport.ReadWrite (
            AbstractWatch.PROP_VARIABLE_NAME,
            String.class,
            getLocalizedString ("PROP_watch_name"),
            getLocalizedString ("HINT_watch_name")
        ) {
            public Object getValue () {
                return getWatch ().getVariableName ();
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    getWatch ().setVariableName (((String)val).trim ());
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
            }
        });
        // add type and value props
        createCommonProperties (bundle, ps);
        // and scope info, finally
        ps.put(new PropertySupport.ReadOnly (
            JavaWatch.PROP_IN_SCOPE,
            Boolean.TYPE,
            getLocalizedString ("PROP_watch_in_scope"),
            getLocalizedString ("HINT_watch_in_scope")
        ) {
            public Object getValue () {
                return new Boolean (getWatch ().isInScope ());
            }
        });
        // and set new sheet
        setSheet(sheet);
    }


    // private methods ...........................................................

    /**
    * Returns watch.
    */
    JavaWatch getWatch () {
        return (JavaWatch) variable;
    }

    /**
    * Returns human presentable name of this watch containing
    * informations about value or valide.
    *
    * @return human presentable name of this variable.
    */
    public String getName () {
        if (getWatch () == null) return "?"; // MOI18N // should not occure
        String name = getWatch ().getVariableName ();
        if (name == null) return "???"; // NOI18N

        if ( (debugger == null) ||
                (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING)
           ) return name + " = " + getLocalizedString ("CTL_NoSession");

        String errorMessage = getWatch ().getErrorMessage ();
        if (errorMessage != null)
            return name + " = >" + errorMessage + "<"; // NOI18N
        String type = getWatch ().getType ();
        String innerType = getWatch ().getInnerType ();
        String value = getWatch ().getAsText ();
        if (getWatch ().isObject ()) {
            if (type.equals (innerType))
                return name + " = (" + innerType + ") " + value; // NOI18N
            else
                return name + " = (" + type + ") (" + innerType + ") " + value; // NOI18N
        }
        else
            return name + " = (" + type + ") " + value; // NOI18N
    }

    // CreateVariableCookie implementation ........................................
    
    public void createVariable () {
        WatchesRootNode n = (WatchesRootNode) DebuggerModule.getNode (WatchesRootNode.class);
        JavaVariable var = (JavaVariable) getWatch ().getVariable ();
        if ((n != null) && (var != null))
            n.createVariable (var.createFixedVariable ());
    }
        
    public boolean canCreateVariable () {
        return getWatch ().isInScope ();
    }    

}
