/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is Forte for Java, Community Edition. The Initial
 * Developer of the Original Code is Sun Microsystems, Inc. Portions
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.modules.debugger.support.java.nodes;

import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Comparator;
import java.lang.reflect.Modifier;
import java.lang.ref.WeakReference;
import java.beans.*;

import org.openide.src.nodes.*;
import org.openide.src.*;
import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.debugger.DebuggerException;
import org.openide.util.actions.SystemAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.nodes.VariableNode;
import org.netbeans.modules.debugger.support.util.Utils;


class LoadedClassFactory extends DefaultFactory {

    private ClassNode node;
    private JavaClass clazz;

    public LoadedClassFactory (JavaClass clazz) {
        super (false);
        this.clazz = clazz;
    }

    public Children createChildren (JavaClass clazz, ClassElement element) {
        Children children = createClassChildren (element, this);
        JavaVariable loader = clazz.getClassLoader ();
        if (loader != null)
            children.add (new Node [] {new JavaVariableNode (loader)});
        return children;    
    }

    public Node createClassNode (ClassElement element) {
        JavaClass inner = clazz.findInnerClass (element);
        if (inner != null)
            return new ClassNode (inner, element, new LoadedClassFactory (inner));
        return new AbstractNode (Children.LEAF);
    }

    public Node createConstructorNode (ConstructorElement element) {
        return new ConstructorNode (element, false, clazz.getName ());
    }

    public Node createMethodNode (MethodElement element) {
        return new MethodNode (element, false, clazz.getName ());
    }    

    public Node createFieldNode (FieldElement element) {
        return  new FieldNode (clazz, element, false);
    }

    /*
    public Node createInitializerNode (InitializerElement element) {
    }
    */

} // LoadedClassFactory

// package private classes .........................................................................

class MethodNode extends MethodElementNode implements SetMethodBreakpointCookie {

    private String className;
    
    MethodNode (MethodElement element, boolean writable, String className) {
        super (element, writable);
        this.className = className;
        getCookieSet ().add (this);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerMethodNode");
    }

    public ConstructorElement getElement () {
        return (ConstructorElement) element;
    }
    
    public String getClassName () {
        return className;
    }
    
    public boolean hasCustomizer () {
        return false;
    }

    protected Sheet createSheet () {
        Sheet sheet = Sheet.createDefault();
        Sheet.Set ps = sheet.get(Sheet.PROPERTIES);
        ps.put(createModifiersProperty(false));
        ps.put(createNameProperty(false));
        ps.put(createParametersProperty(false));
        ps.put(createReturnProperty(false));
        return sheet;
    }

    public SystemAction[] getActions () {
        return new SystemAction[] {
                SystemAction.get (SetMethodBreakpointAction.class),
                // SystemAction.get (ToolsAction.class),
                null,
                SystemAction.get (PropertiesAction.class)
        };
    }

}

class ConstructorNode extends ConstructorElementNode implements SetMethodBreakpointCookie {

    private String className;
    
    ConstructorNode (ConstructorElement element, boolean writable, String className) {
        super (element, writable);
        this.className = className;
        getCookieSet ().add (this);
    }    

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerConstructorNode");
    }

    public ConstructorElement getElement () {
        return (ConstructorElement) element;
    }
    
    public String getClassName () {
        return className;
    }
    
    public boolean hasCustomizer () {
        return false;
    }

    protected Sheet createSheet () {
        Sheet sheet = Sheet.createDefault();
        Sheet.Set ps = sheet.get(Sheet.PROPERTIES);
        ps.put(createModifiersProperty(false));
        ps.put(createNameProperty(false));
        ps.put(createParametersProperty(false));
        return sheet;
    }

    public SystemAction[] getActions () {
        return new SystemAction[] {
                SystemAction.get (SetMethodBreakpointAction.class),
                // SystemAction.get (ToolsAction.class),
                null,
                SystemAction.get (PropertiesAction.class)
        };
    }
    
}

class FieldNode extends FieldElementNode {

    JavaVariable variable;
    
    public FieldNode (JavaClass clazz, FieldElement element, boolean writable) {
        super (element, writable);
        if (isStatic ()) {
            variable = clazz.getStaticVariable (element);
            variable.addPropertyChangeListener (new VariableListener (this));
        }
    }
    
    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerFieldNode");        //NOI18N
    }

    private boolean isStatic () {
        return (((FieldElement) element).getModifiers () & Modifier.STATIC) == Modifier.STATIC;
    }
    
    void updateStaticValue () {
        firePropertyChange ("staticValue", null, null);
    }
    
    public boolean hasCustomizer () {
        return false;
    }
    
    public SystemAction[] getActions () {
        return new SystemAction[] {
                SystemAction.get (PropertiesAction.class)
        };
    }
    
    protected Sheet createSheet () {
        Sheet sheet = Sheet.createDefault();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        ps.put(createModifiersProperty(false));
        ps.put(createNameProperty(false));
        ps.put(createTypeProperty(false));
        if (isStatic ()) {
            ps.put (new PropertySupport.ReadOnly (
                    "staticValue", // NOI18N
                    String.class,
                    ClassNode.getLocalizedString ("PROP_static_value"),
                    ClassNode.getLocalizedString ("HINT_static_value")
                ) {
                    public Object getValue () {
                        return variable.getAsText ();
                    }
                });
        } // if
        return sheet;
    }
    
}

class VariableListener implements PropertyChangeListener {
    
    private WeakReference ref;
    private JavaVariable var;
    
    VariableListener (FieldNode node) {
        ref = new WeakReference (node);
        var = node.variable;
    }
    
    public void propertyChange (PropertyChangeEvent event) {
        FieldNode node = (FieldNode) ref.get ();
        if (node == null)
            var.removePropertyChangeListener (this);
        else
            node.updateStaticValue ();
    }

}