/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is Forte for Java, Community Edition. The Initial
 * Developer of the Original Code is Sun Microsystems, Inc. Portions
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.java.nodes;

import java.util.Hashtable;
import java.util.Enumeration;

import org.netbeans.modules.debugger.support.java.JavaClass;

/**
 *
 * @author  dprusa
 * @version 
 */
class PackageContent extends Object {

    // the full name of the package
    private String name;

    // indicates, if the content has been changed
    private boolean changed;
    
    // list of classes in the package
    private Hashtable classes = new Hashtable ();
    
    // list of sub-packages of the package
    private Hashtable packages = new Hashtable ();
    
    
    // init ........................................................................
    public PackageContent (String name) {
        this.name = name;
        changed = true;
    }
    
    // methods .....................................................................
    
    public void add (JavaClass clazz) {
        add (clazz, null); //NOI18N
    }
    
    private void add (JavaClass clazz, String packageName) {
        if (packageName == null) 
            packageName = clazz.getName ();
        
        // do not add inner classes
        if (packageName.indexOf ('$') > -1)
            return;
        
        int index = packageName.indexOf ('.');
        if (index == -1) {
            // add a class
            Element el = (Element) classes.get (clazz);
            if (el == null) {
                // add a new class
                classes.put (clazz, new Element (clazz));
                changed = true;
            }
            else
                el.confirmed = true;
        }
        else {
            // add a package
            String subPackageName = packageName.substring (0, index);
            PackageContent subPackage = (PackageContent) packages.get (subPackageName);
            if (subPackage == null) {
                // add a new subpackage
                subPackage = new PackageContent (name.concat (".").concat (subPackageName));
                packages.put (subPackageName, subPackage);
                changed = true;
            }
            subPackage.add (clazz, packageName.substring (index + 1));
        }
    }

    /**
    * Returns package name.
    */
    public String getName () {
        return name.substring (name.lastIndexOf ('.') + 1);
    }
    
    /**
    * Returns full package name.
    */
    public String getFullName () {
        return name;
    }
    
    /**
    * Returns a subpackage content.
    */
    public PackageContent getSubContent (String name) {
        return ((PackageContent) packages.get (name));
    }
    
    public boolean isEmpty () {
        return (packages.size () == 0) && (classes.size () == 0);
    }
    
    /**
    * Returns subpackages and classes in this package.
    */
    public Object [] getContent () {
        Object [] res = new Object [classes.size () + packages.size()];
        int index = 0;
        Enumeration enum = packages.elements ();
        while (enum.hasMoreElements ())
            res [index++] = enum.nextElement ();
        enum = classes.elements ();
        while (enum.hasMoreElements ())
            res [index++] = ((Element)enum.nextElement ()).clazz;
        return res;
    }
    
    public void reorganize () {
        Enumeration enum = packages.elements ();
        while (enum.hasMoreElements ()) {
            PackageContent sub = (PackageContent) enum.nextElement ();
            sub.reorganize ();
            if (sub.isEmpty ()) {
                packages.remove (sub.getName());
                changed = true;
            }
        }
        enum = classes.elements ();
        while (enum.hasMoreElements ()) {
            Element el = (Element) enum.nextElement ();
            if (!el.newOne && !el.confirmed) {
                classes.remove (el.clazz);
                changed = true;
            }
            else {
                el.newOne = false;
                el.confirmed = false;
            }
        }
    }
    
    public boolean changed () {
        return changed;
    }

    public void clearChangedFlag () {
        changed = false;
    }
    
    public boolean equals (Object o) {
        if (!(o instanceof PackageContent))
            return false;
        return name.equals (((PackageContent) o).getFullName ());
    }
    
    public int hashCode () {
        return name.hashCode ();
    }
    
    // inner class ..............................................................
    // For storing JavaClass in hashtable, an additional flags are added.
    private class Element {
        
        public JavaClass clazz;
        public boolean confirmed;
        public boolean newOne;
        
        public Element (JavaClass c) {
            clazz = c;
            confirmed = false;
            newOne = true;
        }

    }
    
}