/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;
import java.lang.ref.WeakReference;

import org.openide.TopManager;
import org.openide.debugger.*;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallableSystemAction;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.BeanNode;
import org.openide.nodes.AbstractNode;

import org.netbeans.modules.debugger.CoreBreakpoint;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.DebuggerModule;


/**
* This class represents root of breakpoints.
* This class is final only for performance reasons,
* can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class BreakpointsRootNode extends AbstractNode implements
BreakpointsRootCookie {
    
    /** generated Serialized Version UID */
    static final long serialVersionUID = -6708027507109804744L;


    // static ....................................................................

    private static final String        ICON_BASE =
        "/org/netbeans/core/resources/breakpoints"; // NOI18N

    /** Popup menu actions. */
    private static SystemAction []     staticActions;


    // init ......................................................................

    /**
    * Creates empty BreakpointsRootNode.
    */
    public BreakpointsRootNode () {
        super (new BreakpointsRootChildren());
        String name = DebuggerNode.getLocalizedString ("CTL_Breakpoints_group_root");
        setDisplayName (name);
        setName (name);
        setIconBase (ICON_BASE);
        getCookieSet ().add (this);
        init ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerBreakpointsRootNode");        //NOI18N
    }

    /** deserializes object */
    private void readObject(java.io.ObjectInputStream obis)
    throws java.io.IOException, ClassNotFoundException,
        java.io.NotActiveException {
        obis.defaultReadObject ();
        init ();
    }

    private void init () {
    }


    // Node implementation .......................................................

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
            SystemAction.get (AddBreakpointAction.class),
            null,
            SystemAction.get (EnableAllBreakpointsAction.class),
            SystemAction.get (DisableAllBreakpointsAction.class),
            SystemAction.get (DeleteAllAction.class),
            null,
            SystemAction.get (ToolsAction.class),
            SystemAction.get (PropertiesAction.class),
        };
        return staticActions;
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new BreakpointsHandle ();
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (AddBreakpointAction.class);
    }


    // BreakpointsRootCookie implementation ......................................

    /**
    * Delete all breakpoints.
    */
    public void deleteAll () {
        try {
            // obtain reference to the debugger instance
            Debugger debugger = TopManager.getDefault ().getDebugger ();
            debugger.removeAllBreakpoints ();
            debugger.removeAllBreakpoints ();
        } catch (DebuggerException exc) {
        }
    }

    /**
    * Enable all breakpoints.
    */
    public void enableAll () {
        try {
            // obtain reference to the debugger instance
            Debugger debugger = TopManager.getDefault ().getDebugger ();
            Breakpoint[] b = debugger.getBreakpoints ();
            int i, k = b.length;
            for (i = 0; i < k; i++)
                b [i].setEnabled (true);
        } catch (DebuggerException exc) {
        }
    }

    /**
    * Disable all breakpoints.
    */
    public void disableAll () {
        try {
            // obtain reference to the debugger instance
            Debugger debugger = TopManager.getDefault ().getDebugger ();
            Breakpoint[] b = debugger.getBreakpoints ();
            int i, k = b.length;
            for (i = 0; i < k; i++)
                b [i].setEnabled (false);
        } catch (DebuggerException exc) {
        }
    }

    
    // innerclasses ..............................................................

    /** Special breakpoints context subnodes (children) */
    private static final class BreakpointsRootChildren extends Children.Keys {

        Debugger                    debugger = null;
        ArrayList                   c = new ArrayList ();
        BreakpointComparator        breakpointComparator = new BreakpointComparator ();
        //         BreakpointComparator is not consistent with equals !!!!!!!!!!!!!!!!!!!!!!!!!!
        
        protected void addNotify () {
            try {
                Debugger debugger =
                    (Debugger) TopManager.getDefault ().getDebugger ();
                BRListener l = new BRListener (
                    (BreakpointsRootNode) getNode (), 
                    debugger
                );
            } catch (DebuggerException e) {
            }
        }
        
        private void refresh (Breakpoint[] bs) {
            Arrays.sort (bs, breakpointComparator);
            setKeys (bs);
        }
        
        private void refresh (Breakpoint b) {
            refreshKey (b);
        }
        
        /** Create BreakpointNode for a given key.
        * @param key the key
        * @return one BreakpointNode
        */
        protected Node[] createNodes (Object key) {
            if ((key instanceof Breakpoint) && ((Breakpoint) key).isHidden ())
                return new Node [0];
            Node node = DebuggerNode.createNode (key);
            if (node == null)
                return new Node [0];
            return new Node[] {node};
        }
        
        static class BRListener implements PropertyChangeListener {

            private Debugger                debugger;
            private WeakReference           node;
            private Breakpoint[]            breakpoints = new Breakpoint [0];
            private 


            BRListener (BreakpointsRootNode node, Debugger debugger) {
                this.debugger = debugger;
                this.node = new WeakReference (node);

                // add listeners & set first children
                debugger.addPropertyChangeListener (this);
                refreshBreakpoints (node);
            }

            BreakpointsRootNode getNode () {
                BreakpointsRootNode wrn = (BreakpointsRootNode) node.get ();
                if (wrn != null) return wrn;

                // remove all listeners!
                debugger.removePropertyChangeListener (this);
                int i, k = breakpoints.length;
                for (i = 0; i < k; i++)
                    breakpoints [i].removePropertyChangeListener (this);
                return null;
            }

            public void propertyChange (PropertyChangeEvent e) {
                BreakpointsRootNode node = getNode ();
                if (node == null) return;

                if ( (e.getPropertyName () != null) &&
                     e.getPropertyName ().equals (Debugger.PROP_BREAKPOINTS)
                ) refreshBreakpoints (node);
                else
                if ( (e.getPropertyName () != null) &&
                     e.getPropertyName ().equals (CoreBreakpoint.PROP_HIDDEN)
                ) ((BreakpointsRootChildren) node.getChildren ()).
                    refresh ((Breakpoint) e.getSource ());
            }

            void refreshBreakpoints (BreakpointsRootNode node) {
                int i, k = breakpoints.length;
                for (i = 0; i < k; i++)
                    breakpoints [i].removePropertyChangeListener (this);
                breakpoints = debugger.getBreakpoints ();
                ((BreakpointsRootChildren) node.getChildren ()).refresh (breakpoints);
                k = breakpoints.length;
                for (i = 0; i < k; i++)
                    breakpoints [i].addPropertyChangeListener (this);
            }
        }// BRListener
        
        private class BreakpointComparator implements Comparator, java.io.Serializable {
            public int compare (Object o1, Object o2) {
                String s1 = "";
                try {
                    if (o1 instanceof CoreBreakpoint)
                        s1 = ((CoreBreakpoint) o1).getEvent (debugger).getDisplayName ();
                    else
                        s1 = ((Breakpoint) o1).getClassName ();
                } catch (NullPointerException e) {
                }
                String s2 = "";
                try {
                    if (o2 instanceof CoreBreakpoint)
                        s2 = ((CoreBreakpoint) o2).getEvent (debugger).getDisplayName ();
                    else
                        s2 = ((Breakpoint) o2).getClassName ();
                } catch (NullPointerException e) {
                }
                return s1.compareToIgnoreCase (s2);
            }
        }
    } // end of BreakpointsContextChildren inner class

    /**
    * Serializable node reference. 
    */
    private static class BreakpointsHandle implements Node.Handle {
        static final long serialVersionUID =-4518262935887259653L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return DebuggerModule.getBreakpointsRootNode ();
        }
    }
}
