/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.util.Comparator;
import java.util.ResourceBundle;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.lang.ref.WeakReference;

import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;

import org.netbeans.modules.debugger.*;
//import org.netbeans.modules.debugger.support.*;

/**
* This class represents callstack as a Node.
* Class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class CallStackNode extends AbstractNode {
    /** generated Serialized Version UID */
    static final long                 serialVersionUID = -8259352660663524178L;


    // static variables ..........................................................

    private static String             ICON_BASE =
        "/org/netbeans/core/resources/callstack"; // NOI18N
    private static ResourceBundle     bundle;


    // variables .................................................................

    private AbstractThread            thread;
    private CallStackListener         callStackListener;
//    private Node                      aditionalThreadNode;


    // init ......................................................................

    /**
    * Creates empty CallStackNode node.
    */
    public CallStackNode (AbstractThread thread) {
        super (new CallStackChildren ());
        this.thread = thread;
        String s = getLocalizedString ("CTL_CallStack");
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
        changeChildren ();
//        changeAdditionalNode ();
        callStackListener = new CallStackListener (this);
        thread.addPropertyChangeListener (callStackListener);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerCallStackNode");        //NOI18N
    }


    // helper methods ............................................................

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (CallStackNode.class);
        return bundle.getString (s);
    }

    public void changeChildren () {
        CallStackChildren myChildren = (CallStackChildren) getChildren();
        Location[] callStack = ((CallStackProducer) thread).getFilteredCallStack ();
        Node [] node = myChildren.getNodes ();
        int stackLength = callStack.length;
        int nodesLength = node.length;

        int i, k = Math.min (stackLength, nodesLength);
        for (i = 0; i < k; i++) {
            Node n = node [nodesLength - i - 1];
            if (n instanceof LocationNode)
                ((LocationNode) n).updateLocation (
                    callStack [stackLength - i - 1]
                );
        }
        for (i = k; i < stackLength; i++) {
            myChildren.add (callStack [stackLength - i - 1], i);
        }
        for (i = k; i < node.length; i++) {
            Node n = node [nodesLength - i - 1];
            if (n instanceof LocationNode)
                myChildren.remove (new Node [] {n});
        }
    }

/*    private void changeAdditionalNode () {
        Children ch = getChildren ();
        Node n = thread.getAditionalThreadInfo ();
        if (n == null) {
            if (aditionalThreadNode != null) {
                ch.remove (new Node[] {aditionalThreadNode});
                aditionalThreadNode = null;
            }
        } else {
            if (aditionalThreadNode != null)
                ch.remove (new Node[] {aditionalThreadNode});
            aditionalThreadNode = n;
            ch.add (new Node[] {aditionalThreadNode});
        }
    }*/


    // innerclasses ..............................................................

    /** Special locals subnodes (children) */
    private static final class CallStackChildren extends Children.SortedArray
        implements Comparator {

        CallStackChildren () {
            setComparator (this);
        }

        public int compare (Object o1, Object o2) throws ClassCastException {
            int index1 = (o1 instanceof LocationNode) ?
                         ((LocationNode) o1).getIndex () :
                         Integer.MIN_VALUE;
            int index2 = (o2 instanceof LocationNode) ?
                         ((LocationNode) o2).getIndex () :
                         Integer.MIN_VALUE;

            if (index1 > index2) return -1;
            if (index1 == index2) return 0;
            return 1;
        }

        private void add (Location loc, int index) {
            Node n = NodesRegistry.getNode (loc, new Object[] {loc, new Integer (index)});
            if (n != null) 
                add (new Node [] {n});
        }

    } // end of CallStackChildren inner class

    /**
    * Listens on thread and debugger.
    */
    private static class CallStackListener implements PropertyChangeListener {

        private WeakReference             node;
        private AbstractThread            thread;


        CallStackListener (CallStackNode n) {
            node = new WeakReference (n);
            thread = n.thread;
        }

        public void propertyChange (PropertyChangeEvent e) {
            CallStackNode csn = (CallStackNode) node.get ();
            if (csn == null) {
                thread.removePropertyChangeListener (this);
                return;
            }
/*            if ( (e.getPropertyName () != null) &&
                    (e.getPropertyName ().equals (
                         AbstractThread.PROP_ADITIONAL_THREAD_INFO
                     ))
               ) {
                csn.changeAdditionalNode ();
            } else*/
                if ( (e.getPropertyName () != null) &&
                        (e.getPropertyName ().equals (
                             CallStackProducer.PROP_CALLSTACK
                         ))
                   ) {
                    csn.changeChildren ();
                }
        }
    }
}

