/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.util.Comparator;
import java.util.ResourceBundle;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.lang.ref.WeakReference;

import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.debugger.DebuggerException;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.DebuggerModule;

/**
* This class represents callstack as a Node.
* Class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class CallStackRootNode extends AbstractNode {
    /** generated Serialized Version UID */
    static final long                 serialVersionUID = -8259352660663524178L;


    // static variables ..........................................................

    private static String             ICON_BASE =
        "/org/netbeans/core/resources/callstack"; // NOI18N
    private static ResourceBundle     bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (CallStackRootNode.class);
        return bundle.getString (s);
    }


    // variables .................................................................

    private AbstractThread            thread;
    private ThreadsRoot               debugger;
    private CallStackListener         callStackListener;
//    private Node                      aditionalThreadNode;


    // init ......................................................................

    /**
    * Creates empty CallStackNode node.
    */
    public CallStackRootNode () {
        super (new CallStackChildren ());
        String s = getLocalizedString ("CTL_CallStack");
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
//        changeAdditionalNode ();
        try {
            debugger = (ThreadsRoot) TopManager.getDefault ().getDebugger ();
        } catch (DebuggerException e) {
        }
        changeChildren ();
        callStackListener = new CallStackListener (this);
        thread = debugger.getCurrentThread ();
        if (thread != null)
            thread.addPropertyChangeListener (callStackListener);
        ((AbstractDebugger) debugger).addPropertyChangeListener (callStackListener);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerCallStackRootNode");        //NOI18N
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new CallStackHandle ();
    }

    // helper methods ............................................................

    void changeChildren () {
        CallStackChildren myChildren = (CallStackChildren) getChildren();
        thread = debugger.getCurrentThread ();
        if ( (thread == null) ||
             !(thread instanceof CallStackProducer)
        ) {
            myChildren.remove (myChildren.getNodes ());
            return;
        }
        
        Location[] callStack = ((CallStackProducer) thread).getFilteredCallStack ();
        Node [] node = myChildren.getNodes ();
        int stackLength = callStack.length;
        int nodesLength = node.length;

        int i, k = Math.min (stackLength, nodesLength);
        for (i = 0; i < k; i++) {
            Node n = node [nodesLength - i - 1];
            if (n instanceof LocationNode)
                ((LocationNode) n).updateLocation (
                    callStack [stackLength - i - 1]
                );
        }
        for (i = k; i < stackLength; i++) {
            myChildren.add (callStack [stackLength - i - 1], i);
        }
        for (i = k; i < node.length; i++) {
            Node n = node [nodesLength - i - 1];
            if (n instanceof LocationNode)
                myChildren.remove (new Node [] {n});
        }
    }

/*    private void changeAdditionalNode () {
        Children ch = getChildren ();
        Node n = thread.getAditionalThreadInfo ();
        if (n == null) {
            if (aditionalThreadNode != null) {
                ch.remove (new Node[] {aditionalThreadNode});
                aditionalThreadNode = null;
            }
        } else {
            if (aditionalThreadNode != null)
                ch.remove (new Node[] {aditionalThreadNode});
            aditionalThreadNode = n;
            ch.add (new Node[] {aditionalThreadNode});
        }
    }*/


    // innerclasses ..............................................................

    /** Special locals subnodes (children) */
    private static final class CallStackChildren extends Children.SortedArray
        implements Comparator {

        CallStackChildren () {
            setComparator (this);
        }

        public int compare (Object o1, Object o2) throws ClassCastException {
            int index1 = (o1 instanceof LocationNode) ?
                         ((LocationNode) o1).getIndex () :
                         Integer.MIN_VALUE;
            int index2 = (o2 instanceof LocationNode) ?
                         ((LocationNode) o2).getIndex () :
                         Integer.MIN_VALUE;

            if (index1 > index2) return -1;
            if (index1 == index2) return 0;
            return 1;
        }

        private void add (Location loc, int index) {
            Node n = NodesRegistry.getNode (loc, new Object[] {loc, new Integer (index)});
            if (n != null) 
                add (new Node [] {n});
        }

    } // end of CallStackChildren inner class

    /**
    * Listens on thread and debugger.
    */
    private static class CallStackListener implements PropertyChangeListener {

        private WeakReference             node;


        CallStackListener (CallStackRootNode n) {
            node = new WeakReference (n);
        }

        public void propertyChange (PropertyChangeEvent e) {
            CallStackRootNode csn = (CallStackRootNode) node.get ();
            if (csn == null) {
                csn.thread.removePropertyChangeListener (this);
                return;
            }
/*            if ( (e.getPropertyName () != null) &&
                    (e.getPropertyName ().equals (
                         AbstractThread.PROP_ADITIONAL_THREAD_INFO
                     ))
               ) {
                csn.changeAdditionalNode ();
            } else*/
            if ( (e.getPropertyName () != null) &&
                    (e.getPropertyName ().equals (
                         ThreadsRoot.PROP_CURRENT_THREAD
                     ))
            ) {
                if (csn.thread != null) 
                    csn.thread.removePropertyChangeListener (this);
                csn.thread = csn.debugger.getCurrentThread ();
                if (csn.thread != null)
                    csn.thread.addPropertyChangeListener (this);
                csn.changeChildren ();
            } else
            if ( (e.getPropertyName () != null) &&
                    (e.getPropertyName ().equals (
                         CallStackProducer.PROP_CALLSTACK
                     ))
               ) {
                csn.changeChildren ();
            }
        }
    }

    /**
    * Serializable node reference. 
    */
    private static class CallStackHandle implements Node.Handle {
        static final long serialVersionUID =-45182213487259891L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return DebuggerModule.getCallStackRootNode ();
        }
    }
}

