/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.Vector;
import java.util.Collection;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.debugger.Debugger;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.WeakListener;
import org.openide.nodes.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.DebuggerAdapter;
import org.netbeans.modules.debugger.support.DebuggerModule;

/** The node representing debugger.
*
* @author Jan Jancura, Petr Hamernik, Jaroslav Tulach
*/
public class DebuggerNode extends AbstractNode {


    // static ....................................................................

    /** generated Serialized Version UID */
    static final long serialVersionUID = 6394601904375687521L;

    public static final String    PROP_SHOW_MESSAGES      = "showMessages"; // NOI18N
    public static final String    PROP_SHOW_IN_EDITOR     = "showInEditor"; // NOI18N
    public static final String    PROP_REMOTE_DEBUGGING   = "remoteDebugging"; // NOI18N
    public static final String    PROP_TOTAL_MEMORY       = "totalMemory"; // NOI18N
    public static final String    PROP_FREE_MEMORY        = "freeMemory"; // NOI18N
    public static final String    PROP_SOURCE_PATH        = "sourcePath"; // NOI18N
    public static final String    PROP_EXCEPTION_CATCH_LIST = "exceptionCatchList"; // NOI18N
    public static final String    PROP_DEBUGGER_STATE     = "debuggerState"; // NOI18N

    private static final String   ICON_BASE =
        "/org/netbeans/core/resources/debugger"; // NOI18N

    /** bundle to obtain text information from */
    private static ResourceBundle bundle;

    /*
     * Returns bundle for this package.
     *
     * @return bundle for this package
     */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (DebuggerNode.class);
        return bundle.getString (s);
    }
    
    public static Node createNode (Object w) {
        Node n = NodesRegistry.getNode (w);
        if (n == null) {
            try {
                n = new BeanNode (w);
            } catch (java.beans.IntrospectionException e) {
                e.printStackTrace ();
            }
        }
        return n;
    }
    

    // variables .................................................................

    /** Reference to the debugger */
    private transient AbstractDebugger debugger;
    /** Reference to the debugger listener */
    private transient DebuggerAdapter debuggerListener;
    /** Reference to the top manager listener */
    private transient PropertyChangeListener topManagetListener;


    // init ......................................................................

    /** constructor.
    */
    public DebuggerNode () {
        super (new Children.Array ());
        String name = getLocalizedString ("CTL_Debugger");
        setDisplayName (name);
        setName (name);
        setShortDescription (getLocalizedString ("HINT_Debugger"));
        setIconBase (ICON_BASE);
        initialize ();
    }

    /** Performs initialization - subnodes, properties, listeners */
    private void initialize () {
        // initialize chidren
        Collection c = DebuggerModule.getNodes ();

        getChildren ().add (
            (Node []) c.toArray (new Node [c.size ()])
        );
        // properties...
        createProperties();
        
        TopManager.getDefault ().addPropertyChangeListener (
            WeakListener.propertyChange (
                new PropertyChangeListener () {
                    public void propertyChange (PropertyChangeEvent ev) {
                        if ( (ev.getPropertyName () == null) ||
                             (!ev.getPropertyName ().equals (TopManager.PROP_DEBUGGER))
                        ) return;
                        debuggerChanged ();
                    }
                },
                TopManager.getDefault ()
            )
        );
        debuggerChanged ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerDebuggerNode");        //NOI18N
    }


    // node implementation .......................................................

    /** Creates property sets */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        ps.put(new PropertySupport.ReadOnly (
                   PROP_DEBUGGER_STATE,
                   String.class,
                   getLocalizedString ("PROP_debugger_state"),
                   getLocalizedString ("HINT_debugger_state")
               ) {
                   public Object getValue () {
                       return getDebuggerState ();
                   }
               });
        // and set new sheet
        setSheet (sheet);
    }


    // other methods .............................................................

    /**
     * Returns string description of the debugger state.
     *
     * &return string description of the debugger state
     */
    public String getDebuggerState () {
        if (debugger == null) return ""; // NOI18N
        switch (debugger.getState ()) {
            case Debugger.DEBUGGER_NOT_RUNNING: return getLocalizedString ("VAL_NotRrunning");
            case Debugger.DEBUGGER_STARTING:    return getLocalizedString ("VAL_Starting");
            case Debugger.DEBUGGER_RUNNING:     return getLocalizedString ("VAL_Running");
            case Debugger.DEBUGGER_STOPPED:     return getLocalizedString ("VAL_Stopped");
        }
        return ""; // NOI18N
    }

    /** Fires change of state
    */
    void debuggerStateChanged () {
        firePropertyChange (null, null, null);
    }
    
    void debuggerChanged () {
        if (debugger != null) {
            debugger.removeDebuggerListener (debuggerListener);
        }
        
        try {
            debugger = (AbstractDebugger) TopManager.getDefault ().getDebugger ();
        } catch (DebuggerNotFoundException e) {
        }
        if (debugger == null) return;
        
        if (debuggerListener == null) 
            debuggerListener = new DebuggerAdapter () {
                public void debuggerStateChanged (int debuggerState) {
                    DebuggerNode.this.debuggerStateChanged ();
                }
            };
            
        debugger.addDebuggerListener (debuggerListener);
    }
}

