/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.util.*;
import java.util.*;
import java.beans.*;
import java.lang.ref.WeakReference;

import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
//import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* This class represents locals as a Node.
* This class is final only for performance reasons.
* Can be unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class LocalsRootNode extends AbstractNode {


    // static variables ..........................................................

    /** generated Serialized Version UID */
    static final long                  serialVersionUID = -6346315017458451778L;

    private static String              ICON_BASE =
        "/org/netbeans/core/resources/watches"; // NOI18N

    private static ResourceBundle      bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (LocalsRootNode.class);
        return bundle.getString (s);
    }


    // variables .................................................................

    private AbstractThread            thread;
    private LocalsListener            localsListener;

    // init ......................................................................

    /**
    * Creates empty BreakpointsContext.
    */
    public LocalsRootNode(AbstractThread thread) {
        super (new LocalsRootChildren ());
        this.thread = thread;
        String s = getLocalizedString ("CTL_Locals");
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
        changeChildren ();
        localsListener = new LocalsListener (this);
        thread.addPropertyChangeListener (localsListener);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerLocalsRootNode");        //NOI18N
    }

    
    // helper methods ............................................................

    public void changeChildren () {
        // set locals as keys into our children
        if (thread instanceof VariablesProducer) {
            LocalsRootChildren myChildren = (LocalsRootChildren) getChildren ();
            AbstractVariable [] locals = ((VariablesProducer) thread).getVariables ();
            String info = ((VariablesProducer) thread).getInfo ();
            if (info != null)
                info = getLocalizedString ("CTL_Locals") + " (" + info + ')';
            else
                info = getLocalizedString ("CTL_Locals");
            setDisplayName (info);
            setName (info);
            Arrays.sort (locals, Utils.localsComparator);
            myChildren.setMyKeys (locals);
        }
    }

        // innerclasses ..............................................................

    /** Special locals subnodes (children) */
    private static final class LocalsRootChildren extends Children.Keys implements LeafRefresher {

        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (final Object key) {
            return new Node[] {
                NodesRegistry.getNode (key)
            };
        }

        /** Accessor for LocalsRootNode outer class */
        private void setMyKeys (final Object[] keys) {
            setKeys (keys);
        }

        // interface LeafRefresher

        /** Accessor for VariableNode class */
        public void refreshMyKey (Object key) {
            refreshKey (key);
        }

    } // end of LocalsChildren inner class


    /**
    * Listens on thread and debugger.
    */
    private static class LocalsListener implements PropertyChangeListener {

        private WeakReference             node;
        private AbstractThread            thread;


        LocalsListener (LocalsRootNode n) {
            node = new WeakReference (n);
            thread = n.thread;
        }

        public void propertyChange (PropertyChangeEvent e) {
            LocalsRootNode csn = (LocalsRootNode) node.get ();
            if (csn == null) {
                thread.removePropertyChangeListener (this);
                return;
            }
            if ( (e.getPropertyName () != null) &&
                    (e.getPropertyName ().equals (
                         VariablesProducer.PROP_VARIABLES
                     ))
               ) {
                csn.changeChildren ();
            }
        }
    }
}
