/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.ArrayList;
import java.util.Arrays;

import org.openide.TopManager;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.text.Line;
import org.openide.cookies.LineCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.debugger.DebuggerException;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.PropertySupport;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallableSystemAction;

import org.netbeans.modules.debugger.*;
//import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.GoToSourceAction;
import org.netbeans.modules.debugger.support.actions.GoToSourceCookie;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* This class callstack frame as a Node.
* This class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public class LocationNode extends AbstractNode implements GoToSourceCookie {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = -4901112834613957792L;


    // static ....................................................................

    /** Property name constant */
    public static final String      PROPERTY_CLASS_NAME = "className"; // NOI18N
    /** Property name constant */
    public static final String      PROPERTY_METHOD_NAME = "methodName"; // NOI18N
    /** Property name constant */
    public static final String      PROPERTY_LINE_NUMBER = "lineNumber"; // NOI18N

    /** Icon base for this node */
    private static final String     ICON_BASE =
        "/org/netbeans/core/resources/callstack"; // NOI18N
    private static ResourceBundle   bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (LocationNode.class);
        return bundle.getString (s);
    }

    
    // variables .................................................................

    /** Popup menu actions. */
    private SystemAction[]          staticActions;

    /** CallStack frame of this CallStackLineNode. */
    protected Location              location;

    /** index of line in call stack */
    protected int                   index;

    // init ......................................................................

    /** Creates location node with asociated location
    */
    public LocationNode (Location location, int index) {
        super (location instanceof VariablesProducer ? 
            new LocationChildren (location) : 
            Children.LEAF
        );
        this.location = location;
        this.index = index;
        initialize ();
    }

    /** Performes initialization of this nodeS */
    private void initialize () {
        String s = location.getLocationName ();
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
        createProperties ();
        getCookieSet ().add (this);
        changeChildren ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerLocationNode");        //NOI18N
    }


    // Node implementation .....................................................

    /** Creates properties of this node */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set props = sheet.get (Sheet.PROPERTIES);
        props.put (new PropertySupport.ReadOnly (
            PROPERTY_CLASS_NAME,
            String.class,
            getLocalizedString ("PROP_call_stack_class_name"),
            getLocalizedString ("HINT_call_stack_class_name")
        ) {
            public Object getValue () throws InvocationTargetException {
                if (location.getLine () == null) return "";
                return location.getLine ().getDataObject ().getName ();
            }
        });
        props.put (new PropertySupport.ReadOnly (
            PROPERTY_LINE_NUMBER,
            Integer.TYPE,
            getLocalizedString ("PROP_call_stack_line_number"),
            getLocalizedString ("HINT_call_stack_line_number")
        ) {
            public Object getValue () throws InvocationTargetException {
                try {
                    if (location.getLine () == null) return new Integer (-1);
                    return new Integer (location.getLine ().getLineNumber ());
                } catch (Exception e) {
                    throw new InvocationTargetException (e);
                }
            }
        });
        // and set new sheet
        setSheet (sheet);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                SystemAction.get (GoToSourceAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class),
            };
        return staticActions;
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (GoToSourceAction.class);
    }


    // GoToSourceCookie implementation ...........................................

    /**
    * Goes to source.
    */
    public void goToSource () {
        Line l = location.getLine ();
        if (l != null) l.show (Line.SHOW_GOTO);
    }

    /**
    * Returns true if source is availabled.
    */
    public boolean canGoToSource () {
        Line l = location.getLine ();
        return l != null;
    }


    // helper methods ............................................................

    /** Sets a new Location and updates node. */
    void updateLocation (Location location) {
        this.location = location;
        String s = location.getLocationName ();
        setDisplayName (s);
        setName (s);
        changeChildren ();
    }

    /** Returns index in callstack */
    int getIndex () {
        return index;
    }

    void changeChildren () {
        if (!(location instanceof VariablesProducer)) return;
        // set locals as keys into our children
        LocationChildren myChildren = (LocationChildren) getChildren ();
        AbstractVariable [] locals = ((VariablesProducer) location).getFilteredVariables ();
        Arrays.sort (locals, Utils.localsComparator);
        myChildren.setMyKeys (locals);
    }

    // innerclass ................................................................

    /** Empty list of children. Does not allow anybody to insert a node.
    * Treated especially in the attachTo method.
    */
    private static final class LocationChildren extends Children.Keys implements LeafRefresher {
        /** CallStack frame of this CallStackLineNode. */
        private Location                 location;

        /** Constructs children asociated with given stack frame */
        LocationChildren (
            final Location location
        ) {
            this.location = location;
        }

        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (final Object key) {
            Node n = NodesRegistry.getNode (key);
            if (n == null) return new Node [0];
            return new Node[] {n};
        }

        /** Accessor for LocalsRootNode outer class */
        private void setMyKeys (final Object[] keys) {
            setKeys (keys);
        }

        // interface LeafRefresher

        /** Accessor for VariableNode class */
        public void refreshMyKey (Object key) {
            refreshKey (key);
        }

    }
}

