/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.ref.WeakReference;
import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Arrays;
import javax.swing.SwingUtilities;

import org.openide.TopManager;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.DebuggerException;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.DebuggerModule;
import org.netbeans.modules.debugger.support.actions.SuspendCookie;
import org.netbeans.modules.debugger.support.actions.SuspendAction;
import org.netbeans.modules.debugger.support.actions.ResumeAction;

/**
* This class represents threadGroup as a Node.
* This class is final only for performance reasons,
* can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public class ThreadGroupNode extends AbstractNode {
    /** generated Serialized Version UID */
    static final long                   serialVersionUID = -4736310787523747464L;


    // static ....................................................................

    /** Property names */
    public static final String          PROP_GROUP_NAME       = "threadName"; // NOI18N

    protected static Image              threadGroupIcon;
    protected static Image              threadGroupCurrentIcon;

    /** System actions of this node */
    private static SystemAction[]       staticActions;

    /** Icon paths. */
    protected static final String       ICON_CURRENT =
        "/org/netbeans/modules/debugger/resources/threadGroupCurrent"; // NOI18N
    protected static final String       ICON_BASE =
        "/org/netbeans/modules/debugger/resources/threadGroup"; // NOI18N

    private static ResourceBundle       bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (ThreadGroupNode.class);
        return bundle.getString (s);
    }


    // variables .................................................................

    private transient ThreadsProducer   threadGroup;
    private transient TGListener        TGListener;
    private boolean                     current = false;
    private boolean                     root = false;


    // init ......................................................................

    /**
    * Creates thread group context with given thread group
    */
    public ThreadGroupNode (
        ThreadsProducer threadGroup,
        boolean root
    ) {
        super (new ThreadGroupContextChildren ());
        this.threadGroup = threadGroup;
        this.root = root;
        String s = getTGName ();
        setDisplayName (s);
        setName (s);
        initialize();
    }

    /**
    * Creates thread group context with given thread group
    */
    public ThreadGroupNode (
        ThreadsProducer threadGroup
    ) {
        this (threadGroup, false);
    }

    private void initialize () {
        current = isCurrent ();
        if (current)
            setIconBase (ICON_CURRENT);
        else
            setIconBase (ICON_BASE);
        createProperties();
        if (TGListener == null) TGListener = new TGListener (this);
        if (threadGroup instanceof AbstractThread)
            ((AbstractThread) threadGroup).addPropertyChangeListener (TGListener);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerThreadGroupNode");        //NOI18N
    }


    // Node implementation .......................................................

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class)
            };
        return staticActions;
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new ThreadsHandle ();
    }

    /** Creates property sets */
    private void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
/*        ps.put (new PropertySupport.ReadOnly (
            PROP_GROUP_NAME,
            String.class,
            getLocalizedString ("PROP_group_name"),
            getLocalizedString ("HINT_group_name")
        ) {
            public Object getValue () throws InvocationTargetException {
                try {
                    return getTGName ();
                } catch (Exception e) {
                    throw new InvocationTargetException (e);
                }
            }
        });*/
        // and set new sheet
        setSheet (sheet);
    }

    public void destroy () throws java.io.IOException {
        super.destroy ();
        threadGroup.removeThreadListener (TGListener);
        if (threadGroup instanceof AbstractThread)
            ((AbstractThread) threadGroup).removePropertyChangeListener (TGListener);
    }


    // helper methods ............................................................

    ThreadsProducer getDebuggerThreadGroup () {
        return threadGroup;
    }

    void changeProperties () {
        if ( (threadGroup == null) || 
             (current == isCurrent ())
        ) return;
        current = isCurrent ();
        if (current)
            setIconBase (ICON_CURRENT);
        else
            setIconBase (ICON_BASE);
    }
    
    String getTGName () {
        if (root)
            return getLocalizedString ("CTL_Thread_root");
        if (threadGroup instanceof AbstractThread)
            return ((AbstractThread) threadGroup).getName ();
        return getLocalizedString ("CTL_Thread_group");
    }
    
    boolean isCurrent () {
        if (threadGroup instanceof AbstractThread)
            return ((AbstractThread) threadGroup).isCurrent ();
        return false;//PENDING
    }


    // inner classes .............................................................

    private static class TGListener implements ThreadListener, 
    PropertyChangeListener {

        private WeakReference             node;
        private ThreadsProducer           threadGroup;

        TGListener (ThreadGroupNode n) {
            node = new WeakReference (n);
            threadGroup = n.threadGroup;
        }

        public void threadCreated (AbstractThread tg) {
            updateChindren ();
        }

        public void threadDeath (AbstractThread dt) {
            updateChindren ();
        }

        public void propertyChange (PropertyChangeEvent e) {
            ThreadGroupNode tgn = getTG ();
            if (tgn == null) return;
            tgn.changeProperties ();
            //updateChindren ();
        }

        private void updateChindren () {
            ThreadGroupNode tgn = getTG ();
            if (tgn == null) return;

            ThreadGroupContextChildren myChildren =
                (ThreadGroupContextChildren) tgn.getChildren ();
            myChildren.update ();
        }

        private ThreadGroupNode getTG () {
            ThreadGroupNode tgn = (ThreadGroupNode) node.get ();
            if (tgn == null) {
                threadGroup.removeThreadListener (this);
                if (threadGroup instanceof AbstractThread)
                    ((AbstractThread) threadGroup).removePropertyChangeListener (this);
            }
            return tgn;
        }
    }

    /** Special Thread group context subnodes (children) */
    private static final class ThreadGroupContextChildren extends Children.Keys {

        /** Where to add variables. */
        private boolean                     initialized = false;

        /** Create a node for one data object.
        * @param key DataObject
        */
        protected Node[] createNodes (Object key) {
            Node n = NodesRegistry.getNode (key);
            if (n == null) return new Node [0];
            return new Node [] {n};
        }

        /** Initializes the children.
        */
        protected void addNotify () {
            initialized = true;
            ThreadsProducer threadGroup = ((ThreadGroupNode) getNode ()).
                                         threadGroup;
            TGListener TGListener = ((ThreadGroupNode) getNode ()).TGListener;
            if (TGListener == null)
                TGListener = ((ThreadGroupNode) getNode ()).TGListener =
                     new TGListener (
                         (ThreadGroupNode) getNode ()
                     );
            threadGroup.addThreadListener (TGListener);
            update ();
        }

        /** Deinitializes the children.
        */
        protected void removeNotify () {
            ThreadsProducer threadGroup = ((ThreadGroupNode) getNode ()).
                                         threadGroup;
            TGListener TGListener = ((ThreadGroupNode) getNode ()).TGListener;
            threadGroup.removeThreadListener (TGListener);
            setKeys (java.util.Collections.EMPTY_SET);
            initialized = false;
        }

        void update () {
            if (!initialized) return;
            // add existing threads and groups ...
            ThreadsProducer threadGroup = ((ThreadGroupNode) getNode ()).
                                         threadGroup;
            AbstractThread[] groups = threadGroup.getThreads ();
            ArrayList l = new ArrayList ();
            l.addAll (Arrays.asList (groups));

            setKeys (l);
        }
    } // end of ThreadGroupContextChildren inner class

    /**
    * Serializable node reference. 
    */
    private static class ThreadsHandle implements Node.Handle {
        static final long serialVersionUID = -4518262935887219735L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            if (DebuggerModule.isRegisteredRootNodeClass (ThreadGroupNode.class))
                return DebuggerModule.getThreadGroupsRootNode ();
            return new AbstractNode (Children.LEAF);
        }
    }
}

