/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.lang.ref.WeakReference;

import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.Debugger;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.PropertySupport;
import org.openide.text.Line;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallableSystemAction;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.*;


/**
* This class representates threads and threadGroups as a Node.
* This class is final only for performance reasons, can be happily
* unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class ThreadNode extends AbstractNode implements 
SwitchOnCookie, GoToSourceCookie {

    /** generated Serialized Version UID */
    static final long serialVersionUID = 2202990736781011814L;


    // static ....................................................................

    /** Property names */
    public static final String        PROP_THREAD_NAME         = "threadName"; // NOI18N
    public static final String        PROP_THREAD_STATE        = "threadState"; // NOI18N
    public static final String        PROP_THREAD_CLASS        = "threadClass"; // NOI18N
    public static final String        PROP_THREAD_METHOD       = "threadMethod"; // NOI18N
    public static final String        PROP_THREAD_LINE         = "threadLine"; // NOI18N
    public static final String        PROP_THREAD_STACK_DEPTH  = "threadStackDepth"; // NOI18N
    public static final String        PROP_THREAD_FRAME_INDEX  = "threadFrameIndex"; // NOI18N
    public static final String        PROP_THREAD_SUSPENDED    = "threadSuspended"; // NOI18N

    public static final int           STATE_MAIN               = 1;
    public static final int           STATE_STOPPED            = 2;

    protected static String           ICON_CURRENT             =
        "/org/netbeans/modules/debugger/resources/threadCurrent"; // NOI18N
    protected static String           ICON_RUNNING             =
        "/org/netbeans/modules/debugger/resources/threadSuspended"; // NOI18N

    /** Icons for threads. */
    protected static Image            threadFolderIcon;
    protected static Image            threadFolderIcon32;
    protected static Image            threadCurrentIcon;
    protected static Image            threadCurrentIcon32;
    protected static Image            threadRunningIcon;
    protected static Image            threadRunningIcon32;

    /** System actions of this node */
    private static SystemAction[]     staticActions;
    private static ResourceBundle     bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (ThreadNode.class);
        return bundle.getString (s);
    }


    // variables .................................................................

    private Debugger                  debugger;
    private AbstractThread            thread;
    private ThreadListener            threadListener;
    private Node[]                    children = null;
    private CallStackNode             callStack;
    private LocalsRootNode            locals;
    private int                       state = 4;
    private boolean                   oldSuspended = false;


    // init ......................................................................

    /**
    * Creates ThreadNode.
    */
    public ThreadNode (AbstractThread thread) {
        super (new Children.Array());
        this.thread = thread;
        initialize();
    }

    /**
    * Remove listeners.
    */
    protected void finalize () {
        debugger.removePropertyChangeListener (threadListener);
        thread.removePropertyChangeListener (threadListener);
    }

    /** Performs basic initialization of children, listeners, etc. */
    private void initialize () {
        initializeChildren ();
        try {
            debugger = TopManager.getDefault ().getDebugger ();
        } catch (DebuggerException e) {
        }
        setName ();
        createProperties ();
        getCookieSet ().add (this);
        threadListener = new ThreadListener (this);
        thread.addPropertyChangeListener (threadListener);
    }

    /** Initializes children (subnodes) of this node */
    private void initializeChildren () {
        Children myChildren = getChildren ();
        if (thread instanceof VariablesProducer) {
            if (thread instanceof CallStackProducer)
                myChildren.add (
                    new Node[] {
                        callStack = new CallStackNode (thread),
                        locals = new LocalsRootNode (thread)
                    }
                );
            else
                myChildren.add (
                    new Node[] {
                        locals = new LocalsRootNode (thread)
                    }
                );
        } else
            if (thread instanceof CallStackProducer)
                myChildren.add (
                    new Node[] {
                        callStack = new CallStackNode (thread),
                    }
                );
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerThreadNode");        //NOI18N
    }


    // Node implementation .......................................................

    /** Method that prepares properties. Called from initialize.
    */
    private void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault();
        Sheet.Set props = sheet.get(Sheet.PROPERTIES);
        props.put (new PropertySupport.ReadOnly (
            PROP_THREAD_NAME,
            String.class,
            getLocalizedString ("PROP_thread_name"),
            getLocalizedString ("HINT_thread_name")
        ) {
            public Object getValue () throws InvocationTargetException {
                try {
                    return thread.getName ();
                } catch (java.lang.Exception e) {
                    throw new InvocationTargetException (e);
                }
            }
        });
        if (thread instanceof Location)
            props.put (new PropertySupport.ReadOnly (
                PROP_THREAD_LINE,
                Integer.TYPE,
                getLocalizedString ("PROP_thread_line"),
                getLocalizedString ("HINT_thread_line")
            ) {
                public Object getValue () throws InvocationTargetException {
                    try {
                        return new Integer (getLineNumber ());
                    } catch (java.lang.Exception e) {
                        throw new InvocationTargetException (e);
                    }
                }
            });
        if (thread instanceof CallStackProducer)
            props.put (new PropertySupport.ReadOnly (
                PROP_THREAD_STACK_DEPTH,
                Integer.TYPE,
                getLocalizedString ("PROP_thread_stack_depth"),
                getLocalizedString ("HINT_thread_stack_depth")
            ) {
                public Object getValue () throws InvocationTargetException {
                    try {
                        return new Integer (getStackDepth ());
                    } catch (java.lang.Exception e) {
                        throw new InvocationTargetException (e);
                    }
                }
            });
        // and set new sheet
        setSheet (sheet);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                SystemAction.get (SwitchOnAction.class),
                SystemAction.get (GoToSourceAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class)
            };
        return staticActions;
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (SwitchOnAction.class);
    }

    public void destroy () throws java.io.IOException {
        super.destroy ();
        debugger.removePropertyChangeListener (threadListener);
        thread.removePropertyChangeListener (threadListener);
    }


    // SwitchOnCookie ........................................................................................

    /**
    * Sets this thread to be suspended.
    */
    public boolean canSetCurrent () {
        return !thread.isCurrent ();
    }

    /**
    * Sets this thread to be current.
    */
    public void setCurrent () {
        thread.setCurrent (true);
    }


    // GoToSourceCookie implementation ...........................................

    /**
    * Goes to source.
    */
    public void goToSource () {
        Line l = ((Location) thread).getLine ();
        if (l != null) l.show (Line.SHOW_GOTO);
    }

    /**
    * Returns true if source is availabled.
    */
    public boolean canGoToSource () {
        Line l = ((Location) thread).getLine ();
        if (l != null) return true;
        return false;
    }

    public int getLineNumber () {
        if ( !(thread instanceof Location) ||
             (((Location) thread).getLine () == null)
        ) return -1;
        return ((Location) thread).getLine ().getLineNumber ();
    }
    

    // helper methods ............................................................

    int getStackDepth () {
        if (!(thread instanceof CallStackProducer)) return -1;
        return ((CallStackProducer) thread).getCallStack ().length;
    }
    
    /**
    * Sets curent icon (running thread, current thread, suspended thread).
    */
    void refreshIcon () {
        if ((state & STATE_MAIN) != 0) {
            setIconBase (ICON_CURRENT);
        } else
            setIconBase (ICON_RUNNING);
    }

    /** equals if thread equals o.thread
    * @param o
    */
    public final boolean equals (Object o) {
        return o instanceof ThreadNode && ((ThreadNode) o).thread.equals (thread);
    }

    /**
    * @return hashCode of thread
    */
    public final int hashCode () {
        return thread.hashCode ();
    }

    /**
    * Updates name & icon, and fires change.
    */
    void changeProperties () {
        setName ();
        firePropertyChange (null, null, null);
    }

    /**
    * Updates name & icon.
    */
    void setName () {
        // read thread state
        int newS = 0;
        if (thread.isCurrent ())
            newS = STATE_MAIN;
        
        // change icon
        if (state != newS) {
            state = newS;
            refreshIcon ();
        }

        // update name
        String s = thread.getName ();
        setDisplayName (s);
        setName (s);
    }

    /**
    * Refresh locals & callstack.
    */
    void changeChildren () {
        if (locals != null) locals.changeChildren ();
        if (callStack != null) callStack.changeChildren ();
    }

    AbstractThread getDebuggerThread () {
        return thread;
    }


    // innerclasses ..............................................................

    /**
    * Listens on thread and debugger.
    */
    private static class ThreadListener implements PropertyChangeListener {

        private WeakReference             node;
        private Debugger                  debugger;
        private AbstractThread            thread;


        ThreadListener (ThreadNode n) {
            node = new WeakReference (n);
            debugger = n.debugger;
            thread = n.thread;
        }

        public void propertyChange (PropertyChangeEvent e) {
            ThreadNode tn = (ThreadNode) node.get ();
            if (tn == null) {
                debugger.removePropertyChangeListener (this);
                thread.removePropertyChangeListener (this);
                return;
            }
            tn.changeProperties ();
        }
    }
}

