/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.IOException;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import java.util.Arrays;

import org.openide.actions.DeleteAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.debugger.*;
import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.nodes.*;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.*;
//import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.*;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* This class represents variable as a Node.
*
* @author   Jan Jancura
*/
public class VariableNode extends AbstractNode {

    /** generated Serialized Version UID */
    static final long                       serialVersionUID = -4167730461290769518L;

    protected static String                 ICON_BASE =
        "/org/netbeans/core/resources/variable"; // NOI18N
    protected static String                 ICON_STATIC_VARIABLE =
        "/org/netbeans/modules/debugger/resources/staticVariable"; // NOI18N

    public static final String              PROP_MODIFIERS = "modifiers"; // NOI18N
    public static final String              PROP_INNER_TYPE = "innerType"; // NOI18N
    public static final String              PROP_BASE_INDEX = "baseIndex"; // NOI18N
    public static final String              PROP_DISPLAYED_LENGTH = "displayedLength"; // NOI18N
    private static ResourceBundle           bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (VariableNode.class);
        return bundle.getString (s);
    }

    // variables .................................................................

    protected boolean                       isRoot = false;
    protected AbstractVariable              variable = null;
    protected transient Debugger            debugger;
    private PropertyChangeListener          pcl;
    
    // init ......................................................................

    /**
    * Creates empty Variable context.
    */
    public VariableNode (
        AbstractVariable variable,
        boolean isRoot
    ) {
        super (
            variable instanceof VariablesProducer ?
                ((VariablesProducer) variable).isLeaf () ?
                    Children.LEAF :
                    new VariableContextChildren () :
                Children.LEAF
        );
        this.isRoot = isRoot;
        this.variable = variable;
        init ();
    }

    /**
    * Creates empty Variable context.
    */
    public VariableNode (
        AbstractVariable variable
    ) {
        this (variable, false);
    }

    protected void init () {
        variable.addPropertyChangeListener (pcl = new PropertyChangeListener () {
            public void propertyChange (PropertyChangeEvent e) {
                parameterChanged (e);
            }
        });
        setDisplayName (getName ());
        setName (getName ());
        setIconBase (ICON_BASE);
        // obtain debugger reference
        try {
            debugger = TopManager.getDefault ().getDebugger ();
        } catch (DebuggerException exc) {
            TopManager.getDefault ().notifyException (exc);
        }
        createProperties ();
        changeChildren ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerVariableNode");        //NOI18N
    }

    /** deserializes object */
    private void readObject (java.io.ObjectInputStream obis)
    throws java.io.IOException, ClassNotFoundException,
        java.io.NotActiveException {
        obis.defaultReadObject ();
        init ();
    }


    // implementation of Node ....................................................

    /** Creates properties for this node */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        ps.put (new PropertySupport.ReadOnly (
            Watch.PROP_VARIABLE_NAME,
            String.class,
            getLocalizedString ("PROP_watch_name"),
            getLocalizedString ("HINT_watch_name")
        ) {
            public Object getValue () {
                return variable.getVariableName ();
            }
        });
        createCommonProperties (ps);
        // and set new sheet
        setSheet (sheet);
    }

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (isRoot)
            return new SystemAction[] {
                SystemAction.get (DeleteAction.class),
                null,
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class),
            };
        else
            return new SystemAction[] {
                SystemAction.get (ToolsAction.class),
                SystemAction.get (PropertiesAction.class),
            };
    }

    /** Helper method, creates properties common to variable
    * and watch contexts. */
    void createCommonProperties (
        final Sheet.Set ps
    ) {
        ps.put (new PropertySupport.ReadOnly (
            Watch.PROP_TYPE,
            String.class,
            getLocalizedString ("PROP_watch_type"),
            getLocalizedString ("HINT_watch_type")
        ) {
            public Object getValue () {
                return variable.getType ();
            }
        });
        ps.put (new PropertySupport.ReadWrite (
            Watch.PROP_AS_TEXT,
            String.class,
            getLocalizedString ("PROP_watch_value"),
            getLocalizedString ("HINT_watch_value")
        ) {
            public Object getValue () {
                return variable.getAsText ();
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    variable.setAsText ((String)val);
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
                catch (DebuggerException e) {}
            }
        });
    }

    /**
    * Variable can be removed.
    *
    * @return <CODE>true</CODE>
    */
    public boolean canDestroy () {
        return isRoot;
    }

    /**
    * Removes the Variable from its parent and deletes it.
    */
    public void destroy () throws IOException {
        if (variable == null) return;
        variable.removePropertyChangeListener (pcl);
        variable = null;
        if (getChildren () == Children.LEAF) return;
        VariableContextChildren myChildren = (VariableContextChildren)
                                             getChildren ();
        if (!myChildren.initialized) return;
        Node[] n = myChildren.getNodes ();
        int i, k = n.length;
        for (i = 0; i < k; i++)
            ((VariableNode) n [i]).destroy ();
        super.destroy ();
    }


    // GetVariableCookie implementation ..........................................

    /**
    * Returns variable instance.
    */
    public AbstractVariable getVariable () {
        return variable;
    }


    // other methods .............................................................

    /**
    * Returns human presentable name of this variable containing
    * informations about value or valide.
    *
    * @return human presentable name of this variable.
    */
    public String getName () {
        String name = variable.getVariableName ();
        if ((name == null) || (debugger == null)) return "???"; // NOI18N

        if (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING)
            return name + " " + getLocalizedString ("CTL_NoSession"); // NOI18N
        String type = variable.getType ();
        String value = variable.getAsText ();
        if (value == null) 
            return name + " " + getLocalizedString ("CTL_NotInitialized"); // NOI18N
        if (type == null)
            return name + " = " + value; // NOI18N
        return name + " = (" + type + ") " + value; // NOI18N
    }

    /**
    * Sets display name and fires property change.
    */
    protected void parameterChanged (PropertyChangeEvent e) {
        String s = getName ();
        setDisplayName (s);
        setName (s);
        changeChildren ();
        firePropertyChange (
            e.getPropertyName (),
            e.getOldValue (),
            e.getNewValue ()
        );
    }

    /**
    * Sets current variable fields.
    */
    protected void changeChildren () {
        // set fields of this variable as keys into our children
        if (getChildren() == Children.LEAF) {
            Node n = getParentNode ();
            if (n == null)
                return;
            AbstractVariable v = getVariable ();
            if (!((VariablesProducer) v).isLeaf ()) {
                LeafRefresher parentChildren =
                    (LeafRefresher) n.getChildren ();
                parentChildren.refreshMyKey (v);
            }
            return;
        }
        VariableContextChildren myChildren =
            (VariableContextChildren) getChildren ();
        if (!myChildren.initialized) return;
        
        AbstractVariable [] fields;
        fields = ((VariablesProducer) variable).getFilteredVariables ();
        Arrays.sort (fields, Utils.variablesComparator);
        
        myChildren.setMyKeys (fields);
    }

    // innerclasses ..............................................................

    /** Special subnodes (children) for variable node */
    private static final class VariableContextChildren extends Children.Keys implements LeafRefresher {

        /** Where to add variables. */
        private boolean                         initialized = false;

        /** Initializes children.
        */
        protected void addNotify () {
            initialized = true;
            ((VariableNode) getNode ()).changeChildren ();
        }

        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (Object key) {
            return new Node[] {
                new VariableNode ((AbstractVariable)key)
            };
        }

        /** Accessor for VariableNode outer class */
        private void setMyKeys (Object[] keys) {
            setKeys(keys);
        }

        // interface LeafRefresher

        /** Accessor for VariableNode class */
        public void refreshMyKey (Object key) {
            refreshKey (key);
        }

    } // end of VariableContextChildren inner class
}

