/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.util.ResourceBundle;
import java.util.HashMap;
import java.util.Arrays;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.lang.ref.WeakReference;

import org.openide.TopManager;
import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.debugger.*;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.DebuggerModule;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* This class represents local, static and instance variables as a Node.
* This class is final only for performance reasons.
* Can be unfinaled if desired.
*
* @author   Daniel Prusa
*/
public final class VariablesRootNode extends AbstractNode implements 
Validator.Object {


    // static variables ........................................................

    /** generated Serialized Version UID */
    static final long                  serialVersionUID = -5432455019958452231L;

    private static String              ICON_BASE =
        "/org/netbeans/modules/debugger/resources/variables"; // NOI18N
    private static String              ICON_PENDING_VARIABLES =
        "/org/netbeans/modules/debugger/resources/pendingVariables"; // NOI18N
    private static String              ICON_LOCAL_VARIABLE =
        "/org/netbeans/modules/debugger/resources/localVariable"; // NOI18N

    private static ResourceBundle      bundle;

    private static HashMap             locals = new HashMap ();

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (VariablesRootNode.class);
        return bundle.getString (s);
    }

    
    // init ....................................................................

    /**
    * Creates empty VariablesContext.
    */
    public VariablesRootNode () {
        super (new VariablesRootChildren ());
        String s = getLocalizedString ("CTL_Variables_root");
        setDisplayName (s);
        setName (s);
        setIconBase (ICON_BASE);
        changeChildren ();
        init ();
    }

    private void init () {
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerVariablesRootNode");        //NOI18N
    }


    // helper methods ..........................................................

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new VariablesHandle ();
    }

    void changeChildren () {
        HashMap newLocals = new HashMap ();
        VariablesRootChildren myChildren = (VariablesRootChildren) getChildren ();
        try {
            AbstractVariable [] variables = new AbstractVariable [] {};   // local variables
            
            ThreadsRoot debugger = (ThreadsRoot) TopManager.getDefault ().getDebugger ();
            AbstractThread thread = debugger.getCurrentThread ();
            if (thread != null) {
                if (thread instanceof VariablesProducer) {
                    variables = ((VariablesProducer) thread).getFilteredVariables ();
                    Arrays.sort (variables, Utils.localsComparator);
                }
            }
            myChildren.setMyKeys (variables);
        }
        catch (DebuggerException e) {
        }
    }

    // interface Validator.Object ..............................................

    public void validate () {
        changeChildren ();
    }
    
    public boolean canRemove () {
        return false;
    }

    // innerclasses ............................................................

    /** Special Variables subnodes (children) */
    private static final class VariablesRootChildren extends Children.Keys implements LeafRefresher {

        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (final Object key) {
            AbstractNode node = (AbstractNode) NodesRegistry.getNode (key);
            if (locals.containsKey (key))
                node.setIconBase (ICON_LOCAL_VARIABLE);
            return new Node[] {node};
        }

        protected void addNotify () {
            try {
                AbstractDebugger debugger =
                    (AbstractDebugger) TopManager.getDefault ().getDebugger ();
                VRListener l = new VRListener (getNode (), debugger);
                Validator validator = ((ValidatorHolder) debugger).getValidator ();
                validator.add ((VariablesRootNode) getNode ());
            } catch (DebuggerException e) {
            }
        }
        
        /** Accessor for LocalsRootNode outer class */
        private void setMyKeys (final Object[] keys) {
            setKeys (keys);
        }

        // interface LeafRefresher .....................................................................

        /** Accessor for VariableNode class */
        public void refreshMyKey (Object key) {
            refreshKey (key);
        }

    } // end of VariablesRootChildren inner class

    /**
    * Serializable node reference. 
    */
    private static class VariablesHandle implements Node.Handle {
        static final long serialVersionUID =-45182213487259891L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return DebuggerModule.getVariablesRootNode ();
        }
    }
    
    static class VRListener implements PropertyChangeListener {
        
        private Debugger                debugger;
        private WeakReference           node;
        protected static String         stdIcon = ICON_BASE;
        protected static String         pendingIcon = ICON_PENDING_VARIABLES;
        
        
        VRListener (Node node, Debugger debugger) {
            this.debugger = debugger;
            this.node = new WeakReference (node);
            Validator validator = ((ValidatorHolder) debugger).getValidator ();
            validator.addPropertyChangeListener (this);
        }

        VariablesRootNode getNode () {
            VariablesRootNode wrn = (VariablesRootNode) node.get ();
            if (wrn != null) return wrn;
            Validator validator = ((ValidatorHolder) debugger).getValidator ();
            validator.removePropertyChangeListener (this);
            return null;
        }
        
        public void propertyChange (PropertyChangeEvent e) {
            VariablesRootNode node = getNode ();
            if (node == null) return;
            if (e.getSource () instanceof Validator) {
                // update "pending" icon 
                node.setIconBase (
                    ((Validator) e.getSource ()).isValidated () ?
                    ICON_BASE :
                    ICON_PENDING_VARIABLES
                );
                /*
                setDisplayName (
                  validator.isValidated () ?
                  getLocalizedString ("CTL_Watches_group_root") :
                  getLocalizedString ("CTL_Watches_group_root_refresh")
                );
                */  

                // PATCH, we must force the view to be repainted
                Node [] nodes = node.getChildren ().getNodes ();
                Node n = null;
                String displayName = null;
                if ((nodes != null) && (nodes.length > 0)) {
                    n = nodes [0];
                    displayName = n.getDisplayName ();
                    n.setDisplayName (displayName.concat (" ")); // NOI18N
                }
                DebuggerView view = DebuggerModule.getViewFor (node);
                if (view != null) {
                    view.repaint (70);
                }
                if (n != null)
                    n.setDisplayName (displayName);
                // end of PATCH
                return;
            }
        }
    }// VRListener
}
