/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.IOException;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import javax.swing.JPopupMenu;

import org.openide.actions.DeleteAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.TopManager;
import org.openide.debugger.*;
import org.openide.nodes.BeanNode;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.AbstractWatch;
import org.netbeans.modules.debugger.support.actions.*;


/**
* This class represents watch as a Node.
* This class is final only for performance reasons.
* Can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class WatchNode extends VariableNode {
    /** generated Serialized Version UID */
    static final long               serialVersionUID = 7100337776961302436L;

    protected static String         ICON_BASE =
        "/org/netbeans/core/resources/watch"; // NOI18N
    private static ResourceBundle   bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (WatchNode.class);
        return bundle.getString (s);
    }
    

    // variables ...............................................................

//    protected AbstractWatch         watch;


    // init ....................................................................

    /**
    * Creates empty Watch context.
    */
    public WatchNode (
        final AbstractWatch watch
    ) {
        super (watch, true);
        if (watch == null) throw new NullPointerException ();
        setIconBase (ICON_BASE);
  //      this.watch = watch;
    }

    protected void init () {
//        this.watch = (AbstractWatch) variable;
        super.init ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerWatchNode");        //NOI18N
    }


    // implementation of Node ....................................................

    /**
    * Variable can be removed.
    *
    * @return <CODE>true</CODE>
    */
    public boolean canDestroy () {
        return true;
    }

    /**
    * Removes the Variable from its parent and deletes it.
    */
    public void destroy () throws IOException {
        if (variable == null) return;
        ((Watch) variable).remove ();
        super.destroy ();
    }

    /** Creates properties for watch context
    * (remember that watch extends variable context,
    * so we must pay an attention to its properties too)
    * Overrides createProperties from Variable context.
    */
    protected void createProperties () {
        // default sheet with "properties" property set // NOI18N
        Sheet sheet = Sheet.createDefault ();
        Sheet.Set ps = sheet.get (Sheet.PROPERTIES);
        // and add rw property of the same name
        ps.put(new PropertySupport.ReadWrite (
            AbstractWatch.PROP_VARIABLE_NAME,
            String.class,
            getLocalizedString ("PROP_watch_name"),
            getLocalizedString ("HINT_watch_name")
        ) {
            public Object getValue () {
                return variable.getVariableName ();
            }
            public void setValue (Object val) throws IllegalArgumentException {
                try {
                    ((Watch) variable).setVariableName (((String)val).trim ());
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException ();
                }
            }
        });
        // add type and value props
        createCommonProperties (ps);
        // and set new sheet
        setSheet (sheet);
    }


    // private methods ...........................................................

    /**
    * Returns watch.
    */
    AbstractWatch getWatch () {
        return ((AbstractWatch) variable);
    }

    /**
    * Returns human presentable name of this watch containing
    * informations about value or valide.
    *
    * @return human presentable name of this variable.
    */
    public String getName () {
        if (variable == null) return "?"; // NOI18N // should not occure
        String name = variable.getVariableName ();
        if (name == null) return "???"; // NOI18N

        if ( (debugger == null) ||
                (debugger.getState () == Debugger.DEBUGGER_NOT_RUNNING)
           ) return name + " = " + getLocalizedString ("CTL_NoSession");

        String errorMessage = ((AbstractWatch) variable).getErrorMessage ();
        if (errorMessage != null)
            return name + " = >" + errorMessage + "<"; // NOI18N
        String type = variable.getType ();
        String value = variable.getAsText ();
        return name + " = (" + type + ") " + value; // NOI18N
    }
}
