/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.nodes;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;
import javax.swing.SwingUtilities;
import java.lang.ref.WeakReference;

import org.openide.TopManager;
import org.openide.debugger.*;
import org.openide.actions.AddWatchAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.ToolsAction;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallableSystemAction;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.nodes.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.DebuggerModule;
import org.netbeans.modules.debugger.support.actions.DeleteAllCookie;
import org.netbeans.modules.debugger.support.actions.DeleteAllAction;
import org.netbeans.modules.debugger.support.util.*;

import org.netbeans.modules.debugger.delegator.DelegatingWatch;//[PENDING?]


/**
* This class representates watches root as a Node.
* This class is final only for performance reasons,
* can be happily unfinaled if desired.
*
* @author   Jan Jancura
*/
public final class WatchesRootNode extends AbstractNode implements
DeleteAllCookie {

    // static variables ........................................................

    /** generated Serialized Version UID */
    static final long               serialVersionUID = -3518528079657369038L;

    private static final String     ICON_WATCHES =
        "/org/netbeans/core/resources/watches"; // NOI18N
    private static final String     ICON_PENDING_WATCHES =
        "/org/netbeans/modules/debugger/resources/pendingWatches"; // NOI18N

    /** Popup menu actions. */
    private static SystemAction[]   staticActions;

    private static ResourceBundle   bundle;

    /**
    * @return localized string.
    */
    static String getLocalizedString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (WatchesRootNode.class);
        return bundle.getString (s);
    }


    // init ....................................................................
    /**
    * Creates empty WatchesRootNode.
    */
    public WatchesRootNode () {
        super (new WatchesRootChildren ());
        String name = getLocalizedString ("CTL_Watches_group_root");
        setDisplayName (name);
        setName (name);
        setIconBase (ICON_WATCHES);
        getCookieSet ().add (this);
        init ();
    }

    private void init () {
    }

    /** deserializes object */
    private void readObject(java.io.ObjectInputStream obis)
    throws java.io.IOException, ClassNotFoundException,
        java.io.NotActiveException {
        obis.defaultReadObject ();
        init ();
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("NetbeansDebuggerWatchesRootNode");        //NOI18N
    }


    // Node implementation .....................................................

    /** Getter for set of actions that should be present in the
    * popup menu of this node. This set is used in construction of
    * menu returned from getContextMenu and specially when a menu for
    * more nodes is constructed.
    *
    * @return array of system actions that should be in popup menu
    */
    public SystemAction[] getActions () {
        if (staticActions == null)
            staticActions = new SystemAction[] {
                SystemAction.get(AddWatchAction.class),
                null,
                SystemAction.get (DeleteAllAction.class),
                null,
                SystemAction.get(ToolsAction.class),
                SystemAction.get(PropertiesAction.class),
            };
        return staticActions;
    }

    /** Obtain handle for this node (for serialization).
    * The handle can be serialized and {@link Handle#getNode} used after
    * deserialization to obtain the original node.
    *
    * @return the handle, or <code>null</code> if this node is not persistable
    */
    public Node.Handle getHandle () {
        return new WatchesHandle ();
    }

    /**
    * Returns default action.
    */
    public SystemAction getDefaultAction () {
        return SystemAction.get (AddWatchAction.class);
    }


    // DeleteAllCookie implementation ..........................................

    /**
    * Delete all breakpoints.
    */
    public void deleteAll () {
        try {
            // obtain reference to the debugger instance
            Debugger debugger = TopManager.getDefault ().getDebugger ();
            debugger.removeAllWatches ();
            WatchesRootChildren children = (WatchesRootChildren) getChildren ();
            children.removeAllVariables ();
        } catch (DebuggerException exc) {
        }
    }


    // other methods ...........................................................

    public void createVariable (AbstractVariable variable) {
        ((WatchesRootChildren) getChildren ()).addVariable (
            variable
        );
    }


    // innerclasses ............................................................

    /** Special watches context subnodes (children) */
    static final class WatchesRootChildren extends Children.Keys {

        private ArrayList variables = new ArrayList ();
        
        
        /** Creates nodes for given key.
        * @param key the key that is used
        * @return array of nodes representing the key
        */
        protected Node[] createNodes (Object key) {
            if ((key instanceof Watch) && ((Watch) key).isHidden ())
                return new Node [0];
            if (key instanceof DelegatingWatch) {
                Watch w = ((DelegatingWatch) key).getInnerWatch ();
                if (w != null) {
                    Node node = DebuggerNode.createNode (w);
                    if (node != null)
                        return new Node[] {
                            new DelegatingNode (node, (DelegatingWatch) key)
                        };
                }
            }
            Node node = DebuggerNode.createNode (key);
            if (node == null)
                return new Node [0];
            return new Node[] {node};
        }
        
        protected void addNotify () {
            try {
                Debugger debugger =
                    (Debugger) TopManager.getDefault ().getDebugger ();
                WRListener l = new WRListener ((WatchesRootNode) getNode (), debugger);
            } catch (DebuggerException e) {
            }
        }
        
        void addVariable (AbstractVariable v) {
            Node n = DebuggerNode.createNode (v);
            if (n != null) {
                variables.add (n);
                add (new Node [] {n});
            }
        }
        
        void refreshHelper (Watch w) {
            refreshKey (w);
        }
        
        void set (Object [] ws) {
            try {
            setKeys (ws);
            } catch (Exception e) {
                e.printStackTrace();//PENDING!!!
                
            }
        }
        
        void removeAllVariables () {
            if (variables.size () < 1) return;
            Node[] nn = new Node [variables.size ()];
            nn = (Node[]) variables.toArray (nn);
            remove (nn);
        }
    } // end of WatchesRootNodeChildren inner class

    /**
    * Serializable node reference. 
    */
    private static class WatchesHandle implements Node.Handle {
        static final long serialVersionUID =-4518262478987259653L;

        /** Reconstitute the node for this handle.
        *
        * @return the node for this handle
        * @exception IOException if the node cannot be created
        */
        public Node getNode () {
            return DebuggerModule.getWatchesRootNode ();
        }
    }
    
    static class WRListener implements PropertyChangeListener {
        
        private Debugger                debugger;
        private WeakReference           node;
        private Watch[]                 watches = new Watch [0];
        
        
        WRListener (WatchesRootNode node, Debugger debugger) {
            this.debugger = debugger;
            this.node = new WeakReference (node);
            
            // add listeners & set first children
            debugger.addPropertyChangeListener (this);
            if (debugger instanceof ValidatorHolder) {
                Validator validator = ((ValidatorHolder) debugger).getValidator ();
                validator.addPropertyChangeListener (this);
            }
            setWatches (node);
        }

        WatchesRootNode getNode () {
            WatchesRootNode wrn = (WatchesRootNode) node.get ();
            if (wrn != null) return wrn;
            
            // remove all listeners!
            debugger.removePropertyChangeListener (this);
            if (debugger instanceof ValidatorHolder) {
                Validator validator = ((ValidatorHolder) debugger).getValidator ();
                validator.removePropertyChangeListener (this);
            }
            return null;
        }
        
        public void propertyChange (PropertyChangeEvent e) {
            WatchesRootNode node = getNode ();
            if (node == null) return;
            if (e.getSource () instanceof Validator) {
                // update "pending" icon 
                node.setIconBase (
                    ((Validator) e.getSource ()).isValidated () ?
                    ICON_WATCHES :
                    ICON_PENDING_WATCHES
                );
                /*
                setDisplayName (
                  validator.isValidated () ?
                  getLocalizedString ("CTL_Watches_group_root") :
                  getLocalizedString ("CTL_Watches_group_root_refresh")
                );
                */  

                // PATCH, we must force the view to be repainted
                Node [] nodes = node.getChildren ().getNodes ();
                Node n = null;
                String displayName = null;
                if ((nodes != null) && (nodes.length > 0)) {
                    n = nodes [0];
                    if (n != null) {
                        displayName = n.getDisplayName ();
                        n.setDisplayName (displayName.concat (" ")); // NOI18N
                    }
                }
                DebuggerView view = DebuggerModule.getViewFor (node);
                if (view != null) {
                    view.repaint (70);
                }
                if (n != null)
                    n.setDisplayName (displayName);
                // end of PATCH
                return;
            }
            
            if (e.getPropertyName () == null) return;

            if ( (e.getPropertyName ().equals (Debugger.PROP_STATE) && 
                 debugger.getState () == debugger.DEBUGGER_NOT_RUNNING)
            ) {
                // remove all variables
                ((WatchesRootChildren) node.getChildren ()).
                    removeAllVariables ();
            }
            
            if (e.getPropertyName ().equals (Debugger.PROP_WATCHES)
            ) setWatches (node);
            else
            if (e.getPropertyName ().equals (WatchDelegate.PROP_INNER_WATCH)
            ) refreshWatches (node, (Watch) e.getSource ());
        }
        
        void refreshWatches (WatchesRootNode node, Watch w) {
            ((WatchesRootChildren) node.getChildren ()).refreshHelper (w);
        }
        
        void setWatches (WatchesRootNode node) {
            int i, k = watches.length;
            for (i = 0; i < k; i++)
                watches [i].removePropertyChangeListener (this);
            watches = debugger.getWatches ();
            ((WatchesRootChildren) node.getChildren ()).set (watches);
            k = watches.length;
            for (i = 0; i < k; i++)
                watches [i].addPropertyChangeListener (this);
        }
    }// WRListener
    
    static class DelegatingNode extends FilterNode {
        DelegatingWatch dw;
        private DelegatingNode (Node n, DelegatingWatch dw) {
            super (n);
            this.dw = dw;
        }
        
        public void destroy () {
            dw.remove ();
        }
    }
}

