/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.util;

import java.util.Vector;

import org.openide.debugger.*;
import org.openide.text.Line;
import org.openide.src.ConstructorElement;

import org.netbeans.modules.debugger.CoreBreakpoint;
import org.netbeans.modules.debugger.support.BreakpointSupport;


/**
* Helper methods for debugger.
*
* @author  Jan Jancura
*/
public class Find {

    
    public static Watch watch (
        Debugger d, 
        String watch
    ) {
        Watch[] ws = d.getWatches ();
        int i, k = ws.length;
        for (i = 0; i < k; i++)
            if (watch.equals (ws [i].getVariableName ()))
                return ws [i];
        return null;
    }
    
    /** Find an breakpoint assigned to a specific line.
    * The line is represented by a line object that can change its
    * position as the text is modified.
    *
    * @param l line to find breakpoint at
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    * @return the breakpoint or null
    */
    public static CoreBreakpoint[] breakpoint (
        Debugger d, 
        Line l, 
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        Breakpoint[] b = d.getBreakpoints ();
        Vector br = new Vector ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            if ( (b [i].getLine () != null) && 
                 b [i].getLine ().equals (l) && 
                 ( (enabled & b [i].isEnabled ()) ||
                   (disabled & !b [i].isEnabled ())) &&
                 ( (hidden & b [i].isHidden ()) ||
                   (visible & !b [i].isHidden ()))
            ) br.add (b [i]);
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }

    /** Find an enabled breakpoint assigned to a method.
    * The method is represented by a method source element that
    * must have a declaring class.
    *
    * @param method method with {@link ConstructorElement#getDeclaringClass valid} declaring class
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    * @return the breakpoint or null
    * @exception IllegalArgumentException if the method does not have a declaring class
    */
    public static CoreBreakpoint[] breakpoint (
        Debugger d,
        ConstructorElement method,
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        CoreBreakpoint[] b = (CoreBreakpoint[]) d.getBreakpoints ();
        Vector br = new Vector ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            if ( ( ( (b [i].getMethod () != null) &&
                     b [i].getMethod ().equals (method)) ||
                   ( (b [i] instanceof BreakpointSupport) &&
                     (((BreakpointSupport) b [i]).getMethodName () != null) &&
                     ((BreakpointSupport) b [i]).getMethodName ().equals (method.getName ().getName ()))
                 ) &&
                 ( (enabled & b [i].isEnabled ()) ||
                   (disabled & !b [i].isEnabled ())) &&
                 ( (hidden & b [i].isHidden ()) ||
                   (visible & !b [i].isHidden ()))
            ) br.add (b [i]);
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }

    /**
    * Find an enabled breakpoint assigned to a specific line number.
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    */
    public static CoreBreakpoint[] breakpoint (
        Debugger d, 
        String className, 
        int lineNumber, 
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        String cn = Utils.getTopClassName (className);
        Breakpoint[] b = d.getBreakpoints ();
        Vector br = new Vector ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            if ( (b [i].getClassName () != null) && 
                 (Utils.getTopClassName (b [i].getClassName ()).equals (cn))) {
                    Line line = b[i].getLine();
                    if ((line != null) &&
                        ((line.getLineNumber () + 1) == lineNumber) &&
                        ((enabled & b[i].isEnabled()) || (disabled & !b[i].isEnabled())) &&
                        ((hidden & b[i].isHidden()) || (visible & !b[i].isHidden ()))) {
                            br.add (b[i]);
                    }
            }
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }

    /**
    * Find an enabled breakpoint assigned to a specific method.
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    */
    public static CoreBreakpoint[] breakpoint (
        Debugger d, 
        String className, 
        String methodName, 
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        String cn = Utils.getTopClassName (className);
        Breakpoint[] b = (Breakpoint[]) d.getBreakpoints ();
        Vector br = new Vector ();
        int i, k = b.length;
        for (i = 0; i < k; i++) {
            if ( (b [i].getClassName () != null) && 
                 (Utils.getTopClassName (b [i].getClassName ()).equals (cn)) &&
                 (b [i].getMethod () != null) && 
                 (b [i].getMethod ().getName () != null) && 
                 (b [i].getMethod ().getName ().equals (methodName)) &&
                 ( (enabled & b [i].isEnabled ()) ||
                   (disabled & !b [i].isEnabled ())) &&
                 ( (hidden & b [i].isHidden ()) ||
                   (visible & !b [i].isHidden ()))
            ) br.add (b [i]);
        }
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }

    /**
    * Finds all enabled breakpoints according to className
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    */
    public static CoreBreakpoint[] breakpoints (
        Debugger d, 
        String className, 
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        String cn = Utils.getTopClassName (className);
        Vector br = new Vector ();
        Breakpoint[] b = (Breakpoint[]) d.getBreakpoints ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            if ( (b [i].getClassName () != null) && 
                 (Utils.getTopClassName (b [i].getClassName ()).equals (cn)) &&
                 ( (enabled & b [i].isEnabled ()) ||
                   (disabled & !b [i].isEnabled ())) &&
                 ( (hidden & b [i].isHidden ()) ||
                   (visible & !b [i].isHidden ()))
            ) br.add (b [i]);
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }

    /**
    * Finds all enabled breakpoints according to className
    * @param endbled set true to find enabled breakpoints
    * @param disabled set true to find disabled breakpoints
    * @param visible set true to find visible breakpoints
    * @param hidden set true to find hidden breakpoints
    */
    public static CoreBreakpoint[] breakpoints (
        Debugger d, 
        boolean enabled,
        boolean disabled,
        boolean visible,
        boolean hidden
    ) {
        Vector br = new Vector ();
        Breakpoint[] b = (Breakpoint[]) d.getBreakpoints ();
        int i, k = b.length;
        for (i = 0; i < k; i++)
            if ( ( (enabled & b [i].isEnabled ()) ||
                   (disabled & !b [i].isEnabled ())) &&
                 ( (hidden & b [i].isHidden ()) ||
                   (visible & !b [i].isHidden ()))
            ) br.add (b [i]);
        CoreBreakpoint[] res = new CoreBreakpoint [br.size ()];
        br.copyInto (res);
        return res;
    }
}
