/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.support.util;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.loaders.DataObject;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystemCapability;
import org.openide.text.Line;
import org.openide.text.NbDocument;
import org.openide.nodes.Node;
import org.openide.loaders.DataNode;
import org.openide.src.*;
import org.openide.cookies.LineCookie;
import org.openide.cookies.EditorCookie;
import org.openide.execution.NbClassPath;
import org.openide.execution.NbProcessDescriptor;
import org.openide.debugger.DebuggerInfo;
import org.openide.util.MapFormat;
import org.openide.util.Utilities;

import java.util.*;
import java.io.*;
import java.util.ResourceBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.JavaVariable;


/**
* Helper methods for debugging.
*
* @author  Jan Jancura
*/
public class Utils {

    static ResourceBundle                     bundle = org.openide.util.NbBundle.getBundle
            (org.netbeans.modules.debugger.support.DebuggerSupport.class);

    static boolean hotSpotDetected;

    /**
    * Returns localized Exception.
    */
    public static RuntimeException localizeException (Exception e, String text) {
        return new ExceptionHack (e, text);
    }

    /**
    * Removes leading and ending quote character, if there is any
    */
    private static String removeQuotationMarks (String s) {
        if (s == null)
            return null;
        if (s.startsWith ("\"")) // NOI18N
            s = s.substring (1, s.length ());
        if (s.endsWith("\"")) // NOI18N
            s = s.substring (0, s.length () - 1);
        return s;
    }

    /**
    * Creates map of arguments for NbProcessDescriptor.
    */
    public static HashMap processDebuggerInfo (
        DebuggerInfo   info,
        String         debuggerOptions,
        String         main
    ) {
        // create starting string
        String repository = removeQuotationMarks (NbClassPath.createRepositoryPath (FileSystemCapability.DEBUG).getClassPath ());
        String library = removeQuotationMarks (NbClassPath.createLibraryPath ().getClassPath ());
        String classPath = removeQuotationMarks (NbClassPath.createClassPath ().getClassPath ());
        String bootClassPath = removeQuotationMarks (NbClassPath.createBootClassPath ().getClassPath ());
        String pathSeparator = System.getProperty ("path.separator"); // NOI18N
        String javaHome = System.getProperty ("java.home"); // NOI18N
        String jdkHome = System.getProperty ("jdk.home"); // NOI18N
        boolean classic = false;
        if (info instanceof ProcessDebuggerInfo) {
            ProcessDebuggerInfo processInfo = (ProcessDebuggerInfo)info;
            if (processInfo.getClassPath () != null)
                classPath = removeQuotationMarks (processInfo.getClassPath ());
            if (processInfo.getBootClassPath () != null)
                bootClassPath = removeQuotationMarks (processInfo.getBootClassPath ());
            if (processInfo.getRepositoryPath () != null)
                repository = removeQuotationMarks (processInfo.getRepositoryPath ());
            if (processInfo.getLibraryPath () != null)
                library = removeQuotationMarks (processInfo.getLibraryPath ());
            classic = processInfo.isClassic ();
        }
        //    if (System.getProperty ("java.vm.name", "?").indexOf ("HotSpot") >= 0)
        //      classic = true;
        HashMap map = new HashMap ();
        map.put (ProcessDebuggerType.CLASSIC_SWITCH, classic ? "-classic " : ""); // NOI18N
        map.put (ProcessDebuggerType.DEBUGGER_OPTIONS, debuggerOptions);
        map.put (ProcessDebuggerType.REPOSITORY_SWITCH, repository);
        map.put (ProcessDebuggerType.LIBRARY_SWITCH,
                 ((library.length () > 0) && (repository.length () > 0)) ? pathSeparator + library : library
                );
        map.put (ProcessDebuggerType.CLASS_PATH_SWITCH,
                 ((classPath.length () > 0) && ((library.length () + repository.length ()) > 0)) ? pathSeparator + classPath : classPath
                );
        map.put (ProcessDebuggerType.BOOT_CLASS_PATH_SWITCH, bootClassPath);
        map.put (ProcessDebuggerType.BOOT_CLASS_PATH_SWITCH_SWITCH, (bootClassPath.length () < 1) ? "" : "-Xbootclasspath:"); // NOI18N
        map.put (ProcessDebuggerType.MAIN_SWITCH, main);
        map.put (ProcessDebuggerType.QUOTE_SWITCH, "\""); // NOI18N
        map.put (ProcessDebuggerType.JAVA_HOME_SWITCH, javaHome);
        map.put (ProcessDebuggerType.JDK_HOME_SWITCH, jdkHome);
        map.put (ProcessDebuggerType.FILE_SEPARATOR_SWITCH, System.getProperty ("file.separator")); // NOI18N
        map.put (ProcessDebuggerType.PATH_SEPARATOR_SWITCH, pathSeparator);
        return map;
    }

    /**
    * Returns true if HotSpot is installed for given JVM.
    */
    public static boolean hasHotSpot (String path) {
        try {
            String sub = path.substring (0, path.lastIndexOf (File.separatorChar));
            sub = sub.substring (0, sub.lastIndexOf (File.separatorChar));
            sub = sub.substring (0, sub.lastIndexOf (File.separatorChar));
            // Windows
            File f = new File (sub + File.separatorChar + "bin" + File.separatorChar + "hotspot"); // NOI18N
            if (f.exists ())
                return true;
            // Solaris ...
            f = new File (sub + File.separatorChar + "lib" // NOI18N
                          + File.separatorChar + "sparc" + File.separatorChar + "hotspot"); // NOI18N
            if (f.exists ())
                return true;
            // Solaris, Linux ...
            sub = sub.concat (File.separator).concat ("lib"); // NOI18N
            f = new File (sub);
            String [] list = f.list ();
            if (list != null) {
                for (int x = 0; x < list.length; x++)
                    if (new File (sub + File.separatorChar + list [x] + File.separatorChar + "hotspot").exists ()) // NOI18N
                        return true;
            }
        }
        catch (Exception e) {
        }
        
        hotSpotDetected = false;
        try {
            final Process process = Runtime.getRuntime ().exec (path.concat (" -version")); // NOI18N
            Thread t = new Thread () {
                public boolean checkStream (BufferedReader bufferedReader) {
                    final int LIMIT = 10;
                    int lineCounter = 0;
                    boolean res = false;
                    try {
                        String line = bufferedReader.readLine ();
                        while ((line != null) && (lineCounter < LIMIT)) {
                            if (line.toLowerCase ().indexOf ("hotspot") > -1) { // NOI18N
                                res = true;
                                break;
                            }
                            line = bufferedReader.readLine ();
                            lineCounter++;
                        }
                        bufferedReader.close ();
                    } catch (IOException e) {
                    } catch (SecurityException e) {
                    }
                    return res;
                } // checkStream () method
                
                public void run () {
                    hotSpotDetected = false;
                    // err stream
                    BufferedReader bufferedReader = new BufferedReader (new InputStreamReader (process.getErrorStream ()));
                    hotSpotDetected = checkStream (bufferedReader);
                    if (!hotSpotDetected) {
                        // stdout stream
                        bufferedReader = new BufferedReader (new InputStreamReader (process.getInputStream ()));
                        hotSpotDetected = checkStream (bufferedReader);
                    }
                } // run () method
            }; // Thread

            t.start ();
            int counter = 12;
            while ((counter-- > 0) && (t.isAlive()))
                try {
                    Thread.currentThread ().sleep (200);
                } catch (InterruptedException e) {
                }
            process.destroy ();
            if (t.isAlive ())
                try {
                    Thread.currentThread ().sleep (200);
                } catch (InterruptedException e) {
                }
            t.stop ();
            
        } catch (IOException e) {
        } catch (SecurityException e) {
        }
        return hotSpotDetected;
    }

    /**
    *@param str path to a "java" process (e.g. c:\jdk1.3\bin\java, must point to bin, not jre\bin !!!)
    *@return true if classic VM should be launched by the given process 
    */
    public static boolean classicPreferred (String str) {
        if (Utilities.getOperatingSystem () == Utilities.OS_MAC) 
            return false;
        String version = System.getProperty ("java.version");
        if (version.startsWith ("1.3.0")) { // NOI18N
            if (Utilities.getOperatingSystem () == Utilities.OS_SOLARIS)
                return false;
            else
                return true;
        }
        StringTokenizer tokenizer = new StringTokenizer (version, "."); // NOI18N
        try {
            String token = tokenizer.nextToken ();
            if (Integer.parseInt (token) > 1)
                return false;
            token = tokenizer.nextToken ();
            if (Integer.parseInt (token) >= 3)
                return false;
        } catch (Exception e) {
        }
        if (str == null)
            str = System.getProperty ("jdk.home") + 
                File.separatorChar + "bin" + File.separatorChar + "java"; // NOI18N
        return hasHotSpot (str);
    }
    
    /**
     * Returns unqualified class name.
     * @param  class name (qualified or not qualified)
     * @return  the same (instance of) string if the input class name
     *          was not qualified (i.e. it didnt't contain any dots)
     *          or the part of the class name from the last dot (non-inclusive)
     *          to the end if the input string
    */
    public static String getClassName (
        String className
    ) {
        int i = className.lastIndexOf ('.');
        return ((i > 0) ? className.substring (i + 1) : className);
    }

    /**
     * Returns name of the outermost class.
     * Cuts all '<TT>$<I>xxx</I></TT>' appendixes from the specified classname.
     * E.g.:
     * <PRE>
     *     getTopClassName(&quot;com.company.ClassName$InnerClass&quot;) = &quot;com.company.ClassName&quot;
     *     getTopClassName(&quot;ClassName$InnerClass&quot;) = &quot;ClassName&quot;
     * </PRE>
     * @param  className  class name
     * @return  the same (instance of) string if the class name doesn't contain
     *          the '<TT>$</TT>' character or part of the input string - from the
     *          beginning (inclusive) to the first '<TT>$</TT>' character
     *          (non-inclusive).
    */
    public static String getTopClassName (
        String className
    ) {
        // String cn = getClassName (className);
        int i = className.indexOf ('$');
        return ((i > 0) ? className.substring (0, i) : className);
    }

    /**
     * Returns package name of the specified class.
     * @param  className  fully qualified class name
     * @return  package name of the specified class or an empty string
     *          if the class name didn't contain any dots (was not qualified
     *          or is in a default package)
    */
    public static String getPackageName (
        String className
    ) {
        int i = className.lastIndexOf ('.');
        return ((i > 0) ? className.substring (0, i) : ""); // NOI18N
    }

    /**
     * Selects anonymous inner classes from a list of
     * <A HREF="http://java.sun.com/j2se/1.3/docs/guide/jpda/jdi/com/sun/jdi/ReferenceType.html">reference type</A>s.
     * @param  outerClazzName  qualified class name
     * @param  refTypes  list of reference types to be filtered
     * @return  list of reference types that are anonymous nested classes of the <CODE>outerClazzName</CODE>
     * @exception  java.lang.IllegalArgumentException  either of the arguments is <TT>null</TT>
     */
    public static java.util.List anonymousInnerClasses(String outerClazzName, java.util.List refTypes) {
        if (outerClazzName == null) {
            throw new IllegalArgumentException("outerClazzName == null");       //NOI18N
        }
        if (refTypes == null) {
            throw new IllegalArgumentException("refTypes == null");       //NOI18N
        }
        if (refTypes.isEmpty()) {
            return refTypes;
        }
        if (outerClazzName.trim().length() == 0) {
            return java.util.Collections.EMPTY_LIST;
        }
        String mandatoryPrefix = outerClazzName + '$';
        int mandatoryPrefixLength = mandatoryPrefix.length();
        java.util.List resultList = new java.util.ArrayList(refTypes.size());
        for (java.util.Iterator i = refTypes.iterator(); i.hasNext(); ) {
            com.sun.jdi.ReferenceType refType = (com.sun.jdi.ReferenceType)i.next();
            String refTypeName = refType.name();
            if ((refTypeName.length() > mandatoryPrefixLength)
                && refTypeName.startsWith(mandatoryPrefix)
                && (Character.isJavaIdentifierStart(refTypeName.charAt(mandatoryPrefixLength)) == false)) {
                    resultList.add(refType);
            }
        }
        if (resultList.isEmpty()) {
            return java.util.Collections.EMPTY_LIST;
        }
        return resultList;
    }

    /**
    * Return line for given params.
    * className "java.lang.Object$1" sourceName "Object.java".
    */
    public static Line getLineForSource (
        String className,
        String sourceName,
        int lineNumber
    ) {
        if (sourceName == null) return getLine (className, lineNumber);
        int i = sourceName.lastIndexOf ('.');
        if (i > 0)
            sourceName = sourceName.substring (0, i);
        return getLine (
                   getPackageName (className),
                   sourceName,
                   lineNumber
               );
    }

    /**
    * Return line for given params.
    */
    public static Line getLine (String className, int lineNumber) {
        Line.Set ls = getLineSet (className);
        if (ls == null) return null;
        try {
            return ls.getOriginal (lineNumber - 1);
        } catch (IndexOutOfBoundsException e) {
        } catch (IllegalArgumentException e) {
        }
        return null;
    }

    /**
    * Return line set for given class name.
    */
    private static Line.Set getLineSet (String className) {
        LineCookie lineCookie = null;
        // PATCH: ClassElement.forName currently do not returns Element for "ClassName$1"
        ClassElement cls = ClassElement.forName (
            getTopClassName (className)
        );

        if (cls != null)
            lineCookie = (LineCookie) cls.getCookie (LineCookie.class);
        
        if (lineCookie == null) {
            // for cross-compiler & non-java lang. which do not use src hierarchy
            FileObject file = TopManager.getDefault ().getRepository ().find (
                getPackageName (className), 
                getClassName (getTopClassName (className)), 
                "java" // NOI18N
            );
            if (file == null) {
                // PATCH: for JSP - mappings "JSP class name" =ClassElement.forName=>ClassElement 
                // do not work (it works for "Servlet class name" => ClassElement
                file = TopManager.getDefault ().getRepository ().find (
                    getPackageName (className), 
                    getClassName (getTopClassName (className)), 
                    "jsp" // NOI18N
                );
            }
            if (file == null) return null; 
            DataObject data = null;
            try {
                data = DataObject.find (file);
            } catch (Exception e) {  
                return null;  
            }
            lineCookie = (LineCookie) data.getCookie (LineCookie.class);
            if (lineCookie == null)
                return null;
        }
        return lineCookie.getLineSet ();
    }
    
    /**
    * Return (current) line for given params.
    */
    public static Line getCurrentLine (String className, int lineNumber) {
        Line.Set ls = getLineSet (className);
        if (ls == null) return null;
        try {
            return ls.getCurrent (lineNumber - 1);
        } catch (IndexOutOfBoundsException e) {
        } catch (IllegalArgumentException e) {
        }
        return null;
    }
    
    /**
    * Return line for given params.
    */
    private static Line getLine (
        String packageName,
        String className,
        int lineNumber
    ) {
        String name;
        if (packageName.length () == 0)
            name = className;
        else
            name = packageName + '.' + className;
        return getLine (
                   name,
                   lineNumber
               );
    }

    /**
    * Shows given line in editor, and returns it.
    */
    public static Line showInEditor (
        Line line
    ) {
        if (line == null) return null;
        try {
            line.show (line.SHOW_GOTO, 0);
        } catch (Throwable e ) {
            if (e instanceof ThreadDeath)
                throw (ThreadDeath)e;
            TopManager.getDefault ().notify (new NotifyDescriptor.Exception (
                e,
                bundle.getString ("EXC_Editor")
            ));
        }
        return line;
    }

    /**
    * Obtains exception name from errorText.
    */
    public static String getExceptionName (String errorText) {
        try {
            int index1, index2;
            index1 = errorText.indexOf (':')+1;
            String s = errorText.substring (index1);
            s = s.trim ();
            index1 = s.indexOf (':');
            index2 = s.indexOf (' ');
            if ((index2 >= 0) && (index2 < index1))
                index1 = index2;
            index2 = s.indexOf ('\n');
            if ((index2 >= 0) && (index2 < index1))
                index1 = index2;
            s = (s.substring (0, index1)).trim();
            if (s.length() == 0)
                return null;
            else
                return s;
        }
        catch (IndexOutOfBoundsException e) {
            return null;
        }
    }

    public static String getCurrentClassName (Node[] nodes) {
        if (nodes == null) return "";
        if (nodes.length != 1) return "";
        Node n = nodes [0];
        Element e = (Element) n.getCookie (Element.class);
        if (e != null)
            return getClassNameForElement (e);
        DataObject dO = null;
        if (n instanceof DataNode)
            dO = ((DataNode) n).getDataObject ();
        if (dO == null)
            dO = (DataObject) n.getCookie (DataObject.class);
        if (dO == null) return "";
        return dO.getPrimaryFile ().getPackageName ('.');
    }

    public static String getClassNameForElement (Element e) {
        if (e == null) return "";
        if (e instanceof ClassElement)
            return getClassName ((ClassElement) e);
        if (e instanceof ConstructorElement)
            return getClassName (((ConstructorElement) e).getDeclaringClass ());
        if (e instanceof FieldElement)
            return getClassName (((FieldElement) e).getDeclaringClass ());
        return "";
    }

    private static String getClassName (ClassElement e) {
        String f = e.getName ().getFullName ();
        if (!e.isInner ()) return f;
        SourceElement sourceEl = e.getSource ();
        if (sourceEl == null) return f;            
        Identifier ident = sourceEl.getPackage ();
        String c;
        if (ident == null) c = ""; // NOI18N
        else c = ident.getFullName ();
        if (c.length () > 0)
            return c + '.' + f.substring (c.length () + 1).replace ('.', '$');
        return f.replace ('.', '$');
    }

    public static String getCurrentMethodName (Node[] nodes) {
        if (nodes == null) return "";
        if (nodes.length != 1) return "";

        Node n = nodes [0];
        ConstructorElement e = (ConstructorElement) n.getCookie (
                                   ConstructorElement.class
                               );
        if (e == null) return "";
        String name = ((ConstructorElement) e).getName ().getName ();
        String cls = getCurrentClassName (nodes);
        int i = cls.lastIndexOf ('.');
        return name;
    }

    public static String getCurrentFieldName (Node[] nodes) {
        if (nodes == null) return "";
        if (nodes.length != 1) return "";

        Node n = nodes [0];
        FieldElement e = (FieldElement) n.getCookie (
                             FieldElement.class
                         );
        if (e != null)
            return ((FieldElement) e).getName ().getName ();
        return getCurrentIdentifier (nodes);
    }

    public static int getCurrentLineNumber (Node[] nodes) {
        if (nodes == null) return 0;
        if (nodes.length != 1) return 0;

        Node n = nodes [0];
        EditorCookie e = (EditorCookie) n.getCookie (
                             EditorCookie.class
                         );
        if ((e == null) || (e.getOpenedPanes () == null) || (e.getOpenedPanes ().length < 1)) return 0;
        return NbDocument.findLineNumber (
                   e.getDocument (),
                   e.getOpenedPanes () [0].getCaret ().getDot ()
               ) + 1;
    }

    public static String getCurrentIdentifier (Node[] nodes) {
        if (nodes == null) return "";
        if (nodes.length != 1) return "";

        Node n = nodes [0];
        EditorCookie e = (EditorCookie) n.getCookie (
            EditorCookie.class
        );
        if ((e == null) || (e.getOpenedPanes () == null) || (e.getOpenedPanes ().length < 1)) return "";
        String s = e.getOpenedPanes () [0].getSelectedText ();
        if (s == null) return "";
        if (!Utilities.isJavaIdentifier (s)) return "";
        return s;
    }

    
    // comparators ..........................................................................
    
    /** 
    * Public comparator for sorting of variables.
    * Compares according to name and static modifier.
    */
    public static Comparator variablesComparator = new Comparator () {
        public int compare (Object o1, Object o2) throws ClassCastException {
            boolean firstStatic = ((JavaVariable) o1).getModifiers ().indexOf ("static") >= 0; // NOI18N
            boolean secondStatic = ((JavaVariable) o2).getModifiers ().indexOf ("static") >= 0; // NOI18N
            if (firstStatic == secondStatic)
                return ((AbstractVariable) o1).getVariableName ().compareToIgnoreCase (
                    ((AbstractVariable) o2).getVariableName ()
                );
            else
                return firstStatic ? 1 : -1;
        }
    };

    /** 
    * Public comparator for sorting of variables.
    * Compares according to names.
    */
    public static Comparator localsComparator = new Comparator () {
        public int compare (Object o1, Object o2) throws ClassCastException {
            return ((AbstractVariable) o1).getVariableName ().compareToIgnoreCase (
                ((AbstractVariable) o2).getVariableName ()
            );
        }    
    };
    
    
    // innerclasses .........................................................................

    /**
    * Localised Exception.
    */
    /**
    * Hack for using Exception dialog with Details button.
    */
    static class ExceptionHack extends RuntimeException {

        /** Original exception. */
        private Throwable t;
        /** Localized text. */
        private String text;

        ExceptionHack (Throwable t, String text) {
            super (""); // NOI18N
            this.t = t;
            this.text = text;
        }

        public String getMessage () {
            return text + " "; // NOI18N
        }

        public String getLocalizedMessage () {
            return text;
        }

        public void printStackTrace (java.io.PrintStream s) {
            t.printStackTrace (s);
        }

        public void printStackTrace (java.io.PrintWriter s) {
            t.printStackTrace (s);
        }
    }
}
