/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.PropertyEditor;
import java.util.ResourceBundle;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.HashSet;
import java.text.MessageFormat;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;

import com.sun.jdi.ReferenceType;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.Location;
import com.sun.jdi.Method;
import com.sun.jdi.VMMismatchException;
import com.sun.jdi.AbsentInformationException;
import com.sun.jdi.ObjectCollectedException;
import com.sun.jdi.VMDisconnectedException;
import com.sun.jdi.request.ExceptionRequest;
import com.sun.jdi.request.EventRequestManager;
import com.sun.jdi.request.ClassPrepareRequest;
import com.sun.jdi.request.EventRequest;
import com.sun.jdi.event.ExceptionEvent;
import com.sun.jdi.event.ClassPrepareEvent;
import com.sun.jdi.event.Event;
import com.sun.jdi.event.EventSet;
import com.sun.jdi.event.EventIterator;
import com.sun.jdi.event.StepEvent;

import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.explorer.propertysheet.editors.ChoicePropertyEditor;
import org.openide.filesystems.FileObject;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.jpda.util.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.AddBreakpointAction;
import org.netbeans.modules.debugger.support.java.*;


/**
* Implementation of breakpoint on exception.
*
* @author   Jan Jancura
*/
public class ExceptionBreakpoint extends ClassBreakpointEvent 
implements Executor, StopEvent, InfoProducer {


    // static ......................................................................................

    static final long                           serialVersionUID = -1543264672497993747L;
    
    /** Property name constant. */
    public static final String                  PROP_CATCH_TYPE = "catchType"; // NOI18N
    /** Property name constant. */
    public static final String                  PROP_CONDITION = "condition"; // NOI18N

    /** Catch type property value constant. */
    public static final int                     EXCEPTION_CATCHED = 1;
    /** Catch type property value constant. */
    public static final int                     EXCEPTION_UNCATCHED = 2;


    // variables ......................................................................................

    /** Thread which stops on this breakpoint. */
    private transient ThreadReference     thread;
    /** Stores all EventRequests produced by this Event. */
    private Requestor                     requestor;
    private ReferenceType                 tryClass;
    private JPDAVariable                  exception;
    private int                           catchType = EXCEPTION_UNCATCHED;
    private String                        condition = ""; // NOI18N
    /**
     * Set of <TT>ReferenceType</TT>s this breakpoint has been set on.
     * Used for synchronization in method {@link #set(ReferenceType)}.
     */
    private HashSet                       refTypesWithBreakpoint;
    /**
     * Actions available to this event.
     * Its value is <TT>null</TT> until the {@link #getActions} method is called.
     * @see  #getActions
     */
    private transient CoreBreakpoint.Action[]       actions;

    {
        className = ((AddBreakpointAction) SystemAction.get
                     (AddBreakpointAction.class)).getCurrentIdentifier ();
    }
    
    // init ....................................................................
    
    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        if (condition == null)
            condition = ""; // NOI18N
    }
    
    // Event impl ......................................................................................

    /**
    * Sets breakpoint with specified properties.
    */
    public boolean set () {  //S ystem.out.println("ExceptionBreakpoint.set TRY " + this + " in " + getDebugger () ); // NOI18N
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        if ((getClassName () == null) || (getClassName ().trim ().length () < 1)) return false;
        if (requestor == null) requestor = new Requestor (debugger.requestManager);
        try {
            requestor.removeRequests ();
            if (refTypesWithBreakpoint != null) {
                refTypesWithBreakpoint.clear();
            }

            // For unloaded classes
            ClassPrepareRequest cpr = debugger.requestManager.createClassPrepareRequest ();  //S ystem.out.println ("ExceptionBreakpoint.set CLASSES " + getClassName ()); // NOI18N
            cpr.addClassFilter (getClassName () + '*');     //NOI18N
            cpr.setSuspendPolicy (ExceptionRequest.SUSPEND_ALL);
            debugger.operator.register (cpr, this);
            requestor.add (cpr);
            cpr.enable ();

            List l = debugger.virtualMachine.classesByName (getClassName ());
            if (l.size () == 0) return false;

            // Known classes
            //S ystem.out.println ();
            //S ystem.out.println ("    " + this + ":set " + l.size () + " are loaded"); // NOI18N
            boolean set = false;
            ReferenceType ref;
            List anonymousInner = new ArrayList();
            int i;
            //
            //Set breakpoints on all anonymous inner classes and all their nested classes:
            i = 0;
            int iterCount = l.size();
            while (i < iterCount) {
                ref = (ReferenceType)l.get(i);
                set(ref);
                try {
                    anonymousInner.addAll(Utils.anonymousInnerClasses(ref.name(), ref.nestedTypes()));
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            i = 0;
            while (i < anonymousInner.size()) {
                ref = (ReferenceType)anonymousInner.get(i);
                set(ref);
                try {
                    anonymousInner.addAll(ref.nestedTypes());
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            return true;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {  //S ystem.out.println ("ExceptionBreakpoint.remove " + this); // NOI18N
        if (requestor != null)
            requestor.removeRequests ();
    }


    /**
    * Returns specific properties of this event.
    */
    public Node.Property[] getProperties () {
        final ResourceBundle bundle = NbBundle.getBundle (VariableBreakpoint.class);
        return new Node.Property[] {
            new PropertySupport.ReadWrite (
                PROP_CLASS_NAME,
                String.class,
                bundle.getString ("PROP_Exception_class_name"),
                bundle.getString ("HINT_Exception_class_name")
            ) {
                public Object getValue () {
                    return getClassName ();
                }
                public void setValue (Object val) throws IllegalArgumentException {
                    try {
                        setClassName (((String) val).trim ());
                    } catch (ClassCastException e) {
                        throw new IllegalArgumentException ();
                    }
                }
            },
            new PropertySupport.ReadWrite (
                PROP_CATCH_TYPE,
                Integer.TYPE,
                bundle.getString ("PROP_Exception_type_name"),
                bundle.getString ("HINT_Exception_type_name")
            ) {
                public Object getValue () throws IllegalArgumentException {
                    return new Integer (getCatchType ());
                }
                public void setValue (Object val) throws IllegalArgumentException {
                    try {
                        setCatchType (((Integer) val).intValue ());
                    } catch (ClassCastException e) {
                        throw new IllegalArgumentException ();
                    }
                }
                public PropertyEditor getPropertyEditor () {
                    return new ChoicePropertyEditor (
                        new int[] {
                            EXCEPTION_CATCHED,
                            EXCEPTION_UNCATCHED,
                            EXCEPTION_CATCHED | EXCEPTION_UNCATCHED
                        },
                        new String[] {
                            bundle.getString ("CTL_Exception_type_catched"),
                            bundle.getString ("CTL_Exception_type_uncatched"),
                            bundle.getString ("CTL_Exception_type_both")
                        }
                    );
                }
            },
               new PropertySupport.ReadWrite (
                   PROP_CONDITION,
                   String.class,
                   bundle.getString ("PROP_condition"),
                   bundle.getString ("HINT_condition")
               ) {
                   public Object getValue () throws IllegalArgumentException {
                       return getCondition ();
                   }
                   public void setValue (Object val) throws IllegalArgumentException {
                       try {
                           setCondition (((String) val).trim ());
                       } catch (ClassCastException e) {
                           throw new IllegalArgumentException ();
                       }
                   }
               }
        };
    }

    /**
    * Returns line of breakpoint.
    */
    public Line[] getLines () {
        return null;
    }

    /**
    * Returns actions available specially for this version of event.
    */
    public CoreBreakpoint.Action[] getActions() {
        if (actions == null) {
            CoreBreakpoint.Action[] parentActions = super.getActions();
            actions = new CoreBreakpoint.Action[] {
                new StopAction (),
                new PrintAction (PrintAction.BREAKPOINT_EXCEPTION_TEXT),
            };
            if (parentActions.length != 0) {
                CoreBreakpoint.Action[] myActions = actions;
                actions = new CoreBreakpoint.Action[parentActions.length + myActions.length];
                System.arraycopy(parentActions, 0, actions, 0, parentActions.length);
                System.arraycopy(myActions, 0, actions, parentActions.length, myActions.length);
            }
        }
        return actions;
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaThread getThread () {
        if (thread == null) return null;
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        return debugger.threadManager.getThread (thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public CallStackFrame[] getCallStack () {
        return (CallStackFrame[]) ((JPDAThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaVariable getVariable () {
        return exception;
    }

    /**
    * Returns name of type of this event.
    */
    public String getTypeName () {
        return "Exception"; // NOI18N
    }

    /**
    * Returns display name of this event.
    */
    public String getTypeDisplayName () {
        return NbBundle.getBundle (ExceptionBreakpoint.class).getString ("CTL_Exception_event_name_type_name");
    }

    /**
    * Returns display name of this instance of event. It will be used
    * as the name of the breakpoint.
    */
    public String getDisplayName () {
        if ((getCatchType () & EXCEPTION_CATCHED) != 0)
            if ((getCatchType () & EXCEPTION_UNCATCHED) != 0)
                return new MessageFormat (
                    NbBundle.getBundle (ExceptionBreakpoint.class).getString ("CTL_Exception_event_name")
                ).format (new Object[] {getClassName ()});
            else
                return new MessageFormat (
                    NbBundle.getBundle (ExceptionBreakpoint.class).getString ("CTL_Exception_event_name_catched")
                ).format (new Object[] {getClassName ()});
        else
            return new MessageFormat (
                NbBundle.getBundle (ExceptionBreakpoint.class).getString ("CTL_Exception_event_name_uncatched")
            ).format (new Object[] {getClassName ()});
    }

    /**
    * Returns name of icon.
    */
    public String getIconBase () {
        return "/org/netbeans/modules/debugger/resources/breakpointOnException"; // NOI18N
    }

    /**
    * Returns customizer visual component.
    */
    public JComponent getCustomizer () {
        return new ExceptionBreakpointPanel (this);
    }

    public void setCondition (String cond) {
        String old = condition;
        condition = cond;
        firePropertyChange (PROP_CONDITION, old, cond);
    }
    
    public String getCondition () {
        return condition;
    }
    
    protected void perform () {
        if (!condition.trim ().equals ("")) { // NOI18N
            JPDADebugger debugger = (JPDADebugger) getDebugger ();
            JPDAWatch watch = new JPDAWatch (debugger, true);
            watch.setVariableName (condition);
            watch.refreshValue ((JPDAThread) getThread ());
            String type = watch.getType();
            String value = watch.getAsText();
            if ((type != null) && type.equals ("boolean") && (value != null) && value.equals ("false")) { // NOI18N
                stop (false);
                return;
            }
            if ((type == null) || (!type.equals ("boolean"))) { // NOI18N
                // ivalid condition, report message ...
                final ResourceBundle bundle = org.openide.util.NbBundle.getBundle (LineBreakpoint.class);
                debugger.getIOManager ().println (bundle.getString ("CTL_Incorrect_condition") + ": " + // NOI18N
                                  bundle.getString ("CTL_breakpoint_at") + " " + getDisplayName () + ".", IOManager.DEBUGGER_OUT); // NOI18N
            }
        }
        // perform actions ...
        super.perform ();
    }

    // interface Executor .....................................................................

    /**
    * Executes exception hit event.
    */
    public void exec (com.sun.jdi.event.Event event) {  //S ystem.out.println ("ExceptionBreakpoint.exec! " + this + " : " + event); // NOI18N
        if (event instanceof ClassPrepareEvent) {
            tryClass = ((ClassPrepareEvent) event).referenceType ();
            //
            //Resolve whether the class is being demanded:
            boolean isDemanded;
            String tryClassName = tryClass.name();
            if (tryClassName.equals(getClassName())) {
                isDemanded = true;
            }
            else {
                String mandatoryPrefix = getClassName() + '$';
                isDemanded = tryClassName.startsWith(mandatoryPrefix)
                    && (!Character.isJavaIdentifierStart(tryClassName.charAt(mandatoryPrefix.length())));   //only anonymous inner classes
            }
            //
            if (isDemanded) {
                boolean v = set (tryClass);
                if (v && !getBreakpoint ().isValid ()) setValid (true);
            }
            ((JPDADebugger) getDebugger ()).operator.resume ();
            return;
        }
        thread = ((ExceptionEvent) event).thread ();
        exception = new JPDAVariable (
            (JPDADebugger) getDebugger (),
            "Exception", // NOI18N
            ((ExceptionEvent) event).exception (),
            "" // NOI18N
        );
        perform ();
    }


    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */
    public void stop (boolean stop) {
        ((JPDADebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods ......................................................................................

    /**
    * Returns catchType property value.
    *
    * @return catchType property value.
    */
    public int getCatchType () {
        return catchType;
    }

    /**
    * Sets catchType property value.
    *
    * @param new value of catchType property.
    */
    public void setCatchType (int catchType) {
        if (catchType == this.catchType) return;
        if ( (catchType & (EXCEPTION_CATCHED | EXCEPTION_UNCATCHED)) == 0
           ) throw new IllegalArgumentException  ();
        int old = this.catchType;
        this.catchType = catchType;
        firePropertyChange (PROP_CATCH_TYPE, new Integer (old), new Integer (catchType));
    }

    /**
    * Sets exception breakpoint for given class.
    */
    public boolean set (ReferenceType clazz) {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;

        //Ensure that the breakpoint is not set more than once on the same ReferenceType:
        synchronized (this) {
            if (refTypesWithBreakpoint == null) {
                refTypesWithBreakpoint = new HashSet(17);
            }
            else {
                try {
                    if (refTypesWithBreakpoint.contains(clazz)) {
                        return true;
                    }
                }
                catch (ObjectCollectedException ex) {      //may be thrown from 'ReferenceType.equals(O
                }
            }
            refTypesWithBreakpoint.add(clazz);
        }
        try {  //S ystem.out.println ("ExceptionBreakpoint.set for class " + rt); // NOI18N
            ExceptionRequest er = debugger.requestManager.createExceptionRequest (
                clazz,
                (EXCEPTION_CATCHED & catchType) != 0,
                (EXCEPTION_UNCATCHED & catchType) != 0
            );
            er.setSuspendPolicy (ExceptionRequest.SUSPEND_ALL);
            debugger.operator.register (er, this);
            requestor.add (er);
            er.enable ();
            return true;
        } catch (VMDisconnectedException e) {
        } catch (VMMismatchException e) {
        } catch (Exception e) {  //S ystem.out.println ("ExceptionBreakpoint.set class exc. " + e); // NOI18N
            //e.p rintStackTrace ();
        }
        return false;
    }

    public String toString () {
        return "JPDAExceptionBreakpoint " + getClassName (); // NOI18N
    }
}

