/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.*;
import java.util.*;
import javax.swing.SwingUtilities;

import com.sun.jdi.*;

import org.openide.debugger.DebuggerException;
import org.openide.text.Line;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.support.java.CallStackFrame;
import org.netbeans.modules.debugger.AbstractVariable;
import org.netbeans.modules.debugger.support.util.Protector;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* Class representating one line of callstack.
*/
public class JPDACallStackFrame extends CallStackFrame {

    /** Owning thread. */
    private ThreadReference               threadS;
    private int                           indexS;
    /** Debugger link. */
    private JPDADebugger                  debugger;
    /** Cache for old locals value. */
    HashMap                               oldLocals = new HashMap ();

    /**
    * Creates a new call stack line.
    *
    * @param lineNumber
    * @param methodName
    * @param className
    */
    JPDACallStackFrame (
        JPDADebugger debugger,
        ThreadReference thread,
        int index
    ) {
        this.debugger = debugger;
        this.threadS = thread;
        this.indexS = index;
    }

    StackFrame stackFrame () {
        try {
            return threadS.frame (indexS);
        } catch (IncompatibleThreadStateException e) {
        }
        return null;
    }
    
    /**
    * Returns line number of this frame in this callstack.
    *
    * @return Returns line number of this frame in this callstack.
    */
    public int getLineNumber () {
        try {
            return stackFrame ().location ().lineNumber ();
        } catch (Exception e) {
        }
        return -1;
    }

    /**
    * Returns method name of this frame in this callstack.
    *
    * @return Returns method name of this frame in this callstack.
    */
    public String getMethodName () {
        try {
            Method m = stackFrame ().location ().method ();
            if (m == null) return ""; // NOI18N
            return m.name ();
        } catch (Exception e) {
        }
        return ""; // NOI18N
    }

    /**
    * Returns class name of this frame in this callstack.
    *
    * @return Returns class name of this frame in this callstack.
    */
    public String getClassName () {
        try {
            return stackFrame().location ().declaringType ().name ();
        } catch (Exception e) {
        }
        return ""; // NOI18N
    }

    /**
    * Returns name of file of this frame.
    *
    * @return Returns name of file of this frame.
    * @throws DebuggerException if informations about source are not included or some other error
    *   occurres.
    */
    public String getSourceName () throws DebuggerException {
        try {
            return stackFrame().location ().sourceName ();
        } catch (Exception e) {
            throw new DebuggerException (e);
        }
    }
    
    /**
    * Returns human-readable text info about currently returned variables 
    * or null. This text should be displayed with list of variables. 
    *
    * @return human-readable text info about currently returned variables 
    * or null
    */
    public String getInfo () {
        try {
            if (stackFrame ().location ().method ().isNative ()) return null;
            stackFrame ().visibleVariables ();
            return null;
        } catch (AbsentInformationException e) {
            return NbBundle.getBundle(JPDACallStackFrame.class).getString("CTL_Source_compiled_without_g");
        } catch (Exception e) {
            e.printStackTrace ();
        }
        return null;
    }

    /**
    * Returns current local variables for this frame stack.
    *
    * @return Current local variables for this frame stack.
    */
    public AbstractVariable[] getVariables () {
        if (debugger.isDisconnected ())
            return new AbstractVariable [0];
        try {
            List l = null;
            if (stackFrame ().location ().method ().isNative ())
                l = new LinkedList ();
            else
                try {
                    l = stackFrame().visibleVariables ();
                } catch (AbsentInformationException e) {
                    l = new LinkedList ();
                }
            HashMap newLocals = new HashMap ();
            int i, k = l.size ();
            ObjectReference objectRef = stackFrame ().thisObject ();
            ArrayList result = new ArrayList ();
            JPDAVariable var;

            for (i = 0; i < k; i++) {
                LocalVariable lv = (LocalVariable) l.get (i);
                var = (JPDAVariable) oldLocals.get (lv.name ());
                try {
                    if (var == null)
                        var = new JPDAVariable (
                            debugger,
                            lv.name (),
                            stackFrame().getValue (lv),
                            lv.typeName (),
                            threadS,
                            indexS
                        );
                    else {
                        var.update (
                            lv.name (),
                            stackFrame().getValue (lv),
                            lv.typeName (),
                            threadS,
                            indexS
                        );
                        var.firePropertyChange ();
                    }
                    result.add (var);
                    newLocals.put (lv.name (), var);
                } catch (InconsistentDebugInfoException e) {
                    // do not return this local variable, its value cannot be obtained
                }
            }
            
            // add "this" local variable
            if (objectRef != null) {
                var = (JPDAVariable) oldLocals.get ("this"); // NOI18N
                if (var == null)
                    var = new JPDAVariable (
                        debugger,
                        "this", // NOI18N
                        objectRef,
                        objectRef.type ().name ()
                    );
                else {
                    var.update (
                        "this", // NOI18N
                        objectRef,
                        objectRef.type ().name ()
                    );
                    var.firePropertyChange ();
                }
                result.add (var);
                newLocals.put ("this", var); // NOI18N
            }
            
            oldLocals = newLocals;
            Iterator iter = result.iterator ();
            JPDAVariable [] variables = new JPDAVariable [result.size ()];
            for (int x = 0; x < result.size (); x++)
                variables [x] = (JPDAVariable) iter.next ();
            return variables;
        } catch (Exception e) {
            e.printStackTrace ();
        }
        return new AbstractVariable [0];
    }

    /**
    * Adds listener on the property changing.
    */
    public void addPropertyChangeListener (PropertyChangeListener listener) {
    }

    /**
    * Removes listener on the property changing.
    */
    public void removePropertyChangeListener (PropertyChangeListener listener) {
    }
}
 