/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is Forte for Java, Community Edition. The Initial
 * Developer of the Original Code is Sun Microsystems, Inc. Portions
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.util.*;

import com.sun.jdi.*;

import org.openide.TopManager;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.util.NbBundle;
import org.openide.src.Type;
import org.openide.src.*;

import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.AbstractVariable;

/**
 *
 * @author   Daniel Prusa
 */
public class JPDAClass extends JavaClass {
    
    // static final long serialVersionUID =-121348327319864863L;
    
    /*
    public static final String BREAKPOINT_EXCEPTION_TEXT =
        NbBundle.getBundle (PrintAction.class).getString ("CTL_Default_exception_print_text");
     */
    
    private ReferenceType   referenceType;
    private JPDADebugger    debugger;
    
    private boolean         isClass;
    
    // a storage for already obtained methods, constructors, fields ...
    private MethodElement []            methods = null;
    private ConstructorElement []       constructors = null;
    private InitializerElement          initializer = null;
    private FieldElement []             fields = null;
    private Hashtable                   classes = new Hashtable ();
    
    // init ...............................................................................................
    
    public JPDAClass (ReferenceType ref, JPDADebugger debugger) {
        referenceType = ref;
        this.debugger = debugger;
        isClass = referenceType instanceof ClassType;
    }
    
    // methods ............................................................................................
    
    /**
     * Returns the class name.
     */
    public String getName () {
        if (debugger.isDisconnected ())
            return ""; // NOI18N
        try {
            return referenceType.name ();
        }
        catch (ObjectCollectedException e) {
        }
        return ""; // NOI18N
    }
    
    private String shortName (String name) {
        return name.substring (name.lastIndexOf ('.') + 1);
    }
    
    /**
     * Returns all loaded inner classes of the class
     */
    public JavaClass [] getInnerClasses () {
        if (debugger.isDisconnected ())
            return new JavaClass [0];
        try {
            List list = referenceType.nestedTypes ();
                
            Iterator iter = list.iterator ();
            Hashtable temp = new Hashtable ();
            
            while (iter.hasNext ()) {
                try {
                    ReferenceType ref = (ReferenceType) iter.next ();
                    String name = ref.name ();
                    JavaClass inner = (JavaClass) classes.get (shortName (name));
                    try {
                        if (inner == null) {
                            if (name.substring (getName ().length () + 1).indexOf ('$') == -1) 
                            // is it a direct inner class ?
                                inner = new JPDAClass (ref, debugger);
                        }
                    } catch (IndexOutOfBoundsException e) {
                    }
                    if (inner != null)
                        temp.put (shortName (name), inner);
                } catch (ObjectCollectedException e) {
                }
            }
            classes = temp;
            
            Enumeration enum = classes.elements ();
            JavaClass [] res = new JavaClass [classes.size ()];
            int index = 0;
            while (enum.hasMoreElements ())
                res [index++] = (JPDAClass) enum.nextElement ();
            
            // if (getName ().indexOf ('.') == -1)
            return res;
        } catch (ObjectCollectedException e) {
        } catch (VMDisconnectedException e) {
        }
        return new JavaClass [0];
    }
    
    /**
     * Gets all methods and stores them.
     */
    private void initMethods () {
        if (debugger.isDisconnected ())
            return;
        if ((methods != null) && (constructors != null))
            return; // already inited
        try {
            List methodsList = referenceType.methods ();
            Iterator iter = methodsList.iterator ();
            MethodElement [] tempM = new MethodElement [methodsList.size ()];
            int index = 0;
            ArrayList tempC = new ArrayList (); // temp list of constructors
            while (iter.hasNext ()) {
                Method method = (Method) iter.next ();
                if (method.isSynthetic ()) // do not add a synthetic method
                    continue;
                List args = new ArrayList ();
                MethodParameter [] params;
                boolean failed = true;
                try {
                    args = method.arguments ();
                    failed = false;
                } catch (AbsentInformationException e) {
                } catch (NativeMethodException e) {
                }
                if (failed)
                    args = method.argumentTypeNames ();
                params = new MethodParameter [args.size ()];
                if (!failed)
                    for (int x = 0; x < args.size (); x++) {
                        LocalVariable param = (LocalVariable)args.get (x);
                        params [x] = new MethodParameter (
                        param.name (),
                        Type.createClass (Identifier.create (param.typeName ())),
                        false // we cannot obtain this value (final modifier)
                        );
                    }
                else
                    for (int x = 0; x < args.size (); x++)
                        params [x] = new MethodParameter (
                        "", // argument name cannot be obtained // NOI18N
                        Type.createClass (Identifier.create ((String) args.get (x))),
                        false // we cannot obtain this value (final modifier)
                        );
                if (method.isStaticInitializer ()) {
                    initializer = new InitializerElement ();
                    initializer.setStatic (true);
                } else {
                    ConstructorElement element = null;
                    if (method.isConstructor ()) {
                        element = new ConstructorElement ();
                        tempC.add (element);
                    } else {
                        element = new MethodElement ();
                        ((MethodElement) element).setReturn (Type.createClass (Identifier.create (method.returnTypeName ())));
                        tempM [index++] = (MethodElement) element;
                    }
                    element.setParameters (params);
                    element.setName (Identifier.create (method.name ()));
                    element.setModifiers (method.modifiers ());
                }
            }
            methods = new MethodElement [index];
            for (int x = 0; x < index; x++)
                methods [x] = tempM [x];
            constructors = new ConstructorElement [tempC.size ()];
            for (int x = 0; x < tempC.size (); x++)
                constructors [x] = (ConstructorElement) tempC.get (x);
            return;
        } catch (ObjectCollectedException e) {
        } catch (SourceException e) {
        } catch (ClassNotPreparedException e) {
        }
    }
    
    public MethodElement [] getMethods () {
        initMethods ();
        if (methods != null)
            return methods;
        else
            return new MethodElement [0];
    }
    
    public ConstructorElement [] getConstructors () {
        initMethods ();
        if (constructors != null)
            return constructors;
        else
            return new ConstructorElement [0];
    }
    
    public InitializerElement [] getInitializers () {
        initMethods ();
        return (initializer == null) ? new InitializerElement [0] : new InitializerElement [] {initializer};
    }
    
    public FieldElement [] getFields () {
        if (debugger.isDisconnected ())
            return new FieldElement [0];
        if (fields != null)
            return fields;
        try {
            List list = referenceType.fields ();
            Iterator iter = list.iterator ();
            FieldElement [] res = new FieldElement [list.size ()];
            int index = 0;
            while (iter.hasNext ()) {
                Field field = (Field) iter.next ();
                FieldElement element = new FieldElement ();
                element.setName (Identifier.create (field.name ()));
                element.setType (Type.createClass (Identifier.create (field.typeName ())));
                element.setModifiers (field.modifiers ());
                res [index++] = element;
            }
            return fields = res; // store fields and return them
        } catch (ObjectCollectedException e) {
        } catch (SourceException e) {
        } catch (ClassNotPreparedException e) {
        }
        return new FieldElement [] {};
    }
    
    public Identifier getSuperClass () {
        if (debugger.isDisconnected ())
            return null;
        if (!(referenceType instanceof ClassType))
            return null;
        try {
            ReferenceType ref = ((ClassType)referenceType).superclass ();
            if (ref != null)
                return Identifier.create (ref.name ());
        } catch (ObjectCollectedException e) {
        } catch (ClassNotPreparedException e) {
        }
        return null;
    }
    
    public Identifier [] getInterfaces () {
        if (debugger.isDisconnected ())
            return new Identifier [0];
        try {
            List list = null;
            if (referenceType instanceof ClassType)
                list = ((ClassType) referenceType).interfaces ();
            if (referenceType instanceof InterfaceType)
                list = ((InterfaceType) referenceType).superinterfaces ();
            if (list == null)
                return new Identifier [0];
            Identifier [] res = new Identifier [list.size ()];
            for (int x = 0; x < list.size (); x++)
                res [x] = Identifier.create (((ReferenceType) list.get (x)).name ());
            return res;
        } catch (ObjectCollectedException e) {
        } catch (ClassNotPreparedException e) {
        }
        return new Identifier [0];
    }
    
    public boolean isClass () {
        return isClass;
    }
    
    public boolean isPrepared () {
        if (debugger.isDisconnected ())
            return false;
        return referenceType.isPrepared ();
    }
    
    public int modifiers () {
        if (debugger.isDisconnected ())
            return 0;
        return referenceType.modifiers ();
    }
    
    public String classLoaderTypeName () {
        if (debugger.isDisconnected ())
            return null;
        if (referenceType.classLoader () == null)
            return null; // bootstrap class loader
        try {
            return referenceType.classLoader ().referenceType ().name ();
        } catch (ObjectCollectedException e) {
        }
        return null; // ???
    }
    
    public JavaVariable getClassLoader () {
        if (debugger.isDisconnected ())
            return null;
        ClassLoaderReference ref = referenceType.classLoader ();
        if (ref == null)
            return null; // bootstrap class loader
        JPDAVariable var = new ClassLoaderVariable (debugger, "Classloader", ref, ref.referenceType ().toString ()); // NOI18N
        return var;
    }
    
    public JavaVariable getStaticVariable (FieldElement element) {
        if (debugger.isDisconnected ())
            return new JPDAVariable (
                debugger,
                false
            );
        return new JPDAVariable (
            debugger,
            referenceType,
            referenceType.fieldByName (element.getName ().getFullName ())
        );
    }
    
    // helper methods ................................................................
    
    ReferenceType getReferenceType () {
        return referenceType;
    }
    
    public boolean equals (Object o) {
        if (!(o instanceof JPDAClass))
            return false;
        return (((JPDAClass)o).getReferenceType ()).equals (referenceType);
    }
    
    public int hashCode () {
        return referenceType.hashCode ();
    }
    
    // helper innerclasses ...........................................................
    
    private static class ClassLoaderVariable extends JPDAVariable {
        
        ClassLoaderVariable (
        JPDADebugger  debugger,
        String        name,
        Value         value,
        String        type
        ) {
            super (debugger, name, value, type);
            isCloned = true; // fix it
        }
        
    }
    
}
