/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.*;
import java.io.*;
import java.text.MessageFormat;
import java.awt.Component;
import java.awt.Insets;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Dimension;
import java.awt.Window;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;

import com.sun.jdi.*;
import com.sun.jdi.connect.*;
import com.sun.jdi.connect.Connector.Argument;
import com.sun.jdi.event.*;
import com.sun.jdi.request.*;

import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.text.Line;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.cookies.LineCookie;
import org.openide.debugger.*;
import org.openide.filesystems.*;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.execution.NbProcessDescriptor;
import org.openide.execution.NbClassPath;
import org.openide.util.MapFormat;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.jpda.util.*;


/**
* Main Corona debugger class
*
* @author   Jan Jancura
* @version  0.47, May 26, 1998
*/
public class JPDADebugger extends JavaDebugger implements Executor, LoadedClassesProducer {


    // static ........................................................................

    static final long                          serialVersionUID = 2797853329739651906L;

    public static int                          ACTION_ON_TI_FIND;

    /** bundle to obtain text information from */
    static ResourceBundle                      bundle;

    static String getLocString (String s) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JPDADebugger.class);
        return bundle.getString (s);
    }
    
    private static Random                     random = new Random ();


    // variables .................................................................

    transient VirtualMachine                  virtualMachine = null;
    transient EventRequestManager             requestManager = null;
    transient protected ThreadManager         threadManager;
    transient protected Operator              operator;
    transient private Process                 process;
    transient StepRequest                     stepRequest;
    private transient MethodEntryRequest      findSourceMER;
    private transient StepRequest             findSourceSR;
    private transient int                     findSourceCounter = 0;
    transient private Thread                  debuggerThread;
    transient boolean                         finishVM;
    private transient boolean                 disconnected = false;
    private transient Launcher                launcher;

    private transient AttachingConnector      connector;
    private transient Map                     args;
    private transient String                  mainClassName;
    private transient String                  stopClassName;

    // threads
    private transient JPDAThread              currentThread = null;
    protected transient JPDAThreadGroup       threadGroup = new JPDAThreadGroup (null);

    private transient boolean                 stopOnMain = false;
    private transient DebuggerInfo            debuggerInfo;

    private static transient String []        stopMethodNames = {"main", "start", "init", "<init>"}; // NOI18N
    private transient CoreBreakpoint []       breakpointMain = null;

    
    // init .......................................................................

    /**
    * Deserializes debugger.
    */
    protected void setDebugger (AbstractDebugger debugger) {
        super.setDebugger (debugger);
    }


    // Debugger implementation .................................................................

    /** Starts the debugger. The method stops the current debugging (if any)
    * and takes information from the provided info (containing the class to start and
    * arguments to pass it and name of class to stop debugging in) and starts
    * new debugging session.
    *
    * @param info debugger info about class to start
    * @exception DebuggerException if an error occures during the start of the debugger
    */
    public void startDebugger (DebuggerInfo info) throws DebuggerException {
        debuggerInfo = info;
        if (virtualMachine != null)
            finishDebugger ();
        
        stopOnMain = info.getStopClassName () != null;
        mainClassName = info.getClassName ();  //S ystem.out.println ("JPDADebugger stop on " + info.getStopClassName ()); // NOI18N
        finishVM = true;
        
        // open output window ...
        super.startDebugger (info);

        // stop on main
        if (stopOnMain) {
            try {
                String stopClassName = debuggerInfo.getStopClassName ();
                BreakpointProducer d = (BreakpointProducer) 
                    TopManager.getDefault ().getDebugger ();
                breakpointMain = new CoreBreakpoint [stopMethodNames.length];
                for (int x = 0; x < breakpointMain.length; x++) {
                    breakpointMain [x] = (CoreBreakpoint) d.createBreakpoint (true);
                    MethodBreakpoint mb = new MethodBreakpoint ();
                    mb.setClassName (stopClassName); // NOI18N
                    mb.setMethodName (stopMethodNames [x]);
                    breakpointMain [x].setEvent (mb);
                    CoreBreakpoint.Action[] a = breakpointMain [x].getActions ();
                    int i, ii = a.length;
                    for (i = 0; i < ii; i ++)
                        if (a [i] instanceof PrintAction) {
                            ((PrintAction) a [i]).setPrintText (
                                getLocString ("CTL_Stop_On_Main_print_text")
                            );
                        }
                }

                addPropertyChangeListener (new PropertyChangeListener () {
                    public void propertyChange (PropertyChangeEvent ev) {
                        if (ev.getPropertyName ().equals (PROP_STATE)) {
                            if ( (((Integer)ev.getNewValue ()).intValue () == DEBUGGER_STOPPED) ||
                                 (((Integer)ev.getNewValue ()).intValue () == DEBUGGER_NOT_RUNNING)
                            ) {
                                if (breakpointMain != null) {
                                    for (int x = 0; x < breakpointMain.length; x++)
                                        breakpointMain [x].remove();
                                    breakpointMain = null;
                                }
                                removePropertyChangeListener(this);
                            }
                        }
                    }
                });

            } catch (DebuggerException e) {
                e.printStackTrace();
            }
        }

        // start & init remote debugger ............................................
        boolean launch = false;
        synchronized (getClass ()) {
            if (info instanceof ReconnectDebuggerInfo) {
                virtualMachine = reconnect ((ReconnectDebuggerInfo) info);
            } else
                if (info instanceof RemoteDebuggerInfo) {
                    virtualMachine = connect ((RemoteDebuggerInfo) info);
                } else {
                    virtualMachine = launch (info);
                    getIOManager ().showOutput (
                        process, 
                        IOManager.STD_OUT + IOManager.ERR_OUT, 
                        IOManager.PROCESS_OUT
                    );
                    getIOManager ().connectInput (process);
                    launch = true;
                }
            requestManager = virtualMachine.eventRequestManager ();
            operator = new Operator (
                virtualMachine,
                launch ?
                    new Runnable () {
                        public void run () {
                            startDebugger ();
                        }
                    } : null,
                new Runnable () {
                    public void run () {
                        try {
                            finishVM = false;
                            finishDebugger ();
                        } catch (DebuggerException e) {
                        }
                    }
                }
            );
            operator.start ();
            if (!launch) startDebugger ();                
	}
    }

    /**
    * Finishes debugger.
    */
    synchronized public void finishDebugger () throws DebuggerException {
        disconnected = true;
        if (threadManager != null) threadManager.finish ();
        if (debuggerThread != null) {
            debuggerThread.interrupt ();
            debuggerThread.stop ();
        }
        if (breakpointMain != null) {
            for (int x = 0; x < breakpointMain.length; x++)
                breakpointMain [x].remove ();
            breakpointMain = null;
        }
        if (finishVM)
            try {
                if (virtualMachine != null)
                    if (debuggerInfo instanceof RemoteDebuggerInfo)
                        virtualMachine.dispose ();
                    else {
                        virtualMachine.exit (0);
                        // the process should be destroyed by VM.exit (0), however it seems
                        // there are some problems with hanging processes on Solaris so let's try
                        // to destroy the process explicitly
                        if (process != null)
                            process.destroy ();
                    }
            } catch (VMDisconnectedException e) {
            } catch (UnsupportedOperationException e) {
            }
            
        super.finishDebugger ();
        if (threadManager != null)
            threadManager.refreshStacks ();
    }

    /**
    * Trace into.
     *
     *
     *
    */
    synchronized public void traceInto () throws DebuggerException {
        if ((virtualMachine == null) || (currentThread == null)) return;
        try {
            removeStepRequest ();
            setLastAction (ACTION_TRACE_INTO);
            stepRequest = requestManager.createStepRequest (
                              currentThread.getThreadReference (),
                              StepRequest.STEP_LINE,
                              StepRequest.STEP_INTO
                          );
            stepRequest.addCountFilter (1);
            stepRequest.putProperty ("traceInto", "traceInto"); // NOI18N
            stepRequest.setSuspendPolicy (EventRequest.SUSPEND_ALL);
            operator.register (stepRequest, this);
            stepRequest.enable ();
            virtualMachine.resume ();
            super.traceInto ();
        } catch (VMDisconnectedException e) {
        }
    }

    /**
    * Trace over.
    */
    synchronized public void traceOver () throws DebuggerException {
        if ((virtualMachine == null) || (currentThread == null)) return;
        try {
            removeStepRequest ();
            setLastAction (ACTION_TRACE_OVER);
            stepRequest = requestManager.createStepRequest (
                              currentThread.getThreadReference (),
                              StepRequest.STEP_LINE,
                              StepRequest.STEP_OVER
                          );
            stepRequest.addCountFilter (1);
            stepRequest.setSuspendPolicy (EventRequest.SUSPEND_ALL);
            operator.register (stepRequest, this);
            stepRequest.enable ();
            virtualMachine.resume ();
            super.traceOver ();
        } catch (VMDisconnectedException e) {
        }
    }

    /**
    * Go.
    */
    synchronized public void go () throws DebuggerException {
        if (virtualMachine == null) return;
        try {
            setLastAction (ACTION_GO);
            removeStepRequest ();
            virtualMachine.resume ();
            // threadManager.refreshStacks ();
            super.go ();
        } catch (VMDisconnectedException e) {
        }
    }

    /**
    * Step out.
    */
    synchronized public void stepOut () throws DebuggerException {
        if ((virtualMachine == null) || (currentThread == null)) return;
        try {
            removeStepRequest ();
            setLastAction (ACTION_STEP_OUT);
            stepRequest = requestManager.createStepRequest (
                              currentThread.getThreadReference (),
                              StepRequest.STEP_LINE,
                              StepRequest.STEP_OUT
                          );
            stepRequest.addCountFilter (1);
            stepRequest.setSuspendPolicy (EventRequest.SUSPEND_ALL);
            operator.register (stepRequest, this);
            stepRequest.enable ();
            virtualMachine.resume ();
            super.stepOut ();
        } catch (VMDisconnectedException e) {
        }
    }

    /**
    * Suspend/Resume debugger.
    */
    public void setSuspended (boolean suspended) {
        if (suspended) {
            threadGroup.setSuspended (true);
            if ((currentThread == null) && !isDisconnected ()) {
                JPDAThread thread = threadManager.getAnyThread ();
                if (thread != null)
                    thread.setCurrent (true);
            }
            setState (DEBUGGER_STOPPED);
            final IOManager ioManager = getIOManager ();
            SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    ioManager.println (getLocString ("CTL_Debugger_paused"),
                        IOManager.STATUS_OUT
                    );
                }
            });
            updateWatches ();
            threadManager.refreshStacks ();
        } else {
            threadGroup.markSuspended (false);
            if (virtualMachine != null)
                virtualMachine.resume ();
            setState (DEBUGGER_RUNNING);
        }
    }
    
    // WATCHES ..............................................................

    /** Creates new uninitialized watch. The watch is visible (not hidden).
    *
    * @return new uninitialized watch
    */
    public Watch createWatch () {
        JPDAWatch w = new JPDAWatch (this, false);
        addWatch (w);
        return w;
    }

    /** Creates a watch its expression is set to initial value. Also
    * allows to create a watch not presented to the user, for example
    * for internal usage in the editor to obtain values of variables
    * under the mouse pointer.
    *
    * @param expr expresion to watch for
    * @param hidden allows to specify whether the watch should be presented
    *   to the user or not (be only of internal usage of the IDE).
    * @return new watch
    */
    public Watch createWatch (String expr, boolean hidden) {
        JPDAWatch w = new JPDAWatch (this, false);
        w.setVariableName (expr);
        addWatch (w);
        return w;
    }


    // AbstractDebugger implementation ..................................................


    // properties ........................

    /**
    * @return name of proces for given DebuggerInfo.
    */
    public String getProcessName (DebuggerInfo info) {
        if (info instanceof RemoteDebuggerInfo) {
            if ( hasShmemTransport (((RemoteDebuggerInfo) info).getConnector ())
               ) {
                Argument a = (Argument) ((RemoteDebuggerInfo) info).getArgs ().
                             get ("name"); // NOI18N
                if (a == null)
                    return "localhost:???"; // NOI18N
                else
                    return "localhost:" + a.value (); // NOI18N
            } else
                if ( hasSocketTransport (((RemoteDebuggerInfo) info).getConnector ())
                   ) {
                    Argument name = (Argument) ((RemoteDebuggerInfo) info).getArgs ().
                                    get ("hostname"); // NOI18N
                    Argument port = (Argument) ((RemoteDebuggerInfo) info).getArgs ().
                                    get ("port"); // NOI18N
                    return ((name == null) ? "???:" : (name.value () + ":")) + // NOI18N
                           ((port == null) ? "???" : (port.value ())); // NOI18N
                } else
                    return "???"; // NOI18N
        } else
            return super.getProcessName (info);
    }

    /**
    * @return name of location for given DebuggerInfo.
    */
    public String getLocationName (DebuggerInfo info) {
        if (info instanceof RemoteDebuggerInfo) {
            if ( hasShmemTransport (((RemoteDebuggerInfo) info).getConnector ())
               ) {
                return "localhost"; // NOI18N
            } else
                if ( hasSocketTransport (((RemoteDebuggerInfo) info).getConnector ())
                   ) {
                    Argument name = (Argument) ((RemoteDebuggerInfo) info).getArgs ().
                                    get ("hostname"); // NOI18N
                    return name == null ? "localhost" : name.value (); // NOI18N
                } else
                    return "localhost"; // NOI18N
        } else
            return "localhost"; // NOI18N
    }


    // threads ........................

    /**
    * Returns root of all threads.
    */
    public ThreadsProducer getThreadsRoot () {
        return threadGroup;
    }

    /**
    * Returns current thread or null.
    */
    public AbstractThread getCurrentThread () {
        return currentThread;
    }

    /**
    * Sets current thread. If thread is null, unsets curent thread.
    */
    public void setCurrentThread (AbstractThread thread) {
        if (currentThread == thread) return;
        Object old = currentThread;
        currentThread = (JPDAThread) thread;
        firePropertyChange (PROP_CURRENT_THREAD, old, thread);
    }

    // classes .........................................................................
    
    /**
    * Returns list of loaded classes.
    */
    public java.util.List getLoadedClasses () {
        // return new ArrayList ();
        if ((virtualMachine == null) || isDisconnected ())
            return new ArrayList ();
        try {
            List list = virtualMachine.allClasses ();
            ArrayList res = new ArrayList ();
            for (int i = 0; i < list.size (); i++)
                try {
                    if ( (!(list.get (i) instanceof ArrayType)) && // do not return ArrayTypes
                         (((ReferenceType)list.get (i)).name ().indexOf ('$') == -1) // do not return inner classes
                       )
                        res.add (new JPDAClass ((ReferenceType)list.get (i), this));
                } catch (ObjectCollectedException e) {
                }
            return res;
        } catch (VMDisconnectedException e) {
        }
        return new ArrayList ();
    }

    // interface Executor .....................................................................

    /**
    * Executes breakpoint hit event.
    */
    public void exec (com.sun.jdi.event.Event ev) {
        //S ystem.out.println ("exec " + ev); // NOI18N
        removeStepRequest ();
        StepEvent event = (StepEvent) ev;
        ThreadReference tr = event.thread ();
        com.sun.jdi.Location loc = event.location ();
        int ln = -1;
        String methodName = "?"; // NOI18N
        String className = "?"; // NOI18N
        String lineNumber = "?"; // NOI18N
        String threadName = tr.name ();
        Line l = null;

        if (loc != null) {
            if (loc.method () != null)
                methodName = loc.method ().name ();
            className = loc.declaringType ().name ();
            ln = loc.lineNumber ();
            if (ln >= 0)
                lineNumber = "" + loc.lineNumber ();
        }

        if (ln != -1)
            try {
                l = Utils.getLineForSource (
                        className,
                        loc.sourceName (),
                        ln
                    );
            } catch (AbsentInformationException e) {
                l = Utils.getLine (
                        className,
                        ln
                    );
            }


        if (resolveCanBeCurrent (tr, l))
            // if this line can not be current resolveCanBeCurrent () calls stepOver
            return;
        // line can be current

        if ( (l == null) &&
                (getLastAction () == ACTION_TRACE_INTO)
           ) {
            DebuggerSettings debuggerSettings = (DebuggerSettings) DebuggerSettings.
                                                findObject (DebuggerSettings.class);
            if (debuggerSettings.isActionOnTraceIntoSet ()) {
                // ask user, what to do ...
                ActionTIPanel ap = new ActionTIPanel ();
                NotifyDescriptor.Confirmation mess = new NotifyDescriptor.Confirmation (
                                                         ap,
                                                         NotifyDescriptor.OK_CANCEL_OPTION
                                                     );
                Object o = TopManager.getDefault ().notify (mess);
                if (o.equals (NotifyDescriptor.OK_OPTION))
                    ap.updateSettings ();
            }

            int o = debuggerSettings.getActionOnTraceInto ();
            if (o == ACTION_ON_TI_FIND) {
                //S ystem.out.println();
                //S ystem.out.println("TTSC ========================================================");
                // try to find another "intelligent" line of code
                // you know - intelligent means that one with source code
                traceToSourceCode (tr);
            } else
                if (o == DebuggerSettings.ACTION_ON_TI_TRACE_OUT) {
                    //S ystem.out.println();
                    //S ystem.out.println("Step ========================================================");
                    stepRequest = requestManager.createStepRequest (
                                      tr,
                                      StepRequest.STEP_LINE,
                                      StepRequest.STEP_OUT
                                  );
                    stepRequest.addCountFilter (1);
                    stepRequest.setSuspendPolicy (EventRequest.SUSPEND_ALL);
                    operator.register (stepRequest, this);
                    stepRequest.enable ();
                    operator.resume ();
                } else {
                    //S ystem.out.println();
                    //S ystem.out.println("Stop ========================================================");
                    makeCurrent (
                        threadName,
                        className,
                        methodName,
                        lineNumber,
                        l != null,
                        tr
                    );
                    operator.stopRequest ();
                }
        } else {
            makeCurrent (
                threadName,
                className,
                methodName,
                lineNumber,
                l != null,
                tr
            );
            operator.stopRequest ();
        }
    }



    // support for multisession debugging ................................................................

    /**
    * Disconnects from running debugged process.
    */
    public void disconnect () throws DebuggerException {
        if (breakpointMain != null) {
            for (int x = 0; x < breakpointMain.length; x++)
                breakpointMain [x].remove ();
            breakpointMain = null;
        }
        try {
            if (virtualMachine != null) virtualMachine.dispose ();
        } catch (VMDisconnectedException e) {
        }
        if (threadManager != null) threadManager.finish ();
        if (debuggerThread != null) {
            debuggerThread.interrupt ();
            debuggerThread.stop ();
        }
        super.finishDebugger ();
    }

    /**
    * Reconnects to disconnected Virtual Machine.
    */
    public void reconnect () throws DebuggerException {
        startDebugger (new ReconnectDebuggerInfo (
                           connector,
                           args
                       ));
    }


    // helper private methods .........................................................................

    /**
    * Finds the first executed line with source code.
    */
    public void traceToSourceCode (ThreadReference thread) {
        //S ystem.out.println ("Start finding!!! "); // NOI18N

        // create Step Request for searching a source code
        try {
            findSourceSR = requestManager.createStepRequest (
                               thread,
                               StepRequest.STEP_LINE,
                               StepRequest.STEP_OUT
                           );
            findSourceSR.addCountFilter (1);
            findSourceSR.setSuspendPolicy (EventRequest.SUSPEND_ALL);
            operator.register (findSourceSR, this);
            findSourceSR.enable ();
        } catch (DuplicateRequestException e) {
            e.printStackTrace ();
        }

        // create Method Entry Request for searching a source code
        findSourceMER = requestManager.createMethodEntryRequest ();
        findSourceMER.setSuspendPolicy (EventRequest.SUSPEND_ALL);
        findSourceMER.addThreadFilter (thread);
        findSourceCounter = 0;
        operator.register (findSourceMER, new Executor () {
                               public void exec (com.sun.jdi.event.Event event) {
                                   if (findSourceCounter == 1500) {
                                       // finding source takes a long time
                                       //S ystem.out.println();
                                       //S ystem.out.println("    Stop finding");
                                       operator.resume ();
                                       if (findSourceMER != null) {
                                           requestManager.deleteEventRequest (findSourceMER);
                                           findSourceMER = null;
                                       }
                                       return;
                                   }
                                   findSourceCounter++;

                                   com.sun.jdi.Location loc = ((MethodEntryEvent) event).location ();
                                   if (loc == null) {
                                       // no line => continue finding
                                       operator.resume ();
                                       return;
                                   }
                                   String className = loc.declaringType ().name ();
                                   int ln = loc.lineNumber ();
                                   //S ystem.out.println ("    TRY " + findSourceCounter + " : " + className + " : " + ln); // NOI18N
                                   try {
                                       Line l = null;
                                       if ( (l = Utils.getLineForSource (className, loc.sourceName (), ln))
                                               == null
                                          ) {
                                           // no line => continue finding
                                           operator.resume ();
                                           return;
                                       }

                                       // WOW I have a nice line!
                                       //S ystem.out.println ("    WOW! "); // NOI18N
                                       ThreadReference tr = ((MethodEntryEvent) event).thread ();
                                       if (resolveCanBeCurrent (tr, l))
                                           // if can not be current => steps to some line
                                           return;

                                       // line can be current!
                                       String threadName = tr.name ();
                                       String methodName = loc.method () != null ? loc.method ().name () : ""; // NOI18N
                                       String lineNumber = ln == -1 ? "?" : "" + ln; // NOI18N
                                       makeCurrent (
                                           threadName,
                                           className,
                                           methodName,
                                           lineNumber,
                                           true,
                                           tr
                                       );
                                       operator.stopRequest ();
                                   } catch (AbsentInformationException e) {
                                   }
                               }
                           });
        findSourceMER.enable ();

        operator.resume ();
        return;
    }

    /**
    * if this line can not be current => stepOver & return true.
    * return false on the other hand.
    */
    boolean resolveCanBeCurrent (ThreadReference tr) {
        try {
            com.sun.jdi.Location l = tr.frame (0).location ();
            if (l == null) return false;
            return resolveCanBeCurrent (
                       tr,
                       Utils.getLineForSource (
                           l.declaringType ().name (),
                           l.sourceName (),
                           l.lineNumber ()
                       )
                   );
        } catch (Exception e) {
        }
        return false;
    }

    /**
    * If this line can not be current => stepOver & return true.
    * {support for non java languages}
    *
    * return false on the other hand.
    */
    boolean resolveCanBeCurrent (ThreadReference tr, Line l) {
        if ( (l != null) &&
                (!canBeCurrent (l, false))
           ) {
            try {
                removeStepRequest ();
                findSourceSR = requestManager.createStepRequest (
                                   tr,
                                   StepRequest.STEP_LINE,
                                   StepRequest.STEP_OVER
                               );
                findSourceSR.addCountFilter (1);
                findSourceSR.setSuspendPolicy (EventRequest.SUSPEND_ALL);
                operator.register (findSourceSR, this);
                findSourceSR.enable ();
                operator.resume ();
            } catch (DuplicateRequestException e) {
                e.printStackTrace ();
            }
            return true;
        }
        return false;
    }

    /**
    * Sets curent line. It means: change debugger state to stopped, 
    * shows message, sets current thread and updates watches.
    */
    private void makeCurrent (
        final String  threadName,
        final String  className,
        final String  methodName,
        final String  lineNumber,
        final boolean hasSource,
        final ThreadReference tr
    ) {
        setState (DEBUGGER_STOPPED);
        final IOManager ioManager = getIOManager ();
        SwingUtilities.invokeLater (new Runnable () {
            public void run () {
                // show message
                if (hasSource)
                    ioManager.println (
                        new MessageFormat (getLocString ("CTL_Thread_stopped")).
                        format (
                            new Object[] {
                                threadName,
                                className,
                                methodName,
                                lineNumber
                            }
                        ),
                        IOManager.DEBUGGER_OUT + IOManager.STATUS_OUT
                    );
                else
                    ioManager.println (
                        new MessageFormat (getLocString
                          ("CTL_Thread_stopped_no_source")).format (
                            new Object[] {
                                threadName,
                                className,
                                methodName,
                                lineNumber
                            }
                        ),
                        IOManager.DEBUGGER_OUT + IOManager.STATUS_OUT
                    );

                // refresh all
                JPDAThread tt = threadManager.getThread (tr);
                tt.setCurrent (true);
                updateWatches ();
                threadManager.refreshStacks ();
            } // method run ()
        });
    }

    /**
    * Second part of debugger start procedure.
    */
    private void startDebugger () {
        threadManager = new ThreadManager (this);
        
        // display a warning if variable breakpoints are not supported and some should be set
        boolean access = virtualMachine.canWatchFieldAccess ();
        boolean modif = virtualMachine.canWatchFieldModification ();
        if (!access || !modif) {
            try {
                AbstractDebugger deb = (AbstractDebugger) TopManager.
                    getDefault ().getDebugger ();
                Breakpoint [] breakpoint = deb.getBreakpoints ();
                for (int x = 0; x < breakpoint.length; x++)
                    if (((CoreBreakpoint) breakpoint [x]).isEnabled ()) {
                        CoreBreakpoint.Event event = ((CoreBreakpoint) 
                            breakpoint [x]).getEvent (deb);
                        if (event instanceof VariableBreakpoint)
                            if (!access && !modif) {
                                getIOManager ().println (getLocString ("CTL_Break_on_field_not_supported"), IOManager.DEBUGGER_OUT);
                                break;
                            }
                            else if ((((VariableBreakpoint) event).getType () == VariableBreakpoint.ACCESS_TYPE) && !access) {
                                getIOManager ().println (getLocString ("CTL_Break_on_field_access_not_supported"), IOManager.DEBUGGER_OUT);
                                break;
                            }
                            else if ((((VariableBreakpoint) event).getType () == VariableBreakpoint.MODIFICATION_TYPE) && !modif) {
                                getIOManager ().println (getLocString ("CTL_Break_on_field_modification_not_supported"), IOManager.DEBUGGER_OUT);
                                break;
                            }
                    }
            } catch (DebuggerException e) {
            }
        }
        
        refreshBreakpoints ();
        updateWatches ();
        getIOManager ().println (getLocString ("CTL_Debugger_running"), IOManager.STATUS_OUT); // NOI18N
        setState (DEBUGGER_RUNNING);

        virtualMachine.resume ();

        // start refresh thread .................................................
        if (debuggerThread != null) debuggerThread.stop ();
        debuggerThread = new Thread (
            new Runnable () {
                public void run () {
                    for (;;) {
                        try {
                            Thread.sleep (5000);
                        } catch (InterruptedException ex) {}
                            if (getState () == DEBUGGER_RUNNING)
                            threadManager.refreshThreads ();
                    }
                }
            }, 
            "Debugger refresh thread"  // NOI18N
        );
        debuggerThread.setPriority (Thread.MIN_PRIORITY);
        debuggerThread.start ();
    }

    /**
    * Removes last step request.
    */
    void removeStepRequest () {
        try {
            if (stepRequest != null) {
                requestManager.deleteEventRequest (stepRequest);
                stepRequest = null;
            }
            if (findSourceMER != null) {
                requestManager.deleteEventRequest (findSourceMER);
                findSourceMER = null;
            }
            if (findSourceSR != null) {
                requestManager.deleteEventRequest (findSourceSR);
                findSourceSR = null;
            }
        } catch (VMDisconnectedException e) {
        }
    }

    private static String generatePassword () {
        StringBuffer sb = new StringBuffer ();
        for (int i = 0; i < 4; i++)
            sb.append ((char) (random.nextInt (26) + 'a'));
        return new String (sb);
    }
    
    private VirtualMachine launch (DebuggerInfo info) throws DebuggerException {
        try {
            if (launcher == null)
                launcher = new Launcher (getIOManager ());
            VirtualMachine vm = launcher.launch (info);
            process = launcher.process ();
            connector = launcher.attachingConnector ();
            String address = launcher.address ();
            args = connector.defaultArguments ();
            if (hasShmemTransport (connector)) // NOI18N
                ((Argument) args.get ("name")).setValue (address); // NOI18N
            else
                ((Argument) args.get ("port")).setValue (address); // NOI18N
            return vm;
        } catch (DebuggerException e) {
            finishDebugger ();
            throw e;
        }
    }

    private VirtualMachine reconnect (ReconnectDebuggerInfo info)
    throws DebuggerException {
        return connect (
                   "CTL_Reconnecting_to", // NOI18N
                   info.getConnector (),
                   info.getArgs ()
               );
    }

    private VirtualMachine connect (RemoteDebuggerInfo info)
    throws DebuggerException {
        return connect (
            "CTL_Connecting_to", // NOI18N
            connector = info.getConnector (),
            args = info.getArgs ()
        );
    }

    private VirtualMachine connect (
        String bndlPrefix,
        AttachingConnector connector,
        Map args
    ) throws DebuggerException {
        if (bndlPrefix != null) {
            if (hasShmemTransport (connector)) {
                Argument a = (Argument) args.get ("name"); // NOI18N
                if (a == null)
                    getIOManager ().println (getLocString (bndlPrefix + "_shmem_noargs"), IOManager.DEBUGGER_OUT); // NOI18N
                else
                    getIOManager ().println (
                        new MessageFormat (getLocString (bndlPrefix + "_shmem")). // NOI18N
                        format (new Object[] {
                                    a.value ()
                                }),
                        IOManager.DEBUGGER_OUT
                    );
            } else
                if (hasSocketTransport (connector)) { // NOI18N
                    Argument name = (Argument) args.get ("hostname"); // NOI18N
                    Argument port = (Argument) args.get ("port"); // NOI18N
                    if ((name == null) || (port== null))
                        getIOManager ().println (getLocString (bndlPrefix + "_socket_noargs"), IOManager.DEBUGGER_OUT); // NOI18N
                    else
                        getIOManager ().println (
                            new MessageFormat (getLocString (bndlPrefix + "_socket")). // NOI18N
                            format (new Object[] {
                                        name.value (),
                                        port.value ()
                                    }),
                            IOManager.DEBUGGER_OUT
                        );
                } else
                    getIOManager ().println (getLocString (bndlPrefix), IOManager.DEBUGGER_OUT);
        }
        // launch VM
        try {           //S ystem.out.println ("attach to:" + ac + " : " + password); // NOI18N
            return connector.attach (args);
        } catch (Exception e) {
            finishDebugger ();
            throw new DebuggerException (
                new MessageFormat (getLocString ("EXC_While_connecting_to_debuggee")). // NOI18N
                format (new Object[] {
                            e.toString ()
                        }),
                e
            );
        }
    }

    /**
    * Performs stop action.
    */
    void stop (boolean stop, final AbstractThread thread) {
        final ResourceBundle bundle = NbBundle.getBundle (JPDADebugger.class);
        if (stop) {
            removeStepRequest ();
            setLastAction (ACTION_BREAKPOINT_HIT);
            setState (DEBUGGER_STOPPED);
            operator.stopRequest ();
            SwingUtilities.invokeLater (new Runnable () {
                public void run () {
                    thread.setCurrent (true);
                    updateWatches ();
                    threadManager.refreshStacks ();
                }
            });
        } else
            operator.resume ();
    }

    boolean isDisconnected () {
        if (operator == null)
            return false;
        return disconnected || operator.isDisconnected ();
    }
    
    /**
    * Setter method for debugger state property.
    *
    * @param newState
    */
    synchronized public void setState (final int newState) {
        super.setState (newState);
    }

    // helper methods ..........................................................
    
    private static boolean hasSocketTransport (Connector connector) {
        return connector.transport ().name ().toLowerCase ().indexOf ("socket") > -1;
    }
    
    private static boolean hasShmemTransport (Connector connector) {
        return connector.transport ().name ().toLowerCase ().indexOf ("shmem") > -1;
    }
    
    // innerclasses ............................................................

    private class ReconnectDebuggerInfo extends RemoteDebuggerInfo {
        
        private ReconnectDebuggerInfo (
            AttachingConnector connector,
            Map args
        ) {
            super (connector, args);
        }
    }
}