/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.*;
import java.util.*;

import com.sun.jdi.*;

import javax.swing.SwingUtilities;

import org.openide.debugger.DebuggerException;
import org.openide.text.Line;
import org.openide.util.RequestProcessor.Task;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Utils;
import org.netbeans.modules.debugger.support.util.Protector;

/**
* JPDA Implementation of debugger thread.
*/
public class JPDAThread extends JavaThread implements MonitorInfoProducer {

    // variables .......................................................................

    private JPDADebugger                debugger;
    private ThreadReference             thread;
    private boolean                     current = false;
    /** Cache for old locals value. */
    private HashMap                     oldLocals = new HashMap ();
    /** Cache for old call stack. */
    private JPDACallStackFrame []       oldCallStack = new JPDACallStackFrame [0];
    /** Old current class */
    private Object                      oldCurrentClass = null;
    /** Cache for old context static variables */
    private AbstractVariable []         oldContextFields;

    // init ............................................................................

    JPDAThread (
        JPDADebugger debugger,
        JPDAThreadGroup parentThreadGroup,
        ThreadReference thread
    ) {
        super (debugger, parentThreadGroup);
        this.thread = thread;
        this.debugger = debugger;
    }


    // implementation of DebuggerThread ................................................

    /**
    * Getter for the name of thread property.
    *
    * @throw DebuggerException if some problem occurs.
    * @return name of thread.
    */
    public String getName () {
        Task t = Protector.register ("JPDAThread.getName"); // NOI18N
        try {
            if (isDead ()) return "Thread"; // NOI18N
            return thread.name ();
        } catch (Exception e) {
            return "Thread"; // NOI18N
        } finally {
            t.cancel ();
        }
    }

    /**
    * If this thread is suspended returns line number where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return line number where this thread is stopped.
    */
    public int getLineNumber () {
        Task t = Protector.register ("JPDAThread.getLineNumber"); // NOI18N
        try {
            if (isDead ()) return -1;
            if (thread.frameCount () < 1)
                return -1;
            return thread.frame (0).location ().lineNumber ();
        } catch (Exception e) {
            return -1;
        } finally {
            t.cancel ();
        }
    }

    /**
    * Returns string representation of the current state of this thread (depends on
    * debugger implementation).
    *
    * @throw DebuggerException if some problem occurs.
    * @return string representation of the current state of this thread.
    */
    public String getState () {
        Task t = Protector.register ("JPDAThread.getState"); // NOI18N
        try {
            int s = thread.status ();
            switch (s) {
            case ThreadReference.THREAD_STATUS_UNKNOWN: // [PENDING] - localization!!!
                return "Unknown"; // NOI18N
            case ThreadReference.THREAD_STATUS_ZOMBIE:
                return "Zombie"; // NOI18N
            case ThreadReference.THREAD_STATUS_RUNNING:
                return "Running"; // NOI18N
            case ThreadReference.THREAD_STATUS_SLEEPING:
                return "Sleeping"; // NOI18N
            case ThreadReference.THREAD_STATUS_MONITOR:
                return "Monitor"; // NOI18N
            case ThreadReference.THREAD_STATUS_WAIT:
                return "Wait"; // NOI18N
            case ThreadReference.THREAD_STATUS_NOT_STARTED:
                return "Not started"; // NOI18N
            }
            return "Unknown"; // NOI18N
        } catch (Exception e) {
            return "Undefined"; // NOI18N
        } finally {
            t.cancel ();
        }
    }

    /**
    * If this thread is suspended returns class name where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return class name where this thread is stopped.
    */
    public String getClassName () {
        Task t = Protector.register ("JPDAThread.getClassName"); // NOI18N
        try {
            if (isDead ()) return ""; // NOI18N
            if (thread.frameCount () < 1)
                return ""; // NOI18N
            return thread.frame (0).location ().declaringType ().name ();
        } catch (Exception e) {
            return ""; // NOI18N
        } finally {
            t.cancel ();
        }
    }

    /**
    * If this thread is suspended returns method name where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return method name where this thread is stopped.
    */
    public String getMethod () {
        Task t = Protector.register ("JPDAThread.getMethod"); // NOI18N
        try {
            if (isDead ()) return ""; // NOI18N
            if (thread.frameCount () < 1)
                return ""; // NOI18N
            Method m = thread.frame (0).location ().method ();
            if (m == null)
                return ""; // NOI18N
            return m.name ();
        } catch (Exception e) {
            return ""; // NOI18N
        } finally {
            t.cancel ();
        }
    }

    /**
    * Returns current stack depth.
    *
    * @throw DebuggerException if some problem occurs.
    * @return current stack depth.
    */
    public int getStackDepth () {
        Task t = Protector.register ("JPDAThread.getStackDepth"); // NOI18N
        try {
            if (isDead ()) return 0;
            return thread.frameCount ();
        } catch (Exception e) {
            return 0;
        } finally {
            t.cancel ();
        }
    }

    /**
    * Returns true if this thread is suspended.
    *
    * @throw DebuggerException if some problem occurs.
    * @return true if this thread is suspended.
    */
    public boolean isSuspended () {
        Task t = Protector.register ("JPDAThread.isSuspended"); // NOI18N
        try {
            return thread.isSuspended ();
        } catch (Exception e) {
            return false;
        } finally {
            t.cancel ();
        }
    }

    /**
    * Setter method for the suspend property.
    *
    * @throw DebuggerException if some problem occurs.
    * @param suspend true if this thread might be suspend.
    */
    public void setSuspended (final boolean suspend) {
        Task t = Protector.register ("JPDAThread.setSuspended"); // NOI18N
        try {
            if (isDead ()) return;
            if (thread.name ().equals ("Thread-0")) { // NOI18N
                ThreadGroupReference tg = thread.threadGroup ();
                if ((tg != null) && (tg.name ().equals ("system"))) // NOI18N
                    return;
            }
            if (suspend) {
                if (!isSuspended ()) {
                    thread.suspend ();
                    // firePropertyChange (null, null, null);
                }
            } else
                if (isSuspended ()) {
                    debugger.removeStepRequest ();
                    thread.resume ();
                    // firePropertyChange (null, null, null);
                }
            super.setSuspended (suspend);
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
    }

    void markSuspended (final boolean suspend) throws DebuggerException {
        if (!getState ().equals ("Zombie")) { //NOI18N
            if (!suspend && isSuspended ())
                debugger.removeStepRequest ();
            super.setSuspended (suspend);
        }
    }
    
    /**
    * Stops this thread.
    *
    * @throw DebuggerException if some problem occurs.
    */
    public void stop () {
        Task t = Protector.register ("JPDAThread.stop"); // NOI18N
        try {
            if (isDead ()) return;
            thread.stop (null);
            firePropertyChange (null, null, null);
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
    }

    /**
    * If this thread is suspended returns current call stack.
    *
    * @return current call stack.
    */
    public org.netbeans.modules.debugger.Location[] getCallStack () {
        Task t = Protector.register ("JPDAThread.getCallStack"); // NOI18N
        try {
            if (isDead ()) return new JPDACallStackFrame [0];
            List l = thread.frames ();
            JPDACallStackFrame[] sfs = new JPDACallStackFrame [l.size ()];
            int i, k = Math.min (sfs.length, oldCallStack.length);
            for (i = 0; i < k; i++) {
                sfs [sfs.length-i-1] = new JPDACallStackFrame (debugger, thread, (sfs.length-i-1));
                sfs [sfs.length-i-1].oldLocals = oldCallStack [oldCallStack.length-i-1].oldLocals;
            }
            for (i = k; i < sfs.length; i++)
                sfs [sfs.length-i-1] = new JPDACallStackFrame (debugger, thread, (sfs.length-i-1));
            oldCallStack = sfs;
            return sfs;
        } catch (Exception e) {
            oldCallStack = new JPDACallStackFrame [0];
            return oldCallStack;
        } finally {
            t.cancel ();
        }
    }
    
    /**
    * Returns human-readable text info about currently returned variables 
    * or null. This text should be displayed with list of variables. 
    *
    * @return human-readable text info about currently returned variables 
    * or null
    */
    public String getInfo () {
        try {
            if (isDead () ||
                (thread.frameCount () < 1)
            ) return null;
            
            StackFrame stackFrame = thread.frame (0);
            if (stackFrame.location ().method ().isNative ()) return null;
            stackFrame.visibleVariables ();
            return null;
        } catch (AbsentInformationException e) {
            return NbBundle.getBundle(JPDACallStackFrame.class).getString("CTL_Source_compiled_without_g");
        } catch (IncompatibleThreadStateException e) {
        } catch (Exception e) {
            e.printStackTrace ();
        }
        return null;
    }

    /**
    * If this thread is suspended returns current local variables.
    *
    * @return current local variables.
    */
    public AbstractVariable[] getVariables () {
        Task t = Protector.register ("JPDAThread.getLocals"); // NOI18N
        try {
            if (debugger.isDisconnected () || isDead () ||
                (thread.frameCount () < 1)
            ) return new AbstractVariable [0];
            
            // init
            StackFrame stackFrame = thread.frame (0);
            ObjectReference objectRef = stackFrame.thisObject ();
            HashMap newLocals = new HashMap ();
            int i, k;
            List l = null;
            if (stackFrame.location ().method ().isNative ()) {
                k = 0;
            } else {
                try {
                    l = stackFrame.visibleVariables ();
                    k = l.size ();
                } catch (AbsentInformationException e) {
                    k = 0;
                }
            }
            ArrayList result = new ArrayList ();
            JPDAVariable var;

            // read local values
            for (i = 0; i < k; i++) {
                try {
                    LocalVariable lv = (LocalVariable) l.get (i);
                    var = (JPDAVariable) oldLocals.get (lv.name ());
                    if (var == null)
                        var = new JPDAVariable (
                            debugger,
                            lv.name (),
                            stackFrame.getValue (lv),
                            lv.typeName (),
                            thread,
                            0
                        );
                    else {
                        var.update (
                            lv.name (),
                            stackFrame.getValue (lv),
                            lv.typeName (),
                            thread,
                            0
                        );
                        var.firePropertyChange ();
                    }
                    result.add (var);
                    newLocals.put (lv.name (), var);
                } catch (InconsistentDebugInfoException e) {
                    // do not add this local variable, its value cannot be obtained
                }
            }
            
            // add "this" local variable
            if (objectRef != null) {
                var = (JPDAVariable) oldLocals.get ("this"); // NOI18N
                if (var == null)
                    var = new JPDAVariable (
                        debugger,
                        "this", // NOI18N
                        objectRef,
                        objectRef.type ().name ()
                    );
                else {
                    var.update (
                        "this", // NOI18N
                        objectRef,
                        objectRef.type ().name ()
                    );
                    var.firePropertyChange ();
                }
                result.add (var);
                newLocals.put ("this", var); // NOI18N
            }
            
            oldLocals = newLocals;
            Iterator iter = result.iterator ();
            JPDAVariable [] variables = new JPDAVariable [result.size ()];
            for (int x = 0; x < result.size (); x++)
                variables [x] = (JPDAVariable) iter.next ();
            return variables;
        } catch (IncompatibleThreadStateException e) {
        } catch (Exception e) {
            e.printStackTrace ();
        } finally {
            t.cancel ();
        }
        return new AbstractVariable [0];
    }

    /**
    * If this thread is suspended returns static and instance variables in current context.
    *
    * @return static and instance variables in current context.
    */
    public AbstractVariable[] getContextVariables () {
        AbstractVariable [] result = null;
        Task t = Protector.register ("JPDAThread.getContextVariables"); // NOI18N
        try {
            if (isDead ()) return new AbstractVariable [0];
            ObjectReference objectRef = thread.frame (0).thisObject ();
            if (objectRef != null) {
                if (objectRef.equals (oldCurrentClass))
                    result = oldContextFields;
                else {
                    List list = objectRef.referenceType ().allFields ();
                    result = new AbstractVariable [list.size ()];
                    for (int i = 0; i < list.size (); i++)
                        result [i] = new JPDAVariable (
                                        debugger,
                                        objectRef,
                                        (Field) list.get (i)
                                      );
                    oldCurrentClass = objectRef;
                    oldContextFields = result;
                }
            }
            else { // static context
                ReferenceType reference = thread.frame (0).location ().declaringType ();
                if (reference.equals (oldCurrentClass))
                    result = oldContextFields;
                else {
                    Iterator fields = reference.allFields ().iterator ();
                    LinkedList list = new LinkedList ();
                    Field field;
                    while (fields.hasNext ()) {
                        field = (Field) fields.next ();
                        if (field.isStatic ()) {
                            JPDAVariable var = new JPDAVariable (
                                            debugger,
                                            reference,
                                            field
                                          );
                            list.add (var);
                        }
                    }
                    result = (AbstractVariable []) list.toArray (new AbstractVariable [list.size ()]);      
                    oldCurrentClass = reference;
                    oldContextFields = result;
                }
            }    
        } catch (Exception e) {
            result = new JPDAVariable [0];
        } finally {
            t.cancel ();
        }
        return result;
    }
    
    /**
    * @return variable corresponding to the contended monitor, or null if the thread is not 
    * waiting for a monitor
    */
    public Monitor contendedMonitor () {
        Task t = Protector.register ("JPDAThread.contendedMonitor"); // NOI18N
        try {
            if (isDead ()) return null;
            ObjectReference ref = thread.currentContendedMonitor ();
            if (ref != null)
                return new JPDAMonitor (debugger, ref);
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
        return null;
    }
    
    /**
    * @return an array of owned monitors
    */
    public Monitor [] ownedMonitors () {
        Task t = Protector.register ("JPDAThread.ownedMonitors"); // NOI18N
        try {
            if (isDead ()) return new Monitor [0];
            List list = thread.ownedMonitors ();
            Monitor [] res = new Monitor [list.size ()];
            for (int x = 0; x < list.size (); x++)
                res [x] = new JPDAMonitor (debugger, (ObjectReference) list.get (x));
            return res;
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
        return new Monitor [0];
    }
    
    /**
    * @return variable corresponding to the contended monitor, or null if the thread is not 
    * waiting for a monitor
    */
    public boolean contendsMonitor () {
        Task t = Protector.register ("JPDAThread.contendsMonitor"); // NOI18N
        try {
            if (isDead ()) return false;
            return thread.currentContendedMonitor () != null;
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
        return false;
    }
    
    /**
    * @return an array of owned monitors
    */
    public boolean ownsMonitors () {
        Task t = Protector.register ("JPDAThread.ownsMonitors"); // NOI18N
        try {
            if (isDead ()) return false;
            List list = thread.ownedMonitors ();
            return list.size () > 0;
        } catch (Exception e) {
        } finally {
            t.cancel ();
        }
        return false;
    }
    
    /**
    * Returns name of file of this frame or null if thread has no frame.
    *
    * @return Returns name of file of this frame.
    * @throws DebuggerException if informations about source are not included or some other error
    *   occurres.
    */
    public String getSourceName () {
        Task t = Protector.register ("JPDAThread.getSourceName"); // NOI18N
        try {
            if (isDead ()) return ""; // NOI18N
            if (thread.frameCount () < 1) return null;
            return thread.frame (0).location ().sourceName ();
        } catch (Exception e) {
            return null;
        } finally {
            t.cancel ();
        }
    }

    public ThreadReference getThreadReference () {
        return thread;
    }

    /**
    * Refresh of thread properties - fires changes.
    */
    void refresh () {
        firePropertyChange (null, null, null);
    }

    /**
    * Refresh of stack, locals and properties - fires changes.
    */
    void refreshStack () {
        firePropertyChange (CallStackProducer.PROP_CALLSTACK, null, null);
        firePropertyChange (VariablesProducer.PROP_VARIABLES, null, null);
        firePropertyChange (MonitorInfoProducer.PROP_MONITORS, null, null);
        refresh ();
    }

    boolean isDead () {
        if (debugger.isDisconnected ())
            return true;
        try {
            int s = thread.status ();
            switch (s) {
            case ThreadReference.THREAD_STATUS_UNKNOWN:
            case ThreadReference.THREAD_STATUS_RUNNING:
            case ThreadReference.THREAD_STATUS_SLEEPING:
            case ThreadReference.THREAD_STATUS_MONITOR:
            case ThreadReference.THREAD_STATUS_WAIT:
                return false;
            case ThreadReference.THREAD_STATUS_NOT_STARTED:
            case ThreadReference.THREAD_STATUS_ZOMBIE:
                return true;
            }
            return false;
        } catch (Exception e) {
            return true;
        }
    }
    
    public String toString () {
        try {
            return "Thread: " + getName () + " (" + super.toString () + ")"; // NOI18N
        } catch (Exception e) {
            return super.toString ();
        }
    }
}
