/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.Modifier;
import java.util.ResourceBundle;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

import javax.swing.SwingUtilities;

import com.sun.jdi.Value;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.ArrayReference;
import com.sun.jdi.StringReference;
import com.sun.jdi.Field;
import com.sun.jdi.Method;
import com.sun.jdi.LocalVariable;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.ClassType;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.StackFrame;
import com.sun.jdi.InvalidTypeException;
import com.sun.tools.example.debug.expr.ExpressionParser;
import com.sun.tools.example.debug.expr.ParseException;

import org.openide.*;
import org.openide.util.NbBundle;
import org.openide.debugger.Watch;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.debugger.DebuggerException;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.util.Protector;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* JPDA implemetation of variable.
*
* @author   Jan Jancura
*/
public class JPDAVariable implements JavaVariable {

    // static ..................................................................
    
    /** generated Serialized Version UID */
    static final long serialVersionUID = -4908841115435123749L;
    /** bundle to obtain text information from */
    private static ResourceBundle bundle;
    
    private static void notify (String msg) {
        TopManager.getDefault ().notify (new NotifyDescriptor.Message (
            msg
        ));
    }
    
    private static String getString (String msg) {
        if (bundle == null)
            bundle = NbBundle.getBundle (JPDAVariable.class);
        return bundle.getString (msg);
    }

    
    // variables .........................................................................

    /** Variables parentObject & (name | index) identifies variable. */
    private transient Object                      parentObject; // ObjectReference or ReferenceType
    /** Current value - used for getFields (). */
    protected transient Value                     remoteValue;
    /** Current value - used for getFields (). */
    protected transient String                    name;
    /** Current value - used for getFields (). */
    protected transient String                    value;
    /** Current value - used for getFields (). */
    protected transient String                    innerType;
    /** Current value - used for getFields (). */
    protected transient String                    type;
    /** Current value - used for getFields (). */
    protected transient String                    modifiers;
    /** Current field. */
    protected transient Field                     remoteField;
    /** Thread and frame index for locals. */
    private transient ThreadReference             threadS = null;
    private transient int                         indexS;
    /** Cashing of children variables */
    private transient ObjectReference             oldObject;
    /** Cashing of children variables */
    private transient AbstractVariable[]          oldFields;
    /** Cashing of children variables */
    private transient int                         oldStartIndex;
    /** Cashing of children variables */
    private transient int                         oldLength;
    protected transient boolean                   isCloned;
    private transient boolean                     isArray;
    private transient boolean                     isObject;
    private transient Validator                   validator;
    private transient int                         index = -1;
    
    private transient PropertyChangeSupport       pcs;
    protected transient JPDADebugger              debugger;
    


    // init ...............................................................................

    /**
    * Non public constructor.
    */
    JPDAVariable (JPDADebugger debugger, boolean validate) {
        validator = validate ? debugger.getValidator () : null;
        if (validate) {
            validator = debugger.getValidator ();
            validator.add (this);
        }
        this.debugger = debugger;
        pcs = new PropertyChangeSupport (this);
    }

    /**
    * Creates variable.
    */
    JPDAVariable (
        JPDADebugger     debugger,
        ObjectReference  parentObject,
        Field            field
    ) {
        this (debugger, true);
        update (field, parentObject);
        //S ystem.out.println("new JPDAVariable " + getVariableName ());
    }
   
    /**
    * Creates variable, used in JPDAThread.
    */
    JPDAVariable (
        JPDADebugger     debugger,
        ReferenceType    parentObject,
        Field            field
    ) {
        this (debugger, true);
        update (field, parentObject);
        //S ystem.out.println("new JPDAVariable " + getVariableName ());
    }
    
    /**
    * Creates local variable. Called from thread, LineBreakpoint ...
    */
    JPDAVariable (
        JPDADebugger  debugger,
        String        name,
        Value         value,
        String        type
    ) {
        this (debugger, true);
        update (name, value, type);
        //S ystem.out.println("new JPDAVariable " + getVariableName ());
    }

    /**
    * Creates local variable. Called from thread.
    */
    JPDAVariable (
        JPDADebugger  debugger,
        String        name,
        Value         value,
        String        type,
        ThreadReference threadS,
        int           indexS
    ) {
        this (debugger, true);
        update (name, value, type, threadS, indexS);
        //S ystem.out.println("new JPDAVariable " + getVariableName ());
    }

    /**
    * Creates member varaible of array.
    */
    private JPDAVariable (
        JPDADebugger  debugger,
        ArrayReference array,
        String        parentName,
        int           index,
        String        type
    ) {
        this (debugger, true);
        update (parentName, array, index, type);
        //S ystem.out.println("new JPDAVariable " + getVariableName ());
    }

    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        try {
            debugger = (JPDADebugger) TopManager.getDefault ().getDebugger ();
        } catch (DebuggerNotFoundException e) {
            throw new java.io.IOException ();
        }
    }


    // JavaVariable implementation ...........................................

    // AbstractVariable
    
    /**
    * Returns the name of this variable.
    *
    * @return the name of this variable.
    */
    public String getVariableName () {
        return name;
    }

    /**
    * Getter for textual representation of the value. It converts
    * the value to a string representation. So if the watch represents
    * null reference, the returned string will be for example "null".
    * That is why null can be returned when the watch is not valid
    *
    * @return the value of this watch or null if the watch is not in the scope
    */
    public String getAsText () {
        return value;
    }

    /**
    * Setter that allows to change value of the watched variable.
    *
    * @param value text representation of the value
    * @exception DebuggerException if the value cannot be changed or the
    *    string does not represent valid value
    */
    public void setAsText (final String value) {
        if (debugger.getState () != debugger.DEBUGGER_STOPPED) return;
        if (isCloned) return; // fixed watch
        final JPDAThread tt = (JPDAThread) debugger.getCurrentThread ();
        if (tt == null) return;
        ThreadReference rt = tt.getThreadReference ();
        final StackFrame sf;
        try {
            sf = rt.frame (0);
        } catch (Exception e) {
            return;
        }
        // resolve RVALUE ......................................................
        Value v = null;
        try {
            v = (Value) new Protector ("JPDAVariable.ExpressionParser") { // NOI18N
                    public Object protect () throws Exception {
                        Value vv = ExpressionParser.evaluate (
                            value,
                            debugger.virtualMachine,
                            new ExpressionParser.GetFrame () {
                                public StackFrame get () {
                                    return sf;
                                }
                            }
                        );
                        return vv;
                    }
                }.throwAndWait (null);
        } catch (ParseException e) {
            // a pacth for static variables ....................................
            try {
                ReferenceType refType = sf.location ().declaringType ();
                v = refType.getValue (refType.fieldByName (value));
            } catch (Exception  ee) {
                notify (
                    getString ("EXC_Value_cannt_be_set_info") + //NOI18N
                    " " + e.getMessage () //NOI18N
                );
                return;
            }
            // end of patch ....................................................
        } catch (Throwable e) {
            notify (
                getString ("EXC_Value_cannt_be_set_info") + //NOI18N
                " " + e.getMessage () //NOI18N
            );
            return;
        }
        
        // set value v
        // error codes
        final int RES_OK = 0; // ok
        final int RES_ERR = 1; // cannot set 
        final int RES_EXPR = 2; // watch is an expression
        // find the variable
        Integer result = null;
        final Value v2 = v;
        try {
            final int ind = index;
            result = (Integer) new Protector ("JPDAVariable.ExpressionParser") { // NOI18N
                 public Object protect () throws Exception {
                     if (threadS != null) { // locals
                         StackFrame stackFrame = threadS.frame (indexS);
                         LocalVariable localVariable = stackFrame.visibleVariableByName (
                             getVariableName ()
                         );
                         if (localVariable != null)
                             stackFrame.setValue (localVariable, v2);
                         else // local variable cannot be found
                             return new Integer (RES_ERR);
                     } else {
                         if (parentObject == null) { // watch
                             LocalVariable localVariable = sf.visibleVariableByName (
                                 getVariableName ()
                             );
                             if (localVariable != null)
                                 sf.setValue (localVariable, v2);
                             else {
                                 ObjectReference obj = sf.thisObject ();
                                 ReferenceType ref;
                                 if (obj != null)
                                     ref = obj.referenceType ();
                                 else
                                     ref = sf.location ().declaringType ();
                                 Field field = ref.fieldByName (getVariableName ());
                                 if (field != null) {
                                     if (obj != null)
                                        obj.setValue (field, v2);
                                     else
                                        ((ClassType) ref).setValue (field, v2);
                                 }
                                 else // variable is an expression
                                     return new Integer (RES_EXPR);
                             }
                         }
                         else { // parentObject != null
                             if (parentObject instanceof ArrayReference)
                                 ((ArrayReference) parentObject).setValue (ind, v2);
                             else if (parentObject instanceof ObjectReference)
                                 ((ObjectReference) parentObject).setValue (remoteField, v2);
                             else if (parentObject instanceof ClassType)
                                 ((ClassType) parentObject).setValue (remoteField, v2);
                             else
                                 return new Integer (RES_ERR);
                         }
                     }
                     return new Integer (RES_OK);
                 }
             }.throwAndWait (null);
        } catch (InvalidTypeException e) {
            notify (getString ("EXC_Wrong_type"));
            return;
        } catch (Exception e) {
            notify (getString ("EXC_Value_cannt_be_set"));
            return;
        }
        switch (result.intValue ()) {
            case RES_ERR:
                notify (getString ("EXC_Value_cannt_be_set"));
                return;
            case RES_EXPR:
                // variable is an expression
                notify (getString ("EXC_Cannot_assign_to_expression"));
                return;
        }
        if (v == null) {
            // setNull
            remoteValue = null;
            type = ""; // NOI18N
            update ();
            // end of setNull
        }
        else {
            // update
            remoteValue = v;
            type = v.type ().name ();
            update ();
            // end of update
        }
        if (pcs != null)
            pcs.firePropertyChange (null, null, null);
    }

    /**
    * Returns string representation of type of this variable.
    *
    * @return string representation of type of this variable.
    */
    public String getType () {
        return type;
    }

    /**
    * Adds listener on the property changing.
    */
    public void addPropertyChangeListener (PropertyChangeListener listener) {
        pcs.addPropertyChangeListener (listener);
    }

    /**
    * Removes listener on the property changing.
    */
    public void removePropertyChangeListener (PropertyChangeListener listener) {
        pcs.removePropertyChangeListener (listener);
    }

    // VariableProducer

    /**
    * Returns true if this variable hasn't any fields.
    *
    * @return True if this variable hasn't any fields.
    */
    public boolean isLeaf () {
        return (remoteValue == null) || !(remoteValue instanceof ObjectReference);
    }

    /**
    * If this AbstractVariable object represents instance of some class or array this method
    * returns variables (static and non-static) of this object.
    *
    * @return variables (static and non-static) of this object.
    */
    public AbstractVariable[] getFilteredVariables () {
        return getVariables ();
    }
    
    /**
    * If this AbstractVariable object represents instance of some class or array this method
    * returns variables (static and non-static) of this object.
    *
    * @return variables (static and non-static) of this object.
    */
    public AbstractVariable[] getVariables () {
        final String name = this.name;
        //S ystem.out.println ("getFields " + remoteValue + " " + oldObject); // NOI18N
        if ((remoteValue == null) ||
                !(remoteValue instanceof ObjectReference)
           ) {
            //S ystem.out.println ("getFields prim."); // NOI18N
            return new JPDAVariable [0];
        }
        if ((oldObject != null) && oldObject.equals (remoteValue)) {
            // the same set of children => validate only
            //S ystem.out.println ("getFields return old " + oldFields + " " + oldFields.length); // NOI18N
            //      int i, k = oldFields.length;
            //      for (i = 0; i < k; i++) oldFields [i].validate ();
            return oldFields;
        }
        try {
            AbstractVariable[] variable;
            ObjectReference remoteObject = (ObjectReference)remoteValue;
            if (remoteValue instanceof ArrayReference) {
                final String type = innerType.substring (0, innerType.length () - 2);
                ArrayReference array = (ArrayReference) remoteValue;
                int i, k = array.getValues ().size ();
                variable = new AbstractVariable [k];
                for (i = 0; i < k; i++)
                    variable [i] = new JPDAVariable (
                                       debugger,
                                       array,
                                       name,
                                       i,
                                       type
                                   );
                //S ystem.out.println ("getFields return array " + variable + " " + variable.length); // NOI18N
            } else {
                List fields = ((ObjectReference) remoteValue).referenceType ().allFields ();
                int i, k = fields.size ();
                variable = new AbstractVariable [k];
                for (i = 0; i < k; i++)
                    variable [i] = new JPDAVariable (
                                       debugger,
                                       remoteObject,
                                       (Field) fields.get (i)
                                   );
                //S ystem.out.println ("getFields return fields " + variable + " " + variable.length); // NOI18N
            }
            oldObject = remoteObject;
            oldFields = variable;
            return variable;
        } catch (Exception e) {
            //e.printStackTrace ();
            return new JPDAVariable [0];
        }
    }

    // JavaVariable
    
    /**
    * Returns string representation of inner type of this variable.
    *
    * @return string representation of inner type of this variable.
    */
    public String getInnerType () {
        return innerType;
    }

    /**
    * Returns true, if this variable do not represents primitive type.
    *
    * @return true, if this variable do not represents primitive type.
    */
    public boolean isObject () {
        return isObject;
    }

    /**
    * Returns true, if this variable do not represents primitive type.
    *
    * @return true, if this variable do not represents primitive type.
    */
    public boolean isArray () {
        return isArray;
    }

    /**
    * Returns modifiers of this variable.
    *
    * @return modifiers of this variable.
    */
    public String getModifiers () {
        return modifiers;
    }
    
    /**
    * Returns human-readable text info about currently returned variables 
    * or null. This text should be displayed with list of variables. 
    *
    * @return human-readable text info about currently returned variables 
    * or null
    */
    public String getInfo () {
        return null;
    }
    
    /**
     * If this AbstractVariable object represents instance of some array this method
     * returns (a part of) its items. This method does not check if the given interval
     * is inside the array!
     *
     * @param startindex index of the first item to be returned
     * @param numberofitems number of items to be returned
     * @return array of the items of the array
     */
    public AbstractVariable[] getVariables (int startindex, int numberofitems)
    {
        if ((remoteValue == null) || !(remoteValue instanceof ArrayReference))
            return getVariables ();
        
        final String name = this.name;
        
        if ((oldObject != null) && oldObject.equals (remoteValue) && oldStartIndex==startindex && oldLength==numberofitems) {
            // the same set of children => validate only
            //S ystem.out.println ("getFields return old " + oldFields + " " + oldFields.length); // NOI18N
            //      int i, k = oldFields.length;
            //      for (i = 0; i < k; i++) oldFields [i].validate ();
            return oldFields;
        }
        
        try {
            AbstractVariable[] variable;
            ObjectReference remoteObject = (ObjectReference)remoteValue;

            final String type = innerType.substring (0, innerType.length () - 2);
            ArrayReference array = (ArrayReference) remoteValue;
            variable = new AbstractVariable [numberofitems];
            for (int i = 0; i < numberofitems; i++)
                variable [i] = new JPDAVariable (
                   debugger,
                   array,
                   name,
                   startindex + i,
                   type
               );
            oldObject = remoteObject;
            oldFields = variable;
            oldStartIndex = startindex;
            oldLength = numberofitems;
            return variable;
        } catch (Exception e) {
            //e.printStackTrace ();
            return new JPDAVariable [0];
        }
    }
    
    /**
     * If this AbstractVariable object represents instance of some array this method
     * returns number of its items.
     *
     * @return number of items of the array
     */
    public int getVariablesNumber ()
    {
        if ((remoteValue == null) || !(remoteValue instanceof ArrayReference))
            return 0;
        return ((ArrayReference)remoteValue).length();
    }


    // other methods ....................................................................................

    /**
    * I am member of an array. 
    * Update modifiers, type, remoteValue, parentObject.
    */
    boolean update (
        String         name,
        ArrayReference array,
        int            index,
        String         type
    ) {
        this.name = name + " [" + index + "]"; // NOI18N
        parentObject = array;
        this.index = index;
        this.type = type;
        modifiers = ""; // NOI18N
        try {
            remoteValue = array.getValue (index);
        } catch (Exception e) {
            return false;
        }
        update ();
        return true;
    }

    /**
    * Init for remoteObject
    * modifiers, type, remoteValue, parentObject
    */
    boolean update (
        Field            field,
        Object           parentObject
    ) {
        this.name = field.name ();
        this.parentObject = parentObject;
        try {
            remoteField = field;
            if (remoteField == null) return false;
            if (parentObject instanceof ObjectReference)
                remoteValue = ((ObjectReference) parentObject).getValue (remoteField);
            else
                remoteValue = ((ReferenceType) parentObject).getValue (remoteField);
            modifiers = Modifier.toString (remoteField.modifiers ());
            type = remoteField.typeName ();
        } catch (Exception e) {
            return false;
        }
        update ();
        return true;
    }

    /**
    * For local variables. (locals or watch)
    */
    void update (
        String        name,
        Value         remoteValue,
        String        type
    ) {
        this.name = name;
        this.remoteValue = remoteValue;
        this.parentObject = null;
        this.type = type;
        modifiers = ""; // NOI18N
        update ();
    }

    /**
    * For local variables, called from thread.
    */
    void update (
        String        name,
        Value         remoteValue,
        String        type,
        ThreadReference threadS,
        int           indexS
    ) {
        this.threadS = threadS;
        this.indexS = indexS;
        update (name, remoteValue, type);
    }

    void setNull () {
        remoteValue = null;
        parentObject = null;
        modifiers = ""; // NOI18N
        type = ""; // NOI18N
        update ();
    }

    protected void setError (String description) {
        remoteValue = null;
        parentObject = null;
    }

    /**
    * 
    */
    public void validate () {
        if (isCloned) {
            if (isObject) {
                value = remoteValue.toString ();
            }
        } else {
            if (index != -1)
                if (parentObject instanceof ArrayReference)
                    try {
                        remoteValue = ((ArrayReference) parentObject).getValue (index);
                    } catch (Exception e) {
                    }
                else remoteValue = null;
            else
                try {
                    if (parentObject instanceof ObjectReference)
                        remoteValue = ((ObjectReference) parentObject).getValue (remoteField);
                    else
                        remoteValue = ((ReferenceType) parentObject).getValue (remoteField);
                } catch (Exception e) {
                }
            update ();
        }

        if (pcs != null) pcs.firePropertyChange (null, null, null);
    }

    /**
    * @return true, variable can be removed from validator when debugger is finished
    */
    public boolean canRemove () {
        return true;
    }

    /**
    * remoteValue => isObject, isArray, value, innerType
    * + fire
    */
    private void update () {
        if (remoteValue == null) {
            isObject = false;
            isArray = false;
            value = null;
            innerType = ""; // NOI18N
        } else
            try {
                isObject = remoteValue instanceof ObjectReference;
                isArray = remoteValue instanceof ArrayReference;
                if (isArray) {
                    value = remoteValue.toString ();
                    innerType = ((ObjectReference)remoteValue).referenceType ().name ();
                } else
                    if (isObject) {
                        value = remoteValue.toString ();
                        innerType = ((ObjectReference)remoteValue).referenceType ().name ();
                    } else {
                        innerType = remoteValue.type ().name ();
                        value = remoteValue.toString ();
                    }
            } catch (Exception e) {
                // exception will be notified
                type = null;
                value = e.toString ();
            }
    }

    void setValue (String v) {
        value = v;
    }

    void firePropertyChange () {
        pcs.firePropertyChange (null, null, null);
    }

    /** 
    * If remoteValue is instanceof ObjectReference,
    * its toString () method is invoked and the appropriate value is returned.
    * Not used yet.
    */
    private String toStringValue () {
        final ThreadReference thread = ((JPDAThread) debugger.getCurrentThread ()).getThreadReference ();
        final ObjectReference obj = (ObjectReference) remoteValue;
        ReferenceType ref = ((ObjectReference) remoteValue).referenceType ();
        
        List list = debugger.virtualMachine.classesByName ("java.lang.Object"); // NOI18N
        ReferenceType refObject = (ReferenceType) list.iterator ().next ();
        list = refObject.methodsByName ("toString"); // NOI18N
        final Method method = (Method) list.iterator ().next ();
        
        final Thread mainThread = Thread.currentThread ();
        final LinkedList result = new LinkedList ();
        boolean passed = false;
        
        Thread evaluator = new Thread () {
            public void run () {
                try {
                    result.add (obj.invokeMethod (thread, method, new LinkedList (), 
                                ObjectReference.INVOKE_SINGLE_THREADED));
                } catch (Exception e) {
                }
                mainThread.interrupt ();
            }
        };
        
        try {
            evaluator.start ();
            Thread.sleep (500); // timeout
        } catch (InterruptedException e) {
            passed = true;
        } catch (ThreadDeath e) {
            evaluator.stop ();
            throw (ThreadDeath) e;
        }
        evaluator.stop ();
        Thread.interrupted ();
        Iterator iter = result.iterator ();
        if (passed && iter.hasNext ())
            return ((StringReference)iter.next ()).value ();
        return "???"; // NOI18N
    }

    public JavaVariable createFixedVariable () {
        try {
            JPDAVariable v = (JPDAVariable) clone ();
            v.isCloned = true;
            return v;
        } catch (CloneNotSupportedException e) {
            return null;
        }
    }
    
    public boolean isFixedVariable () {
        return isCloned;
    }
        
    // Helper methods enabling delegating in JPDAWatch

    JPDADebugger getDebugger () {
        return debugger;
    }

    Value getRemoteValue () {
        return remoteValue;
    }

    void setRemoteValue (Value val) {
        remoteValue = val;
    }

    void setError_protected (String description) {
        setError (description);
    }

/*    java.lang.Object clone_protected () {
        return clone ();
    }*/
}
