/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import com.sun.jdi.Value;
import com.sun.jdi.Field;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.ArrayReference;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.StackFrame;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.ObjectCollectedException;
import com.sun.jdi.InvalidStackFrameException;
import com.sun.jdi.ClassNotPreparedException;
import com.sun.jdi.IncompatibleThreadStateException;
import com.sun.tools.example.debug.expr.ExpressionParser;
import com.sun.tools.example.debug.expr.ParseException;

import org.openide.debugger.Watch;
import org.openide.debugger.DebuggerException;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.util.Protector;


/**
* Standart implementation of Watch interface.
* @see org.openide.debugger.Watch
*
* @author   Jan Jancura
* @version  0.18, Feb 23, 1998
*/
public class JPDAWatch extends JavaWatch {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 3439367144447814302L;


    // private variables .....................................................

    private JPDAVariable                       var;
    private String                             errorMessage;

    // init .....................................................................

    /**
    * Non public constructor called from the JavaDebugger only.
    * User must create watch from Debugger.getNewWatch () method.
    */
    JPDAWatch (JPDADebugger debugger, boolean hidden) {
        super (debugger, hidden);
        var = new JPDAVariable (debugger, false);
    }


    // AbstractWatch implementation .........................................................

    /**
    * Returns true if this variable is in scope.
    *
    * @return true if this variable is in scope.
    */
    public boolean isInScope () {
        return (getErrorMessage () == null);
    }

    /**
    * Create AbstractVariable object for this Watch. Can return null, if this Watch currently not
    * represents valide variable.
    *
    * @return AbstractVariable object for this class.
    */
    public AbstractVariable getVariable () {
        return var;
    }

    /**
    * Returns true if this variable hasn't any fields.
    *
    * @return True if this variable hasn't any fields.
    */
    public boolean isLeaf () {
        return false;
    }

    /**
    * Checks value of this watch and if is valide, and sets it.
    */
    public void validate () {
        //S ystem.out.println ("JPDAWatch.validate " + getVariableName () + " : " + // NOI18N
        //debugger + " : " + debugger.getState ()); // NOI18N
        if ((debugger.getState () == debugger.DEBUGGER_NOT_RUNNING) ||
            (((JPDADebugger) debugger).isDisconnected ())
           ) {
            setErrorMessage (NbBundle.getBundle (JPDAWatch.class).getString ("EXC_No_session"));
            firePropertyChange (null, null, null);
        } else
            if (debugger.getState () != debugger.DEBUGGER_STOPPED) {
                setErrorMessage (NbBundle.getBundle (JPDAWatch.class).getString ("CTL_No_context"));
                firePropertyChange (null, null, null);
            } else {
                final JPDAThread tt = (JPDAThread) ((ThreadsRoot) debugger).getCurrentThread ();
                if (tt != null) {
                    refreshValue (tt);
                } else {
                    setErrorMessage (NbBundle.getBundle (JPDAWatch.class).getString ("CTL_No_context"));
                    firePropertyChange (null, null, null);
                }
            }
    }

    /**
    * @return false, watch cannot be removed from validator when debugger is finished
    */
    public boolean canRemove () {
        return false;
    }

    // methods from VariableImpl

/*    protected java.lang.Object clone () {
        return var.clone_protected ();
    }*/

    public String getAsText () {
        String asText = var.getAsText ();
        if ((asText == null) && (errorMessage == null))
            asText = "null"; // NOI18N
        return asText;
    }

    public String getType () {
        return var.getType ();
    }

    public String getInnerType () {
        return var.getInnerType ();
    }

    public boolean isObject () {
        return var.isObject ();
    }

    public boolean isArray () {
        return var.isArray ();
    }

    public String getModifiers () {
        return var.getModifiers ();
    }

    public String getErrorMessage () {
        return errorMessage;
    }

    public void setAsText (final String value) {
        var.setAsText (value);
    }

    public AbstractVariable[] getVariables () {
        return var.getVariables ();
    }

    public AbstractVariable[] getFilteredVariables () {
        return var.getFilteredVariables ();
    }

    public AbstractVariable[] getVariables (int startindex, int numberofitems) {
        return var.getVariables (startindex, numberofitems);
    }
    
    public int getVariablesNumber () {
        return var.getVariablesNumber ();
    }

    public JavaVariable createFixedVariable () {
        return var.createFixedVariable ();
    }
    
    public boolean isFixedVariable () {
        return var.isFixedVariable ();
    }
    
    // other methods ......................................................................

    /**
    * Refresh value of watch in given context.
    */
    void refreshValue (final JPDAThread tt) {
        ThreadReference rt = tt.getThreadReference ();
        final StackFrame sf;
        try {
            sf = rt.frame (0);
        } catch (Exception e) {
            setErrorMessage (NbBundle.getBundle (JPDAWatch.class).getString ("CTL_No_context"));
            firePropertyChange (null, null, null);
            return;
        }
        setErrorMessage (null);
        Value v;
        try {                                   //S ystem.out.println ("EXPRESSION ENTER " + displayName); // NOI18N
            v = (Value) new Protector ("ExpressionParser") { // NOI18N
                    public Object protect () throws Exception {  //S ystem.out.println ("EXPRESSION ENTER1 " + displayName); // NOI18N
                        return evaluateExpression (sf);
                    }
                }.throwAndWait (null);                //S ystem.out.println ("EXPRESSION EXIT  " + displayName); // NOI18N
        } catch (ParseException e) {            //S ystem.out.println ("EXPRESSION EXIT  " + displayName); // NOI18N
            // a pacth for static variables *******************************************
            v = null;
            boolean found = false;
            try {
                ReferenceType refType = sf.location ().declaringType ();
                v = refType.getValue (refType.fieldByName (displayName));
                found = true;
            } catch (Exception  ee) {
            }
            if (!found) {
                setErrorMessage (e.getMessage ());
                firePropertyChange (null, null, null);
                return;
            }
            // end of patch ***********************************************************
        } catch (ThreadDeath e) {               //S ystem.out.println ("EXPRESSION EXIT  " + displayName); // NOI18N
            setErrorMessage ("Deadlock detected while resolving expression."); // NOI18N
            firePropertyChange (null, null, null);
            return;
        } catch (Throwable e) {                 //S ystem.out.println ("EXPRESSION EXIT  " + displayName); // NOI18N
            setErrorMessage (NbBundle.getBundle (JPDAWatch.class).getString ("CTL_Cannot_evaluate"));
            firePropertyChange (null, null, null);
            return;
        }
        if (v == null)
            setNull ();  // really null value?
        else
            update (
                displayName,
                v,
                v.type ().name ()
            );
        firePropertyChange (null, null, null);
    }

    private Value evaluateExpression (final StackFrame sf) throws Exception {
        return ExpressionParser.evaluate (
           displayName,
           ((JPDADebugger) debugger).virtualMachine,
           new ExpressionParser.GetFrame () {
               public StackFrame get () {
                   return sf;
               }
           }
        ); 
    }

    // Delegating methods for JPDAVariable instance ..................................

    boolean update (String name, ArrayReference array, int index, String type) {
        return var.update (name, array, index, type);
    }

    boolean update (Field field, ObjectReference parentObject) {
        return var.update (field, parentObject);
    }

    void update (String name, Value remoteValue, String type) {
        var.update (name, remoteValue, type);
    }

    void setNull () {
        var.setNull ();
    }

    void setValue (String v) {
        var.setValue (v);
    }

    public String toString () {
        return var.toString ();
    }
    
    protected void setErrorMessage (String msg) {
        errorMessage = msg;
    }
}

