 /*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.util.*;
import java.text.MessageFormat;
import java.io.*;

import com.sun.jdi.connect.*;
import com.sun.jdi.VirtualMachineManager;
import com.sun.jdi.VirtualMachine;
import com.sun.jdi.Bootstrap;

import org.openide.execution.NbProcessDescriptor;
import org.openide.util.MapFormat;
import org.openide.util.NbBundle;
import org.openide.util.Utilities;
import org.openide.debugger.DebuggerException;
import org.openide.debugger.DebuggerInfo;

import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.util.Utils;

public class Launcher {
    
    /** bundle to obtain text information from */
    private static ResourceBundle bundle = org.openide.util.NbBundle.getBundle (JPDADebugger.class);
    
    private ListeningConnector connector = null;
    private Map arguments = null;
    private String transport = null;
    private String address;
    private VirtualMachine virtualMachine;
    
    private Process process = null;
    private Exception thrownException = null;
    private boolean exited = false;
    
    private IOManager ioManager;
    
    Launcher (IOManager ioManager) {
        this.ioManager = ioManager;
    }
    
    public synchronized VirtualMachine launch (DebuggerInfo info) throws DebuggerException {
        transport = System.getProperty ("netbeans.debugger.jpda.transport");
        if ((transport == null) && (Utilities.isWindows ()) &&
            System.getProperty ("java.version").startsWith ("1.3")) { // NOI18N
                transport = "dt_socket"; // NOI18N
        }
        if ((transport == null) ||
            ((connector = getListeningConnectorFor (transport)) == null)) {
            VirtualMachineManager vmm = Bootstrap.virtualMachineManager ();
            List lcs = vmm.launchingConnectors ();
            // determine preferred transport type according to default connector
            LaunchingConnector lc = vmm.defaultConnector ();
            if (lc != null) {
                transport = lc.transport ().name ();
                connector = getListeningConnectorFor (transport);
            }
        }
        if (connector != null)
            arguments = connector.defaultArguments ();
                
        if (connector == null) {
            finish ();
            throw new DebuggerException (
                new MessageFormat (bundle.getString ("EXC_Cannot_find_launcher")).
                format (new Object[] {
                            "default" // NOI18N
                        })
            );
        }
                    
        // create commandLine & NbProcessDescriptor ........................................
        NbProcessDescriptor debuggerProcess;
        File workingDirectory = null;
        if (info instanceof ProcessDebuggerInfo) {
            debuggerProcess = ((ProcessDebuggerInfo) info).getDebuggerProcess ();
            workingDirectory = ((ProcessDebuggerInfo) info).getWorkingDirectory ();
        } else
            debuggerProcess = ProcessDebuggerType.DEFAULT_DEBUGGER_PROCESS;
        
        // start listening
        try {
            address = connector.startListening (arguments);
        } catch (Exception e) {
            throwDebuggerException (debuggerProcess.getProcessName (), e);
        }
        
        // create main class & arguments ...................................................
        StringBuffer sb = new StringBuffer ();
        sb.append (info.getClassName ());
        String[] infoArgs = info.getArguments ();
        for (int i = 0; i < infoArgs.length; i ++)
            sb.append (" \"").append (infoArgs [i]).append ('"'); // NOI18N
        String main = new String (sb);
        
        HashMap map = Utils.processDebuggerInfo (
                info,
                "-Xdebug -Xnoagent -Xrunjdwp:transport=" + // NOI18N
                transport +
                ",address=" + // NOI18N
                address +
                ",suspend=y", // NOI18N
                main
        );
        MapFormat format = new MapFormat (map);
        String commandLine = format.format (
                debuggerProcess.getProcessName () + " " + // NOI18N
                debuggerProcess.getArguments ()
        );
        String processName = format.format (debuggerProcess.getProcessName ());
        
        ioManager.println (commandLine, IOManager.DEBUGGER_OUT);
        
        try {
            process = debuggerProcess.exec (format, null, workingDirectory);
        } catch (java.io.IOException exc) {
            throwDebuggerException (processName, exc);
        }
        
        Thread acceptingThread = acceptConnection();
        Thread monitoringThread = monitorTarget();
        virtualMachine = null;
        exited = false;
        thrownException = null;
        
        try {
            while ((virtualMachine == null) && (thrownException == null) && !exited)
                wait();
            
            if (exited) {
                BufferedReader bufferedReader = 
                    new BufferedReader (new InputStreamReader (process.getErrorStream ()));
                checkStream (bufferedReader, IOManager.DEBUGGER_OUT);
                bufferedReader = 
                    new BufferedReader (new InputStreamReader (process.getInputStream ()));
                checkStream (bufferedReader, IOManager.PROCESS_OUT);
                throwDebuggerException (
                    processName, 
                    new IOException (bundle.getString ("CTL_VM_init_failed"))
                );
            }
            if (thrownException != null)
                throwDebuggerException (processName, thrownException);
        } catch (InterruptedException e) {
            throwDebuggerException (processName, e);
        } finally {
            acceptingThread.interrupt();
            monitoringThread.interrupt();
        }
        return virtualMachine;
    }
    
    public void finish () {
        if (connector != null) {
            try {
                connector.stopListening (connector.defaultArguments ());
            } catch (java.io.IOException e) {
            } catch (IllegalConnectorArgumentsException e) {
            }
            connector = null;
        }
    }
    
    public Process process () {
        return process;
    }
    
    public String address () {
        return address;
    }
    
    public AttachingConnector attachingConnector () {
        return getAttachingConnectorFor (transport);
    }
    
    synchronized void notifyOfExit () {
        exited = true;
        notify();
    }
    
    synchronized void notifyOfConnection (VirtualMachine vm) {
        virtualMachine = vm;
        notify();
    }
    
    synchronized void notifyOfThrownException (Exception thrownException) {
        this.thrownException = thrownException;
        notify();
    }
    
    private Thread monitorTarget () {
        Thread thread = new Thread ("Launched target monitor") { // NOI18N
            public void run() {
                try {
                    process.waitFor();
                    notifyOfExit();
                } catch (InterruptedException e) {
                    // Connection has been established, stop monitoring
                }
            }
        };
        thread.start();
        return thread;
    }
    
    private Thread acceptConnection () {
        Thread thread = new Thread ("Connection acceptor") { // NOI18N
            public void run() {
                try {
                    VirtualMachine vm = connector.accept (arguments);
                    finish ();
                    notifyOfConnection (vm);
                } catch (InterruptedIOException e) {
                    // VM terminated, stop accepting
                } catch (IOException e) {
                    // Report any other exception to waiting thread
                    notifyOfThrownException (e);
                } catch (IllegalConnectorArgumentsException e) {
                    notifyOfThrownException (e);
                }
            }
        };
        thread.start();
        return thread;
    }
    
    public static ListeningConnector getListeningConnectorFor (String transport) {
        List lcs = Bootstrap.virtualMachineManager ().
                   listeningConnectors ();
        ListeningConnector lc;
        int i, k = lcs.size ();
        for (i = 0; i < k; i++)
            if ( (lc = (ListeningConnector) lcs.get (i)).transport ().
                    name ().equals (transport)
               ) return lc;
        return null;
    }

    public static AttachingConnector getAttachingConnectorFor (String transport) {
        List acs = Bootstrap.virtualMachineManager ().
                   attachingConnectors ();
        AttachingConnector ac;
        int i, k = acs.size ();
        for (i = 0; i < k; i++)
            if ( (ac = (AttachingConnector) acs.get (i)).transport ().
                    name ().equals (transport)
               ) return ac;
        return null;
    }
    
    private void throwDebuggerException (String processName, Exception e) throws DebuggerException {
        finish ();
        throw new DebuggerException (
            new MessageFormat (
                bundle.getString ("EXC_While_create_debuggee"))
                .format (new Object [] {
                    processName,
                    e.toString ()
            }),    
            e
        );
    }
    
    public boolean checkStream (BufferedReader bufferedReader, int where) {
        final int LIMIT = 20;
        int lineCounter = 0;
        boolean res = false;
        try {
            String line = bufferedReader.readLine ();
            while ((line != null) && (lineCounter < LIMIT)) {
                ioManager.println (line, where);
                line = bufferedReader.readLine ();
                lineCounter++;
            }
            bufferedReader.close ();
        } catch (IOException e) {
        } catch (SecurityException e) {
        }
        return res;
    }
    
}
