/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.util.ResourceBundle;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.HashSet;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;
import org.netbeans.modules.debugger.jpda.util.*;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;

import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.filesystems.FileObject;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;


/**
*
* @author   Jan Jancura
*/
public class LineBreakpoint extends LineBreakpointEvent implements 
Executor, StopEvent, InfoProducer {

    static final long                     serialVersionUID = 1444281317183220550L;
    
    /** Property name constant. */
    public static final String                  PROP_CONDITION = "condition"; // NOI18N
    
    /** Thread which stops on this breakpoint. */
    private transient ThreadReference     thread;
    /** Stores all EventRequests produced by this Event. */
    private Requestor                     requestor;
    private ReferenceType                 tryClass;
    private String                        condition = ""; // NOI18N
    /**
     * Name of the outermost class of the source name.
     * If the breakpoint is set on an outer class then the name is equal
     * to the {@link #org.netbeans.modules.debugger.support.LineBreakpointEvent#sourceName}.
     * This variable is set in method {@link #set()}.
     * @see  #isSetOnNestedClass
     */
    private transient String outerClassName;
    /**
     * Set to true if the {@link #org.netbeans.modules.debugger.support.LineBreakpointEvent#sourceName}
     * contains at least the dollar charecter.
     * The value is set in method {@link #set()}.
     * @see #outerClassName
     */
    private transient boolean isSetOnNestedClass;
    /**
     * Actions available to this event.
     * Its value is <TT>null</TT> until the {@link #getActions} method is called.
     * @see  #getActions
     */
    private transient CoreBreakpoint.Action[]       actions;
    
    // init ....................................................................
    
    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        if (condition == null)
            condition = ""; // NOI18N
    }
    
    /**
    * Returns specific properties of this event.
    */
    public Node.Property[] getProperties () {
        ResourceBundle bundle = NbBundle.getBundle (org.netbeans.modules.debugger.jpda.LineBreakpoint.class);
        return new Node.Property[] {
           new PropertySupport.ReadWrite (
               PROP_SOURCE_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_source_name"),
               bundle.getString ("HINT_breakpoint_source_name")
           ) {
               public Object getValue () {
                   return getSourceName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setSourceName (((String)val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_LINE_NUMBER,
               Integer.TYPE,
               bundle.getString ("PROP_breakpoint_line_number"),
               bundle.getString ("HINT_breakpoint_line_number")
           ) {
               public Object getValue () {
                   return new Integer (getLineNumber ());
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setLineNumber (((Integer)val).intValue ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_CONDITION,
               String.class,
               bundle.getString ("PROP_condition"),
               bundle.getString ("HINT_condition")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return getCondition ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setCondition (((String) val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           }
       };
    }

    /**
     * Set of <TT>ReferenceType</TT>s this breakpoint has been set on.
     * Used for synchronization in method {@link #set(ReferenceType)}.
     */
    private HashSet                       refTypesWithBreakpoint;
        
    // Event impl ......................................................................................

    /**
    * Sets breakpoint with specified properties.
    */
    public boolean set () {
        //S ystem.out.println ();
        //Threa d.dumpStack();
        //S ystem.out.println ("  " + this + ":set TRY  in " + getDebugger () ); // NOI18N


        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if (debugger.isDisconnected ()) return false;
        if ((getSourceName () == null) || (getSourceName ().trim ().length () < 1)) return false;
        if (getLineNumber () < 0) return false;
        if (requestor == null) requestor = new Requestor (debugger.requestManager);
        try {
            if (debugger.virtualMachine == null) return false;
            requestor.removeRequests ();
            if (refTypesWithBreakpoint != null) {
                refTypesWithBreakpoint.clear();
            }

            // For unloaded class
            //S ystem.out.println ();
            //S ystem.out.println ("    " + this + ":set class br. " + getClassName () + '*'); // NOI18N
            ClassPrepareRequest cpr = debugger.requestManager.createClassPrepareRequest ();
            String sourceName = getSourceName();
            //
            //Compute outer class name:
            int dollarIndex = sourceName.indexOf('$');
            outerClassName = (dollarIndex < 0) ? sourceName : sourceName.substring(0, dollarIndex);
            isSetOnNestedClass = (sourceName != outerClassName);
            //
            if (isSetOnNestedClass) {
                //If the source name contains name of a nested class
                //then we are sure that the breakpoint is set on a nested
                //class.
                //We cannot set filter to (sourceName + '$') bacause of the
                //way how Javac generates .class files for inner classes
                //of inner classes.
                cpr.addClassFilter(outerClassName + "$*"); // NOI18N
            }
            else {
                //If the source name does not contain name of a nested
                //class then we do not know anything. Maybe it is set
                //on an inner class, maybe not.
                cpr.addClassFilter(outerClassName + '*'); // NOI18N
            }
            cpr.setSuspendPolicy (ClassPrepareRequest.SUSPEND_ALL);
            debugger.operator.register (cpr, this);
            requestor.add (cpr);
            cpr.enable ();

            //We must always go through all nested classes of the outermost class
            //because of JDK bug #4431170 (ReferenceType.nestedTypes() returns bad data)
            List l = debugger.virtualMachine.classesByName (outerClassName);
            if (l.size () == 0) return false;

            // Known classes
            //S ystem.out.println ();
            //S ystem.out.println ("    " + this + ":set " + l.size () + " are loaded"); // NOI18N
            ArrayList list = new ArrayList (l);
            boolean set = false;
            int i = 0;
            while (i < list.size ()) {
                ReferenceType ref = (ReferenceType) list.get (i);
                if (set (ref))
                    set = true;

                try {
                    list.addAll(ref.nestedTypes());
                } catch (ObjectCollectedException ex) {
                }
                i++;
    	    }
            return set;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {
        //S ystem.out.println (); // NOI18N
        //S ystem.out.println ("  " + this + ":remove "); // NOI18N
        if (requestor != null)
            requestor.removeRequests ();
    }

    /**
    * Returns set of actions available specially for this version of event.
    * @return  list of available actions; the list is always the same
    */
    public CoreBreakpoint.Action[] getActions() {
        if (actions == null) {
            CoreBreakpoint.Action[] parentActions = super.getActions();
            actions = new CoreBreakpoint.Action[] {
                new StopAction (),
                new PrintAction (PrintAction.BREAKPOINT_TEXT)
            };
            if (parentActions.length != 0) {
                CoreBreakpoint.Action[] myActions = actions;
                actions = new CoreBreakpoint.Action[parentActions.length + myActions.length];
                System.arraycopy(parentActions, 0, actions, 0, parentActions.length);
                System.arraycopy(myActions, 0, actions, parentActions.length, myActions.length);
            }
        }
        return actions;
    }

    /**
    * Returns customizer visual component.
    */
    public JComponent getCustomizer () {
        return new LineBreakpointPanel (this);
    }
    
    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaThread getThread () {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        return debugger.threadManager.getThread (thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public CallStackFrame[] getCallStack () {
        return (CallStackFrame[]) ((JPDAThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaVariable getVariable () {
        return null;
    }


    // interface Executor .....................................................................

    /**
    * Executes breakpoint hit event.
    */
    public void exec (com.sun.jdi.event.Event event) {
        //S ystem.out.println ();
        //S ystem.out.println (this + ":exec " + event); // NOI18N
        if (event instanceof ClassPrepareEvent) {
            tryClass = ((ClassPrepareEvent) event).referenceType ();
            //
            //Resolve whether the class is being demanded:
            //
            //If the source name contains name of a nested class, then the filter
            //set in method "set()" ensures that this method is called only
            //for nested classes of the outer class -- see #outerClassName.
            //
            //If the source name does not contain name of a nested class,
            //then the filter set in method "set()" ensures only prefix
            //of the class's name. So we must perform further tests.
            String tryClassName = tryClass.name();
            String sourceName = getSourceName();
            boolean isDemanded = isSetOnNestedClass ||
                                 tryClassName.equals(sourceName) ||
                                 tryClassName.startsWith(sourceName + '$');
            if (isDemanded) {
                boolean v = set (tryClass);
                if (v && !getBreakpoint ().isValid ()) setValid (true);
            }
            ((JPDADebugger) getDebugger ()).operator.resume ();
            return;
        }
        thread = ((BreakpointEvent) event).thread ();
        if (((JPDADebugger) getDebugger ()).resolveCanBeCurrent (thread))
            return;
        perform ();
    }

    public void setCondition (String cond) {
        String old = condition;
        condition = cond;
        firePropertyChange (PROP_CONDITION, old, cond);
    }
    
    public String getCondition () {
        return condition;
    }
    
    protected void perform () {
        if (!condition.trim ().equals ("")) { // NOI18N
            JPDADebugger debugger = (JPDADebugger) getDebugger ();
            JPDAWatch watch = new JPDAWatch (debugger, true);
            watch.setVariableName (condition);
            watch.refreshValue ((JPDAThread) getThread ());
            String type = watch.getType();
            String value = watch.getAsText();
            if ((type != null) && type.equals ("boolean") && (value != null) && value.equals ("false")) { // NOI18N
                stop (false);
                return;
            }
            if ((type == null) || (!type.equals ("boolean"))) { // NOI18N
                // ivalid condition, report message ...
                final ResourceBundle bundle = org.openide.util.NbBundle.getBundle (LineBreakpoint.class);
                debugger.getIOManager ().println (bundle.getString ("CTL_Incorrect_condition") + ": " + // NOI18N
                                  bundle.getString ("CTL_breakpoint_at") + " " + getDisplayName () + ".", IOManager.DEBUGGER_OUT); // NOI18N
            }
        }
        // perform actions ...
        super.perform ();
    }

    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */
    public void stop (boolean stop) {
        ((JPDADebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods .........................................................................

    /**
    * Sets breakpoint for given class.
    */
    public boolean set (ReferenceType clazz) {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        if (getLineNumber () < 0) return false;
        //
        //Ensure that the breakpoint is not set more than once on the same ReferenceType:
        synchronized (this) {
            if (refTypesWithBreakpoint == null) {
                refTypesWithBreakpoint = new HashSet(17);
            }
            else {
                try {
                    if (refTypesWithBreakpoint.contains(clazz)) {
                        return true;
                    }
                }
                catch (ObjectCollectedException ex) {      //may be thrown from 'ReferenceType.equals(O
                }
            }
            refTypesWithBreakpoint.add(clazz);
        }
        try {
            // Known class
            List locL = null;
            try {
                locL = clazz.locationsOfLine (getLineNumber ());
            } catch (Exception e) {
                //S ystem.out.println ("    " + this + ":set2 " + e); // NOI18N
            }
            //S ystem.out.println ("  " + this + ":set2 exec br. for class filter " + clazz + " lines numb(rt.locationsOfLine): " + (locL == null ? "null" : "" + locL.size ())); // NOI18N
            if ((locL == null) || (locL.size () < 1)) return false;
            com.sun.jdi.Location loc = (com.sun.jdi.Location) locL.get (0);

            // PATCH for breakpoints in top-level package private classes, try to obtain Lin
            try {
                CoreBreakpoint breakpoint = getBreakpoint ();
                if ((breakpoint.getLine () == null) && (!breakpoint.isHidden ())) {
                    Line newLine = Utils.getLineForSource (
                        getSourceName (), 
                        loc.sourceName (), 
                        getLineNumber ()
                    );
                    if (newLine != null)
                        ((LineBreakpoint)breakpoint.getEvent ()).setLine (newLine);
                }
            } catch (AbsentInformationException e) {
            }
            // end of PATC
        	
            // add breakpoint
            BreakpointRequest br = debugger.requestManager.createBreakpointRequest (loc);
            br.setSuspendPolicy (BreakpointRequest.SUSPEND_ALL);
            debugger.operator.register (br, this);
            requestor.add (br);
            br.enable ();
            //S ystem.out.println ("  " + this + ":set2 OK! "); // NOI18N
            return true;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    public String toString () {
        return "JPDALineBreakpoint " + getSourceName () + "."  + getLineNumber (); // NOI18N
    }
}

