/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.util.ResourceBundle;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.HashSet;
import java.text.MessageFormat;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;

import com.sun.jdi.ReferenceType;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.Location;
import com.sun.jdi.Method;
import com.sun.jdi.AbsentInformationException;
import com.sun.jdi.ObjectCollectedException;
import com.sun.jdi.VMDisconnectedException;
import com.sun.jdi.request.BreakpointRequest;
import com.sun.jdi.request.MethodEntryRequest;
import com.sun.jdi.request.MethodExitRequest;
import com.sun.jdi.request.EventRequestManager;
import com.sun.jdi.request.ClassPrepareRequest;
import com.sun.jdi.request.EventRequest;
import com.sun.jdi.event.BreakpointEvent;
import com.sun.jdi.event.LocatableEvent;
import com.sun.jdi.event.ClassPrepareEvent;
import com.sun.jdi.event.Event;
import com.sun.jdi.event.EventSet;
import com.sun.jdi.event.EventIterator;
import com.sun.jdi.event.StepEvent;

import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.filesystems.FileObject;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.jpda.util.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Utils;


/**
* Implementation of breakpoint on method.
*
* @author   Jan Jancura
*/
public class MethodBreakpoint extends MethodBreakpointEvent 
implements Executor, StopEvent, InfoProducer {

    // static ....................................................................................

    static final long                     serialVersionUID = -2226362778610894492L;
    
    /** Property name constant. */
    public static final String            PROP_ALL_METHODS = "allMethods"; // NOI18N
    /** Property name constant. */
    public static final String            PROP_CONDITION = "condition"; // NOI18N


    // variables ....................................................................................

    /** Thread which stops on this breakpoint. */
    private transient ThreadReference     thread;
    /** Stores all EventRequests produced by this Event. */
    private Requestor                     requestor;
    private ReferenceType                 tryClass;
    /** Curent method lines. */
    private LinkedList                    lines = new LinkedList ();
    private Line[]                        linesArray;
    private boolean                       allMethods = false;
    private boolean                       methodEntry = true;
    private String                        condition = ""; // NOI18N
    /**
     * If <TT>true</TT>, the method represents a constructor.
     */
    private transient boolean             isOnConstructor;
    /**
     * Actions available to this event.
     * Its value is <TT>null</TT> until the {@link #getActions} method is called.
     * @see  #getActions
     */
    private transient CoreBreakpoint.Action[]       actions;

    // init ....................................................................
    
    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        if (condition == null)
            condition = ""; // NOI18N
    }
    /**
     * Set of <TT>ReferenceType</TT>s this breakpoint has been set on.
     * Used for synchronization in method {@link #set(ReferenceType)}.
     */
    private HashSet                       refTypesWithBreakpoint;


    // Event impl ......................................................................................

    /**
    * Sets breakpoint with specified properties.
    */
    public boolean set () {
        //S ystem.out.println ();
        //S ystem.out.println ("  " + this + ":set TRY  in " + getDebugger () ); // NOI18N
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        if ((getClassName () == null) || (getClassName ().trim ().length () < 1)) return false;
        if ( (!getAllMethods ()) &&
                ( (getMethodName () == null) ||
                  (getMethodName ().trim ().length () < 1)
                )
           ) return false;
        if (requestor == null) requestor = new Requestor (debugger.requestManager);
        try {
            requestor.removeRequests ();
            if (refTypesWithBreakpoint != null) {
                refTypesWithBreakpoint.clear();
            }

            if (allMethods) {
                EventRequest er;
                if (methodEntry) {
                    MethodEntryRequest mer = debugger.requestManager.createMethodEntryRequest ();
                    mer.addClassFilter (getClassName ());
                    er = mer;
                } else {
                    MethodExitRequest mer = debugger.requestManager.createMethodExitRequest ();
                    mer.addClassFilter (getClassName ());
                    er = mer;
                }
                er.setSuspendPolicy (MethodEntryRequest.SUSPEND_ALL);
                debugger.operator.register (er, this);
                requestor.add (er);
                er.enable ();
                //S ystem.out.println ();
                //S ystem.out.println ("    " + this + ":set ALL METHODS " + getClassName ()); // NOI18N
                return true;
            }

            // For unloaded class
            //S ystem.out.println ();
            //S ystem.out.println ("    " + this + ":set class br. " + getClassName ()); // NOI18N
            //
            //Detect if we are setting the breakpoint to a constructor:
            isOnConstructor = false;
            String shortClassName; //unqualified name of the class without names of its outer classes (if any)
            shortClassName = Utils.getClassName(getClassName());
            int dollarIndex = shortClassName.lastIndexOf('$');
            if (dollarIndex >= 0) {
                shortClassName = shortClassName.substring(dollarIndex + 1);
            }
            isOnConstructor = shortClassName.equals(getMethodName());
            shortClassName = null;
            //
            ClassPrepareRequest cpr = debugger.requestManager.createClassPrepareRequest ();
            cpr.addClassFilter (getClassName () + '*'); // NOI18N
            cpr.setSuspendPolicy (BreakpointRequest.SUSPEND_ALL);
            debugger.operator.register (cpr, this);
            requestor.add (cpr);
            cpr.enable ();

            List l = debugger.virtualMachine.classesByName (getClassName ());
            if (l.size () == 0) return false;

            // Known classes
            //S ystem.out.println ();
            //S ystem.out.println ("    " + this + ":set " + l.size () + " are loaded"); // NOI18N
            boolean set = false;
            ReferenceType ref;
            List anonymousInner = new ArrayList();
            int i;
            //
            //Set breakpoints on all anonymous inner classes and all their nested classes:
            i = 0;
            int iterCount = l.size();
            while (i < iterCount) {
                ref = (ReferenceType)l.get (i);
                if (set (ref))
                    set = true;

                try {
                    anonymousInner.addAll(Utils.anonymousInnerClasses(ref.name(), ref.nestedTypes()));
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            i = 0;
            while (i < anonymousInner.size()) {
                ref = (ReferenceType)anonymousInner.get(i);
                if (set (ref))
                    set = true;

                try {
                    anonymousInner.addAll(ref.nestedTypes());
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            return set;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {  //S ystem.out.println ("MethodBreakpoint.remove " + this); // NOI18N
        //S ystem.out.println (); // NOI18N
        //S ystem.out.println ("  " + this + ":remove "); // NOI18N
        if (requestor != null)
            requestor.removeRequests ();
        linesArray = null;
        lines = new LinkedList ();
    }

    /**
    * Returns specific properties of this event.
    */
    public Node.Property[] getProperties () {
        ResourceBundle bundle = NbBundle.getBundle (VariableBreakpoint.class);
        return new Node.Property[] {
           new PropertySupport.ReadWrite (
               PROP_CLASS_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_class_name"),
               bundle.getString ("HINT_breakpoint_class_name")
           ) {
               public Object getValue () {
                   return getClassName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setClassName (((String)val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_METHOD_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_method_name"),
               bundle.getString ("HINT_breakpoint_method_name")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return getMethodName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setMethodName (((String)val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_ALL_METHODS,
               Boolean.TYPE,
               bundle.getString ("PROP_breakpoint_all_methods"),
               bundle.getString ("HINT_breakpoint_all_methods")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return new Boolean (getAllMethods ());
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setAllMethods (((Boolean) val).booleanValue ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_CONDITION,
               String.class,
               bundle.getString ("PROP_condition"),
               bundle.getString ("HINT_condition")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return getCondition ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setCondition (((String) val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           }
       };
    }

    /**
    * Returns line of breakpoint.
    */
    public Line[] getLines () {
        if (linesArray != null) return linesArray;
        if (lines.size () == 0) return null;
        linesArray = (Line[]) lines.toArray (new Line [lines.size ()]);  //S ystem.out.println("MethodBreakpoint.getLines : " + lines.size ()); // NOI18N
        return linesArray;
    }

    /**
    * Returns customizer visual component.
    */
    public JComponent getCustomizer () {
        return new MethodBreakpointPanel (this);
    }

    /**
    * Returns display name of this instance of event. It will be used
    * as the name of the breakpoint.
    */
    public String getDisplayName () {
        if (getAllMethods ())
            return new MessageFormat (
                NbBundle.getBundle (MethodBreakpoint.class).
                    getString ("CTL_All_method_event_name")
            ).format (new Object[] {getClassName ()});
        else
            return new MessageFormat (
                NbBundle.getBundle (MethodBreakpoint.class).
                    getString ("CTL_Method_event_name")
            ).format (new Object[] {getClassName (), getMethodName ()});
    }

    /**
    * Returns actions available specially for this version of event.
    */
    public CoreBreakpoint.Action[] getActions() {
        if (actions == null) {
            CoreBreakpoint.Action[] parentActions = super.getActions();
            actions = new CoreBreakpoint.Action[] {
                new StopAction (),
                new PrintAction (PrintAction.BREAKPOINT_METHOD_TEXT)
            };
            if (parentActions.length != 0) {
                CoreBreakpoint.Action[] myActions = actions;
                actions = new CoreBreakpoint.Action[parentActions.length + myActions.length];
                System.arraycopy(parentActions, 0, actions, 0, parentActions.length);
                System.arraycopy(myActions, 0, actions, parentActions.length, myActions.length);
            }
        }
        return actions;
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaThread getThread () {
        if (thread == null) return null;
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        return debugger.threadManager.getThread (thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public CallStackFrame[] getCallStack () {
        return (CallStackFrame[]) ((JPDAThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaVariable getVariable () {
        return null;
        /*    try {
              RemoteStackVariable rsv = thread.getCurrentFrame ().getLocalVariable ("this");
              return new ToolsVariable (
                (ToolsDebugger) getDebugger (),
                rsv.getName (),
                rsv.getValue (), 
                rsv.getType ().toString ()
              );
            } catch (Exception e) {
              return null;
            }  */
    }

    public void setCondition (String cond) {
        String old = condition;
        condition = cond;
        firePropertyChange (PROP_CONDITION, old, cond);
    }
    
    public String getCondition () {
        return condition;
    }
    
    protected void perform () {
        if (!condition.trim ().equals ("")) { // NOI18N
            JPDADebugger debugger = (JPDADebugger) getDebugger ();
            JPDAWatch watch = new JPDAWatch (debugger, true);
            watch.setVariableName (condition);
            watch.refreshValue ((JPDAThread) getThread ());
            String type = watch.getType();
            String value = watch.getAsText();
            if ((type != null) && type.equals ("boolean") && (value != null) && value.equals ("false")) { // NOI18N
                stop (false);
                return;
            }
            if ((type == null) || (!type.equals ("boolean"))) { // NOI18N
                // ivalid condition, report message ...
                final ResourceBundle bundle = org.openide.util.NbBundle.getBundle (LineBreakpoint.class);
                debugger.getIOManager ().println (bundle.getString ("CTL_Incorrect_condition") + ": " + // NOI18N
                                  bundle.getString ("CTL_breakpoint_at") + " " + getDisplayName () + ".", IOManager.DEBUGGER_OUT); // NOI18N
            }
        }
        // perform actions ...
        super.perform ();
    }
    
    // interface Executor .....................................................................

    /**
    * Executes breakpoint hit event.
    */
    public void exec (com.sun.jdi.event.Event event) {
        //S ystem.out.println ();
        //S ystem.out.println (this + ":exec " + event); // NOI18N
        if (event instanceof ClassPrepareEvent) {
            tryClass = ((ClassPrepareEvent) event).referenceType ();
            //
            //Resolve whether the class is being demanded:
            boolean isDemanded;
            String tryClassName = tryClass.name();
            if (tryClassName.equals(getClassName())) {
                isDemanded = true;
            }
            else {
                String mandatoryPrefix = getClassName() + '$';
                isDemanded = tryClassName.startsWith(mandatoryPrefix)
                    && (!Character.isJavaIdentifierStart(tryClassName.charAt(mandatoryPrefix.length())));
            }
            //
            if (isDemanded) {
                boolean v = set (tryClass);
                if (v && !getBreakpoint ().isValid ()) setValid (true);
            }
            ((JPDADebugger) getDebugger ()).operator.resume ();
            return;
        }
        thread = ((LocatableEvent) event).thread ();
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if (debugger.virtualMachine.canGetSyntheticAttribute()) 
            try {
                if (thread.frame (0).location ().method ().isSynthetic ()) {
                    debugger.operator.resume ();
                    return;
                }
            }
            catch (Exception e) {
            }
        if (debugger.resolveCanBeCurrent (thread))
            return;
        perform ();
    }


    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */
    public void stop (boolean stop) {
        ((JPDADebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods ......................................................................................

    /**
    * 
    */
    public boolean getAllMethods () {
        return allMethods;
    }

    /**
    * 
    */
    public void setAllMethods (boolean all) {
        if (allMethods == all) return;
        boolean old = allMethods;
        allMethods = all;
        firePropertyChange (PROP_ALL_METHODS, new Boolean (old), new Boolean (allMethods));
    }

    /**
    * 
    */
    public boolean getMethodEntry () {
        return methodEntry;
    }

    /**
    * 
    */
    public void setMethodEntry (boolean methodEntry) {
        if (this.methodEntry == methodEntry) return;
        boolean old = this.methodEntry;
        this.methodEntry = methodEntry;
        firePropertyChange (
            PROP_ALL_METHODS, 
            new Boolean (old), 
            new Boolean (methodEntry)
        );
    }

    /**
    * Sets breakpoint for given class.
    */
    public boolean set (ReferenceType clazz) {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        if ( (!getAllMethods ()) &&
                ( (getMethodName () == null) ||
                  (getMethodName ().trim ().length () < 1)
                )
           ) return false;
        //
        //Ensure that the breakpoint is not set more than once on the same ReferenceType:
        synchronized (this) {
            if (refTypesWithBreakpoint == null) {
                refTypesWithBreakpoint = new HashSet(17);
            }
            else {
                try {
                    if (refTypesWithBreakpoint.contains(clazz)) {
                        return true;
                    }
                }
                catch (ObjectCollectedException ex) {      //may be thrown from 'ReferenceType.equals(O
                }
            }
            refTypesWithBreakpoint.add(clazz);
        }
        String stopMethod = isOnConstructor ? "<init>" : getMethodName();
        try {
            List methods = clazz.methods ();
            //S ystem.out.println ("  " + this + ":set2 all methods numb. " +
            //                    methods.size ()); // NOI18N
            boolean ok = false;
            int j, jj = methods.size ();
            for (j = 0; j < jj; j++) {
                Method method = (Method) methods.get (j);
                if (!method.name ().equals (stopMethod)) continue;
                Location loc = method.location ();
                if (loc == null) {
                    //S ystem.out.println ("  " + this + ":set2 method " + method.name () +
                    //                    " :loc. : null"); // NOI18N
                    continue;
                }
                //S ystem.out.println ("  " + this + ":set2 method " + method.name () + " :loc. : " + loc.lineNumber ()); // NOI18N
                //removeRequests ();
                BreakpointRequest br = debugger.requestManager.createBreakpointRequest (loc);
                br.setSuspendPolicy (BreakpointRequest.SUSPEND_ALL);
                debugger.operator.register (br, this);
                requestor.add (br);
                br.enable ();
                ok = true;
                //S ystem.out.println ("  " + this + ":set2 method OK !!!!!!!"); // NOI18N
            }
            return ok;
        } catch (Exception e) {
            //S ystem.out.println ("    " + this + ":set2 " + e); // NOI18N
        }
        return false;
    }

    public String toString () {
        return "JPDAMethodBreakpoint " + getClassName () + "."  + getMethodName (); // NOI18N
    }
}
