/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.util.Map;
import java.util.Iterator;

import com.sun.jdi.connect.AttachingConnector;
import com.sun.jdi.connect.Connector.Argument;
import com.sun.jdi.Bootstrap;

import org.openide.debugger.DebuggerInfo;

import org.netbeans.modules.debugger.delegator.SessionDebuggerInfo;
import org.netbeans.modules.debugger.*;


/**
* Contains information for connectiong to JPDA debugger session.
*
* @author Jan Jancura
*/
public class RemoteDebuggerInfo extends AbstractDebuggerInfo
    implements SessionDebuggerInfo {

    private Map args;
    private AttachingConnector connector;

    /**
    * Construct a new <code>RemoteDebuggerInfo</code>.
    *
    * @param 
    */
    public RemoteDebuggerInfo (
        AttachingConnector connector,
        Map args
    ) {
        super (null, new String [] {});
        this.args = args;
        this.connector = connector;
    }

    /**
    * Construct a new <code>RemoteDebuggerInfo</code> given by a host name and a port number.
    * Socket transport used.
    *
    * @param hostName name of remote computer
    * @param port port
    */
    public RemoteDebuggerInfo (
        String hostName,
        int port
    ) {
        super (null, new String [] {});
        Iterator iter = Bootstrap.virtualMachineManager ().attachingConnectors ().iterator ();
        connector = null;
        while (iter.hasNext ()) {
            AttachingConnector ac = (AttachingConnector) iter.next ();
            if (ac.transport ().name ().toLowerCase ().indexOf ("socket") > -1) { // NOI18N 
                connector = ac;
                break;
            }
        }
        if (connector != null) {
            args = connector.defaultArguments ();
            ((Argument) args.get ("hostname")).setValue (hostName); // NOI18N
            ((Argument) args.get ("port")).setValue (String.valueOf (port)); // NOI18N
        }
    }

    /**
    * Construct a new <code>RemoteDebuggerInfo</code> given by a host name and an address.
    * Shared memory transport used.
    *
    * @param hostName name of remote computer
    * @param address address
    */
    public RemoteDebuggerInfo (
        String hostName,
        String address
    ) {
        super (null, new String [] {});
        Iterator iter = Bootstrap.virtualMachineManager ().attachingConnectors ().iterator ();
        connector = null;
        while (iter.hasNext ()) {
            AttachingConnector ac = (AttachingConnector) iter.next ();
            if (ac.transport ().name ().toLowerCase ().indexOf ("shmem") > -1) { // NOI18N 
                connector = ac;
                break;
            }
        }
        if (connector != null) {
            args = connector.defaultArguments ();
            ((Argument) args.get ("name")).setValue (address); // NOI18N
        }
    }

    /**
    * Construct a new <code>RemoteDebuggerInfo</code> given by an address.
    * Shared memory transport used.
    *
    * @param address address
    */
    public RemoteDebuggerInfo (
        String address
    ) {
        super (null, new String [] {});
        Iterator iter = Bootstrap.virtualMachineManager ().attachingConnectors ().iterator ();
        connector = null;
        while (iter.hasNext ()) {
            AttachingConnector ac = (AttachingConnector) iter.next ();
            if (ac.transport ().name ().toLowerCase ().indexOf ("shmem") > -1) { // NOI18N 
                connector = ac;
                break;
            }
        }
        if (connector != null) {
            args = connector.defaultArguments ();
            ((Argument) args.get ("name")).setValue (address); // NOI18N
        }
    }
    
    /**
    * Get attaching connector.
    *
    * @return attaching connector
    */
    public AttachingConnector getConnector () {
        return connector;
    }

    /**
    * Get arguments for connector.
    *
    * @return map of arguments for connector
    */
    public Map getArgs () {
        return args;
    }

    /**
    * Returns type of debugger.
    */
    public DebuggerImpl getDebuggerImpl () {
        return new JPDADebuggerImpl ();
    }
    
    /**
    * Get the user-visible display name of debugged process.
    *
    * @return display name of debugged process
    */
    public String getProcessName () {
        String s = null;
        String ss;
        Argument arg = (Argument) args.get ("port");
        if (arg != null) {
            // socket transport
            s = arg.value ();
            ss = getLocationName ();
            if (s != null) ss = ss + ':' + s;
        }
        else
            // shmem transport
            ss = ((Argument) args.get ("name")).value ();
        return ss;
    }
    
    public String getLocationName () {
        Argument arg = (Argument) args.get ("hostname");
        if (arg == null)
            return "localhost"; // NOI18N
        String s = arg.value ();
        if (s != null) return s;
        return "localhost";
    }
}
