/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.jpda;

import java.beans.PropertyEditor;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.HashSet;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;

import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.filesystems.FileObject;
import org.openide.explorer.propertysheet.editors.ChoicePropertyEditor;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.actions.AddBreakpointAction;
import org.netbeans.modules.debugger.jpda.util.*;
import org.netbeans.modules.debugger.support.java.*;


/**
* Implementation of breakpoint on variable accecc / modification.
*
* @author   Jan Jancura
*/
public class VariableBreakpoint extends ClassBreakpointEvent 
implements Executor, StopEvent, InfoProducer {


    // static ..................................................................

    static final long                           serialVersionUID = -2579587576443670523L;
    
    /** Property name constant. */
    public static final String                  PROP_FIELD_NAME = "fieldName"; // NOI18N
    /** Property name constant. */
    public static final String                  PROP_FIELD = "field"; // NOI18N
    /** Property name constant. */
    public static final String                  PROP_TYPE = "type"; // NOI18N
    /** Property name constant. */
    public static final String                  PROP_CONDITION = "condition"; // NOI18N
    /** Property type value constant. */
    public static final int                     ACCESS_TYPE = 0;
    /** Property type value constant. */
    public static final int                     MODIFICATION_TYPE = 1;


    // variables ...............................................................

    /** Thread which stops on this breakpoint. */
    private transient ThreadReference     thread;
    /** Stores all EventRequests produced by this Event. */
    private Requestor                     requestor;
    private transient Field               field;
    private String                        fieldName = ((AddBreakpointAction)
            SystemAction.get (AddBreakpointAction.class)).getCurrentFieldName ();
    private int                           type = MODIFICATION_TYPE;
    private transient JPDAVariable        value;
    private transient String              action;
    private String                        condition = ""; // NOI18N
    /**
     * Set of <TT>ReferenceType</TT>s this breakpoint has been set on.
     * Used for synchronization in method {@link #set(ReferenceType)}.
     */
    private HashSet                       refTypesWithBreakpoint;
    /**
     * Actions available to this event.
     * Its value is <TT>null</TT> until the {@link #getActions} method is called.
     * @see  #getActions
     */
    private transient CoreBreakpoint.Action[]       actions;

    // init ....................................................................
    
    private void readObject (java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject ();
        if (condition == null)
            condition = ""; // NOI18N
    }
    
    // Event impl ..............................................................

    /**
    * Sets breakpoint with specified properties.
    */
    public boolean set () { 
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        if ((getClassName () == null) || (getClassName ().trim ().length () < 1)) return false;
        
        // check if a variable breakpoit is supported by the virtual machine
        if ((type == ACCESS_TYPE) && (!debugger.virtualMachine.canWatchFieldAccess ()))
            return false;
        if ((type == MODIFICATION_TYPE) && (!debugger.virtualMachine.canWatchFieldModification ()))
            return false;
        
        if (requestor == null) requestor = new Requestor (debugger.requestManager);
        try {
            requestor.removeRequests ();
            if (refTypesWithBreakpoint != null) {
                refTypesWithBreakpoint.clear();
            }

            // For unloaded class
            ClassPrepareRequest cpr = debugger.requestManager.createClassPrepareRequest ();
            cpr.addClassFilter (getClassName () + '*');     //NOI18N
            cpr.setSuspendPolicy (ClassPrepareRequest.SUSPEND_ALL);
            debugger.operator.register (cpr, this);
            requestor.add (cpr);
            cpr.enable ();

            List l = debugger.virtualMachine.classesByName (getClassName ());
            if (l.size () == 0) return false;

            // Known classes
            boolean set = false;
            ReferenceType ref;
            List anonymousInner = new ArrayList();
            int i;
            //
            //Set breakpoints on all anonymous inner classes and all their nested classes:
            i = 0;
            int iterCount = l.size();
            while (i < iterCount) {
                ref = (ReferenceType)l.get (i);
                if (set (ref))
                    set = true;

                try {
                    anonymousInner.addAll(Utils.anonymousInnerClasses(ref.name(), ref.nestedTypes()));
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            i = 0;
            while (i < anonymousInner.size()) {
                ref = (ReferenceType)anonymousInner.get(i);
                if (set (ref))
                    set = true;

                try {
                    anonymousInner.addAll(ref.nestedTypes());
                }
                catch (ObjectCollectedException ex) {
                }
                i++;
            }
            return set;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {  //S ystem.out.println ("VariableBreakpoint.remove " + this); // NOI18N
        if (requestor != null)
            requestor.removeRequests ();
    }

    /**
    * Returns specific properties of this event.
    */
    public Node.Property[] getProperties () {
        final ResourceBundle bundle = NbBundle.getBundle (VariableBreakpoint.class);
        return new Node.Property[] {
           new PropertySupport.ReadWrite (
               PROP_CLASS_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_class_name"),
               bundle.getString ("HINT_breakpoint_class_name")
           ) {
               public Object getValue () {
                   return getClassName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setClassName (((String)val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_FIELD_NAME,
               String.class,
               bundle.getString ("PROP_breakpoint_field_name"),
               bundle.getString ("HINT_breakpoint_field_name")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return getFieldName ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setFieldName (((String) val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           },
           new PropertySupport.ReadWrite (
               PROP_TYPE,
               Integer.TYPE,
               bundle.getString ("PROP_breakpoint_type_name"),
               bundle.getString ("HINT_breakpoint_type_name")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return new Integer (getType ());
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setType (((Integer) val).intValue ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
               public PropertyEditor getPropertyEditor () {
                   return new ChoicePropertyEditor (
                              new int[] {
                                  ACCESS_TYPE,
                                  MODIFICATION_TYPE
                              },
                              new String[] {
                                  bundle.getString ("CTL_Breakpoint_type_access_value"),
                                  bundle.getString ("CTL_Breakpoint_type_modification_value")
                              }
                          );
               }
           },
           new PropertySupport.ReadWrite (
               PROP_CONDITION,
               String.class,
               bundle.getString ("PROP_condition"),
               bundle.getString ("HINT_condition")
           ) {
               public Object getValue () throws IllegalArgumentException {
                   return getCondition ();
               }
               public void setValue (Object val) throws IllegalArgumentException {
                   try {
                       setCondition (((String) val).trim ());
                   } catch (ClassCastException e) {
                       throw new IllegalArgumentException ();
                   }
               }
           }
       };
    }

    /**
    * Returns actions available specially for this version of event.
    */
    public CoreBreakpoint.Action[] getActions() {
        if (actions == null) {
            CoreBreakpoint.Action[] parentActions = super.getActions();
            actions = new CoreBreakpoint.Action[] {
                new StopAction (),
                new VariablePrintAction (),
            };
            if (parentActions.length != 0) {
                CoreBreakpoint.Action[] myActions = actions;
                actions = new CoreBreakpoint.Action[parentActions.length + myActions.length];
                System.arraycopy(parentActions, 0, actions, 0, parentActions.length);
                System.arraycopy(myActions, 0, actions, parentActions.length, myActions.length);
            }
        }
        return actions;
    }

    /**
    * Returns lines to highlite in the editor.
    */
    public Line[] getLines () {
        return null;
    }

    /**
    * Returns name of type of this event.
    */
    public String getTypeName () {
        return "Variable"; // NOI18N
    }

    /**
    * Returns display name of this event.
    */
    public String getTypeDisplayName () {
        return NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Variable_event_type_name");
    }

    /**
    * Returns display name of this instance of event. It will be used
    * as the name of the breakpoint.
    */
    public String getDisplayName () {
        if (type == MODIFICATION_TYPE)
            return new MessageFormat (
                NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Variable_modification_event_name")
            ).format (new Object[] {getClassName (), getFieldName ()});
        else
            return new MessageFormat (
                NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Variable_access_event_name")
            ).format (new Object[] {getClassName (), getFieldName ()});
    }

    /**
    * Returns name of icon.
    */
    public String getIconBase () {
        return "/org/netbeans/modules/debugger/resources/breakpointOnVariable"; // NOI18N
    }

    /**
    * Returns customizer visual component.
    */
    public JComponent getCustomizer () {
        return new VariableBreakpointPanel (this);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaThread getThread () {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        return debugger.threadManager.getThread (thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public CallStackFrame[] getCallStack () {
        return (CallStackFrame[]) ((JPDAThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaVariable getVariable () {
        return value;
    }


    // interface Executor .....................................................................

    /**
    * Executes breakpoint hit event.
    */
    public void exec (com.sun.jdi.event.Event event) {  //S ystem.out.println ("VariableBreakpoint.exec! " + this + " : " + event); // NOI18N
        if (event instanceof ClassPrepareEvent) {
            ReferenceType tryClass = ((ClassPrepareEvent) event).referenceType ();
            //
            //Resolve whether the class is being demanded:
            boolean isDemanded;
            String tryClassName = tryClass.name();
            if (tryClassName.equals(getClassName())) {
                isDemanded = true;
            }
            else {
                String mandatoryPrefix = getClassName() + '$';
                isDemanded = tryClassName.startsWith(mandatoryPrefix)
                    && (!Character.isJavaIdentifierStart(tryClassName.charAt(mandatoryPrefix.length())));   //only anonymous inner classes
            }
            //
            if (isDemanded) {
                boolean v = set (tryClass);
                if (v && !getBreakpoint ().isValid ()) setValid (true);
            }
            ((JPDADebugger) getDebugger ()).operator.resume ();
            return;
        }
        thread = ((WatchpointEvent) event).thread ();
        value = new JPDAVariable (
                    (JPDADebugger) getDebugger (),
                    "Current value", // NOI18N
                    ((WatchpointEvent) event).valueCurrent (),
                    "" // NOI18N
                );
        action = (event instanceof AccessWatchpointEvent) ?
                 NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Access") :
                 NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Modification");
        perform ();
    }


    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */
    public void stop (boolean stop) {
        ((JPDADebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods .........................................................................

    /**
    * Set name of class the field is in.
    */
    public void setClassName (String cn) {
        if ( (cn == className) ||
             ((cn != null) && (className != null) && className.equals (cn))
        ) return;
        Object old = className;
        className = cn;
        field = null;
        firePropertyChange (PROP_CLASS_NAME, old, className);
    }

    /**
    * Returns name of field in the current class.
    */
    public String getFieldName () {
        return fieldName;
    }

    /**
    * Sets name of field in the current class.
    */
    public void setFieldName (String name) {
        if ( (name == fieldName) ||
             ((name != null) && (fieldName != null) && fieldName.equals (name))
        ) return;
        String old = fieldName;
        fieldName = name;
        field = null;
        firePropertyChange (PROP_FIELD_NAME, old, fieldName);
    }

    /**
    * Returns current field.
    */
    public Field getField () {
        return field;
    }

    /**
    * Sets field and className and fieldName.
    */
    public void setField (Field field) {
        if (this.field == field) return;
        Field old = field;
        this.field = field;
        firePropertyChange (PROP_FIELD, old, field);
        if (field == null) return;
        try {
            setClassName (field.declaringType ().name ());
            setFieldName (field.type ().name ());
        } catch (ObjectCollectedException e) {
        } catch (ClassNotLoadedException e) {
        } catch (VMDisconnectedException e) {
        }
    }

    /**
    * Returns current field.
    */
    public int getType () {
        return type;
    }

    /**
    * Sets field and className and fieldName.
    */
    public void setType (int type) {
        if (this.type == type) return;
        if ( (type != MODIFICATION_TYPE) &&
                (type != ACCESS_TYPE)
           ) throw new IllegalArgumentException  ();
        int old = this.type;
        this.type = type;
        firePropertyChange (PROP_TYPE, new Integer (old), new Integer (type));
    }

    public void setCondition (String cond) {
        String old = condition;
        condition = cond;
        firePropertyChange (PROP_CONDITION, old, cond);
    }
    
    public String getCondition () {
        return condition;
    }
    
    protected void perform () {
        if (!condition.trim ().equals ("")) { // NOI18N
            JPDADebugger debugger = (JPDADebugger) getDebugger ();
            JPDAWatch watch = new JPDAWatch (debugger, true);
            watch.setVariableName (condition);
            watch.refreshValue ((JPDAThread) getThread ());
            String type = watch.getType();
            String value = watch.getAsText();
            if ((type != null) && type.equals ("boolean") && (value != null) && value.equals ("false")) { // NOI18N
                stop (false);
                return;
            }
            if ((type == null) || (!type.equals ("boolean"))) { // NOI18N
                // ivalid condition, report message ...
                final ResourceBundle bundle = org.openide.util.NbBundle.getBundle (LineBreakpoint.class);
                debugger.getIOManager ().println (bundle.getString ("CTL_Incorrect_condition") + ": " + // NOI18N
                                  bundle.getString ("CTL_breakpoint_at") + " " + getDisplayName () + ".", IOManager.DEBUGGER_OUT); // NOI18N
            }
        }
        // perform actions ...
        super.perform ();
    }
    
    /**
    * Sets breakpoint for given class.
    */
    public boolean set (ReferenceType clazz) {
        JPDADebugger debugger = (JPDADebugger) getDebugger ();
        if ((debugger.virtualMachine == null) || (debugger.isDisconnected ())) return false;
        //
        //Ensure that the breakpoint is not set more than once on the same ReferenceType:
        synchronized (this) {
            if (refTypesWithBreakpoint == null) {
                refTypesWithBreakpoint = new HashSet(17);
            }
            else {
                try {
                    if (refTypesWithBreakpoint.contains(clazz)) {
                        return true;
                    }
                }
                catch (ObjectCollectedException ex) {      //may be thrown from 'ReferenceType.equals(O
                }
            }
            refTypesWithBreakpoint.add(clazz);
        }
        try {
            Field f = null;
            boolean ok = false;
            try {
                f = clazz.fieldByName (getFieldName ());
            } catch (ClassNotPreparedException e) {      //S ystem.out.println ("VariableBreakpoint " + e); // NOI18N
            } catch (ObjectCollectedException e) {       //S ystem.out.println ("VariableBreakpoint " + e); // NOI18N
            }          //S ystem.out.println ("VariableBreakpoint.set classinstance: " + rt + " field: " + getFieldName () + " field: " + f); // NOI18N
            if (f == null) return false;
            if (type == ACCESS_TYPE) {
                AccessWatchpointRequest awr = debugger.requestManager.createAccessWatchpointRequest (f);
                awr.setSuspendPolicy (AccessWatchpointRequest.SUSPEND_ALL);
                awr.addClassFilter ("*"); // NOI18N
                debugger.operator.register (awr, this);
                requestor.add (awr);
                awr.enable ();              //S ystem.out.println ("VariableBreakpoint.OK! a"); // NOI18N
                ok = true;
            } else {
                ModificationWatchpointRequest mwr = debugger.requestManager.createModificationWatchpointRequest (f);
                mwr.setSuspendPolicy (ModificationWatchpointRequest.SUSPEND_ALL);
                mwr.addClassFilter ("*"); // NOI18N
                debugger.operator.register (mwr, this);
                requestor.add (mwr);
                mwr.enable ();              //S ystem.out.println ("VariableBreakpoint.OK! m"); // NOI18N
                ok = true;
            }
            return ok;
        } catch (VMDisconnectedException e) {
        }
        return false;
    }

    public String toString () {
        return "JPDAVariableBreakpoint " + getClassName () + "."  + getFieldName () + "."  + getType (); // NOI18N
    }


    // innerclasses ......................................................................................

    class VariablePrintAction extends PrintAction {

        static final long serialVersionUID = -3852634593799541249L;

        /**
        * Creates the new Thrad Print action with default text.
        */
        VariablePrintAction () {
            super (
                NbBundle.getBundle (VariableBreakpoint.class).getString ("CTL_Variable_print_name")
            );
        }

        /**
        * Resolving special tags:
        *   variableName    name variable
        *   action          name of action
        */
        protected void resolveTag (String tag, CoreBreakpoint.Event event, StringBuffer sb) {
            if (tag.equals ("variableName")) // NOI18N
                sb.append (((VariableBreakpoint) event).getFieldName ());
            else
                if (tag.equals ("action")) // NOI18N
                    sb.append (((VariableBreakpoint) event).action);
                else
                    super.resolveTag (tag, event, sb);
        }
    }
}
