/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.debug;

import java.util.ResourceBundle;
import javax.swing.SwingUtilities;

import org.openide.TopManager;
import org.openide.filesystems.FileObject;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;

import sun.tools.debug.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.*;

/**
*
* @author   Jan Jancura
*/
public class LineBreakpoint extends LineBreakpointEvent implements StopEvent, InfoProducer {

    static final long serialVersionUID =-2076421213106468334L;
    /** Class the breakpoint is set on. */
    private transient RemoteClass         remoteClass;
    /** Thread which stops on theis breakpoint. */
    private transient RemoteThread        thread;
    /** Helper field, stores lineNumber property. */
    private transient int                 remoteLineNumber;


    // Event impl ......................................................................................

    /**
    * Returns the new instance of Breakpoint.Event.
    */
    public CoreBreakpoint.Event getNewInstance () {
        return new LineBreakpoint ();
    }

    /**
    * Sets breakpoint with specified properties.
    */
    public boolean set () {
        final ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        if (debugger.synchronizer == null) return false;
        if (remoteClass != null)
            remove ();
        final String className = getSourceName ();
        final int lineNumber = getLineNumber ();
        if ((className == null) || (className.length () < 1) ||
                (lineNumber < 0)) return false;
        try {
            return ((Boolean) new Protector ("JavaDebugger.addBreakpoint") { // NOI18N
                        public Object protect () throws Exception {
                            //
                            //Compute outer class name:
                            int dollarIndex = className.indexOf('$');
                            String outerClassName = (dollarIndex < 0) ? className : className.substring(0, dollarIndex);
                            //
                            // in outerclass
                            remoteClass = debugger.remoteDebugger.findClass (outerClassName);
                            if (remoteClass == null) return new Boolean (false);
                            remoteLineNumber = lineNumber;
                            
                            // PATCH for breakpoints in top-level package private classes, try to obtain Line
                            CoreBreakpoint breakpoint = getBreakpoint ();
                            if ((getLine () == null) && (!breakpoint.isHidden ())) {
                                Line newLine = Utils.getLineForSource (className, remoteClass.getSourceFileName (), lineNumber);
                                if (newLine != null)
                                    setLine (newLine);
                            }
                            // end of PATCH
                            
                            String s = remoteClass.setBreakpointLine (remoteLineNumber);
                            if (s.trim ().equals ("")) return new Boolean (true); // NOI18N

                            // in innerclasses
                            String name = Utils.getClassName (outerClassName);
                            FileObject mainFile = TopManager.getDefault().getRepository ().find (
                                                      Utils.getPackageName (outerClassName),
                                                      name,
                                                      "class" // NOI18N
                                                  );
                            FileObject[] file = mainFile.getParent ().getChildren ();

                            int i, k = file.length;
                            if (k > 0) {
                                String mandatoryPrefix = name + '$';
                                for (i = 0; i < k; i++)
                                    if (file [i].getName ().startsWith (mandatoryPrefix) &&
                                            file [i].getExt ().equals ("class") // NOI18N
                                       ) {
                                        remoteClass = debugger.remoteDebugger.findClass (file [i].getPackageName ('.'));
                                        if (remoteClass == null) return new Boolean (false);
                                        s = remoteClass.setBreakpointLine (lineNumber);
                                        if (s.trim ().equals ("")) return new Boolean (true); // NOI18N
                                    }
                            }

                            //          if (s.trim ().equals ("")) return new Boolean (true); // NOI18N
                            debugger.getIOManager ().println (NbBundle.getBundle (ToolsDebugger.class).getString (
                                                  "CTL_Cannot_set_breakpoint") + ": " + s, // NOI18N
                                              IOManager.DEBUGGER_OUT
                                             );
                            return new Boolean (false);
                        }
                    }.throwAndWait (debugger.synchronizer, debugger.killer)).booleanValue ();
        
        } catch (Exception e) {
            return false;
        }
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {
        final ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        if (debugger.synchronizer == null) return;
        try {
            new Protector ("JavaDebugger.addBreakpoint") { // NOI18N
                public Object protect () throws Exception {
                    String s = remoteClass.clearBreakpointLine (remoteLineNumber);
                    remoteClass = null;
                    if (s.trim ().equals ("")) return null; // NOI18N
                    debugger.getIOManager ().println (NbBundle.getBundle (ToolsDebugger.class).getString (
                                          "CTL_Cannot_clear_breakpoint") + ": " + s, // NOI18N
                                      IOManager.DEBUGGER_OUT
                                     );
                    return null;
                }
            }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Exception e) {
        }
    }

    /**
    * Returns actions available specially for this version of event.
    */
    public CoreBreakpoint.Action[] getActions () {
        CoreBreakpoint.Action[] myActions = new CoreBreakpoint.Action[] {
                                                new StopAction (),
                                                new ToolsPrintAction (PrintAction.BREAKPOINT_TEXT),
                                            };
        CoreBreakpoint.Action[] actions = new CoreBreakpoint.Action [super.getActions ().length + myActions.length];
        System.arraycopy (super.getActions (), 0, actions, 0, super.getActions ().length);
        System.arraycopy (myActions, 0, actions, super.getActions ().length, myActions.length);
        return actions;
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaThread getThread () {
        ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        ToolsThread tt = ((ToolsThreadGroup) debugger.getThreadsRoot ()).getThread (thread);
        if (tt != null) return tt;
        debugger.lastCurrentThread = thread;
        return new ToolsThread ((ToolsDebugger) getDebugger (), null, thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public CallStackFrame [] getCallStack () {
        return (CallStackFrame[]) ((ToolsThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public JavaVariable getVariable () {
        try {
            RemoteStackVariable rsv = thread.getCurrentFrame ().getLocalVariable ("this"); // NOI18N
            return new ToolsVariable (
                       (ToolsDebugger) getDebugger (),
                       rsv.getName (),
                       rsv.getValue (),
                       rsv.getType ().toString ()
                   );
        } catch (Exception e) {
            return null;
        }
    }


    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */ 
    public void stop (boolean stop) {
        ((ToolsDebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods ......................................................................................

    void perform (RemoteThread thread) {
        this.thread = thread;
        perform ();
    }
}
