/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.debug;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.ResourceBundle;
import javax.swing.SwingUtilities;

import org.openide.TopManager;
import org.openide.filesystems.FileObject;
import org.openide.text.Line;
import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;

import sun.tools.debug.*;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.*;

/**
* Implementation of breakpoint on method.
*
* @author   Jan Jancura
*/
public class MethodBreakpoint extends MethodBreakpointEvent implements StopEvent {

    static final long serialVersionUID =4718302661899335262L;

    /**
     * List of method breakpoints connected to this <CODE>MethodBreakpoint</CODE>.
     * Each breakpoint is an instance of class <CODE>MethodBreakpoint.BreakpointInfo</CODE>,
     * i.e. a&nbsp;pair <EM>(remote class, remote method)</EM>.
     * <P>
     * Value is <CODE>null</CODE> if no breakpoint is set.
     * @see  BreakpointInfo
     */
    private transient List                breakpoints;
    /** Thread which stops on theis breakpoint. */
    private transient RemoteThread        thread;


    // Event impl ......................................................................................

    /**
     * This class represent pair (remote class, remote method).
     * @see  #breakpoints
     */
    private static class BreakpointInfo {
        RemoteClass clazz;
        RemoteField method;
        public BreakpointInfo(RemoteClass clazz, RemoteField method) {
            this.clazz = clazz;
            this.method =method;
        }
    }

    /**
    * Returns the new instance of Breakpoint.Event.
    */
    public CoreBreakpoint.Event getNewInstance () {
        return new MethodBreakpoint ();
    }

    /**
     * Finds all methods of the given name within the specified class.
     * <p>
     * This method serves as a patch for method <CODE>RemoteClass.getMethod(String)</CODE>
     * which returns only one method of the given name if the method name is overloaded.
     * @param clazz  the class whose methods are to be found
     * @param methodName  name of the method to be found
     * @return  all methods of the given name, an empty array if no such method exists
     * @exception  java.lang.Exception  if a general exception occurs
     * @see  RemoteClass#getMethod(String)
     */
    private RemoteField[] findMethodsByName(RemoteClass clazz, String methodName) throws Exception {
        RemoteField method;
        String methodNames[] = clazz.getMethodNames();
        int arrLen = methodNames.length;
        int count = 0;
        for (int i = 0; i < arrLen; i++) {      //check if the method name is overloaded
            if (methodNames[i].equals(methodName)) {
                count++;
            }
        }
        if (count == 0) {    //not overloaded
            return new RemoteField[0];
        }
        if (count == 1) {   //not overloaded
            method = clazz.getMethod(methodName);
            if (method == null) {
                return new RemoteField[0];
            }
            return new RemoteField[]{method};
        }
        //
        //Yes - overloaded:
        //Go through all methods and pick up methods with the right name
        RemoteField allMethods[] = clazz.getMethods();
        List methodsFound = new ArrayList(count);
        int iCount = allMethods.length;
        for (int i = 0; i < iCount; i++) {
            method = allMethods[i];
            if (method.getName().equals(methodName)) {
                methodsFound.add(method);
            }
        }
        return (RemoteField[])methodsFound.toArray(new RemoteField[0]);
    }

    /**
     * Finds classes this method breakpoint should be set on.
     * These involve:
     *   <UL>
     *       <LI>the outer class
     *       <LI>anonymous inner classes of the outer class
     *           and all their nested/inner classes
     *   </UL>
     * @return  list of fully qualified class names
     */
    private String[] findClassNames() {
        List result = new ArrayList(10);
        //
        String className = getClassName();
        result.add(className);
        //
        //Find all .class files in the same directory where the main .class file resists:
        String shortClassName = Utils.getClassName(className);
        FileObject mainFile = TopManager.getDefault().getRepository().find(
                                                                Utils.getPackageName(className),
                                                                shortClassName,
                                                                "class"); //NOI18N
        FileObject[] otherFiles = mainFile.getParent().getChildren();
        //
        //Filter the class names - only anonymous classes will be returned:
        String mandatoryPrefix = shortClassName + '$';
        int innerClassNameIndex = mandatoryPrefix.length();  //index of the first letter of the inner class's name
        int iterCount = otherFiles.length;
        for (int i = 0; i < iterCount; i++) {
                String fileName;
                if (otherFiles[i].getExt().equals("class")                //NOI18N
                    && (fileName = otherFiles[i].getName()).startsWith(mandatoryPrefix)
                    && !Character.isJavaIdentifierStart(fileName.charAt(innerClassNameIndex))) {    //only anonymous classes
                        result.add(otherFiles[i].getPackageName('.'));
                }
        }
        return (String[])result.toArray(new String[0]);
    }

    /**
    * Sets breakpoint with specified properties.
    * The breakpoint is set on methods of the class and on its anonymous
    * inner classes.
    * @return  <CODE>true</CODE> if breakpoint was set successfuly on
    *          at least one method of the class or its inner class,
    *          <CODE>false</CODE> if no breakpoint was set
    */
    public boolean set () {
        final ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        if (debugger.synchronizer == null) return false;
        if (breakpoints != null)
            remove ();
        try {
            return ((Boolean) new Protector ("JavaDebugger.addBreakpoint") { // NOI18N
                        public Object protect () throws Exception {
                            boolean successful = false;
                            String methodName = getMethodName();
                            String[] classNames = findClassNames();
                            int iCount = classNames.length;
                            //
                            //Iterate through all matching classes:
                            for (int i = 0; i < iCount; i++) {
                                String className = classNames[i];
                                //
                                RemoteClass remoteClass = debugger.remoteDebugger.findClass(classNames[i]);
                                if (remoteClass == null) {
                                    continue;  //ignore if the .class file is not loaded
                                }
                                RemoteField[] methods = findMethodsByName(remoteClass, methodName);
                                if (methods.length == 0) {
                                    continue;  //not all processed classes must have the required method
                                }
                                //
                                //Try to set a breakpoint on every matching method of the class:
                                int jCount = methods.length;
                                for (int j = 0; j < jCount; j++) {
                                    String errMsg = remoteClass.setBreakpointMethod(methods[j]);
                                    if (errMsg.length() == 0) {
                                        //remember the breakpoint:
                                        if (breakpoints == null) {
                                            breakpoints = new ArrayList(5);
                                        }
                                        breakpoints.add(new BreakpointInfo(remoteClass, methods[j]));
                                        successful = true;
                                    }
                                    else {
                                        debugger.getIOManager().println (NbBundle.getBundle (ToolsDebugger.class).getString (
                                                              "CTL_Cannot_set_breakpoint") + ": " + errMsg, // NOI18N
                                                          IOManager.DEBUGGER_OUT
                                                         );
                                    }
                                } //for each method
                            } //for each class name
                            return new Boolean(successful);
                        }
                    }.throwAndWait (debugger.synchronizer, debugger.killer)).booleanValue ();
        } catch (Exception e) {
            return false;
        }
    }

    /**
    * Removes breakpoint.
    */
    public void remove () {
        if (breakpoints == null) {
            return;
        }
        final ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        if (debugger.synchronizer == null) return;
        try {
            new Protector ("JavaDebugger.addBreakpoint") { // NOI18N
                public Object protect () throws Exception {
                    for (Iterator i = breakpoints.iterator(); i.hasNext(); ) {
                        BreakpointInfo breakpoint = (BreakpointInfo)i.next();
                        String errMsg = breakpoint.clazz.clearBreakpointMethod(breakpoint.method);
                        if (errMsg.length() > 0) {
                            debugger.getIOManager().println (NbBundle.getBundle (ToolsDebugger.class).getString (
                                                  "CTL_Cannot_clear_breakpoint") + ": " + errMsg, // NOI18N
                                              IOManager.DEBUGGER_OUT
                                             );
                        }
                        breakpoint.clazz = null;
                        breakpoint.method = null;
                    }
                    breakpoints = null;
                    return null;
                }
            }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Exception e) {
        }
    }

    /**
    * @return  <CODE>null</CODE> since method breakpoints are not marked in a source code
    */
    public Line[] getLines () {
        return null;
    }

    /**
    * Returns actions available specially for this version of event.
    */
    public CoreBreakpoint.Action[] getActions () {
        CoreBreakpoint.Action[] myActions = new CoreBreakpoint.Action[] {
                                                new StopAction (),
                                                new PrintAction (PrintAction.BREAKPOINT_METHOD_TEXT),
                                            };
        CoreBreakpoint.Action[] actions = new CoreBreakpoint.Action [super.getActions ().length + myActions.length];
        System.arraycopy (super.getActions (), 0, actions, 0, super.getActions ().length);
        System.arraycopy (myActions, 0, actions, super.getActions ().length, myActions.length);
        return actions;
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public AbstractThread getThread () {
        //S ystem.out.println ("LineBreakpoint.getThread " + this + " : " + getDebugger ()); // NOI18N
        ToolsDebugger debugger = (ToolsDebugger) getDebugger ();
        ToolsThread tt = ((ToolsThreadGroup) debugger.getThreadsRoot ()).getThread (thread);
        if (tt != null) return tt;
        debugger.lastCurrentThread = thread;
        return new ToolsThread ((ToolsDebugger) getDebugger (), null, thread);
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public Location [] getCallStack () {
        return ((ToolsThread) getThread ()).getCallStack ();
    }

    /**
    * Aditional ifno about debugger state when this event occures.
    * If event do not produce this type of info, null is returned.
    */
    public AbstractVariable getVariable () {
        try {
            RemoteStackVariable rsv = thread.getCurrentFrame ().getLocalVariable ("this"); // NOI18N
            return new ToolsVariable (
                       (ToolsDebugger) getDebugger (),
                       rsv.getName (),
                       rsv.getValue (),
                       rsv.getType ().toString ()
                   );
        } catch (Exception e) {
            return null;
        }
    }


    // StopEvent impl ......................................................................................

    /**
    * Performs stop action.
    */
    public void stop (boolean stop) {
        ((ToolsDebugger) getDebugger ()).stop (stop, getThread ());
    }


    // other methods ......................................................................................

    void perform (RemoteThread thread) {
        this.thread = thread;
        perform ();
    }

}

/*
* Log
*  16   Jaga      1.13.3.1    3/2/00   Jan Jancura     
*  15   Jaga      1.13.3.0    2/25/00  Daniel Prusa    post 1.0 changes
*  14   Gandalf   1.13        1/14/00  Daniel Prusa    NOI18N
*  13   Gandalf   1.12        1/13/00  Daniel Prusa    Bugfix for 5281
*  12   Gandalf   1.11        1/13/00  Daniel Prusa    NOI18N
*  11   Gandalf   1.10        12/9/99  Daniel Prusa    Default print text
*  10   Gandalf   1.9         11/29/99 Jan Jancura     Bug 3341 - bad \n in 
*       output of debugger
*  9    Gandalf   1.8         11/8/99  Jan Jancura     Somma classes renamed
*  8    Gandalf   1.7         10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  7    Gandalf   1.6         10/1/99  Jan Jancura     Current thread & bug 4108
*  6    Gandalf   1.5         9/2/99   Jan Jancura     
*  5    Gandalf   1.4         8/18/99  Jan Jancura     Localization & Current 
*       thread & Current session
*  4    Gandalf   1.3         8/17/99  Jan Jancura     Actions for session added
*       & Thread group current property
*  3    Gandalf   1.2         8/9/99   Ian Formanek    Generated Serial Version 
*       UID
*  2    Gandalf   1.1         7/21/99  Jan Jancura     
*  1    Gandalf   1.0         7/2/99   Jan Jancura     
* $
*/
