/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.debug;

import java.beans.*;
import java.util.*;
import javax.swing.SwingUtilities;

import sun.tools.debug.RemoteThread;
import sun.tools.debug.RemoteStackFrame;
import sun.tools.debug.RemoteStackVariable;
import sun.tools.debug.RemoteClass;
import sun.tools.debug.RemoteField;
import sun.tools.debug.RemoteObject;

import org.openide.debugger.DebuggerException;
import org.openide.text.Line;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Protector;
import org.netbeans.modules.debugger.support.util.Utils;


public class ToolsThread extends JavaThread {

    // variables .......................................................................

    /** This debugger. */
    private ToolsDebugger               debugger;
    /** Representing thread. */
    private RemoteThread                thread;
    /** Cache for old locals value. */
    private HashMap                     oldLocals = new HashMap ();
    /** Cache for old call stack. */
    private ToolsCallStackFrame []      oldCallStack = new ToolsCallStackFrame [0];
    /** Old current remote class or 'this' object */
    private RemoteObject                oldCurrentContext = null;
    /** Cache for old context static variables */
    private AbstractVariable []         oldContextFields;

    // last stack depth
    transient private int               lastStackDepth = 0;
    // last action performed on the thread
    transient private int               lastAction = AbstractDebugger.ACTION_START;

    // init ............................................................................

    ToolsThread (
        ToolsDebugger debugger,
        ToolsThreadGroup parentThreadGroup,
        RemoteThread thread
    ) {
        super (debugger, parentThreadGroup);
        this.debugger = debugger;
        this.thread = thread;
        if ( (parentThreadGroup != null) &&
                (debugger.lastCurrentThread != null) &&
                thread.equals (debugger.lastCurrentThread)
           ) {
            setCurrent (true);
            debugger.lastCurrentThread = null;
        }
    }


    // implementation of AbstractThread .....................................................

    /**
    * Getter for the name of thread property.
    *
    * @return name of thread.
    */
    public String getName () {
        if (debugger.synchronizer == null) return "Thread"; // NOI18N
        try {
            return (String) new Protector ("ToolsThread.getName") { // NOI18N
                       public Object protect () throws Exception {
                           return thread.getName ();
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return ""; // NOI18N
        }
    }

    /**
    * If this thread is suspended returns line number where this thread is stopped.
    *
    * @return line number where this thread is stopped.
    */
    public int getLineNumber () {
        if (debugger.synchronizer == null) return -1;
        try {
            return ((Integer) new Protector ("ToolsThread.getLineNumber") { // NOI18N
                        public Object protect () throws Exception {
                            return new Integer (thread.getCurrentFrame ().getLineNumber ());
                        }
                    }.throwAndWait (debugger.synchronizer, debugger.killer)).intValue ();
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return -1;
        }
    }

    /**
    * Returns string representation of the current state of this thread (depends on
    * debugger implementation).
    *
    * @throw DebuggerException if some problem occurs.
    * @return string representation of the current state of this thread.
    */
    public String getState () {
        if (debugger.synchronizer == null) return ""; // NOI18N
        try {
            return (String) new Protector ("ToolsThread.getState") { // NOI18N
                       public Object protect () throws Exception {
                           return thread.getStatus ();
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return ""; // NOI18N
        }
    }

    /**
    * If this thread is suspended returns class name where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return class name where this thread is stopped.
    */
    public String getClassName () {
        if (debugger.synchronizer == null) return ""; // NOI18N
        try {
            return (String) new Protector ("ToolsThread.getClassName") { // NOI18N
                       public Object protect () throws Exception {
                           return thread.getCurrentFrame ().getRemoteClass ().getName ();
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return ""; // NOI18N
        }
    }

    /**
    * If this thread is suspended returns method name where this thread is stopped.
    *
    * @throw DebuggerException if some problem occurs.
    * @return method name where this thread is stopped.
    */
    public String getMethod () {
        if (debugger.synchronizer == null) return ""; // NOI18N
        try {
            return (String) new Protector ("ToolsThread.getClassName") { // NOI18N
                       public Object protect () throws Exception {
                           return thread.getCurrentFrame ().getMethodName ();
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return ""; // NOI18N
        }
    }

    /**
    * Returns current stack depth.
    *
    * @throw DebuggerException if some problem occurs.
    * @return current stack depth.
    */
    public int getStackDepth () {
        if (debugger.synchronizer == null) return -1;
        try {
            return ((Integer) new Protector ("ToolsThread.getStackDepth") { // NOI18N
                        public Object protect () throws Exception {
                            return new Integer (thread.dumpStack ().length);
                        }
                    }.throwAndWait (debugger.synchronizer, debugger.killer)).intValue ();
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return -1;
        }
    }

    /**
    * Returns true if this thread is suspended.
    *
    * @throw DebuggerException if some problem occurs.
    * @return true if this thread is suspended.
    */
    public boolean isSuspended () {
        if (debugger.synchronizer == null) return false;
        try {
            return ((Boolean) new Protector ("ToolsThread.isSuspended") { // NOI18N
                        public Object protect () throws Exception {
                            return new Boolean (thread.isSuspended ());
                        }
                    }.throwAndWait (debugger.synchronizer, debugger.killer)).booleanValue ();
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return false;
        }
    }

    /**
    * Setter method for the suspend property.
    *
    * @throw DebuggerException if some problem occurs.
    * @param suspend true if this thread might be suspend.
    */
    public void setSuspended (final boolean suspend) throws DebuggerException {
        if (debugger.synchronizer == null) return;
        final boolean o = isSuspendedIn ();
        if (o == suspend) return;
        try {
            if (suspend)
                suspendIn ();
            else {
                resumeIn ();
                contIn ();
            }
            if (o != isSuspendedIn ())
                super.setSuspended (suspend);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            throw new DebuggerException (e);
        }
    }

    /**
    * Stops this thread.
    *
    * @throw DebuggerException if some problem occurs.
    */
    public void stop () throws DebuggerException {
        if (debugger.synchronizer == null) return;
        try {
            new Protector ("ToolsThread.setSuspend") { // NOI18N
                public Object protect () throws Exception {
                    thread.stop ();
                    return null;
                }
            }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            throw new DebuggerException (e);
        }
    }

    /**
    * If this thread is suspended returns current call stack.
    *
    * @return current call stack.
    */
    public Location [] getCallStack () {
        if (debugger.synchronizer == null) return new ToolsCallStackFrame [0];
        RemoteStackFrame[] rStack = null;
        try {
            if (!isSuspended ()) return new ToolsCallStackFrame [0];
            rStack = (RemoteStackFrame[]) new Protector ("ToolsThread.getCallStack") { // NOI18N
                         public Object protect () throws Exception {
                             return thread.dumpStack ();
                         }
                     }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return new ToolsCallStackFrame [0];
        }
        int i, k = Math.min (rStack.length, oldCallStack.length);
        ToolsCallStackFrame[] theStack = new ToolsCallStackFrame [rStack.length];
        for (i = 0; i < k; i++) {
            theStack [theStack.length-i-1] = new ToolsCallStackFrame (debugger, rStack [theStack.length-i-1]);
            theStack [theStack.length-i-1].oldLocals = oldCallStack [oldCallStack.length-i-1].oldLocals;
        }
        for (i = k; i < theStack.length; i++)
            theStack [theStack.length -i-1] = new ToolsCallStackFrame (debugger, rStack [theStack.length -i-1]);
        oldCallStack = theStack;
        return theStack;
    }

    /**
    * Returns null. 
    *
    * @return null
    */
    public String getInfo () {
        return null;
    }
    
    /**
    * If this thread is suspended returns current local variables.
    *
    * @return current local variables.
    */
    public AbstractVariable[] getVariables () {
        if (debugger.synchronizer == null) return new AbstractVariable [0];
        try {
            return (AbstractVariable[]) new Protector ("ToolsThread.getLocals") { // NOI18N
                       public Object protect () throws Exception {
                           RemoteStackVariable[] v = thread.getStackVariables ();
                           HashMap newLocals = new HashMap ();
                           ToolsVariable[] rVariables = new ToolsVariable [v.length];

                           int i, k = v.length;
                           for (i = 0; i < k; i++) {
                               rVariables [i] = (ToolsVariable) oldLocals.get (v [i].getName ());
                               if (rVariables [i] == null)
                                   rVariables [i] = new ToolsVariable (
                                        debugger,
                                        v [i].getName (),
                                        v [i].getValue (),
                                        v [i].getType ().toString ()
                                    );
                               else {
                                   rVariables [i].update (
                                       v [i].getName (),
                                       v [i].getValue (),
                                       v [i].getType ().toString ()
                                   );
                                   rVariables [i].firePropertyChange ();
                               }
                               newLocals.put (v [i].getName (), rVariables [i]);
                           }
                           oldLocals = newLocals;
                           return rVariables;
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return new ToolsVariable [0];
        }
    }

    /**
    * If this thread is suspended returns static and instance variables in current context.
    *
    * @return static and instance variables in current context.
    */
    public AbstractVariable[] getContextVariables () {
        if (debugger.synchronizer == null) return new AbstractVariable [0];
        AbstractVariable [] result;
            try {
                result = (AbstractVariable[]) new Protector ("ToolsThread.getContextVariables") { // NOI18N
                    public Object protect () throws Exception {
                        RemoteStackFrame stackFrame  = thread.getCurrentFrame ();
                        RemoteStackVariable var = stackFrame.getLocalVariable ("this"); // NOI18N
                        RemoteClass remoteClass = stackFrame.getRemoteClass ();
                        if ((var != null) && (var.inScope ())) {
                            RemoteObject remoteObject = (RemoteObject)var.getValue ();
                            if (remoteObject.equals (oldCurrentContext)) 
                                return oldContextFields;
                            else {
                                RemoteField [] remoteField = remoteObject.getFields ();
                                RemoteField [] sRemoteField = remoteClass.getFields ();
                                AbstractVariable [] variable = new AbstractVariable [remoteField.length + sRemoteField.length];
                                int i, k = remoteField.length;
                                for (i = 0; i < k; i++)
                                    variable [i] = new ToolsVariable (
                                                       debugger,
                                                       remoteObject,
                                                       remoteField [i],
                                                       i
                                                   );
                                int l = sRemoteField.length;
                                for (i = 0; i < l; i++)
                                    variable [i + k] = new ToolsVariable (
                                                           debugger,
                                                           remoteClass,
                                                           sRemoteField [i],
                                                           i
                                                       );
                                oldCurrentContext = remoteObject;
                                oldContextFields = variable;
                                return variable;
                            }
                        }
                        else { // static context
                            if (remoteClass.equals (oldCurrentContext))
                                return oldContextFields;
                            else {
                                RemoteField [] remoteField = remoteClass.getStaticFields ();
                                AbstractVariable [] variables = new AbstractVariable [remoteField.length];
                                int i, k = remoteField.length;
                                for (i = 0; i < k; i++)
                                    variables [i] = new ToolsVariable (
                                                        debugger,
                                                        remoteClass,
                                                        remoteField [i],
                                                        i
                                                    );
                                oldCurrentContext = remoteClass;
                                oldContextFields = variables;
                                return variables;
                            }
                        } // static context
                    }
                }.throwAndWait (debugger.synchronizer, debugger.killer);
            } catch (Throwable e) {
                if (e instanceof ThreadDeath) throw (ThreadDeath)e;
                    result = new ToolsVariable [0];
            }
        return result;
    }
    
    /**
    * Returns name of file of this frame or null if thread has no frame.
    *
    * @return Returns name of file of this frame.
    * @throws DebuggerException if informations about source are not included or some other error
    *   occurres.
    */
    public String getSourceName () {
        if (debugger.synchronizer == null) return ""; // NOI18N
        try {
            return (String) new Protector ("ToolsThread.getSourceName") { // NOI18N
                       public Object protect () throws Exception {
                           return thread.getCurrentFrame ().getRemoteClass ().getSourceFileName ();
                       }
                   }.throwAndWait (debugger.synchronizer, debugger.killer);
        } catch (Throwable e) {
            if (e instanceof ThreadDeath) throw (ThreadDeath)e;
            return ""; // NOI18N
        }
    }


    // other methods ..............................................................

    int getLastStackDepth () {
        return lastStackDepth;
    }

    int getLastAction () {
        return lastAction;
    }

    /** Sets last action performed on the thread and sets lastStackDepth value. */
    void setLastAction (int action) {
        lastAction = action;
        lastStackDepth = getStackDepth ();
    }

    /**
    * Returns RemoteThread for this thread.
    */
    RemoteThread getRemoteThread () {
        return thread;
    }

    /**
    * Updates state of this thread.
    */
    void threadChanged () {
        //    updateDebuggerState ();
        SwingUtilities.invokeLater (new Runnable () {
                                        public void run () {
                                            firePropertyChange ();
                                        }
                                    });
    }

    /**
    * Helper method.
    */
    protected void firePropertyChange () {
        super.firePropertyChange (null, null, null);
    }

    /**
    * Refresh of stack, locales and properties - fires changes.
    */
    void refreshStack () {
        firePropertyChange (CallStackProducer.PROP_CALLSTACK, null, null);
        firePropertyChange (VariablesProducer.PROP_VARIABLES, null, null);
        firePropertyChange (null, null, null);
    }
    
    /**
    * Equals if RemoteThreads equals.
    */
    public boolean equals (Object o) {
        if (!(o instanceof ToolsThread)) return false;
        return thread.equals (((ToolsThread) o).thread);
    }

    /**
    * Helper method.
    */
    public int hashCode () {
        return thread.hashCode ();
    }


    // unsafe methods ..............................................................

    RemoteStackFrame getCurrentStackIn () throws Exception {
        if (debugger.synchronizer == null) return null;
        return (RemoteStackFrame) new Protector ("ToolsThread.getCurrentStackIn") { // NOI18N
                   public Object protect () throws Exception {
                       return thread.getCurrentFrame ();
                   }
               }.throwAndWait (debugger.synchronizer, debugger.killer);
    }

    boolean isSuspendedIn () {
        if (debugger.synchronizer == null) return false;
        return ((Boolean) new Protector ("ToolsThread.dumpStackIn") { // NOI18N
                    public Object protect () {
                        return new Boolean (thread.isSuspended ());
                    }
                }.wait (debugger.synchronizer, debugger.killer)).booleanValue ();
    }

    RemoteStackFrame[] dumpStackIn () throws Exception {
        if (debugger.synchronizer == null) return null;
        return (RemoteStackFrame[]) new Protector ("ToolsThread.dumpStackIn") { // NOI18N
                   public Object protect () throws Exception {
                       return thread.dumpStack ();
                   }
               }.throwAndWait (debugger.synchronizer, debugger.killer);
    }

    void suspendIn () throws Exception {
        if (debugger.synchronizer == null) return;
        new Protector ("ToolsThread.suspendIn") { // NOI18N
            public Object protect () throws Exception {
                thread.suspend ();
                return null;
            }
        }.throwAndWait (debugger.synchronizer, debugger.killer);
    }

    void resumeIn () throws Exception {
        if (debugger.synchronizer == null) return;
        new Protector ("ToolsThread.resumeIn") { // NOI18N
            public Object protect () throws Exception {
                thread.resume ();
                return null;
            }
        }.throwAndWait (debugger.synchronizer, debugger.killer);
    }

    void contIn () throws Exception {
        if (debugger.synchronizer == null) return;
        new Protector ("ToolsThread.contIn") { // NOI18N
            public Object protect () throws Exception {
                thread.cont ();
                return null;
            }
        }.throwAndWait (debugger.synchronizer, debugger.killer);
    }
}
