/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.debugger.debug;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import sun.tools.debug.RemoteClass;
import sun.tools.debug.RemoteThread;
import sun.tools.debug.RemoteField;
import sun.tools.debug.RemoteObject;
import sun.tools.debug.RemoteStackVariable;
import sun.tools.debug.RemoteValue;
import sun.tools.debug.RemoteStackFrame;
import sun.tools.debug.RemoteArray;

import org.openide.debugger.Watch;
import org.openide.debugger.DebuggerException;
import org.openide.util.NbBundle;

import org.netbeans.modules.debugger.*;
import org.netbeans.modules.debugger.support.*;
import org.netbeans.modules.debugger.support.java.*;
import org.netbeans.modules.debugger.support.util.Validator;
import org.netbeans.modules.debugger.support.util.Protector;


/**
* Standart implementation of Watch interface.
* @see org.openide.debugger.Watch
*
* @author   Jan Jancura
* @version  0.18, Feb 23, 1998
*/
public class ToolsWatch extends JavaWatch {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 3439367144447814302L;


    // private variables .....................................................

    private ToolsVariable                       toolsVariable;
    private String                              errorMessage;

    protected transient boolean                 inScope = false;


    // init .....................................................................

    /**
    * Non public constructor called from the JavaDebugger only.
    * User must create watch from Debugger.getNewWatch () method.
    */
    ToolsWatch (ToolsDebugger debugger, boolean hidden) {
        super (debugger, hidden);
        toolsVariable = new ToolsVariable (debugger, false);
    }


    // AbstractWatch implementation .........................................................

    /**
    * Returns true if this variable is in scope.
    *
    * @return true if this variable is in scope.
    */
    public boolean isInScope () {
        return inScope;
    }

    /**
    * Create AbstractVariable object for this Watch. Can return null, if this Watch currently not
    * represents valide variable.
    *
    * @return AbstractVariable object for this class.
    */
    public AbstractVariable getVariable () {
        return toolsVariable;
    }

    /**
    * Returns true if this variable hasn't any fields.
    *
    * @return True if this variable hasn't any fields.
    */
    public boolean isLeaf () {
        return false;
    }

    /**
    * Checks value of this watch and if is valide, and sets it.
    */
    public void validate () {
        if ((toolsVariable.getDebugger ().synchronizer == null) ||
                ( toolsVariable.getDebugger ().getState () ==
                  toolsVariable.getDebugger ().DEBUGGER_NOT_RUNNING
                )
           ) {
            setErrorMessage (NbBundle.getBundle (ToolsWatch.class).getString ("EXC_No_session"));
            firePropertyChange (null, null, null);
        } else
            if (toolsVariable.getDebugger().getState () != toolsVariable.getDebugger().DEBUGGER_STOPPED) {
                setErrorMessage (NbBundle.getBundle (ToolsWatch.class).getString ("CTL_No_context"));
                firePropertyChange (null, null, null);
            } else {
                ToolsThread tt = (ToolsThread) ((ThreadsRoot) toolsVariable.getDebugger()).getCurrentThread ();
                if (tt != null)
                    refreshValue (tt);
                else {
                    setErrorMessage (NbBundle.getBundle (ToolsWatch.class).getString ("CTL_No_context"));
                    firePropertyChange (null, null, null);
                }
            }
    }

    /**
    * @return false, watch cannot be removed from validator when debugger is finished
    */
    public boolean canRemove () {
        return false;
    }

    /**
    * Delegating method for ToolsVariable instance
    */
    public void setAsText (final String value) {
        toolsVariable.setAsText (value);
    }

    /**
    * Delegating method for ToolsVariable instance
    */
    public AbstractVariable[] getVariables () {
        return toolsVariable.getVariables ();
    }
    
    public AbstractVariable[] getFilteredVariables () {
        return toolsVariable.getFilteredVariables ();
    }
    
    public AbstractVariable[] getVariables (int startindex, int numberofitems) {
        return toolsVariable.getVariables (startindex, numberofitems);
    }
    
    public int getVariablesNumber () {
        return toolsVariable.getVariablesNumber ();
    }
    
    public JavaVariable createFixedVariable () {
        return toolsVariable.createFixedVariable ();
    }
    
    public boolean isFixedVariable () {
        return toolsVariable.isFixedVariable ();
    }
    
    // methods from VariableImpl

    /*
    protected java.lang.Object clone () {
        return toolsVariable.clone_protected ();
    }
     */

    public String getAsText () {
        return toolsVariable.getAsText ();
    }

    public String getType () {
        return toolsVariable.getType ();
    }

    public String getInnerType () {
        return toolsVariable.getInnerType ();
    }

    public boolean isObject () {
        return toolsVariable.isObject ();
    }

    public boolean isArray () {
        return toolsVariable.isArray ();
    }

    public String getModifiers () {
        return toolsVariable.getModifiers ();
    }

    public String getErrorMessage () {
        return errorMessage;
    }

    protected void setErrorMessage (String description) {
        errorMessage = description;
    }


    // other methods ......................................................................

    /**
    * Refresh value of watch in given context.
    */
    void refreshValue (final ToolsThread tt) {
        toolsVariable.setErrorMessage (null);
        new Protector ("ToolsWatch.refreshValue") { // NOI18N
            public Object protect () throws Exception {
                RemoteValue oldValue = toolsVariable.getRemoteValue ();
                RemoteThread rt = tt.getRemoteThread ();
                try {
                    int i = getVariableName ().lastIndexOf ('.');
                    if (i >= 0) {
                        RemoteClass clazz = getRemoteClass (getVariableName ().substring (0, i));
                        if (clazz != null) {
                            if (readStaticVariable (clazz, getVariableName ().substring (i + 1, getVariableName ().length ()))) {
                                setErrorHelper(null);
                            }
                            else {
                                setErrorHelper (NbBundle.getBundle (ToolsWatch.class).getString ("CTL_Name_unknown"));
                            }
                        }
                    } else
                        if (readLocalVariable (rt) || readObjectVariable (rt) || readStaticVariable (rt)) {
                            setErrorHelper(null);
                        }
                        else {
                            setErrorHelper (NbBundle.getBundle (ToolsWatch.class).getString ("CTL_Name_unknown"));
                        }
                } catch (Throwable e) {
                    if (e instanceof ThreadDeath) throw (ThreadDeath)e;
                    setErrorHelper (e.toString ());
                    //          toolsVariable.setNull ();
                    //value = e.toString ();
                }
                return null;
            }
        }.wait (toolsVariable.getDebugger().synchronizer, toolsVariable.getDebugger().killer);
        // deadlock prevention
        firePropertyChange (null, null, null);
    }

    private boolean readLocalVariable (RemoteThread rt) throws Exception {
        RemoteStackVariable variable = getRemoteStackVariable (rt, displayName);
        if (variable == null) return false;                 // unknown variable
        inScope = variable.inScope ();
        toolsVariable.update (
            displayName,
            variable.getValue (),
            variable.getType ().toString ()
        );
        return true;
    }

    private boolean readObjectVariable (RemoteThread rt) throws Exception {
        RemoteStackVariable variable = getRemoteStackVariable (rt, "this"); // NOI18N
        if (variable == null) return false;                 // I'am in static method
        RemoteValue thisValue = variable.getValue ();
        if ((thisValue == null) || (!thisValue.isObject ())) return false;  //Illegal state
        return toolsVariable.update (
                   ((RemoteObject) thisValue).getField (displayName),
                   -1, // unknown slot nuber
                   (RemoteObject) thisValue
               );
    }

    private boolean readStaticVariable (RemoteThread rt) throws Exception {
        RemoteStackFrame frame = rt.getCurrentFrame ();
        if (frame == null) return false;
        RemoteClass clazz = frame.getRemoteClass ();
        if (clazz == null) return false;                  //???
        return readStaticVariable (clazz, displayName);
    }

    private boolean readStaticVariable (RemoteClass clazz, String displayName) throws Exception {
        RemoteField remoteField;
        RemoteObject parentObject = null;
        try {
            remoteField = clazz.getField (displayName);
            if (remoteField == null) return false;
            toolsVariable.setRemoteValue (clazz.getFieldValue (displayName));
        } catch (NoSuchFieldException e) {                // to have transparent behaviour
            return false;                                   // unknown variable
        }
        toolsVariable.update (remoteField, -1, clazz);
        return true;
    }

    /**
    * Returns RemoteStackVariable for the name specified or null.
    *
    * @return RemoteStackVariable for the name specified or null.
    */
    private RemoteStackVariable getRemoteStackVariable (
        RemoteThread rt,
        String variableName
    ) throws Exception {
        RemoteStackFrame frame = rt.getCurrentFrame ();
        if (frame == null) return null;
        return frame.getLocalVariable (variableName);
    }

    /**
    * Returns currently debugged RemoteClass or null.
    *
    * @return currently debugged RemoteClass or null.
    */
    private RemoteClass getRemoteClass (final String className) throws Exception {
        if (toolsVariable.getDebugger().remoteDebugger == null) return null;
        return toolsVariable.getDebugger().remoteDebugger.findClass (className);
    }

    /**
    * Sets error message for this watch.
    */
    void setErrorHelper (String description) {
        setErrorMessage (description);
    }

    // Delegating methods for ToolsVariable instance ..................................

    void setValue (String v) {
        toolsVariable.setValue (v);
    }


    public String toString () {
        return toolsVariable.toString ();
    }
}
