/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.editor.ext.java;

import java.awt.Dialog;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import javax.swing.ListCellRenderer;
import javax.swing.JButton;
import javax.swing.JList;
import org.netbeans.editor.DialogSupport;
import org.netbeans.editor.LocaleSupport;
import org.netbeans.editor.ext.ListCompletionView;

/**
 *
 * @author Miloslav Metelka
 * @version 1.0
 */

public class JavaFastOpen implements ActionListener {
    
    private static final int TIMER_DELAY = 500;
    
    private JavaFastOpenPanel panel;
    
    private ListCellRenderer cellRenderer;
    
    private JList resultList;
    
    private Dialog dialog;
    
    private JButton[] buttons;

    private Evaluator evaluator;
    
    public JavaFastOpen() {
    }
    
    public void setDialogVisible(boolean visible) {
        if (dialog == null) {
            dialog = createDialog();
        }
        
        dialog.setVisible(visible);

        if (visible) {
            getPanel().popupNotify();

        } else {
            if (evaluator != null) {
                evaluator.breakLoop();
            }

            dialog.dispose();
        }
    }
    
    protected void openSource(Object item) {
    }
    
    protected ListCellRenderer createCellRenderer() {
        JCCellRenderer rr = new JCCellRenderer();
        rr.setClassDisplayFullName(true);
        return rr;
    }
    
    protected JList createResultList() {
        JList list = new ListCompletionView(getCellRenderer());
        list.addMouseListener(new MouseAdapter() {
             public void mouseClicked(MouseEvent e) {
                 if (e.getClickCount() == 2) {
                     actionPerformed(new ActionEvent(getButtons()[0], 0, ""));
                 }
             }
        });
        return list;
    }
    
    private JButton[] getButtons() {
        if (buttons == null) {
            buttons = new JButton[] {
                new JButton(LocaleSupport.getString("JFO_openSourceButton", "Open Source")),
                new JButton(LocaleSupport.getString("JFO_closeButton", "Close"))
            };
            String mnemonic = LocaleSupport.getString("JFO_openSourceButtonMnemonic", "O");
            if (mnemonic != null && mnemonic.length() > 0) {
                buttons[0].setMnemonic(mnemonic.charAt(0));
            }
            
            mnemonic = LocaleSupport.getString("JFO_closeButtonMnemonic", "C");
            if (mnemonic != null && mnemonic.length() > 0) {
                buttons[1].setMnemonic(mnemonic.charAt(0));
            }
        }
        
        return buttons;
    }
    
    private Dialog createDialog() {
        String title = LocaleSupport.getString("JFO_title", "Open Java Source");

        Dialog dialog = DialogSupport.createDialog(title,
            getPanel(), false, getButtons(), false, 0, 1, this);
        
        return dialog;
    }

    private JavaFastOpenPanel getPanel() {
        if (panel == null) {
            panel = new JavaFastOpenPanel(this);
        }
        return panel;
    }
    
    ListCellRenderer getCellRenderer() {
        if (cellRenderer == null) {
            cellRenderer = createCellRenderer();
        }
        return cellRenderer;
    }
    
    JList getResultList() {
        if (resultList == null) {
            resultList = createResultList();
        }
        return resultList;
    }
    
    private Evaluator getEvaluator() {
        if (evaluator == null) {
            evaluator = new Evaluator();
            evaluator.start();
        }
        return evaluator;
    }

    void postUpdate(String exp) {
        getEvaluator().setExp(exp);
    }
    
    void update(String exp) {
        if (evaluator != null) {
            evaluator.setExp(null);
        }

        List result = evaluate(exp);
        populate(result);
    }
    
    List evaluate(String exp) {
        List ret = null;

        if (exp != null && exp.length() > 0) {
            JCFinder finder = JavaCompletion.getFinder();
            ret = finder.findClasses(null, exp, false);
            // remove innerclasses
            Iterator it = ret.iterator();
            while (it.hasNext()) {
                JCClass cls = (JCClass)it.next();
                if (cls.getName().indexOf('.') >= 0) {
                    it.remove();
                }
            }
        }
        
        return ret;
    }
    
    void populate(List result) {
        if (result != null) {
            if (getResultList() instanceof ListCompletionView) {
                ((ListCompletionView)getResultList()).setResult(result);
            }
        }
    }

    /**
     * Invoked when an action occurs.
     */
    public void actionPerformed(ActionEvent evt) {
        Object src = evt.getSource();

        if (src == buttons[0]) { // Open button
            int selIndex = getResultList().getSelectedIndex();
            if (selIndex >= 0) {
                openSource(getResultList().getModel().getElementAt(selIndex));
            }
            setDialogVisible(false);
            
        } else if (src == buttons[1]) { // Close button
            setDialogVisible(false);
        }
    }
    
    private class Evaluator extends Thread {
        
        String exp;
        
        String lastExp;
        
        private boolean exit;

        void setExp(String exp) {
            this.exp = exp;
        }
        
        void breakLoop() {
            exit = true;
        }
        
        public void run() {
            try {
                while (!exit) {
                    if (exp != lastExp) {
                        lastExp = exp;
                        String locExp = exp;
                        if (locExp != null) {
                            List result = evaluate(locExp);

                            String locExp2 = exp;
                            if (locExp2 == locExp) {
                                populate(result);
                            }
                        }
                    }

                    sleep(500);
                }
            } catch (InterruptedException e) {
            }
        }
        
    }
    
}
