/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.editor;

import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import org.openide.loaders.DataObject;
import java.util.Vector;
import org.openide.cookies.SourceCookie;
import org.openide.src.SourceElement;
import org.openide.util.WeakListener;
import org.openide.util.RequestProcessor;
import org.netbeans.modules.editor.java.JCUpdater;

/**
 * Listener for DataObject.Registry changes.
 *
 * This class listen on modify changes of dataobjects and after
 * save source action and successful parsing of saved source code
 * will add changes to proper parser DB. 
 *
 * @author  Martin Roskanin
 */
class DORegistryListener implements javax.swing.event.ChangeListener {
    
    private Vector unsavedSources;
    
    
    /** Creates new DORegistryListener */
    public DORegistryListener() {
        unsavedSources = new Vector();
    }
    
    /** This method is evoked if some node has been modified or saved */ 
    public void stateChanged(ChangeEvent e){
        // getting all modified data objects
        DataObject dataObj[] = DataObject.getRegistry().getModified();
        Vector newSources = new Vector();
        
        for (int i = 0; i<dataObj.length; i++){
            SourceCookie sc = (SourceCookie)dataObj[i].getCookie(SourceCookie.class);
            if (sc != null) {
                if (!unsavedSources.contains(dataObj[i])){
                    // if there is some new modified source, we add it to unsavedSources vector
                    unsavedSources.add(dataObj[i]);
                }
                // Add all new modified source to newSources vector
                newSources.add(dataObj[i]);
            }
        }
        addSavedSourcesToParserDB(getSavedSources(newSources));
    }
    
    /** Adds all well parsed saved sources to parser database */
    private void addSavedSourcesToParserDB(final DataObject savedSources[]){
        
        final JCUpdater update = new JCUpdater();
        if(savedSources.length>0){
            // Update changes in parser DB on background in thread with minPriority    
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    for ( int k = 0; k< savedSources.length; k++){
                        SourceCookie sc = (SourceCookie)savedSources[k].getCookie(SourceCookie.class);
                        SourceElement se = sc.getSource();
                        se.prepare().waitFinished();
                        if(se.getStatus() == SourceElement.STATUS_OK){
                            // update nodes only if they were well parsed
                            update.processNode(savedSources[k].getNodeDelegate(), null);
                        }
                        
                    }
                }
            }, 0,Thread.MIN_PRIORITY);
        }
    }
    
    
    /** Gets array of saved sources, that we will need to add to parser DB */
    private DataObject[] getSavedSources (Vector newSources){
        Vector ret = new Vector();
        for (int j=0; j<unsavedSources.size(); j++){
            // If new sources doesn't contain checked source, it means that source has been saved.
            if ( !newSources.contains(unsavedSources.get(j)) ){
                // If source has been saved, add it to return array
                ret.add(unsavedSources.get(j));
            }
        }
        
        // Now we have to remove saved sources from unsavedSources vector
        for( int k=0; k<ret.size(); k++){
            unsavedSources.remove(ret.get(k));
        }
        
        DataObject savedSources[] = new DataObject[ret.size()];
        ret.toArray(savedSources);
        
        return savedSources;
    }
}
