/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.editor.java;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Enumeration;
import org.netbeans.editor.ext.java.JavaCompletion;
import org.netbeans.editor.ext.java.JavaCompletionQuery;
import org.netbeans.editor.ext.java.JCUtilities;
import org.netbeans.editor.ext.java.JCClass;
import org.netbeans.editor.ext.java.JCClassProvider;
import org.netbeans.editor.ext.java.JCFileProvider;
import org.netbeans.editor.ext.java.JCFinder;
import org.netbeans.editor.ext.java.JCBaseFinder;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.FileStateInvalidException;

/**
* Management of Java Completion storage
*
* @author Miloslav Metelka
* @version 1.00
*/

public class JCStorage {

    private static final String DB_DIR = "ParserDB"; // NOI18N
    private static final String STORAGE = "storage"; // NOI18N
    private static final String STORAGE_EXT = "ser"; // NOI18N
    
    /** Whether debug messages about loading of the databases should be displayed */
    private static final boolean debug
        = Boolean.getBoolean("netbeans.debug.editor.jcstorage"); // NOI18N

    private static final String PROVIDER_FILE_EXT
    = JCFileProvider.SKEL_FILE_EXT;

    private static boolean inited;

    private static JCStorage storage;

    private FileObject dbFolder; // parser db folder

    private ArrayList elementList = new ArrayList();

    private ArrayList updatingElements = new ArrayList();
    
    public static JCStorage getStorage() {
        return storage;
    }

    public static void init(FileObject rootFolder) {
        storage = new JCStorage(rootFolder);
    }

    JCStorage(FileObject rootFolder) {
        checkDBFolder(rootFolder);
        if (dbFolder != null) {
            if (debug) {
                String fsName;
                try {
                    fsName = dbFolder.getFileSystem().toString();
                } catch (FileStateInvalidException e) {
                    fsName = "<Invalid filesystem>"; // NOI18N
                }

                System.err.println("JCStorage.init(): scanning dbFolder='" // NOI18N
                    + dbFolder + "' of filesystem='" +  fsName + "'"); // NOI18N
            }

            initElements();
            initProviders();
            refresh();
        }
    }

    private void checkDBFolder(FileObject rootFolder) {
        if (rootFolder != null) {
            dbFolder = rootFolder.getFileObject(DB_DIR);
            if (dbFolder == null) {
                try {
                    dbFolder = rootFolder.createFolder(DB_DIR);
                } catch (IOException e) {
                    // probably exists or cannot be created
                }
            }
        }
    }

    private File getStorageFile(boolean createStorageFile) {
        File storage = null;
        if (dbFolder == null) return null;

        FileObject storageFO = dbFolder.getFileObject(STORAGE, STORAGE_EXT);
        try {
            if ( createStorageFile && (storageFO == null) ) {
                storageFO = dbFolder.createData(STORAGE, STORAGE_EXT);
            }
        } catch (IOException e) {
        }
        
        if (storageFO != null) {
            storage = NbClassPath.toFile(storageFO);
            
        }
        return storage;
    }

    private void initElements() {
        try {
            File storage = getStorageFile(false);
            if (storage != null) {
                if (storage.exists()) {
                    ObjectInputStream is = new ObjectInputStream(new FileInputStream(storage));
                    elementList = (ArrayList)is.readObject();
                    is.close();
                }
            }
        } catch (ClassNotFoundException e) {
        } catch (IOException e) {
        }
    }

    public void saveElements() {
        try {
            File storage = getStorageFile(true);
            if (storage != null) {
                ObjectOutputStream os = new ObjectOutputStream(new FileOutputStream(storage));
                os.writeObject(elementList);
                os.close();
            }
        } catch (IOException e) {
        }
    }

    /** Gets list of all stored JCStorageElements */
    public JCStorageElement[] getElements(){
        JCStorageElement jcse[] = new JCStorageElement[elementList.size()];
        elementList.toArray(jcse);
        return jcse;
    }

    public JCStorageElement getElement(String name) {
        int i = elementList.indexOf(new JCStorageElement(name));
        if (i >= 0) {
            return (JCStorageElement)elementList.get(i);
        } else {
            return null;
        }
    }

    private void initProviders() {
        if (dbFolder != null) {
            // Go through all the files with the skeleton file extension
            Enumeration en = dbFolder.getData(false);
            while (en.hasMoreElements()) {
                FileObject fo = (FileObject)en.nextElement();

                if (debug) {
                    System.err.println("JCStorage.initProviders(): fileObject='" // NOI18N
                        + fo + "'");
                }

                if (JCFileProvider.SKEL_FILE_EXT.equals(fo.getExt())) {
                    // Check whether body file also exists
                    if (fo.existsExt(JCFileProvider.BODY_FILE_EXT)) {
                        String name = fo.getName();
                        // Get full name without extension
                        File foFile = NbClassPath.toFile(fo);
                        String prefix = foFile.getAbsolutePath();
                        prefix = prefix.substring(0, prefix.length()
                                - JCFileProvider.SKEL_FILE_EXT.length() - 1);
                        
                        if (debug) {
                            System.err.println("JCStorage.initProviders():   body file exists, " // NOI18N
                                + "full path prefix='" + prefix + "'"); // NOI18N
                        }

                        // Get or create element and set its provider
                        JCClassProvider provider = new JCFileProvider(prefix);
                        JCStorageElement e = getElement(name);
                        if (e != null) { // valid element read from the storage file
                            e.setProvider(provider);
                        } else {
                            e = new JCStorageElement(name);
                            e.setProvider(provider);
                            addElement(e);
                        }
                    }
                }
            }
        }
    }

    public JCStorageElement findFileSystemElement(String fsName) {
        Iterator i = elementList.iterator();
        while (i.hasNext()) {
            JCStorageElement e = (JCStorageElement)i.next();
            if (fsName.equals(e.getFileSystemName())) {
                return e;
            }
        }
        return null;
    }

    void removeFileSystemElement(String fsName) {
        int cnt = elementList.size();
        for (int i = 0; i < cnt; i++) {
            if (fsName.equals(((JCStorageElement)elementList.get(i)).getFileSystemName())) {
                elementList.remove(i);
                cnt--;
            }
        }
    }

    private void addElement(JCStorageElement e) {
        int i = elementList.indexOf(e);
        if (i >= 0) {
            elementList.set(i, e);
        } else {
            elementList.add(e);
        }
    }

    public JCStorageElement addElement(
        String name, String fsName,
        int classLevel, int fieldLevel, int methodLevel) {
        JCStorageElement e = new JCStorageElement(name, fsName,
                             classLevel, fieldLevel, methodLevel);

        removeFileSystemElement(fsName);
        addElement(e);

        saveElements();

        return e;
    }

    public void checkProvider(JCStorageElement e) {
        if (e.getProvider() == null && dbFolder != null) {
            // Check whether skeleton file exists
            String name = e.getName();
            String ext = JCFileProvider.SKEL_FILE_EXT;

            FileObject skelFO = dbFolder.getFileObject(name, ext);
            try {
                if (skelFO == null) {
                    skelFO = dbFolder.createData(name, ext);
                }
            } catch (IOException ex) {
            }

            if (skelFO != null) {
                // Check body file
                ext = JCFileProvider.BODY_FILE_EXT;
                FileObject bodyFO = dbFolder.getFileObject(name, ext);
                try {
                    if (bodyFO == null) {
                        bodyFO = dbFolder.createData(name, ext);
                    }
                } catch (IOException ex) {
                }

                if (bodyFO != null) {
                    JCClassProvider cp = new JCFileProvider(
                            NbClassPath.toFile(skelFO).getAbsolutePath(),
                            NbClassPath.toFile(bodyFO).getAbsolutePath());

                    cp.reset();
                    e.setProvider(cp);
                }
            }
        }
    }

    /** Refresh JCBaseFinder's Class Map */
    public synchronized void refresh(){
        JCFinder finder = new JCBaseFinder();
        JavaCompletion.setFinder(finder);

        Iterator i = elementList.iterator();
        while (i.hasNext()) {
            JCStorageElement e = (JCStorageElement)i.next();
            if ( (e.getProvider() != null) && (!updatingElements.contains(e)) ) {
                // refresh only not updating elements
                finder.append(e.getProvider());
            }
        }
    }
    
    /** Adds Element to ArrayList of currently updating elements*/
    public synchronized void addProcessedElement(JCStorageElement elem){
        updatingElements.add(elem);
    }

    /** Removes Element from ArrayList of currently updating elements*/
    public synchronized void removeProcessedElement(JCStorageElement elem){
        updatingElements.remove(elem);
    }
    
    public static void main(String args[]) {
        if (args.length != 5) {
            System.err.println("arg1=base-source-dir arg2=target-file-prefix arg3=class-level" // NOI18N
                               + " arg4=field-level arg5=method-level\n" // NOI18N
                               + "Storage level: 0=all, 1=not private 2=public and protected, 3=public only\n" // NOI18N
                               + "Example of args: e:\\java\\jdk12\\src e:\\NB\\system\\ParserDB\\jdk12 2 2 2"); // NOI18N
            return;
        }

        int classLevel = Integer.parseInt(args[2]);
        int fieldLevel = Integer.parseInt(args[3]);
        int methodLevel = Integer.parseInt(args[4]);

        System.out.println("Inspecting classes ..."); // NOI18N
        List classNameList = JCUtilities.getClassNameList(args[0]);
        List classList = JCUtilities.getClassList(classNameList, true,
                         classLevel, fieldLevel, methodLevel);

        final int cnt[] = new int[1];
        JCFileProvider fp = new JCFileProvider(args[1]);
        fp.reset();

        JavaCompletion.ListProvider classes = new JavaCompletion.ListProvider(classList) {
                                                  public boolean notifyAppend(JCClass c, boolean appendFinished) {
                                                      if (appendFinished) {
                                                          System.out.print("Building " + c.getFullName() + " ...             \r"); // NOI18N
                                                          cnt[0]++;
                                                      }
                                                      return true;
                                                  }
                                              };

        long tm = System.currentTimeMillis();

        fp.append(classes);

        System.out.println(cnt[0] + " classes rebuilt in " + (System.currentTimeMillis() - tm) + "ms."); // NOI18N
        File skels = new File(args[1] + JCFileProvider.SKEL_FILE_EXT);
        File bodies = new File(args[1] + JCFileProvider.BODY_FILE_EXT);
        System.out.println("Files created:\nSkeleton file: " // NOI18N
                           + skels.getAbsolutePath() + ", length=" + skels.length() // NOI18N
                           + "\nBody file: " + bodies.getAbsolutePath() // NOI18N
                           + ", length=" + bodies.length()); // NOI18N

        System.exit(0);
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        Iterator i = elementList.iterator();
        while (i.hasNext()) {
            sb.append(((JCStorageElement)i.next()).toString());
            sb.append('\n');
        }
        return sb.toString();
    }
    
}

