/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.editor.java;

import java.awt.event.ActionEvent;
import java.net.URL;
import java.text.MessageFormat;
import java.util.List;
import java.util.ResourceBundle;
import javax.swing.JEditorPane;
import javax.swing.JMenuItem;
import javax.swing.Action;
import javax.swing.text.Document;
import javax.swing.text.JTextComponent;
import javax.swing.text.TextAction;
import javax.swing.text.BadLocationException;
import org.netbeans.editor.BaseDocument;
import org.netbeans.editor.BaseAction;
import org.netbeans.editor.BaseKit;
import org.netbeans.editor.Formatter;
import org.netbeans.editor.Syntax;
import org.netbeans.editor.SyntaxSupport;
import org.netbeans.editor.Utilities;
import org.netbeans.editor.ext.ExtEditorUI;
import org.netbeans.editor.ext.ExtUtilities;
import org.netbeans.editor.ext.Completion;
import org.netbeans.editor.ext.CompletionQuery;
import org.netbeans.editor.ext.java.JCPackage;
import org.netbeans.editor.ext.java.JCMethod;
import org.netbeans.editor.ext.java.JCConstructor;
import org.netbeans.editor.ext.java.JavaSyntax;
import org.netbeans.editor.ext.java.JavaFormatter;
import org.netbeans.editor.ext.java.JavaDrawLayerFactory;
import org.netbeans.modules.editor.NbEditorKit;
import org.netbeans.modules.editor.NbEditorDocument;
import org.netbeans.modules.editor.NbEditorUtilities;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.debugger.Debugger;
import org.openide.debugger.Watch;
import org.openide.debugger.Breakpoint;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.windows.TopComponent;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.Presenter;

/**
* Java editor kit with appropriate document
*
* @author Miloslav Metelka
* @version 1.00
*/

public class JavaKit extends NbEditorKit {

    public static final String JAVA_MIME_TYPE = "text/x-java"; // NOI18N

    private static final String[] getSetIsPrefixes = new String[] {
                "get", "set", "is"
            };

    /** Switch first letter of word to capital and insert 'get'
    * at word begining.
    */
    public static final String makeGetterAction = "make-getter"; // NOI18N

    /** Switch first letter of word to capital and insert 'set'
    * at word begining.
    */
    public static final String makeSetterAction = "make-setter"; // NOI18N

    /** Switch first letter of word to capital and insert 'is'
    * at word begining.
    */
    public static final String makeIsAction = "make-is"; // NOI18N

    /** Add the watch depending on the context under the caret */
    public static final String addWatchAction = "add-watch"; // NOI18N

    /** Toggle the breakpoint of the current line */
    public static final String toggleBreakpointAction = "toggle-breakpoint"; // NOI18N

    /** Debug source and line number */
    public static final String abbrevDebugLineAction = "abbrev-debug-line"; // NOI18N

    /** Open dialog for choosing the import statement to be added */
    public static final String fastImportAction = "fast-import"; // NOI18N

    static final long serialVersionUID =-5445829962533684922L;

    public String getContentType() {
        return JAVA_MIME_TYPE;
    }

    public Document createDefaultDocument() {
        BaseDocument doc = new NbEditorDocument(this.getClass());
        // Force '\n' as write line separator // !!! move to initDocument()
        doc.putProperty(BaseDocument.WRITE_LINE_SEPARATOR_PROP, BaseDocument.LS_LF);
        return doc;
    }

    /** Create new instance of syntax coloring scanner
    * @param doc document to operate on. It can be null in the cases the syntax
    *   creation is not related to the particular document
    */
    public Syntax createSyntax(Document doc) {
        return new JavaSyntax();
    }

    /** Create syntax support */
    public SyntaxSupport createSyntaxSupport(BaseDocument doc) {
        return new NbJavaSyntaxSupport(doc);
    }

    public Completion createCompletion(ExtEditorUI extEditorUI) {
        return new NbJavaCompletion(extEditorUI);
    }

    /** Create the formatter appropriate for this kit */
    public Formatter createFormatter() {
        return new JavaFormatter(this.getClass());
    }

    protected void initDocument(BaseDocument doc) {
        doc.addLayer(new JavaDrawLayerFactory.JavaLayer(),
                JavaDrawLayerFactory.JAVA_LAYER_VISIBILITY);
        doc.addDocumentListener(new JavaDrawLayerFactory.LParenWatcher());
    }

    public Class getFocusableComponentClass(JTextComponent c) {
        return TopComponent.class;
    }

    protected Action[] createActions() {
        Action[] javaActions = new Action[] {
                                   new JavaDefaultKeyTypedAction(),
                                   new JavaBuildToolTipAction(),
//                                   new ToggleBreakpointAction(),
                                   new AddWatchAction(),
                                   new JavaGotoHelpAction(),
                                   new JavaGotoSourceAction(),
                                   new JavaGotoDeclarationAction(),
                                   new PrefixMakerAction(makeGetterAction, "get", getSetIsPrefixes),
                                   new PrefixMakerAction(makeSetterAction, "set", getSetIsPrefixes),
                                   new PrefixMakerAction(makeIsAction, "is", getSetIsPrefixes),
                                   new AbbrevDebugLineAction(),
                                   new CommentAction("//"),
                                   new UncommentAction("//"),
                                   new FastImportAction()
                               };
        return TextAction.augmentList(super.createActions(), javaActions);
    }


    public static class JavaDefaultKeyTypedAction extends ExtDefaultKeyTypedAction {

    }

    public static class JavaBuildToolTipAction extends BuildToolTipAction {

        static final long serialVersionUID =-2009277037915948909L;

        public JavaBuildToolTipAction() {
        }

        protected String buildText(JTextComponent target) {
            String text = null;
            NbJavaSyntaxSupport sup = (NbJavaSyntaxSupport)Utilities.getSyntaxSupport(target);
            Debugger debugger = sup.getDebugger();
            if (debugger != null) {
                int state = debugger.getState();
                if (state == Debugger.DEBUGGER_RUNNING || state == Debugger.DEBUGGER_STOPPED) {
                    String word = ExtUtilities.getExtEditorUI(target).getToolTipSupport().getIdentifierUnderCursor();
                    if (word != null) {
                        Watch watch = debugger.createWatch(word, true);
                        if (watch != null) {
                            String asText = watch.getAsText();
                            if ((asText != null) && (!asText.trim ().equals (word.trim ()))) {
                                text = word + '=' + asText;
                            }
                            watch.remove ();
                        }
                    }
                }
            }
            return text;
        }

    }

    public static class ToggleBreakpointAction extends BaseAction {

        public ToggleBreakpointAction() {
            super(toggleBreakpointAction);
            putValue ("helpID", ToggleBreakpointAction.class.getName ());
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                NbJavaSyntaxSupport sup = (NbJavaSyntaxSupport)Utilities.getSyntaxSupport(target);
                int dotPos = target.getCaret().getDot();
                Breakpoint b = sup.getBreakpoint(dotPos);
                if (b == null) {
                    sup.createBreakpoint(dotPos);
                } else {
                    b.remove();
                }
            }
        }

    }

    public static class AddWatchAction extends BaseAction {

        static final long serialVersionUID =4253425227297112737L;

        public AddWatchAction() {
            super(addWatchAction);
            putValue ("helpID", AddWatchAction.class.getName ());
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                int dotPos = target.getCaret().getDot();
                try {
                    String text = Utilities.getSelectionOrIdentifier(target, dotPos);

                    ResourceBundle bundle = NbBundle.getBundle(JavaKit.class);
                    NotifyDescriptor.InputLine il =
                        new NotifyDescriptor.InputLine(bundle.getString ("CTL_Watch_Name"), // NOI18N
                                                       bundle.getString ("CTL_Watch_Title")); // NOI18N
                    il.setInputText (text);
                    TopManager tm = NbEditorUtilities.getTopManager();
                    if (tm != null) {
                        Object r = tm.notify(il);
                        NbJavaSyntaxSupport sup = (NbJavaSyntaxSupport)Utilities.getSyntaxSupport(target);
                        if (r == NotifyDescriptor.OK_OPTION) {
                            text = il.getInputText();
                            if (text != null) {
                                Debugger debugger = sup.getDebugger();
                                if (debugger != null) {
                                    debugger.createWatch(text, false);
                                }
                            }
                        }
                    }
                } catch (BadLocationException e) {
                    // do nothing
                }
            }
        }

    }

    public static class JavaGotoDeclarationAction extends GotoDeclarationAction {

        public JavaGotoDeclarationAction () {
            putValue ("helpID", JavaGotoDeclarationAction.class.getName ());
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                Completion completion = ExtUtilities.getCompletion(target);
                SyntaxSupport sup = Utilities.getSyntaxSupport(target);
                NbJavaSyntaxSupport nbJavaSup = (NbJavaSyntaxSupport)sup.get(NbJavaSyntaxSupport.class);

                if (completion != null) {
                    if (completion.isPaneVisible()) {
                        Object item = completion.getSelectedValue();
                        if (item != null) {
                            nbJavaSup.openSource(item, true, false);
                        }
                    } else { // pane not visible
                        boolean found = false;
                        int dotPos = target.getCaret().getDot();
                        BaseDocument doc = (BaseDocument)target.getDocument();
                        try {
                            int[] idFunBlk = NbEditorUtilities.getIdentifierAndMethodBlock(doc, dotPos);
                            if (idFunBlk != null && idFunBlk.length == 2) { // id but not function
                                int pos = Utilities.getFirstNonWhiteBwd(doc, idFunBlk[0]);
                                if (pos < 0 || doc.getChars(pos, 1)[0] != '.') { // because 'this.var' could search for local var
                                    found = gotoDeclaration(target);
                                }
                            }

                            if (!found) {
                                if (idFunBlk == null) {
                                    idFunBlk = new int[] { dotPos, dotPos };
                                }

                                for (int ind = idFunBlk.length - 1; ind >= 1; ind--) {
                                    CompletionQuery.Result result = completion.getQuery().query(
                                                                        target, idFunBlk[ind], nbJavaSup);
                                    if (result != null && result.getData().size() > 0) {
                                        nbJavaSup.openSource(result.getData().get(0), true, false);
                                        break;
                                    }
                                }
                            }
                        } catch (BadLocationException e) {
                        }
                    }
                }
            }
        }

    }

    public static class JavaGotoSourceAction extends BaseAction {

        public JavaGotoSourceAction() {
            super(gotoSourceAction, SAVE_POSITION);
            putValue ("helpID", JavaGotoSourceAction.class.getName ());
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                String msg = openSource(target, false);
                if (msg != null) { // not found
                    TopManager tm = NbEditorUtilities.getTopManager();
                    if (tm != null) {
                        tm.notify(new NotifyDescriptor.Message(msg));
                    }
                }
            }
        }

        private int findEndOfMethod(BaseDocument doc, int startPos){
            boolean endOfMethod=false;
            try{
                boolean stopIt=false;
                int nwPos=Utilities.getFirstNonWhiteFwd(doc, startPos);
                do{
                    nwPos = Utilities.getFirstNonWhiteFwd(doc, nwPos);
                    if (nwPos >= 0)  {
                        if (doc.getChars(nwPos, 1)[0] == ')'){
                            stopIt = true;
                            endOfMethod=true;
                            return nwPos+1;
                        }else {
                            if(doc.getChars(nwPos, 1)[0] == ';'){
                                stopIt=true;
                            }
                        }
                        nwPos++;
                    }else{ 
                        stopIt = true;
                    }
                }while(!stopIt);
            } catch (BadLocationException e) {
                return -1;
            }
            return -1;
        }
            
        public String openSource(JTextComponent target, boolean simulate) {
            Object item = null;
            String itemDesc = null;
            Completion completion = ExtUtilities.getCompletion(target);
            SyntaxSupport sup = Utilities.getSyntaxSupport(target);
            NbJavaSyntaxSupport nbJavaSup = (NbJavaSyntaxSupport)sup.get(NbJavaSyntaxSupport.class);

            if (completion != null) {
                if (completion.isPaneVisible()) { // completion pane visible
                    item = completion.getSelectedValue();
                    if (item != null) {
                        itemDesc = nbJavaSup.openSource(item, false, simulate);
                    }
                } else { // pane not visible
                    try {
                        int dotPos = target.getCaret().getDot();
                        BaseDocument doc = (BaseDocument)target.getDocument();
                        int[] idFunBlk = NbEditorUtilities.getIdentifierAndMethodBlock(doc, dotPos);
                        if (idFunBlk == null) {
                            idFunBlk = new int[] { dotPos, dotPos };
                        }

                        for (int ind = idFunBlk.length - 1; ind >= 1; ind--) {
                            CompletionQuery.Result result = completion.getQuery().query(
                                                                target, idFunBlk[ind], nbJavaSup);
                            if (result != null && result.getData().size() > 0) {
                                Object itm = result.getData().get(0);
                                if ( ((itm instanceof JCConstructor) || (itm instanceof JCMethod) )&& (result.getData().size() > 1) ){
                                    // It is overloaded method, lets check for the right one
                                    int endOfMethod = findEndOfMethod(doc, idFunBlk[ind]);

                                    if (endOfMethod>-1){
                                        CompletionQuery.Result resultx = completion.getQuery().query(target, endOfMethod, nbJavaSup);
                                        if (resultx != null && resultx.getData().size() > 0) {
                                            itemDesc = nbJavaSup.openSource(resultx.getData().get(0), false, simulate);
                                            break;
                                        }
                                    }

                                } 

                                itemDesc = nbJavaSup.openSource(result.getData().get(0), false, simulate);
                                break;
                            }
                        }
                    } catch (BadLocationException e) {
                    }
                }
            }

            // Complete the messages
            String msg = null;
            if (itemDesc != null) {
                boolean isPkg = (item instanceof JCPackage);
                msg = NbBundle.getBundle(JavaKit.class).getString(
                          simulate ? (isPkg ? "goto_source_explore_package" : "goto_source_open_source") // NOI18N
          : (isPkg ? "goto_source_package_not_found" : "goto_source_source_not_found") // NOI18N
                      );
                msg = MessageFormat.format(msg, new Object [] { itemDesc } );
            }

            return msg;
        }

        public String getPopupMenuText(JTextComponent target) {
            return openSource(target, true); // simulate open
        }

    }

    public static class JavaGotoHelpAction extends BaseAction {

        public JavaGotoHelpAction() {
            super(gotoHelpAction, SAVE_POSITION);
            putValue ("helpID", JavaGotoHelpAction.class.getName ());
        }

        public URL[] getJavaDocURLs(JTextComponent target) {
            Completion completion = ExtUtilities.getCompletion(target);
            SyntaxSupport sup = Utilities.getSyntaxSupport(target);
            NbJavaSyntaxSupport nbJavaSup = (NbJavaSyntaxSupport)sup.get(NbJavaSyntaxSupport.class);

            if (completion != null) {
                if (completion.isPaneVisible()) { // completion pane visible
                    Object item = completion.getSelectedValue();
                    if (item != null) {
                        nbJavaSup.getJavaDocURLs(item);
                    }

                } else { // completion pane not visible
                    int dotPos = target.getCaret().getDot();
                    BaseDocument doc = (BaseDocument)target.getDocument();
                    try {
                        int[] idFunBlk = NbEditorUtilities.getIdentifierAndMethodBlock(doc, dotPos);
                        if (idFunBlk == null) {
                            idFunBlk = new int[] { dotPos, dotPos };
                        }

                        for (int ind = idFunBlk.length - 1; ind >= 1; ind--) {
                            CompletionQuery.Result result = completion.getQuery().query(
                                                                target, idFunBlk[ind], nbJavaSup);
                            if (result != null && result.getData().size() > 0) {
                                return nbJavaSup.getJavaDocURLs(result.getData().get(0));
                            }
                        }
                    } catch (BadLocationException e) {
                    }
                }
            }
            return null;
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                URL[] urls = getJavaDocURLs(target);
                if (urls != null && urls.length > 0) {
                    TopManager tm = NbEditorUtilities.getTopManager();
                    if (tm != null) {
                        tm.showUrl(urls[0]); // show first URL
                    }
                }
            }
        }

        public String getPopupMenuText(JTextComponent target) {
            return NbBundle.getBundle(JavaKit.class).getString("show_javadoc"); // NOI18N
        }

    }

    public static class AbbrevDebugLineAction extends BaseAction {

        public AbbrevDebugLineAction() {
            super(abbrevDebugLineAction);
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                if (!target.isEditable() || !target.isEnabled()) {
                    target.getToolkit().beep();
                    return;
                }
                BaseDocument doc = (BaseDocument)target.getDocument();
                StringBuffer sb = new StringBuffer("System.out.println(\""); // NOI18N
                String title = (String)doc.getProperty(Document.TitleProperty);
                if (title != null) {
                    sb.append(title);
                    sb.append(':');
                }
                try {
                    sb.append(Utilities.getLineOffset(doc, target.getCaret().getDot()) + 1);
                } catch (BadLocationException e) {
                }
                sb.append(' ');

                BaseKit kit = Utilities.getKit(target);
                Action a = kit.getActionByName(BaseKit.insertContentAction);
                if (a != null) {
                    Utilities.performAction(
                        a,
                        new ActionEvent(target, ActionEvent.ACTION_PERFORMED, sb.toString()),
                        target
                    );
                }
            }
        }

    }
    
    public static class FastImportAction extends BaseAction {
    
        public FastImportAction() {
            super(fastImportAction);
        }

        public void actionPerformed(ActionEvent evt, JTextComponent target) {
            if (target != null) {
                new NbJavaFastImport(target).setDialogVisible(true);
            }
        }
        
    }
    
}
