/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.editor.options;

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.List;
import java.util.Map;
import javax.swing.KeyStroke;
import org.netbeans.editor.BaseKit;
import org.netbeans.editor.MultiKeyBinding;
import org.netbeans.modules.editor.java.JavaKit;


/**
* Upgrade older options to the latest version
*
* @author Miloslav Metelka, Petr Nejedly
* @version 1.00
*/

class UpgradeOptions {
    
    /** Latest version of the options.
     * It is defined here so that only
     * this file gets modified
     * when there's a need for additional upgrade.
     */
    static final int LATEST_VERSION = 3;
    
    private static final String[] javaMapping = {
        "text", "java-whitespace",
        "identifier", "java-identifier",
        "float", "java-float-literal",
        "octal", "java-octal-literal",
        "long", "java-long-literal",
        "char", "java-char-literal",
        "double","java-double-literal",
        "line-comment", "java-line-comment",
        "operator", "java-operators",
        "string", "java-string-literal",
        "function", "java-layer-method",
        "keyword", "java-keywords",
        "block-comment", "java-block-comment",
        "error", "java-errors",
        "hex", "java-hex-literal",
        "int", "java-int-literal",
    };
    
    private static final String[] jspMapping = {
        
        "html-html-argument", "jsp-html-argument",
        "java-double", "jsp-java-double-literal",
        "html-html-text", "jsp-html-text",
        "java-float", "jsp-java-float-literal",
        "java-string", "jsp-java-string-literal",
        "java-function", "jsp-java-layer-method",
        "html-html-sgml-declaration", "jsp-html-sgml-declaration",
        "java-long", "jsp-java-long-literal",
        "java-block-comment", "jsp-java-block-comment",
        "html-html-operator", "jsp-html-operator",
        "html-html-value", "jsp-html-value",
        "java-text", "jsp-java-text",
        "java-operator", "jsp-java-operators",
        "java-char", "jsp-java-char-literal",
        "html-html-tag", "jsp-html-tag",
        "java-hex", "jsp-java-hex-literal",
        "text", "jsp-text",
        "java-line-comment", "jsp-java-line-comment",
        "html-html-ws", "jsp-html-ws",
        "html-html-sgml-comment", "jsp-html-sgml-comment",
        "java-int", "jsp-java-int-literal",
        "java-octal", "jsp-java-octal-literal",
        "html-html-block-comment", "jsp-html-block-comment",
        "html-html-error", "jsp-html-error",
        "java-keyword", "jsp-java-keywords",
        "html-html-character", "jsp-html-character",
        "java-identifier", "jsp-java-identifier",
    };
        
        
    static Map patchColorings( Class kitClass, Map oldColoring ) {
        if( kitClass.getName().equals( "org.netbeans.modules.editor.java.JavaKit" ) &&
        ! oldColoring.containsKey( "java-keywords" )
        ) return translateColoringNames( oldColoring, javaMapping  );

        if( kitClass.getName().equals( "org.netbeans.modules.web.core.syntax.JSPKit" ) &&
        ! oldColoring.containsKey( "jsp-java-keywords" )
        ) return translateColoringNames( oldColoring, jspMapping );

        return oldColoring;

    }
        
        
    private static Map translateColoringNames( Map oldMap, String[] nameMap ) {
        for( int i=0; i < nameMap.length; i += 2 ) {
            if( oldMap.containsKey( nameMap[i] ) ) {
                Object value = oldMap.get( nameMap[i] );
                oldMap.remove( nameMap[i]);
                oldMap.put( nameMap[i+1], value );
            }
        }
        return oldMap;
    }
        
        
    /** Upgrade the deserialized options.
     * @param options options to upgrade
     * @param version deserialized version of the options
     * @param latestVersion latest version of the options
     *   that will be set to them after they are upgraded
     */
    static void upgradeOptions(BaseOptions options, int version, int latestVersion) {
        if (options.getClass() == BaseOptions.class) { // upgrade global options
            if (version < 1) { // add VK_KP_??? shortcuts
                List kbl = options.getKeyBindingList();
                
                kbl.add(
                    new MultiKeyBinding(
                        KeyStroke.getKeyStroke(KeyEvent.VK_KP_RIGHT, 0), // keypad right
                        BaseKit.forwardAction
                    )
                );
                kbl.add(
                    new MultiKeyBinding(
                        KeyStroke.getKeyStroke(KeyEvent.VK_KP_LEFT, 0), // keypad right
                        BaseKit.backwardAction
                    )
                );
                kbl.add(
                    new MultiKeyBinding(
                        KeyStroke.getKeyStroke(KeyEvent.VK_KP_UP, 0), // keypad right
                        BaseKit.upAction
                    )
                );
                kbl.add(
                    new MultiKeyBinding(
                        KeyStroke.getKeyStroke(KeyEvent.VK_KP_DOWN, 0), // keypad right
                        BaseKit.downAction
                    )
                );
                
                options.setKeyBindingList(kbl);
            }
        }
        
        // Check FastimportAction shortcut
        if (options.getClass() == JavaOptions.class) {
            if (version < 2) {
                List kbl = options.getKeyBindingList();
            
                kbl.add(
                    new MultiKeyBinding(
                        KeyStroke.getKeyStroke(KeyEvent.VK_I,
                        InputEvent.ALT_MASK | InputEvent.SHIFT_MASK),
                        JavaKit.fastImportAction
                    )
                );

                options.setKeyBindingList(kbl);
            }
        }
            
        // Check macro shortcuts
        if (options.getClass() == BaseOptions.class) {
            if (version < 3) {
                List kbl = options.getKeyBindingList();
            
                kbl.add(
                    new MultiKeyBinding(
                        new KeyStroke[] {
                            KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK),
                            KeyStroke.getKeyStroke(KeyEvent.VK_D, 0)
                        },
                        "macro-debug-var"
                    )
                );
                
                kbl.add(
                    new MultiKeyBinding(
                        new KeyStroke[] {
                            KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK),
                            KeyStroke.getKeyStroke(KeyEvent.VK_S, 0),
                        },
                        BaseKit.startMacroRecordingAction
                    )
                );

                kbl.add(
                    new MultiKeyBinding(
                        new KeyStroke[] {
                            KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK),
                            KeyStroke.getKeyStroke(KeyEvent.VK_E, 0),
                        },
                        BaseKit.stopMacroRecordingAction
                    )
                );
                
                options.setKeyBindingList(kbl);
            }
            
        }
    }
        
}
