/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form;

import java.beans.*;
import java.util.*;

/**
 * A class that manages events for one RADComponent.
 */

public class ComponentEventHandlers
{
    private RADComponent metacomponent;
    private EventSet[] eventSetHandlers;

    private boolean initialized;

    /** Constructs a new ComponentEventHandlers for specified RADComponent.
     * @param component The RADComponent for which this class manages the events
     */ 
    ComponentEventHandlers(RADComponent component) {
        metacomponent = component;
        initialized = false;
    }

    private void initialize() {
        initialized = true;

        BeanInfo beanInfo = metacomponent.getBeanInfo();
        EventSetDescriptor[] eventSets = beanInfo.getEventSetDescriptors();
        
        eventSetHandlers = new EventSet[eventSets.length];
        for (int i = 0; i < eventSets.length; i++) {
            eventSetHandlers[i] = new EventSet(metacomponent, eventSets[i]);
        }

        Arrays.sort(eventSetHandlers, new Comparator() {
            public int compare(Object o1, Object o2) {
                String n1 =((EventSet)o1).getName();
                String n2 =((EventSet)o2).getName();
                return n1.compareTo(n2);
            }
        });
    }

    /** @return an array of all EventSets for this component */
    public EventSet[] getEventSets() {
        if (!initialized) initialize();
        return eventSetHandlers;
    }

    /** @return the Event that represents the default event or null if there is
     * no default event.
     */
    public Event getDefaultEvent() {
        if (!initialized) initialize();
        BeanInfo beanInfo = metacomponent.getBeanInfo();
        int defIndex = beanInfo.getDefaultEventIndex();
        if (defIndex == -1) {
            for (int i = 0; i < eventSetHandlers.length; i++) {
                if ("action".equals(eventSetHandlers[i].getName())) { // NOI18N
                    Event[] events = eventSetHandlers[i].getEvents();
                    for (int j = 0; j < events.length; j++) {
                        if ("actionPerformed".equals(events[j].getName())) // NOI18N
                            return events[j];
                    }
                }
            }
        } else {
            int counter = 0;
            for (int i = 0; i < eventSetHandlers.length; i++) {
                Event[] events = eventSetHandlers[i].getEvents();
                for (int j = 0; j < events.length; j++) {
                    if (counter == defIndex)
                        return events[j];
                    counter++;
                }
            }
        }
        return null;
    }

    /** @return number of events in all event sets of this RADComponent */
    public int getEventCount() {
        if (!initialized) initialize();
        int count = 0;
        for (int i = 0; i < eventSetHandlers.length; i++)
            count += eventSetHandlers[i].getEventCount();
        return count;
    }

    /** @return number of event handlers attached to events of this RADComponent */
    public int getHandlersCount() {
        if (!initialized) initialize();
        int count = 0;
        for (int i=0; i < eventSetHandlers.length; i++) {
            Event[] events = eventSetHandlers[i].getEvents();
            for (int j=0; j < events.length; j++)
                count += events[j].getHandlers().size();
        }
        return count;
    }

    /** Inititializes event handlers for component's events. 
     * Used by GandalfPersistenceManager when reading events and
     * their handlers for a component.
     * @param eventsInfo Collection of Event.EventInfo objects, one object
     * describes one event and its handlers
     */
    void initEvents(Collection eventsInfo) {
        if (!initialized) initialize();
        Iterator it = eventsInfo.iterator();
        while (it.hasNext()) {
            Event.EventInfo eventInfo = (Event.EventInfo)it.next();
            Event event = eventInfo.findEvent(this);
            if (event != null) {
                StringTokenizer st = new StringTokenizer(eventInfo.eventHandlers, ","); // NOI18N
                FormEventHandlers formHandlers = metacomponent.getFormModel().getFormEventHandlers();
                while (st.hasMoreTokens())
                    formHandlers.addEventHandler(event, st.nextToken());
            }
        }
    }


    /** Used by GandalfPersistenceManager when saving events and their handlers
     * for a component.
     * @returns List of EventInfo objects, one object describes one event and its handlers
     */
    List getEventsInfo() {
        if (!initialized) initialize();
        ArrayList list = new ArrayList();

        for (int i = 0; i < eventSetHandlers.length; i++) {
            Event[] events = eventSetHandlers[i].getEvents();
            for (int j = 0; j < events.length; j++) {
                if (events[j].getHandlers().size() > 0)
                    list.add(new Event.EventInfo(events[j]));
            }
        }
        return list;
    }

    /** Initializes the names of event handlers for all events.
     * Used by TuborgPersistenceManager only.
     * @param eventHandlerNames The mapping <String, String> of event name to
     * event handler name
     */
    void initEvents(Hashtable eventHandlerNames) {
        if (!initialized) initialize();
        for (int i = 0; i < eventSetHandlers.length; i++) {
            Event[] events = eventSetHandlers[i].getEvents();
            for (int j = 0; j < events.length; j++) {
                if (eventHandlerNames.get(events[j].getName()) != null) {
                    StringTokenizer st = new StringTokenizer((String)eventHandlerNames.get(events[j].getName()), ","); // NOI18N
                    while (st.hasMoreTokens()) {
                        String handlerName = st.nextToken();
                        metacomponent.getFormModel().getFormEventHandlers()
                            .addEventHandler(events[j], handlerName);
                    }
                }
            }
        }
    }

    /** Used by TuborgPersistenceManager only. 
    * @return the array of all event handler names */ 
    Hashtable getEventNames() {
        if (!initialized) initialize();
        Hashtable names = new Hashtable(20);

        for (int i = 0; i < eventSetHandlers.length; i++) {
            Event[] events = eventSetHandlers[i].getEvents();
            for (int j = 0; j < events.length; j++) {
                Vector handlers = events[j].getHandlers();
                String eventName = events[j].getName();
                String hanlerNames = ""; // NOI18N
                if (handlers.size() > 0) {
                    String handlerNames =((EventHandler) handlers.get(0)).getName();
                    for (int k=1, n=handlers.size(); k<n; k++)
                        handlerNames = handlerNames + "," + ((EventHandler) handlers.get(k)).getName(); // NOI18N
                    names.put(eventName, handlerNames);
                }
            }
        }
        return names;
    }
}
