/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.form;

import java.util.*;
import java.lang.reflect.*;

/**
 * @author Tomas Pavek
 */

public class ConstructorsDescriptor implements CreationDescriptor {

    private Class describedClass;
    private ConstructorCreator[] constructors;
    private Object[] defaultConstrParams;
    private Constructor defaultConstructor;

    private static final String[] emptyNames = { };
    private static final Object[] emptyParams = { };


    public ConstructorsDescriptor(Class descClass,
                                  String[][] constrProps,
                                  Object[] defParams) {
        describedClass = descClass;

        if (constrProps != null && constrProps.length > 0) {
            constructors = new ConstructorCreator[constrProps.length];
            for (int i=0; i < constrProps.length; i++)
                constructors[i] = new ConstructorCreator(describedClass,
                                                         constrProps[i]);
        }

        defaultConstrParams = defParams;
    }

    public ConstructorsDescriptor(Class descClass) {
        describedClass = descClass;

        constructors = new ConstructorCreator[1];
        constructors[0] = new ConstructorCreator(describedClass, emptyNames);

        defaultConstrParams = emptyParams;
    }

    // ---------

    public Class getDescribedClass() {
        return describedClass;
    }

    public Creator[] getCreators() {
        return constructors;
    }

    public Creator findBestCreator(FormProperty[] properties, int style) {
        if (constructors == null) return null;

        int[] evals = CreationFactory.evaluateCreators(
                        constructors, properties, (style & CHANGED_ONLY) != 0);
        int best = CreationFactory.getBestCreator(
                     constructors, evals, (style & PLACE_ALL) != 0);
        return constructors[best];
    }

    public Object createDefaultInstance() throws InstantiationException,
                                                 IllegalAccessException,
                                                 IllegalArgumentException,
                                                 InvocationTargetException
    {
        if (defaultConstrParams == null || defaultConstrParams.length == 0)
            return describedClass.newInstance();

        if (defaultConstructor == null) {
            findDefaultConstructor();
            if (defaultConstructor == null)
                throw new IllegalArgumentException();
        }
        return defaultConstructor.newInstance(defaultConstrParams);
    }

    // ----------

    // finds first constructor that matches defaultConstrParams
    private void findDefaultConstructor() {
        Constructor[] constructors = describedClass.getConstructors();
        for (int i=0; i < constructors.length; i++) {
            Class[] paramTypes = constructors[i].getParameterTypes();
            if (paramTypes.length == defaultConstrParams.length) {
                int ii;
                for (ii=0; ii < paramTypes.length; ii++) {
                    Class cls = paramTypes[ii];
                    Object param = defaultConstrParams[ii];

                    if (cls.isPrimitive()) {
                        if (param == null
                            || (param instanceof Integer && cls != Integer.TYPE)
                            || (param instanceof Boolean && cls != Boolean.TYPE)
                            || (param instanceof Double && cls != Double.TYPE)
                            || (param instanceof Long && cls != Long.TYPE)
                            || (param instanceof Float && cls != Float.TYPE)
                            || (param instanceof Short && cls != Short.TYPE)
                            || (param instanceof Byte && cls != Byte.TYPE)
                            || (param instanceof Character && cls != Character.TYPE))
                        break;
                    }
                    else if (param != null && !cls.isInstance(param))
                        break;
                }
                if (ii == paramTypes.length) {
                    defaultConstructor = constructors[i];
                    return;
                }
            }
        }
    }

    // ----------

    static class ConstructorCreator implements Creator {
        private Class theClass;
        private String[] constructorPropsNames;

        ConstructorCreator(Class cls, String[] propsNames) {
            theClass = cls;
            constructorPropsNames = propsNames;
        }

        public final int getPropertiesCount() {
            return constructorPropsNames.length;
        }

        public final String[] getPropertiesNames() {
            return constructorPropsNames;
        }

        public Object createInstance(FormProperty[] props)
            throws InstantiationException, IllegalAccessException,
                   IllegalArgumentException, InvocationTargetException {

            Constructor constr = null;
            Object[] paramValues;

            try {
                paramValues = new Object[constructorPropsNames.length];
                Class[] paramTypes = new Class[constructorPropsNames.length];

                for (int i=0; i < constructorPropsNames.length; i++) {
                    FormProperty prop = CreationFactory.findProperty(
                                            constructorPropsNames[i], props);
                    if (prop == null) return null; // should not happen

                    paramValues[i] = prop.getRealValue();
                    paramTypes[i] = prop.getValueType();
                }

                constr = theClass.getConstructor(paramTypes);
            }
            catch (Exception ex) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                    ex.printStackTrace();
                throw new InstantiationException(ex.getMessage());
            }

            return constr.newInstance(paramValues);
        }

        public String getJavaCreationCode(FormProperty[] props) {
            StringBuffer buf = new StringBuffer();
            buf.append("new "); // NOI18N
            buf.append(theClass.getName().replace('$', '.'));
            buf.append("("); // NOI18N

            for (int i=0; i < constructorPropsNames.length; i++) {
                FormProperty prop = CreationFactory.findProperty(
                                        constructorPropsNames[i], props);
                if (prop == null) return null; // should not happen

                buf.append(prop.getJavaInitializationString());
                if (i+1 < constructorPropsNames.length)
                    buf.append(", "); // NOI18N
            }

            buf.append(")"); // NOI18N
            return buf.toString();
        }
    }
}
