/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form;

import java.util.Vector;
import java.util.Enumeration;
import java.io.Serializable;

/**
 *
 * @author Tran Duc Trung
 */

/**
 * A class that represents one Event handler method.
 * It holds a list of Events that are associated with the handler.
 */
public class EventHandler implements Serializable
{
    private CodeGenerator codeGenerator;
        
    transient private Vector attachedEvents;
    private String handlerName;
    private String[] parameterTypes;
    private String[] exceptionTypes;
        
    static final long serialVersionUID = -4103904336344477509L;
        
    EventHandler(CodeGenerator codeGenerator,
                 Event event,
                 String name) {
        this.codeGenerator = codeGenerator;
        handlerName = name;
        Class[] parameters = event.getListenerMethod().getParameterTypes();

        parameterTypes = new String[parameters.length];
        for (int i = 0; i < parameters.length; i++)
            parameterTypes [i] = parameters [i].getName();

        Class[] exceptions = event.getListenerMethod().getExceptionTypes();
        exceptionTypes = new String[exceptions.length];
        for (int i = 0; i < exceptions.length; i++)
            exceptionTypes[i] = exceptions[i].getName();

        event.addHandler(this);
        getAttachedEvents().addElement(event);
    }

    /**
     * Called to actually create the handler method in the text
     * @param bodyText The body text of the event or null for
     * default(empty) body
     */
        
    void generateHandler(String bodyText) {
//              if (formModel.isInitialized()) { // [PENDING - what to do during deserializaion about restoring event handlers?]
//                  formModel.getCodeGenerator().generateEventHandler(handlerName, parameterTypes, bodyText);
//              }
        codeGenerator.generateEventHandler(handlerName, parameterTypes, 
                                           exceptionTypes, bodyText);
    }

    /**
     * Called to change the handler text
     * @param bodyText The body text of the event or null for
     * default(empty) body
     */
        
    void setHandlerText(String bodyText) {
        codeGenerator.changeEventHandler(handlerName, parameterTypes, 
                                         exceptionTypes, bodyText);
    }

    void rename(String newName) {
        // 1. regenerate the event handler method
        codeGenerator.renameEventHandler(handlerName, newName,
                                         parameterTypes, exceptionTypes);
        handlerName = newName;

        // 2. set the events
        for (Enumeration e = getAttachedEvents().elements(); e.hasMoreElements();) {
            Event event =(Event)e.nextElement();
        }
    }

    void delete() {
        codeGenerator.deleteEventHandler(handlerName);
    }

    void attachEvent(Event event) {
        event.addHandler(this);
        getAttachedEvents().addElement(event);
    }

    void detachEvent(Event event) {
        event.removeHandler(this);
        getAttachedEvents().removeElement(event);
    }

    public String getName() {
        return handlerName;
    }

    public String toString() {
        return handlerName;
    }

    public Vector getAttachedEvents() {
        if (attachedEvents == null)
            attachedEvents = new Vector();
        return attachedEvents;
    }

    public boolean isEmpty() {
        return(attachedEvents.size() == 0);
    }

    /** Checks whether the specified event is parameter-compatible with this handler
     * @param event the event to check
     * @return true if the event is parameter-compatible with this handler, false otherwise
     */
    boolean checkCompatibility(Event event) {
        Class[] parameters = event.getListenerMethod().getParameterTypes();
        if (parameters.length != parameterTypes.length)
            return false;
        for (int i = 0; i < parameterTypes.length; i++)
            if (!parameterTypes [i].equals(parameters [i].getName()))
                return false;
        return true;
    }

}
