/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form;

import java.io.Serializable;
import java.io.Externalizable;
import java.text.MessageFormat;
import java.util.Vector;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Enumeration;

import org.openide.*;
import org.openide.util.Utilities;

/**
 * A class that manages the pool of event handlers for one form.
 */

public class FormEventHandlers
{
    /** The FormModel that manages the form */
    private CodeGenerator codeGenerator;
    
    /** The list of event handlers(mapping name-> eventHandler) */
    private HashMap nameToHandler = new HashMap();
    
    FormEventHandlers(CodeGenerator codeGenerator) {
        this.codeGenerator = codeGenerator;
    }

    /**
     * Returns event handler of given name.
     */
    EventHandler getEventHandler(String handlerName) {
        return (EventHandler) nameToHandler.get(handlerName);
    }

    /**
     * Returns all event handlers in the form.
     * @return array of EventHandler objects
     */
    Object[] getAllHandlers() {
        return nameToHandler.values().toArray();
    }

    /**
     * Adds a new event handler method. If such event handler already exists,
     * the specified event is attached to it, otherwise a new event handler is
     * created.  If attaching to an existing handler, the attached event must
     * be of the same type as the existing handler(i.e. the parameter types
     * must be same).
     * @param event     The event for which to add the event handler
     * @param eventName The name of the event handler method
     * @return whether the operation was successful (handler is compatible
     * if already exists)
     */
    boolean addEventHandler(Event event, String eventName) {
        return addEventHandler(event, eventName, null);
    }

    /**
     * Adds a new event handler method. If such event handler already exists,
     * the specified event is attached to it, otherwise a new event handler is
     * created.  If attaching to an existing handler, the attached event must
     * be of the same type as the existing handler(i.e. the parameter types
     * must be same).
     * @param event The event for which to add the event handler
     * @param eventName The name of the event handler method
     * @param bodyText The text of the event handler body or null for
     * default(empty) text
     * @return whether the operation was successful (handler is compatible
     * if it already exists)
     */
    boolean addEventHandler(Event event, String eventName, String bodyText) {
        EventHandler handler = (EventHandler) nameToHandler.get(eventName);
        
        if (handler != null) {
             // if the handler already exist, attach to it
            if (!handler.checkCompatibility(event)) {
                TopManager.getDefault().notify(
                    new NotifyDescriptor.Message(
                        FormEditor.getFormBundle().getString("MSG_CannotAttach"),
                        NotifyDescriptor.WARNING_MESSAGE));
                return false;
            }
            handler.attachEvent(event);
        } 
        else {
            // otherwise create a new one
            handler = new EventHandler(codeGenerator, event, eventName);
            handler.generateHandler(bodyText);
            nameToHandler.put(eventName, handler);
        }
        return true;
    }

    /**
     * Detaches the specified event from all of its handlers.
     * @param event the event to deattach from all of its handlers.
     */
    void removeEventHandler(Event event) {
        Vector hands = event.getHandlers();
        for (int i = 0, n = hands.size(); i < n; i++) {
            removeEventHandler(event, (EventHandler) hands.get(0));
        }
    }

    void removeEventHandler(Event event, String handlerName) {
        EventHandler handler = (EventHandler) nameToHandler.get(handlerName);
        removeEventHandler(event, handler);
    }

    /**
     * Detaches the specified event from its handler.
     * @param event the event to deattach from its handler.
     */
    void removeEventHandler(Event event, EventHandler handler) {
        if (handler == null) return;

        if (handler.getAttachedEvents().size() == 1) {
            if (codeGenerator.isEventHandlerEmpty(handler.getName())) {
                handler.detachEvent(event);
                deleteEventHandler(handler);
            }
            else {
                String message = MessageFormat.format(
                    FormEditor.getFormBundle().getString("FMT_MSG_EventDelete"),
                    new Object [] { handler.getName() });
                Object result = TopManager.getDefault().notify(
                    new NotifyDescriptor.Confirmation(message,
                                                      NotifyDescriptor.YES_NO_OPTION,
                                                      NotifyDescriptor.WARNING_MESSAGE));
                if (NotifyDescriptor.YES_OPTION.equals(result)) {
                    handler.detachEvent(event);
                    deleteEventHandler(handler);
                } 
                else
                    handler.detachEvent(event);
            }
        } 
        else
            handler.detachEvent(event);
    }

    void deleteEventHandler(EventHandler handler) {
        handler.delete();
        nameToHandler.remove(handler.getName());
    }

    void renameEventHandler(String oldName, String newName) {
        EventHandler handler = (EventHandler) nameToHandler.get(oldName);
        renameEventHandler(handler, newName);
    }

    /**
     * Renames the specified handler to the given name.
     * @param handler The EventHandler to rename
     * @param newName The new name of the event handler
     */
    void renameEventHandler(EventHandler handler, String newName) {
        if (handler == null) return;

        if (nameToHandler.get(newName) != null) { // new name already exists
            TopManager.getDefault().notify(
                new NotifyDescriptor.Message(
                    FormEditor.getFormBundle().getString("MSG_CannotRename"),
                    NotifyDescriptor.WARNING_MESSAGE)
                );
            return;
        }

        nameToHandler.remove(handler.getName());
        nameToHandler.put(newName, handler);
        handler.rename(newName);
    }

    /**
     * Allows to acquire an EventHandler of specified name.
     * @return The EventHandler of specified name
     */
    EventHandler findEventHandler(String name) {
        return(EventHandler)nameToHandler.get(name);
    }

}
