/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form;

import java.util.ArrayList;

import org.openide.actions.*;
import org.openide.cookies.*;
import org.openide.nodes.*;
import org.openide.util.HelpCtx;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.form.layoutsupport.LayoutSupport;
import org.netbeans.modules.form.layoutsupport.LayoutNode;


/**
 *
 * @author Ian Formanek
 */
    
public class RADChildren extends Children.Keys
{
    /** Icon base for non visuals node. */
    private static final String NON_VISUAL_ICON_BASE = "org/netbeans/modules/form/resources/formNonVisual"; // NOI18N

    private ComponentContainer container;

    private static final Object KEY_NON_VISUAL = new Object();
    private Object keyLayout;

    /** Creates new RADChildren for specified ComponentContainer
     * @param container The component container for which this children should be created
     */
    public RADChildren(ComponentContainer container) {
        super();
        this.container = container;
        updateKeys();
    }

    void updateKeys() {
        RADComponent[] subComps = container.getSubBeans();
        ArrayList keys = new ArrayList(subComps.length + 2);

        if (container instanceof FormContainer)
            keys.add(KEY_NON_VISUAL);

        if (container instanceof RADVisualContainer
                && ((RADVisualContainer)container).shouldHaveLayoutNode()) {
            keyLayout = new Object();
            keys.add(keyLayout);
        }

        for (int i = 0; i < subComps.length; i++)
            keys.add(subComps[i]);

        setKeys(keys);
    }

    /** Create nodes for a given key.
     * @param key the key
     * @return child nodes for this key
     */
    protected Node[] createNodes(Object key) {
        if (key == KEY_NON_VISUAL) {
            NonVisualNode node = new NonVisualNode(
                           ((RADVisualFormContainer)container).getFormModel());
            node.getChildren().getNodes(); // enforce subnodes creation
            return new Node[] { node };
        }
        else if (key == keyLayout) {
            return new Node[] { new LayoutNode((RADVisualContainer)container) };
        }
        else {
            Node newNode = new RADComponentNode((RADComponent)key);
            newNode.getChildren().getNodes(); // enforce subnodes creation
            return new Node[] { newNode };
        }
    }

    private final static class NonVisualNode extends AbstractNode {
        private FormModel manager;

        NonVisualNode(FormModel manager) {
            super(manager.getNonVisualChildren());
            this.manager = manager;
            setIconBase(NON_VISUAL_ICON_BASE);
            setName(FormEditor.getFormBundle().getString("CTL_NonVisualComponents"));
            getCookieSet().add(new NonVisualIndex(manager));
        }

        public HelpCtx getHelpCtx() {
            return new HelpCtx(NonVisualNode.class);
        }

        /** Lazily initialize set of node's actions(overridable).
         * The default implementation returns <code>null</code>.
         * <p><em>Warning:</em> do not call {@link #getActions} within this method.
         * If necessary, call {@link NodeOp#getDefaultActions} to merge in.
         * @return array of actions for this node, or <code>null</code> to use the default node actions
         */
        protected SystemAction [] createActions() {
            return manager.isReadOnly() ? 
                new SystemAction[] {
//                    SystemAction.get(ToolsAction.class),
                    SystemAction.get(PropertiesAction.class),
                }
                :
                new SystemAction[] {
                    SystemAction.get(ReorderAction.class),
                    null,
//                    SystemAction.get(ToolsAction.class),
                    SystemAction.get(PropertiesAction.class),
                };
        }

        /** Get a cookie from the node.
         * Uses the cookie set as determined by {@link #getCookieSet}.
         *
         * @param type the representation class
         * @return the cookie or <code>null</code>
         */
        public Node.Cookie getCookie(Class type) {
            Node.Cookie inh = super.getCookie(type);
            if (inh == null) {
                if (CompilerCookie.class.isAssignableFrom(type) ||
                    SaveCookie.class.isAssignableFrom(type) ||
                    ExecCookie.class.isAssignableFrom(type) ||
                    DebuggerCookie.class.isAssignableFrom(type) ||
                    CloseCookie.class.isAssignableFrom(type) ||
                    ArgumentsCookie.class.isAssignableFrom(type) ||
                    PrintCookie.class.isAssignableFrom(type)) {
                    return manager.getFormDataObject().getCookie(type);
                }
            }
            return inh;
        }

    }

    /** Index support for reordering of file system pool.
     */
    private final static class NonVisualIndex extends org.openide.nodes.Index.Support {
        private FormModel manager;

        NonVisualIndex(FormModel manager) {
            this.manager = manager;
        }

        /** Get the nodes; should be overridden if needed.
         * @return the nodes
         * @throws NotImplementedException always
         */
        public Node[] getNodes() {
            return manager.getNonVisualChildren().getNodes();
        }

        /** Get the node count. Subclasses must provide this.
         * @return the count
         */
        public int getNodesCount() {
            return getNodes().length;
        }

        /** Reorder by permutation. Subclasses must provide this.
         * @param perm the permutation
         */
        public void reorder(int[] perm) {
            manager.reorderNonVisualComponents(perm);
            manager.getNonVisualChildren().updateKeys(manager);
        }

    }
}
