/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileLock;
import org.openide.util.NbBundle;
import org.netbeans.modules.form.forminfo.*;
import org.netbeans.modules.form.*;

import org.netbeans.modules.form.layoutsupport.*;

import org.netbeans.modules.form.compat2.*;
import org.netbeans.modules.form.compat2.layouts.DesignLayout;

import java.beans.BeanInfo;
import java.beans.PropertyDescriptor;
import java.io.*;
import java.util.*;

/**
 *
 * @author Ian Formanek
 */
public class TuborgPersistenceManager extends PersistenceManager {
    /** Constant used for signing the classes loaded by default classloader */
    final static byte LOADER_DEFAULT = 0x01;

    // Constants to recognize form files saved from Tuborg
    private static byte MAGIC_0 = -84;
    private static byte MAGIC_1 = -19;
    private static byte MAGIC_2 = 0;
    private static byte MAGIC_3 = 5;

    // FINALIZE DEBUG METHOD
    public void finalize() throws Throwable {
        super.finalize();
        if (System.getProperty("netbeans.debug.form.finalize") != null) {
            System.out.println("finalized: "+this.getClass().getName()+", instance: "+this); // NOI18N
        }
    } // FINALIZE DEBUG METHOD

    /** A method which allows the persistence manager to provide infotrmation on whether
     * is is capable to store info about advanced features provided from Developer 3.0 
     * - all persistence managers except the one providing backward compatibility with 
     * Developer 2.X should return true from this method.
     * @return true if this PersistenceManager is capable to store advanced form features, false otherwise
     */
    public boolean supportsAdvancedFeatures() {
        return false;
    }

    /** A method which allows the persistence manager to check whether it can read
     * given form format.
     * @return true if this PersistenceManager can load form stored in the specified form, false otherwise
     * @exception IOException if any problem occured when accessing the form
     */
    public boolean canLoadForm(FormDataObject formObject) throws IOException {
        InputStream is = null;
        try {
            is = formObject.getFormEntry().getFile().getInputStream();
            byte[] bytes = new byte[4];
            int len = is.read(bytes);
            return((len == 4) &&(bytes[0] == MAGIC_0) &&(bytes[1] == MAGIC_1) &&(bytes[2] == MAGIC_2) &&(bytes[3] == MAGIC_3));
        } catch (Throwable t) {
            if (t instanceof ThreadDeath) {
                throw(ThreadDeath)t;
            }
            return false;
        }
        finally {
            if (is != null)
                is.close();
        }
    }

    /** Called to actually load the form stored in specified formObject.
     * @param formObject the FormDataObject which represents the form files
     * @return the FormModel representing the loaded form or null if some problem occured
     * @exception IOException if any problem occured when loading the form
     */
    public FormModel loadForm(FormDataObject formObject) throws IOException {
        FormModel formModel = null;
        
        InputStream is = null;
        try {
            is = formObject.getFormEntry().getFile().getInputStream();
        } catch (FileNotFoundException e) {
            throw new IOException("Form file not found"); // NOI18N
        }

        ObjectInputStream ois = null;
        try {
            ois = new org.openide.util.io.NbObjectInputStream(is);

            // deserialization from stream
            Object deserializedForm = null;
            try {
                deserializedForm = ois.readObject();
            } catch (ClassNotFoundException e) {
                throw new IOException("Form type not found: "+e.getMessage()); // NOI18N
            }

            // create new objects from Backward compatibility classes

            if (deserializedForm == null) {
                throw new IOException("The form file does not contain a valid form"); // NOI18N
            } else if (!(deserializedForm instanceof DesignForm)) {
                throw new IOException("Unknown form type: "+deserializedForm.getClass().getName()); // NOI18N
            } else {
                FormInfo info = null;
                if (deserializedForm instanceof JFrameForm) {
                    info = new JFrameFormInfo();
                } else if (deserializedForm instanceof JDialogForm) {
                    info = new JDialogFormInfo();
                } else if (deserializedForm instanceof JPanelForm) {
                    info = new JPanelFormInfo();
                } else if (deserializedForm instanceof JAppletForm) {
                    info = new JAppletFormInfo();
                } else if (deserializedForm instanceof JInternalFrameForm) {
                    info = new JInternalFrameFormInfo();
                } else if (deserializedForm instanceof FrameForm) {
                    info = new FrameFormInfo();
                } else if (deserializedForm instanceof DialogForm) {
                    info = new DialogFormInfo();
                } else if (deserializedForm instanceof PanelForm) {
                    info = new PanelFormInfo();
                } else if (deserializedForm instanceof AppletForm) {
                    info = new AppletFormInfo();
                }

                formModel = new FormModel();
                formModel.setFormInfo(info);
                formModel.setFormDataObject(formObject);

                RADForm radForm = new RADForm(info);
                RADVisualContainer topComp =(RADVisualContainer)radForm.getTopLevelComponent();
                RADFormNode topNode = ((DesignForm)deserializedForm).formManager.rootNode;

                // process non-visual components
                Object[] nonVisualNodes = topNode.nonVisualsNode.nodeArray.array;
                RADComponent[] nonVisualsComps = new RADComponent [nonVisualNodes.length];
                for (int i = 0; i < nonVisualNodes.length; i++) {
                    RADNode node =(RADNode)nonVisualNodes[i];
                    if (node instanceof RADMenuNode) {
                        nonVisualsComps[i] = new RADMenuComponent();
                        //XXX nonVisualsComps[i].initialize(formModel2);
                        nonVisualsComps[i].setComponent(node.beanClass);
                        nonVisualsComps[i].setName(node.componentName);
                        convertComponent(node, nonVisualsComps[i]);
                        //XXX RADComponent[] subs = createMenuHierarchy((RADMenuNode)node, formModel2);
                        //XXX ((RADMenuComponent)nonVisualsComps[i]).initSubComponents(subs);
                    } else {
                        nonVisualsComps[i] = new RADComponent();
                        //XXX nonVisualsComps[i].initialize(formModel2);
                        nonVisualsComps[i].setComponent(node.beanClass);
                        nonVisualsComps[i].setName(node.componentName);
                        convertComponent(node, nonVisualsComps[i]);
                    }
                }
                formModel.initNonVisualComponents(nonVisualsComps);

                convertComponent(topNode, topComp);
                RADComponent[] subComps = createHierarchy((((DesignForm)deserializedForm).formManager.rootNode), formModel);
                topComp.initSubComponents(subComps);
                //XXX topComp.setDesignLayout(((RADContainerNode)topNode).designLayout);
                LayoutSupport layoutSupp =
                    Compat31LayoutFactory.createCompatibleLayoutSupport(
                        ((RADContainerNode)topNode).designLayout);
                topComp.setLayoutSupport(layoutSupp);
                
                String menu =((RADFormNode)topNode).menu;
                if ((menu != null) &&(topComp instanceof RADVisualFormContainer)) {
                    ((RADVisualFormContainer)topComp).setFormMenu(menu);
                }
            }
        } finally {
            if (ois != null) {
                try {
                    ois.close();
                } catch (IOException e) {
                    if (System.getProperty("netbeans.debug.form") != null) {
                        e.printStackTrace();
                    }
                }
            }
        }
        return formModel;
    }

    /** Called to actually save the form represented by specified FormModel into specified formObject.
     * @param formObject the FormDataObject which represents the form files
     * @param manager the FormModel representing the form to be saved
     * @exception IOException if any problem occured when saving the form
     */
    public void saveForm(FormDataObject formObject, FormModel formModel) throws IOException {
        FileLock lock = null;
        try {
            lock = formObject.getFormEntry().getFile().lock();
            OutputStream os = formObject.getFormEntry().getFile().getOutputStream(lock);

            // we first save into memory to prevent corrupting the form file
            // if something goes wrong
            java.io.ByteArrayOutputStream barros = new java.io.ByteArrayOutputStream(10000);
            ObjectOutputStream oos = new X2ObjectOutputStream(barros);

            oos.writeObject(createOldForm(formObject, formModel));
            oos.close();

            // now it is safely written in memory, so we can save it to the file
            barros.writeTo(os);

        } catch (IOException e) {
            TopManager.getDefault().notify(new NotifyDescriptor.Message(
                java.text.MessageFormat.format(
                    NbBundle.getBundle(FormEditorSupport.class).getString("ERR_SavingForm"),
                    new Object[] { formObject.getName(), e.getClass().getName()}
                    ),
                NotifyDescriptor.ERROR_MESSAGE
                )
                                           );
            e.printStackTrace();
        } finally {
            if (lock != null) {
                lock.releaseLock();
            }
        }

    }

    // -----------------------------------------------------------------------------
    // private methods -> Backward compatibility

    private static RADComponent[] createHierarchy(RADContainerNode node, FormModel formModel2) {
        RADNode nodes[] = node.getSubNodes();
        RADComponent[] comps = new RADComponent [nodes.length];
        for (int i = 0; i < nodes.length; i++) {
            if (nodes[i] instanceof RADContainerNode) {
                comps[i] = new RADVisualContainer();
            } else if (nodes[i] instanceof RADVisualNode) {
                comps[i] = new RADVisualComponent();
            } else {
                comps[i] = new RADComponent();
            }
            comps[i].initialize(formModel2);
            comps[i].setComponent(nodes[i].beanClass);
            comps[i].setName(nodes[i].componentName);
            convertComponent(nodes[i], comps[i]);
            if (nodes[i] instanceof RADContainerNode) {
                RADComponent[] subs = createHierarchy((RADContainerNode)nodes[i], formModel2);
                ((ComponentContainer)comps[i]).initSubComponents(subs);
                LayoutSupport layoutSupp =
                    Compat31LayoutFactory.createCompatibleLayoutSupport(
                        ((RADContainerNode)nodes[i]).designLayout);
                ((RADVisualContainer)comps[i]).setLayoutSupport(layoutSupp);
                //XXX ((RADVisualContainer)comps[i]).setDesignLayout(((RADContainerNode)nodes[i]).designLayout);
            }
        }

        return comps;
    }

    private static RADComponent[] createMenuHierarchy(RADMenuNode node, FormModel formModel2) {
        Object nodes[] = node.getSubNodes();
        RADComponent[] comps = new RADComponent [nodes.length];
        for (int i = 0; i < nodes.length; i++) {

            Class beanClass;
            if (nodes[i] instanceof RADMenuNode.MenuSeparatorNode) {
                comps[i] = new RADMenuItemComponent();
                if ((node.type == RADMenuItemNode.T_MENU) ||(node.type == RADMenuItemNode.T_POPUPMENU)) {
                    beanClass = org.netbeans.modules.form.Separator.class;
                } else {
                    beanClass = javax.swing.JSeparator.class;
                }
                comps[i].initialize(formModel2);
                comps[i].setComponent(beanClass);
                if ((node.type == RADMenuItemNode.T_JMENU) ||(node.type == RADMenuItemNode.T_JPOPUPMENU)) {
                    // only swing separators have instance variable
                    comps[i].setName(formModel2.getVariablePool().getNewName(beanClass));
                }
            } else if (nodes[i] instanceof RADNode) {
                beanClass =((RADNode)nodes[i]).beanClass;
                if (nodes[i] instanceof RADMenuNode) {
                    comps[i] = new RADMenuComponent();
                } else if (nodes[i] instanceof RADMenuItemNode) {
                    comps[i] = new RADMenuItemComponent();
                } else {
                    comps[i] = new RADComponent();
                }
                comps[i].initialize(formModel2);
                comps[i].setComponent(beanClass);
                comps[i].setName(((RADNode)nodes[i]).componentName);
                convertComponent((RADNode)nodes[i], comps[i]);
                if (nodes[i] instanceof RADMenuNode) {
                    RADComponent[] subs = createMenuHierarchy((RADMenuNode)nodes[i], formModel2);
                    ((ComponentContainer)comps[i]).initSubComponents(subs);
                }
            }
        }

        return comps;
    }

    private static void convertComponent(RADNode node, RADComponent comp) {
        Map origChanged = node.changedValues;
        BeanInfo bi = comp.getBeanInfo();
        PropertyDescriptor[] pds = bi.getPropertyDescriptors();
        for (Iterator it = origChanged.keySet().iterator(); it.hasNext();) {
            Object key = it.next();
            for (int i = 0; i < pds.length; i++) {
                if (key.equals(pds[i].getName())) {
                    try {
                        RADProperty prop = comp.getPropertyByName(pds[i].getName());
                        if (prop != null)
                            prop.setValue(origChanged.get(key));
//                        comp.restorePropertyValue(pds[i], origChanged.get(key));
                    } catch (IllegalArgumentException e) {
                        // [PENDING]
                    } catch (IllegalAccessException e) {
                        // [PENDING]
                    } catch (java.lang.reflect.InvocationTargetException e) {
                        // [PENDING]
                    }
                    break;
                }
            }
        }

        Hashtable eventHandlers = node.eventHandlers;
        comp.initDeserializedEvents(eventHandlers);

        // process constraints on visual components
        if (node instanceof RADVisualNode) {
            //XXX HashMap map =((RADVisualNode)node).constraints;
            //XXX ((RADVisualComponent)comp).initConstraints(map);
        }
    }

    // -----------------------------------------------------------------------------
    // private methods -> Forward compatibility

    private static DesignForm createOldForm(FormDataObject formObject, FormModel formModel) {
        DesignForm designForm = null;
        FormInfo info = formModel.getFormInfo();
        if (info instanceof JFrameFormInfo) {
            designForm = new JFrameForm();
        } else if (info instanceof JDialogFormInfo) {
            designForm = new JDialogForm();
        } else if (info instanceof JPanelFormInfo) {
            designForm = new JPanelForm();
        } else if (info instanceof JAppletFormInfo) {
            designForm = new JAppletForm();
        } else if (info instanceof JInternalFrameFormInfo) {
            designForm = new JInternalFrameForm();
        } else if (info instanceof FrameFormInfo) {
            designForm = new FrameForm();
        } else if (info instanceof DialogFormInfo) {
            designForm = new DialogForm();
        } else if (info instanceof PanelFormInfo) {
            designForm = new PanelForm();
        } else if (info instanceof AppletFormInfo) {
            designForm = new AppletForm();
        }

        FormManager manager = new FormManager();

        RADFormNode rootNode = new RADFormNode();

        NonVisualsNode nonVisualsNode = new NonVisualsNode();

        RADNodeArray nodeArray = new RADNodeArray();
        // >> RADNodeArray
        // 1. containerNode [RADContainer]
        // 2. array [Object[]]
        nodeArray.containerNode = nonVisualsNode;
        RADComponent[] nonVisualComponents = formModel.getNonVisualComponents();
        RADNode[] nonVisualNodes = new RADNode[nonVisualComponents.length];
        for (int i = 0; i < nonVisualComponents.length; i++) {
            if (nonVisualComponents[i] instanceof RADMenuComponent) {
                nonVisualNodes[i] = new RADMenuNode();
            } else {
                nonVisualNodes[i] = new RADNode();
            }
            convertComponentBack(nonVisualComponents[i], nonVisualNodes [i]);
        }
        nodeArray.array = nonVisualNodes;

        // >> NonVisualsNode
        // 1. designForm [DesignForm]
        // 2. rootNode [RADFormNode]
        nonVisualsNode.formManager = manager;
        nonVisualsNode.nodeArray = nodeArray;

        // >> RADFormNode
        // From RADFormNode:
        //  1. form [DesignForm]
        //  2. nonVisualsNode [NonVisualsNode]
        //  3. menu [String]
        rootNode.form = designForm;
        rootNode.nonVisualsNode = nonVisualsNode;
        rootNode.menu = null;
        if (formModel.getTopRADComponent() instanceof RADVisualFormContainer) {
            rootNode.menu =((RADVisualFormContainer)formModel.getTopRADComponent()).getFormMenu();
        }

        convertComponentBack(formModel.getTopRADComponent(), rootNode);

        // >> FormModel
        // 1. designForm [DesignForm]
        // 2. rootNode [RADFormNode]
        manager.designForm = designForm;
        manager.rootNode = rootNode;


        designForm.formManager = manager;

        return designForm;
    }

    private static void convertComponentBack(RADComponent comp, RADNode contNode) {
        if (contNode instanceof RADMenuNode) {
            // >> RADMenuNode
            //  1. nodeArray [RADNodeArray]

            RADNodeArray nodeArray = new RADNodeArray();
            // >> RADNodeArray
            // 1. containerNode [RADContainer]
            // 2. array [Object[] ]
            nodeArray.containerNode =(RADMenuNode)contNode;
            RADComponent[] children =((RADMenuComponent)comp).getSubBeans();
            RADNode[] childrenNodes = new RADNode [children.length];

            for (int i = 0; i < children.length; i++) {
                if (children[i] instanceof RADMenuComponent) {
                    childrenNodes[i] = new RADMenuNode();
                } else if (children[i] instanceof RADMenuItemComponent) {
                    childrenNodes[i] = new RADMenuItemNode();
                } else {
                    childrenNodes[i] = new RADNode(); // [PENDING - can X2 menu contain RADNodes?]
                }
                convertComponentBack(children[i], childrenNodes[i]);
            }

            nodeArray.array = childrenNodes;

            ((RADMenuNode)contNode).nodeArray = nodeArray;
        }

        if (contNode instanceof RADContainerNode) {
            // >> RADContainerNode
            //  1. nodeArray [RADNodeArray]
            //  2. designLayout [DesignLayout]
            //  3. radLayoutNode [RADLayoutNode]

            RADNodeArray nodeArray = new RADNodeArray();
            // >> RADNodeArray
            // 1. containerNode [RADContainer]
            // 2. array [Object[] ]
            nodeArray.containerNode =(RADContainerNode)contNode;
            RADComponent[] children =((RADVisualContainer)comp).getSubComponents();
            RADNode[] childrenNodes = new RADNode [children.length];

            for (int i = 0; i < children.length; i++) {
                if (children[i] instanceof RADVisualContainer) {
                    childrenNodes[i] = new RADContainerNode();
                } else if (children[i] instanceof RADVisualComponent) {
                    childrenNodes[i] = new RADVisualNode();
                } else {
                    childrenNodes[i] = new RADNode();
                }
                convertComponentBack(children[i], childrenNodes[i]);
            }

            nodeArray.array = childrenNodes;
/* XXX
            DesignLayout designLayout =((RADVisualContainer)comp).getDesignLayout();

            org.netbeans.modules.form.compat2.RADLayoutNode layoutNode = new org.netbeans.modules.form.compat2.RADLayoutNode();
            // >> RADLayoutNode
            // 1. designLayout [DesignLayout]
            // 2. layoutName [String](can be "" ???) // NOI18N
            layoutNode.designLayout = designLayout;
            layoutNode.layoutName = ""; // [PENDING - check] // NOI18N

            ((RADContainerNode)contNode).nodeArray = nodeArray;
            ((RADContainerNode)contNode).designLayout = designLayout;
            ((RADContainerNode)contNode).radLayoutNode = layoutNode;
***/
        }

        if (contNode instanceof RADVisualNode) {
            // >> RADVisualNode
            // 1. designLayout [DesignLayout]
            // 2. layoutName [String](can be "" ???) // NOI18N
//XXX             ((RADVisualNode)contNode).constraints =((RADVisualComponent)comp).getConstraintsMap();
        }

        Hashtable changedProperties = new Hashtable();
        Hashtable handlersTable = new Hashtable();
        HashMap propertiesMap = new HashMap();
        RADProperty[] newProperties = comp.getAllBeanProperties();
        for (int i = 0; i < newProperties.length; i++) {
            if (newProperties[i].isChanged()) {
                Object changedValue = null;
                try {
                    changedValue = newProperties[i].getValue();
                } catch (Exception e) {
                    if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
                    // problem getting value => ignore this property
                    continue;
                }
                String propName = newProperties[i].getName();
                changedProperties.put(propName, propName);
                propertiesMap.put(propName, changedValue);
            }
        }

        // >> RADNode
        // 1. beanClass [Class]
        // 2. changedProperties [Hashtable]
        // 3. handlersTable [Hashtable]
        // 4. componentName [String]
        // 5. hasHiddenState [boolean]
        // 6. propertiesMap [Map]
        contNode.beanClass = comp.getBeanClass();
        contNode.changedProperties = changedProperties;
        contNode.handlersTable = comp.getEventHandlers().getEventNames();
        contNode.componentName = comp.getName();
        contNode.hasHiddenState = false; // [PENDING]
        contNode.propertiesMap = propertiesMap;
    }

    // -----------------------------------------------------------------------------
    // X2 Object Streams
    public static class X2ObjectOutputStream extends ObjectOutputStream {
        public X2ObjectOutputStream(OutputStream os) throws IOException {
            super(os);
        }

        /** Calls super annotateClass and then write the index of the classloader of this class.
         * @see NbObjectInputStream
         */
        protected void annotateClass(Class cl) throws IOException {
            super.annotateClass(cl);
            writeByte(LOADER_DEFAULT);
        }
    }

    // -----------------------------------------------------------------------------
    // Safe Serialization

    public static void writeSafely(ObjectOutput oo, Object obj)
        throws IOException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream(200);
        ObjectOutputStream oos = new X2ObjectOutputStream(bos); // [PENDING!!!]
        oos.writeObject(obj);
        oos.flush();
        bos.close();

        oo.writeInt(bos.size());
        oo.write(bos.toByteArray());
    }

    public static Object readSafely(ObjectInput oi)
        throws IOException, ClassNotFoundException {
        int size = oi.readInt();
        byte[] byteArray = new byte [size];
        oi.readFully(byteArray, 0, size);

        ByteArrayInputStream bis = new ByteArrayInputStream(byteArray);
        ObjectInputStream ois = new org.openide.util.io.NbObjectInputStream(bis);
        Object obj = ois.readObject();
        bis.close();

        return obj;
    }

}
