/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.compat2.layouts;

import java.awt.*;

import org.openide.nodes.Node;
import org.openide.nodes.PropertySupport;

import org.netbeans.lib.awtextra.AbsoluteConstraints;
import org.netbeans.lib.awtextra.AbsoluteLayout;
import org.netbeans.modules.form.util2.NbVersion;
import org.netbeans.modules.form.util2.NbVersionNotCompatibleException;
import org.netbeans.modules.form.*;

/** A design-time support for AbsoluteLayout.
 *
 * @author   Ian Formanek
 */
final public class DesignAbsoluteLayout extends DesignLayout {
    /** A JDK 1.1. serial version UID */
    static final long serialVersionUID = -1148154120636340460L;

    /** Netbeans class version */
    public static final NbVersion nbClassVersion = new NbVersion(1, 0);

    private static final String PROP_NULL_LAYOUT = "useNullLayout";
    private Node.Property nullLayoutProperty;
    private boolean useNullLayout = false;

    public DesignAbsoluteLayout() {
        super();
    }

    public DesignAbsoluteLayout(boolean useNullLayout) {
        this.useNullLayout = useNullLayout;
    }

    public boolean isNullLayout() {
        return useNullLayout;
    }

    public void setNullLayout(boolean useNullLayout) {
        this.useNullLayout = useNullLayout;
    }

    public java.util.List getChangedProperties() {
        getPropertySet(); // enforce creation of properties
        java.util.ArrayList list = new java.util.ArrayList(1);
        list.add(nullLayoutProperty);
        return list;
    }

    /** Method which is called after the layout is loaded with the form to
     * initialize its properties.
     * @param changedProperties map of <String, Object> pairs, where the String
     * is a name of property and the Object its value
     */
    public void initChangedProperties(java.util.Map changedProperties) {
        Boolean nullLayoutVal = (Boolean)changedProperties.get(PROP_NULL_LAYOUT);
        if (nullLayoutVal != null)
            useNullLayout = nullLayoutVal.booleanValue();
    }

    /** Returns a constraint to be used for adding a component to the
     * specified position(e.g. in BorderLayout, the "North", "South", ...
     * will be determined by the position in the container).
     * The default implementation just returns the default constraints.
     * A special constraints object will be returned for layouts that
     * wish to set the position&size of the components rather than add
     * with constraints.
     * @param position The position within the container for which the
     *                 constraints should be returned.
     */
    public DesignLayout.ConstraintsDescription getConstraintsDescription(Point position) {
        return new AbsoluteConstraintsDescription(position);
    }

    /** A display name of the layout will be used for displaying the layout in
     * the components hierarchy during design-time.
     * @return layout's display name.
     */
    public String getDisplayName() {
        return "AbsoluteLayout"; // NOI18N
    }

    /** Returns the global layout's properties(i.e. the properties
     * that are not different for different components in the layout
     * @return the global layout properties
     */
    public Node.PropertySet[] getPropertySet() {
        if (nullLayoutProperty == null) {
            nullLayoutProperty = new PropertySupport.ReadWrite(
                    PROP_NULL_LAYOUT,
                    Boolean.TYPE,
                    "NullLayout", // NOI18N
                    "Null Layout") // NOI18N
            {
                public Object getValue() {
                    return new Boolean(useNullLayout);
                }

                public void setValue(Object val) throws IllegalArgumentException {
                    useNullLayout = ((Boolean)val).booleanValue();
                }
            };
        }

        return new Node.PropertySet[] {
            new Node.PropertySet() {
                public Node.Property[] getProperties() {
                    return new Node.Property[] { nullLayoutProperty };
                }
            }
        };
    }

    /** Returns a class of the layout that this DesignLayout represents(e.g.
     * returns FlowLayout.class from DesignFlowLayout).
     * @return a class of the layout represented by this DesignLayout or null if the 
     *         design layout does not represent a "real" layout(e.g. support layouts for JTabbedPane, ...)
     */
    public Class getLayoutClass() {
        return AbsoluteLayout.class;
    }

    // -----------------------------------------------------------------------------
    // Code generation

    /** Generates the code for initialization of this layout, e.g. panel1.setLayout(new BorderLayout());.
     * @param cont   The container that is managed by this layout
     * @return the init code for the layout or null if it should not be generated
     */
    public String generateInitCode(RADVisualContainer cont) {
        return null;
    }

    /** Generates the code for adding specified component to this layout.
     * @param comp   The component to be added to this layout
     * @param cont   The container that is managed by this layout
     */
    public String generateComponentCode(RADVisualContainer cont, RADVisualComponent comp) {
        return null;
    }

    // -----------------------------------------------------------------------------
    // Serialization

    /** Writes the object to the stream.
     * @param oo output stream to write to
     * @exception IOException Includes any I/O exceptions that may occur
     */
    public void writeExternal(java.io.ObjectOutput oo)
    throws java.io.IOException {
        // store version
        oo.writeObject(nbClassVersion);
    }

    /** Reads the object from stream.
     * @param oi input stream to read from
     * @exception IOException Includes any I/O exceptions that may occur
     * @exception ClassNotFoundException if the class of the read object is not found
     */
    public void readExternal(java.io.ObjectInput oi)
    throws java.io.IOException, ClassNotFoundException {
        // check the version
        NbVersion classVersion =(NbVersion) oi.readObject();
        if (!nbClassVersion.isCompatible(classVersion))
            throw new NbVersionNotCompatibleException(classVersion, nbClassVersion);
    }

    // -----------------------------------------------------------------------------
    // constraints innerclass

    /** The ConstraintsDescription class encapsulates constraints data and
     * operations on a constraints that will be used for adding components
     * to the layout.
     */
    final public static class AbsoluteConstraintsDescription extends DesignLayout.ConstraintsDescription {
        /** A JDK 1.1. serial version UID */
        static final long serialVersionUID = 1265127948437909789L;

        /** Netbeans class version */
        public static final NbVersion nbClassVersion = new NbVersion(1, 0);

        /** For externalization only */
        public AbsoluteConstraintsDescription() {
        }

        public AbsoluteConstraintsDescription(Point pos) {
            this(pos, new Dimension(-1, -1));
        }

        public AbsoluteConstraintsDescription(Point pos, Dimension dim) {
            position = pos;
            size = dim;
        }

        public String getConstraintsString() {
            return "[x="+position.x+", y="+position.y+", width="+size.width+", height="+size.height+"]"; // NOI18N
        }

        public Object getConstraintsObject() {
            return new AbsoluteConstraints(position, size);
        }

        // -----------------------------------------------------------------------------
        // Serialization

        /** Writes the object to the stream.
         * @param oo output stream to write to
         * @exception IOException Includes any I/O exceptions that may occur
         */
        public void writeExternal(java.io.ObjectOutput oo)
            throws java.io.IOException {
            // store version
            oo.writeObject(nbClassVersion);

            oo.writeObject(size);
            oo.writeObject(position);
        }

        /** Reads the object from stream.
         * @param oi input stream to read from
         * @exception IOException Includes any I/O exceptions that may occur
         * @exception ClassNotFoundException if the class of the read object is not found
         */
        public void readExternal(java.io.ObjectInput oi)
            throws java.io.IOException, ClassNotFoundException {
            // check the version
            NbVersion classVersion =(NbVersion) oi.readObject();
            if (!nbClassVersion.isCompatible(classVersion))
                throw new NbVersionNotCompatibleException(classVersion, nbClassVersion);

            size =(Dimension) oi.readObject();
            position =(Point) oi.readObject();
        }

        // -----------------------------------------------------------------------------
        // XML Persistence

        /** Called to load property value from specified XML subtree. If succesfully loaded,
         * the value should be available via the getValue method.
         * An IOException should be thrown when the value cannot be restored from the specified XML element
         * @param element the XML DOM element representing a subtree of XML from which the value should be loaded
         * @exception IOException thrown when the value cannot be restored from the specified XML element
         */
        public void readFromXML(org.w3c.dom.Node element) throws java.io.IOException {
            if (!XML_ABSOLUTE_CONSTRAINTS.equals(element.getNodeName())) {
                throw new java.io.IOException();
            }
            org.w3c.dom.NamedNodeMap attributes = element.getAttributes();
            org.w3c.dom.Node xNode = attributes.getNamedItem(ATTR_X);
            org.w3c.dom.Node yNode = attributes.getNamedItem(ATTR_Y);
            org.w3c.dom.Node wNode = attributes.getNamedItem(ATTR_W);
            org.w3c.dom.Node hNode = attributes.getNamedItem(ATTR_H);
            if ((xNode != null) &&(yNode != null)) position = new Point(Integer.parseInt(xNode.getNodeValue()), Integer.parseInt(yNode.getNodeValue()));
            if ((wNode != null) &&(hNode != null)) size = new Dimension(Integer.parseInt(wNode.getNodeValue()), Integer.parseInt(hNode.getNodeValue()));
        }

        /** Called to store current property value into XML subtree. The property value should be set using the
         * setValue method prior to calling this method.
         * @param doc The XML document to store the XML in - should be used for creating nodes only
         * @return the XML DOM element representing a subtree of XML from which the value should be loaded or null 
         *         if the value does not need to save any additional data and can be created using the default constructor
         */
        public org.w3c.dom.Node storeToXML(org.w3c.dom.Document doc) {
            org.w3c.dom.Element el = doc.createElement(XML_ABSOLUTE_CONSTRAINTS);
            el.setAttribute(ATTR_X, Integer.toString(position.x));
            el.setAttribute(ATTR_Y, Integer.toString(position.y));
            el.setAttribute(ATTR_W, Integer.toString(size.width));
            el.setAttribute(ATTR_H, Integer.toString(size.height));
            return el;
        }

        public Point getPosition() {
            return position;
        }

        public Dimension getSize() {
            return size;
        }

        public static final String XML_ABSOLUTE_CONSTRAINTS = "AbsoluteConstraints"; // NOI18N
        public static final String ATTR_X = "x"; // NOI18N
        public static final String ATTR_Y = "y"; // NOI18N
        public static final String ATTR_W = "width"; // NOI18N
        public static final String ATTR_H = "height"; // NOI18N

        Dimension size;
        Point position;
    }
}
