/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.compat2.layouts;

import java.awt.*;
import java.util.Vector;
import java.lang.reflect.InvocationTargetException;

import org.netbeans.modules.form.util2.NbVersion;
import org.netbeans.modules.form.util2.NbVersionNotCompatibleException;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.netbeans.modules.form.*;

/**
 * A design-time support for BorderLayout.
 * @author Ian Formanek
 */
final public class DesignBorderLayout extends DesignLayout {
    /** A JDK 1.1. serial version UID */
    static final long serialVersionUID = -5510213702084891937L;

    /** Netbeans class version */
    public static final NbVersion nbClassVersion = new NbVersion(1, 0);

    /* Layout Properties */
    public static final String PROP_HGAP = "horizontalGap"; // NOI18N
    public static final String PROP_VGAP = "verticalGap"; // NOI18N

    /* Component Layout Properties */
    public static final String PROP_DIRECTION = FormEditor.LAYOUT_PREFIX + "direction"; // NOI18N

    /* Default layout property values */
    private static final int DEFAULT_VGAP = 0;
    private static final int DEFAULT_HGAP = 0;

    /** The width of the design component if no components were added into it */
    private static final int EMPTY_WIDTH = 120;
    /** The height of the design component if no components were added into it */
    private static final int EMPTY_HEIGHT = 70;

    private static final Color MARKED_COLOR = Color.blue;
    private static final Color BORDER_COLOR = Color.darkGray;

    private static final double HORIZONTAL_RATIO = 0.2;
    private static final double VERTICAL_RATIO = 0.2;

    /** bundle to obtain text information from */
    private static final java.util.ResourceBundle bundle = org.openide.util.NbBundle.getBundle(DesignBorderLayout.class);

    /** icons for the Layout. */
    private static final Image icon = Toolkit.getDefaultToolkit().getImage(
        DesignBorderLayout.class.getResource("/org/netbeans/modules/form/resources/palette/borderLayout.gif")); // NOI18N
    private static final Image icon32 = Toolkit.getDefaultToolkit().getImage(
        DesignBorderLayout.class.getResource("/org/netbeans/modules/form/resources/palette/borderLayout32.gif")); // NOI18N

    /** The Sheet holding layout property sheet */
    private Sheet sheet;

    private Node.Property hgapProperty;
    private Node.Property vgapProperty;

    /** An icon of the design-layout. This icon will be used on the ComponentPalette
     * for this layout's item.
     * @param  type the desired type of the icon(BeanInfo.ICON_???)
     * @return layout's icon.
     */
    public Image getIcon(int type) {
        if ((type == java.beans.BeanInfo.ICON_COLOR_16x16) ||(type == java.beans.BeanInfo.ICON_MONO_16x16))
            return icon;
        else
            return icon32;
    }

    public DesignBorderLayout() {
        this(0, 0);
    }

    public DesignBorderLayout(int hgap, int vgap) {
        this.hgap = hgap;
        this.vgap = vgap;
    }

    /** Assigns this DesignLayout to the specified RADVisualContainer.
     * @param cont The RADVisualContainer that represents a container that will be
     *             managed by this layout or null as a notification that this layout
     *             is not a designLayout for its current container anymore
     */
    public void setRADContainer(RADVisualContainer cont) {
        super.setRADContainer(cont);
        if (cont == null)
            return;
        realLayout = new BorderLayout(hgap, vgap);
        cont.getContainerDelegate(cont.getBeanInstance()).setLayout(realLayout);
    }

    /** Method which allows the design layout to provide list of properties to be saved with the form.
     * @return list of Node.Property objects
     */
    public java.util.List getChangedProperties() {
        getPropertySet(); // enforce creation of properties
        java.util.ArrayList list = new java.util.ArrayList(2);
        if (hgap != DEFAULT_HGAP)
            list.add(hgapProperty);
        if (vgap != DEFAULT_VGAP)
            list.add(vgapProperty);
        return list;
    }

    /** Method which is called after the layout is loaded with the form to
     * initialize its properties.
     * @param changedProperties map of <String, Object> pairs, where the String
     * is a name of property and the Object its value
     */
    public void initChangedProperties(java.util.Map changedProperties) {
        Integer hgapVal = (Integer)changedProperties.get(PROP_HGAP);
        if (hgapVal != null)
            hgap = hgapVal.intValue();
        
        Integer vgapVal =(Integer)changedProperties.get(PROP_VGAP);
        if (vgapVal != null)
            vgap = vgapVal.intValue();
    }


    /** Returns a constraint to be used for adding a component to the specified
     * position(e.g. in BorderLayout, the "North", "South", ...  will be
     * determined by the position in the container).  The default
     * implementation just returns the default constraints.  A special
     * constraints object will be returned for layouts that wish to set the
     * position&size of the components rather than add with constraints.
     * @param position The position within the container for which the
     * constraints should be returned.
     */
    public DesignLayout.ConstraintsDescription getConstraintsDescription(Point position) {
        if (getMode() == REAL_MODE) {
            return findFreeDirection();
        }
        else {
            Dimension size = getContainer().getSize();
            Insets insets = getContainer().getInsets();
            size.width -= insets.left;
            size.width -= insets.right;
            size.height -= insets.top;
            size.height -= insets.bottom;
            int horizontalSize =(int)(DesignBorderLayout.HORIZONTAL_RATIO * size.width);
            int verticalSize =(int)(DesignBorderLayout.VERTICAL_RATIO * size.height);
            if (position.y < verticalSize)
                return north;
            else if (position.y > size.height-verticalSize)
                return south;
            else if (position.x < horizontalSize)
                return west;
            else if (position.x > size.width-horizontalSize)
                return east;
            else
                return center;
        }
    }

    /** A display name of the layout will be used for displaying the layout in
     * the components hierarchy during design-time.
     * @return layout's display name.
     */
    public String getDisplayName() {
        return "BorderLayout"; // NOI18N
    }

    /** Returns the global layout's properties(i.e. the properties
     * that are not different for different components in the layout
     * @return the global layout properties
     */
    public Node.PropertySet[] getPropertySet() {
        if (sheet == null) {
            sheet = new Sheet();
            Sheet.Set set = Sheet.createPropertiesSet();
            set.put(hgapProperty = new PropertySupport.ReadWrite(
                PROP_HGAP,
                Integer.TYPE,
                bundle.getString("PROP_border_hgap"),
                bundle.getString("HINT_border_hgap")
                ) {

                public Object getValue() {
                    return new Integer(hgap);
                }

                public void setValue(Object val) throws IllegalArgumentException {
                    if (val instanceof Integer) {
                        int newValue =((Integer)val).intValue();
                        if (hgap == newValue)
                            return;
                        int oldValue = hgap;
                        hgap = newValue;
                        realLayout.setHgap(hgap);
                        firePropertyChange(null, PROP_HGAP, new Integer(oldValue), new Integer(hgap));
                        getContainer().invalidate();
                        getContainer().validate();
                    }
                    else throw new IllegalArgumentException();
                }

                public boolean supportsDefaultValue() {
                    return true;
                }

                public void restoreDefaultValue() {
                    try {
                        setValue(new Integer(DEFAULT_HGAP));
                    } catch (IllegalArgumentException e) { } // ignore failure
                }
            }
                    );

            set.put(vgapProperty = new PropertySupport.ReadWrite(
                PROP_VGAP,
                Integer.TYPE,
                bundle.getString("PROP_border_vgap"),
                bundle.getString("HINT_border_vgap")
                ) {

                public Object getValue() {
                    return new Integer(vgap);
                }

                public void setValue(Object val) throws IllegalArgumentException {
                    if (val instanceof Integer) {
                        int newValue =((Integer)val).intValue();
                        if (vgap == newValue)
                            return;
                        int oldValue = vgap;
                        vgap = newValue;
                        realLayout.setVgap(vgap);
                        firePropertyChange(null, PROP_VGAP, new Integer(oldValue), new Integer(vgap));
                        getContainer().invalidate();
                        getContainer().validate();
                    }
                    else throw new IllegalArgumentException();
                }

                public boolean supportsDefaultValue() {
                    return true;
                }

                public void restoreDefaultValue() {
                    try {
                        setValue(new Integer(DEFAULT_VGAP));
                    } catch (IllegalArgumentException e) { } // ignore failure
                }
            }
                    );

            sheet.put(set);
        }

        return sheet.toArray();
    }

    /** Returns the layout's properties for specified component.
     * @param comp the RADVisualComponent of the component which properties we request
     * @return the layout-specific properties for specified component
     */
    public Node.Property[] getComponentProperties(RADVisualComponent comp) {
        final RADVisualComponent component = comp;
        return new Node.Property[] {
            new PropertySupport.ReadWrite(PROP_DIRECTION, String.class,
                                          bundle.getString("PROP_bordercomp_direction"), bundle.getString("HINT_bordercomp_direction")) {

                public Object getValue() {
                    BorderConstraintsDescription bcd =(BorderConstraintsDescription)component.getConstraints(DesignBorderLayout.class);
                    if (bcd == null) return null;
                    return bcd.getConstraintsString();
                }

                public void setValue(Object val) {
                    if (val instanceof String) {
                        Object oldValue = getValue();
                        BorderConstraintsDescription bcd = null;
                        if ("Center".equals(val)) bcd = center; // NOI18N
                        else if ("North".equals(val)) bcd = north; // NOI18N
                        else if ("South".equals(val)) bcd = south; // NOI18N
                        else if ("West".equals(val)) bcd = west; // NOI18N
                        else if ("East".equals(val)) bcd = east; // NOI18N
                        if (component != null) {
                            component.setConstraints(DesignBorderLayout.class, bcd);
                            updateComponent(component);
                            getContainer().validate();
                            firePropertyChange(component, PROP_DIRECTION, oldValue, val);
                        } else {
                            throw new IllegalArgumentException();
                        }
                    }
                    else throw new IllegalArgumentException();
                }

                /** Editor for directions. */
                public java.beans.PropertyEditor getPropertyEditor() {
                    return new BorderDirectionEditor();
                }

            }
        };
    }

    /** Returns a class of the layout that this DesignLayout represents(e.g.
     * returns FlowLayout.class from DesignFlowLayout).
     * @return a class of the layout represented by this DesignLayout or null
     * if the design layout does not represent a "real" layout(e.g. support
     * layouts for JTabbedPane, ...)
     */
    public Class getLayoutClass() {
        return BorderLayout.class;
    }

    // -----------------------------------------------------------------------------
    // Child components management

    /** @return a first free unused direction, or "Center" if no free is available */ // NOI18N
    private BorderConstraintsDescription findFreeDirection() {
        boolean isNorth = false, isSouth = false, isWest = false, isEast = false, isCenter = false;
        for (java.util.Enumeration e = constraints.elements(); e.hasMoreElements();) {
            String str =(String)e.nextElement();
            if (str.equals("North")) isNorth = true; // NOI18N
            else if (str.equals("South")) isSouth = true; // NOI18N
            else if (str.equals("West")) isWest = true; // NOI18N
            else if (str.equals("East")) isEast = true; // NOI18N
            else if (str.equals("Center")) isCenter = true; // NOI18N
        }
        if (!isNorth) return north;
        else if (!isWest) return west;
        else if (!isCenter) return center;
        else if (!isEast) return east;
        else if (!isSouth) return south;
        else return center;
    }

    /** Adds specified component to this layout. The constraints for the component
     * are acquired from the component by method getConstraints().
     * If the getConstraints() returns null, a new constraints should be
     * synthesized, and set in the component via setConstraints().
     * @param comp The component to add
     * @see RADVisualComponent#getConstraints
     * @see RADVisualComponent#setConstraints
     * @see #removeComponent
     */
    public void addComponent(RADVisualComponent comp) {
        ConstraintsDescription cd = comp.getConstraints(DesignBorderLayout.class);
        if (cd == null) {
            cd = findFreeDirection();
            comp.setConstraints(DesignBorderLayout.class, cd);
        }
        Object direction = cd.getConstraintsObject();
        Component visual = getFormModel().getVisualRepresentation(comp);
//          if (getMode() == DESIGN_MODE)
//              designComponent.add(visual, direction, 0);
//          else
            getContainer().add(visual, direction, 0);
        constraints.put(comp, direction);
    }

    /** Removes specified component from this layout.
     * @param comp The component to remove
     * @see #addComponent
     */
    public void removeComponent(RADVisualComponent comp) {
        Component visual = getFormModel().getVisualRepresentation(comp);
//          if (getMode() == DESIGN_MODE)
//              designComponent.remove(visual);
//          else
            getContainer().remove(visual);
    }

    /** Updates the layout - readds all the visual components to reflect any global or
     * ordering changes.
     */
    public void updateLayout() {
//          if (getMode() == DESIGN_MODE)
//              designComponent.removeAll();
//          else
            getContainer().removeAll();

        RADVisualComponent[] children = getRADContainer().getSubComponents();
        for (int i=0; i < children.length; i++)
            addComponent(children[i]);
//          if (getMode() == DESIGN_MODE) {
//              designComponent .validate();
//              designComponent .repaint();
//        } else {
            getContainer().validate();
            getContainer().repaint();
//      }
    }

    /** Should be called when specified component's layout properties change.
     * @param comp The component to update
     */
    void updateComponent(RADVisualComponent comp) {
        Component visual = getFormModel().getVisualRepresentation(comp);
        ConstraintsDescription cd = comp.getConstraints(DesignBorderLayout.class);
        if (cd == null) {
            cd = findFreeDirection();
            comp.setConstraints(DesignBorderLayout.class, cd);
        }

        Object direction = cd.getConstraintsObject();
//          if (getMode() == DESIGN_MODE) {
//              designComponent.remove(visual);
//              designComponent.add(visual, direction, 0);
//          } else {
            getContainer().remove(visual);
            getContainer().add(visual, direction, 0);
//      }

        constraints.put(comp, direction);
    }

    // -----------------------------------------------------------------------------
    // Code generation

    /** Generates the code for initialization of this layout,
     * e.g. panel1.setLayout(new BorderLayout());.
     * @param cont The container that is managed by this layout
     * @return the init code for the layout or null if it should not be
     * generated
     */
    public String generateInitCode(RADVisualContainer cont) {
        LayoutManager defaultLM = null;
        try {
            Container defaultCont =(Container)BeanSupport.getDefaultInstance(cont.getBeanClass());
            if (defaultCont != null) defaultLM = defaultCont.getLayout();
        } catch (Exception e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
            // ok, no default
        }
        if (defaultLM != null) {
            if (defaultLM.getClass().equals(BorderLayout.class)) {
                if ((((BorderLayout)defaultLM).getHgap() == hgap) &&(((BorderLayout)defaultLM).getVgap() == vgap)) {
                    return null; // the default layout is the same as current settings => no need to generate layout
                }
            }
        }

        StringBuffer buf = new StringBuffer();
        buf.append(createContainerGenName(cont));
        if ((hgap != DEFAULT_HGAP) ||(vgap != DEFAULT_VGAP)) {
            buf.append("setLayout(new java.awt.BorderLayout("); // NOI18N
            buf.append(hgap);
            buf.append(", "); // NOI18N
            buf.append(vgap);
            buf.append("));\n"); // NOI18N
        }
        else {
            buf.append("setLayout(new java.awt.BorderLayout());\n"); // NOI18N
        }
        return buf.toString();
    }

    /** Generates the code for adding specified component to this layout.
     * @param comp The component to be added to this layout
     * @param cont The container that is managed by this layout
     */
    public String generateComponentCode(RADVisualContainer cont, RADVisualComponent comp) {
        ConstraintsDescription cd = comp.getConstraints(DesignBorderLayout.class);
        String constr;
        if ((cd == null) ||((constr =(String)cd.getConstraintsObject()) == null))
            return "// "+FormEditor.getFormBundle().getString("MSG_ErrorGeneratingCode");

        StringBuffer buf = new StringBuffer();
        buf.append(createContainerGenName(cont));
        buf.append("add("); // NOI18N
        buf.append(comp.getName());
        buf.append(", java.awt.BorderLayout."); // NOI18N
        buf.append(constr.toUpperCase());
        buf.append(");\n"); // NOI18N

        return buf.toString();
    }

    // -----------------------------------------------------------------------------
    // Serialization

    /** Writes the object to the stream.
     * @param oo output stream to write to
     * @exception IOException Includes any I/O exceptions that may occur
     */
    public void writeExternal(java.io.ObjectOutput oo)
        throws java.io.IOException {
        // store version
        oo.writeObject(nbClassVersion);

        oo.writeInt(vgap);
        oo.writeInt(hgap);
    }

    /** Reads the object from stream.
     * @param oi input stream to read from
     * @exception IOException Includes any I/O exceptions that may occur
     * @exception ClassNotFoundException if the class of the read object is not
     * found
     */
    public void readExternal(java.io.ObjectInput oi)
        throws java.io.IOException, ClassNotFoundException {
        org.netbeans.modules.form.FormUtils.DEBUG(">> DesignBorderLayout: readExternal: START"); // NOI18N
        // check the version
        NbVersion classVersion =(NbVersion) oi.readObject();
        if (!nbClassVersion.isCompatible(classVersion))
            throw new NbVersionNotCompatibleException(classVersion, nbClassVersion);

        vgap = oi.readInt();
        hgap = oi.readInt();
        org.netbeans.modules.form.FormUtils.DEBUG("<< DesignBorderLayout: readExternal: END"); // NOI18N
    }

    // -----------------------------------------------------------------------------
    // constraints innerclass

    /** The ConstraintsDescription class encapsulates constraints data and
     * operations on a constraints that will be used for adding components
     * to the layout.
     */
    final public static class BorderConstraintsDescription
        extends DesignLayout.ConstraintsDescription
    {
        /** A JDK 1.1. serial version UID */
        static final long serialVersionUID = -8892681661752848070L;

        /** Netbeans class version */
        public static final NbVersion nbClassVersion = new NbVersion(1, 0);

        /** For Externalization only */
        public BorderConstraintsDescription() {
        }

        public BorderConstraintsDescription(String dir) {
            direction = dir;
        }

        public String getConstraintsString() {
            return direction;
        }

        public Object getConstraintsObject() {
            return direction;
        }

        // -----------------------------------------------------------------------------
        // Serialization

        /** Writes the object to the stream.
         * @param oo output stream to write to
         * @exception IOException Includes any I/O exceptions that may occur
         */
        public void writeExternal(java.io.ObjectOutput oo)
            throws java.io.IOException {
            // store version
            oo.writeObject(nbClassVersion);

            oo.writeObject(direction);
        }

        /** Reads the object from stream.
         * @param oi input stream to read from
         * @exception IOException Includes any I/O exceptions that may occur
         * @exception ClassNotFoundException if the class of the read object is not found
         */
        public void readExternal(java.io.ObjectInput oi)
            throws java.io.IOException, ClassNotFoundException {
            org.netbeans.modules.form.FormUtils.DEBUG(">> BorderConstraintsDescription: readExternal: START"); // NOI18N
            // check the version
            NbVersion classVersion =(NbVersion) oi.readObject();
            if (!nbClassVersion.isCompatible(classVersion))
                throw new NbVersionNotCompatibleException(classVersion, nbClassVersion);

            direction =(String) oi.readObject();
            org.netbeans.modules.form.FormUtils.DEBUG("<< BorderConstraintsDescription: readExternal: END"); // NOI18N
        }

        // -----------------------------------------------------------------------------
        // XML Persistence

        /** Called to load property value from specified XML subtree. If
         * succesfully loaded, the value should be available via the getValue
         * method.  An IOException should be thrown when the value cannot be
         * restored from the specified XML element
         * @param element the XML DOM element representing a subtree of XML
         * from which the value should be loaded
         * @exception IOException thrown when the value cannot be restored from
         * the specified XML element
         */
        public void readFromXML(org.w3c.dom.Node element) throws java.io.IOException {
            if (!XML_BORDER_CONSTRAINTS.equals(element.getNodeName())) {
                throw new java.io.IOException();
            }
            org.w3c.dom.NamedNodeMap attributes = element.getAttributes();
            org.w3c.dom.Node dirNode = attributes.getNamedItem(ATTR_DIRECTION);
            if (dirNode != null) direction = dirNode.getNodeValue();
        }

        /** Called to store current property value into XML subtree. The
         * property value should be set using the setValue method prior to
         * calling this method.
         * @param doc The XML document to store the XML in - should be used for
         * creating nodes only
         * @return the XML DOM element representing a subtree of XML from which
         * the value should be loaded or null if the value does not need to
         * save any additional data and can be created using the default
         * constructor
         */
        public org.w3c.dom.Node storeToXML(org.w3c.dom.Document doc) {
            org.w3c.dom.Element el = doc.createElement(XML_BORDER_CONSTRAINTS);
            el.setAttribute(ATTR_DIRECTION, direction);
            return el;
        }

        public static final String XML_BORDER_CONSTRAINTS = "BorderConstraints"; // NOI18N
        public static final String ATTR_DIRECTION = "direction"; // NOI18N

        /** The direction represented by this BorderConstraintsDescription class */
        private String direction;
    }

    // -----------------------------------------------------------------------------
    // private area

    /** The real BorderLayout LayoutManager that works in real-layout mode*/
    transient private BorderLayout realLayout;

    // predefined constraints
    
    private BorderConstraintsDescription north =
        new BorderConstraintsDescription("North"); // NOI18N
    private BorderConstraintsDescription south =
        new BorderConstraintsDescription("South"); // NOI18N
    private BorderConstraintsDescription west =
        new BorderConstraintsDescription("West"); // NOI18N
    private BorderConstraintsDescription east =
        new BorderConstraintsDescription("East"); // NOI18N
    private BorderConstraintsDescription center =
        new BorderConstraintsDescription("Center"); // NOI18N

    /** Mapping <RADVisualComponent, String constraints> */
    private java.util.Hashtable constraints = new java.util.Hashtable();

    /** The vertical gap property */
    private int vgap = DEFAULT_VGAP;
    /** The horizontal gap property */
    private int hgap = DEFAULT_HGAP;
}
