/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.forminfo;

import java.awt.*;
import javax.swing.*;
import org.netbeans.modules.form.RADMenuItemComponent;

/** FormInfo for java.awt.Frame.
 *
 * @author Ian Formanek
 */
public class FrameFormInfo extends FormInfo implements MenuBarContainer
{
    /** The form instance */
    private Frame formInstance;

    /** The form top-level container */
    private JPanel topPanel;

    /** The form top-level container for adding */
    private JPanel innerPanel;

    private JMenuBar currentMenu = null;
    private boolean initialized = false;

    //
    // FormInfo methods
    //

    /** Used to create the design-time instance of the form object, which is used
     * only for displaing properties and events of the form.  I.e. it is not
     * displayed visually, instead the FormTopComponent is used with the
     * container provided from <code>getTopContainer()</code> method.
     * @return the instance of the form
     * @see #getTopContainer
     */
    public Object getFormInstance() {
        if (formInstance == null) {
            formInstance = new Frame();

            // we must set these explicitly, as the window is never displayed and the
            // fornt and colors would stay null
            formInstance.setBackground(SystemColor.window);
            formInstance.setForeground(SystemColor.windowText);
            formInstance.setFont(new Font("Dialog", Font.PLAIN, 12)); // NOI18N
        }
        return formInstance;
    }

    /** Used to provide the container which is used during design-time as the
     * top-level container.  The container provided by this class should not be a
     * Window, as it is added as a component to the FormTopComponent, rather a
     * JPanel, Panel or JDesktopPane should be used according to the form type.
     * By returning a <code>null</code> value, the form info declares that it
     * does not represent a "visual" form and the visual ediuting should not be
     * used with it.
     * @return the top level container which will be used during design-time or
     * null if the form is not visual
     */
    public Container getTopContainer() {
        if (!initialized) {
            initialize();
        }
        return topPanel;
    }

    /** Used to provide the container which is used during design-time as the
     * top-level container for adding components.  The container provided by this
     * class should not be a Window, as it is added as a component to the
     * FormTopComponent, rather a JPanel, Panel or JDesktopPane should be used
     * according to the form type.  By returning a <code>null</code> value, the
     * form info declares that it does not represent a "visual" form and the
     * visual ediuting should not be used with it.
     * @return the top level container which will be used during design-time or
     * null if the form is not visual
     */
    public Container getTopAddContainer() {
        if (!initialized) {
            initialize();
        }
        return innerPanel;
    }

    private void initialize() {
        topPanel = new JPanel();
        topPanel.setLayout(new java.awt.BorderLayout());
        topPanel.add(innerPanel = new JPanel());
        innerPanel.setLayout(new java.awt.BorderLayout());
        
        initialized = true;
    }

    //
    // JMenuBarContainer implementation
    //

    /** Can be used to set the MenuBar of this form.
     * @param menuBar The MenuBar component or null to set no menu.
     */

    public void setMenuBar(MenuBar menuBar) {
        getFormInstance(); // enforce creation if not initialized yet
        formInstance.setMenuBar(menuBar);
        JMenuBar newMenuBar = null;

        if (menuBar != null)
            newMenuBar =(JMenuBar) RADMenuItemComponent.findDesignTimeMenu(menuBar);

        if (currentMenu == newMenuBar)
            return;

        if (currentMenu != null)
            topPanel.remove(currentMenu);

        currentMenu = newMenuBar;

        if (currentMenu != null)
            topPanel.add(currentMenu, java.awt.BorderLayout.NORTH);

        topPanel.validate();
        topPanel.repaint();
    }

    /** Can be used to obtain the MenuBar of this form.
     * @return The MenuBar component or null if no menu is set.
     */
    public MenuBar getMenuBar() {
        getFormInstance(); // enforce creation if not initialized yet
        return formInstance.getMenuBar();
    }

}
