/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.layoutsupport;

import java.awt.*;
import java.beans.*;
import java.util.*;
import javax.swing.*;

import org.openide.nodes.Node;
import org.netbeans.modules.form.*;

/**
 *
 * @author Tran Duc Trung, Tomas Pavek
 */

/** Support class for BoxLayout. As BoxLayout is not a bean, it must be
 * handled differently from other layout managers...
 */
public class BoxLayoutSupport extends AbstractLayoutSupport
{
    private int axis = BoxLayout.X_AXIS;

    private Node.PropertySet propertySet;
    private FormProperty[] properties;

    public Class getLayoutClass() {
        return BoxLayout.class;
    }

    public int getNewIndex(Container container, Point posInCont,
                           Component comp, Point posInComp) {
        if (!(container.getLayout() instanceof BoxLayout))
            return -1;
        
        Component[] components = container.getComponents();
        for (int i = 0; i < components.length; i++) {
            Rectangle b = components[i].getBounds();
            if (axis == BoxLayout.X_AXIS) {
                if (posInCont.x < b.x + b.width / 2)
                    return i;
            }
            else {
                if (posInCont.y < b.y + b.height / 2)
                    return i;
            }
        }
        
        return components.length;
    }

    public boolean paintDragFeedback(Container container,
                                     Component component,
                                     ConstraintsDesc newConstraints,
                                     int newIndex,
                                     Graphics g) {
        if (!(container.getLayout() instanceof BoxLayout))
            return false;
        
        Component[] components = container.getComponents();
        Rectangle rect;

//        if (newIndex <= 0) {
//            rect = new Rectangle(0, 0, 20, 30);
//        }
//        else
        if (newIndex < 0 || newIndex >= components.length) {
            newIndex = components.length - 1;
            
            Rectangle b = components[newIndex].getBounds();
            if (axis == BoxLayout.X_AXIS) {
                rect = new Rectangle(b.x + b.width - 10, b.y, 20, b.height);
            }
            else {
                rect = new Rectangle(b.x, b.y + b.height - 10, b.width, 20);
            }
        }
        else {
            Rectangle b = components[newIndex].getBounds();
            if (axis == BoxLayout.X_AXIS) {
                rect = new Rectangle(b.x - 10, b.y, 20, b.height);
            }
            else {
                rect = new Rectangle(b.x, b.y - 10, b.width, 20);
            }
        }
        
//          Dimension sz = container.getSize();
//          Insets insets = container.getInsets();
//          sz.width -= insets.left + insets.right;
//          sz.height -= insets.top + insets.bottom;
//          if (axis == BoxLayout.X_AXIS) {
//              rect.y = insets.top + (sz.height - rect.height) / 2;
//          }
//          else {
//              rect.x = insets.left + (sz.width - rect.width) / 2;
//          }
            
        g.drawRect(rect.x, rect.y, rect.width, rect.height);
        return true;
    }
    
    public LayoutManager createDefaultLayoutInstance(Container cont) {
        return new BoxLayout(cont, BoxLayout.X_AXIS);
    }

    public LayoutManager cloneLayoutInstance(Container cont) {
        return new BoxLayout(cont, axis);
    }

    public Node.PropertySet[] getPropertySets() {
        Node.PropertySet[] propSets = super.getPropertySets();

        int i = 0;
        for (int ii=0; ii < propSets.length; ii++)
            if ("properties".equals(propSets[ii].getName())) { // NOI18N
                i = ii;
                break;
            }

        if (propertySet == null) {
            // BoxLayout has no real "bean" properties,
            // we must create a special property set
            propertySet = new Node.PropertySet(
                        propSets[i].getName(),
                        propSets[i].getDisplayName(),
                        propSets[i].getShortDescription()) {

                public Node.Property[] getProperties() {
                    return BoxLayoutSupport.this.getProperties();
                }
            };
        }

        propSets[i] = propertySet;
        return propSets;
    }

    public String getJavaSetLayoutString(/*RADVisualContainer cont*/) {
        StringBuffer buf = new StringBuffer();
        String delegate = getContainer().getJavaContainerDelegateString(); // cont
        if (!"".equals(delegate)) { // NOI18N
            buf.append(delegate);
            buf.append("."); // NOI18N
        }
        buf.append("setLayout(new javax.swing.BoxLayout("); // NOI18N
        buf.append(!"".equals(delegate) ? delegate : "this");
        buf.append(", ");
        buf.append(getProperties()[0].getJavaInitializationString());
        buf.append("));\n"); // NOI18N
        return buf.toString();
    }

    // ------------

    private void updateLayoutInstance() {
        RADVisualContainer container = getContainer();
        Container cont = container.getContainerDelegate(container.getBeanInstance());
        getMetaLayout().updateInstance(new BoxLayout(cont, axis));
    }

    private FormProperty[] getProperties() {
        if (properties == null) {
            // we cannot use RADProperty because "axis" is not a real
            // bean property - we must create a special FormProperty
            properties = new FormProperty[1];

            properties[0] = new FormProperty("axis", // NOI18N
                                             Integer.TYPE,
                                getBundle().getString("PROP_axis"), // NOI18N
                                getBundle().getString("HINT_axis")) { // NOI18N

                public Object getTargetValue() {
                    return new Integer(axis);
                }

                public void setTargetValue(Object value) {
                    axis = ((Integer)value).intValue();
                    updateLayoutInstance();
                }

                public boolean supportsDefaultValue() {
                    return true;
                }

                public Object getDefaultValue() {
                    return new Integer(BoxLayout.X_AXIS);
                }

                public PropertyEditor getExpliciteEditor() {
                    return new BoxAxisEditor();
                }

                // property changes should be propagated to FormModel
                // and LayoutNode (this is normally done by RADProperty)
                protected void firePropertyValueChange(Object old, 
                                                       Object current) {
                    super.firePropertyValueChange(old, current);

                    if (isChangeFiring()) {
                        getContainer().getFormModel()
                            .fireComponentPropertyChanged(getMetaLayout(),
                                              getName(), old, current);

                        LayoutNode node = getContainer().getLayoutNodeReference();
                        if (node != null) node.fireLayoutPropertiesChange();
                    }
                }
            };

            properties[0].setPropertyContext(
                new FormPropertyContext.DefaultImpl(getContainer().getFormModel()));
        }

        return properties;
    }

    public FormProperty getProperty(String propName) {
        return "axis".equals(propName) ? getProperties()[0] : null; // NOI18N
    }

    // --------------

    public static final class BoxAxisEditor extends PropertyEditorSupport {
        private final String[] tags = {
            getBundle().getString("VALUE_axis_x"), // NOI18N
            getBundle().getString("VALUE_axis_y")  // NOI18N
        };
        private final Integer[] values = {
            new Integer(BoxLayout.X_AXIS),
            new Integer(BoxLayout.Y_AXIS)
        };
        private final String[] javaInitStrings = {
            "javax.swing.BoxLayout.X_AXIS", // NOI18N
            "javax.swing.BoxLayout.Y_AXIS"  // NOI18N
        };

        public String[] getTags() {
            return tags;
        }

        public String getAsText() {
            Object value = getValue();
            if (values[0].equals(value)) return tags[0];
            if (values[1].equals(value)) return tags[1];
            return null;
        }

        public void setAsText(String str) {
            if (tags[0].equals(str)) setValue(values[0]);
            else if (tags[1].equals(str)) setValue(values[1]);
        }

        public String getJavaInitializationString() {
            Object value = getValue();
            for (int i=0; i < values.length; i++)
                if (values[i].equals(value))
                    return javaInitStrings[i];
            return null;
        }
    }
}
