/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.layoutsupport;

import java.awt.*;
import java.beans.*;
import java.util.*;
import org.openide.nodes.Node;
import org.openide.explorer.propertysheet.editors.EnhancedPropertyEditor;
import org.netbeans.modules.form.*;

/**
 *
 * @author Tran Duc Trung, Tomas Pavek
 */

public class GridBagLayoutSupport extends AbstractLayoutSupport
{
    private String constraintsVariableName;


    public Class getLayoutClass() {
        return GridBagLayout.class;
    }

    public Class getCustomizerClass() {
        return GridBagCustomizer.class;
    }

    public void addComponent(RADVisualComponent component, ConstraintsDesc desc) {
        if (desc == null)
            desc = new GridBagConstraintsDesc();
        super.addComponent(component, desc);
    }

    public ConstraintsDesc[] convertConstraints(ConstraintsDesc[] constraints,
                                                Component[] components) {
        if (components == null || components.length == 0
                || constraints == null
                || constraints.length != components.length
                || !(constraints[0] instanceof
                    AbsoluteLayoutSupport.AbsoluteConstraintsDesc))
            return null;

        GridBagConstraintsDesc[] newConstraints =
            new GridBagConstraintsDesc[constraints.length];

        for (int i=0; i < constraints.length; i++) {
            Rectangle iBounds = components[i].getBounds();
            Dimension prefSize = components[i].getPreferredSize();
            GridBagConstraints constr = new GridBagConstraints();

            int gx = 0, gy = 0, gw = 1, gh = 1;
            int ix1 = iBounds.x;
            int iy1 = iBounds.y;
            int ix2 = ix1 + iBounds.width;
            int iy2 = iy1 + iBounds.height;
            int fromX = 0, fromY = 0;

            for (int j=0; j < constraints.length; j++) {
                Rectangle jBounds = components[j].getBounds();
                int jx2 = jBounds.x + jBounds.width;
                int jy2 = jBounds.y + jBounds.height;
                if (jx2 <= ix1) {
                    gx++;
                    fromX = Math.max(fromX, jx2);
                }
                if (jy2 <= iy1) {
                    gy++;
                    fromY = Math.max(fromY, jy2);
                }
                if (jx2 > ix1 && jx2 < ix2) gw++;
                if (jy2 > iy1 && jy2 < iy2) gh++;
            }

            constr.gridx = gx;
            constr.gridy = gy;
            constr.gridwidth = gw;
            constr.gridheight = gh;
            constr.insets = new Insets(iy1 - fromY, ix1 - fromX, 0, 0);
            constr.fill = GridBagConstraints.BOTH;
            constr.ipadx = iBounds.width - prefSize.width;
            constr.ipady = iBounds.height - prefSize.height;

            newConstraints[i] = new GridBagConstraintsDesc(constr);
        }

        return newConstraints;
    }

    public String getJavaSetLayoutString(/*RADVisualContainer cont*/) {
        StringBuffer buf = new StringBuffer(super.getJavaSetLayoutString());
        String constrVariable = getConstraintsVariableName();

        buf.append("java.awt.GridBagConstraints "); // NOI18N
        buf.append(constrVariable);
        buf.append(";\n"); // NOI18N
        return buf.toString();
    }

    public String getJavaAddComponentString(/*RADVisualContainer cont,*/
                                            RADVisualComponent comp) {
        StringBuffer buf = new StringBuffer();
        String constrVariable = getConstraintsVariableName();

        buf.append(constrVariable);
        buf.append(" = new java.awt.GridBagConstraints();\n"); // NOI18N

        ConstraintsDesc desc = getConstraints(comp);
        if (desc instanceof GridBagConstraintsDesc) {
            ((GridBagConstraintsDesc)desc).constrVariable = constrVariable;
            buf.append(desc.getJavaInitializationString());
        }

        String delegate = getContainer().getJavaContainerDelegateString(); // cont
        if (!"".equals(delegate)) { // NOI18N
            buf.append(delegate);
            buf.append("."); // NOI18N
        }
        buf.append("add("); // NOI18N
        buf.append(comp.getName());
        buf.append(", "); // NOI18N
        buf.append(constrVariable);
        buf.append(");\n"); // NOI18N

        return buf.toString();
    }

    // --------------------

    private String getConstraintsVariableName() {
        if (constraintsVariableName == null && getContainer() != null) {
            VariablePool pool = getContainer().getFormModel().getVariablePool();
            constraintsVariableName = pool.getNewName(GridBagConstraints.class);
            pool.reserveNameSurviveReparse(constraintsVariableName);
        }

        return constraintsVariableName;
    }

    // -----------------

    public static class GridBagConstraintsDesc
            implements LayoutSupport.ConstraintsDesc 
    {
        private GridBagConstraints constraints;

        private GridBagConstraints defaultConstraints = new GridBagConstraints();
        transient private String constrVariable;

        transient private Node.Property[] properties;


        public GridBagConstraintsDesc() {
            constraints = new GridBagConstraints();
        }

        public GridBagConstraintsDesc(GridBagConstraints constraints) {
            this.constraints = constraints;
        }

        public Node.Property[] getProperties() {
            if (properties == null) {
                createProperties();
                reinstateProperties();
            }
            return properties;
        }

        public Object getConstraintsObject() {
            return constraints;
        }

        public String getJavaInitializationString() {
            StringBuffer buf = new StringBuffer();

            Node.Property[] properties = getProperties();
            for (int i=0; i < properties.length; i++) {
                FormProperty prop = (FormProperty)properties[i];
                if (prop.isChanged()) {
                    String javaStr = prop.getJavaInitializationString();
                    // test it just for sure - it should be always true
                    if (javaStr != null && !"".equals(javaStr)) { // NOI18N
                        buf.append(constrVariable);
                        buf.append("."); // NOI18N
                        buf.append(prop.getName());
                        buf.append(" = "); // NOI18N
                        buf.append(javaStr);
                        buf.append(";\n"); // NOI18N
                    }
                }
            }
            return buf.toString();
        }

        public void readFromXML(org.w3c.dom.Node element)
        throws java.io.IOException {
        }

        public org.w3c.dom.Node storeToXML(org.w3c.dom.Document doc) {
            return null;
        }

        private void createProperties() {
            properties = new Node.Property[] {
                new FormProperty("gridx", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_gridx"), // NOI18N
                                 getBundle().getString("HINT_gridx")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.gridx);
                    }
                    public void setTargetValue(Object value) {
                        constraints.gridx = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.gridx);
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new GridPosEditor();
                    }
                },

                new FormProperty("gridy", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_gridy"), // NOI18N
                                 getBundle().getString("HINT_gridy")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.gridy);
                    }
                    public void setTargetValue(Object value) {
                        constraints.gridy = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.gridy);
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new GridPosEditor();
                    }
                },

                new FormProperty("gridwidth", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_gridwidth"), // NOI18N
                                 getBundle().getString("HINT_gridwidth")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.gridwidth);
                    }
                    public void setTargetValue(Object value) {
                        constraints.gridwidth = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.gridwidth);
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new GridSizeEditor();
                    }
                },

                new FormProperty("gridheight", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_gridheight"), // NOI18N
                                 getBundle().getString("HINT_gridheight")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.gridheight);
                    }
                    public void setTargetValue(Object value) {
                        constraints.gridheight = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.gridheight);
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new GridSizeEditor();
                    }
                },

                new FormProperty("fill", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_fill"), // NOI18N
                                 getBundle().getString("HINT_fill")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.fill);
                    }
                    public void setTargetValue(Object value) {
                        constraints.fill = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.fill);
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new FillEditor();
                    }
                },

                new FormProperty("ipadx", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_ipadx"), // NOI18N
                                 getBundle().getString("HINT_ipadx")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.ipadx);
                    }
                    public void setTargetValue(Object value) {
                        constraints.ipadx = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.ipadx);
                    }
                },

                new FormProperty("ipady", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_ipady"), // NOI18N
                                 getBundle().getString("HINT_ipady")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.ipady);
                    }
                    public void setTargetValue(Object value) {
                        constraints.ipady = ((Integer)value).intValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.ipady);
                    }
                },

                new FormProperty("insets", // NOI18N
                                 Insets.class,
                                 getBundle().getString("PROP_insets"), // NOI18N
                                 getBundle().getString("HINT_insets")) { // NOI18N

                    public Object getTargetValue() {
                        return constraints.insets;
                    }
                    public void setTargetValue(Object value) {
                        constraints.insets = (Insets)value;
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return defaultConstraints.insets;
                    }
                },

                new FormProperty("anchor", // NOI18N
                                 Integer.TYPE,
                                 getBundle().getString("PROP_anchor"), // NOI18N
                                 getBundle().getString("HINT_anchor")) { // NOI18N

                    public Object getTargetValue() {
                        return new Integer(constraints.anchor);
                    }
                    public void setTargetValue(Object value) {
                        constraints.anchor = ((Integer)value).intValue();
                    }
                    public PropertyEditor getExpliciteEditor() {
                        return new AnchorEditor();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Integer(defaultConstraints.anchor);
                    }
                },

                new FormProperty("weightx", // NOI18N
                                 Double.TYPE,
                                 getBundle().getString("PROP_weightx"), // NOI18N
                                 getBundle().getString("HINT_weightx")) { // NOI18N

                    public Object getTargetValue() {
                        return new Double(constraints.weightx);
                    }
                    public void setTargetValue(Object value) {
                        constraints.weightx = ((Double)value).doubleValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Double(defaultConstraints.weightx);
                    }
                },

                new FormProperty("weighty", // NOI18N
                                 Double.TYPE,
                                 getBundle().getString("PROP_weighty"), // NOI18N
                                 getBundle().getString("HINT_weighty")) { // NOI18N

                    public Object getTargetValue() {
                        return new Double(constraints.weighty);
                    }
                    public void setTargetValue(Object value) {
                        constraints.weighty = ((Double)value).doubleValue();
                    }
                    public boolean supportsDefaultValue () {
                        return true;
                    }
                    public Object getDefaultValue() {
                        return new Double(defaultConstraints.weighty);
                    }
                },
            };
        }

        private void reinstateProperties() {
            try {
                for (int i=0; i < properties.length; i++) {
                    FormProperty prop = (FormProperty) properties[i];
                    prop.reinstateProperty();
                }
            }
            catch(IllegalAccessException e1) {} // should not happen
            catch(java.lang.reflect.InvocationTargetException e2) {} // should not happen
        }
    }

    // ------------
    // property editors for properties of GridBagConstraintsDesc

    private abstract static class GridBagConstrEditor extends PropertyEditorSupport {
        String[] tags;
        Integer[] values;
        String[] javaInitStrings;
        boolean otherValuesAllowed;

        public String[] getTags() {
            return tags;
        }

        public String getAsText() {
            Object value = getValue();
            for (int i=0; i < values.length; i++)
                if (values[i].equals(value))
                    return tags[i];

            return otherValuesAllowed && value != null ?
                       value.toString() : null;
        }

        public void setAsText(String str) {
            for (int i=0; i < tags.length; i++)
                if (tags[i].equals(str)) {
                    setValue(values[i]);
                    return;
                }

            if (otherValuesAllowed)
                try {
                    setValue(new Integer(Integer.parseInt(str)));
                } 
                catch (NumberFormatException e) {} // ignore
        }

        public String getJavaInitializationString() {
            Object value = getValue();
            for (int i=0; i < values.length; i++)
                if (values[i].equals(value))
                    return javaInitStrings[i];

            if (!otherValuesAllowed)
                return javaInitStrings[0];
            return value != null ? value.toString() : null;
        }
    }

    public static final class GridPosEditor extends GridBagConstrEditor
                                        implements EnhancedPropertyEditor {
        public GridPosEditor() {
            tags = new String[] {
                getBundle().getString("VALUE_relative") // NOI18N
            };
            values = new Integer[] {
                new Integer(GridBagConstraints.RELATIVE)
            };
            javaInitStrings = new String[] {
                "java.awt.GridBagConstraints.RELATIVE" // NOI18N
            };
            otherValuesAllowed = true;
        }

        public Component getInPlaceCustomEditor() {
            return null;
        }
        public boolean hasInPlaceCustomEditor() {
            return false;
        }
        public boolean supportsEditingTaggedValues() {
            return true;
        }
    }

    public static final class GridSizeEditor extends GridBagConstrEditor
                                        implements EnhancedPropertyEditor {
        public GridSizeEditor() {
            tags = new String[] {
                getBundle().getString("VALUE_relative"), // NOI18N
                getBundle().getString("VALUE_remainder") // NOI18N
            };
            values = new Integer[] {
                new Integer(GridBagConstraints.RELATIVE),
                new Integer(GridBagConstraints.REMAINDER)
            };
            javaInitStrings = new String[] {
                "java.awt.GridBagConstraints.RELATIVE", // NOI18N
                "java.awt.GridBagConstraints.REMAINDER" // NOI18N
            };
            otherValuesAllowed = true;
        }

        public Component getInPlaceCustomEditor() {
            return null;
        }
        public boolean hasInPlaceCustomEditor() {
            return false;
        }
        public boolean supportsEditingTaggedValues() {
            return true;
        }
    }

    public static final class FillEditor extends GridBagConstrEditor {
        public FillEditor() {
            tags = new String[] {
                getBundle().getString("VALUE_fill_none"), // NOI18N
                getBundle().getString("VALUE_fill_horizontal"), // NOI18N
                getBundle().getString("VALUE_fill_vertical"), // NOI18N
                getBundle().getString("VALUE_fill_both") // NOI18N
            };
            values = new Integer[] {
                new Integer(GridBagConstraints.NONE),
                new Integer(GridBagConstraints.HORIZONTAL),
                new Integer(GridBagConstraints.VERTICAL),
                new Integer(GridBagConstraints.BOTH)
            };
            javaInitStrings = new String[] {
                "java.awt.GridBagConstraints.NONE", // NOI18N
                "java.awt.GridBagConstraints.HORIZONTAL", // NOI18N
                "java.awt.GridBagConstraints.VERTICAL", // NOI18N
                "java.awt.GridBagConstraints.BOTH" // NOI18N
            };
            otherValuesAllowed = false;
        }
    }

    public static final class AnchorEditor extends GridBagConstrEditor {
        public AnchorEditor() {
            tags = new String[] {
                getBundle().getString("VALUE_anchor_center"), // NOI18N
                getBundle().getString("VALUE_anchor_north"), // NOI18N
                getBundle().getString("VALUE_anchor_northeast"), // NOI18N
                getBundle().getString("VALUE_anchor_east"), // NOI18N
                getBundle().getString("VALUE_anchor_southeast"), // NOI18N
                getBundle().getString("VALUE_anchor_south"), // NOI18N
                getBundle().getString("VALUE_anchor_southwest"), // NOI18N
                getBundle().getString("VALUE_anchor_west"), // NOI18N
                getBundle().getString("VALUE_anchor_northwest"), // NOI18N
            };
            values = new Integer[] {
                new Integer(GridBagConstraints.CENTER),
                new Integer(GridBagConstraints.NORTH),
                new Integer(GridBagConstraints.NORTHEAST),
                new Integer(GridBagConstraints.EAST),
                new Integer(GridBagConstraints.SOUTHEAST),
                new Integer(GridBagConstraints.SOUTH),
                new Integer(GridBagConstraints.SOUTHWEST),
                new Integer(GridBagConstraints.WEST),
                new Integer(GridBagConstraints.NORTHWEST)
            };
            javaInitStrings = new String[] {
                "java.awt.GridBagConstraints.CENTER", // NOI18N
                "java.awt.GridBagConstraints.NORTH", // NOI18N
                "java.awt.GridBagConstraints.NORTHEAST", // NOI18N
                "java.awt.GridBagConstraints.EAST", // NOI18N
                "java.awt.GridBagConstraints.SOUTHEAST", // NOI18N
                "java.awt.GridBagConstraints.SOUTH", // NOI18N
                "java.awt.GridBagConstraints.SOUTHWEST", // NOI18N
                "java.awt.GridBagConstraints.WEST", // NOI18N
                "java.awt.GridBagConstraints.NORTHWEST" // NOI18N
            };
            otherValuesAllowed = false;
        }
    }
}
