/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.form.layoutsupport.dedicated;


import java.awt.*;
import java.beans.*;
import java.util.*;
import javax.swing.*;

import org.openide.nodes.Node;
import org.netbeans.modules.form.*;
import org.netbeans.modules.form.layoutsupport.*;

/**
 * @author Tomas Pavek
 */

public class JTabbedPaneSupport extends AbstractLayoutSupport
                                implements LayoutSupportArranging {

    private int selectedTab = -1;


    public Class getLayoutClass() {
        return null;
    }

    public int getNewIndex(Container container, Point posInCont,
                           Component component, Point posInComp) {
        if (!(container instanceof JTabbedPane))
            return -1;
        return container.getComponentCount();
    }

    public boolean paintDragFeedback(Container container,
                                     Component component,
                                     ConstraintsDesc newConstraints,
                                     int newIndex,
                                     Graphics g) {
        if (!(container instanceof JTabbedPane))
            return false;

        if (container.getComponentCount() == 0) {
            Dimension sz = container.getSize();
            Insets insets = container.getInsets();
            sz.width -= insets.left + insets.right;
            sz.height -= insets.top + insets.bottom;
            g.drawRect(0, 0, sz.width, sz.height);
        }
        else {
            Rectangle rect = container.getComponent(0).getBounds();
            g.drawRect(rect.x, rect.y, rect.width, rect.height);
        }
        return true;
    }
    
    public void addComponent(RADVisualComponent component, ConstraintsDesc desc) {
        if (desc == null)
            desc = new TabConstraintsDesc(component.getName());
        component.setConstraintsDesc(getClass(), desc);

        Container cont =
            getContainer().getContainerDelegate(getContainer().getBeanInstance());

        if (cont instanceof JTabbedPane) {
            JTabbedPane tabbedPane = (JTabbedPane)cont;
            TabConstraintsDesc tabDesc = (TabConstraintsDesc)desc;
            try {
                tabbedPane.addTab(tabDesc.getTitle(), tabDesc.getIcon(),
                                  component.getComponent(), tabDesc.getToolTip());
            }
            catch (ArrayIndexOutOfBoundsException e) {} // ignore
        }
    }

    public void removeComponent(RADVisualComponent component) {
        if (selectedTab >= 0) selectedTab = -1;
        super.removeComponent(component);
    }

    public String getJavaAddComponentString(/*RADVisualContainer cont,*/
                                            RADVisualComponent comp) {
        StringBuffer buf = new StringBuffer();

        String delegate = getContainer().getJavaContainerDelegateString();
        if (!"".equals(delegate)) { // NOI18N
            buf.append(delegate);
            buf.append("."); // NOI18N
        }
        buf.append("addTab("); // NOI18N

        ConstraintsDesc desc = getConstraints(comp);
        if (desc instanceof TabConstraintsDesc) {
            FormProperty titleProperty = (FormProperty)desc.getProperties()[0];
            FormProperty iconProperty = (FormProperty)desc.getProperties()[1];
            FormProperty toolTipProperty = (FormProperty)desc.getProperties()[2];

            buf.append(titleProperty.getJavaInitializationString());
            buf.append(", "); // NOI18N
            if (iconProperty.isChanged() || toolTipProperty.isChanged()) {
                buf.append(iconProperty.getJavaInitializationString());
                buf.append(", "); // NOI18N
            }
            buf.append(comp.getName());
            if (toolTipProperty.isChanged()) {
                buf.append(", "); // NOI18N
                buf.append(toolTipProperty.getJavaInitializationString());
            }
        }
        else { // no constraints for this component - use its name
            buf.append("\""); // NOI18N
            buf.append(comp.getName());
            buf.append("\", "); // NOI18N
            buf.append(comp.getName());
        }
        buf.append(");\n"); // NOI18N

        return buf.toString();
    }

    // ----------
    // LayoutSupportArranging

    public void processMouseClick(Point p, Container cont) {
        if (!(cont instanceof JTabbedPane)) return;

        JTabbedPane tabbedPane = (JTabbedPane)cont;
        int n = tabbedPane.getTabCount();
        for (int i=0; i < n; i++) {
            if (tabbedPane.getBoundsAt(i).contains(p)) {
                selectedTab = i;
                tabbedPane.setSelectedIndex(i);
                break;
            }
        }
    }

    public void selectComponent(RADVisualComponent comp) {
        int i = getContainer().getIndexOf(comp);
        selectedTab = i;
    }

    public void arrangeContainer(Container cont) {
        if (!(cont instanceof JTabbedPane)
                || selectedTab < 0) return;

        ((JTabbedPane)cont).setSelectedIndex(selectedTab);
    }

    // ----------

    public static class TabConstraintsDesc
            implements LayoutSupport.ConstraintsDesc
    {
        private String title;
        private Icon icon;
        private String toolTip;

        transient private Node.Property[] properties;

        public TabConstraintsDesc(String title) {
            this.title = title;
        }

        public TabConstraintsDesc(String title, Icon icon, String toolTip) {
            this.title = title;
            this.icon = icon;
            this.toolTip = toolTip;
        }

        public String getTitle() { 
            return title;
        }

        public Icon getIcon() {
            return icon;
        }

        public String getToolTip() {
            return toolTip;
        }

        // -----------

        public Node.Property[] getProperties() {
            if (properties == null)
                properties = new Node.Property[] {
                    new FormProperty("tabTitle", // NOI18N
                                     String.class,
                                 getBundle().getString("PROP_tabTitle"), // NOI18N
                                 getBundle().getString("HINT_tabTitle")) { // NOI18N

                        public Object getTargetValue() {
                            return title;
                        }

                        public void setTargetValue(Object value) {
                            title = (String)value;
                        }

                        protected Object getRealValue(Object value) {
                            Object realValue = super.getRealValue(value);
                            if (realValue == FormDesignValue.IGNORED_VALUE)
                                realValue = ((FormDesignValue)value).getDescription();
                            return realValue;
                        }
                    },

                    new FormProperty("tabIcon", // NOI18N
                                     Icon.class,
                                 getBundle().getString("PROP_tabIcon"), // NOI18N
                                 getBundle().getString("HINT_tabIcon")) { // NOI18N

                        public Object getTargetValue() {
                            return icon;
                        }

                        public void setTargetValue(Object value) {
                            icon = (Icon)value;
                        }

                        public boolean supportsDefaultValue() {
                            return true;
                        }

                        public Object getDefaultValue() {
                            return null;
                        }
                    },

                    new FormProperty("tabToolTip", // NOI18N
                                     String.class,
                                 getBundle().getString("PROP_tabToolTip"), // NOI18N
                                 getBundle().getString("HINT_tabToolTip")) { // NOI18N

                        public Object getTargetValue() {
                            return toolTip;
                        }

                        public void setTargetValue(Object value) {
                            toolTip = (String)value;
                        }

                        public boolean supportsDefaultValue() {
                            return true;
                        }

                        public Object getDefaultValue() {
                            return null;
                        }
                    }
                };

            return properties;
        }

        public Object getConstraintsObject() {
            return title;
        }

        public String getJavaInitializationString() {
            return null;
        }

        public void readFromXML(org.w3c.dom.Node element)
        throws java.io.IOException {
        }

        public org.w3c.dom.Node storeToXML(org.w3c.dom.Document doc) {
            return null;
        }
    }
}
