/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.palette;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import org.openide.awt.*;
import org.openide.cookies.InstanceCookie;
import org.openide.loaders.*;
import org.openide.*;
import org.openide.nodes.*;
import org.openide.util.*;
import org.openide.filesystems.*;

import org.netbeans.modules.form.*;

/**
 * The ComponentPalette is a visual component that manages a add
 * component/selection mode for FormEditor.  The current state of the palette
 * is returned by getSelectedItem() method.
 *
 * @author   Ian Formanek, Jan Jancura
 */

public class ComponentPalette extends JPanel
{
    static final long serialVersionUID = 7249158494688903041L;
    
    private static ResourceBundle bundle = null;
    
    // IDs for tasks running in AWT event-queue
    /** ID of task for updating one category (or more - from updateNodes). */
    private static final int UPDATE_CATEGORY = 1;
    /** ID of task for updating all categories of palette. */
    private static final int UPDATE_CATEGORIES = 2;
    /** ID of task for reordering categories. */
    private static final int REORDER_CATEGORIES = 3;
    /** ID of task for showing palette-tabs component (large toolbar). */
    private static final int SHOW_PALETTE = 4;
    /** ID of task for hiding palette-tabs component (showing small toolbar). */
    private static final int HIDE_PALETTE = 5;
    
    /** Set of category-nodes that are to be updated. */
    private Set updateNodes = new HashSet(11);
    
    /** The instance of ComponentPalette (singleton). */
    private static ComponentPalette instance;
    
    /** Whether pallete-tabs component is shown. */
    private boolean showPalette = FormEditor.getFormSettings().getPaletteTabsVisible();
    
    /** Current palette mode - MODE_SELECTION, MODE_CONNECTION or MODE_ADD. */
    private int paletteMode = PaletteAction.MODE_SELECTION;
    
    /** Root palette node - provides palette content (categories and items). */
    private PaletteNode paletteNode;
    
    /** List of all palette items (regardless to categories). */
    java.util.List allItems = null;
    
    /** Palette categories information objects. */
    CategoryInfo[] categories;
    
    CategoryInfo selectCategory;
    
    /** Maps CategoryNode -> CategoryInfo. */
    private HashMap nodeToCategory = new HashMap(10);
    
    /** PaletteItem which is currently selected (in Add mode). */
    private PaletteItem selectedPaletteItem;
    
    /** PaleteButton currently selected on palette. */
    private PaletteButton selectedButton = null;
    
    /** The button that represents Selection mode. */
    private ToolbarToggleButton selectionButton;
    
    /** The button that represents Connection mode. */
    private ToolbarToggleButton connectionButton;
    
    /** The button that represents Add mode and also pops beans menu. */
    private ToolbarToggleButton beansButton;
    
    /** Label with information about current mode and selected item. */
    private JLabel statusLabel;
    
    /** Item that signals that the component is preparing or null. */
    private JLabel wait;
    
    /** Palette tabs component. */
    private JTabbedPane tabbedArea = null;
    
    /** Popup menu for choosing a bean and then switching to Add mode. */
    private JPopupMenu beansMenu;
    
    // listeners
    private ActionListener buttonListener;
    private ActionListener beansMenuListener;
    private NodeListener categoryNodeListener;
    private MouseListener itemRightClickListener;
    private MouseListener categoryRightClickListener;
    private MouseListener paletteRightClickListener;
    
    // hack for resizing tabbedArea
    private int needToResizeState = 1;
    private int prevTabRunCount = 0;
    private int tabHeight = 0;
    private int maxTabWidth = 0;
    private int firstCompWidth = 0;
    
    
    private ComponentPalette() {
        // create listeners for buttons, menu items, nodes etc.
        buttonListener = new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                ToolbarToggleButton pressedButton =
                (ToolbarToggleButton) evt.getSource();
                
                if (pressedButton == selectionButton)
                    setMode(PaletteAction.MODE_SELECTION);
                
                else if (pressedButton == connectionButton)
                    setMode(PaletteAction.MODE_CONNECTION);
                
                else if (pressedButton == beansButton)
                    showBeansMenu();
                
                else { // some palette button pressed
                    if (pressedButton == selectedButton)
                        setMode(PaletteAction.MODE_SELECTION);
                    else {
                        PaletteButton palButton = (PaletteButton) pressedButton;
                        setSelectedItem(palButton.getItem());
                        setMode(PaletteAction.MODE_ADD);
                        statusLabel.setIcon(palButton.getIcon());
                        statusLabel.setText(palButton.getItem().getName());
                        selectedButton = palButton;
                    }
                }
            }
        };
        
        beansMenuListener = new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                PaletteMenuItem item = (PaletteMenuItem)evt.getSource();
                
                PaletteButton button = item.getButton();
                if (button != null && tabbedArea != null) {
                    CategoryInfo category = (CategoryInfo)nodeToCategory.get(
                    button.getCategoryNode());
                    if (category != null)
                        tabbedArea.setSelectedIndex(category.index);
                    else
                        button = null;
                }
                
                setSelectedItem(item.getItem());
                setMode(PaletteAction.MODE_ADD);
                statusLabel.setText(item.getItem().getName());
                statusLabel.setIcon(item.getIcon());
                
                if (button != null) {
                    button.setSelected(true);
                    selectedButton = button;
                }
            }
        };
        
        itemRightClickListener = new MouseAdapter() {
            public void mouseReleased(MouseEvent e) {
                if (MouseUtils.isRightMouseButton(e)) {
                    showMenuOnItem(((PaletteButton)e.getSource()).getItemNode(),
                    e.getComponent(), e.getPoint());
                }
            }
        };
        
        categoryRightClickListener = new MouseAdapter() {
            public void mouseReleased(MouseEvent e) {
                if (org.openide.awt.MouseUtils.isRightMouseButton(e)) {
                    if (e.getSource() == tabbedArea) {
                        int i = tabbedArea.getSelectedIndex();
                        if (i >= 0) {
                            Rectangle r = tabbedArea.getBoundsAt(i);
                            Point p = e.getPoint();
                            if (p.y < r.y
                            || (p.y < r.y+r.height
                            && (p.x < r.x || p.x >= r.x+r.width))) {
                                showMenuOnPalette(e.getComponent(), e.getPoint());
                                return;
                            }
                        }
                    }
                    showMenuOnCategory(e.getComponent(), e.getPoint());
                }
            }
        };
        
        paletteRightClickListener = new MouseAdapter() {
            public void mouseReleased(MouseEvent e) {
                if (org.openide.awt.MouseUtils.isRightMouseButton(e)) {
                    showPaletteVisibilityMenu(e.getComponent(), e.getPoint());
                }
            }
        };
        
        FormEditor.getFormSettings().addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                if (FormLoaderSettings.PROP_PALETTE_TABS_VISIBLE.equals(
                evt.getPropertyName()))
                    runInEventQueue(Boolean.TRUE.equals(evt.getNewValue()) ?
                    SHOW_PALETTE : HIDE_PALETTE);
            }
        });
        
        paletteNode = PaletteNode.getPaletteNode();
        paletteNode.addNodeListener(new NodeAdapter() {
            public void childrenAdded(NodeMemberEvent evt) {
                t("added categories"); // NOI18N
                runInEventQueue(UPDATE_CATEGORIES);
                needToResizeState = 1;
            }
            
            public void childrenRemoved(NodeMemberEvent evt) {
                t("removed categories"); // NOI18N
                runInEventQueue(UPDATE_CATEGORIES);
                needToResizeState = 1;
            }
            
            public void childrenReordered(NodeReorderEvent evt) {
                t("reordered categories"); // NOI18N
                runInEventQueue(REORDER_CATEGORIES);
                needToResizeState = 1;
            }
        });
        
        categoryNodeListener = new NodeAdapter() {
            public synchronized void childrenAdded(NodeMemberEvent evt) {
                t("beans added to: "+((PaletteCategoryNode)evt.getSource()).getName()); // NOI18N
                updateCategoryInEventQueue((PaletteCategoryNode)evt.getSource());
            }
            
            public void childrenRemoved(NodeMemberEvent evt) {
                t("beans removed from: "+((PaletteCategoryNode)evt.getSource()).getName()); // NOI18N
                updateCategoryInEventQueue((PaletteCategoryNode)evt.getSource());
            }
            
            public void childrenReordered(NodeReorderEvent evt) {
                t("beans reordered in: "+((PaletteCategoryNode)evt.getSource()).getName()); // NOI18N
                updateCategoryInEventQueue((PaletteCategoryNode)evt.getSource());
            }
            
            public void propertyChange(PropertyChangeEvent evt) {
                if (Node.PROP_NAME.equals(evt.getPropertyName())
                || Node.PROP_DISPLAY_NAME.equals(evt.getPropertyName())) {
                    t("category "+evt.getOldValue()+" renamed to "+evt.getNewValue()); // NOI18N
                    // whole update??
                    updateCategoryInEventQueue((PaletteCategoryNode)evt.getSource());
                }
            }
        };

        HelpCtx.setHelpIDString(this, "gui.components.palette"); // NOI18N

        if (showPalette) largeLayout();
        else smallLayout();
        
        addMouseListener(paletteRightClickListener);
        
        if (showPalette) {
            wait = new JLabel(getBundle().getString("MSG_PreparingPalette")); // NOI18N
            wait.setHorizontalAlignment(SwingConstants.CENTER);
            add(wait, BorderLayout.CENTER);
        }
        
        runInEventQueue(UPDATE_CATEGORIES);
    }
    
    void largeLayout() {
        if (isCreated())
            removeAll();
        
        setLayout(new BorderLayout());
        
        JPanel controlsArea = new JPanel(new BorderLayout(0,5));
        controlsArea.setBorder(new EmptyBorder(3, 2, 3, 3));
        
        JPanel buttPanel = new JPanel();
        buttPanel.setLayout(new BoxLayout(buttPanel, BoxLayout.X_AXIS));
        
        createControls();
        
        statusLabel.setBorder(new EmptyBorder(2,2,0,0));
        statusLabel.setHorizontalAlignment(SwingConstants.LEFT);
        setStatusLabelSize();
        
        setControlsMode();
        
        buttPanel.add(selectionButton);
        buttPanel.add(Box.createHorizontalStrut(2));
        buttPanel.add(connectionButton);
        buttPanel.add(Box.createHorizontalStrut(2));
        buttPanel.add(beansButton);
        
        controlsArea.add(buttPanel, BorderLayout.CENTER);
        controlsArea.add(statusLabel, BorderLayout.NORTH);
        add(controlsArea, BorderLayout.WEST);
    }
    
    void smallLayout() {
        if (isCreated())
            removeAll();
        
        setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
        
        createControls();
        
        statusLabel.setBorder(new EmptyBorder(2,0,0,0));
        statusLabel.setHorizontalAlignment(SwingConstants.CENTER);
        setStatusLabelSize();
        
        setControlsMode();
        
        add(statusLabel);
        add(Box.createHorizontalStrut(5));
        add(selectionButton);
        add(Box.createHorizontalStrut(2));
        add(connectionButton);
        add(Box.createHorizontalStrut(2));
        add(beansButton);
    }
    
    // -------------------
    
    /** Does Component Palette already exist? */
    public static boolean isCreated() {
        return instance != null;
    }
    
    public static ComponentPalette getDefault() {
        if (instance != null)
            return instance;
        
        synchronized (ComponentPalette.class) {
            if (instance != null)
                return instance;
            
            t("creating Component Palette..."); // NOI18N
            instance = new ComponentPalette();
            t("...Component Palette created\n"); // NOI18N
            return instance;
        }
    }
    
    //    public static DataFolder getPaletteFolder() {
    //        return PaletteNode.getPaletteFolder();
    //    }
    
    //
    // Public interface
    //
    
    public void updatePalette() {
        categories = null;
        runInEventQueue(UPDATE_CATEGORIES);
    }
    
    //    public PaletteCategoryNode[] getCategoriesNodes() {
    //        return paletteNode.getPaletteCategories();
    //    }
    
    public PaletteItem[] getAllItems() {
        // TODO: filter out multiple occurences of same bean
        if (allItems == null) {
            if (categories != null) {
                allItems = new ArrayList(categories.length*10);
                for (int i=0; i < categories.length; i++)
                    if (!categories[i].isEmpty())
                        allItems.addAll(categories[i].items);
            }
            else
                allItems = new ArrayList();
        }
        
        PaletteItem[] items = new PaletteItem[allItems.size()];
        allItems.toArray(items);
        return items;
    }
    
    public int getMode() {
        return paletteMode;
    }
    
    public void setMode(int newMode) {
        paletteMode = newMode;
        
        FormDesigner designer = getDesigner();
        
        if (selectedButton != null) {
            selectedButton.setSelected(false);
            selectedButton = null;
        }
        
        if (newMode == PaletteAction.MODE_SELECTION
        || newMode == PaletteAction.MODE_CONNECTION) {
            setSelectedItem(null);
            statusLabel.setIcon(null);
            
            if (newMode == PaletteAction.MODE_SELECTION) {
                if (designer != null)
                    designer.resetConnection();
            }
            else { // MODE_CONNECTION
                if (designer != null)
                    designer.clearSelection();
            }
        }
        else if (newMode == PaletteAction.MODE_ADD) {
            if (designer != null)
                designer.resetConnection();
        }
        
        setControlsMode();
    }
    
    private FormDesigner getDesigner() {
        Node root = ComponentInspector.getInstance().getExplorerManager()
        .getRootContext();
        RADComponentCookie cookie =
        (RADComponentCookie)root.getCookie(RADComponentCookie.class);
        return cookie == null ? null :
            cookie.getRADComponent().getFormModel().getFormDesigner();
    }
    
    /** Returns the currently selected PaletteNode(that represents a JavaBean to
     * be added to the form) or null for selection mode.
     * @return the currently selected PaletteNode or null for selection mode
     */
    public PaletteItem getSelectedItem() {
        return selectedPaletteItem;
    }
    
    /** Sets the current add component or selection mode.
     * @param value The new add component or null for selection mode
     */
    public void setSelectedItem(PaletteItem item) {
        selectedPaletteItem = item;
    }
    
    // --------------
    
    void updateCategoryInEventQueue(PaletteCategoryNode node) {
        Set nodes = updateNodes;
        boolean noRefreshTask = updateNodes.size() == 0;
        updateNodes.add(node);
        t("added category to update: "+node.getName()); // NOI18N
        if (noRefreshTask || nodes != updateNodes)
            runInEventQueue(UPDATE_CATEGORY);
    }
    
    // called from AWT event-queue
    private void showPaletteTabs() {
        setMode(PaletteAction.MODE_SELECTION);
        largeLayout();
        createPaletteTabs(); // creates tabbedArea
        
        revalidate();
        repaint();
        
        showPalette = true;
    }
    
    // called from AWT event-queue
    private void hidePaletteTabs() {
        if (tabbedArea != null) {
            remove(tabbedArea);
            tabbedArea = null;
            for (int i=0; i < categories.length; i++) {
                categories[i].buttonsPanel = null;
                categories[i].tabComponent = null;
                categories[i].itemsMenu = null;
            }
            beansMenu = null;
            
            setMode(PaletteAction.MODE_SELECTION);
            smallLayout();
            
            revalidate();
            repaint();
        }
        showPalette = false;
    }
    
    // called from AWT event-queue
    private void updateCategories() {
        t("updating categories..."); // NOI18N
        boolean wasEmpty = categories == null || categories.length == 0;
        //        PaletteCategoryNode[] catNodes = paletteNode.getPaletteCategories();
        Node[] catNodes = paletteNode.getCategoriesNodes();
        if (wasEmpty && catNodes.length == 0) {
            t(" - no categories to update (yet)"); // NOI18N
            return;
        }
        
        if (tabbedArea != null) {
            int tabIndex = tabbedArea.getSelectedIndex();
            if (tabIndex >= 0 && !wasEmpty)
                selectCategory = categories[tabIndex];
            remove(tabbedArea);
            tabbedArea.removeAll();
            tabbedArea = null;
        }
        
        beansMenu = null;
        
        CategoryInfo[] newCategories = new CategoryInfo[catNodes.length];
        
        // handle added categories
        for (int i=0; i < catNodes.length; i++) {
            Node node = catNodes[i];
            CategoryInfo category = wasEmpty ? null :
                (CategoryInfo)nodeToCategory.get(node);
                if (category == null) { // new category
                    category = new CategoryInfo(i, (PaletteCategoryNode)node);
                    nodeToCategory.put(node, category);
                    node.addNodeListener(categoryNodeListener);
                    //                updateCategoryInEventQueue(node);
                    updateCategory((PaletteCategoryNode)node);
                }
                else { // already existing category
                    category.index = i;
                }
                
                newCategories[i] = category;
        }
        
        if (!wasEmpty) // handle removed categories
            for (int i=0; i < categories.length; i++) {
                CategoryInfo category = categories[i];
                for (int j=0; j < newCategories.length; j++)
                    if (category == newCategories[j]) {
                        category = null;
                        break;
                    }
                
                if (category != null) {// this category was removed
                    category.node.removeNodeListener(categoryNodeListener);
                    nodeToCategory.remove(category.node);
                    if (category == selectCategory)
                        selectCategory = null;
                }
            }
        
        categories = newCategories;
        
        if (showPalette) {
            //        runInEventQueue(UPDATE_TABS);
            //        /*
            if (wait != null) {
                remove(wait);
                wait = null;
            }
            
            createPaletteTabs(); // creates tabbedArea
            
            revalidate();
            repaint();
        }
        //         */
        
        t("...categories update done"); // NOI18N
    }
    
    // called from AWT event-queue
    private void reorderCategories() {
        if (tabbedArea != null) {
            int tabIndex = tabbedArea.getSelectedIndex();
            if (tabIndex >= 0 && categories != null)
                selectCategory = categories[tabIndex];
        }
        
        beansMenu = null;
        
        // nodes in new order
        //        PaletteCategoryNode[] catNodes = paletteNode.getPaletteCategories();
        Node[] catNodes = paletteNode.getCategoriesNodes();
        
        for (int i=0; i < catNodes.length; i++) {
            CategoryInfo cat = (CategoryInfo)nodeToCategory.get(catNodes[i]);
            for (int j=i; j < categories.length; j++)
                if (categories[j] == cat) {
                    categories[j] = categories[i];
                    categories[i] = cat;
                    cat.index = i;
                    break;
                }
        }
        
        if (tabbedArea != null) {
            remove(tabbedArea);
            tabbedArea.removeAll();
            createPaletteTabs(); // creates tabbedArea
            
            revalidate();
            repaint();
        }
    }
    
    // called from AWT event-queue
    private void updateCategory(PaletteCategoryNode catNode) {
        t("updating category: "+catNode.getName()); // NOI18N
        CategoryInfo category = (CategoryInfo) nodeToCategory.get(catNode);
        if (category == null) return;
        
        if (tabbedArea != null) {
            String catName = catNode.getDisplayName();
            if (!catName.equals(tabbedArea.getTitleAt(category.index)))
                tabbedArea.setTitleAt(category.index, catName);
        }
        
        Node[] categoryItemsNodes = catNode.getChildren().getNodes();
        
        if (categoryItemsNodes.length > 0 || !category.isEmpty()) {
            t(" - "+categoryItemsNodes.length+(categoryItemsNodes.length==1?" bean":" beans")); // NOI18N
            
            allItems = null;
            beansMenu = null;
            category.itemsMenu = null;
            
            if (category.items == null) {
                if (categoryItemsNodes.length > 0)
                    category.items = new ArrayList(categoryItemsNodes.length);
            }
            else
                category.items.clear();
            
            if (category.buttonsPanel != null)
                category.buttonsPanel.removeAll();
            
            for (int i = 0; i < categoryItemsNodes.length; i++) {
                Node itemNode = categoryItemsNodes[i];
                
                InstanceCookie ic =
                (InstanceCookie)itemNode.getCookie(InstanceCookie.class);
                if (ic == null) continue;
                
                PaletteItem palItem = createPaletteItem(itemNode);
                if (palItem != null) { // item is OK
                    category.items.add(palItem);
                    
                    if (category.buttonsPanel != null) {
                        PaletteButton button = new PaletteButton(palItem);
                        button.addActionListener(buttonListener);
                        button.addMouseListener(itemRightClickListener);
                        category.buttonsPanel.add(button);
                    }
                }
                else // problems creating -> do not present in palette
                    System.err.println("[WARNING] Cannot load bean "+ic.instanceName()); // NOI18N
            }
            
            if (getMode() == PaletteAction.MODE_ADD)
                setMode(PaletteAction.MODE_SELECTION);
        }
        else t(" - no beans (yet)"); // NOI18N
        
        if (category.buttonsPanel != null)
            category.buttonsPanel.invalidate();
    }
    
    // -----------------
    
    private static PaletteItem createPaletteItem(Node itemNode) {
        PaletteItem palItem = tryCreatePaletteItem(itemNode);
        if (palItem != null) return palItem;

/*        String dirs[];
        
        if (System.getProperty("netbeans.user").equals( // NOI18N
        System.getProperty("netbeans.home"))) { // NOI18N
            dirs = new String[1];
            dirs[0] = System.getProperty("netbeans.user"); // NOI18N
        }
        else {
            dirs = new String[2];
            dirs[0] = System.getProperty("netbeans.home"); // NOI18N
            dirs[1] = System.getProperty("netbeans.user"); // NOI18N
        }
        
        for (int i = 0; i < dirs.length; i++) {
            Properties libs = new Properties();
            File libsFile = new File(dirs[i] + File.separator + "beans"
                                + File.separator + "libs.properties"); // NOI18N
            
            try {
                libs.load(new FileInputStream(libsFile));
            }
            catch (IOException e) {
                continue;
            }
            
            Enumeration libKeys = libs.propertyNames();
            Repository rep = TopManager.getDefault().getRepository();
            
            while (libKeys.hasMoreElements()) {
                String libName = libs.getProperty((String)libKeys.nextElement());
                JarFileSystem jar = null;
                try {
                    File file = new File(libName);
                    jar = new GlobalJarFileSystem();
                    jar.setJarFile(file);
                }
                catch (Exception ex) {
                    return null;
                }
                
                if (rep.findFileSystem(jar.getSystemName()) != null)
                    continue;
                
                jar.setHidden(true);
                rep.addFileSystem(jar);
                
                palItem = tryCreatePaletteItem(itemNode);
                if (palItem != null) return palItem;
                
                rep.removeFileSystem(jar);
            }
        } */
        return null;
    }

    private static PaletteItem tryCreatePaletteItem(Node itemNode) {
        try {
            return new PaletteItem(itemNode);
        }
        catch (ThreadDeath td) {
            throw td;
        }
        catch (Throwable th) {
        }
        return null;
    }
    
    // ------------------
    
    private void createControls() {
        selectionButton = new ToolbarToggleButton(
            new ImageIcon(getClass().getResource(
                "/org/netbeans/modules/form/resources/selectionMode.gif")),false); // NOI18N
        selectionButton.addActionListener(buttonListener);
        selectionButton.setToolTipText(getBundle().getString("CTL_SelectionButtonHint")); // NOI18N
        HelpCtx.setHelpIDString(selectionButton, "gui.components.palette"); // NOI18N
        
        connectionButton = new ToolbarToggleButton(
            new ImageIcon(getClass().getResource(
                "/org/netbeans/modules/form/resources/connectionMode.gif")),false); // NOI18N
        connectionButton.addActionListener(buttonListener);
        connectionButton.setToolTipText(getBundle().getString("CTL_ConnectionButtonHint")); // NOI18N
        HelpCtx.setHelpIDString(connectionButton, "gui.connecting.intro"); // NOI18N
        
        beansButton = new ToolbarToggleButton(
            new ImageIcon(getClass().getResource(
                "/org/netbeans/modules/form/resources/beansButton.gif")),false); // NOI18N
        beansButton.setToolTipText(getBundle().getString("CTL_BeansButtonHint")); // NOI18N
        beansButton.addActionListener(buttonListener);
        HelpCtx.setHelpIDString(beansButton, "gui.components.adding"); // NOI18N
        
        ButtonGroup g = new ButtonGroup();
        g.add(selectionButton);
        g.add(connectionButton);
        g.add(beansButton);
        
        statusLabel = new JLabel();
        statusLabel.addMouseListener(paletteRightClickListener);
    }
    
    private void setStatusLabelSize() {
        statusLabel.setText(getBundle().getString("CTL_ConnectionMode")); // NOI18N
        Dimension prefSize1 = statusLabel.getPreferredSize();
        statusLabel.setText(getBundle().getString("CTL_SelectionMode")); // NOI18N
        Dimension prefSize2 = statusLabel.getPreferredSize();
        
        if (prefSize1.width < prefSize2.width)
            prefSize1.width = prefSize2.width;
        prefSize1.width += 10;
        if (prefSize1.width < 90)
            prefSize1.width = 90;
        else if (prefSize1.width > 130)
            prefSize1.width = 130;
        
        statusLabel.setPreferredSize(prefSize1);
        statusLabel.setMaximumSize(prefSize1);
    }
    
    private void setControlsMode() {
        int mode = getMode();
        if (mode == PaletteAction.MODE_SELECTION) {
            selectionButton.setSelected(true);
            statusLabel.setText(getBundle().getString("CTL_SelectionMode")); // NOI18N
            statusLabel.setToolTipText(getBundle().getString("CTL_SelectionModeHint")); // NOI18N
        }
        else if (mode == PaletteAction.MODE_CONNECTION) {
            connectionButton.setSelected(true);
            statusLabel.setText(getBundle().getString("CTL_ConnectionMode")); // NOI18N
            statusLabel.setToolTipText(getBundle().getString("CTL_ConnectionModeHint")); // NOI18N
        }
        else {
            beansButton.setSelected(true);
            statusLabel.setToolTipText(getBundle().getString("CTL_AddModeHint")); // NOI18N
        }
    }
    
    private void createPaletteTabs() {
        t("creating palette tabs"); // NOI18N
        tabbedArea = new JTabbedPane();

        if (needToResizeState == 2) {
            tabbedArea.setPreferredSize(new Dimension(0, tabHeight));
            tabbedArea.setSize(new Dimension(0, tabHeight));
            if (needToResizeState == 2) needToResizeState = 0;
        }

        beansMenu = null;
        
        if (categories != null)
            for (int i=0; i < categories.length; i++) {
                CategoryInfo cat = categories[i];
                if (cat.tabComponent == null)
                    createCategoryTab(cat);
                tabbedArea.addTab(cat.node.getDisplayName(), cat.tabComponent);
            }
        
        if (selectCategory != null) {
            tabbedArea.setSelectedIndex(selectCategory.index);
            selectCategory = null;
        }
        
        tabbedArea.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                if (getMode() == PaletteAction.MODE_ADD)
                    setMode(PaletteAction.MODE_SELECTION);
            }
        });
        
        tabbedArea.addMouseListener(categoryRightClickListener);
        
        add(tabbedArea, BorderLayout.CENTER);
        
        org.openide.TopManager.getDefault().getWindowManager().getMainWindow().invalidate();
        org.openide.TopManager.getDefault().getWindowManager().getMainWindow().validate();
        org.openide.TopManager.getDefault().getWindowManager().getMainWindow().repaint();
    }
    
    private void createCategoryTab(CategoryInfo category) {
        t("creating tab for "+category.getName()); // NOI18N
        ScrollPalettePanel buttonsPanel = new ScrollPalettePanel();
        buttonsPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 2, 2));
        buttonsPanel.addMouseListener(categoryRightClickListener);
        ScrollPalette scroll = new ScrollPalette(buttonsPanel);
        scroll.addMouseListener(categoryRightClickListener);
        
        if (!category.isEmpty())
            for (Iterator it = category.items.iterator(); it.hasNext(); ) {
                PaletteButton button = new PaletteButton((PaletteItem)it.next());
                button.addActionListener(buttonListener);
                button.addMouseListener(itemRightClickListener);
                buttonsPanel.add(button);
            }
        
        category.buttonsPanel = buttonsPanel;
        category.tabComponent = scroll;
        category.itemsMenu = null;
    }
    
    private void createBeansMenu() {
        t("create beans menu"); // NOI18N
        beansMenu = new ScrollPopupMenu(531);
        if (categories != null)
            for (int i=0; i < categories.length; i++) {
                CategoryInfo category = categories[i];
                if (category.itemsMenu == null)
                    createCategoryMenu(category);
                
                JMenu menu = category.itemsMenu;
                if (menu.getItemCount() > 0)
                    beansMenu.add(menu);
            }
        
        beansMenu.addPopupMenuListener(new PopupMenuListener() {
            public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        if (getMode() == PaletteAction.MODE_SELECTION)
                            selectionButton.setSelected(true);
                        else if (getMode() == PaletteAction.MODE_CONNECTION)
                            connectionButton.setSelected(true);
                        ComponentPalette.this.requestFocus();
                    }
                });
            }
            public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
            }
            public void popupMenuCanceled(PopupMenuEvent e) {
            }
        });
    }
    
    private void createCategoryMenu(CategoryInfo category) {
        t("create menu for category "+category.getName()); // NOI18N
        JMenu menu = new ScrollMenu(category.node.getDisplayName(), 531);
        PaletteMenuItem pmItem;
        
        if (category.buttonsPanel != null) {
            Component[] buttons = category.buttonsPanel.getComponents();
            for (int i=0; i < buttons.length; i++)
                if (buttons[i] instanceof PaletteButton) {
                    pmItem = new PaletteMenuItem((PaletteButton)buttons[i]);
                    pmItem.addActionListener(beansMenuListener);
                    menu.add(pmItem);
                }
        }
        else if (!category.isEmpty()) {
            for (Iterator it=category.items.iterator(); it.hasNext(); ) {
                pmItem = new PaletteMenuItem((PaletteItem)it.next());
                pmItem.addActionListener(beansMenuListener);
                menu.add(pmItem);
            }
        }
        
        category.itemsMenu = menu;
    }
    
    // -------------
    
    private void showBeansMenu() {
        Point p = selectionButton.getLocation();
        p.y += selectionButton.getHeight()+2;
        
        if (beansMenu == null)
            createBeansMenu();
        
        beansMenu.show(beansButton.getParent(), p.x, p.y);
    }
    
    void showMenuOnItem(Node itemNode, Component comp, Point pos) {
        new ItemPopupMenu(itemNode).show(comp, pos.x, pos.y);
    }
    
    void showMenuOnCategory(Component comp, Point pos) {
        int i = tabbedArea.getSelectedIndex();
        Node[] nodes = paletteNode.getChildren().getNodes();
        if (i < 0 || nodes == null || i >= nodes.length) return;
        
        new CategoryPopupMenu(nodes[i]).show(comp, pos.x, pos.y);
    }
    
    void showMenuOnPalette(Component comp, Point pos) {
        new PalettePopupMenu().show(comp, pos.x, pos.y);
    }
    
    void showPaletteVisibilityMenu(Component comp, Point pos) {
        JPopupMenu menu = new JPopupMenuPlus();
        JMenuItem item = new JMenuItem(getBundle().getString(showPalette ?
                             "CTL_HidePalette" : "CTL_ShowPalette")); // NOI18N
        item.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                FormEditor.getFormSettings().setPaletteTabsVisible(!showPalette);
            }
        });
        menu.add(item);
        
        menu.show(comp, pos.x, pos.y);
    }
    
    // ----------------
    // AWT methods
    
    public Dimension getPreferredSize() {
        Dimension pref = super.getPreferredSize();
        
        if (categories != null && categories.length > 0 && tabbedArea != null
                && categories.length == tabbedArea.getTabCount()) {
            
            int tabbedW = tabbedArea.getWidth();
            int tabbedCount = tabbedArea.getTabCount();
            
            if (tabbedW != 0) {
                // gets width of first panel in ComponentPalette
                if (firstCompWidth == 0)  firstCompWidth = this.getComponent(0).getWidth();
                
                // gets maximum allowed width for tabbedArea
                FontMetrics fm = null;
                Font fnt = tabbedArea.getFont();
                if (fnt != null) fm = tabbedArea.getFontMetrics(fnt);
                if (fm != null) {   // try from fontmetrics
                    maxTabWidth = 50;
                    for (int i=0; i < tabbedCount; i++) {
                        String title = tabbedArea.getTitleAt(i);
                        if (title != null)
                            if (maxTabWidth < fm.stringWidth(title)) maxTabWidth = fm.stringWidth(title);
                    }
                }
                else {  // try from tabs widths (if fontmetrics is not available)
                    for (int i=0; i < tabbedCount; i++) {
                        Rectangle r = tabbedArea.getBoundsAt(i);
                        if (maxTabWidth < r.width) maxTabWidth = r.width;
                    }
                }
                
                // prepare some values...
                int currentTabRunCount = tabbedArea.getTabRunCount();
                int panelY = tabbedArea.getComponentAt(0).getY();
                Rectangle r = tabbedArea.getBoundsAt(0);
                Rectangle s = tabbedArea.getBoundsAt(tabbedCount-1);
                
                // gets tabbedMidder = the space between tabs and panel of beansicons (HACK for MAC)
                int tabbedMidder = panelY - r.y -21;
                if (tabbedMidder<0) tabbedMidder = 0;
                
                // gets tabHeight = correct height of tabbedArea
                int tabbedHeader = r.y - s.y + s.height;
                tabHeight = 36 + tabbedHeader + tabbedMidder + (int)((currentTabRunCount)/2);
                
                // correction of many odd tab lines (more then 7 lines - only odd!!)
                if (currentTabRunCount > 7  &&  ((double)(currentTabRunCount - 1)%2) == 0) tabHeight +=20;
                
                // there was some change from user, it is needed to repaint ComponentPalette 
                // (changes: create tab, remove tab, reorder tab, change of tab lines)
                if ((needToResizeState == 1) || (prevTabRunCount != currentTabRunCount)) {
                    prevTabRunCount = currentTabRunCount;
                    needToResizeState = 2;
                    runInEventQueue(HIDE_PALETTE);
                    runInEventQueue(SHOW_PALETTE);
                }
            }
        }
        
        if (maxTabWidth > 0 && firstCompWidth > 0) pref.width = maxTabWidth + firstCompWidth + 20;
        return pref;
    }
    
    public Dimension getMaximumSize() {
        return new Dimension(1280, super.getPreferredSize().height);
    }
    
    public Dimension getMinimumSize() {
        Dimension pref = super.getPreferredSize();

        if (maxTabWidth > 0 && firstCompWidth > 0) pref.width = maxTabWidth + firstCompWidth + 20;
        return pref;
    }
    
    // -------------
    
    private class CategoryInfo {
        int index;
        PaletteCategoryNode node;
        Container buttonsPanel;
        Component tabComponent;
        java.util.List items;
        JMenu itemsMenu;
        
        CategoryInfo(int index, PaletteCategoryNode node ) {
            this.index = index;
            this.node = node;
            buttonsPanel = null;
            tabComponent = null;
            items = null;
            itemsMenu = null;
        }
        
        String getName() {
            return node.getName();
        }
        
        boolean isEmpty() {
            return items == null || items.size() == 0;
        }
    }
    
    // -------------------
    
    void runInEventQueue(int toDo) {
        t("placing task, id: "+toDo); // NOI18N
        EventQueue.invokeLater(new PaletteTask(toDo));
    }
    
    private class PaletteTask implements Runnable {
        int whatToDo;
        
        PaletteTask(int toDo) {
            whatToDo = toDo;
        }
        
        public void run() {
            t("performing task, id: "+whatToDo); // NOI18N
            if (whatToDo == UPDATE_CATEGORY) {
                Set s = updateNodes;
                updateNodes = new HashSet(11);
                
                Iterator it = s.iterator();
                while (it.hasNext()) {
                    PaletteCategoryNode node = (PaletteCategoryNode) it.next();
                    updateCategory(node);
                }
                
                if (tabbedArea != null) {
                    revalidate();
                    repaint();
                }
            }
            
            else if (whatToDo == REORDER_CATEGORIES) {
                reorderCategories();
            }
            
            else if (whatToDo == UPDATE_CATEGORIES) {
                updateCategories();
            }
            
            else if (whatToDo == SHOW_PALETTE) {
                showPaletteTabs();
            }
            
            else if (whatToDo == HIDE_PALETTE) {
                hidePaletteTabs();
            }
            t(null);
        }
    }
    
    // --------------
    
    static ResourceBundle getBundle() {
        if (bundle == null)
            bundle = NbBundle.getBundle(ComponentPalette.class);
        return bundle;
    }
    
    // ---------------
    
    /** For debugging purposes only. */
    static private final boolean TRACE = false;
    /** For debugging purposes only. */
    static private void t(String str) {
        if (TRACE)
            if (str != null)
                System.out.println("CP: "+str); // NOI18N
            else
                System.out.println(""); // NOI18N
    }
}
