/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.form.palette;

import java.util.ResourceBundle;
import java.text.MessageFormat;
import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.datatransfer.*;

import org.openide.awt.JPopupMenuPlus;
import org.openide.nodes.*;
import org.openide.*;


/** Context menu for palette items (invoked on palette buttons).
 *
 * @author Tomas Pavek
 */
class ItemPopupMenu extends JPopupMenuPlus {

    private Node itemNode;

    public ItemPopupMenu(Node node) {
        itemNode = node;

        JMenuItem menuItem;
        Node parent = itemNode.getParentNode();
        Index indexCookie = parent != null ?
                        (Index)parent.getCookie(Index.class) : null;

        ResourceBundle bundle = ComponentPalette.getBundle();

        menuItem = new JMenuItem(bundle.getString("CTL_MoveLeft")); // NOI18N
        menuItem.setEnabled(indexCookie != null
                                && indexCookie.indexOf(itemNode) > 0);
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                moveLeft();
            }
        });
        add(menuItem);

        menuItem = new JMenuItem(bundle.getString("CTL_MoveRight")); // NOI18N
        if (indexCookie != null) {
            int ind = indexCookie.indexOf(itemNode);
            menuItem.setEnabled(ind >= 0 && ind+1 < indexCookie.getNodesCount());
        }
        else menuItem.setEnabled(false);
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                moveRight();
            }
        });
        add(menuItem);

        addSeparator();

        menuItem = new JMenuItem(bundle.getString("CTL_Cut")); // NOI18N
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                cutNode();
            }
        });
        add(menuItem);

        menuItem = new JMenuItem(bundle.getString("CTL_Copy")); // NOI18N
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                copyNode();
            }
        });
        add(menuItem);

        addSeparator();

        menuItem = new JMenuItem(bundle.getString("CTL_Delete")); // NOI18N
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                deleteNode();
            }
        });
        add(menuItem);
    }

    private void moveLeft() {
        Node parent = itemNode.getParentNode();
        if (parent == null) return;
        Index indexCookie = (Index)parent.getCookie(Index.class);
        if (indexCookie == null) return;
        int ind = indexCookie.indexOf(itemNode);
        if (indexCookie != null && indexCookie.indexOf(itemNode) > 0)
            indexCookie.moveUp(ind);
    }

    private void moveRight() {
        Node parent = itemNode.getParentNode();
        if (parent == null) return;
        Index indexCookie = (Index)parent.getCookie(Index.class);
        if (indexCookie == null) return;
        int ind = indexCookie.indexOf(itemNode);
        if (ind >=0 && ind+1 < indexCookie.getNodesCount())
            indexCookie.moveDown(ind);
    }

    private void cutNode() {
        try {
            Transferable trans = itemNode.clipboardCut();
            if (trans != null) {
                Clipboard clipboard = TopManager.getDefault().getClipboard();
                clipboard.setContents(trans, new StringSelection("")); // NOI18N
            }
        }
        catch (java.io.IOException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                e.printStackTrace();
            TopManager.getDefault().getErrorManager().notify(e);
        }
    }

    private void copyNode() {
        try {
            Transferable trans = itemNode.clipboardCopy();
            if (trans != null) {
                Clipboard clipboard = TopManager.getDefault().getClipboard();
                clipboard.setContents(trans, new StringSelection("")); // NOI18N
            }
        }
        catch (java.io.IOException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                e.printStackTrace();
            TopManager.getDefault().getErrorManager().notify(e);
        }
    }

    private void deleteNode() {
        // first user confirmation...
        String message = MessageFormat.format(
            ComponentPalette.getBundle().getString("FMT_ConfirmBeanDelete"), // NOI18N
            new Object[] { itemNode.getDisplayName() });

        NotifyDescriptor desc = new NotifyDescriptor.Confirmation(message,
                    ComponentPalette.getBundle().getString("CTL_ConfirmBeanTitle"), // NOI18N
                    NotifyDescriptor.YES_NO_OPTION);

        if (NotifyDescriptor.YES_OPTION.equals(TopManager.getDefault().notify(desc))) {
            try {
                itemNode.destroy();
            }
            catch (java.io.IOException e) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                    e.printStackTrace();
                TopManager.getDefault().getErrorManager().notify(e);
            }
        }
    }
}
