/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.form.palette;

import java.awt.Image;
import java.awt.Toolkit;
import java.beans.BeanInfo;
import java.awt.datatransfer.Transferable;
import java.util.*;
import java.text.MessageFormat;

import org.openide.*;
import org.openide.actions.*;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.HelpCtx;
import org.openide.util.Utilities;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.*;

/**
 * The PaletteCategoryNode is a Node representing the ComponentPalette content
 * in the tree under Environment.
 * @author   Ian Formanek
 */
public class PaletteCategoryNode extends DataFolder.FolderNode
{
    /** generated Serialized Version UID */
    //  static final long serialVersionUID = -2098259549820241091L;

    /** icons for the PaletteCategoryNode */
    private static String iconURL =
        "/org/netbeans/modules/form/resources/paletteCategory.gif"; // NOI18N
    private static String icon32URL =
        "/org/netbeans/modules/form/resources/paletteCategory32.gif"; // NOI18N

    private static final String PALETTE_FOLDER_NAME = "Palette"; // NOI18N
    private static final Node.PropertySet[] NO_PROPERTIES = new Node.PropertySet[0];

    private static Image substIcon, substIcon32;
    private static SystemAction[] staticActions;

    private DataFolder folder;

    public PaletteCategoryNode(DataFolder folder) {
        folder.super(new PaletteCategoryNodeChildren(folder));
        this.folder = folder;

        Object catName = folder.getPrimaryFile().getAttribute("categoryName"); // NOI18N
        if (catName != null && catName instanceof String)
            setName((String)catName, false);
    }

    public void setName(String name) {
        setName(name, true);
    }

    public void setName(String name, boolean rename) {
        if (rename) {
            if (!checkCategoryName(name,this))
                return;
            try {
                DataObject dobj = getDataObject();
                FileObject fobj = dobj.getPrimaryFile();
                String folderName = convertToFolderName(name, fobj.getName());
                fobj.setAttribute("categoryName", null); // NOI18N
                dobj.rename(folderName);
                if (!folderName.equals(name))
                    fobj.setAttribute("categoryName", name); // NOI18N
            }
            catch (java.io.IOException ex) {
                RuntimeException e = new IllegalArgumentException();
                org.openide.TopManager.getDefault().getErrorManager().annotate(e, ex);
                throw e;
            }
        }
        super.setName(name, false);
    }

    public HelpCtx getHelpCtx() {
        return new HelpCtx("gui.options.component-palette"); // NOI18N
    }

    public Node[] getItemsNodes() {
        return getChildren().getNodes();
    }

    /**
     * Supports index cookie in addition to standard support.
     * @param type the class to look for
     * @return instance of that class or null if this class of cookie is not
     * supported
     */

    public Node.Cookie getCookie(Class type) {
        if (Index.class.isAssignableFrom(type)) {
            // search for data object
            DataFolder dataObj =(DataFolder) super.getCookie(DataFolder.class);
            if (dataObj != null) {
                return new PaletteCategoryIndex(
                    dataObj, (PaletteCategoryNodeChildren) getChildren());
            }
        }
        return super.getCookie(type);
    }

    public Image getIcon(int type) {
        if ((type == BeanInfo.ICON_COLOR_16x16) ||(type == BeanInfo.ICON_MONO_16x16)) {
            if (substIcon == null)
                substIcon = Toolkit.getDefaultToolkit().getImage(
                    getClass().getResource(iconURL));
            return substIcon;
        }
        else {
            if (substIcon32 == null)
                substIcon32 = Toolkit.getDefaultToolkit().getImage(
                    getClass().getResource(icon32URL));
            return substIcon32;
        }
    }

    public Image getOpenedIcon(int type) {
        return getIcon(type);
    }

    /**
     * Support for new types that can be created in this node.
     * @return array of new type operations that are allowed
     */

    public NewType[] getNewTypes() {
        return new NewType[0];
    }

    /** Actions.
     * @return array of actions for this node
     */

    public SystemAction[] getActions() {
        if (staticActions == null)
            staticActions = new SystemAction [] {
                SystemAction.get(FileSystemAction.class),
                null,
                SystemAction.get(MoveUpAction.class),
                SystemAction.get(MoveDownAction.class),
                SystemAction.get(ReorderAction.class),
                null,
                SystemAction.get(PasteAction.class),
                null,
                SystemAction.get(DeleteAction.class),
                SystemAction.get(RenameAction.class),
                //null,
                //XXX SystemAction.get(NewAction.class),
                //XXX null,
                //XXX(-tdt) SystemAction.get(ToolsAction.class),
                //SystemAction.get(PropertiesAction.class),
            };

        return staticActions;
    }

    /** Creates properties for this node */

    public Node.PropertySet[] getPropertySets() {
        return NO_PROPERTIES;
    }

    protected void createPasteTypes(Transferable t, java.util.List s) {
        Node[] nodes;
        boolean cut;

        nodes = NodeTransfer.nodes(t, NodeTransfer.MOVE);
        if (nodes == null) {
            cut = false;
            nodes = NodeTransfer.nodes(t, NodeTransfer.CLIPBOARD_COPY);
            if (nodes == null) return;
        }
        else cut = true;

        LinkedList ll = new LinkedList();
        for (int i=0; i < nodes.length; i++) {
            Node node = nodes[i];

            DataObject dobj = (DataObject)node.getCookie(DataObject.class);
            InstanceCookie ic = (InstanceCookie)
                                node.getCookie(InstanceCookie.class);
            FileObject fo = dobj != null ? dobj.getPrimaryFile() : null;

            if (getChildren().findChild(node.getName()) == null
                    && dobj != null && dobj.isValid() && ic != null
                    && (!cut
                        || "instance".equals(dobj.getPrimaryFile().getExt()) // NOI18N
                        || dobj instanceof DataShadow))
                ll.add(dobj);
        }

        if (!ll.isEmpty())
            s.add(new CatPaste(ll, cut));
    }

    /** Checks category name if it is valid and if there's already not
     * a category with the same name.
     * @param name name to be checked
     * @param namedNode node which name is checked or null if it doesn't exist yet
     * @return true if the name is OK
     */
    static boolean checkCategoryName(String name, Node namedNode) {
        boolean invalid = false;
        if (name == null || "".equals(name))
            invalid = true;
        else // name should not start with . or contain only spaces
            for (int i=0, n=name.length(); i < n; i++) {
                char ch = name.charAt(i);
                if (ch == '.' || (ch == ' ' && i+1 == n)) {
                    invalid = true;
                    break;
                }
                else if (ch != ' ')
                    break;
            }

        if (invalid) {
            TopManager.getDefault().notify(
                new NotifyDescriptor.Message(MessageFormat.format(
                      ComponentPalette.getBundle().getString("ERR_InvalidName"),
                                         new Object[] { name }),
                      NotifyDescriptor.ERROR_MESSAGE));
            return false;
        }

        Node[] nodes = PaletteNode.getPaletteNode().getChildren().getNodes();
        for (int i=0; i < nodes.length; i++)
            if (name.equals(nodes[i].getName()) && nodes[i] != namedNode) {
                TopManager.getDefault().notify(
                    new NotifyDescriptor.Message(MessageFormat.format(
                          ComponentPalette.getBundle().getString("FMT_CategoryExists"),
                                             new Object[] { name }),
                          NotifyDescriptor.INFORMATION_MESSAGE));
                return false;
            }

        return true;
    }

    /** Converts category name to name that can be used as name of folder
     * for the category (restricted even to package name).
     */ 
    static String convertToFolderName(String name, String currentName) {
        if (name == null || "".equals(name)) return null;

        int i;
        int n = name.length();
        StringBuffer nameBuff = new StringBuffer(n);

        char ch = name.charAt(0);
        if (Character.isJavaIdentifierStart(ch)) {
            nameBuff.append(ch);
            i = 1;
        }
        else {
            nameBuff.append('_');
            i = 0;
        }

        while (i < n) {
            ch = name.charAt(i);
            if (Character.isJavaIdentifierPart(ch))
                nameBuff.append(ch);
            i++;
        }

        String fName = nameBuff.toString();
        if ("_".equals(fName))
            fName = "Category"; // NOI18N
        if (fName.equals(currentName))
            return fName;

        FileObject paletteFO = PaletteNode.getPaletteFolder().getPrimaryFile();
        String freeName = null;
        boolean nameOK = false;

        for (i=0; !nameOK; i++) {
            freeName = i > 0 ? fName + "_" + i : fName;

            if (Utilities.isWindows()) {
                nameOK = true;
                Enumeration en = paletteFO.getChildren(false);
                while (en.hasMoreElements()) {
                    FileObject fo = (FileObject)en.nextElement();
                    String fn = fo.getName();
                    String fe = fo.getExt();

                    // case-insensitive on Windows
                    if ((fe == null || "".equals(fe))
                            && fn.equalsIgnoreCase(freeName)) {
                        nameOK = false;
                        break;
                    }
                }
            }
            else nameOK = paletteFO.getFileObject(freeName) == null;
        }
        return freeName;
    }

    /**
     * Children for the PaletteCategoryNode. Creates PaletteCategoryNodes as
     * filter subnodes...
     */

    private static class PaletteCategoryNodeChildren extends FilterNode.Children
    {
        /** @param original the original node to take children from */
        
        public PaletteCategoryNodeChildren(DataFolder folder) {
            super(folder.getNodeDelegate());
        }

        /**
         * Overriden, returns PaletteCategoryNode filters for folders and copies
         * of other nodes.
         * @param node node to create copy of
         * @return PaletteCategoryNode filter of the original node or Node's clone
         * if it is not a DataFolder
         */

        protected Node copyNode(Node node) {
            if (node.getCookie(DataObject.class) != null
                && node.getCookie(InstanceCookie.class) != null) {
                return new PaletteItemNode(node);
            }
            return node.cloneNode();
        }
    }

    /**
     * This class serves as index cookie implementation for the
     * PaletteCategoryNode object. Allows reordering of palette items.
     */

    static final class PaletteCategoryIndex extends DataFolder.Index
    {
        /** The children we are working with */
        PaletteCategoryNodeChildren children;

        PaletteCategoryIndex(DataFolder df, PaletteCategoryNodeChildren children) {
            super(df);
            this.children = children;
        }

        /**
         * Overrides DataFolder.Index.getNodesCount().
         * Returns count of the nodes from the asociated chidren.
         */
        public int getNodesCount() {
            return children.getNodesCount();
        }

        /**
         * Overrides DataFolder.Index.getNodes().
         * Returns array of subnodes from asociated children.
         * @return array of subnodes
         */
        public Node[] getNodes() {
            return children.getNodes();
        }
    }

   private class CatPaste extends PasteType {
        private List dataObjs;
        private boolean cut;

        public CatPaste(List dataObjs, boolean cut) {
            this.dataObjs = dataObjs;
            this.cut = cut;
        }

        public HelpCtx getHelpCtx() {
            return new HelpCtx(CatPaste.class.getName());
        }

        /** Paste.
        */
        public final Transferable paste() throws java.io.IOException {
            DataFolder df = PaletteCategoryNode.this.folder;
            Iterator it = dataObjs.iterator();
            while (it.hasNext()) {
                DataObject dobj = (DataObject)it.next();
                FileObject fo = dobj.getPrimaryFile();

                if ("instance".equals(fo.getExt()) // NOI18
                        || dobj instanceof DataShadow) {
                    // instance or shadow object (can be copied/moved directly)
                    Object iconAttr = fo.getAttribute("SystemFileSystem.icon"); // NOI18N
                    Object beaninfoAttr = fo.getAttribute("beaninfo"); // NOI18N

                    if (cut) {
                        if (dobj.isMoveAllowed())
                            dobj.move(df);
                        else dobj = null;
                    }
                    else dobj = dobj.copy(df);

                    if (dobj != null) {
                        if (iconAttr != null)
                            dobj.getPrimaryFile().setAttribute(
                                    "SystemFileSystem.icon", iconAttr); // NOI18N
                        if (beaninfoAttr != null)
                            dobj.getPrimaryFile().setAttribute(
                                    "beaninfo", beaninfoAttr); // NOI18N
                    }
                }
                else if (!cut) { // other objects can't be moved
                    if ("java".equals(fo.getExt()) // NOI18N
                            || "class".equals(fo.getExt())) { // NOI18N
                        InstanceCookie ic = (InstanceCookie)
                                            dobj.getCookie(InstanceCookie.class);
                        if (ic != null)
                            BeanInstaller.createInstance(df.getPrimaryFile(),
                                                         ic.instanceName());
                    }
                    else dobj.createShadow(df);
                }
            }
            return null; // clear clipboard
        }
    }
}
