/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.i18n;


import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.JMenuItem;

import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.TopManager;
import org.openide.util.actions.Presenter;
import org.openide.util.RequestProcessor;
import org.openide.windows.TopComponent;


/** 
 * I18n group action which is presnet in nodes Tools popup menu if the nodes 
 * contains some "i18n-izable" data objects. 
 *
 * @author Peter Zavadsky
 */
public class I18nGroupPopupAction extends I18nGroupAction implements Presenter.Popup {

    /** Generated serial version UID. */
    static final long serailVersionUID = -2577653086544960835L;
    
    /** Property name of listener for this action. */
    private static final String PROP_LISTENER = "listener"; // NOI18N
    
    
    /** Initialize the listener. */
    protected void initialize () {
        super.initialize ();
        
        NodesListener listener = new NodesListener(this);
        putProperty(PROP_LISTENER, listener);
    }
    
    /** Overrides superclass method. Test for enablement based on {@link #enable}.
     * @return <code>true</code> to enable */
    public boolean isEnabled() {
        NodesListener listener = (NodesListener)getProperty(PROP_LISTENER);
        listener.checkEnabled(this);
        
        return super.isEnabled();
    }
    
    /** Indicates if action becomes enabled based on activated nodes. Implements superclass abstract method. 
     * @return <code>true</code> if one of activate nodes is data object which is able to i18n-ize
     * or is data folder (it's not scaned inside -> too costly operation), it's supposed it can contain such data objects.
     * @see FactoryRegistry#hasFactory */
    protected boolean enable(Node[] activatedNodes) {
        if(activatedNodes == null && activatedNodes.length == 0)
            return false;

        for(int i = 0; i < activatedNodes.length; i++) {
            DataObject dataObject = (DataObject)activatedNodes[i].getCookie(DataObject.class);

            if(dataObject == null)
                continue;

            if(dataObject instanceof DataFolder) {
                return true;
            } else if(FactoryRegistry.hasFactory(dataObject.getClass().getName()))
                return true;
        }

        return false;
    }

    /** Implements <code>Presenter.Popup</code> interface. */
    public JMenuItem getPopupPresenter() {
        return createLazyPopup(false);
    }
    
    
    /** Node listener to check whether the action is enabled or not */
    private static final class NodesListener
        implements PropertyChangeListener, Runnable {
            
        /** The class we work with. */
        private I18nGroupPopupAction action;
        
        /** Task to delay a while updating of action state. */
        private RequestProcessor.Task task;

        /** Delay before updating */
        private static final int DELAY = 150;

        
        /** Constructor that checks the current state. */
        public NodesListener(I18nGroupPopupAction action) {
            this.action = action;
            
            task = RequestProcessor.createRequest(this);
            task.setPriority(Thread.MAX_PRIORITY - 1);
        }

        /** Property change listener. */
        public void propertyChange (PropertyChangeEvent ev) {
            if(TopComponent.Registry.PROP_ACTIVATED_NODES.equals(ev.getPropertyName ())) {
                task.schedule(DELAY);
                return;
            }
        }

        /** Updates the state of the action. */
        public void run() {
            checkEnabled(action);
        }

        /** Getter for activated nodes
         * @return array of activated nodes
         */
        public Node[] getActivatedNodes() {
            return TopManager.getDefault().getWindowManager().getRegistry().getActivatedNodes();
        }

        /** Checks the state of the action. */
        private void checkEnabled(I18nGroupPopupAction action) {
            // no action no work
            if (action == null) 
                return;

            Node[] nodes = getActivatedNodes();
            boolean enable = nodes != null && action.enable(nodes);
            action.setEnabled(enable);
        }
    } // End of NodeListener class.
    
}


