/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.i18n;


import java.io.IOException;
import java.util.Locale;
import java.util.Map;

import org.openide.execution.ExecInfo;
import org.openide.execution.NbClassPath;
import org.openide.execution.NbProcessDescriptor;
import org.openide.execution.ProcessExecutor;
import org.openide.util.NbBundle;


/**
 * External executor with convenient setting of locale to run app in.
 * It sets arguments user.language and user.region acordingly.
 *
 * @author  Peter Zavadsky
 */
public class LocaleExecutor extends ProcessExecutor {

    /** Locale which to set. */
    private Locale locale;
    
    /** Generated serial version UID. */
    static final long serialVersionUID = -8606172928148451403L;
    

    /** Creates new <code>LocaleExecutor</code>. */
    public LocaleExecutor() {
    }


    /** Setter for <code>locale</code> property. */
    public synchronized void setLocale(Locale newLocale) {
        if(locale == newLocale || (locale != null && locale.equals(newLocale)))
            return;
        
        Locale oldLocale = locale;
        locale = newLocale;
        firePropertyChange("locale", oldLocale, newLocale); // NOI18N
    }
    
    /** Getter for locale property. */
    public Locale getLocale() {
        if(locale == null)
            setLocale(Locale.getDefault());
        
        return locale;
    }

    /** Gets display name. Overrides superclass method. */
    protected String displayName() {
        return NbBundle.getBundle(LocaleExecutor.class).getString("LBL_LocaleExecutorDisplayName");
    }

    /** Called to create the <code>Process</code> for given exec info. Overrides superclass method. */
    protected Process createProcess (ExecInfo info) throws IOException {
        return getExternalExecutor().exec(
            new LocaleFormat(
                info,
                getClassPath(),
                getBootClassPath(),
                getRepositoryPath(),
                getLibraryPath(),
                getLocale()
            ),
            getEnvironmentVariables(),
            getWorkingDirectory()
        );
    }
    
    /** Gets the current external execution command. Overrides superclass method. */
    public NbProcessDescriptor getExternalExecutor() {
        if (externalExecutor == null)
            synchronized(this) {
                if (externalExecutor == null)
                    externalExecutor =  new NbProcessDescriptor (
                        "{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}bin{" + Format.TAG_SEPARATOR + "}java", // /usr/local/bin/java // NOI18N
                        "-cp {" + Format.TAG_REPOSITORY + "}" +  // -cp {REPOSITORY}:{CLASSPATH} {CLASSNAME} {ARGUMENTS} // NOI18N
                        "{" + Format.TAG_PATHSEPARATOR + "}" + "{" + Format.TAG_CLASSPATH + "}" + // NOI18N
                        "{" + Format.TAG_PATHSEPARATOR + "}" + "{" + Format.TAG_LIBRARY + "} " + // NOI18N
                        "{" + LocaleFormat.TAG_USERLANGUAGE + "} " + "{"+LocaleFormat.TAG_USERREGION + "} " + // NOI18N
                        "{" + Format.TAG_CLASSNAME + "} " + // NOI18N
                        "{" + Format.TAG_ARGUMENTS + '}', // NOI18N
                        NbBundle.getBundle(LocaleExecutor.class).getString ("MSG_LocaleExecutorHint")
                    );
            }
            
        return externalExecutor;
    }

    
    /** Format for used by locale executor. Adds setting user.language and user.region values. */
    public static class LocaleFormat extends ProcessExecutor.Format {
        
        /** Tag replaced with -Duser.language=<language_code>. */
        public static final String TAG_USERLANGUAGE = "userlanguage"; // NOI18N
        
        /** Tag replaced with -Duser.region=<country_code>[_<variant>]. */
        public static final String TAG_USERREGION = "userregion"; // NOI18N
        
        
        /** Creates locale format. */
        public LocaleFormat(
            ExecInfo info,
            NbClassPath classPath,
            NbClassPath bootClassPath,
            NbClassPath repository,
            NbClassPath library,
            Locale locale
        ) {
            super(info, classPath, bootClassPath, repository, library);
            
            Map map = getMap();

            map.put(TAG_USERLANGUAGE, "-Duser.language="+locale.getLanguage()); // NOI18N
            
            StringBuffer regionCode = new StringBuffer();
            
            String countryCode = locale.getCountry();
            
            if(countryCode != null && !"".equals(countryCode)) // NOI18N
                regionCode.append(countryCode);
            
            String variant = locale.getVariant();
            
            if(variant != null && !"".equals(variant)) 
                regionCode.append("_"+variant);
            
            map.put(TAG_USERREGION, "-Duser.region="+regionCode.toString()); // NOI18N
        }

    } // End of nested LocalFormat classs.
}
