/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.i18n;


import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.util.Arrays;
import java.util.Enumeration;
import javax.swing.border.TitledBorder;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.loaders.DataFilter;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.nodes.NodeAcceptor;
import org.openide.TopManager;
import org.openide.util.UserCancelException;


/**
 * Customizer for <code>ResourceHolder</code> object which is encapsulated by <code>I18nString</code> instance.
 * It's used inside <code>I18nPanel</code>.
 *
 * @author  Peter Zavadsky
 */
public class ResourcePanel extends JPanel {

    /** Name for resource property. */
    public static final String PROP_RESOURCE = "property_resource";
    
    /** <code>I18nString</code> which resource is cutomized. */
    private I18nString i18nString;
    
    
    /** Creates new form ResourceHolderCustomizer.
     * @param i18nString <code>i18nString</code> which <code>resource</code> will be customized. */
    public ResourcePanel() {
        initComponents();
        
    }
    
    
    /** Setter for <code>i18nString</code> property. */
    public void setI18nString(I18nString i18nString) {        
        this.i18nString = i18nString;
        
        updateValues();
    }

    /** Helper method. Reflects values changes to UI. */
    private void updateValues() {
        // Set resource texfield.
        resourceText.setText(getResourceName(i18nString.getSupport().getResourceHolder().getResource()));
    }

    /** Helper method. Changes resource. */
    private void changeResource(DataObject resource) {
        if(resource == null)
            throw new IllegalArgumentException();

        DataObject oldValue = i18nString.getSupport().getResourceHolder().getResource();
        
        if(oldValue != null && oldValue.equals(resource))
            return;
        
        i18nString.getSupport().getResourceHolder().setResource(resource);
        updateValues();
        
        firePropertyChange(PROP_RESOURCE, oldValue, resource);
        
        I18nUtil.getOptions().setLastResource(resource);
    }

    /** Handler delegate. */
    private void resourceTextActionPerformedDelegate(ActionEvent evt) {
        DataObject resource = findResource(resourceText.getText());
        if(resource != null)
            changeResource(resource);
        else
            resourceText.setText(getResourceName(i18nString.getSupport().getResourceHolder().getResource()));
    }

    /** Helper method. Finds <code>DataObject</code> (resource bundle) for specified name. */
    private DataObject findResource(String bundleName) {
        try {
            // It has to provide package name separated by slashes (not by dots).
            bundleName = bundleName.replace('.', '/');
            
            String resourceName = bundleName + ".properties"; // NOI18N // PENDING

            FileObject fileObject = null;
            
            for(Enumeration en = TopManager.getDefault().getRepository().getFileSystems(); en.hasMoreElements();) {
                fileObject = ((FileSystem)en.nextElement()).findResource(resourceName);
                if (fileObject != null)
                    break;
            }
            
            if(fileObject == null)
                return null;
            
            DataObject dataObject = TopManager.getDefault().getLoaderPool().findDataObject(fileObject);

            // Check for validity of found data object for this resource holder.
            if(isResourceClass(dataObject.getClass()))
                return dataObject;
            else
                return null;
        } catch(IOException e) {
            return null;
        }
    }

    /** Helper method. Checks if specified class is valid resource class. */
    private boolean isResourceClass(Class clazz) {
        return Arrays.asList(i18nString.getSupport().getResourceHolder().getResourceClasses()).contains(clazz);
    }
    
    /** Helper method. Gets name of specified <code>DataObject</code> (resource bundle) . */
    private String getResourceName(DataObject resource) {
        if(resource == null)
            return ""; // NOI18N
        else
            return resource.getPrimaryFile().getPackageName('.');
    }

    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        resourceLabel = new javax.swing.JLabel();
        resourceText = new javax.swing.JTextField();
        jPanel1 = new javax.swing.JPanel();
        browseButton = new javax.swing.JButton();
        newButton = new javax.swing.JButton();
        
        setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints1;
        
        resourceLabel.setText(I18nUtil.getBundle().getString("LBL_BundleName"));
        resourceLabel.setLabelFor(resourceText);
        // setBorder(new TitledBorder(I18nUtil.getBundle().getString("LBL_ResourceTitleBorder")));
        resourceLabel.setDisplayedMnemonic((I18nUtil.getBundle().getString("LBL_BundleName_Mnem")).charAt(0));
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.insets = new java.awt.Insets(12, 12, 0, 0);
        gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
        add(resourceLabel, gridBagConstraints1);
        
        resourceText.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resourceTextActionPerformed(evt);
            }
        });
        
        resourceText.addFocusListener(new java.awt.event.FocusAdapter() {
            public void focusLost(java.awt.event.FocusEvent evt) {
                resourceTextFocusLost(evt);
            }
        });
        
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridwidth = 2;
        gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints1.insets = new java.awt.Insets(12, 12, 0, 11);
        gridBagConstraints1.weightx = 1.0;
        add(resourceText, gridBagConstraints1);
        
        jPanel1.setLayout(new java.awt.GridLayout(1, 2, 5, 0));
        
        browseButton.setMnemonic((I18nUtil.getBundle().getString("CTL_BrowseButton_Mnem")).charAt(0));
        browseButton.setText(I18nUtil.getBundle().getString("CTL_BrowseButton"));
        browseButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                browseButtonActionPerformed(evt);
            }
        });
        
        jPanel1.add(browseButton);
        
        newButton.setMnemonic((I18nUtil.getBundle().getString("CTL_NewButton_Mnem")).charAt(0));
        newButton.setText(I18nUtil.getBundle().getString("CTL_NewButton"));
        newButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                newButtonActionPerformed(evt);
            }
        });
        
        jPanel1.add(newButton);
        
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 1;
        gridBagConstraints1.gridy = 1;
        gridBagConstraints1.gridwidth = 2;
        gridBagConstraints1.insets = new java.awt.Insets(11, 0, 11, 11);
        gridBagConstraints1.anchor = java.awt.GridBagConstraints.EAST;
        add(jPanel1, gridBagConstraints1);
        
    }//GEN-END:initComponents

    private void newButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_newButtonActionPerformed
        try {
            // PENDING at the moment create from first class.
            // It should be possible for user to choose which kind of class.
            DataObject template = i18nString.getSupport().getResourceHolder().getTemplate(i18nString.getSupport().getResourceHolder().getResourceClasses()[0]);
          
            DataObject resource = instantiateTemplate(template);
            
            if(resource != null) {
                changeResource(resource);
            }
            
        } catch (UserCancelException uce) {
            if(Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                System.err.println("I18N module: User cancelled selection"); // NOI18N
        } catch (IOException ioe) {
            TopManager.getDefault().notifyException(ioe);
        }
    }//GEN-LAST:event_newButtonActionPerformed

    private void browseButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_browseButtonActionPerformed
        DataFilter dataFilter = new DataFilter() {
            public boolean acceptDataObject (DataObject dataObject) {
                return (dataObject instanceof DataFolder || isResourceClass(dataObject.getClass()));
            }
        };
      
        Node repositoryNode = TopManager.getDefault().getPlaces().nodes().repository(dataFilter);
      
        // Selects one (resource) data object.
        try {
            DataObject resource = (DataObject)TopManager.getDefault().getNodeOperation().select(
                I18nUtil.getBundle().getString ("CTL_SelectPropDO_Dialog_Title"),
                I18nUtil.getBundle().getString ("CTL_SelectPropDO_Dialog_RootTitle"),
                repositoryNode, 
                new NodeAcceptor() {
                    public boolean acceptNodes(Node[] nodes) {
                        if(nodes == null || nodes.length != 1) {
                            return false;
                        }
                      
                        Node.Cookie cookie = nodes[0].getCookie(DataObject.class);
                        if(cookie == null)
                            return false;
                      
                        return isResourceClass(cookie.getClass());
                    }
            })[0].getCookie(DataObject.class);

            if(resource != null) {
                changeResource(resource);
            }

        } catch (UserCancelException uce) {
            if(Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                System.err.println("I18N module: User cancelled selection"); // NOI18N
        }

    }//GEN-LAST:event_browseButtonActionPerformed

    private void resourceTextFocusLost(java.awt.event.FocusEvent evt) {//GEN-FIRST:event_resourceTextFocusLost
        resourceTextActionPerformedDelegate(new ActionEvent(evt.getSource(), evt.getID(), null));
    }//GEN-LAST:event_resourceTextFocusLost

    private void resourceTextActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_resourceTextActionPerformed
        resourceTextActionPerformedDelegate(evt);
    }//GEN-LAST:event_resourceTextActionPerformed

    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel resourceLabel;
    private javax.swing.JTextField resourceText;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JButton browseButton;
    private javax.swing.JButton newButton;
    // End of variables declaration//GEN-END:variables

    /** Instantiate a template object.
     * Asks user for the target file's folder and creates the file.
     * @param obj the template to use
     * @return the generated DataObject
     * @exception UserCancelException if the user cancels the action
     * @exception IOException on I/O error
     * @see DataObject#createFromTemplate
     */
    public static DataObject instantiateTemplate(DataObject template) throws IOException {
        // Create component for for file name input.
        ObjectNameInputPanel panel = new ObjectNameInputPanel();
        
        DataFilter dataFilter = new DataFilter() {
            public boolean acceptDataObject(DataObject dataObject) {
                return dataObject instanceof DataFolder;
            }
        };
        Node repositoryNode = TopManager.getDefault().getPlaces().nodes().repository(dataFilter);
        
        // Selects one folder from data systems.
        DataFolder dataFolder = (DataFolder)TopManager.getDefault().getNodeOperation().select(
            I18nUtil.getBundle().getString ("CTL_Template_Dialog_Title"),
            I18nUtil.getBundle().getString ("CTL_Template_Dialog_RootTitle"),
            repositoryNode,
            new NodeAcceptor() {
                public boolean acceptNodes(Node[] nodes) {
                    if(nodes == null || nodes.length != 1) {
                        return false;
                    }
                    
                    DataFolder cookie = (DataFolder)nodes[0].getCookie(DataFolder.class);
                    return (cookie != null && !cookie.getPrimaryFile().isReadOnly());
                }
            },
            panel)[0].getCookie(DataFolder.class);
        
        String name = panel.getText();
        
        DataObject newObject;
        
        if(name.equals ("")) { // NOI18N
            newObject = template.createFromTemplate(dataFolder);
        } else {
            newObject = template.createFromTemplate(dataFolder, name);
        }
        
        try {
            return newObject;
        } catch(ClassCastException cce) {
            throw new UserCancelException();
        }
    }

    /** Panel used by <code>instantiateTemplate</code> method. */
    private static class ObjectNameInputPanel extends JPanel {
        
        /** Generated Serialized Version UID. */
        static final long serialVersionUID = 1980214734060402958L;

        /** Text field. */
        JTextField text;

        
        /** Constructs panel. */
        public ObjectNameInputPanel () {
            BorderLayout layout = new BorderLayout();
            layout.setVgap(5);
            layout.setHgap(5);
            setLayout(layout);
            
            // label and text field with mnemonic
            String labelText = I18nUtil.getBundle().getString ("LBL_TemplateName");
            JLabel label = new JLabel(labelText.replace('&', ' '));
            text = new JTextField();
            
            label.setDisplayedMnemonic(labelText.charAt(labelText.indexOf('&') + 1));
            label.setLabelFor(text);
            
            add(BorderLayout.WEST, label);
            add(BorderLayout.CENTER, text);
        }

        
        /** Requets focus for text field. */
        public void requestFocus () {
            text.requestFocus ();
        }
        
        /** Getter for <code>text</code>. */
        public String getText () {
            return text.getText ();
        }

        /** Setter for <code>text</code>. */
        public void setText (String s) {
            setText(s);
        }
    } // End of nested class ObjectNameInputPanel
    
}
