/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.icebrowser;

import java.awt.Color;
import java.awt.Font;
import java.beans.*;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.awt.HtmlBrowser;
import ice.storm.Viewport;

public class IceBrowser implements HtmlBrowser.Factory, java.io.Serializable {
    
    private static final long serialVersionUID = 3477225220592115953L;
    
    private static ICEBrowserSettings settings;
    
    public static final String PROP_NAME = "BrowserName";   // NOI18N
    
    /** helper variable */
    private transient PropertyChangeSupport pcs;
    
    /** name of browser */
    protected String name;
    
    /** default background color */
    private Color defaultBackground;
    
    /** fixed font property value */
    private Font fixedFont;
    
    /** proportional font property value */
    private Font proportionalFont;
    
    /** encoding property value */
    private String encoding;
    
    /** animation on/off property value */
    private boolean animation = false;
    
    /** Creates new IceBrowser */
    public IceBrowser () {
        init ();
    }
    
    /** Initializes this object. 
     * Delegation to ICEBrowserSettings is set up there.
     */
    private void init () {
        pcs = new PropertyChangeSupport (this);
        if (settings == null) {
            settings = (ICEBrowserSettings)ICEBrowserSettings.findObject (ICEBrowserSettings.class, true);
        }
        settings.addPropertyChangeListener (new PropertyChangeListener () {
            public void propertyChange (PropertyChangeEvent evt) {
                IceBrowser.this.pcs.firePropertyChange (evt.getPropertyName (), evt.getOldValue (), evt.getNewValue ());
            }
        });
    }
    
    /** Getter for property name. If name is not set then default name of bean is returned.
     *
     * @return Value of property name.
     */
    public java.lang.String getName () {
        if (name == null) {
            try {
                return Introspector.getBeanInfo (getClass ()).getBeanDescriptor ().getDisplayName ();
            } catch (Exception e) {
                // Catching IntrospectionException, but also maybe NullPointerException...?
                TopManager.getDefault ().getErrorManager ().notify (
                    ErrorManager.INFORMATIONAL,
                    e
                );
            }
        }
        return name;
    }    
    
    /** helper variable for creating new Viewport */
    IceBrowserImpl impl = null;
    
    /**
     * Factory implementation
     *
     * @return Impl a new instance of BrowserImpl implementation.
     */
    public HtmlBrowser.Impl createHtmlBrowserImpl () {
        IceBrowserImpl result = (impl == null)? new IceBrowserImpl (this): impl;
        impl  = null;
        return result;
    }
    
    public IceBrowserImpl prebuildHtmlBrowserImpl (Viewport v) {
        impl = new IceBrowserImpl (this, v);
        return impl;
    }
    
    // Properties delegated to ICEBrowserSettings
    
    
    /**
    * @returns default color of browser background.
    */
    public java.awt.Color getDefaultBackground () {
        if (defaultBackground != null)
            return defaultBackground;
        else
            return settings.getDefaultBackground ();
    }

    /**
    * Set default color of browser background.
    */
    public void setDefaultBackground (java.awt.Color color) {
        Color old = defaultBackground;
        defaultBackground = color;
        pcs.firePropertyChange (ICEBrowserSettings.PROP_DEFAULT_BACKGROUND, old, defaultBackground);
    }

    /**
    * @returns default fixed font for HTML browser.
    */
    public Font getFixedFont () {
        if (fixedFont != null)
            return fixedFont;
        else 
            return settings.getFixedFont ();
    }

    /**
    * Set default fixed font for HTML browser.
    */
    public void setFixedFont (Font font) {
        Font old = fixedFont;
        fixedFont = font;
        pcs.firePropertyChange (ICEBrowserSettings.PROP_FIXED_FONT, old, fixedFont);
    }

    /**
    * @returns default proportional font for HTML browser.
    */
    public Font getProportionalFont () {
        if (proportionalFont != null)
            return proportionalFont;
        else 
            return settings.getProportionalFont ();
    }

    /**
    * Set default proportional font for HTML browser.
    */
    public void setProportionalFont (Font font) {
        Font old = proportionalFont;
        proportionalFont = font;
        pcs.firePropertyChange (ICEBrowserSettings.PROP_PROPORTIONAL_FONT, old, proportionalFont);
    }

    /**
    * @returns encoding for browser.
    */
    public String getEncoding () {
        if (encoding != null)
            return encoding;
        else
            return settings.getEncoding ();
    }

    /**
    * Set encoding for browser.
    */
    public void setEncoding (String encoding) {
        String old = this.encoding;
        this.encoding = encoding;
        pcs.firePropertyChange (ICEBrowserSettings.PROP_ENCODING, old, encoding);
    }
    
    /**
    * @returns true if animation is turned on.
    */
    public boolean isAnimation () {
        return animation;
    }

    /**
    * Set encoing for browser.
    */
    public void setAnimation (boolean animation) {
        boolean old = this.animation;
        this.animation = animation;
        pcs.firePropertyChange (ICEBrowserSettings.PROP_ANIMATION, old, animation);
    }
    
    // OTHER METHODS
    
    /**
     * @param l new PropertyChangeListener */    
    public void addPropertyChangeListener (PropertyChangeListener l) {
        pcs.addPropertyChangeListener (l);
    }
    
    /**
     * @param l PropertyChangeListener to be removed */    
    public void removePropertyChangeListener (PropertyChangeListener l) {
        pcs.removePropertyChangeListener (l);
    }
    
    private void readObject (java.io.ObjectInputStream ois) 
    throws java.io.IOException, ClassNotFoundException {
        ois.defaultReadObject ();
        init ();
    }
}
