
/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.jarpackager;

import java.io.*;
import java.text.*;
import java.util.*;
import java.util.jar.*;
import org.netbeans.modules.jarpackager.api.*;
import org.netbeans.modules.jarpackager.util.JarUtils;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.DataFolder;
import org.openide.filesystems.FileObject;
import org.openide.util.NbBundle;


/** The persistent list of data objects which will be written to a jar file.
*
* @author Mike Schilling
*/
public class ContentMemberList extends ArrayList implements Serializable {

    static final long serialVersionUID = 7194592824986427208L;

    public ContentMemberList() {
        super();
    }

    public ContentMemberList(Collection collection) {
        super(collection);
    }

    public ContentMemberList(int capacity) {
        super(capacity);
    }

    /** Convert a list of DataObjects to ContentMembers.  This is used when
    * deserializing an older jar contents file.  */
    public ContentMemberList(ArrayList list) {
	//System.out.println("Array list contains " + list.size() + " members");
    	for (int i = 0; i < list.size(); i++) {
	    Object o = list.get(i);
	    if (o instanceof ContentMember) {
		//System.out.println("Adding " + ((ContentMember)o).dataObject.getName());
	    	this.add(o);
	    }
	    else if (o instanceof DataObject) {
		addMember((DataObject)o);
	    }
	    else if (o instanceof FileObject) {
	    	FileObject fo = (FileObject)o;
		try {
		    DataObject dobj = DataObject.find(fo);
		    addMember(dobj);
		}
		catch (DataObjectNotFoundException ex) {
		    String fmt = NbBundle.getBundle(ContentMemberList.class).
					getString("FMT_ContentFileNotFound");
		    System.out.println(
		    	MessageFormat.format(fmt, new Object[] {fo} ));
		}
	    }
	    else {
		String fmt = NbBundle.getBundle(ContentMemberList.class).
				    getString("FMT_ContentBadClass`");
		String className = o.getClass().getName();
		System.out.println(
		    MessageFormat.format(fmt, new Object[] {className} ));
		
	    }
	}
    }
    	
   /** add a new member */
   public boolean addMember(DataObject o) {
       if (findDataObject(o) == -1) {
  	    	ContentMember member = new ContentMember();
		member.dataObject = (DataObject)o;
		member.isRecursive = true;
		this.add(member);
                return true;
       }  
       return false;
    }

   /** add a new member */
    public boolean addMember(DataObject o, String target) {
       if (findDataObject(o) == -1) {
  	    	ContentMember member = new ContentMember();
		member.dataObject = (DataObject)o;
		member.targetDirectory = target;
		member.isRecursive = true;
		this.add(member);
                return true;
       }  
       return false;
    }
   	
    /** Add a list of members */
    public void addList(ArrayList list) {
    	for (int i = 0; i < list.size(); i++) {
	    Object o =  list.get(i);
	    if (o instanceof DataObject) {
		addMember((DataObject) o);
	    }
	    else if (o instanceof ContentMember) {
	    	this.add(o);
	    }
	}
    }
    
   /** remove one member */
   public boolean removeMember(DataObject o) {
       int result = findDataObject(o);
       if (result != -1) {
 		this.remove(result);
                return true;
       }  
       return false;
    }
    
    /** remove one member */
    public boolean removeMember(int index) {
       if (index >= 0 && index < size() ) {
 		this.remove(index);
                return true;
       }  
       return false;
    }

    /** return the index for a specific data object */
    public int findDataObject(DataObject o) {
        for (int i=0; i< size(); i++) { 
            ContentMember member = (ContentMember) get(i);
            if (member.getDataObject().equals(o))
                return i;
          }
        return -1;
    }

    /** return all files which are part of this List */
    public List getAllFiles(ArchiveFilter globalFilter) {
    	List files = new ArrayList();
	for (int i = 0; i < this.size(); i++) {
	    ContentMember member = (ContentMember)this.get(i);
	    addFiles(files, member.getDataObject(), member.isRecursive,
		globalFilter, member.getFilter());
	}
	return files;
    }

    /** add the files from one data object to a list */
    private void addFiles(
      List list, DataObject dataObj, boolean recursive,
      ArchiveFilter global, ArchiveFilter local) {

	if (dataObj instanceof DataFolder) {
	    Enumeration enum = ((DataFolder)dataObj).children(false); 
	    while (enum.hasMoreElements()) {
		DataObject curChild = (DataObject)enum.nextElement();
		if (recursive || !(curChild instanceof DataFolder)) {
		    addFiles(list, dataObj, recursive, global, local);
		}
	    }
	}

	Iterator fileIter = dataObj.files().iterator();	
	while (fileIter.hasNext()) {
	    FileObject fo = (FileObject)fileIter.next();
	    if (global != null && !global.accept(fo)) {
	    	fo = null;
	    }
	    if (fo != null && local != null && !local.accept(fo)) {
	    	fo = null;
	    }
	    if (fo != null) {
	    	list.add(fo);
	    }
	}
    }
	    
    /** The list of data objects which will become members of a jar file.
    * This class allows efficient inspection and manipulation of this list.
    */
    static class ArchiveMemberSetImpl implements ArchiveMemberSet {
	/** jar entries */
	private Map entries;

	/** saved copy of entries */
	private ArchiveMember members[] = null;

	/** whether the set of entries has been changed */
	private boolean changed = false;

	/** construct from a ContentMemberList, expanding all folders */
	ArchiveMemberSetImpl(ContentMemberList list) {
	    entries = new HashMap();
	    //System.out.println("Converting " + list.size() + "content members");
	    for (int i = 0; i < list.size(); i++) {
	    	ContentMember origMember = (ContentMember)list.get(i);
		ContentMember member = origMember.resolve();
		if (member.dataObject instanceof DataFolder) {
		    List children = 
		    	expandFolder(member, member.isRecursive);
		    Iterator iter = children.iterator();
		    while (iter.hasNext()) {
			ContentMember child = (ContentMember)iter.next();
			child.setFilter(member.filter);
			addEntry(child);
		    }
		}
		else {
		    //System.out.println("Adding non-folder " + member.dataObject.getName());
		    addEntry(member.dataObject, 
		    	     member.filter, 
		    	     member.targetDirectory);
		}
	    }
	}

	/** construct from DataObject List (temporary) */
	ArchiveMemberSetImpl(DataObjectList dol) {
	    this(new ContentMemberList(dol));
	}

	/** number of members */
	int size() {
	    return entries.size();
	}

	/** set when a client has changed the list */
	void setChanged() {
	    changed = true;
	    members = null;
	}

	/** reset changed indicator */
	void setUnchanged() {
	    changed = false;
	}

	/** did a client change the list? */
	boolean wasChanged() {
	    return changed;
	}

	/* ArchiveMemberSet methods */

	/** return all entries */
	public ArchiveMember[] allMembers() {
	    if (members == null) {
		// members was never set or is no longer valid
		ArchiveMember[] template = new ArchiveMember[entries.size()];
		members = (ArchiveMember[]) entries.values().toArray(template);
	    }
	    return members;
	}

	/** add a member 
	 * @param dataObj the DataObject for the new member.  If there was 
	 * already a member with this DataObject, it will not be added.
	 * @param filter a member-specific filter.  This may be null.
	 * @param traget the target directory in the jar.  If this is null, the
	 * target is determined by the DataObject.
	 * @return the number of members added
	 */
	public int addMember(
	    DataObject dataObj, ArchiveFilter filter, String targetDirectory) {

	    return addMember(dataObj, filter, targetDirectory, false);
	}

	/** add a member 
	 * @param dataObj the DataObject for the new member.  If there was 
	 * already a member with this DataObject, it will not be added.
	 * @param filter a member-specific filter.  This may be null.
	 * @param traget the target directory in the jar.  If this is null, the
	 * target is determined by the DataObject.
	 * @param recursive whether to expand folders recursively
	 * @return the number of members added
	 */
	 private int addMember(
	    DataObject dataObj, ArchiveFilter filter, String targetDirectory,
	    boolean recursive) {

	    int retval = 0;

	    if (dataObj instanceof DataFolder) {
		List children = 
		    expandFolder(
		    	(DataFolder)dataObj, recursive, targetDirectory);
		Iterator iter = children.iterator();
		while (iter.hasNext()) {
		    ContentMember child = (ContentMember)iter.next();
		    child.filter = filter;
		    if (addEntry(child)) {
			retval++;
		    }
		}
	    }
	    else {
		if (addEntry(dataObj, filter, targetDirectory)) {
		    retval++;
		}
	    }
	    return retval;
	}

	/** remove a member from the set
	 * @param dataObj the DataObject for the member to be removed
	 * @return the number of members removed
	 */
	public int removeMember(DataObject dataObj) {
	    int retval = 0;

	    if (dataObj instanceof DataFolder) {
		List children = expandFolder((DataFolder)dataObj, false, null);
		Iterator iter = children.iterator();
		while (iter.hasNext()) {
		    ContentMember child = (ContentMember)iter.next();
		if (entries.remove(child.dataObject) != null) {
			setChanged();
			retval++;
		    }
		}
	    }
	    else {
		if (entries.remove(dataObj) != null) {
		    setChanged();
		    retval++;
		}
	    }
		
	    return retval;
	}

	/** get a member from the set
	 * @param dataObj DataObject of member to get.
	 * @return the member for that DataObject or null.
	 */
	public ArchiveMember getMember(DataObject dataObj) {
	    return (ArchiveMember)entries.get(dataObj);
	}

	/** get the members which are children of a specified folder.
	 * @param dataFolder the folder containing the DataObjects for the
	 * returned members.
	 * @return an array of members (never null)
	 */
	public ArchiveMember[] getMembers(DataFolder dataFolder) {
	    List list = new ArrayList();
	    List children = expandFolder(dataFolder, false, null);
	    Iterator iter = children.iterator();
	    while (iter.hasNext()) {
		ContentMember child = (ContentMember)iter.next();
		Object entry = entries.get(child.dataObject);
		if (entry != null) {
		    list.add(entry);
		}
	    }
	    ArchiveMember retval[] = new ArchiveMember[list.size()];
	    return (ArchiveMember[]) list.toArray(retval);
	}


	/**
	* put a new entry in the map if it isn't already there
	* @return true if the entry was added
	*/
	private boolean addEntry(
	    DataObject dataObj, ArchiveFilter filter, String targetDirectory) {
	    if (entries.get(dataObj) == null) {
		entries.put(dataObj, 
		    new ContentMember(dataObj, filter, targetDirectory));
		setChanged();
		return true;
	    }
	    return false;
	}
		
	/**
	* put a new entry in the map if it isn't already there
	* @return true if the entry was added
	*/
	private boolean addEntry(ContentMember member) {
	    if (entries.get(member.dataObject) == null) {
		entries.put(member.dataObject, member);
		setChanged();
		return true;
	    }
	    return false;
	}
		
	/**
	 * Return all the non-folder children of a folder, optionally
	 * recursively
	 */
	private List expandFolder(ContentMember member, boolean recursive) {
	    return expandFolder(
	    	(DataFolder)member.dataObject, 
		recursive, 
		member.targetDirectory);
	}

	static List expandFolder(
		DataFolder folder, boolean recursive, String targetName) {
	    //System.out.println("Expanding folder " + folder.getName());
	    targetName = JarCreator.getCanonicalTargetName(targetName);
	    List retval = new ArrayList();
	    Enumeration enum = folder.children(false); 
	    while (enum.hasMoreElements()) {
		DataObject curChild = 
			JarUtils.resolve((DataObject)enum.nextElement());
		if (!(curChild instanceof DataFolder)) {
		    //System.out.println("Adding child " + curChild.getName());
		    retval.add(new ContentMember(curChild, null, targetName)); 
		}
		else if (recursive) {
		    //System.out.println("Expanding " + curChild.getName());
		    String childTarget = null;
		    if (ArchiveMember.ROOT.equals(targetName)) {
		    	childTarget = curChild.getName();
		    }
		    else if (targetName != null) {
		    	childTarget = targetName + curChild.getName();
		    }
		    List sublist = 
			expandFolder(
			    (DataFolder)curChild, recursive, childTarget);
		   retval.addAll(sublist);
		}
	    }
	    return retval;
	}
    }
}
