/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager;

import java.util.*;
import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.text.MessageFormat;

import org.openide.filesystems.FileObject;
import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.CompilerGroup;
import org.openide.compiler.ErrorEvent;
import org.openide.compiler.ProgressEvent;
import org.openide.util.NbBundle;
import org.openide.ErrorManager;
import org.openide.TopManager;

import org.netbeans.modules.jarpackager.util.ProgressListener;
import org.netbeans.modules.jarpackager.util.JarUtils;

/** Compiler for the jar archives.
* Compiling of jar archive is actually a repackaging 
* (recreating) the archive to keep the information up to date.
*
* @author Dafe Simonek
*/
final class JarCompiler extends Compiler {

    /** The instance desribing content of the jar */
    JarContent jc;
    
    /** Asociated Jar data object */
    JarDataObject jdo;
    
    /** Creates new JarCompiler */
    public JarCompiler (JarContent jc, JarDataObject jdo) {
        super();
        this.jc = jc;
        this.jdo = jdo;
        dependsOn(JarUtils.createCompilerJob(jc, jdo));        
    }

    /** Check whether the compiler is up to date.
    * Always return false for now...
    *
    * @return false if jar content is null, true otherwise
    */
    protected boolean isUpToDate () {
        return false;
    }

    /** Get the associated <code>CompilerGroup</code> container class.
    * The compiler and compiler group should typically be implemented in parallel.
    * @return a class assignable to {@link CompilerGroup}
    */
    public Class compilerGroupClass () {
        return Group.class;
    }

    /** @return true if two jar compilers have the same file object
    * and jar content attached, false otherwise */
    public boolean equals (Object obj) {
        if (obj instanceof JarCompiler) {
            JarCompiler other = (JarCompiler)obj;
            return super.equals(obj) && (jc == other.jc);
        }
        return false;
    }

    /** @return Hash code for this jar compiler */
    public int hashCode() {
        return jc.hashCode();
    }


    /** Group for jar compilation */
    public static final class Group extends CompilerGroup
        implements ProgressListener {

        /** All jar contents whcih we should compile */
        List contents = new LinkedList();
        /** Worker instance we delegate the work to */
        JarCreator curCreator;

	/** Add a compiler to this group instance */
        public void add (Compiler c) throws IllegalArgumentException {
            // check
            if (!(c instanceof JarCompiler))
                throw new IllegalArgumentException();
            contents.add(c);
        }

        /** Finally compile the jar (update its content) */
        public boolean start () {
            JarCompiler curJc = null;
            curCreator = null;
	    File archiveFile = null;
            try {
                for (Iterator iter = contents.iterator(); iter.hasNext(); ) {
                    curJc = (JarCompiler)iter.next();
                    // now create jar
                    curCreator = new JarCreator(curJc.jc);
                    curCreator.addProgressListener(this);
                    archiveFile = curJc.jc.getTargetFile();
                    FileOutputStream fos = new FileOutputStream(archiveFile);
                    try {
                        curCreator.createJar(fos, archiveFile);
                    } finally {
                        fos.close();
                    }
		    archiveFile = null;
                    curCreator.removeProgressListener(this);
                }
            } catch (IOException exc) {
                if (curCreator != null) {
                    curCreator.removeProgressListener(this);
                }
		if (archiveFile != null && archiveFile.exists()) {
		    // Remove possibly partially created file
		    try {
			archiveFile.delete();
		    }
		    catch(Exception ex) {
		    }
		}
                // some error, notify user...
                ErrorManager errMgr = TopManager.getDefault().getErrorManager();
		String fmt = 
		    NbBundle.getBundle(JarCompiler.class).getString(
					    "EXC_AnnotatedCompilationError"); 
		String msg;
                ErrorManager.Annotation[] msgs = errMgr.findAnnotations(exc);
                if (msgs == null || msgs.length == 0) {
		    msg = exc.getLocalizedMessage();			    
                }
		else
		{
		    msg = msgs[0].getLocalizedMessage();
		}
		errMgr.annotate(
			exc, MessageFormat.format(fmt, new Object[] {msg}));
                errMgr.notify(ErrorManager.ERROR, exc);
                return false;
            } finally {
                curCreator = null;
            }
            return true;
        }

        /** Notification about packaging (compiling) progress. */
        public void progress (int percent, String description) {
            FileObject fo = curCreator.getProcessedFileObject();
            if (fo != null) {
                 fireProgressEvent(new ProgressEvent(this, fo));
                 }
        }

    } // end of Group inner class

}

/*
* <<Log>>
*  14   Jaga      1.12.1.0    4/20/00  David Simonek   
*  13   Gandalf   1.12        1/25/00  David Simonek   Various bugfixes and i18n
*  12   Gandalf   1.11        1/16/00  David Simonek   i18n
*  11   Gandalf   1.10        10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems copyright in file comment
*  10   Gandalf   1.9         10/14/99 David Simonek   manifest updating 
*       bugfixes
*  9    Gandalf   1.8         10/13/99 David Simonek   jar content now primary 
*       file, other small changes
*  8    Gandalf   1.7         10/4/99  David Simonek   
*  7    Gandalf   1.6         7/31/99  David Simonek   lot of bugfixes
*  6    Gandalf   1.5         6/9/99   David Simonek   bugfixes, progress 
*       dialog, compiling progress..
*  5    Gandalf   1.4         6/9/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  4    Gandalf   1.3         6/8/99   David Simonek   
*  3    Gandalf   1.2         6/8/99   David Simonek   bugfixes....
*  2    Gandalf   1.1         6/4/99   Petr Hamernik   temporary version
*  1    Gandalf   1.0         6/4/99   David Simonek   
* $
*/
