/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager;

import java.beans.PropertyEditor;
import java.util.jar.*;
import java.util.*;
import java.io.*;
import java.text.MessageFormat;
import java.net.URL;
import java.net.MalformedURLException;

import org.openide.filesystems.*;
import org.openide.loaders.DataObject;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.DataFolder;
import org.openide.nodes.Node;
import org.openide.util.enum.RemoveDuplicatesEnumeration;
import org.openide.util.NbBundle;
import org.openide.TopManager;
import org.openide.ErrorManager;

import org.netbeans.modules.jarpackager.util.VersionSerializator;
import org.netbeans.modules.jarpackager.util.JarUtils;
import org.netbeans.modules.jarpackager.options.JarPackagerOption;
import org.netbeans.modules.jarpackager.api.*;

/** This class is responsible for holding all the information needed
* to create an archive.
* Archive is fully specified and described by a set of properties
* which this class defines.
* Class is serializable to allow clients to store the information
* about archive for later refreshing of the content of archive.<p>
* Jar package is created from a set of file objects.
* Clients will use putFile method to create set of root file objects
* which will represent the content of resulting jar. If the root file
* object is folder, then all its content will be added into jar.
* Clients can also specify FileObjectFilter implementation to control
* which types of files to put in the resulting jar.<p>
*
* Clients can also specify extra information through callback interface
* ExtraInfoProducer.
*
* @author Dafe Simonek
*/
public class JarContent extends Object implements ArchiveDescriptor, Cloneable {

    /** Constant for filter that accepts class files only */
    public static final FileObjectFilter CLASSES_ONLY = new ClassesOnlyFilter();
    /** Constant for filter that accepts all but java files */
    public static final FileObjectFilter DEFAULT = new AllButJavaFilter();
    /** Constant for filter that accepts all files */
    public static final FileObjectFilter ALL = new AcceptAllFilter();

    /** list of contents (if this is null, use content) */
    ContentMemberList contentList;
    /** The filter for file objects */
    FileObjectFilter filter;
    /** Manifest instance */
    Manifest manifest;
    /** Flag for automatic file list creation in manifest
 * @deprecated moved to ManifestProperty
 */
    boolean manifestFileList;
    /** Flag for automatic main manifest attributes generation
 * @deprecated moved to ManifestProperty
 */
    boolean mainAttributes;

    /** level of the compression */
    CompressionLevel compressionLevel;
    /** target resource for achive */
    URL targetResource;

    /** Extended (archive controller-specific) properties */
    Map extendedProperties;

    /** manager for versioned serialization */
    VersionSerializator serializationManager;

    /** Creates new JarContent with default file object filter which
    * accepts all but java files and with no manifest.
    */
    public JarContent () {
        this(DEFAULT);
    }

    /** Creates new JarContent with given main manifest attributes and
    * with given file object filter.
    * @param filter File object filter
    */
    public JarContent (FileObjectFilter filter) {
        this.filter = filter;
        manifestFileList = false;
        compressionLevel = new CompressionLevel();
        mainAttributes = true;
    }

    /** Adds specified file object to the jar content.
 * When a folder is specified, all its subentries will
 * be inserted into the archive too during the process of
 * creating archive.<br>
 * Duplicates are not allowed (method checks it automatically)
 * @param file File object
 * @deprecated use addDataObject instead
 * @see addDataObject
 */
    public void putFile (FileObject file) {
        try {
            addDataObject(DataObject.find(file));
        } catch (DataObjectNotFoundException exc) {
            // show warning and continue
            TopManager.getDefault().getErrorManager().notify(
                ErrorManager.WARNING, exc
            );
        }
    }

    /** Inserts specified array of file objects into the
    * jar content.
    * @deprecated use addDataObject instead */
    public void putFiles (FileObject[] files) {
        for (int i = 0; i < files.length; i++) {
            putFile(files[i]);
        }
    }

    /** Inserts specified collection of file objects into the
    * jar content.
    * @deprecated use addDataObject instead */
    public void putFiles (Collection files) {
        for (Iterator iter = files.iterator(); iter.hasNext(); ) {
            putFile((FileObject)iter.next());
        }
    }

    /** Removes specified file object from the archive content
    * @deprecated use removeDataObject instead */
    public void removeFile (FileObject file) {
        try {
            removeDataObject(DataObject.find(file));
        } catch (DataObjectNotFoundException exc) {
            // show warning and continue
            TopManager.getDefault().getErrorManager().notify(
                ErrorManager.WARNING, exc
            );
        }
    }

    /** Removes specified array of file objects from the archive content
    * @deprecated use removeDataObject instead */
    public void removeFiles (FileObject[] files) {
        for (int i = 0; i < files.length; i++) {
            removeFile(files[i]);
        }
    }

    /** Removes all file objects that were previously added by
    * calling putFile(..) and putFiles(..) methods */ 
    public void clear () {
        getContentList().clear();
    }

    /** Returns currently asociated file object filter. */
    public FileObjectFilter getFilter () {
        return filter;
    }

    /** Sets new filter for filtering file objects in archive.
    * @filter New filter */
    public synchronized void setFilter (FileObjectFilter filter) {
        this.filter = filter;
    }

    /** Sets the manifest of the archive. */
    public void setManifest (Manifest manifest) {
        this.manifest = manifest;
    }

    /** @return the manifest of archive described by this jar content. */
    public Manifest getManifest () {
        return manifest();
    }

    /** @return String version of manifest */
    public String saveManifest() {
	String result;
    	ByteArrayOutputStream bao = new ByteArrayOutputStream();
	try {
	    getManifest().write(bao);
	    result =  bao.toString();
	}
	catch (IOException ex) {
	    result = "";
	}
	return result;
    }

    /** set manifest from String */
    public void restoreManifest(String str) {
	try {
	    ByteArrayInputStream bai = new ByteArrayInputStream(str.getBytes());
	    Manifest man = new Manifest();
	    man.read(bai);
	    setManifest(man);
	}
	catch (IOException ex) {
	}
    }

		
    /** Enables or disables automatic generation of archive file
    * list to the manifest
    * @deprecated don't use, moved to ManifestProperty */
    public void setManifestFileList (boolean manifestFileList) {
        this.manifestFileList = manifestFileList;
    }

    /** @return true if automatic generation of archive file list
    * in manifest is enabled, false otherwise. Default is false.
    * @deprecated don't use, moved to ManifestProperty */
    public boolean isManifestFileList () {
        return manifestFileList;
    }

    /** Enables or disables automatic generation of various
    * main attributes of the manifest
    * @deprecated don't use, moved to ManifestProperty */
    public void setMainAttributes (boolean mainAttributes) {
        this.mainAttributes = mainAttributes;
    }

    /** @return true if automatic generation of main attributes in
    * enabled, false otherwise. Enabled by default.
    * @deprecated don't use, moved to ManifestProperty */
    public boolean isMainAttributes () {
        return mainAttributes;
    }

    /** @return target file which contains jar
    * (or will contain jar if it is not created yet)
    * @deprecated use getTargetResource instead */
    public File getTargetFile () {
	if (targetResource == null)
	    return null;
        return JarUtils.urlToFile(targetResource);
    }

    /** Sets new target file.
    * @param targetFile new target file
    * @deprecated use setTargetResource instead */
    public void setTargetFile (File targetFile) {
    	if (targetFile == null) {
	    this.targetResource = null;
	}
	else {
	    try {
		this.targetResource = targetFile.toURL();
	    } catch (MalformedURLException exc) {
		// show warning and continue
		TopManager.getDefault().getErrorManager().notify(
		    ErrorManager.WARNING, exc
		);
	    }
	}
    }

    /** @return current compression level (0-9). Property is
    * ignored when compression is disabled
    * @deprecated use getCompLevel instead */
    public int getCompressionLevel() {
        return compressionLevel.getValue().intValue();
    }

    /** Sets new compression level (0-9). Ignored when
    * compression is disabled. Throws IllegalArgumentException
    * if new compression level is invalid.
    * @deprecated use setCompLevel instead */
    public void setCompressionLevel (int compressionLevel) {
        this.compressionLevel = new CompressionLevel(compressionLevel);
    }

    /** Safe accessor for manifest */
    private Manifest manifest () {
        if (manifest == null) {
            manifest = new Manifest();
        }
        return manifest;
    }
    
    /** Get the value of an extended property */
    public Object getExtendedProperty(String name) {
    	if (extendedProperties == null) {
	    return null;
	}
	else {
	    return extendedProperties.get(name);
	}
    }

    /** Set the value of an extended property */
    public void setExtendedProperty(String name, Object value) {
    	Map props = getExtendedProperties();
	props.put(name, value);
    }

    /** Safe accessor for extended properties */
    public Map getExtendedProperties() {
    	if (extendedProperties == null) {
	    extendedProperties = new HashMap();
	}
	return extendedProperties;
    }

    /** Safe accessor for extended properties */
    public void setExtendedProperties(Map properties) {
	extendedProperties = properties;
    }

    /** Versioned deserialization */
    public void readContent (ObjectInput in)
    throws IOException, ClassNotFoundException {
        VersionSerializator vs = serializationManager();
        vs.readVersion(in);
    }

    /** Versioned serialization */
    public void writeContent (ObjectOutput out)
    throws IOException {
        VersionSerializator vs = serializationManager();
        vs.writeLastVersion(out);
    }

    /** Safe getter for serialization manager which
    * manages versioned serisalization */
    private VersionSerializator serializationManager () {
        if (serializationManager == null) {
            serializationManager = new VersionSerializator();
            serializationManager.putVersion(new Version1Serializator(this));
        }
        return serializationManager;
    }

    /** Adds specified data object to the descriptor.
     * Duplicates should not be allowed.
     * @param dataObj Data object to be added.
     * @return true if content was changed as a result of the call
     */
    public boolean addDataObject (DataObject dataObj) {
    	return getContentList().addMember(dataObj);
    }

    /** Removes specified data object from the descriptor.
     * @param dataObj Data object to be removed.
     * @return true if content was changed as a result of the call
     */
    public boolean removeDataObject(DataObject dataObj) {
        return getContentList().removeMember(dataObj);
    }

    /** @return current list of contents */
    public ContentMemberList getContentList() {
	if (contentList == null) {
	    contentList = new ContentMemberList();
	}
	return contentList;
    }

    /** set current list of contents */
    public void setContentList(ContentMemberList list) {
    	contentList = list;
    }

    /** @return resource which contains jar
     * (or will contain jar if it is not created yet)  */
    public String getTarget() {
        return targetResource != null ? targetResource.toString() : null;
    }

    /** @return resource which contains jar
     * (or will contain jar if it is not created yet)  */
    public URL getTargetResource() {
        return targetResource;
    }

    /** Sets new target resource.
     * @param url new target URL  */
    public void setTarget(String url) {
	try {
	    this.targetResource = new URL(url);
	} catch (MalformedURLException exc) {
	    // show warning and continue
	    TopManager.getDefault().getErrorManager().notify(
		ErrorManager.WARNING, exc
	    );
	}
    }

    /** Sets new target resource.
     * @param url new target URL  */
    public void setTargetResource(URL url) {
        this.targetResource = url;
    }

    /** @return current compression level.  */
    public CompressionLevel getCompLevel() {
        return compressionLevel;
    }

    /** Sets new compression level.
     * @compressionLevel new compression level
     */
    public void setCompLevel(CompressionLevel compressionLevel) {
        this.compressionLevel = compressionLevel;
    }

    /** Create a jar content from a (possibly empty) file
    */
    static JarContent readFromFile(FileObject fo) {

	JarContent jc = new JarContent();
	InputStream is = null;
    	ObjectInputStream ois = null;
	try {
	    is = fo.getInputStream();
	    if (is.available() > 2) {
		// the default template should be empty, but being 
		// paranoid in case CRLF was added at some point
		ois = new ObjectInputStream(is);
		jc.readContent(ois);
	    }
	}
	catch (IOException exc) {
	    exc.printStackTrace();
	    return null;
	} 
	catch (ClassNotFoundException exc) {
	    exc.printStackTrace();
	    return null;
	}
	finally {
	    try {
		if (ois != null) {
		    ois.close();
		}
		else if (is != null) {
		    is.close();
		}
	    }
	    catch (IOException ex) {
	    }
	}

	return jc;
    }

    /** make a copy -- a shallow clone, not a deep one
    */
    public JarContent makeCopy() {
	JarContent copy = null;
    	try {
	    copy = (JarContent)super.clone();
	    serializationManager = null;
	}
	catch (CloneNotSupportedException ex) {
	}
	return copy;
    }

    /** Implementation of the filter that accepts only *.class files */
    public static final class ClassesOnlyFilter extends Object implements FileObjectFilter {
        static final long serialVersionUID = 7475557013758392767L;
        public boolean accept (FileObject fo) {
            return "class".equals(fo.getExt()); // NOI18N
        }
        /** Keep the uniquennes of this filter after deserialization */
        private Object readResolve () throws ObjectStreamException {
            return CLASSES_ONLY;
        }
        /** Display name */
        public java.lang.String toString() {
            return NbBundle.getBundle(JarContent.class).getString("CTL_ClassesOnlyFilter");
        }
        
    }

    /** Implementation of the filter that accepts all but *.java, *.form files */
    public static final class AllButJavaFilter implements FileObjectFilter {
        static final long serialVersionUID = -6474655716324211768L;
        public boolean accept (FileObject fo) {
            String extension = fo.getExt();
            return !("java".equals(extension)) && !("jar".equals(extension)) && // NOI18N
                   !("form".equals(extension)) && // NOI18N
                   !(JarPackagerOption.singleton().getContentExt().equals(extension));   
        }
        /** Keep the uniquennes of this filter after deserialization */
        private Object readResolve () throws ObjectStreamException {
            return DEFAULT;
        }
        /** Display name */
        public java.lang.String toString() {
            return NbBundle.getBundle(JarContent.class).getString("CTL_AllButJavaFilter");
        }
    }

    /** Implementation of the filter that accepts all files */
    public static final class AcceptAllFilter implements FileObjectFilter {
        static final long serialVersionUID = 8921981094756492767L;
        public boolean accept (FileObject fo) {
            return true;
        }
        /** Keep the uniquennes of this filter after deserialization */
        private Object readResolve () throws ObjectStreamException {
            return ALL;
        }
        /** Display name */
        public java.lang.String toString() {
            return NbBundle.getBundle(JarContent.class).getString("CTL_AcceptAllFilter");
        }
    }

    /** This class takes care of serializaton of JarContent
    * in first version of jar packager module.
    */
    static final class Version1Serializator implements VersionSerializator.Versionable {
        /** Asociated jar content to manage serialization for */
        JarContent jc;

        public Version1Serializator (JarContent jc) {
            this.jc = jc;
        }

        public String getName () {
            return "Version_1.0"; // NOI18N
        }

        public void readData (ObjectInput in)
        throws IOException, ClassNotFoundException {
            Object obj = in.readObject();
	    //System.out.println("Read a " + obj.getClass().getName());
	    if (obj instanceof ContentMemberList) {
	    	jc.setContentList((ContentMemberList)obj);
	    }
	    else {
	    	jc.setContentList(new ContentMemberList((ArrayList)obj));
            }
            jc.filter = (FileObjectFilter)in.readObject();
            in.readObject(); // was extraProducer
            jc.setTargetFile((File)in.readObject());
            jc.manifestFileList = ((Boolean)in.readObject()).booleanValue();
            jc.mainAttributes = ((Boolean)in.readObject()).booleanValue();
            in.readObject(); // was compressed
            jc.setCompressionLevel(((Integer)in.readObject()).intValue());
            if (((Boolean)in.readObject()).booleanValue()) {
                jc.manifest = new Manifest();
                jc.manifest.read((InputStream)in);
            }
	    try {
		readExtendedProperties(in, jc);
	    }
	    catch (java.io.EOFException ex) {
	    	//System.out.println("Caught EOF");
	    }
        }

        public void writeData (ObjectOutput out)
        throws IOException {
            //System.out.println("in writeData " + jc);
            out.writeObject(jc.getContentList());
            out.writeObject(jc.filter);
            out.writeObject(null); // was extraProducer
            out.writeObject(jc.getTargetFile());
            out.writeObject(new Boolean(jc.manifestFileList));
            out.writeObject(new Boolean(jc.mainAttributes));
            out.writeObject(null); // was compressed
            out.writeObject(new Integer(jc.getCompressionLevel()));
            out.writeObject(new Boolean(jc.manifest != null));
            if (jc.manifest != null)
                jc.manifest.write((OutputStream)out);
	    writeExtendedProperties(out, jc);
        }

	/** Read persistent form of extended properties, and convert to
	 * run-time form. Both are Maps, but in the persistent form, each
	 * property is represented by a string.  The map of current 
	 * defined extended properties held by the JarCreator contains 
	 * PropertyEditors we can use to convert the strings back to objects.
	 */
	private void readExtendedProperties(ObjectInput in, JarContent jc) 
			    throws IOException, ClassNotFoundException {
	    jc.extendedProperties = new HashMap();
	    Map persist = (Map)in.readObject();
	    Map definedProps = JarCreator.getExtendedProperties();
	    if (persist == null || definedProps == null)
		return;
	    Iterator persistKeys = persist.keySet().iterator();

	    while (persistKeys.hasNext()) {
		String key = (String)persistKeys.next();
		ExtendedPropertyFactory fact = 
			(ExtendedPropertyFactory)definedProps.get(key);
		if (fact != null) {
		    Node.Property prop = fact.createProperty(null, jc);
		    PropertyEditor propEd = prop.getPropertyEditor();
		    propEd.setAsText((String)persist.get(key));
		    try {
			prop.setValue(propEd.getValue());
		    }
		    catch(Exception ex) {
		    }
		}
	    }

	}
		
	/** write persistent form of extended properties.
	 */
	private void writeExtendedProperties(ObjectOutput out, JarContent jc) 
				    throws IOException {
	    Map persist = null;
	    Map definedProps = JarCreator.getExtendedProperties();

            //System.out.println("" + jc.extendedProperties.size() + " " + definedProps.size());
	    if (jc.extendedProperties != null && definedProps != null) {
		persist = new HashMap();
		Iterator keys = jc.extendedProperties.keySet().iterator();
		while (keys.hasNext()) {
		    String key = (String)keys.next();
                    //System.out.println(key + " : " + jc.getExtendedProperties().get(key));
		    ExtendedPropertyFactory fact = 
			    (ExtendedPropertyFactory)definedProps.get(key);
		    if (fact != null) {
			Node.Property prop = 
			    fact.createProperty(null, jc);
			PropertyEditor propEd = prop.getPropertyEditor();
			try {
                            //System.out.println("Key: " + key);
			    propEd.setValue(prop.getValue());
			    persist.put(key, propEd.getAsText());
			}
			catch(Exception ex) {
			}
		    }
		}
	    }
	    out.writeObject(persist);
	}
    } // end of Version1Serializator inner class

}

/*
* <<Log>>
*  26   Jaga      1.22.1.0.1.14/20/00  David Simonek   
*  25   Jaga      1.22.1.0.1.04/14/00  David Simonek   first draft of major 
*       update (for forte guys)
*  24   Gandalf-post-FCS1.22.1.0    3/15/00  David Simonek   serialization fix 
*       (serialization UIDs added)
*  23   Gandalf   1.22        1/25/00  David Simonek   Various bugfixes and i18n
*  22   Gandalf   1.21        1/16/00  David Simonek   i18n
*  21   Gandalf   1.20        11/10/99 David Simonek   Workaround for bug - 
*       MultiDataObject secondary entries incorrectly empty (caused missing 
*       files in archive)
*  20   Gandalf   1.19        10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  19   Gandalf   1.18        10/13/99 David Simonek   various bugfixes 
*       concerning primarily manifest
*  18   Gandalf   1.17        10/10/99 Petr Hamernik   console debug messages 
*       removed.
*  17   Gandalf   1.16        10/5/99  David Simonek   various fixes, only 
*       primary entries now resides in 'chosen content'
*  16   Gandalf   1.15        10/4/99  David Simonek   
*  15   Gandalf   1.14        9/16/99  David Simonek   a lot of bugfixes (RE 
*       filters, empty jar content etc)  added templates
*  14   Gandalf   1.13        9/13/99  David Simonek   bugfixes, compressed 
*       on/off support fixed
*  13   Gandalf   1.12        9/8/99   David Simonek   new version of jar 
*       packager
*  12   Gandalf   1.11        8/1/99   David Simonek   another bugfixes...
*  11   Gandalf   1.10        8/1/99   David Simonek   automatic file list 
*       generation to the manifest added
*  10   Gandalf   1.9         6/10/99  David Simonek   progress indocator + 
*       minor bugfixes....
*  9    Gandalf   1.8         6/9/99   David Simonek   bugfixes, progress 
*       dialog, compiling progress..
*  8    Gandalf   1.7         6/9/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  7    Gandalf   1.6         6/8/99   David Simonek   
*  6    Gandalf   1.5         6/8/99   David Simonek   bugfixes....
*  5    Gandalf   1.4         6/4/99   Petr Hamernik   defe's bugfix
*  4    Gandalf   1.3         6/4/99   David Simonek   
*  3    Gandalf   1.2         6/4/99   Petr Hamernik   temporary version
*  2    Gandalf   1.1         6/4/99   David Simonek   manifest ceration now 
*       correctly supported
*  1    Gandalf   1.0         6/3/99   David Simonek   
* $
*/
