/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager;

import java.util.jar.*;
import java.util.zip.CRC32;
import java.util.*;
import java.io.*;
import java.text.MessageFormat;

import org.openide.nodes.*;
import org.openide.filesystems.*;
import org.openide.util.enum.RemoveDuplicatesEnumeration;
import org.openide.execution.NbClassPath;
import org.openide.util.NbBundle;
import org.openide.TopManager;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;

import org.netbeans.modules.jarpackager.api.*;
import org.netbeans.modules.jarpackager.options.JarPackagerOption;
import org.netbeans.modules.jarpackager.util.ProgressListener;

/** Utility class for creating jar packages from a set of file objects.
* Common usage is to create ArchiveDescriptor instance which describes the jar at first
* and then create instance of this class, asociated with previously
* created ArchiveDescriptor instance. Then it's possible to create the jar
* by calling one of the createJar methods.
* This class also supports progress notification. To be notified about
* a progress, implement ProgressListener interface and asociate the
* listener with this instance.
*
* Controlling of packaging process is also possible, just register your 
* ArchiveBuilder using the addBuilder method.
*
* @author Dafe Simonek
*/
public class JarCreator extends Object {

    // Attributes

    /** The block size for buffered reading and writing */
    private static final int BLOCKSIZE = 65536;
    /** Message format for progress info */
    private static MessageFormat progressInfo;
    /** List of ArchiveBuilder instances, which can control and modify
    * packaging process  for all jarCreators */
    private static List builders = new ArrayList();
    /** List of ArchiveBuilder instances, which can control and modify
    * packaging process  for this jarCreator */
    private List myBuilders;

    /** holds the content of the jar */
    ArchiveDescriptor ad;
    /** progress listeners */
    HashSet listeners;
    /** entry currently being processed */
    ArchiveEntry curEntry;

    /** current list of extended properties */
    private static Map extendedProperties = new HashMap();

    /** Creates jar creator instance with asociated content.
     * @param ad Asociated content described by the descriptor
     */
    public JarCreator(ArchiveDescriptor ad) {
        this.ad = ad;
        // message format for progress message
        if (progressInfo == null) {
            progressInfo = new MessageFormat(
                               NbBundle.getBundle(JarCreator.class).getString("FMT_ProgressInfo")
                           );
        }
    }

    /** Writes an archive described by asociated JarContent instance to the
    * specified file object.
    * The content of jar is collected from the root file objects
    * inserted by JarContent.putFile(..) and JarContent.putFiles(...) calls.
    * The manifest file is created automatically from main attributes
    * and per-entry attributes.
    */  
    public void createJar (FileObject fo)
    throws IOException {
        FileLock lock = fo.lock();
        OutputStream os = fo.getOutputStream(lock);
        File file = NbClassPath.toFile(fo);
        try {
            doCreateJar(os, file);
        } finally {
            os.close();
            lock.releaseLock();
        }
    }

    /** Writes an archive described by asociated JarContent instance to the
    * specified output stream.
    */
    public void createJar (OutputStream os, File targetFile)
    throws IOException {
        doCreateJar(os, targetFile);
    }


    /** @return file object currently being added to the archive.
    * Returns null if creation is not in progress or if file object
    * cannot be obtained from processed entry (because ArchiveEntry
    * implementation can exist without any file object).
    * @deprecated Use getProcessedEntry instead, as this method will return
    * null for entries which are not asociated with file object.
    * @see getProcessedEntry */
    public FileObject getProcessedFileObject () {
        ArchiveEntry entry = getProcessedEntry();
        if ((entry == null) || !(entry instanceof FOArchiveEntry)) {
          // play it dirty and return first archived file object,
          // if file object cannot be accessed from entry
          List content = ((JarContent)ad).getContentList();
          if (content.size() > 0) {
	      return (FileObject)content.get(0);            
          }
          return null;
        }
        return ((FOArchiveEntry)entry).getFileObject();
    }

    /** Allows to obtain currently processed entry, which data is being
    * written to the archive. 
    * @return currently processed entry or null if no archiving is in progress
    */
    public ArchiveEntry getProcessedEntry () {
        return curEntry;
    }

    /* Adds new listener which will be notified about creating progress.
    * @param pl new listener
    */
    public synchronized void addProgressListener (ProgressListener pl) {
        if (listeners == null)
            listeners = new HashSet();
        listeners.add(pl);
    }

    /* Removes specified listener from the listener list.
    * @param pl listener to remove
    */
    public synchronized void removeProgressListener (ProgressListener pl) {
        if (listeners == null)
            return;
        listeners.remove(pl);
    }

    /** Adds packaging builder to static list of builders which can
    * control packaging process for all jarCreators.
    * @param controller intance of builder which will be called back to
    * allow it to control the process.
    * @return true if list of builders changed as a result of the call 
    */
    public synchronized static boolean addBuilder 
    					(ArchiveBuilder builder) {
        return builders.contains(builder)
               ? false : builders.add(builder);
    }

    /** Removes specified builder from static builder list.
    * @param builder builder to remove from the list
    * @return true if list of builders changed as a result of the call
    */
    public synchronized static boolean removeBuilder 
    				(ArchiveBuilder builder) {
        return builders.remove(builder);
    }

    /** Adds packaging builder to instance list of builders which
    * control packaging process for this jarCreator only.
    * @param controller intance of builder which will be called back to
    * allow it to control the process.
    * @return true if list of builders changed as a result of the call 
    */
    public synchronized boolean addSpecificBuilder 
    					(ArchiveBuilder builder) {
	if (myBuilders == null) {
	    myBuilders = new ArrayList();
	}
        return myBuilders.contains(builder)
               ? false : myBuilders.add(builder);
    }

    /** Removes specified builder from instance builder list.
    * @param builder builder to remove from the list
    * @return true if list of builders changed as a result of the call
    */
    public synchronized boolean removeSpecificBuilder 
    				(ArchiveBuilder builder) {
        return (myBuilders == null) ? false : myBuilders.remove(builder);
    }

    synchronized ArchiveBuilder [] getBuilders() {
    	List tempList;
	if (myBuilders == null) {
	    tempList = builders;
	}
	else {
	    tempList = new ArrayList(builders);
	    tempList.addAll(myBuilders);
	}
	ArchiveBuilder retval[] = new ArchiveBuilder[tempList.size()];
    	return (ArchiveBuilder[])tempList.toArray(retval);
    }

    /** Fires notification about creating progress.
    * @param pe progress event to fire off
    */
    protected void fireProgressEvent (int percent, String description) {
        if (listeners == null)
            return;
        HashSet cloned;
        // clone listener list
        synchronized (this) {
            cloned = (HashSet)listeners.clone();
        }
        // fire on cloned list to prevent from modifications when firing
        for (Iterator iter = cloned.iterator(); iter.hasNext(); ) {
            ((ProgressListener)iter.next()).progress(percent, description);
        }
    }

    /** get canonical name for targetDirectory string. null if no target 
     * was specified.  Otherwise, ending in '/'.
     */
    public static String getCanonicalTargetName(String target) {
    	if (target == null || target.length() == 0) {
	    return null;
	}
	else if (target.equals(ArchiveMember.ROOT)) {
	    return target;
	}
	else if (!target.endsWith("/")) {
	    return target + "/";
	}
	else {
	    return target;
	}
    }

    /* phases of jar construction */

    /** phase 0 -- find all active and interested builders */
    private ArchiveController[] initControllers(File targetFile, Map props) {
        ArchiveBuilder[] acArray = getBuilders();
	List ctrlList = new ArrayList();
	for (int i = 0; i < acArray.length; i++) {
	    ArchiveController ctl = 
	    	acArray[i].initController(targetFile, props);
	    if (ctl != null) {
		ctrlList.add(ctl);
	    }
	}
	ArchiveController ctrls[] = new ArchiveController[ctrlList.size()];
	//System.out.println("There are " + ctrls.length + "controllers active.");
	return (ArchiveController[])ctrlList.toArray(ctrls);
    }

    /** Phase 1 -- Allow controllers to process jar inputs */
    private ArchiveMember[] processMemberSet(
		ArchiveDescriptor ad, ArchiveController ctrls[]) throws IOException {
	//System.out.println("About to convert " + ad.getDataObjects().getClass().getName());
	ContentMemberList.ArchiveMemberSetImpl memberSet = 
		    new ContentMemberList.ArchiveMemberSetImpl(ad.getContentList());
	while (true) {
	    //System.out.println("There are " + memberSet.size() + "data objects.");
	    memberSet.setUnchanged();
	    for (int i = 0; i < ctrls.length; i++) {
	    	ctrls[i].processMemberSet(memberSet);
	    }
	    if (!memberSet.wasChanged()) {
	    	break;
	    }
	}
	
	return memberSet.allMembers();
    }
	
    /** Phase 1.5 -- make ArchiveEntry's */
    private List makeArchiveEntries(
    	ArchiveFilter globalFilter, ArchiveMember[] members) throws IOException{
	// Convert ArchiveMembers to ArchiveEntries
	List archiveEntries = new ArrayList();
	for (int i = 0; i < members.length; i++) {
	    Iterator iter = members[i].getDataObject().files().iterator();
	    while (iter.hasNext()) {
	    	FileObject file = (FileObject)iter.next();
		if (globalFilter != null && !globalFilter.accept(file)) {
		    file = null;
		}
		else {
		    ArchiveFilter localFilter = members[i].getFilter();
		    if (localFilter != null && !localFilter.accept(file)) {
		    	file = null;
		    }
		}
		if (file != null) {
		    ArchiveEntry ae;
		    String target = 
		    	getCanonicalTargetName(members[i].getTargetDirectory());
		    if (target == null) {
		    	ae = new FOArchiveEntry(file);
		    }
		    else if (target.equals(ArchiveMember.ROOT)) {
		    	ae = new FOArchiveEntry(file, file.getNameExt());
		    }
		    else {
		    	ae = new FOArchiveEntry(file, target+file.getNameExt());
		    }
		    archiveEntries.add(ae);
		}
	    }
	}
	//System.out.println("There are " + archiveEntries.size() + "archive entries.");
	return archiveEntries;
    }

    /** Phase 2 -- let controllers add additional entries */
    private void addEntries(ArchiveController ctrls[], List archiveEntries) throws IOException{
	for (int i = 0; i < ctrls.length; i++) {
	    ArchiveEntry more[] = 
	    	ctrls[i].addEntry();
	    if (more != null) {
		archiveEntries.addAll(Arrays.asList(more));
	    }
	}
	//System.out.println("There are " + archiveEntries.size() + "archive entries.");
    }

    /** Phase 3 -- let controllers express interest in each archive entry */
    private List willPostprocess(
    			ArchiveController ctrls[], List archiveEntries) throws IOException{
	List others = null;
	List annotatedEntries = new ArrayList();

	for (int i = 0; i < archiveEntries.size(); i++) {
	    ArchiveEntry entry = (ArchiveEntry)archiveEntries.get(i);
	    if (others == null) {
	    	others = new ArrayList();
	    }
	    ArchiveController first = null;
	    ArchiveController last = null;
	    for (int j = 0; j < ctrls.length; j++) {

	    	int mode = ctrls[j].willPostprocess(entry);
		switch (mode) {
		    case ArchiveController.WILL_POSTPROCESS:
			others.add(ctrls[j]);
			break;

		    case ArchiveController.MUST_POSTPROCESS_FIRST:
		    	if (first != null) {
			    // Issue error and throw exception
			}
			first = ctrls[j];
			break;

		    case ArchiveController.MUST_POSTPROCESS_LAST:
		    	if (last != null) {
			    // Issue error and throw exception
			}
			last = ctrls[j];
			break;
		}
	    }
	    if (first != null) {
	    	others.add(0, first);
	    }
	    if (last != null) {
	    	others.add(last);
	    }

	    List pp = null;
	    if (others.size() > 0) {
	    	pp = others;
		others = null;
	    }
	    AnnotatedEntry ae = new AnnotatedEntry(entry, pp);
	    annotatedEntries.add(ae);
	}
	return annotatedEntries;
    }

    /** Writes an archive described by asociated JarContent instance to the
    * specified output stream. Called from both createJar methods.
    * @param postProcess true if post processing should be done i
    */
    boolean doCreateJar (OutputStream os, File targetFile)
    throws IOException {
        long now = System.currentTimeMillis();
	
	// Find active controllers
	ArchiveController ctrls[] = 
		initControllers(targetFile, ad.getExtendedProperties());

	// Let the controllers modify the archive members
	ArchiveMember[] members = processMemberSet(ad, ctrls);

	// Make archive entries (i.e. FileObjects) from the 
	// archive members (i.e. DataObjects)
	List archiveEntries = makeArchiveEntries(ad.getFilter(), members);

	// let controllers add entries
	addEntries(ctrls, archiveEntries);

	// let controllers express interest in each archive entry
	List annotatedEntries = willPostprocess(ctrls, archiveEntries);

	// build the archive
        Manifest manifest = ad.getManifest();
        JarOutputStream jos = (manifest == null) ? new JarOutputStream(os)
                              : new JarOutputStream(os, manifest);
        try {
            // set write method - do not use STORED method, because
            // there is a need to compute CRCs of entries, which causes troubles
            jos.setMethod(JarOutputStream.DEFLATED);
            jos.setLevel(ad.getCompLevel().getValue().intValue());
            curEntry = null;
            BufferedInputStream bufIs = null;
            JarEntry curJarEntry = null;
            long curSize = 0;
            // create jar archive from entries found in filled driver
            InputStream in = null;
            long curIndex = 1;
            double fullPercentage = annotatedEntries.size() / 100.0;
            // getting path for *.jarContent file
            String ContentPath=targetFile.getAbsolutePath();
            int offsetOfDot = ContentPath.lastIndexOf(".");
            if (offsetOfDot != -1)
            {
                // remove extension from filename
                ContentPath = ContentPath.substring(0, offsetOfDot);
            }
            // Now append .jarContents extension
            ContentPath = ContentPath + "."
                    + JarPackagerOption.singleton().getContentExt();
            File targetContentFile=new File(ContentPath);
            //System.out.println("ContentPath :"+ContentPath);
            //System.out.println("target:"+targetFile);
            
	    for (Iterator iter = annotatedEntries.iterator(); iter.hasNext();) {
                // try to open input stream for current entry
		AnnotatedEntry currAnn = (AnnotatedEntry)iter.next();
                curEntry = currAnn.entry;
                // check if entry is not archive itself or its content file
                if (curEntry instanceof FOArchiveEntry) {
		    File curFile = NbClassPath.toFile(
			((FOArchiveEntry)curEntry).getFileObject());
		    if (curFile != null) {
			if (targetFile.compareTo(curFile) == 0) { 
			    //System.out.println(
			    //		"Archiving itself!: "+
			    // 		targetFile.getCanonicalPath());
			    continue;
			 }
			 if (targetContentFile.compareTo(curFile) == 0) {
			    //System.out.println(
			    //	"jarcontent:"+curFile.getAbsolutePath());
			   continue;
			 }
		    }
                }     
                try {
		    //System.out.println("Adding entry : "+ curEntry.getName());
                    in = curEntry.createInputStream();
                } catch (IOException exc) {
                    // warn user that this entry cannot be added
                    // and skip to next entry
                    cannotAdd(curEntry);
                    continue;
                }
                bufIs = new BufferedInputStream(in);
                curJarEntry = new JarEntry(curEntry.getName());
                jos.putNextEntry(curJarEntry);
                try {
		    InputStream strm = bufIs;
		    if (currAnn.controllers != null) {
			for (int i = 0; i < currAnn.controllers.size(); i++) {
			    ArchiveController ctrl = 
			    	(ArchiveController)currAnn.controllers.get(i);
			    strm = ctrl.postProcess(curEntry, strm);
			}
		    }
		    //System.out.println("writing " + curEntry.getName());
                    writeEntry(jos, strm);
                } finally {
                    bufIs.close();
                }
                jos.closeEntry();
                // notify progress listeners
                fireProgressEvent(
                    (int)(Math.round(curIndex++ / fullPercentage)),
                    progressInfo.format(new Object[] { curEntry.getName() } )
                );
            }
        } finally {
            jos.finish();
            curEntry = null;
        }
//System.out.println(System.currentTimeMillis() - now);
        return true;
    }

			
    /** Writes the content of given input stream to the specified
    * jar output stream */
    static void writeEntry (JarOutputStream jos, InputStream is)
    throws IOException {
        byte[] content = new byte[BLOCKSIZE];
	boolean eof = false;

	while (true) {
	    int curOff = 0;
	    while (curOff < BLOCKSIZE) {
		int read = is.read(content, curOff, BLOCKSIZE - curOff);
		if (read < 0) {
		    eof = true;
		    if (curOff > 0) {
		    	jos.write(content, 0, curOff);
		    }
		    break;
		}
		else {
		    curOff += read;
		}
	    }
	    if (eof) {
	    	break;
	    }
	    jos.write(content, 0, BLOCKSIZE);
	}
    }

    /** print our warning messages to both output and compiler tabs, so they
    * can't be missed
    */
    private static void showWarning(String warning)
    {
	String compilerTab = 
		NbBundle.getBundle("org.openide.compiler.Bundle").
						getString ("CTL_CompileTab");
    	TopManager.getDefault().getStdOut().println(warning);
	InputOutput compIO = TopManager.getDefault().getIO(compilerTab, false);
	if (compIO != null) {
	    compIO.getOut().println(warning);
	}
    }

    /** Utiltity method, notifies user that given entry
    * cannot be added to the archive */
    private static void cannotAdd (ArchiveEntry entry) {
        String warning = MessageFormat.format(
	    NbBundle.getBundle(JarCreator.class).getString("FMT_CannotAddFo"),
                new Object[] { entry.getName() }
            );
    	showWarning(warning);
    }

    /** Add a new global option
     * @param option description of option to add
     * @return true if property was added; false if it already exists
     */
     synchronized static public boolean addExtendedProperty(
     				ExtendedPropertyFactory option) {
     	String name = option.getName();
	Object oldFactory = extendedProperties.put(name, option);
     	return (oldFactory != option);
     }

    /** remove a global option
     * @param name of option to remove
     * @return true if property was removed; false if it did not exist
     */
    synchronized static public boolean removeExtendedProperty(String name){
     	return (extendedProperties.remove(name) != null);
    }

    /** get current map of all current extended properties
     *
     */
    synchronized static Map getExtendedProperties() {
	return new HashMap(extendedProperties);
    }


    /** archive entry and the controllers which need to 
     * preprocess it
     */
    static class AnnotatedEntry {
    	
	/** The entry to add to the jar file */
	ArchiveEntry entry;

	/** The list of controllers which need to postprocess it */
	List controllers;

    	AnnotatedEntry(ArchiveEntry entry, List controllers) {
	    this.entry = entry;
	    this.controllers = controllers;
	}
    }
}

/*
* <<Log>>
*  21   Jaga      1.18.1.1    4/20/00  David Simonek   
*  20   Jaga      1.18.1.0    4/14/00  David Simonek   first draft of major 
*       update (for forte guys)
*  19   Gandalf   1.18        12/7/99  David Simonek   
*  18   Gandalf   1.17        10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems copyright in file comment
*  17   Gandalf   1.16        10/13/99 David Simonek   various bugfixes 
*       concerning primarily manifest
*  16   Gandalf   1.15        10/13/99 David Simonek   jar content now primary 
*       file, other small changes
*  15   Gandalf   1.14        10/4/99  David Simonek   
*  14   Gandalf   1.13        9/13/99  David Simonek   bugfixes, compressed 
*       on/off support fixed
*  13   Gandalf   1.12        9/8/99   David Simonek   new version of jar 
*       packager
*  12   Gandalf   1.11        8/18/99  Jesse Glick     Trivial.
*  11   Gandalf   1.10        8/1/99   David Simonek   automatic file list 
*       generation to the manifest added
*  10   Gandalf   1.9         6/10/99  David Simonek   progress indocator + 
*       minor bugfixes....
*  9    Gandalf   1.8         6/10/99  David Simonek   progress dialog now 
*       functional
*  8    Gandalf   1.7         6/9/99   David Simonek   bugfixes, progress 
*       dialog, compiling progress..
*  7    Gandalf   1.6         6/9/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  6    Gandalf   1.5         6/5/99   David Simonek   
*  5    Gandalf   1.4         6/4/99   David Simonek   
*  4    Gandalf   1.3         6/4/99   Petr Hamernik   temporary version
*  3    Gandalf   1.2         6/4/99   David Simonek   manifest ceration now 
*       correctly supported
*  2    Gandalf   1.1         6/3/99   David Simonek   
*  1    Gandalf   1.0         5/26/99  David Simonek   
* $
*/
