/*
*                 Sun Public License Notice
* 
* The contents of this file are subject to the Sun Public License
* Version 1.0 (the "License"). You may not use this file except in
* compliance with the License. A copy of the License is available at
* http://www.sun.com/
* 
* The Original Code is NetBeans. The Initial Developer of the Original
* Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
* Microsystems, Inc. All Rights Reserved.
*/

package org.netbeans.modules.jarpackager;

import java.io.*;
import java.text.*;
import java.util.*;

import org.openide.*;
import org.openide.execution.*;
import org.openide.loaders.*;
import org.openide.util.*;


/**
* This is an excutor which knows how to execute a jar file in an external 
* process.  It's a very thin cover over ProcessExecutor, which changes only
* what's necesary to pass to generate the correct command line.
*/
public class JarExecutor extends ProcessExecutor {
    private static final long serialVersionUID = 1449216248334461454L;

    /** generate command line with -jar flag, and omitting -cp flag,
    * which is ignored when running a jar file.
    */
    public NbProcessDescriptor getExternalExecutor() {
	if (externalExecutor == null)
	    synchronized (JarExecutor.class) {
		if (externalExecutor == null)
		    externalExecutor = new NbProcessDescriptor (
			"{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}bin{" + Format.TAG_SEPARATOR + "}java", // /usr/local/bin/java // NOI18N
			" -jar " + // NOI18N
			"{" + Format.TAG_JARFILE + "} " + // NOI18N
			"{" + Format.TAG_ARGUMENTS + '}', // NOI18N
			NbBundle.getBundle(JarExecutor.class).getString ("MSG_JarExecutorHint")
		    );
	    }
	
	return externalExecutor;
    }

    /**
    * overridden to create the correct sort of ExecInfo
    */
    public ExecutorTask execute(DataObject obj) throws IOException {
	checkClass(obj, JarDataObject.class);
	JarDataObject jdo = (JarDataObject)obj;
	String[] params = ExecSupport.getArguments (new FileEntry(null, obj.getPrimaryFile()));
	return execute(new JarExecInfo(jdo, params));
    }

    /**
    * overridden to create the correct sort of Format.
    */
    protected Process createProcess (ExecInfo info) throws IOException {
	return getExternalExecutor ().exec (new Format (
					    info,
					    getClassPath (),
					    getBootClassPath (),
					    getRepositoryPath (),
					    getLibraryPath ()
					), 
					getEnvironmentVariables(), 
					getWorkingDirectory());
    }

    /**
    * overridden to create the correct sort of ExecInfo
    */
    protected Process createProcess (DataObject obj) throws IOException {
	String[] params = ExecSupport.getArguments (new org.openide.loaders.FileEntry(
	    null, obj.getPrimaryFile()
	));
	checkClass(obj, JarDataObject.class);
	return createProcess (new JarExecInfo((JarDataObject)obj, params));
    }

    static void checkClass(Object obj, Class expected) {
    	if (!expected.isInstance(obj)) {
	    String fmt = NbBundle.getBundle(JarExecutor.class).getString(
						    "FMT_JarExecutorBadClass");
	    throw new ClassCastException(
	    	MessageFormat.format(fmt, 
				     new Object[] {obj.getClass().getName(),
				     		   expected.getName()}
				    ));
	}
    }

    /**
    * subclassed to understand the "jarfile" tag.  A more universal
    * Format which can do introspection, or call a standard method when it sees
    * an unknown tag, would be a better idea.
    */
    static class Format extends ProcessExecutor.Format {
	
	/** Tag replaced with ProcessExecutors.getClassPath () */
	public static final String TAG_JARFILE = "jarfile"; // NOI18N

	
	public Format (
	    ExecInfo info,
	    NbClassPath classPath,
	    NbClassPath bootClassPath,
	    NbClassPath repository,
	    NbClassPath library) {

	    super(info, classPath, bootClassPath, repository, library);

	    java.util.Map map = getMap ();
	    JarExecutor.checkClass(info, JarExecInfo.class);
	    map.put(TAG_JARFILE, ((JarExecInfo)info).getJarFile());
	}
    }

    /**
    * subclassed to hold onto the jar file name.
    */
    static class JarExecInfo extends ExecInfo {
    	
	String jarFile;

	public JarExecInfo(JarDataObject jdo,  String[] argv) {
	    super("", argv);
	    this.jarFile = 
	    	jdo.getJarContent().getTargetFile().getPath();
	}

	public JarExecInfo (JarDataObject jdo) {
	    this(jdo, new String[] {});
	}

	String getJarFile() {
	    return (jarFile == null) ? "<none>" : jarFile;
	}
    }
}
