/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.jarpackager.api;

import java.io.*;
import java.util.*;
import org.openide.nodes.Node;

/** Interface for controlling packaging process. The ArchiveController will 
* be called back at defined points during the building of the jar
* file, as described in the individual methods.
*
* Note that if any archive controller throws an exception from any of the 
* callback methods in this interface, the jar file will not be created.
*
* @author Dafe Simonek
*/
public interface ArchiveController {

    /** returned from willPostprocess() if no postprocessing will be done */
    int WILL_NOT_POSTPROCESS = 0;

    /** returned from willPostprocess() if postprocessing will be done */
    int WILL_POSTPROCESS = 1;

    /** returned from willPostprocess() if controller must be the first 
     * postprocessor for the entry
     */
    int MUST_POSTPROCESS_FIRST = 2;

    /** returned from willPostprocess() if controller must be the last 
     * postprocessor for the entry
     */
    int MUST_POSTPROCESS_LAST = 3;

    /** Phase 1. Called to let each archive controller modify the set
     * of data objects which are inputs to the jar.
     * @param context context returned from initController
     * @param ArchiveMemberSet current list of inputs, which this controller
     * can modify
     * @exception IOException java.io.IOException the ArchiveContoller
     * has detected a reason that the jar file should not be created.
     */
    void processMemberSet(ArchiveMemberSet set) throws IOException;

    /** Phase 2.  Called to allow each archive controller to add additional
     * ArchiveEntry's as inputs to the jar file.
     * @param context context returned from initController
     * @return array of entries to be added
     * @exception IOException java.io.IOException the ArchiveContoller
     * has detected a reason that the jar file should not be created.
     */
    ArchiveEntry[] addEntry() throws IOException;

    /** Phase 3.  Called to allow each archive controller to express interest
     * in each ArchiveEntry.  "Interest", in this case, means a desire to 
     * process the contents of the ArchiveEntry before it is written
     * to the jar file.  The controller indicates its interest by returning 
     * one of the POSTPROCESS constants.  Note that if more than one controller
     * needs to be the first postprocessor for any entry (or more than one the
     * last), the packaging fails.
     * @param entry the ArchiveEntry which the controller may want to process
     * @return the controller's interest in this entry.
     * @exception IOException java.io.IOException the ArchiveContoller
     * has detected a reason that the jar file should not be created.
     */
    int willPostprocess(ArchiveEntry entry) throws IOException;

    /** Phase 4.  Called to allow ArchiveController's to postprocess 
     * ArchiveEntry's.  The ArchiveController replaces the InputStream
     * passed in with one of its own, usually by filtering it.  Note that
     * the ArchiveController should being with the InputStream argument, which
     * may be the result of a previous postprocessor, not with the result of
     * ArchiveEntry.createInputStream().
     * @param entry the ArchiveEntry to be modified
     * @param input the input stream from which to read the entry
     * @exception IOException java.io.IOException the ArchiveContoller
     * has detected a reason that the jar file should not be created.
     */
    InputStream postProcess(ArchiveEntry entry, InputStream input) throws IOException;
}
