/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager.filesystem;

import java.beans.*;
import java.util.*;
import java.io.File;
import java.text.MessageFormat;

import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.util.NbBundle;
import org.openide.util.actions.SystemAction;

import org.netbeans.modules.jarpackager.*;


/** Writeable jar file system. This is in fact multi file system
 * consisting of writeable local file system and read only jar file
 * system.
 *
 * @author Dafe Simonek, David Strupl
 */
public final class WriteableJarFileSystem extends MultiFileSystem 
implements PropertyChangeListener, FileSystem.Status, FileSystem.AtomicAction {
    /** serial version UID */
    static final long serialVersionUID=-7096247660760549348L;
    
    /** Holds value of property cacheFileSystem. */
    private FileSystem cacheFileSystem;
    /** Holds value of property jarFileSystem, can be null before user initializes the JAR file. */
    private JarFileSystem jarFileSystem;
    
    /** Creates new WriteableJarFileSystem */
    public WriteableJarFileSystem () {
        super(new FileSystem[1]);
        addPropertyChangeListener(this);
    }
    
    /** Getter for property cacheFileSystem.
     * @return Value of property cacheFileSystem.
     */
    public FileSystem getCacheFileSystem() {
        return cacheFileSystem;
    }
    
    /** Setter for property cacheFileSystem.
     * @param cacheFileSystem New value of property cacheFileSystem.
     */
    public void setCacheFileSystem(FileSystem cacheFileSystem) {
        FileSystem oldCacheFileSystem;
        
        synchronized (this) {
            oldCacheFileSystem = this.cacheFileSystem;
            
            this.cacheFileSystem = cacheFileSystem;
            if (cacheFileSystem != null) {
                cacheFileSystem.addPropertyChangeListener(this);
            }
            if (oldCacheFileSystem != null) {
                oldCacheFileSystem.removePropertyChangeListener (this);
            }
        }
        
        firePropertyChange ("cacheFileSystem", oldCacheFileSystem, cacheFileSystem);
        
    }
    
    /** Getter for the file to display.
     * @return the file or null
     */
    public java.io.File getJarFile () {
        JarFileSystem fs = jarFileSystem;
        return fs == null ? null : fs.getJarFile ();
    }
    
    /** Setter for property jarFileSystem.
     * @param jarFile the jar file to display
     */
    public void setJarFile (java.io.File jarFile) 
    throws java.beans.PropertyVetoException, java.io.IOException {
        java.io.File old = getJarFile ();
        
        JarFileSystem fs;
        synchronized (this) {
            fs = jarFileSystem;
            if (fs == null) {
                fs = new JarFileSystem ();
            }
            fs.setJarFile (jarFile);
            
            jarFileSystem = fs;
            
            if (cacheFileSystem == null) {
                // create temporary filesystem
                FileSystem temp = createTempFileSystem ();
                temp.addPropertyChangeListener(this);
                cacheFileSystem = temp;
            }
        }
        
        firePropertyChange ("jarFile", old, jarFile);
    }
    
    /** If anything changes call setDelegates. Fires PropertyChange and
     * calls refresh().
     */
    public void propertyChange(final java.beans.PropertyChangeEvent ev) {
        if (("cacheFileSystem".equals(ev.getPropertyName())) || 
            ("jarFile".equals(ev.getPropertyName()))) {
            FileSystem arr[] = null;
            if ((cacheFileSystem != null) && (jarFileSystem != null)){
                arr = new FileSystem[] { cacheFileSystem, jarFileSystem };
            } else {
                  if (cacheFileSystem != null) {
                      arr = new FileSystem[] { cacheFileSystem };
                  } else {
                      arr = new FileSystem[] { new LocalFileSystem(), jarFileSystem };
                  }
            }
            if (arr != null) {
                setDelegates(arr);
            }
        }
        getRoot().refresh(true);
        if ( ! FileSystem.PROP_SYSTEM_NAME.equals(ev.getPropertyName()) && ev.getSource() != this) {
	    resetSystemName();
        }
    }
    
    private void resetSystemName() {
	try {
	    if (cacheFileSystem == null ||
		jarFileSystem == null) {
		// not fully initialized 
		setSystemName(""); // NOI18N
		return;
	    }
	    StringBuffer b = new StringBuffer();
	    b.append(cacheFileSystem.getSystemName());
	    b.append(jarFileSystem.getSystemName());	
    	    setSystemName(b.toString());
	} catch (PropertyVetoException ex) {
	}
    }
        /*
    * Provides name of the system that can be presented to the user.
    * @return user presentable name of the file system
    */
    public String getDisplayName () {
        MessageFormat mf = new MessageFormat(getString0("FMT_DisplayName"));
        Object []param = new Object[2];
        if (jarFileSystem != null) {
            param[0] = jarFileSystem.getDisplayName();
        } else {
            param[0] = getString0("LAB_Error");
        }
        if (cacheFileSystem != null) {
            param[1] = cacheFileSystem.getSystemName();
        } else {
            param[1] = getString0("LAB_Error");
        }

        return mf.format(param);
    }
    
    /** I am the status annotator.
     */
    public FileSystem.Status getStatus () {
        return this;
    }
    
    /** Annotate the name of a file cluster.
     * @param name the name suggested by default
     * @param files an immutable set of {@link FileObject}s belonging to this filesystem
     * @return the annotated name (may be the same as the passed-in name)
     * @exception ClassCastException if the files in the set are not of valid types
     */
    public String annotateName(String name, java.util.Set files) {
        Iterator it = files.iterator ();
        int inCache = 1;
        while (it.hasNext ()) {
            FileObject fo = (FileObject)it.next ();
            if (!fo.isRoot () && findSystem (fo) == cacheFileSystem) {
                inCache = 0;
                break;
            }
        }
        return getMessage (inCache).format (new Object[] { name });
    }
    
    /** Does not do any annotations.
     */
    public java.awt.Image annotateIcon(java.awt.Image icon, int iconType, java.util.Set files) {
        return icon;
    }
    
    /** When fire status update event.
     */
    protected void notifyMigration(FileObject fo) {
        fireFileStatusChanged (new FileStatusEvent (this, fo, false, true));
    }
    
    /** buffer of messages */
    private static final MessageFormat[] MESSAGES = new MessageFormat[2];
    
    /** Holds value of property cacheDir. */
    private java.io.File cacheDir;
    
    /** Lazy getter for messages
     * @param msgIndx index of message
     */
    private MessageFormat getMessage (int msgIndx) {
        if (MESSAGES[msgIndx] == null) {
            MESSAGES[msgIndx] = new MessageFormat (getString0 ("MSG_FS_" + msgIndx));
        }
        return MESSAGES[msgIndx];
    }
        
    
    /** Test whether the jarFileSystem is up to date. It scands all files
     * on the primary file system and compares them to the ones from the
     * jar.
     */
    private boolean jarContentNeedsUpdate() {
        FileSystem fs = getCacheFileSystem ();
        if (fs == null) {
            return false;
        }
        
        Enumeration enum = fs.getRoot().getChildren(true);
        while (enum.hasMoreElements()) {
            FileObject fo1 = (FileObject) enum.nextElement();
            // now try to find it on the jarFileSystem
            if (jarFileSystem == null) {
                return true;
            }
            // TODO: add check if the new file qualifies for jar content
            FileObject fo2 = jarFileSystem.findResource(fo1.getPackageNameExt('/','.'));
            if (fo2 == null) {
                return true;
            }
            // comparing just the dates - is it sufficient?
            if (fo1.lastModified().after(fo2.lastModified())) {
                return true;
            }
        }
        // if we didn't find any differences, there is no need to update
        return false;
    }
    
    /** Creates a new jar file with all needed content updated
      * to the current state of the file system.
     */
    private File createNewJar() throws java.io.IOException{
        // First create temp place to store the jar file
        File f = jarFileSystem.getJarFile ();
        File newFile = File.createTempFile ("tmpJarFile" + Math.random (), "jar");
        newFile.deleteOnExit();
        java.io.FileOutputStream fos = new java.io.FileOutputStream(newFile);
        // now use JarPackager to do the job
        JarContent content = new JarContent(JarContent.ALL);
        content.addDataObject(DataObject.find(getRoot()));
        JarCreator creator = new JarCreator(content);
        creator.createJar(fos, newFile);
        //
        return newFile;
    }
    
    /** This method assumes that the newJar is a file with updated
     * state of the file system.
     */
    private void swapToNewJarFile(File newJar) throws java.io.IOException {
        try {
            JarFileSystem jfs = jarFileSystem;
            if (jfs == null) {
                return;
            }
            File oldJar = jfs.getJarFile ();
            
            // unmount the old jar in order to replace by the new one
            jfs.removeNotify ();
            
            if (oldJar != null && !oldJar.delete()) {
                throw new java.io.IOException("Cannot delete old jar file");
            }
            
            java.io.FileInputStream fis = new java.io.FileInputStream(newJar);
            java.io.FileOutputStream fos = new java.io.FileOutputStream(oldJar);
            org.openide.filesystems.FileUtil.copy(fis, fos);
            
            jfs.setJarFile(oldJar);
            fis.close();
            fos.close();
            newJar.delete();
            
            jfs.addNotify ();
            getRoot ().refresh (true);
        } catch (java.beans.PropertyVetoException x) {
              TopManager.getDefault().notifyException(x);
        }
    }
    
    
    /** Create temporary cache filesystem.
     * @return the filesystem to be used as temporary cache
     */
    private FileSystem createTempFileSystem () {
        String pref = "WJFS" + String.valueOf (hashCode ()); // NOI18N
     
        // this creates empty temporary file
        java.io.File tmp;
        
        try {
            tmp = java.io.File.createTempFile (pref, null);
            tmp.delete ();
        
            // which we will convert into a directory
            tmp.mkdirs ();
        } catch (java.io.IOException ex) {
            tmp = null;
        }
        
        LocalFileSystem fs = new LocalFileSystem ();
        try {
            if (tmp != null) {
                fs.setRootDirectory (tmp);
            }
        } catch (java.beans.PropertyVetoException ex) {
            TopManager.getDefault().getErrorManager ().notify (ex);
        } catch (java.io.IOException ex) {
            TopManager.getDefault().getErrorManager ().notify (ex);
        }
        return fs;
    }
            
            
    
    /** Goes thru all files in this file system and if we have
     * identical copies on both file systems delete the file
     * from the base (cache) file system. After finishing there
     * should be no duplicated files on our two filesystems.
     */
    private void clearBaseFileSystem() {
//System.out.println("clearBaseFileSystem");
        FileSystem fs = getCacheFileSystem ();
        if (fs == null) {
            return;
        }
        
        FileSystem jar = jarFileSystem;
        if (jar == null) {
            return;
        }
        
        Enumeration enum = fs.getRoot().getChildren(true);
        while (enum.hasMoreElements()) {
            FileObject fo1 = (FileObject) enum.nextElement();
//System.out.println("    fo1 = " + fo1.getPackageNameExt('/','.'));
            // now try to find it on the jarFileSystem
            FileObject fo2 = jar.findResource(fo1.getPackageNameExt('/','.'));
//System.out.println("  fo1: " + fo1);                
//System.out.println("  fo2: " + fo2);                
            if (fo2 != null) {
                // TODO: ??? (dstrupl)
                boolean delete = fo1.lastModified().before(fo2.lastModified());
//System.out.println("  delete: " + delete);                
                if (delete) {
                    // I hope this is the right way to use FileLock (dstrupl)
                    FileLock lock = null;
                    try {
                        lock = fo1.lock();
                        fo1.delete(lock);
                    } catch (java.io.IOException x) {
                          TopManager.getDefault().notifyException(x);
                    } finally {
                          lock.releaseLock();
                    }
                }
            }
        }
//System.out.println("clearBaseFileSystem --- konec");
    }
    
    public void updateJarFileContent() {
        try {
            runAtomicAction (this);
        } catch (java.io.IOException x) {
              TopManager.getDefault().notifyException(x);
        }
    }
    
    /** Updates the content of a JAR file. Do not call directly.
     */
    public void run () throws java.io.IOException {
        if (jarContentNeedsUpdate()) {
            File f = createNewJar();
            swapToNewJarFile(f);
            clearBaseFileSystem();
            
            fireFileStatusChanged (new FileStatusEvent (this, true, true));
        }
    }
    
  /** Allows filesystems to set up the environment for external execution
    * and compilation.
    * Each filesystem can add its own values that
    * influence the environment. The set of operations that can modify
    * environment is described by the {@link Environment} interface.
    * <P> 
    * Just return the environment of the cache file system.
    *
    * @param env the environment to setup
    * @exception EnvironmentNotSupportedException if external execution
    *    and compilation cannot be supported
    */
    public void prepareEnvironment (Environment env)
    throws EnvironmentNotSupportedException {
        cacheFileSystem.prepareEnvironment(env);
        if (jarFileSystem != null) {
            jarFileSystem.prepareEnvironment(env);
        }
    }

    
    /** Merges actions.
    */
    public SystemAction[] getActions (Set files) {
        SystemAction[] arr = super.getActions (files);
        SystemAction[] a1 = new SystemAction[arr.length + 1];
        a1[0] = SystemAction.get (UpdateWritableJarFileSystemAction.class);
        System.arraycopy (arr, 0, a1, 1, arr.length);
        return a1;
    }

    private static final String getString0(String key) {
        return org.openide.util.NbBundle.getBundle(WriteableJarFileSystem.class).getString(key);
    }
}
