/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager.options;

import java.io.*;

import org.openide.options.*;
import org.openide.util.*;

/** Options for jar packager. Consists of jar content extensions,
* and initial values for new archives - compresseion level,
* generate manifest file list flag, add to repository flag.
*
* @author  Dafe Simonek
*/
public class JarPackagerOption extends SystemOption {

    /** serial version UID */
    static final long serialVersionUID = 2216855924823089903L;

    /** Content extension property */
    public static final String PROP_CONTENT_EXT = "contentExt"; // NOI18N
    private static String contentExt = "jarContent"; // NOI18N

    /** Compression level of the archive. */
    public static final String PROP_COMPRESSION_LEVEL = "compressionLevel"; // NOI18N
    public static final int DEFAULT_COMPRESSION_LEVEL = 6;
    private static int compressionLevel = DEFAULT_COMPRESSION_LEVEL;

    /** Generate manifest file list on/off flag */
    public static final String PROP_MANIFEST_FILELIST = "manifestFileList"; // NOI18N
    private static boolean manifestFileList = true;

    /** Automatic add to repository on/off flag */
    public static final String PROP_ADD_TO_REPOSITORY = "addToRepository"; // NOI18N
    private static boolean addToRepository = false;

    /** Main attributes automatic generation on/off flag */
    public static final String PROP_MAIN_ATTRIBUTES = "mainAttributes"; // NOI18N
    private static boolean mainAttributes;

    /** Confirmation of automatic inspecting on/off flag. */
    public static final String PROP_CONFIRM_AUTO_CREATION = "confirmAutoCreation"; // NOI18N
    private static boolean confirmAutoCreation;

    /** Singleton instance */
    private static JarPackagerOption singletonInstance;

    /** Creates new JarPackagerOption. */
    public JarPackagerOption() {
    }

    /** Returns default instance of jar packager system option */
    public static JarPackagerOption singleton () {
        if (singletonInstance == null) {
            singletonInstance = (JarPackagerOption) SharedClassObject.findObject (JarPackagerOption.class, true);
        }
        return singletonInstance;
    }

    /** Get a human presentable name of the action.
    * This may be presented as an item in a menu.
    * @return the name of the option
    */
    public String displayName () {
        return NbBundle.getBundle(JarPackagerOption.class).
               getString("CTL_RootOption");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (JarPackagerOption.class);
    }

    /** Getter for property compressionLevel.
    * @return Value of property compressionLevel.
    */
    public int getCompressionLevel () {
        return compressionLevel;
    }

    /** Setter for property compressionLevel.
    * @param compressionLevel New value of property compressionLevel.
    */
    public void setCompressionLevel (int compressionLevel) {
        if (this.compressionLevel == compressionLevel)
            return;
        Integer old = new Integer(this.compressionLevel);
        this.compressionLevel= compressionLevel;
        firePropertyChange(PROP_COMPRESSION_LEVEL, old, new Integer(compressionLevel));
    }

    /** Getter for property addToRepository.
     *@return Value of property addToRepository.
     */
    public boolean isAddToRepository() {
        return addToRepository;
    }

    /** Setter for property addToRepository.
     *@param addToRepository New value of property addToRepository.
     */
    public void setAddToRepository(boolean addToRep) {
        if (addToRepository == addToRep)
            return;
        Boolean old = new Boolean(addToRepository);
        addToRepository = addToRep;
        firePropertyChange(PROP_ADD_TO_REPOSITORY, old,
                           new Boolean(addToRepository));
    }

    /** Getter for property manifestFileList.
     *@return Value of property manifestFileList.
     */
    public boolean isManifestFileList() {
        return manifestFileList;
    }

    /** Setter for property manifestFileList.
     *@param manifestFileList New value of property manifestFileList.
     */
    public void setManifestFileList(boolean manFL) {
        if (manifestFileList == manFL)
            return;
        Boolean old = new Boolean(manifestFileList);
        manifestFileList = manFL;
        firePropertyChange(PROP_MANIFEST_FILELIST, old,
                           new Boolean(manifestFileList));
    }

    /** Getter for property contentExt.
     *@return Value of property contentExt.
     */
    public String getContentExt() {
        return contentExt;
    }

    /** Setter for property contentExt.
     *@param contentExt New value of property contentExt.
     */
    public void setContentExt(String contExt) {
        if (contentExt == contExt)
            return;
        String old = contentExt;
        contentExt = contExt;
        firePropertyChange(PROP_CONTENT_EXT, old, contentExt);
    }

    /** Getter for property mainAttributes.
    *@return Value of property mainAttributes.
    */
    public boolean isMainAttributes () {
        return mainAttributes;
    }

    /** Setter for property mainAttributes.
    * @param mainAttributes New value of property mainAttributes.
    */
    public void setMainAttributes (boolean mainAttr) {
        if (mainAttributes == mainAttr)
            return;
        Boolean old = new Boolean(mainAttributes);
        mainAttributes = mainAttr;
        firePropertyChange(PROP_MAIN_ATTRIBUTES, old,
                           new Boolean(mainAttributes));
    }

    /** Getter for property confirmAutoCreation.
    * @return Value of property confirmAutoCreation.
    */
    public boolean isConfirmAutoCreation () {
        return confirmAutoCreation;
    }

    /** Setter for property confirmAutoCreation.
    * @param confirmAutoCreation New value of property confirmAutoCreation.
    */
    public void setConfirmAutoCreation (boolean confirm) {
        if (confirmAutoCreation == confirm)
            return;
        Boolean old = new Boolean(confirmAutoCreation);
        confirmAutoCreation = confirm;
        firePropertyChange(PROP_CONFIRM_AUTO_CREATION, old,
                           new Boolean(confirmAutoCreation));
    }

    /** @return False to signalize that this option should be a part of
    * project settings */
    private boolean isGlobal () {
        return false;
    }
}
