/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager.util;

import java.util.*;
import java.util.jar.*;
import java.io.File;
import java.io.IOException;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.beans.PropertyVetoException;
import java.net.URL;
import java.text.MessageFormat;

import org.openide.util.actions.NodeAction;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.nodes.Node;
import org.openide.loaders.*;
import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.execution.NbClassPath;

import org.openide.compiler.CompilerJob;
import org.openide.cookies.CompilerCookie;
import org.openide.compiler.Compiler;

import org.netbeans.modules.jarpackager.*;
import org.netbeans.modules.jarpackager.options.JarPackagerOption;

/** Static utilities used for common purposes by other classes.
*
* @author Dafe Simonek
*/
public final class JarUtils extends Object {

    public static final String MANIFEST_VERSION = "1.0"; // NOI18N

    /** No need to instantiate */
    private JarUtils () {
    }

    /** @return true if specified nodes all can be added to the
    * archive described by given jar content. */
    public static boolean canAdd (ArchiveDescriptor ad, Node[] nodes,
                                  String targetName) {
        if (nodes.length == 0) {
            return false;
        }
        // check compatibility of selected nodes
        Node parent = null;
        for (int i = 0; i < nodes.length; i++) {
            parent = nodes[i].getParentNode();
            if ((parent != null) &&
                    (parent.getCookie(DataFolder.class) == null) &&
                    (nodes[i].getCookie(DataFolder.class) == null)) {
                return false;
            }
        }
        // check if it is not part of itself
        DataObject curDo = null;
        DataFolder curFolder = null;
        File targetFile =
            (targetName == null) ? null : new File(targetName);
        FileSystem curFs = null;
        for (int i = 0; i < nodes.length; i++) {
            curDo = (DataObject)nodes[i].getCookie(DataObject.class);
            if (curDo != null) {
                // check if processed item is not archive itself
                if ((curDo instanceof JarDataObject) &&
                        isEqual(targetFile, NbClassPath.toFile(curDo.getPrimaryFile()))) {
                    return false;
                }
                // check if processed item is not from the filesystem of the
                // archive itself
                try {
                    curFs = curDo.getPrimaryFile().getFileSystem();
                } catch (FileStateInvalidException exc) {
                    // file system in invalid state, so does not allow
                    // to add itema to archive from such file system
                    return false;
                }
                if ((curFs instanceof JarFileSystem) &&
                        isEqual(targetFile, ((JarFileSystem)curFs).getJarFile())) {
                    return false;
                }
            } else {
                return false;
            }
        }
        for (int i = 0; i < nodes.length; i++) {
            if(nodes[i].getDisplayName().equalsIgnoreCase("META-INF")) return false; // NOI18N
            if ((nodes[i].getParentNode().getDisplayName()!=null) && (nodes[i].getParentNode().getDisplayName().equalsIgnoreCase("META-INF"))) // NOI18N
                    return false;
        }

        return true;
    }

    /** Helper, returns true if given files are the same
    * (their extension can be different) */
    private static boolean isEqual (File file1, File file2) {
        // deal with nulls
        if (file1 == null) {
            return (file2 == null) ? true : false;
        }
        if (file2 == null) {
            return (file1 == null) ? true : false;
        }
        String name1 = file1.getPath();
        String name2 = file2.getPath();
        // strip the extensions
        int index1 = name1.lastIndexOf('.');
        if (index1 < 0) {
            index1 = name1.length();
        }
        int index2 = name2.lastIndexOf('.');
        if (index2 < 0) {
            index2 = name2.length();
        }
        return new File(name1.substring(0, index1)).equals(
                   new File(name2.substring(0, index2)));
    }

    
    /** Adds list of file objects representing all primary and
    * secondary entries of data objects which are represented by
    * specified nodes. 
    * @param jarcontent jar content to which we want to add created file list
    * @deprecated Use addNodes (ArchiveDescriptor ad, Node[] nodes) instead
    */
    public static void addFileList(JarContent jarcontent, Node nodes[])
     {
        // transform nodes to collection of file objects
        ArrayList result = new ArrayList(nodes.length * 2);
        DataObject curDo = null;
        for (int i = 0; i < nodes.length; i++) {
            curDo = (DataObject)nodes[i].getCookie(DataObject.class);
            if (curDo != null) {
                // add only primary file object
                result.add(curDo.getPrimaryFile());
            }
        }
        // add to the jar content
        jarcontent.putFiles(result);
     }


    /** Removes list of file objects representing all primary and
    * secondary entries of data objects which are represented by
    * specified nodes. 
    * @param jarcontent jar content from which we are removing files
    * @deprecated Use removeNodes (ArchiveDescriptor ad, Node[] nodes) instead
    */
    public static void removeFileList(JarContent jarcontent, Node nodes[])
    {
        // transform nodes to collection of file objects
        DataObject curDo = null;
        for (int i = 0; i < nodes.length; i++) {
            curDo = (DataObject)nodes[i].getCookie(DataObject.class);
            if (curDo != null) {
                // remove primary file object
                jarcontent.removeFile(curDo.getPrimaryFile());
            }
        }
     }
    
    
    
    
    /** Adds list of data objects represented by given nodes array to the
     * archive descriptor.
    * @param ad archive descriptor to which we want to add data objects
    */
    public static void addNodes (ArchiveDescriptor ad, Node[] nodes) {
        DataObject curDo = null;
	ContentMemberList members = ad.getContentList();
        // retrieve data object for each node and add it to the descriptor
        for (int i = 0; i < nodes.length; i++) {
            curDo = (DataObject)nodes[i].getCookie(DataObject.class);
            if (curDo != null) {
                if (curDo.getNodeDelegate().getDisplayName().length()==0)
                    curDo.getNodeDelegate().setDisplayName(nodes[i].getDisplayName());
                members.addMember(curDo);
            } else {
                throw new NullPointerException("Added node doesn't support DataObject cookie"); // NOI18N
            }
        }
    }

    /** Removes data objects represented by given nodes from given
     * descriptor.
    * @param ad archive content from which we are removing data objects
    */
    public static void removeNodes (ArchiveDescriptor ad, Node[] nodes) {
        // retrieve data object for each node and try to remove it
        // from archive descriptor
	ContentMemberList members = ad.getContentList();
        DataObject curDo = null;
        for (int i = 0; i < nodes.length; i++) {
            members.removeMember((DataObject)nodes[i].getCookie(DataObject.class));
        }
    }

    /** Utility method that creates new jar file system with given
    * root file and adds created file system to the repository.
    * @param writeable true if file system should be writeable,
    * false if read-only is enough */
    public static FileSystem addJarFSToRepository (File rootFile, boolean writeable)
    throws IOException, PropertyVetoException {
        Repository repo = TopManager.getDefault().getRepository();
        FileSystem foundFS = getMountedJarFS(rootFile);
        if (foundFS == null) {
            JarFileSystem jarFs = new JarFileSystem();
            jarFs.setJarFile(rootFile);
            repo.addFileSystem(jarFs);
            return jarFs;
        }
        return foundFS;
    }

    /** @return FileSystem which has given jar file as its root or
    * null if no such file system could be found in repository */
    public static FileSystem getMountedJarFS (File rootFile) {
        Repository repo = TopManager.getDefault().getRepository();
        // PENDING -> checking is not robust, should be done better..
        // PENDING - writeable file systems not yet supported
        FileSystem[] allFS = repo.toArray();
        for (int i = 0; i < allFS.length; i++) {
            if ((allFS[i] instanceof JarFileSystem) &&
                    rootFile.equals(((JarFileSystem)allFS[i]).getJarFile())) {
                return allFS[i];
            }
        }
        return null;
    }

    /** Tries to inspect given node for JarDataObject cookie
    * and asks for its jar content.
    * @return JarContent connected with specified node or null
    * if no such jar content can be found.
    */
    public static JarContent jarContentFromNode (Node node) {
        JarDataObject jdo =
            (JarDataObject)(node.getCookie(JarDataObject.class));
        return (jdo != null) ? jdo.getJarContent() : null;
    }
    
    /** resolve DataShadows */
    public static DataObject resolve(DataObject source) {
    	if (source instanceof DataShadow) {
	    return ((DataShadow)source).getOriginal();
	}
	else return source;
    }
   
    /** creating compiler job without duplicates */
    static class CompilerJobBuilder {
	/** Dataobjects to compile */
    	private HashSet seen = new HashSet();

	/** Job we are creating */
	private CompilerJob job = new CompilerJob(Compiler.DEPTH_INFINITE);

	/** add a DataOject not to be compiled */
	boolean ignoreDataObject(DataObject ignore) {
	    return seen.add(ignore);
	}

	/** process a DataObject */
	void addDataObject(DataObject dataObj) {
	    dataObj = JarUtils.resolve(dataObj);
	    if (dataObj == null || seen.contains(dataObj))
	    	return;

	    if (dataObj instanceof JarDataObject) {
	    	//System.out.println("processing jar " +dataObj.getPrimaryFile());
		addToCompilerJob(
		    this, ((JarDataObject)dataObj).getJarContent(), dataObj);
	    }
	    else {
		DataFolder folder = 
				(DataFolder)dataObj.getCookie(DataFolder.class);
		if (folder == null) {
		    // Not a folder
		    addScalarObjectToJob(dataObj);
		}
		else if (seen.contains(folder)) {
		    return;
		}
		else {
		    //System.out.println("Adding folder " + folder.getPrimaryFile());
		    seen.add(folder);
		    Enumeration en = folder.children ();
		    while (en.hasMoreElements ()) {
			DataObject child = (DataObject)en.nextElement ();
			addDataObject(child);
		    }
		}
	    }
	}

	/** add a non-folder DataObject, if it's compilable */
	private void addScalarObjectToJob(DataObject dataObj) {
	    seen.add(dataObj);
	    CompilerCookie cc = (CompilerCookie)dataObj.getCookie(
	    			CompilerCookie.Compile.class);
	    if (cc != null) {
		//System.out.println("Scheduling compile of " + dataObj.getClass().getName() + " " + dataObj.getPrimaryFile());
		cc.addToJob(job, Compiler.DEPTH_INFINITE);
		if (dataObj instanceof JarDataObject) {
		    new Throwable().printStackTrace();
		}
	    }
	}

	CompilerJob getJob() {
	    return job;
	}
    }

	
    /** Creates compiler job for given jar content */
    public static CompilerJob createCompilerJob(
    				ArchiveDescriptor ad, DataObject source) {

	CompilerJobBuilder job = new CompilerJobBuilder();
	addToCompilerJob(job, ad, source);
	return job.getJob();
    }

    /** adds contents to an existing job */
    private static void addToCompilerJob(
	    CompilerJobBuilder job, ArchiveDescriptor ad, DataObject source) {

	if (source != null) {
	     if (!job.ignoreDataObject(source)) {
		// Already in the job
	     	return;
	    }
	}

	ContentMemberList members = ad.getContentList();

	for (int i=0; i<members.size(); i++){
	    DataObject dataObj = 
	    	((ContentMember)members.get(i)).getDataObject();
	    job.addDataObject(dataObj);
	}
    }
                                                                       
    
    /** Constructs default file name for archive content. File path is
     * the same as target archive file, except for extension, which is
     * taken from jar packager's options.
     */
    public static String defaultContentName (ArchiveDescriptor desc) {
        String targetPath = desc.getTargetResource().getFile();
        int index = targetPath.lastIndexOf("."); // NOI18N
        if (index < 0) {
            index = targetPath.length();
        }
        StringBuffer contentPath = new StringBuffer();
        contentPath.append(targetPath.substring(0, index));
        contentPath.append("."); // NOI18N
        contentPath.append(JarPackagerOption.singleton().getContentExt());
        return contentPath.toString();
    }

    /** Silly helper method, returns file created from url's getFile() method */
    public static File urlToFile (URL url) {
        return new File(url.getFile());
    }
    

}

/*
* <<Log>>
*  13   Gandalf   1.12        1/25/00  David Simonek   Various bugfixes and i18n
*  12   Gandalf   1.11        1/16/00  David Simonek   i18n
*  11   Gandalf   1.10        12/7/99  David Simonek   
*  10   Gandalf   1.9         11/11/99 David Simonek   add to jar action failure
*       repaired
*  9    Gandalf   1.8         11/9/99  David Simonek   bugfixes and new mount 
*       jar action
*  8    Gandalf   1.7         10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems copyright in file comment
*  7    Gandalf   1.6         10/14/99 David Simonek   manifest updating 
*       bugfixes
*  6    Gandalf   1.5         10/13/99 David Simonek   various bugfixes 
*       concerning primarily manifest
*  5    Gandalf   1.4         10/13/99 David Simonek   jar content now primary 
*       file, other small changes
*  4    Gandalf   1.3         10/10/99 Petr Hamernik   console debug messages 
*       removed.
*  3    Gandalf   1.2         10/5/99  David Simonek   various fixes, only 
*       primary entries now resides in 'chosen content'
*  2    Gandalf   1.1         10/4/99  David Simonek   
*  1    Gandalf   1.0         9/8/99   David Simonek   
* $
*/
