/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager.wizard;

import java.io.*;
import java.awt.event.ActionEvent;
import java.awt.Component;

import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.ResourceBundle;

import org.openide.WizardDescriptor;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.TemplateWizard;

import org.openide.cookies.SourceCookie;
import org.openide.execution.NbClassPath;
import org.openide.src.ClassElement;
import org.openide.src.SourceElement;
import org.openide.src.SourceException;

import org.netbeans.modules.jarpackager.*;

public class JarContentsWizardIterator implements TemplateWizard.Iterator {
    private static final long serialVersionUID = -1987345873459L;

    /* only one panel for now */
    private static final int PANEL_COUNT = 4;

    private transient WizardDescriptor.Panel[] panels;

    private transient String panelNames[];

    /** Index of the current panel. Panels are numbered from 0 to 
     * PANEL_COUNT - 1.  */
    private transient int panelIndex = 0;

    private static JarContentsWizardIterator instance = null;

    private transient TemplateWizard wizardInstance;
    
    public JarContentsWizardIterator() {
    }

    /** Returns JarContentsWizardIterator singleton instance.
     */
    public static synchronized JarContentsWizardIterator singleton() {
        if (instance == null) {
            instance = new JarContentsWizardIterator();
        }
        return instance;
    }

    /** Instantiates the template using informations provided by
     * the wizard.
     *
     * @param wiz the wizard
     * @return set of data objects that has been created (should contain
     *  at least one
     * @exception IOException if the instantiation fails
    */
    public java.util.Set instantiate(TemplateWizard wiz) throws IOException {
        // dig out file from the template.
        DataObject obj = instantiateTemplate(wiz.getTemplate(), wiz.getTargetFolder(), wiz.getTargetName());
	JarContent fromWizard = getJarContent(wiz);
	JarDataObject jdObj = (JarDataObject)obj;
	JarContent fromTemplate = jdObj.getJarContent();

	// Apply wizard-derived changed to template 
	for (int i = 0; i < PANEL_COUNT; i++) {
	    if (panels[i] instanceof JarContentEditor) {
	    	((JarContentEditor)panels[i]).applyChanges(
			fromTemplate, fromWizard);
	    }
	}

	jdObj.save();

        return java.util.Collections.singleton(obj);
    }

    public void initialize(TemplateWizard wizard) {
        this.wizardInstance = wizard;
        
	//System.out.println("Setting JarContent");
	JarDataObject template = ((JarDataObject)wizard.getTemplate());
	//setJarContent(wizard, template.getJarContent().makeCopy());
	setJarContent(wizard, new JarContent());
	if (panels == null) {
	    initialize();

            Component panel = wizard.targetChooser().getComponent();
            if (panel instanceof javax.swing.JComponent) {
                ((javax.swing.JComponent)panel).putClientProperty(
                    "WizardPanel_contentData", panelNames);
            }

	    panels = new WizardDescriptor.Panel[PANEL_COUNT];
	    panels[0] = wizard.targetChooser();
	    panels[1] = new ContentWizardPanel(1);
	    panels[2] = new LocationWizardPanel(2);
	    panels[3] = new ManifestWizardPanel(3);
	}

	panelIndex = 0;
    }

    /** Initializes the Wizard with array of panels.
     */
    protected void initialize() {
        if (panelNames == null) {
            panelNames = new String[PANEL_COUNT];
            panelNames[0] = getString("TIT_ChooseTarget");
            panelNames[1] = getString("TIT_Contents");
            panelNames[2] = getString("TIT_Location");
            panelNames[3] = getString("TIT_Manifest");
        }
    }

    private static ResourceBundle bundle;

    static String getString(String key) {
        if (bundle == null)
             bundle = org.openide.util.NbBundle.getBundle(
	     					JarContentsWizardIterator.class);
        return bundle.getString(key);
    }


    public void uninitialize(TemplateWizard wiz) {
	panels = null;
	panelNames = null;
	//System.out.println("Clearing JarContent");
	setJarContent(wiz, null);
    }

    /** Get the current panel.
     * @return the panel
    */
    public WizardDescriptor.Panel current() {
        return panels[panelIndex];
    }
    
    /** Get the name of the current panel.
     * @return the name
    */
    public String name() {
	return panelNames[panelIndex];
    }
    
    /** Test whether there is a next panel.
     * @return <code>true</code> if so
    */
    public boolean hasNext() {
        return panelIndex < PANEL_COUNT - 1;
    }
    
    /** Test whether there is a previous panel.
     * @return <code>true</code> if so
    */
    public boolean hasPrevious() {
        return panelIndex > 0;
    }
    
    /** Move to the next panel.
     * I.e. increment its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void nextPanel() {
	panelIndex++;
    }
    
    /** Move to the previous panel.
     * I.e. decrement its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void previousPanel() {
	panelIndex--;
    }

    public void addChangeListener (ChangeListener l) {
    }

    public void removeChangeListener (ChangeListener l) {
    }

    /** Instantiates the template. Currently it just delegates to the template DataObject's
     * createFromTemplate method.
     */
    private DataObject instantiateTemplate(DataObject tpl, DataFolder target, String name) throws IOException {
        return tpl.createFromTemplate(target, name);
    }

    private static final String JAR_CONTENT_PROPERTY = new String("jarContent");

    /**
     * Store jarContent in wizard
     */
    public static void setJarContent(TemplateWizard wiz, JarContent jc) {
	wiz.putProperty(JAR_CONTENT_PROPERTY, jc);
    }

    /**
     * Get jarContent from wizard
     */
    public static JarContent getJarContent(TemplateWizard wiz) {
    	return (JarContent)wiz.getProperty(JAR_CONTENT_PROPERTY);
    }

    /**
     * Get file for jar file location, as currently specified.
     */
    public static File getJarFile(TemplateWizard wizard) {
	JarDataLoader loader = 
	    (JarDataLoader)JarDataLoader.findObject(JarDataLoader.class);
	String extension = loader.getArchiveExt(); 
    	try {
	    File folder = 
		NbClassPath.toFile(wizard.getTargetFolder().getPrimaryFile());
	    //System.out.println("folder is " + folder);
	    return (new File(folder.getPath() + "/" + 
			     wizard.getTargetName() +
			      "." + extension));
	}
	catch (IOException ex) {
	    // Leave it null
	    return null;
	}
    }

    /**
     * Get full name for jar content file location, as currently specified.
     */
    public static String getJarContentFile(TemplateWizard wizard) {
	JarDataLoader loader = 
	    (JarDataLoader)JarDataLoader.findObject(JarDataLoader.class);
	String extension = loader.getExtension();

    	try {
	    String folder = 
	    	wizard.getTargetFolder().getPrimaryFile().getPackageName('/');
	    //System.out.println("folder is " + folder);
	    return folder + "/" + wizard.getTargetName() + "." + extension;
	}
	catch (IOException ex) {
	    // Leave it null
	    return null;
	}
    }

    /** interface for applying wizard changes to template-generated data  */
    interface JarContentEditor {

	/* Apply wizard-generated changed
	 * @param fromTemplate the template data
	 * @param fromWizard the wizard's changes
	 */
	void applyChanges(JarContent fromTemplate, JarContent fromWizard);
    }

}
