/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.jarpackager.wizard;

import java.lang.reflect.*;
import java.io.*;

import java.awt.Component;
import java.awt.Container;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.AbstractButton;
import javax.swing.JButton;

import org.openide.explorer.propertysheet.editors.FileOnlyEditor;
import org.openide.loaders.TemplateWizard;
import org.openide.util.NbBundle;
import org.openide.WizardDescriptor;

import org.netbeans.modules.jarpackager.JarContent;

/**
* Embed the location property editor in a wizard pane
* <p>
* We also turn off the cancel and select buttons,  since 
* they make no sense when the editor is embedded in a panel.
*/
class LocationWizardPanel 
	extends PropertyEditorPanel
	implements WizardDescriptor.FinishPanel,
	    JarContentsWizardIterator.JarContentEditor	{

    private FileOnlyEditor2 editor;
    private AbstractButton approveButton = null;

    /* true if user has chose a location other than the default */
    private boolean setByUser = false;

    /** Create panel
     * @param order order of this panel in wizard 
     */
    public LocationWizardPanel(int order) {
	super(new FileOnlyEditor2());

	editor = (FileOnlyEditor2)super.editor;
	JComponent editorComponent = editor.getJComponent();
	editorComponent.putClientProperty(
	    "WizardPanel_contentSelectedIndex", new Integer(order));
	editorComponent.setName(NbBundle.getBundle(LocationWizardPanel.class).
						getString("TIT_LocationPanel"));
	
	JFileChooser chooser = editor.getChooser();
	approveButton = findButton(chooser, chooser.getApproveButtonText());
	makeInvisible(chooser, approveButton);
    }

    /* make approve and cancel buttons invisible */
    private void makeInvisible(
    	JFileChooser chooser, AbstractButton approveButton) {
    	
	Class clazz = chooser.getClass();
	try {
	    Method meth = clazz.getMethod(
				    "setControlButtonsAreShown", //NOI18N
				    new Class[] {boolean.class});
	    meth.invoke(chooser, new Object[] {new Boolean(false)});
	}
	catch (Exception ex) {
	    // Method does not exist before JDK 1.3
	    // Since the only other version supported is 1.2.2, and we know the
	    // implemenation of JFileChooser there, we make the panel
	    // which holds both of the control buttons invisible.
	    approveButton.getParent().setVisible(false);
	}
    }

    	
    /* find button in Container */
    private AbstractButton findButton(Container comp, String label) {
	Component[]comps = comp.getComponents();
	for (int i = 0; i < comps.length; i++) {
	    if (comps[i] instanceof AbstractButton) {
	    	AbstractButton jb = (AbstractButton)comps[i];
		if (jb.getLabel() != null && jb.getLabel().equals(label)){
		    return jb;
		}
	    }
	    else if (comps[i] instanceof Container) {
	    	AbstractButton ab = findButton((Container)comps[i], label);
		if (ab != null)
		    return ab;
	    }
	}
	return null;
    }

    /** no further validation required */
    public boolean isValid() {
	return true;
    }

    /** called when initializing or reinitializing panel
     * @param settings the parent TemplateWizard
     */
    public void readSettings (Object settings) {
	TemplateWizard wizard = (TemplateWizard)settings;
	JarContent fromWizard  =
	    JarContentsWizardIterator.getJarContent(wizard);
	File current;
	if (setByUser) {
	    current = fromWizard.getTargetFile();
	}
	else {
	    // target may have changed, so reapply default
	    current = JarContentsWizardIterator.getJarFile(wizard);
	}
	editor.setValue(current);
    }

    /** called when exiting panel
     * @param settings the parent TemplateWizard
     */
    public void storeSettings(Object settings) {
	if (approveButton != null) {
	    // Kick the button, in case the user has typed a filename and not
	    // hit <return>
	    approveButton.doClick();
	}
	TemplateWizard wizard = (TemplateWizard)settings;
	JarContent fromWizard = 
	    JarContentsWizardIterator.getJarContent(wizard);
	File current = (File)editor.getValue();
	fromWizard.setTargetFile(current);
	File defaultJar = JarContentsWizardIterator.getJarFile(wizard);
	setByUser = !(current.getPath().equals(defaultJar.getPath()));
    }

    /** return associated awt component */
    public Component getComponent() {
	return editor.getJComponent();
    }

    /** apply changes to template-generated object */
    public void applyChanges(JarContent fromTemplate, JarContent fromWizard) {
    	File location = fromWizard.getTargetFile();
	if (location != null) {
	    fromTemplate.setTargetFile(location);
	}
    }

    /**
    * The FileEditor and its subclasses are not directly usable here.
    * The component returned by getComponent above needs to be a JComponent
    * for the wizard code to work, but FileEditor creates a JComponent both
    * too late and too often for our purposes -- we need there to be exactly one
    * associated with the panel.  Also, once the JComponent is created, 
    * FileEditor no longer processes setValue and getValue correctly.  
    * This subclass rectifies all of that,
    *
    */
    static class FileOnlyEditor2 extends FileOnlyEditor {

	/** Remeber out chooser */
	private JFileChooser chooser;

	/** Remember our component */
	private Component customEditor;

	/** Remeber our JComponent */
	private JComponent component;

	/* Create the custom editor immediately, and remeber its 
	 *  associated JComponent 
	 */
    	public FileOnlyEditor2() {
	    getCustomEditor();
	    component = ((JDialog)customEditor).getRootPane();
	}

	/* Cache the chooser, so we only create one*/
	protected JFileChooser createFileChooser () {	
	    if (chooser == null) {
	    	chooser = super.createFileChooser();
	    }
	    return chooser;
	}

	/* Cache the customEditor, so we only create one*/
	public Component getCustomEditor() {
	    if (customEditor == null) {
	    	customEditor = super.getCustomEditor();
	    }
	    return customEditor;
	}

	/* Set the value in the GUI as well as in the property editor */
	public void setValue(Object o) {
	    if (o instanceof File) {
	    	File f = (File) o;
		if (f != null) {
		    chooser.setCurrentDirectory (new File (f.getParent ()));
		    chooser.setSelectedFile (f);
		}
	    }
	    super.setValue(o);
	}

	/* Get the value from the GUI */
	public Object getValue() {
	    return (chooser == null) ?
	    		super.getValue() : chooser.getSelectedFile();
	}

	/** return the associated JComponent */
	public JComponent getJComponent() {
	    return component;
	}

	/** return the chooser */
	public JFileChooser getChooser() {
	    return chooser;
	}

    }
}

