package org.netbeans.modules.jarpackager;

import java.util.*;
import java.io.*;
import java.beans.*;

import org.openide.loaders.*;
import org.openide.nodes.*;
import org.netbeans.modules.jarpackager.api.*;

/**
* Simple example of Node.Property which is an integer-valued property.
* Note the division of labor between the three classes defined here.
*/

/**
* This is a factory object.  It knows how to create IntegerProperty objects
* whose values live in a Map indexed by property name.
*/
public class IntegerPropertyFactory implements ExtendedPropertyFactory {

    private String name;

    /**
     * @param name the name of the property which this factory will create.
     */
    public IntegerPropertyFactory(String name) {
    	this.name = name;
    }

    public String getName() {
    	return name;
    }

    /** create a Node.Property with the same name as the factory, whose value
     * lives in the given Map.
     */
    public Node.Property  createProperty(DataObject dataObj, JarContent jc) {
    	return new IntegerProperty(dataObj, name, jc);
    }

    /**
     * A property which the factory creates.  This example implements the
     * minimum required methods.
     */
    static class IntegerProperty extends Node.Property {

	private JarContent jc;
	private DataObject owner;

	/**
	 * create the property.  Note that each property sheet instance will
	 * use a different instance of this class.  Note also that this class's
	 * setValue and getValue methods access the property in the Map -- no
	 * local copy is created.
	 */
	public IntegerProperty(DataObject dataObj, String name, JarContent jarContent) {
	    super(Integer.class);
	    this.setName(name);
	    this.jc = jarContent;
	    this.owner = dataObj;
	}

	/** can the value of this property be read and thus displayed */
	public boolean canRead() {
	    return true;
	}

	/** can value of this this property be changed. */
	public boolean canWrite() {
	    return true;
	}

	/** get the value from the Map. */
	public Object getValue() {
	    Integer val = (Integer)jc.getExtendedProperty(this.getName());
	    if (val == null) {
	    	Integer defVal = new Integer(0);
		jc.setExtendedProperty(this.getName(), defVal);
		return defVal;
	    }
	    return val;
	}

	/** update the value in the map */
	public void setValue(Object val) {
            System.out.println(" jc = " + jc + " " + val);
            System.out.println(jc.getExtendedProperty(this.getName()));
	    jc.setExtendedProperty(this.getName(), val);
            System.out.println(jc.getExtendedProperty(this.getName()));
            if (owner != null) {
                owner.setModified(true);
            }
	}

	/** return an editor for this property */
	public PropertyEditor getPropertyEditor () {
	    return new Editor();
	}

	/** A basic editor for the property.  Note that it operates
	 * on a local copy of the data.  When updating the data in the Map is
	 * necessary, higher-level code will get the value from this editor
	 * and set it on the property.
	 */
	class Editor extends PropertyEditorSupport {

	    int value;

	    /** convert the local copy to a string */
	    public String getAsText() {
		return Integer.toString(value);
	    }

	    /** get the local copy */
	    public Object getValue() {
		return new Integer(value);
	    }

	    /** set the local copy from a string */
	    public void setAsText(String s) {
		value = Integer.parseInt(s);
	    }

	    /** set the local copy */
	    public void setValue(Object o) {
		value = ((Integer)o).intValue();
	    }
	    
	}
    }
}
