/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.util.Enumeration;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import org.openide.cookies.SourceCookie;

import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileSystemCapability;

import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;

import org.openide.src.*;

/**
 * This class implements Finder interface from org.openide.src and is used
 * to look up ClassElement relevant for the given name. The function is as follows:
 * The finder locates the first .java or .class file in the Repository, looking
 * at filesystems with capability COMPILE. If .java file (source) is found,
 * it is parsed and scanned for the element. The results are then returned.
 * <P>
 * If .class is located first, the scan continues and the finder queries all
 * matching .java files whether they know about the .class that was found first.
 * They might, because of cross-directory compilation.
 * <P>
 * If one of the .java sources claims ownership over the .class, the ClassElement
 * is dug out of the source file, otherwise the finder returns <CODE>null</CODE>
 * and leaves the work to other finders.
 *
 * @author  <A href="mailto:sdedic@netbeans.org">Svatopluk Dedic</A>
 * @version 0.1
 */
class ClassElementFinder implements ClassElement.Finder {
    /**
     * Enables tons of debug messages for tracing bugs and weird results.
     */
    private static final boolean DEBUG = false;
    
    private static ClassElementFinder finder;
    
    /**
     * Hidden to prevent construction from outside.
     */
    private ClassElementFinder() {
    }
    
    /**
     * Returns, creating it if necessary, an instance of Finder. The finder
     * is implemented as a singleton.
     */
    public static ClassElementFinder getInstance() {
        if (finder != null)
            return finder;
        synchronized (ClassElementFinder.class) {
            if (finder == null)
                finder = new ClassElementFinder();
        }
        return finder;
    }
    
    /**
     * @param fullName name as passed to ClassElement.forName. Must contain $
     * as inner class delimiters.
     * @param packageDelimiter basically position of the last . in the fullName,
     * or 0 if there is none.
     * @param innerDelimiter index of the first $ in the fullName. Must be > 
     *  packageDelimiter.
     */
    private ClassElement findSpecifiedInner(String fullName, int packageDelimiter,
        int innerDelimiter) {
        
        // if there's no . in the fullname, packageDelimiter is 0 and 
        String packName = fullName.substring(0, packageDelimiter);
        String topName = fullName.substring(packageDelimiter + 1, innerDelimiter);
        
        FileObject f = FileSystemCapability.COMPILE.find(packName, topName, 
            JavaDataLoader.JAVA_EXTENSION);

        if (DEBUG)
            System.err.println("Searching delimited file " + packName + "/" + topName);
        if (f == null)
            // no such file found...
            return null;

        StringTokenizer tt = new StringTokenizer(fullName.substring(packageDelimiter + 1),
            "$"); // NOI18N
        return findClassInFile(f, tt);
    }
    
    /**
     * Finds the specified ClassElement within the source, travesing through inner
     * classes. The Enumeration should contain Strings - beginning with the name
     * of the top-level class, followed by names of inner classes up to the one
     * the caller wants to locate.
     */
    private ClassElement findInSource(SourceElement s, Enumeration names) {
        Identifier classID;
        ClassElement c;
        
        classID = Identifier.create((String)names.nextElement());
        c = s.getClass(classID);
        if (DEBUG) {
            System.err.println("getClass(" + classID.getName() + ") = " + 
                (c == null ? "nothing" : c.getName().getFullName()));
        }
        if (!names.hasMoreElements() || c == null)
            return c;
        do {
            classID = Identifier.create((String)names.nextElement());
            c = c.getClass(classID);
            if (DEBUG) {
                System.err.println("getClass(" + classID.getName() + ") = " + 
                    (c == null ? "nothing" : c.getName().getFullName()));
            }
        } while (c != null && names.hasMoreElements());
        return c;
    }
    
    /**
     * Tries to find ClassElement with the corresponding name.
     */
    public ClassElement find(String className) {
        int lastDot = className.length() - 1;
        int innerDelimiter = -1;
        ClassElement result;

        if (lastDot < 0)
            return null;
            
        while (lastDot > 0 && className.charAt(lastDot) != '.') {
            if (className.charAt(lastDot) == '$')
                innerDelimiter = lastDot;
            lastDot--;
        }
        if (lastDot < 0)
            return null;
            
        if (DEBUG) {
            System.err.println("[ClassFinder] Searching for " + className + 
                " lastDot = " + lastDot + " delimiter = " + innerDelimiter);
        }
        
        // innerDelimiter > -1 --> specified an inner class.
        // we can find source with package className[0..lastDot - 1],
        // name className[lastDot + 1..innerDelimiter - 1]
        if (innerDelimiter > 0)
            return findSpecifiedInner(className, lastDot, innerDelimiter);

        int nameEnd = className.length();
        String baseName;
        String packName;
        FileObject f;

        baseName = className.substring(lastDot + 1, nameEnd);
        packName = className.substring(0, lastDot);
        while (true) {
            f = searchFile(packName, baseName);
            if (f != null) {
                if (DEBUG) {
                    System.err.println("Finally found file " + f);
                }
                String partName = className.substring(lastDot + 1);
                result = findClassInFile(f, new StringTokenizer(partName, "."));
                if (result != null)
                    return result;
            }
            // dive out from the package:
            if (lastDot <= 0)
                break;

            nameEnd = lastDot;
            lastDot = className.lastIndexOf('.', lastDot - 1);
            if (lastDot == -1) {
                packName = ""; // NOI18N
                baseName = className;
            } else {
                baseName = className.substring(lastDot + 1, nameEnd);
                packName = className.substring(0, lastDot);
            }
        }
        return null;
    }

    /**
     * Attempts to locate the named class within the file. The enumeration
     * should produce String components of the class name, beginning with
     * top-level class name, followed by inner class name(s), if applicable.
     */
    private ClassElement findClassInFile(FileObject f, Enumeration names) {
        DataObject d;
        SourceElement s;
        try {
            d = DataObject.find(f);
            
            SourceCookie sc = (SourceCookie)d.getCookie(SourceCookie.class);
            if (DEBUG)
                System.err.println("SourceCookie = " + sc);
            if (sc == null)
                return null;
            s = sc.getSource();
        } catch (DataObjectNotFoundException ex) {
            return null;
        }
        return findInSource(s, names);
    }
    
    /**
     * Tries to find the designated class. It searches FileSystems in the
     * order defined by the Repository. If it finds a .class file, it
     * does not give up immediately, but tries to locate a matching .java
     * file
     * @param packName name of the package (folder) to look in.
     */
    private FileObject searchFile(String packName, String baseName) {
        Enumeration fss = FileSystemCapability.COMPILE.fileSystems();
        
        FileObject classFile = null;
        Collection javaFiles = null;
        
        if (DEBUG) {
            System.err.println("Searching in " + packName + "/" + baseName);
        }
        while (fss.hasMoreElements()) {
            FileSystem fs = (FileSystem)fss.nextElement();
            if (classFile == null && javaFiles == null) {
                classFile = fs.find(packName, baseName, JavaDataLoader.CLASS_EXTENSION);
                if (DEBUG) {
                    System.err.println("Found: " + classFile);
                }
            }
            FileObject javaFile = fs.find(packName, baseName, JavaDataLoader.JAVA_EXTENSION);
            if (javaFile != null) {
                if (javaFiles == null) {
                    if (classFile == null)
                        // .java is first in compilation path.
                        return javaFile;                    
                    javaFiles = new LinkedList();
                }
                javaFiles.add(javaFile);
                if (DEBUG) {
                    System.err.println("Found: " + javaFile);
                }
            }
        }
        if (javaFiles == null) {
            // no source file... we don't bother with sourceless ones.
            return null;
        } else if (classFile == null) {
            return javaFiles == null ? null : (FileObject)javaFiles.iterator().next();
        }

        // now both classfile and javafile are != null, and java file comes
        // after classfile, or is in the same directory.
        for (Iterator it = javaFiles.iterator(); it.hasNext(); ) {
            FileObject javaFile = (FileObject)it.next();
            DataObject d;

            try {
                d = DataObject.find(javaFile);
            } catch (DataObjectNotFoundException ex) {
                return classFile;
            }
            // iterate through secondaries of the DataObject to find out whether
            // the .class is registered or not:
            if (d.files().contains(classFile))
                return javaFile;
        }
        // no .java claims ownership --> /dev/null.
        return null;
    }
    
    public ClassElement find(Class clazz) {
        return null;
    }
}
