/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.util.*;
import java.io.*;
import java.lang.ref.*;

import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem;
import org.openide.execution.NbClassPath;
import org.openide.ErrorManager;

/**   
 *
 * @author  Marek Sedliak
 * @version 
 */ 
public class ClassPathConfigurator extends Object {
    private static final boolean DEBUG = false;
    
    /** Creates new ClassPathConfigurator */
    private static final String PATH_SEPARATOR = System.getProperty("path.separator"); // NOI18N
    private static final String JAR_FILE_EXTENSION = ".jar"; // NOI18N
    private static final String ZIP_FILE_EXTENSION = ".zip"; // NOI18N

    /** Singleton reference */
    private static ClassPathConfigurator classPathConfigurator = null;
    
    /** Cache for FileSystems */
    private HashMap fileSystemCache = null;
    
    private static final Object TAG_INVALID = new Object();
    
    private ClassPathConfigurator() {
        fileSystemCache = new HashMap(10);
    }

    /** Getter for singleton.
     */
    public static ClassPathConfigurator getInstance() {
        if (classPathConfigurator == null) {
            synchronized (ClassPathConfigurator.class) {
                if (classPathConfigurator == null)
                    classPathConfigurator = new ClassPathConfigurator();
            }
        }
        return classPathConfigurator;
    }
    
    public Collection getFileSystems(NbClassPath classPath) {
	if (classPath == null)
	    return Collections.EMPTY_LIST;
        String cp = classPath.getClassPath();
        if (cp.length() >= 2 && cp.charAt(0) == '"') { // NOI18N
            cp = cp.substring(1, cp.length() - 1);
        }
        return getFileSystems(cp);
    }

    /** Path to FileSystem conversion method.
     */
    public java.util.Collection getFileSystems(String path) {
	if (path == null)
	    return Collections.EMPTY_LIST;
        return getCachedFileSystems(new StringTokenizer(path, PATH_SEPARATOR));
    }
    
    /**
     * Look up in the cache, or create if it is not in the cache, if the
     * filesystem is not known yet, a filesystem rooted at the given filename.
     * The argument must identify a directory or a compressed (zip, jar) file,
     * otherwise the method may throw IllegalArgumentException. The method may
     * throw an IOException if the root is not readable.
     */
    synchronized FileSystem  createFileSystem(String fsRootName) throws IOException {
        FileSystem fs = null;
        boolean report = true;
        if (fileSystemCache.containsKey( fsRootName )) {
            Object o = fileSystemCache.get(fsRootName);            
            if (o == TAG_INVALID) {
                fs = null;
                report = false;
            } else {
                fs = (FileSystem)((Reference)o).get();
            }
        }
        
        if ( fs == null ) {
            try {
                if ( isJarFile( fsRootName ) ) 
                    fs = getJarFileSystem( fsRootName );
                else 
                    fs = getLocalFileSystem( fsRootName );
            } catch (IOException ex) {
                if (report) {
                    fileSystemCache.put(fsRootName, TAG_INVALID);
                    throw ex;
                }
                return null;
            }
            /*
              huh ?
            if ( !fileSystemCache.containsKey( pathName ) ) 
                    fileSystemCache.remove( pathName ); 
             */
            fileSystemCache.put( fsRootName, new SoftReference( fs ) );
        }
        return fs;
    }

    /** Cache management method
     */
    private Vector getCachedFileSystems(Enumeration path) {
        Vector result = new Vector();
        if (path == null) return result;
        String pathName = null;
        FileSystem fs = null;
        
        while (path.hasMoreElements()) {
            pathName = (String)path.nextElement();
            if ((pathName != null) && (pathName.length()>0) ) {
                try {
                    fs = createFileSystem(pathName);
                } catch (IOException ex) {
                    Util.getErrorManager().log(ErrorManager.INFORMATIONAL, 
                        "WARNING: " + pathName + " cannot be opened");
                    Util.getErrorManager().log(ErrorManager.INFORMATIONAL,
                        ex.toString());
                }
                if (fs == null)
                    continue;
                if (!result.contains( fs )) result.add( fs );
            }
        }
        return result;
    }
 
    /** JarFileSystem creation method
     */
    private JarFileSystem getJarFileSystem(String jarFileName) throws IOException {
        JarFileSystem result = new JarFileSystem();
        try {
            if (DEBUG)
                System.err.println("[fscache] creating JARFS: " + jarFileName);
            result.setJarFile( new File( jarFileName ) );
        } catch (java.beans.PropertyVetoException e) {
            System.err.println("Unable to set Jar file "+jarFileName+". "+e.getMessage());
        }
        result.setHidden(true);
        return result;
    }

    /** LocalFileSystem creation method
     */
    private LocalFileSystem getLocalFileSystem(String dirName) throws IOException {
        LocalFileSystem result = new LocalFileSystem();
        try {
            if (DEBUG)
                System.err.println("[fscache] creating LocalFS: " + dirName);
            result.setRootDirectory( new File( dirName ) );
        } catch (java.beans.PropertyVetoException e) {
            System.err.println("Unable to set Root directory "+dirName+". "+e.getMessage());
        }

        result.setHidden(true);
        return result;
    }

    /** JarFile name testing
     */
    private boolean isJarFile(String fileName) {
        boolean result = false;
        if (fileName != null) 
            result = (fileName.endsWith( ZIP_FILE_EXTENSION ) || fileName.endsWith( JAR_FILE_EXTENSION ));
        return result;
    }
}
