/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;

import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.ProgressEvent;
import org.openide.compiler.ErrorEvent;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.filesystems.FileSystem;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.MultiDataObject;
import org.openide.cookies.CompilerCookie;

/**
*
* @author Ales Novak
*/
final class CleanCompiler extends Compiler {

    JavaDataObject jdo;
    CleanCompilerGroup group;
    FileSystem target;

    /**
    * @param job a compiler job
    * @param fo file object to compile (represents .java source)
    * @param constraint - COMPILE, BUILD, ...
    */
    public CleanCompiler(CompilerJob job, JavaDataObject jdo, FileSystem target) {
        super(job);
        this.jdo = jdo;
	this.target = target;
    }

    private boolean isCrossCompiling() {
	return target != null;
    }

    /** inherited */
    public boolean isUpToDate() {
	int size;

	if (isCrossCompiling()) {
	    int cleanables = getCleanables().size();
	    //System.err.println("Got " + cleanables + " cleanables.");
	    return cleanables == 0;
	} else {
	    return jdo.files().size() == 1;
	}
    }
    
    public FileSystem getFileSystem() {
	if (target != null)
	    return target;
	else {
	    try {
		return jdo.getPrimaryFile().getFileSystem();
    	    } catch (org.openide.filesystems.FileStateInvalidException ex) {
		return null;
	    }
	}
    }
    
    protected Collection getCleanables() {
	// include files placed besides sources by default.
	MultiDataObject.Entry entry;
	FileObject fo;
	Collection c = null;
	
	if (!isCrossCompiling()) {
	    Set files = jdo.secondaryEntries();
	    Iterator it = files.iterator();
	    c = new ArrayList(files.size());
	    while (it.hasNext()) {
		entry = (MultiDataObject.Entry)it.next();
		fo = entry.getFile();
		if (fo.getExt().equals(JavaDataLoader.CLASS_EXTENSION)) {
		    c.add(entry);
		}
	    }
	} else if (target.isValid()) {
	    //System.out.println("Target is " + (target == null ? "" : target.isValid() ? "valid" : "invalid"));
	    FileObject primary = jdo.getPrimaryFile();
	    String packName = primary.getParent().getPackageName('.'); // NOI18N
	    String baseName = primary.getName();
	    int baseLength = baseName.length();
	    
	    FileObject targetFolder = target.find(packName, null, null);
	    if (targetFolder != null) {
		FileObject[] children = targetFolder.getChildren();
		FileObject f;
		String n;

		//System.err.println("Searching target fs for " + packName + "/" + baseName);		
		for (int i = 0; i < children.length; i++) {
		    f = children[i];
		    if (f.isValid() && !f.isFolder() && JavaDataLoader.CLASS_EXTENSION.equals(f.getExt()) &&
			(n = f.getName()).startsWith(baseName) &&
			(n.length() == baseLength ||
			n.charAt(baseLength) == '$')) { // NOI18N
			if (c == null) {
			    c = new LinkedList();
			}
			//System.err.println("* found " + f);
			c.add(f);
		    }
		}
	    }
	}
	return c == null || c.size() == 0 ? java.util.Collections.EMPTY_LIST : c;
    }
    
    private void reportError(FileObject fo, IOException ex) {
	String message = java.text.MessageFormat.format(
	    JavaCompilerType.getString("MSG_CannotClean"),
	    new Object[] {
		fo,
		ex.getLocalizedMessage()
	    }
	);
	
	ErrorEvent evt = new ErrorEvent(group, jdo.getPrimaryFile(), 0, 0, message, null);
	group.ireErrorEvent(evt);
    }
    
    private boolean deleteClasses(JavaDataObject dobj) throws IOException {
	Collection c = getCleanables();
	Iterator it = c.iterator();
	MultiDataObject.Entry en;
	Object item;
	FileObject fo;
	FileLock lock;
	FileSystem myFS = jdo.getPrimaryFile().getFileSystem();
	boolean ok = true;

	//System.err.println("Got " + c.size() + " to clean");	
	while (it.hasNext()) {
	    item = it.next();
		if (item instanceof MultiDataObject.Entry) {
		    en = (MultiDataObject.Entry)item;
		    fo = en.getFile();
		} else {
		    en = null;
		    fo = (FileObject)item;
		} 
	    try {
		//System.err.println("Cleaning " + fo);
		lock = fo.lock();
	    } catch (java.io.IOException e) {
		ok = false;
		reportError(fo, e);
		continue;
	    }
	    group.ireProgressEvent(new ProgressEvent(group, fo, ProgressEvent.TASK_CLEANING));
	    try {
		// note: the file deletion is fired as an event that is intercepted by the DataObject
		// (if it exists) and the corresponding entry is removed. No need for a special accessor method.
                fo.delete(lock);
	    } finally {
		lock.releaseLock();
	    }
	    FileObject dir = fo.getParent();
	    if (target != myFS && dir.getFileSystem() == target) {
		while (!dir.isRoot() && dir.getChildren().length == 0) {
		    lock = null;
		    try {
			lock = dir.lock();
			//System.err.println("Deleting directory " + dir);
			dir.delete(lock);
		    } catch (IOException ex) {
			// silently ignore
			break;
		    } finally {
			if (lock != null)
			    lock.releaseLock();
		    }
		    dir = dir.getParent();
		}
	    }
	}
        if (ok) {
            // clear JDO's products attribute.
            jdo.clearProducts();
        }
	return ok;
    }

    public boolean equals (Object other) {
        if (!(other instanceof CleanCompiler))
            return false;
        return (jdo == ((CleanCompiler)other).jdo);
    }

    public int hashCode () {
        return ((jdo == null) ? 0 : jdo.hashCode());
    }

    /**
    */
    boolean compile()  {
        try {
            return deleteClasses(jdo);
        } catch (IOException ioe) {
            group.printThrowable(ioe);
            return false;
        }
    }//compile

    /** @return false used in JavaCompilerGroup */
    public final boolean getClearEnv() {
        return false;
    }

    /** inherited */
    public Class compilerGroupClass() {
        return CleanCompilerGroup.class;
    }

    public Object compilerGroupKey() {
        List l = new ArrayList(2);
        try {
            l.add(super.compilerGroupKey ());
            l.add(jdo.getPrimaryFile().getFileSystem());
        } catch (FileStateInvalidException e) { // could not happen
            if (Boolean.getBoolean("netbeans.debug.exceptions")) { // NOI18N
                e.printStackTrace();
            }
        }
        return l;
    }
}
