/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.java;

import java.beans.*;
import java.util.*;
import java.text.MessageFormat;

import org.openide.TopManager;
import org.openide.filesystems.*;

/** Implements a simple pull-down list of writable local filesystems.  
 *  @author Jesse Glick
 */
public class FileSystemPE extends PropertyEditorSupport {
    private FileSystem fs;

    /**
     * Tags to display in the property editor.
     */
    private String[] tags = null;

    /**
     * Filesystems that match the query.
     */
    private FileSystem[] fss = null;

    /** Message to display if the currently selected filesystem is invalid.
     */
    private static String INVALID_MSG = Util.getString ("LABEL_TargetInvalid");

    /**
     * Message to display if the target is not set == classes should be put right beside
     * their sources.
     */
    private static String DEFAULT_MSG = Util.getString ("LABEL_TargetDefault");

    /** Returns an array of filesystems usable for compilation.
     */
    private final FileSystem[] findUsableFilesystems() {
        Enumeration e = FileSystemCapability.COMPILE.fileSystems();
        LinkedList lst = null;
        
        while (e.hasMoreElements()) {
            FileSystem fs = (FileSystem)e.nextElement();
            if (canUseFilesystem(fs)) {
                if (lst == null)
                    lst = new LinkedList();
                lst.add(fs);
            }
        }
        return lst == null ? null : (FileSystem[])lst.toArray(new FileSystem[lst.size()]);
    }
    
    /** 
     * Checks whether the filesystem is writeable and supports COMPILE capability.
     */
    protected boolean canUseFilesystem(FileSystem fs) {
        return !(fs.isReadOnly() || fs.isDefault()) && !fs.isHidden() &&
            fs.isValid() &&
            fs.getCapability().capableOf(FileSystemCapability.COMPILE);
    }
    
    /** Find all writable local file systems.
     * They must also be valid, not the default system file system, and support compilation.
     * @return a list of all such found
     static LocalFileSystem[] findWritableLocals() {
            Enumeration e = FileSystemCapability.COMPILE.fileSystems ();
            List l = new ArrayList ();
            while (e.hasMoreElements ()) {
                FileSystem f = (FileSystem) e.nextElement ();
                if (f instanceof LocalFileSystem && f.isValid () &&
                        !f.isReadOnly () && !f.isDefault ()) {
                    l.add (f);
                }
            }
            return (LocalFileSystem[]) l.toArray (new LocalFileSystem[] { });
        }
     */

    private synchronized void updateTags() {
        fss = findUsableFilesystems ();
        tags = new String[fss.length + 1];
        for (int i = 0; i < fss.length; i++)
            tags[i + 1] = fss[i].getDisplayName ();
        tags[0] = DEFAULT_MSG;
    }

    public void setValue(Object o) {
        fs = (FileSystem)o;
        firePropertyChange ();
    }

    public Object getValue() {
        return fs;
    }

    public boolean supportsCustomEditor() {
        return false;
    }

    public synchronized String[] getTags() {
        updateTags ();
        return tags;
    }

    public String getAsText() {
        if (fs == null)
            return DEFAULT_MSG;
        
        if (fs == JavaCompilerType.TARGET_INVALID ||
            !fs.isValid())
            return INVALID_MSG;
        else
            return fs.getDisplayName ();
    }

    public synchronized void setAsText(String name) throws IllegalArgumentException {
        updateTags ();
        if (name.equals (DEFAULT_MSG)) {
            setValue (null);
            return;
        }
        
        // Look up the tag:
        for (int i = 0; i < tags.length; i++) {
            if (name.equals (tags[i])) {
                setValue (fss[i - 1]);
                return;
            }
        }
        throw (IllegalArgumentException)
            TopManager.getDefault().getErrorManager().annotate(
                new IllegalArgumentException ("No such fs:" + name),
                MessageFormat.format(Util.getString("MSG_InvalidTargetFS"),
                new Object[] {name})
            );
    }
    
    /**
     * More restrictive PropertyEditor that restricts the usable filesystems to the
     * local ones.
     */
    public static class Local extends FileSystemPE {
        protected boolean canUseFileSystem(FileSystem fs) {
    	    java.io.File f = org.openide.execution.NbClassPath.toFile(fs.getRoot());
	    return f != null && f.exists() && f.isDirectory();
        }
    }
}
