/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.io.IOException;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.text.FieldPosition;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.Iterator;

import org.openide.compiler.Compiler;
import org.openide.compiler.ExternalCompiler;
import org.openide.compiler.ExternalCompilerGroup;
import org.openide.execution.NbProcessDescriptor;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileSystemCapability;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;


/** 
*
* @author Ales Novak, Jaroslav Tulach
*/
public class JExternalCompilerGroup extends ExternalCompilerGroup {
  
    private final HashSet indirectCompilers;

    private final HashSet directCompilers;
  
    // for each specific JavaExternalCompilerType is created designed compiler group
    private JavaExternalCompilerType jext;
    
    /** Set of packages that need to be refreshed after the compilation.
     */
    private Set refreshPackages;
    
    private org.openide.filesystems.FileSystem targetFS;

    /** Create an external compiler group. */
    public JExternalCompilerGroup() {
        indirectCompilers = new HashSet(7);
        directCompilers = new HashSet(7);
    }

    /** Add a compiler */
    public void add(Compiler c) {
        if (c instanceof JavaCompilerType.IndirectCompiler) {
            indirectCompilers.add(c);
        } else {
            checkAndAdd((JExternalCompiler) c);
        }
    }

    /** Checks for duplicity and adds the compiler if it has not been added yet. */ 
    private void checkAndAdd(JExternalCompiler c) {
        if (c.isUpToDate()) {
            // do not add to the compile set.
            return;
        }

        if (directCompilers.add(c)) {
            // initialize settings:
            if (jext == null) {
                initializeOptions(c);
            }
	    if (refreshPackages != null) {
        	FileObject fo = c.getFileObject0();
        	FileObject folder = fo.getParent();
        	addPackage(folder);
	    }
            super.add(c);
        }  
    }
    
    private void addPackage(FileObject fo) {
        if (!jext.isCrossCompile())
            return;
        refreshPackages.add(fo.getPackageName('.'));
    }

    /** Start the group */
    public boolean start() {
        if (indirectCompilers.size() > 0) {
            JavaCompilerType.IndirectCompiler[] cs = 
                new JavaCompilerType.IndirectCompiler[indirectCompilers.size()];
            cs = (JavaCompilerType.IndirectCompiler[]) indirectCompilers.toArray(cs);

            for (int i = 0; i < cs.length; i++) {
                Compiler compiler = cs[i].getCompiler();
                if (compiler != null) {
                    checkAndAdd((JExternalCompiler) compiler);
                }
            }
        }
	
	try {
            if (jext == null) {
                // nothing to do
                return true;
            }
	    jext.assureValidTargetFS();
	} catch (IOException ex) {
	    String msg = java.text.MessageFormat.format(
		JavaCompilerType.getString("MSG_CompilerFatal"), // NOI18N
		new Object[] {
		    ex.getLocalizedMessage()
		}
	    );
	    fireErrorEvent(new org.openide.compiler.ErrorEvent(this, null, 0, 0, msg, null));
	    return false;
	}
        
        ExternalCompiler[] compilers = getAllCompilers();

        boolean success = super.start();

        if (refreshPackages != null)
            refreshTargets();
        refreshDistant();
        return success;
    }
    
    void initializeOptions(Compiler c) {
        jext = ((JExternalCompiler)c).getMyType();
        if (jext.isCrossCompile()) {
	    targetFS = jext.getTargetFileSystem();
            refreshPackages = new HashSet(17);
	}
    }

    void refreshDistant() {
        ExternalCompiler[] compilers = getAllCompilers();
        
        for (int i = 0; i < compilers.length; i++) {
            JExternalCompiler extComp = (JExternalCompiler)compilers[i];
            JavaDataObject d = null;
            
            try {
                DataObject dobj = DataObject.find(extComp.getFileObject0());
                d = (JavaDataObject)dobj.getCookie(JavaDataObject.class);
                if (d == null)
                    // unsupported feature :-)
                    continue;
                if (targetFS == null)
                    d.clearProducts();
                else
                    d.refreshDistantChildren(targetFS);
            } catch (IOException ex) {
                continue;
            }
        }
    }
    
    void refreshTargets() {
        Iterator it = refreshPackages.iterator();
	FileObject current, next;
	FileObject root = targetFS.getRoot();
	
        //System.err.println("Refreshing packages...");
	while (it.hasNext()) {
	    String pkg = (String)it.next();
	    StringTokenizer tukac = new StringTokenizer(pkg, "."); // NOI18N
	    for (current = root; current != null && tukac.hasMoreTokens(); current = next) {
		String piece = tukac.nextToken();
		next = current.getFileObject(piece);
		if (next == null) {
		    //System.err.println("Did not found " + current + "/" + piece + ", refreshing parent");
		    current.refresh();
		    if (tukac.hasMoreTokens()) {
			next = current.getFileObject(piece);
		    }
		}
	    }	    
	    if (current != null) {
    		//System.err.println("--refreshing " + current);
		current.refresh();	
	    }
	}
    }

    /** Allows subclasses to provide their own format for parsing
    * the arguments of NbProcessDescriptor contained in the
    * ExternalCompiler.
    * <P>
    * This implementation creates new format Format with settings
    * from NbClassPath.createXXXX and executes them in the provided
    * process descriptor.
    *
    * @param desc description of program to start
    * @param files the argument to compiler list of files to compile (or reference
    *   to the file with @files)
    * @return format to use for changing the command line of the compiler
    * @exception IOException if exec fails
    */
    protected Process createProcess(NbProcessDescriptor desc, String[] files) throws IOException {
        return desc.exec(new JFormat(files, jext, desc), jext.getEnvironmentVariables(), jext.getWorkingDirectory());
    }

    /** Default format that can format tags related to compilation. These include settings of classpath
    * (can be composed from repository, class path, boot class path and libraries) and 
    * putting somewhere list of files to compile.
    */
    public static class JFormat extends Format {
        /*
         * Tag replaced with -O 
         */
        public static final String TAG_OPTIMIZE = "optimize"; // NOI18N
        
        /** 
         * Tag replaced with -g 
         */
        public static final String TAG_DEBUGINFO = "debuginfo"; // NOI18N
        
        /** 
         * Tag replaced with -deprecation 
         */
        public static final String TAG_DEPRECATION = "deprecation"; // NOI18N
        
        /** 
         * Tag replaced with -encoding 
         */
        public static final String TAG_ENCODING = "encoding"; // NOI18N
        
        /** 
         * Tag replaced with {outputdir-tag-replace}{output-root}
         */
        public static final String TAG_OUTDIR = "outputdir"; // NOI18N
        
        /** 
         * Tag replaced with netbeans.home property value 
         */
        public static final String TAG_IDEHOME = "ide.home"; // NOI18N
        
        /** 
         * Tag replaced with netbeans.home property value 
         * @deprecated deprecated in favor of {@link #TAG_IDEHOME}
         */
        public static final String TAG_NBHOME = "netbeans.home"; // NOI18N
        
        /** Tag replaced with netbeans.home property value 
         * @deprecated The tag is deprecated in favor of {@link #TAG_IDEHOME}
         */
        public static final String TAG_FORTEHOME = "forte.home"; // NOI18N
        /** Tag replaced by "netbeans.home/lib/fastjavac.messages */
        public static final String TAG_MSGFILE = Util.getString("TAG_Msgfile"); // NOI18N
	/** Old property that contains full name for the fastjavac.
         * @deprecated */
        public static final String TAG_FASTJAVAC_BINARY_FULL = "fastjavacBinary"; // NOI18N
        
        /** Tag replaced by fastjavac's OS-wired name */
	public static final String TAG_FASTJAVAC_BINARY = "fastjavac.wired"; // NOI18N

        /**
         * Tag is replaced by "bootclasspath tag replace (?) + bootclasspathlist"
         */
        public static final String TAG_OPTION_BOOTCLASSPATH = "bootclasspath.opt"; // NOI18N

        /**
         * Replacement tag for a doublequote
         */
        public static final String TAG_QUOTE ="q";
        
        static final long serialVersionUID =-8630048324703405233L;

        /** My NbProcessDescriptor */
        private NbProcessDescriptor descriptor;

        /** All values for the paths takes from NbClassPath.createXXX methods.
        *
        * @param files files to compile
        */
        public JFormat (String[] files, JavaExternalCompilerType jext, NbProcessDescriptor descriptor) throws IOException {
            super (
                files,
                jext.getClassPath(),
                jext.getBootClassPath(),
                jext.getRepositoryPath(),
                jext.getLibraryPath()
            );

            this.descriptor = descriptor;

            java.util.Map map = getMap();

            map.put(TAG_OPTIMIZE, jext.getOptimize() ? jext.getOptimizeReplace() : ""); // NOI18N
            map.put(TAG_DEBUGINFO, jext.getDebug() ? jext.getDebuginfoReplace() : ""); // NOI18N
            map.put(TAG_DEPRECATION, jext.getDeprecation() ? jext.getDeprecationReplace() : ""); // NOI18N
            map.put(TAG_OUTDIR, createOutputReplace(jext));
            map.put(TAG_ENCODING, createEncodingReplace(jext));
            map.put(TAG_QUOTE, "\"");
            String nb = System.getProperty(TAG_NBHOME);
            String wiredName = FastJavacCompilerType.getWiredName();
            String wiredFullName = FastJavacCompilerType.getWiredFullName();

            nb = new java.io.File(nb).getCanonicalPath();
            map.put(TAG_FORTEHOME, nb);
            map.put(TAG_NBHOME, nb);
            map.put(TAG_IDEHOME, nb);
            map.put(TAG_MSGFILE, getMsgFile(nb));
            map.put(TAG_FASTJAVAC_BINARY_FULL, wiredFullName);
	    map.put(TAG_FASTJAVAC_BINARY, wiredName);
            map.put(TAG_OPTION_BOOTCLASSPATH, createBootClassPathReplace(jext));
        }

        /** Classpath optimalization */
        public StringBuffer format(Object obj, StringBuffer toAppend, FieldPosition pos) {
            StringBuffer ret = super.format(obj, toAppend, pos);

            String arg = ret.toString();

            int idx = arg.indexOf("-classpath");
            if (idx < 0) {
                return ret;
            }

            // further process
            idx += 11;
            if (idx >= arg.length()) {
                return ret;
            }

            int space = getClassPathEntriesEnd(idx, arg);
            if (space <= idx) {
                return ret;
            }

            String classpath = ret.substring(idx, space);
            HashMap map = new HashMap();
            StringBuffer newpath = new StringBuffer();
            StringTokenizer stoken = new StringTokenizer(classpath, java.io.File.pathSeparator);
            while (stoken.hasMoreTokens()) {
                String token = stoken.nextToken();
                if (map.put(token, token) == null) {
                    newpath.append(token);
                    newpath.append(java.io.File.pathSeparator);
                }
            }
            map = null;
            newpath.insert(0, arg.substring(0, idx));
            newpath.append(arg.substring(space));
            return newpath;
        }
        
        protected String createBootClassPathReplace(JavaExternalCompilerType t) {
            NbClassPath bcp = t.getBootClassPath();
            if (bcp == null)
                return ""; // NOI18N
            String classPath = bcp.getClassPath();
            if (!"".equals(classPath)) {
                return t.getBootClassPathReplace();
            } else
                return "";
        }

        /** Scans from idx to a break */
        private static int getClassPathEntriesEnd(int idx, String searched) {
            boolean slashed = false; // last char was '\'
            boolean inToken = false; // after a '"'
            final int slen = searched.length();

            for (int i = idx; i < slen; i++) {
                switch (searched.charAt(i)) {
                    case ' ' :
                        if (! inToken) {
                            return i;
                        } else {
                            // if (slashed) { ...error... }???
                            break;
                        }
                  case '\\' :
                      slashed = !slashed;
                      break;
                  case '"' :
                      if (!slashed) {
                          inToken = !inToken;
                      }
                      // fall through
                  default :
                      slashed = false;
                      break;
                }
            }

            return -1;
        }

        /** Creates -encoding <spec enc> */
        static String createEncodingReplace(JavaExternalCompilerType jext) {
            String enc = jext.getCharEncoding();
            if (enc == null || "".equals(enc))
                return "";
            return jext.getEncodingReplace() + " " + enc;
        }

        /** 
          * Creates -d outputdir replacement, or "" if no -d is requested.
          */
        static String createOutputReplace(JavaExternalCompilerType jext) {
            if (!jext.isCrossCompile()) {
		//System.err.println("No outdir requested.");
                return ""; // NOI18N
            }
            org.openide.filesystems.FileSystem fs = jext.getTargetFileSystem();
            if (fs == JavaExternalCompilerType.TARGET_INVALID) {
                // it will fail anyway.
                return ""; // NOI18N
            }
            StringBuffer sb = new StringBuffer(fs.getSystemName().length() + 4);
	    String localName;
	    boolean hasSpace;
            File localFile = NbClassPath.toFile(fs.getRoot());
            sb.append(jext.getDReplace());
	    localName = localFile.toString();
	    
	    hasSpace = localName.indexOf(' ') > -1;
	    if (hasSpace)
        	sb.append(" \"");
	    else
		sb.append(" ");
            sb.append(localName);
	    if (hasSpace)
		sb.append('"');
	    //System.err.println("Replacing {outdir} with " + sb.toString());
            return sb.toString();
        }
    }

    /** @return location of mmsgfile fo rfastjavac */
    static String getMsgFile(String nbhome) {
        StringBuffer sb = new StringBuffer(nbhome);
        sb.append(java.io.File.separatorChar);
        sb.append("lib"); // NOI18N
        sb.append(java.io.File.separatorChar);
        sb.append("fastjavac.messages"); // NOI18N
        String ret = sb.toString();
        if (ret.indexOf(' ') >= 0) {
            sb.insert(0, '"');
            sb.append('"');
            ret = sb.toString();
        } 
        return ret;
    }
}
