/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.lang.ref.*;
import java.lang.reflect.Field;
import java.util.ResourceBundle;
import java.io.File;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

import org.openide.filesystems.FileSystem;
import org.openide.filesystems.LocalFileSystem;

import org.openide.cookies.SaveCookie;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.compiler.CompilerType;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.Compiler;
import org.openide.util.NbBundle;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileSystemCapability;
import org.openide.filesystems.FileUtil;
import org.openide.cookies.CompilerCookie;

import sun.tools.java.Constants;

/**
*
* @author Jaroslav Tulach
*/
public abstract class JavaCompilerType extends CompilerType implements Constants, Cloneable {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -1426544702124944362L;

    protected static final int currentVersion = 1;
    
    protected int version;

    /** Constant for compilation. */
    public static final Object COMPILE = CompilerCookie.Compile.class;
    /** Constant for building. */
    public static final Object BUILD = CompilerCookie.Build.class;
    /** Constant for clean. */
    public static final Object CLEAN = CompilerCookie.Clean.class;

    /** bundle to obtain text information from */
    private static ResourceBundle bundle;

    public static final String PROP_F_DEPRECATION = "deprecation"; // NOI18N
    public static final String PROP_F_DEBUG = "debug";  // NOI18N
    public static final String PROP_F_OPTIMIZE = "optimize"; // NOI18N
    public static final String PROP_F_WARNINGS = "warnings"; // NOI18N
    
    public static final String PROP_CHAR_ENCODING = "charEncoding"; // NOI18N
    public static final String PROP_FLAGS = "flag"; // NOI18N
    public static final String PROP_TARGET_FILESYSTEM = "targetFileSystem"; // NOI18N

    private static final String INVALID_TARGET_NAME = "InvalidTarget"; // NOI18N
    
    private static final String DEFAULT_ENCODING = "null"; // NOI18N
    
    /** name of the classpath property */
    public static final String PROP_CLASSPATH = "classPath";  // NOI18N
    
    /** name of bootclasspath property */
    public static final String PROP_BOOTCLASSPATH = "bootClassPath"; // NOI18N

    /**
     * Special tag value for an invalid filesystem.
     */
    public static final FileSystem TARGET_INVALID = new LocalFileSystem();

    /** compiler flags */
    private int flag = F_WARNINGS | F_DEBUG_VARS | F_DEBUG_LINES | F_DEBUG_SOURCE;

    /** character encoding */
    private String charEncoding;

    /** Nonserializable reference to a target FileSystem or null.
     */
    private transient Reference targetFileSystem;
    
    /**
     * Serializable id of the target filesystem. Null if not set.
     */
    private String targetFSName;

    /** class path settings or null 
     */
    private NbClassPath classPath;

    /** boot class path or null 
     */
    private NbClassPath bootClassPath;
    
    /**
     * Bootclasspath obtained from the runtime environment.
     */
    private transient NbClassPath   defaultBootClassPath;
    
    private transient NbClassPath   defaultClassPath;

    /** Clones this compiler type so all settings are properly duplicated. The new CompilerType will have identifying name
     * cleared so it will not intervene with other ServiceTypes.
     */    
    public Object clone() throws CloneNotSupportedException {
	JavaCompilerType other = (JavaCompilerType)super.clone();
        
        // Since ServiceType does not directly support cloning, I have to use reflection
        // to unlink property change support the original:
        try {
            Field f = org.openide.ServiceType.class.getDeclaredField("supp"); // NOI18N
            f.setAccessible(true);
            f.set(other, null);
        } catch (IllegalAccessException ex) {
            throwCloneException(ex);
        } catch (NoSuchFieldException ex) {
            throwCloneException(ex);
        }
        
	other.setName(""); // NOI18N
	return other;
    }

    /** Getter for class path associated with this executor.
    */
    public NbClassPath getClassPath() {
        NbClassPath p = classPath;
        if (p != null)
            return p;
        return createDefaultClassPath();
    }

    /**
     * Creates ClassPath, if none is specified.
     */    
    protected NbClassPath createDefaultClassPath() {
        if (defaultClassPath != null) 
            return defaultClassPath;
        synchronized (this) {
            if (defaultClassPath == null)
                // empty classpath.
                defaultClassPath = new NbClassPath("");
        }
        return defaultClassPath;
    }

    /** Setter for class path for this executor.
    */
    public void setClassPath(NbClassPath path) {
        NbClassPath old;
        
        synchronized (this) {
            if (path == getClassPath() ||
                (path != null && path.equals(getClassPath())))
                // no change.
                return;
            
            old = classPath;
            classPath = path;
        }
        firePropertyChange(PROP_CLASSPATH, old, path); // NOI18N
    }
    
    /** Getter for boot class path.
    */
    public NbClassPath getBootClassPath() {
        NbClassPath bcp = bootClassPath;
        if (bcp != null)
            return bcp;

        return (bootClassPath == null ? getDefaultBootClassPath() : bootClassPath);
    }
    
    /**
     * Creates and returns bootclasspath from the runtime environment.
     */
    protected  NbClassPath getDefaultBootClassPath() {
        if (defaultBootClassPath != null)
            return defaultBootClassPath;
        synchronized (this) {
            if (defaultBootClassPath != null)
                return defaultBootClassPath;
            // make up a bootclasspath.
        }
        return defaultBootClassPath = new NbClassPath("");
    }

    /** Setter for boot class path.
    */
    public void setBootClassPath(NbClassPath path) {
        NbClassPath old;
        
        synchronized (this) {
            if (path == getBootClassPath() ||
                (path != null && path.equals(getBootClassPath())))
                // no change.
                return;

            old = bootClassPath;
            bootClassPath = path;
        }
        firePropertyChange(PROP_BOOTCLASSPATH, old, path); 
    }
    
    /**
     * Checks whether the source file object needs compilation with respect
     * to the passed target filesystem.
     * @param fo source FileObject
     * @param target filesystem for compilation output
     * @return true, if the file is up-to-date (a .class is generated and is
     * newer than the source.
     */
    public static boolean isUpToDate(FileObject fo, FileSystem target) {
        try {
            DataObject mdo = (DataObject) DataObject.find(fo);

            SaveCookie cookie = (SaveCookie) mdo.getCookie(SaveCookie.class);
            if (cookie != null) {
                return false;
            }

            FileObject primary = mdo.getPrimaryFile();
            FileObject folder = primary.getParent();

	    if (target != null) {
		folder = target.find(folder.getPackageName('.'), null, null);
		if (folder == null)
		    return false;
	    }
	    
            FileObject clazz = folder.getFileObject(primary.getName(), JavaDataLoader.CLASS_EXTENSION);
            if (clazz != null) {
                clazz.refresh();
                if (clazz.getSize() > 0 &&
                    clazz.lastModified().compareTo(primary.lastModified()) >= 0) {
                    return true;
                }
            }
        } catch (DataObjectNotFoundException e) {
        }
        return false;
    }

    private void throwCloneException(Throwable t) throws CloneNotSupportedException {
        org.openide.ErrorManager err = org.openide.TopManager.getDefault().getErrorManager();
        CloneNotSupportedException x = new CloneNotSupportedException();
        err.copyAnnotation(x, t);
        throw x;
    }

    public int flag() {
        return flag;
    }

    private void orFlag(int f) {
        int oldflag = flag;
        flag |= f;
        firePropertyChange(PROP_FLAGS, new Integer(oldflag), new Integer(flag));
    }

    private void andFlag(int f) {
        int oldflag = flag;
        flag &= f;
        firePropertyChange(PROP_FLAGS, new Integer(oldflag), new Integer(flag));
    }

    public void setWarnings(boolean x) {
        boolean old = (flag & F_WARNINGS) != 0 ;
        if (x) {
            if (!old) orFlag(F_WARNINGS);
        }
        else if (old) andFlag(~F_WARNINGS);
        if (x != old)
            firePropertyChange(PROP_F_WARNINGS, new Boolean(old), new Boolean(x));
    }

    public boolean getWarnings() {
        return (flag & F_WARNINGS) != 0;
    }

    public void setDebug(boolean x) {
        boolean old = (flag & (F_DEBUG_VARS | F_DEBUG_LINES | F_DEBUG_SOURCE)) != 0;
        if (x) {
            if (!old) orFlag((F_DEBUG_VARS | F_DEBUG_LINES | F_DEBUG_SOURCE));
        }
        else if (old) andFlag(~(F_DEBUG_VARS | F_DEBUG_LINES | F_DEBUG_SOURCE));
        if (x != old)
            firePropertyChange(PROP_F_DEBUG, new Boolean(old), new Boolean(x)); 
    }

    public boolean  getDebug() {
        return (flag & (F_DEBUG_VARS | F_DEBUG_LINES | F_DEBUG_SOURCE)) != 0;
    }

    public void setOptimize(boolean x) {
        boolean old = (flag & (F_OPT | F_OPT_INTERCLASS)) != 0;
        if (x) {
            if (!old) orFlag((F_OPT | F_OPT_INTERCLASS));
        }
        else if (old) andFlag(~(F_OPT | F_OPT_INTERCLASS));
        if (x != old)
            firePropertyChange(PROP_F_OPTIMIZE, new Boolean(old), new Boolean(x));
    }

    public boolean getOptimize () {
        return (flag & (F_OPT | F_OPT_INTERCLASS)) != 0;
    }

    public void setDeprecation(boolean x) {
        boolean old = (flag & F_DEPRECATION) != 0;
        if (x) {
            if (!old) orFlag(F_DEPRECATION);
        }
        else if (old) andFlag(~F_DEPRECATION);
        if (x != old)
            firePropertyChange(PROP_F_DEPRECATION, new Boolean(old), new Boolean(x));
    }

    public boolean getDeprecation () {
        return  (flag & F_DEPRECATION) != 0;
    }

    public void setDependencies(boolean x) {
        boolean old = (flag & F_DEPENDENCIES) != 0;
        if (x) {
            if (!old) orFlag(F_DEPENDENCIES);
        }
        else if (old) andFlag(~F_DEPENDENCIES);
        if (x != old)
            firePropertyChange("dependencies", new Boolean(old), new Boolean(x)); // NOI18N
    }

    public boolean getDependencies () {
        return  (flag & F_DEPENDENCIES) != 0;
    }

    /** sets new character encoding
    * @param enc is a new encoding
    */
    public void setCharEncoding(String enc) {
        if (enc != null) {
            enc = enc.trim();
            if (enc.equals("")) { // NOI18N
                enc = null;
            } else if (DEFAULT_ENCODING.equalsIgnoreCase(enc)) {
		enc = null;
	    }
        }
        String old = charEncoding;
        charEncoding = enc;
        firePropertyChange(PROP_CHAR_ENCODING, old, charEncoding);
    }

    /** returns character encoding
    * @return encoding, null is possible encoding
    */
    public String getCharEncoding() {
        return charEncoding;
    }

    /**
     * Sets the filesystem for storing generated .class files.
     * @param target name of a valid FileSystem that supports COMPILE capability.
     */    
    public void setTargetFileSystem(FileSystem target) {
        FileSystem fs = getTargetFileSystem();
        if (target == fs ||
            (target != null && fs != null &&
             target.equals(fs)))
             return;
        targetFileSystem = target == null ? null : new WeakReference(target);
        firePropertyChange(PROP_TARGET_FILESYSTEM, fs, target);
    } 
    
    /**
     * Checks whether the target FS is valid. It throws a IOException if it the compiler type
     * is not willing to use it for output.
     */
    public void assureValidTargetFS() throws IOException {
	FileSystem fs = getTargetFileSystem();
	
	if (fs == null)
    	    return;
	String key = null;
	
	if (fs == TARGET_INVALID) {
	    key = "MSG_FS_Invalid"; // NOI18N
	} else {
	    java.util.Enumeration en = org.openide.TopManager.getDefault().getRepository().fileSystems();
	    boolean match = false;
	    while (en.hasMoreElements()) {
		FileSystem xfs = (FileSystem)en.nextElement();
		if (xfs == fs) {
		    match = true;
		    break;
		}
	    }
	    if (!match) {
		key = "MSG_FS_NotMounted";
	    }
	}
	if (key == null) {
	    if (fs.isReadOnly()) {
		key = "MSG_FS_ReadOnly"; // NOI18N
	    } else
		return;
	}

	String msg = java.text.MessageFormat.format(
	    getString(key),
	    new Object[] {
		fs.getDisplayName()
	    }
	);
	
	throw (IOException)org.openide.TopManager.getDefault().getErrorManager().annotate(
	    new IOException(msg),
	    org.openide.ErrorManager.USER,
	    null, msg, null, null);
    }
    
    /**
     * Returns the filesystem on which the compiler will store the created files.
     * The target filesystem might not be set (and this function will return null. 
     * In that case, compiler will store the resulting files in the same folders 
     * as their respective sources.
     * @return name of target filesystem, or null if not set.
     */
    public FileSystem getTargetFileSystem() {
        FileSystem fs;
        
        if (targetFileSystem == null)
            return null;
        fs = (FileSystem)targetFileSystem.get();
        if (fs == null) {
            // invalidate the filesystem. The real instance is lost, so mark it as invalid at least.
            fs = TARGET_INVALID;
            targetFileSystem = new WeakReference(fs);
        }
        return fs;
    }
    
    /**
     * Determines whether we are compiling to a specified directory subtree.
     * @return true, if the target filesystem is set (may be invalid!)
     */
    public boolean isCrossCompile() {
        return targetFileSystem != null;
    }
    
    /** Prepare a data object for compilation.
    * Implementations should create an instance of a
    * suitable subclass of {@link Compiler}, passing
    * the compiler job to the constructor so that the job may
    * register the compiler.
    *
    * @param job compiler job to add compilers to
    * @param type the type of compilation task to manage
    * ({@link org.openide.cookies.CompilationCookie.Compile}, etc.)
    * @param obj data object to prepare for compilation
    */
    public void prepareJob(CompilerJob job, Class type, DataObject obj) {
        JavaDataObject jobj = (JavaDataObject) obj;
        if (type == CLEAN) {
            new CleanCompiler(job, jobj, getTargetFileSystem());
        } else {
            Compiler compiler;
            FileObject sourceFile = jobj.getPrimaryFile();
            
            try {
                compiler = createCompiler(type, sourceFile);
                if (type == BUILD) {
                    compiler.dependsOn(new CleanCompiler(job, jobj, getTargetFileSystem()));
                }
            } catch (IllegalArgumentException ex) {
                // the compiler does not permit this file to be compiled.
                // create a bogus compiler that only prints out an error message
                // for the file.
                compiler = new BogusCompiler(sourceFile, ex);
            }
            job.add(compiler);
        }
    }
    
    /** Prepare a data object for compilation.
    * Implementations should create an instance of a
    * suitable subclass of {@link Compiler}, passing
    * the compiler job to the constructor so that the job may
    * register the compiler.
    *
    * @param job compiler job to add compilers to
    * @param type the type of compilation task to manage
    * ({@link org.openide.cookies.CompilationCookie.Compile}, etc.)
    * @param fo file object to prepare for compilation
    */
    public void prepareJobForFileObject(CompilerJob job, Class type, FileObject fo) {
        if (type == CLEAN) {
            throw new IllegalArgumentException();
        } else {
            Compiler c = createCompiler(type, fo);
            job.add(c);
        }
    }

    /** Prepare a data object for compilation.
    * Implementations should create an instance of a
    * suitable subclass of {@link Compiler}, passing
    * the compiler job to the constructor so that the job may
    * register the compiler.
    *
    * @param type the type of compilation task to manage
    * ({@link org.openide.cookies.CompilationCookie.Compile}, etc.)
    * @param fo file in wich to find the resource
    * @param resource file to compile
    */
    public IndirectCompiler prepareIndirectCompiler(Class type, FileSystem fs, String resource) {
        return new IndirectCompiler(type, fs, resource);
    }

    /** Prepare a data object for compilation.
    * Implementations should create an instance of a
    * suitable subclass of {@link Compiler}, passing
    * the compiler job to the constructor so that the job may
    * register the compiler.
    *
    * @param type the type of compilation task to manage
    * ({@link org.openide.cookies.CompilationCookie.Compile}, etc.)
    * @param fo file in wich to find the resource
    * @param resource file to compile
    */
    public IndirectCompiler prepareIndirectCompiler(Class type, String resource) {
        return new IndirectCompiler(type, null, resource);
    }

    /** @return compiler group class */
    protected abstract Class getCompilerGroupClass();

    /** @return compiler for given classtype and file object */
    protected abstract Compiler createCompiler(Class type, FileObject fo);
    
    /** @return localized name */
    static String getString(String name) {
        if (bundle == null) {
            bundle = NbBundle.getBundle(JavaCompilerType.class);
        }
        return bundle.getString(name);
    }
    
    private Object readObject(ObjectInputStream stm) throws IOException, ClassNotFoundException {
        stm.defaultReadObject();
        if (targetFSName == null)
            return this;
        
        FileSystem target = TARGET_INVALID;
        
        if (!INVALID_TARGET_NAME.equals(targetFSName)) {
            java.util.Enumeration en = org.openide.TopManager.getDefault().getRepository().fileSystems();
            while (en.hasMoreElements()) {
                FileSystem fs = (FileSystem)en.nextElement();
                if (fs.getSystemName().equals(targetFSName)) {
                    target = fs;
                    break;
                }
            }
        }
	/*
        if (version < currentVersion) {
            // reset bootclasspath and classpath to the default values:
            this.classPath = null;
            this.bootClassPath = null;
            this.version = currentVersion;
        }
	*/
        targetFileSystem = new WeakReference(target);
        return this;
    }
    
    private void writeObject(ObjectOutputStream stm) throws IOException {
        FileSystem fs = getTargetFileSystem();
        if (fs == null) {
            targetFSName = null;
        } else if (fs == TARGET_INVALID) {
            targetFSName = INVALID_TARGET_NAME;
        } else {
            targetFSName = fs.getSystemName();
        }
        stm.defaultWriteObject();
    }

    public final class IndirectCompiler extends Compiler {

        private final FileSystem dir;
        private final String res;
        private final Class type;

        private FileObject resolvedResource;

        public IndirectCompiler(Class type, FileSystem dir, String res) {
            this.dir = dir;
            this.type = type;
            if ((res != null) && (File.separatorChar != '/')) {
                res = res.replace(File.separatorChar, '/');
            }
            this.res = res;
        }

        public boolean isUpToDate() {
            return false;
        }

        public Class compilerGroupClass() {
            return getCompilerGroupClass();
        }

        public boolean equals(Object other) {
            return (this == other);
        }

        public int hashCode() {
            return (res == null) ? 4371234 /* any number */ : res.hashCode();
        }

        public Compiler getCompiler() {
            FileObject fo = getResolved();
            if (fo == null) {
                return null;
            } else {
                return createCompiler(type, fo);
            }
        }

        public FileObject getResolved() {
            if (resolvedResource == null) {
                if (res != null) {
                    if (dir == null) {
                        resolvedResource = FileSystemCapability.COMPILE.findResource(res);
                    } else {
                        resolvedResource = dir.findResource(res);
                    }
                }
            }
            return resolvedResource;
        }

        public void setResolved(FileObject fo) {
            resolvedResource = fo;
        }
    }
}
