/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.util.*;
import java.text.DateFormat;
import java.io.*;

import org.openide.filesystems.FileObject;
import org.openide.loaders.*;
import org.openide.actions.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;
import org.openide.util.MapFormat;

import org.netbeans.modules.java.settings.JavaSettings;

/** Data loader which recognizes Java source files.
*
* @author Petr Hamernik
*/
public class JavaDataLoader extends MultiFileLoader {
    /** The standard extension for Java source files. */
    public static final String JAVA_EXTENSION = "java"; // NOI18N

    /** The standard extension for Java class files. */
    public static final String CLASS_EXTENSION = "class"; // NOI18N
    
    public static final String PROP_PARSER_ENGINE = "parserEngine";
    
    public static final String PROP_EXTENSIONS = "extensions";

    /** Name of "perfectRecognition" property. */
    public static final String PROP_PERFECT_RECOGNITION = "perfectRecognition"; // NOI18N

    /** Class name of the default parsing engine. The engine is required to implemented
	org.netbeans.modules.java.ParserEngine interface.
    */
    public static final String DEFAULT_PARSER_ENGINE = "org.netbeans.modules.java.gj.JavaParserEngine";
    
    private static final int EXTERNALIZED_VERSION = 1;
    
    /** The list of parsing listener - the instance is in this class, because JavaDataLoader
    * class is prevented to be garbage collected.
    */
    static ArrayList parsingListeners = new ArrayList();
    
    static final long serialVersionUID =-6286836352608877232L;

    private transient ParserEngine	parserEngine;

    /** Create the loader.
    * Should <em>not</em> be used by subclasses.
    */
    public JavaDataLoader() {
        this(JavaDataObject.class);
    }

    /** Create the loader from a subclass.
    * @param recognizedObject the class of data object
    *        recognized by the loader
    */
    public JavaDataLoader(Class recognizedObject) {
        super(recognizedObject);
    }
    
    /**
     * @deprecated Items initialized in initialize() have moved to {@link #defaultActions}
     * and {@link #defaultDisplayName}; this method is no-op to maintain
     * compatibility.
     */
    protected void initialize() {
        // do nothing.
    }

    private SystemAction[] createDefaultActions() {
        return new SystemAction[] {
                       SystemAction.get (OpenAction.class),
                       SystemAction.get (CustomizeBeanAction.class),
                       SystemAction.get (FileSystemAction.class),
                       null,
                       SystemAction.get (CompileAction.class),
                       null,
                       SystemAction.get (BuildAction.class),
                       null,
                       SystemAction.get (ExecuteAction.class),
                       null,
                       SystemAction.get (CutAction.class),
                       SystemAction.get (CopyAction.class),
                       SystemAction.get (PasteAction.class),
                       null,
                       SystemAction.get (NewAction.class),
                       SystemAction.get (DeleteAction.class),
                       SystemAction.get (RenameAction.class),
                       null,
                       SystemAction.get (SaveAsTemplateAction.class),
                       null,
                       SystemAction.get (ToolsAction.class),
                       SystemAction.get (PropertiesAction.class)
                   };
    }
    
    protected String defaultDisplayName() {
        return Util.getString("PROP_JavaLoader_Name");
    }
    
    static SystemAction[] standardActions;
    
    protected SystemAction[] defaultActions() {
        if (standardActions != null)
            return standardActions;
        synchronized (JavaDataLoader.class) {
            if (standardActions == null) {
                standardActions = createDefaultActions();
            }
        }
        return standardActions;
    }

    /** Create the <code>JavaDataObject</code>.
    * Subclasses should rather create their own data object type.
    *
    * @param primaryFile the primary file
    * @return the data object for this file
    * @exception DataObjectExistsException if the primary file already has a data object
    */
    protected MultiDataObject createMultiObject (FileObject primaryFile)
    throws DataObjectExistsException, java.io.IOException {
        return new JavaDataObject(primaryFile, this);
    }

    /** For a given file find the primary file.
    * Subclasses should override this, but still look for the {@link #JAVA_EXTENSION},
    * as the Java source file should typically remain the primary file for the data object.
    * @param fo the file to find the primary file for
    *
    * @return the primary file for this file or <code>null</code> if this file is not
    *   recognized by this loader
    */
    protected FileObject findPrimaryFile (FileObject fo) {
	// never recognize folders.
        if (fo.isFolder()) return null;

        String ext = fo.getExt();
        if (ext.equals(CLASS_EXTENSION)) {
            return Util.findBrother(fo, getExtensions());
        }
        if (getExtensions().isRegistered(fo))
            return fo;
        return null;
    }

    /** Create the primary file entry.
    * Subclasses may override {@link JavaDataLoader.JavaFileEntry} and return a new instance
    * of the overridden entry type.
    *
    * @param primaryFile primary file recognized by this loader
    * @return primary entry for that file
    */
    protected MultiDataObject.Entry createPrimaryEntry (MultiDataObject obj, FileObject primaryFile) {
        primaryFile.setImportant(true);
        return new JavaFileEntry(obj, primaryFile);
    }

    /** Create a secondary file entry.
    * By default, {@link FileEntry.Numb} is used for the class files; subclasses wishing to have useful
    * secondary files should override this for those files, typically to {@link FileEntry}.
    *
    * @param secondaryFile secondary file to create entry for
    * @return the entry
    */
    protected MultiDataObject.Entry createSecondaryEntry (MultiDataObject obj, FileObject secondaryFile) {
        secondaryFile.setImportant(false);
        return new FileEntry.Numb(obj, secondaryFile);
    }

    /** Create the map of replaceable strings which is used
    * in the <code>JavaFileEntry</code>. This method may be extended in subclasses
    * to provide the appropriate map for other loaders.
    * This implementation gets the map from the Java system option;
    * subclasses may add other key/value pairs which may be created without knowledge of the
    * file itself.
    *
    * @return the map of string which are replaced during instantiation
    *        from template
    */
    protected Map createStringsMap() {
        Map map = ((JavaSettings) JavaSettings.findObject(JavaSettings.class, true)).getReplaceableStringsProps();
        map.put("DATE", DateFormat.getDateInstance(DateFormat.LONG).format(new Date())); // NOI18N
        map.put("TIME", DateFormat.getTimeInstance(DateFormat.SHORT).format(new Date())); // NOI18N
        return map;
    }
    
    /** Instructs JavaDataLoader to use new parser engine instead of the old one.
	Old requests are still processed using the engine that was in effect during their creation.
	New requests will be processed and registered with the newly specified engine.
    */
    public void setParserEngine(ParserEngine eng) {
	if (eng == this.parserEngine) {
	    return;
	}
	ParserEngine old;
	synchronized (this) {
	    old = parserEngine;
	    parserEngine = eng;
	}
	old.unregisterAll();
	firePropertyChange(PROP_PARSER_ENGINE, old, eng);
    }
    
    /** Returns instance of parser engine to use.
	The method returns the object set by {@link #setParserEngine}. If no engine was set, it constructs
	instance of class designated by DEFAULT_PARSER_ENGINE and returns it.
	The method might return null -- in that case, parsing is impossible.
    */
    public ParserEngine getParserEngine() {
	synchronized (this) {
	    if (parserEngine != null) {
		return parserEngine;
	    }
	    boolean ok = false;
	    Exception ex = null;
	    
	    try {
		parserEngine = (ParserEngine)Class.forName(DEFAULT_PARSER_ENGINE).newInstance();
		ok = true;
	    } catch (IllegalAccessException e) {
		ex = e;
	    } catch (InstantiationException e) {
		ex = e;
	    } catch (ClassNotFoundException e) {
		ex = e;
	    }
	    if (!ok && ex != null && Boolean.getBoolean("netbeans.debug.exceptions")) {
		ex.printStackTrace();
	    }
	    return parserEngine;
	}
    }

    /** @return The list of extensions this loader recognizes
    * (the default contains only java)
    */
    public ExtensionList getExtensions () {
	ExtensionList extensions = (ExtensionList)getProperty(PROP_EXTENSIONS);
	if (extensions == null) {
            extensions = new ExtensionList();
            extensions.addExtension(JAVA_EXTENSION);
	    putProperty(PROP_EXTENSIONS, extensions, false);
        }
        return extensions;
    }

    /** Sets the extension list for this data loader.
    * @param ext new list of extensions.
    */
    public void setExtensions(ExtensionList ext) {
        putProperty(PROP_EXTENSIONS, ext, true);
    }
    
    /** The "perfect recognition" option, if enabled, makes the Loader to parse and
     * look into .class files to determine, whether there more top-level classes that
     * are implemented by a single source. Setting this option on may severely impact
     * loading folders with .class files without accompanying sources since every file
     * needs to be opened and parsed.
     * @param b turn the feature on (true) or off (false).
     */
    public void setPerfectRecognition(boolean b) {
        Boolean prev = (Boolean)getProperty(PROP_PERFECT_RECOGNITION);
        if ((prev == null && !b) ||
            (prev != null && prev.booleanValue() == b)) {
            // no change needed.
            return;
        }
        putProperty(PROP_PERFECT_RECOGNITION, b ? Boolean.TRUE : Boolean.FALSE, true);
    }

    /** Returns whether the loader parses .class files to find out what source the .class
     * belongs to. See {@link #setPerfectRecognition} for discussion.
     *
     * @return <tt>true</tt> iff deep parsing of class files should be done
     */
    public boolean isPerfectRecognition() {
        Boolean v = (Boolean)getProperty(PROP_PERFECT_RECOGNITION);
        if (v == null)
            return false;
        return v.booleanValue();
    }

    public void writeExternal(ObjectOutput out) throws IOException {
	super.writeExternal(out);
	out.writeInt(EXTERNALIZED_VERSION);
	out.writeObject(getProperty(PROP_EXTENSIONS));	
    }
    
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
	super.readExternal(in);
	if (in.available() > 0) {
	    int v;
	    
	    v = in.readInt();
	    putProperty(PROP_EXTENSIONS, in.readObject(), false);
	}
    }
    
    /** This entry defines the format for replacing text during
    * instantiation the data object.
    * Used to substitute keys in the source file.
    */
    public class JavaFileEntry extends IndentFileEntry {
        static final long serialVersionUID =8244159045498569616L;
        
        /**
         * If true, the Entry refuses to open InputStream to prevent races
         * between readers and attempts to delete the file.
         */
        boolean disableInputStream;
        
        /**
         * Holds a collection of readers that read the file.
         */
        private Collection  activeReaders;

        /** Creates new entry. */
        public JavaFileEntry(MultiDataObject obj, FileObject file) {
            super(obj, file);
        }

        /** Provide suitable format for substitution of lines.
        * Should not typically be overridden.
        * @param target the target folder of the installation
        * @param n the name the file will have
        * @param e the extension the file will have
        * @return format to use for formating lines
        */
        protected java.text.Format createFormat (FileObject target, String n, String e) {
            Map map = createStringsMap();

            modifyMap(map, target, n, e);

            JMapFormat format = new JMapFormat(map);
            format.setLeftBrace("__"); // NOI18N
            format.setRightBrace("__"); // NOI18N
            format.setCondDelimiter("$"); // NOI18N
            format.setExactMatch(false);
            return format;
        }

        /** Modify the replacement map.
        * May be extended in subclasses to provide additional key/value
        * pairs sensitive to the details of instantiation.
        * @param map the map to add to
        * @param target the destination folder for instantiation
        * @param n the new file name
        * @param e the new file extension
        */
        protected void modifyMap(Map map, FileObject target, String n, String e) {
            map.put("NAME", n); // NOI18N
            // Yes, this is package sans filename (target is a folder).
            map.put("PACKAGE", target.getPackageName('.')); // NOI18N
            map.put("PACKAGE_SLASHES", target.getPackageName('/')); // NOI18N
	    // Fully-qualified name:
	    if (target.isRoot ()) {
		map.put ("PACKAGE_AND_NAME", n); // NOI18N
		map.put ("PACKAGE_AND_NAME_SLASHES", n); // NOI18N
	    } else {
		map.put ("PACKAGE_AND_NAME", target.getPackageName ('.') + '.' + n); // NOI18N
		map.put ("PACKAGE_AND_NAME_SLASHES", target.getPackageName ('/') + '/' + n); // NOI18N
	    }
            // No longer needed due to #6025. (You can just put in quotes, they will not
            // prevent things from being escaped.) But leave the token here for
            // compatibility with old templates. --jglick 26/08/00
            map.put("QUOTES","\""); // NOI18N
        }
        
        public synchronized void addReader(InputStream r) {
            if (activeReaders == null) {
                activeReaders = new LinkedList();
            }
            activeReaders.add(r);
        }
        
        public synchronized void removeReader(InputStream r) {
            if (activeReaders == null)
                return;
            activeReaders.remove(r);
        }
        
        public void delete() throws IOException {
            synchronized (this) {
                if (activeReaders != null && activeReaders.size() > 0) {
                    for (Iterator it = activeReaders.iterator(); it.hasNext(); ) {
                        InputStream r = (InputStream)it.next();
                        r.close();
                        it.remove();
                    }
                }
                activeReaders = null;
                disableInputStream = true;
            }
            super.delete();
        }
        
        public InputStream getInputStream() throws FileNotFoundException {
            FileObject fob = getFile();
            synchronized (this) {
                if (disableInputStream) {
                    // refuse to create the stream.
                    throw new FileNotFoundException("File is being deleted."); // NOI18N
                }
                InputStream s = new NotifyInputStream(fob.getInputStream());
                addReader(s);
                return s;
            }
        }
        
        private class NotifyInputStream extends FilterInputStream {
            public NotifyInputStream(InputStream is) {
                super(is);
            }
            
            public void close() throws IOException {
                super.close();
                removeReader(this);
            }
        }
    }
}
