/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
 
package org.netbeans.modules.java;

import java.io.*;
import java.util.*;
import java.beans.*;

import javax.swing.text.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import org.openide.TopManager;
import org.openide.cookies.*;
import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem; // override java.io.FileSystem
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.Task;
import org.openide.actions.OpenAction;

import org.openide.text.CloneableEditorSupport;
import org.openide.text.EditorSupport;

import org.openide.execution.NbClassLoader;
import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.CompilerType;
import org.openide.compiler.ExternalCompiler;
import org.openide.src.*;
import org.openide.src.nodes.SourceChildren;
import org.openide.src.nodes.SourceElementFilter;
import org.openide.src.nodes.ElementNodeFactory;
import org.openide.src.nodes.FilterFactory;
import org.openide.util.actions.SystemAction;
import org.openide.util.*;
import org.openide.util.Utilities;

import org.netbeans.modules.java.codegen.DocumentBinding;
import org.netbeans.modules.java.codesync.ModelEventAdapter;
import org.netbeans.modules.java.parser.JavaParser;
import org.netbeans.modules.java.model.LangModel;
import org.netbeans.modules.java.codesync.SynchronizeCodeCookie;
import org.netbeans.modules.java.codesync.SourceConnectionSupport;

import org.netbeans.modules.java.settings.JavaSettings;

/** Data object representing a Java source file, perhaps with associated class files.
* May be subclassed.
* @author Petr Hamernik, Jaroslav Tulach, Ian Formanek
*/
public class JavaDataObject extends MultiDataObject implements ElementCookie, CookieSet.Factory {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -6035788991669336965L;

    public static final byte CONNECT_NOT = 0;
    public static final byte CONNECT_CONFIRM = 1;
    public static final byte CONNECT_AUTO = 2;

    public static final byte CONNECT_DEFAULT = CONNECT_CONFIRM;

    private static final String EA_CONNECTION = "SourceSync"; // NOI18N
    
    /*
     * @deprecated
     */
    public static final String EA_PRODUCTS = "Products";
    
    private static final String[] SKIP_REGISTRATION_PREFIX = { "java.", "javax."}; // NOI18N

    /**
     * Holds a reference to editor support class. Lazy initialized from getJavaEditor,
     * or a getter for EditorSupport.class cookie.
     */
    transient private JavaEditor editorSupport;
    
    /**
     * Helper object that handles execution bits for us. Lazy initialized from
     * {@link #getExecSupport}.
     */
    transient private ExecSupport execSupport;
    
    /**
     * Helper object that manages connection source mechanics. Lazy initialized from
     * {@link #initializeConnectionSupport}
     */
    transient private ConnectionSupport connectionSupport;

    transient private JavaInstanceSupport instanceSupport;

    transient protected AbstractNode alteranteParent;
    
    /** WeakListener that intercepts external modification while the file is not
     *  opened in the Editor.
     */
    transient private FileChangeListener fileChangeListener;

    /** Hook to keep the WeakListener alive for the lifetime of the DataObject
     */
    transient private FileChangeListener fileChangeListenerHook;
    
    /** TEMPORARY: Filename of the primary file before it was changed (DO was
     * renamed or moved.
     */
    transient private String previousFileName;
    
    /** Options affecting behaviour of the DataObject
     */
    private static JavaSettings jopts;
    
    private transient boolean initialized;
    
    /**
     * Holds a reference to the glue between the parser and the dataobject.
     * Lazy initialized from {@link #initializeParsingSupport}
     */
    private transient JavaParserGlue   parserGlue;

    /**
     * Reference to an adapter and ConnectionSupport implementation, that 
     * adapts model events to JavaConnection events.
     */
    private transient ModelEventAdapter connectionAdapter;
    
    transient SourceConnectionSupport connectionManager;

    /** Create new data object.
    * @param pf primary file object for this data object (i.e. the source file)
    * @param loader the associated loader
    */
    public JavaDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException {
        super(pf, loader);
        init();
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        init();
    }

    private void init() {
        MultiDataObject.Entry entry = getPrimaryEntry();
        CookieSet cookies = getCookieSet();

        PrimaryListener l = new PrimaryListener();
        fileChangeListenerHook = l;
	fileChangeListener = WeakListener.fileChange(fileChangeListenerHook, entry.getFile());
	
	addPropertyChangeListener(l);

        Class[] cookieClasses = new Class[] { 
            CompilerCookie.Build.class, 
            CompilerCookie.Clean.class, 
            CompilerCookie.Compile.class,
	    CompilerSupport.class,
            ExecSupport.class,
            JavaEditor.class,
            ConnectionCookie.class,
            JavaParser.class,
            SourceCookie.Editor.class,
            SynchronizeCodeCookie.class,
        };
        cookies.add(cookieClasses, this);
        
        try {
            refreshDistantChildren(null);
        } catch (IOException ex) {
        }
        
        instanceSupport = new JavaInstanceSupport(entry);

        initialized = true;

        updateInstanceCookie();
        cookies.addChangeListener(new ChangeListener() {
                                      public void stateChanged(ChangeEvent evt) {
                                          // PENDING: fire cookies change on 
                                          // individual Elements.
                                      }
                                  });
        addSourceChangeListener(null);
    }
    
    /**
     * Creates a parsing support.
     */
    private JavaParser initializeParsingSupport() {
        if (parserGlue != null)
            return parserGlue.getParser();
        synchronized (this) {
            if (parserGlue != null) {
                return parserGlue.getParser();
            }

            parserGlue = new JavaParserGlue(getPrimaryEntry());
            if (editorSupport != null) {
                parserGlue.cloneableSupportCreated(extractCloneableEditor(editorSupport));
            }
            
            // initialize connection support -- this can be delayed up to the first commit,
            // if needed, can be captured by a CommitListener.
            LangModel m = parserGlue.getParser().getModel();
            if (connectionSupport == null)
                initializeConnectionSupport(m);
            else
                connectionAdapter.attachToModel(m);
            
            // attach management of incoming dependencies.
            connectionManager = new PersistentConnectionSupport(
                parserGlue.getSource(), m);
            connectionManager.resume(false);
            if (isModified()) {
                connectionManager.syncConnections();
            }
        }
        return parserGlue.getParser();
    }
    
    /**
     * Initializes support for JavaConnections - if a parser is not yet created,
     * it begins its live unattached to any model.
     */
    private ConnectionSupport initializeConnectionSupport(LangModel m) {
        if (connectionAdapter != null)
            return connectionAdapter;
        synchronized (this) {
            if (connectionAdapter != null)
                return connectionAdapter;
            connectionAdapter = new ModelEventAdapter(getPrimaryEntry());
            if (m != null) {
                connectionAdapter.attachToModel(m);
            }
        }
        return connectionAdapter;
    }
    
    public void setValid(boolean v) throws PropertyVetoException {
        super.setValid(v);
        if (!v) {
            synchronized (this) {
                if (parserGlue != null) {
                    parserGlue.suspendDocumentChanges();
                    connectionAdapter.detachFromModel();
                    connectionManager.suspend();
                }
            }
        }
    }
    
    /**
     * Finds a cloneableEditorSupport that holds the displayable/parseable source. The implementation
     * currently uses a hack, that extract package-private member from the EditorSupport's
     * implementation.
     */
    protected CloneableEditorSupport findCloneableEditorSupport() {
        EditorSupport supp = (EditorSupport)getCookie(EditorSupport.class);
        return extractCloneableEditor(supp);
    }

    /**
     * Extract CloneableEditorSupport instance from the given EditorSupport.
     * The implementation is fragile as it uses package-private member and reflection.
     */
    private CloneableEditorSupport extractCloneableEditor(EditorSupport supp) {
        try {
            java.lang.reflect.Field f = EditorSupport.class.getDeclaredField("del");
            f.setAccessible(true);
            return (CloneableEditorSupport)f.get(supp);
        } catch (Exception x) {
        }
        return null;
    }
    
    void updateInstanceCookie() {
        CookieSet cookies = getCookieSet();

        FileObject clazz = Util.findBrother(getPrimaryFile(), JavaDataLoader.CLASS_EXTENSION);
        JavaInstanceSupport prev = (JavaInstanceSupport) cookies.getCookie(InstanceCookie.class);

        if ((prev == null) && (clazz != null)) {
            cookies.add(instanceSupport);
        }
        else if ((prev != null) && (clazz == null)) {
            cookies.remove(instanceSupport);
        }
    }

    /** Attaches a file change listener to the primary (source) file.
     * Optionally removes the listener from previously used file object.
     * @param previousPrimary if not null, then the method removes change listener from this
     *        file object.
     */
    void addSourceChangeListener(FileObject previousPrimary) {
      if (previousPrimary != null) {
        previousPrimary.removeFileChangeListener(fileChangeListener);
      }
      getPrimaryEntry().getFile().addFileChangeListener(fileChangeListener);
    }

    // =================== Connections ====================================

    public byte getSynchronizationType() {
        Object o = getPrimaryFile().getAttribute(EA_CONNECTION);
        return ((o != null) && (o instanceof Byte)) ?
               ((Byte)o).byteValue() : CONNECT_DEFAULT;
    }

    public void setSynchronizationType(byte type) {
        FileObject fo = getPrimaryFile();
        try {
            if ((type != CONNECT_CONFIRM) && (type != CONNECT_NOT) && (type != CONNECT_AUTO))
                throw new IllegalArgumentException();
            fo.setAttribute(EA_CONNECTION, (type == CONNECT_DEFAULT) ? null : new Byte(type));
        }
        catch (IOException e) {
        }
    }

    int getAllListenersMask() {
        //    System.out.println("getAllListenersMask()"+JavaConnections.SETTINGS.isEnabled()); // NOI18N
        if (!JavaConnections.SETTINGS.isEnabled()) // disabled
            return 0;

        Iterator it = getConnectionSupport().getRegisteredTypes().iterator();
        int mask = 0;
        while (it.hasNext()) {
            Object o = it.next();
            /*      System.out.println("nextType:"+o);
                  System.out.println("   first:"+o.getClass().getClassLoader());
                  System.out.println("  second:"+JavaConnections.Type.class.getClassLoader());*/ // NOI18N
            if (o instanceof JavaConnections.Type) {
                mask |= ((JavaConnections.Type)o).getFilter();
            }

            // temporary-begin-------
            mask |= JavaConnections.IMPLEMENTS.getFilter();
            // end ------------------

        }
        return mask;
    }

    /**
     * @deprecated The method will be soon reimplemented in the new connection support
     * classes in the codesync package.
     */
    void startChangeProcessors(LinkedList changeProcessors) {
        if (changeProcessors.size() > 0) {
            if (getSynchronizationType() == CONNECT_AUTO) {
                Iterator it = changeProcessors.iterator();
                while (it.hasNext()) {
                    JavaConnections.ChangeProcessor p = (JavaConnections.ChangeProcessor) it.next();
                    try {
                        p.process();
                    } catch (SourceException e) {
                        // eat the exception
                        // TopManager.getDefault().notify(new NotifyDescriptor.Exception(e));
                    }
                }
            }
            else { // CONNECT_CONFIRM
                final byte oldSynch = getSynchronizationType();
                final JavaConnections.SyncRequest request = 
                    JavaConnections.scheduleShowChangesDialog(changeProcessors, oldSynch);
                request.getTask().addTaskListener(new org.openide.util.TaskListener() {
                    public void taskFinished(Task t) {
                        byte newSynch = request.getSyncType();

                        t.removeTaskListener(this);
                        if (oldSynch != newSynch)
                            setSynchronizationType(newSynch);
                    }
                });
            }
        }
    }

    /**
     * @deprecated Please see the new implementation of the connections in
     * {@link org.netbeans.modules.java.codesync.SourceConnectionSupport} and related
     * classes.
     */
    protected boolean connectionNotify(ConnectionCookie.Event event, LinkedList changeProcessors) {
       
        if (event.getType().getEventClass().isAssignableFrom(JavaConnections.Event.class)) {
            // there are two main sources of notification: external notifications that something this source
            // depends on has changed, and explicit notifications that this source should check itself for changes.
            JavaConnections.Event evt = (JavaConnections.Event)event;
            boolean selfcheck = evt.getType().overlaps(new JavaConnections.Type(JavaConnections.TYPE_SOURCE_CHECK_DEEP | JavaConnections.TYPE_SOURCE_CHECK_SELF));
            if (selfcheck) {
                return InterfaceConnection.sourceCheck(changeProcessors, getSource());
            } else {
                return InterfaceConnection.synchronizeInterfaces((JavaConnections.Event) event, changeProcessors, getSource());
            }
        }
        
        return false;
    }

    /**
     * @deprected The registration implementation was removed from the JavaDataObject.
     * See {@link org.netbeans.modules.java.codesync.SourceConnectionSupport} class for
     * the new implementation.
     */
    protected void registerForName(Identifier id, JavaConnections.Type type) {
        // do nothing
    }

    /**
     * @deprected The registration implementation was removed from the JavaDataObject.
     * See {@link org.netbeans.modules.java.codesync.SourceConnectionSupport} class for
     * the new implementation.
     */
    protected void unregisterForName(Identifier id, JavaConnections.Type type) {
        // do nothing
    }

    /** Set whether this object is considered modified.
    * @param modif <code>true</code> to consider it modified
    */
    public void setModified(boolean modif) {
        super.setModified(modif);
        if (modif && this.connectionManager != null) {
            connectionManager.syncConnections();
        }
    }

    /** Starts the task which update source element
    * depending on the new name and location of the JavaDataObject.
    * @param jdo The javadataobject which should be updated
    * @param dest The current primary file object
    * @param oldName The previous name of the object
    * @param save If document should be saved after updating
    */
    static void updateSourceStart(final JavaDataObject jdo,
                                  final FileObject dest,
                                  final String oldName, final boolean save) {
        
        JavaParser parser = (JavaParser)jdo.getCookie(JavaParser.class);

        Task task = parser.parse(JavaParser.PRIORITY_DEMAND, false, true);
        task.waitFinished();
        
        updateSource(jdo, dest, oldName, save);
        // prevent GC
        task.isFinished();
    }

    /** Update package and name of mail class in the source
    * depending on the new name and location of the JavaDataObject.
    * @param jdo The javadataobject which should be updated
    * @param dest The current primary file object
    * @param oldName The previous name of the object
    * @param save If document should be saved after updating
    */
    static void updateSource(final JavaDataObject jdo, FileObject dest, String oldName, boolean save) {
        boolean resumed = true;
        
        try {
            if (!jdo.isValid() || !dest.isValid()) {
                    // The data object does not exist; it had been probably deleted/moved
                    // before the parser came here.
                    return;
            }
            final JavaParser p = (JavaParser)jdo.getCookie(JavaParser.class);

            resumed = false;
            // suspend everything that is bound to the model and document - save time.
            jdo.suspendSupports();            
            SourceElement src = p.getSource();
            
            boolean wasLoaded = jdo.getJavaEditor().isDocumentLoaded();
            boolean changed = false;

            // update package
            FileObject packFO = dest.getParent();
            Identifier oldPackage = src.getPackage();
            Identifier newPackage = (packFO.isRoot()) ? null : Identifier.create(packFO.getPackageName('.'));
            if ((oldPackage != newPackage) ||
                    (oldPackage != null) && !oldPackage.equals(newPackage)) {
                src.setPackage(newPackage);
            }

            // update class rename
            ClassElement[] classes = src.getClasses();
            if (!dest.getName().equals(oldName)) {
                for (int i = 0; i < classes.length; i++) {
                    if (classes[i].getName().getName().equals(oldName)) {
                        StringBuffer fullName = new StringBuffer(packFO.getPackageName('.'));
                        if (fullName.length() > 0)
                            fullName.append('.');
                        fullName.append(dest.getName());
                        classes[i].setName(Identifier.create(fullName.toString(), dest.getName()));
                        break;
                    }
                }
                //impl.synchronizeEnabled = false;
            }
            if (save) {
                SaveCookie savec = (SaveCookie) jdo.getCookie(SaveCookie.class);
                if (savec != null) {
                    savec.save();
                }
                if (!wasLoaded) {
                    jdo.getJavaEditor().close();
                }
                /*
                // Schedules a task, IF the hierarchy is not clean (should it be done ??)
                p.parse(JavaParser.PRIORITY_BACKGROUND, false, true).addTaskListener(
                new TaskListener() {
                    public void taskFinished(Task t) {
                        t.removeTaskListener(this);
                        */                        
                        /*
                    }
                });
                */
            }
        }
        catch (SourceException e) {
        }
        catch (IOException e) {
        } finally {
            if (!resumed) {
                jdo.resumeSupports();
            }    
        }
    }
    
    public void resumeSupports() {
        connectionAdapter.attachToModel(parserGlue.getParser().getModel());
        connectionManager.resume(false);
        parserGlue.resumeDocumentChanges();
    }

    public void suspendSupports() {
        // suspend everything that is bound to the model and document
        parserGlue.suspendDocumentChanges();
        connectionAdapter.detachFromModel();
        connectionManager.suspend();
    }
    
    boolean isJavaFileReadOnly() {
	FileObject primary = getPrimaryFile();
	if (!isValid() || !primary.isValid()) {
	    return true;
	}
	return primary.isReadOnly();
    }

    // ==================== Handle methods =======================

    /** Copies primary file to the target DataFolder.
     * Since the operation does not rename the object, update source is only scheduled if
     * the target folder's name forms a valid Java identifier. 
    */
    protected synchronized DataObject handleCopy (final DataFolder df) throws IOException {
        DataObject obj = super.handleCopy(df);
        if (obj instanceof JavaDataObject) {
            String oldName = getPrimaryFile().getName();
            JavaDataObject jdo = (JavaDataObject)obj;
            String packageName = df.getPrimaryFile().getPackageName('.');
            if (isValidPackageName(packageName))
                updateSourceStart(jdo, jdo.getPrimaryFile(), oldName, true);
        }
        return obj;
    }

    /* Renames all entries and changes their files to new ones.
    */
    protected FileObject handleRename (String name) throws IOException {
        if (!Utilities.isJavaIdentifier(name))
            throw new IOException(NbBundle.getMessage(JavaDataObject.class, "FMT_Not_Valid_FileName", name));

        FileObject oldFo = getPrimaryFile();
        previousFileName = oldFo.getName();
        FileObject fo = super.handleRename(name);
        return fo;
    }

    /* Moves primary and secondary files to a new folder.
     * May ask for user confirmation before overwriting.
     * @param df the new folder
     * @return the moved primary file object
     * @throws IOException if there was a problem moving
     * @throws UserCancelException if the user cancelled the move
    */
    protected FileObject handleMove (DataFolder df) throws IOException {
        FileObject oldFo = getPrimaryFile();
        previousFileName = oldFo.getName();
        FileObject fo = super.handleMove(df);
        return fo;
    }

    /* Creates new object from template.
    * @exception IOException
    */
    protected DataObject handleCreateFromTemplate (DataFolder df, String name) throws IOException {
        if (name == null) {
            // special case: name is null (unspecified or from one-parameter createFromTemplate)
            name = FileUtil.findFreeFileName(df.getPrimaryFile(),
                getPrimaryFile().getName(), "java");
        } else if (!Utilities.isJavaIdentifier(name))
            throw new IOException(NbBundle.getMessage(JavaDataObject.class, "FMT_Not_Valid_FileName", name));

        DataObject obj = super.handleCreateFromTemplate(df, name);
        if (obj instanceof JavaDataObject) {
            String oldName = getPrimaryFile().getName();
            JavaDataObject jdo = (JavaDataObject)obj;
            String packageName = df.getPrimaryFile().getPackageName('.');
            if (isValidPackageName(packageName))
                updateSourceStart(jdo, jdo.getPrimaryFile(), oldName, true);
        }
        return obj;
    }

    /** Create the editor support for this data object.
    * By default, creates a <code>JavaEditor</code> with the source file entry;
    * subclasses may override this.
    * @return the editor support
    */
    protected JavaEditor createJavaEditor () {
        JavaEditor je = new JavaEditor (getPrimaryEntry ());        
        return je;
    }

    /* Help context for this object.
    * @return help context
    */
    public org.openide.util.HelpCtx getHelpCtx () {
        return new org.openide.util.HelpCtx (JavaDataObject.class);
    }

    /** Provide node that should represent this data object.
    * This implementation creates and returns a {@link JavaNode}.
    * Subclasses may wish to return a specialized subclass of this node type.
    * You should probably make the default action be {@link OpenAction}.
    * @return the node representation for this data object
    */
    protected Node createNodeDelegate () {
        JavaNode node = new JavaNode (this);
        node.setDefaultAction (SystemAction.get (OpenAction.class));
        return node;
    }

    /** Get the parsed representation of this source file.
    * May not be fully parsed yet; the source element itself indicates its status. 
    * @return the source element for this Java source file
    */
    public SourceElement getSource() {
        return ((JavaParser)getCookie(JavaParser.class)).getSource();
    }
    
    /** Get the current editor support.
    * Ought not be subclasses; use {@link #createJavaEditor}.
    * @return the editor support
    */
    public JavaEditor getJavaEditor() {
        if (editorSupport == null) {
            synchronized (this) {
                editorSupport = createJavaEditor();
                if (parserGlue != null)
                    parserGlue.cloneableSupportCreated(extractCloneableEditor(editorSupport));
            }
        }
        return editorSupport;
    }

    /**
     * Returns or creates the exec support.
     */
    public ExecSupport getExecSupport() {
        if (execSupport == null) {
            execSupport = new JavaExecSupport(getPrimaryEntry());
        }
        return execSupport;
    }

    public ConnectionSupport getConnectionSupport() {
        return initializeConnectionSupport(null);
    }

    static JavaSettings getJavaSettings() {
        if (jopts == null) {
            jopts = (JavaSettings) JavaSettings.findObject(JavaSettings.class, true);
        }
        return jopts;
    }
    
    /** @deprecated No longer used. Override {@link #createCookie} if you want to return
     * your own JCompilerSupport descendants.
     */
    protected Compiler createCompiler(CompilerJob job, Class type) {
        return null;
    }

    /** @deprecated Use compiler JCompilerSupport instead.
     */
    public static Compiler createCompilerForFileObject(Compiler[] dependencies,
            CompilerJob job,
            FileObject fo,
            Class type) {
        return null;
    }

    // =============== implementation of ElementCookie ========================

    /*
     * Get the alternate node representation.
     * @return the node
     * @see org.openide.loaders.DataObject#getNodeDelegate
    */
    public Node getElementsParent () {
        if (alteranteParent != null) return alteranteParent;
        synchronized (this) {
            if (alteranteParent != null) return alteranteParent;

            ElementNodeFactory cef = getBrowserFactory();
            SourceChildren sourceChildren = new SourceChildren (cef);
            SourceElementFilter sourceElementFilter = new SourceElementFilter();
            sourceElementFilter.setAllClasses (true);
            sourceChildren.setFilter (sourceElementFilter);
            sourceChildren.setElement (getSource ());

            alteranteParent = new AbstractNode (sourceChildren);
            CookieSet cs = alteranteParent.getCookieSet();
            cs.add (sourceChildren);
            return alteranteParent;
        }
    }

    // =============== The mechanism for regeisteing node factories ==============

    private static ArrayList explorerFactories = new ArrayList();
    private static ArrayList browserFactories = new ArrayList();

    static {
        explorerFactories.add( JavaElementNodeFactory.DEFAULT );

        JavaElementNodeFactory cef = new JavaElementNodeFactory();
        cef.setGenerateForTree (true);

        browserFactories.add( cef ) ;
    }

    public static void addExplorerFilterFactory( FilterFactory factory ) {
        addFactory( explorerFactories, factory );
    }

    public static void removeExplorerFilterFactory( FilterFactory factory ) {
        removeFactory( explorerFactories, factory );
    }

    public static ElementNodeFactory getExplorerFactory( ) {
        return (ElementNodeFactory)explorerFactories.get( explorerFactories.size() - 1);
    }

    public static void addBrowserFilterFactory( FilterFactory factory ) {
        addFactory( browserFactories, factory );
    }

    public static void removeBrowserFilterFactory( FilterFactory factory ) {
        removeFactory( browserFactories, factory );
    }

    public static ElementNodeFactory getBrowserFactory() {
        return (ElementNodeFactory)browserFactories.get( browserFactories.size() - 1 );
    }

    private static synchronized void addFactory( List factories, FilterFactory factory ) {
        factory.attachTo( (ElementNodeFactory)factories.get( factories.size() - 1 ) );
        factories.add( factory );
    }

    private static synchronized void removeFactory( List factories, FilterFactory factory ) {
        int index = factories.indexOf( factory );

        if ( index <= 0 )
            return;
        else if ( index == factories.size() - 1 )
            factories.remove( index );
        else {
            ((FilterFactory)factories.get( index + 1 )).attachTo( (ElementNodeFactory)factories.get( index - 1 ) );
            factories.remove( index );
        }

    }
    
    /* ===================== File -> model dependency handling ===================== */
    protected static boolean isValidPackageName(String str) {
        StringTokenizer tok = new StringTokenizer(str, "."); // NOI18N
        while (tok.hasMoreTokens()) {
            String part = tok.nextToken();
            if (!org.openide.util.Utilities.isJavaIdentifier(part))
                return false;
        }
        return true;
    }
    
    protected void primaryFileMoved(FileObject oldFile, FileObject newFile) {
        String oldName = oldFile.getName();

        addSourceChangeListener(oldFile);
        if (oldName.equals(newFile.getName())) {
            oldName = previousFileName;
        }
        previousFileName = null;
        
        // if the package name is valid --> check package declaration.
        // if name is different -> reset names to match the file's name.
        if (isValidPackageName(newFile.getParent().getPackageName('.')))
            updateSourceStart(this, newFile, oldName, !isModified());
    }
    
    protected void primaryFileChanged() {
        if (!getJavaEditor().isDocumentLoaded()) {
            JavaParser p = (JavaParser)getCookie(JavaParser.class);
            Object hook = p.getSourceImpl();
            
            if (hook == null || p.getSource().getStatus() == SourceElement.STATUS_ERROR)
                return;
            p.parse(JavaParser.PRIORITY_REFRESH, true, true);
        }
    }        
    
    // =============== Primary file monitoring ============
    private class PrimaryListener extends FileChangeAdapter implements PropertyChangeListener{
        public void propertyChange(final java.beans.PropertyChangeEvent evt) {
            if (!initialized)
                return;
            String propName = evt.getPropertyName();
            if ((propName == null) || (propName.equals(PROP_FILES))) {
              updateInstanceCookie();
            } else if (PROP_PRIMARY_FILE.equals(evt.getPropertyName())) {
                primaryFileMoved((FileObject)evt.getOldValue(), (FileObject)evt.getNewValue());
            }
        }
        
        public void fileChanged(FileEvent e) {
            if (!initialized)
                return;
            primaryFileChanged();
        }
    }
    
    /**
     * Extract the date of last modification of the source file. If the file is opened
     * in the editor AND marked as modified (in the editor), it returns true.
     * If the DataObject reports that it was modified, but the editor support knows
     * nothing about it, just ignore -- other parts of the DataObject are changed.
     */
    Date getLastModified() {
        SaveCookie c = (SaveCookie)getCookie(SaveCookie.class);
        if (c != null) {
            EditorCookie ck = (EditorCookie)getCookie(EditorCookie.class);
            if (ck != null && ck.isModified())
                return new Date();
        }
        return getPrimaryFile().lastModified();
    }
    
    /** Creates a Node.Cookie of given class. The method
     * may be called more than once.
     */
    public Node.Cookie createCookie(Class klass) {
        if (CompilerCookie.class.isAssignableFrom(klass)) {
            MultiDataObject.Entry entry = getPrimaryEntry();
            if (CompilerCookie.Build.class.isAssignableFrom(klass)) {
                return new JCompilerSupport.Build(entry);
            } else if (CompilerCookie.Clean.class.isAssignableFrom(klass)) {
                return new JCompilerSupport.Clean(entry);
            } else {
                // ASSERT (klass == CompilerCookie.Compile.class)
                return new JCompilerSupport.Compile(entry);
            }
	// all execution-related services -> getExecSupport
        } else if (ExecCookie.class.isAssignableFrom(klass)) {
            return getExecSupport();
        } else if (klass == ConnectionCookie.class) {
            return getConnectionSupport();
	} else if (klass.isAssignableFrom(JavaEditor.class)) {
	    return getJavaEditor();	
        } else if (klass == ConnectionCookie.class) {
            return initializeConnectionSupport(null);
        } 
        if (klass == SynchronizeCodeCookie.class) {
            initializeParsingSupport();
            return connectionManager;
        }
        
        if (SourceCookie.class.isAssignableFrom(klass) ||
            JavaParser.class.isAssignableFrom(klass)) {
            if (initializeParsingSupport() == null)
                return null;
            if (klass.isAssignableFrom(parserGlue.getClass()))
                return parserGlue;
            else
                return parserGlue.getParser();
        }
        return null;
    }
    
    // =============== Instance ==========================

    static class JavaInstanceSupport extends InstanceSupport {
        MultiDataObject.Entry entry;

        public JavaInstanceSupport(MultiDataObject.Entry entry) {
            super(entry);
        }

        public Class instanceClass ()
        throws java.io.IOException, ClassNotFoundException {
            try {
                return super.instanceClass();
            } catch (RuntimeException ex) {
                throw new ClassNotFoundException(ex.getMessage());
            }
        }

        /** 
         * Creates class loader that permits <<ALL FILES>> and all properties to be read,
         * and is based on currentClassLoader().
        protected ClassLoader createClassLoader() {
            java.security.Permissions perms = new java.security.Permissions();
            perms.add(new java.io.FilePermission("<<ALL FILES>>", "read")); // NOI18N
            perms.add(new java.util.PropertyPermission("*", "read")); // NOI18N
            perms.setReadOnly();

            org.openide.execution.NbClassLoader loader = 
                new org.openide.execution.NbClassLoader(
                    new org.openide.filesystems.FileSystem[] {},
                    org.openide.TopManager.getDefault().currentClassLoader());
            loader.setDefaultPermissions(perms);
            return loader;
        }
         */
    }
    
    // support for cross-compilation :-(((
    /* package-private */ void clearProducts() throws IOException {
        FileObject fo = getPrimaryFile();
        FileSystem myFileSystem = fo.getFileSystem();
        
        /*
        fo.setAttribute(EA_PRODUCTS, null);
         */
        Set secondaries = secondaryEntries();
        
        for (Iterator it = secondaries.iterator(); it.hasNext(); ) {
            Entry en = (Entry)it.next();
            FileObject target = en.getFile();
            if (target.getFileSystem() == myFileSystem)
                continue;
            removeSecondaryEntry(en);
        }
    }
    
    private FileSystem findTargetFileSystem() {
        CompilerType tp = CompilerSupport.getCompilerType(getPrimaryEntry());
        if (tp == null) {
            tp = getJavaSettings().getCompiler();
        }
        if (!(tp instanceof JavaCompilerType))
            return null;
        JavaCompilerType jcompt = (JavaCompilerType)tp;
        return jcompt.getTargetFileSystem();
    }

    /**
     * This method refreshes list of children in different folder managed
     * by this dataobject. STILL PENDING: need more efficient way for handling this
     * preferrably folder-based after compilation completes (refresh would be for the 
     * whole folder, then distributed
     * to proper DataObjects).
     */
    /* package-private */ void refreshDistantChildren(FileSystem fsRoot) throws IOException {
        if (fsRoot == null) {
            fsRoot = findTargetFileSystem();
        }
        if (fsRoot == null) {
            clearProducts();
            return;
        }
        
        // fsroot != null, there can be something that belongs to us underneath.
        FileObject my = getPrimaryFile();
        FileObject parent = my.getParent();
        String pack = parent == null ? "" : parent.getPackageName('.');
        String myName = my.getName();
        
        FileObject cls = fsRoot.find(pack, myName, JavaDataLoader.CLASS_EXTENSION);
        // optimization -- do not look further, if the main class is missing.
        if (cls == null) {
            clearProducts();
            return;
        }

        FileObject folder = pack.length() == 0 ? fsRoot.getRoot() :
            fsRoot.findResource(parent.getPackageName('/'));
        int myNameLength = myName.length();
        
        if (!folder.isFolder())
            return;
        Enumeration en = folder.getChildren(false);
        while (en.hasMoreElements()) {
            cls = (FileObject)en.nextElement();
            if (cls.getExt().equals(JavaDataLoader.CLASS_EXTENSION)) {
                String name = cls.getName();
                
                if (name.startsWith(myName) &&
                    (name.length() == myName.length() ||
                    name.charAt(myNameLength) == '$')) {
                    registerEntry(cls);
                }
            }
        }
        
        /*
        Collection c = new LinkedList();
        c.add(cls);
        my.setAttribute(EA_PRODUCTS, new FileObject[] {
            cls
        });
         */
    }
    
    public Collection getCompiledClasses() {
        Collection c = null;
        
        for (Iterator it = secondaryEntries().iterator(); it.hasNext(); ) {
            FileObject fo = ((Entry)it.next()).getFile();
            if (fo.getExt().equals(JavaDataLoader.CLASS_EXTENSION)) {
                if (c == null)
                    c = new LinkedList();
                c.add(fo);
            }
        }
        return c;
    }
    
    public String toString() {
        return "[JDO for " + getPrimaryFile() + "]";
    }
    
    private static final String EXTATTR_CLASSDEPEND_BASE = "class_dependency_"; // NOI18N
    
    private static class PersistentConnectionHandle implements Node.Handle {
        /** Handle to a DataObject's node.
        */
        Node.Handle        dataNodeHandle;
        
        PersistentConnectionHandle(JavaDataObject o) {
            this.dataNodeHandle = o.getNodeDelegate().getHandle();
        }
        
        public Node getNode() throws IOException {            
            // broken deserialization
            if (dataNodeHandle == null) {
                throw new IOException("Cannot deserialize do handle");
            }
            JavaDataObject d = (JavaDataObject)dataNodeHandle.getNode().getCookie(JavaDataObject.class);
            if (d == null) {
                throw new IOException("JavaDataObject cannot be found");
            }
            d.initializeParsingSupport();
            Node n = d.connectionManager.getListenerNode();
            return n;
        }
    }
    
    private class PersistentConnectionSupport extends SourceConnectionSupport {
        JavaDataObject  target;

        PersistentConnectionSupport(SourceElement src, LangModel mod) {
            super(src, mod, getPrimaryFile().getPackageName('.'));
            target = JavaDataObject.this;
        }
        
        protected Node.Handle createListenerHandle(Node n) {
            return new PersistentConnectionHandle(target);
        }
        
        private String createAttrName(String fqn) {
            return EXTATTR_CLASSDEPEND_BASE + fqn;
        }
        
        public int getSynchronizationMode(ClassElement t) {
            return (int)JavaDataObject.this.getSynchronizationType();
        }
        
        public void setSynchronizationMode(ClassElement t, int type) {
            JavaDataObject.this.setSynchronizationType((byte)type);
        }
        
        protected void saveDependencies(String fqn, String[] deps) {
            String attrName = createAttrName(fqn);
            
            if (deps == null || deps.length == 0) {
                try {
                    target.getPrimaryFile().setAttribute(attrName, null);
                } catch (IOException ex) {
                    // swallow.
                }
                return;
            }
            StringBuffer sb = new StringBuffer();
            sb.append(deps[0]);
            for (int i = 1; i < deps.length; i++) {
                sb.append(';');
                sb.append(deps[i]);
            }
            try {
                target.getPrimaryFile().setAttribute(attrName, sb.toString());
            } catch (IOException ex) {
            }
        }
        
        protected String[] readDependencies(String fqn) {
            String s = (String)target.getPrimaryFile().getAttribute(
                createAttrName(fqn));
            if (s == null) {
                return new String[] {};
            }
            StringTokenizer t = new StringTokenizer(s, ";");
            int toks = t.countTokens();
            String[] deps = new String[toks];
            for (int i = 0; t.hasMoreTokens(); i++) {
                deps[i] = t.nextToken();
            }
            return deps;
        }
        
        protected String[] readDependencyList() {
            Enumeration en = target.getPrimaryFile().getAttributes();
            Collection c = null;
            
            while (en.hasMoreElements()) {
                String name = (String)en.nextElement();
                if (name.startsWith(EXTATTR_CLASSDEPEND_BASE)) {
                    if (c == null)
                        c = new LinkedList();
                    c.add(name.substring(EXTATTR_CLASSDEPEND_BASE.length()));
                }
            }
            if (c == null) {
                return new String[] {};
            }
            return (String[])c.toArray(new String[c.size()]);
        }
    }
}
