/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.beans.PropertyVetoException;
import java.beans.PropertyChangeEvent;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.SwingUtilities;
import javax.swing.JEditorPane;
import javax.swing.Timer;
import javax.swing.event.*;
import javax.swing.text.*;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.actions.ToggleBreakpointAction;
import org.openide.cookies.SaveCookie;
import org.openide.cookies.SourceCookie;
import org.openide.debugger.Debugger;
import org.openide.debugger.Breakpoint;
import org.openide.debugger.DebuggerNotFoundException;
import org.openide.util.Task;
import org.openide.text.*;
import org.openide.nodes.Node;
import org.openide.loaders.DataObject;
import org.openide.loaders.MultiDataObject;
import org.openide.src.SourceElement;
import org.openide.util.TaskListener;
import org.openide.util.actions.ActionPerformer;
import org.openide.util.actions.SystemAction;
import org.openide.windows.CloneableTopComponent;
import org.openide.windows.Workspace;
import org.openide.windows.Mode;

import org.apache.regexp.CharacterArrayCharacterIterator;
import org.apache.regexp.RE;
import org.apache.regexp.RECompiler;
import org.apache.regexp.CharacterIterator;

import org.netbeans.modules.java.settings.JavaSettings;

import org.netbeans.modules.java.parser.JavaParser;
import org.netbeans.modules.java.ui.JEditorToolbar;

/*
* TODO:
* 1) undo support
*/
/** Java source-file extension for handling the Editor.
* The main purpose of this class is to manage guarded sections.
*
*
* @author Petr Hamernik
*/
public class JavaEditor extends EditorSupport {
    /** The prefix of all magic strings */
    private final static String MAGIC_PREFIX = "//GEN-"; // NOI18N

    /** Magic strings - special comments which are inserted during saving
     * and are removed during loading.
     */
    private static String[] SECTION_MAGICS;
    private static final int LONGEST_ITEM = 10;

    static {

        StringBuffer sb = new StringBuffer(MAGIC_PREFIX);
        int size = sb.length();

        SECTION_MAGICS = new String[7];
        SECTION_MAGICS[0] = sb.append("LINE:").toString();
        sb.setLength(size);
        SECTION_MAGICS[1] = sb.append("BEGIN:").toString(); // NOI18N
        sb.setLength(size);
        SECTION_MAGICS[2] = sb.append("END:").toString(); // NOI18N
        sb.setLength(size);
        SECTION_MAGICS[3] = sb.append("HEADER:").toString(); // NOI18N
        sb.setLength(size);
        SECTION_MAGICS[4] = sb.append("HEADEREND:").toString(); // NOI18N
        sb.setLength(size);
        SECTION_MAGICS[5] = sb.append("FIRST:").toString(); // NOI18N
        sb.setLength(size);
        SECTION_MAGICS[6] = sb.append("LAST:").toString(); // NOI18N
    }

    /** Types of the magic comments. */
    private final static int T_LINE      = 0;
    private final static int T_BEGIN     = 1;
    private final static int T_END       = 2;
    private final static int T_HEADER    = 3;
    private final static int T_HEADEREND = 4;
    // obsoleted - only for backward compatibility
    private final static int T_FIRST     = 5;
    private final static int T_LAST      = 6;

    /** Table of the guarded sections. Keys are the names of the sections
    * and values are the GuardedSection classes. The table is null till
    * while document is not in the memory.
    */
    HashMap sections;

    /** Timer which countdowns the auto-reparsing time. */
    Timer timer;

    /** New lines in this file was delimited by '\n' */
    static final byte NEW_LINE_N = 0;

    /** New lines in this file was delimited by '\r' */
    static final byte NEW_LINE_R = 1;

    /** New lines in this file was delimited by '\r\n' */
    static final byte NEW_LINE_RN = 2;

    /** The type of new lines */
    byte newLineType;

    /** Java Settings */
    static final JavaSettings settings = (JavaSettings)JavaSettings.findObject(JavaSettings.class, true);

    /** The hook to the parsed hierarchy - it shouldn't be released
    * when document is loaded.
    */
    Object parsedHook;

    /** True, if the document was saved from the last modification.
      This is like isModified(), but does not get cleared after user chooses Don't save during
      close.
    */
    private boolean documentSaved;
    
    /**
     * Helper variable for document reloading support. If the loadFromStream is called
     * and reloading is true, source reparsing is forced.
     */
    private boolean reloading = false;

    /** True, if there's a visible editor component flying around.
    */
    private boolean componentsCreated = false;
    
    private static final Comparator SECTION_COMPARATOR = new GuardedPositionComparator();

    /** Create a new Editor support for the given Java source.
    * @param entry the (primary) file entry representing the Java source file
    */
    public JavaEditor(MultiDataObject.Entry entry) {
        super(entry);

        sections = null;

        // initialize timer
        timer = new Timer(0, new java.awt.event.ActionListener() {
                              public void actionPerformed(java.awt.event.ActionEvent e) {
                                  changeTimeoutElapsed();
                              }
                          });
        timer.setInitialDelay(settings.getAutoParsingDelay());
        timer.setRepeats(false);

        // create document listener
        final DocumentListener docListener = new DocumentListener() {
                                                 public void insertUpdate(DocumentEvent e) { change(e); }
                                                 public void changedUpdate(DocumentEvent e) { }
                                                 public void removeUpdate(DocumentEvent e) { change(e); }

                                                 private void change(DocumentEvent e) {
                                                     documentSaved = false;
                                                     restartTimer(false);
                                                 }
                                             };

        // add change listener
        addChangeListener(new ChangeListener() {
                              public void stateChanged(ChangeEvent evt) {
                                  if (isDocumentLoaded()) {
                                      getDocument().addDocumentListener(docListener);
                                  } else {
	                                notifyClose();
                                  }
                              }
                          });
        
    }
    
    private JavaDataObject getJavaDataObject() {
        return (JavaDataObject)getJavaEntry().getDataObject();
    }
    
    private JavaParser findParser() {
        return (JavaParser)getJavaDataObject().getCookie(JavaParser.class);
    }

    private void changeTimeoutElapsed() {
        JavaParser parser = findParser();
        if (parser == null)
            return;
        Task t = parser.parse(parser.PRIORITY_REFRESH, false, false);
        t.addTaskListener(new TaskListener() {
            public void taskFinished(Task t2) {
                t2.removeTaskListener(this);
                notifyParsingDone();
            }
        }
        );
    }

    /** Restart the timer which starts the parser after the specified delay.
    * @param onlyIfRunning Restarts the timer only if it is already running
    */
    void restartTimer(boolean onlyIfRunning) {
        if (onlyIfRunning && !timer.isRunning())
            return;

        int delay = settings.getAutoParsingDelay();
        if (delay > 0) {
            timer.setInitialDelay(delay);
            timer.restart();
        }
    }

    /* Calls superclass.
    * @param pos Where to place the caret.
    * @return always non null editor
    */
    protected EditorSupport.Editor openAt(PositionRef pos) {
        EditorSupport.Editor ed = super.openAt(pos);
        ed.requestFocus();
        return ed;
    }

    /** Notify about the editor closing.
    */
    protected void notifyClose() {
        clearSections();
        JavaDataObject jdo = (JavaDataObject) getJavaEntry().getDataObject();
        if (!documentSaved) {
	    documentSaved = true;   
        }
        componentsCreated = false;
        parsedHook = null;
    }

    /** Notify that parsing task has been finished; some dependent data may now
      be refreshed from up-to-date parsing info 
    */
    protected void notifyParsingDone() {
    }

    /** Read the file from the stream, filter the guarded section
    * comments, and mark the sections in the editor.
    *
    * @param doc the document to read into
    * @param stream the open stream to read from
    * @param kit the associated editor kit
    * @throws IOException if there was a problem reading the file
    * @throws BadLocationException should not normally be thrown
    * @see #saveFromKitToStream
    */
    protected void loadFromStreamToKit (StyledDocument doc, InputStream stream, EditorKit kit) throws IOException, BadLocationException {
        boolean filling = false;

        if (sections == null) {
            sections = new HashMap(10);
            filling = true;
        }

        if (filling) {
            GuardedReader reader = new GuardedReader(stream, false);
            kit.read(reader, doc, 0);
            fillSections(reader, doc);
            newLineType = reader.getNewLineType();
        }
        else {
            kit.read(stream, doc, 0);
        }
        // stream is not closed, because is it close outside this method

        // parse it and remember the hook to prevent to garbage collected
        // parsed hierarchy
	final boolean forceUpdate = reloading;

	reloading = false;
	documentSaved = true;
        final JavaParser p = findParser();
        if (p == null)
            return;
        SwingUtilities.invokeLater(new Runnable() {
                                       public void run() {
                                           Task t = p.parse(p.PRIORITY_NORMAL, forceUpdate, false);
                                           parsedHook = p.getSourceImpl();
                                           t.isFinished();
                                           final JavaDataObject jdo = (JavaDataObject) getJavaEntry().getDataObject();
                                       }
                                   });
    }

    /** Store the document and add the special comments signifying
    * guarded sections.
    *
    * @param doc the document to write from
    * @param kit the associated editor kit
    * @param stream the open stream to write to
    * @throws IOException if there was a problem writing the file
    * @throws BadLocationException should not normally be thrown
    * @see #loadFromStreamToKit
    */
    protected void saveFromKitToStream(StyledDocument doc, EditorKit kit, OutputStream stream) throws IOException, BadLocationException {
        OutputStream os = new NewLineOutputStream(stream, newLineType);
        if (sections != null) {
            ArrayList list = new ArrayList(sections.values());
            if (list.size() > 0) {
                GuardedWriter writer = new GuardedWriter(os, list);
                kit.write(writer, doc, 0, doc.getLength());
                return;
            }
        }
        kit.write(os, doc, 0, doc.getLength());
        documentSaved = true;
    }

    /** Reload the document when changed externally */
    protected Task reloadDocumentTask () {
	StyledDocument doc = null;
	    // PENDING: do not openDocument() to obtain doc, as all text is removed from the
	    // document, guarded sections are (probably) erased anyway. This causes two passes
	    // through loadFromStreamToKit!
            try {
                doc = openDocument();
            }
            catch (IOException e) {
                // no reload performed in this case
            }
	    
            if (doc != null) {
                clearSections();
                NbDocument.unmarkGuarded(doc, 0, doc.getLength());
		reloading = true;
                return super.reloadDocumentTask ();
            }
            else
                return null;
    }

    /** Save the document in this thread and start reparsing it.
    * @exception IOException on I/O error
    */
    public void saveDocument () throws IOException {
        saveDocument(true, true);
    }

    /** Save the document in this thread.
    * @param parse true if the parser should be started, otherwise false
    * @exception IOException on I/O error
    */
    protected void saveDocumentIfNecessary(boolean parse) throws IOException {
        saveDocument(parse, false);
    }

    /** Save the document in this thread.
    * @param parse true if the parser should be started, otherwise false
    * @param forceSave if true save always, otherwise only when is modified
    * @exception IOException on I/O error
    */
    private void saveDocument(boolean parse, boolean forceSave) throws IOException {
        if (forceSave || isModified()) {
            super.saveDocument();
        }
    }

    // ==================== SourceCookie.Editor methods =================

    /** Returns a source element describing the hierarchy of the source.
    * @return the element
    * @deprecated Please use DataObject services to obtain java hierarchy.
    */
    public SourceElement getSource() {
        JavaDataObject jdo = (JavaDataObject) entry.getDataObject();
        return jdo.getSource();
    }

    /** Translate a source element to text.
    *
    * @param element an element from the source hierarchy
    * @return a text element
    * @deprecated Please use DataObject's cookies to translate swing <-> org.openide.src
    */
    public javax.swing.text.Element sourceToText(org.openide.src.Element element) {
        return null;
    }

    /** Translate a text element to a source element, if it is possible to do so.
    *
    * @param element a text element
    * @return the element from the source hierarchy
    * @exception NoSuchElementException if the text element doesn't match
    *  any element from the source hierarchy
    * @deprecated Please use DataObject's cookies to translate swing <-> org.openide.src
    */
    public org.openide.src.Element textToSource(javax.swing.text.Element element) throws NoSuchElementException {
        throw new NoSuchElementException();
    }

    /** Find the element at the specified offset in the document.
    * @param offset The position of the element
    * @return the element at the position.
    * @deprecated Please use DataObject's cookies to translate swing <-> org.openide.src
    */
    public org.openide.src.Element findElement(int offset) {
        JavaDataObject jdo = (JavaDataObject) getJavaEntry().getDataObject();
        return null; /* jdo.getSourceElementImpl().findElement(offset); */
    }

    // ==================== Guarded sections public methods =================

    /**
     * Creates an empty simple section at the given position.
     * The position must not be within any existing guarded section or
     * Java Element and the passed name must not be registered to other
     * already existing section. The created section will initially contain
     * one space and a newline.
     * @return SimpleSection instance that can be used for generating text into
     * the protected region
     * @throw IllegalArgumentException if either the name has been already used, or
     * the position is inside another section or Java Element.
     * @throw BadLocationException if pos is outside of document's scope, or
     * the document does not permit creating the guarded section.
     */
    public SimpleSection createSimpleSection(PositionRef pos, String name) 
        throws IllegalArgumentException, BadLocationException {
        checkOverlap(pos);
        return doCreateSimpleSection(pos, name);
    }

    /*
     * Creates a simple section over the given bounds in the source text.
     * The bounds must not overlap with any existing section or Java Element
     * and the passed name must not be registered to other
     * already existing section. The section will then contain all the text
     * inside the passed PositionBounds.
     * @return SimpleSection instance that can be used for generating text into
     * the protected region
     * @throw IllegalArgumentException if either the name has been already used, or
     * the bounds overlap with another section or Java Element.
     * @throw BadLocationException if pos is outside of document's scope, or
     * the document does not permit creating the guarded section.
     */
    public SimpleSection createSimpleSection(PositionBounds bounds, String name) 
        throws IllegalArgumentException, BadLocationException {
        checkOverlap(bounds);
        return doCreateSimpleSection(bounds, name);
    }
    
    private void checkOverlap(PositionRef pos) {
        Iterator it = sections.values().iterator();
        while (it.hasNext()) {
            GuardedSection s = (GuardedSection)it.next();
            if (s.contains(pos, false)) {
                throw new IllegalArgumentException("Sections overlap"); // NOI18N
            }
        }
    }

    /*
     * Tests if bounds overlap with any existing guarded section.
     * @return <code>true</code> if bounds does not overlap with any section, otherwise <code>false</code>
     */    
    public boolean testOverlap(PositionBounds bounds) {
        try {
            StyledDocument loadedDoc = openDocument();
        }
        catch (IOException e) {
            throw new IllegalArgumentException();
        }
        try {
            checkOverlap(bounds);
        } catch (IllegalArgumentException e){
            return false;
        }
        return true;
    }
        
    private void checkOverlap(PositionBounds bounds) throws IllegalArgumentException {
        Collection c = new TreeSet(new GuardedPositionComparator());
        c.addAll(sections.values());
        
        Iterator it = c.iterator();
        PositionRef begin = bounds.getBegin();
        PositionRef end = bounds.getEnd();
        int beginOffset = begin.getOffset();
        int endOffset = end.getOffset();
        GuardedSection starting = null;
        boolean overlapOK = false;
        
        while (it.hasNext()) {
            GuardedSection s = (GuardedSection)it.next();
            if (s.contains(begin, false) || s.contains(end, false))
                throw new IllegalArgumentException("Sections overlap"); // NOI18N
            if (s.getBegin().getOffset() > beginOffset) {
                if (s.getBegin().getOffset() < endOffset) {
                    throw new IllegalArgumentException("Sections overlap"); // NOI18N
                }
                break;
            }
        }
    }

    /**
     * This method is here only to calm down buggy javac from JDK-1.2.2
     */
    protected boolean notifyModified() {
	return super.notifyModified();
    }
    
    private SimpleSection doCreateSimpleSection(final PositionBounds bounds, final String name)
    throws IllegalArgumentException, BadLocationException  {
        StyledDocument loadedDoc = null;
        try {
            loadedDoc = openDocument();
        }
        catch (IOException e) {
            throw new IllegalArgumentException("Cannot load document"); // NOI18N
        }
        final StyledDocument doc = loadedDoc;
        final SimpleSection[] sect = new SimpleSection[] { null };
        
        Util.ExceptionRunnable run = new Util.ExceptionRunnable() {
            public void run() throws Exception {
                sect[0] = new SimpleSection(name,
                createBounds(bounds.getBegin().getOffset(), bounds.getEnd().getOffset(), false)
                );
                sections.put(sect[0].getName(), sect[0]);
                sect[0].markGuarded(doc);
            }
        };
        try {
            Util.runAtomic(doc, run);
	    notifyModified();
        }
        catch (Exception e) {
            if (e instanceof BadLocationException)
                throw (BadLocationException) e;
            else
                throw new IllegalArgumentException();
        }
        return sect[0];
    }
    
    private SimpleSection doCreateSimpleSection(final PositionRef pos, final String name) 
        throws IllegalArgumentException, BadLocationException  {
        StyledDocument loadedDoc = null;
        try {
            loadedDoc = openDocument();
        }
        catch (IOException e) {
            throw new IllegalArgumentException();
        }
        final StyledDocument doc = loadedDoc;
        final SimpleSection[] sect = new SimpleSection[] { null };

        Util.ExceptionRunnable run = new Util.ExceptionRunnable() {
                                         public void run() throws Exception {
                                             int where = pos.getOffset();
                                             doc.insertString(where, "\n \n", null); // NOI18N
                                             sect[0] = new SimpleSection(name,
                                                                         createBounds(where + 1, where + 3, false)
                                                                        );
                                             sections.put(sect[0].getName(), sect[0]);
                                             sect[0].markGuarded(doc);
                                         }
                                     };
        try {
            Util.runAtomic(doc, run);
	    notifyModified();
        }
        catch (Exception e) {
            if (e instanceof BadLocationException)
                throw (BadLocationException) e;
            else
                throw new IllegalArgumentException();
        }
        return sect[0];
    }
    
    /**
     * Create new simple guarded section at a specified place.
     * @param previous section to create the new one after
     * @param name the name of the new section
     * @exception IllegalArgumentException if the name is already in use
     * @exception BadLocationException if it is not possible to create a
     *            new guarded section here
     */
    public SimpleSection createSimpleSectionAfter(final GuardedSection previous,
    final String name)
    throws IllegalArgumentException, BadLocationException {
        PositionBounds bounds;
        if (previous instanceof SimpleSection)
            bounds = ((SimpleSection) previous).bounds;
        else
            bounds = ((InteriorSection) previous).bottom;
        if ((previous == null) || (!previous.valid))
            throw new IllegalArgumentException("Invalid guarded block");

        return doCreateSimpleSection(bounds.getEnd(), name);
    }

    public InteriorSection createInteriorSection(PositionRef pos, String name) 
    throws IllegalArgumentException, BadLocationException {
        checkOverlap(pos);
        return doCreateInteriorSection(pos, name);
    }
    
    public InteriorSection createInteriorSectionAfter(GuardedSection previous,
        String name) throws IllegalArgumentException, BadLocationException {
            PositionBounds bounds;
        if (previous instanceof SimpleSection)
            bounds = ((SimpleSection) previous).bounds;
        else
            bounds = ((InteriorSection) previous).bottom;
        if ((previous == null) || (!previous.valid))
            throw new IllegalArgumentException("Invalid guarded block");
        return doCreateInteriorSection(bounds.getEnd(), name);
    }
    
    public InteriorSection createInteriorSection(PositionBounds bounds, PositionBounds interior,
        String name) throws IllegalArgumentException, BadLocationException {
        checkOverlap(bounds);
        if (bounds.getBegin().getOffset() > interior.getEnd().getOffset() ||
            bounds.getEnd().getOffset() < interior.getEnd().getOffset())
            throw new IllegalArgumentException("Interior is not nested.");
        return doCreateInteriorSection(bounds, interior, name);
    }
    
    private InteriorSection doCreateInteriorSection(final PositionBounds bounds,
    final PositionBounds interiorBounds, final String name)
    throws IllegalArgumentException, BadLocationException {
        StyledDocument loadedDoc = null;
        try {
            loadedDoc = openDocument();
        }
        catch (IOException e) {
            throw new IllegalArgumentException("Cannot load document"); // NOI18N
        }

        final StyledDocument doc = loadedDoc;
        final InteriorSection[] sect = new InteriorSection[] { null };

        Util.ExceptionRunnable run = new Util.ExceptionRunnable() {
            public void run() {
                sect[0] = new InteriorSection(name,
                createBounds(bounds.getBegin().getOffset(), interiorBounds.getBegin().getOffset(), false),
                createBounds(interiorBounds.getBegin().getOffset(), interiorBounds.getEnd().getOffset(), true),
                createBounds(interiorBounds.getEnd().getOffset(), bounds.getEnd().getOffset(), false)
                );
                sections.put(sect[0].getName(), sect[0]);
                sect[0].markGuarded(doc);
            }
        };
        try {
            Util.runAtomic(doc, run);
	    notifyModified();
        }
        catch (Exception e) {
            if (e instanceof BadLocationException)
                throw (BadLocationException) e;
            else
                throw new IllegalArgumentException();
        }
        return sect[0];
    }
    
    /** Create new interior guarded section at a specified place.
    * @param previous section to create the new one after
    * @param name the name of the new section
    * @exception IllegalArgumentException if the name is already in use
    * @exception BadLocationException if it is not possible to create a
    *            new guarded section here
    */
    private InteriorSection doCreateInteriorSection(final PositionRef pos, 
            final String name)
    throws IllegalArgumentException, BadLocationException {
        StyledDocument loadedDoc = null;
        try {
            loadedDoc = openDocument();
        }
        catch (IOException e) {
            throw new IllegalArgumentException();
        }

        final StyledDocument doc = loadedDoc;
        final InteriorSection[] sect = new InteriorSection[] { null };

        Util.ExceptionRunnable run = new Util.ExceptionRunnable() {
            public void run() throws Exception {
                int where = pos.getOffset();
                doc.insertString(where, "\n \n \n \n", null); // NOI18N
                sect[0] = new InteriorSection(name,
                createBounds(where + 1, where + 3, false),
                createBounds(where + 3, where + 5, true),
                createBounds(where + 5, where + 7, false)
                );
                sections.put(sect[0].getName(), sect[0]);
                sect[0].markGuarded(doc);
            }
        };
        try {
            Util.runAtomic(doc, run);
	    notifyModified();
        }
        catch (Exception e) {
            if (e instanceof BadLocationException)
                throw (BadLocationException) e;
            else
                throw new IllegalArgumentException();
        }
        return sect[0];
    }

    /** Try to find the simple section of the given name.
    * @param name the name of the requested section
    * @return the found guarded section or <code>null</code> if there is no section
    *         of the given name
    */
    public SimpleSection findSimpleSection(String name) {
        GuardedSection s = findSection(name);
        return (s instanceof SimpleSection) ? (SimpleSection) s : null;
    }

    /** Try to find the interior section of the given name.
    * @param name the name of the looked-for section
    * @return the found guarded section or <code>null</code> if there is no section
    *         of the given name
    */
    public InteriorSection findInteriorSection(String name) {
        GuardedSection s = findSection(name);
        return (s instanceof InteriorSection) ? (InteriorSection) s : null;
    }

    /** Try to find the section of the given name.
    * @param name the name of the looked-for section
    * @return the found guarded section or <code>null</code> if there is no section
    *         of the given name
    */
    public GuardedSection findSection(String name) {
        try {
            StyledDocument doc = openDocument ();
            synchronized (this) {
                if (sections != null)
                    return (GuardedSection) sections.get(name);
            }
        }
        catch (IOException e) {
        }
        return null;
    }

    /** Get all sections.
    * @return an iterator of {@link JavaEditor.GuardedSection}s
    */
    public Iterator getGuardedSections() {
        try {
            StyledDocument doc = openDocument ();
            synchronized (this) {
                if (sections != null)
                    return ((HashMap)sections.clone()).values().iterator();
            }
        }
        catch (IOException e) {
        }
        return Collections.EMPTY_SET.iterator();
    }

    /** Get all section names.
    * @return an iterator of {@link String}s
    */
    public Iterator getGuardedSectionNames() {
        try {
            StyledDocument doc = openDocument ();
            synchronized (this) {
                if (sections != null)
                    return ((HashMap)sections.clone()).keySet().iterator();
            }
        }
        catch (IOException e) {
        }
        return Collections.EMPTY_SET.iterator();
    }

    /**
     * Finds a position not obscured by a guarded block that is inside the given
     * bounds. Favors positions at the beginning of the passed bounds. Returns null,
     * if the document cannot be loaded.
     * @param bnds bounds to search within
     * @return PositionRef that can be safely written into.
     */
    public PositionRef findFreePosition(PositionBounds bnds) {
        StyledDocument doc;
        try {
            doc = openDocument();
        } catch (IOException ex) {
            return null;
        }
        
        PositionRef beginPos = bnds.getBegin();
        int startOffs = beginPos.getOffset();
        
        TreeSet set = new TreeSet(SECTION_COMPARATOR);
        set.addAll(this.sections.values());
        for (Iterator it = set.iterator(); it.hasNext(); ) {
            GuardedSection s = (GuardedSection)it.next();
            PositionRef start = s.getBegin();
            if (start.getOffset() > startOffs) {
                // no section at the start.
                break;
            }
            if (s.contains(beginPos, false)) {
                // got guarded block that contains
                PositionRef after = s.getPositionAfter();
                if (after.getOffset() >= bnds.getEnd().getOffset()) {
                    return null;
                }
                return after;
            }
        }
        return beginPos;
    }

    // ==================== Misc not-public methods ========================

    /* A method to create a new component. Overridden in subclasses.
    * @return the {@link Editor} for this support
    */
    protected CloneableTopComponent createCloneableTopComponent () {
        prepareDocument();
        return createJavaEditorComponent();
    }

    /** Method for creation of the java editor component
    * - accessible from the innerclass.
    */
    JavaEditorComponent createJavaEditorComponent() {
        DataObject obj = findDataObject ();
        JavaEditorComponent editor = new JavaEditorComponent(obj);

        componentsCreated = true;
        // dock into editor mode if possible
        Workspace current = TopManager.getDefault().getWindowManager().getCurrentWorkspace();
        Mode editorMode = current.findMode(EDITOR_MODE);
        if (editorMode != null)
            editorMode.dockInto(editor);

        return editor;
    }

    /** only accessibility method */
    MultiDataObject.Entry getJavaEntry() {
        return entry;
    }

    /** Set all sections as invalid. It is called from closeLast method
    * of the JavaEditorComponent.
    */
    synchronized void clearSections() {
        if (sections != null) {
            Iterator it = ((HashMap)sections.clone()).values().iterator();
            while (it.hasNext()) {
                GuardedSection sect = (GuardedSection) it.next();
                sect.valid = false;
            }
            sections = null;
        }
    }

    PositionRef findUnguarded(PositionRef fromWhere, boolean allowHoles, boolean after) {
	Iterator it = getGuardedSections();
	
	while (it.hasNext()) {
	    GuardedSection sect = (GuardedSection)it.next();
	    if (sect.contains(fromWhere, allowHoles)) {
		if (after) {
		    return sect.getPositionAfter();
		} else {
		    return sect.getPositionBefore();
		}
	    }
	}
	return fromWhere;
    }
    
    /** The real component of the Java editor.
    * Subclasses should not attempt to work with this;
    * if they require special editing support, separate windows
    * should be created by overriding (e.g.) {@link EditorSupport#open}.
    */
    public static class JavaEditorComponent extends EditorSupport.Editor implements java.beans.PropertyChangeListener {
        /** Default delay between cursor movement and updating selected element nodes. */
        static final int SELECTED_NODES_DELAY = 1000;

        /** Timer which countdowns the "update selected element node" time. */ // NOI18N
        Timer timerSelNodes;

        /** The support, subclass of EditorSupport */
        JavaEditor support;

        /** Listener on caret movements */
        CaretListener caretListener;

        Parsing.Listener parsingListener;
        
        /**
         * Toolbar that is displayed at the top of the editor window.
         * Lazily initialized in enableToolBar / createToolBar.
         */
        java.awt.Component toolBar;

        /** The last caret offset position. */
        int lastCaretOffset = -1;

        static final long serialVersionUID =6223349196427270209L;

        /** Only for externalization */
        public JavaEditorComponent () {
            super();
        }

        /** Creates new editor */
        public JavaEditorComponent (DataObject obj) {
            super(obj);
            initialize();
        }
        
        private transient org.openide.util.RequestProcessor.Task selectionTask = null;

        /** Selects element at the given position. */
        void selectElementsAtOffset(final int offset) {
            SourceElement.Impl holder = support.findParser().getSourceImpl();
            
            if (holder == null)
                return;
            if (getRegistry().getActivated() != JavaEditorComponent.this) {
              return;
            }
            SourceCookie.Editor seditor = (SourceCookie.Editor)support.getJavaDataObject().
                getCookie(SourceCookie.Editor.class);
            
            org.openide.src.Element element = seditor.findElement(offset);
            org.openide.src.nodes.ElementNodeFactory factory = JavaDataObject.getExplorerFactory();
            Node n = null;
            
            if (element instanceof org.openide.src.MethodElement) {
                n = factory.createMethodNode((org.openide.src.MethodElement)element);
            }
            else if (element instanceof org.openide.src.ClassElement) {
                n = factory.createClassNode((org.openide.src.ClassElement)element);
            }
            else if (element instanceof org.openide.src.ConstructorElement) {
                n = factory.createConstructorNode((org.openide.src.ConstructorElement)element);
            }
            else if (element instanceof org.openide.src.FieldElement) {
                n = factory.createFieldNode((org.openide.src.FieldElement)element);
            }
            else if (element instanceof org.openide.src.InitializerElement) {
                n = factory.createInitializerNode((org.openide.src.InitializerElement)element);
            }
            else if (element instanceof org.openide.src.SourceElement) {
                n = support.getJavaEntry().getDataObject().getNodeDelegate();
            }
            setActivatedNodes((n != null) ? new Node[] { n } : new Node[] {} );
        }

        protected void notifyParsingDone() {
            if (lastCaretOffset != -1) {
                selectElementsAtOffset(lastCaretOffset);
            }
        }

        /**
         * Currently only watches PROP_SHOW_TOOLBAR option.
         */
        public void propertyChange(java.beans.PropertyChangeEvent ev) {
            String propName = ev.getPropertyName();
            if (settings.PROP_SHOW_TOOLBAR.equals(propName)) {
                enableToolBar(((Boolean)ev.getNewValue()).booleanValue());
            }
        }
        
        /**
         * Enables or disables toolbar for the editor.
         */
        public final void enableToolBar(boolean enable) {
            if (enable) {
                java.awt.Component c = getToolBar();
                if (c == null)
                    return;
                if (isAncestorOf(toolBar)) {
                    toolBar.setVisible(true);
                } else {
                    add(toolBar, java.awt.BorderLayout.NORTH);
                    toolBar.setVisible(true);
                }
            } else {
                if (toolBar != null) {
                    toolBar.setVisible(false);
                }
            }
        }
        
        private void initializeToolBar() {
            enableToolBar(settings.getShowToolbar());
        }

        /**
         * Creates toolbar for the editor. In theory, the toolbar should
         * load and maintain its contents from %system_fs%/Editors/text|x-java/Toolbar
         */
        protected java.awt.Component createToolBar() {
            javax.swing.JToolBar tb = JEditorToolbar.createToolBar("Editors/text|x-java/Toolbar");
            return tb;
        }
        
        /**
         * Refreshes the activated node immediately. Provides system actions
         * based on the node activated at the time of popu invoke.
         */
        public SystemAction[] getSystemActions() {
            selectElementsAtOffset(lastCaretOffset);
            timerSelNodes.stop();
            return super.getSystemActions();
        }
        
        /**
         * Retrieves the tollolbar component instance.
         */
        protected final java.awt.Component getToolBar() {
            if (toolBar != null)
                return toolBar;
            return toolBar = createToolBar();
        }

        /** Obtain a support for this component */
        private void initialize () {
    	    // This local is to keep javac 1.2 happy.
	        final DataObject myJavaObject = obj;    
            support = (JavaEditor)obj.getCookie(JavaEditor.class);

            timerSelNodes = new Timer(0, new java.awt.event.ActionListener() {
                                          public void actionPerformed(java.awt.event.ActionEvent e) {
                                              selectElementsAtOffset(lastCaretOffset);
                                          }
                                      });
            timerSelNodes.setInitialDelay(SELECTED_NODES_DELAY);
            timerSelNodes.setRepeats(false);

            caretListener = new CaretListener() {
                                public void caretUpdate(CaretEvent e) {
                                    support.restartTimer(true);
                                    restartTimerSelNodes(e.getDot());
                                }
                            };

            parsingListener = new Parsing.Listener() {
                                  public void objectParsed(Parsing.Event info) {
                                      if (info.getJavaDataObject() == myJavaObject) {
                                          notifyParsingDone();
                                      }
                                  }
                              };
            
            initializeToolBar();
            settings.addPropertyChangeListener(this);
        }
        
        public void removeNotify() {
            super.removeNotify();
            settings.removePropertyChangeListener(this);
        }

        /** Restart the timer which updates the selected nodes after the specified delay from
        * last caret movement.
        */
        void restartTimerSelNodes(int pos) {
            timerSelNodes.setInitialDelay(SELECTED_NODES_DELAY);
            timerSelNodes.restart();
            lastCaretOffset = pos;
        }

        /** Returns Editor pane for private use.
        * @return Editor pane for private use.
        */
        private JEditorPane getEditorPane () {
            return pane;
        }

        /* Is called from the clone method to create new component from this one.
        * This implementation only clones the object by calling super.clone method.
        * @return the copy of this object
        */
        protected CloneableTopComponent createClonedObject () {
            return support.createJavaEditorComponent();
        }

        /* This method is called when parent window of this component has focus,
        * and this component is preferred one in it. This implementation adds 
        * performer to the ToggleBreakpointAction.
        */
        protected void componentActivated () {
            try {
                final Debugger debugger = TopManager.getDefault ().getDebugger ();
                ((ToggleBreakpointAction) SystemAction.get (ToggleBreakpointAction.class)).
                setActionPerformer (new ActionPerformer () {
                                        public void performAction (SystemAction a) {
                                            int l = NbDocument.findLineNumber (
                                                        support.getDocument (),
                                                        getEditorPane ().getCaret ().getDot ()
                                                    );
                                            Line ll = support.getLineSet ().getCurrent (l);
                                            if (ll.isBreakpoint ()) {
                                                Breakpoint b = debugger.findBreakpoint (ll);
                                                if (b == null) ll.setBreakpoint (false);
                                                else b.remove ();
                                            } else {
                                                ll.setBreakpoint (true);
                                                debugger.createBreakpoint (ll);
                                            }
                                        }
                                    });
            } catch (DebuggerNotFoundException e) {
            }
            getEditorPane().addCaretListener(caretListener);
            Parsing.addParsingListener(parsingListener);
            super.componentActivated ();
        }

        /*
        * This method is called when parent window of this component losts focus,
        * or when this component losts preferrence in the parent window. This 
        * implementation removes performer from the ToggleBreakpointAction.
        */
        protected void componentDeactivated () {
            ((ToggleBreakpointAction) SystemAction.get (ToggleBreakpointAction.class)).
            setActionPerformer (null);
            getEditorPane().removeCaretListener(caretListener);
            Parsing.removeParsingListener(parsingListener);
            synchronized (this) {
                if (selectionTask != null) {
                    selectionTask.cancel();
                    selectionTask = null;
                }
            }
            super.componentDeactivated ();
        }

        /* When closing last view, also close the document.
         * @return <code>true</code> if close succeeded
        */
        protected boolean closeLast () {
            if (!super.closeLast())
                return false;
            support.componentsCreated = false;
            return true;
        }

        /** Deserialize this top component.
        * @param in the stream to deserialize from
        */
        public void readExternal (ObjectInput in)
        throws IOException, ClassNotFoundException {
            super.readExternal(in);
            initialize();
        }

    } // end of JavaEditorComponent inner class

    /** Takes the section descriptors from the GuardedReader and
    * fills the table 'sections', also marks as guarded all sections
    * in the given document.
    * @param is Where to take the guarded section descriptions.
    * @param doc Where to mark guarded.
    */
    private void fillSections(GuardedReader is, StyledDocument doc) {
        JavaEditor.SectionDesc descBegin = null;

        Iterator it = is.list.iterator();
        while (it.hasNext()) {
            SectionDesc descCurrent = (SectionDesc) it.next();
            GuardedSection sect = null;
            switch (descCurrent.type) {
            case T_LINE:
                sect = new SimpleSection(descCurrent.name,
                                         createBounds(descCurrent.begin,
                                                      descCurrent.end, false));
                break;

            case T_BEGIN:
            case T_HEADER:
            case T_FIRST:
                descBegin = descCurrent;
                break;

            case T_HEADEREND:
                if ((descBegin != null) &&
                        ((descBegin.type == T_HEADER) || (descBegin.type == T_FIRST)) &&
                        (descCurrent.name.equals(descBegin.name))
                   ) {
                    descBegin.end = descCurrent.end;
                }
                else {
                    //SYNTAX ERROR - ignore it.
                    descBegin = null;
                }
                break;

            case T_END:
            case T_LAST:
                if ((descBegin != null) && (descBegin.name.equals(descCurrent.name))) {
                    if ((descBegin.type == T_BEGIN) && (descCurrent.type == T_END)) {
                        // simple section
                        sect = new SimpleSection(descCurrent.name,
                                                 createBounds(descBegin.begin,
                                                              descCurrent.end, false));
                        break;
                    }
                    if (((descBegin.type == T_FIRST) && (descCurrent.type == T_LAST)) ||
                            ((descBegin.type == T_HEADER) && (descCurrent.type == T_END))) {
                        // interior section
                        sect = new InteriorSection(descCurrent.name,
                                                   createBounds(descBegin.begin, descBegin.end, false),
                                                   createBounds(descBegin.end, descCurrent.begin, true),
                                                   createBounds(descCurrent.begin, descCurrent.end, false)
                                                  );
                        break;
                    }
                }
                //SYNTAX ERROR - ignore it.
                descBegin = null;
                break;
            }

            if (sect != null) {
                sections.put(sect.getName(), sect);
                descBegin = null;
                sect.markGuarded(doc);
            }
        }
    }

    /** Simple creates the bounds for the two offsets. */
    public PositionBounds createBounds(int begin, int end, boolean dir) {
        if (dir) {
            return new PositionBounds(
                       createPositionRef(begin, Position.Bias.Forward),
                       createPositionRef(end, Position.Bias.Backward)
                   );
        }
        else {
            return new PositionBounds(
                       createPositionRef(begin, Position.Bias.Backward),
                       createPositionRef(end, Position.Bias.Forward)
                   );
        }
    }

    // ==================== Public inner classes ========================

    /** Represents one guarded section.
    */
    public abstract class GuardedSection extends Object {
        /** Name of the section. */
        String name;

        /** If the section is valid or if it was removed. */
        boolean valid;

        /** Get the name of the section.
        * @return the name
        */
        public String getName() {
            return name;
        }

        /** Creates new section.
        * @param name Name of the new section.
        */
        GuardedSection(String name) {
            this.name = name;
            valid = true;
        }

        /** Set the name of the section.
        * @param name the new name
        * @exception PropertyVetoException if the new name is already in use
        */
        public void setName(String name) throws PropertyVetoException {
            if (!this.name.equals(name)) {
                synchronized (JavaEditor.this) {
                    if (valid) {
                        if (sections.get(name) != null)
                            throw new PropertyVetoException("", new PropertyChangeEvent(this, "name", this.name, name)); // NOI18N
                        sections.remove(this.name);
                        this.name = name;
                        sections.put(name, this);
                    }
                }
            }
        }

        /** Deletes the text of the section and
        * removes it from the table. The section will then be invalid
        * and it will be impossible to use its methods.
        *
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        */
        public boolean deleteSection() {
            synchronized (JavaEditor.this) {
                if (valid) {
                    try {
                        sections.remove(name);
                        // get document should always return the document, when section
                        // is deleted, because it is still valid (and valid is only
                        // when document is loaded.
                        unmarkGuarded(getDocument());
                        deleteText();
                        valid = false;
                        return true;
                    }
                    catch (BadLocationException e) {
                    }
                    catch (IOException e) {
                    }
                }
                return false;
            }
        }
        
        /**
         * Tests if the section is still valid - it is not removed from the
         * source.
         */
        public boolean isValid() {
            return valid;
        }
        
        /**
         * Removes the section from the Document, but retains the text contained
         * within. The method should be used to unprotect a region of code
         * instead of calling NbDocument.
         * @return true if the operation succeeded.
         */
        public boolean removeSection() {
            synchronized (JavaEditor.this) {
                if (!valid)
                    return false;
                sections.remove(name);
                // get document should always return the document, when section
                // is deleted, because it is still valid (and valid is only
                // when document is loaded.
                unmarkGuarded(getDocument());
                valid = false;
                return true;
            }
        }

        /** Delete one new-line character before the specified offset.
        * This method is used when guarded blocks are deleted. When new guarded block is created,
        * there is added one more new-line before it, so this method remove this char in the end of
        * guarded block life cycle.
        * It works only when there is "\n" char before the offset and no problem occured (IOException...)
        * @param offset The begin of removed guarded block.
        */
        void deleteNewLineBeforeBlock(int offset) {
            if (offset > 1) {
                try {
                    PositionBounds b = createBounds(offset - 1, offset, true);
                    String s = b.getText();
                    if (s.equals("\n")) { // NOI18N
                        b.setText(""); // NOI18N
                    }
                }
                catch (IOException e) {
                }
                catch (BadLocationException e) {
                }
            }
        }

        /** Opens the editor and set cursor to this guarded section.
        */
        public void openAt() {
            JavaEditor.this.openAt(getBegin());
        }

        /** Set the text contained in this section.
        * Newlines are automatically added to all text segments handled,
        * unless there was already one.
        * All guarded blocks must consist of entire lines.
        * This applies to the contents of specific guard types as well.
        * @param bounds the bounds indicating where the text should be set
        * @param text the new text
        * @param minLen If true the text has to have length more than 2 chars.
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        */
        protected boolean setText(PositionBounds bounds, String text, boolean minLen) {
            if (!valid)
                return false;

            // modify the text - has to end with new line and the length
            // has to be more then 2 characters
            if (minLen) {
                if (text.length() == 0)
                    text = " \n"; // NOI18N
                else if (text.length() == 1)
                    text = text.equals("\n") ? " \n" : text + "\n"; // NOI18N
            }

            if (!text.endsWith("\n")) // NOI18N
                text = text + "\n"; // NOI18N

            try {
                bounds.setText(text);
                return true;
            }
            catch (BadLocationException e) {
            }
            catch (IOException e) {
            }
            return false;
        }

        /** Marks or unmarks the section as guarded.
        * @param doc The styled document where this section placed in.
        * @param bounds The rangeof text which should be marked or unmarked.
        * @param mark true means mark, false unmark.
        */
        void markGuarded(StyledDocument doc, PositionBounds bounds, boolean mark) {
            int begin = bounds.getBegin().getOffset();
            int end = bounds.getEnd().getOffset();
            if (mark)
                NbDocument.markGuarded(doc, begin, end - begin);
            else
                NbDocument.unmarkGuarded(doc, begin, end - begin);
        }

        /** Marks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        abstract void markGuarded(StyledDocument doc);

        /** Unmarks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        abstract void unmarkGuarded(StyledDocument doc);

        /** Deletes the text in the section.
        * @exception BadLocationException
        * @exception IOException
        */
        abstract void deleteText() throws BadLocationException, IOException;

        /** Gets the begin of section. To this position is set the cursor
        * when section is open in the editor.
        */
        public abstract PositionRef getBegin();

        /** Gets the text contained in the section.
        * @return The text contained in the section.
        */
        public abstract String getText();

        /** Assures that a position is not inside the guarded section. Complex guarded sections
         * that contain portions of editable text can return true if the tested position is
         * inside one of such portions provided that permitHoles is true.
         * @param pos position in question
         * @param permitHoles if false, guarded section is taken as a monolithic block
         * without any holes in it regadless of its complexity.
 */
        public abstract boolean contains(PositionRef pos,boolean permitHoles);
        /** Returns a position after the whole guarded block that is safe for insertions.
 */
        public abstract PositionRef getPositionAfter();
        /** Returns position before the whole guarded block that is safe for insertions.
 */
        public abstract PositionRef getPositionBefore();
    }

    /** Represents a simple guarded section.
    * It consists of one contiguous block.
    */
    public final class SimpleSection extends GuardedSection {
        /** Text range of the guarded section. */
        PositionBounds bounds;

        /** Creates new section.
        * @param name Name of the new section.
        * @param bounds The range of the section.
        */
        SimpleSection(String name, PositionBounds bounds) {
            super(name);
            this.bounds = bounds;
        }

        /** Set the text of the section.
        * @param text the new text
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        * @see JavaEditor.GuardedSection#setText
        */
        public boolean setText(String text) {
            return setText(bounds, text, true);
        }

        /** Deletes the text in the section.
        * @exception BadLocationException
        * @exception IOException
        */
        void deleteText() throws BadLocationException, IOException {
            bounds.setText(""); // NOI18N
            deleteNewLineBeforeBlock(bounds.getBegin().getOffset());
        }

        /** Marks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        void markGuarded(StyledDocument doc) {
            markGuarded(doc, bounds, true);
        }

        /** Unmarks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        void unmarkGuarded(StyledDocument doc) {
            markGuarded(doc, bounds, false);
            JavaEditor.this.notifyModified();
        }

        /** Gets the begin of section. To this position is set the cursor
        * when section is open in the editor.
        * @return the begin position of section.
        */
        public PositionRef getBegin() {
            return bounds.getBegin();
        }

        /** Gets the text contained in the section.
        * @return The text contained in the section.
        */
        public String getText() {
            StringBuffer buf = new StringBuffer();
            try {
                buf.append(bounds.getText());
            }
            catch (Exception e) {
            }
            return buf.toString();
        }

        /*
        public String toString() {
          StringBuffer buf = new StringBuffer("SimpleSection:"+name);
          buf.append("\"");
          try {
            buf.append(bounds.getText());
          }
          catch (Exception e) {
            buf.append("EXCEPTION:");
            buf.append(e.getMessage());
          }
          buf.append("\"");
          return buf.toString();
    }*/
        
        public PositionRef getPositionAfter() {
	    return createPositionRef(bounds.getEnd().getOffset(), Position.Bias.Backward);
	}

        public boolean contains(PositionRef pos,boolean allowHoles) {
	    return bounds.getBegin().getOffset() <= pos.getOffset() &&
		bounds.getEnd().getOffset() >= pos.getOffset();
	}
        
        public PositionRef getPositionBefore() {
            return createPositionRef(bounds.getBegin().getOffset(), Position.Bias.Forward);
	}
    }

    /** Represents an advanced guarded block.
    * It consists of three pieces: a header, body, and footer.
    * The header and footer are guarded but the body is not.
    */
    public final class InteriorSection extends GuardedSection {
        /** Text range of the header. */
        PositionBounds header;

        /** Text range of the header. */
        PositionBounds body;

        /** Text range of the bottom. */
        PositionBounds bottom;

        /** Creates new section.
        * @param name Name of the new section.
        * @param bounds The range of the section.
        */
        InteriorSection(String name, PositionBounds header, PositionBounds body, PositionBounds bottom) {
            super(name);
            this.header = header;
            this.body = body;
            this.bottom = bottom;
        }

        /** Set the text of the body.
        * @param text the new text
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        * @see JavaEditor.GuardedSection#setText
        */
        public boolean setBody(String text) {
            return setText(body, text, false);
        }

        /** Set the text of the header.
        * @param text the new text
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        * @see JavaEditor.GuardedSection#setText
        */
        public boolean setHeader(String text) {
            return setText(header, text, true);
        }

        /** 
         * Returns the contents of the header part of the section. If the
         * section is invalid the method returns null.
         * @return contents of the header or null, if the section is not valid.
         */
        public String getHeader() {
            if (!isValid())
                return null;
            try {
                return header.getText();
            } catch (IOException ex) {
                // should not be never reached.
            } catch (BadLocationException ex) {
                // should not happen.
            }
            return null;
        }

        /** Set the text of the bottom.
        * Note that the bottom of the section must have exactly one line.
        * So, all interior newline characters will be replaced by spaces.
        *
        * @param text the new text
        * @return <code>true</code> if the operation was successful, otherwise <code>false</code>
        * @see JavaEditor.GuardedSection#setText
        */
        public boolean setBottom(String text) {
            boolean endsWithEol = text.endsWith("\n"); // NOI18N
            int firstEol = text.indexOf('\n');
            int lastEol = text.lastIndexOf('\n');

            if ((firstEol != lastEol) || (endsWithEol && (firstEol != -1))) {
                if (endsWithEol) {
                    text = text.substring(0, text.length() - 1);
                }
                text = text.replace('\n', ' ');
            }
            return setText(bottom, text, true);
        }

        /**
         * Returns the contents of the bottom part of the guarded section.
         * The method will return null, if the section is not valid.
         * @return contents of the bottom part, or null if the section is not valid.
         */
        public String getBottom() throws IOException, BadLocationException {
            if (!isValid())
                return null;
            try {
                return bottom.getText();
            } catch (IOException ex) {
                // should not be never reached.
            } catch (BadLocationException ex) {
                // should not happen.
            }
            return null;
        }
        
        /** Gets the begin of section. To this position is set the cursor
        * when section is open in the editor.
        * @return the begin position of the body section - the place where
        *         is possible to edit.
        */
        public PositionRef getBegin() {
            return body.getBegin();
        }

        /** Deletes the text in the section.
        * @exception BadLocationException
        * @exception IOException
        */
        void deleteText() throws BadLocationException, IOException {
            header.setText(""); // NOI18N
            body.setText(""); // NOI18N
            bottom.setText(""); // NOI18N
            deleteNewLineBeforeBlock(header.getBegin().getOffset());
        }

        /** Marks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        void markGuarded(StyledDocument doc) {
            markGuarded(doc, header, true);
            markGuarded(doc, bottom, true);
        }

        /** Unmarks the section as guarded.
        * @param doc The styled document where this section placed in.
        */
        void unmarkGuarded(StyledDocument doc) {
            markGuarded(doc, header, false);
            markGuarded(doc, bottom, false);
            JavaEditor.this.notifyModified();
        }

        /** Gets the text contained in the section.
        * @return The text contained in the section.
        */
        public String getText() {
            StringBuffer buf = new StringBuffer();
            try {
                buf.append(header.getText());
                buf.append(body.getText());
                buf.append(bottom.getText());
            }
            catch (Exception e) {
            }
            return buf.toString();
        }

        /*
        public String toString() {
          StringBuffer buf = new StringBuffer("InteriorSection:"+name);
          try {
            buf.append("HEADER:\"");
            buf.append(header.getText());
            buf.append("\"");
            buf.append("BODY:\"");
            buf.append(body.getText());
            buf.append("\"");
            buf.append("BOTTOM:\"");
            buf.append(bottom.getText());
            buf.append("\"");
          }
          catch (Exception e) {
            buf.append("EXCEPTION:");
            buf.append(e.getMessage());
          }
          return buf.toString();
    }*/
        public boolean contains(PositionRef pos,boolean allowHoles) {
	    if (!allowHoles) {
    		return header.getBegin().getOffset() <= pos.getOffset() &&
		    bottom.getEnd().getOffset() >= pos.getOffset();
	    } else {
		if (header.getBegin().getOffset() <= pos.getOffset() && 
		    header.getEnd().getOffset() >= pos.getOffset()) {
		    return true;
		}
		return bottom.getBegin().getOffset() <= pos.getOffset() &&
		    bottom.getEnd().getOffset() >= pos.getOffset();
	    }
	}
        
        public PositionRef getPositionBefore() {
	    return createPositionRef(header.getBegin().getOffset(), Position.Bias.Forward);
	}
        
        public PositionRef getPositionAfter() {
	    return createPositionRef(bottom.getEnd().getOffset(), Position.Bias.Backward);
	}
    }

    // ==================== Private inner classes ===========================

    /** Comparator of the guarded sections. It compares the begin position
    * of the sections.
    */
    private static class GuardedPositionComparator implements Comparator {
        /** Compare two objects. Both have to be either SimpleSection
        * either InteriorSection instance.
        */
        public int compare(Object o1, Object o2) {
            return getOffset(o1) - getOffset(o2);
        }

        /** Computes the offset of the begin of the section. */
        private int getOffset(Object o) {
            if (o instanceof SimpleSection) {
                return ((SimpleSection)o).bounds.getBegin().getOffset();
            }
            else {
                return ((InteriorSection)o).header.getBegin().getOffset();
            }
        }
    }

    /** Class for holding information about the one special (guarded)
    * comment. It is created by GuardedReader and used by
    * JavaEditor to creating the guarded sections.
    */
    private static class SectionDesc {
        /** Type - one of T_XXX constant */
        int type;

        /** Name of the section comment */
        String name;

        /** offset of the begin */
        int begin;

        /** offset of the end */
        int end;

        /** Simple constructor */
        SectionDesc(int type) {
            this.type = type;
            name = null;
            begin = 0;
            end = 0;
        }
    }

    /** This stream is able to filter special guarded comments.
     * Holding this information is optional and depends on the construction
     * of this stream - the reason of this feature is that
     * GuardedReader is used also for parser (and it doesn't require
     * the storing the guarded block information - just filter the comments).
     */
    static class GuardedReader extends Reader {
        /** Encapsulated reader */
        Reader reader;

        /** Character buffer */
        char[] charBuff;
        char[] readBuff;
        int howmany;
        RE magicsAsRE;

        /** The flag determining if this stream should store the guarded
         * block information (list of the SectionDesc).
         */
        boolean justFilter;

        /** The position at the current line. */
        int position;

        /** The list of the SectionsDesc. */
        LinkedList list;

        /** The count of types new line delimiters used in the file */
        final int[] newLineTypes;

        /** Creates new stream.
         * @param is encapsulated input stream.
         * @param justFilter The flag determining if this stream should
         *        store the guarded block information. True means just filter,
         *        false means store the information.
         */
        public GuardedReader(InputStream is, boolean justFilter) {
            reader = new InputStreamReader(is);
            this.justFilter = justFilter;
            position = 0;
            list = new LinkedList();
            newLineTypes = new int[] { 0, 0, 0 };
        }

        /** Read the array of chars */
        public int read(char[] cbuf, int off, int len) throws IOException {

            if (charBuff == null) {
                readCharBuff();
                translateToCharBuff();
            }

            if (howmany <= 0) {
                return -1;
            } else {
                int min = Math.min(len, howmany);
                System.arraycopy(charBuff, position, cbuf, off, min);
                howmany -= min;
                position += min;
                return min;
            }
        }

        /** Reads readBuff */
        final void readCharBuff() throws IOException {

            char[] tmp = new char[2048];
            int read;
            ArrayList buffs = new ArrayList(20);

            for (;;) {
                read = readFully(tmp);
                buffs.add(tmp);
                if (read < 2048) {
                    break;
                } else {
                    tmp = new char[2048];
                }
            }

            int listsize = buffs.size() - 1;
            int size = listsize * 2048 + read;
            readBuff = new char[size];
            charBuff = new char[size];
            int copy = 0;

            for (int i = 0; i < listsize; i++) {
                char[] tmp2 = (char[]) buffs.get(i);
                System.arraycopy(tmp2, 0, readBuff, copy, 2048);
                copy += 2048;
            }
            System.arraycopy(tmp, 0, readBuff, copy, read);
        }

        /** reads fully given buffer */
        final int readFully(final char[] buff) throws IOException {
            int read = 0;
            int sum = 0;

            do {
                read = reader.read(buff, sum, buff.length - sum);
                sum += read;
            } while ((sum < buff.length) && (read > 0));

            return sum + 1;
        }

        /** Called after raw filling from an underlying reader */
        final void translateToCharBuff() {
            position = 0;

            // points to first unused cell in charBuff
            int charBuffPtr = 0;
            int stop = readBuff.length - 1;

            // read char
            int c;
            // ptr to first not processed char in readBuff
            int i = 0;
            // points to a character right after a newline
            int lastNewLine = 0;

            // final automata
            int fatpos = 0;
            final int MAGICLEN = MAGIC_PREFIX.length();


            //process newlines so only '\n' appears in the charBuff
            //count all kinds of newlines - most used will be used on save
            while (i < stop) {
                c = readBuff[i];
                switch (c) {
                case (int)'\n':
                    newLineTypes[NEW_LINE_N]++;
                    charBuff[charBuffPtr++] = '\n';
                    lastNewLine = charBuffPtr;
                    i++;
                    break;
                case (int)'\r':
                    int c2 = readBuff[i + 1];
                    if (c2 != (int)'\n') {
                        newLineTypes[NEW_LINE_R]++;
                        i++;
                    } else {
                        i +=2;
                        newLineTypes[NEW_LINE_RN]++;
                    }
                    charBuff[charBuffPtr++] = '\n';
                    lastNewLine = charBuffPtr;
                    break;

                default:
                    charBuff[charBuffPtr++] = readBuff[i++];
                }

                switch (fatpos) {
                case 0:
                    if (c == '/') {
                        fatpos++;
                    } else {
                        fatpos = 0;
                    }
                    break;

                case 1:
                    if (c == '/') {
                        fatpos++;
                    } else {
                        fatpos = 0; 
                    }
                    break;

                case 2:
                    if (c == 'G') {
                        fatpos++;
		    } else if (c == '/') {
			fatpos = 2; // what if /////GEN-xxx?
                    } else {
                        fatpos = 0;
                    }
                    break;

                case 3:
                    if (c == 'E') {
                        fatpos++;
                    } else {
                        fatpos = 0;
                    }
                    break;

                case 4:
                    if (c == 'N') {
                        fatpos++;
                    } else {
                        fatpos = 0;
                    }
                    break;

                case 5:
                    if (c == '-') {
                        fatpos++;
                    } else {
                        fatpos = 0;
                    }
                    break;

                default:
                    fatpos = 0;
                }

                // "//GEN-" was reached at this time
                if(fatpos == MAGICLEN) {
                    fatpos = 0;
                    RE magics = getMagicsAsRE();
                    int searchLen = Math.min(LONGEST_ITEM, readBuff.length - i);
                    CharacterIterator chi = new CharacterArrayCharacterIterator(readBuff, i, searchLen);
                    if (magics.match(chi, 0)) {
                        String match = magics.getParen(0).toString();

                        charBuffPtr -= MAGIC_PREFIX.length();
                        i += match.length();

                        if (!justFilter) {
                            int type = string2Type(match);
                            SectionDesc desc = new SectionDesc(type);
                            int start = charBuffPtr + SECTION_MAGICS[type].length();
                            desc.begin = lastNewLine;
                            desc.end = charBuffPtr + 1;
                            int toNl = toNewLine(i);
                            desc.name = new String(readBuff, i, toNl);
                            list.add(desc);
                            i += toNl;
                        } else {
                            i += toNewLine(i);
                        }

                    }
                }
            }

            if (i == stop) {
                c = readBuff[i];
                switch (c) {
                case (int)'\n':
                    newLineTypes[NEW_LINE_N]++;
                    charBuff[charBuffPtr++] = '\n';
                    break;
                case (int)'\r':
                    newLineTypes[NEW_LINE_RN]++;
                    charBuff[charBuffPtr++] = '\n';
                    break;

                default:
                    charBuff[charBuffPtr++] = readBuff[i++];
                }
            }

            // repair last SectionDesc
            if (!justFilter && (list.size() > 0)) {
                SectionDesc desc = (SectionDesc) list.getLast();
                if (desc.end > charBuffPtr) {
                    desc.end = charBuffPtr;
                }
            }

            howmany = charBuffPtr;
            readBuff = null;
        }

        /** Translates a String (//GEN-XXX) to its number */
        static int string2Type(String match) {
            StringBuffer sb = new StringBuffer(MAGIC_PREFIX);
            sb.append(match);
            match = sb.toString();

            final int len = SECTION_MAGICS.length;

            for (int i = 0; i < len; i++) {
                if (match.equals(SECTION_MAGICS[i])) {
                    return i;
                }
            }
            return -1;
        }

        /** Searches for newline from i */
        final int toNewLine(int i) {
            int c;
            int counter = i;
            final int len = readBuff.length;
            while (counter < len) {
                c = readBuff[counter++];
                if (c == '\r' || c == '\n') {
                    counter--;
                    break;
                }
            }

            return counter - i;
        }

        /** @return searching engine for magics */
        final RE getMagicsAsRE() {
            if (magicsAsRE == null) {
                try {
                    RECompiler rec = new RECompiler();
                    magicsAsRE = new RE(rec.compile(makeOrRegexp()));
                } catch (org.apache.regexp.RESyntaxException e) {
                    throw new IllegalStateException();
                }
            }
            return magicsAsRE;
        }

        /** Makes or regular expression for magics */
        final String makeOrRegexp() {
            StringBuffer sb = new StringBuffer(100);
            final int len = MAGIC_PREFIX.length();
            final int slen = SECTION_MAGICS.length - 1;
            for (int i = 0; i < slen; i++) {
                sb.append(SECTION_MAGICS[i].substring(len));
                sb.append('|');
            }
            sb.append(SECTION_MAGICS[slen].substring(len));
            return sb.toString();
        }

        /** @return most frequently type of new line delimiter */
        byte getNewLineType() {
            if (newLineTypes[0] > newLineTypes[1]) {
                return (newLineTypes[0] > newLineTypes[2]) ? (byte) 0 : 2;
            }
            else {
                return (newLineTypes[1] > newLineTypes[2]) ? (byte) 1 : 2;
            }
        }

        /** Close underlayed writer. */
        public void close() throws IOException {
            reader.close();
        }
    }

    /** This stream is able to insert special guarded comments.
    */
    static class GuardedWriter extends Writer {
        /** Encapsulated writer. */
        BufferedWriter writer;

        /** From this iterator is possible to obtain all section
        * descriptions during writing the document.
        */
        Iterator sections;

        /** Current section from the previous iterator. For filling this
        * field is used method nextSection.
        */
        SectionDesc current;

        /** Current offset in the original document (NOT in the encapsulated
        * output stream.
        */
        int offsetCounter;

        /** This flag is used during writing. It is complicated to explain. */
        boolean wasNewLine;

        /** Creates new GuardedWriter.
        * @param os Encapsulated output stream.
        * @param list The list of the guarded sections.
        */
        public GuardedWriter(OutputStream os, ArrayList list) {
            writer = new BufferedWriter(new OutputStreamWriter(os));
            offsetCounter = 0;
            sections = prepareSections(list);
            nextSection();
            wasNewLine = false;
        }

        /** Writes chars to underlayed writer */
        public void write(char[] cbuf, int off, int len) throws IOException {
            for (int i = 0; i < len; i++) {
                writeOneChar(cbuf[i + off]);
            }
        }

        /** Calls underlayed writer flush */
        public void close() throws IOException {
            writer.flush();
        }

        /** Calls underlayed writer flush */
        public void flush() throws IOException {
            writer.flush();
        }

        /** This method prepares the iterator of the SectionDesc classes
        * @param list The list of the GuardedSection classes.
        * @return iterator of the SectionDesc
        */
        private Iterator prepareSections(ArrayList list) {
            LinkedList dest = new LinkedList();
            Collections.sort(list, new GuardedPositionComparator());

            Iterator it = list.iterator();
            while (it.hasNext()) {
                GuardedSection o = (GuardedSection) it.next();
                if (o instanceof SimpleSection) {
                    SectionDesc desc = new SectionDesc(T_LINE);
                    desc.name = o.name;
                    desc.begin = ((SimpleSection)o).bounds.getBegin().getOffset();
                    desc.end = ((SimpleSection)o).bounds.getEnd().getOffset();
                    dest.add(desc);
                }
                else {
                    SectionDesc desc = new SectionDesc(T_HEADER);
                    desc.begin = (((InteriorSection)o).header).getBegin().getOffset();
                    desc.end = (((InteriorSection)o).header).getEnd().getOffset();
                    desc.name = o.name;
                    dest.add(desc);

                    desc = new SectionDesc(T_END);
                    desc.begin = (((InteriorSection)o).bottom).getBegin().getOffset();
                    desc.end = (((InteriorSection)o).bottom).getEnd().getOffset();
                    desc.name = o.name;
                    dest.add(desc);
                }
            }
            return dest.iterator();
        }

        /** Write one character. If there is a suitable place,
        * some special comments are written to the underlaying stream.
        * @param b char to write.
        */
        void writeOneChar(int b) throws IOException {
            if (b == '\r')
                return;

            if (current != null) {
                if (offsetCounter == current.begin) {
                    wasNewLine = false;
                }
                if ((b == '\n') && (current.begin < offsetCounter)) {
                    switch (current.type) {
                    case T_LINE:
                        if (!wasNewLine) {
                            if (offsetCounter + 1 >= current.end) {
                                writeMagic(T_LINE, current.name);
                                nextSection();
                            }
                            else {
                                writeMagic(T_BEGIN, current.name);
                                wasNewLine = true;
                            }
                        }
                        else {
                            if (offsetCounter + 1 >= current.end) {
                                writeMagic(T_END, current.name);
                                nextSection();
                            }
                        }
                        break;

                    case T_HEADER:
                        if (!wasNewLine) {
                            if (offsetCounter + 1 >= current.end) {
                                writeMagic(T_FIRST, current.name);
                                nextSection();
                            }
                            else {
                                writeMagic(T_FIRST, current.name);
                                wasNewLine = true;
                            }
                        }
                        else {
                            if (offsetCounter + 1 >= current.end) {
                                writeMagic(T_HEADEREND, current.name);
                                nextSection();
                            }
                        }
                        break;

                    case T_END:
                        writeMagic(T_LAST, current.name);
                        nextSection();
                        break;
                    }
                }
            }

            writer.write(b);
            offsetCounter++;
        }

        /** Try to get next sectionDesc from the 'sections'
        * If there is no more section the 'current' will be set to null.
        */
        private void nextSection() {
            current = (SectionDesc) (sections.hasNext() ? sections.next() : null);
        }

        /** Writes the magic to the underlaying stream.
        * @param type The type of the magic section - T_XXX constant.
        * @param name name of the section.
        */
        private void writeMagic(int type, String name) throws IOException {
            writer.write(SECTION_MAGICS[type], 0, SECTION_MAGICS[type].length());
            writer.write(name, 0, name.length());
        }
    }

    /** This stream is used for changing the new line delimiters.
    * It replaces the '\n' by '\n', '\r' or "\r\n"
    */
    private static class NewLineOutputStream extends OutputStream {
        /** Underlaying stream. */
        OutputStream stream;

        /** The type of new line delimiter */
        byte newLineType;

        /** Creates new stream.
        * @param stream Underlaying stream
        * @param newLineType The type of new line delimiter
        */
        public NewLineOutputStream(OutputStream stream, byte newLineType) {
            this.stream = stream;
            this.newLineType = newLineType;
        }

        /** Write one character.
        * @param b char to write.
        */
        public void write(int b) throws IOException {
            if (b == '\n') {
                switch (newLineType) {
                case NEW_LINE_R:
                    stream.write('\r');
                    break;
                case NEW_LINE_RN:
                    stream.write('\r');
                case NEW_LINE_N:
                    stream.write('\n');
                    break;
                }
            }
            else {
                stream.write(b);
            }
        }
    }
}
