/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.io.File;
import java.io.ObjectInputStream;
import java.io.IOException;

import org.openide.loaders.DataObject;
import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerType;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.ExternalCompiler;
import org.openide.compiler.ExternalCompiler.ErrorExpression;
import org.openide.compiler.ExternalCompilerGroup;
import org.openide.execution.NbProcessDescriptor;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileSystemCapability;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Utilities;
import org.openide.filesystems.FileObject;

import org.netbeans.modules.java.settings.JavaSettings;

/**
 *
 * @author Ales Novak
 */
public class JavaExternalCompilerType extends JavaCompilerType {
    
    /** generated Serialized Version UID */
    static final long serialVersionUID = 2163925425759211535L;
    
    /** externalCompiler property name */
    public static final String PROP_EXTERNAL_COMPILER = "externalCompilerType"; // NOI18N

    /** errorDescriptions property name 
     * @deprecated incorrect name, not used any more. Should be named PROP_ERROR_EXPRESSION
     */
    public static final String PROP_ERROR_DESCRIPTIONS = "errorDescriptions"; // NOI18N
    
    public static final String PROP_ERROR_EXPRESSION = "errorExpression"; // NOI18N
    
    /** Name of "working directory" property */
    public static final String PROP_WORKING_DIRECTORY = "workingDirectory"; // NOI18N
    /** Name of "Environment variables" property. */
    public static final String PROP_ENVIRONMENT_VARS = "environmentVariables"; // NOI18N
    /** Name of "optimize tag replace" property. */
    public static final String PROP_OPTIMIZE_REPLACE = "optimizeReplace"; // NOI18N
    /** Name of "deprecation tag replace property. */
    public static final String PROP_DEPRECATION_REPLACE = "deprecationReplace"; // NOI18N
    /** Name of "debug info tag replace" property. */
    public static final String PROP_DEBUGINFO_REPLACE = "debuginfoReplace"; // NOI18N
    /** Name of "encoding tag replace" property. */
    public static final String PROP_ENCODING_REPLACE = "encodingReplace"; // NOI18N
    /** Name of "bootclasspath option tag replace" property */
    public static final String PROP_BOOTCLASSPATH_REPLACE = "bootclasspathReplace"; // NOI18N
    
    /** name of outputdir property 
	@deprecated Superseeded by JavaCompilerType.PROP_TARGET_FILESYSTEM
     */
    public static final String PROP_OUTPUTDIR = "outputDir"; // NOI18N
    
    private static final String DEFAULT_BOOTCLASSPATH_REPLACE = "-bootclasspath "; // NOI18N
    
    /** serialzed user defined compilers */
    private ErrorExpression errorExpression = ExternalCompiler.JAVAC;
    
    /** The settings for external compiler's process (executable and classpath */
    protected NbProcessDescriptor externalCompiler;
    
    private String optimizeReplace = "-O"; // NOI18N
    private String debuginfoReplace = "-g"; // NOI18N
    private String deprecationReplace = "-deprecation"; // NOI18N
    private String encodingReplace = "-encoding"; // NOI18N
    private String dReplace = "-d"; // NOI18N
    private String bootclasspathReplace = DEFAULT_BOOTCLASSPATH_REPLACE; // NOI18N
    
    /** class path settings or null 
     * for compatibility only; propagated to JavaCompilerType during readResolve
     */
    private NbClassPath classPath;

    /** boot class path or null 
     * for compatibility only; propagated to JavaCompilerType during readResolve
     */
    private NbClassPath bootClassPath;

    /** environment vars or null */
    private String[] envp = null;

    /** working directory or null */
    private File cwd = null;
    
    private boolean classPathCleared;
    
    public JavaExternalCompilerType() {
        classPathCleared = true;
    }
    
    public Object clone() throws CloneNotSupportedException {
        JavaExternalCompilerType other = (JavaExternalCompilerType)super.clone();
	if (this.errorExpression != null)
    	    other.errorExpression = (ErrorExpression)this.errorExpression.clone();
	if (this.envp != null)
            other.envp = (String[])this.envp.clone();
        return other;
    }
    
    /** human presentable name */
    public String displayName() {
        // Default only matters for regular external, which is
        // instantiated using Class.newInstance:
        return getString("CTL_ExternalCompilerType");
    }
    
    public HelpCtx getHelpCtx () {
        return new HelpCtx(JavaExternalCompilerType.class);
    }
    
    public NbProcessDescriptor getExternalCompiler() {
        if (externalCompiler == null) {
            externalCompiler = createDescriptor();
        } else if (!classPathCleared) {
            // clear {classpath} from the descriptor's command line.
            String arguments = externalCompiler.getArguments();
            String searched = "{" + JExternalCompilerGroup.JFormat.TAG_CLASSPATH + "}";
            String searched2 = "{" + JExternalCompilerGroup.JFormat.TAG_PATHSEPARATOR + "}";

            int pos = arguments.indexOf(searched);
            if (pos != -1) {
                int after = pos + searched.length();
                if (arguments.length() >= after + searched2.length() && 
                    searched2.regionMatches(0, arguments, after, searched2.length())) {
                    after += searched2.length();
                }
                StringBuffer sb = new StringBuffer(arguments);
                sb.delete(pos, after);
                
                externalCompiler = new NbProcessDescriptor(
                    externalCompiler.getProcessName(),
                    sb.toString(),
                    externalCompiler.getInfo()
                );
            }
            classPathCleared = true;
        }
        return externalCompiler;
    }
    
    /** Creates standard NbProcessDescriptor for external javac */
    NbProcessDescriptor createDescriptor() {
        return new NbProcessDescriptor(
            "{" + ExternalCompilerGroup.Format.TAG_JDKHOME + "}{" + // NOI18N
            ExternalCompilerGroup.Format.TAG_SEPARATOR + "}bin{" + ExternalCompilerGroup.Format.TAG_SEPARATOR + "}javac", // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_DEBUGINFO + "} " + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_DEPRECATION + "} " + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_OPTIMIZE + "} " + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_OUTDIR + "} " + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_ENCODING + "} " + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_OPTION_BOOTCLASSPATH + "}" + // NOI18N
            "{" + JExternalCompilerGroup.JFormat.TAG_BOOTCLASSPATH + "} " + // NOI18N
            "-classpath {" + ExternalCompilerGroup.Format.TAG_REPOSITORY + "}{" + ExternalCompilerGroup.Format.TAG_PATHSEPARATOR + "}" + // NOI18N
            "{" + ExternalCompilerGroup.Format.TAG_CLASSPATH + "} " + // NOI18N
            "{" + ExternalCompilerGroup.Format.TAG_FILES + "}", // NOI18N
            getAdditionalHint()
        );
    }
    
    public void setExternalCompiler(NbProcessDescriptor compiler) {
        NbProcessDescriptor old = this.externalCompiler;
        externalCompiler = compiler;
        firePropertyChange(PROP_EXTERNAL_COMPILER, old, compiler);
    }
    
    public ErrorExpression getErrorExpression() {
        return errorExpression;
    }
    
    public void setErrorExpression(ErrorExpression e) {
        if (e == null) {
            throw new IllegalArgumentException();
        }
        ErrorExpression old = errorExpression;
        errorExpression = e;
        firePropertyChange(PROP_ERROR_EXPRESSION, old, e);
    }
    
    public void setOptimizeReplace(String x) {
        String old = optimizeReplace;
        optimizeReplace = x;
        firePropertyChange(PROP_OPTIMIZE_REPLACE, old, x);
    }
    public void setDebuginfoReplace(String x) {
        String old = debuginfoReplace;
        debuginfoReplace = x;
        firePropertyChange(PROP_DEBUGINFO_REPLACE, old, x);
    }
    public void setDeprecationReplace(String x) {
        String old = deprecationReplace;
        deprecationReplace = x;
        firePropertyChange(PROP_DEPRECATION_REPLACE, old, x);
    }
    public void setEncodingReplace(String x) {
        String old = encodingReplace;
        encodingReplace = x;
        firePropertyChange(PROP_ENCODING_REPLACE, old, x);
    }
    
    public void setBootClassPathReplace(String x) {
        String old = bootclasspathReplace;
        bootclasspathReplace = x;
        firePropertyChange(PROP_BOOTCLASSPATH_REPLACE, old, x);
    }
    
    public void setDReplace(String x) {
        dReplace = x;
    }
    
    public String getOptimizeReplace() {
        return optimizeReplace;
    }
    public String getDebuginfoReplace() {
        return debuginfoReplace;
    }
    public String getDeprecationReplace() {
        return deprecationReplace;
    }
    public String getEncodingReplace() {
        return encodingReplace;
    }
    public String getDReplace() {
        return dReplace;
    }
    
    public String getBootClassPathReplace() {
        return bootclasspathReplace;
    }
    
    /** Getter for repository path. It is immutable reflecting
    * NbClassPath.createRepositoryPath (). Is here only to be displayed in property sheet.
    */
    public NbClassPath getRepositoryPath() {
        return NbClassPath.createRepositoryPath(FileSystemCapability.COMPILE);
    }

    /** Getter for repository path. It is immutable reflecting
    * NbClassPath.createLibraryPath (). Is here only to be displayed in property sheet.
    */
    public NbClassPath getLibraryPath() {
        return NbClassPath.createLibraryPath();
    }

    /** Get environment variables.
    * @return the <code><i>NAME</i>=<i>VALUE</i></code> pairs, or <code>null</code> (typically, inherit that of parent)
    */
    public String[] getEnvironmentVariables() {
        return envp;
    }

    /** Set environment variables.
    * @param nue the new variables
    * @see #getEnvironmentVariables
    */
    public synchronized void setEnvironmentVariables(String[] nue) {
        String[] old = envp;
        envp = nue;
        firePropertyChange(PROP_ENVIRONMENT_VARS, old, nue); // NOI18N
    }

    /** Get the working directory.
    * Note that using a nondefault working directory will only work on JDK 1.3.
    * @return the working directory to use, or <code>null</code> (use that of parent)
    */
    public File getWorkingDirectory() {
        return cwd;
    }

    /** Set the working directory.
    * @param nue the new directory
    * @see #getWorkingDirectory
    */
    public synchronized void setWorkingDirectory(File nue) {
        File old = cwd;
        cwd = nue;
        firePropertyChange(PROP_WORKING_DIRECTORY, old, nue); // NOI18N
    }    
    
    /** @return compiler group class */
    protected Class getCompilerGroupClass() {
        return JExternalCompilerGroup.class;
    }
    
    /** @return compiler for given classtype and file object */
    protected Compiler createCompiler(Class type, FileObject fo) {
        return new JExternalCompiler(fo, type, getExternalCompiler(), getErrorExpression(), this, getTargetFileSystem());
    }
    
    public void assureValidTargetFS() throws IOException {
	super.assureValidTargetFS();
	
	FileSystem fs= getTargetFileSystem();
	if (fs == null)
	    return;

	java.io.File f = org.openide.execution.NbClassPath.toFile(fs.getRoot());
	
	if (f == null || !f.exists() || !f.isDirectory()) {
	    String message = 
		java.text.MessageFormat.format(
		    JavaCompilerType.getString("MSG_FS_NonLocal"),
		    new Object[] {
			fs.getDisplayName()
		    });
		    
	    throw (IOException)org.openide.TopManager.getDefault().getErrorManager().annotate(
		new IOException(message),
		org.openide.ErrorManager.USER,
		message,
		null,
		null, null
	    );
	}
    }
    
    protected Object readResolve() {
        this.bootClassPath = null;
        this.classPath = null;
        if (version < 1) {
            // reset the external compiler to the default one:
            externalCompiler = null;
            // set the -bootclasspath replacement:
            bootclasspathReplace = DEFAULT_BOOTCLASSPATH_REPLACE;
            version = currentVersion;
        }
        return this;
    }
    
    /** @return additional hints for java's external compiler types */
    final static String getAdditionalHint() {
        java.util.ResourceBundle bundle = NbBundle.getBundle("org.openide.compiler.Bundle");
        String hint0 = bundle.getString ("MSG_ExternalCompilerHint");
        bundle = NbBundle.getBundle(JavaExternalCompilerType.class);
        String hint1 = bundle.getString("MSG_ExternalCompilerHint");
        return hint0 + hint1;
    }
}
