/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.io.*;
import java.util.*;
import java.beans.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Image;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.Timer;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
  
import org.openide.*;
import org.openide.util.Task;
import org.openide.cookies.OpenCookie;
import org.openide.cookies.CompilerCookie;
import org.openide.compiler.CompilerJob;
import org.openide.filesystems.*;
import org.openide.nodes.*;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObject;
import org.openide.loaders.ExecSupport;
import org.openide.loaders.CompilerSupport;
import org.openide.loaders.MultiDataObject;
import org.openide.src.*;
import org.openide.src.nodes.ClassCustomizer;
import org.openide.src.nodes.SourceChildren;
import org.openide.util.Utilities;
import org.openide.util.WeakListener;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.util.RequestProcessor;
import org.openide.util.datatransfer.NewType;
import org.openide.util.Task;

import org.netbeans.modules.java.tools.BadgeCache;
import org.netbeans.modules.java.parser.JavaParser;
import org.netbeans.modules.java.settings.JavaSettings;

/** The node representation of <code>JavaDataObject</code> for Java sources.
*
* @author Petr Hamernik
*/
public class JavaNode extends DataNode {

    /** generated Serialized Version UID */
    static final long serialVersionUID = -7396485743899766258L;

    private static final String EXECUTION_SET_NAME = "Execution"; // NOI18N

    private static final String PROP_SYNCHRONIZATION_MODE = "SynchMode"; // NOI18N

    private static final String ICON_BASE = "org/netbeans/modules/java/resources/"; // NOI18N
    private static final String BARE_ICON_BASE = "org/netbeans/modules/java/resources/class"; // NOI18N

    private static final String BADGE_MAIN = ICON_BASE + "executable-badge";
    private static final String BADGE_ERROR = ICON_BASE + "error-badge";
    private static final String BADGE_NEEDS_COMPILE = ICON_BASE + "needs-compile";
    
    private static final String[] ICON_ATTRS_NONE = {};
    
    protected static final String ICON_ATTR_NEEDS_COMPILE = "needsCompile";
    /**
     * Icon attribute -- main class.
     */
    protected static final String   ICON_ATTR_MAIN = "mainClass";
    
    /**
     * Icon attribute -- errors in the source.
     */
    protected static final String   ICON_ATTR_ERROR = "error";

    private transient boolean preparsed = false;

    static final java.util.ResourceBundle bundle = NbBundle.getBundle(JavaNode.class);

    private static final String[] NEW_TYPE_NAMES = new String[] {
        bundle.getString("MENU_CREATE_CLASS"), bundle.getString("MENU_CREATE_INTERFACE")
    };

    private static final String[] NEW_TYPE_TITLES = new String[] {
        bundle.getString("TITLE_CREATE_CLASS"), bundle.getString("TITLE_CREATE_INTERFACE")
    };
    
    private static final int ICON_REFRESH_DELAY_MSECS = 1000;

    private BadgeCache  iconCache;
    
    private String[]    currentBadges;
    
    private Task        iconRefreshTask;
    
    private IconResolver iconResolver;

    /**
     * Last known/displayed status of the source.
     */
    private int         lastGoodStatus;
    
    /**
     * Up-to-date status that will be used for badging. 1 - OK, 0 - old,
     * 2 - needs to be determined.
     */
    private int         upToDate = 2;
    
    /** Create a node for the Java data object using the default children.
    * @param jdo the data object to represent
    */
    public JavaNode (JavaDataObject jdo) {
        this(jdo, new SourceChildren(JavaDataObject.getExplorerFactory(), jdo.getSource()));
    }
    
    private JavaParser findParser() {
        return (JavaParser)getDataObject().getCookie(JavaParser.class);
    }
    
    /**
     * This method is called during node initialization and whenever the base icon is
     * replaced to refill the badge cache with custom badges. The JavaNode's implementation
     * adds executable and error badges. Override to add your own badges, or replace the
     * default ones.
     */
    protected void initializeBadges(BadgeCache cache) {
        cache.registerBadge(ICON_ATTR_MAIN, BADGE_MAIN, 10, 5);
        cache.registerBadge(ICON_ATTR_ERROR, BADGE_ERROR, 8, 8);
        cache.registerBadge(ICON_ATTR_NEEDS_COMPILE, BADGE_NEEDS_COMPILE, 16, 0);
    }
    
    /** Create a node for the Java data object with configurable children.
    * Subclasses should use this constructor if they wish to provide special display for the child nodes.
    * Typically this would involve creating a subclass of {@link SourceChildren} based
    * on the {@link JavaDataObject#getSource provided source element}; the children list
    * may have extra nodes {@link Children#add(Node[]) added}, either at the {@link Children.Keys#setBefore beginning or end}.
    * @param jdo the data object to represent
    * @param children the children for this node
    */
    public JavaNode (JavaDataObject jdo, Children children) {
        super (jdo, children);
        initialize();
    }
    
    public void setIconBase(String base) {
        super.setIconBase(base);
        synchronized (this) {
            iconCache = BadgeCache.getCache(base);
            initializeBadges(iconCache);
        }
    }
    
    private PropL              mixedListener;
    private FileChangeListener fileListener;
    
    private class PropL extends FileChangeAdapter implements PropertyChangeListener,
        ChangeListener {
        public void propertyChange(PropertyChangeEvent ev) {
            JavaDataObject d = getJavaDataObject();
            if (ev.getSource() == d) {
                dataPropertyChange(ev);
            }
        }
        
        public void stateChanged(ChangeEvent ev) {
            sourceParsed();
        }
        
        public void fileChanged(FileEvent e) {
            fileSetChanged(false);
        }
    }
    
    private void updateCompiledStatus(boolean indicationEnabled) {
        if (indicationEnabled)
            fileSetChanged(true);
        else {
            StateUpdater.getInstance().updateNode(this);
        }
    }

    /**
     * Reattaches listener to all files in the current set.
     */
    private void attachToFiles() {
        java.util.Collection s = getJavaDataObject().secondaryEntries();
        for (java.util.Iterator it = s.iterator(); it.hasNext(); ) {
            FileObject f = ((MultiDataObject.Entry)it.next()).getFile();
            // remove the listener, if it is there already and reattach
            f.removeFileChangeListener(fileListener);
            f.addFileChangeListener(fileListener);
        }
    }
    
    private void initialize () {
        setIconBase(getBareIconBase());

        SourceElement source = getJavaDataObject().getSource();
        // add a source property change listener that will modify the node's icon
        // as the source's status change.
        mixedListener = new PropL();
        fileListener = WeakListener.fileChange(mixedListener, null);
        getJavaDataObject().addPropertyChangeListener(WeakListener.propertyChange(mixedListener, getJavaDataObject()));
        
        JavaParser p = findParser();
        if (p != null) {
            p.addChangeListener(WeakListener.change(mixedListener, p));
        }
        StateUpdater.getInstance().registerNode(this);
        attachToFiles();
    }

    void sourceParsed() {
        JavaParser p = findParser();
        int status = translateSourceStatus(p.getStatus());
        if (status == SourceElement.STATUS_NOT || lastGoodStatus == status)
            return;
        requestResolveIcons();
    }

    /**
     * Translates the source status so it can be compared with the currently
     * displayed one. Eliminates some transitions that should not bother
     * the end-user watching the Explorer.
     */
    int translateSourceStatus(int sourceStatus) {
        if (sourceStatus == SourceElement.STATUS_NOT)
            return lastGoodStatus;
        if (sourceStatus == SourceElement.STATUS_OK ||
            lastGoodStatus == SourceElement.STATUS_NOT)
            return sourceStatus;
        
        if (getJavaDataObject().isModified())
            return lastGoodStatus;
        else
            return sourceStatus;
    }

    /**
     * Called when something changes in files managed by the dataobject.
     * @param filesAddedRemoved true, if some files were taken in/out of the data object.
     * false, if only file(s) changed.
     */
    protected void fileSetChanged(boolean filesAddedRemoved) {
        // set of files has changed -- wait for a while, and then request icon resolving.
        requestCompileStatus();
    }
    
    private void requestCompileStatus() {
        if (!JavaDataObject.getJavaSettings().isCompileStatusEnabled()) {
            return;
        }
        synchronized (this) {
            if (iconRefreshTask != null) {
                iconResolver.setDetectCompile();
                return;
            }
            iconResolver = new IconResolver(true);
            iconRefreshTask = RequestProcessor.postRequest(iconResolver,
                ICON_REFRESH_DELAY_MSECS);
        }
    }
    
    private class IconResolver implements Runnable {
        boolean     detectCompile;
        
        IconResolver(boolean compile) {
            detectCompile = compile;
        }
        
        public void setDetectCompile() {
            detectCompile = true;
        }
        
        public void run() {
            synchronized (JavaNode.this) {
                iconRefreshTask = null;
                iconResolver = null;
            }
            if (!detectCompile) {
                StateUpdater.getInstance().updateNode(JavaNode.this);
            } else {
                checkUpToDate();
                StateUpdater.getInstance().updateNode(JavaNode.this);
            }
        }
    }
    
    /**
     * Schedules a refresh of node's icon, after ICON_REFRESH_DELAY_MSEC.
     * Requests is not scheduled if another one is still pending.
     */
    protected void requestResolveIcons() {
        synchronized (this) {
            if (iconRefreshTask != null)
                return;
            iconResolver = new IconResolver(false);
            iconRefreshTask = RequestProcessor.postRequest(iconResolver, 
                ICON_REFRESH_DELAY_MSECS);
        }
    }

    void dataPropertyChange(PropertyChangeEvent ev) {
        String propName = ev.getPropertyName();
        if (DataObject.PROP_FILES.equals(propName)) {
            attachToFiles();
            fileSetChanged(true);
        } else if (DataObject.PROP_MODIFIED.equals(propName)) {
            fileSetChanged(false);
        }
    }

    private void readObject(ObjectInputStream is) throws IOException, ClassNotFoundException {
        is.defaultReadObject();
        initialize();
    }

    /** Create the property sheet.
    * Subclasses may want to override this and add additional properties.
    * @return the sheet
    */
    protected Sheet createSheet () {
        Sheet sheet = super.createSheet();

        Sheet.Set ps = sheet.get(Sheet.PROPERTIES);
        ps.put(new PropertySupport.ReadWrite (
                   PROP_SYNCHRONIZATION_MODE,
                   Byte.class,
                   Util.getString("PROP_synchMode"),
                   Util.getString("HINT_synchMode")
               ) {
                   public Object getValue() {
                       return new Byte(getJavaDataObject().getSynchronizationType());
                   }
                   public void setValue (Object val) throws InvocationTargetException {
                       if (val instanceof Byte) {
                           try {
                               getJavaDataObject().setSynchronizationType(((Byte) val).byteValue());
                               return;
                           }
                           catch(IllegalArgumentException e) {
                           }
                       }
                       throw new IllegalArgumentException();
                   }
                   public PropertyEditor getPropertyEditor() {
                       return new SyncModeEditor();
                   }
               });

        ps = new Sheet.Set ();
        ps.setName(EXECUTION_SET_NAME);
        ps.setDisplayName(Util.getString("PROP_executionSetName"));
        ps.setShortDescription(Util.getString("HINT_executionSetName"));

        ExecSupport es = (ExecSupport) getCookie (ExecSupport.class);
        if (es != null)
            es.addProperties (ps);
        CompilerSupport cs = (CompilerSupport) getCookie (CompilerSupport.class);
        if (cs != null)
            cs.addProperties (ps);

        sheet.put(ps);

        return sheet;
    }

    /** Get the associated Java data object.
    * ({@link #getDataObject} is protected; this provides access to it.)
    * @return the data object
    */
    protected JavaDataObject getJavaDataObject() {
        return (JavaDataObject) getDataObject();
    }
    
    public Image getIcon(int type) {
        preparseSource();
      
        // assume setIconBase was already called.
        String[] badges = getBadges();
        return iconCache.getIcon(super.getIcon(type), badges);
    }
    
    public Image getOpenedIcon(int type) {
        preparseSource();
        String[] badges = getBadges();
        return iconCache.getIcon(super.getOpenedIcon(type), badges);
    }
    
    protected void preparseSource() {
        if (preparsed ||
            !JavaDataObject.getJavaSettings().getPrescanSources())
            return;
        preparsed = true;
        // schedule a background reparse so the icon for runnable classes
        // are updated in the background:
        // the listener will be notified after the parse is completed.
        SourceElement source = getJavaDataObject().getSource();
        if (getDataObject().isValid() && source.getStatus() == SourceElement.STATUS_NOT) {
            JavaParser p = findParser();
            if (p != null)
                p.parse(JavaParser.PRIORITY_BACKGROUND, false, false);
    	    // schedule a background reparse so the icon for runnable classes
    	    // are updated in the background:
        }
    }
    
    /** Get the icon base.
    * This should be a resource path, e.g. <code>/some/path/</code>,
    * where icons are held. Subclasses may override this.
    * @return the icon base
    * @see #getIcons
    * @deprecated the function is not consistent with openide's setIconBase() behaviour.
    * Given the icon badging, there's no need for composing icon names - see
    * {@link #setBadges()}, {@link #getBadges} or {@link #getBareIconBase()}
    */
    protected String getIconBase() {
        return ICON_BASE;
    }
    
    /**
     * Returns the base resource name for the basic icon that identifies object type
     * to the user. Additional images may be superimposed over that icon using icon
     * badging specification. Override in subclasses to provide a different icon for
     * the node.
     * @return the icon base 
     */
    protected String getBareIconBase() {
        return getIconBase() + getIcons()[0];
    }

    /** Get the icons.
    * This should be a list of bare icon names (i.e. no extension or path) in the icon base.
    * It should contain five icons in order for:
    * <ul>
    * <li>a regular class
    * <li>a class with a main method
    * <li>a class with a parse error
    * <li>a JavaBean class
    * <li>a JavaBean class with a main method
    * </ul>
    * Subclasses may override this.
    * @return the icons
    * @see #getIconBase
    * @deprecated State icons are now handled using icon badging mechanism. If you need
    * to add or modify the badges attached by the java module, override {@link #getBadges}
    */
    protected String[] getIcons() {
        return new String[] { "class" };
    }

    /**
     * Returns badges that should be combined with the base icon.
     * @return array of badges, can be null if no badges should be used.
     */
    protected String[] getBadges() {
        return currentBadges;
    }
    
    /**
     * Replaces the set of badges applied to the base icon. If the new set differs
     * from the current one, the new value is recorded and icon property changes are
     * fired.
     * @param list of badge identifiers; null to clear.
     */
    protected void setBadges(String[] badges) {
        if (currentBadges == badges)
            return;
        if (currentBadges != null && badges != null &&
            currentBadges.length == badges.length) {
            boolean match = true;
            
            for (int i = 0; i < currentBadges.length; i++) {
                if (!currentBadges[i].equals(badges[i])) {
                    match = false;
                    break;
                }
            }
            if (match) {
                return;
            }
        }
        currentBadges = badges;
        fireIconChange();
        fireOpenedIconChange();
    }
    
    /** Update the icon for this node based on parse status.
    * Called automatically at the proper times.
    * @see #getIconBase
    * @see #getBadges
    * @see #setBadges
    */
    protected void resolveIcons() {
        SourceElement source = getJavaDataObject().getSource();
        String desc = null;
        boolean isTemplate = getDataObject().isTemplate();
        java.util.Collection badges = new java.util.ArrayList(3);
        int status;
        
        if (isTemplate)
            status = SourceElement.STATUS_OK;
        else
            status = translateSourceStatus(source.getStatus());
        
        switch (status) {
        case SourceElement.STATUS_NOT :
            break;

        case SourceElement.STATUS_ERROR:
        case SourceElement.STATUS_PARTIAL:
            desc = Util.getString("HINT_ParsingErrors");
            badges.add(ICON_ATTR_ERROR);
            break;

        case SourceElement.STATUS_OK:
            boolean runnable = false;
            //boolean javaBean = false;

            // NOTE: this code depends on that is called when the parsed data is
            // alive.
            if (source.getStatus() == SourceElement.STATUS_OK) {
                FileObject fo = getJavaDataObject().getPrimaryEntry().getFile();
                String name = fo.getName();
                fo = fo.getParent();
                String pack = (fo == null) ? "" : fo.getPackageName('.'); // NOI18N

                // check the package
                Identifier id = source.getPackage();
                String pack2 = (id == null) ? "" : id.getFullName(); // NOI18N
                if (!pack.equals(pack2)) {
                    desc = new MessageFormat(Util.getString("FMT_Bad_Package")).format(new Object[] { pack2 });
                    //        newIcon = ICON_CLASS_ERROR;
                    break;
                }

                // check the runnable and javabean
                ClassElement[] classes = source.getClasses();
                // A.N. - performance bottleneck during opening of editor window (first time)
                // better solution is to delay the operation...
                for (int i = 0; i < classes.length; i++) {
                  if (classes[i].getName().getName().equals(name)) {
                    runnable = classes[i].hasMainMethod();
                    //javaBean = classes[i].isDeclaredAsJavaBean();
                    break;
                  }
                }
            }

            if (runnable) {
                badges.add(ICON_ATTR_MAIN);
            }
            break;
        }
        lastGoodStatus = status;
        
        // next: check whether the dataobject is up-to-date (if the settings permit that).
        // don't check it if the node corresponds to a template.
        if (!isTemplate) {
            String compBadge = resolveCompileBadge();
            if (compBadge != null)
                badges.add(compBadge);
        }
        
        setShortDescription(desc);
        if (badges.isEmpty()) {
            setBadges(ICON_ATTRS_NONE);
        } else {
            setBadges((String[])badges.toArray(new String[badges.size()]));
        }
    }
    
    protected String resolveCompileBadge() {
        if (!JavaDataObject.getJavaSettings().isCompileStatusEnabled()) {
            return null;
        }
        if (upToDate == 0)
            return ICON_ATTR_NEEDS_COMPILE;
        else if (upToDate != 1)
            requestCompileStatus();
        return null;
    }
    
    private void checkUpToDate() {
        CompilerCookie.Compile c = (CompilerCookie.Compile)getCookie(CompilerCookie.Compile.class);
        if (c == null) {
            this.upToDate = 1;
            return;
        }

	try {            
    	    org.openide.compiler.CompilerJob j = new org.openide.compiler.CompilerJob(
    		org.openide.compiler.Compiler.DEPTH_ONE);
    	    c.addToJob(j, org.openide.compiler.Compiler.DEPTH_ONE);
    	    if (j.isUpToDate()) {
                this.upToDate = 1;
        	return;
    	    }
	} catch (RuntimeException ex) {
	    ErrorManager errMan = TopManager.getDefault().getErrorManager();
	    errMan.notify(ErrorManager.INFORMATIONAL, ex);
            this.upToDate = 1;
	    return;
	}
        this.upToDate = 0;
    }

    public void createPasteTypes(java.awt.datatransfer.Transferable t, java.util.List s) {
        if (getJavaDataObject().isRenameAllowed()) {
            for (int i = 0; i <= 1; i++) {
                final boolean delete = (i == 1);
                final Element addingElement = (Element) NodeTransfer.cookie(t,
                                              delete ? NodeTransfer.MOVE : NodeTransfer.COPY, Element.class);

                if (addingElement != null && (addingElement instanceof ClassElement)) {
                    s.add(new org.openide.util.datatransfer.PasteType() {
                              public java.awt.datatransfer.Transferable paste() throws IOException {
                                  pasteElement((ClassElement)addingElement, delete);
                                  return delete ? org.openide.util.datatransfer.ExTransferable.EMPTY : null;
                              }
                          });
                }
            }
        }
        super.createPasteTypes(t, s);
    }

    private void pasteElement(final ClassElement el, final boolean delete) throws IOException {
	final SourceException ex[] = { null };
	final SourceElement src = getJavaDataObject().getSource();
	Runnable r = new Runnable() {
	    public void run() {
		try {
        	    src.addClass(el);

		    ClassElement target = src.getClass(Identifier.create(el.getName().getSourceName()));
		    int mods = target.getModifiers() & ~Modifier.STATIC;
		    if ((mods & (Modifier.PROTECTED | Modifier.PRIVATE)) > 0) {
			mods = (mods & ~(Modifier.PROTECTED | Modifier.PRIVATE)) | Modifier.PUBLIC;
		    }
		    target.setModifiers(mods);
        	} catch (SourceException e) {
		    ex[0] = e;
		}
	    }
	};
	src.prepare().waitFinished();
	try {
	    src.runAtomicAsUser(r);
	} catch (SourceException e) {
	    ex[0] = e;
	}
	if (ex[0] == null) {
    	    final SourceElement src2 = el.getSource();
            r = new Runnable() {
                public void run() {
                    try {
                        if (delete) {
                            if (el.getDeclaringClass() != null) {
                                el.getDeclaringClass().removeClass(el);
                            } else {
                                src2.removeClass(el);
                            }
                        }
                    } catch (SourceException e) {
                        ex[0] = e;
                    }
                }
            };
            try {
		if (src2 != null) {
                    src2.runAtomicAsUser(r);
		} else {
		    r.run();
		}
            } catch (SourceException e) {
                ex[0] = e;
            }
        }
	if (ex[0] != null) {
	    IOException newex = new IOException(ex[0].getMessage());
	    org.openide.TopManager.getDefault().getErrorManager().copyAnnotation(
		newex, ex[0]);
	    throw newex;
        }
    }

    public NewType[] getNewTypes() {
        if (getJavaDataObject().isJavaFileReadOnly()) {
            return super.getNewTypes();
        }
        return new NewType[] {
            new NewClassType(true),
            new NewClassType(false)
	};
    }
    
    /** Encapsulation of creating new class or interface.
    */
    class NewClassType extends NewType {
        private boolean isClass;

        NewClassType(boolean isClass) {
            this.isClass = isClass;
        }
	
	public String getName() {
	    return NEW_TYPE_NAMES[isClass ? 0 : 1];
	}

        public org.openide.util.HelpCtx getHelpCtx() {
            return null;
        }

        public void create() throws IOException {
            String simpleName = isClass ? "NewClass" : "NewInterface";
	    String outerName = getDataObject().getPrimaryFile().getPackageName('.');
            Identifier newIdent = Identifier.create(outerName + "." + simpleName,
                simpleName);

            final ClassElement e = new ClassElement();
            final SourceElement el = ((JavaDataObject)getDataObject()).getSource();
            Task parsingTask = el.prepare();
            try {
                e.setName(newIdent);
                e.setClassOrInterface(isClass);
                if (openCustomizer(new ClassCustomizer(e), NEW_TYPE_TITLES[isClass ? 0 : 1])) {
                    if (parsingTask.isFinished()) {
                        addClass(el, e);
                    } else {
                        parsingTask.addTaskListener(new org.openide.util.TaskListener() {
                            public void taskFinished(Task t) {
                                t.removeTaskListener(this);
                                try {
                                    addClass(el, e);
                                } catch (IOException exc) {
                                    TopManager.getDefault().notifyException(exc);
                                }
                            }
                        });
                    }
                }
            } catch (SourceException ex) {
		IOException newex = new IOException(ex.getLocalizedMessage());
		org.openide.TopManager.getDefault().getErrorManager().copyAnnotation(
		    newex, ex
		);
		throw newex;
            }
        }

        private void addClass(final SourceElement src, final ClassElement clazz) throws IOException {
            final SourceException[] ex  = new SourceException[] { null };
            try {
                src.runAtomicAsUser(new Runnable() {
                    public void run() {
                        try {
                            src.addClass(clazz);
                        } catch (SourceException e) {
                            ex[0] = e;
                        }
                    }
                });
            } catch (SourceException e) {
            }
            if (ex[0] != null) {
		IOException newex = new IOException(ex[0].getLocalizedMessage());
		org.openide.TopManager.getDefault().getErrorManager().copyAnnotation(
		    newex, ex[0]
		);
		throw newex;
            }
        }
    }

    static boolean openCustomizer(java.awt.Component customizer, String title) {
        NotifyDescriptor desriptor = new NotifyDescriptor(
                                         customizer,
                                         title,
                                         NotifyDescriptor.OK_CANCEL_OPTION,
                                         NotifyDescriptor.PLAIN_MESSAGE,
                                         null, null);

        Object ret = TopManager.getDefault().notify(desriptor);
        return (ret == NotifyDescriptor.OK_OPTION);
    }
    
    /** Simple property editor for two-item pulldown. */
    static class SyncModeEditor extends PropertyEditorSupport {

        private static final String[] tags;

        static {
            tags = new String[3];
            tags[JavaDataObject.CONNECT_NOT] = Util.getString("CTL_CONNECT_NOT");
            tags[JavaDataObject.CONNECT_CONFIRM] = Util.getString("CTL_CONNECT_CONFIRM");
            tags[JavaDataObject.CONNECT_AUTO] = Util.getString("CTL_CONNECT_AUTO");
        }

        public String[] getTags () {
            return tags;
        }

        public String getAsText () {
            return tags[((Byte) getValue ()).byteValue ()];
        }

        public void setAsText (String text) throws IllegalArgumentException {
            for (byte i = 0; i < tags.length; i++) {
                if (tags[i].equals (text)) {
                    setValue (new Byte(i));
                    return;
                }
            }
            throw new IllegalArgumentException ();
        }
    }
    
    private static class StateUpdater implements PropertyChangeListener,
        ActionListener {
        private Set     registeredNodes;
        private static StateUpdater     instance;
        private Collection scheduledNodes;
        private Timer   timer;

        StateUpdater() {
            JavaDataObject.getJavaSettings().addPropertyChangeListener(this);
            scheduledNodes = new LinkedList();
            timer = new Timer(300, this);
            timer.setRepeats(false);
        }
        
        static synchronized StateUpdater getInstance() {
            if (instance == null)
                instance = new StateUpdater();
            return instance;
        }
        
        public void registerNode(Node n) {
            if (registeredNodes == null) {
                synchronized (this) {
                    if (registeredNodes == null)
                        registeredNodes = new org.openide.util.WeakSet(37);
                }
            }
            synchronized (registeredNodes) {
                registeredNodes.add(n);
            }
        }
        
        public void updateNode(JavaNode n) {
            synchronized (this) {
                scheduledNodes.add(new java.lang.ref.WeakReference(n));
                if (scheduledNodes.size() == 1) {
                    timer.start();
                }
            }
            timer.restart();
        }
        
        public void propertyChange(java.beans.PropertyChangeEvent evt) {
            if (!JavaSettings.PROP_SHOW_COMPILE_STATUS.equals(evt.getPropertyName()))
                // not interested.
                return;
            if (registeredNodes == null)
                return;
            Collection v;
            synchronized (registeredNodes) {
                v = new ArrayList(registeredNodes);
            }
            boolean enableStatus = ((Boolean)evt.getNewValue()).booleanValue();
            for (Iterator it = v.iterator(); it.hasNext(); ) {
                ((JavaNode)it.next()).updateCompiledStatus(enableStatus);
            }
        }
        
        public void actionPerformed(java.awt.event.ActionEvent actionEvent) {
            Collection n;
            
            synchronized (this) {
                n = scheduledNodes;
                scheduledNodes = new LinkedList();
            }
            for (Iterator it = n.iterator(); it.hasNext(); ) {
                JavaNode node = (JavaNode)((java.lang.ref.Reference)it.next()).get();
                if (node == null)
                    continue;
                node.resolveIcons();
            }
        }
    }
}
