/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.io.*;
import java.util.*;
import java.beans.PropertyChangeListener;
import javax.swing.text.StyledDocument;
import javax.swing.text.Position;

import sun.tools.java.*;
import sun.tools.javac.*;
import sun.tools.tree.*;

import org.openide.TopManager;
import org.openide.src.*;
import org.openide.text.NbDocument;
import org.openide.text.PositionRef;
import org.openide.text.PositionBounds;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.util.io.NullOutputStream;

/** Parser of Java source code. It generates the hierarchy
* of the implementations of elements.
*
* @author Petr Hamernik
* @deprecated Uses old javac's interface.
*/
class JavaParser extends BatchParser {

    // ===================== Fields ==============================

    /** Mask which contains all possible modifiers used in the hierarchy.
    * It filters internal modifiers used in parser.
    */
    private static final int MODIF_MASK = M_PUBLIC | M_PRIVATE | M_PROTECTED |
                                          M_STATIC | M_FINAL | M_SYNCHRONIZED | M_VOLATILE | M_TRANSIENT |
                                          M_NATIVE | M_ABSTRACT;

    /** The environment for recognizing the errors after parsing. */
    BatchEnvironment environment;

    /** Input stream */
    Util.ParserInputStream input;

    /** Counter of errors which occurs during resolving fully qualified
    * names in parser. If all error occur only during resolving,
    * the parser looks like that trere is no error.
    */
    int errorsInResolving;

    /** Reference to the runnable which has started the parser.
    * It is used in SourceElementImpl.setParsingResult after parser finish.
      @deprecated SourceElementImpl now uses V8-style parser with different binding.
    SourceElementImpl.ParsingRunnable parsingRunnable;
    */

    // ===================== Main methods and constructors ==================

    /** Constructor directly for parsing without compiler.
    * @param fo FileObject where the Java code is stored.
    * @exception IOException if any i/o problem occured during reading
    */
    public JavaParser(FileObject fo) throws IOException {
        this(JavaCompiler.createEnvironment(null), fo);
    }

    /** Constructor directly for parsing without compiler.
    * @param fo FileObject where the Java code is stored.
    * @env environment created by Compiler, which is set to 'environment' field
    * @exception IOException if any i/o problem occured during reading
    */
    private JavaParser(Environment env, FileObject fo) throws IOException {
        this(new Environment(env, new CoronaClassFile(fo)), fo, true);
        if (env instanceof BatchEnvironment) {
            // errors checking - for partially parsed files
            this.environment = (BatchEnvironment) env;
        }
    }

    /** Constructor for compiler.
    * @param environment The environment passed by the compiler or
    * @param fo FileObject where the Java code is stored.
    * @param store if there is required the building and storing the elements
    *              hierarchy
    * @exception IOException if any i/o problem occured during reading
    */
    public JavaParser(Environment environment, FileObject fo, boolean store) throws IOException {
        this(environment, fo, store, createInputStream(fo, store, environment.getCharacterEncoding()));
    }

    /** Private constructor.
    * @param environment The environment passed by the compiler or
    * @param fo FileObject where the Java code is stored.
    * @param store if there is required the building and storing the elements
    *              hierarchy
    * @param input the input stream where to read.
    * @exception IOException if any i/o problem occured during reading
    */
    private JavaParser(Environment environment, FileObject fo, boolean store, Util.ParserInputStream input) throws IOException {
        super(environment, input);

        this.input = input;

        if (!store)
            return;

        if (environment instanceof BatchEnvironment) {
            // errors checking - for partially parsed files
            this.environment = (BatchEnvironment) environment;
        }
        errorsInResolving = 0;
    }

    /** Creates new input stream from the file object.
    * Finds the java data object, checks if the document is loaded and
    * and create the stream either from the file object either from the document.
    * @param fo fileobject with the source
    * @param store if there is required the building and storing the elements
    *              hierarchy
    * @exception IOException if any i/o problem occured during reading
    */
    private static Util.ParserInputStream createInputStream(FileObject fo, boolean store, String encoding) throws IOException {
        return (Util.ParserInputStream)Util.createInputStream(fo, false, true, encoding);
    }

    /** Parses the file - this method starts the parser.
    * After super.parseFile finish, the result is set to
    * the appropriate SourceElementImpl.
    */
    public void parseFile() {
        super.parseFile();
        closeInput();
    }

    /** This method closes the input stream for parsing, which is always opened
    * in the constructor. Method parseFile calls this method, but if anyone just creates
    * the JavaParser class, is also responsible for closing the stream.
    * Otherwise the stream is closed during finalization.
    */
    public void closeInput() {
        try {
            if (input != null) {
                input.close();
                input = null;
            }
        }
        catch (IOException e) {
        }
    }

    /** Finalize the parser and close the input stream if necessary.
    */
    public void finalize() throws Throwable {
        super.finalize();
        closeInput();
    }

    /** Method for the environment - accessibility */
    protected Vector getClassesProtected() {
        return classes;
    }

    /** Method for the environment - accessibility */
    protected Imports getImportsProtected() {
        return imports;
    }

    // ======================== PARSER METHODS ===============================

    // ----------------------- utilities -----------------------------------

    /** Computes the real offset from the long value representing position
    * in the parser.
    * @return the offset
    */
    static int position(long p) {
        return (int)(p & 0xFFFFFFFFL);
    }

    // ==================== The parser input stream ==========================

    /** The input stream which holds all data which are read in the StringBuffer.
    */
    static class ParserInputStream extends InputStream {
        /** The underlaying stream. */
        private InputStream stream;

        /** Whole text */
        private String text;

        /** The string buffer which collect the data. */
        private StringBuffer buffer;

        /** This flag determines if there is used the text field or buffer field.
        * The constructor set it
        */
        private boolean mode;

        /** The counter of read chars */
        private int counter;

        /** Creates the stream from the text. */
        ParserInputStream(String text) {
            stream = new StringBufferInputStream(text);
            this.text = text;
            mode = false;
            counter = 0;
        }

        /** Creates the stream from the another stream. */
        ParserInputStream(InputStream stream) {
            this.stream = stream;
            buffer = new StringBuffer();
            mode = true;
        }

        /** Gets the part of the text which was already read.
        * @param begin the begin index
        * @param end the end index
        */
        public String getString(long begin, long end) {
            int p1 = position(begin);
            int p2 = position(end);
            return mode ? buffer.substring(p1, p2) : text.substring(p1, p2);
        }

        /** Gets the part of the text which was already read.
        * End is last position which was already read.
        * @param begin the begin index
        */
        public String getString(long begin) {
            int p1 = position(begin);

            if (mode) {
                return buffer.substring(p1);
            }
            else {
                int p2 = Math.min(counter - 1, text.length());
                return text.substring(p1, p2);
            }
        }

        /** Read one character from the stream. */
        public int read() throws IOException {
            int x = stream.read();
            if (mode && (x != -1))
                buffer.append((char)x);
            counter++;
            return x;
        }

        /** Closes the stream */
        public void close() throws IOException {
            stream.close();
        }
    }
}
