/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.beans.*;
import java.io.*;
import java.util.*;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.Modifier;

import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Segment;
import javax.swing.text.StyledDocument;

import org.openide.cookies.*;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.nodes.CookieSet;
import org.openide.src.*;
import org.openide.text.*;
import org.openide.util.Task;
import org.openide.util.WeakSet;
import org.openide.util.RequestProcessor;

import org.netbeans.modules.java.parser.LangEnvImpl;
import org.netbeans.modules.java.parser.JavaParser;
import org.netbeans.modules.java.codegen.DocumentBinding;
import org.netbeans.modules.java.codegen.TextBinding;

import org.netbeans.modules.java.model.BindingFactory;
import org.netbeans.modules.java.model.WrapperFactory;
import org.netbeans.modules.java.model.LangModel;
import org.netbeans.modules.java.model.ElementOrder;
import org.netbeans.modules.java.model.ElementImpl;

import org.netbeans.modules.java.parser.ParsingSupport;
import org.netbeans.modules.java.codegen.SourceText;
import org.netbeans.modules.java.model.DefaultLangModel;

import org.netbeans.modules.java.codesync.SynchronizeCodeCookie;


/**
 * A glue between a JavaDataObject and JavaParser
 * 
 * @author  sdedic
 * @version 0.1
 */
class JavaParserGlue implements JavaParser.Env, DocumentBinding.Env, 
    LangModel.Env, SourceCookie.Editor {

    private MultiDataObject.Entry   sourceEntry;
    private JavaDataObject          jdo;
    private DocumentBinding         docBinding;
    private LangEnvImpl             envSupport;
    private ParsingSupport          parser;
    private Map                     cookieMap;
    private SiblingListener         l;
    private boolean                 documentLoaded;
    private CloneableEditorSupport  cloneableEditSupp;
    private int                     suspended;
    private PropertyChangeListener  weakPropListener;
    private boolean 		    disableCookieEvents;
    
    /**
     * Constructs an implementation of a parsing environment and binds it to the
     * data object.
     */
    public JavaParserGlue(MultiDataObject.Entry en) {
        this.sourceEntry = en;
        this.jdo = (JavaDataObject)en.getDataObject();
        this.docBinding = new SourceText(this);
        this.envSupport = new LangEnvImpl(docBinding);
        this.cookieMap = new WeakHashMap(57);
        
        DefaultLangModel model = new DefaultLangModel(this);
        envSupport.setModel(model);
        
        ParsingSupport supp = new ParsingSupport(this, docBinding, model, model);
        supp.setParserEngine(((JavaDataLoader)jdo.getLoader()).getParserEngine());
        parser = supp;
        l = new SiblingListener();
        parser.addChangeListener(l);
        parser.getSource().addPropertyChangeListener(l);
        jdo.addPropertyChangeListener(l);
        // 
        if (parser.getStatus() != SourceElement.STATUS_NOT)
            l.rescanSource(parser.getSource());
    }
    
    /**
     * The implementation provides custom OpenCookies that open the JavaEditor
     * associated with the JavaDataObject at the where the Element starts.
     */
    public Node.Cookie findCookie(Element el, Class clazz) {
        Node.Cookie lookup = null;

        if (el == parser.getSource())
            // handle it specially for SourceElement - it has much more common with DataObject than
            // the rest of Elements.
            return findCookie((SourceElement)el, clazz);
        if (SynchronizeCodeCookie.class == clazz) {
            // disable, if r/o
            if (jdo.getPrimaryFile().isReadOnly())
                return null;
            
            lookup = lookupCookie(el, clazz);
            if (lookup != null)
                return lookup;
            
            if (el instanceof ClassElement) {
                lookup = createClassSyncCookie((ClassElement)el);
            } else if (el == parser.getSource()) {
                return jdo.connectionManager;
            }
        } else if (OpenCookie.class.isAssignableFrom(clazz)) {
            lookup = lookupCookie(el, clazz);
            if (lookup != null)
                return lookup;
            lookup = createOpenCookie(el);
        } else if (!clazz.isAssignableFrom(ElementCookie.class)) {
            return jdo.getCookie(clazz);
        }
        if (lookup != null)
            registerElementCookie(el, lookup);
        return lookup;
    }
    
    public void annotateThrowable(Throwable t, String locMessage, boolean user) {
        Util.annotateThrowable(t, locMessage, user);
    }
    
    public void annotateThrowable(Throwable wrapper, Throwable nested) {
        Util.annotateThrowable(wrapper, nested);
    }
    
    public Node.Cookie findCookie(SourceElement src, Class clazz) {
        return jdo.getCookie(clazz);
    }
    
    /**
     * Notifies the mediator that a CloneablEditorSupport has been created.
     */
    public synchronized void cloneableSupportCreated(CloneableEditorSupport supp) {
        cloneableEditSupp = supp;
    }
    
    public JavaParser getParser() {
        return parser;
    }
    
    /**
     * Environment can find precompiled classes to speed up name resoltion
     */
    public InputStream findCompiledClass(String classFQN) {
        // PENDING: migrate code from the old V8ParseRequest
        return null;
    }
    
    /** 
     * The implementation first tries to get a document - if it is already opened.
     * If it is, then it returns a CharArrayReader constructed on the document's
     * contents. Document is read using {@link StyledDocument#render} to avoid
     * data corruption.<P>
     * If the document is not (yet) opened, a {@link JavaEditor#GuardedReader} is
     * constructed so the output will not contain any guarded block information.
     * <P>
     * The implementation does not handle IOException from lower layers in any
     * way and passes them to the caller.
     */
    public Reader getSourceText() throws IOException {
        CloneableEditorSupport je = findEditorSupport();
        final StyledDocument doc = je.getDocument();
        
        if (doc != null) {
            // can read from the document!
            final Segment s = new javax.swing.text.Segment();
            doc.render(new Runnable() {
                public void run() {
                    try {
                        doc.getText(0, doc.getLength(), s);
                    } catch (BadLocationException ex) {
                        // should never happen
                    }
                }
            });
            return new CharArrayReader(s.array);
        } else {
            JavaDataLoader.JavaFileEntry en = (JavaDataLoader.JavaFileEntry)sourceEntry;
            return new JavaEditor.GuardedReader(
                en.getInputStream(), true);
        }
    }
    
    protected Node.Cookie createClassSyncCookie(Element el) {
        return jdo.connectionManager.createClassSyncCookie((ClassElement)el);
    }
    
    /**
     * Creates an OpenCookie for the specified element. The cookies created are
     * kept in caching WeakHashMap keyed by the element instances for reuse.
     */
    protected OpenCookie createOpenCookie(Element el) {
        ElementImpl impl = (ElementImpl)el.getCookie(ElementImpl.class);
        OpenCookie ck = new OpenCookieImpl((TextBinding)impl.getBinding());
        return ck;
    }
    
    private Node.Cookie lookupCookie(Element el, Class clazz) {
        synchronized (cookieMap) {
            Object o = cookieMap.get(el);
            if (o == null)
                return null;
            
            if (o instanceof CookieSet) 
                return ((CookieSet)o).getCookie(clazz);
            
            if (o.getClass().isAssignableFrom(clazz))
                return (Node.Cookie)o;
            return null;
        }
    }
    
    private void registerElementCookie(Element el, Node.Cookie cookie) {
        synchronized (cookieMap) {
            Object o = cookieMap.get(el);
            if (o == null) {
                cookieMap.put(el, cookie);
                return;
            } else if (o instanceof CookieSet) {
                ((CookieSet)o).add(cookie);
            } else {
                Node.Cookie old = (Node.Cookie)o;
                CookieSet set = new CookieSet();
                set.add(old);
                set.add(cookie);
                cookieMap.put(el, set);
            }
        }
    }
    
    private TextBinding getBinding(Element el) {
        ElementImpl impl = (ElementImpl)el.getCookie(ElementImpl.class);
        if (impl == null) {
            Element.Impl iimpl = (Element.Impl)el.getCookie(Element.Impl.class);
            throw new IllegalArgumentException("Incompatible implementation: " +
                iimpl);
        }
        return (TextBinding)impl.getBinding();
    }


    // ============== Implementation of DocumentBinding.Env ==================
    
    public CloneableEditorSupport findEditorSupport() {
        if (this.cloneableEditSupp == null) 
            this.cloneableEditSupp = jdo.findCloneableEditorSupport();
        
        // attach change listeners to capute document loading/unloading
        cloneableEditSupp.addChangeListener(l);
        documentLoaded = cloneableEditSupp.isDocumentLoaded();
        return cloneableEditSupp;
    }
    
    public PositionRef findFreePosition(PositionBounds bounds) {
        return getJavaEditor().findFreePosition(bounds);
    }
    
    private JavaEditor getJavaEditor() {
        return (JavaEditor)jdo.getCookie(JavaEditor.class);
    }

    // ============== Implementation of LangModel.Env -- DELEGATING ==================
    /**
     * Returns the document binding as the binding factory.
     */
    public BindingFactory getBindingFactory() {
        return envSupport.getBindingFactory();
    }

    public WrapperFactory getWrapperFactory() {
        return envSupport.getWrapperFactory();
    }

    public void complete(Element scope, int informationKind) {
        envSupport.complete(scope, informationKind);
    }
    
    public Type resolveType(Element context, Type original) {
        return envSupport.resolveType(context, original);
    }
    
    public Identifier resolveTypeIdent(Element context, Identifier original) {
        return envSupport.resolveTypeIdent(context, original);
    }
    
    /**
     * Implementation of SourceCookie.
     */
    public SourceElement getSource() {
        return parser.getSource();
    }
    
    /**
     * Implementation of OpenCookie available on individual elements.
     */
    private class OpenCookieImpl implements OpenCookie {
        TextBinding binding;
        
        OpenCookieImpl(TextBinding binding) {
            this.binding = binding;
        }
        
        public void open() {
            PositionBounds elBounds = binding.getElementRange(true);
            getJavaEditor().openAt(elBounds.getBegin());
        }
    }
    
    /**
     * Backfires PROP_COOKIE property change event in the source hierarchy.
     */
    protected void refireCookieChange(PropertyChangeEvent evt) {
        // fire the event on SourceElement.
        parser.fireElementPropertyChange(getSource(), evt);

        SourceElement.Impl impl = parser.getSourceImpl();
        // do not fire on lesser elements - there aren't any.
        if (impl == null)
            return;
        ClassElement[] allClasses = impl.getAllClasses();
        Collection allElements = new ArrayList(40);

        allElements.addAll(Arrays.asList(allClasses));
        for (int i = 0; i < allClasses.length; i++) {
            ClassElement c = allClasses[i];
            allElements.addAll(Arrays.asList(c.getFields()));
            allElements.addAll(Arrays.asList(c.getInitializers()));
            allElements.addAll(Arrays.asList(c.getConstructors()));
            allElements.addAll(Arrays.asList(c.getMethods()));
        }
        // FIRE!
	try {
	    disableCookieEvents = true;
    	    for (Iterator it = allElements.iterator(); it.hasNext(); ) {
    		parser.fireElementPropertyChange((Element)it.next(), evt);
    	    }
	} finally {
	    disableCookieEvents = false;
	}
    }
    
    public StyledDocument getDocument() {
        return findEditorSupport().getDocument();
    }
    
    public Line.Set getLineSet() {
        return findEditorSupport().getLineSet();
    }
    
    public javax.swing.JEditorPane[] getOpenedPanes() {
        return findEditorSupport().getOpenedPanes();
    }
    
    public boolean isModified() {
        return findEditorSupport().isModified();
    }
    
    public void open() {
        findEditorSupport().open();
    }
    
    public StyledDocument openDocument() throws IOException {
        return findEditorSupport().openDocument();
    }
    
    public Task prepareDocument() {
        return findEditorSupport().prepareDocument();
    }
    
    public void saveDocument() throws IOException {
        findEditorSupport().saveDocument();
    }
    
    public boolean close() {
        return findEditorSupport().close();        
    }
    
    public void suspendDocumentChanges() {
        suspended++;
    }
    
    public void resumeDocumentChanges() {
        --suspended;
    }
    
    protected boolean isDocumentSuspended() {
        return suspended > 0;
    }
    
    private void dissolve() {
        suspendDocumentChanges();
        jdo.removePropertyChangeListener(l);

        if (cloneableEditSupp != null) {
            cloneableEditSupp.removeChangeListener(l);
            StyledDocument d = cloneableEditSupp.getDocument();
            if (d != null)
                d.removeDocumentListener(l);
        }
    }
    
    private class SiblingListener implements PropertyChangeListener, ChangeListener,
        DocumentListener {
        
        private StyledDocument doc;
        private Reference   mainClass;
        private Set     knownClasses;
        
        SiblingListener() {
            knownClasses = new WeakSet();
        }
        
        public void propertyChange(PropertyChangeEvent p) {
            String evName = p.getPropertyName();
            Object source = p.getSource();
            
            if (source == jdo)
                dataObjectPropertyChange(evName, p);
            else if (source instanceof ClassElement) {
                classPropertyChange(evName, (ClassElement)source, p);
            } else if (source instanceof SourceElement) {
                sourcePropertyChange(evName, p);
            }
        }
        
        private void classPropertyChange(String name, ClassElement c,
        PropertyChangeEvent e) {
            if (ElementProperties.PROP_VALID.equals(name)) {
                c.removePropertyChangeListener(this);
            } else if (ElementProperties.PROP_NAME.equals(name)) {
                if (mainClass == null || mainClass.get() != c) {
                    // some other class is renamed.
                    rescanSource(c.getSource());
                } else {
                    // rename the dataobject, if possible.
                    String n = jdo.getName();
                    String n2 = (String)((Identifier)e.getNewValue()).getName(); 
                    if (!n.equals(n2)) {
                        try {
                            jdo.rename(n2);
                        } catch (IOException ex) {
                            // swallow.
                            rescanSource(c.getSource());
                        }
                    }
                }
            } else if (ElementProperties.PROP_MODIFIERS.equals(name)) {
                // public class may loose its modifiers -> some other class
                // may become the main one. non-public class may become public
                rescanSource(c.getSource());
            }
        }
        
        private void sourcePropertyChange(String name, PropertyChangeEvent e) {
            if (ElementProperties.PROP_CLASSES.equals(name)) {
                rescanSource((SourceElement)e.getSource());
            } else if (ElementProperties.PROP_STATUS.equals(name)) {
                Integer i = (Integer)e.getOldValue();
                if (i == null || i.intValue() != SourceElement.STATUS_NOT)
                    return;
                rescanSource((SourceElement)e.getSource());
            }
        }
        
        private void rescanSource(SourceElement el) {
            ClassElement[] classes = el.getClasses();
            String currentName = jdo.getName();
            ClassElement main = null;
            
            // sweep through all top-level classes.
            for (int i = 0; i < classes.length; i++) {
                ClassElement c = classes[i];
                String clName = c.getName().getName();
                if (knownClasses.add(c)) {
                    c.addPropertyChangeListener(this);
                }
                if (main == null && clName.equals(jdo.getName()) &&
                    (c.getModifiers() & Modifier.PUBLIC) > 0) {
                    main = c;
                }
            }
            mainClass = main == null ? null : new WeakReference(main);
        }

        private void dataObjectPropertyChange(String evName, PropertyChangeEvent p) {
            if (DataObject.PROP_COOKIE.equals(evName)) {
		if (disableCookieEvents)
		    return;
                refireCookieChange(p);
            } else if (DataObject.PROP_MODIFIED.equals(evName) && !jdo.isModified()) {
                // after save of the object, or reload.
                parser.parse(parser.PRIORITY_NORMAL, false, true);
            } else if (DataObject.PROP_VALID.equals(evName) 
                && ((Boolean)p.getNewValue()).booleanValue() == false) {
                dissolve();
            } else if (DataObject.PROP_NAME.equals(evName)) {
                SourceElement.Impl impl = parser.getSourceImpl();
                SourceElement el = parser.getSource();
                
                if (impl != null) {
                    rescanSource(el);
                }
            }
        }
        
        public void stateChanged(ChangeEvent e) {
            if (e.getSource() == parser) {
                parserStateChanged(e);
            } else {
                documentStateChanged(e);
            }
        }
        
        private void parserStateChanged(ChangeEvent e) {
            final Object hook = ((JavaParser)e.getSource()).getSourceImpl();
            
            RequestProcessor.postRequest(new Runnable() {
                public void run() {
                    Parsing.fireEvent(jdo, hook);
                }
            });
        }
                
            
        private void documentStateChanged(ChangeEvent p) {
            CloneableEditorSupport supp = (CloneableEditorSupport)p.getSource();
            StyledDocument d = supp.getDocument();
            
            if (d != null) {
                if (doc == null) {
                    parser.parse(parser.PRIORITY_NORMAL, true, true);
                    synchronized (this) {
                        doc = d;
                        d.addDocumentListener(this);
                    }
                }
            } else {
                removeDocListener();
                parser.sourceChanged(-1, -1);
                parser.parse(parser.PRIORITY_NORMAL, false, true);
            }
        }
        
        private synchronized void removeDocListener() {
            if (doc != null)
                doc.removeDocumentListener(this);
            doc = null;
        }
        
        public void removeUpdate(javax.swing.event.DocumentEvent p1) {
            if (!isDocumentSuspended())
                parser.sourceChanged(-1, -1);
        }
        
        public void changedUpdate(javax.swing.event.DocumentEvent p1) {
            if (!isDocumentSuspended())
                parser.sourceChanged(-1, -1);
        }
        
        public void insertUpdate(javax.swing.event.DocumentEvent p1) {
            if (!isDocumentSuspended())
                parser.sourceChanged(-1, -1);
        }
    }
    
    private Map     swingElementMap;
    
    // SourceCookie.Editor - specifics
    public org.openide.src.Element findElement(int offset) {
        javax.swing.text.Element swingEl = sourceToText(getSource());
        
        while (swingEl != null) {
            if (swingEl.isLeaf()) {
                return ((TextElement)swingEl).getSourceElement();
            }
            int elIndex = swingEl.getElementIndex(offset);
            if (elIndex == -1)
                return ((TextElement)swingEl).getSourceElement();
            swingEl = swingEl.getElement(elIndex);
        }
        return null;
    }
    
    public org.openide.src.Element textToSource(javax.swing.text.Element element) throws NoSuchElementException {
        if (!(element instanceof TextElement)) {
            throw new NoSuchElementException();
        }
        TextElement el = (TextElement)element;
        return el.getSourceElement();
    }
    
    public javax.swing.text.Element sourceToText(org.openide.src.Element element) {
        if (swingElementMap == null) {
            synchronized (this) {
                if (swingElementMap == null)
                    swingElementMap = new WeakHashMap(75);
            }
        }
        javax.swing.text.Element e = (javax.swing.text.Element)swingElementMap.get(element);
        if (e == null) {
            e = new TextElement(getBinding(element));
            synchronized (this) {
                swingElementMap.put(element, e);
            }
        }
        return e;
    }
    
    private static final Element[] NO_CHILDREN = new Element[0];
    
    private class TextElement implements javax.swing.text.Element {
        private TextBinding    myBinding;
        private Element        myElement;
        
        public TextElement(TextBinding binding) {
            myBinding = binding;
            myElement = binding.getElement();
        }
        
        private PositionBounds getBounds() {
            return myBinding.getElementRange(false);
        }
        
        private Element[] getChildrenElements() {
            ElementOrder ck = (ElementOrder)myElement.getCookie(
                ElementOrder.class);
            if (ck == null)
                return NO_CHILDREN;
            return ck.getElements();
        }
        
        public int getElementIndex(int offset) {
            Element[] children = getChildrenElements();
            javax.swing.text.Element childElement;
            
            for (int i = 0; i < children.length; i++) {
                childElement = sourceToText(children[i]);
                if (childElement.getStartOffset() <= offset &&
                    childElement.getEndOffset() >= offset) {
                    return i;
                }
            }
            return -1;
        }
        
        public javax.swing.text.AttributeSet getAttributes() {
            return null;
        }
        
        public Element getSourceElement() {
            return myElement;
        }
        
        public javax.swing.text.Document getDocument() {
            return JavaParserGlue.this.getDocument();
        }
        
        public javax.swing.text.Element getElement(int index) {
            Element[] els =  getChildrenElements();
            if (index > els.length)
                throw new IllegalArgumentException();
            return sourceToText(els[index]);
        }
        
        public int getElementCount() {
            return getChildrenElements().length;
        }
        
        public int getEndOffset() {
            return getBounds().getEnd().getOffset() - 1;
        }
        
        public String getName() {
            return myElement.getClass().getName();
        }
        
        public javax.swing.text.Element getParentElement() {
            Element parent;

            if (myElement instanceof MemberElement) {
                parent = ((MemberElement)myElement).getDeclaringClass();
                if (parent == null && myElement instanceof ClassElement) {
                    parent = ((ClassElement)myElement).getSource();
                }
            } else if (myElement instanceof InitializerElement) {
                parent = ((InitializerElement)myElement).getDeclaringClass();
            } else
                return null;
            return sourceToText(parent);
        }
        
        public int getStartOffset() {
            return getBounds().getBegin().getOffset();
        }
        
        public boolean isLeaf() {
            return getChildrenElements().length == 0;
        }
    }
}
