/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java;

import java.util.*;
import java.text.MessageFormat;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;

import org.openide.cookies.SourceCookie;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.OperationAdapter;
import org.openide.loaders.OperationEvent;

/**
 * This class encapsulates Java module's specific behaviour for the packages.
 * Currently, the object watches out for package rename and if some package is
 * renamed, the implementation will try to locate and update Java sources
 * contained within the package.
 * @author <A HREF="mailto:svatopluk.dedic@netbeans.com>Svatopluk Dedic</A>, NetBeans.
 */
class PackageHandler extends OperationAdapter  {

    /** Creates new PackageHandler */
    PackageHandler() {
    }


    public void operationRename(OperationEvent.Rename ev) {
        DataObject obj = ev.getObject();
        DataFolder fld = (DataFolder)obj.getCookie(DataFolder.class);
        if (fld == null || fld != obj || obj.getName().equals(ev.getOriginalName())) {
            // not a folder -- not interesting.
            // if fld != obj then there's something that mimics the behaviour of a folder,
            // like a link - we will not update such objects.
            return;
        }
        if (!discoverJavaSources(fld))
            // no Java sources in the folder, don't bother the user with
            // dialogs.
            return;
        handlePackageRename(fld, ev.getOriginalName());
    }
    
    boolean discoverJavaSources(DataFolder f) {
        List c = new LinkedList();
        c.add(f.getPrimaryFile());
        return discoverJavaSources(c);
    }
    
    boolean discoverJavaSources(List folders) {
        FileObject f;
        
        while (!folders.isEmpty()) {
            FileObject fld = (FileObject)folders.remove(0);
            FileObject []children = fld.getChildren();
            for (int i = 0; i < children.length; i++) {
                f = children[i];
                if (!f.isFolder()) {
                    if (JavaDataLoader.JAVA_EXTENSION.equals(f.getExt())) {
                        // something that pretty much resembles Java is here...
                        try {
                            DataObject target = DataObject.find(f);
                            if (target.getCookie(SourceCookie.Editor.class) != null)
                                return true;
                        } catch (DataObjectNotFoundException ex) {
                        }
                    }
                } else
                    folders.add(f);
            }
        }
        return false;
    }

    protected void handlePackageRename(final DataFolder parent, final String originalName) {
        org.openide.filesystems.FileObject parent2 = parent.getPrimaryFile().getParent();
        String parentName;
        
        if (parent2.isRoot()) {
            parentName = "";
        } else {
            parentName = parent2.getPackageName('.') + ".";
        }
        String message = MessageFormat.format(JavaNode.bundle.getString("FMT_RenameConfirmation"),
        new Object[] {
            parentName + originalName,
            parentName + parent.getName()
        });
        final NotifyDescriptor desc = new NotifyDescriptor.Confirmation(message,
        NotifyDescriptor.YES_NO_OPTION, NotifyDescriptor.INFORMATION_MESSAGE
        );
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                if (TopManager.getDefault().notify(desc) != NotifyDescriptor.YES_OPTION)
                    return;
                repackage(parent, originalName);
            }
        });
    }
    
    /**
     * The method that performs the real work -- repackages Java sources beneath the
     * specified folder changing package declarations from `oldName' to the current one.
     */
    private void repackage(DataFolder parent, String oldName) {
        DataObject[] children = parent.getChildren();
        java.util.List subfolders = null;
        for (int i = 0; i < children.length; i++) {
            DataFolder subfolder = (DataFolder)children[i].getCookie(DataFolder.class);
            if (subfolder != null) {
                if (subfolders == null)
                    subfolders = new java.util.LinkedList();
                subfolders.add(subfolder);
                continue;
            }
            JavaDataObject jdo = (JavaDataObject)children[i].getCookie(
            JavaDataObject.class
            );
            if (jdo == null)
                continue;
            // update the source to match the package / filename.
            JavaDataObject.updateSource(jdo, jdo.getPrimaryFile(),
            jdo.getPrimaryFile().getName(), true);
        }
        
        if (subfolders != null) {
            for (java.util.Iterator it = subfolders.iterator(); it.hasNext(); ) {
                repackage((DataFolder)it.next(), oldName);
            }
        }
    }
}
