/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.codesync;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.lang.reflect.Modifier;
import java.util.*;

import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.src.*;
import org.openide.util.RequestProcessor;
import org.openide.util.Task;
import org.openide.util.TaskListener;

import org.netbeans.modules.java.model.CommitListener;

import org.netbeans.modules.java.JavaConnections;
import org.netbeans.modules.java.JavaConnections.ChangeProcessor;
import org.netbeans.modules.java.settings.JavaSynchronizationSettings;
import org.netbeans.modules.java.tools.InheritanceSupport;

/**
 * This class implements a dependency manager between a class and its ancestors.
 * Currently, the implementation is designed as a stateless helper that only separates
 * class-related logic from registration logic (located in {@link SourceConnectionSupport}).
 * It is possible, that the class will be converted into stateful (with the ClassElement bean
 * as the state info) and will manage only the bean's dependencies.
 *
 * @author  sdedic
 * @version 
 */
class ClassDependencyImpl implements PropertyChangeListener, CommitListener {
    private static final int MODIFIERS_ACCESS = Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE;
            
    private static ElementFormat methodFormat;
    
    private static final boolean DEBUG = false;
    
    Registrar registrar;
    
    private JavaSynchronizationSettings SYNC_SETTINGS = null;

    public ClassDependencyImpl(Registrar r) {
        if (SYNC_SETTINGS==null)
            SYNC_SETTINGS=(JavaSynchronizationSettings)JavaSynchronizationSettings.findObject(JavaSynchronizationSettings.class);
        this.registrar = r;
    }
    
    private Collection getDirectSupertypes(ClassElement c) {
        Collection col = new LinkedList();
        col.addAll(Arrays.asList(c.getInterfaces()));
        if (c.getSuperclass() != null) {
            col.add(c.getSuperclass());
        }
        return col;
    }
    
    private void destroy(ClassElement c) {
        if (DEBUG)
            System.err.println("Class " + c.getName().getFullName() + " destructed.");
        Collection col = getDirectSupertypes(c);
        supertypesRemoved(c, col);
        // detach from the class - it is gone for good.
        c.removePropertyChangeListener(this);
    }
    
    private void registerNewClass(ClassElement c) {
        if (DEBUG)
            System.err.println("Class " + c.getName().getFullName() + " constructed.");
        Collection col = getDirectSupertypes(c);
        supertypesAdded(c, col, false);
        c.addPropertyChangeListener(this);
    }
    
    /////////////////////////////////////////////////////////////////////////////
    public void propertyChange(PropertyChangeEvent evt) {
        String propName = evt.getPropertyName();
        if (ElementProperties.PROP_INTERFACES.equals(propName))
            interfacesChanged(evt);
        else if (ElementProperties.PROP_SUPERCLASS.equals(propName))
            superclassChanged((ClassElement)evt.getSource(),
                (Identifier)evt.getOldValue(), (Identifier)evt.getNewValue());
        else if (ElementProperties.PROP_CLASS_OR_INTERFACE.equals(propName)) {
            ClassElement c = (ClassElement)evt.getSource();
            if (c.isClassOrInterface()) {
                supertypesAdded(c, getDirectSupertypes(c), false);
            }
        }
    }
    
    /**
     * Does a full refresh on a particular class' data. Mainly used when the class
     * is first seen (not constructed) in the source - e.g. after the first parse.
     */
    public void refreshClass(ClassElement c, boolean first) {
        if (DEBUG)
            System.err.println("Refreshing class " + c.getName().getFullName());
        if (!first) {
            Collection sups = getDirectSupertypes(c);
            supertypesAdded(c, sups, first);
        }
        // just in case
        c.removePropertyChangeListener(this);
        c.addPropertyChangeListener(this);
    }

    /* package-private */ void synchronizeClass(ClassElement target) {
        Element.Impl2 impl2 = (Element.Impl2)target.getCookie(Element.Impl2.class);
        if (impl2 != null && !impl2.isValid())
            return;
        
        Identifier[] ids = target.getInterfaces();
        Collection c = new ArrayList(ids.length + 1);
        c.addAll(Arrays.asList(ids));
        if (target.getSuperclass() != null)
            c.add(target.getSuperclass());
        
        // get the metadata in sync with the current state.
        supertypesAdded(target, c, true);
        
        String caption = java.text.MessageFormat.format(
            getString("MSG_SynchronizeExplicit"), // NOI18N
            new Object[] {
                target.getName().getFullName(), "" // NOI18N
            });
        // nevertheless, synchronize!
        Synchronizer sync = new Synchronizer(target, c);
        sync.overrideDisable();
        sync.setCaption(caption);
        
        RequestProcessor.postRequest(sync);
        
    }
    
    private class Synchronizer implements Runnable {
        ClassElement target;
        Collection superIDs;
        boolean overrideDisable;
        int mode;
        Collection changeList;
        String caption;
        
        Synchronizer(ClassElement target, Collection superIDs) {
            this.target = target;
            this.superIDs = superIDs;
            this.mode = 1;
        }
        
        Synchronizer(Collection changeList) {
            this.changeList = changeList;
            this.mode = 2;
        }
        
        public void overrideDisable() {
            overrideDisable = true;
        }
        
        public void setCaption(String caption) {
            this.caption = caption;
        }
        
        public void run() {
            switch (mode) {
                case 1:
                    collectSupertypeMethods();
                    break;
                case 2:
                    confirmChanges();
            }
        }
        
        private void collectSupertypeMethods() {
            if (DEBUG)
                System.err.println("collecting supertype methods");
            Collection c = implementFromSuper(target, superIDs);
            if (c == null || c.isEmpty())
                return;
            applyChanges(target, caption, c, overrideDisable);
        }
        
        private void confirmChanges() {
            autoApplyChanges(changeList);
        }
    }
    
    /**
     * Handles action(s) for removed interfaces. Removes them from the
     * "registered set", mainly.
     */
    protected void supertypesRemoved(ClassElement c, Collection removed) {
        for (Iterator it = removed.iterator(); it.hasNext(); ) {
            Identifier id = (Identifier)it.next();
            if (DEBUG)
                System.err.println("Removing dependency: " + c.getName().getFullName() + " -> " + id.getFullName());
            registrar.removeDependency(c, id);
        }
    }
    
    protected void supertypesAdded(final ClassElement c, final Collection added, 
        boolean refreshOnly) {
        Collection resolved = null;
        
        for (Iterator it = added.iterator(); it.hasNext(); ) {
            Identifier id = (Identifier)it.next();
            if (id.getResolutionStatus() != Identifier.RESOLVED) {
                if (DEBUG) {
                    System.err.println("Identifier " + id + " not resolved.");
                }
                continue;
            }
            if (resolved == null)
                resolved = new LinkedList();
            if (DEBUG)
                System.err.println("Adding dependency: " + c.getName().getFullName() + " -> " + id.getFullName());
            if (registrar.addDependency(c, id)) {
                resolved.add(id);
            }
        }
        if (resolved == null || resolved.isEmpty())
            return;

        if (refreshOnly || !c.isClassOrInterface() || (c.getModifiers() & Modifier.ABSTRACT) > 0)
            return;

        // PENDING: handle new interfaces.
        int syncMode = getSynchronizationMode(c);
        
        if (syncMode == SourceConnectionSupport.CONNECT_NOT)
            return;

        String caption = java.text.MessageFormat.format(
            getString("MSG_SynchronizeAddInterfaces"), // NOI18N
            new Object[] {
                c.getName().getFullName(), "" // NOI18N
            });
            
        Synchronizer sync = new Synchronizer(c, resolved);
        sync.setCaption(caption);
        RequestProcessor.postRequest(sync);
    }
    
    private void interfacesChanged(PropertyChangeEvent evt) {
        MultiPropertyChangeEvent e = (MultiPropertyChangeEvent)evt;
        for (Iterator changes = (Iterator)e.getIterator(); changes.hasNext(); ) {
            MultiPropertyChangeEvent part = (MultiPropertyChangeEvent)changes.next();
            int type = part.getEventType();
            
            if (type == MultiPropertyChangeEvent.TYPE_ADD) {
                supertypesAdded((ClassElement)evt.getSource(), part.getAffectedItems(), false);
            } else if (type == MultiPropertyChangeEvent.TYPE_REMOVE) {
                supertypesRemoved((ClassElement)evt.getSource(), part.getAffectedItems());
            }
        }
    }
    
    public boolean dependsOn(ClassElement c, Identifier id) {
        Identifier s = c.getSuperclass();
        if (id == null)
            return false;
        if (s != null && 
            s.compareTo(id, false))
            return true;
        Identifier[] ss = c.getInterfaces();
        for (int i = 0; i < ss.length; i++) {
            s = ss[i];
            if (s.compareTo(id, false))
                return true;
        }
        return false;
    }
    
    private void superclassChanged(ClassElement c, Identifier old, Identifier now) {
        if (old != null)
            supertypesRemoved(c, Collections.singletonList(old));
        if (now != null)
            supertypesAdded(c, Collections.singletonList(now), false);
    }

    public void changesCommited(Set added, Set removed, Map changed) {
        if (removed != null) {
            for (Iterator it = removed.iterator(); it.hasNext(); ) {
                Element el = (Element)it.next();
                if (!(el instanceof ClassElement))
                    continue;
                ClassElement c = (ClassElement)el;
                destroy(c);
            }
        }
        if (added != null) {
            for (Iterator it = added.iterator(); it.hasNext(); ) {
                Element el = (Element)it.next();
                if (!(el instanceof ClassElement))
                    continue;
                ClassElement c = (ClassElement)el;
                registerNewClass(c);
            }
        }
    }
    
    private static MethodElement findMatchingMethod(ClassElement clazz, MethodElement elem,
        boolean superclasses, boolean acceptAbstract) {
        MethodParameter[] params = elem.getParameters();
        Type[] paramTypes = new Type[params.length];
        for (int i = 0; i < paramTypes.length; i++) {
            paramTypes[i] = params[i].getType();
        }
        MethodElement m = clazz.getMethod(elem.getName(), paramTypes);
        // if we do not accept abstract methods and the method is, invalidate
        if (!acceptAbstract && 
            (m != null && (m.getModifiers() & Modifier.ABSTRACT) > 0)) {
            m = null;
        }
        // end the search if a method was found, or the search should not 
        // continue to parents.
        if (m != null || !superclasses)
            return m;
        Identifier s = clazz.getSuperclass();
        if (s == null) {
            s = ClassElement.ROOT_OBJECT;
            superclasses = false;
        } else if (s.equals(ClassElement.ROOT_OBJECT)) {
            superclasses = false;
        }
        clazz = ClassElement.forName(s.getFullName());
        if (clazz == null)
            return null;
        return findMatchingMethod(clazz, elem, superclasses, false);
    }

    private static class AddProcessor extends ChangeProcessor implements Runnable {
        private ClassElement    targetClass;
        private MethodElement   addMethod;
        private SourceException excpt;
        
        AddProcessor(String description, MethodElement model, MethodElement add, 
            ClassElement target) {
            super(description, model);
            addMethod = add;
            targetClass = target;
        }
        
        public void process() throws SourceException {
            SourceElement el = targetClass.getSource();
            el.runAtomicAsUser(this);
            if (excpt != null)
                throw excpt;
        }
        
        public void run() {
            try {
                // override the method -- use the logic inside InheritanceSupport.
                InheritanceSupport.overrideMethod(targetClass, addMethod,
                    true, false);
            } catch (SourceException ex) {
                excpt = ex;
            }
        }
    }
    
    private static class UpdateProcessor extends ChangeProcessor implements Runnable {
        MethodElement target;
        SourceException excpt;
        
        UpdateProcessor(String message, MethodElement target, MethodElement model) {
            super(message, model);
            this.target = target;
        }
        
        public void process() throws SourceException {
            target.getDeclaringClass().getSource().runAtomicAsUser(this);
            if (excpt != null)
                throw excpt;
        }
        
        public void run() {
            try {
                updateMethod();
            } catch (SourceException ex) {
                excpt = ex;
            }
        }
        
        private void updateMethod() throws SourceException {
            // update modifiers, if necessary:
            MethodElement model = (MethodElement)getModel();
            
            int mods = target.getModifiers() & ~ (Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE);
            if (model.getDeclaringClass().isClassOrInterface()) {
                mods |= model.getModifiers() & (Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE);
                mods &= ~Modifier.ABSTRACT;
            } else {
                mods |= Modifier.PUBLIC;
            }
            target.setModifiers(mods);
            
            // update name:
            target.setName(model.getName());
            
            // update return type
            target.setReturn(model.getReturn());
            
            // update exception types:
            target.setExceptions(model.getExceptions());
            
            // update parameters:
            MethodParameter[] pars1 = target.getParameters();
            MethodParameter[] pars2 = model.getParameters();
            MethodParameter[] targetPars = null;
            
            if (pars2.length != pars1.length)
                targetPars = pars2;
            else {
                for (int i = 0; i < pars1.length; i++) {
                    if (!pars1[i].getType().getFullString().equals(pars2[i].getType().
                    getFullString())) {
                        targetPars = pars2;
                        break;
                    }
                }
            }
            if (targetPars != null)
                target.setParameters(targetPars);
        }
    }
    
    private static String getString(String key) {
        return SourceConnectionSupport.getString(key);
    }
    
    private static ChangeProcessor createUpdateProcessor(MethodElement desired, MethodElement found) {
        boolean same = true;
        int access;
        
        if (DEBUG) {
            System.err.println("creating update proc for " +
                ModelEventAdapter.reportElementID(found) + " to " +
                ModelEventAdapter.reportElementID(desired));
        }
        same = desired.getName().equals(found.getName()) &&
            desired.getReturn().getFullString().equals(found.getReturn().getFullString()) &&
            desired.getParameters().length == found.getParameters().length;
        if (DEBUG) {
            System.err.println("basic check: " + same);
        }
        if (desired.getDeclaringClass().isClassOrInterface()) {
            access = desired.getModifiers() & MODIFIERS_ACCESS;
            // PENDING: have the implementing class obey access rules!!
        }  else {
            access = Modifier.PUBLIC;
        }
        int foundAccess = found.getModifiers() & MODIFIERS_ACCESS;
        same &= (foundAccess == access) ||
                (foundAccess == Modifier.PUBLIC && access == Modifier.PROTECTED) ||
                (foundAccess == Modifier.PROTECTED && access == 0);
              
        if (DEBUG) {
            System.err.println("modifier check: " +
                Modifier.toString(found.getModifiers() & MODIFIERS_ACCESS) +
                " vs. " + Modifier.toString(access));
        }
        if (same) {
            MethodParameter[] pars1 = desired.getParameters();
            MethodParameter[] pars2 = found.getParameters();
            for (int i = 0; same && i < pars1.length; i++) {
                same = pars1[i].getType().getFullString().equals(
                    pars2[i].getType().getFullString());
                if (DEBUG && !same) {
                    System.err.println("param mismatch at #" + i + ", " +
                        pars1[i].getType().getFullString() + " vs. " + 
                        pars2[i].getType().getFullString());
                }
            }
            Identifier[] exc1 = desired.getExceptions();
            Identifier[] exc2 = found.getExceptions();

            for (int i = 0; same && i < exc2.length; i++) {
                same = false;
                for (int j = 0; j < exc1.length; j++) {
                    if (exc2[i].compareTo(exc1[j], false)) {
                        same = true;
                        break;
                    }
                }
                if (DEBUG && !same) {
                    System.err.println("exception " + exc2[i] + " not found in supertype ");
                }
            }
        }
        if (same)
            return null;

        MethodElement c = (MethodElement)desired.clone();
        try {
            c.setModifiers((found.getModifiers() & ~MODIFIERS_ACCESS) | access);
        } catch (SourceException ex) {
            // should not happen -- c is memory-represented.
        }
            
        String updateFormat= getString("MSG_UpdateMethod"); // NOI18N
        String msg = java.text.MessageFormat.format(updateFormat, 
            new Object[] { formatMethod(found), formatMethod(c), 
                desired.getDeclaringClass().getName().getFullName()
            });
        
        return new UpdateProcessor(msg, found, desired);
    }
    
    private static ChangeProcessor createAddProcessor(ClassElement target, MethodElement t) {
        // don't add nonabstract methods of superclasses
        if (t.getDeclaringClass().isClassOrInterface() && 
            (t.getModifiers() & Modifier.ABSTRACT) == 0)
            return null;
        
        String addMessageFormat = getString("MSG_AddMethod"); // NOI18N
        String msg = java.text.MessageFormat.format(addMessageFormat, 
            new Object[] { t.getDeclaringClass().getName().getFullName(), formatMethod(t) }
            );
        
        return new AddProcessor(msg, t, t, target);
    }
    
    private static String formatMethod(MethodElement m) {
        if (methodFormat == null) {
            methodFormat = new ElementFormat(getString("FMT_MethodSignature"));
        }
        return methodFormat.format(m);
    }
    
    public void connectionNotify(ClassElement target, ClassElement source, 
        Enumeration newItems, Enumeration changeEvents) {
        if (DEBUG) {
            System.err.println("[" + target.getName().getFullName() + "] got connectionNotify. ");
        }
        
        int syncMode = getSynchronizationMode(target);
        
        if (syncMode == SourceConnectionSupport.CONNECT_NOT)
            return;
        
        final Collection changes = synchronizeMethods(target, newItems, changeEvents, true);
        if (DEBUG) {
            if (changes == null) {
                System.err.println("no changes required");
            } else 
                System.err.println("identified " + changes.size() + " changes");
        }
        if (changes == null || changes.isEmpty())
            return;
        final String caption = java.text.MessageFormat.format(getString("MSG_SynchronizeCaption"),
            new Object[] {
                target.getName().getFullName(), source.getName().getFullName()
            });
        if (DEBUG) {
            System.err.println("caption for the dialog = " +  caption);
        }
        applyChanges(target, caption, changes);
    }

    private void applyChanges(final ClassElement target, final String caption, final Collection changes) {
        applyChanges(target, caption, changes, false);
    }
    
    private void applyChanges(final ClassElement target, final String caption, final Collection changes,
        boolean ignoreNot) {
        final int syncMode = getSynchronizationMode(target);

        if (!ignoreNot && (syncMode == SourceConnectionSupport.CONNECT_NOT || !SYNC_SETTINGS.isEnabled()))
            return;

        if (syncMode == SourceConnectionSupport.CONNECT_AUTO) {
            RequestProcessor.postRequest(new Synchronizer(changes));
        } else {
            List l = new ArrayList(changes);
            JavaConnections.SyncRequest req = JavaConnections.scheduleShowChangesDialog(
                caption, l, (byte)syncMode);
            req.getTask().addTaskListener(new SyncRequestListener(target, req));
        }
    }
    
    private void autoApplyChanges(Collection changes) {
        Collection failed = null;
        
        for (Iterator it = changes.iterator(); it.hasNext(); ) {
            ChangeProcessor proc = (ChangeProcessor)it.next();
            try {
                proc.process();
            } catch (SourceException ex) {
                if (failed == null)
                    failed = new LinkedList();
                failed.add(proc);
            }
        }
        if (failed == null)
            return;
        StringBuffer sb = new StringBuffer(java.text.MessageFormat.format(
            getString("MSG_ChangesFailed"),
            new Object[] { new Integer(failed.size()) }));
        for (Iterator it = failed.iterator(); it.hasNext(); ) {
            sb.append('\n');
            sb.append(((ChangeProcessor)it.next()).getDisplayName());
        }
        NotifyDescriptor desc = new NotifyDescriptor.Message(sb.toString(), 
            NotifyDescriptor.WARNING_MESSAGE);
        TopManager.getDefault().notify(desc);
    }
    
    private class SyncRequestListener implements TaskListener {
        JavaConnections.SyncRequest request;
        ClassElement target;
        
        SyncRequestListener(ClassElement t, JavaConnections.SyncRequest r) {
            request = r;
            target = t;
        }
        
        public void taskFinished(Task t) {
            t.removeTaskListener(this);
            setSynchronizationMode(target, request.getSyncType());
        }
    }
    
    protected Collection synchronizeMethods(ClassElement target, 
        Enumeration neededMethods, boolean deepScan) {
        return synchronizeMethods(target, neededMethods, null, deepScan);
    }
    
    /**
     * Attempts to synchronized the class' implementation with the passed information.
     * @param newItems new methods that should be inserted into the class.
     * @param changeEvents list of changes that should be applied to the class.
     */
    public Collection synchronizeMethods(ClassElement target, Enumeration newItems, 
        Enumeration changeEvents, boolean deepScan) {
        
        // for new items, try to locate the method in the target class.
        Collection changes = null;
        
        if (newItems != null) {
            while (newItems.hasMoreElements()) {
                Object o = newItems.nextElement();
                if (!(o instanceof MethodElement))
                    continue;
                MethodElement m = (MethodElement)o;
                if (DEBUG) {
                    System.err.println("New method: " + 
                        ModelEventAdapter.reportElementID(m));
                }

                MethodElement found = findMatchingMethod(target, m, deepScan, true);
                if (DEBUG) {
                    if (found == null)
                        System.err.println("... no matching method");
                    else
                        System.err.println("Found matching method: " + 
                            ModelEventAdapter.reportElementID(found) + " in " +
                            found.getDeclaringClass().getName().getFullName());
                }
                ChangeProcessor proc = null;

                // PENDING: what if the method is abstract ??
                if (found != null) {
                    // now, the method may have been in one of the ancestors.
                    // this IS a problem, since if it is not EXACTLY the same (exception ids)
                    // we cannot modify it.
                    if (found.getDeclaringClass() != target) {
                        continue;
                    }
                    proc = createUpdateProcessor(m, found);
                    if (proc == null) {
                        // PENDING: link found to m so that the dependency can
                        // be maintained over time.
                        continue;
                    }
                    // the method IS present, but is slightly different.
                } else {
                    if (DEBUG) {
                        System.err.println("creating Add processor...");
                    }
                    proc = createAddProcessor(target, m);
                }
                if (proc != null) {
                    if (changes == null)
                        changes = new LinkedList();
                    changes.add(proc);
                }
            }
        }
        
        // new items have been created. Now consult the changed ones.
        if (changeEvents != null) {
            while (changeEvents.hasMoreElements()) {
                JavaConnections.Change ch = (JavaConnections.Change)changeEvents.nextElement();
                if (ch.getChangeType() != JavaConnections.TYPE_METHODS_CHANGE) {
                    continue;
                }
                MethodElement oldState = (MethodElement)ch.getOldElement();
                MethodElement newState = (MethodElement)ch.getNewElement();
                ChangeProcessor proc = null;

                if (DEBUG) {
                    System.err.println("Change from : " + 
                        ModelEventAdapter.reportElementID(oldState));
                    System.err.println("- to: " + 
                        ModelEventAdapter.reportElementID(newState));
                }
                // first, try to search for the new method element, it may be present.
                MethodElement found = findMatchingMethod(target, newState, deepScan, true);
                if (found != null) {
                    if (target != found.getDeclaringClass()) {
                        // something bad -- possibly conflicting method in a superclass.
                        // PENDING: warn the user.
                        continue;
                    }
                    proc = createUpdateProcessor(newState, found);
                    if (proc == null) {
                        // the method is identical.
                        // PENDING: link the method to the discovered super's one.
                        continue;
                    }
                } else if ((found = findMatchingMethod(target, oldState, deepScan, true)) != null) {
                    if (target != found.getDeclaringClass()) {
                        // if uncompatible, it can pose a problem.
                        // PENDING: warn the user, if this is the case.
                        continue;
                    }
                    proc = createUpdateProcessor(newState, found);
                    if (proc == null) {
                        // huh ? exactly the same method ?
                        continue;
                    }
                } else {
                    proc = createAddProcessor(target, newState);
                }
                if (proc != null) {
                    if (changes == null)
                        changes = new LinkedList();
                    changes.add(proc);
                }
            }
        }
        
        return changes;
    }

    private void addSupertypeMethods(Identifier superID, Map methods, Set seen) {
        if (seen.contains(superID))
            return;

        ClassElement c = ClassElement.forName(superID.getFullName());
        seen.add(superID);
        if (c == null)
            return;
        
        MethodElement[] ms = c.getMethods();
        boolean clazz = c.isClassOrInterface();
        
        if (DEBUG) {
            System.err.println("Collecting supertypes of " + c.getName().getFullName());
        }
        Identifier[] itfs = c.getInterfaces();
        for (int i = 0; i < itfs.length; i++) {
            addSupertypeMethods(itfs[i], methods, seen);
        }
        if (clazz && c.getSuperclass() != null) {
            // adds all superclass' abstract methods.
            addSupertypeMethods(c.getSuperclass(), methods, seen);
        }
        
        if (DEBUG) {
            System.err.println("Adding methods of " + c.getName().getFullName());
        }
        for (int i = 0; i < ms.length; i++) {
            MethodElement m = ms[i];
            MethodElement.Key key = new MethodElement.Key(m);
            if (!clazz || ((m.getModifiers() & Modifier.ABSTRACT) > 0)) {
                if (methods.containsKey(key)) {
                    if (DEBUG) {
                        System.err.println("Method" + ModelEventAdapter.reportElementID(m) + " already recorded");
                    }
                    continue;
                }
                methods.put(key, m);
                if (DEBUG) {
                    System.err.println("Adding " + ModelEventAdapter.reportElementID(m));
                }
            } else {
                if (DEBUG) {
                    System.err.println("Implemented " + ModelEventAdapter.reportElementID(m));
                }
                // un-register the method:
                if (methods.remove(key) != null) {
                    if (DEBUG) {
                        System.err.println("-- removing from list.");
                    }
                }
            }
        }
        if (DEBUG) {
            System.err.println("Done processing of " + c.getName().getFullName());
        }
    }
    
    private Collection implementFromSuper(ClassElement target, Collection supertypes) {
        Map methods = new HashMap(31);
        for (Iterator it = supertypes.iterator(); it.hasNext(); ) {
            Identifier id = (Identifier)it.next();
            if (DEBUG) {
                System.err.println("Top-level: adding methods for: " + id.getFullName());
            }
            addSupertypeMethods(id, methods, new HashSet(11));
        }
        // now, try to match the methods :-)
        Collection c = synchronizeMethods(target, Collections.enumeration(
            methods.values()), null, true);
        if (c == null || c.isEmpty())
            return null;
        else
            return c;
    }
    
    protected byte getSynchronizationMode(ClassElement t) {
        return (byte)registrar.getSynchronizationMode(t);
    }
    
    protected void setSynchronizationMode(ClassElement t, byte value) {
        registrar.setSynchronizationMode(t, value);
    }
}
