/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.environment;

import java.util.*;

import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.Repository;
import org.openide.execution.NbClassPath;
import org.openide.loaders.DataFolder;
import org.openide.loaders.FolderInstance;
import org.openide.util.Lookup;

/**
 * This class implements compile/runtime library path needed for the installed
 * modules to work well with the IDE. The folder watches for changes in the path
 * and mounts/unmounts FileSystems to/from repository.
 *
 * @author  sdedic
 * @version 
 */
public class LibraryFolder extends FolderInstance {
    /**
     * Contains String[systemName] -> FileSystem for file systems mounted by this Folder in the past.
     */
    Map myFileSystems = new HashMap();
    
    /** Cached reference to the Repository.
    */
    Repository repository;
    
    public LibraryFolder(DataFolder fld) {
        super(fld);
        repository = (Repository)org.openide.util.Lookup.getDefault().lookup(Repository.class);
	//System.err.println("Creating FolderInstance for folder " + fld);
	//System.err.println("Repository = " + repository);
        LibraryProcessor.init();
	recreate();
    }

    /*
     * Should the LibraryFolder represent NbClassPath ??
     */
    protected Object createInstance(InstanceCookie[] cookies) 
    throws java.io.IOException, ClassNotFoundException {
        Map toRemove = new HashMap(myFileSystems);
	// will collect filesystems that are not known yet.
        Collection toAdd = new ArrayList(cookies.length);	
	// found map filters out duplicities
	Map found = new HashMap(cookies.length * 4 / 3);
	Collection ordered = new ArrayList(cookies.length);
	
        for (int i = 0; i < cookies.length; i++) {
            try {
                Object o = cookies[i].instanceCreate();
		//System.err.println("Got instance " + o);
                if (o instanceof FileSystem) {
		    FileSystem fs = (FileSystem)o;
		    String sName = fs.getSystemName();
		    if (!found.containsKey(sName)) {
			// first time such entry was seen. Remove it from the remove set.
			FileSystem prev = (FileSystem)toRemove.remove(sName);
			if (prev == null) {
			    // wasn't mounted before.
			    toAdd.add(fs);
			} else {
			    // the filesystem was already mounted, record it into found map.
			    fs = prev;
			}
			found.put(sName, fs);
		    }    
                }
            } catch (Throwable t) {
                t.printStackTrace();
            }
        }
        
        // try to unmount toRemove from the Repository.
        removeFileSystems(toRemove.values());
	
	Collection notMounted = addFileSystems(toAdd);
        // mount the new ones to the repository
	found.values().removeAll(notMounted);
	ordered.removeAll(notMounted);
        myFileSystems = found;
        
        return createNbClassPath(ordered);
    }

    /**
     * Accepts only FileSystem instances.
     */
    protected InstanceCookie acceptCookie(InstanceCookie cookie) throws java.io.IOException, ClassNotFoundException {
        Class c = cookie.instanceClass();
	//System.err.println("AcceptCookie on " + c);
        if (FileSystem.class.isAssignableFrom(c))
            return cookie;
        else
            return null;
    }

    private Collection addFileSystems(Collection fss) {
        if (repository == null)
            return Collections.EMPTY_LIST;
	Collection notMounted = null;
        for (Iterator it = fss.iterator(); it.hasNext(); ) {
            FileSystem fs = (FileSystem)it.next();
            // mount only if the fs is not mounted yet.
            if (repository.findFileSystem(fs.getSystemName()) == null) {
                // do not disturb the user.
		//System.err.println("Mounting fs " + fs.getSystemName());
                fs.setHidden(true);
                repository.addFileSystem(fs);
            } else {
		//System.err.println("fs " + fs.getSystemName() + " already mounted.");
		if (notMounted == null)
		    notMounted = new LinkedList();
		notMounted.add(fs);
	    }
        }
	return notMounted == null ? Collections.EMPTY_LIST : notMounted;
    }
    
    private void removeFileSystems(Collection fss) {
        if (repository == null)
            return;
        for (Iterator it = fss.iterator(); it.hasNext(); ) {
            FileSystem fs = (FileSystem)it.next();
            // note: removes exactly those fileSystems that have been mounted
            // by this object.
            repository.removeFileSystem(fs);
        }
    }
    
    public void remount() {
        /*
	myFileSystems = Collections.EMPTY_LIST;
        //System.err.println("-----------------------------recreating...");
	recreate();
         */
        moveToEnd();
    }
    
    private void moveToEnd() {
        synchronized (repository) {
            FileSystem[] fss = repository.toArray();
            int[] perm = new int[fss.length];
            boolean permute = false;
            int myBase = perm.length - myFileSystems.size();
            int pos = 0;
            int pos2 = 0;

            Collection myMounted = myFileSystems.values();

            /*
            System.err.println("My filesystems:");
            for (Iterator it = myMounted.iterator(); it.hasNext(); ) {
                FileSystem fs = (FileSystem)it.next();
                System.err.println(fs.getSystemName() + "@" + System.identityHashCode(fs));
            }
            System.err.println("--------------");
            System.err.println("myBase = " + myBase + ",  total = " + perm.length);
            */

            for (int i = 0; i < fss.length; i++) {
                FileSystem fs = fss[i];
                if (myMounted.contains(fs)) {
                    perm[myBase + pos2] = i;
                    pos2++;
                    permute |= i < myBase;
                } else {
                    perm[pos] = i;
                    pos++;
                }
            }

            /*
            for (int i = 0; i < perm.length; i++) 
                System.err.println("perm[" + i + "] = " + perm[i]);
                */

            if (permute)
                repository.reorder(perm);
        }
    }
    
    private NbClassPath createNbClassPath(Collection filesystems) {
        StringBuffer sb = new StringBuffer();
        for (Iterator it = filesystems.iterator(); it.hasNext(); ) {
            FileSystem fs = (FileSystem)it.next();
            java.io.File f = NbClassPath.toFile(fs.getRoot());
            if (f == null)
                continue;
            if (sb.length() > 0)
                sb.append(java.io.File.pathSeparatorChar);
            sb.append(f.getPath());
        }
        return new NbClassPath(sb.toString());
    }
}
