/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.environment;

import java.io.IOException;

import org.w3c.dom.*;
import org.xml.sax.*;

import org.openide.filesystems.FileSystem;
import org.openide.filesystems.JarFileSystem;
import org.openide.loaders.XMLDataObject;
import org.openide.cookies.InstanceCookie;

/**
 *
 * @author  sdedic
 * @version 
 */
public class LibraryProcessor implements XMLDataObject.Processor, InstanceCookie {
    static final String JAVALIB_DTD_PUBLIC_ID = 
        "-//NetBeans IDE//DTD JavaLibrary//EN"; // NOI18N
    
    static final String TAG_LIBRARY = "Library"; //NOI18N
    static final String TAG_ARCHIVE = "Archive"; // NOI18N
    static final String ATTR_NAME = "name"; // NOI18N

    private static final char FILE_SEPARATOR = java.io.File.separatorChar;
    /**
     * XML info for the library def document.
     */
    private static XMLDataObject.Info   xmlinfo;
    
    synchronized static void init() {
        if (xmlinfo == null) {
            xmlinfo = new XMLDataObject.Info();
            xmlinfo.addProcessorClass (LibraryProcessor.class);
            XMLDataObject.registerInfo (JAVALIB_DTD_PUBLIC_ID, xmlinfo);
            XMLDataObject.registerCatalogEntry(JAVALIB_DTD_PUBLIC_ID, 
                "nbres:/org/netbeans/modules/java/environment/library.dtd");
        }
    }
    
    /** XML data object. */
    protected XMLDataObject xmlDataObject;
    
    /**
     * Library File System created from the xml-info.
     */
    FileSystem  libraryFileSystem;

    /** Creates new LibraryProcessor */
    public LibraryProcessor() {
    }

    /** When the XMLDataObject creates new instance of the processor,
     * it uses this method to attach the processor to the data object.
     *
     * @param xmlDO XMLDataObject
     */
    public void attachTo(XMLDataObject xmlDO) {
        xmlDataObject = xmlDO;
    }
    
    /** Create an instance.
     * @return the instance of type {@link #instanceClass}
     * @exception IOException if an I/O error occured
     * @exception ClassNotFoundException if a class was not found
     */
    public Object instanceCreate() throws java.io.IOException, ClassNotFoundException {
        if (libraryFileSystem != null)
            return libraryFileSystem;
        try {
            //System.err.println("LibraryProcessor: in Instance create...");
            loadLibrary(
                /*xmlDataObject.getName(), */xmlDataObject.getDocument()
            );
            /*
            xmlDataObject.addPropertyChangeListener(WeakListener.propertyChange (
                tc, xmlDataObject
            ));
            configuration = tc;
             */
            return libraryFileSystem;
        } catch (org.xml.sax.SAXException e) {
            IOException ex = new IOException ();
            org.openide.TopManager.getDefault ().getErrorManager ().copyAnnotation (
                ex, e
            );
            throw ex;
        }
    }
    
    /** The representation type that may be created as instances.
     * Can be used to test whether the instance is of an appropriate
     * class without actually creating it.
     *
     * @return the representation class of the instance
     * @exception IOException if an I/O error occurred
     * @exception ClassNotFoundException if a class was not found
     */
    public Class instanceClass() {
        return LibraryArchive.class;
    }
    
    /** The bean name for the instance.
     * @return the name
     */
    public String instanceName() {
        return instanceClass().getName();
    }

    ////////////////////////////////////////////////////////////////////////
    //
    void loadLibrary (Document doc) {
        Element rootElement = doc.getDocumentElement();
        if (!TAG_LIBRARY.equals(rootElement.getTagName())) {
            //System.err.println(rootElement.getTagName());
            return;
        }
        readLibraryID(rootElement);
    }
    
    void readLibraryID(Element def) {
        //System.err.println("read library id from : " + def);
        NodeList nL = def.getElementsByTagName (TAG_ARCHIVE);
        Node node = nL.item (0);
        Element archiveElement = (Element)node;

	if (archiveElement == null)
	    return;
	            
        String nameString = archiveElement.getAttribute(ATTR_NAME);
        //System.err.println("library archive name = " + nameString);
        
        if (FILE_SEPARATOR != '/') {
            nameString.replace('/', FILE_SEPARATOR);
        }
        
        java.io.File archiveRoot = Utilities.findInstalledFile(nameString);
        
        if (archiveRoot == null) {
            //System.err.println("Cannot find " + nameString);
            return;
        }

        // load the library using LibraryArchive file system:
        JarFileSystem jfs = new LibraryArchive();
        try {
            jfs.setJarFile(archiveRoot);
            libraryFileSystem = jfs;
        } catch (java.beans.PropertyVetoException ex) {
            //System.err.println("Vetoed create " + archiveRoot + ": " + ex);
        } catch (IOException ex) {
            //System.err.println("Cannot create " + archiveRoot + ": " + ex);
        }
    }
}
