/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.imptool;

import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListDataEvent;
import javax.swing.table.TableColumn;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.Dimension;

import org.openide.explorer.view.ListView;
import org.openide.explorer.view.NodeListModel;
import org.openide.explorer.view.Visualizer;
import org.openide.explorer.ExplorerManager;
import org.openide.util.WeakListener;
import org.openide.nodes.*;

/** View compound from two views - list, table. Allows to view list of nodes on the left
 * and its properties in table on the right.
 *
 * @author  jrojcek
 * @version 
 */
public class ListTableView extends ListView {
        
    /** flag that allows group more requests into one */
    private boolean tableChanging = false;
    /** manager is used only for synchronization of header name */
    private ExplorerManager manager;
    /** listener on explorer manager */
    private PropertyChangeListener wlpc;
    /** table view that is controlled by list view */
    private TableView.ControlledTableView controlledTableView;
    /** listener on changes in list model */
    private Listener listener;

    /** Create ListTableView with default NodeTableModel
     */
    public ListTableView() {
        this(null);
    }
    
    /** Creates ListTableView with provided NodeTableModel.
     * @param ntm node table model
     */
    public ListTableView(NodeTableModel ntm) {
        // do not use scroll bars, this scroll pane is dummy, only for border painting
        setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        // remove list from viewportview
        setViewportView(null);

        // insert list into new scrollpane
        JScrollPane listView = new JScrollPane(list);
        listView.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        listView.setBorder(null);
        
        // create table view controlled by new scrollpane
        controlledTableView = ntm == null 
                              ? new TableView.ControlledTableView(listView)
                              : new TableView.ControlledTableView(listView, ntm);
        // hack to avoid moving of compound view
        setLayout(new TableView.MyScrollPaneLayout());
        // add compound scrollpane (JPanel) into this scrollpane
        add(TableView.MyScrollPaneLayout.COMPOUND_SCROLLPANE,
            controlledTableView.compoundScrollPane());

        listener = new Listener();
        delayedFireTableDataChanged();
    }
    
    /** Set columns.
     * @param nodes columns
     */
    public void setProperties(Node.Property[] props) {
        controlledTableView.setProperties(props);
    }

    /** Sets resize mode of table.
     * 
     * @param mode - One of 5 legal values: <pre>JTable.AUTO_RESIZE_OFF,
     *                                           JTable.AUTO_RESIZE_NEXT_COLUMN,
     *                                           JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS, 
     *                                           JTable.AUTO_RESIZE_LAST_COLUMN, 
     *                                           JTable.AUTO_RESIZE_ALL_COLUMNS</pre>
     */    
    public void setAutoResizeMode(int mode) {
        controlledTableView.setAutoResizeMode(mode);
    }

    /** Allows set parameters of column.
     * @param i column index
     * @return table column avaliable to set sizes of columns, ...
     */    
    public TableColumn getTableColumn(int columnIndex) {
        return controlledTableView.getTableColumn(columnIndex);
    }
    
    /** Set preferred size of list view
     * @param dim
     */    
    public void setListViewPreferredSize(Dimension dim) {
        controlledTableView.setControllingViewPreferredSize(dim);
    }
    
    /** Set preferred size of table view
     * @param dim preferred size
     */    
    public void setTableViewPreferredSize(Dimension dim) {
        controlledTableView.setPreferredSize(dim);
    }

    public void addNotify() {
        super.addNotify();

        ExplorerManager newManager = ExplorerManager.find (this);
        if (newManager != manager) {
            if (manager != null) {
                manager.removePropertyChangeListener (wlpc);
            }
            manager = newManager;
            manager.addPropertyChangeListener(wlpc = WeakListener.propertyChange (listener, manager));
            controlledTableView.setHeaderText(manager.getExploredContext().getDisplayName());
        }

        list.getModel().addListDataListener(listener);
        delayedFireTableDataChanged();
    }

    public void removeNotify() {
        super.removeNotify();
        list.getModel().removeListDataListener(listener);
        // clear node listeners
        controlledTableView.setNodes(new Node[] {});
    }

    /** Chnage table data in awt thread */
    private void delayedFireTableDataChanged() {
        if (tableChanging)
            return;
        tableChanging = true;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                if (list.getCellBounds(0, 0) != null)
                    controlledTableView.setRowHeight(list.getCellBounds(0, 0).height);
                changeTableModel();
                tableChanging = false;
            }
        });
    }

    /** Change table model. Sets rows (nodes) of table mmodel */
    private void changeTableModel() {
        Node[] nodes = new Node[list.getModel().getSize()];
        for (int i = 0; i < list.getModel().getSize(); i++) {
            nodes[i] = Visualizer.findNode(list.getModel().getElementAt(i));
        }
        controlledTableView.setNodes(nodes);
    }

    /** Listenes on changes in list model.
     */
    private class Listener implements ListDataListener, PropertyChangeListener {
        /**
         * Sent after the indices in the index0,index1 
         * interval have been inserted in the data model.
         * The new interval includes both index0 and index1.
         *
         * @param e  a ListDataEvent encapuslating the event information
         */
        public void intervalAdded(ListDataEvent e) {
            delayedFireTableDataChanged();
        }

        /**
         * Sent after the indices in the index0,index1 interval
         * have been removed from the data model.  The interval 
         * includes both index0 and index1.
         *
         * @param e  a ListDataEvent encapuslating the event information
         */
        public void intervalRemoved(ListDataEvent e) {
            delayedFireTableDataChanged();
        }

        /**
         * Sent when the contents of the list has changed in a way 
         * that's too complex to characterize with the previous 
         * methods.  Index0 and index1 bracket the change.
         *
         * @param e  a ListDataEvent encapuslating the event information
         */
        public void contentsChanged(ListDataEvent e) {
            delayedFireTableDataChanged();
        }

        /** Updates header name.
         * @param evt event
         */
        public void propertyChange(PropertyChangeEvent evt) {
            if (ExplorerManager.PROP_EXPLORED_CONTEXT.equals(evt.getPropertyName()))
                controlledTableView.setHeaderText(manager.getExploredContext().getDisplayName());
        }
    }
/*    
    public static void main (String[] args) {
System.gc();
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                Node n = new org.netbeans.core.ModuleNode();
                //org.openide.TopManager.getDefault ().getPlaces ().nodes ().repository ();
                ExplorerManager em = new ExplorerManager();
                em.setRootContext(n);
                ListTableView ttv = new ListTableView();
                ttv.model.setDepth(2);
                ttv.setAutoResizeMode(javax.swing.JTable.AUTO_RESIZE_OFF);
                ttv.setProperties(
//                    n.getChildren().getNodes()[0].getPropertySets()[0].getProperties());
                    new Node.Property[]{
                        new PropertySupport.ReadWrite (
                            "enabled", // NOI18N
                            Boolean.TYPE,
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "PROP_modules_enabled"),
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "HINT_modules_enabled")
                        ) {
                            public Object getValue () {
                                return null;
                            }

                            public void setValue (Object o) {
                            }
                        },
                        new PropertySupport.ReadOnly (
                            "specVersion", // NOI18N
                            String.class,
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "PROP_modules_specversion"),
                            org.openide.util.NbBundle.getMessage (org.netbeans.core.Main.class, "HINT_modules_specversion")
                        ) {
                            public Object getValue () {
                                return null;
                            }

                        }
                    }
                );

                org.openide.explorer.ExplorerPanel ep = new org.openide.explorer.ExplorerPanel (em);
                ep.setLayout (new java.awt.BorderLayout ());
                ep.add ("Center", ttv);
                ep.open ();
            }
        });
    }
*/
/*    private static Node.Property[] getPropertyColumnSet(Node node) {
        Node.Property[] retProps = new Node.Property[] {};
        Node.PropertySet[] propSets = node.getPropertySets();
        for (int i = 0; i < propSets.length; i ++) {
            Node.Property[] props = propSets[i].getProperties();
            Node.Property[] propsBuf = new Node.Property[retProps.length + props.length];
            System.arraycopy(retProps, 0, propsBuf, 0, retProps.length);
            System.arraycopy(props, 0, propsBuf, retProps.length, props.length);
            retProps = propsBuf;
        }
        return retProps;
    }
*/
    /*
    public static void main (String[] args) {
System.gc();
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                Node n = org.openide.TopManager.getDefault ().getPlaces ().nodes ().repository ();
                ExplorerManager em = new ExplorerManager();
                em.setRootContext(n);
                ListTableView ttv = new ListTableView();
                ttv.setListViewPreferredSize(new Dimension(50, 50));
                ttv.setTableViewPreferredSize(new Dimension(200, 50));
                ttv.setPreferredSize(new Dimension(300, 100));
                ttv.model.setDepth(3);
                ttv.setAutoResizeMode(javax.swing.JTable.AUTO_RESIZE_OFF);
                ttv.setProperties(
                    n.getChildren().getNodes()[0].getPropertySets()[0].getProperties()
                );

                org.openide.explorer.ExplorerPanel ep = new org.openide.explorer.ExplorerPanel (em);
                ep.setLayout (new java.awt.BorderLayout ());
                ep.add ("Center", ttv);
                ep.open ();
            }
        });
    }
     */
}
