/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.PropertyChangeEvent;

import org.openide.src.*;

/**
 *
 * @author  sdedic
 * @version 
 */
abstract class CallableImpl extends MemberElementImpl implements ConstructorElement.Impl {
    private static final boolean DEBUG = false;
    
    /** Special tag value for methods without parameters (performance enhancement)
     */
    static final MethodParameter[] NO_PARAMETERS = new MethodParameter[0];
    
    /** Array of method parameters. If the method has no parameters, the value is
     * an empty array (NO_PARAMETERS), not null.
     */
    private MethodParameter[]          parameters;
    
    /**
     * Array of checked exceptions thrown from the method/constructor.
     * BAD DESIGN here, this should be ClassType (extends RefType extends Type), if
     * there was one in the OpenAPI.
     */
    private Identifier[]                exceptions;

    /** Flyweight support for PROP_EXCEPTION property.
     */
    private static IdentifierArrayProperty  exceptionSupport;
    
    /** Flyweight support for PROP_PARAMETERS property.
     */
    private static MethodParamSupport paramSupport;
    
    private String                      cachedBody;
    
    private transient long              bodyHash;
    
    // Construction
    ///////////////////////////////////////////////////////////////////////////////////

    /*
    CallableImpl(Identifier name, int modifiers, MethodParameter[] params, 
        Identifier[] exceptions) {
        super(name, modifiers);
        
        if (exceptions == null || exceptions.length == 0) {
            this.exceptions = IdentifierArrayProperty.EMPTY;
        } else {
            this.exceptions = exceptions;
        }
        if (params == null || params.length == 0) {
            this.parameters = NO_PARAMETERS;
        } else {
            this.parameters = params;
        }
    }
     */
    
    CallableImpl(DefaultLangModel model) {
        super(model);
        parameters = NO_PARAMETERS;
        exceptions = IdentifierArrayProperty.EMPTY;
        if (paramSupport == null) {
            paramSupport = new MethodParamSupport();
            exceptionSupport = new ExceptionSupport();
        }
        javadoc = new JavaDocImpl.Method(null, this);
    }

    public JavaDoc.Method getJavaDoc() {
        return (JavaDoc.Method)javadoc;
    }
    
    protected void copyBody(String bodyContent) {
        cachedBody = bodyContent;
    }
    
    void updateBody(String content) {
        if (content == null) {
            bodyHash = -1;
            return;
        }
        long newHash = computeHash(content);
        if (newHash == bodyHash) 
            return;
        if (DEBUG) {
            System.err.println("[" + this + "] updating body to \"" + content + "\"");
            System.err.println("[" + this + "] hash is " + newHash + " old hash = " + bodyHash);
        }
        if (bodyHash != -1) 
            addPropertyChange(new PropertyChangeEvent(getElement(), PROP_BODY, null, null));
        bodyHash = newHash;
    }
    
    // Getters
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Returns parameters for the method.
     * @return array of parameter descriptions.
     */
    public final MethodParameter[] getParameters() {
        if (parameters == NO_PARAMETERS)
            return parameters;
        MethodParameter[] cloneParams = new MethodParameter[parameters.length];
        for (int i = 0; i < parameters.length; i++) {
            MethodParameter p = parameters[i];
            cloneParams[i] = new MethodParameter(
                p.getName(),
                p.getType(),
                p.isFinal()
                );
        }
        return cloneParams;
    }

    /** Retrieves checked exceptions declared to be thrown from the callable.
     * @return Identifiers representing exception types thrown from the method.
     */
    public final Identifier[] getExceptions() {
        if (exceptions == IdentifierArrayProperty.EMPTY)
            return exceptions;
        return (Identifier[])exceptions.clone();
    }
    
    /** Retrieves body of the method.
     */
    public final String getBody() {
        try {
            if (cachedBody != null)
                return cachedBody;
            return getMethodBinding().getBodyContent();
        } catch (SourceException ex) {
            return null;
        }
    }
    
    protected void notifyCreate() {
        super.notifyCreate();
        if (cachedBody != null) {
            String body;
            
            cachedBody = null;
            body = getBody();
            if (body != null)
                bodyHash = computeHash(body);
        }

    }
    
    private long computeHash(String bodyContent) {
        java.util.zip.CRC32 crc = new java.util.zip.CRC32();
        crc.update(bodyContent.getBytes());
        return crc.getValue();
    }
    
    protected void createFromModel(Element model) throws SourceException {
        super.createFromModel(model);

        ConstructorElement m = (ConstructorElement)model;
        setParameters(m.getParameters());
        setExceptions(m.getExceptions());
        copyBody(m.getBody());
        setJavaDocText(m.getJavaDoc().getRawText(), true);
    }
    
    // Setters
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the parameters for the method.
     * @param params array of parameter. null means no parameters at all.
     * @throw SourceException if a VetoableListener disagrees or other error occurs.
     */
    public void setParameters(MethodParameter[] params) throws SourceException {
        // PENDING: translate (or try to translate) parameter types to something
        // appropriate to this class.
        Object token =  takeLock();
        try {
            PropertyChangeEvent evt;

            params = resolveParams(params);
            
            evt = paramSupport.createChangeEvent(getElement(), 
                parameters == null ? NO_PARAMETERS : parameters , params);
            if (evt == null)
                return;
            
            checkVetoablePropertyChange(evt);
            getMethodBinding().changeParameters(params);
            fireOwnPropertyChange(evt);
            this.parameters = params;
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    private MethodParameter resolveParameter(MethodParameter par) {
        Type t = resolveType(par.getType());
        if (t == par.getType())
            return par;
        return new MethodParameter(
            par.getName(), t, par.isFinal());
    }
    
    private MethodParameter[] resolveParams(MethodParameter[] params) {
        if (!isConstrained())
            return params;
        
        int[] map = paramSupport.pairItems(this.parameters, params);
        MethodParameter[] newParams = null;
        
        for (int i = 0; i< map.length; i++) {
            if (map[i] == -1) {
                MethodParameter p = resolveParameter(params[i]);
                
                if (p == params[i])
                    continue;
                if (newParams == null) {
                    newParams = new MethodParameter[params.length];
                    System.arraycopy(params, 0, newParams, 0, params.length);
                }
                newParams[i] = p;
            }
        }
        return newParams == null ? params : newParams;
    }
    
    /**
     * Sets the list of exception declared to be thrown from within the method.
     * @param except array of exception type names.
     * @throw SourceException if a VetoableListener disagrees or other error occurs.
     */
    public void setExceptions(Identifier[] except) throws SourceException {
        // PENDING: translate exceptions to identifiers appropriate for this
        // context, or throw an excption.
        Object token = takeLock();
        try {
            except = resolveIdentifiers(except);
            
            PropertyChangeEvent evt;
            evt = exceptionSupport.createChangeEvent(getElement(), 
                exceptions == null ?
                IdentifierArrayProperty.EMPTY : this.exceptions, except);
            if (evt == null) 
                return;
            checkVetoablePropertyChange(evt);
            getMethodBinding().changeExceptions(except);
            fireOwnPropertyChange(evt);
            this.exceptions = except;
            commit();
        } finally {
            releaseLock(token);
        }
    }

    /**
     * Sets the text of the body to the passed value. Note, that PropertyChangeEvent
     * fired on behalf of this change may contain other value because the underlying
     * binding mechanism may change the exact layout of the body (no semantic changes
     * are permitted).
     * @throw SourceException if a VetoableListener disagrees or other error occurs.
     */
    public final void setBody(String body) throws SourceException {
        Object token = takeLock();
        try {
            String oldBody;
            PropertyChangeEvent evt;

            oldBody = getMethodBinding().getBodyContent();
            if (oldBody == body ||
                (oldBody != null && body != null && oldBody.equals(body)))
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_BODY, oldBody, body);
            checkVetoablePropertyChange(evt);
            if (body == null) {
                // make the method abstract (e.g. removecurly braces and add a semicolon)
                getMethodBinding().makeAbstract();
            } else if (oldBody == null) {
                // turn abstract method to non-abstract removing semicolon and adding
                // curly braces.
                getMethodBinding().createBody(body);
            } else {
                // simply replace the text between the curly braces.
                getMethodBinding().changeBody(body);
            }
            body = getBody();
            evt = new PropertyChangeEvent(getElement(), PROP_BODY, oldBody,
                body);
            addPropertyChange(evt);
            bodyHash = computeHash(body);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    // Private protocol
    /////////////////////////////////////////////////////////////////////////////////
    
    protected void initializeListenerSupport() {
        if (exceptionSupport == null) {
            synchronized (CallableImpl.class) {
                if (exceptionSupport == null)
                    exceptionSupport = new ExceptionSupport();
            }
        }
        if (paramSupport == null) {
            synchronized (CallableImpl.class) {
                if (paramSupport == null)
                    paramSupport = new MethodParamSupport();
            }
        }
    }

    /** Convenience method, casts getBinding() to Binding.Method
     */
    protected final Binding.Method getMethodBinding() {
        return (Binding.Method)getBinding();
    }
    
    // Support classes
    /////////////////////////////////////////////////////////////////////////////////
    
    private static final class MethodParamSupport extends FlyweightIndexedProperty {
        MethodParamSupport() {
            super(PROP_PARAMETERS);
        }
        
        protected final Object[] createEmpty() {
            return NO_PARAMETERS;
        }
        
        protected final Object[] createValue(int size) {
            if (size == 0)
                return createEmpty();
            return new MethodParameter[size];
        }
        
        protected boolean compareValues(Object a, Object b) {
            MethodParameter par1 = (MethodParameter)a;
            MethodParameter par2 = (MethodParameter)b;
            
            Type t1 = par1.getType();
            Type t2 = par2.getType();
            if (!par1.getName().equals(par2.getName()))
                return false;
            if (par1.isFinal() != par2.isFinal())
                return false;
            return compareSourceTypes(t1, t2);
        }
        
        protected final Object[] getValue(ElementImpl impl) {
            return ((CallableImpl)impl).parameters;
        }
    }
    
    private static final class ExceptionSupport extends IdentifierArrayProperty {
        ExceptionSupport() {
            super(PROP_EXCEPTIONS);
        }
        
        public Object[] getAsArray(ElementImpl bean) {
            return ((CallableImpl)bean).exceptions;
        }

        protected final Object[] getValue(ElementImpl impl) {
            return ((CallableImpl)impl).exceptions;
        }
        
        public boolean compareValues(Object one, Object two) {
            return compareSourceIdentifiers((Identifier)one, (Identifier)two);
        }
    }
    
    protected void copyCallableProperties(ConstructorElement el) {
        try {
            el.setName(getName());
            el.setModifiers(getModifiers());
            el.setParameters(getParameters());
            el.setExceptions(getExceptions());
        } catch (SourceException ex) {
        }
    }
}

