/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.*;
import java.util.*;

import org.openide.nodes.Node;
import org.openide.src.*;

/**
 * Models a class in a source - either a nested/local class, top-level class or an inner class.
 *
 * @author  Svatopluk Dedic
 * @version 0.1
 * @since 24/11/2000
 */
class ClassElementImpl extends MemberElementImpl 
implements ClassElement.Impl, ElementOrder {
    /** Uplink to the implementation of the source element. The link is made during
     * attaching the abstract Element to the implementation.
     */
    private SourceElementImpl       sourceImpl;
    
    /**
     * Reference to the enclosing element's implementation; the reference is identical
     * to getDeclaringImpl() if the class is an inner class, but it is different, if
     * the class is a nested or local class.
     */
    private ElementImpl             parentImpl;
    
    /**
     * Identifier of the superclass. Null, if the superclass was not specified.
     */
    private Identifier              superclass;
    
    /**
     * Identifiers of the superinterfaces; if no superinterfaces were specified, 
     * the variable contains an empty array (NO_INTERFACES).
     */
    private Identifier[]            superinterfaces;
    
    /**
     * True, if the element represents a class, false if it represents an interface.
     */
    private boolean                 classFlag;
    
    private FieldCollection         fields;
    
    private MethodCollection        methods;
    
    private ConstructorCollection   constructors;
    
    private PartialCollection       initializers;
    
    private InnerClassCollection    innerClasses;
    
    private MemberCollection        members;

    private static IdentifierArrayProperty    identSupport;
    
    private IdentContextSupport     identSupp;

    // Construction
    ///////////////////////////////////////////////////////////////////////////////////
    
    static {
        // initialize flyweight property support.
        identSupport = new InterfaceSupport();
    }

    /**
     * Constructs a class element from an external data source.
     * @param name name of the class. The name must contain both simple name and
     * fully qualified name parts.
     * @param modifiers initial modifiers for the class.
     * @param superclass declared superclass, null if none.
     * @param superintefaces interfaces implemented directly on the class; may be null,
     * if no interfaces are implemented.
     * @param isClass true, if the constructed item should denote a class. False if it
     * represents an inteface.
    ClassElementImpl(Identifier name, int modifiers, Identifier superclass,
        Identifier[] superinterfaces, boolean isClass) {
        super(name, modifiers);
        this.superclass = superclass;
        if (superinterfaces == null || superinterfaces.length == 0) {
            this.superinterfaces = IdentifierArrayProperty.EMPTY;
        } else {
            this.superinterfaces = superinterfaces;
        }
    }
     */
    
    ClassElementImpl(DefaultLangModel model) {
        super(model);
        identSupp = new IdentContextSupport(37);
        superinterfaces = IdentifierArrayProperty.EMPTY;
        javadoc = new JavaDocImpl.Class(null, this);
    }
    
    protected Identifier createName(String simpleName) {
        ClassElementImpl parentImpl;
        String outerName;
        
        parentImpl = getDeclaringImpl();
        if (parentImpl != null) {
            outerName = parentImpl.getName().getFullName();
        } else if (sourceImpl != null) {
            Identifier id = sourceImpl.getPackage();
            if (id != null)
                outerName = id.getSourceName();
            else
                outerName = ""; // NOI18N
        } else
            outerName = ""; // NOI18N
        StringBuffer sb = new StringBuffer();
        String full;
        
        if (!"".equals(outerName)) {
            sb.append(outerName);
            sb.append('.');
            sb.append(simpleName);
            full = sb.toString();
        } else 
            full = simpleName;
        
        return Identifier.create(full, simpleName);
    }
    
    protected void createFromModel(Element model) throws SourceException {
        ClassElement c = (ClassElement)model;

        // PENDING: set these directly.
        super.createFromModel(model);
        setClassOrInterface(c.isClassOrInterface());
        setSuperclass(c.getSuperclass());

        // member elements need the Element already.
        changeInterfaces(c.getInterfaces(), ClassElement.Impl.ADD);
        changeInitializers(c.getInitializers(), ClassElement.Impl.ADD);
        changeFields(c.getFields(), ClassElement.Impl.ADD);
        changeConstructors(c.getConstructors(), ClassElement.Impl.ADD);
        changeMethods(c.getMethods(), ClassElement.Impl.ADD);
        changeClasses(c.getClasses(), ClassElement.Impl.ADD);
        
        setJavaDocText(c.getJavaDoc().getRawText(), true);
    }
    
    public JavaDoc.Class getJavaDoc() {
        return (JavaDoc.Class)javadoc;
    }
    
    public final void setParent(ElementImpl impl) {
        super.setParent(impl);
        if (impl instanceof SourceElementImpl) {
            sourceImpl = (SourceElementImpl)impl;
        }
    }
    
    public void setBinding(Binding b) {
        super.setBinding(b);
        // initialize all members.
        members = new MemberCollection(this, (Binding.Container)b);
    }
        
    // Public operations
    ///////////////////////////////////////////////////////////////////////////////////
    /*
    public final Node.Cookie getCookie(Class clazz) {
        Node.Cookie ret = super.getCookie(clazz);
        if (ret != null || sourceImpl == null)
            return ret;
        return sourceImpl.getCookie(clazz);
    }
     */

    // Getters
    ///////////////////////////////////////////////////////////////////////////////////
    /**
     * Returns true, iff the class is a proper class. False, if the element represents
     * an interface.
     * @return true, if representing a class.
     */
    public boolean isClassOrInterface() {
        return this.classFlag;
    }
    
    public Identifier getSuperclass() {
        return this.superclass;
    }
    
    public Identifier[] getInterfaces() {
        if (superinterfaces == IdentifierArrayProperty.EMPTY)
            return superinterfaces;
        return (Identifier[])this.superinterfaces.clone();
    }
    
    // Setters/changers
    ///////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the desired superclass. The implementation will attempt to check the
     * passed identifier and will attempt to preserve the semantic meaning of the
     * Identifier when updating the model.
     * @param superclass the identifier of the new superclass.
     * @throw SourceException if the source cannot be locked, veto on the property change is
     * issued, the superclass' identifier is not reachable or there's some I/O or race error
     * during the operation.
     */
    public void setSuperclass(Identifier superclass) throws SourceException {
        // PENDING: resolve the Identifier if it is not this ClassElement-specific.
        // -- this should be a service "orthogonal" to the model management
        // and should be probably done in the representation layer.

        Object token = takeLock();
        try {
            superclass = resolveIdent(superclass);
            PropertyChangeEvent evt;
            if (superclass == this.superclass ||
            (superclass != null && this.superclass != null &&
            compareSourceIdentifiers(this.superclass, superclass)))
                // no change at all.
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_SUPERCLASS,
            this.superclass, superclass);
            checkVetoablePropertyChange(evt);
            getClassBinding().changeSuperclass(superclass);
            fireOwnPropertyChange(evt);
            this.superclass = superclass;
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    Identifier[] filterInterfaces(Identifier[] its) {
        if (its.length == 0)
            return its;
        Identifier[] oldIts = getInterfaces();
        Collection newState = null;
        for (int i = 0; i < its.length; i++) {
            boolean matches = false;
            for (int j = 0; j < oldIts.length; j++) {
                if (its[i].getSourceName().equals(oldIts[j].getSourceName())) {
                    matches = true;
                    break;
                }
            }
            if (!matches) {
                if (newState != null)
                    newState.add(its[i]);
            } else if (newState == null) {
                newState = new ArrayList(its.length);
                for (int k = 0; k < i; k++)
                    newState.add(its[k]);
            }
        }
        if (newState == null)
            return its;
        else
            return (Identifier[])newState.toArray(new Identifier[newState.size()]);
    }
    
    /**
     * Sets the interfaces implemented by the class.
     * @param interfaces collection of affected interfaces. If added, the values are
     * appended at the end of the interface property.
     * @param operation ADD, REMOVE or SET depending on the operation being performed.
     */
    public void changeInterfaces(Identifier[] interfaces, int operation) throws SourceException {
        Object token = takeLock();
        try {
            PropertyChangeEvent evt;

            interfaces = resolveIdentifiers(interfaces);
            Identifier[] newValue;

            // TODO: resolve interface identifiers, one by one, to preserve the semantic
            // meaning and create a suitable representation.
            switch (operation) {
                case ADD:
                    evt = identSupport.add(this, filterInterfaces(interfaces));
                    if (evt == null)
                        return;
                    newValue = (Identifier[])evt.getNewValue();
                    break;
                case REMOVE:
                    evt = identSupport.remove(this, interfaces);
                    if (evt == null)
                        return;
                    newValue = (Identifier[])evt.getNewValue();
                    break;
                // ugly generic case:
                case SET:
                    newValue = interfaces;
                    evt = identSupport.createChangeEvent(getElement(), 
                        superinterfaces == null ? IdentifierArrayProperty.EMPTY :
                        superinterfaces, interfaces);
                    if (evt == null)
                        return;
                    break;
                default:
                    throw new IllegalArgumentException("Unknown operation: " + operation);
            }
            
            checkVetoablePropertyChange(evt);
            getClassBinding().changeInterfaces(newValue);
            fireOwnPropertyChange(evt);
            superinterfaces = newValue;
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    /**
     * Converts a class into an interface or vice-versa.
     * @param classFlag true, if the element should turn itself into a class def,
     * false if it should become an interface.
     */
    public void setClassOrInterface(boolean classFlag) throws SourceException {
        Object token = takeLock();
        try {
            PropertyChangeEvent evt;

            if (this.classFlag == classFlag)
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_CLASS_OR_INTERFACE,
                classFlag ? Boolean.FALSE : Boolean.TRUE,
                classFlag ? Boolean.TRUE : Boolean.FALSE);

            // TODO: constraint the class -> interface change so that methods cannot
            // contain bodies.
            checkVetoablePropertyChange(evt);
            // perform the change in the source.
            getClassBinding().changeClassType(classFlag);
            fireOwnPropertyChange(evt);
            this.classFlag = classFlag;
            commit();
        } finally {
            releaseLock(token);
        }
    }

    // Member management methods
    // - will delegate to collection helpers.
    ///////////////////////////////////////////////////////////////////////////////////
    public FieldElement[] getFields() {
        if (fields == null)
            return FieldCollection.EMPTY;
        return (FieldElement[])fields.getElements().clone();
    }
    
    public FieldElement getField(Identifier name) {
        if (fields == null)
            return null;
        return fields.getField(name);
    }
    
    public void changeFields(FieldElement[] flds, int operation) 
        throws SourceException {
        initializeFields();
        
        Object token = takeMasterLock();
        try {
            fields.changeMembers(flds, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    public MethodElement[] getMethods() {
        if (methods == null)
            return MethodCollection.EMPTY;
        return (MethodElement[])methods.getElements().clone();
    }
    
    public MethodElement getMethod(Identifier id, Type[] argtypes) {
        if (methods == null)
            return null;
        return methods.getMethod(id, argtypes);
    }
    
    public void changeMethods(MethodElement[] m, int operation)
        throws SourceException {
        initializeMethods();
        Object token = takeMasterLock();
        try {
            methods.changeMembers(m, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    public ConstructorElement[] getConstructors() {
        if (constructors == null)
            return ConstructorCollection.EMPTY;
        return (ConstructorElement[])constructors.getElements().clone();
    }
    
    public ConstructorElement getConstructor(Type[] argtypes) {
        if (constructors == null)
            return null;
        return constructors.getConstructor(argtypes);
    }
    
    public void changeConstructors(ConstructorElement[] cons, int operation) 
        throws SourceException {
        initializeConstructors();
        Object token = takeMasterLock();
        try {
            constructors.changeMembers(cons, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    public ClassElement[] getClasses() {
        if (innerClasses == null)
            return InnerClassCollection.EMPTY;
        return (ClassElement[])innerClasses.getElements().clone();
    }
    
    public ClassElement getClass(Identifier id) {
        if (innerClasses == null)
            return null;
        return innerClasses.getInnerClass(id);
    }
    
    public void changeClasses(ClassElement[] classes, int operation)
        throws SourceException {
        Object token = takeMasterLock();
        try {
            initializeInnerClasses();
            innerClasses.changeMembers(classes, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    public InitializerElement[] getInitializers() {
        if (initializers == null) 
            return InitializerCollection.EMPTY;
        else
            return (InitializerElement[])initializers.getElements().clone();
    }
    
    public void changeInitializers(InitializerElement[] el, int operation) 
        throws SourceException {
        Object token = takeMasterLock();
        try {
            initializeInitializers();
            initializers.changeMembers(el, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    // Local element management
    ///////////////////////////////////////////////////////////////////////////////////
    
    /**
     * Local classes of a class are its inner classes.
     */
    public ClassElement[] getLocalClasses() {
        return getClasses();
    }
    
    // Utility methods
    ///////////////////////////////////////////////////////////////////////////////////
    
    /** Return the implementation for the SourceElement - directly, since all classes
     * have a link to the source.
     */
    protected final SourceElementImpl findSourceImpl() {
        return this.sourceImpl;
    }
    
    protected final Binding.Class getClassBinding() {
        return (Binding.Class)getBinding();
    }
    
    public void updateMembers(String propName, Element[] els, int[] indices,
        int[] optMap) {
        if (propName == ElementProperties.PROP_INITIALIZERS) {
            initializeInitializers();
            initializers.updateMembers(els, indices, optMap);
        } else if (propName == ElementProperties.PROP_FIELDS) {
            initializeFields();
            fields.updateMembers(els, indices, optMap);
        } else if (propName == ElementProperties.PROP_CONSTRUCTORS) {
            initializeConstructors();
            constructors.updateMembers(els, indices, optMap);
        } else if (propName == ElementProperties.PROP_METHODS) {
            initializeMethods();
            methods.updateMembers(els, indices, optMap);
        } else if (propName == ElementProperties.PROP_CLASSES) {
            initializeInnerClasses();
            innerClasses.updateMembers(els, indices, optMap);
        }    
    }
    
    public void updateMemberOrder(Element[] ordered) {
        members.updateOrder(ordered);
    }
    
    private void initializeFields() {
        if (fields != null) 
            return;
        synchronized (this) {
            if (fields == null) {
                fields = new FieldCollection(this, getModelImpl(), members);
            }
        }
    }
    
    private void initializeMethods() {
        if (methods != null) 
            return;
        synchronized (this) {
            if (methods == null) {
                methods = new MethodCollection(this, getModelImpl(), members);
            }
        }
    }
                
    private void initializeConstructors() {
        if (constructors != null)
            return;
        synchronized (this) {
            if (constructors == null) {
                constructors = new ConstructorCollection(this, getModelImpl(), members);
            }
        }
    }
    
    private void initializeInitializers() {
        if (initializers != null)
            return;
        synchronized (this) {
            if (initializers == null) {
                initializers = new InitializerCollection(this, getModelImpl(), members);
            }
        }
    }
    
    private void initializeInnerClasses() {
        if (innerClasses != null)
            return;
        synchronized (this) {
            if (innerClasses == null) {
                innerClasses = new InnerClassCollection(this, getModelImpl(), members);
            }
        }
    }
    
    protected final Binding createBinding(Element el) {
        return getModelImpl().getBindingFactory().bindClass((ClassElement)el);
    }
    
    protected final void createAfter(Binding.Container cb, Binding refBinding)
    throws SourceException {
        Element[] els = members.getElements();
        ElementImpl impl;
        Binding ref = null;
        
        for (int i = 0; i < els.length; i++) {
            impl = members.getElementImpl(els[i]);
            impl.createAfter(getClassBinding(), ref);
            ref = impl.getBinding();
        }
    }
    

    /*
    // PENDING: this method is probably not used any more.
    protected final ElementImpl adoptElement(ElementImpl impl) {
        MemberElementImpl mimpl = (MemberElementImpl)impl;
        mimpl.setParent(this);
        return mimpl;
    }
     */
    
    /** Returns a collection of member implementations.
    public Collection getMemberImpls() {
        initializeMembers();
        return members.getElements();
    }
     */

    /*
    /* package private MemberCollection getAllMembers() {
        return null;
    }
     */
    
    // Serialization support
    ///////////////////////////////////////////////////////////////////////////////////
    public Object readResolve() {
        return null;
    }
    
    // Support classes
    ///////////////////////////////////////////////////////////////////////////////////
    private static final class InterfaceSupport extends IdentifierArrayProperty {
        InterfaceSupport() {
            super(PROP_INTERFACES);
        }
        
        public Object[] getValue(ElementImpl bean) {
            Object[] v = ((ClassElementImpl)bean).superinterfaces;
            if (v == null)
                return IdentifierArrayProperty.EMPTY;
            return v;
        }
        
        public PropertyChangeEvent add(ClassElementImpl beanImpl, Identifier[] adding) 
            throws SourceException {
            Identifier[] cur = (Identifier[])getValue(beanImpl);
            Collection remains = null;
            
            for (int i = 0; i < adding.length; i++) {
                for (int j = 0; j < cur.length; j++) {
                    if (adding[i].getSourceName().equals(cur[j].getSourceName())) {
                        if (remains == null)
                            remains = new ArrayList(adding.length - 1);
                        for (int k = 0; k < i; k++) {
                            remains.add(adding[k]);
                        }
                        break;
                    }
                }
            }
            if (remains == null) {
                return super.add(beanImpl, adding);
            } else {
                // PENDING: throw an exception under different mode ??
                return super.add(beanImpl, remains.toArray(new Identifier[remains.size()]));
            }
        }
        
        public boolean compareValues(Object o1, Object o2) {
            return compareSourceIdentifiers((Identifier)o1, (Identifier)o2);
        }
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer();
        
        sb.append("ClassElement[");
        sb.append(getName().getFullName());
        sb.append("/");
        sb.append(getName().getSourceName());
        sb.append("]");
        return sb.toString();
    }
    
    /**
     * This implementation only clones the class element itself, not its subitems.
     */
    protected Element cloneSelf() {
        ClassElement clone = new ClassElement();
        try {
            clone.setClassOrInterface(isClassOrInterface());
            clone.setName(getName());
            clone.setSuperclass(getSuperclass());
            clone.setInterfaces(getInterfaces());
            clone.setModifiers(getModifiers());
        } catch (SourceException ex) {
        }
        return clone;
    }
    
    protected void doSetName(Identifier name) throws SourceException {
        // check whether
        // a) the name is only simple.
        // b) the name is consistent with the outer's name:
        String newFull = name.getFullName();
        if (this.name != null) {
            if  (newFull.equals(name.getSourceName())) {
                String qual = getName().getQualifier();
                if (qual != null && !"".equals(qual)) {
                    StringBuffer sb = new StringBuffer();
                    sb.append(qual);
                    sb.append('.');
                    sb.append(newFull);
                    name = Identifier.create(sb.toString(), newFull);
                    newFull = name.getFullName();
                }
            } else if (this.name != null &&
                !name.getQualifier().equals(getName().getQualifier())) {
                throw new SourceException("Invalid full name " + name.getFullName() + "/" + name.getSourceName() + " my name = "
                    + getName().getFullName() + "/" + getName().getSourceName());
            }
        }
        super.doSetName(name);
        updateInnerClasses(newFull);
    }
    
    protected void changeName(Identifier id) throws SourceException {
        ConstructorElement[] cons = getConstructors();
        int lastChanged = 0;
        Identifier oldName = getName();
        
        try {
            super.changeName(id);
            
            // temporary to make constructor checking code happy.
            name = id;
            
            Identifier newConName = Identifier.create(id.getName());
            
            for (; lastChanged < cons.length; lastChanged++) {
                ConstructorElement el = cons[lastChanged];
                el.setName(newConName);
            }
        } catch (SourceException ex) {
            // try to revert back anything done so far.
            super.changeName(oldName);
            // revert to the old name. This will be done once more in finally
            // but never mind.
            name = oldName;

            Identifier oldConName = Identifier.create(oldName.getName());
            for (int i = 0; i < lastChanged; i++) {
                try {
                    cons[i].setName(oldConName);
                } catch (SourceException ex2) {
                    // what should we do now ??
                }
            }
            // rethrow exception.
            throw ex;
        } finally {
            // clean up temp detritus
            name = oldName;
        }
    }
    
    /**
     * Called by outer classes during their rename operation. The inner classes
     * should fire Property Change on their name property, but without Vetoable change.
     * And, of course, change their names.
     */
    protected void updateName(String outerPrefix) {
        Identifier curName = getName();
        if (curName.getQualifier().equals(outerPrefix))
            return;
        StringBuffer sb = new StringBuffer();
        if (outerPrefix != null && !"".equals(outerPrefix)) {
            sb.append(outerPrefix);
            sb.append('.');
        }
        String simpleName = curName.getName();
        sb.append(simpleName);
        String fullName = sb.toString();
        Identifier newName = Identifier.create(fullName, simpleName);
        fireOwnPropertyChange(new PropertyChangeEvent(getElement(),
            PROP_NAME, curName, newName));
        this.name = newName;
        updateInnerClasses(fullName);
    }
    
    private void updateInnerClasses(String fullName) {
        ClassElement[] cls = getClasses();
        for (int i = 0; i < cls.length; i++) {
            ClassElementImpl impl = (ClassElementImpl)cls[i].getCookie(ClassElementImpl.class);
            impl.updateName(fullName);
        }
    }
    
    protected Element getParent() {
        Element p = super.getParent();
        if (p == null)
            return ((ClassElement)getElement()).getSource();
        else 
            return p;
    }
    
    protected Identifier createLocalIdentifier(Identifier id, int status) {
        return identSupp.create(id, status);
    }
    
    protected boolean checkIdentifierContext(Identifier id) {
        return identSupp.checkContext(id);
    }
    
    protected void checkRemove() throws SourceException {
        super.checkRemove();
        Element[] allElems = members.getElements();
        for (int i = 0; i < allElems.length; i++) {
            ElementImpl impl = members.getElementImpl(allElems[i]);
            impl.checkRemove();
        }
    }
    
    protected void notifyRemove() {
        super.notifyRemove();
        Element[] allElems = members.getElements();
        for (int i = 0; i < allElems.length; i++) {
            ElementImpl impl = members.getElementImpl(allElems[i]);
            impl.notifyRemove();
        }
    }
    
    protected void notifyCreate() {
        Element[] allElems = members.getElements();
        for (int i = 0; i < allElems.length; i++) {
            ElementImpl impl = members.getElementImpl(allElems[i]);
            impl.notifyCreate();
        }
        super.notifyCreate();
        members.sanityCheck();
        if (innerClasses != null)
            innerClasses.sanityCheck();
        if (fields != null)
            fields.sanityCheck();
        if (methods != null)
            methods.sanityCheck();
        if (constructors != null)
            constructors.sanityCheck();
        if (initializers != null)
            initializers.sanityCheck();
    }
    
    protected boolean parentValid() {
        return declaringClassImpl == null ? 
            sourceImpl.isValid() : declaringClassImpl.isValid();
    }
    
    public Element[] getElements() {
        return members.getElements();
    }
}
