/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.*;

import org.openide.nodes.Node;
import org.openide.src.*;

/**
 * Implements properties specific to a Field declared in a class.
 * The field has a <B>type</B> and optionally an <b>initial value</B>
 *
 * @author  Svatopluk Dedic, Petr Hamernik, Jaroslav Tulach
 * @version 0.1
 * @since 24/11/2000
 */
final class FieldElementImpl extends MemberElementImpl implements FieldElement.Impl {
    /** Type of the field. */
    private Type        type;
    
    /** Initializer of the field, or null if none. 
     * PENDING -- should the initValue really be here ? Shouldn't it be taken from
     * the source code just like other unparsed strings in the model ?
     */
    private String      initValue;

    /*
    public FieldElementImpl(Identifier name, int modifiers, Type t, String init) {
        super(name, modifiers);
        this.type = t;
        this.initValue = init;
    }
    */
    
    FieldElementImpl(DefaultLangModel model) {
        super(model);
        initValue = "";
        javadoc = new JavaDocImpl.Field(null, this);
    }
    
    protected Binding createBinding(Element el) {
        return getModelImpl().getBindingFactory().bindField((FieldElement)el);
    }

    protected void createFromModel(Element el) throws SourceException {
        super.createFromModel(el);
        FieldElement f = (FieldElement)el;
        setType(f.getType());
        setInitValue(f.getInitValue());
        setJavaDocText(f.getJavaDoc().getRawText(), true);
    }

    /** Type of the variable.
    * @return the type
    */
    public Type getType() {
        return this.type;
    }
    
    public JavaDoc.Field getJavaDoc() {
        return (JavaDoc.Field)javadoc;
    }
    
    public Object readResolve() {
        return null;
    }
    
    /** Setter for type of the variable.
    * @param type the variable type
    */
    public void setType(Type type) throws SourceException {
        Object token = takeLock();
        try {
            PropertyChangeEvent evt;

            if (this.type == type)
                return;

            type = resolveType(type);
            if (compareSourceTypes(this.type, type))
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_TYPE, this.type, type);
            checkVetoablePropertyChange(evt);
            getFieldBinding().changeType(type);
            fireOwnPropertyChange(evt);
            this.type = type;
            commit();
        } finally {
            releaseLock(token);
        }
    }

    /** Getter for the initial value.
    * @return initial value for the variable or empty string if it is not initialized
    */
    public String getInitValue() {
        return this.initValue;
    }

    /** Setter for the initial value.
    * @param value initial value for the variable
    */
    public void setInitValue(String initValue) throws SourceException {
        Object token = takeLock();
        if (initValue == null)
            initValue = "";
        try {
            PropertyChangeEvent evt;
            if ((initValue == this.initValue) ||
                (initValue != null && initValue.equals(this.initValue)))
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_INIT_VALUE, this.initValue, initValue);
            // no constraings on the Initializer... only check vetoable listeners.
            checkVetoablePropertyChange(evt);
            getFieldBinding().changeInitializer(initValue);
            this.initValue = initValue;
            addPropertyChange(evt);
            commit();
        } finally {
            releaseLock(token);
        }
    }

    // Utility methods.
    //////////////////////////////////////////////////////////////////////////////////
    
    protected Binding.Field getFieldBinding() {
        return (Binding.Field)getBinding();
    }
    
    protected Element createWrapperElement() {
        return null;
    }
    
    public String toString() {
        return "FieldElementImpl[" + getName().getSourceName() + "]";
    }
    
    protected Element cloneSelf() {
        return (Element)((FieldElement)getElement()).clone();
    }
}
