/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.PropertyChangeEvent;
import java.util.*;

import org.openide.src.*;

/**
 *
 * @author  sdedic
 * @version
 */
class ImportCollection extends ObjectCollection {
    Import[]                importCache;
    
    static final ImportElement[] NO_IMPORT_ELEMENTS = new ImportElement[0];
    static final Import[] NO_IMPORTS = new Import[0];
    
    public ImportCollection(Binding.Container binding,
        ElementCreator creator, ElementEvents events) {
        super(binding, creator, events, ElementProperties.PROP_IMPORTS);
    }
    
    protected ElementImpl createElement(Element parent) {
        return null;
    }

    protected Element copyElement(Element el) {
        return el;
    }
    
    private ImportElement createElement(Import im) throws SourceException {
        ImportImpl iimpl = creator.createImport((SourceElement)events.getEventSource());
        iimpl.setImport(im);
        return (ImportElement)iimpl.getElement();
    }
    
    public Element[] createEmpty(int size) {
        return size == 0 ?
            NO_IMPORT_ELEMENTS : new ImportElement[size];
    }
    
    protected void fireVetoableChange(PropertyChangeEvent e) throws SourceException {
    }
    
    public void importsChanged() {
        importCache = null;
    }
    
    protected void doSetMembers(Element[] newElements, Change chng)
    throws SourceException {
        super.doSetMembers(newElements, chng);
        importCache = null;
    }
    
    private void addImports(Import[] els) throws SourceException {
        Collection unique = new ArrayList(els.length);
        ImportElement[] imps = (ImportElement[])getElements();
        
        for (int i = 0; i < els.length; i++) {
            boolean matches = false;
            
            for (int j = 0; j < imps.length; j++) {
                if (els[i].equals(imps[j].getImport())) {
                    matches = true;
                    break;
                }
            }
            if (!matches)
                unique.add(els[i]);
        }
        if (unique.isEmpty()) {
            // no change ever needed.
            return;
        }
        Element[] importEls = new ImportElement[unique.size()];
        
        Iterator it = unique.iterator();
        for (int i = 0; it.hasNext(); i++) {
            importEls[i] = createElement((Import)it.next());
        }
        super.addMembers(importEls);
    }
    
    private void removeImports(Import[] els) throws SourceException {
        Collection toRemove = new ArrayList(els.length);
        ImportElement[] ims = (ImportElement[])getElements();
        BitSet used = new BitSet(ims.length);
        
        for (int i = 0; i < els.length; i++) {
            for (int j = 0; j < ims.length; j++) {
                if (used.get(j))
                    continue;
                if (ims[j].getImport().equals(els[i])) {
                    used.set(j);
                    toRemove.add(ims[j]);
                }
            }
        }
        if (toRemove.isEmpty())
            // no change needed.
            return;
        Element[] ee = (Element[])toRemove.toArray(createEmpty(0));
        super.removeMembers(ee);
    }
    
    public void updateMembers(Element[] members, int[] map) {
        super.updateMembers(members, map);
        importCache = null;
    }
    
    private void setImports(Import[] imports) throws SourceException {
        Element[] members = getElements();
        int[] map = pairItems(members, imports);
        Element[] newEls = new ImportElement[imports.length];
        
        for (int i = 0; i < map.length; i++) {
            if (map[i] == -1) {
                newEls[i] = createElement(imports[i]);
            } else {
                newEls[i] = members[map[i]];
            }
        }
        
        Change chng = computeChanges2(members, newEls, map);
        if (chng.phaseCount == 0)
            return;
        
        doSetMembers(newEls, chng);
    }

    protected boolean compareValues(Object o1, Object o2) {
        if (!(o1 instanceof ImportElement) || !(o2 instanceof Import))
            return false;
        Import i2 = (Import)o2;
        ImportElement i1 = (ImportElement)o1;
        
        return i1.getImport().equals(i2);
    }
    
    public void changeMembers(Import[] elements, int operation) throws SourceException {
        switch (operation) {
            case ClassElement.Impl.ADD:
                addImports(elements);
                break;
            case ClassElement.Impl.REMOVE:
                removeImports(elements);
                break;
            case ClassElement.Impl.SET:
                setImports(elements);
        }
    }
    
    /** This implementation will cowardly add imports to the end of the
     * collection unconditionally.
     */
    public Element[] findPositions(Element[] elements) {
        Element[] markers = new Element[elements.length];
        Element[] cur = getElements();

        // insert after the current import statements, if any.
        if (cur.length > 0)
            markers[0] = cur[cur.length - 1];
        else
            markers[0] = Positioner.FIRST;
        // chain other elements after the first one whenever it gets
        // inserted.
        for (int i = 1; i < markers.length; i++) {
            markers[i] = elements[i - 1];
        }
        return markers;
    }
    
    public Import[] getImports() {
        Import[] cached = importCache;
        if (cached != null)
            return cached;
        synchronized (this) {
            cached = importCache;
            if (cached != null)
                return cached;
            ImportElement[] elems = (ImportElement[])getElements();
            Import[] imp = new Import[elems.length];
            for (int i = 0; i < elems.length; i++)
                imp[i] = elems[i].getImport();
            return importCache = imp;
        }
    }
    
    protected MultiPropertyChangeEvent createPropertyEvent(Change chng,
        Object source, Object[] elems, Object[] newV) {
        Import[] oldImports = getImports();
        Import[] newImports = new Import[newV.length];
        chng = (Change)chng.clone();
        
        for (int i = 0; i < newImports.length; i++) {
            newImports[i] = ((ImportElement)newV[i]).getImport();
        }
        mapImports(chng.inserted);
        mapImports(chng.removed);

        return super.createPropertyEvent(chng, source, oldImports, newImports);
    }
    
    private Collection mapImports(Collection o) {
        if (o == null)
            return null;
        Collection c = new ArrayList(o.size());
        for (Iterator it = o.iterator(); it.hasNext(); ) {
            c.add(((ImportElement)it.next()).getImport());
        }
        return c;
    }
    
}
