/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.lang.reflect.Modifier;
import java.beans.*;

import org.openide.src.*;

/**
 * The class implements property handling for a class' initializer. In fact, the initializer
 * has only two properties: the body and the optional static modifier.
 *
 * @author  Svatopluk Dedic, Petr Hamernik
 * @version 0.1
 * @since 24/11/2000
 */
class InitializerElementImpl extends MemberElementImpl 
    implements InitializerElement.Impl, Element.Impl2 {

    private transient String    copiedBody;
    
    private transient long      bodyHash;
    
    InitializerElementImpl(DefaultLangModel model) {
        super(model);
    }

    protected void createFromModel(Element el) throws SourceException {
        InitializerElement i = (InitializerElement)el;
        copyBody(i.getBody());
        setStatic(i.isStatic());
    }
    
    protected void notifyCreate() {
        super.notifyCreate();
        // recompute the hash using the exact representation (potentially reformatted)
        // in the source text.
        if (copiedBody != null)
            bodyHash = computeHash(getBody());
        copiedBody = null;
    }
    
    private long computeHash(String bodyContent) {
        java.util.zip.CRC32 crc = new java.util.zip.CRC32();
        crc.update(bodyContent.getBytes());
        return crc.getValue();
    }
    
    protected final Binding createBinding(Element el) {
        return getModelImpl().getBindingFactory().bindInitializer((InitializerElement)el);
    }

    public void attachedToElement(Element el) {
        super.attachedToElement(el);
    }

    public boolean isStatic() {
        return (getModifiers() & Modifier.STATIC) > 0;
    }
    
    void copyBody(String s) {
        if (s == null)
            s = ""; // NOI18N
        this.copiedBody = s;
    }
    
    /** Calls the ElementOutput's implementation to retrieve the contents of the body.
     * This is a little hack, since the body *should* be a part of the model,
     * but I wanted to avoid duplicating of (potentially large) text data.
     * @return text of the body.
     */
    public String getBody() {
        if (isCreated())
            return copiedBody;
        try {
            return getInitializerBinding().getBodyContent();
        } catch (SourceException ex) {
            // TODO: Report the exception, the results are *incorrect* !
            return null;
        }
    }
    
    public JavaDoc getJavaDoc() {
        return null;
    }

    public boolean isValid() {
        return super.isValid() && (getDeclaringImpl() == null ||
            getDeclaringImpl().isValid());
    }
    
    void updateBody(String content) {
        if (content == null) {
            bodyHash = -1;
            return;
        }
        long newHash = computeHash(content);
        if (newHash == bodyHash) 
            return;
        bodyHash = newHash;
        // we don't have info about the previous body's contents.
        addPropertyChange(new PropertyChangeEvent(getElement(), PROP_BODY, null, null));
    }
    
    // Change operations.
    ///////////////////////////////////////////////////////////////////////////////
    public void setStatic(boolean enable) throws SourceException {
        Object token = takeLock();
        
        try {
            PropertyChangeEvent evt;

            if (((mod & Modifier.STATIC) > 0) == enable)
                return;
            evt = new PropertyChangeEvent(getElement(), PROP_STATIC, enable ? Boolean.FALSE : Boolean.TRUE, 
                enable ? Boolean.TRUE : Boolean.FALSE);

            checkVetoablePropertyChange(evt);
            getInitializerBinding().changeStatic(enable);
            mod = enable ? Modifier.STATIC : 0;
            fireOwnPropertyChange(evt);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    /** Sets the body. The entire functionality is delegated to the storage binding object.
     * Note that the text sent out in the PropertyChangeEvent may NOT match the text
     * passed in the vetoable change event since the text may be transformed by an 
     * indentation engine.
     * @param body text of the initializer's body.
     */
    public void setBody(String body) throws SourceException {
        Object token = takeLock();
        
        try {
            PropertyChangeEvent evt;
            
            String oldBody = getInitializerBinding().getBodyContent();
            if (oldBody == body ||
                (oldBody != null && body != null && oldBody.equals(body))) {
                // No change
                // TODO: ask the indentation engine whether the body is really different
                // --> ignore whitespace etc.
                return;
            }
            evt = new PropertyChangeEvent(getElement(), PROP_BODY, oldBody, body);
            checkVetoablePropertyChange(evt);
            getInitializerBinding().changeBody(body);
            body = getBody();
            evt = new PropertyChangeEvent(getElement(), PROP_BODY, oldBody,
                body);
            addPropertyChange(evt);
            bodyHash = computeHash(body);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    // Utility methods.
    //////////////////////////////////////////////////////////////////////////////////
    protected final Binding.Initializer getInitializerBinding() {
        return (Binding.Initializer)getBinding();
    }
    
    protected final SourceElementImpl findSource() {
        return getDeclaringImpl().findSource();
    }
    
    protected Element cloneSelf() {
        InitializerElement el = new InitializerElement();
        try {
            el.setStatic(isStatic());
        } catch (SourceException ex) {
        }
        return el;
    }
    
    // Serialization support
    //////////////////////////////////////////////////////////////////////////////////
    public Object readResolve() {
        return null;
    }
}
