/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.PropertyChangeEvent;
import java.lang.reflect.Modifier;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import org.openide.nodes.Node;
import org.openide.src.*;

/** Describes a main elements of java source
 * (variables, methods and classes). Provides support
 * for associating this element with declaring class. Adds `name' and `modifiers' properties
 * implementation for descendant elements as well as some utility methods for type and
 * identifier comparison that can be utilized in change detection.
 *
 * @author Svatopluk Dedic, Petr Hamernik, Jaroslav Tulach
 * @version 0.1
 * @since 24/11/2000
 */
abstract class MemberElementImpl extends ElementImpl {
    /** Modifiers for this element */
    protected int         mod;

    /** Name of this element. This should be a STRING, but - because of old sins in
     * org.openide.src design - it is an Identifier.
     */
    protected Identifier  name;

    /** Implementation of the declaring class. It is just a shorthand for acquiring
     * a cookie from the ClassElement.
     */
    protected ClassElementImpl    declaringClassImpl;
    
    static final long serialVersionUID =6388377681336329844L;
    
    // Construction
    ///////////////////////////////////////////////////////////////////////////////////

    /** Construction of MemberElement from externally supplied data.
     * @param name the name of the Element.
     * @param modifiers initial modifiers of the element.
    protected MemberElementImpl(Identifier name, int modifiers) {
        super(null);
        this.name = name;
        this.mod = modifiers;
    }
     */
    
    protected MemberElementImpl(DefaultLangModel model) {
        super(model);
    }
    
    // Public operations
    ///////////////////////////////////////////////////////////////////////////////////

    protected void setParent(ElementImpl impl) {
        if (impl instanceof ClassElementImpl) {
            this.declaringClassImpl = (ClassElementImpl)impl;
        }
    }
    
    protected boolean parentValid() {
        return declaringClassImpl != null && declaringClassImpl.isValid();
    }

    protected void createFromModel(Element model) throws SourceException {
        MemberElement other = (MemberElement)model;
        setModifiers(other.getModifiers());
        setName(createName(other.getName().getName()));
    }
    
    protected Identifier createName(String n) {
        return Identifier.create(n);
    }
    
        /*
    public Node.Cookie getCookie(Class clazz) {
        Node.Cookie ret = super.getCookie(clazz);
        if (ret != null)
            return ret;
        ClassElementImpl climpl = getDeclaringImpl();
        if (climpl != null)
            return climpl.getCookie(clazz);
        return null;
    }
         */

    // Getters
    ///////////////////////////////////////////////////////////////////////////////////

    /** Getter for modifiers for this element.
    * @see java.lang.reflect.Modifier
    * @return constants from java.lang.reflect.Modifier
    */
    public int getModifiers() {
        return mod;
    }

    /** Getter for name of the field.
    * @return the name
    */
    public Identifier getName() {
        return name;
    }

    /**
     * Extends the basic isValid with a check, whether the ClassElement itself is valid.
     * @return true, if both the element and its decl. class are valid.
     */
    public boolean isValid() {
        if (!super.isValid())
            return false;
        
        ClassElementImpl impl = getDeclaringImpl();
        return impl == null || impl.isValid();
    }
    
    /**
     * Extracts and retrieves local classes for the element.
     * @return array of elements representing local classes.
     */
    public ClassElement[] getLocalClasses() {
        // TODO: implement fetching1 of local classes based on data extracting via
        // model and parsing support.
        return new ClassElement[] {};
    }
    
    // Setters/changers
    ///////////////////////////////////////////////////////////////////////////////////
    /**
     * Overriden in subsclasses so that it check model constraints on modifier value.
     * The default implementation does nothing.
     */
    protected void checkModifierConstraints(int newMods) throws SourceException {
    }

    /** Overload to impose constraint over the member's name. The default implementation
     * does nothing.
     */
    protected void checkNameConstraints(Identifier name) throws SourceException {
    }

    /** Retrieves the binding object.
     */
    protected final Binding.Member getMemberBinding() {
        return (Binding.Member)getBinding();
    }

    /**
     * Resolves, or tries to resolve the type.
     */
    protected final Type resolveType(Type t) {
        if (!isConstrained())
            return t;
        
        if (t.isPrimitive())
            return t;
        return getModelImpl().resolveType(getParent(), t);
    }
    
    protected final Identifier resolveIdent(Identifier i) {
        if (i == null || !isConstrained())
            return i;
        return getModelImpl().resolveIdent(getParent(), i);
    }
    
    protected final Identifier[] resolveIdentifiers(Identifier[] ids) {
        if (!isConstrained())
            return ids;
        
        Identifier[] newIds = null;
        for (int i = 0; i < ids.length; i++) {
            Identifier n = resolveIdent(ids[i]);
            if (n == ids[i])
                continue;
            if (newIds == null) {
                newIds = new Identifier[ids.length];
                System.arraycopy(ids, 0, newIds, 0, ids.length);
            }
            newIds[i] = n;
        }
        if (newIds == null)
            return ids;
        return newIds;
    }
    
    protected Element getParent() {
        return getDeclaringClass();
    }
    
    /** Setter for modifiers for this element.
    * @see java.lang.reflect.Modifier
    * @param mod constants from java.lang.reflect.Modifier
    */
    public void setModifiers(int mod) throws SourceException {
        Object token = takeLock();
        try {
            int old = this.mod;
            if (old == mod)
                return;
            if (isConstrained())
                checkModifierConstraints(mod);
            PropertyChangeEvent evt = new PropertyChangeEvent(getElement(), PROP_MODIFIERS, new Integer(old), new Integer(mod));
            checkVetoablePropertyChange(evt);
            getMemberBinding().changeModifiers(mod);
            fireOwnPropertyChange(evt);
            this.mod = mod;
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    protected void doSetName(Identifier id) throws SourceException {
        PropertyChangeEvent evt;
        Identifier old = this.name;
        if (isConstrained())
            checkNameConstraints(id);
        if (old != null &&
            old.getSourceName() == id.getSourceName())
            return;
        evt = new PropertyChangeEvent(getElement(), PROP_NAME, old, id);
        checkVetoablePropertyChange(evt);
        changeName(id);
        fireOwnPropertyChange(evt);
        this.name = id;
    }

    /**
     * Actually handles out the name change. It is defined as a separate method
     * mainly because ClassElement whose ConstructorElement names depend on
     * their class name
     */
    protected void changeName(Identifier id) throws SourceException {
        getMemberBinding().changeName(id);
    }
    
    /**
     * Implementation of name change for an Element. The implementation will:<UL>
     * <LI>check for model constraints on the name
     * <LI>check vetoable listener acknowledges of the change
     * <LI>ask the Binding for performing the change
     * <LI>fire the PropertyChangedEvent.
     * @param id new name to use with the element.
     */
    public void setName(Identifier id) throws SourceException {
        Object token = takeLock();
        try {
            doSetName(id);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    // Utility methods
    ///////////////////////////////////////////////////////////////////////////////////

    /** Compares source names of the identifiers; it does not pay attention to full names
        specified in the identifier object.
    */
    protected static boolean compareSourceIdentifiers(Identifier oldId, Identifier newId) {
        if (oldId == newId)
            return true;
        if (oldId == null ||
            newId == null)
            return false;
        if (!oldId.getSourceName().equals(newId.getSourceName())) {
            return false;
        }
        int oldRes = oldId.getResolutionStatus();
        int newRes = newId.getResolutionStatus();
        if (oldRes != newRes) {
            if (newRes != Identifier.NOT_YET_RESOLVED)
                return false;
        }
        return oldId.getFullName().equals(newId.getFullName());
    }

    /** Compares types not paying attention to fully qualified names of class types.
    */
    protected static boolean compareSourceTypes(Type oldType, Type newType) {
        if (oldType == newType) {
            return true;
        }
        // no type was ever present ;-)
        if (oldType == null)
            return false;
        // if one of the types is a primitive one, they must be the same instance to match.
        if (oldType.isPrimitive() || newType.isPrimitive()) {
            return false;
        }
        if (oldType.isArray()) {
            if (!newType.isArray()) {
                return false;
            }
            return compareSourceTypes(oldType.getElementType(), newType.getElementType());
        } else if (newType.isArray()) {
            return false;
        }
        if (!oldType.isClass() || !newType.isClass()) {
            throw new InternalError("Unexpected type combination.");
        }
        return compareSourceIdentifiers(oldType.getTypeIdentifier(),
            newType.getTypeIdentifier());
    }

    /**
     * Members source finder delegates to the class' one.
     */
    protected SourceElementImpl findSource() {
        return getDeclaringImpl().findSourceImpl();
    }

    /**
     * Returns the implementation of the declaring class.
     */
    protected ClassElementImpl getDeclaringImpl() {
        return this.declaringClassImpl;
    }
    
    /** Returns the abstract layer for the declaring class implementation.
     */
    protected ClassElement getDeclaringClass() {
        if (declaringClassImpl == null)
            return null;
        return (ClassElement)declaringClassImpl.getElement();
    }
    
    /**
     * Access method to the Element's isValid, may be used by subclasses for overriding
     * isValid semantics.
     */
    protected boolean isElementValid() {
        return super.isValid();
    }
    
    // Support classes
    ///////////////////////////////////////////////////////////////////////////////////
}
