/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.util.*;
import javax.swing.undo.UndoableEdit;

import org.openide.src.*;

/**
 * Implementation class that holds elements of a single type and maps them to a collection
 * of all class members. The class manages operations on the partial collection and maps
 * them to operations on the whole collection. Mechanisms inherited from IndexedPropertyBase
 * allows detection of changes.
 *
 * This collection works in conjunction with {@link #MemberCollection}
 *
 * @author  Svatopluk Dedic (sdedic@netbeans.org)
 * @version 0.1
 */
abstract class PartialCollection extends IndexedPropertyBase {
    /**
     * Threshold for dump identity comparisons. If the avg number of compare ops *2 is
     * more than this value, faster hashing algorithm is used for discovering identities.
     * The value should balance the tradeoff between creating new objects and the comparisons.
     */
    public static final int SMALL_THRESHOLD = 500;

    /** Interface that performs event firing and provides basic information for
     * events (e.g. source of the event).
     */
    ElementEvents   events;
    
    /** Holds positions of individual elements in the ordered collection of all
     * class members.
     */
    private int[]   positions;
    
    /** Cached array of references to the wrapper Elements. This array is invalidated
     * after each modifying operation and constructed whenever some code asks
     * for 
     */
    Element[]  elements;
    
    /** Reference to a collection that contains all members of the declaring class.
     */
    MemberCollection    allMembers;
    
    /** Model used for creation of new elements.
     */
    ElementCreator      creator;

    /** State variable -- causes this partial collection to ignore changes in
     * slot positions broadcasted by allMembers collection.
     */
    boolean             ignoreContentChange;
    
    protected static final boolean DEBUG = false;

    /** Constructs a new collection of elements of a particular type and binds
     * it to a owner's property.
     * @param ownerImpl implementation object of the collection's owner
     * @param allMembers collection of all members of that owner
     * @param propertyName name of the property this collection represents.
     */
    public PartialCollection(ElementEvents events, ElementCreator creator, 
        MemberCollection allMembers, String propertyName) {
        super(propertyName);
        this.events = events;
        this.allMembers = allMembers;
        this.creator = creator;
        
        // tell the member collection about this partial one:
        allMembers.addPartialCollection(this);
    }
    
    /** Returns the Class object for element impls contained within the collection.
     */
    public abstract Class getElementImplClass();
    
    
    protected abstract ElementImpl createElement(Element parent);

    /** Creates empty array of the collection element's type for array-creation
     * operations.
     */
    protected abstract Object[] createEmpty();
    
    protected abstract Element[] createEmpty(int size);
    
    /** Returns indexes to collection of all class members for individual
     * members of this collection/view.
     */
    public int[] getPositions() {
        return positions;
    }

    /** Returns array of Element objects for members contained within this collection.
     * The method optimizes read access in that it caches results of the query; the results
     * are invalidated by the first write operation.
     */
    public final Element[] getElements() {
        if (this.elements == null)
            return (Element[])createEmpty();
        else
            return this.elements;
    }        
    
    public boolean isEmpty() {
        return this.elements == null || this.elements.length == 0;
    }
    
    /** Adds new members to the collection of elements and to the member collection 
     * as well. The method fires VetoablePropertyChange on the property the collection
     * represents and it have the members collection fire VetoablePropertyChange on the
     * whole elements as well.<P>
     * This method is *only* responsible for translating the request into
     * desired insert positions in the whole collection, then delegating to the
     * MemberCollection to do the actual work.
     */
    public void addMembers(Element[] els) throws SourceException {
        if (els.length == 0)
            return;
        Element[] insertingEls = new Element[els.length];
        Element[] elements = getElements();
        
        for (int i = 0; i < els.length; i++) {
            insertingEls[i] = copyElement(els[i]);
        }

        Element[] allElems = allMembers.getElements();
        Element[] precedings = allMembers.findPositions(insertingEls);
        Element[] newElems = createEmpty(insertingEls.length + elements.length);

        int[] insertIndices = new int[els.length];
        int[] newPositions = new int[newElems.length];
        
        // If an element wants to be inserted after another one, that another one
        // needs to be inserted first (if it is also new one)
        Map m = new HashMap(precedings.length * 4 / 3);
        for (int i = 0; i < precedings.length; i++) {
            // throw exception if prev is not a member of this class.
            Object o = m.get(precedings[i]);
            if (o != null) {
                Collection c;
                if (!(o instanceof Collection)) {
                    c = new LinkedList();
                    c.add(o);
                    m.put(precedings[i], c);
                } else {
                    c = (Collection)o;
                }
                c.add(insertingEls[i]);
            } else {
                m.put(precedings[i], insertingEls[i]);
            }
        }

        Element prevElem = Positioner.FIRST;
        int newPos = 0, oldPos = 0;
        int inserted = 0;
        int newGlobalPos = 0;
        int oldGlobalPos = 0;
        int[] newGlobalIndices = new int[els.length];

        for (oldGlobalPos = 0; oldGlobalPos <= allElems.length; oldGlobalPos++) {
            Object dependent = m.get(prevElem);

            if (dependent != null) {
                //Util.log("inserting after global item #" + oldGlobalPos);
                LinkedList toInsert = new LinkedList();
                
                do {
                    if (dependent instanceof Collection)
                        toInsert.addAll((Collection)dependent);
                    else
                        toInsert.add(dependent);

                    Element newEl = (Element)toInsert.removeFirst();
                    newGlobalIndices[inserted] = newGlobalPos;
                    insertIndices[inserted] = newPos;
                    newElems[newPos] = newEl;
                    newPositions[newPos] = newGlobalPos;
                    inserted++;
                    newPos++;
                    newGlobalPos++;

                    dependent = m.get(newEl);
                } while (dependent != null);
                // inserted everything chained after prevElem
            }
            if (oldGlobalPos >= allElems.length)
                break;
            
            prevElem = allElems[oldGlobalPos];
            // won't transfer.
            if (oldPos < elements.length && prevElem == elements[oldPos]) {
                // we processed our OWN element.
                newPositions[newPos] = newGlobalPos;
                newElems[newPos++] = elements[oldPos++];
            }
            newGlobalPos++;
        }
        if (inserted < insertingEls.length)
            throw new IllegalArgumentException("Inconsistent insertion strategy");
        
        Change chng = new Change();
        chng.insertIndices = insertIndices;
        chng.inserted = Arrays.asList(insertingEls);
        chng.phaseCount = 1;
        chng.sizeDiff = insertingEls.length;

        MultiPropertyChangeEvent evt = createInsertionEvent(events.getEventSource(),
            elements, newElems, chng.inserted, insertIndices);
        
        chng.insertIndices = newGlobalIndices;
        
        // are we allowed to do that ??
        events.fireVetoableChange(evt);

        try {
            ignoreContentChange = true;
            allMembers.changeMembers(chng);
        } finally {
            ignoreContentChange = false;
        }
        
        events.addPropertyChange(evt);
        
        this.elements = newElems;
        this.positions = newPositions;
        sanityCheck();
    }
    
    public void removeMembers(Element[] els) throws SourceException {
        Element[] elements = getElements();
        int[] removeIndices = createIdentityMap(els);
        
        // sanity check:
        for (int i = 0; i < removeIndices.length; i++) {
            if (removeIndices[i] == -1)
                throw new SourceException("Element does not exist");
        }

        Element[] newElems = createEmpty(elements.length - els.length);
        //Util.log("newElems = " + newElems.length);
        int[] offsets = new int[elements.length];
        int newPos = 0;
        int removed = 0;
        int nextRemove = removeIndices[0];
        for (int i = 0; i < elements.length; i++) {
            if (i == nextRemove) {
                offsets[i] = -i -1;
                if (++removed < removeIndices.length)
                    nextRemove = removeIndices[removed];
                else
                    nextRemove = -1;
            } else {
                offsets[i] = newPos - i;
                newElems[newPos] = elements[i];
                newPos++;
            }
        }
        Change ch = new Change();
        ch.removed = Arrays.asList(els);
        ch.removeIndices = removeIndices;
        ch.offsets = offsets;
        ch.sizeDiff = - els.length;
        ch.phaseCount = 1;
        
        // elements & positions will be updated from inside doSetMembers.
        doSetMembers(newElems, ch);
    }
    
    /** Sets the members to be *exactly* the members in the array. The operation is
     * in fact performed as a sequence of remove-add-reorder operations and fires exactly
     * one (possibly compound) property change event. This is probably the most complex
     * method, because it must handle arbitrary changes made to the whole set of elements
    public void setMembers(Element[] els) throws SourceException {
        // work on identities.
        int[] mapping = createIdentityMap(els);
        ElementImpl[] impls = getElements();
        Change chng = computeChanges2(impls, els, mapping);
        doSetMembers(createImplList(mapping, els), mapping, impls, chng);
    }
     */
    
    public void updateMembers(Element[] members, int[] indices, int[] map) {
        if (this.elements != null) {
            Element[] els = getElements();
            Element[] newEls = createEmpty(members.length);
            if (map == null)
                map = createIdentityMap(members);

            Change chng = computeChanges2(els, newEls, map);
            if (chng.phaseCount > 0) {
                events.addPropertyChange(createPropertyEvent(chng, events.getEventSource(),
                    els, newEls));
            }
        } else if (!events.getElementImpl().isCreated()) {
            int[] indexes = new int[members.length];
            Collection insert = new ArrayList(members.length);
            for (int i = 0; i < members.length; i++) {
                indexes[i] = i;
                insert.add(members[i]);
            }
            events.addPropertyChange(createInsertionEvent(events.getEventSource(),
                createEmpty(), members, insert, indexes));
        }
        this.elements = members;
        this.positions = indices;
    }
    
    /** Copies an element over to this collection, creating a local implementation of
     * the element.
     */
    protected Element copyElement(Element el) {
        ElementImpl impl = createElement((Element)events.getEventSource());
        try {
            impl.createFromModel(el);
        } catch (SourceException ex) {
        }
        return impl.getElement();
    }

    public void setMembers(Element[] members) throws SourceException {
        Change chng;
        
        Element[] els = getElements();
        Element[] newEls = createEmpty(members.length);
        int[] map = createIdentityMap(members);
        
        int memberPos = 0;
        boolean copied = false;
        for (int i = 0; i < map.length; i++) {
            if (map[i] == -1) {
                newEls[i] = copyElement(members[i]);
                copied = true;
            } else {
                newEls[i] = members[i];
            }
        }
        
        chng = computeChanges2(els, newEls, map);
        
        doSetMembers(newEls, chng);
    }

    /** Helper function, that tries to map insert indices from the partial collection to
     * the global one so that they can be used to update of the global collection.
     * The problem is, that there can be one or more 'foreign' elements in the global collection
     * between an element and the one that immediatelly follows in the partial collection.
     * Another problem is, that some of the elements may be in removal during the set operation.
     * The helper method tries to find a place that can accept the element.
     */
    private void mapInsertIndices(Change chng, int[] newPositions) throws SourceException {
        int[] iindices = chng.insertIndices;
        int[] offsets = chng.offsets;
        
        /*
        for (int i = 0; i < iindices.length; i++) {
            Util.log("insertion at " + iindices[i]);
        }

        for (int i = 0; i < newPositions.length; i++) {
            Util.log("newPosition[" + i + "] = " + newPositions[i]);
        }
         */

        int nextInsert = iindices[0];;
        int oldIndex = 0;
        int prevOldIndex = -1;
        Element[] all = allMembers.getElements();
        Element[] elements = getElements();

        // there IS insert operation, offsets are always != null.
        int newIndex;
        int iPos = 0;
        int globalPrecedingPos = -1;

        while (iPos < iindices.length) {
            // the while condition skips -1, too.
            if (oldIndex < offsets.length) {
                //Util.log("oldIndex = " + oldIndex + ", nextInsert = " + nextInsert);
                while ((newIndex = oldIndex + offsets[oldIndex]) < nextInsert) {
                    //Util.log("old slot #" + oldIndex + " -> slot #" + newIndex);
                    if (newIndex != -1) {
                        prevOldIndex = oldIndex;
                    }
                    oldIndex++;
                    if (oldIndex >= offsets.length) {
                        oldIndex = -1;
                        newIndex = newPositions.length;
                        break;
                    }
                }
                //Util.log("prevOld = " + prevOldIndex + ", afterOld = " + oldIndex);
                // now: oldIndex = index of old element AFTER the inserted run,
                // prevOldIndex == index of old element BEFORE the inserted run.
                // we must collect (old) elements in <prevOldIndex, oldIndex) from
                // the whole collection and ask whether it is possible to insert after
                // any of them.

                int globalBarrier;
                int testPos;
                int myPos;

                if (prevOldIndex == -1) {
                    globalPrecedingPos = testPos = -1;
                    myPos = 0;
                } else {
                    testPos = positions[prevOldIndex];
                    myPos = prevOldIndex + 1;
                    globalPrecedingPos = testPos + offsets[prevOldIndex];
                }
                if (oldIndex == -1) {
                    globalBarrier = all.length;
                } else {
                    globalBarrier = positions[oldIndex];
                }

                //Util.log("starting search from (new) global pos " + globalPrecedingPos);
                //Util.log("myPos = " + myPos + ", globalPrevPos = " + globalPrecedingPos + ", globalAfterPos = " + globalBarrier);
                while (true) {
                    if (testPos >= globalBarrier) {
                        throw new SourceException("cannot insert");
                    }
                    Element ref = testPos < 0 ? null : all[testPos];
                    //Util.log("testing canInsertAfter(" + testPos + ")");
                    if (allMembers.canInsertAfter(ref)) {
                        // got it!!  globalPrevPos is the element we want to insert
                        // after.
                        break;
                    }
                    if (myPos < elements.length && positions[myPos] != testPos) {
                        // the element is NOT being removed. Record for the case we will
                        globalPrecedingPos++;
                        //Util.log("changing prev global pos = " + globalPrecedingPos);
                    } else {
                        myPos++;
                    }
                    testPos++;
                }
            } else {
                newIndex = nextInsert + 1;
            }
            //Util.log("nextInsert = " + nextInsert + ", newIndex = " + newIndex);
            while (nextInsert < newIndex) {
                globalPrecedingPos++;
                //Util.log("Mapping iindices[" + iPos + "] to " + globalPrecedingPos);
                iindices[iPos] = globalPrecedingPos;
                newPositions[nextInsert] = globalPrecedingPos;
                iPos++;
                if (iPos < iindices.length)
                    nextInsert = iindices[iPos];
                else
                    break;
            }
        }
        /*
        for (int i = 0; i < newPositions.length; i++) {
            Util.log("newPosition[" + i + "] = " + newPositions[i]);
        }
         */

    }
    
    private void doSetMembers(Element[] newElements, Change chng) throws SourceException {
        int[] newPositions = new int[newElements.length];
        Element[] els = getElements();
        int[] offsets = chng.offsets;
        
        // removed, inserted, replacements are Elements
        //Util.log("change object= " + chng);
        MultiPropertyChangeEvent multiEvt = createPropertyEvent(chng,
            events.getEventSource(), getElements(), newElements
        );
 
        // notifies Vetoable listeners about the change, possibly failing with
        // a SourceException
        events.fireVetoableChange(multiEvt);
        
        if (offsets != null) {
            for (int i = 0; i < els.length; i++) {
                int newPos = i + offsets[i];
                if (newPos >= 0) {
                    int newGlobPos = positions[i] + offsets[i];
                    //Util.log("slot #" + i + " moving to " + newPos);
                    newPositions[newPos] = newGlobPos;
                }
            }
        } else {
            // maintain the position array
            System.arraycopy(positions, 0, newPositions, 0, els.length);
        }
        
        if (chng.inserted != null) {
            mapInsertIndices(chng, newPositions);
        }
        
        mapIndices(chng.removeIndices, positions);
        mapIndices(chng.replaceIndices, positions);
        mapReorderIndices(chng, newPositions);
        
        try {
            ignoreContentChange = true;

            // delegate the actual change to the MemberCollection. Change object now
            allMembers.changeMembers(chng);

            /*
            if (offsets != null) {
                for (int i = 0; i < els.length; i++) {
                    int newPos = i + offsets[i];
                    if (newPos >= 0)
                        newPositions[newPos] += offsets[i];
                }
            }
             */
        } finally {
            ignoreContentChange = false;
        }
        this.positions = newPositions;
        // revalidate the cached element array.
        this.elements = newElements;
        
        sanityCheck();
        
        events.addPropertyChange(multiEvt);
    }

    /** Maps reorder indices to the collection of all elements. All the required positions
     * are known in advance (except new members, but those do not appear in the reorder
     * operation. The method uses change object, and overwrites its reordered member.
     */
    private void mapReorderIndices(Change chng, int[] newPositions) {
        if (chng.reordered == null)
            return;
        int[] mainReorder = new int[allMembers.getElements().length];
        int[] myOrder = chng.reordered;
        Arrays.fill(mainReorder, -1);
        
        for (int i = 0; i < myOrder.length; i++) {
            // old position = index -> mapped by current positions.
            // new position = value -> mapped by the new positions.
            if (myOrder[i] == -1)
                continue;
            mainReorder[positions[i]] = newPositions[myOrder[i]];
            //Util.log("reorder-map: from " + positions[i] + " to "  + mainReorder[positions[i]]);
        }
        chng.reordered = mainReorder;
    }
    
    /* package private */ void contentSlotsChanged(int[] offsets) {
        if (ignoreContentChange)
            return;
        int l = getElements().length;
        for (int i = 0; i < l; i++) {
            positions[i] += offsets[positions[i]];
        }
        sanityCheck();
    }
    
    private static void mapIndices(int[] indices, int[] positions) {
        if (indices == null)
            return;
        for (int i = 0; i < indices.length; i++ ) {
            if (indices[i] == -1)
                continue;
            indices[i] = positions[indices[i]];
        }
    }
    
    /*
    private void refreshPositions() {
        int p;
        Iterator all = allMembers.getMembers().iterator();
        Collection impls = getElementImpls();
        Iterator it = impls.iterator();
        
        Map m = new HashMap((impls.size() * 3) / 4);
        for (p = 0; it.hasNext(); p++) {
            m.put(it.next(), new Integer(p));
        }
        
        p = 0;
        while (all.hasNext()) {
            Integer i = (Integer)m.get(all.next());
            if (i != null) {
                positions[i.intValue()] = p;
            }
            p++;
        }
    }
     */
    
    /**
     * Returns an implementation of the containing class element.
    protected final ClassElementImpl getClassImpl() {
        return (ClassElementImpl)events.getElementImpl();
    }
     */
    
    /** Creates mapping from els to the currently held elements in the format that
     * computeChanges2() expects. The method implements a dumb n^2 algorithm for small
     * data sets -- that are smaller than {@link #SMALL_THRESHOLD}. HashMap is used
     * for larger data, provided that the ElementImpls' equals() and hashCode() are appropriate
     * for object identity tests (true for all elements).
     */
    protected int[] createIdentityMap(Element[] els) {
        Element[] olds = getElements();
        return super.createIdentityMap(els, olds);
    }

    protected final Object getSource(ElementImpl impl) {
        return impl.getElement();
    }
    
    public void changeMembers(Element[] items, int operation) throws SourceException {
        switch (operation) {
            case ClassElementImpl.ADD:
                addMembers(items);
                break;
            case ClassElementImpl.REMOVE:
                removeMembers(items);
                break;
            case ClassElementImpl.SET:
                setMembers(items);
                break;
            default:
                throw new InternalError("Unknown/unsupported operation: " + operation);
        }
    }
    
    protected void sanityCheck() {
        if (!DEBUG)
            return;
        Element[] els = getElements();
        int[] indices = positions;
        Element[] allElements = allMembers.getElements();
        
        System.err.println("Sanity check " + getClass() + " of " + events);
        if (els.length > 0) {
            if (indices.length < els.length) {
                System.err.println("positions.length = "
                    + positions.length + ", elements = " + els.length);
            }
        }
        ElementImpl parent = events.getElementImpl();
        boolean parentValid = parent.isValid();
        
        for (int i = 0; i < els.length; i++) {
            if (els[i] == null) {
                System.err.println("Element #" + i + " is null");
                continue;
            }
            ElementImpl impl = (ElementImpl)els[i].getCookie(ElementImpl.class);
            if (impl == null) {
                System.err.println("memberimpl #" + i +
                    " is null");
                continue;
            }
            if (parentValid && !impl.isValid()) {
                System.err.println("element #" + i + 
                    " is invalid");
                continue;
            }
            int mainPos = indices[i];
            if (mainPos < 0 || mainPos >= allElements.length) {
                System.err.println("element #" + i +
                    " has invalid position: "+ mainPos);
                continue;
            }
            if (allElements[mainPos] != els[i]) {
                System.err.println("element #" + i + " has incosistent position: " +
                    mainPos + " -> " + allElements[mainPos]);
            }
        }
    }
}
