/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.model;

import java.beans.PropertyChangeEvent;
import java.lang.ref.*;
import java.util.*;

import org.openide.src.*;

/**
 * Represents a source file. This element manages package declaration, zero or more
 * import statements, and zero or more classelements. <B>This element has not any
 * Binding</B> since it has no (direct) properties.<P>
 * Although this implementation *does* implement entire SourceElement.Impl interface,
 * methods that have not anything to do with the model itself are <B>dummy</B> and
 * do nothing (parse, getStatus). 
 * @author  sdedic
 * @version 
 */
public class SourceElementImpl extends ElementImpl implements SourceElement.Impl,
    ElementOrder {
    /**
     * Package identifier. Null if not specified.
     */
    private Identifier   packageID;
    
    /** Import definitions. May be null if there are no imports.
     */
    ImportCollection    imports;
    
    /** Reference keeping a collection of ClassElementImpls residing here.
     */
    private TopClassCollection  topClasses;
    
    private IdentContextSupport identSupp;
    
    // Getters
    ///////////////////////////////////////////////////////////////////////////////////
    
    public SourceElementImpl(DefaultLangModel model) {
        super(model);
        identSupp = new IdentContextSupport(11);
    }
    
    protected void notifyElementCreated() {
        // do nothing -- override ElementImpl's behaviour.
    }
    
    protected void createFromModel(Element model) throws SourceException {
    }
    
    protected Binding createBinding(Element el) {
        return null;
    }
    
    public Element[] getElements() {
        return getClasses();
    }
    
    public Identifier getPackage() {
        return this.packageID;
    }
    
    public Import[] getImports() {
        if (imports == null)
            return ImportCollection.NO_IMPORTS;
        return imports.getImports();
    }
    
    public ImportElement[] getImportElements() {
        if (imports == null) 
            return ImportCollection.NO_IMPORT_ELEMENTS;
        return (ImportElement[])imports.getElements();
    }
    
    public ClassElement getClass(Identifier id) {
        if (topClasses == null) 
            return null;
        return topClasses.getClass(id);
    }
    
    public ClassElement[] getClasses() {
        if (topClasses == null) 
            return InnerClassCollection.EMPTY;
        return (ClassElement[])topClasses.getElements();
    }
    
    public ClassElement[] getAllClasses() {
        if (topClasses == null)
            return InnerClassCollection.EMPTY;
        
        ClassElement[] tops = getClasses();
        Collection list = new LinkedList();
        
        for (int i = 0; i < tops.length; i++) {
            addAllClasses(tops[i], list);
        }
        return (ClassElement[])list.toArray(InnerClassCollection.EMPTY);
    }
    
    private void addAllClasses(ClassElement c, Collection col) {
        col.add(c);
        ClassElement[] inner = c.getClasses();
        for (int i = 0; i < inner.length; i++)
            addAllClasses(inner[i], col);
    }
    
    // Setters/changers
    ///////////////////////////////////////////////////////////////////////////////////
    public void setPackage(Identifier packageID) throws SourceException {
        Object token = takeLock();
        try {
            if (this.packageID == packageID ||
                (this.packageID != null && packageID != null && 
                    this.packageID.getSourceName().equals(packageID.getSourceName())))
                return;
            PropertyChangeEvent evt;
            
            evt = new PropertyChangeEvent(getEventSource(), PROP_PACKAGE,
                this.packageID, packageID);
            checkVetoablePropertyChange(evt);
            getSourceBinding().changePackage(packageID);
            fireOwnPropertyChange(evt);
            this.packageID = packageID;
            updateClassNames();
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    private void updateClassNames() {
        Identifier p = getPackage();
        String prefix;
        
        if (topClasses == null)
            return;
        if (p == null) 
            prefix = ""; // NOI18N
        else
            prefix = p.getSourceName();
        
        ClassElement[] cls = (ClassElement[])topClasses.getElements();
        for (int i = 0; i < cls.length; i++) {
            ClassElementImpl impl = (ClassElementImpl)cls[i].getCookie(ClassElementImpl.class);
            impl.updateName(prefix);
        }
    }
    
    public void changeImports(Import[] defs, int operation) throws SourceException {
        Object token = takeLock();
        try {
            if (imports == null) {
                if (operation != ClassElement.Impl.REMOVE &&
                    defs.length == 0)
                    return;
                initializeImports();
            }
            imports.changeMembers(defs, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    public void changeClasses(ClassElement[] classes, int operation) throws SourceException {
        Object token = takeLock();
        try {
            if (topClasses == null) {
                if (operation != ClassElement.Impl.REMOVE &&
                    classes.length == 0)
                    return;
                initializeClasses();
            }
            topClasses.changeMembers(classes, operation);
            commit();
        } finally {
            releaseLock(token);
        }
    }
    
    private void notifyCreate(Element[] els) {
        for (int i = 0; i < els.length; i++) {
            ElementImpl impl = (ElementImpl)els[i].getCookie(ElementImpl.class);
            impl.notifyCreate();
        }
    }
    
    protected void notifyCreate() {
        Element[] els;
        
        if (this.imports != null) {
            notifyCreate(imports.getElements());
        }
        if (this.topClasses != null) {
            notifyCreate(topClasses.getElements());
        }
        super.notifyCreate();
    }
    
    private Binding.Source getSourceBinding() {
        return (Binding.Source)getBinding();
    }

    /** Dummy method to satisfy typing requirements. This method does not work with the
     * model, but rather serves as interface to the parser. The method may be implemented
     * in subclasses or delegating implementations.
     */
    public int getStatus() {
        return SourceElement.STATUS_NOT;
    }
    
    public org.openide.util.Task prepare() {
        return org.openide.util.Task.EMPTY;
    }
    
    public void runAtomic(Runnable run) {
        // no-op
    }
    
    public void runAtomicAsUser(Runnable run) throws SourceException {
        // no-op
    }
    
    public void updateMembers(String name, Element[] elements, int[] optMap) {
        if (name == ElementProperties.PROP_IMPORTS) {
            if (imports == null) { 
                if (elements.length == 0)
                    return;
                initializeImports();
            }
            imports.updateMembers(elements, optMap);
        } else if (name == ElementProperties.PROP_CLASSES) {
            if (topClasses == null) {
                if (elements.length == 0)
                    return;
                initializeClasses();
            }
            topClasses.updateMembers(elements, optMap);
        } else {
            throw new IllegalArgumentException("Unsupported property: " + name);
        }
    }
    
    // Implementation details
    ///////////////////////////////////////////////////////////////////////////////////
    public void setParent(ElementImpl impl) {
        // no-op, sources cannot have parents.
    }
    
    private void initializeImports() {
        this.imports = new ImportCollection(((Binding.Source)getRawBinding()).getImportsSection(),
            getModelImpl(), this);
    }
    
    private void initializeClasses() {
        this.topClasses = new TopClassCollection(
            ((Binding.Source)getRawBinding()).getClassSection(),
            getModelImpl(), this);
    }
    
    protected SourceElementImpl findSource() {
        return this;
    }
    
    public Object readResolve() {
        return null;
    }
    
    protected Element cloneSelf() {
        return null;
    }
    
    protected Identifier createLocalIdentifier(Identifier id, int status) {
        return identSupp.create(id, status);
    }
    
    protected boolean checkIdentifierContext(Identifier id) {
        return identSupp.checkContext(id);
    }
    
    protected boolean parentValid() {
        return isValid();
    }
}
