/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.parser;

import java.util.*;

import org.openide.nodes.Node;

import org.openide.text.PositionBounds;
import org.openide.src.*;

import org.netbeans.modules.java.ElementFactory;

import org.netbeans.modules.java.model.Binding;
import org.netbeans.modules.java.model.LangModel;

import org.netbeans.modules.java.codegen.DocumentBinding;
import org.netbeans.modules.java.codegen.TextBinding;

/**
 *
 * @author  svata
 * @version 
 */
public abstract class BaseElementInfo implements ElementFactory.Item {
    ElementFactory.Item parent;
    String              name;
    int                 modifiers;
    String              docText;
    Element             el;
    TextBinding         binding;
    String              bodyContent;
    
    PositionBounds      wholeBounds;
    PositionBounds      bodyBounds;
    PositionBounds      headerBounds;
    PositionBounds      docBounds;
    
    BaseElementInfo(String name, int modifiers) {
        this.name = name;
        this.modifiers = modifiers;
    }
    
    public Element createModelElement(Element parent, LangModel.Updater model) 
    throws SourceException {
        Element el = createModelImpl(model, parent);
        updateElement(model, el);
        return this.el = el;
    }
    
    public int comparePositionTo(TextBinding b) {
        if (b == null)
            return -1;
        PositionBounds bounds = b.getElementRange(false);
        
        int begin = wholeBounds.getBegin().getOffset();
        int otherBegin = bounds.getBegin().getOffset();
        int end = wholeBounds.getEnd().getOffset();
        int otherEnd = bounds.getEnd().getOffset();
        
        if (otherBegin == otherEnd)
            // the binding is invalid - extinct ??
            return -1;
        
        if (begin < otherEnd && end > otherBegin)
            // MATCH!!
            return 0;
        if (end == otherBegin)
            return -1;
        return end - otherBegin;
    }
    
    protected Identifier createName() {
        return Identifier.create(name);
    }
    
    public Node.Cookie getCookie(Class desiredClass) {
        return null;
    }
    
    protected abstract Element createModelImpl(LangModel.Updater model, Element parent);
    
    protected Identifier[] mapIdentifiers(LangModel.Updater model, Identifier[] arr, Element t) {
        if (arr.length == 0)
            return arr;
        Identifier[] ids = new Identifier[arr.length];
        for (int i = 0; i < ids.length; i++) {
            ids[i] = createContextID(model, t, arr[i]);
        }
        return ids;
    }
    
    protected Identifier createContextID(LangModel.Updater model, Element context, Identifier orig) {
        return model.createLocalIdentifier(context, orig, orig.getResolutionStatus());
    }
    
    protected Type createContextType(LangModel.Updater model, Element context, Type orig) {
        if (orig.isPrimitive())
            return orig;
        
        Type t2 = orig;
        int depth = 0;
        
        while (t2.isArray()) {
            t2 = t2.getElementType();
            depth++;
        }
        if (t2.isPrimitive())
            return orig;
        Identifier typeid = t2.getTypeIdentifier();
        Identifier newId = model.createLocalIdentifier(context, typeid,
            typeid.getResolutionStatus());
        t2 = Type.createClass(newId);
        while (depth-- > 0)
            t2 = Type.createArray(t2);
        return t2;
    }
    
    public void updateElement(LangModel.Updater model, Element el) throws SourceException {
        this.el = el;
        this.binding = (TextBinding)model.getElementBinding(el);
    }
    
    public void updateBindings(DocumentBinding updater) {
        updateBinding(updater, binding);
    }
    
    public void updateBinding(DocumentBinding updater, TextBinding b) {
        b.updateBounds(TextBinding.BOUNDS_ALL, wholeBounds);
        b.updateBounds(TextBinding.BOUNDS_HEADER, headerBounds);
        b.updateBounds(TextBinding.BOUNDS_BODY, bodyBounds);
        b.updateBounds(TextBinding.BOUNDS_JAVADOC, docBounds);
    }
    
    protected static void updateChildren(DocumentBinding updater,
        Collection infos, TextBinding.Container container) {
        List blist = new ArrayList(infos.size());
        for (Iterator it = infos.iterator(); it.hasNext(); ) {
            BaseElementInfo i = (BaseElementInfo)it.next();
            i.updateBindings(updater);
            blist.add(i.binding);
        }
        container.updateChildren(blist);
    }
    
    protected void updateBase(Element el) throws SourceException {
        this.el = el;
        MemberElement.Impl m = (MemberElement.Impl)el.getCookie(
            MemberElement.Impl.class);
        m.setModifiers(this.modifiers);
        m.setName(createName());
    }
    
    protected void updateJavaDoc(JavaDoc doc) throws SourceException {
        doc.setRawText(docText);
    }

    public static class NameFinder extends ElementMatch.AbstractFinder {
        protected boolean matches(BaseElementInfo info, Element old) {
            MemberElement oldMember = (MemberElement)old;
            return info.name.equals(oldMember.getName().getSourceName());
        }
    }
}
