/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.parser;

import java.util.BitSet;
import java.util.Collection;
import java.util.LinkedList;

import org.openide.src.*;

import org.netbeans.modules.java.model.Binding;
import org.netbeans.modules.java.model.ElementImpl;
import org.netbeans.modules.java.model.LangModel;

import org.netbeans.modules.java.codegen.DocumentBinding;
import org.netbeans.modules.java.codegen.TextBinding;

/**
 * Holds parsed out information about a class.
 *
 * @author  svata
 * @version 
 */
public class ClassInfo extends BaseElementInfo {
    public static final int INITIALIZER = 0;
    public static final int FIELD = 1;
    public static final int CONSTRUCTOR = 2;
    public static final int METHOD = 3;
    public static final int CLASS = 4;
    
    Identifier      superclass;
    Identifier[]    superinterfaces;
    String          fullName;
    boolean         properClass;

    Collection           allMembers;
    ChildCollection[]    memberLists;
    Element[]            allElements;
    
    static final ElementMatch.Finder[] DEFAULT_CLASS_FINDERS = {
        new TextPositionMatch(), new NameFinder()
    };
    
    private static final ElementMatch.Finder[] DEFAULT_INIT_FINDERS = {
        new TextPositionMatch()
    };
    
    private static final ElementMatch.Finder[] DEFAULT_FIELD_FINDERS = 
        DEFAULT_CLASS_FINDERS;
    
    private static final ElementMatch.Finder[] DEFAULT_CONS_FINDERS = {
        new TextPositionMatch(), 
        new MethodInfo.Finder(false, true)
    };
    
    private static final ElementMatch.Finder[] DEFAULT_METHOD_FINDERS = {
        new TextPositionMatch(), new MethodInfo.Finder(true, true),
        new MethodInfo.Finder(true, false),
        new MethodInfo.Finder(false, true)
    };
    
    private static final ElementMatch.Finder[][] FINDER_CLUSTERS = {
        DEFAULT_INIT_FINDERS,
        DEFAULT_FIELD_FINDERS,
        DEFAULT_CONS_FINDERS,
        DEFAULT_METHOD_FINDERS,
        DEFAULT_CLASS_FINDERS
    };
    
    private static final String[] CHILDREN_PROPERTIES = {
        ElementProperties.PROP_INITIALIZERS,
        ElementProperties.PROP_FIELDS,
        ElementProperties.PROP_CONSTRUCTORS,
        ElementProperties.PROP_METHODS,
        ElementProperties.PROP_CLASSES
    };
    
    private static final Class[] CHILDREN_TYPES = {
        InitializerElement.class, FieldElement.class, ConstructorElement.class,
        MethodElement.class, ClassElement.class
    };
    
    public ClassInfo(String name, int modifiers, boolean properClass) {
        super(name, modifiers);
        allMembers = new LinkedList();
        memberLists = new ChildCollection[CLASS + 1];
        this.properClass = properClass;
    }
    
    private void initializeChildren(int kind) {
        memberLists[kind] = new ChildCollection(FINDER_CLUSTERS[kind], 
            CHILDREN_TYPES[kind]);
    }
    
    public void setSuperclasses(Identifier sup, Identifier[] itf) {
        this.superclass = sup;
        this.superinterfaces = itf;
    }
    
    private static final Element[] NO_MEMBERS = new Element[0];
    
    public void updateBinding(DocumentBinding updater, TextBinding b) {
        super.updateBinding(updater, b);
        updateChildren(updater, allMembers, (TextBinding.Container)b);
    }
    
    public void updateElement(LangModel.Updater model, Element target) throws SourceException {
        super.updateElement(model, target);
        super.updateBase(target);
        
        ClassElement clazz = (ClassElement)target;

        Util.log("Updating class properties of " + fullName);
        clazz.setClassOrInterface(this.properClass);
        clazz.setInterfaces(this.superinterfaces);
        clazz.setSuperclass(this.superclass);

        updateJavaDoc(clazz.getJavaDoc());
        
        Element[] whole = new Element[allMembers.size()];
        Element[] newEls;
        
        for (int kind = INITIALIZER; kind <= CLASS; kind++) {
            Element[] curMembers;
            switch (kind) {
                case INITIALIZER:
                    curMembers = clazz.getInitializers();
                    break;
                case FIELD:
                    curMembers = clazz.getFields();
                    break;
                case CONSTRUCTOR:
                    curMembers = clazz.getConstructors();
                    break;
                case METHOD:
                    curMembers = clazz.getMethods();
                    break;
                case CLASS:
                    curMembers = clazz.getClasses();
                    break;
                default:
                    throw new InternalError("Illegal member type");
            }

            ChildCollection col = memberLists[kind];
            if (col == null && curMembers.length == 0)
                continue;
            
            int[] IDs, map;
            if (col == null) {
                newEls = (Element[])java.lang.reflect.Array.newInstance(this.CHILDREN_TYPES[kind], 0);
                map = IDs = null;
            } else {
                IDs = col.getIDs();
                newEls = col.updateChildren(clazz, model, curMembers);
                map = col.getResultMap();
            }
            model.updateMembers(clazz, CHILDREN_PROPERTIES[kind], newEls,
                IDs, map);
            if (col != null)
                col.mapChildren(newEls, whole);
        }
        model.updateMemberOrder(clazz, ElementProperties.PROP_MEMBERS, whole);
    }
    
    protected Identifier createName() {
        return Identifier.create(fullName, name);
    }

    public Element createModelImpl(LangModel.Updater model, Element parent) {
        ElementImpl impl;

        Util.log("Creating class " + fullName);
        if (this.parent instanceof ClassInfo) {
            impl = model.createInnerClass((ClassElement)parent);
        } else {
            impl = model.createTopClass((SourceElement)parent);
        }
        return impl.getElement();
    }
    
    public void addMember(int kind, BaseElementInfo member) {
        member.parent = this;
        int index = allMembers.size();
        allMembers.add(member);
        if (memberLists[kind] == null)
            initializeChildren(kind);
        memberLists[kind].addChild((BaseElementInfo)member, index);
    }
}
