/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.parser;

import javax.swing.text.*;

import org.openide.src.*;

import org.openide.text.CloneableEditorSupport;
import org.openide.text.PositionBounds;
import org.openide.text.PositionRef;

import org.netbeans.modules.java.ElementFactory;
import org.netbeans.modules.java.codegen.SourceText;


/**
 *
 * @author  svata
 * @version 
 */
public class DocumentModelBuilder extends SourceInfo implements ElementFactory {
    CloneableEditorSupport  editSupp;
    char[]  inputText;
    boolean computeHash;
    
    public DocumentModelBuilder(CloneableEditorSupport supp) {
        this.editSupp = supp;
    }
    
    protected void setContent(char[] content, boolean computeHash) {
        this.inputText = content;
        this.computeHash = computeHash;
    }
    
    public void createPackage(Identifier name, int begin, int end) {
        packageDecl = name;
        packageBounds = createStableBounds(begin, end);
    }
    
    public void createImport(Import im, int begin, int end) {
        PositionBounds bnds = createStableBounds(begin, end);
        ImportInfo info = new ImportInfo(im);
        info.wholeBounds = bnds;
        addImport(info);
    }
    
    public Item createClass(boolean isInterface, int modifiers, Identifier name, Identifier superclass, Identifier[] interfaces) {
        ClassInfo info = new ClassInfo(name.getSourceName(), modifiers, !isInterface);
        
        info.superclass = superclass;
        info.superinterfaces = interfaces;
        info.fullName = name.getFullName();
        return info;
    }
    
    public Item createConstructor(int modifiers, Identifier id, MethodParameter[] params, Identifier[] exceptions) {
        MethodInfo info = new MethodInfo(id.getSourceName(), modifiers);
        info.params = params;
        info.exceptions = exceptions;
        info.constructor = true;
        return info;
    }
    
    public Item createField(int modifiers, Identifier name, Type type, String initializer) {
        FieldInfo info = new FieldInfo(name.getSourceName(), modifiers);
        info.type = type;
        info.initValue = initializer;
        return info;
    }
    
    public Item createInitializer(int modifiers) {
        InitializerInfo info = new InitializerInfo(modifiers);
        return info;
    }
    
    public Item createMethod(int modifiers, Identifier name, Type returnType, MethodParameter[] params, Identifier[] exceptions) {
        MethodInfo info = new MethodInfo(name.getSourceName(), modifiers);
        info.retType = returnType;
        info.params = params;
        info.exceptions = exceptions;
        info.constructor = false;
        return info;
    }
    
    public void markError(Item item) {
        // huh ? attach some error-information-decorator to an element ??
    }
    
    public void setBodyBounds(Item item, int begin, int end) {
        BaseElementInfo info = (BaseElementInfo)item;
        info.bodyBounds = createAbsorbingBounds(begin, end);
        if (computeHash && (info instanceof MethodInfo || info instanceof InitializerInfo)) {
            info.bodyContent = String.copyValueOf(this.inputText, begin, end - begin);
        }
    }
    
    public void setBounds(Item item, int begin, int end) {
        BaseElementInfo info = (BaseElementInfo)item;
        info.wholeBounds = createStableBounds(begin, end);
    }
    
    public void setDocumentation(Item item, int begin, int end, String text) {
        BaseElementInfo info = (BaseElementInfo)item;
        info.docText = text;
        if (text != null && begin >= 0 && end > begin)
            info.docBounds = createStableBounds(begin, end);
        else
            info.docBounds = null;
    }
    
    public void setFieldTypeBounds(Item item, int begin, int end) {
        FieldInfo finfo = (FieldInfo)item;
        
        finfo.typeBounds = createStableBounds(begin, end);
    }
    
    public void setHeaderBounds(Item item, int begin, int end) {
        BaseElementInfo info = (BaseElementInfo)item;
        info.headerBounds = createStableBounds(begin, end);
    }
    
    public void setParent(Item child, Item parent) {
        ((BaseElementInfo)child).parent = parent;
        
        if (parent == null) {
            // must be a top-level class:
            addClass((ClassInfo)child);
        } else {
            ClassInfo parentInfo = (ClassInfo)parent;
            
            if (child instanceof InitializerInfo) {
                parentInfo.addMember(ClassInfo.INITIALIZER, (BaseElementInfo)child);
            } else if (child instanceof FieldInfo) {
                parentInfo.addMember(ClassInfo.FIELD, (BaseElementInfo)child);
            } else if (child instanceof ClassInfo) {
                parentInfo.addMember(ClassInfo.CLASS, (BaseElementInfo)child);
            } else if (child instanceof MethodInfo) {
                MethodInfo minfo = (MethodInfo)child;
                parentInfo.addMember(
                    minfo.constructor ? ClassInfo.CONSTRUCTOR : ClassInfo.METHOD,
                    (BaseElementInfo)child
                );
            }
        }
    }
    
    public void setPrecedingField(Item item, Item previous) {
        FieldInfo prev = (FieldInfo)previous;
        FieldInfo current = (FieldInfo)item;
        
        prev.nextField = current;
    }
    
    
    protected PositionBounds createStableBounds(int from, int to) {
        PositionRef posBegin = editSupp.createPositionRef(from, Position.Bias.Backward);
        PositionRef posEnd = editSupp.createPositionRef(to, Position.Bias.Forward);
        return new PositionBounds(posBegin, posEnd);
    }
    
    protected PositionBounds createAbsorbingBounds(int from, int to) {
        PositionRef posBegin = editSupp.createPositionRef(from, Position.Bias.Forward);
        PositionRef posEnd = editSupp.createPositionRef(to, Position.Bias.Backward);
        return new PositionBounds(posBegin, posEnd);
    }
}
