/*
 * ParseSourceRequest.java
 *
 * Created on January 23, 2001, 3:24 PM
 */

package org.netbeans.modules.java.parser;

import org.openide.util.Task;

import java.io.InputStream;
import java.io.IOException;
import java.io.Reader;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.openide.src.SourceElement;
import org.openide.text.CloneableEditorSupport;

import org.netbeans.modules.java.ParserEngine;
import org.netbeans.modules.java.ElementFactory;

/**
 *
 * @author  sdedic
 * @version
 */
public class ParseSourceRequest implements ParsableObjectRequest {
    public static final int     STATE_WAITING = 0;
    public static final int     STATE_READING = 1;
    public static final int     STATE_CANCELLED = 2;
    public static final int     STATE_ANALYSIS = 3;
    public static final int     STATE_UPDATING = 4;
    public static final int     STATE_COMPLETED = 10;
    
    private static final int READ_THRESHOLD = 2048;
    
    JavaParser.Env          environment;
    ChangeListener          listener;
    int                     state;
    boolean                 valid;
    int                     syntaxErrors;
    DocumentModelBuilder    builder;
    CloneableEditorSupport  editSupp;

    public ParseSourceRequest() {
        state = STATE_WAITING;
        valid = true;
    }

    ParseSourceRequest(JavaParser.Env env, CloneableEditorSupport editSupp) {
        this();
        this.editSupp = editSupp;
        this.environment = env;
    }
    
    public synchronized void addChangeListener(ChangeListener l) 
        throws java.util.TooManyListenersException {
        if (listener != null)
            throw new java.util.TooManyListenersException();
        listener = l;
    }
    
    public synchronized void removeChangeListener(ChangeListener l) {
        if (listener == l)
            listener = null;
    }
    
    public void setEnvironment(JavaParser.Env env) {
        environment = env;
    }
    
    public void setEditorSupport(CloneableEditorSupport editor) {
        editSupp = editor;
    }

    /**
     * Notifies the request that the source text has been changed. This causes
     * cancellation of the request in some cases.
     */
    public void sourceChanged() {
        if (state == STATE_READING) {
            // cancel the request only in reading state; if the text is already
            // read, the request can still be cariied out.
            valid = false;
        }
    }
    
    public void modelChanged() {
        if (state != STATE_WAITING && state != STATE_COMPLETED) {
            valid = false;
        }
    }
        
    public void setSyntaxErrors(int errors) {
        this.syntaxErrors = errors;
    }
    
    public int getSyntaxErrors() {
        return syntaxErrors;
    }
    
    public void setSemanticErrors(int errors) {
        // ignore - for now.
    }
    
    /**
     * DocumentModelBuilder actually carries out tasks associated with the Factory.
     */
    public ElementFactory getFactory() {
        return getBuilder();
    }
    
    public DocumentModelUpdater getUpdater() {
        return getBuilder();
    }

    protected DocumentModelBuilder getBuilder() {
        if (builder == null)
            builder = createBuilder(editSupp);
        return builder;
    }

    protected DocumentModelBuilder createBuilder(CloneableEditorSupport supp) {
        return new DocumentModelBuilder(supp);
    }

    public void notifyReschedule() {
        // clear old data.
        builder = null;
        enterState(STATE_WAITING);
    }
    
    protected void enterState(int state) {
        this.state = state;
        if (listener != null)
            listener.stateChanged(new ChangeEvent(this));
    }
    
    /**
     * The method is implemented so that it reads the whole contents from the environment's
     * Reader and returns the resulting char array.
     */
    public char[] getSource() throws java.io.IOException {
        // unsynchronized; the exact sequence really does not matter here too much;
        // the real problem arises that AFTER the contents is read.
        valid = true;
        enterState(STATE_READING);
        
        Reader r = environment.getSourceText();
        
        List buffers = new LinkedList();
        char[] buf;
        int read;
        int offset;
        int total = 0;

        do {
            offset = 0;
            buf = new char[READ_THRESHOLD];
            do {
                read = r.read(buf);
                if (read == -1)
                    break;
                offset += read;
            } while (offset < buf.length);
            if (offset > 0)
                buffers.add(buf);
            total += offset;
        } while (read > 0);
        
        r.close();
        
        buf = new char[total];
        read = 0;
        for (Iterator it = buffers.iterator(); it.hasNext(); ) {
            char[] b = (char[])it.next();
            int size = it.hasNext() ? b.length : total - read;
            System.arraycopy(b, 0, buf, read, size);
            read += size;
        }
        getBuilder().setContent(buf, editSupp.isDocumentLoaded());
        return buf;
    }

    /** This should locate appropriate .class containing the class 'className'.
     * The current implementation will cowardly return false.
     */
    public InputStream findCompiledClass(String className) {
        return environment.findCompiledClass(className);
    }
    
    /**
     * Returns true, IF the request was not invalidated by the ParserSupport,
     * or because of change in the document while parsing.
     */
    public boolean isValid() {
        return this.valid;
    }
    
    public boolean needsProcessing() {
        // always needs to process (?)
        return this.valid;
    }
    
    public void notifyStart() {
    }
    
    public void notifyComplete() {
        // safepoint: everything was processed.
        enterState(STATE_COMPLETED);
    }

    public Object getParserType() {
        return JavaParser.SHALLOW_PARSER;
    }
}
