/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.settings;

import java.io.*;
import java.util.ResourceBundle;
import java.util.Properties;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.openide.options.ContextSystemOption;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.io.ReaderInputStream;
import org.openide.compiler.CompilerType;
import org.openide.ServiceType;
import org.openide.ServiceType.Handle;
import org.openide.TopManager;
import org.openide.execution.Executor;
import org.openide.execution.NbClassPath;
import org.openide.debugger.DebuggerType;
import org.openide.nodes.Node;
import org.openide.util.SharedClassObject;
import org.openide.util.WeakListener;

import org.netbeans.modules.java.FastJavacCompilerType;
import org.netbeans.modules.java.JavaExternalCompilerType;
import org.netbeans.modules.java.gj.JavaCompilerType;
import org.netbeans.modules.java.JavaDataLoader;
import org.netbeans.modules.java.JavaProcessExecutor;
import org.netbeans.modules.java.JavaThreadExecutor;
import org.netbeans.modules.java.imptool.ImpToolSettings;
import org.netbeans.modules.java.ClassPathConfigurator;

/** Settings for java data loader and java source parser
*
* @author Ales Novak, Petr Hamernik
*/
public class JavaSettings extends ContextSystemOption {
    private static final int currentVersion = 1;

    /** serial uid */
    static final long serialVersionUID = -8522143676848697297L;

    public static final String PROP_COMPILER = "compiler"; // NOI18N
    public static final String PROP_EXECUTOR = "executor"; // NOI18N
    public static final String PROP_DEBUGGER = "debugger"; // NOI18N
    public static final String PROP_PARSER_BOOTPATH = "parserBootPath"; // NOI18N
    public static final String PROP_PARSER_CLASSPATH = "parserClassPath"; // NOI18N
    public static final String PROP_PARSER_SOURCEPATH = "parserSourcePath"; // NOI18N
    
    public static final String PROP_SHOW_COMPILE_STATUS = "showCompileStatus"; // NOI18N

    public static final String PROP_REPLACEABLE_STRINGS_TABLE = "replaceableStringsTable"; // NOI18N

    public static final String PROP_AUTO_PARSING_DELAY = "autoParsingDelay"; // NOI18N

    /**
     * @deprecated the property has moved to the {@link JavaDataLoader}
     */
    public static final String PROP_PERFECT_RECOGNITION = "perfectRecognition"; // NOI18N
    
    /** Name of "prescan source" property.
     */
    public static final String PROP_PRESCAN_SOURCES = "prescanSources"; // NOI18N
    
    public static final String PROP_SHOW_TOOLBAR = "showToolbar";

    public static final int DEFAULT_AUTO_PARSING_DELAY = 2000;

    /** The resource bundle for the form editor */
    public static ResourceBundle bundle;

    /** class path settings or null */
    private NbClassPath parserClassPath;

    /** boot class path or null */
    private NbClassPath parserBootPath;
    
    /**
     * Keeps parser's filesystems in memory unless un-configured
     */
    private transient Collection keepClassPath, keepBootPath;

    /**
     * Bootclasspath obtained from the runtime environment.
     */
    private transient NbClassPath   runtimeBootClassPath;

    /**
     * Whether to show the toolbar in the editor or not.
     */
    private boolean showToolbar;
    
    /**
     * Current version of the settings object. If < currentVersion,
     * a settings upgrade is made.
     */
    private int version;
    
    /** If true then external execution is used */
    public String displayName () {
        return getString("CTL_Java_option");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (JavaSettings.class);
    }

    public JavaSettings() {
        addOption(getJavaSynchronizationSettings());
        addOption(getImpToolSettings());
    }

    public boolean isGlobal() {
        return false;
    }
    
    public boolean getShowToolbar() {
        return showToolbar;
    }
    
    public void setShowToolbar(boolean enable) {
        if (showToolbar == enable)
            return;
        showToolbar = enable;
        firePropertyChange(PROP_SHOW_TOOLBAR, 
            enable ? Boolean.FALSE : Boolean.TRUE, enable ? Boolean.TRUE : Boolean.FALSE);
    }

    /** @return CompilerType */
    public CompilerType getCompiler() {
        CompilerType.Handle compilerType = (CompilerType.Handle)getProperty(PROP_COMPILER);
        CompilerType type = null;
        
        if (compilerType != null) {
            type = (CompilerType)compilerType.getServiceType();
        }
        if (type == null) {
            return setDefaultCompilerType();
	}
        return type;
    }

    CompilerType setDefaultCompilerType() {
        CompilerType jcompilerType = CompilerType.find(JavaExternalCompilerType.class);
        
        if (jcompilerType == null)
            jcompilerType = CompilerType.getDefault();
        setCompiler(jcompilerType);
        return jcompilerType;
    }
    
    /** Usese given CompilerType */
    public void setCompiler(CompilerType ct) {
        putProperty(PROP_COMPILER, new CompilerType.Handle(ct), true);
    }

    /** @return Executor */
    public Executor getExecutor() {
        ServiceType.Handle serviceType = (ServiceType.Handle) getProperty(PROP_EXECUTOR);
        if ((serviceType == null) ||
                (serviceType.getServiceType() == null)) {
            // External Execution is the default for Java sources:
            Executor exec = Executor.find (JavaProcessExecutor.class);
            if (exec == null)
                exec = Executor.getDefault();
            serviceType = new ServiceType.Handle(exec);
            putProperty(PROP_EXECUTOR, serviceType, false);
            return exec;
        }
        return (Executor) serviceType.getServiceType();
    }

    /** sets an executor */
    public void setExecutor(Executor ct) {
        putProperty(PROP_EXECUTOR, new ServiceType.Handle(ct), true);
    }

    /** @return debugger */
    public DebuggerType getDebugger() {
        ServiceType.Handle serviceType = (ServiceType.Handle) getProperty(PROP_DEBUGGER);
        if ((serviceType == null) ||
                (serviceType.getServiceType() == null)) {
            DebuggerType debugger = DebuggerType.getDefault();
            serviceType = new ServiceType.Handle(debugger);
            putProperty(PROP_DEBUGGER, serviceType, false);
            return debugger;
        }
        return (DebuggerType) serviceType.getServiceType();
    }

    /** Sets a debugger type */
    public void setDebugger(DebuggerType ct) {
        putProperty(PROP_DEBUGGER, new ServiceType.Handle(ct), true);
    }

    /** Sets the replaceable strings table - used during instantiating
    * from template.
    */
    public void setReplaceableStringsTable(String table) {
        String t = getReplaceableStringsTable();
        if (t.equals(table))
            return;
        putProperty(PROP_REPLACEABLE_STRINGS_TABLE, table, true);
    }

    /** Gets the replacable strings table - used during instantiating
    * from template.
    */
    public String getReplaceableStringsTable() {
        String table = (String)getProperty(PROP_REPLACEABLE_STRINGS_TABLE);
        if (table == null) {
            return "USER="+System.getProperty("user.name")+"\n";
        } else {
            return table;
        }
    }

    /** Gets the replaceable table as the Properties class.
    * @return the properties
    */
    public Properties getReplaceableStringsProps() {
        Properties props = new Properties();
        // PENDING: don't use StringBufferInputStream, it does not encode characters
        // well.
        try {
            props.load(new StringBufferInputStream(getReplaceableStringsTable()));
        }
        catch (IOException e) {
        }
        return props;
    }

    /** Gets the delay time for the start of the parsing.
    * @return The time in milis
    */
    public int getAutoParsingDelay() {
        Integer delay = (Integer)getProperty(PROP_AUTO_PARSING_DELAY);
        if (delay == null)
            return DEFAULT_AUTO_PARSING_DELAY;
        return delay.intValue();
    }

    /** Sets the delay time for the start of the parsing.
    * @param delay The time in milis
    */
    public void setAutoParsingDelay(int delay) {
        if (delay < 0)
            throw new IllegalArgumentException();
        putProperty(PROP_AUTO_PARSING_DELAY, new Integer(delay));
    }
    
    /**
    * @param b Class files are scanned for 'source' property iff <tt>true</tt>
    * @deprecated Please use {@link JavaDataLoader#setPerfectRecognition}
    */
    public void setPerfectRecognition(boolean b) {
        ((JavaDataLoader)SharedClassObject.findObject(JavaDataLoader.class, true)).
            setPerfectRecognition(b);
    }
    
    public boolean isCompileStatusEnabled() {
        Boolean b = (Boolean)getProperty(PROP_SHOW_COMPILE_STATUS);
        if (b == null)
            return true;
        return b.booleanValue();
    }
    
    public void enableCompileStatus(boolean b) {
        if (isCompileStatusEnabled() == b)
            return;
        putProperty(PROP_SHOW_COMPILE_STATUS, b ? Boolean.TRUE : Boolean.FALSE, true);
    }
        

    /**
    * @return <tt>true</tt> iff deep parsing of class files should be done
    * @deprecated Please use {@link JavaDataLoader#getPerfectRecognition}
    */
    public boolean isPerfectRecognition() {
        return ((JavaDataLoader)SharedClassObject.findObject(JavaDataLoader.class, true)).isPerfectRecognition();
    }
    
    /**
     * Returns true, if the sources in a directory should be "pre-scanned" when the
     * folder is opened. For details, see {@link #setPrescanSources}
     * @return true, if sources are parsed after their nodes are shown.
     */
    public boolean getPrescanSources() {
        Boolean b = (Boolean)getProperty(PROP_PRESCAN_SOURCES);
        if (b == null)
            return true;
        else
            return b.booleanValue();
    }

    /** Enables or disables "pre-scan" of source files. The prescan operation is
     * scheduled after the node comes out visible somewhere in order to determine file's
     * type, contents and state. The most common trigger operation is opening of a folder
     * in the Explorer. After the prescan completes, UI of Java source file nodes is
     * updated with icons that reflect the exact state of the file - clean, compiled, with errors,
     * executable, ...<BR>
     * Note that the operation may be seen as costly because it requires parsing all the source
     * files attached to the nodes and on some systems it may have adverse impact on IDE's
     * performance. Default = on.
     * @param enable true if pre-scan should be enabled, false if not.
     */
    public void setPrescanSources(boolean enable) {
        Boolean b = (Boolean)getProperty(PROP_PRESCAN_SOURCES);
        if ((b == null && enable) ||
            (b != null && b.booleanValue() == enable))
            return;
        putProperty(PROP_PRESCAN_SOURCES, 
            enable ? Boolean.TRUE : Boolean.FALSE, true);
    }

    /** @return localized string */
    static String getString(String s) {
        if (bundle == null) {
            bundle = NbBundle.getBundle(JavaSettings.class);
        }
        return bundle.getString(s);
    }

    private static JavaSynchronizationSettings getJavaSynchronizationSettings() {
        return (JavaSynchronizationSettings) JavaSynchronizationSettings.findObject(JavaSynchronizationSettings.class, true);
    }

    private static ImpToolSettings getImpToolSettings() {
        return (ImpToolSettings) ImpToolSettings.findObject(ImpToolSettings.class, true);
    }

    /** The method is to filter ServiceTypes according to its parameters.
     * 
     * @param known Is fastjavac binary (e.g. fastjavac.exe) present? If so then ensure that it is listed among the service types. If not then delete FastJavacCompiler type from the list.
     * @param set Is to be a descendand of java.JavaCompilerType set as default for java?
     */
    private static boolean filterCompilerTypes(boolean known) {
        List list = TopManager.getDefault().getServices().getServiceTypes();
        Iterator iter = list.iterator();
        boolean fastJavacRemoved=false;
        
        // set/remove compiler types
        while (iter.hasNext()) {
            ServiceType stype = (ServiceType) iter.next();
            if (stype instanceof CompilerType) {
                CompilerType ret = (CompilerType) stype;
                Class retClass = ret.getClass();
                if (retClass == FastJavacCompilerType.class) {
                    if (known) {
                        // FastJavac is present in the list - OK
                        return fastJavacRemoved;
                    } else {
                        // FastJavac is present in the list - remove it and and setDefaultCompiler
                        iter.remove();
                        fastJavacRemoved=true;
                    }
                }
             }
        }
        iter = null;

        if (known) { // FastJavacCompilerType not installed - install it
            FastJavacCompilerType type = new FastJavacCompilerType();
            list.add(type);
        }
        TopManager.getDefault().getServices().setServiceTypes(list);
        return fastJavacRemoved;
    }
    
    /** Sets default compiler - called from JavaModule */
    public static void setCompiler(boolean restarted) {
        if (filterCompilerTypes(FastJavacCompilerType.isFastJavacPlatform())) {
            JavaSettings js=(JavaSettings)SharedClassObject.findObject(JavaSettings.class, true);
            CompilerType cType=js.getCompiler();
            
            if (cType==null || cType.getClass()==FastJavacCompilerType.class) {
                js.setDefaultCompilerType();
            }
        }
    }
    
    /** Getter for parser class path.
    */
    public NbClassPath getParserClassPath() {
        NbClassPath cp = parserClassPath;
        return cp != null || isWriteExternal() ? cp : new NbClassPath("");
    }
    
    public Collection getParserClassPathFS() {
        return keepClassPath = ClassPathConfigurator.getInstance().getFileSystems(
            getParserClassPath());
    }

    /**
     * Creates and returns bootclasspath from the runtime environment.
     */
    private synchronized NbClassPath getRuntimeBootPath() {
        if (runtimeBootClassPath != null)
            return runtimeBootClassPath;
        return runtimeBootClassPath = new NbClassPath("");
    }

    /** Setter for parser class path.
    */
    public void setParserClassPath(NbClassPath path) {
        NbClassPath old = parserClassPath;
        parserClassPath = path;
        firePropertyChange(PROP_PARSER_CLASSPATH, old, path); // NOI18N
    }
    
    /** Getter for parser boot class path.
    */
    public NbClassPath getParserBootPath() {
        return (parserBootPath == null && !isWriteExternal()
            ? getRuntimeBootPath() : parserBootPath);
    }

    /**
     * Returns a colleciton of FileSystems that should be used by the parser
     * as runtime classes. If nothing is defined by the user, it will default
     * to IDE's own runtime classes.
     */
    public Collection getParserBootPathFS() {
        NbClassPath pbcp = getParserBootPath();
        if (pbcp == null ||
            "".equals(pbcp.getClassPath()))
            pbcp = NbClassPath.createBootClassPath();
        return keepBootPath = ClassPathConfigurator.getInstance().getFileSystems(
            pbcp);
    }

    /** Setter for parser boot class path.
    */
    public void setParserBootPath(NbClassPath path) {
        if (path == getParserBootPath() ||
            (path != null && path.equals(getParserBootPath())))
            // no change.
            return;
        NbClassPath old = parserBootPath;
        parserBootPath = path;
        firePropertyChange(PROP_PARSER_BOOTPATH, old, path); // NOI18N
    }
    
    public void readExternal(ObjectInput in) 
    throws java.io.IOException, ClassNotFoundException {
        super.readExternal(in);
        if (in.available() > 0) {
            version = in.readInt();
        }
        if (version < currentVersion) {
            // assume version == 0, for now
            this.parserBootPath = null;
            this.parserClassPath = null;
            
            version = currentVersion;
        }
    }
    
    public void writeExternal(ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeInt(version);
    }
    
}
