/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.modules.java.tools;

import java.beans.*;
import java.util.*;

import org.openide.nodes.*;
import org.openide.src.*;
import org.openide.src.nodes.*;
import org.openide.util.TaskListener;
import org.openide.util.Task;

class InheritanceChildren extends Children.Keys implements InheritanceSupport.HierarchyListener {
    /** Special `wait' node that signalizes that inheritance infomation was
        not computed yet.
    */
    private static final Object WAIT = new Object();
    
    /** Text that is printed when the hierarchy of the class changes and the
        superclass computation is not completed yet.
    */
    private static final String WAIT_STRING = OverrideAction.getString("MSG_InheritanceWait");
    
    /** Object that collects class' inheritance data
    */
    private InheritanceSupport data;
    
    /** Comparator used to sort inherited classes
    */
    private Comparator comparator;
    
    private InheritanceSupport.HierarchyListener listener;

    /** Creates new InheritanceChildren */
    public InheritanceChildren(InheritanceSupport data, java.util.Comparator comp) {
        super();
        this.data = data;
        this.comparator = comp;
        setWaitNode();
        listener = this;
    }
    
    public void addNotify() {
        data.addHierarchyListener(listener);
        data.prepareHierarchy().addTaskListener(new TaskListener() {
            public void taskFinished(Task t) {
                t.removeTaskListener(this);
                refreshData();
                
                // attach a listener to listen for future changes.
                data.addHierarchyListener(listener);
            }
        });
    }
    
    public void removeNotify() {
        data.removeHierarchyListener(listener);
    }
    
    public void setComparator(Comparator c) {
        this.comparator = c;
        refreshData();
    } 

    private void refreshData() {
        Collection original = data.getInheritedClasses();
        Collection result;
        if (comparator == null) {
            result = original;
        } else {
            result = new TreeSet(comparator);
            result.addAll(original);
        }
        setKeys(result);
    }

    protected Node[] createNodes(Object key) {
        if (key instanceof ClassElement) {
            return new Node[] {
                data.createClassElementNode((ClassElement)key, Children.LEAF, false)
            };
        } else if (key == WAIT) {
            return new Node[] {
                new AbstractNode(Children.LEAF) {
                    public String getDisplayName() {
                        return WAIT_STRING;
                    }
               }
            };
        } else {
            throw new IllegalArgumentException(key.getClass().toString());
        }
    }
    
    private final void setWaitNode() {
        setKeys(new Object[] {
            WAIT
        });
    }
    
    /** Informs that class' inheritance has changed.
     */
    public void inheritanceChanged() {
        setWaitNode();
    }
    
    /** Informs that collection of inherited classes has been recomputed.
     */
    public void inheritedClassesComputed(java.util.Map classes) {
        // refresh the keys with all the classes, sorted by the comparator.
        refreshData();
    }
    
    /** Informs, that the collection of inherited methods has been recomputed.
     */
    public void inheritedMethodsComputed(Collection methods) {
    }
}
