/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.tools;

import java.lang.reflect.Modifier;
import java.lang.ref.*;
import java.util.*;
 
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.src.*;
import org.openide.src.nodes.*;
import org.openide.util.*;
import org.openide.util.SharedClassObject;

import org.netbeans.modules.java.settings.JavaSynchronizationSettings;

/** Support for navigating through the hierarchy of a class.
 * Currently, the object can listen on changes in the class' inheritance hierarchy
 * and collects class and method information asynchronously.
 * Class information consist of a map of accessible (visible) classes that are directly
 * or inidrectly inherited.
 * Method informations records what class(es) declare a particular method, if the method
 * is defined (and in which class). Contains keyed hashmap of all available methods.
 * The class supports some modification operations like creating an overriden method from
 * the given model MethodElement. Parameter names are pseudo-intelligently created if needed.
 */
public class InheritanceSupport extends Object {
    private static ResourceBundle bundle = NbBundle.getBundle(InheritanceSupport.class);

    /* Inspected (target) class
     */
    public static final String PROP_CLASS = "class"; // NOI18N
    
    /** Collection of all inherited classes.
    */
    public static final String PROP_INHERITED_CLASSES = "inheritedClasses"; // NOI18N
    
    /** Collections of all inherited methods.
    */
    public static final String PROP_INHERITED_METHODS = "inheritedMethods"; // NOI18N

    public static final String PROP_INTERFACES = "interfaces";

    
    private static final String JAVA_LANG_OBJECT_FQN = "java.lang.Object"; // NOI18N
    
    /** Qualified identifier of the java.lang.Object class.
     */
    private static Identifier JAVA_LANG_OBJECT = Identifier.create(JAVA_LANG_OBJECT_FQN);

    /** Strings suitable for parameter names for some well-known data types.
     */
    private static String NAME_ARRAY = bundle.getString("NAME_ArrayOfPrimitives"); // NOI18N
    private static String NAME_PARAM = bundle.getString("NAME_Primitive"); // NOI18N
    private static String NAME_OBJECT_VALUE = bundle.getString("NAME_javaLangValue"); // NOI18N
    private static String NAME_OBJECT = bundle.getString("NAME_javaLangObject"); // NOI18N
    private static String NAME_STRING = bundle.getString("NAME_javaLangString"); // NOI18N
    private static String NAME_CLASS_VALUE = bundle.getString("NAME_javaLangClass"); // NOI18N
    /** Mapping from well-known parameter types to appropriate names.
     */
    private static Map wellKnownClasses;

    static {
        // TODO: read well-known classes from some configuration file or
        // user settings accessible from the GUI
        wellKnownClasses = new HashMap();
        wellKnownClasses.put("java.lang.Object", NAME_OBJECT); // NOI18N
        wellKnownClasses.put("java.lang.String", NAME_STRING); // NOI18N
        wellKnownClasses.put("java.lang.Float", NAME_OBJECT_VALUE); // NOI18N
        wellKnownClasses.put("java.lang.Double", NAME_OBJECT_VALUE); // NOI18N
        wellKnownClasses.put("java.lang.Long", NAME_OBJECT_VALUE); // NOI18N
        wellKnownClasses.put("java.lang.Class", NAME_CLASS_VALUE); // NOI18N
    }

    /** Class that the Wizard operates on. All queries are run againt this class
        element unless stated otherwise.
    */
    ClassElement clazz;

    /** Package name of the class, mainly for accessibility checks
    */
    String  clazzPackage;

    /** Maps fully qualified names of classes/interfaces to their corresponding
        Elements for all inherited superclasses/superinterfaces.
        Note that these are mostly fixed for the class and change only if
        class' superclass changes.
    */
    Map   superInherited;

    /** Map of tree of interfaces implemented on the clazz itself that do not
        appear in the superclass' inheritance tree.
        Interfaces are more likely to change and their inheritance chain
        won't be so deep
    */
    Map   allInherited;

    /** Map of all inherited methods keyed by MethodElement.Key objects. Values
     *  are methods collected from all classes and superclasses.
     *  If more than one superclass or superinterface defines a method, then
     *  the first one encountered will be recorded -- this in effect favor
     *  method definition before abstract declarations.
     */
    Map   inheritedMethods;
    
    /** Map of class inheritance info
     */
    Map   classInheritanceInfo;

    /** Cache of classes uses during the course. This is only for performance
        only. Users of the WizardData should not rely on classes of the same
        name being the same instances.
    */
    Map   classCache = new HashMap(17);

    /** This set contains `faked' symbols that have been created, but do not
        really exist in the repository or the CLASSPATH.
        Users of this object might want to distinguish real classes and faked ones.
    */
    Set   fakeSymbols = new HashSet(19);

    /** Interfaces implemented on the class. The listener checks against this
        array and invalidates inheritance data, if necessary.
    */
    Set   implementedInterfaces;

    /** Task that collects class hierarchy information.
     */
    Task   hierarchyTask;
    
    /** Task that collects information about methods inherited from superclasses
     * and superinterfaces.
     */
    Task   methodTask;
    
    private PropertyChangeListener classPropertyListener;
    private PropertyChangeListener listenerHook;
    
    private static final Task FINISHED = Task.EMPTY;

    /** True, if the superclass was not changed from the last inspection.
     * When true, only directly implemented interfaces are scanned for additional
     * superinterfaces.
     */
    private boolean superIsValid;
    
    private boolean interfacesAreValid;
    
    /** If true, inherited method informations is up-to-date.
     */
    private boolean methodsAreValid;

    /** Request processor that performs hierarchy queries asynchronously.
     */
    private static RequestProcessor wizardRP = new RequestProcessor("JavaWizard query");

    /** Utility field used by event firing mechanism. */
    private java.util.ArrayList listenerList =  null;

    private static JavaSynchronizationSettings syncSettings;
    
    /* ================================================================================== */
    
    public InheritanceSupport() {
    }
    
    /** Creates new WizardData */
    public InheritanceSupport(ClassElement source) {
        setSourceClass(source);
        prepareHierarchy();
    }
    
    /** Changes the inspected class.
     * The object will then fire a series of property changes on interfaces, inheritance
     * in general and on methods, too.
     */
    public void setSourceClass(ClassElement source) {
        ClassElement oldClass = this.clazz;
        
        this.clazz = source;
        this.clazzPackage = source.getName().getQualifier();
        classPropertyListener = WeakListener.propertyChange(listenerHook = new PropL(), source);
        
        if (oldClass != null) {
            oldClass.removePropertyChangeListener(classPropertyListener);
        }
        source.addPropertyChangeListener(classPropertyListener);
        prepareHierarchy();
    }
    
    /** Returns the inspected class for this support.
     */
    public ClassElement getSourceClass() {
        return clazz;
    }
    
    /** forcibly sets the class package. If used, it can affect class/interface
     *  element lookups.
     */
    public void setPackage(String packName) {
        clazzPackage = packName;
	invalidateInheritanceSupport();
    }

    /** Determines if the symbol denoted by the name and the inspected class
     * belongs to the same package.
     * This does not work for inner classes, however.
     */
    public boolean isSamePackage(Identifier name) {
        return name.getQualifier().equals(clazzPackage);
    }
    
    public boolean isSamePackage(ClassElement element) {
        ClassElement outer;
        
        while (true) {
            outer = element.getDeclaringClass();
            if (outer == null)
                break;
            element = outer;
        }
        return isSamePackage(element.getName());
    }

    private synchronized void resetSymbols() {
        classCache.values().removeAll(fakeSymbols);
        fakeSymbols.clear();
    }

    /** Prepares wizard's data in advance. Add a TaskListener to the returned
        task, if you want to be notified about the data completion.
    */
    public synchronized Task prepareHierarchy() {
        if (hierarchyTask != null) {
            return hierarchyTask;
         } else if (!(superIsValid && interfacesAreValid)) {
            return hierarchyTask = wizardRP.post(new HierarchyProducer());
         } else {
            return FINISHED;
         }
    }

    /** Requests preparation of method list only.
     */
    public synchronized Task prepareMethods() {
        if (methodTask != null) {
            return methodTask;
         } else if (!methodsAreValid) {
            return methodTask = wizardRP.post(new MethodListProducer());
         } else {
            return FINISHED;
         }
    }

    /** Returns all inherited (and accessible) classes.
        Note: this call can take SIGNIFICANT amount of time so please execute that
        in other than UI thread.
        Note 2: this returns ONLY classes inherited through the superclass.
        The map does not contain directly implemented interfaces (and their superclasses).
    */
    public Map getInheritedFromSuper() {
        Map m = superInherited;
        if (m != null) 
            return m;
        prepareHierarchy().waitFinished();
        return superInherited;
    }

    /** Returns a map of all inherited classes.
     * The map has a form of <fully qualified name> -> ClassElement
     */
    public Map getAllInherited() {
        Map m = allInherited;
        if (m != null) 
            return m;
        prepareHierarchy().waitFinished();
        return allInherited;
    }        

    /** Returns a collection of all inherited classes.
     * Basically the function returns values of map obtained from {@link getAllInherited}
     */
    public Collection getInheritedClasses() {
        return getAllInherited().values();
    }
    
    public Map getInheritedMethodMap() {
        Map m = inheritedMethods;
        if (m != null) 
            return m;
        prepareMethods().waitFinished();
        return inheritedMethods;
    }
    
    public Collection getInheritedMethods() {
        return getInheritedMethodMap().values();
    }

    /** Determines, if the given method is accessible from the source class.
     * Currently, the method is meant to test methods available thorugh inheritance only.
     */
    boolean isAccessibleMethod(MethodElement m) {
        ClassElement cls = m.getDeclaringClass();
        int modifs;
        
        modifs = cls.getModifiers();
        if ((modifs & Modifier.PRIVATE) != 0 ||
            ((modifs & (Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE)) == 0 &&
             !isSamePackage(cls.getName()))) {
            return false;
        }
        modifs = m.getModifiers();
        if ((modifs & Modifier.PRIVATE) != 0)
            return false;
        if ((modifs & (Modifier.PUBLIC | Modifier.PROTECTED)) != 0)
            return true;
        return isSamePackage(cls.getName());
    }

    /** Determines, if a class is accessible in the source class.
     */
    public boolean isAccessibleClass(ClassElement el) {
        return true;
        /*
        return ((el.getModifiers() & (Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE)) != 0 ||
            isSamePackage(el.getName()));
         */
    }
    
    private synchronized void invalidateInheritanceSupport() {
        superIsValid = false;
        invalidateDirectInheritance();
    }

    private synchronized void invalidateDirectInheritance() {
        interfacesAreValid = false;
        methodsAreValid = false;
        fireInheritanceChanged();
        prepareHierarchy();
    }

    /** Finds appropriate element for an interface. If the interface can't
        be found using normal lookup and createEmpty is true, the method
        creates an empty accessible interface with the requested name and
        returns its representation. In this case, the element is known to be
        `faked' and isFaked(thatElement) returns true.
        @param ident identifier of the interface; if it doesn't contain fully
            qualified name, IllegalArgumentException is thrown.
        @praram createEmpty if true, the method creates interface that are not
            accessible through normal lookup services.
        @return Element for the interface, or null if the element can't be
            looked up and/or created.
    */
    public synchronized ClassElement getInterface(Identifier ident, boolean createEmpty) {
        return lookupClass(ident, false, createEmpty);
    }

    public synchronized ClassElement getClass(Identifier ident, boolean createEmpty) {
        return lookupClass(ident, true, createEmpty);
    }
    
    /** Returns all methods on the current class that override method(s) in its
     * superclasses.
     */
    public Collection getOverridingMethods() {
        Collection c = new LinkedList();
        Collection methods = Arrays.asList(clazz.getMethods());
        Iterator it = methods.iterator();
        while (it.hasNext()) {
            MethodElement m = (MethodElement)it.next();
            if (overrides(m))
                c.add(m);
        }
        return c;
    }

    public boolean overrides(MethodElement m) {
	MethodInfo info = (MethodInfo)getInheritedMethodMap().get(new MethodElement.Key(m));
	if (info == null)
	    return false;
	return info.declared.size() > (info.declared.contains(clazz.getName().getFullName()) ? 1 : 0);
    }

    //================================ IMPLEMENTATION ===================================
    /** Looks up ClassElement identifier by some identifier; if the element can't be found
        using the Repository, the element is constructed if create is true.
    */
    private ClassElement lookupClass(Identifier id, boolean isClass, boolean create) {
        ClassElement el = null;
        Reference ref;
        String fullName = id.getFullName();

        ref = (Reference)classCache.get(fullName);
        if (ref != null) {
            el = (ClassElement)ref.get();
        }
        if (el != null) {
            return el;
        }
        el = ClassElement.forName(fullName);
        if (el == null) {
            if (!create) {
                return null;
            }
            el = createElement(id, isClass);
        }
        classCache.put(id.getFullName(), new SoftReference(el));
        return el;
    }

    private ClassElement createElement(Identifier id, boolean isClass) {   
        ClassElement el;
        el = new ClassElement();
        try {
            el.setClassOrInterface(isClass);
            el.setName(id);
        } catch (SourceException e) {
            // should not happen.
        }
        return el; 
    }

    /** 
    */
    private Map buildSuperHierarchy() {
        Map inh = new HashMap(17);
        Set excl = new HashSet();
        Map inhInfo = new HashMap(23);
	
        ClassElement sup;
        superIsValid = true;
        if (clazz.getSuperclass() != null) {
            sup = getClass(clazz.getSuperclass(), true);
            inh.put(clazz.getSuperclass().getFullName(), sup);
            addInheritedClasses(sup, inh, excl, inhInfo);
            synchronized(this) {
                this.implementedInterfaces = new HashSet(Arrays.asList(clazz.getInterfaces()));
            }
	} else {
            inh = new HashMap();
        }
	if (clazz.isClassOrInterface() && !inh.containsKey(JAVA_LANG_OBJECT_FQN)) {
	    // the inspected element is a class, but no java.lang.Object is present in the inheritance
	    // data -- correct that    
	    sup = getClass(JAVA_LANG_OBJECT, true);
	    inh.put(JAVA_LANG_OBJECT_FQN, sup);
	    inhInfo.put(JAVA_LANG_OBJECT_FQN, new ClassInfo(sup));
	}
        this.superInherited = inh;
        this.classInheritanceInfo = inhInfo;
        return inh;
    }

    /** Uses inheritance tree created for the superclass and creates a trees for
        interfaces implemented on the class itself.
    */
    private Map buildDirectHierarchy() {
        Map base = buildSuperHierarchy();
        Map direct = new HashMap(base.size());
        Set exclusions = new HashSet();
        
        direct.putAll(base);
        Identifier[] impls = clazz.getInterfaces();
        interfacesAreValid = true;
        for (int i = 0; i < impls.length; i++) {
            if (base.containsKey(impls[i].getFullName())) {
                continue;
            }
            ClassElement el = getInterface(impls[i], true);
            direct.put(impls[i].getFullName(), el);
            addInheritedClasses(el, direct, exclusions, classInheritanceInfo);
        }
        this.allInherited = direct;
        fireInheritanceChanged();
        return direct;
    }

    /** Adds everything that is accessible through inheritance from the class
        `cls' to the `inheritanceMap', but not cls itself.
    */
    ClassInfo addInheritedClasses(ClassElement cls, Map inheritanceMap, Set exclusions, Map classInfo) {
        Identifier superId = cls.getSuperclass();
        ClassElement inherited;
        String superName = null;
        ClassInfo info;

        if (!superIsValid)
            return new ClassInfo(cls);
        info = new ClassInfo(cls);
        classInfo.put(cls.getName().getFullName(), info);
        if (superId != null && !superId.getFullName().equals("") &&
            !inheritanceMap.containsKey(superName = superId.getFullName()) &&
            !exclusions.contains(superName)) {
            inherited = getClass(superId, true);
            if (inherited != null && isAccessibleClass(cls)) {
                exclusions.add(superName);
                info.addSuperclass(superId);
                ClassInfo superInfo = addInheritedClasses(inherited, inheritanceMap, exclusions, classInfo);
                inheritanceMap.put(superName, inherited);
                info.addSuperclasses(superInfo.getSuperclasses());
            }
        }
        Identifier[] interfaces = cls.getInterfaces();
        if (interfaces != null && interfaces.length > 0) {
            for (int i = 0; i < interfaces.length; i++) {
                superId = interfaces[i];
                if (inheritanceMap.containsKey(superName = superId.getFullName()) ||
                    exclusions.contains(superName)) {
                    continue;
                }
                inherited = getInterface(superId, true);
                if (inherited == null || !isAccessibleClass(inherited)) {
                    continue;
                }
                inheritanceMap.put(superName, inherited);
                info.addSuperclass(superId);
                info.addSuperclasses(addInheritedClasses(inherited, inheritanceMap, exclusions, 
                    classInfo).getSuperclasses());
            }
        }
        return info;
    }

    private void addMethods(ClassElement clazz, Map result, Set traversed) {
        MethodElement[] methods = clazz.getMethods();
        MethodElement.Key k;

        if (!isAccessibleClass(clazz))
            return;
        for (int i = 0; i < methods.length; i++) {  
            MethodElement m = methods[i];
            int modifiers = m.getModifiers();
            
            // optimization: isAccessibleMethod tests accessibility of a class
            // which is now known.
            if (Modifier.isPrivate(modifiers) ||
                ((modifiers & (Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE)) == 0 &&
                 !isSamePackage(clazz.getName())))
                 continue;
            boolean declOnly = Modifier.isAbstract(modifiers) || !clazz.isClassOrInterface();
            k = new MethodElement.Key(methods[i]);
            if (!result.containsKey(k)) {
                result.put(k, new MethodInfo(methods[i], k, clazz));
            } else {
                ((MethodInfo)result.get(k)).addClass(clazz, declOnly);
            }
        }
        traversed.add(clazz.getName().getFullName());
        Identifier id = clazz.getSuperclass();
        if (id == null && !clazz.getName().equals(ClassElement.ROOT_OBJECT)) {
            // we've at the top of the hierarchy, but don't have explicitly
            // declared extends java.lang.Object there.
            id = ClassElement.ROOT_OBJECT;
        }
        if (id != null && !traversed.contains(id.getFullName())) {
            ClassElement s = getClass(id, true);
            addMethods(s, result, traversed);
        }

        Identifier[] ids = clazz.getInterfaces();
        for (int i = 0; i < ids.length; i++) { 
            if (traversed.contains(ids[i].getFullName()))
                continue;
            ClassElement itf = getInterface(ids[i], true);
            addMethods(itf, result, traversed);
        }
    }

    private void buildMethodList() {
        Set traversed = new HashSet(17);
        Map result = new HashMap(23);

        addMethods(this.clazz, result, traversed);
        Identifier[] ids = clazz.getInterfaces();
        for (int i = 0; i < ids.length; i++) { 
            if (traversed.contains(ids[i].getFullName()))
                continue;
            ClassElement itf = getInterface(ids[i], true);
            addMethods(itf, result, traversed);
        }
        synchronized (this) {
            methodsAreValid = true;
            inheritedMethods = result;
        }
    }

    /** Registers HierarchyListener to receive events.
     * @param listener The listener to register.
     */
    public synchronized void addHierarchyListener(HierarchyListener listener) {
        if (listenerList == null ) {
            listenerList = new java.util.ArrayList();
        }
        listenerList.add (listener);
    }
    
    /** Removes HierarchyListener from the list of listeners.
     * @param listener The listener to remove.
     */
    public synchronized void removeHierarchyListener(HierarchyListener listener) {
        listenerList.remove (listener);
    }

    /** Fires an event informing listeners that inheritance dependecies of the target
        class have changed.
    */
    private void fireInheritanceChanged() {
        Collection c;
        Iterator it;
        Map m = allInherited;
        
        synchronized(this) {
            if (listenerList == null)
                return;
            c = (Collection)listenerList.clone();
        }
        it = c.iterator();
        while (it.hasNext()) {
            HierarchyListener l = (HierarchyListener)it.next();
            l.inheritedClassesComputed(m);
        }
    }
    
    private void fireInheritanceComputed() {
        Collection listeners;
        Iterator it;
        Map hierarchy;
        
        synchronized(this) {
            if (listenerList == null)
                return;
             listeners = (Collection)listenerList.clone();
        }
        it = listeners.iterator();
        while (it.hasNext()) {
            HierarchyListener l = (HierarchyListener)it.next();
            l.inheritanceChanged();
        }
    }
    
    public Node createClassElementNode(ClassElement el, Children ch, boolean writeable) {
        // TODO: create a filter node with a question mark icon if the class is listed in
        // `faked' classes.
        return new ClassElementNode(el, ch, writeable);
    }
    
    private class ClassInfo {
        Set          allInherited;
        ClassElement element;
        
        ClassInfo(ClassElement cls) {
            this.element = cls;
        }
        
        void addSuperclasses(Collection superclasses) {
            if (allInherited == null) {
                allInherited = new HashSet(superclasses.size());
            }
            allInherited.addAll(superclasses);
        }
        
        void addSuperclass(Identifier superId) {
            if (allInherited == null) {
                allInherited = new HashSet(7);
            }
            allInherited.add(superId);
        }
        
        Set getSuperclasses() {
            if (allInherited == null) {
                allInherited = new HashSet();
             }
            return allInherited;
        }
    }

    private class MethodInfo {
        /** Key for the method.
         */
        private MethodElement.Key  key;
        public MethodElement  method;
        // classes the method is declared in
        public Set declared;
        // class that defines the method.
        public ClassElement defined;
	
	public boolean overriden;

        MethodInfo(MethodElement m, MethodElement.Key k, ClassElement cl) {
            this.key = k;
            this.method = m;
            this.declared = new HashSet(17);
            if (cl != null)
                addClass(cl, Modifier.isAbstract(m.getModifiers()) || !cl.isClassOrInterface());
        }

        /** Only the outer class needs to call this method.
         */
        private void addClass(ClassElement cls, boolean declOnly) {
            this.declared.add(cls.getName().getFullName());
            if (!declOnly && defined == null) {
                defined = cls;
            }
        }
        
        public MethodElement.Key getKey() {
            return this.key;
        }
        
        public Set getDeclarations() {
            return declared;
        }
        
        public ClassElement getDefinition() {
            return defined;
        }
    }

    /*****************************************************************************/

    /** Helper class that creates method list for the hierarchy subtree
     */
    private class MethodListProducer implements Runnable {
        public void run() {
            if (!methodsAreValid) {
                buildMethodList();
            }
            synchronized(InheritanceSupport.this) {
                methodTask = null;
            }
        }
    }

    /** Helper class that produces superclass hierarchy.
     */
    private class HierarchyProducer implements Runnable {
        public void run() {
            boolean somethingChanged = false;
            // first try to access all superclasses and superinterfaces.
            while (true) {
                if (!superIsValid) {
                    buildSuperHierarchy();
                    somethingChanged = true;
                }
                if (!interfacesAreValid) {
                    buildDirectHierarchy();
                    somethingChanged = true;
                 }
                 synchronized (InheritanceSupport.this) {
                    if (superIsValid && interfacesAreValid) {
                        hierarchyTask = null;
                        if (somethingChanged) 
                            prepareMethods();
                        return;
                    }
                 }
            }
        }
    }
    
    /** Complete parameter declaration in a method element.
      In case of name clash, the method adds order number to the end of the name.
     */
    public static MethodElement completeParameters(MethodElement el, boolean cloned) {
        MethodParameter[] params = el.getParameters();
        Set synthesized = new HashSet();

        if (params == null || params.length == 0) return el;
        for (int i = 0; i < params.length; i++) {
            MethodParameter p = params[i];
            if (p.getName() == null || p.getName().toString().equals("")) {
                if (!cloned) {
                    el = (MethodElement)el.clone();
                    params = el.getParameters();
                    p = params[i];
                    cloned = true;
                }
                String s;
                s = createName(p.getType());
                if (synthesized.contains(s)) {
                    s = s + i;
                } else {
                    synthesized.add(s);
                }
                p.setName(s);
            }
	    // remove final flag from parameters, if there is one:
	    if (p.isFinal()) {
		if (!cloned) {
		    el = (MethodElement)el.clone();
		    params = el.getParameters();
		    p = params[i];
		    cloned = true;
		}
		p.setFinal(false);
	    }
        }
        return el;
    }

    /** Synthesizes suitable name for type `t'.
      If the type is an array, the name will be based on its element's type.
      If the type is primitive, the name will be `param' or `value'.
      In case of Reference types, the name will be constructed from the simple
      name of the type. If the simple name is composed from several words using
      uppercase (titlecase) letters, the last such name component will be taken.
      The returned name always begins with lowercase letter.
     */
    private static String createName(Type t) {
        String lastComponent;
        boolean array = false;

        if (t.isArray()) {
            array = true;
            do {
                t = t.getElementType();
            } while (t.isArray());
        }
        if (t.isPrimitive()) {
            if (array) {
                return NAME_ARRAY;
            } else {
                return NAME_PARAM;
            }
        }

        int i;
        String fullName = t.getClassName().getFullName();

        if (wellKnownClasses.containsKey(fullName)) {
            return (String)wellKnownClasses.get(fullName);
        }
        lastComponent = t.getClassName().getName();
        for (i = lastComponent.length() - 1; i > 0; i--) {
            if (Character.isTitleCase(lastComponent.charAt(i))) {
                break;
            }
        }
        String n;
        if (i < lastComponent.length() - 1) {
            n = Character.toLowerCase(lastComponent.charAt(i)) + lastComponent.substring(i + 1);
        } else {
            n = lastComponent.toLowerCase();
        }
        if (Utilities.isJavaIdentifier(n))
            return n;
        // it is an identifier because it is a class' name. But it is not a java identifier
        // ==> it is a keyword append a Param which will make the string just an ordinary
        // identifier.
        return java.text.MessageFormat.format(bundle.getString("NAME_typeParam"), new Object[] {
            n
        });
    }
    
    /** Creates an overriding method for the given model in the target class.
     * The function duplicates method header, but creates empty body.
     */
    public MethodElement overrideMethod(MethodElement source) throws SourceException {
        return overrideMethod(source, false);
    }
    
    public MethodElement overrideMethod(MethodElement source, boolean createSuper) throws SourceException {
        return overrideMethod(getSourceClass(), source, createSuper, isImplemented(source));
    }
    
    public static MethodElement overrideMethod(ClassElement target, MethodElement source, 
        boolean createBody, boolean superExists) throws SourceException {
        MethodElement m = completeParameters((MethodElement)source.clone(), true);

	// remove ABSTRACT and NATIVE since the user probably wants to write the method body ;-)
	// SYNCHRONIZED is removed for safety reasons: if the user wants the method to be synchronized,
	// she will add the modifier.
        int mods = m.getModifiers() & ~(Modifier.NATIVE | Modifier.ABSTRACT | Modifier.SYNCHRONIZED);
        if (source.getDeclaringClass() != null &&
            !source.getDeclaringClass().isClassOrInterface())
            mods = mods | Modifier.PUBLIC;
        m.setModifiers(mods);
            
        // to be formatted nicely in the source code...
        if (!createBody) {
            m.setBody("\n");
        } else {
            m.setBody(createSuperCall(m, superExists));
        }
        target.addMethod(m);
        
        MethodParameter[] params = m.getParameters();
        Type[] types = new Type[params.length];
        for (int i = 0; i < types.length; i++) {
            types[i] = params[i].getType();
        }
        
        MethodElement inserted = target.getMethod(m.getName(), types);
        return inserted;
    }
    
    protected static String createSuperCall(MethodElement target, boolean hasSuper) {
	if (syncSettings == null) {
	    syncSettings = (JavaSynchronizationSettings)SharedClassObject.findObject(JavaSynchronizationSettings.class, true);
	}

        String body;
        
        if (hasSuper) {
            // create a parameter list for the superclass' call:
            MethodParameter[] params = target.getParameters();
            String format;
            
            StringBuffer str = new StringBuffer();
            for (int i = 0; i < params.length; i++) {
                if (str.length() > 0)
                    str.append(", "); // NOI18N
                str.append(params[i].getName());
            }
            if (Type.VOID.equals(target.getReturn())) {
                format = bundle.getString("FMT_CallSuper"); // NOI18N
            } else {
                format = bundle.getString("FMT_ReturnCallSuper"); // NOI18N
            }
            body = java.text.MessageFormat.format(format,
                new Object[] {
                    target.getReturn(),
                    target.getName(),
                    str
            });
        } else {
            // no super, let JavaSyncSettings generate a default return:
            body = syncSettings.getGenerateReturnAsString(target.getReturn());
        }
        return body;
    }
    
    /** Retrieves class information for a particular class.
     */
    ClassInfo getClassInfo(ClassElement e) {
        ClassInfo info = (ClassInfo)classInheritanceInfo.get(e.getName().getFullName());
        if (info == null) {
            return new ClassInfo(e);
        }
        return info;
    }

    /** Retrieves method information for a particular method.
     */
    MethodInfo getMethodInfo(MethodElement m) {
        MethodElement.Key mk = new MethodElement.Key(m);
        MethodInfo minfo = (MethodInfo)getInheritedMethodMap().get(mk);
        if (minfo == null) {
            return new MethodInfo(m, mk, null);
        }
        return minfo;
    }

    /** Determines, if the method is implemented in the source class.
     * This test is currently narrowed to check if there's a method in some of
     * the inherited classes that is not declared abstract.
     */
    boolean isImplemented(MethodElement m) {
        MethodInfo info = getMethodInfo(m);
	ClassElement def = info.getDefinition();
	if (def == null) {
	    return false;
	}
	return def != clazz;
    }
    
    /** Tests if the particular method is appears in the hierarchy tree in the 
     * specified class for the first time (it is not declared in any of the 
     * superclasses of e.
     */
    boolean isDeclaredFirst(MethodElement m, ClassElement e) {
        Set ancestors = getClassInfo(e).getSuperclasses();
        Set declarations = getMethodInfo(m).getDeclarations();
        Iterator it = declarations.iterator();
        while (it.hasNext()) {
            if (ancestors.contains(it.next()))
                return false;
        }
        return true;
    }
    
    private void collectClassMethods(ClassElement cls, HashMap result, ElementFilter.Method f) {
        MethodElement[] methods = cls.getMethods();
        for (int i = 0; i < methods.length; i++) {
            MethodElement m = methods[i];
            MethodElement.Key key = new MethodElement.Key(m);
            if (result.containsKey(key))
                continue;
            if (f.accept(m)) 
                result.put(key, m);
        }
    }
    
    private void collectClassMethods(ClassElement root, HashMap result, ElementFilter.Method f,
        boolean recurse, Set traversed) {
        String fullClassName = root.getName().getFullName();
        if (traversed.contains(fullClassName))
            return;
        traversed.add(fullClassName);
        
        // collect the root class' methods first:
        collectClassMethods(root, result, f);
        Identifier superId = root.getSuperclass();
        ClassElement sc;
        
        if (!recurse)
            return;
        if (superId != null && !"".equals(superId.getFullName())) {
            sc = getClass(superId, true);
            collectClassMethods(sc, result, f, recurse, traversed);
        }
        
        Identifier[] interfaceIds = root.getInterfaces();
        for (int i = 0; i < interfaceIds.length; i++) {
            superId = interfaceIds[i];
            if (root.isInterface() &&
                JAVA_LANG_OBJECT.equals(superId.getFullName()))
                continue;
            sc = getInterface(superId, true);
            collectClassMethods(sc, result, f, recurse, traversed);
        }
    }
    
    Collection collectMethods(Collection rootClasses, boolean deep,
        ElementFilter.Method filter) {
        return createMethodMap(rootClasses, deep, filter).values();
    }
    
    public Map createMethodMap(Collection rootClasses, boolean deep,
        ElementFilter.Method filter) {
    
        HashMap result;
        Set traversed = new HashSet();
        
        result = new HashMap(17);
        for (Iterator it = rootClasses.iterator(); it.hasNext(); ) {
            collectClassMethods((ClassElement)it.next(), result, filter, deep,
                traversed);
        }
        return result;
    }
    
    /** Determines if the given method can be overriden in the source class.
     * The method checks accesibility of the method and final modifier.
     */
    public boolean isOverridable(MethodElement m) {
        if (!isAccessibleMethod(m))
            return false;
        int mods = m.getModifiers();
        return !(Modifier.isFinal(mods) || Modifier.isStatic(mods));
    }

    /** Interface that keeps classes informed about inheritance changes for the
        target class.
        Implementing this interface and listening on WizardData's events gives roughly
        the same results as listening on target class' PROP_INTERFACES and 
        PROP_SUPERCLASS properties, but since the information are forwarded through
        a common point, synchronization problems do not occur if listening on
        WizardData.
    */
    public interface HierarchyListener extends java.util.EventListener {
        /** Informs that class' inheritance has changed.
        */
        public void inheritanceChanged();
        
        /** Informs that collection of inherited classes has been recomputed.
        */
        public void inheritedClassesComputed(Map classes);
        
        /** Informs, that the collection of inherited methods has been recomputed.
        */
        
        public void inheritedMethodsComputed(Collection methods);
    }
    
    private class PropL implements PropertyChangeListener {
        public void propertyChange(final PropertyChangeEvent evt) {
            String propertyName = evt.getPropertyName();

            if (propertyName.equals(ElementProperties.PROP_SUPERCLASS)) {
                invalidateInheritanceSupport();
            } else if (propertyName.equals(ElementProperties.PROP_INTERFACES)) {
                if (checkInterfaces(clazz.getInterfaces())) {
                    invalidateDirectInheritance();
                }
            } else if (ElementProperties.PROP_NAME.equals(propertyName)) {
		// test whether the package has changed:
		Identifier newName = (Identifier)evt.getNewValue();
		String packName = newName.getQualifier();
		if (!packName.equals(clazzPackage)) {
		    setPackage(packName);
		}
	    }
        }

        private boolean checkInterfaces(Identifier[] interfaces) {
            int found = 0;
            
            if (implementedInterfaces == null)
                return interfaces.length != 0;
            for (int i = 0; i < interfaces.length; i++) {
                if (!implementedInterfaces.contains(interfaces[i])) {
                    return true;
                }
                found++;
            }
            return found != implementedInterfaces.size();
        }
    }
}
