/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.tools;
 
import java.awt.BorderLayout;
import java.awt.Insets;
import java.beans.*;
import java.util.*;
import java.lang.reflect.Modifier;
import java.awt.*;
import javax.swing.event.ChangeListener;
import javax.swing.*;
import javax.swing.border.*;

import org.openide.explorer.ExplorerManager;
import org.openide.nodes.Node;
import org.openide.src.*;
import org.openide.*;

public class OverridePanelBase extends javax.swing.JPanel {
    private static final Identifier JAVA_LANG_OBJECT = Identifier.create("java.lang.Object"); // NOI18N
    /** Data internally held by the Java Wizard
     */
    private InheritanceSupport support;

    /** Visual representation of a superclass list.
     */
    private SuperclassList superclassList;
    
    private ListCellRenderer selectedRenderer;

    /** Filtering amongst the overriding methods.
     */
    private ElementFilter.Method overrideFilter;
    
    /** Comparator for sorting methods in the overriding list.
     */
    private Comparator overrideComparator;
    
    /** Model for the override list.
     */
    private DefaultListModel overrideModel;
    
    private int buttonPanelWidth = -1;
    
    private boolean initialized;
    
    protected static final ElementFormat FORMAT_AVAILABLE = 
        new ElementFormat(OverrideAction.getString("FMT_DefaultMethodFormat"));
    
    /** Creates new form OverridePanel */
    public OverridePanelBase(InheritanceSupport support) {
        this.support = support;
    }
    
    private static ResourceBundle getBundle() {
	return OverrideAction.getBundle();
    }
    
    protected void initializeGUI() {
        if (initialized)
            return;
        initComponents ();
        addEtchedBorder(firstPanel, OverrideAction.getString("LAB_InheritedClasses"));
        addEtchedBorder(secondPanel, OverrideAction.getString("LAB_OverridableMethods"));
        addEtchedBorder(thirdPanel, OverrideAction.getString("LAB_SelectedMethods"));
        showOptions.setBorder(new javax.swing.border.CompoundBorder(
        new javax.swing.border.TitledBorder(getBundle().getString("LAB_Options")),
        new javax.swing.border.EmptyBorder(new java.awt.Insets(5, 5, 5, 5))));
        
        superclassList = new SuperclassList();
        superclassHolder.add(superclassList, BorderLayout.CENTER);
        superclassList.getExplorerManager().addPropertyChangeListener(
            new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent evt) {
                    String name = evt.getPropertyName();
                    if (ExplorerManager.PROP_SELECTED_NODES.equals(name))
                        refreshAvailableMethods();
                    }
            });
        
        selectedRenderer = new ElementRenderer(getAvailableMethodsFormat());
        overrideFilter = new DefaultMethodFilter();
        overrideComparator = new DefaultMethodComparator();
        overrideList.setCellRenderer(selectedRenderer);
        availableList.setCellRenderer(selectedRenderer);
        
        Dimension d = jPanel3.getPreferredSize();
        buttonPanelWidth = d.width;
        d = jPanel4.getPreferredSize();
        if (buttonPanelWidth < d.width)
            buttonPanelWidth = d.width;
        d = jPanel2.getPreferredSize();
        if (buttonPanelWidth < d.width)
            buttonPanelWidth = d.width;
        initialized = true;
    }
    
    public boolean isGenerateSuperCall() {
        return generateSuper.isSelected();
    }
    
    protected ElementFormat getAvailableMethodsFormat() {
        return FORMAT_AVAILABLE;
    }
    
    public void addNotify() {
        superclassList.setData(support);
        overrideList.setModel(createOverrideModel());
        super.addNotify();
    }
    
    protected InheritanceSupport getSupport() {
        return this.support;
    }
    
    protected void setSupport(InheritanceSupport support) {
        this.support = support;
    }
    
    private void addEtchedBorder(JPanel p, String title) {
        TitledBorder titled = new TitledBorder(title);
        titled.setBorder(new EtchedBorder(EtchedBorder.LOWERED));
        p.setBorder(new CompoundBorder(titled, new EmptyBorder(new Insets(8, 8, 8, 8))));
    }

    protected ListModel createOverrideModel() {
        return overrideModel = createFilterModel(support.getOverridingMethods(), 
            overrideFilter, overrideComparator);
    }
    
    protected Comparator getAvailableMethodsComparator() {
        return overrideComparator;
    }
    
    protected ElementFilter.Method getAvailableMethodsFilter(final boolean includeInherited,
            final boolean filterAbstract) {
        if (includeInherited) {
            return new ElementFilter.Method() {
                public boolean accept(MethodElement m) {
                    boolean ok = support.isOverridable(m);
                    if (filterAbstract) {
                        ok &= !support.isImplemented(m);
                    }
                    return ok;
                }
            };
        } else {
            return new ElementFilter.Method() {
                public boolean accept(MethodElement m) {
                    boolean ok = support.isOverridable(m) &&
                            support.isDeclaredFirst(m, m.getDeclaringClass());
                    if (filterAbstract) {
                        ok &= !support.isImplemented(m);
                    }
                    return ok;
                }
             };
        }
    }
    
    protected void refreshAvailableMethods() {
        ElementFilter.Method f;
        boolean dive = includeInherited.isSelected();

        f = getAvailableMethodsFilter(dive,
            onlyAbstract.isSelected());

        // Collect methods that match the filter
        Collection selected;
        Iterator it;
        Collection methods;
        
        selected = superclassList.getSelectedClasses();
        methods = support.collectMethods(selected, dive, f);
        
        // sort using the comparator
        Set sortedSet = new TreeSet(getAvailableMethodsComparator());
        sortedSet.addAll(methods);
        
        DefaultListModel availModel = new DefaultListModel();
        for (it = sortedSet.iterator(); it.hasNext(); ) {
            availModel.addElement(it.next());
        }
        availableList.setModel(availModel);
    }
    
    /** Converts a collecton to a Model using a filter.
     */
    protected DefaultListModel createFilterModel(Collection items, ElementFilter.Method filter,
        Comparator comparator) {
        
        DefaultListModel model = new DefaultListModel();
        Iterator it;

        Collection sortedFiltered = new TreeSet(comparator);
        for (it = items.iterator(); it.hasNext(); ) {
            MethodElement m = (MethodElement)it.next();
            if (filter.accept(m))
                sortedFiltered.add(m);
        }
        for (it = sortedFiltered.iterator(); it.hasNext(); ) {
            model.addElement(it.next());
        }
        return model;
    }

    /** Actual method override handler. The function is called from {@link overrideSelectedMethods}
     * for individual method overriding.
     * The default implementation just inserts the method object into the model of the
     * override list. No checking for unambigous methods or other restrictions
     * is done.
     */
    protected void overrideMethod(MethodElement source) throws SourceException {
        // find appropriate position in the override list:
        int i;
        for (i = 0; i < overrideModel.getSize(); i++) {
            if (overrideComparator.compare(overrideModel.getElementAt(i), source) < 0)
                break;
        }
        overrideModel.insertElementAt(source, i);
    }
    
    /** Actual handler for removing a method. The default implementation just
     * removes the method object from the override list.
     */
    protected void removeMethod(MethodElement m) throws SourceException {
        overrideList.getSelectionModel().clearSelection();
        overrideModel.removeElement(m);
    }
    
    
    private void selectAllSuperclasses() {
        ExplorerManager m = superclassList.getExplorerManager();
        Node[] nodes = m.getExploredContext().getChildren().getNodes();
        try {
            m.setSelectedNodes(nodes);
        } catch (PropertyVetoException e) {
        }
    }
    
    private void unselectAllSuperclasses() {
        ExplorerManager m = superclassList.getExplorerManager();
        Node[] nodes = new Node[0];
        try {
            m.setSelectedNodes(nodes);
        } catch (PropertyVetoException e) {
        }
    }

    private void removeSelectedMethods() {
      Object[] vals = overrideList.getSelectedValues();
      if (vals == null || vals.length == 0)
        return;
      try {
          for (int i = 0; i < vals.length; i++) {
              MethodElement el = (MethodElement)vals[i];
              removeMethod(el);
          }
      } catch (SourceException e) {
          TopManager.getDefault().notify(
          new NotifyDescriptor.Exception(e));
      }
    }        

    protected void overrideSelectedMethods() {
        Object[] vals = availableList.getSelectedValues();
        if (vals == null || vals.length == 0)
        return;
        for (int i = 0; i < vals.length; i++) {
            MethodElement el = (MethodElement)vals[i];
            try {
                overrideMethod(el);
            } catch (SourceException e) {
                TopManager.getDefault().notify(
                new NotifyDescriptor.Exception(e));
            }
        }
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the FormEditor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        firstPanel = new javax.swing.JPanel();
        superclassHolder = new javax.swing.JPanel();
        jPanel3 = new CommonSizePanel();
        selectAll = new javax.swing.JButton();
        secondPanel = new javax.swing.JPanel();
        availablePanel = new javax.swing.JPanel();
        jScrollPane3 = new javax.swing.JScrollPane();
        availableList = new javax.swing.JList();
        jPanel2 = new CommonSizePanel();
        showOptions = new javax.swing.JPanel();
        includeInherited = new javax.swing.JCheckBox();
        onlyAbstract = new javax.swing.JCheckBox();
        selectAllMethods = new javax.swing.JButton();
        overrideButton = new javax.swing.JButton();
        thirdPanel = new javax.swing.JPanel();
        selectedPanel = new javax.swing.JPanel();
        jScrollPane1 = new javax.swing.JScrollPane();
        overrideList = new javax.swing.JList();
        jPanel4 = new CommonSizePanel();
        removeButton = new javax.swing.JButton();
        generateSuper = new javax.swing.JCheckBox();
        setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints1;
        setPreferredSize(new java.awt.Dimension(700, 550));
        
        firstPanel.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints2;
        
        superclassHolder.setLayout(new java.awt.BorderLayout());
        gridBagConstraints2 = new java.awt.GridBagConstraints();
        gridBagConstraints2.gridx = 0;
        gridBagConstraints2.gridy = 0;
        gridBagConstraints2.gridheight = 0;
        gridBagConstraints2.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints2.anchor = java.awt.GridBagConstraints.NORTHWEST;
        gridBagConstraints2.weightx = 1.0;
        gridBagConstraints2.weighty = 1.0;
        firstPanel.add(superclassHolder, gridBagConstraints2);
        
        
        jPanel3.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints3;
        jPanel3.setMinimumSize(new java.awt.Dimension(124, 33));
        jPanel3.setPreferredSize(new java.awt.Dimension(124, 157));
        
        selectAll.setPreferredSize(new java.awt.Dimension(90, 27));
        selectAll.setText(getBundle().getString("BTN_SelectAllAncestors"));
        selectAll.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                selectAllActionPerformed(evt);
            }
        }
        );
        gridBagConstraints3 = new java.awt.GridBagConstraints();
        gridBagConstraints3.gridwidth = 0;
        gridBagConstraints3.gridheight = 0;
        gridBagConstraints3.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints3.insets = new java.awt.Insets(0, 0, 6, 0);
        gridBagConstraints3.anchor = java.awt.GridBagConstraints.NORTHEAST;
        gridBagConstraints3.weightx = 1.0;
        gridBagConstraints3.weighty = 1.0;
        jPanel3.add(selectAll, gridBagConstraints3);
        
        gridBagConstraints2 = new java.awt.GridBagConstraints();
        gridBagConstraints2.gridx = 1;
        gridBagConstraints2.gridy = 0;
        gridBagConstraints2.insets = new java.awt.Insets(0, 8, 0, 0);
        gridBagConstraints2.anchor = java.awt.GridBagConstraints.NORTHEAST;
        firstPanel.add(jPanel3, gridBagConstraints2);
        
        
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints1.insets = new java.awt.Insets(8, 8, 8, 8);
        gridBagConstraints1.weightx = 1.0;
        gridBagConstraints1.weighty = 1.5;
        add(firstPanel, gridBagConstraints1);
        
        
        secondPanel.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints4;
        
        availablePanel.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints5;
        
        jScrollPane3.setPreferredSize(new java.awt.Dimension(200, 100));
        
        jScrollPane3.setViewportView(availableList);
        
        gridBagConstraints5 = new java.awt.GridBagConstraints();
        gridBagConstraints5.gridx = 0;
        gridBagConstraints5.gridy = 1;
        gridBagConstraints5.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints5.weightx = 1.0;
        gridBagConstraints5.weighty = 1.0;
        availablePanel.add(jScrollPane3, gridBagConstraints5);
        
        gridBagConstraints4 = new java.awt.GridBagConstraints();
        gridBagConstraints4.gridx = 0;
        gridBagConstraints4.gridy = 0;
        gridBagConstraints4.gridheight = 0;
        gridBagConstraints4.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints4.anchor = java.awt.GridBagConstraints.NORTHWEST;
        gridBagConstraints4.weightx = 1.0;
        gridBagConstraints4.weighty = 1.0;
        secondPanel.add(availablePanel, gridBagConstraints4);
        
        
        jPanel2.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints6;
        
        showOptions.setLayout(new javax.swing.BoxLayout(showOptions, 1));
        
        includeInherited.setText(getBundle().getString("BTN_ShowInheritedMethods"));
        includeInherited.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeInheritedActionPerformed(evt);
            }
        }
        );
        showOptions.add(includeInherited);
        
        
        onlyAbstract.setText(getBundle().getString("BTN_ShowOnlyAbstractMethods"));
        onlyAbstract.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                onlyAbstractActionPerformed(evt);
            }
        }
        );
        showOptions.add(onlyAbstract);
        
        gridBagConstraints6 = new java.awt.GridBagConstraints();
        gridBagConstraints6.gridx = 0;
        gridBagConstraints6.gridy = 0;
        gridBagConstraints6.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints6.insets = new java.awt.Insets(0, 0, 8, 0);
        gridBagConstraints6.anchor = java.awt.GridBagConstraints.NORTHEAST;
        gridBagConstraints6.weightx = 1.0;
        gridBagConstraints6.weighty = 1.0;
        jPanel2.add(showOptions, gridBagConstraints6);
        
        
        selectAllMethods.setPreferredSize(new java.awt.Dimension(90, 27));
        selectAllMethods.setText(getBundle().getString("BTN_SelectAllAvailableMethods"));
        selectAllMethods.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                selectAllMethodsActionPerformed(evt);
            }
        }
        );
        gridBagConstraints6 = new java.awt.GridBagConstraints();
        gridBagConstraints6.gridx = 0;
        gridBagConstraints6.gridy = 1;
        gridBagConstraints6.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints6.anchor = java.awt.GridBagConstraints.EAST;
        jPanel2.add(selectAllMethods, gridBagConstraints6);
        
        
        overrideButton.setPreferredSize(new java.awt.Dimension(90, 27));
        overrideButton.setText(getBundle().getString("BTN_OverrideMethod"));
        overrideButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                overrideButtonActionPerformed(evt);
            }
        }
        );
        gridBagConstraints6 = new java.awt.GridBagConstraints();
        gridBagConstraints6.gridx = 0;
        gridBagConstraints6.gridy = 2;
        gridBagConstraints6.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints6.insets = new java.awt.Insets(6, 0, 0, 0);
        gridBagConstraints6.anchor = java.awt.GridBagConstraints.SOUTHEAST;
        jPanel2.add(overrideButton, gridBagConstraints6);
        
        gridBagConstraints4 = new java.awt.GridBagConstraints();
        gridBagConstraints4.gridx = 1;
        gridBagConstraints4.gridy = 0;
        gridBagConstraints4.gridheight = 0;
        gridBagConstraints4.insets = new java.awt.Insets(0, 8, 0, 0);
        gridBagConstraints4.anchor = java.awt.GridBagConstraints.NORTHEAST;
        secondPanel.add(jPanel2, gridBagConstraints4);
        
        
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 1;
        gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints1.insets = new java.awt.Insets(0, 8, 8, 8);
        gridBagConstraints1.weightx = 1.0;
        gridBagConstraints1.weighty = 2.0;
        add(secondPanel, gridBagConstraints1);
        
        
        thirdPanel.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints7;
        
        selectedPanel.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints8;
        
        jScrollPane1.setPreferredSize(new java.awt.Dimension(200, 100));
        
        jScrollPane1.setViewportView(overrideList);
        
        gridBagConstraints8 = new java.awt.GridBagConstraints();
        gridBagConstraints8.gridx = 0;
        gridBagConstraints8.gridy = 1;
        gridBagConstraints8.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints8.weightx = 1.0;
        gridBagConstraints8.weighty = 1.0;
        selectedPanel.add(jScrollPane1, gridBagConstraints8);
        
        gridBagConstraints7 = new java.awt.GridBagConstraints();
        gridBagConstraints7.gridx = 0;
        gridBagConstraints7.gridy = 0;
        gridBagConstraints7.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints7.insets = new java.awt.Insets(8, 0, 0, 0);
        gridBagConstraints7.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints7.weightx = 1.0;
        gridBagConstraints7.weighty = 1.0;
        thirdPanel.add(selectedPanel, gridBagConstraints7);
        
        
        jPanel4.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints9;
        jPanel4.setMinimumSize(new java.awt.Dimension(124, 37));
        jPanel4.setPreferredSize(new java.awt.Dimension(124, 157));
        
        removeButton.setPreferredSize(new java.awt.Dimension(90, 27));
        removeButton.setText(getBundle().getString("BTN_RemoveOverridenMethod"));
        removeButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeButtonActionPerformed(evt);
            }
        }
        );
        gridBagConstraints9 = new java.awt.GridBagConstraints();
        gridBagConstraints9.gridx = 0;
        gridBagConstraints9.gridy = 0;
        gridBagConstraints9.gridheight = 0;
        gridBagConstraints9.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints9.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints9.weightx = 1.0;
        gridBagConstraints9.weighty = 1.0;
        jPanel4.add(removeButton, gridBagConstraints9);
        
        gridBagConstraints7 = new java.awt.GridBagConstraints();
        gridBagConstraints7.gridx = 1;
        gridBagConstraints7.gridy = 0;
        gridBagConstraints7.insets = new java.awt.Insets(8, 8, 0, 0);
        gridBagConstraints7.anchor = java.awt.GridBagConstraints.NORTHEAST;
        thirdPanel.add(jPanel4, gridBagConstraints7);
        
        
        generateSuper.setSelected(true);
        generateSuper.setText(getBundle().getString("BTN_GenerateSuperCall"));
        gridBagConstraints7 = new java.awt.GridBagConstraints();
        gridBagConstraints7.gridx = 0;
        gridBagConstraints7.gridy = 1;
        gridBagConstraints7.anchor = java.awt.GridBagConstraints.NORTHWEST;
        thirdPanel.add(generateSuper, gridBagConstraints7);
        
        
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 2;
        gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints1.insets = new java.awt.Insets(0, 8, 8, 8);
        gridBagConstraints1.weightx = 1.0;
        gridBagConstraints1.weighty = 2.0;
        add(thirdPanel, gridBagConstraints1);
        
    }//GEN-END:initComponents

  private void overrideButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_overrideButtonActionPerformed
      overrideSelectedMethods();
  }//GEN-LAST:event_overrideButtonActionPerformed

  private void onlyAbstractActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_onlyAbstractActionPerformed
    refreshAvailableMethods();
  }//GEN-LAST:event_onlyAbstractActionPerformed

  private void includeInheritedActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_includeInheritedActionPerformed
      refreshAvailableMethods();
  }//GEN-LAST:event_includeInheritedActionPerformed

  private void selectAllMethodsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_selectAllMethodsActionPerformed
    ListSelectionModel m = availableList.getSelectionModel();
    m.setSelectionInterval(0, availableList.getModel().getSize() - 1);
  }//GEN-LAST:event_selectAllMethodsActionPerformed
  
  private void selectAllActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_selectAllActionPerformed
    selectAllSuperclasses();
  }//GEN-LAST:event_selectAllActionPerformed

  private void removeButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeButtonActionPerformed
      Object[] vals = overrideList.getSelectedValues();
      if (vals == null || vals.length == 0)
        return;
      try {
          for (int i = 0; i < vals.length; i++) {
              MethodElement el = (MethodElement)vals[i];
              removeMethod(el);
          }
      } catch (SourceException e) {
          TopManager.getDefault().notify(
          new NotifyDescriptor.Exception(e));
      }
  }//GEN-LAST:event_removeButtonActionPerformed


  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JPanel firstPanel;
  private javax.swing.JPanel superclassHolder;
  private javax.swing.JPanel jPanel3;
  private javax.swing.JButton selectAll;
  private javax.swing.JPanel secondPanel;
  private javax.swing.JPanel availablePanel;
  private javax.swing.JScrollPane jScrollPane3;
  private javax.swing.JList availableList;
  private javax.swing.JPanel jPanel2;
  private javax.swing.JPanel showOptions;
  private javax.swing.JCheckBox includeInherited;
  private javax.swing.JCheckBox onlyAbstract;
  private javax.swing.JButton selectAllMethods;
  private javax.swing.JButton overrideButton;
  private javax.swing.JPanel thirdPanel;
  private javax.swing.JPanel selectedPanel;
  private javax.swing.JScrollPane jScrollPane1;
  private javax.swing.JList overrideList;
  private javax.swing.JPanel jPanel4;
  private javax.swing.JButton removeButton;
  private javax.swing.JCheckBox generateSuper;
  // End of variables declaration//GEN-END:variables

    /** Default method filter. Filters out methods that are declared abstract.
     */
    private static class DefaultMethodFilter implements ElementFilter.Method {
        public boolean accept(MethodElement m) {
            return !Modifier.isAbstract(m.getModifiers());
        }
    }
    
    /** Comparator that orders methods as follows:
     *  1. sorts them in ascending alphabetical order
     *  2. methods with the same name, but less arguments are preceding methods with more arguments
     *  3. methods with the same number of arguments are sorted by their argument declarations
     */
    private static class DefaultMethodComparator implements Comparator {
        public int compare(Object a,Object b) {
            MethodElement ma = (MethodElement)a;
            MethodElement mb = (MethodElement)b;
            int result;

            if (ma == mb)
            return 0;
            result = ma.getName().getSourceName().compareTo(mb.getName().getSourceName());
            if (result != 0)
            return result;
            result = ma.getParameters().length - mb.getParameters().length;
            if (result != 0)
            return result;
            // compare parameter names/types.
            MethodParameter[] paramA = ma.getParameters();
            MethodParameter[] paramB = mb.getParameters();

            for (int i = 0; i < paramA.length; i++) {
                result = paramA[i].getFullString().compareTo(paramB[i].getFullString());
                if (result != 0)
                return result;
            }
            return 0;
        }

        public boolean equals(Object ob) {
            return ob instanceof DefaultMethodComparator;
        }
    }
    
    private class CommonSizePanel extends JPanel {
        public Dimension getPreferredSize() {
            Dimension d = super.getPreferredSize();
            if (OverridePanelBase.this.buttonPanelWidth < 0)
                return d;
            d.width = OverridePanelBase.this.buttonPanelWidth;
            return d;
        }
    }
}
