/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.ui;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;

import java.lang.ref.WeakReference;
import java.lang.ref.Reference;

import java.util.LinkedList;

import javax.swing.ComboBoxModel;
import javax.swing.JComboBox;

import org.openide.cookies.SourceCookie;
import org.openide.cookies.OpenCookie;

import org.openide.explorer.ExplorerPanel;
import org.openide.explorer.ExplorerManager;
import org.openide.explorer.view.ChoiceView;
import org.openide.explorer.view.Visualizer;
import org.openide.explorer.view.NodeListModel;

import org.openide.nodes.AbstractNode;
import org.openide.nodes.*;

import org.openide.windows.TopComponent;

import org.openide.src.*;
import org.openide.src.nodes.*;

import org.netbeans.modules.java.JavaDataObject;

/**
 *
 * @author  sdedic
 * @version 
 */
public class NavigationView extends ChoiceView {
    /**
     * Selection manager for keyboard events.
     */
    SelectionManager        selManager;
    
    /**
     * Listener that binds the View to the explorer manager and to the
     * TopComponent or whatever.
     */
    RootL                   rootChangeListener;
    
    /**
     * Cookie that provides data about the active source.
     */
    SourceCookie.Editor     sourceCookie;
    
    /**
     * Active Children object.
     */
    SourceChildren          sourceChildren;
    
    /**
     * Last observed root node - if changes, contents are reloaded completely.
     */
    Node                    rootNode;
    
    /**
     * Node factory specific for the data source.
     */
    ElementNodeFactory      nodeFactory;
    
    /**
     * Associated ExplorerManager.
     */
    ExplorerManager         manager;

    /**
     * TopComponent to listen to, or null if we should listen to
     * TopComponent.Registry.
     */
    TopComponent            topComponent;
    
    /**
     * Makes the Navigator ignore events from the combobox. This is used
     * during sync of the combo from the activated nodes.
     */
    boolean                 ignoreComboSelect;
    
    /**
     * Makes the Navigator ignore events from the activated nodes. This is
     * mainly used when the node is being "opened".
     */
    boolean                 ignoreActivatedNodes;
    
    private static final int MAX_NAVIGATOR_DEPTH = 4;
    
    public NavigationView() {
        initComponents();
    }
    
    private void initComponents() {
        ((NodeListModel)getModel()).setDepth(MAX_NAVIGATOR_DEPTH);
        setKeySelectionManager(selManager = new SelectionManager());
        sourceChildren = new SourceChildren(nodeFactory = JavaDataObject.getBrowserFactory());
        
        SourceElementFilter f = new SourceElementFilter();
        f.setAllClasses(true);
        
        sourceChildren.setFilter(f);
        rootNode = new AbstractNode(sourceChildren);
    }
    
    /**
     * Finds the context where the view is actually used.
     */
    public void addNotify() {
        super.addNotify();
    
        if (rootChangeListener == null)
            rootChangeListener = new RootL();
        
        topComponent = findParentTopComponent();

        getExplorerManager().setRootContext(rootNode);
        try {
            getExplorerManager().setSelectedNodes(new Node[] {});
        } catch (PropertyVetoException e) {
        }
        activatedNodeChanged(getActivatedNode());        
        if (topComponent == null) {
            TopComponent.getRegistry().addPropertyChangeListener(
                rootChangeListener);
        } else {
            topComponent.addPropertyChangeListener(rootChangeListener);
        }
        getExplorerManager().addPropertyChangeListener(rootChangeListener);
    }
    
    public void removeNotify() {
        if (topComponent != null)
            topComponent.removePropertyChangeListener(rootChangeListener);
        else
            TopComponent.getRegistry().removePropertyChangeListener(rootChangeListener);
        getExplorerManager().removePropertyChangeListener(rootChangeListener);
        super.removeNotify();
    }
        
    
    private ExplorerManager getExplorerManager() {
        if (manager == null) {
            manager = ExplorerManager.find(this);
        }
        return manager;
    }
    
    private TopComponent findParentTopComponent() {
        java.awt.Component p;
        
        for (p = getParent(); p != null && !(p instanceof TopComponent); p = p.getParent())
            ;
        return (TopComponent)p;
    }
    
    private Node getActivatedNode() {
        Node[] actNodes;
        
        if (topComponent == null) {
            actNodes = TopComponent.getRegistry().getActivatedNodes();
        } else {
            actNodes = topComponent.getActivatedNodes();
        }
        if (actNodes == null || actNodes.length == 0)
            return null;
        else 
            return actNodes[0];
    }
    
    private void activatedNodeChanged(Node newNode) {
        if (newNode == null) {
            // PENDING: disable - how ?
            setEnabled(false);
            return;
        }
        
        SourceCookie.Editor srcEditor = (SourceCookie.Editor)newNode.getCookie(
            SourceCookie.Editor.class);
        if (srcEditor == null) {
            // PENDING: disable -- how ?
            setEnabled(false);
            return;
        } else {
            setEnabled(true);
        }
        
        Node curNode = null;

        if (srcEditor != sourceCookie) {
            changeRootContext(srcEditor);
        } else {
            Node[] curNodes = getExplorerManager().getSelectedNodes();
            curNode = curNodes.length == 0 ? null : curNodes[0];
            if (curNode != null && curNode.equals(newNode))
                return;
        }

        // now, determine whether the node has an Element cookie. If it does,
        // select the appropriate element from the list.
        Element el = (Element)newNode.getCookie(Element.class);
        Node toSelect = findNode(getExplorerManager().getRootContext(), el);
        
        if (toSelect == null || (curNode != null && curNode.equals(toSelect)))
            return;
        ExplorerManager m = getExplorerManager();
        try {
            m.removePropertyChangeListener(rootChangeListener);
            getExplorerManager().setSelectedNodes( new Node[] { toSelect } );
        } catch (PropertyVetoException ex) {
        } finally {
            m.addPropertyChangeListener(rootChangeListener);
        }
    }
    
    private void changeRootContext(SourceCookie.Editor cookie) {
        SourceElement s = cookie.getSource();
        // refill the combobox.
        sourceChildren.setElement(s);
        sourceCookie = cookie;
    }
    
    private void choiceItemSelected(Node[] sel) {
        if (sel == null || sel.length == 0)
            sel = getExplorerManager().getSelectedNodes();
        // Do nothing if nothing is selected (it is possible ?)
        if (sel == null || sel.length == 0)
            return;
        
        Node n = sel[0];

        OpenCookie o = (OpenCookie)n.getCookie(OpenCookie.class);
        if (o == null)
            return;
        
        // open the element / locate it in the editor.
        boolean saveActivatedNodes = ignoreActivatedNodes;
        try {
            ignoreActivatedNodes = true;
            o.open();
        } finally {
            ignoreActivatedNodes = saveActivatedNodes;
        }
    }
    
    /**
     * Tries to find a node that represents the element inside the rootNode.
     */
    private Node findNode(Node rootNode, Element target) {
        // suppose that the root node is a SourceElement.
        ClassElement outer;
        String name;
        
        if (target instanceof InitializerElement) {
            InitializerElement i = (InitializerElement)target;
            outer = i.getDeclaringClass();
            name = "<initializer>"; // NOI18N
        } else if (target instanceof MemberElement) {
            MemberElement m = (MemberElement)target;
            name = m.getName().getName();
            outer = m.getDeclaringClass();
        } else {
            return null;
        }
        
        Node parent;
        
        if (outer != null) {
            parent = findElementNode(rootNode, outer);
	    if (parent == null)
		// sorry, bad luck, it isn't there :)
		return null;
        } else {
            parent = rootNode;
        }
        return findElementNode(parent, target);
    }
    
    private Node findElementNode(Node parent, Element el) {
        java.util.Enumeration children = parent.getChildren().nodes();
        while (children.hasMoreElements()) {
            Node n = (Node)children.nextElement();
            if (n.getCookie(Element.class) == el)
                return n;
        }    
        return null;
    }

    private class RootL implements PropertyChangeListener {
        public void propertyChange(PropertyChangeEvent e) {
            Object source = e.getSource();
            String propName = e.getPropertyName();
            if ((source == topComponent ||
                (topComponent == null && source == TopComponent.getRegistry())) &&
                TopComponent.Registry.PROP_ACTIVATED_NODES.equals(propName)
                ) {
                // node changed !!! sync on the SourceCookie.Editor cookie.
                if (!ignoreActivatedNodes) {
                    try {
                        ignoreComboSelect = true;
                        activatedNodeChanged(getActivatedNode());
                    } finally {
                        ignoreComboSelect = false;
                    }
                }
            } else if (source == getExplorerManager() && 
                ExplorerManager.PROP_SELECTED_NODES.equals(propName)) {
                // something was selected ??
                if (!ignoreComboSelect) {
                    boolean saveActivatedNodes = ignoreActivatedNodes;
                    try {
                        ignoreActivatedNodes = true;
                        choiceItemSelected((Node[])e.getNewValue());
                    } finally {
                        ignoreActivatedNodes = saveActivatedNodes;
                    }
                }
            }
        }
    }
    
    private class SelectionManager extends NameNavigator 
        implements JComboBox.KeySelectionManager {

        SelectionManager() {
            super(ExplorerManager.find(NavigationView.this));
        }
        
        public int selectionForKey(char key, ComboBoxModel comboBoxModel) {
            Object selectedItem = comboBoxModel.getSelectedItem();
            // suppose the selected object is a visualizer.
            Node currentNode = Visualizer.findNode(selectedItem);

            checkTimeout();
            Node result = navigate(currentNode, key);
            if (result == null)
                return -1;
            
            int index = ((NodeListModel)comboBoxModel).getIndex(Visualizer.findVisualizer(result));
            return index;
        }
    }
}
