/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.modules.java.ui.wizard;

import java.beans.*;
import java.util.*;

import org.openide.nodes.*;
import org.openide.src.*;
import org.openide.src.nodes.*;
import org.openide.util.TaskListener;
import org.openide.util.Task;

import org.netbeans.modules.java.tools.InheritanceSupport;
import org.netbeans.modules.java.tools.ElementFilter;
//import org.netbeans.modules.java.tools.OverrideAction;

class InheritanceChildren extends Children.Keys implements InheritanceSupport.HierarchyListener {
    /** Special `wait' node that signalizes that inheritance infomation was
        not computed yet.
    */
    private static final Object WAIT = new Object();
    
    /** Text that is printed when the hierarchy of the class changes and the
        superclass computation is not completed yet.
    */
//    private static final String WAIT_STRING = OverrideAction.getString("MSG_InheritanceWait");
    private static final String WAIT_STRING = "wait";
    
    /** Object that collects class' inheritance data
    */
    private InheritanceSupport support;
    
    /** Comparator used to sort inherited classes
    */
    private Comparator comparator;
    private Comparator methodComp;
    private ElementFilter.Method filter;
    private boolean methodsOnly = false;
    
    private InheritanceSupport.HierarchyListener listener;

    /** Creates new InheritanceChildren */
    public InheritanceChildren(InheritanceSupport data, 
                               java.util.Comparator classComp,
                               java.util.Comparator methodComp,
                               ElementFilter.Method filter) {
        super();
        this.support = data;
        this.comparator = classComp;
        this.methodComp = methodComp;
        this.filter = filter;
        setWaitNode();
        listener = this;
    }
    
    public void addNotify() {
        support.addHierarchyListener(listener);
        support.prepareHierarchy().addTaskListener(new TaskListener() {
            public void taskFinished(Task t) {
                t.removeTaskListener(this);
                refreshData();
                
                // attach a listener to listen for future changes.
                support.addHierarchyListener(listener);
            }
        });
        refreshData();
    }
    
    public void removeNotify() {
        support.removeHierarchyListener(listener);
        setKeys(Collections.EMPTY_SET);
    }

    public void setMethodsOnly(boolean mo) {
        if (methodsOnly != mo) {
            methodsOnly = mo;
            refreshData();
        }
    }

    public boolean isMethodsOnly() {
        return methodsOnly;
    }

    public void setFilter(ElementFilter.Method f) {
        if (filter != f) {
            filter = f;
            if (methodsOnly) refreshData();
            else {
                Node[] classNodes = getNodes();
                for(int i = 0; i < classNodes.length; i++) {
                    ((Methods)classNodes[i].getChildren()).refreshKeys();
                }
            }
        }
    }

    public ElementFilter.Method getFilter() {
        return filter;
    }
    
    private void refreshData() {
        Collection original;
        Comparator comp;
        if (methodsOnly) {
            original = support.createMethodMap(support.getInheritedClasses(), false, filter).values();
            comp = methodComp;
        } else {
            original = support.getInheritedClasses();
            comp = comparator;
        }

        Collection result;
        if (comp == null) {
            result = original;
        } else {
            result = new TreeSet(comp);
            result.addAll(original);
        }
        setKeys(result);
    }

    protected Node[] createNodes(Object key) {
        if (key instanceof ClassElement) {
            return new Node[] {
                support.createClassElementNode(
                    (ClassElement)key,
                    new Methods((ClassElement)key),
                    false
                )
            };
        } else if (key instanceof MethodElement) {
            MethodElementNode me = new MethodElementNode((MethodElement)key, false);
            me.setElementFormat(MethodsChildren.METHOD_FORMAT);
            return new Node[] { me };
        } else if (key == WAIT) {
            return new Node[] {
                new AbstractNode(Children.LEAF) {
                    public String getDisplayName() {
                        return WAIT_STRING;
                    }
               }
            };
        } else {
            throw new IllegalArgumentException(key.getClass().toString());
        }
    }
    
    private final void setWaitNode() {
        setKeys(new Object[] {
            WAIT
        });
    }
    
    /** Informs that class' inheritance has changed.
     */
    public void inheritanceChanged() {
        setWaitNode();
    }
    
    /** Informs that collection of inherited classes has been recomputed.
     */
    public void inheritedClassesComputed(java.util.Map classes) {
        // refresh the keys with all the classes, sorted by the comparator.
        refreshData();
    }
    
    /** Informs, that the collection of inherited methods has been recomputed.
     */
    public void inheritedMethodsComputed(Collection methods) {
    }

    
    class Methods extends Children.Keys {

        protected ClassElement classElement;
       
        public Methods(ClassElement clazz) {
            classElement = clazz;
        }

        protected void addNotify() {
            refreshKeys();
        }

        protected void removeNotify () {
            setKeys (Collections.EMPTY_SET);
        }

        protected Node[] createNodes(Object key) {
            if (key instanceof MethodElement) {
                MethodElementNode me = new MethodElementNode((MethodElement)key, false);
                me.setElementFormat(MethodsChildren.METHOD_FORMAT);
                return new Node[] {
                    me
                };
            } else {
                throw new IllegalArgumentException(key.getClass().toString());
            }
        }
        
        void refreshKeys() {
            Collection keys = new ArrayList();
            keys.add(classElement);

            TreeSet subKeys = new TreeSet(methodComp);
            subKeys.addAll(support.createMethodMap(keys, false, filter).values());
            setKeys(subKeys);
        }
    }
    
    static class OverridenMethods extends Children.Keys implements PropertyChangeListener {

        ClassElement classElement;
        InheritanceSupport support;
        Comparator comparator;

        public OverridenMethods(ClassElement clazz, 
                                InheritanceSupport support,
                                Comparator comp) {
            classElement = clazz;
            this.support = support;
            comparator = comp;
        }

        protected void addNotify() {
            classElement.addPropertyChangeListener(this);
            refreshKeys();
        }

        protected void removeNotify () {
            classElement.removePropertyChangeListener(this);
            setKeys (Collections.EMPTY_SET);
        }

        protected Node[] createNodes(Object key) {
            if (key instanceof MethodElement) {
                MethodElementNode me = new MethodElementNode((MethodElement)key, false);
                me.setElementFormat(MethodsChildren.METHOD_FORMAT);
                return new Node[] {
                    me
                };
            } else {
                throw new IllegalArgumentException(key.getClass().toString());
            }
        }

        void refreshKeys() {
            List keys = new ArrayList(support.getOverridingMethods());
            Collections.sort(keys, comparator);
            setKeys(keys);
        }

        public void propertyChange(PropertyChangeEvent evt) {
            if (ClassElement.PROP_METHODS.equals(evt.getPropertyName())) {
                refreshKeys();
            }
        }
    }
}
