/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.ui.wizard;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.ResourceBundle;

import java.awt.Component;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.WizardDescriptor;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.TemplateWizard;
import org.openide.util.Utilities;
import org.openide.util.NbBundle;

import org.openide.cookies.SourceCookie;
import org.openide.src.ClassElement;
import org.openide.src.SourceElement;
import org.openide.src.SourceException;
import org.openide.src.Type;

import org.netbeans.modules.java.JavaDataLoader;

public class JavaWizardIterator implements TemplateWizard.Iterator {
    private static final long serialVersionUID = -1987345873459L;

    /** Array of panels that form the Wizard.
     */
    private transient WizardDescriptor.Panel[] panels;

    /**
     * Names for panels used in the Wizard.
     */
    private static String[] panelNames;
    
    /** Total number of panels that can be displayed.
     */
    private static final int PANEL_COUNT = 5;
    
    private static final int METHODS_PANEL = 4;
    private static final int FIELDS_PANEL = 2;
    
    /** Singleton instance of JavaWizardIterator, should it be ever needed.
     */
    private static JavaWizardIterator instance;
    
    /** Index of the current panel. Panels are numbered from 0 to PANEL_COUNT - 1.
     */
    private transient int panelIndex = 0;
    
    /** Reference to Wizard's private data structures.
     */
    private transient JavaWizardData myData;

    /**
     * Holds a reference to the instance of TemplateWizard we are communicating with.
     */
    private transient TemplateWizard wizardInstance;
    
    /** Preferred size of each panel computed from expected largest one
     */
    private transient java.awt.Dimension prefSize;
    
    /**
     * True, if the iterator was already initialized from the Wizard's
     * input fields.
     */
    private transient boolean initialized;
    
    public JavaWizardIterator() {
    }
    
    /** Returns JavaWizardIterator singleton instance. This method is used
     * for constructing the instance from filesystem.attributes.
     */
    public static synchronized JavaWizardIterator singleton() {
        if (instance == null) {
            instance = new JavaWizardIterator();
        }
        instance.initialized = false;
        return instance;
    }
    // ========================= TemplateWizard.Iterator ============================

    /** Instantiates the template using informations provided by
     * the wizard.
     *
     * @param wiz the wizard
     * @return set of data objects that has been created (should contain
     *  at least one
     * @exception IOException if the instantiation fails
    */
    public java.util.Set instantiate(TemplateWizard wiz) throws IOException,
        IllegalArgumentException {
        // dig out file from the template.
        if (!initialized) {
            initializeTemplateData(wiz.getTemplate());
            initializeTargetData(wiz.getTemplate(), wiz.getTargetFolder(), wiz.getTargetName());
        }                
        DataObject obj = instantiateTemplate(wiz.getTemplate(), wiz.getTargetFolder());
        
        SourceCookie src = (SourceCookie)obj.getCookie(SourceCookie.class);
        if (src == null) {
            return java.util.Collections.singleton(obj);
        }
        ClassElement[] classes = src.getSource().getClasses();
        
        try {
	    ClassElement targetClass;
	    
            // NO classes --> create a new class
	    if (classes == null || classes.length == 0) {
                ClassElement el = new ClassElement();
                el.setName(myData.getCustomizedClass().getName());
                src.getSource().addClass(el);
		targetClass = src.getSource().getClasses()[0];
            } else {
                targetClass = classes[0];
            }
            myData.applyChanges(targetClass);
        } catch (SourceException e) {
            obj.delete();
            throw new IOException(e.getLocalizedMessage());
        }

        // run default action (hopefully should be here)
        org.openide.nodes.Node node = obj.getNodeDelegate ();
        org.openide.util.actions.SystemAction sa = node.getDefaultAction ();
        if (sa != null) {
            sa.actionPerformed (new ActionEvent (node, ActionEvent.ACTION_PERFORMED, "")); // NOI18N
        }
        return java.util.Collections.singleton(obj);
    }
    
    /** Get the current panel.
     * @return the panel
    */
    public WizardDescriptor.Panel current() {
        return panels[panelIndex];
    }
    
    /** Get the name of the current panel.
     * @return the name
    */
    public String name() {
        return ""; // NOI18N
    }
    
    /** Test whether there is a next panel.
     * @return <code>true</code> if so
    */
    
    public boolean hasNext() {
        return panelIndex < PANEL_COUNT - 1;
    }
    
    /** Test whether there is a previous panel.
     * @return <code>true</code> if so
    */
    public boolean hasPrevious() {
        return panelIndex > 0;
    }
    
    /** Move to the next panel.
     * I.e. increment its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void nextPanel() {
	if (panelIndex == FIELDS_PANEL && !myData.getCustomizedClass().isClassOrInterface()) 
	    panelIndex += 2;
	else
    	    panelIndex++;
    }
    
    /** Move to the previous panel.
     * I.e. decrement its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void previousPanel() {
	if (panelIndex == METHODS_PANEL && !myData.getCustomizedClass().isClassOrInterface())
	    panelIndex -= 2;
	else
    	    panelIndex--;
    }
    
    /** Add a listener to changes of the current panel.
     * The listener is notified when the possibility to move forward/backward changes.
     * @param l the listener to add
    */
    public void addChangeListener(ChangeListener l) {
    }
    
    /** Remove a listener to changes of the current panel.
     * @param l the listener to remove
    */
    public void removeChangeListener(ChangeListener l) {
    }

    public void initialize(TemplateWizard wizard) {
        this.wizardInstance = wizard;
        
	if (panels == null) {
            
            myData = new JavaWizardData();
            initializeTemplateData(wizard.getTemplate());
            initialize();
            
            Component panel = wizard.targetChooser().getComponent();
            panelNames[0] = panel.getName();
            if (panel instanceof javax.swing.JComponent) {
                ((javax.swing.JComponent)panel).putClientProperty(
                    "WizardPanel_contentData", panelNames);
            }
            panels = new WizardDescriptor.Panel[] {
                wizard.targetChooser(),
                new DelegatingPanel(ClassPanel.class, true),
                new DelegatingPanel(FieldsPanel.class, false),
                new DelegatingPanel(OverridePanel.class, false),
                new DelegatingPanel(MethodsPanel.class, false)
            };
	}
        reset();
    }
    
    public void uninitialize(TemplateWizard wiz) {
	panels = null;
	myData = null;
    }

    // ========================= IMPLEMENTATION ============================
    
    private Object readResolve() {
        initialize();
        reset();
        return this;
    }
    
    /** Instantiates the template. Currently it just delegates to the template DataObject's
     * createFromTemplate method.
     */
    private DataObject instantiateTemplate(DataObject tpl, DataFolder target) throws IOException {
        String name = myData.getCustomizedClass().getName().getName();
        String packName = target.getPrimaryFile().getPackageNameExt('.', '.');
        checkValidPackageName(target);
        checkTargetName(target, name);
        return tpl.createFromTemplate(target, name);
    }
    
    protected void reset() {
        panelIndex = 0;
    }

    /** Initializes the Wizard with array of panels.
     */
    protected void initialize() {
        if (myData.getCustomizedClass().isClassOrInterface()) {
            panelNames = new String[5];
            panelNames[3] = getString("TIT_OverrideWizardPanel");
            panelNames[4] = getString("TIT_AddMethodsWizardPanel");
        } else {
            panelNames = new String[4];
            panelNames[3] = getString("TIT_AddMethodsWizardPanel");
        }
        panelNames[0] = "";
        panelNames[1] = getString("TIT_InheritanceWizardPanel");
        panelNames[2] = getString("TIT_AddFieldsWizardPanel");
    }
    
    private boolean isValidPackageName(String s) {
        // valid package is an empty one, or well-formed java identifier :-)
        if ("".equals(s))
            return true;
        try {
            Type t = Type.parse(s);
            return true;
        } catch (IllegalArgumentException ex) {
            return false;
        }
    }
    
    private void checkValidPackageName(DataFolder targetFolder) 
        throws IllegalStateException {
        FileObject folder = targetFolder.getPrimaryFile();
        String fullTarget = folder.getPackageNameExt('.', '.');
        // checking for java-compatible name - both the folder name and the target name
        // must be acceptable.
        if (!isValidPackageName(fullTarget)) {
            String msg = MessageFormat.format(getString("FMTERR_IllegalFolderName"),
            new Object[] {
                folder.getPackageNameExt('/', '.'),
                fullTarget
            });
            throw (IllegalStateException)TopManager.getDefault().getErrorManager().annotate(
            new IllegalStateException(msg),
            ErrorManager.USER, null, msg,
            null, null
            );
        }
    }
    
    private void checkTargetName(DataFolder folder, String desiredName) {
        if (!Utilities.isJavaIdentifier(desiredName)) {
            String msg = MessageFormat.format(getString("FMTERR_IllegalTargetName"),
            new Object[] {
                desiredName
            });
            IllegalStateException x = 
                (IllegalStateException)TopManager.getDefault().getErrorManager().annotate(
            new IllegalStateException(msg),
            ErrorManager.USER, null, msg,
            null, null
            );
            x.printStackTrace();
            throw x;
        }
        FileObject f = folder.getPrimaryFile();
        // check whether the name already exists:
        if (f.getFileObject(desiredName, JavaDataLoader.JAVA_EXTENSION) != null) { // NOI18N
            String msg = MessageFormat.format(getString("FMTERR_TargetExists"),
            new Object[] {
                desiredName
            });
            throw new IllegalStateException(msg);
        }
    }
    
    private void initializeTemplateData(DataObject template) throws IllegalStateException {

        if (template == null) {
            throw new IllegalStateException("Expected template, got null");
        }

        myData.setTemplate(template);
    }
    
    private void initializeTargetData(DataObject template, DataFolder targetFolder, String desiredName) 
        throws IllegalStateException {

        FileObject folder = targetFolder.getPrimaryFile();
        if (template == null || targetFolder == null) {
            throw new IllegalStateException("Expected template & target folder, got null(s)");
        }
        checkValidPackageName(targetFolder);
        if (desiredName == null) {
            // make up an name using the Template's name
            String fullTarget = folder.getPackageNameExt('.', '.');
            desiredName = org.openide.filesystems.FileUtil.findFreeFileName(folder,
                template.getName(), JavaDataLoader.JAVA_EXTENSION);
        } else {
            checkTargetName(targetFolder, desiredName);
        }
        myData.setTarget(folder, desiredName);

        prefSize = panels[4].getComponent().getPreferredSize();
        initialized = true;
    }

    private class DelegatingPanel implements WizardDescriptor.Panel, 
        WizardDescriptor.FinishPanel {

        private Class   panelClass;
        private boolean init;
        private WizardDescriptor.Panel  instance;
        
        DelegatingPanel(Class c, boolean initData) {
            panelClass = c;
            init = initData;
        }
        
        private WizardDescriptor.Panel getInstance() {
            if (instance != null)
                return instance;
            try {
                instance = (WizardDescriptor.Panel)panelClass.newInstance();
                if (instance instanceof JavaPanel) {
                    ((JavaPanel)instance).initialize(myData);
                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }
            return instance;
        }
        
        public org.openide.util.HelpCtx getHelp() {
            return getInstance().getHelp();
        }
        
        public void storeSettings(Object s) {
            getInstance().storeSettings(s);
        }
        
        public void addChangeListener(ChangeListener l) {
            getInstance().addChangeListener(l);
        }
        
        public void removeChangeListener(ChangeListener l) {
            getInstance().removeChangeListener(l);
        }
        
        public boolean isValid() {
            return getInstance().isValid();
        }
        
        public Component getComponent() {
            //return (Component)instance;
            ((javax.swing.JComponent) getInstance()).setPreferredSize(prefSize);
            if (panelIndex == panelNames.length) {
                ((javax.swing.JComponent) getInstance())
                    .putClientProperty("WizardPanel_contentSelectedIndex", 
                        new Integer(panelNames.length - 1)); //NOI18N
            }
            return getInstance().getComponent();
        }
        
        public void readSettings(Object settings) {
            if (!init || !(settings instanceof TemplateWizard)) {
                return;
            }
            TemplateWizard wizardInstance = (TemplateWizard)settings;
            
            DataFolder target = null;
            DataObject template = null;
            
            try {
                target = wizardInstance.getTargetFolder();
                template = wizardInstance.getTemplate();
            } catch (java.io.IOException ex) {
                // PENDING: provide better message / annotate through error manager.
                throw new IllegalStateException("Illegal settings");
            }

            initializeTargetData(template, target, wizardInstance.getTargetName());
            
            getInstance().readSettings(settings);
        }
    }

    private static ResourceBundle bundle;

    static String getString(String key) {
        if (bundle == null)
             bundle = org.openide.util.NbBundle.getBundle(JavaWizardIterator.class);
        return bundle.getString(key);
    }
}
