/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.ui.wizard;

import java.awt.*;
import java.beans.*;
import java.lang.reflect.Modifier;

import javax.swing.JPanel;
import java.util.StringTokenizer;

import org.openide.src.ElementProperties;
import org.openide.util.HelpCtx;
import org.openide.explorer.propertysheet.editors.EnhancedCustomPropertyEditor;

/** Property editors for java modifiers.
*
* @author jrojcek
*/
abstract public class ModifierEditor extends PropertyEditorSupport
                                     implements PropertyChangeListener,
                                                EnhancedCustomPropertyEditor {

    /** Name of 'modifier' property (current value) */
    public static final String PROP_MODIFIER = "modifier"; // NOI18N

    /** Names of modifiers */
    private static final String MODIFIER_NAMES[] = {
        "abstract", "final", "static", "synchronized", "transient", "volatile", "native" // NOI18N
    };

    /** Values of modifiers */
    private static final int MODIFIER_VALUES[] = {
        Modifier.ABSTRACT, Modifier.FINAL, Modifier.STATIC, Modifier.SYNCHRONIZED,
        Modifier.TRANSIENT, Modifier.VOLATILE, Modifier.NATIVE
    };

    /** Count of the modifiers */
    private static final int MODIFIER_COUNT = MODIFIER_VALUES.length;

    /** Names of accessibility */
    private static final String ACCESS_NAMES[] = {
        "<default>", "private", "protected", "public" // NOI18N
    };

    /** Values of accessibility */
    private static final int ACCESS_VALUES[] = {
        0, Modifier.PRIVATE, Modifier.PROTECTED, Modifier.PUBLIC
    };

    /** Mask of access modifiers */
    private static final int ACCESS_MASK = Modifier.PRIVATE | Modifier.PROTECTED | Modifier.PUBLIC;

    /** Instance of custom property editor - visual panel. */
    private AccessLevelPanel accessLevelPanel;
    private ModifierPanel modifierPanel;
    
    private JPanel customEditor;

    /** Creates new modifiers editor.
    * @param mask The mask of modifier values which should be possible to change.
    */
    public ModifierEditor(AccessLevelPanel accessLevelPanel, 
                          ModifierPanel modifierPanel) {

        this.accessLevelPanel = accessLevelPanel;
        this.modifierPanel = modifierPanel;
        accessLevelPanel.addPropertyChangeListener(this);
        modifierPanel.addPropertyChangeListener(this);
    }
    
    public AccessLevelPanel getAccessLevelPanel() {
        return accessLevelPanel;
    }
    
    public ModifierPanel getModifierPanel() {
        return modifierPanel;
    }

    /** Set new value */
    public void setValue(Object object) throws IllegalArgumentException {
        if (object instanceof Integer) {
            accessLevelPanel.setAccessLevel(ACCESS_MASK & ((Integer) object).intValue());
            modifierPanel.setTypeModifier(((Integer) object).intValue());
        }
        else {
            throw new IllegalArgumentException ();
        }
    }

    /** @return the java source code representation
    * of the current value.
    */
    public String getJavaInitializationString() {
        return getValue().toString();
    }

    /** Get the value */
    public Object getValue() {
        return new Integer(accessLevelPanel.getAccessLevel() 
                           | modifierPanel.getTypeModifier());
    }

    /** @return textual representation of current value of the modifiers. */
    public String getAsText() {
        return Modifier.toString(((Integer)getValue()).intValue());
    }

    /** Parse the text and sets the modifier editor value */
    public void setAsText(String string) throws IllegalArgumentException {
        int newValue = 0;
        int oldValue = ((Integer)getValue()).intValue();
        int mask = modifierPanel.getMask();

        StringTokenizer tukac = new StringTokenizer(string, ", ", false); // NOI18N
        while (tukac.hasMoreTokens()) {
            String token = tukac.nextToken();
            boolean known = false;
            for (int i = 0; i < MODIFIER_COUNT; i++) {
                if ((MODIFIER_VALUES[i] & mask) != 0) {
                    if (token.equals(MODIFIER_NAMES[i])) {
                        if (((MODIFIER_VALUES[i] == Modifier.FINAL) && ((newValue & Modifier.ABSTRACT) != 0)) ||
                                ((MODIFIER_VALUES[i] == Modifier.ABSTRACT) && ((newValue & Modifier.FINAL) != 0)))
                            break;
                        newValue |= MODIFIER_VALUES[i];
                        known = true;
                        break;
                    }
                }
            }
            if ((newValue & ACCESS_MASK) == 0) {
                for (int i = 1; i <= 3; i++) {
                    if ((ACCESS_VALUES[i] & mask) != 0) {
                        if (token.equals(ACCESS_NAMES[i])) {
                            newValue |= ACCESS_VALUES[i];
                            known = true;
                            break;
                        }
                    }
                }
            }
            if (!known)
                throw new IllegalArgumentException();
        }
        
        if (oldValue != newValue) {
            setValue(new Integer(newValue));
            firePropertyChange();
        }
    }

    /** @return <CODE>this</CODE> */
    public Component getCustomEditor() {
        if (customEditor == null) {
            customEditor = new JPanel(new java.awt.GridBagLayout());

            java.awt.GridBagConstraints gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.insets = new java.awt.Insets(0, 0, 11, 0);
            gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.NORTHWEST;
            customEditor.add(accessLevelPanel, gridBagConstraints1);

            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 1;
            gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.NORTHWEST;
            gridBagConstraints1.weightx = 1.0;
            gridBagConstraints1.weighty = 1.0;
            customEditor.add(modifierPanel.getComponent(), gridBagConstraints1);
        }
        return customEditor;
    }

    /** @return <CODE>true</CODE> */
    public boolean supportsCustomEditor() {
        return true;
    }

    /** Get the customized property value.
     * @return the property value
     * @exception InvalidStateException when the custom property editor does not contain a valid property value
     *           (and thus it should not be set)
     */
    public Object getPropertyValue() throws IllegalStateException {
        return getValue();
    }

    public void propertyChange(PropertyChangeEvent evt) {
        if (ModifierPanel.PROP_TYPE_MODIFIER.equals(evt.getPropertyName())) {
            firePropertyChange();
        } else if (AccessLevelPanel.PROP_ACCESS_LEVEL.equals(evt.getPropertyName())) {
            firePropertyChange();
        }
    }
    
    public static interface ModifierPanel {
        public static final String PROP_TYPE_MODIFIER = "typeModifier";
        
        public void setTypeModifier(int mod);

        public int getTypeModifier();
        
        public int getMask();
        
        public Component getComponent();
        
        public void addPropertyChangeListener(PropertyChangeListener pcl);
        
        public void removePropertyChangeListener(PropertyChangeListener pcl);
    }
}
