/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.wizard;

import java.io.IOException;
import java.util.Arrays;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.openide.nodes.Node;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.IndexedNode;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;

import org.openide.src.*;
import org.openide.src.nodes.ClassElementNode;
import org.openide.util.datatransfer.NewType;

import org.netbeans.modules.java.wizard.Util;

/** Class that creates a proxy node to a ordinary ClassElementNode. Its purpose
 * is to intercept destroy() operations and convert it to a removal of the 
 * interface from the class' implements clause.
 */
class ImplementsNode extends IndexedNode {
    /** Format for individual displayed interfaces. Shows only fully qualified name
        of the interface.
    */
    private static ElementFormat INTERFACE_NODE_FORMAT = new ElementFormat(
        Util.getString("FMT_InterfaceDisplayName")
    );

    /** The class whose interfaces this node represents.
    */
    ClassElement clazz;

    /** Creates new ImplementsNode 
        @param clazz Class the node is created for
        @param data reference to cache/creator of supporting objects.
    */
    public ImplementsNode(ClassElement clazz) {
        this(clazz, new Children(clazz, ElementProperties.PROP_INTERFACES));
    }

    private ImplementsNode(ClassElement clazz, Children ch) {
        super(ch, ch.getIndex());
        this.clazz = clazz;
    }

    public String getName() {
        return Util.getString("NAME_ImplementedInterfacesNode");
    }

    public NewType[] getNewTypes() {
        return new NewType[] {
            new NewInterfaceType(clazz)
        };
    }

    private void addInterface(Identifier id) throws SourceException {
        clazz.addInterface(id);
    }

    private Node createNode(ClassElement el) {
        ClassElementNode original = new ClassElementNode(el, Children.LEAF, false);
        original.setElementFormat(INTERFACE_NODE_FORMAT);
        return new ImplementedInterfaceNode(
            clazz, original, true
        );
    }

    /*************************************************************************/
    private static class Children extends IndexedPropertyChildren {

        public Children(Object bean, String propertyName) {
            super(bean, propertyName);
        }

        /** Extract an identifier from the node's key object.
        */
        private Identifier getIdentifier(Object key) {
            return (Identifier)key;
        }

        /** Create an element based on the identifier.
        */
        private ClassElement getElement(Identifier id) {
            ClassElement el = ClassElement.forName(id.getFullName());
            if (el != null) {
                return el;
            }
            el = new ClassElement();
            try {
                el.setName(id);
                el.setClassOrInterface(false);
           } catch (SourceException e) {
           }
           return el;
        }

        /** Creates a ClassElement node for the identifier found in `key'.
        */
        protected Node[] createNodes(Object key) {
            Identifier id = getIdentifier(key);
            ImplementsNode parent = (ImplementsNode)getNode();
            return new Node[] { 
                ((ImplementsNode)getNode()).createNode(getElement(id))
            };
        }
    }

    public static class NewInterfaceType extends NewType {
        ClassElement target;
        
        public NewInterfaceType(ClassElement target) {
            this.target = target;
        }
        
        public String getName() {
            return Util.getString("LBL_ImplementInterfaceNewType");
        }

        public org.openide.util.HelpCtx getHelpCtx() {
            return null;
        }

        public void create() throws IOException {
            // open up an identifier editor...
            /*
            org.openide.NotifyDescriptor.InputLine input = new org.openide.NotifyDescriptor.InputLine(
                Util.getString("LBL_InterfaceName"), Util.getString("TIT_ImplementInterface")
            );

            for (;;) {
                Object ret = TopManager.getDefault().notify(input);
                if (ret != NotifyDescriptor.OK_OPTION) {
                    return;
                }
                String retValue = input.getInputText();
             */
            InterfaceInputPanel p = new InterfaceInputPanel();
            for (;;) {
                if (!p.display())
                    return;
                String retValue = p.getInterfaceName();
                if ("".equals(retValue)) {
                    // empty identifier is not valid ;-0
                    continue;
                }
                
                if (!Util.isValidTypeIdentifier(retValue)) {
                    String msg = java.text.MessageFormat.format(
                        Util.getString("FMTERR_InvalidIdentifier"),
                        new Object[] {
                            retValue == null ? Util.getString("TXT_EmptyString") : retValue
                        });
                    TopManager.getDefault().notify(new NotifyDescriptor.Message(msg));
                    continue;
                }
                
                // test if the class doesn't contain that interface:
                Identifier id = Identifier.create(retValue);
                if (Util.implementsInterface(target, id)) {
                    String msg = java.text.MessageFormat.format(
                        Util.getString("FMTERR_AlreadyImplemented"),
                        new Object[] {
                            retValue
                        });
                    TopManager.getDefault().notify(new NotifyDescriptor.Message(
                        msg));
                    continue;
                }
                
                // last: test if there's a ClassElement matching that ID and
                // if it is really an interface:
                ClassElement cls = ClassElement.forName(retValue);
                if (cls != null && cls.isClass()) {
                    String msg = java.text.MessageFormat.format(
                        Util.getString("FMTERR_NotAnInterface"),
                        new Object[] {
                            retValue
                        });
                    TopManager.getDefault().notify(new NotifyDescriptor.Message(
                        msg));
                    continue;
                }
                try {
                    target.addInterface(id);
                    break;
                } catch (SourceException e) {
                    throw new IOException(e.getLocalizedMessage());
                }
            }
        }
    }
}