/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.java.wizard;

import java.util.*;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.openide.cookies.SourceCookie;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.util.NbBundle;

import org.openide.src.*;

import org.netbeans.modules.java.tools.InheritanceSupport;

/**
 * The class encapsulates a state of Java Wizard's iterator shared among various
 * wizard components.
 */
public class JavaWizardData extends Object {
    public static final String PROP_ORIGINAL_CLASS = "originalClass"; // NOI18N
    public static final String PROP_TARGET_PACKAGE = "targetPackage"; // NOI18N
    public static final String PROP_CUSTOMIZED_CLASS = "customizedClass"; // NOI18N
    public static final String PROP_OVERRIDEN_METHODS = "overridenMethods";

    /** Template object that is to be customized.
     */
    private DataObject obj;
    
    /** Original class taken from the template object.
     */
    private ClassElement originalClass;

    /** Class that is being customized. This class is cloned from the original class
     * in the template setter.
     */
    private ClassElement customizedClass;
    
    /** Inheritance data kept for the class. If the class model changes (template
     *  changes, the reference is replaced by completely new inheritance data.
     */
    private InheritanceSupport InheritanceSupport;
    
    /** Utility field holding list of PropertyChangeListeners. */
    private transient java.util.ArrayList propertyChangeListenerList;

    /** Creates new JavaWizardData */
    public JavaWizardData() {
    }

    /** Sets the template object to be operated on.
     * This effectively clones the template class into customized class and fires
     * appropriate property changes.
     */
    public void setTemplate(DataObject obj) {
        if (obj.equals(this.obj))
            return;
        this.obj = obj;
        SourceCookie cookie = (SourceCookie)obj.getCookie(SourceCookie.class);
        ClassElement orig = null;

        if (cookie != null) {
            SourceElement el = cookie.getSource();
            el.prepare().waitFinished();
            ClassElement[] classes = el.getClasses();
            if (classes != null &&
            classes.length > 0) {
                orig = classes[0];
            }
        }
        if (orig == null) {
            orig = new ClassElement();
        }
        ClassElement oldOriginal = this.originalClass;
        ClassElement oldCustomized = this.customizedClass;
        this.originalClass = orig;
        firePropertyChange(PROP_ORIGINAL_CLASS, oldOriginal, orig);
        cloneCustomizedClass();
    }

    /** Creates a clone of the original class and replaces old customized class
     * reference.
     * This is usable for changing templates and for implementing features like
     * `Reset' button in the Wizard.
     */
    private void cloneCustomizedClass() {
        this.customizedClass = new ClassElement();
        try {
            String superclassName;
	    
	    if (originalClass.getSuperclass() != null) {
		superclassName = originalClass.getSuperclass().getFullName();
	    } else {
		superclassName = null;
	    }
            customizedClass.setModifiers(originalClass.getModifiers());
	    if (superclassName != null)
        	customizedClass.setSuperclass(Identifier.create(superclassName));
            customizedClass.setInterfaces(originalClass.getInterfaces());
	    customizedClass.setClassOrInterface(originalClass.isClassOrInterface());
        } catch (SourceException ex) {
        }
        this.InheritanceSupport = new InheritanceSupport(customizedClass);
        firePropertyChange(PROP_CUSTOMIZED_CLASS, null, this.customizedClass);
    }
    
    /** Sets the target location and the name of the clss being created.
     * The method propagates name and package changes to the name of the customized
     * class.
     */
    public void setTarget(FileObject folder, String name) {
        StringBuffer fullName = new StringBuffer();
        
        if (!folder.isRoot()) {
            fullName.append(folder.getPackageName('.'));
            fullName.append('.');
        }
        fullName.append(name);
        try {
            customizedClass.setName(Identifier.create(fullName.toString(), name));
        } catch (SourceException e) {
        }
    }

    /** Returns helper object that deals with inheritance graphs.
     */
    public InheritanceSupport getInheritance() {
        return InheritanceSupport;
    }
    
    /** Returns original class obtained from the template object.
     */
    public ClassElement getOriginalClass() {
        return originalClass;
    }
    
    /** Returns customized class that is to be displayed inside Wizard's panels.
     */
    public ClassElement getCustomizedClass() {
        return customizedClass;
    }
    
    /** Registers PropertyChangeListener to receive events.
     * @param listener The listener to register.
     */
    public synchronized void addPropertyChangeListener(java.beans.PropertyChangeListener listener) {
        if (propertyChangeListenerList == null ) {
            propertyChangeListenerList = new java.util.ArrayList ();
        }
        propertyChangeListenerList.add (listener);
    }
    
    /** Removes PropertyChangeListener from the list of listeners.
     * @param listener The listener to remove.
     */
    public synchronized void removePropertyChangeListener(java.beans.PropertyChangeListener listener) {
        if (propertyChangeListenerList != null ) {
            propertyChangeListenerList.remove (listener);
        }
    }

    private void firePropertyChange(String name, Object oldValue, Object newValue) {
        PropertyChangeEvent evt = new PropertyChangeEvent(this, name, oldValue, newValue);
        firePropertyChange(evt);
    }
    
    /** Notifies all registered listeners about the event.
     *
     * @param e The event to be fired
     */
    private void firePropertyChange(java.beans.PropertyChangeEvent event) {
        java.util.ArrayList list;
        synchronized (this) {
            if (propertyChangeListenerList == null)
                return;
            list = (java.util.ArrayList)propertyChangeListenerList.clone ();
        }
        for (int i = 0; i < list.size (); i++) {
            ((java.beans.PropertyChangeListener)list.get (i)).propertyChange (event);
        }
    }
    
    void setCustomizedClass(ClassElement cls) {
        this.customizedClass = cls;
    }
}
