/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.java.wizard;

import java.io.IOException;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;

import org.openide.WizardDescriptor;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.TemplateWizard;

import org.openide.cookies.SourceCookie;
import org.openide.src.ClassElement;
import org.openide.src.SourceElement;
import org.openide.src.SourceException;

public class JavaWizardIterator implements TemplateWizard.Iterator {
    private static final long serialVersionUID = -1987345873459L;
    /** Array of panels that form the Wizard.
     */
    private transient WizardDescriptor.Panel[] panels;
    
    private static String[] panelNames;
    
    /** Total number of panels that can be displayed.
     */
    private static final int PANEL_COUNT = 5;
    
    private static final int INHERITANCE_PANEL = 1;
    private static final int FIELDS_PANEL = 3;
    
    /** Singleton instance of JavaWizardIterator, should it be ever needed.
     */
    private static JavaWizardIterator instance;
    
    /** Index of the current panel. Panels are numbered from 0 to PANEL_COUNT - 1.
     */
    private transient int panelIndex = 0;
    
    /** Reference to Wizard's private data structures.
     */
    private transient JavaWizardData myData;
    
    private transient TemplateWizard wizardInstance;
    
    private transient boolean guiInitialized;
    
    public JavaWizardIterator() {
    }
    
    /** Returns JavaWizardIterator singleton instance.
     */
    public static synchronized JavaWizardIterator singleton() {
        if (instance == null) {
            instance = new JavaWizardIterator();
        }
        return instance;
    }
    // ========================= TemplateWizard.Iterator ============================

    /** Instantiates the template using informations provided by
     * the wizard.
     *
     * @param wiz the wizard
     * @return set of data objects that has been created (should contain
     *  at least one
     * @exception IOException if the instantiation fails
    */
    public java.util.Set instantiate(TemplateWizard wiz) throws IOException {
        // dig out file from the template.
        DataObject obj = instantiateTemplate(wiz.getTemplate(), wiz.getTargetFolder(), wiz.getTargetName());
        SourceCookie src = (SourceCookie)obj.getCookie(SourceCookie.class);
        if (src == null) {
            return java.util.Collections.singleton(obj);
        }
        ClassElement[] classes = src.getSource().getClasses();
        
        try {
	    ClassElement targetClass;
	    
            // NO classes --> create a new class
	    if (classes == null || classes.length == 0) {
                ClassElement el = new ClassElement();
                el.setName(myData.getCustomizedClass().getName());
                src.getSource().addClass(el);
		targetClass = src.getSource().getClasses()[0];
            } else {
                targetClass = classes[0];
            }
            copy(targetClass, myData.getCustomizedClass());
        } catch (SourceException e) {
            obj.delete();
            throw new IOException(e.getLocalizedMessage());
        }

        // run default action (hopefully should be here)
        org.openide.nodes.Node node = obj.getNodeDelegate ();
        org.openide.util.actions.SystemAction sa = node.getDefaultAction ();
        if (sa != null) {
            sa.actionPerformed (new ActionEvent (node, ActionEvent.ACTION_PERFORMED, "")); // NOI18N
        }
        return java.util.Collections.singleton(obj);
    }
    
    /** Get the current panel.
     * @return the panel
    */
    public WizardDescriptor.Panel current() {
        return panels[panelIndex];
    }
    
    /** Get the name of the current panel.
     * @return the name
    */
    public String name() {
        return panelNames[panelIndex];
    }
    
    /** Test whether there is a next panel.
     * @return <code>true</code> if so
    */
    
    public boolean hasNext() {
        return panelIndex < PANEL_COUNT - 1;
    }
    
    /** Test whether there is a previous panel.
     * @return <code>true</code> if so
    */
    public boolean hasPrevious() {
        return panelIndex > 0;
    }
    
    /** Move to the next panel.
     * I.e. increment its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void nextPanel() {
	if (panelIndex == INHERITANCE_PANEL && !myData.getCustomizedClass().isClassOrInterface()) 
	    panelIndex += 2;
	else
    	    panelIndex++;
    }
    
    /** Move to the previous panel.
     * I.e. decrement its index, need not actually change any GUI itself.
     * @exception NoSuchElementException if the panel does not exist
    */
    public void previousPanel() {
	if (panelIndex == FIELDS_PANEL && !myData.getCustomizedClass().isClassOrInterface())
	    panelIndex -= 2;
	else
    	    panelIndex--;
    }
    
    /** Add a listener to changes of the current panel.
     * The listener is notified when the possibility to move forward/backward changes.
     * @param l the listener to add
    */
    public void addChangeListener(ChangeListener l) {
    }
    
    /** Remove a listener to changes of the current panel.
     * @param l the listener to remove
    */
    public void removeChangeListener(ChangeListener l) {
    }

    
    public void initialize(TemplateWizard wizard) {
        this.wizardInstance = wizard;
	if (!guiInitialized) {
    	initialize();
        myData = new JavaWizardData();
        
        panels = new WizardDescriptor.Panel[PANEL_COUNT];

        // just copy the target chooser as the wizard does not provide its own.
        panels[0] = wizard.targetChooser();
        panels[1] = new InheritancePanel(myData);
        panels[2] = new OverridePanel(myData);
        panels[3] = new FieldsPanel(myData);
        panels[4] = new MethodsPanel(myData);
	}
        reset();
    }
    
    public void uninitialize(TemplateWizard wiz) {
	panels = null;
	myData = null;
	guiInitialized = false;
    }

    // ========================= IMPLEMENTATION ============================
    
    private Object readResolve() {
        initialize();
        reset();
        return this;
    }
    
    /** Instantiates the template. Currently it just delegates to the template DataObject's
     * createFromTemplate method.
     */
    private DataObject instantiateTemplate(DataObject tpl, DataFolder target, String name) throws IOException {
        return tpl.createFromTemplate(target, name);
    }
    
    /** Copy properties of a customized class (model) to the target class.
     * Only class own properties, methods, constructors and fields are copied.
     * Inner classes are not handled by this method.
     * No methods are removed or replaced from/in the original class
     */
    private void copy(ClassElement target, ClassElement model) throws SourceException{
        if (target == null || model == null)
            return;
        target.setSuperclass(model.getSuperclass());
        target.setInterfaces(model.getInterfaces());
        target.setModifiers(model.getModifiers());
        
        // copy sub-elements:
        target.addFields(model.getFields());
        target.addConstructors(model.getConstructors());
        target.addMethods(model.getMethods());
    }
           
    protected void reset() {
        panelIndex = 0;
    }

    /** Initializes the Wizard with array of panels.
     */
    protected void initialize() {
        if (panelNames == null) {
            panelNames = new String[5];
            panelNames[0] = Util.getString("TIT_ChooseTarget");
            panelNames[1] = Util.getString("TIT_InheritanceWizardPanel");
            panelNames[2] = Util.getString("TIT_OverrideWizardPanel");
            panelNames[3] = Util.getString("TIT_AddFieldsWizardPanel");
            panelNames[4] = Util.getString("TIT_AddMethodsWizardPanel");
        }
    }
       
}
