/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient;

import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Set;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;

import org.netbeans.lib.cvsclient.admin.AdminHandler;
import org.netbeans.lib.cvsclient.admin.Entry;

import org.netbeans.lib.cvsclient.command.Command;
import org.netbeans.lib.cvsclient.command.CommandAbortedException;
import org.netbeans.lib.cvsclient.command.CommandException;
import org.netbeans.lib.cvsclient.command.GlobalOptions;

import org.netbeans.lib.cvsclient.connection.Connection;
import org.netbeans.lib.cvsclient.connection.ConnectionModifier;
import org.netbeans.lib.cvsclient.connection.GzipModifier;

import org.netbeans.lib.cvsclient.event.EventManager;

import org.netbeans.lib.cvsclient.file.FileDetails;
import org.netbeans.lib.cvsclient.file.FileHandler;
import org.netbeans.lib.cvsclient.file.UncompressedFileHandler;

import org.netbeans.lib.cvsclient.request.CaseRequest;
import org.netbeans.lib.cvsclient.request.Request;
import org.netbeans.lib.cvsclient.request.GzipStreamRequest;
import org.netbeans.lib.cvsclient.request.GzipFileContentsRequest;
import org.netbeans.lib.cvsclient.request.UnconfiguredRequestException;
import org.netbeans.lib.cvsclient.request.UseUnchangedRequest;
import org.netbeans.lib.cvsclient.request.ValidResponsesRequest;

import org.netbeans.lib.cvsclient.response.Response;
import org.netbeans.lib.cvsclient.response.ResponseException;
import org.netbeans.lib.cvsclient.response.ResponseFactory;
import org.netbeans.lib.cvsclient.response.ResponseServices;

import org.netbeans.lib.cvsclient.util.Logger;
import org.netbeans.lib.cvsclient.util.LoggedDataInputStream;
import org.netbeans.lib.cvsclient.util.LoggedDataOutputStream;
import org.netbeans.lib.cvsclient.util.IgnoreFileFilter;
import org.netbeans.lib.cvsclient.util.DefaultIgnoreFileFilter;


/**
 * The main way of communication with a server using the CVS Protocol. The
 * client is not responsible for setting up the connection with the server,
 * only interacting with it.
 * @see org.netbeans.cvsclient.connection.Connection
 * @author  Robert Greig
 * @version $Revision: 1.31.2.4 $
 */
public class Client implements ClientServices, ResponseServices
{
    /**
     * The connection used to interact with the server.
     */
    protected Connection connection;

    /**
     * The file handler to use.
     */
    protected FileHandler fileHandler = new UncompressedFileHandler();

    /**
     * The modified date.
     */
    protected Date modifiedDate = null;
    
    /**
     * The admin handler to use.
     */
    protected AdminHandler adminHandler;

    /**
     * The local path, ie the path to the directory in which this command
     * was executed.
     */
    private String localPath;

    /**
     * Whether any commands have been executed so far. This allows us to
     * send some initialisation requests before the first command.
     */
    private boolean isFirstCommand = true;

    /**
     * The event manager.
     */
    private EventManager eventManager;

    /**
     * The global options for the executing command.
     */
    private GlobalOptions globalOptions;

    /**
     * This is set to true if we should abort the current command.
     */
    private volatile boolean abort;

    private IgnoreFileFilter ignoreFileFilter;
    
    /**
     * Construct a Client using a given connection and file handler.
     * @param c the connection to use
     * @param f the file handler to use
     * @param a the admin handler to use
     */
    public Client(Connection connection, AdminHandler adminHandler)
    {
        setConnection(connection);
        setAdminHandler(adminHandler);
        eventManager = new EventManager();
        ignoreFileFilter = new DefaultIgnoreFileFilter();
    }

    /**
     * Get the connection used for communicating with the server.
     * Connection.
     * @return the connection
     */
    public Connection getConnection()
    {
        return connection;
    }

    /**
     * Set the connection used for communicating with the server.
     * @param c the connection to use for all communication with the server
     */
    public void setConnection(Connection connection)
    {
        this.connection = connection;
        setIsFirstCommand(true);
    }

    /**
     * Get the admin handler uesd to read and write administrative information
     * about files on the local machine.
     * @return the admin handler
     */
    public AdminHandler getAdminHandler()
    {
        return adminHandler;
    }

    /**
     * Set the admin handler used to read and write administrative information
     * about files on the local machine.
     */
    public void setAdminHandler(AdminHandler adminHandler)
    {
        this.adminHandler = adminHandler;
    }

    /**
     * Get the local path; that is, the path to the directory in which
     * the currently executing command is referring.
     */
    public String getLocalPath()
    {
        return localPath;
    }

    /**
     * Set the local path, i.e. the path to the directory in which all
     * commands are given (top level).
     * @param theLocalPath
     */
    public void setLocalPath(String localPath)
    {
        this.localPath = localPath;
    }

    /**
     * Returns true if no previous command was executed using thiz.
     */
    public boolean isFirstCommand()
    {
        return isFirstCommand;
    }

    /**
     * Set whether this is the first command. Normally you do not need to set
     * this yourself - after execution the first command will have set this to
     * false.
     */
    public void setIsFirstCommand(boolean isFirstCommand)
    {
        this.isFirstCommand = isFirstCommand;
    }

    /**
     * Process all the requests. The connection must have been opened and
     * set first.
     * @param requests the requets to process
     * @see org.greig.cvs.Connection.open()
     */
    public void processRequests(List requests)
        throws IOException, UnconfiguredRequestException, ResponseException,
               CommandAbortedException
    {
        if ((requests==null) || (requests.size()==0))
        {
            throw new IllegalArgumentException("[processRequests] requests " +
                                               "was either null or empty.");
        }

        if (abort)
        {
            throw new CommandAbortedException("Aborted during request " +
                                              "processing");
        }

        // send the initialisation requests if we are handling the first
        // command
        if (isFirstCommand())
        {
            setIsFirstCommand(false);
            // MK - sends the global options that are to be sent to server
            // (-q, -Q, -t, -n, l)
            Collection col = globalOptions.createRequestList();
            Iterator reqIt = col.iterator();
            while (reqIt.hasNext())
            {
                requests.add(1, reqIt.next());
            }

            if (globalOptions.getUseGzip())
            {
                //requests.add(1, new GzipStreamRequest());
                requests.add(1, new GzipFileContentsRequest());
            }

            // on Windows, send the Case request.
            if (System.getProperty("os.name").startsWith("Windows"))
            {
                requests.add(1, new CaseRequest());
            }
            requests.add(1, new UseUnchangedRequest());
            requests.add(1, new ValidResponsesRequest());
        }

        Iterator it = requests.iterator();
        LoggedDataOutputStream dos = connection.getOutputStream();

        // this list stores stream modification requests, each to be called
        // to modify the input stream the next time we need to process a
        // response
        LinkedList streamModifierRequests = new LinkedList();

        while (it.hasNext())
        {
            if (abort)
            {
                throw new CommandAbortedException("Aborted during request " +
                                                  "processing");
            }

            Request r = (Request) it.next();
            dos.writeBytes(r.getRequestString());

            // we must modify the outputstream now, but defer modification
            // of the inputstream until we are about to read a response.
            // This is because some modifiers (e.g. gzip) read the header
            // on construction, and obviously no header is present when
            // no response has been sent
            r.modifyOutputStream(connection);
            if (r.modifiesInputStream())
            {
                streamModifierRequests.add(r);
            }
            dos = connection.getOutputStream();

            FileDetails fileDetails = r.getFileForTransmission();
            if (fileDetails != null)
            {
                // only transmit the file if it exists! When committing
                // a remove request you cannot transmit the file
                if (fileDetails.getFile().exists())
                {
                    if (fileDetails.isBinary())
                    {
                        fileHandler.transmitBinaryFile(fileDetails.getFile(),
                                                       dos);
                    }
                    else
                    {
                        fileHandler.transmitTextFile(fileDetails.getFile(),
                                                     dos);
                    }

                    Logger.logOutput("<Sending file: " + fileDetails.getFile().
                                                         getPath() + ">\n");
                }
            }
            if (r.isResponseExpected())
            {
                dos.flush();

                // now perform the deferred modification of the input stream
                Iterator modifiers = streamModifierRequests.iterator();
                while (modifiers.hasNext())
                {
//                    System.err.println("Modifying the inputstream...");
                    final Request request = (Request) modifiers.next();
//                    System.err.println("Request is a: " + request.getClass().getName());
                    request.modifyInputStream(connection);
                }
                streamModifierRequests.clear();
                handleResponse();
            }
        }
        dos.flush();
    }

    /**
     * Handle the response from a request.
     * @throws ResponseException if there is a problem reading the response
     */
    private void handleResponse()
        throws ResponseException, CommandAbortedException
    {
        try
        {
            LoggedDataInputStream dis = connection.getInputStream();

            StringBuffer responseName;
            int ch;
            ch = dis.readByte();
            // read in all the responses

            while (ch != -1)
            {
                responseName = new StringBuffer();
                // read in the response name
                while ((ch != -1) && (((char)ch) != '\n') &&
                       (((char)ch) != ' '))
                {
                    responseName.append((char)ch);
                    ch = dis.readByte();
                }
                //System.err.println("Reader object is: " + dis.hashCode());
                //System.err.println("Response is: " + responseName);
                Response response = ResponseFactory.createResponse(responseName.
                                                                   toString());

                response.process(dis, this);
                if (response.isTerminalResponse() || abort)
                {
                    break;
                }
                ch = dis.readByte();
            }

            if (abort)
            {
                throw new CommandAbortedException("Aborted during request " +
                                                  "processing");
            }
        }
        catch (IOException e)
        {
            throw new ResponseException(e);
        }
    }

    /**
     * Execute a command.
     * @param command the command to execute
     * @param options the global options to use for executing the command
     * @throws CommandException if an error occurs when executing the command
     * @throws CommandAbortedException if the command is aborted
     */
    public void executeCommand(Command command, GlobalOptions options)
        throws CommandException, CommandAbortedException
    {
        globalOptions = options;
        try
        {
            eventManager.addCVSListener(command);
            command.execute(this, eventManager);
        }
        finally
        {
            eventManager.removeCVSListener(command);
        }
    }

    /**
     * Convert a <i>pathname</i> in the CVS sense (see 5.10 in the protocol
     * document) into a local absolute pathname for the file.
     * @param localDirectory the name of the local directory, relative to the
     * directory in which the command was given
     * @param repository the full repository name for the file
     */
    public String convertPathname(String localDirectory, String repository)
    {
        int lastIndexOfSlash = repository.lastIndexOf('/');
        String filename = repository.substring(lastIndexOfSlash+1);
        // note that the localDirectory ends in a slash
        return getLocalPath() + "/" + localDirectory + filename;
    }

    /**
     * Get the repository.
     * @return the repository path, e.g. /home/bob/cvs. Delegated to the
     * Connection in this case
     * @see Connection.getRepository()
     */
    public String getRepository()
    {
        return connection.getRepository();
    }

    /**
     * Create or update the administration files for a particular file.
     * This will create the CVS directory if necessary, and the
     * Root and Repository files if necessary. It will also update
     * the Entries file with the new entry
     * @param localDirectory the local directory, relative to the directory
     * in which the command was given, where the file in question lives
     * @param repositoryPath the path of the file in the repository, in
     * absolute form.
     * @param entry the entry object for that file
     */
    public void updateAdminData(String localDirectory, String repositoryPath,
                                Entry entry)
        throws IOException
    {
        final String absolutePath = localPath + "/" + localDirectory;
        repositoryPath = repositoryPath.substring(getRepository().length()+1);
        adminHandler.updateAdminData(absolutePath, repositoryPath, entry,
                                     globalOptions);
    }

    /**
     * Set the modified date of the next file to be written. The next call
     * to write<Type>File will use this date.
     * @param modifiedDate the date the file should be marked as modified
     */
    public void setNextFileDate(Date modifiedDate)
    {        
        this.modifiedDate = modifiedDate;        
    }

    /**
     * Get the modified date for the next file.
     * @return the date and then null the instance variable.
     */
    public Date getNextFileDate()
    {
        //
        // We null the instance variable so that future calls will not
        // retrieve a date specified for a previous file
        //
        Date copy = modifiedDate;
        modifiedDate = null;
        return copy;
    }
    
    /**
     * Get the Entry for the specified file, if one exists.
     * @param f the file
     * @throws IOException if the Entries file cannot be read
     */
    public Entry getEntry(File f)
        throws IOException
    {
        return adminHandler.getEntry(f);
    }

    /**
     * Get the entries for a specified directory.
     * @param directory the directory for which to get the entries
     * @return an iterator of Entry objects
     */
    public Iterator getEntries(File directory)
        throws IOException
    {
        return adminHandler.getEntries(directory);
    }

    /**
     * Get the repository path for a given directory, for example in
     * the directory /home/project/foo/bar, the repository directory
     * might be /usr/cvs/foo/bar. The repository directory is commonly
     * stored in the file <pre>Repository</pre> in the CVS directory on
     * the client. (This is the case in the standard CVS command-line tool)
     * @param directory the directory
     */
    public String getRepositoryForDirectory(String directory)
        throws IOException
    {
        String localPathRepos;
        if (directory.equals(getLocalPath()))
        {
            try
            {
                localPathRepos = adminHandler.getRepositoryForDirectory(directory,
                                                    connection.getRepository());
            }
            catch (IOException ex)
            {
                localPathRepos = connection.getRepository();
            }
        }
        else
        {
            localPathRepos = adminHandler.getRepositoryForDirectory(directory,
                                                connection.getRepository());
        }
        return localPathRepos;
    }

    /**
     * Set the Entry for the specified file.
     * @param f the file
     * @param e the new entry
     * @throws IOException if an error occurs writing the details
     */
    public void setEntry(File f, Entry e)
        throws IOException
    {
        adminHandler.setEntry(f, e);
    }    

    /**
     * Remove the Entry for the specified file.
     * @param f the file whose entry is to be removed
     * @throws IOException if an error occurs writing the Entries file
     */
    public void removeEntry(File f)
        throws IOException
    {
        adminHandler.removeEntry(f);
    }

    /**
     * Remove the specified file from the local disk.
     * @param pathname the full path to the file to remove
     * @throws IOException if an IO error occurs while removing the file
     */
    public void removeLocalFile(String pathname)
        throws IOException
    {
        fileHandler.removeLocalFile(pathname);
    }

    /**
     * Removes the specified file determined by pathName and repositoryName.
     * In this implementation the pathName is irrelevant.
     */
    public void removeLocalFile(String pathName, String repositoryName)
        throws IOException
    {
        String repository = getRepositoryForDirectory(getLocalPath());
        String relativeRepositoryName = repositoryName.substring(repository.length() + 1);

        File fileToDelete = new File(getLocalPath(), relativeRepositoryName);

        removeLocalFile(fileToDelete.getAbsolutePath());
        removeEntry(fileToDelete);
    }

    /**
     * Rename the local file.
     * @param pathname the full path to the file to rename
     * @param newName the new name of the file (not the full path)
     * @throws IOException if an IO error occurs while renaming the file
     */
    public void renameLocalFile(String pathname, String newName)
        throws IOException
    {
        fileHandler.renameLocalFile(pathname, newName);
    }

    /**
     * Get the CVS event manager. This is generally called by response handlers
     * that want to fire events.
     * @return the eventManager
     */
    public EventManager getEventManager()
    {
        return eventManager;
    }

    /**
     * Call this method to abort the current command. The command will be
     * aborted at the next suitable time
     */
    public synchronized void abort()
    {
        abort = true;
    }

    /**
     * Get all the file and directories contained within a given
     * directory that are <b>known to CVS</b>.
     * @param directory the directory to look in
     * @return a set of all files and directories.
     */
    public Set getAllFiles(File directory) throws IOException
    {
        return adminHandler.getAllFiles(directory);
    }
    
    public void setIgnoreFileFilter(IgnoreFileFilter filter)
    {
        ignoreFileFilter = filter;
    }
    
    public IgnoreFileFilter getIgnoreFileFilter() 
    {
        return ignoreFileFilter;
    }
    
    public boolean shouldBeIgnored(File directory, String noneCvsFile)
    {
        if (ignoreFileFilter != null) 
        {
            return ignoreFileFilter.shouldBeIgnored(directory, noneCvsFile);
        }
        return false;
    }
    
    /**
     * Checks for presence of CVS/Tag file and returns it's value.
     * @returns the value of CVS/Tag file for the specified directory
     *          null if file doesn't exist
     */        
    
    public String getStickyTagForDirectory(File directory) 
    {
        return adminHandler.getStickyTagForDirectory(directory);
    }
}
