/*****************************************************************************
 * Sun Public License Notice

 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Date;
import java.util.TimeZone;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;


import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.admin.Entry;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;

import org.netbeans.lib.cvsclient.request.ArgumentRequest;
import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.StickyRequest;
import org.netbeans.lib.cvsclient.request.EntryRequest;
import org.netbeans.lib.cvsclient.request.ModifiedRequest;
import org.netbeans.lib.cvsclient.request.RootRequest;
import org.netbeans.lib.cvsclient.request.UnchangedRequest;
import org.netbeans.lib.cvsclient.request.UpdateRequest;
import org.netbeans.lib.cvsclient.request.QuestionableRequest;
import org.netbeans.lib.cvsclient.util.IgnoreFileFilter;

/**
 * A class that provides common functionality for many of the CVS command
 * that send similar sequences of requests.
 * @author  Robert Greig
 * @version $Revision: 1.30.2.3 $
 */
public abstract class BasicCommand extends BuildableCommand
{
    /**
     * The requests that are sent and processed.
     */
    protected LinkedList requests = new LinkedList();

    /**
     * The client services that are provided to this command.
     */
    protected ClientServices clientServices;

    /**
     * Whether to update recursively.
     */
    protected boolean recursive = true;

    /**
     * The files and/or directories to operate on.
     */
    protected File[] files;
    
    /**
     * Gets the value of the recursive option.
     * @return true if recursive, false if not
     */
    public boolean getRecursive()
    {
        return recursive;
    }

    /**
     * Sets the value of the recursive option.
     * @param r true if the command should recurse, false otherwise
     */
    public void setRecursive(boolean r)
    {
        recursive = r;
    }

    /**
     * Set the files and/or directories on which to execute the command.
     * The way these are processed is:<P>
     * <UL><LI>Default action (i.e. not setting the files explicitly or
     * setting them to <pre>null</pre>) is to use the directory in which
     * the command was executed (see how directories are treated, below)</LI>
     * <LI>Files are handled how you would expect</LI>
     * <LI>For directories, all files within the directory are sent</LI></UL>
     * @param theFiles the files to operate on. May be null to indicate that the
     * local directory specified in the client should be used. Full, absolute
     * canonical pathnames <b>must</b> be supplied.
     */
    public void setFiles(File[] theFiles)
    {
        // sort array.. files first, directories follow
        if (theFiles == null)
        {
            files = theFiles;
            return;
        }

        files = new File[theFiles.length];
        int fileCount = 0;
        int dirCount = 0;
        int totalCount = theFiles.length;
        for (int index = 0; index < totalCount; index++)
        {
            File currentFile = theFiles[index];
            if (currentFile.isDirectory())
            {
                files[totalCount - (1 + dirCount)] = currentFile;
                dirCount = dirCount + 1;
            }
            else
            {
                files[fileCount] = currentFile;
                fileCount = fileCount + 1;
            }
        }
    }

    /**
     * Get the files and/or directories specified for this command to operate
     * on.
     * @return the array of Files
     */
    public File[] getFiles()
    {
        return files;
    }

    /**
     * Get a single file from the "files" list. returns only files, not directories.
     * This method is used from within the builders, because for single file requests, the
     * cvs server doesn't return us enough information to identify what file has been returned.
     * Thus we sort the "files" array (files come before directories. Then the response froms erver comes in the same order
     * and the files can be found this way.
     * @param index the index of the file in the list.
     *
     */
    public File getXthFile(int index)
    {
        if (index < 0 || index >= files.length)
        {
            return null;
        }
        File file = files[index];
        if (!file.isFile())
        {
            return null;
        }
        return file;
    }

    /**
     * @param ending - the ending part of the file's pathname.. path separator is cvs's default '/'
     */
    public File getFileEndingWith(String ending)
    {
        int index = 0;
        for (index = 0; index < files.length; index++)
        {            
            String path = files[index].getAbsolutePath();
            path = path.replace('\\', '/');
            if (path.endsWith(ending))
            {
                return files[index];
            }
        }
        return null;
    }

    /**
     * Add the appropriate requests for a specified path. For a directory,
     * process all the files within that directory and for a single file,
     * just send it. For each directory, send a directory request. For each
     * file send an Entry request followed by a Modified request.
     * @param path the particular path to issue requests for. May be
     * either a file or a directory.
     */
    protected void addRequests(File path)
        throws FileNotFoundException, IOException
    {
        if (path == null)
        {
            throw new IllegalArgumentException("Cannot add requests for a " +
                                               "null path.");
        }

        if (!path.exists() || path.isFile())
        {
            addRequestsForFile(path);
        }
        else
        {
            addRequestsForDirectory(path);
        }
    }

    /**
     * Should return true if unchanged files should not be sent to server.
     * If false is returned, all files will be sent to server
     * This method is used by <code>sendEntryAndModifiedRequests</code>.
     */
    protected boolean doesCheckFileTime()
    {
        return true;
    }

    /**
     * Send an Entry followed by a Modified or Unchanged request based on
     * whether the file has been untouched on the local machine.
     *
     * @param entry         the entry for the file
     * @param file          the file in question
     */
    protected void sendEntryAndModifiedRequests(Entry   entry,
                                                File    file)
    {
        if (entry == null)
        {
            return;
        }

        Date entryLastModified = entry.getLastModified();
        boolean hadConflicts = entry.hadConflicts();
        if (!hadConflicts)
        {
            // we null out the conflict field if there is no conflict
            // because we use that field to store the timestamp of the
            // file (just like command-line CVS). There is no point
            // in sending this information to the CVS server, even
            // though it should be ignored by the server.
            entry.setConflict(null);
        }
        requests.add(new EntryRequest(entry));

        if (file == null || !file.exists() || entry.isUserFileToBeRemoved())
        {
            return;
        }

        if (doesCheckFileTime() && !hadConflicts && entryLastModified != null)
        {
            // we divide by 1000 because we are only interested in the
            // time to the nearest second
            if (entryLastModified.getTime()/1000 == file.lastModified()/1000)
            {
                requests.add(new UnchangedRequest(file.getName()));
                return;
            }
        }

        requests.add(new ModifiedRequest(file, entry.isBinary()));
    }

    /**
     * Adds the appropriate requests for a given directory. Sends a
     * directory request followed by as many Entry and Modified requests
     * as required
     * @param directory the directory to send requests for
     * @throws IOException if an error occurs constructing the requests
     */
    protected void addRequestsForDirectory(File directory)
        throws IOException
    {
        // remove localPath prefix from directory. If left with
        // nothing, use dot (".") in the directory request. Also remove the
        // trailing slash
        String dir = getRelativeToLocalPathInUnixStyle(directory);
        
        try
        {            
            try
            {
                String repository = clientServices.getRepositoryForDirectory(
                                                   directory.getAbsolutePath());                
                requests.add(new DirectoryRequest(dir, repository));
                String tag = clientServices.getStickyTagForDirectory(directory);
                if (tag != null) {
                    requests.add(new StickyRequest(tag));
                }
            }
            catch (FileNotFoundException e)
            {                
                // we can ignore this exception safely because it just means
                // that the user has deleted a directory referenced in a
                // CVS/Entries file
            }
        }
        catch (IOException e)
        {
//            System.err.println("An error occurred reading the repository " +
//                               "for the directory " + dir + ": " + e);
            e.printStackTrace();
        }

        LinkedList subdirectories = null;
        
        List localFiles = new LinkedList();
        File[] files = directory.listFiles();
        if (files != null) {
        for (int i = 0; i < files.length; i++) 
        {
            if (!files[i].isDirectory()) 
            {
                localFiles.add(files[i]);
            }
        }
        }
        
        if (recursive)
        {
            subdirectories = new LinkedList();
        }
        // get all the entries we know about, and process them
        Iterator it = clientServices.getEntries(directory);
        while (it.hasNext())
        {
            final Entry entry = (Entry)it.next();
            if (entry.isDirectory())
            {
                if (recursive)
                {
                    subdirectories.add(new File(directory, entry.getName()));
                }
            }
            else
            {
                final File theFile = new File(directory, entry.getName());
                sendEntryAndModifiedRequests(entry, theFile);
                localFiles.remove(theFile);
            }
        }
        if (localFiles.size() > 0) 
        {
            Iterator itFiles = localFiles.iterator();
            while (itFiles.hasNext()) 
            {
                File fl = (File)itFiles.next();
                if (!clientServices.shouldBeIgnored(directory, fl.getName()))
                {
                    requests.add(new QuestionableRequest(fl.getName()));
                }
            }
        }
        if (recursive)
        {
            it = subdirectories.iterator();
            while (it.hasNext())
            {
                addRequestsForDirectory((File)it.next());
            }
        }

    }

    /**
     * Add the appropriate requests for a single file. A directory request
     * is sent, followed by an Entry and Modified request
     * @param file the file to send requests for
     * @throws IOException if an error occurs constructing the requests
     */
    protected void addRequestsForFile(File file)
        throws IOException
    {
        final File parentDirectory = file.getParentFile();
        // remove localPath prefix from directory. If left with
        // nothing, use dot (".") in the directory request
        String dir = getRelativeToLocalPathInUnixStyle(parentDirectory);

        try
        {
            String repository = clientServices.getRepositoryForDirectory(
                                             parentDirectory.getAbsolutePath());
            requests.add(new DirectoryRequest(dir, repository));
            String tag = clientServices.getStickyTagForDirectory(parentDirectory);
            if (tag != null) {
                requests.add(new StickyRequest(tag));
            }
        }
        catch (IOException e)
        {
            System.err.println("An error occurred reading the respository " +
                               "for the directory " + dir + ": " + e);
            e.printStackTrace();
        }

        try
        {
            final Entry entry = clientServices.getEntry(file);
            // a non-null entry means the file does exist in the
            // Entries file for this directory
            if (entry != null)
            {
                sendEntryAndModifiedRequests(entry, file);
            }            
        }
        catch (IOException e)
        {
            System.err.println("An error occurred getting the Entry "+
                    "for file " + file + ": " + e);
            e.printStackTrace();
        }
    }

    /**
     * Add the argument requests. The argument requests are created using
     * the original set of files/directories passed in. Subclasses of this
     * class should call this method at the appropriate point in their
     * execute() method. Note that arguments are appended to the list.
     */
    protected void addArgumentRequests()
    {
       //argumentRequests = new LinkedList();
       if (files != null)
       {
           for (int i=0; i < files.length; i++)
           {
               final File f = files[i];
               if (!f.exists() || f.isFile())
               {
                   final File parentDirectory = f.getParentFile();
                    // remove localPath prefix from directory. If left with
                    // nothing, use dot (".") in the directory request
                    final String dir = getRelativeToLocalPathInUnixStyle(
                                                               parentDirectory);
                    requests.add(new ArgumentRequest(dir + "/" +
                                                             f.getName()));
               }
               else
               {
                   String dir = getRelativeToLocalPathInUnixStyle(f);
                   requests.add(new ArgumentRequest(dir));
               }
           }
       }
    }
        
    /**
     * Execute a command. This implementation sends a Root request, followed
     * by as many Directory and Entry requests as is required by the recurse
     * setting and the file arguments that have been set. Subclasses should
     * call this first, and tag on the end of the requests list any further
     * requests and, finally, the actually request that does the command (e.g.
     * <pre>update</pre>, <pre>status</pre> etc.)
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests
     * @throws CommandException if an error occurs executing the command
     */
    public void execute(ClientServices client, EventManager em)
        throws CommandException
    {
        requests.clear();
        super.execute(client, em);

        clientServices = client;               
        
        if (client.isFirstCommand())
        {            
            requests.add(new RootRequest(client.getRepository()));
        }

        try
        {
            if (files != null && files.length > 0)
            {
                for (int i=0; i < files.length; i++)
                {
                    addRequests(files[i]);
                }
            }
            else
            {
                // if no arguments have been specified, then specify the
                // local directory - the "top level" for this command
                addRequests(new File(getLocalPath()));
            }            
        }
        catch (Exception e)
        {
            throw new CommandException(e);
        }
    }
}

