/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.event.TerminationEvent;

import org.netbeans.lib.cvsclient.ClientServices;

/**
 * A class that provides common functionality for many of the CVS command
 * that send similar sequences of requests.
 * @author  Robert Greig
 * @version $Revision: 1.10 $
 */
public abstract class BuildableCommand extends Command {

    /**
     * An implementation of Builder interface that constructs a FileContainerInfo object from
     * the server's output..
     */
    protected Builder builder;

    private StringBuffer taggedLine;

    /**
     * A boolean value indicating if the user has used the setBuilder() method.
     */
    protected boolean builderSet;

    /**
     * Execute a command. This implementation sends a Root request, followed
     * by as many Directory and Entry requests as is required by the recurse
     * setting and the file arguments that have been set. Subclasses should
     * call this first, and tag on the end of the requests list any further
     * requests and, finally, the actually request that does the command (e.g.
     * <pre>update</pre>, <pre>status</pre> etc.)
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests
     * @throws CommandException if an error occurs executing the command
     */
    public void execute(ClientServices client, EventManager eventManager)
        throws CommandException
    {
        super.execute(client, eventManager);

        if (builder == null && !builderSet)
        {
            builder = createBuilder(eventManager);
        }
    }

    /**
     * Method that is called while the command is being executed.
     * Descendants can override this method to return a Builder instance
     * that will parse the server's output and create data structures.
     */
    protected Builder createBuilder(EventManager eventManager)
    {
        return null;
    }

    public void messageSent(MessageEvent e)
    {
        if (builder != null)
        {
            if (e.isTagged())
            {
                String line = e.getMessage();
                if (line.charAt(0) == '+' || line.charAt(0) == '-')
                {
                    return;
                }
                if (taggedLine == null)
                {
                    taggedLine = new StringBuffer();
                }
                if (line.equals("newline")) {
                    builder.parseLine(taggedLine.toString(), false);
                    taggedLine = null;
                }
                int index = line.indexOf(' ');
                if (index > 0)
                {
                    taggedLine.append(line.substring(index + 1));
                }
            }
            else
            {
                builder.parseLine(e.getMessage(), e.isError());
            }
        }
    }

    /**
     * Used for setting user-defined builder.
     * Can be also set null, in that case the builder mechanism is not used at 
     * all.
     */
    public void setBuilder(Builder builder)
    {        
        this.builder = builder;
        builderSet = true;
    }

    /**
     * Called when server responses with "ok" or "error", (when the command finishes).
     */
    public void commandTerminated(TerminationEvent e)
    {
        if (builder != null)
        {
            if (taggedLine != null)
            {
                builder.parseLine(taggedLine.toString(), false);
                taggedLine = null;
            }
            builder.outputDone();
        }
    }
}
