/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command;

import java.io.File;

import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.command.CommandException;

import org.netbeans.lib.cvsclient.event.CVSListener;
import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.event.ModuleExpansionEvent;
import org.netbeans.lib.cvsclient.event.FileAddedEvent;
import org.netbeans.lib.cvsclient.event.FileRemovedEvent;
import org.netbeans.lib.cvsclient.event.FileUpdatedEvent;
import org.netbeans.lib.cvsclient.event.FileInfoEvent;
import org.netbeans.lib.cvsclient.event.TerminationEvent;

/**
 * All commands must extend this class. A command is essentially a
 * collection of requests that make up what is logically known as a CVS
 * command (from a user's perspective). Commands correspond to operations the
 * user can perform with CVS, for example checkout a module or perform a
 * diff on two file versions.<br>
 * Commands are automatically added as CVS event listeners. They can act
 * on particular events and perhaps fire new events.
 * @author  Robert Greig
 * @version $Revision: 1.16 $
 */
public abstract class Command implements CVSListener, Cloneable
{
    /**
     * The local directory from which the command is being run.
     * This gives us the ability to construct a full pathname for the file
     * which we are processing. The information from the responses alone is
     * not enough.
     */
    protected String localPath;

    /**
     * Execute this command.
     * @param client the client services object that provides any necessary
     *               services to this command, including the ability to actually
     *               process all the requests
     * @param e      the event manager. The command can use this to fire events
     *               if necessary - for example, while parsing status responses.
     */
    public void execute(ClientServices client, EventManager e)
        throws CommandException, CommandAbortedException
    {
        setLocalPath(client.getLocalPath());
    }


    /** This method returns how the command would looklike when typed on the command line.
     *  Each command is responsible for constructing this information.
     *  @returns <command's name> [<parameters>] files/dirs. Example: checkout -p CvsCommand.java
     *
     */
    public abstract String getCVSCommand();

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
     */
    public abstract String getCVSArguments();

    /** takes the arguments and sets the command. To be mainly
     *  used for automatic settings (like parsing the .cvsrc file)
     *  @return true if the option (switch) was recognized and set
     */
    public abstract boolean setCVSCommand(char opt, String optArg);

    /**
     * resets all switches in the command. After calling this method,
     * the command should have no switches defined and should behave defaultly.
     */
    public abstract void resetCVSCommand();

    /**
     * String returned by this method defines which options are available for this particular command
     */
    public abstract String getOptString();

    /**
     * This method just calls the Object.clone() and makes it public.
     */
    public Object clone() throws CloneNotSupportedException
    {
        return super.clone();
    }

    /**
     * Called when the server wants to send a message to be displayed to
     * the user. The message is only for information purposes and clients
     * can choose to ignore these messages if they wish.
     * @param e the event
     */
    public void messageSent(MessageEvent e)
    {
    }

    /**
     * Called when a file has been added.
     * @param e the event
     */
    public void fileAdded(FileAddedEvent e)
    {
    }

    /**
     * Called when a file is removed.
     * @param e the event
     */
    public void fileRemoved(FileRemovedEvent e)
    {
    }

    /**
     * Called when a file has been updated.
     * @param e the event
     */
    public void fileUpdated(FileUpdatedEvent e)
    {
    }

    /**
     * Called when file status information has been received.
     */
    public void fileInfoGenerated(FileInfoEvent e)
    {
    }

    /**
     * Called when server responses with "ok" or "error", (when the command finishes).
     */
    public void commandTerminated(TerminationEvent e)
    {
    }

    /**
     * This is called when the servers has responded to an expand-modules 
     * request.
     * @throws CommandException if something goes wrong when processing a 
     * command based on this event.
     */
    public void moduleExpanded(ModuleExpansionEvent e)
    {
    }
    
    /**
     * Returns the local path the command is associated with.
     */
    public final String getLocalPath()
    {
        return localPath;
    }

    /**
     * Returns the relative path of the specified file (relative to the set
     * local path).
     * Backward slashes will be replaced by forward slashes.
     */
    public final String getRelativeToLocalPathInUnixStyle(File file)
    {
        String filePath = file.getAbsolutePath();
        int startIndex = localPath.length() + 1;
        if (startIndex >= filePath.length())
        {
            return ".";
        }
        String relativePath = filePath.substring(startIndex);
        return relativePath.replace('\\', '/');
    }

    /**
     * Sets the local path for the command.
     */
    protected final void setLocalPath(String localPath)
    {
        this.localPath = localPath;
    }            
}
