/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version 
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command;

import org.netbeans.lib.cvsclient.request.GlobalOptionRequest;
import java.util.Collection;
import java.util.LinkedList;

/**
 * Provides access to global options for a specific command.
 * These are options traditionally set in
 * the command line CVS tool before the command name, for example in the
 * command: <pre>cvs -n update -dP</pre> -n is a global options but
 * -dP are options specific to the update command.
 * <br>Note that you can have different global options for each command
 * you execute (just like command-line CVS).
 * @author  Robert Greig
 * @version $Revision: 1.7 $
 */
public class GlobalOptions implements Cloneable
{
    /**
     * Whether to make checked out files read only (read/write is the default)
     */
    private boolean checkedOutFilesReadOnly = false;
    
    /**
     * The CVS root to use
     */
    private String cvsRoot;
    
    /**
     * Whether to use Gzip
     */
    private boolean useGzip;
    
    /**
     * performs the command but doesn't change any files.
     */
    private boolean doNoChanges = false;
    
    /** 
     * Supresses logging of teh command in CVSROOT/history in the repository. 
     */
    private boolean noHistoryLogging = false;
    
    /** 
     * The cvs gets more quiet than without this switch. However it still 
     * prints the important stuff.
     * Note: If this switch is used, the Builder stuff and parsing of the output might break.
     */
    private boolean moderatelyQuiet = false;
    
    /** 
     * Is even more quiet. Commands which primary function is to send info, 
     *  still do print something (diff, etc.), however other command are completely
     * quiet.
     * Note: If this switch is used, the Builder stuff and parsing of the output might break.
     */
    private boolean veryQuiet = false;

    /** 
     * Traces the execution of the command. Useful for tracing down what is 
     * causing problems.
     * Note: If this switch is used, the Builder stuff and parsing of the output might break.
     */
    private boolean traceExecution = false;    
    
    /**
     *  creates a list of requests. Only those global options are included that 
     *  can be sent to server. (-q, -Q, -l, -t, -r, -n).
     *  To be added to the request list sent to the server.
     */
    public Collection createRequestList() 
    {
        LinkedList toReturnList = new LinkedList();
        if (isDoNoChanges())
        {
            toReturnList.add(new GlobalOptionRequest("-n"));
        }
        if (isNoHistoryLogging()) 
        {
            toReturnList.add(new GlobalOptionRequest("-l"));
        }
        if (isModeratelyQuiet())
        {
            toReturnList.add(new GlobalOptionRequest("-q"));
        }
        if (isVeryQuiet())
        {
            toReturnList.add(new GlobalOptionRequest("-Q"));
        }
        if (isTraceExecution())
        {
            toReturnList.add(new GlobalOptionRequest("-t"));
        }
        return toReturnList;
    }
    
    /**
     * Equals to the Command.getCVSCommand() functionality.
     * Returns all the current switches in the command-line cvs style.
     */
    public String getCVSCommand() 
    {
        StringBuffer toReturn = new StringBuffer();
        boolean atLeastOne = false;
        StringBuffer switches = new StringBuffer();
        if (isDoNoChanges())
        {
            switches.append("-n ");
            atLeastOne = true;
        }
        if (isNoHistoryLogging()) 
        {
            switches.append("-l ");
            atLeastOne = true;
        }
        if (isModeratelyQuiet())
        {
            switches.append("-q ");
            atLeastOne = true;
        }
        if (isVeryQuiet())
        {
            switches.append("-Q ");
            atLeastOne = true;
        }
        if (isTraceExecution())
        {
            switches.append("-t ");
            atLeastOne = true;
        }
        if (atLeastOne) 
        {
            toReturn.append("-" + switches.toString());
        }
        return toReturn.toString();
    }    
    
    /**
     * Are checked out files read only
     * @return the answer
     */    
    public boolean areCheckedOutFilesReadOnly()
    {
        return checkedOutFilesReadOnly;
    }
    
    /**
     * Set whether checked out files are read only. False is the default.
     * @param readOnly true for readonly, false for read/write (default)
     */
    public void setCheckedOutFilesReadOnly(boolean readOnly)
    {
        checkedOutFilesReadOnly = readOnly;
    }
    
    /**
     * Get the CVS root
     * @return the CVS root value, e.g. :pserver:user@host@/usr/local/cvs
     */
    public String getCVSRoot()
    {
        return cvsRoot;
    }
    
    /**
     * Set the CVS root
     * @return the value of CVS to use
     */
    public void setCVSRoot(String root)
    {
        cvsRoot = root;
    }
    
    /**
     * Set whether to use Gzip for file transmission/reception
     * @param useGzip true if gzip should be used, false otherwise
     */
    public void setUseGzip(boolean useGzip)
    {
        this.useGzip = useGzip;
    }
    
    /** 
     * Get whether to use Gzip
     * @return true if Gzip should be used, false otherwise
     */
    public boolean getUseGzip()
    {
        return useGzip;
    }
    
    /** Getter for property doNoChanges.
     * @return Value of property doNoChanges.
     */
    public boolean isDoNoChanges() 
    {
        return doNoChanges;
    }
    
    /** Setter for property doNoChanges.
     * @param doNoChanges New value of property doNoChanges.
     */
    public void setDoNoChanges(boolean doNoChanges) 
    {
        this.doNoChanges = doNoChanges;
    }
    

        
    /** Getter for property noHistoryLogging.
     * @return Value of property noHistoryLogging.
     */
    public boolean isNoHistoryLogging() 
    {
        return noHistoryLogging;
    }
    
    /** Setter for property noHistoryLogging.
     * @param noHistoryLogging New value of property noHistoryLogging.
     */
    public void setNoHistoryLogging(boolean noHistoryLogging) 
    {
        this.noHistoryLogging = noHistoryLogging;
    }
    
    /** Getter for property moderatelyQuiet.
     * @return Value of property moderatelyQuiet.
     */
    public boolean isModeratelyQuiet() 
    {
        return moderatelyQuiet;
    }
    
    /** Setter for property moderatelyQuiet.
     * @param moderatelyQuiet New value of property moderatelyQuiet.
     */
    public void setModeratelyQuiet(boolean moderatelyQuiet) 
    {
        this.moderatelyQuiet = moderatelyQuiet;
    }
    
    /** Getter for property veryQuiet.
     * @return Value of property veryQuiet.
     */
    public boolean isVeryQuiet() 
    {
        return veryQuiet;
    }
    
    /** Setter for property veryQuiet.
     * @param veryQuiet New value of property veryQuiet.
     */
    public void setVeryQuiet(boolean veryQuiet) 
    {
        this.veryQuiet = veryQuiet;
    }
    
    /** Getter for property traceExecution.
     * @return Value of property traceExecution.
     */
    public boolean isTraceExecution() 
    {
        return traceExecution;
    }
    
    /** Setter for property traceExecution.
     * @param traceExecution New value of property traceExecution.
     */
    public void setTraceExecution(boolean traceExecution) 
    {
        this.traceExecution = traceExecution;
    }

    /** 
     * This method just calls the Object.clone() and makes it public.
     */
    public Object clone() throws CloneNotSupportedException
    {
        return super.clone();
    }        
}