/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.add;

import java.io.File;
import java.io.IOException;

import java.util.LinkedList;
import java.util.StringTokenizer;
import java.util.Iterator;

import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.admin.Entry;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;

import org.netbeans.lib.cvsclient.request.AddRequest;
import org.netbeans.lib.cvsclient.request.ArgumentRequest;
import org.netbeans.lib.cvsclient.request.ArgumentxRequest;
import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.StickyRequest;
import org.netbeans.lib.cvsclient.request.EntryRequest;
import org.netbeans.lib.cvsclient.request.ModifiedRequest;
import org.netbeans.lib.cvsclient.request.RootRequest;

import org.netbeans.lib.cvsclient.command.BuildableCommand;
import org.netbeans.lib.cvsclient.command.CommandException;
import org.netbeans.lib.cvsclient.command.Builder;

/**
 * Adds a file or directory.
 * @author  Robert Greig
 * @version $Revision: 1.6.2.6 $
 */
public class AddCommand extends BuildableCommand
{
    /**
     * Constants that identify a message that creates a directory in
     * repository.
     */
    private static final String DIR_ADDED = " added to the repository";
    private static final String DIRECTORY = "Directory ";

    /**
     * The requests that are sent and processed.
     */
    private LinkedList requests;

    /**
     * The argument requests that are collected and sent in the end just before the
     * add request.
     */
    private LinkedList argumentRequests;

    /**
     * The list of new directories
     */
    private LinkedList newDirList;

    /**
     * The client services that are provided to this command.
     */
    private ClientServices clientServices;

    /**
     * The files and/or directories to operate on.
     */
    private File[] files;

    /**
     * Holds value of property message, (add's switch -m).
     */
    private String message;

    /**
     * Holds value of property keywordSubst.
     */
    private String keywordSubst;

    /**
     * Constructor.
     */
    public AddCommand()
    {
        argumentRequests = new LinkedList();
    }

    /**
     * Set the files and/or directories on which to execute the command.
     * Sorts the paameter so that directories are first and files follow.
     * That way a directory and it's content will be passed correctly.
     * The user of the library has to specify all the files+dirs being added though.
     * This is just a sanity check, so that no unnessesary errors occur.
     */
    public void setFiles(File[] files)
    {
        this.files = files;
        if (files == null)
        {
            return;
        }

        // sort array: directories first, files follow
        this.files = new File[files.length];
        int dirCount  = 0;
        int fileCount = 0;
        int totalCount = files.length;
        for (int index = 0; index < totalCount; index++)
        {
            File currentFile = files[index];
            if (currentFile.isDirectory())
            {
                this.files[dirCount] = currentFile;
                dirCount++;
            }
            else
            {
                this.files[totalCount - (1 + fileCount)] = currentFile;
                fileCount++;
            }
        }
    }

    /**
     * Get the files and/or directories specified for this command to operate
     * on.
     * @return the array of Files
     */
    public File[] getFiles()
    {
        return files;
    }

    /**
     * @param ending - the ending part of the file's pathname.. path separator is cvs's default '/'
     */
    public File getFileEndingWith(String ending)
    {
        int index = 0;
        for (index = 0; index < files.length; index++)
        {
            String path = files[index].getAbsolutePath();
            path = path.replace('\\', '/');
            if (path.endsWith(ending))
            {
                return files[index];
            }
        }
        return null;
    }

    /**
     * Getter for property message.
     * @return Value of property message.
     */
    public String getMessage()
    {
        return message;
    }

    /**
     * Setter for property message.
     * @param message New value of property message.
     */
    public void setMessage(String message)
    {
        this.message = message;
    }

    /**
     * Getter for property keywordSubst.
     * @return Value of property keywordSubst.
     */
    public String getKeywordSubst()
    {
        return keywordSubst;
    }

    /**
     * Setter for property keywordSubst.
     * @param keywordSubst New value of property keywordSubst.
     */
    public void setKeywordSubst(String keywordSubst)
    {
        if (keywordSubst == null
         || keywordSubst.equals("kv")
         || keywordSubst.equals("k")
         || keywordSubst.equals("o")
         || keywordSubst.equals("v")
         || keywordSubst.equals("b"))
        {
            this.keywordSubst = keywordSubst;
        }
        else
        {
            throw new IllegalArgumentException(
                      "Keyword substitution " + keywordSubst + " not defined!");
        }
    }

    /**
     * Add requests for a particular file or directory to be added.
     */
    protected void addRequests(File f)
        throws IOException, CommandException
    {
        if (f.isDirectory())
        {
            addRequestsForDirectory(f, true);
        }
        else
        {
            addRequestsForFile(f);
        }
    }

    /**
     * Add requests for a particular directory.
     * @param directory the directory to add
     * @param adding - for the directory to be added, set to true.
     *                 used internally to recurse Directory requests.
     * @throws IOException if an error occurs
     */
    protected void addRequestsForDirectory(File directory, boolean adding)
        throws IOException
    {

        File parentDirectory = directory.getParentFile();
        String dir;
        if (adding)
        {
            dir = getRelativeToLocalPathInUnixStyle(parentDirectory);
        }
        else
        {
            dir = getRelativeToLocalPathInUnixStyle(directory);
        }
        String partPath = null;
        if (!dir.equals("."))
        {
            // trim the leading slash from the pathname we end up with
            // (e.g. we end up with something like \banana\foo
            // and this gives us banana\foo). Also replace backslashes with
            // forward slashes. The standard CVS server doesn't like
            // backslashes very much.
            partPath = dir + "/" + directory.getName();
            // recursively scroll back to the localPath..
            addRequestsForDirectory(parentDirectory, false);
        }
        else
        {
            partPath = directory.getName();
        }
        if (!adding) {
            partPath = dir;
        }
//        System.err.println("Dir is: " + dir);
//        System.err.println("Partpath = " + partPath);

        // Note that the repository file for the directory being added has not
        // been created yet, so we are forced to read the repository for
        // the parent directory and build the appropriate entry by tagging
        // on the directory name (called partPath here)
        String repository;
        String tag = null;
        if (adding) 
        {
            repository = clientServices.getRepositoryForDirectory(
                                    parentDirectory.getAbsolutePath()) +
                                    "/" + directory.getName();
            tag = clientServices.getStickyTagForDirectory(parentDirectory);
        } else 
        {
            repository = clientServices.getRepositoryForDirectory(
                                    directory.getAbsolutePath());
            tag = clientServices.getStickyTagForDirectory(directory);
        }
        requests.add(new DirectoryRequest(partPath, repository));
        if (tag != null)
        {
            requests.add(new StickyRequest(tag));
        }

        if (adding)
        {
            argumentRequests.add(new ArgumentRequest(partPath));
            newDirList.add(partPath);
        }
        // MK argument after Dir request.. also with the rel path from the current working dir
    }

    /**
     * Add requests for a particular file.
     */
    protected void addRequestsForFile(File file)
        throws IOException
    {
        File directory = file.getParentFile();
        String dir = getRelativeToLocalPathInUnixStyle(directory);

        String repository = clientServices.getRepositoryForDirectory(
                                                directory.getAbsolutePath());
        requests.add(new DirectoryRequest(dir, repository));
        String tag = clientServices.getStickyTagForDirectory(directory);
        if (tag != null)
        {
            requests.add(new StickyRequest(tag));
        }

        Entry entry = clientServices.getEntry(file);

        if (entry != null)
        {
            requests.add(new EntryRequest(entry));
        }
        else
        {
            String options = getKeywordSubst();
            boolean isBinary = (options != null && "b".equals(options));
            requests.add(new ModifiedRequest(file, isBinary));
        }

        if (dir.equals("."))
        {
            argumentRequests.add(new ArgumentRequest(file.getName()));
        }
        else
        {
            argumentRequests.add(new ArgumentRequest(dir + "/" + file.getName()));
        }
    }

    /**
     * Execute a command.
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests
     */
    public void execute(ClientServices client, EventManager em)
        throws CommandException
    {
        if (files==null || files.length==0)
        {
            throw new CommandException("No files have been specified for " +
                                       "adding.");
        }
        clientServices = client;
        setLocalPath(client.getLocalPath());
        newDirList = new LinkedList();
        super.execute(client, em);

        requests = new LinkedList();

        if (client.isFirstCommand())
        {
            requests.add(new RootRequest(client.getRepository()));
        }

        // sets the message argument -m .. one for all files being sent..
        String message = getMessage();
        if (message != null)
        {
            message = message.trim();
        }
        if (message != null
         && message.length() > 0)
        {
            addMessageRequest(message);
        }

        if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
        {
            requests.add(new ArgumentRequest("-k" + getKeywordSubst()));
        }

        try
        {
            // current dir sent to server BEFORE and AFTER - kinda hack??
            for (int i=0; i < files.length; i++)
            {
                addRequests(files[i]);
            }

            // now add the request that indicates the working directory for the
            // command
            requests.add(new DirectoryRequest(".",
                                  client.getRepositoryForDirectory(getLocalPath())));

            requests.addAll(argumentRequests);
            argumentRequests.clear(); // MK sanity check.
            requests.add(new AddRequest());
            client.processRequests(requests);
            requests.clear();
        }
        catch (CommandException ex)
        {
            requests.clear();
            throw ex;
        }
        catch (Exception ex)
        {
            requests.clear();
            throw new CommandException(ex);
        }
    }

    private void addMessageRequest(String message)
    {
        requests.add(new ArgumentRequest("-m"));
        StringTokenizer token = new StringTokenizer(message, "\n", false);
        boolean first = true;
        while (token.hasMoreTokens())
        {
            if (first)
            {
                requests.add(new ArgumentRequest(token.nextToken()));
                first = false;
            }
            else
            {
                requests.add(new ArgumentxRequest(token.nextToken()));
            }
        }
    }

    /**
     * This method returns how the command would look like when typed on the
     * command line.
     * Each command is responsible for constructing this information.
     * @returns <command's name> [<parameters>] files/dirs. Example: checkout -p CvsCommand.java
     */
    public String getCVSCommand()
    {
        StringBuffer toReturn = new StringBuffer("add ");
        toReturn.append(getCVSArguments());
        File[] files = getFiles();
        if (files != null)
        {
            for (int index = 0; index < files.length; index++)
            {
                toReturn.append(files[index].getName() + " ");
            }
        }
        return toReturn.toString();
    }

    /**
     * Method that is called while the command is being executed.
     * Descendants can override this method to return a Builder instance
     * that will parse the server's output and create data structures.
     */
    protected Builder createBuilder(EventManager eventManager)
    {
        return new AddBuilder(eventManager, this);
    }

    /**
     * Takes the arguments and sets the command.
     * To be mainly used for automatic settings (like parsing the .cvsrc file)
     * @return true if the option (switch) was recognized and set
     */
    public boolean setCVSCommand(char opt, String optArg)
    {
        if (opt == 'm')
        {
            setMessage(optArg);
        }
        else if (opt == 'k')
        {
            setKeywordSubst(optArg);
        }
        else
        {
            return false;
        }
        return true;
    }

    /**
     * Returns a string indicating the available options.
     */
    public String getOptString()
    {
        return "m:k:";
    }

    /**
     * Listens for output of the command.
     * If new directory is added, executes the createCvsFiles() method.
     */
    public void messageSent(MessageEvent e)
    {
        String str = e.getMessage();
        if (str.endsWith(DIR_ADDED))
        {
            str = str.substring(DIRECTORY.length(), str.indexOf(DIR_ADDED)).trim();
//            System.out.println("directory=" + str);
            createCvsFiles(str);
        }
        super.messageSent(e);
    }

    /**
     * For new directory that was added to the repository, creates the admin files in CVS subdir.
     */
    private void createCvsFiles(String newDirInRepository)
    {
        String repository = newDirInRepository;
//        System.out.println("repository=" + repository);
        String dirName = repository;
        if (dirName.lastIndexOf('/') >= 0)
        {
            dirName = dirName.substring(dirName.lastIndexOf('/') + 1,
                                        dirName.length());
        }
//        System.out.println("dirname=" + dirName);
        Iterator it = newDirList.iterator();
        while (it.hasNext())
        {
            String part = (String) it.next();
            String tempDirName = part;
            if (part.lastIndexOf('/') >= 0)
            {
               tempDirName = part.substring(part.lastIndexOf('/') + 1,
                                            part.length());
            }
//            System.out.println("tempDirName=" + tempDirName);
            if (tempDirName.equalsIgnoreCase(dirName))
            {
//                System.out.println("part=" + part);
                try
                {
                    clientServices.updateAdminData(part, repository, null);
                    it.remove(); // hack.. in case 2 dirs being added have the same name??
                    break;
                }
                catch (IOException exc)
                {
                    System.out.println("TODO: couldn't create/update Cvs admin files");
                }
            }
        }
    }

    /**
     * resets all switches in the command. After calling this method,
     * the command should have no switches defined and should behave defaultly.
     */
    public void resetCVSCommand()
    {
        setMessage(null);
        setKeywordSubst(null);
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
     */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer("");
        if (getMessage() != null)
        {
            toReturn.append("-m \"");
            toReturn.append(getMessage());
            toReturn.append("\" ");
        }
        if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
        {
            toReturn.append("-k");
            toReturn.append(getKeywordSubst());
            toReturn.append(" ");
        }
        return toReturn.toString();
    }

}
