/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.
 *
 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.checkout;

import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.io.File;
import java.io.IOException;

import java.util.ArrayList;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.event.ModuleExpansionEvent;

import org.netbeans.lib.cvsclient.request.ArgumentRequest;
import org.netbeans.lib.cvsclient.request.CheckoutRequest;
import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.ExpandModulesRequest;
import org.netbeans.lib.cvsclient.request.RootRequest;

import org.netbeans.lib.cvsclient.ClientServices;
import org.netbeans.lib.cvsclient.command.*;
import org.netbeans.lib.cvsclient.command.update.UpdateBuilder;

/**
 * The checkout command. This handles the sending of the requests and the
 * processing of the responses from the server
 * @author  Robert Greig
 * @version $Revision: 1.16.2.6 $
 */
public class CheckoutCommand extends BasicCommand
                             implements TemporaryFileCreator
{
    /**
     * A store of potentially empty directories. When a directory has a file
     * in it, it is removed from this set. This set allows the prune option
     * to be implemented.
     */
    private HashSet emptyDirectories;

    /**
     * Whether to checkout recursively
     */
    private boolean recursive;

    /**
     * The modules to checkout. These names are unexpanded and will be passed
     * to a module-expansion request.
     */
    private String[] modules;

    /**
     * The expanded modules.
     */
    private LinkedList expandedModules;    
    
    /**
     * Will force the checkout command to display only a list of modules.
     */
    private boolean showModules;

    /**
     * if set, will display just a list of modules with statuses.
     */
    private boolean showModulesWithStatus;

    /**
     * if set, will redirect the output of the command to standard output.
     */
    private boolean pipeToOutput;    

    /**
     * Whether to prune directories, i.e. remove any directories that do not
     * contain any files. This is the -P option in command-line CVS)
     */
    private boolean pruneDirectories;

    /**
     * Resets any sticky tags/dates/options imposed on the updated file(s).
     */
    private boolean resetStickyOnes;

    /**
     * Use head revision if a revision meeting criteria set by switches -r/-D
     * (tag/date) is not found.
     */
    private boolean useHeadIfNotFound;

    /**
     * Forces a checkout of a revision that was current at specified date.
     */
    private String checkoutByDate;

    /**
     * Forces a checkout of specified revision. Can be a number/tag/branch
     */
    private String checkoutByRevision;

    /**
     * Use this keyword substitution for the command.
     * does not include the -k switch part.
     */
    private String keywordSubst;

    /**
     * Construct a new checkout command.
     * @param isRecursive whether to do a recursive checkout
     * @param theModules an array of modules names to checkout
     */
    public CheckoutCommand(boolean isRecursive, String[] theModules)
    {
        recursive = isRecursive;
        modules = theModules;
        emptyDirectories = new HashSet();
    }

    /**
     * Construct a new checkout command
     * @param isRecursive whether to do a recursive checkout
     * @param theModule the module to checkout
     */
    public CheckoutCommand(boolean isRecursive, String theModule)
    {
        this(isRecursive, new String[]{theModule});
    }

    /**
     * Construct a checkout command, with default values for options
     */    
    public CheckoutCommand()
    {
        this(true, (String[])null);
    }

    /**
     * Gets the value of the recursive option
     * @return true if recursive, false if not
     */
    public boolean getRecursive()
    {
        return recursive;
    }

    /**
     * Sets the value of the recursive option
     * @param r true if the command should recurse, false otherwise
     */
    public void setRecursive(boolean r)
    {
        recursive = r;
    }

    /**
     * Set the modules to checkout
     * @param theModules the names of the modules to checkout
     */
    public void setModules(String[] theModules)
    {
        modules = theModules;
    }
    
    
    /**
     * Handle modules that are already checked out. We check whether a 
     * module has been checked out and if so we add it to the list of
     * directories that the superclass must send Modified requests for etc.
     */
    private void processExistingModules(String localPath)
    {
        if (expandedModules != null && expandedModules.size() > 0)
        {            
            ArrayList list = new ArrayList(expandedModules.size());
            final Iterator it = expandedModules.iterator();
            while (it.hasNext())
            {
                String moduleName = (String)it.next();
                File moduleDir = null;
                final File moduleFile = new File(localPath, moduleName);
                if (moduleFile.isFile()) 
                {
                    moduleDir = moduleFile.getParentFile();
                }
                else 
                {
                    moduleDir = moduleFile;
                }
                final File moduleCVSDir = new File(moduleDir, "CVS/Repository");
                if (moduleCVSDir.exists()) 
                {
                    list.add(moduleFile);
                }
            }
            File[] directories = new File[list.size()];
            directories = (File[])list.toArray(directories);
            setFiles(directories);
        }
    }
    
    /**
     * Execute this command
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests
     */
    public void execute(ClientServices client, EventManager em)
                                                        throws CommandException
    {   
        requests = new LinkedList();
        if (client.isFirstCommand())
        {            
            requests.add(new RootRequest(client.getRepository()));            
        }
        
        if (showModules || showModulesWithStatus) {
            // we need to initialize the builder first (is done in BuildableCommand.execute()
            // but we can't run it because of teh BasicCommand's execute and 
            // it's feature that adds the files request to the request list.
            if (builder == null && !builderSet)
            {
                builder = createBuilder(em);
            }
            
            // special handling for -c -s switches.
            if (showModules) {
                requests.add(new ArgumentRequest("-c"));
            }
            if (showModulesWithStatus) {
                requests.add(new ArgumentRequest("-s"));
            }
            requests.add(new CheckoutRequest());
            try {
                client.processRequests(requests);
                requests.clear();
            }
            catch (Exception ex) {
                throw new CommandException(ex);
            }
            return;
        }
        
        if (modules != null)
        {
            // for -c, -s switches no module has to be selected
            for (int i=0; i < modules.length; i++)
            {
                requests.add(new ArgumentRequest(modules[i]));
            }
        }
        
        expandedModules = new LinkedList();
        requests.add(new ExpandModulesRequest()); 
        try
        {
            client.processRequests(requests);
        }
        catch (Exception e)
        {
            throw new CommandException(e);
        }
        client.setIsFirstCommand(false);
        requests.clear();
        postExpansionExecute(client, em);
    }
    
    /**
     * This is called when the server has responded to an expand-modules 
     * request.
     */
    public void moduleExpanded(ModuleExpansionEvent e)
    {
        expandedModules.add(e.getModule());               
    }
    
    /**
     * Execute this command
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests
     */
    private void postExpansionExecute(ClientServices client, EventManager em)
        throws CommandException
    {                                

        // we first test whether the modules specified actually exist
        // checked out already. If so, we must work something like an update
        // command and send modified files to the server.               
        processExistingModules(client.getLocalPath());        
        
        // the sending of the Modified requests and so on is handled in the
        // superclass.
        super.execute(client, em);
        
        //
        // moved modules code to the end of the other arguments --GAR
        // 
        final int FIRST_INDEX = 0;
        final int SECOND_INDEX = 1;
        if (!recursive)
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-l"));
        }
        if (pipeToOutput)
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-p"));
        }
        if (resetStickyOnes)
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-A"));
        }
        if (useHeadIfNotFound)
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-f"));
        }
        if (checkoutByDate != null && (!checkoutByDate.equals("")))
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-D"));
            requests.add(SECOND_INDEX, new ArgumentRequest(getCheckoutByDate()));
        }
        if (checkoutByRevision != null && (!checkoutByRevision.equals("")))
        {
            requests.add(FIRST_INDEX, new ArgumentRequest("-r"));
            requests.add(SECOND_INDEX, new ArgumentRequest(getCheckoutByRevision()));
        }
        if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
        {
            requests.add(new ArgumentRequest("-k" + getKeywordSubst()));
        }

        // Note that modules might be null and still be valid because
        // for -c, -s switches no module has to be selected
        // You might also think that we should pass in expandedModules here
        // but according to the spec that would be wrong because of the -d
        // flag.
        if (modules != null)
        {            
            for (int i=0; i < modules.length; i++)
            {
                requests.add(new ArgumentRequest(modules[i]));
            }
        }

        requests.add(new DirectoryRequest(".", client.getRepository()));
        requests.add(new CheckoutRequest());
        try
        {
            client.processRequests(requests);
            if (pruneDirectories)
            {
                pruneEmptyDirectories();
            }
            requests.clear();
        } catch (Exception e)
        {
            requests.clear();
            throw new CommandException(e);
        }
    }

    /** Getter for property showModules.
     * @return Value of property showModules.
     */
    public boolean isShowModules() {
        return showModules;
    }

    /** Setter for property showModules.
     * @param showModules New value of property showModules.
     */
    public void setShowModules(boolean showModules) {
        this.showModules = showModules;
    }

    /** Getter for property showModulesWithStatus.
     * @return Value of property showModulesWithStatus.
     */
    public boolean isShowModulesWithStatus() {
        return showModulesWithStatus;
    }

    /** Setter for property showModulesWithStatus.
     * @param showModulesWithStatus New value of property showModulesWithStatus.
     */
    public void setShowModulesWithStatus(boolean showModulesWithStatus) {
        this.showModulesWithStatus = showModulesWithStatus;
    }

    /**
     * Set whether to prune directories. This is the -P option in the command-
     * line CVS.
     */
    public void setPruneDirectories(boolean pruneDirectories)
    {
        this.pruneDirectories = pruneDirectories;
    }

    /**
     * Get whether to prune directories.
     * @return true if directories should be removed if they contain no files,
     * false otherwise.
     */
    public boolean getPruneDirectories()
    {
        return pruneDirectories;
    }

    /** Getter for property pipeToOutput.
     * @return Value of property pipeToOutput.
     */
    public boolean isPipeToOutput() {
        return pipeToOutput;
    }

    /** Setter for property pipeToOutput.
     * @param pipeToOutput New value of property pipeToOutput.
     */
    public void setPipeToOutput(boolean pipeToOutput) {
        this.pipeToOutput = pipeToOutput;
    }

    /** Getter for property resetStickyOnes.
     * @return Value of property resetStickyOnes.
     */
    public boolean isResetStickyOnes()
    {
        return resetStickyOnes;
    }

    /** Setter for property resetStickyOnes.
     * @param resetStickyOnes New value of property resetStickyOnes.
     */
    public void setResetStickyOnes(boolean resetStickyOnes)
    {
        this.resetStickyOnes = resetStickyOnes;
    }

    /** Getter for property useHeadIfNotFound.
     * @return Value of property useHeadIfNotFound.
     */
    public boolean isUseHeadIfNotFound()
    {
        return useHeadIfNotFound;
    }

    /** Setter for property useHeadIfNotFound.
     * @param useHeadIfNotFound New value of property useHeadIfNotFound.
     */
    public void setUseHeadIfNotFound(boolean useHeadIfNotFound)
    {
        this.useHeadIfNotFound = useHeadIfNotFound;
    }

    /** Getter for property checkoutByDate.
     * @return Value of property checkoutByDate.
     */
    public String getCheckoutByDate()
    {
        return checkoutByDate;
    }

    /** Setter for property checkoutByDate.
     * @param checkoutByDate New value of property checkoutByDate.
     */
    public void setCheckoutByDate(String checkoutByDate)
    {
        this.checkoutByDate = checkoutByDate;
    }

    /** Getter for property checkoutByRevision.
     * @return Value of property checkoutByRevision.
     */
    public String getCheckoutByRevision()
    {
        return checkoutByRevision;
    }

    /** Setter for property checkoutByRevision.
     * @param checkoutByRevision New value of property checkoutByRevision.
     */
    public void setCheckoutByRevision(String checkoutByRevision)
    {
        this.checkoutByRevision = checkoutByRevision;
    }
    /** Getter for property keywordSubst.
     * @return Value of property keywordSubst.
     */
    public String getKeywordSubst()
    {
        return keywordSubst;
    }

    /** Setter for property keywordSubst.
     * @param keywordSubst New value of property keywordSubst.
     */
    public void setKeywordSubst(String keywordSubst)
    {
        this.keywordSubst = keywordSubst;
    }

    protected Builder createBuilder(EventManager eventMan) {
        if (isShowModules() || isShowModulesWithStatus()) {
            return new ModuleListBuilder(eventMan, this);
        }
        if (isPipeToOutput()) {
            return new PipedFilesBuilder(eventMan, this, this);
        }
        return new UpdateBuilder(eventMan, getLocalPath());
    }

    public File createTempFile() throws IOException  {
        File temp = File.createTempFile("cvs", ".dff");
        return temp;
    }

    /** This method returns how the command would looklike when typed on the command line.
     * Each command is responsible for constructing this information.
     * @returns <command's name> [<parameters>] files/dirs. Example: checkout -p CvsCommand.java
     */
    public String getCVSCommand()
    {
        StringBuffer toReturn = new StringBuffer("checkout ");
        toReturn.append(getCVSArguments());
        if (modules != null && (!isShowModules() && !isShowModulesWithStatus())) {
            for (int index = 0; index < modules.length; index++) {
                toReturn.append(modules[index]);
                toReturn.append(' ');
            }
        }
        return toReturn.toString();
    }

    /** takes the arguments and sets the command. To be mainly
     * used for automatic settings (like parsing the .cvsrc file)
     * @return true if the option (switch) was recognized and set
     */
    public boolean setCVSCommand(char opt,String optArg) {
        if (opt == 'c') {
            setShowModules(true);
        }
        else if (opt == 's') {
            setShowModulesWithStatus(true);
        }
        else if (opt == 'p') {
            setPipeToOutput(true);
        }
        else if (opt == 'R') {
            setRecursive(true);
        }
        else if (opt == 'l') {
            setRecursive(false);
        }
        else if (opt == 'A') {
            setResetStickyOnes(true);
        }
        else if (opt == 'f') {
            setUseHeadIfNotFound(true);
        }
        else if (opt == 'P') {
            setPruneDirectories(true);
        }
        else if (opt == 'D') {
            setCheckoutByDate(optArg.trim());
        }
        else if (opt == 'r') {
            setCheckoutByRevision(optArg.trim());
        }
        else if (opt == 'k') {
            setKeywordSubst(optArg);
        }
        else {
            return false;
        }
        return true;
    }

    /**
     * String returned by this method defines which options are available for this particular command
     */
    public String getOptString()
    {
        return "cpslPRAD:r:fk:";
    }

    /**
     * resets all switches in the command. After calling this method,
     * the command should have no switches defined and should behave defaultly.
     */
    public void resetCVSCommand() {
        setShowModules(false);
        setShowModulesWithStatus(false);
        setPipeToOutput(false);
        setRecursive(true);
        setResetStickyOnes(false);
        setUseHeadIfNotFound(false);
        setCheckoutByDate(null);
        setCheckoutByRevision(null);
        setKeywordSubst(null);
        setPruneDirectories(false);
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
     */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer("");
        if (isShowModules()) {
            toReturn.append("-c ");
        }
        if (isShowModulesWithStatus()) {
            toReturn.append("-s ");
        }
        if (isPipeToOutput()) {
            toReturn.append("-p ");
        }
        if (!getRecursive()) {
            toReturn.append("-l ");
        }
        if (isResetStickyOnes()) {
            toReturn.append("-A ");
        }
        if (isUseHeadIfNotFound()) {
            toReturn.append("-f ");
        }
        if (getPruneDirectories())
        {
            toReturn.append("-P ");
        }

        if (getKeywordSubst() != null && (!getKeywordSubst().equals(""))) {
            toReturn.append("-k");
            toReturn.append(getKeywordSubst());
            toReturn.append(' ');
        }
        if (getCheckoutByRevision() != null && (!getCheckoutByRevision().equals("")))
        {
            toReturn.append("-r ");
            toReturn.append(getCheckoutByRevision());
            toReturn.append(' ');
        }
        if (getCheckoutByDate() != null && (!getCheckoutByDate().equals("")))
        {
            toReturn.append("-D ");
            toReturn.append(getCheckoutByDate());
            toReturn.append(' ');
        }
        return toReturn.toString();
    }

    /**
     * Called when the server wants to send a message to be displayed to
     * the user. The message is only for information purposes and clients
     * can choose to ignore these messages if they wish.
     * @param e the event
     */
    public void messageSent(MessageEvent e)
    {
        super.messageSent(e);
        // we use this event to determine which directories need to be checked
        // for updating
        if (pruneDirectories &&
            e.getMessage().startsWith("cvs server: Updating "))
        {
            File file = new File(getLocalPath(), e.getMessage().substring(21));
            emptyDirectories.add(file);
        }
    }

    /**
     * Prunes a directory, recursively pruning its subdirectories
     * @param directory the directory to prune
     */
    private boolean pruneEmptyDirectory(File directory) throws IOException
    {
        boolean empty = true;

        final File[] contents = directory.listFiles();

        // should never be null, but just in case...
        if (contents!=null)
        {
            for (int i=0; i < contents.length; i++)
            {
                if (contents[i].isFile())
                {
                    empty = false;
                }
                else
                {
                    if (!contents[i].getName().equals("CVS"))
                    {
                        empty = pruneEmptyDirectory(contents[i]);
                    }
                }
                if (!empty)
                {
                    break;
                }
            }

            if (empty)
            {
                // check this is a CVS directory and not some directory the user
                // has stupidly called CVS...
                final File entriesFile = new File(directory, "CVS/Entries");
                if (entriesFile.exists())
                {
                    final File adminDir = new File(directory, "CVS");
                    final File[] adminFiles = adminDir.listFiles();
                    for (int i=0; i < adminFiles.length; i++)
                    {
                        adminFiles[i].delete();
                    }
                    adminDir.delete();
                    directory.delete();
                }
            }
        }

        return empty;
    }

    /**
     * Remove any directories that don't contain any files
     */
    private void pruneEmptyDirectories() throws IOException
    {
        final Iterator it = emptyDirectories.iterator();
        while (it.hasNext())
        {
            final File dir = (File) it.next();
            // we might have deleted it already (due to recursive delete)
            // so we need to check existence
            if (dir.exists())
            {
                pruneEmptyDirectory(dir);
            }
        }
        emptyDirectories.clear();
    }
}
