/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.commit;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.admin.Entry;

import org.netbeans.lib.cvsclient.event.EventManager;

import org.netbeans.lib.cvsclient.request.ArgumentRequest;
import org.netbeans.lib.cvsclient.request.ArgumentxRequest;
import org.netbeans.lib.cvsclient.request.CommitRequest;
import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.StickyRequest;

import org.netbeans.lib.cvsclient.request.EntryRequest;
import org.netbeans.lib.cvsclient.request.ModifiedRequest;
import org.netbeans.lib.cvsclient.request.RootRequest;
import org.netbeans.lib.cvsclient.request.UpdateRequest;

import org.netbeans.lib.cvsclient.command.BasicCommand;
import org.netbeans.lib.cvsclient.command.CommandException;
import org.netbeans.lib.cvsclient.command.Builder;

/**
 * The command to commit any changes that have been made.
 * @author  Robert Greig
 * @version $Revision: 1.9.2.3 $
 */
public class CommitCommand extends BasicCommand
{
    /**
     * The argument requests that must be added at the end.
     * These argument requests indicate the files to be committed
     */
    protected LinkedList argumentRequests;

    /**
     * The log message used for the commit.
     */
    private String message;

    /**
     * Forces the commit of the file(s) even if no changes were done.
     * the standard behaviour is NOT-TO-BE recursive in this case.
     */
    private boolean forceCommit;

    /**
     * The filename for the file that defines the message.
     */
    private String logMessageFromFile;

    /**
     * Determines that no module program should run on the server.
     */
    private boolean noModuleProgram;

    /** Holds value of property toRevisionOrBranch. */
    private String toRevisionOrBranch;

    /**
     * Construct a CommitCommand.
     */
    public CommitCommand()
    {
        argumentRequests = new LinkedList();
    }

    /**
     * Returns the commit message.
     */
    public String getMessage()
    {
        return message;
    }

    /**
     * Sets the commit message.
     */
    public void setMessage(String message)
    {
        this.message = message;
    }

    /**
     * Indicates whether the commit should be forced even if there are no
     * changes.
     */
    public boolean isForceCommit() {
        return forceCommit;
    }

    /**
     * Sets whether the commit should be forced even if there are no changes.
     */
    public void setForceCommit(boolean forceCommit) {
        this.forceCommit = forceCommit;
    }

    /**
     * Adds the appropriate requests for a given directory.
     * Sends a directory request followed by as many Entry and Modified requests
     * as required.
     * @param directory the directory to send requests for
     * @throws IOException if an error occurs constructing the requests
     */
    protected void addRequestsForDirectory(File directory)
        throws IOException
    {
        // remove localPath prefix from directory. If left with
        // nothing, use dot (".") in the directory request. Also remove the
        // trailing slash
        String dir = getRelativeToLocalPathInUnixStyle(directory);

        try
        {
            String repository = clientServices.getRepositoryForDirectory(
                                                   directory.getAbsolutePath());
            requests.add(new DirectoryRequest(dir, repository));
            String tag = clientServices.getStickyTagForDirectory(directory);
            if (tag != null)
            {
                requests.add(new StickyRequest(tag));
            }
        }
        catch (IOException e)
        {
            System.err.println("An error occurred reading the respository " +
                               "for the directory " + dir + ": " + e);
            e.printStackTrace();
        }

        // Obtain a set of all files known to CVS. We union
        // this set with the set of files in the actual filesystem directory
        // to obtain a set of files to commit (or at least attempt to commit).
        Set set = clientServices.getAllFiles(directory);

        // We must add the local files (and directories) because the above
        // command does *not* return cvs controlled directories
        final File[] files = directory.listFiles();

        // get the union of the files in the directory and the files retrieved
        // from the Entries file.
        set.addAll(Arrays.asList(files));

        List subdirectories = null;
        if (recursive)
        {
            subdirectories = new LinkedList();
        }

        for (Iterator it = set.iterator(); it.hasNext(); )
        {
            File file = (File)it.next();
            if (file.isHidden()
             || file.getName().equals("CVS"))
            {
                continue;
            }

            try
            {
                final Entry entry = clientServices.getEntry(file);
                // a non-null entry means the file does exist in the
                // Entries file for this directory
                if (entry == null)
                {
                    continue;
                }

                // here file.isFile() is *not* used, because not existing
                // files (removed ones) should also be sent
                if (!file.isDirectory())
                {
                    sendEntryAndModifiedRequests(entry, file);
                }
                else if (recursive && file.isDirectory())
                {
                    subdirectories.add(file);
                }
            }
            catch (IOException e)
            {
                System.err.println("An error occurred getting the " +
                        "Entry for file " + file + ": " + e);
                e.printStackTrace();
            }
        }

        if (recursive)
        {
            for (Iterator it = subdirectories.iterator(); it.hasNext(); )
            {
                File subdirectory = (File)it.next();
                addRequestsForDirectory(subdirectory);
            }
        }
    }

    /**
     * Add the appropriate requests for a single file.
     * A directory request is sent, followed by an Entry and Modified request.
     * @param file the file to send requests for
     * @throws IOException if an error occurs constructing the requests
     */
    protected void addRequestsForFile(File file)
        throws IOException
    {
        final File parentDirectory = file.getParentFile();
        // remove localPath prefix from directory. If left with
        // nothing, use dot (".") in the directory request
        String dir = getRelativeToLocalPathInUnixStyle(parentDirectory);

        try
        {
            // send a argument request indicating the file to update
            //argumentRequests.add(new ArgumentRequest(getRelativeToLocalPathInUnixStyle(file)));
            requests.add(new DirectoryRequest(dir, clientServices.
                                    getRepositoryForDirectory(parentDirectory.
                                    getAbsolutePath())));
            String tag = clientServices.getStickyTagForDirectory(parentDirectory);
            if (tag != null)
            {
                requests.add(new StickyRequest(tag));
            }
        }
        catch (IOException e)
        {
            System.err.println("An error occurred reading the respository " +
                               "for the directory " + dir + ": " + e);
            e.printStackTrace();
        }

        try
        {
            final Entry entry = clientServices.getEntry(file);
            // a non-null entry means the file does exist in the
            // Entries file for this directory
            if (entry != null)
            {
                sendEntryAndModifiedRequests(entry, file);
            }
        }
        catch (IOException e)
        {
            System.err.println("An error occurred getting the Entry "+
                    "for file " + file + ": " + e);
            e.printStackTrace();
        }
    }

    /**
     * Should return true if unchanged files should not be sent to server.
     * If false is returned, all files will be sent to server
     * This method is used by <code>sendEntryAndModifiedRequests</code>.
     */
    protected boolean doesCheckFileTime()
    {
        if (isForceCommit()) return false;
        return true;
    }
    
    /**
     * Execute the command.
     * @param client the client services object that provides any necessary
     *               services to this command, including the ability to actually
     *               process all the requests
     */
    public void execute(ClientServices client, EventManager em)
        throws CommandException
    {
        super.execute(client, em);

        try
        {
            // add arguments.
            if (isForceCommit())
            {
                requests.add(1, new ArgumentRequest("-f"));
            }
            if (isNoModuleProgram())
            {
                requests.add(1, new ArgumentRequest("-n"));
            }
            if (getToRevisionOrBranch() != null)
            {
                requests.add(1, new ArgumentRequest(getToRevisionOrBranch()));
                requests.add(1, new ArgumentRequest("-r"));
            }

            // build the message to send
            String message = getMessage();
            if (getLogMessageFromFile() != null)
            {
                message = loadLogFile(getLogMessageFromFile());
            }
            if (message != null)
            {
                message = message.trim();
            }
            if (message == null
             || message.length() == 0)
            {
                message = "no message";
            }
            addMessageRequest(message);

            // now add the request that indicates the working directory for the
            // command
            requests.add(new DirectoryRequest(".",
                                  client.getRepositoryForDirectory(getLocalPath())));
            requests.addAll(argumentRequests);
            argumentRequests.clear(); // MK sanity check.
            addArgumentRequests();
            requests.add(new CommitRequest());

            client.processRequests(requests);
            requests.clear();
        }
        catch (Exception ex)
        {
            requests.clear();
            throw new CommandException(ex);
        }
    }

    /**
     * This method returns how the command would looklike when typed on the command line.
     * Example: checkout -p CvsCommand.java
     * @returns <command's name> [<parameters>] files/dirs
     */
    public String getCVSCommand()
    {
        StringBuffer toReturn = new StringBuffer("commit ");
        toReturn.append(getCVSArguments());
        File[] files = getFiles();
        if (files != null)
        {
            for (int index = 0; index < files.length; index++)
            {
                toReturn.append(files[index].getName() + " ");
            }
        }
        return toReturn.toString();
    }

    /**
     * Takes the arguments and sets the command.
     * To be mainly used for automatic settings (like parsing the .cvsrc file).
     * @return true if the option (switch) was recognized and set
     */
    public boolean setCVSCommand(char opt, String optArg)
    {
        if (opt == 'm')
        {
            setMessage(optArg);
        }
        else if (opt == 'l')
        {
            setRecursive(false);
        }
        else if (opt == 'R')
        {
            setRecursive(true);
        }
        else if (opt == 'f')
        {
            setForceCommit(true);
        }
        else if (opt == 'F')
        {
            setLogMessageFromFile(optArg);
        }
        else if (opt == 'r')
        {
            setToRevisionOrBranch(optArg);
        }
        else if (opt == 'n')
        {
            setNoModuleProgram(true);
        }
        else
        {
            return false;
        }
        return true;
    }

    /**
     * Returns a String defining which options are available for this command.
     */
    public String getOptString()
    {
        return "m:flRnF:r:";
    }

    /**
     * Method that is called while the command is being executed.
     * Descendants can override this method to return a Builder instance
     * that will parse the server's output and create data structures.
     */
    protected Builder createBuilder(EventManager eventMan)
    {
        return new CommitBuilder(eventMan, this);
    }

    /**
     * Generates the Argument/Argumentx series of requests depending
     * on the number of lines in the message request.
     */
    private void addMessageRequest(String message)
    {
        requests.add(new ArgumentRequest("-m"));
        StringTokenizer token = new StringTokenizer(message, "\n", false);
        boolean first = true;
        while (token.hasMoreTokens())
        {
            if (first)
            {
                requests.add(new ArgumentRequest(token.nextToken()));
                first = false;
            }
            else
            {
                requests.add(new ArgumentxRequest(token.nextToken()));
            }
        }
    }

    /**
     * Returns the filename for the file that defines the message.
     */
    public String getLogMessageFromFile()
    {
        return logMessageFromFile;
    }

    /**
     * Sets the filename for the file that defines the message.
     */
    public void setLogMessageFromFile(String logMessageFromFile)
    {
        this.logMessageFromFile = logMessageFromFile;
    }

    /**
     * Returns whether no module program should be executed on the server.
     */
    public boolean isNoModuleProgram()
    {
        return noModuleProgram;
    }

    /**
     * Sets whether no module program should run on the server
     */
    public void setNoModuleProgram(boolean noModuleProgram)
    {
        this.noModuleProgram = noModuleProgram;
    }

    /** Getter for property toRevisionOrBranch.
     * @return Value of property toRevisionOrBranch.
     */
    public String getToRevisionOrBranch() {
        return toRevisionOrBranch;
    }

    /** Setter for property toRevisionOrBranch.
     * @param toRevisionOrBranch New value of property toRevisionOrBranch.
     */
    public void setToRevisionOrBranch(String toRevBranch) {
        this.toRevisionOrBranch = toRevBranch;
    }

    private String loadLogFile(String fileName)
        throws CommandException
    {
        StringBuffer buffer = new StringBuffer();
        try
        {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            String line;
            while ((line = reader.readLine()) != null)
            {
                buffer.append(line + "\n");
            }
        }
        catch (FileNotFoundException ex)
        {
            throw new CommandException(ex);
        }
        catch (IOException ex)
        {
            throw new CommandException(ex);
        }
        return buffer.toString();
    }

    /**
     * Resets all switches in the command.
     * After calling this method, the command should have no switches defined
     * and should behave defaultly.
     */
    public void resetCVSCommand()
    {
        setMessage(null);
        setRecursive(true);
        setForceCommit(false);
        setLogMessageFromFile(null);
        setNoModuleProgram(false);
        setToRevisionOrBranch(null);
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name.
     */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer();
        if (!getRecursive())
        {
            toReturn.append("-l ");
        }
        if (isForceCommit())
        {
            toReturn.append("-f ");
        }
        if (isNoModuleProgram())
        {
            toReturn.append("-n ");
        }
        if (getToRevisionOrBranch() != null)
        {
            toReturn.append("-r ");
            toReturn.append(getToRevisionOrBranch() + " ");
        }
        if (getLogMessageFromFile() != null)
        {
            toReturn.append("-F ");
            toReturn.append(getLogMessageFromFile());
            toReturn.append(" ");
        }
        if (getMessage() != null)
        {
            toReturn.append("-m \"");
            toReturn.append(getMessage());
            toReturn.append("\" ");
        }
        return toReturn.toString();
    }



}
