/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.lib.cvsclient.command.diff;

import java.io.File;

import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.event.FileInfoEvent;
import org.netbeans.lib.cvsclient.event.TerminationEvent;

import org.netbeans.lib.cvsclient.file.FileStatus;

import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.DiffRequest;
import org.netbeans.lib.cvsclient.request.ArgumentRequest;

import org.netbeans.lib.cvsclient.command.BasicCommand;
import org.netbeans.lib.cvsclient.command.Builder;
import org.netbeans.lib.cvsclient.command.CommandException;
/**
 * The status command looks up the status of files in the repository
 * @author  Robert Greig
 * @version $Revision: 1.15.2.2 $
 */
public class DiffCommand extends BasicCommand
{
    /**
     * The event manager to use
     */
    protected EventManager eventManager;
   
    /**
     * Holds value of property beforeDate.
     */
    private String beforeDate1 = null;

    /**
     * Holds value of property firstRevision.
     */
    private String revision1 = null;

    /**
     * Holds value of property secondRevision.
     */
    private String revision2 = null;

    /** 
     * Holds value of property beforeDate2. 
     */
    private String beforeDate2 = null;

    /**
     * Keyword substitution. The -k switch in command line cvs.
     */
    private String keywordSubst;

    /**
     * Construct a new diff command
     */
    public DiffCommand() {
    }

    /**
     * Create a builder for this command.
     * @param eventMan the event manager used to receive events.
     */
    protected Builder createBuilder(EventManager eventMan) 
    {        
        return new SimpleDiffBuilder(eventMan, this);
    }

    /**
     * Execute a command
     * @param client the client services object that provides any necessary
     * services to this command, including the ability to actually process
     * all the requests.
     */
    public void execute(ClientServices client, EventManager em)
        throws CommandException
    {
        eventManager = em;

        super.execute(client, em);

        try {
            // parameters come now..
            addRDSwitches();
            if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
            {
                requests.add(new ArgumentRequest("-k" + getKeywordSubst()));
            }
            requests.add(new DirectoryRequest(".", client.getRepositoryForDirectory(getLocalPath())));
            addArgumentRequests();
            requests.add(new DiffRequest());
            client.processRequests(requests);
            requests.clear();
        }
        catch (Exception e) {
            requests.clear();
            throw new CommandException(e);
        }
    }

    /* includes the logic of setting the -r and -D switches to the diff command
     */

    private void addRDSwitches()
    {
        if (getRevision2() != null) {
            requests.add(1, new ArgumentRequest("-r"));
            requests.add(2, new ArgumentRequest(getRevision2()));
        } else {
            if (getBeforeDate2() != null) {
                requests.add(1, new ArgumentRequest("-D " + getBeforeDate2()));
            }
        }
        // -r switch has precendence over the -d switch - is that right??
        if (getRevision1() != null) {
            requests.add(1, new ArgumentRequest("-r"));
            requests.add(2, new ArgumentRequest(getRevision1()));
        } else {
            if (getBeforeDate1() != null) {
                requests.add(1, new ArgumentRequest("-D " + getBeforeDate1()));
            } else {
                // when neither revision nor flag is set for the command, it is assumed
                // that the second parameters are not set either..
                return;
            }
        }
    }


    /** called when server responses with "ok" or "error", (when the command finishes)
     */
    public void commandTerminated(TerminationEvent e)
    {
        if (builder != null)
            builder.outputDone();
    }

    /** Getter for property beforeDate.
     * @return Value of property beforeDate.
     */
    public String getBeforeDate1()
    {
        return beforeDate1;
    }

    /** Setter for property beforeDate.
     * @param beforeDate New value of property beforeDate.
     */
    public void setBeforeDate1(String beforeDate)
    {
        this.beforeDate1 = beforeDate;
    }


    /** Getter for property firstRevision.
     * @return Value of property firstRevision.
     */
    public String getRevision1()
    {
        return revision1;
    }

    /** Setter for property firstRevision.
     * @param firstRevision New value of property firstRevision.
     */
    public void setRevision1(String firstRevision)
    {
        revision1 = firstRevision;
    }

    /** Getter for property secondRevision.
     * @return Value of property secondRevision.
     */
    public String getRevision2()
    {
        return revision2;
    }

    /** Setter for property secondRevision.
     * @param secondRevision New value of property secondRevision.
     */
    public void setRevision2(String secondRevision)
    {
        this.revision2 = secondRevision;
    }

    /** Getter for property beforeDate2.
     * @return Value of property beforeDate2.
     */
    public String getBeforeDate2()
    {
        return beforeDate2;
    }

    /** Setter for property beforeDate2.
     * @param beforeDate2 New value of property beforeDate2.
     */
    public void setBeforeDate2(String beforeDate2)
    {
        this.beforeDate2 = beforeDate2;
    }
    /**
     * Getter for property keywordSubst.
     * @return Value of property keywordSubst.
     */
    public String getKeywordSubst()
    {
        return keywordSubst;
    }

    /**
     * Setter for property keywordSubst.
     * @param keywordSubst New value of property keywordSubst.
     */
    public void setKeywordSubst(String keywordSubst)
    {
        this.keywordSubst = keywordSubst;
    }

   /** This method returns how the command would looklike when typed on the command line.
     * Each command is responsible for constructing this information.
     * @returns <command's name> [<parameters>] files/dirs. Example: checkout -p CvsCommand.java
     *
     */
    public String getCVSCommand() {
        StringBuffer toReturn = new StringBuffer("diff ");
        toReturn.append(getCVSArguments());
        File[] files = getFiles();
        if (files != null) {
            for (int index = 0; index < files.length; index++) {
                toReturn.append(files[index].getName() + " ");
            }
        }
        return toReturn.toString();
    }

    /** takes the arguments and sets the command. To be mainly
     * used for automatic settings (like parsing the .cvsrc file)
     * @return true if the option (switch) was recognized and set
 */
    public boolean setCVSCommand(char opt,String optArg) {
            if (opt == 'R')      setRecursive(true);
            else if (opt == 'l') setRecursive(false);
            else if (opt == 'r')
            {
                if (getRevision1() == null) setRevision1(optArg);
                else setRevision2(optArg);
            } else if (opt == 'D')
            {
                if (getBeforeDate1() == null) setBeforeDate1(optArg);
                else setBeforeDate2(optArg);
            } else if (opt == 'k')
            {
                setKeywordSubst(optArg);
            } else
            {
                return false;
            }
            return true;
    }

    /**
     * String returned by this method defines which options are available for this particular command
 */
    public String getOptString()
    {
        return "Rlr:D:k:";
    }

    /**
     * resets all switches in the command. After calling this method,
     * the command should have no switches defined and should behave defaultly.
 */
    public void resetCVSCommand() {
        setRecursive(true);
        setRevision1(null);
        setRevision2(null);
        setBeforeDate1(null);
        setBeforeDate2(null);
        setKeywordSubst(null);
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
 */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer("");
        if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
        {
            toReturn.append("-k" + getKeywordSubst() + " ");
        }
        if (!getRecursive()) toReturn.append("-l ");
        if (getRevision1() != null) toReturn.append("-r " + getRevision1() + " ");
        if (getBeforeDate1() != null) toReturn.append("-D " + getBeforeDate1() + " ");
        if (getRevision2() != null) toReturn.append("-r " + getRevision2() + " ");
        if (getBeforeDate2() != null) toReturn.append("-D " + getBeforeDate2() + " ");
        return toReturn.toString();
    }

}
