/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.lib.cvsclient.command.diff;

import java.io.File;
import org.netbeans.lib.cvsclient.event.*;
import org.netbeans.lib.cvsclient.command.Builder;

       /**
         * Handles the building of a diff information object and the firing of
         * events when complete objects are built.
        * @author Milos Kleint
         */
 public class SimpleDiffBuilder implements Builder {

    /**
     * The event manager to use
     */
    protected EventManager eventManager;

    protected DiffCommand diffCommand;
        /**
         * The diff object that is currently being built
         */
        protected DiffInformation diffInformation;

        /**
         * The directory in which the file being processed lives. This is
         * relative to the local directory
         */
        protected String fileDirectory;

        protected boolean readingDiffs = false;
        private static final String UNKNOWN = "cvs server: I know nothing about";
        private static final String CANNOT_FIND = "cvs server: cannot find";
        private static final String UNKNOWN_TAG = "cvs server: tag";
        private static final String EXAM_DIR = "cvs server: Diffing";

        private static final String FILE              = "Index: ";
        private static final String RCS_FILE          = "RCS file: ";
        private static final String REVISION          = "retrieving revision ";
        private static final String PARAMETERS        = "diff ";
        private DiffInformation.DiffChange currentChange;

        public SimpleDiffBuilder(EventManager eventMan, DiffCommand diffComm) {
            eventManager = eventMan;
            diffCommand = diffComm;
        }

        public void outputDone() {
            if (diffInformation != null) {
                if (currentChange != null) {
                    diffInformation.addChange(currentChange);
                    currentChange = null;
                }
                eventManager.fireCVSEvent(new FileInfoEvent(this, diffInformation));
                diffInformation = null;
                readingDiffs = false;
            }
        }

        public void parseLine(String line, boolean isErrorMessage) {
            if (readingDiffs) {
                if (line.startsWith(FILE)) {
                    outputDone();
                } else {
                    processDifferences(line);
                    return;
                }
            }
            if (line.startsWith(UNKNOWN)) {
                eventManager.fireCVSEvent(new FileInfoEvent(this, diffInformation));
                diffInformation = null;
                return;
            }
            if (line.startsWith(EXAM_DIR)) {
                fileDirectory = line.substring(EXAM_DIR.length()).trim();
                return;
            }
            if (line.startsWith(FILE)) {
                processFile(line.substring(FILE.length()));
                return;
            }
            if (line.startsWith(RCS_FILE)) {
                processRCSfile(line.substring(RCS_FILE.length()));
                return;
            }
            if (line.startsWith(REVISION)) {
                processRevision(line.substring(REVISION.length()));
                return;
            }
            if (line.startsWith(PARAMETERS)) {
                processParameters(line.substring(PARAMETERS.length()));
                readingDiffs = true;
                return;
            }
        }

/*        protected void processDifferences(String line) {
            diffInformation.addToDifferences(line);
        }
 */
        protected void processFile(String line) {
            outputDone();
            diffInformation = createDiffInformation();
            String fileName = line.trim();
            if (fileName.startsWith("no file")) {
                fileName = fileName.substring(8);
            }
            diffInformation.setFile(new File(diffCommand.getLocalPath(),
//            ((fileDirectory!=null)?fileDirectory:  "") + File.separator +
            fileName));
        }

        protected void processRCSfile(String line) {
            if (diffInformation == null) return;
            diffInformation.setRepositoryFileName(line.trim());
        }

        protected void processRevision(String line) {
            if (diffInformation == null) return;
            line = line.trim();
            // first REVISION line is the from-file, the second is the to-file
            if (diffInformation.getLeftRevision() != null) {
                diffInformation.setRightRevision(line);
            } else {
                diffInformation.setLeftRevision(line);
            }
        }

        protected void processParameters(String line) {
            if (diffInformation == null) return;
            diffInformation.setParameters(line.trim());
        }

        public DiffInformation createDiffInformation() {
            return new DiffInformation();
        }

        protected void assignType(DiffInformation.DiffChange change, String line) {
            int index = 0;
            int cIndex = line.indexOf('c');
            if (cIndex > 0) {
                // change type of change
                change.setType(DiffInformation.DiffChange.CHANGE);
                index = cIndex;
            } else {
                int aIndex = line.indexOf('a');
                if (aIndex > 0) {
                    // add type of change
                    change.setType(DiffInformation.DiffChange.ADD);
                    index = aIndex;
                } else {
                    int dIndex = line.indexOf('d');
                    if (dIndex > 0) {
                        // delete type of change
                        change.setType(DiffInformation.DiffChange.DELETE);
                        index = dIndex;
                    }
                }
            }
            String left = line.substring(0,index);
//            System.out.println("left part of change=" + left);
            change.setLeftRange(getMin(left), getMax(left));
            String right = line.substring(index + 1);
//            System.out.println("right part of change=" + right);
            change.setRightRange(getMin(right), getMax(right));
        }

        private int getMin(String line) {
            String nums = line;
            int commaIndex = nums.indexOf(',');
            if (commaIndex > 0) {
                nums = nums.substring(0, commaIndex);
            }
            int min;
            try {
                min = Integer.parseInt(nums);
            } catch (NumberFormatException exc) {
                min = 0;
            }
//            System.out.println("Min=" + min);
            return min;
        }

        private int getMax(String line) {
            String nums = line;
            int commaIndex = nums.indexOf(',');
            if (commaIndex > 0) {
                nums = nums.substring(commaIndex + 1);
            }
            int max;
            try {
                max = Integer.parseInt(nums);
            } catch (NumberFormatException exc) {
                max = 0;
            }
//            System.out.println("Max=" + max);
            return max;
        }

        protected void processDifferences(String line) {
            char firstChar = line.charAt(0);
            if (firstChar >= '0' && firstChar <= '9') {
                // we got a new difference here
//                System.out.println("new Change=" + line);
                if (currentChange != null) {
                    diffInformation.addChange(currentChange);
                }
                currentChange = diffInformation.createDiffChange();
                assignType(currentChange, line);
            }
            if (firstChar == '<') {
//                System.out.println("Left line=" + line);
                currentChange.appendLeftLine(line.substring(2));
            }
            if (firstChar == '>') {
//                System.out.println("right line=" + line);
                currentChange.appendRightLine(line.substring(2));
            }

        }

 }
