/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.lib.cvsclient.command.log;

    /**
     * Handles the building of a log information object and the firing of
     * events when complete objects are built.
     * @author Milos Kleint
     */
import java.io.File;
import org.netbeans.lib.cvsclient.event.*;
import org.netbeans.lib.cvsclient.command.Builder;
import java.util.StringTokenizer;

public class LogBuilder implements Builder {

    private static final String LOGGING_DIR = "cvs server: Logging ";
    private  int LOGGING_DIR_LENGTH = LOGGING_DIR.length();
    private static final String RCS_FILE = "RCS file: ";
    private int RCS_FILE_LENGTH = RCS_FILE.length();
    private static final String WORK_FILE = "Working file: ";
    private int WORK_FILE_LENGTH = WORK_FILE.length();
    private static final String REV_HEAD = "head: ";
    private int REV_HEAD_LENGTH = REV_HEAD.length();
    private static final String BRANCH = "branch: ";
    private int BRANCH_LENGTH = BRANCH.length();
    private static final String LOCKS = "locks: ";
    private int LOCKS_LENGTH = LOCKS.length();
    private static final String ACCESS_LIST = "access list: ";
    private int ACCESS_LIST_LENGTH = ACCESS_LIST.length();
    private static final String SYM_NAME = "symbolic names:";
    private int SYM_NAME_LENGTH = SYM_NAME.length();
    private static final String KEYWORD_SUBST = "keyword substitution: ";
    private int KEYWORD_SUBST_LENGTH = KEYWORD_SUBST.length();
    private static final String TOTAL_REV = "total revisions: ";
    private int TOTAL_REV_LENGTH = TOTAL_REV.length();
    private static final String SEL_REV = ";	selected revisions: ";
    private int SEL_REV_LENGTH = SEL_REV.length();
    private static final String DESCRIPTION = "description:";
    private int DESCRIPTION_LENGTH = DESCRIPTION.length();

    private static final String REVISION = "revision ";
    private int REVISION_LENGTH = REVISION.length();
    private static final String DATE = "date: ";
    private int DATE_LENGTH = DATE.length();
    private static final String BRANCHES = "branches: ";
    private int BRANCHES_LENGTH = BRANCHES.length();
    private static final String AUTHOR = "  author: ";
    private int AUTHOR_LENGTH = AUTHOR.length();
    private static final String STATE = "  state: ";
    private int STATE_LENGTH = STATE.length();
    private static final String LINES = "  lines: ";
    private int LINES_LENGTH = LINES.length();
    private static final String SPLITTER = "----------------------------";
    private static final String FINAL_SPLIT = "=============";
    private static final String ERROR = "cvs server: nothing known about ";

    /**
     * The event manager to use
     */
    protected EventManager eventManager;

    protected LogCommand logCommand;
        /**
         * The log object that is currently being built
         */
    protected LogInformation logInfo;

    protected LogInformation.Revision revision;

        /**
         * The directory in which the file being processed lives. This is
         * relative to the local directory
         */
    protected String fileDirectory;

    private boolean addingSymNames;
    private boolean addingDescription;
    private boolean addingLogMessage;
    StringBuffer tempBuffer = null;

    public LogBuilder(EventManager eventMan,LogCommand command)
{
    logCommand = command;
    eventManager = eventMan;
    addingSymNames = false;
    addingDescription = false;
    addingLogMessage = false;
    logInfo = null;
    revision = null;

    }

    public void outputDone() {
        if (logInfo != null) {
            eventManager.fireCVSEvent(new FileInfoEvent(this, logInfo));
            logInfo = null;
        }
    }


    public void parseLine(String line, boolean isErrorMessage) {
        if (line.startsWith(FINAL_SPLIT)) {
            if (addingDescription) {
                addingDescription = false;
                logInfo.setDescription(tempBuffer.toString());
            }
            if (addingLogMessage) {
                addingLogMessage = false;
                revision.setMessage(tempBuffer.toString());
            }
            if (revision != null) {
                logInfo.addRevision(revision);
                revision = null;
            }
            // fire the event and exit
            outputDone();
            tempBuffer = null;
            return;
        }
        if (addingLogMessage) {
            // first check for the branches tag
            if (line.startsWith(BRANCHES)) {
                processBranches(line.substring(BRANCHES.length()));
            } else {
                processLogMessage(line);
            }
        }
        if (addingSymNames) {
            processSymbolicNames(line);
        }
        if (addingDescription) {
            processDescription(line);
        }
        // revision stuff first -> will be  the most common to parse
        if (line.startsWith(REVISION)) {
            processRevisionStart(line);
        }
        if (line.startsWith(DATE)) {
            processRevisionDate(line);
        }


        if (line.startsWith(KEYWORD_SUBST)) {
            logInfo.setKeywordSubstitution(line.substring(KEYWORD_SUBST_LENGTH).trim());
            addingSymNames = false;
            return;
        }

        if (line.startsWith(DESCRIPTION)) {
            tempBuffer = new StringBuffer("" + line.substring(DESCRIPTION_LENGTH));
            addingDescription = true;
        }

        if (line.startsWith(LOGGING_DIR)) {
            fileDirectory = line.substring(LOGGING_DIR_LENGTH);
            return;
        }
        if (line.startsWith(RCS_FILE)) {
            processRcsFile(line.substring(RCS_FILE_LENGTH));
            return;
        }
        if (line.startsWith(WORK_FILE)) {
            processWorkingFile(line.substring(WORK_FILE_LENGTH));
            return;
        }
        if (line.startsWith(REV_HEAD)) {
            logInfo.setHeadRevision(line.substring(REV_HEAD_LENGTH).trim());
            return;
        }
        if (line.startsWith(BRANCH)) {
            logInfo.setBranch(line.substring(BRANCH_LENGTH).trim());
        }
        if (line.startsWith(LOCKS)) {
            logInfo.setLocks(line.substring(LOCKS_LENGTH).trim());
        }
        if (line.startsWith(ACCESS_LIST)) {
            logInfo.setAccessList(line.substring(ACCESS_LIST_LENGTH).trim());
        }
        if (line.startsWith(SYM_NAME)) {
            addingSymNames = true;
        }
        if (line.startsWith(TOTAL_REV)) {
            int ind = line.indexOf(';');
            if (ind < 0) { // no selected revisions here..
                logInfo.setTotalRevisions(line.substring(TOTAL_REV_LENGTH).trim());
                logInfo.setSelectedRevisions("0");
            } else {
                String total = line.substring(0,ind);
                String select = line.substring(ind, line.length());
                logInfo.setTotalRevisions(total.substring(TOTAL_REV_LENGTH).trim());
                logInfo.setSelectedRevisions(select.substring(SEL_REV_LENGTH).trim());
            }
        }
    }

    private void processRcsFile(String line) {
        if (logInfo != null) {
            //do fire logcreated event;
        }
        logInfo = new LogInformation();
        logInfo.setRepositoryFilename(line.trim());
    }

    private void processWorkingFile(String line) {
        String fileName = line.trim();
        if (fileName.startsWith("no file"))
            fileName = fileName.substring(8);
        //System.err.println("Filename found is: " + fileName);

        logInfo.setFile(createFile(line));
    }

    private void processBranches(String line) {
        int ind = line.lastIndexOf(';');
        if (ind > 0) {
            line = line.substring(0, ind);
        }
        revision.setBranches(line.trim());
    }

    private void processLogMessage(String line) {
        if (line.startsWith(SPLITTER)) {
            addingLogMessage = false;
            revision.setMessage(tempBuffer.toString());
            return;
        }
        tempBuffer.append(line + "\n");
    }

    private void processSymbolicNames(String line) {
        if (!line.startsWith(KEYWORD_SUBST)) {
            line = line.trim();
            int index = line.indexOf(':');
            if (index > 0) {
                String symName = line.substring(0,index).trim();
                String revName = line.substring(index + 1,line.length()).trim();
                logInfo.addSymbolicName(symName,revName);
            }
        }
    }
    private void processDescription(String line) {
        if (line.startsWith(SPLITTER)) {
            addingDescription = false;
            logInfo.setDescription(tempBuffer.toString());
            return;
        }
        tempBuffer.append(line);
    }

    private void processRevisionStart(String line) {
        if (revision != null) {
            logInfo.addRevision(revision);
        }
        revision = logInfo.createNewRevision(line.substring(REVISION_LENGTH));
    }

    private void processRevisionDate(String line) {
        StringTokenizer token = new StringTokenizer(line, ";", false);
        if (token.hasMoreTokens()) revision.setDate(token.nextToken().substring(DATE_LENGTH));
        if (token.hasMoreTokens()) revision.setAuthor(token.nextToken().substring(AUTHOR_LENGTH));
        if (token.hasMoreTokens()) revision.setState(token.nextToken().substring(STATE_LENGTH));
        if (token.hasMoreTokens()) revision.setLines(token.nextToken().substring(LINES_LENGTH));
        addingLogMessage = true;
        tempBuffer = new StringBuffer("");
    }


    protected File createFile(String fileName) {
        StringBuffer path = new StringBuffer();
        path.append(logCommand.getLocalPath());
        path.append(File.separator);
        if (fileDirectory == null) {
            // happens for single files only
            // (for directories, the dir name is always sent before the actual files)
            File locFile = logCommand.getFileEndingWith(fileName);
            if (locFile == null) {
//                System.out.println("JAVACVS ERROR!! wrong algorithm for assigning path to single files(1)!!");
                path.append(fileName);
            } else {
                path = new StringBuffer(locFile.getAbsolutePath());
            }
        } else {
//            path.append(fileDirectory);
//            path.append(File.separator);
            path.append(fileName);
        }
        return new File(path.toString());
    }
}
