/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.lib.cvsclient.command.status;

import java.io.File;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.FileInfoEvent;
import org.netbeans.lib.cvsclient.command.Builder;

/**
 * Handles the building of a status information object and the firing of
 * events when complete objects are built.
 *
 * @author  Milos Kleint
 * @author  Thomas Singer
 */
public class StatusBuilder implements Builder
{
    private static final String UNKNOWN = "cvs server: nothing known about";
    private static final String EXAM_DIR = "cvs server: Examining";
    private static final String NOT_IN_REPOSITORY = "No revision control file";

    private static final String FILE              = "File: ";
    private static final String WORK_REV          = "   Working revision:";
    private static final String REP_REV           = "   Repository revision:";
    private static final String TAG               = "   Sticky Tag:";
    private static final String DATE              = "   Sticky Date:";
    private static final String OPTIONS           = "   Sticky Options:";
    private static final String EXISTING_TAGS     = "   Existing Tags:";
    private static final String EMPTY_BEFORE_TAGS = "   ";
    private static final String NO_TAGS           = "   No Tags Exist";

    /**
     * The status object that is currently being built.
     */
    private StatusInformation statusInformation;

    /**
     * The event manager to use.
     */
    private EventManager eventManager;

    private StatusCommand statusCommand;
    private String localPath;
    private String relativeDirectory;

    private boolean readingTags;

    public StatusBuilder(EventManager  eventManager,
                         StatusCommand statusCommand)
    {
        this.eventManager    = eventManager;
        this.statusCommand   = statusCommand;
        this.localPath       = statusCommand.getLocalPath();
    }

    public void outputDone()
    {
        if (statusInformation != null)
        {
            eventManager.fireCVSEvent(new FileInfoEvent(this, statusInformation));
            statusInformation = null;
            readingTags = false;
        }
    }

    public void parseLine(String line, boolean isErrorMessage)
    {
        if (readingTags)
        {
            if (line.startsWith(NO_TAGS))
            {
                outputDone();
                return;
            }

            int bracket = line.indexOf('(');
            if (bracket > 0)
            {
                // it's another tag..
                String tag = line.substring(0,bracket - 1).trim();
                String rev = line.substring(bracket + 1, line.length() - 1);

                if (statusInformation == null)
                {
                    statusInformation = new StatusInformation();
                }
                statusInformation.addExistingTag(tag, rev);
            }
            else
            {
                outputDone();
                return;
            }
        }

        if (line.startsWith(UNKNOWN))
        {
            outputDone();
        }
        else if (line.startsWith(EXAM_DIR))
        {
            relativeDirectory = line.substring(22);
        }
        else if (line.startsWith(FILE))
        {
            statusInformation = new StatusInformation();
            processFileAndStatusLine(line.substring(FILE.length()));
        }
        else if (line.startsWith(WORK_REV))
        {
            processWorkRev(line.substring(WORK_REV.length()));
        }
        else if (line.startsWith(REP_REV))
        {
            processRepRev(line.substring(REP_REV.length()));
/*            if (statusInformation.getRepositoryRevision().startsWith(NOT_IN_REPOSITORY))
            {
                outputDone();
            }
 */
        }
        else if (line.startsWith(TAG))
        {
            processTag(line.substring(TAG.length()));
        }
        else if (line.startsWith(DATE))
        {
            processDate(line.substring(DATE.length()));
        }
        else if (line.startsWith(OPTIONS))
        {
            processOptions(line.substring(OPTIONS.length()));
            if (!statusCommand.isIncludeTags())
            {
                outputDone();
            }
        }
        else if (line.startsWith(EXISTING_TAGS))
        {
            readingTags = true;
        }
    }

    private File createFile(String fileName)
    {
        if (relativeDirectory != null)
        {
            if (relativeDirectory.trim().equals(".")) 
            {
                return new File(localPath, fileName);
            }
            return new File(localPath, relativeDirectory + '/' + fileName);
        }

        // happens for single files only
        // (for directories, the dir name is always sent before the actual files)
        File file = statusCommand.getFileEndingWith(fileName);
        if (file != null)
        {
            return file;
        }

//        System.err.println("JAVACVS ERROR!! wrong algorithm for assigning path to single files(1)!!");

        return null;
    }

    private void processFileAndStatusLine(String line)
    {
        int statusIndex = line.lastIndexOf("Status:");
        String fileName = line.substring(0, statusIndex).trim();
        if (fileName.startsWith("no file"))
        {
            fileName = fileName.substring(8);
        }

        statusInformation.setFile(createFile(fileName));

        String status = line.substring(statusIndex + 8).trim();
        statusInformation.setStatusString(status);
    }

    private boolean assertNotNull()
    {
        if (statusInformation == null)
        {
//            System.err.println("Bug: statusInformation must not be null!");
            return false;
        }

        return true;
    }

    private void processWorkRev(String line)
    {
        if (!assertNotNull())
        {
            return;
        }
        statusInformation.setWorkingRevision(line.trim());
    }

    private void processRepRev(String line)
    {
        if (!assertNotNull())
        {
            return;
        }
        line = line.trim();
        if (line.startsWith(NOT_IN_REPOSITORY))
        {
            statusInformation.setRepositoryRevision(line.trim());
            return;
        }

        int firstSpace = line.indexOf('\t');
        if (firstSpace > 0)
        {
            statusInformation.setRepositoryRevision(
                                          line.substring(0, firstSpace).trim());
            statusInformation.setRepositoryFileName(
                                             line.substring(firstSpace).trim());
        }
        else
        {
            statusInformation.setRepositoryRevision("");
            statusInformation.setRepositoryFileName("");
        }
    }

    private void processTag(String line)
    {
        if (!assertNotNull())
        {
            return;
        }
        statusInformation.setStickyTag(line.trim());
    }

    private void processDate(String line)
    {
        if (!assertNotNull())
        {
            return;
        }
        statusInformation.setStickyDate(line.trim());
    }

    private void processOptions(String line)
    {
        if (!assertNotNull())
        {
            return;
        }
        statusInformation.setStickyOptions(line.trim());
    }

}
