/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.status;

import java.io.File;
import java.util.LinkedList;
import java.util.List;
import java.util.Iterator;
import org.netbeans.lib.cvsclient.command.FileInfoContainer;
import org.netbeans.lib.cvsclient.file.FileStatus;

/**
 * Describes status information for a file. This is the result of doing a
 * cvs status command. The fields in instances of this object are populated
 * by response handlers.
 * @author  Robert Greig
 * @version $Revision: 1.6 $
 */
public class StatusInformation extends FileInfoContainer
{
    public static final String UP_TO_DATE                  = "Up-to-date";
    public static final String LOCALLY_MODIFIED            = "Locally Modified";
    public static final String LOCALLY_ADDED               = "Locally Added";
    public static final String LOCALLY_REMOVED             = "Locally Removed";
    public static final String NEEDS_CHECKOUT              = "Needs Checkout";
    public static final String NEEDS_PATCH                 = "Needs Patch";
    public static final String NEEDS_MERGE                 = "Needs Merge";
    public static final String FILE_HAD_CONFLICTS_ON_MERGE = "File had conflicts on merge";
    public static final String UNKNOWN                     = "Unknown";

    /**
     * Holds value of property file.
     */
    private File file;

    /**
     * Holds value of property status.
     */
    private FileStatus status;

    /**
     * The string version of the status.
     */
    private String statusString;

    /**
     * Holds value of property workingRevision.
     */
    private String workingRevision;

    /**
     * Holds value of property repositoryRevision.
     */
    private String repositoryRevision;

    /** Holds value of property repositoryFileName.
     */
    private String repositoryFileName;

    /**
     * Holds value of property stickyDate.
     */
    private String stickyDate;

    /**
     * Holds value of property stickyOptions.
     */
    private String stickyOptions;

    /**
     * Holds value of property stickTag.
     */
    private String stickyTag;

    /**
     * Hold key pairs of existing tags
     */
    private LinkedList tags;

    public StatusInformation()
    {
        tags = new LinkedList();
    }

    /**
     * Getter for property file.
     * @return Value of property file.
     */
    public File getFile()
    {
        return file;
    }

    /**
     * Setter for property file.
     * @param file New value of property file.
     */
    public void setFile(File file)
    {
        this.file = file;
    }

    /**
     * Getter for property status.
     * @return Value of property status.
     */
    public FileStatus getStatus()
    {
        return status;
    }

    /**
     * Setter for property status.
     * @param status New value of property status.
     * @deprecated use setStatusByString instead
     */
    public void setStatus(FileStatus status)
    {
        this.status = status;
    }

    /**
     * Returns the status as a String.
     * The String returned are definitely the static-final-instances.
     */
    public String getStatusString() {
        return statusString;
    }

    /**
     * Sets the status by the specified string.
     */
    public void setStatusString(String statusString)
    {
        if (statusString.equals(UP_TO_DATE))
        {
            setStatus(FileStatus.UP_TO_DATE);
            this.statusString = UP_TO_DATE;
        }
        else if (statusString.equals(LOCALLY_MODIFIED))
        {
            setStatus(FileStatus.MODIFIED);
            this.statusString = LOCALLY_MODIFIED;
        }
        else if (statusString.equals(LOCALLY_ADDED))
        {
            setStatus(FileStatus.ADDED);
            this.statusString = LOCALLY_ADDED;
        }
        else if (statusString.equals(LOCALLY_REMOVED))
        {
            setStatus(FileStatus.REMOVED);
            this.statusString = LOCALLY_REMOVED;
        }
        else if (statusString.equals(NEEDS_CHECKOUT))
        {
            setStatus(FileStatus.NEEDS_CHECKOUT);
            this.statusString = NEEDS_CHECKOUT;
        }
        else if (statusString.equals(NEEDS_PATCH))
        {
            setStatus(FileStatus.NEEDS_PATCH);
            this.statusString = NEEDS_PATCH;
        }
        else if (statusString.equals(NEEDS_MERGE))
        {
            setStatus(FileStatus.NEEDS_MERGE);
            this.statusString = NEEDS_MERGE;
        }
        else if (statusString.equals(FILE_HAD_CONFLICTS_ON_MERGE))
        {
            setStatus(FileStatus.HAS_CONFLICTS);
            this.statusString = FILE_HAD_CONFLICTS_ON_MERGE;
        }
        else if (statusString.equals(UNKNOWN))
        {
            setStatus(FileStatus.UNKNOWN);
            this.statusString = UNKNOWN;
        }
    }

    /**
     * Getter for property workingRevision.
     * @return Value of property workingRevision.
     */
    public String getWorkingRevision()
    {
        return workingRevision;
    }

    /**
     * Setter for property workingRevision.
     * @param workingRevision New value of property workingRevision.
     */
    public void setWorkingRevision(String workingRevision)
    {
        this.workingRevision = workingRevision;
    }

    /**
     * Getter for property repositoryRevision.
     * @return Value of property repositoryRevision.
     */
    public String getRepositoryRevision()
    {
        return repositoryRevision;
    }

    /**
     * Setter for property repositoryRevision.
     * @param repositoryRevision New value of property repositoryRevision.
     */
    public void setRepositoryRevision(String repositoryRevision)
    {
        this.repositoryRevision = repositoryRevision;
    }

    /**
     * Getter for property repositoryFileName.
     * @return Value of property repositoryFileName.
     */
    public String getRepositoryFileName()
    {
        return repositoryFileName;
    }

    /**
     * Setter for property repositoryFileName.
     * @param repositoryRevision New value of property repositoryFileName.
     */
    public void setRepositoryFileName(String repositoryFileName)
    {
        this.repositoryFileName = repositoryFileName;
    }

    /**
     * Getter for property stickyTag.
     * @return Value of property stickyTag.
     */
    public String getStickyTag()
    {
        return stickyTag;
    }

    /**
     * Setter for property stickyTag.
     * @param stickyTag New value of property stickyTag.
     */
    public void setStickyTag(String stickyTag)
    {
        this.stickyTag = stickyTag;
    }

    /**
     * Getter for property stickyDate.
     * @return Value of property stickyDate.
     */
    public String getStickyDate()
    {
        return stickyDate;
    }

    /**
     * Setter for property stickyDate.
     * @param stickyDate New value of property stickyDate.
     */
    public void setStickyDate(String stickyDate)
    {
        this.stickyDate = stickyDate;
    }

    /**
     * Getter for property stickyOptions.
     * @return Value of property stickyOptions.
     */
    public String getStickyOptions()
    {
        return stickyOptions;
    }

    /**
     * Setter for property stickyOptions.
     * @param stickyOptions New value of property stickyOptions.
     */
    public void setStickyOptions(String stickyOptions)
    {
        this.stickyOptions = stickyOptions;
    }

    public void addExistingTag(String tagName, String revisionNumber)
    {
        SymName tag = new SymName();
        tag.setTag(tagName);
        tag.setRevision(revisionNumber);
        tags.add(tag);
    }

    public LinkedList getAllExistingTags()
    {
        return tags;
    }

    public void setAllExistingTags(LinkedList newList)
    {
        tags = newList;
    }


    /** Search the symbolic names by number of revision. If not found, return null.
     */
    public List getSymNamesForRevision(String revNumber)
    {
        StatusInformation.SymName item;
        LinkedList list = new LinkedList();

        for (Iterator it = tags.iterator(); it.hasNext(); )
        {
            item = (StatusInformation.SymName)it.next();
            if (item.getRevision().equals(revNumber))
            {
                list.add(item);
            }
        }
        return list;
    }

    /**
     * Search the symbolic names by name of tag (symbolic name).
     * If not found, return null.
     */
    public StatusInformation.SymName getSymNameForTag(String tagName)
    {
        StatusInformation.SymName item;

        for (Iterator it = tags.iterator(); it.hasNext(); )
        {
            item = (StatusInformation.SymName)it.next();
            if (item.getTag().equals(tagName))
            {
                return item;
            }
        }
        return null;
    }


    /**
     * Return a string representation of this object. Useful for debugging.
     */
    public String toString()
    {
        StringBuffer buf = new StringBuffer();
        buf.append("\nFile: ");
        buf.append((file != null) ? file.getAbsolutePath()
                                  :"null");
        buf.append("\nStatus is: ");
        buf.append(statusString);
        buf.append("\nWorking revision: ");
        buf.append(workingRevision);
        buf.append("\nRepository revision: ");
        buf.append("\nSticky date: ");
        buf.append(stickyDate);
        buf.append("\nSticky options: ");
        buf.append(stickyOptions);
        buf.append("\nSticky tag: ");
        buf.append(stickyTag);
        if (tags.size() > 0)
        {
            // we are having some tags to print
            buf.append("\nExisting Tags:");
            for (Iterator it = tags.iterator(); it.hasNext(); )
            {
                buf.append("\n  ");
                buf.append(it.next().toString());
            }
        }
        return buf.toString();
    }

    /**
     * An inner class storing information about a symbolic name.
     * Consists of a pair of Strings. tag + revision.
     */
    public class SymName
    {
        private String tag;
        private String revision;

        public SymName()
        {
        }

        public String getTag()
        {
            return tag;
        }

        public void setTag(String symName)
        {
            tag = symName;
        }

        public void setRevision(String rev)
        {
            revision = rev;
        }

        public String getRevision()
        {
            return revision;
        }

        public String toString()
        {
            return getTag() + " : " + getRevision();
        }
    }
}
