/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.tag;

import java.io.File;

import org.netbeans.lib.cvsclient.ClientServices;

import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.event.FileInfoEvent;
import org.netbeans.lib.cvsclient.event.TerminationEvent;

import org.netbeans.lib.cvsclient.file.FileStatus;

import org.netbeans.lib.cvsclient.command.BasicCommand;
import org.netbeans.lib.cvsclient.command.Builder;
import org.netbeans.lib.cvsclient.command.CommandException;

import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.TagRequest;
import org.netbeans.lib.cvsclient.request.ArgumentRequest;

/**
 * The tag command adds or deleted a tag to the specified files/directories.
 *
 * @author  Thomas Singer
 * @version $Revision: 1.5.2.1 $
 */
public class TagCommand extends BasicCommand
{
    /**
     * The event manager to use.
     */
    private EventManager eventManager;

    private boolean checkThatUnmodified;

    private boolean deleteTag;

    private boolean makeBranchTag;

    private boolean overrideExistingTag;

    private String tag;

    private String tagByDate;
    
    private String tagByRevision;
    
    /**
     * Construct a new tag command.
     */
    public TagCommand()
    {
    }

    /**
     * Creates the TagBuilder.
     * @param eventManager the event manager used to received cvs events
     */
    protected Builder createBuilder(EventManager eventManager)
    {
        return new TagBuilder(eventManager, getLocalPath());
    }

    /**
     * Returns true if checking for unmodified files is enabled.
     */
    public boolean doesCheckThatUnmodified()
    {
        return checkThatUnmodified;
    }

    /**
     * Enabled the check for unmodified files.
     */
    public void setCheckThatUnmodified(boolean checkThatUnmodified)
    {
        this.checkThatUnmodified = checkThatUnmodified;
    }

    /**
     * Returnes true if the tag should be deleted (otherwise added).
     */
    public boolean doesDeleteTag()
    {
        return deleteTag;
    }

    /**
     * Sets whether the tag should be deleted (true) or added (false).
     */
    public void setDeleteTag(boolean deleteTag)
    {
        this.deleteTag = deleteTag;
    }

    /**
     * Returns true if the tag should be a branch tag.
     */
    public boolean doesMakeBranchTag()
    {
        return makeBranchTag;
    }

    /**
     * Sets whether the tag should be a branch tag.
     */
    public void setMakeBranchTag(boolean makeBranchTag)
    {
        this.makeBranchTag = makeBranchTag;
    }

    /**
     * Returns true to indicate that existing tag will be overridden.
     */
    public boolean doesOverrideExistingTag()
    {
        return overrideExistingTag;
    }

    /**
     * Sets whether existing tags should be overridden.
     */
    public void setOverrideExistingTag(boolean overrideExistingTag)
    {
        this.overrideExistingTag = overrideExistingTag;
    }

    /**
     * Returns the tag that should be added or deleted.
     */
    public String getTag()
    {
        return tag;
    }

    /**
     * Sets the tag that should be added or deleted.
     */
    public void setTag(String tag)
    {
        this.tag = tag;
    }
    
   /** 
     * Returns the latest date of a revision to be tagged. 
     * @return date value. the latest Revision not later ten date is tagged.
     */
    public String getTagByDate() 
    {
        return tagByDate;
    }
    
    /** 
     * Sets the latest date of a revision to be tagged. 
     * @param tagDate New value of property tagDate.
     */
    public void setTagByDate(String tagDate) 
    {
        tagByDate = tagDate;
    }
    
    /** 
     * Sets the latest date of a revision to be tagged. Can be both a number and a tag.
     * @return Value of property tagRevision.
     */
    public String getTagByRevision() 
    {
        return tagByRevision;
    }
    
    /** 
     * Sets the latest date of a revision to be tagged. Can be both a number and a tag.
     * @param tagRevision New value of property tagRevision.
     */
    public void setTagByRevision(String tagRevision) 
    {
        tagByRevision = tagRevision;
    }

    /**
     * Execute the command.
     *
     * @param client the client services object that provides any necessary
     *               services to this command, including the ability to actually
     *               process all the requests.
     */
    public void execute(ClientServices client, EventManager eventManager)
        throws CommandException
    {
        this.eventManager = eventManager;

        super.execute(client, eventManager);

        try
        {
            requests.add(1, new ArgumentRequest(getTag()));

            if (checkThatUnmodified)
            {
                requests.add(1, new ArgumentRequest("-c"));
            }

            if (overrideExistingTag)
            {
                requests.add(1, new ArgumentRequest("-F"));
            }

            if (makeBranchTag)
            {
                requests.add(1, new ArgumentRequest("-b"));
            }

            if (deleteTag)
            {
                requests.add(1, new ArgumentRequest("-d"));
            }
            if (tagByDate != null && (!tagByDate.equals("")))
            {
                requests.add(1, new ArgumentRequest("-D"));
                requests.add(2, new ArgumentRequest(getTagByDate()));
            }
            if (tagByRevision != null && (!tagByRevision.equals("")))
            {
                requests.add(1, new ArgumentRequest("-r"));
                requests.add(2, new ArgumentRequest(getTagByRevision()));
            }

            // now add the request that indicates the working directory for the
            // command

            requests.add(new DirectoryRequest(".",
                                  client.getRepositoryForDirectory(getLocalPath())));

            addArgumentRequests();
            
            requests.add(new TagRequest());

            client.processRequests(requests);
            requests.clear();
        }
        catch (Exception e)
        {
            requests.clear();
            throw new CommandException(e);
        }
    }

    /**
     * Called when server responses with "ok" or "error", (when the command
     * finishes).
     */
    public void commandTerminated(TerminationEvent e)
    {
        if (builder != null)
        {
            builder.outputDone();
        }
    }

    /**
     * This method returns how the tag command would looklike when typed on the
     * command line.
     */
    public String getCVSCommand()
    {
        StringBuffer toReturn = new StringBuffer("tag ");
        toReturn.append(getCVSArguments());
        toReturn.append(getTag());
        toReturn.append(" ");
        File[] files = getFiles();
        if (files != null)
        {
            for (int index = 0; index < files.length; index++)
            {
                toReturn.append(files[index].getName());
                toReturn.append(' ');
            }
        }
        return toReturn.toString();
    }

    /**
     * Takes the arguments and sets the command.
     * To be mainly used for automatic settings (like parsing the .cvsrc file)
     * @return true if the option (switch) was recognized and set
     */
    public boolean setCVSCommand(char opt,String optArg)
    {
        if (opt == 'R')
        {
            setRecursive(true);
        }
        else if (opt == 'l')
        {
            setRecursive(false);
        }
        else if (opt == 'c')
        {
            setCheckThatUnmodified(true);
        }
        else if (opt == 'd')
        {
            setDeleteTag(true);
        }
        else if (opt == 'F')
        {
            setOverrideExistingTag(true);
        }
        else if (opt == 'b')
        {
            setMakeBranchTag(true);
        }
        else if (opt == 'D')
        {
            setTagByDate(optArg.trim());
        }
        else if (opt == 'r')
        {
            setTagByRevision(optArg.trim());
        }
        else
        {
            return false;
        }
        return true;
    }

    /**
     * String returned by this method defines which options are available for
     * this command.
     */
    public String getOptString()
    {
        return "RlcFbdD:r:";
    }

    /**
     * Resets all switches in the command.
     * After calling this method, the command should have no switches defined
     * and should behave defaultly.
     */
    public void resetCVSCommand()
    {
        setRecursive(true);
        setCheckThatUnmodified(false);
        setDeleteTag(false);
        setMakeBranchTag(false);
        setOverrideExistingTag(false);
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
     */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer();
        if (!getRecursive())
        {
            toReturn.append("-l ");
        }
        if (doesCheckThatUnmodified())
        {
            toReturn.append("-c ");
        }
        if (doesOverrideExistingTag())
        {
            toReturn.append("-F ");
        }
        if (doesMakeBranchTag())
        {
            toReturn.append("-b ");
        }
        if (doesDeleteTag())
        {
            toReturn.append("-d ");
        }
        if (getTagByRevision() != null && (!getTagByRevision().equals("")))
        {
            toReturn.append("-r ");
            toReturn.append(getTagByRevision());
            toReturn.append(" ");
        }
        if (getTagByDate() != null && (!getTagByDate().equals("")))
        {
            toReturn.append("-D ");
            toReturn.append(getTagByDate());
            toReturn.append(" ");
        }
        return toReturn.toString();
    }
    

    
}
