/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
package org.netbeans.lib.cvsclient.command.update;

import java.io.File;

import org.netbeans.lib.cvsclient.command.Builder;
import org.netbeans.lib.cvsclient.command.DefaultFileInfoContainer;
import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.FileInfoEvent;

/**
 * Handles the building of update information object and the firing of
 * events when complete objects are built.
 *
 * @author  Milos Kleint, Thomas Singer
 * @version $Revision: 1.13.2.1 $
 */
public class UpdateBuilder implements Builder
{
    /**
     * The status object that is currently being built.
     */
    private DefaultFileInfoContainer fileInfoContainer;

    /**
     * The event manager to use.
     */
    private EventManager eventManager;

    /**
     * The local path the command run in.
     */
    private String localPath;

    private boolean readingTags;

    private static final String UNKNOWN = "cvs server: nothing known about";
    private static final String EXAM_DIR = "cvs server: Updating";
    private static final String TO_ADD = "cvs server: use `cvs add' to create an entry for";
    private static final String STATES = "U P A R M C ? ";
    private static final String WARNING = "cvs server: warning: ";
    private static final String PERTINENT = "is not (any longer) pertinent";

    public UpdateBuilder(EventManager eventManager, String localPath)
    {
        this.eventManager = eventManager;
        this.localPath = localPath;
    }

    public void outputDone()
    {
        if (fileInfoContainer != null)
        {
            eventManager.fireCVSEvent(new FileInfoEvent(this, fileInfoContainer));
            fileInfoContainer = null;
        }
    }

    public void parseLine(String line, boolean isErrorMessage)
    {
        if (line.startsWith(UNKNOWN))
        {
            processUnknownFile(line, UNKNOWN);
        }
        else if (line.startsWith(TO_ADD))
        {
            processUnknownFile(line, TO_ADD);
        }
        else if (line.startsWith(EXAM_DIR))
        {
            return;
        }
        else if (line.startsWith(WARNING))
        {
            if (line.indexOf(PERTINENT) > 0)
            {
                String filename = line.substring(WARNING.length(),
                                                 line.indexOf(PERTINENT)).trim();
                processNotPertinent(filename);
            }
            return;
        }
        else
        {
            // otherwise
            if (line.length() > 2) 
            {
                String firstChar = line.substring(0, 2);
                if (STATES.indexOf(firstChar) >= 0)
                {
                    processFile(line);
                    return;
                }
            }
        }

//        processLog(line);
    }

    private File createFile(String fileName)
    {
        return new File(localPath, fileName);
    }

    private void ensureExistingFileInfoContainer()
    {
        if (fileInfoContainer != null)
        {
            return;
        }
        fileInfoContainer = new DefaultFileInfoContainer();
    }

    private void processUnknownFile(String line, String beforeFile)
    {
        fileInfoContainer = new DefaultFileInfoContainer();
        fileInfoContainer.setType(beforeFile);
        String fileName = (line.substring(beforeFile.length())).trim();
        fileInfoContainer.setFile(createFile(fileName));
        eventManager.fireCVSEvent(new FileInfoEvent(this, fileInfoContainer));
        fileInfoContainer = null;
    }

    private void processFile(String line)
    {
        ensureExistingFileInfoContainer();

        fileInfoContainer.setType(line.charAt(0));
        String fileName = line.substring(2).trim();
        if (fileName.startsWith("no file"))
        {
            fileName = fileName.substring(8);
        }
        if (fileName.startsWith("./")) {
            fileName = fileName.substring(2);        
        }
        fileInfoContainer.setFile(createFile(fileName));

        eventManager.fireCVSEvent(new FileInfoEvent(this, fileInfoContainer));

        fileInfoContainer = null;
    }

    private void processLog(String line)
    {
        ensureExistingFileInfoContainer();

        //fileInfoContainer.addLog(line);
    }

    private void processNotPertinent(String fileName)
    {
        File fileToDelete = createFile(fileName);

        ensureExistingFileInfoContainer();

        // HACK - will create a non-cvs status in order to be able to have consistent info format
        fileInfoContainer.setType('D');
        fileInfoContainer.setFile(fileToDelete);
        //fileInfoContainer.addLog(WARNING);
        //fileInfoContainer.addLog(fileName);
        //fileInfoContainer.addLog(" ");
        //fileInfoContainer.addLog(PERTINENT);

        eventManager.fireCVSEvent(new FileInfoEvent(this, fileInfoContainer));

        fileInfoContainer = null;
    }

}

