/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.command.update;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.ListIterator;

import org.netbeans.lib.cvsclient.ClientServices;
import org.netbeans.lib.cvsclient.admin.Entry;
import org.netbeans.lib.cvsclient.command.BasicCommand;
import org.netbeans.lib.cvsclient.command.Builder;
import org.netbeans.lib.cvsclient.command.CommandException;
import org.netbeans.lib.cvsclient.event.EventManager;
import org.netbeans.lib.cvsclient.event.MessageEvent;
import org.netbeans.lib.cvsclient.request.ArgumentRequest;
import org.netbeans.lib.cvsclient.request.DirectoryRequest;
import org.netbeans.lib.cvsclient.request.UpdateRequest;
import org.netbeans.lib.cvsclient.request.EntryRequest;
import org.netbeans.lib.cvsclient.request.UnchangedRequest;
import org.netbeans.lib.cvsclient.admin.Entry;

import org.netbeans.lib.cvsclient.command.checkout.PipedFilesBuilder;
import org.netbeans.lib.cvsclient.command.checkout.TemporaryFileCreator;

/**
 * The Update command. Updates files that have previously been checked out
 * from the repository with the checkout command. Modified files are not
 * overwritten.
 * @author  Robert Greig
 * @version $Revision: 1.27.2.2 $
 */
public class UpdateCommand extends BasicCommand
    implements TemporaryFileCreator
{
    /**
     * A store of potentially empty directories. When a directory has a file
     * in it, it is removed from this set. This set allows the prune option
     * to be implemented
     */
    private HashSet emptyDirectories;

    /**
     * Whether to build directories, like checkout does (this is the -d option
     * in command-line CVS)
     */
    private boolean buildDirectories;

    /**
     * Determines whether to get a clean copy from the server.
     * This overrides even locally modified files.
     */
    private boolean cleanCopy;

    /**
     * Whether to prune directories, i.e. remove any directories that do not
     * contain any files. This is the -P option in command-line CVS)
     */
    private boolean pruneDirectories;

    /**
     * Determines wheather the output of the command is processed on standard output.
     *  Default is false. If true, nothing is done to local working files.
     */
    private boolean pipeToOutput;

    /**
     * Resets any sticky tags/dates/options imposed on the updated file(s).
     */
    private boolean resetStickyOnes;

    /**
     * Use head revision if a revision meeting criteria set by switches -r/-D
     * (tag/date) is not found.
     */
    private boolean useHeadIfNotFound;

    /**
     * equals the -D switch of command line cvs.
     */
    private String updateByDate;

    /**
     * Equals the -r switch of command-line cvs.
     */
    private String updateByRevision;

    /**
     * Use this keyword substitution for the command.
     * does not include the -k switch part.
     */
    private String keywordSubst;

    /** first of the 2 possible -j switches that merge 2 different revisions.
     *  If only this property is set, the current working file is merged
     * with the specified one.
     */
    private String mergeRevision1;

    /** Second of the 2 possible -j switches that merge 2 different revisions.
     *  Assumes the first -j switch (mergeRevision1 property is set).
     * Then the update commands merges the sources of these 2 revisons specified
     * by the -j switches.
     */
    private String mergeRevision2;

    /**
     * Construct a new update command
     */
    public UpdateCommand()
    {
        emptyDirectories = new HashSet();
    }

    /**
     * Method that is called while the command is being executed.
     * Descendants can override this method to return a Builder instance
     * that will parse the server's output and create data structures.
     */
    protected Builder createBuilder(EventManager eventMan)
    {
        if (isPipeToOutput())
        {
            return new PipedFilesBuilder(eventMan, this, this);
        }
        return new UpdateBuilder(eventMan, getLocalPath());
    }

    /**
     * If <code>getCleanCopy()</code> returns true, the files will be treated
     * as not existing.
     */
    protected void sendEntryAndModifiedRequests(Entry entry, File file)
    {
        if (getCleanCopy())
        {
            file = null;
        }
        super.sendEntryAndModifiedRequests(entry, file);
    }

    /**
     * Set whether to build directories. This is the -d option in command-line
     * CVS.
     */
    public void setBuildDirectories(boolean buildDirectories)
    {
        this.buildDirectories = buildDirectories;
    }

    /**
     * Get whether to build directories.
     * @return true if directories are to be built, false otherwise
     */
    public boolean getBuildDirectories()
    {
        return buildDirectories;
    }

    /**
     * Sets whether to get a clean copy from the server.
     * Even locally modified files will not merged but overridden.
     * This is the -C option in the command-line CVS.
     */
    public void setCleanCopy(boolean cleanCopy)
    {
        this.cleanCopy = cleanCopy;
    }

    /**
     * Returns whether to get a clean copy from the server.
     */
    public boolean getCleanCopy()
    {
        return cleanCopy;
    }

    /**
     * Set whether to prune directories. This is the -P option in the command-
     * line CVS.
     */
    public void setPruneDirectories(boolean pruneDirectories)
    {
        this.pruneDirectories = pruneDirectories;
    }

    /**
     * Get whether to prune directories.
     * @return true if directories should be removed if they contain no files,
     * false otherwise.
     */
    public boolean getPruneDirectories()
    {
        return pruneDirectories;
    }

    /**
     * Execute the command.
     * @param client the client services object that provides any necessary
     *               services to this command, including the ability to actually
     *               process all the requests
     */
    public void execute(ClientServices client, EventManager em)
        throws CommandException
    {
        super.execute(client, em);

        try
        {
            //requests.add(new ArgumentRequest("-u"));
            // now add the request that indicates the working directory for the
            // command
            if (buildDirectories)
            {
                requests.add(1, new ArgumentRequest("-d"));
            }
            if (cleanCopy)
            {
                requests.add(1, new ArgumentRequest("-C"));
            }
            if (pipeToOutput)
            {
                requests.add(1, new ArgumentRequest("-p"));
            }
            if (resetStickyOnes)
            {
                requests.add(1, new ArgumentRequest("-A"));
            }
            if (useHeadIfNotFound)
            {
                requests.add(1, new ArgumentRequest("-f"));
            }
            if (updateByDate != null && (!updateByDate.equals("")))
            {
                requests.add(1, new ArgumentRequest("-D"));
                requests.add(2, new ArgumentRequest(getUpdateByDate()));
            }
            if (updateByRevision != null && (!updateByRevision.equals("")))
            {
                requests.add(1, new ArgumentRequest("-r"));
                requests.add(2, new ArgumentRequest(getUpdateByRevision()));
            }
            if (mergeRevision1 != null && (!mergeRevision1.equals("")))
            {
                requests.add(1, new ArgumentRequest("-j"));
                requests.add(2, new ArgumentRequest(getMergeRevision1()));
            }
            if (mergeRevision2 != null && (!mergeRevision2.equals("")))
            {
                requests.add(1, new ArgumentRequest("-j"));
                requests.add(2, new ArgumentRequest(getMergeRevision2()));
            }
            if (getKeywordSubst() != null && !getKeywordSubst().equals(""))
            {
                requests.add(1, new ArgumentRequest("-k" + getKeywordSubst()));
            }

            requests.add(new DirectoryRequest(".",
                                  client.getRepositoryForDirectory(getLocalPath())));
            addArgumentRequests();
            requests.add(new UpdateRequest());
                 
            // hack - now check for the entry request for removed files 
            // only when p with -r or -D is on
            if (isPipeToOutput() && (getUpdateByRevision() != null || getUpdateByDate() != null))
            {
                ListIterator it = requests.listIterator();
                while (it.hasNext()) 
                {
                    Object req = it.next();
                    if (req instanceof EntryRequest) {
                        EntryRequest eReq = (EntryRequest)req;
                        Entry entry = eReq.getEntry();
                        if (entry.getRevision().startsWith("-")) {
                            entry.setRevision(entry.getRevision().substring(1));
                        }
                        it.set(new EntryRequest(entry));
                        it.add(new UnchangedRequest(entry.getName()));
                    }
                }
            }
            // end of hack..
            client.processRequests(requests);
            if (pruneDirectories)
            {
                pruneEmptyDirectories(client);
            }
            requests.clear();
        }
        catch (Exception e)
        {
            requests.clear();
            throw new CommandException(e);
        }
    }

    /**
     * Getter for property pipeToOutput.
     * @return Value of property pipeToOutput.
     */
    public boolean isPipeToOutput()
    {
        return pipeToOutput;
    }

    /**
     * Setter for property pipeToOutput.
     * @param pipeToOutput New value of property pipeToOutput.
     */
    public void setPipeToOutput(boolean pipeToOutput)
    {
        this.pipeToOutput = pipeToOutput;
    }

    /**
     * Getter for property resetStickyOnes.
     * @return Value of property resetStickyOnes.
     */
    public boolean isResetStickyOnes()
    {
        return resetStickyOnes;
    }

    /**
     * Setter for property resetStickyOnes.
     * @param resetStickyOnes New value of property resetStickyOnes.
     */
    public void setResetStickyOnes(boolean resetStickyOnes)
    {
        this.resetStickyOnes = resetStickyOnes;
    }

    /**
     * Getter for property useHeadIfNotFound.
     * @return Value of property useHeadIfNotFound.
     */
    public boolean isUseHeadIfNotFound()
    {
        return useHeadIfNotFound;
    }

    /**
     * Setter for property useHeadIfNotFound.
     * @param useHeadIfNotFound New value of property useHeadIfNotFound.
     */
    public void setUseHeadIfNotFound(boolean useHeadIfNotFound)
    {
        this.useHeadIfNotFound = useHeadIfNotFound;
    }

    /**
     * Getter for property updateByDate.
     * @return Value of property updateByDate.
     */
    public String getUpdateByDate()
    {
        return updateByDate;
    }

    /**
     * Setter for property updateByDate.
     * @param updateByDate New value of property updateByDate.
     */
    public void setUpdateByDate(String updateByDate)
    {
        this.updateByDate = updateByDate;
    }

    /**
     * Getter for property updateByRevision.
     * @return Value of property updateByRevision.
     */
    public String getUpdateByRevision()
    {
        return updateByRevision;
    }

    /**
     * Setter for property updateByRevision.
     * @param updateByRevision New value of property updateByRevision.
     */
    public void setUpdateByRevision(String updateByRevision)
    {
        this.updateByRevision = updateByRevision;
    }
    /**
     * Getter for property keywordSubst.
     * @return Value of property keywordSubst.
     */
    public String getKeywordSubst()
    {
        return keywordSubst;
    }

    /**
     * Setter for property keywordSubst.
     * @param keywordSubst New value of property keywordSubst.
     */
    public void setKeywordSubst(String keywordSubst)
    {
        this.keywordSubst = keywordSubst;
    }

    /**
     * Method that creates a temporary file.
     */
    public File createTempFile() throws java.io.IOException
    {
        File temp = File.createTempFile("cvs", ".dff");
        return temp;
    }

    /**
     * This method returns how the command would looklike when typed on the
     * command line.
     * Each command is responsible for constructing this information.
     * @returns <command's name> [<parameters>] files/dirs.
     * Example: checkout -p CvsCommand.java
     *
     */
    public String getCVSCommand()
    {
        StringBuffer toReturn = new StringBuffer("update ");
        toReturn.append(getCVSArguments());
        File[] files = getFiles();
        if (files != null)
        {
            for (int index = 0; index < files.length; index++)
            {
                toReturn.append(files[index].getName() + " ");
            }
        }
        return toReturn.toString();
    }

    /**
     * Returns the arguments of the command in the command-line style.
     * Similar to getCVSCommand() however without the files and command's name
     */
    public String getCVSArguments()
    {
        StringBuffer toReturn = new StringBuffer("");
        if (isPipeToOutput())
        {
            toReturn.append("-p ");
        }
        if (getCleanCopy())
        {
            toReturn.append("-C ");
        }
        if (!getRecursive())
        {
            toReturn.append("-l ");
        }
        if (getBuildDirectories())
        {
            toReturn.append("-d ");
        }
        if (getPruneDirectories())
        {
            toReturn.append("-P ");
        }
        if (isResetStickyOnes())
        {
            toReturn.append("-A ");
        }
        if (isUseHeadIfNotFound())
        {
            toReturn.append("-f ");
        }
        if (getKeywordSubst() != null && (!getKeywordSubst().equals("")))
        {
            toReturn.append("-k");
            toReturn.append(getKeywordSubst());
            toReturn.append(" ");
        }
        if (getUpdateByRevision() != null && (!getUpdateByRevision().equals("")))
        {
            toReturn.append("-r ");
            toReturn.append(getUpdateByRevision());
            toReturn.append(" ");
        }
        if (getUpdateByDate() != null && (!getUpdateByDate().equals("")))
        {
            toReturn.append("-D ");
            toReturn.append(getUpdateByDate());
            toReturn.append(" ");
        }
        if (getMergeRevision1() != null)
        {
            toReturn.append("-j ");
            toReturn.append(getMergeRevision1());
            toReturn.append(" ");
        }
        if (getMergeRevision2() != null)
        {
            toReturn.append("-j ");
            toReturn.append(getMergeRevision2());
            toReturn.append(" ");
        }
        return toReturn.toString();
    }

    /**
     * Takes the arguments and by parsing them, sets the command. To be mainly
     * used for automatic settings (like parsing the .cvsrc file)
     */
    public boolean setCVSCommand(char opt, String optArg)
    {
        if (opt == 'R')
        {
            setRecursive(true);
        }
        else if (opt == 'C')
        {
            setCleanCopy(true);
        }
        else if (opt == 'l')
        {
            setRecursive(false);
        }
        else if (opt == 'd')
        {
            setBuildDirectories(true);
        }
        else if (opt == 'P')
        {
            setPruneDirectories(true);
        }
        else if (opt == 'A')
        {
            setResetStickyOnes(true);
        }
        else if (opt == 'f')
        {
            setUseHeadIfNotFound(true);
        }
        else if (opt == 'D')
        {
            setUpdateByDate(optArg.trim());
        }
        else if (opt == 'r')
        {
            setUpdateByRevision(optArg.trim());
        }
        else if (opt == 'k')
        {
            setKeywordSubst(optArg);
        }
        else if (opt == 'p')
        {
            setPipeToOutput(true);
        }
        else if (opt == 'j')
        {
            if (getMergeRevision1() == null)
            {
                setMergeRevision1(optArg);
            }
            else
            {
                setMergeRevision2(optArg);
            }
        }
        else
        {
            // TODO - now silently ignores not recognized switches.
            return false;
        }
        return true;
    }

    /**
     * resets all switches in the command. After calling this method,
     * the command should have no switches defined and should behave defaultly.
     */
    public void resetCVSCommand()
    {
        setRecursive(true);
        setCleanCopy(false);
        setBuildDirectories(false);
        setPruneDirectories(false);
        setResetStickyOnes(false);
        setUseHeadIfNotFound(false);
        setUpdateByDate(null);
        setUpdateByRevision(null);
        setKeywordSubst(null);
        setPipeToOutput(false);
        setMergeRevision1(null);
        setMergeRevision2(null);
    }

    /**
     * Called when the server wants to send a message to be displayed to
     * the user. The message is only for information purposes and clients
     * can choose to ignore these messages if they wish.
     * @param e the event
     */
    public void messageSent(MessageEvent e)
    {
        super.messageSent(e);
        // we use this event to determine which directories need to be checked
        // for updating
        if (pruneDirectories &&
            e.getMessage().startsWith("cvs server: Updating "))
        {
            String relativePath = e.getMessage().substring(21);
            // dont delete the current directory, even if it's empty
            if (relativePath.equals("."))
            {
                return;
            }

            File file = new File(getLocalPath(), relativePath);
            emptyDirectories.add(file);
        }
    }

    /**
     * Prunes a directory, recursively pruning its subdirectories
     * @param directory the directory to prune
     */
    private boolean pruneEmptyDirectory(File directory, ClientServices client)
        throws IOException
    {
        boolean empty = true;

        final File[] contents = directory.listFiles();

        // should never be null, but just in case...
        if (contents != null)
        {
            for (int i = 0; i < contents.length; i++)
            {
                if (contents[i].isFile())
                {
                    empty = false;
                }
                else
                {
                    if (!contents[i].getName().equals("CVS"))
                    {
                        empty = pruneEmptyDirectory(contents[i], client);
                    }
                }
                if (!empty)
                {
                    break;
                }
            }

            if (empty)
            {
                // check this is a CVS directory and not some directory the user
                // has stupidly called CVS...
                final File entriesFile = new File(directory, "CVS/Entries");
                if (entriesFile.exists())
                {
                    final File adminDir = new File(directory, "CVS");
                    final File[] adminFiles = adminDir.listFiles();
                    for (int i=0; i < adminFiles.length; i++)
                    {
                        adminFiles[i].delete();
                    }
                    adminDir.delete();
                    directory.delete();
                    client.removeEntry(directory);
                }
            }
        }

        return empty;
    }

    /**
     * Remove any directories that don't contain any files
     */
    private void pruneEmptyDirectories(ClientServices client)
        throws IOException
    {
        final Iterator it = emptyDirectories.iterator();
        while (it.hasNext())
        {
            final File dir = (File) it.next();
            // we might have deleted it already (due to recursive delete)
            // so we need to check existence
            if (dir.exists())
            {
                pruneEmptyDirectory(dir, client);
            }
        }
        emptyDirectories.clear();
    }

    /**
     * String returned by this method defines which options are available for this particular command
     */
    public String getOptString()
    {
        return "RCldPAfD:r:pj:";
    }

    /** Getter for property mergeRevision1.
     * @return Value of property mergeRevision1.
     */
    public String getMergeRevision1()
    {
        return mergeRevision1;
    }

    /** Setter for property mergeRevision1.
     * @param mergeRevision1 New value of property mergeRevision1.
     */
    public void setMergeRevision1(String mergeRevision1)
    {
        this.mergeRevision1 = mergeRevision1;
    }

    /** Getter for property mergeRevision2.
     * @return Value of property mergeRevision2.
     */
    public String getMergeRevision2()
    {
        return mergeRevision2;
    }

    /** Setter for property mergeRevision2.
     * @param mergeRevision2 New value of property mergeRevision2.
     */
    public void setMergeRevision2(String mergeRevision2)
    {
        this.mergeRevision2 = mergeRevision2;
    }

}
