/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version 
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.commandLine;

import org.netbeans.lib.cvsclient.*;

import org.netbeans.lib.cvsclient.admin.*;

import org.netbeans.lib.cvsclient.connection.*;

import org.netbeans.lib.cvsclient.command.*;

import org.netbeans.lib.cvsclient.file.*;

import java.io.*;

/**
 * An implementation of the standard CVS client utility (command line tool)
 * in Java
 * @author  Robert Greig
 * @version $Revision: 1.7 $
 */
public class CVSCommand
{
    /**
     * The path to the repository on the server
     */
    private String repository;
    
    /**
     * The local path to use to perform operations (the top level)
     */
    private String localPath;
    
    /**
     * The connection to the server
     */
    private Connection connection;
    
    /**
     * The client that manages interactions with the server
     */
    private Client client;
    
    /**
     * The global options being used. GlobalOptions are only global for a
     * particular command.
     */
    private GlobalOptions globalOptions;
    
    /**
     * A struct containing the various bits of information in a CVS root
     * string, allowing easy retrieval of individual items of information
     */
    private static class CVSRoot
    {
        public String connectionType;
        public String user;
        public String host;
        public String repository;
        
        public CVSRoot(String root) throws IllegalArgumentException
        {
            if (!root.startsWith(":"))
                throw new IllegalArgumentException();
            
            int oldColonPosition = 0;
            int colonPosition = root.indexOf(':', 1);
            if (colonPosition==-1)
                throw new IllegalArgumentException();
            connectionType = root.substring(oldColonPosition + 1, colonPosition);
            oldColonPosition = colonPosition;
            colonPosition = root.indexOf('@', colonPosition+1);
            if (colonPosition==-1)
                throw new IllegalArgumentException();
            user = root.substring(oldColonPosition+1, colonPosition);
            oldColonPosition = colonPosition;
            colonPosition = root.indexOf(':', colonPosition+1);
            if (colonPosition==-1)
                throw new IllegalArgumentException();
            host = root.substring(oldColonPosition+1, colonPosition);
            repository = root.substring(colonPosition+1);
            if (connectionType==null || user==null || host==null ||
                repository==null)
                throw new IllegalArgumentException();
        }
    }
    
    /**
     * Execute a configured CVS command
     * @param command the command to execute
     * @throws CommandException if there is an error running the command   
     */
    public void executeCommand(Command command)
        throws CommandException
    {
        client.executeCommand(command, globalOptions);
    }        
    
    public void setRepository(String theRepository)
    {
        repository = theRepository;
    }
    
    public void setLocalPath(String thePath)
    {
        localPath = thePath;
    }
    
    public void setGlobalOptions(GlobalOptions go)
    {
        globalOptions = go;
    }
    
    /**
     * This handles the "server" connection
     */
    private void connectToServer() throws AuthenticationException
    {
        ServerConnection c = new ServerConnection();
        connection = c;        
        c.setRepository(repository);        
        c.open();                
        
        client = new Client(c, new StandardAdminHandler());        
                
        client.setLocalPath(localPath);
        
        // add a listener to the client
        client.getEventManager().addCVSListener(new BasicListener());
    }
    
    /**
     * This handles "pserver" connection
     */
    private void connectToServer(String userName, String encodedPassword,
                                 String hostName)
            throws AuthenticationException
    {        
        PServerConnection c = new PServerConnection();
        connection = c;
        c.setUserName(userName);
        c.setEncodedPassword(encodedPassword);
        c.setHostName(hostName);
        c.setRepository(repository);        
        c.open();        
        
        client = new Client(c, new StandardAdminHandler());     
               
        client.setLocalPath(localPath);
        
        // add a listener to the client
        client.getEventManager().addCVSListener(new BasicListener());
    }
    
    private void close()
    {
        try
        {
            connection.close();
        }
        catch (IOException e)
        {
            System.err.println("Unable to close connection: " + e);
            e.printStackTrace();
        }
    }
    
    /**
     * Obtain the CVS root, either from the -D option cvs.root or from
     * the CVS directory
     * @return the CVSRoot string
     */
    private static String getCVSRoot()
    {
        String root = null;
        BufferedReader r = null;
        try
        {
            File f = new File(System.getProperty("user.dir"));
            File rootFile = new File(f, "CVS/Root");            
            if (rootFile.exists())
            {
                r = new BufferedReader(new FileReader(rootFile));
                root = r.readLine();                
            }
        }
        catch (IOException e)
        {
            // ignore
        }
        finally
        {
            try
            {
                if (r != null)
                    r.close();
            }
            catch (IOException e)
            {
                System.err.println("Warning: could not close CVS/Root file!");
            }
        }
        if (root==null)
        {
            root = System.getProperty("cvs.root");
        }
        return root;        
    }
    
    /**
     * Process global options passed into the application
     * @param args the argument list, complete
     * @param globalOptions the global options structure that will be passed
     * to the command
     */
    private static int processGlobalOptions(String[] args, 
                                            GlobalOptions globalOptions)
    {
        final String getOptString = "Hrwd:z:";
        GetOpt go = new GetOpt(args, getOptString);
        int ch = -1;
        boolean usagePrint = false;
        while ((ch = go.getopt()) != go.optEOF)
        {
            if ((char)ch == 'H')
                usagePrint = true;
            else if ((char)ch == 'r')
                globalOptions.setCheckedOutFilesReadOnly(true);
            else if ((char)ch == 'w')
                globalOptions.setCheckedOutFilesReadOnly(false);
            else if ((char)ch == 'd')
                globalOptions.setCVSRoot(go.optArgGet());            
            else if ((char)ch == 'z')
            {
                globalOptions.setUseGzip(true);
                // we just ignore the zip level but include it for
                // completeness
                Object zipLevel = go.optArgGet();
            }
            else
                usagePrint = true;
        }
        if (usagePrint)
        {
            showUsage();
            System.exit(0);
        }
        return go.optIndexGet();
    }
    
    private static void showUsage()
    {
        System.err.println("Usage: cvs [global options] command [options]");
    }
    
    /**
     * Perform the 'login' command, asking the user for a password. If the
     * login is successful, the password is written to a file. The file's
     * location is user.home, unless the cvs.passfile option is set.
     * @param userName the userName
     * @param hostName the host
     */
    private static void performLogin(String userName, String hostName,
                                     String repository, 
                                     GlobalOptions globalOptions)
    {
        PServerConnection c = new PServerConnection();        
        c.setUserName(userName);
        String password = null;
        try
        {
            BufferedReader in  = new BufferedReader(new 
                                                InputStreamReader(System.in)); 
            System.out.print("Enter password: ");        
            password = in.readLine();
        }
        catch (IOException e)
        {
            System.err.println("Could not read password: " + e);
            System.exit(1);
        }
            
        String encodedPassword = StandardScrambler.getInstance().scramble(
                                                            password);
        c.setEncodedPassword(encodedPassword);
        c.setHostName(hostName);
        c.setRepository(repository);        
        try
        {
            c.verify();
        }
        catch (AuthenticationException e)
        {
            System.err.println("Could not login to host " + hostName);
            System.exit(1);
        }
        // was successful, so write the appropriate file out
        // we look for cvs.passfile being set, but if not use user.dir
        // as the default
        File passFile = new File(System.getProperty("cvs.passfile",
                                    System.getProperty("user.home") + 
                                    "/.cvspass"));
        BufferedWriter writer = null;
        BufferedReader reader = null;
        try
        {
            if (passFile.createNewFile())
            {
                writer = new BufferedWriter(new FileWriter(passFile));
                writer.write(globalOptions.getCVSRoot() + " " + 
                             encodedPassword);
                writer.close();
            }
            else
            {
                File tempFile = File.createTempFile("cvs", "tmp");
                reader = new BufferedReader(new FileReader(passFile));
                writer = new BufferedWriter(new FileWriter(tempFile));
                String line;
                while ((line = reader.readLine()) != null)
                {
                    if (line.startsWith(globalOptions.getCVSRoot()))
                    {
                        writer.write(globalOptions.getCVSRoot() + " " + 
                             encodedPassword);
                    }
                    else
                    {
                        writer.write(line);
                    }
                }
                reader.close();
                writer.close();
                File temp2File = File.createTempFile("cvs", "tmp");
                passFile.renameTo(temp2File);
                tempFile.renameTo(passFile);
                temp2File.delete();
                tempFile.delete();
            }
        }
        catch (IOException e)
        {
            System.err.println("Error: could not write password file to " +
                               passFile);
            System.exit(1);
        }
        finally
        {
            try
            {
                if (writer != null)
                {
                    writer.close();
                }
                if (reader != null)
                {
                    reader.close();
                }
            }
            catch (Exception e)
            {
                // ignore
            }
        }
        
        System.err.println("Logged in successfully to repository " +
                           repository + " on host " + hostName);
    }
    
    /**
     * Lookup the password in the .cvspass file. This file is looked for
     * in the user.home directory if the option cvs.passfile is not set
     * @param CVSRoot the CVS root for which the password is being searched
     * @return the password, scrambled
     */
    private static String lookupPassword(String CVSRoot)
    {
        File passFile = new File(System.getProperty("cvs.passfile",
                                 System.getProperty("user.home") + 
                                 "/.cvspass"));
        
        BufferedReader reader = null;
        String password = null;
        
        try
        {
            reader = new BufferedReader(new FileReader(passFile));
            String line;
            while ((line = reader.readLine()) != null)
            {
                if (line.startsWith(CVSRoot))
                {
                    password = line.substring(CVSRoot.length()+1);
                    break;
                }
            }
        }
        catch (IOException e)
        {
            System.err.println("Could not read password for host: " + e);
            System.exit(1);
        }
        finally
        {
            if (reader != null)
            {
                try
                {
                    reader.close();
                }
                catch (IOException e)
                {
                    System.err.println("Warning: could not close password file.");
                }
            }
        }
        return password;
    }
    
    public static void main(String[] args)
    {        
        // Set up the CVSRoot. Note that it might still be null after this
        // call if the user has decided to set it with the -d command line
        // global option
        GlobalOptions globalOptions = new GlobalOptions();
        globalOptions.setCVSRoot(getCVSRoot());
        
        // Set up any global options specified. These occur before the
        // name of the command to run
        int commandIndex = -1;        
        try
        {
            commandIndex = processGlobalOptions(args, globalOptions);
        }
        catch (IllegalArgumentException e)
        {
            System.err.println("Invalid argument: " + e);
            System.exit(1);
        }
        
        // if we don't have a CVS root by now, the user has messed up
        if (globalOptions.getCVSRoot()==null)
        {
            System.err.println("No CVS root is set. Use the cvs.root " +
                "property, e.g. java -Dcvs.root=\":pserver:user@host:/usr/cvs\"" +
                " or start the application in a directory containing a CVS subdirectory" +
                " or use the -d command switch.");
            System.exit(1);
        }
        
        // parse the CVS root into its constituent parts
        CVSRoot root = null;
        final String cvsRoot = globalOptions.getCVSRoot();
        try
        {
            root = new CVSRoot(cvsRoot);
        }
        catch (IllegalArgumentException e)
        {
            System.err.println("Incorrect format for CVSRoot: " + cvsRoot + 
                "\nThe correct format is: :<connection-type>:user@host:<repository path>" +
                "\nwhere <connection-type> is pserver and <repository path> is the " +
                "path to the cvs repository on the server.");
            System.exit(1);
        }
        
        // if we had some options without any command, then the user messed up
        if (commandIndex >= args.length)
        {
            showUsage();
            System.exit(1);
        }
        
        final String command = args[commandIndex];                
        if (command.equals("login"))
        {
            if (root.connectionType.equals("pserver"))
            {
                performLogin(root.user, root.host, root.repository,
                             globalOptions);
                System.exit(0);
            }
            else
            {
                System.err.println("login does not apply for connection type "+
                                   "\'" + root.connectionType + "\'");
                System.exit(1);
            }
        }                                
        
        // this is not login, but a 'real' cvs command, so construct it, 
        // set the options, and then connect to the server and execute it
        
        Command c = null;
        try
        {            
            c = CommandFactory.getCommand(command, args, ++commandIndex);
        }
        catch (IllegalArgumentException e)
        {
            System.err.println("Illegal argument: " + e.getMessage());            
            System.exit(1);
        }
       
        String password = null;
        
        if (root.connectionType.equals("pserver"))
            password = lookupPassword(cvsRoot);
        CVSCommand cvsCommand = new CVSCommand();
        cvsCommand.setGlobalOptions(globalOptions);
        cvsCommand.setRepository(root.repository);
        // the local path is just the path where we executed the
        // command. This is the case for command-line CVS but not
        // usually for GUI front-ends
        cvsCommand.setLocalPath(System.getProperty("user.dir"));
        try
        {
            if (root.connectionType.equals("pserver"))
                cvsCommand.connectToServer(root.user, password, root.host);
            else
                cvsCommand.connectToServer();
            cvsCommand.executeCommand(c);
        }
        catch (Throwable t)
        {
            System.err.println("Error: " + t);
            t.printStackTrace();
            System.exit(1);
        }
        finally
        {
            if (cvsCommand != null)
            {
                try
                {
                    cvsCommand.close();
                }
                catch (Throwable t)
                {
                }
            }
        }        
        System.exit(0);
    }
}
