/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.connection;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

import java.net.Socket;

import org.netbeans.lib.cvsclient.util.LoggedDataInputStream;
import org.netbeans.lib.cvsclient.util.LoggedDataOutputStream;

/**
 * Implements a connection to a pserver. See the cvs documents for more
 * information about different connection methods. PServer is popular where
 * security is not an issue. For secure connections, consider using a
 * kserver (Kerberos) or the GSSAPI.
 * @author  Robert Greig
 * @version $Revision: 1.7.2.3 $
 */
public class PServerConnection implements Connection
{
    /**
     * The string that is sent at the beginning of the request to open
     * a connection
     */
    protected  static final String OPEN_PREAMBLE = "BEGIN AUTH REQUEST\n";

    /**
     * The string that is sent at the end of the request to open a connection
     */
    protected static final String OPEN_POSTAMBLE = "END AUTH REQUEST\n";

    /**
     * The string that is sent at the beginning of the request to
     * verify a connection. Note the difference between opening a
     * connection and simply verifying.
     */
    protected static final String VERIFY_PREAMBLE =
                                  "BEGIN VERIFICATION REQUEST\n";

    /**
     * The string that is sent at the end of a verify request
     */
    protected static final String VERIFY_POSTAMBLE =
                                  "END VERIFICATION REQUEST\n";

    /**
     * A response indicating that authorisation has succeeded
     */
    protected static final String AUTHENTICATION_SUCCEEDED_RESPONSE =
                                  "I LOVE YOU";

    /**
     * A response indicating that the authorisation has failed
     */
    protected static final String AUTHENTICATION_FAILED_RESPONSE =
                                  "I HATE YOU";

    /**
     * The path to the repository. This is relative to the server
     */
    protected String repository;

    /**
     * The user name to use
     */
    protected String userName;

    /**
     * The password, encoded appropriately
     */
    protected String encodedPassword;

    /**
     * The default port number to use
     */
    public static final int DEFAULT_PORT = 2401;

    /**
     * The port number to use
     */
    protected int port = DEFAULT_PORT;

    /**
     * The host to use
     */
    protected String hostName;

    /**
     * The socket used for the connection
     */
    protected Socket socket;

    /**
     * The socket's input stream
     */
    protected LoggedDataInputStream inputStream;

    /**
     * The socket's output stream
     */
    protected LoggedDataOutputStream outputStream;

    /**
     * Get an input stream for receiving data from the server
     * @return a data inputstream
     */
    public LoggedDataInputStream getInputStream()
    {
        return inputStream;
    }

    /**
     * Get an output stream for sending data to the server
     * @return an outputstream
     */
    public LoggedDataOutputStream getOutputStream()
    {
        return outputStream;
    }

    /**
     * Authenticate a connection with the server, using the specified
     * postamble and preamble
     * @param preamble the preamble to use
     * @param postamble the postamble to use
     * @throws AuthenticationException if an error occurred
     * @return the socket used to make the connection. The socket is
     * guaranteed to be open if an exception has not been thrown
     */
    private void openConnection(String preamble, String postamble)
        throws AuthenticationException
    {
        if (hostName==null)
        {
            throw new AuthenticationException("HostIsNull", 
                AuthenticationException.getBundleString("AuthenticationException.HostIsNull"));
        }
        try
        {
            socket = new Socket(hostName, port);
            BufferedOutputStream bos = new BufferedOutputStream(socket.getOutputStream(), 32768);
            outputStream = new LoggedDataOutputStream(bos);
            BufferedInputStream bis = new BufferedInputStream(socket.getInputStream(), 32768);
            inputStream = new LoggedDataInputStream(bis);

            outputStream.writeBytes(preamble);
            outputStream.writeBytes(repository + "\n");
            outputStream.writeBytes(userName + "\n");
            outputStream.writeBytes(encodedPassword + "\n");
            outputStream.writeBytes(postamble);
            outputStream.flush();
            String response = inputStream.readLine();
            if (!response.equals(AUTHENTICATION_SUCCEEDED_RESPONSE))
            {
                String locMessage = AuthenticationException.getBundleString("AuthenticationException.AuthenticationFailed");
                throw new AuthenticationException("AuthenticationFailed", locMessage);
            }
        }
        catch (AuthenticationException e)
        {
            throw e;
        }
        catch (java.net.ConnectException conExc) 
        {
            String locMessage = getLocalMessage("AuthenticationException.ConnectException", 
                                    new Object[] {hostName});
            throw new AuthenticationException("ConnectException", conExc,locMessage);
        }
        catch (java.net.NoRouteToHostException routeExc) 
        {
            String locMessage = getLocalMessage("AuthenticationException.NoRouteToHostException", 
                                    new Object[] {hostName});
            throw new AuthenticationException("NoRouteToHostException", routeExc,locMessage);
        }
        catch (java.io.IOException ioExc) 
        {
            String locMessage = getLocalMessage("AuthenticationException.IOException", 
                                    new Object[] {hostName});
            throw new AuthenticationException("IOException", ioExc,locMessage);
        }        
        catch (Throwable t)
        {
            String locMessage = AuthenticationException.getBundleString(
                                "AuthenticationException.Throwable");
            throw new AuthenticationException("General error", t, locMessage);
        }
    }

    /**
     * Authenticate with the server. Closes the connection immediately.
     * Clients can use this method to ensure that they are capable of
     * authenticating with the server. If no exception is thrown, you can
     * assume that authentication was successful
     * @throws AutenticationException if the connection with the server
     * cannot be established
     */
    public void verify() throws AuthenticationException
    {
        openConnection(VERIFY_PREAMBLE, VERIFY_POSTAMBLE);
            
        if (socket != null)
        {
            try
            {
                socket.close();
            }
            catch (Throwable t)
            {
               String locMessage = AuthenticationException.getBundleString(
                                "AuthenticationException.Throwable");
                throw new AuthenticationException("General error", t, locMessage);
            }
        }
    }

    /**
     * Authenticate with the server and open a channel of communication
     * with the server. This Client will
     * call this method before interacting with the server. It is up to
     * implementing classes to ensure that they are configured to
     * talk to the server (e.g. port number etc.)
     * @throws AutenticationException if the connection with the server
     * cannot be established
     */
    public void open() throws AuthenticationException
    {
        openConnection(OPEN_PREAMBLE, OPEN_POSTAMBLE);
    }

    /**
     * Get the repository path
     * @return the repository path, e.g. /home/banana/foo/cvs
     */
    public String getRepository()
    {
        return repository;
    }

    /**
     * Set the repository
     * @param root the repository
     */
    public void setRepository(String theRepository)
    {
        repository = theRepository;
    }

    /**
     * Get the username
     */
    public String getUserName()
    {
        return userName;
    }

    /**
     * Set the userName
     * @param name the userName
     */
    public void setUserName(String name)
    {
        userName = name;
    }

    /**
     * Get the encoded password
     * @return the encoded password
     */
    public String getEncodedPassword()
    {
        return encodedPassword;
    }

    /**
     * Set the encoded password
     * @param password the encoded password to use for authentication
     */
    public void setEncodedPassword(String password)
    {
       encodedPassword = password;
    }

    /**
     * Get the port number to use
     * @return the port number
     */
    public int getPort()
    {
        return port;
    }

    /**
     * Set the port number to use
     * @param thePort the port number to use. If you do not set this, 2401
     * is used by default for pserver.
     */
    public void setPort(int thePort)
    {
        port = thePort;
    }

    /**
     * Get the host name to use
     * @return the host name of the server to connect to. If you do not set
     * this, localhost is used by default for pserver.
     */
    public String getHostName()
    {
        return hostName;
    }

    /**
     * Get the host name to use
     * @param theHostName the host name of the server to connect to. If you
     * do not set this, localhost is used by default for pserver.
     */
    public void setHostName(String theHostName)
    {
        hostName = theHostName;
    }

    /**
     * Close the connection with the server
     */
    public void close() throws IOException
    {
        if (socket != null)
        {
            socket.close();
        }
    }

    /**
     * Modify the underlying inputstream
     * @param modifier the connection modifier that performs the modifications
     * @throws IOException if an error occurs modifying the streams
     */
    public void modifyInputStream(ConnectionModifier modifier)
        throws IOException {
        modifier.modifyInputStream(inputStream);
    }

    /**
     * Modify the underlying outputstream
     * @param modifier the connection modifier that performs the modifications
     * @throws IOException if an error occurs modifying the streams
     */
    public void modifyOutputStream(ConnectionModifier modifier)
        throws IOException {
        modifier.modifyOutputStream(outputStream);
    }
    
    private String getLocalMessage(String key, Object[] arguments) 
    {
        String locMessage = AuthenticationException.getBundleString(key);
        if (locMessage != null) 
        {
            locMessage = java.text.MessageFormat.format(locMessage, arguments);
        }
        return locMessage;
    }

}
