/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version 
 * 1.0 (the "License"). You may not use this file except in compliance with 
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.connection;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

import java.net.Socket;

import org.netbeans.lib.cvsclient.util.LoggedDataInputStream;
import org.netbeans.lib.cvsclient.util.LoggedDataOutputStream;

/**
 * Implements a connection to a server. See the cvs documents for more
 * information about different connection methods. Server is popular where
 * the CVS repository exists on the machine where the client library is
 * running.
 * @author  Robert Greig
 * @version $Revision: 1.4.2.1 $
 */
public class ServerConnection implements Connection
{    
    /**
     * The path to the repository. This is relative to the server
     */
    protected String repository;
    
    /**
     * The socket's input stream
     */
    protected LoggedDataInputStream inputStream;
    
    /**
     * The socket's output stream
     */
    protected LoggedDataOutputStream outputStream;
 
    /**
     * The CVS process that is being run
     */
    protected Process process;
    
    /** 
     * Get an input stream for receiving data from the server
     * @return a data inputstream
     */
    public LoggedDataInputStream getInputStream()
    {        
        return inputStream;
    }
    
    /**
     * Get an output stream for sending data to the server
     * @return an outputstream
     */
    public LoggedDataOutputStream getOutputStream()
    {        
        return outputStream;
    }
    
    /**
     * Authenticate a connection with the server, using the specified
     * postamble and preamble
     * @param preamble the preamble to use
     * @param postamble the postamble to use
     * @throws AuthenticationException if an error occurred
     * @return the socket used to make the connection. The socket is
     * guaranteed to be open if an exception has not been thrown
     */
    private void openConnection()
        throws AuthenticationException
    {        
        try
        {            
            process = Runtime.getRuntime().exec("cvs server");
            outputStream = new LoggedDataOutputStream(process.
                                                      getOutputStream());
            inputStream = new LoggedDataInputStream(process.getInputStream());            
        }        
        catch (Throwable t)
        {
            String locMessage = AuthenticationException.getBundleString(
                                "AuthenticationException.ServerConnection");
            throw new AuthenticationException("Connection error", t, locMessage);
        }
    }
    
    /**
     * Authenticate with the server. Closes the connection immediately.
     * Clients can use this method to ensure that they are capable of
     * authenticating with the server. If no exception is thrown, you can
     * assume that authentication was successful
     * @throws AutenticationException if the connection with the server
     * cannot be established
     */
    public void verify() throws AuthenticationException
    {
        try
        {
            process = Runtime.getRuntime().exec("cvs server");
            process.destroy();
        }
        catch (Exception e)
        {
            String locMessage = AuthenticationException.getBundleString(
                                "AuthenticationException.ServerVerification");
            throw new AuthenticationException("Verification error", e, locMessage);
        }
    }
    
    /**
     * Authenticate with the server and open a channel of communication
     * with the server. This Client will
     * call this method before interacting with the server. It is up to
     * implementing classes to ensure that they are configured to
     * talk to the server (e.g. port number etc.)
     * @throws AutenticationException if the connection with the server
     * cannot be established
     */
    public void open() throws AuthenticationException
    {
        openConnection();
    }
    
    /**
     * Get the repository path
     * @return the repository path, e.g. /home/banana/foo/cvs
     */
    public String getRepository()
    {
        return repository;
    }
    
    /**
     * Set the repository
     * @param root the repository
     */
    public void setRepository(String theRepository)
    {
        repository = theRepository;
    }
               
    /**
     * Close the connection with the server
     */
    public void close() throws IOException
    {
        if (process != null)
        {            
            process.destroy();
        }        
    }
    
    /**
     * Modify the underlying inputstream
     * @param modifier the connection modifier that performs the modifications
     * @throws IOException if an error occurs modifying the streams
     */
    public void modifyInputStream(ConnectionModifier modifier)
        throws IOException {
        modifier.modifyInputStream(inputStream);        
    }
    
    /**
     * Modify the underlying outputstream
     * @param modifier the connection modifier that performs the modifications
     * @throws IOException if an error occurs modifying the streams
     */
    public void modifyOutputStream(ConnectionModifier modifier)
        throws IOException {
        modifier.modifyOutputStream(outputStream);
    }
    
}