/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/

 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Robert Greig.
 * Portions created by Robert Greig are Copyright (C) 2000.
 * All Rights Reserved.

 * Contributor(s): Robert Greig.
 *****************************************************************************/
package org.netbeans.lib.cvsclient.file;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;

import java.util.Date;

import org.netbeans.lib.cvsclient.Client;

import org.netbeans.lib.cvsclient.request.Request;

import org.netbeans.lib.cvsclient.util.LoggedDataInputStream;
import org.netbeans.lib.cvsclient.util.LoggedDataOutputStream;

/**
 * Provides a basic implementation of FileHandler, and does much of the
 * handling of reading and writing files and performing CRLF conversions.
 * @author  Robert Greig
 * @version $Revision: 1.10.2.1 $
 */
public class AbstractFileHandler implements FileHandler
{
    /**
     * Whether to emit debug information.
     */
    private static final boolean DEBUG = false;

    /**
     * The size of chunks read from disk.
     */
    protected int CHUNK_SIZE = 32768;

    /**
     * The date the next file written should be marked as being modified on.
     */
    protected Date modifiedDate = null;

    /**
     * Get the string to transmit containing the file transmission length.
     *
     * @return a String to transmit to the server (including carriage return)
     * @param length the amount of data that will be sent
     */
    protected String getLengthString(long length)
    {
         return String.valueOf(length) + "\n";
    }

    protected Reader getProcessedReader(File f)
        throws IOException
    {
        return new FileReader(f);
    }

    protected InputStream getProcessedInputStream(File f)
        throws IOException
    {
        return new FileInputStream(f);
    }

    /**
     * Get any requests that must be sent before commands are sent, to init
     * this file handler.
     *
     * @return an array of Requests that must be sent
     */
    public Request[] getInitialisationRequests()
    {
        return null;
    }

    /**
     * Transmit a text file to the server, using the standard CVS protocol
     * conventions. CR/LFs are converted to the Unix format.
     *
     * @param file the file to transmit
     * @param dos the data outputstream on which to transmit the file
     */
    public void transmitTextFile(File file, LoggedDataOutputStream dos)
        throws IOException
    {
        if (file == null || !file.exists())
        {
            throw new IllegalArgumentException("File is either null or " +
                                            "does not exist. Cannot transmit.");
        }

        final BufferedReader reader = new BufferedReader(new FileReader(file));

        // must write file to temp location first because size might change
        // due to CR/LF changes
        File tempFile = File.createTempFile("cvs", null);
        
        try
        {
            BufferedWriter writer = new BufferedWriter(new FileWriter(tempFile));

            char[] fileChunk = new char[8192];
            char[] fileWriteChunk = new char[8192];

            int readLength;
            int writeLength;
            int index = 0;
            while ((readLength = reader.read(fileChunk)) != -1)
            {
                for (int i=0; i < readLength; i++)
                {
                    if (fileChunk[i]!='\r')
                    {
                        fileWriteChunk[index]= fileChunk[i];
                        index = index + 1;
                    }
                }
                writer.write(fileWriteChunk, 0, index);
                index = 0;
            }
            writer.close();
            reader.close();

            // first write the length of the file
            long length = tempFile.length();
            dos.writeBytes(getLengthString(length));

            BufferedInputStream bis =
                        new BufferedInputStream(new FileInputStream(tempFile));
            // now transmit the file itself
            byte[] chunk = new byte[CHUNK_SIZE];
            while (length > 0)
            {
                int bytesToRead = (length >= CHUNK_SIZE) ? CHUNK_SIZE
                                                         : (int)length;
                int count = bis.read(chunk, 0, bytesToRead);
                length -= count;
                dos.write(chunk, 0, count);
            }
            dos.flush();
            bis.close();
        }
        finally
        {
            tempFile.delete();
        }
    }

    /**
     * Transmit a binary file to the server, using the standard CVS protocol
     * conventions.
     * @param file the file to transmit
     * @param dos the data outputstream on which to transmit the file
     */
    public void transmitBinaryFile(File file, LoggedDataOutputStream dos)
        throws IOException
    {
        if (file == null || !file.exists())
        {
            throw new IllegalArgumentException("File is either null or " +
                                            "does not exist. Cannot transmit.");
        }

        BufferedInputStream bis = null;
        
        try
        {
            bis = new BufferedInputStream(new FileInputStream(file));
            // first write the length of the file
            long length = file.length();

            dos.writeBytes(getLengthString(length));

            // now transmit the file itself
            byte[] chunk = new byte[CHUNK_SIZE];
            while (length > 0)
            {
                int bytesToRead = (length >= CHUNK_SIZE) ? CHUNK_SIZE
                                                         : (int)length;
                int count = bis.read(chunk, 0, bytesToRead);
                length -= count;
                dos.write(chunk, 0, count);
            }
            dos.flush();
        }
        finally
        {
            if (bis != null)
            {
                try
                {
                    bis.close();
                }
                catch (IOException ex)
                {
                    ex.printStackTrace();
                }
            }
        }
    }

    /**
     * Write (either create or replace) a file on the local machine with
     * one read from the server.
     * @param path the absolute path of the file, (including the file name).
     * @param mode the mode of the file
     * @param dis the stream to read the file from, as bytes
     * @param length the number of bytes to read
     */
    public void writeTextFile(String path, String mode,
                              LoggedDataInputStream dis, int length)
        throws IOException
    {
        if (DEBUG)
        {
            System.err.println("[writeTextFile] writing: " + path);
            System.err.println("[writeTextFile] length: " + length);
            System.err.println("Reader object is: " + dis.hashCode());
        }
        File f = new File(path);
        f.getParentFile().mkdirs();
        f.createNewFile();
        // For CRLF conversion, we have to read the file
        // into a temp file, then do the conversion. This is because we cannot
        // perform a sequence of readLines() until we've read the file from
        // the server - the file transmission is not followed by a newline.
        // Bah.
        File tempFile = File.createTempFile("cvsCRLF", "tmp");
        
        try
        {
            BufferedOutputStream tempWriter =
                      new BufferedOutputStream(new FileOutputStream(tempFile));

            byte[] chunk = new byte[CHUNK_SIZE];
            while (length > 0)
            {
                int count = (length >= CHUNK_SIZE) ? CHUNK_SIZE
                                                   :length;
                count = dis.read(chunk, 0, count);
                length -= count;
                if (DEBUG)
                {
                    System.err.println("Still got: " + length + " to read");
                }
                tempWriter.write(chunk, 0, count);
            }
            tempWriter.close();

            // Here we read the temp file in again, doing any processing required
            // (for example, unzipping). We also take the opportunity to
            // convert the bytes to chars
            BufferedReader tempReader =
                              new BufferedReader(getProcessedReader(tempFile));
            BufferedWriter writer = new BufferedWriter(new FileWriter(f));
            
            // this chunk is directly read from the temp file
            char[] cchunk = new char[CHUNK_SIZE];            
            int readLength;
            while ((readLength = tempReader.read(cchunk)) != -1)
            {
                // we must perform our own newline conversion. The file will
                // definitely have unix style CRLF conventions, so if we have
                // a \n this code will write out a \n or \r\n as appropriate for
                // the platform we are running on
                for (int i=0; i < readLength; i++)
                {   
                    if (cchunk[i]=='\n')
                    {
                        writer.newLine();
                    }
                    else
                    {
                        writer.write(cchunk[i]);
                    }
                }                
            }
            tempReader.close();            
            writer.close();
            if (modifiedDate != null)
            {
                f.setLastModified(modifiedDate.getTime());
                modifiedDate = null;
            }
        }
        finally
        {
            tempFile.delete();
        }
    }

    /**
     * Write (either create or replace) a binary file on the local machine with
     * one read from the server.
     * @param path the absolute path of the file, (including the file name).
     * @param mode the mode of the file
     * @param dis the stream to read the file from, as bytes
     * @param length the number of bytes to read
     */
    public void writeBinaryFile(String path, String mode,
                                LoggedDataInputStream dis, int length)
        throws IOException
    {
        if (DEBUG)
        {
            System.err.println("[writeBinaryFile] writing: " + path);
            System.err.println("[writeBinaryFile] length: " + length);
            System.err.println("Reader object is: " + dis.hashCode());
        }
        File f = new File(path);
        f.getParentFile().mkdirs();
        f.createNewFile();

        // FUTURE: optimisation possible - no need to use a temp file if there
        // is no post processing required (e.g. unzipping). So perhaps enhance
        // the interface to allow this stage to be optional
        File tempFile = File.createTempFile("cvsPostConversion", "tmp");

        try
        {
            BufferedOutputStream bos =
                      new BufferedOutputStream(new FileOutputStream(tempFile));

            byte[] chunk = new byte[CHUNK_SIZE];
            while (length > 0)
            {
                int bytesToRead = (length >= CHUNK_SIZE) ? CHUNK_SIZE
                                                         : (int)length;
                int count = dis.read(chunk, 0, bytesToRead);
                if (count < 0)
                {
                    break;
                }

                length -= count;
                if (DEBUG)
                {
                    System.err.println("Still got: " + length + " to read");
                }
                bos.write(chunk, 0, count);
            }
            bos.close();

            // Here we read the temp file in, taking the opportunity to process
            // the file, e.g. unzip the data
            BufferedInputStream tempIS =
                     new BufferedInputStream(getProcessedInputStream(tempFile));
            bos = new BufferedOutputStream(new FileOutputStream(f));

            for (int count = tempIS.read(chunk, 0, CHUNK_SIZE);
                 count > 0;
                 count = tempIS.read(chunk, 0, CHUNK_SIZE))
            {
                bos.write(chunk, 0, count);
            }

            tempIS.close();

            bos.close();

            // now we need to modifiy the timestamp on the file, if specified
            if (modifiedDate != null)
            {
                f.setLastModified(modifiedDate.getTime());
                modifiedDate = null;
            }
        }
        finally
        {
            tempFile.delete();
        }
    }

    /**
     * Remove the specified file from the local disk.
     *
     * @param pathname the full path to the file to remove
     * @throws IOException if an IO error occurs while removing the file
     */
    public void removeLocalFile(String pathname)
        throws IOException
    {
        File fileToDelete = new File(pathname);
        if (!fileToDelete.delete())
        {
            System.err.println("Could not delete file " +
                               fileToDelete.getAbsolutePath());
        }
    }

    /**
     * Rename the local file.
     *
     * @param pathname the full path to the file to rename
     * @param newName the new name of the file (not the full path)
     * @throws IOException if an IO error occurs while renaming the file
     */
    public void renameLocalFile(String pathname, String newName)
        throws IOException
    {
        File sourceFile = new File(pathname);
        File destinationFile = new File(sourceFile.getParentFile(), newName);
        sourceFile.renameTo(destinationFile);
    }

    /**
     * Set the modified date of the next file to be written.
     * The next call to writeFile will use this date.
     *
     * @param modifiedDate the date the file should be marked as modified
     */
    public void setNextFileDate(Date modifiedDate)
    {        
        this.modifiedDate = modifiedDate;
    }
}
