/*****************************************************************************
 * Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License Version
 * 1.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is available at http://www.sun.com/
 *
 * The Original Code is the CVS Client Library.
 * The Initial Developer of the Original Code is Thomas Singer.
 * Portions created by Thomas Singer Copyright (C) 2001.
 * All Rights Reserved.
 *
 * Contributor(s): Thomas Singer, Milos Kleint
 *****************************************************************************/
package org.netbeans.lib.cvsclient.util;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * @author  Thomas Singer
 * @version $Date: 2001/06/21 14:41:34 $
 */
public class SimpleStringPattern
    implements StringPattern {

    private static final char MATCH_EACH = '*';
    private static final char MATCH_ONE  = '?';

    private final List subPatterns = new LinkedList();

    /**
     * Creates a SimpleStringPattern for the specified definition.
     * The definition might contain two special characters ('*' and '?').
     */
    public SimpleStringPattern(String definition) {
        splitInSubPattern(definition);
    }

    /**
     * Returns whether the specified string matches thiz pattern.
     */
    public boolean doesMatch(String string) {
        int index = 0;
        SubPattern subPattern = null;
        for (Iterator it = subPatterns.iterator(); it.hasNext(); ) {
            subPattern = (SubPattern)it.next();
            index = subPattern.doesMatch(string, index);
            if (index < 0) {
                return false;
            }
        }

        if (index == string.length()) {
            return true;
        }
        if (subPattern == null) {
            return false;
        }
        return subPattern.checkEnding(string, index);
    }

    private void splitInSubPattern(String definition) {
        char prevSubPattern = ' ';

        int prevIndex = 0;
        for (int index = 0; index >= 0; ) {
            prevIndex = index;

            index = definition.indexOf(MATCH_EACH, prevIndex);
            if (index >= 0) {
                String match = definition.substring(prevIndex, index);
                addSubPattern(match, prevSubPattern);
                prevSubPattern = MATCH_EACH;
                index++;
                continue;
            }
            index = definition.indexOf(MATCH_ONE, prevIndex);
            if (index >= 0) {
                String match = definition.substring(prevIndex, index);
                addSubPattern(match, prevSubPattern);
                prevSubPattern = MATCH_ONE;
                index++;
                continue;
            }
        }
        String match = definition.substring(prevIndex);
        addSubPattern(match, prevSubPattern);
    }

    private void addSubPattern(String match, char subPatternMode) {
        SubPattern subPattern = null;
        switch (subPatternMode) {
        case MATCH_EACH:
            subPattern = new MatchEachCharPattern(match);
            break;
        case MATCH_ONE:
            subPattern = new MatchOneCharPattern(match);
            break;
        default:
            subPattern = new MatchExactSubPattern(match);
            break;
        }

        subPatterns.add(subPattern);
    }

    public String toString() {
        StringBuffer buffer = new StringBuffer();
        for (Iterator it = subPatterns.iterator(); it.hasNext(); ) {
            SubPattern subPattern = (SubPattern)it.next();
            buffer.append(subPattern.toString());
        }
        return buffer.toString();
    }

    public static void main(String[] arguments) {
        StringPattern sp = new SimpleStringPattern("a*b");

        test(sp, "ab", true);
        test(sp, "aab", true);
        test(sp, "ba", false);
        test(sp, "abc", false);

        sp = new SimpleStringPattern("*.txt");

        test(sp, "datei.txt", true);
        test(sp, ".txt", true);
        test(sp, "datei.tx", false);
        test(sp, "datei.txt.txt", true);

        sp = new SimpleStringPattern("datei*1*");

        test(sp, "datei0.txt", false);
        test(sp, "datei1.txt", true);
        test(sp, "datei.tx", false);
        test(sp, "datei1.txt.txt", true);

        sp = new SimpleStringPattern("Makefile");

        test(sp, "Makefile", true);
        test(sp, "Makefile.mak", false);
        test(sp, "Makefile1", false);
        test(sp, ".Makefile", false);
        test(sp, ".Makefile.", false);

        sp = new SimpleStringPattern("*~");

        test(sp, "datei~", true);
        test(sp, "datei~1", false);
        test(sp, "datei~1~", true);
    }

    private static void test(StringPattern sp, String testString, boolean shouldResult) {
        System.err.print('"' + sp.toString() + '"' + ": " + testString + " " + shouldResult);

        boolean doesMatch = sp.doesMatch(testString);

        if (doesMatch == shouldResult) {
            System.err.println(" proved");
        }
        else {
            System.err.println(" **denied**");
        }
    }

    private static abstract class SubPattern {
        protected final String match;

        protected SubPattern(String match) {
            this.match = match;
        }

        /**
         * @parameter string ... the whole string to test for matching
         * @parameter index  ... the index in string where this' test should begin
         * @returns          ... if successful the next test-position, if not -1
         */
        public abstract int doesMatch(String string, int index);

        public boolean checkEnding(String string, int index) {
            return false;
        }
    }

    private static class MatchExactSubPattern extends SubPattern {
        public MatchExactSubPattern(String match) {
            super(match);
        }

        public int doesMatch(String string, int index) {
            if (!string.startsWith(match, index)) {
                return -1;
            }
            return index + match.length();
        }

        public String toString() {
            return match;
        }
    }

    private static class MatchEachCharPattern extends SubPattern {
        public MatchEachCharPattern(String match) {
            super(match);
        }

        public int doesMatch(String string, int index) {
            int matchIndex = string.indexOf(match, index);
            if (matchIndex < 0) {
                return -1;
            }
            return matchIndex + match.length();
        }

        public boolean checkEnding(String string, int index) {
            return string.endsWith(match);
        }

        public String toString() {
            return MATCH_EACH + match;
        }
    }

    private static class MatchOneCharPattern extends MatchExactSubPattern {
        public MatchOneCharPattern(String match) {
            super(match);
        }

        public int doesMatch(String string, int index) {
            index++;
            if (string.length() < index) {
                return -1;
            }
            return super.doesMatch(string, index);
        }

        public String toString() {
            return MATCH_ONE + match;
        }
    }
}
