/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.cvsclient;

/**
 *
 * @author  mk104111
 */
import org.openide.filesystems.*;
import org.netbeans.modules.cvsclient.commands.remove.*;
import org.netbeans.modules.cvsclient.commands.add.*;
import org.netbeans.modules.cvsclient.commands.*;
import org.netbeans.modules.cvsclient.caching.*;
import org.netbeans.modules.vcscore.commands.VcsCommand;
import org.netbeans.modules.vcscore.cache.*;
import org.netbeans.modules.vcscore.VcsAttributes;
import java.io.*;
import java.util.*;
import java.beans.*;
import org.openide.util.RequestProcessor;



public class JavaCvsAttributes extends DefaultAttributes {
    
    static final long serialVersionUID = 5320181565426569135L;

    private transient JavaCvsFileSystem fileSystem;
    /** Creates new VcsAttributes */
    public JavaCvsAttributes(AbstractFileSystem.Info info, AbstractFileSystem.Change change,
    AbstractFileSystem.List list, JavaCvsFileSystem fs) {
        super(info, change, list);
        fileSystem = fs;
    }

    
  /**
     * Get the file attribute with the specified name.
     * @param name the file name
     * @param attrName name of the attribute
     * @return appropriate (serializable) value or null if the attribute is unset
     *         (or could not be properly restored for some reason).
     *         If the attribute name is the {@link VCS_STATUS}, then the VCS status
     *         of the file is returned.
     */
    public Object readAttribute(String name, String attrName) {
        if (VcsAttributes.VCS_STATUS.equals(attrName)) {
            File toFind = fileSystem.getFile(name);
            if (!toFind.exists()) return VcsAttributes.VCS_STATUS_MISSING;
            JavaCvsCache statusProvider = (JavaCvsCache)CacheHandler.getInstance().getCache(JavaCvsCache.JAVA_CACHE_NAME);
            if (statusProvider != null) {
                CacheFile cacheFile = statusProvider.getCacheFile(toFind, CacheHandler.STRAT_DISK);
                if (cacheFile.getStatus().equals(JavaCvsStatusManager.LOCAL)) {
                    return VcsAttributes.VCS_STATUS_LOCAL;
                }
                return VcsAttributes.VCS_STATUS_UP_TO_DATE;
            }
            return VcsAttributes.VCS_STATUS_UNKNOWN;
        } else {
            return super.readAttribute(name, attrName);
        }
    }
    
    /**
     * Set the file attribute with the specified name. If the name is {@link VCS_ACTION},
     * and the value is an instance of FeatureDescriptor, then it's not set as file attribute,
     * but is interpreted as a VCS command. The name of the command is taken from
     * value.getName() and commands options from attributes of that feature descriptor.
     * @param name the file name
     * @param attrName name of the attribute
     * @param value new value or null to clear the attribute. Must be serializable,
     *        with the exception of VCS command attribute.
     * @throws IOException if the attribute cannot be set. If serialization is
     *                     used to store it, this may in fact be a subclass such
     *                     as NotSerializableException.
     * @throws UnknownServiceException if the requested VCS action is not provided.
     *                                 A subclass of IOException was chosen, since
     *                                 FileObject.setAttribute throws IOException.
     */
    public void writeAttribute(final String name, final String attrName, final Object value) throws IOException, java.net.UnknownServiceException {
       if (VcsAttributes.VCS_ACTION.equals(attrName) && value instanceof FeatureDescriptor) {
            performVcsAction(name, (FeatureDescriptor) value);
        } else if (VcsAttributes.VCS_SCHEDULING_SECONDARY_FO_ACTION.equals(attrName) && value instanceof String) {
            // Set the scheduling action for a secondary file. value is the action name,
            // currently "ADD" and "REMOVE" are the only supported values
            final FileObject fo = fileSystem.findResource(name);
            //System.out.println("scheduleSecondaryFOVcsAction("+name+", "+actionName+") = "+fo);
            if (fo == null) return ;
            FileObject primary;
            try {
                org.openide.loaders.DataObject dobj = org.openide.loaders.DataObject.find(fo);
                primary = dobj.getPrimaryFile();
                //System.out.println("  primary("+primary+").equals("+fo+") = "+primary.equals(fo));
                if (primary.equals(fo)) return ;
            } catch (org.openide.loaders.DataObjectNotFoundException exc) {
                exc.printStackTrace();
                return ;
            }
            // The scheduling is started. When all scheduling actions are done, a refresh is introduced.
            startFileScheduling(name);
            final FileObject primaryFO = primary;
            RequestProcessor.postRequest(new Runnable() {
                public void run() {
                    scheduleSecondaryFOVcsAction(name, (String) value, fo, primaryFO);
                }
            });
            super.writeAttribute(name, VcsAttributes.VCS_SCHEDULED_FILE_ATTR, value);
        } else {
            super.writeAttribute(name, attrName, value);
        }
    }
    
    private void performVcsAction(final String name, final FeatureDescriptor descriptor) throws java.net.UnknownServiceException {
        String cmdName = descriptor.getName();
        FileSystemCommand command = null;
        FileObject fo = fileSystem.findResource(name);
        File[] files = fileSystem.getFiles(new String[] {name});
        CommandDisplayer displ = new AttrCommandDisplayer(descriptor);
//        System.out.println("running.." + cmdName + " file=" + name);
//        System.out.println("fo=" + fo);
        if (cmdName.equals(VcsCommand.NAME_SCHEDULE_ADD)) {
            command = new CvsAdd(files, fileSystem.createClientProvider());
        } else if (cmdName.equals(VcsCommand.NAME_SCHEDULE_REMOVE)) {
            CvsRemove remove = new CvsRemove(files, fileSystem.createClientProvider());
            command = remove;
            // add params..
        } else throw new java.net.UnknownServiceException(cmdName);
        command.setLoadDefaultSwitches(false);
        command.setParamInput(new org.netbeans.modules.cvsclient.commands.BlankParamInput());
        command.setOutputDisplayer(displ);
        command.startCommand();
    }

  /**
     * Schedule a secondary file for "ADD" or "REMOVE". Files scheduled for remove are filtered from children().
     * Files scheduled are added as a special file attribute to the primary file.
     * @param name the name of a secondary file to be scheduled.
     * @param actionName "ADD" or "REMOVE"
     * @param fo the file object associated to the file
     * @param primary the associated primary file
     */
    private boolean scheduleSecondaryFOVcsAction(final String name, final String actionName, FileObject fo, FileObject primary) {
        if (VcsAttributes.VCS_STATUS_LOCAL.equals(primary.getAttribute(VcsAttributes.VCS_STATUS))) {
            // do not schedule local files
            endFileScheduling(name);
            return false;
        }
        int id;
        // create the descriptor of a scheduling action
        FeatureDescriptor descriptor = new FeatureDescriptor() {
            public void setValue(String attrName, Object value) {
                if (VcsAttributes.VCS_ACTION_DONE.equals(attrName)) {
                    // the scheduling action is done, inform the file system
                    fileSystem.removeScheduledFileToBeProcessed(name);
                    endFileScheduling(name);
                }
                super.setValue(attrName, value);
            }
        };
        boolean endOfScheduling = false;
        if (VcsAttributes.VCS_SCHEDULING_ADD.equals(actionName)) {
            //fileSystem.addScheduledSecondaryFO(name, VcsFileSystem.SCHEDULING_ACTION_ADD_ID);
            //FeatureDescriptor descriptor = new FeatureDescriptor();
            descriptor.setName(VcsCommand.NAME_SCHEDULE_ADD);
            try {
                fileSystem.addScheduledFileToBeProcessed(name);
                performVcsAction(name, descriptor);
            } catch (java.net.UnknownServiceException unsExc) {
                fileSystem.removeScheduledFileToBeProcessed(name);
                endOfScheduling = true;
            }
            id = 1;
        } else if (VcsAttributes.VCS_SCHEDULING_REMOVE.equals(actionName)) {
            //fileSystem.addScheduledSecondaryFO(name, VcsFileSystem.SCHEDULING_ACTION_REMOVE_ID);
            //FeatureDescriptor descriptor = new FeatureDescriptor();
            descriptor.setName(VcsCommand.NAME_SCHEDULE_REMOVE);
            try {
                fileSystem.addScheduledFileToBeProcessed(name);
                performVcsAction(name, descriptor);
            } catch (java.net.UnknownServiceException unsExc) {
                fileSystem.removeScheduledFileToBeProcessed(name);
                endOfScheduling = true;
            }
            id = 0;
        } else {
            endFileScheduling(name);
            return false;
        }
        // the file is being scheduled, add it to the primary file attribute.
        Set[] scheduled = (Set[]) primary.getAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR);
        if (scheduled == null) scheduled = new HashSet[2];
        if (scheduled[id] == null) scheduled[id] = new HashSet();
        scheduled[id].add(name);
        try {
            primary.setAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR, scheduled);
            java.io.File file = org.openide.execution.NbClassPath.toFile(primary);
            if (file != null) {
                primary.setAttribute(VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR, file.getAbsolutePath());
            }
            
        } catch (IOException ioExc) {
            if (endOfScheduling) endFileScheduling(name);
            return false;
        }
        if (endOfScheduling) endFileScheduling(name);
        return true;
    }
    
    private transient Map schedulingFilesByFolders;
    
    /**
     * Remember the file as being scheduled for it's folder.
     */
    private void startFileScheduling(String name) {
        synchronized (this) {
            if (schedulingFilesByFolders == null) {
                schedulingFilesByFolders = new HashMap();
            }
            int index = name.lastIndexOf('/');
            String dir = (index < 0) ? "" : name.substring(0, index);
            String file = (index < 0) ? name : index < (name.length() - 1) ? name.substring(index + 1) : "";
            Set files = (Set) schedulingFilesByFolders.get(dir);
            if (files == null) {
                files = new HashSet();
            }
            files.add(file);
            schedulingFilesByFolders.put(dir, files);
        }
    }
    
    /**
     * The scheduling action for this file was done, if no more actions are being
     * processed in the file's folder, do a refresh of that folder.
     */
    private void endFileScheduling(String name) {
        Set files;
        String dir;
        synchronized (this) {
            if (schedulingFilesByFolders == null) {
                schedulingFilesByFolders = new HashMap();
            }
            int index = name.lastIndexOf('/');
            dir = (index < 0) ? "" : name.substring(0, index);
            String file = (index < 0) ? name : index < (name.length() - 1) ? name.substring(index + 1) : "";
            files = (Set) schedulingFilesByFolders.get(dir);
            if (files != null) {
                files.remove(file);
                if (files.size() == 0) files = null;
            }
            if (files == null) {
                schedulingFilesByFolders.remove(dir);
            }
        }
/* no refresh is needed after remove/add
 if (files == null) {
            performRefresh(dir, Boolean.FALSE);
        }
 */
    }    
    
    public class AttrCommandDisplayer implements CommandDisplayer {
            private FeatureDescriptor desc;
            public AttrCommandDisplayer(FeatureDescriptor d) {
                desc = d;
            }
            public void displayOutputData(int moment, Object data) {
                if (CommandDisplayer.MOMENT_EXECUTE_FAILED == moment) {
                    desc.setValue(VcsAttributes.VCS_ACTION_DONE, new Boolean(false));
                }
                if (CommandDisplayer.MOMENT_FINISHED_COMMAND == moment) {
                    desc.setValue(VcsAttributes.VCS_ACTION_DONE, new Boolean(true));
                }
            }
   }
    
}
