/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.cvsclient;

import java.beans.*;
import java.io.*;
import java.util.*;
import java.util.Date;
import java.awt.Image;
import java.awt.Rectangle;
import java.lang.ref.*;
import java.text.MessageFormat;

import org.openide.util.*;
import org.openide.*;
import org.openide.util.actions.SystemAction;

import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem; // override java.io.FileSystem
import org.openide.windows.*;
import org.openide.loaders.DataObject;

import org.netbeans.modules.cvsclient.caching.*;
import org.netbeans.modules.vcscore.cache.*;
import org.netbeans.modules.vcscore.VcsAttributes;
import org.netbeans.modules.vcscore.search.VcsSearchTypeFileSystem;

import org.netbeans.modules.cvsclient.passwd.*;
import org.netbeans.modules.cvsclient.util.*;
import org.netbeans.modules.cvsclient.commands.*;
import org.netbeans.modules.cvsclient.commands.checkout.*;
import org.netbeans.modules.cvsclient.commands.diff.CvsDiff;
import org.netbeans.modules.cvsclient.commands.status.CvsStatus;
import org.netbeans.modules.cvsclient.commands.log.CvsLog;
import org.netbeans.modules.cvsclient.commands.log.RevisionExplorerInfo;
import org.netbeans.modules.cvsclient.commands.add.CvsAdd;
import org.netbeans.modules.cvsclient.commands.remove.CvsRemove;
import org.netbeans.modules.cvsclient.commands.update.CvsUpdate;
import org.netbeans.modules.cvsclient.commands.tag.CvsTag;
import org.netbeans.modules.cvsclient.commands.commit.CvsCommit;


import org.netbeans.lib.cvsclient.Client;
import org.netbeans.lib.cvsclient.connection.*;
import org.netbeans.lib.cvsclient.admin.*;
import org.netbeans.lib.cvsclient.file.*;
import org.netbeans.lib.cvsclient.util.DefaultIgnoreFileFilter;
import org.netbeans.lib.cvsclient.util.IgnoreFileFilter;
import org.netbeans.lib.cvsclient.command.GlobalOptions;

/** A file system.
 *
 * @author mkleint
 */
public class JavaCvsFileSystem extends AbstractFileSystem 
                               implements CacheHandlerListener, VcsSearchTypeFileSystem    {
    static final long serialVersionUID = -5990750897426301780L;
    
    private transient Debug E=new Debug("JavaCvsFileSystem", false); // NOI18N
    private transient Debug D=E; //transient??
    
    public static final String JAVACVS_MODE = "JavaCvsFileSystem.mode";
    private static final String PROP_REL_MOUNT = "RelativeMount";
    private static final String PROP_WORKING_DIR = "WorkingDirectory";
    private static final String PROP_OFF_LINE = "OffLine";
    private static final String PROP_AUTO_REFRESH = "AutoRefresh";
    private static final String PROP_ZIPPED_TRANSFER = "ZippedTransfer";
    private static final String PROP_USER_INTERFACE_MODE = "uiMode";
    private static final String PROP_DISPLAY_TYPE = "displayType";
    private static final String PROP_CVS_IGNORE_LIST = "cvsIgnoreList";
    
    public static final int AUTO_NONE = 0;
    public static final int AUTO_SIMPLE = 1;
    public static final int AUTO_RECURS_ONCE = 2;
    public static final int AUTO_RECURS_EVERYTIME = 3;
    
    public static final int MODE_NOVICE = 0;
    public static final int MODE_ADVANCED = 1;
    public static final int MODE_COMMAND_LINE = 2;
    
    public static final int DISP_TYPE_SIMPLE = 0;
    public static final int DISP_TYPE_FORMATTED = 1;
    
    public static final int TYPE_SERVER = 0;
    public static final int TYPE_PSERVER = 1;
    
    private static final String[] TYPE_NAMES = { "server", "pserver"};
    
    private static final int REFRESH_TIME = 150000; // 15 seconds
    protected int refreshTimeToSet = REFRESH_TIME;
    
    private File rootFile = null;
    private boolean readOnly;
    
    protected transient CvsFsCache cache=null;
    private boolean isRootSet;
    private long cacheId=0;
    
    /* REDO: no cvs root -> into CVS dirs
     */
    //    private static transient String CACHE_ROOT="vcs/cache"; // NOI18N
    
    private transient CvsClientAction cvsAction=null;
    
    private transient Vector unimportantNames = null;
    
    // properties
    private java.io.File workingDir = null;
    private String relMount = "";
    private boolean offLine = false;
    private int autoRefresh = 0;
    private boolean zippedTransfer = true;
    private int uiMode = 1;
    private int displayType = 1;
    
    // due to vcscore's cvs mount wizard.
    private boolean initialCheckout = false;
    //---
    private String repository = "";
    private String serverName = "";
    private int serverType = 1;
    private String userName = "";
    private Status status;
    
    private transient GlobalOptions globalOptions;
    

    
    public JavaCvsFileSystem () {
        // Create and use implementations of file system functionality:
        info = new InfoImpl ();
        change = new ChangeImpl ();
        status = new MyStatus();
        // Handle filesystem.attributes files normally:
//        System.out.println("constr.");
        AbstractFileSystem.List lst = new ListImpl ();
        JavaCvsAttributes a = new JavaCvsAttributes (info, change, lst, this);
        attr = a;
        list = a;        
//        System.out.println("after attr");
        // See below:
        // transfer = new TransferImpl ();
        isRootSet = false;
        setRefreshTime (REFRESH_TIME);
        setWorkingDir(new File(System.getProperty("user.home")));
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                                JavaCvsSettings.class, true); 
// load defaults from the Settings object
        setOffLine(settings.isOffLine());
        setUiMode(settings.getUiMode());
        setAutoRefresh(settings.getAutoRefresh());
        setDisplayType(settings.getDisplayType());
        //        cache = new CvsFsCache(this);
  //      init();
        initCache();
    }
    
    public JavaCvsFileSystem (FileSystemCapability cap) {
        this ();
        setCapability (cap);
        
    }
    
    protected void initCache() {
        CacheHandler handler = CacheHandler.getInstance();
        JavaCvsCache cache = (JavaCvsCache)handler.getCache(JavaCvsCache.JAVA_CACHE_NAME);
        if (cache == null) {
            cache = CvsFsCache.getCacheObject();
            handler.registerCacheType(JavaCvsCache.JAVA_CACHE_NAME, cache);
        }
    }
    
    protected void init() {
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                                JavaCvsSettings.class, true); 
        settings.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(final PropertyChangeEvent event) { 
                settingsChanged(event.getPropertyName(), event.getOldValue(), event.getNewValue());
            }
        });
        initCache();
        cache = new CvsFsCache(this);
        cache.getCacheObject().registerFileSystem(this);
        //        status = new MyStatus();
        globalOptions = new GlobalOptions();
        globalOptions.setCVSRoot(getCvsRootString());
        if (unimportantNames == null) unimportantNames = new Vector();
        if (getAutoRefresh() == JavaCvsFileSystem.AUTO_RECURS_EVERYTIME) {
            // perform recursive refresh each time we start the ide..
            Object[] root = new String[1];
            root[0] = "";
            doRefresh(root, true);
        }
        
        //        if (possibleFileStatuses == null) possibleFileStatuses = cache.getPossibleFileStatuses();
/*        try {
            setRootDirectory(rootFile);
        } catch (PropertyVetoException e) {
            // Could not set root directory
        } catch (IOException e) {
            // Could not set root directory
        }
 */
        //        commandsPool = new CommandsPool(); -- will implement handler for running commands
    }
    
    // --------- NAMING AND PROPERTIES ---------
    
    // This should be something unique based on the filesystem's config:
    protected String computeSystemName (File rootFile) {
        return "org.netbeans.modules.cvsclient.JavaCvsFileSystem" + "[" + rootFile.toString () + "]";
    }
    
    // As displayed to the user:
    public String getDisplayName () {
        if (! isValid ()) {
            return NbBundle.getMessage (JavaCvsFileSystem.class, "JavaCvsFileSystem.invalidFilesystemLabel");
        }
        else {
            return NbBundle.getMessage (JavaCvsFileSystem.class, "JavaCvsFileSystem.validFilesystemLabel", rootFile.toString ());
        }
    }
    
    // Bean getter.
    public File getRootDirectory () {
        return rootFile;
    }
    
    
    public void setupGlobalOptions(Client client) {
/*        if (isZippedTransfer()) {
            client.setFileHandler(new GzippedFileHandler());
        } else {
            client.setFileHandler(new UncompressedFileHandler());
        }
 */
        globalOptions.setCVSRoot(getCvsRootString());
    }
    
    public GlobalOptions getGlobalOptions() {
        return globalOptions;
    }
    
    public JavaCvsCommandFactory getCommandFactory() {
        return JavaCvsCommandFactory.getInstance();
    }
    
    // Bean setter. Changing the root directory (or in general, the identity
    // of the root file object) should cause everything using this filesystem
    // to refresh. The system name must change and refreshRoot should be used
    // to ensure that everything is correctly updated.
    private synchronized void setRootDirectory (File r) throws PropertyVetoException, IOException {
        
        if (! r.exists () || r.isFile ())
            throw new JavaCvsFileSystemException
            (r.toString () + " does not exist",
            NbBundle.getMessage (JavaCvsFileSystem.class, "JavaCvsFileSystem.EXC_root_dir_does_not_exist",
            r.toString ()));
        
        setSystemName (computeSystemName (r));
        rootFile = r;
//        System.out.println("rootFile=" + rootFile.getAbsolutePath());
        FileObject fo = refreshRoot();
        firePropertyChange (PROP_ROOT, null, fo);
//        globalOptions.setCVSRoot(getCvsRootString());
        isRootSet = true;
    }
    
    public String constructRootDirectory() {
        StringBuffer root = new StringBuffer();
        if (getWorkingDir() != null) {
            root.append(getWorkingDir().getAbsolutePath());
        } 
        if (root.charAt(root.length() - 1) != File.separatorChar) {
            root.append(File.separatorChar);
        }
        if (relMount.startsWith(File.separator)) {
            root.append(relMount.substring(1));
        }
        else root.append(relMount);
        if (root.charAt(root.length() - 1) == File.separatorChar) {
            root = root.deleteCharAt(root.length() - 1);
        }
        return root.toString();
    }
    
    // Bean getter.
    public boolean isReadOnly () {
        return readOnly;
    }
    
    // Bean setter.
    public void setReadOnly (boolean flag) {
        if (flag != readOnly) {
            readOnly = flag;
            firePropertyChange (PROP_READ_ONLY, new Boolean (! flag), new Boolean (flag));
        }
    }
    
    // Bean setter.
    public void setWorkingDir(java.io.File newWorking) {
        if (!newWorking.equals(workingDir)) {
            File oldWorking = workingDir;
            workingDir = newWorking;
            firePropertyChange (PROP_WORKING_DIR, oldWorking, newWorking);
            setRelMount("");
            try {
                this.setRootDirectory(new File(constructRootDirectory()));
            } catch (PropertyVetoException prop) {
                //TODO
            } catch (IOException io) {
                //TODO
            }
//            D.deb("setWorkingDir", "new Working=" + newWorking);
            //            this.cache.refreshDir("");
        }
    }
    
    // Bean getter.
    public java.io.File getWorkingDir() {
        return workingDir;
    }
    
    // Bean setter.
    public void setRelMount (String newMount) {
        if (!newMount.equals(relMount)) {
            String oldMount = relMount;
            relMount = newMount;
            firePropertyChange (PROP_REL_MOUNT, oldMount, newMount);
            try {
                this.setRootDirectory(new File(constructRootDirectory()));
            } catch (PropertyVetoException prop) {
                //TODO
            } catch (IOException io) {
                //TODO
            }
//            D.deb("setRelMount", "new relMount=" + newMount);
            //            this.cache.refreshDir("");
        }
    }
    
    // Bean getter.
    public String getRelMount () {
        if (relMount == null) return "";
        return relMount;
    }
    
    //Bean getter
    public boolean isOffLine() {
        return offLine;
    }
    
    //bean setter
    public void setOffLine(boolean off) {
        if (off != offLine) {
            firePropertyChange (PROP_OFF_LINE, new Boolean(!off), new Boolean(off));
            offLine = off;
        }
    }
    
    //Bean getter
    public int getAutoRefresh() {
        return autoRefresh;
    }
    
    //bean setter
    public void setAutoRefresh(int newAuto) {
        if (newAuto != autoRefresh) {
            int oldAuto = autoRefresh;
            autoRefresh = newAuto;
            firePropertyChange (PROP_AUTO_REFRESH, new Integer(oldAuto), new Integer(autoRefresh));
//            D.deb("setAutoRefresh", "new refresh=" + autoRefresh);
        }
    }
    
    //Bean getter
    public boolean isZippedTransfer() {
        return zippedTransfer;
    }
    
    //bean setter
    public void setZippedTransfer(boolean zipped) {
        if (zippedTransfer != zipped) {
            firePropertyChange (PROP_ZIPPED_TRANSFER, new Boolean(!zipped), new Boolean(zipped));
            zippedTransfer = zipped;
            // TODO.. set global options somehow??
        }
    }
    
    //Bean getter
    public int getUiMode() {
        return uiMode;
    }
    
    //bean setter
    public void setUiMode(int newMode) {
        if (newMode != uiMode) {
            int oldMode = uiMode;
            uiMode = newMode;
            firePropertyChange (PROP_USER_INTERFACE_MODE, new Integer(oldMode), new Integer(uiMode));
  //          D.deb("setUiMode", "new uiMode=" + uiMode);
        }
    }
    
        
    /** Getter for property displayType.
     * @return Value of property displayType.
     */
    public int getDisplayType() {
        return displayType;
    }
    
    /** Setter for property displayType.
     * @param displayType New value of property displayType.
     */
    public void setDisplayType(int newType) {
        if (newType != displayType) {
            int oldType = displayType;
            displayType = newType;
            firePropertyChange (PROP_DISPLAY_TYPE, new Integer(oldType), new Integer(displayType));
        }
    }
    
    // -------------------------------------------------------------------------
    
    public void setCvsRepository(String repos) {
        repository = repos;
    }
    public String getCvsRepository() {
        return repository;
    }
    //---
    public void setCvsServerType(int type) {
        serverType = type;
    }
    public int getCvsServerType() {
        return serverType;
    }
    public int getCvsServerType(String typeName) {
        for (int index = 0; index < TYPE_NAMES.length; index++) {
            if (TYPE_NAMES[index].equalsIgnoreCase(typeName)) return index;
        }
        return -1;
    }
    public String[] getCvsServerTypes() {
        return TYPE_NAMES;
    }
    public String getCvsServerTypeName() {
        if ((serverType >= 0) && (serverType < TYPE_NAMES.length)) {
            return TYPE_NAMES[serverType];
        }
        return "";
    }
    public String getCvsServerTypeName(int type) {
        if ((type >= 0) && (type < TYPE_NAMES.length)) {
            return TYPE_NAMES[type];
        }
        return "";
    }
    
    //---
    public void setCvsServerName(String name) {
        serverName = name;
    }
    public String getCvsServerName() {
        return serverName;
    }
    //--
    public void setCvsUserName(String name) {
        userName = name;
    }
    public String getCvsUserName() {
        return userName;
    }
    //-- cvs root string is stored in each CVS directory, also used for authentification
    public String getCvsRootString() {
        StringBuffer toReturn = new StringBuffer(":");
        toReturn.append(getCvsServerTypeName());
        toReturn.append(':');
        toReturn.append(getCvsUserName());
        toReturn.append('@');
        toReturn.append(getCvsServerName());
        toReturn.append(':');
        toReturn.append(getCvsRepository());
        return toReturn.toString();
    }
    
    // ----------- SPECIAL CAPABILITIES --------------
    
    // This is how you can affect the classpath for execution, compilation, etc.:
    public void prepareEnvironment (FileSystem.Environment environment) {
        environment.addClassPath (rootFile.toString ());
    }
    
    
    // Affect the name and icon of files on this filesystem according to their
    // "status", e.g. version-control modification-commit state:
    private class MyStatus implements Status, Serializable {
        
        static final long serialVersionUID = 7525827069203814770L;
        
        public Image annotateIcon (Image icon, int iconType, Set files) {
            // You may first modify it, e.g. by adding a check mark to the icon
            // if that makes sense for this file or group of files.
            return icon;
        }
        
        public String annotateName (String name, Set files) {
            //           D.deb("----------------------------------------------------------------------------");
            String result=name;
            String fullName=""; // NOI18N
            Object[] oo=files.toArray();
            Vector important;
            int len=oo.length;
            if( len==0 || name.indexOf(getRootDirectory().toString())>=0){
                return result;
            }
            Vector importantFiles;
            if( len==1 ) {
/*                FileObject ff=(FileObject)oo[0];
                fullName=ff.getPackageNameExt('/','.');
                important = new Vector();
                important.add(fullName);
 */
                important = getImportantFiles(oo);
            } else {
                important=getImportantFiles(oo);
            }
            String status = cache.getStatus(important).trim();
            //           D.deb("name = "+name+": status = "+status);
            if( status.length()>0 ){
                result=name+" ["+status+"]"; // NOI18N
            }
            String locker = cache.getLocker(important);
            //            D.deb("locker = '"+locker+"'");
            if (locker != null && locker.length() > 0) {
                result += " ("+locker+")";  // NOI18N
            }
            //            D.deb("annotateName("+name+") -> result='"+result+"'"); // NOI18N
            return result;
        }
        
    }
    
    public String getStatus(DataObject dObj) {
        Set files = dObj.files();
        Object[] oo = files.toArray();
        int len = oo.length;
        if (len == 0) return null;
/*        // do convert from file objects to Package+filename+ext strings
        String[] ooStrings = new String[oo.length];
        for (int index = 0; index < len; index++) {
            ooStrings[index] = ((FileObject)oo[index]).getPackageNameExt('/', '.');
        }    
 */
        if (len == 1) {
            String status = cache.getStatus(getImportantFiles(oo)).trim();
                                             // HACK??
            return status;
        }
        else {
            String status = cache.getStatus(getImportantFiles(oo)).trim();
            return status;
        }
        //TODO?? - maybe needs to be redone in order to support different statuses in files in one dataobject
    }
    
    
    public Status getStatus () {
        return status;
    }
    // And use fireFileStatusChanged whenever you know something has changed.
    
 /** VcsSearchTypeFileSystem interface stuff   */
  /** It should return all possible VCS states in which the files in the filesystem
   * can reside.
   */
    public String[] getPossibleFileStatuses() {
        return JavaCvsStatusManager.getInstance().getAllStates();
    }
    
    
    // Filesystem-specific actions, such as version-control operations.
    // The actions should typically be CookieActions looking for DataObject
    // cookies, where the object's primary file is on this type of file system.
    public SystemAction[] getActions () {
        SystemAction[] actions = new SystemAction[1];
        if (cvsAction == null) {
            cvsAction = (CvsClientAction)SharedClassObject.findObject(CvsClientAction.class, true);
            cvsAction.setFileSystem(this);
        }
        actions[0] = cvsAction;
        return actions;
    }
    
        /**
     * Get the VCS actions on a collection of <code>FileObject</code>s.
     * @param fos the collection of <code>FileObject</code>s to act on.
     * @return the actions retrieved from <code>VcsFactory.getActions(fos)</code>
     */
    public SystemAction[] getActions(Set fos) {
        SystemAction[] actions = new SystemAction[1];
        CvsClientAction cvsAction2 = (CvsClientAction)SharedClassObject.findObject(CvsClientAction.class, true);
        cvsAction2.setFileSystem(this);
        cvsAction2.setSelectedFileObjects(fos);
        actions[0] = cvsAction2;
        return actions;
    }
    
    
    // ----------- IMPLEMENTATIONS OF ABSTRACT FUNCTIONALITY ----------
    
    // Utility method only:
    File getFile (String name) {
        if (name.length() > 0) {
            return new File (rootFile, name);
        } else {
            return new File(rootFile.getAbsolutePath());
        }
    }
    
    public CvsFsCache getFsCache() {
        return cache;
    }
    
    /** Converst from String fs-related representation of a file (path+name)
     * to absolute representation in File objects.
     */
    public File[] getFiles(Object[] fileNames) {
        //       D.deb("getFiles() START");
        File[] files = new File[fileNames.length];
        for (int index = 0; index < fileNames.length; index++) {
            String name = (String)fileNames[index];
            //            D.deb("fileName=" + name);
            files[index] = getFile(name);
        }
        return files;
    }
    
    public FileObject getFileObjectForFile(File file) {
        String root = this.constructRootDirectory();
        String absPath = file.getAbsolutePath();
        if (absPath.startsWith(root)) { // it belongs to this FS -> do something
//            D.deb("-------- it is in this filesystem");
            String path;
            if (absPath.equals(root)) {
                path = "";
            } else {
                path = absPath.substring(root.length() + 1,absPath.length());
                if (path.length() > 0 && path.charAt(0) == File.separatorChar) { //another sanity check.
                    path = path.substring(1);
                }
            }
            path = path.replace('\\', '/');
            FileObject fo = findResource(path);
            return fo;
        }
        return null;
    }
    
    private Set getRecursFileObjectSet(File file) {
        FileObject fo = getFileObjectForFile(file);
        if (fo != null) {
            Set toReturn = new HashSet();
            Enumeration enum = fo.getChildren(true);
            while (enum.hasMoreElements()) {
                toReturn.add(enum.nextElement());
            }
            return toReturn;
        }
        return null;
    }
    
    // Information about files and operations on the contents which do
    // not affect the file's presence or name.
    
    // MK: no changes needed for javacvs in order the interface to be functional..
    private class InfoImpl implements AbstractFileSystem.Info {
        
        static final long serialVersionUID = -627867218820299311L;
        
        public boolean folder (String name) {
            return getFile (name).isDirectory ();
        }
        public Date lastModified (String name) {
            File file = getFile(name);
            if (!file.exists()) return new Date(System.currentTimeMillis());
            return new Date (getFile (name).lastModified ());
        }
        public boolean readOnly (String name) {
            File f = getFile (name);
            return f.exists () && ! f.canWrite ();
        }
        // Some filesystems, e.g. based on HTTP, could have better heuristics for this:
        public String mimeType (String name) {
            int i = name.lastIndexOf ('.');
            String type = i == -1 ? null : FileUtil.getMIMEType (name.substring (i + 1));
            return type == null ? "content/unknown" : type;
        }
        public long size (String name) {
            return getFile (name).length ();
        }
        public InputStream inputStream (String name) throws FileNotFoundException {
            //D.deb("inputStream("+name+")"); // NOI18N
            InputStream in = null;
            try {
                in = new FileInputStream (getFile (name));
            } catch (java.io.FileNotFoundException exc) {
                final String fname = name;
                throw (java.io.FileNotFoundException) TopManager.getDefault().getErrorManager().annotate(
                  new java.io.FileNotFoundException() {
                    public String getLocalizedMessage() {
                        return MessageFormat.format (NbBundle.getBundle(JavaCvsFileSystem.class).getString("MSG_FileNotExist"), new Object[] { fname });
                    }
                }, MessageFormat.format (NbBundle.getBundle(JavaCvsFileSystem.class).getString("MSG_FileNotExist"), new Object[] { fname }));
            }
            return in;
        }
        public OutputStream outputStream (String name) throws IOException {
            // JavaCvsFOStream enables us to be notified when an outputstream is closed
            //           -> when file is locally modified
            
            return new JavaCvsFOStream (getFile (name));
        }
        // AbstractFileSystem handles locking the file to the rest of the IDE.
        // This only means that you should define how the file should be locked
        // to the outside world--perhaps it does not need to be.
        public void lock (String name) throws IOException {
            File file = getFile (name);
            
            if (file.exists () && ! file.canWrite ())
                throw new JavaCvsFileSystemException
                ("file " + file + " could not be locked",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_file_could_not_be_locked",
                file.getName ()));
            
        }
        public void unlock (String name) {
            // Nothing special needed to unlock a file to the outside world.
        }
        public void markUnimportant (String name) {
            if (!unimportantNames.contains(name)) {
                unimportantNames.addElement(name);
                //               D.deb("==== unimportant("+name+") ====");
            }
        }
        
        
        
    }
    
    // Operations that change the available files.
    private class ChangeImpl implements AbstractFileSystem.Change {
        
        static final long serialVersionUID = 1647280700071658302L;
        
        public void createFolder (String name) throws IOException {
            File f = getFile (name);
            Object[] errorParams = new Object[] {
                f.getName (),
                getDisplayName (),
                f.getPath ()
            };
            
            if (name.equals (""))
                throw new JavaCvsFileSystemException
                ("cannot create empty name",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_create_empty_name", errorParams));
            
            if (f.exists())
                throw new JavaCvsFileSystemException
                ("folder " + f + " already exists",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_folder_already_exists", errorParams));
            
            boolean b = createRecursiveFolder (f);
            
            if (! b) {
                throw new JavaCvsFileSystemException
                ("folder " + f + " could not be created",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_folder_could_not_be_created", errorParams));
            } else {
                name = name.replace('\\', '/');
                FileObject fo = findResource(name);            
                if (fo != null && fo.getParent() != null) {
                    fo.getParent().refresh();
                }
            }
        }
        
        private boolean createRecursiveFolder (File f) {
            if (f.exists ()) return true;
            if (! f.isAbsolute ())
                f = f.getAbsoluteFile ();
            String par = f.getParent ();
            if (par == null) return false;
            if (! createRecursiveFolder (new File (par))) return false;
            f.mkdir ();
            // TODO here comes the cvscaching stuff..
            if (f.exists()) {
                cache.addToCache(f);
                return true;
            }
            return false;
        }
        
        public void createData (String name) throws IOException {
            File f = getFile (name);
            
            if (! f.createNewFile ())
                throw new JavaCvsFileSystemException
                ("file " + f + " could not be created",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_file_could_not_be_created",
                f.getName (), getDisplayName (), f.getPath ()));
            
            if (f.exists()) {
                name = name.replace('\\', '/');
                FileObject fo = findResource(name);            
                if (fo != null && fo.getParent() != null) {
                    fo.getParent().refresh();
                }
            }
            return;
        }
        
        public void rename (String oldName, String newName) throws IOException {
            File of = getFile (oldName);
            File nf = getFile (newName);
            boolean fails = nf.exists() && !nf.equals(of);
            boolean renamed = false;
            if (!fails) {
                renamed = of.renameTo(nf);
                if (renamed) {
                    String name = newName.replace('\\', '/');
                    FileObject fo = findResource(name);
                    if (fo != null && fo.getParent() != null) {
                        fo.getParent().refresh();
                    }
                }
            }
            if (! renamed || fails)
                throw new JavaCvsFileSystemException
                ("file " + of + " could not be renamed to " + nf,
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_file_could_not_be_renamed",
                new Object[] { of.getName (), nf.getName (), getDisplayName (),
                of.getPath (), nf.getPath () }));
        }
        
        public void delete (String name) throws IOException {
            File file = getFile (name);
            name = name.replace('\\', '/');
            FileObject fo = findResource(name);            
            boolean deleted = deleteFile(file);
            if (deleted && fo != null && fo.getParent() != null) {
                fo.getParent().refresh();
            }
            if (file.exists () && !deleted)
                throw new JavaCvsFileSystemException
                ("file " + file + " could not be deleted",
                NbBundle.getMessage (JavaCvsFileSystem.class, "EXC_file_could_not_be_deleted",
                file.getName (), getDisplayName (), file.getPath ()));
        }
        
        private boolean deleteFile (File file) {
            if (file.isDirectory ()) {
                File[] arr = file.listFiles ();
                for (int i = 0; i < arr.length; i++)
                    if (! deleteFile (arr[i]))
                        return false;
            }
            //update cache..
            //            String flName = file.getName();
//            boolean wasDirectory = file.isDirectory(); // that info gets lost when delete() is performed
            boolean isDone = file.delete();
/*            if (isDone) {
                cache.removeFromCache(file, wasDirectory);
            }
 */
            return isDone;
        }
        
    }
    
    // Operation which provides the directory structure.
    private class ListImpl implements List {
        
        static final long serialVersionUID = -102893327017356802L;
        
        public String[] children (String name) {
            //            D.deb("Children()","name=" + name + " ----------------------------------------------------------------------------");
            if (!isRootSet) return null;
            File f = getFile (name);
            String[] toReturn = null;
            if (f != null && f.isDirectory ()) {
                if (cache != null) {
                    HashMap removedFilesScheduledForRemove = new HashMap();                    
                    toReturn = cache.getDirContent(f);
                    if (toReturn != null) {
                        toReturn = filterScheduledSecondaryFiles(name, toReturn, removedFilesScheduledForRemove);
                    }
//                    System.out.println("children of:" + name);
/*                    if (toReturn != null) {
                        for (int i = 0; i < toReturn.length; i++) {
                            System.out.print(toReturn[i] + " ");
                        }
                        System.out.println("");
                    }
 */
/*                    if (toReturn != null) {
                        toReturn = addLocalFiles(name, vcsFiles, removedFilesScheduledForRemove);
 */
                    
                }
            }
            if (toReturn == null) {
                 toReturn = new String[0];
            }
            return toReturn;
        }
        
    }
    
    /*
    // Optional special implementations of copy and (cross-directory) move.
    private class TransferImpl implements Transfer {
     
        public boolean copy (String name, Transfer target, String targetName) throws IOException {
            // Only permit special implementation within single FS
            // (or you could implement it across file systems if you wished):
            if (target != this) return false;
            // Specially copy the file in an efficient way, e.g. implement
            // a copy-on-write algorithm.
            return true;
        }
     
        public boolean move (String name, Transfer target, String targetName) throws IOException {
            // Only permit special implementation within single FS
            // (or you could implement it across file systems if you wished):
            if (target != this) return false;
            // Specially move the file, e.g. retain rename information even
            // across directories in a version-control system.
            return true;
        }
     
    }
     */
    
    public void setRefreshTimeToSet() {
        setRefreshTime(refreshTimeToSet);
    }
    
    public boolean isImportant(String name) {
        //       D.deb("isImportant("+name+")"+unimportantNames.contains(name));
        return !unimportantNames.contains(name);
    }
    
    /**
     * Get the important files.
     * @return the Vector of important files as Strings (package + name + extension)
     */
    Vector getImportantFiles(Object[] oo){
        //        D.deb("getImportantFiles()"); // NOI18N
        Vector result=new Vector(3);
        int len=oo.length;
        
        for(int i=0;i<len;i++){
            FileObject ff=(FileObject)oo[i];
            boolean isFromThisFs = true;
            try {
                isFromThisFs = ff.getFileSystem().equals(this);
            } catch (org.openide.filesystems.FileStateInvalidException exc) {
                isFromThisFs = true;
            }
            if (isFromThisFs) {
                String fullName=ff.getPackageNameExt('/','.');
                if (isImportant(fullName)) {
                    result.addElement(fullName);
                }            
            }
            Set[] scheduled = (Set[]) ff.getAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR);
            if (scheduled != null && scheduled[0] != null) {
//                System.out.println("adding in getImportant=" + scheduled[0]);
                String filePath = (String) ff.getAttribute(VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR);
                if (filePath != null) {
                    File currentFile = org.openide.execution.NbClassPath.toFile(ff);
                    if (currentFile != null && !filePath.equals(currentFile.getAbsolutePath())) {
                        // the file was moved/copied/renamed to another location.
                        // Delete all it's scheduling attributes, because they are out-of-date
                        try {
                            ff.setAttribute(VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR, null);
                            ff.setAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR, null);
                            scheduled[0].clear();
                        } catch (IOException exc) {}
                    }
                }
                // add all scheduled files
                result.addAll(scheduled[0]);
            }
            
        }
        return result;
    }
    
    //------------------------------------------------------------------------------
    // ---------------- serialization stuff
    //-------------------------------------------
    private void readObject(ObjectInputStream in) throws
    ClassNotFoundException, IOException, NotActiveException{
        // cache is transient
        //        cache = new CvsFsCache(this);
//        System.out.println("readObject");
        in.defaultReadObject();
        AbstractFileSystem.List lst = new ListImpl ();
        JavaCvsAttributes a = new JavaCvsAttributes (info, change, lst, this);
        attr = a;
        list = a;        
        
        init();
//        D.deb("Reading bean done");
        E=new Debug("JavaCvsFileSystem", true); // NOI18N
        D=E; //transient??
        
    }
    
    
    //-------------------------------------------
    private void writeObject(ObjectOutputStream out) throws IOException {
//        D.deb("writeObject() - now save cache"); // NOI18N
        cache.saveToDisk();
  //      D.deb("writeObject() - saving bean"); // NOI18N
        out.defaultWriteObject();
   //     D.deb("writeObject() - saving bean2"); // NOI18N
        
    }
    
    
    //------------------------------------------------------------------------
    // do commands
    
    
    public ClientProvider createClientProvider() {
        IndependantClient client = new IndependantClient();
        client.setCvsRootString(getCvsRootString());
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                                JavaCvsSettings.class, true);         
        client.setUiMode(getUiMode());
        client.setDisplayType(getDisplayType());
        client.setLocalPath(getRootDirectory().getAbsolutePath());
        
        java.util.List ignoreList = JavaCvsFileSystem.constructIgnoreList(settings.getCvsHome(), 
                                    new File(getWorkingDir(), "CVSROOT"));
        IgnoreFileFilter ignoreFilter = new DefaultIgnoreFileFilter(ignoreList);
        client.setIgnoreFilter(ignoreFilter);
        try {
            GlobalOptions opt = (GlobalOptions)getGlobalOptions().clone();
            client.setGlobalOptions(opt);
        } catch (CloneNotSupportedException exc) {
            client.setGlobalOptions(globalOptions);
        }
        return client;
    }
    
    public String[] getCvsIgnoreList() {
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                                JavaCvsSettings.class, true);         
        java.util.List ignoreList = JavaCvsFileSystem.constructIgnoreList(settings.getCvsHome(), 
                                    new File(getWorkingDir(), "CVSROOT"));
        String[] array = new String[ignoreList.size()];
        array = (String[])ignoreList.toArray(array);
        return array;
    }
    
    
    static java.util.List constructIgnoreList(File home, File cvsRootDir) {
        java.util.List ignoreList = null;
        java.io.File ignoreFile = new File(cvsRootDir, "cvsignore");
        // first parse the CVSROOT/cvsignore file
        try {
            ignoreList = DefaultIgnoreFileFilter.parseCvsIgnoreFile(ignoreFile);
        } catch (IOException exc) {/* ignore */}
        // now parse the $home/.cvsignore file
        if (ignoreList == null) {
            ignoreList = new LinkedList();
        }
        ignoreFile = new File(home, ".cvsignore");
        try {
            ignoreList = DefaultIgnoreFileFilter.parseCvsIgnoreFile(ignoreFile);
        } catch (IOException exc) {/* ignore  */ }
        // now parse the Enviroment variable
        String envIgnore = System.getProperty("env-cvsignore");
        if (envIgnore != null) {
            StringTokenizer token = new StringTokenizer(envIgnore, " ", false);
            while (token.hasMoreTokens())
            {
                String tok = token.nextToken();
                ignoreList.add(tok);
            }
        }
        return ignoreList;
    }
    
    public void doRefresh(Object[] fileNames, boolean recursive) {
        // this method is just directing the processing to the cache
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        cache.doRefresh(files, recursive);
    }
    public void doRefresh(File[] files, boolean recursive) {
        // this method is just directing the processing to the cache
        if (isOffLine()) return;
        cache.doRefresh(files, recursive);
    }
    
    public void doStatus(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsStatus myStat = new CvsStatus(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
        //TODO
    }
    
    public void doLog(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsLog myStat = new CvsLog(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
        //TODO
    }
    
    public void doDiff(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsDiff myStat = new CvsDiff(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
        //TODO
    }
    
    public void doUpdate(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsUpdate myStat = new CvsUpdate(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
        //TODO
        
    }
    
    public void doCheckout(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        String start = getRelMount() + File.separator;
        start = start.replace('\\', '/');
        if (start.startsWith("/")) {
            start = start.substring(1);
        }
        IndependantClient ind = (IndependantClient)createClientProvider();
        ind.setLocalPath(getWorkingDir().getAbsolutePath());
        CvsCheckout myStat = new CvsCheckout(ind);
        myStat.setLocalPath(getWorkingDir().getAbsolutePath());
        // this is needed because the checkout command is not working relatively
        // but always requires the whole path
        // -> we want to checkout selected node it iself and not into a different structure
        for (int index = 0; index < fileNames.length; index++) {
            //            if (files[index].isDirectory()) {
            String modName = (String)fileNames[index];
            myStat.addModule(start + modName);
            //            }
        }
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
    }
    
    public void doListModules() {
        if (isOffLine()) return;
        CvsCompleteModList myStat = new CvsCompleteModList(createClientProvider());
        myStat.startCommand();
    }
    
    public void doAdd(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsAdd myStat = new CvsAdd(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
    }

    public void doRemove(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsRemove myStat = new CvsRemove(files, createClientProvider());
        if (CtrlDown) myStat.setForceParamInput(true);
        myStat.startCommand();
    }
    
    public void doCommit(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsCommit myStat = new CvsCommit(files, createClientProvider());
        myStat.setForceParamInput(true); // Do always force param input dialog... for entry of message.
        myStat.startCommand();
    }   
    
    public void doTag(Object[] fileNames, boolean CtrlDown) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        CvsTag myStat = new CvsTag(files, createClientProvider());
        myStat.setForceParamInput(true); // Do always force param input dialog... for entry of tag name.
        myStat.startCommand();
    }
    
    public void doRevisionExplorer(Object[] fileNames) {
        if (isOffLine()) return;
        File[] files = getFiles(fileNames);
        ClientProvider provider = createClientProvider();
        CvsLog myStat = new CvsLog(files, provider);
        myStat.setLoadDefaultSwitches(false);
        myStat.setParamInput(new BlankParamInput());
        myStat.setOutputDisplayer(new RevisionExplorerInfo(this,provider));
        myStat.setFixedDisplayType(true);
        myStat.startCommand();
    }
    
    
    // run when the filesystem is being removed.
    // do write all cache to files..
    public void removeNotify() {
        //        D.deb("removeNofify() - now save cache"); // NOI18N
        if (cache != null) {
            cache.saveToDiskFromNode("");
            cache.removeNotify();
        }
    }

    // run when the filesystem is being added.
    // do write all cache to files..
    public void addNotify() {
        //        D.deb("removeNofify() - now save cache"); // NOI18N
//        cache.saveToDiskFromNode("");
//        System.out.println("addNotify");
        init();
        if (initialCheckout) {
            String[] fls = {""};
            doCheckout(fls, true);
            setInitialCheckout(false);
        }
    }    
    
    //-----------------------------------------------------------------------------
    //-------------------- method from CacheHandlerListener------------------------
    public void cacheAdded(CacheHandlerEvent event) {
        //        D.deb("cacheAdded called for:" + event.getCvsCacheFile().getName());
        CacheFile file = event.getCacheFile();
        String fileString = file.getAbsolutePath();
        String root = constructRootDirectory();
        if (fileString.startsWith(root)) {
            if (fileString.length() > root.length()) {
                fileString = fileString.substring(root.length() + 1 , fileString.length());
            }
            fileString = fileString.replace('\\', '/');
            FileObject fo = findResource("");
            StringTokenizer tok = new StringTokenizer(fileString, "/", false);
            while (fo != null) {
                fo.refresh(true);
//                System.out.println("refreshing parent=" + fo.getName());
                Set foSet = new HashSet();
                Enumeration enum = fo.getChildren(false);
                while (enum.hasMoreElements()) {foSet.add(enum.nextElement());}
                fireFileStatusChanged(new FileStatusEvent(this, foSet, false, true));
                if (tok.hasMoreTokens()) {
//                    System.out.println("next token");
                    FileObject newFo = findResource(fo.getPackageName('/') + "/" + tok.nextToken());
                    if (newFo == null) {
                        fireRecursiveFileStatus(fo);
                    }
                    fo = newFo;
                } else {
                    fireRecursiveFileStatus(fo);
                    fo = null;
                }
            }
        }
    }
    
    private void fireRecursiveFileStatus(FileObject fo) {
       Set foSet = new HashSet();
       Enumeration enum = fo.getChildren(true);
       while (enum.hasMoreElements()) {
           FileObject suFO = (FileObject)enum.nextElement();
           foSet.add(suFO);
           if (suFO.isFolder()) {
               suFO.refresh(true);
//               System.out.println("refreshing child=" + suFO.getName());
           }
       }
       fireFileStatusChanged(new FileStatusEvent(this, foSet, false, true));
    }
    
    public void cacheRemoved(CacheHandlerEvent event) {
        //        D.deb("cacheRemoved called for:" + event.getCvsCacheFile().getName());
        CacheFile file = event.getCacheFile();
        String fileString = file.getParent().getAbsolutePath();
        String root = constructRootDirectory();
        if (fileString.startsWith(root)) {
            if (fileString.length() > root.length()) {
                fileString = fileString.substring(root.length() + 1 , fileString.length());
            }
            fileString = fileString.replace('\\', '/');
            FileObject fo = findResource("");
            StringTokenizer tok = new StringTokenizer(fileString, "/", false);
            while (fo != null) {
                fo.refresh(true);
                Set foSet = new HashSet();
                Enumeration enum = fo.getChildren(false);
                while (enum.hasMoreElements()) {foSet.add(enum.nextElement());}
                fireFileStatusChanged(new FileStatusEvent(this, foSet, false, true));
                if (tok.hasMoreTokens()) {
                    fo = findResource(fo.getPackageName('/') + "/" + tok.nextToken());
                } else {
                    fo = null;
                }
            }
        }
    }
    
    public void statusChanged(CacheHandlerEvent event) {
        FileObject fo = getFileObjectForFile(new File(event.getCacheFile().getAbsolutePath()));
        if (fo == null) {
            // E.err("statusChanged().. could not find FileObject.. name=" + event.getCacheFile().getAbsolutePath());
            return;
        }
        if (!event.isRecursive()) {
            if (fo.isFolder()) {
                Set foSet = new HashSet();
                Enumeration enum = fo.getChildren(false);
                while (enum.hasMoreElements()) {foSet.add(enum.nextElement());}
                fireFileStatusChanged(new FileStatusEvent(this, foSet, false, true));
            } else {
                fireFileStatusChanged (new FileStatusEvent(this, fo, false, true));
            }
        } else {
            Set foSet = getRecursFileObjectSet(new File(event.getCacheFile().getAbsolutePath()));
            Iterator it = foSet.iterator();
            while (it.hasNext()) {
                FileObject fo2 = (FileObject)it.next();
//                System.out.println(fo2.getPackageNameExt('/','.'));
                
            }
            fireFileStatusChanged(new FileStatusEvent(this, foSet, false, true));
        }
    }
    
    // ------ inner class for extending FileOutputStream that signals that a file was modified to the filesystem.
    private class JavaCvsFOStream extends FileOutputStream {
        private File changingFile;
        public JavaCvsFOStream(File file) throws IOException {
            super(file);
            this.changingFile = file;
        }
        
        public void close() throws IOException {
            super.close();
            cache.fileModified(this.changingFile);
        }
    }
    
    
    //---- debug property  - TEMPORARY
    
    public void setDebug(boolean doit) {
        Debug.debGeneralEnabled = doit;
    }
    
    public boolean isDebug() {
        return Debug.debGeneralEnabled;
    }
    
    /** Creates Reference. In FileSystem, which subclasses AbstractFileSystem, you can overload method
     * createReference(FileObject fo) to achieve another type of Reference (weak, strong etc.)
     * @param fo is FileObject. It`s reference yourequire to get.
     * @return Reference to FileObject
     */
    protected Reference createReference(FileObject fo) {
        return cache.createReference(fo);
    }
    
    public String getCacheId() {
        return cache.getCacheId();
    }
    
    
    static public Mode getDockingMode(Workspace workspace) {
        Mode mode =workspace.findMode(JavaCvsFileSystem.JAVACVS_MODE);
        if (mode == null) {
            mode = workspace.createMode(
            JavaCvsFileSystem.JAVACVS_MODE,
            NbBundle.getMessage(JavaCvsFileSystem.class, "JavaCvsFileSystem.modeTitle"),
            JavaCvsFileSystem.class.getResource(
            "/org/netbeans/modules/cvsclient/JavaCvsFileSystemIcon.gif" // NOI18N
            ));
            mode.setBounds(new Rectangle(500, 500 ));
        }
        return mode;
    }
    
    public static void errorMessage(String message) {
        final String mess = message;
        javax.swing.SwingUtilities.invokeLater(new Runnable () {
            public void run () {
                TopManager.getDefault ().notify (new NotifyDescriptor.Message(mess));
            }
        });
    }
    
    /** run when the JavaCvsSettings object fires propertyChange
     */
    private void settingsChanged(String propName, Object oldVal, Object newVal) {
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                        JavaCvsSettings.class, true);
        if (propName == null) return;
        if (propName.equals(JavaCvsSettings.PROP_USE_GLOBAL)) {
            Boolean newBool = (Boolean)newVal;
            if (newBool.booleanValue() == true) {
                setOffLine(settings.isOffLine());
                setUiMode(settings.getUiMode());
                setAutoRefresh(settings.getAutoRefresh());
                setDisplayType(settings.getDisplayType());
            }
        } else {
            if (settings.isUseGlobal()) {
                // do change the fs settings only in case the global ones have priority
                if (propName.equals(JavaCvsSettings.PROP_AUTO_REFRESH)) {
                    Integer intVal = (Integer)newVal;
                    setAutoRefresh(intVal.intValue());
                }
                if (propName.equals(JavaCvsSettings.PROP_UI_MODE)) {
                    Integer intVal = (Integer)newVal;
                    setUiMode(intVal.intValue());
                }
                if (propName.equals(JavaCvsSettings.PROP_OFFLINE)) {
                    Boolean boolVal = (Boolean)newVal;
                    setOffLine(boolVal.booleanValue());
                }
                if (propName.equals(JavaCvsSettings.PROP_DISPLAY_TYPE)) {
                    Integer intVal = (Integer)newVal;
                    setDisplayType(intVal.intValue());
                }
            }
        }
        
    }
    
    /**
     * Stuff added because of the vcscore's CVS Mount wizard
     */
    public static boolean checkLogin(String cvsRoot, String passwd) 
            throws IOException, AuthenticationException {
        String password = null;
        boolean loggedIn = false;
        CVSPasswd passFile = new CVSPasswd();
        passFile.loadPassFile();
        PasswdEntry entry = new PasswdEntry();
        entry.setEntry(cvsRoot + " " + CVSPasswd.scramble(passwd));
        loggedIn = passFile.checkServer(entry);
        if (loggedIn) {
            passFile.add(cvsRoot, passwd);
            passFile.savePassFile();
        }
        return loggedIn;

    }
    
    public void setInitialCheckout(boolean doIt) {
        initialCheckout = doIt;
    }

    
    private void checkScheduledStates(Set fos) {
/*        FileStatusProvider status = getStatusProvider();
        if (status == null) return ;
        VcsConfigVariable schVar = (VcsConfigVariable) variablesByName.get(VAR_STATUS_SCHEDULED_ADD);
        String scheduledStatusAdd = (schVar != null) ? schVar.getValue() : null;
        schVar = (VcsConfigVariable) variablesByName.get(VAR_STATUS_SCHEDULED_REMOVE);
        String scheduledStatusRemove = (schVar != null) ? schVar.getValue() : null;
 */
        JavaCvsStatusManager manager = JavaCvsStatusManager.getInstance();
        String scheduledStatusAdd = manager.getStatus(manager.convertStatus(FileStatus.ADDED));
        String scheduledStatusRemove = manager.getStatus(manager.convertStatus(FileStatus.REMOVED));
        //System.out.println("checkScheduledStates(): scheduledStatusAdd = "+scheduledStatusAdd+", scheduledStatusRemove = "+scheduledStatusRemove);
        for (Iterator it = fos.iterator(); it.hasNext(); ) {
            FileObject fo = (FileObject) it.next();
            //System.out.println("checkScheduledStates("+fo.getPackageNameExt('/', '.')+")");
            String attr = (String) fo.getAttribute(VcsAttributes.VCS_SCHEDULED_FILE_ATTR);
            //System.out.println("attr("+VcsAttributes.VCS_SCHEDULED_FILE_ATTR+") = "+attr);
            Vector vec = new Vector();
            vec.add(fo.getPackageNameExt('/','.'));
            String status = cache.getStatus(vec);
            if (VcsAttributes.VCS_SCHEDULING_ADD.equals(attr) && scheduledStatusAdd != null &&
                !scheduledStatusAdd.equals(status) &&
                isSchedulingDone(fo.getPackageNameExt('/', '.'))) {
                try {
                    fo.setAttribute(VcsAttributes.VCS_SCHEDULED_FILE_ATTR, null);
                } catch (IOException exc) {}
                removeScheduledFromPrimary(fo, 1);
            }
            if (VcsAttributes.VCS_SCHEDULING_REMOVE.equals(attr) && scheduledStatusRemove != null &&
                !scheduledStatusRemove.equals(status)) {
                try {
                    fo.setAttribute(VcsAttributes.VCS_SCHEDULED_FILE_ATTR, null);
                } catch (IOException exc) {}
                removeScheduledFromPrimary(fo, 0);
            }
        }
    }
        /**
     * The file scheduled for remove is onthe disk.
     * If it does not contain VcsAttributes.VCS_SCHEDULING_REMOVE,
     * it will be removed from the list of scheduled files, because it was
     * deleted and reappeared.
     */
/*    private void checkScheduledLocals(String path, Collection locals, Map removedFilesScheduledForRemove) {
        JavaCvsStatusManager manager = JavaCvsStatusManager.getInstance();
        String scheduledStatusRemove = manager.getStatus(manager.convertStatus(FileStatus.REMOVED));

        if (scheduledStatusRemove == null) return ;
        for (Iterator it = locals.iterator(); it.hasNext(); ) {
            String name = path + "/" + it.next();
            if (removedFilesScheduledForRemove.containsKey(name)) {
                String primary = (String) removedFilesScheduledForRemove.get(name);
                //System.out.println("checkScheduledLocals("+name+")");
                String attribute = (String) attr.readAttribute(name, VcsAttributes.VCS_SCHEDULED_FILE_ATTR);
                //System.out.println("attr("+VcsAttributes.VCS_SCHEDULED_FILE_ATTR+") = "+attr);
                if (!VcsAttributes.VCS_SCHEDULING_REMOVE.equals(attribute)) {
                    removeScheduledFromPrimary(name, primary, 0);
                }
            }
        }
    }    
  */  
    private static void removeScheduledFromPrimary(FileObject fo, int id) {
        DataObject dobj;
        try {
            dobj = DataObject.find(fo);
        } catch (org.openide.loaders.DataObjectNotFoundException exc) {
            return ;
        }
        FileObject primary = dobj.getPrimaryFile();
        Set[] scheduled = (Set[]) primary.getAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR);
        if (scheduled != null && scheduled[id] != null) {
            scheduled[id].remove(fo.getPackageNameExt('/', '.'));
            scheduled = cleanScheduledAttrs(scheduled);
            try {
                primary.setAttribute(VcsAttributes.VCS_SCHEDULED_FILES_ATTR, scheduled);
                if (scheduled == null) {
                    primary.setAttribute(VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR, null);
                }                
            } catch (IOException exc) {}
        }
    }
    
    private void removeScheduledFromPrimary(String scheduledFile, String primaryFile, int id) {
        Set[] scheduled = (Set[]) attr.readAttribute(primaryFile, VcsAttributes.VCS_SCHEDULED_FILES_ATTR);
        if (scheduled != null && scheduled[id] != null) {
            scheduled[id].remove(scheduledFile);
            scheduled = cleanScheduledAttrs(scheduled);
            try {
                attr.writeAttribute(primaryFile, VcsAttributes.VCS_SCHEDULED_FILES_ATTR, scheduled);
                if (scheduled == null) {
                    attr.writeAttribute(primaryFile, VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR, null);
                }                
            } catch (IOException exc) {}
        }
    }
        
    
    private transient Vector scheduledFilesToBeProcessed;
    
    void addScheduledFileToBeProcessed(String name) {
        if (scheduledFilesToBeProcessed == null) {
            scheduledFilesToBeProcessed = new Vector();
        }
        scheduledFilesToBeProcessed.add(name);
    }
    
    void removeScheduledFileToBeProcessed(String name) {
        if (scheduledFilesToBeProcessed == null) {
            scheduledFilesToBeProcessed = new Vector();
        }
        scheduledFilesToBeProcessed.remove(name);
    }
    
    private boolean isSchedulingDone(String name) {
        return scheduledFilesToBeProcessed == null || !scheduledFilesToBeProcessed.contains(name);
    }

    private String[] filterScheduledSecondaryFiles(String packageName, String[] files, Map removedFiles) {
        ArrayList filtered = new ArrayList(Arrays.asList(files));
        boolean emptyPackage = (packageName.length() == 0);
        for (int i = 0; i < files.length; i++) {
            String fileName = (emptyPackage) ? files[i] : (packageName + "/" + files[i]);
            Set[] scheduled = (Set[]) attr.readAttribute(fileName, VcsAttributes.VCS_SCHEDULED_FILES_ATTR);
            //System.out.println("filterScheduledSecondaryFiles("+packageName+"): "+ packageName + "/" + files[i]+" scheduled = "+scheduled);
            if (scheduled != null && scheduled[0] != null) {
                LinkedList toRemove = new LinkedList();
                for (Iterator it = scheduled[0].iterator(); it.hasNext(); ) {
                    String secFile = (String) it.next();
                    //System.out.println("secFile = '"+secFile+"'");
                    if (!emptyPackage && secFile.startsWith(packageName) ||
                        emptyPackage && secFile.indexOf('/') < 0) {
                        //System.out.println("removing '"+secFile.substring(packageName.length() + 1)+"'");
                        String nameOnly = (emptyPackage) ? secFile : secFile.substring(packageName.length() + 1);
                        boolean removed = filtered.remove(nameOnly);
   
                        //System.out.println("removed = "+removed+", filtered.contains("+secFile.substring(packageName.length() + 1)+") = "+filtered.contains(secFile.substring(packageName.length() + 1)));
                        if (!removed) {
                            toRemove.add(secFile);
                            try {
                                attr.writeAttribute(secFile, VcsAttributes.VCS_SCHEDULED_FILE_ATTR, null);
                            } catch (IOException exc) {}
                        } else {
                            // the file was removed from the list and put to the list of removed files
                            removedFiles.put(secFile, fileName);

                        }
                    }
                }
                if (toRemove.size() > 0) {
                    scheduled[0].removeAll(toRemove);
                    scheduled = cleanScheduledAttrs(scheduled);
                    try {
                        attr.writeAttribute(fileName, VcsAttributes.VCS_SCHEDULED_FILES_ATTR, scheduled);
                        if (scheduled == null) {
                            attr.writeAttribute(fileName, VcsAttributes.VCS_SCHEDULING_MASTER_FILE_NAME_ATTR, null);
                        }

                    } catch (IOException exc) {}
                }
            }
        }
        return (String[]) filtered.toArray(new String[0]);
    }
    
    private static final Set[] cleanScheduledAttrs(Set[] scheduled) {
        //System.out.print("cleanScheduledAttrs("+scheduled+") = ");
        boolean canClean = true;
        for (int k = 0; k < scheduled.length; k++) {
            if (scheduled[k] != null && scheduled[k].size() == 0) scheduled[k] = null;
            if (scheduled[k] != null) {
                canClean = false;
                break;
            }
        }
        if (canClean) scheduled = null;
        //System.out.println(scheduled);
        return scheduled;
    }

}
