/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.cvsclient.commands;


import org.openide.*;
import org.openide.util.*;
import java.util.*;
import java.io.IOException;
import java.io.File;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.Dialog;
import javax.swing.*;
import java.awt.BorderLayout;
import java.awt.event.KeyEvent;

import org.openide.windows.*;
import org.openide.NotifyDescriptor;

import org.netbeans.lib.cvsclient.event.*;
import org.netbeans.lib.cvsclient.connection.AuthenticationException;
import org.netbeans.modules.cvsclient.*;
import org.netbeans.modules.cvsclient.util.Debug;
import org.netbeans.lib.cvsclient.Client;
import org.netbeans.lib.cvsclient.command.*;
/**
 *
 * @author  mkleint
 * @version
 */
public abstract class FileSystemCommand  extends CVSAdapter 
                                         implements Runnable, Cloneable {
    private Debug E=new Debug("FileSystemCommand", true); // NOI18N
    private Debug D=E;
    
    private LinkedList resultList;
    private Thread thread;
    protected Client libClient;
    protected GlobalOptions globalOptions;
    private EventManager eventManager;
    private JavaCvsFileSystem fileSystem = null;
    private CommandParamInput paramInput;
    protected boolean paramInputOk;
    private boolean dispWasSet;
    private String root;
    private boolean forceParamInput;
    private boolean loadDefaultSwitches;    
    private boolean fixedDisplayType;
    private Command currCommand;
    private ArrayList errListeners;
    private Logger currentLogger;
    private boolean stopCommand;
    private File[] files;
    private Object[] paramOptions;
    private Object[] paramClosingOptions;
    private boolean defaultSettings;
    
    // a class used for displaying output generated by the command
    private CommandDisplayer displayer;
    
    // vector of commands to run. From first to last. Needs to be set up in InitCommand
    protected Vector toDoCommands;
    
    private boolean running;
    
    private ClientProvider provider;
    
    /** Creates new FileSystemCommand */
    
    public FileSystemCommand() {
        initInternals();
        toDoCommands = new Vector(2);
        setForceParamInput(false);
        displayer = null;
        setLoadDefaultSwitches(true);
        setFixedDisplayType(false);
    }    
    
    public FileSystemCommand(ClientProvider provider) {
        this();
        setClientProvider(provider);
    }

    
    /** prepares the internal properties for another run. To be used in CommandFactory
     */
    private void initInternals() {
        setRunning(false);
        stopCommand = false;
        currentLogger = null;
        thread = null;
        setFiles(null);
    }    
    
    public Object clone() throws CloneNotSupportedException {
         FileSystemCommand com = (FileSystemCommand)super.clone();
         com.initInternals(); // it's here to make the command ready for execution
         return com;
    }    
    
    /** set the client provider for the to-be-executed command.
     * In Case the command is created with the constructor without parameters. 
     * -> it needs to be set before starting the command.
     */
    public void setClientProvider(ClientProvider prov) {
        this.provider = prov;
        setGlobalOptions(provider.getGlobalOptions());
    }  
    
    /** set the set of files/dirs the command operates on. Not directly used here,
     * but subclasses use it in 99% of cases. They should set it before starting
     * command, according to the nodes selected or otherwise.. and feed it to 
     * undelying library commands.
     */
    public void setFiles(File[] fls) {
        files = fls;
    }    
    
    public File[] getFiles() {
        return files;
    }    
    
    /** this method is to be called from within the executeFailed() method (or others)
     * if you want to stop the command. (do not perform any other activity after the one where it's called)
     * Note: also closes any Dialogs opened which are inthe same thread.
     */
    public synchronized void stopCommand() {
        stopCommand = true;
        TopManager.getDefault().setStatusText(
        NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.stop") + "  " + getName());
    }
    
    /** Performs a hard stop of all commands running. Similar to CTRL-Break on the command-line.
     *  Not to be called from the  commands's thread but only from outside. (for calles from within the command, use stopCommand().
     *  Ex. a CommandChooser gets displayed and the user presses Cancel before the comand is finished
     *  Note: kills all dialogs or displays that are running in the same thread.
     */
    public synchronized void hardCommandStop() {
        libClient.abort();
        TopManager.getDefault().setStatusText(
        NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.abort") + "  " + getName());
        stopCommand = true; // needs to ne here??
    }
    
    public synchronized boolean isStopped() {
        return stopCommand;
    }
    
    /** sets the file system.. Should be set before stating the command (for commands run from the filesystem only..
     * Fom cache commands, use setClient instead
     * @deprecated
     */
    public void setFileSystem(JavaCvsFileSystem fs) {
        fileSystem = fs;
    }
    
    /**
     * @deprecated
     */
    protected JavaCvsFileSystem getFileSystem() {
        return fileSystem;
    }
    /**
     * @deprecated
     */
    public boolean isFileSystemSet() {
        return (fileSystem != null);
    }
    
    public ClientProvider getClientProvider() {
        return provider;
    }
    
    
    /** Returns the log of messages sent from the server to client
     */
    public StringBuffer getCurrentCommandLog() {
        if (currentLogger != null) {
            return currentLogger.getProtocolCommunication();
        }
        return null;
    }
    
    
    public void setGlobalOptions(GlobalOptions glOpt) {
        globalOptions = glOpt;
    }
    
    public GlobalOptions getGlobalOptions() {
        return globalOptions;
    }
    
    public EventManager getEventManager() throws ClientCreationException {
        if (libClient == null) {
            libClient = createClient();
        }
        return libClient.getEventManager();
    }
    
    public synchronized boolean isRunning() {
        return running;
    }
    
    /** this method is to be called right BEFORE leaving the run() method
     */
    private synchronized void setRunning(boolean run) {
        if (run == false) {
            if (errListeners != null) errListeners.clear();
        }
        running = run;
    }
    
    /** 
     * Adds an item (generally a FileInfoContainer) the the command's resultList.
     */
    
    public void addToResultList(Object item) {
        resultList.add(item);
    }
    
    /**
     * Returns a list with all the returned results.
     */
    public List getResultList() {
        return resultList;
    }
    
    /** 
     * Clears the result list of FileInfoContainers returned and parsed by the cvs client.
     */
    public void clearResultList() {
        resultList = new LinkedList();
    }
    
    
    
    
    public Command getCurrentCommand() {
        return currCommand;
    }
    
    private void setCurrentCommand(Command comm) {
        currCommand = comm;
    }
    
    /** set from outside of the command to incorporate the results received into
     * a user-defined enviroment - Ex. incorporate the Module list into a dialog "checkout module"
     * instead of displaying the list in a simple informative dialog.
     */
    public void setOutputDisplayer(CommandDisplayer disp) {
        displayer = disp;
        dispWasSet = true;
    }
    
    public boolean outputDisplayerWasSet() {
        return dispWasSet;
    }  
    
        
    public Class getDisplayerType() {
        if (displayer != null) {
            return displayer.getClass();
        } else {
            if (getClientProvider().getDisplayType() != JavaCvsFileSystem.DISP_TYPE_SIMPLE 
                 || isFixedDisplayType()) {
                return getDefaultDisplayerType();
            } else {
                return NullDisplayer.class;
            }
        }
    }
    
    protected abstract Class getDefaultDisplayerType();
    
    protected CommandDisplayer getDisplayer() {
        return displayer;
    }
    
    protected CommandDisplayer createNullDisplayer() {
        return new NullDisplayer();
    }
    
    /**
     * When set to true, means that the command must ignore the ClientProvider.getDisplayTYpe()
     * settings. Useful for things like Revision explorer, where a GUI is part of the action
     * and it has no meaning to avoid it.
     */
    public void setFixedDisplayType(boolean fixed) {
        fixedDisplayType = fixed;
    }
    /**
     * Value of "true" means that the command must ignore the ClientProvider.getDisplayTYpe()
     * settings. Useful for things like Revision explorer, where a GUI is part of the action
     * and it has no meaning to avoid it.
     * Default value is false.
     */
    public boolean isFixedDisplayType() {
        return fixedDisplayType;
    }
    
    /** sets from outside of the command a param input, to be used for
     *  different visual representation of a command's parameter input
     *  To have any effect needs to be called before starting the command or 
     *  in the initCommand() section.
     */
    public void setParamInput(CommandParamInput input) {
        paramInput = input;
    }
    
    protected CommandParamInput getParamInput() {
        if (paramInput == null) {
            if (getClientProvider().getUiMode() == JavaCvsFileSystem.MODE_COMMAND_LINE) {
                paramInput = new CommandLineParamInput();
            } else {
                paramInput = createDefaultParamInput();
            }
        }
        return paramInput;
    }    
    
    /** 
     * if set to true, will prompt the Command factory for default switches (as defined by
     * the user) Otherwise loads just the command with no switches.
     * true is default.
     */
    public void setLoadDefaultSwitches(boolean load) {
        loadDefaultSwitches = load;
    }
    public boolean isLoadDefaultSwitches() {
        return loadDefaultSwitches;
    }
    
    /** default way of getting instances of cvs commands for use in the FileSystemObjects
     */
    public Command loadCommand(Class type) {
        return getClientProvider().getCommandFactory().getCommand(type, isLoadDefaultSwitches());
    }
    
    /** starts the thread with the command - executes the run() method.
     */
    public void startCommand() {
        paramInputOk = false;
        setRunning(true);
        //create thread
        thread = new Thread(this, "JavaCvsFileSystem.CVSCommand");
        // register with the filesystems command stack
        // ??
        thread.start();
    }
    
    public abstract String getName();
    
    /** if no param input was set before staring the command, 
     *  the default defined in this method is used
     */
    
    protected abstract CommandParamInput createDefaultParamInput();
    
    
    /** this method is used for setting up the filesystem command..
     * You can put 1 or more cvs commands in the queue. (each will start a new commection to server, sequentially)
     * it's also needed to add this class as listener when underlying cvs command are buildable
     *
     */
    protected  void initCommand() {
        resultList = new LinkedList();
//           getDisplayer().displayOutputData(CommandDisplayer.MOMENT_INIT_COMMAND, null);
    }
    
    /** this method is called when all commands in the queue were already successfully finished.
     *  Good place to decide how to display the results collected.
     *  It's still in the command's thread
     */
    protected void finishedCommand() {
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_FINISHED_COMMAND, 
                                         getResultList());        
        }
    }
    
    /** gets executed before the current command in queue gets executed.
     *  To access that command use the getCurrentCommand() method
     */
    protected void beforeEachExecute() {
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_BEFORE_EACH_EXECUTE, 
                                         getCurrentCommand());        
        }
    }
    
    /** gets executed after the current command in queue successfully finished.
     *  To access that command use the getCurrentCommand() method
     */
    protected void afterEachExecute() {
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_AFTER_EACH_EXECUTE,
                                        getResultList());
        }
    }
    
    /** will run only if any of the commands in the queue failed - for whatever reason.
     * This is the place for error handling. To access the command that failed,
     * use the getCurrentCommand() method
     */
    protected void executeFailed(Exception exc) {
        String excStr = exc.getLocalizedMessage();
        if (excStr == null || excStr.equals("null")) {
            excStr = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.noErrorException");
        }
        if (exc instanceof AuthenticationException) {
            JavaCvsFileSystem.errorMessage(excStr);
        } else {
            FileSystemCommand.errorMessage(excStr, this, exc);
        }
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_EXECUTE_FAILED,
                                         getResultList());
        }
    }
    
    protected void applyInputParams() {
        Iterator it = toDoCommands.iterator();
        // for all commands queued in the toDoCommand list,
        // apply the command line params to them (from the dialog)
        if (provider.getUiMode() != JavaCvsFileSystem.MODE_NOVICE) {        
            while (it.hasNext()) {
                Command sCom = (Command)it.next();
                getParamInput().setData(sCom);
            }
        }
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_AFTER_PARAM_INPUT, null);
        }
    }
    
    public void fileInfoGenerated(FileInfoEvent e) {
        FileInfoContainer fInfo = e.getInfoContainer();
        if (fInfo != null) {
            addToResultList(fInfo);
        } 
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_FILE_INFO_GENERATED, fInfo);
        }
    }
    
    private Client createClient() throws ClientCreationException {
         ClientProvider prov = getClientProvider();
         if (prov == null) {
              throw new ClientCreationException("",
                     NbBundle.getBundle(FileSystemCommand.class).getString(
                              "FileSystemCommand.noClientAvailable"));
          }
          return prov.createClient(prov.getCvsRootString());
    }
    
    
    /** The main loop of the command
     */
    public void run() {
        try {
            if (libClient == null) {
                libClient = createClient();
            }        
            eventManager = libClient.getEventManager();
        } catch (ClientCreationException exc) {
            //TODO
            org.openide.TopManager.getDefault().notify(
                new NotifyDescriptor.Exception(exc, exc.getLocalizedMessage()));
/*            if (exc.getLocalizedMessage() != null) {
                JavaCvsFileSystem.errorMessage(getName() + " : " + exc.getMessage());
            } else {
                JavaCvsFileSystem.errorMessage(getName() + " : " + NbBundle.getMessage(FileSystemCommand.class, "FileSystemCommand.createClientError"));
            }
 */
            D.deb("ERROR WHILE CReating CLIENT!!!!");
            setRunning(false);
            return;
        }
        eventManager.addCVSListener(this);
        initCommand();
        if (isStopped()) {
            setRunning(false);
            return;
        }
        if (getClientProvider().getDisplayType() == JavaCvsFileSystem.DISP_TYPE_SIMPLE 
            && !isFixedDisplayType()) {
                setOutputDisplayer(new CommandLineInfoPanel(this));
        }
        startParamInputDialog();
        if (!paramInputOk) {
            setRunning(false);
            return;
        }
        TopManager.getDefault().setStatusText(
        NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.start") + "  " + getName());
        // create logger to have all the command's output in a String buffer.
        currentLogger = null;
        currentLogger = new Logger();
        // clear the output window.
/*        if (getClientProvider().getDisplayType() == JavaCvsFileSystem.DISP_TYPE_SIMPLE 
            && !isFixedDisplayType()) {
            try {
                getClientProvider().getCVSOutput().getOut().reset();
            } catch (IOException ioExc) {
                // TODO should be no worry about that
            }
        }
 */
        Iterator it = toDoCommands.iterator();
        while (it.hasNext()) {
            Command comm = (Command)it.next();
            setCurrentCommand(comm);
            beforeEachExecute();
            if (isStopped()) {
                setRunning(false);
                return;
            }
/** possible speedup of the commands while using non-formatted output.
 * parsing is omitted, cache updates as well..
 * -> drawbacks!
 * could be done on case-by-case basis, however generally is not acceptable.
 
            if (getClientProvider().getDisplayType() == JavaCvsFileSystem.DISP_TYPE_SIMPLE) {
                BuildableCommand bComm = (BuildableCommand)comm;
                bComm.setBuilder(null);
            }
 */
            //-- main part of loop -------------------------------------------------------------
            try {
                provider.openConnection(libClient, provider.getCvsRootString());
                eventManager.addCVSListener(currentLogger);
                libClient.executeCommand(comm,globalOptions);
                libClient.getConnection().close();
                eventManager.removeCVSListener(currentLogger);
                //----------------------------------------------------------------------------------
            } catch (IOException ioExc)  {
                eventManager.removeCVSListener(currentLogger);
                executeFailed(ioExc);
                fireCommandErrorListener(true);
                if (isStopped()) {
                    setRunning(false);
                    return;
                }

//                E.err(ioExc.getMessage()); // should be removed later
//                ioExc.printStackTrace(); // should be removed later
            } catch (CommandException exc) {
                eventManager.removeCVSListener(currentLogger);
                executeFailed(exc);
                fireCommandErrorListener(true);
                if (isStopped()) {
                    setRunning(false);
                    return;
                }
//                E.err(exc.getMessage()); // should be removed later
//                exc.printStackTrace(); // should be removed later
            } catch (AuthenticationException exc2) {
                eventManager.removeCVSListener(currentLogger);
//                E.err("ERROR WHILE creating Connection!!!!");
                executeFailed(exc2);
                fireCommandErrorListener(true);
                if (isStopped()) {
                    setRunning(false);
                    return;
                }
            }
            afterEachExecute();
            if (isStopped()) {
                setRunning(false);
                return;
            }
        }
        TopManager.getDefault().setStatusText(
              NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.finish") + "  " + getName());
        finishedCommand();
        fireCommandErrorListener(false);
        setRunning(false);
/*        if (getClientProvider().getDisplayType() == JavaCvsFileSystem.DISP_TYPE_SIMPLE
            && !isFixedDisplayType()) {
            String finished = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.commandFinished_output");
            getClientProvider().getCVSOutput().getOut().println(finished);
       }
 */
    }
    
    
    private void startParamInputDialog() {
        // let's prompt for params
        CommandParamInput parInput = getParamInput();
        if (parInput != null) {
            if (parInput.requiresUI() && (provider.getUiMode() != JavaCvsFileSystem.MODE_NOVICE
            || isForceParamInput())
            ) {
                // first call paramInput getData() to load default parameters to 
                // the input dialog
                Iterator it = toDoCommands.iterator();
                  // needs to read the data from all the commands because
                  // many different Commands can be in a FsCommand
                while (it.hasNext()) {
                   Command com = (Command) it.next();
                   parInput.getData(com);
                }  
                // now create option buttons
                paramOptions = new Object[2];
                JButton btnRun = new JButton(NbBundle.getBundle(FileSystemCommand.class).getString(
                                "FileSystemCommand.RunCommandButton"));
                btnRun.setDefaultCapable(true);
                paramOptions[0] = btnRun;
                paramOptions[1] = NbBundle.getBundle(FileSystemCommand.class).getString(
                                "FileSystemCommand.CancelButton");
                
                DialogDescriptor dd = new DialogDescriptor(mainParamInput((JPanel)parInput), 
                                                           parInput.getDialogTitle());
                dd.setValue(btnRun);
                dd.setClosingOptions(paramOptions);
                dd.setOptions(paramOptions);
                dd.setModal(false);
                dd.setButtonListener(new ActionListener() {
                    public void actionPerformed(ActionEvent evt) {
                        D.deb("here I am" + evt.getSource().getClass().getName());
                        if (evt.getSource().equals(paramOptions[1])) {
                            paramInputOk = false;
                            D.deb("Was canceled");
                            return;
                        }    
                        if (evt.getSource().equals(paramOptions[0])) {
                            if (isSetAsDefault()) {
                              doSetDefault();  
                            }    
                            paramInputOk = true;
                            applyInputParams(); // run command was pressed
                            return;
                        }
                        
                    }
                });
                Dialog dial = TopManager.getDefault().createDialog(dd);
                dial.setModal(true);
                dial.show();
            } else if (!parInput.requiresUI()){
                // does not require user interaction.. some kind of automatic parameter setup..
                // will not call paramInput.getData()
                applyInputParams();
                paramInputOk = true;
            } else if (provider.getUiMode() == JavaCvsFileSystem.MODE_NOVICE) {
                applyInputParams();
                paramInputOk = true;
            }
        } else {
            paramInputOk = true;
        }
    }
    
    /** if set the dialog prompting for parameter input will be shown everytime
     * useful for checkout dialog for example, also for the Ctrl-click option in CvsClientAction
     */
    public void setForceParamInput(boolean force) {
        forceParamInput = force;
    }
    
    public boolean isForceParamInput() {
        return forceParamInput;
    }
    
    public void addCommandErrorListener(CommandErrorListener commErrListener) {
        if (errListeners == null) {
            errListeners = new ArrayList ();
        }
        errListeners.add (commErrListener);
    }
    
    public void removeCommandErrorListener(CommandErrorListener commErrListener) {
        if (errListeners != null) {
            errListeners.remove (commErrListener);
        }
    }
    
    private void fireCommandErrorListener(boolean wasError) {
        ArrayList list;
        synchronized (this) {
            if (errListeners == null) return;
            list = (ArrayList) this.errListeners.clone();
        }
        CommandErrorEvent event = new CommandErrorEvent(this, wasError);
        
        for (int i=0; i<list.size(); i++) {
            ((CommandErrorListener)list.get (i)).errorGenerated (event);
        }
    }

    
    protected void doSetDefault() {
        boolean globalDefault = false;
        String message = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.globalDefaultQuestion");
        NotifyDescriptor nd = new NotifyDescriptor.Confirmation(message, NotifyDescriptor.YES_NO_CANCEL_OPTION);
        Object returnValue = TopManager.getDefault().notify(nd);
//        System.out.println("return value=" + returnValue.getClass().getName() + " " + returnValue.toString());
        if (NotifyDescriptor.Confirmation.CANCEL_OPTION.equals(returnValue)) {
            return;
        }    
        JavaCvsCommandFactory fact = getClientProvider().getCommandFactory();
        Iterator it = toDoCommands.iterator();
        while (it.hasNext()) {
            Command comm = (Command)it.next();
            Command newDefault = fact.getCommand(comm.getClass(), true);
            if (newDefault != null) {
              getParamInput().setData(newDefault);
              fact.putCommand(comm.getClass(), newDefault);
//              E.err("set default for=" + comm.getClass().getName());
            }  else {
        //        E.err("Couldn't set as default");
                //TODO.. any error message?
                return;
            }    
        }    
        if (NotifyDescriptor.Confirmation.YES_OPTION.equals(returnValue)) {
            //TODO :)
            fact.writeToDisk(false);
        }    
        if (NotifyDescriptor.Confirmation.NO_OPTION.equals(returnValue)) {
            //TODO :)
        }    

    }    
    
    protected void doShowCommandLine() {
        D.deb("Do show COmmandLIne");
        StringBuffer buff = new StringBuffer();
        Iterator it = toDoCommands.iterator();
        while (it.hasNext()) {
            Command comm = (Command)it.next();
            getParamInput().setData(comm);
            buff.append("cvs " + getGlobalOptions().getCVSCommand() + " ");
            String commandMess = comm.getCVSCommand();
            commandMess = commandMess.replace('\n', ' ');
            buff.append(commandMess + "\n");
        }
        JPanel toReturn = new JPanel();
        toReturn.setLayout(new BorderLayout());
        JScrollPane pane = new JScrollPane();
        JTextArea area = new JTextArea();
        area.setLineWrap(false);
        area.setText(buff.toString());
        area.setEditable(false);
        pane.setPreferredSize(new Dimension(400, 100));
        pane.setViewportView(area);
        toReturn.add(pane, BorderLayout.CENTER);
        String title = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.commandLineTitle");

        DialogDescriptor dd = new DialogDescriptor(toReturn, title);
        JButton btnClose = new JButton(NbBundle.getBundle(FileSystemCommand.class)
                                .getString("FileSystemCommand.closeButton"));
        btnClose.setDefaultCapable(true);
        dd.setValue(btnClose); //HACK
        Object[] options = {btnClose};
        dd.setOptions(options);
        dd.setClosingOptions(options);
        dd.setModal(false);
        Dialog dial = TopManager.getDefault().createDialog(dd);
        dial.setModal(true);
        dial.show();
    }
    
    public void doShowCommandLog() {
        D.deb("Do show command log");
        StringBuffer bf = getCurrentCommandLog();
        // begin visual stuff --------------
        JPanel toReturn = new JPanel();
        toReturn.setLayout(new BorderLayout());
        JScrollPane pane = new JScrollPane();
        JTextArea area = new JTextArea();
        area.setLineWrap(false);
        if (bf != null) {
            area.setText(bf.toString());
        }    
        area.setEditable(false);
        javax.swing.KeyStroke enter = javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_ENTER, 0);
        javax.swing.text.Keymap map = area.getKeymap ();
        map.removeKeyStrokeBinding (enter);

        pane.setPreferredSize(new Dimension(400, 100));
        pane.setViewportView(area);
        toReturn.add(pane, BorderLayout.CENTER);
        // - end visual stuff -------------------
        String title = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.commandLogTitle");
        toReturn.setSize(new Dimension(450, 200));
        toReturn.setMinimumSize(new Dimension(250, 100));
        DialogDescriptor dd = new DialogDescriptor(toReturn, title);
        JButton btnClose = new JButton();
        btnClose.setText(NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.closeButton"));
        btnClose.setDefaultCapable(true);
        dd.setValue(btnClose);
        Object[] options = { btnClose };
        dd.setOptions(options);
        dd.setClosingOptions(options);
        dd.setModal(false);
        Dialog dial = TopManager.getDefault().createDialog(dd);
        dial.setModal(true);
        dial.show();
    }    
    
    /** a generic procedure of displaying results in the javacvs Mode.
     * If the subclass wants to change that behaviour for whatever reason, it can change it.
     * Recommended to keep it this way for UI reasons.
     */
    public static void displayOutputPanel(String title, PersistentCommandDisplayer component) {
        JavaCvsTopComponent comp = findJavaCvsTopComponent(component);
        if (comp == null) {
            Workspace workspace = org.openide.TopManager.getDefault().getWindowManager().getCurrentWorkspace();
            Mode javaMode = JavaCvsFileSystem.getDockingMode(workspace);
            comp = new JavaCvsTopComponent();
            comp.setName(title);
            comp.setLayout(new BorderLayout());
            comp.addDisplayer(component);
            javaMode.dockInto(comp);
            comp.open();
        }
        else {
            comp.requestFocus();
        }
    }

    public static JavaCvsTopComponent findJavaCvsTopComponent(PersistentCommandDisplayer displayer) {
        Set components = TopComponent.getRegistry().getOpened();
        if (components == null || components.size() == 0) return null;
        Iterator it = components.iterator();
        while (it.hasNext()) {
            Object next = it.next();
            if (next instanceof JavaCvsTopComponent) {
                JavaCvsTopComponent jcTop = (JavaCvsTopComponent)next;
                if (jcTop.getDisplayer().equals(displayer)) {
                    return jcTop;
                }
            }
        }
        return null;
        
    }
    
    public static CommandDisplayer findOpenDisplayer(File file, Class type, Object additionalInfo) {
        Set components = TopComponent.getRegistry().getOpened();
        if (components == null || components.size() == 0) return null;
        CommandDisplayer toReturn = null;
        Iterator it = components.iterator();
        while (it.hasNext()) {
            Object next = it.next();
            if (next instanceof JavaCvsTopComponent) {
                JavaCvsTopComponent jcTop = (JavaCvsTopComponent)next;
                toReturn = jcTop.getEqualDisplayer(file, type, additionalInfo);
                if (toReturn != null) {
                    return toReturn;
                }
            }
        }
        return toReturn;
    }
    /** a generic procedure for displaying short-term results like the Update command output
     *
     */
    public static Dialog createDialog(JPanel panel, String title) {
        DialogDescriptor dd = new DialogDescriptor(panel, title);
        Object[] options = new Object[1];
        JButton btnClose = new JButton(NbBundle.getBundle(FileSystemCommand.class)
                           .getString("FileSystemCommand.closeButton"));
        btnClose.setDefaultCapable(true);
        options[0] = btnClose;
        dd.setValue(btnClose);
        dd.setOptions(options);
        dd.setClosingOptions(null); // all are closing
        Dialog dial = TopManager.getDefault().createDialog(dd);
        dial.setModal(false);
        return dial;
    }
    
    public static void errorMessage(String message, FileSystemCommand comm, Exception exc) {
        String mess = message;
        String title = NbBundle.getBundle(FileSystemCommand.class).getString("FileSystemCommand.errorMessage.title");
        final ErrorLogPanel errPanel = new ErrorLogPanel(mess, comm, exc);
        final DialogDescriptor dd = new DialogDescriptor(errPanel, title);
        errPanel.setDialogDescriptor(dd);
        errPanel.initDialogButtons(dd);
        dd.setButtonListener(errPanel);
        javax.swing.SwingUtilities.invokeLater(new Runnable () {
            public void run () {
                Dialog dial = TopManager.getDefault().createDialog(dd);
                dial.setModal(false);
                dial.show();
            }
        });
    }
    
    private JPanel mainParamInput(JPanel concrete) {
        JPanel main = new JPanel();
        JPanel pnlParams = concrete;
        JCheckBox cbDefaultSwitches = new javax.swing.JCheckBox();
        JButton btnCommandLine = new javax.swing.JButton();
        JSeparator jSeparator1 = new javax.swing.JSeparator();
// ---------------- generated stuff        
        main.setLayout(new java.awt.GridBagLayout());
        java.awt.GridBagConstraints gridBagConstraints1;
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridwidth = 2;
        gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
        main.add(pnlParams, gridBagConstraints1);
        cbDefaultSwitches.setText(java.util.ResourceBundle.getBundle("org/netbeans/modules/cvsclient/commands/Bundle").getString("FileSystemCommand.cbDefaultSwitches.text"));
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 1;
        gridBagConstraints1.gridy = 1;
        gridBagConstraints1.insets = new java.awt.Insets(10, 25, 7, 1);
        gridBagConstraints1.anchor = java.awt.GridBagConstraints.EAST;
        main.add(cbDefaultSwitches, gridBagConstraints1);
        btnCommandLine.setText(java.util.ResourceBundle.getBundle("org/netbeans/modules/cvsclient/commands/Bundle").getString("FileSystemCommand.btnCommandLine.text"));
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 1;
        gridBagConstraints1.insets = new java.awt.Insets(10, 2, 7, 25);
        gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
        main.add(btnCommandLine, gridBagConstraints1);
        jSeparator1.setPreferredSize(new java.awt.Dimension(30, 2));
        gridBagConstraints1 = new java.awt.GridBagConstraints();
        gridBagConstraints1.gridx = 0;
        gridBagConstraints1.gridy = 2;
        gridBagConstraints1.gridwidth = 2;
        gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
        main.add(jSeparator1, gridBagConstraints1);

        //--------------- generated stuff
        btnCommandLine.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                doShowCommandLine();
            }
        });
        cbDefaultSwitches.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                JCheckBox box = (JCheckBox)evt.getSource();
                saveAsDefault(box.isSelected());
            }
        });
        cbDefaultSwitches.setSelected(false);
        saveAsDefault(false);
        return main;
    }    
    
    private void saveAsDefault(boolean save) {
        defaultSettings = save;
    }    
    
    private boolean isSetAsDefault() {
        return defaultSettings;
    }    
    
    public void messageSent(MessageEvent e) {
        if (getDisplayer() != null) {
            getDisplayer().displayOutputData(CommandDisplayer.MOMENT_MESSAGE_RECEIVED,
                                        e.getMessage());
        }
    }    

    
    class Logger extends CVSAdapter {
        private StringBuffer buff;
        private org.openide.windows.InputOutput outputCVS; 
        
        public Logger() {
            buff = new StringBuffer();
/*            if (getClientProvider().getDisplayType() == JavaCvsFileSystem.DISP_TYPE_SIMPLE 
                && !isFixedDisplayType()) {
                outputCVS = getClientProvider().getCVSOutput();
            } 
 */
        }
        
        public StringBuffer getProtocolCommunication() {
            return buff;
        }
    /**
     * Called when the server wants to send a message to be displayed to
     * the user. The message is only for information purposes and clients
     * can choose to ignore these messages if they wish.
     * @param e the event
     */
        public void messageSent(MessageEvent e) {
            if (e.isTagged()) 
            {
                String line = e.getMessage();
                if (line.charAt(0) == '+' || line.charAt(0) == '-') 
                {
                    return;
                }
                if (line.equals("newline")) {
                    buff.append("\n");
                    return;
                }
                int index = line.indexOf(' ');
                if (index > 0) 
                {
                    buff.append(line.substring(index + 1));
                }
            }             
            else {
                buff.append(e.getMessage() + "\n");
            }
/*            if (outputCVS != null)  {
                outputCVS.getOut().println(e.getMessage());
            }
 */
        }
    }
    
    
}
