/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.cvsclient.commands;

import org.netbeans.modules.vcscore.util.Debug;
import org.netbeans.modules.cvsclient.*;
import org.netbeans.lib.cvsclient.command.add.AddCommand;
import org.netbeans.lib.cvsclient.command.commit.CommitCommand;
import org.netbeans.lib.cvsclient.command.remove.RemoveCommand;
import org.netbeans.lib.cvsclient.command.Command;
import org.netbeans.lib.cvsclient.command.checkout.CheckoutCommand;
import org.netbeans.lib.cvsclient.command.diff.DiffCommand;
import org.netbeans.lib.cvsclient.command.log.LogCommand;
import org.netbeans.lib.cvsclient.command.status.StatusCommand;
import org.netbeans.lib.cvsclient.command.update.UpdateCommand;
import org.netbeans.lib.cvsclient.command.tag.TagCommand;
import org.netbeans.lib.cvsclient.commandLine.GetOpt;
import org.openide.util.SharedClassObject;
import java.io.*;
import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;


/** factory that creates instances of commands. Can be used for feeding 
 * the instances with default parameters etc.
 * Should be used only for default cvs commands configurable by user.
 * @author  mkleint
 * 
 */
public class JavaCvsCommandFactory extends Object implements Serializable {
    private Debug E=new Debug("CommFactory", true); // NOI18N
    private Debug D=E;    
    
    private static JavaCvsCommandFactory instance = null;
    private String location;
    private HashMap commandMap;

    private String[] commandNames = { "add", "checkout", "diff", "commit", 
                                      "log", "remove", "status", "update", "tag"};
    private Class[] commandClasses = { AddCommand.class, CheckoutCommand.class,
        DiffCommand.class, CommitCommand.class, LogCommand.class, RemoveCommand.class, 
        StatusCommand.class, UpdateCommand.class, TagCommand.class};
    
    /** Creates new JavaCvsCommandFactory */
/*    private JavaCvsCommandFactory(String cvsrcLocation) {
        //TODO.. read up all command defaults.
        location = cvsrcLocation;
        commandMap = new HashMap();
        temporaryInitStuff();
        //TEMP - try to read from cvsrc file
        readFromDisk();
    }
 */
    
    private JavaCvsCommandFactory() {
        File locFile = new File(getHome(), ".cvsrc");
        location = locFile.getAbsolutePath();
        commandMap = new HashMap();
        temporaryInitStuff();
        JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                       JavaCvsSettings.class, true); 
        settings.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(final PropertyChangeEvent event) { 
                settingsChanged(event.getPropertyName(), event.getOldValue(), event.getNewValue());
            }
        });
        //TEMP - try to read from cvsrc file
        readFromDisk();
    }
    
    private void temporaryInitStuff() {
        commandMap.put(AddCommand.class, new AddCommand());
        commandMap.put(RemoveCommand.class, new RemoveCommand());
        commandMap.put(StatusCommand.class, new StatusCommand());
        commandMap.put(LogCommand.class, new LogCommand());
        commandMap.put(DiffCommand.class, new DiffCommand());
        commandMap.put(UpdateCommand.class, new UpdateCommand());
        commandMap.put(CheckoutCommand.class, new CheckoutCommand());
        commandMap.put(CommitCommand.class, new CommitCommand());
        commandMap.put(TagCommand.class, new TagCommand());
    }    
    
    /** could be used as singleton later, that's why this method is here
     */
    
/*    public static JavaCvsCommandFactory createCommandFactory(String cvsrcLocation) {
        return new JavaCvsCommandFactory(cvsrcLocation);
    }  
 */
    
    public static JavaCvsCommandFactory getInstance() {
        if (instance == null) {
            instance = new JavaCvsCommandFactory();
        }
        return instance;
    }

/** retrieves an istance of a command from the factory. If it's in the list,
 * clones default, if not, creates a new instance from the class object
 * @param loadDefault - supresses loading of the default (in case you want to 
 *   put switches automatically
 */
    
    public Command getCommand(Class id, boolean loadDefault) {
        Command comm = (Command)commandMap.get(id);
        if (comm == null) return null; // to make the control over returned instances clearer
          // never create instance  from id, could cause some commands to break..
        if (loadDefault != true) {
            // if the line is not in the map, create a new instance from scratch
            try {
                comm = (Command)comm.getClass().newInstance();
                return comm;
            } catch (InstantiationException exc2) {
                E.err("couldn't instantiate instance of command " + id);
            } catch (IllegalAccessException exc3) {
                E.err("Illegal access while creating command " + id);
            }    
        } else {   
            try {
               comm = (Command)comm.clone();
               return comm;
            } catch (CloneNotSupportedException exc) {
                E.err("exc while clonign command");
            } // WILL NOT HAPPEN
        }    
        return null;
    }   
    
    /** Used for initializaton of command templates. Also good for overriding defaults.
     *  Make sure to have set the files/modules list to null before putting in the map.
     *  *will be written to .cvsrc file*
     *  to have always the right classin the factory, do get and then put. 
     *  (thus never write new Command())
     */
    public void putCommand(Class id, Command command) {
        if (command instanceof CommitCommand) {
            CommitCommand com = (CommitCommand)command;
            com.setMessage(null);
        }
        if (command instanceof AddCommand) {
            AddCommand add = (AddCommand)command;
            add.setMessage(null);
        }
        commandMap.put(id, command);
    } 
    
    /** writes down the commands defaults in the cvsrc format 
     * @param localOnlyDefault - if true, writes to filesystem's root .cvsrc file
     *    otherwise writes to the default .cvsrc file (inhome directory of the user
     */
    public void writeToDisk(boolean localOnlyDefault) {
        String cvsrcFile = null;
        if (localOnlyDefault) {
             cvsrcFile = location;
        } else {
            cvsrcFile = getHome() + File.separator + ".cvsrc";
        }    
        if (cvsrcFile == null) return;
        LinkedList othersList = new LinkedList();
        D.deb("cvsrc file location =" + cvsrcFile);
// first read and store all command that we don't support.
        BufferedReader in = null;
        try {
            in = new BufferedReader
            (new InputStreamReader
            (new FileInputStream(cvsrcFile)));
            String line=null;
            while ((line=in.readLine())!= null ) {
                StringTokenizer tok = new StringTokenizer(line, " ", false);
                if (!tok.hasMoreTokens()) continue;
                String comm = tok.nextToken();
                boolean found = false;
                for (int i = 0; i < commandNames.length; i++) {
                    if (commandNames[i].equals(comm)) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    othersList.add(line);
                }
                
            }
        }
        catch (IOException e) {
            // nothing done.
        } finally {
            try {
                if (in != null) in.close();
            } catch (IOException e) {
                //nothing done?
            }
        }
        BufferedWriter out = null;
        try {
            out = new BufferedWriter(
                      new OutputStreamWriter(
                          new BufferedOutputStream(
                              new FileOutputStream(cvsrcFile)))); // NOI18N
            
            if (!localOnlyDefault) {
                //TODO write global settings as well?? or at least preserve them
            }    
            Collection col = commandMap.values();
            Iterator it = col.iterator();
            while (it.hasNext()) {
                Command comm = (Command)it.next();
                out.write(comm.getCVSCommand());
                out.newLine();
            }   
            // now write those we don't support.
            it = othersList.iterator();
            while (it.hasNext()) {
                String line = (String)it.next();
                out.write(line);
                out.newLine();
            }
            out.flush();
        }
        catch (IOException e){
            E.err(e,"writeCommands failed for " + cvsrcFile); // NOI18N
        } finally {
            try {
                if (out != null) out.close();
            } catch (IOException e) {
                E.err(e,"writeCommand failed for " + cvsrcFile); // NOI18N
            }
        }
    }   
    
    
    
    public void readFromDisk() {
            BufferedReader in = null;
            String fileName = findCvsrcFile();
            if (fileName == null) return;
            D.deb("reading from:" + fileName);
            try {
               in = new BufferedReader
                                   (new InputStreamReader
                                    (new BufferedInputStream
                                     (new FileInputStream(fileName))));
                String line=null;
                while ((line=in.readLine())!=null ){
                    StringTokenizer token = new StringTokenizer(line," ");
                    int count = token.countTokens();
                    if (count > 1) {
                        String commandString = token.nextToken();
                        D.deb("parsing command=" + commandString);
                        // load the array with args
                        String[] args = new String[count - 1];
                        int index = 0;
                        while (token.hasMoreTokens()) {
                            args[index] = token.nextToken();
                            index = index + 1;
                        }    
                        // find the right class + create default instance 
                        // + feed it with params
                        Class foundCommand = findCommandClass(commandString);
                        if (foundCommand == null) {
                            continue;
                        }
                        try {
                          Command defInstance = (Command)foundCommand.newInstance();
                          parseParams(defInstance, args);
                          putCommand(foundCommand, defInstance);
                        } catch (InstantiationException exc) {
                            continue;
                        } catch (IllegalAccessException exc2) {
                            continue;
                        }
                    }    
                }
                in.close();
            }catch (IOException e){
                E.err(e,"readDirFromDiskCache() failed"); // NOI18N
            } finally {
                if (in != null) { 
                    try {in.close();
                    } catch (IOException exc) {}
                    in = null;
                }
            }        
    }    
    
    private Class findCommandClass(String name) {
        for (int index = 0; index < commandNames.length; index++) {
            if (commandNames[index].equalsIgnoreCase(name)) {
                return commandClasses[index];
            }    
        }    
        return null;
    }    
    
    private void parseParams(Command command, String[] params) {
        // TODO
        final String getOptString = command.getOptString();
        
        GetOpt go = new GetOpt(params, getOptString);
        int ch = -1;
        go.optIndexSet(0); // hack or Ok??
        boolean usagePrint = false;
        String arg;
        while ((ch = go.getopt()) != go.optEOF)
        {
            boolean ok = command.setCVSCommand((char)ch, go.optArgGet());
            if (!ok) {
                //TODO ignore, break, rollback??
            } 
        }
    }    
    
    /** TEMPORARY
     */
     private static String getHome() {
          String toReturn;
          JavaCvsSettings settings = (JavaCvsSettings)SharedClassObject.findObject(
                                       JavaCvsSettings.class, true); 
          java.io.File home = settings.getCvsHome();
          if (home != null || !home.exists()) {
              toReturn  = home.getAbsolutePath();
          } else {
          // just to make sure.. should not happen... only when someone deletes the
          // the home dir during work with NB.
            toReturn = System.getProperty("user.home");
          }
          return toReturn;
     }
     
     /** TEMPORARY
      */
     private String findCvsrcFile() {
         String cvsrcString = location;
         File cvsrcFile = new File(cvsrcString);
         if (cvsrcFile.exists()) {
             return cvsrcString;
        } else {
            cvsrcString = getHome() + File.separator + ".cvsrc";
            cvsrcFile = new File(cvsrcString);
            if (cvsrcFile.exists()) {
                return cvsrcString;
            }
        }    
        return null;
     }    
     
         /** run when the JavaCvsSettings object fires propertyChange
     */
    private void settingsChanged(String propName, Object oldVal, Object newVal) {
        if (propName.equals(JavaCvsSettings.PROP_CVS_HOME)) {
            File newHome = (java.io.File)newVal;
            if (newHome != null) {
                location = newHome.getAbsolutePath();
                readFromDisk();
            }
        }
    }
    
}
