#! /usr/local/bin/perl -w

#                 Sun Public License Notice
# 
# The contents of this file are subject to the Sun Public License
# Version 1.0 (the "License"). You may not use this file except in
# compliance with the License. A copy of the License is available at
# http://www.sun.com/
# 
# The Original Code is NetBeans. The Initial Developer of the Original
# Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
# Microsystems, Inc. All Rights Reserved.

use strict;
use English;
use Getopt::Long;
use File::Path;
use File::Find;
use File::Copy;
use IO::Handle;

# platform dependend separators
my $psep;
my $lsep;
my $platform;

# CVS stuff
my $cvsexe = "cvs";
my @work_dirs = ();
my $tag = undef;
my @cvs_roots = ();
my @modules = ();

# command line arguments
my $build_number;
my $work_base;
my $src_base;
my $build_dir;
my $wd_list;
my $mod_list;
my $cvs_list;
my $update;
my $fastjavac;
my $jdk_home;
my $zip;
my $branch;
my $label;
my $date;
my $api;
my $init;
my $javadoc;
my $deprecation;
my $proxy;
my $help;

#
my @user_modules;
my @comp_modules = qw ( core openide );

#
my @source_dirs;
my $sources_main;
my $classpath;
my $build_filename;

#
my @files;
my @makeme_files;

# directories needed for creating build
my %dirs_build;
$dirs_build{'logs'} = 'Logs';                        # all log information
$dirs_build{'manis'} = 'Manifests';                  # all manifests used during build
$dirs_build{'builds_ready'} = 'Builds/Ready';        # folder for ready builds
$dirs_build{'builds_raw'} = 'Builds/Raw';            # folder for raw builds
$dirs_build{'release_eng'}  = 'Release/Eng';         # for creating release
$dirs_build{'release_jap'}  = 'Release/Jap';
$dirs_build{'work'} = 'Work';
$dirs_build{'javadoc'} = 'Javadoc';

# these directories under each module
# will be searched for files to add to CLASSPATH
my @clpth_files = qw ( release/lib/ext 
		       release/modules/ext 
		       release/modules/patches 
		     );

# these dirs under each module will be added to CLASSPATH
my @clpth_dirs = qw ( src libsrc );

# redirection of STDOUT and STDERR
my $OUT = \*STDOUT;
my $ERR = \*STDERR;
open SAVEOUT, '>&STDOUT';
open SAVEERR, '>&STDOUT';
my $save_out = \*SAVEOUT;
my $save_err = \*SAVEERR;
my $main_log;

BEGIN {
   require 'source-directories.pl';
}

sub main () {

    if ($^O eq "MSWin32" or $^O eq "MSDos") {
	$psep = '\\'; $lsep = ';'; $platform = 'win';
    } else {
	$psep = '/'; $lsep = ':'; $platform = 'unix';
    }

    parseArgs ();      
    prepareBuildDir ();
    prepareSrcDirs ();
    prepareWorkDir ();    
    preCompilation ();    
    compilation ();    
    createIntJars ();
    createRelease ();
    createModules ();
    createDocs ();    
    firstStart ();

}

sub parseArgs () {

    &GetOptions ( "num=s",       \$build_number, # you can specify build number
		  "workbase=s",  \$work_base,    # working directory for all building
		  "cachedir=s",  \$src_base,     # where sources will be checked out
		  "builddir=s",  \$build_dir,    # where logs and final zip files will be
		  "workdir=s",   \$wd_list,      # list of working directories with modified sources
		  "modules=s",   \$mod_list,     # list of modules
		  "cvsroot=s",   \$cvs_list,     # list of CVS roots
		  "update=i",    \$update,       # wheather to update from CVS or not
		  "jdk=s",       \$jdk_home,     # jdk location 
		  "fastjavac=s", \$fastjavac,    # executable compiler (not tested)
		  "zip=s",       \$zip,          # executable zip
		  "branch=s",    \$branch,       # build branch for CVS
		  "label=s",     \$label,        # build label for CVS
		  "date=s",      \$date,         # build date for CVS
		  "api",         \$api,          # create OpenAPI documentation		  
		  "init",        \$init,         # stops after synchronization
		  "javadoc",     \$javadoc,      # javadoc location
		  "deprecation", \$deprecation,  # show deprecation warnings
		  "proxy=s",     \$proxy,        # HTTP proxy to use
		  "help",        \$help
		) or usage ();

    usage () if $help;

    if (defined $work_base) {
	usage ("Working directory is not absolute path.") 
	    unless $work_base =~ m!^([a-zA-Z]:)?[/\\]!;
    } else { usage("Working directory is not defined!"); }
    if (defined $mod_list) {
	@user_modules = split /\s+/, $mod_list;
    } else { usage("Set of modules is not defined!"); }
    if (defined $cvs_list) {
	@cvs_roots = split /\s+/, $cvs_list;
    } else { usage("CVSROOT list is not defined!"); } 
    if (defined $wd_list) {
	@work_dirs = split /\s+/, $wd_list;
    } else { }
    if ($update != 1 and $update != 0) {
	$update = 1;
    } else { }
    
    # cvs stuff 
    if (defined ($branch) and not defined ($label) and not defined ($date)) {
	$tag = "T${branch}";
    } elsif (defined ($label) and not defined ($branch) and not defined ($date)) {
	$tag = "N${label}";
    } elsif (defined ($date) and not defined ($branch) and not defined ($label)) {
	$tag = "D${date}";
    }

    if (not defined ($build_number)) { $build_number = 001; }
    $build_filename = "Build$build_number";    
    
}

sub usage () {

    my ($msg) = @_;

    print "Error: $msg\n" if defined ($msg);    
    print <<'USAGE';

Usage:
  --num       ... build number
  --workbase  ... working directory for all building (e.g. tmp)
  --cachedir  ... where sources will be checked out and stored for next,
                  if update == 0, this arg must point to dir, where all
		  modules all stored
  --builddir  ... directory where logs and final zip files will be stored
  --workdir   ... list of working directories with modified sources
  --modules   ... list of modules to check out from CVS
  --cvsroot   ... list of CVS roots to check out sources from
  --update    ... whether to update from CVS or not
  --jdk       ... jdk location 
  --fastjavac ... executable fastjavac compiler (not working yet)
  --zip       ... executable zip (mostly for unix platforms), if not 
                  specified, final build will not be zipped (optional)
  --branch    ... build branch for CVS (optional)
  --label     ... build label (tag) for CVS (optional)
  --date      ... build date for CVS (optional)
                  Only branch or label or date can be specified
  --api       ... create OpenAPI documentation (optional, works only on unix)
  --init      ... build ends after synchronization of source  with CVS,
                  then you can update cache directory or disconnect 
		  from Internet and then run the script again
  --javadoc   ... destination directory for javadoc (optional)
  --deprecation . compile with deprecation warnings on
  --proxy     ... HTTP proxy (+ port) to use (optional, only useful with --api)
  --help      ... print this message

USAGE

    exit;

}

#
# Synchronization of sources from CVS repository with 
# cache directory. If init argument is given, building
# ends here and you can update cache directory with
# binaries
#
sub prepareSrcDirs () {

    # creates list of modules
    foreach my $mod ( @comp_modules, @user_modules ) {
	push @modules, $mod;
    }    
    
    my $with_what = ($update ? 'CVS' : 'cache');
    logMsg1 ("Synchronizing sources with $with_what ... ");
    logSTDOUT (\*SOURCES_LOG); logSTDERR (\*SOURCES_LOG);

    @source_dirs = make_source_directories @cvs_roots, $cvsexe, @work_dirs,
                   $tag, $src_base, @modules, $update,
                   \*SOURCES_LOG, \*SOURCES_LOG, \*SOURCES_LOG;    
	
    if (@source_dirs == 0) {	
	unlogSTDOUT(); unlogSTDERR();	
	logMsg2("failed!\n");
	exit;
    }

    unlogSTDOUT(); unlogSTDERR();

    logMsg2 ("finished.\n");
    
    $sources_main = $source_dirs[0];

    # ends after synchronization, then you can disconnect from Internet
    # or update cache directory and then run script again without init
    if (defined ($init)) {
	logMsg1 ("New CVS cache directory was created: $sources_main\n");
	exit;
    }
    
}

#
# Creates working directory with all checked out modules
# copied from cache direcotry, removes CVS direcotries
#
sub prepareWorkDir () {
    
    # if not exist create
    if (not -e "$work_base") {
	mkpath ("$work_base", 0, 0755) or die ("Cannot create workdir.");
    } else {
	opendir (DIRS, "$work_base");
	my @alldirs = grep !/\.|\.\./, readdir DIRS;
	closedir (DIRS);
	# if exists remove old stuff
	foreach my $dir (@alldirs) {	    	    
	    rmDir ("${work_base}${psep}${dir}");
	}
    }
    
    logMsg1 ("Copying sources to $work_base${psep} ... ");
    autoflush STDOUT, 1;
    foreach my $mod (@modules) {
	mkpath ("${work_base}${psep}${mod}") or die ("Cannot create ${work_base}${psep}${mod} directory!");
	copyDir ("${sources_main}${psep}${mod}", "${work_base}${psep}${mod}");
    }
    logMsg2 ("finished.\n");
    autoflush STDOUT 1;

    logMsg1 ("Removing CVS directories from $work_base/ ... ");
    removeCVSDirs();
    logMsg2 ("finished.\n");
    autoflush STDOUT 1;

}

#
#
#

sub prepareBuildDir () {

    die ("Absolute path is needed for build dir") unless $build_dir =~ m!^([a-zA-Z]:)?[/\\]!;   

    print ("-- [" . localtime() . "] - Preparing build directory and log files ... ");

    mkpath ($build_dir, 0, 0755) or die ("Cannot create build directory $build_dir.")
      unless -d $build_dir;
    chdir ($build_dir);

    opendir (DIRS, ".");
    my @alldirs = grep !/\.|\.\./, readdir DIRS;
    closedir (DIRS);    
    foreach my $rdir (@alldirs) {
	rmDir ("${build_dir}${psep}${rdir}");	
    }

    foreach my $dir (values %dirs_build) {	    
        if (not -e $dir) {
            mkpath ($dir, 0, 0755) or die ("Cannot create build directories.");
	}
    }

    open (MAIN_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}main.log") 
	or die ("Cannot open log file main.log.");
    autoflush MAIN_LOG 1;
    print MAIN_LOG "-- [" . localtime() . "] - START\n";
    $main_log = \*MAIN_LOG;

    open (SOURCES_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}sources.log") 
	or die ("Cannot open log file sources.log.");
    autoflush SOURCES_LOG 1;
    print SOURCES_LOG "-- [" . localtime() . "] - START\n";

    open (COMPERR_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}compile.log") 
	or die ("Cannot open log compile_stderr.log.");
    autoflush COMPERR_LOG 1;
    print COMPERR_LOG "-- [" . localtime() . "] - START\n";

    open (INTJARS_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}intjars.log") 
	or die ("Cannot open log file intjars.log.");
    autoflush INTJARS_LOG 1;
    print INTJARS_LOG "-- [" . localtime() . "] - START\n";

    open (MODULES_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}modules.log")
	or die ("Cannot open log file modules.log.");
    autoflush MODULES_LOG 1;
    print MODULES_LOG "-- [" . localtime() . "] - START\n";

    open (FIRST_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}first.log")
	or die ("Cannot open log file modules.log.");
    autoflush FIRST_LOG 1;
    print FIRST_LOG "-- [" . localtime() . "] - START\n";

    if (defined ($api)) {
	open (OPENAPI_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}openapi.log")
	    or die ("Cannot open log file openapi.log.");
	autoflush OPENAPI_LOG 1;
	print OPENAPI_LOG "-- [" . localtime() . "] - START\n";
    }

    if (defined ($javadoc)) {
	open (JAVADOC_LOG, ">$build_dir${psep}$dirs_build{'logs'}${psep}javadoc.log")
	    or die ("Cannot open log file javadoc.log.");
	autoflush JAVADOC_LOG 1;
	print JAVADOC_LOG "-- [" . localtime() . "] - START\n";
    }

    print ("finished.\n");

}

#
# Removes all CVS directories from given directory structure.
#
sub removeCVSDirs () {

    my @tokill;
    my $ld = "";

    find (sub {
	if ($File::Find::dir =~ /CVS/) {
	    if (not $ld eq $File::Find::dir) {
		push (@tokill, $File::Find::dir);
		$ld = $File::Find::dir;
	    }
	}
    }, $work_base);

    foreach my $tokill (@tokill) {
	rmDir($tokill);
    }

}

#
# Some actions are needed before compilation
#
sub preCompilation() {
    
    replaceString ("${work_base}${psep}core${psep}src${psep}org${psep}netbeans${psep}core${psep}Versioning.properties",
                   '@?BUILD_NUMBER_SUBST@?',
                   "$build_number");
                   
    replaceString ("${work_base}${psep}openide${psep}manifest.mf",
                   'build\d{1,3}|@BUILD_NUMBER_SUBST@',
                   "build$build_number");
    
    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);
    
    logMsg1 ("Starting substitution in manifests ... ");

    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}manifest.mf") {            
	    replaceString ("${work_base}${psep}${dir}${psep}manifest.mf",
			   'OpenIDE-Module-Implementation-Version: (?:\d*(\.\d)*|@BUILD_NUMBER_SUBST@)',
			   "OpenIDE-Module-Implementation-Version: $build_number");
	}
    }

    logMsg2 ("finished.\n");

}

#
#
#

sub compilation () {
    
    # classpath, which contains jar and zip files
    my $jz_classpath = '';
    # classpath, which contains directories
    my $dr_classpath = '';
    # given compiler command
    my $command;

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/\.|\.\./, readdir DIRS;
    closedir (DIRS);

    foreach my $dir (@alldirs) {	
	foreach my $ftp (@clpth_files) {	    
	    if (-e "${work_base}${psep}${dir}${psep}${ftp}") {
		$jz_classpath .= filesToString ("${work_base}${psep}${dir}${psep}${ftp}") . "$lsep";
	    }
	}
	foreach my $dtp (@clpth_dirs) {
	    if (-e "${work_base}${psep}${dir}${psep}${dtp}") {
		$dr_classpath .= "${work_base}${psep}${dir}${psep}${dtp}" . "$lsep";
	    }
	}
    }

    chop $jz_classpath;
    chop $dr_classpath;

    $classpath = "$jz_classpath${lsep}$dr_classpath"; 
    $classpath = "${classpath}${lsep}${jdk_home}${psep}lib${psep}tools.jar";
    $classpath = "${classpath}${lsep}${jdk_home}${psep}jre${psep}lib${psep}ext${psep}jpda.jar";

    # print "$classpath\n";    

    my @compiler = (defined ($fastjavac) ? ($fastjavac, '-jdk', $jdk_home) : ("${jdk_home}${psep}bin${psep}javac"));
    push @compiler, '-deprecation' if defined $deprecation;

    logMsg1 ("Starting compiling ... \n");
    logSTDOUT (\*COMPERR_LOG); logSTDERR (\*COMPERR_LOG);    

    my @erroneous;

    foreach my $dir (@alldirs) {

	chdir ($dir);	

	my @search_dirs; my $comp = undef;
	if (-e "${work_base}${psep}${dir}${psep}src") {
	    push (@search_dirs, "${work_base}${psep}${dir}${psep}src"); $comp = 1; 
	}
	if (-e "${work_base}${psep}${dir}${psep}libsrc") {
	    push (@search_dirs, "${work_base}${psep}${dir}${psep}libsrc"); $comp = 1; 
	}	
	
	if (defined ($comp)) {	    
	    find (\&wanted_compile, @search_dirs);
	    open (FILES, ">${work_base}${psep}${dir}${psep}files.txt") or die ("Cannot open files.txt");
	    foreach my $file (@files) {
		print (FILES "$file\n");
	    }
	    close (FILES);
	    logMsg1 ("Compiling module: $dir\n");
	    system (@compiler, '-classpath', $classpath, '-J-mx200m', "\@${work_base}${psep}${dir}${psep}files.txt")
	      and push @erroneous, $dir;
	}

	@files = ();

    }

    unlogSTDOUT (); unlogSTDERR ();
    if (@erroneous) {
      logMsg1 ("finished (but errors in modules: @erroneous).\n");
    } else {
      logMsg1 ("finished.\n");
    }

}

#
# Gathering of all java sources
#
sub wanted_compile () {

    if ( /\.java$/ && ($_ ne "WeakListener13.java") ) {
	my $dir = $File::Find::dir;	
	push (@files, "$dir/$_");	
    }

}

#
# Removing source files of different types
#
sub removeSources () {

    my @toremove;

    foreach my $mod ( @comp_modules, @user_modules ) {
	if (-e "${work_base}${psep}${mod}${psep}src") {
	    push @toremove, "${work_base}${psep}${mod}${psep}src";	    
	}
        if (-e "${work_base}${psep}${mod}${psep}libsrc") {
	    push @toremove, "${work_base}${psep}${mod}${psep}libsrc";	    
	}
    }

    find (sub {
	if ( ( /\.java$/ || /\.jad$/ || /\.jj$/ || 
	       /\.form$/ || /^Changes.txt$/ || 
	       /^Makefile$/ || /^makefile$/ ) && (not -d)) {	    
	    unlink ($_);
	}
    }, @toremove);

}

#
# Creating of internal jar files
#
sub createIntJars () {

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);

    logMsg1 ("Creating internal jar files ... ");    

    logSTDOUT (\*INTJARS_LOG); logSTDERR (\*INTJARS_LOG);

    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}src") {
	    find (sub {
		if (-f and $File::Find::name =~ m![/\\](\w+)[/\\]\Qsrc\E[/\\](.+)[/\\]([^/\\]+\.jar)\.MAKEME$!) {       
		    push (@makeme_files, $File::Find::name);		    
		}
	    }, "${work_base}${psep}${dir}${psep}src");
	}
    }    

    foreach my $fd (@makeme_files) {
	if ($fd =~ m![/\\](\w+)[/\\]\Qsrc\E[/\\](.+)[/\\]([^/\\]+\.jar)\.MAKEME$!) {
	    my ($mod, $path, $name) = ($1, $2, $3); 	    
	    chdir ("${work_base}${psep}${mod}${psep}tojar${psep}${path}${psep}${name}") or
		print "Cannot change directory to ${work_base}${psep}${mod}${psep}tojar${psep}${path}${psep}${name}";
	    print ("$jdk_home${psep}bin${psep}jar cvf0M ${work_base}${psep}${mod}${psep}src${psep}${path}${psep}$name *");
	    system ("$jdk_home${psep}bin${psep}jar cvf0M ${work_base}${psep}${mod}${psep}src${psep}${path}${psep}$name *");

	    if ($? != 0) { print "jar failed for $name\n"; }
	    unlink ("${work_base}${psep}${mod}${psep}src${psep}${path}${psep}${name}.MAKEME") or 
		die ("Cannot remove MAKEME file ${path}${psep}${name}.MAKEME");
	    print ("\n");
	}
    }

    unlogSTDOUT (); unlogSTDERR ();

    logMsg2 ("finished.\n");

    if (defined ($javadoc)) {
	createJavadoc ();
    }
    
    logMsg1 ("Removing sources ... ");
    removeSources();
    logMsg2 ("finished.\n");

}

#
#
#
sub createRelease () {

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);
    
    logMsg1 ("Creating release directory ... ");

    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}release") {
	    copyDir ("${work_base}${psep}${dir}${psep}release", "${build_dir}${psep}$dirs_build{'release_eng'}");	    
	}
    }    

    logMsg2 ("finished.\n");

    # [PENDING] Not needed if building on Windows; and in fact some
    # files may require that such newlines be *removed* in this case.
    logMsg1 ("Appending DOS newlines ... ");

    foreach my $batch (<${build_dir}${psep}$dirs_build{'release_eng'}${psep}bin${psep}*.bat>) {
        open BATCH_IN, $batch or die ("Fatal error: Cannot open batch file for reading!");
        open BATCH_OUT, ">$batch.NEW" or die ("Fatal error: Cannot open batch file for writing!");
        while (<BATCH_IN>) {
            chomp;
            print BATCH_OUT "$_\r\n";
        }
        close BATCH_IN;
        close BATCH_OUT;
        rename "$batch.NEW", $batch or die ("Fatal error: Cannot rename batch file!");
    }

    logMsg2 ("finished.\n");

    if ($platform eq "win") {    
	foreach my $shell (<${build_dir}${psep}$dirs_build{'release_eng'}${psep}bin${psep}*.sh>) {
        open SHELL_IN, $shell or die ("Fatal error: Cannot open shell file for reading!");
        open SHELL_OUT, ">$shell.NEW" or die ("Fatal error: Cannot open shell file for writing!");
        while (<SHELL_IN>) {
            chomp;
            print SHELL_OUT "$_\n";
        }
        close SHELL_IN;
        close SHELL_OUT;
        rename "$shell.NEW", $shell or die ("Fatal error: Cannot rename shell file!");
    }
    }

}

#
#
#
sub createModules () {

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);
    my @command;    

    my $RELEASE = "$build_dir${psep}$dirs_build{'release_eng'}";
    my $JAR = "$jdk_home${psep}bin${psep}jar";
    my $SRC;

    logMsg1 ("Creating modules ... \n");

    if (not -e "$RELEASE${psep}modules") {
        mkpath ("$RELEASE${psep}modules", 0, 0755) or die ("Cannot create $RELEASE${psep}modules.");
    }
    
    logSTDOUT (\*MODULES_LOG); logSTDERR (\*MODULES_LOG);
    
    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}jars.cfg") {	    
	    $SRC = "${work_base}${psep}${dir}";
	    open (JARS_CFG, "${work_base}${psep}${dir}${psep}jars.cfg");
	    while (<JARS_CFG>) {		
		if (m|^:(\w*): (.*)|) {
		    chdir ("${work_base}${psep}${dir}${psep}$1");
		} else { die ("Don't know how to jar module: $dir, check jars.cfg for this module."); }
		@command = split (/\s+/, $2);
		foreach (@command) {
		    s/\${JAR}/$JAR/g;
		    s/\${RELEASE}/$RELEASE/g;
		    s/\${SRC}/$SRC/g;
		}
		if ($platform eq 'win') {
		    foreach (@command) {
			s!/!\\!g;
		    }
		}
		logMsg1 ("Creating module: $dir\n");
		system @command and die;
	    }
	    close (JARS_CFG);
	}
    }

    unlogSTDOUT (); unlogSTDERR ();
    logMsg1 ("finished.\n");

    if (defined ($api)) {
	logMsg1 ("Creating OpenAPI documentation ... ");
	logSTDOUT (\*OPENAPI_LOG); logSTDERR(\*OPENAPI_LOG);
	my $proxy_addendum;
	if (defined $proxy) {
	  $proxy_addendum = "--proxy $proxy";
	} else {
	  $proxy_addendum = '';
	}
	system ("$^X $work_base${psep}openide${psep}api${psep}doc${psep}build --jdkhome=$jdk_home --stdoc $work_base${psep}openide${psep}api${psep}doc --stsrc $sources_main${psep}openide${psep}src${lsep}$sources_main${psep}openidex${psep}src --output $build_dir${psep}$dirs_build{'release_eng'}${psep}docs${psep}OpenAPIs --noshowstatus --jdkurl=public --zip=zip --zipandkill --zipbare --fastlinks --classpath $classpath $proxy_addendum");
	unlogSTDOUT (); unlogSTDERR ();
	logMsg2 ("finished.\n");
    }

}

#
#
#
sub createDocs () {

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);

    logMsg1 ("Creating documentation ... ");

    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}javahelp") {
	    copyDir ("${work_base}${psep}${dir}${psep}javahelp", "${build_dir}${psep}$dirs_build{'release_eng'}${psep}docs");	    
	}
    }

    logMsg2 ("finished.\n");

}


#
#
#
sub createJavadoc () {

    my @packages = ();
    my @dirs = ();
    my $mdir;

    opendir (DIRS, "$work_base");
    my @alldirs = grep !/CVS|\.|\.\./, readdir DIRS;
    closedir (DIRS);

    logMsg1 ("Creating  Javadoc ...");

    logSTDOUT (\*JAVADOC_LOG); logSTDERR (\*JAVADOC_LOG);

    # create list of all packages
    foreach my $dir (@alldirs) {
	if (-e "${work_base}${psep}${dir}${psep}src") {
	    find (sub {
		if (/\.java/) {
		    foreach my $pdir (@dirs) {
			if ($File::Find::dir eq $pdir) { return; }
		    }
		    $mdir = "$File::Find::dir";		    
		    push (@dirs, $mdir);		    
		    $mdir =~ s|${work_base}${psep}${dir}${psep}src${psep}||;
                    $mdir =~ s|${psep}|\.|g;                    
		    push (@packages, $mdir); 
		    
		}
		
	    }, "${work_base}${psep}${dir}${psep}src");
	}
	if (-e "${work_base}${psep}${dir}${psep}libsrc") {
	    find (sub {
		if (/\.java/) {
                    foreach my $pdir (@dirs) {
	                if ($File::Find::dir eq $pdir) { return; }
                    }
		    $mdir = "$File::Find::dir";
                    push (@dirs, $mdir);
		    $mdir =~ s|${work_base}${psep}${dir}${psep}libsrc${psep}||;
                    $mdir =~ s|${psep}|\.|g;                    
		    push (@packages, $mdir); 
		}
    
	    }, "${work_base}${psep}${dir}${psep}libsrc");
	}
    }

    # write packages to file
    open (PACKAGES, ">${work_base}${psep}packages.txt") or 
        die ("Cannot open file: ${work_base}${psep}packages.txt");
    foreach my $pack (@packages) {
        print (PACKAGES "$pack\n");
    }
    close (PACKAGES);

    # run javadoc
    system ("$jdk_home${psep}bin${psep}javadoc -classpath $classpath -J-Xmx200m -splitindex -windowtitle Build$build_number -d $build_dir${psep}$dirs_build{'javadoc'} \@${work_base}${psep}packages.txt");

    unlogSTDOUT (); unlogSTDERR ();

    logMsg2 ("finished.\n");

}

#
#
#
sub firstStart () {

    my $command;

    $ENV{JAVA_PATH} = $jdk_home;

    copyDir ("${build_dir}${psep}$dirs_build{'release_eng'}", "${build_dir}${psep}$dirs_build{'work'}");

    mkpath ("${build_dir}${psep}$dirs_build{'work'}${psep}lib${psep}patches") or print ("Creating of lib/patches failed.");

    my $bin_path = "${build_dir}${psep}$dirs_build{'work'}${psep}bin";        
    
    if ($platform eq 'unix') {

	logMsg1 ("Creating executables ... ");
	chdir ("${build_dir}${psep}$dirs_build{'work'}${psep}bin");
	opendir (BINS, "${build_dir}${psep}$dirs_build{'work'}${psep}bin");
	my @allbins = grep /[-_A-Za-z0-9]*\.sh/, readdir BINS;
	closedir (BINS);
	my $cnt = chmod (0755, @allbins);	
	if ($cnt == 0) {
	    logMsg2 ("failed!\n");
	} else {
	    logMsg2 ("finished.\n");
	}	

	$command = "${bin_path}${psep}runide.sh -J-Dnetbeans.close=true -J-Dnetbeans.debug.exceptions=true";	

    } elsif ($platform eq 'win') {

	$command = "${bin_path}${psep}runide.exe -J-Dnetbeans.close=true -J-Dnetbeans.debug.exceptions=true";

    }

    logMsg1 ("First start ... ");

    logSTDOUT (\*FIRST_LOG); logSTDERR (\*FIRST_LOG);

    chdir ("${build_dir}${psep}$dirs_build{'work'}");
    
    if (defined ($zip)) {
        print ("\nZipping raw binaries\n--------------------\n\n");
        system ("$zip -9 -r ${build_filename}_NB_raw.zip *");
        copy ("${build_dir}${psep}$dirs_build{'work'}${psep}${build_filename}_NB_raw.zip", 
    	    "${build_dir}${psep}$dirs_build{'builds_raw'}");
        unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}${build_filename}_NB_raw.zip");
    } else {
    	copyDir ("${build_dir}${psep}$dirs_build{'work'}", "${build_dir}${psep}$dirs_build{'builds_raw'}");
    }
    
    system ($command);
       
    unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}system${psep}forte4j.log");
    unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}system${psep}project.last");
    unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}system${psep}project.basic");
    unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}system${psep}filesystem_ja.attributes");
    unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}beans${psep}libs.properties");    

    if (defined ($zip)) {
    	print ("\nZipping binaries\n----------------\n\n");
    	chdir ("${build_dir}${psep}$dirs_build{'work'}");
    	system ("$zip -9 -r ${build_filename}_NB.zip *");
    	copy ("${build_dir}${psep}$dirs_build{'work'}${psep}${build_filename}_NB.zip", 
    	    "${build_dir}${psep}$dirs_build{'builds_ready'}");
    	unlink ("${build_dir}${psep}$dirs_build{'work'}${psep}${build_filename}_NB.zip");
    } else {
    	copyDir ("${build_dir}${psep}$dirs_build{'work'}", "${build_dir}${psep}$dirs_build{'builds_ready'}");
    }

    unlogSTDOUT (); unlogSTDERR ();

    logMsg2 ("finished.\n");

}

#
# Recursive copying of directories
#
sub copyDir () {

    my ($indir, $outdir) = @_;

    $indir =~ s![/\\]$!!;
    $outdir =~ s![/\\]$!!;

    die "$indir must be absolute path." unless $indir =~ m!^([a-zA-Z]:)?[/\\]!;
    die "$outdir must be absolute path." unless $outdir =~ m!^([a-zA-Z]:)?[/\\]!;

    find (sub {
	
	return unless -f;
  
	my $path = $File::Find::name;
	$path =~ s!^\Q$indir\E!$outdir! or die;

	my $dir = $path;
	$dir =~ s![/\\]([^/\\]+)$!! or die;

	mkpath ($dir);
	copy ($File::Find::name, $path)
	    or die "Copying failed: $File::Find::name -> $path: $!";

    }, $indir);

}

#
# Recursive rmdir
#
sub rmDir () {

    my ( $killdir ) = @_;
    my @tokill;

    finddepth (sub {
	push @tokill, $File::Find::name;
    }, $killdir);
    
    foreach my $tokill (@tokill) {
	if (-d $tokill) {
	    rmdir $tokill or die "Delete of directory failed: $tokill: $!";
	} else {
	    unlink $tokill or die "Delete of file failed: $tokill: $!";
	}
    }

} 

#
# Returns all zip and jar files from given dir in one string. Delimiter is platform dependent.
#
sub filesToString () {

    my ($dir) = @_;
    my $ret_str;

    opendir(DIR, $dir);
    my @files = readdir(DIR);
    closedir (DIR);
    foreach my $file (@files) {
	if (-d "${dir}${psep}${file}") { next; }
	if ($file =~ /^[^\.]*(\w*\.jar|\w*\.zip)/) {
	    $ret_str .= "${dir}${psep}${file}${lsep}";
        }
    }
    chop $ret_str;
    return $ret_str;

}
 
#
# Redirect STDERR to given file
#
sub logSTDERR ($) {
    my ($fh) = @_;
    open STDERR, '>&' . fileno ($fh);
}

#
# Unset previous redirection
#
sub unlogSTDERR () {
    logSTDERR (\*SAVEERR);
}

#
# Redirect STDOUT to given file
#
sub logSTDOUT ($) {
    my ($fh) = @_;
    open STDOUT, '>&' . fileno ($fh);
}

#
# Unset previous redirection
#
sub unlogSTDOUT () {
    logSTDOUT (\*SAVEOUT);
}

sub logMsg1 () {

    my ($msg) = @_;
    my $time = "-- [" . localtime() . "] - ";
    
    print ("${time}$msg");
    print ($main_log "${time}$msg");

}

sub logMsg2 () {

    my ($msg) = @_;
        
    print ("$msg");
    print ($main_log "$msg");

}

#
# Replaces regexp in every occurence in file
#
sub replaceString ($$$) {

    my $line;
    my ($file, $tr, $rp) = @_;

    open (IN, "<$file") or die ( "Cannot open file $file for reading." );
    open (OUT, ">$file.bak") or die ( "Cannot open file $file.bak for writing." );
    
    while ( defined ($line = <IN>) ) {
	$line =~ s/$tr/$rp/;
	print OUT $line;
    }
 
    close (IN);
    close (OUT);
    rename "$file.bak", $file or die "rename $file.bak -> $file: $!";
    
}

main ();
1;


