# -*- CPerl -*-
#                 Sun Public License Notice
# 
# The contents of this file are subject to the Sun Public License
# Version 1.0 (the "License"). You may not use this file except in
# compliance with the License. A copy of the License is available at
# http://www.sun.com/
# 
# The Original Code is NetBeans. The Initial Developer of the Original
# Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
# Microsystems, Inc. All Rights Reserved.

# Library file for use in build.pl which handles the creation of build
# work directories containing sources. It interacts with CVS, both
# with developer working directories and with the repository.
# Use the call make_source_directories.

# Sample testing command:
# perl -w -I ~/cvs/nbbuild/scripts -e 'BEGIN {require "source-directories.pl";}' -e '@cvsroots = qw(:pserver:jglick@cvs:/netbeans :pserver:jglick@cvs:/forte4j);' -e '@workdirs = qw(/home/jglick/cvs/nbsrc/org/openide/filesystems/ /home/jglick/cvs/nbsrc/org/netbeans/core/);' -e '@modules = qw(openide core text corba);' -e '@srcdirs = make_source_directories @cvsroots, "cvs", @workdirs, undef, "/home/jglick/scratch/cvswork/srcdirtest", @modules, 1, \*STDERR, \*STDERR, \*STDOUT;' -e 'print "srcdirs=(@srcdirs)\n";'
use strict;
use File::Copy;
use File::Find;
use File::Path;
# Not present in some older Perl dists:
# use File::stat;
use Cwd;

sub make_source_directories (\@$\@$$\@$***);
sub cleanroot ($;);
sub makecvsdir ($$$$$$;\@);
sub readentries ($;);

# Usage:
# BEGIN {
#   require 'source-directories.pl';
# }
# @sourcedirs = make_source_directories @cvsroots, $cvsexe, @workdirs,
#                                       $tag, $sourcebase, @modules, $update,
#                                       \*CRITLOG, \*WARNLOG, \*INFOLOG;
#
# @cvsroots should hold a list of valid CVSROOT's, such as you would
# pass to -d: ':pserver:anoncvs@cvs.netbeans.org:/cvs' and so on. There
# may be just one, or there may be several. If more than one, roots
# later in the list take precedence over those earlier in the list.
# This can mean that files checked out from the later roots can
# override files in earlier roots. For example, a developer who keeps
# a local repository with unofficial customizations can simply include
# this CVSROOT and have overridden files automatically taken from his
# customizations. It is expected that CVS can immediately start using
# all of these roots, namely that preliminary steps such as setting up
# SSH tunnels and logging in with a password have already been done.
# (Only if $update is set, of course. Otherwise CVS is not called at
# all.)
#
# $cvsexe should hold the cvs binary program. It can simply be 'cvs'
# to look in the path.
#
# @workdirs should hold a list of working directories where the
# developer keeps checked-out and possibly modified sources. These
# will all be scanned (recursively) for source files to use in the
# build. Source files which are not checked out in these directories
# will simply be taken from the repository. Any (sub-)directories
# which are found that do not contain CVS/ admin directories will be
# skipped (with a warning). If more than one working directory
# contains files from the same repository path (and root), later ones
# will take precedence (with a warning). If a (sub-)directory is found
# which uses a CVSROOT not listed verbatim among the @cvsroots, that
# directory will be skipped (with a warning). A directory (or file)
# with a tag different from that given in $tag will be skipped with a
# warning.
#
# $tag, if defined, should be a 'T' plus a CVS branch label to use as
# a sticky tag in the source area. Any never-checked-out source
# directories will use this tag, as will any directories coming from
# working dirs, as will any directories checked out from previous
# builds (switching to this tag as needed). If 'N', it is a non-branch
# tag (snapshot label), or if 'D', the rest must be a Universal Time
# date string; with 'N' or 'D', there may not be any working dirs
# specified, and all directories will be updated (frozen) to use this
# label.
#
# $sourcebase should be a directory where sources will be stored.
# Subdirectories under this, named according to the CVSROOTs, will be
# created and updated as needed.
#
# @modules is a list of CVS modules (must be top-level modules, such
# as openide or java) that should be checked out and/or updated.
#
# $update, a flag, should normally be set to true (1) to indicate that
# CVS update should be run. If a complete source directory already
# exists and the only changes are copying from working files, this may
# be turned off to save the overhead of contacting the repository.
#
# CRITLOG is a file handle to print critical error messages to.
# WARNLOG is a handle to print nonfatal warning messages to. INFOLOG
# is a handle for informational messages.
#
# Returns a list of source subdirectories (under $sourcebase), in the
# same order as given for @cvsroots (so that later directories should
# take precedence). Each such subdirectory will contain in turn
# subdirectories for each module requested, and these will have
# substructure as in the repository. Since generally most or all of a
# module lives in just one cvsroot, some such module directories will
# simply be empty. Note that the source directories are CVS working
# directories, so (1) ignore CVS/ subdirs when searching for things,
# and (2) do not modify existing files there.
#
# In case of a critical error, returns an empty list.
sub make_source_directories (\@$\@$$\@$***) {
  die "wrong number of arguments" unless @_ == 10;
  die "must be called in an array context (assign to an array)" unless wantarray;
  my ($cvsroots, $cvsexe, $workdirs,
      $_tag, $sourcebase, $modules, $update,
     $critlog, $warnlog, $infolog) = @_;
  local (*ROOT, *REPOS, *ENTRIES, *TAG, *COPIED, *DIR);
  local ($SIG{__WARN__}) = sub {
    my ($msg) = @_;
    return if $msg =~ m!Use of uninitialized value at .+[/\\]File[/\\]Find\.pm!;
    my $time = localtime;
    $msg =~ s/\n$//;
    print $warnlog "WARN [$time] $msg\n";
  };
  # Call $info->('message...') to show informational messages. For
  # warnings and errors, just use warn() and die().
  my $info = sub {
    my ($msg) = @_;
    my $time = localtime;
    print $infolog "INFO [$time] $msg\n";
  };

  my @sourcedirs;

  eval {
    $_tag = '' unless defined $_tag;
    $cvsexe = 'cvs' unless defined $cvsexe and $cvsexe;
    my ($tagtype, $tag);
    if ($_tag) {
      $tagtype = substr ($_tag, 0, 1);
      $tag = substr ($_tag, 1);
      die "bad tag value $_tag" unless $tagtype =~ m/^[TDN]$/;
      die "bad tag value $_tag" unless $tag;
    } else {
      $tagtype = '-';
      $tag = '';
    }
    die "must give at least one CVSROOT" unless @$cvsroots;
    die "cannot specify working dirs with a nonbranch tag"
      if ($tagtype eq 'D' || $tagtype eq 'N') && @$workdirs;
    die "working directories must be absolute paths" if grep {! m!^([a-zA-Z]:)?[/\\]!} @$workdirs;
    die "source base directory must be an absolute path" unless $sourcebase =~ m!^([a-zA-Z]:)?[/\\]!;
    foreach (@$workdirs) {
      s![/\\]$!!;
    }
    $sourcebase =~ s![/\\]$!!;
    
    # Make the source directories per-CVSROOT. Make them dummy
    # directories so it is possible to update in them directly.
    #
    # Use the last repository arbitrarily: it does not really matter
    # after all.
    makecvsdir ($sourcebase, $cvsroots->[-1], 'CVSROOT/Emptydir', $tagtype, $tag, $info,
                @{[map {cleanroot ($_)} @$cvsroots]});
    my $root; foreach $root (@$cvsroots) {
      my $sourcedir = "$sourcebase/" . cleanroot ($root);
      $info->("Will use source directory `$sourcedir'");
      makecvsdir ($sourcedir, $root, 'CVSROOT/Emptydir', $tagtype, $tag, $info, @$modules);
      # Also make sure there is some dir for every module. If it does
      # not in fact exist in the given repository, then oh well, CVS
      # will give a nasty message but that is all (I hope).
      my $module; foreach $module (@$modules) {
        makecvsdir ("$sourcedir/$module", $root, $module, $tagtype, $tag, $info);
      }
      push @sourcedirs, $sourcedir;
    }
    
    # map from cleanroot to repository to count
    # represents working dirs copied into sources area
    my %repopaths;
    
    # Copy from working dirs.
    $info->("Will search dirs (@$workdirs) for working files...");
    # Precalc dirs since find can get messed up when what it is
    # finding is changing also.
    my @dirs;
    find
      (sub {
         return unless -d;
         if ($_ eq 'CVS') {
           $File::Find::prune = 1;
           return;
         }
         push @dirs, $File::Find::name;
       }, @$workdirs) if @$workdirs;
    my $dir; foreach $dir (@dirs) {
      
      unless (-d "$dir/CVS" && -f "$dir/CVS/Root" && -f "$dir/CVS/Repository" && -f "$dir/CVS/Entries") {
        warn "$dir did not contain CVS administrative files, skipping";
        next;
      }
      
      # Read and verify the CVSROOT.
      my $root;
      open ROOT, "$dir/CVS/Root" or die "open $dir/CVS/Root: $!";
      $root = <ROOT>;
      chomp $root;
      close ROOT;
      unless (grep {$_ eq $root} @$cvsroots) {
        warn "CVSROOT $root found in $dir was not among known roots, skipping";
        next;
      }
      my $cleanroot = cleanroot ($root);
      my $sources = "$sourcebase/" . $cleanroot;
      $root =~ m!(:|^)(/.*)$! or die "strange CVSROOT `$root'";
      my $rootdirpart = $2;
      
      # Determine the path pointed to in the repository.
      my $repos;
      open REPOS, "$dir/CVS/Repository" or die "open $dir/CVS/Repository: $!";
      $repos = <REPOS>;
      chomp $repos;
      close REPOS;
      $repos =~ s!^\Q$rootdirpart/\E!!; # accept either /cvs/foo/bar or foo/bar
      die "strange repository path $repos" if $repos =~ m!^/! or $repos =~ m!/$!;
      if ($repos eq 'CVSROOT/Emptydir') {
        $info->("$dir was a dummy directory, skipping");
        next;
      }
      my @repos = split '/', $repos;
      if (! grep {$_ eq $repos[0]} @$modules) {
        warn "$dir is used for module $repos[0] unused in this build, skipping";
        next;
      }
      
      # Check the existing tag.
      my $existing_tag;
      if (-f "$dir/CVS/Tag") {
        open TAG, "$dir/CVS/Tag" or die "open $dir/CVS/Tag: $!";
        $existing_tag = <TAG>;
        chomp $existing_tag;
        close TAG;
      } else {
        $existing_tag = '-';
      }
      unless ("$tagtype$tag" eq $existing_tag) {
        warn "$dir uses tag $existing_tag rather than $tagtype$tag, skipping";
        next;
      }
      
      $info->("Looking for working files in `$dir' ($root/$repos)...");
      
      # Check for possible overwrites.
      if ($repopaths{$cleanroot}{$repos}++) {
	# Should not be necessary any more...actual clobberings should be caught file-by-file:
        #warn "$dir will be #$repopaths{$cleanroot}{$repos} to copy working files to $sources/$repos";
      }
      
      # Create any needed intermediate directories, incl. that for
      # the target directory itself. If they already have CVS info,
      # just leave them alone.
      for (0 .. $#repos) {
        my $intermed = join '/', @repos[0 .. $_];
        makecvsdir ("$sources/$intermed", $root, $intermed, $tagtype, $tag, $info);
      }
      
      # Copy working files one by one, preserving info from Entries (& Entries.Log).
      my $destdir = "$sources/$repos";
      # Read CVS/Entries from developer's working directory.
      my @entries = readentries ($dir);
      # Read CVS/Entries from target sources directory. We might have
      # just created it, but it might also be there and be real.
      my @entries2 = readentries ($destdir);
      unlink "$destdir/CVS/Entries.Log"
        or warn "could not remove $destdir/CVS/Entries.Log, ignoring"
          if -f "$destdir/CVS/Entries.Log";
      
      # Make a CVS/X-Copied file listing files which we copied. In
      # a subsequent run, if working directories have disappeared,
      # these will be refreshed from the repository.
      my (@oldcopied, @newcopied);
      if (-f "$destdir/CVS/X-Copied") {
        open COPIED, "$destdir/CVS/X-Copied" or die "open $destdir/CVS/X-Copied: $!";
        while (<COPIED>) {
          chomp;
          push @oldcopied, $_;
        }
        close COPIED;
      }
      open COPIED, ">$destdir/CVS/X-Copied" or die "write $destdir/CVS/X-Copied: $!";
      
      # Move developer's versions to master location, replacing
      # only as needed.
      foreach (@entries) {
        my $entry = $_;
        # Ignore directory lines completely.
        if ($entry =~ m!^/(.*)/(.*)/(.*)/(.*)/(.*)$!) {
          my ($name, $rev, $ts, $opt, $enttag) = ($1, $2, $3, $4, $5);
          
          # Check that tags really match.
          my $desiredtag;
          if ($tagtype eq '-') {
            $desiredtag = '';
          } elsif ($tagtype eq 'T' || $tagtype eq 'N') {
            $desiredtag = "T$tag";
          } else {
            $desiredtag = "D$tag";
          }
          unless ($desiredtag eq $enttag) {
            warn "$dir/$name had tag `$enttag' though `$desiredtag' is desired, skipping";
            next;
          }
          
          print COPIED "$name\n";
          push @newcopied, $name;

	  # Check for unintentionally deleted working files.
	  if (! -f "$dir/$name" && $rev !~ /^-/) {
	    warn "$dir/$name was lost, skipping...";
	    next;
	  }

	  # Check to see if both files are there, but the working
	  # dir's file is unmodified--in this case, stop, because the
	  # target file might be modified (from a directory overlap, say)
	  # and we would not want to kill that modification.
	  if (-f "$dir/$name" && -f "$destdir/$name" &&
	      $ts eq gmtime ((stat "$dir/$name")[9])) {
	    $info->("$dir/$name was unmodified, will not overwrite $destdir/$name");
	    next;
	  }
          
          # Check to see if the same version is already there.
          if ((grep {$_ eq $entry} @entries2) &&
              ((! -f "$dir/$name" && ! -f "$destdir/$name") ||
               (-f "$dir/$name" && -f "$destdir/$name" &&
                ((stat "$dir/$name")[9] == (stat "$destdir/$name")[9])))) {
            $info->("$dir/$name and $destdir/$name were already the same, skipping");
            next;
          }

	  # If both files exist, and the target file is modified, then
	  # we know that the working file is modified (else it would
	  # have been skipped above) but differs from the target file
	  # somehow (ditto)--so in this case it is good to issue a
	  # warning, and make a backup of the target file as well,
	  # before copying. Also if working file is intentionally deleted.
	  if (-f "$destdir/$name") {
	    my (@targets) = grep {m!^/\Q$name\E/!} @entries2;
	    if (@targets == 1) {
	      if ($targets[0] =~ m!^/.*/(.*)/(.*)/.*/.*$!) {
		my ($rev, $timestamp) = ($1, $2);
		if ($timestamp ne gmtime ((stat "$destdir/$name")[9])) {
		  my $bak = "$destdir/.#$name.$rev";
		  warn "Both $dir/$name and $destdir/$name were modified (differently); backing up the latter to $bak and overwriting";
		  copy ("$destdir/$name", $bak) or warn "Backup of $destdir/$name -> $bak failed: $!";
		}
		# else OK, unmodified, continue
	      } else {
		warn "Confused CVS/Entries record for $destdir/$name: $targets[0]";
		next;
	      }
	    } else {
	      warn "Confused CVS/Entries record for $destdir/$name: 0 or >1 entries: (@targets)";
	      next;
	    }
	  }
          
          # Change target Entries and move file.
          @entries2 = grep {! m!^/\Q$name\E/!} @entries2;
          push @entries2, $entry;
          if (-f "$dir/$name") {
	    $info->("copying $dir/$name to $destdir/$name");
            copy ("$dir/$name", "$destdir/$name") or die "copy $dir/$name to $destdir/$name: $!";
            my @oldstat = stat "$dir/$name" or die "stat $dir/$name failed: $!";
            utime $oldstat[8], $oldstat[9], "$destdir/$name"
              or warn "changing modification times on $destdir/$name to match $dir/$name failed: $!";
          } elsif (-f "$destdir/$name") {
	    $info->("removing $destdir/$name to match $dir/$name");
            unlink "$destdir/$name" or die "remove $destdir/$name: $!";
          }
        }
      }
      
      # Cleanup: kill previously copied files which were not copied
      # this time. Write new X-Copied and Entries files.
      close COPIED;
      foreach (@oldcopied) {
        my $name = $_;
        unless (grep {$_ eq $name} @newcopied) {
          $info->("$destdir/$name had been copied before, but not this time; reverting to repository version");
          unlink "$destdir/$name" or die "remove $destdir/$name: $!" if -f "$destdir/$name";
          @entries2 = grep {! m!^/\Q$name\E/!} @entries2;
        }
      }
      open ENTRIES, ">$destdir/CVS/Entries" or die "write $destdir/CVS/Entries: $!";
      foreach (@entries2) {
        print ENTRIES "$_\n";
      }
      close ENTRIES;
      
    }                           # foreach dir
    
    if ($update) {
      
      # Clean up source dir: look for directories which had some copied
      # files before, but now are not being copied from any working
      # dirs. For safety, remove all files in them (not subdirectories)
      # and clear their Entries files.
      $info->("Cleaning source dirs of previous working files, to enable reversion");
      my @dirs;
      find
        (sub {
           return unless -d;
           if ($_ eq 'CVS') {
             $File::Find::prune = 1;
             return;
           }
           push @dirs, $File::Find::name;
         }, $sourcebase);
      my $dir; foreach $dir (@dirs) {
        unless (-f "$dir/CVS/Entries") {
          warn "$dir/CVS/Entries should have existed but did not";
          next;
        }
        next unless -f "$dir/CVS/X-Copied";
        $dir =~ m!^\Q$sourcebase\E[/\\]([^/\\]+)[/\\](.+)$! or die "unexpected path $dir (with CVS/X-Copied)";
        my ($cleanroot, $repos) = ($1, $2);
        # If it was not used for working files in this run of the
        # function, revert it.
        unless ($repopaths{$cleanroot}{$repos}) {
          $info->("Reverting files in $dir to repository version");
          opendir DIR, $dir or die "opendir $dir: $!";
          my @files = grep {-f "$dir/$_"} readdir DIR;
          closedir DIR;
          foreach (@files) {
            unlink "$dir/$_" or warn "could not remove $dir/$_: $!";
          }
          open ENTRIES, ">$dir/CVS/Entries" or die "write $dir/CVS/Entries: $!";
          # no entries anymore, revert all from repository
          close ENTRIES;
          unlink "$dir/CVS/X-Copied" or die "remove $dir/CVS/X-Copied: $!";
        }
      }                         # foreach dir
      
      # Update source dir from repository. With CVS after 1.10, it is
      # possible to just update in $sourcebase; however 1.10 screws up
      # without any diagnostics at all in this case, so for
      # compatibility it is necessary to update in each root
      # separately.
      my $oldwd = cwd;
      my $root; foreach $root (@$cvsroots) {
        my $cleanroot = cleanroot ($root);
        my $top = "$sourcebase/$cleanroot";
        chdir $top or die "chdir $top failed: $!";
        #die "CVS binary $cvsexe not present/executable" unless -x $cvsexe;
	# XXX this is not very nice. Rather than using -A/-D/-r options to update,
	# we should run update with no args, and make sure the correct tags are
	# already present on relevant working directories. This would permit
	# someone to build from a partial branch covering only a few packages;
	# most of the build would be updated from the trunk (e.g.) but that
	# area of code would be updated from a private branch.
	# An alternative is to use the -f option to update, but this is not great
	# since it screws up file deletions in some cases.
        my $updatearg;
        if ($tagtype eq '-') {
          $updatearg = '-A';
        } elsif ($tagtype eq 'D') {
          $updatearg = "-D$tag";
        } else {
          $updatearg = "-r$tag";
        }
        $info->("Will run `cvs -q update $updatearg -d -P' in $top");
        # XXX should try to trap output from CVS and redirect to INFOLOG
        # and WARNLOG as appropriate
        system $cvsexe, '-q', 'update', $updatearg, '-d', '-P'
          and die "cvs update failed: exit status $ {\ ($? >> 8)}";
      }
      chdir $oldwd or warn "chdir $oldwd failed: $!";
      
    } else {
      $info->("Skipping CVS update upon request");
    }
    
  };                            # end of eval block to trap fatal errors

  if ($@) {
    my $time = localtime;
    $@ =~ s/\n$//;
    print $critlog "ERROR [$time] $@\n";
    return ();
  } else {
    return @sourcedirs;
  }
    
} # end of make_source_directories

# Given a CVSROOT, produces a cleansed version usable in a file name.
sub cleanroot ($;) {
  my ($cleanroot) = @_;
  $cleanroot =~ s/[^a-zA-Z_0-9.-]/_/g;
  $cleanroot =~ s/__+/_/g;
  $cleanroot =~ s/^_//;
  $cleanroot =~ s/_$//;
  die "no root left from cleaning `$_[0]'!" unless $cleanroot;
  return $cleanroot;
}

# Given a directory, a CVSROOT, and a repository path, and a tagtype
# and tag, and an informational logging callback, makes sure that the
# directory (1) exists and (2) has a CVS subdirectory. If it has to
# fix it, will use provided root and repository (and maybe tag) and
# leave its Entries blank. If CVSROOT/Emptydir is given as the path,
# it will clear the Entries file (assuming it holds only directory
# entries). Last optional arg in this case gives new directory entries
# to write.
sub makecvsdir ($$$$$$;\@) {
  my ($dir, $root, $repos, $tagtype, $tag, $info, $subdirs) = @_;
  local (*ROOT, *REPOS, *ENTRIES, *TAG);

  # Maybe clear useless directory entries:
  if ($repos eq 'CVSROOT/Emptydir' && -f "$dir/CVS/Entries") {
    my @entries = readentries ($dir);
    if (@entries) {
      if (grep {! m!^D!} @entries) {
        warn "will not clear $dir/CVS/Entries, it contains non-directory entries";
      } else {
        $info->("Clearing directory entries from dummy directory $dir");
        unlink "$dir/CVS/Entries" or die "remove $dir/CVS/Entries: $!";
        unlink "$dir/CVS/Entries.Log" or die "remove $dir/CVS/Entries.Log: $!"
          if -f "$dir/CVS/Entries.Log";
        $subdirs = [] unless defined $subdirs;
        open ENTRIES, ">$dir/CVS/Entries" or die "write $dir/CVS/Entries: $!";
        foreach (@$subdirs) {
          print ENTRIES "D/$_////\n";
        }
        print ENTRIES "D\n";
        close ENTRIES;
      }
    }
  }

  # Make admin files if needed:
  return if -d "$dir/CVS";
  $info->("Making CVS admin files in $dir ($root/$repos)");
  mkpath ("$dir/CVS");
  open ROOT, ">$dir/CVS/Root" or die "write $dir/CVS/Root: $!";
  print ROOT "$root\n";
  close ROOT;
  open REPOS, ">$dir/CVS/Repository" or die "write $dir/CVS/Repository: $!";
  print REPOS "$repos\n";
  close REPOS;
  open ENTRIES, ">$dir/CVS/Entries" or die "write $dir/CVS/Entries: $!";
  # No contents: no files yet, nothing known about subdirs.
  close ENTRIES;
  unless ($tagtype eq '-') {
    open TAG, ">$dir/CVS/Tag" or die "write $dir/CVS/Tag: $!";
    print TAG "$tagtype$tag\n";
    close TAG;
  }
}

# Given a directory that has a CVS/Entries file in it, reads contents
# and returns them. Merges in any changes from CVS/Entries.Log too.
sub readentries ($;) {
  my ($dir) = @_;
  local (*ENTRIES, *ENTRIESLOG);
  die "readentries must be called in an array context" unless wantarray;
  my @entries;
  open ENTRIES, "$dir/CVS/Entries" or die "open $dir/CVS/Entries: $!";
  while (<ENTRIES>) {
    chomp;
    push @entries, $_;
  }
  close ENTRIES;
  if (-f "$dir/CVS/Entries.Log") {
    open ENTRIESLOG, "$dir/CVS/Entries.Log" or die "open $dir/CVS/Entries.Log: $!";
    while (<ENTRIESLOG>) {
      chomp;
      if (/^A (.*)/) {
        push @entries, $1;
      } elsif (/^R (.*)/) {
        my $del = $1;
        @entries = grep {$_ ne $del} @entries;
      } else {
        warn "ignoring $dir/CVS/Entries.Log line: $_";
      }
    }
    close ENTRIESLOG;
  }
  my %entries;
  my $entry; foreach $entry (@entries) {
    if ($entries{$entry}++) {
      warn "$dir/CVS/Entries (and maybe Entries.Log) had a duplicate entry for `$entry'";
    }
  }
  return @entries;
}

1;
