#!/usr/bin/perl -w
#                 Sun Public License Notice
# 
# The contents of this file are subject to the Sun Public License
# Version 1.0 (the "License"). You may not use this file except in
# compliance with the License. A copy of the License is available at
# http://www.sun.com/
# 
# The Original Code is NetBeans. The Initial Developer of the Original
# Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
# Microsystems, Inc. All Rights Reserved.

# Creates list of hyperlinks to openide classes & members appropriate
# for documentation files residing in some subdirectory of
# org.openide
# (e.g. $DOCS/org/openide/filesystems/doc-files). Thus, you can
# cut and paste from here without trying to write out the whole thing
# by hand.

# Usage: create-links.pl ~/cvs/nbsrc filesystems/doc-files filesys-linx.html
# Unsorted output displayed to stderr, sorted output sent to named file.
use strict;
use File::Find;

@ARGV==3 or die;
my $srcdir=$ARGV[0];
my $subdir=$ARGV[1];
my $outfile=$ARGV[2];
my @prefix=qw(org openide);
my @subdir=(@prefix, split(m!/!, $subdir));
my $prefix_dots=join '.', @prefix;
my $prefix_slashes=join '/', @prefix;
my $javap='javap';

my $thisdir = $0;
$thisdir =~ s!^(.*/)[^/]+$!$1! or $thisdir = './';

open OUT, "| tee /dev/stderr | sort > $outfile" or die;

# API links.
find (sub {
        return unless -f and -r and /api\.html$/;
        my $filename = $File::Find::name;
        $filename =~ s/^\Q$thisdir\E// or die;
        $filename =~ s/\.html$// or die;
        my @currclass = split m!/!, $filename;
        my $link_base = make_link_base (\@currclass, []);
        $link_base = '' if $link_base eq $_;
#        print STDERR "Link base: $link_base\n";
        open API, $_ or die "$File::Find::name: $!";
        my $title = '<Unknown>';
        while (<API>) {
          if (m!<title>(.*)</title>!i) {
            $title = $1;
            print OUT qq<<a href="$link_base">$title</a>\n>;
          }
          if (m!<a name="(.*)"><h\d>(.*)</h\d>!i) {
            print OUT qq<<a href="$link_base#$1">$title: $2</a>\n>;
          }
        }
        close API;
}, $thisdir);

# Javadoc links.
find(sub {
       # Handle packages.
       if (-d) {
	 my $package=$File::Find::name;
	 $package =~ s!^\Q$srcdir\E/!! or die;
	 $package =~ s!/!.!g;
	 my @pack=split /\./, $package;
	 output(make_link_base([@pack, 'package-summary'], []), undef, $package);
	 return;
       }
       # Only want classfiles for top-level classes.
       return if !/\.class$/ or /\$/;
       local(*JAVAP);
       my $classname=$File::Find::name;
       $classname =~ s!^\Q$srcdir\E/!! or die;
       $classname =~ s/\.class$// or die;
       $classname =~ s!/!.!g;
#       print STDERR "Working on $classname\n";
#       return unless $classname eq 'org.openide.loaders.FileEntry';
       open JAVAP, "$javap -protected -classpath $srcdir $classname |" or die;
       my ($currclass, $currclassrel, $currinner, $currdispname, $currdispnamecompressed);
       while (<JAVAP>) {
	 chomp;
#	 print STDERR "\t\t$_\n";
	 # Javap indicates inner classes w/ an extra space.
	 if (/^ *(public|protected)?.*(class|interface) ([\w.]+\w)((\. \w+)+)?/) {
	   # Skip over non-public classes entirely.
	   my ($access, $classbase, $classinner)=($1, $3, $4);
	   close JAVAP and return unless defined $access;
	   $classinner ||= '';
	   # Class declaration.
	   $currclass=[split /\./, $classbase];
	   $currclassrel=[@$currclass[scalar(@prefix)..$#$currclass]];
#	   print STDERR "\tinnerdef: $classinner\n";
	   $currinner=[grep $_, split /\. /, $classinner];
	   $currdispname=join('.', @$currclass) . join('', map {". $_"} @$currinner);
           $currdispnamecompressed=join('.', @$currclass, @$currinner);
#	   print STDERR "\tclass: (@$currclass)\n";
#	   print STDERR "\trel: (@$currclassrel)\n";
#	   print STDERR "\tinner: (@$currinner)\n";
	   my $link=make_link_base($currclass, $currinner);
	   my $displayname=make_class_displayname($currclass, $currinner);
	   output($link, undef, $displayname);
	 } elsif (defined $currdispname and /(?:\Q$currdispname\E|\Q$currdispnamecompressed\E)\((.*)\)( throws .*)?;/) {
	   # Constructor.
	   my $args=$1 || '';
           my $rawargs = $args;
	   $args =~ s/\. |\$/./g;
	   $args =~ s/,(?=\S)/, /g; # why no spc after commas here?!
	   my $link=make_link_base($currclass, $currinner);
	   my $displayname='new ' . make_class_displayname($currclass, $currinner) .
	     '(' . make_args($rawargs) . ')';
	   my $anchor=join('.', $currclass->[$#$currclass], @$currinner) .
	     '(' . $args . ')';
	   output($link, $anchor, $displayname);
	 } elsif (/(\w+)\((.*)\)(\[\])*( throws .*)?;/) {
	   # Method declaration.
	   my $methodname=$1;
	   my $args=$2 || '';
           my $rawargs = $args;
	   # Inner classes.
	   $args =~ s/\. |\$/./g;
	   my $link=make_link_base($currclass, $currinner);
	   my $displayname=make_class_displayname($currclass, $currinner) .
	     '.' . $methodname . '(' . make_args($rawargs) . ')';
	   my $anchor=$methodname . '(' . $args . ')';
	   output($link, $anchor, $displayname);
	 } elsif (/(\w+)(\[\])*;/) {
	   # Field declaration.
	   my $fieldname=$1;
	   my $link=make_link_base($currclass, $currinner);
	   my $displayname=make_class_displayname($currclass, $currinner) .
	     '.' . $fieldname;
	   output($link, $fieldname, $displayname);
	 }
       }
       close JAVAP;
}, "$srcdir/$prefix_slashes");

close OUT;

sub output($$$;) {
  my ($href, $anchor, $displayname)=@_;
  my $target=defined($anchor) ? "$href#$anchor" : $href;
  print OUT qq{<a href="$target"><code>$displayname</code></a>\n};
}

sub make_link_base($$) {
  my ($currclass, $currinner)=@_;
#  print STDERR "currclass=(@$currclass) currinner=(@$currinner) subdir=(@subdir)\n";
  my $overlap; for ($overlap=0; $overlap < @subdir and $subdir[$overlap] eq $currclass->[$overlap]; $overlap++) {}
  my $dots_up='../' x (@subdir-$overlap);
  my $go_down=join '/', @$currclass[$overlap .. $#$currclass];
  $dots_up . $go_down .
    join('', map {".$_"} @$currinner) .
      '.html';
}

sub make_class_displayname($$) {
  my ($currclass, $currinner)=@_;
  join '.', $currclass->[$#$currclass], @$currinner;
}

sub make_args($;) {
    my ($args)=@_;
    if ($args eq '') {
        '';
    } elsif ($args =~ /,/) {
        '...';                  # '</code>...<code>' does not look good in Navigator, alas
    } else {
        $args =~ s/^.*\.//;
        $args =~ s/\$/./g;
        $args;
    }
}
