/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide;

/** A system of managing, annotating, and classifying errors and log messages.
 * Rather than printing raw exceptions to the console, or popping up dialog boxes
 * with exceptions, or implementing custom debug or logging facililities, code may
 * use the error manager to access logging and error-reporting in a higher-level
 * fashion. Standard error manager implementations can then provide generic ways
 * of customizing logging levels for different components, and so on.
 * <p>Especially important is the attaching of annotations such as stack traces to
 * exceptions, permitting you to throw an exception of a type permitted by your
 * API signature while safely encapsulating the root cause of the problem (in terms
 * of other nested exceptions). Code should use {@link #notify(Throwable)} rather
 * than directly printing caught exceptions, to make sure nested annotations are not lost.
 * <p>Also localized messages may be annotated to exceptions so that code which can deal
 * with a caught exception with a user-visible UI can display a polite and helpful message.
 * Messages with no localized annotation can be handled in a default way while the details
 * are reserved for the log file.
 * <p>A simple example of usage to keep nested stacktraces:
 * <pre>
 * public void doSomething () throws IOException {
 *     try {
 *         doSomethingElse ();
 *     } catch (IllegalArgumentException iae) {
 *         IOException ioe = new IOException ("did not work");
 *         TopManager.getDefault ().getErrorManager ().annotate (ioe, iae);
 *         throw ioe;
 *     }
 * }
 * // ...
 * try {
 *     foo.doSomething ();
 * } catch (IOException ioe) {
 *     TopManager.getDefault ().getErrorManager ().notify (ioe);
 * }
 * </pre>
 * @author Jaroslav Tulach
 * @see <a href="http://openide.netbeans.org/proposals/ErrorManagerUsage.html">Philosophy and usage scenarios</a>
 */
public abstract class ErrorManager extends Object {
    /** Undefined severity. */
    public static final int UNKNOWN = 0x00000000;
    /** Message that would be useful for tracing events but which need not be a problem. */
    public static final int INFORMATIONAL = 0x00000001;
    /** Something went wrong in the software, but it is continuing and the user need not be bothered. */
    public static final int WARNING = 0x00000010;
    /** Something the user should be aware of. */
    public static final int USER = 0x00000100;
    /** Something went wrong, though it can be recovered. */
    public static final int EXCEPTION = 0x00001000;
    /** Serious problem, application may be crippled. */
    public static final int ERROR = 0x00010000;

    /** Associates annotations with an exception.
    *
    * @param t the exception
    * @param arr array of annotations (or <code>null</code>)
    * @return the same exception <code>t</code> (as a convenience)
    */
    public abstract Throwable attachAnnotations (Throwable t, Annotation[] arr);

    /** Finds annotations associated with a given exception.
    * @param t the exception
    * @return array of annotations or <code>null</code>
    */
    public abstract Annotation[] findAnnotations (Throwable t);

    /** Annotates given exception with given values. All the
    * previous annotations are kept and this new one is added at 
    * the top of the annotation stack (index 0 of the annotation
    * array).
    *
    * @param t the exception
    * @param severity integer describing severity, e.g. {@link #EXCEPTION}
    * @param date date or <code>null</code>
    * @param message message to attach to the exception or <code>null</code>
    * @param localizedMessage localized message for the user or <code>null</code>
    * @param stackTrace exception representing the stack trace or <code>null</code>
    * @return the same exception <code>t</code> (as a convenience)
    */
    public abstract Throwable annotate (
        Throwable t, int severity,
        String message, String localizedMessage,
        Throwable stackTrace, java.util.Date date
    );

    /** Prints the exception to the log file and (possibly) notifies the user.
    * @param severity the severity to be applied to the exception (overrides default), e.g. {@link #EXCEPTION}
    * @param t the exception to notify
    */
    public abstract void notify (int severity, Throwable t);


    /** Prints the exception to the log file and (possibly) notifies the user.
     * Guesses at the severity.
    * @param t the exception to notify
    */
    public final void notify (Throwable t) {
        notify (UNKNOWN, t);
    }

    /** Logs the message to a file and (possibly) tells the user.
    * @param severity the severity to be applied (overrides default)
    * @param s the log message
    */
    public abstract void log(int severity, String s);

    /** Logs the message to log file and (possibly) tells the user.
     * Uses a default severity.
    * @param s the log message
    */
    public final void log(String s) {
        log(UNKNOWN, s);
    }
    
    /** Test whether a messages with given severity will be logged in advance.
     * Can be used to avoid the construction of complicated and expensive
     * logging messages.
     *
     * @param severity the severity to check, e.g. {@link #EXCEPTION}
     * @return <code>false</code> if the next call to {@link #log(int,String)} with this severity will
     *    discard the message
     */
    public boolean isLoggable (int severity) {
        return true;
    }

    /** Returns an instance with given name.
     * <p>By convention, you can name error managers the same as packages (or classes)
     * they are designed to report information from.
     * For example, <code>org.netbeans.modules.mymodule.ComplicatedParser</code>.
     * <p>The error manager implementation should provide some way of configuring e.g.
     * the logging level for error managers of different names. For example, in the basic
     * NetBeans core implementation, you can define a system property with the same name
     * as the future error manager (or a package prefix of it) whose value is the numeric
     * logging level (e.g. <samp>-J-Dorg.netbeans.modules.mymodule.ComplicatedParser=0</samp>
     * to log everything). Other implementations may have quite different ways of configuring
     * the error managers.
     * @param name the desired identifying name
     * @return a new error manager keyed off of that name
     */
    public abstract ErrorManager getInstance(String name);
    
    //
    // Helper methods
    //

    /** Annotates given exception with given values. All the
    * previous annotations are kept and this new is added at 
    * the top of the annotation stack (index 0 of the annotation
    * array).
    *
    * @param t the exception
    * @param localizedMessage localized message for the user or null
    * @return the same exception <code>t</code> (as a convenience)
    */
    public final Throwable annotate (
        Throwable t, String localizedMessage
    ) {
        return annotate (t, EXCEPTION, null, localizedMessage, null, null);
    }

    /** Annotates target exception with given exception. All the
    * previous annotations are kept and this new is added at 
    * the top of the annotation stack (index 0 of the annotation
    * array).
    *
    * @param target the exception to be annotated
    * @param t the exception that will be added
    * @return the same exception <code>target</code> (as a convenience)
    */    
    public final Throwable annotate (Throwable target, Throwable t) {
        return annotate (target, EXCEPTION, null, null, t, null);        
    }
    
    /** Takes annotations from one exception and associates
    * them with another one.
    *
    * @param t the exception to annotate
    * @param copyFrom exception to take annotations from
    * @return the same exception <code>t</code> (as a convenience)
    */
    public final Throwable copyAnnotation (Throwable t, Throwable copyFrom) {
        Annotation[] arr = findAnnotations (copyFrom);

        if (arr != null) {
            return attachAnnotations (
                       t, arr
                   );
        } else {
            return annotate (t, UNKNOWN, null, null, copyFrom, null);
        }
    }


    /* Adds annotations to this thread.
    *
    * @param arr array of annotations (or null)
    *
    public final void addAnnotations (Annotation[] arr) {
      Annotation[] old = findAnnotations ();
      
      java.util.LinkedList ll = new java.util.LinkedList ();
       
      ll.addAll (java.util.Arrays.asList(arr));
      if (old != null) {
        ll.addAll (java.util.Arrays.asList(old));
      }

      attachAnnotations ((Annotation[])ll.toArray (arr));
}
    */

    /** Annotation that can be attached to an error.
    */
    public static interface Annotation {
        /** Non-localized message.
        * @return associated message or <code>null</code>
        */
        public abstract String getMessage ();

        /** Localized message.
        * @return message to be presented to the user or <code>null</code>
        */
        public abstract String getLocalizedMessage ();

        /** Stack trace. The stack trace should locate the method
        * and position in the method where the error occurred.
        *
        * @return exception representing the location of the error or <code>null</code>
        */
        public abstract Throwable getStackTrace ();

        /** Time at which the exception occurred.
        * @return the time or <code>null</code>
        */
        public abstract java.util.Date getDate ();

        /** Severity of the exception.
        * @return number representing the severity, e.g. {@link ErrorManager#EXCEPTION}
        */
        public abstract int getSeverity ();
    } // end of Annotation
}
