/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide;

import java.awt.event.*;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.Image;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Component;
import java.awt.Insets;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.net.URL;
import java.lang.ref.WeakReference;

import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;
import org.openide.awt.HtmlBrowser;

/** Implements a basic "wizard" GUI system.
* A list of <em>wizard panels</em> may be specified and these
* may be traversed at the proper times using the "Previous"
* and "Next" buttons (or "Finish" on the last one).
*
* @author Ian Formanek, Jaroslav Tulach
*/
public class WizardDescriptor extends DialogDescriptor {

    /** "Next" button option.
    * @see #setOptions */
    public static final Object NEXT_OPTION = new Object ();
    /** "Finish" button option.
    * @see #setOptions */
    public static final Object FINISH_OPTION = OK_OPTION;
    /** "Previous" button option.
    * @see #setOptions */
    public static final Object PREVIOUS_OPTION = new Object ();

    /** real buttons to be placed instead of the options */
    private final JButton nextButton = new JButton ();
    private final JButton finishButton = new JButton ();
    private final JButton cancelButton = new JButton ();
    private final JButton previousButton = new JButton ();

    private static final ActionListener CLOSE_PREVENTER = new ActionListener () {
                public void actionPerformed (ActionEvent evt) {
                }
            };

    {
        // button init
        ResourceBundle b = NbBundle.getBundle ("org.openide.Bundle"); // NOI18N
        nextButton.setText (b.getString ("CTL_NEXT"));
        previousButton.setText (b.getString ("CTL_PREVIOUS"));
        finishButton.setText (b.getString ("CTL_FINISH"));
        cancelButton.setText (b.getString ("CTL_CANCEL"));
        previousButton.setMnemonic (b.getString ("CTL_PREVIOUS_Mnemonic").charAt(0));
        finishButton.setMnemonic (b.getString ("CTL_FINISH_Mnemonic").charAt(0));

        finishButton.setDefaultCapable (true);
        nextButton.setDefaultCapable (true);
        previousButton.setDefaultCapable (false);
        cancelButton.setDefaultCapable (false);
    }
    
    /** <CODE>Boolean</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to <CODE>true</CODE> for enabling other properties. It is relevant only on 
     * initialization (client property in first panel).
     * When false or not present in JComponent.getClientProperty(), then supplied panel is 
     * used directly without content, help or panel name auto layout.
     */
    private static final String PROP_AUTO_WIZARD_STYLE = "WizardPanel_autoWizardStyle"; // NOI18N
    /** <CODE>Boolean</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to <CODE>true</CODE> for showing help pane in the left pane. It is relevant only on 
     * initialization (client property in first panel).
     */
    private static final String PROP_HELP_DISPLAYED = "WizardPanel_helpDisplayed"; // NOI18N
    /** <CODE>Boolean</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to <CODE>true</CODE> for showing content pane in the left pane. It is relevant only on 
     * initialization (client property in first panel).
     */
    private static final String PROP_CONTENT_DISPLAYED = "WizardPanel_contentDisplayed"; // NOI18N
    /** <CODE>Boolean</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to <CODE>true</CODE> for displaying numbers in the content. It is relevant only on 
     * initialization (client property in first panel).
     */
    private static final String PROP_CONTENT_NUMBERED = "WizardPanel_contentNumbered"; // NOI18N
    /** <CODE>Integer</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Represents index of content item which will be highlited.
     */
    private static final String PROP_CONTENT_SELECTED_INDEX = "WizardPanel_contentSelectedIndex"; // NOI18N
    /** <CODE>String[]</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Represents array of content items.
     */
    private static final String PROP_CONTENT_DATA = "WizardPanel_contentData"; // NOI18N
    /** <CODE>Color</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to background color of content pane.
     */
    private static final String PROP_CONTENT_BACK_COLOR = "WizardPanel_contentBackColor"; // NOI18N
    /** <CODE>Color</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to foreground color of content pane.
     */
    private static final String PROP_CONTENT_FOREGROUND_COLOR = "WizardPanel_contentForegroundColor"; // NOI18N
    /** <CODE>Image</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to image which should be displayed in the left pane (behind the content).
     */
    private static final String PROP_IMAGE = "WizardPanel_image"; // NOI18N
    /** <CODE>String</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Set to side where the image should be drawn.
     */
    private static final String PROP_IMAGE_ALIGNMENT = "WizardPanel_imageAlignment"; // NOI18N
    /** <CODE>Dimension</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Dimension of left pane, should be same as dimension of <CODE>PROP_IMAGE</CODE>.
     * It is relevant only on initialization (client property in first panel).
     */
    private static final String PROP_LEFT_DIMENSION = "WizardPanel_leftDimension"; // NOI18N
    /** <CODE>URL</CODE> property. The value is taken from <CODE>WizardDescriptor.getProperty()</CODE> or
     * <CODE>((JComponent)Panel.getComponent()).getClientProperty()</CODE> in this order.
     * Represents URL of help displayed in left pane.
     */
    private static final String PROP_HELP_URL = "WizardPanel_helpURL"; // NOI18N

    /** Reference to default image */
    private static WeakReference defaultImage;
    
    /** Whether wizard panel will be constructed from <CODE>WizardDescriptor.getProperty()</CODE>/
     * <CODE>(JComponent)Panel.getComponent()</CODE> client properties or returned 
     * <CODE>Component</CODE> will be inserted to wizard dialog directly.
     */
    private boolean autoWizardStyle = false;
    /** Whether properties from first <CODE>(JComponent)Panel.getComponent()</CODE>
     * have been initialized.
     */
    private boolean init = false;
    /** Panel which is used when in <CODE>AUTO_WIZARD_STYLE</CODE> mode.*/
    private WizardPanel wizardPanel;
    /** Image */
    private Image image;
    /** Content data */
    private String[] contentData = new String[] {};
    /** Selected content index */
    private int contentSelectedIndex = -1;
    /** Background color*/
    private Color contentBackColor;
    /** Foreground color*/
    private Color contentForegroundColor;
    /** Help URL displayed in the left pane */
    private URL helpURL;
    /** Listener on a user component client property changes*/
    private PropL propListener;
    /** Whether to use default image or not */
    private boolean useDefaultImage = true;
    /** 'North' or 'South' */
    private String imageAlignment = "North"; // NOI18N
    
    /** Iterator between panels in the wizard */
    private Iterator panels;

    /** Change listener that invokes method update state */
    private Listener listener;

    /** current panel */
    private Panel current;

    /** settings to be used for the panels */
    private Object settings;

    /** message format to create title of the document */
    private MessageFormat titleFormat;

    /** hashtable with additional settings that is usually used
    * by Panels to store their data
    */
    private Map properties;

    /** Create a new wizard from a fixed list of panels, passing some settings to the panels.
    * @param wizardPanels the panels to use
    * @param settings the settings to pass to panels, or <code>null</code>
    * @see #WizardDescriptor(WizardDescriptor.Iterator, Object)
    */
    public WizardDescriptor (Panel[] wizardPanels, Object settings) {
        this (new ArrayIterator (wizardPanels), settings);
    }

    /** Create a new wizard from a fixed list of panels with settings
    * defaulted to <CODE>this</CODE>.
    *
    * @param wizardPanels the panels to use
    * @see #WizardDescriptor(WizardDescriptor.Iterator, Object)
    */
    public WizardDescriptor (Panel[] wizardPanels) {
        // passing CLOSE_PREVENTER which is treated especially
        this (wizardPanels, CLOSE_PREVENTER);
    }

    /** Create wizard for a sequence of panels, passing some settings to the panels.
    * @param panels iterator over all {@link WizardDescriptor.Panel}s that can appear in the wizard
    * @param settings the settings to provide to the panels (may be any data understood by them)
    * @see WizardDescriptor.Panel#readSettings
    * @see WizardDescriptor.Panel#storeSettings
    */
    public WizardDescriptor (Iterator panels, Object settings) {
        super ("", "", true, DEFAULT_OPTION, null, CLOSE_PREVENTER); // NOI18N
        this.settings = settings == CLOSE_PREVENTER ? this : settings;

        listener = new Listener ();

        nextButton.addActionListener (listener);
        previousButton.addActionListener (listener);
        finishButton.addActionListener (listener);
        cancelButton.addActionListener (listener);

        super.setOptions (new Object[] { previousButton, nextButton, finishButton, cancelButton });
        super.setClosingOptions (new Object[] { finishButton, cancelButton });

        this.panels = panels;
        panels.addChangeListener (listener);
        
    }

    /** Create wizard for a sequence of panels, with settings
    * defaulted to <CODE>this</CODE>.
    *
    * @param panels iterator over all {@link WizardDescriptor.Panel}s that can appear in the wizard
    */
    public WizardDescriptor (Iterator panels) {
        // passing CLOSE_PREVENTER which is treated especially
        this (panels, CLOSE_PREVENTER);
    }
    
    /** Initializes settings.
     */
    protected void initialize () {
        super.initialize ();

        updateState ();
    }

    /** Set a different list of panels.
    * Correctly updates the buttons.
    * @param panels the new list of {@link WizardDescriptor.Panel}s 
    */
    public final synchronized void setPanels (Iterator panels) {
        if (panels != null) {
            panels.removeChangeListener (listener);
        }
        this.panels = panels;
        panels.addChangeListener (listener);
        init = false;

        updateState ();
    }

    /** Set options permitted by the wizard considered as a <code>DialogDescriptor</code>.
    * Substitutes tokens such as {@link #NEXT_OPTION} with the actual button.
    *
    * @param options the options to set
    */
    public void setOptions (Object[] options) {
        super.setOptions (convertOptions (options));
    }

    /**
    * @param options the options to set
    */
    public void setAdditionalOptions (Object[] options) {
        super.setAdditionalOptions (convertOptions (options));
    }

    /**
    * @param options the options to set
    */
    public void setClosingOptions (Object[] options) {
        super.setClosingOptions (convertOptions (options));
    }

    /** Converts some options.
    */
    private Object[] convertOptions (Object[] options) {
        options = (Object[])options.clone ();
        for (int i = options.length - 1; i >= 0; i--) {
            if (options[i] == NEXT_OPTION) options[i] = nextButton;
            if (options[i] == PREVIOUS_OPTION) options[i] = previousButton;
            if (options[i] == FINISH_OPTION) options[i] = finishButton;
            if (options[i] == CANCEL_OPTION) options[i] = cancelButton;
        }
        return options;
    }
    /** Overriden to ensure that returned value is one of 
     * the XXX_OPTION constants.
     */
    public Object getValue() {
        return backConvertOption(super.getValue());
    }
    
    /** Converts the option back to one of the constants.
     * It is called from getValue().
     */
    private Object backConvertOption(Object op) {
        if (op == nextButton) {
            return NEXT_OPTION;
        }
        if (op == previousButton) {
            return PREVIOUS_OPTION;
        }
        if (op == finishButton) {
            return FINISH_OPTION;
        }
        if (op == cancelButton) {
            return CANCEL_OPTION;
        }
        // if we don't know just return the original value
        return op;
    }

    /** Sets the message format to create title of the wizard.
    * The format can take two parameters. The name of the
    * current component and the name returned by the iterator that
    * defines the order of panels. The default value is something
    * like
    * <PRE>
    *   {0} wizard {1}
    * </PRE>
    * That can be expanded to something like this
    * <PRE>
    *   EJB wizard (1 of 8)
    * </PRE>
    * This method allows anybody to provide own title format.
    * 
    * @param format message format to the title
    */
    public void setTitleFormat (MessageFormat format) {
        titleFormat = format;
        if (init)
            updateState ();
    }

    /** Getter for current format to be used to format title.
    * @return the format
    * @see #setTitleFormat
    */
    public MessageFormat getTitleFormat () {
        if (titleFormat == null) {
            synchronized (this) {
                if (titleFormat == null) {
                    // ok, initialize the default one
                    titleFormat = new MessageFormat (NbBundle.getBundle (
                                                         WizardDescriptor.class
                                                     ).getString ("CTL_WizardName"));
                }
            }
        }
        return titleFormat;
    }

    /** Allows Panels that use WizardDescriptor as settings object to
    * store additional settings into it.
    *
    * @param name name of the property
    * @param value value of property
    */
    public void putProperty (final String name, Object value) {
        synchronized (this) {
            if (properties == null) {
                properties = new HashMap (7);
            }
            properties.put (name, value);
        }
        if (propListener != null) {
            Mutex.EVENT.readAccess(new Runnable() {
                public void run() {
                    propListener.propertyChange(
                        new PropertyChangeEvent(this, name, null, null)
                    );
                }
            });
        }
    }

    /** Getter for stored property.
    * @param name name of the property
    * @return the value
    */
    public synchronized Object getProperty (String name) {
        return properties == null ? null : properties.get (name);
    }

    /** Updates buttons to reflect the current state of the panels.
    * Can be overridden by subclasses
    * to change the options to special values. In such a case use:
    * <p><code><PRE>
    *   super.updateState ();
    *   setOptions (...);
    * </PRE></code>
    */
    protected synchronized void updateState () {
        Panel p = panels.current ();
        // listeners on the panel
        if (current != p) {
            if (current != null) {
                // remove
                current.removeChangeListener (listener);
                current.storeSettings (settings);
            }
            // Hack - obtain current panel again
            // It's here to allow dynamic change of panels in wizard
            // (which can be done in storeSettings method)
            p = panels.current ();

            // add to new
            p.addChangeListener (listener);

            current = p;
            current.readSettings (settings);
        }

        boolean next = panels.hasNext ();
        boolean prev = panels.hasPrevious ();
        boolean valid = p.isValid ();

        nextButton.setEnabled (next && valid);
        previousButton.setEnabled (prev);
        finishButton.setEnabled (
            valid &&
            (!next || (current instanceof FinishPanel))
        );

        //    nextButton.setVisible (next);
        //    finishButton.setVisible (!next || (current instanceof FinishPanel));

        if (next) {
            setValue (nextButton);
        } else {
//            setValue (finishButton); 
        }

        setHelpCtx (p.getHelp ());

        java.awt.Component c = p.getComponent ();

        //initialize wizardPanel
        if (!init) {
            if (c instanceof JComponent) {
                autoWizardStyle = getBooleanProperty(
                    (JComponent)c, PROP_AUTO_WIZARD_STYLE);
                if (autoWizardStyle) {
                    wizardPanel = new WizardPanel(
                        getBooleanProperty((JComponent)c, PROP_CONTENT_DISPLAYED),
                        getBooleanProperty((JComponent)c, PROP_HELP_DISPLAYED),
                        getBooleanProperty((JComponent)c, PROP_CONTENT_NUMBERED),
                        getLeftDimension((JComponent)c)
                    );
                    initBundleProperties();
                }
            }
            if (propListener == null)
                propListener = new PropL();
            init = true;
        }
        //update wizardPanel
        if (wizardPanel != null) {
            Component oldComp = wizardPanel.getRightComponent();
            if (oldComp != null)
                oldComp.removePropertyChangeListener(propListener);
            if (c instanceof JComponent) {
                setPanelProperties((JComponent)c);
                wizardPanel.setContent(contentData);
                wizardPanel.setSelectedIndex(contentSelectedIndex);
                wizardPanel.setContentBackColor(contentBackColor);
                wizardPanel.setContentForegroundColor(contentForegroundColor);
                wizardPanel.setImage(image);
                wizardPanel.setImageAlignment(imageAlignment);
                wizardPanel.setHelpURL(helpURL);
                c.addPropertyChangeListener(propListener);
            }
            if (wizardPanel.getRightComponent() != c) {
                wizardPanel.setRightComponent(c);
                setMessage(wizardPanel);
                firePropertyChange(DialogDescriptor.PROP_MESSAGE, null, wizardPanel);
            }
        } else if (c != getMessage ())
            setMessage (c);

        String panelName = c.getName ();
        if (panelName == null) {
            panelName = ""; // NOI18N
        }
        Object[] args = {
            panelName,
            panels.name ()
        };
        MessageFormat mf = getTitleFormat ();

        if (autoWizardStyle)
            wizardPanel.setPanelName(mf.format(args));
        else {
            setTitle (mf.format (args));
        }
    }

    /** Tryes to get property from getProperty() if doesn't succeed then tryes at
     * supplied <CODE>JComponent</CODE>s client property.
     * @param c origin of property
     * @param s name of property
     * @return boolean property
     */
    private boolean getBooleanProperty(JComponent c, String s) {
        Object property = getProperty(s);
        if (property instanceof Boolean)
            return ((Boolean)property).booleanValue();
        property = c.getClientProperty(s);
        if (property instanceof Boolean)
            return ((Boolean)property).booleanValue();
        return false;
    }

    /** Tryes to get dimension of wizard panel's left pane from getProperty() 
     * if doesn't succeed then tryes at
     * supplied <CODE>JComponent</CODE>s client property.
     * @return <CODE>Dimension</CODE> dimension of wizard panel's left pane
     */
    private Dimension getLeftDimension(JComponent c) {
        Dimension leftDimension;
        Object property = c.getClientProperty(PROP_LEFT_DIMENSION);
        if (property instanceof Dimension)
            leftDimension = (Dimension)property;
        else
            leftDimension = new Dimension(198, 233);
        return leftDimension;
    }
    
    /** Tryes to get properties from getProperty() if doesn't succeed then tryes at
     * supplied <CODE>JComponent</CODE>s client properties and store them
     * to appropriate fields.
     * @param c origin of property
     * @param s name of property
     * @return boolean property
     */
    private void setPanelProperties(JComponent c) {
        // TODO: Method should be devided into individual setter/getter methods !?
        Object property = getProperty(PROP_CONTENT_SELECTED_INDEX);
        if (property instanceof Integer)
            contentSelectedIndex = ((Integer)property).intValue();
        else {
            property = c.getClientProperty(PROP_CONTENT_SELECTED_INDEX);
            if (property instanceof Integer)
                contentSelectedIndex = ((Integer)property).intValue();
        }
        property = getProperty(PROP_CONTENT_DATA);
        if (property instanceof String[])
            contentData = (String[])property;
        else {
            property = c.getClientProperty(PROP_CONTENT_DATA);
            if (property instanceof String[])
                contentData = (String[])property;
        }
        property = getProperty(PROP_IMAGE);
        if (property instanceof Image) {
            image = (Image)property;
        } else if ((properties == null) || (!properties.containsKey(PROP_IMAGE))) {
            property = c.getClientProperty(PROP_IMAGE);
            if (property instanceof Image)
                image = (Image)property;
            else if (image == null)
                useDefaultImage = true;
        } else {
            useDefaultImage = false;
        }
        property = getProperty(PROP_IMAGE_ALIGNMENT);
        if (property instanceof String) {
            imageAlignment = (String)property;
        } else {
            property = c.getClientProperty(PROP_IMAGE_ALIGNMENT);
            if (property instanceof String)
                imageAlignment = (String)property;
        }
        property = getProperty(PROP_CONTENT_BACK_COLOR);
        if (property instanceof Color)
            contentBackColor = (Color)property;
        else {
            property = c.getClientProperty(PROP_CONTENT_BACK_COLOR);
            if (property instanceof Color)
                contentBackColor = (Color)property;
        }
        property = getProperty(PROP_CONTENT_FOREGROUND_COLOR);
        if (property instanceof Color)
            contentForegroundColor = (Color)property;
        else {
            property = c.getClientProperty(PROP_CONTENT_FOREGROUND_COLOR);
            if (property instanceof Color)
                contentForegroundColor = (Color)property;
        }
        property = c.getClientProperty(PROP_HELP_URL);
        if (property instanceof URL)
            helpURL = (URL)property;
        else if (property == null)
            helpURL = null;
    }
    
    private void initBundleProperties() {
        contentBackColor = new Color(getIntFromBundle("INT_WizardBackRed"), // NOI18N
                                     getIntFromBundle("INT_WizardBackGreen"), // NOI18N
                                     getIntFromBundle("INT_WizardBackBlue")); // NOI18N

        contentForegroundColor = new Color(getIntFromBundle("INT_WizardForegroundRed"), // NOI18N
                                           getIntFromBundle("INT_WizardForegroundGreen"), // NOI18N
                                           getIntFromBundle("INT_WizardForegroundBlue")); // NOI18N
        imageAlignment = bundle.getString("STRING_WizardImageAlignment"); //NOI18N
    }
    
    ResourceBundle bundle;
    
    private int getIntFromBundle(String key) {
        if (bundle == null)
            bundle = NbBundle.getBundle(WizardDescriptor.class);
        return Integer.parseInt(bundle.getString(key));
    }

    private static Image getDefaultImage() {
        Image img = null;
        if (defaultImage != null)
            img = (Image)defaultImage.get();
        if (img == null) {
            java.net.URL url = NbBundle.getLocalizedFile(
                 "org.openide.resources.defaultWizard", // NOI18N
                 "gif", // NOI18N
                 Locale.getDefault(),
                 WizardDescriptor.class.getClassLoader()
             );

            img = url == null ? null : java.awt.Toolkit.getDefaultToolkit().getImage(url);
            if (img != null)
                defaultImage = new WeakReference(img);
        }
        return img;
    }

    /** Iterator on the sequence of panels.
    * @see WizardDescriptor.Panel
    */
    public interface Iterator {
        /** Get the current panel.
        * @return the panel
        */
        public Panel current ();

        /** Get the name of the current panel.
        * @return the name
        */
        public String name ();

        /** Test whether there is a next panel.
        * @return <code>true</code> if so
        */
        public boolean hasNext ();

        /** Test whether there is a previous panel.
        * @return <code>true</code> if so
        */
        public boolean hasPrevious ();

        /** Move to the next panel.
        * I.e. increment its index, need not actually change any GUI itself.
        * @exception NoSuchElementException if the panel does not exist
        */
        public void nextPanel ();

        /** Move to the previous panel.
        * I.e. decrement its index, need not actually change any GUI itself.
        * @exception NoSuchElementException if the panel does not exist
        */
        public void previousPanel ();

        /** Add a listener to changes of the current panel.
        * The listener is notified when the possibility to move forward/backward changes.
        * @param l the listener to add
        */
        public void addChangeListener (ChangeListener l);

        /** Remove a listener to changes of the current panel.
        * @param l the listener to remove
        */
        public void removeChangeListener (ChangeListener l);
    }

    /** One wizard panel with a component on it. */
    public interface Panel {
        /** Get the component displayed in this panel.
        * @return the component
        */
        public java.awt.Component getComponent ();

        /** Help for this panel.
        * When the panel is active, this is used as the help for the wizard dialog.
        * @return the help or <code>null</code> if no help is supplied
        */
        public HelpCtx getHelp ();

        /** Provides the wizard panel with the current data--either
        * the default data or already-modified settings, if the user used the previous and/or next buttons.
        * This method can be called multiple times on one instance of <code>WizardDescriptor.Panel</code>.
        * @param settings the object representing wizard panel state, as originally supplied to {@link WizardDescriptor#WizardDescriptor(WizardDescriptor.Iterator,Object)}
        * @exception IllegalStateException if the the data provided 
        * by the wizard are not valid.
        */
        public void readSettings (Object settings);

        /** Provides the wizard panel with the opportunity to update the
        * settings with its current customized state.
        * Rather than updating its settings with every change in the GUI, it should collect them,
        * and then only save them when requested to by this method.
        * Also, the original settings passed to {@link #readSettings} should not be modified (mutated);
        * rather, the (copy) passed in here should be mutated according to the collected changes.
        * This method can be called multiple times on one instance of <code>WizardDescriptor.Panel</code>.
        * @param settings the object representing a settings of the wizard
        */
        public void storeSettings (Object settings);

        /** Test whether the panel is finished and it is safe to proceed to the next one.
        * If the panel is valid, the "Next" (or "Finish") button will be enabled.
        * @return <code>true</code> if the user has entered satisfactory information
        */
        public boolean isValid ();

        /** Add a listener to changes of the panel's validity.
        * @param l the listener to add
        * @see #isValid
        */
        public void addChangeListener (ChangeListener l);

        /** Remove a listener to changes of the panel's validity.
        * @param l the listener to remove
        */
        public void removeChangeListener (ChangeListener l);
    }

    /** A special interface for panels in middle of the
    * iterators path that would like to have the finish button
    * enabled. So both Next and Finish are enabled on panel
    * implementing this interface.
    */
    public interface FinishPanel extends Panel {
    }

    /** Special iterator that works on an array of <code>Panel</code>s.
    */
    public static class ArrayIterator extends Object implements Iterator {
        /** Array of items.
        */
        private Panel[] panels;

        /** Index into the array
        */
        private int index;

        /* Default constructor. It's here to allow subclasses to
        * be serializable easily. Panel initialization is done
        * through initializePanels() protected method. */
        public ArrayIterator () {
            panels = initializePanels();
            index = 0;
        }

        /** Construct an iterator.
        * @param array the list of panels to use
        */
        public ArrayIterator (Panel[] array) {
            panels = array;
            index = 0;
        }

        /** Allows subclasses to initialize their arrays of panels when
        * constructed using default constructor.
        * (for example during deserialization.
        * Default implementation returns empty array. */
        protected Panel[] initializePanels () {
            return new Panel[0];
        }

        /* The current panel.
        */
        public Panel current () {
            return panels[index];
        }

        /* Current name of the panel */
        public String name () {
            Object[] args = {
                new Integer (index + 1),
                new Integer (panels.length)
            };
            MessageFormat mf = new MessageFormat (NbBundle.getBundle (WizardDescriptor.class).getString ("CTL_ArrayIteratorName"));
            return mf.format (args);
        }

        /* Is there a next panel?
        * @return true if so
        */
        public boolean hasNext () {
            return index < panels.length - 1;
        }

        /* Is there a previous panel?
        * @return true if so
        */
        public boolean hasPrevious () {
            return index > 0;
        }

        /* Moves to the next panel.
        * @exception NoSuchElementException if the panel does not exist
        */
        public synchronized void nextPanel () {
            if (index + 1 == panels.length) throw new java.util.NoSuchElementException ();
            index++;
        }

        /* Moves to previous panel.
        * @exception NoSuchElementException if the panel does not exist
        */
        public synchronized void previousPanel () {
            if (index == 0) throw new java.util.NoSuchElementException ();
            index--;
        }

        /* Ignores the listener, there are no changes in order of panels.
        */
        public void addChangeListener (ChangeListener l) {
        }

        /* Ignored.
        */
        public void removeChangeListener (ChangeListener l) {
        }

        /** Resets this iterator to initial state.
        * Called by subclasses when they need re-initialization of the iterator.
        */
        protected void reset () {
            index = 0;
        }

    }

    /** Listener to changes in the iterator and panels.
    */
    private final class Listener extends Object
        implements ChangeListener, ActionListener {
        /** Change in the observed objects */
        public void stateChanged (ChangeEvent ev) {
            updateState ();
        }
        /** Action listener */
        public void actionPerformed (ActionEvent ev) {
            if (ev.getSource () == nextButton) {
                panels.nextPanel ();
                try {
                    updateState ();
                } catch (IllegalStateException ise) {
                    panels.previousPanel();
                    TopManager.getDefault().notify(new NotifyDescriptor.Message(ise.getMessage()));
                    updateState();
                }
            }

            if (ev.getSource () == previousButton) {
                panels.previousPanel ();
                updateState ();
            }

            if (ev.getSource () == finishButton) {
                current.storeSettings (settings);
                setValue (OK_OPTION);
            }

            if (ev.getSource () == cancelButton) {
                current.storeSettings (settings);
                setValue (CANCEL_OPTION);
            }
        }
    }
    
    /** Listenes on a users client property changes
     */
    private class PropL implements PropertyChangeListener {
        /** Accepts client property changes of user component */
        public void propertyChange(PropertyChangeEvent e) {
            if (wizardPanel == null)
                return;
            String propName = e.getPropertyName();
            setPanelProperties((JComponent)wizardPanel.getRightComponent());
            if (propName.equals(PROP_CONTENT_DATA))
                wizardPanel.setContent(contentData);
            else if (propName.equals(PROP_CONTENT_SELECTED_INDEX))
                wizardPanel.setSelectedIndex(contentSelectedIndex);
            else if (propName.equals(PROP_CONTENT_BACK_COLOR))
                wizardPanel.setContentBackColor(contentBackColor);
            else if (propName.equals(PROP_CONTENT_FOREGROUND_COLOR))
                wizardPanel.setContentForegroundColor(contentForegroundColor);
            else if (propName.equals(PROP_IMAGE))
                wizardPanel.setImage(image);
            else if (propName.equals(PROP_IMAGE_ALIGNMENT))
                wizardPanel.setImageAlignment(imageAlignment);
            else if (propName.equals(PROP_HELP_URL))
                wizardPanel.setHelpURL(helpURL);
        }
    }

    /** Panel which paints image as its background.
     */
    private static class ImagedPanel extends JComponent {
        Image image;
        boolean isDefault = false;
        boolean north = true;

        public ImagedPanel(Image im) {
            setImage(im);
            setLayout(new BorderLayout());
            setOpaque(true);
        }

        protected void paintComponent(Graphics graphics) {
            graphics.setColor(getBackground());
            graphics.fillRect(0, 0, getWidth(), getHeight());
            if (image != null) {
                graphics.drawImage(image, 0, north ? 0 : (getHeight() - image.getHeight(null)), null);
            }
        }
        
        public void setImageAlignment(String align) {
            north = "North".equals(align); // NOI18N
        }

        public void setImage(Image im) {
            if (im != null) {
                image = im;
                isDefault = false;
                return;
            }
            if (!isDefault) {
                image = new ImageIcon(getDefaultImage()).getImage();
                isDefault = true;
            }
        }
    }

    /** Text list cell renderer. Wraps text of items at specified width. Allows numbering
     * of items.
     */
    private static class WrappedCellRenderer extends JPanel implements ListCellRenderer {
        
        JTextArea ta = new JTextArea();
        JLabel numberLabel;
        int selected = -1;
        
        /**
         * @param contentNumbered Whether content will be numbered
         * @param wrappingWidth Width of list item.
         */
        private WrappedCellRenderer(boolean contentNumbered, int wrappingWidth) {
            super(new BorderLayout());
            ta.setOpaque(false);
            ta.setEditable(false);
            ta.setLineWrap(true);
            ta.setWrapStyleWord(true);
            ta.setFont(UIManager.getFont("Label.font"));
            
            int taWidth = wrappingWidth - 12 - 12;
            
            if (contentNumbered) {
                numberLabel = new JLabel();
                numberLabel.setHorizontalAlignment(SwingConstants.LEFT);
                numberLabel.setVerticalAlignment(SwingConstants.TOP);
                numberLabel.setFont(ta.getFont());
                numberLabel.setOpaque(false);
                numberLabel.setPreferredSize(new Dimension(25, 0));
                add(numberLabel, BorderLayout.WEST);            
                taWidth -= 25;
            }
            Insets taInsets = ta.getInsets();
            ta.setSize(taWidth, 
                       taInsets.top + taInsets.bottom + 1);

            add(ta, BorderLayout.CENTER);
            setBorder(BorderFactory.createEmptyBorder(0, 12, 0, 12));
            setOpaque(false);
        }
        
        public Component getListCellRendererComponent(
            JList list,
            Object value,
            int index,
            boolean isSelected,
            boolean cellHasFocus)
        {
            if (numberLabel != null) {
                numberLabel.setText(Integer.toString(index + 1) + ".");
            }
            if (index == selected) {
                ta.setFont(ta.getFont().deriveFont(Font.BOLD));
                if (numberLabel != null)
                    numberLabel.setFont(numberLabel.getFont().deriveFont(Font.BOLD));
            } else {
                ta.setFont(ta.getFont().deriveFont(Font.PLAIN));
                if (numberLabel != null)
                    numberLabel.setFont(numberLabel.getFont().deriveFont(Font.PLAIN));
            }
            ta.setText((String)value);
            return this;
        }
        
        private void setSelectedIndex(int i) {
            selected = i;
        }

        private void setForegroundColor(Color color) {
            if (numberLabel != null) {
                numberLabel.setForeground(color);
                numberLabel.setBackground(color);
            }
            ta.setForeground(color);
        }
    }

    /** Wizard panel. Allows auto layout of content, wizard panel name and input panel.
     */
    private static class WizardPanel extends JPanel {
        
        /** Users panel is inserted into this panel. */
        private JPanel rightPanel = new JPanel(new BorderLayout());
        /** Name of the users panel. */
        private JLabel panelName = new JLabel("Step"); //NOI18N
        /** List of content. */
        private JList contentList;
        /** Users component. Should be held for removing from rightPanel */
        private Component rightComponent;
        /** Panel which paints image */
        private ImagedPanel contentPanel;
        /** Name of content. Can be switched off.  */
        private JPanel contentLabelPanel;
        /** Wrapped list cell renderer */
        private WrappedCellRenderer cellRenderer;
        /** Tabbed pane is used only when both content and help are displayed */
        private JTabbedPane tabbedPane;
        /** HTML Browser is used only when help is displayed in the left pane */
        private HtmlBrowser htmlBrowser;
        /** Each wizard panel have to be larger or same as this */
        private Dimension cachedDimension = new Dimension(600, 365);
        /** Label of steps pane */
        private JLabel label;
        
        /** Creates new <CODE>WizardPanel<CODE>.
         * @param contentDisplayed whether content will be displayed in the left pane
         * @param helpDisplayed whether help will be displayed in the left pane
         * @param contentNumbered whether content will be numbered
         * @param leftDimension dimension of content or help pane
         */
        public WizardPanel(boolean contentDisplayed, boolean helpDispalyed,
                           boolean contentNumbered, Dimension leftDimension) {
            super(new BorderLayout());
            initComponents(contentDisplayed, helpDispalyed,
                           contentNumbered, leftDimension);
            setOpaque(false);
        }
        
        private void initComponents(boolean contentDisplayed, boolean helpDisplayed,
                                    boolean contentNumbered, Dimension leftDimension) {
            if (contentDisplayed) {
                createContentPanel(contentNumbered, leftDimension);
                if (!helpDisplayed)
                    add(contentPanel, BorderLayout.WEST);
            }
            if (helpDisplayed) {
                htmlBrowser = new BoundedHtmlBrowser(leftDimension);
                htmlBrowser.setPreferredSize(leftDimension);
                if (!contentDisplayed)
                    add(htmlBrowser, BorderLayout.WEST);
            }
            if (helpDisplayed && contentDisplayed) {
                tabbedPane = new JTabbedPane(JTabbedPane.BOTTOM);
                tabbedPane.addTab(NbBundle.getBundle(WizardDescriptor.class)
                                               .getString("CTL_ContentName"),
                                  contentPanel);
                tabbedPane.addTab(NbBundle.getBundle(WizardDescriptor.class)
                                               .getString("CTL_HelpName"),
                                  htmlBrowser);
                tabbedPane.setEnabledAt(1, false);
                tabbedPane.setOpaque(false);
//                tabbedPane.setPreferredSize(leftDimension);
                add(tabbedPane, BorderLayout.WEST);
            }
            panelName.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0,
                panelName.getForeground()));
            panelName.setFont(panelName.getFont().deriveFont(Font.BOLD));
            JPanel labelPanel = new JPanel(new BorderLayout());
            labelPanel.add(panelName, BorderLayout.NORTH);
            labelPanel.setBorder(BorderFactory.createEmptyBorder(12, 12, 12, 2));
            rightPanel.setBorder(BorderFactory.createEmptyBorder(0, 12, 11, 11));

            JPanel fullRightPanel = new JPanel(new BorderLayout());
            fullRightPanel.add(labelPanel, BorderLayout.NORTH);
            fullRightPanel.add(rightPanel, BorderLayout.CENTER);
            JSeparator sep = new JSeparator();
            sep.setForeground(Color.darkGray);

            add(fullRightPanel, BorderLayout.CENTER);
            add(sep, BorderLayout.SOUTH);
        }
        /** Creates content panel.
         * @param contentNumbered <CODE>boolean</CODE> whether content will be numbered
         * @param leftDimension <CODE>Dimension</CODE> dimension of content pane
         */
        private void createContentPanel(boolean contentNumbered, Dimension leftDimension) {
            contentList = new JList();
            cellRenderer = new WrappedCellRenderer(contentNumbered,
                                                   leftDimension.width);
            cellRenderer.setOpaque(false);
            contentList.setCellRenderer(cellRenderer);
            contentList.setOpaque(false);
            contentList.setEnabled(false);

            JScrollPane scroll = new JScrollPane(contentList);
            scroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
            scroll.getViewport().setOpaque(false);
            scroll.setBorder(null);
            scroll.setOpaque(false);

            label = new JLabel(NbBundle.getBundle(WizardDescriptor.class)
                                    .getString("CTL_ContentName"));
            label.setForeground(Color.white);
            label.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, label.getForeground()));
            label.setFont(label.getFont().deriveFont(Font.BOLD));
            contentLabelPanel = new JPanel(new BorderLayout());
            contentLabelPanel.setBorder(BorderFactory.createEmptyBorder(12, 12, 11, 11));
            contentLabelPanel.setOpaque(false);
            contentLabelPanel.add(label, BorderLayout.NORTH);
            
            contentPanel = new ImagedPanel(null);
            contentPanel.add(contentLabelPanel, BorderLayout.NORTH);
            contentPanel.add(scroll, BorderLayout.CENTER);

            contentPanel.setPreferredSize(leftDimension);
        }
        /** Setter for lists items. 
         * @param content Array of list items.
         */
        public void setContent(String[] content) {
            if (contentList != null) {
                contentList.setListData(content);
                contentList.revalidate();
                contentList.repaint();
                contentLabelPanel.setVisible(content.length > 0);
            }
        }
        /** Setter for selected list item. 
         * @param index Index of selected item in the list.
         */
        public void setSelectedIndex(int index) {
            if (cellRenderer != null) {
                cellRenderer.setSelectedIndex(index);
                contentList.ensureIndexIsVisible(index);
                // Fix of #10787.
                // This is workaround for swing bug - BasicListUI doesn't ask for preferred
                // size of rendered list cell as a result of property selectedIndex change. 
                // It does only on certain JList property changes (e.g. fixedCellWidth).
                // Maybe subclassing BasicListUI could be better fix.
                contentList.setFixedCellWidth(0);
                contentList.setFixedCellWidth(-1);
            }
        }
        /** Setter for content background color. 
         * @param color content background color.
         */
        public void setContentBackColor(Color color) {
            if (contentPanel != null)
                contentPanel.setBackground(color);
        }
        /** Setter for content foreground color. 
         * @param color content foreground color.
         */
        public void setContentForegroundColor(Color color) {
            if (cellRenderer == null)
                return;
            cellRenderer.setForegroundColor(color);
            label.setForeground(color);
            label.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, label.getForeground()));
        }
        /** Setter for content background image. 
         * @param image content background image
         */
        public void setImage(Image image) {
            if (contentPanel != null)
                contentPanel.setImage(image);
        }
        /** Setter for image alignment. 
         * @param align image alignment - 'North', 'South'
         */
        public void setImageAlignment(String align) {
            if (contentPanel != null)
                contentPanel.setImageAlignment(align);
        }
        /** Setter for user's component. 
         * @param c user's component
         */
        public void setRightComponent(Component c) {
            if (rightComponent != null)
                rightPanel.remove(rightComponent);
            rightComponent = c;
            rightPanel.add(rightComponent, BorderLayout.CENTER);
        }
        /** Getter for user's component. 
         * @return <CODE>Component</CODE> user's component
         */
        public Component getRightComponent() {
            return rightComponent;
        }
        /** Setter for wizard panel name. 
         * @param name panel name
         */
        public void setPanelName(String name) {
            panelName.setText(name);
        }
        /** Setter for help URL. 
         * @param helpURL help URL
         */
        public void setHelpURL(URL helpURL) {
            if (htmlBrowser == null)
                return;
            if (helpURL != null) {
                if (!helpURL.equals(htmlBrowser.getDocumentURL()))
                    htmlBrowser.setURL(helpURL);
                if (tabbedPane != null)
                    tabbedPane.setEnabledAt(tabbedPane.indexOfComponent(htmlBrowser),
                                            true);
            } else if (tabbedPane != null){
                tabbedPane.setSelectedComponent(contentPanel);
                tabbedPane.setEnabledAt(tabbedPane.indexOfComponent(htmlBrowser),
                                        false);
            }
        }
        
        public Dimension getPreferredSize() {
            Dimension dim = super.getPreferredSize();
            if (dim.height > cachedDimension.height)
                cachedDimension.height = dim.height;
            if (dim.width > cachedDimension.width)
                cachedDimension.width = dim.width;
            return cachedDimension;
        }
        
        /** Overriden to delegate call to user component.
         */
        public void requestFocus() {
            if (rightComponent != null)
                rightComponent.requestFocus();
            else
                super.requestFocus();
        }
    }

    /** Overriden to return wished preferred size */
    private static class BoundedHtmlBrowser extends HtmlBrowser {
        Dimension dim;
        boolean firstPage = true;
        
        public BoundedHtmlBrowser(Dimension d) {
            super(false, false);
            dim = d;
        }
        
        public Dimension getPreferredSize() {
            return dim;
        }
    }
}

/*
* Log
*  20   Jaga      1.18.1.0    3/29/00  David Simonek   dynamic panel changing 
*       bugfix, ArrayIterator modified to support better serialization of its 
*       subclasses
*  19   src-jtulach1.18        1/13/00  Ian Formanek    NOI18N
*  18   src-jtulach1.17        1/8/00   Jaroslav Tulach 
*  17   src-jtulach1.16        1/8/00   Jaroslav Tulach 
*  16   src-jtulach1.15        12/15/99 Jesse Glick     There is a now a Help 
*       button automatically added to any dialog (though not notification 
*       dialogs) which provides help either explicitly or on its inner 
*       component.
*  15   src-jtulach1.14        12/9/99  Jaroslav Tulach Can be used more than 
*       once.
*  14   src-jtulach1.13        11/25/99 Jaroslav Tulach Both Next & Finish are 
*       always visible.
*  13   src-jtulach1.12        11/24/99 Jaroslav Tulach New "New From Template" 
*       Dialog
*  12   src-jtulach1.11        11/4/99  Jaroslav Tulach DialogDescriptor.setClosingOptions
*        
*  11   src-jtulach1.10        10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  10   src-jtulach1.9         8/10/99  Martin Ryzl     isValid and readSettings 
*       order in updateState changed
*  9    src-jtulach1.8         6/24/99  Jesse Glick     Gosh-honest HelpID's.
*  8    src-jtulach1.7         6/11/99  Slavek Psenicka storeSettings was not 
*       called at last pane
*  7    src-jtulach1.6         6/8/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  6    src-jtulach1.5         6/8/99   Ian Formanek    oved here from package 
*       awt
*  5    src-jtulach1.4         6/8/99   Ian Formanek    Close on button press 
*       prevented
*  4    src-jtulach1.3         6/4/99   Jaroslav Tulach Improved NbDialog.
*  3    src-jtulach1.2         6/3/99   Ian Formanek    
*  2    src-jtulach1.1         5/15/99  Jesse Glick     [JavaDoc]
*  1    src-jtulach1.0         3/20/99  Jaroslav Tulach 
* $
*/
