/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.actions;

import java.util.*;

import org.openide.TopManager;
import org.openide.cookies.CompilerCookie;
import org.openide.cookies.ExecCookie;
import org.openide.compiler.CompilerTask;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.Compiler;

import org.openide.execution.Executor;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.Mutex;
import org.openide.util.actions.CookieAction;
import org.openide.util.enum.ArrayEnumeration;
import org.openide.windows.WindowManager;
import org.openide.windows.Workspace;

/** Execute a class.
* Is enabled if the only selected node implements
* {@link ExecCookie}.
* @see org.openide.execution
*
* @author   Ian Formanek, Jaroslav Tulach, Jan Jancura
*/
public class ExecuteAction extends CookieAction {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -1861936644244723970L;
    // static ..........................................................................................

    private static String workspace = "Running"; // NOI18N

    /** should we run compilation before execution */
    private static boolean runCompilation = true;

    /** Switches to running workspace */
    private static void changeWorkspace () {
        WindowManager dp = TopManager.getDefault ().getWindowManager ();
        Workspace d = dp.findWorkspace (workspace);
        if (d != null) d.activate ();
    }

    /**
    * Get the name of the workspace in which execution is performed.
    * By default, execution is performed in the "running" workspace.
    * @return the workspace name
    */
    public static String getWorkspace () {
        return workspace;
    }

    /**
    * Set the name of the workspace in which execution is performed.
    * @param workspace the workspace name
    */
    public static void setWorkspace (String workspace) {
        ExecuteAction.workspace = workspace;
    }

    /** Set whether files should be compiled before execution.
    * @param run <code>true</code> if they should
    */
    public static void setRunCompilation (boolean run) {
        runCompilation = run;
    }

    /** Test whether files will be compiled before execution.
    * By default they will.
    * @return <code>true</code> if they will be
    */
    public static boolean getRunCompilation () {
        return runCompilation;
    }

    // init ..........................................................................................

    /* Needs exec cookie.
    */
    protected Class[] cookieClasses () {
        return new Class[] { ExecCookie.class };
    }

    /* Checks the cookies and starts them.
    */
    protected void performAction (final Node[] activatedNodes) {
        execute (activatedNodes, runCompilation, true);
    }

    /* Mode any is enough.
    */
    protected int mode () {
        return MODE_ANY;
    }

    /* Human presentable name of the action. This should be
    * presented as an item in a menu.
    * @return the name of the action
    */
    public String getName() {
        return ActionConstants.BUNDLE.getString("Execute");
    }

    /* Help context where to find more about the acion.
    * @return the help context for this action
    */
    public HelpCtx getHelpCtx() {
        return new HelpCtx (ExecuteAction.class);
    }

    protected String iconResource () {
        return "/org/openide/resources/actions/execute.gif"; // NOI18N
    }

    // utility methods

    /** Execute a list of items by cookie.
    *
    * @param execCookies list of {@link ExecCookie}s (any may be <code>null</code>)
    */
    public static void execute(Iterator execCookies) {
        while (execCookies.hasNext()) {
            ExecCookie cookie = (ExecCookie) execCookies.next();
            if (cookie != null) {
                cookie.start();
            }
        }
    }

    /** Execute some data objects.
    *
    * @param dataObjects the data objects (should have {@link ExecCookie} on them if they are to be used)
    * @param compileBefore <code>true</code> to compile before executing
    * @return true if compilation succeeded or was not performed, false if compilation failed
    */
    public static boolean execute(DataObject[] dataObjects, boolean compileBefore) {
        // search all DataObjects with unique ExecCookies/StartCookies -
        // - it is possible, that multiple activated nodes have the same exec cookie and
        // we have to prevent running it multiple times
        HashSet compile = new HashSet ();
        HashSet execute = new HashSet ();

        for (int i = 0; i < dataObjects.length; i++) {
            ExecCookie exec = (ExecCookie) dataObjects[i].getCookie(ExecCookie.class);
            if (exec != null) {
                execute.add(exec);
                if (compileBefore) {
                    CompilerCookie comp = (CompilerCookie) dataObjects[i].getCookie(CompilerCookie.Compile.class);
                    if (comp != null) {
                        compile.add(comp);
                    }
                }
            }
        }
        // compile
        if (compileBefore) {
            if (! AbstractCompileAction.compile(Collections.enumeration(compile),
                                                AbstractCompileAction.findName(dataObjects))) {
                return false;
            }
        }

        // execute
        execute(execute.iterator());
        return true;
    }

    /** Execute some nodes.
    *
    * @param nodes the nodes (should have {@link ExecCookie} on them if they are to be used)
    * @param compileBefore <code>true</code> to compile before executing
    * @return true if compilation succeeded or was not performed, false if compilation failed
    */
    public static boolean execute(Node[] nodes, boolean compileBefore) {
        return execute(nodes, compileBefore, false);
    }

    /** Execute some nodes.
    *
    * @param nodes the nodes (should have {@link ExecCookie} on them if they are to be used)
    * @param compileBefore <code>true</code> to compile before executing
    * @param switchWorkspace <code>true</code> to switch workspace before executing
    * @return true if compilation succeeded or was not performed, false if compilation failed
    */
    private static boolean execute(Node[] nodes, boolean compileBefore, boolean switchWorkspace) {
        // find all activatedNodes with unique ExecCookies/StartCookies -
        // - it is possible, that multiple activated nodes have the same exec cookie and
        // we have to prevent running it multiple times
        HashSet compile = new HashSet ();
        HashSet execute = new HashSet ();

        for (int i = 0; i < nodes.length; i++) {
            ExecCookie exec = (ExecCookie) nodes[i].getCookie(ExecCookie.class);
            if (exec != null) {
                execute.add(exec);
                if (compileBefore) {
                    CompilerCookie comp = (CompilerCookie) nodes[i].getCookie(CompilerCookie.Compile.class);
                    if (comp != null) {
                        compile.add(comp);
                    }
                }
            }
        }

        // compile
        if (compileBefore) {
            if (! AbstractCompileAction.compile(Collections.enumeration(compile),
                                                AbstractCompileAction.findName(nodes))) {
                return false;
            }
        }

        if (switchWorkspace) {
            Mutex.EVENT.readAccess (new Mutex.Action () {
                    public Object run() {
                        changeWorkspace ();
                        return null;
                    }
                });
        }
        // execute
        execute(execute.iterator());
        return true;
    }
}
