/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.actions;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import org.openide.*;
import org.openide.util.HelpCtx;
import org.openide.util.RequestProcessor;
import org.openide.util.actions.SystemAction;

/** Show help for the currently selected workspace element.
*
* @see HelpCtx#findHelp
* @author Jesse Glick
*/
public class HelpAction extends SystemAction {
    /** generated Serialized Version UID */
    private static final long serialVersionUID = 4658008202517094416L;

    /** Component the mouse is currently over. */
    private static Component globallySelectedComp;

    /* Human presentable name of the action. This should be
    * presented as an item in a menu.
    * @return the name of the action
    */
    public String getName() {
        return ActionConstants.BUNDLE.getString("Help");
    }

    /* Help context where to find more about the action.
    * @return the help context for this action
    */
    public HelpCtx getHelpCtx() {
        return new HelpCtx (HelpAction.class);
    }

    /* Icon resource.
    * @return name of resource for icon
    */
    protected String iconResource () {
        return "/org/openide/resources/actions/showHelp.gif"; // NOI18N
    }

    // This ensures initialize() will be called ASAP, hence that
    // the AWT listener will actually be started quickly and there
    // will already have been interesting mouse-entered events
    // by the time F1 is first pressed. Otherwise only the second
    // F1 actually gets anything other than the main window help.
    static {
        ((HelpAction) SystemAction.get (HelpAction.class)).forceInit ();
    }
    private void forceInit () {
        getProperty ("whatever");
    }

    // Make sure this action works on dialogs, etc.--everywhere.
    protected void initialize () {
        super.initialize ();
        // For simplicity, use the same debug mode as the core does:
        ErrorManager err = TopManager.getDefault ().getErrorManager ().getInstance ("org.netbeans.core.JavaHelp"); // NOI18N
        err.log ("HelpAction.initialize");
        putProperty ("err", err); // NOI18N
        // Cf. org.netbeans.core.windows.frames.NbFocusManager:
        putProperty ("OpenIDE-Transmodal-Action", Boolean.TRUE); // NOI18N
        // For live-tracking, we need to know what component the mouse is over:
        // Make sure it is not null:
        globallySelectedComp = TopManager.getDefault ().getWindowManager ().getMainWindow ();
        Toolkit.getDefaultToolkit ().addAWTEventListener (new AWTEventListener () {
                    public void eventDispatched (AWTEvent ev) {
                        if ((ev instanceof MouseEvent) && ev.getID () == MouseEvent.MOUSE_ENTERED)
                            globallySelectedComp = ((MouseEvent) ev).getComponent ();
                    }
                }, AWTEvent.MOUSE_EVENT_MASK);
	// [PENDING] Should instead use if available (JDK 1.4):
	// java.awt.KeyboardFocusManager.getCurrentKeyboardFocusManager ().getFocusOwner ()
    }

    public void actionPerformed (ActionEvent ev) {
        HelpCtx help = HelpCtx.findHelp (globallySelectedComp);
        final ErrorManager err = (ErrorManager) getProperty ("err"); // NOI18N
        if (err.isLoggable (ErrorManager.UNKNOWN))
            err.log (ErrorManager.UNKNOWN, help.toString () + " from " + globallySelectedComp);
        else if (err.isLoggable (ErrorManager.INFORMATIONAL))
            err.log (ErrorManager.INFORMATIONAL, help.toString ());
        TopManager.getDefault ().setStatusText (ActionConstants.BUNDLE.getString("CTL_OpeningHelp"));
        TopManager.getDefault ().showHelp (help);
        TopManager.getDefault ().setStatusText (""); // NOI18N
        // Copied from MainWindow:
        final MenuSelectionManager msm = MenuSelectionManager.defaultManager ();
        final MenuElement[] path = msm.getSelectedPath ();
        // post request that should after half of second clear the selected menu
        RequestProcessor.postRequest (new Runnable () {
                                          public void run () {
                                              SwingUtilities.invokeLater (new Runnable () {
                                                                              public void run () {
                                                                                  MenuElement[] newPath = msm.getSelectedPath ();
                                                                                  if (newPath.length != path.length) return;
                                                                                  for (int i = 0; i < newPath.length; i++) {
                                                                                      if (newPath[i] != path[i]) return;
                                                                                  }
                                                                                  msm.clearSelectedPath ();
                                                                              }
                                                                          });
                                          }
                                      }, 200);
    }
}
