/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.actions;

import java.beans.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;

import org.openide.TopManager;
import org.openide.awt.JInlineMenu;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.util.WeakListener;
import org.openide.util.actions.*;
import org.openide.awt.JPopupMenuUtils;

/** A "meta-action" that displays (in a submenu) a list of enabled actions provided by modules.
* Such registered actions are called "service actions":
* they are provided externally but seem to provide additional services on existing components.
* Often they will be {@link NodeAction}s or {@link CookieAction}s so that they will
* be enabled based on the node selection, i.e. the node containing this popup.
* It is desirable for most nodes to include this action somewhere in their popup menu.
*
* <p><em>Note:</em> you do not need to touch this class to add a service action!
* Just add the action to a module manifest in an <code>Action</code> section.
*
* <p>The list of registered service actions is provided to this action from the implementation
* by means of {@link ActionManager}.
*
* @author Jaroslav Tulach
*/
public class ToolsAction extends SystemAction
    implements Presenter.Menu, Presenter.Popup {
    static final long serialVersionUID =4906417339959070129L;
    
        
    // Global ActionManager listener monitoring all available actions
    // and their state
    private static G gl = new G();
    
    /* @return name
    */
    public String getName () {
        return getActionName ();
    }

    /* @return help for this action
    */
    public HelpCtx getHelpCtx () {
        return new HelpCtx (ToolsAction.class);
    }

    /* @return menu presenter for the action
    */
    public JMenuItem getMenuPresenter () {
        return new Inline();
    }

    /* @return menu presenter for the action
    */
    public JMenuItem getPopupPresenter () {
        return new Popup();
    }

    /* Does nothing.
    */
    public void actionPerformed (java.awt.event.ActionEvent ev) {
    }


    /* @return name
    */
    private static String getActionName () {
        return ActionConstants.BUNDLE.getString ("CTL_Tools");
    }

    /** Implementation method that regenerates the items in the menu or
    * in the array.
    *
    * @param forMenu true if Presenter.Menu should be used false if Presenter.Popup
    * @param list (can be null)
    */
    private static List generate (boolean forMenu) {
        ActionManager am = TopManager.getDefault ().getActionManager ();
        SystemAction[] actions = am.getContextActions ();
        List list = new ArrayList( actions.length );

        boolean separator = false;
        boolean firstItemAdded = false; // flag to prevent adding separator before actual menu items

        for (int i = 0; i < actions.length; i++) {
            SystemAction a = actions[i];
            if (a == null) {
                if (firstItemAdded) separator = true;
            } else if( forMenu ) {
                if( a instanceof Presenter.Menu && a.isEnabled () ) {
                    if (separator) {
                        list.add (null);
                        separator = false;
                    }

                    JMenuItem mi = ((Presenter.Menu)a).getMenuPresenter ();
                    list.add (mi);

                    firstItemAdded = true;
                }
            } else if( a instanceof Presenter.Popup && a.isEnabled() ) {
                if (separator) {
                    list.add (null);
                    separator = false;
                }

                JMenuItem mi = ((Presenter.Popup)a).getPopupPresenter ();
                list.add (mi);

                firstItemAdded = true;
            }
        }
        return list;
    }    
    
    
    /** Inline menu that watches model changes only when really needed.
     */
    private static final class Inline extends JInlineMenu
        implements PropertyChangeListener, Runnable {
        static final long serialVersionUID =2269006599727576059L;

	/** timestamp of the beginning of the last regeneration */
	private int timestamp = 0;

        /** By calling this method, our parent notifies us we've to be keep
         * updated, so we start listening on SystemAction changes, and 
         * schedule updating Runnable imediately. 
         */
        public void addNotify() {
            SwingUtilities.invokeLater (this);
            gl.addPropertyChangeListener( this );
            super.addNotify();
        }

        /** By calling this method, our parent notifies us we don't have
         * to be up-to-date more, so we switch to lazy mode and discard any
         * pending updates.
         */
        public void removeNotify() {
            gl.removePropertyChangeListener( this );
            super.removeNotify();
        }
        
        /** Change of model.
        */
        public void propertyChange (PropertyChangeEvent ev) {
            String prop = ev.getPropertyName ();
            if (prop == null || prop.equals (G.PROP_STATE)) {
                SwingUtilities.invokeLater (this);
            }
        }


        /** Runs the update */
        public void run () {
	    if( timestamp == gl.getTimestamp() ) return;

	    timestamp = gl.getTimestamp();

            // generate directly list of menu items
            List l = generate ( true );
            JMenuItem[] arr = new JMenuItem[l.size ()];
            l.toArray (arr);
            setMenuItems (arr);
        }
    }

    
    
    //--------------------------------------------------
    
    /** Inline menu that is either empty or contains one submenu.
    */
    private static final class Popup extends JInlineMenu
        implements PropertyChangeListener, Runnable, PopupMenuListener {
        static final long serialVersionUID =2269006599727576059L;

        /** sub menu */
        private JMenu menu;
        
	private int timestamp = 0;
	private int subMenuTimestamp = 0;

        private boolean keepSubMenu = false;

        public Popup () {
            this.menu = new org.openide.awt.JMenuPlus ( getActionName () );
            HelpCtx.setHelpIDString (menu, ToolsAction.class.getName ());
            menu.getPopupMenu().addPopupMenuListener( this );
        }
     
        public void popupMenuCanceled( PopupMenuEvent evt ) {
            keepSubMenu = false;
        }

        public void popupMenuWillBecomeInvisible( PopupMenuEvent evt ) {
            keepSubMenu = false;
        }

        public void popupMenuWillBecomeVisible( PopupMenuEvent evt ) {
            keepSubMenu = true;
            
            // do it directly here, we're in AWT-Event-queue anyway and it
            // will save us one repaint (we'll not popup empty window)
            // * reverted back to invokeLater() because of problems with
            // popupmenu growing off-screen if poped up near screen edge
//            updateSubMenu( true ); 
            
            SwingUtilities.invokeLater (this);
        }

        
        private void updateSubMenu() {
	    if( subMenuTimestamp != gl.getTimestamp() ) {
		subMenuTimestamp = gl.getTimestamp();

                JPopupMenu pop = menu.getPopupMenu();
                boolean wasContained = JPopupMenuUtils.isPopupContained( pop );
                
                menu.removeAll ();
                List l = generate(false);
                Iterator it = l.iterator();
                while( it.hasNext() ) {
                    java.awt.Component item = (java.awt.Component)it.next();
                    if( item == null ) {
                        menu.addSeparator ();
                    } else {
                        menu.add ( item );
                    }
                }
                JPopupMenuUtils.dynamicChange( pop, wasContained );
            }
        }
            
       
        
        public void addNotify() {
            SwingUtilities.invokeLater (this);
            gl.addPropertyChangeListener (this);
            super.addNotify();
        }

        public void removeNotify() {
            gl.removePropertyChangeListener (this);
            super.removeNotify();
        }
        
        /** Change of model. We're listening on this only if we're
         * really visible (only between addNotify and removeNotify)
         */
        public void propertyChange (PropertyChangeEvent ev) {
            String prop = ev.getPropertyName ();
            if (prop == null || prop.equals (G.PROP_STATE)) {
                SwingUtilities.invokeLater (this);
            }
        }


        /** Runs the update */
        public void run () {
	    if( timestamp != gl.getTimestamp() ) {
		timestamp = gl.getTimestamp();

                if( gl.isPopupEnabled() ) {
                    setMenuItems (new JMenuItem[] { menu });
                } else {
                    setMenuItems (new JMenuItem[0]);
                }
            }
	    
            if( keepSubMenu ) updateSubMenu();
        }
    }

    //------------------------------------------
    
    
    
    /** @deprecated Useless, see {@link ActionManager}. */
    public static void setModel (Model m) {
        throw new SecurityException ();
    }
    /** @deprecated Useless, see {@link ActionManager}. */
    public static interface Model {
        public SystemAction[] getActions ();
        public void addChangeListener (javax.swing.event.ChangeListener l);
        public void removeChangeListener (javax.swing.event.ChangeListener l);
    }


    
    //------------------------------------------------
    
    
    //----------------------------------------------------------
    
    
    
    

    private static class G implements PropertyChangeListener {
	private int timestamp = 1;
	
        private boolean enabled = false;
        private boolean dirty = false;
        private SystemAction[] actions = null;
        private PropertyChangeSupport supp = new PropertyChangeSupport( this );
        
        public static final String PROP_STATE = "actionsState";
        
        public G() {
            ActionManager am = TopManager.getDefault().getActionManager();
            am.addPropertyChangeListener ( this );
            actionsListChanged();
        }
        
        public final void addPropertyChangeListener(
                PropertyChangeListener listener
        ) {
            supp.addPropertyChangeListener(listener);
        }
        
        public final void removePropertyChangeListener(
                PropertyChangeListener listener
        ) {
            supp.removePropertyChangeListener (listener);
        }
        
        protected final void firePropertyChange(
            String name , Object o, Object n
        ) {
            supp.firePropertyChange(name, o, n);
        }


        
        
        private void actionsListChanged() {
            // mark us dirty
	    timestamp++;
            dirty = true;
            // deregister all actions listeners
            SystemAction[] copy = actions;
            if( copy != null ) {
                for( int i = 0; i< copy.length; i++ ) {
                    SystemAction act = copy[i];
                    if( act != null ) act.removePropertyChangeListener ( this );
                }
            }

            ActionManager am = TopManager.getDefault ().getActionManager ();
            copy = am.getContextActions ();
            for (int i = 0; i < copy.length; i++) {
                SystemAction act = copy[i];
                if( act != null ) act.addPropertyChangeListener ( this );
            }
            actions = copy;

            firePropertyChange( PROP_STATE, null, null ); // tell the world
        }
        
        private void actionStateChanged() {
	    timestamp++;
            dirty = true;            // mark us dirty
            firePropertyChange( PROP_STATE, null, null );  // tell the world
        }
        
        public void propertyChange (PropertyChangeEvent ev) {
            String prop = ev.getPropertyName ();
            if (prop == null || prop.equals (ActionManager.PROP_CONTEXT_ACTIONS)) {
                actionsListChanged();
            } else if( prop.equals (SystemAction.PROP_ENABLED)) {
                actionStateChanged();
            }
        }

        
        /** Tells if there is any action that is willing to provide
         * Presenter.Popup
         */
        private boolean isPopupEnabled() {
            if( !dirty ) {
                return enabled;
            } else {
                boolean en = false;
                SystemAction[] copy = actions;
                for (int i=0; i<copy.length; i++) {
                    SystemAction act = copy[i];
                    if( act instanceof Presenter.Popup && act.isEnabled() ) {
                        en = true;
                        break;
                    }
                }
                enabled = en;
                dirty = false;
            }
            return enabled;
        }
	
	private int getTimestamp() {
	    return timestamp;
	}
	
    }
}
