/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.awt;

import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.text.MessageFormat;
import javax.swing.*;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.loaders.*;
import org.openide.cookies.InstanceCookie;
import org.openide.nodes.*;
import org.openide.util.actions.Presenter;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.WeakListener;

/** An extended version of swing's JMenuBar. This menubar can
* load its content from the folder where its "disk image" is stored.<P>
* Moreover, menu is <code>Externalizable</code> to restore its persistent
* state with minimal storage expensiveness.
*
* @author  David Peroutka, Dafe Simonek
*/
public class MenuBar extends JMenuBar implements Externalizable {

    /** the folder which represents and loads content of the menubar */
    private MenuBarFolder menuBarFolder;

    static final long serialVersionUID =-4721949937356581268L;
    /** Don't call this constructor or this class will not get
    * initialized properly. This constructor is only for externalization.
    */
    public MenuBar() {
        super();
    }

    /** Creates a new <code>MenuBar</code> from given folder.
    * @param folder The folder from which to create the content of the menubar.
    * If the parameter is null, default menu folder is obtained.
    */
    public MenuBar(DataFolder folder) {
        super();
        // PENDING(david) -- setBorder(new MenuBorder());
        startLoading((folder == null)
                     ? TopManager.getDefault().getPlaces().folders().menus()
                     : folder);
    }

    /** Blocks until the menubar is completely created. */
    public void waitFinished () {
        menuBarFolder.instanceFinished();
    }

    /** Saves the contents of this object to the specified stream.
    *
    * @exception IOException Includes any I/O exceptions that may occur
    */
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeObject(menuBarFolder.getFolder());
    }

    /**
     * Restores contents of this object from the specified stream.
     *
     * @exception ClassNotFoundException If the class for an object being
     *              restored cannot be found.
     */
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        startLoading((DataFolder)in.readObject());
    }

    /** Starts loading of this menu from menu folder */
    void startLoading (final DataFolder folder) {
        menuBarFolder = new MenuBarFolder(folder, this);
    }
    
    /** Convert an array of instance cookies to instances, adds them
     * to given list.
     * @param arr array of instance cookies
     * @param list list to add created objects to
     */
    static void allInstances (InstanceCookie[] arr, java.util.List list) {
        org.openide.ErrorManager err = TopManager.getDefault ().getErrorManager ();
        
        
        Exception ex = null;
        
        for (int i = 0; i < arr.length; i++) {
            
            Exception newEx = null;
            try {
                Object o = arr[i].instanceCreate();
                list.add (o);
            } catch (ClassNotFoundException e) {
                newEx = e;
            } catch (IOException e) {
                newEx = e;
            }
            
            if (newEx != null) {
                if (err.findAnnotations(newEx) == null) {
                    // if the exception is not annotated, assign it low
                    // priority
                    err.annotate(newEx, err.INFORMATIONAL, null, null, null, null);
                }
                err.copyAnnotation(newEx, ex);
                ex = newEx;
            }
        }
     
        // if there was an exception => notify it
        if (ex != null) {
            err.notify (ex);
        }
    }

    /** This class can be used to fill the content of given
    * <code>MenuBar</code> from the given <code>DataFolder</code>.
    */
    static final class MenuBarFolder extends FolderInstance {
        /** Asociated menubar which we should fill */
        MenuBar menuBar;

        /** Creates a new menubar folder on the specified <code>DataFolder</code>.
         * @param folder a <code>DataFolder</code> to work with
         */
        public MenuBarFolder (final DataFolder folder, final MenuBar menuBar) {
            super(folder);
            this.menuBar = menuBar;
            recreate ();
        }

        /** Full name of the data folder's primary file separated by dots.
        * @return the name
        */
        public String instanceName () {
            return menuBar.getClass().getName();
        }

        /** Returns the root class of all objects.
        * @return MenuBar.class
        */
        public Class instanceClass ()
        throws IOException, ClassNotFoundException {
            return MenuBar.class;
        }

        /** Accepts only cookies that can provide <code>MenuBar</code>.
        * @param cookie the instance cookie to test
        * @return true if the cookie can provide <code>MenuBar</code>
        */
        protected InstanceCookie acceptCookie (InstanceCookie cookie)
        throws IOException,
            ClassNotFoundException {
            return JMenu.class.isAssignableFrom(cookie.instanceClass())
                   ? cookie : null;
        }

        /** Returns a <code>MenuFolder</code> cookie for the specified
        * <code>DataFolder</code>.
        * @param df a <code>DataFolder</code> to create the cookie for
        * @return a <code>MenuFolder</code> for the specified folder
        */
        protected InstanceCookie acceptFolder (DataFolder df) {
            return new MenuFolder(df, false);
        }

        /** Updates the <code>MenuBar</code> represented by this folder.
        *
        * @param cookies array of instance cookies for the folder
        * @return the updated <code>MenuBar</code> representee
        */
        protected Object createInstance (InstanceCookie[] cookies)
        throws IOException, ClassNotFoundException {
            final LinkedList ll = new LinkedList ();
            allInstances (cookies, ll);
            
            org.openide.util.Mutex.EVENT.readAccess (new Runnable () {
                public void run () {
                    //      System.out.println ("Creating menu bar:" + menuBar); // NOI18N
                    // clear old content
                    menuBar.removeAll();

                    // hash map of used Integers for mnemonics
                    HashMap mnems = new HashMap ();

                    // fill with new content
                    Iterator it = ll.iterator ();
                    while (it.hasNext()) {
                        Object obj = it.next ();
                        if (obj instanceof JMenu) {
                            JMenu m = (JMenu)obj;
                            
//                            m.setMnemonic(findMnemonic (m, mnems));
                            menuBar.add(m);
                        }
                    }
                    final MenuBar menuBarRef = menuBar;
                    menuBarRef.validate();
                    menuBarRef.repaint();
                }
            });
            
            return menuBar;
        }

        /** For outer class access to the data folder */
        DataFolder getFolder () {
            return folder;
        }

    }

    /** This class can be used to produce a <code>JMenu</code> instance
    * from the given <code>DataFolder</code>.
    */
    static final class MenuFolder extends FolderInstance
        implements NodeListener {

        /** the <code>JMenu</code> to work with */
        private JMenu menu;
        /** whether or not to make an icon for it */
        private boolean icon;

        /**
         * Creates a menu on the specified <code>DataFolder</code>.
         * @param folder a <code>DataFolder</code> to work with
         * @param icon true if the folder should use folder's icon
         */
        public MenuFolder (final DataFolder folder, boolean icon) {
            super(folder);

            this.icon = icon;

            recreate ();
        }

        /**
         * Returns a <code>Menu</code> representee of this folder.
         * @return a <code>Menu</code> representee of this folder
         */
        public final JMenu getMenu() {
            if (menu == null) {
                synchronized (this) {
                    if (menu == null) {
                        Node n = folder.getNodeDelegate ();

                        menu = new org.openide.awt.JMenuPlus();
                        // set the text and be aware of mnemonics
                        Actions.setMenuText(menu, n.getDisplayName (), true);

                        if (icon) {
                            menu.setIcon (new ImageIcon (n.getIcon (
                                java.beans.BeanInfo.ICON_COLOR_16x16
                            )));
                        }

                        n.addNodeListener (
                            WeakListener.node (this, folder.getNodeDelegate ())
                        );
                    }
                }
            }
            return menu;
        }

        /** If the display name changes, than change the name of the menu.
        */
        public void propertyChange (java.beans.PropertyChangeEvent ev) {
            if (
                Node.PROP_DISPLAY_NAME.equals (ev.getPropertyName ()) ||
                Node.PROP_NAME.equals (ev.getPropertyName ()) ||
                Node.PROP_ICON.equals (ev.getPropertyName ())
            ) {
                Node n = folder.getNodeDelegate ();
                Actions.setMenuText (getMenu (), n.getDisplayName (), true);
                if (icon) {
                    getMenu ().setIcon (new ImageIcon (n.getIcon (
                        java.beans.BeanInfo.ICON_COLOR_16x16
                    )));
                }
            }
        }

        /** Fired when a set of new children is added.
        * @param ev event describing the action
        */
        public void childrenAdded (NodeMemberEvent ev) {
        }

        /** Fired when a set of children is removed.
        * @param ev event describing the action
        */
        public void childrenRemoved (NodeMemberEvent ev) {
        }

        /** Fired when the order of children is changed.
        * @param ev event describing the change
        */
        public void childrenReordered(NodeReorderEvent ev) {
        }

        /** Fired when the node is deleted.
        * @param ev event describing the node
        */
        public void nodeDestroyed (NodeEvent ev) {
        }

        /** The name of the menu
        * @return the name
        */
        public String instanceName () {
            return getMenu().getClass().getName();
        }

        /** Returns the class of represented menu.
        * @return Object.class
        */
        public Class instanceClass ()
        throws IOException, ClassNotFoundException {
            return JMenu.class;
        }

        /** If no instance cookie, tries to create execution action on the
        * data object.
        */
        protected InstanceCookie acceptDataObject (DataObject dob) {
            InstanceCookie ic = super.acceptDataObject (dob);
            if (ic == null) {
                JMenuItem item = ExecBridge.createMenuItem (dob);
                return item != null ? new InstanceSupport.Instance (item) : null;
            } else {
                return ic;
            }
        }

        /**
         * Accepts only cookies that can provide <code>Menu</code>.
         * @param cookie an <code>InstanceCookie</code> to test
         * @return true if the cookie can provide accepted instances
         */
        protected InstanceCookie acceptCookie(InstanceCookie cookie)
        throws IOException,
            ClassNotFoundException {
            Class c = cookie.instanceClass();
            return ((Presenter.Menu.class.isAssignableFrom(c)) ||
                    (JMenuItem.class.isAssignableFrom(c)) ||
                    (JSeparator.class.isAssignableFrom(c))) ? cookie : null;
        }

        /**
         * Returns a <code>Menu.Folder</code> cookie for the specified
         * <code>DataFolder</code>.
         * @param df a <code>DataFolder</code> to create the cookie for
         * @return a <code>Menu.Folder</code> for the specified folder
         */
        protected InstanceCookie acceptFolder(DataFolder df) {
            return new MenuFolder(df, true);
        }

        /** Updates the <code>JMenu</code> represented by this folder.
        * @param cookies array of instance cookies for the folder
        * @return the updated <code>JMenu</code> representee
        */
        protected Object createInstance(InstanceCookie[] cookies)
        throws IOException, ClassNotFoundException {
            final LinkedList ll = new LinkedList ();
            allInstances (cookies, ll);
            
            // also initializes the menu (not in the AWT thread)
            JMenu toReturn = getMenu ();
            
            // clear first - refresh the menu's content
            org.openide.util.Mutex.EVENT.readAccess(new Runnable () {
                public void run () {
                    boolean addSeparator = false;
                    JMenu menu = getMenu ();
                    
                    
                    JPopupMenu popup = menu.getPopupMenu ();
                    
                    menu.removeAll();
                    
                    Iterator it = ll.iterator();
                    while (it.hasNext()) {
                        Object obj = it.next();
                        if (obj instanceof Presenter.Menu) {
                            obj = ((Presenter.Menu)obj).getMenuPresenter();
                        }
                        
                        
                        if (obj instanceof JMenuItem) {
                            if(addSeparator) {
                                menu.addSeparator();
                                addSeparator = false;
                            }
                            menu.add((JMenuItem)obj);
                        } else {
                            if (obj instanceof JSeparator) {
                                addSeparator = menu.getMenuComponentCount() > 0;
                            }
                        }
                    }
                    menu.setEnabled((ll.size () == 0) ? false : true);
                    // request menu repaint
                    popup.pack();
                    popup.repaint();
                    
                    // Special help ID dependent on name of menu, since otherwise
                    // there would be no way to configure the ID from the DataFolder.
                    HelpCtx.setHelpIDString (menu, MenuFolder.class.getName () + "." + menu.getText ()); // NOI18N
                }
            });
            
            return toReturn;
        }
    }

}
