/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.explorer;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.datatransfer.*;
import java.io.IOException;

import org.openide.util.datatransfer.*;
import org.openide.loaders.*;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.actions.*;
/*import org.openide.actions.CutAction;
import org.openide.actions.CopyAction;
import org.openide.actions.PasteAction;
import org.openide.actions.DeleteAction;*/
import org.openide.util.actions.ActionPerformer;
import org.openide.nodes.Node;
import org.openide.util.RequestProcessor;
import org.openide.util.NbBundle;

/** This class contains the default implementation of reactions to the standard
* explorer actions. It can be
* attached to any {@link ExplorerManager}. Then this class will listen to changes
* of selected nodes or the explored context of that manager, and update the state
* of cut/copy/paste/delete actions.
* <P>
* An instance of this class can only be attached to one manager at a time. Use
* {@link #attach} and {@link #detach} to make the connection.
*
* @author Jan Jancura, Petr Hamernik, Ian Formanek, Jaroslav Tulach
*/
public class ExplorerActions extends Object {
    /** copy action performer */
    private ActionPerformer copyActionPerformer;
    /** cut action performer */
    private ActionPerformer cutActionPerformer;
    /** delete action performer */
    private ActionPerformer deleteActionPerformer;

    /** tracker for all actions */
    private Listener listener;

    /** the manager we are listening on */
    ExplorerManager manager;

    /** must the delete be confirmed */
    private boolean confirmDelete = true;

    /** actions to work with */
    static CopyAction copy = null;
    static CutAction cut = null;
    static DeleteAction delete = null;
    static PasteAction paste = null;

    /** Attach to new manager.
    * @param m the manager to listen on
    */
    public synchronized void attach (ExplorerManager m) {
        if (manager != null) {
            // first of all detach
            detach ();
        }

        manager = m;

        if (listener == null) listener = new Listener ();
        manager.addPropertyChangeListener (listener);
        ExClipboard clip = TopManager.getDefault().getClipboard();
        clip.addClipboardListener(listener);

        updateActions ();
    }

    /** Detach from manager currently being listened on.
    */
    public synchronized void detach () {
        if (manager == null) return;

        manager.removePropertyChangeListener (listener);
        ExClipboard clip = TopManager.getDefault().getClipboard();
        clip.removeClipboardListener(listener);

        stopActions ();

        manager = null;
    }

    /** Set whether to confirm deletions.
    * @param yes <code>true</code> to confirm deletions
    */
    public final void setConfirmDelete (boolean yes) {
        confirmDelete = yes;
    }

    /** Should deletions be confirmed?
    * @return <code>true</code> if deletions must be confirmed
    */
    public final boolean isConfirmDelete () {
        return confirmDelete;
    }

    /** Stops listening on all actions
    */
    void stopActions () {
        if (copyActionPerformer != null) {
            copy.setActionPerformer (null);
            cut.setActionPerformer (null);
            paste.setPasteTypes (null);
            delete.setActionPerformer (null);
        }
    }

    /** Updates the state of all actions.
    * @param path list of selected nodes
    */
    void updateActions () {
        Node[] path = (manager == null)? (new Node[0]) : (manager.getSelectedNodes ());
        /*    System.out.println("ExplorerActions.updateActions: "+path.length);
            for (int i = 0; i < path.length; i++)
              System.out.println("actions["+i+"]="+path[i].getDisplayName ());

            Thread.dumpStack(); */
        if (copyActionPerformer == null) {
            copyActionPerformer = new CopyCutActionPerformer (true);
            cutActionPerformer = new CopyCutActionPerformer (false);
            deleteActionPerformer = new DeleteActionPerformer();

            copy = (CopyAction) CopyAction.findObject(CopyAction.class, true);
            cut = (CutAction) CutAction.findObject(CutAction.class, true);
            paste = (PasteAction) PasteAction.findObject(PasteAction.class, true);
            delete = (DeleteAction) DeleteAction.findObject(DeleteAction.class, true);
        }

        int i;
        int k = path != null ? path.length : 0;
        if (k > 0) {
            for (i = 0; i < k; i++) {
                if (!path[i].canCopy ()) {
                    copy.setActionPerformer (null);
                    break;
                }
            }
            if (i == k) {
                Object perf = copy.getActionPerformer();
                if (perf != copyActionPerformer) {
                    copy.setActionPerformer (copyActionPerformer);
                }
            }

            for (i = 0; i < k; i++) {
                if (!path[i].canCut ()) {
                    cut.setActionPerformer (null);
                    break;
                }
            }
            if (i == k) {
                Object perf = cut.getActionPerformer();
                if (perf != cutActionPerformer) {
                    cut.setActionPerformer (cutActionPerformer);
                }
            }

            java.util.HashMap allNodes = new java.util.HashMap (101);
            for (i = 0; i < k; i++) {
                if (!path[i].canDestroy () || !checkParents (path[i], allNodes)) {
                    delete.setActionPerformer (null);
                    break;
                }

            }
            if (i == k) {
                Object perf = delete.getActionPerformer();
                if (perf != deleteActionPerformer) {
                    delete.setActionPerformer (deleteActionPerformer);
                }
            }

        } else {
            copy.setActionPerformer (null);
            cut.setActionPerformer (null);
            delete.setActionPerformer (null);
        }
        updatePasteAction(path);
    }
    
    /** Adds all parent nodes into the set.
     * @param set set of all nodes
     * @param node the node to check
     * @return false if one of the nodes is parent of another
     */
    private boolean checkParents (Node node, java.util.HashMap set) {
        if (set.get (node) != null) {
            return false;
        }
        
        // this signals that this node is the original one
        set.put (node, this);
        
        for (;;) {
            node = node.getParentNode ();
            
            if (node == null) {
                return true;
            }
            
        
            if (set.put (node, node) == this) {
                // our parent is a node that is also in the set
                return false;
            }
        }
    }
    

    /** Updates paste action.
    * @param path selected nodes
    */
    private void updatePasteAction (Node[] path) {
        ExplorerManager man = manager;

        if (man == null) {
            paste.setPasteTypes (null);
            return;
        }

        if (path != null && ((path.length > 1) ||
                             ((path.length == 1) && (path [0].isLeaf())))
           ) {
            paste.setPasteTypes(null);
            return;
        } else {
            Node pan = man.getExploredContext ();
            Node[] selectedNodes = man.getSelectedNodes ();
            if (selectedNodes != null && (selectedNodes.length == 1) && (!selectedNodes[0].isLeaf()))
                pan = selectedNodes[0];

            Clipboard clipboard = TopManager.getDefault().getClipboard();
            Transferable trans = clipboard.getContents(TopManager.getDefault()); //hahaha !! who is the requestor ??!
            if (trans != null) {

                // First, just ask the node if it likes this transferable, whatever it may be.
                // If it does, then fine.
                PasteType[] pasteTypes = pan == null ? new PasteType[] { } : pan.getPasteTypes(trans);
                if (pasteTypes.length != 0) {
                    paste.setPasteTypes(pasteTypes);
                    return;
                }

                boolean flavorSupported = false;
                try {

                    flavorSupported = trans.isDataFlavorSupported(ExTransferable.multiFlavor);

                } catch (java.lang.Exception e) {
                    // patch to get the Netbeans start under Solaris
                    // [PENDINGworkaround]
                }

                if (flavorSupported) {
                    // The node did not accept this multitransfer as is--try to break it into
                    // individual transfers and paste them in sequence instead.
                    try {
                        MultiTransferObject obj = (MultiTransferObject) trans.getTransferData(ExTransferable.multiFlavor);
                        int count = obj.getCount();
                        boolean ok = true;
                        Transferable[] t = new Transferable[count];
                        PasteType[] p = new PasteType[count];

                        for (int i = 0; i < count; i++) {
                            t[i] = obj.getTransferableAt(i);
                            pasteTypes = pan == null ? new PasteType[] { } : pan.getPasteTypes(t[i]);
                            if (pasteTypes.length == 0) {
                                ok = false;
                                break;
                            }
                            // [PENDING] this is ugly! ideally should be some way of comparing PasteType's for similarity?
                            p[i] = pasteTypes[0];
                        }
                        if (ok) {
                            paste.setPasteTypes(new PasteType[] { new MultiPasteType(t, p) } );
                            return;
                        }
                    }
                    catch (UnsupportedFlavorException e) {
                        // [PENDING] notify?!
                    }
                    catch (IOException e) {
                        // [PENDING] notify?!
                    }
                }
            }
            if (paste != null) paste.setPasteTypes(null);
            return;
        }
    }

    /** Paste type used when in clipbopard is MultiTransferable */
    private static class MultiPasteType extends PasteType {
        /** Array of transferables */
        Transferable[] t;

        /** Array of paste types */
        PasteType[] p;

        /** Constructs new MultiPasteType for the given content of the clipboard */
        MultiPasteType(Transferable[] t, PasteType[] p) {
            this.t = t;
            this.p = p;
        }

        /** Performs the paste action.
        * @return Transferable which should be inserted into the clipboard after
        *         paste action. It can be null, which means that clipboard content
        *         should be cleared.
        */
        public Transferable paste() throws IOException {
            int size = p.length;
            Transferable[] arr = new Transferable[size];
            for (int i = 0; i < size; i++) {
                Transferable newTransferable = p[i].paste();
                if (newTransferable != null) {
                    arr[i] = newTransferable;
                } else {
                    // keep the orginal
                    arr[i] = t[i];
                }
            }
            return new ExTransferable.Multi (arr);
        }
    }

    /** Class which performs copy and cut actions */
    class CopyCutActionPerformer extends Object implements
        org.openide.util.actions.ActionPerformer {
        /** determine if adapter is used for copy or cut action. */
        boolean copyCut;

        /** Create new adapter */
        public CopyCutActionPerformer (boolean b) {
            copyCut = b;
        }

        /** Perform copy or cut action. */
        public void performAction(org.openide.util.actions.SystemAction action) {
            Transferable trans = null;
            Node[] sel = manager.getSelectedNodes ();

            if (sel.length != 1) {
                Transferable[] arrayTrans = new Transferable[sel.length];

                for (int i = 0; i < sel.length; i++)
                    if ((arrayTrans[i] = getTransferableOwner(sel[i])) == null) return;
                trans = new ExTransferable.Multi (arrayTrans);
            }
            else {
                trans = getTransferableOwner(sel[0]);
            }
            if (trans != null) {
                Clipboard clipboard = TopManager.getDefault().getClipboard();

                clipboard.setContents(trans, new StringSelection ("")); // NOI18N
            }
        }

        private Transferable getTransferableOwner(Node node) {
            try {
                return copyCut ? node.clipboardCopy() : node.clipboardCut();
            } catch (java.io.IOException e) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) { // NOI18N
                    e.printStackTrace();
                }
                org.openide.TopManager.getDefault().notify(
                    new org.openide.NotifyDescriptor.Exception(e)
                );
                return null;
            }
        }
    }// innerclass CopyCutActionAdapter

    /** Class which performs delete action */
    class DeleteActionPerformer extends Object implements
        org.openide.util.actions.ActionPerformer {
        /** Perform delete action. */
        public void performAction(org.openide.util.actions.SystemAction action) {
            final Node[] sel = manager.getSelectedNodes ();
            if ((sel == null) || (sel.length == 0))
                return;

            boolean isConfirmed = !confirmDelete;
            // ask for confirmation if not confirmed
            if (isConfirmed == false) {
                String title;
                String message;

                if (sel.length == 1) {
                    if (sel[0].getCookie (DataShadow.class) != null) {
                        title = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteShadowTitle");
                        DataShadow obj = (DataShadow)sel[0].getCookie (DataShadow.class);
                        message = NbBundle.getMessage
                            (ExplorerActions.class, "MSG_ConfirmDeleteShadow",
                             new Object[] {
                                 obj.getName (), // name of the shadow
                                 sel[0].getDisplayName (), // name of original
                                 obj.getPrimaryFile ().toString (), // full name of file for shadow
                                 obj.getOriginal ().getPrimaryFile ().toString () // full name of original file
                             });
                    } else if (sel[0].getCookie(DataFolder.class) != null) {
                        message = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteFolder",
                                                       sel[0].getDisplayName());
                        title = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteFolderTitle");
                    } else {
                        message = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteObject",
                                                       sel[0].getDisplayName());
                        title = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteObjectTitle");
                    }
                }
                else {
                    message = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteObjects",
                                                   new Integer(sel.length));
                    title = NbBundle.getMessage (ExplorerActions.class, "MSG_ConfirmDeleteObjectsTitle");
                }
                NotifyDescriptor desc = new NotifyDescriptor.Confirmation(
                                            message,
                                            title,
                                            NotifyDescriptor.YES_NO_OPTION
                                        );
                isConfirmed = NotifyDescriptor.YES_OPTION.equals (
                                  TopManager.getDefault().notify(desc)
                              );
            }
            // perform action if confirmed
            if (isConfirmed) {

                // clear selected nodes
                try {
                    manager.setSelectedNodes(new Node[] {});
                } catch (java.beans.PropertyVetoException e) {
                    // never thrown, setting empty selected nodes cannot be vetoed
                }
                try {
                    TopManager.getDefault().getRepository ().getDefaultFileSystem().runAtomicAction (new FileSystem.AtomicAction () {
                                public void run () {
                                    for (int i=0; i<sel.length; i++) {
                                        try {
                                            sel[i].destroy ();
                                        } catch (IOException e) {
                                            TopManager.getDefault ().notifyException (e);                                            
                                        }
                                    }
                                }
                            });
                }
                catch (IOException e) {
                }
                delete.setActionPerformer (null); // fixes bug #673
            }
        }
    }// innerclass DeleteActionAdapter

    /** delay for updating actions */
    private static final int DELAY = 150;

    private class Listener
        implements PropertyChangeListener, ClipboardListener, Runnable {
        /** a task for updating with delay */
        private RequestProcessor.Task task = RequestProcessor.createRequest(this);
        {
            // priority on the level of AWT
            task.setPriority (Thread.MAX_PRIORITY - 1);
        }

        public void propertyChange (PropertyChangeEvent e) {
            if (task.getDelay () == 0) {
                // the task is not waiting in the queue 
                // execute the update immediatelly (maybe somebody is 
                // waiting for it
                
                updateActions ();
                
                // but still schedule the update task into the queue, so
                // second time somebody calls the method the getDelay will
                // probably be > 0 and the update will not be done immediatelly
            }
            
            
            
            task.schedule (DELAY);
        }

        /** This method is called when content of clipboard is changed.
        * @param ev event describing the action
        */
        public void clipboardChanged (ClipboardEvent ev) {
            if (!ev.isConsumed ()) {
                updatePasteAction(manager.getSelectedNodes ());
            }
        }

        /** Updates the actions.
        */
        public void run() {
            updateActions ();
        }
    }
}
