/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.explorer;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.io.*;
import java.text.MessageFormat;

import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.io.*;
import org.openide.windows.TopComponent;
import org.openide.windows.Workspace;
import org.openide.TopManager;

/** Simple top component capable of displaying an Explorer.
* Holds one instance of {@link ExplorerManager} and 
* implements {@link ExplorerManager.Provider} to allow child components to share 
* the same explorer manager.
* <p>Uses {@link java.awt.BorderLayout} by default.
* Pays attention to the selected nodes and explored context as indicated by the manager.
* Cut/copy/paste actions are sensitive to the activation state of the component.
* <p>It is up to you to add a view and other UI apparatus to the panel.
*
* @author Jaroslav Tulach
*/
public class ExplorerPanel extends TopComponent implements ExplorerManager.Provider {
    /** serial version UID */
    static final long serialVersionUID = 5522528786650751459L;

    /** The message formatter for Explorer title */
    private static MessageFormat formatExplorerTitle;

    /** the instance of the explorer manager*/
    private ExplorerManager manager;

    /** listens on the selected nodes in the ExporerManager */
    transient private PropertyChangeListener managerListener;

    /** action handler for cut/copy/paste/delete */
    private static ExplorerActions actions = new ExplorerActions ();

    /** Initialize the explorer panel with the provided manager.
    * @param explorer the explorer manager to use
    */
    public ExplorerPanel (ExplorerManager manager) {
        this ();
        this.manager = manager;
    }

    /** Default constructor. Uses newly created manager.
    */
    public ExplorerPanel () {
        setLayout (new java.awt.BorderLayout ());
    }

    /* Add a listener to the explorer panel in addition to the normal
    * open behaviour.
    */
    public void open () {
        open(TopManager.getDefault().getWindowManager().getCurrentWorkspace());
    }

    /* Add a listener to the explorer panel in addition to the normal
    * open behaviour.
    */
    public void open (Workspace workspace) {
        super.open(workspace);
        if (managerListener == null) {
            managerListener = new PropL ();
            getExplorerManager ().addPropertyChangeListener (managerListener);
        }
        setActivatedNodes (getExplorerManager ().getSelectedNodes ());
        updateTitle ();
    }

    /* Remove listeners to the explorer panel.
    */
    public boolean canClose (Workspace workspace, boolean last) {
        boolean result = super.canClose(workspace, last);
        if (result && last) {
            getExplorerManager ().removePropertyChangeListener (managerListener);
            managerListener = null;
        }
        return result;
    }

    /* Provides the explorer manager to all who are interested.
    * @return the manager
    */
    public ExplorerManager getExplorerManager () {
        if (manager == null) {
            synchronized (this) {
                if (manager == null) {
                    manager = new ExplorerManager ();
                }
            }
        }
        return manager;
    }

    /* Activates copy/cut/paste actions.
    */
    protected void componentActivated () {
        actions.attach (getExplorerManager ());
    }

    /* Deactivates copy/cut/paste actions.
    */
    protected void componentDeactivated () {
        actions.detach ();
    }

    /** Called when the explored context changes.
    * The default implementation updates the title of the window.
    */
    protected void updateTitle () {
        String name = ""; // NOI18N

        ExplorerManager em = getExplorerManager ();
        if (em != null) {
            Node n = em.getExploredContext();
            if (n != null) {
                String nm = n.getDisplayName();
                if (nm != null) {
                    name = nm;
                }
            }
        }

        if (formatExplorerTitle == null) {
            formatExplorerTitle = new MessageFormat (NbBundle.getMessage (ExplorerPanel.class, "explorerTitle"));
        }
        setName(formatExplorerTitle.format (
                    new Object[] { name }
                ));
    }

    /** Get context help for an explorer window.
    * Looks at the manager's node selection.
    * @return the help context
    * @see #getHelpCtx(Node[],HelpCtx)
    */
    public HelpCtx getHelpCtx () {
        return getHelpCtx (getExplorerManager ().getSelectedNodes (),
                           new HelpCtx (ExplorerPanel.class));
    }

    /** Utility method to get context help from a node selection.
    * Tries to find context helps for selected nodes.
    * If there are some, and they all agree, uses that.
    * In all other cases, uses the supplied generic help.
    * @param sel a list of nodes to search for help in
    * @param def the default help to use if they have none or do not agree
    * @return a help context
    */
    public static HelpCtx getHelpCtx (Node[] sel, HelpCtx def) {
        HelpCtx result = null;
        for (int i = 0; i < sel.length; i++) {
            HelpCtx attempt = sel[i].getHelpCtx ();
            if (attempt != null && ! attempt.equals (HelpCtx.DEFAULT_HELP)) {
                if (result == null || result.equals (attempt)) {
                    result = attempt;
                } else {
                    // More than one found, and they conflict. Get general help on the Explorer instead.
                    result = null;
                    break;
                }
            }
        }
        if (result != null)
            return result;
        else
            return def;
    }

    /** Set whether deletions should have to be confirmed on all Explorer panels.
    * @param confirmDelete <code>true</code> to confirm, <code>false</code> to delete at once
    */
    public static void setConfirmDelete (boolean confirmDelete) {
        actions.setConfirmDelete (confirmDelete);
    }

    /** Are deletions confirmed on all Explorer panels?
    * @return <code>true</code> if they must be confirmed
    */
    public static boolean isConfirmDelete () {
        return actions.isConfirmDelete ();
    }

    /** Stores the manager */
    public void writeExternal (ObjectOutput oo) throws IOException {
        super.writeExternal (oo);
        oo.writeObject (new NbMarshalledObject (manager));
    }

    /** Reads the manager.
    * Deserialization may throw {@link SafeException} in case
    * the manager cannot be loaded correctly but the stream is still uncorrupted.
    */
    public void readExternal (ObjectInput oi)
    throws IOException, ClassNotFoundException {
        super.readExternal (oi);
        Object anObj = oi.readObject ();
        if (anObj instanceof ExplorerManager) {
            manager = (ExplorerManager)anObj;
            return;
        }
        NbMarshalledObject obj = (NbMarshalledObject) anObj;
        
        // --- read all data from main stream, it is OK now ---
        try {
            manager = (ExplorerManager) obj.get ();
        } catch (SafeException se) {
//            System.err.println("rethrowing se: " + se.getLocalizedMessage ());
            throw se;
        } catch (IOException ioe) {
//            System.err.println("wrapping non-se: " + ioe.getClass ().getName () + " " + ioe.getLocalizedMessage ());
            throw new SafeException (ioe);
        }
    }

    /** Listener on the explorer manager properties.
    * Changes selected nodes of this frame.
    */
    private final class PropL extends Object implements PropertyChangeListener {
        public void propertyChange(PropertyChangeEvent evt) {
            if (ExplorerManager.PROP_SELECTED_NODES.equals(evt.getPropertyName())) {
                setActivatedNodes (manager.getSelectedNodes ());
                return;
            }
            if (ExplorerManager.PROP_EXPLORED_CONTEXT.equals(evt.getPropertyName())) {
                updateTitle ();
                return;
            }
        }
    }

}
