/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.explorer.propertysheet;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.*;
import java.util.ResourceBundle;
import java.util.Vector;
import java.text.MessageFormat;
import javax.swing.*;

import org.openide.*;
import org.openide.explorer.propertysheet.editors.EnhancedPropertyEditor;
import org.openide.nodes.Node.Property;
import org.openide.nodes.PropertySupport.ReadWrite;
import org.openide.util.Utilities;
import org.openide.util.NbBundle;

/**
* A panel able to show and change property values (all the same property type, maybe for multiple nodes).
* If an application is interested in knowing when the {@link PropertyDetails} changes the value of a property,
* it can register itself as a listener for property change events by calling
* {@link #addPropertyChangeListener}. You may call {@link #getInputState} to determine
* if a property value is currently being read or changed.
*
* @author Jan Jancura, Jaroslav Tulach
* @version 0.34, May 21, 1998
*/
class PropertyDisplayer extends JPanel {
    /** generated Serialized Version UID */
    static final long serialVersionUID = 4059243024814666953L;

    private static final String PROP_CAN_EDIT_AS_TEXT = "canEditAsText"; // NOI18N
    
    private static final String PROP_TITLE = "title";
    
    // private helper variables ...................................................

    /** Links to proper PropertyDetails object. */
    private PropertyDetails               propertyDetails;
    /** proper property editor for this property */
    private PropertyEditor                propertyEditor;
    /** Value of this property. */
    private Object                        value;
    /** string representation of value of this property */
    private String                        stringValue;
    /** is true if the property value can be edited as text */
    private boolean                       canEditAsText;
    private String                        rollingItem = null;
    /** index of edited property in the indexed property or -1 */
    private int                           propertyIndex = -1;
    /** index of edited property in the indexed property or -1 */
    private Object                        lock;
    private Vector                        listener = new Vector (10, 10);

    // Properties

    /** is true if the write component may be automatically changed
        to read comp. after entering some input */
    private boolean                       switchAutomatically = false;
    /** is false when component for showing value is showen
        (not component for changing value) */
    private boolean                       isWriteState = false;
    private int                           paintingStyle;
    /** Foreground color of values. */
    private Color                         foregroundColor;
    /** Foreground color of disabled values. */
    private Color                         disabledColor;
    /** Is plastic property value. */
    private boolean                       plastic;
    /** Text for tool tip. */
    private String                        toolTip;
    private long                          lastDeselectTime;

    // private variables for visual controls ...........................................

    /** component for showing property value is stored here */
    private SheetButton                   readComponent = null;
    private Component                     propertyCustomEditor = null;
    /** component that is actually used for input */
    private Component                     writeComponent = null;

    // constructors ....................................................................
    
    /**
    * Construct a displayer.
    */
    PropertyDisplayer () {
        lock = this;
        paintingStyle = PropertySheet.PAINTING_PREFERRED;
        setLayout (new BorderLayout ());
        setValueAsObject (null);
        setWriteState ();
    }

    /**
    * Construct a displayer for the given property details.
    */
    PropertyDisplayer (PropertyDetails propertyDetails) {
        this (
            propertyDetails,
            new PropertyValue (propertyDetails),
            PropertySheet.PAINTING_PREFERRED,
            null,
            null,
            null,
            false
        );
    }

    /**
    * Constructs PropertyDisplayer.
    */
    PropertyDisplayer (
        PropertyDetails propertyDetails,
        int paintingStyle,
        Object lock,
        Color foregroundColor,
        Color disabledColor,
        boolean plastic
    ) {
        this (
            propertyDetails,
            new PropertyValue (propertyDetails),
            paintingStyle,
            lock,
            foregroundColor,
            disabledColor,
            plastic
        );
    }

    /**
    * Constructs PropertyDisplayer.
    */
    PropertyDisplayer (
        PropertyDetails aPropertyDetails,
        PropertyValue propertyValue,
        int aPaintingStyle,
        Object lock,
        Color foregroundColor,
        Color disabledColor,
        boolean plastic
    ) {
        this.lock = lock;
        this.plastic = plastic;
        this.foregroundColor = foregroundColor;
        this.disabledColor = disabledColor;
        if (this.lock == null) this.lock = this;
        propertyDetails = aPropertyDetails;
        paintingStyle = aPaintingStyle;
        propertyEditor = propertyDetails.getPropertyEditor ();
        setLayout (new BorderLayout ());
        setValue (propertyValue);
        setSwitchAutomatically (true);
    }


    // main public methods ................................................................

    /**
    * Set the value of displayed property.
    * @param value the new value
    */
    public void setValueAsObject (Object value) {
        this.value = value;
        setValueAsProperty (
            new ReadWrite (
                "", // NOI18N
                (PropertyDisplayer.this.value == null) ?
                Object.class :
                PropertyDisplayer.this.value.getClass (),
                "", // NOI18N
                "" // NOI18N
            ) {
                public Object getValue () {
                    return PropertyDisplayer.this.value;
                }
                public void setValue (Object val) throws IllegalArgumentException {
                    PropertyDisplayer.this.value = val;
                }
            }
        );
    }

    /**
    * Set the value of the displayed details based on a node property.
    * @param value the node property used to set the displayed property (or property set)
    */
    public void setValueAsProperty (Property value) {
        if (value == null) return;
        propertyDetails = new PropertyDetails (
                              null,
                              value
                          );
        propertyEditor = propertyDetails.getPropertyEditor ();
        setValue (new PropertyValue (propertyDetails));
    }


    /**
    * Sets a new value of property as PropertyValue.
    */
    void setValue (PropertyValue value) {
        this.value = null;
        if (readComponent != null) releaseReadComponent (readComponent);
        if (isWriteState) {
            readComponent = null;

            releaseWriteComponent(false);
            JComponent inputComp = getWriterComponent ();

            // null means that input component wasn't change since last time asked for it
            if (inputComp != null) {
                removeAll();
                add ("Center", inputComp); // NOI18N
            }
        } else {
            removeAll ();
            add ("Center", readComponent = getReaderComponent (value)); // NOI18N
            attachReadComponent (readComponent);
        }
    }

    /**
    * Get the value of the displayed property.
    * @return the value
    */
    public Object getValue () {
        return value;
    }

    /**
    * Set whether component is being used to enter a value.
    * @param b <code>true</code> to enter a value, <code>false</code> to only display it
    */
    public void setInputState (boolean b) {
        if (b) setWriteState ();
        else setReadState ();
    }

    /**
    * Test whether component is displaying or entering a value.
    * @return <code>true</code> if entering, <code>false</code> if only displaying
    */
    public boolean getInputState () {
        return isWriteState;
    }

    /*
    * Sets tool tip.
    */
    public void setToolTipText (String text) {
        toolTip = text;
        if (readComponent != null) readComponent.setToolTipText (text);
    }

    /*
    * Sets enabled property value.
    */
    public void setEnabled (boolean enabled) {
        if (readComponent != null) readComponent.setEnabled (enabled);
        super.setEnabled (enabled);
    }

    /**
    * Set whether the component is in the pressed state.
    * @param pressed <code>true</code> if so
    */
    public void setPressed (boolean pressed) {
        if (readComponent != null) readComponent.setPressed (pressed);
    }

    /**
    * Set whether to switch beetween the read and wtite component automatically.
    * For example, this would apply to an input line which should be replaced
    * by property sheet button after hitting Enter.
    * @param b <code>true</code> if so
    */
    public void setSwitchAutomatically (boolean b) {
        if (b == switchAutomatically) return;
        switchAutomatically = b;
    }

    /**
    * Test whether to close automatically.
    * @return <code>true</code> if so
    * @see #setCloseAutomatically
    */
    public boolean getSwitchAutomatically () {
        return switchAutomatically;
    }

    /**
    * Get the time of the last deselection.
    * @return the time
    */
    public long getLastDeselectTime () {
        return lastDeselectTime;
    }


    // private helper methods ....................................................................

    /**
    * Switches from reading component to writing one.
    */
    void setWriteState () {
        rollingItem = null;

        // detach listeners from input component and clear input related fields
        releaseWriteComponent(true);
        
        removeAll ();
        JComponent c = getWriterComponent ();
        //    if (toolTip != null) c.setToolTipText (toolTip);
        add ("Center", c); // NOI18N

        Container co = getParent ();
        if (co != null) {
            co.invalidate ();
            co.getParent ().validate ();
        }
        repaint ();

        c.requestDefaultFocus ();
        isWriteState = true;
    }

    /**
    * Switches from writing component to reading one.
    */
    void setReadState () {
        if (!isWriteState) return;
        isWriteState = false;

        // #10797
        // This assures that value typed in textfield/inputChoice will be set if propertysheet switches
        // this displayer to readState
        if (writeComponentListener != null) {
            if (textField != null) writeComponentListener.changeValue(textField.getText());
            else if ((inputChoice != null) && (inputChoice.isEditable())) 
                writeComponentListener.changeValue((String)inputChoice.getEditor().getItem());
        }

        // detach listeners from input component and clear input related fields
        releaseWriteComponent(true);

        removeAll ();
        if (readComponent == null) {
            PropertyValue value = new PropertyValue (propertyDetails);
            readComponent = getReaderComponent (value);
            attachReadComponent (readComponent);
        }
        add ("Center", readComponent); // NOI18N

        Component c = getParent ();
        if (c != null) {
            c.invalidate ();
            c = getParent ();
            if (c != null) c.validate ();
        }
        repaint ();

        lastDeselectTime = System.currentTimeMillis ();
        requestFocus();
    }

    // reader component ........................

    /**
    * Creates Reader component.
    */
    private SheetButton getReaderComponent (PropertyValue value) {
        stringValue = null;
        canEditAsText = true;
        FeatureDescriptor p = propertyDetails.getFirstProperty();
        Object customEditAsText = p.getValue(PROP_CAN_EDIT_AS_TEXT);
        if (customEditAsText instanceof Boolean) {
            canEditAsText = ((Boolean)customEditAsText).booleanValue();
        }

        SheetButton c = null;
        if (!value.canRead ()) {//except., prop.editor == null or nonequals values
            if (value.getException () != null) //exception while canRead ()
                c = getTextView (getExceptionString (value.getException ()));
            else
                c =  getTextView (getTypeString (propertyDetails.getValueType ()));
        } else
            try {
                propertyEditor.setValue (this.value = value.getValue ());
                stringValue = propertyEditor.getAsText ();
                if ((stringValue == null) && (customEditAsText == null)){
                    canEditAsText = false;
                }
                if ( propertyEditor.isPaintable () &&
                        ( (paintingStyle == PropertySheet.PAINTING_PREFERRED) ||
                          ( (paintingStyle == PropertySheet.STRING_PREFERRED) &&
                            (stringValue == null)
                          )
                        )
                   )
                    c = getPaintView ();
                else
                    c = getTextView (
                            stringValue == null ?
                            getTypeString (propertyDetails.getValueType ()) :
                            stringValue
                        );
            } catch (Throwable e) {
                //exception while setValue () | getAsText () | isPaintable ()
                if (e instanceof ThreadDeath)
                    throw (ThreadDeath)e;
                if (this.value == null) {
                    // PE can not show null value
                    c = getTextView ("null"); // NOI18N
                } else {
                    Throwable t = TopManager.getDefault().getErrorManager().annotate(
                            e,
                            getString("PS_ExcIn") +
                            " " + propertyEditor.getClass ().getName () + // NOI18N
                            " " + getString ("PS_Editor") + "."
                    );
                    TopManager.getDefault().getErrorManager().notify(t);
                }
            }
        // we should never return null (dstrupl)
        if (c == null) {
            c = getTextView("null");
        }
        return c;
    }

    /**
    * Creates SheetButton with text representing current value of property.
    */
    private SheetButton getTextView (String str) {
        SheetButton c = new SheetButton (str, plastic, plastic);
        if (toolTip != null) c.setToolTipText (toolTip);
        if (foregroundColor != null) c.setActiveForeground (foregroundColor);
        if (disabledColor != null) c.setInactiveForeground (disabledColor);
        return c;
    }

    /**
    * Creates SheetButton with PropertyShow representing current value of property.
    */
    private SheetButton getPaintView () {
        PropertyShow propertyShow = new PropertyShow (propertyEditor);
        propertyShow.setForeground(foregroundColor);
        SheetButton c = new SheetButton ();
        c.add (propertyShow);
        if (toolTip != null) c.setToolTipText (toolTip);
        return c;
    }

    // writer component ........................

    /**
    * This method returns property value editor Component like input line (if property supports
    * setAsText (String string) method) or some others.
    *
    * @return property value editor Component
    */
    private JComponent getWriterComponent () {
        if (propertyEditor == null) return getDisabledWriterComponent();

        stringValue = null;
        canEditAsText = true;
        FeatureDescriptor p = propertyDetails.getFirstProperty();
        Object customEditAsText = p.getValue(PROP_CAN_EDIT_AS_TEXT);
        if (customEditAsText instanceof Boolean) {
            canEditAsText = ((Boolean)customEditAsText).booleanValue();
        }

        try {
            propertyEditor.setValue (
                this.value = new PropertyValue (propertyDetails).getValue ()
            );
            stringValue = propertyEditor.getAsText ();
            if ((stringValue == null) && (customEditAsText == null)){
                canEditAsText = false;
            }
        } catch (Throwable e) {//exception while getAsText ()
            if (e instanceof ThreadDeath)
                throw (ThreadDeath)e;
        }
        boolean existsCustomEditor = propertyEditor.supportsCustomEditor ();
        if ((propertyEditor instanceof EnhancedPropertyEditor) && (((EnhancedPropertyEditor)propertyEditor).hasInPlaceCustomEditor ()))
            return getInput (getInPlace (), existsCustomEditor);

        if (!propertyDetails.canWrite ()) {
            if (!existsCustomEditor) return getDisabledWriterComponent();
            return getInput (getDisabledWriterComponent(), true);// read-only
        }

        String[] tags;// tags
        int k;
        boolean editable = (propertyEditor instanceof EnhancedPropertyEditor) &&
                           (((EnhancedPropertyEditor)propertyEditor).supportsEditingTaggedValues ());
        if ( ((tags = propertyEditor.getTags ()) != null) && ((k = tags.length) > 0) )
            return getInput (getInputTag (tags, rollingItem = stringValue, editable), existsCustomEditor);

        if (canEditAsText) return getInput (getInputLine ((stringValue == null) ? "???" : stringValue, true), // NOI18N
                                                existsCustomEditor);
        if (existsCustomEditor) return getInput (getDisabledWriterComponent(), true);
        
        return getDisabledWriterComponent();
    }

    private JComponent getDisabledWriterComponent() {
       PropertyValue value = new PropertyValue (propertyDetails);
       return getReaderComponent (value);
    }

    JComboBox inputChoice;
    
    /**
    * This is helper method for method getWriterComponent () which returns Panel with Choice
    * in the "Center" and enhanced property editor open button on the "East". This Panel
    * is then returned as property value editor Component.
    *
    * @param tags There are lines for Choice stored.
    * @param selected Line to be selected.
    *
    * @return Choice Component
    */
    private JComponent getInputTag (String[] tags, final String selected, boolean editable) {
        if (inputChoice == null)
            inputChoice = new JComboBox ();
        inputChoice.setModel(new DefaultComboBoxModel(tags));
        
        inputChoice.setMaximumRowCount(tags.length <= 12 ? tags.length : 8);

        int i, k = tags.length;
        for (i = 0; i < k; i++) if (tags [i].equals (selected)) break;
        if (i < k) inputChoice.setSelectedIndex (i); // setSelectedItem cann't be used...
        // inputChoice.setSelectedItem (selected); selects the first item if item
        //  isn't in list...

        if (writeComponentListener == null)
            writeComponentListener = new WriteComponentListener();
        writeComponentListener.setPropertyStringValue(selected);
        inputChoice.addActionListener (writeComponentListener);

        if (editable) {
            inputChoice.setEditable (true);
            inputChoice.setSelectedItem (selected);
            inputChoice.getEditor ().setItem (selected);
            inputChoice.getEditor ().getEditorComponent().addFocusListener(writeComponentListener);
        } else {
            // patch
            inputChoice.addFocusListener (writeComponentListener);
            inputChoice.putClientProperty (
                "JComboBox.lightweightKeyboardNavigation", // NOI18N
                "Lightweight" // NOI18N
            );  // ugly hack for JDK bug 4199622
        }

        inputChoice.setToolTipText (toolTip);
        return inputChoice;
    }

    private Component getInPlace () {
        Component c = ((EnhancedPropertyEditor) propertyEditor).
                      getInPlaceCustomEditor ();
        propertyEditor.addPropertyChangeListener (
            new PropertyChangeListener () {
                public void propertyChange (PropertyChangeEvent e) {
                    propertyEditor.removePropertyChangeListener (this);
                    if (!isWriteState) return;
                    Object o = propertyEditor.getValue ();
                    synchronized (lock) {
                        if (switchAutomatically) setReadState ();
                        setPropertyValue (o);
                        notifyPropertyChange (e);
                    }
                    if (!switchAutomatically) 
                      propertyEditor.addPropertyChangeListener (this);
                }
            }
        );
        c.addFocusListener (new FocusAdapter () {
            public void focusLost (FocusEvent e) {
                if (!isWriteState) return;
                if (switchAutomatically) setReadState ();
            }
        });
        if (c instanceof JComponent) ((JComponent) c).setToolTipText (toolTip);
        return c;
    }

    JTextField textField;
    /** Listener on textField and inputChoice */
    WriteComponentListener writeComponentListener;
    
    /**
    * This is helper method for method getWriterComponent () which returns Panel with TextField
    * in the "Center" and enhanced property editor open button on the "East". This Panel
    * is then returned as property value editor Component.
    *
    * @param String propertyStringValue initial property value.
    * @param boolean editable is true if string editing should be allowed.
    * @param boolean existsCustomEditor is true if enhanced property editor open button
    *  should be showen.
    *
    * @return Panel Component
    */
    private JComponent getInputLine (final String propertyStringValue, boolean editable) {
        if (textField == null)
            textField = new JTextField ();

        if (writeComponentListener == null)
            writeComponentListener = new WriteComponentListener();

        textField.addActionListener (writeComponentListener);
        textField.addKeyListener(writeComponentListener);
        textField.addFocusListener (writeComponentListener);

        writeComponentListener.setPropertyStringValue(propertyStringValue);
        textField.setText(propertyStringValue);
        textField.setEditable (editable);
        textField.setToolTipText (toolTip);

//        if (!isWriteState)
//            textField.selectAll();

        return textField;
    }
    
    private void releaseWriteComponent(boolean clear) {
        if (textField != null) {
            textField.removeActionListener (writeComponentListener);
            textField.removeKeyListener(writeComponentListener);
            textField.removeFocusListener (writeComponentListener);
        }
        if (inputChoice != null) {
            inputChoice.removeActionListener (writeComponentListener);
            inputChoice.removeFocusListener (writeComponentListener);
            inputChoice.getEditor().getEditorComponent().removeFocusListener(writeComponentListener);
        }
        if (clear) {
            textField = null;
            inputChoice = null;
            writeComponentListener = null;
            writeComponent = null;
        }
    }

    /**
    * This is helper method for method getInput () and getInputTag () which returns Panel
    * with enhanced property editor open button on the "East".
    *
    * @param Component leftComponent this component will be added to the "Center" of this panel
    * @param boolean existsCustomEditor is true if enhanced property editor open button
    *  should be showen.
    *
    * @return Panel Component
    */
    private JComponent getInput (
        Component leftComponent,
        boolean existsCustomEditor
    ) {
        // if left component haven't changed
        if ((leftComponent != null) && (leftComponent == writeComponent)) return null;
        else writeComponent = leftComponent;

        JPanel panel;
        if ( (leftComponent == null) &&
                (propertyEditor != null) &&
                (propertyEditor.isPaintable ()) &&
                (paintingStyle != PropertySheet.ALWAYS_AS_STRING)
           ) {
            panel = new PropertyShow (propertyEditor);
        } else
            panel = new JPanel ();

        panel.setLayout (new BorderLayout());
        if (leftComponent != null) panel.add ("Center", leftComponent); // NOI18N

        if (existsCustomEditor) {
            SheetButton button = new SheetButton ("..."); // NOI18N
            Font currentFont = button.getFont ();
            button.setFont (
                new java.awt.Font (
                    currentFont.getName (),
                    currentFont.getStyle () | java.awt.Font.BOLD,
                    currentFont.getSize ()
                )
            );
            if (switchAutomatically)
                button.addFocusListener (new FocusAdapter () {
                    public void focusLost (FocusEvent e) {
                        if (!isWriteState) return;
                        setReadState ();
                    }
                });
            button.addSheetButtonListener (new CustomPEListener ());
            button.setToolTipText (toolTip);
            panel.add ("East", button); // NOI18N
        }
        panel.setToolTipText (toolTip);
        return panel;
    }

    // other .........................

    private void releaseReadComponent (SheetButton readComponent) {
        int i, k = listener.size ();
        for (i = 0; i < k; i++)
            readComponent.removeSheetButtonListener (
                (SheetButtonListener) listener.elementAt (i)
            );
    }

    private void attachReadComponent (SheetButton readComponent) {
        int i, k = listener.size ();
        for (i = 0; i < k; i++)
            readComponent.addSheetButtonListener (
                (SheetButtonListener) listener.elementAt (i)
            );      
    }

    boolean rolling (boolean down) {
        if (rollingItem == null) return false;
        String[] tags = propertyEditor.getTags ();
        int i, k = tags.length;
        for (i = 0; i < k; i++) if (tags [i].equals (rollingItem)) break;
        if (i >= (k - 1)) i = 0;
        else i++;
        setAsText (tags [i]);
        return true;
    }

    void setAsText (String value) {
        try {
            propertyEditor.setAsText (value);
        } catch (IllegalArgumentException ex) {
            ErrorManager em = TopManager.getDefault().getErrorManager();
            ErrorManager.Annotation[] an = em.findAnnotations(ex);
            String userMessage = null;
            int userSeverity = org.openide.ErrorManager.INFORMATIONAL;
            
            if (an != null) {
                for (int i = 0; i < an.length; i++) {
                    String msg = an[i].getLocalizedMessage();
                    if (msg != null) {
                        userMessage = msg;
                        userSeverity = an[i].getSeverity();
                        break;
                    }
                }
            }
            if (userMessage != null && (userSeverity == org.openide.ErrorManager.USER
                    || userSeverity == org.openide.ErrorManager.WARNING
                    || userSeverity == org.openide.ErrorManager.ERROR) ) {
                em.notify(ex);
            }
            else if (userMessage != null) {
                Throwable t = TopManager.getDefault().getErrorManager().annotate(
                    ex,
                    MessageFormat.format(
                        getString("FMT_ErrorSettingProperty"),
                        new Object[] { userMessage }
                    )
                );
                em.notify (org.openide.ErrorManager.USER, t);
            } else {
                em.notify (org.openide.ErrorManager.INFORMATIONAL, ex);
            }
            // return;
        } catch (Throwable ee) {
            if (ee instanceof ThreadDeath) throw (ThreadDeath)ee;
            if (switchAutomatically) setReadState ();
            Throwable t = TopManager.getDefault().getErrorManager().annotate(
                ee,
                MessageFormat.format(
                    getString("FMT_ErrorSettingProperty"),
                    new Object[] { ee.getLocalizedMessage() }
                )
            );
            TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, t);
            return;
        }
        Object o = propertyEditor.getValue ();
        synchronized (lock) {
            if (switchAutomatically) setReadState ();
            Object origPropertyValue = null;
            try {
                origPropertyValue = propertyDetails.getPropertyValue();
            } catch (Exception x) {
            }
      boolean oldIsNull = (origPropertyValue == null);
      boolean newIsNull = (o == null);
if ((newIsNull != oldIsNull) || (!newIsNull && !o.equals(origPropertyValue))) {
            //if (o != origPropertyValue) {
                setPropertyValue (o);
                notifyPropertyChange (new PropertyChangeEvent (this, null, null, null));
            }
        }
    }

    private String getExceptionString (Throwable exception) {
        if (exception instanceof java.lang.reflect.InvocationTargetException)
            exception = ((java.lang.reflect.InvocationTargetException) exception).
                        getTargetException ();
        return "<" + exception.getClass().getName() + ">"; // NOI18N
    }

    private String getTypeString (Class clazz) {
        if (propertyDetails.isArray ())
            return "[" + getString ("PS_ArrayOf") +" " +
                   propertyDetails.getIndexedValueType ().getName () + "]"; // NOI18N
        return "[" + clazz.getName() + "]"; // NOI18N
    }

    // delegating methods

    /**
    * This method is called when property value should be changed. It redirects this
    * call to the proper PropertyDetails object.
    */
    void setPropertyValue (Object o) {

        //S ystem.out.println ("\nsetPropertyValue : " + o); // NOI18N
        //T hread.dumpStack ();
        try {
            propertyDetails.setPropertyValue (value = o);
            if (textField != null) {
                try {
                    textField.setText(propertyEditor.getAsText());
                } catch (Exception x) {
                    TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, x);
                }
            }
        } catch (Exception e) {
            if (switchAutomatically) setReadState ();
            org.openide.TopManager.getDefault ().notifyException (e);
        }
    }

    /**
    * Delegate.
    * @return delegated value
    * @see PropertyDetails#supportsDefaultValue
    */
    public boolean supportsDefaultValue () {
        return propertyDetails.supportsDefaultValue ();
    }

    /**
    * Delegate.
    * @see PropertyDetails#restoreDefaultValue
    */
    public void restoreDefaultValue () {
        propertyDetails.restoreDefaultValue ();
        notifyPropertyChange (new PropertyChangeEvent (this, null, null, null));
    }

    /**
    * Get the property details object delegated to.
    * @return the property details
    */
    public PropertyDetails getPropertyDetails () {
        return propertyDetails;
    }

    public String toString () {
        if (propertyDetails == null)
            return "No PropertyDetails"; // NOI18N
        return "PropertyDetails [" + propertyDetails.getName () + "]"; // NOI18N
    }


    // listeners .................................................................

    /** Vector of all PropertyChngeListeners. */
    private Vector propertyChangeListeners = new Vector (3,10);

    /**
    * Standart helper method.
    */
    void notifyPropertyChange (PropertyChangeEvent e) {
        int i, k = propertyChangeListeners.size ();
        for (i = 0; i < k; i++)
            ((PropertyChangeListener)propertyChangeListeners.elementAt (i)).
            propertyChange (e);
    }

    /*
    * Standart helper method.
    */
    public void addPropertyChangeListener (PropertyChangeListener al) {
        propertyChangeListeners.addElement (al);
    }

    /*
    * Standart helper method.
    */
    public void removePropertyChangeListener (PropertyChangeListener al) {
        propertyChangeListeners.removeElement (al);
    }

    /**
    * Standart helper method.
    */
    private void fireSheetButtonClicked (ActionEvent e) {
        Vector listener = (Vector) this.listener.clone ();
        int i, k = listener.size ();
        for (i = 0; i < k; i++)
            ((SheetButtonListener)listener.elementAt (i)).sheetButtonClicked (e);
    }

    void addSheetButtonListener (SheetButtonListener sheetListener) {
        listener.addElement (sheetListener);
        if (readComponent != null)
            readComponent.addSheetButtonListener (sheetListener);
    }

    void removeSheetButtonListener (SheetButtonListener sheetListener) {
        listener.removeElement (sheetListener);
        if (readComponent != null)
            readComponent.removeSheetButtonListener (sheetListener);
    }

    private static String getString(String key) {
        return NbBundle.getBundle(PropertyDisplayer.class).getString(key);
    }

    // innerclasses ..............................................................

    private final class CustomPEListener implements SheetButtonListener {
        public void sheetButtonEntered (ActionEvent e) {}
        public void sheetButtonExited (ActionEvent e) {}

        public void sheetButtonClicked (ActionEvent e) {
            if (switchAutomatically) setReadState ();

            final PropertyEditor propertyEditor = propertyDetails.getNewPropertyEditor ();
            if (!(propertyEditor instanceof IndexedPropertyEditor))
                try {
                    value = propertyDetails.getPropertyValue ();
                    propertyEditor.setValue (value);
                } catch (Throwable ee) {
                    if (ee instanceof ThreadDeath) throw (ThreadDeath)ee;
                    Throwable t = TopManager.getDefault().getErrorManager().annotate(
                        ee,
                        // XXX this is what MessageFormat (NbBundle.getMessage(...)) is for!
                        getString ("PS_ExcIn") +
                        " " + propertyEditor.getClass ().getName () +
                        " " + getString ("PS_Editor") + "."
                    );
                    TopManager.getDefault().getErrorManager().notify(t);
                    return;
                }

            Component propertyCustomEditor = propertyDetails.getPropertyCustomEditor
                                             (propertyEditor);
            if (propertyCustomEditor == null) return;

            Window dialog = null;
            if (propertyCustomEditor instanceof Window) {
                final PropertyChangeListener pcl = new PropertyChangeListener () {
                    public void propertyChange (PropertyChangeEvent ev) {
                        Object o = null;
                        o = propertyEditor.getValue ();
                        synchronized (lock) {
                            setPropertyValue (o);
                            notifyPropertyChange (ev);
                        }
                    }
                };
                propertyEditor.addPropertyChangeListener (pcl);
                ((Window)propertyCustomEditor).addWindowListener (new WindowAdapter () {
                    public void windowClosed (WindowEvent evv) {
                        propertyEditor.removePropertyChangeListener (pcl);
                    }
                });
                dialog = (Window) propertyCustomEditor;
            } else { // it is not a window --> create dialog
                String title = null;
                Object[] titleParams = new Object[] {
                    propertyDetails.getName(),
                    Utilities.getShortClassName(propertyDetails.getValueType())
                };
                if (propertyCustomEditor instanceof JComponent) {
                    // try to get the title from JComponents client property
                    JComponent jc = (JComponent)propertyCustomEditor;
                    Object to = jc.getClientProperty(PROP_TITLE);
                    if (to instanceof String) {
                        title = (String)to;
                    }
                }
                if (propertyDetails.canWrite ()) {
                    if (title == null) {
                        title = MessageFormat.format(
                            getString ("PS_EditorTitle"),
                            titleParams
                        );
                    }
                    dialog = new PropertyDialogManager (
                        title,
                        propertyCustomEditor,
                        true,
                        PropertyDisplayer.this,
                        propertyEditor,
                        lock
                    ).getDialog ();
                }
                else {
                    if (title == null) {
                        title = MessageFormat.format(
                            getString ("PS_ViewerTitle"),
                            titleParams
                        );
                    }
                    dialog = new PropertyDialogManager (
                        title,
                        propertyCustomEditor,
                        true
                    ).getDialog ();
                }
            }
            dialog.pack ();
            dialog.show ();
            
            if (isWriteState)
                requestDefaultFocus();
        }
    }

    /** Listener on textField(JTextField) or inputChoice(JComboBox) - input line components
     * controlled by displayer.
     */
    private final class WriteComponentListener extends KeyAdapter implements ActionListener, FocusListener {
        
        private String oldStringValue = null;
        // preventing cycles with this flag - note that this
        // is really ugly (just hot fix for #11680)
        private boolean changingValue = false;
        
        /** Setter for old value.
         */
        public void setPropertyStringValue(String oldValue) {
            oldStringValue = oldValue;
        }
        
        public void actionPerformed (ActionEvent e) {
            if (!isWriteState) return;

            if (e.getSource() == inputChoice) {
                changeValue((String) inputChoice.getSelectedItem ());
                return;
            }
            
            if (e.getSource() == textField)
                changeValue (textField.getText ());
        }
        
        public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
                textField.setText(oldStringValue == null ? "" : oldStringValue);
                if (switchAutomatically) setReadState ();
            }
        }
        
        public void focusLost (FocusEvent e) {
            if (!isWriteState || (!switchAutomatically && e.isTemporary())) return;

            if (e.getSource() == textField) {
                changeValue(textField.getText ());
            } else if (inputChoice != null) {
                changeValue((String) inputChoice.getEditor().getItem());
                return;
            }
        }

        public void focusGained(FocusEvent e)  {
            if (e.getSource() == textField) {
                textField.selectAll();
            } else if (inputChoice != null) {
                inputChoice.getEditor ().selectAll ();
            }
        }

        /** Changes value if needed.
         */
        public void changeValue(String s) {
            if (changingValue) {
                return;
            }
            changingValue = true;
            try {
                if ( (s != null) && 
                     !s.equals (oldStringValue)
                ) setAsText (s);
                else if (switchAutomatically) setReadState ();
            } finally { // the try/finally here is probably paranoic
                        //  but we want to make sure that changingValue is
                        //  always set to false
                changingValue = false;
            }
        }
    }
}
