/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.explorer.propertysheet.editors;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.util.*;
import java.net.URL;
import java.util.ResourceBundle;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.border.*;

import org.openide.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.actions.CallableSystemAction;
import org.openide.util.HelpCtx;
import org.openide.explorer.propertysheet.editors.EnhancedCustomPropertyEditor;

/**
* PropertyEditor for Icons. Depends on existing DataObject for images.
* Images must be represented by some DataObject which returns itselv
* as cookie, and has image file as a primary file. File extensions
* for images is specified in isImage method.
*
* @author Jan Jancura
*/
public class IconEditor extends Object implements PropertyEditor, XMLPropertyEditor {

    public static final int TYPE_URL = 1;
    public static final int TYPE_FILE = 2;
    public static final int TYPE_CLASSPATH = 3;

    static final String URL_PREFIX = "URL"; // NOI18N
    static final String FILE_PREFIX = "File"; // NOI18N
    static final String CLASSPATH_PREFIX = "Classpath"; // NOI18N

    /** Standard variable for localization. */
    static java.util.ResourceBundle bundle = org.openide.util.NbBundle.getBundle(
                IconEditor.class);

    public static boolean isImage(String s) {
        s = s.toLowerCase();
        return s.endsWith(".jpg") || s.endsWith(".gif") || // NOI18N
               s.endsWith(".jpeg") || s.endsWith(".jpe") || // NOI18N
               s.equals("jpg") || s.equals("gif") || // NOI18N
               s.equals("jpeg") || s.equals("jpe"); // NOI18N
    }

    static String convert(String s) {
        StringTokenizer st = new StringTokenizer(s, "\\"); // NOI18N
        StringBuffer sb = new StringBuffer();
        if (st.hasMoreElements()) {
            sb.append(st.nextElement());
            while (st.hasMoreElements())
                sb.append("\\\\").append(st.nextElement()); // NOI18N
        }
        return new String(sb);
    }


    // variables .................................................................................

    private Icon icon;

    // init .......................................................................................

    public IconEditor() {
    }

    // Special access methods......................................................................

    /** @return the type of image source - one of TYPE_CLASSPATH, TYPE_FILE, TYPE_URL */
    public int getSourceType() {
        if (getValue() instanceof NbImageIcon)
            return ((NbImageIcon)getValue()).type;
        else
            return TYPE_FILE;
    }

    /** @return the name of image's source - depending on the type it can be a URL, file name or
    * resource path to the image on classpath */
    public String getSourceName() {
        if (getValue() instanceof NbImageIcon)
            return ((NbImageIcon)getValue()).name;
        else
            return null;
    }

    // PropertyEditor methods .....................................................................

    /**
    * @return The value of the property.  Builtin types such as "int" will
    * be wrapped as the corresponding object type such as "java.lang.Integer".
    */
    public Object getValue() {
        return icon;
    }

    /**
    * Set (or change) the object that is to be edited.  Builtin types such
    * as "int" must be wrapped as the corresponding object type such as
    * "java.lang.Integer".
    *
    * @param value The new target object to be edited.  Note that this
    *     object should not be modified by the PropertyEditor, rather
    *     the PropertyEditor should create a new object to hold any
    *     modified value.
    */
    public void setValue(Object object) {
        icon = (Icon) object;
    }

    /**
    * @return The property value as a human editable string.
    * <p>   Returns null if the value can't be expressed as an editable string.
    * <p>   If a non-null value is returned, then the PropertyEditor should
    *       be prepared to parse that string back in setAsText().
    */
    public String getAsText() {
        Object val = getValue();

        if (val == null) return "null"; // NOI18N

        if (val instanceof NbImageIcon) {
            NbImageIcon ii = (NbImageIcon)val;
            switch (ii.type) {
                case TYPE_URL:
                    return URL_PREFIX + ": " + ii.name; // NOI18N
                case TYPE_FILE:
                    return FILE_PREFIX + ": " + ii.name; // NOI18N
                case TYPE_CLASSPATH:
                    return CLASSPATH_PREFIX + ": " + ii.name; // NOI18N
            }
        }
        return null;
    }

    /**
    * Set the property value by parsing a given String.  May raise
    * java.lang.IllegalArgumentException if either the String is
    * badly formatted or if this kind of property can't be expressed
    * as text.
    * @param text  The string to be parsed.
    */
    public void setAsText(String string) throws IllegalArgumentException {
        setValue(iconFromText(string));
    }

    private NbImageIcon iconFromText(String string) throws IllegalArgumentException {
        NbImageIcon ii;
        try {
            if (string.startsWith(FILE_PREFIX)) {
                String s = string.substring(FILE_PREFIX.length() + 1).trim();
                ii = new NbImageIcon(s);
                ii.type = TYPE_FILE;
                ii.name = s;
            } else
                if (string.startsWith(CLASSPATH_PREFIX)) {
                    String s = string.substring(CLASSPATH_PREFIX.length() + 1).trim();
                    URL url = TopManager.getDefault().currentClassLoader().getResource(s);
                    ii = new NbImageIcon(url);
                    ii.type = TYPE_CLASSPATH;
                    ii.name = s;
                } else
                    if (string.startsWith(URL_PREFIX)) {
                        String s = string.substring(URL_PREFIX.length() + 1).trim();
                        URL url = new URL(s);
                        ii = new NbImageIcon(url);
                        ii.type = TYPE_URL;
                        ii.name = s;
                    } else
                        if (string.equals("null")) {
                            ii = null;
                        }
                        else {
                            ii = new NbImageIcon(string.trim());
                            ii.type = TYPE_FILE;
                            ii.name = string;
                        }
        } catch (Exception e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
            throw new IllegalArgumentException(e.toString());
        }
        return ii;
    }
        
    /**
    * This method is intended for use when generating Java code to set
    * the value of the property.  It should return a fragment of Java code
    * that can be used to initialize a variable with the current property
    * value.
    * <p>
    * Example results are "2", "new Color(127,127,34)", "Color.orange", etc.
    *
    * @return A fragment of Java code representing an initializer for the
    *     current value.
    */
    public String getJavaInitializationString() {
        if (getValue() instanceof NbImageIcon) {
            NbImageIcon ii = (NbImageIcon)getValue();
            switch (ii.type) {
                case TYPE_URL: return
                    "new javax.swing.JLabel() {\n" + // NOI18N
                    "  public javax.swing.Icon getIcon() {\n" + // NOI18N
                    "    try {\n" + // NOI18N
                    "      return new javax.swing.ImageIcon(\n" + // NOI18N
                    "        new java.net.URL(\"" + convert(ii.name) + "\")\n" + // NOI18N
                    "      );\n" + // NOI18N
                    "    } catch (java.net.MalformedURLException e) {\n" + // NOI18N
                    "    }\n" + // NOI18N
                    "    return null;\n" + // NOI18N
                    "  }\n" + // NOI18N
                    "}.getIcon()"; // NOI18N
                case TYPE_FILE: return
                    "new javax.swing.ImageIcon(\"" + convert(ii.name) + "\")"; // NOI18N
                case TYPE_CLASSPATH: return
                    "new javax.swing.ImageIcon(getClass().getResource(\"" + convert(ii.name) + "\"))"; // NOI18N
            }
        }
        return "null"; // NOI18N
    }

    /**
    * If the property value must be one of a set of known tagged values,
    * then this method should return an array of the tags.  This can
    * be used to represent (for example) enum values.  If a PropertyEditor
    * supports tags, then it should support the use of setAsText with
    * a tag value as a way of setting the value and the use of getAsText
    * to identify the current value.
    *
    * @return The tag values for this property.  May be null if this
    *   property cannot be represented as a tagged value.
    *
    */
    public String[] getTags() {
        return null;
    }

    /**
    * @return  True if the class will honor the paintValue method.
    */
    public boolean isPaintable() {
        return false;
    }

    /**
    * Paint a representation of the value into a given area of screen
    * real estate.  Note that the propertyEditor is responsible for doing
    * its own clipping so that it fits into the given rectangle.
    * <p>
    * If the PropertyEditor doesn't honor paint requests (see isPaintable)
    * this method should be a silent noop.
    * <p>
    * The given Graphics object will have the default font, color, etc of
    * the parent container.  The PropertyEditor may change graphics attributes
    * such as font and color and doesn't need to restore the old values.
    *
    * @param gfx  Graphics object to paint into.
    * @param box  Rectangle within graphics object into which we should paint.
    */
    public void paintValue(Graphics g, Rectangle rectangle) {
    }

    /**
    * @return  True if the propertyEditor can provide a custom editor.
    */
    public boolean supportsCustomEditor() {
        return true;
    }

    /**
    * A PropertyEditor may choose to make available a full custom Component
    * that edits its property value.  It is the responsibility of the
    * PropertyEditor to hook itself up to its editor Component itself and
    * to report property value changes by firing a PropertyChange event.
    * <P>
    * The higher-level code that calls getCustomEditor may either embed
    * the Component in some larger property sheet, or it may put it in
    * its own individual dialog, or ...
    *
    * @return A java.awt.Component that will allow a human to directly
    *      edit the current property value.  May be null if this is
    *      not supported.
    */
    public java.awt.Component getCustomEditor() {
        return new IconPanel();
    }

    /**
    * Register a listener for the PropertyChange event.  When a
    * PropertyEditor changes its value it should fire a PropertyChange
    * event on all registered PropertyChangeListeners, specifying the
    * null value for the property name and itself as the source.
    *
    * @param listener  An object to be invoked when a PropertyChange
    *    event is fired.
    */
    public void addPropertyChangeListener(PropertyChangeListener propertyChangeListener) {
    }

    /**
    * Remove a listener for the PropertyChange event.
    *
    * @param listener  The PropertyChange listener to be removed.
    */
    public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener) {
    }


    // innerclasses ...............................................................

    public static class NbImageIcon extends ImageIcon implements Externalizable {
        /** generated Serialized Version UID */
        static final long serialVersionUID = 7018807466471349466L;
        int type;
        String name;

        public NbImageIcon() {
        }

        NbImageIcon(URL url) {
            super(url);
            type = TYPE_URL;
        }

        NbImageIcon(String file) {
            super(file);
            type = TYPE_FILE;
        }

        String getName() {
            return name;
        }

        public void writeExternal(ObjectOutput oo) throws IOException {
            oo.writeObject(new Integer(type));
            oo.writeObject(name);
        }

        public void readExternal(ObjectInput in)
        throws IOException, ClassNotFoundException {
            type = ((Integer)in.readObject()).intValue();
            name = (String) in.readObject();
            ImageIcon ii = null;
            switch (type) {
            case TYPE_URL:
                try {
                    ii = new ImageIcon(new URL(name));
                } catch (java.net.MalformedURLException e) {
                    if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
                    return;
                }
                break;
            case TYPE_FILE:
                ii = new ImageIcon(name);
                break;
            case TYPE_CLASSPATH:
                ii = new ImageIcon(TopManager.getDefault().currentClassLoader().getResource(name));
                break;
            }
            setImage(ii.getImage());
        }
    }

 
    class IconPanel extends javax.swing.JPanel implements EnhancedCustomPropertyEditor {

        static final long serialVersionUID = -6904264999063788703L;        
        
        private javax.swing.JPanel jPanel1;
        
        private javax.swing.JLabel jLabel1;
        
        private javax.swing.JRadioButton rbUrl;
        
        private javax.swing.JRadioButton rbFile;
        
        private javax.swing.JRadioButton rbClasspath;
        
        private javax.swing.JRadioButton rbNoPicture;
        
        private javax.swing.JLabel jLabel2;
        
        private javax.swing.JLabel jLabel3;
        
        private javax.swing.JLabel jLabel4;
        
        private javax.swing.JLabel jLabel5;
        
        private javax.swing.JPanel jPanel2;
        
        private javax.swing.JLabel lName;
        
        private javax.swing.JTextField tfName;
        
        private javax.swing.JButton bSelect;
        
        private javax.swing.JPanel jPanel3;
        
        private javax.swing.JLabel jLabel7;
        
        private javax.swing.JScrollPane spImage;
        
        /** Creates new form IconPanel  */
        public IconPanel() {
            initComponents();
            jLabel1.setText(bundle.getString("CTL_ImageSourceType"));           //NoI18N
            rbUrl.setText(bundle.getString("CTL_URL"));                         //NoI18N
            rbFile.setText(bundle.getString("CTL_File"));                       //NoI18N
            rbClasspath.setText(bundle.getString("CTL_Classpath"));             //NoI18N
            rbNoPicture.setText(bundle.getString("CTL_NoPicture"));             //NoI18N
            jLabel2.setText(bundle.getString("CTL_URLExample"));                //NoI18N
            jLabel3.setText(bundle.getString("CTL_FileExample"));               //NoI18N
            jLabel4.setText(bundle.getString("CTL_ClasspathExample"));          //NoI18N
            jLabel5.setText(bundle.getString("CTL_Null"));                      //NoI18N
            lName.setText(bundle.getString("CTL_ImageSourceName"));             //NoI18N
            lName.setDisplayedMnemonic(bundle.getString("CTL_ImageSourceName_mnemonic").charAt(0)); //NoI18N
            lName.setLabelFor(tfName);
            jLabel7.setText(bundle.getString("CTL_Preview"));        //NoI18N
            bSelect.setText(bundle.getString("CTL_ButtonSelect"));   //NoI18N
            bSelect.setMnemonic(bundle.getString("CTL_ButtonSelect_mnemonic").charAt(0));   //NoI18N
            
            ButtonGroup bg = new ButtonGroup();
            bg.add(rbUrl);
            bg.add(rbFile);
            bg.add(rbClasspath);
            bg.add(rbNoPicture);
            
            Icon i = (Icon)getValue();
            if (i == null || !(i instanceof NbImageIcon)) {
                rbNoPicture.setSelected(true);
                bSelect.setEnabled(false);
                tfName.setEnabled(false);
                return;
            }

            switch (((NbImageIcon)i).type) {
            case TYPE_URL:
                rbUrl.setSelected(true);
                bSelect.setEnabled(false);
                break;
            case TYPE_FILE:
                rbFile.setSelected(true);
                bSelect.setEnabled(true);
                break;
            case TYPE_CLASSPATH:
                rbClasspath.setSelected(true);
                bSelect.setEnabled(true);
                break;
            }
            tfName.setText(((NbImageIcon)i).name);

            HelpCtx.setHelpIDString(this, "gui.csh.icon"); // NOI18N

            updateIcon();
        }
        
        /** This method is called from within the constructor to
         * initialize the form.
         * WARNING: Do NOT modify this code. The content of this method is
         * always regenerated by the Form Editor.
 */
        private void initComponents() {
            jPanel1 = new javax.swing.JPanel();
            jLabel1 = new javax.swing.JLabel();
            rbUrl = new javax.swing.JRadioButton();
            rbFile = new javax.swing.JRadioButton();
            rbClasspath = new javax.swing.JRadioButton();
            rbNoPicture = new javax.swing.JRadioButton();
            jLabel2 = new javax.swing.JLabel();
            jLabel3 = new javax.swing.JLabel();
            jLabel4 = new javax.swing.JLabel();
            jLabel5 = new javax.swing.JLabel();
            jPanel2 = new javax.swing.JPanel();
            lName = new javax.swing.JLabel();
            tfName = new javax.swing.JTextField();
            bSelect = new javax.swing.JButton();
            jPanel3 = new javax.swing.JPanel();
            jLabel7 = new javax.swing.JLabel();
            spImage = new javax.swing.JScrollPane();
            
            setLayout(new java.awt.GridBagLayout());
            java.awt.GridBagConstraints gridBagConstraints2;
            
            jPanel1.setLayout(new java.awt.GridBagLayout());
            java.awt.GridBagConstraints gridBagConstraints1;
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.insets = new java.awt.Insets(12, 12, 0, 0);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(jLabel1, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 1;
            gridBagConstraints1.insets = new java.awt.Insets(12, 24, 0, 0);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(rbUrl, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 2;
            gridBagConstraints1.insets = new java.awt.Insets(0, 24, 0, 0);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(rbFile, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 3;
            gridBagConstraints1.insets = new java.awt.Insets(0, 24, 0, 0);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(rbClasspath, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 0;
            gridBagConstraints1.gridy = 4;
            gridBagConstraints1.insets = new java.awt.Insets(0, 24, 0, 0);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(rbNoPicture, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 1;
            gridBagConstraints1.insets = new java.awt.Insets(12, 5, 0, 12);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(jLabel2, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 2;
            gridBagConstraints1.insets = new java.awt.Insets(5, 5, 0, 12);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(jLabel3, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 3;
            gridBagConstraints1.insets = new java.awt.Insets(5, 5, 0, 12);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            jPanel1.add(jLabel4, gridBagConstraints1);
            
            gridBagConstraints1 = new java.awt.GridBagConstraints();
            gridBagConstraints1.gridx = 1;
            gridBagConstraints1.gridy = 4;
            gridBagConstraints1.insets = new java.awt.Insets(5, 5, 0, 12);
            gridBagConstraints1.anchor = java.awt.GridBagConstraints.WEST;
            gridBagConstraints1.weightx = 1.0;
            gridBagConstraints1.weighty = 1.0;
            jPanel1.add(jLabel5, gridBagConstraints1);
            
            gridBagConstraints2 = new java.awt.GridBagConstraints();
            gridBagConstraints2.fill = java.awt.GridBagConstraints.BOTH;
            add(jPanel1, gridBagConstraints2);
            
            jPanel2.setLayout(new java.awt.GridBagLayout());
            java.awt.GridBagConstraints gridBagConstraints3;
            
            gridBagConstraints3 = new java.awt.GridBagConstraints();
            gridBagConstraints3.insets = new java.awt.Insets(12, 12, 0, 0);
            gridBagConstraints3.anchor = java.awt.GridBagConstraints.WEST;
            jPanel2.add(lName, gridBagConstraints3);
            
            gridBagConstraints3 = new java.awt.GridBagConstraints();
            gridBagConstraints3.fill = java.awt.GridBagConstraints.HORIZONTAL;
            gridBagConstraints3.insets = new java.awt.Insets(12, 5, 0, 0);
            gridBagConstraints3.anchor = java.awt.GridBagConstraints.WEST;
            gridBagConstraints3.weightx = 1.0;
            jPanel2.add(tfName, gridBagConstraints3);
            
            gridBagConstraints3 = new java.awt.GridBagConstraints();
            gridBagConstraints3.insets = new java.awt.Insets(12, 5, 0, 17);
            gridBagConstraints3.anchor = java.awt.GridBagConstraints.WEST;
            jPanel2.add(bSelect, gridBagConstraints3);
            
            gridBagConstraints2 = new java.awt.GridBagConstraints();
            gridBagConstraints2.gridx = 0;
            gridBagConstraints2.gridy = 1;
            gridBagConstraints2.fill = java.awt.GridBagConstraints.BOTH;
            add(jPanel2, gridBagConstraints2);
            
            jPanel3.setLayout(new java.awt.GridBagLayout());
            java.awt.GridBagConstraints gridBagConstraints4;
            
            gridBagConstraints4 = new java.awt.GridBagConstraints();
            gridBagConstraints4.insets = new java.awt.Insets(12, 12, 0, 0);
            gridBagConstraints4.anchor = java.awt.GridBagConstraints.WEST;
            jPanel3.add(jLabel7, gridBagConstraints4);
            
            gridBagConstraints4 = new java.awt.GridBagConstraints();
            gridBagConstraints4.gridx = 0;
            gridBagConstraints4.gridy = 1;
            gridBagConstraints4.fill = java.awt.GridBagConstraints.BOTH;
            gridBagConstraints4.insets = new java.awt.Insets(5, 12, 0, 12);
            gridBagConstraints4.weightx = 1.0;
            gridBagConstraints4.weighty = 1.0;
            jPanel3.add(spImage, gridBagConstraints4);
            
            gridBagConstraints2 = new java.awt.GridBagConstraints();
            gridBagConstraints2.gridx = 0;
            gridBagConstraints2.gridy = 2;
            gridBagConstraints2.fill = java.awt.GridBagConstraints.BOTH;
            gridBagConstraints2.weightx = 1.0;
            gridBagConstraints2.weighty = 1.0;
            add(jPanel3, gridBagConstraints2);
        
            // listeners .................................................

            tfName.addActionListener(new ActionListener() {
                                          public void actionPerformed(ActionEvent e) {
                                              setValue();
                                          }
                                      });
            rbUrl.addActionListener(new ActionListener() {
                                         public void actionPerformed(ActionEvent e) {
                                             bSelect.setEnabled(false);
                                             tfName.setEnabled(true);
                                             setValue();
                                         }
                                     });
            rbFile.addActionListener(new ActionListener() {
                                          public void actionPerformed(ActionEvent e) {
                                              bSelect.setEnabled(true);
                                              tfName.setEnabled(true);
                                              setValue();
                                              updateIcon();
                                          }
                                      });
            rbClasspath.addActionListener(new ActionListener() {
                                               public void actionPerformed(ActionEvent e) {
                                                   bSelect.setEnabled(true);
                                                   tfName.setEnabled(true);
                                                   setValue();
                                               }
                                           });
            rbNoPicture.addActionListener(new ActionListener() {
                                               public void actionPerformed(ActionEvent e) {
                                                   bSelect.setEnabled(false);
                                                   tfName.setEnabled(false);
                                                   icon = null; //IconEditor.this.setValue(null);
                                                   updateIcon();
                                               }
                                           });
            bSelect.addActionListener(new ActionListener() {
                                           public void actionPerformed(ActionEvent e) {
                                               if(rbFile.isSelected()) {
                                                   JFileChooser chooser = new JFileChooser();
                                                   FileFilter filter = new FileFilter() {
                                                                           public boolean accept(java.io.File f) {
                                                                               return f.isDirectory() || isImage(f.getName());
                                                                           }
                                                                           public String getDescription() {
                                                                               return bundle.getString("CTL_ImagesExtensionName");
                                                                           }
                                                                       };
                                                   chooser.setFileFilter(filter);
                                                   chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES );
                                                   HelpCtx.setHelpIDString(chooser, "csh.openimage"); // NOI18N
                                                   int returnVal = chooser.showOpenDialog(
                                                                       IconPanel.this
                                                                   );
                                                   if (returnVal != JFileChooser.APPROVE_OPTION) return;
                                                   tfName.setText(chooser.getSelectedFile().getAbsolutePath());
                                                   setValue();
                                               } else
                                                   if (rbClasspath.isSelected()) {
                                                       //            InputPanel ip = new InputPanel();
                                                       Places places = TopManager.getDefault().getPlaces();
                                                       Node ds = places.nodes().repository(new DataFilter() {
                                                                                                 public boolean acceptDataObject(DataObject obj) {
                                                                                                     // accept only data folders but ignore read only roots of file systems
                                                                                                     if (obj instanceof DataFolder)
                                                                                                         return !obj.getPrimaryFile().isReadOnly() ||
                                                                                                                obj.getPrimaryFile().getParent() != null;
                                                                                                     return isImage(obj.getPrimaryFile().getExt());
                                                                                                 }
                                                                                             });

                                                       String name;
                                                       try {
                                                           // selects one folder from data systems
                                                           DataObject d = (DataObject)
                                                                          TopManager.getDefault().getNodeOperation().select(
                                                                              bundle.getString("CTL_OpenDialogName"),
                                                                              bundle.getString("CTL_FileSystemName"),
                                                                              TopManager.getDefault().getPlaces().nodes().repository(),
                                                                              new NodeAcceptor() {
                                                                                  public boolean acceptNodes(Node[] nodes) {
                                                                                      if ((nodes == null) || (nodes.length != 1))
                                                                                          return false;
                                                                                      return nodes[0].getCookie(DataFolder.class) == null;
                                                                                  }
                                                                              },
                                                                              null
                                                                          )[0].getCookie(DataObject.class);
                                                           name = (d.getPrimaryFile().getPackageNameExt('/', '.'));
                                                       } catch (org.openide.util.UserCancelException ex) {
                                                           return;
                                                       }
                                                       tfName.setText("/" + name); // NOI18N
                                                       setValue();
                                                   }
                                           }
                                       });
        }

        
        
        
        
        /**
         * @return Returns the property value that is result of the CustomPropertyEditor.
         * @exception InvalidStateException when the custom property editor does not
         * represent valid property value (and thus it should not be set)
 */
        public Object getPropertyValue() throws IllegalStateException {
            NbImageIcon ii = null;
            String s = tfName.getText().trim();
            if ("".equals(s)) { // NOI18N
                return null;
            }
            try {
                if (rbFile.isSelected()) {
                    ii = new NbImageIcon(s);
                    ii.type = TYPE_FILE;
                    ii.name = s;
                } else
                    if (rbClasspath.isSelected()) {
                        URL url = TopManager.getDefault().currentClassLoader().getResource(s);
                        ii = new NbImageIcon(url);
                        ii.type = TYPE_CLASSPATH;
                        ii.name = s;
                    } else
                        if (rbUrl.isSelected()) {
                            URL url = new URL(s);
                            ii = new NbImageIcon(url);
                            ii.type = TYPE_URL;
                            ii.name = s;
                        }
            } catch (Exception e) {
//                if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
//                throw new IllegalStateException(e.toString());
            }
            return ii;
        }
        
        void updateIcon() {
            Icon i = (Icon)getValue();
            spImage.setViewportView((i == null) ? new JLabel() : new JLabel(i));
            //      repaint();
            validate();
        }
        
        void setValue() {
            String val = tfName.getText();
            val.trim();
            if ("".equals(val)) { // NOI18N
                icon = null;
                return;
            }
            
            String pref = "";   // NOI18N
            if (rbUrl.isSelected()) pref = URL_PREFIX + ": "; // NOI18N
            else
                if (rbFile.isSelected()) pref = FILE_PREFIX + ": "; // NOI18N
                else
                    if (rbClasspath.isSelected()) pref = CLASSPATH_PREFIX + ": "; // NOI18N
            try {
                icon = iconFromText(pref + val);
            } catch (IllegalArgumentException ee) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) ee.printStackTrace(); // NOI18N
            }
            updateIcon();
        }
    }
    
    // end of IconPanel

    //--------------------------------------------------------------------------
    // XMLPropertyEditor implementation

    public static final String XML_IMAGE = "Image"; // NOI18N

    public static final String ATTR_TYPE = "iconType"; // NOI18N
    public static final String ATTR_NAME = "name"; // NOI18N

    /** Called to load property value from specified XML subtree. If succesfully loaded,
    * the value should be available via the getValue method.
    * An IOException should be thrown when the value cannot be restored from the specified XML element
    * @param element the XML DOM element representing a subtree of XML from which the value should be loaded
    * @exception IOException thrown when the value cannot be restored from the specified XML element
    */
    public void readFromXML(org.w3c.dom.Node element) throws java.io.IOException {
        if (!XML_IMAGE.equals(element.getNodeName())) {
            throw new java.io.IOException();
        }
        org.w3c.dom.NamedNodeMap attributes = element.getAttributes();
        try {
            int type = Integer.parseInt(attributes.getNamedItem(ATTR_TYPE).getNodeValue());
            String name = attributes.getNamedItem(ATTR_NAME).getNodeValue();
            switch (type) {
            case TYPE_URL: setAsText(URL_PREFIX + ": " + name); break; // NOI18N
            case TYPE_FILE: setAsText(FILE_PREFIX + ": " + name); break; // NOI18N
            case TYPE_CLASSPATH: setAsText(CLASSPATH_PREFIX + ": " + name); break; // NOI18N
            }
        } catch (NullPointerException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) e.printStackTrace(); // NOI18N
            throw new java.io.IOException(e.toString());
        }
    }

    /** Called to store current property value into XML subtree. The property value should be set using the
    * setValue method prior to calling this method.
    * @param doc The XML document to store the XML in - should be used for creating nodes only
    * @return the XML DOM element representing a subtree of XML from which the value should be loaded
    */
    public org.w3c.dom.Node storeToXML(org.w3c.dom.Document doc) {
        org.w3c.dom.Element el = doc.createElement(XML_IMAGE);
        NbImageIcon ii = (NbImageIcon)getValue();
        el.setAttribute(ATTR_TYPE, Integer.toString(ii.type));
        el.setAttribute(ATTR_NAME, ii.name);
        return el;
    }
}
