/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.explorer.view;

import java.awt.*;
import java.awt.event.*;
import java.awt.dnd.DnDConstants;
import java.beans.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import org.openide.*;
import org.openide.actions.PopupAction;
import org.openide.awt.MouseUtils;
import org.openide.explorer.*;
import org.openide.util.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.CallbackSystemAction;
import org.openide.nodes.Node;
import org.openide.nodes.NodeOp;
import org.openide.nodes.Children;


/** Tree view abstract class.
*
* @author   Petr Hamernik, Ian Formanek, Jaroslav Tulach
*/
public abstract class TreeView extends JScrollPane {
    //
    // static fields
    //


    /** generated Serialized Version UID */
    static final long serialVersionUID = -1639001987693376168L;

    /** How long it takes before collapsed nodes are released from the tree's cache
    */
    private static final int TIME_TO_COLLAPSE = System.getProperty ("netbeans.debug.heap") != null ? 0 : 15000;


    /** Minimum width of this component. */
    private static final int MIN_TREEVIEW_WIDTH = 400;

    /** Minimum height of this component. */
    private static final int MIN_TREEVIEW_HEIGHT = 400;

    //
    // components
    //

    /** Main <code>JTree</code> component. */
    transient protected JTree tree;

    /** model */
    transient private NodeTreeModel treeModel;

    /** Explorer manager, valid when this view is showing */
    transient private ExplorerManager manager;


    // Attributes

    /** not null if default action on nodes allowed */
    transient ClickAdapter defaultActionListener;

    /** not null if popup menu enabled */
    transient PopupAdapter popupListener;




    /** the most important listener (on four types of events */
    transient TreePropertyListener managerListener = null;
    /** weak variation of the listener for property change on the explorer manager */
    transient PropertyChangeListener wlpc;
    /** weak variation of the listener for vetoable change on the explorer manager */
    transient VetoableChangeListener wlvc;


    /** true if drag support is active */
    transient boolean dragActive = false;
    /** true if drop support is active */
    transient boolean dropActive = false;
    /** Drag support */
    transient TreeViewDragSupport dragSupport;
    /** Drop support */
    transient TreeViewDropSupport dropSupport;



    /** Constructor.
    */
    public TreeView () {
        this (true, true);
    }

    /** Constructor.
    * @param defaultAction should double click on a node open its default action?
    * @param popupAllowed should right-click open popup?
    */
    public TreeView (boolean defaultAction, boolean popupAllowed) {
        initializeTree ();

        // hack - DnD not stable now...
        // setDragSource(true);
        // setDropTarget(true);

        setPopupAllowed (popupAllowed);
        setDefaultActionAllowed (defaultAction);
    }

    /** Initializes the tree & model.
    */
    private void initializeTree () {
        // initilizes the JTree
        treeModel = createModel ();
        
        tree = new AutoscrollJTree(treeModel) {
                   public void updateUI () {
                       setCellRenderer(new NodeRenderer ());
                       super.updateUI ();
                   }
               };
        setViewportView (tree);

        NodeRenderer rend = NodeRenderer.sharedInstance ();
        tree.setCellRenderer(rend);
        tree.setCellEditor(new TreeViewCellEditor(tree, new NodeRenderer.Tree ()));
        tree.putClientProperty("JTree.lineStyle", "Angled"); // NOI18N
        tree.setEditable(true);

        ToolTipManager.sharedInstance().registerComponent(tree);


        // init listener & attach it to closing of
        managerListener = new TreePropertyListener();
        tree.addTreeExpansionListener (managerListener);

        // do not care about focus
        setRequestFocusEnabled (false);
        
        tree.addFocusListener(new FocusAdapter());
    }


    /** Is it permitted to display a popup menu?
     * @return <code>true</code> if so
     */
    public boolean isPopupAllowed () {
        return popupListener != null;
    }

    /** Enable/disable displaying popup menus on tree view items.
    * Default is enabled.
    * @param value <code>true</code> to enable
    */
    public void setPopupAllowed (boolean value) {
        if (popupListener == null && value) {
            // on
            popupListener = new PopupAdapter ();
            tree.addMouseListener (popupListener);
            return;
        }
        if (popupListener != null && !value) {
            // off
            tree.removeMouseListener (popupListener);
            popupListener = null;
            return;
        }
    }

    /** Does a double click invoke the default node action?
    * @return <code>true</code> if so
    */
    public boolean isDefaultActionEnabled () {
        return defaultActionListener != null;
    }

    /** Also requests focus for the tree component */
    public void requestFocus () {
        super.requestFocus();
        tree.requestFocus();
    }

    /** Enable/disable double click to invoke default action.
    * @param value <code>true</code> to enable
    */
    public void setDefaultActionAllowed (boolean value) {
        if (defaultActionListener == null && value) {
            // on
            defaultActionListener = new ClickAdapter ();
            tree.addMouseListener (defaultActionListener);
            tree.registerKeyboardAction(
                defaultActionListener,
                KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false),
                JComponent.WHEN_FOCUSED
            );
            return;
        }
        if (defaultActionListener != null && !value) {
            // off
            tree.removeMouseListener (defaultActionListener);
            tree.unregisterKeyboardAction(
                KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false)
            );
            defaultActionListener = null;
            return;
        }

    }

    /**
    * Is the root node of the tree displayed?
    *
    * @return <code>true</code> if so
    */
    public boolean isRootVisible() {
        return tree.isRootVisible();
    }

    /** Set whether or not the root node from
    * the <code>TreeModel</code> is visible.
    *
    * @param rootVisible <code>true</code> if it is to be displayed
    * @beaninfo
    *        bound: true
    *  description: Whether or not the root node
    *               from the TreeModel is visible.
    */
    public void setRootVisible (boolean visible) {
        tree.setRootVisible (visible);
    }


    /********** Support for the Drag & Drop operations *********/

    /** @return true if dragging from the view is enabled, false
    * otherwise.<br>
    * Drag support is disabled by default.
    */
    public boolean isDragSource () {
        return dragActive;
    }

    /** Enables/disables dragging support.
    * @param state true enables dragging support, false disables it.
    */
    public void setDragSource (boolean state) {
        if (state == dragActive)
            return;
        dragActive = state;
        // create drag support if needed
        if (dragActive && (dragSupport == null))
            dragSupport = new TreeViewDragSupport(this, tree);
        // activate / deactivate support according to the state
        dragSupport.activate(dragActive);
    }

    /** @return true if dropping to the view is enabled, false
    * otherwise<br>
    * Drop support is disabled by default.
    */
    public boolean isDropTarget () {
        return dropActive;
    }

    /** Enables/disables dropping support.
    * @param state true means drops into view are allowed,
    * false forbids any drops into this view.
    */
    public void setDropTarget (boolean state) {
        if (state == dropActive)
            return;
        dropActive = state;
        // create drop support if needed
        if (dropActive && (dropSupport == null))
            dropSupport = new TreeViewDropSupport(this, tree);
        // activate / deactivate support according to the state
        dropSupport.activate(dropActive);
    }

    /** @return Set of actions which are allowed when dragging from
    * asociated component.
    * Actions constants comes from DnDConstants.XXX constants.
    * All actions (copy, move, link) are allowed by default.
    */
    public int getAllowedDragActions () {
        // PENDING
        return DnDConstants.ACTION_MOVE | DnDConstants.ACTION_COPY |
               DnDConstants.ACTION_LINK;
    }

    /** Sets allowed actions for dragging
    * @param actions new drag actions, using DnDConstants.XXX 
    */  
    public void setAllowedDragActions (int actions) {
        // PENDING
    }

    /** @return Set of actions which are allowed when dropping
    * into the asociated component.
    * Actions constants comes from DnDConstants.XXX constants.
    * All actions are allowed by default.
    */
    public int getAllowedDropActions () {
        // PENDING
        return DnDConstants.ACTION_MOVE | DnDConstants.ACTION_COPY |
               DnDConstants.ACTION_LINK;
    }

    /** Sets allowed actions for dropping.
    * @param actions new allowed drop actions, using DnDConstants.XXX 
    */  
    public void setAllowedDropActions (int actions) {
        // PENDING
    }


    //
    // Control over expanded state
    //

    /** Collapses the tree under given node.
    * 
    * @param n node to collapse
    */
    public void collapseNode (Node n) {
        TreePath treePath = new TreePath (
                                treeModel.getPathToRoot (
                                    VisualizerNode.getVisualizer (null, n)
                                )
                            );
        tree.collapsePath (treePath);
    }

    /** Expandes the node in the tree.
    * 
    * @param n node 
    */
    public void expandNode (Node n) {
        TreePath treePath = new TreePath (
                                treeModel.getPathToRoot (
                                    VisualizerNode.getVisualizer (null, n)
                                )
                            );

        tree.expandPath (treePath);
    }

    /** Test whether a node is expanded in the tree or not
    * @param n the node to test
    * @return true if the node is expanded
    */
    public boolean isExpanded (Node n) {
        TreePath treePath = new TreePath (
                                treeModel.getPathToRoot (
                                    VisualizerNode.getVisualizer (null, n)
                                )
                            );
        return tree.isExpanded (treePath);
    }


    /** Expands all paths.
    */
    public void expandAll () {
        int i = 0, j, k = tree.getRowCount ();
        do {
            do {
                j = tree.getRowCount ();
                tree.expandRow (i);
            } while (j != tree.getRowCount ());
            i++;
        } while (i < tree.getRowCount ());
    }

    //
    // Processing functions
    //

    /* Initializes the component.
    */
    public void addNotify () {
        super.addNotify ();

        // Enter key in the tree


        ExplorerManager newManager = ExplorerManager.find (TreeView.this);

        if (newManager != manager) {
            if (manager != null) {
                manager.removeVetoableChangeListener (wlvc);
                manager.removePropertyChangeListener (wlpc);
            }

            manager = newManager;

            manager.addVetoableChangeListener(wlvc = WeakListener.vetoableChange (managerListener, manager));
            manager.addPropertyChangeListener(wlpc = WeakListener.propertyChange (managerListener, manager));

            synchronizeRootContext ();
            synchronizeExploredContext ();
            synchronizeSelectedNodes ();
        }

        tree.getSelectionModel().addTreeSelectionListener(managerListener);
    }

    /* Deinitializes listeners.
    */
    public void removeNotify () {
        super.removeNotify ();

        //System.out.println ("Calling remove notify..."); // NOI18N
        tree.getSelectionModel().removeTreeSelectionListener(managerListener);
    }

    /* Defines new way how to compute preffered size
    */
    public Dimension getPreferredSize() {
        Dimension dim = null;
        try {
            dim = super.getPreferredSize();
        } catch (NullPointerException npe) {
            dim = new Dimension(MIN_TREEVIEW_WIDTH, MIN_TREEVIEW_HEIGHT);
        } 
        if (dim.width < MIN_TREEVIEW_WIDTH)
            dim.width = MIN_TREEVIEW_WIDTH;
        if (dim.height < MIN_TREEVIEW_HEIGHT)
            dim.height = MIN_TREEVIEW_HEIGHT;
        return dim;
    }

    // *************************************
    // Methods to be overriden by subclasses
    // *************************************


    /** Allows subclasses to provide own model for displaying nodes.
    * @return the model to use for this view
    */
    protected abstract NodeTreeModel createModel();


    /** Called to allow subclasses to define the behaviour when a
    * node(s) are selected in the tree.
    *
    * @param nodes the selected nodes
    * @param em explorer manager to work on (change nodes to it)
    * @throws PropertyVetoException if the change cannot be done by the explorer
    *    (the exception is silently consumed)
    */
    protected abstract void selectionChanged (Node[] nodes, ExplorerManager em) throws PropertyVetoException;

    /** Called when explorer manager is about to change the current selection.
    * The view can forbid the change if it is not able to display such
    * selection.
    *
    * @param nodes the nodes to select
    * @return false if the view is not able to change the selection
    */
    protected abstract boolean selectionAccept (Node[] nodes);


    /** Show a given path in the screen. It depends on the kind of <code>TreeView</code>
    * if the path should be expanded or just made visible.
    *
    * @param path the path
    */
    protected abstract void showPath(TreePath path);

    /** Shows selection to reflect the current state of the selection in the explorer.
    *
    * @param paths array of paths that should be selected
    */
    protected abstract void showSelection (TreePath[] paths);


    /** Should a context menu of the explored context be used?
    * Applicable when no nodes are selected and the user wants to invoke
    * a context menu (clicks right mouse button).
    *
    * @return <code>true</code> if so; <code>false</code> in the default implementation
    */
    protected boolean useExploredContextMenu() {
        return false;
    }


    //
    // synchronizations
    //

    /** Called when selection in tree is changed.
    */
    final void callSelectionChanged (Node[] nodes) {
        manager.removePropertyChangeListener (wlpc);
        manager.removeVetoableChangeListener (wlvc);
        try {
            selectionChanged (nodes, manager);
        } catch (PropertyVetoException e) {
            synchronizeSelectedNodes ();
        } finally {
            manager.addPropertyChangeListener (wlpc);
            manager.addVetoableChangeListener (wlvc);
        }
    }


    /** Synchronize the root context from the manager of this Explorer.
    */
    final void synchronizeRootContext() {
        treeModel.setNode (manager.getRootContext ());
    }

    /** Synchronize the explored context from the manager of this Explorer.
    */
    final void synchronizeExploredContext() {
        Node n = manager.getExploredContext ();
        if (n != null) {
            TreePath treePath = new TreePath (treeModel.getPathToRoot (VisualizerNode.getVisualizer (null, n)));
            showPath(treePath);
        }
    }

    /** Synchronize the selected nodes from the manager of this Explorer.
    * The default implementation does nothing.
    */
    final void synchronizeSelectedNodes() {
        Node[] arr = manager.getSelectedNodes ();
        TreePath[] paths = new TreePath[arr.length];

        for (int i = 0; i < arr.length; i++) {
            TreePath treePath = new TreePath (treeModel.getPathToRoot (VisualizerNode.getVisualizer (null, arr[i])));
            paths[i] = treePath;
        }

        tree.getSelectionModel().removeTreeSelectionListener(managerListener);
        showSelection (paths);
        tree.getSelectionModel().addTreeSelectionListener(managerListener);
    }



    /** Listens to the property changes on tree */
    private class TreePropertyListener implements VetoableChangeListener,
                PropertyChangeListener,
                TreeExpansionListener,
                TreeSelectionListener, Runnable
    {
        public void vetoableChange(PropertyChangeEvent evt) throws PropertyVetoException {
            if (evt.getPropertyName().equals(ExplorerManager.PROP_SELECTED_NODES)) {
                if (!selectionAccept ((Node[])evt.getNewValue ())) {
                    throw new PropertyVetoException ("", evt); // NOI18N
                }
            }
        }

        public final void propertyChange(PropertyChangeEvent evt) {
            if (manager == null) return; // the tree view has been removed before the event got delivered
            if (evt.getPropertyName().equals(ExplorerManager.PROP_ROOT_CONTEXT))
                synchronizeRootContext();
            if (evt.getPropertyName().equals(ExplorerManager.PROP_EXPLORED_CONTEXT))
                synchronizeExploredContext();
            if (evt.getPropertyName().equals(ExplorerManager.PROP_SELECTED_NODES))
                synchronizeSelectedNodes();
        }

        public void treeExpanded (TreeExpansionEvent ev) {
            final TreePath path = ev.getPath ();
            
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    SwingUtilities.invokeLater (new Runnable () {
                        public void run () {
                            if (!tree.isVisible(path)) {
                                // if the path is not visible - don't check the children
                                return;
                            }
                            if (treeModel == null) {
                                // no model, no action, no problem
                                return;
                            }
                            TreeNode myNode = (TreeNode)path.getLastPathComponent();
                            if (treeModel.getPathToRoot(myNode)[0]
                            != treeModel.getRoot()) {
                                // the way from the path no longer
                                // goes to the root, probably someone
                                // has removed the node on the way up
                                // System.out.println("different roots.");
                                return;
                            }
                            int lastChildIndex = myNode.getChildCount()-1;
                            if (lastChildIndex >= 0) {
                                TreeNode lastChild = myNode.getChildAt(lastChildIndex);
                                // scroll to see the last child
                                tree.scrollPathToVisible(new TreePath(treeModel.getPathToRoot(lastChild)));
                                // in case the list is long scroll back to see the original path
                                tree.scrollPathToVisible(path);
                            }
                        }
                    });
                }
            }, 250); // hope that all children are there after this time
        }

        public void treeCollapsed (final TreeExpansionEvent ev) {
            final TreePath path = ev.getPath ();
            
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    SwingUtilities.invokeLater (new Runnable () {
                        public void run () {
                            if (tree.isExpanded(path)) {
                                // the tree shows the path - do not collapse
                                // the tree
                                return;
                            }

                            if (!tree.isVisible(path)) {
                                // if the path is not visible do not collapse
                                // the tree
                                return;
                            }

                            if (treeModel == null) {
                                // no model, no action, no problem
                                return;
                            }

                            TreeNode myNode = (TreeNode)path.getLastPathComponent();

                            if (treeModel.getPathToRoot(myNode)[0]
                            != treeModel.getRoot()) {
                                // the way from the path no longer
                                // goes to the root, probably someone
                                // has removed the node on the way up
                                // System.out.println("different roots.");
                                return;
                            }

                            treeModel.nodeStructureChanged(myNode);
                        }
                    });
                }
            }, TIME_TO_COLLAPSE);
        }

        /* Called whenever the value of the selection changes.
        * @param ev the event that characterizes the change.
        */
        public void valueChanged(TreeSelectionEvent ev) {
            TreePath[] paths = tree.getSelectionPaths ();
            if (paths == null) {
                callSelectionChanged (new Node[0]);
            } else {
                // we need to force no changes to nodes hierarchy =>
                // we are requesting read request, but it is not necessary
                // to execute the next action immediatelly, so postReadRequest
                // should be enough
                readAccessPaths = paths;
                Children.MUTEX.postReadRequest(this);
            }
        }
        
        private TreePath[] readAccessPaths;
       
        /** Called under Children.MUTEX to refresh the currently selected nodes.
        */
        public void run () {

            if (readAccessPaths == null) {
                return;
            }
            
            TreePath[] paths = readAccessPaths;
            // non null value caused leak in
            // ComponentInspector
            // When the last Form was closed then the ComponentInspector was
            // closed as well. Since this variable was not null - 
            // last selected Node (RADComponentNode) was held ---> FormManager2 was held, etc.
            readAccessPaths = null;
            java.util.List ll = new ArrayList(paths.length);
            
            for (int i = 0; i < paths.length; i++) {
                Node n = Visualizer.findNode (paths[i].getLastPathComponent ());
                if (n == manager.getRootContext() || n.getParentNode() != null) {
                    ll.add (n);
                }
            }

            callSelectionChanged ((Node[])ll.toArray (new Node[ll.size ()]));
        }

    } // end of TreePropertyListener


    /** Popup adapter.
    */
    private class PopupAdapter extends MouseUtils.PopupMouseAdapter {
        protected void showPopup (MouseEvent e) {
            int selRow = tree.getRowForLocation(e.getX(), e.getY());

            if (!tree.isRowSelected(selRow)) {
                tree.setSelectionRow(selRow);
            }
            
            // refresh the actions on the selected nodes by
            // deselecting and selecting again the selected nodes
            // [dstrupl]: didn't find better way how to do this
            // without an API change, something like manager.refresh() ???
            try {
                Node[] na = manager.getSelectedNodes();
                manager.setSelectedNodes(new Node[0]);
                manager.setSelectedNodes(na);    
            } catch (PropertyVetoException pve) {
                TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, pve);
            }
                
            if (selRow != -1) {
                createPopup(e.getX(), e.getY());
            }
        }
    }

    void createPopup(int xpos, int ypos) {
        JPopupMenu popup = NodeOp.findContextMenu(manager.getSelectedNodes());
        if ((popup != null) && (popup.getSubElements().length > 0)) {
            java.awt.Point p = getViewport().getViewPosition();
            p.x = xpos - p.x;
            p.y = ypos - p.y;
            SwingUtilities.convertPointToScreen (p, TreeView.this);
            Dimension popupSize = popup.getPreferredSize ();
            Dimension screenSize = Toolkit.getDefaultToolkit ().getScreenSize ();
            if (p.x + popupSize.width > screenSize.width) p.x = screenSize.width - popupSize.width;
            if (p.y + popupSize.height > screenSize.height) p.y = screenSize.height - popupSize.height;
            SwingUtilities.convertPointFromScreen (p, TreeView.this);
            popup.show(TreeView.this, p.x, p.y);
        }
    }

    final class PopupPerformer implements org.openide.util.actions.ActionPerformer {
        public void performAction(SystemAction act) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    int i = tree.getLeadSelectionRow();
                    if (i < 0) {
                        return;
                    }
                    Rectangle rect = tree.getRowBounds(i);
                    if (rect == null) {
                        return;
                    }
                    createPopup(rect.x, rect.y);
                }
            });
        }
    }
    
    final class FocusAdapter implements java.awt.event.FocusListener {
        
        CallbackSystemAction csa;
        PopupPerformer performer;

        public void focusGained(java.awt.event.FocusEvent ev) {
            if (csa == null) {
                csa = (CallbackSystemAction) SystemAction.get (PopupAction.class);
                performer = new PopupPerformer();
            }
            csa.setActionPerformer(performer);
        }
        
        public void focusLost(java.awt.event.FocusEvent ev) {
            if (csa != null) {
                csa.setActionPerformer(null);
            }
        }
    }    
    
    /** clicking adapter
    */
    private class ClickAdapter extends MouseAdapter implements ActionListener {
        public void mouseClicked(MouseEvent e) {
            int selRow = tree.getRowForLocation(e.getX(), e.getY());

            //Default action
            if ((selRow != -1) && SwingUtilities.isLeftMouseButton(e)) {

                TreePath selPath = tree.getPathForLocation(e.getX(), e.getY());
                Node node = Visualizer.findNode (selPath.getLastPathComponent());

                if (defaultActionListener != null && MouseUtils.isDoubleClick(e)) {
                    //System.out.println ("Double clicked???"); // NOI18N
                    SystemAction sa = node.getDefaultAction ();
                    if (sa != null) {
                        TopManager.getDefault ().getActionManager ().invokeAction
                            (sa, new ActionEvent (node, ActionEvent.ACTION_PERFORMED, "")); // NOI18N
                        e.consume ();
                    } else { // this is to be uncommented if we can force
                        // JTree not to expand / collapse on double click

                        Package p = Package.getPackage ("java.lang"); // NOI18N
                        if (p.isCompatibleWith("1.3")) { // NOI18N
                            if (tree.isExpanded(selRow))
                                tree.collapseRow(selRow);
                            else
                                tree.expandRow(selRow);
                        }
                    }
                }
            }
        }

        public void actionPerformed(ActionEvent evt) {
            Node[] nodes = manager.getSelectedNodes();
            if (nodes.length == 1) {
                SystemAction sa = nodes[0].getDefaultAction ();
                if (sa != null) {
                    TopManager.getDefault ().getActionManager ().invokeAction
                        (sa, new ActionEvent (nodes[0], ActionEvent.ACTION_PERFORMED, "")); // NOI18N
                }
            }
        }

    }

}
