/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.filesystems;
import java.io.*;

import org.openide.util.enum.QueueEnumeration;
/**
 *
 * @author  rmatous
 * @version
 */
class EventControl {
    /** number of requests posted and not processed. to
    * know what to do in sync.
    */
    private  int requests;
    /** List of requests */
    private QueueEnumeration requestsQueue;
     
    /**
     * Method that can fire events directly, postpone them, fire them in 
     * standalone thread or in RequestProcessor
     */
    void dispatchEvent (FileSystem.EventDispatcher dispatcher) {
      if (postponeFiring (dispatcher)) return;        
      dispatcher.run();
    }
    
    /**
     * Begin of block, that should be performed without firing events.
     * Firing of events is postponed after end of block .
     * There is strong necessity to use always both methods: beginAtomicAction
     * and finishAtomicAction. It is recomended use it in try - finally block.
     * @see FileSystemt#beginAtomicAction
     */        
    void beginAtomicAction () {
        enterAtomicAction ();
    }
    
    /**
     * End of block, that should be performed without firing events.
     * Firing of events is postponed after end of block .
     * There is strong necessity to use always both methods: beginAtomicAction
     * and finishAtomicAction. It is recomended use it in try - finally block.
     * @see FileSystemt#finishAtomicAction
     */        
    void finishAtomicAction () {
        exitAtomicAction ();
    }
        
    /** Executes atomic action. The atomic action represents a set of
    * operations constituting one logical unit. It is guaranteed that during
    * execution of such an action no events about changes in the file system
    * will be fired.*/    
    void runAtomicAction (final FileSystem.AtomicAction run) throws IOException {
        try {
            enterAtomicAction ();
            run.run ();
        } finally {
            exitAtomicAction ();
        }
    }

    /** Enters atomic action.
    */
    private  synchronized void enterAtomicAction () {
        if (requests++ == 0) {
            requestsQueue = new QueueEnumeration ();
        }
    }

    /** Exits atomic action.
    */
    private  void exitAtomicAction () {
        java.util.Enumeration myQueue;
        synchronized (this) {
            if (--requests == 0) {
                myQueue = requestsQueue;
                requestsQueue = null;
            } else {
                myQueue = null;
            }
        }

        if (myQueue != null) {
            while (myQueue.hasMoreElements()) {
                Runnable r = (Runnable)myQueue.nextElement();
                r.run();
            }
        }
    }
    /* Adds dispatcher to queue.*/
    private  synchronized boolean postponeFiring (Runnable run) {
        if (requestsQueue != null) {
            // run later
            requestsQueue.put (run);
            return true;
        }        
        return false;
    }
    
}
