/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.filesystems;

import org.openide.util.Lookup;

/** Contains utility methods to deal with repository and error manager,
* so we do not need to directly contact 
*
* @author Jaroslav Tulach
*/
final class ExternalUtil extends Object {
    /** value for the repository & error manager */
    private static Repository repository;
    private static org.openide.ErrorManager errMan;
    
    /** Static method to find the Repository to use.
     * @return Repository instance
     */
    public static Repository getRepository () {
        initialize ();
        return repository;
    }
    
    /** Notifies an exception.
     */
    public static void exception (Exception ex) {
        if (errMan != null) {
            errMan.notify (errMan.INFORMATIONAL, ex);
        } else {
            ex.printStackTrace();
        }
    }
    
    /** Copies anotation.
     * @param newEx new exception to annotate
     * @param oldEx old exception to take annotation from
     * @return newEx
     */
    public static Throwable copyAnnotation (Throwable newEx, Throwable oldEx) {
        if (errMan != null) {
            return errMan.copyAnnotation (newEx, oldEx);
        } else {
            // just return the new one
            return newEx;
        }
    }
    
    
    /** Annotates the exception with a message.
     */
    public static void annotate (Throwable ex, String msg) {
        if (errMan != null) {
            errMan.annotate (ex, msg);
        }
    }
    
    /** Annotates the exception with a message.
     */
    public static void annotate (Throwable ex, Throwable stack) {
        if (errMan != null) {
            errMan.annotate (ex, stack);
        }
    }
    
    
    /** Logs a text.
     */
    public static void log (String msg) {
        if (errMan != null) {
            errMan.log (errMan.INFORMATIONAL, msg);
        }
    }
    
    /** Loads a class of given name
     * @param name name of the class
     * @return the class
     * @exception ClassNotFoundException if class was not found
     */
    public static Class findClass (String name) throws ClassNotFoundException {
        initialize ();
        
        ClassLoader c = (ClassLoader)Lookup.getDefault ().lookup (ClassLoader.class);
        
        if (c == null) {
            return Class.forName (name);
        } else {
            return Class.forName (name, true, c);
        }
    }
        
    
    /** Initializes the context and errManager
     */
    private static void initialize () {
        if (repository == null) {
            initializeSynch ();
        }
    }
    
    /** Init the manager.
     */
    private static synchronized void initializeSynch () {
        if (repository != null) {
            return;
        }
        
        Lookup l = Lookup.getDefault ();
        repository = (Repository)l.lookup (org.openide.filesystems.Repository.class);
        errMan = (org.openide.ErrorManager)l.lookup (org.openide.ErrorManager.class);
        
        if (repository == null) {
            // if not provided use default one
            repository = new Repository (new XMLFileSystem ());
        }
    }
    
    /** Static method to find out the ErrorManager.
     */
    private static org.openide.ErrorManager getErrorManager () {
        return errMan;
    }

}
