/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.filesystems;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import java.net.UnknownServiceException;
import java.util.ResourceBundle;

/** Special URL connection directly accessing an internal file object.
*
* @author Ales Novak, Petr Hamernik, Jan Jancura, Jaroslav Tulach
*/
final class FileURL extends URLConnection {
    /** Protocol name for this type of URL. */
    public static final String PROTOCOL = "nbfs"; // NOI18N

    /** url separator */
    private static final char SEPARATOR = '/';
    
    /** Default implemenatation of handler for this type of URL.
     */
    private static URLStreamHandler HANDLER = new URLStreamHandler () {
        /**
        * @param u - URL to open connection to.
        * @return new URLConnection.
        */
        public URLConnection openConnection(URL u) throws IOException {
            return new FileURL (u);
        }
    };
    

    /** FileObject that we want to connect to. */
    private FileObject fo;


    /**
    * Create a new connection to a {@link FileObject}.
    * @param u URL of the connection. Please use {@link #encodeFileObject(FileObject)} to create the URL.
    */
    private FileURL(URL u) {
        super (u);
    }

    /** Provides a URL to access a file object.
    * @param fo the file object
    * @return a URL using the correct syntax and {@link #PROTOCOL protocol}
    * @exception FileStateInvalidException if the file object is not valid (typically, if its filesystem is inconsistent or no longer present)
    */
    public static URL encodeFileObject (FileObject fo) throws FileStateInvalidException {
        return encodeFileObject (fo.getFileSystem (), fo);
    }

    /** Encodes fileobject into URL.
    * @param fs file system the object is on
    * @param fo file object
    * @return URL
    */
    private static URL encodeFileObject (FileSystem fs, FileObject fo) {
        String fsName = encodeFileSystemName (fs.getSystemName());
        String fileName = fo.getPackageNameExt('/', '.');
        String url = fsName + SEPARATOR + fileName; // NOI18N
        try {
            return new java.net.URL (PROTOCOL, null, -1, url, HANDLER);
        } catch (java.net.MalformedURLException stillException) {
            // should not happend because this is our HANDLER, our connection, 
            // there should be no errors
            stillException.printStackTrace();
            throw new InternalError ();
        }
    }

    /** Retrieves the file object specified by an internal URL.
    * @param u the url to decode
    * @return the file object that is represented by the URL, or <code>null</code> if the URL is somehow invalid or the file does not exist
    */
    public static FileObject decodeURL (URL u) {
        if (!u.getProtocol ().equals (PROTOCOL)) return null;
        // resource name
        String resourceName = u.getFile ();
        if (resourceName.startsWith ("/")) resourceName = resourceName.substring (1); // NOI18N

        // first part is FS name
        int first = resourceName.indexOf ('/');
        if (first == -1) return null;

        String fileSystemName = decodeFileSystemName (resourceName.substring (0, first));
        resourceName = resourceName.substring (first);
        FileSystem fsys = ExternalUtil.getRepository ().findFileSystem(fileSystemName);

        return (fsys == null) ? null : fsys.findResource (resourceName);
    }

    /* A method for connecting to a FileObject.
    */
    public void connect() throws IOException {
        if (fo != null) return;

        fo = decodeURL (url);
        if (fo == null) {
            throw new FileNotFoundException("Cannot find: " + url); // NOI18N
        }
    }

    /*
    * @return InputStream or given FileObject.
    */
    public InputStream getInputStream()
    throws IOException, UnknownServiceException {
        connect ();
        try {
            if (fo.isFolder()) return new FIS (fo);
            return fo.getInputStream();
        } catch (FileNotFoundException e) {
            ExternalUtil.exception (e);
            throw e;
        }
    }

    /*
    * @return OutputStream for given FileObject.
    */
    public OutputStream getOutputStream()
    throws IOException, UnknownServiceException {
        connect();
        if (fo.isFolder()) throw new UnknownServiceException();
        org.openide.filesystems.FileLock flock = fo.lock();
        return new LockOS (fo.getOutputStream(flock), flock);
    }

    /*
    * @return length of FileObject.
    */
    public int getContentLength() {
        try {
            connect();
            return (int)fo.getSize();
        } catch (IOException ex) {
            return 0;
        }
    }


    /** Get a header field (currently, content type only).
    * @param name the header name. Only <code>content-type</code> is guaranteed to be present.
    * @return the value (i.e., MIME type)
    */
    public String getHeaderField(String name) {
        if (name.equalsIgnoreCase("content-type")) { // NOI18N
            try {
                connect();
                if (fo.isFolder())
                    return "text/html"; // NOI18N
                else
                    return fo.getMIMEType ();
            }
            catch (IOException e) {
            }
        }
        return super.getHeaderField(name);
    }

    /** Encodes filesystem name.
    * @param fs original filesystem name
    * @return new encoded name
    */
    static String encodeFileSystemName (String fs) {
	// [PENDING] this ought to use standard URL encoding, not this weird scheme
        StringBuffer sb = new StringBuffer ();
        for (int i = 0; i < fs.length (); i++) {
            switch (fs.charAt (i)) {
            case 'Q':
                sb.append ("QQ"); // NOI18N
                break;
            case '/':
                sb.append ("QB"); // NOI18N
                break;
            case ':':
                sb.append ("QC"); // NOI18N
                break;
            case '\\':
                sb.append ("QD"); // NOI18N
                break;
            case '#':
                sb.append ("QE"); // NOI18N
                break;                
            default:
                sb.append (fs.charAt (i));
                break;
            }
        }
        return sb.toString ();
    }

    /** Decodes name to FS one.
    * @param name encoded name
    * @return original name of the filesystem
    */
    static String decodeFileSystemName (String name) {
        StringBuffer sb = new StringBuffer ();
        int i = 0;
        int len = name.length ();
        while (i < len) {
            char ch = name.charAt (i++);
            if (ch == 'Q' && i < len) {
                switch (name.charAt (i++)) {
                case 'B':
                    sb.append ('/');
                    break;
                case 'C':
                    sb.append (':');
                    break;
                case 'D':
                    sb.append ('\\');
                    break;
                case 'E':
                    sb.append ('#');
                    break;                    
                default:
                    sb.append ('Q');
                    break;
                }
            } else {
                // not Q
                sb.append (ch);
            }
        }
        return sb.toString ();
    }


    /** Stream that also closes the lock, if closed.
     */
    private static class LockOS extends java.io.BufferedOutputStream {
        /** underlying OutputStream */
        private OutputStream fos;

        /** lock */
        private FileLock flock;

        /**
        * @param os is an OutputStream for writing in
        * @param lock is a lock for the stream
        */
        public LockOS (OutputStream os, FileLock lock)
        throws IOException {
            super(os);
            flock = lock;
        }

        /** overriden */
        public void close()
        throws IOException {
            flock.releaseLock();
            super.close();
        }
    }
    
    /** The class allows reading of folder via URL. Because of html
    * oriented user interface the document has html format.
    *
    * @author Ales Novak
    * @version 0.10 May 15, 1998
    */
    private static final class FIS extends InputStream {

        /** delegated reader that reads the document */
        private StringReader reader;

        /**
        * @param folder is a folder
        */
        public FIS (FileObject folder)
        throws IOException {
            reader = new StringReader(createDocument(folder));
        }

        /** creates html document as string */
        private String createDocument(FileObject folder)
        throws IOException {
            StringBuffer buff = new StringBuffer(150);
            StringBuffer lit = new StringBuffer(15);
            FileObject[] fobia = folder.getChildren();
            String name;

            buff.append("<HTML>\n"); // NOI18N
            buff.append("<BODY>\n"); // NOI18N

            // [PENDING] these links are all broken when from nbfs: protocol;
            // they point to "server-root" but nbfs: (oddly) puts the file system
            // name as part of the file path, not as a "host", so this does not work

            FileObject parent = folder.getParent();
            if (parent != null) {
                lit.setLength(0);
                lit.append('/').append(parent.getPackageName('/'));
                buff.append("<P>"); // NOI18N
                buff.append("<A HREF=").append(lit).append(">").append("..").append("</A>").append("\n"); // NOI18N
                buff.append("</P>"); // NOI18N
            }

            for (int i = 0; i < fobia.length; i++) {
                lit.setLength(0);
                lit.append('/').append(fobia[i].getPackageName('/'));
                name = fobia[i].getName();
                if (!fobia[i].isFolder()) {
                    lit.append(".").append(fobia[i].getExt()); // NOI18N
                    name = name + "." + fobia[i].getExt(); // NOI18N
                }
                buff.append("<P>"); // NOI18N
                buff.append("<A HREF=").append(lit).append(">").append(name).append("</A>").append("\n"); // NOI18N
                buff.append("</P>"); // NOI18N
            }

            buff.append("</BODY>\n"); // NOI18N
            buff.append("</HTML>\n"); // NOI18N
            return buff.toString();
        }

        //************************************** stream methods **********
        public int read() throws IOException {
            return reader.read();
        }

        public int read(byte[] b, int off, int len) throws IOException {
            char[] ch = new char[len];
            int r = reader.read(ch, 0, len);
            for (int i = 0; i < r; i++)
                b[off + i] = (byte) ch[i];
            return r;
        }

        public long skip(long skip) throws IOException {
            return reader.skip(skip);
        }

        public void close() throws IOException {
            reader.close();
        }

        public void reset() throws IOException {
            reader.reset();
        }

        public void mark(int i) {
            try {
                reader.mark(i);
            } catch (IOException ex) {
            }
        }

        public boolean markSupported() {
            return reader.markSupported();
        }
    } // end of FIS
    
}

