/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights ReFserved.
 */

package org.openide.filesystems;

import java.beans.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.util.jar.*;
import java.lang.ref.*;

import org.openide.filesystems.*;
import org.openide.filesystems.FileSystem;
import org.openide.util.actions.SystemAction;
import org.openide.util.enum.EmptyEnumeration;
import org.openide.util.NbBundle;
import org.openide.util.Task;
import org.openide.ErrorManager;

import  org.openide.util.*;
import  org.openide.util.enum.*;

/** A virtual file system based on a JAR archive.
*
* @author Jan Jancura, Jaroslav Tulach, Petr Hamernik
*/
public class JarFileSystem extends AbstractFileSystem
    /*nbif compat
    implements AbstractFileSystem.List, AbstractFileSystem.Info, AbstractFileSystem.Change, AbstractFileSystem.Attr
    /*nbend*/
{
    /** generated Serialized Version UID */
    static final long serialVersionUID = -98124752801761145L;

    /** File system name prefix */
    private static final String JAR_FS = "JAR_FS "; // NOI18N

    /**
    * Opened zip file of this file system is stored here or null.
    */
    private transient JarFile jar;

    /** Manifest file for jar
    */
    private transient Manifest manifest;

    /** Archive file.1
    */
    private File root = new File ("."); // NOI18N

    
    private transient EntryRef retRef;

    /** One request proccesor shared for all instances of JarFileSystem*/
    private static RequestProcessor req = new RequestProcessor ("JarFs - modification watcher");            

    /** Task that scans the content of jar.*/
    private  transient Task parserTask = RequestProcessor.Task.EMPTY ;    
    
    /** Watches modification on root file */    
    private  transient RequestProcessor.Task watcherTask = null;
    
    private transient long lastModification = 0;
    
    /* Advantage usage in children method - modifies behaviour of this method according to current thread (Another behaviour for parsing thread). */
    private static ThreadLocal thrStatus = new ThreadLocal ();
    
    /*Should help to prevent closing JarFile if anybody has InputStream. Also this variable 
     is used as object for synchronization: synchronized(closeSync)*/
    private transient Object closeSync = new Object () ;
    
    /** Controlls the LocalFileSystem's automatic refresh.
    * If the refresh time interval is set from the System.property, than this value is used.
    * Otherwise, do the 15000 ms refresh. If the refresh time interval is set to 0,
    * the refresh is disabled. */    
    private static final int REFRESH_TIME  = Integer.getInteger
    ("org.openide.filesystems.JarFileSystem.REFRESH_TIME",15000).intValue();
    
    /**
    * Default constructor.
    */
    public JarFileSystem () {
        Impl impl = new Impl (this);

        this.list = impl;
        this.info = impl;
        this.change = impl;
        this.attr = impl; 
        
        parserTask = RequestProcessor.Task.EMPTY ;            
    }

    /**
    * Constructor that can provide own capability for the file system.
    * @param cap the capability
    */
    public JarFileSystem (FileSystemCapability cap) {
        this ();
        setCapability (cap);
    }

    /* Anonymous Runnable class - responsible for parsing => parsing = standalone thread. 
     * Parsing is invoked in setJarFile method, or when JarFile was modified (checks watcherTask periodically).
     */
    private Runnable parserTask () {        
        return new Runnable () {
            private transient AbstractFolder parentFo;            
            private static final int    STATUS_PARSER = 0;            
            public  void run () {
                thrStatus.set(new Integer (STATUS_PARSER));
                lastModification = getJarFile().lastModified();
                String name;
                
                if (jar == null || root == null) return;
                 
		try {
		    try {
			jar.close();
		    }
		    catch (IOException ex) {
		    }
		    JarFile newJar;
		    try {
			newJar = new JarFile(root);
		    }
		    catch (IOException ex) {
			// force watcher to reschedule us
			lastModification = 0;
			return;
		    }
		    jar = newJar;

//		    System.out.println(new Date().toString() +
//			"parsing jar file: " +
//			root.getPath());
		    boolean parseWorked = true;
		    java.util.List entries = new ArrayList();
		    try {
			Enumeration en = jar.entries ();
			while (en.hasMoreElements ()) {
			    JarEntry je = (JarEntry)en.nextElement ();
			    name = je.getName ();
			    entries.add(name);
			}
		    }
		    catch(Throwable t) {
			// jar is invalid; perhaps it's being rebuilt
			// don't touch file system
			parseWorked = false;
			t.printStackTrace();
			// force watcher to reschedule us
			lastModification = 0;
		    }
		    if (parseWorked) {
			((AbstractFolder )getRoot()).refreshFolder(null,null,true,true,new String[] {});      
			Iterator iter = entries.iterator();
			while (iter.hasNext()) {
			    String entry = (String)iter.next();
			    addFile (entry);
			}
		    }
		}
		finally {

		    // If the refreshTime is 0, then the refresh is disabled.
		    //
		    if (REFRESH_TIME  > 0 && watcherTask == null) 
			watcherTask = req.post (watcherTask (),REFRESH_TIME );
//		    System.out.println(new Date().toString() +
//			"done parsing jar file: " +
//			root.getPath());
		}
            }
            
            private String[] addChildren (AbstractFolder parent,String name ) {
                String[]        strChildren;
                String[]        tempChildren = parent.getChildrenArray();

                if (tempChildren == null) strChildren = new String[1];
                else strChildren = new String[tempChildren.length+1];
                for(int i = 0; i <strChildren.length-1;i++) {
                    strChildren[i] =tempChildren[i];
                }
                strChildren[strChildren.length-1] = name;
                return strChildren;
            }
            
            /** Adds a file together with all subdirectories to the hashtable
             * @param hash hashtable
             * @param name name of the file to add
             */
            private void addFile (String name) {
                String chld,parentName;
                int prev,enumIdx,enumCount;
                boolean isFold = false;
                
                name=name.trim();
                if (name.endsWith ("/")) {
                    name = name.substring (0, name.length () - 1);
                    isFold = true;
                }
                // work only with slashes
                name = name.replace ('\\', '/');
                if ( "".equals (name) ) return;
                prev = name.lastIndexOf ('/');
                if (JarFileSystem.super.findResource(name) != null) {
		    return;
		}
                
                parentName = prev < 0 ? "" : name.substring (0, prev);
                
                parentFo = (AbstractFolder)JarFileSystem.super.findResource(parentName);
                if ( parentFo != null ) { // It should almost go through this block
                    retRef = new EntryRef(isFold);
                    parentFo = addResource(parentFo,name.substring (prev + 1));
                    return;
                }
                parentFo = (AbstractFolder)JarFileSystem.super.findResource( (parentName="") );
                StringTokenizer en = new StringTokenizer(name,"/");
                
                enumIdx = 0;
                enumCount = en.countTokens();
                while (en.hasMoreElements ()) {
                    enumIdx++;
                    chld = (String)en.nextElement ();
                    
                    if (JarFileSystem.super.findResource("/".equals(parentName)?parentName+chld:parentName+"/"+chld) != null) {
                        parentName="/".equals(parentName)?parentName+chld:parentName+"/"+chld;
                        continue;
                    }
                    
                    parentFo = (AbstractFolder)JarFileSystem.super.findResource(parentName);
                    if (parentFo == null) {
			return;
		    }
                    
                    retRef = new EntryRef(enumIdx < enumCount?true:isFold);
                    parentFo = addResource(parentFo,chld);
                    
                    parentName = "/".equals(parentName)?parentName + chld:parentName + "/" + chld;
                }
            }
            
        /* Adds new FileObject under its parent
         * @param name Is name of FileObject, which should be added
         * @param parent Is parent of FileObject, which should be added
         * @return new added object or null if any problem
         */
            private final  AbstractFolder addResource(AbstractFolder parent,String name) {
                AbstractFolder retParent;
                name = name.trim();
                parent.refreshFolder(null,null,true,true,addChildren(parent, name));//internalChildren(parent, name,STATUS_ADD)
                
                retParent = (AbstractFolder)parent.getFileObject(name);
                if (retParent != null)
                    retParent.getFileObject("EMPTY");//NOI18N - there can be any other string
                                
                return retParent;
            }
            
        };
    }

    /* Anonymous Runnable class - responsible for checking whether JarFile was modified => standalone thread. 
     * If JarFile was modified, parsing is invoked.
     */    
    private Runnable watcherTask () {
        return new Runnable () {
            public    void run () {
                File rootJar = null;
                long curModif;
                
                rootJar = getJarFile();
                if (rootJar == null) return;
                
                curModif = rootJar.lastModified();
                if (curModif != lastModification) {
                    lastModification = curModif;
		    watcherTask = null;
                    parserTask = req.post (parserTask (),0);
                    return;
                }               
                watcherTask.schedule (REFRESH_TIME );
            }            
        };
    }
    
    /** Getter for root folder in the filesystem.
    *
    * @return root folder of whole filesystem
    */    
    public FileObject getRoot () {
        parserTask.waitFinished ();
        return super.getRoot ();  
    }
    
    
    /* 
    * Finds file when its resource name is given. This method is prevented to call, if
    * some jar file is parsed.
    * @param name resource name
    *
    * @return FileObject that represents file with given name or
    *   <CODE>null</CODE> if the file does not exist
    */
    public FileObject findResource (String name) {        
        parserTask.waitFinished ();
        return super.findResource(name);
    }
    /* Finds file when its name is provided. DESCRIPTION FROM: AbstractFileSystem.
    *
    * @param aPackage package name where each package is separated by a dot
    * @param name name of the file (without dots) or <CODE>null</CODE> if
    *    one want to obtain name of package and not file in it
    * @param ext extension of the file or <CODE>null</CODE> if one needs
    *    package and not file name
    *
    * @warning when one of name or ext is <CODE>null</CODE> then name and
    *    ext should be ignored and scan should look only for a package
    *
    * @return FileObject that represents file with given name or
    *   <CODE>null</CODE> if the file does not exist
    */
    public  FileObject find (String aPackage, String name, String ext) {
        parserTask.waitFinished ();
        return super.find(aPackage,name,ext);        
    }
    
    /** Getter for entry.
    */
    private final JarEntry getEntry (String file) {
        JarFile j = reOpenJarFile ();
        JarEntry je = j == null ? null : j.getJarEntry (file);
        if (je == null) {
            return new JarEntry (file);
        } else {
            return je;
        }
    }

    /** Creates Reference. In FileSystem, which subclasses AbstractFileSystem, you can overload method
     * createReference(FileObject fo) to achieve another type of Reference (weak, strong etc.)
     * @param fo is FileObject. It`s reference yourequire to get.
     * @return Reference to FileObject
     */
    protected  Reference createReference(FileObject fo) {
        EntryRef ref = retRef;       
        if ( retRef != null ) {
            retRef.setFileObject(fo);
            ref = retRef;
        } else {
            /*it should be unreachable code - now I can not decide what to do*/
            /*
            ErrorManager em = ExternalUtil.getErrorManager();
            RuntimeException ex = new NullPointerException ("JarFileSystem.createReference:  retRef == null");            
            em.annotate(ex,"Unexpected behaviour");            
            throw ex;
             */
            ref = new EntryRef(false);
            ref.setFileObject(fo);
        }
        return(ref);
    }
    
    /** Get the JAR manifest.
    * It will be lazily initialized.
    * @return parsed manifest file for this archive
    */
    public Manifest getManifest() {
        parserTask.waitFinished ();        
        if (manifest == null) {
            try {
                JarFile j = reOpenJarFile ();;
                manifest = j == null ? null : j.getManifest ();
            } catch (IOException ex) {
            }
            if (manifest == null) {
                manifest = new Manifest ();
            }
        }

        return manifest;
    }

    /**
    * Set name of the ZIP/JAR file.
    * @param aRoot path to new ZIP or JAR file
    * @throws IOException if the file is not valid
    */
    public  synchronized void setJarFile (File aRoot) throws IOException, PropertyVetoException {
        setRefreshTime(0);        
        parserTask.waitFinished ();
        
        
        if (aRoot == null) {
            closeCurrentRootAndNull();
            return;
        }
                
        if (!aRoot.exists ())
            FSException.io ("EXC_FileNotExists", aRoot.toString ()); // NOI18N
        if (!aRoot.canRead ())
            FSException.io ("EXC_CanntRead", aRoot.toString ()); // NOI18N
        if (!aRoot.isFile ())
            FSException.io ("EXC_NotValidFile", aRoot.toString ()); // NOI18N

        String s;
        try {
            s = aRoot.getCanonicalPath ();
        } catch (IOException e) {
            FSException.io ("EXC_NotValidFile", aRoot.toString ()); // NOI18N
            s = null;
        }

        // try to close previous jar file first
        closeCurrentRoot ();
        s = s.intern();   
        //    System.out.println("setJarFile, systemName:"+JAR_FS + s); // NOI18N

        try {
            jar = new JarFile (s);
        } catch (ZipException e) {
            closeCurrentRootAndNull();                        
            FSException.io ("EXC_NotValidJarFile"); // NOI18N
        }
        
         try {
            setSystemName (s);        
         } catch (PropertyVetoException vex) {
            jar = null;
            reOpenJarFile ();
            throw vex;
         }
         
        root = new File (s);

        firePropertyChange ("root", null, refreshRoot ()); // NOI18N
        parserTask = req.post (parserTask (),0);
    }

    /** Get the file path for the ZIP or JAR file.
    * @return the file path
    */
    public  File getJarFile () { // JST
        return root;
    }

    

    
    /*
    * Provides name of the system that can be presented to the user.
    * @return user presentable name of the file system
    */
    public String getDisplayName () {
        if (root == null) return getString ("JAR_NotValidJarFileSystem");
        return root.getName();
    }

    /** This file system is read-only.
    * @return <code>true</code>
    */
    public boolean isReadOnly () {
        return true;
    }

    /** Closes associated JAR file on cleanup, if possible. */ 
    public void removeNotify () {
	// [PENDING] this is not very polite! If you create a JAR, add
	// it, remove it, and readd it, then the root will have
	// disappeared and it will be useless!
	//
	// Also, closing the root does not clear file entries, yet
	// trying to access these throws exceptions.
	//
	// Also, when jar==null isValid() still returns true,
	// confusingly.
	//
	// -jglick
        closeCurrentRoot ();
    }

    /** initialization of jar variable, that is necessary after JarFileSystem was removed from Repository */ 
    public void addNotify () {
        super.addNotify ();
        reOpenJarFile (); 
    }
    
    /** may return null */
    
    private   JarFile reOpenJarFile () {        
        RequestProcessor.Task reOpenJarTask;
        if (jar == null && root != null) {            
            reOpenJarTask = req.post (new Runnable () {
                public  void run () {        
                            if (jar == null && root != null) {
                                try {
                                   jar = new JarFile (root);
                                } catch (Exception e) {}
                            }                                      
                }},0);
            reOpenJarTask.waitFinished ();                            
        }        
        return jar;
    }

    
    

    /** Prepare environment for external compilation or execution.
    * <P>
    * Adds name of the ZIP/JAR file, if it has been set, to the class path.
    */
    public void prepareEnvironment (Environment env) {
        if (root != null) {
            env.addClassPath (root.toString ());
        }
    }

    /** Initializes the root of FS.
    */
    private void readObject (ObjectInputStream ois)
    throws IOException, ClassNotFoundException {
        ois.defaultReadObject ();
        closeSync = new Object();
        try {
            parserTask = RequestProcessor.Task.EMPTY ;                        
            setJarFile (root);
        } catch (PropertyVetoException ex) {
            throw new IOException (ex.getMessage ());
        }
    }

    /** Performs a clean-up */
    private void closeCurrentRootAndNull () {
        closeCurrentRoot ();
        root = null;
    }

    /** Performs a clean-up */
    private void closeCurrentRoot () {
        RequestProcessor.Task closeRootTask;
        Runnable close = new Runnable () {
                    public  void run () {
                        if (jar != null) {
                            if (watcherTask != null) watcherTask.cancel ();
                            try {  
                                // Prevents to close JarFile if anybody uses InputStream (finalizer will close it anyway)
                                StreamPool pool = StreamPool.find (JarFileSystem.this);

                                if (pool != null && 
                                    (pool.isInputStreamOpen () || pool.isOutputStreamOpen ())
                                ) {
                                    IOException e = new IOException (); 
                                    pool.annotate (e);
                                    ExternalUtil.exception (e);
                                } else {
                                    jar.close();
                                }

                            } catch (Exception exc) {
                                // ignore exception during closing, just log it
                                ExternalUtil.exception (exc);
                            } finally {
                                jar = null;
                            }
                        }                                        
                    }
                };
        
        synchronized (closeSync) {
            if (jar != null) {
                closeRootTask = req.post ( close,0);
                closeRootTask.waitFinished ();
            }
        }        
    }

    //
    // List
    //

    /**
     *
     * Scans children for given name
     */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    String[] children (String name) {
        FileObject  fo2name;
        String[] emptyChildren = {};

        Object obj = thrStatus.get();
        if ( obj == null) {
            parserTask.waitFinished ();
            if (( fo2name = super.findResource(name) ) == null) return emptyChildren;;                
            return ((AbstractFolder)fo2name).getChildrenArray();
        }
                
        return emptyChildren;
    }

    //
    // Change
    //

    /**
    * Creates new folder named name.
    * @param name name of folder
    * @throws IOException if operation fails
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void createFolder (String name) throws java.io.IOException {
        throw new IOException ();
    }

    /**
    * Create new data file.
    *
    * @param name name of the file
    *
    * @return the new data file object
    * @exception IOException if the file cannot be created (e.g. already exists)
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void createData (String name) throws IOException {
        throw new IOException ();
    }

    /**
    * Renames a file.
    *
    * @param oldName old name of the file
    * @param newName new name of the file
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void rename(String oldName, String newName) throws IOException {
        throw new IOException ();
    }

    /**
    * Delete the file. 
    *
    * @param name name of file
    * @exception IOException if the file could not be deleted
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void delete (String name) throws IOException {
        throw new IOException ();
    }

    //
    // Info
    //

    /**
    * Get last modification time.
    * @param name the file to test
    * @return the date
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    java.util.Date lastModified(String name) {
        /** JarEntry.getTime returns wrong value: already reported in bugtraq 4319781
         *  Fixed in jdk1.4  */
        return new java.util.Date (getEntry (name).getTime ());
    }

    /**
    * Test if the file is folder or contains data.
    * @param name name of the file
    * @return true if the file is folder, false otherwise
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    boolean folder (String name) {
        if ("".equals (name)) return true; // NOI18N
        EntryRef eref = (EntryRef)findReference(name);
        if (eref == null) return false;
        return eref.folder();
    }

    /**
    * Test whether this file can be written to or not.
    * @param name the file to test
    * @return <CODE>true</CODE> if file is read-only
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    boolean readOnly (String name) {
        return true;
    }

    /** Get the MIME type of the file. If filesystem has no special support 
    * for MIME types then can simply return null. FileSystem can register 
    * MIME types for a well-known extensions: FileUtil.setMIMEType(String ext, String mimeType)
    * or together with filesystem supply some resolvers subclassed from MIMEResolver.
    *
    * @param name the file to test
    * @return the MIME type textual representation (e.g. <code>"text/plain"</code>)
    * or null if no special support for recognizing MIME is implemented.
     */
    
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    String mimeType (String name) {
        return null;
    }

    /**
    * Get the size of the file.
    *
    * @param name the file to test
    * @return the size of the file in bytes or zero if the file does not contain data (does not
    *  exist or is a folder).
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    long size (String name) {
        long retVal = getEntry (name).getSize ();
        return  (retVal == -1) ? 0 : retVal;
    }

/**
    * Get input stream.
    *
    * @param name the file to test
    * @return an input stream to read the contents of this file
    * @exception FileNotFoundException if the file does not exists or is invalid
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
  InputStream inputStream (String name) throws java.io.FileNotFoundException {
        InputStream is = null;
        AbstractFolder fo = null; 
        try {
            JarFile j = reOpenJarFile ();            
            if (j != null) {
                synchronized (closeSync) { 
                    if (j != null) {
                        JarEntry je = j.getJarEntry (name);
                        if (je != null) {
                            is = j.getInputStream (je);
                            if (is != null) {
                                EntryRef eref = (EntryRef)findReference(name);        
                                if (eref != null) fo = (AbstractFolder) eref.get ();                                                                
                                if (fo != null) is = StreamPool.createInputStream(fo, is);
                            }
                        }
                    }
                }
            }
        } catch (java.io.FileNotFoundException e) {
            throw e;
        } catch (IOException e) {
            throw new java.io.FileNotFoundException (e.getMessage ());
        } 

        if (is == null) {
          throw new java.io.FileNotFoundException (name);
        }

        return is;
    }
            
    
    /**
    * Get output stream.
    *
    * @param name the file to test
    * @return output stream to overwrite the contents of this file
    * @exception IOException if an error occures (the file is invalid, etc.)
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    OutputStream outputStream (String name) throws java.io.IOException {
        throw new IOException ();
    }

    /**
    * Does nothing.
    *
    * @param name name of the file
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void lock (String name) throws IOException {
        FSException.io ("EXC_CannotLock", name, getDisplayName (), name); // NOI18N
    }

    /**
    * Unlock the file. Does nothing.
    *
    * @param name name of the file
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void unlock (String name) {
    }

    /**
    * Does nothing.
    *
    * @param name the file to mark
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void markUnimportant (String name) {
    }

    /**
    * Get the file attribute with the specified name.
    * @param name the file
    * @param attrName name of the attribute
    * @return appropriate (serializable) value or <CODE>null</CODE> if the attribute is unset (or could not be properly restored for some reason)
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    Object readAttribute(String name, String attrName) {
        Attributes attr = getManifest ().getAttributes (name);
        return attr == null ? null : attr.getValue (attrName);
    }

    /**
    * Set the file attribute with the specified name.
    * @param name the file
    * @param attrName name of the attribute
    * @param value new value or <code>null</code> to clear the attribute. Must be serializable, although particular file systems may or may not use serialization to store attribute values.
    * @exception IOException if the attribute cannot be set. If serialization is used to store it, this may in fact be a subclass such as {@link NotSerializableException}.
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void writeAttribute(String name, String attrName, Object value) throws IOException {
        throw new IOException ();
    }

    /**
    * Get all file attribute names for the file.
    * @param name the file
    * @return enumeration of keys (as strings)
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    Enumeration attributes(String name) {
        Attributes attr = getManifest ().getAttributes (name);
        
        if (attr != null) {
            return  new AlterEnumeration(Collections.enumeration (attr.keySet ())) {
                public Object alter(Object obj) {
                    return obj.toString();
                }
            };
            
        } else {
            return EmptyEnumeration.EMPTY;
        }
    }

    /**
    * Called when a file is renamed, to appropriatelly update its attributes.
    * <p>
    * @param oldName old name of the file
    * @param newName new name of the file
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void renameAttributes (String oldName, String newName) {
    }

    /**
    * Called when a file is deleted to also delete its attributes.
    *
    * @param name name of the file
    */
    /*nbif compat
    public
    nbelse*/
    protected
    /*nbend*/
    void deleteAttributes (String name) {
    }

    /** Close the jar file when we go away...*/
    protected void finalize () throws Throwable {
        super.finalize();
        closeCurrentRoot ();
    }

    /** Implementation of all interfaces List, Change, Info and Attr
    * that delegates to JarFileSystem
    */
    public static class Impl extends Object
                implements AbstractFileSystem.List, AbstractFileSystem.Info,
        AbstractFileSystem.Change, AbstractFileSystem.Attr {
        /** generated Serialized Version UID */
        static final long serialVersionUID = -67233308132567232L;

        /** the pointer to file system */
        private JarFileSystem fs;

        /** Constructor.
        * @param fs the file system to delegate to
        */
        public Impl (JarFileSystem fs) {
            this.fs = fs;
        }

        /**
        *
        * Scans children for given name
        */
        public String[] children (String name) {
            return fs.children (name);
        }

        //
        // Change
        //

        /**
        * Creates new folder named name.
        * @param name name of folder
        * @throws IOException if operation fails
        */
        public void createFolder (String name) throws java.io.IOException {
            fs.createFolder (name);
        }


        /**
        * Create new data file.
        *
        * @param name name of the file
        *
        * @return the new data file object
        * @exception IOException if the file cannot be created (e.g. already exists)
        */
        public void createData (String name) throws IOException {
            fs.createData (name);
        }

        /**
        * Renames a file.
        *
        * @param oldName old name of the file
        * @param newName new name of the file
        */
        public void rename(String oldName, String newName) throws IOException {
            fs.rename (oldName, newName);
        }

        /**
        * Delete the file. 
        *
        * @param name name of file
        * @exception IOException if the file could not be deleted
        */
        public void delete (String name) throws IOException {
            fs.delete (name);
        }

        //
        // Info
        //

        /**
        *
        * Get last modification time.
        * @param name the file to test
        * @return the date
        */
        public java.util.Date lastModified(String name) {
            return fs.lastModified (name);
        }

        /**
        * Test if the file is folder or contains data.
        * @param name name of the file
        * @return true if the file is folder, false otherwise
        */
        public boolean folder (String name) {
            return fs.folder (name);
        }

        /**
        * Test whether this file can be written to or not.
        * @param name the file to test
        * @return <CODE>true</CODE> if file is read-only
        */
        public boolean readOnly (String name) {
            return fs.readOnly (name);
        }

        /**
        * Get the MIME type of the file.
        * Uses {@link FileUtil#getMIMEType}.
        *
        * @param name the file to test
        * @return the MIME type textual representation, e.g. <code>"text/plain"</code>
        */
        public String mimeType (String name) {
            return fs.mimeType (name);
        }

        /**
        * Get the size of the file.
        *
        * @param name the file to test
        * @return the size of the file in bytes or zero if the file does not contain data (does not
        *  exist or is a folder).
        */
        public long size (String name) {
            return fs.size (name);
        }

        /**
        * Get input stream.
        *
        * @param name the file to test
        * @return an input stream to read the contents of this file
        * @exception FileNotFoundException if the file does not exists or is invalid
        */
        public InputStream inputStream (String name) throws java.io.FileNotFoundException {
            return fs.inputStream (name);
        }

        /**
        * Get output stream.
        *
        * @param name the file to test
        * @return output stream to overwrite the contents of this file
        * @exception IOException if an error occures (the file is invalid, etc.)
        */
        public OutputStream outputStream (String name) throws java.io.IOException {
            return fs.outputStream (name);
        }

        /**
        * Does nothing to lock the file.
        *
        * @param name name of the file
        */
        public void lock (String name) throws IOException {
            fs.lock (name);
        }

        /**
        * Does nothing to unlock the file.
        *
        * @param name name of the file
        */
        public void unlock (String name) {
            fs.unlock (name);
        }

        /**
        * Does nothing to mark the file as unimportant.
        *
        * @param name the file to mark
        */
        public void markUnimportant (String name) {
            fs.markUnimportant (name);
        }

        /**
        * Get the file attribute with the specified name.
        * @param name the file
        * @param attrName name of the attribute
        * @return appropriate (serializable) value or <CODE>null</CODE> if the attribute is unset (or could not be properly restored for some reason)
        */
        public Object readAttribute(String name, String attrName) {
            return fs.readAttribute (name, attrName);
        }

        /**
        * Set the file attribute with the specified name.
        * @param name the file
        * @param attrName name of the attribute
        * @param value new value or <code>null</code> to clear the attribute. Must be serializable, although particular file systems may or may not use serialization to store attribute values.
        * @exception IOException if the attribute cannot be set. If serialization is used to store it, this may in fact be a subclass such as {@link NotSerializableException}.
        */
        public void writeAttribute(String name, String attrName, Object value) throws IOException {
            fs.writeAttribute (name, attrName, value);
        }

        /**
        * Get all file attribute names for the file.
        * @param name the file
        * @return enumeration of keys (as strings)
        */
        public Enumeration attributes(String name) {
            return fs.attributes (name);
        }

        /**
        * Called when a file is renamed, to appropriatelly update its attributes.
        * <p>
        * @param oldName old name of the file
        * @param newName new name of the file
        */
        public void renameAttributes (String oldName, String newName) {
            fs.renameAttributes (oldName, newName);
        }

        /**
        * Called when a file is deleted to also delete its attributes.
        *
        * @param name name of the file
        */
        public void deleteAttributes (String name) {
            fs.deleteAttributes (name);
        }
    }

    /** Strong Reference to FileObject, which holds additional info about if FileObject is folder or file   
    */ 

    private static class EntryRef extends WeakReference {
        private FileObject fo;
        private boolean     folder;
        
        public EntryRef (boolean folder) {
            super (null);
            this.folder = folder;
        }
        public boolean folder() { return folder; }        
        public void setFileObject(FileObject fo) {  this.fo = fo; }        
        public Object get () { return fo; }
    }

    
    
}
