/*
 * MIMESupport.java
 *
 * Created on January 25, 2001, 2:33 PM
 */

package org.openide.filesystems;
import java.util.*;
import java.io.*;
import java.lang.ref.*;

import org.openide.util.Lookup;

/**
 * This class is intended to enhance MIME resolving. This class offers
 * only one method: findMIMEType(FileObject fo). If this method is called, then 
 * registered subclasses of MIMEResolver are asked one by one to resolve MIME type of this FileObject.
 * Resolving is finished right after first resolver is able to resolve this FileObject or if all registered
 * resolvers returns null (not recognized).
 *
 *Resolvers are registered if they have their record in IDE_HOME\system\Services
 * in form *.instance e.g.: org-some-package-JavaResolver.instance
 *
 * @author  rmatous
 */

final class MIMESupport extends Object {
    static private Map cache;
    static private LinkedList recentlist;
    static private int cachesize;
    
    static private MIMESupport defaultInst;
    private static final int MAX_CACHE_SIZE = 5;
    
    
    /** Asks all registered subclasses of MIMEResolver to resolve FileObject passed as parameter.
     * @param fo is FileObject, whose MIME should be resolved
     * @return  MIME type or null if not resolved*/
    static String findMIMEType(FileObject fo) {
        if (fo.isFolder ()) return null;
        if (!CachedFileObject.isAnyResolver ()) return null;
        
        CachedFileObject cfo =  getDefault().getCachedFileObject (fo);                
        return cfo.getMIMEType ();
    }
        
    
    /** Creates new MIMESupport if not exist */
    private static MIMESupport getDefault() {
        if (defaultInst == null) defaultInst = new MIMESupport ();
        return defaultInst;
    }
    
    private MIMESupport() {
        cachesize = MAX_CACHE_SIZE;
        cache = new HashMap (cachesize);
        recentlist = new LinkedList ();
    }
    
    private  CachedFileObject getCachedFileObject (FileObject  fo) {
        CachedFileObject retVal = null;
        synchronized (cache) {
            retVal = (CachedFileObject)cache.get(fo);
            if (retVal == null) {
                if (recentlist.size() > cachesize) {
                    cache.remove(recentlist.getLast());
                    recentlist.removeLast();
                }
                cache.put(fo,retVal = new CachedFileObject (fo));
                recentlist.addFirst(fo);
            } else {
                boolean isModified;
                isModified = (fo.lastModified().getTime() != retVal.lastModified ().getTime());
                if (isModified) {
                    cache.put(fo,retVal = new CachedFileObject (fo));
                }
                recentlist.remove(fo);
                recentlist.addFirst(fo);
            }
        }
        return  retVal;
    }
    
    
    //    private void removeFromCache (FileObject fo)
    
    
    private static class CachedFileObject extends FileObject implements FileChangeListener{
        /*Variables  for caching*/
        Reference   refForStream;
        String      mimeType;
        java.util.Date lastModified;
        static Lookup.Result result;
        
        static {
            result = Lookup.getDefault().lookup(new Lookup.Template (MIMEResolver.class));
        }        
        
        /*All calls delegated to this object.
         Except few methods, that returns cached values*/
        FileObject    fileObj;
        
        CachedFileObject (FileObject fo) {
            fileObj = fo;
            lastModified = fileObj.lastModified();
            fileObj.addFileChangeListener (this);
        }
        
        public static boolean isAnyResolver () {
            return (!result.allInstances ().isEmpty ());
        }
            
        public void freeCaches () {
            refForStream = null;
            mimeType = null;
            lastModified = null;
        }
        
        public String getMIMEType() {
            String retVal;
            if (mimeType == null) mimeType = resolveMIME ();
            return mimeType;
        }
        
        private String resolveMIME () {
            String retVal = null;
            
            if (result.allInstances ().isEmpty ()) return null;
             /* because InputStream is hold as WeakReference,
              I require here the reference was not freed in this method*/
            InputStream fixIt;
            try {
                fixIt = getInputStream ();
            } catch (FileNotFoundException fex) {
                return null;
            }
            if (fixIt == null) return null;
            
            Iterator instances = result.allInstances ().iterator();
            while (instances.hasNext()) {
                MIMEResolver resolv = (MIMEResolver) instances.next();
                retVal = resolv.findMIMEType(this);
                if (retVal != null) return retVal;
            }
            return FileUtil.getMIMEType (this.getExt ());
        }
        
        public java.util.Date lastModified() {
            if (lastModified != null) return lastModified;
            return lastModified = fileObj.lastModified();
        }
        
        public InputStream getInputStream() throws java.io.FileNotFoundException {
            CachedInputStream cachedIStream = null;
            if (refForStream == null || (cachedIStream = (CachedInputStream ) refForStream.get()) == null) {
                InputStream is = fileObj.getInputStream();
                if (!(is instanceof BufferedInputStream)) is = new BufferedInputStream (is);
                refForStream = new WeakReference (cachedIStream = new CachedInputStream(is));
            }
            cachedIStream.cacheToStart();
            return cachedIStream;
        }
        
        public void fileChanged(FileEvent fe) {
            freeCaches ();
        }
        
        public void fileDeleted(FileEvent fe) {
            freeCaches ();
            //removeFromCache (fe.getFile ());
        }
        
        public void fileRenamed(FileRenameEvent fe) {
            freeCaches ();
        }
        
        
        /*All other methods only delegate to fileObj*/
        
        public FileObject getParent() {
            return fileObj.getParent ();
        }
        
        public String getPackageNameExt(char separatorChar,char extSepChar) {
            return fileObj.getPackageNameExt(separatorChar, extSepChar);
        }
        
        public FileObject copy(FileObject target,String name,String ext) throws IOException {
            return fileObj.copy(target, name, ext);
        }
        
        protected void fireFileDeletedEvent(Enumeration en,FileEvent fe) {
            fileObj.fireFileDeletedEvent(en, fe);
        }
        
        protected void fireFileFolderCreatedEvent(Enumeration en,FileEvent fe) {
            fileObj.fireFileFolderCreatedEvent(en, fe);
        }
        
        public void setImportant(boolean b) {
            fileObj.setImportant(b);
        }
        
        public boolean isData() {
            return fileObj.isData();
        }
        
        public Object getAttribute(String attrName) {
            return fileObj.getAttribute(attrName);
        }
        
        public Enumeration getFolders(boolean rec) {
            return fileObj.getFolders(rec);
        }
        
        
        public void delete(FileLock lock) throws IOException {
            fileObj.delete (lock);
        }
        
        public boolean isRoot() {
            return fileObj.isRoot();
        }
        
        public Enumeration getData(boolean rec) {
            return fileObj.getData(rec);
        }
        
        public FileObject[] getChildren() {
            return fileObj.getChildren();
        }
        
        public String getNameExt() {
            return fileObj.getNameExt();
        }
        
        public boolean isValid() {
            return fileObj.isValid();
        }
        
        public boolean isReadOnly() {
            return fileObj.isReadOnly();
        }
        
        public String getExt() {
            return fileObj.getExt();
        }
        
        public String getName() {
            return fileObj.getName();
        }
        
        public void removeFileChangeListener(FileChangeListener fcl) {
            fileObj.removeFileChangeListener(fcl);
        }
        
        protected void fireFileRenamedEvent(Enumeration en,FileRenameEvent fe) {
            fileObj.fireFileRenamedEvent(en, fe);
        }
        
        public void refresh(boolean expected) {
            fileObj.refresh(expected);
        }
        
        protected void fireFileAttributeChangedEvent(Enumeration en,FileAttributeEvent fe) {
            fileObj.fireFileAttributeChangedEvent(en, fe);
        }
        
        public long getSize() {
            return fileObj.getSize() ;
        }
        
        public Enumeration getAttributes() {
            return fileObj.getAttributes() ;
        }
        
        public void rename(FileLock lock,String name,String ext) throws IOException {
            fileObj.rename (lock,name,ext);
        }
        
        protected void fireFileChangedEvent(Enumeration en,FileEvent fe) {
            fileObj.fireFileChangedEvent(en, fe);
        }
        
        public FileObject getFileObject(String name,String ext) {
            return fileObj.getFileObject(name,ext);
        }
        
        public void refresh() {
            fileObj.refresh();
        }
        
        public FileObject createData(String name,String ext) throws IOException {
            return fileObj.createData(name,ext);
        }
        
        public void addFileChangeListener(FileChangeListener fcl) {
            fileObj.addFileChangeListener(fcl);
        }
        
        protected void fireFileDataCreatedEvent(Enumeration en,FileEvent fe) {
            fileObj.fireFileDataCreatedEvent(en, fe);
        }
        
        public boolean isFolder() {
            return fileObj.isFolder();
        }
        
        public FileObject createFolder(String name) throws IOException {
            return fileObj.createFolder(name);
        }
        
        public Enumeration getChildren(boolean rec) {
            return fileObj.getChildren(rec);
        }
        
        public void setAttribute(String attrName,Object value) throws IOException {
            fileObj.setAttribute(attrName,value);
        }
        
        public String getPackageName(char separatorChar) {
            return fileObj.getPackageName(separatorChar);
        }
        
        public FileSystem getFileSystem() throws FileStateInvalidException {
            return fileObj.getFileSystem();
        }
        
        public OutputStream getOutputStream(FileLock lock) throws java.io.IOException {
            return fileObj.getOutputStream(lock);
        }
        
        public boolean existsExt(String ext) {
            return fileObj.existsExt(ext);
        }
        
        public FileObject move(FileLock lock,FileObject target,String name,String ext) throws IOException {
            return fileObj.move(lock, target, name, ext);
        }
        
        public FileLock lock() throws IOException {
            return fileObj.lock();
        }
        
        public void fileFolderCreated(FileEvent fe) {
        }
        
        public void fileDataCreated(FileEvent fe) {
        }
        
        public void fileAttributeChanged(FileAttributeEvent fe) {
        }
    }
    
    
    private static class CachedInputStream extends java.io.InputStream {
        java.io.InputStream inputStream;
        InnerCache   cache;
        int pos =  0;
        int size;
        
        public CachedInputStream (java.io.InputStream is) {
            inputStream = is;
            try {
                cache = new InnerCache (size = is.available());
            } catch (Exception e) {} // domyslet
        }
        
        
        public int available() throws IOException {
            return size - pos;
        }
        
        
        public long skip(long n) throws IOException { //implement better
            long loop = n / Integer.MAX_VALUE;
            int mod = (int)n % Integer.MAX_VALUE;
            
            for (long i = 0; i < loop; i++) {
                read (new byte[Integer.MAX_VALUE]);
            }
            if (mod > 0) read (new byte[mod]);
            return n;
        }
        
        /*prevents stream to be closed, because this stream is
         expected to be used many types. Next calls get data from cache*/
        public void close() throws IOException {
        }
        
        protected void finalize () {
            try {
                inputStream.close();
            } catch (IOException ioe) {
                // I don`t intend any special handling here
            }
        }
        
        public void cacheToStart () {
            pos = 0;
            cache.cacheToStart ();
        }
        
        public int read(byte[] b) throws IOException {
            return read (b,0,b.length);
        }
        
        
        
        public int read() throws IOException {
            int retVal;
            if ((retVal =  cache.intRead()) != -1) return retVal;
            
            if ((retVal = inputStream.read ()) != -1)  cache.intWrite(retVal);
            return retVal;
        }
        
        public int read(byte[] b,int off,int len) throws IOException {
            int retVal;
            
            retVal = cache.read(b,off,len);
            if (retVal < len) {
                int incr = inputStream.read(b,off + retVal,len - retVal);
                cache.write(b,off + retVal,len - retVal);
                retVal += incr;
            }
            return retVal;
        }
        
        public boolean markSupported() {
            return false;
        }
        
        public void mark(int readlimit) {
            //throw new IOException ("mark: not supported");
        }
        
        public void reset() throws IOException {
            //throw new IOException ("reset: not supported");
        }
        
        
        private static class InnerCache {
            byte[] cache;
            int wpos = 0;//cache.length;
            int rpos = 0;//cache.length;
            
            InnerCache (int size) throws IOException {
                cache = new byte[size];
            }
            
            void cacheToStart () {
                rpos = 0;
            }
            
            public void write(byte b)  {
                if (wpos < cache.length) cache[wpos++] = b;
            }
            void write (byte[] buffer, int offset, int len)  { // new added method
                int retVal = Math.min((cache.length - wpos), Math.min(buffer.length-offset, len));
                if (retVal > 0) System.arraycopy(buffer,offset,cache,wpos,retVal);//(cache[rpos++])
                wpos += retVal;
                //return retVal;
            }
            
            public int intRead() {
                byte retVal;
                if (rpos >= wpos)  return -1;
                retVal = cache[rpos++];
                return (retVal < 0)? (retVal+256):retVal;
            }
            
            public void intWrite(int i)  {
                if (wpos < cache.length)
                    cache[wpos++] = (byte)((i > 127)?i-256:i);
            }
            
            public int read (byte[] buffer, int offset ,int len) {
                int retVal = Math.min((wpos - rpos), Math.min(buffer.length-offset, len));;
                if (retVal > 0) System.arraycopy(cache,0,buffer,offset,retVal);//(cache[rpos++])
                rpos += retVal;
                return retVal;
            }
        }
    }
    
}
