/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.filesystems;

import java.beans.*;
import java.io.*;
import java.util.*;

import org.openide.util.NbBundle;
import org.openide.util.io.NbMarshalledObject;

/** This singleton object contains all {@link FileSystem}s in the IDE.
* It corresponds to the <b>Filesystems</b> tab in the Explorer, or more precisely
* to <b>Filesystems&nbsp;Settings</b> in Project Settings.
* <P>
* At any given time, no two file systems in the pool may share the same {@link FileSystem#getSystemName name}
* (unless all but one are {@link FileSystem#isValid invalid}).
*
* <p>Use {@link org.openide.TopManager#getRepository} to retrieve the default instance.
*
* <p>Note that you may construct additional instances of the Repository if you
* wish. The serialization replacer of this class specifically deals with the
* default instance as stored by the top manager; however you may safely call
* the {@link #readExternal} and {@link #writeExternal} methods directly to implement
* persistence of a non-default instance.
*
* @author Jaroslav Tulach, Petr Hamernik
*/
public final class Repository extends Object implements java.io.Serializable 
/*nbif compat
, org.openide.nodes.Node.Cookie
/*nbend*/
{
    /** list of file systems (FileSystem) */
    private ArrayList fileSystems;
    private transient ArrayList fileSystemsClone;

    /** the system file system */
    private FileSystem system;

    /** hashtable that maps system names to FileSystems */
    private Hashtable names;

    // [PENDING] access to this hashtable is apparently not propertly synched
    // should use e.g. Collections.synchronizedSet, or just synch methods using it
    /** hashtable for listeners on changes in the file system.
    * Its elements are of type (RepositoryListener, RepositoryListener)
    */
    private Hashtable listeners = new Hashtable ();

    /** vetoable listener on systemName property of file system */
    private VetoableChangeListener vetoListener = new VetoableChangeListener () {
                /** @param ev event with changes */
                public void vetoableChange (PropertyChangeEvent ev)
                throws PropertyVetoException {
                    if (ev.getPropertyName ().equals ("systemName")) {
                        final String ov = (String)ev.getOldValue ();
                        final String nv = (String)ev.getNewValue ();
                        if (names.get (nv) != null) {
                            throw new PropertyVetoException ("system name already exists", ev) { // NOI18N
                                public String getLocalizedMessage () {
                                    return NbBundle.getMessage (Repository.class, "EXC_duplicate_system_name", ov, nv);
                                }
                            };
                        }
                    }
                }
            };

    /** property listener on systemName property of file system */
    private PropertyChangeListener propListener = new PropertyChangeListener () {
                /** @param ev event with changes */
                public void propertyChange (PropertyChangeEvent ev) {
                    if (ev.getPropertyName ().equals ("systemName")) {
                        // assign the property to new name
                        String ov = (String)ev.getOldValue ();
                        String nv = (String)ev.getNewValue ();
                        FileSystem fs = (FileSystem)ev.getSource ();
                        if (fs.isValid ()) {
                            // when a file system is valid then it is attached to a name
                            names.remove (ov);
                        }
                        // register name of the file system
                        names.put (nv, fs);
                        // the file system becomes valid
                        fs.setValid (true);
                    }
                }
            };

    static final long serialVersionUID =-6344768369160069704L;
    /** Creates new instance of file system pool and
    * registers it as the default one. Also registers the default file system.
    *
    * @param def the default filesystem
    */
    public Repository (FileSystem def) {
        this.system = def;
        init ();
    }
    
    /** Access method to get default instance of repository in the system.
     * The instance is either taken as a result of 
     * <CODE>org.openide.util.Lookup.getDefault ().lookup (Repository.class)</CODE>
     * or (if the lookup query returns null) a default instance is created.
     *
     * @return default repository for the system
     */
    public static Repository getDefault () {
        return ExternalUtil.getRepository ();
    }

    /** Initialazes the pool.
    * @exception SecurityException if CoronaStreamHandlerFactory is already registered.
    */
    private void init () {
        // empties the pool
        fileSystems = new ArrayList ();
        names = new Hashtable ();
        addFileSystem (system);
    }

    /** Gets the default filesystem of the IDE.
    * @return the default filesystem
    */
    public FileSystem getDefaultFileSystem () {
        return system;
    }

    /** Adds new file system to the pool.
    * <em>Note</em> that a file system cannot be assigned to more than one file
    *   system pool at one time (though currently there is only one pool anyway).
    * @param fs file system to add
    */
    public final void addFileSystem (FileSystem fs) {
        // if the file system is not assigned yet
        boolean fireIt = false;
        synchronized (this) {
            if (!fs.assigned && !fileSystems.contains(fs)) {
                // new file system
                fileSystems.add(fs);
                fileSystemsClone = (ArrayList)fileSystems.clone();
                                  
                String systemName = fs.getSystemName ();

                boolean isReg = names.get (systemName) == null;
                if (isReg && !systemName.equals ("")) { // NOI18N
                    // file system with the same name is not there => then it is valid
                    names.put (systemName, fs);
                    fs.setValid (true);
                } else {
                    // there is another file system with the same name => it is invalid
                    fs.setValid (false);
                }
                // mark the file system as being assigned
                fs.assigned = true;
                // mark as a listener on changes in the file system
                fs.addPropertyChangeListener (propListener);
                fs.addVetoableChangeListener (vetoListener);

                // notify filesystem itself that it has been added
                fs.addNotify(); 

                // fire info about new file system
                fireIt = true;                     
            }
        }
        // postponed firing after synchronized  block to prevent deadlock
        if (fireIt) 
            fireFileSystem (fs, true); 
        
    }

    /** Removes a file system from the pool.
    * @param fs file system to remove
    */
    public final void removeFileSystem (FileSystem fs) {
        boolean fireIt = false;
        synchronized (this) {
            if (fs.isDefault()) return;
            if (fireIt = fileSystems.remove(fs)) {
                fileSystemsClone = (ArrayList)fileSystems.clone();
                                  
                // the file system realy was here
                if (fs.isValid ()) {
                    // if the file system is valid then is in names hashtable
                    names.remove (fs.getSystemName ());
                }
                // in all cases remove it from listeners
                fs.removePropertyChangeListener (propListener);
                fs.removeVetoableChangeListener (vetoListener);

                // notify filesystem itself that it has been removed
                fs.removeNotify();                 
            }
            // unassign the file system
            fs.assigned = false;
        }
        // postponed firing after synchronized  block to prevent deadlock
        if (fireIt) 
            fireFileSystem (fs, false);
    }

    /** Reorders {@link FileSystem}s by given permutation.
     * For example, if there are three file systems, <code>new int[] {2, 0, 1}</code> cycles the file systems forwards.
    * @param perm an array of integers
    * @throws IllegalArgumentException if the array is not a permutation, or is not the same length as the current number of file systems in the pool
    */
    public final void reorder(int[] perm) {
        synchronized (this) {
            if (perm == null || perm.length != fileSystems.size() || !isPermutation(perm))
                throw new IllegalArgumentException();

            ArrayList newList = new ArrayList (fileSystems.size ());
            int len = perm.length;
            for (int i = 0; i < len; i++) {
                newList.add (fileSystems.get (perm[i]));
            }
            fileSystems = newList;
            fileSystemsClone = (ArrayList)fileSystems.clone();                        
        }
        // postponed firing after synchronized  block to prevent deadlock
        fireFileSystemReordered(perm);
    }


    /** @return true if the parameter describes a permutation */
    private static boolean isPermutation(int[] perm) {
        final int len = perm.length;
        boolean[] bool = new boolean[len];
        try {
            for (int i = 0; i < len; i++) {
                if (bool[perm[i]]) return false;
                else bool[perm[i]] = true;
            }
            return true;
        } catch (IndexOutOfBoundsException e) {
            return false;
        }
    }

    /** Returns enumeration of all file systems.
    * @return enumeration of type {@link FileSystem}
    */
    public final Enumeration getFileSystems () {
         ArrayList tempFileSystems = fileSystemsClone;
         return java.util.Collections.enumeration (tempFileSystems);
    }

    /** Returns enumeration of all file systems.
    * @return enumeration of type {@link FileSystem}
    */
    public final Enumeration fileSystems () {
        return getFileSystems ();
    }


    /** Returns a sorted array of file systems. */
    public final FileSystem[] toArray() {
        ArrayList tempFileSystems = fileSystemsClone;

        FileSystem[] fss = new FileSystem[tempFileSystems.size()];
        tempFileSystems.toArray(fss);
        return fss;
    }
  
    /** Finds file system when only its system name is known.
    * @param systemName {@link FileSystem#getSystemName name} of the file system
    * @return the file system or <CODE>null</CODE> if there is no such
    *   file system
    */
    public final FileSystem findFileSystem (String systemName) {
        FileSystem fs = (FileSystem)names.get (systemName);
        return fs;
    }

    /** Saves pool to stream by saving all file systems.
    * The default (system) file system, or any persistent file systems, are skipped.
    *
    * @param oos object output stream
    * @exception IOException if an error occures
    */
    public synchronized void writeExternal (ObjectOutput oos) throws IOException {
        Iterator iter = fileSystems.iterator();
        while (iter.hasNext()) {
            FileSystem fs = (FileSystem)iter.next();

            if (!fs.isDefault () && !fs.isPersistent ()) {
                oos.writeObject (new NbMarshalledObject (fs));
            }
        }
        oos.writeObject (null);
    }

    /** Reads object from stream.
    * Reads all filesystems. Persistent and system file systems are untouched; all others are removed and possibly reread.
    * @param oos object output stream
    * @exception IOException if an error occures
    * @exception ClassNotFoundException if read class is not found
    */
    public synchronized void readExternal (ObjectInput ois)
    throws IOException, ClassNotFoundException {
        ArrayList temp = new ArrayList(10);

        for (;;) {
            Object obj = ois.readObject ();
            if (obj == null) {
                // all system has been read in
                break;
            }
            FileSystem fs;
            if (obj instanceof FileSystem) {
                fs = (FileSystem)obj;
            } else {
                try {
                    NbMarshalledObject mar = (NbMarshalledObject)obj;
                    fs = (FileSystem)mar.get ();
                } catch (IOException ex) {
                    ExternalUtil.exception (ex);
                    fs = null;
                } catch (ClassNotFoundException ex) {
                    ExternalUtil.exception (ex);
                    fs = null;
                }
            }
            if (fs != null) {
                // add the new file system
                temp.add(fs);
            }
        }

        Enumeration ee = getFileSystems();
        FileSystem fs;
        while (ee.hasMoreElements()) {
            fs = (FileSystem) ee.nextElement();
            if (!fs.isPersistent ()) {
                removeFileSystem (fs);
            }
        }
        // in init assigned is checked and we force 'system' to be added again
        system.assigned = false;
        init ();

        // all is successfuly read
        for (Iterator iter = temp.iterator(); iter.hasNext();)
            addFileSystem ((FileSystem) iter.next());
    }



    /** Finds file when its name is provided. It scans in the list of
    * file systems and asks them for the specified file by a call to
    * {@link FileSystem#find find}. The first object that is found is returned or <CODE>null</CODE>
    * if none of the file systems contain such a file.
    *
    * @param aPackage package name where each package is separated by a dot
    * @param name name of the file (without dots) or <CODE>null</CODE> if
    *    one wants to obtain the name of a package and not a file in it
    * @param ext extension of the file or <CODE>null</CODE> if one needs
    *    a package and not a file name
    *
    * @return {@link FileObject} that represents file with given name or
    *   <CODE>null</CODE> if the file does not exist
    */
    public final FileObject find (String aPackage, String name, String ext) {
        Enumeration en = getFileSystems ();
        while (en.hasMoreElements ()) {
            FileSystem fs = (FileSystem)en.nextElement ();
            FileObject fo = fs.find (aPackage, name, ext);
            if (fo != null) {
                // object found
                return fo;
            }
        }
        return null;
    }


    /** Searches for the given resource among all file systems.
    * @see FileSystem#findResource
    * @param name a name of the resource
    * @return file object or <code>null</code> if the resource can not be found
    */
    public FileObject findResource(String name) {
        Enumeration en = getFileSystems ();
        while (en.hasMoreElements ()) {
            FileSystem fs = (FileSystem)en.nextElement ();
            FileObject fo = fs.findResource(name);
            if (fo != null) {
                // object found
                return fo;
            }
        }
        return null;
    }

    /** Searches for the given resource among all file systems, returning all matches.
    * @param name name of the resource
    * @return enumeration of {@link FileObject}s
    */
    public Enumeration findAllResources(String name) {
        Vector v = new Vector(8);
        Enumeration en = getFileSystems ();
        while (en.hasMoreElements ()) {
            FileSystem fs = (FileSystem)en.nextElement ();
            FileObject fo = fs.findResource(name);
            if (fo != null) {
                v.addElement(fo);
            }
        }
        return v.elements();
    }

    /** Finds all files among all file systems matching a given name, returning all matches.
    * All file systems are queried with {@link FileSystem#find}.
    *
    * @param aPackage package name where each package is separated by a dot
    * @param name name of the file (without dots) or <CODE>null</CODE> if
    *    one wants to obtain the name of a package and not a file in it
    * @param ext extension of the file or <CODE>null</CODE> if one needs
    *    a package and not a file name
    *
    * @return enumeration of {@link FileObject}s
    */
    public final Enumeration findAll (String aPackage, String name, String ext) {
        Enumeration en = getFileSystems ();
        Vector ret = new Vector();
        while (en.hasMoreElements ()) {
            FileSystem fs = (FileSystem)en.nextElement ();
            FileObject fo = fs.find (aPackage, name, ext);
            if (fo != null) {
                ret.addElement(fo);
            }
        }
        return ret.elements();
    }

    /** Fire info about changes in the file system pool.
    * @param fs file system
    * @param add <CODE>true</CODE> if the file system is added,
    *   <CODE>false</CODE> if it is removed
    */
    private void fireFileSystem (FileSystem fs, boolean add) {
        Enumeration en = ((Hashtable)listeners.clone ()).elements ();
        RepositoryEvent ev = new RepositoryEvent (this, fs, add);
        while (en.hasMoreElements ()) {
            RepositoryListener list = (RepositoryListener)en.nextElement ();
            if (add) {
                list.fileSystemAdded (ev);
            } else {
                list.fileSystemRemoved (ev);
            }
        }
    }

    /** Fires info about reodering
    * @param perm
    */
    private void fireFileSystemReordered(int[] perm) {
        Enumeration en = ((Hashtable)listeners.clone ()).elements ();
        RepositoryReorderedEvent ev = new RepositoryReorderedEvent(this, perm);
        while (en.hasMoreElements ()) {
            RepositoryListener list = (RepositoryListener)en.nextElement ();
            list.fileSystemPoolReordered(ev);
        }
    }

    /** Adds new listener.
    * @param list the listener
    */
    public void addRepositoryListener (RepositoryListener list) {
        listeners.put (list, list);
    }

    /** Removes listener.
    * @param list the listener
    */
    public void removeRepositoryListener (RepositoryListener list) {
        listeners.remove (list);
    }

    /** Writes the object to the stream.
    */
    private Object writeReplace () {
        return new Replacer ();
    }

    private static class Replacer implements java.io.Serializable {
        /** serial version UID */
        static final long serialVersionUID=-3814531276726840241L;

        private void writeObject (ObjectOutputStream oos) throws IOException {
            ExternalUtil.getRepository ().writeExternal (oos);
        }

        private void readObject (ObjectInputStream ois)
        throws IOException, ClassNotFoundException {
            ExternalUtil.getRepository ().readExternal (ois);
        }

        /** @return the default pool */
        public Object readResolve () {
            return ExternalUtil.getRepository ();
        }
    }
}
