/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.filesystems;

import java.lang.ref.*;
import java.io.*;
import java.lang.reflect.*;

import java.beans.*;
import java.util.*;
import java.net.*;

import org.openide.util.enum.EmptyEnumeration;
import org.openide.util.NbBundle;

import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;

/** XML-based file system.
 * <PRE>
 *  Description of format of XML file (which can be parsed by XMLFileSystem)
 * ==================================================================
 * Allowed Elements:	filesystem,file,folder,attr
 *
 * Mandatory attributes:	
 *         -for filesystem    version=... (e.g. "1.0")
 *         -for file,folder,attr name=....  (e.g.: &lt;folder name="Config"&gt;)
 *         -for attr is mandatory one of bytevalue,shortvalue,intvalue,longvalue,floatvalue,doublevalue,boolvalue,charvalue,stringvalue,methodvalue,serialvalue,urlvalue
 * 
 * Allowed atributes:
 *         -for file:	url=.... (e.g.: &lt;file name="sample.xml" url="file:/c:\sample.xml"&gt;)
 * 	-for folder,filesystem	nothing allowed
 *
 *
 *
 * Note: file can contain 	content e.g.:
 *  &lt; file name="sample.java"&gt;
 * &lt; ![CDATA[
 * package org.sample;
 * import java.io;
 * ]]&gt;
 * &lt; /file&gt;
 *
 *
 *
 * This class implements virtual FileSystem. It is special case of FileSystem in XML format.
 *
 * Description of this format best ilustrate DTD file that is showed in next lines: 
 * &lt; !ELEMENT filesystem (file | folder)*&gt;
 * &lt; !ATTLIST filesystem version CDATA #REQUIRED&gt; //version not checkked yet
 * &lt; !ELEMENT folder (file |folder | attr)*&gt;
 * &lt; !ATTLIST folder name CDATA #REQUIRED&gt; //name of folder
 * &lt; !ELEMENT file (#PCDATA | attr)*&gt;
 * &lt; !ATTLIST file name CDATA #REQUIRED&gt; //name of file
 * &lt; !ATTLIST file url CDATA #IMPLIED&gt; //content of the file can be find at url
 * &lt; !ELEMENT attr EMPTY&gt;
 * &lt; !ATTLIST attr name CDATA #REQUIRED&gt; //name of attribute
 * &lt; !ATTLIST attr bytevalue CDATA #IMPLIED&gt;//the rest - types of attributes
 * &lt; !ATTLIST attr shortvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr intvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attribute longvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr floatvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr doublevalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr boolvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr charvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr stringvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr methodvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr serialvalue CDATA #IMPLIED&gt;
 * &lt; !ATTLIST attr urlvalue CDATA #IMPLIED&gt;
 * </PRE>
 * @author Radek Matous
 */

public final class XMLFileSystem extends AbstractFileSystem {
    
    static final long serialVersionUID =  28974107313702326L;

    private static final String[] NO_CHILDREN = {};    
    
    private static final int    STATUS_INIT = 0;
    private static final int    STATUS_ADD = 1;
    private static final int    STATUS_REMOVE = 2;
    private static final int    STATUS_CHILDS_FROM_SUPERCLS = 3;
    
    private transient int       status = STATUS_INIT;

    private transient FileObjRef retRef;
    private  transient AbstractFolder parentFo = null;

    
    /**  Url location of XML document    */            
    private URL uriId;

    private transient int rdOnlySrc = -1;        
    
      
    // <?xml version="1.0"?>
    // <!DOCTYPE filesystem PUBLIC "-//NetBeans//DTD Filesystem 1.0//EN" "http://www.netbeans.org/dtds/filesystem-1_0.dtd">
    // <filesystem>...</filesystem>
    
    private static final String PUBLIC_ID = "-//NetBeans//DTD Filesystem 1.0//EN";// NOI18N
    private static final String DTD_PATH = "org/openide/filesystems/filesystem.dtd";// NOI18N

    /** Creates new XMLFileSystem */
    public XMLFileSystem() {
        Impl impl = new Impl (this);
        this.list = impl;
        this.info = impl;
        this.change = impl;
        this.attr = impl;
    }
    
   public XMLFileSystem(String uri)  throws  SAXException{
           this();
           try {
              setXmlUrl(new URL(uri));
           } catch(Exception e) {
                throw (SAXException)ExternalUtil.copyAnnotation (new SAXException (e.getMessage()),e);                                                              
           }
       }
       
       public XMLFileSystem(URL url) throws  SAXException{
           this();
           try {
               setXmlUrl(url);
           } catch(Exception e) {
                throw (SAXException)ExternalUtil.copyAnnotation (new SAXException (e.getMessage()),e);                                              
           }         
       }    

       
    /** Constructor. Allows user to provide own capabilities
    * for this file system.
    * @param cap capabilities for this file system
    */
    public XMLFileSystem(FileSystemCapability cap) {
        this ();
        setCapability (cap);
    }

   
    
    /** Getter of url field. 
    */
    public URL getXmlUrl () { // JST
        return uriId;
    }

    /**
     * Setter of url field. Set name of the XML file.
     * @param aRoot path to new ZIP or JAR file
     * @throws IOException if the file is not valid
     */    
    public synchronized void setXmlUrl (URL url) throws IOException, PropertyVetoException {    
        setXmlUrl (url, false);
    }
    /**
     * Setter of url field. Set name of the XML file.
     * @param aRoot path to new ZIP or JAR file
     * @param validate if true uses validating parser else nonvalidating
     * @throws IOException if the file is not valid
     */
    public synchronized void setXmlUrl (URL url, boolean validate) throws IOException, PropertyVetoException {
        String uri = url.toString().trim();
        uriId = url;
        rdOnlySrc = -1;
        
        status = STATUS_INIT;
        refreshResource("",true);// NOI18N
                                        
        setSystemName("XML_"+url.toString());// NOI18N
        try {
            Stack  attrStack = new Stack();
            ElementHandler[] elmKeyService = {fileSystemElement(),folderElement(attrStack),fileElement(attrStack),attrElement(attrStack)};            
            String dtd = getClass().getClassLoader().getResource( DTD_PATH ).toString();
            (new InnerParser(PUBLIC_ID, dtd, elmKeyService)).parseXML(uri, validate);
        } catch (Exception e) { 
            IOException x = new IOException (e.getMessage()+" : "+url.toString());
            ExternalUtil.copyAnnotation (x,e);
            throw x;
        }
    }
        
    
    /**
     * Test if the file is folder or contains data.
     * @param name name of the file
     * @return true if the file is folder, false otherwise
     */
    private boolean isFolder(String name) {
        Reference ref = findReference(name);
        
        if (ref != null && (ref instanceof FileObjRef) ) {
            return ((FileObjRef) ref).isFolder();
        }
        return false;
    }
    
    /**
    * Get input stream.
    *
    * @param name the file to test
    * @return an input stream to read the contents of this file
    * @exception FileNotFoundException if the file does not exists or is invalid
    */    
    private InputStream getInputStream(String name) throws java.io.FileNotFoundException  {
        Reference ref = findReference(name);        
        if (ref != null && (ref instanceof FileObjRef ) ) {
            return ((FileObjRef) ref).getInputStream(getXmlUrl(),name);
        }        
        throw new FileNotFoundException (NbBundle.getMessage(XMLFileSystem.class,"EXC_CanntRead"));// NOI18N
    }

    /** Get size of stream*/
    private long getSize (String name)   {
        Reference ref = findReference(name);        
        if (ref != null && (ref instanceof FileObjRef ) ) {
            return ((FileObjRef) ref).getSize(getXmlUrl(),name);
        }        
        return 0;
    }
    
    
    /* Human presentable name */    
    public String getDisplayName() {
        if (uriId == null || uriId.toString().length() == 0) return  NbBundle.getMessage(XMLFileSystem.class,"XML_NotValidXMLFileSystem");// NOI18N
        return "XML:"+uriId.toString().trim();// NOI18N
    }
    
    /** Test if the filesystem is read-only or not.
     * @return true if the system is read-only
     */
    public boolean isReadOnly() {
        return true;
    }
        
    /** Initializes the root of FS.
    */
    private void readObject (ObjectInputStream ois) throws IOException, ClassNotFoundException {
        ois.defaultReadObject ();
        try {
            setXmlUrl(uriId);
        } catch (PropertyVetoException ex) {
            IOException x = new IOException (ex.getMessage());
            ExternalUtil.copyAnnotation (x,ex);
            throw x;

//            throw (IOException)ExternalUtil.copyAnnotation (new IOException (),ex);              
        }
    }
    
    /** Notifies this file system that it has been added to the repository.
     * Various initialization tasks should go here. Default implementation is noop.
     */
    public void addNotify() {}
    
    /** Notifies this file system that it has been removed from the repository.
     * Concrete file system implementations should perform clean-up here.
     * Default implementation is noop.
     */
    public void removeNotify() {}
    
    
    /** Creates Reference. In FileSystem, which subclasses AbstractFileSystem, you can overload method
     * createReference(FileObject fo) to achieve another type of Reference (weak, strong etc.)
     * @param fo is FileObject. It`s reference yourequire to get.
     * @return Reference to FileObject
     */
    protected  Reference createReference(FileObject fo) {
        FileObjRef ref = retRef;
        if ( retRef != null ) {
            ((FileObjRef) retRef).setFileObject(fo); 
            ref = retRef;
            retRef = null;
        } else {
            ref = new FileObjRef(fo,true);  
        }
        return(ref);
    }
    
    
    /** Adds file/folder resource to hiearchy
     * @param parent under parent is added new resource as its child
     * @param name  name of new resource
     * @return Reference to new created AbstractFolder
     */
    private final  AbstractFolder addResource(AbstractFolder parent,String name) throws SAXException{
        AbstractFolder retParent;        
        name = name.trim();                        
                        
        if (parent == null ) parent = (AbstractFolder)findResource("");// NOI18N
        
        retParent = (AbstractFolder)parent.getFileObject (name);        
        if (retParent != null) {
            // folders can be duplicated and  content of them will be combined
            // files cannot be duplicated
            if (retParent.isFolder () ) return retParent;
            else throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_DuplElement")+" "+name);            
        }
        
        parent.refresh(null,null,true,true,internalChildren(parent, name,STATUS_ADD));
        
        status = STATUS_INIT;
        retParent = (AbstractFolder)parent.getFileObject(name);
        if (retParent != null) retParent.getFileObject("EMPTY");// NOI18N - tricky way how to be sure that  createReference was called
        else throw new InternalError (NbBundle.getMessage(XMLFileSystem.class,"EXC_UnreachableCode"));// NOI18N
        
        status = STATUS_CHILDS_FROM_SUPERCLS;        
        return (retParent != null)?retParent:(AbstractFolder)findResource("");// NOI18N
    }
    
    
    /** Does work for children 
     * @param parent under parent is added new resource as its child
     * @param name  name of new resource
     * @return array of children`s names 
     */
    private String[] internalChildren(AbstractFolder parent,String name, int internalStatus ) {
        String[]        strChildren;
        synchronized (parent) {
            if ( internalStatus == STATUS_ADD ) {
                strChildren = new String[parent.getChildrenArray().length+1];
                for(int i = 0; i <strChildren.length-1;i++) {
                    strChildren[i] = parent.getChildrenArray()[i];
                }
                strChildren[strChildren.length-1] = name;
            } else strChildren = parent.getChildrenArray();
        }        
        return strChildren;
    }
    
    //private void debugInfo(String dbgStr) { System.out.println(dbgStr);}
        
    /** Implementation of all interfaces List, Change, Info and Attr
     * that delegates to JarFileSystem
     */
    public static class Impl extends Object
    implements AbstractFileSystem.List, AbstractFileSystem.Info,
    AbstractFileSystem.Change, AbstractFileSystem.Attr {
        /** generated Serialized Version UID */
        static final long serialVersionUID = -67233358102597232L;
        private long urlDate = 0;
        
        /** the pointer to file system */
        private XMLFileSystem fs;
        
        /** Constructor.
         * @param fs the file system to delegate to
         */
        public Impl (XMLFileSystem fs) {
            this.fs = fs;
        }
        
        /**
         *
         * Scans children for given name
         * @return array of children`s names 
         */
        public String[] children (String name) {
            String[]    strChildren;
            FileObject  fo2name;
            FileObject[]  foChildren;
            
            switch(fs.status) {
                case STATUS_CHILDS_FROM_SUPERCLS://In normal cases all List structures are hold in AbstractFileSystem
                    if (( fo2name = fs.findResource(name) ) == null) return NO_CHILDREN;                    
                    return fs.internalChildren((AbstractFolder)fo2name,"",fs.status);// NOI18N
            }
            return NO_CHILDREN;
        }
        
        //
        // Change
        //
        
        /**
         * Creates new folder named name.
         * @param name name of folder
         * @throws IOException if operation fails
         */
        public void createFolder (String name) throws java.io.IOException {
            throw new IOException ();
        }
        
        
        /**
         * Create new data file.
         *
         * @param name name of the file
         *
         * @return the new data file object
         * @exception IOException if the file cannot be created (e.g. already exists)
         */
        public void createData (String name) throws IOException {
            throw new IOException ();
        }
        
        /**
         * Renames a file.
         *
         * @param oldName old name of the file
         * @param newName new name of the file
         */
        public void rename(String oldName, String newName) throws IOException {
            throw new IOException ();
        }
        
        /**
         * Delete the file.
         *
         * @param name name of file
         * @exception IOException if the file could not be deleted
         */
        public void delete (String name) throws IOException {
            throw new IOException ();
        }
        
// Info
        //
        private java.util.Date localDate = null;
        
        /**
         *
         * Get last modification time.
         * @param name the file to test
         * @return the date
         */
        public java.util.Date lastModified(String name) {
            switch (fs.rdOnlySrc) {
                case -1 :
                    if (isRdOnlyProtocol(fs.uriId.getProtocol())) {
                        fs.rdOnlySrc = 1;
                        return localTime();
                    } else {
                        fs.rdOnlySrc = 0;
                        return trueTime();
                    }
                case 0 :
                    return trueTime();
                default :  
                    return localTime();
            }
        }
        
        private boolean isRdOnlyProtocol(String proto) {
            return proto.equals("jar");//NOI18N
        }

        private java.util.Date localTime() {
            if (localDate == null) {
                localDate = new java.util.Date();
            }
            return localDate;
        }
                    
        private java.util.Date trueTime() {
            URLConnection urlConn;
            if (urlDate == 0) {
                try {
                    urlConn = fs.uriId.openConnection ();
                    urlDate = urlConn.getDate();                    
                } catch (IOException ie) {
                    Calendar cal = Calendar.getInstance();
                    cal.set(2000,0,1);
                    urlDate = cal.getTime().getTime();
                }
                
            }
            return new java.util.Date (urlDate);//System.currentTimeMillis()
        }
                
        /**
         * Test if the file is folder or contains data.
         * @param name name of the file
         * @return true if the file is folder, false otherwise
         */
        public boolean folder (String name) {
            return fs.isFolder(name);
        }
        
        /**
         * Test whether this file can be written to or not.
         * @param name the file to test
         * @return <CODE>true</CODE> if file is read-only
         */
        public boolean readOnly (String name) {
            return true;
        }
        
    /** Get the MIME type of the file. If filesystem has no special support 
    * for MIME types then can simply return null. FileSystem can register 
    * MIME types for a well-known extensions: FileUtil.setMIMEType(String ext, String mimeType)
    * or together with filesystem supply some resolvers subclassed from MIMEResolver.
    *
    * @param name the file to test
    * @return the MIME type textual representation (e.g. <code>"text/plain"</code>)
    * or null if no special support for recognizing MIME is implemented.
     */
        public String mimeType (String name) {
            return null;
        }
        
        /**
         * Get the size of the file.
         *
         * @param name the file to test
         * @return the size of the file in bytes or zero if the file does not contain data (does not
         *  exist or is a folder).
         */
        public long size (String name) {
            if (fs.isFolder(name)) return 0;               
            return fs.getSize (name); 
        }
        
        /**
         * Get input stream.
         *
         * @param name the file to test
         * @return an input stream to read the contents of this file
         * @exception FileNotFoundException if the file does not exists or is invalid
         */
        public InputStream inputStream (String name) throws java.io.FileNotFoundException {            
            InputStream is = fs.getInputStream(name);
            if (is == null) throw new java.io.FileNotFoundException (name);
            return is; 
        }
        
        /**
         * Get output stream.
         *
         * @param name the file to test
         * @return output stream to overwrite the contents of this file
         * @exception IOException if an error occures (the file is invalid, etc.)
         */
        public OutputStream outputStream (String name) throws java.io.IOException {
            throw new IOException ();
        }
        
        /**
         * Does nothing to lock the file.
         *
         * @param name name of the file
         */
        public void lock (String name) throws IOException {
            FSException.io ("EXC_CannotLock", name, fs.getDisplayName (), name); // NOI18N
        }
        
        /**
         * Does nothing to unlock the file.
         *
         * @param name name of the file
         */
        public void unlock (String name) {
        }
        
        /**
         * Does nothing.
         *
         * @param name the file to mark
         */
        public void markUnimportant (String name) {
        }
        
        /** Get the file attribute with the specified name.
         * @param name the file
         * @param attrName name of the attribute
         * @return appropriate (serializable) value or <CODE>null</CODE> if the attribute is unset (or could not be properly restored for some reason)
         */
        public Object readAttribute(String name,String attrName) {
            FileObjRef ref = (FileObjRef)fs.findReference(name);            
            if (ref == null  )  return null;
            return ref.readAttribute(attrName);            
        }
        /** Set the file attribute with the specified name.
         * @param name the file
         * @param attrName name of the attribute
         * @param value new value or <code>null</code> to clear the attribute. Must be serializable, although particular file systems may or may not use serialization to store attribute values.
         * @exception IOException if the attribute cannot be set. If serialization is used to store it, this may in fact be a subclass such as {@link NotSerializableException}.
         */
        public void writeAttribute(String name,String attrName,Object value) throws IOException {
            throw new IOException ();            
        }
        
        /** Get all file attribute names for the file.
         * @param name the file
         * @return enumeration of keys (as strings)
         */
        public Enumeration attributes(String name) {
            FileObjRef ref = (FileObjRef)fs.findReference(name);            
            if (ref == null  )  return EmptyEnumeration.EMPTY;
            return ref.attributes();
        }
        
        /** Called when a file is renamed, to appropriately update its attributes.
 * @param oldName old name of the file
 * @param newName new name of the file
 */
        public void renameAttributes(String oldName,String newName) {
        }
        
        /** Called when a file is deleted, to also delete its attributes.
 *
 * @param name name of the file
 */
        public void deleteAttributes(String name) {
        }
        
    }

    /** Creates new anonymous subclass of ElementHandler. 
     * Processes: "FILESYSTEM" element
     */    
        private  ElementHandler fileSystemElement() {
            ElementHandler elemService = new ElementHandler() {
                private final  String[]  ELM_KEYS = {"FILESYSTEM"};// NOI18N
                private final  String[]  MANDAT_ATTR_KEYS =  {};                //"VERSION"
                
                public void internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {
                    if (parentFo != null ) throw new SAXException(NbBundle.getMessage(XMLFileSystem.class,"XML_RootInvalid"));// NOI18N 
                }
                protected  String[] getKeys() {return ELM_KEYS;}
                protected  String[] getMandatoryAttrs() {return MANDAT_ATTR_KEYS;}                
            };
            return elemService;
        }

    /** Creates new anonymous subclass of ElementHandler. 
     * Processes: "FOLDER" element
     */            
        private  ElementHandler folderElement(final Stack  attrStack) {
            ElementHandler elemService = new ElementHandler() {
                private final  String[]  ELM_KEYS = {"FOLDER"};// NOI18N
                private final  String[]  MANDAT_ATTR_KEYS =  {"NAME"};// NOI18N
                Stack          folderStack = new Stack();
                //
                public void internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {
                    attrStack.push(new XMLMapAttr());                                        
                    String nameAttr = (String)mapMandatory.get("NAME");// NOI18N
                    if (nameAttr == null) nameAttr = (String)mapMandatory.get("name");// NOI18N
                    
                    if (nameAttr == null) throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_MisssingAttr"));// NOI18N 
                    
                    folderStack.push(retRef = new FileObjRef(mapAllowed,true));
                    parentFo = addResource(parentFo,nameAttr);                    
                    
                    if (parentFo != null && retRef != null && retRef.get () == null) {
                        Reference ref = findReference(parentFo.getPackageNameExt('/','.'));
                        if (ref != null) {
                            folderStack.pop ();                             
                            folderStack.push (ref);
                        }
                    }
                    retRef = null;                    
                }                
                public void endElement(String elementName) throws SAXException {                    
                    FileObjRef tempRef = (FileObjRef)folderStack.pop();
                    XMLMapAttr     foAttrs = (XMLMapAttr)attrStack.pop();
                    
                    if (tempRef != null && foAttrs != null) {
                        tempRef.setFOAttributes(foAttrs);
                    }
                    parentFo = (AbstractFolder)parentFo.getParent ();                                        
                }
                
                protected  String[] getKeys() {return ELM_KEYS;}
                protected  String[] getMandatoryAttrs() {return MANDAT_ATTR_KEYS;}
            };
            return elemService;
        }

    /** Creates new anonymous subclass of ElementHandler. 
     * Processes: "FILE" element
     */                    
        private  ElementHandler fileElement(final Stack  attrStack) {
            ElementHandler elemService = new ElementHandler() {
                private final  String[]  ELM_KEYS = {"FILE"};// NOI18N
                private final  String[]  MANDAT_ATTR_KEYS =  {"NAME"};// NOI18N
                private final  String[]  ALLOWED_ATTR_KEYS =  {"URL"};// NOI18N
                HashMap   tempMap;
                String  nameAttr;                               
                
                private ByteArrayOutputStream os;
                private OutputStreamWriter wr;

                public void internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {
                    attrStack.push(new XMLMapAttr());                    
                    String nameAttr = (String)mapMandatory.get("NAME");// NOI18N
                    if (nameAttr == null) nameAttr = (String)mapMandatory.get("name");// NOI18N
                    
                    tempMap = mapAllowed;
                    this.nameAttr = nameAttr;                    
                    os = new ByteArrayOutputStream();
                    wr = new OutputStreamWriter(os);
                }
                public void characters(char[] ch, int start, int length) throws SAXException {
                    try {
                        wr.write(ch,start,length);
                    } catch(IOException e) {
                        SAXException x = new SAXException (e.getMessage());
                        ExternalUtil.copyAnnotation (x,e);
                        throw x;                                                
                    }
                }
                
                public void endElement(String elementName) throws SAXException {
                    XMLMapAttr     foAttrs = (XMLMapAttr)attrStack.pop();
                    try {
                        wr.close();
                    } catch (Exception e ) {
                        SAXException x = new SAXException (e.getMessage());
                        ExternalUtil.copyAnnotation (x,e);
                        throw x;                                                                        
                    }
                    
                    if (os.size() != 0) {
                        retRef = new OnlyFileRef(tempMap,os.toByteArray());
                        if (foAttrs != null) {
                            retRef.setFOAttributes(foAttrs);
                        }
                    }
                    else {
                        retRef = new FileObjRef(tempMap,false);
                        if (foAttrs != null) {
                            retRef.setFOAttributes(foAttrs);
                        }
                    }
                    
                    addResource(parentFo,nameAttr);                    
                    retRef = null;                     
                }
                
                protected   String[] getKeys() {return ELM_KEYS;}
                protected  String[] getMandatoryAttrs() {return MANDAT_ATTR_KEYS;}
                protected  String[] getAllowedAttrs() {return ALLOWED_ATTR_KEYS;}
            };
            return elemService;
        }
        
    /** Creates new anonymous subclass of ElementHandler. 
     * Processes: "ATTRIBUTE" element
     */                    
        private  ElementHandler attrElement(final Stack  attrStack) {
            ElementHandler elemService = new ElementHandler() {
                private final  String[]  ELM_KEYS = {"ATTR"};// NOI18N
                private final  String[]  MANDAT_ATTR_KEYS =  {"NAME"};// NOI18N

                public void internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {
                    XMLMapAttr foAttrs = (XMLMapAttr)attrStack.peek();//XMLMapAttr
                    String nameAttr = (String)mapMandatory.get("NAME");// NOI18N
                    if (nameAttr == null) nameAttr = (String)mapMandatory.get("name");// NOI18N
                    
                    if (mapAllowed.size() == 0) return;
                    Iterator it = mapAllowed.entrySet ().iterator ();
                    while (it.hasNext ()) {
                        Map.Entry pair = (Map.Entry) it.next ();
                        if (XMLMapAttr.Attr.isValid((String)pair.getKey()) != -1) {
                            XMLMapAttr.Attr attr = new XMLMapAttr.Attr((String)pair.getKey(),(String)pair.getValue());
                            foAttrs.put(nameAttr,attr);
                            return;
                        }
                    }
                    
                }
                protected   String[] getKeys() {return ELM_KEYS;}
                protected  String[] getMandatoryAttrs() {return MANDAT_ATTR_KEYS;}
                protected  String[] getAllowedAttrs() {return  XMLMapAttr.Attr.getAttrTypes ();}//ALLOWED_ATTR_KEYS;
            };
            return elemService;
        }            
        
    
    
    /** Strong reference to FileObject. To FileObject may be attached attributes (XMLMapAttr) 
     *  and info about if it is folder or not.
     */
    private static class FileObjRef  extends WeakReference {
        private FileObject fo;
        private boolean     isFolder;
        private XMLMapAttr foAttrs;
        protected HashMap     attr;        
        
        public FileObjRef (FileObject fo,boolean isFolder) {
            super (null);
            this.isFolder = isFolder;
            this.fo = fo;
        }
        
        public FileObjRef (HashMap    attr,boolean isFolder) {
            super (null);
            this.isFolder = isFolder;
            this.attr = attr;            
        }

        public void attachInfo(HashMap  attr) {this.attr = attr;}        
        public HashMap getAttributeList() { return attr;}        
        public boolean isFolder() {return isFolder;}        
        public void setFileObject(FileObject fo) {this.fo = fo;}        
        public Object get () { return fo; }        
        public void setFOAttributes(XMLMapAttr foAttrs) {
            if (this.foAttrs != null) {
                this.foAttrs.putAll (foAttrs);
            } else this.foAttrs = foAttrs;
        }

        public Enumeration attributes() {            
            if (foAttrs == null) {
                return EmptyEnumeration.EMPTY;
            } else {
                HashSet s = new HashSet (foAttrs.keySet ());
                return Collections.enumeration (s);
            }
        }
        
        public Object readAttribute(String attrName) {
            Object o;
            if (foAttrs == null) return null;
            Object[] objs = new Object[] {fo,attrName};
            o = foAttrs.get(attrName,objs);//fo,attrName
           return o;
        }
    /**
    * Get input stream.
    *
    * @param name the file to test
    * @return an input stream to read the contents of this file
    * @exception FileNotFoundException if the file does not exists or is invalid
    */        
        public InputStream getInputStream(URL context,String name) throws java.io.FileNotFoundException  {
            String urlStr;
            URL url;
            
            InputStream is;
            urlStr = (String) attr.get("URL");// NOI18N
            if (urlStr == null) urlStr = (String) attr.get("url");// NOI18N
            if (urlStr == null ) {
                //throw new java.io.FileNotFoundException (name);
                return new ByteArrayInputStream (new byte[] {});
            }
            
            try {
                url = new URL(context,urlStr);
                is = url.openStream();
            } catch(IOException ex) {// neni koser osetreni - RM
                FileNotFoundException x = new FileNotFoundException (name);
                ExternalUtil.copyAnnotation (x,ex);
                throw x;                                                                
            }
            if (is == null) throw new FileNotFoundException (name);
            return is;
        }
        
        public long getSize (URL context,String name)  {
            URL url0;
            String urlStr = (String) attr.get("URL");// NOI18N
            long retVal = 0;
            if (urlStr == null) urlStr = (String) attr.get("url");// NOI18N
            if (urlStr == null ) return 0;
            
            try {
                url0 = new URL(context,urlStr);
                retVal = url0.openConnection().getContentLength();
            } catch (IOException iex) {
                return 0;
            }
            
            return retVal;
        }
    }
    /** Strong reference to FileObject. To FileObject may be attached attributes (XMLMapAttr) 
     *  and info about if it is folder or not. In addition there is attched content of a file as a byte array.
     */
    private static class  OnlyFileRef extends  FileObjRef{
        byte[]  content;
        
        public OnlyFileRef (FileObject fo,byte[]  content) {
            super (fo,false);
            this.content = content;
        }
        
        public OnlyFileRef (HashMap    attr,byte[]  content) {
            super (attr,false);
            this.content = content;
        }
        
        public void content(byte[]  content) {
            this.content = content;
        }
        public InputStream getInputStream(URL context,String name) throws java.io.FileNotFoundException  {
            if (content == null || content.length == 0) return super.getInputStream(context,name);
            return new ByteArrayInputStream(content);
        }

        public long getSize (URL context,String name)  {
            if (content == null || content.length == 0) return super.getSize (context,name);            
            return content.length;
        }                
    }
    
    /** Element handler should be used as superclass for future classes. These future classes should be passed 
     * to constructor of InnerParser as Array. (ElementHandler[]). Each subclass of ElementHandler is responsible for 
     * processing one or more elements in XML file.
     * Each subclass of ElementHandler should overwrite one or more of these methods:
     * - protected   String[] getKeys() 
     * - protected   String[] getMandatoryAttrs() 
     * - protected   String[] getAllowedAttrs() 
     * - protected  void endElement(String name) throws SAXException {}        
     * - protected  void characters(char[] ch, int start, int length) throws SAXException {}        
     * - protected  void  internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {}        
     */    
    static abstract class  ElementHandler {
        private int mandatAttrCount;
                        
        public  void  startElement(String elemName, AttributeList attrList) throws SAXException {
            HashMap mapAllowed =new HashMap();
            HashMap mapMandatory =new HashMap();
            
            if ( checkAttributes(attrList,mapMandatory,mapAllowed) == false ) {
                throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_InaccurateParam")+": "+elemName);// NOI18N 
            }
            internalStartElement(elemName, mapMandatory,mapAllowed);
        }
        
    /** Inner parser calls this method to notify this class that start element was parsed (<someelement>)
     * @param elemName  name of element
     * @param mapMandatory  map(String attributeName,String attributeValue) which holds pairs attributeName and attributeValue, which are mandatory for this element
     * @param mapAllowed map(String attributeName,String attributeValue) which holds pairs attributeName and attributeValue, which are optional for this element
     * @throws SAXException
     */        
        protected  void  internalStartElement(String elemName, HashMap mapMandatory,HashMap mapAllowed) throws SAXException {}                
        
    /** Inner parser calls this method to notify this class that there is content between start element and end element
     * @param ch[]  array of characters found between start and end element
     * @param start is start position in ch[] 
     * @param length is length of content
     * @throws SAXException
     */        
        protected  void characters(char[] ch, int start, int length) throws SAXException {}                
    /** Inner parser calls this method to notify this class that end element was parsed 
     * @param elemName  name of element
     * @throws SAXException
     */                
        protected  void endElement(String elemName) throws SAXException {}        

    /** @return names of elements which this class can process
     */        
        protected   String[] getKeys() {
            String[] empty = {};
            return empty;
        }
    /** @return names of attributes which are checked and are mandatory
     */        
        protected   String[] getMandatoryAttrs() {
            String[] empty = {};
            return getKeys();
        }               
    /** @return names of attributes which are allowed, are expected, but are not mandatory
     */        
        protected   String[] getAllowedAttrs() {
            String[] empty = {};
            return empty;
        }
        

        private   int isMyTag(String name) {return isInArray(name,getKeys());}                
        private   int isAllowedAttr(String name) {return isInArray(name,getAllowedAttrs());}        
        private   boolean isMandatOK() {return (mandatAttrCount == getMandatoryAttrs().length);}        
        private   int isMandatoryAttr(String name) {
            int retValue = isInArray(name,getMandatoryAttrs());
            if ( retValue != -1 )  mandatAttrCount++; 
            return retValue;
        }
                
        private int isInArray(String name,String[] arr) {
            if (arr == null || name == null) return -1;
            String correctStr = name.trim();
            for( int i = 0; i < arr.length; i++) {
                if (correctStr.equalsIgnoreCase(arr[i]) == true)  return i;                 
            }
            return -1;
        }
        
        private  boolean checkAttributes(AttributeList attrList,HashMap mapMandatory,HashMap mapAllowed) {
            String temp;
            mandatAttrCount = 0;
            if (attrList == null) return false;
            for (int i = 0; i < attrList.getLength(); i++) {
                if (isMandatoryAttr(attrList.getName(i))  != -1) {
                    temp = attrList.getName(i).toUpperCase();
                    mapMandatory.put(temp,attrList.getValue(i));
                    continue;                    
                }
                if ( isAllowedAttr(attrList.getName(i)) != -1) {
                    temp = attrList.getName(i).toUpperCase();
                    mapAllowed.put(temp,attrList.getValue(i));
                    continue;
                }                
            }
            return isMandatOK();
        }        
    }
    /** Class that can be used to parse XML document (Expects array of ElementHandler clasess).  Calls handler methods of ElementHandler clasess.
     */                    
    static class InnerParser extends HandlerBase  {
        private  ElementHandler[] elmKeyService;// = {fileSystemElement(attrStack),folderElement(attrStack),fileElement(attrStack),attrElement(attrStack)};        
        private String tagInProcess = "";
        private String publicId;
        private String publicURL;        
        
        
        InnerParser(String publicId, String publicURL, ElementHandler[] elmKeyService) {
            this.elmKeyService = elmKeyService;
            this.publicId = publicId;
            this.publicURL = publicURL;
        }
        
    /** Starts parsing document, that can be localized by means of uri parameter
     * @param uri adress of document, that will be parsed 
     * @throws IOException
     * @throws SAXException
     */        
        public void parseXML(String uri, boolean validate) throws IOException,SAXException,ParserConfigurationException {
            Parser parser = getParser(validate);
            parser.parse(uri);
        }

    /** Starts parsing document - if you have document`s InputStream
     * @param is document`s InputStream 
     * @throws IOException
     * @throws SAXException     
     */                
        public void parseXML(InputStream is, boolean validate) throws IOException,SAXException,ParserConfigurationException {
            InputSource iSource = new InputSource(is);
            Parser parser = getParser(validate);
            parser.parse(iSource);
        }
                
        private Parser getParser(boolean validate) throws IOException,SAXException,ParserConfigurationException{
            Parser parser;
            SAXParserFactory    factory = SAXParserFactory.newInstance ();
            factory.setValidating (validate);
            factory.setNamespaceAware(false);            

            
            parser = factory.newSAXParser ().getParser (); 

            if (parser == null) {
                throw new InternalError (NbBundle.getMessage(XMLFileSystem.class,"EXC_UnreachableCode"));
            }
            // create document handler and register it
            //parser.setEntityResolver(entityRes);                                    
            parser.setEntityResolver(this);            
            parser.setDocumentHandler(this);//before new InnerParser() - now this            
            parser.setErrorHandler(this);            
            
            return parser;
        }
        
        public void error(SAXParseException exception) throws SAXException {
            throw exception;
        }

        public void warning(SAXParseException exception) throws SAXException {
            throw exception;                    
        }

        public void fatalError(SAXParseException exception) throws SAXException {
            throw exception;
        }
                
        public void startDocument() throws SAXException {}
        public void endDocument() throws SAXException {}        
        public void startElement(String name, AttributeList amap) throws SAXException {
            tagInProcess = name = name.trim();
            for (int i = 0; i < elmKeyService.length;i++) {
                if ( elmKeyService[i].isMyTag(name)  != -1) {
                    elmKeyService[i].startElement(name,amap);
                    return;
                }
            }
            throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_UnknownElement")+" "+name);// NOI18N 
        }
        
        public void endElement(String name) throws SAXException {
            for (int i = 0; i < elmKeyService.length;i++) {
                if ( elmKeyService[i].isMyTag(name.trim())  != -1) {
                    elmKeyService[i].endElement(name.trim());
                    return;
                }
            }
            throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_UnknownElement")+" "+name);// NOI18N 
        }
        
        public void characters(char[] ch, int start, int length) throws SAXException {
            String str = new String(ch,start,length).trim();
            if (str.length() != 0) {
                for (int i = 0; i < elmKeyService.length;i++) {
                    if ( elmKeyService[i].isMyTag(tagInProcess)  != -1) {
                        elmKeyService[i].characters(ch,start,length);
                        return;
                    }
                }
            throw new SAXException (NbBundle.getMessage(XMLFileSystem.class,"XML_UnknownElement")+" "+tagInProcess);// NOI18N
            }
        }
        
        public InputSource resolveEntity(java.lang.String pid,java.lang.String sid) throws SAXException   {
            if (pid != null && pid.equals(publicId)) 
                return new InputSource (publicURL);
            
            return new InputSource (sid);            
        }
        
    }
    
}
