/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;

import org.openide.*;
import org.openide.compiler.*;
import org.openide.compiler.Compiler;
import org.openide.filesystems.*;
import org.openide.cookies.CompilerCookie;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;

/** Support for compilation of data objects.
*
* @author Jaroslav Tulach
*/
public class CompilerSupport extends Object implements CompilerCookie {
    /** extended attribute for the type of compiler */
    // [PENDING] this is a misleading name, since it actually points to a handle to a CompilerType
    private static final String EA_COMPILER_MANAGER = "NetBeansAttrDataObjectCompilerManager"; // NOI18N
    /** Name of property providing a custom {@link CompilerType} for a file. */
    public static final String PROP_COMPILER_TYPE = "compiler"; // NOI18N

    /** entry to be associated with */
    private MultiDataObject.Entry entry;

    /** cookie class for the compilation */
    private Class cookie;

    /** New support for given entry. The file is taken from the
    * entry and is updated if the entry moves or renames itself.
    * @param entry entry to create instance from
    * @param cookie cookie class for the compilation (e.g. {@link CompilerCookie.Build})
    */
    protected CompilerSupport(MultiDataObject.Entry entry, Class cookie) {
        this.entry = entry;
        this.cookie = cookie;
    }

    /** Supports only {@link Compiler#DEPTH_ONE depth one}.
    */
    public boolean isDepthSupported (Compiler.Depth depth) {
        return Compiler.DEPTH_ONE == depth;
    }

    /* Adds the right compiler to the job.
    */
    public void addToJob (CompilerJob job, Compiler.Depth depth) {
        CompilerType comp = getCompilerType (entry);
        if (comp == null) {
            comp = defaultCompilerType ();
        }

        comp.prepareJob (job, cookie, entry.getDataObject ());
    }

    /** Allows subclasses to override the default compiler that should
    * be used for this support.
    */
    protected CompilerType defaultCompilerType () {
        return CompilerType.getDefault ();
    }

    /** Set compiler manager for a given file.
    * @param entry entry to set the compiler for
    * @param man type to use
    * @exception IOException if compiler cannot be set
    */
    public static void setCompilerType (MultiDataObject.Entry entry, CompilerType man)
    throws IOException {
        entry.getFile ().setAttribute (
            EA_COMPILER_MANAGER, man == null ? null : new ServiceType.Handle (man)
        );
    }

    /** Get compiler manager associated with a given file.
    * @param entry entry to obtain the compiler for
    * @return associated manager or null
    */
    public static CompilerType getCompilerType (MultiDataObject.Entry entry) {
        Object man = entry.getFile ().getAttribute (EA_COMPILER_MANAGER);
        if (man == null || ! (man instanceof ServiceType.Handle)) return null;
        ServiceType type = ((ServiceType.Handle) man).getServiceType ();
        if (type instanceof CompilerType) {
            return (CompilerType)type;
        } else {
            return null;
        }
    }

    /** Helper method that creates default properties for compilation of
    * given file entry.
    *
    * @param set sheet set to add properties to
    */
    public void addProperties (Sheet.Set set) {
        set.put (createCompilerProperty ());
    }

    /** Creates the compiler property.
    * @return the property
    */
    private PropertySupport createCompilerProperty () {
        return (PropertySupport)new PropertySupport.ReadWrite (
                   PROP_COMPILER_TYPE,
                   CompilerType.class,
                   DataObject.getString("PROP_compilerType"),
                   DataObject.getString("HINT_compilerType")
               ) {
                   public Object getValue() {
                       CompilerType ct = getCompilerType(entry);
                       if (ct == null)
                           return defaultCompilerType ();
                       else
                           return ct;
                   }
                   public void setValue (Object val) throws InvocationTargetException {
                       try {
                           setCompilerType(entry, (CompilerType) val);
                       } catch (IOException ex) {
                           throw new InvocationTargetException (ex);
                       }
                   }

                   public boolean supportsDefaultValue () {
                       return true;
                   }

                   public void restoreDefaultValue () throws InvocationTargetException {
                       setValue (null);
                   }
                   
                   public boolean canWrite () {
                       return !entry.getFile().isReadOnly();
                   }
               };
    }


    /** Compile cookie support.
    * Note that as a special case, when {@link Compiler#DEPTH_ONE} is requested,
    * a {@link CompilerCookie.Build} will actually be sent to the compiler manager,
    * rather than a {@link CompilerCookie.Compile}, on the assumption that the user
    * wished to force (re-)compilation of the single data object.
    */
    public static class Compile extends CompilerSupport
        implements CompilerCookie.Compile {
        /** New support for given entry. The file is taken from the
        * entry and is updated if the entry moves or renames itself.
        * @param entry entry to create instance from
        */
        public Compile (MultiDataObject.Entry entry) {
            super (entry, CompilerCookie.Compile.class);
        }
    }

    /** Build cookie support.
    */
    public static class Build extends CompilerSupport
        implements CompilerCookie.Build {
        /** New support for given entry. The file is taken from the
        * entry and is updated if the entry moves or renames itself.
        * @param entry entry to create instance from
        */
        public Build (MultiDataObject.Entry entry) {
            super (entry, CompilerCookie.Build.class);
        }
    }

    /** Clean cookie support.
    */
    public static class Clean extends CompilerSupport
        implements CompilerCookie.Clean {
        /** New support for given entry. The file is taken from the
        * entry and is updated if the entry moves or renames itself.
        * @param entry entry to create instance from
        */
        public Clean (MultiDataObject.Entry entry) {
            super (entry, CompilerCookie.Clean.class);
        }
    }

}
