/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.io.*;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import org.openide.*;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.*;
import org.openide.util.RequestProcessor;
import org.openide.util.enum.ArrayEnumeration;
import org.openide.util.enum.FilterEnumeration;
import org.openide.util.enum.SingletonEnumeration;
import org.openide.util.enum.SequenceEnumeration;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;
import org.openide.util.SharedClassObject;

/** Pool of data loaders.
 * Provides access to set of registered
 * {@link DataLoader loaders} in the system. They are used to find valid data objects
 * for given files.
 * <P>
 * The default instance can be retrieved with
 * {@link org.openide.TopManager#getLoaderPool}.
 *
 * @author Jaroslav Tulach, Petr Hamernik, Dafe Simonek
 */
public abstract class DataLoaderPool extends Object
implements java.io.Serializable {
    /** SUID */
    static final long serialVersionUID=-360141823874889956L;
    /** standard system loaders. Accessed by getSystemLoaders method only */
    private static DataLoader[] systemLoaders;
    /** standard default loaders. Accessed by getDefaultLoaders method only */
    private static DataLoader[] defaultLoaders;
    
    /** Cache of loaders for faster toArray() method. */
    private transient DataLoader[] loaderArray;
    /** List of listeners (ChangeListener) */
    private transient HashSet listeners;
    /** set of listeners (OperationListener) */
    private transient HashSet operations;
    
    /** prefered loader */
    private transient DataLoader preferredLoader;
    
    /** Create new loader pool.
     */
    protected DataLoaderPool () {
    }
    
    /** Create new loader pool and set preferred loader.
     * The preferred loader will be asked before any other to recognize files (also before the system
     * loader).
     *
     * @param loader the preferred loader
     */
    protected DataLoaderPool (DataLoader loader) {
        preferredLoader = loader;
    }
    
    /** Get an enumeration of data loaders.
     * Must be overridden in subclasses to provide a list of additional loaders.
     * The list should <em>not</em> include the preferred loader.
     *
     * @return enumeration of {@link DataLoader}s
     */
    protected abstract Enumeration loaders ();
    
    /** Add a new listener to the listener list. A listener is notified of
     * any change which was made to the loader pool (add, remove, or reorder).
     *
     * @param chl new listener
     */
    public final synchronized void addChangeListener (ChangeListener chl) {
        if (listeners == null)
            listeners = new HashSet();
        listeners.add(chl);
    }
    
    /** Remove a listener from the listener list.
     *
     * @param chl listener to remove
     */
    public final synchronized void removeChangeListener (ChangeListener chl) {
        if (listeners == null)
            return;
        listeners.remove(chl);
    }
    
    // Clears loaderArray before firing a change.
    /** Fire change event to all listeners. Asynchronously.
     * @param che change event
     */
    protected final void fireChangeEvent (final ChangeEvent che) {
        if (listeners == null)
            return;
        loaderArray = null;
        
        HashSet cloned;
        // clone listener list
        synchronized (this) {
            cloned = (HashSet)listeners.clone();
        }
        
        // fire on cloned list to prevent from modifications when firing
        Iterator iter = cloned.iterator();
        while (iter.hasNext()) {
            final ChangeListener l = (ChangeListener)iter.next();
            // separates the task to small pieces not to slow down the
            // rest of the IDE
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    l.stateChanged(che);
                }
            });
        }
    }
    
    /** Add a listener for operations on data objects.
     * @param l the listener
     */
    public synchronized final void addOperationListener (OperationListener l) {
        if (operations == null) operations = new HashSet ();
        operations.add (l);
    }
    
    /** Remove a listener for operations on data objects.
     * @param l the listener
     */
    public synchronized final void removeOperationListener (OperationListener l) {
        if (operations != null) {
            operations.remove (l);
        }
    }
    
    /** Fires operation event to all listeners.
     * Clears loaderArray before firing a change.
     * @param ev event to fire
     * @param type the type of the event
     */
    final void fireOperationEvent (OperationEvent ev, int type) {
        if (operations == null)
            return;
        HashSet cloned;
        // clone listener list
        synchronized (this) {
            cloned = (HashSet)operations.clone();
        }
        // fire on cloned list to prevent from modifications when firing
        for (Iterator iter = cloned.iterator(); iter.hasNext(); ) {
            OperationListener l = (OperationListener)iter.next ();
            switch (type) {
                case OperationEvent.COPY:
                    l.operationCopy ((OperationEvent.Copy)ev);
                    break;
                case OperationEvent.MOVE:
                    l.operationMove ((OperationEvent.Move)ev);
                    break;
                case OperationEvent.DELETE:
                    l.operationDelete (ev);
                    break;
                case OperationEvent.RENAME:
                    l.operationRename ((OperationEvent.Rename)ev);
                    break;
                case OperationEvent.SHADOW:
                    l.operationCreateShadow ((OperationEvent.Copy)ev);
                    break;
                case OperationEvent.TEMPL:
                    l.operationCreateFromTemplate ((OperationEvent.Copy)ev);
                    break;
                case OperationEvent.CREATE:
                    l.operationPostCreate (ev);
                    break;
            }
        }
    }
    
    /** Get an enumeration of all loaders, including the preferred and system loaders.
     * This should be the list of loaders as actually used by the system.
     * Typically it will consist of, in this order:
     * <ol>
     * <li>The preferred loader, if any.
     * <li>The system loaders, such as may be used for folders, shadows, etc.
     * <li>Module-specified loaders.
     * <li>The loader for instance data objects.
     * <li>Default loaders, which may handle files not otherwise recognizable.
     * </ol>
     * Applications should not rely on the exact contents of the pool,
     * rather the fact that this contains all the loaders which are
     * capable of recognizing files in the order in which they are
     * called.
     * @return enumeration of {@link DataLoader}s */
    public final Enumeration allLoaders () {
        // enumeration of systemloaders followed by normal loaders
        Enumeration en =new SequenceEnumeration (
        new SequenceEnumeration (
        new ArrayEnumeration (getSystemLoaders ()),
        loaders ()
        ),
        new ArrayEnumeration (getDefaultLoaders ())
        );
        if (preferredLoader != null) {
            // prepends preferred loader
            Enumeration queue = new SingletonEnumeration (preferredLoader);
            return new SequenceEnumeration (queue, en);
        } else {
            // enumeration without pref. loader
            return en;
        }
    }
    
    /** Get an array of loaders that are currently registered.
     * Does not include special system loaders, etc.
     * @return array of loaders
     * @see #loaders
     */
    public DataLoader[] toArray () {
        DataLoader[] localArray = loaderArray;
        if (localArray != null)
            return localArray;
        ArrayList loaders = new ArrayList ();
        Enumeration en = loaders ();
        while (en.hasMoreElements ()) {
            loaders.add(en.nextElement ());
        }
        localArray = new DataLoader[loaders.size()];
        localArray = (DataLoader[])loaders.toArray(localArray);
        loaderArray = localArray;
        return localArray;
    }
    
    /** Finds the first producer of a representation class.
     * Scans through the list of all loaders and returns the first one
     * whose representation class is a superclass of <code>clazz</code>.
     *
     * @param clazz class to find producer for
     * @return data loader or <CODE>null</CODE> if there is no loader that
     *   can produce the class
     */
    public final DataLoader firstProducerOf (Class clazz) {
        Enumeration en = allLoaders ();
        while (en.hasMoreElements ()) {
            DataLoader dl = (DataLoader)en.nextElement ();
            if (dl.getRepresentationClass ().isAssignableFrom (clazz)) {
                // representation class is super class of clazz
                return dl;
            }
        }
        return null;
    }
    
    /** Get an enumeration of all producers of a representation class.
     * @see #firstProducerOf
     *
     * @param clazz class to find producers for
     * @return enumeration of {@link DataLoader}s
     */
    public final Enumeration producersOf (final Class clazz) {
        return new FilterEnumeration (allLoaders ()) {
                   /** Accepts only those loaders that produces superclass of clazz
                    */
            public boolean accept (Object o) {
                DataLoader dl = (DataLoader)o;
                return clazz.isAssignableFrom( dl.getRepresentationClass() );
            }
        };
    }
    
    
    /** private class for next method. Empty implementation of
     * DataLoaderRecognized.
     */
    private static final DataLoader.RecognizedFiles emptyDataLoaderRecognized =
    new DataLoader.RecognizedFiles () {
            /** No op. replacement.
             *
             * @param fo file object to exclude
             */
        public void markRecognized (FileObject fo) {
        }
    };
    
    /** Find a data object for this file object (not for normal users of the APIs).
     * <strong>DO NOT USE THIS</strong> as a normal user of the APIs!
     * Unless you really know what you are doing, use {@link DataObject#find} instead.
     * This call will throw an exception if it already exists, and it is normally
     * only for use by the loader infrastructure.
     * <p>All loaders are asked to recognize it according to their priority.
     * @param fo file object to recognize
     * @return the data object for this object or <CODE>null</CODE> if
     *   no loader recognizes this file
     * @exception DataObjectExistsException if the object for this primary file
     *   already exists
     * @exception IOException if the data object is recognized but
     *   an error occurs during instantiation
     * @see #findDataObject(FileObject, DataLoader.RecognizedFiles)
     */
    public DataObject findDataObject (FileObject fo) throws IOException {
        return findDataObject (fo, emptyDataLoaderRecognized);
    }
    
    /** Find a data object for this file object, considering already-recognized files (not for normal users of the APIs).
     * <strong>DO NOT USE THIS</strong> as a normal user of the APIs!
     * Unless you really know what you are doing, use {@link DataObject#find} instead.
     * This call will throw an exception if it already exists, and it is normally
     * only for use by the loader infrastructure.
     * <p>First of all looks at the
     * file extended attribute <code>NetBeansDataLoader</code>; if it is set and it
     * contains the class name of a valid {@link DataLoader}, that loader is given preference.
     * For all loaders used, the first to return non-<code>null</code> from {@link DataLoader#findDataObject}
     * is used.
     *
     * @param fo file object to recognize
     * @param r recognized files buffer
     * @return the data object for this object
     * @exception DataObjectExistsException if the object for this primary file
     *   already exists
     * @exception IOException if the data object is recognized but
     *   an error occurs during instantiation
     */
    public DataObject findDataObject (
    FileObject fo, DataLoader.RecognizedFiles r
    ) throws IOException {
        // try to find assigned loader
        DataLoader pref = getPreferredLoader (fo);
        if (pref != null) {
            DataObject obj = pref.findDataObject (fo, r);
            if (obj != null) {
                DataObjectPool.POOL.notifyCreation ();
                // file has been recognized
                return obj;
            }
        }
        
        // scan through loaders
        java.util.Enumeration en = allLoaders ();
        while (en.hasMoreElements ()) {
            DataLoader l = (DataLoader)en.nextElement ();
            DataObject obj = l.findDataObject (fo, r);
            if (obj != null) {
                // the loader recognized the file
                
                // notify it
                DataObjectPool.POOL.notifyCreation ();
                
                return obj;
            }
        }
        return null;
    }
    
    /** Utility method to mark a file as belonging to a loader.
     * When the file is to be recognized this loader will be used first.
     *
     * @param fo file to mark
     * @param loader the loader to assign to the file or null if previous
     *    association should be cleared
     * @exception IOException if setting the file's attribute failed
     */
    public static void setPreferredLoader (FileObject fo, DataLoader loader)
    throws IOException {
        fo.setAttribute (DataObject.EA_ASSIGNED_LOADER, loader.getClass ().getName ());
    }
    
    /** Get the preferred loader for a file.
     * @param file the file to get loader from
     * @return the loader or null if no has been found
     */
    public static DataLoader getPreferredLoader (FileObject fo) {
        String assignedLoaderName = (String)fo.getAttribute (DataObject.EA_ASSIGNED_LOADER);
        if (assignedLoaderName != null) {
            try {
                Class loaderClass = Class.forName (
                assignedLoaderName,
                true,
                TopManager.getDefault().systemClassLoader()
                );
                return DataLoader.getLoader(loaderClass);
            } catch (Exception ex) {
                TopManager.getDefault ().getErrorManager ().notify (
                ErrorManager.INFORMATIONAL,
                ex
                );
            }
        }
        return null;
    }
    
    
    /** Lazy getter for system loaders.
     */
    private static DataLoader[] getSystemLoaders () {
        if (systemLoaders == null) {
            systemLoaders = new DataLoader [] {
                new ShadowLoader ()
            };
        }
        return systemLoaders;
    }
    
    /** Lazy getter for default loaders.
     */
    private static DataLoader[] getDefaultLoaders () {
        if (defaultLoaders == null) {
            defaultLoaders = new DataLoader [] {
                DataLoader.getLoader(FolderLoader.class),
                DataLoader.getLoader(XMLDataObject.Loader.class),
                DataLoader.getLoader(InstanceLoader.class),
                DataLoader.getLoader(DefaultLoader.class)
            };
        }
        return defaultLoaders;
    }
    
    /** Getter for default file loader
     * @return the default file loader
     */
    static DataLoader getDefaultFileLoader () {
        return getDefaultLoaders ()[3];
    }
    
    /** Getter for folder loader
     * @return the folder loader
     */
    static DataLoader getFolderLoader () {
        return getDefaultLoaders ()[0];
    }
    
    /** Getter for shadow loader.
     */
    static DataLoader getShadowLoader () {
        return getSystemLoaders ()[0];
    }
    
    //
    // Default loaders
    //
    
    /** Loader for folders */
    /*nbif compat
    public
    nbelse*/
    private
    /*nbend*/
    static class FolderLoader extends DataLoader {
        static final long serialVersionUID =-8325525104047820255L;
        
        /* Representation class is DataFolder */
        public FolderLoader () {
            super (DataFolder.class);
        }
        
        protected void initialize () {
        }
        
        /** Get default actions.
        * @return array of default system actions or <CODE>null</CODE> if this loader does not have any
        *   actions
        */
        protected SystemAction[] defaultActions () {
            return new SystemAction[] {
                    SystemAction.get (org.openide.actions.OpenLocalExplorerAction.class),
                    SystemAction.get (org.openide.actions.FindAction.class),
                    SystemAction.get (org.openide.actions.FileSystemAction.class),
                    null,
                    SystemAction.get (org.openide.actions.CompileAction.class),
                    SystemAction.get (org.openide.actions.CompileAllAction.class),
                    null,
                    SystemAction.get (org.openide.actions.BuildAction.class),
                    SystemAction.get (org.openide.actions.BuildAllAction.class),
                    null,
                    SystemAction.get (org.openide.actions.CutAction.class),
                    SystemAction.get (org.openide.actions.CopyAction.class),
                    SystemAction.get (org.openide.actions.PasteAction.class),
                    null,
                    SystemAction.get (org.openide.actions.DeleteAction.class),
                    SystemAction.get (org.openide.actions.RenameAction.class),
                    null,
                    // JST: there should be a template Package (DataFolder) instead of this action
                    SystemAction.get (org.openide.actions.NewAction.class),
                    SystemAction.get (org.openide.actions.NewTemplateAction.class),
                    null,
                    SystemAction.get (org.openide.actions.ToolsAction.class),
                    SystemAction.get (org.openide.actions.PropertiesAction.class)
                };
        }
        
        /** Get the default display name of this loader.
        * @return default display name
        */
        protected String defaultDisplayName () {
            return NbBundle.getBundle (DataLoaderPool.class).getString ("LBL_folder_loader_display_name");
        }
        
        protected DataObject handleFindDataObject (
        FileObject fo, DataLoader.RecognizedFiles recognized
        ) throws IOException {
            if (fo.isFolder ()) {
                return new DataFolder (fo);
            } else {
                return null;
            }
        }
    }
    
    /* Instance loader to recognize instances.
     */
    /*nbif compat
    public
    nbelse*/
    private
    /*nbend*/
    static class InstanceLoader extends UniFileLoader {
        static final long serialVersionUID =-3462727693843631328L;
        
        /* Creates new InstanceLoader */
        public InstanceLoader() {
            super(InstanceDataObject.class);
        }
        
        protected void initialize () {
            
            ExtensionList ext = new ExtensionList();
            ext.addExtension ("instance"); // NOI18N
            ext.addExtension ("ser"); // NOI18N
            setExtensions(ext);
            
        }
        
        /** Get default actions.
        * @return array of default system actions or <CODE>null</CODE> if this loader does not have any
        *   actions
        */
        protected SystemAction[] defaultActions () {
            return new SystemAction[] {
                SystemAction.get (org.openide.actions.CustomizeBeanAction.class),
                null,
                SystemAction.get (org.openide.actions.FileSystemAction.class),
                null,
                SystemAction.get(org.openide.actions.CutAction.class),
                SystemAction.get(org.openide.actions.CopyAction.class),
                SystemAction.get(org.openide.actions.PasteAction.class),
                null,
                SystemAction.get(org.openide.actions.DeleteAction.class),
                //      SystemAction.get(RenameAction.class),
                null,
                SystemAction.get (org.openide.actions.ToolsAction.class),
                SystemAction.get(org.openide.actions.PropertiesAction.class)
            };
        }
        
        /** Get the default display name of this loader.
        * @return default display name
        */
        protected String defaultDisplayName () {
            return NbBundle.getBundle (DataLoaderPool.class).getString ("LBL_instance_loader_display_name");
        }

        /* Creates the right data object for given primary file.
         * It is guaranteed that the provided file is realy primary file
         * returned from the method findPrimaryFile.
         *
         * @param primaryFile the primary file
         * @return the data object for this file
         * @exception DataObjectExistsException if the primary file already has data object
         */
        protected MultiDataObject createMultiObject (FileObject primaryFile)
        throws DataObjectExistsException, java.io.IOException {
            InstanceDataObject obj = new InstanceDataObject(primaryFile, this);
            return obj;
        }
        
        public void writeExternal (ObjectOutput oo) throws IOException {
            // does not use super serialization of extensions
            oo.writeObject (this);
            
            super.writeExternal (oo);
        }
        
        public void readExternal (ObjectInput oi) throws IOException, ClassNotFoundException {
            // the result of following code is either ExtensionList (original version)
            // or this (current version).
            Object o = oi.readObject ();
            if (o instanceof SystemAction[]) {
                //added for compatibility with FFJ2.0
                setActions ((SystemAction[]) o);            
            } else if (o instanceof ExtensionList) {
                // old serialization, add new extension
                ExtensionList list = (ExtensionList)o;
                list.addExtension ("ser"); // NOI18N
                setExtensions (list);
            } else {
                // newer serialization, everything should be ok, just read
                // the original value
                super.readExternal (oi);
            }
        }
    }
    
    
    /** Loader for file objsects not recognized by any other loader */
    /*nbif compat
    public
    nbelse*/
    private
    /*nbend*/
    static class DefaultLoader extends MultiFileLoader {
        static final long serialVersionUID =-6761887227412396555L;
        
        /* Representation class is DataFolder */
        public DefaultLoader () {
            super (DefaultDataObject.class);
        }
        
        protected void initialize () {            
        }
        
        /** Get default actions.
        * @return array of default system actions or <CODE>null</CODE> if this loader does not have any
        *   actions
        */
        protected SystemAction[] defaultActions () {
            return new SystemAction[] {
                    SystemAction.get (org.openide.actions.FileSystemAction.class),
                    null,
                    SystemAction.get (org.openide.actions.CutAction.class),
                    SystemAction.get (org.openide.actions.CopyAction.class),
                    SystemAction.get (org.openide.actions.PasteAction.class),
                    null,
                    SystemAction.get (org.openide.actions.DeleteAction.class),
                    SystemAction.get (org.openide.actions.RenameAction.class),
                    null,
                    SystemAction.get (org.openide.actions.ToolsAction.class),
                    SystemAction.get (org.openide.actions.PropertiesAction.class)
                };
        }
        
        /** Get the default display name of this loader.
        * @return default display name
        */
        protected String defaultDisplayName () {
            return NbBundle.getBundle (DataLoaderPool.class).getString ("LBL_default_loader_display_name");
        }
        
        /** Get the primary file.
         * @param fo the file to find the primary file for
         *
         * @return the primary file
         */
        protected FileObject findPrimaryFile (FileObject fo) {
            // never recognize folders
            if (fo.isFolder()) return null;
            return fo;
        }
        
        /* Creates the right data object for given primary file.
         * It is guaranteed that the provided file is realy primary file
         * returned from the method findPrimaryFile.
         *
         * @param primaryFile the primary file
         * @return the data object for this file
         * @exception DataObjectExistsException if the primary file already has data object
         */
        protected MultiDataObject createMultiObject (FileObject primaryFile)
        throws DataObjectExistsException, java.io.IOException {
            return new DefaultDataObject(primaryFile, this);
        }
        
        /* Creates the right primary entry for given primary file.
         *
         * @param obj requesting object
         * @param primaryFile primary file recognized by this loader
         * @return primary entry for that file
         */
        protected MultiDataObject.Entry createPrimaryEntry (MultiDataObject obj, FileObject primaryFile) {
            return new FileEntry (obj, primaryFile);
        }
        
        /** Do not create a seconday entry.
         *
         * @param obj ignored
         * @param secondaryFile ignored
         * @return never returns
         * @exception UnsupportedOperationException because this loader supports only a primary file object
         */
        protected MultiDataObject.Entry createSecondaryEntry (MultiDataObject obj, FileObject secondaryFile) {
            throw new UnsupportedOperationException ();
        }
        
        /** Does nothing because this loader works only with objects
         * with one file => primary file so it is not necessary to search
         * for anything else.
         *
         * @param obj the object to test
         */
        void checkFiles (MultiDataObject obj) {
        }
    }
    
    /** Loader for shadows */
    /*nbif compat
    public
    nbelse*/
    private
    /*nbend*/
    static class ShadowLoader extends DataLoader {
        static final long serialVersionUID =-11013405787959120L;
        
        /* Adapter for listening on filesystems changes */
        private static ShadowChangeAdapter changeAdapter = new ShadowChangeAdapter();
        
        /* Representation class is DataShadow */
        public ShadowLoader () {
            super (DataShadow.class);
        }
        
        protected void initialize () {            
        }
        
        /** Get the default display name of this loader.
        * @return default display name
        */
        protected String defaultDisplayName () {
            return NbBundle.getBundle (DataLoaderPool.class).getString ("LBL_shadow_loader_display_name");
        }
        
        protected DataObject handleFindDataObject (
        FileObject fo, DataLoader.RecognizedFiles recognized
        ) throws IOException {
            if (fo.hasExt (DataShadow.SHADOW_EXTENSION)) {
                try {
                    DataObject d = DataShadow.deserialize (fo);
                    if (d != null) return new DataShadow (fo, d, this);
                } catch (IOException ex) {
                    /* Link is broken, create BrokenDataShadow */
                    return new BrokenDataShadow (fo, this);
                }
            }
            return null;
        }
    }
    
    
}
