/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.io.IOException;
import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.filesystems.*;
import org.openide.util.RequestProcessor;
import java.lang.ref.*;

/** Registraction list of all data objects in the system.
* Maps data objects to its handlers.
*
* @author Jaroslav Tulach
*/
final class DataObjectPool extends Object
    implements javax.swing.event.ChangeListener, RepositoryListener, PropertyChangeListener {
    /** hashtable that maps FileObject to DataObjectPool.Item */
    private HashMap map = new HashMap ();
    /** the pool for all objects */
    static final DataObjectPool POOL = new DataObjectPool ();
    static {
        org.openide.TopManager.getDefault ().getLoaderPool ().addChangeListener (
            POOL
        );
        org.openide.TopManager.getDefault ().getRepository().addRepositoryListener(
            POOL
        );
    }
    /** Collection of all objects that has been created but their
    * creation has not been yet notified to OperationListener.postCreate
    * method.
    *
    * Set<Item>
    */
    private Collection toNotify;


    /** validator */
    private static final Validator VALIDATOR = new Validator ();

    /** Checks whether there is a data object with primary file
    * passed thru the parameter.
    *
    * @param fo the file to check
    * @return data object with fo as primary file or null
    */
    public DataObject find (FileObject fo) {
        if (toNotify != null) {
            return null;
        }
        synchronized (this) {
            Item doh = (Item)map.get (fo);
            return doh == null ? null : doh.getDataObjectOrNull ();
        }
    }
    
    
    /** Calls handle find on data loader. The method is here just because
     * historical reasons (hopefully). It used to synchronize itself
     * on this instance, but it could call very strange code and it was
     * hard to prevent some strange deadlocks. That is why now it is not
     * synchronized but instead the find method returns null if toNotify
     * is not empty.
     *
     * @param loader the loader to invoke the action on
     * @param fo file object to find data object for
     * @param rec object to notify recognized files to
     * @return data object or null
     */
    public DataObject callHandleFind (
        DataLoader loader,
        FileObject fo,
        DataLoader.RecognizedFiles rec
    ) throws IOException {
        try {
            return loader.handleFindDataObject(fo, rec);
        } catch (IOException ioe) {
            throw ioe;
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable t) {
            // Some strange error, perhaps an unexpected exception in
            // MultiFileLoader.findPrimaryFile. Such an error ought
            // not cause whole folder recognizer to die! Assume that
            // file/loader is kaput and continue.
            IOException ioe = new IOException ();
            TopManager.getDefault ().getErrorManager ().annotate (ioe, t);
            throw ioe;
        }
    }

    /** Refresh of all folders.
    */
    private void refreshAllFolders () {
        Set files;
        synchronized (this) {
            files = new HashSet (map.keySet ());
        }

        Iterator it = files.iterator ();
        while (it.hasNext ()) {
            FileObject fo = (FileObject)it.next ();
            if (fo.isFolder ()) {
                DataObject obj = find (fo);
                if (obj instanceof DataFolder) {
                    DataFolder df = (DataFolder)obj;
                    if ( toNotify == null || !toNotify.contains((Item)map.get(df.getPrimaryFile())))
                        df.refresh ();                    
                }
            }
        }
    }

    /** Rescans all fileobjects in given set.
    * @param s set of FileObjects
    * @return set of DataObjects that refused to be revalidated
    */
    public Set revalidate (Set s) {
        return VALIDATOR.revalidate (s);
    }

    /** Rescan all primary files of currently existing data
    * objects.
    *
    * @return set of DataObjects that refused to be revalidated
    */
    public Set revalidate () {
        Set files;
        synchronized (this) {
            files = createSetOfAllFiles (map.values ());
        }

        return revalidate (files);
    }

    /** Notifies all lastly created objects.
    */
    public void notifyCreation () {
        Iterator it;

        synchronized (this) {
            if (toNotify == null) {
                return;
            }
            it = toNotify.iterator();
            toNotify = null;
        }

        DataLoaderPool pool =  TopManager.getDefault ().getLoaderPool ();

        while (it.hasNext()) {
            Item item = (Item)it.next();
            DataObject obj = item.getDataObjectOrNull ();
            if (obj != null) {
                pool.fireOperationEvent (
                    new OperationEvent (obj), OperationEvent.CREATE
                );
            }
        }
    }

    /** Registers new DataObject instance.
    * @param fo primary file for obj
    * @param loader the loader of the object to be created
    *
    * @return object with common information for this <CODE>DataObject</CODE>
    * @exception DataObjectExistsException if the file object is already registered
    */
    public synchronized Item register (FileObject fo, DataLoader loader)
    throws DataObjectExistsException {
        Item doh = (Item)map.get (fo);
        // if Item for this file has not been created yet
        if (doh == null) {
            doh = new Item (fo);
            map.put (fo, doh);
            if (toNotify == null) {
                toNotify = new LinkedList ();
            }
            toNotify.add (doh);

            VALIDATOR.notifyRegistered (fo);

            return doh;
        } else {
            DataObject obj = doh.getDataObjectOrNull ();

            if (obj == null || VALIDATOR.reregister (obj, loader)) {
                // the item is to be finalize => create new

                doh = new Item (fo);
                map.put (fo, doh);
                if (toNotify == null) {
                    toNotify = new LinkedList ();
                }
                toNotify.add (doh);

                return doh;
            }


            // PENDING only for debug
            //      System.err.println ("Original stack (" + doh.getDataObject () + "): ");
            //      System.err.println (doh.toString ());
            //      Thread.dumpStack();
            // throw exception with the existing data object
            throw new DataObjectExistsException (obj);
        }
    }

    /** Notifies all newly created objects to

    /** Deregister.
    * @param item the item with common information to deregister
    * @param refresh true if the parent folder should be refreshed
    */
    private synchronized void deregister (Item item, boolean refresh) {
        FileObject fo = item.primaryFile;

        Item previous = (Item)map.remove (fo);

        if (previous != null && previous != item) {
            // ops, mistake,
            // return back the original
            map.put (fo, previous);
            return;
        }

        // refresh of parent folder
        if (refresh) {
            fo = fo.getParent ();
            if (fo != null) {
                Item item2 = (Item)map.get (fo);
                if (item2 != null) {
                    DataFolder df = (DataFolder) item2.getDataObjectOrNull();
                    if (df != null) {
                        VALIDATOR.refreshFolderOf (df);
                    }
                }
            }
        }
    }

    /** Changes the primary file to new one.
    * @param item the item to change
    * @param newFile new primary file to set
    */
    private synchronized void changePrimaryFile (
        Item item, FileObject newFile
    ) {
        map.remove (item.primaryFile);
        item.primaryFile = newFile;
        map.put (newFile, item);
    }

    /** When the loader pool is changed, then all objects are rescanned.
    */
    public void stateChanged (javax.swing.event.ChangeEvent ev) {
        Set set;
        synchronized (this) {
            // copy the values synchronously
            set = new HashSet (map.values ());
        }
        set = createSetOfAllFiles (set);
        revalidate (set);
    }
    
    /** Create list of all files for given collection of data objects.
    * @param c collection of DataObjectPool.Item
    * @return set of files
    */
    private static Set createSetOfAllFiles (Collection c) {
        HashSet set = new HashSet (c.size () * 7);
        
        Iterator it = c.iterator();
        while (it.hasNext()) {
            Item item = (Item)it.next ();
            DataObject obj = item.getDataObjectOrNull ();
            if (obj != null) {
                set.addAll (obj.files ());                
            }
        }
        return set;
    }
    
    /** Remove DataObjects which became invalid thanks
     * to unmounting a FileSystem
     */
    private void removeInvalidObjects() {
        Set files;
        synchronized (this) {
            files = new HashSet (map.values ());
        }
        files = createSetOfAllFiles (files);
        
        VALIDATOR.removeInvalidObject(files);
    }

    //
    // Repository listener changes
    //
    /** Called when new file system is added to the pool.
     * @param ev event describing the action
     */
    public void fileSystemAdded(RepositoryEvent ev) {
        ev.getFileSystem().addPropertyChangeListener( POOL );
    }
    
    /** Called when a file system is removed from the pool.
     * @param ev event describing the action
     */
    public void fileSystemRemoved(RepositoryEvent ev) {
        ev.getFileSystem().removePropertyChangeListener( POOL );
        removeInvalidObjects();
    }

    /** Called when a file system pool is reordered. */
    public void fileSystemPoolReordered(RepositoryReorderedEvent ev) {
    }

    /** Called when a file system property changed.
     * If it's property root, check validity.
     * @param ev event describing the action     
    */
    public void propertyChange (final PropertyChangeEvent ev) {
        if (FileSystem.PROP_SYSTEM_NAME.equals (ev.getPropertyName ())) {
            removeInvalidObjects();
        }        
        if (FileSystem.PROP_ROOT.equals (ev.getPropertyName ())) {
            removeInvalidObjects();
        }        
    }
    
    /** Returns all currently existing data
    * objects.
    *
    * @return iterator of DataObjectPool.Item
    */    
    Iterator getActiveDataObjects () {
        synchronized (this) {
            ArrayList alist = new ArrayList(map.values());
            return alist.iterator();
        }
    }

    /** One item in object pool.
    */
    static final class Item extends Object implements Runnable {
        /** primary file */
        FileObject primaryFile;
        /** weak reference data object with this primary file */
        private Reference obj;
        
        // [PENDING] hack to check the stack when the DataObject has been created
        //    private Exception stack;

        /** @param fo primary file
        * @param pool object pool
        */
        public Item (FileObject fo) {
            this.primaryFile = fo;

            // [PENDING] // stores stack
            /*      java.io.StringWriter sw = new java.io.StringWriter ();
                  stack = new Exception ();
                }

                // [PENDING] toString returns original stack
                public String toString () {
                  return stack.toString ();*/
        }

        /** Setter for the data object. Called immediatelly as possible.
        * @param obj the data object for this item
        */
        public void setDataObject (DataObject obj) {
            this.obj = new ItemReference (obj, this);
            
            if (obj != null && !obj.getPrimaryFile ().isValid ()) {
                // if the primary file is already invalid =>
                // mark the object as invalid
                deregister (false);
            }
        }

        /** Getter for the data object.
        * @return the data object or null
        */
        DataObject getDataObjectOrNull () {
            return this.obj == null ? null : (DataObject)this.obj.get ();
        }
        /** Getter for the data object.
        * @return the data object
        * @exception IllegalStateException if the data object has been lost
        *   due to weak references (should not happen)
        */
        public DataObject getDataObject () {
            DataObject obj = getDataObjectOrNull ();
            if (obj == null) {
                throw new IllegalStateException ();
            }
            return obj;
        }

        /** Deregister one reference.
        * @param refresh true if the parent folder should be refreshed
        */
        public void deregister (boolean refresh) {
            POOL.deregister (this, refresh);
        }

        /** Changes the primary file to new one.
        * @param newFile new primary file to set
        */
        public void changePrimaryFile (FileObject newFile) {
            POOL.changePrimaryFile (this, newFile);
        }

        /** Is the item valid?
        */
        public boolean isValid () {
            if (POOL.map.get (primaryFile) == this) {
                return true;
            } else {
                return false;
            }
            
        }
        
        /** Initialize the data object's node delegate.
         */
        public void run () {
            DataObject obj = getDataObjectOrNull ();
            if (obj != null) {
                synchronized (this) {
                    obj.doInitNodeDelegate ();
                }
            }
        }

        public String toString () {
            DataObject obj = (DataObject)this.obj.get ();
            if (obj == null) {
                return " nothing"; // NOI18N
            }
            return obj.toString ();
        }
    }

    /** WeakReference - references a DataObject, strongly references an Item */
    static final class ItemReference extends WeakReference {
        /** ReferenceQueue of all References of DataObjects */
        private static ReferenceQueue refqueue;
        
        /** Reference to an Item */
        private Item item;
        
        /** Getter for the refqueue */
        private static ReferenceQueue getRefqueue() {
            if (refqueue == null) {
                try {
                    Object sync = Class.forName("org.openide.loaders.DataObjectPool$ItemReference");
                    synchronized (sync) {
                        if (refqueue == null) {
                            refqueue = new ReferenceQueue();
                            ReferenceHandler.start();
                        }
                    }
                } catch (ClassNotFoundException e) {
                    throw new Error();
                }
            }
            return refqueue;
        }
        
        /** Deregisters Items of discarded DataObjects */
        static void checkQueue() {
            if (refqueue == null) {
                return;
            }
            
            ItemReference ref = (ItemReference) refqueue.poll();
            while (ref != null) {
                ref.deregister();
                ref = (ItemReference) refqueue.poll();
            }
        }
        
        ItemReference(DataObject dobject, Item item) {
            super(dobject, getRefqueue());
            this.item = item;
            checkQueue();
        }
        
        private void deregister() {
            item.deregister(false);
            item = null;
        }
    }
    
    /** Cleans up ItemReferences every 5000 ms */
    static final class ReferenceHandler implements Runnable {
        static void start() {
            RequestProcessor.postRequest(new ReferenceHandler(), 5000);
        }
        
        private void restart() {
            RequestProcessor.postRequest(this, 5000);
        }
        
        /** Cleans up all references */
        public void run() {
            ItemReference.checkQueue();
            restart();
        }
    }

    /** Validator to allow rescan of files.
    */
    private static final class Validator extends Object
        implements DataLoader.RecognizedFiles {
        /** set of all files that should be revalidated (FileObject) */
        private Set files;
        /** current thread that is in the validator */
        private Thread current;
        /** number of threads waiting to enter the validation */
        private int waiters;
        /** set of files that has been marked recognized (FileObject) */
        private HashSet recognizedFiles;
        /** set with all objects that refused to be discarded (DataObject) */
        private HashSet refusingObjects;
        /** set of files that has been registered during revalidation */
        private HashSet createdFiles;

        /** Enters the section.
        * @param set set of files that should be processed
        * @return the set of files concatenated with any previous sets
        */
        private synchronized Set enter (Set set) {
            waiters++;
            while (current != null) {
                try {
                    wait ();
                } catch (InterruptedException ex) {
                }
            }
            current = Thread.currentThread ();
            waiters--;

            if (files == null) {
                files = set;
            } else {
                files.addAll (set);
            }

            return files;
        }

        /** Leaves the critical section.
        */
        private synchronized void exit () {
            current = null;
            if (waiters == 0) {
                files = null;
            }
            notify ();
        }

        /** If there is another waiting thread, then I can
        * cancel my computation.
        */
        private synchronized boolean goOn () {
            return waiters == 0;
        }

        /** Called to either refresh folder, or register the folder to be
        * refreshed later is validation is in progress.
        */
        public void refreshFolderOf (DataFolder df) {
            if (createdFiles == null) {
                // no validator in progress
                df.refresh ();
            }
        }

        /** Mark this file as being recognized. It will be excluded
        * from further processing.
        *
        * @param fo file object to exclude
        */
        public void markRecognized (FileObject fo) {
            recognizedFiles.add (fo);
        }

        public void notifyRegistered (FileObject fo) {
            if (createdFiles != null) {
                createdFiles.add (fo);
            }
        }

        /** Reregister new object for already existing file object.
        * @param obj old object existing
        * @param loader loader of new object to create
        * @return true if the old object has been discarded and new one can
        *    be created
        */
        public boolean reregister (DataObject obj, DataLoader loader) {
            if (recognizedFiles == null) {
                // revalidation not in progress
                return false;
            }

            if (obj.getLoader () == loader) {
                // no change in loader =>
                return false;
            }

            if (createdFiles.contains (obj.getPrimaryFile ())) {
                // if the file already has been created
                return false;
            }

            if (refusingObjects.contains (obj)) {
                // the object has been refused before
                return false;
            }

            try {
                obj.setValid (false);
                return true;
            } catch (java.beans.PropertyVetoException ex) {
                refusingObjects.add (obj);
                return false;
            }

        }

        /** Rescans all fileobjects in given set.
        * @param s set of FileObjects
        * @return set of objects that refused to be revalidated
        */
        public Set revalidate (Set s) {
            // holds all created object, so they are not garbage
            // collected till this method ends
            LinkedList createObjects = new LinkedList ();
            try {
                s = enter (s);

                recognizedFiles = new HashSet ();
                refusingObjects = new HashSet ();
                createdFiles = new HashSet ();

                HashSet allFS = new HashSet (java.util.Arrays.asList(
                                                 org.openide.TopManager.getDefault().getRepository().toArray()
                                             ));

                DataLoaderPool pool = org.openide.TopManager.getDefault ().getLoaderPool ();
                Iterator it = s.iterator ();
                while (it.hasNext () && goOn ()) {
                    FileObject fo = (FileObject)it.next ();
                    if (!recognizedFiles.contains (fo)) {
                        // first of all test if the file is on a valid filesystem
                        boolean invalidate;
                        try {
                            FileSystem fs = fo.getFileSystem ();
                            invalidate = !allFS.contains (fs);
                        } catch (FileStateInvalidException ex) {
                            invalidate = true;
                        }

                        try {
                            // the previous data object should be canceled
                            DataObject orig = POOL.find (fo);
                            if (orig == null) {
                                // go on
                                continue;
                            }

                            if (!invalidate) {
                                // findDataObject
                                // is not using method DataObjectPool.find to locate data object
                                // directly for primary file, that is good
                                DataObject obj = pool.findDataObject (fo, this);
                                createObjects.add (obj);

                                invalidate = obj != orig;
                            }


                            if (invalidate) {
                                try {
                                    orig.setValid (false);
                                } catch (java.beans.PropertyVetoException ex) {
                                    refusingObjects.add (orig);
                                }
                            }
                        } catch (java.io.IOException ex) {
                        }

                    }

                }
                return refusingObjects;
            } finally {
                recognizedFiles = null;
                refusingObjects = null;
                createdFiles = null;

                exit ();

                if ( s.size() > 1 )
                    POOL.refreshAllFolders ();
            }
        }
        
        /** Remove DataObjects which became invalid thanks
         * to unmounting a FileSystem
         */
        void removeInvalidObject(Set files) {
            try {
                files = enter(files);
                HashSet allFS = new HashSet();
                FileSystem[] fss = org.openide.TopManager.getDefault().getRepository().toArray();
                for (int i = 0; i < fss.length; i++) {
                    allFS.add(fss[i]);
                }
                
                Iterator it = files.iterator();
                while (it.hasNext() && goOn()) {
                    FileObject fo = (FileObject) it.next();
                    
                    boolean invalidate = !fo.isValid();
                    if ( !invalidate ) {
                        try {
                            FileSystem fs = fo.getFileSystem ();
                            invalidate = !allFS.contains (fs);
                        } catch (FileStateInvalidException ex) {
                            invalidate = true;
                        }
                    }
                    
                    // the previous data object should be canceled
                    DataObject orig = POOL.find(fo);
                    if (orig == null) {
                        // go on
                        continue;
                    }
                    
                    if (invalidate) {
                        try {
                            orig.setValid (false);
                        } catch (java.beans.PropertyVetoException ex) {
                            // silently ignore?
                            TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
                        }
                    }
                }
            } finally {
                exit();
            }
        }


    } // end of Validator
}
