/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.io.*;
import java.util.HashSet;

import org.openide.filesystems.*;
import org.openide.util.HelpCtx;
import org.openide.nodes.Node;

/** An implementation of a data object which consumes file objects not recognized by any other loaders.
*
* @author Ian Formanek
*/
class DefaultDataObject extends MultiDataObject {
    static final long serialVersionUID =-4936309935667095746L;
    /** generated Serialized Version UID */
    //  static final long serialVersionUID = 6305590675982925167L;

    /** Constructs new data shadow for given primary file and referenced original.
    * @param fo the primary file
    * @param original original data object
    */
    DefaultDataObject (FileObject fo, MultiFileLoader loader) throws DataObjectExistsException {
        super (fo, loader);
    }

    /* Creates node delegate.
    */
    protected Node createNodeDelegate () {
        return new DataNode (this, org.openide.nodes.Children.LEAF);
    }

    /** Get the name of the data object.
    * <p>The implementation uses the name of the primary file and its exten.
    * @return the name
    */
    
    public String getName() {
        return getPrimaryFile ().getNameExt ();
    }

    /* Help context for this object.
    * @return help context
    */
    public HelpCtx getHelpCtx () {
        return new HelpCtx (DefaultDataObject.class);
    }

    
    /* Handles renaming of the object.
    * Must be overriden in children.
    *
    * @param name name to rename the object to
    * @return new primary file of the object
    * @exception IOException if an error occures
    */
    protected FileObject handleRename (String name) throws IOException {
        FileLock lock = getPrimaryFile ().lock ();
        int pos = name.lastIndexOf('.');
        
        try {
            if (pos < 0){
                // file without separator
                getPrimaryFile ().rename (lock, name, null);
            } else if (pos == 0){
                getPrimaryFile ().rename (lock, name, getPrimaryFile ().getExt ());
            } else {
                if (!name.equals(getPrimaryFile ().getNameExt())){
                    getPrimaryFile ().rename (lock, name.substring(0, pos), 
                        name.substring(pos+1, name.length()));
                    DataObjectPool.POOL.revalidate(
                        java.util.Collections.singleton(getPrimaryFile ())
                    );
                }
            }
        } finally {
            lock.releaseLock ();
        }
        return getPrimaryFile ();
    }
}

/*
 * Log
 *  11   Gandalf   1.10        1/15/00  Jaroslav Tulach annotation applied only 
 *       once.
 *  10   Gandalf   1.9         1/12/00  Ian Formanek    NOI18N
 *  9    Gandalf   1.8         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun
 *       Microsysteims Copyright in File Comment
 *  8    Gandalf   1.7         10/7/99  Jesse Glick     Create-from-template 
 *       needs to clear template status on result.
 *  7    Gandalf   1.6         9/3/99   Jaroslav Tulach #3649
 *  6    Gandalf   1.5         8/18/99  Ian Formanek    Generated serial version
 *       UID
 *  5    Gandalf   1.4         7/21/99  Ian Formanek    Fixed display name with 
 *       empty extension
 *  4    Gandalf   1.3         6/24/99  Jesse Glick     Gosh-honest HelpID's.
 *  3    Gandalf   1.2         6/8/99   Ian Formanek    ---- Package Change To 
 *       org.openide ----
 *  2    Gandalf   1.1         5/3/99   Ian Formanek    Class made 
 *       package-private
 *  1    Gandalf   1.0         5/2/99   Ian Formanek    
 * $
 */
