/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.beans.*;
import java.awt.event.*;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import javax.swing.JDialog;

import org.openide.*;
import org.openide.compiler.*;
import org.openide.compiler.Compiler;
import org.openide.debugger.*;
import org.openide.execution.*;
import org.openide.explorer.propertysheet.*;
import org.openide.filesystems.*;
import org.openide.cookies.CompilerCookie;
import org.openide.cookies.ExecCookie;
import org.openide.cookies.ArgumentsCookie;
import org.openide.cookies.DebuggerCookie;
import org.openide.nodes.Sheet;
import org.openide.nodes.PropertySupport;
import org.openide.util.NbBundle;
import org.openide.util.Mutex;
import org.openide.util.Utilities;

/** Support for execution of a class file. Looks for the class with
* the same base name as the primary file, locates a main method
* in it, and starts it.
*
* @author Jaroslav Tulach
*/
public class ExecSupport extends Object
    implements ExecCookie, ArgumentsCookie, DebuggerCookie {
    /** extended attribute for the type of executor */
    private static final String EA_EXECUTOR = "NetBeansAttrExecutor"; // NOI18N
    /** extended attribute for attributes */
    private static final String EA_ARGUMENTS = "NetBeansAttrArguments"; // NOI18N
    /** extended attribute for debugger type */
    private static final String EA_DEBUGGER_TYPE = "NetBeansAttrDebuggerType"; // NOI18N

    // copy from JavaNode
    /** Name of property providing argument parameter list. */
    public static final String PROP_FILE_PARAMS   = "params"; // NOI18N
    /** Name of property providing a custom {@link Executor} for a file. */
    public static final String PROP_EXECUTION     = "execution"; // NOI18N
    /** Name of property providing a custom {@link DebuggerType} for a file. */
    public static final String PROP_DEBUGGER_TYPE     = "debuggerType"; // NOI18N

    /** bundle */
    static ResourceBundle bundle;

    /** entry to be associated with */
    protected MultiDataObject.Entry entry;
    
    /** If the filesystem is readonly executor and debugger properties will
     * be read only. */
    private boolean isReadOnly = false;

    /** Create new support for given entry. The file is taken from the
    * entry and is updated if the entry moves or renames itself.
    * @param entry entry to create instance from
    */
    public ExecSupport (MultiDataObject.Entry entry) {
        this.entry = entry;
        isReadOnly = entry.getFile().isReadOnly(); // used to be getFileSystem()....
    }

    /* Starts the class.
    */
    public void start () {
        Executor exec = getExecutor (entry);
        if (exec == null) {
            exec = defaultExecutor ();
        }

        try {
            exec.execute(entry.getDataObject());
        } catch (final IOException ex) {
            Mutex.EVENT.readAccess (new Runnable () {
                                        public void run () {
                                            if (startFailed (ex)) {
                                                // restart
                                                ExecSupport.this.start ();
                                            }
                                        }
                                    });
        }
    }

    /* Start debugging of associated object.
    * @param stopOnMain if <code>true</code>, debugger stops on the first line of debugged code
    * @exception DebuggerException if the session cannot be started
    */
    public void debug (final boolean stopOnMain) throws DebuggerException {
        DebuggerType t = getDebuggerType (entry);
        if (t == null) {
            t = defaultDebuggerType ();
        }

        try {
            t.startDebugger (entry.getDataObject(), stopOnMain);
            // ok, debugger started
            return;
        } catch (final DebuggerException ex) {
            try {
                Mutex.EVENT.readAccess (new Mutex.ExceptionAction () {
                                            public Object run () throws DebuggerException {
                                                if (debugFailed (ex)) {
                                                    // restart
                                                    debug (stopOnMain);
                                                }
                                                return null;
                                            }
                                        });
            } catch (org.openide.util.MutexException mx) {
                throw (DebuggerException)mx.getException ();
            }
        }
    }

    /** Called when invocation of the executor fails. Allows to do some
    * modifications to the type of execution and try it again.
    *
    * @param ex exeception that occured during execution
    * @return true if the execution should be restarted
    */
    protected boolean startFailed (IOException ex) {
        Executor e = (Executor)choose (getExecutor (entry), Executor.class, ex);
        if (e == null) {
            return false;
        } else {
            try {
                setExecutor (entry, e);
                return true;
            } catch (IOException exc) {
                return false;
            }
        }
    }

    /** Called when invocation of the debugger fails. Allows to do some
    * modifications to the type of debugging and try it again.
    *
    * @param ex exeception that occured during execution
    * @return true if the debugging should be started again
    */
    protected boolean debugFailed (DebuggerException ex) {
        DebuggerType e = (DebuggerType)choose (getDebuggerType (entry), DebuggerType.class, ex);
        if (e == null) {
            return false;
        } else {
            try {
                setDebuggerType (entry, e);
                return true;
            } catch (IOException exc) {
                return false;
            }
        }
    }

    /** Check if this object is up to date or in need of compilation.
    * Should compile it if necessary.
    * <p>The default implementation checks whether {@link CompilerCookie} is provided and
    * if so, creates a job and compiles the object. This behavior may be 
    * overridden by subclasses.
    *
    * @return <code>true</code> if the object was successfully brought up to date, <code>false</code> if the attempt failed (and it may be still be out of date)
    * @deprecated The check should be done in an action - ExecAction, ...
    */
    protected boolean checkCompiled () {
        DataObject obj = entry.getDataObject ();

        CompilerCookie c = (CompilerCookie)obj.getCookie (CompilerCookie.class);

        if (c != null) {
            CompilerJob job = new CompilerJob (Compiler.DEPTH_ZERO);
            job.setDisplayName (obj.getName ());
            c.addToJob (job, Compiler.DEPTH_ZERO);
            if (!job.isUpToDate ()) {
                // add name
                // compile it
                CompilerTask t = job.start ();
                return t.isSuccessful ();
            }
        }
        return true;
    }

    /** This method allows subclasses to override the default
    * debugger type they want to use for debugging.
    *
    * @return current implementation returns DebuggerType.getDefault ()
    */
    protected DebuggerType defaultDebuggerType () {
        return DebuggerType.getDefault ();
    }

    /** This method allows subclasses to override the default
    * executor they want to use for debugging.
    *
    * @return current implementation returns Executor.getDefault ()
    */
    protected Executor defaultExecutor () {
        return Executor.getDefault ();
    }


    /** Set the executor for a given file object.
     * Uses file attributes to store this information.
    * @param entry entry to set the executor for
    * @param exec executor to use
    * @exception IOException if executor cannot be set
    */
    public static void setExecutor (MultiDataObject.Entry entry, Executor exec) throws IOException {
        entry.getFile ().setAttribute (EA_EXECUTOR,
                                       exec == null ? null : new ServiceType.Handle (exec)
                                      );
    }

    /** Get the executor for a given file object.
    * @param entry entry to obtain the executor for
    * @return executor associated with the file, or <code>null</code> if the default should be used
    */
    public static Executor getExecutor (MultiDataObject.Entry entry) {
        Object handle = entry.getFile ().getAttribute (EA_EXECUTOR);
        if (handle != null && (handle instanceof ServiceType.Handle)) {
            ServiceType exec = ((ServiceType.Handle) handle).getServiceType ();
            if (exec instanceof Executor) {
                return (Executor)exec;
            }
        }
        return null;
    }

    /* Sets execution arguments for the associated entry.
    * @param args array of arguments
    * @exception IOException if arguments cannot be set
    */
    public void setArguments (String[] args) throws IOException {
        entry.getFile ().setAttribute (EA_ARGUMENTS, args);
    }

    /** Set execution arguments for a given entry.
    * @param entry the entry
    * @param args array of arguments
    * @exception IOException if arguments cannot be set
    */
    public static void setArguments (MultiDataObject.Entry entry, String[] args) throws IOException {
        entry.getFile ().setAttribute (EA_ARGUMENTS, args);
    }

    /* Getter for arguments associated with given file.
    * @return the arguments or empty array if no arguments associated
    */
    public String[] getArguments () {
        return getArguments (entry);
    }

    /** Get the arguments associated with a given entry.
    * @param entry the entry
    * @return the arguments, or an empty array if no arguments are specified
    */
    public static String[] getArguments(MultiDataObject.Entry entry) {
        Object o = entry.getFile ().getAttribute (EA_ARGUMENTS);
        if (o != null && (o instanceof String[])) {
            return (String[]) o;
        } else {
            return new String[] { };
        }
    }

    //
    // debugger support
    //

    /** Assignes a debugger type to an entry.
    * @param entry the object's entry
    * @param type the debugger type for this entry
    * @exception IOException if arguments cannot be set
    */
    public static void setDebuggerType (MultiDataObject.Entry entry, DebuggerType type) throws IOException {
        entry.getFile ().setAttribute (EA_DEBUGGER_TYPE,
                                       type == null ? null : new ServiceType.Handle (type)
                                      );
    }

    /** Retrieves the debugger type for this entry.
    * @param entry the entry
    * @return the debugger type or null if no type assigned
    */
    public static DebuggerType getDebuggerType (MultiDataObject.Entry entry) {
        Object handle = entry.getFile ().getAttribute (EA_DEBUGGER_TYPE);
        if (handle != null && (handle instanceof ServiceType.Handle)) {
            ServiceType dbg = ((ServiceType.Handle) handle).getServiceType ();
            if (dbg instanceof DebuggerType) {
                return (DebuggerType) dbg;
            }
        }
        return null;
    }

    /** Helper method that creates default properties for execution of
    * a given support.
    * Includes properties to set the executor; debugger; and arguments.
    *
    * @param set sheet set to add properties to
    */
    public void addProperties (Sheet.Set set) {
        set.put(createParamsProperty());
        set.put(createExecutorProperty ());
        set.put(createDebuggerProperty ());
    }

    /** Creates the fileparams property for entry.
    * @return the property
    */
    private PropertySupport createParamsProperty () {
        return new PropertySupport.ReadWrite (
                    PROP_FILE_PARAMS,
                    String.class,
                    getString("PROP_fileParams"),
                    getString("HINT_fileParams")
                ) {
                    public Object getValue() {
                        String[] args = getArguments ();
                        /*
                        StringBuffer b = new StringBuffer(50);
                        for (int i = 0; i < args.length; i++) {
                            b.append(args[i]).append(' ');
                        }
                        return b.toString();
                         */
                        return Utilities.escapeParameters(args);
                    }
                    public void setValue (Object val) throws InvocationTargetException {
                        if (val instanceof String) {
                            try {
                                setArguments(Utilities.parseParameters((String)val));
                            } catch(IOException e) {
                                throw new InvocationTargetException (e);
                            }
                        }
                        else {
                            throw new IllegalArgumentException();
                        }
                    }

                    public boolean supportsDefaultValue () {
                        return true;
                    }

                    public void restoreDefaultValue () throws InvocationTargetException {
                        try {
                            setArguments(null);
                        } catch(IOException e) {
                            throw new InvocationTargetException (e);
                        }
                    }
                    
                    public boolean canWrite () {
                       return !entry.getFile().isReadOnly();
                    }
                };
               
    }
    
    /** Creates the executor property for entry.
    * @return the property
    */
    private PropertySupport createExecutorProperty () {
        return new PropertySupport.ReadWrite (
                   PROP_EXECUTION,
                   Executor.class,
                   getString("PROP_execution"),
                   getString("HINT_execution")
               ) {
                   public Object getValue() {
                       Executor e = getExecutor (entry);
                       if (e == null)
                           return defaultExecutor ();
                       else
                           return e;
                   }
                   public void setValue (Object val) throws InvocationTargetException {
                       try {
                           setExecutor(entry, (Executor) val);
                       } catch (IOException ex) {
                           throw new InvocationTargetException (ex);
                       }
                   }
                   public boolean supportsDefaultValue () {
                       return true;
                   }

                   public void restoreDefaultValue () throws InvocationTargetException {
                       setValue (null);
                   }
                   
                   public boolean canWrite () {
                       return !entry.getFile().isReadOnly();
                   }
               };
    }

    /** Creates the debugger property for entry.
    * @return the property
    */
    private PropertySupport createDebuggerProperty () {
        return new PropertySupport.ReadWrite (
                   PROP_DEBUGGER_TYPE,
                   DebuggerType.class,
                   getString("PROP_debuggerType"),
                   getString("HINT_debuggerType")
               ) {
                   public Object getValue() {
                       DebuggerType dt = getDebuggerType (entry);
                       if (dt == null)
                           return defaultDebuggerType ();
                       else
                           return dt;
                   }
                   public void setValue (Object val) throws InvocationTargetException {
                       try {
                           setDebuggerType (entry, (DebuggerType) val);
                       } catch (IOException ex) {
                           throw new InvocationTargetException (ex);
                       }
                   }
                   public boolean supportsDefaultValue () {
                       return true;
                   }

                   public void restoreDefaultValue () throws InvocationTargetException {
                       setValue (null);
                   }
                   
                   public boolean canWrite () {
                       return !entry.getFile().isReadOnly();
                   }
               };
    }

    /** @return a localized String */
    static String getString(String s) {
        if (bundle == null) {
            bundle = NbBundle.getBundle(ExecSupport.class);
        }
        return bundle.getString(s);
    }

    /** Opens dialog and asks for different executor or debugger.
    * @param current current value
    * @param clazz the class to use to locate property editor
    * @param ex the exception that caused the problem
    * @return the new value or null if choosing failed
    */
    private static ServiceType choose (ServiceType current, Class clazz, final Exception ex) {
        TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);

        PropertyEditor ed = PropertyEditorManager.findEditor (clazz);
        if (ed == null) return null;

        ed.setValue (current);

        java.awt.Component c = ed.getCustomEditor ();
        if (c == null) return null;

        String configure = org.openide.util.NbBundle.getBundle(ExecSupport.class).getString("CTL_ServiceConfigure");
        NotifyDescriptor exc = new NotifyDescriptor.Exception (ex);
        exc.setTitle (org.openide.util.NbBundle.getBundle(ExecSupport.class).getString("CTL_Service_Configuration_Title"));
        // [PENDING] this is not necessary a *useful* message: e.g. for a ClassNotFoundException: foo.Bar,
        // this shows: "foo.Bar   ... Configure? Cancel?" which is not very helpful.
        exc.setMessage (ex.getLocalizedMessage ());
        exc.setOptions (new Object[] { configure, NotifyDescriptor.CANCEL_OPTION });
        Object res = TopManager.getDefault ().notify (exc);
        if (!configure.equals (res)) {
            return null;
        }

        DialogDescriptor d = new DialogDescriptor (
                                 c,
                                 getString ("MSG_ConfigureService")
                             );
        d.setOptions (new Object[] {
                          DialogDescriptor.OK_OPTION,
                          DialogDescriptor.CANCEL_OPTION
                      });

        java.awt.Dialog dialog = TopManager.getDefault ().createDialog (d);

        dialog.show ();

        if (d.getValue () == NotifyDescriptor.OK_OPTION) {
            // get the current value
            return (ServiceType)ed.getValue ();
        } else {
            // canceled
            return null;
        }
    }
}
