/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.beans.*;
import java.util.*;
import javax.swing.SwingUtilities;

import org.openide.*;
import org.openide.filesystems.*;
import org.openide.util.Mutex;
import org.openide.util.WeakListener;
import org.openide.util.RequestProcessor;
import org.openide.util.Task;
import org.openide.util.TaskListener;
import org.openide.nodes.Children;
import org.openide.nodes.Node;

/** Watches over a folder and represents its
* child data objects by nodes.
*
* @author Jaroslav Tulach
*/
final class FolderChildren extends Children.Keys 
implements PropertyChangeListener {
    /** the folder */
    private DataFolder folder;
    /** filter of objects */
    private DataFilter filter;
    /** initialization of children task */
    private Task initTask;
    /** last task that refreshes the children nodes */
    private Task refreshTask;
    /** listener on changes in nodes */
    private PropertyChangeListener listener;

    /**
    * @param f folder to display content of
    * @param map map to use for holding of children
    */
    public FolderChildren (DataFolder f) {
        this (f, DataFilter.ALL);
    }

    /**
    * @param f folder to display content of
    * @param filter filter of objects
    */
    public FolderChildren (DataFolder f, DataFilter filter) {
        this.folder = f;
        this.filter = filter;
        this.listener = WeakListener.propertyChange (this, folder);
    }
    
    /** used from DataFolder */
    DataFilter getFilter () {
        return filter;
    }

    /** If the folder changed its children we change our nodes.
    */
    public void propertyChange (final PropertyChangeEvent ev) {
        if (DataFolder.PROP_CHILDREN.equals (ev.getPropertyName ())) {
            refreshChildren ();
            return;
        }
        if (
            DataFolder.PROP_SORT_MODE.equals (ev.getPropertyName ()) ||
            DataFolder.PROP_ORDER.equals (ev.getPropertyName ())
        ) {
            refreshChildren ();
            return;
        }
    }

    /** Refreshes the children.
    */
    void refreshChildren () {
        initialize (true, false);
    }

    /** Creates a key for given data object.
    * This method is here to create something different then data object,
    * because the data object should be finalized when not needed and
    * that is why it should not be used as a key.
    *
    * @param obj data object
    * @return key representing the data object.
    */
    static Object createKey (DataObject obj) {
        return new Pair (obj.getClass (), obj.getPrimaryFile ());
    }

    /** This method takes the key created by createKey and converts it
    * into primary file.
    *
    * @param key the key
    * @return primary file of the key
    */
    private static FileObject getFile (Object key) {
        return ((Pair)key).primaryFile;
    }

    /** Create a node for one data object.
    * @param key DataObject
    */
    protected Node[] createNodes (Object key) {
        FileObject fo = getFile (key);
        DataObject obj;
        try {
            obj = DataObject.find (fo);

            if (filter == null || filter.acceptDataObject (obj)) {
                return new Node[] { obj.getClonedNodeDelegate (filter) };
            } else {
                return new Node[0];
            }
        } catch (DataObjectNotFoundException e) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, e);
            return new Node[0];
        }
    }

    /** Improves the searching capability to wait till all children
    * are found and then searching thru them.
    */
    public Node findChild (final String name) {
        // start the initialization
        Node[] forget = getNodes ();
        
        // waits till the list of children is created
        
        // JST: the folowing code is replacement for
        // initialize (false, false).waitFinished ()
        // 
        // the orgiginal code caused deadlocks when called with readAccess,
        // because the refreshChildren (which we wait for to be finished)
        // method calls setKeys and they require write request
        //
        // this code could probably be replaced by 
        // Children.MUTEX.isReadAccess, if such call would be added to Mutex
        class Jst implements Runnable {
            public boolean inReadAccess = true;
            public void run () {
                inReadAccess = false;
            }
        }
        Jst t = new Jst ();
        // the code will run either immediatally or after we leave readAccess
        // section
        Children.MUTEX.postWriteRequest(t);
        
        if (!t.inReadAccess) {
            // if we are not in readAccess we can safely wait till
            // refreshChildren finishes and sets new keys for the 
            initialize (false, false).waitFinished ();
            refreshTask.waitFinished ();
        }
        
        // do the regular child lookup
        Node node = super.findChild (name);
        return node;
    }

    /** Initializes the children.
    */
    protected void addNotify () {
        initialize (true, true);
        // add as a listener for changes on nodes
        folder.addPropertyChangeListener (listener);
    }

    /** Deinitializes the children.
    */
    protected void removeNotify () {
        // removes the listener
        folder.removePropertyChangeListener (listener);
        setKeys (java.util.Collections.EMPTY_SET);
    }

    /** Ensures that the content of children will be filled sometime.
    * @param force true if the content should be filled immediatelly
    */
    private Task initialize (boolean force, boolean waitFirst) {
        Task t = initTask;
        if (t != null && !force) {
            return t;
        }

        final Addition add = new Addition (waitFirst);
        initTask = t = folder.computeChildrenList (add);
        t.addTaskListener (add);

        if (waitFirst) {
            add.waitFirst ();
        }
        
        refreshTask = add.refTask;

        return t;
    }

    /** Display name */
    public String toString () {
        return folder.getPrimaryFile ().toString ();
    }

    /** Javac 1.2 workaround */
    private final void setKeys0 (Collection c) {
        setKeys (c);
    }

    /** time delay between two inserts of nodes */
    private static final int TIME_DELAY = 256;

    /** Support for incremental adding of new nodes.
    *
    * <P>
    * There is a deadlock warning:
    * <OL>
    *   <LI>A thread waiting in the waitFirst method can have MUTEX.readAccess
    *   <LI>Thread running run () needs access to MUTEX.writeAccess (because of setKeys)
    *   <LI>Be sure that the thread leaves waitFirst before writeAccess is needed
    * </OL>
    */
    private class Addition extends Object
    implements TaskListener, FolderListListener, Runnable {
        static final long serialVersionUID =-4194617547214845940L;

        /** last time of addition */
        private long time = System.currentTimeMillis () + TIME_DELAY;
        /** delay */
        private int delay = TIME_DELAY;

        /** update the nodes during processing or only at the end */
        private boolean processingUpdate;
        
        /** a keys to be later refreshed */
        private List refKeys;
        
        /** a task that is used to request refreshing of keys */
        private RequestProcessor.Task refTask = RequestProcessor.createRequest (
            this
        );

        /** @param processingUpdate update the nodes during
        *  processing or only at the end 
        */
        public Addition (boolean processingUpdate) {
            this.processingUpdate = processingUpdate;
        }

        /** Another object has been recognized.
         * @param obj the object recognized
         * @param arr array where the implementation should add the
         *    object
         */
        public void process(DataObject obj, java.util.List arr) {
            if (!filter.acceptDataObject (obj)) {
                return;
            }

            // first accepted object is notified to the waiting thread in
            boolean first = arr.isEmpty ();
            arr.add (obj);

            if (!processingUpdate) {
                // if we should not notify during processing update
                // skip the rest
                return;
            }

            if (first) {
                synchronized (this) {
                    this.notify ();
                    refreshKeys (arr);
                }
                return;
            }

            if (System.currentTimeMillis () > time) {
                if (!arr.isEmpty ()) {
                    // add the nodes
                    synchronized (this) {
                        refreshKeys (arr);
                    }
                    delay *= 2;
                }

                time = System.currentTimeMillis () + delay;
            }
        }

        /** All objects has been recognized.
         * @param arr list of DataObjects
         */
        public void finished(java.util.List arr) {
            synchronized (this) {
                this.notify ();
                // change the order because initialize method has already finished
                refreshKeys (arr);
            }
        }

        /** Getter for first map.
        */
        public synchronized void waitFirst () {
            try {
                this.wait (50);
            } catch (InterruptedException e) {
                // cannot happen
                throw new InternalError ();
            }
        }

        /** Called when a task finishes running.
         * @param task the finished task
         */
        public void taskFinished(Task task) {
            initTask = Task.EMPTY;
        }
        
        /** Refreshes the children.
        * @param ch collection of children data objects
        */
        private void refreshKeys (List ch) {
            ListIterator it = ch.listIterator ();
            LinkedList l = new LinkedList ();

            while (it.hasNext ()) {
                DataObject obj = (DataObject)it.next ();
                l.add (createKey (obj));
            }

            // remember the keys
            refKeys = l;
            // request method run to be called from request processor thread
            refTask.schedule (0);
        }
        
        /** Refreshes the keys in safe thread (that nobody should wait for)
        */
        public void run() {
            List toSet;
            synchronized (this) { // 1
                toSet = refKeys;
                refKeys = null;
            }

            if (toSet != null) {
                // toSet can be null if 
                //   1. a task blocks on synchronized(this) // 1
                //   2. refreshKeys calls refKey = <newvalue> + refTask.schedule
                //   3. a task uses the new value, but is already planned =>
                //   4. it will be run next time with toSet == null
                setKeys0 (toSet);
            }
        }
        
    }

    /** Pair of dataobject class and primary file.
    */
    private static final class Pair extends Object {
        public Class clazz;
        public FileObject primaryFile;

        public Pair (Class clazz, FileObject primaryFile) {
            this.clazz = clazz;
            this.primaryFile = primaryFile;
        }

        public int hashCode () {
            return clazz.hashCode () + 2 * primaryFile.hashCode ();
        }

        public boolean equals (Object o) {
            if (o instanceof Pair) {
                Pair p = (Pair)o;
                return clazz == p.clazz && primaryFile.equals (p.primaryFile);
            }
            return false;
        }
    }
}
