/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;
import java.lang.ref.SoftReference;
import java.lang.ref.WeakReference;
import java.lang.ref.Reference;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.util.datatransfer.*;
import org.openide.filesystems.*;
import org.openide.util.*;
import org.openide.util.actions.CallableSystemAction;

/** Watches over a folder and recognizes its children.
*
* @author Jaroslav Tulach
*/
final class FolderList extends Object implements FileChangeListener {
    /** serial version UID */
    static final long serialVersionUID = -592616022226761148L;

    /** priority for tasks that can be run later */
    private static final int LATER_PRIORITY = Thread.NORM_PRIORITY;

    /** request processor for recognizing of folders */
    private static final RequestProcessor PROCESSOR = new RequestProcessor (
                "Folder recognizer" // NOI18N
            );


    /** data folder to work with */
    protected DataFolder folder;

    /** last time when this folder has been refreshed */
    transient private long time;

    /** The task that computes the content of FolderList. There is also
    * only one computation task in the PROCESSOR for each FolderList.
    * Whenever a new change notification arrives (thru file listener)
    * the previous task is canceled (if not running) and new is created.
    */
    transient private RequestProcessor.Task refreshTask;

    /** Primary files in this folder. Maps (FileObject, Reference (DataObject))
    */
    transient private HashMap primaryFiles = null;

    /** order of primary files (FileObject) */
    transient private List order;

    transient private ErrorManager err;

    /**
    * @param df data folder to show
    */
    public FolderList (DataFolder df, boolean attach) {
        folder = df;
        err = TopManager.getDefault ().getErrorManager ().getInstance
            ("org.openide.loaders.FolderList." + df.getPrimaryFile ().getPackageName ('.')); // NOI18N
        if (attach) {
            FileObject fo = df.getPrimaryFile ();

            // creates object that handles all elements in array and
            // assignes it to the
            fo.addFileChangeListener (WeakListener.fileChange (this, fo));
        }
    }

    final void reassign(DataFolder df, FileObject fo) {
        folder = df;
        // reassign is called from DataFolder.handleMove()
        // in this time the folder - df - does not have
        // setup the right primary file
        // so the fo is the new primary file for df
        fo.addFileChangeListener (WeakListener.fileChange (this, fo));
    }


    /** List all children.
    * @return array with children
    */
    public DataObject[] getChildren () {
        List res = getChildrenList ();
        DataObject[] arr = new DataObject[res.size ()];
        res.toArray (arr);
        return arr;
    }

    /** Computes array of children associated
    * with this folder.
    */
    public List getChildrenList () {
        ListTask lt = getChildrenList (null);
        lt.task.waitFinished ();
        return lt.result;
    }

    /** Starts computation of children list asynchronously.
    */
    public RequestProcessor.Task computeChildrenList (FolderListListener filter) {
        return getChildrenList (filter).task;
    }

    private ListTask getChildrenList (FolderListListener filter) {
        ListTask lt = new ListTask (filter);
        int priority = Thread.currentThread().getPriority();

        // and then post your read task and wait
        lt.task = PROCESSOR.post (lt, 0, priority);
        return lt;
    }

    /** Setter for sort mode.
    */
    public void changeComparator () {
        err.log ("changeComparator");
        PROCESSOR.post (new Runnable () {
                            public void run () {
                                // if has been notified
                                // change mode and regenerated children
                                if (primaryFiles != null) {
                                    // the old children
                                    err.log ("changeComparator: get old");
                                    List v = getObjects (null);
                                    if (v.size () != 0) {
                                        // the new children - also are stored to be returned next time from getChildrenList ()
                                        order = null;
                                        err.log ("changeComparator: get new");
                                        List r = getObjects (null);
                                        err.log ("changeComparator: fire change");
                                        fireChildrenChange (r, v);
                                    }
                                }
                            }
                        }, 0, Thread.MIN_PRIORITY);
    }

    /** Refreshes the list of children.
    */
    public void refresh () {
        refresh (Long.MAX_VALUE);
    }

    /** Does refreshes if not done.
    */
    private RequestProcessor.Task doRefreshIfNotDone () {
        if (refreshTask == null) {
            synchronized (this) {
                if (refreshTask == null) {
                    refresh ();
                }
            }
        }
        return refreshTask;
    }

    /** Refreshes the list of children but only if the
    * lastest refresh time < the passed argument
    * @param now the time
    */
    private void refresh (final long now) {
        err.log ("refresh @" + now);
        refreshTask = PROCESSOR.post (new Runnable () {
            public void run () {
                err.log ("refresh: now=" + now + " time=" + time);
                if (now > time - 10) {
                    // the change has been done after our last refresh
                    // => we have to refresh again
                    time = System.currentTimeMillis();
                    if (primaryFiles != null) {
                        // list of children is created, recreate it for new files
                        createBoth (null, true);
                    }
                }
            }
        }, 0, LATER_PRIORITY);
    }

    //
    // FileChangeListener methods
    //

    /** Fired when a file has been changed.
    * @param fe the event describing context where action has taken place
    */
    public void fileChanged (FileEvent fe) {
        err.log ("fileChanged: " + fe);
        
        FileObject fo = fe.getFile ();
        if (fo.isData ()) {
            // when a data on the disk has been changed, look whether we
            // should reparse children
            if (primaryFiles != null) {
                // a file has been changed and the list of files is created
                try {
                    DataObject obj = DataObject.find (fo);
                    if (!primaryFiles.containsKey (obj.getPrimaryFile ())) {
                        // BUGFIX: someone who recognized the file and who isn't registered
                        // yet =>
                        // may be still not O.K.

                        // this primary file is not registered yet
                        // so recreate list of children
                        refresh (fe.getTime ());
                    }
                } catch (DataObjectNotFoundException ex) {
                    TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
                    // file without data object => no changes
                }
            }
        }
    }

    /** Fired when a file has been deleted.
    * @param fe the event describing context where action has taken place
    */
    public void fileDeleted (FileEvent fe) {
        err.log ("fileDeleted: " + fe);
        //    boolean debug = fe.getFile().toString().equals("P"); // NOI18N
        //if (debug) System.out.println ("fileDeleted: " + fe.getFile ()); // NOI18N
        //if (debug) System.out.println ("fileList: " + fileList + " file: " + fileList.get (fe.getFile ())); // NOI18N
        if (primaryFiles == null || primaryFiles.containsKey (fe.getFile ())) {
            // one of main files has been deleted => reparse
            //if (debug) System.out.println ("RecreateChildenList"); // NOI18N
            refresh (fe.getTime ());
            //if (debug) System.out.println ("Done"); // NOI18N
        }
    }

    /** Fired when a new file has been created. This action can only be
    * listened in folders containing the created file up to the root of
    * file system.
    *
    * @param fe the event describing context where action has taken place
    */
    public void fileDataCreated (FileEvent fe) {
        err.log ("fileDataCreated: " + fe);
        refresh (fe.getTime ());
    }

    /** Fired when a new file has been created. This action can only be
    * listened in folders containing the created file up to the root of
    * file system.
    *
    * @param fe the event describing context where action has taken place
    */
    public void fileFolderCreated (FileEvent fe) {
        err.log ("fileFolderCreated: " + fe);
        refresh (fe.getTime ());
    }

    /** Fired when a new file has been renamed.
    *
    * @param fe the event describing context where action has taken place
    */
    public void fileRenamed (FileRenameEvent fe) {
        err.log ("fileRenamed: " + fe);
        refresh (fe.getTime ());
    }

    /** Fired when a file attribute has been changed.
    *
    * @param fe the event describing context where action has taken place
    */
    public void fileAttributeChanged (FileAttributeEvent fe) {
        err.log ("fileAttributeChanged: " + fe);
        // [PENDING] probably this ought to do something, since if a
        // file attribute changes on disk (e.g. at request of
        // filesystem, ultimately of module installer?) and it affects
        // ordering/sort mode, then the order should be refreshed.
        // However, DataFolder.set{Order,RelativeOrder} already call
        // changeComparator() for this purpose and we do not want to
        // do it twice.
    }


    //
    // Processing methods
    //

    /** Getter for list of children.
    * @param f filter to be notified about additions
    * @return List with DataObject types
    */
    private List getObjects (FolderListListener f) {
        err.log ("getObjects");
        List res;
        if (primaryFiles == null) {
            res = createBoth (f, false);
        } else {
            if (order != null) {
                res = createObjects (order, primaryFiles, f);
            } else {
                res = createObjects (primaryFiles.keySet (), primaryFiles, f);
                res = carefullySort (res, folder.getComparator ());
                order = createOrder (res);
            }
        }
        return res;
        /* createChildrenAndFiles ();/*
        ArrayList v = (Collection)childrenList.get ();
        //if (debug) System.out.println ("Children list xxxxxxxxxxxxxx");
        if (v == null) {
        //if (debug) System.out.println ("Create them xxxxxxxxxxxx");
          v = createChildrenList (f);
        //if (debug) System.out.println ("result: " + v);
    }
        return v;*/
    }

    /** Sort a list of DataObject's carefully.
     * The supplied comparator should supply a basic ordering,
     * and may also have an associated overriding partial ordering.
     * If the partial ordering is given and is self-contradictory,
     * it will be ignored and a warning issued.
     * @param l the list to sort
     * @param c a comparator and maybe partial comparator to use
     * @return the sorted list (may or may not be the same)
     */
    private /*static*/ List carefullySort (List l, DataFolder.PartialComparator c) {
        err.log ("carefullySort");
        Collections.sort (l, c);
        Comparator p = c.getPartialComparator ();
        if (p == null) {
            return l;
        } else {
            err.log ("carefullySort: partial orders");
            try {
                // [PENDING] this is not really particularly efficient.
                // Utilities.partialSort is quadratic since it looks for
                // all possible partial orderings. In this concrete case
                // we can easily enumerate the partial orderings (and there
                // are probably not many of them). So a more efficient
                // implementation would surely be possible. On the other hand,
                // most folders with partial orderings are not likely to have
                // all that many files in them, probably. Settings would have
                // the most. However if Settings is split up this will probably
                // not be a significant issue.
                return Utilities.partialSort (l, p, true);
            } catch (Utilities.UnorderableException ue) {
                err.annotate(ue, ErrorManager.WARNING, "Folder cannot be sorted: " + folder.getPrimaryFile (), null, null, null); // NOI18N
                err.notify (ErrorManager.WARNING, ue);
                return l;
            }
        }
    }

    /** Creates list of primary files from the list of data objects.
    * @param list list of DataObject
    * @return list of FileObject
    */
    private static List createOrder (List list) {
        int size = list.size ();
        List res = new ArrayList (size);

        for (int i = 0; i < size; i++) {
            res.add (((DataObject)list.get (i)).getPrimaryFile ());
        }

        return res;
    }

    /** Creates array of data objects from given order
    * and mapping between files and data objects.
    *
    * @param order list of FileObjects that define the order to use
    * @param map mapping (FileObject, Reference (DataObject)) to create data objects from
    * @param f filter that is notified about additions - only items
    * which are accepted by the filter will be added. Null means no filtering.
    * @return array of data objects
    */
    private /*static*/ List createObjects (
        Collection order, Map map, FolderListListener f
    ) {
        err.log ("createObjects");
        int size = order.size ();

        Iterator it = order.iterator ();

        List res = new ArrayList (size);
        for (int i = 0; i < size; i++) {
            FileObject fo = (FileObject)it.next ();

            Reference ref = (Reference)map.get (fo);
            DataObject obj = (DataObject)ref.get ();

            if (obj == null) {
                // try to find new data object
                try {
                    obj = DataObject.find (fo);
                    ref = new SoftReference (obj);
                } catch (DataObjectNotFoundException ex) {
                    TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
                }
            }
            // add if accepted
            if (obj != null) {

                // JST: Cannot be avoided otherwise DataObject.files () can be unconsistent
                // avoid to checkFiles(this)
                //obj.recognizedByFolder();

                if (f == null) {
                    // accept all objects
                    res.add (obj);
                } else {
                    // allow the listener f to filter
                    // objects in the array res
                    f.process (obj, res);
                }
            }
        }

        if (f != null) {
            f.finished (res);
        }

        return res;
    }

    /** Scans for files in the folder and creates representation for
     * children. Fires info about changes in the nodes.
     *
     * @param filter listener to addition of nodes or null
     * @param notify true if changes in the children should be fired
     * @return vector of children
     */
    private List createBoth (FolderListListener filter, boolean notify) {
        err.log ("createBoth");
        // map for (FileObject, DataObject)
        final HashMap file = new HashMap ();

        // list of all processed objects
        List all = new ArrayList ();
        // result list to return from the method
        List res = new ArrayList ();

        // map of current objects (FileObject, DataObject)
        final HashMap remove = primaryFiles == null ?
                               new HashMap () : (HashMap)primaryFiles.clone ();

        // list of new objects to add
        final List add = new ArrayList ();

        DataLoaderPool pool = TopManager.getDefault ().getLoaderPool ();

        // hashtable with FileObjects that are marked to be recognized
        // and that is why being out of enumeration
        final HashSet marked = new HashSet ();
        DataLoader.RecognizedFiles recog = new DataLoader.RecognizedFiles () {
                                               /** Adds the file object to the marked hashtable.
                                               * @param fo file object (can be <CODE>null</CODE>)
                                               */
                                               public void markRecognized (FileObject fo) {
                                                   if (fo != null) {
                                                       marked.add (fo);
                                                   }
                                               }
                                           };
        // enumeration of all files in the folder
        Enumeration en = folder.getPrimaryFile ().getChildren (false);
        while (en.hasMoreElements ()) {
            FileObject fo = (FileObject)en.nextElement ();
            if (!marked.contains (fo)) {
                // the object fo has not been yet marked as recognized
                // => continue in computation
                DataObject obj;
                try {
                    obj = pool.findDataObject (fo, recog);
                } catch (DataObjectExistsException ex) {
                    // use existing data object
                    obj = ex.getDataObject ();
                } catch (IOException ex) {
                    // data object not recognized or not found
                    obj = null;
                }

                if (obj != null) {
                    // adds object to data if it is not already there

                    // avoid to checkFiles(this)
                    obj.recognizedByFolder();

                    // primary file
                    FileObject primary = obj.getPrimaryFile ();

                    boolean doNotRemovePrimaryFile = false;
                    if (!file.containsKey (primary)) {
                        // realy added object, test if it is new

                        // if we have not created primaryFiles before, then it is new
                        boolean goIn = primaryFiles == null;
                        if (!goIn) {
                            Reference r = (Reference)primaryFiles.get (primary);
                            // if its primary file is not between original primary files
                            // then data object is new
                            goIn = r == null;
                            if (!goIn) {
                                // if the primary file is there, but the previous data object
                                // exists and is different, then this one is new
                                DataObject obj2 = (DataObject)r.get ();
                                goIn = obj2 == null || obj2 != obj;
                                if (goIn) {
                                    doNotRemovePrimaryFile = true;
                                }
                            }
                        }

                        if (goIn) {
                            // realy new
                            add.add (obj);
                            /* JST: In my opinion it should not be here
                            * so I moved this out of this if. Is it ok?

                            if (filter != null) {
                              // fire info about addition
                              filter.acceptDataObject (obj);
                        }
                            */
                        }
                        // adds the object
                        all.add (obj);
                        if (filter == null) {
                            res.add (obj);
                        } else {
                            filter.process (obj, res);
                        }
                    }

                    if (!doNotRemovePrimaryFile) {
                        // this object exists it should not be removed
                        remove.remove (primary);
                    }

                    // add it to the list of primary files
                    file.put (primary, new WeakReference (obj));
                } else {
                    // 1. nothing to add to data object list
                    // 2. remove this object if it was in list of previous ones
                    // 3. do not put the file into list of know primary files
                    // => do nothing at all
                }
            }
        }

        // !!! section that fires info about changes should be here !!!

        // now file contains newly inserted files
        // data contains data objects
        // remove contains data objects that should be removed
        // add contains data object that were added

        primaryFiles = file;

        all = carefullySort (all, folder.getComparator ());
        order = createOrder (all);
        if (all.size () == res.size ()) {
            // assume no filtering has been done
            res = all;
        } else {
            // sort also content of res
            res = carefullySort (res, folder.getComparator ());
        }
            

        ////if (debug) System.out.println ("Notified: " + notified + " added: " + add.size () + " removed: " + remove.size ()); // NOI18N
        if (notify) {
            fireChildrenChange (add, createObjects (new ArrayList (remove.keySet ()), remove, null));
        }

        // notify the filter
        if (filter != null) {
            filter.finished (res);
        }

        return res;
    }

    /** Fires info about change of children to the folder.
    * @param add added data objects
    * @param removed removed data objects
    */
    private void fireChildrenChange (List add, List removed) {
        /*    if (!add.isEmpty () || !removed.isEmpty ()) {
              System.out.println("Firing: " + folder.getPrimaryFile ());
              System.out.println("   add: " + add);
              System.out.println("   rem: " + removed);
            } else {
              System.out.println("No ch.: " + folder.getPrimaryFile ());
            }
        */    
        folder.fireChildrenChange (add, removed);
    }

    /** Task that holds result and also task. Moreover
    * can do the computation.
    */
    private final class ListTask implements Runnable {
        private FolderListListener filter;

        public ListTask (FolderListListener filter) {
            this.filter = filter;
        }

        public List result;
        public RequestProcessor.Task task;

        public void run () {
            err.log ("ListTask.run 1");
            // invokes the refresh task before we do anything else
            if (refreshTask != null) {
                refreshTask.waitFinished ();
            }
            err.log ("ListTask.run 2");

            result = getObjects (filter);
            err.log ("ListTask.run 3");
        }
    }

}
