/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.awt.Image;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.PropertyEditor;
import java.lang.reflect.InvocationTargetException;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.Properties;
import java.util.Enumeration;
import javax.swing.Icon;
import javax.swing.ImageIcon;

import org.openide.*;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.actions.OpenAction;
import org.openide.util.*;
import org.openide.util.actions.*;
import org.openide.nodes.*;

/** A data object whose only purpose is to supply <code>InstanceCookie</code>.
* The instances are created by default instantiation; the name of the class
* to instantiate is stored on disk, typically right in the file name.
* <p>This data object is generally used to configure menus and toolbars,
* though it could be used in any situation requiring instances to be present in
* a folder; for example, anything using {@link FolderInstance}.
* <p>Typical instance classes are subclasses of {@link SystemAction} to make
* menu items or toolbar buttons; {@link javax.swing.JSeparator} for a menu
* separator; or {@link javax.swing.JToolBar.Separator} for a toolbar
* separator.
* <p>Use {@link #create} and {@link #remove} to make the objects.
* Better yet, use an XML filesystem to install them declaratively.
*
* @author Ian Formanek
*/
public class InstanceDataObject extends MultiDataObject implements InstanceCookie {
    /** generated Serialized Version UID */
    static final long serialVersionUID = -6134784731744777123L;

    private static final String EA_INSTANCE_CLASS = "instanceClass"; // NOI18N
    private static final String EA_INSTANCE_CREATE = "instanceCreate"; // NOI18N

    /** opening symbol */
    private static final char OPEN = '[';
    /** closing symbol */
    private static final char CLOSE = ']';

    /** File extension for instance data objects. */
    public static final String INSTANCE = "instance"; // NOI18N

    /** File extension for serialized files. */
    private static final String SER_EXT = "ser"; // NOI18N
    
    /** optional property file key for icon resource */
    private static final String ICON_NAME = "icon"; // NOI18N
    /** optional property file key for display name (localizable with suffix) */
    private static final String DISPLAY_NAME = "name"; // NOI18N
    
    
    /** the object that handles instance cookie manipulation */
    private Ser ser;
    
    /** Create a new instance.
    * Do not use this to make instances; use {@link #create}.
    * @param pf primary file object for this data object
    * @param loader the loader
    * @throws DataObjectExistsException if it already exists
    */
    public InstanceDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException {
        super (pf, loader);
        
        ser = new Ser (getPrimaryEntry());
        
        if (pf.hasExt ( SER_EXT )) {
            // only if we have ser extension the Ser (InstanceCookie.Origin)
            // should be accessible
            getCookieSet ().add (ser);
        } else {
            // otherwise we implement just the InstanceCookie directly
        }
    }

    /** Finds instance of specified name in a given folder.
    * @param folder the folder to create the instance data object in
    * @param name the name to give to the object (can be <code>null</code> if no special name besides the class name is needed)
    * @param className the name of the class the new object should provide an instance of
    * @return the found instance data object or null if it does not exist
    */
    public static InstanceDataObject find (DataFolder folder, String name, String className) {
        FileObject fo = folder.getPrimaryFile ();
        String fileName;
        if (name == null) {
            fileName = OPEN + className.replace ('.', '-') + CLOSE;
        } else {
            fileName = escape (name) + OPEN + className.replace ('.', '-') + CLOSE;
        }

        FileObject file = fo.getFileObject (fileName, INSTANCE);
        if (file != null) {
            try {
                return (InstanceDataObject)DataObject.find (file);
            } catch (DataObjectNotFoundException e) {
                return null;
            }
        } else {
            return null;
        }
    }

    /** Finds instance of specified name in a given folder.
    * @param folder the folder to create the instance data object in
    * @param name the name to give to the object (can be <code>null</code> if no special name besides the class name is needed)
    * @param clazz the class to create instance for (see class header for details)
    * @return the found instance data object or null if it does not exist
    */
    public static InstanceDataObject find (DataFolder folder, String name, Class clazz) {
        return find (folder, name, clazz.getName ());
    }

    /** Create a new <code>InstanceDataObject</code> in a given folder. If object with specified name already exists, it is returned.
    * You should specify the name if there is a chance another file of the same
    * instance class already exists in the folder; or just to provide a more
    * descriptive name, which will appear in the Explorer for example.
    * @param folder the folder to create the instance data object in
    * @param name the name to give to the object (can be <code>null</code> if no special name besides the class name is needed)
    * @param className the name of the class the new object should provide an instance of (see class header for details)
    * @return the newly created or eisting instance data object
    * @exception IOException if the file cannot be created
    * @deprecated Use of XML filesystems to install instances is preferred.
    */
    public static InstanceDataObject create (DataFolder folder, String name, String className) throws IOException {
        FileObject fo = folder.getPrimaryFile ();
        String fileName;
        if (name == null) {
            fileName = OPEN + className.replace ('.', '-') + CLOSE;
        } else {
            fileName = escape (name) + OPEN + className.replace ('.', '-') + CLOSE;
        }

        FileObject newFile = fo.getFileObject (fileName, INSTANCE);
        if (newFile == null) newFile = fo.createData (fileName, INSTANCE);

        return (InstanceDataObject)DataObject.find (newFile);
    }

    /** Create a new <code>InstanceDataObject</code> in a given folder. If object with specified name already exists, it is returned.
    * You should specify the name if there is a chance another file of the same
    * instance class already exists in the folder; or just to provide a more
    * descriptive name, which will appear in the Explorer for example.
    * @param folder the folder to create the instance data object in
    * @param name the name to give to the object (can be <code>null</code> if no special name besides the class name is needed)
    * @param clazz the class to create instance for (see class header for details)
    * @return the newly created or eisting instance data object
    * @exception IOException if the file cannot be created  
    * @deprecated Use of XML filesystems to install instances is preferred.
    */
    public static InstanceDataObject create (DataFolder folder, String name, Class clazz) throws IOException {
        return create (folder, name, clazz.getName ());
    }

    /** Remove an existing instance data object.
    * If you have the exact file name, just call {@link DataObject#delete};
    * this method lets you delete an instance you do not have an exact record
    * of the file name for, based on the same information used to create it.
    * @param folder the folder to remove the file from
    * @param name the name of the instance (can be <code>null</code>)
    * @param className the name of class the object referred to (see class header for details)
    * @return <code>true</code> if the instance was succesfully removed, <code>false</code> if not
    * @deprecated Use of XML filesystems to install instances is preferred.
    */
    public static boolean remove (DataFolder folder, String name,
                                  String className) {
        FileLock lock = null;
        try {
            String fileName;
            if (name == null) {
                fileName = OPEN + className.replace ('.', '-') + CLOSE;
            } else {
                fileName = escape (name) + OPEN + className.replace ('.', '-') + CLOSE;
            }
            FileObject fileToRemove =
                folder.getPrimaryFile().getFileObject (fileName, INSTANCE);
            if (fileToRemove == null) // file not found
                return false;
            lock = fileToRemove.lock();
            fileToRemove.delete(lock);
        } catch (IOException exc) {
            // something is bad, instance wasn't removed
            return false;
        } finally {
            if (lock != null)
                lock.releaseLock();
        }
        return true;
    }

    /** Remove an existing instance data object.
    * If you have the exact file name, just call {@link DataObject#delete};
    * this method lets you delete an instance you do not have an exact record
    * of the file name for, based on the same information used to create it.
    * @param folder the folder to remove the file from
    * @param name the name of the instance (can be <code>null</code>)
    * @param className the name of class the object referred to (see class header for details)
    * @return <code>true</code> if the instance was succesfully removed, <code>false</code> if not
    * @deprecated Use of XML filesystems to install instances is preferred.
    */
    public static boolean remove (DataFolder folder, String name, Class clazz) {
        return remove (folder, name, clazz.getName ());
    }

    /* Help context for this object.
    * @return help context
    */
    public HelpCtx getHelpCtx () {
        HelpCtx test = InstanceSupport.findHelp (this);
        if (test != null)
            return test;
        else
            return new HelpCtx (InstanceDataObject.class);
    }

    /* Provides node that should represent this data object. When a node for representation
    * in a parent is requested by a call to getNode (parent) it is the exact copy of this node
    * with only parent changed. This implementation creates instance
    * <CODE>DataNode</CODE>.
    * <P>
    * This method is called only once.
    *
    * @return the node representation for this data object
    * @see DataNode
    */
    protected Node createNodeDelegate () {
        // Instances of Node or Node.Handle should be used as is.
        try {
            if (Node.class.isAssignableFrom (instanceClass ())) {
                Node n = (Node)instanceCreate ();
                return n;
            } else if (Node.Handle.class.isAssignableFrom (instanceClass ())) {
                Node.Handle h = (Node.Handle) instanceCreate ();
                return h.getNode ().cloneNode ();
            }
        } catch (IOException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        } catch (ClassNotFoundException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        }

        return new InstanceNode (this);
    }
    
    /** Finds delegate instance cookie/if provided in cookie set.
    * @return instance cookie or null
    */
    private InstanceCookie delegateIC () {
        return ser;
    }
    
    /* The name of the bean for this file or null if the class name is not encoded
    * in the file name and rather the CLASS_NAME property from the file content should be used.
    *
    * @return the name for the instance or null if the class name is not defined in the name
    */
    public String instanceName () {
        InstanceCookie delegateIC = delegateIC ();
        return delegateIC.instanceName ();
    }

    /* The class of the instance represented by this cookie.
    * Can be used to test whether the instance is of valid
    * class before it is created.
    *
    * @return the class of the instance
    * @exception IOException an I/O error occured
    * @exception ClassNotFoundException the class has not been found
    */
    public Class instanceClass ()
    throws java.io.IOException, ClassNotFoundException {
        InstanceCookie delegateIC = delegateIC ();
        return delegateIC.instanceClass ();
    }

    /*
    * @return an object to work with
    * @exception IOException an I/O error occured
    * @exception ClassNotFoundException the class has not been found
    */
    public Object instanceCreate ()
    throws java.io.IOException, ClassNotFoundException {
        InstanceCookie delegateIC = delegateIC ();
        return delegateIC.instanceCreate ();
    }

    /* Overriden to return only first part till the bracket */
    public String getName () {
        String superName = super.getName();
        int bracket = superName.indexOf (OPEN);
        return unescape ((bracket < 0) ? superName : superName.substring (0, bracket));
    }
    
    // [PENDING] probably setName also needs to be overridden!
    /* Renames all entries and changes their files to new ones.
     */
    protected FileObject handleRename (String name) throws IOException {
        String className = super.getName();
        int bracket = className.indexOf (OPEN);
        className = ((bracket < 0) ? "" : className.substring(bracket));
        
        String mangled = escape(name) + className;
        return super.handleRename(mangled);
    }
    

    /** Hex-escapes anything potentially nasty in some text. */
    private static String escape (String text) {
        int len = text.length ();
        StringBuffer escaped = new StringBuffer (len);
        for (int i = 0; i < len; i++) {
            char c = text.charAt (i);
            // For some reason Windoze throws IOException if angle brackets in filename...
            if (c == '/' || c == ':' || c == '\\' || c == OPEN || c == CLOSE || c == '<' || c == '>' ||
                    c == '.' || c == '"' || c < '\u0020' || c > '\u007E' || c == '#') {
                // Hex escape.
                escaped.append ('#');
                String hex = Integer.toString (c, 16).toUpperCase ();
                if (hex.length () < 4) escaped.append ('0');
                if (hex.length () < 3) escaped.append ('0');
                if (hex.length () < 2) escaped.append ('0');
                escaped.append (hex);
            } else {
                escaped.append (c);
            }
        }
        return escaped.toString ();
    }

    /** Removes hex escapes and regenerates displayable Unicode. */
    private static String unescape (String text) {
        int len = text.length ();
        StringBuffer unesc = new StringBuffer (len);
        for (int i = 0; i < len; i++) {
            char c = text.charAt (i);
            if (c == '#') {
                if (i + 4 >= len) {
                    if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                        System.err.println("trailing garbage in instance name: " + text); // NOI18N
                    break;
                }
                try {
                    char[] hex = new char[4];
                    text.getChars (i + 1, i + 5, hex, 0);
                    unesc.append ((char) Integer.parseInt (new String (hex), 16));
                } catch (NumberFormatException nfe) {
                    if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                        nfe.printStackTrace ();
                }
                i += 4;
            } else {
                unesc.append (c);
            }
        }
        return unesc.toString ();
    }
    
    /** Creates a save task for periodic updates of the changed bean.
     * Used in InstanceNode.
     */
    final RequestProcessor.Task saveTask () {
        return RequestProcessor.createRequest (ser);
    }

    /* Creates new object from template.
    * @exception IOException
    */
    protected DataObject handleCreateFromTemplate (
        DataFolder df, String name
    ) throws IOException {        
        FileObject fo = df.getPrimaryFile ();
        try {
            if ( java.io.Serializable.class.isAssignableFrom( instanceClass()) ) {            
                InstanceCookie ic = (InstanceCookie)this.getCookie(InstanceCookie.class);
                Object obj = ic.instanceCreate();                
                                
                return DataObject.find( createSerFile( df, name, obj ) ); 
            }
        } catch (ClassNotFoundException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        }        
        
        return super.handleCreateFromTemplate(df, name);
    }
    
    private FileObject createSerFile(
        DataFolder df, String name, Object obj
    ) throws IOException {      
        FileLock lock = null;
        OutputStream ostream = null;
        FileObject newFile = null;
        try {
            FileObject fo = df.getPrimaryFile ();
            
            if (name == null) {
                name = FileUtil.findFreeFileName(
                           df.getPrimaryFile (), getPrimaryFile ().getName (), getPrimaryFile ().getExt ()
                       );
            }
            
            newFile = fo.getFileObject (name, SER_EXT);
            if (newFile == null) newFile = fo.createData (name, SER_EXT);
                
            lock = newFile.lock ();
            ostream = newFile.getOutputStream(lock);
                
            ObjectOutputStream p = new ObjectOutputStream(ostream);
            p.writeObject(obj);                
            p.flush();
        } finally {
            if (ostream != null)
                ostream.close();
            if (lock != null)
                lock.releaseLock ();
        }
        return newFile;
    }
    
    /** Support for serialized objects.
     */
    private static final class Ser extends InstanceSupport.Origin 
    implements Runnable {
        /** the reference to the bean, so it is created just once when used */
        private java.lang.ref.Reference bean = new java.lang.ref.WeakReference (null);
        /** last time the bean was read from a file */
        private long saveTime;

        /** @param en entry to work on */
        public Ser (MultiDataObject.Entry en) {
            super (en);
        }
        
        public String instanceName () {
            FileObject fo = entry ().getFile ();
            if (!fo.hasExt (INSTANCE)) {
                return super.instanceName ();
            }
            
            // first of all try "instanceClass" property of the primary file
            Object attrName = fo.getAttribute (EA_INSTANCE_CLASS);
            if (attrName instanceof String) {
                return org.openide.util.Utilities.translate((String)attrName);
            }

            // otherwise extract the name from the filename
            String name = fo.getName ();

            int first = name.indexOf (OPEN) + 1;

            int last = name.indexOf (CLOSE);
            if (last < 0) {
                last = name.length ();
            }

            // take only a part of the string
            if (first < last) {
                name = name.substring (first, last);
            }

            name = name.replace ('-', '.');
            name = org.openide.util.Utilities.translate(name);

            //System.out.println ("Original: " + getPrimaryFile ().getName () + " new one: " + name); // NOI18N
            return name;
        }

        public Object instanceCreate () throws IOException, ClassNotFoundException {
            FileObject fo = entry ().getFile ();
            
            
            Object o;
            if (fo.lastModified ().getTime () <= saveTime) {
                o = bean.get ();
            } else {
                o = null;
            }
            
            if (o != null) {
                return o;
            }
            
            saveTime = fo.lastModified ().getTime ();
            if (saveTime < System.currentTimeMillis ()) {
                saveTime = System.currentTimeMillis ();
            }
            if (fo.hasExt (INSTANCE)) {
                // try to ask for instance creation attribute
                o = fo.getAttribute (EA_INSTANCE_CREATE);
            }

            if (o == null) {
                // try super method
                o = super.instanceCreate ();
            } 
            
            // remember the created value
            bean = new java.lang.ref.WeakReference (o);
            return o;
        }
        
        /** Saves the bean to disk.
         */
        public void run () {
            Object bean = this.bean.get ();
            if (bean == null) {
                // nothing to save
                return;
            }
            
            try {
                FileLock lock = entry ().takeLock ();
                try {
                    ObjectOutputStream oos = new ObjectOutputStream (
                        entry ().getFile ().getOutputStream (lock)
                    );
                    try {
                        oos.writeObject (bean);
                        // avoid bean reloading
                        saveTime = entry ().getFile ().lastModified ().getTime ();
                    } finally {
                        oos.close ();
                    }
                } finally {
                    lock.releaseLock ();
                }
            } catch (IOException ex) {
                ErrorManager err = TopManager.getDefault ().getErrorManager ();
                err.annotate (ex, NbBundle.getMessage (
                    InstanceDataObject.class, "EXC_CannotSaveBean", 
                    instanceName (), entry ().getFile ().getPackageNameExt ('/', '.')
                ));
                err.notify (ex);
            }
                
        }
        
    } // end of Ser
        
}
