/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.awt.Image;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.PropertyEditor;
import java.lang.reflect.InvocationTargetException;
import java.io.IOException;
import java.net.URL;
import java.util.Properties;
import java.util.Enumeration;
import javax.swing.Icon;
import javax.swing.ImageIcon;

import org.openide.*;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.actions.OpenAction;
import org.openide.util.*;
import org.openide.util.actions.*;
import org.openide.nodes.*;

/** Node to represent a .ser or .instance file.
 *
 * @author  Jaroslav Tulach
 */
final class InstanceNode extends DataNode implements Runnable {
    /** if a property is modified, what is the delay before it is saved? */
    private static final int SAVE_DELAY = 5000;
    
    /** icon base */
    private static final String INSTANCE_ICON_BASE =
        "/org/openide/resources/instanceObject"; // NOI18N
    
    
    /** icon from bean info */
    private Image beanInfoIcon;

    /** Constructor */
    public InstanceNode (InstanceDataObject obj) {
        super (obj, Children.LEAF);
        
        setIconBase(INSTANCE_ICON_BASE);
        // This is not enabled; just serves to ensure that
        // CustomizeBeanAction is *not* the default!
        setDefaultAction (SystemAction.get (OpenAction.class));

        // bean info is not used only if the file specifies 
        // <attr name="beaninfo" booleanvalue="false" />
        //
        boolean noBeanInfo = Boolean.FALSE.equals (
            obj.getPrimaryFile ().getAttribute ("beaninfo") // NOI18N
        );

        if (!noBeanInfo) {
            // bean info processing in different thread
            RequestProcessor.postRequest (this, 0, Thread.MIN_PRIORITY);
        }
    }
    
    /** Getter for instance data object.
     * @return instance data object
     */
    private InstanceDataObject i () {
        return (InstanceDataObject)getDataObject ();
    }

    /** Find an icon for this node (in the closed state).
    * @param type constant from {@link java.beans.BeanInfo}
    * @return icon to use to represent the node
    */
    public Image getIcon (int type) {
        if (beanInfoIcon != null) {
            return beanInfoIcon;
        } else {
            return super.getIcon (type);
        }
    }

    /** Find an icon for this node (in the open state).
    * This icon is used when the node may have children and is expanded.
    *
    * @param type constant from {@link java.beans.BeanInfo}
    * @return icon to use to represent the node when open
    */
    public Image getOpenedIcon (int type) {
        return getIcon (type);
    }

    /** Runs the get for the icon and name */
    public void run () {
        initIconName ();
    }
        
    private void initIconName () {
        try {
            Class clazz = i ().instanceClass ();
            //Fixed bug #5610
            //Class javax.swing.JToolBar$Separator does not have icon
            //we will use temporarily icon from javax.swing.JSeparator
            //New icon is requested.

            String className = clazz.getName ();
            BeanInfo bi;
            if (
                className.equals ("javax.swing.JSeparator") || 
                className.equals ("javax.swing.JToolBar$Separator")
            ) {
                setDisplayName (NbBundle.getMessage (InstanceDataObject.class, "LBL_separator_instance"));

                Class clazzTmp = Class.forName ("javax.swing.JSeparator");
                bi = Utilities.getBeanInfo (clazzTmp);
            } else {
                bi = Utilities.getBeanInfo (clazz);
            }

            if (bi != null) {
                beanInfoIcon = bi.getIcon (BeanInfo.ICON_COLOR_16x16);
                if (beanInfoIcon != null) {
                    beanInfoIcon = toBufferedImage(beanInfoIcon, true);
                }
                // [PENDING] could also try to set display name based on bean descriptor
            }
            // Also specially handle SystemAction's.
            if (SystemAction.class.isAssignableFrom (clazz)) {
                SystemAction action = SystemAction.get (clazz);
                if (beanInfoIcon == null) {
                    Icon icon = action.getIcon ();
                    // [PENDING] not very pretty, but there is no good way to
                    // get an Image from an Icon that I know of
                    if (icon instanceof ImageIcon) {
                        beanInfoIcon = ((ImageIcon) icon).getImage ();
                    }
                }
                // Set node's display name.
                String name = action.getName ();
                int amper = name.indexOf ((char) '&');
                if (amper != -1)
                    name = name.substring (0, amper) + name.substring (amper + 1);
                if (name.endsWith ("...")) // NOI18N
                    name = name.substring (0, name.length () - 3);
                name = name.trim ();
                setDisplayName (name);
            }
        } catch (Exception e) {
            // Problem ==>> use default icon
            TopManager.getDefault ().getErrorManager ().notify (
                ErrorManager.INFORMATIONAL, e
            );
        }

        if (beanInfoIcon != null) {
            // fire change
            fireIconChange ();
        }
    }

    protected Sheet createSheet () {
        Sheet orig = super.createSheet ();
    
        if (getDataObject ().getPrimaryFile ().hasExt ("ser")) { // NOI18N
            // it is possible that the sheet change will take longer time
            RequestProcessor.postRequest (new Runnable () {
                public void run () {
                    changeSheet (getSheet ());
                    firePropertySetsChange(null, null);
                }
            }, 0, Thread.MIN_PRIORITY + 1);
        } else {
            // just instance file, change here
            Sheet.Set props = orig.get (Sheet.PROPERTIES);
            props.put (new PropertySupport.ReadOnly (
                "className", String.class,
                NbBundle.getMessage (InstanceDataObject.class, "PROP_instance_class"),
                NbBundle.getMessage (InstanceDataObject.class, "HINT_instance_class")
            ) {
                public Object getValue () {
                    return i ().instanceName ();
                }
            });
        }
        
        return orig;
    }
        
        
    private void changeSheet (Sheet orig) {
        Sheet.Set props = orig.get (Sheet.PROPERTIES);

        // default value is to modify props, if introspection does not 
        // succeed or is not computed at all
        Sheet.Set files = props;
        
        try {
            // properties
            BeanInfo beanInfo = Utilities.getBeanInfo (i ().instanceClass ());
            BeanNode.Descriptor descr = BeanNode.computeProperties (i ().instanceCreate (), beanInfo);

            // ok, introspection succeeded, we will have properties tab
            orig.remove (Sheet.PROPERTIES);
            files.setName ("files");  // NOI18N
            files.setDisplayName (
                NbBundle.getMessage (InstanceDataObject.class, "LBL_instance_file_info")
            );

            props = Sheet.createPropertiesSet();
            if (descr.property != null) {
                convertProps (props, descr.property, i ().saveTask ());
            }
            orig.put (props);

            if (descr.expert != null && descr.expert.length != 0) {
                Sheet.Set p = Sheet.createExpertSet();
                convertProps (p, descr.expert, i ().saveTask ());
                orig.put (p);
            }
        } catch (ClassNotFoundException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        } catch (IOException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        } catch (IntrospectionException ex) {
            TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
        }


        if (props != files) {
            orig.put (files);
        }
    }
    
    
    /** Method that converts properties of an object.
     * @param set set to add properties to
     * @param arr array of Node.Property and Node.IndexedProperty
     * @param task to invoke when a property changes
     */
    private static final void convertProps (
        Sheet.Set set, Node.Property[] arr, final RequestProcessor.Task t
    ) {
        for (int i = 0; i < arr.length; i++) {
            if (arr[i] instanceof Node.IndexedProperty) {
                set.put (new I ((Node.IndexedProperty)arr[i], t));
            } else {
                set.put (new P (arr[i], t));
            }
        }
    }        
    
    /** The method creates a BufferedImage which represents the same Image as the
     * parameter but consumes less memory.
     */
    private static final java.awt.Image toBufferedImage(Image img, boolean load) {
        // load the image
        if (load) {
            new javax.swing.ImageIcon(img);
        }
        
        java.awt.image.BufferedImage rep = createBufferedImage();
        java.awt.Graphics g = rep.createGraphics();
        g.drawImage(img, 0, 0, null);
        g.dispose();
        img.flush();
        return rep;
    }

    /** Creates BufferedImage 16x16 and Transparency.BITMASK */
    private static final java.awt.image.BufferedImage createBufferedImage() {
        java.awt.image.ColorModel model = java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment().
                                          getDefaultScreenDevice().getDefaultConfiguration().getColorModel(java.awt.Transparency.BITMASK);
        java.awt.image.BufferedImage buffImage = new java.awt.image.BufferedImage(model,
                model.createCompatibleWritableRaster(16, 16), model.isAlphaPremultiplied(), null);
        return buffImage;
    }
    
    //
    // inner classes - properties
    //
    
    /** A property that delegates every call to original property
     * but when modified, also starts a saving task.
     */
    private static final class P extends Node.Property {
        /** delegate */
        private Node.Property del;
        /** task to executed */
        private RequestProcessor.Task t;

        public P (Node.Property del, RequestProcessor.Task t) {
            super (del.getValueType ());
            this.del = del;
            this.t = t;
        }

        public void setName(java.lang.String str) {
            del.setName(str);
        }

        public void restoreDefaultValue() throws IllegalAccessException, InvocationTargetException {
            del.restoreDefaultValue();
        }

        public void setValue(java.lang.String str, java.lang.Object obj) {
            del.setValue(str, obj);
        }

        public boolean supportsDefaultValue() {
            return del.supportsDefaultValue();
        }

        public boolean canRead() {
            return del.canRead ();
        }

        public PropertyEditor getPropertyEditor() {
            return del.getPropertyEditor();
        }

        public boolean isHidden() {
            return del.isHidden();
        }

        public Object getValue() throws IllegalAccessException, InvocationTargetException {
            return del.getValue ();
        }

        public void setExpert(boolean param) {
            del.setExpert(param);
        }

        /** Delegates the set value and also saves the bean.
         */
        public void setValue(Object val) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
            del.setValue (val);
            t.schedule (SAVE_DELAY);
        }

        public void setShortDescription(java.lang.String str) {
            del.setShortDescription(str);
        }

        public boolean isExpert() {
            return del.isExpert();
        }

        public boolean canWrite() {
            return del.canWrite ();
        }

        public Class getValueType() {
            return del.getValueType();
        }

        public java.lang.String getDisplayName() {
            return del.getDisplayName();
        }

        public java.util.Enumeration attributeNames() {
            return del.attributeNames();
        }

        public java.lang.String getShortDescription() {
            return del.getShortDescription();
        }

        public java.lang.String getName() {
            return del.getName();
        }

        public void setHidden(boolean param) {
            del.setHidden(param);
        }

        public void setDisplayName(java.lang.String str) {
            del.setDisplayName(str);
        }

        public boolean isPreferred() {
            return del.isPreferred();
        }

        public java.lang.Object getValue(java.lang.String str) {
            return del.getValue(str);
        }

        public void setPreferred(boolean param) {
            del.setPreferred(param);
        }

    } // end of P

    /** A property that delegates every call to original property
     * but when modified, also starts a saving task.
     */
    private static final class I extends Node.IndexedProperty {
        /** delegate */
        private Node.IndexedProperty del;
        /** task to executed */
        private RequestProcessor.Task t;

        public I (Node.IndexedProperty del, RequestProcessor.Task t) {
            super (del.getValueType (), del.getElementType ());
            this.del = del;
            this.t = t;
        }

        public void setName(java.lang.String str) {
            del.setName(str);
        }

        public void restoreDefaultValue() throws IllegalAccessException, InvocationTargetException {
            del.restoreDefaultValue();
        }

        public void setValue(java.lang.String str, java.lang.Object obj) {
            del.setValue(str, obj);
        }

        public boolean supportsDefaultValue() {
            return del.supportsDefaultValue();
        }

        public boolean canRead() {
            return del.canRead ();
        }

        public PropertyEditor getPropertyEditor() {
            return del.getPropertyEditor();
        }

        public boolean isHidden() {
            return del.isHidden();
        }

        public Object getValue() throws IllegalAccessException, InvocationTargetException {
            return del.getValue ();
        }

        public void setExpert(boolean param) {
            del.setExpert(param);
        }

        /** Delegates the set value and also saves the bean.
         */
        public void setValue(Object val) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
            del.setValue (val);
            t.schedule (SAVE_DELAY);
        }

        public void setShortDescription(java.lang.String str) {
            del.setShortDescription(str);
        }

        public boolean isExpert() {
            return del.isExpert();
        }

        public boolean canWrite() {
            return del.canWrite ();
        }

        public Class getValueType() {
            return del.getValueType();
        }

        public java.lang.String getDisplayName() {
            return del.getDisplayName();
        }

        public java.util.Enumeration attributeNames() {
            return del.attributeNames();
        }

        public java.lang.String getShortDescription() {
            return del.getShortDescription();
        }

        public java.lang.String getName() {
            return del.getName();
        }

        public void setHidden(boolean param) {
            del.setHidden(param);
        }

        public void setDisplayName(java.lang.String str) {
            del.setDisplayName(str);
        }

        public boolean isPreferred() {
            return del.isPreferred();
        }

        public java.lang.Object getValue(java.lang.String str) {
            return del.getValue(str);
        }

        public void setPreferred(boolean param) {
            del.setPreferred(param);
        }

        public boolean canIndexedRead () {
            return del.canIndexedRead ();
        }

        public Class getElementType () {
            return del.getElementType ();
        }

        public Object getIndexedValue (int index) throws
        IllegalAccessException, IllegalArgumentException, InvocationTargetException {
            return del.getIndexedValue (index);
        }

        public boolean canIndexedWrite () {
            return del.canIndexedWrite ();
        }

        public void setIndexedValue (int indx, Object val) throws
        IllegalAccessException, IllegalArgumentException, InvocationTargetException {
            del.setIndexedValue (indx, val);
            t.schedule (SAVE_DELAY);
        }

        public PropertyEditor getIndexedPropertyEditor () {
            return del.getIndexedPropertyEditor ();
        }
    } // end of I
    
}
