/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.awt.event.ActionEvent;
import java.io.*;
import java.lang.ref.*;
import java.net.*;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.Component;
import javax.swing.JComponent;

import org.openide.*;
import org.openide.loaders.*;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.*;

/** Wizard for creation of new objects from a template.
*
* @author Jaroslav Tulach
*/
public class TemplateWizard extends WizardDescriptor {
    /** EA that defines the wizards description */
    private static final String EA_DESCRIPTION = "templateWizardURL"; // NOI18N
    /** EA that defines custom iterator */
    private static final String EA_ITERATOR = "templateWizardIterator"; // NOI18N
    /** EA that defines resource string to the description instead of raw URL
     * @deprecated
     */
    private static final String EA_DESC_RESOURCE = "templateWizardDescResource"; // NOI18N

    /** See org.openide.WizardDescriptor.PROP_CONTENT_SELECTED_INDEX
     */
    private static final String PROP_CONTENT_SELECTED_INDEX = "WizardPanel_contentSelectedIndex"; // NOI18N
    /** See org.openide.WizardDescriptor.PROP_CONTENT_DATA
     */
    private static final String PROP_CONTENT_DATA = "WizardPanel_contentData"; // NOI18N

    /** prefered dimmension of the panels */
    static java.awt.Dimension PREF_DIM = new java.awt.Dimension (560, 350);

    /** panel */
    private Panel templateChooser;
    /** panel */
    private Panel targetChooser;
    /** whether to show target chooser */
    private boolean showTargetChooser = true;
    
    /** Iterator for the targetChooser */
    private Iterator targetIterator;
    /** whole iterator */
    private TemplateWizardIterImpl iterator;

    /** values for wizards */
    private DataObject template;
    
    /** root of all templates */
    private DataFolder templatesFolder;

    /** class name of object to create */
    private String className = null; //org.openide.util.NbBundle.getBundle(TemplateWizard.class).getString("LAB_TemplateClassName");
    /** package name */
    private String packageName = org.openide.util.NbBundle.getBundle(TemplateWizard.class).getString("LAB_TemplatePackageName");
    /** file system to create object on Reference (FileSystem) */
    private Reference system = new WeakReference (null);

    /** This is true if we have already set a value to the title format */
    private boolean titleFormatSet = false;

    /** listenes on property (steps, index) changes and updates steps pane */
    private PropertyChangeListener pcl;
    
    /** Component which we are listening on for changes of steps */
    private Component lastComp;
    
    /** Creates new TemplateWizard */
    public TemplateWizard () {
        this (new TemplateWizardIterImpl ());
    }
     
    /** Constructor to be called from public default one.
    */
    private TemplateWizard (TemplateWizardIterImpl it) {
        super (it);
        
        this.iterator = it;

        putProperty("WizardPanel_autoWizardStyle", Boolean.TRUE); // NOI18N
        putProperty("WizardPanel_contentDisplayed", Boolean.TRUE); // NOI18N
        putProperty("WizardPanel_contentNumbered", Boolean.TRUE); // NOI18N
        setTitle(org.openide.util.NbBundle.getBundle(TemplateWizard.class).getString("CTL_TemplateTitle")); //NOI18N
        setTitleFormat(new java.text.MessageFormat("{0}")); // NOI18N
    }

    /** Constructor
     *  For wizards that requires the target chooser or template
     *  chooser panel.
     *  @param TemplateIterator.Iterator iterator panel iterator instance
     */
    protected TemplateWizard (TemplateWizard.Iterator it) {
        super (it);
    }
    
    /** Initializes important settings.
     */
    protected void initialize () {
        if (iterator != null) {
            iterator.initialize(this);
        }
        super.initialize ();
    }

    /** Getter for current name.
    * @return the name or null if default name should be used
    */
    final String getClassName () {
        return className;
    }

    /** Setter for class name.
    * @return name name or null if default name should be used
    */
    final void setClassName (String name) {
        className = name;
    }

    /** Getter for package name.
    */
    final String getPackageName () {
        return packageName;
    }

    /** Getter for current system to create objects on
    */
    final FileSystem getSystem () {
        FileSystem fs = (FileSystem)system.get ();
        if (fs == null) {
            // search for a file system
            Enumeration en = TopManager.getDefault ().getRepository().fileSystems ();
            while (en.hasMoreElements()) {
                fs = (FileSystem)en.nextElement();
                if (!fs.isHidden()) {
                    // found first non hidden file system
                    system = new WeakReference (fs);
                    return fs;
                }
            }

            // use the default if it is the only one
            fs = TopManager.getDefault ().getRepository().getDefaultFileSystem ();
            system = new WeakReference (fs);
        }
        return fs;
    }

    /** Changes the system & name */
    final void setNameSystem (String packageName, FileSystem system) {
        this.packageName = packageName;
        this.system = new WeakReference (system);
    }

    /** This is method used by TemplateWizard1 to change the template
    */
    final void setTemplateImpl (DataObject obj, boolean notify) {
        DataObject old = template;

        if (template != obj) {
            template = obj;

            if (templateChooser != null) {
                templateChooser.readSettings(this);
            }
        }
        if (old != template) {
            Iterator it;
            if (
                obj == null ||
                (it = getIterator (obj)) == null
            ) {
                it = defaultIterator ();
            }
            this.iterator.setIterator (it, notify);
        }
    }

    /** Getter for template to create object from.
    */
    public DataObject getTemplate () {
        return template;
    }

    /** Sets the template. If under the TopManager.getPlaces ().folders ().
    * templates () directory it will be selected by the dialog.
    *
    * @param obj the template to start with
    */
    public void setTemplate (DataObject obj) {
        if (obj != null) {
            setTemplateImpl (obj, true);
        }
    }
    
    /** Setter for the folder with templates. If not specified the 
     * TopManager.getPlaces ().folders ().templates () is used.
     *
     * @param folder the root folder for all templates if null the
     *    default folder is used
     */
    public void setTemplatesFolder (DataFolder folder) {
        templatesFolder = folder;
        
        if (templateChooser != null) {
            templateChooser.readSettings(this);
        }
    }
    
    /** Getter for the folder with templates.
     * @return the folder with templates that should be used as root
     */
    public DataFolder getTemplatesFolder () {
        DataFolder df = templatesFolder;
        return df == null ? 
            TopManager.getDefault ().getPlaces().folders().templates() : df;
    }
        
        

    /** Getter for target folder. If the folder does not
    * exists it is created at this point.
    *
    * @return the target folder
    * @exception IOException if the possible creation of the folder fails
    */
    public DataFolder getTargetFolder () throws IOException {
        DataFolder folder = DataFolder.findFolder (getSystem ().getRoot());

        if (packageName.length () > 0) {
            String f = packageName.replace ('.', '/');
            folder = DataFolder.create (folder, f);
        }

        return folder;
    }

    /** Sets the target folder.
    *
    * @param f the folder
    */
    public void setTargetFolder (DataFolder f) {
        try {
            FileSystem system = f.getPrimaryFile().getFileSystem();

            String newName = f.getPrimaryFile ().getPackageName('.');

            setNameSystem (newName, system);
            if (targetChooser != null) {
                targetChooser.readSettings(this);
            }
        } catch (FileStateInvalidException ex) {
        }
    }

    /** Getter for the name of the target template.
    * @return the name
    */
    public String getTargetName () {
        return className;
    }

    /** Setter for the name of the template.
    * @param name name for the new object
    */
    public void setTargetName (String name) {
        className = name;
    }

    /** Panel that is used to choose a template.
    */
    public Panel templateChooser () {
        if (templateChooser == null) {
            synchronized (this) {
                if (templateChooser == null) {
                    templateChooser = createTemplateChooser ();
                }
            }
        }
        return templateChooser;
    }

    /** Panel that is used to choose target package and
    * name of the template.
    */
    public Panel targetChooser () {
        if (targetChooser == null) {
            synchronized (this) {
                if (targetChooser == null) {
                    targetChooser = createTargetChooser ();
                }
            }
        }
        return targetChooser;
    }
    
    /** Access method to the default iterator.
    */
    final Iterator defaultIterator () {
        if (targetIterator == null) {
            synchronized (this) {
                if (targetIterator == null) {
                    targetIterator = createDefaultIterator ();
                }
            }
        }
        return targetIterator;
    }
    
    /** Method that allows subclasses to provide their own panel
     * for choosing the template (the first panel).
     * 
     * @return the panel
     */
    protected Panel createTemplateChooser () {
        return new TemplateWizard1 ();
    }

    /** Method that allows subclasses to second (default) panel.
     * 
     * @return the panel
     */
    protected Panel createTargetChooser () {
        return showTargetChooser ? (Panel)new TemplateWizard2 () : (Panel)new ClassNamePanel ();
    }
    
    /** Allows subclasses to provide their own default iterator
     * the one that will be used if not special iterator is associated
     * with selected template.
     * <P>
     * This implementation creates iterator that just shows the targetChooser
     * panel.
     *
     * @return the iterator to use as default one
     */
    protected Iterator createDefaultIterator () {
        return new DefaultIterator ();
    }
    
    /** Chooses the template and instantiates it.
    * @return set of instantiated data objects (DataObject) 
    *   or null if user canceled the dialog
    * @exception IOException I/O error
    */
    public java.util.Set instantiate () throws IOException {
        showTargetChooser = true;
        return instantiateImpl (null, null);
    }

    /** Chooses the template and instantiates it.
    *
    * @param template predefined template that should be instantiated
    * @return set of instantiated data objects (DataObject) 
    *   or null if user canceled the dialog
    * @exception IOException I/O error
    */
    public java.util.Set instantiate (DataObject template) throws IOException {
        showTargetChooser = true;
        return instantiateImpl (template, null);
    }

    /** Chooses the template and instantiates it.
    *
    * @param template predefined template that should be instantiated
    * @param targetFolder the target folder
    *
    * @return set of instantiated data objects (DataObject) 
    *   or null if user canceled the dialog
    * @exception IOException I/O error
    */
    public java.util.Set instantiate (
        DataObject template, DataFolder targetFolder
    ) throws IOException {
        showTargetChooser = false;
        return instantiateImpl (template, targetFolder);
    }

    /** Chooses the template and instantiates it.
    * @param template predefined template or nothing
    * @return set of instantiated data objects (DataObject) 
    *   or null if user canceled the dialog
    * @exception IOException I/O error
    */
    private java.util.Set instantiateImpl (
        DataObject template, DataFolder targetFolder
    ) throws IOException {

        showTargetChooser |= targetFolder == null;
        targetChooser = null;

        while (true) {
            // it should loop only in case of IllegalStateException
                
            if (iterator != null) {
                iterator.initialize(this);
                iterator.first();
            }

            if (template != null){
                // force new template selection 
                this.template = null;
                setTemplate (template);
            }

            if (targetFolder != null) {
                setTargetFolder (targetFolder);
            }

            try {
                updateState();
                java.awt.Dialog d = TopManager.getDefault().createDialog(this);
                d.show ();
                d.dispose();

                if (lastComp != null) {
                    lastComp.removePropertyChangeListener(propL());
                    lastComp = null;
                }

                if (getValue () != CANCEL_OPTION && getValue () != CLOSED_OPTION) {
                    return handleInstantiate();
                } else {
                    return null;
                }
            } catch (IllegalStateException ise) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message(ise.getMessage()));
            } finally {
                if (iterator != null)
                    iterator.uninitialize();
            }
        }
    }

    /** Overriden to be able to set own default value for the title format.*/
    public void setTitleFormat (MessageFormat format) {
        titleFormatSet = true; // someone have set the title format
        super.setTitleFormat(format);
    }

    /** Overriden to be able to set a default value for the title format.*/
    public MessageFormat getTitleFormat () {
        if (!titleFormatSet) {
            // we want to call this just for the first time getTitleFormat was called
            // and not to call it when someone else called setTitleFormat
            setTitleFormat (new MessageFormat (org.openide.util.NbBundle.getBundle(TemplateWizard.class).getString("CTL_TemplateTitle")));
        }
        return super.getTitleFormat();
    }

    
    /** Calls iterator's instantiate. It is called when user selects
     * a option which is not CANCEL_OPTION or CLOSED_OPTION.
     */
    protected java.util.Set handleInstantiate() throws IOException {
        return iterator.getIterator ().instantiate (this);
    }
    
    /** Method to attach a description to a data object.
    * It is suggested that the URL use the <code>nbrescurrloc</code> protocol.
    * @param obj data object to attach description to
    * @param url the url with description or null if there should be
    *   no description
    * @exception IOException if I/O fails
    */
    public static void setDescription (DataObject obj, URL url) throws IOException {
        obj.getPrimaryFile().setAttribute(EA_DESCRIPTION, url);
    }

    /** Method to get a description for a data object.
    * @param obj data object to attach description to
    * @return the url with description or null
    */
    public static URL getDescription (DataObject obj) {
        URL desc = (URL)obj.getPrimaryFile().getAttribute(EA_DESCRIPTION);
        if (desc != null) return desc;
	// Backwards compatibility:
        String rsrc = (String) obj.getPrimaryFile ().getAttribute (EA_DESC_RESOURCE);
	if (rsrc != null) {
	    try {
		URL better = new URL ("nbrescurrloc:/" + rsrc); // NOI18N
		try {
		    setDescription (obj, better);
		} catch (IOException ioe) {
		    // Oh well, just ignore.
		}
		return better;
	    } catch (MalformedURLException mfue) {
		TopManager.getDefault ().notifyException (mfue);
	    }
	}
        return null;
    }

    /** Set a description for a data object by resource path rather than raw URL.
    * @param obj data object to set description for
    * @param rsrc a resource string, e.g. "com/foo/MyPage.html", or <code>null</code> to clear
    * @deprecated Use {@link #setDescription} instead.
    */
    public static void setDescriptionAsResource (DataObject obj, String rsrc) throws IOException {
        if (rsrc != null && rsrc.startsWith ("/")) { // NOI18N
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                System.err.println ("Warning: auto-stripping leading slash from resource path in TemplateWizard.setDescriptionAsResource");
            rsrc = rsrc.substring (1);
        }
        obj.getPrimaryFile ().setAttribute (EA_DESC_RESOURCE, rsrc);
    }

    /** Get a description as a resource.
    * @param obj the data object
    * @return the resource path, or <code>null</code> if unset (incl. if only set as a raw URL)
    * @deprecated Use {@link #getDescription} instead.
    */
    public static String getDescriptionAsResource (DataObject obj) {
        return (String) obj.getPrimaryFile ().getAttribute (EA_DESC_RESOURCE);
    }

    /** Allows to attach a special Iterator to a template. This allows
    * templates to completelly control the way they are instantiated.
    *
    * @param obj data object
    * @param iter TemplateWizard.Iterator to use for instantiation of this
    *    data object, or <code>null</code> to clear
    * @exception IOException if I/O fails
    */
    public static void setIterator (DataObject obj, Iterator iter)
    throws IOException {
        obj.getPrimaryFile().setAttribute(EA_ITERATOR, iter);
    }

    /** Finds a custom iterator attached to a template that should
    * be used to instantiate the object.
    * @param obj the data object
    * @return custom iterator or null
    */
    public static Iterator getIterator (DataObject obj) {
        return (Iterator)obj.getPrimaryFile ().getAttribute(EA_ITERATOR);
    }
    
    /** Overriden to add/remove listener to/from displayed component. Also make recreation
     * of steps and content.
     */
    protected void updateState() {
        super.updateState();

        if (lastComp != null)
            lastComp.removePropertyChangeListener(propL());

        // listener
        lastComp = iterator.current().getComponent();
        lastComp.addPropertyChangeListener(propL());
        
        // compoun steps pane info
        putProperty(PROP_CONTENT_SELECTED_INDEX,
            new Integer(getContentSelectedIndex()));
        if (getContentData() != null)
            putProperty(PROP_CONTENT_DATA, getContentData());
    }

    /**
     * @return String[] content taken from first panel and delegated iterator or null if
     * delegated iterator doesn't supplied content steps name
     */
    private String[] getContentData() {
        Component first = templateChooser.getComponent();
        if (iterator.current() == templateChooser)
            // return first panel steps
            return (String[])((JComponent)first).getClientProperty(PROP_CONTENT_DATA);
        String[] cd = null;
        Component c = iterator.current().getComponent();
        if (c instanceof JComponent) {
            // merge first panel name with delegated iterator steps
            Object property = ((JComponent)c).getClientProperty(PROP_CONTENT_DATA);
            if (property instanceof String[]) {
                String[] cont = (String[])property;
                cd = new String[cont.length + 1];
                cd[0] = ((String[])((JComponent)first).getClientProperty(PROP_CONTENT_DATA))[0];
                System.arraycopy(cont, 0, cd, 1, cont.length);
            }
        }
        return cd;
    }
    
    /** Returns selected item in content
     * @return int selected index of content
     */
    private int getContentSelectedIndex() {
        if (iterator.current() == templateChooser)
            return 0;
        int index = 1;
        Component c = iterator.current().getComponent();
        if (c instanceof JComponent) {
            // increase supplied selected index by one (template chooser)
            Object property = ((JComponent)c).getClientProperty(PROP_CONTENT_SELECTED_INDEX);
            if ((property instanceof Integer)) {
                index += ((Integer)property).intValue();
            }
        }
        return index;
    }
    
    /** Listens on content property changes in delegated iterator. Updates Wizard
     * descriptor properties.
     */
    private PropertyChangeListener propL() {
        if (pcl == null)
            pcl = new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent ev) {
                    if (PROP_CONTENT_SELECTED_INDEX.equals(ev.getPropertyName()))
                        putProperty(PROP_CONTENT_SELECTED_INDEX,
                            new Integer(getContentSelectedIndex()));
                    else if ((PROP_CONTENT_DATA.equals(ev.getPropertyName())) && (getContentData() != null))
                        putProperty(PROP_CONTENT_DATA, getContentData());
                }
            };
        return pcl;
    }

    /** The interface for custom iterator. Enhances to WizardDescriptor.Iterator
    * by serialization and ability to instantiate the object.
    * <P>
    * All Panels provided by this iterator will receive a TemplateWizard
    * as settings object and they are encourage to store its data by the 
    * use of <CODE>putProperty</CODE> method.
    *
    */
    public interface Iterator extends WizardDescriptor.Iterator,
        java.io.Serializable {
        /** Instantiates the template using informations provided by
        * the wizard.
        *
        * @param wiz the wizard
        * @return set of data objects that has been created (should contain
        *   at least one
        * @exception IOException if the instantiation fails
        * @exception IllegalStateException if the file name supplied
        * by the wizard is not valid.
        */
        public java.util.Set instantiate (TemplateWizard wiz)
            throws IOException;
        
        /** Initializes the Iterator after it is constructed.
         * @param wiz instance of TemplateWizard that wishes to use the Iterator.
         *  The Iterator can for example obtain {@link #targetChooser target Chooser}
         *  from the Wizard if it does not provide its own.
         */
        public void initialize(TemplateWizard wiz);
        
        /** Informs the Iterator that the TemplateWizard finished using the Iterator.
         * The main purpose of this method is to perform cleanup tasks that should
         * not be left on the garbage collector / default cleanup mechanisms.
         */
        public void uninitialize(TemplateWizard wiz);
    } // end of Iterator

    /** Implementation of default iterator.
    */
    private final class DefaultIterator extends Object
    implements Iterator {
        /** Name */
        public String name () {
            return ""; // NOI18N
        }

        /** Instantiates the template using informations provided by
        * the wizard.
        *
        * @param wiz the wizard
        * @return set of data objects that has been created (should contain
        *   at least one) 
        * @exception IOException if the instantiation fails
        */
        public java.util.Set instantiate(TemplateWizard wiz) throws IOException {
            String n = wiz.getTargetName ();
            DataFolder folder = wiz.getTargetFolder ();
            DataObject template = wiz.getTemplate ();
            DataObject obj = n == null ?
                             template.createFromTemplate (folder)
                             :
                             template.createFromTemplate (folder, n);

            // run default action (hopefully should be here)
            org.openide.nodes.Node node = obj.getNodeDelegate ();
            org.openide.util.actions.SystemAction sa = node.getDefaultAction ();
            if (sa != null) {
                sa.actionPerformed (new ActionEvent (node, ActionEvent.ACTION_PERFORMED, "")); // NOI18N
            }

            return java.util.Collections.singleton(obj);
        }
        
        /** No-op implementation.
         */
        public void initialize(TemplateWizard wiz) {
        }
        
        /** No-op implementation.
         */
        public void uninitialize(TemplateWizard wiz) {
        }
        
        /** Get the current panel.
        * @return the panel
        */
        public Panel current() {
            return targetChooser ();
        }
        
        /** Test whether there is a next panel.
        * @return <code>true</code> if so
        */
        public boolean hasNext() {
            return false;
        }
        
        /** Test whether there is a previous panel.
        * @return <code>true</code> if so
        */
        public boolean hasPrevious() {
            return false;
        }
        
        /** Move to the next panel.
        * I.e. increment its index, need not actually change any GUI itself.
        * @exception NoSuchElementException if the panel does not exist
        */
        public void nextPanel() {
            throw new java.util.NoSuchElementException ();
        }
        
        /** Move to the previous panel.
        * I.e. decrement its index, need not actually change any GUI itself.
        * @exception NoSuchElementException if the panel does not exist
        */
        public void previousPanel() {
            throw new java.util.NoSuchElementException ();
        }
        
        /** Add a listener to changes of the current panel.
        * The listener is notified when the possibility to move forward/backward changes.
        * @param l the listener to add
        */
        public void addChangeListener(javax.swing.event.ChangeListener l) {
        }
        
        /** Remove a listener to changes of the current panel.
        * @param l the listener to remove
        */
        public void removeChangeListener(javax.swing.event.ChangeListener l) {
        }
    }

    /*
      public static void main (String[] args) throws java.lang.Exception {
        TemplateWizard wiz = new TemplateWizard ();
        
        
        FileObject fo = FileSystemCapability.ALL.findResource(
          "Templates/AWTForms/Frame.java"
        );
        DataObject obj = DataObject.find (fo);

        fo = FileSystemCapability.ALL.findResource(
          "test"
        );
        DataFolder f = DataFolder.findFolder(fo);
        
        
        wiz.instantiate();
      }
    */  

}
