/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.loaders;

import java.net.URL;
import java.io.*;
import java.util.*;
import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.lang.reflect.*;
import java.text.MessageFormat;
import java.beans.*;

import javax.swing.event.*;
import javax.xml.parsers.*;


import org.xml.sax.*;
import org.xml.sax.ext.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;

import org.openide.*;
import org.openide.actions.*;
import org.openide.cookies.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.text.*;
import org.openide.util.*;
import org.openide.util.actions.SystemAction;
import org.openide.xml.*;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.CookieSet;
import org.openide.windows.CloneableOpenSupport;

/** 
 * Object that provides main functionality for xml documents.
 * These object are recognized by <var>xml</var> extension and
 * by <var>text/xml</var> MIME type. 
 * <p>
 * It is declaratively extensible by an Info {@link XMLDataObject.Info Info}. 
 * The Info is assigned to document instances using sibling <var>xmlInfo</var> file or it
 * can be {@link #registerInfo(String, Info) registered} for whole class of XML documents
 * by DOCTYPE's public ID.
 * <p>
 * 
 * @see XMLUtil
 * @see EntityCatalog
 *
 * @author  Libor Kramolis, Jaroslav Tulach, Petr Kuzel
 * @version lightweight
 */
public class XMLDataObject extends MultiDataObject {
    /** Public ID of xmlinfo dtd. */
    public static final String XMLINFO_DTD_PUBLIC_ID_FORTE = "-//Forte for Java//DTD xmlinfo//EN"; // NOI18N
    public static final String XMLINFO_DTD_PUBLIC_ID = "-//NetBeans IDE//DTD xmlinfo//EN"; // NOI18N

    /** Mime type of XML documents. */
    public static final String MIME = "text/xml";  //NOI18N
    
    /** the static instance for users that do not want to have own processor */
    private static final RequestProcessor XML_RP = new RequestProcessor ("XMLDataObject/parser"); // NOI18N

    /** Not parsed yet. Constant for getStatus method. */
    public static final int STATUS_NOT     = 0;
    /** Parsed ok. Constant for getStatus method. */
    public static final int STATUS_OK      = 1;
    /** Parsed with warnings. Constant for getStatus method. */
    public static final int STATUS_WARNING = 2;
    /** Parsed with errors. Constant for getStatus method. */
    public static final int STATUS_ERROR   = 3;
    
    /** property name of document property */
    public static final String PROP_DOCUMENT = "document"; // NOI18N

    /** property name of info property */
    public static final String PROP_INFO = "info"; // NOI18N

    /** generated Serialized Version UID */
    static final long serialVersionUID = 8757854986453256578L;

    /** xmlinfo */
    private Info info;
    
    /** has been above info obtained from xmlInfo. */    
    private boolean isXmlInfo;

    /** XmlDocument created from 'xml' file
    * Weaker reference to org.w3c.dom.Document 
    */
    private Reference xmlDocument = new SoftReference (null);

    /** XML parse error handler */
    private static ErrorPrinter errorHandler = new ErrorPrinter();
        
    /**
     * Chain of resolvers contaning all EntityResolvers registred by a user.
     */
    private static XMLEntityResolverChain chainingEntityResolver;
    

    /** the result of parsing */
    private int status;

    /** editor support */
    private EditorCookie editor = createEditorCookie();

    /** 
     * Task body triggered by file change (primaryFile() or xmlinfo) parsing document
     * for extension (info) assigment information (xmlinfo or public id)
     */
    private InfoParser infoParser = new InfoParser();

    /** 
     * Task executed XMLDataObject::Info request processor in for parse xmlinfo file.
     */
    private RequestProcessor.Task infoTask = XML_RP.create (infoParser);
    {
        infoTask.setPriority(Thread.NORM_PRIORITY - 1);
    }
    
    /** map of DTD publicID => ResInfo. */
    private static HashMap infos = new HashMap();

    /** DTD public ID or null.  */    
    private String publicID;
    private Object publicIDLock = new PublicIDLock();

    /* Lazy initialized. For logging and debugging. */
    private ErrorManager err;
    // the lock can be seamlesly shared between all instances
    private static Object emgrLock = new ErrorManagerLock();

    /** cookie set that holds cookies that has been created from
    * processor classes in Info class.
    */
    private CookieSet cachedCookies;
               

    // enable access to private method used by back compatability code
/*    static {
        try {
            Method method = XMLDataObject.class.getDeclaredMethod(
                "createEntityCatalog", new Class[0]
            );
            method.setAccessible(true);  //avoidable but practical            
        } catch (NoSuchMethodException ex) {
           System.err.println("Missing method");
        }
    }
*/
    
    /** Create new XMLDataObject
     *
     * @param fo the primary file object
     * @param loader loader of this data object
     */
    public XMLDataObject (FileObject fo, MultiFileLoader loader)
    throws DataObjectExistsException {
        super (fo, loader);
        
        fo.getParent ().addFileChangeListener (
            WeakListener.fileChange (infoParser, fo.getParent ())
        );

        status = STATUS_NOT;
        info = null;

        //remove my brother from loader recognization as standalone DataObject
        FileObject foe = FileUtil.findBrother (fo, Loader.XMLINFO_EXT);
        registerEntry (foe);

        //add editor cookie if provided
        if (editor != null) {
            getCookieSet().add(editor);
        }
        
        // set info for this file
        infoParser.resolveInfo (fo);

        
    }

    /** If the Info associated with this data object (if any) provides 
    * a subclass of Node, then this object is created to represent the
    * XML data object, otherwise DataNode is created.
    *
    * @return the node representation for this data object
    * @see DataNode
    */
    protected Node createNodeDelegate () {
        Node ownNode = (Node)findProcessorObject (Node.class);
        emgr().log ("createNodeDelegate: ownNode=" + ownNode);
        if (ownNode != null) {
            // the info class provided its own object
            return ownNode;
        } 

        Info i = info;
        if (i != null) {
            return new XMLNode (this, i.getIconBase ());
        } else {
            return new XMLNode (this, null);
        }
    }

    /** Called when the info file is parsed and the icon should change.
    * @param res resource for the icon
    */
    protected void updateIconBase (String res) {
    }

    /*
     * Wait until background parsing terminates.
     * It should terminate very early if just running, we can wait for it.
     */
    protected void handleDelete() throws IOException {

        if (infoTask.cancel() == false) {    // we catch it in queue
            infoTask.waitFinished();         // too late wait for finnish
        }
        super.handleDelete();
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (XMLDataObject.class);
    }

    /**     
    * Cookies from info are not placed into cookie set and can
    * be obtained only by invoking this method. Cookie order in info is handled 
    * consistently with CookieSet i.e. FIFO.
    * @return a cookie (instanceof cls) that has been found in info or
    * super.getCookie(cls).
    */
    public Node.Cookie getCookie (Class cls) {
        infoTask.waitFinished();

        Node.Cookie cake = null;
        CookieSet cs = cachedCookies;
        
        // try cached volatile cookies provided by Info objects
        if (cs != null) {
            cake = cs.getCookie (cls);
        }
        if (cake == null) {
            cake = getVolatileCookie (cls);
        }

        if (cake instanceof InstanceCookie) {
            cake = originCookie ((InstanceCookie)cake);
        }
        
        if (cake == null) {
            cake = super.getCookie (cls);
        }
        
        
        return cake;
    }

    /** Special support of InstanceCookie.Origin. If the Info class
     * provides InstanceCookie but not Origin, we add the origin to be
     * this data object.
     *
     * @param ic instance cookie
     * @return instance of InstanceCookie.Origin 
     */
    private InstanceCookie originCookie (final InstanceCookie ic) {
        class ICDel extends Object implements InstanceCookie.Origin {
            public String instanceName () {
                return ic.instanceName ();
            }

            public Class instanceClass ()
            throws java.io.IOException, ClassNotFoundException {
                return ic.instanceClass ();
            }

            public Object instanceCreate ()
            throws java.io.IOException, ClassNotFoundException {
                return ic.instanceCreate ();
            }
            
            public FileObject instanceOrigin () {
                return getPrimaryFile ();
            }
        };
        
        if (ic instanceof InstanceCookie.Origin) {
            return ic;
        } else {
            ICDel d = new ICDel ();
            CookieSet cs = cachedCookies;
            if (cs != null) {
                cs.add (d);
            }
            return d;
        }
    }

    /** @return a cookie (instanceof templ) from current info. */
    private Node.Cookie getVolatileCookie(Class templ) {
        if (useRegisteredInfo ()) {
            // new info has been created => clear cache
            cachedCookies = null;
        }
        
        Object o = findProcessorObject (templ);
        if (o != null) {
            CookieSet cs = cachedCookies;
            if (cs == null) {
                // not sychronized, because I am not sure whether needed
                cachedCookies = cs = new CookieSet ();
            }
            Node.Cookie nc = (Node.Cookie)o;
            cs.add (nc);
            return nc;
        }
        return null;
    }
    
    /** Finds and construct an object 
     * @param clazz the class to construct
     */
    private Object findProcessorObject (Class clazz) {
        useRegisteredInfo();

        if (info == null) return null;
        
        for (Iterator it = info.processorClasses(); it.hasNext(); ) {
        
            try {
                Class next = (Class)it.next();
                
                if (clazz.isAssignableFrom(next)) {
                    if (Processor.class.isAssignableFrom (next)) {
                        // the class implements Processor interface, so use
                        // default constructor to construct instance
                        Object o = next.newInstance ();
                        Processor proc = (Processor) o;
                        proc.attachTo (this);
                        return o;
                    } else {
                        // does not implement processor, try to search
                        // for constructor with one argument of DataObject or
                        // XMLDataObject
                        
                        Constructor[] arr = next.getConstructors();
                        for (int i = 0; i < arr.length; i++) {
                            Class[] params = arr[i].getParameterTypes();
                            if (params.length == 1) {
                                if (
                                    params[0] == DataObject.class || 
                                    params[0] == XMLDataObject.class 
                                ) {
                                    return arr[i].newInstance(
                                        new Object[] { this }
                                    );
                                }
                            }
                        }
                    }
                    throw new InternalError ("XMLDataObject processor class " + next + " invalid"); // NOI18N
                }
            } catch (InvocationTargetException e) {
                notifyEx (e);
            } catch (InstantiationException e) {
                notifyEx(e);
            } catch (IllegalAccessException e) {
                notifyEx(e);
            }
        }

        return null;
    }
    
    private void notifyEx(Exception e) {
        ErrorManager emgr = TopManager.getDefault().getErrorManager();        
        emgr.annotate(e, "Cannot resolve following class in xmlinfo.");        
        emgr.notify(e);
    }
    
    /** Allows subclasses to provide their own editor cookie.
     * @return an editor cookie to use
     */
    protected EditorCookie createEditorCookie () {
        return new XMLEditorSupport (this);
    }
    private final void addSaveCookie (SaveCookie save) {
        getCookieSet ().add (save);
    }
    private final void removeSaveCookie (SaveCookie save) {
        getCookieSet ().remove (save);
    }
    private static class XMLEditorSupport extends DataEditorSupport implements OpenCookie, EditorCookie, PrintCookie {
        public XMLEditorSupport (XMLDataObject obj) {
            super (obj, new XMLEditorEnv (obj));
            setMIMEType ("text/xml"); // NOI18N
        }
        class Save implements SaveCookie {
            public void save () throws IOException {
                saveDocument ();
                getDataObject ().setModified (false);
            }
        }
        protected boolean notifyModified () {
            if (! super.notifyModified ()) {
                return false;
            }
            if (getDataObject ().getCookie (SaveCookie.class) == null) {
                ((XMLDataObject) getDataObject ()).addSaveCookie (new Save ());
                getDataObject ().setModified (true);
            }
            return true;
        }
        protected void notifyUnmodified () {
            super.notifyUnmodified ();
            SaveCookie save = (SaveCookie) getDataObject ().getCookie (SaveCookie.class);
            if (save != null) {
                ((XMLDataObject) getDataObject ()).removeSaveCookie (save);
                getDataObject ().setModified (false);
            }
        }
        private static class XMLEditorEnv extends DataEditorSupport.Env {
            private static final long serialVersionUID = 6593415381104273008L;
            
            public XMLEditorEnv (DataObject obj) {
                super (obj);
            }
            protected FileObject getFile () {
                return getDataObject ().getPrimaryFile ();
            }
            protected FileLock takeLock () throws IOException {
                return ((XMLDataObject) getDataObject ()).getPrimaryEntry ().takeLock ();
            }
            public CloneableOpenSupport findCloneableOpenSupport () {
                return (XMLEditorSupport) getDataObject ().getCookie (XMLEditorSupport.class);
            }
        }
    }

    /** Creates w3c's document for the xml file. Either returns cached reference
    * or parses the file and creates new document.
    * 
    * @return the parsed document
    * @exception SAXException if there is a parsing error
    * @exception IOException if there is an I/O error
    */
    public final Document getDocument () throws IOException, SAXException {
        emgr().log ("getDocument");
        Document doc = (Document)xmlDocument.get ();
        if (doc != null)
            return doc;

        synchronized (this) {
            doc = (Document)xmlDocument.get ();
            if (doc != null)
                return doc;

            status = STATUS_OK;
            try {
                doc = parsePrimaryFile ();
            } catch (SAXException e) {
                status = STATUS_ERROR;
                throw e;
            } catch (IOException e) {
                status = STATUS_ERROR;
                throw e;
            }
            //        if (doc == null)      // never
            //  	status = STATUS_ERRO+++++++++++++++++++++++++++++++++++++++++R;
            // store it in the cache
            xmlDocument = new SoftReference (doc);

            return doc;
        }
    }

    /** Clears the document. Called when the document file is changed.
     */
    final void clearDocument () {
        emgr().log ("clearDocument");
        //err.notify (ErrorManager.INFORMATIONAL, new Throwable ("stack dump"));
        xmlDocument.clear ();
        firePropertyChange (PROP_DOCUMENT, null, null);
    }

    /** @return one of STATUS_XXX constants  */
    public final int getStatus () {
        return status;
    }

    /* @return info's clone. Returns null if file has no xmlinfo. */
    public final Info getInfo () {
        infoTask.waitFinished();
        if (info == null)
            return info;
        return (Info)info.clone();
    }

    /* Sets info persistently (write it to fs)*/
    public final synchronized void setInfo (Info ii) throws IOException {
        
        setInfoImpl (ii, false);
        writeInfo();
    }

    private final void setInfoImpl (Info ii, boolean fromRegistry) throws IOException {
        if (info == ii)
            return;
        if ((info != null) && info.equals (ii))
            return;
        Info prevInfo = info;
        info = ii;
        isXmlInfo = !fromRegistry;

        if (info != null) {
/*!!! done dynamically, there was a memory leak (it just add no remove)
            IOException allEx = new IOException("Collection of:");            
            ErrorManager emgr = TopManager.getDefault().getErrorManager();
            
            for (Iterator it = info.processorClasses(); it.hasNext(); ) {
            
                try {
                    Class c = (Class)it.next();
                    Object o = c.newInstance ();
                    Processor proc = (Processor)o;
                    proc.attachTo (this);
                    getCookieSet().add (proc); //??? what about remove?
                } catch (InstantiationException e) {
                    emgr.annotate(allEx, new IOException (e.getClass().getName() + ": " + e.getMessage())); // NOI18N
                } catch (IllegalAccessException e) {
                    emgr.annotate(allEx, new IOException (e.getClass().getName() + ": " + e.getMessage())); // NOI18N
                }
                
            }

            // notify only if there were some errors
            if (TopManager.getDefault().getErrorManager().findAnnotations(allEx) != null) {
                TopManager.getDefault().getErrorManager().annotate(allEx, "Can not resolve following classes in info:");
                throw allEx;
            }
*/            
            cachedCookies = null;
            updateIconBase (info.getIconBase ());
        }
        firePropertyChange (PROP_INFO, prevInfo, info);
    }

    private void writeInfo () throws IOException {
        if (info == null)
            return;

        final FileObject primary = getPrimaryFile();
        final FileObject parent = primary.getParent();
        final org.openide.filesystems.FileSystem FS = parent.getFileSystem();

        FS.runAtomicAction (new org.openide.filesystems.FileSystem.AtomicAction () {
                                public void run () throws IOException {
                                    FileLock lock = null;
                                    OutputStream os = null;

                                    FileObject infoFO = FS.find (parent.getName(), primary.getName(), Loader.XMLINFO_EXT);
                                    if (infoFO == null)
                                        infoFO = parent.createData (primary.getName(), Loader.XMLINFO_EXT);
                                    try {
                                        lock = infoFO.lock ();
                                        os = infoFO.getOutputStream (lock);
                                        PrintWriter writer = new PrintWriter (os);
                                        info.write (writer);
                                        writer.close();
                                    } finally {
                                        if (os != null)
                                            os.close ();
                                        if (lock != null)
                                            lock.releaseLock ();
                                    }
                                }
                            });
    }

    /** Parses the primary file of this data object.
    * and provide different implementation.
    *
    * @return the document in the primary file
    * @exception IOException if error during parsing occures
    */
    final Document parsePrimaryFile () throws IOException, SAXException {
        emgr().log ("parsePrimaryFile");
        String loc = getPrimaryFile().getURL().toExternalForm();

        return XMLUtil.parse(new InputSource(loc), false, false, errorHandler, getSystemResolver());
    }


    /** Return ErrorManager for this instance. */
    private ErrorManager emgr() {

        if (err == null) {
            synchronized(emgrLock) {
                if (err == null) {
                    err = TopManager.getDefault ().getErrorManager ().getInstance(
                        "org.openide.loaders.XMLDataObject[" + getPrimaryFile().getPackageNameExt ('/','.') + "]" // NOI18N
                    ); 
                }
            }
        }
    
        return err;
    }
    
    // ~~~~~~~~~~~~~~~~~~~~ Start of Utilities ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ~~~~~~~~~~~~~~~~~~ To be replaced by XMLUtil ~~~~~~~~~~~~~~~~~~~~~~
    
    
    /** Provides access to internal XML parser.
    * This method takes URL. After successful finish the
    * document tree is returned. Used non validating parser.
    *
    * @param url the url to read the file from
    * @deprecated Use {@link XMLUtil#parse(InputSource, boolean, boolean, ErrorHandler, EntityResolver) XMLUtil} instead
    * setting null error handler and validation to false.
    */
    public static Document parse (URL url) throws IOException, SAXException {
        return parse (url, errorHandler, false);
    }

    /** Provides access to internal XML parser.
    * This method takes URL. After successful finish the
    * document tree is returned. Used non validating parser.
    *
    * @param url the url to read the file from
    * @param validate if true validating parser is used
    * @deprecated Use {@link XMLUtil#parse(InputSource, boolean, boolean, ErrorHandler, EntityResolver) XMLUtil} instead
    * setting null handler.
    */
    public static Document parse (URL url, boolean validate) throws IOException, SAXException {
        return parse (url, errorHandler, validate);
    }

    /** Provides access to internal XML parser.
    * This method takes URL. After successful finish the
    * document tree is returned.
    *
    * @param url the url to read the file from
    * @param eh error handler to notify about exception
    * @deprecated Use {@link XMLUtil#parse(InputSource, boolean, boolean, ErrorHandler, EntityResolver) XMLUtil} instead
    * setting validation to false.
    */
    public static Document parse (URL url, ErrorHandler eh) throws IOException, SAXException {
        return parse (url, eh, false);
    }

    /** Factory a DocumentBuilder and let it create a org.w3c.dom.Document
    * This method takes URL. After successful finish the
    * document tree is returned.
    * A parser producing the Document has
    * set entity resolver to system entity resolver chain.
    *
    * @param url the url to read the file from
    * @param eh error handler to notify about exception
    * @param validate if true validating parser is used
    * @throw SAXException annotated if thrown due to configuration problem
    * @throw FactoryConfigurationError
    * @return org.w3c.dom.Document
    * @deprecated Use {@link XMLUtil#parse(InputSource, boolean, boolean, ErrorHandler, EntityResolver) XMLUtil} instead.     
    */
    public static Document parse (URL url, ErrorHandler eh, boolean validate) throws IOException, SAXException {
        
        DocumentBuilder builder = XMLDataObjectImpl.makeBuilder(validate);
        builder.setErrorHandler(eh);
        builder.setEntityResolver(getChainingEntityResolver());
        
        return builder.parse (new InputSource(url.toExternalForm()));

    }

    /** Creates SAX parse that can be used to parse XML files.
     * @return sax parser
     * @deprecated Use {@link XMLUtil#createXMLReader() XMLUtil} instead.
     * It will create a SAX XMLReader that is SAX Parser replacement.
     * You will have to replace DocumentHandler by ContentHandler
     * besause XMLReader accepts just ContentHandler. 
     * <p>Alternatively if not interested in new callbacks defined by
     * SAX 2.0 you can wrap returned XMLReader into XMLReaderAdapter
     * that implements Parser.
     */
    public static Parser createParser () {
        return createParser (false);
    }
   
    
    /** Factory SAX parser that can be used to parse XML files.
     * The factory is created according to javax.xml.parsers.SAXParserFactory property.
     * The parser has set entity resolver to system entity resolver chain.
     * @param validate if true validating parser is returned
     * @throw FactoryConfigurationError 
     * @return sax parser or null if no parser can be created
     * @deprecated Use {@link XMLUtil#createXMLReader(boolean,boolean ) Util} instead
     * setting ns to false.
     * For more details see {@link #createParser() createParser}
     */
    public static Parser createParser (boolean validate) {
        
        Parser parser = XMLDataObjectImpl.makeParser(validate);
        parser.setEntityResolver(getChainingEntityResolver());
        return parser;
        
    }


    /** 
     * Creates empty DOM Document using JAXP factoring.
     * @return Document or null on problems with JAXP factoring
     * @deprecated Replaced with {@link XMLUtil#createDocument(String,String,String,String) XMLUtil}
     *             It directly violates DOM's root element reference read-only status.
     *             If you can not move to XMLUtil for compatabilty reasons please
     *             replace with following workaround:
     * <pre>
     * String templ = "<myroot/>";
     * InputSource in = new InputSource(new StringReader(templ));
     * in.setSystemId("StringReader");  //workaround
     * DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
     * Document doc = builder.parse(in);
     * </pre>
     */
    public static Document createDocument() {
        
        deprecated();
        
        try {
            return XMLDataObjectImpl.makeBuilder(false).newDocument();
        } catch (IOException ex) {
            return null;
        } catch (SAXException ex) {
            return null;
        }
    }

    /** 
     * Writes DOM Document to writer. 
     *
     * @param doc DOM Document to be written
     * @param writer OutoutStreamWriter preffered otherwise
     *        encoding will be left for implementation specific autodection
     *
     * @deprecated Encoding used by Writer
     * may be in direct conflict with encoding
     * declared in document. Replaced with {@link XMLUtil#write(Document, OutputStream, String) Util}.
     */
    public static void write (Document doc, Writer writer) throws IOException {

        deprecated();

        // WARNING: back compatability code
        
        //use reflection to access "friendly" implementation in other package
        
        final String FAILURE = "org.openide.xml.XMLUtilImpl.write() invocation failed.";  //NOI18N
        
        try {
            Class clzz = Class.forName("org.openide.xml.XMLUtilImpl");  //NOI18N

            Method impl = clzz.getDeclaredMethod("write", new Class[] {//NOI18N
                Document.class, Object.class, String.class
            }); 
            impl.setAccessible(true);
            impl.invoke(null, new Object[] {doc, writer, null});                    

        } catch (IllegalAccessException ex) {
            throw new IOException(FAILURE);
        } catch (IllegalArgumentException ex) {
            throw new IOException(FAILURE);
        } catch (NoSuchMethodException ex) {
            throw new IOException(FAILURE);
        } catch (ClassNotFoundException ex) {
            throw new IOException(FAILURE);
        } catch (InvocationTargetException ex) {
            Throwable t = ex.getTargetException();
            if (t instanceof IOException) {
                throw (IOException) t;
            } else if (t instanceof RuntimeException) {
                throw (RuntimeException) t;
            } else if (t instanceof Error) {
                throw (Error) t;
            }
            throw new IOException(FAILURE);
        }
    }

    /**
     * Write Document into OutputStream using given encoding. 
     * It is a shortcut for writing configurations etc. It guarantee 
     * just that data will be written. Structure and indentation
     * may change.
     *
     * @param doc DOM Document to be written
     * @param out data sink     
     * @param enc - XML defined encoding name (i.e. IANA defined, one of UTF-8, UNICODE, ASCII).
     * @deprecated Moved to {@link XMLUtil#write(Document, OutputStream, String) XMLUtil}.
     */
    public static void write(Document doc, OutputStream out, String enc) throws IOException {
        XMLUtil.write(doc, out, enc);
    }
    
    
    /** 
     * Creates SAX InputSource for specified URL 
     * @deprecated Deprecated as it was a workaround method. Replace
     * with <code>new InputSource(url.toExternalForm())</code>.
     */
    public static org.xml.sax.InputSource createInputSource (URL url) throws IOException {                
        return new InputSource(url.toExternalForm());
    }

    /**
     * Registers the given public ID as corresponding to a particular 
     * URI, typically a local copy.  This URI will be used in preference
     * to ones provided as system IDs in XML entity declarations.  This
     * mechanism would most typically be used for Document Type Definitions
     * (DTDs), where the public IDs are formally managed and versioned.
     *
     * <P> Any created parser use global entity resolver and you can
     * register its catalog entry.
     *
     * @param publicId The managed public ID being mapped
     * @param uri The URI of the preferred copy of that entity
     *
     * @deprecated use module's XMLlayer instead registering at "Services" folder
     *             an XML document with EntityCatalog.PUBLIC_ID      
     */
    public static void registerCatalogEntry (String publicId, String uri) {
        
        if (publicId == null) 
            throw new IllegalArgumentException("null public ID is not allowed."); //NOI18N

        XMLDataObjectImpl.registerCatalogEntry(publicId, uri);
        
    }

    /**
     * Registers a given public ID as corresponding to a particular Java
     * resource in a given class loader, typically distributed with a
     * software package.  This resource will be preferred over system IDs
     * included in XML documents.  This mechanism should most typically be
     * used for Document Type Definitions (DTDs), where the public IDs are
     * formally managed and versioned.
     *
     * <P> If a mapping to a URI has been provided, that mapping takes
     * precedence over this one.
     *
     * <P> Any created parser use global entity resolver and you can
     * register its catalog entry.
     *
     * @param publicId The managed public ID being mapped
     * @param resourceName The name of the Java resource
     * @param loader The class loader holding the resource, or null if
     *  it is a system resource.
     *
     * @deprecated use module's XMLlayer instead registering at "Services" folder
     *             an XML document with EntityCatalog.PUBLIC_ID 
     */
    public static void registerCatalogEntry (String publicId, String resourceName, ClassLoader loader) {
        if (publicId == null) 
            throw new IllegalArgumentException("null public ID is not allowed."); //NOI18N

        XMLDataObjectImpl.registerCatalogEntry(publicId, "nbres:/" + resourceName);  //NOI18N
    }

    /**
     * Add a given entity resolver to IDE resolver chain.
     * The resolver chain is searched by private chaining resolver
     * until some registered resolver succed.
     *
     * <P>Every created parser use global entity resolver and then chain.
     *
     * @deprecated EntityResolver is a parser user responsibility. 
     *             Every time set a EntityResolver to an XML parser you use.
     *             A utility class {@link EntityResolverChain EntityResolverChain}
     *             is provided. IDE also defines default entity catalog reachable
     *             at {@link EntityCatalog#getDefault() EntityCatalog}.
     *
     * @param resolver non null resolver to be added
     *
     * @return true if successfully added
     */
    public static final boolean addEntityResolver(EntityResolver resolver) {
        return getChainingEntityResolver().addEntityResolver(resolver);
    }

    /**
     * Remove a given entity resolver from IDE resolver chain.
     *
     * <P>Every created parser use global entity resolver and then chain.
     *
     * @deprecated EntityResolver is a parser user responsibility.
     *
     * @param resolver non null resolver to be removed
     * @return removed resolver instance or null if not present
     */    
    public static final EntityResolver removeEntityResolver(EntityResolver resolver) {
        return getChainingEntityResolver().removeEntityResolver(resolver);
    }


    /** Accessor method for chaining entity resolver implementation. */
    private static synchronized XMLEntityResolverChain getChainingEntityResolver() {

        if (chainingEntityResolver == null) {                    
            chainingEntityResolver = new XMLEntityResolverChain();
            chainingEntityResolver.addEntityResolver(getSystemResolver());
        }
        
        return chainingEntityResolver;
        
    }
    
    /** Lazy initialized system resolver. */
    private static EntityResolver getSystemResolver() {        
        return  EntityCatalog.getDefault();
    }
    
    /**
    * Registers new Info to particular XML document content type as 
    * recognized by DTD public id. The registration is valid until end.
    *
    * @param publicId used as key
    * @param info associated value or null to unregister
    */    
    public static void registerInfo (String publicId, Info info) {
        synchronized (infos) {
            if (info == null) {
                infos.remove(publicId);
            } else {
                infos.put(publicId, info);
            }
        }
    }

    /**
    * Obtain registered Info for particular DTD public ID.
    *
    * @param publicId key which value is required
    * @return Info clone that is used for given publicId or null
    */
    public static Info getRegisteredInfo(String publicId) {
        synchronized (infos) {
            Info ret = (Info) infos.get(publicId);
            return ret == null ? null : (Info) ret.clone();
        }
    }

    
    // ~~~~~~~~~~~~~~~~~~~ PRIVATE AREA ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // dangerous, enter on your own risk
    
    /** Used for creating RuntimeCatalog instance */
    private static EntityCatalog createEntityCatalog() {
        return XMLDataObjectImpl.createEntityCatalog();
    }
    
    private void setPublicID(String id) {
        synchronized (publicIDLock) {
            publicID = id;        
        }
    }        
    
    private String getPublicID() {
        synchronized (publicIDLock) {
            return publicID;
        }
    }

    /** Just give better name to this lock at thread dump. */
    private static class PublicIDLock {}

    /** lazy recognizition of changed info of a public ID 
    * @return true if the info has been updated
    */
    private boolean useRegisteredInfo() {
        
        //info get from xmlInfo file or set via API, it prevails
        if (isXmlInfo) return false;

        //info get from registry
        infoTask.waitFinished();
        
        Info newInfo = getRegisteredInfo(getPublicID());
        if ((info != null && !info.equals(newInfo)) || (newInfo != info)) {
            try {
                setInfoImpl(newInfo, true);
                return  true;
            } catch (IOException ex) {
                emgr().notify(ex);
            }
        }        
        
        return false;
    }                    

    /**
      * Guess from stack trace who calls deprecated method and print it.
      */
    private static void deprecated() {
        StringWriter wr = new StringWriter();
        PrintWriter pr = new PrintWriter(wr);
        new Exception("").printStackTrace(pr);
        pr.flush();
        String stack = wr.toString().trim();

        int start = stack.indexOf("\n");
        int end = stack.indexOf("\n", start + 1);     
        while (stack.indexOf("XMLDataObject", start + 1 )>0) {        
            start = end;
            end = stack.indexOf("\n", start + 1);     
        }        

        String line =  stack.substring(start + 1, end).trim();

//        System.out.println("Warning: deprecated method called " + line);
    }
    
    /**
     * Default ErrorHandler reporting to log.
     */
    static class ErrorPrinter implements org.xml.sax.ErrorHandler {
        
        private void message (final String level, final org.xml.sax.SAXParseException e) {
            
            final String msg = NbBundle.getMessage(
                XMLDataObject.class,
                "PROP_XmlMessage",  //NOI18N
                new Object [] {
                    level,
                    e.getMessage(),
                    e.getSystemId() == null ? "" : e.getSystemId(), // NOI18N
                    "" + e.getLineNumber(), // NOI18N
                    "" + e.getColumnNumber() // NOI18N
                }
            );
                
            TopManager.getDefault().getErrorManager().log(msg);
        }

        public void error (org.xml.sax.SAXParseException e) {
            message (NbBundle.getMessage(XMLDataObject.class, "PROP_XmlError"), e);  //NOI18N
        }

        public void warning (org.xml.sax.SAXParseException e) {
            message (NbBundle.getMessage(XMLDataObject.class, "PROP_XmlWarning"), e); //NOI18N
        }

        public void fatalError (org.xml.sax.SAXParseException e) {
            message (NbBundle.getMessage(XMLDataObject.class, "PROP_XmlFatalError"), e); //NOI18N
        }
    } // end of inner class ErrorPrinter
    
    
    //~~~~~~~~~~~~~~~~~~~~~~~~ PARSER ----------------------------------

    // internally stops documet parsing when looking for public id
    static class StopSaxException extends SAXException {
        public StopSaxException() { super("STOP"); } //NOI18N
    }

    // static fields that that are logically a part of InfoParser
    
    private static final StopSaxException STOP = new StopSaxException();
    
    /** We are guaranteed to be executed in one thread let reuse parser, etc. */
    private static XMLReader sharedParserImpl = null;
   
    static {
        try {
            sharedParserImpl = XMLUtil.createXMLReader();        
            sharedParserImpl.setEntityResolver(new EmptyEntityResolver());
        } catch (SAXException ex) {
            ErrorManager err = TopManager.getDefault().getErrorManager();
            err.annotate(ex, "System does not contain JAXP 1.1 compliant parser!");
            err.notify(err.ERROR, ex);            
        }
        
        
        //initialize stuff possibly needed by libs that do not use 
        //JAXP but SAX 2 directly
        try {
            final Properties props = System.getProperties();
            final String SAX2_KEY = "org.xml.sax.driver";  //NOI18N
            if (props.getProperty(SAX2_KEY) == null) {
                props.put(SAX2_KEY, sharedParserImpl.getClass().getName());                
            }
        } catch (RuntimeException ex) {
            //ignore it (we did the best efford)
        }
    }
    
    /** 
     * Parser that assigns info *or* public ID.
     * Implementation assumes that all parsing tasks are enqueued into single request processor.
     */
    private final class InfoParser
        implements Runnable, FileChangeListener {
            
        /** the name of info tag */
        private static final String TAG_INFO = "info"; // NOI18N
        /** the name of processor tag */
        private static final String TAG_PROCESSOR = "processor"; // NOI18N
        /** the class attribute */
        private static final String ATT_PROCESSOR_CLASS = "class"; // NOI18N
        /** icon tag */
        private static final String TAG_ICON = "icon"; // NOI18N
        /** base attribute */
        private static final String ATT_ICON_BASE = "base"; // NOI18N

        /** file object of xmlinfo to parse or null */
        private FileObject infoFileObject;
        
        /** 
          * Changes the info for a file that starts new parsing.
          * Dispatch change to my thread.
          * @param fo primary fileobject to resolve.
          */
        public synchronized void resolveInfo (FileObject fo) {
            // synchronized to allow only one parsing to be running at one time.
            // wait till previous parsing finishes
            //infoTask.waitFinished ();

            if (Loader.XMLINFO_EXT.equals(fo.getExt())) {
                infoFileObject = null; //xmlInfo could use info assigned via registry
            } else {
                infoFileObject = FileUtil.findBrother (fo, Loader.XMLINFO_EXT);
            }
            
            // start new parsing enqueue itself
            infoTask.schedule (0);
        }
                
        //~~~~~~~~~~~~~~~~~~~~~ Task body ~~~~~~~~~~~~~~~~~~~
        
        /** 
          * Starts the parsing in info task request processor thread.
          */
        public void run () {
            if (infoFileObject != null) {
                parseInfo();
            } else {
                parseHeader();
            }      
        }

        /*
         * Parse xmlinfo, errors are not so critical.
         */
        private void parseInfo() {

            URL url = null;
            XMLReader parser = sharedParserImpl;
            FileLock infoFileLock = null;
            InfoHandler handler = new InfoHandler();

            try {

                //infoFileLock = infoFileObject.lock();  //??? should I lock when reading

                try {                                        
                    url = infoFileObject.getURL();
                } catch (IOException ex) {
                    warning(ex, "I/O exception while retrieving xmlinfo URL."); //NOI18N
                    return;  // info will be ignored
                }

                configureParser(parser, true, null);
                parser.setContentHandler(handler);
                parser.setErrorHandler(handler);

                try {
                    parser.parse(new InputSource(url.toExternalForm()));
                } catch (SAXException ex) {
                    warning(ex, "Unexpected SAX exception in InfoHandler while parsing xmlinfo."); //NOI18N
                } catch (IOException ex) {
                    warning(ex, "I/O exception while parsing xmlinfo.");  //NOI18N
                }

                try {
                    XMLDataObject.this.setInfoImpl (handler.getInfo(), false);
                } catch (Exception ex) {
                    warning(ex, "Can not assign info");  //NOI18N
                }

            } finally {

                if (infoFileLock != null) {
                    infoFileLock.releaseLock();
                }
            }


        }

        /*
         * Find out DTD public ID.
         * Info is then assigned according to it from registry.
         */
        private void parseHeader() {

            URL url = null;
            InputStream in = null;
            XMLReader parser = sharedParserImpl;
            FileObject myFileObject = getPrimaryFile();
            FileLock myFileLock = null;

            try {

                //myFileLock = myFileObject.lock();  //??? should I lock when reading

                DocListener lis = new DocListener();
                lis.reset();

                try {
                    url = myFileObject.getURL();
                    in =  myFileObject.getInputStream();
                } catch (IOException ex) {
                    warning(ex, "I/O exception while retrieving xml FileObject URL."); //NOI18N
                    return;  // cannot parse
                }

                configureParser(parser, false, lis);
                parser.setContentHandler(lis);
                parser.setErrorHandler(lis);                    

                try {

                    InputSource input =  new InputSource(url.toExternalForm());
                    input.setByteStream(in);
                    parser.parse(input);

                } catch (StopSaxException stopped) {
                    try {
                        XMLDataObject.this.setPublicID(lis.getPublicID());
                        Info tempInfo = getRegisteredInfo(lis.getPublicID());
                        XMLDataObject.this.setInfoImpl (tempInfo, true);

                    } catch (IOException ex) {
                        // the exception will be notified for every document
                        // that uses public id that has registered wrong info
                        emgr().notify(ex);
                    }
                } catch (SAXException checkStop) {
                    // stop parsing anyway
                    emgr().annotate(checkStop, "DocListener should not throw SAXException but STOP one.");  //NOI18N
                    emgr().notify(emgr().INFORMATIONAL, checkStop);
                } catch (FileNotFoundException ex) {
                    // thrown when there is a problem with URL for example
                    emgr().notify(emgr().INFORMATIONAL, ex);
                } catch (IOException ex) {
                    // error while parsing for public id hide it.
                    // somebody have deleted the file meanwhile, because I do not lock?
                    emgr().notify(emgr().INFORMATIONAL, ex);
                } finally {
                    parser = null;
                    lis = null;
                }

            } finally {
                if (myFileLock != null) {
                    myFileLock.releaseLock();
                }
                try {
                    if (in != null) in.close();
                } catch (IOException ex) {
                    emgr().notify(emgr().INFORMATIONAL, ex);
                }
            }
        }


        /*
         * We reuse the parser so it must be reconfigured prior every parsing task.
         */
        private void configureParser(XMLReader parser, boolean validation, LexicalHandler lex) {

            try {
                parser.setFeature("http://xml.org/sax/features/validation", validation);  //NOI18N
            } catch (SAXException sex) {
                emgr().log("Warning: XML parser does not support validation feature.");  //NOI18N                    
            }

            try {
                parser.setProperty("http://xml.org/sax/properties/lexical-handler", lex);  //NOI18N
            } catch (SAXException sex) {
                emgr().log("Warning: XML parser does not support lexical-handler feature.");  //NOI18N
                //throw new Error("");
            }
        }


        // ~~~~~~~~~~ ERROR REPORTING ~~~~~~~~~~~~~~~~~~~~~~

        public void warning (Throwable ex) {
            warning(ex, null);
        }

        public void warning (Throwable ex, String annotation) {
            ErrorManager emgr = emgr();
            if (annotation != null)
                emgr.annotate(ex, annotation);
            emgr.notify(ErrorManager.INFORMATIONAL, ex);  //do not show until in debug mode
        }
            
        /**
          * Content & error handler recognizing .xmlinfo grammar.
          */ 
        private class InfoHandler extends DefaultHandler {            

            private Info tempInfo = null;

            Info getInfo() {
                return tempInfo;
            }

            /** Must be called before reuse. */
            void reset() {
                tempInfo = new Info();
            }

            /** Accepts module item */
            public void startElement (String uri, String lName, String qName, Attributes attr) {
                if (qName.equals (TAG_PROCESSOR)) {
                    String className = attr.getValue (ATT_PROCESSOR_CLASS);
                    if (className != null) {
                        try {
                            Class c = Class.forName (className, true, TopManager.getDefault().systemClassLoader ());
                            tempInfo.addProcessorClass (c);
                        } catch (Exception e) {
                            InfoParser.this.warning(e, "Can not load class " + className);  //NOI18N
                        }
                    }
                    return;
                }
                if (qName.equals (TAG_ICON)) {
                    String file = attr.getValue (ATT_ICON_BASE);
                    tempInfo.setIconBase (file);
                    return;
                }
            }

            public void error(final SAXParseException ex) throws SAXException {
                InfoParser.this.warning(ex, "Validity error in: " + infoFileObject);  //NOI18N            
                tempInfo = null;
                throw ex; // stop parsing
            }

            public void fatalError(final SAXParseException ex) throws SAXException {

                InfoParser.this.warning(ex, "Wellformess error in: " + infoFileObject);  //NOI18N
                tempInfo = null;
                throw ex; //stop parsing
            }

        }


        /**
          * Content & error handler and lexical listener looking up for DTD public id.
          */ 
        private class DocListener extends DefaultHandler implements LexicalHandler {

            String publicId = null;

            /** @return  recognized system id. */
            String getPublicID() {
                return publicId;
            }

            /** Must be called before reuse. */
            void reset() {
                publicId = null;
            }

            // LexicalHandler

            public void startDTD(String root, String pID, String sID) throws SAXException {
                publicId = pID;
                stop();
            }

            public void endDTD() throws SAXException {
                stop();
            }

            public void startEntity(String name) throws SAXException {
            }

            public void endEntity(String name) throws SAXException {
            }

            public void startCDATA() throws SAXException {
            }

            public void endCDATA() throws SAXException {                
            }

            public void comment(char[] ch, int start, int length) throws SAXException {
            }

            // redefine DefaultHandler

            //!!! should we stop on error?
            public void error(final org.xml.sax.SAXParseException p1) throws org.xml.sax.SAXException {
                stop();
            }

            public void fatalError(final org.xml.sax.SAXParseException p1) throws org.xml.sax.SAXException {
                stop();
            }

            public void endDocument() throws org.xml.sax.SAXException {
                stop();
            }

            public void startElement(String uri, String lName, String qName, Attributes atts) throws org.xml.sax.SAXException {
                // no DTD present
                stop(); 
            }

            private void stop() throws SAXException {
                throw STOP;
            }

        }                            

        
        //~~~~~~~~~~~~~~~~~~ FS LISTENER ~~~~~~~~~~~~~~~~~
        //listening at parent folder
        
        public void fileFolderCreated (FileEvent fe) {
            // not interesting
        }

        public void fileDataCreated (FileEvent fe) {
            FileObject fo = fe.getFile();
            fileCreated(fo);
        }

        private void fileCreated(FileObject fo) {
            if (
                fo.hasExt (Loader.XMLINFO_EXT) &&
                fo.getName ().equals (getPrimaryFile ().getName ())
            ) {
                // new info file created => force it to be reparsed
                resolveInfo(getPrimaryFile());
            }
        }

        /** Fired when a file is changed.
        * @param fe the event describing context where action has taken place
        */
        public void fileChanged (FileEvent fe) {
            if (fe.getFile().equals (infoFileObject)) {
                // repase the file
                resolveInfo(getPrimaryFile ());  //reparse info again
            } else {
                if (getPrimaryFile ().equals (fe.getFile ())) {
                    // the main file changed => invalidate DOM document
                    //resolveInfo (getPrimaryFile ());  //reparse info again                    
                    clearDocument ();
                }
            }
        }

        public void fileDeleted (FileEvent fe) {
            FileObject fo = fe.getFile();

            if (fo.equals (infoFileObject)) {
                // reparse the file
                resolveInfo(getPrimaryFile ());
            }
        }

        public void fileRenamed (FileRenameEvent fe) {
            // the same behaviour as when the file is deleted
            fileDeleted (fe);
            // and new created
            fileCreated(fe.getFile());
        }

        public void fileAttributeChanged (FileAttributeEvent fe) {
            // not interested in
        }
        
    } // end of InfoParser


    /** Avoid Internet connections */
    private static class EmptyEntityResolver implements EntityResolver {
        public InputSource resolveEntity(String publicId, String systemID) {
            InputSource ret = new InputSource(new StringReader(""));
            ret.setSystemId("StringReader");  //NOI18N
            return ret;
        }
    }
    
    
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~ private Loader ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
        
    /** The DataLoader for XmlDataObjects.
     * This class is final only for performance reasons,
     * can be happily unfinaled if desired.
     */
    static class Loader extends MultiFileLoader {
        /** Extension constants */
        static final String XML_EXT = "xml"; // NOI18N
        static final String XMLINFO_EXT = "xmlinfo"; // NOI18N

        static final long serialVersionUID =3917883920409453930L;
        /** Creates a new XMLDataLoader */
        public Loader () {
            super (XMLDataObject.class);
        }

        /** Initialize XMLDataLoader: name, actions, ...
        */
        protected void initialize () {            
        }

        /** Get default actions.
        * @return array of default system actions or <CODE>null</CODE> if this loader does not have any
        *   actions
        */
        protected SystemAction[] defaultActions () {
            return new SystemAction[] {
                           SystemAction.get(OpenAction.class),
                           SystemAction.get(FileSystemAction.class),
                           null,
                           SystemAction.get(CutAction.class),
                           SystemAction.get(CopyAction.class),
                           SystemAction.get(PasteAction.class),
                           null,
                           SystemAction.get(DeleteAction.class),
                           SystemAction.get(RenameAction.class),
                           null,
                           SystemAction.get(SaveAsTemplateAction.class),
                           null,
                           SystemAction.get(ToolsAction.class),
                           SystemAction.get(PropertiesAction.class)
                       };
        }
        
        /** Get the default display name of this loader.
        * @return default display name
        */
        protected String defaultDisplayName () {
            return NbBundle.getBundle (XMLDataObject.class).getString ("PROP_XmlLoader_Name");
        }
        
        /** For a given file finds a primary file.
        * @param fo the file to find primary file for
        *
        * @return the primary file for the file or null if the file is not
        *   recognized by this loader
        */
        protected FileObject findPrimaryFile (FileObject fobj) {
            FileObject fo = accept(fobj);
            if (fo == null) return null;
            return fo;
                        
        }

        /** @return recognized fileobject or null. */
        private FileObject accept(FileObject fo) {

            // this must have higher priority than MIME rule
            if (XMLINFO_EXT.equals(fo.getExt())) {
                FileObject f = FileUtil.findBrother (fo, XML_EXT);
                return f == null ? fo : f;
            }            
            if (MIME.equals(fo.getMIMEType())) {
                return fo;
            }
            if (XML_EXT.equals(fo.getExt())) {
                return fo;
            }            
            /** @deprecated register MIME instead. */
            if ("tld".equals(fo.getExt())) { // NOI18N
                return fo; // JSP Tag Library Descriptor
            }
            // not recognized
            return null;            
        }
        
        /** Creates the right data object for given primary file.
        * It is guaranteed that the provided file is realy primary file
        * returned from the method findPrimaryFile.
        *
        * @param primaryFile the primary file
        * @return the data object for this file
        * @exception DataObjectExistsException if the primary file already has data object
        */
        protected MultiDataObject createMultiObject (FileObject primaryFile)
        throws DataObjectExistsException {
            return new XMLDataObject (primaryFile, this);
        }

        /** Creates the right primary entry for given primary file.
        *
        * @param primaryFile primary file recognized by this loader
        * @return primary entry for that file
        */
        protected MultiDataObject.Entry createPrimaryEntry (MultiDataObject obj, FileObject primaryFile) {
            return new FileEntry (obj, primaryFile);
        }

        /** Creates right secondary entry for given file. The file is said to
        * belong to an object created by this loader.
        *
        * @param secondaryFile secondary file for which we want to create entry
        * @return the entry
        */
        protected MultiDataObject.Entry createSecondaryEntry (MultiDataObject obj, FileObject secondaryFile) {
            return new FileEntry (obj, secondaryFile);
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~ extension support via info ~~~~~~~~~~~~~~~~~~~~~~~~
    // i would like to throw it away sometimes in future and replace it by better one

    /** This class has to be implemented by all processors in the
    * xmlinfo file. It is cookie, so after parsing such class is instantiated
    * and put into data objects cookie set.
    */
    public static interface Processor extends Node.Cookie {
        /** When the XMLDataObject creates new instance of the processor,
        * it uses this method to attach the processor to the data object.
        *
        * @param xmlDO XMLDataObject
        */
        public void attachTo (XMLDataObject xmlDO);
    }
    
    
    /**
     * Representation of xmlinfo file holding container of Processors.
     */
    public static final class Info implements Cloneable {
        List processors;
        String iconBase;

        /** Create info */
        public Info () {
            processors = new OrderedSet();
            iconBase = null;
        }

        public Object clone () {
            Info ii = new Info();
            for (Iterator it = processors.iterator(); it.hasNext();) {
                Class proc = (Class)it.next();
                ii.processors.add (proc);
            }
            ii.iconBase = iconBase;
            return ii;
        }

        /** Add processor class to info. 
        * The class should be public and either implement the Processor 
        * interface or should
        * have public constructor with one argument (DataObject or XMLDataObject).
        *
        * @param proc the class to add to this info
        * @exception IllegalArgumentException if the class does not seem to be valid
        */
        public synchronized void addProcessorClass (Class proc) {
            if (!Processor.class.isAssignableFrom (proc)) {
                Constructor[] arr = proc.getConstructors();
                for (int i = 0; i < arr.length; i++) {
                    Class[] params = arr[i].getParameterTypes();
                    if (params.length == 1) {
                        if (
                            params[0] == DataObject.class || 
                            params[0] == XMLDataObject.class 
                        ) {
                            arr = null;
                            break;
                        }
                    }
                }
                
                if (arr != null) {
                    // no suitable constructor
                    throw new IllegalArgumentException();
                }
            }
            
            processors.add (proc);
        }

        /** Remove processor class from info.
         * @return true if removed
         */
        public boolean removeProcessorClass (Class proc) {
            return processors.remove (proc);
        }

        public Iterator processorClasses () {
            return processors.iterator();
        }

        /** Set icon base */
        public void setIconBase (String base) {
            iconBase = base;
        }

        /** @return icon base */
        public String getIconBase () {
            return iconBase;
        }

        /** Write specified info to writer */
        public void write (Writer writer) throws IOException {
            writer.write ("<?xml version=\"1.0\"?>\n\n"); // NOI18N
            writer.write (MessageFormat.format ("<!DOCTYPE {0} PUBLIC \"{1}\" \"\">\n\n", // NOI18N
                                                new Object [] { InfoParser.TAG_INFO, XMLINFO_DTD_PUBLIC_ID }));
            writer.write (MessageFormat.format ("<{0}>\n", // NOI18N
                                                new Object [] { InfoParser.TAG_INFO }));
            for (Iterator it = processors.iterator(); it.hasNext();)
                writer.write (MessageFormat.format ("  <{0} {1}=\"{2}\" />\n", // NOI18N
                                                    new Object [] { InfoParser.TAG_PROCESSOR,
                                                                    InfoParser.ATT_PROCESSOR_CLASS,
                                                                    ((Class)it.next()).getName() }));
            if (iconBase != null)
                writer.write (MessageFormat.format ("  <{0} {1}=\"{2}\" />\n", // NOI18N
                                                    new Object [] { InfoParser.TAG_ICON,
                                                                    InfoParser.ATT_ICON_BASE,
                                                                    iconBase }));
            writer.write (MessageFormat.format ("</{0}>\n", // NOI18N
                                                new Object [] { InfoParser.TAG_INFO }));
        }
        
        public boolean equals (Object obj) {
            if (obj == null) return false;
            if (obj instanceof Info == false) return false;
            
            Info i = (Info) obj;
            
            return ((iconBase != null && iconBase.equals(i.iconBase)) || (i.iconBase == iconBase)) 
                    && processors.equals(i.processors);
        }

        /** Set that just preserves order of elemenst added by add(). */
        private class OrderedSet extends ArrayList {
            public OrderedSet() {
                super(7);
            }
            
            public boolean add(Object object) {
                if (contains(object)) return false;
                return super.add(object);
            }
        }
        
    } // end of inner class XMLInfoCreator

    
    
    
    /** Node that listens to changes in info and in such case it updates
     * the icon base, etc.
     */
    private static class XMLNode extends DataNode {
        public XMLNode (XMLDataObject obj, String icon) {
            super (obj, Children.LEAF);
            setDefaultAction (SystemAction.get (OpenAction.class));
            update (icon);
        }
        
        /** Called when there is a property change.
         */
        void fireChange (PropertyChangeEvent ev) {
            if (PROP_INFO.equals (ev.getPropertyName())) {
                // update icon base
                XMLDataObject obj = (XMLDataObject)getDataObject ();
                Info info = obj.getInfo();
                update (info == null ? null : info.getIconBase());
                return;
            }
            
            super.fireChange (ev);
        }
        
        private void update (String iconBase) {
            if (iconBase == null) {
                setIconBase ("/org/openide/resources/xmlObject"); // NOI18N
            } else {
                setIconBase (iconBase);
            }
        }
        
    }

    /** For better monitor dumps. */
    private static class ErrorManagerLock {}
}
