/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.nodes;

import java.util.*;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.EventListenerList;

/** Support class for storing cookies and
* retriving them by representation class.
* Provides simple notifications about changes
* in cookies.
*
* @author Jaroslav Tulach
*/
public final class CookieSet extends Object {
    /** list of cookies (Class, Node.Cookie) */
    private HashMap map = new HashMap (31);

    /** set of listeners */
    private EventListenerList listeners = new EventListenerList ();

    /** Default constructor. */
    public CookieSet() {}

    /** Add a new cookie to the set. If a cookie of the same
    * <em>actual</em> (not representation!) class is already there,
    * it is replaced.
    * <p>Cookies inserted earlier are given preference during lookup,
    * in case a supplied representation class matches more than one cookie
    * in the set.
    *
    * @param cookie cookie to add
    */
    public void add (Node.Cookie cookie) {
        synchronized (this) {
            Class clazz = cookie.getClass ();
            Node.Cookie previous = (Node.Cookie)map.get (clazz);
            if (previous != null) {
                unregisterCookie (clazz, previous);
            }

            // insert the cookie to be accessible from all superclass
            Class c = cookie.getClass ();
            registerCookie (c, cookie);
        }

        fireChangeEvent ();
    }

    /** Remove a cookie from the set.
    * @param cookie the cookie to remove
    */
    public void remove (Node.Cookie cookie) {
        synchronized (this) {
            unregisterCookie (cookie.getClass (), cookie);
        }

        fireChangeEvent ();
    }

    /** Get a cookie.
    *
    * @param clazz the representation class
    * @return a cookie assignable to the representation class, or <code>null</code> if there is none
    */
    public synchronized Node.Cookie getCookie (Class clazz) {
        Node.Cookie ret = (Node.Cookie) map.get(clazz);
        if (ret instanceof CookieEntry) {
            ret = ((CookieEntry) ret).getCookie();
        }
        
        return ret;
    }

    /** Add a listener to changes in the cookie set.
    * @param l the listener to add
    */
    public void addChangeListener (ChangeListener l) {
        listeners.add (ChangeListener.class, l);
    }

    /** Remove a listener to changes in the cookie set.
    * @param l the listener to remove
    */
    public void removeChangeListener (ChangeListener l) {
        listeners.remove (ChangeListener.class, l);
    }

    /** Fires change event
    */
    private void fireChangeEvent () {
        Object[] arr = listeners.getListenerList();
        if (arr.length > 0) {
            ChangeEvent ev = new ChangeEvent (this);
            // Process the listeners last to first, notifying
            // those that are interested in this event
            for (int i = arr.length-1; i>=0; i-=2) {
                ((ChangeListener)arr[i]).stateChanged (ev);
            }
        }
    }

    /** Attaches cookie to given class and all its superclasses and
    * superinterfaces.
    *
    * @param c class or null
    * @param cookie cookie to attach
    */
    private void registerCookie (Class c, Node.Cookie cookie) {
        if ((c == null) || !nodeCookieClass.isAssignableFrom(c)) return;

        Object orig = map.put (c, cookie);
        if (orig != null) {
            // return it back and finish
            map.put (c, orig);
            return;
        }

        registerCookie (c.getSuperclass (), cookie);

        Class[] inter = c.getInterfaces ();
        for (int i = 0; i < inter.length; i++) {
            registerCookie (inter[i], cookie);
        }
    }

    private static Class nodeCookieClass = Node.Cookie.class;
    
    /** Attaches cookie to given class only.
    *
    * @param c class or null
    * @param cookie cookie to attach
    */
    /*
    private void simpleRegisterCookie (Class c, Node.Cookie cookie) {
        if (c == null) return;

        Object orig = map.put (c, cookie);
        if (orig != null) {
            // return it back and finish
            map.put (c, orig);
            return;
        }
    }*/
        
    /** Removes cookie from the class and all its superclasses and
    * superinterfaces.
    *
    * @param c class or null
    * @param cookie cookie to attach
    */
    private void unregisterCookie (Class c, Node.Cookie cookie) {
        if (c == null) return;

        // if different cookie is attached to class c stop removing
        if (cookie != map.get (c)) return;

        // remove the cookie
        map.remove (c);

        unregisterCookie (c.getSuperclass (), cookie);

        Class[] inter = c.getInterfaces ();
        for (int i = 0; i < inter.length; i++) {
            unregisterCookie (inter[i], cookie);
        }
    }

    /** Checks whether this set contains given cookie
     * @param cookie Cookie class to test
     * @return <tt>true</tt> if this set contains the cookie
     */
    /*public boolean contains(java.lang.Class cookie) {
        return map.containsKey(cookie);
    }*/
    
    /** Registers a Factory for given cookie class */
    public void add(Class cookieClass, Factory factory) {
        if (factory == null) {
            throw new IllegalArgumentException();
        }
        
        synchronized (this) {
            addCookieNoFire(cookieClass, factory);
        }

        fireChangeEvent ();
    }

    /** Registers a Factory for given cookie classes */
    public void add(Class[] cookieClass, Factory factory) {
        if (factory == null) {
            throw new IllegalArgumentException();
        }
        
        synchronized (this) {
            for (int i = 0; i < cookieClass.length; i++) {
                addCookieNoFire(cookieClass[i], factory);
            }
        }

        fireChangeEvent ();
    }    
    
    /** Adds a cookie does not fire changes */
    private void addCookieNoFire(Class cookieClass, Factory factory) {
        Node.Cookie previous = (Node.Cookie)map.get (cookieClass);
        if (previous != null) {
            unregisterCookie (cookieClass, previous);
        }

        // insert the cookie to be accessible from all superclass
        registerCookie (cookieClass, new CookieEntry(factory, cookieClass));
    }
    
    /** Factory for creating cookies of given Class */
    public static interface Factory {
        /** Creates a Node.Cookie of given class. The method
         * may be called more than once.
         */
        public Node.Cookie createCookie(Class klass);
    }
    
    /** Entry for one Cookie */
    private static class CookieEntry implements Node.Cookie {
        /** Factory for the cookie */
        private final Factory factory;
        /** Class of the cookie */
        private final Class klass;
        /** A Referenec to the cookie */
        private Reference cookie;
        
        /** Constructs new FactoryEntry */
        public CookieEntry(Factory factory, Class klass) {
            this.factory = factory;
            this.klass = klass;
        }
        
        /** Getter for the cookie */
        public Node.Cookie getCookie() {
            Node.Cookie ret;
            if ((cookie == null) || ((ret = (Node.Cookie) cookie.get()) == null)) {
                ret = factory.createCookie(klass);
                cookie = new WeakReference(ret);
            }
            
            return ret;
        }
    }
}
