/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.nodes;

import org.openide.util.Mutex;

/** Simple implementation of <code>Node.Handle</code>.
* When created by {@link #createHandle} it
* looks for the parent of the node and stores the node's name.
* When {@link #getNode} is then called, it tries to restore the
* parent and then to walk down to the child.
* <p>Note that if most nodes use <code>DefaultHandle</code>, this
* may walk up to the root recursively. Otherwise, some other sort
* of handle may provide the termination case.
*
* @author Jaroslav Tulach, Jesse Glick
*/
public final class DefaultHandle extends Object implements Node.Handle {
    /** parent handle */
    private Node.Handle parent;
    /** path to the node (just one hop, really) */
    private String path;

    private static final long serialVersionUID = -8739127064355983273L;
    /* Create a new handle.
    * @param parent handle for the parent node
    * @param path desired name of child
    */
    DefaultHandle (Node.Handle parent, String path) {
        this.parent = parent;
        this.path = path;
    }

    /** Find the node.
    * @return the found node
    * @exception IOException if the parent cannot be recreated
    * @exception NodeNotFoundException if the path is not valid (exception may be examined for details)
    */
    public Node getNode () throws java.io.IOException {
        Node parentNode = parent.getNode ();
        Node child = parentNode.getChildren ().findChild (path);
        if (child != null)
            return child;
        else
            throw new NodeNotFoundException (parentNode, path, 0);
    }

    /** Create a handle for a given node.
    * A handle cannot be created under these circumstances:
    * <ol>
    * <li>The node has no name.
    * <li>The node has no parent.
    * <li>The parent has no handle.
    * <li>The parent is incapable of finding its child by the supplied name.
    * </ol>
    * @param node the node to create a handler for
    * @return the handler, or <code>null</code> if a handle cannot be created
    */
    public static DefaultHandle createHandle (final Node node) {
        // read access to ensure that the computed parent and path
        // will be the same
        return (DefaultHandle)Children.MUTEX.readAccess (new Mutex.Action () {
                    public Object run () {
                        String childPath = node.getName ();
                        if (childPath == null) return null;
                        Node parentNode = node.getParentNode ();
                        if (parentNode == null) return null;
                        if (parentNode.getChildren ().findChild (childPath) != node) {
                            if (Boolean.getBoolean ("netbeans.debug.exceptions")) { // NOI18N
                                // PLEASE DO NOT COMMENT OUT:
                                System.err.println("WARNING: parent could not find own child in DefaultHandle.createHandle"); // NOI18N
                                System.err.println("\tnode=" + node); // NOI18N
                                System.err.println("\tparentNode=" + parentNode); // NOI18N
                                System.err.println("\tchildPath=" + childPath); // NOI18N
                                System.err.println("\tfound child=" + parentNode.getChildren ().findChild (childPath)); // NOI18N
                            }
                            return null;
                        }
                        Node.Handle parentHandle = parentNode.getHandle ();
                        if (parentHandle == null) return null;
                        return new DefaultHandle (parentHandle, childPath);
                    }
                });
    }

    public String toString () {
        return "DefaultHandle[" + parent + "|" + path + "]"; // NOI18N
    }

}

/*
* Log
*  8    Jaga      1.6.1.0     3/9/00   Jesse Glick     ExplorerManager + node 
*       persistence overhaul.
*  7    Gandalf   1.6         10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  6    Gandalf   1.5         8/9/99   Ian Formanek    Generated Serial Version 
*       UID
*  5    Gandalf   1.4         6/8/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  4    Gandalf   1.3         3/18/99  Jaroslav Tulach Compiles under jikes.
*  3    Gandalf   1.2         3/17/99  Jesse Glick     Constructor ought not to 
*       have been public.
*  2    Gandalf   1.1         3/17/99  Jesse Glick     [JavaDoc]
*  1    Gandalf   1.0         1/5/99   Ian Formanek    
* $
*/
