/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.util;

import java.awt.Image;
import java.awt.Toolkit;
import java.util.HashMap;
import java.lang.ref.*;

import org.openide.util.Utilities;

/** Registers all loaded images into the AbstractNode, so nothing is loaded twice.
*
* @author Jaroslav Tulach,
*/
class IconManager implements Runnable {
    /** a value that indicates that the icon does not exists */
    private static final Object NO_ICON = new Object();

    /** map of resource name to loaded icon (String, SoftRefrence (Image)) or (String, NO_ICON) */
    private static final HashMap map = new HashMap ();

    private static final HashMap reverseMap = new HashMap();
    
    private static final ReferenceQueue compositeQueue = new ReferenceQueue();
    
    private static Task  cleaningTask;
    
    /**
     * Interval for regular cache clean-ups, in milliseconds.
     */
    private static final int CACHE_CLEAN_INTERVAL = 10000;

    static Image getIcon (String name) {
        ClassLoader loader =  (ClassLoader)Lookup.getDefault().lookup(ClassLoader.class);
        return getIcon(name, loader);
    }
    
    private static void startCleaner() {
        if (cleaningTask != null)
            return;
        cleaningTask = RequestProcessor.postRequest(new IconManager(), CACHE_CLEAN_INTERVAL, Thread.MIN_PRIORITY);
    }
    
    /** Finds imager for given resource.
    * @param name name of the resource
    * @param loader classloader to use for locating it
    */
    static Image getIcon (String name, ClassLoader loader) {
        Object img = map.get (name);

        // no icon for this name (already tested)
        if (img == NO_ICON) return null;

        if (img != null) {
            // then it is SoftRefrence
            img = ((Reference)img).get ();
        }

        // icon found
        if (img != null) return (Image)img;
        
        cleanUpMap();

        synchronized (map) {
            // again under the lock
            img = map.get (name);
            if (img != null) {
                // then it is SoftRefrence
                img = ((Reference)img).get ();
            }
            
            if (img != null)
                // cannot be NO_ICON, since it never disappears from the map.
                return (Image)img;

            // path for bug in classloader
            String n;
            if (name.startsWith ("/")) { // NOI18N
                n = name.substring (1);
            } else {
                n = name;
            }

            // we have to load it
            java.net.URL url = loader != null ?
                               loader.getResource (n) : IconManager.class.getClassLoader ().getResource (n);

            img = url == null ? null : Toolkit.getDefaultToolkit ().getImage (url);
            if (img != null) {
                Image img2 = toBufferedImage((Image) img);
                //System.err.println("loading icon " + n + " = " + img2);
                Reference r = new SoftReference (img2);
                map.put (name, r);
                startCleaner();
                return (Image) img2;
            } else {
                // no icon found
                map.put (name, NO_ICON);
                return null;
            }
        }
    }

    /**
     * Cleans image cache.
     */    
    public void run() {
        cleanUpMap();
        cleaningTask = RequestProcessor.postRequest(this, CACHE_CLEAN_INTERVAL, Thread.MIN_PRIORITY + 1);
    }
    
    private static void cleanUpMap() {
        Reference r;
        
        while ((r = compositeQueue.poll()) != null) {
            Object k;
            synchronized (reverseMap) {
                k = reverseMap.remove(r);
            }
            //System.err.println("removing " + k);
            synchronized (map) {
                map.remove(k);
            }
       }
    }

    /**
     * Key used for composite images -- it holds image identities
     */    
    private static class CompositeImageKey {
        Image baseImage, overlayImage;
        int x, y;
        
        CompositeImageKey(Image base, Image overlay, int x, int y) {
            this.x = x;
            this.y = y;
            this.baseImage = base;
            this.overlayImage = overlay;
        }
        
        public boolean equals(Object other) {
            if (!(other instanceof CompositeImageKey))
                return false;
            CompositeImageKey k = (CompositeImageKey)other;
            return (x == k.x) && (y == k.y) && (baseImage == k.baseImage) &&
                (overlayImage == k.overlayImage);
        }
        
        public int hashCode() {
            int hash =  ((x << 3) ^ y) << 4;
            hash = hash ^ baseImage.hashCode() ^ overlayImage.hashCode();
            return hash;
        }
        
        public String toString() {
            return "Composite key for " + baseImage + " + " + overlayImage + " at [" + x + ", " + y + "]";
        }
    }

    /**
     * Method that attempts to find the merged image in the cache first, then
     * creates the image if it was not found.
     */
    static final Image mergeImages(Image im1, Image im2, int x, int y) {
        CompositeImageKey k = new CompositeImageKey(im1, im2, x, y);
        Image cached;
        
        cleanUpMap();
        synchronized (map) {
            Reference r = (Reference)map.get(k);
            if (r != null) {
                cached = (Image)r.get();
                if (cached != null)
                    return cached;
            }
            cached = doMergeImages(im1, im2, x, y);
            r = new SoftReference(cached, compositeQueue);
            map.put(k, r);
            reverseMap.put(r, k);
            return cached;
        }
    }

    /** The method creates a BufferedImage which represents the same Image as the
     * parameter but consumes less memory.
     */
    static final Image toBufferedImage(Image img) {
        // load the image
        new javax.swing.ImageIcon(img);
        java.awt.image.BufferedImage rep = createBufferedImage(img.getWidth(null), img.getHeight(null));
        java.awt.Graphics g = rep.createGraphics();
        g.drawImage(img, 0, 0, null);
        g.dispose();
        img.flush();
        return rep;
    }

    private static final Image doMergeImages (Image image1, Image image2, int x, int y) {
        int w = 0, h = 0;
        //System.err.println("merging images " + image1 + " + " + image2 + " at [" + x + ", " + y + "]");
        if (image1 != null) {
            w = image1.getWidth(null);
            h = image1.getHeight(null);
        }
        if (image2 != null) {
            w = image2.getWidth(null)+x > w ? image2.getWidth(null)+x : w;
            h = image2.getHeight(null)+y > h ? image2.getHeight(null)+y : h;
        }
        if (w < 1) w = 16;
        if (h < 1) h = 16;
        
        java.awt.image.ColorModel model = java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment ().
                                          getDefaultScreenDevice ().getDefaultConfiguration ().
                                          getColorModel (java.awt.Transparency.BITMASK);
        java.awt.image.BufferedImage buffImage = new java.awt.image.BufferedImage (model,
             model.createCompatibleWritableRaster (w, h), model.isAlphaPremultiplied (), null);
        
        java.awt.Graphics g = buffImage.createGraphics ();
        if (image1 != null) g.drawImage (image1, 0, 0, null);
        if (image2 != null) g.drawImage (image2, x, y, null);
        g.dispose();
        
        return buffImage;
    }
    
    /** Creates BufferedImage 16x16 and Transparency.BITMASK */
    private static final java.awt.image.BufferedImage createBufferedImage(int width, int height) {
        java.awt.image.ColorModel model = java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment().
                                          getDefaultScreenDevice().getDefaultConfiguration().getColorModel(java.awt.Transparency.BITMASK);
        java.awt.image.BufferedImage buffImage = new java.awt.image.BufferedImage(model,
                model.createCompatibleWritableRaster(width, height), model.isAlphaPremultiplied(), null);
        return buffImage;
    }
}
