/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.util;

import java.util.*;


/** This is abstract interfaces to give us access to lookup capabilities.
 * <P>
 * I admit that the class is inspirated by JINI(tm). The difference 
 * is that methods are not allowed to throw exceptions and also the license.
 * Also we are mostly concentrated on the lookup, not on the registration.
 * And last difference that comes to my mind is that our classes are not 
 * serializable (we are not distributed).
 *
 * @author  Jaroslav Tulach
 */
public abstract class Lookup {
    /** default instance */
    private static Lookup defaultLookup;
    
    /** Static method to obtain the global lookup in the whole system.
     * @return the global lookup in the system
     */
    public static Lookup getDefault () {
        if (defaultLookup != null) {
            return defaultLookup;
        }
        
        return initializeLookup ();
    }
    
    /** Initializes the top manager.
    */
    private static synchronized Lookup initializeLookup () {
        // synchronization check
        if (defaultLookup != null) {
            return defaultLookup;
        }

        String className = System.getProperty(
                               "org.openide.util.Lookup" // NOI18N
                           );
        
        if (className == null) {
            defaultLookup = new Empty ();
            return defaultLookup;
        }

        try {
            Class c = Class.forName(className);
            defaultLookup = (Lookup)c.newInstance();
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new InternalError ();
        }

        // late initialization of the lookup if needed
        if (defaultLookup instanceof Runnable) {
            ((Runnable)defaultLookup).run ();
        }

        return defaultLookup;
    }
    
    
    /** Lookups an object of given interface. This is the simplest method
     * for the lookuping, if more registered objects implement the given
     * class any of them can be returned. 
     *
     * @param clazz class of the object we are searching for
     * @return the object implementing given class or null if no such 
     *    has been found
     */
    public abstract Object lookup (Class clazz);
    
    /** The general lookup method.
     * @param template the template describing the services we are looking for
     * @return object containing the results
     */
    public abstract Result lookup (Template template);
    
    /** Template to define pattern of object to look for. 
     * <P>
     * I expect this class to grow in the future, but for now, it is 
     * enough to start with something simple.
     */
    public static final class Template extends Object {
        /** type of the service */
        private Class type;
        
        /** General template to find all possible instances.
         */
        public Template () {
        }
        
        /** Constructor to create new template.
         * @param serviceClass the class of service we are looking for
         */
        public Template (Class type) {
            this.type = type;
        }
        
        /** @return the class we are looking for */
        public Class getType () {
            return type;
        }
    }
    
    /** Result of a lookup request. Allows access to single object
     * that was found (not too useful) and also to all objects found
     * (more useful).
     */
    public static abstract class Result extends Object {
        /** Registers a listener that is invoked when there is a possible 
         * change in this result.
         *
         * @param l listener to invoke when there is an change
         */
        public abstract void addLookupListener (LookupListener l);

        /** Unregisters a listener previously added by addLookupListener.
         * @param l the listener
         */
        public abstract void removeLookupListener (LookupListener l);
        
        /** Access to all instances in the result.
         * @return collection of all instances
         */
        public abstract java.util.Collection allInstances ();
        
        /** Classes of all results. Set of the most concreate classes
         * that are registered in the system.
         * @return set of Class objects
         */
//JST: For now commented out
//        public abstract java.util.Set allClasses ();
        
        /** All registered items. The collection of all pairs of 
         * instances and their classes.
         * @return collection of Lookup.Item
         */
//JST: For now commented out        
//        public abstract java.util.Collection allItems ();
    }
    
//JST: Not needed now:    
    /** A single item in the result. An object to bind together the 
     * instance and its class.
     * <P>
     * In future it could also get access to additional attributes.
     *
    public static abstract class Item extends Object {
        /** The class of the result item.
         * @return the instance of the object.
         *
        public abstract Object getInstance ();
        
        /** The class of the result item.
         * @return the class of the item
         *
        public abstract Class getType ();
    }
     */
    
    
    //
    // Implementation of the default lookup
    //
    
    private static final class Empty extends Lookup {
        private static final Result NO_RESULT = new Result () {
            /** Registers a listener that is invoked when there is a possible 
             * change in this result.
             *
             * @param l listener to invoke when there is an change
             */
            public void addLookupListener (LookupListener l) {
            }

            /** Unregisters a listener previously added by addChangeListener.
             * @param l the listener
             */
            public void removeLookupListener (LookupListener l) {
            }

            /** Access to all instances in the result.
             * @return collection of all instances
             */
            public java.util.Collection allInstances () {
                return java.util.Collections.EMPTY_LIST;
            }
        };
        
        /** Lookups an object of given interface. This is the simplest method
         * for the lookuping, if more registered objects implement the given
         * class any of them can be returned. 
         *
         * @param clazz class of the object we are searching for
         * @return the object implementing given class or null if no such 
         *   has been found
         */
        public Object lookup(Class clazz) {
            return null;
        }
        
        /** The general lookup method.
         * @param template the template describing the services we are looking for
         * @return object containing the results
         */
        public Result lookup(Template template) {
            return NO_RESULT;
        }
    }
}