/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.util;

import java.lang.ref.*;
import java.util.*;

import org.openide.ErrorManager;

/** Request processor that is capable to execute actions in
* special thread.
*
* @author Jaroslav Tulach
*/
public final class RequestProcessor extends Object {
    // JST: final can be removed if needed
    /** the static instance for users that do not want to have own processor */
    private static RequestProcessor DEFAULT = new RequestProcessor ();

    /** number of processors */
    private static int processorCount = 0;

    /** sorted set of all task that are waiting to be processed
    * objects of type (Holder), sorted by time
    */
    private final TreeSet waiting = new TreeSet (new Cmp (true));

    /** thread to process requests */
    private ProcessorThread thread;

    /** counter of all created Holders */
    private int counter;

    /** name of the request processor or null */
    private String name;

    /** Default constructor.
    */
    public RequestProcessor () {
    }

    /** Constructor.
    * @param name the name to use for the request processor thread
    */
    public RequestProcessor (String name) {
        this.name = name;
    }

    /** When finalized, stops the thread.
    */
    protected void finalize () {
        stop ();
    }

    /** This methods asks the request processor to start given
    * runnable immediately. The default priority is {@link Thread#MIN_PRIORITY}.
    *
    * @param run class to run
    * @return the task to control the request
    */
    public Task post (Runnable run) {
        return post (run, 0, Thread.MIN_PRIORITY);
    }

    /** This methods asks the request processor to start given
    * runnable after <code>timeToWait</code> milliseconds. The default priority is {@link Thread#MIN_PRIORITY}.
    *
    * @param run class to run
    * @param timeToWait to wait before execution
    * @return the task to control the request
    */
    public Task post (final Runnable run, int timeToWait) {
        return post (run, timeToWait, Thread.MIN_PRIORITY);
    }

    /** This methods asks the request processor to start given
    * runnable after <code>timeToWait</code> milliseconds. Given priority is assigned to the
    * request.
    *
    *
    * @param run class to run
    * @param timeToWait to wait before execution
    * @param priority the priority from {@link Thread#MIN_PRIORITY} to {@link Thread#MAX_PRIORITY}
    * @return the task to control the request
    */
    public Task post (final Runnable run, int timeToWait, int priority) {
        Task t = new Task (run, timeToWait, priority);
        synchronized (waiting) {
            Holder h = t.createHolder ();
            waiting.add (h);
            checkTimerQueue ();
        }
        /* debug code
        if (debug != null && RequestProcessor.this == DEFAULT) {              
          debug.notifyPost (t);
    }
        // end of debug
        */
        return t;
    }

    /** Creates request that can be later started by setting its delay.
    * The request is not immediatelly put into the queue. It is planned after
    * setting its delay by setDelay method.
    *
    * @param run action to run in the process
    * @return the task to control execution of given action
    */
    public Task create (Runnable run) {
        return new Task (run, 0, Thread.MIN_PRIORITY);
    }

    /** Tests if the current thread is request processor thread.
    * This method could be used to prevent the deadlocks using
    * <CODE>waitFinished</CODE> method. Any two tasks created
    * by request processor must not wait for themself.
    *
    * @return <CODE>true</CODE> if the current thread is request processor
    *          thread, otherwise <CODE>false</CODE>
    */
    public boolean isRequestProcessorThread () {
        return Thread.currentThread().equals(thread);
    }

    /** Stops processing of runnables processor.
    * The currently running runnable is finished and no new is started.
    */
    public void stop () {
        synchronized (waiting) {
            if (thread != null) {
                thread.stopProcessing ();
                thread = null;
            }
        }
    }


    //
    // Static methods communicating with default request processor
    //

    /** This methods asks the request processor to start given
    * runnable after <code>timeToWait</code> milliseconds. The default priority is {@link Thread#MIN_PRIORITY}.
    *
    * @param run class to run
    * @return the task to control the request
    */
    public static Task postRequest (Runnable run) {
        return DEFAULT.post (run);
    }

    /** This methods asks the request processor to start given
    * runnable after <code>timeToWait</code> milliseconds. The default priority is {@link Thread#MIN_PRIORITY}.
    *
    * @param run class to run
    * @param timeToWait to wait before execution
    * @return the task to control the request
    */
    public static Task postRequest (final Runnable run, int timeToWait) {
        return DEFAULT.post (run, timeToWait);
    }

    /** This methods asks the request processor to start given
    * runnable after <code>timeToWait</code> milliseconds. Given priority is assigned to the
    * request.
    *
    *
    * @param run class to run
    * @param timeToWait to wait before execution
    * @param priority the priority from {@link Thread#MIN_PRIORITY} to {@link Thread#MAX_PRIORITY}
    * @return the task to control the request
    */
    public static Task postRequest (final Runnable run, int timeToWait, int priority) {
        return DEFAULT.post (run, timeToWait, priority);
    }

    /** Creates request that can be later started by setting its delay.
    * The request is not immediatelly put into the queue. It is planned after
    * setting its delay by setDelay method.
    *
    * @param run action to run in the process
    * @return the task to control execution of given action
    */
    public static Task createRequest (Runnable run) {
        return DEFAULT.create (run);
    }


    //
    // Implementation of the queue
    //

    /** Checks the timer queue. First of all sees whether there is
    * the processor thread running. If not, starts one. Then checks if
    * the time of execution of the first task has not changed. If so,
    * wakeups the thread so it can replan itself.
    * <P>
    * The methods is always called with synchronization on waiting.
    */
    final void checkTimerQueue () {
        if (thread == null) {
            // only start processor thread
            thread = new ProcessorThread (name, waiting);
            thread.start ();
        } else {
            if (waiting.size () == 0) {
                // nothing to plan
                return;
            }
            // wakeup the thread to change what should be changed
            waiting.notify ();
        }
    }
    
    /** Increments the counter of holders created for this RequestProcessor.
     * If the number of created olders is equal to Integer.MAX_VALUE, it 
     * decreases it and numbers of all holders as much as possible.
     * Can be called only with waiting lock.
     *
     * @return counter value
     * 
     */
    final int incrementCounter () {
        if (counter != Integer.MAX_VALUE) {
            return ++counter;
        }

        // a set of all handlers sorted by their count numbers
        TreeSet allHandlers = new TreeSet (new Comparator () {
            public int compare (Object o1, Object o2) {
                Holder h1 = (Holder)o1;
                Holder h2 = (Holder)o2;
                
                return h1.count - h2.count;
            }
        });
        allHandlers.addAll (waiting);
        if (thread != null) {
            allHandlers.addAll (thread.pending);
        }
        
        Iterator it = allHandlers.iterator ();
        int i = 0;
        while (it.hasNext ()) {
            Holder h = (Holder)it.next ();
            h.count = i++;
        }
        
        if (i == Integer.MAX_VALUE) {
            // should not happen, just to be really sure
            throw new IllegalStateException ("Request processor is full of tasks > Integer.MAX_VALUE"); // NOI18N
        }
        
        counter = i + 1;
        return counter;
    }
    
    /** Holder for a task */
    private static final class Holder extends Exception {
        public Task task;
        public int priority;
        public long time;
        public int count;

        public Holder (Task t, int count) {
            super ("Posted StackTrace"); // NOI18N
            task = t;
            priority = t.priority;
            time = t.time;
            this.count = count;
        }

        final String toString (long time) {
            Task t = task;
            return super.toString () + "(" + (t == null ? "null" : t.toString (time)) + ")"; // NOI18N
        }
        
        public String toString () {
            return toString (System.currentTimeMillis ());
        }
    }

    /** The task describing the request task send to the processor.
    */
    public final class Task extends org.openide.util.Task {
        /** time to run */
        long time;
        /** priority */
        int priority;

        /** holder that is currently representing the task in the queue */
        Holder holder;

        /** @param run runnable to start
        * @param delay amount of millis to wait
        * @param priority the priorty of the task
        */
        Task (Runnable run, long delay, int priority) {
            super (run);
            time = System.currentTimeMillis () + delay;
            this.priority = priority;
        }

        /** Creates new holder canceling the previous one (if any).
        * !Can be called only under waiting lock!
        */
        Holder createHolder () {
            if (holder != null) {
                holder.task = null;
            }
            return holder = new Holder (this, incrementCounter ());
        }

        /** Getter for amount of millis till this task
        * is started.
        * @return amount of millis
        */
        public int getDelay () {
            long delay = time - System.currentTimeMillis ();
            if (delay < 0L) return 0;
            if (delay > (long)Integer.MAX_VALUE) return Integer.MAX_VALUE;
            return (int)delay;
        }

        /** Changes the delay to different level,
        * if the task has not been run yet, it is replaned to
        * the new time. If it has been finished, it is replaned
        * to be started again.
        *
        * @param delay time in millis to wait
        */
        public void schedule (int delay) {
            synchronized (waiting) {
                time = System.currentTimeMillis () + delay;
                waiting.add (createHolder ());
                checkTimerQueue ();
            }
        }

        /** Removes the task from the queue.
        *
        * @return true if the task has been removed from the queue,
        *   false it the task has already been processed
        */
        public boolean cancel () {
            synchronized (waiting) {
                if (holder != null) {
                    holder.task = null;
                    holder = null;

                    // mark this task as being finished
                    notifyFinished ();

                    return true;
                } else {
                    return false;
                }
            }
        }

        /** Current priority of the task.
        */
        public int getPriority () {
            return priority;
        }

        /** Changes priority to new one. If the task has been
        * already run, do not plan it again.
        *
        */
        public void setPriority (int priority) {
            if (this.priority != priority) {
                if (priority < Thread.MIN_PRIORITY) priority = Thread.MIN_PRIORITY;
                if (priority > Thread.MAX_PRIORITY) priority = Thread.MAX_PRIORITY;

                synchronized (waiting) {
                    this.priority = priority;
                    if (holder != null) {
                        waiting.add (createHolder ());
                        checkTimerQueue ();
                    }
                }
            }
        }

        /** This method is an implementation of the waitFinished method
        * in the RequestProcessor.Task. It check the current thread if it is
        * request processor thread.
        *
        * This implemetation run the task and then returns.
        */
        void waitFinishedImpl () {
            if (isRequestProcessorThread()) {
                // one thread runnig
                boolean toRun = false;

                synchronized (waiting) {
                    if (holder != null) {
                        holder.task = null;
                        holder = null;
                        toRun = true;
                    }
                }
                if (toRun) {
                    run();
                }
            }
            else {
                /* JST: new threading model does not need this.
                        if (System.getProperty("netbeans.debug.threads") != null) {
                          if (javax.swing.SwingUtilities.isEventDispatchThread()) {
                            System.out.println ("WARNING: EventDispathThread is waiting for request processor task.");
                            Thread.dumpStack();
                          }
                        }
                */        
                super.waitFinishedImpl();
            }
        }

        /** @return string representation */
        final String toString (long currentTime) {
            Holder h = holder;
            int cnt = h == null ? -1 : h.count;
            return super.toString () + " [" + (time - currentTime) + ", " + priority + ", " + cnt + ']'; // NOI18N
        }
        
        public String toString () {
            return toString (System.currentTimeMillis ());
        }
    }

    /** Processor thread.
    */
    private final static class ProcessorThread extends Thread {
        /** sorted set of all task that are waiting to be processed
        * objects of type (Holder), sorted by priority
        */
        private TreeSet pending = new TreeSet (new Cmp (false));
        
        /** a reference to request processor's waiting set. Serves as 
         * an assurance that the request processor will not be garbage
         * collected, if there is a job that should be executed. Because:
         *
         * If waiting.isEmpty () => there is no reference from the thread
         *   to the RequestProcessor
         * If !waiting.isEmpty () => there is at least one Holder in the
         *   set and it references a Task and each task has pointer to 
         *   RequestProcessor
         */
        private TreeSet waiting;
        

        /** reference to the processor */
        //private Reference requestProcessor;

        /** stops the processor if true */
        private boolean stop;
        /** previous priority */
        private int priority;
        /** last sleep time */
        Object sleep;
        /** last delay */
        long at;


        /** Constructor. */
        public ProcessorThread (String name, TreeSet waiting) {
            super (
                getTopLevelThreadGroup(),
                name == null ?
                ("OpenIDE Request Processor-" + processorCount++) // NOI18N
                :
                name
            );
            setDaemon (true);
            priority = getPriority ();

            this.waiting = waiting;
        }

        /** Stops the processing (must have synch on the waiting) */
        public void stopProcessing () {
            stop = true;
            interrupt ();
        }


        /** Copies objects from the waiting queue that are ready to be
        * processed into pending queue. This method blocks when the
        * the pending queue is empty.
        *
        * <P>
        * Called under lock on waiting.
        * @return the amount of time to wait till next event should be
        *   posted from waiting to pending (0 means infinity)
        */
        private long waitingToPending () {
            if (waiting.isEmpty ()) {
                // wait for ever
                return 0L;
            }

            Iterator it = waiting.iterator ();
            while (it.hasNext ()) {
                Holder holder = (Holder)it.next ();

                if (holder.task == null) {
                    // continue
                    it.remove ();
                    continue;
                }

                long diff = holder.time - System.currentTimeMillis ();
                /*        if (debug != null && pending.size () > 15) {
                System.err.println ("waitingToPending: " + holder + " has diff: " + diff);
                        }
                */        
                if (diff > 0L) {
                    // this task should be run in future =>
                    // return the amount of time to wait
                    return diff;
                }

                it.remove ();
                // put the holder into the pending queue
                pending.add (holder);
            }

            // pending is not empty => do not wait
            // pending is empty => wait forever
            return 0L;
        }

        /** Synchronization object or exception.
        */
        private Object synch () {
            return waiting;
        }

        /** Processor of requests.
        */
        public void run () {
            int priority = getPriority ();

            while (!stop) {
                Task t;
                Holder h;
                Iterator first;


                synchronized (synch ()) {

                    for (;;) {
                        h = null;
                        t = null;
                        first = null;
                        
                        // either fills the pending set or return time to
                        // wait for another request processing
                        long w = waitingToPending ();

                        if (!pending.isEmpty ()) {
                            break;
                        }

                        // wait the given time
                        try {
                            /*
                            if (debug != null && requestProcessor.get () == DEFAULT) {              
                              debug.notifySleep ((int)w);
                        }
                            */             
                            synch ().wait (w);
                        } catch (InterruptedException ex) {
                            if (stop) {
                                return;
                            } else {
                                throw new InternalError ();
                            }
                        }
                    }

                    // take first holder (pending is not empty)
                    first = pending.iterator ();
                    h = (Holder)first.next ();
                    first.remove ();

                    // take the task
                    t = h.task;
                    if (t != null) {
                        // mark as being processed
                        t.holder = null;
                    } else {
                        continue;
                    }
                }

                // run the task
                int p = t.getPriority ();
                if (priority != p) {
                    setPriority (priority = p);
                }

                try {
                    /*
                    // debug code
                    if (debug != null && requestProcessor.get () == DEFAULT) {
                    System.err.println ("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv");
                      debug.notifyRun (t, h);
                      h.printStackTrace();
                }
                    // end of debug    
                    */          
                    t.run ();
                    /*
                    if (debug != null && requestProcessor.get () == DEFAULT) {
                      System.err.println ("Task finished: " + t);
                      debug.printRequestProcessor ();
                    System.err.println ("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^");
                }
                    */
                } catch (ThreadDeath td) {
                    // do not catch this
                    throw td;
                } catch (Throwable ex) {
                    ErrorManager err = Utilities.errMan ();
                    if (err != null) {
                        // copies annotation from exception ex to the handler (which holds)
                        // the stack trace of the caller of the exception
                        err.copyAnnotation (h, ex);
                        // adds a description that this exception happened in
                        // request processor
                        err.annotate (
                          h, NbBundle.getMessage (RequestProcessor.class, "EXC_IN_REQUEST_PROCESSOR")
                        );
                        err.notify (
                          h
                        );
                    } else {
                        ex.printStackTrace();
                    }
                    // memory leak!!!
                    ex = null;
                }
            }
        } // run
    }

    /**
    * @return a top level ThreadGroup. The method ensures that even
    * RequestProcessors created by internal execution will survive the
    * end of the task.
    */
    static ThreadGroup getTopLevelThreadGroup() {
        java.security.PrivilegedAction run = new java.security.PrivilegedAction() {
                                                 public Object run() {
                                                     ThreadGroup current = Thread.currentThread().getThreadGroup();
                                                     while (current.getParent() != null) {
                                                         current = current.getParent();
                                                     }
                                                     return current;
                                                 }
                                             };

        return (ThreadGroup) java.security.AccessController.doPrivileged(run);
    }




    /** Comparator that compares times & priorities */
    private static final class Cmp extends Object implements Comparator {
        private boolean timeFirst;
        
        public Cmp (boolean timeFirst) {
            this.timeFirst = timeFirst;
        }
        
        /** Compares two holders.
        *
        */
        public int compare (Object o1, Object o2) {
            Holder h1 = (Holder)o1;
            Holder h2 = (Holder)o2;
            

            int ret;
            if (timeFirst) {
                ret = compareTimes (h1, h2, true);
            } else {
                ret = comparePriorities (h1, h2, true);
            }
            
            /*
            if (ret == 0 && h1 != h2) {
                long time = System.currentTimeMillis ();
                System.err.println("!!!RequestProcessorMistake!!!");
                System.err.println("   h1: " + h1.toString (time) + " id: " + System.identityHashCode (h1));
                System.err.println("   h2: " + h2.toString (time) + " id: " + System.identityHashCode (h2));
                System.err.println("   eq: " + (h1 == h2));
                Thread.dumpStack();
            }
            */
            
            return ret;
        }
        
        /** Compares times.
         */
        private static int compareTimes (Holder h1, Holder h2, boolean goOn) {
            if (h1.time == h2.time) {
                if (goOn) {
                    return comparePriorities (h1, h2, false);
                } else {
                    // compare counts
                    return h1.count - h2.count;
                }
            } else {
                if (h1.time > h2.time) {
                    return 1;
                } else {
                    return -1;
                }
            }                
        }
        
        /** Compares priorities.
         */
        private static int comparePriorities (Holder h1, Holder h2, boolean goOn) {
            if (h1.time == h2.time) {
                if (goOn) {
                    return compareTimes (h1, h2, false);
                } else {
                    return h1.count - h2.count;
                }
            } else {
                if (h1.time > h2.time) {
                    return 1;
                } else {
                    return -1;
                }
            }                
        }
    }
}

