/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.windows;

import java.beans.*;
import java.text.MessageFormat;

import org.openide.util.*;
import org.openide.TopManager;
import org.openide.cookies.OpenCookie;
import org.openide.cookies.ViewCookie;
import org.openide.cookies.CloseCookie;
import org.openide.util.WeakListener;

/** Simple support for an openable objects.
* Can be used either as an {@link OpenCookie}, {@link ViewCookie}, or {@link CloseCookie},
* depending on which cookies the subclass implements.
*
* @author Jaroslav Tulach
*/
public abstract class CloneableOpenSupport extends Object {
    /** the environment that provides connection to outside world */
    protected Env env;

    /** All opened editors on this file. */
    protected CloneableTopComponent.Ref allEditors;

    private static java.awt.Container container;

    /** New support for a given environment. 
    * @param env environment to take all date from/to
    */
    public CloneableOpenSupport(Env env) {
        this.env = env;

        Listener l = new Listener (env);
        this.allEditors = l;

        // attach property change listener to be informed about loosing validity
        env.addPropertyChangeListener (WeakListener.propertyChange (
                                           l, env
                                       ));
        // attach vetoable change listener to be cancel loosing validity when modified
        env.addVetoableChangeListener (WeakListener.vetoableChange (
                                           l, env
                                       ));
    }

    /** Focuses existing component to open, or if none exists creates new.
    * @see OpenCookie#open
    */
    public void open () {
        //Bugfix #10688 open() is now run in AWT thread
        Mutex.EVENT.writeAccess (new Runnable () {
            public void run () {
                CloneableTopComponent editor = openCloneableTopComponent();
                editor.requestFocus();
            }
        });
    }

    /** Focuses existing component to view, or if none exists creates new.
    * The default implementation simply calls {@link #open}.
    * @see ViewCookie#view
    */
    public void view () {
        open ();
    }

    /** Focuses existing component to view, or if none exists creates new.
    * The default implementation simply calls {@link #open}.
    * @see ViewCookie#view
    */
    public void edit () {
        open ();
    }

    /** Closes all components.
    * @return <code>true</code> if every component is successfully closed or <code>false</code> if the user cancelled the request
    * @see CloseCookie#close
    */
    public boolean close () {
        return close (true);
    }

    /** Closes all opened windows.
    * @param ask true if we should ask user
    * @return true if sucesfully closed
    */
    protected boolean close (final boolean ask) {
        if (allEditors.isEmpty ()) {
            return true;
        }
        
        //Bugfix #10688 close() is now run in AWT thread
        //also bugfix of 10714 - whole close (boolean) is run in AWT thread
        Boolean ret = (Boolean) Mutex.EVENT.writeAccess (new Mutex.Action () {
            public Object run () {
                //synchronized (allEditors) {
                synchronized (getLock()) {
                    // user canceled the action
                    if (ask && !canClose ()) {
                        return Boolean.FALSE;
                    }

                    java.util.Enumeration en = allEditors.getComponents ();
                    while (en.hasMoreElements ()) {
                        TopComponent c = (TopComponent)en.nextElement ();
                        if (!c.close ()) {
                            return Boolean.FALSE;
                        }
                    }
                }
                return Boolean.TRUE;
            }
        });
        return ret.booleanValue();
    }

    /** Should test whether all data is saved, and if not, prompt the user
    * to save.
    * The default implementation returns <code>true</code>.
    *
    * @return <code>true</code> if everything can be closed
    */
    protected boolean canClose () {
        return true;
    }

    /** Simply open for an editor. */
    protected final CloneableTopComponent openCloneableTopComponent() {
        //synchronized (allEditors) {
        synchronized (getLock()) {
            try {
                CloneableTopComponent ret = (CloneableTopComponent)allEditors.getAnyComponent ();
                ret.open();
                return ret;
            } catch (java.util.NoSuchElementException ex) {
                // no opened editor

                String msg = messageOpening ();
                if (msg != null) {
                    TopManager.getDefault().setStatusText(msg);
                }

                CloneableTopComponent editor = createCloneableTopComponent ();
                editor.setReference (allEditors);
                editor.open();

                msg = messageOpened ();
                if (msg == null) {
                    msg = "";
                }
                TopManager.getDefault ().setStatusText (msg);

                return editor;
            }
        }
    }

    /** Creates lock object used in close and openCloneableTopComponent. */
    private Object getLock() {
        if (container == null) {
            container = new java.awt.Container();
        }
        return container.getTreeLock();
    }

    /** A method to create a new component. Must be overridden in subclasses.
    * @return the cloneable top component for this support
    */
    protected abstract CloneableTopComponent createCloneableTopComponent ();

    /** Message to display when an object is being opened.
    * @return the message or null if nothing should be displayed
    */
    protected abstract String messageOpening ();
    

    /** Message to display when an object has been opened.
    * @return the message or null if nothing should be displayed
    */
    protected abstract String messageOpened ();



    /** Abstract interface that is used by CloneableOpenSupport to
    * talk to outside world.
    */
    public static interface Env extends java.io.Serializable {
        /** that is fired when the objects wants to mark itself as
        * invalid, so all components should be closed.
        */
        public static final String PROP_VALID = org.openide.loaders.DataObject.PROP_VALID;
        /** that is fired when the objects wants to mark itself modified
        * or not modified.
        */
        public static final String PROP_MODIFIED = org.openide.loaders.DataObject.PROP_MODIFIED;

        /** Adds property listener.
        */
        public void addPropertyChangeListener (PropertyChangeListener l);
        /** Removes property listener.
        */
        public void removePropertyChangeListener (PropertyChangeListener l);

        /** Adds veto listener.
        */
        public void addVetoableChangeListener (VetoableChangeListener l);
        /** Removes veto listener.
        */
        public void removeVetoableChangeListener (VetoableChangeListener l);

        /** Test whether the support is in valid state or not.
        * It could be invalid after deserialization when the object it
        * referenced to does not exist anymore.
        *
        * @return true or false depending on its state
        */
        public boolean isValid ();
        
        /** Test whether the object is modified or not.
        * @return true if the object is modified
        */
        public boolean isModified ();

        /** Support for marking the environement modified.
        * @exception IOException if the environment cannot be marked modified
        *    (for example when the file is readonly), when such exception
        *    is the support should discard all previous changes
        */
        public void markModified () throws java.io.IOException;

        /** Reverse method that can be called to make the environment 
        * unmodified.
        */
        public void unmarkModified ();

        /** Method that allows environment to find its 
        * cloneable open support.
        */
        public CloneableOpenSupport findCloneableOpenSupport ();
    }

    /** Property change & veto listener. To react to dispose/delete of
    * the data object.
    */
    private static final class Listener extends CloneableTopComponent.Ref
    implements PropertyChangeListener, VetoableChangeListener, Runnable {
        /** generated Serialized Version UID */
        static final long serialVersionUID = -1934890789745432531L;
        /** environment to use as connection to outside world */
        private Env env;

        /** Constructor.
        */
        public Listener (Env env) {
            this.env = env;
        }
        
        /** Getter for the associated CloneableOpenSupport 
        * @return the support or null if none was found
        */
        private CloneableOpenSupport support () {
            return env.findCloneableOpenSupport ();
        }

        public void propertyChange (PropertyChangeEvent ev) {
            if (Env.PROP_VALID.equals (ev.getPropertyName ())) {
                // do not check it if old value is not true
                if (Boolean.FALSE.equals (ev.getOldValue ())) return;

                Mutex.EVENT.readAccess (this);
            }
        }
        
        /** Closes the support in AWT thread.
         */
        public void run () {
            // loosing validity
            CloneableOpenSupport os = support ();
            if (os != null) {
                // mark the object as not being modified, so nobody
                // will ask for save
                env.unmarkModified ();

                os.close (false);
            }
        }

        /** Forbids setValid (false) on data object when there is an
        * opened editor.
        *
        * @param ev PropertyChangeEvent
        */
        public void vetoableChange (PropertyChangeEvent ev)
        throws PropertyVetoException {
            if (Env.PROP_VALID.equals (ev.getPropertyName ())) {
                // do not check it if old value is not true
                if (Boolean.FALSE.equals (ev.getOldValue ())) return;

                if (env.isModified ()) {
                    // if the object is modified 
                    CloneableOpenSupport os = support ();
                    if (os != null && !os.canClose ()) {
                        // is modified and has not been sucessfully closed
                        throw new PropertyVetoException (
                            // [PENDING] this is not a very good detail message!
                            "", ev // NOI18N
                        );
                    }
                }
            }
        }

        /** Resolvable to connect to the right data object. This
        * method is used for connectiong CloneableTopComponents via
        * their CloneableTopComponent.Ref
        */
        public Object readResolve () {
            CloneableOpenSupport os = support ();
            if (os == null) {
                // problem! no replace!?
                return this;
            }
            // use the editor support's CloneableTopComponent.Ref
            return os.allEditors;
        }
    }
}

/*
* Log
*  21   Jaga      1.19.1.0    3/27/00  Jesse Glick     Only added a comment.
*  20   src-jtulach1.19        12/8/99  Jaroslav Tulach TopComponent enhanced.
*  19   src-jtulach1.18        11/5/99  Jaroslav Tulach WeakListener has now 
*       registration methods.
*  18   src-jtulach1.17        10/22/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  17   src-jtulach1.16        9/30/99  Jaroslav Tulach 
*  16   src-jtulach1.15        9/30/99  Jaroslav Tulach OpenSupport is attached 
*       to setValid veto change of its data object.
*  15   src-jtulach1.14        9/13/99  Jaroslav Tulach #2628
*  14   src-jtulach1.13        9/13/99  Jan Jancura     Do not write "Opening 
*       blablabla" when no new document is opened - only focused.
*  13   src-jtulach1.12        7/13/99  Ales Novak      new win sys change
*  12   src-jtulach1.11        7/11/99  David Simonek   window system change...
*  11   src-jtulach1.10        6/22/99  Ales Novak      creating of editors is 
*       centralized
*  10   src-jtulach1.9         6/8/99   Ian Formanek    ---- Package Change To 
*       org.openide ----
*  9    src-jtulach1.8         4/2/99   Jaroslav Tulach New from template opens 
*       the editor.
*  8    src-jtulach1.7         3/31/99  David Simonek   ugly ugly ugly 
*       requestFocus bungs fixed
*  7    src-jtulach1.6         3/17/99  Jaroslav Tulach Output Window fixing.
*  6    src-jtulach1.5         3/15/99  Jesse Glick     [JavaDoc]
*  5    src-jtulach1.4         3/14/99  Jaroslav Tulach Change of 
*       MultiDataObject.Entry.
*  4    src-jtulach1.3         3/10/99  Jesse Glick     [JavaDoc]
*  3    src-jtulach1.2         2/3/99   Jaroslav Tulach 
*  2    src-jtulach1.1         1/17/99  Jaroslav Tulach open/close/canClose 
*       methods
*  1    src-jtulach1.0         1/7/99   Jaroslav Tulach 
* $
*/
