/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openide.xml;

import java.io.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * Utility class collecting library methods related to XML processing.
 *
 * @author  Petr Kuzel
 * @since   release 3.2
 * @version 1.0
 */
public final class XMLUtil extends Object {

    /** Forbids creating new XMLUtil */
    private XMLUtil() {
    }

    
    // ~~~~~~~~~~~~~~~~~~~~~ SAX related ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
    /**
      * @return <code>createXMLReader(false, false);</code>
      */
    public static XMLReader createXMLReader() throws SAXException {
        return createXMLReader(false, false);
    }

    /**
     * @param validate if true validating parser is returned
     * @return <code>createXMLReader(validate, false);</code>
     */
    public static XMLReader createXMLReader(boolean validate) throws SAXException{
        return createXMLReader(validate, false);
    }

    /**
     * Factory SAX XMLReader using JAXP. Result can be used to parse XML files.
     * 
     * The parser has set entity resolver to system entity resolver chain.
     *
     * @param validate if true validating parser is returned
     * @param ns if true namespace aware parser is returned
     *
     * @throw FactoryConfigurationError Application developers should never need to directly catch errors of this type.
     * @throw SAXException if a parser fulfilling given parameters can not be created
     *
     * @return XMLReader configured according to passed prameters
     */
    public static XMLReader createXMLReader(boolean validate, boolean namespaceAware) throws SAXException {
        
        SAXParserFactory factory = SAXParserFactory.newInstance();

        factory.setValidating(validate);
        factory.setNamespaceAware(namespaceAware);

        try {
             return factory.newSAXParser().getXMLReader();
        } catch (ParserConfigurationException ex) {
            throw new SAXException("Cannot create parser satisfying configuration parameters", ex);  //NOI18N                        
        }
        
    }

    
    // ~~~~~~~~~~~~~~~~~~~~~ DOM related ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    
    /** 
     * Creates empty DOM Document using JAXP factoring. E.g.:
     * <p><pre>
     * Document doc = createDocument("book",null,null,null);
     * </pre><p>
     * creates new DOM of a well-formed document with root element named book.
     *
     * @param rootQName qualified name of root element. e.g. <code>myroot</code> or <code>ns:myroot</code>
     * @param namespaceURI URI of root element namespace or <code>null</code>
     * @param doctypePublicID public ID of DOCTYPE or <code>null</code>
     * @param doctypeSystemID system ID of DOCTYPE or <code>null</code> if no DOCTYPE 
     *        required and doctypePublicID is also <code>null</code>
     *
     * @throw DOMException if new DOM with passed parameters can not be created
     * @throw FactoryConfigurationError Application developers should never need to directly catch errors of this type.     
     *
     * @return new DOM Document
     */        
    public static Document createDocument(String rootQName, String namespaceURI, String doctypePublicID, String doctypeSystemID) throws DOMException {
        
        DOMImplementation impl = getDOMImplementation();

        if (doctypePublicID != null && doctypeSystemID == null) {
            throw new IllegalArgumentException("System ID cannot be null if public ID specified. ");  //NOI18N
        }
        
        DocumentType dtd = null;
        if (doctypeSystemID != null) {
            dtd = impl.createDocumentType(rootQName, doctypePublicID, doctypeSystemID);
        }
        
        return impl.createDocument(namespaceURI, rootQName, dtd);        
    }

    /**
     * Obtains DOMImpementaton interface providing a number of methods for performing 
     * operations that are independent of any particular DOM instance. 
     *
     * @throw DOMException <code>NOT_SUPPORTED_ERR</code> if can not get DOMImplementation
     * @throw FactoryConfigurationError Application developers should never need to directly catch errors of this type.          
     *
     * @return DOMImplementation implementation
     */
    private static DOMImplementation getDOMImplementation() throws DOMException { //can be made public
        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    
        try {
            return factory.newDocumentBuilder().getDOMImplementation();
        } catch (ParserConfigurationException ex) {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR , "Cannot create parser satisfying configuration parameters");  //NOI18N
        }        
    }

    /** 
     * Factory a DocumentBuilder and let it create a org.w3c.dom.Document.
     * This method takes InputSource. After successful finish the document tree is returned.
     *
     * @param input a parser input (for URL users use: <code>new InputSource(url.toExternalForm())</code>
     * @param validate if true validating parser is used
     * @param namespaceAware if true DOM is created by namespace aware parser
     * @param errorHandler a error handler to notify about exception or <code>null</code>
     * @param entityResolver - SAX entity resolver (e.g. <code>EntityCatalog.getDefault()</code>)or <code>null</code>
     *
     * @throw IOException if I/O problem during parsing occures
     * @throw SAXException is thrown if a parser error occures.
     * @throw FactoryConfigurationError Application developers should never need to directly catch errors of this type.
     *
     * @return org.w3c.dom.Document representing given inout or null if a parsing error occures
     */
    public static Document parse (
            InputSource input, 
            boolean validate, 
            boolean namespaceAware,
            ErrorHandler errorHandler,             
            EntityResolver entityResolver
        ) throws IOException, SAXException {
        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();        
        factory.setValidating(validate);
        factory.setNamespaceAware(namespaceAware);            
    
        DocumentBuilder builder = null;
        try {
             builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException ex) {
            throw new SAXException("Cannot create parser satisfying configuration parameters", ex);  //NOI18N
        }
            
        if (errorHandler != null) {
            builder.setErrorHandler(errorHandler);
        }
        
        if (entityResolver != null) {
            builder.setEntityResolver(entityResolver);
        }

        return builder.parse(input);            
    }
    
    /**
     * Write Document into OutputStream using given encoding. 
     * It is a shortcut for writing configurations etc. It guarantee 
     * just that data will be written. Structure and indentation may change.
     *
     * @param doc DOM Document to be written
     * @param out data sink     
     * @param enc - XML defined encoding name (i.e. IANA defined, one of UTF-8, UNICODE, ASCII).
     *
     * @throw IOException if an I/O exception occures
     */
    public static void write(Document doc, OutputStream out, String enc) throws IOException {
        XMLUtilImpl.write(doc, out, enc);
    }
    
}
