/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.openidex.jarpackager;

import java.util.List;
import java.io.File;
import java.util.jar.Manifest;

/** Driver that allows users to control and manipulate packaging process.
 * It contains information about what data objects were chosen to be included
 * in the archive and also contains methods for manipulating with the
 * content of the archive. ArchiveDriver holds list of entries which is then
 * used by packager tool to create the archive from.
 *
 * @deprecated Note that the interfaces and classes contained in 
 * org.openidex.jarpackager are for use with version 1 of the jarpackager 
 * module.  Starting with version 2, use the versions in 
 * org.netbeans.modules.jarpackager.api .
 * @author  Dafe Simonek
 */
public abstract class ArchiveDriver extends Object {

    /** Methods that gives information about the archive being created */

    /** Obtains list of data objects (for example directories, files)
    * chosen by the user to be included in the archive.
    * @return list of the data objects
    */
    public abstract List getDataObjects ();

    /** Obtains archive filter which should be used to filter file objects
    * that should be included in the archive.
    * @return the filter for filtering of file objects
    */
    public abstract ArchiveFilter getFilter ();

    /** Methods to manipulate the content of the archive */

    /** Adds specified entry to the entry list. Duplicates are not allowed
    * and will not be added.
    * @param entry entry to add
    * @return true if entry list changed as a result of the call,
    * false otherwsise
    */
    public abstract boolean addEntry (ArchiveEntry entry);

    /** Removes specified entry from the entry list.
    * @param entry entry to remove
    * @return true if entry list changed as a result of the call,
    * false otherwsise
    */
    public abstract boolean removeEntry (ArchiveEntry entry);

    /** Removes an entry with specified name from the entry list.
    * @param entryName name of the entry to remove
    * @return true if entry list changed as a result of the call,
    * false otherwsise */
    public abstract boolean removeEntry (String entryName);

    /** Obtains and returns list of entries.
    * @return List of the ArchiveEntry instances that will be included
    * in the resulting archive */
    public abstract List getEntries ();

    /** Sets all entries of the archive at once. Use this with care,
    * as this method replaces all old entries.
    * @param entries list of ArchiveEntry instances
    */ 
    public abstract void setEntries (List entries);

    /** Accessor for the manifest of the archive. Manifest already contains
    * information supplied by the user. Manifest modifications are possible
    * with direct modifications of returned manifest instance.
    * @return manifest of the archive
    */
    public abstract Manifest getManifest ();

    /** Registers post processor that will have opportunity to
     * post-process packaging and modify or manipulate with created archive file.
     * @param pp post processor to be notified when packaging is done 
     */
    public abstract boolean addPostProcessor(PostProcessor pp);

    /** Interface to allow post-process control over
     * already created archive.
     * @deprecated This is not preferred way to modify jar archive. All
     * modifications should be done through ArchiveController interface
     * (pre and in-process modifications). This interface may dissapear
     * in future */
    public interface PostProcessor {

        /** This method is called after archive file is written and
        * closed. Implementors must provide proper clean-up itself 
        * if they opened and modified archive file.
        * @param archiveFile archive file containing created archive
        * @return true if list of post processors changed as a result of this call
        * @deprecated All enhanced jar processing should be done via
        * ArchiveController interface. */
        public void postProcess (File archiveFile);

    } // end of PostProcessor inner interface

}
