/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects;

import java.beans.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.text.MessageFormat;

import org.openide.*;
import org.openide.compiler.CompilerJob;
import org.openide.cookies.*;
import org.openide.debugger.DebuggerException;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.windows.WindowManager;
import org.openide.windows.Workspace;
import org.openide.modules.ManifestSection;
import org.openide.util.*;
import org.openide.util.enum.*;
import org.openide.util.io.NbMarshalledObject;

import org.netbeans.modules.projects.content.*;

/**
 * The heart of the project module.
 * @author  mryzl
 */

public class PSupport implements ProjectCookie, ExecCookie, DebuggerCookie, SaveCookie {

    /** Serial version UID. */
    static final long serialVersionUID = -8907373672945248595L;
    /** Name of the method used to determine whether an option is global or not. */
    static final String GLOBAL_METHOD_NAME = "isGlobal"; // NOI18N

        /** Extended attribute dexcribing what should be stored in this project. */
    private static final String EA_SAVE = "Project-Save"; // NOI18N
    /** Extended attribute allowing deletion of objects under projects tab
     * of this project. */
    private static final String EA_DELETE = "Project-DeleteFiles"; // NOI18N
    
    private static final boolean DELETE_DEFAULT = false;

    public static final int SAVE_CONTROL_PANEL = 1;
    public static final int SAVE_LOADERS = 2;
    public static final int SAVE_MODULES = 4;
    public static final int SAVE_REPOSITORY = 8;
    public static final int SAVE_SERVICES = 16;
    public static final int SAVE_WINDOW_MANAGER = 32;
    public static final int SAVE_ALL = SAVE_CONTROL_PANEL |
                                       SAVE_LOADERS |
                                       SAVE_MODULES |
                                       SAVE_REPOSITORY |
                                       SAVE_SERVICES |
                                       SAVE_WINDOW_MANAGER;

    /** Info type. Indicates that open of something has started. */
    public static final int INFO_OPEN_START = 1;
    /** Info type. Indicates that open of something has finished. */
    public static final int INFO_OPEN_DONE = 2;
    /** Info type. Indicates that open of something has finished. */
    public static final int INFO_OPEN_ERR = 3;
    /** Info type. Indicates that open of something has started. */
    public static final int INFO_SAVE_START = 4;
    /** Info type. Indicates that open of something has finished. */
    public static final int INFO_SAVE_DONE = 5;
    /** Info type. Indicates that open of something has finished. */
    public static final int INFO_SAVE_ERR = 6;

    /** Information concerning Control Panel. */
    public static final int INFO_CONTROL_PANEL = 1;
    /** Information concerning Loaders. */
    public static final int INFO_LOADERS = 2;
    /** Information concerning Modules. */
    public static final int INFO_MODULES = 4;
    /** Information concerning Repository. */
    public static final int INFO_REPOSITORY = 8;
    /** Information concerning Services. */
    public static final int INFO_SERVICES = 16;
    /** Information concerning Window Manager. */
    public static final int INFO_WINDOW_MANAGER = 32;
    
    private static final String ATTR_GLOBAL = "global"; // NOI18N
    private static final String ATTR_VERSION = "version"; // NOI18N
    private static final String ATTR_TRANSIENT = "transient"; // NOI18N
    
    /** Global project content. */
    private static transient ProjectContent globalContent = null;
    /** Error manager used for notification. */
    private static transient ProjectErrorManager pem;
    
    /** Project project content. */
    private transient ProjectContent projectContent = null;
    private transient Node rootNode = null;

    private static boolean doNotSave = false;
    
    private static boolean checkServicesVersion = true;
    
    /** Data object that provides data for the object. */
    private DataObject dobj;
    
    /** Creates a new PSupport.
    */
    public PSupport(DataObject dobj) {
        this.dobj = dobj;
    }

    /** Get data object representing this project.
     */
    private ProjectDataObject getProjectDataObject() {
        return (ProjectDataObject) getDataObject();
    }
    
    /** Get node representing this project.
    */
    public Node projectDesktop() {
        if (rootNode == null) {
            ProjectDataObject pdo = (ProjectDataObject) getDataObject();
            Children children = Children.LEAF;
            try {
                children = pdo.getFileFolder().createNodeChildren (DataFilter.ALL);
            } catch (java.io.IOException ex) {
                TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
            }
            ProjectDataNode pdn = new ProjectDataNode(pdo, children, true);
            rootNode = new ProjectRootFilterNode(pdn, new ProjectFilterChildren(pdn));
        }
        return rootNode;
    }
    
    /** */
    private static final class ProjectFilterChildren extends FilterNode.Children {
        /** Creates new ProjectFilterChildren.  */
        public ProjectFilterChildren(Node node) {
            super(node);
        }
        
        protected Node copyNode(Node node) {
            return new ProjectFilterNode(
            node,
            (node.isLeaf()) ? org.openide.nodes.Children.LEAF: new ProjectFilterChildren(node)
            );
        }
    }

    /** Get current project error manager.
     * @return project manager
     */
    private static ProjectErrorManager getProjectErrorManager() {
        if (pem == null) pem = new ProjectErrorManager();
        return pem;
    }
    
    /**
     * Get current ProjectContent for this project.
     */
    ProjectContent getProjectContent() {
        if (projectContent == null) {
            projectContent = new XMLProjectContent((DataFolder) getDataObject(), getGlobalProjectContent(), getProjectErrorManager());
        }
        return projectContent;
    }

    /**
     * Get global ProjectContent.
     * @return ProjectContent common for all projects
     */
    static ProjectContent getGlobalProjectContent() {
        if (globalContent == null) {
            DataFolder folder = org.openide.TopManager.getDefault().getPlaces().folders().projects();
            globalContent = new XMLProjectContent(folder, null, getProjectErrorManager());
        }
        return globalContent;
    }

    /** Close the project. This method instructs the project that another project
     * is becoming the active project and that the project can drop allocated 
     * resources.
     *
     * @exception IOException if an error occurs during saving
     */
    public void projectClose () throws IOException {
        // drop resources
        projectContent = null; 
    }

    /** Initializes new project*/
    void initialize() {
        ProjectContent pc = getProjectContent();
        pc.getRepository().setClear(false);
        pc.getServices().setClear(false);
        ProjectContent gpc = getGlobalProjectContent();
        gpc.getRepository().setClear(true);
        gpc.getServices().setClear(true);

        TopManager.getDefault().getServices().setServiceTypes(null); // default services
        fillDiffSetFromServices(getProjectContent().getServices(), false);
    }

    /**
     */
    public void projectOpen() throws java.io.IOException {
        TopManager tm = TopManager.getDefault();

        tm.setStatusText(NbBundle.getMessage(PSupport.class, "MSG_OpenProject", getDataObject().getName())); // NOI18N

        // Get current content
        ProjectContent project = getProjectContent();
        
        // initialize project error manager
        getProjectErrorManager().clear();

        // for all settings set
        // create OptionProcessor for all SettingsSet and all options
        OptionProcessor.Set cpop = new OptionProcessor.Set();
        OptionProcessor.Set lpop = new OptionProcessor.Set();

        updateRepository();

        // for current project content and all parent
        for(ProjectContent pc = project; pc != null; pc = pc.getSuperProject()) {
            SettingsSet lp;
            lp = pc.getLoaderPool();
            lp.read(lpop);
            lpop.addAll(lp.getObjects());
        }
        // for current project content and all parent
        for(ProjectContent pc = project; pc != null; pc = pc.getSuperProject()) {
                SettingsSet cp;
                cp = pc.getControlPanel();
                cp.read(cpop);
                cpop.addAll(cp.getObjects());
        }

        // the flag is set to false in initialSave
        updateServices(checkServicesVersion);
        // and we want to have it true for next project open
        checkServicesVersion = true;

        // and the best ... - window manager
        NbMarshalledObject nbo = null;

        for(ProjectContent pc = project; pc != null; pc = pc.getSuperProject()) {
            if ((nbo = pc.getWindowManager()) != null) break;
        }
        
        if ((getSave() & SAVE_WINDOW_MANAGER) != 0) {
            updateWindowManager(nbo);
        }

        // shows all errors
        getProjectErrorManager().show();
        
        tm.setStatusText(NbBundle.getMessage(PSupport.class, "MSG_OpenProjectDone", getDataObject().getName())); // NOI18N
        
    }

    /**
     * Save the project.
     */
    public void projectSave() throws java.io.IOException {
        if (doNotSave) {
            // when creating new project we don't want it being saved
            // by the NbProjectOperation while openning the new one
            doNotSave = false;
            return;
        }
        
        TopManager tm = TopManager.getDefault();

        tm.setStatusText(NbBundle.getMessage(PSupport.class, "MSG_SaveProject", getDataObject().getName())); // NOI18N

        ProjectContent gpc = getGlobalProjectContent();
        SharedClassObject[] scos = null;
        int save = getSave();

        // update global project
        // update control panel
        scos = tm.getControlPanel().getSystemOptions();
        updateSettingsSet(gpc.getControlPanel(), scos);

        // update loaders
        scos = tm.getLoaderPool().toArray();
        updateSettingsSet(gpc.getLoaderPool(), scos);

        // update repository
        fillDiffSetFromRepository(gpc.getRepository(), true);
        gpc.storeRepository();
        
        // update services
        fillDiffSetFromServices(gpc.getServices(), true);
        gpc.storeServices();
        
        // update window manager
        // if null, save current manager
        
        // update project
        ProjectContent pc = getProjectContent();

        // update control panel - only project specific options
        scos = tm.getControlPanel().getSystemOptions();
        scos = getProjectObjects(scos);
        updateSettingsSet(pc.getControlPanel(), scos);

        // update loaders
        scos = tm.getLoaderPool().toArray();
        scos = getProjectObjects(scos);
        updateSettingsSet(pc.getLoaderPool(), scos);

        // update repository
        if ((save & SAVE_REPOSITORY) != 0) {
            fillDiffSetFromRepository(pc.getRepository(), false);
            pc.storeRepository();
        }

        // update services
        if ((save & SAVE_SERVICES) != 0) {
            fillDiffSetFromServices(pc.getServices(), false);
            pc.storeServices();
        }

        // create OptionProcessor for all SettingsSet and all options
        OptionProcessor.Set cpop = new OptionProcessor.Set();
        OptionProcessor.Set lpop = new OptionProcessor.Set();

        if ((save & SAVE_CONTROL_PANEL) != 0) {
            // for current project content and all parent
            for(ProjectContent pc2 = pc; pc2 != null; pc2 = pc2.getSuperProject()) {
                // save SettingsSets
                // update OptionProcessor
                SettingsSet cp = pc2.getControlPanel();
                cp.write(cpop);
                cpop.addAll(cp.getObjects());
                pc2.storeControlPanel();
            }
            
        }
        
        if ((save & SAVE_LOADERS) != 0) {
            // for current project content and all parent
            for(ProjectContent pc2 = pc; pc2 != null; pc2 = pc2.getSuperProject()) {
                SettingsSet lp = pc2.getLoaderPool();
                lp.write(lpop);
                lpop.addAll(lp.getObjects());
                pc2.storeLoaderPool();
            }
        }
        
        if ((save & SAVE_WINDOW_MANAGER) != 0) {
            pc.storeWindowManager();
        }

        tm.setStatusText(NbBundle.getMessage(PSupport.class, "MSG_SaveProjectDone", getDataObject().getName())); // NOI18N
        
    }

    /**
     * Saves the project just after creation. Does
     * not save windows and some other stuff - for full
     * save see projectSave method above.
     */
    void initialSave(boolean creatingDefault) throws java.io.IOException {
        if (!creatingDefault) {
            // this is here because we want to have the global
            // services updated
            updateServices(false);
        }
        // setting this static flag to true disables next
        // regular save (it is invoked by project open)
        doNotSave = true;
        
        TopManager tm = TopManager.getDefault();
        
        ProjectContent gpc = getGlobalProjectContent();
        SharedClassObject[] scos = null;

        // update global project
        // update control panel
        scos = tm.getControlPanel().getSystemOptions();
        updateSettingsSet(gpc.getControlPanel(), scos);

        // update loaders
        scos = tm.getLoaderPool().toArray();
        updateSettingsSet(gpc.getLoaderPool(), scos);

        // update repository
        if (creatingDefault) {
            fillDiffSetFromRepository(gpc.getRepository(), true);
        }
        gpc.storeRepository();
        
        // update services
        if (creatingDefault) {
            fillDiffSetFromServices(gpc.getServices(), true);
        }
        gpc.storeServices();
        
        // update window manager
        // if null, save current manager
        
        // update project
        ProjectContent pc = getProjectContent();
        
        // update control panel - only project specific options
        scos = tm.getControlPanel().getSystemOptions();
        scos = getProjectObjects(scos);
        updateSettingsSet(pc.getControlPanel(), scos);

        // update loaders
        scos = tm.getLoaderPool().toArray();
        scos = getProjectObjects(scos);
        updateSettingsSet(pc.getLoaderPool(), scos);

        pc.storeRepository();

        // update services
        fillDiffSetFromServices(pc.getServices(), false);
        pc.storeServices();
        
        checkServicesVersion = false;
    }

    /**
     * Set repository to current state.
     */
    private void fillDiffSetFromRepository(DiffSet diffset, boolean global) {
        Enumeration en;
        TopManager tm = TopManager.getDefault();

        diffset.clear();
        en = tm.getRepository().getFileSystems();
        while (en.hasMoreElements()) {
            FileSystem fs = (FileSystem) en.nextElement();
            if (fs.isDefault()) {
                continue;
            }
            if (isTransientFromBeanInfo(fs.getClass())) {
                continue;
            }
            if (isProjectObject(fs.getClass(), fs, true) != global) {
                diffset.add(fs.getSystemName(), fs);
            }
        }
    }

    /**
     * Set services to current state.
     */
    private void fillDiffSetFromServices(DiffSet diffset, boolean global) {
        Enumeration en;
        TopManager tm = TopManager.getDefault();

        diffset.clear();
        en = tm.getServices().services();
        while (en.hasMoreElements()) {
            ServiceType st = (ServiceType) en.nextElement();
            if (isProjectObject(st.getClass(), st, true) != global) {
                diffset.add(st.getName(), st);
            }
        }
    }

    /**
     * Update SettingsSet to new values.
     * @param set SettingsSet
     * @param scos new values
     */
    private void updateSettingsSet(SettingsSet set, SharedClassObject[] scos) {
        set.clear();
        for(int i = 0; i < scos.length; i++) {
            set.add(scos[i]);
        }
    }

    /**
     * Filter out objects that are not in project.
     * @param scos set of all objects
     * @return filtered set
     */
    public static SharedClassObject[] getProjectObjects(SharedClassObject[] scos) {
        int i,j;
        for(i = 0, j = 0; i < scos.length; i++ ){
            if (isProjectObject(scos[i].getClass(), scos[i], false)) j++;
        }
        SharedClassObject[] objs = new SharedClassObject[j];
        for(i = 0, j = 0; i < scos.length; i++ ){
            if (isProjectObject(scos[i].getClass(), scos[i], false)) objs[j++] = scos[i];
        }
        return objs;
    }

    /**
     * Test if the object is Project specific.
     * @param obj an object
     * @param clazz class from which the beaninfo is used
     * @return true if the object is Project specific
     */
    public static boolean isProjectObject(Class clazz, Object obj, boolean defaultValue) {
        try {
            if (hasGlobalAtrributeInBeanInfo(clazz)) {
                return !isGlobalFromBeanInfo(clazz);
            }
            if (obj == null) {
                return defaultValue;
            }
            // the old hack with undocumented method isGlobal
            Method m = clazz.getMethod(GLOBAL_METHOD_NAME, new Class[] {});
            m.setAccessible(true);
            Boolean b = (Boolean) m.invoke(obj, new Object[] {});
            return !b.booleanValue();
        } catch (Exception ex) {
            // ignore and return default
        }
        return defaultValue;
    }

    /** 
     * Checks the beaninfo for presence of the special
     * attribute.
     */
    private static boolean hasGlobalAtrributeInBeanInfo(Class clazz) {
        try {
            BeanInfo b = Introspector.getBeanInfo(clazz);
            if (b != null) {
                BeanDescriptor bd = b.getBeanDescriptor();
                if (bd != null) {
                    Object v = bd.getValue(ATTR_GLOBAL);
                    if (v != null) {
                        return true;
                    }
                }
                BeanInfo []infos = b.getAdditionalBeanInfo();
                if (infos != null) {
                    for (int i = 0; i < infos.length; i++) {
                        BeanDescriptor bd1 = infos[i].getBeanDescriptor();
                        if (bd1 != null) {
                            Object v = bd1.getValue(ATTR_GLOBAL);
                            if (v != null) {
                                return true;
                            }
                        }
                    }
                }
            }
        } catch (IntrospectionException ie) {
            // simply ignore and return false
        }
        return false;
    }
    
    /**
     * This method can be called only if hasGlobalAtrributeInBeanInfo
     * returns true for the given object o. It returns the
     * value from the beaninfo.
     */
    private static boolean isGlobalFromBeanInfo(Class clazz) {
        try {
            BeanInfo b = Introspector.getBeanInfo(clazz);
            if (b != null) {
                BeanDescriptor bd = b.getBeanDescriptor();
                if (bd != null) {
                    Object v = bd.getValue(ATTR_GLOBAL);
                    if (v instanceof Boolean) {
                        return ((Boolean)v).booleanValue();
                    }
                }
                BeanInfo []infos = b.getAdditionalBeanInfo();
                if (infos != null) {
                    for (int i = 0; i < infos.length; i++) {
                        BeanDescriptor bd1 = infos[i].getBeanDescriptor();
                        if (bd1 != null) {
                            Object v = bd1.getValue(ATTR_GLOBAL);
                            if (v instanceof Boolean) {
                                return ((Boolean)v).booleanValue();
                            }
                        }
                    }
                }
            }
        } catch (IntrospectionException ie) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, ie);
        }
        throw new IllegalArgumentException("Trying to get a value from beaninfo but the value is not there"); // NOI18N
    }
    
    /**
     * It returns the  value of optional attribute "version" from the beaninfo.
     */
    private static String versionFromBeanInfo(Class clazz) {
        try {
            BeanInfo b = Introspector.getBeanInfo(clazz);
            if (b != null) {
                BeanDescriptor bd = b.getBeanDescriptor();
                if (bd != null) {
                    Object v = bd.getValue(ATTR_VERSION);
                    if (v instanceof String) {
                        return (String)v;
                    }
                }
                BeanInfo []infos = b.getAdditionalBeanInfo();
                if (infos != null) {
                    for (int i = 0; i < infos.length; i++) {
                        BeanDescriptor bd1 = infos[i].getBeanDescriptor();
                        if (bd1 != null) {
                            Object v = bd1.getValue(ATTR_VERSION);
                            if (v instanceof String) {
                                return (String)v;
                            }
                        }
                    }
                }
            }
        } catch (IntrospectionException ie) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, ie);
        }
        return null;
    }
    /**
     * returns true if the class has attribute transient set to true
     */
    private static boolean isTransientFromBeanInfo(Class clazz) {
        try {
            BeanInfo b = Introspector.getBeanInfo(clazz);
            if (b != null) {
                BeanDescriptor bd = b.getBeanDescriptor();
                if (bd != null) {
                    Object v = bd.getValue(ATTR_TRANSIENT);
                    if (v instanceof Boolean) {
                        return ((Boolean)v).booleanValue();
                    }
                }
                BeanInfo []infos = b.getAdditionalBeanInfo();
                if (infos != null) {
                    for (int i = 0; i < infos.length; i++) {
                        BeanDescriptor bd1 = infos[i].getBeanDescriptor();
                        if (bd1 != null) {
                            Object v = bd1.getValue(ATTR_TRANSIENT);
                            if (v instanceof Boolean) {
                                return ((Boolean)v).booleanValue();
                            }
                        }
                    }
                }
            }
        } catch (IntrospectionException ie) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, ie);
        }
        return false;
    }
    // ====== Implementation of ExecCookie ======
    
    /**
     * Start execution of the project.
     */
    public void start() {
        ProjectDataObject df = (ProjectDataObject) getDataObject();
        DataObject dobj = MainClassHelper.getMainClass(df);
        if (dobj == null) {
            try {
                MainClassHelper.setMainClassDialog(df);
                dobj = MainClassHelper.getMainClass(df);
            } catch (java.io.IOException ex) {
                //
            }
        }
        if (dobj != null) {
            ExecCookie ec = (ExecCookie)dobj.getCookie(ExecCookie.class);
            if (ec != null) {
                ec.start();
            }
        }
    }

    // ===== Implementation of DebuggerCookie =====

    /**
     * Start debugging of associated object.
     * @param stopOnMain if <code>true</code>, debugger stops on the first line of debugged code
     * @exception DebuggerException if the session cannot be started
     */
    public void debug(boolean stopOnMain) throws DebuggerException {
        ProjectDataObject df = (ProjectDataObject) getDataObject();
        DataObject dobj = MainClassHelper.getMainClass(df);
        if (dobj == null) {
            try {
                MainClassHelper.setMainClassDialog(df);
                dobj = MainClassHelper.getMainClass(df);
            } catch (java.io.IOException ex) {
                //
            }
        }
        if (dobj != null) {
            DebuggerCookie dc = (DebuggerCookie)dobj.getCookie(DebuggerCookie.class);
            if (dc != null) {
                dc.debug(stopOnMain);
            }
        }
    }

    // ===== Implementation of CompilerCookie =====

    CompilerCookie getCompileCookie(Class c) {
        if (CompilerCookie.Compile.class.isAssignableFrom(c) ||
            CompilerCookie.Build.class.isAssignableFrom(c) ||
            CompilerCookie.Clean.class.isAssignableFrom(c)) {
            return new CompileCookie(c);
        } else {
            return null;
        }
    }

    // CompilerCookie

    private final class CompileCookie
    implements CompilerCookie.Compile, CompilerCookie.Build, CompilerCookie.Clean {

        private CompilerCookie cookie;

        /** Get delegatee CompilerCookie.
         * @return CompilerCookie or null
         */
        public CompileCookie(Class c) {
            ProjectDataObject pdo = getProjectDataObject();
            try {
                cookie = (CompilerCookie) pdo.getFileFolder().getCookie(c);
            } catch (Exception ex) {
                TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
            }
        }

        public void addToJob(org.openide.compiler.CompilerJob job,
                             org.openide.compiler.Compiler.Depth depth) {

            if (cookie != null) {
                cookie.addToJob (job, org.openide.compiler.Compiler.DEPTH_INFINITE);
            }
        }

        public boolean isDepthSupported(org.openide.compiler.Compiler.Depth depth) {
            return true;
        }
    }
    
   // ===== OpenCookie implementation =====

    /**
     * Opens project
     */
    public void open() {
//       RequestProcessor.postRequest(new Runnable() {
//            public void run() {
                 try {
                     org.openide.TopManager.getDefault().openProject(PSupport.this);
                 } catch (UserCancelException ex) {
                     // user canceled -> nothing to do
                 } catch (IOException ex) {
                     TopManager.getDefault().getErrorManager().notify(ex);
                 }
//            }
//        });
    }

    // ===== SaveCookie implementation =====

    /**
     * Save project.
     */
    public void save() throws java.io.IOException {
        RequestProcessor.postRequest(new Runnable() {
            public void run() {
                try {
                    projectSave();
                } catch (IOException ex) {
                    TopManager.getDefault().getErrorManager().notify(ex);
                }
            }
        });
    }

    // serialization
    
    private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
        stream.defaultReadObject();
        
        // workaround for #10945; tries to obtain deserialized project DataObject
        // if it is impossible then tries create or use default project.
        if (dobj == null && stream instanceof CompatibilityStream) {
            dobj = ((CompatibilityStream) stream).getDataObject();
            
            if ((dobj == null) || (!(dobj instanceof ProjectDataObject)) || (!dobj.getPrimaryFile().isValid())) {
                dobj = ProjectsModule.createDefaultProject();
                getProjectDataObject().support.initialize();
            }
        } else {
            if (dobj != null && ((!(dobj instanceof ProjectDataObject)) || (!dobj.getPrimaryFile().isValid()))) {
                dobj = ProjectsModule.createDefaultProject();
                getProjectDataObject().support.initialize();
            }
        }
    }
    
    /** workaround for #10945; if serialized version of PSupport in project.last is 
     * incompatible (< NB3.2) then the method tries to read the file again with
     * CompatibilityStream.
     */
    private Object readResolve() throws ObjectStreamException {
        PSupport project = this;
        
        if (dobj == null) {
            FileObject fo = getLastProjectFile();
            if (fo != null) {
                try {
                    CompatibilityStream ois = new CompatibilityStream(fo.getInputStream ());
                    project = (PSupport) ois.readObject();
                } catch (IOException ex) {
                    InvalidObjectException ioex =
                        new InvalidObjectException("Cannot resolve object."); // NOI18N
                    TopManager.getDefault().getErrorManager().annotate(ioex, ex);
                } catch (ClassNotFoundException ex) {
                    InvalidObjectException ioex =
                        new InvalidObjectException("Cannot resolve object."); // NOI18N
                    TopManager.getDefault().getErrorManager().annotate(ioex, ex);
                }
            }
        }
        return project;
    }
    
    /** workaround for #10945; Stream solving incompatibility PSupport classes
     * (NB3.1 => NB3.2).
     */
    private class CompatibilityStream extends org.openide.util.io.NbObjectInputStream {
        private DataObject dobj;
        
        public CompatibilityStream(InputStream in) throws IOException, StreamCorruptedException {
            super(in);
        }
        
        protected Object resolveObject(Object obj) throws IOException {
            if (obj instanceof org.netbeans.modules.projects.ProjectDataObject) {
                dobj = (DataObject) obj;
            }
            
            return obj;
        }
        /** Gets project DataObject. Can be <code>null</code>.*/
        public DataObject getDataObject() {
            return dobj;
        }
    }
    
    /** Returns file object (file) where the name of last opened project is
    * stored. It is part of workaround for #10945.
    * @param create true if the file should be created
    * @return the file object
    */
    private FileObject getLastProjectFile() {
        org.openide.filesystems.FileSystem def =
            TopManager.getDefault().getRepository().getDefaultFileSystem ();

        return def.find ("", // NOI18N
                        ProjectsModule.PROJECT_LAST_NAME,
                        ProjectsModule.PROJECT_LAST_EXT
        );

    }
    
    /**
     * Get asociated entry.
     */
    private DataObject getDataObject() {
        return dobj;
    }

    /**
     */
    private void updateWindowManager(NbMarshalledObject nbo) {
        try {
            resetWindowManager(true);
            if (nbo != null) {
                nbo.get();
            } 
        } catch (ThreadDeath td) {
            throw td;
        } catch (Throwable th) {
            getProjectErrorManager().annotate(ProjectErrorManager.TYPE_WM, "Error deserializing window system.", th); // NOI18N
            resetWindowManager(true);
        }
    }
    
    /** ResetWindowManager. */
    public static void resetWindowManager(boolean createFromScratch) {
        WindowManager wm = TopManager.getDefault().getWindowManager();
        try {
            // remove Workspaces
            Workspace[] ws = wm.getWorkspaces();
            for(int i = 0; i < ws.length; i++) {
                ws[i].remove();
            }
            if (createFromScratch) {
                Method m = wm.getClass().getMethod("createFromScratch", new Class[] {}); // NOI18N
                m.invoke(null, new Object[] {});
            }
        } catch (Exception ex) {
            TopManager.getDefault().getErrorManager().notify(ex);
        }
    }


    /**
     * Apply diffset on repository.
     */
    private void updateRepository() {
        Stack stack = new Stack();
        List list = new LinkedList();   // list of names of filesystems
        Map map = new HashMap();  // names and filesystems
        DiffSet diffset = null;
        Enumeration en;
        Iterator it;
        FileSystem fs;

        for(ProjectContent pc = getProjectContent(); pc != null; pc = pc.getSuperProject()) {
            diffset = pc.getRepository();
            if ((pc == getProjectContent()) && ("1.0".equals(diffset.getVersion()))) {
                // fixing wrong clear attribute for old filesystems
                //
                diffset.setClear(false);
            }
            stack.push(diffset);
            if (diffset.isClear()) break;
        }

        Repository repository = org.openide.TopManager.getDefault().getRepository();

        if (diffset != null) {
            if (!diffset.isClear()) {
                // get filesystems from repository
                en = repository.fileSystems();
                while (en.hasMoreElements()) {
                    fs = (FileSystem) en.nextElement();
                    if (!fs.isDefault()) {
                        String fsname = fs.getSystemName();
                        list.add(fsname);
                        map.put(fsname, fs);
                    }
                }
            }

            while (!stack.empty()) {
                diffset = (DiffSet) stack.pop();
                // remove items
                it = diffset.removedItems().iterator();
                while (it.hasNext()) {
                    Object obj = it.next();
                    list.remove(obj);
                    map.remove(obj);
                }

                // add items
                it = diffset.addedItems().iterator();
                while (it.hasNext()) {
                    Object obj = it.next();
                    if (!list.contains(obj)) {
                        list.add(obj);
                        map.put(obj, diffset.addedItem((String) obj));
                    }
                }
            }
            
            clearRepository();

            // add new content
            it = list.iterator();
            while (it.hasNext()) {
                String fsname = (String) it.next();
                if (repository.findFileSystem( fsname ) == null) {
                    fs = (FileSystem) map.get(fsname);
                    // workaround for #10991
                    // translate {netbeans.user} to real path
                    if (fs instanceof LocalFileSystem &&
                            fsname.indexOf("{netbeans.user}") == 0) { // NOI18N
                        fsname = ((LocalFileSystem) fs).getRootDirectory()
                                 .getAbsolutePath()
                                 .replace(File.separatorChar, '/');
                        if (repository.findFileSystem( fsname ) != null) {
                            continue;
                        }
                    }
                    
                    if (fs != null) repository.addFileSystem(fs);
                }
            }
        }
    }

    private void clearRepository() {
        Repository repository = TopManager.getDefault().getRepository();
        Enumeration en = repository.fileSystems();
        while (en.hasMoreElements()) {
            FileSystem fs = (FileSystem) en.nextElement();
            if (fs.isDefault()) {
                continue;
            }
            if (isProjectObject(fs.getClass(), fs, true)) {
                repository.removeFileSystem(fs);
            }
        }
    }
    
    /**
     * Apply diffset on services.
     */
    private void updateServices(boolean checkForOldProject) {
        Stack stack = new Stack();
        List list = new LinkedList();   // list of names of filesystems
        Map map = new HashMap();
        DiffSet diffset = null;
        Enumeration en;
        Iterator it;
        FileSystem fs;
        for(ProjectContent pc = getProjectContent(); pc != null; pc = pc.getSuperProject()) {
            diffset = pc.getServices();
            if ((pc == getProjectContent()) && ("1.0".equals(diffset.getVersion()))) {
                // fixing wrong clear attribute for old services
                //
                diffset.setClear(false);
            }
            stack.push(diffset);
            if (diffset.isClear()) break;
        }
        ServiceType.Registry registry = TopManager.getDefault().getServices();
        if (diffset != null) {
            if (!diffset.isClear()) {
                // get current services
                en = registry.services();
                while (en.hasMoreElements()) {
                    ServiceType st = (ServiceType) en.nextElement();
                    String name = st.getName();
                    list.add(name);
                    map.put(name, st);
                }
            }

            while (!stack.empty()) {
                diffset = (DiffSet) stack.pop();
                // remove items
                it = diffset.removedItems().iterator();
                while (it.hasNext()) {
                    String name = (String) it.next();
                    list.remove(name);
                    map.remove(name);
                }

                // add items
                it = diffset.addedItems().iterator();
                while (it.hasNext()) {
                    String name = (String) it.next();
                    Object instance = diffset.addedItem(name);
                    if (instance instanceof org.openide.text.IndentEngine) {
                        // fix for #11189
                        // we should ignore indent engines saved as
                        // part of a project
                        if (diffset == getProjectContent().getServices()) {
                            continue;
                        }
                    }
                    if (!list.contains(name)) {
                        list.add(name);
                        map.put(name, diffset.addedItem(name));
                    } 
                }
            }
            if (checkForOldProject) {
                String projectVersion = getProjectContent().getServices().getVersion();

                if (!XMLProjectContent.VERSION.equals(projectVersion)) {
                    Lookup.Result r = Lookup.getDefault().lookup(
                        new Lookup.Template(ManifestSection.ServiceSection.class)
                    );
                    // add items
                    it = r.allInstances().iterator();
                    while (it.hasNext()) {
                        ManifestSection.ServiceSection ss = (ManifestSection.ServiceSection)it.next();
                        ServiceType instance = null;
                        try {
                            instance = ss.getServiceType();
                        } catch (InstantiationException ex) {
                            TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
                            continue;
                        }
                        String name = instance.getName();
                        String version = versionFromBeanInfo(instance.getClass());
                        if (instance instanceof org.openide.text.IndentEngine) {
                              // fix for #11189
                              // indent engines are added from manifest regardless of version
                              //  (because they were changed from project to global)
                            if (!list.contains(name)) {
                                list.add(name);
                                map.put(name, instance);
                                continue;
                            }
                        }
                        if (version == null) {
                            continue;
                        }
                        if ((!list.contains(name))&&(XMLSupport.isNew(version, projectVersion))) {
                            list.add(name);
                            map.put(name, instance);
                        }
                    }
                }
            }

            List services = new ArrayList(list.size());
            it = list.iterator();
            while (it.hasNext()) {
                Object service = map.get(it.next());
                services.add(service);
            }
            registry.setServiceTypes(services);
        }
    }

    /** Getter for property save.
     * @return Value of property save.
     */
    public int getSave() {
        Integer save = (Integer) dobj.getPrimaryFile().getAttribute(EA_SAVE);
        return (save != null) ? save.intValue(): SAVE_ALL;
    }

    /** Setter for property save.
     * @param save New value of property save.
     */
    public void setSave(int save) throws java.io.IOException {
        dobj.getPrimaryFile().setAttribute(EA_SAVE, new Integer(save));
    }
    
    /** Getter for attribute allowing deletion of objects under projects tab.
     * @return if True, delete files from the system when deleting from the project.
     */
    public boolean getDelete() {
        Boolean delete = (Boolean) dobj.getPrimaryFile().getAttribute(EA_DELETE);
        return (delete != null) ? delete.booleanValue(): DELETE_DEFAULT;
    }
    
    /** Setter for attribute allowing deletion of objects under projects tab.
     * @param delete if True, delete files from the system when deleting from the project.
     * @throws IOException if the attribute cannot be set.
     */
    public void setDelete(boolean delete) throws java.io.IOException {
        dobj.getPrimaryFile().setAttribute(EA_DELETE, new Boolean(delete));
    }

    /**
     */
    private String getString(String key) {
        return NbBundle.getBundle(PSupport.class).getString(key);
    }
}
