/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects;

import java.io.*;
import java.util.Enumeration;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.cookies.*;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.*;

import org.netbeans.modules.projects.content.*;

/** Object that provides main functionality for project data loader.
* This class is final only for performance reasons,
* can be unfinaled if desired.
*
* @author Jaroslav Tulach
*/
public final class ProjectDataObject extends DataFolder {

    static final long serialVersionUID = 3984322918015509710L;

    public static final String FILES_FOLDER = "Files"; // NOI18N

    final PSupport support;

    public ProjectDataObject (FileObject obj, DataLoader loader) throws DataObjectExistsException {
        super (obj, loader);
        support = new PSupport(this);
    }

    /** Creates new project in specified folder under given name.
    * @param f the folder
    * @param n name of the project
    * @return the project data object
    * @exception IOException if it fails
    */
    public static ProjectDataObject createProject(final DataFolder f, final String n)
    throws IOException {
        final FileObject[] primary = new FileObject[1];
        final FileObject pf = f.getPrimaryFile ();

        pf.getFileSystem ().runAtomicAction (new FileSystem.AtomicAction () {
                                                 public void run () throws IOException {
                                                     // find free name
                                                     String name = FileUtil.findFreeFolderName (
                                                                       f.getPrimaryFile (), n
                                                                   );

                                                     FileObject ff = FileUtil.createFolder (pf, name);
                                                     createFileFolder(ff);

                                                     if (ProjectsModule.DEFAULT_PROJECT_NAME.equals (name)) {
                                                         try {
                                                             ff.setAttribute ("SystemFileSystem.localizingBundle", "org.netbeans.modules.projects.Bundle"); // NOI18N
                                                         } catch (IOException ioe) {
                                                             if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                                                                 ioe.printStackTrace ();
                                                         }
                                                     }

                                                     // mark the folder as belonging to project loader
                                                     ProjectDataLoader.INSTANCE.markFile (ff);
                                                     primary[0] = ff;
                                                 }
                                             });
        DataObject obj1 = DataObject.find(primary[0]);
        if (!(obj1 instanceof ProjectDataObject)) {
            // the code should NOT normally end up here
            TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL,
                new FileStateInvalidException("Project data object not recognized " + primary[0].getPackageName('/'))); // NOI18N
            try {
                // trying to force new recognition by ProjectDataLoader
                obj1.setValid(false);
            } catch (Exception x) {
                TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, x);
            }
            obj1 = DataObject.find(primary[0]);
        }
        ProjectDataObject obj = (ProjectDataObject)obj1;
        return obj;
    }

    /** Add DataObject to the project.
    */
    public void add(DataObject dobj) throws IOException {
        dobj.createShadow(getFileFolder());
    }

    /**
    */
    protected DataFolder getFileFolder() throws java.io.IOException {
        FileObject pf = getPrimaryFile();
        FileObject ff = pf.getFileObject(FILES_FOLDER);
        if (ff == null) {
            ff = createFileFolder(pf);
        }
        return (DataFolder) DataObject.find(ff);
    }

    /**
    * @param pf - file object representing project folder
    */
    protected static FileObject createFileFolder(FileObject pf) throws java.io.IOException {
        FileObject ff = pf.createFolder(FILES_FOLDER);
        try {
            ff.setAttribute ("SystemFileSystem.localizingBundle", "org.netbeans.modules.projects.ProjectDataObject$FilesFolderBundle"); // NOI18N
        } catch (IOException ioe) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                ioe.printStackTrace ();
        }
        return ff;
    }

    public static class FilesFolderBundle extends java.util.ResourceBundle {
        private final String FILES_FOLDER_LOC = NbBundle.getBundle (ProjectDataObject.class).getString ("LBL_files_folder");
        public java.util.Enumeration getKeys () {
            return org.openide.util.enum.EmptyEnumeration.EMPTY;
        }
        protected Object handleGetObject (String key) {
            return FILES_FOLDER_LOC;
        }
    }
    
    /** Copy this object to a folder.
     * The copy of the object is required to
    * be deletable and movable.
    *
    * @param f the folder to copy object to
    * @exception IOException if something went wrong
    * @return the new object
    */
    protected DataObject handleCopy (DataFolder f) throws IOException {
        String name = getPrimaryFile ().getName ();
        DataFolder newFolder = createProject(f, name);
        copyChildren(this, newFolder);
        return newFolder;
    }

    /** Test if this DataObject is used as a current project.
     * @return true if node of this DO is the same as system project node.
     */
    protected boolean isCurrentProject() {
        Node node = TopManager.getDefault ().getPlaces ().nodes ().projectDesktop ();
        return (node != null) && (node.getCookie(ProjectDataObject.class) == this);
    }
    
    /** Delete this object. It must not be current project.
    * @exception IOException if something went wrong
     */
    protected void handleDelete() throws IOException {
        if (!isCurrentProject()) super.handleDelete();
    }
    
    /** Copy all children of a data folder into another.
     */
    public static void copyChildren(DataFolder src, DataFolder target) throws IOException {
        Enumeration en = src.children ();

        while (en.hasMoreElements ()) {
            try {
                DataObject obj = (DataObject)en.nextElement ();
                if (obj.isCopyAllowed()) { 
                    obj.copy (target);
                } else {
                    // data object can not be copied, inform user
                    TopManager.getDefault().getStdOut().println(
                        java.text.MessageFormat.format(
                            NbBundle.getBundle(DataFolder.class).getString("FMT_CannotCopyDo"),
                            new Object[] { obj.getName() }
                        )
                    );
                }
            } catch (IOException ex) {
                TopManager.getDefault ().notifyException (ex);
            }
        }

        try {
            FileUtil.copyAttributes (src.getPrimaryFile (), target.getPrimaryFile());
        } catch (IOException ioe) {
            TopManager.getDefault ().notifyException (ioe);
        }

    }

    /** Provides node that should represent this data object. When a node for representation
    * in a parent is requested by a call to getNode (parent) it is the exact copy of this node
    * with only parent changed. This implementation creates instance
    * <CODE>DataNode</CODE>.
    * <P>
    * This method is called only once.
    *
    * @return the node representation for this data object
    * @see DataNode
    */
    protected Node createNodeDelegate () {
        return new ProjectDataNode(this);
    }

    public boolean isDeleteAllowed () {
        return super.isDeleteAllowed () && !isCurrentProject();
    }

    public boolean isMoveAllowed () {
        return isDeleteAllowed ();
    }

    /** Special handling for compile cookie
    */
    public Node.Cookie getCookie (Class c) {
        if (CompilerCookie.class.isAssignableFrom(c)) {
            return support.getCompileCookie(c);
        }
        if (c.isInstance(support)) return support;
        if (getClass().equals(c)) return this;
        // return files data folder if somebody wants data folder cookie
        if (DataFolder.class.isAssignableFrom(c)) {
            try {
                return getFileFolder();
            } catch (IOException exc) {
                if (System.getProperty("netbeans.debug.exceptions") != null) {
                    exc.printStackTrace();
                }
                return null;
            }
        }

        return super.getCookie (c);
    }

    /** Help context for this object.
    * @return help context
    */
    public HelpCtx getHelpCtx () {
        return new HelpCtx (ProjectDataObject.class);
    }

    /** Getter for bundle string.
    */
    static String getLocalizedString (String s) {
        return NbBundle.getBundle (ProjectDataObject.class).getString (s);
    }

    /** Getter for bundle string.
    */
    static String getLocalizedString (String s, Object val) {
        return java.text.MessageFormat.format (getLocalizedString (s), new Object[] { val });
    }

    /** Test whether the object is accessible from folder.
    */
    public boolean isAccessibleFromFolder(DataObject dobj) throws IOException {
        return isAccessibleFromFolder(getFileFolder(), dobj);
    }

    /** Test whether the object is accessible from folder.
    */
    public static boolean isAccessibleFromFolder(DataFolder folder, DataObject dobj) {
        // get all objects under FilesFolder
        // dobj must be either the non-folder object or the folder should be
        // somewhere on path to the root
        DataObject dobjs[] = folder.getChildren();

        for(int i = 0; i < dobjs.length; i++) {
            FileObject fo = getOriginal(dobj).getPrimaryFile(),
                            ff = getOriginal(dobjs[i]).getPrimaryFile();
            while (fo != null) {
                if (fo.equals(ff)) return true;
                fo = fo.getParent();
            }
        }
        return false;
    }

    /** De-shadow the object.
    * @return return the original if the obj is DataShadow, otherwise return obj
    */
    public static DataObject getOriginal(DataObject obj) {
        while (obj instanceof DataShadow) obj = ((DataShadow) obj).getOriginal();
        return obj;
    }
}
