package org.netbeans.modules.projects;

import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import org.openide.*;
import org.openide.nodes.*;
import org.openide.loaders.*;
import org.openide.util.*;
import org.openide.util.actions.CallableSystemAction;
import org.openide.cookies.ProjectCookie;
import org.openide.util.actions.SystemAction;

/** 
 * Presents dialog allowing to manipulate with projects.
 * Action that can always be invoked and work procedurally.
 * @author  dstrupl
 */
public class ProjectManagerAction extends CallableSystemAction {

    /**
     * Shows ProjectManagerPanel in a dialog.
     */
    public void performAction () {
        final ProjectManagerPanel panel = new ProjectManagerPanel();
        panel.getExplorerManager().setRootContext(getProjectRoot(new ProjectFilter()));
        final Object [] options = new Object [] {
            DialogDescriptor.CLOSED_OPTION,
        };

        final DialogDescriptor descriptor = new DialogDescriptor (
            panel,
            getString("CTL_ProjectManager_Title"),
            true,
            options,
            options[0],
            DialogDescriptor.BOTTOM_ALIGN,
            new HelpCtx(ProjectManagerAction.class), // HelpCtx
            null // Action Listener
        );
        
        javax.swing.SwingUtilities.invokeLater(
            new Runnable() {
                public void run() {
                    final Dialog d = TopManager.getDefault().createDialog(descriptor);
                    // Button Open will close the dialog
                    panel.addPropertyChangeListener(
                        ProjectManagerPanel.PROP_CLOSE_DIALOG,
                        new PropertyChangeListener() {
                            public void propertyChange(PropertyChangeEvent evt) {
                                javax.swing.SwingUtilities.invokeLater(
                                    new Runnable() {
                                        public void run() {
                                            d.dispose();
                                        }
                                    });
                            }
                        }
                    );
                    d.show();
                }
            }
        );
        
    }

    public String getName () {
        return getString("LBL_ProjectManagerAction");
    }

    protected String iconResource () {
        return "/org/netbeans/modules/projects/resources/newProject.gif"; // NOI18N
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (ProjectManagerAction.class);
    }

    /** Perform extra initialization of this action's singleton.
     * PLEASE do not use constructors for this purpose!
    protected void initialize () {
	super.initialize ();
        putProperty (Action.SHORT_DESCRIPTION, NbBundle.getMessage (ProjectManagerActionAction.class, "HINT_Action"));
    }
    */

    /** 
     * Create a hierarchy of nodes.
     * @return a node representing the root of the hierarchy
     */
    private static Node getProjectRoot (DataFilter ff) {
        DataFolder f = TopManager.getDefault ().getPlaces ().folders ().projects ();
        // listener used as filter (has method acceptDataObject)
        Children ch = f.createNodeChildren (ff);
        // filter the children
        ch = new ProjectChildren (new AbstractNode (ch));
        // create the root
        return new DoNothingFilterNode(f.getNodeDelegate (), ch);
    }

    /** */
    private static class ProjectFilter extends Object
        implements NodeAcceptor, DataFilter {
        /** Data filter impl.
        */
        public boolean acceptDataObject (DataObject obj) {
            return obj.getCookie (ProjectCookie.class) != null;
        }

        public boolean acceptNodes (Node[] n) {
            if (n.length != 1) {
                return false;
            }
            return n[0].getCookie (ProjectCookie.class) != null;
        }
    }

    /** Filter node children, that stops on data objects (does not go futher)
     */
    private static class ProjectChildren extends FilterNode.Children {
        public ProjectChildren (Node or) {
            super (or);
        }

        protected Node copyNode (Node n) {
            DataFolder df = (DataFolder)n.getCookie (DataFolder.class);
            if (df == null) {
                // on normal nodes stop recursion
                return new DoNothingFilterNode (n, LEAF);
            } else {
                // on folders use normal filtering
                return new DoNothingFilterNode (n, new ProjectChildren (n));
            }
        }
    }
    
    /* Do nothing? ;-) */
    private static class DoNothingFilterNode extends FilterNode {
        
        public DoNothingFilterNode(org.openide.nodes.Node node, org.openide.nodes.Children children) {
            super(node, children);
        }
        
        public DoNothingFilterNode(org.openide.nodes.Node node) {
            super(node);
        }

        public boolean canCopy() {
            return false;
        }
        
        public boolean canCut()  {
            return false;
        }

        public boolean canRename() {
            return false;
        }
        
        public SystemAction getDefaultAction() {
            return null;
        }
        
        public SystemAction[] getActions () {
            return new SystemAction[0];
        }

        public SystemAction[] getContextActions () {
            return new SystemAction[0];
        }
    }

    private static String getString(String key) {
        try {
            return NbBundle.getMessage (ProjectManagerAction.class, key);
        } catch (Exception x) {
        }
        return "Bundle not available";
    }
}
