/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects;

import java.io.*;
import java.util.LinkedList;

import org.openide.TopManager;
import org.openide.actions.*;
import org.openide.cookies.ExecCookie;
import org.openide.cookies.ProjectCookie;
import org.openide.execution.ThreadExecutor;
import org.openide.loaders.*;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.*;
import org.openide.modules.ModuleInstall;
import org.openide.util.RequestProcessor;
import org.openide.util.io.NbMarshalledObject;

import org.netbeans.modules.projects.settings.*;
import org.netbeans.modules.projects.content.XMLProjectContent;

/**
 * @author Ian Formanek
 */
public class ProjectsModule extends ModuleInstall {

    static int numberOfStarts = 0;

    public static final String DEFAULT_PROJECT_NAME = "Default"; // NOI18N
    public static final String PROJECT_LAST_NAME = "project"; // NOI18N
    public static final String PROJECT_LAST_EXT = "last"; // NOI18N

    /** */
    private static final int WAIT_PISHVEITSH = 650;

    static final long serialVersionUID = -8515854707471314270L;

    private static final boolean DEBUG = true;

    /** This variable indicates that the initial filesystem was created  */
    static boolean defaultProjectCreated = false;

    /**
     * Module initialization for the first time.
     */
    public void restored () {
        numberOfStarts++;

        try {
            if (!lastExists()) createDefaultProject();
        } catch (Exception ex) {
            TopManager.getDefault().getErrorManager().notify(ex);
        }

        // save default workspaces
        NbMarshalledObject nbo = null;
        nbo = PSupport.getGlobalProjectContent().getWindowManager();
        if (nbo == null) {
            final java.awt.Window mainWindow =  TopManager.getDefault ().getWindowManager ().getMainWindow ();
            mainWindow.addWindowListener(
                new java.awt.event.WindowAdapter () {
                    public void windowOpened (java.awt.event.WindowEvent ev) {
                        // notify me no more
                        mainWindow.removeWindowListener(this);
                        // wait
                        RequestProcessor.postRequest(new WMSaver(), WAIT_PISHVEITSH);
                    }
                }
            );
        }
    }

    /**
     * Module installed for the first time.
     */
    public void installed () {
        // everything should be handled in restore
    }

    /**
     * Module was uninstalled.
     */
    public void uninstalled () {
        removeLast();
    }

    // -----------------------------------------------------------------------------------
    // Private methods

    static ProjectDataObject createDefaultProject() throws IOException {
        DataFolder projects = org.openide.TopManager.getDefault().getPlaces().folders().projects();
        FileObject folder = projects.getPrimaryFile();
        FileObject fo = folder.getFileObject(DEFAULT_PROJECT_NAME);
        ProjectDataObject pdo = null;
        if (fo != null) {
            pdo = (ProjectDataObject) DataObject.find(fo);
        } else {
            pdo = ProjectDataObject.createProject(projects, DEFAULT_PROJECT_NAME);
            pdo.support.initialize();
            pdo.support.initialSave(true);
            pdo.support.open();
            defaultProjectCreated = true;
        }
        saveLast(pdo.support);
        return pdo;
    }

    static boolean lastExists() {
        FileSystem fs = org.openide.TopManager.getDefault().getRepository().getDefaultFileSystem();
        FileObject system = fs.getRoot();
        FileObject last = system.getFileObject(PROJECT_LAST_NAME, PROJECT_LAST_EXT);
        return last != null;
    }

    static void saveLast(ProjectCookie project) {
        FileLock lock = null;
        ObjectOutputStream oos = null;
        try {
            FileSystem fs = org.openide.TopManager.getDefault().getRepository().getDefaultFileSystem();
            FileObject system = fs.getRoot();
            FileObject last = system.getFileObject(PROJECT_LAST_NAME, PROJECT_LAST_EXT);

            if (last == null) {
                last = system.createData(PROJECT_LAST_NAME, PROJECT_LAST_EXT);
                lock = last.lock();
                oos = new ObjectOutputStream(last.getOutputStream(lock));
                oos.writeObject(project);
            }
        } catch (IOException ex) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) ex.printStackTrace (); // NOI18N
        } finally {
            if (lock != null) lock.releaseLock();
            try {
                if (oos != null) oos.close();
            } catch (IOException ex) {
                if (Boolean.getBoolean ("netbeans.debug.exceptions")) ex.printStackTrace (); // NOI18N
            }
        }
    }

    static void removeLast() {
        FileLock lock = null;
        try {
            FileSystem fs = org.openide.TopManager.getDefault().getRepository().getDefaultFileSystem();
            FileObject system = fs.getRoot();
            FileObject last = system.getFileObject(PROJECT_LAST_NAME, PROJECT_LAST_EXT);
            if (last != null) {
                lock = last.lock();
                last.delete(lock);
            }
        } catch (IOException ex) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) ex.printStackTrace (); // NOI18N
        } finally {
            if (lock != null) lock.releaseLock();
        }
    }

    /** */
    private static class WMSaver implements Runnable {
        public void run() {
            PSupport.getGlobalProjectContent().storeWindowManager();
        }
    }

    // Implementation of java.io.Externalizable ------------------

    public void readExternal(final java.io.ObjectInput objectInput )
    throws java.io.IOException, java.lang.ClassNotFoundException {

        super.readExternal( objectInput );

        numberOfStarts = objectInput.readInt();

    }

    public void writeExternal(final java.io.ObjectOutput objectOutput )
    throws java.io.IOException {
        super.writeExternal( objectOutput );

        objectOutput.writeInt( numberOfStarts );
    }
}
