/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects.content;

import java.util.*;
import java.awt.datatransfer.*;
import java.text.MessageFormat;
import java.io.IOException;
import javax.swing.*;

import org.openide.*;
import org.openide.nodes.*;
import org.openide.loaders.DataFolder;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;
import org.openide.util.HelpCtx;
import org.openide.util.datatransfer.ExTransferable;

/**
 * We store all exceptions during deserialization
 * here. They can be shown all at once by calling
 * show().
 * @author  mryzl, David Strupl
 */
public class ProjectErrorManager extends Object {
    
    public static final int TYPE_FILE = 1;
    public static final int TYPE_SCO = 2;
    public static final int TYPE_ITEM = 3;
    public static final int TYPE_WM = 5;
    
    static final String PROP_ERROR_DETAILS = "details"; // NOI18N
    
    private List exceptions = new ArrayList();
    private AllErrorsNode errorsRootNode;
    
    private ProjectErrorsPanel panel;
    
    /** Creates new ProjectErrorManager */
    public ProjectErrorManager() {
    }
    
    /** Add an exception that can occur during one
     * project operation.
     *
     * @param type - type of the exception
     * @param t - an exception
     */
    public void annotate(int type, String name, Throwable t) {
        annotate(type, name, t, true);
    }
    
    /** Add an exception that can occur during one
     * project operation.
     *
     * @param type - type of the exception
     * @param t - an exception
     * @param notify - also notify using ErrorManager
     */
    public void annotate(int type, String name, Throwable t, boolean notify) {
        ListEntry li = new ListEntry(type, name, t);
        exceptions.add(li);
        getErrorsRootNode().getChildren().add(new Node[] { new ErrorNode(li)});
        if (notify) {
            TopManager.getDefault().getErrorManager().annotate(t, li.toString());
            TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, t);
        }
    }
    /**
     * Clear all.
     */
    public void clear() {
        exceptions.clear();
        errorsRootNode = null;
    }
    
    /** Shows the dialog with accumulated exceptions.*/
    public void show() {
        if (! isEmpty() ) {
            ProjectErrorsPanel p = getPanel();
            p.getExplorerManager().setRootContext(getErrorsRootNode());
            p.updateDescription();

            DialogDescriptor dd = new DialogDescriptor(p, getString("CTL_ErrorsDialogTitle"));
            dd.setModal(false);
            java.awt.Dialog d = TopManager.getDefault().createDialog(dd);
            d.show();
        }
    }
    
    public boolean isEmpty() {
        return exceptions.isEmpty();
    }
    
    /** Lazy init of the JPanel */
    private ProjectErrorsPanel getPanel() {
        if (panel == null) {
            panel = new ProjectErrorsPanel();
        }
        return panel;
    }
    
    /** */
    private AllErrorsNode getErrorsRootNode() {
        if (errorsRootNode == null) {
            errorsRootNode = new AllErrorsNode();
        }
        return errorsRootNode;
    }
    
    private static String getString(String key) {
        try {
            return NbBundle.getBundle(ProjectErrorManager.class).getString(key);
        } catch (MissingResourceException mre) {
            // [PENDING] remove this (the whole try)
            mre.printStackTrace();
        }
        return "Missing resource";
    }
    
    /** Data entry in the exception list */
    static class ListEntry {
        int type;
        String name;
        Throwable t;
        public ListEntry(int type, String name, Throwable t) {
            this.type = type;
            this.name = name;
            this.t = t;
        }
        
        public String toString() {
            return MessageFormat.format(
            getString("FMT_ErrorText"),
            new Object[] {
                new Integer(type),
                name,
                t
            }
            );
        }
    }
    
    /**
     * Node to be used as a root node in ProjectErrorsPanel.
     */
    private class AllErrorsNode extends AbstractNode {
        public AllErrorsNode() {
            super(new Children.Array());
            setIconBase("/org/netbeans/modules/projects/resources/AllErrorsNodeIcon"); // NOI18N
            // Whatever is most relevant to a user:
//            setDefaultAction (SystemAction.get (PropertiesAction.class));
            Node n = TopManager.getDefault().getPlaces().nodes().projectDesktop();
            String name = NbBundle.getMessage(ProjectErrorManager.class, "LBL_AllErrorsNode", n.getName());
            setName(name);
            setDisplayName(name);
            setShortDescription(
                ProjectErrorManager.getString("HINT_AllErrorsNode"));
        }
        
        public HelpCtx getHelpCtx() {
            return HelpCtx.DEFAULT_HELP;
            // When you have help, change to:
            // return new HelpCtx (AllErrorsNode.class);
        }
        
        public Node cloneNode() {
            return new AllErrorsNode();
        }
    } // end of AllErrorsNode
    
    /**
     * Node to be used in the ProjectErrorsPanel.
     */
    private static class ErrorNode extends AbstractNode {
        private ListEntry entry;
        public ErrorNode(ListEntry entry) {
            super (Children.LEAF);
            this.entry = entry;
            
            setValue(PROP_ERROR_DETAILS, entry);
            setIconBase("/org/netbeans/modules/projects/resources/ErrorNodeIcon"); // NOI18N
            // Whatever is most relevant to a user:
//            setDefaultAction (SystemAction.get (PropertiesAction.class));
            setName(ProjectErrorManager.getString("LBL_ErrorNode"));
            setDisplayName(entry.name);
            String description = ProjectErrorManager.getString("HINT_ErrorNode");
            switch (entry.type) {
                case TYPE_FILE: 
                    description = ProjectErrorManager.getString("MSG_ErrorTypeFile");
                    break;
                case TYPE_SCO: 
                    description = ProjectErrorManager.getString("MSG_ErrorTypeSCO");
                    break;
                case TYPE_ITEM: 
                    description = ProjectErrorManager.getString("MSG_ErrorTypeItem");
                    break;
                case TYPE_WM: 
                    description = ProjectErrorManager.getString("MSG_ErrorTypeWindowManager");
                    break;
            }
            setShortDescription(description);
        }
        
        public HelpCtx getHelpCtx() {
            return HelpCtx.DEFAULT_HELP;
            // When you have help, change to:
            // return new HelpCtx (ErrorNode.class);
        }
        
        public Node cloneNode() {
            return new ErrorNode(entry);
        }
        
        public boolean canCopy() {
            return true;
        }
        
        public boolean canCut() {
            return false;
        }
        
        public Transferable clipboardCopy() {
            try {
                // Add to, do not replace, the default node copy flavor:
                ExTransferable et = ExTransferable.create (super.clipboardCopy ());
                et.put (new ExTransferable.Single (DataFlavor.stringFlavor) {
                    protected Object getData () {
                        return ErrorNode.this.getDisplayName ();
                    }
                });
                return et;
            } catch (IOException ex) {
                TopManager.getDefault().getErrorManager().notify(ex);
            }
            return ExTransferable.EMPTY;
        }
    } // end of ErrorNode
}
