/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects.content;

import java.io.*;
import java.util.*;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.options.*;
import org.openide.util.SharedClassObject;
import org.openide.util.io.NbMarshalledObject;

import org.xml.sax.*;

import org.netbeans.modules.projects.*;



/**
 *
 * @author  mryzl
 */

public class XMLProjectContent implements ProjectContent {

    public static final String VERSION = "1.1";
    
    /** Debug flag. */
    private static final boolean DEBUG = false;
    
    /** Extension for project data files. */
    public static final String EXT_PROJECT_DATA = "xml"; // NOI18N
    /** Extension for project data files. */
    public static final String EXT_SER = "ser"; // NOI18N
    /** ControlPanel file name. */
    public static final String NAME_CONTROL_PANEL = "cpanel"; // NOI18N
    /** Loaders file name. */
    public static final String NAME_LOADERS = "loaders"; // NOI18N
    /** Repository file name. */
    public static final String NAME_REPOSITORY = "repository"; // NOI18N
    /** Services file name. */
    public static final String NAME_SERVICES = "services"; // NOI18N
    /** Services file name. */
    public static final String NAME_WORKSPACE = "workspace"; // NOI18N

    /** SettingsSet element name. */
    public static final String SETTINGS_SET_ELEMENT = "SettingsSet"; // NOI18N
    
    private XMLSettingsSet cpanel, loaders;
    private XMLDiffSet repository, services;
    private static ProjectErrorManager pem;
    private DataFolder dfolder;
    private ProjectContent superProject;

    /** Creates new XMLProjectContent.
    * @param folder - folder where project files are stored
    * @param superProject - superProject or null for main project
    */
    public XMLProjectContent(DataFolder dfolder, ProjectContent superProject, ProjectErrorManager pem) {
        this.dfolder = dfolder;
        this.superProject = superProject;
        this.pem = pem;
    }

    /**
    * @return super project content or null if there is not one
    */
    public ProjectContent getSuperProject() {
        return superProject;
    }
    
    /** Get control panel.
    *
    * @return control panel
    */
    public SettingsSet getControlPanel() {
        if (cpanel == null) {
            cpanel = loadSettingsSet(new XMLSettingsSet(), NAME_CONTROL_PANEL);
        }
        return cpanel;
    }

    /** Get loaders pool.
    *
    * @return loaders
    */
    public SettingsSet getLoaderPool() {
        if (loaders == null) {
            loaders = loadSettingsSet(new XMLSettingsSet(), NAME_LOADERS);
        }
        return loaders;
    }

    /** Get repository.
    *
    * @return repository
    */
    public DiffSet getRepository() {
        if (repository == null) {
            repository = loadDiffSet(new XMLDiffSet(), NAME_REPOSITORY);
        }
        return repository;
    }

    /** Get services.
    *
    * @return services
    */
    public DiffSet getServices() {
        if (services == null) {
            services = loadDiffSet(new XMLDiffSet(), NAME_SERVICES);
        }
        return services;
    }

    /**
    * @return window manager or null if there is no one.
    */
    public void storeWindowManager() {
        
        NbMarshalledObject windowManager = null;
        try {
            windowManager = new NbMarshalledObject(TopManager.getDefault().getWindowManager());
        }  catch (ThreadDeath td) {
            throw td;
        } catch (Throwable th) {
            pem.annotate(pem.TYPE_WM, null, th);
            return;
        }
        
        // [REMIND - potentially risky - should be done in runAtomicAction
        FileObject ff = dfolder.getPrimaryFile();
        FileLock lock = null;
        ObjectOutputStream oos = null;
        try {
            FileObject fo = ff.getFileObject(NAME_WORKSPACE, EXT_SER);
            if (fo == null) {
                fo = ff.createData(NAME_WORKSPACE, EXT_SER);
            }
            lock = fo.lock();
            oos = new ObjectOutputStream(fo.getOutputStream(lock));
            oos.writeObject(windowManager);
            oos.close();
        } catch (Exception ex) {
            pem.annotate(pem.TYPE_FILE, ff.getPackageName('/') + '/' + NAME_WORKSPACE + '.' + EXT_SER, ex, true);
        } finally {
            if (lock != null) lock.releaseLock();
            try {
                if (oos != null) oos.close();
            } catch (Exception ex) {
                // ignore
            }
        }
    }

    /**
    */
    public NbMarshalledObject getWindowManager() {
        NbMarshalledObject nmo = null;
        FileObject ff = dfolder.getPrimaryFile();
        FileObject fo = ff.getFileObject(NAME_WORKSPACE, EXT_SER);
        if (fo != null) {
            ObjectInputStream ois = null;
            try {
                ois = new org.openide.util.io.NbObjectInputStream(fo.getInputStream());
                nmo = (NbMarshalledObject) ois.readObject();
            } catch (Exception ex) {
                pem.annotate(pem.TYPE_FILE, ff.getPackageName('/') + '/' + NAME_WORKSPACE + '.' + EXT_SER, ex, true);
            } finally {
                try {
                    if (ois != null) ois.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }
        return nmo;
    }

    /** Load settings set from the project content's folder, file <name>.EXT_PROJECT_DATA.
     * @param set - a set (not null)
     * @param name
     * @return the set
     */
    private XMLSettingsSet loadSettingsSet(XMLSettingsSet set, String name) {
        FileObject fo = getFileObject(name, EXT_PROJECT_DATA, false);
        if (fo != null) {
            BufferedReader reader = null;
            try {
                reader = new BufferedReader(new InputStreamReader(fo.getInputStream()));
                Map map = XMLSupport.loadObjects(reader, new HashMap(), SETTINGS_SET_ELEMENT);
                set.setRawData(map);
            } catch (Exception ex) {
                pem.annotate(pem.TYPE_FILE, fo.getPackageNameExt('/', '.'), ex, true);
            } finally {
                try {
                    if (reader != null) reader.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }
        return set;
    }

    /** Save settings set to the project content's folder, file <name>.EXT_PROJECT_DATA.
     * @param set - a set (not null)
     * @param name
     */
    private void saveSettingsSet(XMLSettingsSet set, String name) {
        FileObject fo = getFileObject(name, EXT_PROJECT_DATA, true);
        if (fo != null) {
            BufferedWriter writer = null;
            FileLock lock = null;
            try {
                lock = fo.lock();
                writer = new BufferedWriter(new OutputStreamWriter(fo.getOutputStream(lock)));
                Map map = set.getRawData();
                XMLSupport.saveObjects(writer, map, SETTINGS_SET_ELEMENT);
            } catch (Exception ex) {
                pem.annotate(pem.TYPE_FILE, fo.getPackageNameExt('/', '.'), ex, true);
            } finally {
                if (lock != null) lock.releaseLock();
                try {
                    if (writer != null) writer.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }
    }
    
    /** Load diff set from the project content's folder, file <name>.EXT_PROJECT_DATA.
     * @param set - a set (not null)
     * @param name
     * @return the set
     */
    private XMLDiffSet loadDiffSet(XMLDiffSet set, String name) {
        FileObject fo = getFileObject(name, EXT_PROJECT_DATA, false);
        if (fo != null) {
            BufferedReader reader = null;
            try {
                reader = new BufferedReader(new InputStreamReader(fo.getInputStream()));
                XMLSupport.loadDiffSet(reader, set);
            } catch (Exception ex) {
                pem.annotate(pem.TYPE_FILE, fo.getPackageNameExt('/', '.'), ex, true);
            } finally {
                try {
                    if (reader != null) reader.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }
        return set;
    }

    /** Save diff set to the project content's folder, file <name>.EXT_PROJECT_DATA.
     * @param set - a set (not null)
     * @param name
     */
    private void saveDiffSet(XMLDiffSet set, String name) {
        FileObject fo = getFileObject(name, EXT_PROJECT_DATA, true);
        if (fo != null) {
            BufferedWriter writer = null;
            FileLock lock = null;
            try {
                lock = fo.lock();
                writer = new BufferedWriter(new OutputStreamWriter(fo.getOutputStream(lock)));
                XMLSupport.saveDiffSet(writer, set);
            } catch (Exception ex) {
                pem.annotate(pem.TYPE_FILE, fo.getPackageNameExt('/', '.'), ex, true);
            } finally {
                if (lock != null) lock.releaseLock();
                try {
                    if (writer != null) writer.close();
                } catch (Exception ex) {
                    // ignore
                }
            }
        }
    }
    
    
    /** Get file object if Project Content folder and mark it as DumbXML.
     * @param name - name of the file
     * @param ext - extension of the file
     * @param create - if create is true, try to create a new file when it doesn' exist
     * @return file object or null
     */
    private FileObject getFileObject(final String name, final String ext, boolean create) {
        final FileObject ff = dfolder.getPrimaryFile();
        final FileObject[] fo = new FileObject[] { ff.getFileObject(name, ext) };
        if ((fo[0] == null) && create) {
            try {
                ff.getFileSystem().runAtomicAction(new FileSystem.AtomicAction() {
                    public void run() throws IOException {
                        try {
                            fo[0] = ff.createData(name, ext);
                        } catch (Exception ex) {
                            pem.annotate(pem.TYPE_FILE, ff.getPackageName('/') + '/' + name + '.' + ext, ex, true);
                            return;
                        }
                        
                        // f[0] is not null here because createData returns either a file object
                        // or it throws an exception
                        
                        // mark it as DumbXML - it is better but not important
                        try {
                            DumbXMLDataLoader.getInstance().markFile(fo[0]);
                            DataObject dobj = DataObject.find(fo[0]);
                            if (dobj.getClass() != DumbXMLDataLoader.class) {
                                dobj.setValid(false);
                            }
                            dobj = DataObject.find(fo[0]);
                        } catch (Exception ex) {
                            // ignore
                        }
                    }
                });
            } catch (Exception ex) {
                // probably problem with filesystem
                pem.annotate(pem.TYPE_FILE, ff.getPackageName('/') + '/' + name + '.' + ext, ex, true);
            }
        }
        return fo[0];
    }
    
    /** Store the project.
     */
    public void store() {
        // save all
    }

    /** Store control panel.
     */
    public void storeControlPanel() {
        if (cpanel != null) saveSettingsSet(cpanel, NAME_CONTROL_PANEL);
    }
    
    /** Store loaders.
     */
    public void storeLoaderPool() {
        if (loaders != null) saveSettingsSet(loaders, NAME_LOADERS);
    }
    
    /** Store repository.
     */
    public void storeRepository() {
        if (repository != null) saveDiffSet(repository, NAME_REPOSITORY);
    }
    
    /** Store services.
     */
    public void storeServices() {
        if (services != null) saveDiffSet(services, NAME_SERVICES);
    }
    
    /**
     */
    public static class XMLDiffSet implements DiffSet {
        
        private List items;
        private Map addMap;
        private Set removeSet;
        private boolean clear = true;
        private String version;
        
        /** Add the item. If it was previously removed, cancel its record in "removed" set.
        * @param name
        * @param value
        */
        public synchronized void add(String name, Object value) {
            if (removeSet != null) {
                removeSet.remove(name);
            }
            if (items == null) {
                items = new LinkedList();
            }
            if (!items.contains(name)) {
                items.add(name);
            }
            if (addMap == null) {
                addMap = new HashMap(16);
            }
            addMap.put(name, value);
        }
        
        public synchronized void addRaw(String name, String rawString) {
            try {
                if (rawString != null) {
                    Object obj = XMLSupport.decodeValue(rawString);
                    add(name, obj);
                }
            } catch (ThreadDeath td) {
                throw td;
            } catch (ClassNotFoundException ex) {
                // #9944 fix
                if (!ex.getMessage().equals("org.netbeans.core.SystemFileSystem")) { // NOI18N
                    pem.annotate(pem.TYPE_ITEM, name, ex);
                }
            } catch (Throwable th) {
                pem.annotate(pem.TYPE_ITEM, name, th);
            }
        }

        /** Remove the item. If it was previously added, cancel its record in "added" set.
        * @param name
        */
        public synchronized void remove(String name) {
            if (items != null) items.remove(name);
            if (addMap != null) addMap.remove(name);
            if (removeSet == null) {
                removeSet = new HashSet(16);
            }
            removeSet.add(name);
        }

        /**
        * @return Map of added items
        */
        public synchronized List addedItems() {
            if (items != null) {
                return new ArrayList(items);
            } else {
                return new ArrayList(0);
            }
        }

        /** Added item by name.
        * @param name - name of the item
        * @return value of the item.
        */
        public synchronized Object addedItem(String name) {
            if (addMap != null) return addMap.get(name);
            else return null;
        }
        
        public synchronized String addedRawItem(String name) {
            Object obj = addedItem(name);
            if (obj != null) {
                try {
                    return XMLSupport.encodeValue(obj);
                } catch (ThreadDeath td) {
                    throw td;
                } catch (Throwable th) {
                    pem.annotate(pem.TYPE_ITEM, name, th);
                }
            }
            return null;
        }

        /**
        * @return set of removed items
        */
        public synchronized Set removedItems() {
            if (removeSet != null) return new HashSet(removeSet);
            else return new HashSet(4);
        }

        /** Test whether the diff set should be applied on fresh settings or not.
         * @return true if it should be applied on fresh settings
         */
        public boolean isClear() {
            return clear;
        }

        /** Set clear flag.
        * @param clear 
        */
        public void setClear(boolean clear) {
            this.clear = clear;
// ???            if (clear) removeSet = null;
        }

        public String getVersion() {
            return version;
        }
        
        public void setVersion(String ver) {
            version = ver;
        }
        
        /** Clear DiffSet.
        */
        public synchronized void clear() {
            items = null;
            removeSet = null;
            addMap = null;
        }
    }
    
    
    /** Internal implementation of SettingsSet.
     */
    class XMLSettingsSet implements SettingsSet {

        /** Contains set of active SCOs.
         */
        private HashSet set;
        
        /** Cache. Contains
         * <classname_of_SCO, encoded_string>
         *
         * when the SCO is successfully read its value is set to null
         * when the SCO is being written, null value indicates that 
         *   the current value should be serialized
         */
        private Map rawData;

        /** Test whether the set contains given shared class object.
        * @param obj object
        * @return true if the set contains the object
        */
        public synchronized boolean contains(SharedClassObject obj) {
            if (set != null) {
                return set.contains(obj.getClass().getName());
            }
            return false;
        }

        /** Add shared object to the set.
        * @param obj - object to add.
        */
        public synchronized void add(SharedClassObject obj) {
            if (set == null) {
                set = new HashSet(16);
            }
            set.add(obj);
        }

        /** Remove shared object from the set.
        * @param obj - object to remove.
        */
        public synchronized void remove(SharedClassObject obj) {
            if (set != null) {
                set.remove(obj);
            }
        }

        /** Clear SettingsSet.
        */
        public synchronized void clear() {
            set = null;
        }

        /** Get all options of the set.
         */
        public synchronized java.util.Collection getObjects() {
            if (set != null) {
                return new HashSet(set);
            } else {
                return new HashSet(4);
            }
        }
        
        public synchronized void setRawData(Map map) {
            rawData = new HashMap(map);
            set = new HashSet(rawData.size());
            Iterator it = rawData.entrySet().iterator();
            ClassLoader cl = TopManager.getDefault().systemClassLoader();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry)it.next();
                String name = entry.getKey().toString();
                try {
                    Class clazz = cl.loadClass(name);
                    set.add(SharedClassObject.findObject(clazz, true));
                } catch (Exception ex) {
                    it.remove(); // remove data for invalid SCO
                    
                    // icebrowser module was removed from product - it is OK that its settings can't be deserialized
                    if ("org.netbeans.modules.icebrowser.ICEBrowserSettings".equals (name)) // NOI18N
                        TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, ex);
                    else
                        pem.annotate(pem.TYPE_SCO, name, ex);
                }
            }
        }
        
        public synchronized Map getRawData() {
            if (rawData != null) {
                return new HashMap(rawData);
            } else {
                return new HashMap(4);
            }
        }
        
        /** 
        * @param processor - processor that controls writing of an option
        */
        public synchronized void read(OptionProcessor processor) {
            if ((set != null) && (rawData != null)) {
                Iterator it = set.iterator();
                while (it.hasNext()) {
                    SharedClassObject sco = (SharedClassObject) it.next();
                    if (processor.canProcess(sco)) {
                        try {
                            String rawString = (String) rawData.get(sco.getClass().getName());
                            if (rawString != null) {
                                XMLSupport.decodeValue(rawString);
                            }
                        } catch (ThreadDeath td) {
                            throw td;
                        } catch (Throwable th) {
                            pem.annotate(pem.TYPE_SCO, sco.getClass().getName(), th);
                        }
                    }
                }
            }
        }

        /*
        */
        public synchronized void write(OptionProcessor processor) {
            if (set != null) {
                rawData = new HashMap(set.size() * 2);
                Iterator it = set.iterator();
                while (it.hasNext()) {
                    SharedClassObject sco = (SharedClassObject) it.next();
                    if (processor.canProcess(sco)) {
                        try {
                            rawData.put(sco.getClass().getName(), XMLSupport.encodeValue(sco));
                        } catch (ThreadDeath td) {
                            throw td;
                        } catch (Throwable th) {
                            pem.annotate(pem.TYPE_SCO, sco.getClass().getName(), th);
                        }
                    }
                }
            }
        }
    }
    
    
    // ===================================== Tests ============================================
    
    private static void assert(boolean cond, String msg) throws AssertException {
        if (!cond) throw new AssertException(msg);
    }
    
    /** Test class. 
     * It checks whether it is possible to store and load settings set.
     */
    private static boolean test1() {
        try {
            // find temporary folder
            FileObject ff = TopManager.getDefault().getRepository().findResource("/tmp");
            assert((ff != null), "No /tmp folder.");
            DataFolder tmp = DataFolder.findFolder(ff);
            
            // find cpanel and delete it
            FileObject fo = ff.getFileObject("cpanel", "xml");
            if (fo != null) {
                FileLock lock = null;
                try {
                    lock =  fo.lock();
                    fo.delete(lock);
                } finally {
                    if (lock != null) lock.releaseLock();
                }
            }
            
            // create a new XMLProjectContent with one item
            XMLProjectContent pc = new XMLProjectContent(tmp, null, new ProjectErrorManager());
            SettingsSet ss = pc.getControlPanel();
            org.netbeans.modules.projects.test.TestSharedClassObject tsco = (org.netbeans.modules.projects.test.TestSharedClassObject) SharedClassObject.findObject(org.netbeans.modules.projects.test.TestSharedClassObject.class, true);
            assert((tsco != null), "No TSCO.");
            
            tsco.setMessage("t1");
            ss.add(tsco);
            Collection col = ss.getObjects();
            assert((col.size() == 1), "col.size != 1");
            
            ss.write(OptionProcessor.ALL);
            pc.storeControlPanel();
            
            FileObject fo2 = ff.getFileObject("cpanel", "xml");
            assert((fo2 != null), "No cpanel.xml");
            
            pc.cpanel = null;
            tsco.setMessage("t2");
            SettingsSet ss2 = pc.getControlPanel();
            ss2.read(OptionProcessor.ALL);
            Collection col2 = ss2.getObjects();
            assert(((col2.size() == 1) && (col2.toArray()[0] instanceof org.netbeans.modules.projects.test.TestSharedClassObject)),
                "col2.size != 1 or col2[0] doesn't contain TSCO"
            );
            assert(("t1".equals(tsco.getMessage())), "TSCO has not been deserialized properly.");
            
            
            // test writing an object with wrong serialization
            org.netbeans.modules.projects.test.ErrorSharedClassObject esco = (org.netbeans.modules.projects.test.ErrorSharedClassObject) SharedClassObject.findObject(org.netbeans.modules.projects.test.ErrorSharedClassObject.class, true);
            assert((esco != null), "No ESCO.");
            ss2.add(esco);
            esco.setException(new Error("test error"));
            esco.setWrite(true);
            long foLen = fo.getSize();
            ss2.write(OptionProcessor.ALL);
            // check PEM for an exception
            pc.storeControlPanel();
            
            long foLen2 = fo.getSize();
            assert((foLen == foLen2), "Size of cpanel.xml has changed.");

            // test reading an object with wrong deserialization
            esco.setWrite(false);
            tsco.setMessage("t1");
            ss2.write(OptionProcessor.ALL);
            pc.storeControlPanel();
            long foLen3 = fo.getSize();
            assert((foLen2 != foLen3), "Size of cpanel.xml has not changed.");
            
            pc.cpanel = null;
            tsco.setMessage("t2");
            SettingsSet ss3 = pc.getControlPanel();
            ss3.read(OptionProcessor.ALL);
            Collection col3 = ss3.getObjects();
            assert((col3.size() == 2), "Col3.size != 2");
            assert("t1".equals(tsco.getMessage()), "TSCO has not been deserialized properly."); 
            return true;
        } catch (Exception ex) {
            ex.printStackTrace();
            return false;
        }
    }

    
    /** Test error handling during serialization/deserialization. 
     */
    private static boolean test2() {
        try {
            return true;
        } catch (Exception ex) {
            ex.printStackTrace();
            return false;
        }
    }
    
    public static void main(String[] args) throws Exception {
        System.err.println(">> Test1 status = " + test1());
    }

    
    /** AssertException.
     */
    private static class AssertException extends RuntimeException {
        public AssertException(String msg) {
            super(msg);
        }
    }
    
    
}
