/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.projects.content;

import java.io.*;
import java.text.MessageFormat;
import java.util.*;

import org.openide.TopManager;
import org.openide.ErrorManager;
import org.openide.loaders.XMLDataObject;
import org.openide.util.SharedClassObject;
import org.openide.util.Utilities;
import org.openide.util.io.*;

import org.xml.sax.*;

/**
 *
 * @author  mryzl
 * @version 
 */
public class XMLSupport extends Object {

    public static final String EMPTY_STRING = "";  // NOI18N

    public static final String FMT_XML_HEADER = "<?xml version=\"1.0\"?>"; // NOI18N
    public static final String FMT_OBJECTS_START = "<{0} version=\""+XMLProjectContent.VERSION+"\">"; // NOI18N
    public static final String FMT_OBJECTS_END = "</{0}>"; // NOI18N
    public static final String ELEMENT_CONTROL_PANEL = "control-panel"; // NOI18N
    public static final String ELEMENT_MODULES= "modules"; // NOI18N
    public static final String ELEMENT_LOADERS = "loaders"; // NOI18N

    /** Creates new XMLSupport */
    private XMLSupport() {
    }

    static String encodeToProperXML(String text) {
        if (text.indexOf ('&') != -1) text = Utilities.replaceString (text, "&", "&amp;"); // must be the first to prevent changes in the &XX; codes // NOI18N

        if (text.indexOf ('<') != -1) text = Utilities.replaceString (text, "<", "&lt;"); // NOI18N
        if (text.indexOf ('>') != -1) text = Utilities.replaceString (text, ">", "&gt;"); // NOI18N
        if (text.indexOf ('\'') != -1) text = Utilities.replaceString (text, "\'", "&apos;"); // NOI18N
        if (text.indexOf ('\"') != -1) text = Utilities.replaceString (text, "\"", "&quot;"); // NOI18N
        if (text.indexOf ('\n') != -1) text = Utilities.replaceString (text, "\n", "&#xa;"); // NOI18N
        if (text.indexOf ('\t') != -1) text = Utilities.replaceString (text, "\t", "&#x9;"); // NOI18N
        return text;
    }

    /** Decode string form to the byte array.
     * @param value - list of decimal numbers separated by commas.
     * @return Object or null
     * @exception Throwable
     */
    static Object decodeValue(String value) throws Throwable {
        if ((value == null) || (value.length () == 0)) return null;

        char[] bisChars = value.toCharArray ();
        byte[] bytes = new byte[bisChars.length];
        char[] singleNum = new char[4];
        int snCount = 0;
        int count = 0;
        for (int i = 0; i < bisChars.length; i++) {
            if (',' == bisChars[i]) {
                bytes[count++] = parseByte(singleNum, snCount);
                snCount = 0;
            } else {
                singleNum[snCount++] = bisChars[i];
            }
        }
        // add the last byte
        bytes[count++] = parseByte(singleNum, snCount);
        return deserializeValue(bytes);
    }
    
    /** Decodes a value of from the specified String containing textual representation of serialized stream.
     * @param bytes - 
     * @return deserialized object (never null)
     * @exception Throwable, NullPointerException when bytes is null.
     */
    private static Object deserializeValue (byte[] bytes) throws Throwable {
        if (bytes == null) throw new NullPointerException();
        ObjectInputStream ois = null;
        try {
            ois = new NbObjectInputStream (new ByteArrayInputStream (bytes, 0, bytes.length));
            Object ret = ois.readObject ();
            return ret;
        } finally {
            if (ois != null) ois.close();
        }
    }
    
    /** e.g. makes from String "-123" byte -123 
     * @exception NumberFormatException
     */
    static byte parseByte(char[] chars, int len) throws NumberFormatException {
        int idx = 0;
        boolean negative = false;
        if (chars[0] == '-') {
            negative = true;
            idx++;
        }
        
        int ret = 0;
        int tmp;
        
        for ( ; idx < len; idx++) {
            tmp = chars[idx] - '0';
            if ((tmp < 0) || (tmp > 9)) {
                throw new NumberFormatException();
            }
            ret *= 10;
            ret += tmp;
        }
        
        if (negative) {
            ret = -ret;
        }
        
        if ((ret < Byte.MIN_VALUE) || (ret > Byte.MAX_VALUE)) {
            throw new NumberFormatException();
        }
        
        return (byte) ret;
    }

    /** Encodes specified value to a String containing textual representation of serialized stream.
     * @return String containing textual representation of the serialized object
     */
    private static byte[] serializeValue (Object value) throws Throwable {
        
        if (value == null) return null; // NOI18N
        
        ByteArrayOutputStream bos = new ByteArrayOutputStream ();
        ObjectOutputStream oos = null;
        try {
            oos = new NbObjectOutputStream (bos);
            oos.writeObject (value);
        } finally {
            if (oos != null) oos.close();
        }
        
        return bos.toByteArray();
        
    }
    
    
    /** Encode a byte array to comma separated decimal numbers string.
     * @param bosBytes - the array
     * @return the string or null if obj is null
     */
    static String encodeValue(Object obj) throws Throwable {
        if (obj == null) return null;
        byte[] bosBytes = serializeValue(obj);
        StringBuffer sb = new StringBuffer (bosBytes.length * 4 + 2);
        for (int i = 0; i < bosBytes.length - 1; i++) {
            sb.append(bosBytes[i]).append(',');
        }
        sb.append(bosBytes[bosBytes.length - 1]);
        
        return sb.toString ();
    }

    /** Encode strings to one. Comma will be used as a delimiter.
    * @param list of strings
    */
    public static String encodeStrings(List list) {
        Iterator it = list.iterator();
        StringBuffer sb = new StringBuffer();

        boolean comma = false;
        for(int i = 0; it.hasNext();) {
            if (comma) {
                sb.append(',');
            } else {
                comma = true;
            }
            sb.append(encodeToProperXML((String)it.next()));
        }
        return sb.toString();
    }

    /** Decode strings from one string with comma as a delimiter.
    * @return list of strings
    */
    public static List decodeStrings(String value) {
        List list = new LinkedList();
        StringTokenizer st = new StringTokenizer(value, ","); // NOI18N
        while (st.hasMoreTokens()) {
            list.add(st.nextToken());
        }
        return list;
    }


    /** Save given array of objects to the XML file.
    * @param writer 
    * @param map with <classname_of_SCO,  byte[]>
    * @param mainElement main element of the XML
    */
    public static void saveObjects(Writer writer, Map map, String mainElement) {
        Object[] formatObjs = new Object[] { mainElement };
        PrintWriter pw = new PrintWriter(writer);
        pw.println(FMT_XML_HEADER);
        pw.println(MessageFormat.format(FMT_OBJECTS_START, formatObjs));
        Iterator it = map.keySet().iterator();
        while (it.hasNext()) {
            String name = (String) it.next();
            String rawString = (String) map.get(name);
            if (rawString != null) {
                pw.print("  <object name=\""); // NOI18N
                pw.print(encodeToProperXML(name));
                pw.print("\" value=\""); // NOI18N
                pw.print(rawString);
                pw.println("\"/>"); // NOI18N
            }
        }
        pw.println(MessageFormat.format(FMT_OBJECTS_END, formatObjs));
        pw.flush();
    }

    /** Load objects from XML file.
    * @param reader
    * @param map an empty map. It will be filled by processed values and returned back.
    * @param mainElement main element of the XML
    * @return a map - <classname_of_SCO, String>
    */
    public static Map loadObjects(Reader reader, Map map, String mainElement) throws IOException, SAXException {
        Parser parser = XMLDataObject.createParser();
        HandlerObjects handler = new HandlerObjects(map, mainElement);
        parser.setDocumentHandler(handler);
        parser.parse(new InputSource(reader));
        return map;
    }

    /**
    */
    public static void loadDiffSet(Reader reader, XMLProjectContent.XMLDiffSet diffset) throws IOException, SAXException {
        Parser parser = XMLDataObject.createParser();
        HandlerDiffSet handler = new HandlerDiffSet(diffset);
        parser.setDocumentHandler(handler);
        parser.parse(new InputSource(reader));
    }

    /**
    */
    public static void saveDiffSet(Writer writer, XMLProjectContent.XMLDiffSet diffset) throws IOException {
        // write header
        PrintWriter pw = new PrintWriter(writer);
        pw.println(FMT_XML_HEADER);
        // write diff set start element
        pw.println("<" + HandlerDiffSet.ELEMENT_DIFF_SET + " clear=\"" + diffset.isClear() + "\" version=\""+XMLProjectContent.VERSION+"\">"); // NOI18N
        // write add
        Iterator it = diffset.addedItems().iterator();
        while (it.hasNext()) {
            String item = (String) it.next();
            pw.print("  <add name=\""); // NOI18N
            pw.print(encodeToProperXML(item));
            pw.print("\" value=\""); // NOI18N
            pw.print(diffset.addedRawItem(item));
            pw.println("\"/>"); // NOI18N
        }
        // write remove
        it = diffset.removedItems().iterator();
        while (it.hasNext()) {
            String item = (String) it.next();
            pw.print("  <remove name=\""); // NOI18N
            pw.print(encodeToProperXML(item));
            pw.println("\"/>"); // NOI18N
        }
        // write order
        /** Not used anymore.
        pw.print("  <order value=\""); // NOI18N
        pw.print(encodeStrings(diffset.getOrder()));
        pw.println("\"/>"); // NOI18N
        */
        
        // write diff set end element
        pw.println("</" + HandlerDiffSet.ELEMENT_DIFF_SET +">"); // NOI18N

    }

    /**
     * Compare parameter version with projectVersion.
     * It returns true if
     * version number is greater than the projectVersion. <p>
     *
     * Version numbers are compared by sequentially comparing corresponding
     * components of the strings.
     * Each component is converted as a decimal integer and the values
     * compared.
     *
     * @param version the version string to be compared
     * @param projectVersion the version string to be compared
     * @return true if this version number is greater
     * 		than the projectVersion param<br>  
     *		
     * @exception NumberFormatException if the version
     *		is not of the correct dotted form.
     */
    public static boolean isNew(String version, String projectVersion)
	throws NumberFormatException
    {
    	if (version == null || version.length() < 1) {
	    throw new NumberFormatException("Empty version string"); // NOI18N
	}

    	// Until it matches scan and compare numbers
    	StringTokenizer vtok = new StringTokenizer(version, ".", true);
    	StringTokenizer tok = new StringTokenizer(projectVersion, ".", true);
	while (vtok.hasMoreTokens() || tok.hasMoreTokens()) {
	    int vver;
	    int ver;
	    if (vtok.hasMoreTokens()) {
		vver = Integer.parseInt(vtok.nextToken());
	    } else
		vver = 0;

	    if (tok.hasMoreTokens()) {
		ver = Integer.parseInt(tok.nextToken());
	    } else
		ver = 0;

    	    if (vver < ver)
    	    	throw new IllegalArgumentException("Specified version is older than the projecVersion"); // NOI18N
    	    if (vver > ver)
    	    	return true; // it is NEW!

    	    // Check for and absorb seperators
    	    if (vtok.hasMoreTokens())
    	    	vtok.nextToken();
    	    if (tok.hasMoreTokens())
    	    	tok.nextToken();
    	    // Compare next component
    	}
    	// All components numerically equal
	return false;
    }
    
    
}