/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.properties;


import java.awt.Color;
import java.awt.Event;
import java.awt.event.KeyEvent;
import java.awt.SystemColor;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.KeyStroke;

/**
 * Factory for creating <code>TableViewSettings</code> implementation. If editor module
 * is available the editor module dependent implementation is created, if case of not
 * default (independent) implemenation is created.
 * Note: Don't instantaniate, use <code>getTableViewSettings</code> static method.
 * @author  Peter Zavadsky
 */
public class TableViewSettingsFactory extends Object {

    /** Default key color. */
    public static final Color KEY_DEFAULT_COLOR = Color.blue;
    /** Default key background. */    
    public static final Color KEY_DEFAULT_BACKGROUND = Color.white;
    /** Default value color. */
    public static final Color VALUE_DEFAULT_COLOR = Color.magenta;
    /** Default value background. */
    public static final Color VALUE_DEFAULT_BACKGROUND = Color.white;
    /** Default highlight color. */
    public static final Color HIGHLIGHT_DEFAULT_COLOR = Color.black;
    /** Default highlight background. */
    public static final Color HIGHLIGHT_DEFAULT_BACKGROUND = Color.yellow;
    /** Default shadow color. */
    public static final Color SHADOW_DEFAULT_COLOR = new Color(SystemColor.controlHighlight.getRGB());
    /** Default keystrokes for find next action. */
    public static final KeyStroke[] FIND_NEXT_DEFAULT_KEYSTROKES = new KeyStroke[] {KeyStroke.getKeyStroke(KeyEvent.VK_F3, 0)};
    /** Default keystrokes for find previous action. */
    public static final KeyStroke[] FIND_PREVIOUS_DEFAULT_KEYSTROKES = new KeyStroke[] {KeyStroke.getKeyStroke(KeyEvent.VK_F3, Event.SHIFT_MASK)};
    /** Default keystrokes for toggle highliht action. */
    public static final KeyStroke[] TOGGLE_HIGHLIGHT_DEFAULT_KEYSTROKES = new KeyStroke[] {KeyStroke.getKeyStroke(KeyEvent.VK_H, Event.SHIFT_MASK | Event.ALT_MASK)};

    /** Cache. */
    private static TableViewSettings settings;
    
    
    /** Private constructor for preventing to instantaniate by agressive programmer. */
    private TableViewSettingsFactory() {
    }
    
    
    /** @return <code>TableViewSettings</code> implementation. */
    public static TableViewSettings getTableViewSettings() {
        if(settings == null) {
            synchronized(TableViewSettings.class) {
                if(settings == null) {
                        // Editor module is not present or some problems happened.
                        // Init defaults.
                        settings = new TableViewSettings() {
                            public Color getKeyColor() {return KEY_DEFAULT_COLOR;}
                            public Color getKeyBackground() {return KEY_DEFAULT_BACKGROUND;}
                            public Color getValueColor() {return VALUE_DEFAULT_COLOR;}
                            public Color getValueBackground() {return VALUE_DEFAULT_BACKGROUND;}
                            public Color getHighlightColor() {return HIGHLIGHT_DEFAULT_COLOR;}
                            public Color getHighlightBackground() {return HIGHLIGHT_DEFAULT_BACKGROUND;}
                            public Color getShadowColor() {return SHADOW_DEFAULT_COLOR;}
                            
                            public KeyStroke[] getKeyStrokesFindNext() {return FIND_NEXT_DEFAULT_KEYSTROKES;}
                            public KeyStroke[] getKeyStrokesFindPrevious() {return FIND_PREVIOUS_DEFAULT_KEYSTROKES;}
                            public KeyStroke[] getKeyStrokesToggleHighlight() {return TOGGLE_HIGHLIGHT_DEFAULT_KEYSTROKES;}
                            
                            public void settingsUpdated() {}
                            public void addPropertyChangeListener(PropertyChangeListener listener) {}
                            public void removePropertyChangeListener(PropertyChangeListener listener) {}
                        };
                }
            }
        }
        
        return settings;        
    }

    // For access from syntax module:
    public static synchronized void setTableViewSettings (TableViewSettings tvs) throws SecurityException {
        if (settings != null) throw new SecurityException ("Too late!"); // NOI18N
        if (tvs == null) throw new NullPointerException ();
        settings = tvs;
    }
    
    /** Interface used for gaining colors and keystrokes for table view. */
    public interface TableViewSettings {
        public Color getKeyColor();
        public Color getKeyBackground();
        public Color getValueColor();
        public Color getValueBackground();
        public Color getHighlightColor();
        public Color getHighlightBackground();
        public Color getShadowColor();

        public KeyStroke[] getKeyStrokesFindNext();
        public KeyStroke[] getKeyStrokesFindPrevious();
        public KeyStroke[] getKeyStrokesToggleHighlight();
        
        public void settingsUpdated();
        public void addPropertyChangeListener(PropertyChangeListener listener);
        public void removePropertyChangeListener(PropertyChangeListener listener);
   } // End of interface PropertiesSettings.

}
