/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.properties.syntax;


import java.awt.Color;
import java.awt.Event;
import java.awt.event.KeyEvent;
import java.awt.SystemColor;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import javax.swing.KeyStroke;
import javax.swing.text.JTextComponent;

import org.netbeans.editor.BaseKit;
import org.netbeans.editor.Coloring;
import org.netbeans.editor.MultiKeyBinding;
import org.netbeans.editor.Settings;
import org.netbeans.editor.SettingsChangeEvent;
import org.netbeans.editor.SettingsChangeListener;
import org.netbeans.editor.SettingsNames;
import org.netbeans.modules.properties.TableViewSettingsFactory;
import org.netbeans.modules.properties.TableViewSettingsFactory.TableViewSettings;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.util.SharedClassObject;


/**
 * Class which stores copy of some settings of colors and keystrokes from editor module. 
 * The copy of values is used in <code>BundleEditPanel</code>.
 * The static method <code>getSettings</code> returns <code>TableViewSettings</code> of this implementation
 * if editor module is available or default implementation if not.
 * @author Peter Zavadsky
 * @see org.netbeans.modules.BundleEditPanel.TableViewSettings
 */
public class EditorSettingsCopy implements TableViewSettings {

    /** Value of key color retrieved from settings in editor module. */
    private Color keyColor;
    /** Value of key background retrieved from settings in editor module. */
    private Color keyBackground;
    /** Value of value color retrieved from settings in editor module. */
    private Color valueColor;
    /** Value of value background retrieved from settings in editor module. */
    private Color valueBackground;
    /** Value of highlight color retrieved from settings in editor module. */
    private Color highlightColor;
    /** Value of highlight bacground retrieved from settings in editor module. */
    private Color highlightBackground;
    /** Value of shadow color retrieved from settings in editor module. */
    private Color shadowColor;
    
    /** Key strokes for find next action rerieved from editor module. */
    private KeyStroke[] keyStrokesFindNext;
    /** Key strokes for find previous action retrieved from editor module. */
    private KeyStroke[] keyStrokesFindPrevious;
    /** Key strokes for toggle search highlight action retrieved from editor module. */
    private KeyStroke[] keyStrokesToggleHighlight;

    /** Support for property changes. */
    private final PropertyChangeSupport support = new PropertyChangeSupport(this);
        
    /** Listener for changes on settings. Updates our settings. */
    private SettingsChangeListener settingsListener = null;
    
    /** Singleton object. */
    private static TableViewSettings singleton;

    
    /** Private constructor. */
    private EditorSettingsCopy() {
    }

    
    /** Implements <code>EditorSetings</code> interface method. */
    public Color getKeyColor() {
        return keyColor;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */    
    public Color getKeyBackground() {
        return keyBackground;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */
    public Color getValueColor() {
        return valueColor;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */
    public Color getValueBackground() {
        return valueBackground;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */
    public Color getHighlightColor() {
        return highlightColor;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */ 
    public Color getHighlightBackground() {
        return highlightBackground;
    }
    
    /** Implements <code>EditorSetings</code> inaterface method. */ 
    public Color getShadowColor() {
        return shadowColor;
    }

    /** Implements <code>EditorSetings</code> interface method. */     
    public KeyStroke[] getKeyStrokesFindNext() {
        return keyStrokesFindNext;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */     
    public KeyStroke[] getKeyStrokesFindPrevious() {
        return keyStrokesFindPrevious;
    }
    
    /** Implements <code>EditorSetings</code> interface method. */
    public KeyStroke[] getKeyStrokesToggleHighlight() {
        return keyStrokesToggleHighlight;
    }

    /** Implements <code>EditorSetings</code> interface method. */    
    public void settingsUpdated() {
        support.firePropertyChange(new PropertyChangeEvent(this, null, null, null));
    }

    /** Implements <code>EditorSetings</code> interface method. */     
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        support.addPropertyChangeListener(listener);
    }

    /** Implements <code>EditorSetings</code> interface method. */    
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        support.removePropertyChangeListener(listener);
    }
    
    /** @return editor-dependent implementation of <code>TableViewSettings</code> interface or
     * null if some problems during init happened.
     * Attention. These method is save to invoke after initializing modules only (not from Module restored like method). 
     * Otherwise could lead to exceptional behavior.*/
    public static TableViewSettings getSettings() {
        if(singleton == null) {
            synchronized(EditorSettingsCopy.class) {
                if(singleton == null) {
                    // Editor module is available -> it's OK to call dependent code.
                    singleton = new EditorSettingsCopy();

                    // Set listening on changes of settings.
                    if(((EditorSettingsCopy)singleton).settingsListener == null)
                        Settings.addSettingsChangeListener(((EditorSettingsCopy)singleton).settingsListener = new SettingsChangeListener() {
                                public void settingsChange(SettingsChangeEvent evt) {
                                    // maybe could be refined
                                    ((EditorSettingsCopy)singleton).updateSettings();
                                }
                            }
                        );

                    // Init settings.                            
                    ((EditorSettingsCopy)singleton).updateSettings();

                    try {
                        TableViewSettingsFactory.setTableViewSettings (singleton);
                    } catch (SecurityException se) {
                        // Came too late.
                        TopManager.getDefault ().getErrorManager ().notify (ErrorManager.INFORMATIONAL, se);
                    }
                }
            }
        }
        
        return singleton;
    }

    /** Uninstalls settings. Called when module is uninstalled. */
    static void uninstallSettings() {
        if(singleton != null)
            ((EditorSettingsCopy)singleton).removeSettingsListenerFromEditor();
    }
    
    /** Removes settings listener from editor explicitly. */
    private void removeSettingsListenerFromEditor() {
        // Settings listener is WeakListener so it should be enough to assign null.
        settingsListener = null;
    }
    
    
    
    /** Updates settings from properties options. Only editor module dependent code. */
    private void updateSettings() {
        try {
            updateColors();
            updateKeyStrokes();
        } catch (InternalError ierr) {
            if(Boolean.getBoolean("netbeans.debug.exception")) // NOI18N
                TopManager.getDefault().notifyException(ierr);

            // Problems -> set singleton to null.
            synchronized(EditorSettingsCopy.class) {
                singleton = null;
            }
        }
    }

    /** Updates keystrokes. */
    private void updateColors() {
        PropertiesOptions propertiesOptions = (PropertiesOptions)SharedClassObject.findObject(PropertiesOptions.class, false);
        if(propertiesOptions == null)
            throw new InternalError("Properties options are not available."); // NOI18N
        
        // Update colors.
        Map map = propertiesOptions.getColoringMap();
        Coloring keyColoring = (Coloring)map.get(PropertiesTokenContext.contextPath.getFullTokenName(
            PropertiesTokenContext.KEY));
        keyColor = keyColoring.getForeColor();
        keyBackground = keyColoring.getBackColor();
        
        Coloring valueColoring = (Coloring)map.get(PropertiesTokenContext.contextPath.getFullTokenName(
            PropertiesTokenContext.VALUE));
        valueColor = valueColoring.getForeColor();
        valueBackground = valueColoring.getBackColor();
        
        Coloring highlightColoring = (Coloring)map.get(SettingsNames.HIGHLIGHT_SEARCH_COLORING);
        highlightColor = highlightColoring.getForeColor();
        highlightBackground = highlightColoring.getBackColor();

        shadowColor = propertiesOptions.getShadowTableCell();

        // If there is not the colors specified use default inherited colors.
        Color defaultForeground = ((Coloring)map.get("default")).getForeColor(); // NOI18N
        Color defaultBackground = ((Coloring)map.get("default")).getBackColor(); // NOI18N
        
        if(keyColor == null) keyColor = defaultForeground;
        if(keyBackground == null) keyBackground = defaultBackground;
        if(valueColor == null) valueColor = defaultForeground;
        if(valueBackground == null) valueBackground = defaultBackground;
        if(highlightColor == null) highlightColor = new Color(SystemColor.textHighlightText.getRGB());
        if(highlightBackground == null) highlightBackground = new Color(SystemColor.textHighlight.getRGB());
        if(shadowColor == null) shadowColor = new Color(SystemColor.controlHighlight.getRGB());
        
        // In case of some problems load default table colors.
        if(keyColor == null || keyBackground == null) {
            keyColor = TableViewSettingsFactory.KEY_DEFAULT_COLOR;
            keyBackground = TableViewSettingsFactory.KEY_DEFAULT_BACKGROUND;
        }
        if(valueColor == null || valueBackground == null) {
            valueColor = TableViewSettingsFactory.VALUE_DEFAULT_COLOR;
            valueBackground = TableViewSettingsFactory.VALUE_DEFAULT_BACKGROUND;
        }
        if(highlightColor == null || highlightBackground == null) {
            highlightColor = TableViewSettingsFactory.HIGHLIGHT_DEFAULT_COLOR;
            highlightBackground = TableViewSettingsFactory.HIGHLIGHT_DEFAULT_BACKGROUND;
        }
        if(shadowColor == null)
            shadowColor = TableViewSettingsFactory.SHADOW_DEFAULT_COLOR;
    }

    /** Updates keystrokes. Dependent code. */
    private void updateKeyStrokes() {
        // Update keyStrokes.
        // Retrieve key bindings for Propeties Kit and super kits.
        Settings.KitAndValue kv[] = Settings.getValueHierarchy(
            PropertiesKit.class, SettingsNames.KEY_BINDING_LIST);

        // Go through all levels (PropertiesKit and its supeclasses) and collect key bindings.
        HashSet nextKS = new HashSet();
        HashSet prevKS = new HashSet();
        HashSet toggleKS = new HashSet();
        // Loop thru each keylist for the kit class.
        for (int i = kv.length - 1; i >= 0; i--) {
            List keyList = (List)kv[i].value;
            
            JTextComponent.KeyBinding[] bindings = new JTextComponent.KeyBinding[keyList.size()];
            
            keyList.toArray(bindings);
            
            // Loop thru all bindings in the kit class.
            for(int j=0; j<bindings.length; j++) {
                JTextComponent.KeyBinding binding = bindings[j];
                if(binding == null) 
                    continue;
                
                // Find key keystrokes for find next action.
                if(binding.actionName.equals(BaseKit.findNextAction)) {
                    if(binding instanceof MultiKeyBinding && ((MultiKeyBinding)binding).keys != null)
                        for (int k=0; k<((MultiKeyBinding)binding).keys.length; k++)
                            nextKS.add(((MultiKeyBinding)binding).keys[k]);
                    else
                        nextKS.add(binding.key);
                }
                // Find key keystrokes for find previous action.
                if(binding.actionName.equals(BaseKit.findPreviousAction)) {
                    if(binding instanceof MultiKeyBinding && ((MultiKeyBinding)binding).keys != null)
                        for (int k=0; k<((MultiKeyBinding)binding).keys.length; k++)
                            prevKS.add(((MultiKeyBinding)binding).keys[k]);
                    else
                        prevKS.add(binding.key);
                }
                // Find key keystrokes for toggle highlight action.
                if(binding.actionName.equals(BaseKit.toggleHighlightSearchAction)) {
                    if(binding instanceof MultiKeyBinding && ((MultiKeyBinding)binding).keys != null)
                        for (int k=0; k<((MultiKeyBinding)binding).keys.length; k++)
                            toggleKS.add(((MultiKeyBinding)binding).keys[k]);
                    else
                        toggleKS.add(binding.key);
                }
                
            } // End of inner loop.
        } // End of outer loop.
        
        // Copy found values to our variables.
        nextKS.toArray(keyStrokesFindNext = new KeyStroke[nextKS.size()]);
        prevKS.toArray(keyStrokesFindPrevious = new KeyStroke[prevKS.size()]);
        toggleKS.toArray(keyStrokesToggleHighlight = new KeyStroke[toggleKS.size()]);

        // In case of some problem set the deafutl value.
        if(keyStrokesFindNext == null || keyStrokesFindNext.length == 0) 
            keyStrokesFindNext = TableViewSettingsFactory.FIND_NEXT_DEFAULT_KEYSTROKES;
        if(keyStrokesFindPrevious == null || keyStrokesFindPrevious.length == 0)
            keyStrokesFindPrevious = TableViewSettingsFactory.FIND_PREVIOUS_DEFAULT_KEYSTROKES;
        if(keyStrokesToggleHighlight == null || keyStrokesToggleHighlight.length == 0)
            keyStrokesToggleHighlight = TableViewSettingsFactory.TOGGLE_HIGHLIGHT_DEFAULT_KEYSTROKES;
        
        // notify listeners about update
        settingsUpdated();
    }
    
}
