/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.rmi;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.text.MessageFormat;
import java.util.*;

import org.openide.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.*;
import org.openide.actions.*;
import org.openide.src.*;

import org.netbeans.modules.rmi.settings.*;
import org.netbeans.modules.java.*;

/** The DataLoader for RMIDataObjects.
* This class is final only for performance reasons,
* can be happily unfinaled if desired.
*
* @author Martin Ryzl, Dafe Simonek
*/
public final class RMIDataLoader extends JavaDataLoader {

    transient private static final boolean DEBUG = false;

    /** Serial version UID. */
    static final long serialVersionUID = -3325329576021256358L;

    /** Extended attribute. */
    public static final String EA_DETECT_REMOTE = "NetBeansAttrDetectRemote"; // NOI18N

    /** Name of the property stubFormat. */
    public static final String PROP_STUB_FORMATS = "stubFormats"; // NOI18N
    /** Name of the property hideStubs. */
    public static final String PROP_HIDE_STUBS = "hideStubs"; // NOI18N

    /** Extension for rmi object. */
    public static final String RMI_EXTENSION = "rmi"; // NOI18N
    /** Default extension for remote implementation. */
    public static final String IMPL_SUFFIX = "Impl"; // NOI18N
    /** Inner class divider. */
    static final char INNER_CLASS_DIVIDER = '$';
    /** Default value for stubFormats. */
    public static final String[] DEFAULT_STUB_FORMATS = {"{0}_Stub", "{0}_Skel", "_{0}_Tie", "_{0}_Stub"}; // NOI18N

    private static final int STREAM_VERSION = 1;
    
    private static Parsing.Listener parsingListener = null;
    private static RMISettings settings = (RMISettings) RMISettings.findObject(RMISettings.class, true);

    /** Creates a new RMIDataLoader
     * 
     */ 
    public RMIDataLoader() {
        this (RMIDataObject.class);
    }

    /** Creates a new RMIDataLoader
     * 
     */ 
    public RMIDataLoader(Class recognizedObject) {
        super (recognizedObject);
    }

    protected void initialize () {
        setParsingListener();
        settings.addPropertyChangeListener(new SettingsListener());

        setDisplayName(NbBundle.getMessage(RMIDataLoader.class, "CTL_RMIDataLoader")); // NOI18N
        setActions (new SystemAction [] {
                        SystemAction.get(OpenAction.class),
                        SystemAction.get(CustomizeBeanAction.class),
                        SystemAction.get(FileSystemAction.class),
                        null,
                        SystemAction.get(CompileAction.class),
                        null,
                        SystemAction.get(ExecuteAction.class),
                        null,
                        SystemAction.get(CutAction.class),
                        SystemAction.get(CopyAction.class),
                        SystemAction.get(PasteAction.class),
                        null,
                        SystemAction.get(NewAction.class),
                        SystemAction.get(DeleteAction.class),
                        SystemAction.get(RenameAction.class),
                        null,
                        SystemAction.get(SaveAsTemplateAction.class),
                        null,
                        SystemAction.get(ToolsAction.class),
                        SystemAction.get(PropertiesAction.class),
                    });
    }
    
      /** Creates the right data object for given primary file.
      * It is guaranteed that the provided file is realy primary file
      * returned from the method findPrimaryFile.
      *
      * @param primaryFile the primary file
      * @return the data object for this file
      * @exception DataObjectExistsException if the primary file already has data object
      */
      protected MultiDataObject createMultiObject (FileObject primaryFile)
      throws DataObjectExistsException, IOException {
          return new RMIDataObject(primaryFile, this);
      }

    /** For a given file finds a primary file.
    * @param fo the file to find primary file for
    *
    * @return the primary file for the file or null if the file is not
    *   recognized by this loader
    */
    protected FileObject findPrimaryFile (FileObject fo) {

        // ignore folder
        if (fo.isFolder()) return null;
        
        final String ext = fo.getExt(), fname = fo.getName();
        String name = null;
        FileObject primaryFO;
        boolean rmiFile = false; //
        int index;

        // only for .class, .java or *.rmi
        if (ext.equals(CLASS_EXTENSION) || ext.equals(JAVA_EXTENSION) || ext.equals(RMI_EXTENSION)) {

            // check for inner class
            if ((index = fname.indexOf(INNER_CLASS_DIVIDER)) != -1) {
                name = fname.substring(0, index);
            } else {
                if (isHideStubs()) name = checkStub(fname);
                if (name == null) name = fname;
            }
        }

        // if .rmi doesn't exist it is not an RMI object
        FileObject parent = fo.getParent();
        if ((parent != null) && (name != null)) {
            if ((rmiFile = (parent.getFileObject(name, RMI_EXTENSION)) != null)) {
                if ((primaryFO = parent.getFileObject(name, JAVA_EXTENSION)) != null) {
                    return primaryFO;
                }
            }
        }

        // not recognized
        return null;
    }

    /** Check stub.
    * @return name of the main file or null
    */
    protected String checkStub(final String name) {
        String[] formats = getStubFormats();

        for(int i = 0; i < formats.length; i++) {
            MessageFormat fmt = new MessageFormat(formats[i]);
            try {
                Object[] objs = fmt.parse(name);
                if (objs != null) return objs[0].toString();
            } catch (java.text.ParseException ex) {
                // continue
            }
        }
        return null;
    }

    /** Getter for property hideStubs.
     * @return true if stubs and skeletons should be hidden.
     */
    public boolean isHideStubs() {
        Boolean b = (Boolean) getProperty(PROP_HIDE_STUBS);
        if (b == null) return true;
        return b.booleanValue();
    }
    
    /** Setter for property hideStubs.
     * @param hideStubs - 
     */
    public void setHideStubs(boolean hideStubs) {
        putProperty(PROP_HIDE_STUBS, new Boolean(hideStubs), true);
    }
    
    /** Getter for property stubFormats.
     * @return a list of formats that will be hidden by this loaders
     */
    public String[] getStubFormats() {
        String[] fmts = (String[]) getProperty(PROP_STUB_FORMATS);
        if (fmts == null) return DEFAULT_STUB_FORMATS;
        else return fmts;
    }
    
    /** Setter for property stubFormats.
     */
    public void setStubFormats(String[] formats) {
        putProperty(PROP_STUB_FORMATS, formats, true);
    }

    /** Creates the right primary entry for given primary file.
    *
    * @param primaryFile primary file recognized by this loader
    * @return primary entry for that file
    */
    protected MultiDataObject.Entry createPrimaryEntry (MultiDataObject obj, FileObject primaryFile) {
        return new RMIFileEntry(obj, primaryFile);
    }

    /** Creates right secondary entry for given file. The file is said to
    * belong to an object created by this loader.
    *
    * @param secondaryFile secondary file for which we want to create entry
    * @return the entry
    */
    protected MultiDataObject.Entry createSecondaryEntry (MultiDataObject obj, FileObject secondaryFile) {
        if (secondaryFile.getExt().equals(RMI_EXTENSION)) return new FileEntry(obj, secondaryFile);
        return super.createSecondaryEntry (obj, secondaryFile);
    }

    /** @return The list of extensions this loader recognizes
    * (default list constains class, ser extensions)
    */
    /*
    public ExtensionList getExtensions () {
        if (extensions == null) {
            extensions = new ExtensionList();
            extensions.addExtension(RMI_EXTENSION);
            extensions.addExtension(CLASS_EXTENSION);
            extensions.addExtension(JAVA_EXTENSION);
        }
        return extensions;
    }
     */

    /** Sets the extension list for this data loader.
    * @param ext new list of extensions.
    */
    /*
    public void setExtensions(ExtensionList ext) {
        extensions = ext;
    }
     */

    /** This entry defines the format for replacing the text during
    * instantiation the data object.
    */
    public class RMIFileEntry extends JavaDataLoader.JavaFileEntry {

        /** Serial version UID. */
        static final long serialVersionUID = 2704205254221237611L;

        /** Creates new JavaFileEntry */
        protected RMIFileEntry(MultiDataObject obj, FileObject file) {
            super(obj, file);
        }

        protected void modifyMap(Map map, FileObject target, String n, String e) {
            super.modifyMap(map, target, n, e);

            RMIDataObject obj = (RMIDataObject) getDataObject();
            String iName = n;
            int i;

            // interface name
            if ((i = iName.indexOf(IMPL_SUFFIX)) != -1)
                iName = iName.substring(0, i);
            else
                iName = RMIHelper.REMOTE;

            map.put("INTERFACENAME", iName);   // NOI18N
            map.put("STRING", "\""); // NOI18N

            // string versions of package and name
            String string;
            if ((string = (String) map.get("NAME")) != null) { // NOI18N
                map.put("NAMESTRING", "\"" + string + "\""); // NOI18N
            }
            if ((string = (String) map.get("PACKAGE")) != null) { // NOI18N
                map.put("PACKAGESTRING", "\"" + string + "\""); // NOI18N
            }
            if ((string = getFSPath(getFile())) != null) {
                map.put("FSPATH", string); // NOI18N
            }
        }
    }

    /** Returns path to the filesystem of the file object.
    * @return path to the filesystem*/
    public static String getFSPath(FileObject fo) {
        final StringBuffer sb = new StringBuffer();

        // getPackage
        try {
            FileSystem fs = fo.getFileSystem();
            fo.getFileSystem().prepareEnvironment(new FileSystem.Environment() {
                                                      public void addClassPath(String element) {
                                                          sb.append(element);
                                                      }
                                                  });
            return sb.toString();
        } catch (FileStateInvalidException ex) {
        } catch (EnvironmentNotSupportedException ex) {
        }
        return null;
    }

    /** Mark JDO as RMI.
    * @param jdo - data object to mark
    * @param set - if true, set it as RMI, if false, unset
    */
    public static void markRMI(JavaDataObject jdo, boolean set) throws IOException, java.beans.PropertyVetoException {
        if (jdo == null) return;
        FileObject fo = jdo.getPrimaryFile();

        if (jdo instanceof RMIDataObject) {
            if (!set) {
                // delete rmi marker and revalidate
                FileLock lock = null;
                try {
                    FileObject rmi = FileUtil.findBrother(fo, RMI_EXTENSION);
                    if (rmi != null) {
                        lock = rmi.lock();
                        rmi.delete(lock);
                    }
                    setDetectRemoteEnabled(jdo, true);
                    jdo.setValid(false);
                } catch (Exception ex) {
                    TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
                } finally {
                    if (lock != null) lock.releaseLock();
                }
            }
        } else {
            if (set) {
                // create rmi marker and revalidate
                FileObject parent = fo.getParent();
                try {
                    fo.getParent().createData(fo.getName(), RMI_EXTENSION);
                    jdo.setValid(false);
                } catch (Exception ex) {
                    TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
                }
            }
        }
    }
    
    
    /** Write stubFormats and hideStubs properties. 
     *
     * @param oos - stream
     */
    public void writeExternal(ObjectOutput oos) throws IOException {
        /* 1) call JDO serialization
         * 2) store version (it is also necessary for compatibility with previous
         *    version
         * 3) store properties - name + value
         * 4) null indicates that there is no additional property
         */
        super.writeExternal(oos);
        oos.writeInt(STREAM_VERSION);
        writeProperty(PROP_STUB_FORMATS, oos);
        writeProperty(PROP_HIDE_STUBS, oos);
        oos.writeObject(null);
    }

    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        if (in.available() > 0) {
            in.readInt();
            if (readProperty(in) == null) return;
            if (readProperty(in) == null) return;
            firePropertyChange(null, null, null);
        }
    }
    
    /** Writes a property to the object output.
     * @param name - name of the property
     * @param oos - stream
     */
    private void writeProperty(String name, ObjectOutput oos) throws IOException {
        Object value = getProperty(name);
        if (value != null) {
            oos.writeObject(name);
            oos.writeObject(value);
        }
    }
    
    
    /** Reads a property from the object input.
     * @param ois - stream
     * @return a name of the property or null
     */
    private String readProperty(ObjectInput ois) throws IOException, ClassNotFoundException {
        String name = null;
        Object value;
        name = (String) ois.readObject();
        if (name != null) {
            value = ois.readObject();
            putProperty(name, value, false);
        }
        return name;
    }
    
    /** Tests whether the object should be tested or not.
     */
    static boolean isDetectRemoteEnabled(DataObject dobj) {
        Boolean b = (Boolean) dobj.getPrimaryFile().getAttribute(EA_DETECT_REMOTE);
        if (b != null) return b.booleanValue();
        else return true;
    }

    /**
     */
    static void setDetectRemoteEnabled(DataObject dobj, boolean value) {
        try {
            if (!value) dobj.getPrimaryFile().setAttribute(EA_DETECT_REMOTE, new Boolean(value));
            else dobj.getPrimaryFile().setAttribute(EA_DETECT_REMOTE, null);
        } catch (IOException ex) {
            // ignore
        }
    }

    /** Add parsing listener.
    */
    private static void setParsingListener() {
        boolean isEnabled = settings.isDetectRemote();
        if (isEnabled) {
            if (parsingListener == null) {
                parsingListener = new ParsingListener();
                Parsing.addParsingListener(parsingListener);
            }
        } else { // not enabled
            if (parsingListener != null) {
                Parsing.removeParsingListener(parsingListener);
                parsingListener = null;
            }
        }
    }
    
    /** Cancel the listener. */
    static void cancelParsingListener() {
        if (parsingListener != null) {
            Parsing.removeParsingListener(parsingListener);
        }
    }

    /** The listener interface for everybody who want to control all
    * parsed JavaDataObjects.
    */
    private static class ParsingListener implements Parsing.Listener {

        /** Method which is called everytime when some object is parsed.
        * @param evt The event with the details.
        */
        public void objectParsed(final Parsing.Event evt) {
            RequestProcessor.postRequest(new ParsingPerformer(evt));
        }
    }
    
    private static class ParsingPerformer implements Runnable {

        private String OPTION_YES = null;
        private String OPTION_YES_ALL = null;
        private String OPTION_NO = null;
        private String OPTION_NO_ALL = null;
        private Parsing.Event ev;
        
        public ParsingPerformer(Parsing.Event ev) {
            this.ev = ev;
        }
        
        public synchronized void run() {
            JavaDataObject jdo = ev.getJavaDataObject();
            try {
                 // only JavaDataObject
                 if (!jdo.getClass().equals(JavaDataObject.class) || !isDetectRemoteEnabled(jdo)) return;
                 SourceElement se = ev.getSourceElement();
                 final ClassElement[] ce = se.getAllClasses();
                 out: for(int i = 0; i < ce.length; i++) {
                     if (ce[i].isClass()) {
                         if (RMIHelper.implementsClass(ce[i], RMIHelper.REMOTE)) {
                             // ask user if necessary, eventually modify settings
                            if (getConfirm()) {
                                try {
                                    Object result = TopManager.getDefault().notify(createDescriptor(ce[i]));
                                    if (result == OPTION_YES) {
                                        markRMI(jdo, true);
                                    } else if (result == OPTION_YES_ALL) {
                                        markRMI(jdo, true);
                                        setConfirm(false);
                                    } else if (result == OPTION_NO_ALL) {
                                        setDetect(false);
                                    } else if (result == OPTION_NO) {
                                        setDetectRemoteEnabled(jdo, false);
                                    }
                                } catch (Exception ex) {
                                    TopManager.getDefault().getErrorManager().notify(ex);
                                }
                             } else {
                                 // confirm without asking
                                 try {
                                     markRMI(jdo, true);
                                 } catch (Exception ex) {
                                     TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
                                 }
                             }
                             break out;
                         }
                     }
                 }
             } catch (Exception ex) {
                 TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
             }
        }

        /** Getter for confirm.
        * @return true if the action should be confirmed.
        */
        protected boolean getConfirm() {
            return settings.isConfirmConvert();
        }

        /** Setter for confirm.
        * @param confirm if true, subsequent actions should be confirmed
        */
        protected void setConfirm(boolean confirm) {
            settings.setConfirmConvert(confirm);
        }

        /** Setter for detect.
        * @param detect if true, detect all parsing events and check for implements
        *               Remote
        */
        protected void setDetect(boolean detect) {
            settings.setDetectRemote(detect);
        }

        /** Get notify descriptor.
        * @return a NotifyDesriptor for the confirmation dialog
        */
        protected NotifyDescriptor createDescriptor(ClassElement ce) {
            if (OPTION_YES == null) {
                OPTION_YES = NbBundle.getMessage(RMIDataLoader.class, "CTL_OPTION_YES"); // NOI18N
                OPTION_YES_ALL = NbBundle.getMessage(RMIDataLoader.class, "CTL_OPTION_YES_ALL"); // NOI18N
                OPTION_NO = NbBundle.getMessage(RMIDataLoader.class, "CTL_OPTION_NO"); // NOI18N
                OPTION_NO_ALL = NbBundle.getMessage(RMIDataLoader.class, "CTL_OPTION_NO_ALL"); // NOI18N
            }
            
            return new NotifyDescriptor(
                NbBundle.getMessage(RMIDataLoader.class, "FMT_CONVERT_MESSAGE", ce.getName().getFullName()),   // NOI18N
                NbBundle.getMessage(RMIDataLoader.class, "CTL_CONFIRMATION_TITLE"), // NOI18N
                NotifyDescriptor.DEFAULT_OPTION,
                NotifyDescriptor.QUESTION_MESSAGE,
                new Object[] { OPTION_YES, OPTION_YES_ALL, OPTION_NO, OPTION_NO_ALL },
                null
            );
        }
       
    }

    /** Listens on RMISettings. */
    private static class SettingsListener implements java.beans.PropertyChangeListener {
        public void propertyChange(final java.beans.PropertyChangeEvent ev) {
            if (RMISettings.PROP_DETECT_REMOTE.equals(ev.getPropertyName())) {
                setParsingListener();
            }
        }
    }
}
