/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.rmi;

import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerGroup;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.filesystems.FileSystem;
import org.openide.loaders.MultiDataObject;

/** Compiler that cleans all stubs.
*
* @author Martin Ryzl
*/
public class RMIStubCleanCompiler extends Compiler {

    private RMIDataObject dobj;
    
    /**
    * @param fo FileObject to compile
    * @param fname name of file if different from fo's name (or null)
    * @param nbDescriptor external process
    * @param err error expression
    */
    public RMIStubCleanCompiler(RMIDataObject dobj) {
        super();
        this.dobj = dobj;
    }
    
    /**
     */
    public Class compilerGroupClass() {
        return CleanGroup.class;
    }

    /** Checks if stub and skeleton are up to date.
     *
     */
    public boolean isUpToDate() {

        // always compile - it is necessary because user can change to
        // different stub model
        return false;
    }

    /** Clean given file object.
     * @param formatString - use this format to create stub name then delete it.
     */
    protected void cleanStubs() throws IOException {
        final RMIDataLoader loader = (RMIDataLoader) dobj.getMultiFileLoader();
        final Iterator it = dobj.secondaryEntries().iterator();

        dobj.getPrimaryFile().getFileSystem().runAtomicAction(
            new FileSystem.AtomicAction() {
                public void run() throws IOException {
                    while (it.hasNext()) {
                        MultiDataObject.Entry en = (MultiDataObject.Entry) it.next();
                        FileObject fo = en.getFile();
                        if (loader.checkStub(fo.getName()) != null) {
                            deleteFileObject(fo);
                            dobj.removeSecondaryEntry2(en);
                        }
                    }
                }
            }
        );
    }
    

    /** Delete file object.
    */
    public static void deleteFileObject(FileObject fo) throws IOException {
        FileLock lock = null;
        if (fo == null) {
            return;
        } else {
            try {
                lock = fo.lock();
                fo.delete(lock);
            } finally {
                if (lock != null) {
                    lock.releaseLock();
                }
            }
        }
    }

    /** 
     * @return <tt>true</tt> if the compilers are equal 
     */
    public boolean equals(Object obj) {
        if (obj instanceof RMIStubCleanCompiler) {
            RMIStubCleanCompiler rsc = (RMIStubCleanCompiler) obj;
            return rsc.dobj.equals(dobj);
        }
        return false;
    }

    /** Hash code is delegated to file object.
     */
    public int hashCode() {
        return dobj.hashCode();
    }
  
    /**
     */
    public static class CleanGroup extends CompilerGroup {
        
        private HashSet compilers = new HashSet();
        
        public void add(Compiler c) throws IllegalArgumentException {
            if (c instanceof RMIStubCleanCompiler) {
                synchronized (compilers) {
                    compilers.add(c);
                }
            }
            else throw new IllegalArgumentException();
        }
        
        public boolean start() {
            boolean status = true;
            Iterator it = compilers.iterator();
            while (it.hasNext()) {
                RMIStubCleanCompiler rsc = (RMIStubCleanCompiler) it.next();
                try {
                    rsc.cleanStubs();
                } catch (IOException ex) {
                    if (Boolean.getBoolean("netbeans.debug.exceptions")) ex.printStackTrace(); // NOI18N
                    status = false;
                }
            }
            return status;
        }
    }

}

/*
 * <<Log>>
 *  7    Gandalf-post-FCS1.5.1.0     3/20/00  Martin Ryzl     localization
 *  6    Gandalf   1.5         1/24/00  Martin Ryzl     compilation of inner 
 *       classes added
 *  5    Gandalf   1.4         1/21/00  Martin Ryzl     compilation fixed (new 
 *       API)
 *  4    Gandalf   1.3         12/23/99 Jaroslav Tulach mergeInto deleted from 
 *       the CompilerJob.
 *  3    Gandalf   1.2         10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun
 *       Microsystems Copyright in File Comment
 *  2    Gandalf   1.1         10/7/99  Martin Ryzl     
 *  1    Gandalf   1.0         10/6/99  Martin Ryzl     
 * $
 */
