/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.rmi;

import java.beans.*;
import java.util.*;

import org.openide.compiler.*;
import org.openide.compiler.Compiler;
import org.openide.cookies.SourceCookie;
import org.openide.filesystems.FileObject;
import org.openide.loaders.*;
import org.openide.execution.NbProcessDescriptor;
import org.openide.src.*;
import org.openide.util.*;


/**
 *
 * @author  mryzl
 */

public class RMIStubCompilerType extends ExternalCompilerType {

    /** Serial version UID. */
    static final long serialVersionUID = 3511317103361547621L;

    /** Resource bundle. */
    private static final ResourceBundle bundle = NbBundle.getBundle(RMIStubCompilerType.class);

    /** Name of the property stubCompiler. */
    public static final String PROP_StubCompiler = "stubCompiler"; // NOI18N

    /** Name of the property. */
    public static final String PROP_KEEP_GENERATED = "keepGenerated"; // NOI18N

    /** Holds value of property stubCompiler. */
    private NbProcessDescriptor stubCompiler;

    /** Utility field used by bound properties. */
    private java.beans.PropertyChangeSupport propertyChangeSupport = new java.beans.PropertyChangeSupport (this);

    /** Default process. */
    public static final String DEFAULT_COMPILER_PROCESS = "{jdk.home}{/}bin{/}rmic"; // NOI18N

    /** Default arguments. */
    public static final String DEFAULT_COMPILER_ARGUMENTS = "-classpath \"{java.home}{/}..{/}lib{/}rt.jar\"{:}{filesystems} " + // NOI18N
        "{" + RMIStubCompilerGroup.TAG_KEEP_GENERATED + "} " +  // NOI18N
        "-d \"{" + RMIStubCompilerGroup.TAG_PACKAGEROOT + "}\" {files}"; // NOI18N

    /** Info. */
    public static final String INFO = bundle.getString("MSG_CompilerHint");
    
    /** Holds value of property keepGenerated. */
    private boolean keepGenerated;
    
 // NOI18N


    /** Creates new RMIStubCompilerType. */
    public RMIStubCompilerType() {
        setExternalCompiler(new NbProcessDescriptor(
            DEFAULT_COMPILER_PROCESS,
            DEFAULT_COMPILER_ARGUMENTS,
            INFO
        ));
    }

    public void prepareJob(CompilerJob job, Class type, DataObject obj) {
        if (type.equals(RMIStubCompiler.CLEAN)) return; // handle clean separately

        FileObject fo = obj.getPrimaryFile();
        String[] classes = getRemoteClasses(obj, false);

        if (classes.length > 0) {
            for(int i = 0; i < classes.length; i++) {
                job.add(createCompiler(fo, classes[i], type));
            }
        } /* else {
            String foName = '/' + fo.getPackageName('/') + '.' + RMIDataLoader.CLASS_EXTENSION;
            job.add(createCompiler(fo, null, type));
        }
           */
    }
    
    private RMIStubCompiler createCompiler(FileObject src, String name, Class type) {
        return new RMIStubCompiler(src, name, type, getExternalCompiler(), getErrorExpression(), this);
    }

    /** Return names of remote classes in the file.
     * @param obj - source data object
     * @param pkg - if true, it returns fully qualified names (with package)
     *              if false, it returns fully qualified names without the package
     * @return names of remote classes in form [pkg.pkg.]class$inner$inner (
     *         [] means optional
     */
    protected static String[] getRemoteClasses(DataObject obj, boolean pkgFlag) {
        LinkedList list = new LinkedList();

        SourceCookie sc = (SourceCookie) obj.getCookie(SourceCookie.class);
        SourceElement se = sc.getSource();
        Identifier pkg = se.getPackage();
        String pname = null;
        int index = 0;
        if (pkg != null) {
            pname = pkg.getFullName();
            index = pname.length();
        }

        ClassElement[] ces = se.getAllClasses();
        for(int i = 0; i < ces.length; i++) {
            if (ces[i].isClass() && RMIHelper.implementsClass(ces[i], RMIHelper.REMOTE)) {
                String name = ces[i].getName().getFullName();
                if (pkg != null) {
                    name = name.substring(index + 1);
                    name = name.replace('.', '$');
                    if (pkgFlag) name = pname + '.' + name;
                } else {
                    name = name.replace('.', '$');
                }
                list.add(name);
            }
        }

        return (String[])list.toArray(new String[list.size()]);
    }

    /** Getter for process descriptor describing the compiler.
     * @returns a process descriptor. It returns a default value instead of null.
     */
    public NbProcessDescriptor getExternalCompiler() {
        NbProcessDescriptor nbp = super.getExternalCompiler();
        if (nbp == null) { // patch because of backward compatibility
            return new NbProcessDescriptor(
                    DEFAULT_COMPILER_PROCESS,
                    DEFAULT_COMPILER_ARGUMENTS,
                    INFO
                   );
        } else return nbp; 
    }
    
    /** Getter for current error expression.
     * @returns an error expression. It returns a default value instead of null.
     */
    public ExternalCompiler.ErrorExpression getErrorExpression() {
        ExternalCompiler.ErrorExpression ee = super.getErrorExpression();
        if (ee != null) return ee;
        else return ExternalCompiler.JAVAC;  
    }
    
    /** human presentable name */
    public String displayName() {
        return bundle.getString("CTL_RMIStubCompilerType"); // NOI18N
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx(RMIStubCompilerType.class);
    }
    
    /** Getter for property keepGenerated.
     * @return Value of property keepGenerated.
     */
    public boolean isKeepGenerated() {
        return keepGenerated;
    }
    
    /** Setter for property keepGenerated.
     * @param keepGenerated New value of property keepGenerated.
     */
    public void setKeepGenerated(boolean keepGenerated) {
        boolean oldValue = this.keepGenerated;
        this.keepGenerated = keepGenerated;
        firePropertyChange(PROP_KEEP_GENERATED, new Boolean(oldValue), new Boolean(keepGenerated));
    }
    
}

/*
* <<Log>>
*  10   Gandalf-post-FCS1.8.1.0     3/20/00  Martin Ryzl     localization
*  9    Gandalf   1.8         1/31/00  Martin Ryzl     problems with compilation
*       files with null package removed
*  8    Gandalf   1.7         1/28/00  Martin Ryzl     if RMICompiler used with 
*       non remote object it writes error message
*  7    Gandalf   1.6         1/24/00  Martin Ryzl     compilation of inner 
*       classes added
*  6    Gandalf   1.5         1/21/00  Martin Ryzl     compilation fixed (new 
*       API)
*  5    Gandalf   1.4         12/23/99 Jaroslav Tulach mergeInto deleted from 
*       the CompilerJob.
*  4    Gandalf   1.3         11/27/99 Patrik Knakal   
*  3    Gandalf   1.2         10/23/99 Ian Formanek    NO SEMANTIC CHANGE - Sun 
*       Microsystems Copyright in File Comment
*  2    Gandalf   1.1         10/6/99  Martin Ryzl     debug info removed
*  1    Gandalf   1.0         10/6/99  Martin Ryzl     
* $ 
*/ 
