/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.rmi.wizard;

import java.awt.event.ActionEvent;
import java.beans.*;
import java.io.*;
import java.lang.ref.*;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;


import org.openide.*;
import org.openide.cookies.*;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileObject;
import org.openide.loaders.*;
import org.openide.src.*;
import org.openide.util.NbBundle;

/** 
 *
 * @author  mryzl, Jan Pokorsky
 */

public class RMIWizard extends WizardDescriptor {
    
    /** Wizard panel index. */
    public static final int RMIWIZARDINTRO = 0;
    /** Wizard panel index. */
    public static final int RMIWIZARD1_2 = 1;
    /** Wizard panel index. */
    public static final int RMIWIZARD2 = 2;
    /** Wizard panel index. */
    public static final int RMIWIZARD3 = 4;
    /** Wizard panel index. */
    public static final int RMIWIZARD4 = 3;
    /** Wizard panel index. */
    public static final int RMIWIZARDFINISH = 5;
    
    /** An empty string. */
    public static final String EMPTY_STRING = ""; // NOI18N
    
    /** Default name for interface. */
    public static final String DEFAULT_INTERFACE = "NewRemote"; // NOI18N
    
    /** Default name for method. */
    public static final String DEFAULT_METHOD = "newMethod"; // NOI18N
    
    /** Location of RMI templates. */
    public static final String TEMPLATES_RMI = "RMI"; // NOI18N
    
    /** Location of interface templates. */
    public static final String TEMPLATES_INTERFACES = "interfaces"; // NOI18N
    
    /** Location of implementation templates. */
    public static final String TEMPLATES_IMPLEMENTATIONS = "implementations"; // NOI18N

        /** Preferred dimension. */
//    public static final java.awt.Dimension DIM = new java.awt.Dimension (640, 320);
    public static final java.awt.Dimension DIM = new java.awt.Dimension (500, 380);
    
    /** Name of the property. */
    public static final String PROP_TARGET_FOLDER = "targetFolder"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_IMPLEMENTATION_NAME = "implementationName"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_IMPLEMENTATION_TEMPLATE = "implementationTemplate"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_INTERFACE_TEMPLATE = "interfaceTemplate"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_INTERFACE_NAME = "interfaceName"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_REMOTE_METHODS = "remoteMethods"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_SUPERINTERFACE_OBJECT = "superInterfaceObject"; // NOI18N
    
    /** Name of the property. */
    public static final String PROP_SUPERINTERFACE_NAME = "superInterfaceName"; // NOI18N
    
    /** Property change support. */
    private PropertyChangeSupport support;
    
    /** Holds value of property targetFolder. */
    private DataFolder targetFolder;

    /** Holds value of property interfaceName. */
    private String interfaceName;

    /** Holds value of property superInterfaceName. */
    private String superInterfaceName;

    /** Holds value of property superInterfaceObject. */
    private DataObject superInterfaceObject;

    /** Holds value of property interfaceTemplate. */
    private DataObject interfaceTemplate;
    
    /** Holds value of property newInterface. */
    private boolean newInterface = true;
    
    /** Holds value of property implemenationName. */
    private String implementationName;
    
    /** Holds value of property implementationTemplate. */
    private DataObject implementationTemplate;
    
    /** Holds value of property remoteMethods. */
    private MethodElement[] remoteMethods;
    
    /** Template data filter. */
    private static DataFilter templateFilter = new TemplateFilter();
    
    /** Wizard iterator. */
    private static RMIIterator iterator;
    
    /** Creates new RMIWizard. 
    */
    public RMIWizard() {
        //super(createDefaultIterator());
        super(iterator = new RMIIterator());
        setTitleFormat(new MessageFormat(NbBundle.getBundle(RMIWizard.class).getString("FMT_Title"))); // NOI18N
        putProperty("WizardPanel_autoWizardStyle", new Boolean(true)); // NOI18N
        putProperty("WizardPanel_contentDisplayed", new Boolean(true)); // NOI18N
        putProperty("WizardPanel_contentNumbered", new Boolean(true)); // NOI18N
        putProperty("WizardPanel_contentData", new String[] { // NOI18N
            getString("RMIWizardIntro.Form.step"), // NOI18N
            getString("RMIWizard1.Form.step"), // NOI18N
            getString("RMIWizard2.Form.step"), // NOI18N
            getString("RMIWizard4.Form.step"), // NOI18N
            getString("RMIWizard3.Form.step"), // NOI18N
            getString("RMIWizardFinish.Form.step") // NOI18N
        });
        setTitle(getString("RMIWizard.title")); // NOI18N
        //setModal(false);
    }
    
    /** Localization */
    private static String getString(String key) {
        return org.openide.util.NbBundle.getMessage(RMIWizard.class, key);
    }
    
    /** Add an interfac to a ClassElement. 
     * @param ce c - the ClassElement
     * @param ifcs - interface
     * @return the ClassElement
     */
    private ClassElement addInterface(ClassElement ce, Identifier ifc) {
        try {
            Identifier[] old = ce.getInterfaces();
            for(int i = 0; i < old.length; i++) {
                if (old[i].equals(ifc)) return ce;
            }
            ce.addInterface(ifc);
        } catch (java.lang.Exception ex) {
            // ignore
        }
        return ce;
    }
    
    /** Add a listener. 
    */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        if (support == null) support = new PropertyChangeSupport(this);
        support.addPropertyChangeListener(listener);
    }
    
    /** Remove the listener. 
    */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        if (support != null) support.removePropertyChangeListener(listener);
    }
    
    /* Fire property change. 
    * @param name - name of the property
    * @param oldValue - the old value
    * @param newValue - a new value
    */
    protected void firePropertyChange(String name, Object oldValue, Object newValue) {
        if (support != null) support.firePropertyChange(name, oldValue, newValue);
    }
    
    /** Create an interface.
     * @param current current panel index
    */
    protected DataObject instantiateInterface(int current) throws IOException, SourceException {
        DataObject tmpl = getInterfaceTemplate(), dobj;

        if (tmpl != null) {  // create data object from template
            dobj = tmpl.createFromTemplate(getTargetFolder(), getInterfaceName());
        } else { // create data object from scratch

            // create an empty class element
            ClassElement el = new ClassElement(); // new class element
            el.setName(createIdentifier(getTargetFolder(), getInterfaceName())); // name
            el.setClassOrInterface(false);  // interface
            el.setModifiers(java.lang.reflect.Modifier.PUBLIC); // public

            // superinterfac
            /*
            Identifier[] exInterfaces;
            exInterfaces = new Identifier[] {
                Identifier.create("java.rmi.Remote"), // NOI18N
            };
            el.setInterfaces(exInterfaces);
             */

            // create an empty java file
            FileObject target = getTargetFolder().getPrimaryFile();
            FileObject fo = target.createData(getInterfaceName(), org.netbeans.modules.rmi.RMIDataLoader.JAVA_EXTENSION); // NOI18N
            dobj = DataObject.find(fo);

            // put it into
            SourceCookie cookie = (SourceCookie) dobj.getCookie(SourceCookie.class);
            if (cookie != null) {
                SourceElement se = cookie.getSource();

                // set package
                String pkgName = target.getPackageName('.').trim();
                if (pkgName.length() > 0) se.setPackage(Identifier.create(pkgName));
                se.addClass(el);
            }
        }

        SourceCookie cookie = (SourceCookie) dobj.getCookie(SourceCookie.class);
        if (cookie != null) {
            SourceElement se = cookie.getSource();
            ClassElement el = se.getClass(createIdentifier(dobj));
            if (el != null) {
                // add superinterface
                if (getSuperInterfaceName() != null) {
                    addInterface(el, Identifier.create(getSuperInterfaceName()));
                }

                addInterface(el, Identifier.create("java.rmi.Remote")); // NOI18N

                // add methods
                MethodElement[] mes = getRemoteMethods(), mes2;
                if (mes != null && current > 1) {
                    // clone methods, set null bodies
                    mes2 = createInterfaceMethods(mes);
                    // commentMethods(mes2);
                    el.addMethods(mes2);
                }
            }
        }

        saveDataObject(dobj);
        return dobj;
    }
    
    /** Create an implementation.
     */
    protected DataObject instantiateImplementation() throws IOException, SourceException {
        DataObject dobj = null;
        
        if (getImplementationName() != null) {
            DataObject tmpl = getImplementationTemplate();

            if (tmpl != null) {  // create data object from template
                dobj = tmpl.createFromTemplate(getTargetFolder(), getImplementationName());
            } else { // create data object from scratch

                // create an empty class element
                ClassElement el = new ClassElement(); // new class
                Identifier id = createIdentifier(getTargetFolder(), getImplementationName());
//                    System.err.println("id = " + id + ", full = " + id.getFullName());
                el.setName(id); // name
                el.setClassOrInterface(true);  // class
                el.setModifiers(java.lang.reflect.Modifier.PUBLIC); // public

                // create an empty rmi file
                final FileObject target = getTargetFolder().getPrimaryFile();
                final FileObject[] fos = new FileObject[1];

                target.getFileSystem().runAtomicAction(new FileSystem.AtomicAction() {
                    public void run() throws IOException {
                        target.createData(getImplementationName(), org.netbeans.modules.rmi.RMIDataLoader.RMI_EXTENSION);
                        fos[0] = target.createData(getImplementationName(), org.netbeans.modules.rmi.RMIDataLoader.JAVA_EXTENSION);
                    }
                });
                dobj = DataObject.find(fos[0]);

                // put it into
                SourceCookie cookie = (SourceCookie) dobj.getCookie(SourceCookie.class);
                if (cookie != null) {
                    SourceElement se = cookie.getSource();

                    // set package
                    String pkgName = target.getPackageName('.').trim();
                    if (pkgName.length() > 0) se.setPackage(Identifier.create(pkgName));

                    se.addClass(el);
                }
            }

            // add methods
            MethodElement[] mes2;
            MethodElement[] mes = getRemoteMethods();

            SourceCookie cookie = (SourceCookie) dobj.getCookie(SourceCookie.class);
            if (cookie != null) {
                SourceElement se = cookie.getSource();
                ClassElement el = se.getClass(createIdentifier(dobj));
                if (el != null) {
                    if (mes != null) {
                        mes2 = createImplementationMethods(mes);
                        //  commentMethods(mes2);
                        el.addMethods(mes2);
                    }

                    addInterface(el, Identifier.create(getInterfaceName()));
                    // add methods from superinterfaces
                    mes2 = createImplementationMethods(
                        getInterfaceMethods((isNewInterface())? getSuperInterfaceName(): getInterfaceName()));
                    el.addMethods(mes2);
                    mes2 = createImplementationMethods(getTemplateMethods());
                    el.addMethods(mes2);
                    el.setInterfaces(new Identifier[] { Identifier.create(getInterfaceName()) });
                }
                /*
                System.err.println("RMIWizard: el = " + el);
                Identifier id1 = createIdentifier(getTargetFolder(), getImplementationName()),
                           id2 = createIdentifier(dobj);
                System.err.println("id1 = " + id1 + ", full = " + id1.getFullName());
                System.err.println("id2 = " + id2 + ", full = " + id2.getFullName());
                System.err.println("compareTo = " + (id1.compareTo(id2, false)));
                ClassElement[] ces = se.getClasses();
                for(int i = 0; i < ces.length; i++) {
                    System.err.println("ces[" + i + "] " + ces[i].getName() + ", full = " + ces[i].getName().getFullName() + ", compareTo(id2) = " + id2.compareTo(ces[i].getName(), false));
                }
                 */
            }

            saveDataObject(dobj);
        }
        return dobj;
    } 

    /** Instantiates the template using informations provided by
    * the wizard.
    *
    * @param wiz the wizard
    * @return set of data objects that has been created (should contain
    *   at least one) 
    * @exception IOException if the instantiation fails
    */
    public java.util.Set instantiate() throws IOException {
        return instantiate(null);
    }
    
    /** Instantiates the template using informations provided by
    * the wizard.
    *
    * @param wiz the wizard
    * @return set of data objects that has been created (should contain
    *   at least one) 
    * @exception IOException if the instantiation fails
    */
    public java.util.Set instantiate(DataFolder targetFolder) throws IOException {
        setTargetFolder(targetFolder);
        
        java.awt.Dialog d = TopManager.getDefault().createDialog(this);
        d.show ();
        d.dispose();

        if (getValue() == FINISH_OPTION) {
            return handleInstantiate();
        } else {
            return null;
        }
    }
    
    /** Handle instantiate.
    */
    private java.util.Set handleInstantiate() throws IOException {
        java.util.HashSet set = new java.util.HashSet();
        try {
            int current = ((Integer) ((javax.swing.JComponent) iterator.current().getComponent()).getClientProperty("WizardPanel_contentSelectedIndex")).intValue(); // NOI18N

            DataObject dobj;
            if (isNewInterface()) {
                dobj = instantiateInterface(current);
                if (dobj != null) {
                    openDataObject(dobj);
                    set.add(dobj);
                }
            }

            if (current < 3) return set;
            dobj = instantiateImplementation();
            if (dobj != null) {
                openDataObject(dobj);
                set.add(dobj);
            }
        } catch (SourceException ex) {
            TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, ex);
        }
        return set;
    }

    /** Getter for property targetFolder.
     * @return Value of property targetFolder.
     */
    public DataFolder getTargetFolder() {
//        if (targetFolder == null) targetFolder = getDefaultFolder();
        return targetFolder;
    }

    /** Setter for property targetFolder.
     * @param targetFolder New value of property targetFolder.
     */
    public void setTargetFolder(DataFolder targetFolder) {
        DataFolder oldValue = this.targetFolder;
        this.targetFolder = targetFolder;
        firePropertyChange(PROP_TARGET_FOLDER, oldValue, targetFolder);
    }

    /** Getter for property interfaceName.
     * @return Value of property interfaceName.
     */
    public String getInterfaceName() {
        return interfaceName;
    }

    /** Setter for property interfaceName.
     * @param interfaceName New value of property interfaceName.
     */
    public void setInterfaceName(String interfaceName) {
        String oldValue = this.interfaceName;
        this.interfaceName = interfaceName;
        firePropertyChange(PROP_INTERFACE_NAME, oldValue, interfaceName);
    }
    
    /** Is the interface new or existing? */
    public boolean isNewInterface() {
        return newInterface;
    }
    
    /** Set interface type. True means create new, false means use existing. */
    public void setNewInterface(boolean newInterface) {
        this.newInterface = newInterface;
    }    
    
    /** Getter for property superInterfaceName.
     * @return Value of property superInterfaceName.
     */
    public String getSuperInterfaceName() {
        return superInterfaceName;
    }

    /** Setter for property superInterfaceName.
     * @param superInterfaceName New value of property superInterfaceName.
     */
    public void setSuperInterfaceName(String superInterfaceName) {
        String oldValue = this.superInterfaceName;
        this.superInterfaceName = superInterfaceName;
        firePropertyChange(PROP_SUPERINTERFACE_NAME, oldValue, superInterfaceName);
    }

    /** Getter for property superInterfaceObject.
     * @return Value of property superInterfaceObject.
     */
    public DataObject getSuperInterfaceObject() {
        return superInterfaceObject;
    }

    /** Setter for property superInterfaceObject.
     * @param superInterfaceObject New value of property superInterfaceObject.
     */
    public void setSuperInterfaceObject(DataObject superInterfaceObject) {
        DataObject oldValue = this.superInterfaceObject;
        this.superInterfaceObject = superInterfaceObject;
        firePropertyChange(PROP_SUPERINTERFACE_OBJECT, oldValue, superInterfaceObject);
    }
    
    /** Getter for property interfaceTemplate.
     * @return Value of property interfaceTemplate.
     */
    public DataObject getInterfaceTemplate() {
        return interfaceTemplate;
    }
    
    /** Setter for property interfaceTemplate.
     * @param interfaceTemplate New value of property interfaceTemplate.
     */
    public void setInterfaceTemplate(DataObject interfaceTemplate) {
        DataObject oldValue = this.interfaceTemplate;
        this.interfaceTemplate = interfaceTemplate;
        firePropertyChange(PROP_INTERFACE_TEMPLATE, oldValue, interfaceTemplate);
    }
    
    /** Getter for property implemenationName.
     * @return Value of property implemenationName.
     */
    public String getImplementationName() {
        return implementationName;
    }
    
    /** Setter for property implemenationName.
     * @param implemenationName New value of property implemenationName.
     */
    public void setImplementationName(String implementationName) {
        String oldValue = this.implementationName;
        this.implementationName = implementationName;
        firePropertyChange(PROP_IMPLEMENTATION_NAME, oldValue, implementationName);
    }
    
    /** Getter for property implementationTemplate.
     * @return Value of property implementationTemplate.
     */
    public DataObject getImplementationTemplate() {
        return implementationTemplate;
    }
    
    /** Setter for property implementationTemplate.
     * @param implementationTemplate New value of property implementationTemplate.
     */
    public void setImplementationTemplate(DataObject implementationTemplate) {
        DataObject oldValue = this.implementationTemplate;
        this.implementationTemplate = implementationTemplate;
        firePropertyChange(PROP_IMPLEMENTATION_TEMPLATE, oldValue, implementationTemplate);
    }
    
    /** Getter for property remoteMethods.
     * @return Value of property remoteMethods.
     */
    public MethodElement[] getRemoteMethods() {
        return remoteMethods;
    }
    
    /** Setter for property remoteMethods.
     * @param remoteMethods New value of property remoteMethods.
     */
    public void setRemoteMethods(MethodElement[] remoteMethods) {
        Object oldValue = this.remoteMethods;
        this.remoteMethods = remoteMethods;
        firePropertyChange(PROP_REMOTE_METHODS, oldValue, remoteMethods);
    }
    
    
    /** Get default target folder. 
     * @return default data folder or null if there is no filesystem
     */
    public static DataFolder getDefaultFolder() {
        try {
            FileSystem fs = null;
            // search for a file system
            java.util.Enumeration en = org.openide.TopManager.getDefault ().getRepository().fileSystems ();
            while (en.hasMoreElements()) {
                fs = (FileSystem)en.nextElement();
                if (!fs.isHidden()) {
                    // found first non hidden file system
                    return DataFolder.findFolder(fs.getRoot());
                }
            }
        } catch (java.lang.Exception ex) {
              // return null
        }
        return null;
    }
    
    //
    
    /** Create identifier for data object. For example for
     * data object Test in package test will be created identifier
     * by Identifier.create("test.Test", "Test");
     * @param dobj - data object
     * @return an identifier
     */
    public static Identifier createIdentifier(DataObject dobj) {
        FileObject fo = dobj.getPrimaryFile();
        return Identifier.create(fo.getPackageName('.'), fo.getName());
    }
    
    /** Create identifier for data object. For example for
     * data folder test and name Test will be created identifier
     * by Identifier.create("test.Test", "Test");
     * @param folder - a folder. May be null.
     * @param name - name 
     * @return an identifier
     */
    public static Identifier createIdentifier(DataFolder folder, String name) {

        if (folder == null) return Identifier.create(name); // default behaviour
        
        // folder may be root ...
        String pkgName = folder.getPrimaryFile().getPackageName('.').trim(), fullName;
        if (pkgName.length() > 0) fullName = pkgName + '.' + name;
        else fullName = name;
        
        return Identifier.create(fullName, name);
    }
    
    /**
     */
    public static MethodElement[] cloneMethods(MethodElement[] mes) {
        MethodElement[] mes2 = new MethodElement[mes.length];
        for(int i = 0; i < mes2.length; i++) mes2[i] = (MethodElement) mes[i].clone();
        return mes2;
    }

    /** 
     * @return an array of method elements
     */
    protected MethodElement[] getTemplateMethods() {
        DataObject dobj = getInterfaceTemplate();
        MethodElement[] mes = null;
        if (dobj != null) {
            SourceCookie cookie = (SourceCookie) dobj.getCookie(SourceCookie.class);
            if (cookie != null) {
                SourceElement el = cookie.getSource();
                Identifier id = Identifier.create(dobj.getPrimaryFile().getName());
                ClassElement ce = el.getClass(id);
                if (ce != null) {
                    mes = RMIWizard.cloneMethods(ce.getMethods());
                }
            }
        }
        if (mes != null) return mes;
        else return new MethodElement[] {};
    }
    
    /** Find all methods that are defined in superinterfaces
     */
    protected MethodElement[] getInterfaceMethods(String inerfaceName) {
        // [TODO] depth search through all interfaces, warnings
        java.util.LinkedList list = new java.util.LinkedList();
        
        if (inerfaceName != null) {
            ClassElement ce = ClassElement.forName(inerfaceName);
            if ((ce != null) && (!ce.isClassOrInterface())) {
                MethodElement[] mes = ce.getMethods();
                for(int i = 0; i < mes.length; i++) {
                    list.add(mes[i]);
                }
            }
        }
        
        return (MethodElement[]) list.toArray(new MethodElement[list.size()]);
    }
    
    /** Create interface methods from original.
     */
    protected MethodElement[] createInterfaceMethods(MethodElement[] mes) {
        MethodElement[] mes2 = new MethodElement[mes.length];
        for(int i = 0; i < mes2.length; i++) {
            mes2[i] = (MethodElement) mes[i].clone();
            // only public modifiers
            // no body
            try {
                mes2[i].setModifiers(java.lang.reflect.Modifier.PUBLIC);
                mes2[i].setBody(null);
            } catch (SourceException ex) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) ex.printStackTrace(); // NOI18N
            }
        }
        return mes2;
    }
    
    /** Create implementation methods from original.
     */
    protected MethodElement[] createImplementationMethods(MethodElement[] mes) {
        MethodElement[] mes2 = new MethodElement[mes.length];
        for(int i = 0; i < mes2.length; i++) {
            mes2[i] = (MethodElement) mes[i].clone();
            // only public modifiers
            // no body
            int mod = mes2[i].getModifiers();
            int mask = Modifier.PRIVATE | Modifier.PROTECTED;
            try {
                mes2[i].setModifiers((mod | Modifier.PUBLIC | mask) ^ mask ); // only public
                if ((mod & (Modifier.ABSTRACT | Modifier.NATIVE)) == 0) mes2[i].setBody("\n"); // NOI18N
                else mes2[i].setBody(null);
            } catch (SourceException ex) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) ex.printStackTrace(); // NOI18N
            }
        }
        return mes2;
    }
    
    /** Create default comments for methods. 
     */
    protected void commentMethods(MethodElement[] mes) {
        // [TODO] doesn't work, don't use it
        for(int i = 0; i < mes.length; i++) {
            // create params
            MethodParameter[] mps = mes[i].getParameters();
            JavaDocTag.Param[] params = new JavaDocTag.Param[mps.length];
            for(int j = 0; j < params.length; j++) {
                params[j] = JavaDocSupport.createParamTag(mps[j].getName(), EMPTY_STRING);
            }
            // create return
            // create throws
            Identifier[] is = mes[i].getExceptions();
            JavaDocTag.Throws[] exs = new JavaDocTag.Throws[is.length];
            for(int j = 0; j < exs.length; j++) {
               exs[j] = JavaDocSupport.createThrowsTag(is[j].getFullName(), EMPTY_STRING);
            }
            // set javadoc
            try {
                mes[i].getJavaDoc().changeTags(params, JavaDoc.ADD);
                mes[i].getJavaDoc().changeTags(exs, JavaDoc.ADD);
            } catch (SourceException ex) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) ex.printStackTrace(); // NOI18N
            }
        }
    }
    
    /** Check whether a file with given name and extension does exist in the 
     * target folder.
     * @param name - name of the file
     * @param ext - extension (may be null)
     * @return true if the file exists, false if the file doesn't exist or
     *               there is no target folder
     */
    public boolean isFileExist(String name, String ext) {
        DataFolder df = getTargetFolder();
        if (df != null) {
            FileObject ffolder = df.getPrimaryFile();
            if (ext != null) return ffolder.getFileObject(name, ext) != null;
            else return ffolder.getFileObject(name) != null;
        }
        return false;
    }
    
    /** Open a data object.
     */
    public static void openDataObject(DataObject dobj) {
        OpenCookie cookie = (OpenCookie) dobj.getCookie(OpenCookie.class);
        if (cookie != null) cookie.open();
    }
    
    
    /** Save a data object.
    */
    public static void saveDataObject(DataObject dobj) throws IOException {
        SaveCookie cookie = (SaveCookie) dobj.getCookie(SaveCookie.class);
        if (cookie != null) cookie.save();
    }

    /** Get data objects (not folders) from given folder.
     * @param folder - parent data folder, may be null
     * @return an array of data objects or null if parent folder is null
     */
    public static DataObject[] getDataObjects(DataFolder folder, DataFilter filter) {

        if (folder == null) return null;
        
        LinkedList list = new LinkedList();
        java.util.Enumeration en = folder.children();
        while (en.hasMoreElements()) {
            DataObject dobj = (DataObject) en.nextElement();
            if (filter.acceptDataObject(dobj)) list.add(dobj);
        }
        return (DataObject[]) list.toArray(new DataObject[list.size()]);
    }
    
    /** Get folder by name.
     * @param df - parent data folder
     * @param name - name of the folder to be looked for
     * @return data folder or null
     */
    public static DataFolder getFolder(DataFolder df, String name) {
        DataObject[] children = df.getChildren();
        for(int i = 0; i < children.length; i++) {
            if ((children[i] instanceof DataFolder) && children[i].getName().equals(name)) {
                return (DataFolder) children[i];
            }
        }
        return null;
    }

    /** Get interface templates. 
    * @return an array of interface templates
    */
    public static DataObject[] getInterfaceTemplates() {
        DataFolder rmi = getFolder(TopManager.getDefault().getPlaces().folders().templates(), TEMPLATES_RMI);
        DataFolder interfaces =  null;
        if (rmi != null) interfaces = getFolder(rmi, TEMPLATES_INTERFACES);
        if (interfaces != null) return getDataObjects(interfaces, templateFilter);
        return new DataObject[] {};
    }
    
    /** Get implementation templates.
    * @return an array of templates
    */
    public static DataObject[] getImplementationTemplates() {
        DataFolder rmi = getFolder(TopManager.getDefault().getPlaces().folders().templates(), TEMPLATES_RMI);
        DataFolder implementations = null;
        if (rmi != null) implementations = getFolder(rmi, TEMPLATES_IMPLEMENTATIONS);
        if (implementations!= null) return getDataObjects(implementations, templateFilter);
        return new DataObject[] {};
    }
        
    /** For debugging.
    */
    public static void main(String[] args) throws java.lang.Exception {
        RMIWizard wizard = new RMIWizard();
        System.err.println("loader = " + wizard.getClass().getClassLoader());
        wizard.instantiate(null);
        
    } 
    
    // inner classes
    static class DataObjectListModel implements javax.swing.ComboBoxModel {

        DataObject selected;
        DataObject[] objs;
        final String NONE;
        
        public DataObjectListModel(DataObject[] objs) {
            this.objs = objs;
            NONE = NbBundle.getBundle(RMIWizard.class).getString("LBL_NONE"); // NOI18N
        }
        
        /** Utility field used by event firing mechanism. */
        private javax.swing.event.EventListenerList listenerList =  null;

        public int getSize() {
            return objs.length + 1;
        }

        public java.lang.Object getElementAt(int p1) {
            if (p1 == 0) return NONE;
            else return objs[p1 - 1].getName();
        }
        
        public DataObject getByName(String name) {
            for(int i = 0; i < objs.length; i++) {
                if (objs[i].getName().equals(name)) return objs[i];
            }
            return null;
        }
        
        /** Registers ListDataListener to receive events.
         * @param listener The listener to register.
         */
        public synchronized void addListDataListener(javax.swing.event.ListDataListener listener) {
            if (listenerList == null ) {
                listenerList = new javax.swing.event.EventListenerList();
            }
            listenerList.add (javax.swing.event.ListDataListener.class, listener);
        }
        /** Removes ListDataListener from the list of listeners.
         * @param listener The listener to remove.
         */
        public synchronized void removeListDataListener(javax.swing.event.ListDataListener listener) {
            listenerList.remove (javax.swing.event.ListDataListener.class, listener);
        }
        
        /** Notifies all registered listeners about the event.
         *
         * @param source - the source object
         * @param type - type of the event
         * @param index0 - the bottom of the range
         * @param index1 - top of the range
         */
        protected void fireListDataListenerIntervalAdded(int index0, int index1) {
            javax.swing.event.ListDataEvent e = null;
            Object[] listeners = listenerList.getListenerList ();
            for (int i = listeners.length - 2; i >= 0; i -= 2) {
                if (listeners[i]==javax.swing.event.ListDataListener.class) {
                    if (e == null)
                    e = new javax.swing.event.ListDataEvent (this, e.INTERVAL_ADDED, index0, index1);
                    ((javax.swing.event.ListDataListener)listeners[i + 1]).intervalAdded (e);
                }
            }
        }
        
        /** Notifies all registered listeners about the event.
         *
         * @param source - the source object
         * @param index0 - the bottom of the range
         * @param index1 - top of the range
         */
        protected void fireListDataListenerIntervalRemoved(int index0, int index1) {
            javax.swing.event.ListDataEvent e = null;
            Object[] listeners = listenerList.getListenerList ();
            for (int i = listeners.length - 2; i >= 0; i -= 2) {
                if (listeners[i]==javax.swing.event.ListDataListener.class) {
                    if (e == null)
                    e = new javax.swing.event.ListDataEvent (this, e.INTERVAL_REMOVED, index0, index1);
                    ((javax.swing.event.ListDataListener)listeners[i + 1]).intervalRemoved (e);
                }
            }
        }
        
        /** Notifies all registered listeners about the event.
         *
         * @param source - the source object
         * @param type - type of the event
         * @param index0 - the bottom of the range
         * @param index1 - top of the range
         */
        protected void fireListDataListenerContentsChanged(int index0, int index1) {
            javax.swing.event.ListDataEvent e = null;
            Object[] listeners = listenerList.getListenerList ();
            for (int i = listeners.length - 2; i >= 0; i -= 2) {
                if (listeners[i]==javax.swing.event.ListDataListener.class) {
                    if (e == null)
                    e = new javax.swing.event.ListDataEvent (this, e.CONTENTS_CHANGED, index0, index1);
                    ((javax.swing.event.ListDataListener)listeners[i + 1]).contentsChanged (e);
                }
            }
        }
        
        public void setSelectedItem(final java.lang.Object p1) {
            if (p1 != null) selected = getByName((String) p1);
            else selected = null;
        }
        
        public java.lang.Object getSelectedItem() {
            if (selected != null) return selected.getName();
            else return NONE;
        }
        
        /**
        * @return  
        */
        public DataObject getSelectedDataObject() {
            return selected;
        }
        
        public void setSelectedDataObject(DataObject dobj) {
            for(int i = 0; i < objs.length; i++) {
                if (objs[i].equals(dobj)) {
                    selected = dobj;
                    return;
                }
            }
            selected = null;
        }
    }
    
    static class EmptyPanel implements Panel {
        
        /** Get the component displayed in this panel.
         * @return the component
         */
        public java.awt.Component getComponent() {
            return new javax.swing.JPanel();
        }
        
        /** Help for this panel.
         * When the panel is active, this is used as the help for the wizard dialog.
         * @return the help or <code>null</code> if no help is supplied
         */
        public org.openide.util.HelpCtx getHelp() {
            return org.openide.util.HelpCtx.DEFAULT_HELP;
        }
        
        /** Provides the wizard panel with the current data--either
         * the default data or already-modified settings, if the user used the previous and/or next buttons.
         * This method can be called multiple times on one instance of <code>WizardDescriptor.Panel</code>.
         * @param settings the object representing wizard panel state, as originally supplied to {@link WizardDescriptor#WizardDescriptor(WizardDescriptor.Iterator,Object)}
         */
        public void readSettings(Object settings) {
        }
        
        /** Provides the wizard panel with the opportunity to update the
         * settings with its current customized state.
         * Rather than updating its settings with every change in the GUI, it should collect them,
         * and then only save them when requested to by this method.
         * Also, the original settings passed to {@link #readSettings} should not be modified (mutated);
         * rather, the (copy) passed in here should be mutated according to the collected changes.
         * This method can be called multiple times on one instance of <code>WizardDescriptor.Panel</code>.
         * @param settings the object representing a settings of the wizard
         */
        public void storeSettings(Object settings) {
        }
        
        /** Test whether the panel is finished and it is safe to proceed to the next one.
         * If the panel is valid, the "Next" (or "Finish") button will be enabled.
         * @return <code>true</code> if the user has entered satisfactory information
         */
        public boolean isValid() {
            return true;
        }
        
        /** Add a listener to changes of the panel's validity.
         * @param l the listener to add
         * @see #isValid
         */
        public void addChangeListener(javax.swing.event.ChangeListener l) {
        }
        
        /** Remove a listener to changes of the panel's validity.
         * @param l the listener to remove
         */
        public void removeChangeListener(javax.swing.event.ChangeListener l) {
        }
    }
    
    /** Default filter for templates.
     */
    static class TemplateFilter implements DataFilter {
        /** Should the data object be displayed or not?
         * @param obj the data object
         * @return <CODE>true</CODE> if the object should be displayed,
         *   <CODE>false</CODE> otherwise
         */
        public boolean acceptDataObject(DataObject obj) {
            return obj instanceof org.netbeans.modules.java.JavaDataObject;
        }
    }
    
    static class RMIIterator implements Iterator {
        
        WizardDescriptor.Panel[] panels;
        int current = 0;

        /** Creates new RMIIterator */
        public RMIIterator() {
            panels = new WizardDescriptor.Panel[] {
                new RMIWizardIntro(),
                new RMIWizard1_2(),
                new RMIWizard2(),
                new RMIWizard4(),
                new RMIWizard3(),
                new RMIWizardFinish()
            };
        }

        public WizardDescriptor.Panel current() {
            return panels[current];
        }

        public void previousPanel() {
            if (current == RMIWIZARD3) {
                RMIWizard2 panel = (RMIWizard2) panels[RMIWIZARD2];
                if (panel.isNewSelected()) current--;
                else current = RMIWIZARD2;
            } else current--;
        }

        public void nextPanel() {
            if (current == RMIWIZARD2) {
                RMIWizard2 panel = (RMIWizard2) panels[RMIWIZARD2];
                if (panel.isNewSelected()) current++;
                else current = RMIWIZARD3;
            } else current++;
        }

        public String name() {
            return EMPTY_STRING;
        }

        public boolean hasPrevious() {
            return (current > 0);
        }

        public boolean hasNext() {
            return (current < panels.length - 1);
        }
        
        public void addChangeListener(javax.swing.event.ChangeListener p1) {}
        public void removeChangeListener(javax.swing.event.ChangeListener p1) {}
    }
    
}
