/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting;

import java.io.*;
import java.lang.reflect.InvocationTargetException;

import org.openide.filesystems.*;
import org.openide.execution.*;
import org.openide.loaders.*;
import org.openide.*;

import org.netbeans.modules.scripting.navigation.NavigationSupport;
import org.netbeans.modules.scripting.navigation.NodeNavigation;

/**
 * Abstract implementation serving as a base class for custom ScriptTypes.
 * You have to override at least one form of eval(...) in subclasses.
 * @author  David Strupl
 */
public abstract class AbstractScriptType extends ScriptType {
    
    static final long serialVersionUID = 3955375519065452860L;
    /** Names of variables that you can refer to from the scripts.
     */
    public static final String[] NAVIGATION_SUPPORT_NAMES = { "fjscript", "Fjscript" };
    
    /** Different forms of eval can be called from each other.
     * This variable prevents indefinite loop.
     */
    private transient boolean delegatingEval = false;
    
    /** Different forms of eexec can be called from each other.
     * This variable prevents indefinite loop.
     */
    private transient boolean delegatingExec = false;
    
    /** */
    private transient boolean navigationSupportRegistered = false;
    
    public AbstractScriptType() {
        extensions = new ExtensionList();
    }
    
    /** Holds value of property extensions. */
    private ExtensionList extensions;
    
    /** Utility field used by constrained properties. */
    private java.beans.VetoableChangeSupport vetoableChangeSupport =  new java.beans.VetoableChangeSupport (this);
    
    /** Get context help for this executor type.
     * @return context help
     */
    public org.openide.util.HelpCtx getHelpCtx() {
        return new org.openide.util.HelpCtx (AbstractScriptType.class);
    }
    
    /**
     * @param fileName
     * @return
     */
    public boolean acceptFileObject(FileObject fo) {
        return extensions.isRegistered(fo);
    }
    
    /**
     * @param r
     * @param context
     * @return whatever is the result of the script. It can be null.
     */
    public Object eval(java.io.Reader r, Context context) throws InvocationTargetException {
        if (!delegatingEval) {
            delegatingEval = true;
            // anyone know better way how to copy a stream into a string? (dstrupl)
            try {
                final StringWriter w = new StringWriter();
                int c = 0;
                while ((c = r.read()) != -1) {
                    w.write(c);
                }
                Object ret = eval(w.toString(), context);
                delegatingEval = false;
                return ret;
            } catch (java.io.IOException ex) {
                throw new InvocationTargetException(ex, "IOException in eval: " + ex.getMessage());
            } finally {
                delegatingEval = false;
            }
        } else {
            throw new UnsupportedOperationException("Wrong ScriptType implementation - eval not implemented");
        }
    }
    
    /**
     * @param script
     * @param context
     * @return whatever is the result of the script. It can be null.
     */
    public Object eval(String script, Context context) throws InvocationTargetException {
        if (!delegatingEval) {
            delegatingEval = true;
            Object ret = null;
            try {
                ret = eval(new StringReader(script), context);
            } finally {
                delegatingEval = false;
            }
            return ret;
        } else {
            throw new UnsupportedOperationException("Wrong ScriptType implementation - eval not emplemented");
        }
    }
    
    /**
     * @param is
     * @param context
     * @return whatever is the result of the script. It can be null.
     */
    public void exec(java.io.Reader r, Context context) throws InvocationTargetException {
        if (!delegatingExec) {
            delegatingExec = true;
            // anyone know better way how to copy a stream into a string? (dstrupl)
            try {
                final StringWriter w = new StringWriter();
                int c = 0;
                while ((c = r.read()) != -1) {
                    w.write(c);
                }
                exec(w.toString(), context);
                delegatingEval = false;
            } catch (java.io.IOException ex) {
                throw new InvocationTargetException(ex, "IOException in eval: " + ex.getMessage());
            } finally {
                delegatingExec = false;
            }
        } else {
            eval(r, context);
        }
    }
    
    /**
     * @param script
     * @param context
     * @return whatever is the result of the script. It can be null.
     */
    public void exec(java.lang.String script, Context context) throws InvocationTargetException {
        if (!delegatingExec) {
            delegatingExec = true;
            try {
                exec(new StringReader(script), context);
                delegatingExec = false;
            } finally {
                delegatingEval = false;
            }
        } else {
            eval(script, context);
        }
    }
    
    /** Add a VetoableChangeListener to the listener list.
     * @param l The listener to add.
     */
    public final void addVetoableChangeListener(java.beans.VetoableChangeListener l) {
        vetoableChangeSupport.addVetoableChangeListener (l);
    }
    
    /** Removes a VetoableChangeListener from the listener list.
     * @param l The listener to remove.
     */
    public final void removeVetoableChangeListener(java.beans.VetoableChangeListener l) {
        vetoableChangeSupport.removeVetoableChangeListener (l);
    }
    
    /** Getter for property extensions.
     * @return Value of property extensions.
     */
    public ExtensionList getExtensions() {
        return extensions;
    }
    
    /** Setter for property extensions.
     * @param extensions New value of property extensions.
     *
     * @throws PropertyVetoException
     */
    public void setExtensions(ExtensionList extensions) throws java.beans.PropertyVetoException {
        ExtensionList oldExtensions = this.extensions;
        vetoableChangeSupport.fireVetoableChange("extensions", oldExtensions, extensions);
        this.extensions = extensions;
        firePropertyChange ("extensions", oldExtensions, extensions);
    }
    
    /** */
    private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
        stream.defaultReadObject();
    }
    
    /** Calls addVariable for each name from NAVIGATION_SUPPORT_NAMES.
     * It registers an object of type NavigationSupport.
     */
    protected void registerNavigationSupport() {
        if (navigationSupportRegistered) {
            return;
        }
        NavigationSupport nav = new NavigationSupport(
            System.in, System.out
        );
        for (int i = 0; i < NAVIGATION_SUPPORT_NAMES.length; i++) {
            addVariable(NAVIGATION_SUPPORT_NAMES[i], nav);
        }
        
        NodeNavigation nn = new NodeNavigation ();
        addVariable ("ide", nn); // NOI18N
        
        navigationSupportRegistered = true;
    }
}