/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting;

import java.util.*;
import java.io.Reader;
import java.lang.reflect.*;

import org.openide.*;

//import bsh.*;

/**
 * This is an interface to the BeanShell (http://www.beanshell.org/)
 * @author David Strupl
 */

public class BeanShellScriptType extends AbstractScriptType {
    
    static final long serialVersionUID = -3650287473944191532L;
    
    /** The compile type is object - using reflection to call methods. 
     * Should hold reference to a bsh.Interpreter instance.*/
    private transient Object interpreter;
    
    /** We call interpreter using reflection.*/
    private transient Class interpreterClass;
    
    /** 
     * When this is true we no longer log exceptions from the addVariable
     * method to the ErrorManager.
     */
    private transient boolean problemReported = false;

    /**
     * Registers bsh file extension.
     */
    public BeanShellScriptType() {
        // Register extensions to be recognized for this script type
        getExtensions().addExtension("bsh"); // NOI18N
    }
    
    /** Fills in variables interpreter and interpreterClass. */
    private Object getInterpreter() throws Exception {
        if (interpreter == null) {
            interpreterClass = Class.forName("bsh.Interpreter"); //NOI18N
            interpreter = interpreterClass.newInstance();
            // register to gain access to the IDE objects
            registerNavigationSupport();
        }
        return interpreter;
    }
    
    /** Under this name the type will be known.*/
    protected String displayName () {
        return "BeanShell"; // NO-I18N
    }
    
    /**
     * Evaluate an expression. Uses reflection to call interpreter.eval(...);
     */
    public Object eval (String s, Context context) throws InvocationTargetException {
        try {
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("eval", new Class[] { String.class }); // NOI18N
            return m.invoke(o, new Object[] { s });
        } catch (InvocationTargetException e) {
            handleException(e);
        } catch (Exception e) {
            if (problemReported) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message("Maybe the file lib/ext/bsh-1_0.jar is missing."));
            }
            throw new InvocationTargetException(e, displayName());
        }
        return null;
    }
    
    /**
     * Evaluate an expression. Uses reflection to call interpreter.eval(...);
     */
    public Object eval (java.io.Reader r, Context context) throws InvocationTargetException {
        try {
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("eval", new Class[] { java.io.Reader.class }); // NOI18N
            return m.invoke(o, new Object[] { r });
        } catch (InvocationTargetException e) {
            handleException(e);
        } catch (Exception e) {
            if (problemReported) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message("Maybe the file lib/ext/bsh-1_0.jar is missing."));
            } 
            throw new InvocationTargetException(e, displayName());
        }
        return null;
    }
    
    /**
     * Uses reflection to call interpreter.eval(...);
     */
    public void exec (java.io.Reader r, Context context) throws InvocationTargetException {
        try {
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("eval", new Class[] { java.io.Reader.class }); // NOI18N
            m.invoke(o, new Object[] { r });
            //getInterpreter().eval(r);
        } catch (InvocationTargetException e) {
            handleException(e);
        } catch (Exception e) {
            if (problemReported) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message("Maybe the file lib/ext/bsh-1_0.jar is missing."));
            } 
            throw new InvocationTargetException(e, displayName());
        }
    }
    
    /**
     * Uses reflection to call interpreter.setVariable(...).
     */
    public void addVariable(String name, Object value) {
        //getInterpreter().setVariable(name, value);
        try {
            // getInterpreter().set(name, value);
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("setVariable", new Class[] { String.class, Object.class }); // NOI18N
            m.invoke(o, new Object[] { name, value });
        } catch (Exception e) {
            if (!problemReported) {
                //TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, e);
                problemReported = true;
            }
        }
    }
    
    /** Check whether exception is a bsh.TargetError. If it is rethrows with getTarget().*/
    private void handleException(InvocationTargetException e) throws InvocationTargetException {
        Throwable e1 = e.getTargetException();

        try {
            Class targetErrorClass = Class.forName("bsh.TargetError"); // NOI18N
            if ((targetErrorClass != null) && (e1.getClass().isAssignableFrom(targetErrorClass))) {
                Method m = targetErrorClass.getMethod("getTarget", new Class[0]); // NOI18N
                Object target = m.invoke(e1, new Object[0]);
                Throwable t = (Throwable)target;
                throw new InvocationTargetException(t, t.getMessage());
            }
        } catch (Exception e2) {
            throw new InvocationTargetException(e1, e1.getMessage());
        }
    }
}
