/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import org.openide.*;

/**
 * This is an interface to the JPython (http://www.jpython.org/)
 * @author David Strupl
 */

public class JPythonScriptType extends AbstractScriptType {
    
    static final long serialVersionUID = 5438294236761951778L;
    
    /** The compile type is object - using reflection to call methods. 
     * Should hold reference to a org.python.util.PythonInterpreter instance.*/
    private transient Object interpreter;
    
    /** We call interpreter using reflection.*/
    private transient Class interpreterClass;
    
    /** 
     * When this is true we no longer log exceptions from the addVariable
     * method to the ErrorManager.
     */
    private transient boolean problemReported = false;
    
    /** 
     * Registers py file extension.
     */
    public JPythonScriptType() {
        // Register extensions to be recognized for this script type
        getExtensions().addExtension("py"); //NOI18N
    }
    
    /** Fills in variables interpreter and interpreterClass. */
    private Object getInterpreter() throws Exception {
        if (interpreter == null) {
            // set the cache dir to the user home
            System.setProperty("python.cachedir",
                System.getProperty("netbeans.user") + "/system/jpythoncachedir");
            // create an interpreter
            interpreterClass = Class.forName("org.python.util.PythonInterpreter"); //NOI18N
            interpreter = interpreterClass.newInstance();
            // register to gain access to the IDE objects
            registerNavigationSupport();
        }
        return interpreter;
    }
    
    /** Under this name the type will be known.*/
    protected String displayName () {
        return "JPython"; // NO-I18N
    }

    /** Uses reflection to call interpreter.eval(...).
     */
    public Object eval(String s, Context context) throws InvocationTargetException {
        try {
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("eval", new Class[] { String.class }); // NOI18N
            return m.invoke(o, new Object[] { s });
        } catch (Exception e) {
            if (problemReported) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message("Maybe the file lib/ext/djava.jar is missing."));
            } 
            throw new InvocationTargetException(e, e.getMessage());
        }
    }
    
    /** Uses reflection to call interpreter.exec(...).
     */
    public void exec(String s, Context context) throws InvocationTargetException {
        try {
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("exec", new Class[] { String.class }); // NOI18N
            m.invoke(o, new Object[] { s });
        } catch (Exception e) {
            if (problemReported) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message("Maybe the file lib/ext/jpython.jar is missing."));
            } 
            throw new InvocationTargetException(e, e.getMessage());
        }
    }
    
    /** Uses reflection to call interpreter.set(...).
     */
    public void addVariable(String name, Object value) {
        try {
            // getInterpreter().set(name, value);
            Object o = getInterpreter();
            Method m = interpreterClass.getMethod("set", new Class[] { String.class, Object.class }); // NOI18N
            m.invoke(o, new Object[] { name, value });
        } catch (Exception e) {
            if (!problemReported) {
                //TopManager.getDefault().getErrorManager().notify(ErrorManager.INFORMATIONAL, e);
                problemReported = true;
            }
        }
    }
}
