/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.util.ResourceBundle;
import javax.swing.text.*;
import javax.swing.JEditorPane;

import org.openide.TopManager;
import org.openide.actions.*;
import org.openide.cookies.*;
import org.openide.execution.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.util.actions.*;
import org.openide.text.EditorSupport;
import org.openide.util.Utilities;
import org.openide.util.NbBundle;

/** Object that represents one script file. Allows to edit and execute the file.
*
* @author  Jaroslav Tulach, David Strupl
*/

public final class ScriptObject extends MultiDataObject implements ExecCookie, ArgumentsCookie {
        /** bundle */
    private static ResourceBundle bundle;

        /** extended attribute for the type of executor */
//    private static final String EA_EXECUTOR = "NetBeansAttrExecutor"; // NOI18N
    /** extended attribute for attributes */
    private static final String EA_ARGUMENTS = "NetBeansAttrArguments"; // NOI18N

    // copy from JavaNode
    /** Name of property providing argument parameter list. */
    public static final String PROP_FILE_PARAMS   = "params"; // NOI18N
    /** Name of property providing a custom {@link Executor} for a file. */
    public static final String PROP_EXECUTION     = "execution"; // NOI18N

    /** property name */
    public static final String PROP_SCRIPT_TYPE = "scriptType";
  
    /** editor support to be able to get selected text */
    private EditorSupport editor;

    /** language to use for interpreting */
    private String language;

    /** Creates new ScriptObject. */
    public ScriptObject(FileObject pf, MultiFileLoader loader, String language)
    throws DataObjectExistsException {
        super(pf, loader);
    
        this.language = language;
        init();
    }
  
    private void init() {
        CookieSet cookies = getCookieSet();
    
        // editor support for the file
        editor = new EditorSupport (getPrimaryEntry ());
        cookies.add (editor);
    }
  
    /** Help context for this object.
     * @return help context
     */
    public org.openide.util.HelpCtx getHelpCtx () {
        return org.openide.util.HelpCtx.DEFAULT_HELP;
    }
  
    /** Getter for the assigned script type 
     */
    public String getScriptLanguage () {
        return language;
    }
  
    /** Setter to allow change of given script type.
     */
    public void setScriptLanguage (String t) {
        language = t;
    }
  
   /** Provides node that should represent this data object. When a node for representation
     * in a parent is requested by a call to getNode (parent) it is the exact copy of this node
     * with only parent changed. This implementation creates instance
     * <CODE>DataNode</CODE>.
     * <P>
     * This method is called only once.
     *
     * @return the node representation for this data object
     * @see DataNode
     */
    protected Node createNodeDelegate () {
        Node node = new ScriptNode (this, Children.LEAF);
        return node;
    }

    /** Starts the execution
     */
    public void start () {
        try {
            // save first
            SaveCookie save = (SaveCookie)getCookie (SaveCookie.class);
            if (save != null) {
                save.save ();
            }
            ScriptSupport sup = new ScriptSupport(getPrimaryEntry());
            sup.start();
        } catch (java.io.IOException ex) {
            // notify the exeception
            org.openide.TopManager.getDefault ().notifyException (ex);
        }
    }
    
   /* Sets execution arguments for the primary entry.
    * @param args array of arguments
    * @exception IOException if arguments cannot be set
    */
    public void setArguments (String[] args) throws IOException {
        getPrimaryEntry().getFile ().setAttribute (EA_ARGUMENTS, args);
    }

      /* Getter for arguments associated with given file.
    * @return the arguments or empty array if no arguments associated
    */
    public String[] getArguments () {
        try {
            String[] args = (String[])getPrimaryEntry().getFile ().getAttribute (EA_ARGUMENTS);
            if (args != null) {
                return args;
            }
        } catch (Exception ex) {
            // null pointer or IOException
        }
        return new String[0];
    }
    /** Helper method that creates default properties for execution of
    * a given support.
    * Includes properties to set the executor and arguments.
    *
    * @param set sheet set to add properties to
    */
    public void addProperties (Sheet.Set set) {
        set.put(new PropertySupport.ReadWrite (
                    PROP_FILE_PARAMS,
                    String.class,
                    getString("PROP_fileParams"),
                    getString("HINT_fileParams")
                ) {
                    public Object getValue() {
                        String[] args = getArguments ();
                        StringBuffer b = new StringBuffer(50);
                        for (int i = 0; i < args.length; i++) {
                            b.append(args[i]).append(' ');
                        }
                        return b.toString();
                    }
                    public void setValue (Object val) throws InvocationTargetException {
                        if (val instanceof String) {
                            try {
                                setArguments(Utilities.parseParameters((String)val));
                            } catch(IOException e) {
                                throw new InvocationTargetException (e);
                            }
                        }
                        else {
                            throw new IllegalArgumentException();
                        }
                    }

                    public boolean supportsDefaultValue () {
                        return true;
                    }

                    public void restoreDefaultValue () throws InvocationTargetException {
                        try {
                            setArguments(null);
                        } catch(IOException e) {
                            throw new InvocationTargetException (e);
                        }
                    }

                }
               );
    }
     /** @return a localized String */
    static String getString(String s) {
        if (bundle == null) {
            bundle = NbBundle.getBundle(ScriptSupport.class);
        }
        return bundle.getString(s);
    }
 
}