/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting;

import java.io.*;
import java.util.*;
import java.lang.reflect.InvocationTargetException;
import javax.swing.*;
import javax.swing.text.Utilities;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import org.openide.*;
import org.openide.execution.*;
import org.openide.filesystems.*;
import org.openide.windows.InputOutput;
import org.openide.util.NbBundle;

/**
* @author jtulach, dstrupl
*/
public class ScriptUtils extends Object {

    public static String LINE_CONTINUATION_CHAR = "\\"; // NO-I18N

    
    /** Performs execution of a selected block or current line
     * in the provided editor pane.
     *
     * @param pane editor pane
     * @param type the script type to use
     */
    public static void execute (JEditorPane pane, String lang) {
        evaluate0 (pane, lang, false);
    }
  
    /** Performs evaluation of a selected block or current line 
     * in provided editor pane and inserts the result back into
     * the editor.
     *
     * @param pane editor pane to evaluate
     * @param type the script type to use
     */
    public static void evaluate (final JEditorPane pane, final String lang) {
        Thread t = new Thread(new Runnable() {
            public void run() {
                evaluate0 (pane, lang, true);
            }
        });
        t.start();
    }

    /** Evaluates or executes the text in the pane 
     */
    private static void evaluate0 (final JEditorPane pane, String lang, boolean insert) {
        if (lang == null) {
            lang = ScriptType.getDefault().getName();
        }
        final Object []obj = new Object[1];
    
        try {
            final int pos = pane.getCaretPosition ();
            String sel = pane.getSelectedText ();

            if (sel == null) {
                // get current line
                int beg = Utilities.getRowStart (pane, pos);
                int end = Utilities.getRowEnd (pane, pos);
                sel = pane.getText (beg, end - beg);
                insert = false;
            }
            final String textToEval = sel;
            final ScriptType s = getScriptType(lang);
            final InputOutput io = TopManager.getDefault().getIO(s.getName(), false);
            // execute only selected text
            if (insert) {
                ExecutorTask task = TopManager.getDefault().getExecutionEngine().execute("scripting", new Runnable() {
                    public void run() {
                        try {
                            obj[0] = s.eval(textToEval);
                        } catch (InvocationTargetException ite) {
                            Throwable t = ite.getTargetException();
                            if (t != null) {
                                t.printStackTrace(io.getOut());
                            } else {
                                ite.printStackTrace(io.getOut());
                            }
                        }
                    }
                }, io);
                task.result();
                final String res = obj[0] == null ? "null" : obj[0].toString ();
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        Document doc = pane.getDocument ();
                        try {
                            doc.insertString (pos, res, null);
                        } catch (BadLocationException ex) {
                            TopManager.getDefault().getErrorManager ().notify (ex);
                        }
                        pane.select (pos, pos + res.length ());
                    }
                });
            } else {
                // execute
                ExecutorTask task = TopManager.getDefault().getExecutionEngine().execute("scripting", new Runnable() {
                    public void run() {
                        try {
                            s.exec(textToEval);
                        } catch (InvocationTargetException ite) {
                            Throwable t = ite.getTargetException();
                            if (t != null) {
                                t.printStackTrace(io.getOut());
                            } else {
                                ite.printStackTrace(io.getOut());
                            }
                        }
                    }
                }, io);
            }
        } catch (Exception ex) {
            TopManager.getDefault().getErrorManager ().notify (ex);
        }
    } 

    public static String getLangFromFilename(FileObject fo) {
        String lang = null;
        Enumeration e = ScriptType.scriptTypes();
        while (e.hasMoreElements()) {
            ScriptType s = (ScriptType)e.nextElement();
            if (s.acceptFileObject(fo)) {
                lang = s.getName();
            } else {
            }
        }
        return lang;
    }

    public static String[] getLanguages() {
        List list = new ArrayList();
        Enumeration e = ScriptType.scriptTypes();
        while (e.hasMoreElements()) {
            ScriptType s = (ScriptType)e.nextElement();
            list.add(s.getName());
        }
        return (String[])list.toArray(new String[0]);
    }
    
   /** Get the script type for a given language.
    * @param entry entry to obtain the scirpt type for
    * @return script type associated with the file, or null if the default should be used
    */
    public static ScriptType getScriptType (String lang) {
        Enumeration e = ScriptType.scriptTypes();
        while (e.hasMoreElements()) {
            ScriptType s = (ScriptType)e.nextElement();
            if (s.getName().equals(lang)) {
                return s;
            }
        }
        return null;
    }

    /** Tries to read parameters from the input stream. The
     * parameters have to be typed as key=value pairs each
     * on separate line. It reads until all keys in requiredKeys
     * are filled in. Returns Map containing all the pairs. It
     * uses out to communicate with the user prompting him/her
     * for additional input.
     */
    public static Map readParameters(InputStream in, PrintStream out, 
        String []requiredKeys, String []optionalKeys) throws IOException {
        out.print("Please enter following keys (type key=value each on separate line): \n\r");
        out.println("\n\r");
        out.flush();
        Properties result = new Properties();
        boolean finished = false;
        while (!finished) {
            Properties tmp = new Properties();
            String line = null;
            String lines = "";
            BufferedReader br = new BufferedReader(new InputStreamReader(in));
            do {
                line = readLine(br);
                lines += line;
            } while (!"".equals(line));
            tmp.load(new StringBufferInputStream(lines));
            for (Enumeration keys = tmp.keys(); keys.hasMoreElements(); ) {
                Object key = keys.nextElement();
                Object value = tmp.get(key);
                result.put(key, value);
            }
            finished = true;
            for (int i = 0; i < requiredKeys.length; i++) {
                if (!result.containsKey(requiredKeys[i])) {
                    finished = false;
                    out.print("Key " + requiredKeys[i] + " not suplied\n\r");
                    out.flush();
                }
            }
            out.flush();
        }
        return result;
    }

   /** Reads one line or more from the given reader. To read more
     * lines than one the lines have to be termineated with
     * LINE_CONTINUATION_CHAR.
     */
    public static String readLine(BufferedReader br) throws IOException {
        String line = ""; // NO-I18N
        String tmpLine = null;
        do {
            tmpLine = br.readLine();
            if (tmpLine != null) {
                line += tmpLine;
                if (line.endsWith(LINE_CONTINUATION_CHAR)) {
                    line = line.substring(0, line.length()-1);
                }
            }
        } while ((tmpLine != null)&&(tmpLine.endsWith(LINE_CONTINUATION_CHAR)));
        return line;
    }

    /** Utitility method for removing 'bad' characters from the
     * given string.
     */
    public static String convertLine(String line) {
        StringBuffer sb = new StringBuffer(line);
        for (int i = 0; i < sb.length(); i++) {
            // in case of \b delete two characters
            while ((i < sb.length()) && (sb.charAt(i) == '\b')) {
                if (i > 0) {
                    sb.deleteCharAt(i - 1);
                    sb.deleteCharAt(i - 1);
                    i--;
                } else {
                    if (i == 0) {
                        sb.deleteCharAt(0);
                    }
                }
            }
            // for non printables - delete them
            while ((i < sb.length()) && (Character.isISOControl(sb.charAt(i)))) {
                sb.deleteCharAt(i);
            }
            // foreign alphabets --- sorry
            while ((i < sb.length()) && (sb.charAt(i)>127)) {
                sb.deleteCharAt(i);
            }
        }
        return sb.toString().trim();
    }
    
    /** Add mime types for known languages to show correct editor.
     * For now it registers DynamicJava and BeanShell.
     */
    public static void registerMimeTypes() {
        org.openide.filesystems.FileUtil.setMIMEType("djava", "text/x-java");
        org.openide.filesystems.FileUtil.setMIMEType("bsh", "text/x-java");
    }    
}