/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting.navigation;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.beans.PropertyVetoException;
import java.lang.reflect.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Stack;

import javax.swing.*;

import org.openide.TopManager;
import org.openide.actions.ActionManager;
import org.openide.actions.ToolsAction;
import org.openide.execution.NbClassPath;
import org.openide.explorer.ExplorerManager;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.Node;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;

// Sorry - I cannot avoid this (dstrupl)
import org.netbeans.core.*;

/** 
 * A  class for driving command-line access to the IDE.
 * It has built-in support for browsing mounted file systems, via the findObject,
 * findChild, and findParent methods.
 * There is also support for loading "command classes" for objects found 
 * while browsing the file systems.
 *
 * @author Roger Blumer
 */
public class NavigationSupport
{
	public  Object	 curObj		= null;
	public DataObject curDataObj	= null;
	public Properties forteProps	= new Properties(System.getProperties());
	public final String moduleName = "fjscript";	// NOI18N	

	/**
	* This ExplorerManager is the context for selected nodes in Fjscript.
	* This is used by setCurObject to activate the currently selected object.
	*/
	public ExplorerManager	explorerMgr = null;	// used to select Nodes

	private boolean	exited = false; 
	private String	commandSuffix = "Command";	// NOI18N
	private Stack	stack = null;

	/**
	* Success status for commands.
	*/
	public static final int				CMD_OK			= 0;

	/**
	* Failure status for commands.
	*/
	public static final int				CMD_FAIL		= 1;


	/**
	 * CONFIGURATION flag for printing stacktraces for exceptions. Value is 1.
	 */
	public static final int CFG_PRINT_STACKTRACE = 1;// print exception traces

	/**
	 * CONFIGURATION flag for avoiding the auto-start of TopManager. Value is 2.
	 */
	public static final int CFG_NO_AUTOLOAD = 2;	// don't start TopManager



	// Command classes and test scripts can use these to get input
	// and show output.  They may be redirected to files by interpreter/wrapper.
	// 
	/**
	 * Input stream for the interpreter/wrapper to Fjscript.
	 */
	public InputStream inStrm	= System.in;
	/**
	 * Output stream for giving messages to the user/output file.
	 */
	public PrintStream outStrm   = System.out;

	/**
	 * used to create objects
	 */
	private Dialog                 templateDialog = null;
	private NonGuiTemplateWizard   templateWizard = null;

        public NavigationSupport(InputStream in,PrintStream out) {
            inStrm = in;
            outStrm = out;
        }
        
	/**
	 * Exits the IDE (if it was loaded), then call System.exit().
	 *
         * @param status Exit status for the program.
	 */
	public synchronized void exit(int status) {
		boolean wasExited = exited;
		exited = true;

		if (!wasExited && TopManager.isInitialized())
					TopManager.getDefault().exit();
		System.exit(status);
	}


	void checkCurObj() throws IllegalArgumentException {
		if (curObj == null)
			throw new IllegalArgumentException(getString("MSG_NullObject"));
	}
	void checkCurDataObj() throws IllegalArgumentException {
		if (curDataObj == null)
			throw new IllegalArgumentException(getString("MSG_NullDataObject"));
	}
	
	public void setTopManager() {
		// set the default TopManager implementation to NonGui version
		System.getProperties().put (
			"org.openide.TopManager", // NOI18N
			"org.netbeans.core.NonGui" // NOI18N
			);
	}

	public Class loadClass(String className) {
		return loadClass(className, true);
	}

	public Class loadClass(String className, boolean showException) {
		Class  cls = null;
		try {
			//
			// use the current class loader here.  In the case that modules
			// were loaded from the Repository/mounted fileSystem,
			// the system class loader won't be able to find them.
			// 
			ClassLoader curLoader;
			if (TopManager.isInitialized()) {
				curLoader = TopManager.getDefault().currentClassLoader();
				if (curLoader != null) {
					cls = curLoader.loadClass(className);
				}
			}
			else {
				cls = Class.forName(className);
			}
		}
		catch (ClassNotFoundException e) {
			if (showException)
				printException("NavigationSupport.loadClass", e);	// NOI18N
		}
		return(cls);
	}

	public Object createObject(String className) {
		return createObject(className, true);
	}

	public Object createObject(String className, boolean showException) {
		Object obj = null;
		Class  cls;

		try {
			cls = loadClass(className, showException);
			if (cls != null) {
				obj = cls.newInstance();
			}
		}
		catch (IllegalAccessException iae) {
			printException("NavigationSupport.createObject", iae);	// NOI18N
		}
		catch (InstantiationException ie) {
			printException("NavigationSupport.createObject", ie);	// NOI18N
		}
		
		return obj;
	}
	
	public int listMethods(Object obj) {
	 	if (obj == null)
			return CMD_FAIL;

		Class cls = obj.getClass();
		return listMethods(cls, false);
	}

	public int listMethods(Class cls) {
		return listMethods(cls, false);
	}

	public int listMethods(Class c, boolean details) {
		Method 		ms[];
		if (c == null) {
			print("NavigationSupport.listMethods: ");	// NOI18N
			println(getString("MSG_NullClass"));
			return CMD_FAIL;
		}
		try
		{
			ms = c.getMethods();

			if (details) {
				for (int i = 0; i < ms.length; i++)
					println(ms[i].toString());
			}
			else {
				for (int i = 0; i < ms.length; i++)
					println(ms[i].getName());
			}
			
		}
		catch(Exception e)
		{
			printException("NavigationSupport.listMethods", e);	// NOI18N
			return CMD_FAIL;
		}
		return CMD_OK;
	}

	public int listMethods(String clsName, boolean details) {
	 	if (clsName == null)
			return CMD_FAIL;

		Class cls = loadClass(clsName);
		return listMethods(cls, details);
	}

	public int listMethods(String clsName) {
		return listMethods(clsName, false);
	}

	public int listConstructors(String clsName) {
	 	if (clsName == null)
			return CMD_FAIL;

		Class cls = loadClass(clsName);
		return listConstructors(cls);
	}

	public int listConstructors(Class c) {
		Constructor cs[];

		try
		{
			cs = c.getConstructors();
			for (int i = 0; i < cs.length; i++)
				println(cs[i].toString());
		}
		catch(Exception e)
		{
			printException("NavigationSupport.listConstructors", e);	// NOI18N
			return CMD_FAIL;
		}
		return CMD_OK;
	}

	public int listFileSystems() {
		return listFileSystems(null);
	}
	
	public int listFileSystems(String option1) {
		return listFileSystems(option1, null);
	}

	public int listFileSystems(String option1, String option2) {
		boolean showHidden = false;
		boolean showSystem = false;
		
		if (option1 != null)
		{
			if (option1.equals("hidden"))	// NOI18N
				showHidden = true;
			else if (option1.equals("system"))	// NOI18N
				showSystem = true;
			if (option2 != null) {
				if (option2.equals("hidden"))	// NOI18N
					showHidden = true;
				else if (option2.equals("system"))	// NOI18N
					showSystem = true;
			}
		}

		Enumeration ee = TopManager.getDefault().getRepository().getFileSystems();
		FileSystem fs;
		while (ee.hasMoreElements()) {
			fs = (FileSystem) ee.nextElement();
			if (fs.isHidden() && !showHidden)
				continue;
			print(fs.getDisplayName());
			if (showSystem) {
				print("\t\t");				// NOI18N
				print(getString("MSG_SystemName"));
				print(": ");				// NOI18N
				println(fs.getSystemName());
			}
			else {
				println();
			}
		}
		return CMD_OK;
	}

	// for use as "type" in mountFileSystem method
	// 
	public final static int DIRECTORY	= 1;
	public final static int JAR 		= 2;

	public FileSystem mountFileSystem (String fileName, int type) {
		FileSystem fs;
		fileName = expandVars(fileName);

		try {
			if (type == DIRECTORY) {
				LocalFileSystem lfs=new LocalFileSystem();
				lfs.setRootDirectory(new File(fileName));
				fs = lfs;
			}
			else if (type == JAR) {
				JarFileSystem jfs=new JarFileSystem();
				jfs.setJarFile(new File(fileName));
				fs = jfs;
			}
			else {
				println(getString("MSG_InvalidType",
										  new Integer(type),
										  new Integer(DIRECTORY),
										  new Integer(JAR)));
				return null;
			}
		  
		}
		catch (IOException ioe) {
			printException("NavigationSupport.mountDirectory", ioe);	// NOI18N
			return null;
		}
		catch (java.beans.PropertyVetoException pe){
			printException("NavigationSupport.mountFileSystem", pe);	// NOI18N
			return null;
		}

		TopManager.getDefault().getRepository().addFileSystem(fs);
		return fs;
	}

	public FileSystem mountDir(String fileName) {
		return mountDirectory(fileName);
	}

	public FileSystem mountDirectory(String fileName) {
		return mountFileSystem(fileName, DIRECTORY);
	}

	public FileSystem mountJar(String fileName) {
		return mountFileSystem(fileName, JAR);
	}

	public int unmountJar (String fileName) {
		return unmountDirectory(fileName);
	}

	public int unmountDir (String fileName) {
		return unmountDirectory(fileName);
	}

	public int unmountDirectory (String fileName) {
		fileName = expandVars(fileName);

		FileSystem fs = null;
		File theFile = new File(fileName);
		Repository repo = TopManager.getDefault().getRepository();

		Enumeration all=repo.getFileSystems();
		while (all.hasMoreElements()) {
			Object o=all.nextElement();
			if (o instanceof LocalFileSystem) {
				LocalFileSystem lfs=(LocalFileSystem) o;
				if (lfs.getRootDirectory().equals(theFile)) {
					fs=lfs;
					break;
				}
			}
			else if (o instanceof JarFileSystem) {
                JarFileSystem jfs=(JarFileSystem) o;
                File jfsFile = jfs.getJarFile();
                boolean equal = jfsFile.equals(theFile);
                //
                // check for canonical file, too, to handle case
                // where mounted jar is a symbolic link (b4363096)
                //
                if (!equal) {
                    try {
                    equal = jfsFile.equals(theFile.getCanonicalFile());
                    }
                    catch (IOException io) {
                    //ignore
                    }
                }
                if (equal) {
                    fs=jfs;
                    break;
                }
			}
			//else  need to handle other file systems (CVS, Teamware, etc)
			//--why isn't there a generic equals(fileName) for FileSystem?
			//--could use equals(displayName) though that may not be unique
			//--could use equals(systemName); that's probably unique.
		}

		if (fs == null)
		{
			println(getString("MSG_NoFileSystem", fileName));
			return CMD_FAIL;
		}
	  
		repo.removeFileSystem(fs);
		return CMD_OK;
	}

    public int unmountAll() {
        Repository repo = TopManager.getDefault().getRepository();
        Enumeration all=repo.getFileSystems();
        while (all.hasMoreElements()) {
            FileSystem fs = (FileSystem)all.nextElement();
            if (!fs.isHidden() && !fs.isDefault())
                repo.removeFileSystem(fs);
        }
        return CMD_OK;
    }

	public Object loadCommandClass (Object dataObj) {

		if (dataObj == null)
			return null;

		String clsName;
		Object obj = null;
		Class  cls = null;
		Class  sup = dataObj.getClass();

		clsName = dataObj.getClass().getName() + commandSuffix;

		while (cls == null) {
			cls = loadClass(clsName, false);
			if (cls == null) {
				sup = sup.getSuperclass();
				if (sup == null)
					return null;
				clsName = sup.getName() + commandSuffix;
			}
		}

		if (CommandHelper.class.isAssignableFrom(cls)) {
			obj = createObject(clsName.toString(), false);
			if (obj != null)
				((CommandHelper)obj).setObject(dataObj);
		}
		else {
			// try to find a constructor that takes a DataObject
			//
			Constructor cs;
			try {
				cs = cls.getConstructor(new Class[] { Object.class });
				if (cs == null)
					obj = cls.newInstance();
				else
					obj = cs.newInstance(new Object[] { dataObj });
			}
			catch (NoSuchMethodException nsm) {
			}
			catch (IllegalAccessException iae) {
			}
			catch (InstantiationException ie) {
			}
			catch (InvocationTargetException ite) {
			}
		}
		
		return obj;
	}
	
	// 

    /** 
     * Sets the current Fjscript context to the give object.
	 * This includes setting up the curObj & curDataObj attributes,
	 * loading the command class for curObj (if one exists),
	 * and setting curObj as the selectedNodes in the explorerManager
	 * (if the object is a node or a DataObject).
     *
     * @param obj The object that will become the current object.
     */
	public Object setCurObject (Object obj) {

		curObj = obj;

		if (obj instanceof DataObject)
			curDataObj = (DataObject)(obj);
		else
			curDataObj = null;

		Object cmdInstance = loadCommandClass(curObj);
		if (cmdInstance != null)
			curObj = cmdInstance;

		setInterpObject();

		// set up the current object as the selected node,
		// if it is a node or dataobject.
		//
		if (explorerMgr != null) {
			if (obj instanceof Node)
				setSelectedNodes((Node) obj);
			else if (curDataObj != null)
				setSelectedNodes(curDataObj.getNodeDelegate());
		}
				
		return curObj;
	}
        
            private static final class EMRun implements Runnable {
                private static EMRun INSTANCE = new EMRun ();
                private static ExplorerManager em;
                
                
                /** Implements the runnable interface. Is called twice. Once 
                 * from the Thread (this) and second time from the
                 * attachExplorer (this) method.
                 * For the first call em == null for the second em != null.
                 */
                public synchronized void run () {
                    if (em == null) {
                        // the code executed when called from Thread (this)
                        em = new ExplorerManager ();
                        notify ();
                        
                        NbTopManager.get ().attachExplorer(this, em);
                        return;
                    }
                    
                    // the code executed from attachExplorer (this)
                    try {
                        // waits until somebody calls stop
                        wait ();
                    } catch (InterruptedException ex) {
                    }
                }

                public static void stop () {
                    synchronized (INSTANCE) {
                        // clears the manager and notifies the waiting 
                        // thread if any
                        em = null;
                        INSTANCE.notify ();
                    }
                }
                
                public static ExplorerManager init () {
                    synchronized (INSTANCE) {
                        while (em == null) {
                            // called for the first time => create 
                            // the explorer manager (in run () method)
                            new Thread (EMRun.INSTANCE).start ();
                            try {
                                // wait till the em is filled
                                INSTANCE.wait ();
                            } catch (InterruptedException ex) {
                            }
                        }
                        // return the explorer manager
                        return em;
                    }
                }
            }
	
	/**
	 * Makes the nodes the currently selected nodes in the explorerMgr.
	 * The first node in the array (or its root-parent) becomes the
	 * rootContext of the explorer manager.
	 *
	 * @param value the node to be selected.  If null, the selection is cleared.
	 */
	public void setSelectedNodes (Node[] arr) {
            /** Explorer nodes hack.
             */
            
            if (arr == null) {
                EMRun.stop ();
                return;
            } 
            
            ExplorerManager em = EMRun.init ();
            
            
            
		if (arr == null)
			arr = new Node[] {};	// set to empty array, to clear selection

		try {
			// attempt to set the root context based on the first node
			//
			if (arr.length >= 1) {
				Node root = arr[0];
				if (root != null) {
					while (root.getParentNode() != null)
						root = root.getParentNode();
					explorerMgr.setRootContext(root);
				}
			}

			explorerMgr.setSelectedNodes(arr);
		}
		catch (PropertyVetoException e) {
			printException("NavigationSupport.setSelectedNodes", e);	// NOI18N
		}
		catch (IllegalArgumentException e) {
			printException("NavigationSupport.setSelectedNodes", e);	// NOI18N
		}
	}

	/**
	 * Makes the node the currently selected node in the explorerMgr.
	 * The node or its root-parent becomes the rootContext of the
	 * explorer manager.
	 *
	 * @param value the node to be selected
	 */
	public void setSelectedNodes (Node value) {
		if (value == null)
			setSelectedNodes((Node[]) null);

		setSelectedNodes(new Node[] { value } );
	}

	public  Object interp = null; 
	private Method meth = null;

	// if the interpreter object is set, and has a method called setVariable,
	// then set curObj & curDataObj variables in the interpreter.
	// 
	private void setInterpObject () {
		if (interp == null)
			return;
		if (meth == null) {
			try {
				meth = interp.getClass().getMethod("setVariable",  // NOI18N
												   new Class[] { String.class,
																 Object.class });
			}
			catch (NoSuchMethodException e) {
				// it's OK if method doesn't exist
				meth = null;
				interp = null;	// so we don't check again
			}
		}
		if (meth == null)
			return;

		try {
			meth.invoke(interp,
						new Object[] { "curObj", curObj }); // NOI18N
			meth.invoke(interp, new Object[] { "curDataObj",		 // NOI18N
											   curDataObj });
		}
		catch (InvocationTargetException e) {
			// it's OK if call fails
		}
		catch (IllegalAccessException e) {
			// it's OK if call fails
		}
	}

	public Object findObject (String fileName) {
		DataObject	dataObj = getObject(fileName);

		if (dataObj != null) {
			return setCurObject(dataObj);
		} else {
			return null;
		}
	}

	public DataObject getObject (String fileName) {
		fileName = expandVars(fileName);

		try {
			Repository repo = TopManager.getDefault().getRepository();
			FileObject file = repo.findResource(fileName);

			if (file == null) {
				println(getString("MSG_FileNotFound", file));
				return null;
			}

			DataObject dataObj = DataObject.find(file);
			return dataObj;
		}
		catch(DataObjectNotFoundException donf) {
			printException("NavigationSupport.findObject", donf);	// NOI18N
		}
		return null;
	}

	public Object findParent () {
		if (curDataObj == null) {
			println(getString("MSG_NullObject"));
			return null;
		}
		else if (!(curDataObj instanceof DataObject)) {
			println(getString("MSG_NotDataObject"));
			return null;
		}
		DataFolder parent = ((DataObject) curDataObj).getFolder();
		if (parent != null)
			setCurObject(parent);
		return(parent);
	}
	
	public Object findChild (String childName) {
		DataObject child;

		child = getChild(curDataObj, childName);
		if (child != null) {
			return setCurObject(child);
		} else {
			return null;
		}
	}

	public DataObject getChild (DataObject parent, String childName) {

		if (parent == null) {
			println(getString("MSG_NullDataObject"));
			return null;
		}
		else if (!(parent instanceof DataFolder)) {
			println(getString("MSG_NotDataFolder"));
			return null;
		}

		DataObject curChild;
		DataFolder folder = (DataFolder) parent;
		Enumeration ee = folder.children();

		while (ee.hasMoreElements()) {
			curChild = (DataObject) ee.nextElement();
			if (childName.equals(curChild.getName())) {
				return curChild;
			}
		}
		println(getString("MSG_ChildNotFound", 
								  childName, folder.getName()));
		return null;
	}

	public int listChildren () {
		if (curDataObj == null) {
			println(getString("MSG_NullObject"));
			return CMD_FAIL;
		}
		else if (!(curDataObj instanceof DataFolder)) {
			println(getString("MSG_NotDataFolder"));
			return CMD_FAIL;
		}

		DataObject arr[] = ((DataFolder) curDataObj).getChildren();
		DataObject a;
		String className;
		for (int i =0; i < arr.length; i++)
		{
			if (arr[i] == null)
				continue;
			a = arr[i];
			className = a.getClass().getName();
			className = className.substring(className.lastIndexOf('.') + 1);
			println(a.getName() + "\t" + className);
		}
		return CMD_OK;
	}
	
	void printActions (DataFolder df, int level)  {
		DataObject curChild;
		Enumeration ee = df.children();
		while (ee.hasMoreElements()) {
			curChild = (DataObject) ee.nextElement();
			for (int i = 0; i < level; i++)
				print("    ");
			print(curChild.getName());
			//println("\t"+ curChild.getClass().getName());

			if (curChild instanceof InstanceDataObject) {
				InstanceDataObject inst = (InstanceDataObject) curChild;
				try {
					println("\t"+ inst.instanceCreate().getClass().getName());
				}
				catch (Exception e) {
				}
			}
			else {
				println();
				if (curChild instanceof DataFolder)
					printActions((DataFolder)curChild, level+1);
			}
		}
	}

    /**
     * Run the specified action on the current data object (curDataObj).
     * Loads the class <actionName> and calls its actionPerformed method
     * on the current data object. 
     *
     * @param actionName full class name of the action to run
     */
    public int doAction (String actionName)  {
        return doAction(actionName, 0);
    }
        
    /**
     * A variant of doAction for passing modifiers such as ActionEvent.CTRL_MASK.
     * Run the specified action on the current data object (curDataObj).
     * Loads the class <actionName> and calls its actionPerformed method
     * on the current data object. 
     *
     * @param actionName full class name of the action to run
     * @param modifiers the modifier keys held down during this action
     */
    public int doAction (String actionName, int modifiers)  {
        return doAction(actionName, modifiers, 0, "");
    }
        
    /**
     * Run the specified action on the current data object (curDataObj).
     * Loads the class <actionName> and calls its actionPerformed method
     * on the current data object. 
     * A variant of doAction with the full set of options for creating the ActionEvent passed to the action.
     *
     * @param actionName full class name of the action to run
     * @param modifiers the modifier keys held down during this action
     * @param id an integer that identifies the event
     * @param command a string that may specify a command (possibly one of several) associated with the event
     *
     * @see java.awt.event.ActionEvent
     */
    public int doAction (String actionName, int modifiers, int id, String command)  {
        String origName = actionName;

		checkCurDataObj();

		Class cls = loadClass(actionName, false);
		if (cls == null) {
		    actionName = "org.openide.actions." + actionName;
		    cls = loadClass(actionName, false);
		}
		if (cls == null) {
			println(getString("MSG_NoClass", origName));
			return CMD_FAIL;
		}
	  
		Object obj = createObject(actionName);
		if (obj == null) {
			println(getString("MSG_CouldNotCreate", actionName));
			return CMD_FAIL;
		}
		else if (!(obj instanceof SystemAction)) {
			println(getString("MSG_NotSubClass", actionName));
			return CMD_FAIL;
		}
		
		Node node = curDataObj.getNodeDelegate();
		((SystemAction) obj).actionPerformed(new ActionEvent(node, id, command, modifiers)); // NOI18N

		return CMD_OK;
	}

	public int listActions ()  {
		DataFolder df = TopManager.getDefault().getPlaces().folders().actions();

		printActions(df, 0);
		return CMD_OK;
	}

	public int listActionsFromLoader ()  {
		checkCurDataObj();
		return listActionsFromLoader(curDataObj.getLoader());
	}

	public int listActionsFromLoader (DataLoader loader)  {

		if (loader == null) {
			println ("loader is null");
			return CMD_OK;
		}

		SystemAction[] actions = loader.getActions();

		if (actions == null) {
			println ("no actions");
			return CMD_OK;
		}

		ArrayList subMenus = new ArrayList(5);
		  
		println ("# actions: " + actions.length);
		for (int i = 0; i < actions.length; i++) {
			SystemAction a = actions[i];
			if (a != null) {
				println ("action " + i + " : " + 
								 a.getName () + " class name "
								 + a.getClass().getName()
								 + (a.isEnabled() ? "" : "  DISABLED"));
			} else {
				println ("-----------------");
			}
			if (a instanceof ToolsAction) {
				subMenus.add(a);
			}
		}
		for (int i = 0; i < subMenus.size(); i++) {
			println ("--------subMenu " + i + " : ");
		  
			ActionManager am = TopManager.getDefault ().getActionManager ();
			SystemAction[] subActions = am.getContextActions ();
			for (i = 0; i < subActions.length; i++) {
				SystemAction a = subActions[i];
				if (a == null)
					continue;
				if (a.isEnabled ()) {
					println ("\taction " + i + " : " + 
									 a.getName () + " class name "
									 + a.getClass().getName());
				
					// do submenus here again--but the code above doesn't seem
					// specific to the current menu!!  so how do we handle it? 
				}
			}
		}
		return CMD_OK;
	}

	public int showCurObj () {
		if (curObj == null)
			println("curObject: null");	// NOI18N
		else {
			println("curObject: " + curObj.getClass().getName()); // NOI18N
		}
		if (curDataObj == null)
			println("curDataObject: null");	// NOI18N
		else {
			print("curDataObject: " + curDataObj.getName()); // NOI18N
			FileObject fo = curDataObj.getPrimaryFile();
			if (fo != null) {
				print("\t " + fo.getName());	// NOI18N
				String ext = fo.getExt();
				if ((ext == null) || (ext.length() == 0))
					println();
				else
					println("." + fo.getExt());	// NOI18N
			}
		}
		return CMD_OK;
	}

	public int help ()  {
		return help(false, false);
	}
	
	public int help (boolean details)  {
		return help(details, false);
	}
	
	public int help (boolean details, boolean all)  {
		
		println();
		if (curObj != null) {
			String clsName = curObj.getClass().getName();
			String name = clsName.substring(clsName.lastIndexOf('.')+1);
            String helpStr;

            if (curObj instanceof CommandHelper
                && (helpStr = ((CommandHelper)curObj).help()) != "") {
				println(getString("MSG_CmdClass", name));
				println(getString("MSG_UndClass"));
				println(helpStr);
			}
			else {
				println(getString("MSG_MethodsOn", name));
				listMethods(curObj.getClass(), details);
			}
		}

		if (all) {
			println();
			println(getString("MSG_MethodsOnFJ"));
			listMethods(NavigationSupport.class, details);
		}
		return CMD_OK;
	}

    /* installing direct from JARs now --jglick
	File getManifestFile (String fileName) throws IOException {
		//
		//  this was taken from InstalledModules.execute()
		//	(package src_internal/com/netbeans/developer/modules/apisupport)
		//	
		fileName = expandVars(fileName);

		String name = fileName.replace ('.', '/');
		Repository repo = TopManager.getDefault ().getRepository ();

		String jarName = name + ".jar";		// NOI18N
		FileObject jarfo = repo.findResource (jarName);
		FileObject mffo	 = repo.findResource (name + ".txt");	// NOI18N
		if (mffo == null)
			mffo = repo.findResource (name + ".mf");	// NOI18N
		File f;

/*******  will not do JAR load yet  -rjb
 *******  (else we have to add jarpackager.jar to base classpath)
		if (jarfo != null) {
			DataObject dobj = DataObject.find (jarfo);
			if (dobj == null)
				throw new IOException("Couldn't find DataObject for " + jarName);
			if (!(dobj instanceof JarDataObject))
			{
				throw new IOException ("Invalid jar: " + jarName
									   + ".	 Object returned was a "
									   + dobj.getClass().getName() + ", not a JarDataObject.");
			}
	  
			JarDataObject jdo = (JarDataObject) dobj;
			Manifest m = jdo.getJarContent ().getManifest ();
			f = File.createTempFile ("manif", "txt");
			OutputStream os = new FileOutputStream (f);
			try {
				m.write (os);
			} finally {
				os.close ();
			}
		} else if (mffo != null) {
****** /
		if (mffo != null) {
			f = NbClassPath.toFile (mffo);
			if (f == null)
				throw new IOException (getString("MSG_PlainManifest", name));
		} else {
			throw new IOException (getString("MSG_NoManifest", name));
		}
		return f;
	}
    */

	public int loadPropertyFile (String fileName) {
		final String methodName = "NavigationSupport.loadPropertyFile"; // NOI18N

		Exception exc = null;
		fileName = expandVars(fileName);

		if (fileName != null) {
			try {
				FileInputStream fis = null;
				fis = new FileInputStream(fileName);
				forteProps.load(fis);
			}
			catch (FileNotFoundException e) {
				exc = e;
				println(
					getString("MSG_NoPropFile", fileName));
				printException(methodName, e);
			}
			catch (IOException e) {
				exc = e;
				println(
					getString("MSG_IOPropFile", fileName));
				printException(methodName, e);
			}
			catch (IllegalArgumentException e) {
				exc = e;
				println(
					getString("MSG_ErrPropFile", fileName));
				printException(methodName, e);
			}
			if (exc != null)
				return CMD_FAIL;
		}
		return CMD_OK;
	}

	public String getProperty (String propName) {
		return getProperty(propName, null);
	}

	public String getProperty (String propName, String defaultValue) {
		return forteProps.getProperty(propName, defaultValue);
	}

	public Object setProperty (String propName, String value) {
		return forteProps.setProperty(propName, value);
	}

	/**
	 * Expands any property values in this string that have the format
	 * %{<i>propName</i>} or ${<i>propName</i>}.
	 * These are sometimes also called environment variables.
	 *
	 * @param input the string to process for environment variables.
	 * 
	 * @return A string in which the environment variables have been expanded.
	 * 
	 */
	// CHANGES
	//  19-jun-2000  (Roger Blumer)
	//  	fixed $-at-end-string bug (used to get ArrayOutOfBounds)
	//  
	public String expandVars (String str) {
		String	res;
		String	var;
		boolean	xlate;
		byte	buf[];
		int		startptr;
		int		ptr;
		int		ptr2;
		int		ptr3;
		int		leadLen;
		int		len;

		if (str == null)
			return null;

		startptr = 0;
		res = new String();

		//
		// Look for ${var} or %{var}
		//
		ptr = startptr;
		buf = str.getBytes();
		len = buf.length;
		while (ptr < len)
		{
			if (buf[ptr] == '%')
				xlate = true;
			else if (buf[ptr] == '$')
				xlate = false;
			else
			{
				ptr++;
				continue;
			}
			
			ptr2 = ptr + 1;
			if (ptr2 == len) {
				ptr++;	// hit $ at end of string
				continue;
			}

			if (buf[ptr2] != '{')
			{
				// No open curly, this is not a real variable.
				res = res.concat(str.substring(startptr, ptr2));
				ptr = startptr = ptr2;
				continue;
			}
			ptr2++;
			
			// No end curly, this is a bad variable.
			if ((ptr3 = str.indexOf('}', ptr2)) == -1)
			{
				println(getString("MSG_IncompleteVar", str));
				return(str);
			}

			// Expand the var
			var = getProperty(str.substring(ptr2, ptr3));

			// Put any part of the string before the var into the result.
			leadLen = (ptr2 - 2) - startptr;
			if (leadLen > 0)
				res = res.concat(str.substring(startptr, startptr + leadLen));

			// If something was expanded, add it to the result.
			if ((var != null) && (var.length() > 0))
			{
				//
				// Translate to a local name if needed.
				//
			/// I don't think we need to bother with translation
			/// (this was from SynerJ)
			/*	
				if (xlate)
				{
					res = res +
						ThreadContext.shell().io.resolvePortableFileName(var,
							false);
				}
				else
					res = res + var;
			*/
				res = res + var;
			}
			startptr = ptr3 + 1;
			ptr = startptr;
		}
		res = res.concat(str.substring(startptr));
		return res;
	}

	public int loadModule (String fileName) {
		//
		// got the idea to use deploy from InstalledModules.execute()
		//	(package src_internal/com/netbeans/developer/modules/apisupport)
		//
            //File f;
		fileName = expandVars(fileName);
		try {
                    //f = getManifestFile(fileName);
                    //final String toDeploy = f.getAbsolutePath ();
                    // New style (untested!): --jglick
                    ModuleInstaller.deployTestModule (new File (fileName).getAbsoluteFile ());
		}
		catch (IOException ioe) {
			printException("NavigationSupport.loadModule", ioe);
			return CMD_FAIL;
		}
		
		return CMD_OK;
	}

	public int unloadModule  (String fileName) {
		fileName = expandVars(fileName);
		throw new IllegalArgumentException ("unLoadModule: Can't implement yet");
/********
//	the following code uses things with Package visibility (ModuleItem,etc.),
//	so it won't compile.
		File f = getManifestFile(fileName);
		final String manifestFileName = f.getAbsolutePath ();
		ModuleItem exists = ModuleInstaller.getByName( manifestFileName );
		ModuleInstaller.deleteModule (exists);
**********/
	}

	public void printException(String methodName, Throwable exc) {
		println(getString("MSG_ExceptionIn", methodName));
		if (wantStackTrace())
			exc.printStackTrace(outStrm);
		else
			outStrm.println(exc.toString());
                outStrm.flush();
	}
	

	public boolean wantStackTrace() {
		//
		// convert C4 to "cli" or "org.netbeans.commandline.Fjscript"
		// once we have updated logger
		//
                // TODO: fix this (david.strupl)
                return true;
//		TraceLogger lgr = TraceLogger.getTraceLogger();
//		return lgr.test(TraceLogger.CONFIGURATION, lgr.getModuleId(moduleName),
//						CFG_PRINT_STACKTRACE, 1);
	}


        /** Prints and flushes. This is a fix for not sending data 
         * immediately down to the stream.
         */
        private void print(String s) {
            outStrm.print(s);
            outStrm.flush();
        }
        
        /** Prints and flushes. This is a fix for not sending data 
         * immediately down to the stream.
         */
        private void println(String s) {
            outStrm.println(s);
            outStrm.flush();
        }

        /** Prints and flushes. This is a fix for not sending data 
         * immediately down to the stream.
         */
        private void println() {
            outStrm.println();
            outStrm.flush();
        }

        
        /** Getter for a text from resource.
	 * @param resName resource name
	 * @return string with resource
	 */
	static String getString (String resName) {
            try {
		return NbBundle.getBundle (NavigationSupport.class).getString (resName);
            } catch (Exception x) {
                return "Res not found: " + resName;
            }
	}

	/** Getter for a text from resource with one argument.
	 * @param resName resource name
	 * @return string with resource
	 * @param arg the argument
	 */
	static String getString (String resName, Object arg) {
		MessageFormat mf = new MessageFormat (getString (resName));
		return mf.format (new Object[] { arg });
	}
  
	/** Getter for a text from resource with one argument.
	 * @param resName resource name
	 * @return string with resource
	 * @param arg1 the argument
	 * @param arg2 the argument
	 */
	static String getString (String resName, Object arg1, Object arg2) {
		MessageFormat mf = new MessageFormat (getString (resName));
		return mf.format (new Object[] { arg1, arg2 });
	}

	static String getString (String resName, Object arg1, Object arg2, Object arg3) {
		MessageFormat mf = new MessageFormat (getString (resName));
		return mf.format (new Object[] { arg1, arg2, arg3 });
	}

	/** used to keep stack information for Push & Pop
	 */
	static class StackData {
		Object curObj;
		DataObject curDataObj;
	}
	
	public void push() {
		if (stack == null)
			stack = new Stack();

		StackData stData = new StackData();
		stData.curObj     = curObj;
		stData.curDataObj = curDataObj;

		stack.push(stData);		
	}
	public void pop() {
		if (stack == null) {
			curObj = null;
			curDataObj = null;
		}
		StackData stData = (StackData) stack.pop();
		curObj     = stData.curObj;
		curDataObj = stData.curDataObj;
	}

	/** 
	 * Get a template from the system templates folder.
	 *
	 * @param folder The name of the folder within system/Templates (e.g.
	 *               EJB)
	 * @param name The name of the template to find (e.g. SessionBean)
	 * @return The found template or null if not found.
	 */
	public DataObject getTemplate(String folder, String name) {
		DataFolder tempFolder;

		tempFolder = TopManager.getDefault().getPlaces().folders().templates();
		tempFolder = (DataFolder) getChild(tempFolder, folder);
		return getChild(tempFolder, name);
	}

	/** 
	 * Find a template from the system templates folder and make curobj.
	 *
	 * @param folder The name of the folder within system/Templates (e.g.
	 *               EJB)
	 * @param name The name of the template to find (e.g. SessionBean)
	 * @return The found template or null if not found.
	 */
	public Object findTemplate(String folder, String name) {
		DataObject child;

		child = getTemplate(folder, name);
		if (child != null) {
			return setCurObject(child);
		} else {
			return null;
		}
	}
        
        
    /**
     * Run the new from template with the given template.
     * You can use the templateSetXXX commands to set the values. When
     * done do templateInstantiate.
     */
    public void templateStart(String folder, String name) {
        DataObject child = getTemplate(folder, name);

        if (child == null) {
            outStrm.println(  "Template " 
                            + folder
                            + "/"
                            + name
                            + "not found.");
            return;
        }
        templateWizard = new NonGuiTemplateWizard();
        templateWizard.setTemplate(child);
        templateDialog = TopManager.getDefault().createDialog(templateWizard);
    }

    public void templateFinish() throws Throwable {
        if (templateWizard == null) {
            outStrm.println(
                "You must do a templateStart before you can run this command.");
            return;
        }
        templateWizard.finish();
        templateWizard = null;
        templateDialog = null;
    }

    public Dialog templateGetDialog() {
        return templateDialog;
    }

    public TemplateWizard templateGetWizard() {
        return templateWizard;
    }

    public void templatePrintDialog() {
        if (templateDialog == null) {
            outStrm.println("You must run templateStart before this command.");
            return;
        }
        outStrm.println(templateDialog.getClass().getName() + ":");
        printDialog(templateDialog, "");
    }

    public void printDialog(Container cont, String indent) {
        Component[] comps;

        indent = indent + "    ";
        comps = cont.getComponents();
        for (int i = 0; i < comps.length; i++) {
            String txt = "";

            try {
                Method getText;

                getText = comps[i].getClass().getMethod("getText", null);
                txt = (String) getText.invoke(comps[i], null);
            } catch (Throwable t) {
                txt = "";
            }
            outStrm.println(  indent
                            + comps[i].getClass().getName()
                            + ":"
                            + comps[i].getName()
                            + ":"
                            + txt);
            if (comps[i] instanceof Container) {
                printDialog((Container) comps[i], indent);
            }
        }
    }

    public void templatePrintTextField(String name) {
        JTextField tFld = getTextField(name);

        if (tFld != null) {
            outStrm.println(tFld.getText());
        }
    }

    public void templateSetTextField(String name, String value) {
        JTextField tFld = getTextField(name);

        if (tFld != null) {
            tFld.setText(value);
        }
    }

    public void templateClickButton(String name) {
        AbstractButton b = getButton(name);

        if (b != null) {
            b.doClick();
        }
    }


    protected JTextField getTextField(String name) {
        if (templateDialog == null) {
            outStrm.println(
                    "You must run templateStart before running this command.");
            return null;
        }
        JTextField tFld = getTextField(templateDialog, name);
        if (tFld == null) {
            outStrm.println("Text field " + name + " not found.");
            return null;
        }
        return tFld;
    }

    protected JTextField getTextField(Container cont, String name) {
        Component[] comps;

        comps = cont.getComponents();
        for (int i = 0; i < comps.length; i++) {
            if (       comps[i] instanceof JLabel
                    && i + 1 < comps.length
                    && comps[i+1] instanceof JTextField) {
                JLabel lab = (JLabel) comps[i];

                if (lab.getText().equals(name)) {
                    return (JTextField) comps[i+1];
                }
            }
            if (comps[i] instanceof Container) {
                JTextField rval;

                if ((rval = getTextField((Container) comps[i], name)) != null) {
                    return rval;
                }
            }
        }
        return null;
    }

    protected AbstractButton getButton(String name) {
        if (templateDialog == null) {
            outStrm.println(
                    "You must run templateStart before running this command.");
            return null;
        }
        AbstractButton button = (AbstractButton) getComponent(templateDialog,
                                                          name,
                                                          AbstractButton.class);
        if (button == null) {
            outStrm.println("Button " + name + " not found.");
            return null;
        }
        return button;
    }

    protected Component getComponent(Container cont,
                                            String name,
                                            Class theClass) {
        Component[] comps;

        comps = cont.getComponents();
        for (int i = 0; i < comps.length; i++) {
            if (theClass.isAssignableFrom(comps[i].getClass())) {
                String txt = null;

                if (       comps[i].getName() != null
                        && comps[i].getName().equals(name)) {
                    return comps[i];
                }
                try {
                    Method getText = null;

                    getText = comps[i].getClass().getMethod("getText", null);
                    txt = (String) getText.invoke(comps[i], null);
                } catch (Throwable t) {
                    txt = null;
                }
                if (txt != null && txt.equals(name)) {
                    return comps[i];
                }
            }
            if (comps[i] instanceof Container) {
                Component rval;

                rval = getComponent((Container) comps[i], name, theClass);
                if (rval != null) {
                    return rval;
                }
            }
        }
        return null;
    }

    public class NonGuiTemplateWizard extends TemplateWizard {
        public void finish() throws Throwable {
            handleInstantiate();
        }
    }

    // The equivalent to the unix cat command for InputStream
    public int printInputStream(InputStream is) {
        byte[] buf = new byte[1024];
        int len;
        try {
            while ((len = is.read(buf)) > 0) {
                outStrm.write(buf, 0, len);
            }
        } catch (IOException ioe) {
            printException("NonGui.printInputStream", ioe);
            return CMD_FAIL;
        }
        return CMD_OK;
    }
}

//
// CHANGES
//  01-may-2000  (Roger Blumer)
//      Created.
//  04-may-2000  (Roger Blumer)  
//      Added Sun Public License.
//  08-may-2000  (Roger Blumer)  
//      Removed Sun Public License text;
//      this won't be part of initial Open Source release.
//  10-may-2000  (Roger Blumer)  
//      Changed setCurObject to load command class for any object,
//      not just DataObjects.  Changed loadCommandClass to check up superclass
//      chain for command classes.
//  11-may-2000  (Roger Blumer)  
//      Added property handling: loadPropertyFile, getProperty, expandVars.
//      Use expandVars for fileName parameters in all methods.

