/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.scripting.navigation;

import java.awt.event.ActionEvent;
import java.beans.PropertyVetoException;
import java.lang.reflect.*;
import java.security.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.Action;

import org.openide.*;
import org.openide.actions.ActionManager;
import org.openide.actions.ToolsAction;
import org.openide.cookies.SourceCookie;
import org.openide.cookies.InstanceCookie;
import org.openide.execution.ScriptType;
import org.openide.explorer.ExplorerManager;
import org.openide.explorer.ExplorerActions;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.NodeAdapter;
import org.openide.nodes.NodeMemberEvent;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;
import org.openide.util.enum.*;

// Sorry - I cannot avoid this (dstrupl)
import org.netbeans.core.NbTopManager;

/** Class for navigation in nodes.
 *
 * @author Jaroslav Tulach
 */
public class NodeNavigation extends Object {
    /** our error manager */
    private static final org.openide.ErrorManager ERR =
        TopManager.getDefault ().getErrorManager ();
//        TopManager.getDefault ().getErrorManager ().getInstance (
//            "org.netbeans.modules.scripting.NodeNavigation"
//        );
    
    /** the current node. Set to some default value */
    private Node node = TopManager.getDefault ().getPlaces ().nodes ().repository ();
    

    /**
     * Makes the node the currently selected node in the explorerMgr.
     * The node or its root-parent becomes the rootContext of the
     * explorer manager.
     *
     * @param value the node to be selected or null if explorer should be cleared
     */
    public final void setCurrent (Node value) {
        if (value == null) {
                setSelectedNodes((Node[]) null);
                return;
        }
        
        node = value;

        setSelectedNodes(new Node[] { value } );
    }
    
    /** Getter for currently selected node
     * @return selected node (never null)
     */
    public final Node getCurrent () {
        return node;
    }
    
    /** Lists all children in the node.
     * @param path relative path to print content of
     */
    public final void ls (String path) {
        Node c = getCurrent ();
        if (path != null) {
            c = findPath (c, path);
        }
        if (c == null) {
            System.out.println("Path not found");
        } else {
            Node[] arr = findChildren (c);
            System.out.println("Children of: " + c.getDisplayName ());
            for (int i = 0; i < arr.length; i++) {
                System.out.println("  " + arr[i].getDisplayName ());
            }
        }
        
        System.out.flush();
    }
    
    /** List childrens' display names.
     */
    public final void ls () {
        ls (null);
    }
    
    /** Searches the hierarchy of nodes.
     * @param name the names of nodes to find separated with slash (/)
     * @return the node found or null if no such exists (but always tries to
     *    update the selected node property
     */
    public final Node cd (String name) {
        Node n = findPath (getCurrent (), name);
        if (n != null) {
            setCurrent (n);
        }
        return n;
    }
    
    /** Prints current node.
    */
    public final void pwd () {
        System.out.println(getCurrent ().getDisplayName ());
        System.out.flush();
    }
    
    /** Prints names of commands that can be invoked on this instance.
     */
    public final void commands () {
        Enumeration en = findActions ();
        while (en.hasMoreElements ()) {
            Action a = (Action)en.nextElement();
            
            String name = (String)a.getValue(Action.NAME);
            
            System.out.println("  " + name);
        }
        System.out.flush ();
    }
    

    /** Invokes one command. 
     * @param name name of the command to invoke (either full class name,
     *    or full display name or part of the names)
     */
    public final void invoke (String name) {
        ActionCmp cmp = new ActionCmp ();
        
        for (int i = 0; i < ActionCmp.STATES; i++) {
            cmp.state = i;
            Action a = (Action)scanEnum (findActions (), name, cmp);
            if (a != null) {
                a.actionPerformed(new ActionEvent (this, 0, null));
                return;
            }
        }
    }

    /** Sets current node to given "well known" node.
     */
    public final void root(String name) {
        if (name.equalsIgnoreCase("controlPanel")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().controlPanel());
        }
        if (name.equalsIgnoreCase("environment")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().environment());
        }
        if (name.equalsIgnoreCase("loaderPool")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().loaderPool());
        }
        if (name.equalsIgnoreCase("project")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().project());
        }
        if (name.equalsIgnoreCase("projectDesktop")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().projectDesktop());
        }
        if (name.equalsIgnoreCase("repository")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().repository());
        }
        if (name.equalsIgnoreCase("repositorySettings")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().repositorySettings());
        }
        if (name.equalsIgnoreCase("session")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().session());
        }
        if (name.equalsIgnoreCase("workspaces")) {
            setCurrent(TopManager.getDefault().getPlaces().nodes().workspaces());
        }
    }
    //
    // Implementation of node traversing
    //
    
    /** Finds a node following path from starting node.
    * @param n node to start at
    * @param path the string separated with "/"
    * @return found node (closest as possible) or null if no match has been found
    */
    private static Node findPath (Node current, String path) {
        String name = path.trim ();
        
        StringTokenizer tok = new StringTokenizer (name, "/");
        
        if (name.startsWith("/")) {
            current = root (current);
            if (name.length () == 1) {
                // just request for root
                return current;
            }
        }

        
        Node the = current;
        boolean change = false;
        
        while (the != null && tok.hasMoreTokens ()) {
            String s = tok.nextToken ();
            
            the = findChild (the, s);
            if (the != null) {
                change = true;
                current = the;
            }
        }
        
        if (!change) {
            // no node has been found, try to look for whole name
            the = findChild (current, name);
            if (the != null) {
                current = the;
            }
        }
        
        return current;
    }
        
    
    /** Finds a child of a node by name. Improvements possible in future.
     * @param n node to find child for
     * @param name name of child to find
     * @return the found child or null
     */
    private static Node findChild (Node n, String name) {
        if ("..".equals (name)) {
            // request for parent node
            return n.getParentNode ();
        }

        Node[] arr = findChildren (n);
        
        NodeCmp cmp = new NodeCmp ();

        // what this actually does is it tries to find a node by 
        // its getName (), then by its getDisplayName (), then it looks 
        // whether the name is not part of a node's getName and lastly 
        // part of node's getDisplayName ()
        for (int i = 0; i < NodeCmp.STATES; i++) {
            cmp.state = i;
            Object o = scanEnum (
                new ArrayEnumeration (arr), name, cmp
            );
            
            if (o != null) {
                return (Node)o;
            }
        }
        return null;
    }
    
    
    /** Helper finder in enumerations. Receives enumeration and also
    * comparator. The method goes thru the enumeration and tries to find 
    * the first matching object.
    *
    * @param en enumeration of objects
    * @param pattern the pattern object to look for
    * @param comp comparator to use for comparing of objects
    * @return the found object or null
    */
    private static Object scanEnum (
        Enumeration en, Object pattern, java.util.Comparator comp
    ) {
        while (en.hasMoreElements()) {
            Object o = en.nextElement ();
            if (comp.compare(pattern, o) == 0) {
                return o;
            }
        }
        return null;
    }
    
    /** Finds root for given node.
     * @param n the node
     * @return root of n
     */
    private static Node root (Node n) {
        for (;;) {
            Node nn = n.getParentNode ();
            if (nn == null) {
                return n;
            }
            n = nn;
        }
    }
    
    /** Computes children of a node. Waits until all of them has been computed,
     * by calling findChild method, which should block when complex computation
     * is performed. 
     *
     * @param node the node to find children for
     * @return array of children for given node
     */
    private static Node[] findChildren (Node n) {
        /** This class holds all creates nodes, so they will not be garbage
         * collected.
         */
        class Collector extends NodeAdapter {
            private java.util.List ll = new java.util.LinkedList ();
            
            /** Fired when a set of new children is added.
            * @param ev event describing the action
            */
            public void childrenAdded (NodeMemberEvent ev) {
                ll.add (ev.getDelta ());
            }
        }
        
        Collector c = new Collector ();
        try {
            n.addNodeListener (c);
            Node[] tempArr = n.getChildren ().getNodes ();
            // finds some dummy child (which should block until the children are
            // initialized
            Node any = n.getChildren ().findChild (null);

            return n.getChildren ().getNodes ();
        } finally {
            // remove the listener
            n.removeNodeListener (c);
        }
    }
            

    /**
     * Makes the nodes the currently selected nodes in the explorerMgr.
     * The first node in the array (or its root-parent) becomes the
     * rootContext of the explorer manager.
     *
     * @param value the node to be selected.  If null, the selection is cleared.
     */
    private static final void setSelectedNodes (Node[] arr) {
        if (arr == null) {
            EMRun.stop ();
            return;
        } 

        ExplorerManager em = EMRun.init ();

        if (arr == null)
                arr = new Node[] {};	// set to empty array, to clear selection

        try {
            // attempt to set the root context based on the first node
            //
            if (arr.length >= 1) {
                Node root = arr[0];
                if (root != null) {
                        while (root.getParentNode() != null)
                                root = root.getParentNode();
                        em.setRootContext(root);
                }
            }

            em.setSelectedNodes(arr);
        }
        catch (PropertyVetoException e) {
            ERR.notify (e);
        }
        catch (IllegalArgumentException e) {
            ERR.notify (e);
        }
    }
    
    
    /** Creates enumeration of all available actions.
     * @return enum of javax.swing.Action
     */
    private static Enumeration findActions () {
	DataFolder df = findActionsFolder();
        
        Enumeration allObjects = df.children (true);
        
        Enumeration allActions = new AlterEnumeration (allObjects) {
            protected Object alter (Object o) {
                DataObject obj = (DataObject)o;
                InstanceCookie ic = (InstanceCookie)obj.getCookie (
                    InstanceCookie.class
                );
                if (ic == null) return null;
                
                try {
                    Object instance = ic.instanceCreate ();
                    if (instance instanceof Action) {
                        Action a = (Action)instance;
                        if (a.isEnabled()) {
                            return instance;
                        }
                    }
                } catch (ClassNotFoundException ex) {
                    ERR.notify (ERR.INFORMATIONAL, ex);
                } catch (IOException ex) {
                    ERR.notify (ERR.INFORMATIONAL, ex);
                }
                
                return null;
            }
        };
        
        return new FilterEnumeration (allActions) {
            protected boolean accept (Object obj) {
                return obj != null;
            }
        };
    }

   /**
    * Finds folder with scripting actions. It is ScriptActions under
    * default file system. If the folder does not exist creates a new
    * one by copying the actions folder.
    */
    private static DataFolder findActionsFolder() {
        FileObject f = null;
        f = TopManager.getDefault().getRepository().findResource("ScriptActions"); //NOI18N
        if (f == null) {
            // resource not found, try to create new folder
            f = (FileObject)AccessController.doPrivileged(new PrivilegedAction() {
                public Object run() {
                    try {
                        DataObject systemActions = TopManager.getDefault().getPlaces().folders().actions();
                        FileObject tmpFolder = TopManager.getDefault().getRepository ().getDefaultFileSystem().getRoot().createFolder("tmp");
                        FileLock lock = tmpFolder.lock();
                        DataObject tmpActions = systemActions.copy(DataFolder.findFolder(tmpFolder));
                        tmpActions.rename("ScriptActions"); //NOI18N
                        tmpActions.move(DataFolder.findFolder(TopManager.getDefault().getRepository ().getDefaultFileSystem().getRoot()));
                        tmpFolder.delete(lock);
                        return tmpActions.getPrimaryFile();
                    } catch (Exception e) {
                        TopManager.getDefault().getErrorManager().notify(ErrorManager.EXCEPTION, e);
                    }
                    return null;
                }
            });
        }
        DataFolder df = DataFolder.findFolder(f);
        return df;
    }

    /** Hack to set activated nodes to getCurrent value.
     */
    private static final class EMRun implements Runnable {
        private static EMRun INSTANCE = new EMRun ();
        private static ExplorerManager em;
        private static ExplorerActions ea = new ExplorerActions ();

        /** Implements the runnable interface. Is called twice. Once 
         * from the Thread (this) and second time from the
         * attachExplorer (this) method.
         * For the first call em == null for the second em != null.
         */
        public synchronized void run () {
            if (em == null) {
                // the code executed when called from Thread (this)
                em = new ExplorerManager ();
                ea.attach (em);
                notify ();

                NbTopManager.get ().attachExplorer(this, em);
                return;
            }

            // the code executed from attachExplorer (this)
            try {
                // waits until somebody calls stop
                wait ();
            } catch (InterruptedException ex) {
            }
        }

        public static void stop () {
            synchronized (INSTANCE) {
                // clears the manager and notifies the waiting 
                // thread if any
                ea.detach ();
                em = null;
                INSTANCE.notify ();
            }
        }

        public static ExplorerManager init () {
            synchronized (INSTANCE) {
                while (em == null) {
                    // called for the first time => create 
                    // the explorer manager (in run () method)
                    new Thread (EMRun.INSTANCE).start ();
                    try {
                        // wait till the em is filled
                        INSTANCE.wait ();
                    } catch (InterruptedException ex) {
                    }
                }
                // return the explorer manager
                return em;
            }
        }
    } // end of EMRun

    
    /** Comparator of nodes. Has 4 states.
    */
    private static class NodeCmp implements java.util.Comparator {
        public static final int STATES = 4;
        
        public int state;

        public int compare (Object o1, Object o2) {
            String name = (String)o1;
            Node n = (Node)o2;

            switch (state) {
                case 0: 
                    // comparing on name of node
                    return name.equals (n.getName ()) ? 0 : 1;
                case 1: 
                    // comparing on display name of node
                    return name.equals (n.getDisplayName ()) ? 0 : 1;
                case 2: 
                    // comparing on part of name
                    {
                        String nn = n.getName ();
                        return nn != null && nn.indexOf (name) >= 0 ? 0 : 1;
                    }
                case 3:
                    // on part of display name
                    {
                        String nn = n.getDisplayName ();
                        return nn != null && nn.indexOf (name) >= 0 ? 0 : 1;
                    }
                default:
                    throw new InternalError ();
            }
        }
    }
    
    /** Comparator of nodes. Has 4 states.
    */
    private static class ActionCmp implements java.util.Comparator {
        public static final int STATES = 4;
        
        public int state;

        public int compare (Object o1, Object o2) {
            String name = (String)o1;
            Action a = (Action)o2;

            switch (state) {
                case 0: 
                    // comparing on name of class of the action
                    return name.equals (a.getClass ().getName ()) ? 0 : 1;
                case 1: 
                    // comparing on display name 
                    return name.equals (a.getValue(Action.NAME)) ? 0 : 1;
                case 2: 
                    // comparing on part of class name
                    {
                        String nn = a.getClass ().getName ();
                        return nn != null && nn.indexOf (name) >= 0 ? 0 : 1;
                    }
                case 3:
                    // on part of display name
                    {
                        String nn = (String)a.getValue (Action.NAME);
                        return nn != null && nn.indexOf (name) >= 0 ? 0 : 1;
                    }
                default:
                    throw new InternalError ();
            }
        }
    }
    
/*    
    public static void main (String[] args) throws Exception {
        NodeNavigation nn = new NodeNavigation ();
System.out.println("identity: " + System.identityHashCode(nn));        
        
        Enumeration en = ScriptType.scriptTypes ();
        while (en.hasMoreElements()) {
            ScriptType t = (ScriptType)en.nextElement();
System.out.println("modifying " + t);            
            t.addVariable ("ide", nn);
        }
    }
 */
}

