/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.text;


import java.beans.PropertyVetoException;
import java.io.IOException;
import java.text.MessageFormat;

import org.openide.loaders.DataLoader;
import org.openide.loaders.DataObject;
import org.openide.loaders.ExtensionList;
import org.openide.nodes.Node;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.util.actions.CookieAction;


/**
 * Action which converts <code>TXTDataObject</code>'s to <code>DefaultDataObject</code>'s.
 * @author Peter Zavadsky
 */
public class ConvertToDefaultAction extends CookieAction {

    /** Generated serail version UID. */
    private static final long serialVersionUID = 3495147047528675606L;

    
    /** Overrides superclass method. Enables action only on that <code>TXTDataObect</code> which were
     * <code>DefaultDataObject</code> before. */
    protected boolean enable(Node[] activatedNodes) {
        if(!super.enable(activatedNodes))
            return false;
        
        for(int i=0; i<activatedNodes.length; i++) {
            DataObject txtDataObject = (DataObject)activatedNodes[i].getCookie(TXTDataObject.class);
            if(txtDataObject == null)
                continue;

            // Don't convert read-only objects.
            if(txtDataObject.getPrimaryFile().isReadOnly())
                continue;
            
            String extension = txtDataObject.getPrimaryFile().getExt();
            ExtensionList list = ((TXTDataLoader)DataLoader.getLoader(TXTDataLoader.class)).getExtensions();
            
            if(extension.length()==0 || !(list.isRegistered(extension)))
                return true;
        }
        return false;
    }

    /** Implements superclass abstract method. Cookies which presence requires this action. 
     * @return array containing <code>TXTDataObject</code> class (i.e. cookie class) only */
    protected Class[] cookieClasses () {
        return new Class[] { TXTDataObject.class };
    }

    /** Implements superclass abstract method. Gets which exactly from cookies action require.(In our case all.) 
     * @return CookieAction.MODE_ALL value
     * @see org.openide.util.CookieAction#MODE_ALL */
    protected int mode () {
        return MODE_ALL;
    }
    
    /** Implements superclass abstract method. Actually perform action. */
    protected void performAction (Node[] nodes) {
        for (int i=0; i<nodes.length; i++) {
            DataObject txtDataObject = (DataObject)nodes[i].getCookie(TXTDataObject.class);
            if (txtDataObject == null)
                continue;

            // Don't convert read-only objects.
            if(txtDataObject.getPrimaryFile().isReadOnly())
                continue;
            
            // Ask user for confirmation first.
            NotifyDescriptor.Confirmation descriptor = new NotifyDescriptor.Confirmation(
                MessageFormat.format(NbBundle.getBundle(TXTModule.class).getString("MSG_ConvertToDefaultQuestion"),
                    new Object[] {txtDataObject.getPrimaryFile().getName()+"."+txtDataObject.getPrimaryFile().getExt()}), // NOI18N
                NotifyDescriptor.OK_CANCEL_OPTION
            );

            // If user confirmed change do the dirty job.
            if(TopManager.getDefault().notify(descriptor).equals(NotifyDescriptor.OK_OPTION)) {
                try {
                    txtDataObject.getPrimaryFile().setAttribute(TXTDataLoader.ATTR_IS_TEXT_FILE, null);
                    txtDataObject.setValid(false);
                } catch (IOException ioe) {
                    TopManager.getDefault().notifyException(ioe);
                } catch (PropertyVetoException pve) {
                    TopManager.getDefault().notifyException(pve);
                }
            }
        }
    }

    /** Overrides superclass method. Gets name of icon resource for action. */
    protected String iconResource () {
        return "txtObject.gif"; // NOI18N
    }
    
    /** Implements superclass abstract method. Gets name of action. */
    public String getName () {
        return NbBundle.getBundle (ConvertToTextAction.class).getString("LBL_ConvertToDefault");
    }

    /** Implements superclass abstract method. Gets help context. */
    public HelpCtx getHelpCtx() {
        return new HelpCtx(ConvertToTextAction.class); // TEMP
    }

}

