/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.text;


import java.beans.PropertyVetoException;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Enumeration;

import org.openide.loaders.DataLoader;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.util.actions.CookieAction;
import org.openide.util.HelpCtx;
import org.openide.util.NbBundle;
import org.openide.filesystems.FileStateInvalidException;


/** 
 * Action which converts <code>DefaultDataObject</code>s to <code>TXTDataObject</code>s.
 *
 * @author Jesse Glick
 */
public class ConvertToTextAction extends CookieAction {

    /** Generated serial version UID. */
    private static final long serialVersionUID = 3495147047528675606L;

    
    /** Overrides superclass method. */
    protected boolean enable (Node[] activatedNodes) {
        if(!super.enable(activatedNodes))
            return false;
        
        for(int i=0; i<activatedNodes.length; i++) {
            DataObject ddo = (DataObject)activatedNodes[i].getCookie(getDDOClazz());
            if(ddo == null)
                continue;

            if (isSystemObject (ddo)) continue;
                
            // If at least one from the selected ddo is not read only enable action.
            if(!ddo.getPrimaryFile().isReadOnly())
                return true;
        }
        
        return false;
    }

    /** Implements superclass abstract method. Cookies which presence requires this action. 
     * @return array containing default data object class (i.e cookie class) only */
    protected Class[] cookieClasses () {
        return new Class[] { getDDOClazz() };
    }

    /** Implements superclass abstract method. Gets which exactly from cookies action require.(In our case all.) 
     * @return CookieAction.MODE_ALL value
     * @see org.openide.util.CookieAction#MODE_ALL */
    protected int mode () {
        return MODE_ALL;
    }

    /** Implements superclass abstract method. Actually perform action. */
    protected void performAction (Node[] nodes) {
        for (int i = 0; i < nodes.length; i++) {
            DataObject ddo = (DataObject)nodes[i].getCookie(getDDOClazz());
            if (ddo == null)
                continue;

            if (isSystemObject (ddo)) continue;

            // Don't perform action on read-only objects.
            if(ddo.getPrimaryFile().isReadOnly())
                continue;
            
            // Ask user for confirmation first.
            NotifyDescriptor.Confirmation descriptor = new NotifyDescriptor.Confirmation(
                MessageFormat.format(NbBundle.getBundle(TXTModule.class).getString("MSG_ConvertToTextQuestion"),
                    new Object[] {ddo.getPrimaryFile().getName()+"."+ddo.getPrimaryFile().getExt()}), // NOI18N
                NotifyDescriptor.OK_CANCEL_OPTION
            );
        
            // If user confirmed change do the dirty job.
            if(TopManager.getDefault().notify(descriptor).equals(NotifyDescriptor.OK_OPTION)) {
                try {
                    ddo.getPrimaryFile().setAttribute(TXTDataLoader.ATTR_IS_TEXT_FILE, Boolean.TRUE);
                    ddo.setValid(false);
                } catch (IOException ioe) {
                    TopManager.getDefault().notifyException(ioe);
                } catch (PropertyVetoException pve) {
                    TopManager.getDefault().notifyException(pve);
                }
            }
        }
    }

    /** true if this one lives on SFS */
    private static final boolean isSystemObject (DataObject obj) {
        try {
            return obj.getPrimaryFile ().getFileSystem () ==
                TopManager.getDefault ().getRepository ().getDefaultFileSystem ();
        } catch (FileStateInvalidException fsie) {
            // Never mind.
            return false;
        }
    }

    /** Overrides superclass method. Gets name of icon resource for action. */
    protected String iconResource () {
        return "txtObject.gif"; // NOI18N
    }

    /** Implements superclass abstract method. Gets name of action. */
    public String getName () {
        return NbBundle.getBundle (ConvertToTextAction.class).getString("LBL_ConvertToText");
    }
    
    /** Implements superclass abstract method. Gets help context. */
    public HelpCtx getHelpCtx() {
        return new HelpCtx(ConvertToTextAction.class);
    }

    /** Gets <code>DefaultDataObject</code> class. */
    private Class getDDOClazz () {
        return ConvertToTextAction.getDefaultDataLoader().getRepresentationClass();
    }
    
    /** Helper method. Gets default data loader. 
     * @return default data loader instance or null (shouldn't happen) in case of some error */
    static DataLoader getDefaultDataLoader() {
        // This is not usual way how to get the data loader. Our problem is that default data
        // object is package private and therefore its class is not accessible in normal way
        // On the other hand there should be a guarranty that default data loader is the last one from 
        // all data loaders in data loader pool, so we get it just the way we get the last loader
        // assumnig it is the default data loader.
        // Note: in case of changes in data loaders API we would have to use the way which is at the next commented out row
        // DataLoader DDOLoader =  TopManager.getDefault().getLoaderPool().firstProducerOf(Class.forName("org.openide.loaders.DefaultDataObject")); // NOI18N
        Enumeration loaders = TopManager.getDefault().getLoaderPool().allLoaders();
        DataLoader DDOLoader = null;
        // get last data loader from enumeration which have to be default data loader
        for(; loaders.hasMoreElements(); )
            DDOLoader = (DataLoader)loaders.nextElement();

        return DDOLoader;
    }
    
}
