/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.text;


import java.beans.PropertyVetoException;
import java.io.IOException;
import java.util.ResourceBundle;

import org.openide.loaders.UniFileLoader;
import org.openide.loaders.ExtensionList;
import org.openide.loaders.DataObject;
import org.openide.loaders.MultiDataObject;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.actions.*;
import org.openide.util.actions.SystemAction;
import org.openide.util.NbBundle;
import org.openide.TopManager;


/** 
 * Data loader which recognizes txt files.
 * This class is final only for performance reasons,
 * can be unfinaled if desired.
 *
 * @author Petr Hamernik, Jaroslav Tulach, Dafe Simonek
 */
public final class TXTDataLoader extends UniFileLoader {

    /** Generated serial version UID. */
    static final long serialVersionUID =-3658061894653334886L;    
    
    /** file attribute which forces a file to be considered a text file */
    static final String ATTR_IS_TEXT_FILE = "org.netbeans.modules.text.IsTextFile"; // NOI18N

    
    /** Creates new <code>TXTDataLoader</code>. */
    public TXTDataLoader() {
        super(TXTDataObject.class);
    }

    /** Does initialization. Initializes extension list. */
    protected void initialize () {
        ExtensionList ext = new ExtensionList();
        ext.addExtension("txt"); // NOI18N
        ext.addExtension("doc"); // NOI18N
        ext.addExtension("me"); // for read.me files // NOI18N
        ext.addExtension("policy"); // NOI18N
        ext.addExtension("properties"); // provided for case when PropertiesModule is not installed // NOI18N
        ext.addExtension("mf"); // for manifest.mf files // NOI18N
        ext.addExtension("MF"); //  -""- // NOI18N
        ext.addExtension("dtd"); // when xml module isn't installed, you can view dtd as plain text // NOI18N
        setExtensions(ext);
    }
    
    /** Gets default display name. Overrides superclass method. */
    protected String defaultDisplayName() {
        return NbBundle.getBundle(TXTDataLoader.class).getString("PROP_TXTLoader_Name");
    }
    
    /** Gets default system actions. Overrides superclass method. */
    protected SystemAction[] defaultActions() {
        return new SystemAction[] {
            SystemAction.get(OpenAction.class),
            SystemAction.get (FileSystemAction.class),
            null,
            SystemAction.get(CutAction.class),
            SystemAction.get(CopyAction.class),
            SystemAction.get(PasteAction.class),
            null,
            SystemAction.get(DeleteAction.class),
            SystemAction.get(RenameAction.class),
            null,
            SystemAction.get(SaveAsTemplateAction.class),
            null,
            SystemAction.get(ToolsAction.class),
            SystemAction.get(PropertiesAction.class),
        };
    }

    /** Check whether a file is recognized.
     * It will be if the extension matches, or if it is marked to be a text file. */
    protected FileObject findPrimaryFile (FileObject fo) {
        boolean isSysFile;
        try {
            isSysFile = fo.getFileSystem () == TopManager.getDefault ().getRepository ().getDefaultFileSystem ();
        } catch (FileStateInvalidException fsie) {
            // Never mind.
            isSysFile = false;
        }
        if (! isSysFile && Boolean.TRUE.equals (fo.getAttribute (ATTR_IS_TEXT_FILE)))
            return fo;
        return super.findPrimaryFile (fo);
    }

    /** Creates new <code>TXTDataObject</code> for specified <code>FileObject</code>.
     * @param fo FileObject
     * @return new TXTDataObject
     */
    protected MultiDataObject createMultiObject(final FileObject fo)
    throws IOException {
        return new TXTDataObject(fo, this);
    }

}
