/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.text;


import java.awt.Image;

import org.openide.actions.OpenAction;
import org.openide.cookies.CompilerCookie;
import org.openide.filesystems.FileObject;
import org.openide.loaders.*;
import org.openide.nodes.Children;
import org.openide.nodes.CookieSet;
import org.openide.nodes.Node;
import org.openide.util.actions.SystemAction;
import org.openide.util.HelpCtx;


/** 
 * Object that provides main functionality for txt data loader.
 * This class is final only for performance reasons,
 * can be unfinaled if desired.
 * @author Petr Hamernik
 */
public final class TXTDataObject extends MultiDataObject implements CookieSet.Factory {

    /** Generated Serialized Version UID */
    static final long serialVersionUID = 4795737295255253334L;

    /** Editor support for text data object. */
    private transient TXTEditorSupport editorSupport;
    

    /** Constructor. */
    public TXTDataObject(final FileObject obj, final MultiFileLoader loader) throws DataObjectExistsException {
        super(obj, loader);
        
        getCookieSet().add(TXTEditorSupport.class, this);
    }

    
    /** Implements <code>CookieSet.Factory</code> interface. */
    public Node.Cookie createCookie(Class clazz) {
        if(clazz.isAssignableFrom(TXTEditorSupport.class))
            return getEditorSupport();
        else
            return null;
    }
    
    /** Gets editor support for this data object. */
    private TXTEditorSupport getEditorSupport() {
        if(editorSupport == null) {
            synchronized(this) {
                if(editorSupport == null)
                    editorSupport = new TXTEditorSupport(this);
            }
        }
        
        return editorSupport;
    }

    /** Provides node that should represent this data object. When a node for representation
     * in a parent is requested by a call to getNode (parent) it is the exact copy of this node
     * with only parent changed. This implementation creates instance <code>DataNode</code>.
     * <p>
     * This method is called only once.
     *
     * @return the node representation for this data object
     * @see DataNode
     */
    protected Node createNodeDelegate () {
        return new TXTNode(this);
    }

    /** Help context for this object.
     * @return help context
     */
    public HelpCtx getHelpCtx () {
        return new HelpCtx (TXTDataObject.class);
    }


    /** Text node implementation.
     * Leaf node, default action opens editor or instantiates template.
     * Icons redefined.
     */
    public static final class TXTNode extends DataNode {
        /** Icon base for the TXTNode node */
        private static final String TXT_ICON_BASE = "org/netbeans/modules/text/txtObject"; // NOI18N

        /** Constructs node. */
        public TXTNode (final DataObject dataObject) {
            super(dataObject, Children.LEAF);
            setIconBase(TXT_ICON_BASE);
        }

        /** Overrides default action from DataNode. */
        public SystemAction getDefaultAction () {
            SystemAction result = super.getDefaultAction();
            return result == null ? SystemAction.get(OpenAction.class) : result;
        }
    } // End of nested class TXTNode.

}
