/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.usersguide;

import java.awt.Image;
import java.beans.*;
import java.io.*;
import java.net.URL;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.openide.ErrorManager;
import org.openide.TopManager;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.LocalFileSystem;
import org.openide.filesystems.Repository;
import org.openide.modules.ModuleInstall;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;

/** Installs online help. Unpacks & mounts samples as needed.
 * @author Jesse Glick
 */
public class UGModule extends ModuleInstall implements PropertyChangeListener, Runnable {

    private static final long serialVersionUID = -5353846631812949705L;

    private static final String PROP_INSTALL_COUNT = "installCount"; // NOI18N

    private static ErrorManager err = TopManager.getDefault ().getErrorManager ();

    private static int icount = 0;

    private Node projDesk;

    private FileSystem toAdd = null;
    
    public void installed () {
        restored ();
    }

    private static FileSystem createLocalFS (File dir) throws Exception {
        LocalFileSystem fs = new GlobalLocalFileSystem ();
        fs.setRootDirectory (dir);
        return fs;
    }

    public void restored () {
        // Mount docs, or remount if project was discarded:
        icount++;
        //System.err.println ("usersguide.restored: count=" + icount);
        // 1: first install (project is discarded anyway)
        // 2: first restore as actual user
        // 3: next restore (project settings incl. Repository loaded)
        if (icount <= 2) {
            File f;
            try {
                f = makeExamplesDir ();
            } catch (IOException ioe) {
                err.notify (ErrorManager.INFORMATIONAL, ioe);
                f = null;
            }
            if (f != null) {
                try {
                    toAdd = createLocalFS (f);
                    Repository repo = TopManager.getDefault ().getRepository ();
                    if (repo.findFileSystem (toAdd.getSystemName ()) == null) {
                        err.log (ErrorManager.INFORMATIONAL, "Adding samples directory " + f + " to Filesystems.");
                        repo.addFileSystem (toAdd);
                    } else {
                        err.log (ErrorManager.INFORMATIONAL, "Note: samples directory was already present in Filesystems.");
                    }
                    // HACK: according to whim of project module, that might be removed again
                    // later! So try to make sure...
                    projDesk = TopManager.getDefault ().getPlaces ().nodes ().projectDesktop ();
                    TopManager.getDefault ().addPropertyChangeListener (this);
                    RequestProcessor.postRequest (this, 10000);
                } catch (Exception e) {
                    err.notify (ErrorManager.INFORMATIONAL, e);
                }
            } else {
                err.log (ErrorManager.INFORMATIONAL, "No examples to add to Filesystems.");
            }
        }
    }

    public void propertyChange (PropertyChangeEvent ev) {
        if (projDesk != TopManager.getDefault ().getPlaces ().nodes ().projectDesktop ()) {
            err.log ("Project initialized, check samples directory...");
            run ();
        }
    }

    public void run () {
        if (toAdd != null) {
            Repository repo = TopManager.getDefault ().getRepository ();
            if (repo.findFileSystem (toAdd.getSystemName ()) == null) {
                err.log (ErrorManager.INFORMATIONAL, "Adding samples directory to Filesystems again.");
                repo.addFileSystem (toAdd);
            } else {
                err.log (ErrorManager.INFORMATIONAL, "OK, samples directory is still in Filesystems.");
            }
            toAdd = null;
        }
    }

    public void uninstalled () {
        // Unmount examples:
        Repository repo = TopManager.getDefault ().getRepository ();
        Enumeration e = repo.fileSystems ();
        while (e.hasMoreElements ()) {
            Object o = e.nextElement ();
            if (o instanceof GlobalLocalFileSystem) {
                repo.removeFileSystem ((FileSystem) o);
            }
        }
    }
    
    /** Creates a ${netbeans.user}/examples/ directory, if it
     * does not already exist, and returns that dir. If we are
     * running in single-user mode, does not try to make the
     * directory and simply returns null.
     */
    private static File makeExamplesDir () throws IOException {
        String homedir = System.getProperty ("netbeans.home");
        String userdir = System.getProperty ("netbeans.user");
        if (userdir == null || userdir.equals (homedir)) {
            // Single-user mode; whatever.
            return null;
        } else {
            // Handle old Development directory sanely, do not duplicate.
            File olddir = new File (userdir, "Development"); // NOI18N
            if (olddir.isDirectory ()) {
                return olddir;
            }
            // OK, will try to produce the examples dir.
            String egdir = NbBundle.getMessage (UGModule.class, "FILE_examples");
            File dir = new File (userdir, egdir);
            if (! dir.isDirectory ()) {
                err.log (ErrorManager.INFORMATIONAL, "Creating samples directory " + dir);
                // Need to create it & unpack examples into it.
                if (! dir.mkdirs ()) throw new IOException ();
                byte[] buf = new byte[4096];
                URL u = new URL ("nbresloc:/org/netbeans/modules/usersguide/demosrc.zip"); // NOI18N
                ZipInputStream zis = new ZipInputStream (u.openStream ());
                try {
                    ZipEntry entry;
                    while ((entry = zis.getNextEntry ()) != null) {
                        String name = entry.getName ().replace ('/', File.separatorChar);
                        if (entry.isDirectory ()) {
                            if (! new File (dir, name).mkdirs ()) throw new IOException ();
                        } else {
                            OutputStream os = new FileOutputStream (new File (dir, name));
                            try {
                                int read;
                                while ((read = zis.read (buf, 0, buf.length)) != -1) {
                                    os.write (buf, 0, read);
                                }
                            } finally {
                                os.close ();
                            }
                        }
                    }
                } finally {
                    zis.close ();
                }
            }
            return dir;
        }
    }

    public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal (in);
        try {
            Integer i = (Integer) in.readObject ();
            icount = ((i == null) ? 0 : i.intValue ());
        } catch (OptionalDataException ode) {
            if (ode.eof) {
                // Old, old UGModule with no such property.
                err.notify (ErrorManager.INFORMATIONAL, ode);
                icount = 1;
            } else {
                throw ode;
            }
        }
        //System.err.println ("usersguide.readExternal: count=" + icount);
    }

    public void writeExternal (ObjectOutput out) throws IOException {
        super.writeExternal (out);
        out.writeObject (new Integer (icount));
        //System.err.println ("usersguide.writeExternal: count=" + icount);
    }
    
    /** Exists only for the sake of its bean info. */
    public static final class GlobalLocalFileSystem extends LocalFileSystem {
        private static final long serialVersionUID = -2365058849573910139L;
    }
    /** Marks it as global (not project-specific). */
    public static final class GlobalLocalFileSystemBeanInfo extends SimpleBeanInfo {
        public BeanDescriptor getBeanDescriptor () {
            BeanDescriptor bd = new BeanDescriptor (GlobalLocalFileSystem.class);
            bd.setValue ("global", Boolean.TRUE); // NOI18N
            return bd;
        }
        public BeanInfo[] getAdditionalBeanInfo () {
            try {
                return new BeanInfo[] { Introspector.getBeanInfo (LocalFileSystem.class) };
            } catch (IntrospectionException ie) {
                err.notify (ie);
                return null;
            }
        }
        public Image getIcon (int kind) {
            try {
                return Introspector.getBeanInfo (LocalFileSystem.class).getIcon (kind);
            } catch (IntrospectionException ie) {
                err.notify (ie);
                return null;
            }
        }
    }

}
