/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.openfile;


import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;
import javax.swing.JEditorPane;
import javax.swing.SwingUtilities;

import org.openide.cookies.EditorCookie;
import org.openide.loaders.DataObject;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.util.NbBundle;
import org.openide.windows.TopComponent;


/** 
 * Acts as an UDP server for requests from the Opend file launcher{@link Main}.
 *
 * @author Jaroslav Tulach, Jesse Glick
 * @see Main
 */
public class Server extends Object implements Runnable {

    /** Max length of transferred data. */
    private static final int LENGTH = 512;
    
    /** Flag indicating if we should stop due to uninstallation. */
    private static boolean stop;
    
    /** Socket to use. */
    private static DatagramSocket socket;
    
    /** <code>Waiting</code>'s to provide callback notifications. */
    private final static Set waitings = new HashSet();
    

    /** Satrts server. */
    static void startup () {
        //System.err.println("Server.startup");
        stop = false;
        if (! Settings.DEFAULT.isActualRunning ()) {
            //System.err.println("Server.startup: starting thread");
            new Thread(new Server (), "OpenFile Server").start (); // NOI18N
        }
    }

    /** Shuts down open file server. */
    public static void shutdown() {
        shutDownWaitings();
        
        stop = true;
        
        if(Settings.DEFAULT.isActualRunning()) {
            try {
                int port = Settings.DEFAULT.getActualPort ();
                if (port == 0) {
                    if(Boolean.getBoolean("netbeans.debug.exceptions")) // NOI18N
                        System.err.println("OpenFile: Port should not have been zero during shutdown!"); // NOI18N
                    return;
                }
                
                DatagramPacket shutDownPacket = new DatagramPacket (new byte[] { (byte) 'X' }, 1, InetAddress.getLocalHost (), port);
                DatagramSocket shutDownSocket = new DatagramSocket ();
                
                try {
                    shutDownSocket.send(shutDownPacket);
                } finally {
                    shutDownSocket.close();
                }
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }
    }

    /** Sets up the socket */
    private static void initSocket () {
        if(socket != null)
            socket.close ();
        
        int port = Settings.DEFAULT.getPort ();
        
        try {
            socket = new DatagramSocket(port);
            Settings.DEFAULT.setActualPort0(port);
        } catch(SocketException se) {
            TopManager.getDefault().notify(new NotifyDescriptor.Message (SettingsBeanInfo.getString ("MSG_cannotBind", new Integer(port))));
            Settings.DEFAULT.setRunning(false);
            Settings.DEFAULT.setActualPort0(0);
        }
    }
    
    /** Waits on the connection. */
    public void run() {
        initSocket();
        
        DatagramPacket packet = new DatagramPacket(new byte[LENGTH], LENGTH);
        try {
REQUESTS: while (! stop && Settings.DEFAULT.isRunning()) {
                if (Settings.DEFAULT.getPort () != Settings.DEFAULT.getActualPort ())
                    initSocket ();
                
                if (! Settings.DEFAULT.isActualRunning ())
                    Settings.DEFAULT.setActualRunning0 (true);
                
                packet.setLength (LENGTH);
                socket.receive(packet);
                
                // Check access:
                if(Settings.DEFAULT.getAccess () == Settings.ACCESS_LOCAL) {
                    if (!packet.getAddress ().equals (InetAddress.getLocalHost ())) {
                        TopManager.getDefault ().notify (new NotifyDescriptor.Message (SettingsBeanInfo.getString ("MSG_rejectHost", packet.getAddress())));
                        continue;
                    }
                }
                
                // Try to open the requested file:
                String fileName = new String (packet.getData(), packet.getOffset() + 1, packet.getLength() - 1);
                int lineNumber;
                int index = fileName.indexOf ('@');
                if (index == -1) {
                    lineNumber = -1;
                } else {
                    try {
                        lineNumber = Integer.parseInt (fileName.substring (index + 1)) - 1;
                        fileName = fileName.substring (0, index);
                    } catch (NumberFormatException nfe) {
                        TopManager.getDefault ().notifyException (nfe);
                        lineNumber = -1;
                    }
                }
                
                final boolean wait;
                switch ((char) (packet.getData ()[packet.getOffset()])) {
                case 'Y':
                    wait = true;
                    break;
                case 'N':
                    wait = false;
                    break;
                case 'X':
                    if (stop)
                        break REQUESTS;
                    else
                        continue REQUESTS;
                default:
                    throw new IOException (NbBundle.getBundle (Server.class).getString ("EXC_bad_lead_char"));
                }

                byte res;
                final File f = new File (fileName);
                if (f.exists () && f.isFile ()) {
                    res = 0;
                    final InetAddress addr = packet.getAddress ();
                    final int port = packet.getPort ();
                    // Opening may be slow, incl. prompting for mount point. Should not time out launcher while waiting for user.
                    final int _lineNumber = lineNumber;
                    new Thread (new Runnable () {
                        public void run () {
                            //System.err.println("Server.run: to call open on f=" + f + " wait=" + wait + " addr=" + addr + " port=" + port + " lineNumber=" + _lineNumber);
                            OpenFile.open (f, wait, addr, port, _lineNumber);
                        }
                    }).start ();
                } else {
                    // Do not time out launcher while waiting for user to press OK.
                    // XXX SwingUtilities.invokeLater?
                    final String _fileName = fileName;
                    new Thread (new Runnable () {
                        public void run () {
                            TopManager.getDefault ().notify (new NotifyDescriptor.Message (SettingsBeanInfo.getString ("MSG_fileNotFound", _fileName)));
                        }
                    }).start ();
                    
                    res = 1;
                }

                // send reply (unless we are waiting for file to be saved)
                if (!wait || res != 0) {
                    //System.err.println("Server.run: will send response res=" + res);
                    packet.getData ()[0] = res;
                    packet.setLength (1);
                    socket.send (packet);
                }
            }
        } catch (IOException ioe) {
            ioe.printStackTrace ();
            Settings.DEFAULT.setActualRunning0 (false);
            Settings.DEFAULT.setRunning (false);
        } finally {
            //System.err.println("Server.run: done");
            if(socket != null)
                socket.close ();
            
            socket = null;
            Settings.DEFAULT.setActualRunning0(false);
            Settings.DEFAULT.setActualPort0(0);
        }
    }

    
    /** Registers a callback so that the launcher will be notified when the file is modified & saved & or its pane closed.
     * @param obj the object to wait for
     * @param addr the address to send a message back to
     * @param port the port to send a message back to
     */
    static void waitFor(DataObject dataObject, InetAddress address, int port) {
        waitings.add(new Waiting(dataObject, address, port));
    }
    
    
    /** Notuify waitings while server shuttings down. */
    private static void shutDownWaitings() {
        Waiting[] allWaitings = (Waiting[])waitings.toArray(new Waiting[waitings.size()]);
        
        for(int i = 0; i < allWaitings.length; i++) {
            // Notify laucher server is shutting down.
            allWaitings[i].unWait((byte)2);
        }
        
        waitings.clear();
    }
    
    
    /** Class which which monitors changes of data object and  notifies 
     * wainting launchers when theirs data objects was seved, deleted or closed. */
    private static class Waiting implements PropertyChangeListener, AncestorListener {

        /** DataObject to monitor. */
        private DataObject dataObject;
        
        /** Address of host to notify. */
        private InetAddress address;
        
        /** Port at which to notify. */
        private int port;
        
        /** Reference to editor cookie if any. */
        private EditorCookie editorCookie;
        
        /** Editor pane we listen on. */
        private JEditorPane observedPane;

        
        /** Constructor. */
        public Waiting(DataObject dataObject, InetAddress address, int port) {
            this.dataObject = dataObject;
            this.address = address;
            this.port = port;
            
            init();
        }
       
                
        /** Initializes listener. */
        private void init() {
            dataObject.addPropertyChangeListener(this);
            
            editorCookie = (EditorCookie)dataObject.getCookie(EditorCookie.class);
            
            if(editorCookie == null)
                return;
            
            JEditorPane[] panes = editorCookie.getOpenedPanes();
            
            if(panes == null || panes.length == 0)
                return;

            // Find top component.
            TopComponent topComponent = (TopComponent)SwingUtilities.getAncestorOfClass(TopComponent.class, panes[0]);
            
            if(topComponent == null)
                return;

            topComponent.addAncestorListener(this);
        }
        
        
        /** Provides cleanup. */
        private void clean() {
            dataObject.removePropertyChangeListener(this);
        }

        /** Notifies waiting launcher about saving, deleting or closing data object. */
        public void unWait(byte status) {
            clean();
            
            waitings.remove(this);
            
            if(socket != null) {
                DatagramPacket packet = new DatagramPacket (new byte[] { status }, 1, address, port);
                
                try {
                    socket.send(packet);
                } catch (IOException ioe) {
                    TopManager.getDefault().notifyException(ioe);
                }
            } else {
                TopManager.getDefault().notify(
                    new NotifyDescriptor.Message(SettingsBeanInfo.getString("MSG_serverNotRunningWhenSaved", dataObject.getName()))
                );
            }
        }


        /** Implements <code>PropertyChangeListener</code> interface. */
        public void propertyChange (PropertyChangeEvent evt) {
            if(dataObject != evt.getSource())
                return; 
            
            if(DataObject.PROP_VALID.equals (evt.getPropertyName ())) {
                // If destroyed, report an error.
                if(!dataObject.isValid ()) {
                    unWait((byte)1);
                }
            } else if(DataObject.PROP_MODIFIED.equals (evt.getPropertyName())) {
                // Don't do anything when it *becomes* modified, only when unmodified, i.e. saved.
                if(!dataObject.isModified()) {
                    unWait((byte)0);
                }
            }
        }
        
        /** Dummy implementation of <code>AncestorListener</code> interface method. */
        public void ancestorAdded(AncestorEvent evt) {}

        /** Dummy implementation of <code>AncestorListener</code> interface method. */
        public void ancestorMoved(AncestorEvent evt) {}

        /** Implements <code>AncestorListener</code> interface method. */
        public void ancestorRemoved(AncestorEvent evt) {
            JEditorPane[] panes = editorCookie.getOpenedPanes();
            
            // Note: Doesn't work 100% since org.netbeans.core.windows.frames.(Default)SplittedContainerImpl
            // doesn't fire events properly. Submit bug?
            if(panes == null || panes.length == 0) {
                // It was really closed, not only switched to another workspace.
                // Notify launcher.
                unWait((byte)0);
            }
        }
    } // End of Waiting class.
    
    
    /** Test run. */
//    public static void main (String[] args) {
//      startup ();
//    }

}
