/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.openfile;

import java.io.IOException;
import java.io.ObjectInput;
import java.text.MessageFormat;

import org.openide.options.SystemOption;
import org.openide.util.HelpCtx;
import org.openide.util.Utilities;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;

/** Settings for the Open File utility.
* @author Jesse Glick
*/
public class Settings extends SystemOption {

    public static final String PROP_RUNNING = "running"; // NOI18N
    public static final String PROP_PORT = "port"; // NOI18N
    public static final String PROP_ACCESS = "access"; // NOI18N
    public static final int ACCESS_LOCAL = 0;
    public static final int ACCESS_ANY = 1;
    public static final String PROP_ACTUAL_RUNNING = "actualRunning"; // NOI18N
    public static final String PROP_ACTUAL_PORT = "actualPort"; // NOI18N

    private static boolean running, actualRunning;
    private static int port, actualPort;

    // Weird nonsense half-copied from HttpServerSettings.
    private static boolean inited, pendingRunning;

    // default settings object, avoids calling the constructor repeatedly
    public static final Settings DEFAULT = (Settings) findObject (Settings.class, true);

    private static final long serialVersionUID = 7655861665922160177L;

    private boolean isGlobal () {
        return true;
    }

    public String displayName () {
        return SettingsBeanInfo.getString ("LBL_openFileSettings");
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (Settings.class);
    }

    protected void initialize () {
        //System.err.println("Settings.initialize");
        // Not what it looks like:
        running = false;
        inited = false;
        port = Main.DEFAULT_PORT;
        putProperty (PROP_ACCESS, new Integer (ACCESS_LOCAL), false);
        actualRunning = false;
        actualPort = 0;
    }

    // The nasty part:
    public boolean isRunning () {
        synchronized (DEFAULT) {
            //System.err.println("Settings.isRunning: inited=" + inited + " running=" + running + " writeExternal=" + isWriteExternal ());
            if (! inited) {
                //System.err.println("Settings.isRunning: will automatically call setRunning(true)");
                inited = true;
                // Only defaults to running on Windoze:
                setRunning (Utilities.isWindows ());
            }
            return running;
        }
    }
    public void setRunning (boolean r) {
        synchronized (DEFAULT) {
            if (isReadExternal ()) {
                //System.err.println("Settings.setRunning (readExternal): r=" + r);
                pendingRunning = r;
            } else {
                //System.err.println("Settings.setRunning: " + running + " -> " + r);
                boolean old = running;
                running = r;
                firePropertyChange (PROP_RUNNING, new Boolean (old), new Boolean (r));
                if (old && ! r)
                    Server.shutdown ();
                else if (! old && r)
                    Server.startup ();
            }
        }
    }
    public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException {
        //System.err.println("Settings.readExternal");
        super.readExternal (in);
        inited = true;
        //System.err.println("Settings.readExternal: pendingRunning=" + pendingRunning);
        setRunning (pendingRunning);
    }

    public int getPort () {
        return port;
    }
    public void setPort (int p) {
	if (p <= 0 || p > 0xFFFF) {
            String msg = MessageFormat.format(SettingsBeanInfo.getString("MSG_portOutOfRange"), // NOI18N
                                              new Object[] { new Integer(p), new Integer(1), new Integer(65535) } );
            TopManager.getDefault().notify(new NotifyDescriptor.Message(msg,NotifyDescriptor.ERROR_MESSAGE));
            return;
        }

        if (p < 1024 && p != port && inited && Utilities.isUnix()) {
            String msg = SettingsBeanInfo.getString("MSG_onlyRootOnUnix"); // NOI18N
            TopManager.getDefault().notify(new NotifyDescriptor.Message(msg,NotifyDescriptor.WARNING_MESSAGE));
        }
        //
        // (Is it suitable to display GUI stuff here?)

        synchronized (DEFAULT) {
            boolean restart = (p != port && isActualRunning ());
            if (restart) Server.shutdown ();
            int old = port;
            port = p;
            firePropertyChange (PROP_PORT, new Integer (old), new Integer (p));
            if (restart && running) Server.startup ();
        }
    }

    public int getAccess () {
        return ((Integer) getProperty (PROP_ACCESS)).intValue ();
    }
    public void setAccess (int a) {
        putProperty (PROP_ACCESS, new Integer (a), true);
    }

    public boolean isActualRunning () {
        return actualRunning;
    }
    void setActualRunning0 (boolean r) {
        synchronized (DEFAULT) {
            //System.err.println("Settings.actualRunning: " + actualRunning + " -> " + r);
            boolean old = actualRunning;
            actualRunning = r;
            firePropertyChange (PROP_ACTUAL_RUNNING, new Boolean (old), new Boolean (r));
        }
    }

    public int getActualPort () {
        return actualPort;
    }
    void setActualPort0 (int p) {
        synchronized (DEFAULT) {
            //System.err.println("Settings.actualPort: " + actualPort + " -> " + p);
            int old = actualPort;
            actualPort = p;
            firePropertyChange (PROP_ACTUAL_PORT, new Integer (old), new Integer (p));
        }
    }

}
