/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.search;


import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.*;
import javax.swing.DefaultSingleSelectionModel;
import javax.swing.SingleSelectionModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.openide.nodes.Node;
import org.openide.TopManager;
import org.openide.util.HelpCtx;
import org.openidex.search.SearchType;

import org.netbeans.modules.search.res.Res;
import org.netbeans.modules.search.types.FullTextType;
import org.netbeans.modules.search.types.ObjectNameType;
import org.netbeans.modules.search.types.ObjectTypeType;
import org.netbeans.modules.search.types.ModificationDateType;


/**
 * Holds data of customized criteria. Collects
 * <code>CriterionModel</code>'s. Provides context help switching.
 *
 * <p>Presumtion: View is organized as TabbedPane so help
 * context switching is supported because criteria are
 * determined at runtime.
 *
 * <p>Listens: at each criterion model for PROP_CUSTOMIZED
 * <p>Fires: PROP_CUSTOMIZED if any criterion fired it
 *
 * @author  Petr Kuzel
 */
public class CriteriaModel implements PropertyChangeListener, ChangeListener {
    /** Set of seatch type classes. */
    private Set criteriaClasses;

    /** The starting point of this criteria. */
    private Node[] nodes;

    /** Current map SearchType.class -> current <code>CriterionModel</code>. */
    private HashMap criteria;

    /** Vector of <code>CriterionModel</code>'s. */
    private Vector orderedCriteria;

    /** Utility field used by bound properties. */
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    /** Selection model. */
    private SingleSelectionModel tabModel;

    /** Customized property name. */
    public static final String PROP_CUSTOMIZED = "customized"; // NOI18N

    /** Whether some criterion is customized. */
    private boolean customized = false;

    
    /** Creates new <code>CriteriaModel</code>.
     * @param nodes starting point.
     * @param classes search types available.
     */
    public CriteriaModel(Node[] nodes, Set classes) {
        this.nodes = nodes;
        criteriaClasses = classes;
        criteria = new HashMap();
        orderedCriteria = new Vector();
        tabModel = new DefaultSingleSelectionModel();
        tabModel.addChangeListener(this);

        // default values of criteria
        Iterator it = classes.iterator();

        while (it.hasNext()) {
            Class next = (Class) it.next();

            CriterionModel cust = new CriterionModel(this, next);
            cust.addPropertyChangeListener(this);
            criteria.put(next, cust);
        }

        initOrderedCriteria();
    }

    
    /** Initialized the vector. Includes only supported search types in
     * order defined by search services order. */
    private void initOrderedCriteria() {
        Enumeration en = TopManager.getDefault().getServices().services(SearchType.class);

        while (en.hasMoreElements()) {
            Class next = en.nextElement().getClass();

            if (criteriaClasses.contains(next)) {
                CriterionModel assoc = (CriterionModel) criteria.get(next);
               if ( ! orderedCriteria.contains(assoc) ) {
                    orderedCriteria.add(assoc);
                }
            }
            
            Collections.sort(orderedCriteria, new CriterionModelComparator());
        }
    }

    /** Query models handling customization.
     * @return properly ordered <code>CriterionModel</code>'s. */
    public Iterator getCustomizers() {
        return orderedCriteria.iterator();
    }


    /** Query modified criteria.
     * @return current state of customized criteria. */
    public SearchType[] getCustomizedCriteria() {
        // convert entered criteria to array
        int size = 0;

        Iterator it = criteria.values().iterator();
        while(it.hasNext()) {
            CriterionModel next = (CriterionModel) it.next();
            if (next.isCustomized()) size++;
        }

        SearchType[] criteriaArray = new SearchType[size];

        int index = 0;

        Iterator it2 = criteria.values().iterator();
        while(it2.hasNext()) {
            CriterionModel next = (CriterionModel) it2.next();
            if (next.isCustomized()) {
                criteriaArray[index++] = (SearchType) next.getCriterion();
            }
        }

        return criteriaArray;
    }

    /** @return node[] representing starting nodes. */
    public Node[] getNodes() {
        return nodes;
    }

    /** @return true if some criterion customized. */
    public boolean isCustomized() {
        return customized;
    }

    /** @return name of criterion at index is modified. */
    public String getTabText(int index) {
        try {
            return ((CriterionModel) orderedCriteria.get(index)).getName();
        } catch (ArrayIndexOutOfBoundsException ex) {
            return null;
        }
    }

    /** Add a PropertyChangeListener to the listener list.
     * @param l listener to add.
     */
    public void addPropertyChangeListener(java.beans.PropertyChangeListener l) {
        propertyChangeSupport.addPropertyChangeListener(l);
    }

    /** Removes a PropertyChangeListener from the listener list.
     * @param l The listener to remove.
     */
    public void removePropertyChangeListener(java.beans.PropertyChangeListener l) {
        propertyChangeSupport.removePropertyChangeListener(l);
    }

    /** Listens for "customized" from all included <code>CriterionModel</code>'s. 
     * Implemenst <code>PropertyChangeListener</code> interface. */
    public void propertyChange(final java.beans.PropertyChangeEvent event) {
        if (CriterionModel.PROP_CUSTOMIZED.equals(event.getPropertyName())) {
            customized = getCustomizedCriteria().length != 0;

            Boolean oldVal = null; // must be null
            Boolean newVal = new Boolean(customized);

            propertyChangeSupport.firePropertyChange(PROP_CUSTOMIZED, oldVal, newVal);
        }
    }

    /** View support method - it provides milked SingleSelectionModel.
     * Context help is assigned according it.
     * @return SingleSelectionModel that is monitored by this.
     */
    public SingleSelectionModel getTabModel() {
        return tabModel;
    }

    /** Overrides suprclass method.
     * @return this as string. */
    public String toString() {
        String ret;
        ret = "CriteriaModel: \n"; // NOI18N
        Iterator it = criteria.values().iterator();
        while(it.hasNext()) {
            ret = ret + it.next().toString() + "\n"; // NOI18N
        }
        ret = ret + "/CriteriaModel"; // NOI18N

        return ret;
    }

    /** The context help must be changed.
    */
    public void stateChanged(ChangeEvent e) {
        propertyChangeSupport.firePropertyChange("help", null, getHelpCtx()); // NOI18N
    }

    /** Return current HelpCtx. */
    public HelpCtx getHelpCtx() {
        int index = tabModel.getSelectedIndex();
        CriterionModel model = (CriterionModel) orderedCriteria.get(index);
        return model.getHelpCtx();
    }


    /** Comparator which compares the criterion models showing as tabs in search dialog in this order
     * <p>1st <code>FullTextType</code>
     * <p>2nd <code>ObjectNameType</code>
     * <p>3rd <code>ObjectTypeType</code>
     * <p>4th <code>ModificationDateType</code>
     * <p>rest custom criterion's ordered according their names.
     */
    private static class CriterionModelComparator implements Comparator {

        /** Implements <code>Comparator</code> interface method. */
        public int compare(Object obj1, Object obj2) {
            if(obj1 == obj2)
                return 0;
            
            if(obj1 == null)
                return 1;
            
            if(obj2 == null)
                return -1;
            
            // Has to be criterion model type.
            CriterionModel cm1 = (CriterionModel)obj1;
            CriterionModel cm2 = (CriterionModel)obj2;

            if(cm1.getCriterion().getClass().equals(FullTextType.class))
                return -1;
            
            if(cm2.getCriterion().getClass().equals(FullTextType.class))
                return 1;
            
            if(cm1.getCriterion().getClass().equals(ObjectNameType.class))
                return -1;
            
            if(cm2.getCriterion().getClass().equals(ObjectNameType.class))
                return 1;

            if(cm1.getCriterion().getClass().equals(ObjectTypeType.class))
                return -1;

            if(cm2.getCriterion().getClass().equals(ObjectTypeType.class))
                return 1;
            
            if(cm1.getCriterion().getClass().equals(ModificationDateType.class))
                return -1;

            if(cm2.getCriterion().getClass().equals(ModificationDateType.class))
                return 1;

            return cm1.getName().compareTo(cm2.getName());
        }
        
        /** Implements <code>Comparator</code> interface method. */
        public boolean equals(java.lang.Object obj) {
            if(this == obj)
                return true;
            else
                return false;
        }
        
    } // End of CriterionModelComparator class.
    
    
    /** For debugging purposes only. */
    static private final boolean TRACE = false;
    /** For debugging purposes only. */
    static private void t(String str) {
        if (TRACE)
            System.out.println("CriteriaModel> "+str); // NOI18N
    }
    
}
