/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.search;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

import org.openide.*;
import org.openide.util.HelpCtx;

import org.netbeans.modules.search.res.*;


/**
 * Creates tabbed dialog based on passed criteria. 
 * Mark changed criteria as customized.
 *
 * <p>Listens: PROP_CUSTOMIZED on model
 *
 * @author  Petr Kuzel
 * @version 1.0
 */
public class CriteriaView extends javax.swing.JPanel implements PropertyChangeListener
{
    /** A return status code - returned if Cancel button has been pressed */
    public static final int RET_CANCEL = 0;
    /** A return status code - returned if OK button has been pressed */
    public static final int RET_OK = 1;

    /** This shows state of: */
    private CriteriaModel model;

    private JTabbedPane pane;

    /** Nb equalent for dialog. */
    private DialogDescriptor desc;
    private JButton okButton;
    private JButton cancelButton;

    /** Java equavalent. */
    private Dialog dlg;

    private int returnStatus = RET_CANCEL;

    /** Creates new form CriteriaView upon model */
    public CriteriaView(CriteriaModel cModel) {
        model = cModel;

        initComponents();

        setName(Res.text("TITLE_CUSTOMIZE")); // NOI18N

        okButton = new JButton(Res.text("BUTTON_SEARCH")); // NOI18N
//        okButton.setIcon (Res.icon("SEARCH")); // NOI18N
        okButton.setEnabled(model.isCustomized());

        cancelButton = new JButton(Res.text("BUTTON_CANCEL")); // NOI18N

        Object options[] = new Object[] { okButton, cancelButton };

        //create representing dialog descriptor
        desc = new DialogDescriptor(this, getName(), true, options, options[0],
                                    DialogDescriptor.BOTTOM_ALIGN, model.getHelpCtx(), 
                                    new AL());
    }

    /** Create dynammically set of tabs representing criteria customizers. */
    private void initComponents() {
        setLayout(new BorderLayout());
        setBorder(new EmptyBorder(2,5,8,5));

        pane = new JTabbedPane();
        pane.setModel(model.getTabModel());
        pane.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                Component comp = getTypeCustomizer(pane.getSelectedIndex());
                if (comp != null)
                    comp.requestFocus();
            }
        });

        // for each criterion create one tab
        Iterator it = model.getCustomizers();

        while(it.hasNext()) {
            CriterionModel next = (CriterionModel) it.next();
            Component comp = new CriterionView(next);

            try {
                pane.add(comp);
            } catch (ArrayIndexOutOfBoundsException ex) {
                //TODO why? a bug in JTabbedPane?
            }
        }

        add(pane, BorderLayout.CENTER);
    }

    /** @return the return status of this dialog - one of RET_OK or RET_CANCEL */
    public int getReturnStatus () {
        return returnStatus;
    }


    private void doClose (int retStatus) {
        model.removePropertyChangeListener(this);
        returnStatus = retStatus;

        dlg.setVisible(false);
        dlg.dispose();
    }

    /** Wrap itself to DialogDescriptor and show itself.
    */ // I have overwritten deprecated method -> Ignore such warning.
    public void show()  {
        model.addPropertyChangeListener(this);

        dlg = TopManager.getDefault().createDialog(desc);
        dlg.setModal(true);
        dlg.addComponentListener(new java.awt.event.ComponentAdapter() {
            public void componentShown(java.awt.event.ComponentEvent e) {
                int i = pane.getSelectedIndex();
                if (i < 0) i = 0;
                Component comp = getTypeCustomizer(i);
                if (comp != null)
                    comp.requestFocus();
                dlg.removeComponentListener(this);
            }
        });
        dlg.pack();
        dlg.show();
    }

    public void propertyChange(PropertyChangeEvent event) {
        if (CriteriaModel.PROP_CUSTOMIZED.equals(event.getPropertyName())) {
            okButton.setEnabled(model.isCustomized());
        }
        else if (event.getPropertyName().equals("help")) {
            desc.setHelpCtx( (HelpCtx) event.getNewValue());
        }

        for (int i = 0; i<pane.getTabCount(); i++ ) {
            pane.setTitleAt(i, model.getTabText(i));
            
            pane.setIconAt(i, null);
        }
    }

    private Component getTypeCustomizer(int index) {
        CriterionModel cm = null;
        Iterator it = model.getCustomizers();
        while (index >= 0 && it.hasNext()) {
            cm = (CriterionModel) it.next();
            index--;
        }
        return cm != null ? cm.getComponent() : null;
    }

    /** Listen for provided options*/
    private class AL implements ActionListener {

        public void actionPerformed(final java.awt.event.ActionEvent e) {
            if (e.getSource() == okButton) {
                doClose(RET_OK);
            } else {
                doClose(RET_CANCEL);
            }
        }
    }
}
