/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.search;

import java.awt.*;
import java.beans.*;
import java.util.*;

import org.openide.*;
import org.openide.nodes.*;
import org.openide.util.*;

import org.openidex.search.*;

import org.netbeans.modules.search.res.*;


/**
 * Represent one criterion during customization. The criterion
 * may be in different states: [un]modified. Also links the criterion
 * with its customizer and storage of own presets.
 *
 * <p>Listens: <every property on customized bean>
 * <p>Fires: PROP_CUSTOMIZED, PROP_RESTORED
 *
 * @author  Petr Kuzel
 * @version  
 */
public class CriterionModel implements PropertyChangeListener
{
    /** Customized property name. */
    public static final String PROP_CUSTOMIZED = "customized";
    public static final String PROP_RESTORED = "restored";
    private boolean customized;

    private static final String MODIFICATOR_SUFFIX = " *";

    private SearchType criterion;
    private CriteriaModel parent;
    private Class type;

    // beaninfo of criterion and customizer obtained from it
    private BeanInfo beanInfo;
    private Customizer customizer;
    private Component customizerComponent;

    /** Utility field used by bound properties. */
    private PropertyChangeSupport propertyChangeSupport =
        new PropertyChangeSupport(this);

    /** Creates new PresetModel */
    public CriterionModel(CriteriaModel parent, Class type) {
        this.parent = parent;
        this.type = type;

        customized = false;
        criterion = getInstance(type, null);

        try {
            beanInfo = org.openide.util.Utilities.getBeanInfo(type);

            if (hasCustomizer()) {
                customizer = getCustomizer();
                customizerComponent = (Component) customizer;
            } else {
                // TODO use property sheet as it will implement Customizer
                // allow hiding tabs, ....
                System.err.println("No customizer for " + criterion.getName() + ", skipping...");
            }
        } catch (java.beans.IntrospectionException ex) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                ex.printStackTrace();
        }

        customizer.setObject(criterion);
        criterion.addPropertyChangeListener(this);
    }

    public boolean hasCustomizer () {
        // true if we have already computed beanInfo and it has customizer class
        return beanInfo.getBeanDescriptor().getCustomizerClass() != null;
    }

    /**
    * Get name of this criterion used as tab name, 
    * it is mangled if the criterion is customized.
    * @return my name. 
    */
    public String getName() {
        String name;
        if (criterion.getTabText() == null)
            name = criterion.getDisplayName();
        else
            name = criterion.getTabText();

        if (isCustomized())
            return  name + MODIFICATOR_SUFFIX;
        else
            return  name;
    }

    /** @return relaled customizer compoment. */
    public Component getComponent() {
        return customizerComponent;
    }

    /** @return customizer object. */
    public Customizer getCustomizer() {
        if (customizer != null) return customizer;

        Class clazz = beanInfo.getBeanDescriptor ().getCustomizerClass ();
        if (clazz == null) return null;

        Object o;
        try {
            o = clazz.newInstance ();
        } catch (InstantiationException e) {
            return null;
        } catch (IllegalAccessException e) {
            return null;
        }

        if (!(o instanceof java.awt.Component) ||
                !(o instanceof java.beans.Customizer)) return null;

        return (Customizer) o;
    }

    public CriteriaModel getParent() {
        return parent;
    }

    public Class getCriterionClass() {
        return criterion.getClass();
    }

    public void restoreCriterion(String name) {
        criterion.removePropertyChangeListener(this);

        criterion = getInstance(type, name);
        getCustomizer().setObject(criterion);
        criterion.addPropertyChangeListener(this);

        setCustomized(true);
        propertyChangeSupport.firePropertyChange(PROP_RESTORED, null, name);
    }

    /**
    * @return true if new value was created
    */
    public boolean saveAs(String name) throws IllegalArgumentException {
        boolean savedNew;

        SearchType copy = (SearchType)criterion.clone();
        copy.setName(name);

        // overwrite existing
        if (Registry.exist(copy)) {
            Registry.remove(copy);
            savedNew = false;
        }
        else savedNew = true;

        Registry.append(copy);

        return savedNew;
    }

    public void propertyChange(PropertyChangeEvent event) {
        if (event.getSource() == criterion) {
            // if Type fires valid property change listens for
            // its invalidity -> mark itself as unmodified
            if (SearchType.PROP_VALID.equals(event.getPropertyName()) ) {
                if (event.getNewValue().equals(new Boolean(false))) {
                    setCustomized(false);
                }
            } 
            else { // any modification
                setCustomized(true);
            }
        }
    }


    /** Adds a PropertyChangeListener to the listener list.
     * @param l The listener to add.
     */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        propertyChangeSupport.addPropertyChangeListener(l);
    }

    /** Removes a PropertyChangeListener from the listener list.
     * @param l The listener to remove.
     */
    public void removePropertyChangeListener(java.beans.PropertyChangeListener l) {
        propertyChangeSupport.removePropertyChangeListener(l);
    }

    public boolean isCustomized () {
        return customized;
    }

    public void setCustomized(boolean cust) {
        if (customized != cust) {
            customized = cust;
            propertyChangeSupport.firePropertyChange(PROP_CUSTOMIZED, !cust, cust);
        }
    }

    /** Return currently hold bean. */
    public SearchType getCriterion () {
        return criterion;
    }


    /** Class equality
    * @return this.bean.getClass().equals(bean.getClass());
    */
    public boolean equals(Object obj) {
        try {
            return criterion.getClass().equals(((CriterionModel)obj).getCriterionClass());
        } catch (ClassCastException ex) {
            return false;
        }
    }


    /**
    * @param type class os SearchType which instance is required
    * @param name of instance (from control panel)
    * @return preset instance of search type or default instance if not found
    */
    private SearchType getInstance(Class type, String name) {
        // use named instance
        if ( type != null ) {
            Enumeration en = TopManager.getDefault().getServices().services(type);

            while (en.hasMoreElements()) {
                SearchType next = (SearchType) en.nextElement();

                if (next.getName().equals(name)) {
                    return (SearchType) next.clone();
                }
            }
        }

        // otherwise use default instance
        Object ret = null;

        try {
            ret = type.newInstance();
        } catch (InstantiationException ex) {
            // ignore such search type
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                ex.printStackTrace();
        } catch (IllegalAccessException ex) {
            // ignore such search type
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                ex.printStackTrace();
        }

        return (SearchType) ret;
    }

    public String toString() {
        return "CriterionModel: " + type + criterion; // NOI18N
    }

    public HelpCtx getHelpCtx() {
        return criterion.getHelpCtx();
    }


    /** For debugging purposes only. */
    static private final boolean TRACE = false;
    /** For debugging purposes only. */
    static private void t(String str) {
        if (TRACE)
            System.out.println("CriterionModel> "+str); // NOI18N
    }
}
