/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.search;

import java.beans.*;

import org.openide.actions.*;
import org.openide.util.actions.*;
import org.openide.explorer.*;
import org.openide.windows.*;

/**
 * Hooks search presenter to FindAction.
 *
 * Listens at TopComponent.Registry on CURRENT_NODES property.
 * On change evaluates all registered search types and
 * if some of them is enabled enable the Find action.
 *
 * @author  Petr Kuzel
 * @version 1.0
 */
public class SearchHook implements PropertyChangeListener {

    private SearchPerformer performer;
    private static SearchHook theHook = null;

    /** Creates new SearchHook
    * @param performer knows if search type is enabled and so on.
    * If null is passed hook() does nothing
    */
    public SearchHook(SearchPerformer performer) {
        this.performer = performer;
    }

    /** Hook performer at FindAction.
    * Conditionally hooks performer to FindAction.
    * Condition: active top component is ExplorerManager.Provider AND
    * some criteria is enabled be current nodes
    */
    public void hook() {
        setHookListener(this);
//        TopComponent.getRegistry().addPropertyChangeListener(this);
    }

    /**
    * Finnish hooking job. After invoking, this class is useless.
    */
    public void unhook() {
        performer = null;
        setHookListener(null);
//        hook();
//        TopComponent.getRegistry().removePropertyChangeListener(this);
    }

    private static void setHookListener(SearchHook hook) {
        TopComponent.Registry regs = TopComponent.getRegistry();

        if (theHook != null) {
            regs.removePropertyChangeListener(theHook);
        }

        if (hook != null) {
            regs.addPropertyChangeListener(hook);
        }

        theHook = hook;
    }

    /** Determine if activated topcomponent represent Editor
    * if not call overwriting routine.
    * It prevents battle about FindAction performer.
    */
    private void someoneActivated() {
        TopComponent activated = TopComponent.getRegistry().getActivated();

        if (! (activated instanceof org.openide.text.EditorSupport.Editor) ) {
            overwriteFindPerformer();
        }
    }

    /** Set performer as performer if no performer is set.
    */
    private void overwriteFindPerformer() {
        CallbackSystemAction action = (CallbackSystemAction) SystemAction.get(FindAction.class);

        if ( action == null ) { // hook target does not exist
            throw new RuntimeException("Should not happen: Cannot locate FindAction."); // NOI18N
        }

        if (performer == null) { // handle deinstalation
            action.setActionPerformer(null);
            return;
        }

        Object currPerformer = action.getActionPerformer();

        if (performer.enabled(TopComponent.getRegistry().getCurrentNodes())) {
            action.setActionPerformer(performer);
        } else if (currPerformer != null && currPerformer.getClass().equals(SearchPerformer.class)) {
            action.setActionPerformer(null);
        }
    }


    /** Be interested in Current_nodes property change.
    */
    public void propertyChange(final java.beans.PropertyChangeEvent p1) {
        if (p1.getPropertyName() == TopComponent.Registry.PROP_CURRENT_NODES ) {

            // PREVENT deadlock
            // do not depend on locking order of
            // CookieSet[.add/remove()] and TopComponent[.attach()]
            org.openide.util.RequestProcessor.postRequest( new Runnable() {
                        public void run() {
                            someoneActivated();
                        }
                    });

        }
    }

    /**
    * @param args the command line arguments
    */
    public static void main (String args[]) {
        SearchHook me = new SearchHook(new SearchPerformer());
        me.hook();
    }
}
