/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.search.scanners;


import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;

import org.openide.*;
import org.openide.nodes.*;
import org.openide.loaders.*;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.*;
import org.openide.util.enum.*;
import org.openidex.search.*;

import org.netbeans.modules.search.types.*;
import org.netbeans.modules.search.*;
import org.netbeans.modules.search.res.Res;


/**
 * Scanner that can scan all collection nodes in repository.
 * It tests DataObjectType by <code>test(DataObject)</code>.
 *
 * @see DataObjectType
 * @author  Petr Kuzel
 */
public class RepositoryScanner extends Scanner {
    
    /** DataObjectTypes to be applied */
    private Vector dot;
    
    /** Criteria. */
    private DataObjectType[] criteria;

    /** Cache of file objects to skip due they were recognized */
    private HashSet cache;

    /** Not used detail object can be reused */
    private SearchDetail reuseDetail = null;

    
    /** Creates new RepositoryScanner */
    public RepositoryScanner() {
        dot = new Vector();
        cache = new HashSet();
    }

    
    /** Add new search type. The type will be applied on
     * scanned object if all other such added types are eveluated to true. */
    public void add(SearchType type) {

        t("add(" + type + ")"); // NOI18N

        if (type instanceof DataObjectType) {
            dot.add(type);
        } else {
            throw new ClassCastException("Unsupported search type"); // NOI18N
        }
    }

    /** Scan particular nodes. Test if their children
     * match added() search types. */
    public void scan(Node[] nodes) {

        t("scan()"); // NOI18N

        try {

            criteria = new DataObjectType[dot.size()];
            dot.copyInto(criteria);

            nodes = normalizeNodes(nodes);

            // test whether scan whole repository

            if (nodes.length == 1) {
                InstanceCookie ic = (InstanceCookie)nodes[0].getCookie (InstanceCookie.class);

                try {
                  if (ic != null && Repository.class.isAssignableFrom (ic.instanceClass ())) {
                      scanRepository();
                      return;
                  }
                } catch (IOException ex) {
                } catch (ClassNotFoundException ex) {
                }
            }

            for (int i = 0; i<nodes.length; i++) {
                Node.Cookie cake = nodes[i].getCookie(org.openide.loaders.DataFolder.class);
                if ( cake != null) {
                    DataFolder folder = (DataFolder) cake;
                    scan(folder.getPrimaryFile());
                }
            }

        } catch ( InterruptedException ex) {
            // finnish scanning
            t("interrupted."); // NOI18N
        }
    }

    /** Scan all visible filesystems. */
    private void scanRepository() throws InterruptedException {
        Repository rep = TopManager.getDefault().getRepository();
        Enumeration fss = rep.getFileSystems();
        while (fss.hasMoreElements()) {
            org.openide.filesystems.FileSystem fs = (org.openide.filesystems.FileSystem)fss.nextElement();
            if (fs.isValid() && !fs.isHidden())
                scan(fs.getRoot());
        }
    }

    /**
     * Scan recursively particular FileObject.
     * Fires found notifications. */
    private void scan(FileObject fo) throws InterruptedException {

        if (Thread.interrupted()) throw new InterruptedException();

        if (fo.isData()) { //leaf
            t("scanning " + fo); // NOI18N
            DataObject dobj = getDataObject(fo);

            if (dobj == null) return;

            SearchDetail detail;
            if (reuseDetail != null) {
                detail = reuseDetail;
                reuseDetail = null;
            }
            else detail = new SearchDetail();

            // test obtained object on all criteria
            for (int i = 0; i<criteria.length; i++) {
                if ( ! criteria[i].test(dobj, detail) ) return;
            }

            Node node = dobj.getNodeDelegate();

            t("Found node: "+ node); // NOI18N
            if (node != null) {
                // mark with detail cookie
                notifyFound( new Node[] {
                                 new FoundNode (node, detail.isEmpty () ? null : detail)
                             } );
            }

            if (detail.isEmpty()) reuseDetail = detail;
            return;

        } else { //data folder
            FileObject[] fos = fo.getChildren();
            if (fos.length == 0) return;
            for(int i=0; i<fos.length; i++)
                scan(fos[i]);
        }
    }

    /** Caching model: cache holds recognized DataObjects */
    private boolean cacheHit(DataObject dobj) {
        return cache.contains(dobj);
    }

    /** Cache add. */
    private void cacheAdd(DataObject dobj) {
        cache.add(dobj);
    }


    /**
     * Translates FileObject to DataObject.
     * @return DataObject or null if it was already recognized */
    private DataObject getDataObject(FileObject fo) {

        DataObject dobj = null;

        try {
            try {
                dobj = org.openide.loaders.DataObject.find(fo);
            }
            catch (DataObjectNotFoundException ex1) {
                t("Creating dobj from scratch"); // NOI18N
                dobj = TopManager.getDefault().getLoaderPool().findDataObject(fo);
            }

            if (cacheHit(dobj)) {
                return null;
            } else {
                cacheAdd(dobj);
            }

        } catch (DataObjectExistsException ex) {
            throw new RuntimeException("Should not occur."); // NOI18N

        } catch (IOException ex) {
            if (Boolean.getBoolean ("netbeans.debug.exceptions")) // NOI18N
                ex.printStackTrace();
        }

        return dobj;

    }

    /** Remove kids from set. */
    private Node[] normalizeNodes(Node[] nodes) {

        Vector ret = new Vector();

        for (int i = 0; i<nodes.length; i++) {
            if (! hasParent(nodes[i],nodes)) ret.add(nodes[i]);
        }

        Node[] newNodes = new Node[ret.size()];
        ret.copyInto(newNodes);
        return newNodes;
    }

    /** Has parent. */
    private boolean hasParent(Node node, Node[] nodes) {
        for (Node parent = node.getParentNode(); parent != null; parent = parent.getParentNode()) {
            for (int i = 0; i<nodes.length; i++) {
                if (nodes[i].equals(parent)) return true;
            }
        }
        return false;
    }

    /** The name given to my thread. */
    public String toString() {
        return "RepositoryScanner"; // NOI18N
    }


    /** Name of found node destroyed property. */
    public static final String PROP_NODE_DESTROYED = "foundNodeDestroyed"; // NOI18N
    /** Name of found node valid property. */
    public static final String PROP_NODE_VALID = "foundNodeValid"; // NOI18N
    

    /** Node that may know search result details. */
    public static class FoundNode extends FilterNode implements PropertyChangeListener {

        /** DeatailCookie for this node. */
        private DetailCookie detail;

        /** Original data object if there is any. */
        private DataObject originalDataObject;
        
        /** Name of original file object. */
        private String originalFileObjectName;
        
        
        /** Constructs node. */
        public FoundNode(Node original, DetailCookie detail) {
            super(original);
            this.detail = detail;
            this.originalDataObject = (DataObject) original.getCookie (DataObject.class);
            
            if (originalDataObject == null)
                return; 

            this.originalDataObject.addPropertyChangeListener(this);

            this.originalFileObjectName = this.originalDataObject.getPrimaryFile().getPackageNameExt('/','.');
            
            FileObject fileFolder = originalDataObject.getPrimaryFile().getParent();
            if(fileFolder != null) {
                String packageName = fileFolder.getPackageName ('.');
                String hint;
                if(packageName.equals("")) // NOI18N
                    hint = Res.hint("result_default_package"); // NOI18N
                else
                    hint = MessageFormat.format(Res.hint ("result_package"), // NOI18N
                         new Object[] { packageName });

                disableDelegation (DELEGATE_SET_SHORT_DESCRIPTION |
                                   DELEGATE_GET_SHORT_DESCRIPTION);
                setShortDescription (hint);
            }

        }
        
        /** Destroys node. Overrides superclass method. */
        public void destroy() throws IOException {
            super.destroy();
            
            if(originalDataObject != null)
                originalDataObject.removePropertyChangeListener(this);
            
            firePropertyChange(PROP_NODE_DESTROYED, this, null);
        }
        
        /** Gets cookie. Overrides superclass method. */
        public Node.Cookie getCookie(Class type) {
            if (type.equals(DetailCookie.class) && detail != null) return detail;
            return super.getCookie(type);
        }
        
        
        /** Gets original file object name. */
        public String getOriginalFileObjectName() {
            return originalFileObjectName;
        }
        
        /** Implements <code>PropertyChangeListener</code>. */
        public void propertyChange(PropertyChangeEvent evt) {
            if(DataObject.PROP_PRIMARY_FILE.equals(evt.getPropertyName())) {
                // primary file has been changed
                firePropertyChange(PROP_NODE_VALID, this, null);
            } else if (DataObject.PROP_VALID.equals(evt.getPropertyName())) {
                // data object might be deleted
                if (!originalDataObject.isValid()) { // link becomes invalid
                    if(originalDataObject != null)
                        originalDataObject.removePropertyChangeListener(this);
                    
                    firePropertyChange(PROP_NODE_VALID, this, null);
                    //children.update();
                }
            }
        }
        
    } // End of FoundNode class.

    
    /** For debugging pusposes only. */
    private boolean TRACE = false;
    
    /** For debugging pusposes only. */
    private void t(String msg) {
        if (TRACE)
            System.err.println("RepositoryScanner> " + msg);
    }
}
