/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */


package org.netbeans.modules.search.types;


import javax.swing.JEditorPane;
import javax.swing.SwingUtilities;
import javax.swing.text.Caret;

import org.openide.cookies.EditorCookie;
import org.openide.cookies.LineCookie;
import org.openide.loaders.DataObject;
import org.openide.text.Line;


/** 
 * Holds details about one search hit in the text document.
 *
 * @author Tomas Pavek
 * @see DetailHandler
 */
public class TextDetail implements DetailHandler {
    /** Full description of detail. */
    private String fullDesc;
    /** Short descriptions of detail. */
    private String shortDesc;
    
    /** Data object. */
    private DataObject dobj;
    /** Line number where search result occures.*/
    private int line;
    /** Column where search result starts. */
    private int column;
    
    /** Length of search result which to mark. */
    private int markLength;
    /** Line. */
    private Line lineObj;
    
    
    /** Constructor without data object. */
    public TextDetail() {
        this(null);
    }

    /** Constructor using data object. */
    public TextDetail(DataObject obj) {
        dobj = obj;
        fullDesc = null;
        shortDesc = null;
        line = 0;
        column = 0;
        markLength = 0;
        lineObj = null;
    }

    /** Shows the search detail on the DataObject. Implements <code>DetailHandler</code> interface method
     * The document is opened in the editor, the caret is positioned on the right line and column 
     * and searched string is marked.
     * @param how indicates how to show detail. 
     * @see DetailHandlet#DH_GOTO 
     * @see DetailHandler#DH_SHOW 
     * @see DetailHandler#DH_HIDE */
    public void showDetail(int how) {
        if (dobj == null) return;
        
        if (lineObj == null) { // try to get Line from DataObject
            LineCookie lineCookie = (LineCookie)dobj.getCookie(LineCookie.class);
            if (lineCookie != null)
                lineObj = lineCookie.getLineSet().getOriginal(line-1);

            if (lineObj == null) return;
        }

        if (column > 0) { // exact position on the line is known
            if (how == DH_HIDE) return;

            if (how == DH_SHOW)
                lineObj.show(Line.SHOW_TRY_SHOW, column-1);
            else if (how == DH_GOTO)
                lineObj.show(Line.SHOW_GOTO, column-1);

            if (markLength > 0) {
                EditorCookie edCookie = (EditorCookie)dobj.getCookie(EditorCookie.class);
                if (edCookie != null) {
                    JEditorPane[] panes = edCookie.getOpenedPanes();
                    if (panes != null && panes.length > 0) {
                        final Caret caret = panes[0].getCaret();

                        // Necessary since above lineObj.show leads to invoke later as well.
                        SwingUtilities.invokeLater(new Runnable() {
                            public void run() {
                                caret.moveDot(caret.getDot() + markLength);
                            }
                        });
                    }
                }
            }
        } else { // go to the line without column specified
            if (how == DH_SHOW) {
                lineObj.show(Line.SHOW_TRY_SHOW, 0);
                lineObj.markCurrentLine();
            } else if (how == DH_GOTO) {
                lineObj.show(Line.SHOW_GOTO, 0);
                lineObj.unmarkCurrentLine();
            } else if (how == DH_HIDE) {
                lineObj.unmarkCurrentLine();
            }
        }
    }

    /** Gets the full description of the detail. Implements <code>DetailHanlder</code> interface method.
     * @return full description or <code>null</code> if no is available */
    public String getDescription() {
        return fullDesc;
    }

    /** Sets the full description of the detail. */
    public void setDescription(String desc) {
        fullDesc = desc;
    }
    
    /** Gets the short description of the detail. Implements <code>DetailHanlder</code> interface method.
     * @return short description or <code>null</code> if no is available */
    public String getShortDescription() {
        return shortDesc;
    }

    /** Sets the short description of the detail. */
    public void setShortDescription(String desc) {
        shortDesc = desc;
    }

    /** Gets the <code>DataObject</code> where the searched text was found. 
     * @return data object or <code>null</code> if no data object is available */
    public DataObject getDataObject() {
        return dobj;
    }

    /** Sets the <code>DataObject</code> where the searched text was found. */
    void setDataObject(DataObject obj) {
        dobj = obj;
    }

    /** Gets the line position of the text. */
    public int getLine() {
        return line;
    }

    /** Sets the line position of the text. */
    public void setLine(int line) {
        this.line = line;
    }

    /** Gets the column position of the text. */
    public int getColumn() {
        return column;
    }

    /** Sets the column position of the text. */
    public void setColumn(int col) {
        column = col;
    }

    /** Set the length of the text that should be marked when the detail is shown. */
    public int getMarkLength() {
        return markLength;
    }

    /** Gets the length of the text that should be marked when the detail is shown. */
    public void setMarkLength(int len) {
        markLength = len;
    }

    /** Gets the <code>Line</code> object for locating the detail in the document. */
    public Line getLineObj() {
        return lineObj;
    }

    /** Sets the <code>Line</code> object for locating the detail in the document. */
    public void setLineObj(Line line) {
        lineObj = line;
    }
}
