/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.search.types;

import java.io.*;

import org.openide.loaders.*;
import org.openide.filesystems.*;

import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

/**
 * Abstract text test.
 *
 * <p>There are mutually exclusive criteria: substring and re.
 * One of substring or re must be null.
 *
 * <p>Internally uses null as wildcard. It is presented as WILDCARD string.
 *
 * @author  Petr Kuzel
 * @version 1.0
 */
public abstract class TextType extends DataObjectType {

    public static final long serialVersionUID = 3L;

    public final String PROP_SUBSTRING = "matchString"; // NOI18N
    public final String PROP_CASE_SENSITIVE = "caseSensitive"; // NOI18N
    public final String PROP_WHOLE_WORDS = "wholeWords"; // NOI18N
    public final String PROP_RE = "RE"; // NOI18N

    /** Holds value of subtring criterion. */
    protected String matchString;
    protected String ciMatchString; // case insensitive match string
    protected boolean caseSensitive;
    protected boolean wholeWords;

    private static String wordSeparators = " \t,;.:+-*/\\()[]{}<>=&|\"\'`~!?@#%^\n\r";

    /** Holds value of re criterion. */
    protected transient RE re;
    protected String reString;


    /** Creates new FullTextType */
    public TextType() {
    }

    /** @return string desribing current state.
    */
    public String toString() {
        return "TextType: substring:" + matchString + " REstring:" + reString + " re:" + re; // NOI18N
    }

    //
    // Match methods
    //

    protected boolean match(String text) {
        return matchString(text, 0) >= 0 && matchRE(text);
    }

    protected boolean matchRE(String line) {
        if ( re == null ) return true;
        return re.match(line);
    }

    protected int matchString(String text, int fromIndex) {
        if (matchString == null) return 0;

        int index = caseSensitive ?
                    text.indexOf(matchString, fromIndex) :
                    text.toUpperCase().indexOf(ciMatchString, fromIndex);

        if (wholeWords && index >= 0) { // test "whole words only" condition
            if (index > 0 && wordSeparators.indexOf(text.charAt(index-1)) < 0)
                index = -1;
            else {
                int matchLen = matchString.length();
                if (index + matchLen < text.length()
                    && wordSeparators.indexOf(text.charAt(index+matchLen)) < 0)
                index = -1;
            }
        }

        return index;
    }


    //
    // Properties
    //


    /** Getter for matchString property.
     *@return Value of property matchString.
     */
    public String getMatchString() {

        if (matchString == null)
            return WILDCARD;
        else
            return matchString;
    }

    /** Setter for matchString.
     * @param matchString New value of criterion substring.
     */
    public void setMatchString(String matchString) {

        if (matchString == null) {
            setValid(false);
            throw new IllegalArgumentException();
        }

        if (matchString.equals(WILDCARD))
            matchString = null;

        String old = this.matchString;
        this.matchString = matchString;
        if (!caseSensitive && matchString != null)
            ciMatchString = matchString.toUpperCase();

        re = null;
        reString = null;
        firePropertyChange(PROP_SUBSTRING, old, matchString);
        firePropertyChange(PROP_RE, null, null);
        setValid(true);
    }

    /** Getter for caseSensitive property.
     * @return value of caseSensitive property
     */
    public boolean getCaseSensitive() {
        return caseSensitive;
    }

    /** Setter for caseSensitive property.
     * @param caseSensitive New value of "case sensitivity" condition.
     */
    public void setCaseSensitive(boolean caseSensitive) {
        boolean old = this.caseSensitive;
        this.caseSensitive = caseSensitive;
        if (!caseSensitive && matchString != null)
            ciMatchString = matchString.toUpperCase();

        firePropertyChange(PROP_CASE_SENSITIVE,
                           new Boolean(old), new Boolean(caseSensitive));
    }

    /** Getter for wholeWords property.
     * @return Value of wholeWords property.
     */
    public boolean getWholeWords() {
        return wholeWords;
    }

    /** Setter for caseSensitive property.
     * @param caseSensitive New value of "whole words only" condition.
     */
    public void setWholeWords(boolean wholeWords) {
        boolean old = this.wholeWords;
        this.wholeWords = wholeWords;
        firePropertyChange(PROP_CASE_SENSITIVE,
                           new Boolean(old), new Boolean(wholeWords));
    }

    /** Getter for property re1.
     *@return Value of property re1.
     */
    public String getRe() {

        if (reString == null)
            return WILDCARD;
        else
            return reString;
    }

    /** Setter for property re1.
     *@param exp New value of criterion re.
     *@throw IllegalArgumentException if not valid regexp or null.
     */
    public void setRe(String re) {
        try {
            setReImpl(re);
            setValid(true);
        } catch (IllegalArgumentException ex) {
            setValid(false);
            throw ex;
        }
    }

    private void setReImpl(String exp) {

        if (exp == null) throw new IllegalArgumentException();

        String old = reString;

        if (exp.equals(WILDCARD)) {
            reString = null;
            exp = null;
            matchString=null;  //MUX
            firePropertyChange(PROP_RE, old, reString);
            firePropertyChange(PROP_SUBSTRING, null, null);
            return;
        }


        try {
            re = new RE(exp);
            reString = exp;
            matchString = null; //MUX
            firePropertyChange(PROP_RE, old, reString);
            firePropertyChange(PROP_SUBSTRING, null, null);

        } catch (RESyntaxException ex) {
            throw new IllegalArgumentException();
        }
    }

}
