/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.actions;

import java.beans.BeanInfo;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;
import javax.swing.JMenuItem;

import org.openide.TopManager;
import org.openide.awt.JInlineMenu;
import org.openide.actions.FileSystemAction;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.actions.SystemAction;
import org.openide.util.actions.Presenter;
import org.openide.util.enum.*;
import org.openide.windows.TopComponent.Registry;

import org.netbeans.modules.vcscore.VcsFileSystem;

/**
 * Action, that contains all VCS actions.
 * @author  Martin Entlicher
 */
public class VcsCommandsAction extends SystemAction implements Presenter.Menu, Presenter.Popup {

    static final long serialVersionUID = -2345116396734900262L;
    /** empty array of menu items */
    static JMenuItem[] NONE = new JMenuItem[] {};

    /** Creates menu for currently selected nodes.
    * @param popUp create popup or normal menu
    */
    static JMenuItem[] createMenu (boolean popUp) {
        Node[] n = TopManager.getDefault().getWindowManager().getRegistry ().getActivatedNodes ();
        Node.Cookie cookie;
        HashMap fsSet = new HashMap ();
        Enumeration result = null;
        

        if (n != null) {
            for (int i = 0; i < n.length; i++) {
                 DataObject obj = (DataObject)n[i].getCookie (DataObject.class);
                 if (obj != null) {
                     try {
                         FileSystem fs = obj.getPrimaryFile ().getFileSystem ();
                         BeanInfo info = org.openide.util.Utilities.getBeanInfo(fs.getClass());
                         info.getBeanDescriptor();
                         Object vcs = info.getBeanDescriptor().getValue(VcsFileSystem.VCS_PROVIDER_ATTRIBUTE);
                         if (!Boolean.TRUE.equals(vcs)) continue;
                         //if (!(fs instanceof VcsFileSystem)) continue;
                         HashSet    foSet = (HashSet) fsSet.get (fs);
                         if (foSet == null ) {
                             fsSet.put (fs,foSet = new HashSet ());
                             foSet.addAll(obj.files ());                     
                         } else { 
                             foSet.addAll(obj.files ());                     
                         }
                     } catch (FileStateInvalidException ex) {
                         continue;
                     } catch (java.beans.IntrospectionException intrexc) {
                         continue;
                     }
                 }  
            }
            /* At present not allowed to construct actions for selected nodes on more filesystems - its safe behaviour
             * If this restriction will be considered as right solution, then code of this method can be simplified
             */
            if (fsSet.size () == 0 || fsSet.size() > 1) return createMenu (EmptyEnumeration.EMPTY, popUp);            
            
            Iterator entrySetIt = fsSet.entrySet ().iterator();
            
            while (entrySetIt.hasNext()) {
                Map.Entry entry = (Map.Entry)entrySetIt.next();
                FileSystem fs = (FileSystem)entry.getKey();
                HashSet    foSet = (HashSet)entry.getValue();

                if (foSet != null) {
                    if (result == null) result = new SequenceEnumeration (new ArrayEnumeration (fs.getActions (foSet)),EmptyEnumeration.EMPTY);
                    else result = new SequenceEnumeration (result,new ArrayEnumeration (fs.getActions (foSet))); 
                }
            }
            
            return createMenu ((result != null) ? result : EmptyEnumeration.EMPTY, popUp);            
        }
        return NONE;
    }

    /** Creates list of menu items that should be used for given
    * data object.
    * @param en enumeration of SystemAction that should be added
    *   into the menu if enabled and if not duplicated
    */
    static JMenuItem[] createMenu (Enumeration en, boolean popUp) {
        en = new RemoveDuplicatesEnumeration (en);

        ArrayList items = new ArrayList ();
        while (en.hasMoreElements ()) {
            SystemAction a = (SystemAction)en.nextElement ();
            
            if (a.isEnabled ()) {
                JMenuItem item = null;
                if (popUp) {
                    if (a instanceof Presenter.Popup) {
                        item = ((Presenter.Popup)a).getPopupPresenter ();
                    }
                } else {
                    if (a instanceof Presenter.Menu) {
                        item = ((Presenter.Menu)a).getMenuPresenter ();
                    }
                }
                // test if we obtained the item
                if (item != null) {
                    items.add (item);
                }
            }
        }
        JMenuItem[] array = new JMenuItem [items.size ()];
        items.toArray (array);
        return array;
    }

    /* @return menu presenter.
    */
    public JMenuItem getMenuPresenter () {
        return new Menu (false);
    }

    /* @return popup presenter.
    */
    public JMenuItem getPopupPresenter () {
        return new Menu (true);
    }

    /* Getter for name
    */
    public String getName () {
        return org.openide.util.NbBundle.getBundle(VcsMountAction.class).getString("CTL_VcsFSActionName");
    }

    /* Getter for help.
    */
    public HelpCtx getHelpCtx () {
        return new HelpCtx (FileSystemAction.class);
    }

    /* Do nothing.
    * This action itself does nothing, it only presents other actions.
    * @param ev ignored
    */
    public void actionPerformed (java.awt.event.ActionEvent e) {}

    /** Presenter for this action.
    */
    private class Menu extends JInlineMenu {
        /** menu presenter (true) or popup presenter (false) */
        private boolean popup;
        /** last registered items */
        private JMenuItem[] last = NONE;
        /** own property change listner */
        private PropL propL = new PropL ();

        static final long serialVersionUID =2650151487189209766L;

        /** Creates new instance for menu/popup presenter.
        * @param popup true if this should represent popup
        */
        Menu (boolean popup) {
            this.popup = popup;
            changeMenuItems (createMenu (popup));

            Registry r = TopManager.getDefault().getWindowManager().getRegistry ();

            r.addPropertyChangeListener (
                org.openide.util.WeakListener.propertyChange (propL, r)
            );
        }

        /** Changes the selection to new items.
        * @param items the new items
        */
        synchronized void changeMenuItems (JMenuItem[] items) {
            removeListeners (last);
            addListeners (items);
            last = items;
            setMenuItems (items);
        }
        
        /** Add listeners to menu items.
        * @param items the items
        */
        private void addListeners (JMenuItem[] items) {
            int len = items.length;
            for (int i = 0; i < len; i++) {
                items[i].addPropertyChangeListener (propL);
            }
        }

        /** Remove all listeners from menu items.
        * @param items the items
        */
        private void removeListeners (JMenuItem[] items) {
            int len = items.length;
            for (int i = 0; i < len; i++) {
                items[i].removePropertyChangeListener (propL);
            }
        }
        
        boolean needsChange = false;
        
        public void addNotify() {
            if (needsChange) {
                changeMenuItems (createMenu (popup));
                needsChange = false;
            }
            super.addNotify();
        }

        public void removeNotify() {
            removeListeners (last);
            last = NONE;
        }


        /** Property listnener to watch changes of enable state.
        */
        private class PropL implements PropertyChangeListener {
            public void propertyChange (PropertyChangeEvent ev) {
                String name = ev.getPropertyName ();
                if (
                    name == null ||
                    name.equals (SystemAction.PROP_ENABLED) ||
                    name.equals (Registry.PROP_ACTIVATED_NODES)
                ) {
                    // change items later
                    needsChange = true;
                }
            }
        }
    }
}
