/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.cache;



import org.netbeans.modules.vcscore.util.Debug;
import java.util.*;
import java.io.*;

/** this class displays a directory in the cache and ensures the consistency of the cache. 
 *  You can add/remove directories/files, write to disk cache etc. Also keeps track if all children are stil in the cache.
 *  Your subclasses should mainly just override the abstract methods.
 *
 * @author  mkleint
 * @version
 */

public abstract class CacheDir extends CacheFile {
    private Debug E=new Debug("CacheDir", true); // NOI18N
    private Debug D=E;
    
    private int strategy;
    /** absolute pathname of the directory.
     */
    protected String dirName;
    /** File instance of this directory.
     */
    protected File dirFile;
    private boolean modified = false;

    /** hashmap of subdirectories - are kept separately from files */
    protected HashMap childDirs = null;

    /** hashmap of files  - are kept separately from  subdirectories */
    protected HashMap cachedFiles;
    private boolean local = true;
    /** flag that is switched off when a file/subdir is removed by the CacheQueue
     */
    protected boolean complete;
//    protected int childCount;
    //    private CacheDir parent = null;
    protected FileSystemCache cacheObject;
    
    /** Creates new CacheDir */
    public CacheDir(String cacheName, File dirFile) {
        super(cacheName);
        D.deb("Costructor", "dirName=" + dirFile.getAbsolutePath());
        setAppliedLevel(CacheHandler.STRAT_NONE);
        cachedFiles = new HashMap(50);
        childDirs = new HashMap(10);
        modified = false;
//        childCount = 0;
        this.dirFile = dirFile;
        dirName = dirFile.getAbsolutePath();
        this.setName(dirFile.getName());
        complete = true;
        cacheObject = CacheHandler.getInstance().getCache(getCacheName());
        D.deb("cacheObjectName = " + getCacheName());
    }
    
    // ------ dir stuff -----------------------------------------------------------
    
    /** return all subdirectories in an array
     */
    public CacheDir[] getSubDirs() {
        Collection col;
        synchronized (this) {
            col = childDirs.values();
        }
        CacheDir[] dirs = new CacheDir[col.size()];
        Iterator it = col.iterator();
        for (int index = 0; index < col.size(); index++) {
            dirs[index] = (CacheDir)it.next();
        }
        return dirs;
    }
    
    /** returns a single subdir by name (just name not absolute path)
     */
    public synchronized CacheDir getSubDir(String name) {
        CacheDir toReturn = (CacheDir)childDirs.get(name);
        return toReturn;
    }
    
/*    //--
    public String[] getDirContentNames() {
        D.deb("getDirContentName() -- beginning of " + this.getAbsolutePath());
        Collection colDr = childDirs.values();
        Collection colFl = cachedFiles.values();
        int dirSize = colDr.size() + colFl.size();
        int size1 = colDr.size();
        String[] content = new String[dirSize];
        Iterator it = colDr.iterator();
        int index = 0;
        while (index < size1) {
            CacheDir subDir = (CacheDir)it.next();
            content[index] = subDir.getName();
            index = index + 1;
        }
        it = colFl.iterator();
        for (int index2 = 0; index2 < colFl.size(); index2++) {
            content[index2 + size1] = ((CacheFile)it.next()).getName();
        }
        return content;
    }
 */   
    
    /** Adds a subdirectory to the current one.
     * @param fireEvent - if set to true, fires cacheAdded event + also causes the parent directory to get modified
     */
    public void addChildDir(CacheDir subDir, boolean fireEvent) {
        CacheDir dir = getCacheObject().getDir(subDir.getAbsolutePath());
        if (dir != null && dir.getCacheName().equals(getCacheName())) {
          /* directory is already in the cache - could be done by other fs mounted at different point
           *  or when reloading the directory (we don't want to destroy subdir structure, just find new nodes
           */
            subDir = dir;
        } else { // is new dir.. do register it..
            getCacheObject().registerDir(subDir);
        }
        childDirs.put(subDir.getName(),subDir);
        subDir.setParent(this);
        if (fireEvent) {
            getCacheObject().fireCacheHandlerEvent(FileSystemCache.EVENT_ADD, subDir);
            // also do change this dir to modified
            this.setModified(true);
        }
    }
    
    /** Removes a directory from the current one.
     * @param fireEvent - if set to true, fires cacheRemoved event;
     * also causes the parent directory to get modified
     *
     */
    
    public void removeChildDir(String subdirName, boolean fireEvent) {
        CacheDir subDir = this.getSubDir(subdirName);
        if (subDir != null) {
            childDirs.remove(subDir.getName());
//??            subDir.setParent(null);
            getCacheObject().unregisterDir(subDir);
            if (fireEvent) {
                getCacheObject().fireCacheHandlerEvent(FileSystemCache.EVENT_REMOVE, subDir);
                this.setModified(true);
            }
        }
    }
    
    /**
     * @param fireEvent - if set to true, fires cacheAdded event;
     * also causes the parent directory to get modified
     *
     */
    public void renameChildDirs(CacheDir newParent, boolean fireEvent) {
        CacheDir[] dirs = this.getSubDirs();
        for (int index = 0; index < dirs.length; index++) {
            CacheDir childDir = dirs[index];
            removeChildDir(childDir.getName(),false);
            childDir.rename(new File(newParent.getAbsolutePath() + File.separator + childDir.getName()));
            newParent.addChildDir(childDir,false);
            childDir.renameChildDirs(childDir,false);
        }
    }
    
    /** Renames the current instance. 
     * make sure you register-unregister the dir in the cache.
     */
    public void rename(File newLocation) {
        D.deb("rename()", "Old=" + this.getAbsolutePath());
        dirFile = newLocation;
        dirName = dirFile.getAbsolutePath();
        this.setName(dirFile.getName());
        D.deb("rename()", "New=" + this.getAbsolutePath());
    }
    
    
    //----- file stuff -------------------------------------------------------

    public CacheFile[] getFiles() {
        Collection col = cachedFiles.values();
        CacheFile[] files = new CacheFile[col.size()];
        Iterator it = col.iterator();
        for (int index = 0; index < col.size(); index++) {
            files[index] = (CacheFile) it.next();
        }
        return files;
    }
    /** gets a file that is inthe directory. If it's not there , returns null
     * @param name - filename of the file - no path
     */ 
    public CacheFile getFile(String name) {
        CacheFile toReturn = (CacheFile)cachedFiles.get(name);
        return toReturn;
    }
    
    /** adds a file to the directory.
     * @param fireEvent - if true, fires EVENT_ADD event + sets this dir as modified.
     */
    public void addFile(CacheFile fl, boolean fireEvent) {
        cachedFiles.put(fl.getName(), fl);
        fl.setParent(this);
        if (fireEvent) {
            getCacheObject().fireCacheHandlerEvent(FileSystemCache.EVENT_ADD, fl);
            setModified(true);
        }
    }
    /** removes a file from the directory.
     * @param fireEvent - if true, fires EVENT_REMOVE event + sets this dir as modified.
     */
    
    public void removeFile(String flName, boolean fireEvent) {
        CacheFile fl = getFile(flName);
        if (fl != null) {
            cachedFiles.remove(flName);
//?            fl.setParent(null);
            if (fireEvent) {
                getCacheObject().fireCacheHandlerEvent(FileSystemCache.EVENT_REMOVE, fl);
                setModified(true);
            }
        } else {
            System.out.println("ERROR=File WAS NOT removed :(");
        }
    }
    
    /**
     * Removes all files in this directory.
     */
    public void removeFiles() {
        LinkedList colFile = new LinkedList(cachedFiles.values());
        Iterator it = colFile.iterator();
        while (it.hasNext()) {
            CacheFile file = (CacheFile) it.next();
            removeFile(file.getName(), false);
        }
    }

    /** removes all subdir and file cache objects in this directory.
     */
    public void removeAll(boolean recursively) {
        removeFiles();
        // same for dirs.
        LinkedList colDir = new LinkedList(childDirs.values());
        Iterator it = colDir.iterator();
        while (it.hasNext()) {
            CacheDir dir = (CacheDir)it.next();
            if (recursively) {
                dir.removeAll(true);
            }
            removeChildDir(dir.getName(), false);
/*            dir.setParent(null);
            D.deb("removeAll() - removed:" + dir.getAbsolutePath());
            getCacheObject().unregisterDir(dir);
            it.remove();
 */
        }
    }
    
    
    /** sets the strategy level taht this directory resides at.
     */
    //--
    public void setAppliedLevel(int strat) {
        strategy = strat;
    }
    /** gets the strategy level taht this directory resides at.
     */
    public int getAppliedLevel() {
        return strategy;
    }
    
    /** returns the name of the directory
     */
    public String getName() {
        return dirFile.getName(); // return short name, just name of file..
        //        return dirName;
    }
    
    
    /** Absolute path to the directory.
     * By this string the directory can be found by the cache Handler
     */
    public String getAbsolutePath() {
        return dirName;
    }
    
    public String toString() {
        return this.getName();
    }
    
    /** Sets the dir as modified. Do so when you want the dir to be written to disk cache later.
     */
    public void setModified(boolean mod) {
        modified = mod;
    }
    
    /** can be used when deciding wheater it should be written to disk cache
     */
    public boolean isModified() {
        return modified;
    }
    
    /** sets indication about the completeness of the directory in the cache.
     * is set by the CacheQueue when removing the dir's file/subdir.
     * If it was previously modified, then it's first saved to disk.
     */
    public void setComplete(boolean compl) {
        if (this.complete == true && compl == false) {
            D.deb("WE GOT UNCOMPLETE FOR " + getName());
            if (isModified()) {
                writeToDisk();
            }
            D.deb(this.getName() + " got uncomplete");
            //TODO .. fine tuning might be needed when files that still exist are requested
        }
        complete = compl;
    }
    
    /** check wheather the directory is complete 
     */
    
    public boolean isComplete() {
        return complete;
    }
    

   public boolean isEmpty() {
//       return (childCount <= 0);
        boolean ok = true;
        if (cachedFiles.size() > 0 || childDirs.size() > 0) ok = false;
        return ok;
    }
 
    
    public boolean isLocal() {
        return local;
    }
    public void setLocal(boolean loc) {
        local = loc;
    }
    
    public FileSystemCache getCacheObject() {
        return cacheObject;
    }
    
    // where to find the file that the cache is written to (for this directory)
    
    protected abstract String getCacheFileName();
    
    /** from CacheFile, generates a line that's written to disk cache.
     * By default nothing is written.
     */
    public String writeLineToDisk() {
        // directories don't get written to cache by default
        return "";
    }
    
    public void writeToDiskRecursively() {
        writeToDisk();
//        System.out.println("writing..." + this.getName());
        Iterator it = childDirs.values().iterator();
        while (it.hasNext()) {
            CacheDir child = (CacheDir)it.next();
            child.writeToDiskRecursively();
        }
    }
    
    /** Specify there how the directory should be written to disk.
     */
    public abstract void writeToDisk();
    
    /** associates with the STRAT_LOCAL, if that strategy is requested this method is called.
     * It should populate the the directory with files on the local disk physically there.
     * No status retrieved.
     */
    public abstract void populateWithLocal();
    
    /** associates with the STRAT_DISK, if that strategy is requested this method is called.
     * do load the disk cache in this method. I suggest not to add any files there, just update status etc.
     * adding new files that are not physically presnt can cause trouble.
     */
    public abstract boolean readFromDisk();
    /** associates with the STRAT_REFRESH, if that strategy is requested this method is called.
     * do a server check. Connect to server and request status update etc..
     */
    public abstract void checkServer();
  
    /** associates with the STRAT_REFRESH_RECURS, if that strategy is requested this method is called.
     * do a recursive server check. Connect to server and request status update etc..
     * Note: within this method implementations need to make sure
     */
    public abstract void checkServerRecursive(); 
    
}
