/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.cache;

import org.netbeans.modules.vcscore.util.Debug;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.Reference;
import java.io.*;
import java.util.*;

/** main entry point into the cache. The class is a singleton to be found via the getInstance() static method.
 *  Stores all registered caches + makes sure the requests are processed by the right cache instance.
 *
 * 
 * @author  mkleint
 */

public class CacheHandler extends java.lang.Object {
    private Debug E=new Debug("CacheHandler", false); // NOI18N
    private Debug D=E;
    
    
    /** Strategy constant, to be used during Filesystem.children()
     * do just locla stuff.. no server comm, or reading disk cache   
     * 
     */
    public  static final int STRAT_LOCAL = 0;

    /** Strategy constant, to be called during FileSystem.annotateName().
     * Calls the readFromDisk() in CacheDir.
     * just local stuff + disk cache reading   
     */
    public  static final int STRAT_DISK = 3;
    
    /**  Strategy constant, do not apply any strategy. just search the current state of the cache.
     *  To be called from within the ReferenceQueue only to check if 
     *  the file is in cache without recreating it.(probably)
     */
    public  static final int STRAT_NONE = -1;

    /** 
     *  Strategy constant that causes a refresh of the file/directory.
     *  fires the checkServer() method in the FileSystemCache.
     *  For this strategy all files in the parent directory are forgotten and reloaded.
     *  
     */
    public static final int STRAT_REFRESH = 11;

    /** 
     *  Strategy constant that causes a recursive refresh of the directory.
     *  fires the checkServerRecursively() method in the FileSystemCache.
     *  For this strategy all files/dirs in the parent directory are forgotten and reloaded.
     */
    public static final int STRAT_REFRESH_RECURS = 12;
    
    /**
     * This strategy constant is to be used when accessing cache during any kind of refresh.
     * 
     */
    public static final int STRAT_REFRESHING = 10;
    
    /**
     * This strategy constant is to be used when we want to perform the refresh 
     * only on directories where the reading from disk fails.
     */
    public static final int STRAT_DISK_OR_REFRESH = 5;
    
    private static CacheHandler instance = null;
    
    private ArrayList listeners = new ArrayList ();
    
    private HashMap cacheMap;
    private CacheQueue queue;
    
    /** Creates new CacheHandler */
    private CacheHandler() {
        D.deb("Constructor");
        cacheMap = new HashMap(5);
        queue = new CacheQueue();
        queue.start();
    }
    
    /** returns an instance of CacheHandler. Only one instance is needed throughout the IDE.
     */
    public static CacheHandler getInstance() {
        if (instance == null) {
            instance = new CacheHandler();
        }
        return instance;
    }
    
    private ReferenceQueue getRefQueue() {
        return queue.getQueue();
    }
    
    /** Creates a reference object. To be used in the Filesystem's createReference method to ensure
     * that the cache shrinks well.
     */
    public Reference createReference(Object obj, File file, String cacheId) {
        return new CacheReference(obj, getRefQueue(), file, cacheId);
    }

    /** The preffered way of getting files/dirs from the cache.
     * @param toFind     the file we want to get from the cache.
     * @param strategy   crutial parameter. Defines what measures should be taken by the cache
     *                   to locate the file needed.
     *                   Following strategies are defined:</p>
     *                   STRAT_NONE - nothing is done (currently used only in the reference queue so that the cache isn't rebuild when shrinking.</p>
     *                   STRAT_LOCAL - loads the file only from the local dir structure.</p>
     *                   STRAT_DISK - reads the file(s) that stores the content of cache previously written to disk.</p>
     *                   STRAT_REFRESH - checks the server for new states. </p>
     *                   STRAT_REFRESH_RECURS - for directories performs recursive refresh.
     * 
     * @param cacheFs the calling filesystem (interface CacheFileSystem) that provides the cacheID string in order to find the cache instance in the registry
     * <p> For STRAT_REFRESH, STRAT_REFRESH_RECURS and STRAT_REFRESHING or higher, 
     * the disk cache is not consulted (REFRESH/R_R also does clear the cache and relaods)
     * 
     */
    
    public CacheFile getCacheFile(File toFind, int strategy, CachedFileSystem cacheFs) {
        return getCacheFile(toFind, strategy, cacheFs.getCacheId());
    }
    
    /** Way of getting files/dirs from the cache. For details see...
     *
     */
    
    public CacheFile getCacheFile(File toFind, int strategy, String type) {
        FileSystemCache cache = getCache(type);
        if (cache == null) {
            return null;
        }
        return cache.getCacheFile(toFind,strategy);
    }  
    
    /** adds a new cache to the registry. To be called when the filesystem gets created/upon restarting the IDE.
     * @param type  IDString that identifies the cache. (the cacheID can be 1 for a single filesystem or for all filesystems of one kind.
     */
    public void registerCacheType(String type, FileSystemCache cacheObject) {
        cacheMap.put(type, cacheObject);
    }
    
    /** Opposite action to registerCache. Do when unmounting filesystem that has it's own cache.
     */
    public void unregistedType(String type) {
        cacheMap.remove(type);
    }
    
    /** utility method for grabbing the right cache for the specified ID.
     *  Not to be used for file retrieval tasks.
     */
    public FileSystemCache getCache(CachedFileSystem type) {
        FileSystemCache cache = (FileSystemCache)cacheMap.get(type.getCacheId());
        return cache;
    }
    
    /** utility method for grabbing the right cache for the specified ID.
     *  Not to be used for file retrieval tasks.
     */
    
    public FileSystemCache getCache(String type) {
        FileSystemCache cache = (FileSystemCache)cacheMap.get(type);
        return cache;
    }
    
    
}