/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.caching;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.io.File;

import org.netbeans.modules.vcscore.cache.*;
import org.netbeans.modules.vcscore.commands.VcsCommand;
import org.netbeans.modules.vcscore.commands.VcsCommandIO;
import org.netbeans.modules.vcscore.Variables;
import org.netbeans.modules.vcscore.util.VcsUtilities;

/**
 * This class provides static methods for creation cache data from refresh command.
 *
 * @author  Martin Entlicher
 */
public class RefreshCommandSupport extends Object {
    
    /**
     * The index of file name in the cache.
     */
    public static final int ELEMENT_INDEX_FILE_NAME = 0;
    /**
     * The index of file status in the cache.
     */
    public static final int ELEMENT_INDEX_STATUS = 1;
    /**
     * The index of file locker in the cache.
     */
    public static final int ELEMENT_INDEX_LOCKER = 2;
    /**
     * The index of file revision in the cache.
     */
    public static final int ELEMENT_INDEX_REVISION = 3;
    /**
     * The index of file sticky information in the cache (i.e. the current branch)
     */
    public static final int ELEMENT_INDEX_STICKY = 4;
    /**
     * The index of file additional attribute in the cache.
     */
    public static final int ELEMENT_INDEX_ATTR = 5;
    /**
     * The index of file size in the cache.
     */
    public static final int ELEMENT_INDEX_SIZE = 6;
    /**
     * The index of file date in the cache.
     */
    public static final int ELEMENT_INDEX_DATE = 7;
    /**
     * The index of file time in the cache.
     */
    public static final int ELEMENT_INDEX_TIME = 8;
    /**
     * The number of file attributes in the cache.
     */
    public static final int NUM_ELEMENTS = 9;
    
    /**
     * The annotation pattern of a file name.
     */
    public static final String ANNOTATION_PATTERN_FILE_NAME = "fileName";
    /**
     * The annotation pattern of a file status.
     */
    public static final String ANNOTATION_PATTERN_STATUS = "status";
    /**
     * The annotation pattern of a file locker.
     */
    public static final String ANNOTATION_PATTERN_LOCKER = "locker";
    /**
     * The annotation pattern of a file revision.
     */
    public static final String ANNOTATION_PATTERN_REVISION = "revision";
    /**
     * The annotation pattern of a file sticky info.
     */
    public static final String ANNOTATION_PATTERN_STICKY = "sticky";
    /**
     * The annotation pattern of a file additional attribute.
     */
    public static final String ANNOTATION_PATTERN_ATTR = "attribute";
    /**
     * The annotation pattern of a file size.
     */
    public static final String ANNOTATION_PATTERN_SIZE = "size";
    /**
     * The annotation pattern of a file date.
     */
    public static final String ANNOTATION_PATTERN_DATE = "date";
    /**
     * The annotation pattern of a file time.
     */
    public static final String ANNOTATION_PATTERN_TIME = "time";
    
    /**
     * When the attributes of the set of files differ, do not show them.
     */
    public static final int MULTI_FILES_ANNOTATION_EMPTY = 0;
    /**
     * When the attributes of the set of files differ, show them as a list.
     */
    public static final int MULTI_FILES_ANNOTATION_LIST = 1;
    /**
     * When the attributes of the set of files differ, show a "Not In Synch" status instead.
     */
    public static final int MULTI_FILES_ANNOTATION_NOT_SYNCH_ATTR = 2;

    /* ${fileName} $[? status] [[${status}$[? revision] [, revision] []]] []$[? revision] [revision] []$[? locker] [(${locker})] [] */
    public static final String DEFAULT_ANNOTATION_PATTERN = "${"+ANNOTATION_PATTERN_FILE_NAME+"}"+
    "$[? "+ANNOTATION_PATTERN_STATUS+"] "+
        "[ [${"+ANNOTATION_PATTERN_STATUS+"}$[? "+ANNOTATION_PATTERN_REVISION+"] [; ${"+ANNOTATION_PATTERN_REVISION+"}] []]] "+
        "["+"$[? "+ANNOTATION_PATTERN_REVISION+"] [ ${"+ANNOTATION_PATTERN_REVISION+"}] []"+"]"+
    "$[? "+ANNOTATION_PATTERN_LOCKER+"][ (${"+ANNOTATION_PATTERN_LOCKER+"})] []" +
    "$[? "+ANNOTATION_PATTERN_STICKY+"][ (${"+ANNOTATION_PATTERN_STICKY+"})] []";
    public static final int[] DEFAULT_MULTI_FILES_ANNOTATION_TYPES = { MULTI_FILES_ANNOTATION_NOT_SYNCH_ATTR, MULTI_FILES_ANNOTATION_LIST, MULTI_FILES_ANNOTATION_EMPTY,
                                                                       MULTI_FILES_ANNOTATION_EMPTY, MULTI_FILES_ANNOTATION_EMPTY, MULTI_FILES_ANNOTATION_EMPTY,
                                                                       MULTI_FILES_ANNOTATION_EMPTY, MULTI_FILES_ANNOTATION_EMPTY, MULTI_FILES_ANNOTATION_EMPTY };
    public static final String DEFAULT_MULTI_FILES_ANNOTATION_DELIMETER = ", ";
    
    private static final String DIRECTORY_CACHE_ID = "D";
    
    /** Creates new RefreshCommandSupport */
    private RefreshCommandSupport() {
    }

    /*
     * Create a new VcsCacheFile or VcsCacheDir from reader data
     * @param elements the data obtained from the reader
     * @param list the list command
     * @return the created VcsFile
     *
    public static CacheFile matchToFile(String[] elements, String cacheName, File parent) {
        return matchToFile(elements, null, cacheName, parent);
    }
     */
    
    /**
     * Extract the file name from the array of elements.
     */
    public static String getFileName(String[] elements) {
        if (elements.length < NUM_ELEMENTS) return null;
        return elements[ELEMENT_INDEX_FILE_NAME];
    }

    /**
     * Create a new VcsCacheFile or VcsCacheDir from reader data
     * @param elements the data obtained from the reader
     * @param list the list command
     * @param statusTr the transformation table of file statuses.
     * @return the created VcsFile
     * @throws IllegalArgumentException if the file name element is <code>null</code>
     *                                  or if the number of elements is not sufficient
     */
    public static CacheFile matchToFile(String[] elements, //HashMap statusTr,
                                        String cacheName, File parent) {
        if (elements.length < NUM_ELEMENTS) {
            throw new IllegalArgumentException("The number of elements ("+elements.length+") < "+NUM_ELEMENTS);
        }
        //System.err.println("matchToFile("+VcsUtilities.arrayToString(elements)+")");

        /*
        CacheFile file = (CacheFile) CacheHandler.getInstance().getCacheFile(
                    fileSystem.getFile(path),
                    CacheHandler.STRAT_TEMP_DISK, fileSystem.getCacheIdStr());
         */
        CacheFile file = new VcsCacheFile(cacheName);

        //if (VcsUtilities.withinRange(0, fileNameIndex, elements.length - 1)) {
            String name = elements[ELEMENT_INDEX_FILE_NAME];
            if (name == null) throw new IllegalArgumentException("File name element is missing.");
            file.setName (name);
            if (name.endsWith("/")) { // NOI18N
                file = (VcsCacheDir) CacheHandler.getInstance().getCacheFile(
                        new File(parent, name.substring(0,name.length()-1)),
                        CacheHandler.STRAT_REFRESHING, cacheName);
                if (file == null) {
                    file = new VcsCacheDir(cacheName, new File(parent, name.substring(0,name.length()-1)));
                }
            }
        //}
        
        //if (VcsUtilities.withinRange(0, statusIndex, elements.length - 1)) {
            String status = elements[ELEMENT_INDEX_STATUS];
            if (status != null) file.setStatus(status);
            /*
            String trans = (statusTr != null) ? (String) statusTr.get(status) : null;
            if (trans == null) {
                file.setStatus (status);
            } else {
                file.setStatus(trans);
            }
             */
        //}
        
        String locker = elements[ELEMENT_INDEX_LOCKER];
        if (locker != null) file.setLocker (locker);
        
        String revision = elements[ELEMENT_INDEX_REVISION];
        if (revision != null) file.setRevision(revision);
        
        String sticky = elements[ELEMENT_INDEX_STICKY];
        if (sticky != null) file.setSticky(sticky);
        
        String attr = elements[ELEMENT_INDEX_ATTR];
        if (attr != null) file.setAttr (attr);
        
        String date = elements[ELEMENT_INDEX_DATE];
        if (date != null) file.setDate (date);
        
        String time = elements[ELEMENT_INDEX_TIME];
        if (time != null) file.setTime (time);
        
        String size = elements[ELEMENT_INDEX_SIZE];
        if (size != null) {
            try {
                file.setSize (Integer.parseInt(size));
            }
            catch (NumberFormatException e) {
                file.setSize (0);
            }
        }
        //System.err.println("file="+file);
        return file;
    }

    public static String[] makeElements(CacheFile file) {
        String[] elements = new String[NUM_ELEMENTS];
        elements[ELEMENT_INDEX_FILE_NAME] = file.getName();
        elements[ELEMENT_INDEX_STATUS] = file.getStatus();
        elements[ELEMENT_INDEX_LOCKER] = file.getLocker();
        elements[ELEMENT_INDEX_REVISION] = file.getRevision();
        elements[ELEMENT_INDEX_STICKY] = file.getSticky();
        elements[ELEMENT_INDEX_ATTR] = file.getAttr();
        elements[ELEMENT_INDEX_DATE] = file.getDate();
        elements[ELEMENT_INDEX_TIME] = file.getTime();
        elements[ELEMENT_INDEX_SIZE] = Integer.toString(file.getSize());
        return elements;
    }
    
    /*
    public static String[] getElementsFromLine(String line) {
        boolean dir = false;
        StringTokenizer tok = new StringTokenizer(line, "/");
        String[] elements = new String[NUM_ELEMENTS];
        for(int i = 0; tok.hasMoreTokens() && i < NUM_ELEMENTS; i++) {
            String element = tok.nextToken();
            if (i == 0 && DIRECTORY_CACHE_ID.equals(element)) {
                dir = true;
                i--;
                continue;
            }
            elements[i] = element;
        }
        if (dir) elements[ELEMENT_INDEX_FILE_NAME] += "/";
        return elements;
    }
     */
    
    public static String[] getElementsFromLine(String line) {
        boolean dir = false;
        int index = 0;
        if (line.charAt(index) == '/') index++;
        String[] elements = new String[NUM_ELEMENTS];
        for(int i = 0; index < line.length() && i < NUM_ELEMENTS; i++) {
            int next = line.indexOf('/', index);
            if (next < 0) break;
            String element = new String(line.substring(index, next)); // I want to have a completely new object so that line can be G.C.
            index = next + 1;
            if (i == 0 && DIRECTORY_CACHE_ID.equals(element)) {
                dir = true;
                i--;
                continue;
            }
            elements[i] = element;
        }
        if (dir) elements[ELEMENT_INDEX_FILE_NAME] += "/";
        return elements;
    }
    
    public static String getLineFromElements(String[] elements) {
        if (elements.length < NUM_ELEMENTS) return null;
        boolean dir = elements[ELEMENT_INDEX_FILE_NAME].endsWith("/");
        StringBuffer line = new StringBuffer();
        if (dir) line.append(DIRECTORY_CACHE_ID);
        line.append("/");
        for(int i = 0; i < NUM_ELEMENTS; i++) {
            line.append(elements[i]);
            if (i != 0 || !dir) line.append("/");
        }
        return line.toString();
    }
    
    /**
     * Get the annotation line for a file.
     * @param name the object file name
     * @param fullName the full path of the file with respect to the filesystem root
     * @param annotationPattern the pattern how the annotation should be displayed
     * @param statusProvider the provider of the status attributes information
     * @return the annotation pattern filled up with proper attributes
     */
    public static String getStatusAnnotation(String name, String fullName, String annotationPattern,
                                             FileStatusProvider statusProvider) {
        return getStatusAnnotation(name, fullName, annotationPattern, statusProvider, null);
    }
    
    /**
     * Get the annotation line for a file.
     * @param name the object file name
     * @param fullName the full path of the file with respect to the filesystem root
     * @param annotationPattern the pattern how the annotation should be displayed
     * @param statusProvider the provider of the status attributes information
     * @return the annotation pattern filled up with proper attributes
     */
    public static String getStatusAnnotation(String name, String fullName, String annotationPattern,
                                             FileStatusProvider statusProvider, Hashtable additionalVars) {
        Hashtable vars = new Hashtable();
        if (additionalVars != null) vars.putAll(additionalVars);
        vars.put(ANNOTATION_PATTERN_FILE_NAME, name);
        String status = statusProvider.getFileStatus(fullName);
        String trans = null;
        HashMap possibleFileStatusesMap = statusProvider.getPossibleFileStatusesTable();
        if (possibleFileStatusesMap != null && status != null) {
            synchronized (possibleFileStatusesMap) {
                trans = (String) possibleFileStatusesMap.get(status);
            }
        }
        if (trans != null) {
            status = trans;
        }
        if (status != null) vars.put(ANNOTATION_PATTERN_STATUS, status);
        status = statusProvider.getFileLocker(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_LOCKER, status);
        status = statusProvider.getFileRevision(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_REVISION, status);
        status = statusProvider.getFileSticky(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_STICKY, status);
        status = statusProvider.getFileSize(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_SIZE, status);
        status = statusProvider.getFileAttribute(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_ATTR, status);
        status = statusProvider.getFileDate(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_DATE, status);
        status = statusProvider.getFileTime(fullName);
        if (status != null) vars.put(ANNOTATION_PATTERN_TIME, status);
        //System.out.println("vars = "+vars+",\npattern = "+annotationPattern+",\nexpansion = "+Variables.expandFast(vars, annotationPattern, false));
        return Variables.expand(vars, annotationPattern, false);
    }
    
    /**
     * Get the annotation line for a list of files.
     * @param name the object file name
     * @param fullName the full path of the file with respect to the filesystem root
     * @param annotationPattern the pattern how the annotation should be displayed
     * @param statusProvider the provider of the status attributes information
     * @param multiFilesAnnotationTypes the annotation types for individual attributes.
     *        Values of this files are <code>MULTI_FILES_ANNOTATION_*<code> constants.
     * @return the annotation pattern filled up with proper attributes
     */
    public static String getStatusAnnotation(String name, ArrayList files, String annotationPattern,
                                             FileStatusProvider statusProvider, int[] multiFilesAnnotationTypes) {
        Hashtable vars = new Hashtable();
        ArrayList attributes = new ArrayList();
        int n = files.size();
        vars.put(ANNOTATION_PATTERN_FILE_NAME, name);
        String trans = null;
        HashMap possibleFileStatusesMap = statusProvider.getPossibleFileStatusesTable();
        if (possibleFileStatusesMap != null) {
            synchronized (possibleFileStatusesMap) {
                for(int i = 0 ; i < n; i++) {
                    String status = statusProvider.getFileStatus((String) files.get(i));
                    trans = (String) possibleFileStatusesMap.get(status);
                    if (trans != null) {
                        status = trans;
                    }
                    attributes.add(status);
                }
            }
        } else {
            for(int i = 0 ; i < n; i++) {
                attributes.add(statusProvider.getFileStatus((String) files.get(i)));
            }
        }
        String attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_STATUS],
                                   statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_STATUS, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileLocker((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_LOCKER],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_LOCKER, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileRevision((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_REVISION],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_REVISION, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileSticky((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_STICKY],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_STICKY, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileAttribute((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_ATTR],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_ATTR, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileSize((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_SIZE],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_SIZE, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileDate((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_DATE],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_DATE, attr);
        attributes.clear();
        for(int i = 0 ; i < n; i++) {
            attributes.add(statusProvider.getFileTime((String) files.get(i)));
        }
        attr = getAttribute(attributes, multiFilesAnnotationTypes[ELEMENT_INDEX_TIME],
                            statusProvider.getNotInSynchStatus());
        if (attr != null) vars.put(ANNOTATION_PATTERN_TIME, attr);
        return Variables.expand(vars, annotationPattern, false);
    }
    
    private static String getAttribute(ArrayList attributes, int annotationType, String nSynch) {
        boolean differ;
        Object firstObj = attributes.get(0);
        if (firstObj == null) return null;
        String first = (String) firstObj;
        int n = attributes.size();
        int i = 1;
        for (i = 1; i < n; i++) {
            if (!firstObj.equals(attributes.get(i))) {
                //System.out.println("first = "+firstObj+" != "+attributes.get(i));
                break;
            }
        }
        differ = i < n;
        switch (annotationType) {
            case MULTI_FILES_ANNOTATION_EMPTY:
                if (differ) return "";
                else return first;
            case MULTI_FILES_ANNOTATION_NOT_SYNCH_ATTR:
                if (differ) return nSynch;
                else return first;
            case MULTI_FILES_ANNOTATION_LIST:
                if (!differ) return first;
                StringBuffer buf = new StringBuffer(first);
                for (int j = 1; j < n; j++) {
                    buf.append(DEFAULT_MULTI_FILES_ANNOTATION_DELIMETER + (String) attributes.get(j));
                }
                return buf.toString();
            default: return "";
        }
    }
}
