/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.caching;

import java.lang.ref.Reference;
import java.io.File;
import java.io.FilenameFilter;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;

import org.openide.filesystems.FileObject;

import org.netbeans.modules.vcscore.cache.*;
import org.netbeans.modules.vcscore.util.Debug;
import org.netbeans.modules.vcscore.util.VcsUtilities;
import org.netbeans.modules.vcscore.VcsFactory;
import org.netbeans.modules.vcscore.VcsDirReader;
import org.netbeans.modules.vcscore.DirReaderListener;
import org.netbeans.modules.vcscore.FileReaderListener;
import org.netbeans.modules.vcscore.VcsDirContainer;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.commands.VcsCommandExecutor;

/**
 *
 * @author  Martin Entlicher
 */
public class VcsCache extends FileSystemCache implements DirReaderListener, FileReaderListener {
    
    private Debug E=new Debug("VcsCache", true); // NOI18N
    private Debug D=E;

    //public static final String VCS_CACHE_NAME = "VCS_Cache";
    
    private Reference fileSystem;
    private String fsRoot;
    private int rootNameLength;

   /** A hashtable of cache files, identified by full path.
     * Key: full path (string)  for Files File.getAbsolutePath()
     *                          for VcsCacheDirs VcsCacheDir.getAbsolutePath()
     * Value: instance of VcsCacheDir
     */
    private HashMap cacheDirs;
    
    /**
     * The map of paths, which should be refreshed after they become non-local
     * with their strategies.
     */
    private HashMap refreshStrategyWhenNonLocal;
    
    /** Creates new VcsCache with default name <code>VCS_CACHE_NAME</code> *
    public VcsCache(Reference fileSystem) {
        this(fileSystem, VCS_CACHE_NAME);
    }
     */

    /** Creates new VcsCache of a custom name
     * @args cacheName the cache name.
     */
    public VcsCache(Reference fileSystem, String cacheName) {
        super(cacheName);
        this.fileSystem = fileSystem;
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        fsRoot = fsystem.getRootDirectory().getAbsolutePath(); // filesystem has to exist during construction
        fsystem.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                updateFSRoot();
            }
        });
        rootNameLength = fsRoot.length();
        cacheDirs = new HashMap();
        refreshStrategyWhenNonLocal = new HashMap();
    }
    /*
    void setFSRoot(File fsRoot) {
        this.fsRoot = fsRoot;
    }
    
    File getFSRoot() {
        return fsRoot;
    }
    */
    
    public void updateFSRoot() {
        //System.out.println("VcsCache.updateFSRoot()");
        fsRoot = ((VcsFileSystem) fileSystem.get()).getRootDirectory().getAbsolutePath(); // filesystem has to exist during construction
        rootNameLength = fsRoot.length();
    }
    
    /**
     * Get the path from the root of the filesystem for the given file.
     */
    String getPath(File file) {
        String filePath = file.getAbsolutePath();
        //System.out.println("VcsCache.getPath("+filePath+"), rootNameLength = "+rootNameLength);
        if (filePath.length() <= rootNameLength) return "";
        String path = filePath.substring(rootNameLength + 1); // skip the path separator also
        path.replace(File.separatorChar, '/');
        //System.out.println(" -> path = "+path);
        return path;
    }
    
    String getAbsolutePath(File file) {
        return file.getAbsolutePath();//.replace(File.separatorChar, '/');
    }
    
    private VcsFactory getFactory() {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return null;
        return fsystem.getVcsFactory();
    }
    
    /** Run the directory reader on the directory
     * @param dir the absolute path to the directory
     */
    synchronized void runVcsDirReader(File dir) {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return ;
        VcsCacheDir cacheDir = (VcsCacheDir) getDir(getAbsolutePath(dir));
        //System.out.println("runVcsDirReader("+dir+"), absolutePath = "+getAbsolutePath(dir)+", cahceDir = "+cacheDir);
        //System.out.println("runVcsDirReader: cacheDir = "+cacheDir+", isBeingLoaded = "+cacheDir.isBeingLoaded()+", isLoaded = "+cacheDir.isLoaded());
        if (cacheDir.isBeingLoaded()) return ;
        cacheDir.setBeingLoaded(true);
        VcsFactory factory = fsystem.getVcsFactory();
        VcsCommandExecutor exec = factory.getVcsDirReader(this, getPath(dir));
        if (exec != null) fsystem.getCommandsPool().startExecutor(exec);
    }

    /** Run the recursive directory reader on the directory
     * @param dir the absolute path to the directory
     */
    synchronized void runVcsDirReaderRecursive(File dir) {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return ;
        VcsFactory factory = fsystem.getVcsFactory();
        VcsCommandExecutor exec = factory.getVcsDirReaderRecursive(this, getPath(dir));
        if (exec != null) fsystem.getCommandsPool().startExecutor(exec);
    }
    
    String getCacheFileName(File dir) {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return null;
        return fsystem.getCacheFileName(getPath(dir));
    }
    
    FilenameFilter getLocalFileFilter() {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return null;
        return fsystem.getLocalFileFilter();
    }
    
    String getLocalFileStatus() {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return "";
        return fsystem.getStatusProvider().getLocalFileStatus();
    }
    
    public void writeAllToDisk() {
    }

    protected void loadDir(CacheDir dir, int strategy) {
        super.loadDir(dir, strategy);
        if (dir.isLocal()) {
            if (doesStrategyApply(dir, strategy, CacheHandler.STRAT_REFRESH)
                || doesStrategyApply(dir, strategy, CacheHandler.STRAT_DISK)) {
                refreshStrategyWhenNonLocal.put(dir.getAbsolutePath(), new Integer(strategy));
            }
        }
    }
    
    public CacheFile getCacheFile(File toFind, int strategy) {
        //D.deb("getCacheFile:" + toFind.toString() + "  Strategy=" + strategy);
        //System.out.println("getCacheFile:" + toFind.toString() + "  Strategy=" + strategy);
        //CacheFile cacheFile = (CacheFile) cacheDirs.get(toFind.getAbsolutePath());
        //if (cacheFile != null) {
        //}
      /*
      if (!toFind.exists()) {
          return null; 
      }
      if (toFind.isDirectory()) {
        D.deb("getCacheFile- isDirectory");
        String parentDir = toFind.getParentFile().getAbsolutePath();
        CacheDir cacheDir = getDir(parentDir);
        if (cacheDir == null) {
            //it doesn't have a parent -> is some kind of root.
          cacheDir = initCacheDir(toFind);  
          loadDir(cacheDir, strategy);
          return cacheDir;
        }
        loadDir(cacheDir, strategy);
        if (cacheDir == null) return null;
        return cacheDir.getSubDir(toFind.getName());
      }
       */
      //else {
        D.deb("getCacheFile- isFile");
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem == null) return null;
        //System.out.println(toFind+".equals FS root = "+fsystem.getRootDirectory().equals(toFind));
        if (fsystem.getRootDirectory().equals(toFind)) {
            return initCacheDir(toFind);
        }
        String parentDir = toFind.getParentFile().getAbsolutePath();
        CacheDir cacheDir = getDir(parentDir);
        //System.out.println("parentDir = "+parentDir+" => cacheDir = "+cacheDir);
        if (cacheDir == null) {
            cacheDir = initCacheDir(toFind.getParentFile());  
            //cacheDir = new VcsCacheDir(getId(), toFind.getParentFile());
            //cacheDir.setParent(null);
            //registerDir(cacheDir);
            if (cacheDir == null) return null;
        }
        loadDir(cacheDir, strategy);
        //if (cacheDir == null) return null;
        String name = toFind.getName();
        CacheFile file = cacheDir.getFile(name);
        if (file != null) return file;
        CacheDir dir = cacheDir.getSubDir(name);
        return dir;
        //return cacheDir.getFile(toFind.getName());
      //}
    }

    public void doRefreshDir(CacheDir dir, boolean recursively) {
        int strategy = recursively ? CacheHandler.STRAT_REFRESH_RECURS : CacheHandler.STRAT_REFRESH;
        loadDir(dir, strategy);
    }
    
    /*
    public CacheDir getCacheDir(File toFind, int strategy) {
        D.deb("getCacheDir:" + toFind.toString() + "  Strategy=" + strategy);
        String parentDir = toFind.getParentFile().getAbsolutePath();
        CacheDir cacheDir = getDir(parentDir);
        if (cacheDir == null) {
            //it doesn't have a parent -> is some kind of root.
          cacheDir = initCacheDir(toFind);  
          loadDir(cacheDir, strategy);
          return cacheDir;
        }
        loadDir(cacheDir, strategy);
        if (cacheDir == null) return null;
        return cacheDir.getSubDir(toFind.getName());        
    }
    */
     private CacheDir initCacheDir(File toInit) {
         //System.out.println("initCacheDir("+toInit+")");
         File parentFile = toInit.getParentFile();
         CacheDir parentDir = null;
         if (parentFile != null) {
            getDir(parentFile.getAbsolutePath());
         }
         CacheDir initDir = null;
         if (parentDir == null) {
             //System.out.println("initCacheDir without parent.");
             //D.deb("initCacheDir", "is without Parent");
             //          CacheDir parentofParent = initCacheDir(toInit.getParentFile(), strategy); //TEMP
             // the current dir will be prolly a root directory of sorts..
             VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
             if (fsystem == null) return null;
             VcsFactory factory = fsystem.getVcsFactory();
             initDir = new VcsCacheDir(getId(), toInit);
             initDir.setParent(null);
             // I have to suppose, that the initial directory is not local,
             // otherwise no refresh would be performed on it.
             initDir.setLocal(false);
             registerDir(initDir);
         } else {
             // now let's see in the disc cache first.. parent has to have the curr. dir in the children list
             //D.deb("initCacheDir", "has parent=" + parentDir.getName());
             //System.out.println("initCacheDir has parent = "+parentDir.getName());
             initDir = parentDir.getSubDir(toInit.getName());
             if (initDir == null) {
                 initDir = new VcsCacheDir(getId(), toInit);
                 parentDir.addChildDir(initDir, false);
                 initDir.setParent(parentDir);
                 // I have to suppose, that the initial directory is not local,
                 // otherwise no refresh would be performed on it.
                 initDir.setLocal(false);
                 registerDir(initDir);
             }
         }
         //System.out.println("initCacheDir: initDir = "+initDir);
         return initDir;
    }

    //public boolean isCacheFile(File toFind) {
    //}
    
    public void registerDir(CacheDir dir) {
        if (dir != null) {
            //D.deb("registerDir", "Directory \"" + dir.getAbsolutePath() + "\" registered");
            String fullName = dir.getAbsolutePath();
            while (fullName.endsWith(File.separator)) fullName = fullName.substring(0, fullName.length() - 1);
            //System.out.println("VcsCache.registerDir("+fullName+", "+dir+")");
            cacheDirs.put(fullName, dir);
        }
    }
    
    public void registerDirRecursive(CacheDir dir) {
        registerDir(dir);
        CacheDir[] dirs = dir.getSubDirs();
        for (int i = 0; i < dirs.length; i++) {
            registerDirRecursive(dirs[i]);
        }
    }
    
    public void unregisterDir(CacheDir dir) {
        if (dir != null) {
            //D.deb("UNregisterDir", "Directory \"" + dir.getAbsolutePath() + "\" UNregistered");  
            cacheDirs.remove(dir.getAbsolutePath());
        }
    }
    
    /**
     * Get the cache directory
     * @param fullName the absolute path to the folder you want to obtain the cache for
     */
    public CacheDir getDir(String fullName) {
        while (fullName.endsWith(File.separator)) fullName = fullName.substring(0, fullName.length() - 1);
        int index;
        while((index = fullName.indexOf(File.separator+File.separator)) > 0) fullName = fullName.substring(0, index) + fullName.substring(index + 1, fullName.length());
        CacheDir cacheDir = (CacheDir)cacheDirs.get(fullName);
        //System.out.println("VcsCache.getDir("+fullName+") = "+cacheDir);
        return cacheDir;
    }
    
    /**
    * To be called when any change on directory occures that results in diferent appearence.
    */
    private void heyDoRefreshDir(String path) {
        //System.out.println("heyDoRefreshDir("+path+")");
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem != null) {
            org.openide.filesystems.FileObject fo = fsystem.findResource(path);
            if (fo != null) {
                //System.out.println("heyDoRefreshDir("+path+") REFRESH CALLED.");
                fo.refresh();
            }
        }
    }

    /**
    * To be called when any change on directory occures that results in diferent appearence.
    */
    private void heyDoRefreshDirRecursive(VcsCacheDir dir) {
        heyDoRefreshDir(getPath(dir.getFile()));
        CacheDir[] subdirs = dir.getSubDirs();
        for(int i = 0; i < subdirs.length; i++) {
            //CacheDir subdir = subdirs[i];
            //D.deb("Refreshing '"+subdirs[i]+"'");
            //if (subdir != null) {
            heyDoRefreshDirRecursive((VcsCacheDir) subdirs[i]);
            //}
        }
    }
    
    /** @param path the path relative to the fs root */
    private VcsCacheDir createSubFiles(VcsFileSystem fs, String path, Collection rawData) {
        // Cache provider has to be != null, otherwise I would not get here
        //System.out.println("createSubFiles("+path+", "+rawData+")");
        VcsCacheDir dir = (VcsCacheDir) fs.getCacheProvider().getDir(path);
        //System.out.println("dir = "+dir+", path = "+((dir != null) ? ""+dir.getFile() : "null"));
        if (dir == null) {
            dir = (VcsCacheDir) initCacheDir(fs.getFile(path));
        }
        //CacheDir[] subDirs = dir.getSubDirs();
        //String[] subDirPaths = new String[subDirs.length];
        //ArrayList knownSubDirPaths = new ArrayList();
        //for(int i = 0; i < subDirs.length; i++) knownSubDirPaths.add(subDirs[i].getAbsolutePath());
        for (Iterator it = rawData.iterator(); it.hasNext(); ) {
            String[] elements = (String[]) it.next();
            CacheFile file = RefreshCommandSupport.matchToFile(elements, fs.getCacheIdStr(), dir.getFile());
            if (file instanceof VcsCacheFile) ((VcsCacheFile) file).setLocal(false);
            if (file instanceof VcsCacheDir) {
                VcsCacheDir sDir = (VcsCacheDir) file;
                sDir.setLocal(false);
                Integer refresh = (Integer) refreshStrategyWhenNonLocal.get(sDir.getAbsolutePath());
                if (refresh != null) {
                    loadDir(sDir, refresh.intValue());
                }
            }
            //System.out.println("createSubFiles("+path+"): I've got file = "+file+"\nfile path = "+file.getAbsolutePath());
            //System.out.println("elements = "+VcsUtilities.arrayToString(elements)+" => file = "+file);
            dir.add(file);
            if (file instanceof CacheDir) {
                registerDir((CacheDir) file);
                //knownSubDirPaths.remove(file.getAbsolutePath());
            }
        }
        /*
        for(Iterator it = knownSubDirPaths.iterator(); it.hasNext(); ) {
            String subPath = (String) it.next();
            dir.removeChildDir(VcsUtilities.getFileNamePart(subPath), true);
        }
         */
        //System.out.println("VcsCache.createDir("+path+"): dir = "+dir);
        dir.setLoaded(true);
        //System.out.println("createSubFiles("+path+") = "+dir);
        return dir;
    }

    private VcsCacheDir createSubFiles(VcsFileSystem fs, VcsDirContainer filesByName) {
        Hashtable fileElements = (Hashtable) filesByName.getElement();
        Collection rawData;
        if (fileElements == null) rawData = Collections.EMPTY_LIST;
        else rawData = fileElements.values();
        VcsCacheDir dir = createSubFiles(fs, filesByName.getPath(), rawData);
        VcsDirContainer[] subfilesByName = filesByName.getSubdirContainers();
        for(int i = 0; i < subfilesByName.length; i++) {
            //VcsDirContainer subfilesByName = filesByName.getSubdir(subdirs[i]);
            dir.addChildDir(createSubFiles(fs, subfilesByName[i]), true);
        }
        return dir;
    }
    
    /**
     * The reading of file attributes was finished.
     * @param path the path of the file relative to the file system root.
     * @param rawData the data with attributes of one or more files.
     */
    public void readFileFinished(String path, Collection rawData) {
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs == null) return; // FS no longer exists
        VcsCacheDir dir = createSubFiles(fs, path, rawData);
        //System.out.println("readDirFinished: cacheDir = "+dir+", isBeingLoaded = "+dir.isBeingLoaded()+", isLoaded = "+dir.isLoaded());
        registerDir(dir);
        fireCacheHandlerEvent(EVENT_CHANGED, dir);
        heyDoRefreshDir(getPath(dir.getFile()));
    }
    
    /**
     * Called by VcsDirReader when asynchronous reading of directory is completed.
     * @param dir directory that was read by VcsDirReader relative to the filesystem root
     * @param rawData vector of <CODE>String[]</CODE> that describes files and subdirectories
     */
    public void readDirFinished(/*VcsCacheDir dir,*/String path, Collection rawData, boolean success) {
        //System.out.println("VcsCache.readDirFinished("+path+", rawData = "+rawData+", success = "+success);
        /*
        for (Iterator it = rawData.iterator(); it.hasNext(); ) {
            String[] elements = (String[]) it.next();
            //System.out.print(VcsUtilities.arrayToString(elements)+", ");
        } System.out.println("");
         */
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs == null) return; // FS no longer exists
        VcsCacheDir dir = createSubFiles(fs, path, rawData);
        //System.out.println("readDirFinished: cacheDir = "+dir+", isBeingLoaded = "+dir.isBeingLoaded()+", isLoaded = "+dir.isLoaded());
        registerDir(dir);
        dir.writeToDisk();
        dir.setBeingLoaded(false);
        fireCacheHandlerEvent(EVENT_CHANGED, dir);
        heyDoRefreshDir(getPath(dir.getFile()));
    }
    
    public void readDirFinishedRecursive(/*VcsCacheDir dir,*/VcsDirContainer rawData, boolean success) {
        //System.out.println("VcsCache.readDirFinishedRecursive("+rawData.getPath()+", rawData = "+rawData+", success = "+success);
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs == null) return; // FS no longer exists
        VcsCacheDir dir = createSubFiles(fs, rawData);
        registerDirRecursive(dir);
        dir.writeToDiskRecursively();
        fireCacheHandlerEvent(EVENT_CHANGED_RECURSIVELY, dir);
        heyDoRefreshDirRecursive(dir);
    }
    
    
    /**
     * Creates a list of FileObjects that are contents of the topNode.
     * Adds that list to the map of nodes being locked. 
     * If such node is being already locked, the lock is multiplied.
     */
    boolean lockFileObjects(CacheDir topNode, boolean recursively) {
        String path = topNode.getAbsolutePath();
        // make path relative to fs..
        String found = this.fsRoot;
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (path.length() <= found.length()) return false;
        path = path.substring(found.length());
        path = path.replace(File.separatorChar,'/');
        if (path.startsWith("/")) path = path.substring(1);
        FileObject topFO = fsystem.findResource(path);
//        System.out.println("topFO = " + topFO);
        return doLockFileObjects(topFO, topNode.getAbsolutePath(), recursively);
    }    
    
}
