/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.caching;

import java.lang.ref.*;
import java.util.HashMap;
import java.util.ArrayList;
import java.io.File;
import java.text.MessageFormat;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.util.NbBundle;

import org.netbeans.modules.vcscore.util.Debug;
import org.netbeans.modules.vcscore.util.VcsUtilities;
import org.netbeans.modules.vcscore.cache.*;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.VcsDirReader;

/**
 * This class creates and registeres cache for VCS filesystem.
 * It creates one VcsCache object for each filesystem, therefore each filesystem should provide
 * different cache ID.
 *
 * @author  Martin Entlicher
 */
public class VcsFSCache extends Object implements FileStatusProvider, FileCacheProvider {
    private Debug E=new Debug("CvsFsCache", true); // NOI18N
    private Debug D=E; 
        
    public final String localStatusStr = g("CTL_StatusLocal"); // NOI18N
    private String modifiedStatus = g("CTL_DefaultModifiedStatus"); // NOI18N
    private String notInSynchStatus = g("CTL_StatusNotInSync"); // NOI18N
    private String unknownStatus = g("CTL_StatusUnknown"); // NOI18N
    private HashMap fileStatuses = new HashMap();

    //private boolean localFilesAdd = true;

    private Reference fileSystem;
    private CacheHandler handler;
    private String cacheId;
    private String fsRoot;
    private String relativeMountPoint;

    /** Creates new VcsFSCache */
    public VcsFSCache(VcsFileSystem fileSystem) {
        this.fileSystem = new WeakReference(fileSystem);
        cacheId = fileSystem.getCacheIdStr();
        fsRoot = fileSystem.getFSRoot();
        relativeMountPoint = fileSystem.getRelativeMountPoint();
        handler = CacheHandler.getInstance();
        initFileStatuses();
        VcsCache cacheObj = new VcsCache(this.fileSystem, getCacheId());
        handler.registerCacheType(getCacheId(), cacheObj);
        cacheObj.addCacheHandlerListener((VcsFileSystem) this.fileSystem.get());
    }

    /* FileStatusProvider */
    
    private void initFileStatuses() {
        fileStatuses.put(localStatusStr, localStatusStr);
        fileStatuses.put(modifiedStatus, modifiedStatus);
        fileStatuses.put(notInSynchStatus, notInSynchStatus);
        fileStatuses.put(unknownStatus, unknownStatus);
    }
    
    public HashMap getPossibleFileStatusesTable() {
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs != null) {
            HashMap fst = fs.getPossibleFileStatusesTable();
            if (fst == null) return new HashMap(fileStatuses);
            else return fst;
        } else {
            return new HashMap();
        }
    }
    
    public HashMap getStatusIconMap() {
        HashMap iconStatusMap = new HashMap();
        iconStatusMap.put(modifiedStatus, new javax.swing.ImageIcon(getClass().getResource("/org/netbeans/modules/vcscore/caching/badgeLocModified.gif")).getImage());
        return iconStatusMap;
        //return new HashMap();
    }

    /**
     * Get the status that is displayed instead of the attribute value, when this
     * value differs for multiple files contained in the same data object file.
     */
    public String getNotInSynchStatus() {
        return (String) fileStatuses.get(notInSynchStatus);
    }
    
    public void setFileModified(String path) {
        setFileStatus(path, modifiedStatus);
    }
    
    public String getLocalFileStatus() {
        return localStatusStr;
    }
    
    public void setFileStatus(String path, String status) {
        CacheFile file = getFileOrDir(path);
        //System.out.println("setFileStatus("+path+", "+status+": file = '"+file+"'");
        if (file != null) {
            file.setStatus(status);
            file.getParent().writeToDisk();
        }
    }
    
    public String getFileStatus(String path){
        D.deb("getFileStatus('"+path+"')"); // NOI18N
        //System.out.println("  VcsFSCache.getFileStatus("+path+")");
        CacheFile file = getFileOrDir(path);
        String stat;
        //System.out.println("  VcsFSCache.getFileOrDir("+path+") = "+file);
        if (file != null) stat = file.getStatus();
        else stat = localStatusStr;
        //System.out.println("   VcsFSCache.status = "+stat);
        return stat;
    }
    
    public String getFileLocker(String path) {
        String locker = null;
        //if (isFile(path)) {
        CacheFile file = getFileOrDir(path);
        if (file != null) locker = file.getLocker();
        //}
        return locker;
    }
    
    public String getFileRevision(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return file.getRevision();
        else return null;
    }
    
    public String getFileSticky(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return file.getSticky();
        else return null;
    }
    
    public String getFileAttribute(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return file.getAttr();
        else return null;
    }

    public String getFileSize(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return ""+file.getSize();
        else return null;
    }

    public String getFileDate(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return file.getDate();
        else return null;
    }

    public String getFileTime(String path) {
        CacheFile file = getFileOrDir(path);
        if (file != null) return file.getTime();
        else return null;
    }

    public void refreshDir(String path) {
        //System.out.println("VcsFSCache.refreshDir("+path+")");
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs != null) {
            FileCacheProvider cache = fs.getCacheProvider();
            if (cache != null) cache.refreshCacheDir(path);
        }
    }
    
    public void refreshDirRecursive(String path) {
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs != null) {
            FileCacheProvider cache = fs.getCacheProvider();
            if (cache != null) cache.refreshCacheDirRecursive(path);
        }
    }
    
    /* FileCacheProvider */
    
    public String getCacheId() {
        return cacheId;
    }
    
    public void setFSRoot(String fsRoot) {
        this.fsRoot = fsRoot;
    }
    
    public void setRelativeMountPoint(String relativeMountPoint) {
        this.relativeMountPoint = relativeMountPoint;
    }
    
    private String getAbsolutePath(String path) {
        //System.out.println("VcsFSCache.getAbsolutePath("+path+")");
        //if (path.length() > 0) {
        path = fsRoot + File.separator + relativeMountPoint + File.separator + path.replace('/', File.separatorChar);
        while (path.endsWith(File.separator)) path = path.substring(0, path.length() - 1);
        //} else {
        //    path = fsRoot;
        //}
        //System.out.println(" -> return '"+path+"'");
        return path;
    }
    
    private CacheFile getFileOrDir(String path, int strategy) {
        //int currentStrategy = CacheHandler.STRAT_ALL;
        //System.out.println("VcsFSCache.getFileOrDir("+path+", "+strategy+")");
        VcsFileSystem fs = (VcsFileSystem) fileSystem.get();
        if (fs != null) {
            FileCacheProvider cache = fs.getCacheProvider();
            if (cache != null) {
                path = getAbsolutePath(path);
                //System.out.println("Calling handler.getCacheFile("+new File(path)+", "+adjustStrategy(strategy)+")");
                CacheFile file = handler.getCacheFile(new File(path), adjustStrategy(strategy), cache);
                //System.out.println("VcsFSCache.getFileOrDir: file = "+file);
                return file;
            }
        }
        return null;
    }
    
    private CacheFile getFileOrDir(String path) {

        return getFileOrDir(path, CacheHandler.STRAT_DISK_OR_REFRESH);

        /*
        FileSystemCache cache = handler.getCache(getCacheId());
        path = getAbsolutePath(path);
        String pDirName = VcsUtilities.getDirNamePart(path);
        CacheDir dir = cache.getDir(pDirName);
        if (dir == null) return null;
        String name = VcsUtilities.getFileNamePart(path);
        return dir.getFile(name);
         */
    }

    public CacheFile getFile(String path) {
        FileSystemCache cache = handler.getCache(getCacheId());
        path = getAbsolutePath(path);
        int sep = path.lastIndexOf(File.separatorChar);
        String pDirName = (sep < 0) ? "" : path.substring(0, sep);
        CacheDir dir = cache.getDir(pDirName);
        if (dir == null) return null;
        String name = (sep < 0) ? path : path.substring(sep + 1);
        return dir.getFile(name);
    }
    
    /**
     * Get the cache for the given directory.
     * @param path the directory path relative to the file system root
     */
    public CacheDir getDir(String path) {
        FileSystemCache cache = handler.getCache(getCacheId());
        path = getAbsolutePath(path);
        CacheDir dir = cache.getDir(path);
        return dir;
    }

    public boolean isFile(String path) {
        return (getFile(path) != null);
    }
    
    /**
     * Whether the path is a directory in the cache.
     * @param path the directory path relative to the file system root
     */
    public boolean isDir(String path) {
        //FileSystemCache cache = handler.getCache(getCacheId());
        //path = getAbsolutePath(path);
        return (getDir(path) != null);
    }
    
    /**
     * Try to asynchronously read the directory.
     */
    public void readDir(String path) {
        getFileOrDir(path); // downloads the dir
    }

    /*
     * Try to asynchronously read the directory.
     * @param strategy the desired strategy
     *
    public void readDir(String path, int strategy) {
        getFileOrDir(path, strategy); // downloads the dir
    }
     */

    /** For strategies higher than STRAT_DISK, make adjust depending
     * on the value of filesystem's offLine property.
     */
    private int adjustStrategy (int reqStrategy) {
        if (reqStrategy > CacheHandler.STRAT_DISK) {
            VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
            if (fsystem.isOffLine()) return CacheHandler.STRAT_DISK;
        }
        return reqStrategy;
    }
    /** For strategies higher than STRAT_DISK, permit intended action depending
     * on the value of filesystem's offLine property.
     */
    
    private boolean permitAction(int reqStrategy) {
        if (reqStrategy > CacheHandler.STRAT_DISK) {
            VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
            if (fsystem.isOffLine()) return false;
        }
        return true;
    }
    
    /*
    public void setLocalFilesAdd (boolean localFilesAdd) {
        this.localFilesAdd = localFilesAdd;
    }

    public boolean isLocalFilesAdd () {
        return localFilesAdd;
    }
     */
    
    public String[] getFilesAndSubdirs(String path) {
        VcsCacheDir dir = (VcsCacheDir) getDir(path);
        if (dir == null) return null;
        return dir.getFilesAndSubdirs();
    }

    /** Force refresh of directory from disk cache. Delete the directory only from memory cache
     * and read it.
     * @param path complete name of directory to refresh
     *
    public void refreshDirFromDiskCache(File dir) {
        CacheFile cfile = handler.getCacheFile(dir, CacheHandler.STRAT_DISK, getCacheId());
        //cfile.
    }
     */

    public void refreshCacheDir(String path) {
        //System.out.println("VcsFSCache.refreshCacheDir("+path+")");
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem.isOffLine()) {
            if (NotifyDescriptor.Confirmation.YES_OPTION.equals (
                TopManager.getDefault ().notify (new NotifyDescriptor.Confirmation (
                    g("DLG_RefreshCommandDisabled"), NotifyDescriptor.Confirmation.YES_NO_OPTION)))) { // NOI18N
                fsystem.setOffLine(false);
            } else {
                return ;
            }
        }
        FileSystemCache cache = handler.getCache(getCacheId());
        String apath = getAbsolutePath(path);
        CacheDir dir = cache.getDir(apath);
        //System.out.println("cache for path = "+apath+" is "+dir);
        if (dir == null) {
            apath = VcsUtilities.getDirNamePart(apath);
            dir = cache.getDir(apath);
            if (dir == null) {
                //dir = ((VcsCache) cache).initCacheDir(path);
                getFileOrDir(VcsUtilities.getDirNamePart(path));
                return ;
            }
        }
/* is part of FileSystemCache.loadDir()
 * happens automagically
         dir.removeAll();
 */
    // dunnot what this refreshDir() method  is  used for.. I guess the followng code could replace following line.
//        ((VcsCache) cache).runVcsDirReader(new File(path));
        ((VcsCacheDir) dir).setLoaded(false); // to force the reload
        //dir.checkServer();
        ((VcsCache) cache).doRefreshDir(dir, false);
        
        //VcsDirReader reader = ((VcsCache) cache).getVcsDirReader(new File(path));
        //CvsCacheClient client = cache.getCacheClient();
        //client.doRefreshFor(new File(getAbsolutePath()));
        //if (reader != null) {
        //    new Thread (reader, "VCS-DirReader").start(); // NOI18N
        //}
        //dir.checkServer();
    }

    public void refreshCacheDirRecursive(String path) {
        VcsFileSystem fsystem = (VcsFileSystem) fileSystem.get();
        if (fsystem.isOffLine()) {
            if (NotifyDescriptor.Confirmation.YES_OPTION.equals (
                TopManager.getDefault ().notify (new NotifyDescriptor.Confirmation (
                    g("DLG_RefreshCommandDisabled"), NotifyDescriptor.Confirmation.YES_NO_OPTION)))) { // NOI18N
                fsystem.setOffLine(false);
            } else {
                return ;
            }
        }
        FileSystemCache cache = handler.getCache(getCacheId());
        //CacheFile file = handler.getCacheFile(new File(path), adjustStrategy(strategy), cache);
        String fullPath = getAbsolutePath(path);
        CacheDir dir = cache.getDir(fullPath);
        if (dir == null) {
            fullPath = VcsUtilities.getDirNamePart(fullPath);
            dir = cache.getDir(fullPath);
            if (dir == null && "".equals(path)) {
                dir = (CacheDir) cache.getCacheFile(fsystem.getFile(path), CacheHandler.STRAT_DISK);
            }
            if (dir == null) return ;
        }
/* is part of FileSystemCache.loadDir()
 * happens automagically
         dir.removeAll();
 */
    // dunnot what this refreshDirRecursive() method  is  used for.. I guess the followng code could replace following line.
    //    ((VcsCache) cache).runVcsDirReaderRecursive(new File(path));
        ((VcsCacheDir) dir).setLoaded(false); // to force the reload
        //dir.checkServerRecursive();
        ((VcsCache) cache).doRefreshDir(dir, true);
        //VcsDirReader reader = ((VcsCache) cache).getVcsDirReaderRecursive(new File(path));
        //CvsCacheClient client = cache.getCacheClient();
        //client.doRefreshFor(new File(getAbsolutePath()));
        //if (reader != null) {
        //    new Thread (reader, "VCS-DirReader").start(); // NOI18N
        //}
    }

    public void refreshDirFromDiskCache(File fpath) {
        FileSystemCache cache = handler.getCache(getCacheId());
        String path = fpath.getAbsolutePath();
        CacheDir dir = cache.getDir(path);
        if (dir == null) {
            path = VcsUtilities.getDirNamePart(path);
            dir = cache.getDir(path);
            if (dir == null) return;
        }
        dir.readFromDisk();
    }
    
    public void addFolder(String path) {
        FileSystemCache cache = handler.getCache(getCacheId());
        cache.getCacheFile(new File(getAbsolutePath(path)), CacheHandler.STRAT_LOCAL);
    }

    public void addFile(String path) {
        FileSystemCache cache = handler.getCache(getCacheId());
        cache.getCacheFile(new File(getAbsolutePath(path)), CacheHandler.STRAT_LOCAL);
    }
    
    public void rename(String oldPath, String newPath) {
    }
    
    public void remove(String path, boolean wasDirectory) {
    //public void removeFromCache(File deletedFile, boolean wasDirectory) {
        CacheDir parent = null;
        if (wasDirectory) {
            CacheDir dir = getDir(path);
            if (dir != null) {
                parent = dir.getParent();
                if (parent != null) parent.removeChildDir(dir.getName(), true);
            }
        } else {
            CacheFile file = getFile(path);
            if (file != null) {
                parent = file.getParent();
                parent.removeFile(file.getName(), true);
            }
        }
        if (parent != null) parent.setComplete(false); //writeToDisk();
        /*
        D.deb("removeFromCache()","===================================================================");
        File parent = deletedFile.getParentFile();
        D.deb("removeFromCache()","parentpath="+parent.getAbsolutePath());
        CvsCacheDir parentCache = (CvsCacheDir)handler.getCacheFile(parent, CvsCacheHandler.STRAT_LOCAL, this);
        if (parentCache == null) {
            // we want to remove a chache node that doesn't have a parent -> needs to be root!
            // do remove it manually..
            CvsCacheDir deletedCache = (CvsCacheDir)handler.getCacheFile(deletedFile, CvsCacheHandler.STRAT_LOCAL, this);
            handler.getCache(this).unregisterDir(deletedCache);
            return;
        }
        if (wasDirectory) {
            D.deb("removeFromCache()","isDirectory");
            parentCache.removeChildDir(deletedFile.getName(), true);
        }
        else {
            D.deb("removeFromCache()","isFile");
            parentCache.removeFile(deletedFile.getName(), true);
        }
        parentCache.setComplete(false);
//        parentCache.writeToDisk(); // write to disk immediately .. could be removed..
        D.deb("removeFromCache()","--------------------------------------------------------------------");
         */
    }

    
    String g(String s) {
        //D.deb("getting "+s);
        return NbBundle.getBundle(VcsFSCache.class).getString (s);
    }
    String  g(String s, Object obj) {
        return MessageFormat.format (g(s), new Object[] { obj });
    }
    String g(String s, Object obj1, Object obj2) {
        return MessageFormat.format (g(s), new Object[] { obj1, obj2 });
    }
    String g(String s, Object obj1, Object obj2, Object obj3) {
        return MessageFormat.format (g(s), new Object[] { obj1, obj2, obj3 });
    }
}
