/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.cmdline.diff;

import java.util.*;
import java.io.*;

import org.netbeans.modules.vcscore.diff.AbstractDiff;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.Variables;
import org.netbeans.modules.vcscore.commands.*;
import org.netbeans.modules.vcscore.cmdline.*;
import org.netbeans.modules.vcscore.cmdline.exec.*;
import org.netbeans.modules.vcscore.util.*;

/**
 * The abstract diff command, that use a command line command to get the differences.
 * @author  Martin Entlicher
 */
public abstract class AbstractDiffCmdline extends Object implements VcsAdditionalCommand, CommandDataOutputListener {

    private Debug D=new Debug("AbstractDiffCmdline",true); // NOI18N
    private Debug E=D;

    private static final String TEXT_MIMETYPE = "text/plain";

    private transient String TMP_ROOT;
    private File tmpDir = null;
    private File tmpDir2 = null;
    private String tmpDirName = ""; // NOI18N
    private String tmpDir2Name = ""; // NOI18N
    Hashtable vars = null;

    private String rootDir = null;
    private String file = null;
    private String dir = null;
    
    private transient String diffDataRegex = null;
    private transient CommandOutputListener stdoutNRListener = null;
    private transient CommandOutputListener stderrNRListener = null;
    private transient CommandDataOutputListener stdoutListener = null;
    protected transient CommandDataOutputListener stderrListener = null;
    private String dataRegex = null;
    private String errorRegex = null;
    private String checkoutCmd = null;
    private String diffCmd = null;
    
    protected String diffOutRev1 = null;
    protected String diffOutRev2 = null;
    protected int outputType = -1;
    
    protected AbstractDiff diff = null;
    
    private VcsFileSystem fileSystem = null;

    private transient java.util.ResourceBundle resBundle = null;

    //static final long serialVersionUID =4356723902547341894L;
    /** Creates new AbstractDiffCmdline */
    public AbstractDiffCmdline() {
    }
    
    /** Set the VCS file system to use to execute commands.
     */
    public void setFileSystem(VcsFileSystem fileSystem) {
        this.fileSystem = fileSystem;
    }

    protected boolean checkEmpty(String str, String element) {
        if (str == null || str.length() == 0) {
            if (this.stderrListener != null) {
                String[] elements = { "Bad format of diff result: "+element }; // NOI18N
                stderrListener.outputData(elements);
            }
            E.deb("Bad format of diff result: "+element); // NOI18N
            return true;
        }
        return false;
    }

    private boolean checkOut(Hashtable vars, String file, String revision, String tmpDir) {
        VcsCommand cmd = fileSystem.getCommand(checkoutCmd);
        String varRevision = ""; // NOI18N
        //if (revision != null) varRevision = " -r "+revision+" "; // NOI18N
        if (revision != null) varRevision = ""+revision; // NOI18N
        vars.put("REVISION", varRevision); // NOI18N
        D.deb("varRevision = "+varRevision); // NOI18N
        vars.put("TEMPDIR", tmpDir); // NOI18N
        D.deb("checkOut Command: "+cmd); // NOI18N
        VcsCommandExecutor vce = fileSystem.getVcsFactory().getCommandExecutor(cmd, new Hashtable(vars));
        fileSystem.getCommandsPool().startExecutor(vce);
        /*
        String prepared = Variables.expand(vars,cmd, true);
        D.deb("checkOut prepared: "+prepared); // NOI18N
        if (stderrListener != null) {
            String[] command = { "CHECKOUT: "+prepared }; // NOI18N
            stderrListener.outputData(command);
        }
        //if (stderrNRListener != null) stderrNRListener.match("CHECKOUT: "+prepared); // NOI18N
        ExternalCommand ec = new ExternalCommand(prepared);
        //ec.setTimeout(((Long) vars.get("TIMEOUT")).longValue()); // NOI18N
        if (this.stdoutNRListener != null) ec.addStdoutListener(this.stdoutNRListener);
        if (this.stderrNRListener != null) ec.addStderrListener(this.stderrNRListener);
        if (this.stdoutListener != null) {
            try {
                ec.addStdoutRegexListener(this.stdoutListener, this.dataRegex);
            } catch (BadRegexException e) {
                if (stderrListener != null) {
                    String[] elements = { "CHECKOUT: Bad data regex "+dataRegex+"\n" }; // NOI18N
                    stderrListener.outputData(elements);
                }
            }
        }
        if (this.stderrListener != null) {
            try {
                ec.addStderrRegexListener(this.stderrListener, this.errorRegex);
            } catch (BadRegexException e) {
                String[] elements = { "CHECKOUT: Bad error regex "+errorRegex+"\n" }; // NOI18N
                stderrListener.outputData(elements);
            }
        }
         */
        fileSystem.getCommandsPool().waitToFinish(vce);
        if (vce.getExitStatus() != VcsCommandExecutor.SUCCEEDED) {
            //E.err("exec failed "+ec.getExitStatus()); // NOI18N
            return false;
        } else return true;
    }

    private boolean performDiff(String revision1, String revision2) {
        VcsCommand cmd = fileSystem.getCommand(diffCmd);
        if (revision1 != null) vars.put("REVISION1", revision1); // NOI18N
        else vars.put("REVISION1", ""); // Put an empty variable to disable the error output about undefined variables // NOI18N
        if (revision2 != null) vars.put("REVISION2", revision2); // NOI18N
        else vars.put("REVISION2", ""); // Put an empty variable to disable the error output about undefined variables // NOI18N
        D.deb("diff command: "+cmd); // NOI18N
        //System.out.println("diff command: "+cmd); // NOI18N
        VcsCommandExecutor vce = fileSystem.getVcsFactory().getCommandExecutor(cmd, new Hashtable(vars));
        vce.addDataOutputListener(this);
        fileSystem.getCommandsPool().startExecutor(vce);
        fileSystem.getCommandsPool().waitToFinish(vce);
        diffFinished();
        if (vce.getExitStatus() != VcsCommandExecutor.SUCCEEDED) {
            //D.deb("exec failed "+ec.getExitStatus()); // NOI18N
            return false;
        }
        return true;
    }
    
    /**
     * This method is called when the diff command finishes.
     * Allows subclasess to do some cleanup of diff actions
     */
    protected void diffFinished() {
    }
    
    /**
     * Executes the checkout and diff commands and display differences.
     * @param vars variables needed to run cvs commands
     * @param args the arguments. At least three arguments have to be present:
     *             the diff output type, the checkout command and the diff command.
     *             Additionally, the first two arguments can be the revision tags to be compared.
     * @param stdoutNRListener listener of the standard output of the command
     * @param stderrNRListener listener of the error output of the command
     * @param stdoutListener listener of the standard output of the command which
     *                       satisfies regex <CODE>dataRegex</CODE>
     * @param dataRegex the regular expression for parsing the standard output
     * @param stderrListener listener of the error output of the command which
     *                       satisfies regex <CODE>errorRegex</CODE>
     * @param errorRegex the regular expression for parsing the error output
     * @return true if the command was succesfull,
     *         false if some error has occured.
     */
    public boolean exec(Hashtable vars, String[] args,
                        CommandOutputListener stdoutNRListener,
                        CommandOutputListener stderrNRListener,
                        CommandDataOutputListener stdoutListener, String dataRegex,
                        CommandDataOutputListener stderrListener, String errorRegex) {
        boolean status = true;
        this.stdoutNRListener = stdoutNRListener;
        this.stderrNRListener = stderrNRListener;
        this.stdoutListener = stdoutListener;
        this.dataRegex = dataRegex;
        this.stderrListener = stderrListener;
        this.errorRegex = errorRegex;
        this.vars = vars;
        int arglen = args.length;
        //System.out.println("DIFF: args = "+VcsUtilities.arrayToString(args));
        if (arglen < 3) {
            String message = "Too few arguments to Diff command !";
            String[] elements = { message };
            if (stderrListener != null) stderrListener.outputData(elements);
            if (stderrNRListener != null) stderrNRListener.outputLine(message);
            return false;
        }
        try {
            this.outputType = Integer.parseInt(args[arglen - 3]);
        } catch(NumberFormatException e) {
            String message = "Bad output type specification:"+args[arglen - 3];
            String[] elements = { message };
            if (stderrListener != null) stderrListener.outputData(elements);
            if (stderrNRListener != null) stderrNRListener.outputLine(message);
            return false;
        }
        //System.out.println("outputType = "+outputType);
        D.deb("outputType = "+outputType);
        this.checkoutCmd = args[arglen - 2];
        this.diffCmd = args[arglen - 1];
        String mime = (String) vars.get("MIMETYPE"); // NOI18N
        if (mime == null || mime.indexOf("unknown") >= 0) { // NOI18N
            vars.put("MIMETYPE", TEXT_MIMETYPE);
            mime = TEXT_MIMETYPE;
        }
        final String mimeType = mime;
        this.rootDir = (String) vars.get("ROOTDIR"); // NOI18N
        String module = (String) vars.get("MODULE"); // NOI18N
        if (module == null) module = ""; // NOI18N
        if (module.length() > 0) module += File.separator;
        //this.dir = (String) vars.get("DIR"); // NOI18N
        this.dir = module + (String) vars.get("DIR"); // NOI18N
        this.file = (String) vars.get("FILE"); // NOI18N
        tmpDir = VcsUtilities.createTMP();
        //tmpDirName = tmpDir.getName();
        tmpDirName = tmpDir.getAbsolutePath();
        String path = rootDir+File.separator+dir+File.separator+file;
        this.diffDataRegex = (String) vars.get("DATAREGEX"); // NOI18N
        if (this.diffDataRegex == null) //this.diffDataRegex = "(^.*)$"; // NOI18N
            //this.diffDataRegex = "(^[0-9]+(,[0-9]+|)[d][0-9]+$)|(^[0-9]+(,[0-9]+|)[c][0-9]+(,[0-9]+|)$)|(^[0-9]+[a][0-9]+(,[0-9]+|)$)|(^retrieving revision.*)"; // NOI18N
            this.diffDataRegex = "(^.*$)"; // NOI18N
        String revision1 = null;
        String revision2 = null;
        if (args.length > 3) {
            revision1 = args[0];
            if (args.length > 4) {
                revision2 = args[1];
            }
        }
        diff = new AbstractDiff();
        boolean diffStatus = performDiff(revision1, revision2);
        if (diff.getNumActions() == 0) {
            if (diffStatus == true) {
                org.openide.TopManager.getDefault ().notify (new org.openide.NotifyDescriptor.Message(java.text.MessageFormat.format (
                    org.openide.util.NbBundle.getBundle(AbstractDiff.class).getString("NoDifferenceInFile"), new Object[] { file } )));
                return true;
            } else {
                return false;
            }
        }
        if (diffOutRev1 != null) revision1 = diffOutRev1;
        if (diffOutRev2 != null) revision2 = diffOutRev2;
        //System.out.println("revision1 = "+revision1+", revision2 = "+revision2);
        if (revision2 != null) {
            tmpDir2 = VcsUtilities.createTMP();
            tmpDir2Name = tmpDir2.getAbsolutePath();
        }
        status = checkOut(vars, dir+File.separator+file, revision1, tmpDirName);
        if (!status) {
            closing();
            return status;
        }
        if (revision2 != null) {
            status = checkOut(vars, dir+File.separator+file, revision2, tmpDir2Name);
            if (!status) {
                closing();
                return status;
            }
        }
        final String file1Title = (revision1 == null) ? g("Diff.titleHeadRevision") : g("Diff.titleRevision", revision1); // NOI18N
        final String file2Title = (revision2 == null) ? g("Diff.titleWorkingFile") : g("Diff.titleRevision", revision2); // NOI18N
        javax.swing.SwingUtilities.invokeLater(new Runnable () {
                                                   public void run () {
                                                       String file1 = tmpDir+File.separator+file;
                                                       String file2;
                                                       if (tmpDir2 == null) {
                                                           file2 = rootDir+File.separator+dir+File.separator+file;
                                                       } else {
                                                           file2 = tmpDir2+File.separator+file;
                                                       }
                                                       diff.open(java.text.MessageFormat.format (org.openide.util.NbBundle.getBundle(AbstractDiff.class).
                                                                                        getString("DiffComponent.titleFile"), new Object[] { file }),
                                                            mimeType, file1, file2, file1Title, file2Title);
                                                       diff.addCloseListener(new TopComponentCloseListener() {
                                                           public void closing() {
                                                               AbstractDiffCmdline.this.closing();
                                                           }
                                                       });
                                                   }
                                               });
        D.deb("exec return = "+status); // NOI18N
        return status;
    }

    /*
     * Called when the diff is being closed. Cleanup of temporary files.
     */
    public void closing() {
        //new File(tmpDir, file).delete();
        //System.out.println("AbstractDiffCmdline.closing(), delete "+tmpDir+" & "+tmpDir2);
        D.deb("deleting "+tmpDir); // NOI18N
        VcsUtilities.deleteRecursive(tmpDir);
        if (tmpDir2 != null) {
            D.deb("deleting "+tmpDir2); // NOI18N
            VcsUtilities.deleteRecursive(tmpDir2);
        }
    }

    protected void setBundle(java.util.ResourceBundle bundle) {
        this.resBundle = bundle;
    }

    private String g(String s) {
        D.deb("getting "+s);
        if (resBundle == null) {
            throw new java.util.MissingResourceException(s, "?", s);
        }
        return resBundle.getString(s);
    }

    private String g(String s, Object obj) {
        return java.text.MessageFormat.format (g(s), new Object[] { obj });
    }

    abstract public void outputData(String[] elements);

}
