/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.diff;

import java.util.*;
import java.io.*;
import java.net.URL;
import javax.swing.*;
import javax.swing.text.*;

import org.openide.nodes.*;
import org.openide.loaders.DataObject;
import org.openide.filesystems.FileObject;
import org.openide.cookies.*;
import org.openide.util.RequestProcessor;

import org.netbeans.modules.vcscore.util.*;

/**
 *
 * @author  Martin Entlicher
 * @version 1.0
 * This class implements color diff display of two files.
 */
public class AbstractDiff extends Object implements Serializable {

    private Debug D=new Debug("AbstractDiff",true); // NOI18N

    private static final java.awt.Color colorMissing = new java.awt.Color(255, 160, 180);
    private static final java.awt.Color colorAdded = new java.awt.Color(180, 255, 180);
    private static final java.awt.Color colorChanged = new java.awt.Color(160, 200, 255);
    
    //private static transient String TMP_ROOT="vcs/tmp"; // NOI18N
    private String mimeType = null;
    private URL url1 = null;
    private URL url2 = null;
    private String title1 = null;
    private String title2 = null;
    private String mainTitle = null;

    //private StringBuffer diffBuffer = new StringBuffer(4096);
    private Vector diffActions = new Vector();
    private volatile transient DiffComponent diffFrame = null;
    private int currentDiffLine = -1;
    
    //private transient java.util.ResourceBundle resBundle = null;

    static final long serialVersionUID =2364195735462564757L;
    /** Creates new CvsDiff */
    public AbstractDiff() {
    }

    protected DiffComponent createDiffComponent(AbstractDiff diff) {
        return new DiffComponent(diff);
    }
    
    /**
     * Should be called after deserialization to initialize the TopComponent
     */
    boolean setDiffComponent(DiffComponent diffFrame) {
        this.diffFrame = diffFrame;
        if (mimeType != null) {
            diffFrame.setMimeType1(mimeType);
            diffFrame.setMimeType2(mimeType);
        }
        try {
            diffFrame.setFile1(url1);
            diffFrame.setFile2(url2);
        } catch (IOException e) {
            //D.err("IO Exception "+e.getMessage()); // NOI18N
            //org.openide.TopManager.getDefault().notifyException(e);
            return false;
        }
        diffFrame.setFile1Title(title1);
        diffFrame.setFile2Title(title2);
        diffFrame.setName(mainTitle/*java.text.MessageFormat.format (org.openide.util.NbBundle.getBundle(AbstractDiff.class).
                                  getString("DiffComponent.titleFile"), new Object[] { file })*/); // NOI18N
        javax.swing.SwingUtilities.invokeLater(new Runnable () {
            public void run () {
                insertEmptyLines(false);
                javax.swing.SwingUtilities.invokeLater(new Runnable () {
                    public void run () {
                        setDiffHighlight(true);
                    }
                });
            }
        });
        return true;
    }
    
    public void addAddAction(int n1, int n3, int n4) {
        DiffAction action = new DiffAction();
        action.setAddAction(n1, n3, n4);
        diffActions.add(action);
    }
    
    public void addDeleteAction(int n1, int n2, int n3) {
        DiffAction action = new DiffAction();
        action.setDeleteAction(n1, n2, n3);
        diffActions.add(action);
    }
    
    public void addChangeAction(int n1, int n2, int n3, int n4) {
        DiffAction action = new DiffAction();
        action.setChangeAction(n1, n2, n3, n4);
        diffActions.add(action);
    }
    
    public int getNumActions() {
        return diffActions.size();
    }
    
    /**
     * Open the diff view.
     */
    public void open(final String mainTitle, final String mimeType,
                     final String file1, final String file2,
                     final String title1, final String title2) {
        setTitle(mainTitle);
        setMIMEType(mimeType);
        RequestProcessor.Task task = RequestProcessor.postRequest(new Runnable() {
            public void run() {
                openEditors(file1, file2, title1, title2);
                insertEmptyLines(true);
                setDiffHighlight(true);
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        diffFrame.open();
                    }
                });
            }
        });
        //task.run();
    }
    
    public void addCloseListener(TopComponentCloseListener listener) {
        if (diffFrame == null) diffFrame = createDiffComponent(this);
        diffFrame.addCloseListener(listener);
    }

    /**
     * Set the MIME type of the display content.
     */
    private void setMIMEType(String mimeType) {
        this.mimeType = mimeType;
    }
    
    /**
     * Set the title of the visual diff component.
     */
    private void setTitle(String title) {
        this.mainTitle = title;
        if (diffFrame == null) diffFrame = createDiffComponent(this);
        diffFrame.setName(title);
    }
    
    /** Open the files in two editor panes. Should NOT be called from an AWT thread.
     */
    private boolean openEditors(String file1, String file2, final String title1, final String title2) {
        D.deb("I have MIME = "+mimeType); // NOI18N
        this.title1 = title1;
        this.title2 = title2;
        url1 = null;
        url2 = null;
        try {
            url1 = new File(file1).toURL();
            url2 = new File(file2).toURL();
        } catch (java.net.MalformedURLException e) {
            D.deb("MalformedURLException "+e.getMessage()); // NOI18N
            return false;
        }
        if (diffFrame == null) {
            try {
                SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        diffFrame = createDiffComponent(AbstractDiff.this);
                    }
                });
            } catch (InterruptedException iexc) {
                return false;
            } catch (java.lang.reflect.InvocationTargetException invexc) {
                return false;
            }
        }
        //VcsUtilities.centerWindow(diffFrame);
        if (mimeType != null) {
            diffFrame.setMimeType1(mimeType);
            diffFrame.setMimeType2(mimeType);
        }
        try {
            diffFrame.setFile1(url1);
            diffFrame.setFile2(url2);
        } catch (IOException e) {
            org.openide.TopManager.getDefault().notifyException(e);
            return false;
        }
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                diffFrame.setFile1Title(title1);
                diffFrame.setFile2Title(title2);
            }
        });
        return true;
    }

    private void insertEmptyLines(boolean updateActionLines) {
        int n = diffActions.size();
        int ins1 = 0;
        int ins2 = 0;
        D.deb("insertEmptyLines():"); // NOI18N
        for(int i = 0; i < n; i++) {
            DiffAction action = (DiffAction) diffActions.get(i);
            int n1 = action.getF1Line1() + ins1;
            int n2 = action.getF1Line2() + ins1;
            int n3 = action.getF2Line1() + ins2;
            int n4 = action.getF2Line2() + ins2;
            D.deb("Action: "+action.getAction()+": ("+n1+","+n2+","+n3+","+n4+")"); // NOI18N
            D.deb("ins1 = "+ins1+", ins2 = "+ins2); // NOI18N
            switch (action.getAction()) {
            case DiffAction.DELETE:
                diffFrame.addEmptyLines2(n3, n2 - n1 + 1);
                ins2 += n2 - n1 + 1;
                break;
            case DiffAction.ADD:
                diffFrame.addEmptyLines1(n1, n4 - n3 + 1);
                ins1 += n4 - n3 + 1;
                break;
            case DiffAction.CHANGE:
                int r1 = n2 - n1;
                int r2 = n4 - n3;
                if (r1 < r2) {
                    diffFrame.addEmptyLines1(n2, r2 - r1);
                    ins1 += r2 - r1;
                } else if (r1 > r2) {
                    diffFrame.addEmptyLines2(n4, r1 - r2);
                    ins2 += r1 - r2;
                }
                break;
            }
            if (updateActionLines) {
                action.setF1Line1(n1);
                action.setF1Line2(n2);
                action.setF2Line1(n3);
                action.setF2Line2(n4);
            }
        }
    }

    private void setDiffHighlight(boolean set) {
        int n = diffActions.size();
        D.deb("Num Actions = "+n); // NOI18N
        for(int i = 0; i < n; i++) {
            DiffAction action = (DiffAction) diffActions.get(i);
            int n1 = action.getF1Line1();
            int n2 = action.getF1Line2();
            int n3 = action.getF2Line1();
            int n4 = action.getF2Line2();
            D.deb("Action: "+action.getAction()+": ("+n1+","+n2+","+n3+","+n4+")"); // NOI18N
            switch (action.getAction()) {
            case DiffAction.DELETE:
                if (set) diffFrame.highlightRegion1(n1, n2, colorMissing);
                else diffFrame.highlightRegion1(n1, n2, java.awt.Color.white);
                break;
            case DiffAction.ADD:
                if (set) diffFrame.highlightRegion2(n3, n4, colorAdded);
                else diffFrame.highlightRegion2(n3, n4, java.awt.Color.white);
                break;
            case DiffAction.CHANGE:
                if (set) {
                    diffFrame.highlightRegion1(n1, n2, colorChanged);
                    diffFrame.highlightRegion2(n3, n4, colorChanged);
                } else {
                    diffFrame.highlightRegion1(n1, n2, java.awt.Color.white);
                    diffFrame.highlightRegion2(n3, n4, java.awt.Color.white);
                }
                break;
            }
        }
    }

    private int getActionLine(int actionNum) {
        DiffAction action = (DiffAction) diffActions.get(actionNum);
        int line = action.getF1Line1();
        if (action.getAction() == DiffAction.ADD) line++;
        return line;
    }

    int getNextDiffLine() {
        if (diffActions.size() <= 0) return -1;
        currentDiffLine++;
        if (currentDiffLine >= diffActions.size()) currentDiffLine = 0;
        return getActionLine(currentDiffLine);
    }

    int getPrevDiffLine() {
        if (diffActions.size() <= 0) return -1;
        currentDiffLine--;
        if (currentDiffLine < 0) currentDiffLine = diffActions.size() - 1;
        return getActionLine(currentDiffLine);
    }
    /*
    public void diffAgain() {
      //setDiffHighlight(false);
      diffFrame.unhighlightAll();
      diffActions.removeAllElements();
      performDiff();
      setDiffHighlight(true);
}
    */
    
    /**
     * Called when the diff is being closed. Allows subclasses to make some cleanup. Empty by default.
     *
    void closing() {
    }
    */
    /*
    public void setBundle(java.util.ResourceBundle bundle) {
        this.resBundle = bundle;
    }

    private String g(String s) {
        D.deb("getting "+s);
        if (resBundle == null) {
            throw new java.util.MissingResourceException(s, "?", s);
        }
        return resBundle.getString(s);
    }

    private String g(String s, Object obj) {
        return java.text.MessageFormat.format (g(s), new Object[] { obj });
    }
     */
    
    private final class DiffAction extends Object implements Serializable {
        public static final int DELETE = 0;
        public static final int CHANGE = 1;
        public static final int ADD = 2;

        private int action = 0;
        private int f1Line1 = 0;
        private int f1Line2 = 0;
        private int f2Line1 = 0;
        private int f2Line2 = 0;

        public DiffAction()  {
        }

        public void setDeleteAction(int f1Line1, int f1Line2, int f2Line1) {
            this.action = DELETE;
            this.f1Line1 = f1Line1;
            this.f1Line2 = f1Line2;
            this.f2Line1 = f2Line1;
        }

        public void setAddAction(int f1Line1, int f2Line1, int f2Line2) {
            this.action = ADD;
            this.f1Line1 = f1Line1;
            this.f2Line1 = f2Line1;
            this.f2Line2 = f2Line2;
        }

        public void setChangeAction(int f1Line1, int f1Line2, int f2Line1, int f2Line2) {
            this.action = CHANGE;
            this.f1Line1 = f1Line1;
            this.f1Line2 = f1Line2;
            this.f2Line1 = f2Line1;
            this.f2Line2 = f2Line2;
        }

        public int getAction() {
            return this.action;
        }

        public int getF1Line1() {
            return this.f1Line1;
        }

        public void setF1Line1(int value) {
            this.f1Line1 = value;
        }

        public int getF1Line2() {
            return this.f1Line2;
        }

        public void setF1Line2(int value) {
            this.f1Line2 = value;
        }

        public int getF2Line1() {
            return this.f2Line1;
        }

        public void setF2Line1(int value) {
            this.f2Line1 = value;
        }

        public int getF2Line2() {
            return this.f2Line2;
        }

        public void setF2Line2(int value) {
            this.f2Line2 = value;
        }
    }
}
