/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.runtime;

import java.util.Hashtable;

import org.openide.TopManager;
import org.openide.nodes.*;

import org.netbeans.modules.vcscore.commands.VcsCommandExecutor;
import org.netbeans.modules.vcscore.commands.CommandsPool;

/**
 * Takes care about nodes in Runtime tab.
 *
 * @author  Martin Entlicher
 */
public class RuntimeSupport extends Object {

    private static final String VCS_RUNTIME_NODE_NAME = "VcsRuntime";
    
    /** The table containing pairs of executors as keys and their runtime command nodes
     * as values.
     */
    private static Hashtable commands = new Hashtable();

    /** Creates new RuntimeSupport */
    private RuntimeSupport() {
    }
    
    /**
     * Initializes the runtime folder for command nodes from the filesystem.
     * @param fsName the name of the filesystem
     * @return the node of the runtime folder
     */
    public static RuntimeFolderNode initRuntime(String fsName) {
        Node runtime = TopManager.getDefault().getPlaces().nodes().environment();
        Children runtimeCh = runtime.getChildren();
        Node vcsRuntime = getVcsRuntime(runtimeCh);
        Children fsCh = new Index.ArrayChildren();
        RuntimeFolderNode fsRuntime = new RuntimeFolderNode(fsCh);
        fsRuntime.setName(fsName);
        fsRuntime.setDisplayName(fsName);
        vcsRuntime.getChildren().add(new Node[] { fsRuntime });
        return fsRuntime;
    }
    
    private static Node getVcsRuntime(Children ch) {
        Node[] nodes = ch.getNodes();
        for(int i = 0; i < nodes.length; i++) {
            if (VCS_RUNTIME_NODE_NAME.equals(nodes[i].getName())) return nodes[i];
        }
        AbstractNode vcsRuntime = new AbstractNode(new Children.Array());
        vcsRuntime.setName(VCS_RUNTIME_NODE_NAME);
        vcsRuntime.setDisplayName(g("CTL_VcsRuntime"));
        ch.add(new Node[] { vcsRuntime });
        return vcsRuntime;
    }
    
    /**
     * Update the runtime folder name with the new file system display name.
     */
    public static void updateRuntime(Node node, String fsName) {
        node.setName(fsName);
        node.setDisplayName(fsName);
    }
    
    /**
     * Add a waiting command.
     */
    public static void addWaiting(Node fsRuntime, VcsCommandExecutor vce, CommandsPool cpool) {
        if (fsRuntime == null) return ;
        Children ch = fsRuntime.getChildren();
        RuntimeCommandNode cmd = new RuntimeCommandNode(vce, cpool);
        cmd.setState(RuntimeCommandNode.STATE_WAITING);
        commands.put(vce, cmd);
        ch.add(new Node[] { cmd });
    }
    
    /**
     * Add a running command.
     */
    public static void addRunning(Node fsRuntime, VcsCommandExecutor vce, CommandsPool cpool) {
        if (fsRuntime == null) return ;
        synchronized (fsRuntime) {
        Children ch = fsRuntime.getChildren();
        RuntimeCommandNode cmd = (RuntimeCommandNode) commands.get(vce);
        if (cmd == null) {
            cmd = new RuntimeCommandNode(vce, cpool);
            commands.put(vce, cmd);
            ch.add(new Node[] { cmd });
        }
        //ch.remove(new Node[] { cmd });
        cmd.setState(RuntimeCommandNode.STATE_RUNNING);
        //ch.add(new Node[] { cmd });
        }
    }
    
    /**
     * Add a done command.
     */
    public static void addDone(Node fsRuntime, VcsCommandExecutor vce, CommandsPool cpool) {
        if (fsRuntime == null) return ;
        synchronized (fsRuntime) {
        Children ch = fsRuntime.getChildren();
        RuntimeCommandNode cmd = (RuntimeCommandNode) commands.get(vce);
        if (cmd == null) {
            cmd = new RuntimeCommandNode(vce, cpool);
            commands.put(vce, cmd);
            ch.add(new Node[] { cmd });
        }
        //ch.remove(new Node[] { cmd });
        cmd.setState(RuntimeCommandNode.STATE_DONE);
        //ch.add(new Node[] { cmd });
        }
    }
    
    /**
     * Add a done command.
     */
    public static void addCancelled(Node fsRuntime, VcsCommandExecutor vce, CommandsPool cpool) {
        if (fsRuntime == null) return ;
        synchronized (fsRuntime) {
        Children ch = fsRuntime.getChildren();
        RuntimeCommandNode cmd = (RuntimeCommandNode) commands.get(vce);
        if (cmd == null) {
            cmd = new RuntimeCommandNode(vce, cpool);
            commands.put(vce, cmd);
            ch.add(new Node[] { cmd });
        }
        //ch.remove(new Node[] { cmd });
        cmd.setState(RuntimeCommandNode.STATE_CANCELLED);
        //ch.add(new Node[] { cmd });
        }
    }
    
    /**
     * Remove a command.
     */
    public static void removeDone(VcsCommandExecutor vce) {
        Node cmd = (Node) commands.remove(vce);
        if (cmd == null) return ;
        try {
            cmd.destroy();
        } catch (java.io.IOException exc) {
            TopManager.getDefault().notifyException(exc);
        }
        /*
        Children ch = cmd.getParentNode().getChildren();
        ch.remove(new Node[] { cmd });
         */
    }

    private static String g(String s) {
        return org.openide.util.NbBundle.getBundle(RuntimeSupport.class).getString(s);
    }
    
}
