/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.wizard.mountcvs;

import java.lang.reflect.*;

import org.netbeans.modules.vcscore.VcsFileSystem;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.Repository;
import org.openide.filesystems.RepositoryEvent;
import org.openide.filesystems.RepositoryReorderedEvent;
import org.openide.filesystems.RepositoryListener;
import org.openide.filesystems.AbstractFileSystem;
import org.openide.util.SharedClassObject;
import org.netbeans.modules.vcscore.settings.VcsSettings;

import org.openide.util.NbBundle;
import org.openide.util.WeakListener;

/**
 * A "virtual" filesystem, that does nothing, but have a joined CVS wizard
 * as a customizer. After it's added to the Repository, it replace itself by the
 * right CVS filesystem.
 *
 * @author  Martin Entlicher
 */
public class CvsMountFS extends AbstractFileSystem {

    private static final String FS_CMDLINE = "org.netbeans.modules.vcs.cmdline.CvsFileSystem";
    private static final String FS_JAVA = "org.netbeans.modules.cvsclient.JavaCvsFileSystem";

    //private static final String SET_WORK = "setRootDir
    private CvsWizardData data = null;

    static final long serialVersionUID = 1135495669383440118L;

    /** Creates new CvsMountFS */
    public CvsMountFS() {
        //System.out.println("CvsMountFS()");
        /*
        Repository r = TopManager.getDefault().getRepository();
        r.addRepositoryListener(WeakListener.repository(new RepositoryListener() {
            public void fileSystemAdded(RepositoryEvent ev) {
                System.out.println("fs = "+ev.getFileSystem()+" added.");
                if (ev.getFileSystem() == CvsMountFS.this) {
                    mountClientFS();
                    TopManager.getDefault().getRepository().removeFileSystem(CvsMountFS.this);
                }
            }
            public void fileSystemRemoved(RepositoryEvent ev) {
            }
            public void fileSystemPoolReordered(RepositoryReorderedEvent rev) {
            }
        }, r)); 
         */
        list = new AbstractFileSystem.List() {
            public String[] children(String name) {
                return new String[0];
            }
        };
    }

    /**
     * Notifies this file system that it has been added to the repository.
     */
    public void addNotify() {
        TopManager.getDefault().getRepository().removeFileSystem(this);
        if (data != null) return ;
        try {
            org.openide.nodes.BeanNode bn = new org.openide.nodes.BeanNode(this);
            java.awt.Component c = (java.awt.Component) bn.getCustomizer ();
            if (c instanceof java.awt.Window) {
                c.setVisible (true);
            }
        } catch (java.beans.IntrospectionException exc) {
            TopManager.getDefault().notifyException(exc);
        }
    }
    
    /**
     * Get a human presentable name of the file system
     */
    public String getDisplayName() {
        return NbBundle.getMessage(CvsMountFS.class, "LAB_CVSFileSystem");
    }
    
    /* Test whether file system is read only.
     * @return <code>true</code>
     */
    public boolean isReadOnly() {
        return true;
    }
    
    public void setWizardData(final CvsWizardData data) {
        this.data = data;
        org.openide.util.RequestProcessor.postRequest(new Runnable() {
            public void run() {
                mountClientFS(data);
            }
        });
    }
    
    /**
     * @return the client filesystem properly setted up or null, when not found.
     */
    public static Object setupClientFS(CvsWizardData data) {
        if (data != null) {
            Class mountFSClass;
            try {
                if (data.isBuildInCvs()) {
                    mountFSClass = Class.forName(FS_JAVA, true, TopManager.getDefault().currentClassLoader());
                } else {
                    mountFSClass = Class.forName(FS_CMDLINE, true, TopManager.getDefault().currentClassLoader());
                }
            } catch (ClassNotFoundException exc) {
                mountFSClass = null;
            }
            if (mountFSClass == null) {
                TopManager.getDefault().notify(new NotifyDescriptor.Message(NbBundle.getMessage(CvsMountFS.class, "MSG_FS_Not_Available")));
                return null;
            }
            Object mountFS = createClientFS(mountFSClass);
            if (mountFS == null) return null;
            if (data.isBuildInCvs()) {
                setBuiltInFSParams(mountFS, mountFSClass, data);
            } else {
                setCmdLineFSParams(mountFS, mountFSClass, data);
            }
            return mountFS;
        }
        return null;
    }
    
    /** Mount the client filesystem instead of this */
    private static void mountClientFS(CvsWizardData data) {
        //System.out.println("mountClientFS(): data = "+data);
        String[] mountPoints = data.getMountPoints();
        if (mountPoints == null || mountPoints.length == 0) {
            mountPoints = new String[] { data.getMountPoint() };
        }
        for (int i = 0; i < mountPoints.length; i++) {
            data.setMountPointOnly(mountPoints[i]);
            Object mountFS = setupClientFS(data);
            if (mountFS != null) {
                TopManager.getDefault().getRepository().addFileSystem((FileSystem) mountFS);
            }
        }
    }
    
    private static Object createClientFS(Class mountFSClass) {
        Object mountFS;
        try {
            mountFS = mountFSClass.newInstance();
        } catch (final InstantiationException instExc) {
            TopManager.getDefault().notifyException(new Exception() {
                public String getLocalizedMessage() {
                    return instExc.getMessage();
                }
            });
            mountFS = null;
        } catch (final IllegalAccessException iaExc) {
            TopManager.getDefault().notifyException(new Exception() {
                public String getLocalizedMessage() {
                    return iaExc.getMessage();
                }
            });
            mountFS = null;
        }
        return mountFS;
    }
    
    private static void setBuiltInFSParams(Object mountFS, Class mountFSClass, CvsWizardData data) {
        // needs to be setup rightly before login is performed.
        VcsSettings settings = (VcsSettings)SharedClassObject.findObject(
                                   VcsSettings.class, true); 
        if (data.getHome() != null) {
            settings.setHome(data.getHome());
            try {
                final String SETT_CLIENT = "org.netbeans.modules.cvsclient.JavaCvsSettings";
                Class jSettingsClass = Class.forName(SETT_CLIENT, true, TopManager.getDefault().currentClassLoader());
                
                SharedClassObject jSettings = SharedClassObject.findObject(jSettingsClass, true);
                setFSParam(jSettings, jSettingsClass, "setCvsHome", java.io.File.class, new java.io.File(data.getHome()));
            } catch (ClassNotFoundException exc) {
            }
        }
        setFSParam(mountFS, mountFSClass, "setWorkingDir", java.io.File.class, new java.io.File(data.getWorkingDir()));
        setFSParam(mountFS, mountFSClass, "setRelMount", String.class, data.getMountPoint());
        int type = data.getMethodType();
        if (type == org.netbeans.modules.vcscore.wizard.mountcvs.util.CvsHelper.T_PSERVER) {
            setFSParam(mountFS, mountFSClass, "setCvsServerType", Integer.TYPE, new Integer(1));
        }
        if (type == org.netbeans.modules.vcscore.wizard.mountcvs.util.CvsHelper.T_SERVER) {
            setFSParam(mountFS, mountFSClass, "setCvsServerType", Integer.TYPE, new Integer(0));
        }
        setFSParam(mountFS, mountFSClass, "setCvsServerName", String.class, data.getHostName());
        setFSParam(mountFS, mountFSClass, "setCvsUserName", String.class, data.getUserName());
        setFSParam(mountFS, mountFSClass, "setCvsRepository", String.class, data.getCvsRepository());
        setFSParam(mountFS, mountFSClass, "setOffLine", Boolean.TYPE, new Boolean(data.isOffLine()));
        setFSParam(mountFS, mountFSClass, "setInitialCheckout", Boolean.TYPE, new Boolean(data.getPerformInitialCheckout()));
    }
    
    private static void setCmdLineFSParams(Object mountFS, Class mountFSClass, CvsWizardData data) {
        // needs to be setup rightly before
        VcsSettings settings = (VcsSettings)SharedClassObject.findObject(
                                   VcsSettings.class, true); 
        if (data.getHome() != null) {
            settings.setHome(data.getHome());
        }
        setFSParam(mountFS, mountFSClass, "setCvsModule", String.class, data.getMountPoint());
        setFSParam(mountFS, mountFSClass, "setRootDirectory", java.io.File.class, new java.io.File(data.getWorkingDir()));
        setFSParam(mountFS, mountFSClass, "setCvsServerType", String.class, data.getMethod());
        setFSParam(mountFS, mountFSClass, "setCvsServer", String.class, data.getHostName());
        setFSParam(mountFS, mountFSClass, "setCvsUserName", String.class, data.getUserName());
        setFSParam(mountFS, mountFSClass, "setCvsRoot", String.class, data.getCvsRepository());
        setFSParam(mountFS, mountFSClass, "setCvsExePath", String.class, data.getCvsExecutable());
        if (data.getShellExecutalbe() != null) {
            setFSParam(mountFS, mountFSClass, "setCygwinPath", String.class, data.getShellExecutalbe());
        }
        setFSParam(mountFS, mountFSClass, "setUseUnixShell", Boolean.TYPE, new Boolean(data.isWindows()));
        setFSParam(mountFS, mountFSClass, "setOffLine", Boolean.TYPE, new Boolean(data.isOffLine()));
        setFSParam(mountFS, mountFSClass, "setInitialCheckout", Boolean.TYPE, new Boolean(data.getPerformInitialCheckout()));
	
	String cvs_rsh = data.getRsh();
	if (cvs_rsh != null && cvs_rsh.length() > 0)
	    ((VcsFileSystem)mountFS).setEnvironmentVar ("CVS_RSH",cvs_rsh);
    }
    
    private static void setFSParam(Object mountFS, Class mountFSClass, String methodName, Class parameterType, Object value) {
        Method setter;
        try {
            setter = mountFSClass.getMethod(methodName, new Class[] { parameterType });
        } catch (final NoSuchMethodException nsmExc) {
            nsmExc.printStackTrace();
            TopManager.getDefault().notifyException(new Exception() {
                public String getLocalizedMessage() {
                    return nsmExc.getMessage();
                }
            });
            setter = null;
        } catch (final SecurityException sExc) {
            sExc.printStackTrace();
            TopManager.getDefault().notifyException(new Exception() {
                public String getLocalizedMessage() {
                    return sExc.getMessage();
                }
            });
            setter = null;
        }
        if (setter != null) {
            try {
                //System.out.println("method = "+methodName+", value = "+value);
                setter.invoke(mountFS, new Object[] { value });
            } catch (final Exception exc) {
                exc.printStackTrace();
                TopManager.getDefault().notifyException(new Exception() {
                    public String getLocalizedMessage() {
                        return exc.getMessage();
                    }
                });
            }
        }
    }

}
