/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.vcscore.wizard.mountcvs.util;

import java.util.StringTokenizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author  tzezula
 * @version 
 */
public class CvsHelper extends Object {
    
    public static final String LOCAL = "local";     // No I18N
    public static final String SERVER = "server";   // No I18N
    public static final String PSERVER = "pserver"; // No I18N
    public static final String EXTERNAL = "ext";    // No I18N
    
    public static final int T_UNKNOWN = 0;
    public static final int T_LOCAL = 1;
    public static final int T_SERVER = 2;
    public static final int T_PSERVER = 3;
    public static final int T_EXTERNAL = 4;
    
    private String cvsRoot_;
    private String serverName_;
    private String userName_;
    private String repository_;
    private String method_;
    private boolean modified_;

    private ArrayList listeners_;
    
    /** Creates new Helper */
    public CvsHelper(String cvsRoot) {
        this.cvsRoot_ = cvsRoot;
        this.modified_ = false;
        this.listeners_ = new ArrayList ();
        this.serverName_ = this.userName_ = this.repository_ = this.method_ = "";
        if (this.cvsRoot_ != null)
            this.parseCvsRoot ();
        else
            this.cvsRoot_ = "";
    }
    
    public CvsHelper () {
        this (null);
    }
    
    /** Returns CVS root
     */
    public String getCvsRoot () {
        if (modified_) 
            buildCvsRoot();
        return this.cvsRoot_;
    }
    
    /** Returns CVS connection method or empty string in case
     * of invalid string given in constructor or in case
     * of default constructor used
     */
    public String getMethod () {
        return this.method_;
    }
    
    public int getMethodType () {
        if (this.method_.equals(LOCAL))
            return T_LOCAL;
        else if (this.method_.equals (SERVER))
            return T_SERVER;
        else if (this.method_.equals (PSERVER))
            return T_PSERVER;
        else if (this.method_.equals (EXTERNAL))
            return T_EXTERNAL;
        else
            return T_UNKNOWN;
    }
    
    public String getServerName () {
        return this.serverName_;
    }
    
    public String getUserName () {
        return this.userName_;
    }
    
    public String getRepository () {
        return this.repository_;
    }
    
    public void setCvsRoot (String cvsRoot) {
        String oldCvsRoot = this.cvsRoot_;
        this.userName_ = this.serverName_ = this.method_ = this.repository_ = "";
        this.cvsRoot_ = cvsRoot;
        if (cvsRoot != null)
            parseCvsRoot ();
        else
            cvsRoot_ = "";
        this.modified_ = false;
        this.fireChange ("cvsRoot", oldCvsRoot, this.cvsRoot_);
    }
    
    public void setMethod (String method) {
        String oldMethod = this.method_;
        this.method_ = method;
        this.modified_ = true;
        this.fireChange ("method", oldMethod, this.method_);
    }
    
    public void setUserName (String userName) {
        String oldUserName = this.userName_;
        this.userName_ = userName;
        this.modified_ = true;
        this.fireChange ("userName", oldUserName, this.userName_);
    }
    
    public void setServerName (String serverName) {
        String oldServerName = this.serverName_;
        this.serverName_ = serverName;
        this.modified_ = true;
        this.fireChange ("serverName", oldServerName, this.serverName_);
    }
    
    public void setRepository (String repository) {
        String oldRepository = this.repository_;
        this.repository_ = repository;
        this.modified_ = true;
        this.fireChange ("repository", oldRepository, this.repository_);
    }
    
    public synchronized void addPropertyChangeListener (PropertyChangeListener listener) {
        this.listeners_.add (listener);
    }
    
    public synchronized void removePropertyChangeListener (PropertyChangeListener listener) {
        this.listeners_.remove (listener);
    }
    
    private void buildCvsRoot () {
        String tmp = ":"+this.method_+":";
        if (!this.method_.equals(LOCAL)) {
            if (this.serverName_.length()>0 || this.userName_.length()>0)
                tmp = tmp + this.userName_+"@"+this.serverName_;
            tmp = tmp + ":";
        }
        this.cvsRoot_ = tmp+this.repository_;
        this.modified_ = false;
    }
    
    private void parseCvsRoot () {
        this.cvsRoot_ = this.cvsRoot_.trim();
        if (this.cvsRoot_.charAt(0) != ':') {
            this.cvsRoot_ = ":"+LOCAL+":"+this.cvsRoot_;
        }
        StringTokenizer tk = new StringTokenizer (this.cvsRoot_,":",true);        
        if (tk.countTokens() < 4) {
            this.cvsRoot_ = "";
            throw new IllegalArgumentException ();
        }
        tk.nextToken(); // Skeep :
        this.method_ = tk.nextToken(); // the method
        tk.nextToken(); // Skeep :
        if (this.method_.equals(LOCAL)) {
            this.repository_ = this.restTokens(tk);;
        }
        else {
            if (tk.countTokens() < 2) {
                this.cvsRoot_ = this.method_ = "";
                throw new IllegalArgumentException ();
            }
            String tmp = tk.nextToken();
            if (!tmp.equals(":")) {
                StringTokenizer tk2 = new StringTokenizer (tmp,"@");
                if (tk2.countTokens()!=2) {
                    this.cvsRoot_ = this.method_ = "";
                    throw new IllegalArgumentException ();
                }
                this.userName_ = tk2.nextToken();
                this.serverName_ = tk2.nextToken();
                tk.nextToken (); // Skeep :
            }
            this.repository_ = this.restTokens(tk);
        }
    }
    
    private void fireChange (String name, Object oldValue, Object newValue) {
        PropertyChangeEvent event = new PropertyChangeEvent (this, name, oldValue, newValue);
        Iterator it;
        synchronized (this) {
            it = ((ArrayList)this.listeners_.clone()).iterator();
        }
        while (it.hasNext()) {
            ((PropertyChangeListener)it.next()).propertyChange (event);
        }
    }
    
    private String restTokens (StringTokenizer tk) {
        String result = "";
        while (tk.hasMoreTokens())
            result = result + tk.nextToken ();
        return result;
    }
}
