/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2001 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package vcs.commands;

import java.util.*;

import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.util.NbBundle;

import org.netbeans.modules.vcscore.Variables;
import org.netbeans.modules.vcscore.VcsFileSystem;
import org.netbeans.modules.vcscore.cmdline.VcsAdditionalCommand;
import org.netbeans.modules.vcscore.commands.*;
import org.netbeans.modules.vcscore.revision.*;
import org.netbeans.modules.vcscore.util.VcsUtilities;
import org.netbeans.modules.vcscore.util.TopComponentCloseListener;

/**
 * The Revision Explorer data parser for VSS.
 * @author  Martin Entlicher
 */
public class VssRevisionExplorerLocalized extends Object implements VcsAdditionalCommand, CommandDataOutputListener {

    private static final String revisionStr = org.openide.util.NbBundle.getBundle(VssRevisionExplorerLocalized.class).getString("HISTORY_Version");
    private static final String userStr = org.openide.util.NbBundle.getBundle(VssRevisionExplorerLocalized.class).getString("HISTORY_User");
    private static final String dateStr = org.openide.util.NbBundle.getBundle(VssRevisionExplorerLocalized.class).getString("HISTORY_Date");
    private static final String timeStr = org.openide.util.NbBundle.getBundle(VssRevisionExplorerLocalized.class).getString("HISTORY_Time");
    private static final String commentStr = org.openide.util.NbBundle.getBundle(VssRevisionExplorerLocalized.class).getString("HISTORY_Comment");

    private ArrayList revisionItems = new ArrayList();
    private ArrayList lastRevisionItems = null;
    private VcsFileSystem fileSystem = null;
    private VcsCommand logCmd = null;
    private CommandOutputListener stdoutNRListener = null;
    private CommandOutputListener stderrNRListener = null;

    /** Creates new VssRevisionExplorer */
    public VssRevisionExplorerLocalized() {
    }

    public void setFileSystem(VcsFileSystem fileSystem) {
        this.fileSystem = fileSystem;
    }
    
    /**
     * Run the vlog command and return the exit status.
     */
    private boolean runLog(Hashtable vars) {
        //System.out.println("runLog("+prepared+", "+timeout+", "+logDataRegex+")");
        revisionItems = new ArrayList();
        VcsCommandExecutor vce = fileSystem.getVcsFactory().getCommandExecutor(logCmd, vars);
        vce.addDataOutputListener(this);
        fileSystem.getCommandsPool().startExecutor(vce);
        fileSystem.getCommandsPool().waitToFinish(vce);
        if (vce.getExitStatus() != VcsCommandExecutor.SUCCEEDED) {
            //E.err("exec failed "+ec.getExitStatus()); // NOI18N
            return false;
        }
        //addTags();
        return true;
    }
    
    /**
     * Executes the vlog command to get the logging informations.
     * @param vars variables needed to run cvs commands
     * @param args the arguments,
     * @param stdoutNRListener listener of the standard output of the command
     * @param stderrNRListener listener of the error output of the command
     * @param stdoutListener listener of the standard output of the command which
     *                       satisfies regex <CODE>dataRegex</CODE>
     * @param dataRegex the regular expression for parsing the standard output
     * @param stderrListener listener of the error output of the command which
     *                       satisfies regex <CODE>errorRegex</CODE>
     * @param errorRegex the regular expression for parsing the error output
     * @return true if the command was succesfull,
     *         false if some error has occured.
     */
    public boolean exec(final Hashtable vars, String[] args,
                        CommandOutputListener stdoutNRListener, CommandOutputListener stderrNRListener,
                        CommandDataOutputListener stdoutListener, String dataRegex,
                        CommandDataOutputListener stderrListener, String errorRegex) {

        this.stdoutNRListener = stdoutNRListener;
        this.stderrNRListener = stderrNRListener;
        this.logCmd = fileSystem.getCommand(args[0]);
        boolean status = runLog(new Hashtable(vars));
        if (false == status) return status;
        //ArrayList revisionItems = createRevisionItems();
        final RevisionList list = new RevisionList();
        list.setFileSystem(fileSystem);
        final RevisionNode node = new RevisionNode(list);
        node.setName((String) vars.get("FILE"));
        node.setFileNode(true);
        String path1 = (String) vars.get("PATH");
        final String path = path1.replace(java.io.File.separatorChar, '/');
        final FileObject fo = fileSystem.findFileObject(path);
        list.setFileObject(fo);
        node.setDisplayName(fileSystem.annotateName(path));
        final FileStatusListener statusListener = new FileStatusListener() {
            public void annotationChanged(FileStatusEvent evt) {
                if (evt.hasChanged(fo)) {
                    node.setDisplayName(fileSystem.annotateName(path));
                }
            }
        };
        fileSystem.addFileStatusListener(statusListener);
        RevisionItem rootItem = new RevisionItem(null);
        node.setItem(rootItem);
        //RevisionExplorer explorer = new RevisionExplorer(node, fileSystem.annonateDOName(displayName));
        String rootDisplayName = path;
        try {
            rootDisplayName = fileSystem.annotateDOName(path);
        } catch (org.openide.loaders.DataObjectNotFoundException exc) {
            // silently ignored
        }
        java.awt.Image rootIcon;
        try {
            rootIcon = DataObject.find(fo).getNodeDelegate().getIcon(java.beans.BeanInfo.ICON_COLOR_16x16);
        } catch (DataObjectNotFoundException exc) {
            rootIcon = null;
        }
        RevisionExplorer.Panel explorer = RevisionExplorer.getRevisionExplorer(node, rootDisplayName, rootIcon);
        explorer.open();
        Object[] revisionItemsSorted = revisionItems.toArray(new RevisionItem[0]);
        Arrays.sort(revisionItemsSorted);
        revisionItems = new ArrayList(Arrays.asList(revisionItemsSorted));
        lastRevisionItems = revisionItems;
        for(Iterator it = revisionItems.iterator(); it.hasNext(); ) {
            RevisionItem item = (RevisionItem) it.next();
            //System.out.println("CvsRevisionViewer:list.add("+item.getRevision()+")");
            list.add(item);
        }
        final RevisionListener revisionListener = new RevisionListener() {
            public void revisionsChanged(int whatChanged, FileObject chfo, Object info) {
                //System.out.println("revisionsChanged("+whatChanged+"), fo = "+fo+", info = "+info);
                if (fo == chfo) {
                    runLog(new Hashtable(vars));
                    if ((whatChanged & RevisionListener.NUM_REVISIONS_CHANGED) != 0) {
                        //adjustItems(list);
                        lastRevisionItems = revisionItems;
                    }
                    if ((whatChanged & RevisionListener.ONE_REVISION_CHANGED) != 0) {
                        //adjustItem(list, (String) info);
                    }
                }
            }
        };
        fileSystem.addRevisionListener(revisionListener);
        explorer.addCloseListener(new TopComponentCloseListener() {
            public void closing() {
                fileSystem.removeRevisionListener(revisionListener);
                fileSystem.removeFileStatusListener(statusListener);
            }
        });
        return true;
    }
    
    /**
     * Get the last revision item obtained.
     */
    private RevisionItem getLastRevisionItem() {
        RevisionItem item = null;
        for(int i = revisionItems.size() - 1; i >= 0; i--) {
            RevisionItem it = (RevisionItem) revisionItems.get(i);
            if (!it.isBranch()) {
                //System.out.println("getLastRevision: "+it.getRevision());
                item = it;
                break;
            }
        }
        return item;
    }
    
    public void outputData(String[] elements) {
        if (elements == null || elements.length == 0 || elements[0] == null) return;
        if (elements[0].indexOf(revisionStr) == 0) {
            int endRev = elements[0].indexOf('*', revisionStr.length());
            if (endRev < 0) endRev = elements[0].length();
            String revision = elements[0].substring(revisionStr.length(), endRev).trim();
            String revisionVCS = revision;
            revision += ".0";
            RevisionItem item = new RevisionItem(revision);
            item.setRevisionVCS(revisionVCS);
            revisionItems.add(item);
        } else if (elements[0].indexOf(userStr) == 0) {
            if (revisionItems.size() > 0) {
                RevisionItem item = getLastRevisionItem();
                int index1 = userStr.length();
                int index2 = elements[0].indexOf(dateStr, index1);
                if (index2 < index1) return;
                String user = elements[0].substring(index1, index2).trim();
                item.setAuthor(user);
                index1 = index2 + dateStr.length(); // date
                index2 = elements[0].indexOf(timeStr, index1);
                if (index2 < index1) return;
                int len = elements[0].length();
                if (index2 < index1) index2 = len;
                item.setDate(elements[0].substring(index1, index2).trim());
                index1 = index2 + timeStr.length(); // date
                item.setDate(elements[0].substring(index1).trim());
            }
        } else if (elements[0].indexOf(commentStr) == 0) {
            RevisionItem item = getLastRevisionItem();
            String msg = elements[0].substring(commentStr.length());
            item.setMessage(msg);
        } else if (revisionItems.size() > 0) {
            RevisionItem item = getLastRevisionItem();
            String msg = item.getMessage();
            if (msg == null) return ;
            msg += "\n" + elements[0].substring(commentStr.length());
            item.setMessage(msg);
        }
    }
}
