/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.execution;

import java.beans.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;

import org.openide.util.NbBundle;
import org.openide.loaders.DataObject;
import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.util.HttpServer;
import org.openide.TopManager;

import org.netbeans.modules.web.core.WebExecInfo;
import org.netbeans.modules.web.core.Util;
import org.netbeans.modules.web.core.WebExecUtil;
import org.netbeans.modules.web.core.WebExecSupport;
import org.netbeans.modules.web.core.jsploader.JspCompileUtil;
import org.netbeans.modules.web.core.jsploader.JspDataObject;

/**
 *
 * @author  Petr Jiricka
 * @version 
 */
public class DefaultExecParams implements ExecParams {

    /** serialVersionUID for serialization compatibility */
    static final long serialVersionUID=8566973176961655058l;
    
    public static final String PROP_PORT = "port";
    public static final String PROP_CONTEXT_URI = "contextURI";

    /** Set of contexts describing the currently executing server instance. */
    protected static Set serverInstanceContexts;
    
    protected PropertyChangeSupport propertySupport;
    
    /** server.xml config file template */
    private static String defaultJakartaConfig;
    /** default web.xml file */
    private static String defaultDD;

    /** Port of the server. */
    private int port = 8080;
    
    /** Mapping of the web application. */
    private String contextURI = "";
    
    /** Creates new ExecParams */
    public DefaultExecParams() {
        propertySupport = new PropertyChangeSupport ( this );
    }

    /** Getter for property port.
     * @return Value of property port.
     */
    public int getPort() {
        return port;
    }
    
    /** Setter for property port.
     * @param port New value of property port.
     */
    public void setPort(int port) {
        int oldPort = this.port;
        this.port = port;
        propertySupport.firePropertyChange (PROP_PORT, new Integer (oldPort), new Integer (port));
    }
    
    /** Returns the context URI (mapping). */
    public String getContextURI() {
        return contextURI;
    }
    
    /** Sets the context URI (mapping). */
    public void setContextURI(String contextURI) {
        String oldContextURI = this.contextURI;
        this.contextURI = canonizeContextURI(contextURI);
        propertySupport.firePropertyChange (PROP_CONTEXT_URI, oldContextURI, this.contextURI);
    }
    
    /** Converts a context URI to an appropriate form. */
    protected final String canonizeContextURI(String uri) {
        while (!uri.startsWith("/"))
            uri = "/" + uri;
        if (uri.endsWith("/"))
            uri = uri.substring(0, uri.length() - 1);
        return uri;
    }

    
    public void addPropertyChangeListener (PropertyChangeListener listener) {
        propertySupport.addPropertyChangeListener (listener);
    }

    public void removePropertyChangeListener (PropertyChangeListener listener) {
        propertySupport.removePropertyChangeListener (listener);
    }
    
    // --------- implementation of the ExecParams interface ------------------

    /** Returns the MIME type of the content language for object
     * represented by <code>executedObject</code>.
     * This is used to determine which browser (WWW, WAP, ...) to display the URL in.
     */
    
    public String getContentLanguage(WebExecInfo executedObject) throws IOException {
        DataObject obj = executedObject.getDataObject();
        if (obj == null) {
            return executedObject.getContentLanguage();
        }    
        if (obj instanceof JspDataObject) {
            return ((JspDataObject)obj).getContentLanguage();
        }
        if (isHtmlObject(executedObject)) {
            return "text/html";
        }
        return null;
    }
    
    /** Should determine whether the parameters have changed sufficiently since the
     * last time so the server should be restarted.
     * @param executedObject object which is now being executed
     * @param lastExecParams execution parameters used last time to start the server.
     *    may be null if the server was not executed yet
     * @return true if the server should be restarted
     */
    public boolean serverRestartRequired(WebExecInfo executedObject, ExecParams lastExecParams) {
        if (executedObject.getDataObject() == null) {
            return false;
        }
        
        // resolve the new contexts for this object
        Set newContexts = getContextsToExecute(executedObject);
        
        if (lastExecParams == null) {
            return true;
        }
        
        if (!compareContexts(newContexts, serverInstanceContexts)) {
            return true;
        }
        
        if (getClass() != lastExecParams.getClass()) {
            return true;
        }
        
        return false;
    }
    
    /** Compares two sets of contexts, returns true if they are the same. */
    protected boolean compareContexts(Set ctx1, Set ctx2) {
        if (ctx1.size() != ctx2.size())
            return false;
        Iterator it = ctx1.iterator();
        for (; it.hasNext(); ) {
            if (!ctx2.contains(it.next()))
                return false;
        }
        return true;
    }
    
    protected WebContext getContextForObject(WebExecInfo webExecInfo) {
        DataObject dobj = webExecInfo.getDataObject();
        if (dobj == null) {
            throw new IllegalArgumentException();
        }
        try {
            FileObject docRoot = JspCompileUtil.getContextRoot(dobj.getPrimaryFile());
            return new WebContext(docRoot, getContextURI());
        }
        catch (FileStateInvalidException e) {
            throw (IllegalArgumentException)TopManager.getDefault().getErrorManager().
                annotate(new IllegalArgumentException(), e);
        }
    }
    
    protected Set getContextsToExecute(WebExecInfo webExecInfo) {
        HashSet set = new HashSet();
        WebContext wc = getContextForObject(webExecInfo);
        set.add(wc);
        if (!wc.getPath().equals("")){
            wc=getContextForObject(webExecInfo);
            wc.setPath("");
            set.add(wc);
        }
        return set;
    }
    
    protected String getRequestInterceptorsCode() {
        return "";
    }
    
    protected String getContextsCode(Set contexts) {
        StringBuffer sb = new StringBuffer();
        Iterator it = contexts.iterator();
        for (; it.hasNext();) {
            WebContext ctx = (WebContext)it.next();
            String contextDocRoot = NbClassPath.toFile(ctx.getDocumentRoot()).getAbsolutePath();
            sb.append(getSingleContextCode(ctx.getPath(), contextDocRoot));
        }
        return sb.toString();
    }
    
    /** Creates and deploys the Tomcat configuration file (server.xml)
     * @return the configuration file
     */
    public File deployTomcatConfig(WebExecInfo executedObject) throws IOException {
        Set contexts = getContextsToExecute(executedObject);
        String contextsCode = getContextsCode(contexts);
        File configFileName = deployJakartaConfig0(getPort(), contextsCode, 
            getRequestInterceptorsCode());
        
        // internal server port number
        tellInternalServerPortNumber();
        
        // file containing information about temporary directories for contexts
        FileObject[] docRoots = new FileObject[contexts.size()];
        Iterator it = contexts.iterator();
        for (int i = 0; it.hasNext(); i++) {
            docRoots[i] = ((WebContext)it.next()).getDocumentRoot();
        }
        deployTempDirFile(docRoots);
        
        // deploy the default deployment descriptor
        deployDefaultDD();
        
        // set the new server instance
        serverInstanceContexts = contexts;
        
        return configFileName;
    }
    
    protected static void deployTempDirFile(FileObject contextRoots[]) throws IOException {
        String dirName = WebExecUtil.getTempDirectory() + File.separator + "tomcat";
        File ff = new File(dirName);
        WebExecUtil.myMkdirs(ff);
        ff = new File(ff.getAbsolutePath() + File.separator + "tempDirs.properties");
        OutputStream propStream = new FileOutputStream(ff);
        java.util.Properties props = new java.util.Properties();
        for (int i = 0; i < contextRoots.length; i++) {
            props.put(NbClassPath.toFile(contextRoots[i]).getCanonicalPath(), 
                      NbClassPath.toFile(JspCompileUtil.getContextOutputRoot(
                          contextRoots[i])).getAbsolutePath());
        }
        props.save(propStream, null);
        propStream.close();
    }
    
    private static void initDefaultJakartaConfig() throws IOException {
        if (defaultJakartaConfig == null) {
            // read the config from resource first
            StringBuffer sb = new StringBuffer();
            String lineSep = System.getProperty("line.separator");
            BufferedReader br = new BufferedReader(new InputStreamReader(
                DefaultExecParams.class.getClassLoader().getResourceAsStream(
                "org/netbeans/modules/web/core/resources/defaultserver.xml")));
            String line = br.readLine();
            while (line != null) {
                sb.append(line);
                sb.append(lineSep);
                line = br.readLine();
            }
            br.close();
            defaultJakartaConfig = sb.toString();
        }
    }
    
    /** Writes server.xml in the temp/tomcat/server.xml file. Formats the default template
    * from resource 'org/netbeans/modules/web/core/execution/defaultserver.xml' by MessageFormat
    * and replaces parameters by the following values:
    * The params are:
    * <ul>
    * <li>{0}  port</li>
    * <li>{1}  code to initialize contexts</li>
    * <li>{2}  serverTempDir</li>
    * <li>{3}  code to initialize additional request interceptors</li>
    * </ul>
    * @return name of the file created
    */
    protected static File deployJakartaConfig0(int port, String contextsCode, String interceptorsCode) throws IOException {
        initDefaultJakartaConfig();
        
        String result = MessageFormat.format(
            defaultJakartaConfig, new Object[]
            {"" + port, contextsCode, 
             WebExecUtil.getTomcatTemp().getAbsolutePath(), interceptorsCode});
        String dirName = WebExecUtil.getTempDirectory() + File.separator + "tomcat";
        File ff = new File(dirName);
        WebExecUtil.myMkdirs(ff);
        ff = new File(ff.getAbsolutePath() + File.separator + "server.xml");
        BufferedWriter bw = new BufferedWriter(new FileWriter(ff));
        bw.write(result);
        bw.close();
        return ff;
    }
    
    private static void initDefaultDD() throws IOException {
        if (defaultDD == null) {
            // read the config from resource first
            StringBuffer sb = new StringBuffer();
            String lineSep = System.getProperty("line.separator");
            BufferedReader br = new BufferedReader(new InputStreamReader(
                DefaultExecParams.class.getClassLoader().getResourceAsStream(
                "org/netbeans/modules/web/core/resources/defaultweb.xml")));
            String line = br.readLine();
            while (line != null) {
                sb.append(line);
                sb.append(lineSep);
                line = br.readLine();
            }
            br.close();
            defaultDD = sb.toString();
        }
    }
        
    public static File deployDefaultDD() throws IOException {
        initDefaultDD();
        String dirName = WebExecUtil.getTempDirectory() + File.separator + "tomcat" + File.separator + "conf";
        File ff = new File(dirName);
        WebExecUtil.myMkdirs(ff);
        ff = new File(ff.getAbsolutePath() + File.separator + "web.xml");
        BufferedWriter bw = new BufferedWriter(new FileWriter(ff));
        bw.write(defaultDD);
        bw.close();
        return ff;
    }

    protected static File tellInternalServerPortNumber() throws IOException {

        File ff = WebExecUtil.getTomcatTemp();
        ff = new File(ff.getAbsolutePath() + File.separator + "port");

        // read the config from resource first
        URL internalServerRoot = HttpServer.getRepositoryRoot();
        String port = (internalServerRoot == null) ? "" : "" + internalServerRoot.getPort();

        // write the port file
        // pending - somehow listen on changes of the port number
        BufferedWriter bw = new BufferedWriter(new FileWriter(ff));
        bw.write(port);
        bw.close();
        return ff;
    }

/*    protected String getAdditionalRequestInterceptors() {
        return "";
    }*/
    
    /** Return code for server.xml to initialize contexts. */
/*    protected String getContextsCode() {
        String contextDocRoot = NbClassPath.toFile(serverInstanceDocumentRoot).getAbsolutePath();
        return getSingleContextCode(getContextURI(), contextDocRoot);
    }*/
    
    protected final String getSingleContextCode(String path, String docBase) {
        StringBuffer code = new StringBuffer();
        String lineSep = System.getProperty("line.separator");
        code.append("        <Context path=\"");
        code.append(path);
        code.append("\" docBase=\"");
        code.append(docBase);
        code.append("\" debug=\"0\" reloadable=\"true\" > ");
        code.append(lineSep);
        code.append("        </Context>");
        code.append(lineSep);
        return code.toString();
    }

    /** Returns a URL which should be displayed in the browser
     * when executing object represented by <code>executedObject</code>.
     */
    public URL getURL(WebExecInfo executedObject) throws MalformedURLException, IOException {
        DataObject obj = executedObject.getDataObject();
        if (obj == null) {
            return executedObject.getURL();
        }    
        if (obj instanceof JspDataObject) {
            return constructJspURL(obj, getContextURI(), getPort());
        }
        if (isHtmlObject(executedObject)) {
            return constructHtmlURL(obj, getContextURI());
        }
        return null;
    }
    
    protected boolean isHtmlObject(WebExecInfo executedObject) {
        DataObject obj = executedObject.getDataObject();
        try {
            Class htmlDO = Class.forName("org.netbeans.modules.html.HtmlDataObject");
            return (htmlDO.isInstance(obj));
        }
        catch (ClassNotFoundException e) {
            return false;
        }
    }
    
    /** Returns URL scheme for execution of a given object. */
    protected String getScheme(DataObject dobj) {
        // PENDING - there should be a way for the user to set this, 
        // probably on a per-object basis
        return "http";
    }
    
    public static URL constructJspURL(DataObject jspDo, String contextBase, int port) throws MalformedURLException, IOException {
        String queryString = WebExecSupport.getQueryString(jspDo.getPrimaryFile());

        if ((queryString.length() > 0) && (!queryString.startsWith("?"))) // NOI18N
            queryString = "?" + queryString; // NOI18N

        URL url = new URL("http", Util.getUtil().getLocalHost(), port, contextBase + "/" + // NOI18N
                          jspDo.getPrimaryFile().getPackageNameExt('/', '.') + queryString);
        return url;
    }

    protected URL constructHtmlURL(DataObject dobj, String contextBase) throws MalformedURLException, IOException {
        URL url = new URL(getScheme(dobj), Util.getUtil().getLocalHost(), getPort(), 
            contextBase + "/" + dobj.getPrimaryFile().getPackageNameExt('/', '.')); // NOI18N
        return url;
    }
    
    /** Display name for the Tomcat process.
     */
    public String getProcessDisplayName() {
        return java.text.MessageFormat.format(NbBundle.getBundle(DefaultExecParams.class).
            getString("CTL_Servlet_Process"), new Object[] {new Integer(getPort())});
    }
    
    /** Directory containing the Tomcat home directory.
     */
    public File getTomcatHome() throws IOException {
        return WebExecUtil.getTomcatHome();
    }
    
    public Object clone() throws CloneNotSupportedException {
        DefaultExecParams newParams = new DefaultExecParams();
        newParams.setPort(getPort());
        newParams.setContextURI(getContextURI());
        return newParams;
    }
    
    public void runObject(WebExecInfo executedObject) {
    }
    
    public String executorDisplayName() {
        return NbBundle.getBundle(JakartaExecutor.class).getString("CTL_Exec_Name");
    }
    
    public String debuggerTypeDisplayName() {
        return "-";
    }
    
    /** A class describing context running in the server. */
    protected static class WebContext {
        
        protected FileObject documentRoot;
        protected String path;
        
        public WebContext(FileObject documentRoot, String path) {
            this.documentRoot = documentRoot;
            this.path = path;
        }
        
        public FileObject getDocumentRoot() {
            return documentRoot;
        }
        
        public String getPath() {
            return path;
        }
        
        public void setPath(String path) {
            this.path=path;
        }
        
        public boolean equals(Object obj) {
            WebContext ctx = (WebContext)obj;
            return (documentRoot.equals(ctx.documentRoot) &&
                    path.equals(ctx.path));
        }
        
        public int hashCode() {
            return documentRoot.hashCode() + path.hashCode();
        }
    }
    
}