/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.execution;

import java.beans.*;
import java.lang.reflect.*;
import java.io.IOException;
import java.util.Enumeration;
import java.net.URL;
import java.net.MalformedURLException;

import org.openide.TopManager;
import org.openide.filesystems.*;
import org.openide.execution.NbProcessDescriptor;
import org.openide.util.HelpCtx;
import org.openide.util.MapFormat;
import org.openide.util.NbBundle;
import org.openide.util.TaskListener;
import org.openide.util.Task;
import org.openide.loaders.DataObject;
import org.openide.execution.Executor;
import org.openide.execution.ProcessExecutor;
import org.openide.execution.NbClassPath;
import org.openide.execution.ExecInfo;
import org.openide.execution.ExecutorTask;
import org.openide.util.WeakListener;

import org.netbeans.modules.web.core.FeatureFactory;
import org.netbeans.modules.web.core.WebExecUtil;
import org.netbeans.modules.web.core.WebExecInfo;
import org.netbeans.modules.web.core.jsploader.JspCompileUtil;

/** Executor for objects (JSPs, web apps, servlets, HTML pages, ...) 
 * executed using the Jakarta server.
 *
 * @author  Petr Jiricka
 * @version 
 */
public class JakartaExecutor extends ProcessExecutor {
    
    /** serialVersionUID for serialization compatibility */
    static final long serialVersionUID=6533546813743662354l;
    
    /** Needs some more design work : serves to distinguish between "normal" run action
    *   and 'restart server and run' action. */
    private static boolean restartServerRunAction = false;

    /** The currently executing server. */
    private static ExecutorTask serverInstance;
    
    /** The currently executing server process. */
    private static Process serverProcess;
    
    /** Executor used to execute the server last time. */
    private static JakartaExecutor serverInstanceExecutor;
    
    /** Execution paramaters used to execute the server last time. */
    private static ExecParams serverInstanceExecParams;
    
    /** Listener which listens for property changes on the executor which 
     * was last used to execute the server. Server restart is forced 
     * if any property changes. 
     */
    private static PropertyChangeListener pcl;
    

    /** 
     * The interval during which we wait for the server to come up in
     * order to launch the browser.
     */
    private static final int timeout = 30000;
    
    // fields related to ExecParams and dynamic changes of ExecParams
    public static final String PROP_EXEC_PARAMS = "execParams";
    private ExecParams execParams;
    private static BeanInfo execParamsBeanInfo = null;
    protected transient PropertyChangeListener paramsListener;
    
    /** Setting for pointbase database */
    private String getPointbaseSetting(){
        String pointbaseIni=System.getProperty("pointbase.ini");
        if (pointbaseIni==null) return "";
        else return "-Dpointbase.ini="+pointbaseIni+" ";
    }
           
    /** Creates new JakartaExecutor */
    public JakartaExecutor() {
        super();      
        initialize();        
    }
    
    /* Default human-presentable name of the executor.
    * In the default implementation, just the class name.
    * @return initial value of the human-presentable name
    */
    public String displayName() {
        return execParams.executorDisplayName();
    }

    public HelpCtx getHelpCtx () {
        return new HelpCtx (JakartaExecutor.class);
    }

    private void initialize() {
        paramsListener = new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                myFirePropertyChange(evt.getPropertyName(), evt.getOldValue(), evt.getNewValue());
            }
        };
        if (execParams != null)
            execParams.addPropertyChangeListener(paramsListener);
        checkParams();
    }
    
    public void checkParams() {
//System.out.println("checking params for " + this);        
        ExecParams fromFactory = FeatureFactory.getFactory().createExecParams(this);
        if (execParams == null) {
            setExecParams(fromFactory);
        }
        else {
            if (execParams.getClass() != fromFactory.getClass()) {
                setExecParams(fromFactory);
            }
        }
    }
    
    private void readObject(java.io.ObjectInputStream in) 
        throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        initialize();       
    }

     public ExecParams getExecParams() {
        return execParams;
    }
    
    public void setExecParams(ExecParams execParams) {
//System.out.println("setting params " + execParams);
        ExecParams oldExecParams = this.execParams;
        if (oldExecParams != null)
            oldExecParams.removePropertyChangeListener(paramsListener);
        this.execParams = execParams;
        execParams.addPropertyChangeListener(paramsListener);
        JakartaExecutorBeanInfo.prop = null;
        execParamsBeanInfo = null;
        Introspector.flushFromCaches(getClass());
        firePropertyChange (PROP_EXEC_PARAMS, oldExecParams, execParams);
    }
    
    void myFirePropertyChange(String name, Object o, Object n) {
        super.firePropertyChange(name, o, n);
    }
    
    private Object callGetter(String propertyName) {
        try {
            if (execParamsBeanInfo == null)
                execParamsBeanInfo = Introspector.getBeanInfo(execParams.getClass());
            PropertyDescriptor[] prop = execParamsBeanInfo.getPropertyDescriptors();
            for (int i = 0; i < prop.length; i++) {
                if (prop[i].getName().equals(propertyName)) {
                    Method getter = prop[i].getReadMethod();
                    return getter.invoke(execParams, new Object[0]);
                }
            }
        }
        catch (IntrospectionException e) {}
        catch (IllegalAccessException e) {}
        catch (InvocationTargetException e) {}
        return null;
    }

    private void callSetter(String propertyName, Object value) {
        try {
            if (execParamsBeanInfo == null)
                execParamsBeanInfo = Introspector.getBeanInfo(execParams.getClass());
            PropertyDescriptor[] prop = execParamsBeanInfo.getPropertyDescriptors();
            for (int i = 0; i < prop.length; i++) {
                if (prop[i].getName().equals(propertyName)) {
                    Method setter = prop[i].getWriteMethod();
                    setter.invoke(execParams, new Object[] {value});
                    //firePropertyChange(propertyName, null, value);
                }
            }
        }
        catch (IntrospectionException e) {}
        catch (IllegalAccessException e) {}
        catch (InvocationTargetException e) {}
    }
    
    public static void changeExecParams() {
        Enumeration executors = Executor.executors();
        for (; executors.hasMoreElements(); ) {
            Object exec = executors.nextElement();
            if (exec instanceof JakartaExecutor) {
                ((JakartaExecutor)exec).checkParams();
            }
        }
    }
    
    // methods actually related to execution
    public static void forceRestart() {
        //System.out.println("forcing restart");
        restartServerRunAction = true;
    }

    static Process getServerProcess() {
        return serverProcess;
    }
    
    public static void killServerIfRunning() {
        ExecutorTask tempServer = serverInstance;
        if (serverInstance != null) {
            serverInstance.stop();
            // now serverInstance is null, restarting -> have to close the output of the previous instance
            tempServer.getInputOutput().closeInputOutput();
            //cleanServerInstance();
        }
    }

    static void cleanServerInstance(Process caller) {
        serverProcess = null;
        serverInstance = null;
        serverInstanceExecParams = null;
        if ((serverInstanceExecutor != null) && (pcl != null))
            serverInstanceExecutor.removePropertyChangeListener(pcl);
        serverInstanceExecutor = null;
        pcl = null;
    }

    /** Called to create the java.lang.Process for given exec info.
    * Current implementation scans creates new Format with provided
    * exec info and asks the current executor to start with that
    * format.
    * <P>
    * Subclasses can override this to achive the right behaviour, add
    * system properties, own format, etc.
    * 
    * @param info exec info 
    * @return the executed process
    * @exception IOException if the action fails
    */
    protected Process createProcess (ExecInfo info) throws IOException {
        JakartaExecInfo info2 = (JakartaExecInfo)info;
        String pointbaseSetting=getPointbaseSetting();
        
        Process p = getExternalExecutor ().exec (            
                    new JakartaFormat(new ExecInfo(WebAppMain.class.getName(),
                                           new String[] {info2.getConfigFileName(), info2.getTomcatHomeDir()}),pointbaseSetting)
               );
        serverProcess = new ServerProcess(this, p, info2.getConfigFileName(), info2.getTomcatHomeDir());
        return serverProcess;
    }
    
    
    /* Executes given class by creating new process in underlying operating system.
    * @param obj DataObject to execute
    */
    public ExecutorTask execute(DataObject obj) throws IOException {
        return execute(new WebExecInfo(obj));
    }
    /* Executes given class by creating new process in underlying operating system.
    * @param obj DataObject to execute
    */
    public ExecutorTask execute(WebExecInfo webExecInfo) throws IOException {
        try {
            // force restart if the executor has changed
            if (serverInstanceExecutor != this)
                forceRestart();

            // force restart if the execution parameters say so
            if (execParams.serverRestartRequired(webExecInfo, serverInstanceExecParams))
                forceRestart();

            /*  PENDING
            // force restart if this is a servlet
            if (EditServletParamsAction.isServlet(info))
              forceRestart();
            */

            // stop the previously executed instance
            if (restartServerRunAction) {
                killServerIfRunning();
            }

            // construct the URL for the web browser
            final URL url = execParams.getURL(webExecInfo);
            if (url == null) {
                throw new IOException(NbBundle.getBundle(JakartaExecutor.class).
                    getString("EXC_ExecObjectNotSupported"));
            }
            
            final String contentLanguage = execParams.getContentLanguage(webExecInfo);
            final WebExecInfo webExecInfo2 = webExecInfo;

            // start the web browser
            Thread browser = new Thread(
                new Runnable() {
                    public void run() {

                        if (!WebExecUtil.waitAndShowInBrowser(url, contentLanguage, timeout)) {
                            // consider server not running
                            //cleanServerInstance();
                        }
                        else {
                            execParams.runObject(webExecInfo2);
                        }
                    }
                }
            );
            browser.start();

            //System.out.println("(re)starting server " + (serverInstance == null));
            if (serverInstance == null) {
                // execute the process
                serverInstance = super.execute(
                    new JakartaExecInfo(
                        execParams.getProcessDisplayName(),
                        execParams.deployTomcatConfig(webExecInfo).getAbsolutePath(),
                        execParams.getTomcatHome().getAbsolutePath()
                    ));

                serverInstanceExecParams = (ExecParams)execParams.clone();
                serverInstanceExecutor = this;
                pcl = new PropertyChangeListener() {
                          public void propertyChange(PropertyChangeEvent evt) {
                              forceRestart();
                          }
                      };
                serverInstanceExecutor.addPropertyChangeListener(pcl);

                serverInstance.addTaskListener(new TaskListener() {
                                                   public void taskFinished(Task task){
                                                       /*if (task == serverInstance)
                                                           cleanServerInstance();*/
                                                   }
                                               });
            }
            return serverInstance;
        }
        catch (IOException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions"))
                e.printStackTrace();
            throw e;
        }
        catch (Throwable e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions"))
                e.printStackTrace();
            throw (IOException)TopManager.getDefault ().getErrorManager ().copyAnnotation (
                new IOException(e.getMessage()), e);
        }
        finally {
            restartServerRunAction = false;
        }
    }


    // much as I hate to do this, there have to be getters and setters duplicating
    // all getters and setters in all known ExecParams implementations.
    // The JavaBeans concept does not allow for a cleaner solution

    // properties from DefaultExecParams
    
    public int getPort() {
        return ((Integer)callGetter("port")).intValue();
    }
    
    public void setPort(int port) {
        callSetter("port", new Integer(port));
    }
    
    public String getContextURI() {
        return (String)callGetter("contextURI");
    }
    
    public void setContextURI(String contextURI) {
        callSetter("contextURI", contextURI);
    }
    
    // properties from IEExecParams
    public boolean isRunMonitor() {
        return ((Boolean)callGetter("runMonitor")).booleanValue();
    }

    public void setRunMonitor(boolean runMonitor) {
        callSetter("runMonitor", new Boolean(runMonitor));
    }
    
    public Object getContextExecutionType() {
        return callGetter("contextExecutionType");
    }

    public void setContextExecutionType(Object contextExecutionType) {
        callSetter("contextExecutionType", contextExecutionType);
    }
    
    /** Get the current external execution command.
     * The default Java launcher associated with this VM's installation will be used,
     * and the user repository entries will be used for the class path if customized command is not set.
     * @return the settings for the current external executor
     */
    public NbProcessDescriptor getExternalExecutor () {
        if (externalExecutor == null) {
            synchronized (JakartaExecutor.class) {
                if (externalExecutor == null) 
                    externalExecutor = new NbProcessDescriptor (
                        // {java.home}{/}bin{/}java
                        "{" + Format.TAG_JAVAHOME + "}{" + Format.TAG_SEPARATOR + "}bin{" + Format.TAG_SEPARATOR + // NOI18N
                        "}java",
                        // {database} -cp {filesystems}{:}{classpath}{:}{library} {classname} {arguments}
                        "{"+JakartaFormat.TAG_POINTBASE+"} "+
                        "-cp {"+Format.TAG_REPOSITORY+"}"+
                        "{"+Format.TAG_PATHSEPARATOR+"}"+
                        "{"+Format.TAG_CLASSPATH+"}"+
                        "{"+Format.TAG_PATHSEPARATOR+"}"+
                        "{"+Format.TAG_LIBRARY+"} "+
                        "{"+Format.TAG_CLASSNAME+"} "+
                        "{"+Format.TAG_ARGUMENTS+"}",
                        NbBundle.getBundle(JakartaExecutor.class).getString("MSG_JakartaExecutor")
                    );
            }
        }
        return externalExecutor;
    }
    
    /** ExecInfo class for running the Jakarta Tomcat server.
     * Provides information about Tomcat config file (server.xml) location
     * and Tomcat home directory. */
    public static class JakartaExecInfo extends ExecInfo {

        private String configFileName, tomcatHomeDir;

        public JakartaExecInfo(String className, String configFileName, String tomcatHomeDir) {
            super(className);
            this.configFileName = configFileName;
            this.tomcatHomeDir = tomcatHomeDir;
        }

        public String getConfigFileName() {
            return configFileName;
        }

        public String getTomcatHomeDir() {
            return tomcatHomeDir;
        }
    }
    
    /** Default format that can format tags related to execution. Currently this is only the URL.
    */
    public static class JakartaFormat extends Format {
        /** Tag replaced with the URL */
        public static final String TAG_POINTBASE = "database"; // NOI18N

        //static final long serialVersionUID =4315554797414856261L;
        /** @param info exec info about class to execute
        * @param classPath to substitute instead of CLASSPATH
        * @param bootClassPath boot class path
        * @param repository repository path
        * @param library library path
        */
        public JakartaFormat (ExecInfo info, String pointbaseSetting) {
            super(info);
            java.util.Map map = getMap ();
            map.put (TAG_POINTBASE, pointbaseSetting);
        }

    }
    
}


