/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.execution;

import javax.servlet.Servlet;
import javax.servlet.ServletContext;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.SingleThreadModel;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.jsp.HttpJspPage;
import javax.servlet.jsp.JspFactory;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import java.lang.reflect.*;
import java.net.URL;
import java.net.URLEncoder;
import java.net.MalformedURLException;

import org.apache.jasper.JasperException;
import org.apache.jasper.Constants;
import org.apache.jasper.Options;
import org.apache.jasper.EmbededServletOptions;
import org.apache.jasper.JspCompilationContext;
import org.apache.jasper.JspEngineContext;

import org.apache.jasper.compiler.Mangler;
import org.apache.jasper.compiler.Compiler;
import org.apache.jasper.compiler.JspCompiler;

import org.apache.jasper.runtime.JspLoader;

import org.netbeans.modules.web.core.Util;

/**
 * JSP Servlet which can reload classes precompiled by an external tool
 *
 * @author Petr Jiricka
 */
public class JspServlet extends org.apache.jasper.servlet.JspServlet {
    
    private static final String COMPILE_SERVLET_NAME = "org.netbeans.modules.web.core.jsploader.CompileServlet";
    public static final String COMPILE_SERVLET_PARAM_NAME = "jsp";
    public static final String COMPILE_SERVLET_JSP_URI    = "jsp_uri";
    public static final String COMPILE_SERVLET_OK_RESPONSE = "OK";

    Hashtable loadedJSPTimes = new Hashtable();

    public void init(ServletConfig config)
	throws ServletException {
	super.init(config);
    }

    /** Our service methods delegates to super, only checks if compilation succeeded. If it didn't, 
    * outputs an error page. */
    public void service (HttpServletRequest request, 
    			 HttpServletResponse response)
	throws ServletException, IOException {
        try {
            super.service(request, response);
        }
        catch (CompilationFailedException e) {
            outputCompilationFailedPage(request, response);
        }
    }
    
    /*  Check if we need to reload a JSP page.
     *
     *  Side-effect: re-compile the JSP page.
     *
     *  @param classpath explicitly set the JSP compilation path.
     *  @return true if JSP files is newer
     */
    public boolean doLoadJSP(String jspUri, String classpath, 
	boolean isErrorPage, HttpServletRequest req, HttpServletResponse res) 
	throws JasperException, FileNotFoundException {
            
        getServletClassInWrapper(jspUri);
            
	Long jspClassTime = (Long) loadedJSPTimes.get(jspUri);
        
        // find out the last modified time of the class file
        JspCompilationContext ctxt = new JspEngineContext(loader, classpath,
                                                     context, jspUri, 
                                                     isErrorPage, options,
                                                     req, res);
	boolean outDated = false; 
        JspCompiler compiler = (JspCompiler)ctxt.createCompiler();
        outDated = compiler.isOutDated();
        
        String classFileName = compiler.getClassFileName();

        // hack - taken from JspCompiler.compile()
        Mangler mangler = (Mangler)compiler;
        String pkgName = mangler.getPackageName();
        ctxt.setServletPackageName(pkgName);
        String javaFileName = mangler.getJavaFileName();
        ctxt.setServletJavaFileName(javaFileName);
	String className = mangler.getClassName();
        ctxt.setServletClassName(className);
        // end hack

        File classFile = new File(classFileName);
        
        // find out the file name 
        File jsp = new File(ctxt.getJspFile());
        String fileName = ctxt.getRealPath(jsp.getPath());
        if (fileName == null)
            throw new FileNotFoundException(jsp.toString());
        
        File jspReal = new File(fileName);
        
        // if reload not needed, just return
        if (classFile.exists()) {
            if ((jspClassTime != null) && 
                (jspClassTime.longValue() >= jspReal.lastModified()) &&
                (jspClassTime.longValue() >= classFile.lastModified())
            ) {
//System.out.println("jspClassTime  " + jspClassTime.longValue());
//System.out.println("jspFileTime   " + jspReal.lastModified());
//System.out.println("classFileTime " + classFile.lastModified());
//System.out.println("reload not necessary");                
                return false;
            }
        }
        
        // now check whether the page needs to be compiled or reloaded
        // the need for compilation implies the need for reload
        boolean needsCompile = false;
        boolean needsReload  = false;
        // class file does not exist -> recompile
        if (!classFile.exists()) 
            needsCompile = true;
        // already loaded, but JSP modified since -> reload
        if ((jspClassTime != null) && 
            (jspClassTime.longValue() < jspReal.lastModified()))
            needsReload = true;
        // already loaded, but class modified since -> reload
        if ((jspClassTime != null) && 
            (classFile.exists()) && 
            (jspClassTime.longValue() < classFile.lastModified()))
            needsReload = true;
        // loaded or not, but JSP modified after the class was modified
        if ((classFile.exists()) &&
            (classFile.lastModified() < jspReal.lastModified()))
            needsCompile = true;
        // enforce the implication
        if (needsCompile)
            needsReload = true;
        // now do the necessary actions
        if (needsReload) {
            if (needsCompile) {
//System.out.println("compiling in the IDE");
                // try to compile in the IDE
                // now ask the ide to compile this page for us
                // call the compilation servlet to find out the result of the compilation
                if (!callCompilationServlet(jspReal.getAbsolutePath(), jspUri)) {
                    // compilation failed
                    throw new CompilationFailedException();
                }
            }

            // now reload, whether is was just compiled or not
            // find out the class name again
            ctxt = new JspEngineContext(loader, classpath,
                                                         context, jspUri, 
                                                         isErrorPage, options,
                                                         req, res);
            compiler = (JspCompiler)ctxt.createCompiler();

            classFileName = compiler.getClassFileName();

            // hack - taken from JspCompiler.compile()
            mangler = (Mangler)compiler;
            pkgName = mangler.getPackageName();
            ctxt.setServletPackageName(pkgName);
            javaFileName = mangler.getJavaFileName();
            ctxt.setServletJavaFileName(javaFileName);
            className = mangler.getClassName();
            ctxt.setServletClassName(className);
            // end hack
        }
        
        try {
            /*if( null ==ctxt.getServletClassName() ) {
                compiler.computeServletClassName();
            }*/
//System.out.println("setting full class name " + ctxt.getFullClassName());            
            setServletClassInWrapper(jspUri, loader.loadClass(ctxt.getFullClassName()));
            //loadClass(ctxt.getFullClassName(), true);
        }
        catch (ThreadDeath tdth) {
            throw tdth;
        } 
        catch (Throwable cex) {
            return super.doLoadJSP(jspUri, classpath, isErrorPage, req, res);
            //throw new JasperException(Constants.getString("jsp.error.unable.load"), cex);
        }
        loadedJSPTimes.put(jspUri, new Long(classFile.lastModified()));
//System.out.println("returning reload YES");                
        return true;
            
    }
    
    /** Calls the servlet which compiles the page in the IDE.
    * @param jspFileName JSP file to compile
    * @return true if the compilation succeeded
    */
    private boolean callCompilationServlet(String jspFileName, String jspUri) {
        try {
            File tempFile = (File)getServletConfig().getServletContext().getAttribute("javax.servlet.context.tempdir");
            File portFile = new File(tempFile.getParentFile(), "port");
            BufferedReader br = new BufferedReader(new FileReader(portFile));
            String portStr = br.readLine();
            br.close();
            int port = Integer.parseInt(portStr);
            
            // construct the URI of the request for the compilation servlet
            StringBuffer uri = new StringBuffer();
            uri.append("/servlet/");
            uri.append(COMPILE_SERVLET_NAME);
            uri.append("?");
            uri.append(COMPILE_SERVLET_PARAM_NAME);
            uri.append("=");
            uri.append(URLEncoder.encode(jspFileName));
            uri.append("&");
            uri.append(COMPILE_SERVLET_JSP_URI);
            uri.append("=");
            uri.append(URLEncoder.encode(jspUri));
            
            URL url = new URL("http", Util.getUtil().getLocalHost(), port, uri.toString());
            String response = Util.getUtil().issueGetRequest(url);
            if (response == null)
                return false;
            return (response.startsWith(COMPILE_SERVLET_OK_RESPONSE));
        }
        catch (MalformedURLException e) {
            return false;
        }
        catch (IOException e) {
            return false;
        }
    }

    /** Outputs a page saying "Compilation failed, look at the IDE for the errors. */
    private void outputCompilationFailedPage(HttpServletRequest request, HttpServletResponse response)
	throws ServletException, IOException {
        response.setContentType("text/html");
        PrintWriter pw = response.getWriter();
        pw.println(org.openide.util.NbBundle.getBundle(JspServlet.class).getString("MSG_CompilationFailed"));
        pw.close();
        /*response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
            org.openide.util.NbBundle.getBundle(JspServlet.class).getString("MSG_CompilationFailed"));*/
    }
    
    private Class getServletClassInWrapper(String jspUri) throws JasperException {
	Object jsw= jsps.get(jspUri);
	if( jsw==null ) {
	    throw new JasperException("Can't happen - JspServletWrapper=null");
	}
        try {
            Class jswClass = jsw.getClass();
            Field ff = jswClass.getDeclaredField("servletClass");
            ff.setAccessible(true);
            return (Class)ff.get(jsw);
        }
        catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }
    
    private void setServletClassInWrapper(String jspUri, Class servletClass) throws JasperException {
	Object jsw= jsps.get(jspUri);
	if( jsw==null ) {
	    throw new JasperException("Can't happen - JspServletWrapper=null");
	}
        try {
            Class jswClass = jsw.getClass();
            Field ff = jswClass.getDeclaredField("servletClass");
            ff.setAccessible(true);
            ff.set(jsw, servletClass);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    
}
