/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.Map;

import javax.servlet.jsp.tagext.TagInfo;
import javax.servlet.jsp.tagext.TagLibraryInfo;
import javax.servlet.jsp.tagext.TagAttributeInfo;
import javax.servlet.jsp.PageContext;

import org.apache.jasper.compiler.ParseEventListener;
import org.apache.jasper.compiler.Mark;
import org.apache.jasper.compiler.ParseException;
import org.apache.jasper.compiler.CompileException;
import org.apache.jasper.compiler.JspReader;
import org.apache.jasper.compiler.TagLibraries;
import org.apache.jasper.compiler.TagLibraryInfoImpl;
import org.apache.jasper.compiler.JspUtil;
import org.apache.jasper.JasperException;
import org.apache.jasper.Constants;
import org.apache.jasper.JspCompilationContext;

import org.openide.util.NbBundle;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.cookies.SaveCookie;
import org.openide.loaders.DataObject;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.TopManager;

/** Analyzer for JSP files. Looks for beans, included files,
* include and forward directives, error pages. Included files 
* are processed recursively. */
class AnalyzerParseEventListener implements ParseEventListener {
    
    /** Mode in which some errors (such as error parsing a tag library) are ignored. */
    public static final int ERROR_IGNORE = 1;
    /** Mode in which some errors (such as error parsing a tag library) are reported, 
     * but no accurate error description is needed. */
    public static final int ERROR_REPORT_ANY = 2;
    /** Mode in which an accurate description of all errors is required, so an actual attempt to parse all
     * tag libraries is done, so the parser throws a root cause exception. */
    public static final int ERROR_REPORT_ACCURATE = 3;

    protected JspReader reader;
    protected AnalyzerCompilerContext ctxt;
    //protected CompilationDescriptor cd;
    //protected String pageContextURL;  // the JSP page URL within the context

    private String[] includedFiles;
    private BeanDataImpl[] beans;
    private String[] errorPage;
    private String[] referencedPages;
    private TagLibraryDataImpl[] tagLibraryData;
    private boolean isErrorPage;
    private String servletContentType;
    private String encoding;

    private boolean iserrorpageDir;
    private boolean errorpageDir;
    private boolean contentTypeDir;

    private ArrayList includedFileList;
    private ArrayList beanList;
    private ArrayList errorPageList;
    private ArrayList referencedPageList;
    private ArrayList tagLibraryDataList;

    TagLibraries libraries;

    private boolean doSave;
    private int errorReportingMode;

    /** Constructs a new AnalyzerParseEventListener. 
     * @param jspreader a reader used to read the JSP
     * @param ctxt compiler content to use with this listener
     * @param doSave whether files included in the compiled page by the include directive should be saved before parsing
     * @param errorReportingMode mode of reporting errors (ignore, any exception, accurate exception)
     */
    AnalyzerParseEventListener(JspReader jspreader, AnalyzerCompilerContext ctxt, boolean doSave, int errorReportingMode) {
        reader = jspreader;
        this.ctxt = ctxt;
        this.doSave = doSave;
        this.errorReportingMode = errorReportingMode;
        //this.cd = cd;
        //this.pageContextURL = pageContextURL;
        /*    int lastSlash = pageContextURL.lastIndexOf('/');
            if (lastSlash == -1)
              throw new IllegalArgumentException();
            pageDirURL = pageContextURL.substring(0, lastSlash + 1);  */
    }

    public void setTemplateInfo(Mark start, Mark stop) {
    }

    public void beginPageProcessing() throws JasperException {
        includedFileList = new ArrayList();
        beanList = new ArrayList();
        errorPageList = new ArrayList();
        referencedPageList = new ArrayList();
        tagLibraryDataList = new ArrayList();
        isErrorPage = false;
        servletContentType = Constants.SERVLET_CONTENT_TYPE;
        // attributes
        encoding = "8859_1";
        // duplicate attribute guards
        iserrorpageDir = false;
        errorpageDir = false;
        contentTypeDir = false;
        libraries = new TagLibraries(TopManager.getDefault().currentClassLoader());
    }

    public void endPageProcessing() throws JasperException {
        includedFiles = (String[])includedFileList.toArray(new String[includedFileList.size()]);
        beans = (BeanDataImpl[])beanList.toArray(new BeanDataImpl[beanList.size()]);
        errorPage = (String[])errorPageList.toArray(new String[errorPageList.size()]);
        referencedPages = (String[])referencedPageList.toArray(new String[referencedPageList.size()]);
        tagLibraryData = (TagLibraryDataImpl[])tagLibraryDataList.toArray(new TagLibraryDataImpl[tagLibraryDataList.size()]);
        String newEncoding = extractEncoding(servletContentType);
        if (newEncoding != null) encoding = newEncoding;
    }

    public void handleComment(Mark start, Mark stop) throws JasperException {
    }

    // Page directive handlers - cut & paste from Jakarta
    interface PageDirectiveHandler {
        void handlePageDirectiveAttribute(AnalyzerParseEventListener listener,
                                          String value,
                                          Mark start, Mark stop)
        throws JasperException;
    }

    static final class PageDirectiveHandlerInfo {
        String attribute;
        PageDirectiveHandler handler;
        PageDirectiveHandlerInfo(String attribute, PageDirectiveHandler handler) {
            this.attribute = attribute;
            this.handler = handler;
        }
    }

    static final String errorPageStr = "errorPage";
    static final String isErrorPageStr = "isErrorPage";
    static final String contentTypeStr = "contentType";


    PageDirectiveHandlerInfo[] pdhis = new PageDirectiveHandlerInfo[] {
                                           new PageDirectiveHandlerInfo(isErrorPageStr, new IsErrorPageHandler()),
                                           new PageDirectiveHandlerInfo(contentTypeStr, new ContentTypeHandler()),
                                           new PageDirectiveHandlerInfo(errorPageStr, new ErrorPageHandler())
                                       };

    static final class ContentTypeHandler implements PageDirectiveHandler {
        public void handlePageDirectiveAttribute(AnalyzerParseEventListener listener,
                String contentType,
                Mark start, Mark stop)
        throws JasperException
        {
            if (listener.contentTypeDir == true)
                throw new CompileException(start, Constants.getString("jsp.error.page.multiple.contenttypes"));
            listener.contentTypeDir = true;
            if (contentType == null)
                throw new CompileException(start, Constants.getString("jsp.error.page.invalid.contenttype"));
            listener.servletContentType = contentType;
        }
    }

    static final class IsErrorPageHandler implements PageDirectiveHandler {
        public void handlePageDirectiveAttribute(AnalyzerParseEventListener listener,
                String iserrorpage,
                Mark start, Mark stop)
        throws JasperException
        {
            if (listener.iserrorpageDir == true)
                throw new CompileException (start, Constants.getString("jsp.error.page.multiple.iserrorpage"));

            listener.iserrorpageDir = true;
            if (iserrorpage == null)
                throw new CompileException(start, Constants.getString("jsp.error.page.invalid.iserrorpage"));

            if (iserrorpage.equalsIgnoreCase("true"))
                listener.isErrorPage = true;
            else if (iserrorpage.equalsIgnoreCase("false"))
                listener.isErrorPage = false;
            else
                throw new CompileException(start, Constants.getString("jsp.error.page.invalid.iserrorpage"));
        }
    }

    static final class ErrorPageHandler implements PageDirectiveHandler {
        public void handlePageDirectiveAttribute(AnalyzerParseEventListener listener,
                String errorpage,
                Mark start, Mark stop)
        throws JasperException
        {
            if (listener.errorpageDir == true)
                throw new CompileException(start, Constants.getString("jsp.error.page.multiple.errorpage"));

            listener.errorpageDir = true;
            if (errorpage != null) {
                String absolute = listener.ctxt.resolveRelativeUri(errorpage);
                listener.errorPageList.add(absolute);
            }
        }
    }

    public void handleDirective(String directive, Mark start, Mark stop, Hashtable attrs)
    throws JasperException {

        // page directive
        if (directive.equals("page")) {
            Enumeration e = attrs.keys();
            String attr;
            while (e.hasMoreElements()) {
                attr = (String) e.nextElement();
                for(int i = 0; i < pdhis.length; i++) {
                    PageDirectiveHandlerInfo pdhi = pdhis[i];
                    if (attr.equals(pdhi.attribute)) {
                        String value = (String) attrs.get(pdhi.attribute);
                        pdhi.handler.handlePageDirectiveAttribute(this, value,
                                start, stop);
                    }
                }
            }
        }

        // taglib directive
        if (directive.equals("taglib")) {
            String uri = (String) attrs.get("uri");
            String prefix = (String) attrs.get("prefix");
            try {
                // try to get a "cached" copy first, if that fails, then in case accurate error is not required,
                // exception is throws. If an accurate error is required, TagLibraryInfoImpl is created so
                // it throws an accurate exception
                TagLibraryInfo tl = null;
                String resolvedURI = resolveTagLibURI(uri);
                if (resolvedURI != null) {
                    TagLibraryDataImpl tldi = new TagLibraryDataImpl(prefix, uri, resolvedURI, null);
                    tl = getTagLibraryInfo(tldi);
                }
                if (tl == null) {
                    switch (errorReportingMode) {
                        case ERROR_IGNORE:
                            tl = getDummyTagLibraryInfo(prefix, uri);
                            break;
                        case ERROR_REPORT_ANY:
                            throw new CompileException(start, NbBundle.getBundle(AnalyzerParseEventListener.class).
                                getString("EXC_TagLibProcessError"));
                        case ERROR_REPORT_ACCURATE:
                            // this will throw an "accurate" exception
                            tl = new TagLibraryInfoImpl(ctxt, prefix, uri);
                            break;
                    }
                }
                
                libraries.addTagLibrary(prefix, tl);
                // add to my structure too
                tagLibraryDataList.add(new TagLibraryDataImpl(prefix, uri, resolvedURI, tl));
            } catch (Exception ex) {
                if (Boolean.getBoolean("netbeans.debug.exceptions")) {
                    ex.printStackTrace();
                }
                Object[] args = new Object[] { uri, ex.getMessage() };
                throw (CompileException)TopManager.getDefault().getErrorManager().annotate(
                    new CompileException(start, Constants.getString("jsp.error.badtaglib", args)), ex);
            }
        }

        // include directive
        if (directive.equals("include")) {
            String file = (String) attrs.get("file");
            String encoding = (String) attrs.get("encoding");

            if (file == null)
                throw new CompileException(start, Constants.getString("jsp.error.include.missing.file"));

            // add to the file list, maybe save
            String absFile = ctxt.resolveRelativeUri(file);
            includedFileList.add(absFile);

            if (doSave) {
                try {
                    // find the corresponding DO and save it if it needs saving
                    FileObject fo = ctxt.getResourceAsObject(file);
                    if (fo != null) {
                        DataObject dobj = DataObject.find(fo);
                        SaveCookie sc = (SaveCookie) dobj.getCookie (SaveCookie.class);
                        if (sc != null)
                            sc.save();
                    }
                }
                catch (DataObjectNotFoundException e) {
                    // not a bad error
                }
                catch (IOException e) {
                    // not a bad error
                }
            }

            // jsp.error.include.bad.file needs taking care of here??
            try {
                reader.pushFile(file, encoding);
            } catch (FileNotFoundException fnfe) {
                throw new CompileException(start, Constants.getString("jsp.error.include.bad.file"));
            }
        }

    }

    public void handleDeclaration(Mark start, Mark stop, Hashtable attrs) throws JasperException {
    }

    public void handleScriptlet(Mark start, Mark stop, Hashtable attrs) throws JasperException {
    }

    public void handleExpression(Mark start, Mark stop, Hashtable attrs) throws JasperException {
    }

    public void handleBean(Mark start, Mark stop, Hashtable attrs) throws JasperException {
        if (attrs == null)
            throw new ParseException(start,
                                     NbBundle.getBundle(AnalyzerParseEventListener.class).
                                     getString("EXC_ClassTypeMissing"));
        String clazz = (String)attrs.get("class"); // NOI18N
        String type  = (String)attrs.get("type"); // NOI18N
        if(clazz == null && type == null) {
            throw new ParseException(start,
                                     NbBundle.getBundle(AnalyzerParseEventListener.class).
                                     getString("EXC_ClassTypeMissing"));
        }
        if (clazz == null)
            clazz = type;

        String name  = (String)attrs.get("id");

        String scope = (String)attrs.get("scope");
        int scopeInt;
        if (scope == null || scope.equals("page")) {
            scopeInt = PageContext.PAGE_SCOPE;
        } else if (scope.equals("request")) {
            scopeInt = PageContext.REQUEST_SCOPE;
        } else if (scope.equals("session")) {
            scopeInt = PageContext.SESSION_SCOPE;
        } else if (scope.equals("application")) {
            scopeInt = PageContext.APPLICATION_SCOPE;
        } else {
            String m = Constants.getString("jsp.error.usebean.invalid.scope",
                                           new Object[] {name, scope});
            throw new CompileException (start, m);
        }

        beanList.add(new BeanDataImpl(name, scopeInt, clazz));
    }

    public void handleBeanEnd(Mark start, Mark stop, Hashtable attrs)
    throws JasperException {
    }

    public void handleGetProperty(Mark mark, Mark mark1, Hashtable attrs)
    throws JasperException {
    }

    public void handleSetProperty(Mark mark, Mark mark1, Hashtable attrs)
    throws JasperException {
    }

    public void handlePlugin(Mark start, Mark stop, Hashtable attrs, Hashtable param, String fallback)
    throws JasperException {
    }

    public void handleCharData(Mark start, Mark stop, char[] chars) throws JasperException {
    }

    public void handleTagBegin(Mark start, Mark stop, Hashtable attrs, String prefix, String shortTagName,
                               TagLibraryInfo tli, TagInfo ti) throws JasperException {
    }

    public void handleTagEnd(Mark start, Mark stop, String prefix, String shortTagName,
                             Hashtable attrs, TagLibraryInfo tli, TagInfo ti) throws JasperException {
    }

    public void handleForward(Mark start, Mark stop, Hashtable attrs, Hashtable param)
    throws JasperException {
        handleIncludeForward(start, stop, attrs, param);
    }


    public void handleInclude(Mark start, Mark stop, Hashtable attrs, Hashtable param)
    throws JasperException {
        handleIncludeForward(start, stop, attrs, param);
    }

    private void handleIncludeForward(Mark start, Mark stop, Hashtable attrs, Hashtable param)
    throws JasperException {
        boolean isExpression = false;
        String page = (String)attrs.get("page"); // NOI18N
        if (page == null)
            throw new ParseException(start, Constants.getString("jsp.error.include.tag"));
        isExpression = JspUtil.isExpression(page);
        if (!isExpression) {
            page = ctxt.resolveRelativeUri(page); // NOI18N
            referencedPageList.add(page);
        }
        else {
            referencedPageList.add(page);
        }
    }

    public TagLibraries getTagLibraries() {
        return libraries;
    }

    /** Tries to resolve a given relative URI of a tag library against this JSP, with respect to web.xml
     * depends on errorReportingMode in the following way:
     * If errorReportingMode == ERROR_IGNORE, returns always non-null value regradless of any errors
     * Otherwise returns null in case of an error.
     */
    private String resolveTagLibURI(String relativeURI) {
        boolean errorOccurred = false;
        
        // look up web.xml to see if it contains my URI
        MultiDataObject webXML = null;
        FileObject webXMLFile = ctxt.getResourceAsObject("/WEB-INF/web.xml");
        if (webXMLFile != null) {
            try {
                webXML = (MultiDataObject)DataObject.find(webXMLFile);
            }
            catch (DataObjectNotFoundException e) {  
                errorOccurred = true;
            }
            catch (ClassCastException e) { 
                errorOccurred = true;
            }
        }
        
        WebXMLSupport sup = null;
        if (webXML != null) {
            sup = WebXMLSupport.getWebXMLSupport(webXML);
            if (sup == null)
                errorOccurred = true;
        }
        Map map = (sup == null) ? null : sup.getTagLibraryMap();
        String realLocation = (map == null) ? null : (String)map.get(relativeURI);
        // real location is already absolute, don't need to resolve against the location of web.xml
        if (realLocation != null)
            return realLocation;
        
        if (errorOccurred && (errorReportingMode != ERROR_IGNORE))
            return null;
        
        return ctxt.resolveRelativeUri(relativeURI);
    }

    /** Exctracts encoding from content type.
    *  Returns null if there is no encoding specified.
    */
    private String extractEncoding(String myContentType) {
        if (myContentType != null) {
            int loc = myContentType.indexOf("charset=");
            if (loc > 0) {
                return myContentType.substring(loc + 8);
            }
        }
        return null;
    }

    /** Gets the JspInfo retrieved from the parsed text */
    public JspInfo getJspInfo() {
        return new JspInfo() {

                   public String[] getIncludedFiles() {
                       return includedFiles;
                   }

                   public JspInfo.BeanData[] getBeans() {
                       return beans;
                   }

                   public String[] getErrorPage() {
                       return errorPage;
                   }

                   public String[] getReferencedPages() {
                       return referencedPages;
                   }

                   public JspInfo.TagLibraryData[] getTagLibraryData() {
                       return tagLibraryData;
                   }

                   public boolean isErrorPage() {
                       return isErrorPage;
                   }

                   public String getEncoding() {
                       return encoding;
                   }

               };
    }

/*    private TagLibraries instantiateTagLibraries(ClassLoader cl) {
        try {
            java.lang.reflect.Constructor constr = TagLibraries.class.getDeclaredConstructor(new Class[] {ClassLoader.class});
            constr.setAccessible(true);
            return (TagLibraries)constr.newInstance(new Object[] {cl});
        }
        catch (Exception e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions"))
                e.printStackTrace();
            return null;
        }
    }

    private TagLibraryInfoImpl instantiateTagLibraryInfoImpl(JspCompilationContext ctxt,
            String prefix, String uriIn) throws IOException, JasperException{
        try {
            java.lang.reflect.Constructor constr = TagLibraryInfoImpl.class.getDeclaredConstructor(
                                                       new Class[] {JspCompilationContext.class, String.class, String.class});
            constr.setAccessible(true);
            return (TagLibraryInfoImpl)constr.newInstance(new Object[] {ctxt, prefix, uriIn});
        }
        catch (Exception e) {
            if (e instanceof java.lang.reflect.InvocationTargetException) {
                Throwable e2 = ((java.lang.reflect.InvocationTargetException)e).getTargetException();
                if (e2 instanceof IOException)
                    throw (IOException)e2;
                if (e2 instanceof JasperException)
                    throw (JasperException)e2;
            }
            if (Boolean.getBoolean("netbeans.debug.exceptions"))
                e.printStackTrace();
            return null;
        }
    }

    private void addTagLibraryToLibraries(TagLibraries libraries, String prefix, TagLibraryInfoImpl tli) {
        try {
            java.lang.reflect.Method adder = TagLibraries.class.getDeclaredMethod("addTagLibrary", new Class[] {String.class, TagLibraryInfoImpl.class});
            adder.setAccessible(true);
            adder.invoke(libraries, new Object[] {prefix, tli});
        }
        catch (Exception e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions"))
                e.printStackTrace();
        }
    }*/
    
    /** Returns a TagLibraryInfo for a given JspInfo.TagLibraryData. 
     *  Returns null if the creation failed for whatever reason.
     */
    private TagLibraryInfo getTagLibraryInfo(JspInfo.TagLibraryData info) {
        // find the tag library            
        FileObject tagLibFile = ctxt.getResourceAsObject(info.getResolvedURI());
        if (tagLibFile != null) {
            try {
                MultiDataObject tagLib = (MultiDataObject)DataObject.find(tagLibFile);
                TagLibraryInfoSupport sup = TagLibraryInfoSupport.getTagLibraryInfoSupport(tagLib);
                if (sup != null) {
                    return sup.getTagLibraryInfo(info.getPrefix(), info.getUnresolvedURI());
                }
            }
            catch (DataObjectNotFoundException e) { /* ignore */ }
            catch (ClassCastException e) { /* ignore */ }
        }
        return null;
    }
    
    private TagLibraryInfo getDummyTagLibraryInfo(String prefix, String uri) {
        // new tag library info which does almost nothing, but recognizes all tags in its namespace
        return new DummyTagLibraryInfo(prefix, uri);
    }
    
    private class DummyTagLibraryInfo extends TagLibraryInfo {
        
        public DummyTagLibraryInfo(String prefix, String uri) {
            super (prefix, uri);
        }
        
        public TagInfo getTag(String shortName) {
            return new TagInfo(shortName, null, TagInfo.BODY_CONTENT_JSP, "", this, null, new TagAttributeInfo[0]);
        }
    }

    class BeanDataImpl implements JspInfo.BeanData {

        String id;
        int scope;
        String className;

        public BeanDataImpl(String id, int scope, String className) {
            this.id = id;
            this.scope = scope;
            this.className = className;
        }

        /** Identifier of the bean in the page (variable name). */
        public String getId() {
            return id;
        }

        /** Scope for this bean. Returns constants defined in {@link javax.servlet.jsp.PageContext}. */
        public int getScope() {
            return scope;
        }

        /** Returns the class name for this bean. */
        public String getClassName() {
            return className;
        }
    }

    class TagLibraryDataImpl implements JspInfo.TagLibraryData {

        String prefix;
        String unresolvedURI;
        String resolvedURI;
        TagLibraryInfo tagLib;

        public TagLibraryDataImpl(String prefix, String unresolvedURI, String resolvedURI, TagLibraryInfo tagLib) {
            this.prefix = prefix;
            this.unresolvedURI = unresolvedURI;
            this.resolvedURI = resolvedURI;
            this.tagLib = tagLib;
        }

        /** Returns prefix with which the taglib was imported. */
        public String getPrefix() {
            return prefix;
        }

        /** Returns literal value of the uri attribute. */
        public String getUnresolvedURI() {
            return unresolvedURI;
        }

        /** Returns URI resolved to a context-relative URI. */
        public String getResolvedURI() {
            return resolvedURI;
        }

        /** non-standard method, which gives the implementation of TagLibraryInfo, because
        * Jakarta does not allow me NOT to create this, so I might as well use it. 
        * I'd rather use Simran's TagLibraryInfo implementation, but since the parser instantiates it ...
        */
        public TagLibraryInfo getTagLibraryInfo() {
            return tagLib;
        }

    }
}