/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.beans.PropertyChangeEvent;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.JEditorPane;
import javax.swing.Timer;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.SwingUtilities;

import org.openide.actions.*;
import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.DataObject;
import org.openide.text.EditorSupport;
import org.openide.windows.*;
import org.openide.util.TaskListener;
import org.openide.util.Task;
import org.openide.util.actions.SystemAction;


/* This ediror handles calling a parser to reparse tag library information about the page.
*
* @author Petr Jiricka
*/
public class BaseJspEditor extends EditorSupport {

  private static final int AUTO_PARSING_DELAY = 2000;
  
  /** Timer which countdowns the auto-reparsing time. */
  Timer timer;

  /** Create a new Editor support for the given Java source.
  * @param entry the (primary) file entry representing the Java source file
  */
  public BaseJspEditor(MultiDataObject.Entry entry) {
    super(entry);

    // initialize timer
    timer = new Timer(0, new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent e) {
//System.out.println("timer off");
        DataObject dObj = getFileEntry().getDataObject();
        final TagLibParseSupport sup = (TagLibParseSupport)dObj.getCookie(TagLibParseSupport.class);
        if (sup != null) {
            sup.autoParse().addTaskListener(new TaskListener() {
              public void taskFinished(Task t) {
                notifyParsingDone(sup);
              }
            });
        }
      }
    });
    timer.setInitialDelay(AUTO_PARSING_DELAY);
    timer.setRepeats(false);

    // create document listener
    final DocumentListener docListener = new DocumentListener() {
      public void insertUpdate(DocumentEvent e) { change(e); }
      public void changedUpdate(DocumentEvent e) { }
      public void removeUpdate(DocumentEvent e) { change(e); }
      
      private void change(DocumentEvent e) {
        restartTimer(false);
        DataObject dObj = getFileEntry().getDataObject();
        TagLibParseSupport sup = (TagLibParseSupport)dObj.getCookie(TagLibParseSupport.class);
        if (sup != null) {
            sup.setDocumentDirty(true);
        }
      }
    };

    // add change listener
    addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent evt) {
        if (isDocumentLoaded()) {
          getDocument().addDocumentListener(docListener);
        }
      }
    });
  }
  
  /** Restart the timer which starts the parser after the specified delay.
  * @param onlyIfRunning Restarts the timer only if it is already running
  */
  void restartTimer(boolean onlyIfRunning) {
    if (onlyIfRunning && !timer.isRunning())
      return;
    
    int delay = AUTO_PARSING_DELAY;
    if (delay > 0) {
      timer.setInitialDelay(delay);
      timer.restart();
    }
  }
  
    /** Notify about the editor closing.
    */
    protected void notifyClose() {
        DataObject dObj = getFileEntry().getDataObject();
        TagLibParseSupport sup = (TagLibParseSupport)dObj.getCookie(TagLibParseSupport.class);
        if (sup != null) {
            sup.prepare();
        }
        //parsedHook = null;
    }

    /** Notify that parsing task has been finished; some dependent data may now
      be refreshed from up-to-date parsing info */
    protected void notifyParsingDone(TagLibParseSupport sup) {
//System.out.println("parsing done");
        if (sup.isDocumentDirty()) {
            restartTimer(false);
        }
    }
    
    protected String getObjectEncoding() {
        DataObject dObj = getFileEntry().getDataObject();
        if (dObj instanceof JspDataObject)
            return ((JspDataObject)dObj).getEncoding();
        else
            return JspDataObject.getDefaultEncoding();
    }
    
  
  /** Read the file from the stream, filter the guarded section
  * comments, and mark the sections in the editor.
  *
  * @param doc the document to read into
  * @param stream the open stream to read from
  * @param kit the associated editor kit
  * @throws IOException if there was a problem reading the file
  * @throws BadLocationException should not normally be thrown
  * @see #saveFromKitToStream
  */
    protected void loadFromStreamToKit (StyledDocument doc, InputStream stream, EditorKit kit) throws IOException, BadLocationException {
        Reader reader = null;
        try {
            reader = new InputStreamReader(stream, getObjectEncoding());
            kit.read(reader, doc, 0);
        }
        finally {
            if (reader != null) 
                reader.close();
        }
        
        /**
         NOT NEEDED, HOPEFULLY

        // parse it and remember the hook to prevent to garbage collected
        // parsed hierarchy
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                DataObject dObj = getFileEntry().getDataObject();
                final TagLibParseSupport sup = (TagLibParseSupport)dObj.getCookie(TagLibParseSupport.class);
                if (sup != null) {
                    Task task = sup.prepare();
                }
            }
        });*/
    }

    protected void saveFromKitToStream (StyledDocument doc, EditorKit kit, OutputStream stream) throws IOException, BadLocationException {
        Writer wr = null;
        try {
            wr = new OutputStreamWriter(stream, getObjectEncoding());
            kit.write(wr, doc, 0, doc.getLength());
        }
        finally {
            if (wr != null)
                wr.close();
        }
    }

  
    /** Save the document in this thread and start reparsing it.
    * @exception IOException on I/O error
    */
    public void saveDocument () throws IOException {
        saveDocument(true, true);
    }

  /** Save the document in this thread.
  * @param parse true if the parser should be started, otherwise false
  * @exception IOException on I/O error
  */
  protected void saveDocumentIfNecessary(boolean parse) throws IOException {
      saveDocument(parse, false);
  }

  /** Save the document in this thread.
  * @param parse true if the parser should be started, otherwise false
  * @param forceSave if true save always, otherwise only when is modified
  * @exception IOException on I/O error
  */
  private void saveDocument(boolean parse, boolean forceSave) throws IOException {
    if (forceSave || isModified()) {
      super.saveDocument();
      if (parse) {
        DataObject dObj = getFileEntry().getDataObject();
        TagLibParseSupport sup = (TagLibParseSupport)dObj.getCookie(TagLibParseSupport.class);
        if (sup != null) {
            sup.prepare();
        }
      }
    }
  }

 // ==================== Misc not-public methods ========================

  /* A method to create a new component. Overridden in subclasses.
  * @return the {@link Editor} for this support
  */
  protected CloneableTopComponent createCloneableTopComponent () {
        // initializes the document if not initialized
        prepareDocument ();
        DataObject dObj = getFileEntry().getDataObject();
        return new BaseJspEditorComponent(dObj);
  }

    /** @return JSP Data Object */
    public DataObject getDataObject() {
        return findDataObject();
    }


    /** only accessibility method */
    protected MultiDataObject.Entry getFileEntry() {
        return entry;
    }

    /** The real component of the XML bean editor.
    * Subclasses should not attempt to work with this;
    * if they require special editing support, separate windows
    * should be created by overriding (e.g.) {@link EditorSupport#open}.
    */
    public static class BaseJspEditorComponent extends EditorSupport.Editor {
    
        /** The support, subclass of EditorSupport */
        protected BaseJspEditor support;

        /** Listener on caret movements */
        CaretListener caretListener;

        /** Only for externalization */
        public BaseJspEditorComponent () {
            super();
        }

        /** Creates new editor */
        public BaseJspEditorComponent (DataObject obj) {
            super(obj);
            initialize();
        }

        public SystemAction[] getSystemActions() {
            SystemAction[] sa = super.getSystemActions ();
            SystemAction[] jspActions = new SystemAction[] {
                SystemAction.get(CompileAction.class),
                SystemAction.get(ExecuteAction.class),
                SystemAction.get(ToolsAction.class),
                SystemAction.get(PropertiesAction.class)
            };
            return SystemAction.linkActions (sa, jspActions);
        }
        
        protected void notifyParsingDone() {
        }

        /** Obtain a support for this component */
        private void initialize () {
            support = (BaseJspEditor)obj.getCookie(BaseJspEditor.class);

            caretListener = new CaretListener() {
                public void caretUpdate(CaretEvent e) {
                    support.restartTimer(true);
                }
            };
        }

        /** Returns Editor pane for private use.
        * @return Editor pane for private use.
        */
        public JEditorPane getEditorPane () {
            return pane;
        }

        /* Is called from the clone method to create new component from this one.
        * This implementation only clones the object by calling super.clone method.
        * @return the copy of this object
        */
        protected CloneableTopComponent createClonedObject () {
            return support.createCloneableTopComponent();
        }

        /* This method is called when parent window of this component has focus,
        * and this component is preferred one in it. This implementation adds 
        * performer to the ToggleBreakpointAction.
        */
        protected void componentActivated () {
            getEditorPane().addCaretListener(caretListener);
            super.componentActivated ();
        }

        /*
        * This method is called when parent window of this component losts focus,
        * or when this component losts preferrence in the parent window. This 
        * implementation removes performer from the ToggleBreakpointAction.
        */
        protected void componentDeactivated () {
            getEditorPane().removeCaretListener(caretListener);
            super.componentDeactivated ();
        }

        /* When closing last view, also close the document.
         * @return <code>true</code> if close succeeded
        */
        protected boolean closeLast () {
            if (!super.closeLast())
                return false;

            support.notifyClose();
            return true;
        }

        /** Deserialize this top component.
        * @param in the stream to deserialize from
        */
        public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException {
            super.readExternal(in);
            initialize();
        }  
    
    } // end of JavaEditorComponent inner class

}

