/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.MalformedURLException;

import org.openide.execution.NbClassPath;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.TopManager;

/** Descriptor of a web application context. This is a per-context structure,
 *  independent on a particular request.
 */

public class ContextDescriptor {

    private FileSystem baseFs;

    /** Constructs new ContextDescriptor.
    *  @param fs filesystem for which to create the ServletContext. 
    *  If this is null ServletContext for the whole repository is created.
    */
    public ContextDescriptor(FileSystem fs) {
        baseFs = fs;
    }


    /** Returns a file: url
    * @param path a context-relative URI path
    */
    public URL getResource(String path) throws MalformedURLException {
        FileObject fo = getResourceAsObject(path);
        if (fo == null)
            return null;
        try {
            return getFileExternalURL(fo);
        }
        catch (FileStateInvalidException ex) {
            throw new MalformedURLException(ex.getClass().getName() + " : " + ex.toString()); // NOI18N
        }
    }

    /** Returns inputstream of a FileObject.
    * @param path a context-relative URI path
    */
    public InputStream getResourceAsStream(String path) {
        FileObject fo = getResourceAsObject(path);
        if (fo == null)
            return null;
        try {
            return fo.getInputStream();
        }
        catch (FileNotFoundException ex) {
            return null;
        }
    }

    /** Returns a file object for a given path, or <code>null</code> if not found.
    * @param path a context-relative URI path
    */
    public FileObject getResourceAsObject(String path) {
        path = path.replace('\\', '/');
        if (baseFs == null)
            return TopManager.getDefault().getRepository().findResource(path);
        else
            return baseFs.findResource(path);
    }

    private URL getFileExternalURL(FileObject fo) throws FileStateInvalidException {
        File ff = NbClassPath.toFile(fo);
        if (ff != null)
            try {
                return ff.toURL();
            }
            catch (MalformedURLException e) {
                return fo.getURL();
            }
        return fo.getURL();
    }


    /** Returns a real file path for a uri path. The path parameter must be a
    * context-relative URI. */
    public String getRealPath(String path) {
        String realPath = null;
        try {
            URL url = getResource(path);
            if (url != null && url.getProtocol().equals("file")) {// NOI18N
                FileObject fo = getResourceAsObject(path);
                if (fo != null) {
                    File ff = NbClassPath.toFile(fo);
                    if (ff != null)
                        realPath = ff.getAbsolutePath();
                }
            }
        }
        catch(Exception ex) {
        }
        return realPath;
    }

}

