/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.lang.reflect.*;
import java.net.URLEncoder;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.text.MessageFormat;

import org.apache.jasper.JspCompilationContext;
import org.apache.jasper.compiler.JspReader;
import org.apache.jasper.compiler.Parser;
import org.apache.jasper.runtime.JspLoader;

import org.openide.TopManager;
import org.openide.NotifyDescriptor;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.LocalFileSystem;
import org.openide.filesystems.FileSystemCapability;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.Repository;
import org.openide.loaders.DataObject;
import org.openide.compiler.Compiler;
import org.openide.execution.NbClassPath;
import org.openide.util.NbBundle;

// THIS IS THE ONLY ALLOWED IMPORT OF JAKARTA IN THIS PACKAGE
import org.netbeans.modules.web.core.jsploader.jakarta.JakartaServerPlugin;
import org.netbeans.modules.web.core.WebExecUtil;

/** JSP compilation utilities
*
* @author Petr Jiricka
*/
public class JspCompileUtil {

    private static ServerPlugin sp = null;

    public static ServerPlugin getCurrentServerPlugin() {
        if (sp == null)
            sp = new JakartaServerPlugin();
        return sp;
    }

    /** Finds a fileobject for an absolute file name or null if not found */
    public static FileObject findFileObjectForFile(String fileName) {
        Repository rep = TopManager.getDefault().getRepository();
        for (Enumeration en = rep.getFileSystems(); en.hasMoreElements(); ) {
            FileSystem fs = (FileSystem)en.nextElement();
            FileObject fo = findFileObjectForFile(fs, fileName);
            if (fo != null)
                return fo;
        }
        return null;
    }

    /** Finds a fileobject for an absolute file name or null if not found */
    public static FileObject findFileObjectForFile(FileSystem fs, String fileName) {
        String canName = ""; // NOI18N
        String canName2 = ""; // NOI18N
        try {
            canName = new File(fileName).getCanonicalPath();
        }
        catch (IOException e) {
            return null;
        }
        String resFriendly = canName.replace(File.separatorChar, '/');
        int sepFound = -1;
        while (true) {
            sepFound = resFriendly.indexOf('/', sepFound + 1);
            if (sepFound == -1)
                return null;
            String resName = resFriendly.substring(sepFound);
            FileObject fo = fs.findResource(resName);
            if (fo == null) continue;
            File ff = NbClassPath.toFile(fo);
            if (ff == null) continue;
            // verify that they're the same
            try {
                canName2 = ff.getCanonicalPath();
            }
            catch (IOException e) {
                continue;
            }
            if (!(canName2.equals(canName)))
                continue;
            return fo;
        }
    }

    public static final String getFileObjectFileName(FileObject fo) throws FileStateInvalidException {
        File ff = NbClassPath.toFile(fo);
        if (ff == null)
            throw new FileStateInvalidException(NbBundle.getBundle(JspCompileUtil.class).getString("CTL_NotLocalFile"));
        return ff.getAbsolutePath();
    }

    /** Gets the folder which is at the root of the context into which fo belongs
    */
    public static final FileObject getContextRoot(FileObject fo) throws FileStateInvalidException {
        // pending
        return fo.getFileSystem().getRoot();
    }

    public static final String getContextPath(FileObject fo) {
        return "/" + fo.getPackageNameExt('/','.');
    }

    /** Gets the folder which is the root output folder for the context into which fo belongs */
    public static final FileObject getContextOutputRoot(FileObject fo) throws IOException {
        File serverRoot = getOutputRootFolder();
        FileSystem fs = fo.getFileSystem();
        int hc = fs.getSystemName().hashCode();
        String dirName = Long.toHexString(hc <= 0 ? -hc : hc);
        File contextRoot = new File(serverRoot, URLEncoder.encode(dirName));
        return getAsRootOfFileSystem(contextRoot);
    }


    /** Gets the root OUTPUT folder for all contexts in NB systm FileSystem. */
    public static File getOutputRootFolder() {
        String path = WebExecUtil.getTempDirectory() + File.separator + "jspwork"; // NOI18N
        File myRoot = (new File(path)).getAbsoluteFile();
        return myRoot;
    }

    /** Does the following:
    * <ul>
    * <li>creates a hidden LocalFileSystem (with compile, execute and debug capabilities)
    * with root in <code>intendedRoot</code>, if it does not exist yet</li>
    * <li>returns the root of this filesystem
    * </ul> */
    private static FileObject getAsRootOfFileSystem(File intendedRoot) {
        // try to find it among current filesystems
        for (Enumeration en = TopManager.getDefault().getRepository().getFileSystems(); en.hasMoreElements(); ) {
            FileSystem fs = (FileSystem)en.nextElement();
            File root = NbClassPath.toFile(fs.getRoot());
            if (root != null) {
                if (root.equals(intendedRoot))
                    return fs.getRoot();
            }
        }

        // does not exist in repository
        if (!intendedRoot.exists()) {
            boolean success = WebExecUtil.myMkdirs(intendedRoot);
        }

        FileSystemCapability.Bean cap = new FileSystemCapability.Bean();
        cap.setCompile(true);
        cap.setExecute(true);
        cap.setDebug(true);
        cap.setDoc(false);
        LocalFileSystem newFs = new LocalFileSystem(cap);
        try {
            newFs.setRootDirectory(intendedRoot);
        }
        catch (Exception e) {
            NotifyDescriptor.Message message = new NotifyDescriptor.Message(
                                                   MessageFormat.format(NbBundle.getBundle(JspCompileUtil.class).
                                                                        getString("EXC_JspFSNotCreated"),
                                                                        new Object[] {intendedRoot.getAbsolutePath()}), NotifyDescriptor.ERROR_MESSAGE);
            TopManager.getDefault().notify(message);
            return null;
        }
        newFs.setHidden(true);
        TopManager.getDefault().getRepository().addFileSystem(newFs);
        return newFs.getRoot();
    }

    /** Returns an absolute context URL (starting with '/') for a relative URL and base URL.
    *  @param relativeTo url to which the relative URL is related. Treated as directory iff
    *    ends with '/'
    *  @param url the relative URL by RFC 2396
    *  @exception IllegalArgumentException if url is not absolute and relativeTo 
    * can not be related to, or if url is intended to be a directory
    */
    public static String resolveRelativeURL(String relativeTo, String url) {
        //System.out.println("- resolving " + url + " relative to " + relativeTo);
        String result;
        if (url.startsWith("/")) { // NOI18N
            result = "/"; // NOI18N
            url = url.substring(1);
        }
        else {
            // canonize relativeTo
            if ((relativeTo == null) || (!relativeTo.startsWith("/"))) // NOI18N
                throw new IllegalArgumentException();
            relativeTo = resolveRelativeURL(null, relativeTo);
            int lastSlash = relativeTo.lastIndexOf('/');
            if (lastSlash == -1)
                throw new IllegalArgumentException();
            result = relativeTo.substring(0, lastSlash + 1);
        }

        // now url does not start with '/' and result starts with '/' and ends with '/'
        StringTokenizer st = new StringTokenizer(url, "/", true); // NOI18N
        while(st.hasMoreTokens()) {
            String tok = st.nextToken();
            //System.out.println("token : \"" + tok + "\""); // NOI18N
            if (tok.equals("/")) { // NOI18N
                if (!result.endsWith("/")) // NOI18N
                    result = result + "/"; // NOI18N
            }
            else
                if (tok.equals("")) // NOI18N
                    ;  // do nohing
                else
                    if (tok.equals(".")) // NOI18N
                        ;  // do nohing
                    else
                        if (tok.equals("..")) { // NOI18N
                            String withoutSlash = result.substring(0, result.length() - 1);
                            int ls = withoutSlash.lastIndexOf("/"); // NOI18N
                            if (ls != -1)
                                result = withoutSlash.substring(0, ls + 1);
                        }
                        else {
                            // some file
                            result = result + tok;
                        }
            //System.out.println("result : " + result); // NOI18N
        }
        //System.out.println("- resolved to " + result);
        return result;
    }


    public static JspInfo analyzePage(JspDataObject jspPage, String compilationURI, 
    boolean doSaveIncluded, int errorReportingMode) throws Exception {
        OptionsImpl options = new OptionsImpl(jspPage);
        
        CompilationDescriptor cd = new CompilationDescriptor(
                                       getContextRoot(jspPage.getPrimaryFile()).getFileSystem(), compilationURI);
        String jspResource = getContextPath(jspPage.getPrimaryFile());

        AnalyzerCompilerContext ctxt = new AnalyzerCompilerContext(jspResource, cd, options);
        JspReader reader = JspReader.createJspReader(jspResource, ctxt, "8859_1");
        ctxt.setReader(reader);

        AnalyzerParseEventListener listener = new AnalyzerParseEventListener(reader, ctxt, 
            doSaveIncluded, errorReportingMode);
        Parser parser = new Parser(reader, listener);
        listener.beginPageProcessing();
        parser.parse();
        listener.endPageProcessing();
        return listener.getJspInfo();
    }


    public static FileSystem mountJarIfNotMounted(String jarFileName) throws IOException {
        //if(libs.containsKey(jarFileName)) return (FileSystem) libs.get(jarFileName);
        org.openide.filesystems.JarFileSystem jfs = new org.openide.filesystems.JarFileSystem();
        File f = new File(jarFileName);
        try {
            jfs.setJarFile(f);
        } catch (java.beans.PropertyVetoException jbp) {}
        FileSystem fs = mountFS(jfs);
        //libs.put(jarFileName, fs);
        return fs;
    }

    private static FileSystem mountFS(FileSystem fs) throws IOException {
        // compare with the other filesystems
        Enumeration en = TopManager.getDefault().getRepository().getFileSystems();
        for (;en.hasMoreElements(); ) {
            FileSystem f = (FileSystem)en.nextElement();
            if (f.getSystemName().equals(fs.getSystemName()))
                return f;
        }
        
        // add to the repository
        fs.setHidden(true);
        TopManager.getDefault().getRepository().addFileSystem(fs);
        return fs;
    }

}

