/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.util.Set;
import java.util.Iterator;
import java.util.Comparator;
import java.util.Collection;
import java.util.ArrayList;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.io.ObjectInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.NotActiveException;
import java.io.FileNotFoundException;
import java.io.File;
import javax.swing.text.StyledDocument;
import javax.swing.text.BadLocationException;

import org.openide.*;
import org.openide.cookies.EditorCookie;
import org.openide.cookies.CompilerCookie;
import org.openide.cookies.SaveCookie;
import org.openide.cookies.SourceCookie;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.windows.*;
import org.openide.actions.OpenAction;
import org.openide.actions.ViewAction;
import org.openide.text.*;
import org.openide.util.*;
import org.openide.util.actions.*;
import org.openide.nodes.Node;
import org.openide.nodes.Children;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.NodeListener;
import org.openide.compiler.Compiler;
import org.openide.compiler.CompilerJob;
import org.openide.compiler.CompilerType;

import org.netbeans.modules.web.core.ServletSettings;
import org.netbeans.modules.web.core.WebExecSupport;

import org.netbeans.modules.java.JavaDataObject;
import org.netbeans.modules.java.JavaCompilerType;
import java.util.Hashtable;

/** Object that provides main functionality for internet data loader.
*
* @author Petr Jiricka
*/
public class JspDataObject extends MultiDataObject {

    public static final String EA_CONTENT_LANGUAGE = "AttrJSPContentLanguage";
    public static final String EA_SCRIPTING_LANGUAGE = "AttrJSPScriptingLanguage";
    public static final String EA_ENCODING = "AttrEncoding";
    public static final String EA_JSP_ERRORPAGE = "jsp_errorpage"; // NOI18N
    // property for the servlet dataobject corresponding to this page
    public static final String PROP_SERVLET_DATAOBJECT = "servlet_do"; // NOI18N
    public static final String PROP_CONTENT_LANGUAGE   = "contentLanguage"; // NOI18N
    public static final String PROP_SCRIPTING_LANGUAGE = "scriptingLanguage"; // NOI18N
    public static final String PROP_ENCODING = "encoding"; // NOI18N

    transient private SourceCookie.Editor servletEdit;
    transient protected JspServletDataObject servletDataObject;
    transient private ServerPlugin.JspPlugin jspPlugin;
    
    // this hashtable is necessery for clonning compilerTypes
    // and for setting parameter -encoding to UTF8 to these cloned compilerTypes
    transient private static Hashtable clonedCompilers;
    
    public JspDataObject (FileObject pf, final UniFileLoader l) throws DataObjectExistsException {
        super (pf, l);
        initialize();
    }

    protected org.openide.nodes.Node createNodeDelegate () {
        return new JspNode (this);
    }
    
    /** Creates a EditorSupport for this page. May return null. */
    protected EditorSupport createJspEditor() {
        return new BaseJspEditor(getPrimaryEntry());
    }
    
    protected SourceCookie.Editor createServletEditor() {
        return new ServletEditor(this);
    }
    
    /** Creates an execution (and debugging) support for this page. May return null. */
    protected Node.Cookie createExecSupport() {
        return new WebExecSupport(getPrimaryEntry ());
    }
    
    /** Creates a compiler which will be called after the JSP to servlet translation has finished, but
    * before the servlet to class compilation. May return null. 
    * @param type type of compiler cookie which is creating the compiler.
    */
    protected Compiler createPostTranslateCompiler(Class type, JspCompilationInfo compInfo) {
        return null;
    }

    /** Creates a compiler which will be called before the JSP to servlet translation is called. 
     * If CleanCompiler is created, will be called before the CleanCompiler.
     * May return null. 
     * @param type type of compiler cookie which is creating the compiler.
     */
    protected Compiler createPreTranslateCompiler(Class type) {
        return null;
    }

    /** Creates a compiler and adds it into the job: for beans on which this JSP depends,
    *  generating servlet for this JSP, for compiling this JSP
    *  into a servlet, for other JSPs used by this JSP by &lt;jsp:include&gt; and &lt;jsp:forward&gt;.
    */
    protected void createCompiler(CompilerJob job, java.lang.Class type, /*Compiler.Depth depth,*/
                                  boolean individual) {

//System.out.println("creating JSP compiler for " + getPrimaryFile().getPackageNameExt('/','.'));
        // clean compiler, if any
        Compiler cleanCompiler = null;
        // check if we should clean first
        if (type == JspCompiler.CLEAN || type == JspCompiler.BUILD) {
            // construct clean compiler
            cleanCompiler = new CleanCompiler(this);
            if (type == JspCompiler.CLEAN) {
                job.add(cleanCompiler);
            }
        }

        // the real type
        Class jspType = individual ? JspCompiler.BUILD : type;
        if (jspType == JspCompiler.CLEAN)
            jspType = JspCompiler.BUILD;

        try {
            // save first
            SaveCookie sc = (SaveCookie)getCookie(SaveCookie.class);
            if (sc != null)
                sc.save();
            
            // save WEB-INF/web.xml
            FileObject webXml = getPrimaryFile().getFileSystem().findResource("WEB-INF/web.xml");
            if (webXml != null) {
                try {
                    DataObject webXmlDo = DataObject.find(webXml);
                    sc = (SaveCookie)webXmlDo.getCookie(SaveCookie.class);
                    if (sc != null)
                        sc.save();
                }
                catch (DataObjectNotFoundException e) {
                    // not bad, just won't be saved
                }
            }

            // compiler which creates the line mapping
            Compiler preTranslate = createPreTranslateCompiler(jspType);
	    if ( null != preTranslate) {
		job.add(preTranslate);
                if (cleanCompiler != null)
                    cleanCompiler.dependsOn(preTranslate);
	    }

            // create the compiler for this page
            JspCompiler jspCompiler = new JspCompiler(this, jspType);
            if (cleanCompiler != null)
                jspCompiler.dependsOn(cleanCompiler);
            if (preTranslate != null)
                jspCompiler.dependsOn(preTranslate);

            if (type == JspCompiler.CLEAN) {
                return;
            }
            
            // check if this compiler has already been added
            Collection compilers = job.compilers();
            for (Iterator it = compilers.iterator(); it.hasNext(); ) {
                Object other = it.next();
                if (jspCompiler.equals(other)) {
                    JspCompiler otherJspComp = (JspCompiler)other;
                    // make the cookies consistent
                    if (jspCompiler.getType () != otherJspComp.getType()) {
                        otherJspComp.setType(JspCompiler.BUILD);
                    }
                    // make the dependencies consistent - it's enough to depend on the cleancompiler
                    otherJspComp.dependsOn(jspCompiler.dependsOn());
                    return;
                }
            }

            // get the JspCompilationInfo
            JspInfo jspInfo = JspCompileUtil.analyzePage(this, JspCompileUtil.getContextPath(getPrimaryFile()), 
                true, AnalyzerParseEventListener.ERROR_REPORT_ACCURATE);
            JspCompilationInfo compInfo = new JspCompilationInfo(jspInfo, getPrimaryFile());
            
            // acquire compilers for the beans
            DataObject beans[] = compInfo.getBeans();
            CompilerJob beansJob = new CompilerJob(Compiler.DEPTH_ZERO);
            for (int i = 0; i < beans.length; i++) {
                CompilerCookie c = (CompilerCookie)beans[i].getCookie(CompilerCookie.Compile.class);
                if (c != null) {
                    c.addToJob(beansJob, Compiler.DEPTH_ZERO);
                }
                else {
                }
            }
            // now refresh the folders
            CompilerJob refreshBeansJob = new CompilerJob(Compiler.DEPTH_ZERO);
            RefreshCompiler rc;
            for (int i = 0; i < beans.length; i++) {
                rc = new RefreshCompiler(beans[i].getPrimaryFile().getParent());
                rc.dependsOn(beansJob);
                refreshBeansJob.add(rc);
            }


            // add it, add dependencies, error page info
            jspCompiler.setErrorPage(compInfo.isErrorPage());
            jspCompiler.setEncoding(compInfo.getEncoding());
            jspCompiler.setCompilationURI(JspCompileUtil.getContextPath(getPrimaryFile()));
            jspCompiler.dependsOn (refreshBeansJob);
            job.add(jspCompiler);

            // compiler which creates the line mapping
            Compiler lmc = createPostTranslateCompiler(jspType, compInfo); // new LineManglerCompiler(this, jspType);
	    if ( null != lmc) {
		lmc.dependsOn(jspCompiler);
		job.add(lmc);
	    }

            // add the compiler for the generated servlet
            JavaCompilerType.IndirectCompiler servletComp = getServletCompiler(CompilerCookie.Build.class);
            if (servletComp == null) {
                Compiler error = new ErrorCompiler(getPrimaryFile(), new Exception(
                                                       NbBundle.getBundle(JspDataObject.class).getString("CTL_BadCompilerType")), true);
                job.add(error);
            }
            else {
                // compile the servlet
                servletComp.dependsOn(jspCompiler);
                if (lmc != null) {
                    servletComp.dependsOn(lmc);
                }
                jspCompiler.servletCompiler = servletComp;
                job.add(servletComp);
                // rename the class
                Compiler ren = new RenameCompiler(jspCompiler);
                ren.dependsOn(servletComp);
                job.add(ren);
                // refresh the folder
                Compiler refr = new RefreshCompiler(JspCompileUtil.getContextOutputRoot(getPrimaryFile()));
                refr.dependsOn(ren);
                job.add(refr);
            }

            // add compilers for referenced pages - jsp:include and jsp:forward
            if (ServletSettings.OPTIONS.isCompileIncludedForwarded()) {
                JspDataObject usedPages[] = compInfo.getReferencedPages();
                for (int i = 0; i < usedPages.length; i++) {
                    JspDataObject jspdo = usedPages[i];
                    jspdo.createCompiler(job, CompilerCookie.Compile.class, /*depth,*/ false);
                }
            }

            // add compilers for error pages
            if (ServletSettings.OPTIONS.isCompileErrorPage()) {
                JspDataObject errorPage[] = compInfo.getErrorPage();
                for (int i = 0; i < errorPage.length; i++) {
                    JspDataObject jspdo = errorPage[i];
                    jspdo.createCompiler(job, CompilerCookie.Compile.class, /*depth,*/ false);
                }
            }
        }
        catch (FileStateInvalidException e) {
            //e.printStackTrace();
            Compiler error = new ErrorCompiler(getPrimaryFile(), e, true);
            job.add(error);
        }
        catch (IOException e) {
            //e.printStackTrace();
            Compiler error = new ErrorCompiler(getPrimaryFile(), e, false);
            job.add(error);
        }
        catch (Throwable e) {
            //e.printStackTrace();
            Compiler error = new ErrorCompiler(getPrimaryFile(), e, false);
            job.add(error);
        }
//printJob(job);
//System.out.println("created JSP compiler for " + getPrimaryFile().getPackageNameExt('/','.'));
    }

    public boolean isUpToDate() {
        if (getServletDataObject() == null)
            return false;
        return !getPlugin().isOutDated();
    }

    /** Return the compiler for the resulting servlet.*/
    private JavaCompilerType.IndirectCompiler getServletCompiler(Class type) throws IOException {
        CompilerType ct = CompilerSupport.getCompilerType (getPrimaryEntry());
        
        if (ct == null) {
            JspCompilerSupport ccookie = (JspCompilerSupport)getCookie(JspCompilerSupport.Compile.class);
            if (ccookie == null)
                return null;
            ct = ccookie.defaultCompilerType();
        }

        if (ct instanceof JavaCompilerType) {
            
            JavaCompilerType jct = (JavaCompilerType)ct;
            JavaCompilerType javaCompilerType=null;
            
            if (clonedCompilers == null)
                clonedCompilers = new Hashtable();
            
            // test if compilerType need to be cloned
            if (!clonedCompilers.containsKey(jct)){ //some property was changed
                //System.out.println("new clone is running");
                try {
                    javaCompilerType = (JavaCompilerType)jct.clone();
                } catch (CloneNotSupportedException ex) {
                    return jct.prepareIndirectCompiler(type, null);
                }
                clonedCompilers.put(jct,javaCompilerType);
                // setting encoding to UTF8
                javaCompilerType.setCharEncoding("UTF8");
                jct.addPropertyChangeListener(new PropertyChangeListener(){
                    public void propertyChange(PropertyChangeEvent e){
                        if (!PROP_ENCODING.equals(e.getPropertyName())) {
                            JavaCompilerType jct2 = (JavaCompilerType)e.getSource();
                            jct2.removePropertyChangeListener(this);
                            clonedCompilers.remove(jct2);                          
                        }
                    }
                });
            }else {
                javaCompilerType = (JavaCompilerType)clonedCompilers.get(jct);
            }

            // prepare indirect compiler for cloned compilerType
            JavaCompilerType.IndirectCompiler comp = javaCompilerType.prepareIndirectCompiler(type, null);
            return comp;
        }
        else {
            return null;
        }
    }

    public ServerPlugin.JspPlugin getPlugin() {
        if (jspPlugin == null) {
            JspCompileUtil.getCurrentServerPlugin().updateJspPlugin(this);
            jspPlugin = JspCompileUtil.getCurrentServerPlugin().getJspPlugin(this);
            checkRefreshServlet();
        }
        return jspPlugin;
    }

    /** Invalidates the current copy of server plugin for this JSP.
    * @param reload true if the new version of the plugin should be loaded.
    */
    public void refreshPlugin(boolean reload) {
//System.out.println("REFRESHING PLUGIN " + reload);
        jspPlugin = null;
        if (reload)
            getPlugin();
    }

    public void refreshPlugin() {
        refreshPlugin(true);
    }

    /*  public Main.ClassFileData getClassFileData() {
        return classFileData;
      }*/

    public JspServletDataObject getServletDataObject() {
        // force registering the servlet
        getPlugin();
        return servletDataObject;
    }

    /** Returns the MIME type of the content language for this page set in this file's attributes. 
     * If nothing is set, defaults to 'text/html'.
     */
    public String getContentLanguage() {
        try {
            String contentLanguage = (String)getPrimaryFile ().getAttribute (EA_CONTENT_LANGUAGE);
            if (contentLanguage != null) {
                return contentLanguage;
            }
        } catch (Exception ex) {
            // null pointer or IOException
        }
        return "text/html";
    }

    /** Sets the MIME type of the content language for this page. The language is stored 
     * in this page's filesystem attributes set in this file's attributes. 
     */
    public void setContentLanguage(String contentLanguage) throws IOException {
        getPrimaryFile ().setAttribute (EA_CONTENT_LANGUAGE, contentLanguage);
        firePropertyChange(PROP_CONTENT_LANGUAGE, null, contentLanguage);
    }

    /** Returns the MIME type of the scripting language for this page set in this file's attributes. 
     * If nothing is set, defaults to 'text/x-java'.
     */
    public String getScriptingLanguage() {
        try {
            String scriptingLanguage = (String)getPrimaryFile ().getAttribute (EA_SCRIPTING_LANGUAGE);
            if (scriptingLanguage != null) {
                return scriptingLanguage;
            }
        } catch (Exception ex) {
            // null pointer or IOException
        }
        return "text/x-java";
    }

    /** Sets the MIME type of the scripting language for this page. The language is stored 
     * in this page's filesystem attributes set in this file's attributes. 
     */
    public void setScriptingLanguage(String scriptingLanguage) throws IOException {
        getPrimaryFile ().setAttribute (EA_SCRIPTING_LANGUAGE, scriptingLanguage);
        firePropertyChange(PROP_SCRIPTING_LANGUAGE, null, scriptingLanguage);
    }
    
    public String getEncoding() {
        try {
            String encoding = (String)getPrimaryFile ().getAttribute (EA_ENCODING);
            if (encoding != null) {
                return encoding;
            }
        } catch (Exception ex) {
            // null pointer or IOException
        }
        return getDefaultEncoding();
    }
    
    public static String getDefaultEncoding() {
        String language = System.getProperty("user.language");
        if ("ja".equals(language)) {
            // we are Japanese
            if (org.openide.util.Utilities.isUnix())
                return "EUC-JP";
            else
                return "Shift_JIS";
        }
        else
            // we are English
            return "ISO-8859-1";
    }

    public void setEncoding(String encoding) throws IOException {
        getPrimaryFile ().setAttribute (EA_ENCODING, encoding);
        firePropertyChange(PROP_ENCODING, null, encoding);
    }
    
    private void printJob(CompilerJob job) {
/*        System.out.println("-- compilers --");
        java.util.Iterator compilers = job.compilers().iterator();
        for (; compilers.hasNext();) {
            org.openide.compiler.Compiler comp = (org.openide.compiler.Compiler)compilers.next();
            CompilerJob newJob = new CompilerJob(Compiler.DEPTH_ZERO);
            newJob.add(comp);
            System.out.println(comp.toString() + " upToDate=" + newJob.isUpToDate());
        }
        System.out.println("-- x --");*/
    }

    private void initialize() {
        Listener l = new Listener();
        getPrimaryFile().addFileChangeListener(l);
        addPropertyChangeListener(l);
        refreshPlugin(false);
    }


    /** Updates classFileData, servletDataObject, servletEdit */
    private void checkRefreshServlet() {

        final DataObject oldServlet = getServletDataObject();

        // dataobject
        try {
            FileObject servletFileObject = updateServletFileObject();
            if (servletFileObject != null) {
                if ((oldServlet != null) && (oldServlet.getPrimaryFile() == servletFileObject)) return; // performance
                // maybe this is necessary, but on the other hand it probably slows down the compilation:
                // servletFileObject.refresh(true);
                DataObject dObj= TopManager.getDefault().getLoaderPool().findDataObject(servletFileObject);
                if (dObj instanceof JspServletDataObject) {
                    servletDataObject = (JspServletDataObject)dObj;
                    servletDataObject.setSourceJspPage(this);
                }
            }
            else
                servletDataObject = null;
        }
        catch (IOException e) {
            servletDataObject = null;
        }

        // editor
        if ((oldServlet == null)&&(servletDataObject != null)) {
        } else {
            RequestProcessor.postRequest(
                new Runnable() {
                    public void run() {
                        updateServletEditor();
                        JspDataObject.this.firePropertyChange0(PROP_SERVLET_DATAOBJECT, oldServlet, getServletDataObject());
                    }
                }
            );
        }
    }
    
    /** JDK 1.2 compiler hack. */
    public void firePropertyChange0(String propertyName, Object oldValue, Object newValue) {
        super.firePropertyChange(propertyName, oldValue, newValue);
    }
    
    /** Returns an editor for the servlet. Architecturally, a better solution would be to attach a cookie for 
     * editing the servlet, but we choose this approach for performance reasons - this allows lazy initialization of
     * the editor (unlike the cookie). */
    public SourceCookie.Editor getServletEditor() {
        DataObject obj = getServletDataObject();
        if ((obj == null) != (servletEdit == null))
            updateServletEditor();
        return servletEdit;
    }
    
    private void updateServletEditor() {
        if (servletDataObject == null) {
            if (servletEdit != null) {
                servletEdit.close();
                servletEdit = null;
            }
        }
        else {
            if (servletEdit == null) {
                servletEdit = createServletEditor();
            }
        }
    }


    /** Gets the current fileobject of the servlet corresponding to this JSP or null if may not exist.
    * Note that the file still doesn't need to exist, even if it's not null. */
    private FileObject updateServletFileObject() throws IOException {
        String servletFileName = jspPlugin.getLocation(false).getServletFileName();
        if (servletFileName == null)
            return null;
        FileObject servletFo;
        int dotIndex = servletFileName.lastIndexOf('.');
        if (dotIndex == -1)
            servletFo = jspPlugin.getServletDirectory().getFileObject(servletFileName, "java");
        else
            servletFo = jspPlugin.getServletDirectory().getFileObject(servletFileName.substring(0, dotIndex),
                        servletFileName.substring(dotIndex + 1));
        return servletFo;
    }

    private class Listener extends FileChangeAdapter implements PropertyChangeListener {
        
        Listener() {
        }
        
        public void propertyChange(PropertyChangeEvent evt) {
            // listening on properties which could affect the server plugin
            // saving the file
            if (PROP_MODIFIED.equals(evt.getPropertyName())) {
                if ((new Boolean(false)).equals(evt.getNewValue())) {
                    refreshPlugin(false);
                }
            }
            // primary file changed or files changed
            if (PROP_PRIMARY_FILE.equals(evt.getPropertyName()) || 
                PROP_FILES.equals(evt.getPropertyName())) {
                if (evt.getOldValue() instanceof FileObject)
                    ((FileObject)evt.getOldValue()).removeFileChangeListener(this);
                if (evt.getNewValue() instanceof FileObject)
                    ((FileObject)evt.getNewValue()).addFileChangeListener(this);
                refreshPlugin(true);
            }
        }
        
        public void fileRenamed(FileRenameEvent fe) {
            refreshPlugin(true);
        }
        
    }
    
}

