/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.NoSuchElementException;
import java.text.MessageFormat;

import javax.swing.text.*;

import org.openide.cookies.*;
import org.openide.filesystems.*;
import org.openide.loaders.*;
import org.openide.nodes.*;
import org.openide.actions.OpenAction;
import org.openide.util.actions.SystemAction;
import org.openide.util.Task;
import org.openide.text.EditorSupport;
import org.openide.text.Line;
import org.openide.src.SourceElement;
import org.openide.compiler.Compiler;
import org.openide.compiler.ExternalCompiler;
import org.openide.execution.NbProcessDescriptor;
import org.openide.util.NbBundle;

import org.netbeans.modules.java.JavaDataObject;
import org.netbeans.modules.java.JavaEditor;
import org.netbeans.modules.java.JavaDataLoader;

/** Dataobject representing a servlet generated from a JSP page
*
* @author Petr Jiricka
*/
public class JspServletDataObject extends JavaDataObject {

    public static final String EA_ORIGIN_JSP_PAGE = "NetBeansAttrOriginJspPage"; // NOI18N

    private transient ServletEditorCookie servletEditor;

    /** New instance.
    * @param pf primary file object for this data object
    */
    public JspServletDataObject(FileObject pf, MultiFileLoader loader)
    throws DataObjectExistsException {
        super(pf, loader);
        init();
    }

    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        init();
    }

    public Node.Cookie getCookie(Class c) {
        if (c.isAssignableFrom(ServletEditorCookie.class)) {
            if (servletEditor == null) {
                SourceCookie.Editor ed = (SourceCookie.Editor)super.getCookie(SourceCookie.Editor.class);
                if (ed != null)
                    servletEditor = new ServletEditorCookie(ed, this);
            }
            if (servletEditor != null)
                return servletEditor;
        }
        return super.getCookie(c);
    }

    private void init() {
        changeCookies(getSourceJspPage());
    }

    private void changeCookies(DataObject jspPage) {
        CookieSet cookies = getCookieSet();

        // remove exec+debugger cookie
        Node.Cookie es = getCookie(ExecCookie.class); // same as debugger cookie
        if (es != null)
            cookies.remove(es);

        // remove compiler
        Node.Cookie cs;
        cs = getCookie(CompilerCookie.Compile.class);
        if (cs != null) cookies.remove(cs);
        cs = getCookie(CompilerCookie.Build.class);
        if (cs != null) cookies.remove(cs);
        cs = getCookie(CompilerCookie.Clean.class);
        if (cs != null) cookies.remove(cs);

        /*// remove editor and add the new one
        SourceCookie.Editor ed = (SourceCookie.Editor)getCookie(SourceCookie.Editor.class);
        if (ed != null) cookies.remove(ed);
        cookies.add(new ServletEditorCookie(ed, this));*/

        // now add the cookies corresponding to the JSP page
        if (jspPage != null) {
            DebuggerCookie jspDebug = (DebuggerCookie)jspPage.getCookie(DebuggerCookie.class);
            if (jspDebug != null)
                cookies.add(new JspServletExecSupport.JspServletExecDebugSupport(jspPage));
            else
                cookies.add(new JspServletExecSupport(jspPage));

            cookies.add(new JspServletCompilerSupport.Compile(jspPage));
            cookies.add(new JspServletCompilerSupport.Build(jspPage));
            cookies.add(new JspServletCompilerSupport.Clean(jspPage));
        }

    }

    /** Get the name of the data object.
    * Uses the name of the source JSP
    * @return the name
    */
    public String getName () {
        DataObject jsp = getSourceJspPage();
        if (jsp == null)
            return super.getName();
        int markIndex = getPrimaryFile().getName().lastIndexOf(JspServletDataLoader.JSP_MARK);
        String fileIndex = (markIndex == -1) ? "?" : getPrimaryFile().getName().substring(
                               markIndex + JspServletDataLoader.JSP_MARK.length());
        return MessageFormat.format(NbBundle.getBundle(JspServletDataObject.class).
                                    getString("LBL_ServletDisplayName"), new Object[] {fileIndex, jsp.getPrimaryFile().getName()});
    }

    /** Help context for this object.
    * @return help context
    */
    public org.openide.util.HelpCtx getHelpCtx () {
        return org.openide.util.HelpCtx.DEFAULT_HELP;
    }
    
    /** Allows subclasses to provide own JavaEditor for this servlet. */
/*    protected JavaEditor createJavaEditor() {
	return new BaseServletEditor(getPrimaryEntry());
    }*/

    /** Sets the source JSP page for this servlet */
    public void setSourceJspPage(DataObject jspPage) throws IOException {
        changeCookies(jspPage);
        getPrimaryFile().setAttribute(EA_ORIGIN_JSP_PAGE, jspPage);
    }

    /** Returns the source JSP page for this servlet */
    public DataObject getSourceJspPage() {
        return (DataObject)getPrimaryFile().getAttribute(EA_ORIGIN_JSP_PAGE);
    }

    private static class ServletEditorCookie implements SourceCookie.Editor {
        
        private SourceCookie.Editor original;
        private JspServletDataObject servlet;
        
        public ServletEditorCookie(SourceCookie.Editor original, JspServletDataObject servlet) {
            this.original = original;
            this.servlet = servlet;
        }

        private SourceCookie.Editor currentEditorCookie() {
            DataObject jsp = servlet.getSourceJspPage();
            if ((jsp != null) && (jsp instanceof JspDataObject)) {
                if (((JspDataObject)jsp).getServletDataObject() == servlet) {
                    SourceCookie.Editor newCookie = ((JspDataObject)jsp).getServletEditor();
                    if (newCookie != null)
                        return newCookie;
                }
            }
            return original;
        }

        // implementation of EditorCookie
        public Line.Set getLineSet() {
            return currentEditorCookie().getLineSet();
        }

        public void open() {
            currentEditorCookie().open();
        }

        public boolean close() {
            return currentEditorCookie().close();
        }

        public Task prepareDocument() {
            return currentEditorCookie().prepareDocument();
        }

        public javax.swing.text.StyledDocument openDocument() throws java.io.IOException {
            return currentEditorCookie().openDocument();
        }

        public javax.swing.text.StyledDocument getDocument() {
            return currentEditorCookie().getDocument();
        }

        public void saveDocument() throws java.io.IOException {
            currentEditorCookie().saveDocument();
        }

        public boolean isModified() {
            return currentEditorCookie().isModified();
        }

        public javax.swing.JEditorPane[] getOpenedPanes() {
            return currentEditorCookie().getOpenedPanes();
        }

        // implementation of SourceCookie.Editor

        /** Translate a source element to text.
         *
         * @param element an element from the source hierarchy
         * @return a text element
         */
        public javax.swing.text.Element sourceToText(org.openide.src.Element element) {
            return currentEditorCookie().sourceToText(element);
        }

        /** Translate a text element to a source element, if it is possible to do so.
         *
         * @param element a text element
         * @return the element from the source hierarchy
         * @exception NoSuchElementException if the text element doesn't match
         * any element from the source hierarchy
         */
        public org.openide.src.Element textToSource(javax.swing.text.Element element) throws NoSuchElementException {
            return currentEditorCookie().textToSource(element);
        }

        /** Find the element at the specified offset in the document.
         * @param offset The position of the element
         * @return the element at the position.
         */
        public org.openide.src.Element findElement(int offset) {
            return currentEditorCookie().findElement(offset);
        }

        /** Returns a source element describing the hierarchy
         * of the source.
         *
         * @return the element
         */
        public SourceElement getSource() {
            return currentEditorCookie().getSource();
        }

    }

}

