/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.io.Reader;
import java.io.IOException;

import org.openide.filesystems.FileObject;
import org.openide.compiler.CompilerGroup;
import org.openide.compiler.ErrorEvent;

/** Interface, which describes information specific to a particular server. This interface
* must be implemented by support plugins for various servers. The implementation is responsible for caching a copy 
* of its subinterfaces and updating it.
*
* @author Petr Jiricka
*/
public interface ServerPlugin {

    /** Creates a JspLocatorPlugin for a given JSP page, with the option to create it for the current state
    * or the future state used for the next compilation cycle.
    *   @param jspPage page for which to create the plugin
    *   @param future if true, creates the data for the next iteration of the compilation cycle, otherwise
    *      reflects the current state
    */
    public JspPlugin getJspPlugin(JspDataObject jspPage);

    /** Updates the JSP plugin for a given page. */
    public void updateJspPlugin(JspDataObject jspPage);

    /** Creates a source mangler for a given servlet, represented by a reader to its content. */
    public SourceMangler getSourceMangler(Reader servletReader) throws  IOException;

    /** Construct an error event based on an exception thrown by the compilation engine.
    *   May return null if the exception should be handled by the default way.
    */
    public ErrorEvent constructCompilerError(CompilerGroup group,
            Throwable ex, FileObject file);

    /** This inner interface takes care of locating files corresponding to a JSP, such as
    * generated servlets and class files. This interface can be only valid for one compilation 
    * cycle, i.e. when the page is recompiled, the old instance does not need to return correct values.
    */
    public interface JspPlugin {

        /** Returns a Location object related to either current or next location of the related files. */
        public Location getLocation(boolean future);

        /** Describes location of the related files. */
        public interface Location {

            /** Returns a name of the servlet with extension without package.
            * Null if does not exist.
            */
            public String     getServletFileName();

            /** Returns a full class name of the servlet or null if does not exist. */
            public String getFullClassName();

        }

        /** Returns whether the JSP page is outdated */
        public boolean    isOutDated();

        /** Returns main classes directory for target class. */
        public FileObject getClassesDirectory();

        /** Returns a class file name (without package and extension) */
        public String     getTargetClassFileName();

        /** Returns servlet directory for the servlet (including any subpackage directories). */
        public FileObject getServletDirectory();

        /** Cleans this JSP. */
        public void       cleanJsp() throws IOException;

        /** Compiles this page */
        public void       compile(boolean isErrorPage, String compilationURI, String encoding)
        throws Exception;

    }
}

/*
* Log
*  8    Jaga      1.7         4/10/00  Petr Jiricka    Fixes in execution and 
*       debugging, integrating monitor.
*  7    Jaga      1.6         3/29/00  Petr Jiricka    Clean, incremental 
*       compilation fixes
*  6    Jaga      1.5         3/29/00  Petr Jiricka    Compilation fixes and 
*       enhancements.
*  5    Jaga      1.4         3/23/00  Petr Jiricka    SourceMangler method 
*       added
*  4    Jaga      1.3         3/22/00  Petr Jiricka    Bugfixes in finding the 
*       servlet.
*  3    Jaga      1.2         3/7/00   Petr Jiricka    Fixing compilation errors
*       under 1.2.2, more interfaces cleanup.
*  2    Jaga      1.1         3/6/00   Petr Jiricka    Changes in interfaces to 
*       plugins for external servers
*  1    Jaga      1.0         3/2/00   Petr Jiricka    
* $
*/
