/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader;

import java.beans.*;
import java.util.Map;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Vector;
import java.util.zip.*;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

import org.w3c.dom.*;
import org.w3c.dom.Node; // ambiguity with org.openide.nodes.Node
import org.xml.sax.*;
import javax.xml.parsers.*;
import javax.servlet.jsp.tagext.*;

import org.openide.nodes.*;
import org.openide.loaders.DataObject;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.XMLDataObject;
import org.openide.filesystems.FileChangeAdapter;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileRenameEvent;
import org.openide.TopManager;

/** Support for getting information about DataObjects representing tag
 *  libraries (.tld and .jar files). The main goal is to provide an implementation
 *  of javax.servlet.jsp.tagext.TagLibraryInfo for the purpose of syntax coloring
 *  and code completion.
 *
 * @author  Petr Jiricka
 * @version 
 */
public class TagLibraryInfoSupport implements org.openide.nodes.Node.Cookie {

    private static final String TLD = "META-INF/taglib.tld";
    
    public static final String PROP_TAGLIBRARYINFO = "tagLibraryInfo";

    protected DataObject dobj;
    protected TagLibraryInfoInternal tagLibraryInfo;
    private transient PropertyChangeSupport pcs;
    
    /** Map of DataObject -> TagLibraryInfoSupport, serving as a cache for TagLibraryInfoSupport objects. */
    private static HashMap supportInstances;
    
    /** Creates new TagLibraryInfoSupport */
    public TagLibraryInfoSupport(DataObject dobj) {
        this.dobj = dobj;
        dobj.getPrimaryFile().addFileChangeListener(
            new FileChangeAdapter() {
            
                public void fileChanged(FileEvent fe) {
                    invalidate();
                }
                
                public void fileDeleted(FileEvent fe) {
                    invalidate();
                }
                
                public void fileRenamed(FileRenameEvent fe) {
                    invalidate();
                }
            }
        );
          
        /*dobj.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                if (DataObject.PROP_MODIFIED.equals(evt.getPropertyName()) &&
                    new Boolean(false).equals(evt.getNewValue())) {
                    invalidate();
                }
            }
        });*/
    }
    
    /** Returns tag library information represented by this tag library object
    * as TagLibraryInfo. May return null if parsing fails. */
    public TagLibraryInfo getTagLibraryInfo(String prefix, String uri) {
        if (tagLibraryInfo == null) {
            synchronized (this) {
                if (tagLibraryInfo == null)
                    tagLibraryInfo = createTagLibraryInfo();
            }
        }
        return (tagLibraryInfo == null) ? 
            null :
            new TagLibraryInfoExternal(tagLibraryInfo, prefix, uri);
    }
    
    private TagLibraryInfoInternal createTagLibraryInfo() {
        try {
            ErrorHandler eh = 
                new ErrorHandler() {
                    public void error(SAXParseException exception) throws SAXException {
                    }

                    public void fatalError(SAXParseException exception) throws SAXException {
                        throw exception;
                    }

                    public void warning(SAXParseException exception) throws SAXException {
                    }
                };
            Document tld;
            if ("jar".equals(dobj.getPrimaryFile().getExt())) {
                tld = parse(new InputSource(getInputStreamFromJar()),
                    eh, true);
            }
            else {
                tld = parse(/*dobj.getPrimaryFile().getURL()*/
                    new InputSource(dobj.getPrimaryFile().getInputStream()), eh, true);
            }
            return new TagLibraryInfoInternal(tld);
        }
        catch (IOException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) {
                e.printStackTrace();
            }
        }
        catch (SAXException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) {
                e.printStackTrace();
            }
        }
        catch (ParserConfigurationException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) {
                e.printStackTrace();
            }
        }
        catch (TagLibFormatException e) {
            if (Boolean.getBoolean("netbeans.debug.exceptions")) {
                e.printStackTrace();
            }
        }
        return null;
    }
    
    /** Treats this file as jar, gets an input stream corresponding to
     * the META-INF/taglib.tld entry of this jar file. 
     */
    private InputStream getInputStreamFromJar() throws IOException {
        ZipInputStream zin = new ZipInputStream(dobj.getPrimaryFile().getInputStream());
        ZipEntry entry;
        ByteArrayInputStream tldIs = null;
        while ((entry = zin.getNextEntry()) != null) {
            if (entry.getName().equals(TLD)) {
                /*******
                 * This hack is necessary because XML reads until the end 
                 * of an inputstream -- does not use available()
                 * -- and closes the inputstream when it can't
                 * read no more.
                 */

                // BEGIN HACK
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                int b;
                while (zin.available() != 0) {
                    b = zin.read();
                    if (b == -1)
                        break;
                    baos.write(b);
                }

                baos.close();
                tldIs = new ByteArrayInputStream(baos.toByteArray());
                // END HACK
            } else {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                int b;
                while (zin.available() != 0) {
                    b = zin.read();
                    if (b == -1)
                        break;
                    baos.write(b);
                }
                baos.close();
                //jarEntries.put(entry.getName(), baos.toByteArray());
            }
            zin.closeEntry();
        }
        return tldIs;
    }
    
    private static Document parse (InputSource source, ErrorHandler eh, boolean validate) 
    throws IOException, SAXException, ParserConfigurationException {
        return parseXMLDocJaxp(source, 
            "org/netbeans/modules/web/core/resources/web-jsptaglib_1_1.dtd",
            "-//Sun Microsystems, Inc.//DTD JSP Tag Library 1.1//EN",
            validate, eh);
        /*Parser parser;
        com.sun.xml.tree.XmlDocumentBuilder builder;

        parser = XMLDataObject.createParser (validate);
        parser.setErrorHandler (eh);
        builder = new com.sun.xml.tree.XmlDocumentBuilder();
        builder.setDisableNamespaces (true);
        builder.setParser (parser);
        parser.parse (source);

        return builder.getDocument ();*/
    }
    
    private static Document parseXMLDocJaxp(InputSource is, String dtdResource, 
    String dtdId, boolean validate, ErrorHandler eh)
    throws IOException, SAXException, ParserConfigurationException {
        Document tld;
        DocumentBuilderFactory docFactory = DocumentBuilderFactory.
            newInstance();
        docFactory.setValidating(validate);
        docFactory.setNamespaceAware(true);
        DocumentBuilder builder = docFactory.newDocumentBuilder();
        builder.setErrorHandler(eh);

        /***
         * These lines make sure that we have an internal catalog entry for
         * the taglib.dtdfile; this is so that jasper can run standalone 
         * without running out to the net to pick up the taglib.dtd file.
         */
        MyEntityResolver resolver =
            new MyEntityResolver(dtdId, dtdResource);
        builder.setEntityResolver(resolver);
        tld = builder.parse(is);
        return tld;
    }
    
    private synchronized void invalidate() {
        tagLibraryInfo = null;
        fireTagLibraryInfoChange();
    }
    
    public void addPropertyChangeListener(PropertyChangeListener pcl) {
        if (pcs == null)
            pcs = new PropertyChangeSupport(this);
        pcs.addPropertyChangeListener(pcl);
    }
    
    public void removePropertyChangeListener(PropertyChangeListener pcl) {
        if (pcs == null)
            return;
        pcs.removePropertyChangeListener(pcl);
    }
    
    void fireTagLibraryInfoChange() {
        if (pcs != null)
            pcs.firePropertyChange(PROP_TAGLIBRARYINFO, null, null);
    }
    
    /** Returns TagLibraryInfoSupport attached to DataObject dobj. If no support is attached,
    * creates a new one and attaches it to the object. If obj is a MultaDataObject, will also be
    * available as a cookie for this object.
    * @param dobj tag library object for which the support should be created
    */
    public static TagLibraryInfoSupport getTagLibraryInfoSupport(DataObject dobj) {
        if (dobj instanceof MultiDataObject) {
            MultiDataObject mobj = (MultiDataObject)dobj;
            TagLibraryInfoSupport sup = (TagLibraryInfoSupport)mobj.getCookie(TagLibraryInfoSupport.class);
            if (sup == null) {
                sup = new TagLibraryInfoSupport(mobj);
                mobj.getCookieSet().add(sup);
            }
            return sup;
        }
        else {
            if (supportInstances == null)
                supportInstances = new HashMap();
            TagLibraryInfoSupport sup = (TagLibraryInfoSupport)supportInstances.get(dobj);
            if (sup == null) {
                sup = new TagLibraryInfoSupport(dobj);
                supportInstances.put(dobj, sup);
            }
            return sup;
        }
    }
    
    /** Internal class which represents a tag library as it is regardless of the page which includes it,
    * so getPrefixString() and getURI() return null. 
    */
    private static class TagLibraryInfoInternal extends TagLibraryInfo {
        
        TagLibraryInfoInternal(Document tld) throws TagLibFormatException {
            super(null, null);
            processTLD(tld);
        }
        
        private ClassLoader getTagLibClassLoader() {
            return TopManager.getDefault().currentClassLoader();
        }

        private void processTLD(Document tld) throws TagLibFormatException {
            Vector tagVector = new Vector();
            NodeList list = tld.getElementsByTagName("taglib");

            if (list.getLength() != 1)
                throw new TagLibFormatException();

            Element elem = (Element) list.item(0);
            list = elem.getChildNodes();

            for(int i = 0; i < list.getLength(); i++) {
                Node tmp = list.item(i);
                if (! (tmp instanceof Element)) continue;
                Element e = (Element) tmp;
                String tname = e.getTagName();
                if (tname.equals("tlibversion")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        this.tlibversion = t.getData();
                } else if (tname.equals("jspversion")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        this.jspversion = t.getData();
                } else if (tname.equals("shortname")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        this.shortname = t.getData();
                } else if (tname.equals("urn")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        this.urn = t.getData();
                } else if (tname.equals("info")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        this.info = t.getData();
                } else if (tname.equals("tag"))
                    tagVector.addElement(createTagInfo(e));
                else
                    tagLibWarning(null);
            }

            this.tags = new TagInfo[tagVector.size()];
            tagVector.copyInto (this.tags);
        }

        private TagInfo createTagInfo(Element elem) throws TagLibFormatException {
            String name = null, tagclass = null, teiclass = null;
            String bodycontent = "JSP"; // Default body content is JSP
            String info = null;

            Vector attributeVector = new Vector();
            NodeList list = elem.getChildNodes();
            for(int i = 0; i < list.getLength(); i++) {
                Node tmp  =  list.item(i);
                if (! (tmp instanceof Element)) continue;
                Element e = (Element) tmp;
                String tname = e.getTagName();
                if (tname.equals("name")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        name = t.getData();
                } else if (tname.equals("tagclass")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        tagclass = t.getData();
                } else if (tname.equals("teiclass")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        teiclass = t.getData();
                } else if (tname.equals("bodycontent")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        bodycontent = t.getData();
                } else if (tname.equals("info")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        info = t.getData();
                } else if (tname.equals("attribute"))
                    attributeVector.addElement(createAttribute(e));
                else 
                    tagLibWarning(null);
            }
            TagAttributeInfo[] tagAttributeInfo 
                = new TagAttributeInfo[attributeVector.size()];
            attributeVector.copyInto (tagAttributeInfo);

            TagExtraInfo tei = null;

            if (teiclass != null && !teiclass.equals(""))
                try {
                    Class teiClass = getTagLibClassLoader().loadClass(teiclass);
                    tei = (TagExtraInfo) teiClass.newInstance();
                } catch (ClassNotFoundException cex) {
                    tagLibWarning(null);
                } catch (IllegalAccessException iae) {
                    tagLibWarning(null);
                } catch (InstantiationException ie) {
                    tagLibWarning(null);
                }

            TagInfo taginfo = new TagInfo(name, tagclass, bodycontent,
                                          info, this, 
                                          tei,
                                          tagAttributeInfo);
            return taginfo;
        }

        TagAttributeInfo createAttribute(Element elem) {
            String name = null;
            boolean required = false, rtexprvalue = false, reqTime = false;
            String type = null;

            NodeList list = elem.getChildNodes();
            for(int i = 0; i < list.getLength(); i++) {
                Node tmp  = list.item(i);
                if (! (tmp instanceof Element)) continue;
                Element e = (Element) tmp;
                String tname = e.getTagName();
                if (tname.equals("name"))  {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        name = t.getData();
                } else if (tname.equals("required"))  {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        required = Boolean.valueOf(t.getData()).booleanValue();
                } else if (tname.equals("rtexprvalue")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        rtexprvalue = Boolean.valueOf(t.getData()).booleanValue();
                } else if (tname.equals("type")) {
                    Text t = (Text) e.getFirstChild();
                    if (t != null)
                        type = t.getData();
                } else 
                    tagLibWarning(null);
            }

            //     return new TagAttributeInfo(name, required, rtexprvalue, type);
            return new TagAttributeInfo(name, required, type, rtexprvalue);
        }
        
        String getTLibVersion() {
            return tlibversion;
        }

        /** Warning about something in the tag lib descriptor. 
        * @param warningText text of the warning, may be null
        */
        private void tagLibWarning(String warningText) {
//System.out.println("warning " + warningText);
//Thread.dumpStack();        
        }
        
    } // end of inner class TagLibraryInfoInternal
    
    /** "Real class for TagLibraryInfo, delegates most of the functionality to 
    * the internal tag library info.
    */
    private static class TagLibraryInfoExternal extends TagLibraryInfo {
        
        TagLibraryInfoExternal(TagLibraryInfoInternal delegator, String prefix, String uri) {
            super(prefix, uri);
            tags = delegator.getTags();
            tlibversion = delegator.getTLibVersion();
            jspversion = delegator.getRequiredVersion();
            shortname = delegator.getShortName();
            urn = delegator.getReliableURN();
            info = delegator.getInfoString();
        }
        
        public String toString() {
            StringBuffer result = new StringBuffer();
            result.append("[TagLibraryInfoExternal\n    prefix = ");
            result.append(prefix);
            result.append("\n    uri = ");
            result.append(uri);
            result.append("\n    tlibversion = ");
            result.append(tlibversion);
            result.append("\n    jspversion = ");
            result.append(jspversion);
            result.append("\n    shortname = ");
            result.append(shortname);
            result.append("\n    urn = ");
            result.append(urn);
            result.append("\n    info = ");
            result.append(info);
            
            result.append("\n    TAGS:\n");
            TagInfo[] tags = getTags();
            for (int i = 0; i < tags.length; i++) {
                result.append("        ");
                result.append(tags[i].toString());
                result.append("\n");
            }
            result.append("]");
            return result.toString();
        }
        
    }

    
    static class MyEntityResolver implements EntityResolver {

        String dtdId;
        String dtdResource;

        public MyEntityResolver(String id, String resource) {
            this.dtdId = id;
            this.dtdResource = resource;
        }

        public InputSource resolveEntity(String publicId, String systemId)
            throws SAXException, IOException
        {
            if (publicId.equals(dtdId)) {
                InputStream input =
                    TopManager.getDefault().systemClassLoader().getResourceAsStream(dtdResource);
                InputSource isrc =
                    new InputSource(input);
                return isrc;
            }
            else {
                return null;
            }
        }
    } // end of MyEntityResolver

}

