/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader.jakarta;

import java.io.*;

import org.apache.jasper.JspCompilationContext;
import org.apache.jasper.JasperException;
import org.apache.jasper.Options;
import org.apache.jasper.compiler.JspReader;
import org.apache.jasper.compiler.ServletWriter;
import org.apache.jasper.compiler.Compiler;
import org.apache.jasper.compiler.Mangler;
import org.apache.jasper.servlet.JasperLoader;

import org.netbeans.modules.web.core.jsploader.CompilationDescriptor;
/**
 * Holds data used on a per-page compilation context that would otherwise spill
 * over to other pages being compiled.  Things like the taglib classloaders
 * and directives.
 *
 *@author Petr Jiricka, Danno Ferrin
 */
public class InternalCompilerContext implements JspCompilationContext {

    JspReader reader;
    ServletWriter writer;
    JasperLoader loader;
    boolean errPage;
    String jspFile;
    String servletClassName;
    String servletPackageName;
    String servletJavaFileName;
    String contentType;
    Options options;

    String uriBase;
    File uriRoot;

    boolean packageNameLocked;
    boolean classNameLocked;
    boolean outputInDirs;

    CompilationDescriptor cd;
    Mangler mangler;

    public InternalCompilerContext(JasperLoader newLoader,
                                   String newJspFile, CompilationDescriptor compDesc,
                                   boolean newErrPage, Options newOptions,
                                   Mangler newMangler)
    throws JasperException
    {
        loader = newLoader;
        //uriBase = newUriBase;
        //String tUriRoot = newUriRoot;
        jspFile = newJspFile;
        errPage = newErrPage;
        options = newOptions;
        cd = compDesc;
        mangler = newMangler;
    }

    /**
     * The classpath that is passed off to the Java compiler. 
     */
    public String getClassPath() {
        return "";
        //throw new UnsupportedOperationException();
    }

    public void addJar( String jar ) throws IOException  {
	loader.addJar( jar );
    }

    /**
     * Get the input reader for the JSP text. 
     */
    public JspReader getReader() {
        return reader;
    }

    /**
     * Where is the servlet being generated?
     */
    public ServletWriter getWriter() {
        return writer;
    }

    /**
     * What class loader to use for loading classes while compiling
     * this JSP? I don't think this is used right now -- akv. 
     */
    public ClassLoader getClassLoader() {
        return loader;
    }

    /**
     * Are we processing something that has been declared as an
     * errorpage? 
     */
    public boolean isErrorPage() {
        return errPage;
    }

    /**
     * What is the scratch directory we are generating code into?
     * FIXME: In some places this is called scratchDir and in some
     * other places it is called outputDir.
     */
    public String getOutputDir() {
        return options.getScratchDir().toString();
    }

    /**
     * Path of the JSP URI. Note that this is not a file name. This is
     * the context rooted URI of the JSP file. 
     */
    public String getJspFile() {
        return jspFile;
    }

    /**
     * Just the class name (does not include package name) of the
     * generated class. 
     */
    public String getServletClassName() {
        return servletClassName;
    }

    /**
     * The package name into which the servlet class is generated. 
     */
    public String getServletPackageName() {
        return servletPackageName;
    }

    /**
     * Utility method to get the full class name from the package and
     * class name. 
     */
    public String getFullClassName() {
        String pkg = getServletPackageName();
        if (pkg != null) {
            pkg += ".";
        } else {
            pkg = "";
        };
        return pkg + getServletClassName();
    }

    /**
     * Full path name of the Java file into which the servlet is being
     * generated. 
     */
    public String getServletJavaFileName() {
        if (outputInDirs) {
            return getServletPackageName().replace('.', File.separatorChar)
                   + servletJavaFileName;
        } else {
            return servletJavaFileName;
        }
    }

    /**
     * Are we keeping generated code around?
     */
    public boolean keepGenerated() {
        return options.getKeepGenerated();
    }

    /**
     * What's the content type of this JSP? Content type includes
     * content type and encoding. 
     */
    public String getContentType() {
        return contentType;
    }

    /**
     * Get hold of the Options object for this context. 
     */
    public Options getOptions() {
        return options;
    }

    public void setContentType(String contentType) {
        this.contentType = contentType;
    }

    public void setReader(JspReader reader) {
        this.reader = reader;
    }

    public void setWriter(ServletWriter writer) {
        this.writer = writer;
    }

    public void setServletClassName(String servletClassName) {
        if (classNameLocked) {
            //System.out.println("Did not change clazz to " + servletClassName);
        } else {
            this.servletClassName = servletClassName;
        };
    }

    public void setServletPackageName(String servletPackageName) {
        if (packageNameLocked) {
            //System.out.println("Did not change pkg to " + servletPackageName);
        } else {
            this.servletPackageName = servletPackageName;
        };
    }

    public void setServletJavaFileName(String servletJavaFileName) {
        this.servletJavaFileName = servletJavaFileName;
    }

    public void setErrorPage(boolean isErrPage) {
        errPage = isErrPage;
    }

    public void lockPackageName() {
        packageNameLocked = true;
    }

    public void lockClassName() {
        classNameLocked = true;
    }

    public void setOutputInDirs(boolean newValue) {
        outputInDirs = true;
    }

    public boolean isOutputInDirs() {
        return outputInDirs;
    }

    /**
     * Create a "Compiler" object based on some init param data. This
     * is not done yet. Right now we're just hardcoding the actual
     * compilers that are created. 
     */
    public Compiler createCompiler() throws JasperException {
        Compiler comp = new InternalCompiler(this);
        comp.setMangler(mangler);
        // PENDING
        //setServletPackageName(...pkgName);
        return comp;
    }


    /**
     * Get the full value of a URI relative to this compilations context
     * uses current file as the base.
     */
    public String resolveRelativeUri(String uri) {
        return cd.resolveRelativeURI(uri);
    }


    /**
     * Gets a resource as a stream, relative to the meanings of this
     * context's implementation.
     *@returns a null if the resource cannot be found or represented 
     *         as an InputStream.
     */
    public java.io.InputStream getResourceAsStream(String res) {
        return cd.getResourceAsStream(res);
    }


    /**
     * Gets the actual path of a URI relative to the context of
     * the compilation.
     */
    public String getRealPath(String path) {
        return cd.getRealPath(path);
    }

}

