/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.jsploader.jakarta;

import java.io.IOException;
import java.io.Reader;
import java.beans.*;

import org.openide.filesystems.FileObject;
import org.openide.compiler.CompilerGroup;
import org.openide.compiler.ErrorEvent;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.nodes.Node;
import org.openide.TopManager;
import org.netbeans.modules.web.core.FeatureFactory;
import org.netbeans.modules.web.core.jsploader.ServerPlugin;
import org.netbeans.modules.web.core.jsploader.JspDataObject;
import org.netbeans.modules.web.core.jsploader.JspCompileUtil;
import org.netbeans.modules.web.core.jsploader.ErrorCompiler;
import org.netbeans.modules.web.core.jsploader.CompilationDescriptor;
import org.netbeans.modules.web.core.jsploader.OptionsImpl;
import org.netbeans.modules.web.core.jsploader.SourceMangler;

import org.apache.jasper.servlet.JasperLoader;
import org.apache.jasper.compiler.Compiler;
import org.apache.jasper.compiler.Mangler;
import org.apache.jasper.compiler.ParseException;
import org.apache.jasper.compiler.JspParseEventListener;
import org.apache.jasper.JasperException;

/** Interface, which describes information specific to a particular server. This interface
 * must be implemented by support plugins for various servers
 *
 * @author Petr Jiricka
 */

public class JakartaServerPlugin implements ServerPlugin {

    // Static initialization of the Jakarta (set special comment generator)
    static {
        JspParseEventListener.setCommentGenerator(new VerboseCommentGenerator());
    }

    /** Creates a JspLocatorPlugin for a given JSP page, with the option to create it for the current state
     * or the future state used for the next compilation cycle.
     *   @param jspPage page for which to create the plugin
     *   @param future if true, creates the data for the next iteration of the compilation cycle, otherwise
     *      reflects the current state
     */
    public ServerPlugin.JspPlugin getJspPlugin(JspDataObject jspPage) {
        JakartaJspPluginCookie cookie = (JakartaJspPluginCookie)jspPage.getCookie(JakartaJspPluginCookie.class);
        if (cookie == null) {
            cookie = new JakartaJspPluginCookie(jspPage);
            jspPage.getCookieSet().add(cookie);
        }
        return cookie.getPlugin();
    }

    public void updateJspPlugin(JspDataObject jspPage) {
        JakartaJspPluginCookie cookie = (JakartaJspPluginCookie)jspPage.getCookie(JakartaJspPluginCookie.class);
        if (cookie != null) {
            cookie.invalidate();
        }
    }

    /** Creates a source mangler for a given servlet, represented by a reader to its content. */
    public SourceMangler getSourceMangler(Reader servletReader) throws IOException {
        return FeatureFactory.getFactory().getSourceMangler(servletReader); // return new SourceMapper(servletReader);
    }

    public ErrorEvent constructCompilerError(CompilerGroup group,
            Throwable ex, FileObject file) {
        return ErrorCompiler.constructJakartaCompilerError(group, ex, file);
    }

    /** Private helper cookie, which is attached to the JSP object. */
    class JakartaJspPluginCookie implements Node.Cookie {

        private JakartaJspPlugin plugin;
        private JspDataObject jspPage;

        JakartaJspPluginCookie(JspDataObject jspPage) {
            this.jspPage = jspPage;
        }

        JakartaJspPlugin getPlugin() {
            if (plugin == null) {
                plugin = new JakartaJspPlugin(jspPage);
            }
            return plugin;
        }

        void invalidate() {
            plugin = null;
        }
    }

    /** Implementation of JspLocatorPlugin  */
    public class JakartaJspPlugin implements ServerPlugin.JspPlugin {

        private JspDataObject jspPage;
        private FileObject classesDirectory, servletDirectory;
        //private String classFileName
        private String futureServletClassName, currentServletClassName;
        private String targetClassFileName;
        private String packageName;
        private boolean outDated;

        public JakartaJspPlugin(JspDataObject jspPage) {
            this.jspPage = jspPage;
            //System.out.println("constructing new jsp plugin for " + jspPage.getPrimaryFile().getPackageNameExt('/','.'));
            try {
                ManglerImpl mangler = new ManglerImpl(jspPage);
                classesDirectory = mangler.getBaseOutputDir();
                targetClassFileName = mangler.getClassFileNameSansPackageSansExt();
                servletDirectory = mangler.getJavaOutputDir();
                outDated = mangler.isOutDated();
                futureServletClassName = mangler.getClassName();
                packageName = mangler.getPackageName();
                computeCurrentServletFileName();
            }
            catch (Exception e) {
                e.printStackTrace();
                /*classesDirectory = JspCompileUtil.getContextOutputRoot(jspPage.getPrimaryFile());
                servletDirectory = classesDirectory;
                outDated = true;*/
            }
            //System.out.println(">successfully created jsp plugin");
            //System.out.println(">servlet directory " + getServletDirectory().getPackageNameExt('/','.'));
            //System.out.println(">current servlet " + getLocation(false).getServletFileName());
            //System.out.println(">future servlet " + getLocation(true).getServletFileName());
            //System.out.println(">outdated " + isOutDated());
        }

        private void computeCurrentServletFileName() {
            servletDirectory.refresh();
            FileObject futureObj = servletDirectory.getFileObject(futureServletClassName, "java");
            if (futureObj != null) {
                currentServletClassName = futureServletClassName;
                return;
            }
            int lastUnder = futureServletClassName.lastIndexOf('_');
            if (lastUnder == -1) {
                currentServletClassName = null;
                return;
            }
            try {
                int servletNr = Integer.parseInt(futureServletClassName.substring(lastUnder + 1));
                while (servletNr > 0) {
                    servletNr--;
                    currentServletClassName = futureServletClassName.substring(0, lastUnder + 1) + servletNr;
                    FileObject currentObj = servletDirectory.getFileObject(currentServletClassName, "java");
                    if (currentObj != null) {
                        return;
                    }
                }
                currentServletClassName = null;
                return;
            }
            catch (NumberFormatException e) {
                currentServletClassName = null;
                return;
            }
        }

        /** Returns a Location object related to either current or next location of the related files. */
        public ServerPlugin.JspPlugin.Location getLocation(boolean future) {
            return new JakartaJspLocation(future);
        }

        /** Describes location of the related files. */
        class JakartaJspLocation implements ServerPlugin.JspPlugin.Location {

            private boolean future;

            JakartaJspLocation(boolean future) {
                this.future = future;
            }

            /** Returns a name of the servlet with extension without package. */
            public String     getServletFileName() {
                String className = (future ? futureServletClassName : currentServletClassName);
                return (className == null) ? null : (className + ".java");
            }

            public String getFullClassName() {
                String className = (future ? futureServletClassName : currentServletClassName);
                if (className == null)
                    return null;
                return ((packageName == null) || packageName.equals("")) ? className : packageName + "." + className;
            }

        }

        /** Returns whether the JSP page is outdated */
        public boolean    isOutDated() {
            return outDated;
        }

        public FileObject getClassesDirectory() {
            return classesDirectory;
        }

        /** Returns a class file name (without package and extension) */
        public String     getTargetClassFileName() {
            return targetClassFileName;
        }

        public FileObject getServletDirectory() {
            return servletDirectory;
        }

        public void       cleanJsp() throws IOException {
            // delete the class
            FileObject deletingClass = classesDirectory.getFileObject(targetClassFileName, "class");
            if (deletingClass != null) {
                try {
                    DataObject classDo = DataObject.find(deletingClass);
                    classDo.delete();
                }
                catch (DataObjectNotFoundException e) {
                    throw new IOException(e.getClass().getName() + " : " + e.getMessage());
                }
            }
            // PENDING - inner classes

            // now delete the servlets
            String currentServlet = getLocation(false).getServletFileName();
            if (currentServlet == null) return;
            int lastDot = currentServlet.lastIndexOf('.');
            if (lastDot == -1) return;
            currentServlet = currentServlet.substring(0, lastDot);
            int lastUnder = currentServlet.lastIndexOf('_');
            if (lastUnder == -1) return;
            int servletNr = Integer.parseInt(currentServlet.substring(lastUnder + 1));
            currentServlet = currentServlet.substring(0, lastUnder + 1);
            while (servletNr >= 0) {
                String deletingServlet = currentServlet + servletNr;
                FileObject deletingObj = servletDirectory.getFileObject(deletingServlet, "java");
                if (deletingObj != null) {
                    try {
                        DataObject servletDo = DataObject.find(deletingObj);
                        servletDo.delete();
                    }
                    catch (DataObjectNotFoundException e) {
                        throw new IOException(e.getClass().getName() + " : " + e.getMessage());
                    }
                }
                servletNr--;
            }
        }

        public void       compile(boolean isErrorPage, String compilationURI, String encoding) throws Exception {
            // PENDING - encoding ?

            OptionsImpl options = new OptionsImpl(jspPage);
            
            JasperLoader loader = new JasperLoader();
            loader.setParentClassLoader(TopManager.getDefault().currentClassLoader());
            loader.setOptions(options);
            
            Mangler mangler = new ManglerImpl(jspPage);

            CompilationDescriptor cd = new CompilationDescriptor(
                                           JspCompileUtil.getContextRoot(jspPage.getPrimaryFile()).getFileSystem(),
                                           compilationURI);

            String jspResource = JspCompileUtil.getContextPath(jspPage.getPrimaryFile());
            InternalCompilerContext clctxt = new InternalCompilerContext(
                                                 loader, jspResource, cd, isErrorPage, options, mangler);

            // PENDING
            /*            if ((targetClassName != null) && (targetClassName.length() > 0)) {
                            clctxt.setServletClassName(targetClassName);
                            clctxt.lockClassName();
                        }
                        if (targetPackage != null) {
                            clctxt.setServletPackageName(targetPackage);
                            clctxt.lockPackageName();
                        }*/

            Compiler comp = clctxt.createCompiler();
            comp.compile();

        }

    }
}
