/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.syntax;


import java.util.*;
import javax.swing.JList;
import javax.swing.JLabel;
import javax.swing.text.*;
import java.awt.Color;
import java.awt.Component;

import javax.servlet.jsp.tagext.*;

import org.netbeans.editor.*;
import org.netbeans.editor.ext.CompletionQuery;

/**
 *
 * @author  Petr Jiricka, Petr Nejedly
 * @version 
 */

/* ------------------------------------------------------------------------ */
/** An interface used as a generic value returned in CompletionQuery.Result
* as an data item
*/
class CompletionItem {
    
    public static class DelegatingResult implements CompletionQuery.Result {
        
        /** The List of the ResultItem instances - the content of the result */
        private List data;

        /** The title of the result */
        private String title;

        public DelegatingResult(String title) {
            data = new ArrayList();
            this.title = title;
        }
        
        public void addResult(CompletionQuery.Result result) {
            if (result == null)
                return;
            List newData = result.getData();
            for (int i = 0; i < newData.size(); i++) {
                data.add(new DelegatingResultItem(result, i));
            }
        }

        public List getData() {
            return data;
        }

        public String getTitle() {
            return title;
        }

        public boolean substituteText(int dataIndex, boolean shift) {
            DelegatingResultItem item = (DelegatingResultItem)getData().get(dataIndex);
            return item.getResult().substituteText(item.getIndex(), shift);
        }

        public boolean substituteCommonText(int dataIndex) {
            DelegatingResultItem item = (DelegatingResultItem)getData().get(dataIndex);
            return item.getResult().substituteCommonText(item.getIndex());
        }
    }
    
    static class DelegatingResultItem implements CompletionQuery.ResultItem {
        
        CompletionQuery.Result result;
        int index;
        CompletionQuery.ResultItem resultItem;
        
        public DelegatingResultItem(CompletionQuery.Result result, int index) {
            this.result = result;
            this.index = index;
            this.resultItem = (CompletionQuery.ResultItem)result.getData().get(index);
        }
        
        public CompletionQuery.Result getResult() {
            return result;
        }
        
        public int getIndex() {
            return index;
        }
        
        public boolean substituteCommonText( JTextComponent c, int offset, int len, int subLen ) {
            return resultItem.substituteCommonText(c, offset, len, subLen);
        }

        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            return resultItem.substituteText(c, offset, len, shift);
        }

        public String getItemText() {
            return resultItem.getItemText();
        }

        public Component getPaintComponent( JList list, boolean isSelected, boolean cellHasFocus) {
            return resultItem.getPaintComponent(list, isSelected, cellHasFocus);
        }
        
        public String toString() {
            if (resultItem instanceof JspResultItem)
                return resultItem.toString();
            else
                return resultItem.getItemText();
        }
        
    }
    
    static class JspResultItem extends CompletionQuery.DefaultResultItem {
        
        public JspResultItem( String text, Color foreColor ) {
            super( text, foreColor );
        }
        
        static JLabel rubberStamp1 = new JLabel();

        static {
            rubberStamp1.setOpaque( true );
        }
        
        protected boolean substituteText( JTextComponent c, int offset, int len, String fill, int moveBack) {
            BaseDocument doc = (BaseDocument)c.getDocument();
            try {
                doc.atomicLock();
                try {
                    doc.remove( offset, len );
                    doc.insertString( offset, fill, null);
                } finally {
                    doc.atomicUnlock();
                }
                Caret caret = c.getCaret();
                int dot = caret.getDot();
                caret.setDot(dot - moveBack);
            } catch( BadLocationException exc ) {
                return false;    //not sucessfull
            }
            return true;
        }
        
        
        public Component getPaintComponent( JList list, boolean isSelected, boolean cellHasFocus ) {
            rubberStamp1.setText( " " + getPaintText() );
            if (isSelected) {
                rubberStamp1.setBackground(list.getSelectionBackground());
                rubberStamp1.setForeground(list.getSelectionForeground());
            } else {
                rubberStamp1.setBackground(list.getBackground());
                rubberStamp1.setForeground( foreColor );
            }
            return rubberStamp1;
        }
        
        public String getPaintText() {
            return getItemText();
        }

        /*public String toString() {
            return getItemText();
        }*/
        
    }
    
    /** Item representing prefix of a JSP tag. */
/*    static class Prefix extends JspResultItem {
        
        Prefix( String text ) {
            super(text, Color.black);
        }
        
        public String toString() {
            return getItemText();
        }
    }*/

    /** Item representing a JSP tag including its prefix. */
    static class PrefixTag extends JspResultItem {
        
        private TagInfo tagInfo;
        private boolean isEmpty = false;
        
        PrefixTag( String text ) {
            super(text, Color.blue);
        }
        
        PrefixTag (String prefix, TagInfo ti) {
            super(prefix + ":" + ti.getTagName (), Color.blue);
            tagInfo = ti;
            if ((tagInfo != null) &&
                (tagInfo.getBodyContent ().equalsIgnoreCase (TagInfo.BODY_CONTENT_EMPTY)))
                isEmpty = true;
        }
        
        public String getPaintText() {
            return "<" + getItemText() + (isEmpty? "/>": ">");  // NOI18N
        }

        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            String suffix = isEmpty? "/>": ">"; // NOI18N
            
            if (!getItemText ().startsWith ("/")) {  // NOI18N
                if (!shift)
                    return substituteText(c, offset, len, getItemText(), 0);  // NOI18N

                boolean hasAttrs = true;
                if (tagInfo != null) {
                    TagAttributeInfo[] tAttrs = tagInfo.getAttributes ();
                    hasAttrs = (tAttrs != null)? (tAttrs.length > 0): true;
                }
                if (hasAttrs)
                    return substituteText(c, offset, len, getItemText() + " " + suffix, suffix.length ());  // NOI18N
                else
                    return substituteText(c, offset, len, getItemText() + suffix, 0);  // NOI18N
            }
            else
                // closing tag
                return substituteText(c, offset, len, getItemText().substring (1) + ">", 0);  // NOI18N
        }
        
    }

    /** Item representing a JSP tag (without prefix) or JSP directive. */
    abstract static class TagDirective extends JspResultItem {
        
        TagDirective( String text ) {
            super(text, Color.blue);
        }
        
        /*public String toString() {
            return "<" + getItemText() + ">";
        }*/
        
        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            return substituteText(c, offset, len, getItemText() + " ", 0);
        }
        
    }

    /** Item representing a JSP tag (without prefix). */
    static class Tag extends TagDirective {
        
        Tag( String text ) {
            super(text);
        }
        
        public String getPaintText() {
            return "<" + getItemText() + ">";
        }

        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            if (!getItemText ().startsWith ("/"))   // NOI18N
                return substituteText(c, offset, len, getItemText() + " ", 0);
            else
                return substituteText(c, offset, len, getItemText().substring (1) + ">", 0);
        }
        
    }

    /** Item representing a JSP tag (without prefix). */
    static class Directive extends TagDirective {
        
        Directive( String text ) {
            super(text);
        }
        
        public String getPaintText() {
            return "@" + getItemText();
        }

    }

    /** Item representing an attribute of a  JSP tag or directive. */
    static class Attribute extends JspResultItem {
        
        Attribute( String text ) {
            super(text, Color.green.darker());
        }
        
        Attribute (TagAttributeInfo tai) {
            super(tai.getName (), (tai.isRequired ())? Color.red: Color.green.darker());
        }
        
        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            if (shift)
                return substituteText(c, offset, len, getItemText() + "=\"\"", 1);
            else
                return substituteText(c, offset, len, getItemText(), 0);
        }
        
    }

    /** Item representing a JSP attribute value. */
    static class AttributeValue extends JspResultItem {
        
        AttributeValue( String text ) {
            super(text, Color.red);
        }
        
        public boolean substituteText( JTextComponent c, int offset, int len, boolean shift ) {
            return substituteText(c, offset, len, getItemText(), 0);
        }
        
    }


}
