/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.core.syntax;

import java.util.*;
import javax.swing.text.JTextComponent;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;

import javax.servlet.jsp.tagext.TagInfo;
import javax.servlet.jsp.tagext.TagAttributeInfo;

import org.netbeans.editor.*;
import org.netbeans.editor.ext.*;
import org.netbeans.editor.ext.java.JavaSyntaxSupport;

/**
* JSP completion support finder
*
* @author Petr Nejedly
* @version 1.00
*/

public class JspCompletionQuery implements CompletionQuery {
    
    protected CompletionQuery contentQuery;
    protected CompletionQuery scriptingQuery;
    
    public JspCompletionQuery(CompletionQuery contentQuery, CompletionQuery scriptingQuery ) {
        super();
        this.contentQuery = contentQuery;
        this.scriptingQuery = scriptingQuery;
    }

    /** Perform the query on the given component. The query usually
    * gets the component's document, the caret position and searches back
    * to find the last command start. Then it inspects the text up to the caret
    * position and returns the result.
    * @param component the component to use in this query.
    * @param offset position in the component's document to which the query will
    *   be performed. Usually it's a caret position.
    * @param support syntax-support that will be used during resolving of the query.
    * @return result of the query or null if there's no result.
    */
    public CompletionQuery.Result query(JTextComponent component, int offset, SyntaxSupport support) {

//System.err.println("--------------------------------Query called------------------------");

        BaseDocument doc = (BaseDocument)component.getDocument();
        JspSyntaxSupport sup = (JspSyntaxSupport)support.get(JspSyntaxSupport.class);
        //JavaSyntaxSupport javaSup = (JavaSyntaxSupport)support.get(JavaSyntaxSupport.class);
        //System.out.println("sup = " + sup);
        //System.out.println("javaSup = " + javaSup);
        //System.out.println(sup.toString());

        try {
            // find last separator position

            //TokenItem item = sup.getTokenChain( offset, Math.min( offset+100, doc.getLength() )/*, false*/ );

            SyntaxElement elem = sup.getElementChain( offset );
//System.err.println("element = " + elem );
            
            if (elem == null)
                // this is a legal option, when I don't have anything to say just return null
                return null;
            
//System.err.println("element text = '" + elem.getImage() + "'");
            CompletionData jspData;
            switch (elem.getCompletionContext()) {
                // TAG COMPLETION
                case JspSyntaxSupport.TAG_COMPLETION_CONTEXT :
                    jspData = queryJspTag(offset, sup, 
                        (SyntaxElement.Tag)elem, doc);
//System.out.println(jspData.toString());
                    return new CompletionQuery.DefaultResult(component, 
                        "JSP completion", jspData.completionItems, 
                        offset - jspData.removeLength, jspData.removeLength);
                // ENDTAG COMPLETION
                case JspSyntaxSupport.ENDTAG_COMPLETION_CONTEXT :
                    jspData = queryJspEndTag(offset, sup, 
                        (SyntaxElement.EndTag)elem, doc);
//System.out.println(jspData.toString());
                    return new CompletionQuery.DefaultResult(component, 
                        "JSP completion", jspData.completionItems, 
                        offset - jspData.removeLength, jspData.removeLength);
                // DIRECTIVE COMPLETION
                case JspSyntaxSupport.DIRECTIVE_COMPLETION_CONTEXT :
                    jspData = queryJspDirective(offset, sup, 
                        (SyntaxElement.Directive)elem, doc);
//System.out.println(jspData.toString());
                    return new CompletionQuery.DefaultResult(component, 
                        "JSP completion", jspData.completionItems, 
                        offset - jspData.removeLength, jspData.removeLength);
                // CONTENT LANGUAGE
                case JspSyntaxSupport.CONTENTL_COMPLETION_CONTEXT :
                    // html results
                    CompletionQuery.Result contentLResult = null; 
                    contentLResult = (contentQuery == null) ? 
                        null :
                        contentQuery.query(component, offset, support);
                    // JSP results
                    jspData = queryJspTagInContent(offset, sup,
                        (SyntaxElement.ContentL)elem, doc);
//System.out.println(jspData.toString());
                    CompletionQuery.Result jspRes = new CompletionQuery.DefaultResult(component, 
                        "JSP completion", jspData.completionItems, 
                        offset - jspData.removeLength, jspData.removeLength);
                    if (jspData.completionItems.isEmpty () && (contentLResult == null))
                        return null;
                    CompletionItem.DelegatingResult result = new CompletionItem.DelegatingResult("JSP completion");
                    result.addResult(jspRes);
                    result.addResult(contentLResult);
                    // PENDING merge with JSP results
                    return result;
                // SCRIPTING LANGUAGE
                case JspSyntaxSupport.SCRIPTINGL_COMPLETION_CONTEXT :
                    CompletionQuery.Result scriptingLResult = (scriptingQuery == null) ? 
                        null :
                        scriptingQuery.query(component, offset, support);
                    return scriptingLResult;
            }

        } catch (BadLocationException e) {
            e.printStackTrace();
        }
        return null;
    }
    
    /** Gets a list of completion items for JSP tags.
    * @param offset position of caret
    * @param sup JSP syntax support
    * @param elem syntax element representing the current JSP tag
    * @return list of completion items
    */
    protected CompletionData queryJspTag(int offset, JspSyntaxSupport sup, 
        SyntaxElement.Tag elem, BaseDocument doc) throws BadLocationException {
        // find the current item
        List compItems = new ArrayList();
        int removeLength = 0;
        
        TokenItem item = sup.getItemAtOrBefore(offset);
        if (item == null) {
            return new CompletionData(compItems, 0);
        }
        
        TokenID id = item.getTokenID();
        String tokenPart = item.getImage().substring(0, offset - item.getOffset());
        String token = item.getImage().trim();
//System.out.println("tokenPart = " + tokenPart);

        // SYMBOL
        if (id == JspTagTokenContext.SYMBOL) {
            if (tokenPart.equals("<")) {
                // just after the beginning of the tag
                removeLength = 0;
                addTagPrefixItems(sup, compItems, sup.getTagPrefixes(""));
            }
            if (tokenPart.endsWith("\"")) {
                // try an attribute value
                String attrName = findAttributeForValue(sup, item);
                if (attrName != null) {
                    AttributeValueSupport attSup = 
                        AttributeValueSupport.getSupport(true, elem.getName(), attrName);
                    if (attSup != null) {
                        addAttributeValues(compItems, attSup.getCompletionItems(offset, sup, elem, ""));
                        removeLength = 0;
                    }
                }
            }
        }
        
        // TAG
        if (id == JspTagTokenContext.TAG) {
            // inside a JSP tag name
            if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                // blank character - do attribute completion
                removeLength = 0;
                addAttributeItems(sup, compItems, elem, sup.getTagAttributes(elem.getName(), ""), null);
            }
            else {
                int colonIndex = tokenPart.indexOf(":");
                if (colonIndex == -1) {
                    removeLength = tokenPart.length();
                    addTagPrefixItems(sup, compItems, sup.getTagPrefixes(tokenPart));
                }
                else {
                    String prefix = tokenPart.substring(0, colonIndex);
                    removeLength = tokenPart.length();
                    addTagPrefixItems(sup, compItems, prefix, sup.getTags(tokenPart));
                }
            }
        }
        
        // ATTRIBUTE
        if (id == JspTagTokenContext.ATTRIBUTE) {
            // inside or after an attribute
            if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                // blank character - do attribute completion
                removeLength = 0;
                addAttributeItems(sup, compItems, elem, sup.getTagAttributes(elem.getName(), ""), null);
            }
            else {
                removeLength = tokenPart.length();
                addAttributeItems(sup, compItems, elem, sup.getTagAttributes(elem.getName(), tokenPart), token);
            }
        }
        
        // ATTRIBUTE VALUE
        if (id == JspTagTokenContext.ATTR_VALUE) {
            // inside or after an attribute
            String valuePart = tokenPart;
            item = item.getPrevious();
            while ((item != null) && (item.getTokenID() == JspTagTokenContext.ATTR_VALUE)) {
                valuePart = item.getImage() + valuePart;
                item = item.getPrevious();
            }
            // get rid of the first quote
            valuePart = valuePart.substring(1);
            removeLength = valuePart.length ();
            String attrName = findAttributeForValue(sup, item);
            if (attrName != null) {
                AttributeValueSupport attSup = 
                    AttributeValueSupport.getSupport(true, elem.getName(), attrName);
                if (attSup != null) {
                    addAttributeValues(compItems, attSup.getCompletionItems(offset, sup, elem, valuePart));
                }
            }
            
            /*if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                // blank character - do attribute completion
                removeLength = 0;
                addAttributeItems(sup, compItems, elem, sup.getTagAttributes(elem.getName(), ""), null);
            }*/
        }
        
        return new CompletionData(compItems, removeLength);
    }
    
    /** Gets a list of completion items for JSP tags.
    * @param offset position of caret
    * @param sup JSP syntax support
    * @param elem syntax element representing the current JSP tag
    * @return list of completion items
    */
    protected CompletionData queryJspEndTag(int offset, JspSyntaxSupport sup, 
        SyntaxElement.EndTag elem, BaseDocument doc) throws BadLocationException {
        // find the current item
        List compItems = new ArrayList();
        int removeLength = 0;
        
        TokenItem item = sup.getItemAtOrBefore(offset);
        if (item == null) {
            return new CompletionData(compItems, 0);
        }
        
        TokenID id = item.getTokenID();
        String tokenPart = item.getImage().substring(0, offset - item.getOffset());
//System.out.println("tokenPart = " + tokenPart);

        removeLength = tokenPart.length();
        return new CompletionData (sup.getPossibleEndTags (offset, tokenPart), removeLength);
    }
    
    /** Gets a list of completion items for JSP directives.
    * @param offset position of caret
    * @param sup JSP syntax support
    * @param elem syntax element representing the current JSP tag
    * @return list of completion items
    */
    protected CompletionData queryJspDirective(int offset, JspSyntaxSupport sup, 
        SyntaxElement.Directive elem, BaseDocument doc) throws BadLocationException {
        // find the current item
        List compItems = new ArrayList();
        int removeLength = 0;
        
        TokenItem item = sup.getItemAtOrBefore(offset);
        if (item == null) {
            return new CompletionData(compItems, 0);
        }
        
        TokenID id = item.getTokenID();
        String tokenPart = item.getImage().substring(0, offset - item.getOffset());
        String token = item.getImage().trim();
//System.out.println("tokenPart = " + tokenPart + ", id="+id.toString ());

        // SYMBOL
        if (id == JspTagTokenContext.SYMBOL) {
            if (tokenPart.equals("<%@")) {
                // just after the beginning of the directive
                removeLength = 0;
                addDirectiveItems(sup, compItems, sup.getDirectives(""));
            }
            if (tokenPart.endsWith("\"")) {
                // try an attribute value
                String attrName = findAttributeForValue(sup, item);
                if (attrName != null) {
                    AttributeValueSupport attSup = 
                        AttributeValueSupport.getSupport(false, elem.getName(), attrName);
                    if (attSup != null) {
                        addAttributeValues(compItems, attSup.getCompletionItems(offset, sup, elem, ""));
                        removeLength = 0;
                    }
                }
            }
        }
        
        // DIRECTIVE
        if (id == JspTagTokenContext.TAG) {
            // inside a JSP directive name
            if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                TokenItem prevItem = item.getPrevious ();
                TokenID prevId = prevItem.getTokenID ();
                String prevToken = prevItem.getImage().trim();
                if (prevId == JspTagTokenContext.TAG
                ||  prevId == JspTagTokenContext.ATTR_VALUE) {
                    // blank character - do attribute completion
                    removeLength = 0;
                    addAttributeItems(sup, compItems, elem, sup.getDirectiveAttributes(elem.getName(), ""), null);
                }
                else if (prevId == JspTagTokenContext.SYMBOL && prevToken.equals("<%@")) {
                    // just after the beginning of the directive
                    removeLength = 0;
                    addDirectiveItems(sup, compItems, sup.getDirectives(""));
                }
            }
            else {
                removeLength = tokenPart.length();
                addDirectiveItems(sup, compItems, sup.getDirectives(tokenPart));
            }
        }
        
        // ATTRIBUTE
        if (id == JspTagTokenContext.ATTRIBUTE) {
            // inside or after an attribute
            if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                // blank character - do attribute completion
                removeLength = 0;
                addAttributeItems(sup, compItems, elem, sup.getDirectiveAttributes(elem.getName(), ""), null);
            }
            else {
                removeLength = tokenPart.length();
                addAttributeItems(sup, compItems, elem, sup.getDirectiveAttributes(elem.getName(), tokenPart), token);
            }
        }
        
        // ATTRIBUTE VALUE
        if (id == JspTagTokenContext.ATTR_VALUE) {
            // inside or after an attribute
            String valuePart = tokenPart;
            item = item.getPrevious();
            while ((item != null) && (item.getTokenID() == JspTagTokenContext.ATTR_VALUE)) {
                valuePart = item.getImage() + valuePart;
                item = item.getPrevious();
            }
            // get rid of the first quote
            valuePart = valuePart.substring(1);
            removeLength = valuePart.length ();
            String attrName = findAttributeForValue(sup, item);
            if (attrName != null) {
                AttributeValueSupport attSup = 
                    AttributeValueSupport.getSupport(false, elem.getName(), attrName);
                if (attSup != null) {
                    addAttributeValues(compItems, attSup.getCompletionItems(offset, sup, elem, valuePart));
                }
            }
            
            /*if (isBlank(tokenPart.charAt(tokenPart.length() - 1))) {
                // blank character - do attribute completion
                removeLength = 0;
                addAttributeItems(sup, compItems, elem, sup.getDirectiveAttributes(elem.getName(), ""), null);
            }*/
        }
        
        return new CompletionData(compItems, removeLength);
    }
    
    
    protected CompletionData queryJspTagInContent(int offset, JspSyntaxSupport sup, 
        SyntaxElement.ContentL elem, BaseDocument doc) throws BadLocationException {
        // find the current item
        List compItems = new ArrayList();
        int removeLength = 0;
        
        TokenItem item = sup.getItemAtOrBefore(offset);
        if (item == null) {
            return new CompletionData(compItems, 0);
        }
        
        String tokenPart = item.getImage().substring(0, offset - item.getOffset());
//System.out.println("tokenPart = " + tokenPart);
        int ltIndex = tokenPart.lastIndexOf('<');
        if (ltIndex != -1) {
            tokenPart = tokenPart.substring(ltIndex + 1);
        }
        while (ltIndex == -1) {
            item = item.getPrevious();
            if (item == null) {
                return new CompletionData(compItems, 0);
            }
            String newImage = item.getImage();
            ltIndex = newImage.lastIndexOf('<');
            if (ltIndex != -1)
                tokenPart = newImage.substring(ltIndex + 1) + tokenPart;
            else {
                tokenPart = newImage + tokenPart;
            }
            if (tokenPart.length() > 20) {
                return new CompletionData(compItems, 0);
            }
        }
        // we found ltIndex, tokenPart is either the part of the token we are looking for
        // or '/' + what we are looking for
        if (tokenPart.startsWith("/")) {
            tokenPart = tokenPart.substring(1);
//System.out.println("found tokenPart = " + tokenPart);
            compItems = sup.getPossibleEndTags (offset, tokenPart);
        }
        else {
            addTagPrefixItems(sup, compItems, sup.getTagPrefixes(tokenPart));
        }
        removeLength = tokenPart.length();
        return new CompletionData(compItems, removeLength);
    }
    
    private boolean isBlank(char c) {
        return c == ' ';
    }
    
    /** Finds an attribute name, assuming that the item is either
     * SYMBOL after the attribute name or ATTR_VALUE after this attribute name. 
     * May return null if nothing found.
     */
    protected String findAttributeForValue(JspSyntaxSupport sup, TokenItem item) {
        // get before any ATTR_VALUE
        while ((item != null) && (item.getTokenID() == JspTagTokenContext.ATTR_VALUE))
            item = item.getPrevious();
        // now collect the symbols
        String symbols = "";
        while ((item != null) && (item.getTokenID() == JspTagTokenContext.SYMBOL)) {
            symbols = item.getImage() + symbols;
            item = item.getPrevious();
        }
        // two quotes at the end are not allowed
        if (!sup.isValueBeginning(symbols)) 
            return null;
        String attributeName = "";
        while ((item != null) && (item.getTokenID() == JspTagTokenContext.ATTRIBUTE)) {
            attributeName = item.getImage() + attributeName;
            item = item.getPrevious();
        }
        if (attributeName.trim().length() > 0)
            return attributeName.trim();
        return null;
    }
    
    /** Adds to the list of items <code>compItemList</code> new TagPrefix items with prefix 
     * <code>prefix</code> for list of tag names <code>tagStringItems</code>. 
     */
    private void addTagPrefixItems(JspSyntaxSupport sup, List compItemList, String prefix, List tagStringItems) {
        for (int i = 0; i < tagStringItems.size(); i++) {
            Object item = tagStringItems.get(i);
            if (item instanceof TagInfo)
                compItemList.add(new CompletionItem.PrefixTag(prefix , (TagInfo)item));
            else
                compItemList.add(new CompletionItem.PrefixTag(prefix + ":" + (String)item));
        }
    }

    /** Adds to the list of items <code>compItemList</code> new TagPrefix items for prefix list
     * <code>prefixStringItems</code>, followed by all possible tags for the given prefixes. 
     */
    private void addTagPrefixItems(JspSyntaxSupport sup, List compItemList, List prefixStringItems) {
        for (int i = 0; i < prefixStringItems.size(); i++) {
            String prefix = (String)prefixStringItems.get(i);
            // now get tags for this prefix
            List tags = sup.getTags(prefix, "");
            for (int j = 0; j < tags.size(); j++) {
                Object item = tags.get(j);
                if (item instanceof TagInfo)
                    compItemList.add(new CompletionItem.PrefixTag(prefix , (TagInfo)item));
                else
                    compItemList.add(new CompletionItem.PrefixTag(prefix + ":" + (String)item));
            }
        }
    }

    /** Adds to the list of items <code>compItemList</code> new TagPrefix items with prefix 
     * <code>prefix</code> for list of tag names <code>tagStringItems</code>. 
     */
    private void addDirectiveItems(JspSyntaxSupport sup, List compItemList, List directiveStringItems) {
        for (int i = 0; i < directiveStringItems.size(); i++) 
            compItemList.add(new CompletionItem.Directive( 
                (String)directiveStringItems.get(i)));
    }
    
    /** Adds to the list of items <code>compItemList</code> new Attribute items. 
     * Only those which are not already in tagDir 
     * @param sup the syntax support
     * @param compItemList list to add to
     * @param tagDir tag or directive element
     * @param attributeItems list of strings containing suitable values (String or TagAttributeInfo)
     * @param currentAttr current attribute, may be null
     */
    private void addAttributeItems(JspSyntaxSupport sup, List compItemList, 
        SyntaxElement.TagDirective tagDir, List attributeItems, String currentAttr) {
        for (int i = 0; i < attributeItems.size(); i++) {
            Object item = attributeItems.get(i);
            String attr;
            if (item instanceof TagAttributeInfo)
                attr = ((TagAttributeInfo)item).getName ();
            else
                attr = (String)item;
            if (!tagDir.getAttributes().keySet().contains(attr) ||
                attr.equals(currentAttr)) {
                if (item instanceof TagAttributeInfo)
                    compItemList.add(new CompletionItem.Attribute((TagAttributeInfo)item));
                else
                    compItemList.add(new CompletionItem.Attribute((String)item));
            }
        }
    }
    
    private void addAttributeValues(List compItemList, List newCompList) {
        compItemList.addAll(newCompList);
    }
    
    /** Class which encapsulates a list of completion items and length of 
     * the part which should be replaced. */
    public static class CompletionData {
        
        public List completionItems;
        public int removeLength;
        
        public CompletionData(List items, int length) {
            this.completionItems = items;
            this.removeLength = length;
        }
        
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("------ completion items, remove " + removeLength + " : ----------\n");
            for (int i = 0; i < completionItems.size(); i++) {
                CompletionQuery.ResultItem item = 
                    (CompletionQuery.DefaultResultItem)completionItems.get(i);
                sb.append(item.getItemText());
                sb.append("\n");
            }
            return sb.toString();
        }

        
    }

}

/*
 * $Log: 
 *  2    Jaga      1.1         3/16/00  Miloslav Metelka patch for 1.2 compiler
 *  1    Jaga      1.0         3/16/00  Miloslav Metelka 
 * $
 */



