/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.templates.model;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import javax.swing.JFileChooser;
import java.beans.PropertyVetoException;
import java.util.ResourceBundle;
import java.net.URL;
import java.lang.ClassLoader;
import java.util.jar.*;

import org.openide.util.NbBundle;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.LocalFileSystem;
import org.openide.execution.NbClassPath;
import org.openide.TopManager;

// ------------------------------------

/**
 * DWBeanTemplate class provides functionality to:
 *   -- Access PrintWriter objects for each region of template.
 *   -- Save generated code written to PrintWriters into 
 *      right regions of template.
 *   -- Receive entire contents of a template after code generation.
 *
 * A typical scenario for using this class:
 * (Refer to the main() method at the end of this file for an example.)
 *
 *   -- Get a DWBeanTemplate object using one of the static get methods:
 *      DWBeanTemplate.getDefaultTemplate() loads default
 *      template for input and output pages.
 *      DWBeanTemplate.getErrorPageTemplate() loads default
 *      template for error pages. 
 *      DWBeanTemplate.getCustomTemplate(arg) loads a custom
 *      template using the argument supplied.
 *   -- If a DWBeanTemplate object is successfully initialized,
 *      get individual PrintWriter objects for each region.
 *   -- Generate code and write it to the appropriate PrintWriter.
 *   -- If needed, reset PrintWriter objects to start fresh.  Code
 *      written to PrintWriter objects will be lost after a call
 *      to resetTemplateWriters() method.
 *   -- Call getJSPPageContents() to receive a String
 *      containing the original contents of the template plus
 *      any generated code inserted into appropriate regions.
 */

// ------------------------------------

public class DWBeanTemplate {

    /**
     * Used to denote status of accessing and validating templates.
     */
    public final static int TEMPLATE_PARSED          = 0;
    public final static int ERROR_ACCESSING_TEMPLATE = 1;

    // -------------------------
    // Private Member Variables

    private final static String i18nBundle =          // NOI18N
	"org.netbeans.modules.web.templates.resources.Bundle";
    private static ResourceBundle resBundle = NbBundle.getBundle(i18nBundle);
    private static String PAGE_DIRECTIVE_END =
	resBundle.getString("EI_PageDirectiveEndTag");
    private static String USE_BEAN_END =
	resBundle.getString("EI_UseBeanEndTag");
    private static String PROP_INIT_END =
	resBundle.getString("EI_PropertyInitEndTag");
    private static String METHOD_CALL_END =
	resBundle.getString("EI_MethodCallEndTag");
    private static String DYNAMIC_FORM_END =
	resBundle.getString("EI_DynamicFormEndTag");
    private static String CLOSING_COMMENT_TAG =
	resBundle.getString("EI_ClosingCommentTag");
    private static String END_BODY_TAG = resBundle.getString("EI_EndBodyTag"); 

    private BufferedReader templateContents;
    private DWBeanTemplateParser templateParser; 

    // These Writers are output streams to which
    // users of this class write generated code.
    //
    private CharArrayWriter pageDirectiveWriter;
    private CharArrayWriter useBeanWriter;
    private CharArrayWriter propertyInitWriter;
    private CharArrayWriter methodCallWriter;
    private CharArrayWriter dynamicFormWriter;

    // These buffers contain the contents of template
    // which is initially read and validated.
    //
    private StringBuffer pageDirectiveBuffer;
    private StringBuffer useBeanBuffer;
    private StringBuffer propertyInitBuffer;
    private StringBuffer methodCallBuffer;
    private StringBuffer dynamicFormBuffer;
    private StringBuffer beginBuffer;
    private StringBuffer initRegionBuffer;
    private StringBuffer bodyBuffer;
    private StringBuffer dynamicRegionBuffer;
    private StringBuffer endBuffer;

    // The following member variables are used
    // to determine which default template must be used.
    //
    private static final int DEFAULT_TEMPLATE    = 0;
    private static final int ERROR_PAGE_TEMPLATE = 1;
    private String defaultTemplateName;

    // The following flags are set by the parser when
    // an editable DW region is found in the template.
    // If these flags are not set, DW editable regions
    // aren't added to the JSP file.
    //
    private boolean foundDWEditableInit;
    private boolean foundDWEditableDynamicForm;

    // This member designates whether a Dreamweaver object
    // has been successfully initialized or not.
    // Possible values are TEMPLATE_PARSED and
    // ERROR_ACCESSING_TEMPLATE
    //
    private int initStatus;

    // -------------------------

    /**
     * Consturctors are private.
     * Use one of the public static methods to create and
     * initialize a DWBeanTemplate object. 
     */

    /**
     *  The default constructor initializes a DWBeanTemplate
     *  object to read from and access the default template.
     *  This default template is a system resource available
     *  to the Dreamweaver Integration Module in Forte4j.
     *  Upon successful installation of this module, this system
     *  resource becomes available and accessible.
     */
    private DWBeanTemplate() {
        setDefaultTemplate(DEFAULT_TEMPLATE);
        loadDefaultTemplate();
        processTemplate();
    }

    private DWBeanTemplate(int templateType) {
        setDefaultTemplate(templateType);
        loadDefaultTemplate();
        processTemplate();
    }

    /**
     * Creates a DWBeanTemplate object,
     * initializes the PrintWriter objects, and
     * sets templateContents.
     *
     * @param
     *       File temp - The File object representing the
     *       template file.
     */
    private DWBeanTemplate(File template) {
        initializeWriters();
        setTemplateContents(template);
    }

    /**
     * Creates a DWBeanTemplate object,
     * initializes the PrintWriter objects, and
     * sets templateContents.
     *
     * @param
     *       String tempParent - Directory path where template exists.
     *       String tempName - File name of the template.
     */
    private DWBeanTemplate(String tempParent, String tempName) {
        initializeWriters();
        File template = new File(tempParent, tempName);
        setTemplateContents(template);
    }

    /**
     * Creates a DWBeanTemplate object and
     * initializes the PrintWriter objects.
     *
     * @param
     *       FileObject fileObj - FileObject representing a template.
     */
    private DWBeanTemplate(FileObject fileObj) {
        initializeWriters();

        try {
	    templateContents = new BufferedReader(
						  new InputStreamReader(fileObj.getInputStream()));
	    processTemplate();
        } catch (IOException e) {
	    initStatus = ERROR_ACCESSING_TEMPLATE;
        }
    }

    // -------------------------
    private void setTemplateContents(File template) {
	if (template.exists()) {
	    try {
		templateContents = new BufferedReader(new FileReader(template));
		processTemplate();
	    } catch (IOException e) {
		initStatus = ERROR_ACCESSING_TEMPLATE;
	    }
	} else {
	    initStatus = ERROR_ACCESSING_TEMPLATE;
	}
    }

    // -------------------------

    /**
     *  Returns a DWBeanTemplate object preloaded to use
     *  the default template for input and output pages.
     */
    public static DWBeanTemplate getDefaultTemplate() throws IOException {

	DWBeanTemplate dwObject = new DWBeanTemplate(DEFAULT_TEMPLATE);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException("Unable to initialize DWBeanTemplate object!");
	}
	return dwObject;
    }

    /**
     * Returns a DWBeanTemplate object preloaded to use
     * the default template for error pages.
     */
    public static DWBeanTemplate getErrorPageTemplate() throws IOException {

	DWBeanTemplate dwObject = new DWBeanTemplate(ERROR_PAGE_TEMPLATE);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException("Unable to initialize DWBeanTemplate object!");
	}
	return dwObject;
    }

    /**
     * Returns a DWBeanTemplate object preloaded to use
     * the custom template supplied by the argument.
     */
    public static DWBeanTemplate getCustomTemplate(File temp) throws IOException {

	DWBeanTemplate dwObject = new DWBeanTemplate(temp);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException("Unable to initialize DWBeanTemplate object!");
	}
	return dwObject;
    }

    /**
     * Returns a DWBeanTemplate object preloaded to use
     * the custom template supplied by the argument.
     */
    public static DWBeanTemplate getCustomTemplate(String tempParent,
						   String tempName)
	throws IOException {

	DWBeanTemplate dwObject = new DWBeanTemplate(tempParent,tempName);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException("Unable to initialize DWBeanTemplate object!");
	}
	return dwObject;
    }

    /**
     * Returns a DWBeanTemplate object preloaded to use
     * the custom template supplied by the argument.
     */
    public static DWBeanTemplate getCustomTemplate(FileObject fileObj)
	throws IOException {

	DWBeanTemplate dwObject = new DWBeanTemplate(fileObj);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException("Unable to initialize DWBeanTemplate object!");
	}
	return dwObject;
    }

    // -------------------------
    private void setDefaultTemplate(int templateType) {

	// A switch statement is being used in order to support
	// addition of other default templates easily.
	// 
	switch (templateType) {

	case DEFAULT_TEMPLATE:
	    defaultTemplateName = resBundle.getString("EI_DWDefaultTemplate");
	    break;
	case ERROR_PAGE_TEMPLATE:
	    defaultTemplateName = resBundle.getString("EI_DWErrorPageTemplate");
	    break;
	default:
	    defaultTemplateName = resBundle.getString("EI_DWDefaultTemplate");
	    break;
	}
    }

    // -------------------------
    //  This method accesses the default template as a system
    //  resource.  It then initializes templateContents with
    //  the contents of this default template.
    //
    private void loadDefaultTemplate() {

	initializeWriters();
	JarInputStream jarIstream;
	boolean usingDefaultTemplate = false;

        try {
            FileSystem defaultFs = TopManager.getDefault().getRepository().getDefaultFileSystem();
            FileObject fo = defaultFs.findResource (defaultTemplateName);
            if (fo != null) {
                templateContents = new BufferedReader (new InputStreamReader (fo.getInputStream ()));
            }

        } catch (IOException e) {
            org.openide.TopManager.getDefault ().notifyException (e);
            initStatus = ERROR_ACCESSING_TEMPLATE;
        }
    }

    // -------------------------
    private void initializeWriters() {
	pageDirectiveWriter = new CharArrayWriter();
	useBeanWriter       = new CharArrayWriter();
	propertyInitWriter  = new CharArrayWriter();
	methodCallWriter    = new CharArrayWriter();
	dynamicFormWriter   = new CharArrayWriter();
    }

    private int getInitStatus() {
	return initStatus;
    }

    // -------------------------

    /**
     * Clears the contents of the Writer objects used by
     * PrintWriter objects.  After this method is called,
     * the code stored in each Writer is lost.
     */
    public void resetTemplateWriters() {
	initializeWriters();
    }

    /**
     *  Creates and returns a PrintWriter object for
     *  PageDirective region.
     */
    public PrintWriter getPageDirectiveWriter() {
	return new PrintWriter(pageDirectiveWriter,true);
    }

    /**
     *  Creates and returns a PrintWriter object for
     *  UseBean region.
     */
    public PrintWriter getUseBeanWriter() {
	return new PrintWriter(useBeanWriter,true);
    }

    /**
     *  Creates and returns a PrintWriter object for
     *  PropertyInitialization region.
     */
    public PrintWriter getPropertyInitWriter() {
	return new PrintWriter(propertyInitWriter,true);
    }

    /**
     *  Creates and returns a PrintWriter object for
     *  MethodCall region.
     */
    public PrintWriter getMethodCallWriter() {
	return new PrintWriter(methodCallWriter,true);
    }

    /**
     *  Creates and returns a PrintWriter object for
     *  DynamicForm region.
     */
    public PrintWriter getDynamicFormWriter() {
	return new PrintWriter(dynamicFormWriter,true);
    }

    // -------------------------

    /**
     * Saves contents of all buffers in the required order. 
     * Regions are appended one after the other in order.
     * At the end, contents of endBuf is added.
     */ 
    public StringBuffer getJSPPageContents() {

        // Insert generated code in appropriate buffers.
        insertPageDirectiveCode();
        insertUseBeanCode();
        insertPropertyInitCode();
        insertMethodCallCode();
        insertDynamicFormCode();

        StringBuffer jspPageBuffer = new StringBuffer();
        int bufferOffset;

        jspPageBuffer.append(beginBuffer.toString());

        if (initRegionBuffer.length() > 0) {

            // Custom regions must be appended after editable region tag
            bufferOffset = findOffsetAfterEditableTag(initRegionBuffer);
            initRegionBuffer.insert(bufferOffset, methodCallBuffer.toString());
            initRegionBuffer.insert(bufferOffset, propertyInitBuffer.toString());
            initRegionBuffer.insert(bufferOffset, useBeanBuffer.toString());
            initRegionBuffer.insert(bufferOffset, pageDirectiveBuffer.toString());
            jspPageBuffer.append(initRegionBuffer.toString());

        } else {

            jspPageBuffer.append(pageDirectiveBuffer.toString());
            jspPageBuffer.append(useBeanBuffer.toString());
            jspPageBuffer.append(propertyInitBuffer.toString());
            jspPageBuffer.append(methodCallBuffer.toString());
        }

        if (dynamicRegionBuffer.length() > 0) {

            jspPageBuffer.append(bodyBuffer.toString());

            // Custom regions must be appended after editable region tag
            bufferOffset = findOffsetAfterEditableTag(dynamicRegionBuffer);
            dynamicRegionBuffer.insert(bufferOffset, dynamicFormBuffer.toString());
            jspPageBuffer.append(dynamicRegionBuffer.toString());

        } else {

            // Custom regions must be inserted after body tag
            bufferOffset = findOffsetBeforeEndBodyTag(bodyBuffer);
            bodyBuffer.insert(bufferOffset, dynamicFormBuffer.toString());
            jspPageBuffer.append(bodyBuffer.toString());
        }

        if (endBuffer.length() > 0) {
            jspPageBuffer.append(endBuffer.toString());
        }
        return jspPageBuffer;
    }

    // -------------------------

    // Inserts generated code before PAGE_DIRECTIVE_END 
    private void insertPageDirectiveCode() {
	int bufferLength = pageDirectiveBuffer.length(); 
	int endTagLength = PAGE_DIRECTIVE_END.length();
	int bufferOffset = bufferLength - endTagLength - 1;
	if ( (bufferLength > 0) && (bufferLength >= endTagLength) ) {
	    pageDirectiveBuffer.insert(bufferOffset, pageDirectiveWriter.toString());
	}
    }

    // Inserts generated code before USE_BEAN_END
    private void insertUseBeanCode() {
	int bufferLength = useBeanBuffer.length();
	int endTagLength = USE_BEAN_END.length();
	int bufferOffset = bufferLength - endTagLength - 1;
	if ( (bufferLength > 0) && (bufferLength >= endTagLength) ) {
	    useBeanBuffer.insert(bufferOffset, useBeanWriter.toString());
	}
    }

    // Inserts generated code before PROP_INIT_END
    private void insertPropertyInitCode() {
	int bufferLength = propertyInitBuffer.length();
	int endTagLength = PROP_INIT_END.length();
	int bufferOffset = bufferLength - endTagLength - 1;
	if ( (bufferLength > 0) && (bufferLength >= endTagLength) ) {
	    propertyInitBuffer.insert(bufferOffset, propertyInitWriter.toString());
	}
    }

    // Inserts generated code before METHOD_CALL_END
    private void insertMethodCallCode() {
	int bufferLength = methodCallBuffer.length();
	int endTagLength = METHOD_CALL_END.length();
	int bufferOffset = bufferLength - endTagLength - 1;
	if ( (bufferLength > 0) && (bufferLength >= endTagLength) ) {
	    methodCallBuffer.insert(bufferOffset, methodCallWriter.toString());
	}
    }

    // Inserts generated code before DYNAMIC_FORM_END
    private void insertDynamicFormCode() {
	int bufferLength = dynamicFormBuffer.length();
	int endTagLength = DYNAMIC_FORM_END.length();
	int bufferOffset = bufferLength - endTagLength - 1;
	if ( (bufferLength > 0) && (bufferLength >= endTagLength) ) {
	    dynamicFormBuffer.insert(bufferOffset, dynamicFormWriter.toString());
	}
    }

    // ----------------------
    // Finds the offset after DW editable tag in the argument buf.
    // At this offset, custom tags will be inserted.
    //
    private int findOffsetAfterEditableTag(StringBuffer buf) {
	int offset = buf.toString().indexOf(CLOSING_COMMENT_TAG);
	return offset + CLOSING_COMMENT_TAG.length() + 1;
    }

    // Finds the offset after body tag in bodyBuffer.
    // At this offset, custom tags will be inserted.
    //
    private int findOffsetBeforeEndBodyTag(StringBuffer buf) {
	String lowerCaseEndBodyTag = END_BODY_TAG.toLowerCase();
	String upperCaseEndBodyTag = END_BODY_TAG.toUpperCase();

	int index = buf.toString().indexOf(lowerCaseEndBodyTag);
	if (index != -1) {
	    return index;
	} else {
	    return buf.toString().indexOf(upperCaseEndBodyTag);
	}
    }

    // ----------------------
    // This method uses a DWBeanTemplateParser to process
    // a template.
    //
    // Returns:
    //          ERROR_ACCESSING_TEMPLATE if can't access template
    //          TEMPLATE_PARSED if successful
    // 
    private void processTemplate() {

        if (templateContents == null) {
            initStatus = ERROR_ACCESSING_TEMPLATE;
            return;
        }

        // Ask the parser to parse and validate the template.
        templateParser = new DWBeanTemplateParser();
        initStatus = templateParser.parseTemplate(templateContents, this);
    }

    // ----------------------
    public void printParsedRegions() {
	if (templateParser != null) {
	    templateParser.printRegions();
	}
    }

    // ----------------------
    // This method is no longer being used.
    // Bezita - 3/9/2000
    //
    // Mounts the dir where the template exists onto the File System.
    // Verifies read/write permissions on dir and template.
    // If the dir is already mounted, only verifies read/write
    // permissions of dir and template.
    //
    // Returns:
    //       ERROR_MOUNTING_FS if dir read/write permissions not set
    //       ERROR_ACCESSING_TEMPLATE if template read/write permissions not set
    //       FS_MOUNT_SUCCESSFUL if dir is mounted successfully
    //

    /**
       private int setRootFileSystem() {

       boolean fileSystemMounted = false;

       // Is the FileSystem already mounted onto the repository? 
       for (Enumeration en =
       TopManager.getDefault().getRepository().getFileSystems();
       en.hasMoreElements(); ) {

       FileSystem fs = (FileSystem)en.nextElement();
       File root = NbClassPath.toFile(fs.getRoot());
       if (root != null) {
       if (root.equals(templateParent))
       fileSystemMounted = true; 
       }
       }

       // Is the directory accessible?
       if ( !(templateParent.canRead() && templateParent.canWrite()) )
       return ERROR_MOUNTING_FS;

       // Is the file accessible?
       if ( !(template.canRead() && template.canWrite()) )
       return ERROR_ACCESSING_TEMPLATE;

       LocalFileSystem fileSystem = new LocalFileSystem();

       if (!fileSystemMounted) {
       try {
       fileSystem.setRootDirectory(templateParent);
       TopManager.getDefault().getRepository().addFileSystem(fileSystem);
       }  catch (PropertyVetoException ex) {
       return ERROR_MOUNTING_FS;
       } catch (IOException ex) {
       return ERROR_MOUNTING_FS;
       }
       }
       return FS_MOUNT_SUCCESSFUL;

       }  // end of setRootFileSystem()

    */

    // ----------------------
    protected void setPageDirectiveBuffer(String pageDir) {
        pageDirectiveBuffer = new StringBuffer(pageDir);
    }

    protected void setUseBeanBuffer(String useBean) {
        useBeanBuffer = new StringBuffer(useBean);
    }

    protected void setPropertyInitBuffer(String PropInit) {
        propertyInitBuffer = new StringBuffer(PropInit);
    }

    protected void setMethodCallBuffer(String methodCall) {
        methodCallBuffer = new StringBuffer(methodCall);
    }

    protected void setDynamicFormBuffer(String dynamicForm) {
        dynamicFormBuffer = new StringBuffer(dynamicForm);
    }

    protected void setBeginBuffer(String beginStr) {
        beginBuffer = new StringBuffer(beginStr);
    }

    protected void setInitRegionBuffer(String initRegionStr) {
        initRegionBuffer = new StringBuffer(initRegionStr);
    }

    protected void setBodyBuffer(String bodyStr) {
        bodyBuffer = new StringBuffer(bodyStr);
    }

    protected void setDynamicRegionBuffer(String dynamicRegionStr) {
        dynamicRegionBuffer = new StringBuffer(dynamicRegionStr);
    }

    protected void setEndBuffer(String endStr) {
        endBuffer = new StringBuffer(endStr);
    }

    // ----------------------
    protected void setFoundDWEditableInit(boolean flag) {
        foundDWEditableInit = flag;
    }

    protected void setFoundDWEditableDynamic(boolean flag) {
        foundDWEditableDynamicForm = flag;
    }

    // ----------------------

    /**
     * Tests DWBeanTemplate and DWBeanTemplateParser classes.
     */ 
    public static void main(String[] args) {

	if(args.length == 0) {
	    System.out.println(                //NOI18N
			       "Main-Template name (full path) is required to run this test.");
	    System.exit(0);
	}

	File template = new File(args[0]);
	DWBeanTemplate dwTemplate;

	try {
	    dwTemplate = DWBeanTemplate.getCustomTemplate(template);
	    dwTemplate.printParsedRegions();

	    // Get writers and simulate code generation.
	    PrintWriter pageDirWriter     = dwTemplate.getPageDirectiveWriter();
	    PrintWriter useBeanWriter     = dwTemplate.getUseBeanWriter();
	    PrintWriter propInitWriter    = dwTemplate.getPropertyInitWriter();
	    PrintWriter methodCallWriter  = dwTemplate.getMethodCallWriter();
	    PrintWriter dynamicFormWriter = dwTemplate.getDynamicFormWriter();

	    //NOI18N
	    pageDirWriter.println("This is generated code for Page Directive region");  // NOI18N
	    useBeanWriter.println("This is generated code for Use Bean region");        // NOI18N
	    propInitWriter.println("This is generated code for Property Init region");  // NOI18N
	    methodCallWriter.println("This is generated code for Method Call region");  // NOI18N
	    dynamicFormWriter.println("This is generated code for Dynamic Form region");// NOI18N

	    // Get contents of JSP page.
	    StringBuffer jspPageContents = dwTemplate.getJSPPageContents();
	    System.out.println("Main--------------------------------");  //NOI18N
	    System.out.println("Main-JSP Page contains:");               //NOI18N
	    System.out.println(jspPageContents.toString());
	    System.out.println("Main--------------------------------");  //NOI18N

	} catch (IOException e) {
            System.out.println(e);
	}

    }  // main
}

