/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.templates.model;

import java.io.*;
import java.util.ResourceBundle;
import java.lang.StringIndexOutOfBoundsException;
import org.openide.util.NbBundle;

// ------------------------------------

/**
 * This class parses a Dreamweaver template used by JSP Bean
 * wizard.  A Dreamweaver template may contain two DW editable
 * regions and five custom regions.  Dreamweaver editable regions
 * that may appear in the template are:
 *
 *          Initialization Region
 *          Dynamic Form Region
 *
 * If the template contains these DW editable regions, they
 * are preserved.  Otherwise, no action is taken.
 *
 * Custom regions are defined by the integration module and
 * have no meaning in DW.  Custom regions in the template are
 * identified by tags containing $$BEGIN and $$END.  Bean
 * wizard generates code and inserts the code into these
 * predefined regions.  If custom regions aren't in the template,
 * they are generated and added.  If custom regions do exist in
 * the template but are at the wrong location, they are moved.
 *
 * There are two types of custom regions:
 * Four custom regions used for initialization and one custom region
 * used for Dynamic Form.  Custom regions used for initialization
 * must appear before <body> tag.  If there is one or more DW
 * editable regions before <body> tag, custom regions are placed
 * inside the first DW editable region after the BeginEditable tag.
 * Otherwise, custom regions are simply placed before <body> tag.
 *
 * The custom region used for Dynamic Form must appear after <body>
 * tag. If there is one or more DW editable regions after <body> tag,
 * this custom region is placed inside the first DW editable region
 * after the BeginEditable tag.  Otherwise, the custom region
 * is simply placed after <body> tag.
 *
 * As the template is parsed, regions are identified and contents
 * of each region is stored in a buffer.  These buffers are then
 * given to a DWBeanTemplate object so that generated code
 * may be added to each region.
 *
 */


public class DWBeanTemplateParser {

   private static final String i18nBundle =          // NOI18N
      "org.netbeans.modules.web.templates.resources.Bundle";
   private static ResourceBundle resBundle = NbBundle.getBundle(i18nBundle);

   private static boolean parserInitialized;
   private DWBeanTemplate dwTemplate;

   // Strings identifying regions used by Bean Wizard
   //
   private static String PAGE_DIRECTIVE_BEGIN_TAG;
   private static String PAGE_DIRECTIVE_END_TAG;
   private static String USE_BEAN_BEGIN_TAG;
   private static String USE_BEAN_END_TAG;
   private static String PROPERTY_INIT_BEGIN_TAG;
   private static String PROPERTY_INIT_END_TAG;
   private static String METHOD_CALL_BEGIN_TAG;
   private static String METHOD_CALL_END_TAG;
   private static String DYNAMIC_FORM_BEGIN_TAG;
   private static String DYNAMIC_FORM_END_TAG;
   private static String BODY_TAG;
   private static String END_BODY_TAG;
   private static String HTML_TAG;

   // Strings identifying DW regions
   //
   private static String DW_EDITABLE_REGION_TAG;
   private static String DW_END_REGION_TAG;
   private static int dwEditableRegionTagLength;

   // These buffers contain the contents of template
   // which is initially read.
   //
   private StringBuffer pageDirectiveBuffer;
   private StringBuffer useBeanBuffer;
   private StringBuffer propertyInitBuffer;
   private StringBuffer methodCallBuffer;
   private StringBuffer dynamicFormBuffer;

   private StringBuffer beginBuffer;       // holds code before first editable region
   private StringBuffer initRegionBuffer;  // holds code in first editable region 
   private StringBuffer bodyBuffer;        // holds code in <body>
   private StringBuffer dynamicRegionBuffer;  // holds code in second editable region
   private StringBuffer endBuffer;            // holds code after last region

   // These static members are used to determine which
   // region of the template is being parsed currently.
   //
   private final static int BEGINNING           = 0;
   private final static int INIT_REGION         = 1;  // Identifies a DW editable region
   private final static int PAGE_DIRECTIVE      = 2;
   private final static int USEBEAN             = 3;
   private final static int PROPERTY_INIT       = 4;
   private final static int METHOD_CALL         = 5;
   private final static int DYNAMIC_FORM        = 6;
   private final static int BODY                = 7;
   private final static int DYNAMIC_FORM_REGION = 8;  // Identifies a DW editable region
   private final static int ENDING              = 9;

   // ---------------------
   private void initializeParser() {
       PAGE_DIRECTIVE_BEGIN_TAG = resBundle.getString("EI_PageDirectiveBeginTag");
       PAGE_DIRECTIVE_END_TAG = resBundle.getString("EI_PageDirectiveEndTag");
       USE_BEAN_BEGIN_TAG = resBundle.getString("EI_UseBeanBeginTag");
       USE_BEAN_END_TAG = resBundle.getString("EI_UseBeanEndTag");
       PROPERTY_INIT_BEGIN_TAG = resBundle.getString("EI_PropertyInitBeginTag");
       PROPERTY_INIT_END_TAG = resBundle.getString("EI_PropertyInitEndTag");
       METHOD_CALL_BEGIN_TAG = resBundle.getString("EI_MethodCallBeginTag");
       METHOD_CALL_END_TAG = resBundle.getString("EI_MethodCallEndTag");
       DYNAMIC_FORM_BEGIN_TAG = resBundle.getString("EI_DynamicFormBeginTag");
       DYNAMIC_FORM_END_TAG = resBundle.getString("EI_DynamicFormEndTag");
       DW_EDITABLE_REGION_TAG = resBundle.getString("EI_DWEditableRegionTag");
       DW_END_REGION_TAG = resBundle.getString("EI_DWEndTag");
       BODY_TAG = resBundle.getString("EI_BodyTag");
       END_BODY_TAG = resBundle.getString("EI_EndBodyTag");
       HTML_TAG = resBundle.getString("EI_HtmlTag");
       dwEditableRegionTagLength = DW_EDITABLE_REGION_TAG.length();
       parserInitialized = true;
   }

   // ---------------------
   private void createBuffers() {
          beginBuffer         = new StringBuffer();
          initRegionBuffer    = new StringBuffer();
          pageDirectiveBuffer = new StringBuffer();
          useBeanBuffer       = new StringBuffer();
          propertyInitBuffer  = new StringBuffer();
          methodCallBuffer    = new StringBuffer();
          bodyBuffer          = new StringBuffer();
          dynamicRegionBuffer = new StringBuffer();
          dynamicFormBuffer   = new StringBuffer();
          endBuffer           = new StringBuffer();
   }

   // ---------------------

   /**
    * This method parses a Dreamweaver template and populates
    * buffers as it identifies different regions.
    *
    * @param
    *       template - File object representing a Dreamweaver template
    *       dwTemp - DWBeanTemplate object providing public
    *                interface to the integration module.
    * @return
    *       ERROR_ACCESSING_TEMPLATE
    *       TEMPLATE_PARSED
    */
   protected int parseTemplate(BufferedReader templateContents,
                               DWBeanTemplate dwTemp) {

       if ( (templateContents == null) || (dwTemp == null) ) {
           return DWBeanTemplate.ERROR_ACCESSING_TEMPLATE;
       }

       if (!parserInitialized) {
           initializeParser();
       }

       createBuffers();
       dwTemplate = dwTemp;
       String line;
       boolean foundHtmlTag = false;
       boolean foundDWEditableInit = false;
       boolean foundBodyTag = false;
       boolean afterBodyTag = false;
       boolean foundDWEditableDynamic = false;
       boolean addToEndRegion = false; 

       try {
            while ( (line = templateContents.readLine()) != null) {

                 // Blank lines are ignored.
                 if (line.length() == 0) {
                    continue;
                 }

                 // Search for tags that are significant:
                 //      HTML_TAG to populate beginBuffer
                 //      DW_EDITABLE_REGION_TAG to identify editable regions 
                 //      BODY_TAG to populate bodyBuffer
                 //      END_BODY_TAG to populate endBuffer
                 //
                 if ( containsHtmlTag(line) ) {
                      foundHtmlTag = true;
                 } else if ( containsDWEditableRegionTag(line) &&
                             (!foundBodyTag) && (!afterBodyTag) ) {
                      foundHtmlTag = false;
                      foundDWEditableInit = true;
                      dwTemplate.setFoundDWEditableInit(true);
                 } else if ( containsBodyTag(line) ) {
                      foundHtmlTag = false;
                      foundDWEditableInit = false;
                      foundBodyTag = true;
                 } else if ( containsDWEditableRegionTag(line) &&
                             (foundBodyTag || afterBodyTag) ) {
                      foundHtmlTag = false;
                      foundDWEditableInit = false;
                      foundBodyTag = false;
                      afterBodyTag = true;
                      foundDWEditableDynamic = true;
                      dwTemplate.setFoundDWEditableDynamic(true);
                 } else if ( containsEndBodyTag(line) &&
                             (foundDWEditableDynamic) ) {
                      foundHtmlTag = false;
                      foundDWEditableInit = false;
                      foundBodyTag = false;
                      foundDWEditableDynamic = false;
                      addToEndRegion = true; 
                 }

                 // Insert the line into the appropriate buffer.
                 if (foundHtmlTag) {
                     appendLineToRegion(BEGINNING, line);
                 } else if (foundDWEditableInit) {
                     appendLineToRegion(INIT_REGION, line);
                 } else if (foundBodyTag) {
                     appendLineToRegion(BODY, line);
                 } else if (foundDWEditableDynamic) {
                     appendLineToRegion(DYNAMIC_FORM_REGION, line);
                 } else if ( (addToEndRegion) ) {
                     appendLineToRegion(ENDING, line);
                 }

            }  // while

            templateContents.close();

       } catch (IOException e) {
            return DWBeanTemplate.ERROR_ACCESSING_TEMPLATE;
       }

       // Insert custom tags into appropriate buffers
       populateCustomBuffers();
       setDWTemplateBuffers();
       return DWBeanTemplate.TEMPLATE_PARSED;
   }

   // ---------------------
   private boolean containsBodyTag(String str) {
        if ( str.compareToIgnoreCase(BODY_TAG) == 0 ) {
           return true;
        } else {
           String subInputStr = str.substring(0,5);
           String subBodyStr = BODY_TAG.substring(0,5);
           if (subInputStr.compareToIgnoreCase(subBodyStr) == 0) {
               return true;
           }
        }
        return false;
   }

   // ---------------------
   // Determines whether a string contains the Dreamweaver
   // tag identifying the beginning of an editable region.
   //
   private boolean containsDWEditableRegionTag(String str) {

      if (str.length() >= dwEditableRegionTagLength) {
          try {
               String tag = str.substring(0, dwEditableRegionTagLength);
               if ( tag.equals(DW_EDITABLE_REGION_TAG) ) {
                    return true;
               }
          } catch (StringIndexOutOfBoundsException e) {
               return false;
          }
      }
      return false;
   }

   // ---------------------
   // Determines whether a string contains the html tag.
   // This tag may appear in upper or lower case.
   //
   private boolean containsHtmlTag(String str) {
      if ( (str.equals(HTML_TAG.toLowerCase())) ||
           (str.equals(HTML_TAG.toUpperCase())) ) {
         return true;
      } else {
         return false;
      }
   }

   // Determines whether a string contains </body>. 
   // This tag may appear in upper or lower case.
   //
   private boolean containsEndBodyTag(String str) {
      if ( (str.equals(END_BODY_TAG.toLowerCase())) ||
           (str.equals(END_BODY_TAG.toUpperCase())) ) {
         return true;
      } else {
         return false;
      }
   }

   // ---------------------
   // This method populates buffers that will be used by Bean
   // wizard to generate and store code.
   //
   private void populateCustomBuffers() {
       int currentRegion = PAGE_DIRECTIVE;
       appendLineToRegion(currentRegion, PAGE_DIRECTIVE_BEGIN_TAG);
       appendLineToRegion(currentRegion++, PAGE_DIRECTIVE_END_TAG);
       appendLineToRegion(currentRegion, USE_BEAN_BEGIN_TAG);
       appendLineToRegion(currentRegion++, USE_BEAN_END_TAG);
       appendLineToRegion(currentRegion, PROPERTY_INIT_BEGIN_TAG);
       appendLineToRegion(currentRegion++, PROPERTY_INIT_END_TAG);
       appendLineToRegion(currentRegion, METHOD_CALL_BEGIN_TAG);
       appendLineToRegion(currentRegion++, METHOD_CALL_END_TAG);
       appendLineToRegion(currentRegion, DYNAMIC_FORM_BEGIN_TAG);
       appendLineToRegion(currentRegion, DYNAMIC_FORM_END_TAG);
   }

   // ---------------------
   private void setDWTemplateBuffers() {
       dwTemplate.setPageDirectiveBuffer(pageDirectiveBuffer.toString());
       dwTemplate.setUseBeanBuffer(useBeanBuffer.toString());
       dwTemplate.setPropertyInitBuffer(propertyInitBuffer.toString());
       dwTemplate.setMethodCallBuffer(methodCallBuffer.toString());
       dwTemplate.setDynamicFormBuffer(dynamicFormBuffer.toString());
       dwTemplate.setEndBuffer(endBuffer.toString());
       dwTemplate.setBeginBuffer(beginBuffer.toString());
       dwTemplate.setInitRegionBuffer(initRegionBuffer.toString());
       dwTemplate.setBodyBuffer(bodyBuffer.toString());
       dwTemplate.setDynamicRegionBuffer(dynamicRegionBuffer.toString());
   }

   // ---------------------
   // This method is used for testing and debugging
   // to verify contents of each buffer.
   //
   protected void printRegions() {
       System.out.println("beginBuffer contains:");
       System.out.println(beginBuffer);

       System.out.println("initRegionBuffer contains:");
       System.out.println(initRegionBuffer);

       System.out.println("PageDirectiveBuffer contains:");
       System.out.println(pageDirectiveBuffer);

       System.out.println("UseBeanBuffer contains:");
       System.out.println(useBeanBuffer);

       System.out.println("PropertyInitializationBuffer contains:");
       System.out.println(propertyInitBuffer);

       System.out.println("MethodCallBuffer contains:");
       System.out.println(methodCallBuffer);

       System.out.println("bodyBuffer contains:");
       System.out.println(bodyBuffer);

       System.out.println("DynamicFormBuffer contains:");
       System.out.println(dynamicFormBuffer);

       System.out.println("dynamicRegionBuffer contains:");
       System.out.println(dynamicRegionBuffer);

       System.out.println("endBuffer contains:");
       System.out.println(endBuffer);
   }

   // ---------------------
   private void appendLineToRegion(int currentRegion, String line) {

      // Save line in the current buffer
      switch (currentRegion) {
           case BEGINNING:
                beginBuffer.append(line);
                beginBuffer.append("\n");
                break;
           case INIT_REGION:
                if (!isCustomTag(line)) {
                    initRegionBuffer.append(line);
                    initRegionBuffer.append("\n");
                }
                break;
           case BODY:
                bodyBuffer.append(line);
                bodyBuffer.append("\n");
                break;
           case DYNAMIC_FORM_REGION:
                if (!isCustomTag(line)) {
                    dynamicRegionBuffer.append(line);
                    dynamicRegionBuffer.append("\n");
                }
                break;
           case PAGE_DIRECTIVE:
                pageDirectiveBuffer.append(line);
                pageDirectiveBuffer.append("\n");
                break;
           case USEBEAN:
                useBeanBuffer.append(line);
                useBeanBuffer.append("\n");
                break;
           case PROPERTY_INIT:
                propertyInitBuffer.append(line);
                propertyInitBuffer.append("\n");
                break;
           case METHOD_CALL:
                methodCallBuffer.append(line);
                methodCallBuffer.append("\n");
                break;
           case DYNAMIC_FORM:
                dynamicFormBuffer.append(line);
                dynamicFormBuffer.append("\n");
                break;
           default:
                endBuffer.append(line);
                endBuffer.append("\n");
                break;
      } // switch

   }    // appendLinetoRegion

   // ---------------------
   // Determines whether a string is one of the custom tags
   // used by Bean wizards.
   // 
   private boolean isCustomTag(String str) {

       if ( (str.equals(PAGE_DIRECTIVE_BEGIN_TAG)) ||
            (str.equals(PAGE_DIRECTIVE_END_TAG)) ||
            (str.equals(USE_BEAN_BEGIN_TAG)) ||
            (str.equals(USE_BEAN_END_TAG)) ||
            (str.equals(PROPERTY_INIT_BEGIN_TAG)) ||
            (str.equals(PROPERTY_INIT_END_TAG)) ||
            (str.equals(METHOD_CALL_BEGIN_TAG)) ||
            (str.equals(METHOD_CALL_END_TAG)) ||
            (str.equals(DYNAMIC_FORM_BEGIN_TAG)) ||
            (str.equals(DYNAMIC_FORM_END_TAG)) ) {
           return true;
       } else {
           return false;
       }
   }
}

