/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.templates.model;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import javax.swing.JFileChooser;
import java.beans.PropertyVetoException;
import java.util.ResourceBundle;
import java.net.URL;
import java.lang.ClassLoader;
import java.util.jar.*;

import org.openide.util.NbBundle;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.LocalFileSystem;
import org.openide.execution.NbClassPath;
import org.openide.TopManager;

// ------------------------------------

/**
 * DatabaseTemplate class provides functionality to:
 *   -- Access PrintWriter objects for each region of template.
 *   -- Save generated code written to PrintWriters into 
 *      regions of template.
 *   -- Receive entire contents of a template after code generation.
 *
 * A typical scenario for using this class:
 * (Refer to the main() method at the end of this file for an example.)
 *
 *   -- Get a DatabaseTemplate object using one of the static get methods:
 *      DatabaseTemplate.getDefaultTemplate(arg) loads a
 *      default template using the argument supplied.
 *      DatabaseTemplate.getTemplate(arg) loads a custom
 *      template using the argument supplied.
 *   -- If a DatabaseTemplate object is successfully initialized,
 *      get individual PrintWriter objects for each region.
 *   -- Generate code and write it to the appropriate PrintWriter.
 *   -- Call getJSPPageContents() to receive a StringBuffer
 *      containing the original contents of the template plus
 *      any generated code inserted into appropriate regions.
 */

// ------------------------------------

public class DatabaseTemplate {

    /**
     * Used to denote status of accessing and validating templates.
     */
    public final static int TEMPLATE_PARSED          = 0;
    public final static int ERROR_ACCESSING_TEMPLATE = 1;

    /**
     *  Member variables that identify default templates
     */
    public final static int SINGLE_LEVEL_DBTOJSP        = 0;
    public final static int SINGLE_LEVEL_QUERY_RESULT   = 1;
    public final static int MASTER_DETAIL_DBTOJSP       = 2;
    public final static int MASTER_DETAIL_QUERY_RESULT  = 3;

    // -------------------------
    // Private Member Variables

    private final static String i18nBundle =          // NOI18N
	"org.netbeans.modules.web.templates.resources.Bundle";

    private static ResourceBundle resBundle = NbBundle.getBundle(i18nBundle);

    private BufferedReader templateContents;
    private DatabaseTemplateParser templateParser;
    private String defaultTemplateName;
    private Vector regionInfoVector;

    // Designates whether the DatabaseTemplate object has been
    // successfully initialized.
    private int initStatus;

    // -------------------------
    // Public Methods

    public static DatabaseTemplate getDefaultTempalte(int templateType)
	throws IOException {

	DatabaseTemplate dwObject = new DatabaseTemplate(templateType);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException(
				  "Unable to initialize DatabaseTemplate object!");
	}
	return dwObject;
    }

    public static DatabaseTemplate getTemplate(FileObject tempFileObj)
	throws IOException {

	DatabaseTemplate dwObject = new DatabaseTemplate(tempFileObj);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException(
				  "Unable to initialize DatabaseTemplate object!");
	}
	return dwObject;
    }

    public static DatabaseTemplate getTemplate(File tempFile)
	throws IOException {

	DatabaseTemplate dwObject = new DatabaseTemplate(tempFile);
	if (dwObject.getInitStatus() == ERROR_ACCESSING_TEMPLATE) {
	    throw new IOException(
				  "Unable to initialize DatabaseTemplate object!");
	}
	return dwObject;
    }

    // -------------------------
    public PrintWriter getPrintWriter(String regionName,
				      String regionPosition) {

	CharArrayWriter writer = new CharArrayWriter();
	PrintWriter printWriter = new PrintWriter(writer, true);
	regionInfoVector.add(new RegionInfo(regionName, null, writer, regionPosition));
	//printVectorElements();
	return printWriter;
    }

    // -------------------------
    public PrintWriter getPrintWriter(String regionName,
				      String alternativeRegion,
				      String regionPosition) {

	CharArrayWriter writer = new CharArrayWriter();
	PrintWriter printWriter = new PrintWriter(writer, true);
	regionInfoVector.add(new RegionInfo(regionName,
					    alternativeRegion,
					    writer,
					    regionPosition));
	//printVectorElements();
	return printWriter;
    }


    // -------------------------
    public StringBuffer getJSPPageContents() throws IOException {

	if ( (templateContents == null) ||
	     (initStatus == ERROR_ACCESSING_TEMPLATE) ) {
	    throw new IOException("Unable to create contents of JSP page.");
	} else {

	    templateParser = new DatabaseTemplateParser();
	    initStatus = templateParser.parseTemplate(regionInfoVector, templateContents);
	    if (initStatus != ERROR_ACCESSING_TEMPLATE) {
		return templateParser.getJSPPageBuffer();
	    } else {
		throw new IOException("Unable to create contents of JSP page.");
	    }
	}
    }

    // -------------------------
    // Private Methods

    /**
     * Constructors are private.
     * Use one of the public static methods to create and
     * initialize a DatabaseTemplate object.
     */

    private DatabaseTemplate(int templateType) {
	setDefaultTemplate(templateType);
	loadDefaultTemplate();
	regionInfoVector = new Vector();
    }

    private DatabaseTemplate(FileObject fileObj) {

	try {
	    templateContents = new BufferedReader(
						  new InputStreamReader(fileObj.getInputStream()));
	    regionInfoVector = new Vector();
	} catch (IOException e) {
	    initStatus = ERROR_ACCESSING_TEMPLATE;
	}
    }

    private DatabaseTemplate(File template) {

	if (template.exists()) {
	    try {
		templateContents = new BufferedReader(new FileReader(template));
		regionInfoVector = new Vector();
	    } catch (IOException e) {
		initStatus = ERROR_ACCESSING_TEMPLATE;
	    }
	} else {
	    initStatus = ERROR_ACCESSING_TEMPLATE;
	}
    }

    // -------------------------
    private int getInitStatus() {
	return initStatus;
    }

    // -------------------------
    private void setDefaultTemplate(int templateType) {
	// A switch statement is being used in order to support
	// addition of other default templates easily.
	//
	switch (templateType) {

	case SINGLE_LEVEL_DBTOJSP:
	    defaultTemplateName =
		resBundle.getString("EI_DWSingleLevelDBtoJSP");
	    break;
	case SINGLE_LEVEL_QUERY_RESULT:
	    defaultTemplateName =
		resBundle.getString("EI_DWSingleLevelQueryResult");
	    break;
	case MASTER_DETAIL_DBTOJSP:
	    defaultTemplateName =
		resBundle.getString("EI_DWMasterDetailDBtoJSP");
	    break;
	case MASTER_DETAIL_QUERY_RESULT:
	    defaultTemplateName =
		resBundle.getString("EI_DWMasterDetailQueryResult");
	    break;
	default:
	    defaultTemplateName =
		resBundle.getString("EI_DWSingleLevelDBtoJSP");
	    break;
	}
    }

    // -------------------------
    //  This method accesses the default template as a system
    //  resource.  It then initializes templateContents with
    //  the contents of this default template.
    //
    private void loadDefaultTemplate() {
	JarInputStream jarIstream;
	boolean usingDefaultTemplate = false;

        try {
            FileSystem defaultFs = TopManager.getDefault().getRepository().getDefaultFileSystem();
            FileObject fo = defaultFs.findResource (defaultTemplateName);
            if (fo != null) {
                templateContents = new BufferedReader (new InputStreamReader (fo.getInputStream ()));
            }

        } catch (IOException e) {
            org.openide.TopManager.getDefault ().notifyException (e);
            initStatus = ERROR_ACCESSING_TEMPLATE;
        }
    }

    // -------------------------
    // This method is used for debugging and verification
    // 
    private void printVectorElements() {

	System.out.println("regionInfoVector contains:");

	for (Enumeration en = regionInfoVector.elements(); en.hasMoreElements(); ) {
	    RegionInfo regInfo = (RegionInfo)en.nextElement();
	    System.out.println("RegionInfo.region " + regInfo.getRegion());
	    System.out.println("RegionInfo.alternativeRegion " +
			       regInfo.getAlternativeRegion());
	    System.out.println("RegionInfo.position " + regInfo.getPosition());
	    System.out.println("RegionInfo.generatedCodeInserted " +
			       regInfo.getGeneratedCodeInserted());
	    System.out.println("-----------------------");
	}
    }

    // -------------------------
    // Protected inner class

    protected class RegionInfo {

	private String region;
	private String alternativeRegion;
	private CharArrayWriter writer;
	private String position;
	private boolean generatedCodeInserted;

	protected RegionInfo(String regName,
			     String altReg,
			     CharArrayWriter charWriter,
			     String pos) {
	    region = regName;
	    alternativeRegion = altReg;
	    writer = charWriter;
	    position = pos;
	    generatedCodeInserted = false;
	}

	protected String getRegion() {
	    return region;
	}

	protected String getAlternativeRegion() {
	    return alternativeRegion;
	}

	protected CharArrayWriter getWriter() {
	    return writer;
	}

	protected String getPosition() {
	    return position;
	}

	protected boolean getGeneratedCodeInserted() {
	    return generatedCodeInserted;
	}

	protected void setGeneratedCodeInserted(boolean status) {
	    generatedCodeInserted = status;
	}
    }

    // -------------------------
    // Main Method for Testing

    /**
     * Tests DatabaseTemplate and DatabaseTemplateParser classes.
     */ 
    public static void main(String[] args) {

	if(args.length == 0) {
	    System.out.println(                //NOI18N
			       "Main-Template name (full path) is required to run this test.");
	    System.exit(0);
	}

	File template = new File(args[0]);
	DatabaseTemplate dbTemp;

	try {
	    dbTemp = DatabaseTemplate.getTemplate(template);

	    // ------------------------------------
	    // Uncomment the test you wish to run!
	    // Only one test at a time may be run!

	    // ------------------------------------
	    // Testing Single Level Query
	    // Get writers and simulate code generation.
	    //PrintWriter queryWriter =
	    //dbTemp.getPrintWriter("query", "AllOperations", "body");  //NOI18N
	    //PrintWriter resultWriter =
	    //dbTemp.getPrintWriter("result", "AllOperations", "bodby"); //NOI18N
	    //PrintWriter toolbarWriter =
	    //dbTemp.getPrintWriter("toolbar", "body"); //NOI18N

	    //queryWriter.println("This is generated code for query.");     //NOI18N
	    //resultWriter.println("This is generated code for result.");   //NOI18N
	    //toolbarWriter.println("This is generated code for toolbar."); //NOI18N

	    // ------------------------------------
	    // Testing Master/Detail Query
	    // Get writers and simulate code generation.
	    PrintWriter queryMasterWriter =
		dbTemp.getPrintWriter("Query Master",
				      "AllOperations Master", "header");  //NOI18N
	    PrintWriter resultMasterWriter =
		dbTemp.getPrintWriter("Result Master", 
				      "AllOperations Master", "header"); //NOI18N
	    PrintWriter detailWriter =
		dbTemp.getPrintWriter("AllOperations Detail", "header"); //NOI18N
	    PrintWriter toolbarWriter =
		dbTemp.getPrintWriter("toolbar", "body"); //NOI18N

	    //NOI18N
	    queryMasterWriter.println("This is generated code for query master.");
	    resultMasterWriter.println("This is generated code for result master.");
	    detailWriter.println("This is generated code for detail.");   //NOI18N
	    toolbarWriter.println("This is generated code for toolbar."); //NOI18N

	    // ------------------------------------
	    // Get contents of JSP page.
	    StringBuffer jspPageContents = dbTemp.getJSPPageContents();
	    System.out.println("Main--------------------------------");  //NOI18N
	    System.out.println("Main-JSP Page contains:");               //NOI18N
	    System.out.println(jspPageContents.toString());
	    System.out.println("Main--------------------------------");  //NOI18N

	} catch (IOException e) {
            System.out.println(e);
	}

    }  // main
}

