/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2000 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package org.netbeans.modules.web.templates.model;

import java.io.*;
import java.util.ResourceBundle;
import java.lang.StringIndexOutOfBoundsException;
import java.util.Vector;
import java.util.Enumeration;
import org.openide.util.NbBundle;

import org.netbeans.modules.web.templates.model.DatabaseTemplate.RegionInfo;

// ------------------------------------

/**
 * This class is used by DatabaseTemplate to parse
 * a template and insert generated code in appropriate regions or
 * position within the template.  Default or custom templates are
 * used by JSP Database wizard to stored generated JSP code.
 * A template may or maynot contain predefined regions.
 *
 * Predefined regions are:
 *    Dreamweaver Editable Regions
 *    Custom Regions
 *
 * Custom regions are defined by the Integration module and
 * have no meaning in DW.  Custom regions in the template are
 * identified by tags containing $$BEGIN and $$END.
 *
 * Database wizard identifies the regions for which it is going
 * to generate code.  The wizard may also identify alternative regions
 * or positions in the template where the generated code may be
 * inserted if the original regions don't exist.
 *
 * The parser first attempts to insert the generated code in
 * the regions identified by the Database wizard.  If these regions
 * don't exist, the parser then attempts to insert the generated
 * code into the alternative regions.  If alternative regions
 * don't exist or haven't been identified (null strings), the
 * parser attempts to insert the generated code into positions
 * identified by the Database wizard.  If these positions don't
 * exist or haven't been identified (null strings), the parser
 * inserts the generated code into a default position.
 *
 */

public class DatabaseTemplateParser {

    private static final String i18nBundle =          // NOI18N
	"org.netbeans.modules.web.templates.resources.Bundle";
    private static ResourceBundle resBundle = NbBundle.getBundle(i18nBundle);

    private static String JASPER_BEGIN_REGION_TAG = 
	resBundle.getString("EI_JasperBeginRegionTag");
    private static String JASPER_END_REGION_TAG =
	resBundle.getString("EI_JasperEndRegionTag");
    private static String DEFAULT_POSITION =
	resBundle.getString("EI_BodyTag");
    private static String HEADER_POSITION =
	resBundle.getString("EI_HeaderPosition");
    private static String BODY_POSITION =
	resBundle.getString("EI_BodyPosition");

    // Platform-dependent newline character
    private static String NEWLINE = System.getProperty("line.separator");

    // jspPageBuffer contains the contents of template
    // and generated code.
    private StringBuffer jspPageBuffer;

    // regionTagVector contains a list of region tags
    // for which code has been generated.
    private Vector regionTagVector;

    // matchedRegionTag identifies the region tag
    // that has been found in a template.
    private String matchedRegionTag;

    // ---------------------
    protected int parseTemplate(Vector regionInfoVector,
				BufferedReader templateContents) {

	if ( (templateContents == null) || (regionInfoVector == null) ) {
	    return DatabaseTemplate.ERROR_ACCESSING_TEMPLATE;
	}

	initializeParser(regionInfoVector);
	jspPageBuffer = new StringBuffer();
	String line;

	try {
	    while ( (line = templateContents.readLine()) != null) {

		// Blank lines are ignored.
		if (line.length() == 0) {
		    continue;
		}

		// If line contains a region tag,
		// add the contents of that region's Writer to
		// the jspPageBuffer.  Otherwise, just add the line
		// to the jspPageBuffer.
		//
		if (containsRegionTag(line)) {
		    jspPageBuffer.append(line + NEWLINE);
		    addGeneratedCode(regionInfoVector);
		} else {
		    jspPageBuffer.append(line + NEWLINE);
		}

	    }  // while

	    templateContents.close();

	} catch (IOException e) {
	    return DatabaseTemplate.ERROR_ACCESSING_TEMPLATE;
	}

	insertUnusedGeneratedCode(regionInfoVector);
	return DatabaseTemplate.TEMPLATE_PARSED;
    }

    // ---------------------
    protected StringBuffer getJSPPageBuffer() {
	return jspPageBuffer;
    }

    // ---------------------
    // Builds a list of region tags and alternative region tags
    // for which code has been generated.
    //
    private void initializeParser(Vector regionInfoVector) {

	regionTagVector = new Vector();

	for (Enumeration en = regionInfoVector.elements(); en.hasMoreElements(); ) {

	    RegionInfo regInfo = (RegionInfo)en.nextElement();
	    String regionTag = regInfo.getRegion();
	    String altRegionTag = regInfo.getAlternativeRegion(); 
 
	    if (regionTag != null) {
		regionTag = JASPER_BEGIN_REGION_TAG +
		    " " +
		    regionTag;
		if (!regionTagVector.contains(regionTag)) {
                    regionTagVector.add(regionTag);
		}
	    }

	    if (altRegionTag != null) {
		altRegionTag = JASPER_BEGIN_REGION_TAG +
		    " " +
		    altRegionTag; 
		if (!regionTagVector.contains(altRegionTag)) {
                    regionTagVector.add(altRegionTag);
		}
	    }

	}  // for

	//printRegionTagVectorElements();
    }

    // ---------------------
    // Finds the Writer which holds the generated code
    // for matchedRegionTag.  Adds the generated code to jspPageBuffer.
    // Marks the generated code as used.
    // 
    private void addGeneratedCode(Vector regionInfoVector) {

	// First we'll attempt to insert generated code into
	// the region identified by RegionInfo.getRegion().
	// 
	for (Enumeration en = regionInfoVector.elements(); en.hasMoreElements(); ) {
	    RegionInfo regInfo = (RegionInfo)en.nextElement();

	    if (matchedRegionTag.indexOf(regInfo.getRegion()) != -1) {
		jspPageBuffer.append(regInfo.getWriter().toString());
		regInfo.setGeneratedCodeInserted(true);
	    }
	}

	// Now we'll attempt to insert generated code into
	// the region identified by RegionInfo.getAlternativeRegion()
	// if and only if the generated code isn't already used.
	//
	// Generated code is considered to be used if a region has 
	// already been found in the template and the code has
	// been inserted in that region.
	//
	for (Enumeration en = regionInfoVector.elements(); en.hasMoreElements(); ) {
	    RegionInfo regInfo = (RegionInfo)en.nextElement();

	    if ( (!regInfo.getGeneratedCodeInserted()) &&
		 (regInfo.getAlternativeRegion() != null) ) {
		if (matchedRegionTag.indexOf(regInfo.getAlternativeRegion()) != -1) {
		    jspPageBuffer.append(regInfo.getWriter().toString());
		    regInfo.setGeneratedCodeInserted(true);
		}
	    }
	}
    }

    // ---------------------
    // Searches regionTagVector to determine whether a
    // line contains any region tags for which code has
    // been generated.
    //
    // Returns true if the region tag is found and stores
    // the value in matchedRegionTag.  Otherwise, returns
    // false.
    //
    private boolean containsRegionTag(String line) {

	for (Enumeration en = regionTagVector.elements(); en.hasMoreElements();) {

	    String regionTag = (String)en.nextElement();

	    if (line.regionMatches(true, 0, regionTag, 0, regionTag.length())) {
		matchedRegionTag = regionTag;
		return true;
	    }
	}
	matchedRegionTag = null;
	return false;
    }

    // ---------------------
    // Inserts generated code into jspPageBuffer if
    // the generated code hasn't already been inserted in
    // a region of the template.  This condition can only
    // occur if neither the region nor the alternative region
    // is in the template.
    //
    private void insertUnusedGeneratedCode(Vector regionInfoVector) {

	for (Enumeration en = regionInfoVector.elements(); en.hasMoreElements(); ) {
	    RegionInfo regInfo = (RegionInfo)en.nextElement();

	    if (!regInfo.getGeneratedCodeInserted()) {
                int bufferOffset = findOffset(regInfo.getPosition());
                jspPageBuffer.insert(bufferOffset,
                                     regInfo.getWriter().toString());
                regInfo.setGeneratedCodeInserted(true);
	    }
	}
    }

    // ---------------------
    // Finds the offset of positionStr in jspPageBuffer.
    // If this offset can't be found because positionStr
    // doesn't exist in jspPageBuffer, the offset of a
    // default position will be returned.
    //
    // Default position is identified by DEFAULT_POSITION
    // member variable.
    //
    private int findOffset(String positionStr) {

	// Find the offset at which the default position starts. 
	// The default position string may appear in upper or
	// lower case.
	//
	int offset = jspPageBuffer.toString().indexOf(
						      DEFAULT_POSITION.toUpperCase());

	if (offset == -1) {
	    offset = jspPageBuffer.toString().indexOf(
						      DEFAULT_POSITION.toLowerCase());
	}

	if (positionStr == null) {
	    return offset;
	}

	if (positionStr.equals(HEADER_POSITION)) {
	    // Return the offset before default position 
	    return offset;
	} else {
	    // Return the offset after default position 
	    return offset + DEFAULT_POSITION.length() + 1;
	}
    }

    // ---------------------
    // This method is used for debugging and verification
    //
    private void printRegionTagVectorElements() {

	System.out.println("regionTagVector contains:");
	for (Enumeration en = regionTagVector.elements(); en.hasMoreElements();) {
	    String regionTag = (String)en.nextElement();
	    System.out.println("        " + regionTag);
	}
    }
}

