
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde.editor;

import javax.swing.*;
import java.io.IOException;

/**
 * Represents the core editor functionality that AJDE depends on.  Should be extended
 * in order to support richer editing.
 *
 * @author Mik Kersten
 */
public abstract class AjdeEditor {

    public abstract void showSourceLine(String filePath, int lineNumber, boolean highlight);

    public abstract void showSourceLine(int lineNumber, boolean highlight);

    public abstract void showSourcelineAnnotation(String filePath, int lineNumber, JPopupMenu annotation);

    public abstract String getCurrFile();

    public abstract void saveContents() throws IOException;

    public abstract void pasteToCaretPos(String text);

    public abstract JPanel getPanel();

    /**
     * TODO: refactor (used by debugger)
     * Show the file given by the full name <code>filePath</code> in the IDE.
     *
     * @param filePath    The <b>full</b> name of the file to show.
     */
    public void showSourceForFile(String filePath) {
        showSourceLine(filePath, 0, false);
    }

    /**
     * TODO: refactor (used by debugger)
     * Seek to the line given by <code>lineNumber</code> in the file
     * the currently has the IDE's focus and decide whether to highlight
     * it based on <code>highlight</code>.
     *
     * @param lineNumber  The line to which we seek.
     * @param highlight   Whether this line should be highlighted.
     */
    public void showSourceForLine(int lineNumber, boolean highlight) {
        showSourceLine(getCurrFile(), 0, false);
    }

    /**
     * TODO: refactor (used by debugger)
     * Show the file given by the full name <code>filePath</code> in the IDE,
     * seek to the line given by <code>lineNumber</code> in the file, and decide
     * whether to highlight it based on <code>highlight</code>.
     *
     * @param filePath    The <b>full</b> name of the file to show.
     * @param lineNumber  The line to which we seek.
     * @param highlight   Whether this line should be highlighted.
     */
    public void showSourceForSourceLine(String filePath, int lineNumber, boolean highlight) {
        showSourceLine(filePath, lineNumber, highlight);
    }

    /**
     * TODO: refactor (used by debugger)
     * Returns the current source path.
     *
     * @return The current source path.
     */
    public String getCurrSourceFilePath() {
        return getCurrFile();
    }

    /**
     * TODO: refactor (used by debugger)
     * When a breakpoint is requested this method is called.  The breakpoint occurs
     * in the file given by the <b>full</b> name <code>filePath</code> at the line
     * given by <code>lineNumber</code>.  The arguments <code>isDeferred</code> tells
     * whether this request has been <i>set</i> or <i>deferred</i>.
     *
     * @param filePath    The <b>full</b> name of the file to show.
     * @param lineNumber  The line to which we seek.
     * @param isDeferred  <code>true</code> means this breakpoint was <i>deferred</i>,
     *                    and <code>false</code> means this breakpoint was <i>set</i>.
     */
    public void setBreakpointRequest(String filePath, int lineNumber, boolean isDeferred) {
        // ignored
    }

    /**
     * TODO: refactor (used by debugger)
     * When a breakpoint is cleared this method is called.  The breakpoint was previously
     * in the file given by the <b>full</b> name <code>filePath</code> at the line
     * given by <code>lineNumber</code>.
     *
     * @param filePath    The <b>full</b> name of the file to show.
     * @param lineNumber  The line to which we seek.
     */
    public void clearBreakpointRequest(String filePath, int lineNumber) {
        // ignored
    }

    protected class GoToLineThread extends Thread {
        private int lineNumber = 0;
        private String fileToSeekTo = "";
        private boolean finished = false;

        public boolean isFinished() {
            return finished;
        }

        public boolean needsRetry() {
            return !this.isAlive() && !finished;
        }

        public GoToLineThread(String fileToSeekTo, int lineNumber) {
            this.lineNumber = lineNumber;
            this.fileToSeekTo = fileToSeekTo;
        }

        public void run() {

            while(true) {
                String file = getCurrFile();
                if (file != null) {
                    if (file.equals(this.fileToSeekTo)) {
                        try {
                        SwingUtilities.invokeAndWait( new Runnable() {
                            public void run() {
                                showSourceLine(lineNumber, true);
                            }
                        });
                        } catch (Exception e) { e.printStackTrace(); }
                        finished = true;
                        break;
                    }
                    shortPause();
                }
            }
        }

        private void shortPause() {
            try {
                this.sleep(100);
            } catch (InterruptedException e) {
                throw new RuntimeException(e.getMessage());
            }
        }
    }
}