
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm;

import java.util.*;
import java.io.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.ast.Type;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.base.parser.*;
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.crosscuts.ast.*;

import org.aspectj.asm.associations.*;

/**
 * @author Mik Kersten
 */
public class StructureManager {
    protected StructureModel model = new StructureModel();
    private List structureListeners = new ArrayList();
    private List associations = new ArrayList();

    public static Correspondences correspondences = null; // HACK: remove this

    private static final StructureNode NO_STRUCTURE = new StructureNode("", "", null);
//    private modelProperties properties = new StructureViewProperties();

    public StructureManager() {
        associations.add(new Advice());
        associations.add(new Introduction());
        associations.add(new Inheritance());
        associations.add(new Reference());
    }

    public StructureNode getStructureModel() {
        if (!model.isValid()) {
            return NO_STRUCTURE;
        } else {
            return model.getRoot();
        }
    }

    /**
     * @todo    make generic
     */
    public StructureNode getStructureModel(String filePath) {
        if (!model.isValid()) {
            return null;
        } else {
            String correctedPath = filePath.replace('\\', '/');
            StructureNode node = (StructureNode)model.getFileMap().get(correctedPath);//findFileNode(filePath, model);
            if (node != null) {
                return node;
            } else {
                return NO_STRUCTURE;
            }
        }
    }

    public StructureNode buildStructureModel(AspectJCompiler compiler) {
//        model = null;
        StructureNodeFactory.clear();
        StructureManager.correspondences = compiler.getCorrespondences();
        World world = compiler.getWorld();
        String configFile = compiler.getBuildConfigFile();

        buildProgramStructurePass(world, configFile);
        buildRelationsPass(model.getRoot());
        notifyListeners();

        if (configFile != null) {
            writeStructureModel(configFile);
        }
        return model.getRoot();
    }

    /**
     * Constructs map each time it's called.
     */
    public HashMap getInlineAnnotations(String sourceFile) {
        if (!model.isValid()) return null;

        HashMap annotations = new HashMap();
        StructureNode node = getStructureModel(sourceFile);
        if (node == NO_STRUCTURE) {
            return null;
        } else {
            ProgramElementNode fileNode = (ProgramElementNode)node;
            ArrayList peNodes = new ArrayList();
            getAllStructureChildren(fileNode, peNodes);
            for (Iterator it = peNodes.iterator(); it.hasNext(); ) {
                ProgramElementNode peNode = (ProgramElementNode)it.next();
                List entries = new ArrayList();
                entries.add(peNode);
                Integer hash = new Integer(peNode.getBeginLine());
                List existingEntry = (List)annotations.get(hash);
                if (existingEntry != null) {
                    entries.addAll(existingEntry);
                }
                annotations.put(hash, entries);
            }
            return annotations;
        }
    }

    private void getAllStructureChildren(ProgramElementNode node, List result) {
        List children = node.getChildren();
        for (Iterator it = children.iterator(); it.hasNext(); ) {
            StructureNode next = (StructureNode)it.next();
            if (next instanceof ProgramElementNode) {
                ProgramElementNode pNode = (ProgramElementNode)next;
                if (pNode.isCode() && pNode.getRelations() != null && pNode.getRelations().size() > 0) {
                    result.add(next);
                }
                getAllStructureChildren((ProgramElementNode)next, result);
            }
        }
    }

    public void addStructureListener(StructureModelListener listener) {
        structureListeners.add(listener);
    }

    public void removeStructureListener(StructureModelListener listener) {
        structureListeners.remove(listener);
    }

    private void notifyListeners() {
        for (Iterator it = structureListeners.iterator(); it.hasNext(); ) {
            ((StructureModelListener)it.next()).modelUpdated();
        }
    }

    public List getAssociations() {
        return associations;
    }

    private void buildProgramStructurePass(World world, String configFile) {
        String rootName = "<root name>";
        List files = new ArrayList();
        HashMap modelFileMap = new HashMap();
        for (Iterator it = world.getCompilationUnits().iterator(); it.hasNext(); ) {
            CompilationUnit compilationUnit = (CompilationUnit)it.next();
            ProgramElementNode fileNode = makeProgramElementNode(compilationUnit, correspondences);
            files.add(fileNode);
            modelFileMap.put(compilationUnit.getSourceFileName().replace('\\', '/'), fileNode);
        }
        model.setFileMap(modelFileMap);
        model.setRoot(mapFilesIntoPackages(' ' + configFile));
    }

    private StructureNode mapFilesIntoPackages(String rootName) {
        String rootSig = new File(rootName).getName();
        StructureNode root = new StructureNode(rootSig, "project");
        ArrayList packages = new ArrayList();
        Set fileSet = model.getFileMap().entrySet();
        TreeSet packageSet = new TreeSet();
        for (Iterator it = fileSet.iterator(); it.hasNext(); ) {
            ProgramElementNode peNode = (ProgramElementNode)((Map.Entry)it.next()).getValue();
            if (peNode.getPackageName() != null) {
                packageSet.add(peNode.getPackageName());
            }
        }
        for (Iterator it2 = packageSet.iterator(); it2.hasNext(); ) {
            addPackageToRoot(root, (String)it2.next());
        }
        for (Iterator it3 = fileSet.iterator(); it3.hasNext(); ) {
            addFileToPackage(root, (ProgramElementNode)((Map.Entry)it3.next()).getValue());
        }
        return root;
    }

    private void addPackageToRoot(StructureNode node, String packageName) {
        List children = node.getChildren();
        int dotIndex = packageName.indexOf('.');
//        if (packageName == "") {
//            node.addChild();
//        } else
        if (dotIndex == -1) {
            if (!node.getChildren().contains(packageName)) {
                node.addChild(new StructureNode(packageName, "package"));
            }
        } else {
            String pkgTop = packageName.substring(0, dotIndex);
            String pkgRest = packageName.substring(dotIndex+1, packageName.length());
            boolean added =false;
            for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
                StructureNode child = (StructureNode)it.next();
                if (child.getName().equals(pkgTop)) {
                    addPackageToRoot(child, pkgRest);
                    added = true;
                }
            }
            if (!added) {
                StructureNode newChild = new StructureNode(pkgTop, "package");
                node.addChild(newChild);
                addPackageToRoot(newChild, pkgRest);
            }
        }
    }

    private void addFileToPackage(StructureNode root, ProgramElementNode fileNode) {
        if (fileNode.getPackageName() == null) {
            root.addChild(fileNode);
        } else {
            addFileToPackageHelper(root, fileNode, new StringTokenizer(fileNode.getPackageName(), "."));
        }
    }

    private void addFileToPackageHelper(StructureNode root, ProgramElementNode fileNode, StringTokenizer pkgPath) {
        if (pkgPath == null) {
            root.addChild(fileNode);
        } else {
            while (pkgPath.hasMoreElements()) {
                String nextElement = (String)pkgPath.nextElement();
                for (Iterator it = root.getChildren().iterator(); it.hasNext(); ) {
                    StructureNode pkgNode = (StructureNode)it.next();
                    if (pkgNode.getName().equals(nextElement)) {
                        if (!pkgPath.hasMoreElements()) {
                            pkgNode.addChild(fileNode);
                        } else {
                            addFileToPackageHelper(pkgNode, fileNode, pkgPath);
                        }
                    }
                }
            }
        }
    }

    private void buildRelationsPass(StructureNode node) {
        if (node instanceof ProgramElementNode) {
            ProgramElementNode peNode = (ProgramElementNode)node;
            List relations = new ArrayList();
            for (Iterator it = associations.iterator(); it.hasNext(); ) {
                Association association = (Association)it.next();
                relations.addAll(association.getRelationNodes(peNode.getAstObject()));
            }
            peNode.setRelations(relations);
        }
        List children = node.getChildren();
        if (children != null) {
            for (Iterator it = children.iterator(); it.hasNext(); ) {
                buildRelationsPass((StructureNode)it.next());
            }
        }
    }

    /**
     * @todo    make into "findNode" using callback
     */
    private StructureNode findFileNode(String filePath, StructureNode node) {
        if (node == null) return node;
        for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
            StructureNode child = (StructureNode)it.next();
            if (child instanceof ProgramElementNode &&
                ((ProgramElementNode)child).getSourceFilePath().replace('\\', '/').equals(filePath)) {
                return child;
            } else {
                findFileNode(filePath, child);
            }
        }
        return NO_STRUCTURE;
    }

    /**
     * TODO: avoid the childrenNodes list traversal?
     */
    private ProgramElementNode makeProgramElementNode(ASTObject astObject, Correspondences correspondences) {
        List children = new ArrayList();
        List childNodes = CompositionHierarchy.makeChildren(astObject, correspondences);
        for (Iterator it = childNodes.iterator(); it.hasNext(); ) {
            children.add(makeProgramElementNode((ASTObject)it.next(), correspondences));
        }
        return StructureNodeFactory.makeNode(astObject, null, children);
    }

    public void writeStructureModel(String configFilePath) {
        try {
            String filePath = genExternFilePath(configFilePath);
            ObjectOutputStream s = new ObjectOutputStream(new FileOutputStream(filePath));
            s.writeObject(model);
            s.flush();
        } catch (Exception e) {
            System.err.println("AJDE: could not write structure for: " + configFilePath);
            e.printStackTrace();
        }
    }

    public void readStructureModel(String configFilePath) {
        try {
            if (configFilePath == null) return;
            String filePath = genExternFilePath(configFilePath);
            FileInputStream in = new FileInputStream(filePath);
            ObjectInputStream s = new ObjectInputStream(in);
            model = (StructureModel)s.readObject();
        } catch (Exception e) {
            model.setRoot(NO_STRUCTURE);
        }
    }

    private String genExternFilePath(String configFilePath) {
        return configFilePath.substring(0, configFilePath.lastIndexOf(".lst")) + ".ajsym";
    }

    public static class UpdateStructurePass extends AbstractCompilerPass {
        private StructureManager structureManager = null;

        public UpdateStructurePass(JavaCompiler jc, StructureManager structureManager) {
            super(jc);
            this.structureManager = structureManager;
        }

        public String getDisplayName() { return "updating structure"; }

        public void transformWorld() {
            structureManager.buildStructureModel((AspectJCompiler)super.getCompiler());
        }
    }

    /**
     * Will not add InitializerDec(s) because these are assumed to be equivalent to
     * TypeDecs in the structure.
     */
    private static class CompositionHierarchy {
       static List makeChildren(ASTObject astObject, Correspondences correspondeces) {
            List children = new ArrayList();
            if (astObject instanceof CompilationUnit) {
                List definedTypes = ((CompilationUnit)astObject).getDefinedTypes();
                for (Iterator it = definedTypes.iterator(); it.hasNext(); ) {
                    TypeDec typeDec = (TypeDec)it.next();
                    children.add(typeDec);
                }
            } else if (astObject instanceof TypeDec) {
                TypeDec typeDec = (TypeDec)astObject;
                List decs = typeDec.getBody().getList();
                for (Iterator it = decs.iterator(); it.hasNext(); ) {
                    Dec bodyDec = (Dec)it.next();
                    if (bodyDec != null && !(bodyDec instanceof InitializerDec)) {
                        if (!bodyDec.isSynthetic()) {
                            if (bodyDec.getBeginLine() != bodyDec.getDeclaringType().getCorrespondingDec().getBeginLine()) {
                                children.add(bodyDec);
                            }
                        }
                    }
                }
            } else if (astObject instanceof MethodDec) {
                // HACK: this should not use the "affects" correspondeces db
                Set affectsSet = correspondeces.getAffects(astObject);
                if (affectsSet != null && affectsSet.size() > 0) {
                    for (Iterator it = affectsSet.iterator(); it.hasNext(); ) {
                        ASTObject stmtDec = (ASTObject)it.next();
                        children.add(stmtDec);
                    }
                }
            } else if (astObject instanceof AdviceDec) {
                // HACK: this should not use the "affects" correspondeces db
                Set affectsSet = correspondeces.getAffects(astObject);
                if (affectsSet != null && affectsSet.size() > 0) {
                    for (Iterator it = affectsSet.iterator(); it.hasNext(); ) {
                        ASTObject stmtDec = (ASTObject)it.next();
                        if (   (astObject.getSourceFileName().equals(stmtDec.getSourceFileName()))
                            && (astObject.getBeginLine() < stmtDec.getBeginLine())
                            && (astObject.getEndLine() > stmtDec.getEndLine())) {
                            children.add(stmtDec);
                        }
                    }
                }
            }
            return children;
        }
    }
}

