
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm.views;

import java.util.*;
import java.io.*;
import org.aspectj.tools.ide.*;
import org.aspectj.asm.associations.*;
import org.aspectj.asm.*;
import org.aspectj.compiler.crosscuts.AspectJCompiler;
//import org.aspectj.ajde.compiler.AjdeCompiler;

/**
 * @author Mik Kersten
 */
public class EmacsStructureManager extends StructureManager {

    private static final String EXTERN_FILE_SUFFIX = ".ajesym";

    public EmacsStructureManager() {
        super();
    }

    public void externalizeStructureView(final AspectJCompiler ajdeCompiler) {
        super.buildStructureModel(ajdeCompiler);
        if (!model.isValid()) return;
        try {
            Set fileSet = model.getFileMap().entrySet();
            for (Iterator it = fileSet.iterator(); it.hasNext(); ) {
                ProgramElementNode peNode = (ProgramElementNode)((Map.Entry)it.next()).getValue();
                dumpStructureToFile(peNode);
            }
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }

    private void dumpStructureToFile(ProgramElementNode node) throws IOException {
        String sourceName = node.getSourceFilePath();
        String fileName = sourceName.substring(0, sourceName.lastIndexOf(".")) + EXTERN_FILE_SUFFIX;
        BufferedWriter writer = new BufferedWriter(new FileWriter(new File(fileName)));
        new SExpressionPrinter(writer).printDecls(node);
        writer.flush();
    }

    /**
     * This class was not written in an OO style.
     */
    private static class SExpressionPrinter {

        private BufferedWriter writer = null;

        public SExpressionPrinter(BufferedWriter writer) {
            this.writer = writer;
        }

        private void printDecls(ProgramElementNode node) {
            print("(");
            for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
                // this ignores relations on the compile unit
                Object nodeObject = it.next();
                if (nodeObject instanceof ProgramElementNode) {
                    ProgramElementNode child = (ProgramElementNode)nodeObject;
                    printDecl(child, true);
                } else if (nodeObject instanceof LinkNode) {
                    LinkNode child = (LinkNode)nodeObject;
                    printDecl(child.getProgramElementNode(), false);
                }
            }
            print(") ");
        }

        private void printDecls(RelationNode node) {
            for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
                // this ignores relations on the compile unit
                Object nodeObject = it.next();
                if (nodeObject instanceof LinkNode) {
                    LinkNode child = (LinkNode)nodeObject;
                    if (//!child.getProgramElementNode().getKind().equals("stmnt") &&
                        !child.getProgramElementNode().getKind().equals("<undefined>")) {
                        printDecl(child.getProgramElementNode(), false);
//                        printDecl(child.getProgramElementNode(), false);
                    }
                }
            }
        }

        /**
         * @param   structureNode   can be a ProgramElementNode or a LinkNode
         */
        private void printDecl(ProgramElementNode node, boolean recurse) {
            String kind = node.getKind().toLowerCase();
            print("(");
            print("(" + node.getBeginLine() + " . " + node.getBeginColumn() + ") ");
            print("(" + node.getBeginLine() + " . " + node.getBeginColumn() + ") ");
            print(kind + " ");                                     //2

            // HACK:
            String displayName = node.toString().replace('\"', ' ');

            print("\"" + displayName + "\" ");
            if (node.getSourceFilePath() != null) {
                print("\"" + fixFilename(node.getSourceFilePath()) + "\"");  //4
            } else {
                print("nil");
            }
            if (node.getSignature() != null) {
                print("\"" + node.getDeclaringType() + "\" ");         //5
            } else {
                print("nil");
            }

            if (!recurse) {
                print("nil");
                print("nil");
                print("nil");
            } else {
                print("(");
                if (node instanceof ProgramElementNode) {
                    java.util.List relations = ((ProgramElementNode)node).getRelations();
                    if (relations != null) {
                        for (Iterator it = relations.iterator(); it.hasNext(); ) {
                            RelationNode relNode = (RelationNode)it.next();
                            if (relNode.getRelation().getAssociationName().equals(Advice.NAME) ||
                                relNode.getRelation().getAssociationName().equals(Introduction.NAME)) {
                                printDecls(relNode);                                   // 6
                            }
                        }
                    }
                }
                print(") ");
                print("(");
                print(") ");
                print("(");
                Iterator it3 = node.getChildren().iterator();
                if (it3.hasNext()) {
                    while (it3.hasNext()) {
                        // this ignores relations on the compile unit
                        Object nodeObject = it3.next();
                        if (nodeObject instanceof ProgramElementNode) {
                            ProgramElementNode currNode = (ProgramElementNode)nodeObject;
                            if (//!currNode.isStmntKind() &&
                                !currNode.getKind().equals("<undefined>")) {
                                printDecl(currNode, true);
                            }
                        }
                    }
                }
                print(") ");
            }

            print(node.getKind().equals("class") ? "t " : "nil ");        // 9
//            print(node.getKind().equals("introduction") ? "t " : "nil "); // 10
            print(node.getKind().equals("introduction") ? "nil " : "nil "); // 10
            print("nil ");            // 11
            print("nil ");       // 12
            print(")");
        }

        String fixFilename(String filename) {
            return subst("\\\\", "\\", filename);
        }

        private void print(String string) {
            try {
                writer.write(string + "\n");
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }

        private String subst(String n, String o, String in) {
            int pos = in.indexOf(o);
            if (pos == -1)
                return in;
            return in.substring(0, pos) +
                   n +
                   subst(n, o, (in.substring(pos + o.length())));
        }

        private void lose(Error e) {
            try {
                print("(ERROR \"" + e.toString() + "\")");
            }
            catch(Error ex) { }
        }
    }
}

