/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;

/**
 * To allow multiple compilers or for purposes of changing
 * compiler mode (like change of classpath or incremental/non-incremental
 * modes) we need to avoid usage of static fields in our classes,
 * because the only other way (usage of custom ClassLoader) may
 * conflict with different environment that uses AJ compiler
 * (like IDE that have it's own ClassLoader and security policy).
 *
 * So, each object, that needs access to 'global' data - like
 * compiler instance itself, compilation environment, type system
 * and so on needs a reference to root object. The root object is
 * the instance of compiler, that refers others instances of main
 * objects - World, TypeManager and Options.
 *
 * This class - is an abstract superclass of each non-root object
 * in our compiler, and it grants access to main objects and 'global'
 * data through the compiler instance.
 *
 * Note: CompilerObject class is declared to be abstract to disable
 * creation of instances of this class (even if it does not countaince
 * abstract methods). Also, it's methods are declared to be 'final'
 * to allow JIT compiler to inline them.
 */

public abstract class CompilerObject {
    /** The root of our 'global' data - the JavaCompiler instance */
    private final JavaCompiler compiler;

    /** The constructor of CompilerObject - requires compiler instance */
    public CompilerObject(JavaCompiler compiler) {
        this.compiler = compiler;
    }

    /** This method returns the compiler itself */
    public final JavaCompiler getCompiler() {
        return compiler;
    }

    /** This method returns World object - the root of
     *  misceleneous global data
     */
    public final World getWorld() {
        return compiler.getWorld();
    }

    /** This method returns TypeManager object - the root
     *  of all defined and loaded types. It also includes
     *  set of primitive and core types, and is responsible
     *  for types creation and managment.
     */
    public final TypeManager getTypeManager() {
        return compiler.getTypeManager();
    }

    /** This method returns current compiler settings (options) */
    public final Options getOptions() {
        return compiler.getOptions();
    }
    
    private SourceLocation mySource; 
    
    /** Convenience method for AST factory **/
    public AST getAST() {
        //XXX should be abstract
        if (mySource == null) mySource = new DummySourceLocation(compiler);
        return new AST(new SyntheticSourceLocation(this));
    }
}

