/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;

import java.util.*;

/**
 *  Deals with inner classes.
 *
 *  <ul>
 *
 *    <li>Adds a field to all inner classes (see {@link
 *    TypeDec#walkMemberMunger}).</li>
 *
 *    <li>Converts qualified this expressions into chains of
 *    references to those fields (see {@link
 *    QualifiedThisExpr#walkMemberMunger}).</li>
 *
 *    <li>Adds a new formal to constructors of inner classes and sets
 *    the value passed in to the class' field for constructors with
 *    super constructor calls (see {@link
 *    ConstructorDec#walkMemberMunger}).  </li>
 *
 * </ul>
 *
 * <p> The apropriate qualified this expressions (for new instance
 * expressions and the like) have already been set in the
 * InnerInfoPass. </p>
 *
 * */

public class MemberClassMunger extends WalkerPass {

    public MemberClassMunger(JavaCompiler compiler) {
        super(compiler);
    }

    public String getDisplayName() {
        return "munging inner classes";
    }

    private Stack types = new Stack();
    public void enterType(NameType t) {
        types.push(t);
    }
    public void leaveType() {
        types.pop();
    }
    public NameType currentType() {
        return (NameType) types.peek();
    }

    private List/*<FieldDec>*/ fields = new ArrayList();
    public void pushField(FieldDec dec) {
        fields.add(dec);
    }
    public void popField() {
        fields.remove(fields.size() - 1);
    }
    public FieldDec currentField() {
        return (FieldDec) fields.get(fields.size() - 1);
    }
    // possibly overkill, but will show us errors earlier
    public List saveFields() {
        List fields = this.fields;
        this.fields = new ArrayList();
        return fields;
    }
    public void restoreFields(List fields) {
        this.fields = fields;
    }

    public Expr buildExprFromEnclosing(Expr seed, NameType targetType) {
        return buildExpr(seed, targetType, fields.size() - 2);
    }
    public Expr buildExprFromThis(Expr seed, NameType targetType) {
        return buildExpr(seed, targetType, fields.size() - 1);
    }

    private Expr buildExpr(Expr seed, NameType targetType, int i) {
        final AST ast = getAST();
        while (true) {
            Type currType = seed.getType();
            if (currType == targetType) {
                return seed;
            } else {
                if (i < 0) {
                    throw new RuntimeException("Can't get to " + targetType
                                               + " from " + seed);
                }
                seed = ast.makeGet(seed, (FieldDec) fields.get(i--));
            }
        }
    }

    public ASTObject process(ASTObject object) {
        return object.walkMemberMunger(this);
    }
}

