/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.bytecode.*;

import java.util.*;
import java.io.*;

public final class TypeManager extends CompilerObject {
    protected SourceLocation dummySource = new DummySourceLocation(getCompiler());
    public final TypeDec TYPE_DEC_NOT_FOUND = new ClassDec(dummySource, 
        new Modifiers(dummySource,0),
        "*NF*",null, null, new Decs(dummySource));
    
    public final NameType TYPE_NOT_FOUND = new NotFoundType(TYPE_DEC_NOT_FOUND);
    {
        TYPE_DEC_NOT_FOUND.setType(TYPE_NOT_FOUND);
        //((NameType)TYPE_NOT_FOUND).builtTypeGraph = true;
        //((NameType)TYPE_NOT_FOUND).isFinished = true;
    }
    
    private ClassPathManager classPathManager;
    private Map packages = new HashMap();
    private PackageSO defaultPackage;
    private Collection loadedTypes = new HashSet();
    
    public TypeManager(JavaCompiler compiler) {
        super(compiler);
        Options options = compiler.getOptions();
        if (options.classpath == null) options.classpath = options.defaultClasspath;
        classPathManager =
          new ClassPathManager(compiler, 
                               options.classpath, options.bootclasspath, 
                               options.extdirs);
		defaultPackage = new PackageSO(compiler, null, classPathManager);
        //astConnection = new ASTConnection(compiler);
    }
    
    private Collection allVisibleTypeNames = null;
    
    /** return all type names that are visible, meaning passed in, or on the classpath.
     * 
     * This returns a name twice if shadowing occurs on the classpath.
     * This must only be called after {@link World#getTypes()}.
     */
    public Collection getAllVisibleTypeNames() {
		if (allVisibleTypeNames == null) {
			allVisibleTypeNames = classPathManager.getAllPossibleTypeNames();
			for (Iterator i = getWorld().getTypes().iterator(); i.hasNext();) {
				TypeDec t = (TypeDec) i.next();
				allVisibleTypeNames.add(t.getType().getNamePieces());
			}
			for (Iterator i = getPrimitiveTypeMap().values().iterator(); i.hasNext(); ) {
				Type t = (Type)i.next();
				allVisibleTypeNames.add(t.getNamePieces());
			}
		}
    	return allVisibleTypeNames;
    }
    
    /**
     * returns all Types that have been loaded by this TypeManager
     */
    public Collection getLoadedTypes() {
    	return loadedTypes;
    }
    
        
    public Type getType(String packageName, String className) {
        Type type = findType(packageName, className);
        if (type == null) {
            getCompiler().showError("missing type " + packageName + '.' + className);
            return TYPE_NOT_FOUND;
        } else {
            return type;
        }
    }

    public synchronized Type findType(String packageName, String className) {
        //System.out.println("finding " + packageName + "::" + className);
        PackageSO p = findPackage(packageName);
        Type ret = p.findType(className);
        if (ret != null) loadedTypes.add(ret);
        //if (ret == null) { System.out.println("   NOT FOUND"); }
        return ret;
    }
    
    public void addType(Type type) {
        //System.out.println("adding " + type.getPackageName() + "::" + type.getId());
        
        PackageSO p = findPackage(type.getPackageName());
        p.addType(type);
        loadedTypes.add(type);
    }

    PackageSO findTopLevelPackage(String name) {
        PackageSO p = (PackageSO)packages.get(name);
        if (p != null) return p;

        p = new PackageSO(getCompiler(), name,
                        classPathManager.makeSubPackageManager(name));
        //getCompiler().showMessage("made top package: " + name);
//         System.out.println("sub: " + classPathManager + ", " + 
//                            classPathManager.makeSubPackageManager(name));
        packages.put(name, p);
        return p;
    }
    
    public PackageSO findPackage(String packageName) {
        if (packageName == null) return defaultPackage;

        PackageSO p = (PackageSO)packages.get(packageName);
        if (p != null) return p;
        
        String[] names = splitDottedName(packageName);
        p = findTopLevelPackage(names[0]);
        for (int i=1; i<names.length; i++) {
            //System.out.println(p + ": " + names[i]);
            p = p.findPackage(names[i]);
        }
        //getCompiler().showMessage("adding " + packageName);

        packages.put(packageName, p);
        return p;
    }
    
    String[] splitDottedName(String name) {
        //getCompiler().showMessage("splitting: " + name);
        int nameCount = 0;
        int lastDot=-1;
        do {
            lastDot = name.indexOf('.', lastDot+1);
            nameCount++;
        } while (lastDot != -1);

        String[] names = new String[nameCount];
        lastDot = 0;
        int index=0;
        while (true) {
            int nextDot = name.indexOf('.', lastDot);
            if (nextDot == -1) {
                names[index] = name.substring(lastDot);
                break;
            } else {
                names[index++] = name.substring(lastDot, nextDot);
                lastDot = nextDot+1;
            }
        }

        return names;
    }
    
    /** Check that specified type exists in classpath and is loadable */
    public boolean checkLoadable(String packageName, String className) {
        PackageSO p = findPackage(packageName);
        return p.findTypeOnClassPath(className) != null;
    }

    
	private final Map primitiveTypeMap = new HashMap();
    
    public final Type nullType = new NullType(getCompiler());
    public final Type voidType = new VoidType(getCompiler());
    { primitiveTypeMap.put("void", voidType); }
    public final Type anyType = TYPE_NOT_FOUND;// new AnyType(getCompiler());

    public final PrimitiveType intType     = new IntType(getCompiler());
    { primitiveTypeMap.put("int", intType); }
    public final PrimitiveType byteType    = new ByteType(getCompiler());
    { primitiveTypeMap.put("byte", byteType); }
    public final PrimitiveType shortType   = new ShortType(getCompiler());
    { primitiveTypeMap.put("short", shortType); }
    public final PrimitiveType longType    = new LongType(getCompiler());
    { primitiveTypeMap.put("long", longType); }
    public final PrimitiveType doubleType  = new DoubleType(getCompiler());
    { primitiveTypeMap.put("double", doubleType); }
    public final PrimitiveType floatType   = new FloatType(getCompiler());
    { primitiveTypeMap.put("float", floatType); }
    public final PrimitiveType charType    = new CharType(getCompiler());
    { primitiveTypeMap.put("char", charType); }
    public final PrimitiveType booleanType = new BooleanType(getCompiler());
    { primitiveTypeMap.put("boolean", booleanType); }


	public Type findPrimitiveType(String id) {
		return (Type)getPrimitiveTypeMap().get(id);
	}

	public Map getPrimitiveTypeMap() {
		return primitiveTypeMap;
	}
	

    public final NameType getObjectType() {
        return (NameType)getType("java.lang", "Object");
    }
    public final NameType getStringType() {
        return (NameType)getType("java.lang", "String");
    }
    public final NameType getClassType() {
        return (NameType)getType("java.lang", "Class");
    }
    public final NameType getCloneableType() {
        return (NameType)getType("java.lang", "Cloneable");
    }
    public final NameType getSerializableType() {
        return (NameType)getType("java.io", "Serializable");
    }
    public final NameType getRuntimeExceptionType() {
        return (NameType)getType("java.lang", "RuntimeException");
    }
    public final NameType getErrorType() {
        return (NameType)getType("java.lang", "Error");
    }
    public final NameType getThrowableType() {
        return (NameType)getType("java.lang", "Throwable");
    }
    public final NameType getExceptionType() {
        return (NameType)getType("java.lang", "Exception");
    }
    public final NameType getStringBufferType() {
        return (NameType)getType("java.lang", "StringBuffer");
    }

    //ASPECT introduce from crosscuts
    public final Type getJoinPointType() {
        return getType("org.aspectj.lang", "JoinPoint");
    }    
    public final Type getJoinPointStaticPartType() {
        return getType("org.aspectj.lang", "JoinPoint$StaticPart");
    }    
	public final Type getAroundClosureType() {
		return getType("org.aspectj.runtime.internal", "AroundClosure");
	}
	public final Type getConversionsType() {
		return getType("org.aspectj.runtime.internal", "Conversions");
	}

//    public final NameType objectType             = new NameType(getCompiler(), null);
//    public final NameType stringType             = new NameType(getCompiler(), null);
//    public final NameType cloneableType          = new NameType(getCompiler(), null);
//    public final NameType runtimeExceptionType   = new NameType(getCompiler(), null); 
//    public final NameType errorType              = new NameType(getCompiler(), null);
        
    
    /**
     * Get bytecode for predefined types, except those were compiled
     */
//    public void fixPredefinedTypes() {
//        fixPredefinedType("java.lang.Object");
//        fixPredefinedType("java.io.Serializable");
//        fixPredefinedType("java.lang.Comparable");
//        fixPredefinedType("java.lang.String");
//        fixPredefinedType("java.lang.Cloneable");
//        fixPredefinedType("java.lang.Throwable");
//        fixPredefinedType("java.lang.Exception");
//        fixPredefinedType("java.lang.RuntimeException");
//        fixPredefinedType("java.lang.Error");
//    }
//
//    private void fixPredefinedType(String fullName) {
//        TypeDec typeDec = (TypeDec)typeDecCache.get(fullName);
//        if (typeDec.getBody() == null) {
//            internalFindTypeDec(fullName);
//            typeDec.rebuildTypeGraph();
//        }
//    }

    public Type unaryNumericPromotion(Type ty) {
        if (ty == doubleType || ty == floatType || ty == longType) {
            return ty;
        }
        return intType;
    }

    public Type binaryNumericPromotion(Type ty1, Type ty2) {
        if (ty1 == doubleType || ty2 == doubleType) {
            return doubleType;
        }
        if (ty1 == floatType || ty2 == floatType) {
            return floatType;
        }
        if (ty1 == longType || ty2 == longType) {
            return longType;
        }
        return intType;
    }
}

