/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 *
 * @grammar x += y
 */
public class AddAssignExpr extends AssignExpr {

    public Type discoverType() {
        Type ty1 = lhs.getType();
        Type ty2 = rhs.getType();
        if (! ((ty1.isString() && !ty2.isVoid())
               || ((ty1 instanceof NumericType)
                   && (ty2 instanceof NumericType)))) {
            showOperatorTypeError(op, ty1, ty2);
        }
        return ty1;
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        cgValueEffect(cb, true);
    }
    protected void cgEffect(CodeBuilder cb) {
        cgValueEffect(cb, false);
    }
    private void cgValueEffect(CodeBuilder cb, boolean needsValue) {
        AssignableExpr lhs = getLhs();
        Expr rhs = getRhs();
        Type lhsTy = lhs.getType();
        if (lhsTy.isString()) {
            NameType stringBufferType = getTypeManager().getStringBufferType();
            lhs.cgLvalue(cb);
            lhs.cgDupLvalue(cb);
            lhs.cgLtoRvalue(cb);
            cb.emitNEW(stringBufferType);
            cb.emitDUP();
            cb.emitINVOKESPECIAL(stringBufferType, "<init>", "()V", -1);
            cb.emitSWAP();
            cb.emitINVOKEVIRTUAL(stringBufferType, "append",
                                 "(Ljava/lang/String;)Ljava/lang/StringBuffer;",
                                 -1);
            rhs.cgBuffer(cb);
            cb.emitINVOKEVIRTUAL(stringBufferType, "toString", "()Ljava/lang/String;", 0);
            if (needsValue) lhs.cgDupRvalue(cb);
            lhs.cgAssignment(cb);
        } else if ((rhs instanceof IntLiteralExpr)
                   && lhsTy.hasFastIncOp(lhs, ((IntLiteralExpr) rhs).getIntValue())) {
            lhsTy.emitFastIncOp(cb, lhs, ((IntLiteralExpr) rhs).getIntValue());
            if (needsValue) lhs.cgValue(cb);
        } else {
            Type rhsTy = rhs.getType();
            Type ty = getTypeManager().binaryNumericPromotion(lhsTy, rhsTy);
            lhs.cgLvalue(cb);
            lhs.cgDupLvalue(cb);
            lhs.cgLtoRvalue(cb);
            lhsTy.emitCast(cb, ty);
            rhs.cgValue(cb, ty);
            ty.emitAdd(cb);
            ty.emitCast(cb, lhsTy);
            if (needsValue) lhs.cgDupRvalue(cb);
            lhs.cgAssignment(cb);
        }
    }

    //BEGIN: Generated from @child and @property

    public AddAssignExpr(SourceLocation location, AssignableExpr _lhs, String _op, Expr _rhs) {
        super(location, _lhs, _op, _rhs);

    }
    protected AddAssignExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        AddAssignExpr ret = new AddAssignExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (lhs != null) ret.setLhs( (AssignableExpr)walker.process(lhs) );
        ret.op = op;
        if (rhs != null) ret.setRhs( (Expr)walker.process(rhs) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "AddAssignExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
