/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;


import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar expr[index]
 * @child Expr expr      the expression
 * @child Expr index     the index
 */
public class ArrayExpr extends AssignableExpr {

    protected Type discoverType() {
        Type intType = getTypeManager().intType;
        if (! index.isAssignableTo(intType)) {
            index.showTypeError(index.getType(), intType);
        }
        Type exprType = expr.getType();

        if (exprType.isAnyType()) return getTypeManager().anyType;

        if (exprType instanceof ArrayType) {
            return ((ArrayType)exprType).getComponentType();
        } else {
            expr.showError("array required, but " + exprType.getString() + " found");
            return getTypeManager().anyType;
        }
    }

    public void unparse(CodeWriter writer) throws IOException {
        writer.write(expr);
        writer.write('[');
        writer.write(getIndex());
        writer.write(']');
    }

    // ------------------------------
    // bcg
    protected void cgLvalue(CodeBuilder cb) {
        getExpr().cgValue(cb);
        getIndex().cgValue(cb);
    }
    protected void cgLtoRvalue(CodeBuilder cb) {
        getType().emitAload(cb);
    }
    protected void cgAssignment(CodeBuilder cb) {
        getType().emitAstore(cb);
    }
    protected void cgDupLvalue(CodeBuilder cb) {
        // array, index --> array, index, array, index
        cb.emitDUP2();
    }
    protected void cgDupRvalue(CodeBuilder cb) {
        // array, index, val --> val, array, index, val
        getType().emitDupX2(cb);
    }
    // end bcg

    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    protected Expr index;
    public Expr getIndex() { return index; }
    public void setIndex(Expr _index) {
        if (_index != null) _index.setParent(this);
        index = _index;
    }

    public ArrayExpr(SourceLocation location, Expr _expr, Expr _index) {
        super(location);
        setExpr(_expr);
        setIndex(_index);
    }
    protected ArrayExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ArrayExpr ret = new ArrayExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        if (index != null) ret.setIndex( (Expr)walker.process(index) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        case 1: return index;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        case 1: return "index";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        case 1: setIndex((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "ArrayExpr()";
    }

    //END: Generated from @child and @property
}
