/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;

/**
 */
public class ArrayType extends RefType {
    private Type componentType;
    private int arrayDimCount;

    public ArrayType(JavaCompiler compiler, Type componentType) {
        super(compiler);
        this.componentType = componentType;
        this.arrayDimCount = 1 + componentType.getArrayDimCount();
    }

    public Type getComponentType() {
        return componentType;
    }
    public int getArrayDimCount() {
        return arrayDimCount;
    }

    public Type getBaseComponentType() {
        return getComponentType().getBaseComponentType();
    }


    /* The next five methods overlap too much with NameType */
    public Expr getClassExpr() {
        return getAST().makeLiteral(this); //new ClassExpr(getCompiler(), this.makeTypeD());
    }

    public Expr fromObject(Expr expr) {
        return getAST().makeCast(this, expr); //new CastExpr(getCompiler(), this.makeTypeD(), expr);
    }

    public Expr makeObject(Expr expr) { return expr; }
    public Expr getNullExpr() { return getAST().makeNull(); }
    public Type getRefType() { return this; }



    private final FieldDec lengthFieldDec =
        getAST().makeField(getAST().makeModifiers(Modifiers.PUBLIC),
                getTypeManager().intType, "length", null);
    {
        lengthFieldDec.setAllEnclosingTypes(this);
    }
    private final Field lengthField = (Field)lengthFieldDec.getCorrespondingSemanticObject();


    public Field getField(String id, ASTObject fromWhere, boolean showError) {
        if (id.equals("length")) { return lengthField; }
        if (showError) {
            getCompiler().showError(fromWhere, "no field " + id + " on arrays");
        }
        return null; //getWorld().FIELD_DEC_NOT_FOUND;
    }

    private final MethodDec cloneMethodDec =
        getAST().makeMethod(getAST().makeModifiers(Modifiers.PUBLIC),
                getTypeManager().getObjectType(), "clone", getAST().makeFormals(), null);
    {
        cloneMethodDec.setAllEnclosingTypes(this);
    }
    private final Method cloneMethod = (Method)cloneMethodDec.getCorrespondingSemanticObject();

    public Method getMethod(String id, ASTObject fromWhere, Exprs params,
                               boolean showError)
    {
        //XXX might want to improve type names in error messages
        if (id.equals("clone") && params.size() == 0) {
            return cloneMethod;
        }
        return getTypeManager().getObjectType().getMethod(id, fromWhere, params, showError);
    }

    public SemanticObject findMatchingSemanticObject(SemanticObject o) {
        if (o == cloneMethod) return cloneMethod;
        else if (o == lengthField) return lengthField;
        else return getTypeManager().getObjectType().findMatchingSemanticObject(o);
    }

    {
        addDirectSuperType(getTypeManager().getObjectType());
    }


    public String getString() {
        return componentType.getString()+"[]";
    }

    public String toShortString() {
        return componentType.toShortString()+"[]";
    }

    public void unparse(CodeWriter writer) {
        writer.write(componentType.getString());
        writer.write("[]");
    }

    //XXX we don't have unique array types yet
    public boolean isEquivalent(Type _other) {
        if (!(_other instanceof ArrayType)) return false;
        return ((ArrayType)_other).componentType.isEquivalent(componentType);
    }


    public boolean isSubtypeOf(Type other) {
        return other.isObject()
            || other == getTypeManager().getCloneableType()
            || other == getTypeManager().getSerializableType();
    }

    public boolean isAssignableFrom(Type _other) {
        if (_other.isAnyType()) return true;
        if (_other instanceof NullType) return true;

        if (!(_other instanceof ArrayType)) return false;
        ArrayType other = (ArrayType)_other;
        if (this.isEquivalent(other)) return true;
        // pass the question on to component types iff they're reference types
        if (componentType.isReferenceType() && other.getComponentType().isReferenceType()) {
            return componentType.isAssignableFrom(other.getComponentType());
        } else {
            return false;
        }
    }

    public boolean isCloneable() { return true; }

    // ------------------------------
    // bcg

    void emitMultiNewarray(CodeBuilder cb, int filledDims) {
        cb.emitMULTIANEWARRAY(this, filledDims);
    }

    // ------------------------------
    // vm internals
    private String externalName;
    private String internalName;
    private String descriptor;

    public synchronized String getExternalName() {
        if (externalName == null) {
            externalName = getDescriptor().replace('/', '.');
        }
        return externalName;
    }

    public String getInternalName() {
        return getDescriptor();
    }

    public synchronized String getDescriptor() {
        if (descriptor == null) {
            descriptor = "[" + getComponentType().getDescriptor();
        }
        return descriptor;
    }
}
