/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 *
 * @grammar
 * @property String op
 * @child Expr rhs      the new value
 */
public abstract class AssignExpr extends BangExpr {

    //INTRO from AccessFixer
    public ASTObject fixAccessToFieldSet(FieldAccessExpr expr) {
            return makeInPlaceSet(expr, getOp(), getRhs());
    }

    public void unparse(CodeWriter writer) {
        writer.write(lhs);
        writer.writeOp(op + "=");
        writer.write(rhs);
    }

    public Expr makeReference() {
        if (lhs instanceof VarExpr) return lhs.makeReference();
        return super.makeReference();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    // this is big and dumb, but careful.  And overridden in BasicAssignExpr.
    public void walkFlow(FlowCheckerPass w) {
        w.process(getLhs());
        w.process(getRhs());
        if (getLhs() instanceof VarExpr) {
            VarExpr lhs = (VarExpr)getLhs();
            VarDec dec = lhs.getVarDec();
            if (dec.isFinal()
                && ((! dec.isBlank())
                    || (! w.getVars().isDefinitelyUnassigned(dec)))) {
                w.showVarError(this, dec, "Final variable " + dec.getId() +
                               " already has a value");
            }
        } else if (getLhs() instanceof FieldAccessExpr) {
            FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
            FieldDec dec = lhs.getFieldDec();
            if (dec.isFinal()
                && ((! dec.isBlank())
                    || (! w.getVars().isDefinitelyUnassigned(dec)))) {
                w.showVarError(this, dec, "Final field " + dec.getId() +
                               " already has a value");
            }
        }
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        // ???
        Expr lhs = getLhs();
        Type lhsType = lhs.getType();
        Type rhsType = rhs.getType();

        return this;
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        if (! (getLhs() instanceof FieldAccessExpr)) return this;

        FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
        Expr q = lhs.getExpr();
        FieldDec dec = lhs.getFieldDec();
        if (w.isAccessible(dec, q)) return this;

        final AST ast = getAST();
        Type qType = q.getType();
        Expr rhs = getRhs();
        MethodDec newMethodDec = w.getAccessMethod(qType, dec, getOp(), this);
        Exprs newArgs = ast.makeExprs(rhs);
        Expr newExpr = w.makeOutsidePrimary(dec.isStatic(), newArgs, q);
        return ast.makeCall(newMethodDec, newExpr, newArgs);
    }

    // move this down.
    public MethodDec buildAccessMethod(InnerAccessFixer w) {
        final AST ast = getAST();
        FieldAccessExpr lhs = (FieldAccessExpr) getLhs();
        Expr q = lhs.getExpr();
        Type qType = q.getType();
        FieldDec dec = lhs.getFieldDec();
        Type fieldType = dec.getType();

        FormalDec valFormal = ast.makeFormal(fieldType, "val$");
        Formals newFormals = ast.makeFormals(valFormal);
        Expr newExpr = w.makeInsidePrimary(dec.isStatic(), newFormals, qType);

        return
            w.makeAccessMethod(fieldType,
                               newFormals,
                               ast.makeSet(ast.makeGet(newExpr, dec),
                                           getOp(),
                                           ast.makeVar(valFormal)));
    }

    // ------------------------------
    // static factory

    public static AssignExpr build(SourceLocation source,
                                   AssignableExpr rand1, String op, Expr rand2) {
        op = op.intern();
        if        (op == "") {
            return new BasicAssignExpr(source, rand1, op, rand2);
        } else if (op == "+") {
            return new AddAssignExpr(source, rand1, op, rand2);
        } else if (op == "*" || op == "/" || op == "%" || op == "-") {
            return new NumericAssignExpr(source, rand1, op, rand2);
        } else if (op == "&" || op == "|" || op == "^") {
            return new BitwiseAssignExpr(source, rand1, op, rand2);
        } else if (op == "<<" || op == ">>" || op == ">>>") {
            return new ShiftAssignExpr(source, rand1, op, rand2);
        } else {
            throw new RuntimeException("bad op " + op);
        }
    }

    //BEGIN: Generated from @child and @property
    protected String op;
    public String getOp() { return op; }
    public void setOp(String _op) { op = _op; }

    protected Expr rhs;
    public Expr getRhs() { return rhs; }
    public void setRhs(Expr _rhs) {
        if (_rhs != null) _rhs.setParent(this);
        rhs = _rhs;
    }

    public AssignExpr(SourceLocation location, AssignableExpr _lhs, String _op, Expr _rhs) {
        super(location, _lhs);
        setOp(_op);
        setRhs(_rhs);
    }
    protected AssignExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 1: return rhs;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 1: return "rhs";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 1: setRhs((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "AssignExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
