/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;
import org.aspectj.compiler.base.ExceptionFinder;

import org.aspectj.compiler.base.*;

import java.util.*;

import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.MethodBuilder;
import org.aspectj.compiler.base.bcg.ClassfileBuilder;

/**
  * @grammar ...
  * @child Modifiers modifiers
  * @child Formals formals
  * @child TypeDs throws
  * @child CodeBody body
  */
public abstract class CodeDec extends Dec implements PossibleSoftThrowable {
    private boolean _isSoft = false;
    public boolean isSoftThrowable() { return _isSoft; }
    public void setSoftThrowable() { _isSoft = true; }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public final void walkFlow(FlowCheckerPass w) {
        if (getBody() != null) {
            setupFlowWalker(w);
            w.setLive(true);
            w.process(getBody());

            if (! getResultType().isVoid() && w.isLive()) {
                showError("missing return statement");
            }
        }
    }

    /** set up the flow walker, by adding implicit variables like
        thisJoinPoint, etc */

    protected void setupFlowWalker(FlowCheckerPass w) { }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public final void walkCleanup(ByteCodeCleanupPass w) {
        w.setLive(true);
        super.walkCleanup(w);
    }

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getBody() != null
                && getResultType().isVoid()
                && w.isLive()) {
            ReturnStmt s =
                (ReturnStmt) getAST().makeReturn(null).setSource(getBody());
            s.setCompletesNormally(false);
            getBody().getStmts().add(s);
            getBody().setCompletesNormally(false);
        }
        return this;
    }

    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inMember(isStatic());
        super.walkInnerInfo(w);
        w.restoreContext(context);
    }
    // end intro

    public static CodeBody blockToBody(BlockStmt block) {
        if (block == null) return null;
        block.getStmts().clearParent();
        CodeBody body = new CodeBody(block.getSourceLocation(),block.getStmts(),true);
        body.setSource(block);
        return body;
    }

    public void setBody(BlockStmt block) {
        setBody(blockToBody(block));
    }


    public abstract String getId();
    public abstract TypeD getResultTypeD();

    //XXX ugly
    public Type getResultType() {
        return getResultTypeD().getType();
    }

    public boolean canThrow(Type t) {
        if (t.isUncheckedThrowable()) return true;
        if (_throws == null) return false;
        for (int i=0; i<_throws.size(); i++) {
            if (t.isSubtypeOf(_throws.get(i).getType())) return true;
        }
        return false;
    }
    
    /**
     * returns the checked exceptions that the context expects could be thrown here
     * 
     * guarantees that if x is a member then no subtype(x) is a member
     * 
     * XXX figure out what the performance issues are and if we should cache this
     */
    public Set getPossibleCheckedExceptions() {
        if (_throws == null) return Collections.EMPTY_SET;
        Set ret = new HashSet();
        
        for (int i=0; i<_throws.size(); i++) {
        	ret.add(_throws.get(i).getType());
        }
        
        return Type.filterTopTypes(ret);
    }

    public void checkSpec() {
        // checks that all entries in 'throws' are subtypes of Throwable
        if (_throws == null) return;
        Type throwableType = getTypeManager().getThrowableType();
        for (int i=0; i<_throws.size(); i++) {
            TypeD throwsTypeD = _throws.get(i);
            if (!throwsTypeD.getType().isSubtypeOf(throwableType)) {
                throwsTypeD.showTypeError(throwsTypeD.getType(), throwableType);
            }
        }
    }

    public boolean appearsStaticToCaller() { return isStatic(); }


    public boolean hasThis() {
        return !isStatic();
    }

    public boolean isAlmostApplicable(Exprs params) {
        return params.size() == getFormals().size();
    }

    public boolean isApplicable(Exprs parameters) {
        return getFormals().canBeCalledWith(parameters);
    }

    public boolean conflictsWith(Dec otherDec) {
        if (!getId().equals(otherDec.getId())) return false;
        if (!this.isAccessible(otherDec) && !otherDec.isAccessible(this)) return false;

        if (!(otherDec instanceof CodeDec)) return false;

        CodeDec otherCodeDec = (CodeDec)otherDec;

//        System.out.println("matching formals: " + getFormals().toShortString() + " with " +
//            otherCodeDec.getFormals().toShortString() + " == " +
//            getFormals().matches(otherCodeDec.getFormals()));

        return getFormals().matches(otherCodeDec.getFormals());
    }


    public boolean isMoreSpecificThan(Dec otherDec) {
        if (!(otherDec instanceof CodeDec)) return false;

        CodeDec other = (CodeDec)otherDec;

        if (!super.isMoreSpecificThan(other)) return false;
        // is m1 more specific than m2?

        // require that A1i is method convertable to A2i
        Formals formals1 = getFormals();
        Formals formals2 = other.getFormals();
        if (formals1.size() != formals2.size()) return false;
        for(int i=0; i<formals1.size(); i++) {
            if (!formals1.get(i).getType().isMethodConvertableTo(formals2.get(i).getType())) {
                return false;
            }
        }

        return true;
    }

    //INTRO from ScopePass
    public void preScope(ScopeWalker walker) {
        walker.pushScope(makeBlockScope(walker));
    }
    public ASTObject postScope(ScopeWalker walker) { walker.popScope(); return this; }

    //XXX yet another non-thread-safe hack
    boolean computingMinimalThrows = false;
    public void computeMinimalThrows() {
        //getCompiler().showMessage("compute minimal throws for: " + this.getId());
        //System.out.println("compute minimal throws for: " + this.toShortString());
        //this.display(2);
        if (computingMinimalThrows || getBody() == null) return;

        computingMinimalThrows = true;
        Set exceptions = ExceptionFinder.getPossibleExceptions(getBody(), false);
        //System.out.println(exceptions);
        computingMinimalThrows = false;

        final int N = exceptions.size();
        if (N == 0) {
            setThrows(null);
            return;
        }

        exceptions = Type.filterTopTypes(exceptions);

        TypeDs newThrows = getAST().makeTypeDs(); //new TypeDs(getCompiler());
        for(Iterator iter = exceptions.iterator(); iter.hasNext(); ) {
            Type type = (Type)iter.next();
            newThrows.add(type.makeTypeD());
        }
        setThrows(newThrows);
    }

//    private Set callExprs; // = new HashSet();
    public Set getCallExprs() {
//        return callExprs;
        return null;
    }
//
//    public void noteCallExpr(Expr callExpr) {
//        if (callExprs == null)
//            callExprs = new HashSet();
//        callExprs.add(callExpr);
//    }
//
//    public void cleanupCrossRefs() {
//        callExprs = null;
//    }

    public void cleanup() {
        //callExprs = null;
        if (body != null)
            body.cleanup();
        super.cleanup();
    }

    // ------------------------------
    // Intro:  FrameLocPass

    private int frameSize;
    public int getFrameSize() { return frameSize; }
    public void setFrameSize(int frameSize) { this.frameSize = frameSize; }

    public void walkFrameLoc(FrameLocPass walker) {
        walker.setfs((!isStatic() || this instanceof ConstructorDec) ?
                     1 : 0);
        super.walkFrameLoc(walker); // this will walk the formals first.
        setFrameSize(walker.getmaxfs());
    }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walker.enterCodeDec(this);
        walker.inCode(isStatic());
        this.walk(walker);
        walker.leaveCodeDec();
    }

    // ------------------------------
    // bcg

    /** This method takes care of generating code for Member decs.  At
        this point, we have a number of invariants:

        <ul>
          <li> implicit return statements are inserted</li>

          <li> all return statements are typechecked, with appropriate
               casts inserted, so we know what kind of return bytecode to
               generate based on the return stmt rather than on the return
               type of the method.</li>
        </ul>
    */
    protected void cgMember(ClassfileBuilder cfb) {
        if (getOptions().bcgverbose) System.err.println("generating for " + this);
        MethodBuilder mb = cfb.getMethodBuilder();
        mb.addAccessFlags(getModifiers().getValue());
        mb.setName(getInternalId());
        mb.setDescriptor(getDescriptor());
        if (isDeprecated()) mb.setDeprecated();
        //XXX get this right -- if (isSynthetic())  mb.setSynthetic();
        TypeDs exns = getThrows();
        if (exns != null) {
            for (Iterator i = exns.iterator(); i.hasNext(); ) {
                TypeD typeD = (TypeD)i.next();
                mb.addToExceptionsAttribute((NameType)typeD.getType());
            }
        }

        if (getBody() != null) {
            CodeBuilder cb = mb.getCodeBuilder();
            cgCodeMember(cb);
            mb.setCode(cb);
        }
        cfb.addMethod(mb);
    }

    /** generate code for the body.  Assumes there is a body.

        @param cb the codeBuilder into which to generate code
    */
    protected void cgCodeMember(CodeBuilder cb) {
        throw new RuntimeException("unsupported " + this);
    }

    /** return the bytecode type descriptor for the member. */
    public String getDescriptor() {
        throw new RuntimeException("unsupported " + this);
    }

    /** return the change in stack that would result from invoking
        this codeDec.  This is the size of the return value minus the
        size of the argument values (including the implicit this
        object).  It is typically negative for coded members with
        arguments. */
    public int getStackDelta() {
        throw new RuntimeException("unsupported " + this);
    }

    String getInternalId() {
        return getBytecodeId();
    }

    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }

    protected Formals formals;
    public Formals getFormals() { return formals; }
    public void setFormals(Formals _formals) {
        if (_formals != null) _formals.setParent(this);
        formals = _formals;
    }

    protected TypeDs _throws;
    public TypeDs getThrows() { return _throws; }
    public void setThrows(TypeDs __throws) {
        if (__throws != null) __throws.setParent(this);
        _throws = __throws;
    }

    protected CodeBody body;
    public CodeBody getBody() { return body; }
    public void setBody(CodeBody _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    public CodeDec(SourceLocation location, Modifiers _modifiers, Formals _formals, TypeDs __throws, CodeBody _body) {
        super(location);
        setModifiers(_modifiers);
        setFormals(_formals);
        setThrows(__throws);
        setBody(_body);
    }
    protected CodeDec(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return formals;
        case 2: return _throws;
        case 3: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "formals";
        case 2: return "throws";
        case 3: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setFormals((Formals)child); return;
        case 2: setThrows((TypeDs)child); return;
        case 3: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 4;
    }

    public String getDefaultDisplayName() {
        return "CodeDec()";
    }

    //END: Generated from @child and @property
}
