/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import org.aspectj.compiler.base.LocalClassPass;

/**
  * @grammar enclosingInstanceExpr.[this|super](args)
  * @child Expr enclosingInstanceExpr
  * @property boolean isSuper
  * @child Exprs args
  * @property Constructor constructor
  */

public class ConstructorCallExpr extends Expr {
    public ConstructorDec getConstructorDec() { return (ConstructorDec)constructor.getCorrespondingDec(); }
    public void setConstructorDec(ConstructorDec md) {
        setConstructor((Constructor)md.getCorrespondingSemanticObject());
    }

    public boolean inStaticContext() { return true; }


    // ------------------------------
    // INTRO from SpecCheckerPass

    public void checkSpec() {
        if (getEnclosingInstanceExpr() != null) {
            NameType ty = (NameType) getConstructorDec().getDeclaringType();
            if (! ty.isInner()) {
                showError("Class " + ty.getId() + " is not an inner class");
            }
            Type exprTy = getEnclosingInstanceExpr().getType();
            NameType enclosingTy = ty.getEnclosingType();
            if (! exprTy.isSubtypeOf(enclosingTy)) {
                showTypeError(exprTy, enclosingTy);
            }
        }
        super.checkSpec();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        w.process(getEnclosingInstanceExpr());
        w.process(getArgs());
        if (getConstructor().getThrows() != null) {
            TypeDs ts = getConstructor().getThrows();
            for (int i = 0, len = ts.size(); i < len; i++) {
                w.setExns(w.getExns().add((NameType) ts.get(i).getType()));
            }
        }
    }

    //INTRO from ScopeWalker
    public ASTObject postScope(ScopeWalker walker) {
        if (constructor != null) return this;

        Type myType = getDeclaringType();
        if (isSuper) {
            myType = myType.getTypeDec().getSuperClassType();
        }
        constructor = myType.getConstructor(this, args, true);
        return this;
    }

    public Type discoverType() {
        return getTypeManager().voidType;
    }

    public void fixForIntroductions() {
        //System.out.println("fixing: " + constructor.toShortString() );
        Expr extraArg = null;
        if (constructor != null) extraArg = constructor.getExtraArgExpr();
        //System.out.println("constructor " + constructor.toShortString() + ": " + extraArg);
        if (extraArg != null) args.add(extraArg);
    }


    //INTRO from ASTFixerPass
    public ASTObject postFixAST(ASTFixerPass fixer) {
        fixForIntroductions();
        return super.postFixAST(fixer);
    }

    public void unparse(CodeWriter writer) {
        if (getEnclosingInstanceExpr() != null && isSuper) {
            writer.write(getEnclosingInstanceExpr());
            writer.write(".");
        }
        if (isSuper) {
            writer.writeKeyword("super");
        } else {
            writer.writeKeyword("this");
        }
        writer.openParen('(');
        writer.write(args);
        writer.closeParen(')');
        writer.closeStmt();
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        ConstructorDec dec = getConstructorDec();
        if (! w.isAccessible(dec, this)) {
            // perhaps should be creating a cookie, but won't for now.
            dec.getModifiers().setPrivate(false);
        }
        return this;
    }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walker.inConstructorCall();
        walk(walker);
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        if (getConstructorDec() == null) {
            this.showWarning("no constructor dec bound: " + this);
            this.getParent().getParent().display(2);
            return;
        }
        walker.addArgs(getArgs(),
                       getConstructorDec().getDeclaringType().getTypeDec());
    }

    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inCCall();
        super.walkInnerInfo(w);
        w.restoreContext(context);
    }

    public void postInnerInfo(InnerInfoPass w) {
        if (getConstructor() == null) return;
        NameType target = (NameType) getConstructor().getDeclaringType();
        if (w.isInner(target) && getEnclosingInstanceExpr() == null) {
            NameType enclosingTarget = target.getEnclosingType();
            if (! w.isAccessible(enclosingTarget)) {
                showError("no enclosing instance available");
            } else {
                setEnclosingInstanceExpr(getAST()
                                         .makeEnclosingPrimary(enclosingTarget,
                                                               w.currentType()));
            }
        }
    }

    //------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        throw new RuntimeException("should only be called in effect context "
                                   + this);
    }
    protected void cgEffect(CodeBuilder cb) {
        cb.enterLocation(getSourceLocation());        
        Constructor dec = getInternalConstructor();
        NameType declaringType = (NameType) dec.getDeclaringType();
        cb.emitALOAD(0);
        if (getEnclosingInstanceExpr() != null) {
            getEnclosingInstanceExpr()
                .cgValue(cb, dec.getEnclosingInstanceFormal().getType());
        }
        getArgs().cgValues(cb, dec.getFormals());
        cb.emitINVOKESPECIAL(declaringType, "<init>",
                             dec.getDescriptor(), dec.getStackDelta());
    }
    public Constructor getInternalConstructor() {
        if (constructor != null) return constructor;

        Type myType = getDeclaringType(); //getContext().getEnclosingTypeDec(this);
        if (isSuper) {
            myType = myType.getTypeDec().getSuperClassType();
        }
        constructor = myType.getConstructor(this, args, true);
        return constructor;
    }

    // ------------------------------
    // convenience constructors

    public ConstructorCallExpr(SourceLocation source, boolean _isSuper, Exprs _args) {
        this(source, null, _isSuper, _args, null);
    }

    public ConstructorCallExpr(SourceLocation location, boolean _isSuper,
                               Exprs _args, Constructor _constructor) {
        this(location, null, _isSuper, _args, _constructor);
    }

    //BEGIN: Generated from @child and @property
    protected Expr enclosingInstanceExpr;
    public Expr getEnclosingInstanceExpr() { return enclosingInstanceExpr; }
    public void setEnclosingInstanceExpr(Expr _enclosingInstanceExpr) {
        if (_enclosingInstanceExpr != null) _enclosingInstanceExpr.setParent(this);
        enclosingInstanceExpr = _enclosingInstanceExpr;
    }

    protected boolean isSuper;
    public boolean getIsSuper() { return isSuper; }
    public void setIsSuper(boolean _isSuper) { isSuper = _isSuper; }

    protected Exprs args;
    public Exprs getArgs() { return args; }
    public void setArgs(Exprs _args) {
        if (_args != null) _args.setParent(this);
        args = _args;
    }

    protected Constructor constructor;
    public Constructor getConstructor() { return constructor; }
    public void setConstructor(Constructor _constructor) { constructor = _constructor; }

    public ConstructorCallExpr(SourceLocation location, Expr _enclosingInstanceExpr, boolean _isSuper, Exprs _args, Constructor _constructor) {
        super(location);
        setEnclosingInstanceExpr(_enclosingInstanceExpr);
        setIsSuper(_isSuper);
        setArgs(_args);
        setConstructor(_constructor);
    }
    protected ConstructorCallExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ConstructorCallExpr ret = new ConstructorCallExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (enclosingInstanceExpr != null) ret.setEnclosingInstanceExpr( (Expr)walker.process(enclosingInstanceExpr) );
        ret.isSuper = isSuper;
        if (args != null) ret.setArgs( (Exprs)walker.process(args) );
        ret.constructor = constructor;
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return enclosingInstanceExpr;
        case 1: return args;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "enclosingInstanceExpr";
        case 1: return "args";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setEnclosingInstanceExpr((Expr)child); return;
        case 1: setArgs((Exprs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "ConstructorCallExpr(isSuper: "+isSuper+", "+"constructor: "+constructor+")";
    }

    //END: Generated from @child and @property
}
