/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;


import java.io.StringWriter;
import org.aspectj.util.JavaStrings;

import org.aspectj.compiler.base.bcg.ClassfileBuilder;

/**
  * @grammar ...
  * @##child Modifiers modifiers
  *
  */
public abstract class Dec extends Stmt {
    protected SemanticObject semanticObject;

    public SemanticObject getCorrespondingSemanticObject() {
        if (semanticObject == null) semanticObject = makeCorrespondingSemanticObject();
        return semanticObject;
    }

    //XXX should be abstract
    public SemanticObject makeCorrespondingSemanticObject() { return null; }

    protected BlockScope makeBlockScope(ScopeWalker walker) {
        Scope parent = walker.getScope();
        if (!walker.useJavaScopes() && fromLexicalScope != null) {
            parent = new TypeScope(getCompiler(),
                fromLexicalScope, 
                getDeclaringType());
        }
        return new BlockScope(getCompiler(), parent);
    }

    private boolean isDeprecated;

    public boolean isDeprecated() {
        return isDeprecated;
    }

    public void setDeprecated(boolean isDeprecated) {
        this.isDeprecated = isDeprecated;
    }

    public boolean inStaticContext() {
        //System.out.println("static? " + toShortString() + " == " + isStatic());
        return isStatic();
    }

    private boolean explicitlyLanguageVisible = false;

    public boolean isLanguageVisible() {
        return !isSynthetic() || explicitlyLanguageVisible;
    }

    public void setLanguageVisible() {
        explicitlyLanguageVisible = true;
    }
    
    
    private boolean explicitlyNonSynthetic = false;

    public boolean isSynthetic() {
    	if (explicitlyNonSynthetic) return false;
    	else return super.isSynthetic();
    }

    public void setExplicitlyNonSynthetic() {
        explicitlyNonSynthetic = true;
    }

    protected String addIntroducedFromType(String s) {
        if (!isIntroduced()) return s;
        return s + " from " + getLexicalType().toShortString();
    }

    public boolean isIntroduced() {
        // this is a good working definition of an introduction
        return getLexicalType() != getDeclaringType();
    }

    public Dec owner = null;


    private Type declaringType = null;
    private Type lexicalType = null;

    private Scope fromLexicalScope = null;
    public void setFromLexicalScope(Scope fromScope) {
        fromLexicalScope = fromScope;
    }

    public void setAllEnclosingTypes(Type type) {
        setDeclaringType(type);
        setLexicalType(type);
    }

    public void setDeclaringType(Type type) { declaringType = type; }

    public Type getDeclaringType() {
        if (declaringType == null) declaringType = lookupDeclaringType();
        return declaringType;
    }

    public void setLexicalType(Type type) { lexicalType = type; }

    public Type getLexicalType() {
        if (lexicalType == null) lexicalType = lookupLexicalType();
        return lexicalType;
    }

    protected Type lookupDeclaringType() {
        //System.out.println(getParent());
        if (getParent() == null) {
            this.showError("no parent: "); // + this);
            return null;
        }

        if (getParent().getParent() instanceof TypeDec) {
            return ((TypeDec)getParent().getParent()).getType();
        } else {
            return super.getDeclaringType();
        }
    }

    protected Type lookupLexicalType() {
        //System.out.println(getParent());
        if (getParent() == null) {
            //this.showError("no parent: "); // + this);
            return null;
        }

        if (getParent().getParent() instanceof TypeDec) {
            return ((TypeDec)getParent().getParent()).getType();
        } else {
            return super.getLexicalType();
        }
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        this.walk(w.createBodyChecker());
    }

    //INTRO from CopyWalker
    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        this.declaringType = null; //walker.moveType(((Dec)oldObject).declaringType);
//        System.out.println("moveType: "
//        + walker.moveType(((Dec)oldObject).declaringType));
        this.lexicalType = null; //walker.moveType(((Dec)oldObject).lexicalType);
        this.explicitlyLanguageVisible = ((Dec)oldObject).explicitlyLanguageVisible;
        this.explicitlyNonSynthetic = ((Dec)oldObject).explicitlyNonSynthetic;


        //??? how do we handle copying of SO's???
        walker.addMapping(((Dec)oldObject).getCorrespondingSemanticObject(),
                                getCorrespondingSemanticObject());
        return super.postCopy(walker, oldObject);
    }

    public void preCopy(CopyWalker walker, ASTObject oldObject) {
        walker.addMapping(oldObject, this);
        walker.addMapping(((Dec)oldObject).getCorrespondingSemanticObject(),
                                getCorrespondingSemanticObject());
    }
    //INTRO from MovingWalker ???
    public ASTObject postMove(MovingWalker walker) {
        walker.addMapping(this, this);
        walker.addMapping(getCorrespondingSemanticObject(),
                                getCorrespondingSemanticObject());
        return this;
    }

    // delegate all of these to modifiers
    public boolean isPublic() {
        return getModifiers().isPublic();
    }
    public boolean isFinal() {
        return getModifiers().isFinal();
    }
    public boolean isPrivate() {
        return getModifiers().isPrivate();
    }
    public boolean isProtected() {
        return getModifiers().isProtected();
    }
    public boolean isStatic() {
        return getModifiers().isStatic();
    }
    public boolean isAbstract() {
        return getModifiers().isAbstract();
    }

    public abstract Modifiers getModifiers();

    public boolean forcePublic = false;
    public void forcePublic() {
        if (!isPublic()) forcePublic = true;
    }

    public boolean isMoreAccessibleThan(Dec other) {
        return getModifiers().isMoreAccessibleThan(other.getModifiers());
    }


    public boolean isMoreSpecificThan(Dec other) {
        // require that T1 is method convertable to T2
        Type type1 = this.getDeclaringType();
        Type type2 = other.getDeclaringType();
        return type1.isMethodConvertableTo(type2);
    }


    public boolean isApplicable(Exprs parameters) {
        return true;
    }

    public boolean isAlmostApplicable(Exprs params) { return true; }


    boolean hasPrivilegedAccess(ASTObject fromWhere) {
        //XXX needs to be cleaned up over time
        if (fromWhere == null) return true;

        Type fromType = fromWhere.getLexicalType();
        while (fromType != null) {
            TypeDec td = fromType.getTypeDec();
            if (td == null) return false;
            if (td.getModifiers().isPrivileged()) return true;
            fromType = td.getLexicalType();
        }
        return false;
    }

    public boolean isAccessible(ASTObject fromWhere) {
        return isAccessible(fromWhere, false);
    }

    public boolean isAccessible(ASTObject fromWhere, boolean inBytecode) {
//        if (fromWhere instanceof TypeDec) {
//            fromWhere.showError("checking access from type dec");
//        }
        if (isPublic()) return true;

        if (!inBytecode && hasPrivilegedAccess(fromWhere)) return true;

        Type toType;
        Type fromType;
        if (inBytecode) {
            toType = getBytecodeType();
            if (fromWhere.getBytecodeTypeDec() == null) fromType = null;
            else fromType = fromWhere.getBytecodeTypeDec().getType();
        } else {
            toType = getLexicalType();
            fromType = fromWhere.getLexicalType();
        }

//        fromWhere.showWarning(toShortString() + ": " +
//                toType + " -> " + fromType);

        // special case for import statements
        if (fromType == null) {
            //System.out.println(toShortString() + ", " + toType + ", " + fromWhere);
            if (isPrivate()) {
                return false;
            } else {
                return samePackage(fromType, toType, fromWhere);
            }
        }

        if (isPrivate()) {
            /* Will add this if I can figure out if inBytecode does what it should...
            if (inBytecode) {
                return toType == fromType;
            } else {
            */
            if (fromType == null || toType == null) return false;
            return fromType.getOutermostLexicalType()
                == toType.getOutermostLexicalType();
                /*
            }
                */
        } else if (isProtected()) {
            if (samePackage(fromType, toType, fromWhere)) return true;
            if (fromType == null || toType == null) return false;
//            System.out.println("    subtype? " + fromType.isSubtypeOf(toType));
            while (true) {
                if (fromType.isSubtypeOf(toType)) break;
                //XXXif (inBytecode) fromType = fromType.getTypeDec().getBytecodeType();
                fromType = fromType.getTypeDec().getLexicalType();
                if (fromType== null) return false;
            }

            // special rules from JLSv2 6.6.2 only for instance fields and methods
            // and constructors
            if (isStatic()) return true;
            if (this instanceof TypeDec) return true;

            return fromWhere.hasLegalProtectedAccess(fromType);
        } else { // package private
            return samePackage(fromType, toType, fromWhere);
        }
    }

    public boolean samePackage(Type fromType, Type thisType, ASTObject fromWhere) {
        String p1, p2;
        if (thisType == null) {
            p1 = getCompilationUnit().getPackageName();
        } else {
            p1 = thisType.getPackageName();
        }
        if (fromType == null) {
            p2 = fromWhere.getCompilationUnit().getPackageName();
        } else {
            p2 = fromType.getPackageName();
        }
        return nameMatches(p1, p2);
    }

    static boolean nameMatches(String p1, String p2) {
        //System.err.println(toShortString() + ", " + p1 + "::" + p2);
        if (p1 == null) return p2 == null;
        else return p1.equals(p2);
    }


    public boolean isInherited(Type inType) {
        if (isPrivate()) return this.isIntroduced();
        
        if (this.isIntroduced()) return true;
        
        if (this.isAccessible(inType.getTypeDec().getBody())) return true;
        
        if (!(this instanceof TypeDec) && this.isAbstract()) {
        	inType.getTypeDec().showError("The member '" + toShortString() + 
        		"' has default access and hence is not inherited and can not be made concrete");
        }
        
        return false;
    }

    public boolean conflictsWith(Dec otherDec) {
        if (!getId().equals(otherDec.getId())) return false;

        if (!this.isAccessible(otherDec) && !otherDec.isAccessible(this)) return false;
        if (this.getDeclaringType() == otherDec.getDeclaringType()) return true;
        return false;
    }

    public boolean checkOverride(Type inType, Dec otherDec) { return false; }

    public boolean dominates(Dec otherDec) {
        if (this == otherDec) return false;

        if (getDeclaringType() != otherDec.getDeclaringType()) return false;

        if (isIntroduced() && otherDec.isIntroduced()) {
            return getLexicalType().dominates(otherDec.getLexicalType());
        }
        return false;
    }


//     public boolean canOverride(Dec otherDec) {
//         if (this == otherDec) return true;

//         if (this.getDeclaringType().isSubtypeOf(otherDec.getDeclaringType())) return true;

//         if (isIntroduced() && otherDec.isIntroduced()) {
//             return (getLexicalType().dominates(otherDec.getLexicalType()));
//         }

//         return false;
//     }

    public String getSignatureString() {
        //StringWriter stringBuf = new StringWriter();
        CodeWriter writer = new CodeWriter(getCompiler());
        writer.setOnlySignatures(true);
        writer.write(this);

        return writer.getString();
    }

    //XXX should be abstract
    public abstract String getKind();

    private String bytecodeId = null;
    public String getBytecodeId() {
        if (bytecodeId == null) return getId();
        return bytecodeId;
    }

    public void setBytecodeId(String id) {
        bytecodeId = id;
    }

    //XXX should be abstract
    public void setId(String id) {}

    //!!! remove
    public String getName() {
        return getId();
    }

    //!!! should be abstract
    public abstract String getId();
    public abstract String toShortString();

    public void writeNames(CodeWriter writer, String label, TypeDs names)  {
        if (names == null || names.size() == 0) return;

        writer.requiredSpace();
        writer.writeKeyword(label);
        writer.requiredSpace();
        writer.write(names);
    }

    public void writeModifiers(CodeWriter writer) {
        if (getModifiers() == null) return;

        if (forcePublic) {
            Modifiers mods = (Modifiers)getModifiers().copy();
            mods.setPublic(true);
            writer.write(mods);
        } else {
            writer.write(getModifiers());
        }
    }

    public void cleanup() {
        super.cleanup();
    }

    //XXX don't belong here
    public String getDescriptor() { return null; }
    public int getStackDelta() { return 0; }

    // ------------------------------
    // bcg

    /** This method takes care of generating code for Member decs.  At
        this point in the compilation, we should only have two kinds
        of Member decs, {@link FieldDec} and {@link CodeDec}.
        Moreover, {@link FieldDec}s should not have initializers
        unless they're constants.  This method throws an error if it
        is ever invoked:  {@link FieldDec} and {@link CodeDec} override it.
    */
    protected void cgMember(ClassfileBuilder maker) {
        //throw new RuntimeException("Invalid dec " + this);
    }

    //BEGIN: Generated from @child and @property

    public Dec(SourceLocation location) {
        super(location);

    }


    public String getDefaultDisplayName() {
        return "Dec()";
    }

    //END: Generated from @child and @property
}
