/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

public final class DoubleType extends RealType {

    protected boolean isAssignableFromOtherType(Type other) {
        return (other instanceof NumericType);
    }

    public String getName()      { return "double"; }
    public int    getTypeIndex() { return Type.DOUBLE; }
    public Type   getRefType()   { return getTypeManager().getType("java.lang", "Double"); }
    public Expr   getNullExpr()  { return getAST().makeLiteral(0.0); }

    public DoubleType(JavaCompiler compiler) {
        super(compiler);
    }

    // vm internals
    public String getDescriptor() { return "D"; }
    public int getSlotCount() { return 2; }

    // folding
    private LiteralExpr makeLit(double d) {
        return getAST().makeLiteral(d);
    }
    private LiteralExpr makeBooleanLit(boolean b) {
        return getAST().makeLiteral(b);
    }

    LiteralExpr foldCast(LiteralExpr rand) {
        return makeLit(rand.getDoubleValue());
    }

    strictfp LiteralExpr foldPlusOp(LiteralExpr rand) {
        return makeLit(+ rand.getDoubleValue());
    }
    strictfp LiteralExpr foldMinusOp(LiteralExpr rand) {
        return makeLit(- rand.getDoubleValue());
    }

    strictfp LiteralExpr foldAddOp(LiteralExpr rand1, LiteralExpr rand2) {
        return makeLit(rand1.getDoubleValue() + rand2.getDoubleValue());
    }

    strictfp LiteralExpr foldNumericOp(String op,
                                       LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        double a = lit1.getDoubleValue();
        double b = lit2.getDoubleValue();
        if      (op == "-") return ast.makeLiteral(a - b);
        else if (op == "%") return ast.makeLiteral(a % b);
        else if (op == "/") return ast.makeLiteral(a / b);
        else if (op == "*") return ast.makeLiteral(a * b);
        else throw new RuntimeException("bad numeric op " + op);
    }

    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        double a = lit1.getDoubleValue();
        double b = lit2.getDoubleValue();
        if      (op == "==") return ast.makeLiteral(a == b);
        else if (op == "!=") return ast.makeLiteral(a != b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        double a = lit1.getDoubleValue();
        double b = lit2.getDoubleValue();
        if      (op == "<")  return ast.makeLiteral(a < b);
        else if (op == "<=") return ast.makeLiteral(a <= b);
        else if (op == ">=") return ast.makeLiteral(a >= b);
        else if (op == ">")  return ast.makeLiteral(a > b);
        else throw new RuntimeException("bad op " + op);
    }

    // code gen

    void emitAdd(CodeBuilder cb) { cb.emitDADD(); }

    void emitNumericOp(CodeBuilder cb, String op) {
        if      (op == "-") cb.emitDSUB();
        else if (op == "%") cb.emitDREM();
        else if (op == "/") cb.emitDDIV();
        else if (op == "*") cb.emitDMUL();
    }
    void emitNeg(CodeBuilder cb) { cb.emitDNEG(); }

    public void emitLoad(CodeBuilder cb, int loc) { cb.emitDLOAD(loc); }
    public void emitStore(CodeBuilder cb, int loc) { cb.emitDSTORE(loc); }
    public void emitReturn(CodeBuilder cb) { cb.emitDRETURN(); }

    void emitCast(CodeBuilder cb, Type castTo) {
        ((NumericType) castTo).emitCastFromDouble(cb);
    }
    void emitCastFromInt(CodeBuilder cb)    { cb.emitI2D(); }
    void emitCastFromFloat(CodeBuilder cb)  { cb.emitF2D(); }
    void emitCastFromLong(CodeBuilder cb)   { cb.emitL2D(); }
    void emitCastFromDouble(CodeBuilder cb) {  }

    void emitPop(CodeBuilder cb) { cb.emitPOP2(); }
    void emitDup(CodeBuilder cb) { cb.emitDUP2(); }
    void emitDupX1(CodeBuilder cb) { cb.emitDUP2_X1(); }
    void emitDupX2(CodeBuilder cb) { cb.emitDUP2_X2(); }

    void emitZero(CodeBuilder cb) { cb.emitDoubleConstant(0.0); }
    void emitOne(CodeBuilder cb) { cb.emitDoubleConstant(1.0); }
    void emitMinusOne(CodeBuilder cb) { cb.emitDoubleConstant(-1.0); }

    void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        cb.emitDCMPL(); // -1 if either are NaN
        if      (op == "==") cb.emitIFEQ(t, f);
        else if (op == "!=") cb.emitIFNE(t, f);
    }

    void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "<")  { cb.emitDCMPG(); cb.emitIFLT(t, f); }
        else if (op == "<=") { cb.emitDCMPG(); cb.emitIFLE(t, f); }
        else if (op == ">=") { cb.emitDCMPL(); cb.emitIFGE(t, f); }
        else if (op == ">")  { cb.emitDCMPL(); cb.emitIFGT(t, f); }
    }

    void emitAload(CodeBuilder cb) { cb.emitDALOAD(); }
    void emitAstore(CodeBuilder cb) { cb.emitDASTORE(); }

}


