/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar rand1 <op> rand2
 */
public class EqualityTestOpExpr extends BinopExpr {

    protected Type discoverType() {
        Type ty1 = getRand1().getType();
        Type ty2 = getRand2().getType();

        if (ty1 instanceof RefType && ty2 instanceof RefType) {
            if (! (ty1.isCoercableTo(ty2) || ty2.isCoercableTo(ty1))) {
                showError("incomparable types: " + ty1.getString() +
                          " and " + ty2.getString());
            }
        } else if ((ty1 instanceof RefType || ty2 instanceof RefType) ||
                   ! ((ty1.isBoolean() && ty2.isBoolean()) ||
                      (ty1.isNumeric() && ty2.isNumeric()))) {
            showOperatorTypeError(op, ty1, ty2);
        }
        return getTypeManager().booleanType;
    }

    protected Type getLiftType() {
        Type ty1 = getRand1().getType();
        Type ty2 = getRand2().getType();

        if (ty1.isBoolean() && ty2.isBoolean()) {
            return getTypeManager().booleanType;
        } else if (ty1.isNumeric() && ty2.isNumeric()) {
            return getTypeManager().binaryNumericPromotion(ty1, ty2);
        } else if (ty1 instanceof RefType && ty2 instanceof RefType) {
            return getTypeManager().getObjectType();
        } else {
            showOperatorTypeError(op, ty1, ty2);
            return getTypeManager().anyType;
        }
    }

    protected LiteralExpr halfFold(Type type, LiteralExpr lit1, LiteralExpr lit2) {
        return type.foldEqualityTestOp(getOp(), lit1, lit2);
    }

    // ------------------------------
    // bcg

    protected void cgTest(CodeBuilder cb, Label t, Label f) {
        Type liftType = getLiftType();
        if (liftType.hasFastEqualityTestOp()) {
            liftType.emitFastEqualityTestOp(cb, this, t, f);
        } else {
            getRand1().cgValue(cb, liftType);
            getRand2().cgValue(cb, liftType);
            liftType.emitEqualityCompare(cb, getOp(), t, f);
        }
    }

    protected void cgOp(CodeBuilder cb, Type ty) {
        throw new RuntimeException("Attempt to generate code for effect op "
                                   + getOp());
    }

    //BEGIN: Generated from @child and @property

    public EqualityTestOpExpr(SourceLocation location, Expr _rand1, String _op, Expr _rand2) {
        super(location, _rand1, _op, _rand2);

    }
    protected EqualityTestOpExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        EqualityTestOpExpr ret = new EqualityTestOpExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (rand1 != null) ret.setRand1( (Expr)walker.process(rand1) );
        ret.op = op;
        if (rand2 != null) ret.setRand2( (Expr)walker.process(rand2) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "EqualityTestOpExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
