/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.util.*;

/**
 * @grammar ...
 */

public class InterfaceDec extends TypeDec {
    public String getKind() { return "interface"; }

    public Type getSuperClassType() {
        return getTypeManager().getObjectType();
    }

    public boolean isAnonymous() { return false; }

    protected void addFieldDec(FieldDec fieldDec) {
        if (fieldDec.fromSource()) {
            fieldDec.getModifiers().setPublic(true);
            fieldDec.getModifiers().setStatic(true);
            fieldDec.getModifiers().setFinal(true);
        }
        super.addFieldDec(fieldDec);
    }

    protected void addMethodDec(MethodDec methodDec) {
        if (methodDec.fromSource()) {
            methodDec.getModifiers().setPublic(true);
            methodDec.getModifiers().setAbstract(true);
        }
        super.addMethodDec(methodDec);
    }

    /* This collection of methods provides a helper class inside of
     * an interface as a place to hold introduced static methods and fields.
     * It might also hows a static initializer???
     */
    public void addToHelperClass(Dec dec) {
        ClassDec helper = getHelperClass();
        dec.setDeclaringType(helper.getType());
        helper.getBody().add(dec);
    }

    static final String HELPER_CLASS_NAME = "AJC_HELPER_CLASS";
    ClassDec getHelperClass() {
        Decs body = getBody();
        for (int i = 0; i<body.size(); i++) {
            if (isHelperClass(body.get(i))) return (ClassDec)body.get(i);
        }
        ClassDec helper = makeHelperClass();
        body.add(helper);

        return helper;
    }

    public static boolean isHelperClass(Dec dec) {
        return dec != null && dec instanceof ClassDec && dec.getName().equals(HELPER_CLASS_NAME);
    }

    ClassDec makeHelperClass() {
        final AST ast = getAST();
        ClassDec ret = ast.makeClass(ast.makeModifiers(Modifiers.PUBLIC|Modifiers.STATIC),
                HELPER_CLASS_NAME, getTypeManager().getObjectType().makeTypeD(),
                ast.makeTypeDs());
        ret.setEnclosingTypeDec(this);

        // deliberately creating a circular reference to the helper class for
        // order of class loading issues
        FieldDec ref1 = ast.makeField(ast.makeModifiers(Modifiers.PRIVATE|Modifiers.FINAL|Modifiers.STATIC),
                getTypeManager().getObjectType(), "AJC_INNER_REF");
        FieldDec ref2 = ast.makeField(ast.makeModifiers(Modifiers.PUBLIC|Modifiers.FINAL|Modifiers.STATIC),
                getTypeManager().getObjectType(), "AJC_OUTER_REF");
        ret.getBody().add(ref1);
        getBody().add(ref2);
        ref1.setInitializer(ast.makeGet(ref2.getField()));
        ref2.setInitializer(ast.makeGet(ref1.getField()));
        return ret;
    }

    public void unparse(CodeWriter writer) {
        writer.write(modifiers);
        writer.writeKeyword("interface");
        writer.requiredSpace();
        writer.write(id);

        writeNames(writer, "extends", superInterfaces);

        if (!writer.isOnlySignatures()) {
            writer.optionalSpace();
            writer.openBlock();
            writer.write(body);
            writer.closeBlock();
        }
    }

    public boolean isStatic() {
        return true;
    }

    // ------------------------------
    // INTRO: ByteCodeCleanupPass

    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        final int len = getBody().size();
        for (int i = 0; i < len; i++ ) {
            Dec dec = getBody().get(i);
            Modifiers m = dec.getModifiers();
            if (dec instanceof FieldDec) {
                m.setPublic(true);
                m.setStatic(true);
                m.setFinal(true);
            } else if (dec instanceof MethodDec) {
                m.setPublic(true);
                m.setAbstract(true);
            }
        }
        return this;
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        return this;
    }

    //BEGIN: Generated from @child and @property

    public InterfaceDec(SourceLocation location, Modifiers _modifiers, String _id, TypeDs _superInterfaces, Decs _body) {
        super(location, _modifiers, _id, _superInterfaces, _body);

    }
    protected InterfaceDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        InterfaceDec ret = new InterfaceDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        ret.id = id;
        if (superInterfaces != null) ret.setSuperInterfaces( (TypeDs)walker.process(superInterfaces) );
        if (body != null) ret.setBody( (Decs)walker.process(body) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "InterfaceDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
