/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

public abstract class IntishType extends IntegralType {

    public Expr getNullExpr() {
        return new IntLiteralExpr(getAST().getSourceLocation(), this, "0", 0);
    }

    public IntishType(JavaCompiler compiler) {
        super(compiler);
    }

    // folding
    private LiteralExpr makeLit(int i) {
        return getAST().makeLiteral(i);
    }
    private LiteralExpr makeBooleanLit(boolean b) {
        return getAST().makeLiteral(b);
    }

    LiteralExpr foldPlusOp(LiteralExpr rand) {
        return makeLit(+ rand.getIntValue());
    }
    LiteralExpr foldMinusOp(LiteralExpr rand) {
        return makeLit(- rand.getIntValue());
    }
    LiteralExpr foldBitNotOp(LiteralExpr rand) {
        return makeLit(~ rand.getIntValue());
    }

    LiteralExpr foldAddOp(LiteralExpr rand1, LiteralExpr rand2) {
        return makeLit(rand1.getIntValue() + rand2.getIntValue());
    }

    LiteralExpr foldNumericOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        int a = lit1.getIntValue();
        int b = lit2.getIntValue();
        if      (op == "-") return ast.makeLiteral(a - b);
        else if (op == "%") return ast.makeLiteral(a % b);
        else if (op == "/") return ast.makeLiteral(a / b);
        else if (op == "*") return ast.makeLiteral(a * b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldBitwiseOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        int a = lit1.getIntValue();
        int b = lit2.getIntValue();
        if      (op == "&") return ast.makeLiteral(a & b);
        else if (op == "|") return ast.makeLiteral(a | b);
        else if (op == "^") return ast.makeLiteral(a ^ b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldShiftOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        int a = lit1.getIntValue();
        int b = lit2.getIntValue();
        if      (op == "<<") return ast.makeLiteral(a << b);
        else if (op == ">>") return ast.makeLiteral(a >> b);
        else if (op == ">>>") return ast.makeLiteral(a >>> b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        int a = lit1.getIntValue();
        int b = lit2.getIntValue();
        if      (op == "==") return ast.makeLiteral(a == b);
        else if (op == "!=") return ast.makeLiteral(a != b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        int a = lit1.getIntValue();
        int b = lit2.getIntValue();
        if      (op == "<")  return ast.makeLiteral(a < b);
        else if (op == "<=") return ast.makeLiteral(a <= b);
        else if (op == ">=") return ast.makeLiteral(a >= b);
        else if (op == ">")  return ast.makeLiteral(a > b);
        else throw new RuntimeException("bad op " + op);
    }

    // code gen
    final void emitAdd(CodeBuilder cb) { cb.emitIADD(); }

    final void emitNumericOp(CodeBuilder cb, String op) {
        if      (op == "-") cb.emitISUB();
        else if (op == "%") cb.emitIREM();
        else if (op == "/") cb.emitIDIV();
        else if (op == "*") cb.emitIMUL();
    }
    final void emitBitwiseOp(CodeBuilder cb, String op) {
        if      (op == "&") cb.emitIAND();
        else if (op == "|") cb.emitIOR();
        else if (op == "^") cb.emitIXOR();
    }
    final void emitShiftOp(CodeBuilder cb, String op) {
        if      (op == "<<")  cb.emitISHL();
        else if (op == ">>")  cb.emitISHR();
        else if (op == ">>>") cb.emitIUSHR();
    }
    final void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "==") cb.emitIF_ICMPEQ(t, f);
        else if (op == "!=") cb.emitIF_ICMPNE(t, f);
    }
    final void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "<")  cb.emitIF_ICMPLT(t, f);
        else if (op == "<=") cb.emitIF_ICMPLE(t, f);
        else if (op == ">=") cb.emitIF_ICMPGE(t, f);
        else if (op == ">")  cb.emitIF_ICMPGT(t, f);
    }

    final void emitNeg(CodeBuilder cb) { cb.emitINEG(); }

    void emitCast(CodeBuilder cb, Type castTo) {
        if (castTo != this) {
            ((NumericType) castTo).emitCastFromInt(cb);
        }
    }

    final public void emitLoad(CodeBuilder cb, int loc) { cb.emitILOAD(loc); }
    final public void emitStore(CodeBuilder cb, int loc) { cb.emitISTORE(loc); }
    final public void emitReturn(CodeBuilder cb) { cb.emitIRETURN(); }

    final void emitZero(CodeBuilder cb) { cb.emitIntConstant(0); }
    final void emitOne(CodeBuilder cb) { cb.emitIntConstant(1); }
    final void emitMinusOne(CodeBuilder cb) { cb.emitIntConstant(-1); }


    // ------------------------------
    // "fast" opcode handling

    boolean hasFastEqualityTestOp() { return true; }
    void emitFastEqualityTestOp(CodeBuilder cb, EqualityTestOpExpr e,
                                Label t, Label f) {
        Expr rand1 = e.getRand1();
        Expr rand2 = e.getRand2();
        String op = e.getOp();
        if        (rand1.isConstantZero()) {
            rand2.cgValue(cb, this);
            emitEqualityCompare0(cb, op, t, f);
        } else if (rand2.isConstantZero()) {
            rand1.cgValue(cb, this);
            emitEqualityCompare0(cb, op, t, f);
        } else {
            rand1.cgValue(cb, this);
            rand2.cgValue(cb, this);
            emitEqualityCompare(cb, op, t, f);
        }
    }
    private void emitEqualityCompare0(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "==") cb.emitIFEQ(t, f);
        else if (op == "!=") cb.emitIFNE(t, f);
    }

    boolean hasFastNumericTestOp() { return true; }
    void emitFastNumericTestOp(CodeBuilder cb, NumericTestOpExpr e,
                                Label t, Label f) {
        Expr rand1 = e.getRand1();
        Expr rand2 = e.getRand2();
        String op = e.getOp();
        if        (rand1.isConstantZero()) {
            rand2.cgValue(cb, this);
            emit0Compare(cb, op, t, f);
        } else if (rand2.isConstantZero()) {
            rand1.cgValue(cb, this);
            emitCompare0(cb, op, t, f);
        } else {
            rand1.cgValue(cb, this);
            rand2.cgValue(cb, this);
            emitNumericCompare(cb, op, t, f);
        }
    }
    private void emit0Compare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "<")  cb.emitIFGT(t, f);
        else if (op == "<=") cb.emitIFGE(t, f);
        else if (op == ">=") cb.emitIFLE(t, f);
        else if (op == ">")  cb.emitIFLT(t, f);
    }
    private void emitCompare0(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "<")  cb.emitIFLT(t, f);
        else if (op == "<=") cb.emitIFLE(t, f);
        else if (op == ">=") cb.emitIFGE(t, f);
        else if (op == ">")  cb.emitIFGT(t, f);
    }
}
