/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar new typeD[dimExprs, ..] {initializers, ..}
 * @child ArrayInitializer initializer
 * @child Exprs dimExprs
 */

// note carefully that typeD holds the ARRAY TYPED, not the typeD from
// the original concrete syntax.

// Also note that there are TWO distinct types hiding here, because of
// the invariant that either dimExprs or initializers (but never both)
// will be null.  So we either have, from the concrete syntax:
//   new TypeId[]... { Expr, ... }, or
//   new TypeId[Expr]...[]...
public class NewArrayExpr extends NewExpr {

    public Type discoverType() {
        //System.out.println("discover type: " + this);
        Exprs dimExprs = getDimExprs();
        ArrayInitializer inits = getInitializer();
        ArrayType type = (ArrayType)getTypeD().getType();

        if (initializer == null) {
            // at least one dimExpr, all dimExprs are ints followed by emptys
            final int dimExprCount = dimExprs.size();
            for (int i = 0; i < dimExprCount; i++) {
                Expr dimExpr = dimExprs.get(i);
                if (dimExpr instanceof EmptyExpr) break;
                if (! dimExpr.isAssignableTo(getTypeManager().intType)) {
                    dimExpr.showTypeError(dimExpr.getType(), getTypeManager().intType);
                }
            }
        }
        return type;
    }

    public Exprs getArgs() { return new Exprs(getSourceLocation()); }
    public CodeDec getCodeDec() { return null; } //XXX this might be trouble
    public Type getCalledType() {
        return getTypeD().getType();
    }


    public void unparse(CodeWriter writer) throws IOException {
        ArrayType arrayType = (ArrayType)getTypeD().getType();
        Exprs exprs = getDimExprs();

        writer.writeKeyword("new");
        writer.requiredSpace();
        if (exprs == null) {
            writer.write(arrayType.getString());
            writer.optionalSpace();
            writer.write(initializer);
        } else {
            Type componentType = arrayType.getComponentType();
            while (componentType instanceof ArrayType) {
                componentType = ((ArrayType)componentType).getComponentType();
            }
            writer.write(componentType.getString());

            for(int i=0; i<exprs.size(); i++) {
                writer.write('['); writer.write(exprs.get(i)); writer.write(']');
            }
        }
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public void preAssignmentCheck(AssignmentCheckerPass walker) {
        if (getInitializer() != null) {
            Type componentType = ((ArrayType)getType()).getComponentType();
            if (componentType instanceof ArrayType) {
                ArrayType arrayComponentType = (ArrayType) componentType;
                Exprs exprs = initializer.getExprs();
                for (int i = exprs.size() - 1; i >= 0; i--) {
                    Expr expr = exprs.get(i);
                    if (expr instanceof ArrayInitializer) {
                        exprs.set(i,
                                  getAST().makeNewArray(arrayComponentType,
                                                        (ArrayInitializer)expr));
                    }
                }
            }
        }
    }

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        if (getInitializer() != null) {
            Type componentType = ((ArrayType)getType()).getComponentType();
            Exprs exprs = initializer.getExprs();
            for (int i = exprs.size() - 1; i >= 0; i--) {
                Expr expr = exprs.get(i);
                if (! expr.isAssignableTo(componentType)) {
                    expr.showTypeError(expr.getType(), componentType);
                }
            }
        }
        return this;
    }

    // ------------------------------
    // bytecode generation fixing


    /* We can't actually check that a varDec is typesafe until we
       constant fold, due to the strange static downcast allowable
       from integer constants.  */
    private void typeCheck(Type ty, Expr e) {
        if (e instanceof ArrayInitializer) {
            // ARRAY INITIALIZERS
            if (! (ty instanceof ArrayType)) {
                showError("illegal initializer for " + ty.getString());
            } else {
                Type componentTy = ((ArrayType)ty).getComponentType();
                Exprs exprs = ((ArrayInitializer)e).getExprs();
                // make sure expressions were scanned...
                if (exprs != null) {
                    int len = exprs.size();
                    for (int i = 0; i < len; i++) {
                        Expr arrayInit = exprs.get(i);
                        typeCheck(componentTy, arrayInit);
                    }
                }
            }
        } else {
        }
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        ArrayType type = (ArrayType)getTypeD().getType();
        int dims = type.getArrayDimCount();
        Type compTy = type.getComponentType();

        if (initializer == null) {
            // invariant:  (dimExprs != null)
            int filledDims = 0;
            for (int i = 0; i < dims; i++) {
                Expr e = dimExprs.get(i);
                if (e instanceof EmptyExpr) break;
                e.cgValue(cb);
                filledDims++;
            }
            if (dims == 1) {
                compTy.emitNewarray(cb);
            } else {
                type.emitMultiNewarray(cb, filledDims);
            }
        }
        else {
            // invariant: (dimExprs == null)
            // assumption:  we've desugared the array initializer
            Exprs es = initializer.getExprs();
            int size = es.size();
            cb.emitIntConstant(size);
            compTy.emitNewarray(cb);
            for (int i = 0; i < size; i++) {
                Expr e = es.get(i);
                // optimizing new int[] {0, 0}.
                // and new Int[][] {null, null, null}
                if (e.isConstantZero()) continue;
                cb.emitDUP(); // dup the array
                cb.emitIntConstant(i);
                e.cgValue(cb, compTy);
                compTy.emitAstore(cb);
            }
        }
    }

    //BEGIN: Generated from @child and @property
    protected ArrayInitializer initializer;
    public ArrayInitializer getInitializer() { return initializer; }
    public void setInitializer(ArrayInitializer _initializer) {
        if (_initializer != null) _initializer.setParent(this);
        initializer = _initializer;
    }

    protected Exprs dimExprs;
    public Exprs getDimExprs() { return dimExprs; }
    public void setDimExprs(Exprs _dimExprs) {
        if (_dimExprs != null) _dimExprs.setParent(this);
        dimExprs = _dimExprs;
    }

    public NewArrayExpr(SourceLocation location, TypeD _typeD, ArrayInitializer _initializer, Exprs _dimExprs) {
        super(location, _typeD);
        setInitializer(_initializer);
        setDimExprs(_dimExprs);
    }
    protected NewArrayExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        NewArrayExpr ret = new NewArrayExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (typeD != null) ret.setTypeD( (TypeD)walker.process(typeD) );
        if (initializer != null) ret.setInitializer( (ArrayInitializer)walker.process(initializer) );
        if (dimExprs != null) ret.setDimExprs( (Exprs)walker.process(dimExprs) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 1: return initializer;
        case 2: return dimExprs;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 1: return "initializer";
        case 2: return "dimExprs";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 1: setInitializer((ArrayInitializer)child); return;
        case 2: setDimExprs((Exprs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "NewArrayExpr()";
    }

    //END: Generated from @child and @property
}
