/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

/**
 * @grammar case label: stmts
 * @children SwitchClause clause
 */

public class SwitchClauses extends ASTObject {

    //BEGIN: Generated from @child and @property
    protected int size;
    public SwitchClause[] children;

    public SwitchClauses(SourceLocation location, SwitchClause[] _children) {
        super(location);
        for(int i=0; i<_children.length; i++) {
            if (_children[i] != null) _children[i].setParent(this);
        }
        children = _children;
        size = _children.length;
    }

    public SwitchClauses(SourceLocation location) {
        this(location, new SwitchClause[] {});
    }

    public SwitchClauses(SourceLocation location, SwitchClause child1) {
        this(location, new SwitchClause[] {child1});
    }

    public SwitchClauses(SourceLocation location, SwitchClause child1, SwitchClause child2) {
        this(location, new SwitchClause[] {child1, child2});
    }

    public SwitchClauses(SourceLocation location, SwitchClause child1, SwitchClause child2, SwitchClause child3) {
        this(location, new SwitchClause[] {child1, child2, child3});
    }

    public ASTObject copyWalk(CopyWalker walker) {
        final int N = size;
        SwitchClause[] copiedChildren = new SwitchClause[N];
        int newIndex = 0;
        for(int oldIndex=0; oldIndex<N; oldIndex++) {
            SwitchClause newChild = (SwitchClause)walker.process(children[oldIndex]);
            if (newChild != null) copiedChildren[newIndex++] = newChild;
        }
        SwitchClauses ret = new SwitchClauses(getSourceLocation(),copiedChildren);
        ret.size = newIndex;
        ret.setSource(this);
        return ret;
    }

    public ASTObject getChildAt(int childIndex) { return get(childIndex); }
    public void setChildAt(int childIndex, ASTObject child) { set(childIndex, (SwitchClause)child); }
    public String getChildNameAt(int childIndex) { return "clause"+childIndex; }

    public int getChildCount() { return size; }
    public int size() { return size; }

    public SwitchClause get(int index) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        return children[index];
    }

    public void set(int index, SwitchClause child) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        children[index] = child;
        child.setParent(this);
    }

    public void resize(int newSize) {
        if (newSize > children.length) {
            SwitchClause[] newChildren = new SwitchClause[children.length*2 + 1];
            System.arraycopy(children, 0, newChildren, 0, children.length);
            children = newChildren;
        }
        size = newSize;
    }

    public void addAll(SwitchClauses collection) {
        addAll(size, collection);
    }

    public void addAll(int index, SwitchClauses collection) {
        for(int i=0; i<collection.size(); i++) {
            add(index+i, collection.get(i));
        }
    }

    public void add(SwitchClause child) {
        add(size, child);
    }

    public void add(int index, SwitchClause child) {
        if (child == null) return;

        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        resize(size+1);

        for(int moveIndex = size-1; moveIndex > index; moveIndex--) {
            children[moveIndex] = children[moveIndex-1];
        }

        children[index] = child;
        child.setParent(this);
    }

    public void remove(int index) {
        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        size -= 1;

        for(int moveIndex = index; moveIndex < size; moveIndex++) {
            children[moveIndex] = children[moveIndex+1];
        }
    }

    public String getDefaultDisplayName() {
        return "SwitchClauses()";
    }

    //END: Generated from @child and @property
}


