/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import java.util.*;
import java.io.File;

public class TextSourceLocation extends SourceLocation {
    protected int startPosition = -1;
    protected int endPosition = -1;
    protected CompilationUnit compilationUnit;

    // multiple comments form a linked list
    protected Comment comment;

    public String toString() {
        return "TextSourceLocation(" + getSourceFileName() + ":" +
        	getBeginLine() + ":" + getBeginColumn() + ")";
    }


    public TextSourceLocation(CompilationUnit compilationUnit, int start, int end) {
        this.compilationUnit = compilationUnit;
        this.startPosition = start;
        this.endPosition = end;
    }

    public JavaCompiler getCompiler() { return getCompilationUnit().getCompiler(); }

    public int getStartPosition() { return startPosition; }
    public int getEndPosition() { return endPosition; }

    public int getBeginLine() {
        if (startPosition == -1) return -1;
        return getCompilationUnit().getLine(startPosition);
    }

    public int getEndLine() {
        if (endPosition == -1) return -1;
        return getCompilationUnit().getLine(endPosition);
    }

    public int getBeginColumn() {
        if (startPosition == -1) return -1;
        return getCompilationUnit().getColumn(startPosition);
    }

    public int getEndColumn() {
        if (endPosition == -1) return -1;
        return getCompilationUnit().getColumn(endPosition);
    }

    public Comment getComment() {
        return comment;
    }

    public void addComment(Comment _comment) {
        if (comment == null) {
            comment = _comment;
        } else {
            comment.addComment(_comment);
        }
    }

    public void clearComment() {
        comment = null;
    }

    public void setFormalComment(String comment) {
        //System.out.println("adding formal comment: " + comment + " to " + this.getDefaultDisplayName());
        addComment(new FormalComment(comment));
    }

    public String getFormalComment() {
        if (comment == null) return "";
        return comment.getFormalComment();
    }

    public CompilationUnit getCompilationUnit() {
        return compilationUnit;
    }

    public java.io.File getSourceFile() {
        return getCompilationUnit().getSourceFile();
    }

    public String getSourceFileName() {
        return getCompilationUnit().getSourceCanonicalPath();
    }

    public String getSourceDirectoryName() {
        return getCompilationUnit().getSourceDirectory();
    }

    public boolean isSynthetic() {
        return false;
    }

    public boolean hasSource() {
        return getCompilationUnit().getSourceFile() != null;
    }
    public boolean fromSource() {
        return hasSource() && getCompilationUnit().getDecs() != null;
    }
}
