/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import java.util.Iterator;

/**
 * @grammar try {body} finally {finally}
 * @child Stmt body
 * @child Stmt finally
 */
public class TryFinallyStmt extends TryStmt implements WindingStmt {
	
	public void checkSpec() {
		super.checkSpec();
		if (_finally == null) {
       		showError("'try' without 'catch' or 'finally'");
		}
	}

    public void unparse(CodeWriter writer) {
        if ((body instanceof TryCatchStmt) && body.isSynthetic()) {
            TryCatchStmt tcb = (TryCatchStmt) body;
            writer.writeKeyword("try");
            writer.optionalSpace();
            writer.write(tcb.getBody());
            writer.writeChildren(tcb.getCatches());
            writer.writeKeyword("finally");
            writer.optionalSpace();
            writer.write(_finally);
        } else {
            writer.writeKeyword("try");
            writer.optionalSpace();
            writer.write(body);
            writer.writeKeyword("finally");
            writer.optionalSpace();
            writer.write(_finally);
        }
    }

    // ------------------------------
    // INTRO from ASTFixerPass

    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (! isOptional()) return super.postFixAST(fixer);
        setIsOptional(false);
        return this;
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        FlowCheckerPass.Vars beforeV = w.getVars();
        FlowCheckerPass.ESet beforeExns = w.popExns();
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();
        boolean bodyLive = w.isLive();
        FlowCheckerPass.ESet bodyExns = w.popExns();

        FlowCheckerPass.Vars cvExit = w.getVars();
        FlowCheckerPass.Vars fv =
            FlowCheckerPass.Vars.makeSets(beforeV.getDa(),
                                          cvExit.getDua());
        w.setLive(true);
        w.setVars(fv);
        w.process(getFinally());

        if (w.isLive()) {
            w.releasePendingTransfers(this);
            w.setExns(bodyExns.union(w.getExns()).union(beforeExns));
        } else {
            if (! getOptions().strict) {
                w.setExns(bodyExns.union(w.getExns()).union(beforeExns));
            } else {
                w.setExns(w.getExns().union(beforeExns));
            }
        }
        boolean liveSoFar = bodyLive & w.isLive();
        w.setVars(FlowCheckerPass.Vars.makeSets(cvExit.getDa().union(w.getVars().getDa()),
                                                w.getVars().getDua()));
        w.setLive(liveSoFar);
    }

    // ------------------------------
    // Intro: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        w.enterContext(this);
        setBody((Stmt) w.process(getBody()));
        w.leaveContext();
        boolean bodyLive = w.isLive();
        w.setLive(true);
        setFinally((Stmt) w.process(getFinally()));
        boolean liveSoFar = bodyLive & w.isLive();
        w.setLive(liveSoFar);
    }

    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        if (getBody() instanceof EmptyStmt) {
            return getFinally();
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label startBody = cb.genAnchor();
        Label endBody = cb.genAnchor();
        Label startExn = cb.genAnchor();
        Label startFinally = cb.genAnchor();
        Label end = cb.genLabel();

        cb.emitLabel(startBody);
        cb.enterWindingContext(this, startFinally);
        getBody().cgTop(cb);
        cb.leaveContext();
        cb.emitLabel(endBody);
        if (body.completesNormally()) {
            cb.emitJSR(startFinally);
            if (getFinally().completesNormally()) {
                cb.emitJump(end);
            }
        }            

        cb.emitLabel(startExn);  // where we go if an exn happens
        cb.pushStack(1);          // there's an exn on the stack
        cb.emitASTORE(retValLoc); // so store it
        cb.emitJSR(startFinally); // make the jump
        cb.emitALOAD(retValLoc);  // reload the exn
        cb.emitATHROW();          // rethrow

        cb.emitLabel(startFinally);          // always target of JSR instruction
        cb.pushStack(1);                     // stack holds JSR return address
        cb.emitASTORE(retPtrLoc);            // store it.
        getFinally().cgTop(cb);              // execute the body
        if (getFinally().completesNormally()) {
            cb.emitRET(retPtrLoc);           // return from JSR
        }
        cb.addHandler(startBody, endBody, startExn, null); // add default handler

        cb.emitLabel(end);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    // We need to allocate space for the entry to a finally block,
    // which could be entered in two different ways.

    // If it is entered abruptly, through an exception being thrown,
    // we need space to squirrel away the exception, do the JSR, store
    // the RET pointer, and go on.

    // If it is entered normally, then it is possible that it was
    // entered through "return".  Which means that we need to
    // "allocate" space for the return value.  The return code will
    // actually squirrel away in this point, but we need to allocate
    // the space now.  And on top of _that_, we need to save space to
    // store the RET pointer.

    // For now, we'll save _two_ spaces for the return value.  Really,
    // we should be allocating (max 1 retvalSlotSize).

    // needless to say, this is weird enough that we're going to walk
    // it ourselves.

    private int retValLoc;  // same as exn loc
    private int retPtrLoc;

    public int getRetValLoc() { return retValLoc; }

    public void walkFrameLoc(FrameLocPass walker) {
        boolean inTryFinally = walker.inTryFinally;
        walker.inTryFinally = true;
        int start = walker.getfs();
        walker.process(getBody());
        //if (walker.getfs() != start) throw new RuntimeException("inv. broken");
        retValLoc = walker.allocate(2); // Math.max(1, retvalSlotSize)
        retPtrLoc = walker.allocate(1);
        walker.process(getFinally());
        walker.inTryFinally = inTryFinally;
        walker.setfs(start);
    }    
    
    //BEGIN: Generated from @child and @property
    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }
    
    protected Stmt _finally;
    public Stmt getFinally() { return _finally; }
    public void setFinally(Stmt __finally) {
        if (__finally != null) __finally.setParent(this);
        _finally = __finally;
    }
    
    public TryFinallyStmt(SourceLocation location, Stmt _body, Stmt __finally) {
        super(location);
        setBody(_body);
        setFinally(__finally);
    }
    protected TryFinallyStmt(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        TryFinallyStmt ret = new TryFinallyStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        if (_finally != null) ret.setFinally( (Stmt)walker.process(_finally) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return body;
        case 1: return _finally;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "body";
        case 1: return "finally";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setBody((Stmt)child); return;
        case 1: setFinally((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "TryFinallyStmt()";
    }
    
    //END: Generated from @child and @property
}


