/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import java.util.*;
import org.aspectj.compiler.base.cst.*;

/**
 * @grammar ...
 */
public abstract class TypeD extends ASTObject {
    public abstract Type getType();

    public String getString() {
        return getType().getString();
    }

    public String toShortString() {
        return getType().toShortString();
    }

    public final void checkSpec() {
        //showWarning("lex: " + toShortString()); 

        //!!! a lot of caution here to avoid breaking things
        if (fromSource() &&
            getType() != null &&
            !getType().isAnyType() &&
            !getType().isMissing() &&
            getType().getTypeDec() != null &&
            !getType().getTypeDec().isAccessible(this)) {
            showError("can't access " + getType().getTypeDec().toShortString());
        }
    }

    //XXX want this to go away
    public boolean isEquivalent(TypeD other) {
        return getType().isEquivalent(other.getType());
    }

    private static final int UNQUALIFIED = 1;
    private static final int FULLY_QUALIFIED = 0;
    private static final int NULL_TRICK = 2;

    private int unparseRule; // = FULLY_QUALIFIED;
    private boolean checkedQualification = false;

    void makePublic(TypeDec typeDec) {
        //XXX
        if (typeDec == null) return;

        typeDec.getModifiers().setPublic(true);
        if (typeDec.getEnclosingTypeDec() != null) {
            makePublic(typeDec.getEnclosingTypeDec());
        }
    }


    void makeAccessible(Type type, ASTObject fromWhere) {
        //TypeDec fromTypeDec = fromWhere.getBytecodeTypeDec();
        //getCompiler().showMessage("  isAccessible? " + type + " from " + fromTypeDec);
        //if (fromTypeDec == null || type.getTypeDec() == null) return;

        if (type.isAccessible(fromWhere, true)) return ;

        getCompiler().showMessage("  fixing access to type: " + type.toShortString());

        makePublic(type.getTypeDec());
    }


    //INTRO from NameHygienePass
    public ASTObject postScope(ScopeWalker walker) {
        if (!(walker instanceof NameHygienePass)) return this;

        NameHygienePass pass = (NameHygienePass)walker;

        //System.out.println("postHygeniene on " + this);

        if (checkedQualification) {
            //showError("already checked qualification");
            //System.out.println("already checked qualification: " + getType());
            unparseRule = FULLY_QUALIFIED;
            return this;
        }
        checkedQualification = true;
        makeAccessible(getType(), this);
        if (getType().getTypeDec() != null && getType().getTypeDec().hasGlobalName()) {
            Type checkType = getType().getOutermostType();

            if (pass.canUseUnqualifiedName(checkType, this)) {
                unparseRule = UNQUALIFIED;
            } else if (pass.mustUseNullTrick(checkType, this)) {
                unparseRule = NULL_TRICK;
            } else {
                unparseRule = FULLY_QUALIFIED;
            }
        } else {
            unparseRule = FULLY_QUALIFIED;
        }
        //System.out.println("--------->" + unparseRule);
        return this;
    }


    public void unparse(CodeWriter writer) {
        unparse(writer, false);
    }

    public void unparse(CodeWriter writer, boolean inExprPosition) {

        Type type = getType();
        int realUnparseRule = unparseRule;
        if (!writer.isOnlySignatures()) {
            if (!checkedQualification) {
                //showWarning("didn't check qualification: " + unparseRule + " type: " + getType());
                //writer.write("/*!!!*/");
            }
        }

        if (!inExprPosition && realUnparseRule == NULL_TRICK) {
            realUnparseRule = FULLY_QUALIFIED;
        }

        switch (realUnparseRule) {
            case UNQUALIFIED:
                writer.write(getType().getExtendedId().replace('$', '.'));
                break;
            case FULLY_QUALIFIED:
                writer.write(getType().getString());
                break;
            case NULL_TRICK:
                // this pattern is the only way that I know to trick Java into treating
                // this as a type name
                writer.write("((");
                writer.write(getType().getString());
                writer.write(")null)");
                break;
        }
    }
    /*
    private void writeTypeName(CodeWriter writer, String name, boolean inExprPosition) {
        if (!inExprPosition) {
            writer.write(name);
            return;
        }

        String prefix = name;
        int dot = prefix.indexOf('.');
        if (dot != -1) prefix = prefix.substring(0, dot);

        if (getContext().findVarDec(prefix, this) != null) {
            // this pattern is the only way that I know to trick Java into treating
            // this as a type name
            writer.write("((");
            writer.write(name);
            writer.write(")null)");
        } else {
            writer.write(name);
            return;
        }
    }
    */

    //BEGIN: Generated from @child and @property

    public TypeD(SourceLocation location) {
        super(location);

    }


    public String getDefaultDisplayName() {
        return "TypeD()";
    }

    //END: Generated from @child and @property
}
