/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bcg;

import org.aspectj.compiler.base.bcg.pool.*;

import java.lang.reflect.Modifier;

import java.util.ArrayList;
import java.util.List;
import java.util.HashSet;
import java.util.Iterator;

import java.io.IOException;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.File;

import org.aspectj.compiler.base.ast.SourceLocation;
import org.aspectj.compiler.base.JavaCompiler;

// XXX annoying big modularity violation.
import org.aspectj.compiler.base.ast.RefType;
import org.aspectj.compiler.base.ast.NameType;
import org.aspectj.compiler.base.ast.TypeDec;

//XXX JBuilder's compiler doesn't really understand inner classes well enough
//XXX for this to work, so we use the semi-qualified name below
//XXX but look at ForStmt.java for some interesting caveats... ?????????????

//import org.aspectj.compiler.base.bcg.ConstantPool.Constant;

/** Lifecycle: creation, addition of everything, output to file,
    processing of inner classes.  It is an error for these steps to
    occur out of order, but this is not currently checked. */
public class ClassfileBuilder {

    private JavaCompiler compiler;
    public ClassfileBuilder(JavaCompiler compiler) {
        this.compiler = compiler;
    }
    public JavaCompiler getCompiler() { return compiler; }

    // ------------------------------
    // state

    private int magic = (int)0xcafebabeL;
    private short minorVersion = 3;
    private short majorVersion = 45;
    ConstantPool pool = new ConstantPool(this);
    private int accessFlags = 0; // 0x0020; // ACC_SUPER
    private Constant thisClass;
    private Constant superClass;
    private /* Constant */ List interfaces = new ArrayList();
    private /* FieldBuilder */ List fields = new ArrayList();
    private /* MethodBuilder */ List methods = new ArrayList();
    private Attributes attributes = new Attributes(pool);

    // used for delaying generation of inner types until we've closed our port.
    private /* TypeDec */ List innerTypeDecs = new ArrayList();

    // ------------------------------
    // building methods

    public void addAccessFlags(int newFlags) {
        accessFlags = accessFlags | newFlags;
    }
    public void setClassName(NameType nameType) {
        thisClass = pool.addClass(nameType);
    }
    public void setSuperClassName(NameType nameType) {
        superClass = pool.addClass(nameType);
    }
    public void addInterface(NameType nameType) {
        interfaces.add(pool.addClass(nameType));
    }
    public FieldBuilder getFieldBuilder() {
        return new FieldBuilder(this);
    }
    public void addField(FieldBuilder maker) {
        fields.add(maker);
    }
    public MethodBuilder getMethodBuilder() {
        return new MethodBuilder(this);
    }
    public void addMethod(MethodBuilder maker) {
        methods.add(maker);
    }
    public void setSynthetic() {
        attributes.addSyntheticAttribute();
    }
    public void setSourceFile(SourceLocation loc) {
        sourceFiles = new SourceFile(loc);
    }
    public void setSourceDebugExtension(String data) {
    	attributes.addSouceDebugExtensionAttribute(data);
    }

    // ------------------------------
    // irritating inner class stuff

    private /* NameType */ HashSet innerClasses = new HashSet();
    public void addInnerClassRef(NameType nameType) {
        if (! innerClasses.contains(nameType)) {
            innerClasses.add(nameType);
            attributes.addToInnerClassesAttribute(nameType);
        }
    }

    public void delayInnerClassGeneration(TypeDec typeDec) {
        addInnerClassRef(typeDec.getNameType());
        innerTypeDecs.add(typeDec);
    }

    public void generateBytecodeForDelayedInnerClasses(File outputDir) throws IOException {
        for (Iterator i = innerTypeDecs.iterator(); i.hasNext(); ) {
            TypeDec typeDec = (TypeDec) i.next();
            typeDec.generateBytecode(outputDir);
        }
    }

    // ------------------------------
    // output methods

    public void writeTo(DataOutputStream stream) throws IOException {
        stream.writeInt(magic);
        stream.writeShort(minorVersion);
        stream.writeShort(majorVersion);
        pool.writeTo(stream);
        stream.writeShort(accessFlags);
        thisClass.writeIndex(stream);
        superClass.writeIndex(stream);
        stream.writeShort((short)interfaces.size());
        for (Iterator i = interfaces.iterator(); i.hasNext(); ) {
            Constant iface = (Constant) i.next();
            iface.writeIndex(stream);
        }
        stream.writeShort((short)fields.size());
        for (Iterator i = fields.iterator(); i.hasNext(); ) {
            FieldBuilder field = (FieldBuilder) i.next();
            field.writeTo(stream);
        }
        stream.writeShort((short)methods.size());
        for (Iterator i = methods.iterator(); i.hasNext(); ) {
            MethodBuilder method = (MethodBuilder) i.next();
            method.writeTo(stream);
        }
        attributes.writeTo(stream);
        if (compiler.getOptions().bcgtrace) {
            print();
        }
    }

    // ------------------------------
    // debugging printing

    void print() {
        display(0, false);
        System.err.println();
    }

    static void between(int indent, boolean inline) {
        if (inline) {
            System.err.print(" ");
        } else {
            System.err.println();
            for (int s=indent; s >= 0; s--) System.err.print(" ");
        }
    }

    void display(int indent, boolean inline) {
        indent += 2;
        System.err.print("(class "); thisClass.display(0, true);
        between(indent, inline);
        System.err.print("(modifiers " + Modifier.toString(accessFlags) + ")");
        between(indent, inline);
        System.err.print("(extends "); superClass.display(0, true); System.err.print(")");
        between(indent, inline);
        System.err.print("(implements");
        ConstantPool.display(interfaces, 0, true);
        System.err.print(")");
        between(indent, inline);
        System.err.print("(members");
        for (Iterator i = fields.iterator(); i.hasNext(); ) {
            FieldBuilder b = (FieldBuilder) i.next();
            between(indent + 2, inline);
            b.display(indent + 2, inline);
        }
        for (Iterator i = methods.iterator(); i.hasNext(); ) {
            MethodBuilder b = (MethodBuilder) i.next();
            between(indent + 2, inline);
            b.display(indent + 2, inline);
        }
        System.err.print(")");
        between(indent, inline);
        attributes.display(indent, inline);
        between(indent, inline);
        pool.display(indent, inline);
        System.err.print(")");
    }

    // ------------------------------
    // line number table resolution

    // we need to resolve stuff for the line number table and local
    // variable table.  Before resolution, the LineNumberTable holds
    // maximum lines.  Afterward, it holds offsets.  The Local
    // Variable table attribute is simply created at this time.

    public void resolve() {
        pool.resolve();
        int offset = 0;
        int fileNumber = 0;
        for (SourceFile t = sourceFiles; t != null; t = t.next) {
            int i = t.i;
            t.offset = offset;
            t.fileNumber = fileNumber++;
            offset = offset + i + 1000 - (i % 1000);
        }
        for(Iterator i = methods.iterator(); i.hasNext();) {
            MethodBuilder m = (MethodBuilder)i.next();
            m.resolve();
        }
        String simpleName = sourceFiles.getSimpleName();
        String s = simpleName;
        for (SourceFile t = sourceFiles.next; t != null; t = t.next) {
            s += ";" + t.getFullName() + '(' + (t.offset / 1000) + "k)";
        }
        attributes.addSourceFileAttribute(s);
        attributes.addSouceDebugExtensionAttribute(createSourceDebugExtensionData(simpleName));
    }

    private String createSourceDebugExtensionData(String simpleName) {
    	StringBuffer s = new StringBuffer("SMAP\n");
    	s.append(simpleName + "\n");
    	s.append("AspectJ\n");

    	//if (sourceFiles.next != null && sourceFiles.getFileNumber() != 0) {
    		s.append("*S AspectJ\n");
    		s.append("*F\n");
    		for (SourceFile t = sourceFiles; t != null; t = t.next) {
				if (t.getFileNumber() == 0) continue;
                if (t.sourcePackage == null) {
                    s.append(t.getFileNumber() + " " + t.getSimpleName() + "\n");
                } else {
                    s.append("+ " + t.getFileNumber() + " " + t.getSimpleName() + "\n" 
                             + t.getFullName() + "\n");
                }
    		}
			s.append("*L\n");
    		for (SourceFile t = sourceFiles; t != null; t = t.next) {
				if (t.getFileNumber() == 0) continue;
				s.append("1#" + t.getFileNumber() + "," + (t.i) + ":"
							+ (t.offset+1) + ",1\n");
    		}
    	//}
		s.append("*E\n");
		return s.toString();
    }

    // this class is shared by CodeBuilder
    static class SourceFile {
        SourceFile next = null;
        final String sourcePackage;
        final String sourceFile;
        int fileNumber = 0;
        int i = 0;
        int offset = 0;
        String getSimpleName() { return sourceFile; }
        String getFullName() {
            if (sourcePackage == null) return sourceFile;
            else return sourcePackage.replace('.', '/') + '/' + sourceFile;
        }
        SourceFile(SourceLocation loc) {
            if (loc.hasSource()) {
                this.sourcePackage = loc.getCompilationUnit().getPackageName();
                this.sourceFile = loc.getSourceFile().getName();
            } else {
                this.sourcePackage = null;
                this.sourceFile = "NoSourceFile";
            }
        }
        SourceFile(SourceLocation loc, int i) {
            this(loc); this.i = i;
        }
        int getFileNumber() { return fileNumber + 1; /* !!!!!!!!!! */ }
        int getOutputLine(int line) { return line + i; }
        SourceFile get(SourceLocation loc) {
            String currPackage = loc.getCompilationUnit().getPackageName();
            if (((sourcePackage == null)
                 ? currPackage == null
                 : sourcePackage.equals(currPackage))
                && sourceFile.equals(loc.getSourceFile().getName())) {
                int line = loc.getBeginLine();
                if (line > i) i = line;
                return this;
            } else if (next == null) {
                next = new SourceFile(loc, loc.getBeginLine());
                return next;
            } else {
                return next.get(loc);
            }
        }
    }

    SourceFile sourceFiles;

    SourceFile getSourceFile(SourceLocation loc) {
        if (! loc.hasSource()) return null; // ??? is this right?
        return sourceFiles.get(loc);
    }

    // ------------------------------
    // disassembly methods

    void readFrom(DataInputStream stream) throws IOException {
        int magic = stream.readInt();
        minorVersion = (short)stream.readUnsignedShort();
        majorVersion = (short)stream.readUnsignedShort();
        pool.readFrom(stream);
        accessFlags = stream.readUnsignedShort();
        thisClass = pool.get(stream.readUnsignedShort());
        superClass = pool.get(stream.readUnsignedShort());
        for (int i = stream.readUnsignedShort(); i > 0; i--) {
            interfaces.add(pool.get(stream.readUnsignedShort()));
        }
        for (int i = stream.readUnsignedShort(); i > 0; i--) {
            fields.add(getFieldBuilder().readFrom(stream));
        }
        for (int i = stream.readUnsignedShort(); i > 0; i--) {
            methods.add(getMethodBuilder().readFrom(stream));
        }
        attributes.readFrom(stream);
    }

    public static void main(String[] args) throws Exception {
        ClassfileBuilder cfb = new ClassfileBuilder(null);
        cfb.readFrom(new DataInputStream(new java.io.FileInputStream(args[0])));
        cfb.print();
    }
}
