/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bcg;

import org.aspectj.compiler.base.bcg.pool.*;
import java.lang.reflect.Modifier;

import java.io.IOException;
import java.io.DataOutputStream;

// XXX annoying big modularity violation.  
import org.aspectj.compiler.base.ast.NameType;

//import org.aspectj.compiler.base.bcg.ConstantPool.Constant;

/** Lifecycle: creation, addition of code (through the addX methods),
    resolution of constant pool (done in the constant pool, when all
    constant pool entries are set), output.  It is an error for these
    steps to occur out of order, but this is not currently checked. */
public class MethodBuilder {

    // ------------------------------
    // state

    ConstantPool pool;
    ClassfileBuilder cfb;
    Attributes attributes;

    MethodBuilder(ClassfileBuilder cfb) {
        this.cfb = cfb;
        this.pool = cfb.pool;
        attributes = new Attributes(pool);
    }

    int accessFlags = 0;
    Constant name;
    Constant descriptor;

    // ------------------------------
    // output methods

    public void writeTo(DataOutputStream stream) throws IOException {
        stream.writeShort((short) accessFlags);
        name.writeIndex(stream);
        descriptor.writeIndex(stream);
        attributes.writeTo(stream);
    }

    public String toString() {
        return "(method " +
            Modifier.toString(accessFlags) + " " +
            name + " " +
            descriptor + " " + attributes + ")";
    }

    void display(int indent, boolean inline) {
        indent += 2;
        System.err.print("(method "); name.display(0, true); 
        System.err.print(" "); descriptor.display(0, true);
        between(indent, inline);
        System.err.print("(modifiers " + Modifier.toString(accessFlags) + ")");
        between(indent, inline);
        attributes.display(indent, inline);
        System.err.print(")");
    }

    static void between(int indent, boolean inline) {
        if (inline) {
            System.err.print(" ");
        } else {
            System.err.println();
            for (int s=indent; s >= 0; s--) System.err.print(" ");
        }
    }

    // ------------------------------
    // building methods

    public void addAccessFlags(int newFlags) {
        accessFlags = accessFlags | newFlags;
    }

    public void setName(String name) {
        this.name = pool.addUtf8(name);
    }

    public void setDescriptor(String descriptor) {
        this.descriptor = pool.addUtf8(descriptor);
    }

    public void setDeprecated() {
        attributes.addDeprecatedAttribute();
    }

    public void setSynthetic() {
        attributes.addSyntheticAttribute();
    }

    public void addToExceptionsAttribute(NameType exn) {
        attributes.addToExceptionsAttribute(exn);
    }

    public CodeBuilder getCodeBuilder() {
        return new CodeBuilder(cfb);
    }

    CodeBuilder cb = null;
    public void setCode(CodeBuilder cb) {
        this.cb = cb;
        attributes.addCodeAttribute(cb);
    }

    void resolve() {
        if (cb != null) cb.resolve();
    }

    // ------------------------------
    // disassembly methods

    MethodBuilder readFrom(java.io.DataInputStream stream) throws java.io.IOException {
        accessFlags = stream.readUnsignedShort();
        name = pool.get(stream.readUnsignedShort());
        descriptor = pool.get(stream.readUnsignedShort());
        attributes.readFrom(stream);
        return this;
    }
}
