/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bcg.pool;

import org.aspectj.compiler.base.bcg.ClassfileBuilder;
import org.aspectj.compiler.base.bcg.Asserts;

import java.io.*;
import java.util.*;

import org.aspectj.compiler.base.ast.ArrayType;
import org.aspectj.compiler.base.ast.NameType;
import org.aspectj.compiler.base.ast.RefType;

public final class ConstantPool {

    // ------------------------------
    // state

    private boolean isResolvedFlag = false;
    public boolean isResolved() { return isResolvedFlag; }

    // ------------------------------
    // state
    private ClassfileBuilder cfb;
    public ConstantPool(ClassfileBuilder cfb) {
        this.cfb = cfb;
    }

    private List table = new ArrayList();
    {
        table.add(Placeholder.it);
    }
    private Map cache = new Hashtable();

    public void resolve() {
        if (isResolved()) {
            throw new RuntimeException("already resolved");
        }
        Object[] a = table.toArray();
        table.clear();
        Arrays.sort(a);
        table.add(Placeholder.it);
        for (int i = 0, len = a.length; i < len; i++) {
            add1((Constant) a[i]);
        }
        isResolvedFlag = true;
    }

    public Constant get(int i) { return (Constant) table.get(i); }

    Constant get(int i, Constant young) { 
        Constant c = (Constant) table.get(i);
        if (c instanceof Placeholder) {
            table.set(i, young);
            young.setIndex(i);
            return young;
        } else {
            return c;
        }
    }

    private Constant add(Constant constant) {
        Constant cachedConstant = (Constant) cache.get(constant);
        if (cachedConstant != null) {
            return cachedConstant;
        } else {
            add1(constant);
            return constant;
        }
    }

    private void add1(Constant constant) {
        if (constant instanceof Placeholder) return;
        Asserts.assertUU2(table.size(), "Too many constants in constant pool");
        constant.setIndex(table.size());
        cache.put(constant, constant);
        table.add(constant);
        if (constant.isBig()) {
            table.add(Placeholder.it);
        }
    }


    // ------------------------------
    // output

    public static void writeIndex(Constant constant, DataOutputStream stream) 
            throws IOException {
        if (constant == null) {
            stream.writeShort(0);
        } else {
            constant.writeIndex(stream);
        }
    }

    public void writeTo(DataOutputStream stream) throws IOException {
        if (! isResolved()) {
            throw new RuntimeException("must resolve constant pool first");
        }
        stream.writeShort((short) table.size());
        for (Iterator i = table.iterator(); i.hasNext(); ) {
            Constant c = (Constant) i.next();
            c.writeTo(stream);
        }
    }

    // ------------------------------
    // debugging

    public void display(int indent, boolean inline) {
        System.err.print("(pool");
        display(table, indent + 2, inline);
        System.err.print(")");
    }

    public static void display(List l, int indent, boolean inline) {
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            Constant c = (Constant) i.next();
            if (c instanceof Placeholder) continue;
            between(indent, inline);
            c.display(indent, inline);
        }
    }

    static void between(int indent, boolean inline) {
        if (inline) {
            System.err.print(" ");
        } else {
            System.err.println();
            for (int s=indent; s >= 0; s--) System.err.print(" ");
        }
    }        

    // ------------------------------
    // building methods

    // ---------- classes

    public ClassConstant addClass(RefType refType) {
        if (refType instanceof ArrayType)
            return addClass((ArrayType)refType);
        else
            return addClass((NameType)refType);
    }

    /** only called for ANEWARRAY, MULTINEWARRAY, CHECKCAST, INSTANCEOF bytecodes */
    public ClassConstant addClass(ArrayType arrayType) {
        return addClassNoInnerCheck(arrayType);
    }

    public ClassConstant addClass(NameType nameType) {
        if (! nameType.isPackageMember()) {
            cfb.addInnerClassRef(nameType);
        }
        return addClassNoInnerCheck(nameType);
    }

    public ClassConstant addClassNoInnerCheck(RefType refType) {
        return (ClassConstant) add(new ClassConstant(addUtf8(refType.getInternalName())));
    }

    // ---------- others

    public Utf8Constant addUtf8(String name) {
        return (Utf8Constant) add(new Utf8Constant(name));
    }

    public StringConstant addString(String value) {
        return (StringConstant) add(new StringConstant(addUtf8(value)));
    }
    public IntConstant addInt(int value) {
        return (IntConstant) add(new IntConstant(value));
    }
    public FloatConstant addFloat(float value) {
        return (FloatConstant) add(new FloatConstant(value));
    }
    public LongConstant addLong(long value) {
        return (LongConstant) add(new LongConstant(value));
    }
    public DoubleConstant addDouble(double value) {
        return (DoubleConstant) add(new DoubleConstant(value));
    }

    public MethodrefConstant addMethodRef(NameType nameType, String name, String descriptor) {
        return (MethodrefConstant)
            add(new MethodrefConstant(addClass(nameType),
                                      addNameAndType(name, descriptor)));
    }

    public InterfaceMethodrefConstant
          addInterfaceMethodRef(NameType nameType, String name, String descriptor) {
        return (InterfaceMethodrefConstant)
            add(new InterfaceMethodrefConstant(addClass(nameType),
                                               addNameAndType(name, descriptor)));
    }

    public FieldrefConstant addFieldRef(NameType nameType, String name, String descriptor) {
        return (FieldrefConstant)
            add(new FieldrefConstant(addClass(nameType),
                                     addNameAndType(name, descriptor)));
    }

    public NameAndTypeConstant addNameAndType(String name, String descriptor) {
        return (NameAndTypeConstant)
            add(new NameAndTypeConstant(addUtf8(name), addUtf8(descriptor)));
    }

    // ------------------------------
    // disassembly methods

    public void readFrom(DataInputStream stream) throws IOException {
        int size = stream.readUnsignedShort();
        table.clear();
        for (int i = size - 1; i >= 0; i--) table.add(Placeholder.it);
        for (int i = 1; i < size; i++) {
            int tag = stream.readUnsignedByte();
            Constant c;
            switch (tag) {
            case Constant.UTF_TAG:
                c = get(i, new Utf8Constant()); break;
            case Constant.INT_TAG:
                c = get(i, new IntConstant()); break;
            case Constant.FLOAT_TAG:
                c = get(i, new FloatConstant()); break;
            case Constant.LONG_TAG:
                c = get(i, new LongConstant()); i++; break;
            case Constant.DOUBLE_TAG:
                c = get(i, new DoubleConstant()); i++; break;
            case Constant.CLASS_TAG:
                c = get(i, new ClassConstant()); break;
            case Constant.STRING_TAG:
                c = get(i, new StringConstant()); break;
            case Constant.FIELD_REF_TAG:
                c = get(i, new FieldrefConstant()); break;
            case Constant.METHOD_REF_TAG:
                c = get(i, new MethodrefConstant()); break;
            case Constant.INTERFACE_METHOD_REF_TAG:
                c = get(i, new InterfaceMethodrefConstant()); break;
            case Constant.NAME_AND_TYPE_TAG:
                c = get(i, new NameAndTypeConstant()); break;
            default:
                throw new RuntimeException("unknown constant pool tag " + tag);
            }
            c.readFrom(stream, this);
        }
        for (Iterator i = table.iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof Placeholder) continue;
            cache.put(o, o);
        }
    }
}
