/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CompilerObject;
import org.aspectj.compiler.base.parser.SourceInfo;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;


import java.util.*;
import java.lang.reflect.Modifier;

public class ASTConnection extends CompilerObject {

    public ASTConnection(JavaCompiler compiler) {
        super(compiler);
    }
    
    protected SourceLocation dummySource = new DummySourceLocation(getCompiler());

    // at the top are more general methods that might be moved elsewhere

    private TypeDs classesToTypeDs(ConstantPool.Class[] classes) {
        if (classes == null || classes.length == 0) {
            return null;
        }

        TypeDs names = new TypeDs(dummySource);

        final int N = classes.length;
        for(int i=0; i<N; i++) {
            //System.out.println(i + ": " + makeTypeD(classes[i]));
            names.add(makeTypeD(classes[i]));
        }
        return names;
    }

    // at the bottom are clear AST connections
    public TypeDec makeTypeDec(ClassFile classFile) {
        int modifiers = classFile.modifiers;
        boolean isInterface = Modifier.isInterface(modifiers);
        modifiers = modifiers & ~Modifier.INTERFACE;
        Decs body = new Decs(dummySource);

        String fullName = classFile.thisClass.getName();
        String name = fullName;
        String packageName = null;
        int dot = fullName.lastIndexOf('.');
        if (dot != -1) {
            packageName = fullName.substring(0, dot);
            name = fullName.substring(dot+1);
        }
        
        String id = name;
        String enclosingName = null;
        int dollar = id.lastIndexOf('$');
        if (dollar != -1) {
            name = id.substring(dollar+1);
            enclosingName = id.substring(0, dollar);
        }

        CompilationUnit cu = new CompilationUnit(dummySource,
                packageName, null, new Decs(dummySource), 
                new SourceInfo(null, null),true);
        SourceLocation loc = new TextSourceLocation(cu, -1, -1);
        //cu.setContext(new SourceContext(getCompiler(),cu));

        //System.err.println(packageName+" :: "+name);

        TypeD superClass;
        if (classFile.superClass == null) {
            superClass = null;
        } else {
            superClass = makeTypeD(classFile.superClass);
        }

        TypeDs _implements = classesToTypeDs(classFile.interfaces);

        TypeDec typeDec;
        if (isInterface) {
            typeDec = getAST().makeInterface(getAST().makeModifiers(modifiers), 
                        /*fullName,*/ name, _implements);
        } else {
            typeDec = getAST().makeClass(getAST().makeModifiers(modifiers), 
                        /*fullName,*/ name, superClass, _implements);
        }
        typeDec.setBody(body);

        typeDec.setSourceLocation(loc);
        if (enclosingName != null) {
            Type t = getTypeManager().getType(packageName, enclosingName);
            t.getTypeDec().getBody().add(typeDec);
            typeDec.setEnclosingTypeDec(t.getTypeDec());
        } else {
            cu.getDecs().add(typeDec);
        }
        //typeDec.setContext(cu.getContext());

        //Context typeContext = new TypeContext(getCompiler(),typeDec.getContext(), typeDec.getType());
        //typeDec.fullName = fullName;
        //typeDec.cacheDecs();
        //body.setContext(typeContext);

        for(int i=0; i<classFile.fields.length; i++) {
            FieldDec dec = makeFieldDec(classFile.fields[i]);
            dec.setSourceLocation(loc);
            //dec.setContext(typeContext);
            body.add(dec);
            //typeDec.addDec(dec, true);
            //body.append(makeFieldDec(classFile.fields[i]));
        }

        for(int i=0; i<classFile.methods.length; i++) {
            CodeDec dec = makeCodeDec(classFile.methods[i]);
            if (dec == null) {
                //!!!ystem.err.println("null code in "+fullName);
                continue;
            }
            dec.setSourceLocation(loc);
            //dec.setContext(typeContext);
            body.add(dec);
            //typeDec.addDec(dec, true);
            //body.append(makeMethodDec(classFile.methods[i]));
        }

        Attributes.InnerClassesAttribute attr =
            (Attributes.InnerClassesAttribute)
            classFile.attributes.findAttribute(Attributes.InnerClassesAttribute.class);
        if (attr != null) {
            typeDec.memberTypeNames = attr.getContainedTypeNames(fullName);
//              System.out.println("Inners of class " + fullName); 
//              System.out.println(attr);
//              System.out.println("  " + typeDec.memberTypeNames);
            if (! attr.isPackageLevel(fullName)) {
                Modifiers m = new Modifiers(dummySource, attr.getModifiers(fullName));
//                  System.out.println("  " + "modifiers " + m);
                typeDec.setModifiers(m);
                if (! m.isStatic() && ! (typeDec instanceof InterfaceDec)) {
                    for (Iterator i = typeDec.getBody().iterator(); i.hasNext(); ) {
                        Object dec = i.next();
                        if (dec instanceof ConstructorDec) {
                            ConstructorDec cdec = (ConstructorDec) dec;
                            Formals f = cdec.getFormals();
                            cdec.setEnclosingInstanceFormal(f.get(0));
                            f.remove(0);
                        }
                    }
                    typeDec.setInnerDiscoveries(false, true, false);
                } else {
                    typeDec.setInnerDiscoveries(false, false, false);
                }
            } else {
                typeDec.setInnerDiscoveries(true, false, false);
            }
        } else {
            typeDec.setInnerDiscoveries(true, false, false);
        }

        //typeDec.cacheDecs();
        //typeDec.buildTypeGraph();

        //System.err.println("about to build context for: "+typeDec.getId());

        //typeDec.buildContext(new EmptyContext(), Context.CODE_BLOCKS);

        return typeDec;
    }

    public TypeD makeTypeD(ConstantPool.Class cpClass) {
        String name = cpClass.getName();
        int lastDot = name.lastIndexOf('.');
        if (lastDot == -1) {
            return new LazyNameTypeD(dummySource,null, name);
        } else {
            return new LazyNameTypeD(dummySource,name.substring(0, lastDot), name.substring(lastDot+1));
        }
    }

    public FieldDec makeFieldDec(ClassFile.FieldInfo field) {
        int modifiers = field.modifiers;
        TypeD typeD = makeTypeD(field.descriptor.value);
        String name = field.name.value;
        Expr initializer = null;

        Attributes.ConstantValueAttribute attr =
            (Attributes.ConstantValueAttribute)
            field.attributes.findAttribute(Attributes.ConstantValueAttribute.class);
        if (attr != null) {
            ConstantPool.Entry entry = attr.value;
            if (entry instanceof ConstantPool.String) {
                ConstantPool.String stringEntry = (ConstantPool.String)entry;
                initializer = new StringLiteralExpr(dummySource,stringEntry.value.value);
            } else if (entry instanceof ConstantPool.Float) {
                ConstantPool.Float floatEntry = (ConstantPool.Float)entry;
                initializer = new FloatLiteralExpr(dummySource,floatEntry.value);
            } else if (entry instanceof ConstantPool.Long) {
                ConstantPool.Long longEntry = (ConstantPool.Long)entry;
                initializer = new LongLiteralExpr(dummySource,longEntry.value);
            } else if (entry instanceof ConstantPool.Double) {
                ConstantPool.Double doubleEntry = (ConstantPool.Double)entry;
                initializer = new DoubleLiteralExpr(dummySource,doubleEntry.value);
            } else /* (entry instanceof ConstantPool.Integer) */ {
                ConstantPool.Integer intEntry = (ConstantPool.Integer)entry;
                int init = intEntry.value;
                if (typeD.getType().isBoolean()) {
                    initializer = new BooleanLiteralExpr(dummySource,init != 0);
                } else {
                    initializer = new IntLiteralExpr(dummySource,init);
                }
            }
        }
        return new FieldDec(dummySource,new Modifiers(dummySource,modifiers), typeD, name, initializer);
    }


    public CodeDec makeCodeDec(ClassFile.MethodInfo method) {
        int modifiers = method.modifiers;
        String name = method.name.value;

        TypeDs _throws = classesToTypeDs(method.getExceptions());
//        if (name.equals("finalize")) System.out.println("  .class throws: " + _throws +
//             ", " + method.getExceptions().length + ": " + method.getExceptions()[0]);
        Formals formals = getFormals(method.descriptor.value);

        if (name.equals("<init>")) {
            CodeDec ret = new ConstructorDec(dummySource,new Modifiers(dummySource,modifiers), formals, _throws, null);
            if (method.isDeprecated()) ret.setDeprecated(true);
            return ret;
        } else if (name.equals("<clinit>")) {
            //!!!System.err.println("found class initializer");
            return null;
        }

        //!!! this dualness probably indicates that a signature class is needed
        TypeD typeD = getReturnTypeD(method.descriptor.value);


        CodeDec ret = new MethodDec(dummySource,new Modifiers(dummySource,modifiers), typeD, name,
                             formals, _throws, null);
        if (method.isDeprecated()) {
            ret.setDeprecated(true);
            //System.out.println("added deprecated method: " + ret.toShortString() + ", " + ret.isDeprecated());
        }

        return ret;
    }

    TypeD getReturnTypeD(String descriptor) {
        int lastParen = descriptor.lastIndexOf(')');
        return makeTypeD(descriptor.substring(lastParen+1));
    }

    class IndexedChars {
        char[] chars;
        int index;
        public IndexedChars(char[] chars) {
            this.chars = chars;
            index = 0;
        }

        public char read() {
            return chars[index++];
        }

        public String readTo(char end) {
            for(int j=index; j<chars.length; j++) {
                if (chars[j] == end) {
                    String ret = new String(chars, index, j-index);
                    index = j+1;
                    return ret;
                }
            }
            String ret = new String(chars, index, chars.length-index);
            index = chars.length;
            return ret;
        }
    }

    TypeD readNameTypeD(IndexedChars ic) {
        char[] chars = ic.chars;
        int startIndex = ic.index;
        int endIndex = startIndex;
        int idStart = startIndex;
        while (endIndex < chars.length) {
            if (chars[endIndex] == '/') {
                chars[endIndex] = '.';
                idStart = endIndex+1;
            }
            if (chars[endIndex] == ';') break;
            endIndex++;
        }

        String id = new String(chars, idStart, endIndex-idStart);
        String pname;
        if (idStart == startIndex) pname = null;
        else pname = new String(chars, startIndex, idStart-startIndex-1);
        //System.err.println("read name: "+pname+", "+id);
        ic.index = endIndex+1;
        return new LazyNameTypeD(dummySource, pname, id);
    }


    TypeD readTypeD(IndexedChars ic) {
        switch(ic.read()) {
        case 'B': return getTypeManager().byteType.makeTypeD();
        case 'C': return getTypeManager().charType.makeTypeD();
        case 'D': return getTypeManager().doubleType.makeTypeD();
        case 'F': return getTypeManager().floatType.makeTypeD();
        case 'I': return getTypeManager().intType.makeTypeD();
        case 'J': return getTypeManager().longType.makeTypeD();
        case 'S': return getTypeManager().shortType.makeTypeD();
        case 'Z': return getTypeManager().booleanType.makeTypeD();
        case 'V': return getTypeManager().voidType.makeTypeD();
        case '[': return new ArrayTypeD(dummySource,readTypeD(ic));
        case 'L': return readNameTypeD(ic);
        }
        throw new RuntimeException("unknown descriptor: "+ic.chars[ic.index-1]+" in "+new String(ic.chars));
    }


    Formals getFormals(String descriptor) {
        IndexedChars ic = new IndexedChars(descriptor.toCharArray());
        ic.index++;
        List formals = new ArrayList();
        int n = 0;
        while (ic.chars[ic.index] != ')') {
            FormalDec formal = new FormalDec(dummySource,new Modifiers(dummySource,0), readTypeD(ic), "__" + n, null);
            n += 1;
            formals.add(formal);
        }
        return new Formals(dummySource,(FormalDec[])formals.toArray(new FormalDec[formals.size()]));
    }

    public TypeD makeTypeD(String descriptor) {
        return readTypeD(new IndexedChars(descriptor.toCharArray()));
    }
}
