/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import java.io.*;
import java.lang.reflect.Modifier;

class Attributes {
    //static NoAttributes = new Attributes(new Attribute[0]);
    Attribute[] data;
    ClassFile classFile;

    static Attributes createFrom(ClassFile classFile, DataInputStream stream) throws IOException {
        Attributes attributes = new Attributes();
        attributes.classFile = classFile;
        attributes.readFrom(stream);
        return attributes;
    }


    public Attribute findAttribute(Class ofType) {
        for(int i=0; i<data.length; i++) {
            if (ofType.isInstance(data[i])) {
                return data[i];
            }
        }
        return null;
    }

    void readFrom(DataInputStream stream) throws IOException {
        int attributeCount = stream.readUnsignedShort();
        //System.err.println("constants: "+constantCount);
        data = new Attribute[attributeCount];
        for(int index = 0; index<attributeCount; index++) {
            int nameIndex = stream.readUnsignedShort();
            //!!! this can be an unsigned int
            long attributeLength = (256*256)*stream.readUnsignedShort() +
                                    stream.readUnsignedShort();
            String name = classFile.pool.getUTF(nameIndex).value;

            Attribute attribute;

            if (name.equals("SourceFile")) {
                attribute = new SourceFileAttribute();
                attribute.readFrom(stream);
            } else if (name.equals("Deprecated")) {
                attribute = new DeprecatedAttribute();
                attribute.readFrom(stream);
            } else if (name.equals("Exceptions")) {
                attribute = new ExceptionsAttribute();
                attribute.readFrom(stream);
            } else if (name.equals("ConstantValue")) {
                attribute = new ConstantValueAttribute();
                attribute.readFrom(stream);
            } else if (name.equals("InnerClasses")) {
                attribute = new InnerClassesAttribute();
                attribute.readFrom(stream);
            } else {
                attribute = new UnknownAttribute(name, attributeLength);
                attribute.readFrom(stream);
            }
            data[index] = attribute;
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("Attributes(");
        for(int i=0; i<data.length; i++) {
            if (i > 0) buf.append(", ");
            buf.append(data[i]);
        }
        buf.append(")");
        return buf.toString();
    }

    abstract class Attribute {
        abstract void readFrom(DataInputStream stream) throws IOException;
    }

    class SourceFileAttribute extends Attribute {
        ConstantPool.UTF filename;

        void readFrom(DataInputStream stream) throws IOException {
            filename = classFile.pool.getUTF(stream.readUnsignedShort());
        }

        public String toString() {
            return "SourceFile("+filename.value+")";
        }
    }

    public class ExceptionsAttribute extends Attribute {
        public ConstantPool.Class[] exceptions;

        void readFrom(DataInputStream stream) throws IOException {
            int n = stream.readUnsignedShort();
            exceptions = new ConstantPool.Class[n];
            for(int i=0; i<n; i++) {
                exceptions[i] = classFile.pool.getClass(stream.readUnsignedShort());
            }
        }

        public String toString() {
            return "Exceptions("+exceptions.length+")";
        }
    }

    class DeprecatedAttribute extends Attribute {
        void readFrom(DataInputStream stream) throws IOException {
            ;
        }

        public String toString() {
            return "Deprecated()";
        }
    }

    class ConstantValueAttribute extends Attribute {
        public ConstantPool.Entry value;

        void readFrom(DataInputStream stream) throws IOException {
            value = classFile.pool.getEntry(stream.readUnsignedShort());
        }
        
        public String toString() {
            return "ConstantValue(...)";
        }
    }

    class InnerClassesAttribute extends Attribute {
        public ConstantPool.Class[] innerClasses;
        public ConstantPool.Class[] outerClasses;
        public ConstantPool.UTF[] innerNames;
        public int[] innerAccessFlags;
            
        void readFrom(DataInputStream stream) throws IOException {
            int size = stream.readUnsignedShort();
            innerClasses = new ConstantPool.Class[size];
            outerClasses = new ConstantPool.Class[size];
            innerNames = new ConstantPool.UTF[size];
            innerAccessFlags = new int[size];

            for (int i = 0; i < size; i++) {
                innerClasses[i] = classFile.pool.getClass(stream.readUnsignedShort());
                outerClasses[i] = classFile.pool.getClass(stream.readUnsignedShort());
                innerNames[i] = classFile.pool.getUTF(stream.readUnsignedShort());
                innerAccessFlags[i] = stream.readUnsignedShort();
            }
        }
        
        public String toString() {
            return "InnerClassesAttribute(...)";
        }
        
        public boolean isPackageLevel(String str) {
            int len = innerAccessFlags.length;
            for (int i = 0; i < len; i++) {
                ConstantPool.Class inner = innerClasses[i];
                if (str.equals(inner.getName())) return false;
            }
            return true;
        }            

        public int getModifiers(String str) {
            int len = innerAccessFlags.length;
            for (int i = 0; i < len; i++) {
                ConstantPool.Class inner = innerClasses[i];
                if (str.equals(inner.getName())) return innerAccessFlags[i];
            }
            throw new RuntimeException("Unknown non-package-level class " + str);
        }            

        public java.util.Set getContainedTypeNames(String str) {
            java.util.HashSet set = new java.util.HashSet();
            int len = innerAccessFlags.length;
            for (int i = 0; i < len; i++) {
                ConstantPool.Class outer = outerClasses[i];
                if (outer == null) continue;
                ConstantPool.UTF name = innerNames[i];
                if (name == null) continue;
                if (str.equals(outer.getName())) {
                    set.add(name.value);
                }
            }
            return set;
        }

        /*
          // Testing code, clean me out when unnecessary
        public void print() {
            System.out.println("  " + innerAccessFlags.length + " inner references");
            for (int i = 0; i < innerAccessFlags.length; i++) {
            String inner = getClassName(innerClasses[i]);
            String outer = getClassName(outerClasses[i]);
            String name =  getUTFValue(innerNames[i]);
            String mods = getMods(innerAccessFlags[i]);
            System.out.println("     " + mods + name + " (" + inner + " in " + outer + ")");
            }
        }
        private String getClassName(ConstantPool.Class x) {
            if (x == null) return "NOWHERE";
            else return x.getName();
        }
        private String getUTFValue(ConstantPool.UTF x) {
            if (x == null) return "ANONYMOUS";
            else return x.value;
        }
        private String getMods(int x) {
            if (x == 0) return "";
            return new org.aspectj.compiler.base.ast.Modifiers(null, x).toString() + " ";
        }
        */
    }

    class UnknownAttribute extends Attribute {
        String name;
        long size;
        public UnknownAttribute(String name, long size) {
            this.name = name;
            this.size = size;
        }
        void readFrom(DataInputStream stream) throws IOException {
            stream.skip(size);
        }

        public String toString() {
            return "UnknownAttribute("+name+", "+size+")";
        }
    }
}
