/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import java.io.*;
import java.lang.reflect.Modifier;

public class ClassFile {
    ConstantPool pool;

    int modifiers;
    ConstantPool.Class thisClass;
    ConstantPool.Class superClass;
    ConstantPool.Class[] interfaces;

    FieldInfo[] fields;
    MethodInfo[] methods;

    Attributes attributes;

    //public ClassFile(String filename) {
    //}
    
    public String getClassName() {
        return thisClass.getName();
    }

    private void assertion(boolean test, String message) {
        if (!test) throw new RuntimeException(message);
    }

    void readConstantPool(DataInputStream stream) throws IOException {
        pool = new ConstantPool();
        pool.readFrom(stream);
    }

    void readInterfaces(DataInputStream stream) throws IOException {
        int interfacesCount = stream.readUnsignedShort();
        interfaces = new ConstantPool.Class[interfacesCount];
        for(int i=0; i<interfacesCount; i++) {
            interfaces[i] = pool.getClass(stream.readUnsignedShort());
        }
    }

    class MemberInfo {
        int modifiers;
        ConstantPool.UTF name;
        ConstantPool.UTF descriptor;
        Attributes attributes;

        void readFrom(DataInputStream stream) throws IOException {
            modifiers = stream.readUnsignedShort();
            name = pool.getUTF(stream.readUnsignedShort());
            descriptor = pool.getUTF(stream.readUnsignedShort());
            attributes = Attributes.createFrom(ClassFile.this, stream);
        }

        public ConstantPool.Class[] getExceptions() {
            Attributes.ExceptionsAttribute excs =
                    (Attributes.ExceptionsAttribute)
                        attributes.findAttribute(Attributes.ExceptionsAttribute.class);
            if (excs == null) return null;
            else return excs.exceptions;
        }

        public boolean isDeprecated() {
            return attributes.findAttribute(Attributes.DeprecatedAttribute.class) != null;
        }


        public String toString() {
            return Modifier.toString(modifiers)+" "+name.value+" "+descriptor.value+": "+attributes;
        }
    }

    class FieldInfo extends MemberInfo {}

    void readFields(DataInputStream stream) throws IOException {
        int fieldsCount = stream.readUnsignedShort();
        fields = new FieldInfo[fieldsCount];
        for(int i=0; i<fieldsCount; i++) {
            fields[i] = new FieldInfo();
            fields[i].readFrom(stream);
        }
    }

    class MethodInfo extends MemberInfo {
        //public ConstantPool.Class[] getExceptions() {
        //
        //}
    }

    void readMethods(DataInputStream stream) throws IOException {
        int methodsCount = stream.readUnsignedShort();
        methods = new MethodInfo[methodsCount];
        for(int i=0; i<methodsCount; i++) {
            methods[i] = new MethodInfo();
            methods[i].readFrom(stream);
        }
    }

    void readAttributes(DataInputStream stream) throws IOException {
        attributes = Attributes.createFrom(this, stream);
    }

    public void readFrom(DataInputStream stream) throws IOException {
        int magic = stream.readInt();
        int minorVersion = stream.readUnsignedShort();
        int majorVersion = stream.readUnsignedShort();
        assertion(magic == 0xCAFEBABE, "bad magic number: "+Integer.toString(magic, 16));
        //XXX think more carefully about these constraints
        //XXX we need to support 45, 46, and 47 at a minimum to handle jdk1.1, 1.2, and 1.3
        assertion(majorVersion >= 45, "bad major version: "+majorVersion);
        //assert(minorVersion >= 3, "bad minor version: "+minorVersion);

        readConstantPool(stream);

        modifiers = stream.readUnsignedShort();

        thisClass = pool.getClass(stream.readUnsignedShort());
        superClass = pool.getClass(stream.readUnsignedShort());

        readInterfaces(stream);
        readFields(stream);
        readMethods(stream);
        readAttributes(stream);
    }

    void print() {
        System.err.println("modifiers: "+Modifier.toString(modifiers));
        System.err.println("this: "+thisClass);
        System.err.println("extends: "+superClass);
        System.err.println("interfaces");
        printObjects(interfaces);
        System.err.println("fields");
        printObjects(fields);
        System.err.println("methods");
        printObjects(methods);
        System.err.println("attributes: "+attributes);
    }

    static void printObjects(Object[] objects) {
        for(int i=0; i<objects.length; i++) {
            System.err.println(""+i+": "+objects[i]);
        }
    }

    public static void main(String[] args) throws Exception {
        ClassFile cf = new ClassFile();
        cf.readFrom(new DataInputStream(new FileInputStream(args[0])));
        cf.print();
        for (int i = 0; i < cf.pool.constants.length; i++) {
            System.err.println(cf.pool.constants[i]);
        }
    }
}
