/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;

import java.io.*;
import java.util.zip.*;
import java.util.StringTokenizer;

import java.util.*;

public class ClassPathManager extends ClassManager {
    List classManagers;

    private ClassPathManager(JavaCompiler compiler, List classManagers) {
        super(compiler);
        this.classManagers = classManagers;
    }

    public ClassPathManager(JavaCompiler compiler, String classpath, 
                            String bootclasspath, String extdirs) {
        super(compiler);
        //this.astConnection = new ASTConnection(compiler);
        // Build up the appropriate ClassManagers for this classpath
        List managers = new LinkedList();

        if (bootclasspath == null) {
            bootclasspath = System.getProperty("sun.boot.class.path", "");
        }
        addClasspathManagers(managers, bootclasspath);

        if (extdirs == null) {
            extdirs = System.getProperty("java.ext.dirs", "");
        }
        addExtensionManagers(managers, extdirs);

        if (classpath == null) {
            classpath = System.getProperty("java.class.path", "");
        }
        addClasspathManagers(managers, classpath);

        classManagers = managers;
    }

    protected void addClasspathManagers(List managers, String classpath) {
        StringTokenizer tok = new StringTokenizer(classpath, java.io.File.pathSeparator);
        while  (tok.hasMoreTokens())  {
            String entry = tok.nextToken().trim();
            //System.out.println("entry: " + entry);
            
            if (entry.endsWith(".jar") || entry.endsWith(".zip")) {
                managers.add(new JarClassManager(getCompiler(), new File(entry)));
            } else {
                managers.add(new DirectoryClassManager(getCompiler(), new File(entry)));
            }
        }
    }


    protected void addExtensionManagers(List managers, String extdirs) {
        StringTokenizer tok = new StringTokenizer(extdirs, java.io.File.pathSeparator);
        while  (tok.hasMoreTokens())  {
            String entry = tok.nextToken().trim();
            addExtensionDirManagers(managers, new File(entry));
        }
    }

    protected void addExtensionDirManagers(List managers, File extdir) {
        if (!extdir.isDirectory()) return;

        String[] names = extdir.list();
        if (names == null) return;

        for(int i=0; i<names.length; i++) {
            // choose to skip all files that don't end in .jar or .zip
            if (!names[i].endsWith(".jar") && !names[i].endsWith(".zip")) {
                continue;
            }

            File file = new File(extdir, names[i]);
            if (file.isFile()) {
                managers.add(new JarClassManager(getCompiler(), file));
            }
        }
    }

    private String join(String[] names) {
        StringBuffer buf = new StringBuffer();
        buf.append(names[0]);
        for(int i=1; i<names.length; i++) {
            buf.append('.');
            buf.append(names[i]);
        }
        return buf.toString();
    }

    public TypeDec findTypeDec(String name) {
        for (Iterator i = classManagers.iterator(); i.hasNext(); ) {
            ClassManager cm = (ClassManager)i.next();
            TypeDec ret = cm.findTypeDec(name);
            if (ret != null) return ret;
        }
        return null;
    }

    public ClassManager makeSubPackageManager(String subPackageName) {
        List ret = new LinkedList();
//         System.out.println("total: " + Runtime.getRuntime().totalMemory() +
//                            ", free: " + Runtime.getRuntime().freeMemory());
        for (Iterator i = classManagers.iterator(); i.hasNext(); ) {
            ClassManager cm = (ClassManager)i.next();
            ClassManager newCM = cm.makeSubPackageManager(subPackageName);
            if (newCM != null) ret.add(newCM);
        }
//         System.out.println("total: " + Runtime.getRuntime().totalMemory() +
//                            ", free: " + Runtime.getRuntime().freeMemory());
//         getCompiler().showMessage("new cpm: " + ret);
        return new ClassPathManager(getCompiler(), ret);
    }

    public Collection/*List<String>*/ getAllPossibleTypeNames() {
    	Collection ret = new ArrayList();
    	for (Iterator i = classManagers.iterator(); i.hasNext(); ) {
            ClassManager cm = (ClassManager)i.next();
            ret.addAll(cm.getAllPossibleTypeNames());
    	}
    	return ret;
    }
}


class DirectoryClassManager extends ClassManager {
    File directory;
    public DirectoryClassManager(JavaCompiler compiler, File directory) {
        super(compiler);
        this.directory = directory;
    }

    public ClassManager makeSubPackageManager(String subPackageName) {
        File dir = new File(directory, subPackageName);
        if (!dir.isDirectory()) return null;
        return new DirectoryClassManager(getCompiler(), dir);
    }

    public TypeDec findTypeDec(String name) {
        File testFile = new File(directory, name+".class");
        // special work-around for the strange handling of 'nul.class'
        // on Windows
        if (name.equals("nul")) return null;
        if (testFile.isFile()) {
            try {
                //XXX add SourceLocation to this
                return makeTypeDec(name, new FileInputStream(testFile));
            } catch (FileNotFoundException e) {
                return null;
            }
        } else {
            return null;
        }
    }
    public Collection/*List<String>*/ getAllPossibleTypeNames() {
		Collection temp = new ArrayList();
		collectNames(directory, Collections.EMPTY_LIST, temp);
		//System.err.println("XXXX: classPathManager");
		//System.err.println(temp);
		return temp;
    }
    
    private void collectNames(File directory, List prefixName, Collection names) {
    	File[] files = directory.listFiles();
		if (files == null) return;
		for (int i = files.length - 1; i >= 0; i--) {
			File f = files[i];
			if (f.isDirectory()) {
				collectNames(f, newList(prefixName, f.getName()), names);
			} else if (isClassFile(f)) {
				names.add(newList(prefixName, filenameToClassName(f.getName())));
			}
		}
    }
    
    private List newList(List left, Object right) {
    	ArrayList temp = new ArrayList(left);
    	temp.add(right);
    	return temp;
    }
    
    private List newList(List left, List right) {
    	ArrayList temp = new ArrayList(left);
    	temp.addAll(right);
    	return temp;
    }
    
    private boolean isClassFile(File f) {
    	return f.isFile() && f.getName().toLowerCase().endsWith(".class");
    }
    
    public String toString()  { return "dir(" + directory + ")"; }
}
	


class JarClassManager extends ClassManager {
    ZipFile zipFile;
    String prefix = null;

    private JarClassManager(JavaCompiler compiler, ZipFile zipFile, String prefix) {
        super(compiler);
        this.zipFile = zipFile;
        this.prefix = prefix;
    }

    public JarClassManager(JavaCompiler compiler, File jarFile) {
        super(compiler);
        //System.out.println("checking: "+jarFile);
        if (jarFile.isFile()) {
            //System.out.println("is file");
            try {
                zipFile = new ZipFile(jarFile);
            } catch (IOException e) {
                zipFile = null;
            }
        } else {
            zipFile = null;
        }
    }

    public ClassManager makeSubPackageManager(String subPackageName) {
        if (zipFile == null) return null;

        String newPrefix;
        if (prefix == null) newPrefix = subPackageName + '/';
        else newPrefix = prefix + subPackageName + '/';
        
        //XXX for performance, we'd really like a test of "possibility" here

        JarClassManager ret = new JarClassManager(getCompiler(), zipFile, newPrefix);
        return ret;
    }

    public TypeDec findTypeDec(String name) {
        if (zipFile == null) return null;

        String fileName = prefix + name + ".class";

        ZipEntry entry = zipFile.getEntry(fileName);
        if (entry == null) return null;

        try {
            return makeTypeDec(name, zipFile.getInputStream(entry));
        } catch (IOException e) {
            return null;
        }
    }
    
    public Collection/*List<String>*/ getAllPossibleTypeNames() {
    	Collection ret = new ArrayList();
    	for (Enumeration i = zipFile.entries(); i.hasMoreElements(); ) {
    		ZipEntry entry = (ZipEntry)i.nextElement();
    		if (entry.isDirectory()) continue; 
    		String name = entry.getName();
    		if (name.toLowerCase().endsWith(".class")) {
    			ret.add(filenameToClassName(name));
    		}
    	}
    	return ret;
    }
  
  

    
    public String toString()  { return "jar(" + zipFile.getName() + ")"; }
}
