/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.cst;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import java.util.*;

/**
 * The scope of a compilation unit, typically a single source file
 */
public class CUScope extends Scope {
    private CompilationUnit cu;
    
    public CUScope(JavaCompiler compiler, Scope parent, CompilationUnit cu) {
        super(compiler, parent);
        this.cu = cu;
    }
    public String shortToString() { return "CU(" + cu.getSourceFileName() + ")"; }
    
    public Expr bindUnqualifiedName(String name, ASTObject fromWhere) {
        return null;
    }

    public Type findMethodLookupType(String name, ASTObject fromWhere) {
        return null;
    }
    
    HashMap nameToType = new HashMap();    
    public Type findType(String name, ASTObject fromWhere) {         
        Type type = (Type)nameToType.get(name);
        if (type != null) return type;

        type = internalFindType(name, fromWhere);
        if (type != null) {
            nameToType.put(name, type);
        }
        return type;
    }
    
    protected boolean isPackageAccessible(TypeDec typeDec) {
        String p1 = typeDec.getPackageName();
        String p2 = cu.getPackageName();
        if (p1 == null) return p2 == null;
        else return p1.equals(p2);
    }        
    
    protected boolean checkAccessibility(TypeDec typeDec, ASTObject fromWhere) {
        if (typeDec.isPublic()) return true;
        if (typeDec.isPrivate()) return false;
        if (typeDec.isProtected()) {
            if (isPackageAccessible(typeDec)) return true;
            return typeDec.isAccessible(fromWhere);
        }
        return isPackageAccessible(typeDec);
    }

        
    
    protected Type checkFoundType(String id, Type oldType, Type foundType, ASTObject fromWhere) {
        if (foundType == null) return oldType;
        if (foundType == getTypeManager().TYPE_NOT_FOUND) return oldType;
        
        if (!checkAccessibility(foundType.getTypeDec(), fromWhere)) return oldType;
        
        if (oldType != null && foundType != oldType) {
            return handleAmbiguousImport(id, foundType, oldType, fromWhere);
        } else {
            return foundType;
        }
    }
        

    protected Type internalFindType(String id, ASTObject fromWhere) {
        Type type = null;

        //System.err.println("looking in source for "+id);

        // first loop for single import names
        Imports imports = cu.getImports();
        if (imports != null) {
            for (int i=0; i<imports.size(); i++) {
                Import _import = imports.get(i);
                if (!_import.getStar()) {
                    if (_import.capturesId(id)) {
                        return _import.getType();
                    }
                }
            }
        }
        
        // check the current package for this name
        String packageName = cu.getPackageName();
        type = getTypeManager().findType(packageName, id);
        if (type != null) return type;


        //!!! This should raise an error if the name is found in more than
        //!!! one import-on-demand package

        // now loop for import-on-demand names
        //Type foundOnDemandType = null;
        if (imports != null) {
            for (int i=0; i<imports.size(); i++) {
                Import _import = imports.get(i);
                if (_import.getStar()) {
                    Object o = _import.getTypeName().resolve(_import.makeTypeNameScope(), false);
                    Type tryType = null;
                    if (o == null) continue;
                    
                    if (o instanceof String) {
                        tryType = getTypeManager().findType((String)o, id);
                    } else if (o instanceof Type) {
                        tryType = ((Type)o).getInnerType(id, _import, false);
                    }
                    type = checkFoundType(id, type, tryType, fromWhere);
                }
            }
        }
        
        //if (type != null) return type;

        // finally check in java.lang which is always import-on-demand
        Type javaLangType = getTypeManager().findType("java.lang", id);
        type = checkFoundType(id, type, javaLangType, fromWhere);

        // finally, look for the name itself
        //Type noPackageType = getTypeManager().findType(null, id);
        //type = checkFoundType(id, type, noPackageType, fromWhere);

        return type;
    }
    
    protected Type handleAmbiguousImport(String id, Type type1, Type type2, ASTObject showWhere) {
        getCompiler().showError(showWhere, "ambiguous type reference, could be either " + 
                        type1.getString() + " or " + type2.getString());
        return null;
    }
}
