/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.cst;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import java.util.*;

/**
  * @grammar parent.id
  * @child Name parentName
  * @property String id
  * @property boolean allowExpr
  */
public class Name extends ASTObject {
    boolean allowType = true;

    public Name(SourceLocation location, Name parentName, String id) {
        this(location, parentName, id, true);
    }

    public void forceType() {
        allowExpr = false;
        if (parentName != null) parentName.forceType();
    }    
    
    /********************************************************************************
     * New Scope-based design
     */

    private Object resolveSimple(Scope scope, boolean showError) {
        if (allowExpr) {
            Expr expr = scope.bindUnqualifiedName(id, this);    
    
            if (expr != null) {
                expr.setSourceLocation(getSourceLocation());
                return expr;
            } else if (showError) {
                this.showError("can't bind name: " + id);
            }
        }
        if (allowType) {
            Type type = scope.findType(id, this);
            if (type != null) return type;

            if (type == null && showError && !allowExpr) {
                this.showError("type name not found: " + id);
            }
        }

        return id;
    }

    private Object resolve1(Expr parentExpr, boolean showError) {
        Field field = parentExpr.getType().getField(id, this, showError);
        ASTObject ret = getAST().makeGet(parentExpr, field);
        ret.setSourceLocation(getSourceLocation());
        return ret;
    }

    private Object resolve1(String parentPackage, boolean showError) {
        Type type = getTypeManager().findType(parentPackage, id);
        if (type == null) {
            return parentPackage + '.' + id;
        } else {
            return type;
        }
    }

    private Object resolve1(Type parentType, boolean showError) {
        if (allowExpr) {
            Field field = parentType.getField(id, this, false);
            if (field != null) {
                TypeExpr typeExpr = getAST().makeTypeExpr(parentType);
                typeExpr.setSourceLocation(parentName.getSourceLocation());
                ASTObject ret = getAST().makeGet(typeExpr, field); 
                ret.setSourceLocation(getSourceLocation());
                return ret;
            }
        }
        if (allowType) {
            Type type = parentType.getInnerType(id, this, false);
            if (type != getTypeManager().TYPE_NOT_FOUND) {
                return type;
            }
        }
        
        if (!showError) return null;

        // some code to generate better error messages
        if (allowExpr && allowType) {
            return getAST().makeGet(null, parentType.getField(id, this, true));
        } else if (allowType && !allowExpr) {
            if (parentType.getField(id, this, false) != null) {
                this.showError("type required, but field found: " + id);
                return getTypeManager().TYPE_NOT_FOUND;
            } else {
                return parentType.getInnerType(id, this, true);
            }
        } else {
            this.showError("name not found: " + id);
            return null;
        }
    }


    public Type resolveType(Scope scope) {
    	return resolveType(scope, true);
    }
    
    public Type resolveType(Scope scope, boolean showError) {
        forceType();
        Object ret = internalResolve(scope, showError);
        if (ret instanceof Type) {
            return (Type)ret;
        }
        if (showError) showNiceError(scope);
        return getTypeManager().TYPE_NOT_FOUND;
    }

    public Expr resolveExpr(Scope scope) {
        Object ret = internalResolve(scope, true);
        if (ret instanceof Expr) {
            return (Expr)ret;
        } else if (ret instanceof Type) {
            //??? is this really legal ???
            Expr expr = getAST().makeTypeExpr((Type)ret);
            expr.setSourceLocation(getSourceLocation());
            return expr;
        }
        
        showNiceError(scope);
        return getAST().makeNotFoundExpr();
    }
    
    private void showNiceError(Scope scope) {
        //System.out.println("error in " + scope);
        if (parentName == null) {
            resolveSimple(scope, true);
        } else {
            showError("can't bind name: " + toString());
        }
    }


	/**
	 * returns one of
	 * <ul>
	 *   <li> Type if this is a type name
	 *   <li> Expr (VarExpr or FieldAccessExpr) if this is a variable or field name
	 *   <li> String if this is a package name
	 *   <li> null if nothing matches
	 * </ul>
	 */
    public Object resolve(Scope scope, boolean showError) {
        return internalResolve(scope, showError);
    }
    
    private final Object internalResolve(Scope scope, boolean showError) {
        if (parentName == null) {
        	//!!! why can't this be showError???
            return resolveSimple(scope, false);
        } else {
            Object parentO = parentName.internalResolve(scope, showError);
            if (parentO == null) {
                return null;
            } else if (parentO instanceof Type) {
                return resolve1((Type)parentO, showError);
            } else if (parentO instanceof String) {
                return resolve1((String)parentO, showError);
            } else {
                return resolve1((Expr)parentO, showError);
            }
        }
    }
    
    
    
    public String toString() {
        if (parentName == null) {
            return id;
        } else {
            return parentName.toString() + '.' + id;
        }
    }    
    
    //BEGIN: Generated from @child and @property
    protected Name parentName;
    public Name getParentName() { return parentName; }
    public void setParentName(Name _parentName) {
        if (_parentName != null) _parentName.setParent(this);
        parentName = _parentName;
    }
    
    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }
    
    protected boolean allowExpr;
    public boolean getAllowExpr() { return allowExpr; }
    public void setAllowExpr(boolean _allowExpr) { allowExpr = _allowExpr; }
    
    public Name(SourceLocation location, Name _parentName, String _id, boolean _allowExpr) {
        super(location);
        setParentName(_parentName);
        setId(_id);
        setAllowExpr(_allowExpr);
    }
    protected Name(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        Name ret = new Name(getSourceLocation());
        ret.preCopy(walker, this);
        if (parentName != null) ret.setParentName( (Name)walker.process(parentName) );
        ret.id = id;
        ret.allowExpr = allowExpr;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return parentName;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "parentName";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setParentName((Name)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "Name(id: "+id+", "+"allowExpr: "+allowExpr+")";
    }
    
    //END: Generated from @child and @property
}
