/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.parser;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.base.*;


import java.util.*;
import java.io.*;
import java.util.Stack;

public class JavaParser extends CompilerObject implements JavaConstants {
    public boolean parseInterfaceOnly = false;

    public JavaParser(JavaCompiler compiler) {
        super(compiler);
        if (compiler.getOptions().source.equals("1.4")) {
            stmtParsers.put("assert", new AssertStmtParser());
            keywords.add("assert");
        }
    }

    public CompilationUnit interfaceParse(CompilationUnit cu) throws IOException {
        tokenSource.setSourceInfo(cu.getSourceInfo());
        currentCompilationUnit = cu;
        currentIndex = 0;
        lastValidIndex = -1;

        parseCompilationUnit();
        currentCompilationUnit.setScanned(true);
        return currentCompilationUnit;
    }

    /* Global variable to identify that we are parsing interface */
    protected boolean inInterface;
    /* Interface to the Token Manager */
    private JavaTokenizer tokenSource = new JavaTokenizer(getCompiler());
    // this size must be a power of two
    private static final int LOOKAHEAD_BUFFER_SIZE = 256;
    private static final int LOOKAHEAD_MASK = LOOKAHEAD_BUFFER_SIZE-1;
    private Token[] lookaheadBuffer = new Token[LOOKAHEAD_BUFFER_SIZE];
    public int currentIndex = 0;
    private int lastValidIndex = -1;

    private final void initTokenSource(int offset) {
        currentIndex = 0;
        lastValidIndex = -1;
        tokenSource.setOffset(offset);
    }

    public void cleanup(boolean fully) {
        tokenSource.cleanup(fully);
        for (int i=0; i < LOOKAHEAD_BUFFER_SIZE; i++) {
            lookaheadBuffer[i] = null;
        }
        lastValidIndex = -1;
    }

    private final Token getToken(int index) {
        while (index > lastValidIndex) {
            lastValidIndex++;
            Token newToken = tokenSource.getNextToken();
            //System.err.println("t: "+newToken.startPosition+", "+newToken+", "+newToken.image+", "+newToken.isIdentifier());
            lookaheadBuffer[lastValidIndex&LOOKAHEAD_MASK] = newToken;
        }
        return lookaheadBuffer[index&LOOKAHEAD_MASK];
    }

    private int savedIndex = -1;
    //!!! this scheme needs some error checking
    protected final void pushCurrentIndex() {
        savedIndex = currentIndex;
    }

    protected final void popCurrentIndex() {
        currentIndex = savedIndex;
    }

    public final void eatTopToken(int kind) {
        if (peekToken().kind != kind) {
            throwError(peekToken(), "expected "+JavaConstants.tokenImage[kind]+", found "+peekToken().image);
        }
        currentIndex++;
    }

    public final boolean maybeEatToken(int kind) {
        if (peekToken().kind == kind) {
            currentIndex++;
            return true;
        } else {
            return false;
        }
    }

    public final void eatTopToken() {
        currentIndex++;
    }

    public final Token popToken(int kind) {
        return getToken(currentIndex++);
    }

    public final Token popToken() {
        return getToken(currentIndex++);
    }

    public final Token peekToken() {
        return getToken(currentIndex);
    }

    public final Token peekToken(int i) {
        return getToken(currentIndex+i);
    }

    public int scanTillMatched()  {
        Token beginToken = peekToken();
        int kind = RBRACE;
        if (beginToken.kind != LBRACE)
            throwError(beginToken,"internal error: expected '{'");

        // TODO: scan all lookahead tokens
        // I don't check lookahead tokens right now, because
        // the is no meaning to lookahead after '{' token.
        if (currentIndex != lastValidIndex)
            throwError(beginToken,"internal error: lookahead of '{'");

        // eat the left brace
        eatTopToken();

        // ask tokenizer scan all data until matched 'kind' token
        try {
            Token endToken = tokenSource.scanTillMatched(kind);
            lastValidIndex++;
            lookaheadBuffer[lastValidIndex&LOOKAHEAD_MASK] = endToken;
            // eat the right brace
            eatTopToken();
            return endToken.endPosition;
        } catch (Exception e) {
            showError(beginToken,"Unexpected end of file while looking matched close brace");
        }
        return beginToken.endPosition;
    }

    private Stack enclosingTypeDecs = new Stack();

    public void pushEnclosingTypeDec(TypeDec dec) {
        enclosingTypeDecs.push(dec);
        //currentCompilationUnit.addDefinedType(dec);
    }

    public void popEnclosingTypeDec() {
        enclosingTypeDecs.pop();
    }

    public TypeDec peekEnclosingTypeDec() {
        if (enclosingTypeDecs.size() == 0) return null;
        return (TypeDec)enclosingTypeDecs.peek();
    }

    protected void parseIdMatchingEnclosingTypeName(String where) {
        Token beginToken = peekToken();
        String id;
        if (beginToken.kind == NEW) {
            id = "new";
            popToken();
        } else {
            id = parseId();
        }
        if ( id.equals(peekEnclosingTypeDec().getId())) return;

        throwError(beginToken, where + " \"" + id +
                                 "\" doesn't match enclosing type name \"" +
                                 peekEnclosingTypeDec().getId()+"\"");
    }




    /* Dynamic parser stuff */

    // table of all keywords from JLSv2 3.9
    protected Set keywords = new HashSet(Arrays.asList(new String[]
        { "abstract", "boolean", "break", "byte", "case", "catch",
          "char", "class", "const", "continue",
          "default", "do", "double", "else", "extends",
          "final", "finally", "float", "for", "goto",
          "if", "implements", "import", "intstanceof", "int",
          "interface", "long", "native", "new", "package",
          "private", "protected", "public", "return", "short",
          "static", "strictfp", "super", "switch", "synchronized",
          "this", "throw", "throws", "transient", "try",
          "void", "volatile", "while",
          "true", "false", "null", "instanceof"// technically not keywords, but not legal ids anyway
        }));

    private Map primitiveTypeNames = new HashMap();
    {
        primitiveTypeNames.put("int", getTypeManager().intType);
        primitiveTypeNames.put("byte", getTypeManager().byteType);
        primitiveTypeNames.put("short", getTypeManager().shortType);
        primitiveTypeNames.put("long", getTypeManager().longType);
        primitiveTypeNames.put("double", getTypeManager().doubleType);
        primitiveTypeNames.put("float", getTypeManager().floatType);
        primitiveTypeNames.put("char", getTypeManager().charType);
        primitiveTypeNames.put("boolean", getTypeManager().booleanType);
        primitiveTypeNames.put("void", getTypeManager().voidType);
    }
    public TypeD lookupPrimitiveTypeD(String name) {
        Type type = (Type)primitiveTypeNames.get(name);
        if (type == null) return null;
        return type.makeTypeD();
    }

    protected static Map modifiers = new HashMap();
    static {
        modifiers.put("public", new Integer(Modifiers.PUBLIC));
        modifiers.put("protected", new Integer(Modifiers.PROTECTED));
        modifiers.put("private", new Integer(Modifiers.PRIVATE));
        modifiers.put("static", new Integer(Modifiers.STATIC));
        modifiers.put("final", new Integer(Modifiers.FINAL));
        modifiers.put("abstract", new Integer(Modifiers.ABSTRACT));
        modifiers.put("native", new Integer(Modifiers.NATIVE));
        modifiers.put("synchronized", new Integer(Modifiers.SYNCHRONIZED));
        modifiers.put("transient", new Integer(Modifiers.TRANSIENT));
        modifiers.put("volatile", new Integer(Modifiers.VOLATILE));
        modifiers.put("strictfp", new Integer(Modifiers.STRICT));
    }
    public final boolean isModifier(Token token) {
        return getModifier(token) != null;
    }

    public final Integer getModifier(Token token) {
        if (!isIdentifier(token)) return null;
        return (Integer)modifiers.get(token.image);
    }

    protected Map decParsers = new HashMap();
    {
        decParsers.put("class", new ClassDecParser());
        decParsers.put("interface", new InterfaceDecParser());
    }

    protected Map stmtDecParsers = new HashMap();
    {
        stmtDecParsers.put("class", new ClassDecParser());
        stmtDecParsers.put("interface", new InterfaceDecParser());
    }


    protected Map stmtParsers = new HashMap();
    {


        stmtParsers.put("if", new IfStmtParser());
        stmtParsers.put("while", new WhileStmtParser());
        stmtParsers.put("do", new DoStmtParser());
        stmtParsers.put("for", new ForStmtParser());

        stmtParsers.put("synchronized", new SynchronizedStmtParser());
        stmtParsers.put("try", new TryStmtParser());
        stmtParsers.put("switch", new SwitchStmtParser());

        stmtParsers.put("break", new BreakStmtParser());
        stmtParsers.put("continue", new ContinueStmtParser());
        stmtParsers.put("return", new ReturnStmtParser());
        stmtParsers.put("throw", new ThrowStmtParser());

        stmtParsers.put("goto", new GotoStmtParser());
    }

    protected SourceLocation dummySource = new DummySourceLocation(getCompiler());

    protected class NewOperator extends PrefixOperator {
        public NewOperator(int precedence) {
            super("new", precedence);
        }

        private NewArrayExpr parseNewArrayExprWithInitializer(ArrayTypeD typeD) {
            ArrayInitializer init = null;
            if (peekToken().kind == LBRACE) {
                init = parseArrayInitializer();
            } else {
                showError(peekToken(), "'{' expected");
            }
            //System.out.println("parsed new array: " + peekToken());
            return new NewArrayExpr(dummySource, typeD, init, null);
        }

        private NewArrayExpr parseNewArrayExpr(TypeD typeD) {
            Exprs dimExprs = null;
            if (peekToken().kind == LBRACKET) {
                dimExprs = parseArrayDimExprs();
                for(int i=0; i<dimExprs.size(); i++) {
                    typeD = new ArrayTypeD(dummySource,typeD);
                }
            }
            return new NewArrayExpr(dummySource,(ArrayTypeD)typeD, null, dimExprs);
        }

        public Expr parse() {
            Token beginToken = peekToken();

            eatTopToken();  //eats the 'new'
            TypeD typeD = parseTypeD();
            if (typeD == null) {
                throwError(peekToken(), "type expected");
            }

            if (maybeEatToken(LPAREN)) {
                Exprs args = parseExprs(RPAREN);
                eatTopToken(RPAREN);

                TypeDec typeDec = maybeParseClassBody(typeD);
                return new NewInstanceExpr(dummySource, null, (TypeD)typeD, args, typeDec, null);
            } else if (typeD instanceof ArrayTypeD) {
                // already ate a bunch of [][][]...s
                return parseNewArrayExprWithInitializer((ArrayTypeD)typeD);
            } else if (peekToken().kind == LBRACKET) {
                // have yet to eat any []s.
                return parseNewArrayExpr(typeD);
            } else {
                throwError(peekToken(), "( or [ expected");
                return null;
            }
        }
    }

    TypeDec maybeParseClassBody(TypeD superTypeD) {
        if (peekToken().kind == LBRACE) {
            Token beginToken = peekToken();
            Modifiers mods = new Modifiers(makeSourceLocation(beginToken), 0);
            if (superTypeD != null) superTypeD = (TypeD)superTypeD.copy();
            ClassDec classDec =
                new ClassDec(dummySource, mods,
                             "ANONYMOUS", superTypeD, 
                             null, null);
            classDec.setIsAnonymous();
            classDec.setLocal();
            Decs body = null;
            boolean oldInInterface = inInterface;
            try {
                inInterface = false;
                pushEnclosingTypeDec(classDec);
                body = parseDecs();
            } finally {
                inInterface = oldInInterface;
                popEnclosingTypeDec(); //classDec);
            }
            classDec.setBody(body);
            
            classDec.setEnclosingTypeDec(peekEnclosingTypeDec());
            addContext(classDec, beginToken);
            return classDec;
        } else {
            return null;
        }
    }
        

    protected class ParenOperator extends PrefixOperator {
        protected boolean allowCasts = false;

        public ParenOperator(int precedence, boolean allowCasts) {
            super("(", precedence);
            this.allowCasts = allowCasts;
        }
        public ParenOperator(int precedence) {
            this(precedence, false);
        }

        private boolean notPlusMinus(int kind) {
            switch (kind) {
                case PLUS:
                case MINUS:
                case INCR:
                case DECR:
                //XXX I don't think these are really needed, but they can't hurt
                case HOOK:
                case SEMICOLON:
                    return false;
                default:
                    return true;
            }
        }

        // Handle's both cast op and simple parenthesized exp
        public Expr parse() {
            eatTopToken(); // eats the '('

            if (!allowCasts) return basicParse();

            // follow the language spec closely, even though I disagree with their rules
            // if next identifier is a numerictype, then it's definately a cast
            // otherwise, peek through a typed and a close paren, and if the next
            // token is not a +, ++, -, or --, then it's a cast
            int savedIndex = currentIndex; //pushCurrentIndex();
            Token nextToken = peekToken();

            // cast precedence is between unary ops and multiplicative binops

            if ((nextToken.isIdentifier() && primitiveTypeNames.get(nextToken.image) != null)) {
                TypeD typeD = parseTypeD();
                eatTopToken(RPAREN);
                return new CastExpr(dummySource,typeD, exprParser.parse(1500));
            } else if (eatTypeD() && popToken().kind == RPAREN &&
                       notPlusMinus(popToken().kind)) {
                currentIndex = savedIndex;
                TypeD typeD = parseTypeD();
                eatTopToken(RPAREN);
                Expr expr = exprParser.parse(1500); // parseExpr(true);
                if (expr != null) return new CastExpr(dummySource,typeD, expr);
            }

            currentIndex = savedIndex;
            return basicParse();
        }

        protected Expr basicParse() {
            Expr expr = parentParse();
            eatTopToken(RPAREN);
            return new ParenExpr(dummySource,expr);
        }

    }


    abstract protected class Operator {
        protected int precedence;
        protected String image;
        protected int tokenKind;

        private ExprParser parent;

        protected Operator(int tokenKind, int precedence) {
            this.precedence = precedence;
            this.tokenKind = tokenKind;
            this.image = tokenImage[tokenKind];
        }

        protected Operator(String image, int precedence) {
            int tokenKind = 0;
            String testImage = "\"" + image + "\"";

            for(;tokenKind < tokenImage.length; tokenKind++) {
                String maybeImage = tokenImage[tokenKind];
                if (maybeImage != null && maybeImage.equals(testImage)) {
                    break;
                }
            }

            if (tokenKind == tokenImage.length) {
                tokenKind = tokenSource.addOperator(image);
            }

            this.tokenKind = tokenKind;
            this.image = image;
            this.precedence = precedence;
        }

        public int getTokenKind() {
            return tokenKind;
        }

        public void setParent(ExprParser parent) {
            this.parent = parent;
        }

        public Expr parentParse() {
            Expr ret = parent.parse();
            if (ret == null) throwError(null, "expression expected");
            return ret;
        }

        public Expr parentParse(int precedence) {
            Expr ret = parent.parse(precedence);
            if (ret == null) throwError(null, "expression expected");
            return ret;
        }
     }

    protected class MinusOperator extends PrefixOperator {
        public MinusOperator(int precedence) {
            super("-", precedence);
        }

        public Expr parse() {
            eatTopToken();
            //!!! should be this, not image
            if (peekToken().kind == INTEGER_LITERAL) {
                Token t = popToken();
                return t.getNegativeExpr(makeSourceLocation(peekToken(-1)));
            } else {
                return new MinusOpExpr(makeSourceLocation(peekToken(-1)), 
                                       image, parentParse(precedence));
            }                          
        }
    }

    protected class PrefixOperator extends Operator {
        protected PrefixOperator(int tokenKind, int precedence) {
            super(tokenKind, precedence);
        }

        public PrefixOperator(String image, int precedence) {
            super(image, precedence);
        }

        public Expr parse() {
            eatTopToken();
            //!!! should be this, not image
            return UnopExpr.build(dummySource, image, parentParse(precedence));
        }
    }


    protected abstract class LiteralOp extends PrefixOperator {
        public LiteralOp(int tokenKind) {
            super(tokenKind, -1);
        }

        public LiteralOp(String image) {
            super(image, -1);
        }

        public Expr parse() {
            Token token = peekToken();
            Expr expr = parse(token.image);
            return expr;
        }

        public abstract Expr parse(String image);
    }


    protected class TrueLiteral extends LiteralOp {
        public TrueLiteral() { super("true"); }

        public Expr parse(String image) {
            eatTopToken();
            return new BooleanLiteralExpr(dummySource,true);
        }
    }

    protected class FalseLiteral extends LiteralOp {
        public FalseLiteral() { super("false"); }

        public Expr parse(String image) {
            eatTopToken();
            return new BooleanLiteralExpr(dummySource,false);
        }
    }

    protected class NullLiteral extends LiteralOp {
        public NullLiteral() { super("null"); }

        public Expr parse(String image) {
            eatTopToken();
            return new NullExpr(dummySource);
        }
    }

    protected class ThisLiteral extends LiteralOp {
        public ThisLiteral() { super("this"); }

        public Expr parse(String image) {
            eatTopToken();
            return new ThisExpr(dummySource);
        }
    }

    protected class SuperLiteral extends LiteralOp {
        public SuperLiteral() { super("super"); }

        public Expr parse(String image) {
            eatTopToken();
            return new SuperExpr(dummySource);
        }
    }

    public Expr parseCallExpr(Name name) {
        //System.out.println("call name: " + name);

        String id = name.getId();
        Name parentName = name.getParentName();
        Exprs args = parseExprs(RPAREN);
        eatTopToken(RPAREN);
        Expr expr = null;
        if (parentName != null) {
            parentName.clearParent();
            expr = new UnresolvedExpr(dummySource,parentName);
            expr.setSource(parentName);
        }

        if (id.equals("this")) {
            if (parentName != null) throwError(null, "bad this");
            return new ConstructorCallExpr(dummySource, null, false, args, null);
        } else if (id.equals("super")) {
            if (parentName != null) throwError(null, "bad super");
            return new ConstructorCallExpr(dummySource, null, true, args, null);
        }

        return new CallExpr(dummySource,expr, id, args);
    }


    protected class IdentifierOp extends LiteralOp {
        public IdentifierOp() { super(IDENTIFIER); }

        public Expr parse(String image) {
            TypeD typeD = lookupPrimitiveTypeD(image);
            if (typeD != null) {
                eatTopToken();
                return new TypeExpr(dummySource, typeD);
            }

            Name name = parseExtendedName();

            if (maybeEatToken(LPAREN)) {
                return parseCallExpr(name);
            }
            return new UnresolvedExpr(dummySource,name);
        }
    }

    protected class InfixOperator extends Operator {
        protected InfixOperator(int tokenKind, int precedence) {
            super(tokenKind, precedence);
        }

        public InfixOperator(String image, int precedence) {
            super(image, precedence);
        }

        public Expr parse(Expr lhs, int minimumPrecedence) {
            if (precedence <= minimumPrecedence) return lhs;
            eatTopToken();

            return parse(lhs);
        }

        public Expr parse(Expr lhs) {
            //!!! should be this, not image
            return BinopExpr.build(dummySource,this.image, lhs, parentParse(precedence));
        }
    }

    private Name buildName(Expr expr) {
        if (expr instanceof UnresolvedExpr) {
            return ((UnresolvedExpr)expr).getName();
        } else {
            return null;
        }
    }

    private TypeD exprToTypeD(Expr lhs) {
        if (lhs instanceof TypeExpr) {
            return ((TypeExpr)lhs).getTypeD();
        }

        Name id = buildName(lhs);
        TypeD ret = null;
        if (id == null) return null;
        if (id.getParentName() == null) {
            TypeD primTypeD = lookupPrimitiveTypeD(id.getId());
            if (primTypeD != null) ret = primTypeD;
        }
        id.clearParent();
        if (ret == null) ret = new UnresolvedNameTypeD(dummySource,id);
        ret.setSource(lhs);
        return ret;
    }

    private TypeD exprToTypeD(Expr lhs, String name) {
        TypeD ret = exprToTypeD(lhs);
        if (ret == null) {
            throwError(peekToken(-2), "type required before "+name);
        }

        addContext(ret, lhs);
        return ret;
    }

    protected class DotOperator extends InfixOperator {
        public DotOperator(int precedence) {
            super(".", precedence);
        }

        public Expr parse(Expr lhs) {
            // here we handle the ever popular inner constructor syntax
            if (maybeEatKeyword("new")) {
                String innerName = parseId();
                eatTopToken(LPAREN);
                Exprs args = parseExprs(RPAREN);
                eatTopToken(RPAREN);
                TypeDec typeDec = maybeParseClassBody(null);
                return new NewInnerInstanceExpr(dummySource,
                                                lhs, null, args, typeDec, 
                                                null, innerName);
            }

            String id = parseIdOrKeywords_this_super_class();
            if (id.equals("this")) {
                return new QualifiedThisExpr(dummySource,exprToTypeD(lhs, ".this"));
            } else if (id.equals("super")) {
                if (inConstructor() && maybeEatToken(LPAREN)) {
                    Exprs args = parseExprs(RPAREN);
                    eatTopToken(RPAREN);
                    return new ConstructorCallExpr(dummySource, lhs, true, args, null);
                } else {
                    return new QualifiedSuperExpr(dummySource,exprToTypeD(lhs, ".super"));
                }
            } else if (id.equals("class")) {
                return new ClassExpr(dummySource,exprToTypeD(lhs, ".class"));
            } else if (maybeEatToken(LPAREN)) {
                Exprs args = parseExprs(RPAREN);
                eatTopToken(RPAREN);
                return new CallExpr(dummySource,lhs, id, args);
            } else {
                return new UnresolvedFieldAccessExpr(dummySource,lhs, id);
            }
        }
    }

    protected class AssignOperator extends InfixOperator {
        public AssignOperator(String image, int precedence) {
            super(image, precedence);
        }

        public Expr parse(Expr lhs) {
            Expr rhs = parentParse(precedence-1);
            String s = this.image;
            return AssignExpr.build(dummySource,checkAssignable(lhs), 
                                    s.substring(0, s.length() - 1),
                                    rhs);
        }
    }

    protected class ArrayOperator extends InfixOperator {
        public ArrayOperator(int precedence) {
            super("[", precedence);
        }

        public Expr parse(Expr lhs) {
            // add in a special case for an array type followed by .class
            if (peekToken().kind == RBRACKET) {
                //System.out.println("lhs: "+lhs);
                TypeD typeD = exprToTypeD(lhs, ".class");
                //!!! violates an abstraction, should be something like unpopToken()
                currentIndex--;
                typeD = addBracketsToTypeD(typeD);
                eatTopToken(DOT);
                eatKeyword("class");
                return new ClassExpr(dummySource,typeD);
            }

            Expr index = parseExpr();
            eatTopToken(RBRACKET);
            return new ArrayExpr(dummySource,lhs, index);
        }
    }

    protected class QuestionOperator extends InfixOperator {
        public QuestionOperator(int precedence) {
            super("?", precedence);
        }

        public Expr parse(Expr lhs) {
            // parse anything up to a :
            Expr trueExpr = parentParse(0);
            eatTopToken(COLON);
            Expr falseExpr = parseExpr();
            return new TriTestExpr(dummySource,lhs, trueExpr, falseExpr);
        }
    }

    protected class InstanceOfOperator extends InfixOperator {
        public InstanceOfOperator(int precedence) {
            super("instanceof", precedence);
        }

        public Expr parse(Expr lhs) {
            TypeD typeD = parseTypeD();
            return new InstanceofExpr(dummySource,lhs, typeD);
        }
    }

    AssignableExpr checkAssignable(Expr expr) {
        if (expr instanceof AssignableExpr) return (AssignableExpr)expr;
        expr.showError("invalid lhs for assign: "+expr.getClass().getName());
        return null;
    }

    protected class PreOperator extends PrefixOperator {
        public PreOperator(String image, int precedence) {
            super(image, precedence);
        }

        public Expr parse() {
            eatTopToken(); // eats the '--' or the '++'
            return new PrefixExpr(dummySource,checkAssignable(parentParse(precedence)), image);
        }
    }

    protected class PostOperator extends InfixOperator {
        public PostOperator(String image, int precedence) {
            super(image, precedence);
        }

        public Expr parse(Expr lhs) {
            return new PostfixExpr(dummySource,checkAssignable(lhs), image);
        }
    }

    private static final int MAX_TOKEN_KINDS = 512;

    protected class ExprParser {
        // these should be hidden as implementation details as much as possible
        protected PrefixOperator[] prefixOps;
        protected InfixOperator[] infixOps;

        private boolean allowLiterals = true;

        public ExprParser() {
            int tokenTypes = tokenImage.length;

            prefixOps = new PrefixOperator[MAX_TOKEN_KINDS];
            infixOps = new InfixOperator[MAX_TOKEN_KINDS];
            //terminals = new PrefixOperator[tokenTypes];
        }

        private final void fillOps(Operator[] ops, Operator[] toArray) {
            for(int i=0; i<ops.length; i++) {
                Operator op = ops[i];
                toArray[op.getTokenKind()] = op;
                op.setParent(this);
            }
        }

        public ExprParser(PrefixOperator[] prefixes, InfixOperator[] infixes,
                    boolean allowLiterals)
        {
            this(prefixes, infixes);
            this.allowLiterals = allowLiterals;
        }

        public ExprParser(PrefixOperator[] prefixes, InfixOperator[] infixes) {
            this();

            fillOps(prefixes, prefixOps);
            fillOps(infixes, infixOps);
        }

        public void addInfixOperator(InfixOperator op) {
            infixOps[op.getTokenKind()] = op;
            op.setParent(this);
        }

        public void addPrefixOperator(PrefixOperator op) {
            prefixOps[op.getTokenKind()] = op;
            op.setParent(this);
        }

        //public void addTerminal(PrefixOperator op) {
        //    terminals[op.getTokenKind()] = op;
        //    op.setParent(this);
        //}


        public Expr parseTerminal() {
            Token token = peekToken();
            int kind = token.kind;

            Expr literalExpr = token.getExpr(makeSourceLocation(token));
            if (literalExpr != null && allowLiterals) {
                eatTopToken();
                return literalExpr;
            }

            PrefixOperator op = prefixOps[kind];
            if (op == null && token.isIdentifier()) {
                op = prefixOps[IDENTIFIER];
            }

            if (op != null) {
                return op.parse();
            }

            return null;
        }

        public Expr parseOperator(Expr lhs, int minimumPrecedence) {
            Token token = peekToken();
            int kind = token.kind;

            InfixOperator op = infixOps[kind];
            if (op != null) {
                return op.parse(lhs, minimumPrecedence);
            }
            return lhs;
        }

        public Expr parseAll(Expr lhs, int minimumPrecedence, Token beginToken) {
            Expr ret;
            // keep parsing until a fixed-point is reached
            //Token beginToken = peekToken();
            while ((ret = parseOperator(lhs, minimumPrecedence)) != lhs && ret != null) {
                addContext(ret, beginToken);
                beginToken = peekToken();
                lhs = ret;
            }
            return ret;
        }

        protected Expr parse(int minimumPrecedence) {
            Token beginToken = peekToken();
            Expr lhs = parseTerminal();
            if (lhs == null) return null;
            addContext(lhs, beginToken);
            return parseAll(lhs, minimumPrecedence, beginToken);
        }

        public Expr parse() {
            return parse(0);
        }
    }

    private InfixOperator[] infixOperators = {
        new QuestionOperator(200),

        new InfixOperator("||", 300),
        new InfixOperator("&&", 400),

        new InfixOperator("|", 500),
        new InfixOperator("^", 600),
        new InfixOperator("&", 700),

        new InfixOperator("==", 800),
        new InfixOperator("!=", 800),

        new InstanceOfOperator(900),

        new InfixOperator(">", 1000),
        new InfixOperator("<", 1000),
        new InfixOperator("<=", 1000),
        new InfixOperator(">=", 1000),

        new InfixOperator("<<", 1100),
        new InfixOperator(">>", 1100),
        new InfixOperator(">>>", 1100),

        new InfixOperator("+", 1200),
        new InfixOperator("-", 1200),

        new InfixOperator("*", 1300),
        new InfixOperator("/", 1300),
        new InfixOperator("%", 1300),

        new PostOperator("++", 4000),  //??? postfix is a degenerate case of infix
        new PostOperator("--", 4000),

        //new CallOperator(5000),
        new ArrayOperator(5000),
        new DotOperator(5000),

        // the claim is that assignments are just infix operators
        new AssignOperator("=", 100),
        new AssignOperator("+=", 100),
        new AssignOperator("-=", 100),
        new AssignOperator("*=", 100),
        new AssignOperator("/=", 100),
        new AssignOperator("&=", 100),
        new AssignOperator("|=", 100),
        new AssignOperator("^=", 100),
        new AssignOperator("%=", 100),
        new AssignOperator("<<=", 100),
        new AssignOperator(">>=", 100),
        new AssignOperator(">>>=", 100),
    };

    private PrefixOperator[] prefixOperators = {
        new PrefixOperator("+", 2000),
        new MinusOperator(2000),
        new PrefixOperator("!", 3000),
        new PrefixOperator("~", 3000),

        new PreOperator("++", 2500),
        new PreOperator("--", 2500),

        new ParenOperator(1, true),
        new NewOperator(1),

        new TrueLiteral(),
        new FalseLiteral(),
        new NullLiteral(),
        new ThisLiteral(),
        new SuperLiteral(),

        new IdentifierOp()
    };


    protected ExprParser exprParser = new ExprParser(prefixOperators, infixOperators);

    /* Useful predicates */
    public final boolean eatKeyword(String keyword, boolean isMaybe) {
        Token topToken = peekToken();
        if (topToken.isIdentifier() && topToken.image.equals(keyword)) {
            eatTopToken();
            return true;
        } else {
            if (!isMaybe) {
                throwError(topToken, "expected keyword: '"+keyword+"'");
            }
            return false;
        }
    }

    public final boolean peekKeyword(String keyword) {
        Token topToken = peekToken();
        if (topToken.isIdentifier() && topToken.image.equals(keyword)) {
            return true;
        } else {
            return false;
        }
    }

    public final boolean eatKeyword(String keyword) {
        return eatKeyword(keyword, false);
    }

    public final boolean maybeEatKeyword(String keyword) {
        return eatKeyword(keyword, true);
    }

    public final boolean isIdentifier(Token token) {
        //??? make sure it's not a keyword?
        return token.isIdentifier() && !token.image.equals("new");
    }

    public final boolean isIdentifierOrNew(Token token) {
        return token.isIdentifier();
    }

    public final boolean isKeyword(Token token) {
        return keywords.contains(token.image);
    }

    /* Parse different AST nodes */
    protected CompilationUnit currentCompilationUnit;

    private CompilationUnit parseCompilationUnit() {
        Token beginToken0 = peekToken();

        Decs decs = new Decs(dummySource);
        currentCompilationUnit.setDecs(decs);

        // first check for a package
        if (maybeEatKeyword("package")) {
            currentCompilationUnit.setPackageName(parseExtendedString(false));
            eatTopToken(SEMICOLON);
        }

        Token beginImportsToken = peekToken();
        Imports imports = new Imports(dummySource);
        while (peekKeyword("import")) {
            imports.add(parseImport());
        }
        addContext(imports, beginImportsToken);

        while (peekToken().kind != EOF) {
            if (peekToken().kind == SEMICOLON) {
                eatTopToken(); continue;
            }
            Token beginToken = peekToken();
            Dec dec = (Dec)parseTopDec(null);
            addContext(dec, beginToken);

            if (dec instanceof TypeDec) {
                decs.add(dec);
            } else {
                throwError(null, "expected import, package, type or aspect declaration");
            }
        }

        addContext(decs, beginToken0);

        currentCompilationUnit.setImports(imports);
        addContext(currentCompilationUnit, beginToken0);
        return currentCompilationUnit;
    }

    public Import parseImport() {
        Token beginToken = peekToken();
        eatKeyword("import");
        Name name = parseExtendedName();

        boolean star = false;
        if (peekToken(0).kind == DOT && peekToken(1).kind == STAR) {
            eatTopToken(); eatTopToken();
            star = true;
        }
        eatTopToken(SEMICOLON);
        return new Import(makeSourceLocation(beginToken), name, star);
    }


    protected boolean eatModifiers() {
        Token token;
        while (isModifier(token=peekToken())) {
            eatTopToken();
        }
        return true;
    }

    public void parseNoModifiers() {
        parseModifiers(0);
    }

    public Modifiers parseModifiers(int legalModifiers) {
        Token beginToken = peekToken();
        int modifierValue = 0;
        while (true) {
            Integer ivalue = getModifier(peekToken());
            if (ivalue == null) break;

            int value = ivalue.intValue();
            if ((legalModifiers & value) != value) {
                showError(peekToken(), "modifier " + peekToken().image + " not allowed here");
            } else if ((modifierValue & value) != 0) {
                showError(peekToken(), "repeated modifier");
            } else {
                modifierValue |= value;
            }
            eatTopToken();
        }
        return new Modifiers(makeSourceLocation(beginToken), modifierValue);
    }

    public TypeDs parseTypeDs() {
        Token beginToken = peekToken();
        TypeD nameTypeD;
        List names = null;
        while ((nameTypeD = parseTypeD()) != null) {
            if (names == null) names = new ArrayList();
            names.add(nameTypeD);
            if (peekToken().kind != COMMA) break;
            eatTopToken();
        }
        if (names == null) {
            return null;
        } else {
            TypeDs ret = new TypeDs(dummySource,(TypeD[])names.toArray(new TypeD[0]));
            addContext(ret, beginToken);
            return ret;
        }
    }

    public String parseIdPattern() {
        if (maybeEatToken(STAR)) {
            return "*";
        } else {
            return parseId();
        }
    }

    protected boolean eatIdentifier(boolean onlyDec) {
        if (!isIdentifier(peekToken())) return false;
        eatTopToken();
        return true;
    }

    public boolean lookaheadVarDes() {
        pushCurrentIndex();
        eatModifiers();
        boolean isVarDes = (eatModifiers() && eatTypeD() && eatIdentifier(true));
        popCurrentIndex();
        return isVarDes;
    }

    public Name parseExtendedName() {
        if (!isSimpleIdentifier(peekToken())) return null;
        return parseExtendedName(null);
    }


    public boolean isSimpleIdentifier(Token t) {
        //System.out.println("t: " + t + " keyword: " + isKeyword(t));
        return t.isIdentifier() &&
            (lookupPrimitiveTypeD(t.image) != null || !isKeyword(t));
    }

    //ASPECT this is a cflow aspect
    private boolean inConstructor = false;
    public boolean inConstructor() {
        return inConstructor;
    }


    public Name parseExtendedName(Name parentName) {
        Token beginToken = peekToken();
        Name name = new Name(dummySource, parentName, parseId());
        addContext(name, beginToken);
        if (peekToken(0).kind == DOT && isSimpleIdentifier(peekToken(1))) {
            eatTopToken();
            return parseExtendedName(name);
        } else {
            return name;
        }
    }

    public String parseExtendedString(boolean allowStar) {
        if (!isIdentifier(peekToken())) return null;
        StringBuffer buf = new StringBuffer(parseId());
        while (peekToken().kind == DOT) {
            eatTopToken();
            buf.append(".");
            if (allowStar && peekToken().kind == STAR) {
                eatTopToken();
                buf.append("*");
            } else {
                buf.append(parseId());
            }
        }
        return buf.toString();
    }


    public TypeD parseNameTypeD() {
        Token beginToken = peekToken();
        Name name = parseExtendedName();
        if (name == null) {
            throwError(null, "name expected");
        }
        TypeD ret = new UnresolvedNameTypeD(dummySource,name);
        addContext(ret, beginToken);
        return ret;
    }

    public TypeD addBracketsToTypeD(TypeD typeD) {
        while (peekToken().kind == LBRACKET && peekToken(1).kind == RBRACKET) {
            eatTopToken(); eatTopToken();
            TypeD oldTypeD = typeD;
            typeD = new ArrayTypeD(dummySource,typeD);
            addContext(typeD, oldTypeD);
        }
        return typeD;
    }


    public TypeD parseTypeD() {
        Token beginToken = peekToken();
        TypeD typeD = lookupPrimitiveTypeD(beginToken.image);
        if (typeD != null) {
            eatTopToken();
        } else {
            typeD = parseNameTypeD();
        }

        addContext(typeD, beginToken);
        typeD = addBracketsToTypeD(typeD);
        return typeD;
    }

    protected boolean eatTypeD() {
        if (!isSimpleIdentifier(peekToken())) return false;
        eatTopToken();
        while (peekToken().kind == DOT) {
            eatTopToken();
            if (peekToken().kind == DOT) continue;
            if (!isSimpleIdentifier(peekToken())) {
                currentIndex -= 1;
                return true;
            }
            eatTopToken();
        }

        while (peekToken().kind == LBRACKET && peekToken(1).kind == RBRACKET) {
            eatTopToken(); eatTopToken();
        }
        return true;
    }

    public FormalDec parseFormal() {
        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiers(Modifiers.FINAL);
        TypeD typeD = parseTypeD();
        String id = parseId();
        //Identifier id = parseIdentifier();
        typeD = addBracketsToTypeD(typeD);
        FormalDec ret = new FormalDec(dummySource,modifiers, typeD, id, null);
        addContext(ret, beginToken);
        return ret;
    }

    public Formals parseFormals() {
        Token beginToken = peekToken();
        eatTopToken(LPAREN);

        Formals ret=new Formals(dummySource);
        if (peekToken().kind == RPAREN) {
            eatTopToken();
            addContext(ret, beginToken);
            return ret;
        }

        while (true) {
            FormalDec formal = parseFormal();
            ret.add(formal);

            Token t = popToken();
            if (t.kind == RPAREN) break;
            else if (t.kind == COMMA) continue;
            throwError(t, "parsing formals");
        }
        addContext(ret, beginToken);
        return ret;
    }

    public BlockStmt parseOptionalBlockStmt() {
        if (peekToken().kind == SEMICOLON) {
            eatTopToken();
            return null;
        } else {
            return parseBlockStmt();
        }
    }

    public Stmts parseStmts(int endToken) {
        Token beginToken = peekToken();
        Stmts stmts = new Stmts(dummySource);
        while (peekToken().kind != endToken) {
            stmts.add(parseStmt());
        }
        addContext(stmts, beginToken);
        eatTopToken(endToken);
        return stmts;
    }

    public BlockStmt parseBlockStmt() {
        Token beginToken = peekToken();
        eatTopToken(LBRACE);

        Stmts stmts = parseStmts(RBRACE);
        BlockStmt ret = new BlockStmt(dummySource,stmts);
        addContext(ret, beginToken);
        return ret;
    }



    public CodeBody parseOptionalCodeBody() {
        Token beginToken = peekToken();
        if (maybeEatToken(SEMICOLON)) return null;

        return parseCodeBody();
    }

    public CodeBody parseCodeBodyFully() {
        boolean oldParseInterfaceOnly = parseInterfaceOnly;
        parseInterfaceOnly = false;
        CodeBody body = null;
        try {
            body = parseCodeBody();
        } finally {
            parseInterfaceOnly = oldParseInterfaceOnly;
        }
        return body;
    }

    public CodeBody parseCodeBody() {
        Token beginToken = peekToken();

        CodeBody ret;
//        if (parseInterfaceOnly) {
//            ret = new CodeBody(getCompiler(),null,false);
//            if (beginToken.startPosition < 0)
//                throwError(beginToken, "wrong offset "+beginToken.startPosition);
//            ret.setStartPosition(beginToken.startPosition);
//            int endPos = scanTillMatched();
//            ret.setEndPosition(endPos);
//            ret.setParsed(false);
//        } else {
            eatTopToken(LBRACE);
            Stmts stmts = parseStmts(RBRACE);
            ret = new CodeBody(dummySource,stmts,true);
            ret.setParsed(true);
            addContext(ret, beginToken);
//        }
        return ret;
    }

    public ConstructorBody parseConstructorBody() {
        Token beginToken = peekToken();
        if (maybeEatToken(SEMICOLON)) {
            showError(beginToken, "constructor must have a body");
            return null;
        }

        ConstructorBody ret;
//        if (parseInterfaceOnly) {
//            ret = new ConstructorBody(dummySource,null, null, false);
//            if (beginToken.startPosition < 0)
//                throwError(beginToken, "wrong offset "+beginToken.startPosition);
//            ret.setStartPosition(beginToken.startPosition);
//            int endPos = scanTillMatched();
//            ret.setEndPosition(endPos);
//            ret.setParsed(false);
//        } else {
            eatTopToken(LBRACE);

            ConstructorCallExpr constructorCall = null;

            Token firstToken = peekToken();
            if (firstToken.image.equals("this") || firstToken.image.equals("super")) {
                boolean isSuper = firstToken.image.equals("super");
                if (peekToken(1).kind == LPAREN) {
                    eatTopToken();
                    constructorCall = new ConstructorCallExpr(dummySource, null, isSuper, parseParenExprs(), null);
                    addContext(constructorCall, firstToken);
                }
            }

            inConstructor = true;
            Stmts stmts = parseStmts(RBRACE);
            inConstructor = false;
            if (constructorCall == null && stmts.size() > 0) {
                Stmt stmt = stmts.get(0);
                if (stmt instanceof ExprStmt && ((ExprStmt)stmt).getExpr() instanceof ConstructorCallExpr) {
                    constructorCall = (ConstructorCallExpr) ((ExprStmt)stmt).getExpr();
                    stmts.remove(0);
                }
            }
            ret = new ConstructorBody(dummySource,stmts, constructorCall, true);
            ret.setParsed(true);
            addContext(ret, beginToken);
//        }
        return ret;
    }




    // too much like the above!!!
    public Decs parseDecs() {
        return parseDecs(null);
    }

    public Decs parseDecs(Map extraDecParsers) {
        Token beginToken = peekToken();
        eatTopToken(LBRACE);

        Decs decs = new Decs(dummySource);
        while (peekToken().kind != RBRACE) {
            if (peekToken().kind == SEMICOLON) {
                eatTopToken(); continue;
            }
            Token decBeginToken = peekToken();
            Stmt dec = parseStmt(extraDecParsers, true, true);
            addContext(dec, decBeginToken);
            if (dec instanceof Dec) {
                decs.add((Dec)dec);
            } else if (dec instanceof Decs) {
                decs.addAll((Decs)dec);
            } else {
                throwError(decBeginToken, "expected declaration");
                return null;
            }
        }

        eatTopToken(RBRACE);
        addContext(decs, beginToken);
        return decs;
    }

    public TypeDs parseThrows() {
        if (maybeEatKeyword("throws")) {
            return parseTypeDs();
        } else {
            return null;
        }
    }

    public InitializerDec parseInitializerDec() {
        int legalModifiers = Modifiers.STATIC;
        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiers(legalModifiers);
        Formals formals = new Formals(dummySource);
        addContext(formals, beginToken);
        TypeDs _throws = new TypeDs(dummySource);
        addContext(_throws, beginToken);
        return new InitializerDec(dummySource, modifiers, formals, _throws, parseOptionalCodeBody());
    }

    public Dec parseConstructorDec() {
        int legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE;

        Token beginToken = peekToken();

        if (inInterface) { showError(beginToken, "constructor not allowed in interface"); }

        Modifiers modifiers = parseModifiers(legalModifiers);
        parseIdMatchingEnclosingTypeName("constructor name");
        Formals formals = parseFormals();
        TypeDs _throws = parseThrows();

        return new ConstructorDec(dummySource,modifiers, formals, _throws, parseConstructorBody());
    }

    public Dec parseMethodDec() {
        int legalModifiers;
        if (inInterface) {
            legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT;
        } else {
            legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
                Modifiers.ABSTRACT | Modifiers.STATIC | Modifiers.FINAL |
                Modifiers.SYNCHRONIZED | Modifiers.NATIVE | Modifiers.STRICT;
        }

        Modifiers modifiers = parseModifiers(legalModifiers);
        if (inInterface) {
            modifiers.setAbstract(true);
            modifiers.setPublic(true);
        }


        TypeD typeD = parseTypeD();
        String id = parseId();

        Formals formals = parseFormals();

        // handle funky "int foo()[] {" return type specifier
        typeD = addBracketsToTypeD(typeD);

        TypeDs _throws = parseThrows();

        return new MethodDec(dummySource,modifiers, typeD, id, formals, _throws,
                             parseOptionalCodeBody());
    }

    public Stmt parseVarDec() { return parseVarDec(false); }
    public Stmt parseFieldDec() { return parseVarDec(true); }


    // returns either a multivardec or a vardec
    public Stmt parseVarDec(boolean isField) {
        int legalModifiers;

        if (isField) {
            if (inInterface) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.STATIC | Modifiers.FINAL;
            } else {
                legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
                    Modifiers.STATIC | Modifiers.FINAL |
                    Modifiers.TRANSIENT | Modifiers.VOLATILE;
            }
        } else {
            legalModifiers = Modifiers.FINAL;
        }

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiers(legalModifiers);

        if (inInterface) {
            modifiers.setPublic(true);
            modifiers.setStatic(true);
        }

        TypeD baseTypeD = parseTypeD();

        Decs decs = null;
        VarDec dec = null;

        while (true) {
            String id = parseId();
            TypeD typeD = baseTypeD;

            Expr initializer = null;
            typeD = addBracketsToTypeD(typeD);
            //VarDes des = new VarDes(modifiers, id, typeD);
            //??? what should the context really be for a var des in a list
            //addContext(des, beginToken);

            Token endToken = popToken();
            if (endToken.kind == ASSIGN) {
                if (typeD instanceof ArrayTypeD && peekToken().kind == LBRACE) {
                    initializer = parseArrayInitializer();
                } else {
                    initializer = parseExpr();
                }
                endToken = popToken();
            }

            if (isField) {
                dec = new FieldDec(dummySource,(Modifiers)modifiers.copy(), typeD, id, initializer);
            } else {
                dec = new VarDec(dummySource,(Modifiers)modifiers.copy(), typeD, id, initializer);
            }
            addContext(dec, beginToken);

            if (endToken.kind == SEMICOLON) {
                if (decs != null) decs.add(dec);
                break;
            } else if (endToken.kind == COMMA) {
                if (decs == null) decs = new Decs(dummySource,dec);
                else decs.add(dec);

                // this copy is needed to keep the AST a tree and not a graph
                baseTypeD = (TypeD)baseTypeD.copy();
                continue;
            } else {
                throwError(endToken, "expected ',' or ';'");
            }
        }

        if (decs != null) return decs;
        else return dec;
    }

    public String parseAnyId() {
        Token token = popToken();
        if (!isIdentifier(token)) {
            throwError(token, "identifier expected");
        }
        return token.image;
    }

    public String parseId() {
        Token token = popToken();
        if (!isIdentifier(token) || isKeyword(token)) {
            throwError(token, "identifier expected");
        }
        return token.image;
    }

    public String parseIdOrKeywords_this_super_class() {
        Token token = popToken();
        if (!isIdentifier(token) || isKeyword(token)) {
            if (!(  token.image.equals("this")
                 || token.image.equals("super")
                 || token.image.equals("class")
                ))
            throwError(token, "identifier expected");
        }
        return token.image;
    }


    public FormalDec parseParenFormal() {
        eatTopToken(LPAREN);
        FormalDec formal = parseFormal();
        eatTopToken(RPAREN);

        return formal;
    }

    public Exprs parseExprs(int stopKind) {
        Token beginToken = peekToken();
        Exprs exprs = new Exprs(dummySource);

        if (beginToken.kind != stopKind) {
            while (true) {
                Expr expr = parseExpr();
                exprs.add(expr);
                if (peekToken().kind != COMMA) break;
                eatTopToken();
            }
        }
        addContext(exprs, beginToken);
        return exprs;
    }

    //public Exprs parseExprs(int stopKind) {
    //    return new Exprs(parseExprArray(stopKind));
    //}

    public Exprs parseParenExprs() {
        eatTopToken(LPAREN);
        Exprs exprs = parseExprs(RPAREN);
        eatTopToken(RPAREN);
        return exprs;
    }


    public ArrayInitializer parseArrayInitializer() {
        ArrayInitializer ret;
        eatTopToken(LBRACE);

        Exprs exprs = new Exprs(dummySource);

        if (maybeEatToken(RBRACE)) {
            return new ArrayInitializer(dummySource,exprs);
        }

        while (true) {
            Expr expr;
            if (peekToken().kind == LBRACE) {
                expr = parseArrayInitializer();
            } else if (maybeEatToken(COMMA)) {
                break;
            } else {
                expr = parseExpr();
            }
            exprs.add(expr);
            if (!maybeEatToken(COMMA)) break;
            if (peekToken().kind == RBRACE) break;
        }
        eatTopToken(RBRACE);
        return new ArrayInitializer(dummySource,exprs);
    }

    public Exprs parseArrayDimExprs() {
        // check for an array type
        Exprs exprs = new Exprs(dummySource);
        while (peekToken().kind == LBRACKET) {
            eatTopToken();
            if (peekToken().kind == RBRACKET) {
                exprs.add(new EmptyExpr(dummySource));
            } else {
                exprs.add(parseExpr());
            }
            eatTopToken(RBRACKET);
        }
        return exprs;
    }

    public Expr parseParenExpr() {
        eatTopToken(LPAREN);
        Expr expr = parseExpr();
        eatTopToken(RPAREN);

        return expr;
    }

    public Expr parseExpr() {
        return parseExpr(false);
    }

    public Expr parseExpr(boolean allowNulls) {
        Expr ret = exprParser.parse();
        if (!allowNulls && ret == null) {
            throwError(null, "expression required");
        }
        return ret;
    }

    protected String grabFormalComment(Token token) {
        while ((token = token.specialToken) != null) {
            if (token.kind == FORMAL_COMMENT) {
                return token.image;
            }
        }
        return null;
    }

    protected void addComments(ASTObject ast, Token beginToken) {
        //System.out.println(ast + ": " + beginToken);

        // we only stick formal comments on Dec's
        if (ast instanceof Dec) {
            String comment = grabFormalComment(beginToken);
            if (comment != null) ast.setFormalComment(comment);
        }
    }

    protected SourceLocation makeSourceLocation(Token beginToken) {
        Token endToken = peekToken(-1);
        if (endToken == null) endToken = beginToken;
        SourceLocation ret = new TextSourceLocation(currentCompilationUnit,
                                                beginToken.startPosition,
                                                endToken.endPosition);
        return ret;
    }

    protected SourceLocation makeSourceLocation(SourceLocation beginSourceLoc) {
        Token endToken = peekToken(-1);
        int startPosition = -1;
        if (beginSourceLoc != null) startPosition = beginSourceLoc.getStartPosition();
        SourceLocation ret = new TextSourceLocation(currentCompilationUnit,
                                                startPosition,
                                                endToken.endPosition);
        return ret;
    }

    protected void addContext(ASTObject ast, ASTObject beginAST) {
        if (ast == null) return;
        ast.setSourceLocation(makeSourceLocation(beginAST.getSourceLocation()));
    }

    protected void addContext(ASTObject ast, Token beginToken) {
        if (ast == null) return;
        ast.setSourceLocation(makeSourceLocation(beginToken));
        addComments(ast, beginToken);
    }

    public Stmt parseStmt() {
        Token beginToken = peekToken();
        Stmt ret = parseStmt(null, false, false);
        addContext(ret, beginToken);
        return ret;
    }

    public boolean lookaheadFormals() {
        if (peekToken().kind != LPAREN) return false;
        if (peekToken(1).kind == RPAREN) return true;

        int start = currentIndex;
        eatTopToken();

        if (eatModifiers() && eatTypeD() && isIdentifier(peekToken())) {
            currentIndex = start;
            return true;
        }
        currentIndex = start;
        return false;
    }


    private Stmt parseStmt(Map extraDecParsers,
                           boolean allowConstructors, boolean onlyDec) {
        Token topToken = peekToken();

        // Step 1. check for a statement keyword or a labeled statement
        if (!onlyDec && isIdentifier(topToken)) {
            String keyword = topToken.image;
            StmtParser stmtParser = (StmtParser)stmtParsers.get(keyword);
            if (stmtParser != null) {
                eatTopToken();
                return stmtParser.parse();
            } else if (peekToken(1).kind == COLON) {
                if (isKeyword(popToken())) { //label
                    showError(peekToken(-1), "keyword not allowed as label");
                }
                eatTopToken(); // :
                Stmt stmt = parseStmt(extraDecParsers, allowConstructors, onlyDec);
                return new LabeledStmt(dummySource,keyword, stmt);
            }
        }

        // Step 2. check for the variety of possible decs
        // method: typed id(
        // constructor: id(
        // var: typed id
        // initialized: {
        // named: keyword ...

        pushCurrentIndex();
        eatModifiers();
        topToken = peekToken();

        // check for named dec
        if (isIdentifier(topToken)) {
            String keyword = topToken.image;

            DecParser decParser;
            if (allowConstructors) {
                decParser = (DecParser)decParsers.get(keyword);
            } else {
                decParser = (DecParser)stmtDecParsers.get(keyword);
            }

            if (extraDecParsers != null && decParser == null) {
                decParser = (DecParser)extraDecParsers.get(keyword);
            }
            if (decParser != null) {
                popCurrentIndex();
                return decParser.parse(onlyDec);
            }
        }

        if (allowConstructors) {
            //!!! inserting some strange code here for the sake of introduction blocks
            if (isIdentifierOrNew(topToken)) {
                eatTopToken();
                if (lookaheadFormals()) {
                    popCurrentIndex();
                    return parseConstructorDec();
                } else {
                    currentIndex -= 1;
                }
            }
            if (topToken.kind == LBRACE) {
                popCurrentIndex();
                return parseInitializerDec();
                //return new InitializerDec(new Des[] {new InitializerDes(modifiers, null)},
                //  parseBlockStmt());
            }
        }

        if (eatTypeD() && eatIdentifier(onlyDec)) {
            if (peekToken().kind == LPAREN) {
                popCurrentIndex();
                return parseMethodDec();
            } else {
                popCurrentIndex();
                if (onlyDec) {
                    return parseFieldDec();
                } else {
                    return parseVarDec();
                }
            }
        }

        // give up on looking for a dec
        popCurrentIndex();
        if (onlyDec) {
            return noDecFound();
        }

        // Step 3. parse unnamed stmts
        if (topToken.kind == SEMICOLON) {
            eatTopToken();
            return new EmptyStmt(dummySource);
        }

        if (topToken.kind == LBRACE) {
            return parseBlockStmt();
        }

        Expr expr = parseExpr();
        eatTopToken(SEMICOLON);
        return new ExprStmt(dummySource,expr);
    }

    protected Dec noDecFound() {
        throwError(peekToken(), "declaration or '}' expected");
        return null;
    }


    private Stmt parseTopDec(Map extraDecParsers) {
        Token topToken = peekToken();

        pushCurrentIndex();
        eatModifiers();
        topToken = peekToken();

        // check for named dec
        if (isIdentifier(topToken)) {
            String keyword = topToken.image;
            DecParser decParser = (DecParser)decParsers.get(keyword);
            if (extraDecParsers != null && decParser == null) {
                decParser = (DecParser)extraDecParsers.get(keyword);
            }
            if (decParser != null) {
                popCurrentIndex();
                return decParser.parse(true);
            }
        }

        // give up on looking for a dec
        popCurrentIndex();
        throwError(topToken, "type declaration expected");
        return null;
    }


    /* error handling code */
    public void throwError(Token t, String message) {
        if (t == null) t = peekToken();
        //System.out.println(t);
        //if (getCompiler().getOptions().dumpstack) Thread.dumpStack();
        throw new ParseException(t, tokenSource.sourceInfo, message);
    }


    public void warnVersion(String version, Token t, String message) {
        getCompiler().warnVersion(version, tokenToASTObject(t), message);
    }

    public void showError(Token t, String message) {
         getCompiler().showError(tokenToASTObject(t), message);
    }

    // utility function for interacting with code that expects an ASTObject
    static class FakeASTObject extends ASTObject {
        public FakeASTObject(SourceLocation loc) {
            super(loc);
            //setSourceLocation(loc);
            //XXX should be new CompilationUnit(file);
//            source = new CompilationUnit(compiler,null, null, null, _sourceInfo,true);
//            startPosition = _startPosition;
        }
        public ASTObject copy() { return null; }
        public int getChildCount() { return 0; }
    }

    public ASTObject tokenToASTObject(Token t) {
        ASTObject fake = new FakeASTObject(makeSourceLocation(t));//tokenSource.sourceInfo, t.startPosition);
        return fake;
    }


    // dec parsers
    public static abstract class DecParser {
        public Dec parse(boolean isNotMethodLocal) {
            // actually, boolean doesn't matter, since class/interface/etc
            // parsers should override parse(boolean) method
            return parse();
        }
        abstract public Dec parse();
    }

    class InterfaceDecParser extends DecParser {
        public Dec parse() {
            // not used, we override parse(boolean)
            throw new RuntimeException();
        }
        public Dec parse(boolean isNotMethodLocal) {
            int legalModifiers;
            if (!isNotMethodLocal) {
                legalModifiers = Modifiers.ABSTRACT | Modifiers.STRICT;
            } else if (peekEnclosingTypeDec() == null) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.STRICT;
            } else if (inInterface) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT |
                    Modifiers.STRICT | Modifiers.STATIC;
            } else {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.STRICT |
                    Modifiers.PRIVATE | Modifiers.PROTECTED | Modifiers.STATIC;
            }

            Modifiers modifiers = parseModifiers(legalModifiers);
            if (inInterface) {
                modifiers.setPublic(true);
            }
            eatKeyword("interface");
            String id = parseId();

            TypeDs _extends = null;
            if (maybeEatKeyword("extends")) {
                _extends = parseTypeDs();
            }

            TypeDec ret = new InterfaceDec(dummySource, modifiers, id, _extends, null);
            ret.setEnclosingTypeDec(peekEnclosingTypeDec());

            pushEnclosingTypeDec(ret);
            Decs body = null;
            boolean oldInInterface = inInterface;
            try {
                inInterface = true;
                body = parseDecs();
            } finally {
                inInterface = oldInInterface;
            }
            popEnclosingTypeDec();
            ret.setBody(body);

            if (peekEnclosingTypeDec() == null) currentCompilationUnit.addDefinedType(ret);
            return ret;
        }
    }

    class ClassDecParser extends DecParser {
        public Dec parse() {
            // not used, we override parse(boolean)
            throw new RuntimeException();
        }
        public Dec parse(boolean isNotMethodLocal) {
            int legalModifiers;
            if (!isNotMethodLocal) {
                legalModifiers = Modifiers.ABSTRACT | Modifiers.FINAL | Modifiers.STRICT;
            } else if (peekEnclosingTypeDec() == null) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT;
            } else if (inInterface) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT | Modifiers.STATIC;
            } else {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT |
                    Modifiers.PRIVATE | Modifiers.PROTECTED | Modifiers.STATIC;
            }

            Modifiers modifiers = parseModifiers(legalModifiers);
            if (inInterface) {
                modifiers.setPublic(true);
            }

            eatKeyword("class");
            String id = parseId();

            TypeD superClass = null;
            if (maybeEatKeyword("extends")) {
                superClass = parseTypeD();
            }
            TypeDs _implements = null;
            if (maybeEatKeyword("implements")) {
                _implements = parseTypeDs();
            }

            TypeDec ret =
                new ClassDec(dummySource, modifiers, id, superClass, _implements, null);
            if(!isNotMethodLocal) ret.setLocal();
            ret.setEnclosingTypeDec(peekEnclosingTypeDec());

            pushEnclosingTypeDec(ret);

            Decs body = null;
            boolean oldInInterface = inInterface;
            try {
                inInterface = false;
                body = parseDecs();
            } finally {
                inInterface = oldInInterface;
            }
            popEnclosingTypeDec();

            ret.setBody(body);
            if (peekEnclosingTypeDec() == null) currentCompilationUnit.addDefinedType(ret);
            return ret;
        }
    }


    // stmt parsers
    public static abstract class StmtParser {
        abstract public Stmt parse();
    }

    class BreakStmtParser extends StmtParser {
        public Stmt parse() {
            String label = null;
            if (peekToken().kind != SEMICOLON) {
                label = parseId();
            }
            eatTopToken(SEMICOLON);
            return new BreakStmt(dummySource,label);
        }
    }

    class ContinueStmtParser extends StmtParser {
        public Stmt parse() {
            String label = null;
            if (peekToken().kind != SEMICOLON) {
                label = parseId();
            }
            eatTopToken(SEMICOLON);
            return new ContinueStmt(dummySource,label);
        }
    }

    class DoStmtParser extends StmtParser {
        public Stmt parse() {
            Stmt body = parseStmt();

            eatKeyword("while");
            Expr test = parseParenExpr();
            eatTopToken(SEMICOLON);

            return new DoStmt(dummySource, body, test);
        }
    }

    class AssertStmtParser extends StmtParser {
        public Stmt parse() {
            Expr test = parseExpr();
            Expr message = null;
            if (maybeEatToken(COLON)) {
                message = parseExpr();
            }
            eatTopToken(SEMICOLON);

            return new AssertStmt(dummySource, test, message);
        }
    }

    class ForStmtParser extends StmtParser {
        public Stmt parse() {
            eatTopToken(LPAREN);
            ASTObject init;
            Expr test;
            Exprs update;

            // check for a var dec
            if (lookaheadVarDes()) {
                init = parseVarDec();
            } else {
                init = parseExprs(SEMICOLON);
                eatTopToken(SEMICOLON);
            }

            test = parseExpr(true);
            eatTopToken(SEMICOLON);
            update = parseExprs(RPAREN);
            eatTopToken(RPAREN);
            Stmt body = parseStmt();

            ForStmt forStmt = new ForStmt(dummySource,init, test, update, body);
            if (test == null) {
                forStmt.setTest((Expr)
                                new BooleanLiteralExpr(dummySource, true)
                                    .setSyntheticSource(forStmt));
            }
            return forStmt;
        }
    }

    class IfStmtParser extends StmtParser {
        public Stmt parse() {
            Expr test = parseParenExpr();
            Stmt _then = parseStmt();
            Stmt _else = null;
            if (maybeEatKeyword("else")) {
                _else = parseStmt();
            }
            IfStmt ifStmt = new IfStmt(dummySource,test, _then, _else);
            if (_else == null) {
                ifStmt.setElse((Stmt)
                               new EmptyStmt(dummySource).setSyntheticSource(ifStmt));
            }
            return ifStmt;
        }
    }

    class ReturnStmtParser extends StmtParser {
        public Stmt parse() {
            Expr expr = null;
            if (peekToken().kind != SEMICOLON) {
                expr = parseExpr();
            }
            eatTopToken(SEMICOLON);
            return new ReturnStmt(dummySource,expr);
        }
    }

    class SwitchStmtParser extends StmtParser {
        public Stmt parse() {
            Expr expr = parseParenExpr();

            eatTopToken(LBRACE);
            SwitchClauses clauses = new SwitchClauses(dummySource);
            Expr label = null;
            Stmts stmts = null;
            SwitchClause switchClause=null;

            while (true) {
                if (maybeEatKeyword("case")) {
                    label = parseExpr();
                    eatTopToken(COLON);
                    if (switchClause != null) clauses.add(switchClause);
                    stmts = new Stmts(dummySource);
                    switchClause = new SwitchClause(dummySource,label, stmts);
                } else if (maybeEatKeyword("default")) {
                    eatTopToken(COLON);
                    if (switchClause != null) clauses.add(switchClause);
                    stmts = new Stmts(dummySource);
                    switchClause = new SwitchClause(dummySource,null, stmts);
                } else if (peekToken().kind == RBRACE) {
                    if (switchClause != null) clauses.add(switchClause);
                    eatTopToken();
                    break;
                } else {
                    if (stmts == null) {
                        throwError(peekToken(), "'case', 'default' or '}' expected");
                        break;
                    }
                    stmts.add(parseStmt());
                }
            }

            return new SwitchStmt(dummySource,expr, clauses);
        }
    }

    class SynchronizedStmtParser extends StmtParser {
        public Stmt parse() {
            Expr expr = parseParenExpr();
            Stmt body = parseStmt();

            return new SynchronizedStmt(dummySource,expr, body);
        }
    }

    class ThrowStmtParser extends StmtParser {
        public Stmt parse() {
            Expr expr = parseExpr();
            eatTopToken(SEMICOLON);
            return new ThrowStmt(dummySource,expr);
        }
    }

    class GotoStmtParser extends StmtParser {
        public Stmt parse() {
            showError(peekToken(),"'goto' is java reserved keyword");
            return null;
        }
    }

    class TryStmtParser extends StmtParser {
        public Stmt parse() {
            Stmt body = parseStmt();

            CatchClauses catches = new CatchClauses(dummySource);
            Token beginToken = peekToken();
            while (maybeEatKeyword("catch")) {
                FormalDec formal = parseParenFormal();
                Stmt catchBody = parseStmt();
                CatchClause clause = new CatchClause(dummySource,formal, catchBody);
                addContext(clause, beginToken);
                catches.add(clause);
                beginToken = peekToken();
            }

            Stmt _finally = null;
            if (maybeEatKeyword("finally")) {
                _finally = parseStmt();
            }

            if (catches.size() == 0) {
                return new TryFinallyStmt(dummySource, body, _finally);
            } else if (_finally == null) {
                return new TryCatchStmt(dummySource, body, catches);
            } else {
                TryFinallyStmt s = new TryFinallyStmt(dummySource, null, _finally);
                TryCatchStmt b = new TryCatchStmt(dummySource, body, catches);
                b.setSyntheticSource(s);
                s.setBody(b);
                return s;
            }
        }
    }

    class WhileStmtParser extends StmtParser {
        public Stmt parse() {
            Expr test = parseParenExpr();
            Stmt body = parseStmt();
            return new WhileStmt(dummySource,test, body);
        }
    }

}


/* Thoughts on macros

macro for (ExprList init; Expr[boolean] test; ExprList update) {Stmt @code} => {
  @init;
  while (@test) {
    @code;
    @update;
  }
}

macro do {@code} while (@test); => {
  while (true) {
    @code;
    if (@test) break;
  }
}

macro foreach (@type @var) in (@sequence) {@code} => {
  Collection temp = @sequence;
  for (Iterator iter = temp.iterator(); iter.hasNext(); ) {
    @type @var = (@type)iter.next();
    @code;
  }
}

*/
