/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts;

import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.util.CollectionUtil;

import java.util.*;

/**************************************************************************
 * Weave into a single TypeDec
 */
public class JoinPointCollector extends CompilerObject {
    TypeDec typeDec;
    
    public JoinPointCollector(TypeDec typeDec) {
        super(typeDec.getCompiler());
        this.typeDec = typeDec;
    }
    
    public void collect() {
        if (typeDec.isSynthetic()) return;
        getCompiler().showMessage("  collecting join point shadows for " + typeDec.toShortString());
        collectJoinPoints();
        //System.out.println("wove: " + typeDec.unparse());
    }
    
    /* Need to replicate this for synthetic callee-side call points *
    void planSyntheticReceptionPoints() {
        Collection inheritedMethods = typeDec.getType().getInheritedMethods();
        
        for (Iterator iter = inheritedMethods.iterator(); iter.hasNext(); ) {
            Method method = (Method)iter.next();
            if (method.isStatic()) continue;
            
            //XXX only make synthetic reception points when the parent
            //XXX method isn't available to us
            if (method.getMethodDec().fromSource()) continue;
            
            //XXX don't make synthetic reception points for final methods
            if (method.getMethodDec().isFinal()) continue;
            
            planJoinPoint(new SyntheticReceptionPoint(getCompiler(),typeDec, method.getMethodDec()));
        }
    }
    /**/
    
    void collectJoinPoints() {
//         JoinPoint staticInitializerExecutionPoint =
//                 new StaticInitializerExecutionPoint(getCompiler(),typeDec);

        //!!! interfaces don't have static initializers
        if (!(typeDec instanceof InterfaceDec)) {
            planInitializerExecution(typeDec.getSingleInitializerDec(true));
        }
        planInitializerExecution(typeDec.getSingleInitializerDec(false));
        
        //XXX this design depends on typeDec.getBody() not being modified during loop
        Decs decs = typeDec.getBody();
        //final int N = decs.size();
        for (int i = 0; i < decs.size(); i++) {
            Dec dec = decs.get(i);
            //System.out.println("dec: " + dec.toShortString());
            
            if (dec instanceof InitializerDec) {
                //System.out.println("planning body");
                continue;
            } else if (dec instanceof CodeDec) {
                planJoinPoints((CodeDec)dec);
            } else if (dec instanceof TypeDec) {
                new JoinPointCollector((TypeDec)dec).collect();
            } else if (dec instanceof FieldDec) {
                FieldDec field = (FieldDec)dec;
                // don't do anything with fields
                // their initializers are handled by the initializer decs
            } else {
                continue;
            }
        }
    }


    void planInitializerExecution(InitializerDec dec) {
        JoinPoint jp;
        if (dec.isStatic()) {
            jp = new StaticInitializerExecutionJp(dec);
        } else {
            jp = new InstanceInitializerExecutionJp(dec);
        }
        planBodyJoinPoints(dec, jp);
        planJoinPoint(jp);
    }
    
    JoinPoint makeExecutionJoinPoint(CodeDec codeDec) {
        //System.out.println(codeDec.toShortString() + ", " + codeDec.isAdvisable());            
        if (codeDec.isAbstract() || codeDec.getBody() == null) return null;
        if (!codeDec.isLanguageVisible()) return null;
        
        //System.out.println(codeDec.toShortString());
        
        if (codeDec instanceof MethodDec) {
            return new MethodExecutionJp((MethodDec)codeDec);
        } else if (codeDec instanceof ConstructorDec) {
            return new ConstructorExecutionJp((ConstructorDec)codeDec);
        } else if (codeDec instanceof AdviceDec) {
            return new AdviceExecutionJp((AdviceDec)codeDec);
        } else if (codeDec instanceof InitializerDec) {
            // initializers are handled specially
            return null;
        } else {
            throw new RuntimeException("unexpected code: " + codeDec.getClass().getName());
        }
    }
    
    CalleeSideCallJp planCalleeSideCallPoint(MethodDec methodDec) {
        if (!methodDec.fromSource()) return null;
        
        //???
        if (methodDec.isAbstract()) return null;
        
        CalleeSideCallJp point = (CalleeSideCallJp)getWorld().calleeSideCallPoints.get(methodDec);
        if (point == null) {
            point = new CalleeSideCallJp(methodDec.getDeclaringType().getTypeDec(), methodDec);
            getWorld().calleeSideCallPoints.put(methodDec, point);
        }
        
        return point;
    }
    
    void planCallPoint(AnyCallExpr callExpr, JoinPoint enclosingJp) {
        if (callExpr instanceof CallExpr) {
            planBodyJoinPoint(new MethodCallJp((CallExpr)callExpr, enclosingJp));
        } else {
            planBodyJoinPoint(new ConstructorCallJp((NewInstanceExpr)callExpr, enclosingJp));
        }
    }
    
    void planJoinPoints(CodeDec codeDec) {
        if (!codeDec.isLanguageVisible()) return;
        
        JoinPoint executionPoint = makeExecutionJoinPoint(codeDec);
        
        if (executionPoint != null) {
            planBodyJoinPoints(codeDec.getBody(), executionPoint);
            
            //XXX order matters here because of implementation for enclosingExecutionPoint
            planJoinPoint(executionPoint);
        }
        
        if (codeDec instanceof ConstructorDec) {
            planJoinPoint(new InitializationJp((ConstructorDec)codeDec));
        }
        
        if (codeDec instanceof MethodDec) {
            JoinPoint calleeCallPoint = planCalleeSideCallPoint((MethodDec)codeDec);
            if (calleeCallPoint != null) typeDec.joinPoints2.add(calleeCallPoint);            
        }
    }
    
    void planBodyJoinPoints(ASTObject node, JoinPoint executionPoint) {
        //if (!anyBodyPoints) return;
        
        new BodyPointPlanner(getCompiler(), executionPoint).process(node);
    }
    
    void planJoinPoint(JoinPoint jp) {
        typeDec.joinPoints2.add(jp);
    }
    
    void planBodyJoinPoint(CodeBodyJp jp) {
        typeDec.joinPoints1.add(jp);
        if (jp.getEnclosingExecutionJoinPoint() instanceof AdviceExecutionJp) {
            AdviceDec ad =
                (AdviceDec)jp.getEnclosingExecutionJoinPoint().getTargetDec();
            if (ad instanceof AroundAdviceDec) {
                ((AroundAdviceDec)ad).joinPoints1.add(jp);
            }
        }
    }
    

    
    class BodyPointPlanner extends Walker {
        JoinPoint enclosingExecutionJoinPoint;
        
        public BodyPointPlanner(JavaCompiler compiler, JoinPoint enclosingExecutionJoinPoint) {
            super(compiler);
            this.enclosingExecutionJoinPoint = enclosingExecutionJoinPoint;
        }
        
        public boolean preProcess(ASTObject node) {           
            if (node instanceof ConstructorCallExpr) {
                if (!(enclosingExecutionJoinPoint instanceof PreInitializationJp)) {
                    JoinPoint preInitPoint =
                        new PreInitializationJp((ConstructorDec)node.getEnclosingCodeDec());
                    planJoinPoint(preInitPoint);
                    new BodyPointPlanner(this.getCompiler(), preInitPoint).process(node);
                    return false;
                } else {
                    return true;
                }
            }
            
            //XXX don't capture new array operations
            if (node instanceof AnyCallExpr && !(node instanceof NewArrayExpr)) {
                AnyCallExpr callExpr = (AnyCallExpr)node;
                
                Expr expr = callExpr.getExpr();
                if (callExpr instanceof CallExpr && ((CallExpr)callExpr).getIsSuper()) {
                    // super calls don't count as "calls" joinpoints
                    // but we have to give them very special handling regarding
                    // callee-side call joinpoints
                    //????maybeFixSuperCall((CallExpr)callExpr);
                    return true;
                }
                
                //make sure we don't count proceed as a call
                if (expr instanceof ProceedExpr) {
                    return true;
                }
                
                planCallPoint(callExpr, enclosingExecutionJoinPoint);
            } else if (node instanceof CatchClause) {
                CatchClause catchClause = (CatchClause)node;                
                planBodyJoinPoint(new ExceptionHandlerJp(catchClause, enclosingExecutionJoinPoint));
            } else if (node instanceof TypeDec) {
                new JoinPointCollector((TypeDec)node).collect();
                return false;
            } else if (node instanceof FieldAccessExpr) {
                FieldAccessExpr fieldAccessExpr = (FieldAccessExpr)node;                
                if (fieldAccessExpr.isLhs()) {
                    BangExpr bangExpr = (BangExpr)node.getParent();
                    
                    if (bangExpr instanceof BasicAssignExpr) {
                        addSetPoint((AssignExpr) bangExpr);
                    } else {
                        SyntheticGetSet synExpr = new SyntheticGetSet(this.getCompiler(), bangExpr);
                        addSetPoint(fieldAccessExpr, synExpr);
                        addGetPoint(fieldAccessExpr, synExpr);
                    }
                } else {
                    addGetPoint(fieldAccessExpr);
                }
            }
            
            return true;
        }
        
        void addGetPoint(final FieldAccessExpr fieldAccessExpr) {
            FieldGetJp point = new FieldGetJp(
                new FieldGetJp.FieldGetExprPromise() {
                    public FieldAccessExpr getFieldAccessExpr() { return fieldAccessExpr; }
                    
                    public FieldDec getFieldDec() { return getFieldAccessExpr().getFieldDec(); }
                    public Expr getSourceExpr() { return fieldAccessExpr; }
                    public Type getTargetType() { return getFieldAccessExpr().getTargetType(); }
                    public ASTObject getSourceLocation() { return fieldAccessExpr; }
                },
                enclosingExecutionJoinPoint);
            planBodyJoinPoint(point);
        }
        
        FieldSetJp.FieldSetExprPromise makeFieldSetExprPromise(
              final FieldAccessExpr fieldAccessExpr,
              final SyntheticGetSet synGetSet, final boolean useBang)
        {
            return new FieldSetJp.FieldSetExprPromise() {
                    public AssignExpr getAssignExpr() { synGetSet.makeConcrete(); return synGetSet.makeSetExpr(); }
                    public FieldAccessExpr getFieldAccessExpr() {
                        synGetSet.makeConcrete(); return synGetSet.makeGetExpr(); }
                    
                    public FieldDec getFieldDec() { return fieldAccessExpr.getFieldDec(); }
                    public Expr getSourceExpr() {
                        if (useBang) return synGetSet.getBangExpr();
                        else return fieldAccessExpr;
                    }
                    public Type getTargetType() { return fieldAccessExpr.getTargetType(); }
                    public ASTObject getSourceLocation() { return fieldAccessExpr; }
                };
        }
        
        void addGetPoint(FieldAccessExpr fieldAccessExpr, SyntheticGetSet synGetSet) {
            planBodyJoinPoint(new FieldGetJp(
                makeFieldSetExprPromise(fieldAccessExpr, synGetSet, false),
                enclosingExecutionJoinPoint));
        }
        
        void addSetPoint(FieldAccessExpr fieldAccessExpr, SyntheticGetSet synGetSet) {
            //FieldAccessExpr fieldAccessExpr = (FieldAccessExpr)assignExpr.getLhs();
            planBodyJoinPoint(new FieldSetJp(
                        makeFieldSetExprPromise(fieldAccessExpr, synGetSet, true),
                        enclosingExecutionJoinPoint));
        }
        void addSetPoint(final AssignExpr assignExpr) {
            FieldSetJp point = new FieldSetJp(
                new FieldSetJp.FieldSetExprPromise() {
                    public AssignExpr getAssignExpr() { return assignExpr; }
                    public FieldAccessExpr getFieldAccessExpr() {
                        return (FieldAccessExpr)assignExpr.getLhs(); }
                    
                    public FieldDec getFieldDec() { return getFieldAccessExpr().getFieldDec(); }
                    public Expr getSourceExpr() { return assignExpr; }
                    public Type getTargetType() { return getFieldAccessExpr().getTargetType(); }
                    public ASTObject getSourceLocation() { return assignExpr; }
                },
                enclosingExecutionJoinPoint);
            
            planBodyJoinPoint(point);
        }
    }
}

