/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.base.JavaCompiler;

/**
 * @grammar after (formals) returning (formal)
 * @child FormalDec extraFormal @after formals
 */

public class AfterReturningAdviceDec extends AdviceDec {    
    protected String getAdviceKind() { return "afterReturning"; }

    protected boolean isAfterAdvice() { return true; }   

    protected Stmts wrapStmts(Stmts stmts, AdvicePlan plan) {
        //System.out.println("wrapping: " + plan.getJoinPoint());
        //stmts.display(0);
        ReturnFixerWalker walker = new ReturnFixerWalker(getCompiler(), plan);
        Stmts ret = walker.removeReturns(stmts);
        //ret.display(4);
        return ret;
    }
    
    //XXX don't generate advice when normal return is impossible
    static class ReturnFixerWalker extends Walker {
        AdvicePlan plan;

        public ReturnFixerWalker(JavaCompiler compiler, AdvicePlan plan) {
            super(compiler);
            this.plan = plan;
        }
        
        public Stmts removeReturns(Stmts stmts) {
            Stmts ret = (Stmts)process(stmts);
            //seenAnyReturns && !seenVoidReturn || 
            if (!plan.getJoinPoint().getResultType().isVoid()) {
                return ret;
            } else {
                final AST ast = getAST();
                Expr var = plan.getJoinPoint().makeVoidReturnExpr();

                //XXX wrap in an if (true) so we don't have to do flow analysis
                return ast.makeStmts(
                    ast.makeIf(ast.makeLiteral(true), ret.getSingleStmt()),
                    plan.makeCall(var));
            }
        }

        boolean seenVoidReturn = false;
        boolean seenAnyReturns = false;
        public Stmt wrapReturn(ReturnStmt returnStmt) {
            seenAnyReturns = true;
            if (returnStmt.getExpr() == null) seenVoidReturn = true;
            return buildReturn(returnStmt.getExpr());
        }
        
        public Stmt buildReturn(Expr returnExpr) {
            final AST ast = this.getAST();
            
            Expr returningParam ;
            Expr returnValue = null;
            VarDec retDec = null;
            if (returnExpr != null) {
                Type returnType = plan.getJoinPoint().getResultType();
                retDec = ast.makeVarDec(returnType, "_return", returnExpr);
                returningParam = ast.makeVar(retDec);
                returnValue = ast.makeVar(retDec);
            } else {
                returningParam = plan.getJoinPoint().makeVoidReturnExpr();
            }
            
            BlockStmt block = 
                getAST().makeBlock(plan.makeCall(returningParam), ast.makeReturn(returnValue));
            if (retDec != null) block.getStmts().add(0, retDec);
            return block;
        }
        
        public ASTObject process(ASTObject object) {
            // make sure we don't walk into inners
            if (object instanceof TypeDec) return object;
            if (object instanceof AnonymousMethodExpr) return object;
            
            if (object instanceof ReturnStmt) {
                return wrapReturn((ReturnStmt)object);
            }

            return super.process(object);
        }
    }    
    
    // ------------------------------
    // INTRO from FlowCheckerPass

    protected void setupFlowWalker(FlowCheckerPass w) {
        if (getExtraFormal() != null) 
            w.setVars(w.getVars().addAssigned(getExtraFormal()));
        super.setupFlowWalker(w);
    }    
    
    //BEGIN: Generated from @child and @property
    protected FormalDec extraFormal;
    public FormalDec getExtraFormal() { return extraFormal; }
    public void setExtraFormal(FormalDec _extraFormal) {
        if (_extraFormal != null) _extraFormal.setParent(this);
        extraFormal = _extraFormal;
    }
    
    public AfterReturningAdviceDec(SourceLocation location, Modifiers _modifiers, Formals _formals, FormalDec _extraFormal, TypeDs __throws, Pcd _pcd, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _pcd, _body);
        setExtraFormal(_extraFormal);
    }
    protected AfterReturningAdviceDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        AfterReturningAdviceDec ret = new AfterReturningAdviceDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (extraFormal != null) ret.setExtraFormal( (FormalDec)walker.process(extraFormal) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return formals;
        case 2: return extraFormal;
        case 3: return _throws;
        case 4: return pcd;
        case 5: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "formals";
        case 2: return "extraFormal";
        case 3: return "throws";
        case 4: return "pcd";
        case 5: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setFormals((Formals)child); return;
        case 2: setExtraFormal((FormalDec)child); return;
        case 3: setThrows((TypeDs)child); return;
        case 4: setPcd((Pcd)child); return;
        case 5: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 6;
    }
    
    public String getDefaultDisplayName() {
        return "AfterReturningAdviceDec()";
    }
    
    //END: Generated from @child and @property
}
