/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.base.*;

import java.util.*;

/**
 * @grammar after (formals) throwing (formal)
 * @child FormalDec extraFormal @after formals
 */

public class AfterThrowingAdviceDec extends AdviceDec {    
    protected String getAdviceKind() { return "afterThrowing"; }

    protected boolean isAfterAdvice() { return true; }
    
    protected Stmts wrapStmts(Stmts stmts, AdvicePlan plan) {
        return wrapAfterThrowingCall(plan, stmts);
    }
    
    Stmt makeSingleRethrow(VarDec formal, Type excType, boolean needTest) {
        final AST ast = getAST();
        
        Expr castExpr = ast.makeCast(excType, ast.makeVar(formal));

        Stmt rethrow = ast.makeThrow(castExpr);
        if (needTest) {
            return ast.makeIf(ast.makeInstanceof(ast.makeVar(formal), excType), rethrow);
        } else {
            return rethrow;
        }
    }

    Stmts makeRethrows(VarDec formal, Set exceptions) {
        Stmts stmts = getAST().makeStmts();

        exceptions = Type.filterSubTypes(formal.getType(), exceptions);
        exceptions = Type.filterTopTypes(exceptions);

        if (exceptions.size() == 1) {
            Type checkType = (Type)exceptions.iterator().next();
            stmts.add(makeSingleRethrow(formal, checkType, false));
            return stmts;
        }

        for(Iterator iter = exceptions.iterator(); iter.hasNext(); ) {
            Type checkType = (Type)iter.next();
            stmts.add(makeSingleRethrow(formal, checkType, iter.hasNext()));
        }
        return stmts;
    }

    private static int excIndex = 0;
    private CatchClause makeCatchClause(Type catchType, AdvicePlan plan, int index, Set exceptions) {
        final AST ast = getAST();
        
        String newName = "ajcexc$" + excIndex++;
        FormalDec newFormal = ast.makeFormal(catchType, newName);
        Stmt doCall = plan.makeCall(ast.makeVar(newFormal));

        Stmts rethrows = makeRethrows(newFormal, exceptions);

        rethrows.add(0, doCall);
        return ast.makeCatch(newFormal, ast.makeBlock(rethrows));
    }

    boolean isUncheckedException(Type baseType) {
        return baseType.isSubtypeOf(getTypeManager().getRuntimeExceptionType())
          || baseType.isSubtypeOf(getTypeManager().getErrorType());
    }


    boolean canThrow(Set types, Type baseType) {
//        if (isUncheckedException(baseType)) {
//            return true;
//        }

        for(Iterator iter = types.iterator(); iter.hasNext(); ) {
            Type checkType = (Type)iter.next();
            if (checkType != null && checkType.isSubtypeOf(baseType) ||
                baseType.isSubtypeOf(checkType)) {
                return true;
            }
        }
        return false;
    }

    public Stmts wrapAfterThrowingCall(AdvicePlan plan, Stmts baseStmts) {
        FormalDec formal = ((AfterThrowingAdviceDec)plan.getAdviceDec()).getExtraFormal();

        Set exceptions = ExceptionFinder.getPossibleExceptions(baseStmts, true);
        exceptions.add(getTypeManager().getRuntimeExceptionType());
        exceptions.add(getTypeManager().getErrorType());
        //System.out.println("body might throw: "+exceptions);
        
        Type catchType;            
        if (formal != null) {
            catchType = formal.getType();
        } else {
            catchType = getTypeManager().getThrowableType();
        }

        if (!canThrow(exceptions, catchType)) {
            //System.out.println("can't throw: " + formal.getType());
            return baseStmts;
        }

        int index = 0; //XXX this doesn't quite work...
        CatchClause catches = makeCatchClause(catchType, plan, index, exceptions);
        
        final AST ast = getAST();
        return ast.makeStmts(ast.makeTryCatch(ast.makeBlock(baseStmts), catches));
    }    
    
    // ------------------------------
    // INTRO from FlowCheckerPass

    protected void setupFlowWalker(FlowCheckerPass w) {
        if (getExtraFormal() != null) 
            w.setVars(w.getVars().addAssigned(getExtraFormal()));
        super.setupFlowWalker(w);
    }    
    
    //BEGIN: Generated from @child and @property
    protected FormalDec extraFormal;
    public FormalDec getExtraFormal() { return extraFormal; }
    public void setExtraFormal(FormalDec _extraFormal) {
        if (_extraFormal != null) _extraFormal.setParent(this);
        extraFormal = _extraFormal;
    }
    
    public AfterThrowingAdviceDec(SourceLocation location, Modifiers _modifiers, Formals _formals, FormalDec _extraFormal, TypeDs __throws, Pcd _pcd, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _pcd, _body);
        setExtraFormal(_extraFormal);
    }
    protected AfterThrowingAdviceDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        AfterThrowingAdviceDec ret = new AfterThrowingAdviceDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (extraFormal != null) ret.setExtraFormal( (FormalDec)walker.process(extraFormal) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return formals;
        case 2: return extraFormal;
        case 3: return _throws;
        case 4: return pcd;
        case 5: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "formals";
        case 2: return "extraFormal";
        case 3: return "throws";
        case 4: return "pcd";
        case 5: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setFormals((Formals)child); return;
        case 2: setExtraFormal((FormalDec)child); return;
        case 3: setThrows((TypeDs)child); return;
        case 4: setPcd((Pcd)child); return;
        case 5: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 6;
    }
    
    public String getDefaultDisplayName() {
        return "AfterThrowingAdviceDec()";
    }
    
    //END: Generated from @child and @property
}
