/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;
import org.aspectj.compiler.base.cst.*;

import java.util.*;

/**
 * @grammar ...
 * @child GenTypeName dominates @after superInterfaces
 * @child PerClause perClause @after dominates
 */
public class AspectDec extends ClassDec {  
    public String getKind() { return "aspect"; }        
      
    public Type getSuperClassType() {
        //??? better error handling here
        return super.getSuperClassType();
    }
    
    public PerClause findPerClause() { return getPerClause(); }

    public AspectDec getParentAspect() {
        Type superType = getSuperClassType();
        if (superType.isAspect()) return (AspectDec)((NameType)superType).getTypeDec();
        return null;
    }

    public void checkSpec() {
        super.checkSpec();
        if (getType().isSubtypeOf(getTypeManager().getSerializableType()) &&
            !getOptions().XserializableAspects) {
            showError("aspects may not implement java.io.Serializable");
        }
        if (getType().isSubtypeOf(getTypeManager().getCloneableType())) {
            showError("aspects may not implement java.lang.Cloneable");
        }

        if (isInnerType() && !isStatic()) {
            showError("inner aspects must be static");
        }
    }


    protected boolean explicitDominates(AspectDec otherAspect) {
        if (dominates == null) {
            if (getParentAspect() != null) {
                return getParentAspect().explicitDominates(otherAspect);
            } else {
                return false;
            }
        }
        return dominates.matches(otherAspect.getType());
    }

    public boolean dominates(AspectDec otherAspect) {
        if (otherAspect == this) return false;
        
        if (explicitDominates(otherAspect)) return true;
        if (otherAspect.explicitDominates(this)) return false;

        if (getType().isStrictSubtypeOf(otherAspect.getType())) return true;
        //??? might want to do some extra work here for explict dominates specs
        return false;
    }

    public boolean hasAdvicesOrIntroductions() {
        return true;
    }

    protected void maybeShowWarning(Dec dec) { }

    public List getIntroductions() {
        return introductions;
    }
    
    List introductions = new LinkedList();
    public void addDec(Dec dec) {
        if (dec instanceof AdviceDec) {
            return;
        } else if (dec instanceof IntroducedDec) {
            introductions.add(dec);
        } else if (dec instanceof ShowErrorDec) {
            return;
        } else if (dec instanceof SoftThrowableDec) {
            return;
        } else if (dec instanceof IntroducedSuperDec) {
            introductions.add(dec);
        } else {
            super.addDec(dec);
        }
    }
    

    public void postIntroductionFinish() {   
        super.postIntroductionFinish();
                     
        if (perClause == null) {
            if (getParentAspect() == null) {
                setPerClause(new PerSingleton(getSourceLocation()));
            } else {
                setPerClause((PerClause)getParentAspect().getPerClause().copy());
            }                
        }
        
        if (getParentAspect() != null && !getParentAspect().isAbstract()) {
            superClass.showError("can only extend abstract aspects, not "
                + getParentAspect().getType().getString());
            return;
        }
       
        
        if (!isAbstract()) {
            getPerClause().setupAspect();
        }
    }

    private Map extraPlanners = new HashMap();
    public Map getExtraPlanners() { return extraPlanners; }

    protected void walkExtendsAndImplements(ScopeWalker walker) {
        //XXX extremely odd location for this
        if (perClause instanceof PerThisOrTarget) {
            getModifiers().setPublic(true);
        }

        if (dominates != null) walker.process(dominates);
        super.walkExtendsAndImplements(walker);
    }

    protected void walkBody(ScopeWalker walker) {
        if (perClause != null) walker.process(perClause);
        super.walkBody(walker);
    }
    
    public void addConstructorDec(ConstructorDec constructorDec) {
        if (constructorDec.getFormals().size() != 0) {
            constructorDec.showError("only 0-argument constructors allowed in an aspect");
            return;
        }
        super.addConstructorDec(constructorDec);
    }
    



    protected List myJpPlannerMakers = null;
    public List getJpPlannerMakers() {
        if (myJpPlannerMakers != null) return myJpPlannerMakers;

        List ret = new LinkedList();
        final int N = getBody().size();
        for (int i = 0; i < N; i++) {
            Dec dec = getBody().get(i);
            if (dec instanceof JpPlannerMaker) {
                ret.add(dec);
            }
        }

        AspectDec parentAspect = getParentAspect();
        if (parentAspect != null) {
            if (!parentAspect.isAbstract()) {
                showError("only abstract aspects can be extended, " + parentAspect.toShortString() + " is not abstract");
            }

            ret.addAll(parentAspect.getJpPlannerMakers());
        }

        myJpPlannerMakers = ret;
        return ret;
    }    
    

    /*
    protected List myAdvice = null;
    public List getAdvice() {
        if (myAdvice != null) return myAdvice;

//        if (isAbstract() && ofClause != null) {
//            showError("abstract aspects can not have of clauses");
//        }

        if (isInnerType() && !isStatic()) {
            showError("inner aspects must be static");
        }

        List ret = new LinkedList();
        final int N = getBody().size();
        for (int i = 0; i < N; i++) {
            Dec dec = getBody().get(i);
            if (dec instanceof AdviceDec) {
                ret.add(dec);
            }
        }

        //??? add in non-copies of all advice from my parent
        AspectDec parentAspect = getParentAspect();
        if (parentAspect != null) {
            if (!parentAspect.isAbstract()) {
                showError("only abstract aspects can be extended, " + parentAspect.toShortString() + " is not abstract");
            }
            
            List extraAdvice = parentAspect.getAdvice();
            for(Iterator i = extraAdvice.iterator(); i.hasNext(); ) {
                AdviceDec adviceDec = (AdviceDec)i.next(); //).copy();

                //this will go away when we remove deprecated support for static advice
                if (adviceDec.isStatic()) continue;

                //adviceDec.buildContext(getBody().getContext());
                ret.add(adviceDec);
            }
        }

        myAdvice = ret;
        return ret;
    }
    */
    
    // ------------------------------
    // Inner class stuff

    // AAARGH!  This shouldn't be a subtype of class dec. 

    public boolean isInner() { return false; }

    //BEGIN: Generated from @child and @property
    protected GenTypeName dominates;
    public GenTypeName getDominates() { return dominates; }
    public void setDominates(GenTypeName _dominates) {
        if (_dominates != null) _dominates.setParent(this);
        dominates = _dominates;
    }
    
    protected PerClause perClause;
    public PerClause getPerClause() { return perClause; }
    public void setPerClause(PerClause _perClause) {
        if (_perClause != null) _perClause.setParent(this);
        perClause = _perClause;
    }
    
    public AspectDec(SourceLocation location, Modifiers _modifiers, String _id, TypeD _superClass, TypeDs _superInterfaces, GenTypeName _dominates, PerClause _perClause, Decs _body) {
        super(location, _modifiers, _id, _superClass, _superInterfaces, _body);
        setDominates(_dominates);
        setPerClause(_perClause);
    }
    protected AspectDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        AspectDec ret = new AspectDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        ret.id = id;
        if (superClass != null) ret.setSuperClass( (TypeD)walker.process(superClass) );
        if (superInterfaces != null) ret.setSuperInterfaces( (TypeDs)walker.process(superInterfaces) );
        if (dominates != null) ret.setDominates( (GenTypeName)walker.process(dominates) );
        if (perClause != null) ret.setPerClause( (PerClause)walker.process(perClause) );
        if (body != null) ret.setBody( (Decs)walker.process(body) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return superClass;
        case 2: return superInterfaces;
        case 3: return dominates;
        case 4: return perClause;
        case 5: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "superClass";
        case 2: return "superInterfaces";
        case 3: return "dominates";
        case 4: return "perClause";
        case 5: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setSuperClass((TypeD)child); return;
        case 2: setSuperInterfaces((TypeDs)child); return;
        case 3: setDominates((GenTypeName)child); return;
        case 4: setPerClause((PerClause)child); return;
        case 5: setBody((Decs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 6;
    }
    
    public String getDefaultDisplayName() {
        return "AspectDec(id: "+id+")";
    }
    
    //END: Generated from @child and @property
}

