/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;


import java.util.*;

import org.aspectj.util.*;

/**
 * @grammar cflow(pcd)
 * @child Pcd pcd
 * @property boolean includesRoot
 */

public class CFlowPcd extends Pcd {
    public String toShortString() {
        return "cflow" + (includesRoot ? CFlowPlanner.NO_ROOT_NAME : "")
                  +  "(" + getPcd().toShortString() + ")";
    }
    
    public void checkStatic() { showNonStaticError(); }    
    
    public JpPlanner makePlanner(PlanData planData) {
        findFlowState();
        Map extraPlanners = planData.inAspect.getExtraPlanners();
        if (!extraPlanners.containsKey(this)) {
            extraPlanners.put(this,
                              makeEntryPointInitializer(planData));
        }

        final CflowField field = getCflowField(planData.inAspect);
        return new JpPlanner() {
            public FastMatch fastMatch(JoinPoint jp) {
                return MAYBE;
            }
            public JpPlan makePlan(JoinPoint jp) {
                JpPlan plan = new JpPlan(jp);
                plan.test = field.testStackExpr();
                plan.addDependency(field);
                field.remapExprs(plan);
                return plan;
            }
        };
    }
    
    private JpPlanner makeEntryPointInitializer(PlanData planData) {
        findFlowState();

        final JpPlanner planner = getPcd().makePlanner(planData);

        final CflowField field = getCflowField(planData.inAspect);
        return new WrappedJpPlanner(planner) {
            public JpPlan makePlan(JoinPoint jp) {
                JpPlan plan = super.makePlan(jp);
                if (!plan.isPossible()) return plan;
                return new CFlowEntryPlan(plan, field); //CFlowPcd.this);
            }
        };
    }
    
    //INTRO from ScopeWalker
    public ASTObject postScope(ScopeWalker walker) {
        if (walker.walkBodies()) {
            findFlowState();
            //getStackField();
        }
        return this;
    }

    List flowState = null;

    void findFlowState() {
        if (flowState != null) return;

        flowState = new ArrayList();

        new Walker(getCompiler()) {
            public void postProcess(ASTObject node) {
                if (node instanceof VarTypeName) {
                    VarTypeName varTypeName = (VarTypeName)node;
                    if (varTypeName.getFormalDec() != null) {
                        addFlowState(varTypeName);
                    }
                }
            }
        }.process(getPcd());
    }

    void addFlowState(VarTypeName varTypeName) {
        flowState.add(varTypeName);
    }

    Map cflowFields = new HashMap();
    
    public CflowField getCflowField(TypeDec inTypeDec) {
        CflowField ret = (CflowField)cflowFields.get(inTypeDec);
        if (ret == null) {
            ret = new CflowField(inTypeDec);
            cflowFields.put(inTypeDec, ret);
        }
        return ret;
    }

    private static int counter=0;
    private static int ecounter = 0;

    class CflowField implements CFlowPlanner {
        private TypeDec declaringTypeDec = null;
        private FieldDec stack;

        public CflowField(TypeDec inTypeDec) {
            this.declaringTypeDec = inTypeDec;
        }

        public boolean getIncludesRoot() {
            return CFlowPcd.this.getIncludesRoot();
        }                                    

        FieldDec getStackField() {
            if (stack != null) return stack;
            
            final AST ast = getAST();
            String name = "cflow$ajc" + counter++;
            Type type = getTypeManager().getType("org.aspectj.runtime.internal", "CFlowStack");
            
            Modifiers modifiers =
                ast.makeModifiers(Modifiers.PUBLIC | Modifiers.STATIC | Modifiers.FINAL);
            stack = ast.makeField(modifiers, type, name, ast.makeNew(type));
            //stack.setSynthetic();
            
            //System.out.println(typeDec +", " + typeDec.getBody());
            declaringTypeDec.getBody().add(0, stack);
            return stack;
        }


        Expr getStackVar() {
            return getAST().makeGet(getStackField());
        }

        Expr testStackExpr() {
            return getAST().makeCall(getStackVar(), "isValid");
        }

        Expr pushStackExpr(Expr expr) {
            return getAST().makeCall(getStackVar(), "push", expr);
        }
        
        Expr popStackExpr() {
            return getAST().makeCall(getStackVar(), "pop");
        }

        Expr peekStackExpr() {
            return getAST().makeCall(getStackVar(), "peekCFlow");
        }

        void remapExprs(JpPlan plan) {
            int index = 0;
            for (Iterator i = flowState.iterator(); i.hasNext(); ) {
                VarTypeName varTypeName = (VarTypeName)i.next();
                FormalDec formal = varTypeName.getFormalDec();
                
                Expr expr = makeFlowStateExpr(formal.getType(), index++);
                plan.bindExpr(varTypeName.getFormalDec(), expr);
            }
        }

        // this is called by the CFlowAdviceDec for the "entry points"
        public Stmts wrapPushPop(JpPlan plan, Stmts body) {
            final AST ast = getAST();
        
            Expr newCFlowExpr = makeNewCFlowExpr(plan);

            Expr test = plan.getDynamicTest();
            if (test == null) {
                return ast.makeStmts(
                                     ast.makeStmt(pushStackExpr(newCFlowExpr)),
                                     ast.makeTryFinally(ast.makeBlock(body), 
                                                        ast.makeBlock(popStackExpr())));
            } else {
                VarDec tmp = ast.makeFinalVar(getTypeManager().booleanType,
                                              "ajc_enter_cflow_" + ecounter++, test);
                return ast.makeStmts(
                                     tmp,
                                     ast.makeIf(ast.makeVar(tmp), 
                                                ast.makeStmt(pushStackExpr(newCFlowExpr))),
                                     ast.makeTryFinally(ast.makeBlock(body),
                                                        ast.makeBlock(ast.makeIf(ast.makeVar(tmp), 
                                                                                 ast.makeStmt(popStackExpr())))));
            }
        }

        Exprs initializeFlowState(JpPlan plan) {
            final AST ast = getAST();

            int index = 0;
            //List planExprs = plan.exprs;
            Exprs stateExprs = ast.makeExprs();

            for (Iterator i = flowState.iterator(); i.hasNext(); ) {
                VarTypeName varTypeName = (VarTypeName)i.next();
                FormalDec formal = varTypeName.getFormalDec();
                Expr expr = (Expr)plan.bindings.get(formal);

                //Expr expr = (Expr)planExprs.get(formal.getIndex());
                expr = expr.getType().makeObject(expr);
                stateExprs.add(expr);
            }

            Exprs ret = ast.makeExprs();
            if (stateExprs.size() != 0) {
                ret.add(ast.makeObjectArray(stateExprs));
            } else {
                return null;
            }
            return ret;
        }

        Expr makeFlowStateExpr(Type desiredType, int index) {
            Expr objectExpr =
                getAST().makeCall(peekStackExpr(), "get", getAST().makeLiteral(index));
            return desiredType.fromObject(objectExpr);
        }

        public Expr makeNewCFlow(Exprs baseExprs) {
            Type cflowType;
            if (baseExprs == null) {
                baseExprs = getAST().makeExprs();
                cflowType = getTypeManager().getType("org.aspectj.runtime", "CFlow");
            } else {
                cflowType = getTypeManager().getType("org.aspectj.runtime.internal", "CFlowPlusState");
            }

            return getAST().makeNew(cflowType, baseExprs);
        }

        public Expr makeNewCFlowExpr(JpPlan plan) {
            return makeNewCFlow(initializeFlowState(plan));
        }



    }



    
    //-----------------------------------------------------------------------------------    
    
    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    protected boolean includesRoot;
    public boolean getIncludesRoot() { return includesRoot; }
    public void setIncludesRoot(boolean _includesRoot) { includesRoot = _includesRoot; }
    
    public CFlowPcd(SourceLocation location, Pcd _pcd, boolean _includesRoot) {
        super(location);
        setPcd(_pcd);
        setIncludesRoot(_includesRoot);
    }
    protected CFlowPcd(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        CFlowPcd ret = new CFlowPcd(getSourceLocation());
        ret.preCopy(walker, this);
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        ret.includesRoot = includesRoot;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "CFlowPcd(includesRoot: "+includesRoot+")";
    }
    
    //END: Generated from @child and @property
}

