/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.crosscuts.joinpoints.TypeDecPlanner;
import org.aspectj.compiler.crosscuts.*;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.bcg.ClassfileBuilder;

import java.util.*;

/**
 * @grammar introduction ...
 * @child GenTypeName targets
 * @property Dec dec
 */

public class IntroducedDec extends Dec implements TypeDecPlanner {
    private List introducedDecs = new LinkedList();


    public Modifiers getModifiers() { return dec.getModifiers(); }

    Scope myLexicalScope = null;
    public void walkScope(ScopeWalker walker) {
        walker.process(targets);
        boolean saveWalkBodies = walker.walkBodies();
        walker.setWalkBodies(false);
        walker.process(dec);
        walker.setWalkBodies(saveWalkBodies);


        myLexicalScope = walker.getScope();
        dec.setParent(this);
        getWorld().addTypeDecPlanner(this);
    }


    /*
    //INTRO from IntroductionPlannerPass
    public void preIntroduction(IntroductionPlannerPass pass) {
        ScopeWalker.bindNames(targets, getBytecodeTypeDec().getTypeScope());
        dec.setParent(this);
        getWorld().addTypeDecPlanner(this);
    }
    */

//    public Dec getSavedDec() { return savedDec; }
//
//    public Dec savedDec = null;

    public boolean isStatic() { return true; }

    public String getUniqueIdString(Type onType) {
        String ret = null;

        if (isPrivate()) {
            ret = getLexicalType().getPrivateCookieType().getPrettyString();
        } else if (isProtected()) {
            showError("protected introduction not allowed");
            ret = "";
        } else if (isPublic()) {
             ret = "";
        } else { // package protected
            ret = getLexicalType().getPackageName();
        }

        if (onType.isInterface() && dec instanceof FieldDec) {
            ret += "_" + onType.getPrettyString();
        }

        return ret;
    }

    public void plan(TypeDec typeDec, int phase) {
        if (phase != TypeDecPlanner.SIGNATURE) return;

        Type toType = typeDec.getType();

        if (!getTargets().matches(toType)) return;

        //CopyWalker cw = CopyWalker.rawCopy(object);
        //new CopyWalker(getCompiler(), toType, toType);

        Dec newDec = (Dec)CopyWalker.rawCopy(dec, toType);
        newDec.setSourceLocation(getAST().getSourceLocation());
        newDec.setLanguageVisible();
        
        // public introductions need to not generate synthetic members
        // so they will be visible to other compilers
        if (isPublic()) {
        	newDec.setExplicitlyNonSynthetic();
        }
        newDec.setDeclaringType(toType);
        newDec.setLexicalType(getLexicalType());
        newDec.setFromLexicalScope(myLexicalScope);


        if (!isPublic() || (newDec instanceof FieldDec && typeDec instanceof InterfaceDec)) {
            String newId = getAST().makeGeneratedName(newDec.getBytecodeId() + '_' +
                                                      getUniqueIdString(typeDec.getType()));
            newDec.setBytecodeId(newId);
        }

        // add this to the typeDec's signature and body
        typeDec.addIntroducedDec(newDec);
        typeDec.addExtraWithinType(this.getLexicalType());

        ((AspectJCompiler)getCompiler()).getCorrespondences().addPointsTo(this, typeDec);
        getCompiler().showMessage("    introduced " + newDec.toShortString() + " on " +
                typeDec.getType().toShortString());
    }

    public String getKind() {
        return "introduction";
    }

    public String getId() { return dec.getId(); }

    public String toShortString() {
        return targets.toShortString()+"."+dec.getId();
    }

    // ------------------------------
    // bcg

    // I don't know why these are still around, but it doesn't seem to
    // be a problem if we don't generate code for them.
    protected void cgMember(ClassfileBuilder maker) {}

    //BEGIN: Generated from @child and @property
    protected GenTypeName targets;
    public GenTypeName getTargets() { return targets; }
    public void setTargets(GenTypeName _targets) {
        if (_targets != null) _targets.setParent(this);
        targets = _targets;
    }

    protected Dec dec;
    public Dec getDec() { return dec; }
    public void setDec(Dec _dec) { dec = _dec; }

    public IntroducedDec(SourceLocation location, GenTypeName _targets, Dec _dec) {
        super(location);
        setTargets(_targets);
        setDec(_dec);
    }
    protected IntroducedDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        IntroducedDec ret = new IntroducedDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (targets != null) ret.setTargets( (GenTypeName)walker.process(targets) );
        ret.dec = dec;
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return targets;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "targets";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTargets((GenTypeName)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }

    public String getDefaultDisplayName() {
        return "IntroducedDec(dec: "+dec+")";
    }

    //END: Generated from @child and @property
}



