/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;
import org.aspectj.util.InvertableSet;

/**
  * @grammar modifiers returnTypeName [declaringTypeName].id(formals) throws
  * @child GenTypeName returnTypeName
  * @property NamePattern id
  * @child FormalsPattern formalsPattern
  * @child NameTypeDsPattern throwsPattern
  */
public class MethodPattern extends CodePattern {
    public String toShortString() {
        //XXX this ignores throws for now
        String mods = modifiers.toShortString();
        if (mods.length() > 0) mods += " ";
        mods += returnTypeName.toShortString()+" ";
        String ret = mods + makeDeclaringTypeString() + id.toShortString() + formalsPattern.toShortString();

        if (throwsPattern != null) {
            ret += " throws ...";
        }

        return ret;
    }

    public void checkSpec() {
        if (id.getSimpleName() != null && id.getSimpleName().equals("new")) {
            this.showError("constructor patterns don't have a return type");
        }
    }

    public String getLookupId() { return id.getLookupString(); }

    public JpPlan makePlan(JoinPoint joinPoint) {
        if (! (joinPoint.getTargetSO() instanceof Method) ) return JpPlan.NO_PLAN;
        
        Method m = (Method)joinPoint.getTargetSO();
        
        if (!id.matches(m.getName())) return JpPlan.NO_PLAN;
        if (!returnTypeName.matches(m.getReturnType())) return JpPlan.NO_PLAN;
        if (!formalsPattern.matches(m.getFormals())) return JpPlan.NO_PLAN;
        if (throwsPattern != null && !throwsPattern.matches(m.getThrows())) return JpPlan.NO_PLAN;
        
        return super.makePlan(joinPoint);
    }    
    
    //BEGIN: Generated from @child and @property
    protected GenTypeName returnTypeName;
    public GenTypeName getReturnTypeName() { return returnTypeName; }
    public void setReturnTypeName(GenTypeName _returnTypeName) {
        if (_returnTypeName != null) _returnTypeName.setParent(this);
        returnTypeName = _returnTypeName;
    }
    
    protected NamePattern id;
    public NamePattern getId() { return id; }
    public void setId(NamePattern _id) { id = _id; }
    
    protected FormalsPattern formalsPattern;
    public FormalsPattern getFormalsPattern() { return formalsPattern; }
    public void setFormalsPattern(FormalsPattern _formalsPattern) {
        if (_formalsPattern != null) _formalsPattern.setParent(this);
        formalsPattern = _formalsPattern;
    }
    
    protected NameTypeDsPattern throwsPattern;
    public NameTypeDsPattern getThrowsPattern() { return throwsPattern; }
    public void setThrowsPattern(NameTypeDsPattern _throwsPattern) {
        if (_throwsPattern != null) _throwsPattern.setParent(this);
        throwsPattern = _throwsPattern;
    }
    
    public MethodPattern(SourceLocation location, Modifiers _modifiers, GenTypeName _declaringTypeName, GenTypeName _returnTypeName, NamePattern _id, FormalsPattern _formalsPattern, NameTypeDsPattern _throwsPattern) {
        super(location, _modifiers, _declaringTypeName);
        setReturnTypeName(_returnTypeName);
        setId(_id);
        setFormalsPattern(_formalsPattern);
        setThrowsPattern(_throwsPattern);
    }
    protected MethodPattern(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        MethodPattern ret = new MethodPattern(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (declaringTypeName != null) ret.setDeclaringTypeName( (GenTypeName)walker.process(declaringTypeName) );
        if (returnTypeName != null) ret.setReturnTypeName( (GenTypeName)walker.process(returnTypeName) );
        ret.id = id;
        if (formalsPattern != null) ret.setFormalsPattern( (FormalsPattern)walker.process(formalsPattern) );
        if (throwsPattern != null) ret.setThrowsPattern( (NameTypeDsPattern)walker.process(throwsPattern) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 2: return returnTypeName;
        case 3: return formalsPattern;
        case 4: return throwsPattern;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 2: return "returnTypeName";
        case 3: return "formalsPattern";
        case 4: return "throwsPattern";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 2: setReturnTypeName((GenTypeName)child); return;
        case 3: setFormalsPattern((FormalsPattern)child); return;
        case 4: setThrowsPattern((NameTypeDsPattern)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 5;
    }
    
    public String getDefaultDisplayName() {
        return "MethodPattern(id: "+id+")";
    }
    
    //END: Generated from @child and @property
}

