/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;


import java.util.*;

import org.aspectj.util.*;

/**
 * @grammar percflow(pcd)
 * @child Pcd pcd
 * @property boolean includesRoot
 */

public class PerCFlow extends PerClause implements CFlowPlanner {
    public String toShortString() {
        return "percflow" + (includesRoot ? NO_ROOT_NAME : "") 
                + "(" + getPcd().toShortString() + ")";
    }
    
    public JpPlanner makeInnerPlanner(PlanData planData) {
        return new JpPlanner() {
            public FastMatch fastMatch(JoinPoint jp) {
                //??? might want to include a better filter test here as an optimization
                return MAYBE;
            }
            public JpPlan makePlan(JoinPoint jp) {                
                JpPlan plan = new JpPlan(jp);
                plan.test = testStackExpr();
                plan.setInstanceExpr(makeAspectInstance());
                plan.addDependency(this);
                return plan;
            }
        };
    }
    
    public JpPlanner makeInitializerPlanner(PlanData planData) {
        return new WrappedJpPlanner(getPcd().makePlanner(planData)) {
                public JpPlan makePlan(JoinPoint jp) {
                    JpPlan plan = super.makePlan(jp);
                    return new CFlowEntryPlan(plan, PerCFlow.this);
                }
            };
    }

    private FieldDec stack;
    FieldDec getStackField() {
        if (stack != null) return stack;

        final AST ast = getAST();
        String name = "eachcflow$ajc";
        Type type = getTypeManager().getType("org.aspectj.runtime.internal", "CFlowStack");

        Modifiers modifiers =
            ast.makeModifiers(Modifiers.PUBLIC | Modifiers.STATIC | Modifiers.FINAL);
        stack = ast.makeField(modifiers, type, name, ast.makeNew(type));
        //stack.setSynthetic();
        this.getAspectDec().getBody().add(0, stack);

        return stack;
    }

    Expr getStackVar() {
        return getAST().makeGet(getStackField());
    }

    Expr testStackExpr() {
        return getAST().makeCall(getStackVar(), "isValid");
    }

    Expr pushStackExpr() {
        return pushStackExpr(getAST().makeNew(getTypeManager().getObjectType()));
    }
    Expr pushStackExpr(Expr expr) {
        return getAST().makeCall(getStackVar(), "push", expr);
    }

    Expr popStackExpr() {
        return getAST().makeCall(getStackVar(), "pop");
    }

    Expr peekStackExpr() {
        return getAST().makeCall(getStackVar(), "peek");
        //return new CallExpr(getCompiler(),getStackVar(), "peek");
    }

    // this is called by the CFlowAdviceDec for the "entry points"
    private static int ecounter = 0;
    public Stmts wrapPushPop(JpPlan plan, Stmts body) {
        final AST ast = getAST();
        
        Expr test = plan.getDynamicTest();
        if (test == null) {
            return ast.makeStmts(
                  ast.makeStmt(pushStackExpr(makeNewAspectExpr())),
                  ast.makeTryFinally(ast.makeBlock(body), 
                                     ast.makeBlock(popStackExpr())));
        } else {
            VarDec tmp = ast.makeFinalVar(getTypeManager().booleanType,
                    "ajc_enter_ecflow_" + ecounter++, test);
            return ast.makeStmts(
                tmp,
                ast.makeIf(ast.makeVar(tmp), 
                           ast.makeStmt(pushStackExpr(makeNewAspectExpr()))),
                 ast.makeTryFinally(ast.makeBlock(body),
                       ast.makeBlock(ast.makeIf(ast.makeVar(tmp), 
                                                ast.makeStmt(popStackExpr())))));
        }
    }
    

    //-----------------------------------------------------------------------------------

    Expr makeAspectInstance() {
        return getAST().makeCast(getAspectType(), peekStackExpr());
    }

    public Expr makeAspectOfExpr(Expr fromObject) { return makeAspectInstance(); }



    Expr makeNewAspectExpr() {
        return getAST().makeNew(getAspectType());
    }

    protected MethodDec makeHasAspectMethod() {
        /* public static boolean hasAspect() {
               return stackVar.isValid();
           }
        */
        final AST ast = getAST();
        BlockStmt body = ast.makeBlock(ast.makeReturn(testStackExpr()));
        return makeHasAspectMethod(ast.makeFormals(), body);
    }

    protected MethodDec makeAspectOfMethod() {
        /* public static A aspectOf() {
               //if (stackVar.isValid()) {
                   Object o = stackVar.getAspect();
                   if (o != null) return o;
               //}
               throw new org.aspectj.runtime.NoAspectException();
           }
        */
        final AST ast = getAST();
        
        VarDec aspectDec = ast.makeVarDec(getAspectType(), "asp", makeAspectInstance());

        BlockStmt body = ast.makeBlock(
            aspectDec,
            ast.makeIf(ast.makeNonNullTest(ast.makeVar(aspectDec)),
                       ast.makeReturn(ast.makeVar(aspectDec))),
            ast.makeThrow(getTypeManager().getType("org.aspectj.lang", "NoAspectBoundException")));

        return makeAspectOfMethod(ast.makeFormals(), body);
    }    
    
    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    protected boolean includesRoot;
    public boolean getIncludesRoot() { return includesRoot; }
    public void setIncludesRoot(boolean _includesRoot) { includesRoot = _includesRoot; }
    
    public PerCFlow(SourceLocation location, Pcd _pcd, boolean _includesRoot) {
        super(location);
        setPcd(_pcd);
        setIncludesRoot(_includesRoot);
    }
    protected PerCFlow(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        PerCFlow ret = new PerCFlow(getSourceLocation());
        ret.preCopy(walker, this);
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        ret.includesRoot = includesRoot;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "PerCFlow(includesRoot: "+includesRoot+")";
    }
    
    //END: Generated from @child and @property
}

