/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
  * @grammar withincode(codePattern)
  * @child CodePattern codePattern
  */
public class WithinCodePcd extends Pcd {
    public String toShortString() {
        return "withincode(" + getCodePattern().toShortString() + ")";
    }
    public void checkStatic() { }

    /**
     * loop to find enclosing code decs of inner types
     */
    private CodeDec matchingEnclosingCodeDec(CodeDec codeDec) {
        while (codeDec != null) {
            //System.out.println("within? " + codeDec.toShortString() + " in " + point);
            JoinPoint jp;
            if (codeDec instanceof ConstructorDec) {
                jp = new ConstructorExecutionJp((ConstructorDec)codeDec);
            } else if (codeDec instanceof MethodDec) {
                jp = new MethodExecutionJp((MethodDec)codeDec);
            } else {
                ///???
                return null;
            }
            
            if (getCodePattern().matches(jp)) return codeDec;
            
            if (!codeDec.getDeclaringType().getTypeDec().isLocal()) break;
            codeDec = codeDec.getEnclosingCodeDec();           
        }
        
        return null;
    }
    

    public JpPlanner makePlanner(PlanData planData) {
        return new JpPlanner() {
            public FastMatch fastMatch(JoinPoint jp) {
                ASTObject sourceLocation = jp.getSourceLocation();
                if (sourceLocation == null) return NO;
                
                CodeDec codeDec;
                if (sourceLocation instanceof CodeDec) {
                    codeDec = (CodeDec)sourceLocation;
                } else {
                    codeDec = sourceLocation.getEnclosingCodeDec();
                }
                if (codeDec == null) return NO;
                
                codeDec = matchingEnclosingCodeDec(codeDec);
                if (codeDec == null) return NO;
                else return YES;                
            }
            public JpPlan makePlan(JoinPoint jp) {
                if (jp.getSourceLocation() == null) return JpPlan.NEVER_PLAN;
                return super.makePlan(jp);
            }
        };
    }    
    
    //BEGIN: Generated from @child and @property
    protected CodePattern codePattern;
    public CodePattern getCodePattern() { return codePattern; }
    public void setCodePattern(CodePattern _codePattern) {
        if (_codePattern != null) _codePattern.setParent(this);
        codePattern = _codePattern;
    }
    
    public WithinCodePcd(SourceLocation location, CodePattern _codePattern) {
        super(location);
        setCodePattern(_codePattern);
    }
    protected WithinCodePcd(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        WithinCodePcd ret = new WithinCodePcd(getSourceLocation());
        ret.preCopy(walker, this);
        if (codePattern != null) ret.setCodePattern( (CodePattern)walker.process(codePattern) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return codePattern;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "codePattern";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setCodePattern((CodePattern)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "WithinCodePcd()";
    }
    
    //END: Generated from @child and @property
}

