/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
 
package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.util.JavaStrings;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;


public class InitializationJp extends CodeDecJp {
    public InitializationJp(ConstructorDec constructorDec) {
        super(constructorDec);
        constructorDec.getBytecodeTypeDec().getInitializerExecutionJoinPoints().add(this);
    }
    
    public static final Kind KIND = new Kind("initialization", INITIALIZER_EXECUTION);

    public Kind getKind() { return KIND; }
 
    //!!! bad name
    public TypeDec getEnclosingTypeDec() { return getCodeDec().getBytecodeTypeDec(); }

    public ASTObject getSourceLocation() { return getEnclosingTypeDec().getBody(); }  

    public void setStmts(Stmts stmts) {
        makeInnerAndOuterConstructorDecs();
        innerConstructorDec.getBody().setStmts(stmts);
    }

    public Stmts getStmts() {
        makeInnerAndOuterConstructorDecs();        
        return innerConstructorDec.getBody().getStmts();
    }
    
    private ConstructorDec innerConstructorDec = null;
    private ConstructorDec outerConstructorDec = null;
    
    public void inlineBodies() {
        ConstructorDec origConstructorDec = (ConstructorDec)getCodeDec();
        outerConstructorDec = makeInitializer(origConstructorDec);
    }
    
    
    public void finishInline() {
        ConstructorDec origConstructorDec = (ConstructorDec)getCodeDec();
        ConstructorDec topDec = outerConstructorDec;

        origConstructorDec.setBody(topDec.getBody());          
        if (instanceInitializer != null && instanceInitializer.getBody().getStmts().size() > 0) {
            Stmt initializerStmt =
                getAST().makeBlock(instanceInitializer.getBody().getStmts().copy());
            innerConstructorDec.getBody().getStmts().add(0, initializerStmt);
        }
        
        // move formals to innerConstructorDec and move references in there as well
        // this will deal better with advice that holds on to these references
        Formals oldFormals = origConstructorDec.getFormals();
        origConstructorDec.setFormals(topDec.getFormals());
        MovingWalker mover = new MovingWalker(getCompiler());
        for (int i = 0; i<oldFormals.size(); i++) {
            FormalDec oldFormal = oldFormals.get(i);
            FormalDec newFormal = innerConstructorDec.getFormals().get(i);
            mover.addMapping(newFormal, oldFormal);
            newFormal.replaceWith(oldFormal);
        }
        mover.process(innerConstructorDec.getBody());
    }

    private InitializerDec instanceInitializer;
    public void makeInnerAndOuterConstructorDecs() {
        //System.out.println("making inners and outers: " + this);
        
        if (outerConstructorDec != null) return;
        
        instanceInitializer = getEnclosingTypeDec().getSingleInitializerDec(false);
        
        for (Iterator i = getEnclosingTypeDec().getInitializerExecutionJoinPoints().iterator(); i.hasNext(); ) {
            InitializationJp point = (InitializationJp)i.next();
            point.instanceInitializer = instanceInitializer;
            point.inlineBodies();
        }
        
        for (Iterator i = getEnclosingTypeDec().getInitializerExecutionJoinPoints().iterator(); i.hasNext(); ) {
            InitializationJp point = (InitializationJp)i.next();
            point.finishInline();
        }
        
        getEnclosingTypeDec().getBody().remove(instanceInitializer);
        
        //System.out.println("get: " + innerConstructorDec.getBody().unparse());
    }
    

    /**
     *  Takes one step towards making a constructor a reasonable target for
     *  a initializer join point.
     *
     * There are three cases
     * <ul>
     * <li> C(args) { super(...); ... } <p> This will return a copy of C
     * <li> C(args) { this(1, "hi"); ... } C(int, String) { ... } <p>
     *    This will inline C(int, String) in the copy of C and return that.
     * <li> C(args) { this(x+y, s.substring(...)); ... } C(int, String) { ... } <p>
     * This will produce C(args) { this(x, y, s, x+y, s.substring(...)); } and internally
     * will produce a C(int, int, String, int, String) { merged bodies }
     */
    ConstructorDec makeInitializer(ConstructorDec dec) {
        // we don't do a simple copy because that will remap references
        // to this constructor from its body
        CopyWalker cw = new CopyWalker(getCompiler());
        ConstructorDec ret =
            new ConstructorDec(dec.getSourceLocation(),
                               (Modifiers)cw.cp(dec.getModifiers()),
                               (Formals)cw.cp(dec.getFormals()),
                               (TypeDs)cw.cp(dec.getThrows()),
                               (CodeBody)cw.cp(dec.getBody()));

        ConstructorBody body = (ConstructorBody)ret.getBody();
        ConstructorCallExpr thisCallExpr = body.getConstructorCall();
        
        //ret.setFormalComment("/* from " + this.toString() + " */");
        
        if (thisCallExpr.getIsSuper()) {
            body.setStmts(body.getStmts().removeReturns());
            innerConstructorDec = ret;
            return ret;
        }
        
        ConstructorDec innerDec = 
                makeInitializer(thisCallExpr.getConstructor().getConstructorDec());

        Formals oldFormals = ret.getFormals();
        CopyWalker copier = new CopyWalker(getCompiler());
        Formals newFormals = (Formals)copier.process(oldFormals);
        thisCallExpr = (ConstructorCallExpr)copier.process(thisCallExpr);
        thisCallExpr.getArgs().addAll(0, newFormals.makeExprs());
        thisCallExpr.setConstructor(innerDec.getConstructor());
        body.setConstructorCall(thisCallExpr);
        ret.setFormals(newFormals);
        Stmts stmts = body.getStmts();
        body.setStmts(getAST().makeStmts());
        inlineStmts(stmts, innerDec.getBody().getStmts());
        innerDec.getFormals().addAll(0, oldFormals);
        
        addUniquelyToBody(dec.getBytecodeTypeDec().getBody(), innerDec, thisCallExpr);
        
        return ret;
    }
    
    void inlineStmts(Stmts addStmts, Stmts baseStmts) {
        baseStmts.addAll(addStmts.removeReturns());
    }
    
    void addUniquelyToBody(Decs body, ConstructorDec dec, ConstructorCallExpr thisCallExpr) {
        final AST ast = getAST();
        
        body.add(dec);
        FormalDec cookieFormal = null;
        while (true) {
            int i = 0;
            for (; i < body.size(); i++) {
                Dec idec = body.get(i);
                if (! (idec instanceof ConstructorDec) || idec == dec) continue;
                
                ConstructorDec constructorDec = (ConstructorDec)idec;
                if (constructorDec.getFormals().matches(dec.getFormals())) break;
            }
            if (i >= body.size()) break;
            
            if (cookieFormal == null) {
                cookieFormal =
                  ast.makeFormal(getTypeManager().shortType.getArrayType(), "_ajcookie");
                dec.getFormals().add(cookieFormal);
            } else {
                cookieFormal.setTypeD(cookieFormal.getType().getArrayType().makeTypeD());
            }
        }
        
        if (cookieFormal != null) {
            thisCallExpr.getArgs().add(ast.forceCast(cookieFormal.getType(), ast.makeNull()));
        }
    } 
}
