/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.crosscuts.AspectJCompiler;
import org.aspectj.util.PartialOrder;

import java.util.*;

/**
 * Each JoinPoint object is the static shadow of N possible dyanmic join points
 * in the executing program.  There should be exactly one JoinPoint object for each
 * such possible static shadow.
 */
public abstract class JoinPoint extends CompilerObject {
    public final static int METHOD_EXECUTION = 0X001;
    //public final static int METHOD_RECEPTION      = 0x002;
    public final static int METHOD_CALL = 0x004;
    public final static int CONSTRUCTOR_EXECUTION = 0x008;
    //public final static int CONSTRUCTOR_RECEPTION = 0x010;
    public final static int CONSTRUCTOR_CALL = 0x020;
    public final static int FIELD_GET = 0x040;
    public final static int FIELD_SET = 0x080;
    public final static int HANDLER = 0x100;

    public final static int ADVICE_EXECUTION = 0x1000;
    public final static int STATIC_INITIALIZER_EXECUTION = 0x2000;
    public final static int INSTANCE_INITIALIZER_EXECUTION = 0x4000;
    public final static int INITIALIZER_EXECUTION = 0x8000;

    public final static int SUPER_METHOD_CALL = 0x10000;
    public final static int SUPER_CONSTRUCTOR_CALL = 0x20000;

    public static class Kind {
        private String name;
        private int id;

        public Kind(String name, int id) {
            this.name = name;
            this.id = id;
        }

        public int getIntCode() {
            return id;
        }
        public String toString() {
            return name;
        }
        public String toFullString() {
            return name;
        }
        public String toShortString() {
            return name;
        }
    }

    public abstract Kind getKind();

    //initial size of 0 is a compromise optimization for a rarely used list
    List plans /*JpPlan*/
    = new ArrayList(0);
    private boolean implementedPlans = false;

    private VarDec dynamicJoinPointDec = null;
    private FieldDec staticJoinPointDec = null;

    public JoinPoint(JavaCompiler compiler) {
        super(compiler);
    }

    public void addPlan(JpPlan plan) {
        //System.out.println(this + " add " + plan);
        plans.add(plan);
    }

    /*********************************************************************************
     * get information about the join point
     */
    public abstract ASTObject getSourceLocation();

    // This encapsulates the "signature" of the JoinPoint
    public abstract Dec getTargetDec(); // { return getEnclosingCodeDec(); }
    public SemanticObject getTargetSO() {
        return getTargetDec().getCorrespondingSemanticObject();
    }
    // overridden only be CalleSideCallPoint
    public ASTObject getTargetNode() {
        return getSourceLocation();
    }

    // This is the "kind"
    public final int getTypeCode() {
        return getKind().getIntCode();
    }

    public abstract Type getResultType();

    public Type getExceptionType() {
        return null;
    }

    public Type getBytecodeType() {
        return getSourceLocation().getBytecodeType();
    }

    public Type getDeclaringType() {
        return getSourceLocation().getDeclaringType();
    }

    public abstract Type getTargetType();

    /**
     * In the default case target == this
     * @return the type of the target expr or null if there is no target
     */
    public Type getTargetExprType() {
        return getThisExprType();
    }

    /**
     * In the default case target == this
     * @return a new expr that refers to the target or null if there is no target
     */
    public Expr makeTargetExpr() {
        return makeThisExpr();
    }

    /**
     * @return the type of the this expr or null if there is no this
     */
    public Type getThisExprType() {
        if (isStaticContext())
            return null;
        else
            return getDeclaringType();
    }

    public abstract boolean isStaticContext();

    /**
     * @return a new expr that refers to the target or null if there is no target
     */
    public Expr makeThisExpr() {
        Type thisType = getThisExprType();
        if (thisType == null)
            return null;
        else
            return getAST().makeThis(thisType);
    }

    /**
     * @return the Types of the args
     */
    //     public List/*Type*/ getArgsTypes() {
    //     }

    /**
     * @return the args exprs
     */
    public abstract Exprs makeArgsExprs();

    /* Used only by CalleeSideCallPoint XXX move it there */
    public Expr makeThisExprOrType() {
        if (isStaticContext()) {
            return getAST().makeTypeExpr(getBytecodeType());
        } else {
            return getAST().makeThis(getBytecodeType());
        }
    }

    public Expr makeVoidReturnExpr() {
        return getAST().makeNull();
    }

    public abstract boolean canThrow(Type t);
    public abstract Collection getPossibleCheckedExceptions();

    public void showError(ASTObject node, String message) {
        //XXX need to produce a crosscutting error here
        node.showError("on target " + this +" " + message);
    }

    /*********************************************************************************
     * get reflective exprs for run-time information about the join point
     */
    public Type makeFactoryType() {
        return getTypeManager().getType("org.aspectj.runtime.reflect", "Factory");
    }

    public FieldDec makeFactoryDec() {
        final AST ast = getAST();

        java.io.File f = getStaticFieldHolder().getSourceFile();
        String fname;
        if (f == null)
            fname = "<missing>";
        else
            fname = f.getName();

        Modifiers modifiers = ast.makeModifiers(Modifiers.STATIC | Modifiers.FINAL);
        if (getStaticFieldHolder().getType().isInterface())
            modifiers.setPublic(true);

        return ast.makeField(
            modifiers,
            makeFactoryType(),
            "ajc$JPF",
            ast.makeNew(
                makeFactoryType(),
                ast.makeLiteral(fname),
                ast.makeLiteral(getStaticFieldHolder().getType())));
    }

    public FieldDec getFactoryDec() {
        if (getStaticFieldHolder().joinPointFactoryDec == null) {
            FieldDec factoryDec = makeFactoryDec();
            //factoryDec.setSynthetic();
            getStaticFieldHolder().joinPointFactoryDec = factoryDec;
            getStaticFieldHolder().getBody().add(0, factoryDec);
        }
        return getStaticFieldHolder().joinPointFactoryDec;
    }

    public Expr makeFactoryInstance() {
        return getAST().makeGet(getFactoryDec());
    }

    public Expr makeObjectsArray(Formals formals) {
        final AST ast = getAST();

        Exprs exprs = ast.makeExprs();
        for (int i = 0; i < formals.size(); i++) {
            exprs.add(ast.makeVar(formals.get(i)));
        }
        return ast.makeObjectArray(exprs);
    }

    public String makeString(int i) {
        return Integer.toString(i, 16);
    }

    public String makeString(Type t) {
        return t.getExternalName();
    }

    public String makeString1(ASTObject ast, boolean asType) {
        if (ast instanceof TypeD) {
            return makeString(((TypeD) ast).getType());
        } else if (ast instanceof FormalDec) {
            FormalDec formal = (FormalDec) ast;
            if (asType)
                return makeString(formal.getType());
            else
                return formal.getId();
        }
        return "?";
    }

    public String makeString(TypeDs list) {
        return makeString(list, true);
    }

    public String makeString(ASTObject list, boolean asType) {
        if (list == null)
            return "";
        StringBuffer buf = new StringBuffer();
        final int N = list.getChildCount();
        for (int i = 0; i < N; i++) {
            buf.append(makeString1(list.getChildAt(i), asType));
            buf.append(':');
        }
        return buf.toString();
    }

    public Expr makeFactoryCall(String name, String stringRep) {
        //        return getAST().makeLiteral(stringRep);
        return getAST().makeCall(
            makeFactoryInstance(),
            name,
            getAST().makeLiteral(stringRep));
    }

    public Expr makeSignatureExpr(Dec dec) {
        if (dec instanceof MethodDec) {
            return makeSignatureExpr((MethodDec) dec);
        } else if (dec instanceof ConstructorDec) {
            return makeSignatureExpr((ConstructorDec) dec);
        } else if (dec instanceof AdviceDec) {
            return makeSignatureExpr((AdviceDec) dec);
        } else if (dec instanceof FieldDec) {
            return makeSignatureExpr((FieldDec) dec);
        } else if (dec instanceof InitializerDec) {
            return makeSignatureExpr((InitializerDec) dec);
        } else {
            return null;
        }
    }

    public Expr makeSignatureExpr(ConstructorDec dec) {
        StringBuffer buf = new StringBuffer();
        buf.append(makeString(dec.getModifiers().getValue()));
        buf.append('-');
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), true));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), false));
        buf.append('-');
        buf.append(makeString(dec.getThrows()));
        buf.append('-');
        return makeFactoryCall("makeConstructorSig", buf.toString());
    }

    public Expr makeSignatureExpr(MethodDec dec) {
        StringBuffer buf = new StringBuffer();
        buf.append(makeString(dec.getModifiers().getValue()));
        buf.append('-');
        buf.append(dec.getName());
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), true));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), false));
        buf.append('-');
        buf.append(makeString(dec.getThrows()));
        buf.append('-');
        buf.append(makeString(dec.getResultType()));
        buf.append('-');
        return makeFactoryCall("makeMethodSig", buf.toString());
    }

    public Expr makeSignatureExpr(AdviceDec dec) {
        StringBuffer buf = new StringBuffer();
        buf.append(makeString(dec.getModifiers().getValue()));
        buf.append('-');
        buf.append(dec.getName());
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), true));
        buf.append('-');
        buf.append(makeString(dec.getFormals(), false));
        buf.append('-');
        buf.append(makeString(dec.getThrows()));
        buf.append('-');
        buf.append(makeString(dec.getResultType()));
        buf.append('-');
        return makeFactoryCall("makeAdviceSig", buf.toString());
    }

    public Expr makeSignatureExpr(FieldDec dec) {
        StringBuffer buf = new StringBuffer();
        buf.append(makeString(dec.getModifiers().getValue()));
        buf.append('-');
        buf.append(dec.getName());
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        buf.append(makeString(dec.getType()));
        buf.append('-');
        return makeFactoryCall("makeFieldSig", buf.toString());
    }

    public Expr makeSignatureExpr(InitializerDec dec) {
        StringBuffer buf = new StringBuffer();
        buf.append(makeString(dec.getModifiers().getValue()));
        buf.append('-');
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        return makeFactoryCall("makeInitializerSig", buf.toString());
    }

    public Expr makeSignatureExpr(CatchClause dec) {
        StringBuffer buf = new StringBuffer();
        buf.append('-');
        buf.append('-');
        buf.append(makeString(dec.getDeclaringType()));
        buf.append('-');
        buf.append(makeString(dec.getFormal().getType()));
        buf.append('-');
        buf.append(dec.getFormal().getId());
        buf.append('-');
        return makeFactoryCall("makeCatchClauseSig", buf.toString());
    }

    public Expr makeExecutingObjectExpr() {
        Expr ret = makeThisExpr();
        if (ret == null)
            ret = getAST().makeNull();
        return ret;
    }

    public JoinPoint getEnclosingExecutionJoinPoint() {
        return null;
    }

    protected final Type getJoinPointType() {
        return getTypeManager().getJoinPointType();
    }

    protected final Type getJoinPointStaticPartType() {
        return getTypeManager().getJoinPointStaticPartType();
    }

    protected Expr makeSignatureExpr() {
        return makeSignatureExpr(getTargetDec());
    }
    protected final String getKindString() {
        return getKind().toFullString();
    }

    protected Expr makeStaticEnclosingExecutionJoinPointExpr() {
        return null;
    }

    public final Expr makeStaticJoinPointExpr() {
        final AST ast = getAST();

        Exprs args =
            ast.makeExprs(ast.makeLiteral(getKindString()), makeSignatureExpr());

        ASTObject sourceLocation = getSourceLocation();
        if (sourceLocation == null) {
            args.add(ast.makeNull());
        } else {
            args.add(ast.makeLiteral(sourceLocation.getBeginLine()));
            args.add(ast.makeLiteral(sourceLocation.getBeginColumn()));
        }

        return ast.makeCall(makeFactoryInstance(), "makeSJP", args);
    }

    private Expr reflectNull(Expr expr) {
        if (expr == null)
            return getAST().makeNull();
        else
            return expr;
    }

    public final Expr makeDynamicJoinPointExpr() {
        final AST ast = getAST();

        Exprs args =
            ast.makeExprs(
                makeStaticJoinPointVarExpr(),
                reflectNull(makeThisExpr()),
                reflectNull(makeTargetExpr()),
                ast.makeObjectArray(makeArgsExprs()));
        return ast.makeStaticCall(makeFactoryType(), "makeJP", args);
    }

    public Expr makeStaticEnclosingJoinPointVarExpr() {
        if (getStaticEnclosingJoinPointDec() == null) {
            return getAST().makeNull();
        }
        return getAST().makeGet(getStaticEnclosingJoinPointDec());
    }

    public Expr makeStaticJoinPointVarExpr() {
        return getAST().makeGet(getStaticJoinPointDec());
    }

    public Expr makeDynamicJoinPointVarExpr() {
        return getAST().makeVar(getDynamicJoinPointDec());
    }

    private boolean isSyntheticJoinPoint(Dec dec) {
        if (dec.isSynthetic() && dec instanceof FieldDec) {
            FieldDec field = (FieldDec) dec;
            if (field.getId().indexOf("$ajcjp") != -1)
                return true;
            if (field.getId().startsWith("ajc$"))
                return true;
        }
        return false;
    }

    public TypeDec getStaticFieldHolder() {
        return getBytecodeType().getOutermostType().getTypeDec();
    }

    FieldDec makeStaticFinalField(Type type, String name, Expr value) {
        final AST ast = getAST();
        Modifiers modifiers =
            ast.makeModifiers(Modifiers.FINAL | Modifiers.STATIC | Modifiers.PRIVATE);
        if (getStaticFieldHolder().getType().isAspect())
            modifiers.setPublic(true);
        if (getStaticFieldHolder().getType().isInterface())
            modifiers.setPublic(true);
        FieldDec ret = ast.makeField(modifiers, type, name, value);
        //ret.setSynthetic();
        return ret;
    }

    void addStaticJoinPointDecToBody() {
        if (staticJoinPointDec != null) {
            staticJoinPointDec.setInitializer(makeStaticJoinPointExpr());
            Decs body = getStaticFieldHolder().getBody();
            int i = 0;
            for (; i < body.size(); i++) {
                if (!isSyntheticJoinPoint(body.get(i))) {
                    staticJoinPointDec.setId(staticJoinPointDec.getId() + i);
                    body.add(i, staticJoinPointDec);
                    break;
                }
            }
            if (i == body.size())
                body.add(staticJoinPointDec);
        }
    }

    public FieldDec getStaticJoinPointDec() {
        if (staticJoinPointDec == null) {
            String name = getAST().makeGeneratedName(getTargetDec().getId() + "$ajcjp");
            staticJoinPointDec =
                makeStaticFinalField(getJoinPointStaticPartType(), name, null);
            staticJoinPointDec.setAllEnclosingTypes(getStaticFieldHolder().getType());
        }
        return staticJoinPointDec;
    }

    public VarDec getDynamicJoinPointDec() {
        if (dynamicJoinPointDec == null) {
            dynamicJoinPointDec =
                getAST().makeFinalVar(
                    getJoinPointType(),
                    "thisJoinPoint",
                    makeDynamicJoinPointExpr());
        }
        return dynamicJoinPointDec;
    }

    public FieldDec getStaticEnclosingJoinPointDec() {
        if (getEnclosingExecutionJoinPoint() == null) {
            //System.out.println("no enclosing execution for: " + this);
            return getStaticJoinPointDec();
        }
        return getEnclosingExecutionJoinPoint().getStaticJoinPointDec();
    }

    /**********************************************************************************/

    //XXX this is the bare minimum for a useful error message here
    void showPrettyCircularityError(List plans) {
        for (Iterator i = plans.iterator(); i.hasNext();) {
            AdvicePlan plan = (AdvicePlan) i.next();
            AdviceDec adviceDec = plan.getAdviceDec();
            AspectDec aspectDec = plan.getAspectDec();

            if (adviceDec != null) {
                getCompiler().showError(
                    adviceDec,
                    "circularity in advice precedence applied to " + this);
            } else {
                getCompiler().showError(
                    null,
                    plan + ": circularity in advice precedence applied to " + this);
            }
        }
    }

    public boolean hasPlans() {
        return plans != null && plans.size() > 0;
    }
    
    // only called from MethodCallJp.preImplement()
    public boolean plansMatch(JoinPoint other) {
        if (plans.size() != other.plans.size())
            return false;

        List otherPlans = new LinkedList(other.plans);

        outer : for (Iterator i1 = plans.iterator(); i1.hasNext();) {
            JpPlan plan1 = (JpPlan) i1.next();
            // if one of the plans is for advice that needs call site context
            // then we can't merge these sets
            if (plan1 instanceof AdvicePlan) {
                AdvicePlan advicePlan = (AdvicePlan)plan1;
                if (advicePlan.getAdviceDec().needsCallSiteContext()) {
                    return false;
                }
            }
            
            for (Iterator i2 = otherPlans.iterator(); i2.hasNext();) {
                JpPlan plan2 = (JpPlan) i2.next();
                if (plan1.matches(plan2)) {
                    i2.remove();
                    continue outer;
                }
            }
            return false;
        }
        return true;
    }

    public void forgetPlans() {
        plans = new ArrayList(0);
    }

    public void makeCorrespondences() {
        if (plans == null)
            return;

        for (Iterator i = plans.iterator(); i.hasNext();) {
            JpPlan plan = (JpPlan) i.next();
            if (plan instanceof AdvicePlan) {
                AdviceDec adviceDec = ((AdvicePlan) plan).getAdviceDec();

                ASTObject codePoint = getTargetNode();

                //            System.out.println("corr: " + adviceDec.toShortString() + " to "
                //            + codePoint.toShortString());
                ((AspectJCompiler) getCompiler()).getCorrespondences().addPointsTo(
                    adviceDec,
                    codePoint);

            }
        }
    }

    void sortPlans() {
        if (plans.size() < 2)
            return;

        //??? could do this more efficiently by pre-sorting by kind,
        //??? but we might remove that property from the language, so don't optimize yet.
        if (PartialOrder.sort(plans) == null) {
            showPrettyCircularityError(plans);
        }
    }
    
    public boolean needToImplement() {
        return plans.size() != 0
            || dynamicJoinPointDec != null
            || staticJoinPointDec != null;
    }

    public void implementPlans() {
        if (implementedPlans || !needToImplement()) return;
        implementedPlans = true;
        sortPlans();
        Collections.reverse(plans);
        getCompiler().showMessage(
            "  implementing plans on " + this +System.identityHashCode(this));

        preImplement();
        if (!needToImplement()) return;

        for (Iterator i = plans.iterator(); i.hasNext();) {
            JpPlan plan = (JpPlan) i.next();
            plan.wrapJoinPoint(this);
        }
        makeCorrespondences();
        finishJoinPoint();
    }

    protected void preImplement() {
    }


    public void finishJoinPoint() {
        if (dynamicJoinPointDec != null) {
            getStmts().add(0, dynamicJoinPointDec);
        }

        addStaticJoinPointDecToBody();
    }

    public abstract Stmts getStmts();

    public abstract void setStmts(Stmts stmts);
}