/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.util.JavaStrings;

import java.util.*;


public class MethodCallJp extends CallJp {
    public MethodCallJp(CallExpr callExpr, JoinPoint enclosingJp) {
        super(callExpr, enclosingJp);
    }

    public static final Kind KIND = new Kind("method-call", METHOD_CALL);

    public Kind getKind() { return KIND; }

    public boolean isMethodCall() { return true; }

    public Type getTargetExprType() {
        if (getCalledCodeDec().isStatic()) return null;
        return getCallExpr().getExpr().getType();
    }

    public Type getTargetType() {
        return getCallExpr().getExpr().getType();
    }

    protected AnyCallExpr makeBasicInnerCall(Expr targetExpr, Exprs argsExprs) {
        AnyCallExpr innerCall = getCallExpr();
        innerCall = (AnyCallExpr)innerCall.copy();
        innerCall.setArgs(argsExprs);
        innerCall.setExpr(targetExpr);
        return innerCall;
    }
    
    protected ASTObject makeInnerCall(Expr targetExpr, Exprs argsExprs) {
        AnyCallExpr innerCall = makeBasicInnerCall(targetExpr, argsExprs);
        
        if (skipTypes != null) {
            ASTObject ret = innerCall;
            boolean isVoid = innerCall.getType().isVoid();
            
            for (Iterator i = skipTypes.entrySet().iterator(); i.hasNext(); ) {
                Map.Entry entry = (Map.Entry)i.next();
                Type type = (Type)entry.getKey();
                MethodDec md = (MethodDec)entry.getValue();
                
                final AST ast = getAST();
                if (md.isStatic()) {
                    return ast.makeStaticCall(md.getMethod(), (Exprs)innerArgs.copy());
                }
                
                Expr innerExpr = (Expr)targetExpr.copy();
                
                if (!innerExpr.getType().isCoercableTo(type)) continue;
                
                Expr skipCall = ast.makeCall(md.getMethod(), 
                                        ast.makeCast(type, innerExpr), 
                                        (Exprs)innerArgs.copy());
                
                if (innerExpr.getType().isSubtypeOf(type)) {
                    return skipCall;
                } else {
                    Expr instanceTest = ast.makeInstanceof((Expr)innerExpr.copy(), type);
                    if (isVoid) {
                        Stmt stmt;
                        if (ret instanceof Stmt) stmt = (Stmt)ret;
                        else stmt = ast.makeStmt((Expr)ret);
                        ret = ast.makeIf(instanceTest, ast.makeStmt(skipCall), stmt);
                    } else {
                        ret = ast.makeTriTest(instanceTest, skipCall, (Expr)ret);
                    }
                }
            }
            return ret;
        }
        return innerCall;
    }

    protected void preImplement() {
        if (!hasPlans()) return;
        
        CodeDec codeDec = getCalledCodeDec();
        MethodDec targetMethodDec = (MethodDec)codeDec;
        
        boolean needsCallSiteAdvice = false;
        Map skipAdviceOnTypes = new HashMap();
        
        //System.out.println("call point: " + this);        
        
        for (Iterator i = targetMethodDec.getMatchingMethods().iterator(); i.hasNext(); ) {
            MethodDec dec = (MethodDec)i.next();
            //System.out.println("  md: " + dec.toShortString());
            CalleeSideCallJp targetPoint =
                (CalleeSideCallJp)getWorld().calleeSideCallPoints.get(dec);
            //System.out.println("  tp: " + targetPoint + ", " + targetPoint.hasPlans());
            if (targetPoint == null || !targetPoint.hasPlans()) {
                needsCallSiteAdvice = true;
            } else {
                skipAdviceOnTypes.put(targetPoint.getCodeDec().getDeclaringType(), 
                                      targetPoint.getPostMethodDec());
                if (this.plansMatch(targetPoint)) {
                    //System.out.println("    plans match"); // do nothing
                } else {
                    needsCallSiteAdvice = true;
                }
            }
        }
        
        if (!needsCallSiteAdvice) {
            //???makeCorrespondences();
            forgetPlans();
        } else if (skipAdviceOnTypes.size() > 0) {
            Set keepTypes = Type.filterTopTypes(skipAdviceOnTypes.keySet());
            skipAdviceOnTypes.keySet().retainAll(keepTypes);
            skipAdviceOnTypes(skipAdviceOnTypes);
        }
        
        //System.err.println("planning: " + this + " plans: " + plans);
    }

    private Map skipTypes = null;
    public void skipAdviceOnTypes(Map onTypes) {
        //System.out.println("  top skipAdviceOnTypes: " + onTypes);
        skipTypes = onTypes;
    }   
}
