/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.tools.ide;


import java.util.Vector;
import java.util.Enumeration;
import java.lang.String;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.FileNotFoundException;

public class StringBasedSymbolManager {

    public static void main(String[] args) {
	fetchDeclarations(args[0]);
	//fetchSourceLine(args[0], 154);
    }

    public static void fetchDeclarations(String filename) {
	try {
	    SymbolManager mgr = SymbolManager.getSymbolManager();
	    Declaration[] decls = mgr.getDeclarations(filename);
	    if ( decls == null )
		throw new Error("Can't find declarations file for " + filename);
	    printDecls(decls, true);
	}
	catch (Error e) {
	    lose(e);
	}
    }

    public static void fetchSourceLine(String filename, int line) {
	try {
	    SymbolManager mgr = SymbolManager.getSymbolManager();
	    SourceLine srl = mgr.mapToSourceLine(filename, line);
	    if ( srl == null )
		print("nil");
	    else {
		String fn = srl.filename;
		int    ln = srl.line;
		print("(\"" + fixFilename(fn) + "\" . " + ln + ")");
	    }
	}
	catch (Error e) {
	    lose(e);
	}
    }

    static void lose(Error e) {
	try {
	    print("(ERROR \"" + e.toString() + "\")");
	}
	catch(Error ex) { }
    }

    static void printDecls(Declaration[] decls, boolean recurse) {
        print("(");
	for (int i = 0; i < decls.length; i++) {
	    Declaration decl = decls[i];
	    /*
	     * null means that one symbol file pointed to a declaration
	     * that another symbol file should have contained but didn't.
	     * This can happen if the symbol files are out of date with
	     * each other.
	     */
	    if ( decl != null )
		printDecl(decl, recurse);
        }
        print(") ");
    }

    static void printDecl(Declaration decl, boolean recurse) {
        String crosscutDesignator = decl.getCrosscutDesignator();
	String kind = decl.getKind().toLowerCase();
	Declaration mappedDecl = nearestMappedDecl(decl);
        print("(");
	print("(" + mappedDecl.getBeginLine() + " . " + mappedDecl.getBeginColumn() + ") ");
        print("(" + mappedDecl.getEndLine() + " . " + mappedDecl.getEndColumn()   + ") ");
        print(kind + " ");                                     //2
	if (crosscutDesignator == null)
	    print("\"" + decl.getSignature() + "\" ");           //3
	else if (kind.equals("introduction"))
	    print("\"" + decl.getSignature() + " "  + crosscutDesignator + "\" ");  //3
        else                 /* advice */
	    print("\"" + decl.getSignature() + ": " + crosscutDesignator + "\" ");  //3
        print("\"" + fixFilename(decl.getFilename()) + "\"");  //4
	print("\"" + decl.getDeclaringType() + "\" ");         //5
	if (recurse) {
	    printDecls(decl.getTargets(),      false);         //6
	    printDecls(decl.getPointedToBy(),  false);         //7
	    printDecls(decl.getDeclarations(), true);          //8
	}
	else {
	    print("nil");
	    print("nil");
	    print("nil");
	}
	print(fixBoolean(decl.isType()));       //9
	print(fixBoolean(decl.isIntroduced())); //10
	print(fixBoolean(decl.hasBody()));      //11
	print(fixBoolean(decl.hasSignature())); //12

        print(")");
    }

    /*
     * this method and maybe the next should really be in Declaration.
     */

    static boolean legalLineNumber(int ln) { return (ln != -1); }

    /*
     * Searches up declaration hierarchy until it finds a declaration
     * that has useable line numbers.
     */
    static Declaration nearestMappedDecl(Declaration decl) {
        Declaration parentDecl;
        if (legalLineNumber(decl.getBeginLine()))
            return decl;
        else if ((parentDecl = decl.getParentDeclaration()) != null)
            return nearestMappedDecl(parentDecl);
        else
            return decl;   // bail with self if no parent
    }

    /*
     * this should really be in Declaration.  it only does something for
     * introduction?
     */

    static String fixBoolean(boolean flag) {
	if ( flag ) { return "t "; } else { return "nil ";}
    }


    static String fixFilename(String filename) {
        return subst("\\\\", "\\", filename);
    }

    static int convertLineToPosition(String filename, int lineNumber)
	throws FileNotFoundException {

	FileReader     freader = new FileReader(filename);
	BufferedReader breader = new BufferedReader(freader);
	int position = 0;
	try {
	    for(int n = 0; n < lineNumber; n++) {
		String line = breader.readLine();
		position += line.length();
	    }
	}
	catch (IOException e) {}
	return position;
    }


    static void print(String string) {
	System.out.println(string);
    }

    static String subst(String n, String o, String in) {
	int pos = in.indexOf(o);
	if (pos == -1)
	    return in;
	return in.substring(0, pos) +
	       n +
	       subst(n, o, (in.substring(pos + o.length())));
    }


}
