/*
 * Copyright (c) 2002 PARC Inc.  All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works based
 * upon this software are permitted.  Any distribution of this software or
 * derivative works must comply with all applicable United States export
 * control laws.
 *
 * This software is made available AS IS, and PARC Inc. makes no
 * warranty about the software, its performance or its conformity to any
 * specification.
 */

package com.xerox.printing; 

import org.aspectj.lang.*;

/** test cases for controlling field writes */
public class RuntimeWrites {

    public static void main(String[] args) {
        final String PASS      = "---- PASS: ";
        final String FAIL      = "---- FAIL: ";
        final String invalid   = "invalid write, " ;
        final SubPrinterStream me = new SubPrinterStream();
        String label = null;
        try {
            label = invalid + "outside init - nonstatic";
            me.setOne(1);
            Util.log(FAIL + label);
        } catch (Error e) {
            Util.log(PASS + label + ": " + e.getMessage());
        }

        try {
            label = invalid + "outside init - static";
            me.one = 0;
            Util.log(FAIL + label);
        } catch (Error e) {
            Util.log(PASS + label + ": " + e.getMessage());
        }

        try {
            label = invalid + "caller is not same as target";
            PrinterStream ps = new PrinterStream();
            new PrinterStream(ps);
            Util.log(FAIL + label);
        } catch (Error e) {
            Util.log(PASS + label + ": " + e.getMessage());
        }
    }
}


/** 
 * Control field writes.
 * This implementation restricts writes to the same object 
 * during initialization.
 * This is like having field values be final, except that 
 * they may be set set outside the constructor.
 */
aspect ControlFieldWrites {
    public static boolean throwError;
    
    // article page 42 - field writes

    /** any write to a non-static field in PrinterStream itself */
    pointcut fieldWrites() : set(!static * PrinterStream.*);

    /** execution of any constructor for PrinterStream or subclasses */
    pointcut init() : execution(PrinterStream+.new(..));

    /** 
     * Handle any situation where fields are written
     * outside of the control flow of initialization
     * @throws Error when run.
     */
    before() : fieldWrites() && !cflow(init()) { 
        throw new Error("set outside of init");
    }

    /** 
     * Handle any situation where fields are written
     * by another object.
     * @throws Error if caller and targ are not the same reference
     */
    before(Object caller, PrinterStream targ) : this(caller) 
        && target(targ) && fieldWrites() {
        if (caller != targ) {
            throw new Error(caller + " setting fields in targ " + targ);
        }
    }
}

