/*
 * Copyright (c) 2002 PARC Inc.  All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works based
 * upon this software are permitted.  Any distribution of this software or
 * derivative works must comply with all applicable United States export
 * control laws.
 *
 * This software is made available AS IS, and PARC Inc. makes no
 * warranty about the software, its performance or its conformity to any
 * specification.
 */
package com.xerox.printing; 
    
import java.io.*;

/** 
 * Demonstrate technique of fault-injection 
 * as coordinated by test driver.
 */
aspect InjectingIOException { 
    /** the test starts when the driver starts executing */
    pointcut testEntryPoint(TestDriver driver) :
        target(driver) && execution(* startTest()); 
    
    /** 
     * The fault may be injected at the execution of any
     * (non-static) PrinterStream  method that throws an IOException
     */
    pointcut testCheckPoint(PrinterStream stream) : target(stream) 
        && execution(public * *(..) throws IOException);

    /*
      In the article pointcut, PrinterStream+ is redundant because of
      target(PrinterStream), so we can simplify the execution pcd - 
        
        from: execution(public * PrinterStream+.*(..) throws IOException)
          to: execution(public * *(..) throws IOException)
       
     */

    // article page 43 - fault injection
    /**
     * After the method returns normally, ask the test driver for an 
     * exception to throw (a fault to inject) if this is a test case.
     */
    after (TestDriver driver, PrinterStream stream) returning 
                                                     throws IOException :
            cflow(testEntryPoint(driver)) && testCheckPoint(stream) {
        IOException e = driver.createException(stream);
        if (null != e) {
            throw e;
        }
    }

    /* Comment on the after advice IOException declaration:

       "throws IOException" is a declaration of the advice,
       not the pointcut.

       Since the advice might throw the injected fault, it
       must declare that it throws IOException. When advice declares
       exceptions thrown, the compiler will emit an error if any 
       join point is not also declared to throw an IOException.  

       In this case, the testCheckPoint pointcut only picks out 
       methods that throw IOException, so the compile will not
       signal any errors.
    */
}

/** this runs the test case */
public class Injection {
    /** Run three print jobs, two as a test and one normally */
    public static void main(String[] args) throws Exception {
        final String label = "Injection.main() - ";
        Runnable r = new Runnable() {
                public void run() { 
                    try { new TestDriver().startTest(); } 
                    catch (IOException e) {
                        Util.log("ok, got injected error " + e.getMessage());
                    }
                }
            };

        Util.log(label + "starting separate test thread");
        Thread t = new Thread(r);
        t.start();

        Util.log(label + "running test in this thread");
        r.run();
        t.join();

        Util.log(label + "running job normally (not TestDriver)");
        new PrintJob().runPrintJob();
    }
}

/** handle starting of test and determining whether to inject failure */
class TestDriver {

    /** start a new test */
    public void startTest() throws IOException {
        new PrintJob().runPrintJob();
    }

    /** this implementation always injects a failure */
    public IOException createException(PrinterStream p) {
        return new IOException(""+p);
    }
}

