/*
 * Copyright (c) 2002 PARC Inc.  All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works based
 * upon this software are permitted.  Any distribution of this software or
 * derivative works must comply with all applicable United States export
 * control laws.
 *
 * This software is made available AS IS, and PARC Inc. makes no
 * warranty about the software, its performance or its conformity to any
 * specification.
 */

package com.xerox.printing; 

public class RoundTrip {
    public static void main(String[] args) throws java.io.IOException {
        PrinterStream p = testing();
        p.write();
        Util.log("did create and public write");
    } 
    static PrinterStream testing() { return new PrinterStream(1); }

} 

aspect VerifyPrinterStreamIntegrity {
    // article page 76 - round trip

    /** Check PrinterStream after any public call. */
    after (PrinterStream stream) returning : target(stream)
            && call(public * *(..)) {
        check(stream);
    }

    /** Check any PrinterStream returned from any call in our packages. */
    after () returning (PrinterStream stream) :
        call (PrinterStream+ com.xerox..*.*(..)) 
        && !call (PrinterStream PrinterStream.make(BufferedPrinterStream)) 
    {
        check(stream);
    }  

    /*
      Note one difference from the article pointcut:

         article: call (PrinterStream+ com.xerox.printing..*(..)) 
            here: call (PrinterStream+ com.xerox..*.*(..)) 

      The syntax for method call pcd's is something like this:

           call(TypePattern {TypePattern.}Identifier(TypePattern..))

      where '..' in a TypePattern means "anything between a start and end 
      '.', and '*' in a TypePattern or identifier means "any contiguous 
      name, not including any '.'".

      While '..*' works to specify '{TypePattern.}Identifier', it is more
      correct to use '..*.*', i.e.,       '{..*}.<*>'.  
                                                |
                                     TypePattern.Identifier delimiter

      Further, to specify any type, you can omit the TypePattern:

             call(boolean equals(Object))  // any call to equals()
             call(* *(..))                 // any call whatsoever

      For information on syntax, see the Programming Guide.
     */

    /** 
     * Verify PrinterStream.
     * This uses a round-trip as a way to verify the
     * integrity of PrinterStream, but one could also use
     * a self-test (built-in or otherwise) coded specifically 
     * for validating the object (without changing state).
     * @throws Error if input PrinterStream is invalid 
     */
    void check(PrinterStream stream) {
        BufferedPrinterStream bufferStream 
            = new BufferedPrinterStream(stream);
        PrinterStream newStream = PrinterStream.make(bufferStream);
        if (newStream.number != stream.number) {
            throw new Error("round-trip failed for " + stream);
        }
    }
    /*
      check(..) avoids using equals(Object) for two reasons:
      - the implementation of equals(Object) may change independently
      - the first advice pointcut picks out calls to all public methods,
        including the call to equals(Object) in check(..), resulting
        in a StackOverflowError.  It is better to code methods 
        defensively than to require all pointcuts to avoid infinite
        recursion, but you could avoid the recursion here as follows:

           target(stream) && call(public * *(..)) 
           && !call(boolean equals(Object)) // do not check after equals(..)
      
     */
}
