
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde;

import java.util.*;
import java.io.*;
import org.aspectj.ajde.ui.ErrorHandler;
import org.aspectj.asm.associations.*;

/**
 * Represents configurable project properties.  Extend and override with
 * IDE-specific implementations.
 *
 * @author Mik Kersten
 */
public class ProjectProperties {
    public static final String FILE_NAME = "/.ajbrowser";
    public static final String DEFAULT_CONFIG_NAME = "<all project files>";
    private final String VALUE_SEP = ";";
    private Properties properties = new Properties();

    public ProjectProperties() {
        try {
            if (new File(getPropertiesFilePath()).exists()) {
                properties.load(new FileInputStream(getPropertiesFilePath()));
            }
        } catch (IOException ioe) {
            ioe.printStackTrace();
            ErrorHandler.handleError("Could not read properties", ioe);
        }
    }

    public boolean isGlobalMode() {
        return true;
    }

    public String getProjectName() {
        return null;
    }

    public String getClassToRun() {
        return getProperty("runtime.mainClass");
    }

    public void setClassToRun(String mainClass) {
        storeProperty("runtime.mainClass", mainClass);
    }

    public String getProjectDir() {
        return new File(TopManager.BROWSER_MANAGER.getCurrConfigFile()).getParent();
    }

    public List getProjectFiles() {
        return null;
    }

    public String getProjectSourcePath() {
        return null;
    }

    public String getBootClasspath() {
        return null;
    }

    public void setCompilerFlags(String flags) {
        storeProperty("build.flags", flags);
    }

    public String getCompilerFlags() {
        return getProperty("build.flags");
    }

    public String getRunParams() {
        return null;
    }

    public String getVMParams() {
        return null;
    }

    public String getOutputPath() {
        String outputPath = getProperty("build.outputpath");
        if (outputPath == null) {
            return ".";
        } else {
            return outputPath;
        }
    }

    public void setOutputPath(String path) {
        storeProperty("build.outputpath", path);
    }

    public String getUserClasspath() {
        return getProperty("build.classpath");
    }

    public String getClasspath() {
        String systemPath = System.getProperties().getProperty("java.class.path", ".");
        String userPath = getProperty("build.classpath");
        return userPath + File.pathSeparator + systemPath;
    }

    public void setUserClasspath(String path) {
        storeProperty("build.classpath", path);
    }

    public String getWorkingDir() {
        String workingDir = getProperty("build.workingdir");
        if (workingDir == null || workingDir.equals(getOutputPath())) {
            return getOutputPath() + "./ajworkingdir";
        } else {
            return workingDir;
        }
    }

    public void setWorkingDir(String path) {
        storeProperty("build.workingdir", path);
    }

    public List getStructureViewOrdering() {
        return getPropertyValues("structureView.ordering");
    }

    public void setStructureViewOrdering(List ordering) {
        storeProperty("structureView.ordering", ordering);
    }

    /**
     * Returns default values if no associations are saved.
     */
    public List getActiveAssociations() {
        List associations = getPropertyValues("structureView.associations");
        if (associations.size() == 0) {
            associations.add(Advice.METHOD_CALL_SITE_RELATION.toString());
            associations.add(Advice.METHOD_RELATION.toString());
            associations.add(Advice.CONSTRUCTOR_CALL_SITE_RELATION.toString());
            associations.add(Advice.CONSTRUCTOR_RELATION.toString());
            associations.add(Advice.FIELD_ACCESS_RELATION.toString());
            associations.add(Advice.INITIALIZER_RELATION.toString());
            associations.add(Advice.HANDLER_RELATION.toString());
            associations.add(Advice.INTRODUCTION_RELATION.toString());
            associations.add(Introduction.INTRODUCES_RELATION.toString());
            associations.add(Inheritance.IMPLEMENTS_RELATION.toString());
            associations.add(Inheritance.INHERITS_RELATION.toString());
            associations.add(Inheritance.INHERITS_MEMBERS_RELATION.toString());
            associations.add(Reference.USES_POINTCUT_RELATION.toString());
            associations.add(Reference.IMPORTS_RELATION.toString());
        }
        return associations;
    }

    public void setActiveAssociations(List associations) {
        storeProperty("structureView.associations", associations);
    }

    public void setActiveFilteredMemberKinds(List associations) {
        storeProperty("structureView.filtering.memberKinds", associations);
    }

    public String getActiveHierarchy() {
       return getProperty("structureView.hierarchy");
    }

    public void setActiveHierarchy(String hierarchy) {
        storeProperty("structureView.hierarchy", hierarchy);
    }

    public List getActiveVisibility() {
        return getPropertyValues("structureView.filtering.accessibility");
    }

    public void setActiveVisiblity(List visibility) {
        storeProperty("structureView.filtering.accessibility", visibility);
    }

    public List getActiveModifiers() {
        return getPropertyValues("structureView.filtering.modifiers");
    }

    public List getActiveFilteredMemberKinds() {
        return getPropertyValues("structureView.filtering.memberKinds");
    }

    public String getActiveGranularity() {
        return getProperty("structureView.granularity");
    }

    public void setActiveGranularity(String granularity) {
    storeProperty("structureView.granularity", granularity);
    }

    public void setActiveModifiers(List modifiers) {
        storeProperty("structureView.filtering.modifiers", modifiers);
    }

    protected String getProperty(String name) {
        return properties.getProperty(name);
    }

    protected List getPropertyValues(String name) {
        List values = new ArrayList();
        String valuesString = properties.getProperty(name);
        if (valuesString != null && !valuesString.trim().equals("")) {
            StringTokenizer st = new StringTokenizer(valuesString, VALUE_SEP);
            while (st.hasMoreTokens()) {
                values.add(st.nextToken());
            }
        }
        return values;
    }

    private void storeProperty(String name, String value) {
        properties.setProperty(name, value);
        saveProperties();
    }

    private void storeProperty(String name, List values) {
        String valuesString = "";
        for (Iterator it = values.iterator(); it.hasNext(); ) {
            valuesString += (String)it.next() + ';';
        }
        properties.setProperty(name, valuesString);
        saveProperties();
    }

    private void saveProperties() {
        try {
            properties.store(new FileOutputStream(getPropertiesFilePath()), "AJDE Settings");
        } catch (IOException ioe) {
            ErrorHandler.handleError("Could not write properties", ioe);
        }
    }

    protected String getPropertiesFilePath() {
        String path = System.getProperty("user.home");
        if (path == null) {
            path = ".";
        }
        return path + FILE_NAME;
    }
}