
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde.compiler;

import java.io.*;
import java.util.*;
import org.aspectj.compiler.base.*;
import org.aspectj.util.ConfigParser;
import org.aspectj.ajde.ui.ErrorHandler;
import org.aspectj.ajde.TopManager;

/**
 * Responsible for the build process, including compiler invocation, threading, and error
 * reporting.
 *
 * @author Mik Kersten
 */
public class CompilerManager {

    private AjdeCompiler currCompiler = null;
    private CompilerMessages compilerMessages = null;
    private CompileProgressMonitor progressMonitor = null;
    private ArrayList compilerListeners = new ArrayList();
    private String configFile = "";
    private int lastCompileTime = 50;

    private final CompilerListener STATUS_TEXT_UPDATER = new CompilerListener() {
        public void compileFinished(String buildConfig, int buildTime, boolean succeeded) {
            int timeInSeconds = buildTime/1000;
            if (succeeded) {
                TopManager.IDE_MANAGER.setStatusText(" Build succeeded in " + timeInSeconds + " second(s).");
                TopManager.IDE_MANAGER.hideMessages();
            } else {
                TopManager.IDE_MANAGER.setStatusText(" Compile failed in " + timeInSeconds + " second(s)");
                TopManager.IDE_MANAGER.showMessages();
            }
        }
    };

//	/**
//	 * Use the default compiler messages panel and compile progress monitor.
//	 */
//    public CompilerManager() {
//        this(null);
//    }

// 	/**
// 	 * Use the default progress monitor.
// 	 */
//	public CompilerManager(CompilerMessages compilerMessages) {
//		this(compilerMessages, null);
//	}

	/**
	 * Use specialized messages and progress monitor.
	 */
    public CompilerManager(CompilerMessages compilerMessages, CompileProgressMonitor progressMonitor) {
        this.compilerMessages = compilerMessages;
        this.progressMonitor = progressMonitor;
        addListener(STATUS_TEXT_UPDATER);
    }

    public void build() {
        if (TopManager.BROWSER_MANAGER.getCurrConfigFile() == null) {
            ErrorHandler.handleWarning("Nothing to compile, please add a \".lst\" file.", TopManager.IDE_MANAGER.getRootFrame());
            return;
        } else {
            if (TopManager.BROWSER_MANAGER.getCurrConfigFile().equals("<all project files>")) {
                build(TopManager.IDE_MANAGER.getDefaultConfigFile());
            } else {
                build(TopManager.BROWSER_MANAGER.getCurrConfigFile());
            }
        }
    }

    public void build(final String configFile) {
        TopManager.IDE_MANAGER.saveAll();
        if (configFile == null) {
            ErrorHandler.handleWarning("Please add a configuration file to compile.", TopManager.IDE_MANAGER.getRootFrame());
        } else {
            this.configFile = configFile;
            CompilerThread compilerThread = new CompilerThread();
            compilerThread.start();
        }
    }

    public void abortBuild() {
        if (currCompiler != null) {
            currCompiler.requestCompileExit();
        }
    }

    public String getLastCompiledConfig() {
        return configFile;
    }

//    public JPanel getCompilerMessagesPanel() {
//        if (compilerMessages instanceof CompilerMessagesPanel) {
//            return (CompilerMessagesPanel)compilerMessages;
//        } else {
//            return null;
//        }
//    }

    public AjdeCompiler getCurrCompiler() {
        return currCompiler;
    }

    public boolean isStructureDirty() {
        if (currCompiler != null) {
            return currCompiler.structureDirty;
        } else {
            return false;
        }
    }

    public void setStructureDirty(boolean structureDirty) {
        if (currCompiler != null) {
            currCompiler.structureDirty = structureDirty;
        }
    }

    public void addListener(CompilerListener compilerListener) {
        compilerListeners.add(compilerListener);
    }

    public void removeListener(CompilerListener compilerListener) {
        compilerListeners.remove(compilerListener);
    }

    private void notifyCompileFinished(String configFile, int buildTime, boolean succeeded) {
        for (Iterator it = compilerListeners.iterator(); it.hasNext(); ) {
            ((CompilerListener)it.next()).compileFinished(configFile, buildTime, succeeded);
        }
    }

    private void displayMessages(CompileResult compileResult) {
		String[]  descriptions = compileResult.getDescriptions();
        String[]  files        = compileResult.getfiles();
        Integer[] lineNumbers  = compileResult.getLineNumbers();
        if (descriptions.length == 0 && compileResult.getResult().trim() != "") {
            compilerMessages.addMessage(compileResult.getResult(), "", 0, 0, CompilerMessages.ERROR_MESSAGE	);
			return;
        }
        TopManager.IDE_MANAGER.showMessages();
		for ( int i = 0; i < descriptions.length &&
						 i < files.length &&
						 i < lineNumbers.length; i++ ) {
            String message = "";
            if (files[i] != "") {
                message += "\"" + files[i] + "\": ";
            }
            if (lineNumbers[i].intValue() != -1 && lineNumbers[i].intValue() != 0) {
                message += descriptions[i] + ", at line: " + lineNumbers[i];
            } else {
                message += descriptions[i];
            }

            if (message.startsWith("Nothing to compile.")) {
                message = "Nothing to compile, please select the project, package(s), or class(es) to compile.";
            }

            boolean isWarning = false;
            if (descriptions[i].endsWith("(warning)")) {
                compilerMessages.addMessage(message, files[i], lineNumbers[i].intValue(), 0, CompilerMessages.WARNING_MESSAGE);
            } else {
                compilerMessages.addMessage(message, files[i], lineNumbers[i].intValue(), 0, CompilerMessages.ERROR_MESSAGE);
            }
		}
    }

    /**
     * @todo    clean up this mess.
     */
    public class CompilerThread extends Thread {

        public void run() {
            //CompileProgressThread progressThread = new CompileProgressThread(this, TopManager.IDE_MANAGER.getRootFrame());

            try {
//                progressThread.startCompileProgress();
                progressMonitor.start();

                TopManager.IDE_MANAGER.hideMessages();
                long timeStart = System.currentTimeMillis();
                TopManager.IDE_MANAGER.setStatusText(" Building: " + configFile + "...");

                compilerMessages.clearMessages();
                AjdePrintWriter writer = new AjdePrintWriter(new PrintWriter(new OutputStream() {
                    public void write(int b) { } } ));

                AjdeCompiler compiler = new AjdeCompiler(TopManager.STRUCTURE_MANAGER,
                    new org.aspectj.compiler.base.ErrorHandler(writer),
                    progressMonitor);
                CompilerManager.setCompilerOptions(compiler);

                compiler.setConfigFile(configFile);
                currCompiler = compiler;
                try {
                    File config = new File(configFile);
                    if (!config.exists()) {
                        ErrorHandler.handleWarning("Config file \"" + configFile + "\" does not exist.", TopManager.IDE_MANAGER.getRootFrame());
                        return;
                    }
                    ConfigParser configParser = new ConfigParser();
                    configParser.parseConfigFile(config);
                    compiler.compile(configParser.getFiles());
                } catch (ConfigParser.ParseException pe) {
                    ErrorHandler.handleWarning("Config file entry invalid, file: " + pe.getFile().getPath() + ", line number: " + pe.getLine(), TopManager.IDE_MANAGER.getRootFrame());
                } catch (CompilerErrors ce) {
                    System.err.println("AJDE: compile failed: " + ce.toString());
                }
                CompileResult compileResult = new CompileResult(writer.getContents());
                boolean succeeded = true;
                if (compileResult.getResult() != "") {
                    System.err.println("<ajde> compile result: " + compileResult.getResult());
                    displayMessages(compileResult);
                    succeeded = false;
                }

                long timeEnd = System.currentTimeMillis();
                lastCompileTime = (int)(timeEnd - timeStart);
                notifyCompileFinished(configFile, lastCompileTime, succeeded);
            } catch (ExitRequestException ere) {
                if (ere.getValue() == 0) {
                    TopManager.IDE_MANAGER.setStatusText("Build cancelled by user.");
                } else {
                    ErrorHandler.handleError("To view message details please see the console. "
                        + "If no console is available re-launch the application from the command line.",
                        ere,
                        TopManager.IDE_MANAGER.getRootFrame());
                }
            } catch (InternalCompilerError compilerError) {
                if (compilerError.uncaughtThrowable instanceof OutOfMemoryError) {
                    ErrorHandler.handleError("Out of memory.  " +
                        "Increase memory by setting the -Xmx parameter that this VM was launched with.",
                        compilerError.uncaughtThrowable,
                        TopManager.IDE_MANAGER.getRootFrame());
                } else  if (compilerError.uncaughtThrowable instanceof MissingRuntimeError) {
                    ErrorHandler.handleWarning("Compilation aborted because the AspectJ runtime was not found. "
                        + "Please place aspectjrt.jar in the lib/ext directory.",
                        TopManager.IDE_MANAGER.getRootFrame());
                } else  if (compilerError.uncaughtThrowable instanceof BadRuntimeError) {
                    ErrorHandler.handleWarning("Compilation aborted because an out-of-date version of " +
                        "the AspectJ runtime was found.  "
                        + "Please place a current version of aspectjrt.jar in the lib/ext directory.",
                        TopManager.IDE_MANAGER.getRootFrame());
                } else {
                    ErrorHandler.handleError("Compile error.", compilerError.uncaughtThrowable,
                        TopManager.IDE_MANAGER.getRootFrame());
                }
            } catch (Throwable e) {
                ErrorHandler.handleError("Compile error, caught Throwable: " + e.toString(), e);
            } finally {
//                progressThread.stopCompileProgress();
                progressMonitor.finish();
            }
        }
    }

    public static void setCompilerOptions(AjdeCompiler compiler) {
        if (TopManager.projectProperties.getBootClasspath() != null) compiler.getOptions().bootclasspath = TopManager.projectProperties.getBootClasspath();
        if (TopManager.projectProperties.getClasspath() != null) compiler.getOptions().classpath = TopManager.projectProperties.getClasspath();
        if (TopManager.projectProperties.getOutputPath() != null) compiler.getOptions().outputDir = new File(TopManager.projectProperties.getOutputPath());
        if (TopManager.projectProperties.getWorkingDir() != null) compiler.getOptions().workingDir = new File(TopManager.projectProperties.getWorkingDir());

        String compilerFlags = TopManager.projectProperties.getCompilerFlags();
        if (compilerFlags != null && !compilerFlags.trim().equals("")) {
            StringTokenizer st = new StringTokenizer(compilerFlags, " ");
            while (st.hasMoreTokens()) {
                String flag = (String)st.nextToken();
                if (flag.equals("-source")) {
                    compiler.getOptions().source = (String)st.nextToken();
                } else {
                    compiler.getOptions().set(flag.substring(1, flag.length()), Boolean.TRUE);
                }
            }
        }
    }

    /**
     * Used for writing converting text written to an output stream into
     * a string.
     */
    public static class AjdePrintWriter extends PrintWriter {
        private String contents = "";

        public AjdePrintWriter(Writer out) {
            super(out);
        }

        public String getContents() {
            return contents;
        }

        public void flushBuffer() {
            contents = "";
            super.flush();
        }

        public void print(char x) {
            contents += x + "\n";
        }

        public void print(char[] x) {
            contents += new String( x );
        }

        public void print(int x) {
            contents += x;
        }

        public void print(String x) {
            contents += x;
        }

        public void println(char x) {
            contents += x + "\n";
        }

        public void println(char[] x) {
            contents += new String( x ) + "\n";
        }

        public void println(int x) {
            contents += x + "\n";
        }

        public void println(String x) {
            contents += x + "\n";
        }

        public void write( byte[] x ) {
            contents += new String( x );
        }

        public void write( byte[] x, int i1, int i2 ) {
            StringWriter writer = new StringWriter();
            String s = new String( x );
            writer.write( s.toCharArray(), i1, i2 );
            contents += writer.getBuffer().toString();
        }

        public void write( int c ) {
            contents += c;
        }

        public void write( String s ) {
            contents += s;
        }

        public void write( String s, int i1, int i2 ) {
            contents += s;
        }
    }

    static class CompileResult {
        private String[]  files        = null;
        private Integer[] lineNumbers  = null;
        private String[]  descriptions = null;
        private String resultString = "";
        private boolean resultContainsErrors = false;

        /**
         * Parses out warning messages, error messages, "file not found" messages, javac "Note:" messages.
         *
         * @todo    get error message structure directly from compiler
         */
        public CompileResult( String result )
        {
            resultString = result;

            BufferedReader reader = new BufferedReader( new StringReader( result ) );
            Vector fileV = new Vector();
            Vector lineV = new Vector();
            Vector descV = new Vector();
            try {
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    String originalLine = line;
                    String  description = "";
                    String  file = "";
                    Integer lineNo = new Integer(0);
                    int index = line.indexOf( ":", 2 ); // @todo    skip the initial drive ":" (fix, Windows only)
                    try {
                        if (line.indexOf("Note: ") != -1) {
                            int index1 = line.indexOf(".java");
                            if (index1 != -1) {
                                description = line.substring(index1+5) + " (warning)";
                                file = line.substring("Note: ".length(), index1+5);
                                lineNo = new Integer(0);
                            } else {
                                description = line + " (warning)";
                                file = "";
                                lineNo = new Integer(-1);
                            }
                        }
                        else if (line.indexOf("file not found: ") != -1) {
                            description = "file not found: ";
                            file = line.substring("file not found: ".length());
                            lineNo = new Integer(0);
                        }
                        else if (index != -1 && line.indexOf( "java" ) != -1) {
                            file =  line.substring( 0, index );
                            line = line.substring( index+1 );

                            index = line.indexOf( ":" );
                            lineNo = new Integer( Integer.parseInt( line.substring( 0, index ) ) ) ;
                            line = line.substring( index+1 );

                            if (!resultContainsErrors) {
                                if (!line.endsWith("(warning)")) {
                                    resultContainsErrors = true;
                                }
                            }
                            description = line.substring( line.indexOf( ":" ) + 2 );
                        }
                    } catch (Exception e) {
                        description = "Internal ajc message: " + originalLine;
                        file = "";
                        lineNo = new Integer(-1);
                    }
                    if (description.trim() != "") {
                        descV.addElement(description);
                        fileV.addElement(file);
                        lineV.addElement(lineNo);
                    }
                }
            }
            catch ( IOException ioe ) {
                resultString = "ERROR: could not parse result at line for string: " + result;
            }
            files        = new String[fileV.size()];
            lineNumbers  = new Integer[lineV.size()];
            descriptions = new String[descV.size()];
            fileV.copyInto(files);
            lineV.copyInto(lineNumbers);
            descV.copyInto(descriptions);
        }

        public String toString()
        {
            return resultString;
        }

        public String[] getfiles()
        {
            return files;
        }

        public Integer[] getLineNumbers()
        {
            return lineNumbers;
        }

        public String[] getDescriptions()
        {
            return descriptions;
        }

        public String getResult()
        {
            return resultString;
        }

        public boolean containsErrors() {
            return resultContainsErrors;
        }
    }
}