/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */


package org.aspectj.ajde.ui;

import java.util.*;
import javax.swing.JPanel;
import javax.swing.JFrame;
import java.io.*;
import org.aspectj.ajde.*;

/**
 * IDE manager for standalone AJDE application.
 *
 * @author  Mik Kersten
 */
public class AjdeManager implements IdeManager {

    private static TopFrame topFrame = null;
    private List configs = new ArrayList();

    public AjdeManager(String[] configFiles) {
        setConfigFiles(configFiles);
    }

    public void initTopFrame(JPanel browserPanel, JPanel compilerMessagesPanel, JPanel editorPanel, boolean visible) {
        topFrame = new TopFrame(browserPanel, compilerMessagesPanel, editorPanel);
        if (visible) {
            topFrame.setVisible(true);
        }
    }

    public void resetEditorFrame() {
        topFrame.resetSourceEditorPanel();
    }

    public void resetGUI() {
        topFrame.setSize(topFrame.getWidth()+1, topFrame.getHeight()+1);
        topFrame.doLayout();
        topFrame.repaint();
    }

    public void setStatusText(String text) {
        topFrame.statusText_label.setText(text);
    }

    public void setEditorStatusText(String text) {
        topFrame.setTitle(IdeManager.TITLE + " - " + text);
    }

    public void build() {
        TopManager.COMPILER_MANAGER.build();
    }

    public void editConfigFile(String configFile) {
        ConfigFileEditorPanel cfe = new ConfigFileEditorPanel(configFile);
        topFrame.setEditorPanel(cfe);
    }

    public String getDefaultConfigFile() {
        return null;
    }

    public void run() {
        Runner.invoke(TopManager.projectProperties.getClassToRun());
    }

    public void saveAll() {
        TopManager.EDITOR_MANAGER.saveContents();
    }

    public Icons getIcons() {
        return Icons.INSTANCE;
    }

    public List getConfigFiles() {
        return configs;
    }

    public void setConfigFiles(String[] configFiles) {
        for (int i = 0; i < configFiles.length; i++) {
            if (configFiles[i].endsWith(IdeManager.CONFIG_FILE_SUFFIX)) {
                configs.add(configFiles[i]);
            }
        }
    }

    public void mountConfigFile(String configFile) {
        configs.add(0, configFile);
        TopManager.BROWSER_MANAGER.updateConfigsList();
    }

    public void removeConfigFile(String configFile) {
        configs.remove(configFile);
        TopManager.BROWSER_MANAGER.updateConfigsList();
    }

    public void showMessages() {
        topFrame.showMessagesPanel();
    }

    public void hideMessages() {
        topFrame.hideMessagesPanel();
    }

    public String getDumpFilePath() {
        return new File("ajdedump.ser").getAbsolutePath();
    }

    public JFrame getRootFrame() {
        return topFrame;
    }

    private static class Runner {
        public static void run(String className) {
            try {
                Process p = Runtime.getRuntime().exec("java " + className);
            } catch(IOException ioe) {
                ErrorHandler.handleError("Coud not run: " + className, ioe, TopManager.IDE_MANAGER.getRootFrame());
            }
        }

        public static void invoke(String className) {
            try {
                System.err.println(":" + className + ":");
                if (className == null || className.length() == 0) {
                    ErrorHandler.handleWarning("No main class specified, please select a class to run.", TopManager.IDE_MANAGER.getRootFrame());
                    TopManager.showOptionsFrame(OptionsFrame.BUILD_PROPERTIES);
                } else {
                    Class[] argTypes = { String[].class };
                    java.lang.reflect.Method method = Class.forName(className).getDeclaredMethod("main", argTypes);
                    Object[] args = { new String[0] };
                    method.invoke(null, args);
                }
            } catch(ClassNotFoundException cnfe) {
                ErrorHandler.handleError("Main class not found: " + className +
                "\nMake sure that you have \".\" on your classpath.", cnfe, TopManager.IDE_MANAGER.getRootFrame());
            } catch(NoSuchMethodException nsme) {
                ErrorHandler.handleError("Class: " + className + " does not declare public static void main(String[])", nsme, TopManager.IDE_MANAGER.getRootFrame());
            } catch(java.lang.reflect.InvocationTargetException ite) {
                ErrorHandler.handleError("Could not execute: " + className, ite, TopManager.IDE_MANAGER.getRootFrame());
            } catch(IllegalAccessException iae) {
                ErrorHandler.handleError("Class: " + className + " does not declare public main method", iae, TopManager.IDE_MANAGER.getRootFrame());
            }
        }
    }
}
