/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde.views;

import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import org.aspectj.asm.*;
import org.aspectj.asm.associations.*;
import org.aspectj.ajde.*;
import org.aspectj.ajde.ui.Icons;

/**
 * Responsible for displaying and controlling the configuration and output of a
 * master and slave structure view.
 *
 * @author Mik Kersten
 */
public class BrowserManager {

    private StructureView masterView;
    private StructureView slaveView;

    private boolean globalMode = true;
    private boolean splitViewMode = false;
    private Icons icons;
    private BrowserPanel browserPanel = null;
    private Stack backHistory = new Stack();
    private Stack forwardHistory = new Stack();
    private ProgramElementNode currNode = null;

    private final ViewProperties COMPOSITION_VIEW = new ViewProperties("composition");
    private final ViewProperties CROSSCUTTING_VIEW = new ViewProperties("crosscutting");
    private final ViewProperties INHERITANCE_VIEW = new ViewProperties("inheritance");
    private final ViewProperties CUSTOM_VIEW = new ViewProperties("<custom>");
    private Map views = new HashMap();

    public BrowserManager(boolean showSlaveView) {
        this(Icons.INSTANCE);
    }

    public final StructureModelListener STRUCTURE_LISTENER = new StructureModelListener() {
            public void modelUpdated() {
                TopManager.COMPILER_MANAGER.setStructureDirty(false);
                updateView();
                resetHistory();
            }
        };

    public BrowserManager(Icons icons) {
        this.icons = icons;
        initViewProperties();
        masterView = new StructureView(icons, views, "<custom>");
        slaveView = new StructureView(icons, views, "crosscutting");
        browserPanel = new BrowserPanel(masterView, slaveView, icons);
        TopManager.STRUCTURE_MANAGER.addStructureListener(STRUCTURE_LISTENER);
    }

    public void removeStructureListener() {
        TopManager.STRUCTURE_MANAGER.removeStructureListener(STRUCTURE_LISTENER);
    }

    private void initViewProperties() {
        CROSSCUTTING_VIEW.addAssociation(Advice.METHOD_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.METHOD_CALL_SITE_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.CONSTRUCTOR_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.CONSTRUCTOR_CALL_SITE_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.HANDLER_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.INITIALIZER_RELATION.toString());
        CROSSCUTTING_VIEW.addAssociation(Advice.FIELD_ACCESS_RELATION.toString());

        INHERITANCE_VIEW.addAssociation(Inheritance.IMPLEMENTS_RELATION.toString());
        INHERITANCE_VIEW.addAssociation(Inheritance.INHERITS_MEMBERS_RELATION.toString());
        INHERITANCE_VIEW.addAssociation(Inheritance.INHERITS_RELATION.toString());

        java.util.List customAssocs = TopManager.projectProperties.getActiveAssociations();
        for (int i = 0; i < customAssocs.size(); i++) {
            CUSTOM_VIEW.addAssociation((String)customAssocs.get(i));
        }

        views.put(COMPOSITION_VIEW.getName(), COMPOSITION_VIEW);
        views.put(CROSSCUTTING_VIEW.getName(), CROSSCUTTING_VIEW);
        views.put(INHERITANCE_VIEW.getName(), INHERITANCE_VIEW);
        views.put(CUSTOM_VIEW.getName(), CUSTOM_VIEW);
    }


    public void setSplitViewMode(boolean mode) {
        splitViewMode = mode;
        browserPanel.setSlaveViewVisible(mode);
    }

    public boolean isSplitViewMode() {
        return splitViewMode;
    }

    public java.util.List getViews() {
        return new ArrayList(views.values());
    }

    public ViewProperties getViewProperties(String viewName) {
        return (ViewProperties)views.get(viewName);
    }

    public void showSourcesNodes(java.util.List nodes) {
        browserPanel.setSlaveViewVisible(false);
        for (Iterator it = nodes.iterator(); it.hasNext(); ) {
            ProgramElementNode currNode = null;
            StructureNode structureNode = (StructureNode)it.next();
            if (structureNode instanceof LinkNode) {
                currNode = ((LinkNode)structureNode).getProgramElementNode();
            } else {
                currNode = (ProgramElementNode)structureNode;
            }
            TopManager.EDITOR_MANAGER.addViewForSourceLine(currNode.getSourceFilePath(),
                currNode.getBeginLine());
        }
    }

    public void showLink(LinkNode linkNode) {
        if (browserPanel.isSlaveViewVisible()) {
            slaveView.highlightNode(linkNode.getProgramElementNode());
        }
    }

    public void showNodeInMasterView(ProgramElementNode node) {
        masterView.highlightNode(node);
    }

     public void showNodeInMasterView(LinkNode node) {
        if (currNode != null) backHistory.push(currNode);
        showNodeInMasterView(node.getProgramElementNode());
        currNode = node.getProgramElementNode();
    }

    public void showNodeInSlaveView(ProgramElementNode node) {
        slaveView.highlightNode(node);
    }

    public void navigateBackAction() {
        if (!backHistory.isEmpty()) {
            ProgramElementNode backNode = (ProgramElementNode)backHistory.pop();
            forwardHistory.push(currNode);
            masterView.highlightNode(backNode);
            currNode = backNode;
        }
    }

    public void navigateForwardAction() {
        if (!forwardHistory.isEmpty()) {
            ProgramElementNode forwardNode = (ProgramElementNode)forwardHistory.pop();
            backHistory.push(currNode);
            masterView.highlightNode(forwardNode);
            currNode = forwardNode;
        }
    }

    public void resetHistory() {
        backHistory.clear();
        forwardHistory.clear();
    }

    public void updateView() {
        if (TopManager.COMPILER_MANAGER.isStructureDirty()) {
            TopManager.STRUCTURE_MANAGER.buildStructureModel(TopManager.COMPILER_MANAGER.getCurrCompiler());
            TopManager.COMPILER_MANAGER.setStructureDirty(false);
        }

        if (globalMode) {
            updateViews(null);
        } else {
            String currFile = TopManager.EDITOR_MANAGER.getCurrFile();
            if (currFile == null || currFile.equals("")) {
                masterView.updateTree(null);
            } else {
                if (splitViewMode) {
                    slaveView.updateTree(currFile);
                } else {
                    updateViews(currFile);
                }
            }
        }
    }

    public void displayProbe() {
        masterView.displayProbe();
    }

    private void updateViews(String currFile) {
        if (currFile == null) {
            masterView.updateTree();
            slaveView.updateTree();
        } else {
            masterView.updateTree(currFile);
            slaveView.updateTree(currFile);
        }
    }

    public void dumpStructureView() {
        TopManager.STRUCTURE_MANAGER.writeStructureModel(getCurrConfigFile());
    }

    public void readStructureView() {
        TopManager.STRUCTURE_MANAGER.readStructureModel(getCurrConfigFile());
    }

    public void updateConfigsList() {
        browserPanel.updateConfigs(TopManager.IDE_MANAGER.getConfigFiles());
        browserPanel.setSelectedConfig(TopManager.COMPILER_MANAGER.getLastCompiledConfig());
    }

    Icons getIcons() {
        return icons;
    }


    public String getCurrConfigFile() {
       if (browserPanel.getSelectedConfig() != null && browserPanel.getSelectedConfig().equals("<all project files>")) {
            return TopManager.IDE_MANAGER.getDefaultConfigFile();
        } else {
            return browserPanel.getSelectedConfig();
        }
    }

    public JPanel getBrowserPanel() {
        return browserPanel;
    }

    public void setGlobalMode(boolean globalMode) {
        this.globalMode = globalMode;
        browserPanel.zoomToFile_button.setSelected(!globalMode);
    }

    public boolean isGlobalMode() {
        return globalMode;
    }

    public void extractAndInsertSignatures(java.util.List signatures, boolean calls) {
        PointcutWizard pointcutWizard = new PointcutWizard(signatures);
        pointcutWizard.setVisible(true);
        pointcutWizard.setLocation(TopManager.IDE_MANAGER.getRootFrame().getX()+100, TopManager.IDE_MANAGER.getRootFrame().getY()+100);
    }
}

interface BrowserNavigationListener extends java.util.EventListener {
    public void navigated(ProgramElementNode fromNode, ProgramElementNode toNode);
}


class PointcutWizard extends JFrame {
    private StructureView typeTreeView = null;
    private java.util.List signatures = null;

    JPanel jPanel1 = new JPanel();
    JPanel jPanel2 = new JPanel();
    JPanel jPanel4 = new JPanel();
    JLabel jLabel1 = new JLabel();
    BorderLayout borderLayout1 = new BorderLayout();
    BorderLayout borderLayout2 = new BorderLayout();
    BorderLayout borderLayout3 = new BorderLayout();
    JLabel jLabel4 = new JLabel();
    JPanel jPanel3 = new JPanel();
    JCheckBox jCheckBox5 = new JCheckBox();
    JCheckBox jCheckBox4 = new JCheckBox();
    JCheckBox jCheckBox3 = new JCheckBox();
    JCheckBox jCheckBox2 = new JCheckBox();
    JCheckBox jCheckBox1 = new JCheckBox();
    JButton cancel_button = new JButton();
    JButton ok_button = new JButton();
    JPanel jPanel5 = new JPanel();

    public PointcutWizard(java.util.List signatures) {
        this.signatures = signatures;
        typeTreeView = new StructureView(Icons.INSTANCE, getViewProperties(), "inheritance");
        typeTreeView.updateTree(TopManager.EDITOR_MANAGER.getCurrFile());
        try {
            jbInit();
        }
        catch(Exception e) {
            e.printStackTrace();
        }
        this.setSize(400, 400);
        this.setIconImage(((ImageIcon)Icons.INSTANCE.getJoinpointPublicIcon()).getImage());
    }

    private Map getViewProperties() {
        Map views = new HashMap();
        ViewProperties INHERITANCE_VIEW = new ViewProperties("inheritance");
        INHERITANCE_VIEW.addAssociation(Inheritance.INHERITS_RELATION.toString());
        views.put(INHERITANCE_VIEW.getName(), INHERITANCE_VIEW);
        return views;
    }

    private void jbInit() throws Exception {
        jLabel1.setFont(new java.awt.Font("Dialog", 0, 11));
        jLabel1.setText("Generate pointcut designator for corresponding joinpoints:");
        jPanel1.setLayout(borderLayout1);
        jPanel4.setLayout(borderLayout2);
        jPanel2.setLayout(borderLayout3);
        jLabel4.setText("Select the target type that will host the generated pointcut:");
        jLabel4.setFont(new java.awt.Font("Dialog", 0, 11));
        jLabel4.setToolTipText("");
        jPanel3.setMaximumSize(new Dimension(32767, 34));
        jCheckBox5.setEnabled(false);
        jCheckBox5.setFont(new java.awt.Font("Dialog", 0, 11));
        jCheckBox5.setSelected(true);
        jCheckBox5.setText("call");
        jCheckBox4.setEnabled(false);
        jCheckBox4.setFont(new java.awt.Font("Dialog", 0, 11));
        jCheckBox4.setText("execution");
        jCheckBox3.setEnabled(false);
        jCheckBox3.setFont(new java.awt.Font("Dialog", 0, 11));
        jCheckBox3.setText("initialization");
        jCheckBox2.setEnabled(false);
        jCheckBox2.setFont(new java.awt.Font("Dialog", 0, 11));
        jCheckBox2.setText("static initialization");
        jCheckBox1.setEnabled(false);
        jCheckBox1.setFont(new java.awt.Font("Dialog", 0, 11));
        jCheckBox1.setText("field get/set");
        cancel_button.setFont(new java.awt.Font("Dialog", 0, 11));
        cancel_button.setText("Cancel");
        cancel_button.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                cancel_button_actionPerformed(e);
            }
        });
        ok_button.setText("OK");
        ok_button.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ok_button_actionPerformed(e);
            }
        });
        ok_button.setFont(new java.awt.Font("Dialog", 0, 11));
        this.setTitle("Pointcut Wizard");
        this.getContentPane().add(jPanel1, BorderLayout.CENTER);
        jPanel1.add(jPanel4, BorderLayout.NORTH);
        jPanel4.add(jLabel1,  BorderLayout.NORTH);
        jPanel4.add(jPanel3, BorderLayout.CENTER);
        jPanel3.add(jCheckBox5, null);
        jPanel3.add(jCheckBox4, null);
        jPanel3.add(jCheckBox3, null);
        jPanel3.add(jCheckBox2, null);
        jPanel3.add(jCheckBox1, null);
        jPanel1.add(jPanel2,  BorderLayout.CENTER);
        jPanel2.add(jLabel4, BorderLayout.NORTH);
        jPanel2.add(typeTreeView,  BorderLayout.CENTER);
        jPanel1.add(jPanel5,  BorderLayout.SOUTH);
        jPanel5.add(ok_button, null);
        jPanel5.add(cancel_button, null);
    }

    private void ok_button_actionPerformed(ActionEvent e) {
        TopManager.EDITOR_MANAGER.pasteToCaretPos(generatePcd());
        this.dispose();
    }

    private void cancel_button_actionPerformed(ActionEvent e) {
        this.dispose();
    }

    private String generatePcd() {
        String pcd = "\n\n" +
            "    pointcut temp(): \n";
        for (Iterator it = signatures.iterator(); it.hasNext(); ) {
            pcd += "        call(* " + it.next() + ")";
            if (it.hasNext()) {
                pcd += " ||";
            } else {
                pcd += ";";
            }
            pcd += "\n";
        }
        return pcd;
//            "        call(void Point.setX(int)) ||\n" +
//            "call(void Point.setY(int)) ||\n" +
//            "call(void Line.setP1(Point)) ||\n" +
//            "call(void Line.setP2(Point))";
    }
}