
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.ajde.views;

import java.util.*;
import java.awt.Component;
import java.awt.Color;
import java.awt.Font;
import java.awt.font.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import org.aspectj.ajde.compiler.AjdeCompiler;
import org.aspectj.ajde.TopManager;
import org.aspectj.asm.*;
import org.aspectj.asm.associations.*;
import org.aspectj.ajde.ui.Icons;

/**
 * @todo    resolve accessiblity/visibility name schitzophrenia
 * @author  Mik Kersten
 */
class StructureTreeManager {

    private List associations = null;
    private List modifiers = null;
    private List visibility = null;
    private List ordering = null;
    private List filteredMemberKinds = null;
    private String hierarchy = null;
    private String granularity = null;

    private StructureTree structureTree;
    private ViewProperties viewProperties;
    private boolean sortNodes;
    private StructureNodeComparator structureNodeComparator = new StructureNodeComparator();
    private String rootFile = null;
    private Icons icons = null;
    private StructureTreeCellRenderer cellRenderer = null;
    private StructureTreeListener treeListener = null;
    private ProgramElementNode currNode = null;

    private final StructureTreeModel NO_STRUCTURE_MODEL
        = new StructureTreeModel(new StructureTreeNode(new StructureNode("", "")));
    private final StructureTreeModel NO_FILE_SELECTED_MODEL
        = new StructureTreeModel(new StructureTreeNode(new StructureNode("open a file to view structure...", "")));

    public StructureTreeManager(Icons icons, ViewProperties viewProperties) {
        this.icons = icons;
        this.viewProperties = viewProperties;
        this.associations = viewProperties.getAssociations();
        structureTree = new StructureTree();
        structureTree.setModel(NO_STRUCTURE_MODEL);
        treeListener = new StructureTreeListener(structureTree);
        cellRenderer = new StructureTreeCellRenderer(icons);
        structureTree.setCellRenderer(cellRenderer);
        structureTree.addTreeSelectionListener(treeListener);
        structureTree.addMouseListener(treeListener);
    }

    public void setActiveView(ViewProperties viewProperties) {
        this.viewProperties = viewProperties;
        this.associations = viewProperties.getAssociations();
    }

    public StructureNode getSelectedStructureNode() {
        return (StructureNode)((StructureTreeNode)structureTree.getLastSelectedPathComponent()).getUserObject();
    }

    public void scrollToHighlightedNode() {
        structureTree.scrollPathToVisible(structureTree.getSelectionPath());
    }

    private void highlightNode(StructureTreeNode parent, ProgramElementNode node) {
        for (int i = 0; i < parent.getChildCount(); i++) {
            StructureTreeNode currNode = (StructureTreeNode)parent.getChildAt(i);
            StructureNode sNode = (StructureNode)currNode.getUserObject();
            if (sNode instanceof ProgramElementNode &&
                ((ProgramElementNode)sNode).equals(node)) {
                TreePath path = new TreePath(currNode.getPath());
                structureTree.setSelectionPath(path);
                int currRow = structureTree.getRowForPath(path);
                structureTree.expandRow(currRow);
//                structureTree.expandRow(currRow+2);
//                if (!structureTree.isVisible(path)) {
                    structureTree.scrollRowToVisible(currRow);
//                } else {
//                    highlightNode(currNode, node);
//                }
//                structureTree.scrollRowToVisible(currRow-1);
            } else {
                highlightNode(currNode, node);
            }
        }
    }

    void updateTree(int depth, String hierarchy) {
        updateTree(TopManager.STRUCTURE_MANAGER.getStructureModel(), depth, hierarchy);
    }

    void updateTree(String filePath, int depth, String hierarchy) {
        if (filePath == null || filePath.equals("")) {
            structureTree.setModel(NO_FILE_SELECTED_MODEL);
        } else {
            structureTree.setRootFilePath(filePath);
            updateTree(TopManager.STRUCTURE_MANAGER.getStructureModel(filePath), depth, hierarchy);
        }
    }

    private void updateTree(StructureNode rootNode, int depth, String hierarchy) {
        this.hierarchy = hierarchy;
        modifiers = TopManager.projectProperties.getActiveModifiers();
        visibility = TopManager.projectProperties.getActiveVisibility();
        ordering = TopManager.projectProperties.getStructureViewOrdering();
        filteredMemberKinds = TopManager.projectProperties.getActiveFilteredMemberKinds();
        granularity = TopManager.projectProperties.getActiveGranularity();
        if (granularity == null) granularity = "member";
        displayTree(rootNode, depth);
    }

    StructureTree getStructureTree() {
        return structureTree;
    }

    void displayProbe() {
        StructureNode rootNode = TopManager.STRUCTURE_MANAGER.getStructureModel();
        for (Iterator it = rootNode.getChildren().iterator(); it.hasNext(); ) {
            StructureNode currNode = (StructureNode)it.next();
            if (currNode instanceof ProgramElementNode) {
                ProgramElementNode probe = (ProgramElementNode)currNode;
                displayTree((StructureNode)probe.getChildren().get(0), 4);
                return;
            }
        }
    }

    /**
     * @param   node    if null assume root
     */
    public void navigationAction(ProgramElementNode node, boolean followedLink, boolean forward) {
        if (node == null) {
            structureTree.setSelectionRow(0);
            structureTree.scrollRowToVisible(0);
        } else if (node.getSourceFilePath() != null) {
            final String fileName = node.getSourceFilePath();
            final int lineNumber = node.getBeginLine();
            if (fileName != null && lineNumber > 0) {
                Runnable update = new Runnable() {
                    public void run() {
                        org.aspectj.ajde.TopManager.EDITOR_MANAGER.showSourceLine(fileName, lineNumber, true);
                   }
                };

                if (SwingUtilities.isEventDispatchThread()) {
                    update.run();
                } else {
                    try {
                        SwingUtilities.invokeAndWait(update);
                    } catch (Exception ee) {
                        ee.printStackTrace();
                    }
                }
            }
            if (followedLink) {
                highlightNode((StructureTreeNode)structureTree.getModel().getRoot(), node);
//                TopManager.BROWSER_MANAGER.navigatedAction(node, forward);
            }
        }
    }

    public ProgramElementNode getRootProgramElementNode() {
        StructureNode node = (StructureNode)((StructureTreeNode)structureTree.getModel().getRoot()).getUserObject();
        if (node instanceof ProgramElementNode) {
            return (ProgramElementNode)node;
        } else {
            return null;
        }
    }

    private void displayTree(final StructureNode rootNode, final int depth) {
        if (viewProperties.getName().equals("<custom>")) {
            associations = TopManager.projectProperties.getActiveAssociations();
        }
        if (TopManager.projectProperties.getStructureViewOrdering().contains("sort")) {
            sortNodes = true;
        }

        Runnable update = new Runnable() {
            public void run() {
                StructureTreeNode rootTreeNode = null;
                if (hierarchy.equals("crosscutting view") || hierarchy.equals("inheritance view")) {
                    rootTreeNode = buildCustomTree();
                } else {
                    rootTreeNode = buildTree(rootNode);
                }

                structureTree.removeAll();
                TreeModel treeModel = new StructureTreeModel(rootTreeNode);
                structureTree.setModel(treeModel);
                expandTree(depth);
           }
        };

        if (SwingUtilities.isEventDispatchThread()) {
            update.run();
        } else {
            try {
                SwingUtilities.invokeAndWait(update);
            } catch (Exception ee) {
                ee.printStackTrace();
            }
        }
    }

    /**
     * @todo    HACK: this is a workaround and can break
     */
    private static ProgramElementNode mapResult = null;
    private ProgramElementNode getNodeForLink(LinkNode node, StructureNode rootNode) {
        ProgramElementNode result = null;
        if (rootNode instanceof ProgramElementNode &&
            ((ProgramElementNode)rootNode).getName().equals(node.getProgramElementNode().getName())) {
            mapResult = (ProgramElementNode)rootNode;
        } else {
            ProgramElementNode linkedNode = node.getProgramElementNode();
            for (Iterator it = rootNode.getChildren().iterator(); it.hasNext(); ) {
                StructureNode child = (StructureNode)it.next();
                getNodeForLink(node, child);
            }
        }
        return mapResult;
    }


    private StructureTreeNode buildTree(StructureNode node) {
        StructureTreeNode treeNode = new StructureTreeNode(node);
        if (node instanceof ProgramElementNode) {
            List relations = ((ProgramElementNode)node).getRelations();
            if (relations != null) {
                for (Iterator it = relations.iterator(); it.hasNext(); ) {
                    RelationNode relationNode = (RelationNode)it.next();
                    if (associations.contains(relationNode.getRelation().toString())) {
                        treeNode.add(buildTree(relationNode));
                    }
                }
            }
        }
        if (node != null) {
            List children = null;
            children = node.getChildren();
            if (children != null) {
                List childList = new ArrayList();
                for (Iterator itt = children.iterator(); itt.hasNext(); ) {
                    StructureNode child = (StructureNode)itt.next();
                    if (child instanceof ProgramElementNode) {
                        ProgramElementNode progNode = (ProgramElementNode)child;
                        if (acceptNode(progNode)) {
                            childList.add(buildTree(child));
                        }
                    } else {
                        childList.add(buildTree(child));
                    }
                }
                sortNodes(childList);
                for (Iterator it = childList.iterator(); it.hasNext(); ) {
                    treeNode.add((StructureTreeNode)it.next());
                }
            }

        }
        return treeNode;
    }

    private StructureTreeNode buildCustomTree() {
        StructureNode rootNode = TopManager.STRUCTURE_MANAGER.getStructureModel();
        StructureTreeNode treeNode = new StructureTreeNode(rootNode);

        List rootNodes = new ArrayList();
        getRoots(rootNode, rootNodes);

        for (Iterator it = rootNodes.iterator(); it.hasNext(); ) {
            if (hierarchy.equals("crosscutting view")) {
                treeNode.add(getCrosscuttingChildren((StructureNode)it.next()));
            } else if (hierarchy.equals("inheritance view")) {
                treeNode.add(getInheritanceChildren((StructureNode)it.next()));
            }
        }
        return treeNode;
    }

    private void getRoots(StructureNode rootNode, List roots) {
        if (rootNode.getChildren() != null) {
            for (Iterator it = rootNode.getChildren().iterator(); it.hasNext(); ) {
                StructureNode node = (StructureNode)it.next();
                if (node instanceof ProgramElementNode) {
                    if (acceptNodeAsRoot((ProgramElementNode)node)) {
                        ProgramElementNode pNode = (ProgramElementNode)node;
                        List relations = pNode.getRelations();
                        String delimiter = "";
                        if (hierarchy.equals("crosscutting view")) {
                            delimiter = "uses pointcut";
                        } else if (hierarchy.equals("inheritance view")) {
                            delimiter = "inherits";
                        }
                        if (relations != null && relations.toString().indexOf(delimiter) == -1) {
                            boolean found = false;
                            for (Iterator it2 = roots.iterator(); it2.hasNext(); ) {
                                if (((ProgramElementNode)it2.next()).equals(pNode)) found = true;
                            }
                            if (!found) roots.add(pNode);
                        }
                    }
                }
                getRoots(node, roots);
            }
        }
    }

    public boolean acceptNodeAsRoot(ProgramElementNode node) {
        if (hierarchy.equals("crosscutting view")) {
            return node.getKind().equals(ProgramElementNode.KIND_ADVICE)
                || node.getKind().equals(ProgramElementNode.KIND_POINTCUT);
        } else if (hierarchy.equals("inheritance view")) {
            return node.getKind().equals(ProgramElementNode.KIND_CLASS);
        } else {
            return false;
        }
    }

    private StructureTreeNode getInheritanceChildren(StructureNode node) {
        StructureTreeNode treeNode = new StructureTreeNode(node);
        List relations = ((ProgramElementNode)node).getRelations();
        if (relations != null) {
            for (Iterator it = relations.iterator(); it.hasNext(); ) {
                RelationNode relation = (RelationNode)it.next();
                if (relation.getName().equals("is inherited by")) {
                    for (Iterator it2 = relation.getChildren().iterator(); it2.hasNext(); ) {
                        ProgramElementNode pNode = ((LinkNode)it2.next()).getProgramElementNode();
                        StructureTreeNode newNode = getInheritanceChildren(pNode);
                        StructureTreeNode typeChildren = buildTree((StructureNode)newNode.getUserObject());
                        for (int i = 0; i < typeChildren.getChildCount(); i++) {
                            newNode.add((StructureTreeNode)typeChildren.getChildAt(i));
                        }

                        treeNode.add(newNode);
                    }
                }
            }
        }
        return treeNode;
    }

    private StructureTreeNode getCrosscuttingChildren(StructureNode node) {
        StructureTreeNode treeNode = new StructureTreeNode(node);
        List relations = ((ProgramElementNode)node).getRelations();
        //System.err.println(node + ": " + relations);
        if (relations != null) {
            for (Iterator it = relations.iterator(); it.hasNext(); ) {
                RelationNode relation = (RelationNode)it.next();
                if (relation.getName().equals("pointcut used by")) {
                    for (Iterator it2 = relation.getChildren().iterator(); it2.hasNext(); ) {
                        ProgramElementNode pNode = ((LinkNode)it2.next()).getProgramElementNode();
                        StructureTreeNode newNode = getCrosscuttingChildren(pNode);
                        for (Iterator it3 = pNode.getRelations().iterator(); it3.hasNext(); ) {
                            RelationNode relationNode = (RelationNode)it3.next();
                            if (relationNode.getName().indexOf("pointcut") == -1) {
                                newNode.add(getRelations(relationNode));
                            }
                        }
                        treeNode.add(newNode);
                    }
                } else if (relations.toString().indexOf("uses pointcut") == -1) {
                    for (Iterator it4 = relations.iterator(); it4.hasNext(); ) {
                        RelationNode relationNode = (RelationNode)it4.next();
                        if (relationNode.getName().indexOf("pointcut") == -1) {
                            treeNode.add(getRelations(relationNode));
                        }
                    }
                }
            }
        }
        return treeNode;
    }

    private StructureTreeNode getRelations(RelationNode node) {
        StructureTreeNode treeNode = new StructureTreeNode(node);
        for (Iterator it = node.getChildren().iterator(); it.hasNext(); ) {
            treeNode.add(new StructureTreeNode((StructureNode)it.next()));
        }
        return treeNode;
    }

    private void sortNodes(List nodes) {
        if (sortNodes) {
            Collections.sort(nodes, structureNodeComparator);
        }
    }

    private class StructureNodeComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            StructureNode t1 = (StructureNode)((StructureTreeNode)o1).getUserObject();
            StructureNode t2 = (StructureNode)((StructureTreeNode)o2).getUserObject();
            if (t1 instanceof RelationNode) {
                return -1;
            } else {
                return 1;
            }
        }
    }

    /**
     * Represents the node filtering mechanism.
     *
     * @todo    stop using strings, fix flow
     */
    private boolean acceptNode(ProgramElementNode node) {
        if (node.getKind().equals("package")) return true;

        if (node.getKind().equals("file")) {
            if (granularity.equals("package") || node.getName().equals("_JoinpointProbe.java")) {
                return false;
            } else {
                return true;
            }
        }

        if (node.getKind().equals("class") || node.getKind().equals("aspect") || node.getKind().equals("interface")) {
            if (granularity.equals("file") || granularity.equals("package")) {
                return false;
            } else {
                return true;
            }
        }

        if (node.isMemberKind()) {
            if (granularity.equals("member")) {
                for (Iterator it = modifiers.iterator(); it.hasNext(); ) {
                    if (node.getModifiers().contains((String)it.next())) return false;
                }
                for (Iterator it2 = visibility.iterator(); it2.hasNext(); ) {
                    if (node.getAccessibility().contains((String)it2.next()))  return false;
                }
                if (filteredMemberKinds.contains(node.getKind())) {
                    return false;
                } else {
                    return true;
                }
            } else {
                return false;
            }
        }

        if (node.isCode()) return false;

        return false;
    }

    private void expandTree(int depth) {
        for (int i = 0; i < structureTree.getRowCount(); i++) {
            TreePath path = structureTree.getPathForRow(i);
            StructureTreeNode node = (StructureTreeNode)path.getLastPathComponent();
            if (path.getPath().length-1 > depth || node.getUserObject() instanceof RelationNode) {
                structureTree.collapsePath(path);
            } else {
                structureTree.expandPath(path);
            }
        }
        structureTree.expandPath(structureTree.getPathForRow(0));
    }

    private class StructureTreeNode extends DefaultMutableTreeNode {
        private StructureNode nodeInfo = null;

        public StructureTreeNode(StructureNode nodeInfo) {
            super(nodeInfo, true);
            this.nodeInfo = nodeInfo;
        }

        public String toString() {
            if (nodeInfo != null) {
                return nodeInfo.toString();
            } else {
                return "";
            }
        }
    }

    private class StructureTreeModel extends DefaultTreeModel implements TreeModel {
        public StructureTreeModel(TreeNode newRoot) {
            super(newRoot);
        }

        public void valueForPathChanged(TreePath path, Object newValue) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode)path.getLastPathComponent();
            nodeChanged(node);
        }
    }

    private class StructureTreeCellRenderer extends DefaultTreeCellRenderer {

        private Icons icons = null;

        public StructureTreeCellRenderer(Icons icons) {
            this.icons = icons;
        }

        /**
         * @todo:   probably should get rid of "instanceof" tests
         */
        public Component getTreeCellRendererComponent(JTree tree,
                                                        Object treeNode,
                                                        boolean sel,
                                                        boolean expanded,
                                                        boolean leaf,
                                                        int row,
                                                        boolean hasFocus) {
            if (treeNode == null) { // instanceof test here?
                return null;
            }
            StructureNode node = (StructureNode)((StructureTreeNode)treeNode).getUserObject();

            if (node instanceof LinkNode) {
                if (((LinkNode)node).getProgramElementNode().getSourceFilePath() == null) {
                    setTextNonSelectionColor(new Color(150, 150, 255));
                } else {
                    setTextNonSelectionColor(new Color(0, 0, 255));
                }
            }

            super.getTreeCellRendererComponent(tree, treeNode, sel, expanded, leaf, row, hasFocus);
            setTextNonSelectionColor(new Color(0, 0, 0));
            this.setFont(StructureTree.DEFAULT_FONT);

            if (node instanceof ProgramElementNode) {
                ProgramElementNode lexicalNode = (ProgramElementNode)node;
                setIcon(getProgramElementNodeIcon(lexicalNode));
            } else if (node instanceof RelationNode) {
                RelationNode relationNode = (RelationNode)node;
                if (relationNode.getKind().equals(org.aspectj.asm.associations.Advice.NAME) ||
                    relationNode.getKind().equals(org.aspectj.asm.associations.Introduction.NAME)) {
                    if (relationNode.getRelation().getBackNavigationName().equals(relationNode.getName()) ){
                        setIcon(icons.getRelationAdviceBackIcon());
                    } else {
                        setIcon(icons.getRelationAdviceForwardIcon());
                    }
                } else if (relationNode.getKind().equals(org.aspectj.asm.associations.Inheritance.NAME)) {
                    if (relationNode.getRelation().getBackNavigationName().equals(relationNode.getName()) ){
                        setIcon(icons.getRelationInheritanceBackIcon());
                    } else {
                        setIcon(icons.getRelationInheritanceForwardIcon());
                    }
                }
                else {
                    if (relationNode.getRelation().getBackNavigationName().equals(relationNode.getName()) ){
                        setIcon(icons.getRelationReferenceBackIcon());
                    } else {
                        setIcon(icons.getRelationReferenceForwardIcon());
                    }
                }
		        this.setFont(new Font(this.getFont().getName(), Font.ITALIC, this.getFont().getSize()));

            } else if (node instanceof LinkNode) {
                LinkNode link = (LinkNode)node;
                setIcon(getProgramElementNodeIcon(link.getProgramElementNode()));
            } else {
                if (node != null && node.getKind().equals("package")) {
                    setIcon(icons.getPackageIcon());
                } else if (node != null && node.getKind().equals("project")) {
                    setIcon(icons.getProjectIcon());
                } else if (node != null && node.getKind().equals("file")) {
                    setIcon(icons.getFileClassIcon());
                } else {
                    setIcon(null);
                }
            }
            return this;
        }

        private Icon getProgramElementNodeIcon(ProgramElementNode pNode) {
            int accessibility = 0;
            if (pNode.getAccessibility().contains(ProgramElementNode.ACCESSIBILITY_PUBLIC)) {
                accessibility = 1;
            } else if (pNode.getAccessibility().contains(ProgramElementNode.ACCESSIBILITY_PROTECTED)) {
                accessibility = 2;
            } else if (pNode.getAccessibility().contains(ProgramElementNode.ACCESSIBILITY_PRIVATE)) {
                accessibility = 3;
            } else if (pNode.getAccessibility().contains(ProgramElementNode.ACCESSIBILITY_PRIVILEGED)) {
                accessibility = 3;
            }

            if (pNode == null || pNode.getKind() == null) {
                return null;
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_PROJECT)) {
                return icons.getProjectIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_PACKAGE)) {
                return icons.getPackageIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_FILE)) {
                return icons.getFileClassIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_CLASS)) {
                switch (accessibility) {
                    case 1: return icons.getClassPublicIcon();
                    case 2: return icons.getClassProtectedIcon();
                    case 3: return icons.getClassPrivateIcon();
                    default: return icons.getClassPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_INTERFACE)) {
                switch (accessibility) {
                    case 1: return icons.getInterfacePublicIcon();
                    case 2: return icons.getInterfaceProtectedIcon();
                    case 3: return icons.getInterfacePrivateIcon();
                    default: return icons.getInterfacePackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_ASPECT)) {
                switch (accessibility) {
                    case 1: return icons.getAspectPublicIcon();
                    case 2: return icons.getAspectProtectedIcon();
                    case 3: return icons.getAspectPrivateIcon();
                    case 4: return icons.getAspectPrivilegedIcon();
                    default: return icons.getAspectPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_METHOD)
                || pNode.getKind().equals(ProgramElementNode.KIND_INITIALIZER)
                || pNode.getKind().equals(ProgramElementNode.KIND_CONSTRUCTOR)) {
                switch (accessibility) {
                    case 1: return icons.getMethodPublicIcon();
                    case 2: return icons.getMethodProtectedIcon();
                    case 3: return icons.getMethodPrivateIcon();
                    default: return icons.getMethodPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_FIELD)) {
                switch (accessibility) {
                    case 1: return icons.getFieldPublicIcon();
                    case 2: return icons.getFieldProtectedIcon();
                    case 3: return icons.getFieldPrivateIcon();
                    default: return icons.getFieldPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_INTRODUCTION)) {
                switch (accessibility) {
                    case 1: return icons.getIntroductionPublicIcon();
                    case 2: return icons.getIntroductionProtectedIcon();
                    case 3: return icons.getIntroductionPrivateIcon();
                    default: return icons.getIntroductionPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_POINTCUT)) {
                switch (accessibility) {
                    case 1: return icons.getJoinpointPublicIcon();
                    case 2: return icons.getJoinpointProtectedIcon();
                    case 3: return icons.getJoinpointPrivateIcon();
                    default: return icons.getJoinpointPackageIcon();
                }
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_ADVICE)) {
               return icons.getAdviceIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_DECLARE_PARENTS)) {
               return icons.getDeclareParentsIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_DECLARE_ERROR)) {
               return icons.getDeclareErrorIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_DECLARE_WARNING)) {
               return icons.getDeclareWarningIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_DECLARE_SOFT)) {
               return icons.getDeclareSoftIcon();
            } else if (pNode.getKind().equals(ProgramElementNode.KIND_CODE)) {
               return icons.getCodeIcon();
            } else {
                System.err.print("> AJDE Warning: unknown node kind - " + pNode.getKind());
                return null;
            }
        }
    }

    private class StructureTreeListener implements TreeSelectionListener, MouseListener {
        private StructureTree tree = null;

        public StructureTreeListener(StructureTree tree) {
            this.tree = tree;
        }

        public void valueChanged(TreeSelectionEvent e) { }

        public void mouseEntered(MouseEvent e) { }

        public void mouseExited(MouseEvent e) { }

        public void mousePressed(MouseEvent e) { }

        public void mouseReleased(MouseEvent e) { }

        public void mouseClicked(MouseEvent e) {
            singleClickNavigation(e);
//            doubleClickNavigation(e);
            maybeShowPopup(e);
        }

        public void singleClickNavigation(MouseEvent e) {
            StructureTreeNode treeNode = (StructureTreeNode)tree.getLastSelectedPathComponent();
            if (treeNode != null && !e.isControlDown() && !e.isShiftDown() && e.getModifiers() != 4) {
                StructureNode currNode = (StructureNode)treeNode.getUserObject();
                if (currNode instanceof ProgramElementNode && !e.isControlDown()
                    && !e.isShiftDown() && e.getModifiers() != 4) {
                    TopManager.BROWSER_MANAGER.showNodeInMasterView((ProgramElementNode)currNode);
                    if (TopManager.BROWSER_MANAGER.isSplitViewMode()) {
                        TopManager.BROWSER_MANAGER.showNodeInSlaveView((ProgramElementNode)currNode);
                    }
                } else if (currNode instanceof LinkNode) {
                    if (!TopManager.BROWSER_MANAGER.isSplitViewMode()) {
                        TopManager.BROWSER_MANAGER.showNodeInMasterView((LinkNode)currNode);
                    } else {
                        TopManager.BROWSER_MANAGER.showNodeInSlaveView(((LinkNode)currNode).getProgramElementNode());
                    }
               }
            }
        }

        public void doubleClickNavigation(MouseEvent e) {
            int clickCount = e.getClickCount();
            StructureTreeNode treeNode = (StructureTreeNode)tree.getLastSelectedPathComponent();
            if (treeNode != null) {
                StructureNode currNode = (StructureNode)treeNode.getUserObject();
                if (currNode instanceof ProgramElementNode && !e.isControlDown() && !e.isShiftDown()
                    && e.getModifiers() != 4) {
                    TopManager.BROWSER_MANAGER.showNodeInMasterView(((LinkNode)currNode).getProgramElementNode());
                    TopManager.BROWSER_MANAGER.showNodeInSlaveView(((LinkNode)currNode).getProgramElementNode());
                } else if (currNode instanceof LinkNode) {
                    if (clickCount == 1) {
                        TopManager.BROWSER_MANAGER.showLink((LinkNode)currNode);
                    } else if (clickCount == 2) {
                        navigationAction((ProgramElementNode)((LinkNode)currNode).getProgramElementNode(), true, true);
                    }
                }
            }
        }

        /**
         * @todo    this should probably use <CODE>e.isPopupTrigger()</CODE> but that
         * doesn't work for some reason, so we just check if the right mouse button
         * has been clicked.
         */
        private void maybeShowPopup(MouseEvent e) {
            if (e.getModifiers() == e.BUTTON3_MASK && tree.getSelectionCount() > 0) {
                TreePath[] selectionPaths = tree.getSelectionPaths();
                final List signatures = new ArrayList();
                for (int i = 0; i < selectionPaths.length; i++) {
                    StructureNode currNode = (StructureNode)((StructureTreeNode)selectionPaths[i].getLastPathComponent()).getUserObject();
                    if (currNode instanceof LinkNode || currNode instanceof ProgramElementNode) {
                        signatures.add(currNode);
                    }
                }

                JPopupMenu popup = new JPopupMenu();
                JMenuItem showSourcesItem = new JMenuItem("Display sources", icons.getCodeIcon());
                showSourcesItem.setFont(new java.awt.Font("Dialog", 0, 11));
                showSourcesItem.addActionListener(new AbstractAction() {
                    public void actionPerformed(ActionEvent e) {
                        TopManager.BROWSER_MANAGER.showSourcesNodes(signatures);
                    }
                });
                popup.add(showSourcesItem);

                popup.addSeparator();
                JMenuItem generatePCD = new JMenuItem("Pointcut Wizard (alpha)...", icons.getJoinpointPublicIcon());
                generatePCD.setFont(new java.awt.Font("Dialog", 0, 11));
                generatePCD.addActionListener(new AbstractAction() {
                    public void actionPerformed(ActionEvent e) {
                        TopManager.BROWSER_MANAGER.extractAndInsertSignatures(signatures, true);
                    }
                });
                popup.add(generatePCD);

                popup.show(e.getComponent(), e.getX(), e.getY());
            }
        }
    }
}

