
/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.asm;

import java.util.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;

/**
 * @author Mik Kersten
 */
public class StructureNodeFactory {

    private static Hashtable programElementNodes = new Hashtable();

    public static void clear() {
        programElementNodes.clear();
    }

    public static ProgramElementNode makeNode(ASTObject astObject, List relations, List children) {
        return makeNode(astObject, relations, children, false);
    }

    public static LinkNode makeLink(ASTObject astObject, boolean terminal) {
        if (terminal) {
            return new LinkNode(makeNode(astObject, null, null, false));
        } else {
            return new LinkNode(makeNode(astObject, null, null, true));
        }
    }

    private static ProgramElementNode makeNode(ASTObject astObject, List relations, List children, boolean resolve) {
        if (resolve) {
            if (astObject instanceof InitializerDec) {
                InitializerDec initDec = (InitializerDec)astObject;
                return (ProgramElementNode)programElementNodes.get(initDec.getDeclaringType().getTypeDec());
            } else if (astObject instanceof Decs) {
                Decs decs = (Decs)astObject;
                return (ProgramElementNode)programElementNodes.get(decs.getDeclaringType().getTypeDec());
            } else {
                ProgramElementNode peNode = (ProgramElementNode)programElementNodes.get(astObject);
                if (peNode == null) {
                    return makeNode(astObject, null, null, false);
                } else {
                    return peNode;
                }
            }
        } else {
            String declaringType = "";
            if (astObject.getDeclaringType() != null) {
                declaringType = astObject.getDeclaringType().toShortString();
            }
            ProgramElementNode newNode = new ProgramElementNode(
                genSignature(astObject).trim(),
                genKind(astObject),
                genModifiers(astObject),
                genAccessibility(astObject),
                declaringType,
                genPackageName(astObject),
                genFormalComment(astObject),
                astObject.getSourceLocation().getSourceFileName(),
                astObject.getSourceLocation().getBeginLine(),
                relations,
                children,
                isMemberKind(astObject),
                astObject);
            programElementNodes.put(astObject, newNode);
            return newNode;
        }
    }

    private static boolean genIsStmntKind(ASTObject astObject) {
        return astObject instanceof CatchClause || astObject instanceof SOLink;
    }

    private static List genModifiers(ASTObject astObject) {
        List modifiers = new ArrayList();
        if (astObject instanceof Dec) {
            Dec dec = (Dec)astObject;
            if (dec.getModifiers().isStrict()) modifiers.add("strictfp");
            if (dec.getModifiers().isAbstract()) modifiers.add("abstract");
            if (dec.getModifiers().isSynchronized()) modifiers.add("synchronized");
            if (dec.getModifiers().isNative()) modifiers.add("native");
            if (dec.getModifiers().isFinal()) modifiers.add("final");
            if (dec.getModifiers().isTransient()) modifiers.add("transient");
            if (dec.getModifiers().isStatic()) modifiers.add("static");
            if (dec.getModifiers().isVolatile()) modifiers.add("volatile");
        }
        return modifiers;
    }

    private static List genAccessibility(ASTObject astObject) {
        List modifiers = new ArrayList();
        if (astObject instanceof Dec) {
            Dec dec = (Dec)astObject;
            if (dec.getModifiers().isPublic()) modifiers.add("public");
            if (dec.getModifiers().isProtected()) modifiers.add("protected");
            if (dec.getModifiers().isPrivileged()) modifiers.add("privileged");
            if (dec.getModifiers().isPackagePrivate()) modifiers.add("package");
            if (dec.getModifiers().isPrivate()) modifiers.add("private");
        }
        return modifiers;
    }

    /**
     * @todo	special cases should be fixes to AST nodes, this should have no instanceof tests.
     * @todo	get rid of "<undefined>" string
     */
    private static String genKind(ASTObject astObject) {
        if (astObject instanceof CompilationUnit) {
            return "file";
        } else if (genIsStmntKind(astObject)) {
            return "decBodyElement";
        } else if (astObject instanceof Dec) {
            String kind = ((Dec)astObject).getKind();
//            if (kind != null) {
//                if (kind.startsWith("declare")) {
//                    return "declare";
//                } else {
                    return ((Dec)astObject).getKind();
//                }
//            } else {
//                return "<undefined>";
//            }
        } else {
            return "<undefined>";
        }
    }

    private static boolean isMemberKind(ASTObject astObject) {
        if (astObject instanceof Dec) {
            Dec dec = (Dec)astObject;
            return dec.getDeclaringType() != null && !dec.getDeclaringType().equals(dec.getName());
        } else {
            return false;
        }
    }

    private static String genPackageName(ASTObject astObject) {
        if (astObject instanceof TypeDec) {
            return ((TypeDec)astObject).getPackageName();
        } else if (astObject instanceof CompilationUnit) {
            return ((CompilationUnit)astObject).getPackageName();
        } else if (astObject.getDeclaringType() != null) {
            return astObject.getDeclaringType().getPackageName();
        } else {
            return "";
        }
    }

    private static String genDeclaringType(ASTObject astObject) {
        if (astObject != null && astObject.getDeclaringType() != null) {
            return astObject.getDeclaringType().toShortString();
        } else {
            return null;
        }
    }

    /**
     * Tries to return the ajdoc generated comment, otherwise returns the raw comment.
     */
    private static String genFormalComment(ASTObject astObject) {
        try {
            return (String)astObject.getComment().getClass().getMethod("commentText", new Class[]{}).invoke(astObject.getComment(), new Object[]{});
        } catch (Throwable t) {
            if (astObject != null) {
                return astObject.getFormalComment();
            } else {
                return "";
            }
        }
    }

    /**
     * Specialized signature generation for nodes in the structure model.
     *
     * @todo    the compiler should generate these names, doing it this way is atrocious
     */
    private static String genSignature(ASTObject astObject) {
        String name = "";
        if (astObject instanceof CompilationUnit) {
            return astObject.getSourceFile().getName();
        } else if (astObject instanceof MethodDec) {
            Method method = ((MethodDec)astObject).getMethod();
            return method.getName() + method.getFormals().toShortString();
        } else if (astObject instanceof TypeDec) {
            return ((TypeDec)astObject).getSourceExtendedId();
        } else if (astObject instanceof FieldDec) {
            return ((FieldDec)astObject).getName();
        } else if (astObject instanceof ConstructorDec) {
            ConstructorDec constructorDec = (ConstructorDec)astObject;
            return constructorDec.getDeclaringType().getSourceExtendedId() + constructorDec.getFormals().toShortString();
      }  else if (astObject instanceof IntroducedDec) {
            IntroducedDec introDec = (IntroducedDec)astObject;
            return introDec.getTargets().toShortString() + '.' + genSignature(introDec.getDec());
//            introDec.toShortString();
        } else if (astObject instanceof PointcutDec) {
            PointcutDec pointcutDec = (PointcutDec)astObject;
            return pointcutDec.getName() + pointcutDec.getFormals().toShortString();
//        } else if (astObject instanceof CallExpr) {
//            CallExpr call = (CallExpr)astObject;
//            name = call.get;
        } else if (astObject instanceof ShowErrorDec) {
            ShowErrorDec errorDec = (ShowErrorDec)astObject;
            return errorDec.toShortString();
        } else if (astObject instanceof SoftThrowableDec) {
            SoftThrowableDec softThrowableDec = (SoftThrowableDec)astObject;
            return softThrowableDec.toShortString();
        } else if (astObject instanceof IntroducedSuperDec) {
            IntroducedSuperDec introducedSuperDec = (IntroducedSuperDec)astObject;
            return introducedSuperDec.toShortString();
        } else if (astObject instanceof AdviceDec) {
            AdviceDec adviceDec = (AdviceDec)astObject;
            return adviceDec.toShortString();
        } else if (astObject instanceof SOLink) {
            SOLink soLink = (SOLink)astObject;
            return genSignature(soLink.getTarget().getCorrespondingDec());
        } else if (astObject instanceof CatchClause) {
            CatchClause catchClause = (CatchClause)astObject;
            return catchClause.getFormal().getType().getSourceExtendedId();
//    	} else if (genIsStmntKind(astObject)) {
//            name = astObject.unparse();
//            name = name.replace('/', ' ');
//            name = name.replace('*', ' ');
//            name = name.replace('\n', ' ');
//            name.trim();
//            java.util.StringTokenizer st = new java.util.StringTokenizer(name, " ");
//            String s = "";
//            while (st.hasMoreElements()) {
//                s += ((String)st.nextElement()).trim() + " ";
//            }
//            name = s;
//            int endIndex = name.indexOf(')');
//            if (endIndex != -1) {
//                name = name.substring(0, endIndex+1);
//            }
//            if (name.startsWith("this.")) {
//                name = name.substring(5);
//            }
        } else {
            return "? " + astObject.toShortString();
        }
    }
}