/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;


import java.util.*;


public class CanThrowWalker extends Walker {
    Stack canThrow = null;
    
    public static void checkThrows(CodeDec codeDec) {
        if (codeDec.isSoftThrowable()) return ;
        //System.out.println("checking: " + codeDec.toShortString());
        new CanThrowWalker(codeDec).process(codeDec);
    }
    
    
    public CanThrowWalker(CodeDec codeDec) {
        super(codeDec.getCompiler());
        canThrow = new Stack();
        canThrow.push(makeThrowableSet(codeDec.getThrows()));
    }
    
    Collection makeThrowableSet(TypeDs _throws) {        
        if (_throws == null) return Collections.EMPTY_SET;

        Collection ret = new HashSet();        
        
        final int N = _throws.size();
        for(int i=0; i<N; i++) {
            Type exceptionType = _throws.get(i).getType();
            ret.add(exceptionType);
        }
        return ret;
    }
    
    Collection makeThrowableSet(CatchClauses catchClauses) {        
        if (catchClauses == null) return Collections.EMPTY_SET;

        Collection ret = new HashSet();        
        
        final int N = catchClauses.size();
        for(int i=0; i<N; i++) {
            Type exceptionType = catchClauses.get(i).getFormal().getType();
            ret.add(exceptionType);
        }
        return ret;
    }

    
    boolean checkLegalThrow(ASTObject where, Type throwType) {
        if (throwType.isUncheckedThrowable()) return true;
        
        for (Iterator i = canThrow.iterator(); i.hasNext(); ) {
            for (Iterator j = ((Collection)i.next()).iterator(); j.hasNext(); ) {
                Type canThrowType = (Type)j.next();
                if (throwType.isSubtypeOf(canThrowType)) return true;
            }
        }
        where.showError("unreported exception " + throwType.getString() +
                "; must be caught or declared to be thrown");
        return false;
    }
    
    
    void checkLegalThrows(ASTObject where, TypeDs throwTypeDs) {
        if (throwTypeDs == null) return;
        
        for (int i = 0; i < throwTypeDs.size(); i++) {
            if (!checkLegalThrow(where, throwTypeDs.get(i).getType())) return;
        }
    }
        
    public ASTObject process(ASTObject node) {
        if (node instanceof TryCatchStmt) {
            TryCatchStmt tryStmt = (TryCatchStmt)node;
            canThrow.push(makeThrowableSet(tryStmt.getCatches()));
            super.process(tryStmt.getBody());
            canThrow.pop();
            super.process(tryStmt.getCatches());
            return node;
        }

        if (node instanceof ThrowStmt) {
            ThrowStmt throwStmt = (ThrowStmt)node;
            checkLegalThrow(throwStmt, throwStmt.getExpr().getType());
        } else if (node instanceof CallExpr) {
            CallExpr callExpr = (CallExpr)node;
            if (callExpr.isSoftThrowable()) return super.process(node);
            if (callExpr.getId().equals("aspectOf")) {
                return super.process(node);
            }

            Method method = callExpr.getMethod();

            if (method == null) {
                //System.out.println(">>>>>>>>>>>>>>>>>>>>>unbound call to: " + callExpr.getId());
                return super.process(node);
            }
            checkLegalThrows(callExpr, method.getThrows());
        } else if (node instanceof NewInstanceExpr) {
            NewInstanceExpr newExpr = (NewInstanceExpr)node;
            if (newExpr.isSoftThrowable()) return super.process(node);
            
            Constructor init = newExpr.getConstructor();
            checkLegalThrows(newExpr, init.getThrows());
        } else if (node instanceof ConstructorCallExpr) {
            ConstructorCallExpr newExpr = (ConstructorCallExpr)node;
            //if (newExpr.isSoftThrowable()) return super.process(node);
            
            Constructor init = newExpr.getConstructor();
            checkLegalThrows(newExpr, init.getThrows());
            
        }
        if (node instanceof TypeDec) return node;
        return super.process(node);
    }
}

