/* -*- Mode: Java; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import java.util.*;
import org.aspectj.compiler.base.ast.*;

public class ClassExprPass extends WalkerPass {
    public ClassExprPass(JavaCompiler jc) { super(jc); }
    public String getDisplayName() { return "desugaring T.class exprs"; }

    TypeDec dec = null;
    HashMap/*<FieldDec>*/ fields = null;
    MethodDec method = null;
    TypeDec holder = null;  // used by interfaces

    public ASTObject process(ASTObject object) {
	if (object instanceof TypeDec) {
	    TypeDec savedDec = this.dec;
	    HashMap savedFields = this.fields;
	    MethodDec savedMethod = this.method;
	    TypeDec savedHolder = this.holder;
	    dec    = (TypeDec) object;
	    fields = new HashMap();
	    if (object instanceof ClassDec) {
		holder = dec;
	    }
	    object.walk(this);
	    this.dec = savedDec;
	    this.fields = savedFields;
	    this.method = savedMethod;
	    this.holder = savedHolder;
	    return object;
	}
	if (object instanceof ClassExpr) {
	    object = handleClassExpr((ClassExpr) object);
	}
        object.walk(this);
        return object;
    }

    ASTObject handleClassExpr(ClassExpr object) {
        final AST ast = getAST();
	Type ty = object.getTypeD().getType();
        if (ty instanceof PrimitiveType) return ty.getClassExpr();
	FieldDec field  = genField(ty);
	MethodDec maker = genMethod();
	return ast.makeTriTest(ast.makeBinop("==",
                                             ast.makeGet(field),
					     ast.makeNull()),
			       ast.makeSet(field,
					   ast.makeStaticCall(method,
							      ast.makeString(ty.getExternalName()))),
			       ast.makeGet(field));
    }

    FieldDec genField(Type ty) {
        final AST ast = getAST();
        final TypeManager tm = getTypeManager();
	FieldDec field = (FieldDec) fields.get(ty);
	if (field == null) {
	    if (holder == null) genHolder();
	    field = ast.makeField(ast.makeModifiers(Modifiers.STATIC),
				  tm.getClassType(),
                                  genFieldName(ty));
	    holder.getBody().add(field);
	    fields.put(ty, field);
	}
	return field;
    }

    String genFieldName(Type ty) {
        String str = ty.getExternalName();
        if (ty instanceof ArrayType) {
            str = "array" + str;
        } else {
            str = "class$" + str;
        }
        str = str.replace('[','$').replace('.','$');
        if (str.charAt(str.length() - 1) == ';') {
            str = str.substring(0, str.length() - 1);
        }
        return str;
    }

    MethodDec genMethod() {
        final AST ast = getAST();
        final TypeManager tm = getTypeManager();
	if (method == null) {
	    if (holder == null) genHolder();
            FormalDec formal = ast.makeFormal(tm.getStringType(), "name$");
            FormalDec exnFormal = ast.makeFormal(tm.getType("java.lang", "ClassNotFoundException"),
                                                 "exn$");
            method = ast.makeMethod(
                         ast.makeModifiers(Modifiers.STATIC),
                         tm.getClassType(),
                         "class$",
                         ast.makeFormals(formal),
                         ast.makeCodeBody(
                             ast.makeTryCatch(
                                 ast.makeBlock(
                                     ast.makeReturn(
                                         ast.makeStaticCall(tm.getClassType(),
                                                            "forName",
                                                            ast.makeExprs(ast.makeVar(formal))))),
                                 exnFormal,
                                 ast.makeBlock(
                                     ast.makeThrow(
                                         ast.makeNew(
                                             tm.getType("java.lang", "NoClassDefFoundError"),
                                             ast.makeCall(ast.makeVar(exnFormal), 
                                                          "getMessage",
                                                          ast.makeExprs())))))));
            holder.getBody().add(method);
	}
	return method;
    }

    void genHolder() {
        final AST ast = getAST();
        final TypeManager tm = getTypeManager();
        
        holder = ast.makeClass(ast.makeModifiers(Modifiers.STATIC | Modifiers.PRIVATE),
                               "00", 
                               tm.getObjectType().makeTypeD(), 
                               ast.makeTypeDs());
        holder.setIsInner(false);
        dec.addMemberTypeDec(holder);
    }
}
