/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;
import org.aspectj.compiler.base.ast.*;

import java.util.*;
import java.io.*;

import org.aspectj.tools.ide.SourceLine;
import org.aspectj.util.CollectionUtil;

public class CodeWriter extends CompilerObject {
    private boolean writeComments = false;
    private LineMap lineMap = null;
    private CodeText out;
    //private Writer out = null;

    public CodeWriter(JavaCompiler compiler) {
        this(compiler, 1024, null);
    }

    public CodeWriter(JavaCompiler compiler, int expectedSize, LineMap lineMap) {
        super(compiler);
        out = new CodeText(expectedSize);
        this.lineMap = lineMap;
        //XXX bogus
        this.writeComments = !compiler.getOptions().nocomments;
    }

    public String getString() {
        return out.getString();
    }
    
    public void writeTo(OutputStream ostream) throws IOException {
        out.writeTo(ostream);
    }

    private boolean onlySignatures = false;
    public boolean isOnlySignatures() { return onlySignatures; }
    public void setOnlySignatures(boolean v) { onlySignatures = v; }

    private boolean onlyPatch = false;
    public boolean isOnlyPatch() { return onlyPatch; }
    public void setOnlyPatch(boolean v) { onlyPatch = v; }

    public void parenExpr(Expr expr) {
        openParen('(');
        write(expr);
        closeParen(')');
    }

    public void parenExprs(Exprs exprs) {
        openParen('(');
        write(exprs);
        closeParen(')');
    }

    public void writeChildren(ASTObject parent) {
        if (parent == null) return;
        final int N = parent.getChildCount();
        if (N > 0) write(parent.getChildAt(0));

        for(int i=1; i<N; i++) {
            write(parent.getChildAt(i));
        }
    }

    public void writeChildrenWithCommas(ASTObject parent) {
        if (parent == null) return;
        final int N = parent.getChildCount();
        if (N > 0) write(parent.getChildAt(0));

        for(int i=1; i<N; i++) {
            comma();
            write(parent.getChildAt(i));
        }
    }


    public void write(ASTObject node) {
        try {
            if (node == null) return;
    
            if (writeComments) {
                Comment comment = node.getComment();
                while (comment != null) {
                    comment.unparse(this);
                    comment = comment.getNextComment();
                }
            }
    
            // This is to output line-number correspondences for the debugger
            if (lineMap != null) {
                lineMap.noteNode(node);
            }
    
            node.unparse(this);
        } catch (IOException ioe) {
            //XXX this clause needs to go away
        }
    }

    public int getIndentSize() { return 2; }

    private int lineStart = 0;
    private int lastGoodBreak = -1;
    private int lastPossibleBreak = -1;    
    
    public void noteGoodBreak() {
        breakTooLongLine();
        lastGoodBreak = out.getIndex();
    }
    
    public void notePossibleBreak() {
        breakTooLongLine();
        lastPossibleBreak = out.getIndex();
    }
    
    public void breakTooLongLine() {
        int column = out.getIndex() - lineStart;
        //System.out.println("column: " + column);
        if (column < MAX_COLUMNS) return;
     
//        int goodMetric = MAX_COLUMNS - lastGoodBreak;
//        int possibleMetric = MAX_COLUMNS - lastPossibleBreak;
        
        int breakPoint = -1;
        if (lastGoodBreak != -1) {
            breakPoint = lastGoodBreak;
        } else if (lastPossibleBreak != -1) {
            breakPoint = lastPossibleBreak;
        } else {
            // we can't manage to break this line
            return;
        }
        
        insertNewLine(breakPoint);        
    }
    
    public void insertNewLine(int breakAtIndex) {
        byte[] saveData = out.backup(breakAtIndex);
        // would like to also include pretty-breaking using parens, but not now
        out.write(newLine);
        lineStart = out.getIndex();
        lastPossibleBreak = lastGoodBreak = -1;        
        writeSpaces(indents + 4);
        out.write(saveData);
        lastPossibleBreak = out.getIndex();
    }
    
    
    
    public void comma() {
        write(',');
        optionalSpace();
        noteGoodBreak();
    }

    public void requiredSpace() {
        //column++;
        out.write(' ');
        notePossibleBreak();
    }
    public void optionalSpace() {
        //column++;
        out.write(' ');
        notePossibleBreak();     
    }

    public void writeKeyword(String keyword) {
        write(keyword);
    }

    public void writeOp(String op) {
        optionalSpace();
        write(op);
        optionalSpace();
    }

    // These fields keep a stack of open paren locations to use for pretty-printing lines
    // that are too long
    final int MAX_PAREN_DEPTH = 32;
    private int[] parenPositions = new int[MAX_PAREN_DEPTH+1];
    private int parenDepth = 0;
    
    public void openParen(char paren) {
        write(paren);
        parenDepth++;
        parenPositions[parenDepth > MAX_PAREN_DEPTH ? MAX_PAREN_DEPTH : parenDepth] = out.getIndex();
        notePossibleBreak();
    }

    public void closeParen(char paren) {
        write(paren);
        parenDepth--;
    }


    public void closeStmt() {
        write(';');
        //newLine();
    }


    private int indents=0;
    public int getIndents() {
        return indents;
    }
    public void setIndents(int indents) {
        this.indents = indents;
    }

    public void openBlock() {
        //if (!needIndent) write(" ");
        writeln("{");
        indent();
    }

    public void closeBlock() {
        dedent();
        write("} ");
    }

    public void indent() {
        indents += getIndentSize();
    }

    public void dedent() {
        indents -= getIndentSize();
    }

    // text may include newlines and will be nicely broken up on those
    public void writeLines(String text) {
        if (text == null) return;

        // break text up into lines and write each individually
        int currentPos = 0;
        while (true) {
            int newlinePos = text.indexOf('\n', currentPos);
            if (newlinePos == -1) break;
            writeln(text.substring(currentPos, newlinePos));
            currentPos = newlinePos + 1;
        }

        write(text.substring(currentPos));
    }


    //int column = 0;
    // text can not include new lines
    public void write(String text) {
        doIndent();
        out.write(text);
        //column += text.length();
    }

    public void write(char ch) {
        doIndent();
        out.write(ch);
        //column += 1;
    }

    // text can not include new lines (but one is added)
    public void writeln(String text) {
        write(text);
        newLine();
    }

    //XXX this should be made platform independent
    final String newLine = "\n";

    final int COMMENT_COLUMN = 67;
    final int MAX_COLUMNS    = 100;
    public void newLine() {
        // deal with long lines
        breakTooLongLine();
        
        if (writeComments && lineMap != null && lineMap.hasCurrentSourceLine()) {
            int column = out.getIndex() - lineStart;
            if (column < COMMENT_COLUMN) {
                writeSpaces(COMMENT_COLUMN - column);
            } else {
                out.write(' ');
            }
            
            out.write("//");
            // should add in a check to see if the file matches and use a shorter form
            // that step should also optimize the computation of this
            out.write(new File(lineMap.getCurrentSourceFile()).getName());
            out.write(":" + lineMap.getCurrentSourceLine());
        }
        out.write(newLine);
        if (lineMap != null) {
            lineMap.noteNewLine();
        }
        lineStart = out.getIndex();
        needIndent = true;
        lastGoodBreak = -1;
        lastPossibleBreak = -1;
    }

    private boolean needIndent = false;
    void doIndent() {
        if (!needIndent) return;
        needIndent = false;
        writeSpaces(indents);
    }

    //??? do this with an array for speed
    final void writeSpaces(final int N) {
        //column += N;
        for (int i = 0; i < N; i++) {
            out.write(' ');
        }
    }

    public static final String HEADER_START =
        "/*   Generated by AspectJ version ";

    public static boolean isGeneratedSource(File file) {
        try {
            Reader reader = new FileReader(file);
            for(int i=0; i<HEADER_START.length(); i++) {
                int ch = reader.read();
                if (ch != HEADER_START.charAt(i)) return false;
            }
            reader.close();
            return true;
        } catch (IOException ioe) {
            //!!! should I worry about closing the file under these circumstances?
            return false;
        }
    }


    public void writeHeader() {
        writeln(HEADER_START+Options.version+" */");
    }
}


class CodeText {
    protected byte[] data;
    protected int index;
    protected boolean addSlash;

    public CodeText(int expectedSize) {
        data = new byte[expectedSize];
        index = 0;
        addSlash = false;
    }

    public int getIndex() {
        return index;
    }
    
    public byte[] backup(int toIndex) {
        int backupSize = index - toIndex;
        byte[] ret = new byte[backupSize];
        
        System.arraycopy(data, toIndex, ret, 0, backupSize);
        
        index = toIndex;
        return ret;
    }
    
    private final void expandTo(int newSize) {
        if (newSize < data.length) return;

        int newLength = data.length * 2;
        while (newLength < newSize) newLength *= 2;
        
        byte[] newData = new byte[newLength];
        System.arraycopy(data, 0, newData, 0, index);
        data = newData;
    }
    
    
    public void write(byte b) {
        expandTo(index + 1);
        data[index++] = b;
    }
        
    public void write(byte[] bytes) {
        expandTo(index + bytes.length);
        System.arraycopy(bytes, 0, data, index, bytes.length);
        index += bytes.length;
    }
    
    public void write(String s) {
        final int N = s.length();
        for (int i = 0; i < N; i++) {
            write(s.charAt(i));
        }
    }
    
    public void write(int ch) {
        expandTo(index + 7);
        
        if (addSlash && ch == 'u') {
            data[index++] = (byte)'\\';
            //data[index++] = 'u';
        }

        if (ch == '\\') {
            addSlash = !addSlash;
        } else {
            addSlash = false;
        }

        if (ch < 1 || ch > 126) {
            int b1 = ch>>12;
            int b2 = ch>>8 & 0xf;
            int b3 = ch>>4 & 0xf;
            int b4 = ch & 0xf;

            data[index++] = (byte)'\\';
            data[index++] = (byte)'u';
            data[index++] = (byte)Character.forDigit(b1, 16);
            data[index++] = (byte)Character.forDigit(b2, 16);
            data[index++] = (byte)Character.forDigit(b3, 16);
            data[index++] = (byte)Character.forDigit(b4, 16);
        } else {
            data[index++] = (byte)ch;
        }
    }

    public String getString() {
        try {
            return new String(data, 0, index, "US-ASCII");
        } catch (UnsupportedEncodingException e) {
            return new String(data, 0, index);
        }
    }
    
    public void writeTo(OutputStream ostream) throws IOException {
        ostream.write(data, 0, index);
        ostream.flush();
    }
}
