/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;


import org.aspectj.compiler.base.ast.*;


import java.util.*;
import java.io.*;

public class ErrorHandler {
    private PrintWriter errorWriter;
    private JavaCompiler compiler;

    private static final String PREFIX = "";

    public void setCompiler(JavaCompiler compiler) {
        this.compiler = compiler;
    }
    
    public JavaCompiler getCompiler() { return compiler; }

    public void cleanupErrorWarningCounters() {
        errors = 0;
        warnings = 0;
    }

    public ErrorHandler() {
        this(null);
    }

    public ErrorHandler(Writer errorWriter) {
        if (errorWriter == null) {
            errorWriter = makePrintWriter();
        }
        if (!(errorWriter instanceof PrintWriter)) {
            errorWriter = new PrintWriter(errorWriter, true);
        }
        this.errorWriter = (PrintWriter)errorWriter;
    }

    protected PrintWriter makePrintWriter() {
        return new PrintWriter(System.err, true);
    }


    /* This section of code handles errors that occur during compilation */
    static final String internalErrorMessage =
    "Please copy the following text into an email message and send it,\n" +
    "along with any additional information you can add to:            \n" +
    "                                                                 \n" +
    "       support@aspectj.org                            \n" +
    "                                                                 \n";

    static final String specificCodeRequest =
    "If you could send us at least that fragment of code, provided it isn't\n"+
    "confidential, that would be helpful to us in finding the bug.\n";

    static final String generalCodeRequest =
    "(If possible, please include a small fragment of code that can \n"+
    "reproduce the error, provided that code isn't confidential.)   \n";


    public synchronized void internalError(Throwable uncaughtException, ASTObject where) {
        throw new InternalCompilerError(compiler, uncaughtException, where);
    }

    //public synchronized void internalError(Throwable uncaughtException) {
    //  internalError(uncaughtException, compiler.getCurrentNode());
    //}

    private int printLine(File source, int beginLine) {
        return printLines(source, beginLine, beginLine);
    }

    private static int count(String s, char ch) {
        int index = s.indexOf(ch);
        int n = 0;
        while (index != -1) {
            n += 1;
            index = s.indexOf(ch, index+1);
        }
        return n;
    }


    private int printLines(File source, int beginLine, int endLine) {
        int tabCount = 0;
        try {
            BufferedReader reader = new BufferedReader(new FileReader(source));

            try {
                int i;

                for(i=1; i<beginLine; i++) reader.readLine();

                for(;i<=endLine; i++) {
                    String line = reader.readLine();
                    if (line == null) {
                        errorWriter.println(PREFIX + "<at end of file>");
                        return -1;
                    }
                    errorWriter.println(PREFIX + line);
                    tabCount += count(line, '\t');
                }
            } finally {
                if (reader != null) reader.close();
            }
        } catch (IOException e) {
            errorWriter.println(PREFIX + "can't read text from "+source.getPath());
        }

        return tabCount;
    }

    private void printCaretLine(int pos) {
        errorWriter.print(PREFIX);
        for(int i=1; i<pos; i++) errorWriter.print(" ");
        errorWriter.println("^");
    }

    public synchronized boolean showMessage(String message) {
        return showMessage(null, message, false, false);
    }

    public synchronized boolean showMessage(ASTObject where, String message) {
        return showMessage(where, message, false, false);
    }


    private Set firstLines = new HashSet();
    public synchronized boolean showMessage(File source, int line, int endLine, int column, String message, boolean isError) {
        if (isError && compiler.getOptionDumpStack()) {
            Thread.dumpStack();
        }

        if (source == null) {
            errorWriter.println(PREFIX + message);
            return true;
        }

        String head = source.getPath() + ":" + line + ":" + column + ": ";
        String firstLine = head+message;
        if (firstLines.contains(firstLine)) return false;
        firstLines.add(firstLine);
        errorWriter.println(PREFIX + firstLine);
        if (endLine > line) {
            printLines(source, line, endLine);
        } else {
            int tabcount = printLine(source, line);
            if (tabcount != -1) {
               column += tabcount * 7;
               printCaretLine(column);
            }
        }
        return true;
    }

    public synchronized boolean showMessage(ASTObject where, String message, boolean showAll, boolean isError) {
        if (where == null) {
            return showMessage(null, -1, -1, -1, message, isError);
        }
        if (where.getSourceFile() == null || where.getBeginLine() == -1) {
            
            ASTObject newWhere = where.getSourceLocation().getSourceObject();
            if (newWhere != null) {
                
                if (showMessage(newWhere, message, showAll, isError)) {
                    if (newWhere.getBeginLine() == -1 || compiler.getOptions().developer) {
                        errorWriter.println("    in generated code: " + where.unparse());
                    }
                    return true;
                } else {
                    return false;
                }
            }
        }
        int endLine = showAll ? where.getEndLine() : -1;
        return showMessage(where.getSourceFile(), where.getBeginLine(), endLine,
                    where.getBeginColumn(), message, isError);
    }


    protected int errors = 0;
    public synchronized void showError(ASTObject where, String message) {
        if (showMessage(where, message, false, true)) errors += 1;
    }

    public synchronized void showError(File source, int line, int column, String message) {
        if (showMessage(source, line, -1, column, message, true)) errors += 1;
    }

    protected int warnings = 0;
    public synchronized void showWarning(ASTObject where, String message) {
        // we only show warnings when there are no pending errors
        if (errors > 0) return;
        if (showMessage(where, message+" (warning)")) warnings += 1;
    }


	public synchronized void showWarning(ASTObject where, Message message) {
		if (message.shouldDisplay(getCompiler())) {
			showWarning(where, message.getString());
		}
	}
			


    public synchronized void showWarnings() {
        if (warnings > 0) {
            errorWriter.println(PREFIX + Integer.toString(warnings)+" warnings");
        }
    }

    public synchronized boolean willExitWithErrors() {
        return errors > 0;
    }

    public synchronized void exitOnErrors() {
        if (errors > 0) {
            showWarnings();
            errorWriter.println(PREFIX + Integer.toString(errors)+" errors");
            if (compiler.getOptionIgnoreErrors()) return;

            throw new CompilerErrors(errors);
        }
    }

	public static class Message {
		private String message;
		public Message(String message) {
			this.message = message;
		}
		
		public boolean shouldDisplay(JavaCompiler compiler) {
			return compiler.getOptions().Xlint;
		}
		public String getString() { return message; }
	}		


	public static final Message invalidAbsoluteTypeName =
		new Message("No match for this type");
		
	public static final Message invalidWildcardTypeName =
		new Message("No matches for this type pattern");
		

}
