/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;

import org.aspectj.compiler.crosscuts.ast.ProceedExpr;

import java.util.*;


public class ExceptionFinder extends Walker {
    /**
     * Finds all exceptions that can be thrown from ast
     * 
     * @param ast      where to search for exceptions
     * @param deFacto  true == exceptions that make actually be thrown
     *                 false == exceptions that JLS flow analysis wants to know about
     */
    
    public static Set getPossibleExceptions(ASTObject ast, boolean deFacto) {
        if (ast == null) {
            return new HashSet();
        }

        ExceptionFinder finder = new ExceptionFinder(ast.getCompiler());
        finder.deFacto = deFacto;
        finder.process(ast);
        return finder.exceptions;
    }


    boolean deFacto;
    Set exceptions = new HashSet();

    Set findingInMethods = new HashSet();

    private ExceptionFinder(JavaCompiler compiler) {
        super(compiler);
    }

    void addAll(Set set, TypeDs typeDs) {
        if (typeDs == null) return;

        final int N = typeDs.size();
        for(int i=0; i<N; i++) {
            set.add(typeDs.get(i).getType());
        }
    }

    void removeSubtypes(Set types, Type baseType) {
        for(Iterator iter = types.iterator(); iter.hasNext(); ) {
            Type checkType = (Type)iter.next();
            if (checkType.isSubtypeOf(baseType)) {
                iter.remove();
            }
        }
    }

    void removeCaughtExceptions(Set exceptions, CatchClauses catchClauses) {
        if (catchClauses == null) return;

        final int N = catchClauses.size();
        for(int i=0; i<N; i++) {
            Type exceptionType = catchClauses.get(i).getFormal().getType();
            removeSubtypes(exceptions, exceptionType);
        }
    }

    public ASTObject process(ASTObject node) {
        if (node instanceof TryCatchStmt) {
            TryCatchStmt tryStmt = (TryCatchStmt) node;
            Set bodyExceptions = getPossibleExceptions(tryStmt.getBody(), deFacto);
            removeCaughtExceptions(bodyExceptions, tryStmt.getCatches());
            exceptions.addAll(bodyExceptions);
            super.process(tryStmt.getCatches());
            return node;
        }

        if (node instanceof ThrowStmt) {
            ThrowStmt throwStmt = (ThrowStmt)node;
            exceptions.add(throwStmt.getExpr().getType());
        } else if (node instanceof CallExpr) {
            CallExpr callExpr = (CallExpr)node;
            if (callExpr.getId().equals("aspectOf")) {
                return super.process(node);
            }

            Method method = callExpr.getMethod();

            if (method == null) {
                //System.out.println(">>>>>>>>>>>>>>>>>>>>>unbound call to: " + callExpr.getId());
                return super.process(node);
            }
            addAll(exceptions, method.getThrows());
        } else if (node instanceof NewInstanceExpr) {
            NewInstanceExpr newExpr = (NewInstanceExpr)node;
            Constructor init = newExpr.getConstructor();
            addAll(exceptions, init.getThrows());
        } else if (node instanceof ProceedExpr) {
            exceptions.addAll(((ProceedExpr)node).getPossibleExceptions(deFacto));
        }
            
        if (node instanceof TypeDec) return node;
        return super.process(node);
    }
}

