/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import java.util.*;

/** <p> Checks adherence to JLS 8.3.2.3, "Restrictions on the use of
 *  Fields during Initialization", which says:</p>
 *
 *  <p>The declaration of a member needs to appear before it is used only if
 *  the member is an instance (respectively static) field of a class or
 *  interface C and all of the following conditions hold:</p>
 *
 *  <ul>
 *   <li> The usage occurs in an instance (respectively static) variable
 *    initializer of C or in an instance (respectively static) initializer
 *    of C.</li>
 *
 *   <li> The usage is not on the left hand side of an assignment.</li>
 *
 *   <li> C is the innermost class or interface enclosing the usage.  </li>
 *  </ul>
 *
 *  <p> A compile-time error occurs if any of the three requirements
 *  above are not met.</p>
 *
 *  <p> There are four contexts, each represented by a class:</p>
 *
 *  <ul>
 *    <li>ForwardReferenceChecker: The outermost, just walks compilation
 *    units. </li>
 *
 *    <li>TypeChecker: walks contents of TypeDecs, can create (and
 *    transfer control to) TypeCheckers (for member types),
 *    InitializerCheckers (for initializers) and BodyCheckers (for methods
 *    and constructors). </li>
 *
 *    <li>InitializerChecker: walks contents of initializers.  Can create
 *    (and transfer control to) TypeCheckers (for local types). </li>
 *
 *    <li>BodyChecker: walks methods and constructors. Can create (and
 *    transfer control to) TypeCheckers (for local types). </li>
 *  </ul>
 *
 */

public class ForwardReferenceChecker extends WalkerPass {
    public ForwardReferenceChecker(JavaCompiler compiler) {
        super(compiler);
    }

    public String getDisplayName() {
        return "checking forward references";
    }

    public ASTObject process(ASTObject node) {
        //System.err.println("entering " + node + " with " + this);
        node.walkForwardReference(this);
        return node;
    }

    public ForwardReferenceChecker createTypeChecker(TypeDec ty) {
        return new TypeChecker(getCompiler(), ty);
    }
    public ForwardReferenceChecker createInitializerChecker(boolean isStatic) {
        throw new RuntimeException("found initializer while walking with " + this);
    }
    public ForwardReferenceChecker createBodyChecker() {
        throw new RuntimeException("found body while walking with " + this);
    }
    public void addFieldDec(FieldDec f) {
        throw new RuntimeException("found field while walking with " + this);
    }
    public void checkReference(FieldAccessExpr f) { }

    public String toString() {
        return "<ForwardReferenceChecker in Compilation Unit>";
    }

    // ------------------------------

    static class TypeChecker extends ForwardReferenceChecker {
        TypeDec context;
        TypeChecker(JavaCompiler compiler, TypeDec ty) {
            super(compiler);
            context = ty;
        }
        Set statics = new HashSet();
        Set dynamics = new HashSet();

        public ForwardReferenceChecker createInitializerChecker(boolean isStatic) {
            return new InitializerChecker(getCompiler(), isStatic, context,
                                          isStatic ? statics : dynamics);
        }
        public ForwardReferenceChecker createBodyChecker() {
            return new BodyChecker(getCompiler());
        }
        public void addFieldDec(FieldDec f) {
            (f.isStatic() ? statics : dynamics).add(f);
        }
        public String toString() {
            return "<ForwardReferenceChecker in TypeDec with dynamics " +
                dynamics + " and statics " + statics + ">";
        }
    }

    static class InitializerChecker extends ForwardReferenceChecker {
        boolean isStatic;
        TypeDec context;
        Set fields;
        InitializerChecker(JavaCompiler compiler, boolean isStatic, TypeDec context, Set fields) {
            super(compiler);
            this.isStatic = isStatic;
            this.context = context;
            this.fields = fields;
        }
        public ForwardReferenceChecker createBodyChecker() {
            return this;
        }
        public void checkReference(FieldAccessExpr e) {
            FieldDec dec = e.getFieldDec();
            if (dec.isStatic() != isStatic) return;
            if (dec.getDeclaringType().getTypeDec() != context) return;
            if (fields.contains(dec)) return;
            e.showError("Illegal forward reference");
        }
        public String toString() {
            return "<ForwardReferenceChecker in " +
                (isStatic ? "static" : "dynamic") +
                " Init with fields " + fields + ">";
        }
    }

    static class BodyChecker extends ForwardReferenceChecker {
        BodyChecker(JavaCompiler compiler) { super(compiler); }

        public ForwardReferenceChecker createBodyChecker() {
            return this;
        }
        public String toString() {
            return "<ForwardReferenceChecker in Body>";
        }

    }
}

