/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;

import java.util.*;

/** Provides backdoor methods for access to methods and fields that
    otherwise would not be accessible.

    <p> This follows the javac model rather than the jikes model.
    This model generates more methods, but will result in fewer method
    calls.  In particular, the code <code>foo1.x += 5</code> will be
    translated into 
    </p>

<pre>
Foo.incX(foo1, 5)
static int Foo.incX(Foo obj, int val) {
    return obj.x += 5;
}
</pre>

    <p> rather than </p>

<pre>
Foo.setX(Foo.getX() + 5)
static int Foo.setX(Foo obj, int val) {
    return obj.x = val;
}
static int Foo.getX(Foo obj) {
    return obj.x;
}
</pre>

    <p> which means we might have, for every private variable, one
    method for each kind of assignment expression, prefix expression,
    and postfix expression, in addition to one for plain reference.
    The jikes/ajc model would just have two.  Ah well. </p>

    <p> This creates the access methods on the <em>qualifying</em>
    type of the field access or method expression, not the type that
    necessarily declares the field.  So, the expression
    <code>foo1.m()</code>, when referring to the protected method m
    defined in a superclass of foo from outside the package, will
    result in:</p>

<pre>
package p1:
  public class Goo { protected void m() {} }
package p2:
  class Foo extends Goo { 
      class Inner {
          ... Foo.this.m() ...
      }
  }
--&gt;
package p2:
  class Foo extends Goo {
      static void access$m(Foo f) {
          f.m(); return;
      }
      class Inner {
         ... Foo.access$m(Foo.this) ...
      }
  }
</pre>

    <p> the access method going into Foo, the qualifying type of the
    method call, not Goo, the defining type of the protected method.</p>

    <p> This pass was written to deal with inner access (access to
    private and protected members), so it will always generate
    package-protected access methods (see {@link #makeAccessMethod})
    when access is needed (see {@link #isAccessible}).  It could be
    used (with some adaptation) as a cleanup from the crosscutting,
    and in such a case would generate some public access methods. </p>

    <p> Currently it handles private constructors by simply making
    them package-protected.  This will not work in general for
    crosscutting, but works just fine for inners.  We may want to
    change this to add "cookied" access constructors when needed. </p>

*/

public class InnerAccessFixer extends WalkerPass {

    public InnerAccessFixer(JavaCompiler compiler) {
        super(compiler);
    }

    public String getDisplayName() {
        return "adding inner access backdoors";
    }

    public ASTObject process(ASTObject object) {
        object.walk(this);
        return object.postInnerAccess(this);
    }

    private Map/*<Triple, MethodDec>*/ accessMethods = new HashMap();

    public MethodDec getAccessMethod(Type qType, Dec dec, String op, Expr src) {
        Triple key = new Triple(qType, dec, op);
        MethodDec methodDec = (MethodDec) accessMethods.get(key);
        if (methodDec == null) {
            methodDec = src.buildAccessMethod(this);
            ((NameType)qType).getTypeDec().addToBody(methodDec);
            accessMethods.put(key, methodDec);
        }
        return methodDec;
    }

    private int accessCounter = 0;
    public MethodDec makeAccessMethod(Type returnType, Formals formals, Expr bodyExpr) {
        final AST ast = getAST();
        return 
            ast.makeMethod(ast.makeModifiers(Modifiers.STATIC),
                           returnType,
                           "access$" + accessCounter++,
                           formals,
                           returnType.isVoid() 
                           ? ast.makeCodeBody(ast.makeStmt(bodyExpr), ast.makeReturn()) 
                           : ast.makeCodeBody(ast.makeReturn(bodyExpr)));
    }        

    /** make a primary expression that is usable in calling the access
        method.  For static resources, this is the primary expression
        originally used to get access to the resource.  For non-static
        resources, this is a TypeExpr of the type of the primary
        expression, and in such cases we add the primary expression as
        the first argument to the call to the access method. 

        @param isStatic is the resource static?
        @param args the arguments to the call we are about to make,
        which me may have to add to
        @param e the original primary expression
        
        @return e or a TypeExpr

    */
    public Expr makeOutsidePrimary(boolean isStatic, Exprs args, Expr e) {
        if (e.getType() instanceof PrimitiveType) throw new RuntimeException();
        if (isStatic) {
            return e;
        } else {
            args.add(0, e);
            return getAST().makeTypeExpr(e.getType());
        }
    }

    /** make a primary expression that is usable inside the access
        method to get at the accessed resource (field or method).  For
        static resources, this is a typeExpr, but for non-static
        resources, we add a new formal and return a reference to that
        formal. 

        @param isStatic is the resource static?
        @param f the formals of the access method, which we may add to
        @param ty the type of the object with access to the resource

        @return a TypeExpr or a VarExpr
    */
    public Expr makeInsidePrimary(boolean isStatic, Formals f, Type ty) {
        final AST ast = getAST();
        if (isStatic) {
            return ast.makeTypeExpr(ty);
        } else {
            FormalDec obj = ast.makeFormal(ty, "obj$");
            f.add(0, obj);
            return ast.makeVar(obj);
        }
    }
            
    /** Is the dec accessible from the qualifying expression?  Note
        that this assumes that if q.getType() does not enclose
        q.getDeclaringType() (i.e., if we're dealing with a
        non-inner-type issue) that the dec <em>is</em> accessible.  */
    public boolean isAccessible(Dec dec, Expr q) {
        if (dec.isPrivate()) {
            return q.getBytecodeTypeDec() == dec.getBytecodeTypeDec();
        } else if (dec.isProtected()) {
            TypeDec target = dec.getBytecodeTypeDec();
            TypeDec src = q.getBytecodeTypeDec();
            if (src.getType().isSubtypeOf(target.getType())) return true;
            String pkg1 = src.getPackageName();
            String pkg2 = target.getPackageName();
            return (pkg1 == null) ? (pkg2 == null) : pkg1.equals(pkg2);
        } else {
            return true;
        }
    }

    // ------------------------------
    
    /** A type used to create keys for the access method map */
    static class Triple {
        Object one, two, three;
        Triple(Object one, Object two, Object three) { 
            this.one = one; this.two = two; this.three = three; 
        }
        public int hashCode() {
            return one.hashCode() + two.hashCode() * 2 + three.hashCode() * 3;
        }
        public boolean equals(Object other) {
            if (other instanceof Triple) {
                Triple otherTriple = (Triple) other;
                return one.equals(otherTriple.one) 
                    && two.equals(otherTriple.two)
                    && three.equals(otherTriple.three);
            } else {
                return false;
            }
        }
    }
}
