/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;

import java.util.*;

/**
 * This pass sets up information for TypeDec's about wherer they are as inners
 * It is also used to check whether or not non-static references are allowed.
 */
public final class InnerInfoPass extends WalkerPass {
    public InnerInfoPass(JavaCompiler jc) { super(jc); }

    public String getDisplayName() { return "setting inner info"; }

    private static final int CU = 0;           // compilation unit
    private static final int MEMBER = 1;       // class or aspect
    private static final int STATICMEMBER = 2; // interface
    private static final int DYNAMIC = 3;      // local, enclosing guy available
    private static final int STATIC = 4;       // local, no enclosing guy available
    private static final int NIGH_STATIC = 5;  // local, no enclosing guy available,
                                               // but one guy up is available

    private int context = CU;

    public ASTObject process(ASTObject object) {
        object.walkInnerInfo(this);
        object.postInnerInfo(this);        
        return object;
    }

    public int inCCall() { return setContext(NIGH_STATIC); }
    public int inType(boolean isInterface) {
        return setContext(isInterface ? STATICMEMBER : MEMBER);
    }
    public int inMember(boolean isStatic) { 
        return setContext(isStatic ? STATIC : DYNAMIC); 
    }
    private int setContext(int c) { 
        int context = this.context;
        this.context = c;
        return context;
    }
    public void restoreContext(int context) {
        this.context = context;
    }

    public boolean isStatic() {
        return context != DYNAMIC;
    }
    
    public boolean hasNonStaticAccess(ASTObject node, Type type) {        
        return isAccessible(type);
//          Type lexType = node.getLexicalType();
//          if (lexType == type) {
//              return !isStatic();
//          }
//          //XXX insert more agressive handling here for inners
//          return true;
    }
    
    public void checkStaticAccess(ASTObject node, SemanticObject so) {
        if (so.isStatic()) return;
        
        if (!hasNonStaticAccess(node, so.getDeclaringType())) {
            node.showError("non-static " + so.getKind() + " " + so.getName() + 
            " cannot be referenced from a static context");
        }
    }

    public static boolean isInner(Type t) {
        if (! (t instanceof NameType)) return false;
        NameType nt = (NameType) t;
        return nt.isInner()
            // we might not have processed it yet, if it's a member
            // type.  So we check...
            || (nt.getTypeDec().fromSource() 
                && ! nt.isAspect()
                && ! nt.isPackageMember()
                && ! nt.isLocal()
                && ! nt.getTypeDec().isStatic()
                && ! nt.getDeclaringType().isInterface());
    }

    public boolean isInDynamicContext() {
        return (context == MEMBER) || (context == DYNAMIC);
    }

    // ------------------------------

    private List types = new ArrayList();
    public void enterType(TypeDec dec) { types.add(dec.getType()); }
    public void leaveType() { types.remove(types.size() - 1); }

    public NameType currentType() {
        return (NameType) types.get(types.size() - 1);
    }
    public NameType enclosingType() {
        return (NameType) types.get(types.size() - 2);
    }

    public boolean isAccessibleExactly(Type t) {
        return isAccessible(t, false);
    }
    public boolean isAccessible(Type t) {
        return isAccessible(t, true);
    }
    public boolean isAccessible(Type t, boolean allowSubs) {
        NameType current = (NameType) types.get(types.size() - 1);
        if (current == t || (allowSubs && current.isSubtypeOf(t))) {
            return (context == DYNAMIC);
        }
        if (! current.isInner()) return false;
        for (int i = types.size() - 2; i >= 0; i--) {
            current = (NameType) types.get(i);
            if (current == t || (allowSubs && current.isSubtypeOf(t))) {
                return (context != STATIC);
            } else if (! current.isInner()) {
                return false;
            }
        }
        return false; //???
    }
}
