/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base;

import java.util.*;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;


public class Options {
    public static final String version = org.aspectj.compiler.Version.text;
    public static final long build = org.aspectj.compiler.Version.time;

    public File outputDir = new File(".");
    public File workingDir = new File("ajworkingdir");

    public String classpath = null;
    //{ addStringOption("classpath", "Specify where to find user class files"); }
    public String defaultClasspath = null;

    public String bootclasspath = null;
    public String extdirs = null;

    public String source = "1.3";

    public boolean lenient = false;
    public boolean strict = false;

    public String encoding = null;

    public boolean porting = false;

    public boolean ignoreErrors = false;

    public boolean runForever = false;

    public boolean mangler = false;

    public boolean force = true;
    public boolean nocomments = false; // This could be fancier...
    public boolean verbose = false; // This should be more sophisticated...
    public boolean preprocess = false;
    public boolean usejavac = false;
    public boolean bcgverbose = false;
    public boolean bcgtrace = false;

    public boolean newWeaving = true;

    public boolean strictTree = false;

    public boolean O = false;
    public boolean deprecation = false;

    public boolean noMetaJoinPointOptimization = false;
    public boolean alwaysMakeAroundClosures = false;

    // lintish stuff
    public boolean Xlint = false;

    // special hack added for Arno Schmidmeier
    public boolean XserializableAspects = false;

    // safer names for bean and EJB reflective code
    public boolean XaddSafePrefix = false;

    //XXX need something fancier for g:{...} and g:none
    public boolean g = false;

    public String target = null;

    // for internal development use
    public boolean dumpstack = false;
    public boolean timings = false;
    public boolean noVersionWarnings = false;

    public boolean prettyPrint = false;

    public boolean nosymbols = false;

    public boolean emacssym = false;

    /**
     * Makes the compiler likelier ot blow-up
     *  1. will show stack traces even if there's been an error
     *  2. does more things
     */
    public boolean torture = false;
    public boolean developer = false;

    // generate code that will cause some jvm's to reveal their bugs
    public boolean jvmTorture = false;


    // number of threads, threads==0 means disable multi-threaded
    // concurrent compilation
    public int threads = 0;
    public int procs = 1;

    // incremental compilation requested
    public boolean incremental = false;

    public boolean isStringOption(String key) {
        try {
            Field field = Options.class.getField(key);
            if (Modifier.isFinal(field.getModifiers())) return false;
            if (field.getType() != String.class) return false;
            return true;
        } catch (NoSuchFieldException e) {
            return false;
        }
    }

    public boolean isIntegerOption(String key) {
        try {
            Field field = Options.class.getField(key);
            if (Modifier.isFinal(field.getModifiers())) return false;
            if (field.getType() != Integer.TYPE) return false;
            return true;
        } catch (NoSuchFieldException e) {
            return false;
        }
    }

    public boolean set(String key, Object value) {
        try {
            Field field = Options.class.getField(key);
            if (Modifier.isFinal(field.getModifiers())) return false;
            if (Modifier.isStatic(field.getModifiers())) {
                field.set(null,value);
            } else {
                field.set(this,value);
            }
            return true;
        } catch (IllegalArgumentException e) {
            return false;
        } catch (NoSuchFieldException e) {
            return false;
        } catch (IllegalAccessException e) {
            throw new RuntimeException("internal error setting option: "+key);
        }
    }

    public String setOptions(String[] args, File currentWorkingDir) {
        for (int i = 0; i < args.length ; i++) {
            String arg = args[i];
            //System.out.println("arg: "+arg);
            String key = arg.substring(1);
            Object value = Boolean.TRUE;

            if (arg.equals("-d")) {
                if ((i+1) >= args.length) return arg;
                outputDir = qualifiedFile(args[++i], currentWorkingDir);
                continue;
            } else if (arg.equals("-workingdir")) {
                if ((i+1) >= args.length) return arg;
                workingDir = qualifiedFile(args[++i], currentWorkingDir);
                continue;
            } else if (isStringOption(key)) {
                if ((i+1) >= args.length) return arg;
                value = args[++i];
            }
            else if (isIntegerOption(key)) {
                if ((i+1) >= args.length) return arg;
                value = Integer.valueOf(args[++i]);
            }
            if (!set(key, value)) {
                return arg;
            }
        }
        return null;
    }

    private File qualifiedFile(String name, File currentWorkingDir) {
        name = name.replace('/', File.separatorChar);
        File file = new File(name);
        if (!file.isAbsolute() && currentWorkingDir != null) {
            file = new File(currentWorkingDir, name);
        }
        return file;
    }


}
