/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import java.util.*;

/**
 *
 * @grammar
 * @child AssignableExpr lhs    the value to be mutated
 */
public abstract class BangExpr extends Expr {
    public Type discoverType() {
        return lhs.getType();
    }

    public boolean isLegalStmt() {
        return true;
    }

    //INTRO from ASTFixerPass
    public ASTObject postFixAST(ASTFixerPass fixer) {
        if (!(getLhs() instanceof FieldAccessExpr)) return this;

        FieldAccessExpr fieldAccessExpr = (FieldAccessExpr)getLhs();
        Field field = fieldAccessExpr.getField();

        if (field == null) return this; //XXX should be an error

        if (field.getDeclaringType().isInterface() && !field.isStatic()) {
            return fixAccessToFieldSet(fieldAccessExpr).setSource(this);
        }

        return this;
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessPost(AccessFixer fixer) {
        if (!(getLhs() instanceof FieldAccessExpr)) return this;

        FieldAccessExpr fieldAccessExpr = (FieldAccessExpr)getLhs();
        Field field = fieldAccessExpr.getField();

        if (field == null) return this; //XXX should be an error
        //TypeDec fromTypeDec = getBytecodeTypeDec(); //fixer.fromTypeDec;
        if (field.isAccessible(this, true)) return this;

        getCompiler().showMessage("  fixing privileged set: " + field.toShortString());
//        System.out.println(field.getBytecodeTypeDec() + " not accessible from " +
//                        fixer.fromTypeDec);

        if (!fixer.apply) {
            field.getBackdoorGetterMethod();
            field.getBackdoorSetterMethod();
            return this;
        }

        return fixAccessToFieldSet(fieldAccessExpr).setSource(this);
    }

    //INTRO from AccessFixer
    public abstract ASTObject fixAccessToFieldSet(FieldAccessExpr fieldAccessExpr);

    //INTRO from AccessFixer
    Expr makeInPlaceSet(FieldAccessExpr expr, String op, Expr rhs) {
        Expr lhs = expr.getExpr();
        Field field = expr.getField();

        //VarDec tmp = lhs.getEnclosingCodeDec().getBody().makeTemporary(lhs.getType());
        Expr newExpr = lhs.makeReference();

        Method getMethod = field.getBackdoorGetterMethod();
        Method setMethod = field.getBackdoorSetterMethod();
        final AST ast = getAST();

        Expr newValue = ast.makeBinop(op, ast.makeCall(getMethod, newExpr), rhs);

        //Expr setExpr = ast.makeParen(ast.makeSet(ast.makeVar(tmp), lhs));
        return ast.makeCall(setMethod, lhs, newValue);
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    // this is overridden in AssignExpr, but is used for both ++x and x++.
    public void walkFlow(FlowCheckerPass w) {
        w.process(getLhs());
        if (getLhs() instanceof VarExpr) {
            VarExpr lhs = (VarExpr)getLhs();
            VarDec dec = lhs.getVarDec();
            if (dec.isFinal()
                && ((! dec.isBlank())
                    || ! w.getVars().isDefinitelyUnassigned(dec))) {
                w.showVarError(this, dec, "Final variable " + dec.getId() +
                               " already has a value");
            }
        } else if (getLhs() instanceof FieldAccessExpr) {
            FieldAccessExpr fa = (FieldAccessExpr) getLhs();
            FieldDec dec = fa.getFieldDec();
            if (dec.isFinal()
                && ((! dec.isBlank())
                    || ! w.getVars().isDefinitelyUnassigned(dec))) {
                w.showVarError(this, dec, "Final field " + dec.getId() +
                               " already has a value");
            }
        }
    }

    //BEGIN: Generated from @child and @property
    protected AssignableExpr lhs;
    public AssignableExpr getLhs() { return lhs; }
    public void setLhs(AssignableExpr _lhs) {
        if (_lhs != null) _lhs.setParent(this);
        lhs = _lhs;
    }

    public BangExpr(SourceLocation location, AssignableExpr _lhs) {
        super(location);
        setLhs(_lhs);
    }
    protected BangExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return lhs;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "lhs";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setLhs((AssignableExpr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }

    public String getDefaultDisplayName() {
        return "BangExpr()";
    }

    //END: Generated from @child and @property
}
