/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the source and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar rand1 <op> rand2
 * @child Expr rand1
 * @property String op
 * @child Expr rand2
 */
public abstract class BinopExpr extends Expr {
    /** Only called from {@link AssignExpr}, overrided in {@link
        PlusOpExpr}, since += is the only operator with weird typing
        behaviour. */
    void checkAssignmentOperatorType() {
        discoverType();
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass (defined in AspectJCompiler)

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        return tryToFold();
    }

    private ASTObject tryToFold() {
        if (getLiftType().isAnyType()) return this;
        if (rand1 instanceof LiteralExpr && rand2 instanceof LiteralExpr) {
            try {
                return
                    halfFold(getLiftType(), (LiteralExpr)rand1, (LiteralExpr)rand2)
                    .setSource(this);
            } catch (ArithmeticException e) {
                this.showError(e.getMessage());
                return this;
            }
        } else {
            return this;
        }
    }

    abstract protected Type getLiftType();

    /** returns a new unfinished ASTObject, that still needs to have
        {@link ASTObject#setSource} called on it. */
    abstract protected LiteralExpr halfFold(Type type, LiteralExpr lit1, LiteralExpr lit2);

    public static BinopExpr build(SourceLocation source, String op,
                                  Expr rand1, Expr rand2) {
        op = op.intern();
        if        (op == "+") {
            return new AddOpExpr(source, rand1, op, rand2);
        } else if (op == "&&") {
            return new AndAndOpExpr(source, rand1, op, rand2);
        } else if (op == "||") {
            return new OrOrOpExpr(source, rand1, op, rand2);
        } else if (op == "*" || op == "/" || op == "%" || op == "-") {
            return new NumericOpExpr(source, rand1, op, rand2);
        } else if (op == "&" || op == "|" || op == "^") {
            return new BitwiseOpExpr(source, rand1, op, rand2);
        } else if (op == "<<" || op == ">>" || op == ">>>") {
            return new ShiftOpExpr(source, rand1, op, rand2);
        } else if (op == "<" || op == "<=" || op == ">=" || op == ">") {
            return new NumericTestOpExpr(source, rand1, op, rand2);
        } else if (op == "==" || op == "!=") {
            return new EqualityTestOpExpr(source, rand1, op, rand2);
        } else {
            throw new RuntimeException("bad op " + op);
        }
    }

    public void unparse(CodeWriter writer) {
        writer.write(rand1);
        writer.writeOp(op);
        writer.write(rand2);
    }

    // ------------------------------
    // bcg
    /** Generate code that takes the top two values on the stack and
        replaces it with the one value that is appropriate.  This is
        only used by non-effect operators, and is sometimes called
        from {@link AssignExpr}. */
    abstract protected void cgOp(CodeBuilder cb, Type ty);

    // XXX this turns off effect optimization of + just for the bad case of
    //   ("foo" + new Blah()).staticField
    // there must be a better way to do this.
    protected void cgEffect(CodeBuilder cb) {
        if (op == "/" || this instanceof ConditionalOpExpr || op == "+") {
            super.cgEffect(cb);
        } else {
            getRand1().cgEffect(cb);
            getRand2().cgEffect(cb);
        }
    }

    //BEGIN: Generated from @child and @property
    protected Expr rand1;
    public Expr getRand1() { return rand1; }
    public void setRand1(Expr _rand1) {
        if (_rand1 != null) _rand1.setParent(this);
        rand1 = _rand1;
    }

    protected String op;
    public String getOp() { return op; }
    public void setOp(String _op) { op = _op; }

    protected Expr rand2;
    public Expr getRand2() { return rand2; }
    public void setRand2(Expr _rand2) {
        if (_rand2 != null) _rand2.setParent(this);
        rand2 = _rand2;
    }

    public BinopExpr(SourceLocation location, Expr _rand1, String _op, Expr _rand2) {
        super(location);
        setRand1(_rand1);
        setOp(_op);
        setRand2(_rand2);
    }
    protected BinopExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return rand1;
        case 1: return rand2;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "rand1";
        case 1: return "rand2";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setRand1((Expr)child); return;
        case 1: setRand2((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "BinopExpr(op: "+op+")";
    }

    //END: Generated from @child and @property
}
