/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

public final class BooleanType extends PrimitiveType {

    public String getName()      { return "boolean"; }
    public int    getTypeIndex() { return Type.BOOLEAN; }
    public Type   getRefType()   { return getTypeManager().getType("java.lang", "Boolean"); }
    public Expr   getNullExpr()  { return getAST().makeLiteral(false); }
    public String getDescriptor() { return "Z"; }

    public BooleanType(JavaCompiler compiler) {
        super(compiler);
    }

    // ------------------------------
    // folding
    final LiteralExpr foldPlusOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldMinusOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldBitNotOp(LiteralExpr rand) { return unsupportedFold(); }
    final LiteralExpr foldAddOp(LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldNumericOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldShiftOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }
    final LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) { return unsupportedFold(); }

    LiteralExpr foldLogNotOp(LiteralExpr rand) {
        return getAST().makeLiteral(! rand.getBooleanValue());
    }
    LiteralExpr foldCast(LiteralExpr rand) {
        return getAST().makeLiteral(rand.getBooleanValue());
    }

    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        boolean a = lit1.getBooleanValue();
        boolean b = lit2.getBooleanValue();
        if      (op == "==") return ast.makeLiteral(a == b);
        else if (op == "!=") return ast.makeLiteral(a != b);
        else throw new RuntimeException("bad op " + op);
    }
    LiteralExpr foldBitwiseOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        boolean a = lit1.getBooleanValue();
        boolean b = lit2.getBooleanValue();
        if      (op == "&") return ast.makeLiteral(a & b);
        else if (op == "|") return ast.makeLiteral(a | b);
        else if (op == "^") return ast.makeLiteral(a ^ b);
        else throw new RuntimeException("bad bitwise op " + op);
    }

    // ------------------------------
    // code generation
    final void emitCast(CodeBuilder cb, Type castTo) {}

    final void emitLogNot(CodeBuilder cb) { cb.emitIntConstant(1); cb.emitIXOR(); }

    final void emitAload(CodeBuilder cb) { cb.emitBALOAD(); }
    final void emitAstore(CodeBuilder cb) { cb.emitBASTORE(); }

    public final void emitLoad(CodeBuilder cb, int loc) { cb.emitILOAD(loc); }
    public final void emitStore(CodeBuilder cb, int loc) { cb.emitISTORE(loc); }
    public final void emitReturn(CodeBuilder cb) { cb.emitIRETURN(); }

    void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "==") cb.emitIF_ICMPEQ(t, f);
        else if (op == "!=") cb.emitIF_ICMPNE(t, f);
    }
    void emitBitwiseOp(CodeBuilder cb, String op) {
        if      (op == "&") cb.emitIAND();
        else if (op == "|") cb.emitIOR();
        else if (op == "^") cb.emitIXOR();
    }

    final void emitZero(CodeBuilder cb)       { unsupportedEmit(); }
    final void emitOne(CodeBuilder cb)        { unsupportedEmit(); }
    final void emitMinusOne(CodeBuilder cb)   { unsupportedEmit(); }
    final void emitAdd(CodeBuilder cb)        { unsupportedEmit(); }
    final void emitBitNot(CodeBuilder cb)     { unsupportedEmit(); }
    final void emitNumericOp(CodeBuilder cb, String op) { unsupportedEmit(); }
    final void emitShiftOp(CodeBuilder cb, String op)   { unsupportedEmit(); } 
    final void emitNeg(CodeBuilder cb)                  { unsupportedEmit(); }
    final void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) { 
        unsupportedEmit(); 
    }

    // ------------------------------
    // "fast" opcode handling

    boolean hasFastEqualityTestOp() { return true; }
    void emitFastEqualityTestOp(CodeBuilder cb, EqualityTestOpExpr e,
                                Label t, Label f) {
        Expr rand1 = e.getRand1();
        Expr rand2 = e.getRand2();
        String op = e.getOp();
        if (rand1 instanceof LiteralExpr) {
            if (rand1.isConstantTrue() ^ (op != "==")) rand2.cgTest(cb, t, f);
            else                                       rand2.cgTest(cb, f, t);
        } else if (rand2 instanceof LiteralExpr) {
            if (rand2.isConstantTrue() ^ (op != "==")) rand1.cgTest(cb, t, f);
            else                                       rand1.cgTest(cb, f, t);
        } else {
            rand1.cgValue(cb, this);
            rand2.cgValue(cb, this);
            emitEqualityCompare(cb, op, t, f);
        }
    }
}
