/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.base.bcg.CodeBuilder;
import java.util.*;


/**
  * @grammar modifier $(formals) throws throws
  * @child FormalDec enclosingInstanceFormal @before formals
  */
public class ConstructorDec extends CodeDec {

    public boolean isSuper() {
        ConstructorBody body = (ConstructorBody) getBody();
        return body.getConstructorCall().getIsSuper();
    }

    public ConstructorDec getNextConstructorDec() {
        ConstructorBody body = (ConstructorBody) getBody();
        return body.getConstructorCall().getConstructor().getConstructorDec();
    }

    //INTRO from ASTFixerPass
    public ASTObject postFixAST(ASTFixerPass fixer) {
        fixIntroducedDec(null);
        return this;
    }

    public String getId() { return "new"; }

    public TypeD getResultTypeD() {
        return getTypeManager().voidType.makeTypeD();
    }

    public boolean isFinal() { return true; }
    //XXX picking just one is SO hard
    public boolean isStatic() { return false; }

    public boolean appearsStaticToCaller() { return true; }

    public boolean hasThis() {
        return true;
    }

    private String getDeclaringTypeD() {
        if (getDeclaringType() == null) return "";
        //if (getDeclaringTypeDec().getNameTypeD() == null) return "";

        return getDeclaringType().toShortString();
    }

    public String toShortString() {
        return modifiers.toShortString()+" "+getDeclaringTypeName()+formals.toShortString();
    }

    public String getKind() { return "constructor"; }

    public SemanticObject makeCorrespondingSemanticObject() {
        return new Constructor(this);
    }

    public Constructor getConstructor() {
        return (Constructor)getCorrespondingSemanticObject();
    }

    public String toString() { return toShortString(); }


    String getDeclaringTypeName() {
        return getDeclaringType().getId();
//        ASTObject a = getParent();
//        if (a == null) return "";
//        a = a.getParent();
//        if (a == null || !(a instanceof TypeDec)) return "";
//        return ((TypeDec)a).getId();
    }


    FormalDec extraFormalDec = null;
    public FormalDec getExtraFormalDec() {
        if (extraFormalDec != null) return extraFormalDec;
        if (!isIntroduced()) return null;

        if (isPrivate()) {
            extraFormalDec =
              getAST().makeFormal(getLexicalType().getPrivateCookieType(), "ajc$cookie");
        } else if (isProtected()) {
            showError("protected introduction not allowed");
            return null;
        } else if (isPublic()) {
            return null;
        } else {
            extraFormalDec =
              getAST().makeFormal(getLexicalType().getPackageCookieType(), "ajc$cookie");
        }
        return extraFormalDec;
    }


    public Expr getExtraArgExpr() {
        FormalDec extraFormal = getExtraFormalDec();
        if (extraFormal == null) return null;

        return new CastExpr(getSourceLocation(), extraFormal.getTypeD(), getAST().makeNull());
    }

    protected void fixIntroducedAccessibility() {
        if (isPublic()) return;

        getFormals().add(getExtraFormalDec());
        getModifiers().setPublic(true);
    }


    protected void makeMixinConcrete(TypeDec onTypeDec) {
        CodeDec newDec = (CodeDec)this.copy();
        //??? any binding/contexts needed ???
        onTypeDec.getBody().add(newDec);
    }

    public void makePublicAccessible() {
        if (isIntroduced()) fixIntroducedAccessibility();
        else getModifiers().setPublic(true);
    }

    public void fixIntroducedDec(TypeDec onTypeDec) {
        if (!isIntroduced()) return;

        if (onTypeDec == null || getDeclaringType().getTypeDec() == onTypeDec) {
            fixIntroducedAccessibility();
        } else if (getDeclaringType().isInterface()) {
            // check that this is a top-most implementor
            Type superType = onTypeDec.getSuperClassType();
            if (!superType.getTypeDec().fromSource() ||
                    !superType.isSubtypeOf(getDeclaringType())) {
                makeMixinConcrete(onTypeDec);
            }
        }
    }

    //INTRO from MixinImplementationPass
    public ASTObject postImplementMixin(MixinImplementationPass fixer) {
        final TypeDec inTypeDec = getBytecodeTypeDec();

        if (!(inTypeDec instanceof InterfaceDec) || getBody() == null || isStatic()) return this;

        Set topmostImplementors =
            Type.filterTopTypes(Type.filterConcreteTypes(inTypeDec.getType().getSubTypes()));

        for (Iterator i = topmostImplementors.iterator(); i.hasNext(); ) {
            final Type implType = (Type)i.next();
            TypeDec implDec = implType.getTypeDec();

            CodeDec newDec = (CodeDec)fixer.copyToClass(this, implDec);
            newDec = getAST().makeInitializer(newDec.getBody().getStmts());

            //XXX the position to add is important, this algorithm is a total hack
            Decs decs = implDec.getBody();
            int j = decs.size()-1;
            for (; j >= 0; j--) {
                if (decs.get(j) instanceof InitializerDec) {
                    implDec.getBody().add(j,  newDec);
                    break;
                }
            }
            if (j < 0) implDec.getBody().add( newDec );

        }

        return null;
    }

    public void unparse(CodeWriter writer) {
//        fixIntroducedDec(null);
        setAllEnclosingTypes(null);

        writer.write(modifiers);
        writer.write(getDeclaringTypeName()); //ParentTypeD());

        if (getEnclosingInstanceFormal() == null) {
            writer.write(formals);
        } else {
            Formals f = (Formals) formals.copy();
            f.add(0, (FormalDec) getEnclosingInstanceFormal().copy());
            writer.write(f);
        }

        writeNames(writer, "throws", getThrows());

        if (!writer.isOnlySignatures()) {
            writer.optionalSpace();

            writer.write(body);
            //writer.newLine();
        }
    }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walker.enterCodeDec(this);
        walker.inConstructor();
        this.walk(walker);
        walker.leaveCodeDec();
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        walker.pushConstructorEnv(getFormals());
    }

    public ASTObject postThreading(LocalClassPass.ThreadingWalker walker) {
        walker.popEnv();
        return this;
    }

    // ------------------------------
    // INTRO from MemberClassMunger

    public ASTObject walkMemberMunger(MemberClassMunger w) {
        NameType currentType = w.currentType();
        if (currentType.isInner()) {
            final AST ast = getAST();
            NameType enclosingInstanceType = currentType.getEnclosingInstanceType();
            FormalDec formalDec =
                ast.makeFormal(enclosingInstanceType, "enclosing$instance");
            setEnclosingInstanceFormal(formalDec);

            ConstructorBody body = (ConstructorBody) getBody();
            ConstructorCallExpr call = body.getConstructorCall();

            if (call.getIsSuper()) {
                Stmt stmt = ast.makeStmt(ast.makeSet(ast.makeThis(currentType),
                                                     w.currentField(),
                                                     ast.makeVar(formalDec)));
                body.getStmts().add(0, stmt);
            }
        }
        return super.walkMemberMunger(w);
    }

    // ------------------------------
    // bcg
    String getInternalId() {
        return "<init>";
    }

    protected void cgCodeMember(CodeBuilder cb) {
        cb.setMaxFrame(getFrameSize());
        cb.enterBlock();
        // ??? do I really want to put the enclosing instance formal in the var table?
        if (getEnclosingInstanceFormal() != null) {
            cb.enterVar(getEnclosingInstanceFormal());
        }
        Formals f = getFormals();
        for (int i = 0, len = f.size(); i < len; i++) {
            cb.enterVar(f.get(i));
        }
        getBody().cgTop(cb);
        cb.exitBlock();
    }

    private String descriptor;
    public synchronized String getDescriptor() {
        if (descriptor == null) {
            descriptor = "(";
            if (getEnclosingInstanceFormal() != null) {
                descriptor += getEnclosingInstanceFormal().getType().getDescriptor();
            }
            for (Iterator i = getFormals().iterator(); i.hasNext(); ) {
                FormalDec dec = (FormalDec)i.next();
                descriptor += dec.getType().getDescriptor();
            }
            descriptor += ")";
            descriptor += getResultType().getDescriptor();
        }
        return descriptor;
    }

    private boolean isDeltaComputed = false;
    private int stackDelta;
    public synchronized int getStackDelta() {
        if (!isDeltaComputed) {
            // start with 1 or 2 things on the stack
            stackDelta = (getEnclosingInstanceFormal() == null) ? -1 : -2;
            // then we add the formals
            for (Iterator i = getFormals().iterator(); i.hasNext(); ) {
                FormalDec dec = (FormalDec)i.next();
                stackDelta -= dec.getType().getSlotCount();
            }
            // we're void, so we don't leave anything on the stack
            isDeltaComputed = true;
        }
        //System.err.println("(CDEC) S-delta for " + this + " is " + stackDelta);
        return stackDelta;
    }

    // ------------------------------
    // convienience constructors

    public ConstructorDec(SourceLocation location, Modifiers _modifiers, Formals _formals, TypeDs __throws, CodeBody _body) {
        this(location, _modifiers, null, _formals, __throws, _body);
    }

    //BEGIN: Generated from @child and @property
    protected FormalDec enclosingInstanceFormal;
    public FormalDec getEnclosingInstanceFormal() { return enclosingInstanceFormal; }
    public void setEnclosingInstanceFormal(FormalDec _enclosingInstanceFormal) {
        if (_enclosingInstanceFormal != null) _enclosingInstanceFormal.setParent(this);
        enclosingInstanceFormal = _enclosingInstanceFormal;
    }

    public ConstructorDec(SourceLocation location, Modifiers _modifiers, FormalDec _enclosingInstanceFormal, Formals _formals, TypeDs __throws, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _body);
        setEnclosingInstanceFormal(_enclosingInstanceFormal);
    }
    protected ConstructorDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ConstructorDec ret = new ConstructorDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (enclosingInstanceFormal != null) ret.setEnclosingInstanceFormal( (FormalDec)walker.process(enclosingInstanceFormal) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return enclosingInstanceFormal;
        case 2: return formals;
        case 3: return _throws;
        case 4: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "enclosingInstanceFormal";
        case 2: return "formals";
        case 3: return "throws";
        case 4: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setEnclosingInstanceFormal((FormalDec)child); return;
        case 2: setFormals((Formals)child); return;
        case 3: setThrows((TypeDs)child); return;
        case 4: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 5;
    }

    public String getDefaultDisplayName() {
        return "ConstructorDec()";
    }

    //END: Generated from @child and @property
}
