/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import org.aspectj.compiler.base.cst.*;

import java.util.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar dec*
  * @children Dec dec
  */

//!!! this should extend ASTObject...
public class Decs extends Stmt {

    public List getList() {
        return Arrays.asList(children);
    }

    public ListIterator iterator() {
        return new ListIterator() {
                private int i = -1;
                private boolean isRemovable = false;

                public int previousIndex() {
                    return i == -1 ? size - 1 :  i - 1;
                }
                public boolean hasPrevious() {
                    int i = previousIndex();
                    return i >= 0 && i < size;
                }
                public Object previous() {
                    if (! hasPrevious()) {
                        throw new NoSuchElementException();
                    }
                    isRemovable = true;
                    return children[--i];
                }

                public int nextIndex() {
                    return i == -1 ? 0 : i + 1;
                }
                public boolean hasNext() {
                    int i = nextIndex();
                    return i >= 0 && i < size;
                }
                public Object next() {
                    if (! hasNext()) {
                        throw new NoSuchElementException();
                    }
                    isRemovable = true;
                    return children[++i];
                }

                public void remove() {
                    if (! isRemovable) {
                        throw new IllegalStateException();
                    }
                    isRemovable = false;
                    Decs.this.remove(i--);
                }
                public void set(Object o) {
                    if (! isRemovable) {
                        throw new IllegalStateException();
                    }
                    Decs.this.set(i, (Dec)o);
                }
                public void add(Object o) {
                    if (i == -1) i = 0;
                    isRemovable = false;
                    Decs.this.add(i++, (Dec)o);
                }
            };
    }

    //!!! should go away
    public void append(Decs decs) {
        for(int i=0; i<decs.size(); i++) {
            append(decs.get(i));
        }
    }

    public void append(Dec dec) {
        add(dec);
    }

    public void prepend(Dec dec) {
        add(0, dec);
    }

    public void unparse(CodeWriter writer) {
        final int N = size;
        for (int i = 0; i < N; i++) {
            if (i > 0) writer.newLine();
            writer.write(children[i]);
        }
    }

    public void remove(Dec dec) {
        int i = indexOf(dec);
        if (i != -1) remove(i);
    }

    public void cleanup() {
        for(int i=0; i < size; i++) {
            children[i].cleanup();
        }
        super.cleanup();
    }

    // ------------------------------
    // bcg
    // should only be a list of varDecs here...
    protected void cgStmt(CodeBuilder cb) {
        for (int i = 0, len = size; i < len; i++) {
            children[i].cgTop(cb);
        }
    }

    //BEGIN: Generated from @child and @property
    protected int size;
    public Dec[] children;

    public Decs(SourceLocation location, Dec[] _children) {
        super(location);
        for(int i=0; i<_children.length; i++) {
            if (_children[i] != null) _children[i].setParent(this);
        }
        children = _children;
        size = _children.length;
    }

    public Decs(SourceLocation location) {
        this(location, new Dec[] {});
    }

    public Decs(SourceLocation location, Dec child1) {
        this(location, new Dec[] {child1});
    }

    public Decs(SourceLocation location, Dec child1, Dec child2) {
        this(location, new Dec[] {child1, child2});
    }

    public Decs(SourceLocation location, Dec child1, Dec child2, Dec child3) {
        this(location, new Dec[] {child1, child2, child3});
    }

    public ASTObject copyWalk(CopyWalker walker) {
        final int N = size;
        Dec[] copiedChildren = new Dec[N];
        int newIndex = 0;
        for(int oldIndex=0; oldIndex<N; oldIndex++) {
            Dec newChild = (Dec)walker.process(children[oldIndex]);
            if (newChild != null) copiedChildren[newIndex++] = newChild;
        }
        Decs ret = new Decs(getSourceLocation(),copiedChildren);
        ret.size = newIndex;
        ret.setSource(this);
        return ret;
    }

    public ASTObject getChildAt(int childIndex) { return get(childIndex); }
    public void setChildAt(int childIndex, ASTObject child) { set(childIndex, (Dec)child); }
    public String getChildNameAt(int childIndex) { return "dec"+childIndex; }

    public int getChildCount() { return size; }
    public int size() { return size; }

    public Dec get(int index) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        return children[index];
    }

    public void set(int index, Dec child) {
        if (index >= size) throw new ArrayIndexOutOfBoundsException();
        children[index] = child;
        child.setParent(this);
    }

    public void resize(int newSize) {
        if (newSize > children.length) {
            Dec[] newChildren = new Dec[children.length*2 + 1];
            System.arraycopy(children, 0, newChildren, 0, children.length);
            children = newChildren;
        }
        size = newSize;
    }

    public void addAll(Decs collection) {
        addAll(size, collection);
    }

    public void addAll(int index, Decs collection) {
        for(int i=0; i<collection.size(); i++) {
            add(index+i, collection.get(i));
        }
    }

    public void add(Dec child) {
        add(size, child);
    }

    public void add(int index, Dec child) {
        if (child == null) return;

        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        resize(size+1);

        for(int moveIndex = size-1; moveIndex > index; moveIndex--) {
            children[moveIndex] = children[moveIndex-1];
        }

        children[index] = child;
        child.setParent(this);
    }

    public void remove(int index) {
        if (index < 0 || index > size) throw new ArrayIndexOutOfBoundsException();

        size -= 1;

        for(int moveIndex = index; moveIndex < size; moveIndex++) {
            children[moveIndex] = children[moveIndex+1];
        }
    }

    public String getDefaultDisplayName() {
        return "Decs()";
    }

    //END: Generated from @child and @property
}
