/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar ...
  */
public abstract class Expr extends ASTObject {
    /** is this expression zero or null? */
    public boolean isConstantZero() { return false; }

    /** is this expression the boolean literal value true? */
    public boolean isConstantTrue() { return false; }

    /** is this expression the boolean literal value false? */
    public boolean isConstantFalse() { return false; }

    /** Normally this is a type consideration, but this is overridden
        by {@link IntLiteralExpr} to handle JLS 5.2's requirement that
        some int literals are assignable to byte, short, and char
        variables. */
    public boolean isAssignableTo(Type type) {
        return type.isAssignableFrom(getType());
    }

    /** Method invocation conversion, JLS 5.3, is the same as
        assignment conversion without the numeric narrowing
        conversion.  So all this does is delegate to the purely
        type-based {@link Type#isAssignableFrom} method.
    */
    public final boolean isMethodConvertableTo(Type type) {
        return type.isAssignableFrom(getType());
    }

    public boolean isLegalStmt() {
        return false;
    }

    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        if (this.type == null) {
            this.type = ((Expr)oldObject).type;
        }
        return super.postCopy(walker, oldObject);
    }

    //XXX total hack to avoid understanding inners too much
    protected Expr makeQualifiedThis(Type declaringType) {
        if (getLexicalType() != null && getLexicalType().isInnerType()) {
            return getAST().makeQualifiedThis(declaringType);
        } else {
            return getAST().makeThis(declaringType);
        }
    }

    public boolean isInExprStmt() {
        ASTObject parent = this.getParent();
        if (parent instanceof ExprStmt) return true;
        if (parent instanceof Exprs) {
            ASTObject grandparent = parent.getParent();
            return (grandparent instanceof ForStmt ? true : false);
        }
        return false;
    }

    public void checkSpec() {
        if (type == null) type = discoverType();
    }

    public void assertType(Type type) {
        if (!(type.isAssignableFrom(getType()))) {
            showTypeError(getType(), type);
        }
    }

    protected Type type;
    public Type getType() {
        if (type == null) type = discoverType();
          if (type == null) {
              //this.showError("my type is null");
//              System.out.println("My type is null: ");
//              this.display(2);
          }
        return type;
    }

    //XXX used when?
    public void setType(Type type) {
        this.type = type;
    }

    public void showOperatorTypeError(String op, Type a, Type b) {
        if (a.isMissing() || b.isMissing()) return;
        showError("operator " + op + " cannot be applied to " + a.getString() + "," + b.getString());
    }

    public Expr makeReference() {
        return new ReferenceExpr(getSourceLocation(), this);
    }

    public boolean canBeCopied() {
        return false;
    }


    abstract protected Type discoverType();

//    public ASTObject finishCopy(ASTObject newObject) {
//        if (newObject instanceof Expr) {
//            ((Expr)newObject).type = this.type;
//        }
//        return super.finishCopy(newObject);
//    }

    public void cleanup() {
        type = null;
        super.cleanup();
    }

    // ------------------------------
    // Intro from ByteCodeCleanupPass

    // we don't need to clean up expressions, so don't traverse them.
    public void walkCleanup(ByteCodeCleanupPass w) {
        return;
    }

    // ------------------------------
    // bcg

    /** Leaves the value of this expression on the stack.  May self-call
     {@link #cgTest(CodeBuilder, Label, Label)}.  */
    protected void cgValue(CodeBuilder cb) {
        Label trueLab = cb.genLabel();
        Label falseLab = cb.genLabel();
        Label endLab = cb.genLabel();
        cgTest(cb, trueLab, falseLab);
        cb.emitLabel(trueLab);
        cb.emitIntConstant(1);
        cb.emitJump(endLab);
        cb.emitLabel(falseLab);
        cb.popStack(1);
        cb.emitIntConstant(0);
        cb.emitLabel(endLab);
    }

    /** Leaves the value of this expression on the stack and casts the
        value to castTo. May self-call {@link #cgValue(CodeBuilder)}.  */
    protected void cgValue(CodeBuilder cb, Type castTo) {
        cgValue(cb);
        getType().emitCast(cb, castTo);
    }

    /** Evaluates the expression in a context where its value is
        unnecessary (i.e., from {@link ExprStmt}, or in a {@link
        FieldAccessExpr} when the field is static but there is an
        expression there nonetheless).  May self-call {@link
        #cgValue(CodeBuilder)}.  */
    protected void cgEffect(CodeBuilder cb) {
        cgValue(cb);
        getType().emitPop(cb);
    }

    /** Evaluates the expression in a context where the top of the
        stack contains a {@link StringBuffer} object.  The value of
        this expression is then appended into the StringBuffer.  This
        is used in evaluating the String + operator.  May self-call
        {@link #cgValue(CodeBuilder)}.  */
    protected void cgBuffer(CodeBuilder cb) {
        cgValue(cb);
        NameType stringBufferType = getTypeManager().getStringBufferType();
        Type ty = getType();
        if (ty instanceof ByteType || ty instanceof ShortType) {
            ty = getTypeManager().intType;
        } else if (ty.isString()) {
        } else if (ty instanceof RefType) {
            ty = getTypeManager().getObjectType();
        }
        cb.emitINVOKEVIRTUAL(stringBufferType, "append",
                             "(" + ty.getDescriptor() + ")Ljava/lang/StringBuffer;",
                             - ty.getSlotCount());
    }

    /** Evaluates the expression in a context where its value will
        only be used to branch (from {@link IfStmt}, {@link
        TriTestExpr}, {@link WhileStmt}, {@link DoStmt}, and {@link
        ForStmt}).  The expression, of course, will have boolean type.
        By default, this evaluates the expression to the stack and
        then emits an EQ/NE branch bytecode, but many expressions such
        as tests in {@link BinopExpr} will be able to do much better.
        May self-call {@link #cgValue(CodeBuilder)}.  */
    protected void cgTest(CodeBuilder cb, Label tdest, Label fdest) {
        cgValue(cb);
        cb.emitIFNE(tdest, fdest);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    // We don't assign frame locations here
    public void walkFrameLoc(FrameLocPass walker) {
    }

    //BEGIN: Generated from @child and @property

    public Expr(SourceLocation location) {
        super(location);

    }


    public String getDefaultDisplayName() {
        return "Expr()";
    }

    //END: Generated from @child and @property
}
