/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.AccessFixer;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar expr.name;
  * @child Expr expr
  * @property Field field
  * @property boolean isSuper
  */
public class FieldAccessExpr extends AssignableExpr implements SOLink {
    public FieldDec getFieldDec() { return (FieldDec)field.getCorrespondingDec(); }

    protected Type discoverType() {
        return field.getFieldType();
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        Expr e = getExpr();
        FieldDec dec = getFieldDec();
        if (e != null) w.process(e);
        if (dec.isFinal()
                && isBare()
                && w.isCurrent(dec)
                && ! w.getVars().isDefinitelyAssigned(dec)) {
            showError("Field " + getId() + " might not have a value");
        }
    }

    public boolean isBare() {
        Expr e = getExpr();
        if (e == null) return true;
        if (getField().isStatic()) {
            return e instanceof TypeExpr && e.isSynthetic();
        } else {
            return e instanceof ThisExpr
                || ((e instanceof QualifiedThisExpr) && e.isSynthetic());
        }
    }

    // ------------------------------
    // INTRO from InnerInfoPass -- should have a more generic name

    public void postInnerInfo(InnerInfoPass w) {
        if (expr == null /* || expr.isSynthetic() */) {
            w.checkStaticAccess(this, field);
        }
        if (expr == null) {
            setExpr(getAST().makePrimary(getField(), w.currentType()));
        }
    }

    /**
     * Section 6.6.2.1 from JLSv2
     * E.Id requires that E is subtype of S
     */
    public boolean hasLegalProtectedAccess(Type fromType) {
        //this.display(0);
        if (expr == null || getIsSuper()) return true;
        return expr.getType().isSubtypeOf(fromType);
    }


    public void checkSpec() {
        if (field == null) return;

        if (field.getModifiers().isProtected() && !field.isAccessible(this)) {
            showError(field.toShortString() + " has protected access");
        }

        if (expr instanceof TypeExpr && !field.isStatic()) {
            showError("non-static field " + field.toShortString() +
                        " cannot be accessed through a static reference");
        }
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        super.walkForwardReference(w);
        if (isLhs() && (getParent() instanceof BasicAssignExpr))
            return;
        if (getExpr() != null && ! getExpr().isSynthetic()) return;
        //showWarning("synthetic expr");
        w.checkReference(this);
    }


    // ------------------------------
    // INTRO from AssignmentCheckerPass (defined in AspectJCompiler)

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        if (isLhs()) return this;
        Field field = getField();
        FieldDec fieldDec = getFieldDec();
        fieldDec.ensureFolded(walker);
        Expr initExpr = fieldDec.getInitializer();

        if (fieldDec.isConstant() && isLegalConstantPrimary(getExpr())) {
            return
                fieldDec.getType().foldCast((LiteralExpr)initExpr).setSource(this);
        } else {
            return this;
        }
    }

    private boolean isLegalConstantPrimary(Expr expr) {
        if (expr == null) return true;

        if (expr instanceof ParenExpr) {
            return isLegalConstantPrimary(((ParenExpr)expr).getExpr());
        }
        if (expr instanceof TypeExpr) {
            return true;
        }
        // ignore generated ThisExprs for purposes of folding
        if (expr.isSynthetic() || !expr.fromSource()) {
            return (expr instanceof ThisExpr);
        }

        return false;
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessPost(AccessFixer fixer) {
        if (isLhs()) return this; // will be handled by the assign expr I'm in

        Field field = getField();
        if (field == null) return this;  //XXX should be an error
        //TypeDec fromTypeDec = getBytecodeTypeDec(); //fixer.fromTypeDec;
        //getCompiler().showMessage("  isAccessible? " + field.toShortString());
        if (field.isAccessible(this, true)) return this;

        getCompiler().showMessage("  fixing privileged get: " + field.toShortString());

        Method method = field.getBackdoorGetterMethod();
        if (!fixer.apply) return this;

        return getAST().makeCall(method, getExpr()).setSource(this);
    }

    public FieldAccessExpr(SourceLocation source, Type type, String name) {
        this(source, null, type.getField(name, null, true), false);
    }

    public FieldAccessExpr(SourceLocation source, Expr expr, String name) {
        this(source, expr, expr.getType().getField(name, expr, true), false);
    }

    public String getId() { return getField().getId(); }

    //INTRO from MovingWalker
    public ASTObject postMove(MovingWalker walker) {
        return walker.moveLinkExpr(this);
    }

    //INTRO from ASTFixerPass
    public ASTObject postFixAST(ASTFixerPass fixer) {
        // handle introduced statics on interfaces
        setExpr(field.updateTargetExpr(getExpr()));

        if (!isLhs() && getField().getDeclaringType().isInterface() && !getField().isStatic()) {
            return getAST().makeCall(getField().getBackdoorGetterMethod(), getExpr());
        } else {
            return this;
        }
    }

    //SOLink interface
    public SemanticObject getTarget() {
        return getField();
    }

    public void setTarget(SemanticObject newTarget) {
        setField((Field)newTarget);
    }

    public Type getTargetType() {
        Type declaringType = getField().getDeclaringType();
        if (declaringType.isInterface() && ! getOptions().strict) {
            // dealing with buggy sun jdk 1.2.2 vm
            return declaringType;
        }
        if (expr != null) 
        	declaringType = expr.getType();
        if (getIsSuper()) 
        	declaringType = ((NameType)declaringType).getSuperClassType();
        return declaringType;
    }

    public void unparse(CodeWriter writer) {
    	if (getIsSuper()) {
    		//??? add a test that expr is appropriate for this
    		writer.write(expr.getType().getId());
    		writer.write(".super.");
    	} else {
    		writer.write(expr);
    		writer.write('.');
    	}
        writer.write(getField().getBytecodeId());
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        if (isLhs()) return this;

        Expr q = getExpr();
        FieldDec dec = getFieldDec();
        if (w.isAccessible(dec, q)) return this;

        final AST ast = getAST();
        Type qType = q.getType();
        MethodDec newMethodDec = w.getAccessMethod(qType, dec, "x", this);
        Exprs newArgs = ast.makeExprs();
        Expr newExpr = w.makeOutsidePrimary(dec.isStatic(), newArgs, q);
        return ast.makeCall(newMethodDec, newExpr, newArgs);
    }

    public MethodDec buildAccessMethod(InnerAccessFixer w) {
        final AST ast = getAST();
        Expr q = getExpr();
        Type qType = q.getType();
        FieldDec dec = getFieldDec();
        Type fieldType = dec.getType();

        Formals newFormals = ast.makeFormals();
        Expr newExpr = w.makeInsidePrimary(dec.isStatic(), newFormals, qType);

        return
            w.makeAccessMethod(fieldType,
                               newFormals,
                               ast.makeGet(newExpr, dec));
    }

    // ------------------------------
    // bcg

    protected void cgLvalue(CodeBuilder cb) {
        if (field.isStatic()) expr.cgEffect(cb);
        else                  expr.cgValue(cb);
    }
    protected void cgLtoRvalue(CodeBuilder cb) {
        if (expr.getType() instanceof ArrayType) {
            cb.emitARRAYLENGTH();
        } else {
            NameType qualifyingType = (NameType) getTargetType();
            String name = field.getBytecodeId();
            String descriptor = field.getDescriptor();
            int delta = getType().getSlotCount();
            if (field.isStatic()) {
                cb.emitGETSTATIC(qualifyingType, name, descriptor, delta);
            } else {
                cb.emitGETFIELD(qualifyingType, name, descriptor, delta - 1);
            }
        }
    }
    protected void cgAssignment(CodeBuilder cb) {
        NameType qualifyingType = (NameType) getTargetType();     
        String name = field.getBytecodeId();
        String descriptor = field.getDescriptor();
        int delta = getType().getSlotCount();
        if (field.isStatic()) {
            cb.emitPUTSTATIC(qualifyingType, name, descriptor, - delta);
        } else {
            cb.emitPUTFIELD(qualifyingType, name, descriptor, - delta - 1);
        }
    }
    protected void cgDupLvalue(CodeBuilder cb) {
        if (! field.isStatic()) cb.emitDUP();
    }
    protected void cgDupRvalue(CodeBuilder cb) {
        if (field.isStatic()) getType().emitDup(cb);
        else                  getType().emitDupX1(cb);
    }    
    
    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }
    
    protected Field field;
    public Field getField() { return field; }
    public void setField(Field _field) { field = _field; }
    
    protected boolean isSuper;
    public boolean getIsSuper() { return isSuper; }
    public void setIsSuper(boolean _isSuper) { isSuper = _isSuper; }
    
    public FieldAccessExpr(SourceLocation location, Expr _expr, Field _field, boolean _isSuper) {
        super(location);
        setExpr(_expr);
        setField(_field);
        setIsSuper(_isSuper);
    }
    protected FieldAccessExpr(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        FieldAccessExpr ret = new FieldAccessExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        ret.field = field;
        ret.isSuper = isSuper;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "FieldAccessExpr(field: "+field+", "+"isSuper: "+isSuper+")";
    }
    
    //END: Generated from @child and @property
}
