/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.util.JavaStrings;
import org.aspectj.compiler.base.cst.*;

// to get the passes for introductions
import org.aspectj.compiler.crosscuts.*;

import org.aspectj.compiler.base.bcg.Label;
import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.ClassfileBuilder;
import org.aspectj.compiler.base.bcg.FieldBuilder;

import java.util.*;

/**
  * @grammar modifiers typeD id initializer
  *
  */

public class FieldDec extends VarDec {
    public Dec getEnclosingDec() {
        return this;
    }

    /**
    * We need to do this because we can't properly do name binding of
    * a reference to a field until we've done constant folding.
    * Actually what we can't do is type checking of assignment
    * statements until we've done folding.
    *
    * This isn't true for method look-up...
    *
    * <pre> static final int x = 2;
    *       byte b = x + 1; // this can't be type checked until we know whether or not x
    *                       // is a constant
    * </pre>
    *
    * We can fix this by splitting into:
    *   name-binding
    *   constant-folding: this unfortunately can't be just a walker
    *   assignment type checking */

    // ------------------------------
    // INTRO from AssignmentCheckerPass (defined in AspectJCompiler)

    public void walkAssignmentCheck(AssignmentCheckerPass walker) {
        this.ensureFolded(walker);
    }

    private boolean isFoldedFlag = false;
    public boolean isFolded() {
        //System.err.println("in FieldDec E " + this + " " + isFoldedFlag);
        return isFoldedFlag;
    }
    private void setFolded(boolean isFoldedFlag) {
        this.isFoldedFlag = isFoldedFlag;
    }

    public /* syncrhonized */ void ensureFolded(AssignmentCheckerPass walker) {
        //System.err.println("in FieldDec C " + this);
        if (! isFolded()) {
            //System.err.println("    in FieldDec D ");
            setFolded(true);
            preAssignmentCheck(walker);
            this.walk(walker);
            postAssignmentCheck(walker);
        }
    }
    // end intro

    public boolean isField() {
        return true;
    }

    public String toShortString() {
        return addIntroducedFromType(
             getDeclaringType().toShortString() + "." + getId());
    }

    public SemanticObject makeCorrespondingSemanticObject() {
        return new Field(this);
    }

    public Field getField() { return (Field)getCorrespondingSemanticObject(); }

    public String getKind() { return "field"; }

    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (getBytecodeTypeDec() instanceof InterfaceDec) {
            getModifiers().setValue(Modifiers.FINAL|Modifiers.PUBLIC|Modifiers.STATIC);
        }
        //??? don't output any blank finals to javac or .class files for now
        if (isFinal() && initializer == null) {
            modifiers.setFinal(false);
        }
        return this;
    }

    //INTRO from MixinImplementationPass
    public ASTObject postImplementMixin(MixinImplementationPass fixer) {
        final TypeDec inTypeDec = getBytecodeTypeDec();

        if (!(inTypeDec instanceof InterfaceDec)) return this;

        if (isFinal() && isStatic()) return this;

        if (isStatic()) {
            ((InterfaceDec)inTypeDec).addToHelperClass(this);
            return null;
        }

        Type onInterfaceType = getDeclaringType();

        getBackdoorSetterMethod().setBody(null);
        getBackdoorGetterMethod().setBody(null);

        Set topmostImplementors =
            Type.filterTopTypes(Type.filterConcreteTypes(inTypeDec.getType().getSubTypes()));

        for (Iterator i = topmostImplementors.iterator(); i.hasNext(); ) {
            final Type implType = (Type)i.next();
            TypeDec implDec = implType.getTypeDec();

            FieldDec newDec = (FieldDec)fixer.copyToClass(this, implDec);
            implDec.getBody().add(0,  newDec);
            newDec.setDeclaringType(implType); //this.getLexicalType());
            //newDec.setLexicalType(implType);
            newDec.getBackdoorGetterMethod();
            newDec.getBackdoorSetterMethod();
        }

        return null;
    }

    //XXX We need to better handle the block/code scope of field initializers
    //INTRO from ScopePass
    public void walkScope(ScopeWalker walker) {
        //XXX there must be a more elegant way to do this
        Expr saveInitializer = initializer; initializer = null;
        super.walkScope(walker);
        initializer = saveInitializer;

        if (!walker.walkBodies() || initializer == null) return;

        walker.pushScope(makeBlockScope(walker));
        //System.out.println("walking scope for init: " + initializer);
        setInitializer((Expr)walker.process(initializer));
//        System.out.println("      walked: " + initializer + ", " + initializer.unparse() +
//            ", " + initializer.getType());
        walker.popScope();
    }

    private String makeBackdoorId(String kind) {
        //??? need to add type into mangled name for no inheritance surprises
        String ret = getBytecodeId() + "_aj" + kind;
        return getAST().makeGeneratedName(ret);
    }


    /** This refers backdoor method for privileged aspects and
     *  for access from inner classes (for bytecode compilation)
     *  for getting field's value
     */
    private MethodDec backdoorGetterMethod;

    /** This refers backdoor method for privileged aspects and
     *  for access from inner classes (for bytecode compilation)
     *  for setting field's value
     */
    private MethodDec backdoorSetterMethod;

    public MethodDec getBackdoorGetterMethod() {
        if (backdoorGetterMethod != null) return backdoorGetterMethod;

        final AST ast = getAST();

        String id = makeBackdoorId("get");
        Modifiers modifiers = ast.makeModifiers(Modifiers.PUBLIC);
        if (isStatic()) modifiers.setStatic(true);

        Formals formals = ast.makeFormals();
        Expr thisExpr;
        if (isStatic()) thisExpr = ast.makeTypeExpr(getBytecodeTypeDec().getType());
        else thisExpr = ast.makeThis(getBytecodeTypeDec().getType());
        Stmt stmt = ast.makeReturn(ast.makeGet(thisExpr, this));

        MethodDec ret = ast.makeMethod(modifiers, getTypeD().getType(), id, formals,
                   ast.makeBlock(stmt));

        getBytecodeTypeDec().addToBody(ret);

        backdoorGetterMethod = ret;
        return backdoorGetterMethod;
    }

    public MethodDec getBackdoorSetterMethod() {
        if (backdoorSetterMethod != null) return backdoorSetterMethod;

        final AST ast = getAST();

        String id = makeBackdoorId("set");
        Modifiers modifiers = ast.makeModifiers(Modifiers.PUBLIC);
        if (isStatic()) modifiers.setStatic(true);

        FormalDec newValue = ast.makeFormal(getTypeD().getType(), "newValue");
        Formals formals = ast.makeFormals(newValue);

        Stmt stmt = ast.makeReturn(ast.makeSet(this, ast.makeVar(newValue)));

        MethodDec ret = ast.makeMethod(modifiers, getTypeD().getType(), id, formals,
                   ast.makeBlock(stmt));

        getBytecodeTypeDec().addToBody(ret);

        backdoorSetterMethod = ret;
        return backdoorSetterMethod;
    }

    public void unparse(CodeWriter writer) {
        super.unparse(writer);
    }

    public boolean isConstant() {
        // according to section 15.28 of the JLS non-static finals are constants
        return isFinal() && /*isStatic() &&*/ !getType().isObject() &&
            initializer != null && initializer instanceof LiteralExpr;
    }

    private AssignExpr assignExpr = null;

    public AssignExpr getAssignExpr() {
        if (assignExpr != null) return assignExpr;

        //XXX this is a mistake
        if (isStatic() && getDeclaringType().isInterface() &&getOptions().usejavac) {
            return null;
        }

        Expr fieldInit = getInitializer();
        final AST ast = getAST();

        if (fieldInit != null && !isConstant()) {
            setInitializer(null);
            if (isFinal() && isSynthetic()) getModifiers().setFinal(false);
            fieldInit.clearParent();
            assignExpr = isStatic() ? ast.makeSet(this, fieldInit)
                                    : ast.makeSet(ast.makeThis(getBytecodeTypeDec().getType()), this, fieldInit);
            Stmt stmt = new FromDecStmt(ast.getSourceLocation(),
                ast.makeStmt(assignExpr), this);

            Stmts stmts = ast.makeStmts(stmt);
            Dec init;
            if (isStatic()) init = ast.makeStaticInitializer(stmts);
            else init = ast.makeInitializer(stmts);

            init.setSource(this);

            int index = getBytecodeTypeDec().getBody().indexOf(this);
            getBytecodeTypeDec().getBody().add(index+1, init);
            assignExpr.setSource(this); //Location(getSourceLocation());
            return assignExpr;
        } else {
            return null;
        }
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        this.walk(w.createInitializerChecker(isStatic()));
        w.addFieldDec(this);
    }

    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inMember(isStatic());
        super.walkInnerInfo(w);
        w.restoreContext(context);
    }

    // ------------------------------
    // Intro:  FrameLocPass

    // We don't assign frame locations here

    public void walkFrameLoc(FrameLocPass walker) {
    }

    // ------------------------------
    // bcg
    /** This method takes care of generating code for Member decs.  At
        this point, we should not have initializers unless they're
        constants.  */
    protected void cgMember(ClassfileBuilder cb) {
        if (getOptions().bcgverbose) System.err.println("generating for " + this);
        FieldBuilder fb = cb.getFieldBuilder();
        fb.addAccessFlags(getModifiers().getValue());
        fb.setName(getBytecodeId());
        fb.setDescriptor(getDescriptor());
        if (isDeprecated()) fb.setDeprecated();
        //XXX get this right -- if (isSynthetic()) fb.setSynthetic();
        Expr init = getInitializer();
        if (init != null) {
            if (!(init instanceof LiteralExpr)) {
                this.showError("expected literal here: " + toShortString());
            }
            getType().foldCast((LiteralExpr) init).addConstant(fb);
        }
        cb.addField(fb);
    }

    //BEGIN: Generated from @child and @property

    public FieldDec(SourceLocation location, Modifiers _modifiers, TypeD _typeD, String _id, Expr _initializer) {
        super(location, _modifiers, _typeD, _id, _initializer);

    }
    protected FieldDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        FieldDec ret = new FieldDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (typeD != null) ret.setTypeD( (TypeD)walker.process(typeD) );
        ret.id = id;
        if (initializer != null) ret.setInitializer( (Expr)walker.process(initializer) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "FieldDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
