/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

public final class FloatType extends RealType {

    protected boolean isAssignableFromOtherType(Type other) {
        return (other instanceof NumericType) && ! (other instanceof DoubleType);
    }

    public String getName()      { return "float"; }
    public int    getTypeIndex() { return Type.FLOAT; }
    public Type   getRefType()   { return getTypeManager().getType("java.lang", "Float"); }
    public Expr   getNullExpr()  { return new FloatLiteralExpr(getAST().getSourceLocation(), 0); }

    public FloatType(JavaCompiler compiler) {
        super(compiler);
    }

    // vm internals
    public String getDescriptor() { return "F"; }

    // folding
    private LiteralExpr makeLit(float f) {
        return new FloatLiteralExpr(getAST().getSourceLocation(), f);
    }
    private LiteralExpr makeBooleanLit(boolean b) {
        return getAST().makeLiteral(b);
    }

    LiteralExpr foldCast(LiteralExpr rand) {
        return makeLit(rand.getFloatValue());
    }

    strictfp LiteralExpr foldPlusOp(LiteralExpr rand) {
        return makeLit(+ rand.getFloatValue());
    }
    strictfp LiteralExpr foldMinusOp(LiteralExpr rand) {
        return makeLit(- rand.getFloatValue());
    }

    strictfp LiteralExpr foldAddOp(LiteralExpr rand1, LiteralExpr rand2) {
        return makeLit(rand1.getFloatValue() + rand2.getFloatValue());
    }

    strictfp LiteralExpr foldNumericOp(String op,
                                       LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        float a = lit1.getFloatValue();
        float b = lit2.getFloatValue();
        if      (op == "-") return ast.makeLiteral(a - b);
        else if (op == "%") return ast.makeLiteral(a % b);
        else if (op == "/") return ast.makeLiteral(a / b);
        else if (op == "*") return ast.makeLiteral(a * b);
        else throw new RuntimeException("bad numeric op " + op);
    }

    LiteralExpr foldEqualityTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        float a = lit1.getFloatValue();
        float b = lit2.getFloatValue();
        if      (op == "==") return ast.makeLiteral(a == b);
        else if (op == "!=") return ast.makeLiteral(a != b);
        else throw new RuntimeException("bad op " + op);
    }

    LiteralExpr foldNumericTestOp(String op, LiteralExpr lit1, LiteralExpr lit2) {
        final AST ast = getAST();
        float a = lit1.getFloatValue();
        float b = lit2.getFloatValue();
        if      (op == "<")  return ast.makeLiteral(a < b);
        else if (op == "<=") return ast.makeLiteral(a <= b);
        else if (op == ">=") return ast.makeLiteral(a >= b);
        else if (op == ">")  return ast.makeLiteral(a > b);
        else throw new RuntimeException("bad op " + op);
    }

    // code gen

    void emitAdd(CodeBuilder cb) { cb.emitFADD(); }

    void emitNumericOp(CodeBuilder cb, String op) {
        if      (op == "-") cb.emitFSUB();
        else if (op == "%") cb.emitFREM();
        else if (op == "/") cb.emitFDIV();
        else if (op == "*") cb.emitFMUL();
    }

    void emitNeg(CodeBuilder cb) { cb.emitFNEG(); }

    public void emitLoad(CodeBuilder cb, int loc) { cb.emitFLOAD(loc); }
    public void emitStore(CodeBuilder cb, int loc) { cb.emitFSTORE(loc); }
    public void emitReturn(CodeBuilder cb) { cb.emitFRETURN(); }

    void emitCast(CodeBuilder cb, Type castTo) {
        ((NumericType) castTo).emitCastFromFloat(cb);
    }
    void emitCastFromInt(CodeBuilder cb)    { cb.emitI2F(); }
    void emitCastFromFloat(CodeBuilder cb)  {               }
    void emitCastFromLong(CodeBuilder cb)   { cb.emitL2F(); }
    void emitCastFromDouble(CodeBuilder cb) { cb.emitD2F(); }

    void emitZero(CodeBuilder cb) { cb.emitFloatConstant(0.0f); }
    void emitOne(CodeBuilder cb) { cb.emitFloatConstant(1.0f); }
    void emitMinusOne(CodeBuilder cb) { cb.emitFloatConstant(-1.0f); }

    void emitEqualityCompare(CodeBuilder cb, String op, Label t, Label f) {
        cb.emitFCMPL(); // -1 if either are NaN
        if      (op == "==") cb.emitIFEQ(t, f);
        else if (op == "!=") cb.emitIFNE(t, f);
    }

    void emitNumericCompare(CodeBuilder cb, String op, Label t, Label f) {
        if      (op == "<")  { cb.emitFCMPG(); cb.emitIFLT(t, f); }
        else if (op == "<=") { cb.emitFCMPG(); cb.emitIFLE(t, f); }
        else if (op == ">=") { cb.emitFCMPL(); cb.emitIFGE(t, f); }
        else if (op == ">")  { cb.emitFCMPL(); cb.emitIFGT(t, f); }
    }

    void emitAload(CodeBuilder cb) { cb.emitFALOAD(); }
    void emitAstore(CodeBuilder cb) { cb.emitFASTORE(); }

}
