/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
  * @grammar for (init; test; update) {body}
  * @child ASTObject init
  * @child Expr test
  * @child Exprs update
  * @child Stmt body
  */

public class ForStmt extends TestStmt {

    public boolean isBreakable() { return true; }
    public boolean isContinuable() { return true; }

    private void writeInit(CodeWriter writer) {
        if (init == null) return;

        if (init instanceof Decs) {
            Decs decs = (Decs)init;
            ((VarDec)decs.get(0)).unparse(writer, false, true);
            for(int i=1; i<decs.size(); i++) {
                writer.comma();
                writer.optionalSpace();
                ((VarDec)decs.get(i)).unparse(writer, false, false);
            }
        } else if (init instanceof VarDec) {
            ((VarDec)init).unparse(writer, false, true);
        } else {
            writer.write(init);
        }
    }

    public void checkSpec() {
        super.checkSpec();
        for (int i=0; i<update.size(); i++) {
            Expr expr = update.get(i);
            if (!expr.isLegalStmt()) {
                expr.showError("not a statement");
            }
        }
    }


    public void unparse(CodeWriter writer) {
        writer.writeKeyword("for");
        writer.requiredSpace();
        writer.openParen('(');

        writeInit(writer);
        writer.write("; ");
        if (! (test.isSynthetic() && test.isConstantTrue())) {
            writer.write(test);
        }
        writer.write("; ");
        writer.write(update);
        writer.closeParen(')');
        writer.optionalSpace();
        writer.write(body);
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        w.process(getInit());

        FlowCheckerPass.Set enteringPossiblyAssigned = w.popPossiblyAssigned();
        w.process(getTest());
        FlowCheckerPass.Vars p = w.getVars();
        FlowCheckerPass.Vars tv = p.getTrue();
        FlowCheckerPass.Vars fv = p.getFalse();

        w.setLive(! getTest().isConstantFalse());
        w.setVars(p.getTrue());
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();

        FlowCheckerPass.Vars v = w.getVars();
        FlowCheckerPass.Vars cv = w.getContinueVars(this);
        FlowCheckerPass.Vars bv = w.getBreakVars(this);

        w.setLive(true);  // should be isContinued || isLive
        w.setVars(v.join(cv));
        w.process(getUpdate());
        FlowCheckerPass.Vars uv = w.getVars();

        FlowCheckerPass.Set loopPossiblyAssigned = w.popPossiblyAssigned();

        w.checkLoopingFinals(this, loopPossiblyAssigned, uv);
        w.mergePossiblyAssigned(enteringPossiblyAssigned);
        w.mergePossiblyAssigned(loopPossiblyAssigned);

        w.setLive(! getTest().isConstantTrue() || w.isBroken(this));
        w.setVars(fv.join(bv));
    }

    // ------------------------------
    // INTRO: ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantFalse()) {
            w.process(getInit());
            return;
        }
        setInit(w.process(getInit()));
        w.setLive(true);
        w.enterContext(this);
        setBody((Stmt) w.process(getBody()));
        w.leaveContext();

        w.setLive(! getTest().isConstantTrue() || w.isBroken(this));
    }

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantFalse()) {
            return getInit();
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label testLab = cb.genLabel();
        Label startLab = cb.genLabel();
        Label updateLab = cb.genLabel();
        Label endLab = cb.genLabel();

        cb.enterBlock();
        if      (init instanceof Stmt)  ((Stmt) init).cgTop(cb);
        else if (init instanceof Exprs) ((Exprs) init).cgEffects(cb);
        cb.emitJump(testLab);
        cb.emitLabel(startLab);
        cb.enterNonWindingContext(this, endLab, updateLab);
        body.cgTop(cb);
        cb.leaveContext();
        cb.emitLabel(updateLab);
        update.cgEffects(cb);
        cb.emitLabel(testLab);
        test.cgTest(cb, startLab, endLab);
        cb.emitLabel(endLab);
        cb.exitBlock();
    }

    // ------------------------------
    // Intro:  FrameLocPass

    // we might allocate something, so we must clean up.

    public void walkFrameLoc(FrameLocPass walker) {
        int start = walker.getfs();
        super.walkFrameLoc(walker);
        walker.setfs(start);
    }

    //BEGIN: Generated from @child and @property
    protected ASTObject init;
    public ASTObject getInit() { return init; }
    public void setInit(ASTObject _init) {
        if (_init != null) _init.setParent(this);
        init = _init;
    }

    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    protected Exprs update;
    public Exprs getUpdate() { return update; }
    public void setUpdate(Exprs _update) {
        if (_update != null) _update.setParent(this);
        update = _update;
    }

    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    public ForStmt(SourceLocation location, ASTObject _init, Expr _test, Exprs _update, Stmt _body) {
        super(location);
        setInit(_init);
        setTest(_test);
        setUpdate(_update);
        setBody(_body);
    }
    protected ForStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        ForStmt ret = new ForStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (init != null) ret.setInit( (ASTObject)walker.process(init) );
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        if (update != null) ret.setUpdate( (Exprs)walker.process(update) );
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return init;
        case 1: return test;
        case 2: return update;
        case 3: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "init";
        case 1: return "test";
        case 2: return "update";
        case 3: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setInit((ASTObject)child); return;
        case 1: setTest((Expr)child); return;
        case 2: setUpdate((Exprs)child); return;
        case 3: setBody((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 4;
    }

    public String getDefaultDisplayName() {
        return "ForStmt()";
    }

    //END: Generated from @child and @property
}


