/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
// to get the passes for introductions
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.base.bcg.CodeBuilder;
import java.util.*;

import org.aspectj.compiler.base.bcg.ClassfileBuilder;
/**
  * @grammar [static] {body}
  */

public class InitializerDec extends CodeDec {

    public InitializerDec(SourceLocation location, Modifiers _modifiers, CodeBody _body) {
        this(location, _modifiers, new Formals(location), new TypeDs(location), _body);
    }

    public String getId() { return getModifiers().isStatic() ? "<clinit>" :  "<init>"; }

    public TypeD getResultTypeD() {
        return getTypeManager().voidType.makeTypeD();
    }

    public boolean conflictsWith(Dec otherDec) {
        return false;
    }

    public String toShortString() {
        return "{}";
    }

    public String getKind() { return "initializer"; }

    //INTRO from MixinImplementationPass
    public ASTObject postImplementMixin(MixinImplementationPass fixer) {
        final TypeDec inTypeDec = getBytecodeTypeDec();

        if (!(inTypeDec instanceof InterfaceDec) || getBody() == null || isStatic()) return this;

        Set topmostImplementors =
            Type.filterTopTypes(Type.filterConcreteTypes(inTypeDec.getType().getSubTypes()));

        for (Iterator i = topmostImplementors.iterator(); i.hasNext(); ) {
            final Type implType = (Type)i.next();
            TypeDec implDec = implType.getTypeDec();

            CodeDec newDec = (CodeDec)fixer.copyToClass(this, implDec);

            //XXX the position to add is important, this algorithm is a total hack
            Decs decs = implDec.getBody();
            int j = decs.size()-1;
            for (; j >= 0; j--) {
                if (decs.get(j) instanceof InitializerDec) {
                    implDec.getBody().add(j,  newDec);
                    break;
                }
            }
            if (j < 0) implDec.getBody().add( newDec );

        }

        return null;
    }


    public void unparse(CodeWriter writer) {
        if (body == null || body.getStmts() == null || body.getStmts().size() == 0) return;

        // odd unparser for static initializers in interfaces
        // this doesn't work because of the blank final rules in Java (not JVM)
//        if (isStatic() && getBytecodeTypeDec() instanceof InterfaceDec) {
//            writer.write("private static class AJC_INIT");
//            writer.openBlock();
//            writer.write("static");
//            writer.write(body);
//            writer.writeln("static int ajc_field = 0;");
//            writer.closeBlock();
//            writer.writeln("static int ajc_init = AJC_INIT.ajc_field;");
//        } else {
            writeModifiers(writer);
            writer.write(body);
            writer.newLine();
//        }
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        this.walk(w.createInitializerChecker(isStatic()));
    }

    // ------------------------------
    // bcg

    public ASTObject postCleanup(ByteCodeCleanupPass walker) {
        if (getBody().isEmpty()) {
            return null;
        } else {
            return super.postCleanup(walker);
        }
    }

    protected void cgCodeMember(CodeBuilder cb) {
        // no args, nothing on frame
        cb.setMaxFrame(getFrameSize());
        cb.enterBlock();
        getBody().cgTop(cb);
        cb.exitBlock();
    }

    public String getDescriptor() {
        return "()V";
    }

    public int getStackDelta() { return 0; }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walker.enterCodeDec(this);
        if (isStatic()) walker.inCode(true);
        else            walker.inConstructor();
        this.walk(walker);
        walker.leaveCodeDec();
    }

    public String toString() {
        return "InitializerDec(" + (isStatic() ? "static" : "non-static") + ")";
    }

    //BEGIN: Generated from @child and @property

    public InitializerDec(SourceLocation location, Modifiers _modifiers, Formals _formals, TypeDs __throws, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _body);

    }
    protected InitializerDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        InitializerDec ret = new InitializerDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }


    public String getDefaultDisplayName() {
        return "InitializerDec()";
    }

    //END: Generated from @child and @property
}
