/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar expr instanceof typeD
 * @child Expr expr
 * @child TypeD testTypeD
 */

public class InstanceofExpr extends Expr {

    public Type discoverType() {
        return getTypeManager().booleanType;
    }

    public void unparse(CodeWriter writer) throws IOException {
        writer.write(expr);
        writer.write(" instanceof ");
        writer.write(testTypeD);
    }

    public boolean isAlwaysFalse() {
        return !expr.getType().isCoercableTo(testTypeD.getType());
    }

    public void checkSpec() {
        if (!testTypeD.getType().isReferenceType()) {
            testTypeD.showError("not a reference type");
        }

        if (isAlwaysFalse()) {
            showError("Impossible for " + expr.getType().getString()
                      + " to be instanceof " + testTypeD.getType().getString());
        }
    }

    //INTRO from MovingWalker
    public ASTObject postMove(MovingWalker walker) {
        //XXX this might simplify too much
        //XXX this is needed because some instanceof tests can become illegal when moved
        //XXX (only illegal for Java source code, not bytecode)
        return simplify();
    }

    // special simplify method...
    public Expr simplify() {
        //System.out.println("simplifying " + this.unparse());
        if (expr.getType() == null) return this;

        if (testTypeD.getType().isAssignableFrom(expr.getType())) {
            return getAST().makeLiteral(true);
        } else if (isAlwaysFalse()) {
            return getAST().makeLiteral(false);
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        expr.cgValue(cb);
        testTypeD.getType().emitInstanceof(cb);
    }

    //BEGIN: Generated from @child and @property
    protected Expr expr;
    public Expr getExpr() { return expr; }
    public void setExpr(Expr _expr) {
        if (_expr != null) _expr.setParent(this);
        expr = _expr;
    }

    protected TypeD testTypeD;
    public TypeD getTestTypeD() { return testTypeD; }
    public void setTestTypeD(TypeD _testTypeD) {
        if (_testTypeD != null) _testTypeD.setParent(this);
        testTypeD = _testTypeD;
    }

    public InstanceofExpr(SourceLocation location, Expr _expr, TypeD _testTypeD) {
        super(location);
        setExpr(_expr);
        setTestTypeD(_testTypeD);
    }
    protected InstanceofExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        InstanceofExpr ret = new InstanceofExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (expr != null) ret.setExpr( (Expr)walker.process(expr) );
        if (testTypeD != null) ret.setTestTypeD( (TypeD)walker.process(testTypeD) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return expr;
        case 1: return testTypeD;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "expr";
        case 1: return "testTypeD";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setExpr((Expr)child); return;
        case 1: setTestTypeD((TypeD)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "InstanceofExpr()";
    }

    //END: Generated from @child and @property
}
