/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import java.util.*;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;
// to get the passes for introductions
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.base.bcg.CodeBuilder;

/**
  * @grammar modifiers resultType id(formals) throws body
  * @child TypeD resultTypeD @after modifiers
  * @property String id @after resultTypeD
  */
public class MethodDec extends CodeDec {
    //!!! temporary work-around
    public MethodDec(SourceLocation source,Modifiers modifiers, TypeD resultTypeD, String id, Formals formals,
                     TypeDs _throws, BlockStmt block)
    {
        this(source,modifiers, resultTypeD, id, formals, _throws, blockToBody(block));
    }
    public MethodDec(SourceLocation source,Modifiers modifiers, Type resultType, String id, Formals formals,
                     TypeDs _throws, BlockStmt block)
    {
        this(source,modifiers, resultType.makeTypeD(), id, formals, _throws, blockToBody(block));
    }

    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (this.isIntroduced()) this.getModifiers().setPublic(true);

        if (getBytecodeTypeDec() instanceof InterfaceDec) {
            this.getModifiers().setValue(Modifiers.ABSTRACT|Modifiers.PUBLIC);
            this.setBody(null);
        }
        return this;
    }

    //INTRO from MixinImplementationPass
    public ASTObject postImplementMixin(MixinImplementationPass fixer) {
        TypeDec inTypeDec = getBytecodeTypeDec();

        if (!(inTypeDec instanceof InterfaceDec)) return this;

        if (getBody() == null) return this;

        if (isStatic()) {
            ((InterfaceDec)inTypeDec).addToHelperClass(this);
            return null;
        }


        this.getModifiers().setFinal(false);
        this.getModifiers().setPublic(true);

        Set topmostImplementors =
            Type.filterTopTypes(Type.filterConcreteTypes(inTypeDec.getType().getSubTypes()));

        for (Iterator i = topmostImplementors.iterator(); i.hasNext(); ) {
            Type implType = (Type)i.next();
            TypeDec implDec = implType.getTypeDec();
            //System.out.println("adding: " + this.toShortString() + " to " + implDec);

            // don't implement if I'm already there in some other form
            if (!shouldImplement(implDec.getType(), this)) continue;

            MethodDec newDec = (MethodDec)fixer.copyToClass(this, implDec);
//            System.out.println("   really adding");
            implDec.getBody().add( newDec );
        }

        this.setBody(null);
        this.getModifiers().setSynchronized(false);
        return this;
    }


    boolean shouldImplement(Type onType, Dec dec) {
        Dec existingDec = onType.findMatchingDec(dec);
        if (existingDec != null && existingDec != dec) return false;
        
        Dec owner = dec.owner;
        if (owner != null) return shouldImplement(onType, owner);
        else return true;
    }

    //INTRO from ScopeWalker
    public void preScope(ScopeWalker walker) {
        walker.addMethodDec(this);
        super.preScope(walker);
    }

    public Expr getExpr() { return null; }

    public String toShortString() {
        return addIntroducedFromType(
            modifiers.toShortString() + " " + resultTypeD.toShortString() + " " +
                    getDeclaringType().toShortString()+"." + id +
                    formals.toShortString());
    }

    public String getKind() { return "method"; }

    public SemanticObject makeCorrespondingSemanticObject() {
        return new Method(this);
    }

    public Method getMethod() { return (Method)getCorrespondingSemanticObject(); }

    MethodDec makeSuperCaller() {
        final AST ast = getAST();

        Modifiers modifiers = ast.makeModifiers(0);
        if (isProtected()) modifiers.setProtected();
        if (isPublic()) modifiers.setPublic(true);

        Type resultType = getResultType();
        Formals formals = (Formals)getFormals().copy();
        CallExpr superCall = ast.makeSuperCall(getMethod(), formals.makeExprs());

        Stmt stmt;
        if (resultType.isVoid()) {
            stmt = ast.makeStmt(superCall);
        } else {
            stmt = ast.makeReturn(superCall);
        }

        MethodDec ret = ast.makeMethod(modifiers, resultType, getId(),
                formals, ast.makeBlock(stmt));
        if (getThrows() != null) {
            ret.setThrows((TypeDs)getThrows().copy());
        }
        return ret;
    }

    public void checkSpec() {
        super.checkSpec();
        if (getModifiers().isAbstract()) {
            if (getModifiers().isStrict()) {
                getModifiers().showError("illegal combination of modifiers: abstract strictfp");
            } else if (getModifiers().isPrivate()) {
                if (!isIntroduced()) {
                    showError("abstract private method can never be implemented");
                }
            } else if (getModifiers().isStatic()) {
                getModifiers().showError("illegal combination of modifiers: abstract static");
            }
        }
        if (isAbstract() || getModifiers().isNative()) {
            if (getBody() != null) {
                showError("abstract or native methods cannot have a body");
            }
        } else {
            if (getBody() == null) {
                showError("non-abstract/native methods must have a body");
            }
        }
    }

    void showOverrideError(Type inType, MethodDec other, String message) {
        if (getDeclaringType() != inType) {
            inType.getTypeDec().showError("inherited method " + this.toShortString() +
                " is incompatible with inherited method " + other.toShortString() +
                "; " + message);
        } else {
            String overrideKind = "override";
            if (other.getDeclaringType().isInterface()) {
                overrideKind = "implement";
            }
            showError(toShortString() + " cannot " + overrideKind + " " +
                        other.toShortString() + "; " + message);
        }
    }

    public boolean isEffectivelyAbstract(Type inType) {
        return isAbstract() || 
            (inType.isInterface() && getDeclaringType().isObject());
    }

    public boolean checkOverride(Type inType, Dec otherDec) {
        MethodDec other = (MethodDec)otherDec;

//         System.out.println("check override: " + this.toShortString() + ", " +
//                            other.toShortString() + " inherited? " + 
//                            (getDeclaringType() != inType));

        // both are inherited
        if (getDeclaringType() != inType) {
            if (isEffectivelyAbstract(inType)) {
                if (!other.isEffectivelyAbstract(inType)) {
                    other.internalCheckOverride(inType, this);
                    return true;
                } else {
                    if (getModifiers().isWeakerThan(other.getModifiers())) {
                        other.internalCheckOverride(inType, this);
                        return true;
                    } else {
                        this.internalCheckOverride(inType, other);
                        return false;
                    }
                }
            } else {
                if (other.isEffectivelyAbstract(inType)) {
                    this.internalCheckOverride(inType, other);
                    return false;
                } else if (this.overrides(other)) {
                    this.internalCheckOverride(inType, other);
                    return false;
                } else if (other.overrides(this)) {
                    other.internalCheckOverride(inType, this);
                    return true;
                } else {
                    showOverrideError(inType, other, "both methods are concrete");
                    return false;
                }
            }
        } else {
            this.internalCheckOverride(inType, other);
            return false;
        }
    }

    private void internalCheckOverride(Type inType, MethodDec other) {
        if (!getResultType().isEquivalent(other.getResultType())) {
            showOverrideError(inType, other,
                        "attempting to use incompatible return type\n" +
                  "found   : " + getResultType().getString() + "\n" +
                  "required: " + other.getResultType().getString());
        } else if (other.getModifiers().isFinal()) {
            showOverrideError(inType, other, "overridden method is final");
        } else if (getModifiers().isWeakerThan(other.getModifiers())) {
            showOverrideError(inType, other,
                              "attempting to assign weaker access privileges; was " +
                              other.getModifiers().toShortString());
        } else if (_throws != null) {
            // check that throws clause is not > super's
            for (int i=0; i<_throws.size(); i++) {
                if (!other.canThrow(_throws.get(i).getType())) {
                    showOverrideError(inType, other,
                        "overridden method does not throw " +
                        _throws.get(i).getType().getString());
                    break;
                }
            }
        }

        return;
    }


    public boolean overrides(MethodDec other) {
        for (Iterator i = getDeclaringType().getDirectSuperTypes().iterator(); i.hasNext(); ) {
            Type type = (Type)i.next();
            MethodDec dec = (MethodDec)type.findMatchingDec(this);
            if (dec == null) continue;
            if (dec == other) return true;
            if (dec == this) {
                getCompiler().internalError(this, "nonsense");
                return false;
            }
            if (dec.overrides(other)) return true;
        }
        return false;
    }

    public Set getMatchingMethods() {
        Set ret = new HashSet();

        for (Iterator i = getDeclaringType().getSubTypes().iterator(); i.hasNext(); ) {
            Type type = (Type)i.next();
            MethodDec dec = (MethodDec)type.findMatchingDec(this);
            if (dec != null) ret.add(dec);
        }

        return ret;
    }


    public MethodDec getRootMethodDec() {
        MethodDec ret = null;
        for (Iterator i = getDeclaringType().getDirectSuperTypes().iterator(); i.hasNext(); ) {
            Type type = (Type)i.next();
            MethodDec dec = (MethodDec)type.findMatchingDec(this);
            if (dec != null) {
                if (ret == null) {
                    ret = dec;
                } else {
                    return null;
                }
            }
        }
        if (ret == null) return this;
        else return ret.getRootMethodDec();
    }

    /** This refers backdoor method for privileged aspects and
     *  for access from inner classes (for bytecode compilation)
     */
    private MethodDec backdoorMethod;

    public MethodDec getBackdoorMethod() {
        if (backdoorMethod != null) return backdoorMethod;

        final AST ast = getAST();
        String id = getBytecodeId() + "$ajc$backdoor";

        Modifiers modifiers = ast.makeModifiers(Modifiers.FINAL|Modifiers.PUBLIC);
        if (isStatic()) modifiers.setStatic(true);

        Formals formals = (Formals)getFormals().copy();
        Expr thisExpr;
        if (isStatic()) thisExpr = ast.makeTypeExpr(getBytecodeTypeDec().getType());
        else thisExpr = ast.makeThis(getBytecodeTypeDec().getType());

        CallExpr innerCall = ast.makeCall(this, thisExpr, formals.makeExprs());
        Stmt stmt;
        if (resultTypeD.getType().isEquivalent(getTypeManager().voidType)) {
            stmt = ast.makeStmt(innerCall);
        } else {
            stmt = ast.makeReturn(innerCall);
        }

        MethodDec ret = ast.makeMethod(modifiers, getResultType(), id, formals,
                ast.makeBlock(stmt));
        ret.getBody().setParsed(true);
        if (getThrows() != null) {
            ret.setThrows((TypeDs)getThrows().copy());
        }

        getDeclaringType().getTypeDec().addToBody(ret);

        backdoorMethod = ret;
        return backdoorMethod;
    }


    public void unparse(CodeWriter writer) {
        if (getBytecodeTypeDec() instanceof InterfaceDec && (isPrivate() || isFinal())) {
            return;
        }

        writeModifiers(writer);
        writer.write(resultTypeD);
        writer.requiredSpace();
        writer.write(getBytecodeId());
        writer.write(formals);

        writeNames(writer, "throws", getThrows());

        if (!writer.isOnlySignatures()) {
            if (body == null) {
                writer.closeStmt();
            } else {
                writer.optionalSpace();
                writer.write(body);
            }
            writer.newLine();
        }
    }

    // ------------------------------
    // bcg

    protected void cgCodeMember(CodeBuilder cb) {
        cb.setMaxFrame(getFrameSize());
        cb.enterBlock();
        Formals f = getFormals();
        for (int i = 0, len = f.size(); i < len; i++) {
            cb.enterVar(f.get(i));
        }
        getBody().cgTop(cb);
        cb.exitBlock();
    }

    private String descriptor;
    public synchronized String getDescriptor() {
        if (descriptor == null) {
            descriptor = "(";

            for (Iterator i = getFormals().iterator(); i.hasNext(); ) {
                FormalDec dec = (FormalDec)i.next();
                descriptor += dec.getType().getDescriptor();
            }
            descriptor += ")";
            descriptor += getResultType().getDescriptor();
        }
        return descriptor;
    }

    private boolean isDeltaComputed = false;
    private int stackDelta;
    public synchronized int getStackDelta() {
        if (!isDeltaComputed) {
            stackDelta = getResultType().getSlotCount();
            if (!isStatic()) stackDelta--;
            for (Iterator i = getFormals().iterator(); i.hasNext(); ) {
                FormalDec dec = (FormalDec)i.next();
                stackDelta -= dec.getType().getSlotCount();
            }
            isDeltaComputed = true;
        }
        return stackDelta;
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        walker.pushNonConstructorEnv();
    }

    public ASTObject postThreading(LocalClassPass.ThreadingWalker walker) {
        walker.popEnv();
        return this;
    }

    //BEGIN: Generated from @child and @property
    protected TypeD resultTypeD;
    public TypeD getResultTypeD() { return resultTypeD; }
    public void setResultTypeD(TypeD _resultTypeD) {
        if (_resultTypeD != null) _resultTypeD.setParent(this);
        resultTypeD = _resultTypeD;
    }

    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }

    public MethodDec(SourceLocation location, Modifiers _modifiers, TypeD _resultTypeD, String _id, Formals _formals, TypeDs __throws, CodeBody _body) {
        super(location, _modifiers, _formals, __throws, _body);
        setResultTypeD(_resultTypeD);
        setId(_id);
    }
    protected MethodDec(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        MethodDec ret = new MethodDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (modifiers != null) ret.setModifiers( (Modifiers)walker.process(modifiers) );
        if (resultTypeD != null) ret.setResultTypeD( (TypeD)walker.process(resultTypeD) );
        ret.id = id;
        if (formals != null) ret.setFormals( (Formals)walker.process(formals) );
        if (_throws != null) ret.setThrows( (TypeDs)walker.process(_throws) );
        if (body != null) ret.setBody( (CodeBody)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return resultTypeD;
        case 2: return formals;
        case 3: return _throws;
        case 4: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "resultTypeD";
        case 2: return "formals";
        case 3: return "throws";
        case 4: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setResultTypeD((TypeD)child); return;
        case 2: setFormals((Formals)child); return;
        case 3: setThrows((TypeDs)child); return;
        case 4: setBody((CodeBody)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 5;
    }

    public String getDefaultDisplayName() {
        return "MethodDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
