/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

//import java.lang.reflect.Modifier;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;
import java.io.IOException;


/**
  * @grammar (public|private|...)*
  * @property int value
  *
  */
public class Modifiers extends ASTObject {
    /*
     * Access modifier flag constants from <em>The Java Virtual
     * Machine Specification</em>, Table 4.1.
     */

    /**
     * The <code>int</code> value representing the <code>public</code>
     * modifier.
     */
    public static final int PUBLIC           = 0x00000001;

    /**
     * The <code>int</code> value representing the <code>private</code>
     * modifier.
     */
    public static final int PRIVATE          = 0x00000002;

    /**
     * The <code>int</code> value representing the <code>protected</code>
     * modifier.
     */
    public static final int PROTECTED        = 0x00000004;

    /**
     * The <code>int</code> value representing the <code>static</code>
     * modifier.
     */
    public static final int STATIC           = 0x00000008;

    /**
     * The <code>int</code> value representing the <code>final</code>
     * modifier.
     */
    public static final int FINAL            = 0x00000010;

    /**
     * The <code>int</code> value representing the <code>synchronized</code>
     * modifier.
     */
    public static final int SYNCHRONIZED     = 0x00000020;

    /**
     * The <code>int</code> value representing the <code>volatile</code>
     * modifier.
     */
    public static final int VOLATILE         = 0x00000040;

    /**
     * The <code>int</code> value representing the <code>transient</code>
     * modifier.
     */
    public static final int TRANSIENT        = 0x00000080;

    /**
     * The <code>int</code> value representing the <code>native</code>
     * modifier.
     */
    public static final int NATIVE           = 0x00000100;

    /**
     * The <code>int</code> value representing the <code>interface</code>
     * modifier.
     */
    public static final int INTERFACE        = 0x00000200;

    /**
     * The <code>int</code> value representing the <code>abstract</code>
     * modifier.
     */
    public static final int ABSTRACT         = 0x00000400;

    /**
     * The <code>int</code> value representing the <code>strictfp</code>
     * modifier.
     */
    public static final int STRICT           = 0x00000800;

    //XXX I'm claiming the upper 16-bits of modifier space here for any
    //XXX aspect-specific modifiers.  This certainly isn't in line with
    //XXX notions of an extensible compiler
    public static final int PRIVILEGED = 0x00010000;

    /**
     *   toString masks for different kinds of output
     */

    /** Default print mask */
    public static final int DEFAULT_PRINT_MASK =
        PUBLIC | PRIVATE | PROTECTED | STATIC | FINAL | SYNCHRONIZED |
        VOLATILE | TRANSIENT | NATIVE | ABSTRACT | STRICT;

    /** Print mask for interface modifiers - avoid public+abstract+strictfp */
    public static final int INTERFACE_METHOD_PRINT_MASK =
        STATIC | FINAL | SYNCHRONIZED | VOLATILE | TRANSIENT | NATIVE;

    /** Print mask for constructor modifiers - avoid strictfp */
    public static final int CONSTRUCTOR_PRINT_MASK =
        PUBLIC | PRIVATE | PROTECTED | STATIC | FINAL | SYNCHRONIZED |
        VOLATILE | TRANSIENT | NATIVE | ABSTRACT;


        /*
    protected static int modifierValue(String[] labels) {
        int modifier = 0;
        for (int i=0; i<labels.length; i++) {
            String label = labels[i];
            if (label.equals("abstract"))          modifier |= Modifiers.ABSTRACT;
            else if (label.equals("final"))        modifier |= Modifiers.FINAL;
            else if (label.equals("interface"))    modifier |= Modifiers.INTERFACE;
            else if (label.equals("native"))       modifier |= Modifiers.NATIVE;
            else if (label.equals("private"))      modifier |= Modifiers.PRIVATE;
            else if (label.equals("protected"))    modifier |= Modifiers.PROTECTED;
            else if (label.equals("public"))       modifier |= Modifiers.PUBLIC;
            else if (label.equals("static"))       modifier |= Modifiers.STATIC;
            else if (label.equals("synchronized")) modifier |= Modifiers.SYNCHRONIZED;
            else if (label.equals("transient"))    modifier |= Modifiers.TRANSIENT;
            else if (label.equals("volatile"))     modifier |= Modifiers.VOLATILE;
            else if (label.equals("strictfp"))     modifier |= Modifiers.STRICT;
            else throw new RuntimeException("unknown modifier name: "+label);
        }
        return modifier;
    }

    public Modifiers(SourceLocation location, String[] labels) {
        this(location,modifierValue(labels));
    }
    */

    public static boolean isPublic(int value) {
        return (value & PUBLIC) != 0;
    }
    public boolean isPublic() {
        return (value & PUBLIC) != 0;
    }
    public static boolean isFinal(int value) {
        return (value & FINAL) != 0;
    }
    public boolean isFinal() {
        return (value & FINAL) != 0;
    }
    public static boolean isPrivate(int value) {
        return (value & PRIVATE) != 0;
    }
    public boolean isPrivate() {
        return (value & PRIVATE) != 0;
    }
    public static boolean isProtected(int value) {
        return (value & PROTECTED) != 0;
    }
    public boolean isProtected() {
        return (value & PROTECTED) != 0;
    }

    public static boolean isPackagePrivate(int value) {
        return (value & (PUBLIC | PROTECTED | PRIVATE)) == 0;
    }
    public boolean isPackagePrivate() {
        return (value & (PUBLIC | PROTECTED | PRIVATE)) == 0;
    }

    public static boolean isInterface(int value) {
        return (value & INTERFACE) != 0;
    }
    public boolean isInterface() {
        return (value & INTERFACE) != 0;
    }

    public static boolean isStatic(int value) {
        return (value & STATIC) != 0;
    }
    public boolean isStatic() {
        return (value & STATIC) != 0;
    }
    public static boolean isAbstract(int value) {
        return (value & ABSTRACT) != 0;
    }
    public boolean isAbstract() {
        return (value & ABSTRACT) != 0;
    }
    public static boolean isStrict(int value) {
        return (value & STRICT) != 0;
    }
    public boolean isStrict() {
        return (value & STRICT) != 0;
    }

    public static boolean isPrivileged(int value) {
        return (value & PRIVILEGED) != 0;
    }
    public boolean isPrivileged() {
        return (value & PRIVILEGED) != 0;
    }

    public boolean isSynchronized() {
        return (value & SYNCHRONIZED) != 0;
    }

    public boolean isNative() {
        return (value & NATIVE) != 0;
    }

    public boolean isTransient() {
        return (value & TRANSIENT) != 0;
    }

    public boolean isVolatile() {
        return (value & VOLATILE) != 0;
    }

    protected void setAttribute(int mask, boolean set) {
        if (set) value |= mask;
        else value &= ~mask;
    }

    public void setPublic(boolean v) {
        setAttribute(PRIVATE, false);
        setAttribute(PROTECTED, false);
        setAttribute(PUBLIC, v);
    }
    public void setProtected() {
        setAttribute(PRIVATE, false);
        setAttribute(PUBLIC, false);
        setAttribute(PROTECTED, true);
    }

    public void setPrivate(boolean v) {
        setAttribute(PROTECTED, false);
        setAttribute(PUBLIC, false);
        setAttribute(PRIVATE, v);
    }
    public void setInterface(boolean v) {
        setAttribute(INTERFACE, v);
    }
    public void setFinal(boolean v) {
        setAttribute(FINAL, v);
    }
    public void setAbstract(boolean v) {
        setAttribute(ABSTRACT, v);
    }
    public void setStatic(boolean v) {
        setAttribute(STATIC, v);
    }
    public void setStrict(boolean v) {
        setAttribute(STRICT, v);
    }
    public void setSynchronized(boolean v) {
        setAttribute(SYNCHRONIZED, v);
    }

    public boolean isWeakerThan(Modifiers other) {
        if (isPublic()) return false;
        if (isProtected()) return other.isPublic();
        if (isPackagePrivate()) return other.isPublic() || other.isProtected();
        if (isPrivate()) return !other.isPrivate();

        //XXX this is probably an error
        return false;
    }


    public boolean isMoreAccessibleThan(Modifiers other) {
        if (isPublic()) return !other.isPublic();
        if (isProtected()) return !(other.isPublic() || other.isProtected());
        if (isPackagePrivate()) return other.isPrivate();
        if (isPrivate()) return false;

        //XXX this is probably an error
        return false;
    }

    public int getFalseValue() { return 0; }

    /** Returns an integer that is acceptable to store as a
        classfile's access flags.  Only modifiers of TypeDecs will go
        through this. This strips out any value disallowed by the JVM
        spec, and turns protected modifiers (which may occur on inner
        classes) public. */
    public int getAcceptableClassValue() {
        int val = getValue() & (PUBLIC | FINAL | INTERFACE | ABSTRACT);
        if (isProtected()) val |= PUBLIC;
        if (isInterface()) {
            val &= PUBLIC;
            val |= ABSTRACT | INTERFACE;
        } else {
            val |= 0x0020; // ACC_SUPER
        }
        return val;
    }

    /** Returns an integer that is acceptable to store in the
        InnerClass attribute of a classfile.  Only modifiers of
        TypeDecs will go through this. This strips out any value
        disallowed by the JVM. */
    public int getAcceptableInnerClassValue() {
        int val = getValue() &
            (PUBLIC | PRIVATE | PROTECTED | STATIC | FINAL | INTERFACE | ABSTRACT);
        if (isInterface()) {
            val |= STATIC | ABSTRACT;
        }
//        else {
//        	val |= 0x0020; // ACC_SUPER
//        }
        return val;
    }

    /** Returns an integer that only encapsulates the access
        modifiers.  That is, only PUBLIC, PRIVATE, or PROTECTED. */
    public int getAccessValue() {
        return getValue() & (PUBLIC | PRIVATE | PROTECTED);
    }

    public String getAccessValueString() {
        if (isPublic()) return "public";
        else if (isPrivate()) return "private";
        else if (isProtected()) return "protected";
        else return "package";
    }

    public void checkIllegalCombination(int c1, int c2) {
        if ((getValue() & c1) != 0) {
            if ((getValue() & c2) != 0) {
                showError("illegal combination of modifiers: " +
                          toString(c1, false) + toString(c2, true));
            }
        }
    }

    public void checkSpec() {
        checkIllegalCombination(PUBLIC, PRIVATE);
        checkIllegalCombination(PUBLIC, PROTECTED);
        checkIllegalCombination(PROTECTED, PRIVATE);
        //checkIllegalCombination(ABSTRACT, PRIVATE); handle in MethodDec
        //checkIllegalCombination(ABSTRACT, STATIC); handle in MethodDec
        checkIllegalCombination(ABSTRACT, FINAL);
        checkIllegalCombination(ABSTRACT, NATIVE);
        //checkIllegalCombination(ABSTRACT, STRICT); handle in MethodDec
        checkIllegalCombination(ABSTRACT, SYNCHRONIZED);

        checkIllegalCombination(NATIVE, STRICT);
    }


    //!!! strange definition of matching here...
    public boolean matches(Modifiers otherModifiers) {
        // a match is when other modifiers define everything that this does
        // they're also allowed to define more
        return (value & otherModifiers.getValue()) == value;
    }

    public String toShortString() {
        return toString();
    }

    public void unparse(CodeWriter writer) throws IOException {
        //XXX JJH - this reliance on getParent() for unparsing makes me uncomfortable
        //XXX JJH - we should revisit this some time
        String modText;
        if (getParent() instanceof ConstructorDec) {
            modText = toString(CONSTRUCTOR_PRINT_MASK,false);
        }
        else if (getParent() instanceof MethodDec &&
                 getParent().getParent().getParent() instanceof TypeDec &&
                 !((TypeDec)(getParent().getParent().getParent())).isConcrete()
                 ) {
            modText = toString(INTERFACE_METHOD_PRINT_MASK,false);
        }
        else {
            modText = toString(DEFAULT_PRINT_MASK,false);
        }
        if (modText.length() > 0) writer.write(modText);
    }

    public String toString() {
        return toString(DEFAULT_PRINT_MASK,true);
    }

    private String toString(int mask, boolean trim) {
        int val = value & mask;
        if (val == 0) return "";
        StringBuffer sb = new StringBuffer();

        if ((val & PUBLIC) != 0)         sb.append("public ");
        else if ((val & PRIVATE) != 0)   sb.append("private ");
        else if ((val & PROTECTED) != 0) sb.append("protected ");

        /* Canonical order */
        if ((val & STATIC) != 0)         sb.append("static ");
        if ((val & ABSTRACT) != 0)       sb.append("abstract ");
        if ((val & FINAL) != 0)          sb.append("final ");
        if ((val & TRANSIENT) != 0)      sb.append("transient ");
        if ((val & VOLATILE) != 0)       sb.append("volatile ");
        if ((val & NATIVE) != 0)         sb.append("native ");
        if ((val & SYNCHRONIZED) != 0)   sb.append("synchronized ");

        if ((val & STRICT) != 0 && (val & (ABSTRACT|NATIVE)) == 0) {
            sb.append("strictfp ");
        }

        String s = sb.toString();
        if (trim) return s.substring(0, s.length()-1);
        return s;
    }

    //BEGIN: Generated from @child and @property
    protected int value;
    public int getValue() { return value; }
    public void setValue(int _value) { value = _value; }

    public Modifiers(SourceLocation location, int _value) {
        super(location);
        setValue(_value);
    }
    protected Modifiers(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        Modifiers ret = new Modifiers(getSourceLocation());
        ret.preCopy(walker, this);
        ret.value = value;
        return ret;
    }


    public String getDefaultDisplayName() {
        return "Modifiers(value: "+value+")";
    }

    //END: Generated from @child and @property
}
