/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.CodeBuilder;

/**
 * @grammar enclosingInstanceExpr.new typeD(args) { typeDec }
 * @child Expr enclosingInstanceExpr @before typeD
 * @child Exprs args
 * @child TypeDec typeDec
 * @property Constructor constructor
 */

public class NewInstanceExpr extends NewExpr implements SOLink {

    public CodeDec getCodeDec() { return getConstructorDec(); }
    public ConstructorDec getConstructorDec() {
        return getConstructor().getConstructorDec();
    }
    public void setConstructorDec(ConstructorDec md) {
        setConstructor((Constructor)md.getCorrespondingSemanticObject());
    }
    public SemanticObject getTarget() { return getConstructor(); }
    public void setTarget(SemanticObject newTarget) {
        setConstructor((Constructor)newTarget);
    }

    public boolean isLegalStmt() {
        return true;
    }

    // ------------------------------
    // INTRO from InnerInfoPass

    public void walkInnerInfo(InnerInfoPass w) {
//         if (typeDec != null && ! (enclosingInstanceExpr instanceof ThisExpr)) {
//             final AST ast = getAST();
//             // we're a qualified anonymous type, so let's do the rewrite now.
//             ConstructorDec dec = getConstructorDec();
//             ConstructorCallExpr cc =
//                 ((ConstructorBody) dec.getBody()).getConstructorCall();
//             FormalDec f = ast.makeFormal(enclosingInstanceExpr.getType(),
//                                          "super$enc$obj");
//             dec.getFormals().add(0, f);
//             dec.setEnclosingInstanceFormal(null);
//             cc.setEnclosingInstanceExpr(ast.makeVar(f));
//             getArgs().add(0, enclosingInstanceExpr);
//             setEnclosingInstanceExpr(null);
//         }
        super.walkInnerInfo(w);

    }

    public void postInnerInfo(InnerInfoPass w) {
        // the only time getConstructor() can be null is if we've
        // already signalled an error for this expression that we
        // couldn't find the constructor.  In such cases, we'll just
        // punt since we're about to exit anyway.
        if (getConstructor() == null) return;
        NameType target = (NameType) getConstructor().getDeclaringType();
        if (w.isInner(target) && getEnclosingInstanceExpr() == null) {
            NameType enclosingTarget = target.getEnclosingType();
            if (! w.isAccessible(enclosingTarget)) {
                showError("no enclosing instance available");
            } else {
                setEnclosingInstanceExpr(getAST().makePrimary(enclosingTarget,
                                                              w.currentType(),
                                                              false));
            }
        }
    }

    // ------------------------------
    // INTRO from spec checker pass

    public void checkSpec() {
        if (getEnclosingInstanceExpr() != null) {
            NameType ty;
            if (getTypeDec() == null ||
                (getEnclosingInstanceExpr() instanceof ThisExpr)) {
                ty = (NameType) getConstructorDec().getDeclaringType();
            } else {
                ty = (NameType) getTypeD().getType();
                //!!! Erik can remove this if he gets the rest to work
                showError("anonymous qualified inner creation expressions not supported (compiler limitation");
            }
            if (! ty.isInner()) {
                showError("Class " + ty.getId() + " is not an inner class");
            }
            // if the primary is not of the correct type then we wouldn't
            // have resolved the constructor.
        }
        super.checkSpec();
    }

    //INTRO from AccessFixer
    public ASTObject fixAccessPost(AccessFixer fixer) {
        if (getConstructor() == null) return this; //XXX should be error
        //TypeDec fromTypeDec = getBytecodeTypeDec(); //fixer.fromTypeDec;
        if (getTypeDec() != null || getConstructor().isAccessible(this, true)) return this;

        getCompiler().showMessage("  fixing privileged new: " +
                                  getConstructor().toShortString());
        //this.getParent().getParent().display(2);

        getConstructor().getConstructorDec().makePublicAccessible();
        return this;
    }


    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        super.walkFlow(w); // walks the args
        if (getConstructor().getThrows() != null) {
            TypeDs ts = getConstructor().getThrows();
            for (int i = 0, len = ts.size(); i < len; i++) {
                w.setExns(w.getExns().add((NameType) ts.get(i).getType()));
            }
        }
    }

    /**
     * Section 6.6.2.2 from JLSv2
     * access to protected is only allowed from outside package
     * if typeDec != null
     */
    public boolean hasLegalProtectedAccess(Type fromType) {
        return typeDec != null;
    }

    //INTRO from MovingWalker
    public ASTObject postMove(MovingWalker walker) {
        //System.out.println("moving: " + this.unparse());
        walker.moveLink(this);

        /* !!!This nonsense is to make local type in around advice to "move"
           nicely. */
        if (enclosingInstanceExpr != null && walker.getToType() != null &&
            (getTypeDec() != null || getType().getTypeDec().isLocal())) {
            if (!walker.isStatic) {
                enclosingInstanceExpr = getAST().makeThis(walker.getToType());
            } else {
                enclosingInstanceExpr = null;
            }
        }

        if (getTypeDec() != null) {
            setType(null);
            getType();
        }

        return this;
    }

    //INTRO from ScopeWalker
    public ASTObject postScope(ScopeWalker walker) {
        if (constructor != null) return this;

        Type myType;
        if (typeDec != null) {
            myType = typeDec.getType();
        } else {
            myType = typeD.getType();
        }

        if (typeDec == null) {
            if (typeD.getType().isInterface()) {
                showError("can't instantiate interface " + myType.toShortString()); }
            else {
                constructor = typeD.getType().getConstructor(this, args, true);
            }
        } else {
            Constructor superConstructor;
            if (typeD.getType().isInterface()) {
                superConstructor =
                    getTypeManager().getObjectType().getConstructor(this, args, true);
            } else {
                superConstructor =
                    typeD.getType().getConstructor(typeDec.getBody(), args, true);
            }
            ConstructorDec dec = buildAnonDec(superConstructor);
            //dec.setLanguageVisible();
            dec.setAllEnclosingTypes(typeDec.getType());
            typeDec.addConstructorDec(dec);
            typeDec.getBody().add(dec);
            constructor = (Constructor) dec.getCorrespondingSemanticObject();
            if (getEnclosingInstanceExpr() != null) {
                // we've already made the constructor to take the enclosing
                // instance expression as an argument, so let's make
                // it an argument
                getArgs().add(0, enclosingInstanceExpr);
                setEnclosingInstanceExpr(null);
            }
        }

        if (fromSource() && myType.isAspect()) {
            showError("can't explicitly instantiate an aspect");
        }

        if (myType.getTypeDec().isAbstract()) {
            showError("can't instantiate abstract type " + myType.toShortString());
        }
        type = discoverType();

        return this;
    }

    private ConstructorDec buildAnonDec(Constructor superConstructor) {
        final AST ast = superConstructor.getAST();
        Formals oldFormals = superConstructor.getFormals();
        TypeDs oldThrows = superConstructor.getThrows();

        Formals newFormals = ast.makeFormals();
        Exprs newArgs = ast.makeExprs();
          TypeDs newThrows =
              oldThrows == null ? ast.makeTypeDs() : (TypeDs) oldThrows.copy();

        final int len = oldFormals.size();
        for (int i = 0; i < len; i++) {
            FormalDec newFormal = (FormalDec) oldFormals.get(i).copy();
            newFormal.getModifiers().setFinal(true);
            Expr newArg = ast.makeVar(newFormal);
            newFormals.add(newFormal);
            newArgs.add(newArg);
        }

        Expr enclosingExpr = null;
        ConstructorCallExpr cce = null;
        if (getEnclosingInstanceExpr() != null) {
            FormalDec newFormal =
                // ??? Is it wise to use getType() in this pass (scope
                // walker)?  Shouldnn't we use 
                // ((NameType) superConstructor.getDeclaringType())
                // .getEnclosingType()
                // instead?
                ast.makeFinalFormal(getEnclosingInstanceExpr().getType(),
                                    "super$enc$obj");
            newFormals.add(0, newFormal);
            enclosingExpr = ast.makeVar(newFormal);
            cce = new ConstructorCallExpr(ast.getSourceLocation(), enclosingExpr,
                                          true, newArgs, superConstructor);
        } else {
            cce = ast.makeSuperConstructorCall(newArgs, superConstructor);
        }


        return
            ast.makeConstructor(ast.makeModifiers(0),
                                newFormals,
                                newThrows,
                                cce,
                                ast.makeStmts());
    }


    public Type discoverType() {
        if (typeDec == null) {
            return typeD.getType();
        } else {
            return typeDec.getType();
        }
    }

    public Type getCalledType() {
        return getType();
    }

    public void fixForIntroductions() {
//        System.out.println("fixing: " + constructor.toShortString() + System.identityHashCode(this));
        Expr extraArg = null;
        if (constructor != null) extraArg = constructor.getExtraArgExpr();
//        System.out.println("    :" + extraArg);
//        this.showError("fixing");
        if (extraArg != null) args.add(extraArg);
    }


    //INTRO from ASTFixerPass
    public ASTObject postFixAST(ASTFixerPass fixer) {
        fixForIntroductions();
        return this; //super.postFixAST(fixer);
    }

    public void unparse(CodeWriter writer) {
        if (getEnclosingInstanceExpr() == null) {
            writer.writeKeyword("new");
            writer.requiredSpace();
            writer.write(typeD);
            writer.parenExprs(args);
        } else {
            writer.write(enclosingInstanceExpr);
            writer.write('.');
            writer.writeKeyword("new");
            writer.requiredSpace();
            writer.write(getTypeD().getType().getId());
            writer.openParen('(');
            writer.write(args);
            writer.closeParen(')');
        }
        if (typeDec != null) {
            writer.openBlock();
            writer.write(typeDec.getBody());
            writer.closeBlock();
        }
    }

    // ------------------------------
    //INTRO from InnerAccessFixer

    public ASTObject postInnerAccess(InnerAccessFixer w) {
        ConstructorDec dec = getConstructorDec();
        if (! w.isAccessible(dec, this)) {
            // perhaps should be creating a cookie, but won't for now.
            dec.getModifiers().setPrivate(false);
        }
        return this;
    }

    // ------------------------------
    // bcg

    protected void cgValue(CodeBuilder cb) {
        cgValueEffect(cb, true);
    }
    protected void cgEffect(CodeBuilder cb) {
        cgValueEffect(cb, false);
    }
    private void cgValueEffect(CodeBuilder cb, boolean needsValue) {
        cb.enterLocation(getSourceLocation());
        NameType declaringType = (NameType) constructor.getDeclaringType();
        cb.emitNEW(declaringType);
        if (needsValue) cb.emitDUP();
        if (enclosingInstanceExpr != null) {
            enclosingInstanceExpr
                .cgValue(cb, constructor.getEnclosingInstanceFormal().getType());
        }
        getArgs().cgValues(cb, constructor.getFormals());
        cb.emitINVOKESPECIAL(declaringType, "<init>",
                             constructor.getDescriptor(),
                             constructor.getStackDelta());
    }

    // ------------------------------
    // INTRO:  LocalClassPass.LiftWalker

    public void preLift(LocalClassPass.LiftWalker walker) {
        TypeDec anonTypeDec = getTypeDec();
        if (anonTypeDec == null) return;
        NameType nameType = anonTypeDec.getNameType();
        // XXX isn't this done in scopeWalker now?
        this.setTypeD(nameType.makeTypeD());
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        walker.addArgs(getArgs(),
                       getConstructor().getDeclaringType().getTypeDec());
    }

    // ------------------------------
    //XXX HACK

    public void remove(int index) {
        if (index == 3) {
            setTypeDec(null);
        } else {
            super.remove(index);
        }
    }

    // ------------------------------
    // convienience constructors

    public NewInstanceExpr(SourceLocation location, TypeD _typeD,
                           Exprs _args, TypeDec _typeDec, Constructor _constructor) {
        this(location, null, _typeD, _args, _typeDec, _constructor);
    }

    public NewInstanceExpr(SourceLocation source, Constructor constructor, Exprs args) {
        this(source, null, constructor.getDeclaringType().makeTypeD(), args, null,
             constructor);
    }

    public NewInstanceExpr(SourceLocation source, TypeD typeD, Exprs args) {
        this(source, null, typeD, args, null, null);
    }

    //BEGIN: Generated from @child and @property
    protected Expr enclosingInstanceExpr;
    public Expr getEnclosingInstanceExpr() { return enclosingInstanceExpr; }
    public void setEnclosingInstanceExpr(Expr _enclosingInstanceExpr) {
        if (_enclosingInstanceExpr != null) _enclosingInstanceExpr.setParent(this);
        enclosingInstanceExpr = _enclosingInstanceExpr;
    }

    protected Exprs args;
    public Exprs getArgs() { return args; }
    public void setArgs(Exprs _args) {
        if (_args != null) _args.setParent(this);
        args = _args;
    }

    protected TypeDec typeDec;
    public TypeDec getTypeDec() { return typeDec; }
    public void setTypeDec(TypeDec _typeDec) {
        if (_typeDec != null) _typeDec.setParent(this);
        typeDec = _typeDec;
    }

    protected Constructor constructor;
    public Constructor getConstructor() { return constructor; }
    public void setConstructor(Constructor _constructor) { constructor = _constructor; }

    public NewInstanceExpr(SourceLocation location, Expr _enclosingInstanceExpr, TypeD _typeD, Exprs _args, TypeDec _typeDec, Constructor _constructor) {
        super(location, _typeD);
        setEnclosingInstanceExpr(_enclosingInstanceExpr);
        setArgs(_args);
        setTypeDec(_typeDec);
        setConstructor(_constructor);
    }
    protected NewInstanceExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        NewInstanceExpr ret = new NewInstanceExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (enclosingInstanceExpr != null) ret.setEnclosingInstanceExpr( (Expr)walker.process(enclosingInstanceExpr) );
        if (typeD != null) ret.setTypeD( (TypeD)walker.process(typeD) );
        if (args != null) ret.setArgs( (Exprs)walker.process(args) );
        if (typeDec != null) ret.setTypeDec( (TypeDec)walker.process(typeDec) );
        ret.constructor = constructor;
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return enclosingInstanceExpr;
        case 1: return typeD;
        case 2: return args;
        case 3: return typeDec;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "enclosingInstanceExpr";
        case 1: return "typeD";
        case 2: return "args";
        case 3: return "typeDec";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setEnclosingInstanceExpr((Expr)child); return;
        case 1: setTypeD((TypeD)child); return;
        case 2: setArgs((Exprs)child); return;
        case 3: setTypeDec((TypeDec)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 4;
    }

    public String getDefaultDisplayName() {
        return "NewInstanceExpr(constructor: "+constructor+")";
    }

    //END: Generated from @child and @property
}
