/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;


import java.util.*;

import org.aspectj.compiler.base.*;

public class SemanticGroup extends CompilerObject {
    protected String name;
    protected SemanticMap map;
    protected List members;

    public SemanticGroup(SemanticMap map, SemanticObject o) {
        this(map, o.getName());
        members.add(o);
    }

    public SemanticGroup(SemanticMap map, String name) {
        this(map, name, new LinkedList());
    }

    private SemanticGroup(SemanticMap map, String name, List members) {
        super(map.getCompiler());
        this.map = map;
        this.name = name;
        this.members = members;
    }

    public Collection getMembers() {
        return members;
    }

    public String getName() {
        return name;
    }

    public SemanticGroup makeInheritedCopy(SemanticMap newMap) {
        SemanticGroup ret = new SemanticGroup(newMap, name);
        ret.inheritFrom(this);
        return ret;
    }

    public void inheritFrom(SemanticGroup otherGroup) {
        if (otherGroup == null || otherGroup == this) {
//            getCompiler().showError(
//                "INTERNAL: " + this + " inheriting from " + otherGroup);
            return;
        }

        boolean skipObject = !map.getType().isInterface() && 
            otherGroup.map.getType().isInterface();

        for(Iterator i = otherGroup.members.iterator(); i.hasNext(); ) {
            SemanticObject so = (SemanticObject)i.next();
            if (skipObject && so.getDeclaringType().isObject()) continue;
            if (so.isInherited(map.getType())) {
                addSO(so, inheritMode);
            }
        }
    }

    private static abstract class Mode {
        abstract SemanticObject handleConflict(Type inType, SemanticObject currentSO, SemanticObject newSO);
    }

    private static final Mode inheritMode = new Mode() {
        SemanticObject handleConflict(Type inType, SemanticObject currentSO, SemanticObject newSO) {
            if (currentSO.checkOverride(inType, newSO)) {
                return currentSO;
            } else {
                return null;
            }
        }
    };

    private static final Mode addMode = new Mode() {
        SemanticObject handleConflict(Type inType, SemanticObject currentSO, SemanticObject newSO) {
            newSO.showConflictError(currentSO, "defined in same type");
            return null;
        }
    };

    private static final Mode introduceMode = new Mode() {
        SemanticObject handleConflict(Type inType, SemanticObject currentSO, SemanticObject newSO) {
            if (currentSO.isAbstract() && !newSO.isAbstract()) {
                //currentSO.replaceContentsWith(newSO);
                return currentSO;
            }
            if (newSO.isAbstract()) { return newSO; }

            if (newSO.dominates(currentSO)) { return currentSO; }
            if (currentSO.dominates(newSO)) { return newSO; }

            newSO.showConflictError(currentSO, "introduced member");
            return newSO;
        }
    };

    private SemanticObject addSO(SemanticObject newSO, Mode mode) {
        for(Iterator i = members.iterator(); i.hasNext(); ) {
            SemanticObject currentSO = (SemanticObject)i.next();

            //XXX there is a nasty inefficiency that this hides
            if (currentSO == newSO) return newSO;

            // this means that these both effectively name the same member
            if (currentSO.conflictsWith(newSO)) {
                SemanticObject removeSO = mode.handleConflict(map.getType(), currentSO, newSO);
                if (removeSO != null) {
                    if (removeSO == currentSO) {
                        i.remove();
                        members.add(newSO);
                        return currentSO;
                    } else if (removeSO == newSO) {
                        return newSO;
                    }
                }
                return null;
            }
        }
        members.add(newSO);
        return null;
    }

    public void addDeclaredSO(SemanticObject newSO) {
        addSO(newSO, addMode);
    }

    public SemanticObject addIntroducedSO(SemanticObject newSO) {
        return addSO(newSO, introduceMode);
    }

    private boolean containsMissingTypes(Exprs exprs) {
        if (exprs == null) return false;

        for (int i=0; i < exprs.size(); i++) {
            if (exprs.get(i).getType().isMissing()) {
                return true;
            }
        }

        return false;
    }


    public SemanticObject resolve(ASTObject fromWhere, Exprs params, boolean showError) {
        SemanticObject foundSemanticObject = null;
        List possibleSemanticObjects = null;

        for(Iterator i = members.iterator(); i.hasNext(); ) {
            SemanticObject object = (SemanticObject)i.next();

            if (object.isApplicable(params) && object.isAccessible(fromWhere)) {
                if (foundSemanticObject == null) {
                    foundSemanticObject = object;
                } else if (object.isMoreSpecificThan(foundSemanticObject)) {
//                    System.out.println(prettyIdString(name, params) +
//                        " replacing " + foundSemanticObject.toShortString() +
//                        " with more specific " + object.toShortString());
                    foundSemanticObject = object;
                } else if (foundSemanticObject.isMoreSpecificThan(object)) {
//                    System.out.println(prettyIdString(name, params) +
//                        " not replacing " + foundSemanticObject.toShortString() +
//                        " with less specific " + object.toShortString());
                    continue;
                } else if (containsMissingTypes(params)) {
                    return map.getNotFoundSemanticObject();
                } else {
                    // no clear most specific object
                    if (possibleSemanticObjects == null) possibleSemanticObjects = new ArrayList();
                    possibleSemanticObjects.add(foundSemanticObject);
                    possibleSemanticObjects.add(object);
                }
            }
        }

        if (foundSemanticObject == null) {
            //System.out.println("not found in group: " + name);
            if (showError && !containsMissingTypes(params)) showNotFoundError(fromWhere, params);
            return map.getNotFoundSemanticObject();
        }

        if (possibleSemanticObjects != null) {
            for (Iterator i = possibleSemanticObjects.iterator(); i.hasNext(); ) {
                SemanticObject possibleSemanticObject = (SemanticObject)i.next();
                if (!foundSemanticObject.isMoreSpecificThan(possibleSemanticObject)) {
                    if (showError && !containsMissingTypes(params)) getCompiler().showError(
                            fromWhere,
                            "two applicable and accessible " + map.getSemanticObjectTypeName() +
                            "s are equally specific: " + possibleSemanticObject.toShortString() + " and " +
                            foundSemanticObject.toShortString() + " in "
                            + map.getType().getPrettyString()
                    );
                    return map.getNotFoundSemanticObject();
                }
            }
        }
        return foundSemanticObject;
    }

    private void showNotFoundError(ASTObject fromWhere, Exprs params) {
        String message;
        String kind = map.getSemanticObjectTypeName();
        if (kind.equals("constructor")) {
            message = prettyIdString(map.getType().getPrettyString(), params);
        } else {
            message = prettyIdString(name, params);
        }

        message = getExplanationString(message, fromWhere, params);
        getCompiler().showError(fromWhere, message);
    }

    private String prettyIdString(String id, Exprs args) {
        if (args == null) return id;

        StringBuffer message = new StringBuffer();
        if (id.equals("new")) {
            id = map.getType().getPrettyString();
        }
        message.append(id);
        if (args != null) {
            message.append("(");
            for(int i=0; i<args.size(); i++) {
                if (i > 0) message.append(", ");
                message.append(args.get(i).getType().getString());
            }
            message.append(")");
        }
        return message.toString();
    }


    private String getExplanationString(String message, ASTObject fromWhere, Exprs parameters) {
        StringBuffer buf = new StringBuffer(message);
        String shortString = null;
        int notAccessibleMethods = 0;
        int notApplicableMethods = 0;

        buf.append(" not found in " + map.getType().getPrettyString());

        if (fromWhere == null) return buf.toString();

        for(Iterator i = members.iterator(); i.hasNext(); ) {
            SemanticObject object = (SemanticObject)i.next();

            if (object.isApplicable(parameters)) {
                if (!object.isAccessible(fromWhere)) {
                    if (notAccessibleMethods == 0) {
                        shortString =
                            object.toShortString() + " has " +
                            object.getModifiers().getAccessValueString() + " access";
                    }
                    buf.append("\n  inaccessible " + map.getSemanticObjectTypeName() +
                               " found: " + object.toShortString()
                               +" from "+fromWhere.getLexicalType().toShortString());
                    notAccessibleMethods += 1;
                }
            } else if (object.isAlmostApplicable(parameters)) {
                if (notAccessibleMethods == 0 && notApplicableMethods == 0) {
                    shortString =
                        message + " doesn't match expected formals for " + object.toShortString();
                }

                buf.append("\n  not applicable " + map.getSemanticObjectTypeName() +
                            " found: " + object.toShortString());
                notApplicableMethods += 1;
            }
        }
        if (shortString != null && (notAccessibleMethods == 1 || notApplicableMethods <= 1)) {
            return shortString;
        }

        return buf.toString();
    }
}
