/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.AssignmentCheckerPass;
import org.aspectj.compiler.base.CodeWriter;
import org.aspectj.compiler.base.FlowCheckerPass;
import org.aspectj.compiler.base.ByteCodeCleanupPass;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar test ? true: false
 * @child Expr test
 * @child Expr true
 * @child Expr false
 */
public class TriTestExpr extends Expr {

    // XXX this is wrong, most of this checking should be done after constant folding.
    public Type discoverType() {
        Type testTy = getTest().getType();
        if (testTy != getTypeManager().booleanType) {
            showError("Incompatable type for ?: Can't convert " + testTy.getString() + " to boolean");
        }
        Expr trueExpr = getTrue();
        Expr falseExpr = getFalse();
        Type trueTy = trueExpr.getType();
        Type falseTy = falseExpr.getType();

        if (trueTy == falseTy) {
            return trueTy;
        } else if (trueTy.isNumeric() && falseTy.isNumeric()) {
            if ((trueTy instanceof ShortType && falseTy instanceof ByteType) ||
                (trueTy instanceof ByteType && falseTy instanceof ShortType)) {
                return getTypeManager().shortType;
            } else if (trueExpr instanceof IntLiteralExpr &&
                       trueTy instanceof IntType &&
                       trueExpr.isAssignableTo(falseTy)) {
                return falseTy;
            } else if (falseExpr instanceof IntLiteralExpr &&
                       falseTy instanceof IntType &&
                       falseExpr.isAssignableTo(trueTy)) {
                return trueTy;
            } else {
                return getTypeManager().binaryNumericPromotion(trueTy, falseTy);
            }
        } else if (trueTy.isReferenceType() && falseTy.isReferenceType()) {
            if (trueTy.isAssignableFrom(falseTy)) {
                return trueTy;
            } else if (falseTy.isAssignableFrom(trueTy)) {
                return falseTy;
            } else {
                showError("Incompatable type for ?: Can't convert between " + trueTy.getString() + " and " + falseTy.getString());
                return trueTy;
            }
        } else {
            showError("Incompatable type for ?: Can't convert between " + trueTy.getString() + " and " + falseTy.getString());
            return trueTy;
        }
    }

    public void preAssignmentCheck(AssignmentCheckerPass walker) {
        getType();
    }

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        setType(null);
        if ((getTrue() instanceof LiteralExpr) &&
            (getFalse() instanceof LiteralExpr)) {
            if (getTest().isConstantTrue()) {
                return getTrue();
            } else if (getTest().isConstantFalse()) {
                return getFalse();
            }
        }
        return this;
    }

    public void unparse(CodeWriter writer) {
        writer.write(test);
        writer.writeOp("?");
        writer.write(_true);
        writer.writeOp(":");
        writer.write(_false);
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        // w.setArgs(w.getArgs());
        w.process(getTest());
        FlowCheckerPass.Vars p0 = w.getVars();

        if (getType().isBoolean()) {
            w.setVars(p0.getTrue());
            w.process(getTrue());
            FlowCheckerPass.Vars p1 = w.getVars();

            w.setVars(p0.getFalse());
            w.process(getFalse());
            FlowCheckerPass.Vars p2 = w.getVars();

            w.setVars(p1.getTrue().join(p2.getTrue()),
                      p1.getFalse().join(p2.getFalse()));
        } else {
            w.setVars(p0.getTrue());
            w.process(getTrue());
            FlowCheckerPass.Vars t1 = w.getVars();

            w.setVars(p0.getFalse());
            w.process(getFalse());
            FlowCheckerPass.Vars t2 = w.getVars();

            w.setVars(t1.join(t2));
        }
    }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantTrue()) {
            return getTrue();
        } else if (getTest().isConstantFalse()) {
            return getFalse();
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg
    protected void cgValue(CodeBuilder cb) {
        Type ty = getType();
        Label trueLab = cb.genLabel();
        Label falseLab = cb.genLabel();
        Label endLab = cb.genLabel();
        test.cgTest(cb, trueLab, falseLab);
        cb.emitLabel(trueLab);
        _true.cgValue(cb, ty);
        cb.emitJump(endLab);
        cb.emitLabel(falseLab);
        cb.popStack(ty.getSlotCount());
        _false.cgValue(cb, ty);
        cb.emitLabel(endLab);
    }
    protected void cgEffect(CodeBuilder cb) {
        Label trueLab = cb.genLabel();
        Label falseLab = cb.genLabel();
        Label endLab = cb.genLabel();
        test.cgTest(cb, trueLab, falseLab);
        cb.emitLabel(trueLab);
        _true.cgEffect(cb);
        cb.emitJump(endLab);
        cb.emitLabel(falseLab);
        _false.cgEffect(cb);
        cb.emitLabel(endLab);
    }
    protected void cgTest(CodeBuilder cb, Label tdest, Label fdest) {
        Label trueLab = cb.genLabel();
        Label falseLab = cb.genLabel();
        test.cgTest(cb, trueLab, falseLab);
        cb.emitLabel(trueLab);
        _true.cgTest(cb, tdest, fdest);
        cb.emitLabel(falseLab);
        _false.cgTest(cb, tdest, fdest);
    }

    //BEGIN: Generated from @child and @property
    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    protected Expr _true;
    public Expr getTrue() { return _true; }
    public void setTrue(Expr __true) {
        if (__true != null) __true.setParent(this);
        _true = __true;
    }

    protected Expr _false;
    public Expr getFalse() { return _false; }
    public void setFalse(Expr __false) {
        if (__false != null) __false.setParent(this);
        _false = __false;
    }

    public TriTestExpr(SourceLocation location, Expr _test, Expr __true, Expr __false) {
        super(location);
        setTest(_test);
        setTrue(__true);
        setFalse(__false);
    }
    protected TriTestExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        TriTestExpr ret = new TriTestExpr(getSourceLocation());
        ret.preCopy(walker, this);
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        if (_true != null) ret.setTrue( (Expr)walker.process(_true) );
        if (_false != null) ret.setFalse( (Expr)walker.process(_false) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return test;
        case 1: return _true;
        case 2: return _false;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "test";
        case 1: return "true";
        case 2: return "false";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTest((Expr)child); return;
        case 1: setTrue((Expr)child); return;
        case 2: setFalse((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "TriTestExpr()";
    }

    //END: Generated from @child and @property
}
