/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.bcg.ClassfileBuilder;

import java.io.*;
import java.util.*;

import java.lang.reflect.Modifier;

// INTRO from crosscuts
import org.aspectj.compiler.crosscuts.ast.PointcutDec;
import org.aspectj.compiler.crosscuts.*;

/**
 * Represents declared classes, interfaces and aspects
 *
 * One per declaration in the source or in the bytecodes
 * ??? Should bytecodes produce something else ???
 *
 * @grammar ...
 * @child Modifiers modifiers
 * @property String id
 * @child TypeDs superInterfaces
 * @child Decs body
 */
public abstract class TypeDec extends Dec {
    // ------------------------------
    // InnerInfo stuff

    public void walkInnerInfo(InnerInfoPass w) {
        int context = w.inType(this instanceof InterfaceDec);
        w.enterType(this);
        super.walkInnerInfo(w);
        w.leaveType();
        w.restoreContext(context);
    }

    // ------------------------------
    // some spec checking

    public void checkSpec() {
        super.checkSpec();
        checkConstructorLoops();
        checkParentNonFinal();
        if (isInner() || isLocal()) {
            checkNoStaticMembers();
        }
        if (!isAnonymous()) {
            TypeDec td = getEnclosingTypeDec();
            while (td != null) {
                if (td.getId().equals(getId())) {
                    showError("the name of this inner type conflicts with the name of an enclosing type: " + getId());
                }
                td = td.getEnclosingTypeDec();
            }
        }

        if (fromSource()) {
            if (!getType().isAbstract()) {
                getNameType().ensureNoAbstractDecs();
            }
            if (isPackageMember()) {
                PackageSO p = getTypeManager().findPackage(getPackageName());
                if (p.containsSubPackage(getId())) {
                    showError(getFullName() + " clashes with package of same name");
                }
            }
        }
    }

    private void checkNoStaticMembers() {
        for (int i=0, N=body.size(); i < N; i++) {
            Dec dec = body.get(i);
            if (!dec.isStatic()) continue;
            if (dec instanceof FieldDec) {
                FieldDec fd = (FieldDec)dec;
                if (!fd.isConstant()) {
                    dec.showError("all static fields in inners must be constant static finals");
                }
            } else {
                dec.showError("inner classes cannot have static declarations");
            }
        }
    }

    private void checkParentNonFinal() {
        if (((NameType)getSuperClassType()).isFinal()) {
            showError("May not extend a final class");
        }
    }

    private void checkConstructorLoops() {
        for (int i = 0, len = body.size(); i < len; i++) {
            Dec d = body.get(i);
            if (! (d instanceof ConstructorDec)) continue;
            ConstructorDec cd = (ConstructorDec) d;
            Set alreadySeen = new java.util.HashSet();
            while (! cd.isSuper()) {
                if (alreadySeen.contains(cd)) {
                    ((ConstructorBody)cd.getBody()).getConstructorCall()
                        .showError("a constructor may not directly or indirectly invoke itself");
                    break;
                }
                alreadySeen.add(cd);
                cd = cd.getNextConstructorDec();
            }
        }
    }

    // ------------------------------
    // INTRO from MemberClassMunger

    public ASTObject walkMemberMunger(MemberClassMunger w) {
        w.enterType(getNameType());
        if (isInner()) {
            final AST ast = getAST();
            NameType enclosingInstanceType = getEnclosingInstanceType();
            FieldDec enclosingInstanceField =
                ast.makeField(ast.makeModifiers(Modifiers.FINAL | Modifiers.PRIVATE),
                              enclosingInstanceType,
                              "this$" + enclosingInstanceType.getId());
            w.pushField(enclosingInstanceField);
            super.walkMemberMunger(w);
            w.popField();

            addFieldDec(enclosingInstanceField);
            getBody().add(0, enclosingInstanceField);
        } else {
            List fields = w.saveFields();
            super.walkMemberMunger(w);
            w.restoreFields(fields);
        }
        w.leaveType();
        return this;
    }

    // ------------------------------
    // INTRO from FlowCheckerPass
    public void walkFlow(FlowCheckerPass ww) {
        // ------------------------------
        // load up the body and the length into locals

        Decs body = getBody();
        int len = body.size();

        // ------------------------------
        // first collect up the static and non-static final fields

        FlowCheckerPass.Set staticFinalFields = FlowCheckerPass.Set.getNone();
        FlowCheckerPass.Set dynamicFinalFields = FlowCheckerPass.Set.getNone();

        for (int i = 0; i < len; i++) {
            Dec d = body.get(i);
            if ((d instanceof FieldDec) && d.isFinal()) {
                FieldDec dd = (FieldDec) d;
                if (d.isStatic()) {
                    staticFinalFields = staticFinalFields.add(dd);
                } else {
                    dynamicFinalFields = dynamicFinalFields.add(dd);
                }
            }
        }

//         System.err.println("S: " + staticFinalFields);
//         System.err.println("D: " + dynamicFinalFields);

        // ------------------------------
        // start with an empty flow checker

        FlowCheckerPass w = new FlowCheckerPass(getCompiler(), this);

        // ------------------------------
        // process the static world with no context

        // process our static fields and initializers
        for (int i = 0; i < len; i++) {
            Dec d = body.get(i);
            if (! d.isStatic()) continue;
            if ((d instanceof FieldDec) && d.isFinal()) {
                FieldDec fd = (FieldDec) d;
                if (fd.getInitializer() != null) {
                    // constants
                    w.setVars(w.getVars().addAssigned(fd));
                } else {
                    w.setVars(w.getVars().addUnassigned(fd));
                }


            } else if (d instanceof InitializerDec) {
                w.process(d);
            }
        }

        // now, go through the static final fields and complain if any of them
        // are uninitialized
        FlowCheckerPass.Set staticAssigned = w.getVars().getDa();
        for (FlowCheckerPass.Set x = staticFinalFields; ! x.isEmpty(); x = x.rest()) {
            if (! staticAssigned.contains(x.first())) {
                w.showVarError(this,
                               x.first(),
                               "Field " + x.first().getId()
                               + " might not be initialized");
            }
        }

        // ------------------------------
        // now, start with a flow-checker pass assuming all statics
        // and whatever is coming in from our context are definitely
        // assigned

        staticAssigned = ww.getVars().getDa().union(staticFinalFields);
        w = new FlowCheckerPass(getCompiler(), staticAssigned, this);

        // process our non-static fields and initializers
        for (int i = 0; i < len; i++) {
            Dec d = body.get(i);
            if (d.isStatic()) continue;
            if ((d instanceof FieldDec) && d.isFinal()) {
                FieldDec fd = (FieldDec) d;
                if (fd.getInitializer() != null) {
                    // constants
                    w.setVars(w.getVars().addAssigned(fd));
                } else {
                    w.setVars(w.getVars().addUnassigned(fd));
                }
            } else if (d instanceof InitializerDec) {
                w.process(d);
            }
        }

        FlowCheckerPass.Vars preConstructorVars = w.getVars();

        FlowCheckerPass.Set dynamicAssigned =
            staticAssigned.union(dynamicFinalFields);

        // now process each of our constructors.  For constructors
        // calling super, start w/ preConstructorVars, and check after
        // each super constructor that all the dynamicfinalfields are
        // initialized.  For constructors calling this, start w/
        // dynamicAssigned and don't check anything afterward.
        for (int i = 0; i < len; i++) {
            Dec d = body.get(i);
            if (! (d instanceof ConstructorDec)) continue;
            if (((ConstructorDec)d).isSuper()) {
                w = new FlowCheckerPass(getCompiler(), preConstructorVars, this);
                w.process(d);
                // check to make sure all dynamic finals are assigned
                FlowCheckerPass.Set currentAssigned = w.getVars().getDa();
                for (FlowCheckerPass.Set x = dynamicFinalFields; ! x.isEmpty(); x = x.rest()) {
                    if (! currentAssigned.contains(x.first())) {
                        w.showVarError(this,
                                       x.first(),
                                       "Field " + x.first().getId()
                                       + " might not be initialized");
                    }
                }
            } else {
                w = new FlowCheckerPass(getCompiler(), dynamicAssigned, this);
                w.process(d);
            }
        }

        // ------------------------------
        // now, start with all the other guys (methods, advice, etc),
        // assuming all the fields are definitely assigned

        for (int i = 0; i < len; i++) {
            Dec d = body.get(i);
            if ((d instanceof FieldDec)
                    || (d instanceof InitializerDec)
                    || (d instanceof ConstructorDec))
                continue;
            w = new FlowCheckerPass(getCompiler(), dynamicAssigned, this);
            w.process(d);
        }

        // and we're done, without side-effecting ww.
    }

    // a helper for AfterReturningAdviceDec and maybe others to generate unique labels
    public int getDepth() {
        if (getEnclosingTypeDec() == null) return 0;
        else return 1 + getEnclosingTypeDec().getDepth();
    }

    public void preMove(MovingWalker walker) {
        walker.pushLexicalType(getType());
    }

    public void preCopy(CopyWalker walker, ASTObject oldObject) {
        walker.pushLexicalType(getType());
    }

    /**
     * We now explicitly can't copy local types from one type dec to another
     */
    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        TypeDec oldTypeDec = (TypeDec)oldObject;
        //this.isLocal = oldTypeDec.isLocal;
        //this.isLocalFlag = oldTypeDec.isLocalFlag;
        //this.isInnerFlag = oldTypeDec.isInnerFlag;
        //this.isPackageMemberFlag = oldTypeDec.isPackageMemberFlag;
        //this.hasBeenInnerDiscovered = oldTypeDec.hasBeenInnerDiscovered;
        //System.out.println("walker: " + walker + " has " + walker.hasToType());
        if (walker.hasToType()) {
            //System.out.println(walker.getToType());
            this.setEnclosingTypeDec(walker.getToType().getTypeDec());
            //this.isInnerFlag = !walker.isStatic;
        } else {
            this.setEnclosingTypeDec(oldTypeDec.getEnclosingTypeDec());
        }
        return super.postCopy(walker, oldObject);
    }

    public ASTObject postMove(MovingWalker walker) {
        // walk all decs and set my type as declaring type
        for (int i = 0; i < body.size(); i++) {
            body.get(i).setDeclaringType(getType());
            //System.out.println(body.get(i).unparse());
        }


        walker.popLexicalType();
        return super.postMove(walker);
    }

    public /* String */ Set memberTypeNames = null;

    protected NameType type = null;
    public Type getType() {
        return getNameType();
    }

    public void setType(NameType type) {
        this.type = type;
    }

    public NameType getNameType() {
        if (type == null) {
            type = new NameType(this);
            //System.out.println("made type for: " + this.getExtendedId());
        }
        return type;
    }

    //XXX wrong
    public TypeDec getBytecodeTypeDec() { return this; }

    public SemanticObject makeCorrespondingSemanticObject() { return getType(); }

    private Set extraWithinTypes = new HashSet();
    public Set getExtraWithinTypes() { return extraWithinTypes; }
    public void addExtraWithinType(Type type) { extraWithinTypes.add(type); }

    public Type getSuperClassType() { return getTypeManager().getObjectType(); }

    public Collection getSuperInterfaceTypes() {
        return makeInterfaceTypesFromTypeDs(superInterfaces);
    }

    public void addSuperInterfaceType(Type newSuperType) {
        TypeDs list = this.getSuperInterfaces();
        if (list == null) {
            list = getAST().makeTypeDs(); //new TypeDs(getCompiler());
            this.setSuperInterfaces(list);
        }
        list.add(newSuperType.makeTypeD());
    }



    //public boolean isLocal = false;
    private TypeDec enclosingTypeDec = null;

    public void setEnclosingTypeDec(TypeDec dec) {
        enclosingTypeDec = dec;
        //??? add dec to enclosingTypeDecs inner types
    }

    public TypeDec getEnclosingTypeDec() {
        return enclosingTypeDec;
    }

    public TypeDec getDeclaringTypeDec() {
        return getEnclosingTypeDec();
    }

    public TypeDec getOutermostTypeDec() {
        TypeDec immediateParent = getEnclosingTypeDec();
        if (immediateParent == null) return this;
        return immediateParent.getOutermostTypeDec();
    }

    public Type getOutermostType() {
        TypeDec immediateParent = getEnclosingTypeDec();
        if (immediateParent == null) return this.getType();
        return immediateParent.getOutermostType();
    }

    public Type getOutermostLexicalType() {
        Type parentLexicalType = getLexicalType();
        if (parentLexicalType == null) return getType();
        return parentLexicalType.getTypeDec().getOutermostLexicalType();
    }

    public boolean isInnerType() {
        return getDeclaringTypeDec() != null;
    }

    public boolean isInnerTypeOf(TypeDec parentType) {
        TypeDec immediateParent = getDeclaringTypeDec();
        if (immediateParent == null) return false;
        if (immediateParent == parentType) return true;
        return immediateParent.isInnerTypeOf(parentType);
    }

    /** Does this typeDec not have an accessible typeName from the
        rest of the universe?  That is, is it a local or anonymous
        class, or is one of its declaring typedecs local or anonymous?
    */
    public boolean hasGlobalName() {
        if (isLocal()) return false;
        TypeDec immediateParent = getDeclaringTypeDec();
        if (immediateParent == null) return true;
        return immediateParent.hasGlobalName();
    }

    /** Is this typeDec in local context?  A local or anonymous class? */
    public boolean isLocallyDefined() {
        return isLocal();
    }

    public abstract boolean isAnonymous();


    public boolean canOverride(Dec otherDec) {
        return true;
    }

    /*
    public boolean isAccessible(ASTObject fromWhere, boolean inBytecode) {
        if (!isPackageMember()) {
            return super.isAccessible(fromWhere, inBytecode);
        }

        if (!inBytecode && hasPrivilegedAccess(fromWhere)) return true;
        if (isPublic()) return true;

        Type fromType;
        if (inBytecode) {
            if (fromWhere.getBytecodeTypeDec() == null) fromType = null;
            else fromType = fromWhere.getBytecodeTypeDec().getType();
        } else {
            fromType = fromWhere.getLexicalType();
        }

        //???if (isPrivate() || isProtected()) return false;

        String fromPackage=null;
        if (fromType != null) fromPackage = fromType.getPackageName();
        else fromPackage = fromWhere.getCompilationUnit().getPackageName();

        return nameMatches(getCompilationUnit().getPackageName(), fromPackage);
    }
    */


    public boolean isRoot() {
        return this == getTypeManager().getObjectType().getTypeDec();
    }

    public boolean isConcrete() {
        return !(this instanceof InterfaceDec);
    }

    private int localTypesCounter = 0;
    public int allocateLocalTypeIndex() {
        return ++localTypesCounter;
    }


	/**
	 * Returns the name of this class minus the package name
	 * 
	 * uses '$' as a separator for inners
	 * 
	 * java.util.Map.Entry goes to Map$Entry
	 */
    public String getExtendedId() {
        if (getEnclosingTypeDec() != null) {
            return getEnclosingTypeDec().getExtendedId() + '$' + id;
        } else {
            return id;
        }
    }

	/**
	 * Returns the name of this class minus the package name
	 * 
	 * uses '.' as a separator for inners
	 * 
	 * java.util.Map.Entry goes to Map.Entry
	 */
    public String getSourceExtendedId() {
        if (getEnclosingTypeDec() != null) {
            return getEnclosingTypeDec().getSourceExtendedId() + '.' + id;
        } else {
            return id;
        }
    }


    protected Collection makeInterfaceTypesFromTypeDs(TypeDs typeDs) {
        if (typeDs == null || typeDs.size() == 0) return Collections.EMPTY_LIST;

        List ret = new ArrayList(typeDs.size());

        for (int i = 0; i < typeDs.size(); i++) {
            Type type = typeDs.get(i).getType();
            if (!type.isInterface()) {
                typeDs.get(i).showError("interface required, not " + type.toShortString());
            } else {
                ret.add(type);
            }
        }

        return ret;
    }

    protected void addPointcutDec(PointcutDec pointcutDec) {
        type.addPointcut(pointcutDec);
    }

    protected void addFieldDec(FieldDec fieldDec) {
        type.addField(fieldDec);
    }

    protected void addMethodDec(MethodDec methodDec) {
         if (getModifiers().isStrict() && !methodDec.getModifiers().isAbstract()) {
             methodDec.getModifiers().setStrict(true);
         }
         type.addMethod(methodDec);
    }

    protected void addConstructorDec(ConstructorDec constructorDec) {
        //System.out.println("adding: " + constructorDec.getId() + " to " + type);
        if (getModifiers().isStrict()) constructorDec.getModifiers().setStrict(true);
        type.addConstructor(constructorDec);
    }

    protected void addTypeDec(TypeDec typeDec) {
        if (getModifiers().isStrict()) typeDec.getModifiers().setStrict(true);
        typeDec.enclosingTypeDec = this;
        typeDec.setAllEnclosingTypes(this.getType());
        ((NameType)getType()).addInnerType(typeDec);
    }

    public void addMemberTypeDec(TypeDec memberDec) {
        addTypeDec(memberDec);
        getBody().add(memberDec);
    }

    public void addMemberMethodDec(MethodDec memberDec) {
        addMethodDec(memberDec);
        getBody().add(memberDec);
    }

    public void addIntroducedDec(Dec dec) {
        getBody().add(dec);

        SemanticObject oldSO = null;
        if (dec instanceof MethodDec) {
            oldSO = type.methods.addIntroduced(dec.getCorrespondingSemanticObject());
        } else if (dec instanceof FieldDec) {
            oldSO = type.fields.addIntroduced(dec.getCorrespondingSemanticObject());
        } else if (dec instanceof ConstructorDec) {
            oldSO = type.constructors.addIntroduced(dec.getCorrespondingSemanticObject());
        }

        if (oldSO != null) {
            getBody().remove(oldSO.getCorrespondingDec());
        }
    }

    //??? don't like this being public
    protected void addDec(Dec dec) {
        //??? is this always right ???
        if (!dec.isLanguageVisible() && !(
            dec.getId().equals("aspectOf") || dec.getId().equals("hasAspect"))) return;

         if (dec instanceof TypeDec) {
            addTypeDec((TypeDec)dec);
        } else if (dec instanceof MethodDec) {
            addMethodDec((MethodDec)dec);
        } else if (dec instanceof FieldDec) {
            addFieldDec((FieldDec)dec);
        } else if (dec instanceof ConstructorDec) {
            addConstructorDec((ConstructorDec)dec);
        } else if (dec instanceof PointcutDec) {
            addPointcutDec((PointcutDec)dec);
        } else if (dec instanceof InitializerDec) {
            return;
        } else if (dec instanceof VarDec) {
            showWarning("shouldn't be found in a class");
        } else {
            dec.showWarning("shouldn't be found in " + this.toShortString());
        }
    }

    public final String getPackageName() {
        if (getEnclosingTypeDec() != null) {
            return getEnclosingTypeDec().getPackageName();
        } else {
            return getCompilationUnit().getPackageName();
        }
    }


    public void addInnerTypes() {
        // first add any inner classes for .class classes
        if (memberTypeNames != null) {
            for (Iterator i = memberTypeNames.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                if (name == null || name.length() == 0 ||
                        !Character.isJavaIdentifierStart(name.charAt(0))) {
                    //??? should we show a message here ???
                    getCompiler().showMessage("ignoring invalid inner class attribute: " +
                        getExtendedId()+'$'+name);
                    continue;
                }

                //System.out.println("=============" + name);
                Type innerType = getTypeManager().findType(getPackageName(),
                                        getExtendedId()+'$'+name);
                if (innerType == null) {
                    getCompiler().showMessage("ignoring invalid inner class attribute: " +
                        getExtendedId()+'$'+name);
                    continue;
                }
                addTypeDec(innerType.getTypeDec());
            }
        }

        Decs decs = getBody();
        if (decs == null) {
            if (getOptions().torture) showWarning("no body for " + this);
            return;
        }
        final int N = decs.size();
        for(int i=0; i < N; i++) {
            if (decs.get(i) instanceof TypeDec) {
                addDec(decs.get(i));
            }
        }
    }


    public void addDecs(NameType toThis) {
        Decs decs = getBody();
        if (decs == null) {
            if (getOptions().torture) showWarning("no body for " + this);
            return;
        }
        final int N = decs.size();
        for(int i=0; i < N; i++) {
            //System.out.println(">>>>" + decs.get(i));
            if (! (decs.get(i) instanceof TypeDec) ) {
                addDec(decs.get(i));
            }
        }
    }

    private ConstructorDec makeDefaultConstructor() {
        final AST ast = getAST();

        ConstructorCallExpr superCall = null;
        if (!isRoot()) {
            Constructor superConstructor =
                  getSuperClassType().getConstructor(getBody(), ast.makeExprs(), true);
            superCall = ast.makeSuperConstructorCall(superConstructor);
        }
        ConstructorDec ret = ast.makeConstructor(
                // JLS 8.8.7 says to replace Modifiers.PUBLIC with
                //the commented-out version
                ast.makeModifiers(getModifiers().getAccessValue()),
                ast.makeFormals(), ast.makeTypeDs(),
                superCall, ast.makeStmts());
        ret.setLanguageVisible();
        //XXX this gets package protection to work "right", but it's ugly
        ret.setSourceLocation(getSourceLocation());
        return ret;
    }

    void addDefaultConstructor() {
        //XXX should be an error
        if (body == null || getType().isObject()) return;

        //XXX
        if (isAnonymous()) return;

        //System.out.println("adding default constructor to: " + this + ", " + this.getBody().unparse());

        ConstructorDec ret = makeDefaultConstructor();
        addConstructorDec(ret);
        body.add(ret);
    }

    // everybody gets a default constructor, even interfaces...
    public void postIntroductionFinish() {
        if (((NameType)getType()).getConstructors().size() == 0) {
            if (!isAnonymous()) addDefaultConstructor();
        } else if (isAnonymous()) {
            showError("Anonymous classes cannot define constructors");
        }
    }

    //XXX how much work should I do here?
    public void addToBody(Dec dec) {
        getBody().add(dec);
    }

    public void addToBodyAndType(Dec dec) {
        addToBody(dec);
        addDec(dec);
    }


    public String getPrettyString() {
        return getType().getPrettyString();
    }


    public String toString() {
        return getType().getString();
    }

    public String toShortString() {
        return getType().getString();
    }

    public void fixAST(ASTFixerPass fixer) {
        TypeDec saveTypeDec = fixer.getInTypeDec();
        fixer.setInTypeDec(this);
        this.walk(fixer);
        fixer.setInTypeDec(saveTypeDec);

        // if we inherit any methods from interfaces, build
        // "concrete-abstract" versions in this class
        if (getOptions().jvmTorture) return;

        NameType myType = getNameType();
        if (myType.isInterface()) return;
        if (!myType.isAbstract()) return;

        for (Iterator i = myType.getMethods().iterator(); i.hasNext(); ) {
            Method m = (Method)i.next();
            if (m.getDeclaringType().isInterface() && 
                m.getMethodDec().getBody() == null &&
                !m.getMethodDec().isIntroduced())
            {
                getBody().add((MethodDec)m.getMethodDec().copy());
            }
        }
    }

    protected void walkExtendsAndImplements(ScopeWalker walker) {
        if (superInterfaces != null) walker.process(superInterfaces);
    }

    protected void walkBody(ScopeWalker walker) {
        walker.process(body);
    }

    //XXX very inefficient, but probably doesn't matter
    protected Scope getEnclosingScope(boolean justTypeGraph) {
        if (isLocal()) {
            getCompiler().internalError(this, "trying to get simple scope for a local type");
        }
        TypeDec td = getEnclosingTypeDec();
        if (td == null) {
            return getCompilationUnit().getScope();
        } else {
            Scope ret = new TypeScope(getCompiler(), td.getEnclosingScope(justTypeGraph), td.getType());
            if (justTypeGraph) td.addToTypeGraph();
            else td.buildSignatures();
            return ret;
        }
    }

	//don't do any scope walks on synthetic type decs
    public void addToTypeGraph() {
        if (!fromSource()) getNameType().addToTypeGraph();
        else addToTypeGraph(new ScopeWalker(getCompiler(), getEnclosingScope(true)));
    }

    public void buildSignatures() {
        if (!fromSource()) getNameType().buildSignatures();
        else buildSignatures(new ScopeWalker(getCompiler(), getEnclosingScope(false)));
    }

    private boolean addToTypeGraphDone = false;
    private boolean addToTypeGraphInProgress = false;

    /**
     *  1. Bind names in extends/implements clauses
     *  2. Insert this type into type graph
     *  3. Build maps of inner types including inherited inner types
     */
    public void addToTypeGraph(ScopeWalker walker) {
        if (addToTypeGraphDone) return;
        if (addToTypeGraphInProgress) {
            showError("circularity detected, already adding type to type graph");
            return;
        }
        addToTypeGraphInProgress = true;

        walker.addTypeDec(this);

        // bind names in extends and implements
        walkExtendsAndImplements(walker);
        getNameType().addToTypeGraph(); // includes building inner type map
        //getType().buildInnerTypeMap();

        addToTypeGraphDone = true;

        walker.pushScope(new TypeScope(getCompiler(),
                                       walker.getScope(), getType()));

        Decs decs = getBody();
        for (int i=0, N=decs.size(); i<N; i++) {
            Dec dec = decs.get(i);
            if (dec instanceof TypeDec) ((TypeDec)dec).addToTypeGraph(walker);
        }

        //??? the timing on this is not ideal...
        //XXX really might want this to be a separate pass
        boolean saveWalkBodies = walker.walkBodies();
        walker.setWalkBodies(false);
        walkBody(walker);
        walker.setWalkBodies(saveWalkBodies);

        walker.popScope();
    }


    private boolean buildSignaturesDone = false;
    private boolean buildSignaturesInProgress = false;

    /**
     *  1. Bind names in member signatures
     *  2. Build maps of members of this type and of inherited members
     *     This step also handles introduced members
     *  3. Add a default constructor if required
     *
     *  4. Do a spec check for no abstract decs in a concrete type
     *     ??? can we move this to checkSpec() method ???
     */
    public void buildSignatures(ScopeWalker walker) {
        if (buildSignaturesDone) return;
        if (buildSignaturesInProgress) {
            showError("circularity detected, already building signatures");
            return;
        }
        buildSignaturesInProgress = true;

        addToTypeGraph(walker);

        getNameType().buildSignatures();
        //getType().buildIntroductions();
        //getType().buildInheritedSignatures();

        postIntroductionFinish();

        buildSignaturesDone = true;

        walker.pushScope(new TypeScope(getCompiler(),
                                       walker.getScope(), getType()));

        Decs decs = getBody();
        for (int i=0, N=decs.size(); i<N; i++) {
            Dec dec = decs.get(i);
            if (dec instanceof TypeDec) ((TypeDec)dec).buildSignatures(walker);
        }
        walker.popScope();

    }

    /**
     *  This is the simplest scope pass of them all.
     *  It just walks the body and binds all names (after ensuring that
     *  prvious name binding has been done).
     */
    public void walkScope(ScopeWalker walker) {
        if (! (walker instanceof NameHygienePass) ) {
            if (walker.walkBodies()) buildSignatures(walker);
            else addToTypeGraph(walker);
        } else {
            walkExtendsAndImplements(walker);
        }

        // bind names in bodies
        walker.pushScope(new TypeScope(getCompiler(),
                                       walker.getScope(), getType()));
        walkBody(walker);
        walker.popScope();
    }

    // INTRO from JoinPoint
    public FieldDec joinPointFactoryDec = null;

    // INTRO from JoinPointCollector
    public List joinPoints0 = new ArrayList();
    public List joinPoints1 = new ArrayList();
    public List joinPoints2 = new ArrayList();

    //private String fullName = null;
    public String getFullName() {
        String fullName = null;
        //System.out.println("id: " + getId() + " in " + getEnclosingTypeDec());

        if (fullName == null) {
            if (getEnclosingTypeDec() != null) {
                fullName = getEnclosingTypeDec().getFullName() + '$' + getId();
            } else {
                String packageName = getPackageName();
                String id = getId();
                fullName = packageName == null ? id : packageName + '.' + id;
            }
            //return "null"; //XXXthrow new RuntimeException("fullName of "+this+" == null");
        }
        return fullName;
    }

    // ------------------------------
    // non-package-level type stuff

    /** should only be called from ASTConnection (or here, I suppose) */
    public void setInnerDiscoveries(boolean a, boolean b, boolean c) {
        setIsInner(b);
    }

    public void setIsInner(boolean b) {
        if (b) {
            throw new RuntimeException("TypeDec " + this + " should never be inner");
        }
    }

	/**
	 * Only classes can be inners, so this is overridden by ClassDec
	 */
    public boolean isInner() { return false; }
    public void setLocal() { } //isLocalFlag = true; }
    public boolean isLocal() {
        ASTObject parent = getParent();
        if (parent == null) return false;
        parent = parent.getParent();
        if (parent == null || parent instanceof TypeDec
            || parent instanceof CompilationUnit) return false;
        return true;
    }

    public boolean isPackageMember() {
        ASTObject parent = getParent();
        if (parent == null) return false;
        parent = parent.getParent();
        if (parent == null || !(parent instanceof CompilationUnit)) return false;
        return true;
    }

    // Only valid if isInner()
    public NameType getEnclosingInstanceType() {
        return (NameType)getDeclaringType();
    }
    public TypeDec getEnclosingInstanceTypeDec() {
        return getEnclosingInstanceType().getTypeDec();
    }

    //INTRO: InitializerExecutionPoint
    private Set initializerExecutionJoinPoints = new HashSet();
    public Set getInitializerExecutionJoinPoints() {
        return initializerExecutionJoinPoints;
    }

    // ------------------------------
    // Intro: ForwardReferenceChecker

    public void walkForwardReference(ForwardReferenceChecker w) {
        this.walk(w.createTypeChecker(this));
    }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        if (isLocal()) {
            walker.enterTypeDec(this);
            setId(walker.makeNewId(this));
        }
        walker.inType();
        this.walk(walker);
        if (isLocal()) walker.leaveTypeDec();
    }

    // ------------------------------
    // INTRO:  LocalClassPass.LiftWalker

    public void preLift(LocalClassPass.LiftWalker walker) {
        walker.pushPendingDecs();
    }

    public ASTObject postLift(LocalClassPass.LiftWalker walker) {
        initializersToConstructors();
        walker.popPendingDecsInto(this);
        if (isLocal()) {
            walker.addToPendingDecs(this);
            return null;
        } else {
            return this;
        }
    }

    public void collectInitializers(boolean collectSynthetics) {
        collectInitializers(true, collectSynthetics);
        collectInitializers(false, collectSynthetics);
    }

    Field assertionsDisabledField = null;
    /** This field is actually added in postFixAST
     */
    public Field getAssertionsDisabledField() {
        if (assertionsDisabledField != null) return assertionsDisabledField;

        final AST ast = getAST();
        Modifiers mods =
            ast.makeModifiers(Modifiers.STATIC|Modifiers.FINAL);
        Type type = getTypeManager().booleanType;
        String id = "$assertionsDisabled";
        Expr init = ast.makeUnop("!",
                                 ast.makeCall(ast.makeLiteral(getType()),
                                 "desiredAssertionStatus"));
        FieldDec adf = ast.makeField(mods, type, id, init);
        adf.setDeclaringType(getType());
        //getBody().add(0, adf);
        assertionsDisabledField = adf.getField();
        return assertionsDisabledField;
    }
    
    public ASTObject postFixAST(final ASTFixerPass fixer) {
        if (assertionsDisabledField != null) {
            getBody().add(0, assertionsDisabledField.getFieldDec());
        }
        return super.postFixAST(fixer);
    }


    public void collectInitializers(boolean isStatic, boolean collectSynthetics) {
        // don't collect static initializers in interfaces if we're going to
        // be generating source code
        if (isStatic &&
            (!collectSynthetics || getCompiler().willGenerateSourceCode())) {
            if (this instanceof InterfaceDec) return;
        }


        final AST ast = getAST();
        Stmts stmts = ast.makeStmts();
        InitializerDec ret = null;
        boolean singleInitializer = true;

        for (ListIterator i = body.iterator(); i.hasNext(); ) {
            Dec memDec = (Dec)i.next();
            if (isStatic != memDec.isStatic()) continue;
            //XXXif (!memDec.isLanguageVisible()) continue; //XXX trouble for bcg

            if (memDec instanceof InitializerDec) {
                InitializerDec iDec = (InitializerDec)memDec;
                //if (iDec.isStatic() && inInterface) continue;
                //if (iDec.getBody().getStmts().size() == 0) continue;
                stmts.addAll(iDec.getBody().getStmts());
                if (ret == null) ret = iDec;
                else singleInitializer = false;
                i.remove();
            } else if (memDec instanceof FieldDec) {
                FieldDec fDec = (FieldDec)memDec;
                if (collectSynthetics || fDec.isLanguageVisible()) {
                    fDec.getAssignExpr(); // this creates the initializer code by magic
                }
            }
        }

        if (ret == null || true) { // || !singleInitializer) {
            //System.out.println(stmts.size() + ": ");
            //if (stmts.size() == 1) System.out.println(stmts.get(0));
            if (stmts.size() == 1 && stmts.get(0) instanceof BlockStmt) {
                stmts = ((BlockStmt)stmts.get(0)).getStmts();
                //System.out.println("shrunk: " + stmts.unparse());
            }

            if (isStatic) ret = ast.makeStaticInitializer(stmts);
            else ret = ast.makeInitializer(stmts);
            //ret.setLanguageVisible();
        }
        ret.setSourceLocation(getSourceLocation());
        body.add(ret);
        //return ret;
    }

    public InitializerDec getSingleInitializerDec(boolean isStatic) {
        for (ListIterator i = body.iterator(); i.hasNext(); ) {
            Dec memDec = (Dec)i.next();
            if (memDec instanceof InitializerDec &&
                isStatic == memDec.isStatic()) return (InitializerDec)memDec;
        }
        showError("no initializer for: " + this);
        return null;
    }


    // Assumes there is zero or one static initializer, and leaves it
    // alone (or erase it if it's empty).  also assumes that there are
    // NO field-decs in the initializers... i.e., that there are no
    // unlifted local/anonymous classes.
    private void initializersToConstructors() {
        final AST ast = getAST();
        Stmts stmts = ast.makeStmts();

        // first collect the initializers
        for (Iterator i = getBody().iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof InitializerDec) {
                InitializerDec iDec = (InitializerDec) o;
                if (iDec.isStatic()) {
                    if (iDec.getBody().getStmts().size == 0) i.remove();
                    continue;
                }
                i.remove();
                Stmt iStmt = iDec.getBody();
                if (iStmt instanceof BlockStmt &&
                    ((BlockStmt)iStmt).getStmts().size() == 0)
                    continue;
                stmts.add(0, iStmt);
            }
        }
        // then stick them in the constructors
        for (Iterator i = getBody().iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof ConstructorDec) {
                ConstructorDec cDec = (ConstructorDec) o;
                // we only do this for constructors that don't call this
                ConstructorCallExpr call =
                                ((ConstructorBody)cDec.getBody()).getConstructorCall();
                if (call != null && !call.getIsSuper()) continue;

                Stmts newStmts = (Stmts) CopyWalker.copy(stmts);
                // !!! this _must_ set up links to copied local vars.
                cDec.getBody().getStmts().addAll(0, newStmts);
            }
        }
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public void preThreading(LocalClassPass.ThreadingWalker walker) {
        walker.pushNonConstructorEnv();
        walker.pushTypeDec(this);
    }

    public ASTObject postThreading(LocalClassPass.ThreadingWalker walker) {
        walker.popTypeDec();
        walker.popEnv();
        return this;
    }

    // ------------------------------
    // bcg

    protected void cgMember(ClassfileBuilder maker) {
        maker.delayInnerClassGeneration(this);
    }

    /** Used to generate .class files in TypeDec.generateBytecode method
        Used to generate .java files in CodeGenerator.?
       */
    public File getPackageDir(File outputdir) {
        String packagename = getPackageName();
        //System.out.println(this + " p " + packagename);

        if (packagename == null) return outputdir;

        int lastDot = 0;
        int dot = 0;
        while ((dot = packagename.indexOf('.', lastDot)) != -1) {
            outputdir = new File(outputdir, packagename.substring(lastDot, dot));
            lastDot = dot+1;
        }
        return new File(outputdir, packagename.substring(lastDot));
    }


    /** should only be called from {@link CompilationUnit} and from
        {@link ClassfileBuilder} */
    public final void generateBytecode(File outputDir) throws IOException {
        String filename = getExtendedId()+".class";
        getCompiler().showMessage("  writing " + filename);

        getCompiler().beginSection("bcg:AST-directed");
        // start AST-directed generation
        ClassfileBuilder cfb = new ClassfileBuilder(getCompiler());
        cfb.setSourceFile(getSourceLocation());
        if (!this.isConcrete()) getModifiers().setInterface(true);
        cfb.addAccessFlags(getModifiers().getAcceptableClassValue());

        cfb.setClassName((NameType)getType());
        cfb.setSuperClassName((NameType)getSuperClassType());
        for (Iterator i = getSuperInterfaceTypes().iterator(); i.hasNext(); ) {
            NameType iface = (NameType)i.next();
            cfb.addInterface(iface);
        }
        for (Iterator i = getBody().iterator(); i.hasNext(); ) {
            Dec memberDec = (Dec)i.next();
            memberDec.cgMember(cfb);
        }
        //XXX get this right -- if (isSynthetic())      cfb.setSynthetic();
        // end AST-directed generation

        // start assembly
        getCompiler().beginSection("bcg:assembly");
        cfb.resolve();
        // end assembly

        // start IO
        getCompiler().beginSection("bcg:output");
        File packagedir = getPackageDir(outputDir);
        packagedir.mkdirs();
        File outFile = new File(packagedir, filename);
        FileOutputStream fileStream = new FileOutputStream(outFile);
        DataOutputStream stream =
            new DataOutputStream(new BufferedOutputStream(fileStream));
        cfb.writeTo(stream);
        stream.close();
        // end IO

        cfb.generateBytecodeForDelayedInnerClasses(outputDir);
    }

    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }

    protected String id;
    public String getId() { return id; }
    public void setId(String _id) { id = _id; }

    protected TypeDs superInterfaces;
    public TypeDs getSuperInterfaces() { return superInterfaces; }
    public void setSuperInterfaces(TypeDs _superInterfaces) {
        if (_superInterfaces != null) _superInterfaces.setParent(this);
        superInterfaces = _superInterfaces;
    }

    protected Decs body;
    public Decs getBody() { return body; }
    public void setBody(Decs _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    public TypeDec(SourceLocation location, Modifiers _modifiers, String _id, TypeDs _superInterfaces, Decs _body) {
        super(location);
        setModifiers(_modifiers);
        setId(_id);
        setSuperInterfaces(_superInterfaces);
        setBody(_body);
    }
    protected TypeDec(SourceLocation source) {
        super(source);
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return superInterfaces;
        case 2: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "superInterfaces";
        case 2: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setSuperInterfaces((TypeDs)child); return;
        case 2: setBody((Decs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 3;
    }

    public String getDefaultDisplayName() {
        return "TypeDec(id: "+id+")";
    }

    //END: Generated from @child and @property
}
