/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.cst.BlockScope;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

import org.aspectj.compiler.base.LocalClassPass;

/**
 * @grammar id
 * @property VarDec varDec
 */

public class VarExpr extends AssignableExpr {
    protected Type discoverType() {
        return getVarDec().getType();
    }

    public Expr makeReference() {
        return getAST().makeVar(getVarDec());
    }

    public boolean canBeCopied() {
        return true;
    }

    public String getId() { return varDec.getId(); }

    //INTRO from CopyWalker
//    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
//        VarDec newDec = (VarDec)walker.lookupMapping(((VarExpr)oldObject).varDec);
//        if (newDec != varDec) setVarDec(newDec);
//        return this;
//    }

    //INTRO from MovingWalker
    public ASTObject postMove(MovingWalker walker) {
        return walker.moveVarExpr(this);
    }

    public void unparse(CodeWriter writer) {
        writer.write(varDec.getId());
        return;
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        VarDec dec = getVarDec();
        if (! (dec instanceof FormalDec)
                && ! w.getVars().isDefinitelyAssigned(getVarDec())) {
            showError("Variable " + getId() + " might not have a value");
        }
    }

    // ------------------------------
    // INTRO from AssignmentCheckerPass

    public ASTObject postAssignmentCheck(AssignmentCheckerPass walker) {
        if (isLhs()) return this;
        Expr initExpr = varDec.getInitializer();
        if (initExpr != null && varDec.isFinal() &&
            initExpr instanceof LiteralExpr && ! getType().isObject() ) {
            return initExpr.copy().setSource(this);
        } else {
            return this;
        }
    }

    // ------------------------------
    // INTRO:  LocalClassPass.AnalysisWalker

    public void walkAnalysis(LocalClassPass.AnalysisWalker walker) {
        walker.processVarExpr(this);
    }

    // ------------------------------
    // INTRO:  LocalClassPass.ThreadingWalker

    public ASTObject postThreading(LocalClassPass.ThreadingWalker walker) {
        Object o = walker.lookup(getVarDec());
        if (o == null) {
            return this;
        } else if (o instanceof FieldDec) {
            FieldDec fieldDec = (FieldDec)o;
//            System.out.println("threading: " + this + ", " + fieldDec + ", " +
//                fieldDec.getDeclaringType() + "::" + getLexicalType());
            return getAST().makeGet(getAST().makePrimary(
                                                  (NameType)fieldDec.getDeclaringType(),
                                                  (NameType)getLexicalType(), false),
                        fieldDec);
        } else {
            return getAST().makeVar((FormalDec) o);
        }
    }

    // ------------------------------
    // bcg

    protected void cgEffect(CodeBuilder cb) {}
    protected void cgLvalue(CodeBuilder cb) {}
    protected void cgLtoRvalue(CodeBuilder cb) {
        getType().emitLoad(cb, getVarDec().getFrameLoc());
    }
    protected void cgAssignment(CodeBuilder cb) {
        getType().emitStore(cb, getVarDec().getFrameLoc());
    }
    protected void cgDupLvalue(CodeBuilder cb) {}
    protected void cgDupRvalue(CodeBuilder cb) {
        getType().emitDup(cb);
    }
    protected boolean isIncable() {
        return getType() instanceof IntType &&
            getVarDec().getFrameLoc() < 0xFF;
    }

    //BEGIN: Generated from @child and @property
    protected VarDec varDec;
    public VarDec getVarDec() { return varDec; }
    public void setVarDec(VarDec _varDec) { varDec = _varDec; }

    public VarExpr(SourceLocation location, VarDec _varDec) {
        super(location);
        setVarDec(_varDec);
    }
    protected VarExpr(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        VarExpr ret = new VarExpr(getSourceLocation());
        ret.preCopy(walker, this);
        ret.varDec = varDec;
        return ret;
    }


    public String getDefaultDisplayName() {
        return "VarExpr(varDec: "+varDec+")";
    }

    //END: Generated from @child and @property
}
