/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.ast;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CodeWriter;

import org.aspectj.compiler.base.bcg.CodeBuilder;
import org.aspectj.compiler.base.bcg.Label;

/**
 * @grammar while (test) {body}
 * @child Expr test
 * @child Stmt body
 */

public class WhileStmt extends TestStmt {

    public boolean isBreakable() { return true; }
    public boolean isContinuable() { return true; }

    public void unparse(CodeWriter writer) {
        writer.writeKeyword("while");
        writer.optionalSpace();
        writer.parenExpr(test);
        writer.write(body);
    }

    // ------------------------------
    // INTRO from FlowCheckerPass

    public void walkFlow(FlowCheckerPass w) {
        // w.setArgs(w.getArgs());
        FlowCheckerPass.Set enteringPossiblyAssigned = w.popPossiblyAssigned();
        w.process(getTest());
        FlowCheckerPass.Vars p = w.getVars();
        FlowCheckerPass.Vars tv = p.getTrue();
        FlowCheckerPass.Vars fv = p.getFalse();

        w.setLive(! getTest().isConstantFalse());
        w.setVars(tv);
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();

        FlowCheckerPass.Vars v = w.getVars();
        FlowCheckerPass.Vars cv = w.getContinueVars(this);
        FlowCheckerPass.Vars bv = w.getBreakVars(this);

        FlowCheckerPass.Set loopPossiblyAssigned = w.popPossiblyAssigned();

        w.checkLoopingFinals(this, loopPossiblyAssigned, v.join(cv));
        w.mergePossiblyAssigned(enteringPossiblyAssigned);
        w.mergePossiblyAssigned(loopPossiblyAssigned);

        w.setLive(! getTest().isConstantTrue() || w.isBroken(this));
        w.setVars(fv.join(bv));
    }

    // ------------------------------
    // INTRO from ByteCodeCleanupPass

    public void walkCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantFalse()) return;
        w.enterContext(this);
        w.process(getBody());
        w.leaveContext();
        w.setLive(! getTest().isConstantTrue() || w.isBroken(this));
    }

    public ASTObject postCleanup(ByteCodeCleanupPass w) {
        if (getTest().isConstantFalse()) {
            return getAST().makeEmptyStmt().setSource(this);
        } else {
            return this;
        }
    }

    // ------------------------------
    // bcg

    protected void cgStmt(CodeBuilder cb) {
        Label startLab = cb.genLabel();
        Label testLab = cb.genLabel();
        Label endLab = cb.genLabel();
        cb.emitJump(testLab);
        cb.emitLabel(startLab);
        cb.enterNonWindingContext(this, endLab, testLab);
        body.cgTop(cb);
        cb.leaveContext();
        cb.emitLabel(testLab);
        test.cgTest(cb, startLab, endLab);
        cb.emitLabel(endLab);
    }

    // ------------------------------
    // flow

//      protected FlowEnv checkFlow(FlowEnv env) {
//          return checkLabeledFlow(env, null);
//      }

//      protected FlowEnv checkLabeledFlow(FlowEnv env, String label) {
//          FlowEnv testEnv = env.testEnv().check(getTest());
//          FlowEnv trueEnv = testEnv.getTrueEnv();
//          FlowEnv falseEnv = testEnv.getFalseEnv();
//          AssignExprs assignExprs = testEnv.getAssignExprs();

//          FlowEnv preBodyEnv = trueEnv.continuableBreakable(label);
//          FlowEnv bodyEnv = preBodyEnv.check(getBody());
//          FlowEnv breakEnv = bodyEnv.getBreaks();
//          FlowEnv continueEnv = bodyEnv.getContinues();
//          FlowEnv finishingEnv = bodyEnv.stripBreakContinue();

//          // check to make sure that if something was not definitely
//          // unassigned in a continue statement, then we don't assign it
//          // in the test expression.
//          continueEnv.checkAssignExprs(assignExprs);

//          return falseEnv.join(breakEnv);
//      }

    //BEGIN: Generated from @child and @property
    protected Expr test;
    public Expr getTest() { return test; }
    public void setTest(Expr _test) {
        if (_test != null) _test.setParent(this);
        test = _test;
    }

    protected Stmt body;
    public Stmt getBody() { return body; }
    public void setBody(Stmt _body) {
        if (_body != null) _body.setParent(this);
        body = _body;
    }

    public WhileStmt(SourceLocation location, Expr _test, Stmt _body) {
        super(location);
        setTest(_test);
        setBody(_body);
    }
    protected WhileStmt(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        WhileStmt ret = new WhileStmt(getSourceLocation());
        ret.preCopy(walker, this);
        if (test != null) ret.setTest( (Expr)walker.process(test) );
        if (body != null) ret.setBody( (Stmt)walker.process(body) );
        return ret;
    }

    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return test;
        case 1: return body;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "test";
        case 1: return "body";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTest((Expr)child); return;
        case 1: setBody((Stmt)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }

    public String getDefaultDisplayName() {
        return "WhileStmt()";
    }

    //END: Generated from @child and @property
}
