/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bcg;

import org.aspectj.compiler.base.bcg.pool.*;

import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.PrintStream;
import java.io.IOException;

import java.util.HashSet;
import java.util.Set;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;

// XXX annoying big modularity violation.  
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.ast.ControlContext;
import org.aspectj.compiler.base.ast.VarDec;
import org.aspectj.compiler.base.ast.LabeledStmt;
import org.aspectj.compiler.base.ast.WindingStmt;
import org.aspectj.compiler.base.ast.Stmt;
import org.aspectj.compiler.base.ast.Type;
import org.aspectj.compiler.base.ast.ArrayType;
import org.aspectj.compiler.base.ast.RefType;
import org.aspectj.compiler.base.ast.NameType;
import org.aspectj.compiler.base.ast.SourceLocation;

/** Lifecycle: creation, addition of code (through the emitX methods),
    resolution of constant pool (done in the constant pool, when all
    constant pool entries are set), resolution of this, output.  It is
    an error for these steps to occur out of order.  It is checked
    that this is resolved after the constant pool, and printed after
    it is resolved.  It is still possible to mess up by resolving the
    constant pool and adding new codes to this, but that shouldn't
    matter, since the constant pool can't be resolved twice. */
public final class CodeBuilder {

    // ------------------------------
    // state

    private boolean isResolvedFlag = false;
    public boolean isResolved() { return isResolvedFlag; }

    // ------------------------------
    // state

    ConstantPool pool;
    ClassfileBuilder cfb;
    Attributes attributes;
    CodeBuilder(ClassfileBuilder cfb) {
        this.cfb = cfb;
        this.pool = cfb.pool;
        this.attributes = new Attributes(pool);
    }

    public JavaCompiler getCompiler() { return cfb.getCompiler(); }

    int maxFrame;
    public void setMaxFrame(int maxFrame) {
        this.maxFrame = maxFrame;
    }
    int maxStack = 0;
    int stackSize = 0;

    /* Label */    List starts   = new ArrayList();
    /* Label */    List ends     = new ArrayList();
    /* Label */    List handlers = new ArrayList();
    /* Constant */ List names    = new ArrayList();

    List codes = new ArrayList();
    Code[] codea;
    
    List labels = new ArrayList();

    // ------------------------------
    // stack 

    public int getStackSize() {
        return stackSize;
    }
    public void pushStack(int delta) {
        stackSize += delta;
        stackInvariant();
    }
    public void adjustStack(int delta) {
        pushStack(delta);
    }
    public void popStack(int delta) {
        stackSize -= delta;
    }
    public void setStackSize(int ss) {
        stackSize = ss;
        stackInvariant();
    }
    public void clearStack() {
        stackSize = 0;
    }
    private void stackInvariant() {
        if (stackSize > maxStack) maxStack = stackSize;
    }

    // ------------------------------
    // control context

    private ControlContext context = new ControlContext();
    private Map breakLabels = new HashMap();
    private Map continueLabels = new HashMap();
    private Map windingLabels = new HashMap();

    public void enterNonWindingContext(Stmt s, Label b, Label c) {
        context.enter(s);
        breakLabels.put(s, b);
        if (c != null) continueLabels.put(s, c);
    }
    public void enterWindingContext(Stmt s, Label w) {
        context.enter(s);
        windingLabels.put(s, w);
    }
    public void leaveContext() { context.exit(); }

    public void doBreak(String label) {
        Stmt target = context.getBreakTarget(label);
        doJsrUntil(target);
        emitJump((Label)breakLabels.get(target));
    }
    public void doContinue(String label) {
        Stmt target = context.getContinueTarget(label);
        doJsrUntil(target);
        emitJump((Label)continueLabels.get(target));
    }
    private void doJsrUntil(Stmt target) {
        for (Iterator i = context.getWindsUntil(target); i.hasNext(); ) {
            emitJSR((Label)windingLabels.get(i.next()));
        }
    }
    public void doReturn(Type t) {
        for (Iterator i = context.getWindsUntil(null); i.hasNext(); ) {
            WindingStmt w = (WindingStmt) i.next();
            int rvl = w.getRetValLoc();
            t.emitStore(this, rvl);
            emitJSR((Label)windingLabels.get(w));
            t.emitLoad(this, rvl);
        }
        t.emitReturn(this);
    }

    // ------------------------------
    // local variable table stuff

    // We handle local variables by using a Stack, and pushing and
    // popping as appropriate.  When we enter a block, for stmt, or
    // catch block, we push a pair of null markers to indicate the
    // block.  When we leave a block, we pop down to those markers.
    // The stack also contains varDec x label.  Popping through a
    // VarDec x label ends its scope.  Pushing a VarDec x label enters
    // its scope, with that label.  Easy peasy.

    // We're using java.util.Stack, but just it's stack-like
    // operations.  So this would work even if java.util.Sack
    // encapsulated java.util.Vector, instead of making the mistake of
    // extending it.

    private Stack localVarStack = new Stack();

    public void enterBlock() {
        //System.err.println("CB/entering block");
        localVarStack.push(null); 
        localVarStack.push(null);
    }

    public void enterVar(VarDec v) {
        //System.err.println("CB/entering var " + v);
        Label start = genAnchor();
        emitLabel(start);
        localVarStack.push(v);
        localVarStack.push(start);
    }        

    public void exitBlock() {
        //System.err.println("CB/leaving block");
        Label start = (Label) localVarStack.pop();
        VarDec v = (VarDec) localVarStack.pop();
        for (; v != null; 
             start = (Label) localVarStack.pop(),
             v = (VarDec) localVarStack.pop()) {
            //System.err.println(" of var " + v);
            Label end = genAnchor();
            emitLabel(end);
            localVars = new LocalVar(start, end,
                                      pool.addUtf8(v.getBytecodeId()),
                                      pool.addUtf8(v.getDescriptor()),
                                      v.getFrameLoc(),
                                      localVars);
        }
    }
    
    // the local vars themselves are stored as a list of four-element
    // tuples, as we do the line number table later. 
        
    private static final class LocalVar {
        Label start;
        Label end;
        Constant name;
        Constant descriptor;        
        int index;
        LocalVar next;
        LocalVar(Label start,
                 Label end,
                 Constant name,
                 Constant descriptor,        
                 int index,
                 LocalVar next) {
            this.start = start;
            this.end = end;
            this.name = name;
            this.descriptor = descriptor;        
            this.index = index;
            this.next = next;
        }
        public String toString() {
            return "(" + start + " " + end + " " + name + " " + descriptor + " "
                + index + "):" + next;
        }
    }

    private LocalVar localVars = null;
    
    void createLocalAttribute() {
        //System.err.println("CB/Creating");
        //System.err.println(localVars);
        // find the size
        int size = 0;
        for (LocalVar in = localVars; in != null; in = in.next) {
            size++;
        }
        if (size == 0) return;
        // create the arrays
        Label[] starts = new Label[size];
        Label[] ends = new Label[size];
        Constant[] names = new Constant[size];
        Constant[] descriptors = new Constant[size];
        int[] indices = new int[size];
        // fill them
        LocalVar lvs = localVars;
        int i = 0;
        for (; lvs != null; i++, lvs = lvs.next) {
            starts[i] = lvs.start;
            ends[i] = lvs.end;
            names[i] = lvs.name;
            descriptors[i] = lvs.descriptor;
            indices[i] = lvs.index;
        }
        // and create the attribute
        attributes.addLocalVariableTableAttribute(starts, ends, names, 
                                                  descriptors, indices);
    }

    // ------------------------------
    // line number table context

    private static final class LineNumber {
        ClassfileBuilder.SourceFile fileInfo;
        Label pc;
        int line;
        LineNumber next;
        LineNumber(ClassfileBuilder.SourceFile fileInfo, Label pc, int line, LineNumber next) {
            this.fileInfo = fileInfo;
            this.pc = pc;
            this.line = line;
            this.next = next;
        }
    }

    LineNumber lineNumbers = null;
        
    public void enterLocation(SourceLocation loc) {
        //System.err.println("hiya: " + loc);
        ClassfileBuilder.SourceFile fileInfo = cfb.getSourceFile(loc);
        if (fileInfo == null) return; // why?
        int currLine = loc.getBeginLine();        
        if (lineNumbers != null) {
            if (currLine == -1) return; // hmm.  This sounds dangerous.
            if (currLine == lineNumbers.line && fileInfo == lineNumbers.fileInfo) return;
        }
        Label pc = genAnchor();
        emitLabel(pc);
        lineNumbers = new LineNumber(fileInfo, pc, currLine, lineNumbers);
    }

    void createLineAttribute() {
        // This could be done in one pass with recursion.
        // reverse, set aright, and get size.  
        int endPc = getEndPc();
        int size = 0;
        LineNumber in = lineNumbers;
        LineNumber out = null;
		// increment the line number according to the offset of its source file.
        while (in != null) {
            if (in.pc.getPc() >= endPc) {
                in = in.next;
            } else {
//				cfb.addLineNumber(in.line, in.fileInfo);
                in.line += in.fileInfo.offset;
                LineNumber temp = in.next;
                in.next = out;
                out = in;
                in = temp;
                size++;
            }
        }
        // create the arrays
        Label[] pcs = new Label[size];
        int[] lines = new int[size];
        // fill them
        for (int i = 0; i < size; i++, out = out.next) {
            pcs[i] = out.pc;
            lines[i] = out.line;
        }
        // and create the attribute
        attributes.addLineNumberTableAttribute(pcs, lines);
    }

    // ------------------------------
    // exception handlers

    public void addHandler(Label start, Label end, Label handler, NameType exn) {
        starts.add(start);    
        ends.add(end);
        handlers.add(handler); 
        names.add((exn == null) ? null : pool.addClass(exn));
    }

    void writeExceptionTableTo(DataOutputStream stream) throws IOException {
        stream.writeShort((short)starts.size());
        Iterator i = starts.iterator();
        Iterator j = ends.iterator();
        Iterator k = handlers.iterator();
        Iterator l = names.iterator();
        for (; i.hasNext(); ) {
            stream.writeShort(((Label)i.next()).getPc());
            stream.writeShort(((Label)j.next()).getPc());
            stream.writeShort(((Label)k.next()).getPc());
            Constant name = (Constant)l.next();
            if (name == null) {
                stream.writeShort(0);
            } else {
                name.writeIndex(stream);
            }
        }
    }

    // ------------------------------
    // labels

    JumpCode prevJump = null;

    public void emitLabel(Label l) {
        if (isResolved()) throw new RuntimeException("already resolved"); 
        labels.add(l);
        if (prevJump != null) {
            if (prevJump.getDest() == l) {
                labels.addAll(prevJump.getLabels());
                codes.remove(codes.size() - 1);
                if (! codes.isEmpty() 
                        && (codes.get(codes.size() - 1) instanceof JumpCode)) {
                    prevJump = (JumpCode) codes.get(codes.size() - 1);
                } else {
                    prevJump = null;
                }
            }
        }
    }

    public Label genLabel() {
        return new Label(false);
    }
    public Label genAnchor() {
        return new Label(true);
    }


    // ------------------------------
    // Codes

    private void addCode(Code c, int stackDelta) {
        addCode(c);
        adjustStack(stackDelta);
    }
    private void addCode(Code c) {
        if (isResolved()) {
            throw new RuntimeException("already resolved");
        }
        for (Iterator i = labels.iterator(); i.hasNext(); ) {
            ((Label) i.next()).setTarget(c);
        }
        c.index = codes.size();
        labels.clear();
        codes.add(c);
        prevJump = null;
    }

    // ------------------------------
    // assembly and output

    public void resolve() {
        if (isResolved()) {
            throw new RuntimeException("already resolved");
        }
        if (! pool.isResolved()) {
            throw new RuntimeException("pool not resolved yet");
        }

        addCode(new SentinelCode());
        int len = codes.size();
        Code[] a = new Code[len];
        for (int i = 0; i < len; i++) {
            a[i] = (Code) codes.get(i);
        }
        codea = a;
        codes = null; 

        keelbackwards();

        int pc = 0;
        try {
            for (int i = 0; i < len; i++) {
                if (DEBUG) System.err.print("(" + i + ") pc " + pc);
                Code c = a[i];
                pc = c.resolveBest(pc);
                if (DEBUG) System.err.println(" -> " + pc);
            }
            
            // XXX let's try writing to a dummy stream to make sure we
            // can without generating a tooWideJumpException.  This is
            // fairly inefficient; it would be better to run through
            // and just check jump/branch/etc codes. 
            DataOutputStream sink = new DataOutputStream(new OutputStream() {
                    public void write(int b) {}
                });
            try {
                for (int i = 0; i < len; i++) {
                    a[i].writeTo(sink);
                }
            } catch (IOException e) {}
        } catch (TooWideJumpException e) {
            pc = 0;
            for (int i = 0; i < len; i++) {
                Code c = a[i];
                pc = c.resolveWorst(pc);
            }
        }
        endPc = pc;
        if (DEBUG) {
            System.err.println(">>>>");
            display(0, false);
            System.err.println();
        
            for (int i = 0; i < len; i++) {
                Code c = a[i];
                for (Iterator ii = c.getLabels().iterator(); ii.hasNext(); ) {
                    if (((Label) ii.next()).getTarget() != c) {
                        System.err.println("hi");
                    }
                }
            }
        }
        isResolvedFlag = true;

        createLineAttribute();
        createLocalAttribute();
    }

    private int endPc;
    int getEndPc() {
        if (! isResolvedFlag) throw new RuntimeException("not yet resolved");
        return endPc;
    }

    static class TooWideJumpException extends RuntimeException {}

    //!!! should clean this up
    static final boolean DEBUG = false;
    private void keelbackwards() {
        if (DEBUG) {
            isResolvedFlag = true;
            System.err.println("---");
            display(0, false);
            System.err.println();
            System.err.println(">>>");
        }
        Code[] a = codea;
        // there's a sentinel at the end
        Code next;
        Code curr = a[a.length - 1];
        for (int i = a.length - 2; i >= 0; i--) {
            next = curr;
            curr = a[i];
            curr.resolve(next.getRealNext());
            if (curr.jumps()) {
                next.notifyNonLive();
            }
        }
        if (DEBUG) {
            display(0, false);
            System.err.println();
        }
    }

    int getByteSize() { 
        return 2 + 2 + getCodeByteSize() 
            + getExnTableByteSize() 
            + attributes.getByteSize(); 
    }
    private int getCodeByteSize() {
        return 4 + getEndPc();
    }
    private int getExnTableByteSize() {
        return 2 + (8 * starts.size());
    }

    void writeTo(DataOutputStream stream) throws IOException {
        if (! isResolved()) {
            throw new RuntimeException("not yet resolved");
        }
        stream.writeShort((short) maxStack);
        stream.writeShort((short) maxFrame);
        writeCodeTo(stream);
        writeExceptionTableTo(stream);
        attributes.writeTo(stream);
    }

    private void writeCodeTo(DataOutputStream stream) throws IOException {
        stream.writeInt(getEndPc());
        Code[] a = codea;
        for (int i = 0, len = a.length; i < len; i++) {
            a[i].writeTo(stream);
        }
    }

    // ------------------------------
    // building methods

    // jumps

    public void emitJump(Label dest) {
        JumpCode c = new JumpCode(dest);
        addCode(c);
        prevJump = c;
    }

    public void emitJSR(Label dest) {
        addCode(new JsrCode(dest));
    }

    public void emitRET(int n) {
        addCode(new FrameCode(RET, n), 0);
    }

    // switches

    public void emitSwitch(int[] keys, Label[] cases, Label defaultCase) {
        long min = keys[0];
        long max = keys[keys.length - 1];

        if ((1 + 1 + max - min + 1) // max + min + 1 word per (max - min + 1)
            <=
            (1 + keys.length * 2)) { // size + 2 words per entry
            Label[] realCases = new Label[(int)(max - min + 1)];
            for (int kv = (int)min, ki = 0, ci = 0, oi = 0; kv <= (int)max; kv++, oi++) {
                if (kv == keys[ki]) {
                    realCases[oi] = cases[ci++];
                    ki++; 
                } else {
                    realCases[oi] = defaultCase;
                }
            }
            addCode(new TableSwitchCode((int)min, (int)max, realCases, defaultCase), -1);

        } else {
            addCode(new LookupSwitchCode(keys, cases, defaultCase), -1);
        }
    }

    // fields and methods

    public void emitINVOKESTATIC(NameType nameType, String name,
                                 String descriptor, int stackDelta) {
        addCode(new PoolCode(INVOKESTATIC,
                             pool.addMethodRef(nameType, name, descriptor)),
                stackDelta);
    }
    public void emitINVOKESPECIAL(NameType nameType, String name,
                                 String descriptor, int stackDelta) {
        addCode(new PoolCode(INVOKESPECIAL,
                             pool.addMethodRef(nameType,
                                               name, descriptor)),
                stackDelta);
    }
    public void emitINVOKEINTERFACE(NameType nameType, String name,
                                    String descriptor, int stackDelta, int arglen) {
        addCode(new InvokeInterfaceCode(pool.addInterfaceMethodRef(nameType,
                                                                   name, 
                                                                   descriptor),
                                        arglen),
                stackDelta);
    }
    public void emitINVOKEVIRTUAL(NameType nameType, String name,
                                 String descriptor, int stackDelta) {
        addCode(new PoolCode(INVOKEVIRTUAL,
                             pool.addMethodRef(nameType,
                                               name, descriptor)),
                stackDelta);
    }
    public void emitGETSTATIC(NameType nameType, String name,
                              String descriptor, int stackDelta) {
        addCode(new PoolCode(GETSTATIC,
                             pool.addFieldRef(nameType,
                                              name, descriptor)),
                stackDelta);
    }
    public void emitPUTSTATIC(NameType nameType, String name,
                              String descriptor, int stackDelta) {
        addCode(new PoolCode(PUTSTATIC,
                             pool.addFieldRef(nameType,
                                              name, descriptor)),
                stackDelta);
    }
    public void emitGETFIELD(NameType nameType, String name,
                             String descriptor, int stackDelta) {
        addCode(new PoolCode(GETFIELD,
                             pool.addFieldRef(nameType,
                                              name, descriptor)),
                stackDelta);
    }
    public void emitPUTFIELD(NameType nameType, String name,
                             String descriptor, int stackDelta) {
        addCode(new PoolCode(PUTFIELD,
                             pool.addFieldRef(nameType,
                                              name, descriptor)),
                stackDelta);
    }

    // pools

    public void emitNEW(NameType nameType) {
        addCode(new PoolCode(NEW, pool.addClass(nameType)), 1);
    }
    public void emitCHECKCAST(RefType refType) {
        addCode(new PoolCode(CHECKCAST, pool.addClass(refType)));
    }
    public void emitINSTANCEOF(RefType refType) {
        addCode(new PoolCode(INSTANCEOF, pool.addClass(refType)), 0);
    }

    // basics
    public void emitIAND()        { addCode(new BasicCode(IAND),        -1); }
    public void emitLAND()        { addCode(new BasicCode(LAND),        -2); }
    public void emitNULL()        { addCode(new BasicCode(ACONST_NULL),  1); }
    public void emitI2L()         { addCode(new BasicCode(I2L),          1); }
    public void emitI2F()         { addCode(new BasicCode(I2F),          0); }
    public void emitI2D()         { addCode(new BasicCode(I2D),          1); }
    public void emitL2I()         { addCode(new BasicCode(L2I),         -1); }
    public void emitL2F()         { addCode(new BasicCode(L2F),         -1); }
    public void emitL2D()         { addCode(new BasicCode(L2D),          0); }
    public void emitF2I()         { addCode(new BasicCode(F2I),          0); }
    public void emitF2L()         { addCode(new BasicCode(F2L),          1); }
    public void emitF2D()         { addCode(new BasicCode(F2D),          1); }
    public void emitD2I()         { addCode(new BasicCode(D2I),         -1); }
    public void emitD2L()         { addCode(new BasicCode(D2L),          0); }
    public void emitD2F()         { addCode(new BasicCode(D2F),         -1); }
    public void emitI2B()         { addCode(new BasicCode(I2B),          0); }
    public void emitI2C()         { addCode(new BasicCode(I2C),          0); }
    public void emitI2S()         { addCode(new BasicCode(I2S),          0); }
    public void emitPOP()         { addCode(new BasicCode(POP),         -1); }
    public void emitPOP2()        { addCode(new BasicCode(POP2),        -2); }
    public void emitLCMP ()       { addCode(new BasicCode(LCMP),        -3); }
    public void emitFCMPL()       { addCode(new BasicCode(FCMPL),       -1); }
    public void emitFCMPG()       { addCode(new BasicCode(FCMPG),       -1); }
    public void emitDCMPL()       { addCode(new BasicCode(DCMPL),       -3); }
    public void emitDCMPG()       { addCode(new BasicCode(DCMPG),       -3); }
    public void emitNOP()         { addCode(new BasicCode(NOP),          0); }
    public void emitACONST_NULL() { addCode(new BasicCode(ACONST_NULL),  1); }
    public void emitIALOAD()      { addCode(new BasicCode(IALOAD),      -1); }
    public void emitLALOAD()      { addCode(new BasicCode(LALOAD),       0); }
    public void emitFALOAD()      { addCode(new BasicCode(FALOAD),      -1); }
    public void emitDALOAD()      { addCode(new BasicCode(DALOAD),       0); }
    public void emitAALOAD()      { addCode(new BasicCode(AALOAD),      -1); }
    public void emitBALOAD()      { addCode(new BasicCode(BALOAD),      -1); }
    public void emitCALOAD()      { addCode(new BasicCode(CALOAD),      -1); }
    public void emitSALOAD()      { addCode(new BasicCode(SALOAD),      -1); }
    public void emitIASTORE()     { addCode(new BasicCode(IASTORE),     -3); }
    public void emitLASTORE()     { addCode(new BasicCode(LASTORE),     -4); }
    public void emitFASTORE()     { addCode(new BasicCode(FASTORE),     -3); }
    public void emitDASTORE()     { addCode(new BasicCode(DASTORE),     -4); }
    public void emitAASTORE()     { addCode(new BasicCode(AASTORE),     -3); }
    public void emitBASTORE()     { addCode(new BasicCode(BASTORE),     -3); }
    public void emitCASTORE()     { addCode(new BasicCode(CASTORE),     -3); }
    public void emitSASTORE()     { addCode(new BasicCode(SASTORE),     -3); }
    public void emitSWAP()        { addCode(new BasicCode(SWAP),         0); }
    public void emitIADD()        { addCode(new BasicCode(IADD),        -1); }
    public void emitLADD()        { addCode(new BasicCode(LADD),        -2); }
    public void emitFADD()        { addCode(new BasicCode(FADD),        -1); }
    public void emitDADD()        { addCode(new BasicCode(DADD),        -2); }
    public void emitISUB()        { addCode(new BasicCode(ISUB),        -1); }
    public void emitLSUB()        { addCode(new BasicCode(LSUB),        -2); }
    public void emitFSUB()        { addCode(new BasicCode(FSUB),        -1); }
    public void emitDSUB()        { addCode(new BasicCode(DSUB),        -2); }
    public void emitIMUL()        { addCode(new BasicCode(IMUL),        -1); }
    public void emitLMUL()        { addCode(new BasicCode(LMUL),        -2); }
    public void emitFMUL()        { addCode(new BasicCode(FMUL),        -1); }
    public void emitDMUL()        { addCode(new BasicCode(DMUL),        -2); }
    public void emitIDIV()        { addCode(new BasicCode(IDIV),        -1); }
    public void emitLDIV()        { addCode(new BasicCode(LDIV),        -2); }
    public void emitFDIV()        { addCode(new BasicCode(FDIV),        -1); }
    public void emitDDIV()        { addCode(new BasicCode(DDIV),        -2); }
    public void emitIREM()        { addCode(new BasicCode(IREM),        -1); }
    public void emitLREM()        { addCode(new BasicCode(LREM),        -2); }
    public void emitFREM()        { addCode(new BasicCode(FREM),        -1); }
    public void emitDREM()        { addCode(new BasicCode(DREM),        -2); }
    public void emitINEG()        { addCode(new BasicCode(INEG),         0); }
    public void emitLNEG()        { addCode(new BasicCode(LNEG),         0); }
    public void emitFNEG()        { addCode(new BasicCode(FNEG),         0); }
    public void emitDNEG()        { addCode(new BasicCode(DNEG),         0); }
    public void emitISHL()        { addCode(new BasicCode(ISHL),        -1); }
    public void emitLSHL()        { addCode(new BasicCode(LSHL),        -1); }
    public void emitISHR()        { addCode(new BasicCode(ISHR),        -1); }
    public void emitLSHR()        { addCode(new BasicCode(LSHR),        -1); }
    public void emitIUSHR()       { addCode(new BasicCode(IUSHR),       -1); }
    public void emitLUSHR()       { addCode(new BasicCode(LUSHR),       -1); }
    public void emitIOR()         { addCode(new BasicCode(IOR),         -1); }
    public void emitLOR()         { addCode(new BasicCode(LOR),         -2); }
    public void emitIXOR()        { addCode(new BasicCode(IXOR),        -1); }
    public void emitLXOR()        { addCode(new BasicCode(LXOR),        -2); }
    public void emitIRETURN()     { addCode(new ReturnCode(IRETURN),     -1); }
    public void emitLRETURN()     { addCode(new ReturnCode(LRETURN),     -2); }
    public void emitFRETURN()     { addCode(new ReturnCode(FRETURN),     -1); }
    public void emitDRETURN()     { addCode(new ReturnCode(DRETURN),     -2); }
    public void emitARETURN()     { addCode(new ReturnCode(ARETURN),     -1); }
    public void emitRETURN()      { addCode(new ReturnCode(RETURN),       0); }
    public void emitARRAYLENGTH() { addCode(new BasicCode(ARRAYLENGTH),  0); }
    public void emitATHROW()      { addCode(new BasicCode(ATHROW),      -1); }
    public void emitMONITORENTER(){ addCode(new BasicCode(MONITORENTER),-1); }
    public void emitMONITOREXIT() { addCode(new BasicCode(MONITOREXIT), -1); }
    public void emitDUP()         { addCode(new BasicCode(DUP),          1); }
    public void emitDUP_X1()      { addCode(new BasicCode(DUP_X1),       1); }
    public void emitDUP_X2()      { addCode(new BasicCode(DUP_X2),       1); }
    public void emitDUP2()        { addCode(new BasicCode(DUP2),         2); }
    public void emitDUP2_X1()     { addCode(new BasicCode(DUP2_X1),      2); }
    public void emitDUP2_X2()     { addCode(new BasicCode(DUP2_X2),      2); }

    // branches

    public void emitIFEQ(Label tdest, Label fdest) {
        addCode(new BranchCode(IFEQ, IFNE, tdest, fdest), -1);
    }
    public void emitIFNE(Label tdest, Label fdest) {
        addCode(new BranchCode(IFNE, IFEQ, tdest, fdest), -1);
    }
    public void emitIFLT(Label tdest, Label fdest) {
        addCode(new BranchCode(IFLT, IFGE, tdest, fdest), -1);
    }
    public void emitIFGE(Label tdest, Label fdest) {
        addCode(new BranchCode(IFGE, IFLT, tdest, fdest), -1);
    }
    public void emitIFGT(Label tdest, Label fdest) {
        addCode(new BranchCode(IFGT, IFLE, tdest, fdest), -1);
    }
    public void emitIFLE(Label tdest, Label fdest) {
        addCode(new BranchCode(IFLE, IFGT, tdest, fdest), -1);
    }

    public void emitIF_ICMPEQ(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPEQ, IF_ICMPNE, tdest, fdest), -2);
    }
    public void emitIF_ICMPNE(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPNE, IF_ICMPEQ, tdest, fdest), -2);
    }
    public void emitIF_ICMPLT(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPLT, IF_ICMPGE, tdest, fdest), -2);
    }
    public void emitIF_ICMPGE(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPGE, IF_ICMPLT, tdest, fdest), -2);
    }
    public void emitIF_ICMPGT(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPGT, IF_ICMPLE, tdest, fdest), -2);
    }
    public void emitIF_ICMPLE(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ICMPLE, IF_ICMPGT, tdest, fdest), -2);
    }

    public void emitIF_ACMPEQ(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ACMPEQ, IF_ACMPNE, tdest, fdest), -2);
    }
    public void emitIF_ACMPNE(Label tdest, Label fdest) {
        addCode(new BranchCode(IF_ACMPNE, IF_ACMPEQ, tdest, fdest), -2);
    }
    public void emitIFNULL(Label tdest, Label fdest) {
        addCode(new BranchCode(IFNULL, IFNONNULL, tdest, fdest), -1);
    }
    public void emitIFNONNULL(Label tdest, Label fdest) {
        addCode(new BranchCode(IFNONNULL, IFNULL, tdest, fdest), -1);
    }

    // constants

    public void emitIntConstant(int val) {
        Code code;
        switch (val) {
        case -1: code = new BasicCode(ICONST_M1); break;
        case  0: code = new BasicCode(ICONST_0); break;
        case  1: code = new BasicCode(ICONST_1); break;
        case  2: code = new BasicCode(ICONST_2); break;
        case  3: code = new BasicCode(ICONST_3); break;
        case  4: code = new BasicCode(ICONST_4); break;
        case  5: code = new BasicCode(ICONST_5); break;
        default:
            if (Asserts.isByte(val)) {
                code = new ByteCode(BIPUSH, (byte)val);
            } else if (Asserts.isShort(val)) {
                code = new ShortCode(SIPUSH, (short)val);
            } else {
                code = new LDCCode(pool.addInt(val));
            }
            break;
        }
        addCode(code, 1);
    }

    public void emitLongConstant(long val) {
        Code code;
        if      (val == 0L) code = new BasicCode(LCONST_0);
        else if (val == 1L) code = new BasicCode(LCONST_1);
        else     code = new PoolCode(LDC2_W, pool.addLong(val));
        addCode(code, 2);
    }

    public void emitFloatConstant(float val) {
        Code code;
        if      (val == 0.0f) {
            if ((1 / val) > 0) {
                code = new BasicCode(FCONST_0);
            } else {
                code = new LDCCode(pool.addFloat(val));
            }
        }
        else if (val == 1.0f) code = new BasicCode(FCONST_1);
        else if (val == 2.0f) code = new BasicCode(FCONST_2);
        else       code = new LDCCode(pool.addFloat(val));
        addCode(code, 1);
    }

    public void emitDoubleConstant(double val) {
        Code code;
        if      (val == 0.0D) {
            if ((1 / val) > 0) {
                code = new BasicCode(DCONST_0);
            } else {
                code = new PoolCode(LDC2_W, pool.addDouble(val));
            }
        }
        else if (val == 1.0D) code = new BasicCode(DCONST_1);
        else   code = new PoolCode(LDC2_W, pool.addDouble(val));
        addCode(code, 2);
    }

    public void emitStringConstant(String val) {
        addCode(new LDCCode(pool.addString(val)), 1);
    }

    // arrays

    public void emitNEWARRAY(int tyIndex) {
        addCode(new ByteCode(NEWARRAY, (byte)tyIndex), 0);
    }

    public void emitANEWARRAY(RefType componentType) {
        addCode(new PoolCode(ANEWARRAY, pool.addClass(componentType)), 0);
    }

    public void emitMULTIANEWARRAY(ArrayType arrayType, int dimCount) {
        addCode(new PoolByteCode(MULTIANEWARRAY,
                                 pool.addClass(arrayType),
                                 (byte)dimCount),
                -(dimCount - 1));
    }

    // loads and stores

    private final byte[] aloads =  { ALOAD_0, ALOAD_1, ALOAD_2, ALOAD_3, ALOAD };
    private final byte[] iloads =  { ILOAD_0, ILOAD_1, ILOAD_2, ILOAD_3, ILOAD };
    private final byte[] lloads =  { LLOAD_0, LLOAD_1, LLOAD_2, LLOAD_3, LLOAD };
    private final byte[] floads =  { FLOAD_0, FLOAD_1, FLOAD_2, FLOAD_3, FLOAD };
    private final byte[] dloads =  { DLOAD_0, DLOAD_1, DLOAD_2, DLOAD_3, DLOAD };

    private final byte[] astores = { ASTORE_0, ASTORE_1, ASTORE_2, ASTORE_3, ASTORE };
    private final byte[] istores = { ISTORE_0, ISTORE_1, ISTORE_2, ISTORE_3, ISTORE };
    private final byte[] lstores = { LSTORE_0, LSTORE_1, LSTORE_2, LSTORE_3, LSTORE };
    private final byte[] fstores = { FSTORE_0, FSTORE_1, FSTORE_2, FSTORE_3, FSTORE };
    private final byte[] dstores = { DSTORE_0, DSTORE_1, DSTORE_2, DSTORE_3, DSTORE };

    public void emitALOAD(int n) { emitFrameCode(aloads, n, 1); }
    public void emitILOAD(int n) { emitFrameCode(iloads, n, 1); }
    public void emitLLOAD(int n) { emitFrameCode(lloads, n, 2); }
    public void emitFLOAD(int n) { emitFrameCode(floads, n, 1); }
    public void emitDLOAD(int n) { emitFrameCode(dloads, n, 2); }

    public void emitASTORE(int n) { emitFrameCode(astores, n, -1); }
    public void emitISTORE(int n) { emitFrameCode(istores, n, -1); }
    public void emitLSTORE(int n) { emitFrameCode(lstores, n, -2); }
    public void emitFSTORE(int n) { emitFrameCode(fstores, n, -1); }
    public void emitDSTORE(int n) { emitFrameCode(dstores, n, -2); }

    public void emitFrameCode(byte[] insts, int n, int delta) {
        if (n <= 3) addCode(new BasicCode(insts[n]), delta);
        else        addCode(new FrameCode(insts[4], n), delta);
    }

    // iinc

    public void emitIINC(int n, int incr) {
        addCode(new IncCode(n, incr), 0);
    }

    // ------------------------------
    // ------------------------------
    // debugging

    public String toString() {
        return "(st/fr " + maxStack + " " + maxFrame + ") " +
            attributes;
    }

    public void emitStackSize() {
        addCode(new BasicCode(NOP));
        addCode(new ByteCode(BIPUSH, (byte)getStackSize()));
        addCode(new BasicCode(POP));
        addCode(new BasicCode(NOP));
    }

    int labelCount = 0;
    public void displayInstructions(int indent) {
        List codes = (isResolved()) 
            ? Arrays.asList(codea) 
            : this.codes;
        for (int i = 0, len = codes.size() - 1; i < len; i++) {
            Code c = (Code) codes.get(i);
            if (! c.isActive()) continue;
            between(indent, false);
            if (DEBUG) System.err.print("(" + i + ") ");
            if (c.getLabels().size() != 0) {
                if (DEBUG) {
                    for (Iterator ii = c.getLabels().iterator(); ii.hasNext(); ) {
                        System.err.print(ii.next() + " ");
                    }
                } else {
                    String s = c.getLabels().get(0) + " ";
                    System.err.print(s);
                    for (int p = 4 - s.length(); p > 0; p--) {
                        System.err.print(" ");
                    }
                }
            } else {
                System.err.print("    ");
            }
            System.err.print(c);
        }
        for (int i = 0, len = starts.size(); i < len; i++) {
            between(indent, false);
            String name = (names.get(i) == null) 
                ? "any" 
                : names.get(i).toString();
            System.err.print("(throwing " + name 
                               + " from " + starts.get(i)
                               + " to " + ends.get(i)
                               + " goto " + handlers.get(i) + ")");
        }
    }

    public void display(int indent, boolean inline) {
        System.err.print("(stack " + maxStack + ") (locals " + maxFrame + ")");
        between(indent, inline);
        attributes.display(indent, inline);
        if (! inline) {
            displayInstructions(indent);
        }
    }

    static void between(int indent, boolean inline) {
        if (inline) {
            System.err.print(" ");
        } else {
            System.err.println();
            for (int s=indent; s >= 0; s--) System.err.print(" ");
        }
    }

    // ------------------------------
    // ------------------------------
    // classes

    static class SentinelCode extends Code {
        int resolveBest(int pc) { setPc(pc); return pc; }
        void writeTo(DataOutputStream stream) {}
        public String toString() { return "SENTINEL"; }
        boolean isActive() { return false; }
    }

    static class BranchCode extends Code {
        byte tcode, fcode;
        Label tdest, fdest;
        BranchCode(byte tcode, byte fcode, Label tdest, Label fdest) {
            this.tcode = tcode; this.tdest = tdest;
            this.fcode = fcode; this.fdest = fdest;
        }

        boolean oneArmed;
        boolean arm;
        void resolve(Code next) {
            if (tdest.getTarget() == next) {
                oneArmed = true;
                arm = false;
            } else if (fdest.getTarget() == next) {
                oneArmed = true;
                arm = true;
            } else {
                oneArmed = false;
            }
        }

        boolean jumps() { return true; }

        boolean small = true;
        int resolveBest(int pc) { this.pc = pc; return pc + (oneArmed ? 3 : 6); }
        int resolveWorst(int pc) { small = false; this.pc = pc; return pc + (oneArmed ? 8 : 13); }

        void writeTo(DataOutputStream stream) throws IOException {
            if (small) {
                if (! oneArmed) {
                    stream.writeByte(tcode);
                    tdest.writeRelativeShort(pc, stream);
                    stream.writeByte(GOTO);
                    fdest.writeRelativeShort(pc + 3, stream);
                } else if (arm) {
                    // branch on true code, fallthrough on false code
                    stream.writeByte(tcode);
                    tdest.writeRelativeShort(pc, stream);
                } else {
                    stream.writeByte(fcode);
                    fdest.writeRelativeShort(pc, stream);
                }
            } else {
                if (! oneArmed) {
                    stream.writeByte(tcode);
                    stream.writeShort(8);
                    stream.writeByte(GOTO_W);
                    fdest.writeRelativeInt(pc + 3, stream);
                    stream.writeByte(GOTO_W);
                    tdest.writeRelativeInt(pc + 8, stream);
                } else if (arm) {
                    // branch on true code, fallthrough on false code
                    stream.writeByte(fcode);
                    stream.writeShort(8);
                    stream.writeByte(GOTO_W);
                    tdest.writeRelativeInt(pc + 3, stream);
                } else {
                    stream.writeByte(tcode);
                    stream.writeShort(8);
                    stream.writeByte(GOTO_W);
                    fdest.writeRelativeInt(pc + 3, stream);
                }
            }
        }

        public String toString() {
            if (! oneArmed) {
                return debugString(tcode) + " " + tdest + " else " + fdest;
            } else if (arm) {
                return debugString(tcode) + " " + tdest + " ";
            } else {
                return debugString(fcode) + " " + fdest + " ";
            }
        }
    }

    static class JumpCode extends Code {
        boolean active = true;
        boolean isActive() { return active; }
        Label dest;
        JumpCode(Label dest) { this.dest = dest; }
        Label getDest() { return dest; }
        void setDest(Label dest) { this.dest = dest; }

        boolean small = true;
        int resolveBest(int pc) { this.pc = pc; return (active) ? (pc + 3) : pc; }
        int resolveWorst(int pc) { 
            small = false; this.pc = pc; 
            return (active) ? (pc + 5) : pc; 
        }
        void writeTo(DataOutputStream stream) throws IOException {
            if (! active) return;
            if (small) {
                stream.writeByte(GOTO);
                dest.writeRelativeShort(pc, stream);
            } else {
                stream.writeByte(GOTO_W);
                dest.writeRelativeInt(pc, stream);
            }
        }
        public String toString() {
            return "GOTO " + dest;
        }
        boolean jumps() { return active; }

        void notifyNonLive() {
            active = false;
        }
        Code getRealNext() {
            return (active) ? this : dest.getTarget();
        }
        void resolve(Code next) {
            Code target = dest.getTarget();
            if (target != this) {
                for (Iterator i = getLabels().iterator(); i.hasNext(); ) {
                    Label l = (Label) i.next();
                    if (! l.isAnchored()) {
                        i.remove();
                        l.setTarget(target);
                    }
                }
            } 
            if (next == target) {
                active = false;
            }
        } 
    }

    static class JsrCode extends Code {
        Label dest;
        JsrCode(Label dest) { this.dest = dest; }
        Label getDest() { return dest; }

        int resolveBest(int pc) { this.pc = pc; return pc + 3; }
        int resolveWorst(int pc) { small = false; this.pc = pc; return pc + 5; }

        boolean small = true;
        void writeTo(DataOutputStream stream) throws IOException {
            if (small) {
                stream.writeByte(JSR);
                dest.writeRelativeShort(pc, stream);
            } else {
                stream.writeByte(JSR_W);
                dest.writeRelativeInt(pc, stream);
            }
        }

        public String toString() {
            return "JSR " + dest;
        }
    }

    static class IncCode extends Code {
        int loc, val;
        int size;
        IncCode(int loc, int val) { 
            this.loc = loc; this.val = val;
            size = (Asserts.isUByte(loc) && Asserts.isByte(val)) ? 3 : 6;
        }
        int resolveBest(int pc) { this.pc = pc; return pc + size; }

        void writeTo(DataOutputStream stream) throws IOException {
            if (size == 6) {
                stream.writeByte(WIDE);
                stream.writeByte(IINC);
                stream.writeShort(loc);
                stream.writeShort(val);
            } else {
                stream.writeByte(IINC);
                stream.writeByte(loc);
                stream.writeByte(val);

            }
        }
        public String toString() {
            return ((size == 6) ? "WIDE IINC " : "IINC ") 
                + loc + " " + val;
        }
    }

    static class FrameCode extends Code {
        byte code;
        int loc;
        int size;
        FrameCode(byte code, int loc) { 
            this.code = code; this.loc = loc; 
            size = Asserts.isUByte(loc) ? 2 : 4;
        }
        int resolveBest(int pc) { this.pc = pc; return pc + size; }

        void writeTo(DataOutputStream stream) throws IOException {
            if (size == 4) {
                stream.writeByte(WIDE);
                stream.writeByte(code);
                stream.writeShort(loc);
            } else {
                stream.writeByte(code);
                stream.writeByte(loc);
            }
        }
        public String toString() {
            return ((size == 4) ? "WIDE " : "") 
                + debugString(code) + " " + loc;
        }
    }

    static class BasicCode extends Code {
        byte code;
        BasicCode(byte code) { this.code = code; }

        int resolveBest(int pc) { this.pc = pc; return pc + 1; }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(code);
        }
        public String toString() {
            return debugString(code);
        }
    }

    static class ReturnCode extends BasicCode {
        ReturnCode(byte code) { super(code); }
        boolean jumps() { return true; }
    }

    static class ByteCode extends Code {
        byte code;
        byte data;
        ByteCode(byte code, byte data) { 
            this.code = code; this.data = data; 
        }

        int resolveBest(int pc) { this.pc = pc; return pc + 2; }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(code);
            stream.writeByte(data);
        }
        public String toString() {
            return debugString(code) + " " + data;
        }
    }

    static class ShortCode extends Code {
        byte code;
        short data;
        ShortCode(byte code, short data) { 
            this.code = code; this.data = data; 
        }

        int resolveBest(int pc) { this.pc = pc; return pc + 3; }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(code);
            stream.writeShort(data);
        }
        public String toString() {
            return debugString(code) + " " + data;
        }
    }

    static class PoolCode extends Code {
        byte code;
        Constant constant;
        PoolCode(byte code, Constant constant) {
            this.code = code; this.constant = constant;
        }
        int resolveBest(int pc) {
            this.pc = pc;
            return pc + 3;
        }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(code);
            constant.writeIndex(stream);
        }
        public String toString() {
            return debugString(code) + " " + constant;
        }
    }

    static class InvokeInterfaceCode extends Code {
        Constant constant;
        int arglen;
        InvokeInterfaceCode(Constant constant, int arglen) {
            this.constant = constant;
            this.arglen = arglen;
        }
        int resolveBest(int pc) {
            this.pc = pc;
            return pc + 5;
        }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(INVOKEINTERFACE);
            constant.writeIndex(stream);
            stream.writeByte((byte)arglen);
            stream.writeByte(0);
        }
        public String toString() {
            return "INVOKEINTERFACE (args " + arglen + ") " + constant;
        }
    }

    static class LDCCode extends Code {
        Constant constant;
        LDCCode(Constant constant) {
            this.constant = constant;
            constant.makePopular(); // popular constants are more likely to be LDCs rather than LDC_Ws
        }
        int resolveBest(int pc) {
            this.pc = pc; return pc + (Asserts.isUByte(constant.getIndex()) ? 2 : 3);
        }
        void writeTo(DataOutputStream stream) throws IOException {
            if (Asserts.isUByte(constant.getIndex())) {
                stream.writeByte(LDC);
                constant.writeBriefIndex(stream);
            } else {
                stream.writeByte(LDC_W);
                constant.writeIndex(stream);
            }
        }
        public String toString() {
            return "LDC " + constant + (Asserts.isUByte(constant.getIndex()) ? "" : " (!!WIDE!!)");
        }
    }

    static class PoolByteCode extends Code {  // multiNewArray
        byte code;
        Constant constant;
        byte data;
        PoolByteCode(byte code, Constant constant, byte data) {
            this.code = code; this.constant = constant; this.data = data;
        }
        int resolveBest(int pc) { this.pc = pc; return pc + 4; }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(code);
            constant.writeIndex(stream);
            stream.writeByte(data);
        }
        public String toString() {
            return debugString(code) + " " + constant + " " + data;
        }
    }

    public static class TableSwitchCode extends Code {
        int padBytes = 0;
        int min;
        int max;
        Label[] cases;
        Label defaultCase;
        TableSwitchCode(int min, int max, Label[] cases, Label defaultCase) {
            this.min = min; this.max = max;
            this.cases = cases; this.defaultCase = defaultCase;
        }
        int resolveBest(int pc) {
            this.pc = pc;
            padBytes = (-pc) % 4 + 3;
            return pc + 1 + padBytes + 4 + 4 + 4 + (4 * cases.length);
        }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(TABLESWITCH);
            for (int i = padBytes; i > 0; i--) {
                stream.writeByte(0);
            }
            defaultCase.writeRelativeInt(pc, stream);
            stream.writeInt(min);
            stream.writeInt(max);
            for (int i = 0; i < cases.length; i++) {
                cases[i].writeRelativeInt(pc, stream);
            }
        }
        public String toString() {
            return "tableswitch";
        }

        boolean jumps() { return true; }
    }

    public static class LookupSwitchCode extends Code {
        int padBytes = 0;
        int[] keys;
        Label[] cases;
        Label defaultCase;
        LookupSwitchCode(int[] keys, Label[] cases, Label defaultCase) {
            this.keys = keys; this.cases = cases;
            this.defaultCase = defaultCase;
        }
        int resolveBest(int pc) {
            this.pc = pc;
            padBytes = (-pc) % 4 + 3;
            return pc + 1 + padBytes + 4 + 4 + (8 * cases.length);
        }
        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(LOOKUPSWITCH);
            for (int i = padBytes; i > 0; i--) {
                stream.writeByte(0);
            }
            defaultCase.writeRelativeInt(pc, stream);
            stream.writeInt(cases.length);
            for (int i = 0; i < cases.length; i++) {
                stream.writeInt(keys[i]);
                cases[i].writeRelativeInt(pc, stream);
            }
        }
        public String toString() {
            return "lookupswitch";
        }
        boolean jumps() { return true; }
    }

    // ------------------------------
    // constants

    // BasicCode
    protected static final byte NOP          = (byte) 0x00;
    protected static final byte ACONST_NULL  = (byte) 0x01;
    protected static final byte ICONST_M1    = (byte) 0x02;
    protected static final byte ICONST_0     = (byte) 0x03;
    protected static final byte ICONST_1     = (byte) 0x04;
    protected static final byte ICONST_2     = (byte) 0x05;
    protected static final byte ICONST_3     = (byte) 0x06;
    protected static final byte ICONST_4     = (byte) 0x07;
    protected static final byte ICONST_5     = (byte) 0x08;
    protected static final byte LCONST_0     = (byte) 0x09;
    protected static final byte LCONST_1     = (byte) 0x0a;
    protected static final byte FCONST_0     = (byte) 0x0b;
    protected static final byte FCONST_1     = (byte) 0x0c;
    protected static final byte FCONST_2     = (byte) 0x0d;
    protected static final byte DCONST_0     = (byte) 0x0e;
    protected static final byte DCONST_1     = (byte) 0x0f;
    protected static final byte ILOAD_0      = (byte) 0x1a;
    protected static final byte ILOAD_1      = (byte) 0x1b;
    protected static final byte ILOAD_2      = (byte) 0x1c;
    protected static final byte ILOAD_3      = (byte) 0x1d;
    protected static final byte LLOAD_0      = (byte) 0x1e;
    protected static final byte LLOAD_1      = (byte) 0x1f;
    protected static final byte LLOAD_2      = (byte) 0x20;
    protected static final byte LLOAD_3      = (byte) 0x21;
    protected static final byte FLOAD_0      = (byte) 0x22;
    protected static final byte FLOAD_1      = (byte) 0x23;
    protected static final byte FLOAD_2      = (byte) 0x24;
    protected static final byte FLOAD_3      = (byte) 0x25;
    protected static final byte DLOAD_0      = (byte) 0x26;
    protected static final byte DLOAD_1      = (byte) 0x27;
    protected static final byte DLOAD_2      = (byte) 0x28;
    protected static final byte DLOAD_3      = (byte) 0x29;
    protected static final byte ALOAD_0      = (byte) 0x2a;
    protected static final byte ALOAD_1      = (byte) 0x2b;
    protected static final byte ALOAD_2      = (byte) 0x2c;
    protected static final byte ALOAD_3      = (byte) 0x2d;
    protected static final byte IALOAD       = (byte) 0x2e;
    protected static final byte LALOAD       = (byte) 0x2f;
    protected static final byte FALOAD       = (byte) 0x30;
    protected static final byte DALOAD       = (byte) 0x31;
    protected static final byte AALOAD       = (byte) 0x32;
    protected static final byte BALOAD       = (byte) 0x33;
    protected static final byte CALOAD       = (byte) 0x34;
    protected static final byte SALOAD       = (byte) 0x35;
    protected static final byte ISTORE_0     = (byte) 0x3b;
    protected static final byte ISTORE_1     = (byte) 0x3c;
    protected static final byte ISTORE_2     = (byte) 0x3d;
    protected static final byte ISTORE_3     = (byte) 0x3e;
    protected static final byte LSTORE_0     = (byte) 0x3f;
    protected static final byte LSTORE_1     = (byte) 0x40;
    protected static final byte LSTORE_2     = (byte) 0x41;
    protected static final byte LSTORE_3     = (byte) 0x42;
    protected static final byte FSTORE_0     = (byte) 0x43;
    protected static final byte FSTORE_1     = (byte) 0x44;
    protected static final byte FSTORE_2     = (byte) 0x45;
    protected static final byte FSTORE_3     = (byte) 0x46;
    protected static final byte DSTORE_0     = (byte) 0x47;
    protected static final byte DSTORE_1     = (byte) 0x48;
    protected static final byte DSTORE_2     = (byte) 0x49;
    protected static final byte DSTORE_3     = (byte) 0x4a;
    protected static final byte ASTORE_0     = (byte) 0x4b;
    protected static final byte ASTORE_1     = (byte) 0x4c;
    protected static final byte ASTORE_2     = (byte) 0x4d;
    protected static final byte ASTORE_3     = (byte) 0x4e;
    protected static final byte IASTORE      = (byte) 0x4f;
    protected static final byte LASTORE      = (byte) 0x50;
    protected static final byte FASTORE      = (byte) 0x51;
    protected static final byte DASTORE      = (byte) 0x52;
    protected static final byte AASTORE      = (byte) 0x53;
    protected static final byte BASTORE      = (byte) 0x54;
    protected static final byte CASTORE      = (byte) 0x55;
    protected static final byte SASTORE      = (byte) 0x56;
    protected static final byte POP          = (byte) 0x57;
    protected static final byte POP2         = (byte) 0x58;
    protected static final byte DUP          = (byte) 0x59;
    protected static final byte DUP_X1       = (byte) 0x5a;
    protected static final byte DUP_X2       = (byte) 0x5b;
    protected static final byte DUP2         = (byte) 0x5c;
    protected static final byte DUP2_X1      = (byte) 0x5d;
    protected static final byte DUP2_X2      = (byte) 0x5e;
    protected static final byte SWAP         = (byte) 0x5f;
    protected static final byte IADD         = (byte) 0x60;
    protected static final byte LADD         = (byte) 0x61;
    protected static final byte FADD         = (byte) 0x62;
    protected static final byte DADD         = (byte) 0x63;
    protected static final byte ISUB         = (byte) 0x64;
    protected static final byte LSUB         = (byte) 0x65;
    protected static final byte FSUB         = (byte) 0x66;
    protected static final byte DSUB         = (byte) 0x67;
    protected static final byte IMUL         = (byte) 0x68;
    protected static final byte LMUL         = (byte) 0x69;
    protected static final byte FMUL         = (byte) 0x6a;
    protected static final byte DMUL         = (byte) 0x6b;
    protected static final byte IDIV         = (byte) 0x6c;
    protected static final byte LDIV         = (byte) 0x6d;
    protected static final byte FDIV         = (byte) 0x6e;
    protected static final byte DDIV         = (byte) 0x6f;
    protected static final byte IREM         = (byte) 0x70;
    protected static final byte LREM         = (byte) 0x71;
    protected static final byte FREM         = (byte) 0x72;
    protected static final byte DREM         = (byte) 0x73;
    protected static final byte INEG         = (byte) 0x74;
    protected static final byte LNEG         = (byte) 0x75;
    protected static final byte FNEG         = (byte) 0x76;
    protected static final byte DNEG         = (byte) 0x77;
    protected static final byte ISHL         = (byte) 0x78;
    protected static final byte LSHL         = (byte) 0x79;
    protected static final byte ISHR         = (byte) 0x7a;
    protected static final byte LSHR         = (byte) 0x7b;
    protected static final byte IUSHR        = (byte) 0x7c;
    protected static final byte LUSHR        = (byte) 0x7d;
    protected static final byte IAND         = (byte) 0x7e;
    protected static final byte LAND         = (byte) 0x7f;
    protected static final byte IOR          = (byte) 0x80;
    protected static final byte LOR          = (byte) 0x81;
    protected static final byte IXOR         = (byte) 0x82;
    protected static final byte LXOR         = (byte) 0x83;
    protected static final byte I2L          = (byte) 0x85;
    protected static final byte I2F          = (byte) 0x86;
    protected static final byte I2D          = (byte) 0x87;
    protected static final byte L2I          = (byte) 0x88;
    protected static final byte L2F          = (byte) 0x89;
    protected static final byte L2D          = (byte) 0x8a;
    protected static final byte F2I          = (byte) 0x8b;
    protected static final byte F2L          = (byte) 0x8c;
    protected static final byte F2D          = (byte) 0x8d;
    protected static final byte D2I          = (byte) 0x8e;
    protected static final byte D2L          = (byte) 0x8f;
    protected static final byte D2F          = (byte) 0x90;
    protected static final byte I2B          = (byte) 0x91;
    protected static final byte I2C          = (byte) 0x92;
    protected static final byte I2S          = (byte) 0x93;
    protected static final byte LCMP         = (byte) 0x94;
    protected static final byte FCMPL        = (byte) 0x95;
    protected static final byte FCMPG        = (byte) 0x96;
    protected static final byte DCMPL        = (byte) 0x97;
    protected static final byte DCMPG        = (byte) 0x98;
    protected static final byte IRETURN      = (byte) 0xac;
    protected static final byte LRETURN      = (byte) 0xad;
    protected static final byte FRETURN      = (byte) 0xae;
    protected static final byte DRETURN      = (byte) 0xaf;
    protected static final byte ARETURN      = (byte) 0xb0;
    protected static final byte RETURN       = (byte) 0xb1;
    protected static final byte ARRAYLENGTH  = (byte) 0xbe;
    protected static final byte ATHROW       = (byte) 0xbf;
    protected static final byte MONITORENTER = (byte) 0xc2;
    protected static final byte MONITOREXIT  = (byte) 0xc3;
    // ByteCode
    protected static final byte BIPUSH       = (byte) 0x10;
    protected static final byte NEWARRAY     = (byte) 0xbc;
    // ByteCode and FrameCode
    protected static final byte ILOAD        = (byte) 0x15;
    protected static final byte LLOAD        = (byte) 0x16;
    protected static final byte FLOAD        = (byte) 0x17;
    protected static final byte DLOAD        = (byte) 0x18;
    protected static final byte ALOAD        = (byte) 0x19;
    protected static final byte ISTORE       = (byte) 0x36;
    protected static final byte LSTORE       = (byte) 0x37;
    protected static final byte FSTORE       = (byte) 0x38;
    protected static final byte DSTORE       = (byte) 0x39;
    protected static final byte ASTORE       = (byte) 0x3a;
    protected static final byte RET          = (byte) 0xa9;
    // ShortCode
    protected static final byte SIPUSH       = (byte) 0x11;
    // PoolCode
    protected static final byte LDC          = (byte) 0x12;
    protected static final byte LDC_W        = (byte) 0x13;
    protected static final byte LDC2_W       = (byte) 0x14;
    protected static final byte GETSTATIC    = (byte) 0xb2;
    protected static final byte PUTSTATIC    = (byte) 0xb3;
    protected static final byte GETFIELD     = (byte) 0xb4;
    protected static final byte PUTFIELD     = (byte) 0xb5;
    protected static final byte INVOKEVIRTUAL   = (byte) 0xb6;
    protected static final byte INVOKESPECIAL   = (byte) 0xb7;
    protected static final byte INVOKESTATIC    = (byte) 0xb8;
    protected static final byte INVOKEINTERFACE = (byte) 0xb9;
    protected static final byte ANEWARRAY    = (byte) 0xbd;
    protected static final byte NEW          = (byte) 0xbb;
    protected static final byte CHECKCAST    = (byte) 0xc0;
    protected static final byte INSTANCEOF   = (byte) 0xc1;
    // BytePoolCode
    protected static final byte MULTIANEWARRAY  = (byte) 0xc5;
    // BranchCode
    protected static final byte IFEQ         = (byte) 0x99;
    protected static final byte IFNE         = (byte) 0x9a;
    protected static final byte IFLT         = (byte) 0x9b;
    protected static final byte IFGE         = (byte) 0x9c;
    protected static final byte IFGT         = (byte) 0x9d;
    protected static final byte IFLE         = (byte) 0x9e;
    protected static final byte IF_ICMPEQ    = (byte) 0x9f;
    protected static final byte IF_ICMPNE    = (byte) 0xa0;
    protected static final byte IF_ICMPLT    = (byte) 0xa1;
    protected static final byte IF_ICMPGE    = (byte) 0xa2;
    protected static final byte IF_ICMPGT    = (byte) 0xa3;
    protected static final byte IF_ICMPLE    = (byte) 0xa4;
    protected static final byte IF_ACMPEQ    = (byte) 0xa5;
    protected static final byte IF_ACMPNE    = (byte) 0xa6;
    protected static final byte IFNULL       = (byte) 0xc6;
    protected static final byte IFNONNULL    = (byte) 0xc7;
    // JumpCode and JsrCode
    protected static final byte GOTO         = (byte) 0xa7;
    protected static final byte JSR          = (byte) 0xa8;
    protected static final byte GOTO_W       = (byte) 0xc8;
    protected static final byte JSR_W        = (byte) 0xc9;
    // SwitchCode
    protected static final byte TABLESWITCH  = (byte) 0xaa;
    protected static final byte LOOKUPSWITCH = (byte) 0xab;
    // IncCode
    protected static final byte IINC         = (byte) 0x84;
    // unused
    protected static final byte WIDE         = (byte) 0xc4;
    protected static final byte XXXUNUSEDXXX1   = (byte) 0xba;
    protected static final byte BREAKPOINT   = (byte) 0xca;
    protected static final byte IMPDEP1      = (byte) 0xfe;
    protected static final byte IMPDEP2      = (byte) 0xff;

    // ------------------------------
    // debug

    // if I cared about efficiency of printing I'd turn this into an
    // array, I suppose.
    private static String debugString(byte b) {
        switch (b) {
        case NOP: return "NOP";
        case ACONST_NULL: return "ACONST_NULL";
        case ICONST_M1: return "ICONST_M1";
        case ICONST_0: return "ICONST_0";
        case ICONST_1: return "ICONST_1";
        case ICONST_2: return "ICONST_2";
        case ICONST_3: return "ICONST_3";
        case ICONST_4: return "ICONST_4";
        case ICONST_5: return "ICONST_5";
        case LCONST_0: return "LCONST_0";
        case LCONST_1: return "LCONST_1";
        case FCONST_0: return "FCONST_0";
        case FCONST_1: return "FCONST_1";
        case FCONST_2: return "FCONST_2";
        case DCONST_0: return "DCONST_0";
        case DCONST_1: return "DCONST_1";
        case ILOAD_0: return "ILOAD_0";
        case ILOAD_1: return "ILOAD_1";
        case ILOAD_2: return "ILOAD_2";
        case ILOAD_3: return "ILOAD_3";
        case LLOAD_0: return "LLOAD_0";
        case LLOAD_1: return "LLOAD_1";
        case LLOAD_2: return "LLOAD_2";
        case LLOAD_3: return "LLOAD_3";
        case FLOAD_0: return "FLOAD_0";
        case FLOAD_1: return "FLOAD_1";
        case FLOAD_2: return "FLOAD_2";
        case FLOAD_3: return "FLOAD_3";
        case DLOAD_0: return "DLOAD_0";
        case DLOAD_1: return "DLOAD_1";
        case DLOAD_2: return "DLOAD_2";
        case DLOAD_3: return "DLOAD_3";
        case ALOAD_0: return "ALOAD_0";
        case ALOAD_1: return "ALOAD_1";
        case ALOAD_2: return "ALOAD_2";
        case ALOAD_3: return "ALOAD_3";
        case IALOAD: return "IALOAD";
        case LALOAD: return "LALOAD";
        case FALOAD: return "FALOAD";
        case DALOAD: return "DALOAD";
        case AALOAD: return "AALOAD";
        case BALOAD: return "BALOAD";
        case CALOAD: return "CALOAD";
        case SALOAD: return "SALOAD";
        case ISTORE_0: return "ISTORE_0";
        case ISTORE_1: return "ISTORE_1";
        case ISTORE_2: return "ISTORE_2";
        case ISTORE_3: return "ISTORE_3";
        case LSTORE_0: return "LSTORE_0";
        case LSTORE_1: return "LSTORE_1";
        case LSTORE_2: return "LSTORE_2";
        case LSTORE_3: return "LSTORE_3";
        case FSTORE_0: return "FSTORE_0";
        case FSTORE_1: return "FSTORE_1";
        case FSTORE_2: return "FSTORE_2";
        case FSTORE_3: return "FSTORE_3";
        case DSTORE_0: return "DSTORE_0";
        case DSTORE_1: return "DSTORE_1";
        case DSTORE_2: return "DSTORE_2";
        case DSTORE_3: return "DSTORE_3";
        case ASTORE_0: return "ASTORE_0";
        case ASTORE_1: return "ASTORE_1";
        case ASTORE_2: return "ASTORE_2";
        case ASTORE_3: return "ASTORE_3";
        case IASTORE: return "IASTORE";
        case LASTORE: return "LASTORE";
        case FASTORE: return "FASTORE";
        case DASTORE: return "DASTORE";
        case AASTORE: return "AASTORE";
        case BASTORE: return "BASTORE";
        case CASTORE: return "CASTORE";
        case SASTORE: return "SASTORE";
        case POP: return "POP";
        case POP2: return "POP2";
        case DUP: return "DUP";
        case DUP_X1: return "DUP_X1";
        case DUP_X2: return "DUP_X2";
        case DUP2: return "DUP2";
        case DUP2_X1: return "DUP2_X1";
        case DUP2_X2: return "DUP2_X2";
        case SWAP: return "SWAP";
        case IADD: return "IADD";
        case LADD: return "LADD";
        case FADD: return "FADD";
        case DADD: return "DADD";
        case ISUB: return "ISUB";
        case LSUB: return "LSUB";
        case FSUB: return "FSUB";
        case DSUB: return "DSUB";
        case IMUL: return "IMUL";
        case LMUL: return "LMUL";
        case FMUL: return "FMUL";
        case DMUL: return "DMUL";
        case IDIV: return "IDIV";
        case LDIV: return "LDIV";
        case FDIV: return "FDIV";
        case DDIV: return "DDIV";
        case IREM: return "IREM";
        case LREM: return "LREM";
        case FREM: return "FREM";
        case DREM: return "DREM";
        case INEG: return "INEG";
        case LNEG: return "LNEG";
        case FNEG: return "FNEG";
        case DNEG: return "DNEG";
        case ISHL: return "ISHL";
        case LSHL: return "LSHL";
        case ISHR: return "ISHR";
        case LSHR: return "LSHR";
        case IUSHR: return "IUSHR";
        case LUSHR: return "LUSHR";
        case IAND: return "IAND";
        case LAND: return "LAND";
        case IOR: return "IOR";
        case LOR: return "LOR";
        case IXOR: return "IXOR";
        case LXOR: return "LXOR";
        case I2L: return "I2L";
        case I2F: return "I2F";
        case I2D: return "I2D";
        case L2I: return "L2I";
        case L2F: return "L2F";
        case L2D: return "L2D";
        case F2I: return "F2I";
        case F2L: return "F2L";
        case F2D: return "F2D";
        case D2I: return "D2I";
        case D2L: return "D2L";
        case D2F: return "D2F";
        case I2B: return "I2B";
        case I2C: return "I2C";
        case I2S: return "I2S";
        case LCMP: return "LCMP";
        case FCMPL: return "FCMPL";
        case FCMPG: return "FCMPG";
        case DCMPL: return "DCMPL";
        case DCMPG: return "DCMPG";
        case IRETURN: return "IRETURN";
        case LRETURN: return "LRETURN";
        case FRETURN: return "FRETURN";
        case DRETURN: return "DRETURN";
        case ARETURN: return "ARETURN";
        case RETURN: return "RETURN";
        case ARRAYLENGTH: return "ARRAYLENGTH";
        case ATHROW: return "ATHROW";
        case MONITORENTER: return "MONITORENTER";
        case MONITOREXIT: return "MONITOREXIT";
        case BIPUSH: return "BIPUSH";
        case NEWARRAY: return "NEWARRAY";
        case ILOAD: return "ILOAD";
        case LLOAD: return "LLOAD";
        case FLOAD: return "FLOAD";
        case DLOAD: return "DLOAD";
        case ALOAD: return "ALOAD";
        case ISTORE: return "ISTORE";
        case LSTORE: return "LSTORE";
        case FSTORE: return "FSTORE";
        case DSTORE: return "DSTORE";
        case ASTORE: return "ASTORE";
        case RET: return "RET";
        case SIPUSH: return "SIPUSH";
        case LDC: return "LDC";
        case LDC_W: return "LDC_W";
        case LDC2_W: return "LDC2_W";
        case GETSTATIC: return "GETSTATIC";
        case PUTSTATIC: return "PUTSTATIC";
        case GETFIELD: return "GETFIELD";
        case PUTFIELD: return "PUTFIELD";
        case INVOKEVIRTUAL: return "INVOKEVIRTUAL";
        case INVOKESPECIAL: return "INVOKESPECIAL";
        case INVOKESTATIC: return "INVOKESTATIC";
        case INVOKEINTERFACE: return "INVOKEINTERFACE";
        case ANEWARRAY: return "ANEWARRAY";
        case NEW: return "NEW";
        case CHECKCAST: return "CHECKCAST";
        case INSTANCEOF: return "INSTANCEOF";
        case MULTIANEWARRAY: return "MULTIANEWARRAY";
        case IFEQ: return "IFEQ";
        case IFNE: return "IFNE";
        case IFLT: return "IFLT";
        case IFGE: return "IFGE";
        case IFGT: return "IFGT";
        case IFLE: return "IFLE";
        case IF_ICMPEQ: return "IF_ICMPEQ";
        case IF_ICMPNE: return "IF_ICMPNE";
        case IF_ICMPLT: return "IF_ICMPLT";
        case IF_ICMPGE: return "IF_ICMPGE";
        case IF_ICMPGT: return "IF_ICMPGT";
        case IF_ICMPLE: return "IF_ICMPLE";
        case IF_ACMPEQ: return "IF_ACMPEQ";
        case IF_ACMPNE: return "IF_ACMPNE";
        case IFNULL: return "IFNULL";
        case IFNONNULL: return "IFNONNULL";
        case GOTO: return "GOTO";
        case JSR: return "JSR";
        case GOTO_W: return "GOTO_W";
        case JSR_W: return "JSR_W";
        case TABLESWITCH: return "TABLESWITCH";
        case LOOKUPSWITCH: return "LOOKUPSWITCH";
        case IINC: return "IINC";
        case WIDE: return "WIDE";
        case XXXUNUSEDXXX1: return "XXXUNUSEDXXX1";
        case BREAKPOINT: return "BREAKPOINT";
        case IMPDEP1: return "IMPDEP1";
        case IMPDEP2: return "IMPDEP2";
        default: throw new RuntimeException("weird bytecode");
        }
    }
}
