/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.base.bytecode;

import java.io.*;
import java.lang.reflect.Modifier;

class ConstantPool {
    Entry[] constants;

    public Entry getEntry(int index) {
        return constants[index];
    }

    public Class getClass(int index) {
        return (Class)constants[index];
    }

    public UTF getUTF(int index) {
        //System.err.println("getUTF: "+index);
        if (index > constants.length) {
            throw new InvalidClassFileException("invalid constant: "+index);
        }
        return (UTF)constants[index];
    }

    private UTF readUTF(DataInputStream stream) throws IOException {
        int index = stream.readUnsignedShort();
        //System.err.println("readUTF: "+index+", "+constants[index+1]+", "+constants[index-1]);
        Entry entry = constants[index];
        if (entry == null) {
            entry = new UTF();
            constants[index] = entry;
        } else {
            // check that entry is a UTF
        }
        return (UTF)entry;
    }

    private Class readClass(DataInputStream stream) throws IOException {
        int index = stream.readUnsignedShort();
        //System.err.println("clas: "+index);

        Entry entry = constants[index];
        if (entry == null) {
            entry = new Class();
            constants[index] = entry;
        } else {
            // check that entry is a Class
        }
        return (Class)entry;
    }

    private NameAndType readNameAndType(DataInputStream stream) throws IOException {
        int index = stream.readUnsignedShort();
        //System.err.println("nat: "+index);    
        Entry entry = constants[index];
        if (entry == null) {
            entry = new NameAndType();
            constants[index] = entry;
        } else {
            // check that entry is a NameAndType
        }
        return (NameAndType)entry;
    }

    public Entry readEntry(int tag, DataInputStream stream) throws IOException {
        Entry entry=null;
        switch (tag) {
        case CLASS_TAG:
            entry = new Class(); break;
        case FIELD_REF_TAG:
            entry = new FieldRef(); break;
        case METHOD_REF_TAG:
            entry = new MethodRef(); break;
        case INTERFACE_METHOD_REF_TAG:
            entry = new InterfaceMethodRef(); break;
        case STRING_TAG:
            entry = new String(); break;
        case INTEGER_TAG:
            entry = new Integer(); break;
        case FLOAT_TAG:
            entry = new Float(); break;
        case LONG_TAG:
            entry = new Long(); break;
        case DOUBLE_TAG:
            entry = new Double(); break;
        case NAME_AND_TYPE_TAG:
            entry = new NameAndType(); break;
        case UTF_TAG:
            entry = new UTF(); break;
        }
        entry.readFrom(stream);
        return entry;
    }


    public void readFrom(DataInputStream stream) throws IOException {
        int constantCount = stream.readUnsignedShort();
        //System.err.println("constants: "+constantCount);
        constants = new Entry[constantCount];
        for(int index = 1; index<constantCount; index++) {
            int tag = stream.readUnsignedByte();
            //System.err.println(""+index+": "+tag);

            Entry entry = constants[index];
            if (entry != null) {
                //System.err.println("entry exists: "+entry+", "+tag);
                // check tag vs. entry.getTag()
                entry.readFrom(stream);
            } else {
                entry = readEntry(tag, stream);
                constants[index] = entry;
            }
            if (tag == DOUBLE_TAG || tag == LONG_TAG) {
                index++;
            }
        }
    }


    abstract class Entry {
        int index;
        void writeIndex(DataOutputStream stream) throws IOException {
            stream.writeShort(index);
        }
        abstract void readFrom(DataInputStream stream) throws IOException;
        abstract void writeTo(DataOutputStream stream) throws IOException;

        public int getIndex() {
            return index;
        }
    }

    static final int CLASS_TAG = 7;
    class Class extends Entry {
        UTF name;

        void readFrom(DataInputStream stream) throws IOException {
            name = readUTF(stream);
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(CLASS_TAG);
            name.writeIndex(stream);
        }

        public java.lang.String getName() {
            return name.value.replace('/', '.');
        }

        public java.lang.String toString() {
            if (name == null) return "Class()";
            return "Class("+name.value+")";
        }
    }

    abstract class MemberRef extends Entry {
        Class _class;
        NameAndType nameAndType;

        void readFrom(DataInputStream stream) throws IOException {
            _class = readClass(stream);
            nameAndType = readNameAndType(stream);
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(getTag());
            _class.writeIndex(stream);
            nameAndType.writeIndex(stream);
        }

        abstract int getTag();

        public java.lang.String toString() {
            return "MemberRef(" + _class + " " + nameAndType + ")";
        }
    }

    static final int FIELD_REF_TAG = 9;
    class FieldRef extends MemberRef {
        int getTag() { return FIELD_REF_TAG; }
    }

    static final int METHOD_REF_TAG = 10;
    class MethodRef extends MemberRef {
        int getTag() { return METHOD_REF_TAG; }
    }

    static final int INTERFACE_METHOD_REF_TAG = 11;
    class InterfaceMethodRef extends MemberRef {

        int getTag() { return INTERFACE_METHOD_REF_TAG; }
    }

    static final int STRING_TAG = 8;
    class String extends Entry {
        UTF value;

        void readFrom(DataInputStream stream) throws IOException {
            value = readUTF(stream);
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(STRING_TAG);
            value.writeIndex(stream);
        }
    }

    static final int INTEGER_TAG = 3;
    class Integer extends Entry {
        int value;

        void readFrom(DataInputStream stream) throws IOException {
            value = stream.readInt();
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(INTEGER_TAG);
            stream.writeInt(value);
        }
    }

    static final int FLOAT_TAG = 4;
    class Float extends Entry {
        float value;

        void readFrom(DataInputStream stream) throws IOException {
            value = stream.readFloat();
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(FLOAT_TAG);
            stream.writeFloat(value);
        }
    }

    static final int LONG_TAG = 5;
    class Long extends Entry {
        long value;

        void readFrom(DataInputStream stream) throws IOException {
            value = stream.readLong();
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(LONG_TAG);
            stream.writeLong(value);
        }
    }

    static final int DOUBLE_TAG = 6;
    class Double extends Entry {
        double value;

        void readFrom(DataInputStream stream) throws IOException {
            value = stream.readDouble();
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(DOUBLE_TAG);
            stream.writeDouble(value);
        }
    }

    static final int NAME_AND_TYPE_TAG = 12;
    class NameAndType extends Entry {
        UTF name;
        UTF descriptor;

        void readFrom(DataInputStream stream) throws IOException {
            name = readUTF(stream);
            descriptor = readUTF(stream);
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(NAME_AND_TYPE_TAG);
            name.writeTo(stream);
            descriptor.writeTo(stream);
        }
        public java.lang.String toString() {
            return "NameAndType(" + name + " " + descriptor + ")";
        }
    }

    static final int UTF_TAG = 1;
    class UTF extends Entry {
        java.lang.String value;

        void readFrom(DataInputStream stream) throws IOException {
            value = stream.readUTF();
        }

        void writeTo(DataOutputStream stream) throws IOException {
            stream.writeByte(UTF_TAG);
            stream.writeUTF(value);
        }
        public java.lang.String toString() {
            return "UTF(" + value + ")";
        }
    }
}
