/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.parser.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import java.util.*;
import java.io.*;

public class AspectJCompiler extends JavaCompiler {
//    public CompilationUnit newCompilationUnit(File _file) {
//        return new CompilationUnit(this,_file);
//    }

    public AspectJCompiler(ErrorHandler errorHandler) {
        super(errorHandler);
        this.scanner = AspectJParser.class;
        //addPasses();
    }

    public AspectJCompiler() {
        this(null);
    }

    public JavaParser makeJavaParser() {
        return new AspectJParser(this);
    }

    //protected List/*CompilerPass*/ passes = null;
    public void addPass(CompilerPass pass) { addPass(pass, false); }
    public void addPass(CompilerPass pass, boolean exitOnErrors) {
        passes.add(pass);
        totalWorkEstimate += pass.getWorkEstimate();
        //if (getOptions().torture) passes.add(new TreeCheckerPass(...));
        if (exitOnErrors) passes.add(new ExitOnErrorsPass(this));
    }

    protected void addPreSymbolPasses() {
        // This allows ajdoc to print out the files, classes, and
        // packages being documented
        addPass(createParserPass(), true);
        addPass(new TypeGraphBuilderPass(getCompiler()));
        //addPass(new IntroductionPlannerPass(getCompiler()));
        addPass(new SignatureBinderPass(getCompiler()));
        addPass(new BodyBinderPass(getCompiler()));
        addPass(new ForwardReferenceChecker(getCompiler()));
        addPass(new InnerInfoPass(getCompiler()));
        addPass(new AssignmentCheckerPass(getCompiler()));
        addPass(new SpecChecker(getCompiler()), true);

        addPass(new BuildAdvicePlannersPass(getCompiler()), true);
        addPass(new InitializerCollector(getCompiler(), false));

        addPass(new JoinPointCollectorPass(getCompiler()), true);
        // this planner also softens exceptions, which is why it's
        // before flow analysis.
        addPass(new StaticJpPlannerPass(getCompiler()));

        addPass(new FlowCheckerPass(getCompiler()));
        addPass(new ThrowCheckerPass(getCompiler()));

        addPass(new AdvicePlannerPass(getCompiler()), true);

        addPass(new AdviceWeaver(getCompiler()), true);
    }

    protected void addPostSymbolPasses() {
//        addPass(new SymbolWriter(getCompiler()));

        addPass(new AbstractCompilerPass(getCompiler()) {
            public String getDisplayName() {
                return "emacssym ";
            }
            public void transformWorld() {
                if (super.getCompiler().getOptions().emacssym) {
                    new org.aspectj.asm.views.EmacsStructureManager().externalizeStructureView((AspectJCompiler)super.getCompiler());
                }
            }
        });

        //XXX there is an assumption in the implementation of introduction that this
        //XXX pass is run in a topo-sorted order.  obviously that's not true yet.
        addPass(new MixinImplementationPass(getCompiler()));

        // move these after SymbolWriter because they can change the symbols
        addPass(new PrivilegeFixer(getCompiler(), true));
        addPass(new ASTFixerPass(getCompiler()), true);

        addPass(new InitializerCollector(getCompiler(), true));

        addPass(new LocalClassPass(getCompiler()), true);
    }

    protected void addSourceGenerationPasses() {
        addPass(new NameHygienePass(getCompiler()), true);
        addPass(new CodeGenerator(getCompiler()), true);
        // runs javac on generated sources
        addPass(new RunJavacPass(getCompiler()));
        //new CodeGenerator(getCompiler()).transformWorld();
    }

    protected void addBytecodeGenerationPasses() {
        addPass(new MemberClassMunger(getCompiler()));

        if (getOptions().bcgverbose) {
            // XXX DANGER:  unsets FINAL, somehow...
            addPass(new CodeGenerator(getCompiler()));
        }

        addPass(new ClassExprPass(getCompiler()));
        addPass(new InnerAccessFixer(getCompiler()));


        addPass(new NameHygienePass(getCompiler()), true);

        // XXX this pass ELIMINATES some statements, which will
        // mess up the control pass when we start checking for
        // checked exceptions.
        addPass(new ByteCodeCleanupPass(getCompiler()));
        addPass(new FrameLocPass(getCompiler()));

        addPass(new ByteCodeGenerator(getCompiler()));

        // Note that we're not writing line maps.  We're going to hell.
    }

    protected void addPasses() {
        passes = new ArrayList();
        addPreSymbolPasses();
        addPostSymbolPasses();
        if (getOptions().usejavac || getOptions().preprocess) {
            addSourceGenerationPasses();
        } else {
            addBytecodeGenerationPasses();
        }
    }


    protected JavaCompiler getCompiler() { return this; }

    /**
     *  Override this to return true for compilers that don't want to generate code.
     */
    //protected boolean stopAfterWeaving() { return false; }


    private Correspondences correspondences = new Correspondences(this);
    public Correspondences getCorrespondences() {
        return correspondences;
    }

    public String getBuildConfigFile() { return null; }

    protected class ExitOnErrorsPass extends AbstractCompilerPass {
        public ExitOnErrorsPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return null; }
        public double getWorkEstimate() { return 0.0; }

        public void transformWorld() {
            errorHandler.exitOnErrors();
        }
    }

    protected AbstractCompilerPass createParserPass() {
        return new ParserPass(getCompiler());
    }

    protected static class ParserPass extends AbstractCompilerPass {
        public ParserPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "parse sources"; }
        public double getWorkEstimate() { return 4.0; }

        public void transform(CompilationUnit cu) {
            this.getCompiler().showMessage("    parsing " + cu.getSourceFile());
            try {
                JavaParser parser = this.getCompiler().allocateParser();
                parser.parseInterfaceOnly = false;
                parser.interfaceParse(cu);
                this.getCompiler().freeParser(parser);
            } catch (ParseException pe) {
                if (getOptions().dumpstack) {
                    pe.printStackTrace();
                }
                this.getCompiler().showError(cu.getSourceFile(),
                                        pe.getLine(),
                                        pe.getColumn(),
                                        pe.getMessage());
            } catch (IOException e) {
                this.getCompiler().showError(null,e.toString());
            }

            for (Iterator j = cu.getDefinedTypes().iterator(); j.hasNext(); ) {
                TypeDec td = (TypeDec)j.next();
                getWorld().addSourceType(td.getType());
                addInnerTypes(td); //, td.getId());
            }
        }

        private void addInnerTypes(TypeDec typeDec) { //, String id) {
            Decs body = typeDec.getBody();
            for (int i=0; i < body.size(); i++) {
                if (body.get(i) instanceof TypeDec) {
                    TypeDec innerTypeDec = (TypeDec)body.get(i);
//                    String innerId = id + '$' + innerTypeDec.getId();
//                    innerTypeDec.setExtendedId(innerId);
                    getTypeManager().addType( innerTypeDec.getType() );
                    addInnerTypes(innerTypeDec); //, innerId);
                }
            }
        }
    }

    protected static class TypeGraphBuilderPass extends AbstractCompilerPass {
        public TypeGraphBuilderPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "build type graph"; }
        public void transform(CompilationUnit cu) {
            cu.addTypesToTypeGraph();
        }
    }

    protected static class SignatureBinderPass extends WalkerPass {
        public SignatureBinderPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "resolve type signatures"; }

        public void transform(CompilationUnit cu) {
            cu.buildSignatures();
        }
    	/*
		 * @see CompilerPass#transformWorld()
		 */
		public void transformWorld() {
			super.transformWorld();
			getWorld().collectedTypeDecPlanners = true;
		}
 	}

    protected static class BodyBinderPass extends AbstractCompilerPass {
        public BodyBinderPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "bind names in bodies"; }
        public void transform(CompilationUnit cu) {
            new ScopeWalker(this.getCompiler()).process(cu);
        }
    }

    protected static class JoinPointCollectorPass extends AbstractCompilerPass {
        public JoinPointCollectorPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "collect join point shadows"; }
        public void transform(TypeDec td) {
            new JoinPointCollector(td).collect();
        }
    }

    protected static class ThrowCheckerPass extends WalkerPass {
        public ThrowCheckerPass(JavaCompiler jc) { super(jc); }

        public String getDisplayName() { return "checking throws"; }
        public ASTObject process(ASTObject node) {

            if (node instanceof CodeDec) {
                CanThrowWalker.checkThrows((CodeDec)node);
            }
            return super.process(node);
        }
    }
}

