/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts;
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.base.parser.*;
import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.TypeManager;

import java.util.*;
import java.io.*;

public class AspectJParser extends JavaParser {

    public AspectJParser(JavaCompiler compiler) {
        super(compiler);
    }

    // aspect, pointcut and privileged are virtually new keywords
    // this is part of the implementation of ajc-1.0, not necessarily
    // part of the language design of AspectJ-1.0
    {
        decParsers.put("aspect", new AspectDecParser());

        decParsers.put("pointcut", new PointcutDecParser());

        modifiers.put("privileged", new Integer(Modifiers.PRIVILEGED));
    }


    protected Map aspectDecParsers = new HashMap();
    {
        aspectDecParsers.put("declare", new DeclareParser());

        aspectDecParsers.put("before", new BeforeDecParser());
        aspectDecParsers.put("after", new AfterDecParser());
        aspectDecParsers.put("around", new AroundDecParser());

        // this is both in AspectOf and class/aspect
        aspectDecParsers.put("pointcut", new PointcutDecParser());
    }

    protected void eatModifiersPattern() {
        Token token;
        while (true) {
            token = peekToken();
            if (!(token.kind == BANG || isModifier(token))) break;
            eatTopToken();
        }
    }

    public ModifiersPattern parseModifiersPattern(int legalModifiers) {
        int trueValue = 0;
        int falseValue = 0;
        Token beginToken = peekToken();

        while (true) {
            boolean assertion = true;

            if (peekToken().kind == BANG) {
                if (getModifier(peekToken(1)) == null) break;
                eatTopToken();
                assertion = false;
            }

            Integer ivalue = getModifier(peekToken());
            if (ivalue == null) {
                if (!assertion) {
                    throwError(null, "! in modifier pattern not followed by modifier");
                }
                break;
            }

            int value = ivalue.intValue();
            if ((legalModifiers & value) != value) {
                throwError(null, "illegal modifier for this dec");
            }
            eatTopToken();
            if (assertion) trueValue  |= value;
            else        falseValue |= value;
        }
        ModifiersPattern modifiers = new ModifiersPattern(dummySource,trueValue, falseValue);
        addContext(modifiers, beginToken);
        return modifiers;
    }

    public NameTypeDsPattern parseNameTypeDsPattern() {
        if (peekToken(0).kind != BANG && !(isIdentifier(peekToken(0)))) {
            return null;
        }

        TypeDs trueNames = new TypeDs(dummySource);
        TypeDs falseNames = new TypeDs(dummySource);

        while (true) {
            Token token = peekToken(0);
            boolean assertion = true;
            if (token.kind == BANG) {
                eatTopToken();
                token = peekToken(0);
                assertion = false;
            }

            TypeD nameTypeD = parseNameTypeD();
            if (nameTypeD == null) {
                if (!assertion) throwError(token, "illegal name pattern");
            }

            if (assertion) {
                trueNames.add(nameTypeD);
            } else {
                falseNames.add(nameTypeD);
            }
            if (peekToken(0).kind != COMMA) break;
            eatTopToken();
        }

        return new NameTypeDsPattern(dummySource,trueNames, falseNames);
    }

    public NameTypeDsPattern parseThrowsPattern() {
        if (maybeEatKeyword("throws")) {
            return parseNameTypeDsPattern();
        } else {
            return null;
        }
    }

    public GenTypeName parseFormalPattern() {
        Token beginToken = peekToken();
        if (beginToken.kind == DOT && peekToken(1).kind == DOT) {
            eatTopToken(); eatTopToken();
            // an arbitrary number of formals
            return null;
        } else {
            GenTypeName genTypeName = parseGenTypeName();
            addContext(genTypeName, beginToken);
            return genTypeName;
        }
    }


    public FormalsPattern parseFormalsPattern() {
        Token beginToken = peekToken();
        eatTopToken(LPAREN);

        GenTypeNames names = new GenTypeNames(dummySource);
        FormalsPattern formalsPattern = new FormalsPattern(dummySource,names); //, -1);

        if (peekToken().kind != RPAREN) {
            while (true) {
                GenTypeName pattern = parseFormalPattern();
                if (pattern == null) {
                    pattern = new EllipsesFakeTypeName(dummySource);
                    addContext(pattern, peekToken(-1));
                }
                names.add(pattern);
                
                Token t = popToken();
                if (t.kind == RPAREN) break;
                else if (t.kind == COMMA) continue;
                throwError(t, "expected ',' or ')'");
            }
        } else {
            eatTopToken();
        }

        addContext(names, beginToken);
        addContext(formalsPattern, beginToken);
        return formalsPattern;
    }


    public Expr parseVarInitializer() {
        if (peekToken(0).kind == ASSIGN) {
            eatTopToken();
            if (peekToken(0).kind == LBRACE) {
                return parseArrayInitializer();
            } else {
                return parseExpr();
            }
        } else {
            return null;
        }
    }

    public TypeD parseOptionalReturnTypeD() {
        if (!maybeEatKeyword("returns")) return null;

        Token beginToken = peekToken();
        return parseTypeD();
    }

    public boolean isAdjacent(Token t1, Token t2) {
        //System.out.println("is adjacent? " + t1.endPosition + ", " + t2.startPosition);
        return t1.endPosition == t2.startPosition;
    }

    String makeNamePatternPiece(Token t) {
        if (t.isIdentifier()) return t.image;
        else if (t.kind == STAR) return "*";
        else if (t.kind == INTEGER_LITERAL) return t.image;
        //??? probably need to do some tricky handling of float literals here
        else return null;
    }
    
    public NamePattern parseNamePattern(Token t1, String firstPiece) {
        List pieces = new ArrayList();
        pieces.add(firstPiece);
        
        Token previousToken = t1;
        while (isAdjacent(previousToken, peekToken())) {
            previousToken = peekToken();
            String piece = makeNamePatternPiece(previousToken);

            if (piece == null) break;
            eatTopToken();
            pieces.add(piece);
        }
        
        if (pieces.size() == 1) {
            String piece = (String)pieces.get(0);
            if (piece == "*") {
                return new AnyName(makeSourceLocation(t1));
            } else {
                return new SimpleName(makeSourceLocation(t1), piece);
            }
        } else {
            return new WildName(makeSourceLocation(t1), pieces);
        }
    }    
    public NamePattern parseNamePattern() {
        Token t1 = peekToken();
        String firstPiece = makeNamePatternPiece(t1);
        if (firstPiece == null) {
            throwError(t1, "identifier pattern expected");
            return null;
        }
        eatTopToken();
        return parseNamePattern(t1, firstPiece);
    }
    
    protected boolean inAspect() {
        if (peekEnclosingTypeDec() != null && peekEnclosingTypeDec() instanceof AspectDec) {
            return true;
        }
        return false;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////    

    //XXX dangerous use of this...
    protected boolean eatIdentifier(boolean onlyDec) {
        //System.out.println(peekToken() + ": " + inAspect() + ": " + onlyDec);
        if (!inAspect() || !onlyDec) return super.eatIdentifier(onlyDec);
        
        if (eatGenTypeName()) {
            if (peekToken().kind == DOT) {
                eatTopToken(); eatTopToken();
            }
            return true;
        }
        return false;
    }    
    
    
    static class TypeNameAndId {
        public GenTypeName typeName;
        public NamePattern namePattern;
        
        public String getId() {
            if (namePattern.getSimpleName() == null) {
                namePattern.showError("identifier expected");
            }
            return namePattern.getSimpleName();
        }
    }
    
//    public String parseIdOrNew() {
//        if (maybeEatToken(NEW)) {
//            return "new";
//        } else {
//            return parseId();
//        }
//    }
    
    public TypeNameAndId parseTypeNameAndId() {
        TypeNameAndId ret = new TypeNameAndId();
        Token beginToken = peekToken();
        
        if (peekToken().kind == LPAREN || peekKeyword("subtypes")) {
            ret.typeName = parseGenTypeName();
            eatTopToken(DOT);
            ret.namePattern = parseNamePattern();
        } else {
            NamePatterns namePatterns = parseDottedNamePatterns();
            
            //namePatterns.display(0);
            
            //System.out.println("parsed: " + namePatterns);
            
            boolean includeSubTypes = false;
            if (peekToken().kind == PLUS || peekToken().kind == LBRACKET) {
                ret.typeName =  finishParsingPatternTypeName(beginToken, namePatterns);
                eatTopToken(DOT);
                ret.namePattern = parseNamePattern();
            } else {
                //XXX validate that is is a true simple name
                ret.namePattern = namePatterns.pop(); //.getSimpleName();
                if (namePatterns.size() > 0) {
                    ret.typeName = new PatternTypeName(dummySource, namePatterns, false, 0);
                    ret.typeName.setSource(namePatterns);
                }
            }
        }
        
        return ret;
    }
    
    protected boolean eatGenTypeName() {
        if (peekKeyword("import")) return false;
        
        //XXX very heuristic approach
        if (peekToken(0).kind == LPAREN && !(peekToken(1).kind == RPAREN)) {
            parseGenTypeName();
            return true;             
        } else {
            if (!eatTypeD()) return false;
            maybeEatToken(PLUS);
            return true;
        }
    }
    

    protected Dec noDecFound() {
        Token beginToken = peekToken();
        
        if (inAspect()) {
            if (lookaheadIntroducedSuperDec()) {                
                return parseIntroducedSuperDec();
            } else if (lookaheadIntroducedConstructorDec()) {
                return parseIntroducedConstructorDec();
            }
        }
        
        throwError(beginToken, "declaration or '}' expected");
        return null;
    }
    
    protected boolean lookaheadIntroducedSuperDec() {
        pushCurrentIndex();
        
        if (eatGenTypeName()) {
            if (maybeEatKeyword("implements") || maybeEatKeyword("extends")) {
                throwError(peekToken(-1), "expected +implements or +extends");
                return false;
            } else if (this.maybeEatToken(PLUS) &&
                (maybeEatKeyword("implements") || maybeEatKeyword("extends")))
            {
                popCurrentIndex();        
                return true;
            }
        }
        popCurrentIndex();
        return false;
    }
    
    public Dec parseIntroducedSuperDec() {
        getCompiler().warnVersion("1.0alpha1", tokenToASTObject(peekToken()),
                "use declare parents instead");     
        
        GenTypeName typeName = parseGenTypeName();
        eatTopToken(PLUS);
        boolean isImplements;
        if (maybeEatKeyword("implements")) {
            isImplements = true;
        } else {
            eatKeyword("extends");
            isImplements = false;
        }
        TypeDs typeDs = parseTypeDs();
        Dec ret = new IntroducedSuperDec(dummySource, typeName, typeDs, isImplements);
        eatTopToken(SEMICOLON);
        return ret;
    }
    
    
    protected boolean lookaheadIntroducedConstructorDec() {
        pushCurrentIndex();
    
        this.eatModifiers();
        //this.eatGenTypeName();
        //System.out.println(peekToken(0) + ", " + peekToken(1) + ", " + peekToken(2));
        if (this.eatGenTypeName() && maybeEatToken(DOT) && maybeEatToken(NEW)) {
            popCurrentIndex();
            return true;
        }
        
        popCurrentIndex();
        return false;
    }
    
    
    
    /* Mostly cut from JavaParser, this also allows for introduced Decs */
    public Dec parseIntroducedConstructorDec() {
        int legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE;

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiers(legalModifiers);
        GenTypeName genTypeName = null;
        if (inAspect()) {
            TypeNameAndId gid = parseTypeNameAndId();
            genTypeName = gid.typeName;
            String id = gid.getId();
            if (genTypeName != null) {
                if (!id.equals("new")) {
                    throwError(beginToken, "bad introduced constructor syntax");
                }
            } else {
                if (!id.equals(peekEnclosingTypeDec().getId())) {
                    throwError(beginToken, "constructor name doesn't match enclosing type " + 
                            peekEnclosingTypeDec().getId());
                }
            }
        } else {
            parseIdMatchingEnclosingTypeName("constructor name");
        }        
        
        
        Formals formals = parseFormals();
        TypeDs _throws = parseThrows();

        Dec ret = new ConstructorDec(dummySource,modifiers, formals, _throws, 
                    parseConstructorBody());
    
        if (genTypeName != null) {
            //System.out.println("making introduced constructor on " + genTypeName);
            return new IntroducedDec(dummySource, genTypeName, ret);
        } else {
            return ret;
        }    
    }    
    
    public Dec parseConstructorDec() {
        pushCurrentIndex();
        eatModifiers();
        if (peekKeyword("subtypes")) {
            popCurrentIndex();
            return parseIntroducedConstructorDec();
        } else {
            popCurrentIndex();
            return super.parseConstructorDec();
        }
    }   
    
    public Dec parseMethodDec() {
        int legalModifiers;
        if (inInterface) {
            legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT;
        } else {
            legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
                Modifiers.ABSTRACT | Modifiers.STATIC | Modifiers.FINAL |
                Modifiers.SYNCHRONIZED | Modifiers.NATIVE | Modifiers.STRICT;
        }

        Modifiers modifiers = parseModifiers(legalModifiers);
        if (inInterface) {
            modifiers.setAbstract(true);
            modifiers.setPublic(true);
        }        
        
        TypeD typeD = parseTypeD();
        
        GenTypeName genTypeName = null;
        String id = null;
        if (inAspect()) {
            TypeNameAndId gid = parseTypeNameAndId();
            genTypeName = gid.typeName;
            id = gid.getId();
            //XXX this is a horrible design
            if (genTypeName == null && id.equals("around")) {
                return parseAroundAdviceDec(modifiers, typeD);
            }
        } else {
            id = parseId();
        }

        Formals formals = parseFormals();

        // handle funky "int foo()[] {" return type specifier
        typeD = addBracketsToTypeD(typeD);

        TypeDs _throws = parseThrows();

        MethodDec ret = new MethodDec(dummySource, modifiers, typeD, id, formals, _throws,
                                 parseOptionalCodeBody());
        
        if (genTypeName != null) {
            //System.out.println("making introduced dec: " + id + " on " + genTypeName);
            return new IntroducedDec(dummySource, genTypeName, ret);
        } else {
            return ret;
        }
    }
    
    
    public Stmt parseFieldDec() {
        int legalModifiers;

        if (inInterface) {
            legalModifiers = Modifiers.PUBLIC | Modifiers.STATIC | Modifiers.FINAL;
        } else {
            legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
                Modifiers.STATIC | Modifiers.FINAL |
                Modifiers.TRANSIENT | Modifiers.VOLATILE;
        }

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiers(legalModifiers);
        
        if (inInterface) {
            modifiers.setPublic(true);
            modifiers.setStatic(true);
            modifiers.setFinal(true);
        }           
        
        TypeD baseTypeD = parseTypeD();

        Decs decs = null;
        Dec dec = null;

        while (true) {
            GenTypeName genTypeName = null;
            String id = null;
            if (inAspect()) {
                TypeNameAndId gid = parseTypeNameAndId();
                genTypeName = gid.typeName;
                id = gid.getId();
            } else {
                id = parseId();
            }
            TypeD typeD = baseTypeD;

            Expr initializer = null;
            typeD = addBracketsToTypeD(typeD);
            //VarDes des = new VarDes(modifiers, id, typeD);
            //??? what should the context really be for a var des in a list
            //addContext(des, beginToken);

            Token endToken = popToken();
            if (endToken.kind == ASSIGN) {
                if (typeD instanceof ArrayTypeD && peekToken().kind == LBRACE) {
                    initializer = parseArrayInitializer();
                } else {
                    initializer = parseExpr();
                }
                endToken = popToken();
            }

            dec = new FieldDec(dummySource,(Modifiers)modifiers.copy(), typeD, id, initializer);
            addContext(dec, beginToken);
            if (genTypeName != null) {
                dec = new IntroducedDec(dummySource, genTypeName, dec);
            }

            if (endToken.kind == SEMICOLON) {
                if (decs != null) decs.add(dec);
                break;
            } else if (endToken.kind == COMMA) {
                if (decs == null) decs = new Decs(dummySource,dec);
                else decs.add(dec);

                // this copy is needed to keep the AST a tree and not a graph
                baseTypeD = (TypeD)baseTypeD.copy();
                continue;
            } else {
                throwError(endToken, "expected ',' or ';'");
            }
        }

        if (decs != null) return decs;
        else return dec;
    }    
    
    
    
    
    
    ///////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////
    public ConstructorPattern parseConstructorPattern() {
        int legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE
            | Modifiers.STRICT;    // constructor may be implicitly strictfp by class decl.

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiersPattern(legalModifiers);
        
        TypeNameAndId gid = parseTypeNameAndId();
        GenTypeName declaringType = gid.typeName;
        
        //System.out.println(gid.getId());
        
        if (!"new".equals(gid.namePattern.getSimpleName())) {
            this.throwError(peekToken(-1), "expected new");
        }
        FormalsPattern formals = parseFormalsPattern();
        NameTypeDsPattern _throws = parseThrowsPattern();

        ConstructorPattern ret = new ConstructorPattern(dummySource,modifiers, declaringType, formals, _throws);
        addContext(ret, beginToken);
        return ret;
    }

    public MethodPattern parseMethodPattern() {
        int legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
            Modifiers.ABSTRACT | Modifiers.STATIC | Modifiers.FINAL |
            Modifiers.SYNCHRONIZED | Modifiers.NATIVE | Modifiers.STRICT;

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiersPattern(legalModifiers);
        GenTypeName returnType = parseGenTypeName();
        
        TypeNameAndId gid = parseTypeNameAndId();
        GenTypeName declaringType = gid.typeName;
        NamePattern id = gid.namePattern;
        if ("new".equals(id.getSimpleName())) {
            this.tokenToASTObject(beginToken).showError("constructor patterns don't have a return type");
        }

        FormalsPattern formals = parseFormalsPattern();
        NameTypeDsPattern _throws = parseThrowsPattern();

        MethodPattern ret = new MethodPattern(dummySource,modifiers, declaringType, returnType, id, formals, _throws);
        addContext(ret, beginToken);
        return ret;
    }

    public CodePattern parseConstructorOrMethodPattern() {
        //XXX this is clearly an inefficient design
        //XXX is this the straw that broke the hand-written parser's back?
        pushCurrentIndex();
        try {
            CodePattern ret = parseConstructorPattern();
            //XXX don't really want this
            //XXX popCurrentIndex();
            return ret;
        } catch (ParseException pe) {
            popCurrentIndex();
            return parseMethodPattern();
        }
    }

    public FieldPattern parseFieldPattern() {
        int legalModifiers = Modifiers.PUBLIC | Modifiers.PROTECTED | Modifiers.PRIVATE |
            Modifiers.STATIC | Modifiers.FINAL |
            Modifiers.NATIVE | Modifiers.TRANSIENT;

        Token beginToken = peekToken();
        Modifiers modifiers = parseModifiersPattern(legalModifiers);
        GenTypeName fieldType = parseGenTypeName();
        
        TypeNameAndId gid = parseTypeNameAndId();
        GenTypeName declaringType = gid.typeName;
        NamePattern id = gid.namePattern;
//        if (declaringType == null) {
//            throwError(peekToken(-1), "must include declaring type (or pattern)");
//        }

        FieldPattern ret = new FieldPattern(dummySource,modifiers, declaringType, fieldType, id);
        addContext(ret, beginToken);
        return ret;
    }

//    boolean peekComma() {
//        int nesting = 0;
//        final int MAX_LOOKAHEAD = 20;
//        for (int i = 0; i < MAX_LOOKAHEAD; i++) {
//            Token t = peekToken(i);
//            if (t.kind == COMMA) {
//                return true;
//            }
//            if (t.kind == RPAREN) {
//                nesting -= 1;
//                if (nesting < 0) return false;
//            }
//            if (t.kind == LPAREN) {
//                nesting += 1;
//            }
//        }
//        return false;
//    }

    public void skipToMatching(int lparen, int rparen, int depth) {
        do {
            Token t = popToken();
            if (t.kind == lparen) depth++;
            if (t.kind == rparen) depth--;
        } while (depth > 0);
    }
    
    
    public void skipToMatchingParen() {
        skipToMatching(LPAREN, RPAREN, 0);
    }
    
    public void warnNameChange(String name, String newName) {
        if (name.equals(newName)) return;
        
        getCompiler().warnVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
            "change from " + name + " to " + newName + "");        
    }

    public Pcd parseFunctionPcd() {
        parseNoModifiers();
        Token beginToken = peekToken();
        String name = parseAnyId();

        if (name.equals("receptions")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "replaced by call, see release notes");
            skipToMatchingParen();
            return new EmptyPcd(dummySource);
        } else if (name.equals("calls") || name.equals("call")) {
            warnNameChange(name, "call");
            
            eatTopToken(LPAREN);
            CodePattern codePattern = parseConstructorOrMethodPattern();
            eatTopToken(RPAREN);
            return new CallPcd(dummySource, codePattern); //, null);
         } else if (name.equals("gets") || name.equals("get")) {
            warnNameChange(name, "get");
            eatTopToken(LPAREN);
            FieldPattern fieldPattern = parseFieldPattern();
            eatTopToken(RPAREN);
            if (maybeEatToken(LBRACKET)) {
                getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                    "[oldValue] not supported use field access to get this value");
                skipToMatching(LBRACKET, RBRACKET, 1);
            }
            return new FieldGetPcd(dummySource,fieldPattern); //, oldVarTypeName);
         } else if (name.equals("sets") || name.equals("set")) {
            warnNameChange(name, "set");
            eatTopToken(LPAREN);
            FieldPattern fieldPattern = parseFieldPattern();
            eatTopToken(RPAREN);

            if (maybeEatToken(LBRACKET)) {
                getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                    "[oldValue][newValue] not supported use field access to get oldValue and args(newValue)");
                skipToMatching(LBRACKET, RBRACKET, 1);                
                skipToMatching(LBRACKET, RBRACKET, 0);                
            }
            return new FieldSetPcd(dummySource,fieldPattern); //, oldVarTypeName, newVarTypeName);
        } else if (name.equals("executions") || name.equals("execution")) {
            warnNameChange(name, "execution");
            eatTopToken(LPAREN);
            CodePattern codePattern = parseConstructorOrMethodPattern();
            eatTopToken(RPAREN);
            return new ExecutionPcd(dummySource,codePattern);
        } else if (name.equals("initialization")) {        
            eatTopToken(LPAREN);
            ConstructorPattern codePattern = parseConstructorPattern();
            eatTopToken(RPAREN);
            return new InitializerPcd(dummySource, codePattern);
        } else if (name.equals("initializations")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "replaced by initialization with different meaning, see release notes");
            skipToMatchingParen();
            return new EmptyPcd(dummySource);
        } else if (name.equals("staticinitializations") || name.equals("staticinitialization")) {
            warnNameChange(name, "staticinitialization");
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new StaticInitializerPcd(dummySource,genTypeName);
        } else if (name.equals("handlers") || name.equals("handler")) {
            warnNameChange(name, "handler");
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new HandlerPcd(dummySource,genTypeName);
        } else if (name.equals("hasaspect")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "change to if(A.hasAspect(o))");            
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new EmptyPcd(dummySource);
        } else if (name.equals("instanceof") || name.equals("this")) {
            warnNameChange(name, "this");
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new ThisPcd(dummySource,genTypeName);
        } else if (name.equals("within")) {
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new WithinPcd(dummySource,genTypeName);
        } else if (name.equals("target")) {
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            eatTopToken(RPAREN);
            return new TargetPcd(dummySource,genTypeName);
        } else if (name.equals("args")) {
            //eatTopToken(LPAREN);
            FormalsPattern genTypeNames = parseFormalsPattern();
            //eatTopToken(RPAREN);
            return new ArgsPcd(dummySource,genTypeNames);
        } else if (name.equals("withinall")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "replaced with with(T+)");
            eatTopToken(LPAREN);
            GenTypeName genTypeName = parseGenTypeName();
            CodePattern codePattern = null;
            eatTopToken(RPAREN);
            return new EmptyPcd(dummySource);
        } else if (name.equals("withincode")) {
            eatTopToken(LPAREN);
            CodePattern codePattern = parseConstructorOrMethodPattern();
            eatTopToken(RPAREN);
            return new WithinCodePcd(dummySource,codePattern);
        } else if (name.equals("callsto")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "not supported any more due to the removal of reception join points");
            skipToMatchingParen();
            return new EmptyPcd(dummySource);
        } else if (name.equals("cflow")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            return new CFlowPcd(dummySource,pointcut, true);
        } else if (name.equals("if")) {
            eatTopToken(LPAREN);
            Expr test = parseExpr();
            eatTopToken(RPAREN);
            return new IfPcd(dummySource, test);
        } else if (name.equals("cflowbelow")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            return new CFlowPcd(dummySource,pointcut, false);
        } else if (name.equals("cflowtop")) {
            getCompiler().warnVersion("0.8beta3", tokenToASTObject(peekToken(-1)),
                                  "use common idiom cflow(P && !cflowbelow(P))");            
            return new EmptyPcd(dummySource);
        } else {
            return new NamePcd(dummySource,null, name, parseFormalsPattern(), null);
        }
    }

    private TypeD typeDFromName(Name name) {
        UnresolvedNameTypeD ret = new UnresolvedNameTypeD(dummySource,name);
        ret.setSource(name);
        return ret;
    }
    
    
    private NamePcd parseNamePcd() {
        Name name = parseExtendedName();
        String pointcutName = name.getId();

        TypeD inTypeD = null;
        if (name.getParentName() != null) {
            inTypeD = typeDFromName(name.getParentName());
        }
        FormalsPattern formals = parseFormalsPattern();
        return new NamePcd(dummySource,inTypeD, pointcutName, formals, null);
    }

    public Pcd parsePcdTerminal() {
        if (peekToken(1).kind == DOT) {
            return parseNamePcd();
        } else {
            return parseFunctionPcd();
        }
    }


    class PcdOp extends PrefixOperator {
        public PcdOp(int kind) {
            super(kind, -1);
        }

        public Expr parse() {
            return parsePcdTerminal();
        }
    }

    class UnionPcdOperator extends InfixOperator {
        public UnionPcdOperator(int precedence) {
            super("||", precedence);
        }

        public Expr parse(Expr lhs) {
            Pcd rhsCC = (Pcd)parentParse(precedence);
            return new OrPcd(dummySource,(Pcd)lhs, rhsCC);
        }
    }

    class IntersectPcdOperator extends InfixOperator {
        public IntersectPcdOperator(int precedence) {
            super("&&", precedence);
        }

        public Expr parse(Expr lhs) {
            Pcd rhsCC = (Pcd)parentParse(precedence);
            return new AndPcd(dummySource,(Pcd)lhs, rhsCC);
        }
    }

    class ComplementPcdOperator extends PrefixOperator {
        public ComplementPcdOperator(int precedence) {
            super("!", precedence);
        }

        public Expr parse() {
            // if this is a modifiers pattern, then treat it specially
            // this is a hack for an ugly part of the syntax
            if (isModifier(peekToken(1))) {
                return parsePcdTerminal();
            }

            eatTopToken(); // eats the '!'
            return new NotPcd(dummySource,(Pcd)parentParse(precedence));
        }
    }

    class ParenPcdOperator extends PrefixOperator {
        public ParenPcdOperator(int precedence) {
            super("(", precedence);
        }

        public Expr parse() {
            eatTopToken(); // eats the '('
            Pcd cc = (Pcd)parentParse(precedence);
            eatTopToken(RPAREN);
            return cc;
        }
    }

    private InfixOperator[] infixPcdOperators = {
        new UnionPcdOperator(500),
        new IntersectPcdOperator(700),
    };

    private PrefixOperator[] prefixPcdOperators = {
        new PcdOp(IDENTIFIER),
        new PcdOp(STAR),
        new PcdOp(NEW),
        new ComplementPcdOperator(3000),
        new ParenPcdOperator(1),
    };


    protected ExprParser pointcutParser =
        new ExprParser(prefixPcdOperators, infixPcdOperators, false);

    public Pcd parsePcd() {
        Pcd ret;
        ret = (Pcd)pointcutParser.parse();
        if (ret == null) {
            throwError(null, "pointcut designator required");
        }
        return ret;
    }


    class TypeNameOp extends PrefixOperator {
        public TypeNameOp(int kind) {
            super(kind, -1);
        }

        public Expr parse() {
            return parseTypeNameTerminal();
        }
    }

    class UnionTypeNameOperator extends InfixOperator {
        public UnionTypeNameOperator(int precedence) {
            super("||", precedence);
        }

        public Expr parse(Expr lhs) {
            GenTypeName rhsCC = (GenTypeName)parentParse(precedence);
            return new UnionTypeName(dummySource,(GenTypeName)lhs, rhsCC);
        }
    }

    class IntersectTypeNameOperator extends InfixOperator {
        public IntersectTypeNameOperator(int precedence) {
            super("&&", precedence);
        }

        public Expr parse(Expr lhs) {
            GenTypeName rhsCC = (GenTypeName)parentParse(precedence);
            return new IntersectTypeName(dummySource,(GenTypeName)lhs, rhsCC);
        }
    }

    class ComplementTypeNameOperator extends PrefixOperator {
        public ComplementTypeNameOperator(int precedence) {
            super("!", precedence);
        }

        public Expr parse() {
            eatTopToken(); // eats the '!'
            return new ComplementTypeName(dummySource,(GenTypeName)parentParse(precedence));
        }
    }

    class ParenTypeNameOperator extends PrefixOperator {
        public ParenTypeNameOperator(int precedence) {
            super("(", precedence);
        }

        public Expr parse() {
            eatTopToken(); // eats the '('
            GenTypeName cc = (GenTypeName)parentParse(precedence);
            eatTopToken(RPAREN);
            return cc;
        }
    }

    private InfixOperator[] infixTypeNameOperators = {
        new UnionTypeNameOperator(500),
        new IntersectTypeNameOperator(700),
    };

    private PrefixOperator[] prefixTypeNameOperators = {
        new TypeNameOp(IDENTIFIER),
        new TypeNameOp(STAR),
        new ComplementTypeNameOperator(3000),
        new ParenTypeNameOperator(1),
    };


    protected ExprParser typeNameParser =
        new ExprParser(prefixTypeNameOperators, infixTypeNameOperators, false);


    public GenTypeName parseGenTypeName() {
        GenTypeName ret = (GenTypeName)typeNameParser.parse();
        if (ret == null) {
            throwError(peekToken(), "expected type name");
        }
        return ret;
    }

    public GenTypeName parseTypeNameTerminal() {
        Token beginToken = peekToken();

        if (maybeEatKeyword("subtypes")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(peekToken(-1)),
                "subtypes(T) should be replaced by T+");
            skipToMatchingParen();
            return null;
        }

        GenTypeName baseType = parsePatternTypeName();
//        if (beginToken.isIdentifier()
//                && peekToken(1).kind != DOT && peekToken(1).kind != STAR
//                && peekToken(1).kind != PLUS)
//        {
//            String id = parseId();
//            TypeD typeD = lookupPrimitiveTypeD(id);
//            if (typeD == null) {
//                baseType = new VarTypeName(dummySource,id);
//            } else {
//                baseType = new SimpleTypeName(dummySource,typeD);
//            }
//        } else {
//            baseType = parsePatternTypeName();
//        }

        //??? this is an ugly way to handle []s
        //baseType = addBracketsToTypeName(baseType);
        //addContext(baseType, beginToken);
        return baseType;
    }

    public GenTypeName parsePatternTypeName() {
        Token beginToken = peekToken();
        NamePatterns names = parseDottedNamePatterns();
        
        return finishParsingPatternTypeName(beginToken, names);
    }

    protected PatternTypeName finishParsingPatternTypeName(Token beginToken, NamePatterns names) {
        boolean includeSubTypes = false;
        if (maybeEatToken(PLUS)) includeSubTypes = true;
        
        int dimensions = 0;
        while (peekToken().kind == LBRACKET && peekToken(1).kind == RBRACKET) {
            eatTopToken(); eatTopToken();
            dimensions += 1;
        }
        
        PatternTypeName ret = new PatternTypeName(dummySource, names, includeSubTypes, dimensions);
        addContext(ret, beginToken);
        return ret;
    }        
    
    /** We mess around with FLOATING_POINT_LITERAL in here in order to handle
     *   *1.f() kinds of patterns
     */
    public NamePatterns parseDottedNamePatterns() {
        Token beginToken = peekToken();

        NamePatterns ret = new NamePatterns(dummySource,parseNamePattern());

        String extraName = null;
        while (peekToken().kind == DOT || peekToken().kind == FLOATING_POINT_LITERAL) {
            Token t = popToken();
            if (t.kind == FLOATING_POINT_LITERAL) {
                String i = t.image;
                int dot = i.indexOf('.');
                NamePattern p = ret.get(ret.size()-1);
                if (p instanceof AnyName) {
                    List pieces = new ArrayList();
                    pieces.add("*");
                    pieces.add(i.substring(0, dot));
                    p.replaceWith(new WildName(dummySource, pieces));
                } else {
                    ((WildName)p).getPieces().add(i.substring(0, dot));
                }
                if (dot != i.length()-1) {
                    extraName = i.substring(dot+1);
                }
            }
            if (extraName == null && maybeEatToken(DOT)) {
                ret.add(new EllipsesFakeNamePattern(dummySource));
            }
            if (extraName != null) {
                ret.add(parseNamePattern(t, extraName));
                extraName = null;
            } else {
                ret.add(parseNamePattern());
            }
        }

        addContext(ret, beginToken);
        return ret;
    }



    public PerClause parsePerClause() {
        PerClause ret = null;
        Token beginToken = peekToken();
        
        if (!beginToken.isIdentifier()) return null;
        
        String name = parseAnyId();
        if (name.equals("perthis")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            //ret = new PerObject(dummySource,pointcut, true);
            ret = new PerThisOrTarget(dummySource,pointcut, true);
        } else if (name.equals("pertarget")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            //ret = new PerObject(dummySource,pointcut, false);
            ret = new PerThisOrTarget(dummySource,pointcut, false);
        } else if (name.equals("percflow")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            ret = new PerCFlow(dummySource,pointcut, true);
        } else if (name.equals("percflowbelow")) {
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            ret = new PerCFlow(dummySource,pointcut, false);
        } else if (name.equals("issingleton")) {
            eatTopToken(LPAREN);
            eatTopToken(RPAREN);
            ret = new PerSingleton(dummySource);
        } else {
            throwError(beginToken,
                "expected issingleton, pertarget, perthis, percflow or percflowbelow");
        }

        addContext(ret, beginToken);
        return ret;
    }
    
    public PerClause parseOldOfClause() {
        PerClause ret = null;
        Token beginToken = peekToken(-1);
        
        String name = parseAnyId();
        if (name.equals("eachobject")) {
            getCompiler().errorVersion("1.0alpha1", tokenToASTObject(beginToken),
                "change to perthis or pertarget as appropriate");
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            return null;
        } else if (name.equals("eachcflow")) {
            getCompiler().warnVersion("1.0alpha1", tokenToASTObject(beginToken),
                "change 'of eachcflow' to 'percflow'");
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            ret = new PerCFlow(dummySource,pointcut, true);
        } else if (name.equals("eachcflowbelow")) {
            getCompiler().warnVersion("1.0alpha1", tokenToASTObject(beginToken),
                "change 'of eachcflowbelow' to 'percflowbelow'");
            eatTopToken(LPAREN);
            Pcd pointcut = parsePcd();
            eatTopToken(RPAREN);
            ret = new PerCFlow(dummySource,pointcut, false);
        } else if (name.equals("eachJVM")) {
            getCompiler().warnVersion("1.0alpha1", tokenToASTObject(beginToken),
                "change 'of eachJVM' to 'issingleton'");
            eatTopToken(LPAREN);
            eatTopToken(RPAREN);
            ret = new PerSingleton(dummySource);
        } else {
            throwError(beginToken,
                "expected issingleton, pertarget, perthis, percflow or percflowbelow");
        }

        addContext(ret, beginToken);
        return ret;
    }

    class PointcutDecParser extends DecParser {
        public Dec parse() {
            final int legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT |
                Modifiers.FINAL | Modifiers.STATIC |
                Modifiers.PRIVATE | Modifiers.PROTECTED;

            Modifiers modifiers = parseModifiers(legalModifiers);
            eatKeyword("pointcut");

            String name = parseId();
            Formals formals = parseFormals();
            TypeD resultTypeD = parseOptionalReturnTypeD();

            if (modifiers.isAbstract()) {
                if (peekToken().kind == COLON) {
                    throwError(null, "abstract pointcut can't have body");
                }
                eatTopToken(SEMICOLON);
                return new PointcutDec(dummySource,modifiers, name, resultTypeD, formals, new EmptyPcd(dummySource));
            } else {
                Pcd pointcut;
                
                if (maybeEatToken(COLON)) {
                    pointcut = parsePcd();
                    eatTopToken(SEMICOLON);
                } else {
                    pointcut = new EmptyPcd(dummySource);
                }

                return new PointcutDec(dummySource,modifiers, name, resultTypeD, formals, pointcut);
            }
        }
    }


    class AspectDecParser extends DecParser {
        public Dec parse() {
            // not used, we override parse(boolean)
            throw new RuntimeException();
        }
        public Dec parse(boolean isNotMethodLocal) {
            int legalModifiers;
            if (peekEnclosingTypeDec() == null) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT | Modifiers.PRIVILEGED;
            } else if (inInterface) {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT | Modifiers.PRIVILEGED | Modifiers.STATIC;

            } else {
                legalModifiers = Modifiers.PUBLIC | Modifiers.ABSTRACT | Modifiers.FINAL |
                    Modifiers.STRICT | Modifiers.PRIVILEGED |
                    Modifiers.PRIVATE | Modifiers.PROTECTED | Modifiers.STATIC;
            }

            Modifiers modifiers = parseModifiers(legalModifiers);
            if (inInterface) modifiers.setPublic(true);

            eatKeyword("aspect");
            String id = parseId();

            TypeD superClass = null;
            if (maybeEatKeyword("extends")) {
                superClass = parseTypeD();
            }
            TypeDs _implements = null;
            if (maybeEatKeyword("implements")) {
                _implements = parseTypeDs();
            }
            GenTypeName _dominates = null;
            if (maybeEatKeyword("dominates")) {
                _dominates = parseGenTypeName();
            }

            // the of clause is optional but it must go at the end of the list
            PerClause perClause = null;
            if (maybeEatKeyword("of")) {
                //XXX show version error
                perClause = parseOldOfClause();
            } else {
                perClause = parsePerClause();
            }

            TypeDec ret =
                new AspectDec(dummySource, modifiers, id, superClass, _implements,
                               _dominates, perClause, null);

            ret.setEnclosingTypeDec(peekEnclosingTypeDec());
            pushEnclosingTypeDec(ret);
            Decs body = null;
            boolean oldInInterface = inInterface;
            boolean oldParseInterfaceOnly = parseInterfaceOnly;
            try {
                inInterface = false;
                parseInterfaceOnly = false;
                body = parseDecs(aspectDecParsers);
            } finally {
                inInterface = oldInInterface;
                parseInterfaceOnly = oldParseInterfaceOnly;
            }
            popEnclosingTypeDec();

            ret.setBody(body);
            if (peekEnclosingTypeDec() == null) currentCompilationUnit.addDefinedType(ret);
            return ret;
        }
    }

    class DeclareParser extends DecParser {
        public Dec parse() {
            parseNoModifiers();
            eatKeyword("declare");
            
            String kind = parseId();
            
            if (kind.equals("parents")) {
                eatTopToken(COLON);
                
                GenTypeName typeName = parseGenTypeName();
                boolean isImplements;
                if (maybeEatKeyword("implements")) {
                    isImplements = true;
                } else {
                    eatKeyword("extends");
                    isImplements = false;
                }
                TypeDs typeDs = parseTypeDs();
                Dec ret = new IntroducedSuperDec(dummySource, typeName, typeDs, isImplements);
                eatTopToken(SEMICOLON);
                return ret;         
            } else if (kind.equals("error") || kind.equals("warning")) {
                boolean isWarning = kind.equals("warning");
                eatTopToken(COLON);
                Pcd pointcut = parsePcd();
                eatTopToken(COLON);
                Expr message = parseExpr();
                return new ShowErrorDec(dummySource, pointcut, message, isWarning);
            } else if (kind.equals("soft")) {
                eatTopToken(COLON);
                GenTypeName typeName = parseGenTypeName();
                eatTopToken(COLON);
                Pcd pointcut = parsePcd();
                return new SoftThrowableDec(dummySource, typeName, pointcut);
            } else {
                throwError(peekToken(-1), "no declare " + kind + " is defined");
                return null;
            }
        }
    }
    
    class BeforeDecParser extends DecParser {
        public Dec parse() {
            Modifiers modifiers = parseModifiers(Modifiers.STATIC | Modifiers.STRICT);
            eatKeyword("before");

            Formals formals = parseFormals();
            TypeDs _throws = parseThrows();
            //TypeD resultTypeD = parseOptionalReturnTypeD();
            eatTopToken(COLON);

            Pcd pointcut = parsePcd();

            return new BeforeAdviceDec(dummySource,modifiers, formals, _throws, pointcut, parseCodeBodyFully());
        }
    }

    //XXX used by parseMethodDec to deal with some ambiguities
    public AroundAdviceDec parseAroundAdviceDec(Modifiers modifiers, TypeD resultTypeD) {
        // would be nice to verify that modifiers are appropriate
        
        Formals formals = parseFormals();
        TypeDs _throws = parseThrows();
        eatTopToken(COLON);
        
        Pcd pointcut = parsePcd();
        
        return new AroundAdviceDec(dummySource,modifiers, resultTypeD, formals, _throws, pointcut, parseCodeBodyFully());  
    }
    
    class AroundDecParser extends DecParser {
        public Dec parse() {
            Modifiers modifiers = parseModifiers(Modifiers.STATIC | Modifiers.STRICT);
            
            boolean needReturns = false;
            TypeD resultTypeD = null;
            if (maybeEatKeyword("around")) {
                needReturns = true;
                getCompiler().warnVersion("1.0alpha1",
                        tokenToASTObject(peekToken(-1)),
                        "the return type now goes before the around keyword,\n" +
                        "i.e. Object around() instead of around() returns Object");
            } else {
                resultTypeD = parseTypeD();
                eatKeyword("around");
            }

            Formals formals = parseFormals();
            // I believe this is always required for around advice


            if (needReturns) {
                if (!maybeEatKeyword("returns") ) {
                    getCompiler().showError(formals, "return type required for around");
                } else {
                    resultTypeD = parseTypeD();
                }
            }
            TypeDs _throws = parseThrows();
            eatTopToken(COLON);

            Pcd pointcut = parsePcd();

            return new AroundAdviceDec(dummySource,modifiers, resultTypeD, formals, _throws, pointcut, parseCodeBodyFully());  
        }
    }

    class AfterDecParser extends DecParser {
        public Dec parse() {
            Modifiers modifiers = parseModifiers(Modifiers.STATIC | Modifiers.STRICT);
            eatKeyword("after");

            Formals formals = parseFormals();

            if (maybeEatKeyword("returning")) {
                FormalDec formal = null;
                if (maybeEatToken(LPAREN)) {
                    if (peekToken().kind != RPAREN) {
                        formal = parseFormal();
                    }
                    eatTopToken(RPAREN);
                }

                TypeDs _throws = parseThrows();
                eatTopToken(COLON);
                
                Pcd pointcut = parsePcd();
                return new AfterReturningAdviceDec(dummySource,modifiers, formals, formal, _throws, pointcut, parseCodeBodyFully());
            } else if (maybeEatKeyword("throwing")) {
                FormalDec formal = null;
                if (maybeEatToken(LPAREN)) {
                    if (peekToken().kind != RPAREN) {
                        formal = parseFormal();
                    }
                    eatTopToken(RPAREN);
                }
                TypeDs _throws = parseThrows();
                eatTopToken(COLON);
                Pcd pointcut = parsePcd();
                return new AfterThrowingAdviceDec(dummySource,modifiers, formals, formal, _throws, pointcut, parseCodeBodyFully());
            } else {
                TypeDs _throws = parseThrows();
                eatTopToken(COLON);
                Pcd pointcut = parsePcd();
                return new AfterAdviceDec(dummySource,modifiers, formals, _throws, pointcut, parseCodeBodyFully());
            }
        }
    }
}
