/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts;

import org.aspectj.compiler.base.*;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.util.CollectionUtil;

import org.aspectj.tools.ide.Declaration;
import org.aspectj.tools.ide.SymbolManager;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.base.CompilerObject;



import java.io.*;
//import java.io.Writer;
//import java.io.File;
//import java.io.FileWriter;
//import java.io.BufferedWriter;
//import java.io.IOException;


import java.util.*;

// Make sure that references to Vector use the standard version
// This lets us generate serialized files which interoperate with other code better
import java.util.Vector;


// ultimate goal is probably to do dependency tracking here to make enviro
// much nicer to use...

class CompilationUnitCorrespondences extends CompilerObject {
    // a map of cu's that this one depends on, where each entry in the map
    // is the ast node that causes the depend
    // this should allow sophisticated dependency checking sometime in the future
    private CompilationUnit compilationUnit;
    public CompilationUnitCorrespondences(CompilationUnit cu) {
        super(cu.getCompiler());
        compilationUnit = cu;
    }

    //private Map dependencies = new HashMap();

    //private Map correspondences = new HashMap();

    /** the code in this cu at myNode either impacts, or is impacted by
      the code at otherNode **/
    //public void addCorrespondence(ASTObject myNode, ASTObject otherNode) {
    //    CollectionUtil.getSetInMap(correspondences, myNode).add(otherNode);
    //}

    private Map pointsTo = new HashMap();
    private Map pointedToBy = new HashMap();

    public Set getPointsToSet(ASTObject fromNode) {
        return CollectionUtil.getSetInMap(pointsTo, fromNode);
    }

    public Set getPointedToBySet(ASTObject fromNode) {
        return CollectionUtil.getSetInMap(pointedToBy, fromNode);
    }

    public void addPointsTo(ASTObject fromNode, ASTObject toNode) {
        // backwards compatibility
        //addCorrespondence(fromNode, toNode);

        CollectionUtil.getSetInMap(pointsTo, fromNode).add(toNode);
    }

    public void addPointedToBy(ASTObject toNode, ASTObject fromNode) {
        // backwards compatibility
        //addCorrespondence(toNode, fromNode);

        CollectionUtil.getSetInMap(pointedToBy, toNode).add(fromNode);
    }

    private String quoteString(String s) {
        return "\""+s+"\" ";
    }

    private String getKind(ASTObject node) {
        if (node instanceof TypeDec) {
            TypeDec typeDec = (TypeDec)node;
            if (typeDec instanceof InterfaceDec) return "interface";
            if (typeDec instanceof ClassDec) return "class";
            if (typeDec instanceof AspectDec) return "aspect";
        }

        if (node instanceof VarDec) return "field";
        if (node instanceof InitializerDec) return "initializer";
        if (node instanceof ConstructorDec) return "constructor";
        if (node instanceof MethodDec) return "method";


//        if (node instanceof IntroduceOnDec) return "introduction";
        if (node instanceof IntroducedDec) return "introduction";
        if (node instanceof IntroducedSuperDec) return "introduction";

        //********************The new types**********************

        if (node instanceof AdviceDec) return "advice";
        if (node instanceof PointcutDec) return "pointcut";

        // internal error
        return "unknown";
    }

    private String getDeclaringType(ASTObject node) {
        Type declaringType = node.getDeclaringType();
        if (declaringType == null) return null;
        return declaringType.getTypeDec().getExtendedId().replace('$', '.');
    }

    private String getSignature(ASTObject node) {
        if (node instanceof TypeDec) {
            return ((TypeDec)node).getId();
        }


//        if (node instanceof IntroduceOnDec) {
//            return "introduction";
//        }

        if (node instanceof PointcutDec) {
            PointcutDec ccDec = (PointcutDec)node;
            String name = ccDec.getName();

            return name+ccDec.getFormals().toShortString();
        }


        if (node instanceof AdviceDec) {
            AdviceDec aDec = (AdviceDec)node;
            return aDec.toShortString();
        }


        if (node instanceof VarDec) {
            return ((VarDec)node).getName();
        }

        if (node instanceof CodeDec) {
            CodeDec codeDec = (CodeDec)node;
            String name = codeDec.getName();
            if (name.equals("<init>")) name = getDeclaringType(node);

            return name+codeDec.getFormals().toShortString();
        }

        if (node instanceof IntroducedDec) {
            IntroducedDec idec = (IntroducedDec)node;
            return idec.getTargets().toShortString()+"."+getSignature(idec.getDec());
        }
        if (node instanceof IntroducedSuperDec) {
            IntroducedSuperDec idec = (IntroducedSuperDec)node;

            return idec.getTargets().toShortString()+"+" +
               ( idec.getIsImplements() ? "implements" : "extends" );
        }

        return "...";
    }

    public String trimSig(String sig) {
        int index = sig.length()-1;
        while (index > 0) {
            char ch = sig.charAt(index);
            if (ch == '{' || ch == '}' || ch == ';' || Character.isWhitespace(ch)) {
                index--;
                continue;
            }
            break;
        }
        return sig.substring(0, index+1);
    }

    public String getModifiers(ASTObject node) {
        if (node instanceof IntroducedDec) {
            return ((IntroducedDec)node).getDec().getModifiers().toShortString();
        } else if (node instanceof Dec) {
            return ((Dec)node).getModifiers().toShortString();
        } else {
            return "";
        }
    }

    /*
    public Crosscut getCrosscut(ASTObject node) {
        if (node instanceof CrosscutActionDec) {
            CrosscutActionDec ccDec = (CrosscutActionDec)node;
            return ccDec.getCrosscut();
        } else if (node instanceof AdviceDec1) {
            AdviceDec1 adviceDec = (AdviceDec1)node;
            return adviceDec.getCrosscut();
        } else {
            return null;
        }
    }

    public String getCrosscutDesignator(ASTObject node) {
        Crosscut cc = getCrosscut(node);
        if (cc == null) return null;
        else return cc.toShortString();
    }

    public Declaration getCrosscutDeclaration(ASTObject node) {
        Crosscut cc = getCrosscut(node);
        if (cc == null) return null;
        if (!(cc instanceof NameCrosscut)) return null;

        NameCrosscut nameCC = (NameCrosscut)cc;
        return makeBaseDeclaration(nameCC.getCrosscutDec());
    }
    */

    public Pcd getPcd(ASTObject node) {
        if (node instanceof AdviceDec) {
            AdviceDec adviceDec = (AdviceDec)node;
            return adviceDec.getPcd();
        } else if (node instanceof PointcutDec) {
            return ((PointcutDec)node).getPcd();
        } else {
            return null;
        }
    }

    public String getPcdString(ASTObject node) {
        Pcd cc = getPcd(node);
        if (cc == null) return null;
        return cc.toShortString();
    }

    public Declaration getPointcutDeclaration(ASTObject node) {
        Pcd cc = getPcd(node);
        if (cc == null) return null;
        if (!(cc instanceof NamePcd)) return null;

        NamePcd nameCC = (NamePcd)cc;
        return makeBaseDeclaration(nameCC.getPointcutDec());
    }


    public String getFullSignature(ASTObject node) {
        if (node instanceof Dec) {
            Dec dec = (Dec)node;
            return dec.getSignatureString();
        } else {
            return null;
        }
    }

    public void doIndent(Writer writer, int size) throws IOException {
        for(int i=0; i<size; i++) writer.write(" ");
    }

    public Declaration makeBaseDeclaration(ASTObject node) {
        if (node == null) return null;

        //XXX hack for IDE folks for now
        if (node instanceof InitializerDec) return null;

        //??? I don't think we want to show these
        if (node instanceof Dec) {
            if ( ((Dec)node).isIntroduced()) return null;
        }

        String kindPrefix = "";
//        if (node.flag == IntroductionWeaver.INTRODUCED) {
//            kindPrefix = "introduced-";
//        }
        if (node.getSourceFile() == null) {
            //!!!node.showWarning("no source file");
            return null;
        }

        Declaration dec =
               new Declaration(node.getBeginLine(), node.getEndLine(),
                               node.getBeginColumn(), node.getEndColumn(),
                               getModifiers(node),
                               getSignature(node),
                               getFullSignature(node),
                               getPcdString(node),
                               getDeclaringType(node),
                               kindPrefix+getKind(node),
                               node.getSourceFile().getAbsolutePath(),
                               node.getFormalComment(),
                               node.getCompilationUnit().getPackageName()
                               );
        dec.setCrosscutDeclaration(getPointcutDeclaration(node));
        return dec;
    }

    private Declaration[] makeDeclarations(Collection nodes) {
        if (nodes == null) return new Declaration[0];

        List newDeclarations = new ArrayList();
        for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
            Declaration dec = makeBaseDeclaration((ASTObject)iter.next());
            //System.err.println("  add: "+dec.getSignature());
            if (dec == null) continue;
            newDeclarations.add(dec);
        }

        return (Declaration[])newDeclarations.toArray(new Declaration[0]);
    }

    private Declaration[] decsToDeclarations(Decs decs) {
        if (decs == null) return new Declaration[0];

        List newDeclarations = new ArrayList();
        for (int i=0; i<decs.size(); i++) {
            if (!decs.get(i).isLanguageVisible()) {
                continue;
            }
            Declaration dec = makeDeclaration(decs.get(i));
            //System.err.println("  add: "+dec.getSignature());
            if (dec == null) continue;
            newDeclarations.add(dec);
        }

        return (Declaration[])newDeclarations.toArray(new Declaration[newDeclarations.size()]);
    }

    private Declaration makeDeclaration(ASTObject node) {
        Declaration declaration = makeBaseDeclaration(node);
        if (declaration == null) return null;

        declaration.setPointsTo(makeDeclarations((Collection)pointsTo.get(node)));
        declaration.setPointedToBy(makeDeclarations((Collection)pointedToBy.get(node)));

        if (node instanceof TypeDec) {
            Decs enclosedDecs = ((TypeDec)node).getBody();
            declaration.setDeclarations(decsToDeclarations(enclosedDecs));
        }
//        if (node instanceof IntroduceOnDec) {
//            Decs enclosedDecs = ((IntroduceOnDec)node).getBody();
//            declaration.setDeclarations(decsToDeclarations(enclosedDecs));
//        }

        return declaration;
    }

    public void dumpSymbolFile(OutputStream stream) throws IOException {
        Declaration[] declarations = decsToDeclarations(compilationUnit.getDecs());
        //System.err.println("size: "+ declarations.size());

        ObjectOutputStream objectStream = new ObjectOutputStream(stream);
        objectStream.writeObject(declarations);
        objectStream.close();
    }
}
