/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
  * @grammar ???
  * @child Modifiers modifiers
  * @child GenTypeName declaringTypeName
  */
public abstract class DecPattern extends ASTObject {
    public abstract String toShortString();

    protected String makeDeclaringTypeString() {
        if (getDeclaringTypeName() == null) return "";
        return getDeclaringTypeName().toShortString() + '.';
    }

    public abstract String getLookupId();

    private boolean shownWarning = false;
    
    //XXX this shouldn't actually need to make a JpPlan
    public boolean matches(JoinPoint jp) {
        return makePlan(jp).isPossible();
    }

    public JpPlan makePlan(JoinPoint jp) {        
        if (!(modifiers.matches(jp.getTargetSO().getModifiers()))) return JpPlan.NO_PLAN;

        JpPlan plan;
        if (getDeclaringTypeName() != null) {
            if (jp.getTargetType() == null) return JpPlan.NEVER_PLAN;
            plan = matchDeclaringTypeName(jp, jp.getTargetType(), 
                                                jp.getTargetSO());
        } else {
            plan = new JpPlan(jp);
        }

        //System.out.println("formals: " + joinPoint.getCodeFormals().unparse() + " plan: " + plan);
//        if (plan != null && !shownWarning && matchCodeDec.isDeprecated()) {
//            String message =
//              "This pattern targets deprecated methods (including \n" +
//              " " + matchCodeDec.toShortString() + ").  Even if you don't actually refer\n" +
//              " to this method in your code, you will get a warning from the Java compiler at\n" +
//              " the end of the compilation.  This will be fixed in a future compiler.";
//
//            showMessage(message);
//            shownWarning = true;
//        }
        return plan;
    }

    protected boolean isStaticMatch(SemanticObject so) {
        return so.isStatic();
    }
    
    JpPlan matchDeclaringTypeName(JoinPoint joinPoint, Type inType, SemanticObject so) {        
        //System.out.println("  matching dec? " + inType + ": " + so.toShortString());
        
        //XXX This design is very brittle and depends on the details of SemanticObjects
        //XXX and the way that they are held in Types
        SemanticObject matchingSO = inType.findMatchingSemanticObject(so);
        if (matchingSO == null) return JpPlan.NO_PLAN;
        
        if (isStaticMatch(so)) {
            if (matchingSO != so) return JpPlan.NO_PLAN;
        }

        if (getDeclaringTypeName().matches(inType)) {
            return new JpPlan(joinPoint);
        }

        //??? for fields, static methods and constructors this can probably be optimized
        //??? to only go up the class hierarchy ignoring interfaces
        for (Iterator i = inType.getDirectSuperTypes().iterator(); i.hasNext(); ) {
            JpPlan plan = matchDeclaringTypeName(joinPoint, (Type)i.next(), so);
            if (plan.isPossible()) return plan;
        }

        return JpPlan.NO_PLAN;
    }    
    
    //BEGIN: Generated from @child and @property
    protected Modifiers modifiers;
    public Modifiers getModifiers() { return modifiers; }
    public void setModifiers(Modifiers _modifiers) {
        if (_modifiers != null) _modifiers.setParent(this);
        modifiers = _modifiers;
    }
    
    protected GenTypeName declaringTypeName;
    public GenTypeName getDeclaringTypeName() { return declaringTypeName; }
    public void setDeclaringTypeName(GenTypeName _declaringTypeName) {
        if (_declaringTypeName != null) _declaringTypeName.setParent(this);
        declaringTypeName = _declaringTypeName;
    }
    
    public DecPattern(SourceLocation location, Modifiers _modifiers, GenTypeName _declaringTypeName) {
        super(location);
        setModifiers(_modifiers);
        setDeclaringTypeName(_declaringTypeName);
    }
    protected DecPattern(SourceLocation source) {
        super(source);
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return modifiers;
        case 1: return declaringTypeName;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "modifiers";
        case 1: return "declaringTypeName";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setModifiers((Modifiers)child); return;
        case 1: setDeclaringTypeName((GenTypeName)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "DecPattern()";
    }
    
    //END: Generated from @child and @property
}

