/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
  * @grammar ([typename|.., ]*)
  * @child GenTypeNames names
  */
public class FormalsPattern extends ASTObject {
    private boolean isEllipses(GenTypeName name) {
        return name instanceof EllipsesFakeTypeName;
    }
    
    
    public JpPlan finishPlan(JpPlan plan, ASTObject nodes, JoinPoint point, 
                                int namesIndex, int formalsIndex) {
        if (namesIndex >= names.size()) {
            if (formalsIndex >= nodes.getChildCount()) return plan;
            else return JpPlan.NO_PLAN;
        }
        
        GenTypeName name = names.get(namesIndex);
        if (isEllipses(name)) {
            JpPlan plan1 = plan;
            JpPlan ret = JpPlan.NO_PLAN;
            for (; formalsIndex <= nodes.getChildCount(); formalsIndex++) {
                JpPlan tmp = finishPlan(plan1, nodes, point, namesIndex+1, formalsIndex);
                //??? bairly adequate ambiguity handling
                if (tmp.isPossible()) {
                    ret = tmp;
                    plan1 = tmp;
                }
            }
            return ret;
        } else {
            if (formalsIndex >= nodes.getChildCount()) return JpPlan.NO_PLAN;
            JpPlan paramPlan = name.makePlan(point, (Expr)nodes.getChildAt(formalsIndex));
            if (paramPlan.isPossible()) {
                return finishPlan(plan.and(paramPlan), nodes, point, namesIndex+1, formalsIndex+1);
            } else {
                return JpPlan.NO_PLAN;
            }
        }
    }
    
    
    public JpPlan finishPlan(JpPlan plan, Exprs exprs, JoinPoint point) {
        return finishPlan(plan, exprs, point, 0, 0);
    }

    public JpPlan makePlan(Exprs exprs, JoinPoint jp) {
        return finishPlan(JpPlan.ANY_PLAN, exprs, jp);
    }
    
    
    public boolean matches(Formals formals) {
        return matches(formals, 0, 0);
    }
    
    public Type extractType(ASTObject node) {
        if (node instanceof FormalDec) {
            return ((FormalDec)node).getType();
        } else if (node instanceof Expr) {
            return ((Expr)node).getType();
        } else {
            return null;
        }
    }
    
    public boolean matches(ASTObject nodes, int namesIndex, int formalsIndex) {
        if (namesIndex >= names.size()) {
            if (formalsIndex >= nodes.getChildCount()) return true;
            else return false;
        }
        
        GenTypeName name = names.get(namesIndex);
        if (isEllipses(name)) {
            for (; formalsIndex <= nodes.getChildCount(); formalsIndex++) {
                if (matches(nodes, namesIndex+1, formalsIndex)) return true;
            }
            return false;
        } else {
            if (formalsIndex >= nodes.getChildCount()) return false;
            if (name.matches(extractType(nodes.getChildAt(formalsIndex)))) {
                return matches(nodes, namesIndex+1, formalsIndex+1);
            } else {
                return false;
            }
        }
    }
    
    
    public String toShortString() {
        StringBuffer buf = new StringBuffer();
        buf.append("(");
        final int N = names.size();
        for(int i=0; i<=N; i++) {
            if (i < N) {
               buf.append(names.get(i).toShortString());
            }
            if (i < (N-1)) {
                buf.append(", ");
            }
        }
        buf.append(")");
        return buf.toString();
    }    
    
    //BEGIN: Generated from @child and @property
    protected GenTypeNames names;
    public GenTypeNames getNames() { return names; }
    public void setNames(GenTypeNames _names) {
        if (_names != null) _names.setParent(this);
        names = _names;
    }
    
    public FormalsPattern(SourceLocation location, GenTypeNames _names) {
        super(location);
        setNames(_names);
    }
    protected FormalsPattern(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        FormalsPattern ret = new FormalsPattern(getSourceLocation());
        ret.preCopy(walker, this);
        if (names != null) ret.setNames( (GenTypeNames)walker.process(names) );
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return names;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "names";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setNames((GenTypeNames)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "FormalsPattern()";
    }
    
    //END: Generated from @child and @property
}
