/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.crosscuts.joinpoints.TypeDecPlanner;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.crosscuts.*;

/**
 * @grammar introduction ...
 * @child GenTypeName targets
 * @child TypeDs typeDs
 * @property boolean isImplements
 */

public class IntroducedSuperDec extends Dec implements TypeDecPlanner {

    public Modifiers getModifiers() { return getAST().makeModifiers(0); }

//     void addMissingName(TypeDs namesToAdd, TypeDs toList, TypeDec typeDec) {
//     }

    public void walkScope(ScopeWalker walker) {
        super.walkScope(walker);
        getWorld().addTypeDecPlanner(this);
    }
        
    /*
    //INTRO from IntroductionPlannerPass
    public void preIntroduction(IntroductionPlannerPass pass) {
        ScopeWalker.bindNames(targets, getBytecodeTypeDec().getTypeScope());
        ScopeWalker.bindNames(typeDs, getBytecodeTypeDec().getTypeScope());
        getWorld().addTypeDecPlanner(this);
    }
    */

    protected void addSuperType(Type toType, Type newSuperType) {
        if (toType.isSubtypeOf(newSuperType)) return;

        if (toType.isInterface()) {
            addSuperTypeToInterfaceDec((InterfaceDec)toType.getTypeDec(), newSuperType);
        } else if (toType.isClass()) {
            addSuperTypeToClassDec((ClassDec)toType.getTypeDec(), newSuperType);
        } else {
            showWarning("can't apply to a " + toType.getTypeDec().getClass().getName());
        }
    }

    boolean checkExistingSubtypes(Type thisType, Type newSuperType) {
        if (thisType.isSubtypeOf(newSuperType)) return true;
        if (newSuperType.isSubtypeOf(thisType)) {
            showError("circular inheritance: " + newSuperType.toShortString() +
                " is a subtype of " + thisType.toShortString());
            return true;
        }
        return false;
    }

    protected void addSuperTypeToInterfaceDec(InterfaceDec typeDec, Type newSuperType) {
        if (!newSuperType.isInterface()) {
            showError("interface can't extend a class");
            return;
        }
        if (checkExistingSubtypes(typeDec.getType(), newSuperType)) return;
        typeDec.addSuperInterfaceType(newSuperType);
        typeDec.getType().addDirectSuperType(newSuperType);
    }

    protected void addSuperTypeToClassDec(ClassDec typeDec, Type newSuperType) {
        if (checkExistingSubtypes(typeDec.getType(), newSuperType)) return;

        if (newSuperType.isInterface()) {
            typeDec.addSuperInterfaceType(newSuperType);
            typeDec.getType().addDirectSuperType(newSuperType);
        } else if (newSuperType.isClass()) {
            Type oldParent = typeDec.getSuperClassType();
            if (oldParent == null) return;

            if (newSuperType.isSubtypeOf(oldParent)) {
                typeDec.setSuperClass(newSuperType.makeTypeD());
                typeDec.getType().addDirectSuperType(newSuperType);
            } else {
                showError("introduced superclass must extend old superclass: " +
                        oldParent.getPrettyString());
            }
        } else {
            showError("classes can't extend " + newSuperType.getTypeDec().getClass().getName());
        }
    }



    public void plan(TypeDec typeDec, int phase) {
        if (phase != TypeDecPlanner.TYPE) return;
        
        if (!getTargets().matches(typeDec.getType())) return;

        final int N = typeDs.size();
        for (int i = 0; i < N; i++) {
            TypeD name = typeDs.get(i);
            getCompiler().showMessage("    introducing " + name.toShortString() +
                    " on " + typeDec.toShortString());

            addSuperType(typeDec.getType(), name.getType());
        }
        ((AspectJCompiler)getCompiler()).getCorrespondences().addPointsTo(this, typeDec);
    }

    public String getKind() {
        return "declare parents";
    }

    public String getId() {
        return targets.toShortString();
    }

    public String toShortString() {
        return "declare parents: " + targets.toShortString();
    }  
    
    //BEGIN: Generated from @child and @property
    protected GenTypeName targets;
    public GenTypeName getTargets() { return targets; }
    public void setTargets(GenTypeName _targets) {
        if (_targets != null) _targets.setParent(this);
        targets = _targets;
    }
    
    protected TypeDs typeDs;
    public TypeDs getTypeDs() { return typeDs; }
    public void setTypeDs(TypeDs _typeDs) {
        if (_typeDs != null) _typeDs.setParent(this);
        typeDs = _typeDs;
    }
    
    protected boolean isImplements;
    public boolean getIsImplements() { return isImplements; }
    public void setIsImplements(boolean _isImplements) { isImplements = _isImplements; }
    
    public IntroducedSuperDec(SourceLocation location, GenTypeName _targets, TypeDs _typeDs, boolean _isImplements) {
        super(location);
        setTargets(_targets);
        setTypeDs(_typeDs);
        setIsImplements(_isImplements);
    }
    protected IntroducedSuperDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        IntroducedSuperDec ret = new IntroducedSuperDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (targets != null) ret.setTargets( (GenTypeName)walker.process(targets) );
        if (typeDs != null) ret.setTypeDs( (TypeDs)walker.process(typeDs) );
        ret.isImplements = isImplements;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return targets;
        case 1: return typeDs;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "targets";
        case 1: return "typeDs";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setTargets((GenTypeName)child); return;
        case 1: setTypeDs((TypeDs)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "IntroducedSuperDec(isImplements: "+isImplements+")";
    }
    
    //END: Generated from @child and @property
}



