/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;
import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.*;
import org.aspectj.util.*;

import org.aspectj.compiler.base.TypeManager;
import java.util.*;

/**
 * @grammar typename
 * @property NamePatterns names
 * @property boolean includeSubTypes
 * @property int dimensions
 */

public class PatternTypeName extends GenTypeName {
    /**
     *  These maps cache the result of matching with different types.  This is an
     *  extremely important performance optimization.  Removing this optimization
     *  can easily increase compile times by 20X.
     */
    Map matchCache = new HashMap();
    Map instanceMatchCache = new HashMap();

    public String toShortString() {
        String ret = names.toShortString();
        if (includeSubTypes)
            ret += "+";
        return ret;
    }

    private Formals getEnclosingFormals(ASTObject node) {
        if (node == null)
            return null;
        if (node instanceof Formals)
            return (Formals) node;
        if (node instanceof AdviceDec)
            return ((AdviceDec) node).getFormals();
        if (node instanceof PointcutDec)
            return ((PointcutDec) node).getFormals();
        return getEnclosingFormals(node.getParent());
    }

    private FormalDec getEnclosingAfterFormal(ASTObject node) {
        if (node == null)
            return null;
        if (node instanceof AdviceDec)
            return ((AdviceDec) node).getExtraFormal();
        return getEnclosingAfterFormal(node.getParent());
    }

    private Pcd getEnclosingPcd(ASTObject node) {
        if (node == null)
            return null;
        if (node instanceof Pcd)
            return (Pcd) node;
        return getEnclosingPcd(node.getParent());
    }

    private Scope savedScope;

    //INTRO from ScopeWalker

    /* Specializing on three cases:
       .  STAR
       .  A single name that is a varref of a pointcut formal
       .  A single name that will be matched to a type
       .  otherwise, it's a starred pattern
    
       The specialization is used for:
       .  lint checking now
       .  performance optimization of matching later in the compiler
     */
    public ASTObject postScope(ScopeWalker walker) {
        //System.out.println(this + ", " + walker.getScope());
        if (savedScope != null)
            return this;

        savedScope = walker.getScope();

        // if we have nothing but one star
        if (names.size() == 1 && names.get(0).isStar() && dimensions == 0) {
            return new AnyTypeName(getSourceLocation()).setSource(this);
        }

        // if we're a formal
        if (names.size() == 1 && dimensions == 0) {
            Formals formals = getEnclosingFormals(this);
            String id = names.get(0).getSimpleName();
            if (formals != null) {
                FormalDec formalDec = formals.findName(id);
                if (formalDec != null) {
                    return makeVarTypeName(formalDec);
                } else {
                    FormalDec afterFormal = getEnclosingAfterFormal(this);
                    if (afterFormal != null && afterFormal.getId().equals(id)) {
                        getCompiler().showError(
                            this,
                            "can't refer to '" + afterFormal.toShortString() + "' from this context");
                        return makeSimpleTypeName(getTypeManager().TYPE_NOT_FOUND);
                    }
                }
            }
        }

        // if we're a primitive type name
        if (names.size() == 1) {
            String id = names.get(0).getSimpleName();
            if (id != null) {
                Type primType = getTypeManager().findPrimitiveType(id);
                if (primType != null) {
                    return makeSimpleTypeName(primType);
                }
            }
        }

        return this;
    }

    private SimpleTypeName makeSimpleTypeName(Type t) {
        for (int i = 0; i < dimensions; i++) {
            t = t.getArrayType();
        }
        return new SimpleTypeName(getSourceLocation(), t, includeSubTypes);
    }

    /** returns true if any visible type name matches this pattern.  
     * 
     * Note that this doesn't care about subtyping, since we only care about existence.
     * 
     * @see TypeManager#getAllVisibleTypeNames()
     */
    private boolean anyTypeMatches() {
        // as a performance improvement, first check loaded types
        for (Iterator i = getTypeManager().getLoadedTypes().iterator(); i.hasNext();) {
            Type t = (Type) i.next();
            if (this.matches(null, t.getNamePieces()))
                return true;
        }

        //System.out.println(this.toShortString() + " not in loaded");

        for (Iterator i = getTypeManager().getAllVisibleTypeNames().iterator();
            i.hasNext();
            ) {
            List pieces = (List) i.next();
            if (this.matches(null, pieces))
                return true;
        }
        return false;
    }

    /* 
     * 
     * @see ASTObject#checkSpec()
     */
    public void checkSpec() {
        Name cstName = names.makeName();

        // first check for non-wildcarded names
        if (cstName != null) {
            //XXX we need the parent to be set so that accessibility checks in scopes
            //XXX will treat the name as in my lexical scope
            //XXX this breaks an important invariant that parents
            //XXX take responsibility for their children
            cstName.setParent(this);

            Type t = cstName.resolveType(savedScope, false);
            if (t.isMissing()) {
                getCompiler().showWarning(this, ErrorHandler.invalidAbsoluteTypeName);
                super.checkSpec();
                return;
            }
            SimpleTypeName s = makeSimpleTypeName(t);
            this.replaceWith(s);
            s.checkSpec();
            return;
        } else {
            // check for wildcarded names matching anything
            if (ErrorHandler.invalidWildcardTypeName.shouldDisplay(getCompiler())) {
                if (!anyTypeMatches()) {
                    getCompiler().showWarning(this, ErrorHandler.invalidWildcardTypeName);
                }
            }
        }
        super.checkSpec();
    }

    protected GenTypeName makeVarTypeName(FormalDec formalDec) {
        formalDec.isBound = true;

        if (includeSubTypes || getParent() instanceof GenTypeName) {
            showError("formal binding can not be done as part of a GTN");
            return this;
        }

        if (!getEnclosingPcd(this).allowsNameBinding()) {
            showError("name binding is only allowed in target, this, and args PCDs");
            return this;
        }
        return new VarTypeName(getSourceLocation(), formalDec);
    }

    public ASTObject postCopy(CopyWalker walker, ASTObject oldObject) {
        savedScope = ((PatternTypeName) oldObject).savedScope;
        return this;
    }

    public FuzzyBoolean matchesInstance(Type type) {
        FuzzyBoolean ret = (FuzzyBoolean) instanceMatchCache.get(type);

        if (ret != null)
            return ret;

        Set tests = new HashSet();
        boolean isMatch = matches(null, type, false, tests);
        if (!isMatch) {
            ret = FuzzyBoolean.NO;
        } else if (tests.size() == 0) {
            ret = FuzzyBoolean.YES;
        } else {
            ret = FuzzyBoolean.MAYBE;
        }

        instanceMatchCache.put(type, ret);
        
        //System.out.println(type + " instance match " + ret);
        return ret;
    }

    void addInstanceTests(Set instanceTests, Expr expr, JpPlan plan) {
        if (instanceTests.size() == 0)
            return;

        instanceTests = Type.filterTopTypes(instanceTests);

        Expr testExpr = null;
        final AST ast = getAST();
        for (Iterator i = instanceTests.iterator(); i.hasNext();) {
            Type matchingType = (Type) i.next();
            if (expr.getType().isSubtypeOf(matchingType))
                continue;
            Expr instanceTest = ast.makeParen(ast.makeInstanceof(expr, matchingType));
            if (testExpr == null)
                testExpr = instanceTest;
            else
                testExpr = ast.makeBinop("||", testExpr, instanceTest);
        }
        if (testExpr != null)
            plan.addExprTest(testExpr);
    }

    public JpPlan makePlan(JoinPoint jp, Expr expr) {
        //System.out.println(">>>makePlan " + this);
        //XXX bad structure
        JpPlan plan = new JpPlan(jp);
        //        System.out.println(">>>>" + expr.getType());
        if (expr.getType() == null) {
            //System.out.println("bad expr: " + expr.unparse());
            return JpPlan.NO_PLAN;
        }
        Set instanceTests = new HashSet();
        if (matches(plan, expr.getType(), false, instanceTests)) {
            //System.out.println(this + " matches " + expr.getType() + " tests " + instanceTests);
            addInstanceTests(instanceTests, expr, plan);
            return plan;
        } else {
            //System.out.println("  NO match " + instanceTests);
            return JpPlan.NO_PLAN;
        }
    }

    /**
     * asType -- the type to match
     * needsTest -- insert a test even if we DEFINATELY match
     *              always passed in false at the top
     * instanceofTests -- an accumulator
     */
    boolean matches(
        JpPlan plan,
        Type asType,
        boolean needsTest,
        Set instanceofTests)
   {
        Type matchingType = matches(asType, true);
        //System.out.println(asType + ", " + matchingType);
        if (matchingType != null) {
            if (needsTest) {
                instanceofTests.add(matchingType);
            }
            // bind???
            return true;
        }

        // otherwise we need to walk all subtypes ;-(
        boolean couldMatch = false;
        Collection subTypes = asType.getDirectSubTypes();
        //System.out.println(subTypes);

        if (subTypes == null)
            return false;

        for (Iterator i = subTypes.iterator(); i.hasNext();) {
            if (matches(plan, (Type) i.next(), true, instanceofTests))
                couldMatch = true;
        }

        return couldMatch;
    }

    public boolean matches(Type type) {
        return matches(type, includeSubTypes) != null;
    }

    public Type matches(Type type, boolean includeSubTypes) {
        //System.out.println("matches? " + type + ", " + names.toShortString());
        if (basicMatch(type))
            return type;

        if (includeSubTypes) {
            //System.out.println(type + " supers: " + type.getDirectSuperTypes());
            for (Iterator i = type.getDirectSuperTypes().iterator(); i.hasNext();) {
                Type tryType = (Type) i.next();
                Type matchType = matches(tryType, includeSubTypes);
                if (matchType != null)
                    return matchType;
            }
        }
        //System.out.println("    false");
        return null;
    }

    protected boolean basicMatch(Type type) {
        Boolean cachedResult = (Boolean) matchCache.get(type);
        if (cachedResult != null)
            return cachedResult.booleanValue();

        boolean ret = innerBasicMatch(type);
        matchCache.put(type, ret ? Boolean.TRUE : Boolean.FALSE);
        return ret;
    }

    private boolean innerBasicMatch(Type type) {
        // first check that dimensions match
        //System.out.println(dimensions, type.getArrayDimCount());
        if (dimensions != type.getArrayDimCount()) {
            if (dimensions == 0)
                type = getTypeManager().getObjectType();
            else
                return false;
        }

        // by doing this match backwards we finesse some efficiency issues
        List pieces = type.getBaseComponentType().getNamePieces();
        //System.out.println(pieces);
        if (!matches(type.getBaseComponentType(), pieces)) {
            return false;
        }
        return true;
    }

    private boolean isEllipses(NamePattern name) {
        return name instanceof EllipsesFakeNamePattern;
    }

    protected boolean matches(Type type, List pieces) {
        return matches(type, pieces, pieces.size() - 1, names.size() - 1);
    }

    protected boolean matches(
        Type type,
        List pieces,
        int pieceIndex,
        int nameIndex) {
        if (nameIndex < 0) {
            // we need to see if the last name matched is at the top level or not
            pieceIndex += 1;
            // we've matched to the "real" top level
            if (pieceIndex == 0)
                return true;

            // otherwise let's see if this is a legal unqualified name
            // and that it binds to the correct type based on lexical scope
            if (savedScope == null) {
                showError("no saved scope");
            }

            Type importType = savedScope.findType((String) pieces.get(pieceIndex), this);

            if (type == null)
                return true;
            return importType == type;
        }

        if (pieceIndex < 0)
            return false;

        NamePattern name = names.get(nameIndex);
        if (isEllipses(name)) {
            for (; pieceIndex >= 0; pieceIndex--) {
                // no need for ambiguity handling since we don't do binding
                if (matches(type, pieces, pieceIndex, nameIndex - 1))
                    return true;
            }
            return false;
        } else {
            if (name.matches((String) pieces.get(pieceIndex))) {
                return matches(type, pieces, pieceIndex - 1, nameIndex - 1);
            } else {
                return false;
            }
        }
    }

    //BEGIN: Generated from @child and @property
    protected NamePatterns names;
    public NamePatterns getNames() {
        return names;
    }
    public void setNames(NamePatterns _names) {
        names = _names;
    }

    protected boolean includeSubTypes;
    public boolean getIncludeSubTypes() {
        return includeSubTypes;
    }
    public void setIncludeSubTypes(boolean _includeSubTypes) {
        includeSubTypes = _includeSubTypes;
    }

    protected int dimensions;
    public int getDimensions() {
        return dimensions;
    }
    public void setDimensions(int _dimensions) {
        dimensions = _dimensions;
    }

    public PatternTypeName(
        SourceLocation location,
        NamePatterns _names,
        boolean _includeSubTypes,
        int _dimensions) {
        super(location);
        setNames(_names);
        setIncludeSubTypes(_includeSubTypes);
        setDimensions(_dimensions);
    }
    protected PatternTypeName(SourceLocation source) {
        super(source);
    }

    public ASTObject copyWalk(CopyWalker walker) {
        PatternTypeName ret = new PatternTypeName(getSourceLocation());
        ret.preCopy(walker, this);
        ret.names = names;
        ret.includeSubTypes = includeSubTypes;
        ret.dimensions = dimensions;
        return ret;
    }

    public String getDefaultDisplayName() {
        return "PatternTypeName(names: "
            + names
            + ", "
            + "includeSubTypes: "
            + includeSubTypes
            + ", "
            + "dimensions: "
            + dimensions
            + ")";
    }

    //END: Generated from @child and @property

}