/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import java.util.*;
import org.aspectj.util.*;

/**
 * @grammar per[this|target](pointcut)
 * @child Pcd pcd
 * @property boolean onThis
 */

public class PerObject extends PerClause {
    public String toShortString() {
        return "per" + (onThis ? "this" : "target") +
                "(" + getPcd().toShortString() + ")";
    }
    
    public JpPlanner makeInnerPlanner(PlanData planData) {
        return new JpPlanner() {
            public FastMatch fastMatch(JoinPoint jp) {
                //??? might want to include a better filter test here as an optimization
                return MAYBE;
            }
            public JpPlan makePlan(JoinPoint jp) {
                Expr onExpr = makeOnExpr(jp);
                if (onExpr == null) return JpPlan.NO_PLAN;
                
                JpPlan plan = new JpPlan(jp);
                plan.test = testMapExpr(makeOnExpr(jp));
                plan.setInstanceExpr(getAspectOfExpr(makeOnExpr(jp)));
                return plan;
            }
        };
    }
    
    public JpPlanner makeInitializerPlanner(PlanData planData) {
        return new WrappedJpPlanner(getPcd().makePlanner(planData)) {
                public JpPlan makePlan(JoinPoint jp) {
                    if (makeOnExpr(jp) == null) return JpPlan.NO_PLAN;
                    JpPlan plan = super.makePlan(jp);
                    return new EachObjectPlan(plan, PerObject.this);
                }  
            };
    }

    public Expr makeOnExpr(JoinPoint point) {
        if (onThis) return point.makeThisExpr();
        else return point.makeTargetExpr();
    }

    
    private FieldDec map;
    FieldDec getMapField() {
        if (map != null) return map;

        final AST ast = getAST();
        String name = "permap$ajc";
        Type type = getTypeManager().getType("org.aspectj.runtime.internal", "PerObjectMap");

        Modifiers modifiers =
            ast.makeModifiers(Modifiers.PUBLIC | Modifiers.STATIC | Modifiers.FINAL);
        map = ast.makeField(modifiers, type, name, ast.makeNew(type));
        //stack.setSynthetic();        
        this.getAspectDec().addToBody(map);

        return map;
    }
    
    private MethodDec bindMethod;
    MethodDec getBindMethod() {
        if (bindMethod != null) return bindMethod;
        /* synchronized public static void bindToAspect(Object toObject) {
               if (!permap$ajc.hasAspect(toObject)) {
                   permap$ajc.bind(new A(), toObject);
               }
           }*/
        final AST ast = getAST();
        
        FormalDec toObjectFormal = ast.makeFormal(getTypeManager().getObjectType(),
                "toObject");
        
        BlockStmt body = ast.makeBlock(
              ast.makeIf(
                ast.makeUnop("!", testMapExpr(ast.makeVar(toObjectFormal))),
                ast.makeBlock(
                  ast.makeCall(getMapVar(), "bind",
                                        ast.makeVar(toObjectFormal),
                                        ast.makeNew(getAspectType())))
               ));

        Modifiers modifiers =
          ast.makeModifiers(Modifiers.STATIC | Modifiers.PUBLIC | Modifiers.SYNCHRONIZED);

        bindMethod = ast.makeMethod(modifiers, getTypeManager().voidType,
            "bind$ajc", ast.makeFormals(toObjectFormal), body);
        this.getAspectDec().addToBody(bindMethod);
        return bindMethod;
    }

    Expr getMapVar() {
        return getAST().makeGet(getMapField());
    }

    Expr testMapExpr(Expr onExpr) {
        return getAST().makeCall(getMapVar(), "hasAspect", onExpr);
    }
    
    Expr bindExpr(Expr onExpr) {
        return getAST().makeStaticCall(getBindMethod(), onExpr);
    }
    
    Expr getAspectOfExpr(Expr onExpr) {
        final AST ast = getAST();
        return ast.makeParen(ast.forceCast(
                getAspectType(), ast.makeCall(getMapVar(), "aspectOf", onExpr)));
    }
    
    protected MethodDec makeHasAspectMethod() {
        /* public static boolean hasAspect(Object onObject) {
               return map.hasAspect(onObject);
           }
        */
        final AST ast = getAST();
        FormalDec onObjectFormal = ast.makeFormal(getTypeManager().getObjectType(),
                "onObject");
        
        BlockStmt body =
            ast.makeBlock(ast.makeReturn(testMapExpr(ast.makeVar(onObjectFormal))));
        return makeHasAspectMethod(ast.makeFormals(onObjectFormal), body);
    }

    protected MethodDec makeAspectOfMethod() {
        /* public static A aspectOf(Object onObject) {
               Object asp = map.aspectOf(onObject);
               if (asp != null) return asp;
               throw new org.aspectj.runtime.NoAspectException();
           }
        */
        final AST ast = getAST();
        FormalDec onObjectFormal = ast.makeFormal(getTypeManager().getObjectType(),
                "onObject");
        
        VarDec aspectDec = ast.makeVarDec(getAspectType(), "asp", 
                                getAspectOfExpr(ast.makeVar(onObjectFormal)));

        BlockStmt body = ast.makeBlock(
            aspectDec,
            ast.makeIf(ast.makeNonNullTest(ast.makeVar(aspectDec)),
                       ast.makeReturn(ast.makeVar(aspectDec))),
            ast.makeThrow(getTypeManager().getType("org.aspectj.lang", "NoAspectBoundException")));

        return makeAspectOfMethod(ast.makeFormals(onObjectFormal), body);
    }
    //public Expr makeAspectOfExpr(Expr fromObject) { return makeAspectInstance(fromObject); }


    /***********************************************************************************/

    // this is called by the EachObjectPlan for the "entry points"
    public Stmts wrapCheckAndSet(JpPlan plan, JoinPoint jp, Stmts body) {
        final AST ast = getAST();
        Expr onExpr = makeOnExpr(jp);
        if (onExpr == null) return body;
        
        Stmt stmt = ast.makeStmt(bindExpr(onExpr));
        stmt = plan.wrapDynamicTest(ast.makeBlock(stmt));
        
        body.add(0, stmt);
        return body;
    }    
    
    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    protected boolean onThis;
    public boolean getOnThis() { return onThis; }
    public void setOnThis(boolean _onThis) { onThis = _onThis; }
    
    public PerObject(SourceLocation location, Pcd _pcd, boolean _onThis) {
        super(location);
        setPcd(_pcd);
        setOnThis(_onThis);
    }
    protected PerObject(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        PerObject ret = new PerObject(getSourceLocation());
        ret.preCopy(walker, this);
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        ret.onThis = onThis;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "PerObject(onThis: "+onThis+")";
    }
    
    //END: Generated from @child and @property
}

