/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */
package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.*;
import org.aspectj.compiler.crosscuts.joinpoints.*;

import java.util.*;
import org.aspectj.util.*;

/**
 * @grammar per[this|target](pointcut)
 * @child Pcd pcd
 * @property boolean onThis
 */

public class PerThisOrTarget extends PerClause {
    public String toShortString() {
        return "per" + (onThis ? "this" : "target") +
                "(" + getPcd().toShortString() + ")";
    }
    
    private Set onTypes = new HashSet();
    
    public JpPlanner makeInnerPlanner(PlanData planData) {
        return new JpPlanner() {
            public FastMatch fastMatch(JoinPoint jp) {
                //??? might want to include a better filter test here as an optimization
                return MAYBE;
            }
            public JpPlan makePlan(JoinPoint jp) {
                Expr onExpr = makeOnExpr(jp);
                if (onExpr == null) return JpPlan.NO_PLAN;
                
                JpPlan plan = new JpPlan(jp);
                plan.test = getHasAspectExpr(makeOnExpr(jp));
                plan.setInstanceExpr(getAspectOfExpr(makeOnExpr(jp)));
                return plan;
            }
        };
    }
    
    public JpPlanner makeInitializerPlanner(PlanData planData) {
        return new WrappedJpPlanner(getPcd().makePlanner(planData)) {
                public JpPlan makePlan(JoinPoint jp) {
                    Expr onExpr = makeOnExpr(jp);
                    if (onExpr == null) return JpPlan.NO_PLAN;
                    // don't plan on me or my supers
                    if (getAspectType().isSubtypeOf(onExpr.getType())) {
                        getPcd().showWarning("will not match containing aspect (compiler limitation)");
                        return JpPlan.NO_PLAN;
                    }
                    
                    JpPlan plan = super.makePlan(jp);
                    
                    if (plan.isPossible()) {
                        addHasAspectType(onExpr.getType(), jp);
                    }
                    return new InitializerPlan(plan);
                }
            };
    }
    
    private class InitializerPlan extends JpPlan {
        JpPlan innerPlan;
        
        public InitializerPlan(JpPlan innerPlan) {
            super(innerPlan.joinPoint);
            this.innerPlan = innerPlan;
        }
        
        public int getPreSortOrder() { return EACHOBJECT; }
        
        public String toString() {
            return "perobject" + super.toString();
        }
        
        public void wrapJoinPoint(JoinPoint jp) {
            jp.setStmts(this.wrapCheckAndSet(jp, jp.getStmts()));
        }
        
        public Stmts wrapCheckAndSet(JoinPoint jp, Stmts stmts) {
            return PerThisOrTarget.this.wrapCheckAndSet(innerPlan, jp, stmts);
        }
    }
    
    void addHasAspectType(Type toType, JoinPoint jp) {
        if (onTypes.contains(toType)) return;
        if (toType == getHasAspectInterfaceType()) return;
        onTypes.add(toType);
        if (!toType.getTypeDec().fromSource()) {
            //??? should there be a compiler error or warning here
            return;
        }
        toType.getTypeDec().addSuperInterfaceType(getHasAspectInterfaceType());
        toType.addDirectSuperType(getHasAspectInterfaceType());
    }
    

        /*
    public void plan(TypeDec typeDec, int phase) {
        //??? is this too early
        if (phase != TypeDecPlanner.SIGNATURE) return;
        
        JoinPoint jp = ???;
        
        if (getPcd().matches(jp)) {
            
        }
        
        if (!getTargets().matches(typeDec.getType())) return;

        final int N = typeDs.size();
        for (int i = 0; i < N; i++) {
            TypeD name = typeDs.get(i);
            getCompiler().showMessage("    introducing " + name.toShortString() +
                    " on " + typeDec.toShortString());

            addSuperType(typeDec.getType(), name.getType());
        }
        ((AspectJCompiler)getCompiler()).getCorrespondences().addPointsTo(this, typeDec);
    }
    private Set plannedTypes = new HashSet();
    
    void planType(Type type) {
        if (plannedTypes.contains(type)) return;
        plannedTypes.add(type);
        if (type.isInterface()) return;
        
        //???
        if (type.has
    }
    */
    

    public Expr makeOnExpr(JoinPoint point) {
        if (onThis) return point.makeThisExpr();
        else return point.makeTargetExpr();
    }    
    
    /**
     * <code>
     * public static interface HasAspect {
     *     private transient AspectType _aspect;
     *     private synchronized void bind() {
     *         if (_aspect != null) return;
     *         _aspect = new AspectType();
     *     }
     *    
     *     private AspectType getAspect() {
     *         return _aspect;
     *     }
     * }
     * </code>
     */
    private InterfaceDec hasAspectInterface;
    private MethodDec hasAspectBindMethod;
    private MethodDec hasAspectGetMethod;
    
    public Type getHasAspectInterfaceType() {
        return getHasAspectInterface().getType();
    }
    
    public InterfaceDec getHasAspectInterface() {
        if (hasAspectInterface != null) return hasAspectInterface;
        hasAspectInterface = makeHasAspectInterface();
        return hasAspectInterface;
    }
    
    String makeName(String base) {
        return getAST().makeGeneratedName(getAspectType().getString() + "_" + base);
    }
    
    int getFieldModifiers() {
        if (getOptions().XserializableAspects) {
            if (getAspectType().isSubtypeOf(getTypeManager().getSerializableType())) {
                return Modifiers.PRIVATE;
            }
        }
        return Modifiers.PRIVATE | Modifiers.TRANSIENT;
    }

    InterfaceDec makeHasAspectInterface() {
        final AST ast = getAST();
        
        hasAspectInterface = ast.makeInterface(
            ast.makeModifiers(Modifiers.PUBLIC), "AJC_HasAspect", ast.makeTypeDs());
        hasAspectInterface.setEnclosingTypeDec(getAspectDec());
        getAspectDec().addToBody(hasAspectInterface);
        
        //System.out.println("package: " + hasAspectInterface.getPackageName());
        
        FieldDec aspectField = ast.makeField(
            ast.makeModifiers(getFieldModifiers()), getAspectType(), makeName("_aspect"));
        hasAspectInterface.getBody().add(aspectField);
        
        hasAspectBindMethod = ast.makeMethod(
            ast.makeModifiers(Modifiers.PRIVATE|Modifiers.SYNCHRONIZED), 
            getTypeManager().voidType, makeName("bind"),
            ast.makeFormals(),
            ast.makeBlock(
                ast.makeIf(
                        ast.makeNonNullTest(ast.makeGet(aspectField.getField())),
                        ast.makeReturn()),
                ast.makeSet(aspectField.getField(), ast.makeNew(getAspectType()))));
        hasAspectInterface.getBody().add(hasAspectBindMethod);
                
        hasAspectGetMethod = ast.makeMethod(
            ast.makeModifiers(Modifiers.PRIVATE), getAspectType(), makeName("getAspect"),
            ast.makeFormals(),
            ast.makeBlock(
                ast.makeReturn(ast.makeGet(aspectField.getField()))));        
        hasAspectInterface.getBody().add(hasAspectGetMethod);
        
        return hasAspectInterface;
    }

    /** <code>
     *  public static void bind$ajc(Object obj) {   
     *    if (!(obj instanceof HasAspect)) return; //XXX this seems deadly
     *    ((HasAspect)obj).bind();
     * }
     *
     *  public static boolean hasAspect(Object obj) {   
     *    if (!(obj instanceof HasAspect)) return false;
     *    return ((HasAspect)obj).getAspect() != null;
     * }
     *
     *  public static AspectType aspectOf(Object obj) {
     *    AspectType _aspect = null;
     *    if (obj instanceof HasAspect) {
     *        _aspect = ((HasAspect)obj).getAspect();
     *    }
     *    if (_aspect == null) {
     *        throw new org.aspectj.lang.NoAspectBoundException();
     *    }
     *    return _aspect;
     * }
     *
     * </code>
     */
    private MethodDec bindMethod;
    MethodDec getBindMethod() {
        if (bindMethod != null) return bindMethod;

        final AST ast = getAST();
        
        FormalDec objFormal = ast.makeFormal(getTypeManager().getObjectType(), "obj");
        bindMethod = ast.makeMethod(
            ast.makeModifiers(Modifiers.STATIC|Modifiers.PUBLIC|Modifiers.SYNCHRONIZED),
            getTypeManager().voidType, "bind$ajc", ast.makeFormals(objFormal),
            ast.makeBlock(
              ast.makeIf(
                ast.makeNotInstanceofTest(ast.makeVar(objFormal),
                                          getHasAspectInterfaceType()),
                ast.makeReturn()),
              ast.makeCall(hasAspectBindMethod, 
                ast.makeCast(getHasAspectInterfaceType(), ast.makeVar(objFormal)))));
        getAspectDec().addToBody(bindMethod);
        return bindMethod;
    }
    
    protected MethodDec makeHasAspectMethod() {
        final AST ast = getAST();
        FormalDec objFormal = ast.makeFormal(getTypeManager().getObjectType(), "obj");
        
        BlockStmt body = ast.makeBlock(
            ast.makeIf(
              ast.makeNotInstanceofTest(ast.makeVar(objFormal), getHasAspectInterfaceType()),
              ast.makeReturn(ast.makeLiteral(false))),
            ast.makeReturn(ast.makeNonNullTest(ast.makeCall(hasAspectGetMethod, 
                ast.makeCast(getHasAspectInterfaceType(), ast.makeVar(objFormal))))));
        return makeHasAspectMethod(ast.makeFormals(objFormal), body);
    }

    protected MethodDec makeAspectOfMethod() {
        getHasAspectInterfaceType();
        
        final AST ast = getAST();
        VarDec aspDec = ast.makeVarDec(getAspectType(), "asp", ast.makeNull());
        FormalDec objFormal = ast.makeFormal(getTypeManager().getObjectType(), "obj");
        
        BlockStmt body = ast.makeBlock(
            aspDec,
            ast.makeIf(
              ast.makeInstanceof(ast.makeVar(objFormal), getHasAspectInterfaceType()),
              ast.makeBlock(
                ast.makeSet(aspDec, ast.makeCall(hasAspectGetMethod, 
                  ast.makeCast(getHasAspectInterfaceType(), ast.makeVar(objFormal)))))),
            ast.makeIf(
              ast.makeNullTest(ast.makeVar(aspDec)),
              ast.makeBlock(
                ast.makeThrow(ast.makeNew(
                    getTypeManager().getType("org.aspectj.lang", "NoAspectBoundException"))))),
            ast.makeReturn(ast.makeVar(aspDec)));

        return makeAspectOfMethod(ast.makeFormals(objFormal), body);
    }
    
    Expr bindExpr(Expr onExpr) {
        return getAST().makeStaticCall(getBindMethod(), onExpr);
    }
    
    Expr getHasAspectExpr(Expr onExpr) {
        final AST ast = getAST();
        return ast.makeStaticCall(hasAspectMethod, onExpr);
    }
    
    Expr getAspectOfExpr(Expr onExpr) {
        final AST ast = getAST();
        return ast.makeParen(ast.forceCast(
                getAspectType(), ast.makeStaticCall(aspectOfMethod, onExpr)));
    }

    /***********************************************************************************/

    // this is called by the EachObjectPlan for the "entry points"
    public Stmts wrapCheckAndSet(JpPlan plan, JoinPoint jp, Stmts body) {
        final AST ast = getAST();
        Expr onExpr = makeOnExpr(jp);
        if (onExpr == null) return body;
        
        Stmt stmt = ast.makeStmt(bindExpr(onExpr));
        stmt = plan.wrapDynamicTest(ast.makeBlock(stmt));
        
        body.add(0, stmt);
        return body;
    }    
    
    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    protected boolean onThis;
    public boolean getOnThis() { return onThis; }
    public void setOnThis(boolean _onThis) { onThis = _onThis; }
    
    public PerThisOrTarget(SourceLocation location, Pcd _pcd, boolean _onThis) {
        super(location);
        setPcd(_pcd);
        setOnThis(_onThis);
    }
    protected PerThisOrTarget(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        PerThisOrTarget ret = new PerThisOrTarget(getSourceLocation());
        ret.preCopy(walker, this);
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        ret.onThis = onThis;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return pcd;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "pcd";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setPcd((Pcd)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 1;
    }
    
    public String getDefaultDisplayName() {
        return "PerThisOrTarget(onThis: "+onThis+")";
    }
    
    //END: Generated from @child and @property
}

