/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.ast;

import org.aspectj.compiler.crosscuts.joinpoints.*;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.base.cst.*;

import org.aspectj.compiler.base.JavaCompiler;
import org.aspectj.compiler.crosscuts.AspectJCompiler;

/**
 * @grammar declare error|warning: pointcut: message;
 * @child Pcd pcd
 * @child Expr message
 * @property boolean isWarning
 */

public class ShowErrorDec extends Dec implements JpPlannerMaker {

    public Modifiers getModifiers() { return getAST().makeModifiers(0); }
    
    public String getId() { return "declare"; }
    public String toShortString() {
        return getKind() + ": " + pcd.toShortString();
    }
    public String getKind() {
        return "declare " + (isWarning ? "warning" : "error");
    }


    public void checkSpec() {
        pcd.checkStatic();
        
        if (! (message instanceof StringLiteralExpr) ) {
            message.showError("error message must be a constant string");
        }
    }

    public JpPlanner makePlanner(PlanData planData) {
        return new ErrorPlanner(getPcd().makePlanner(planData));
    }
        
    private String getMessageString() {
        if (! (message instanceof StringLiteralExpr)) return "";
        return ((StringLiteralExpr)message).getStringValue();
    }
    
    private class ErrorPlanner extends WrappedJpPlanner {
        public ErrorPlanner(JpPlanner planner) { super(planner); }

        public JpPlan makePlan(JoinPoint jp) {
            //??? this is controversial design
            if (jp.getSourceLocation() == null) return JpPlan.NO_PLAN;
            if (alwaysMatches(jp)) {
                showErrorAt(jp);
            }
            return JpPlan.NO_PLAN;
        }

        public boolean isStaticPlanner() { return true; }
    }
    
    public void showErrorAt(JoinPoint point) {
        if (isWarning) {
            getCompiler().showWarning(point.getSourceLocation(), getMessageString());
        } else {
            getCompiler().showError(point.getSourceLocation(), getMessageString());
        }
    }    
    
    //BEGIN: Generated from @child and @property
    protected Pcd pcd;
    public Pcd getPcd() { return pcd; }
    public void setPcd(Pcd _pcd) {
        if (_pcd != null) _pcd.setParent(this);
        pcd = _pcd;
    }
    
    protected Expr message;
    public Expr getMessage() { return message; }
    public void setMessage(Expr _message) {
        if (_message != null) _message.setParent(this);
        message = _message;
    }
    
    protected boolean isWarning;
    public boolean getIsWarning() { return isWarning; }
    public void setIsWarning(boolean _isWarning) { isWarning = _isWarning; }
    
    public ShowErrorDec(SourceLocation location, Pcd _pcd, Expr _message, boolean _isWarning) {
        super(location);
        setPcd(_pcd);
        setMessage(_message);
        setIsWarning(_isWarning);
    }
    protected ShowErrorDec(SourceLocation source) {
        super(source);
    }
    
    public ASTObject copyWalk(CopyWalker walker) {
        ShowErrorDec ret = new ShowErrorDec(getSourceLocation());
        ret.preCopy(walker, this);
        if (pcd != null) ret.setPcd( (Pcd)walker.process(pcd) );
        if (message != null) ret.setMessage( (Expr)walker.process(message) );
        ret.isWarning = isWarning;
        return ret;
    }
    
    public ASTObject getChildAt(int childIndex) {
        switch(childIndex) {
        case 0: return pcd;
        case 1: return message;
        default: return super.getChildAt(childIndex);
        }
    }
     public String getChildNameAt(int childIndex) {
        switch(childIndex) {
        case 0: return "pcd";
        case 1: return "message";
        default: return super.getChildNameAt(childIndex);
        }
    }
     public void setChildAt(int childIndex, ASTObject child) {
        switch(childIndex) {
        case 0: setPcd((Pcd)child); return;
        case 1: setMessage((Expr)child); return;
        default: super.setChildAt(childIndex, child); return;
        }
    }
     public int getChildCount() {
        return 2;
    }
    
    public String getDefaultDisplayName() {
        return "ShowErrorDec(isWarning: "+isWarning+")";
    }
    
    //END: Generated from @child and @property
}



