/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.base.ast.*;
import org.aspectj.compiler.crosscuts.ast.*;

import java.util.*;

public class CalleeSideCallJp extends CodeDecJp {
    TypeDec typeDec;
    public CalleeSideCallJp(TypeDec typeDec, MethodDec methodDec) {
        super(methodDec);
        this.typeDec = typeDec;
    }

    public Kind getKind() { return MethodCallJp.KIND; }
    
    public void addPlan(JpPlan plan) {
        if (plan instanceof AdvicePlan) {
            AdvicePlan advicePlan = (AdvicePlan)plan;
            if (advicePlan.getAdviceDec() != null &&
                advicePlan.getAdviceDec().needsStaticEnclosingJoinPointFormal())
            {
                getCompiler().showMessage("    skip plan on " + this + " needs enclosing join point");
                return;
            }
        }
        plans.add(plan);
    }

    public String toString() {
        return super.toString() + "*";
    }

    public Type getBytecodeType() {
        return getCodeDec().getBytecodeType();
    }

    public Type getDeclaringType() {
        return getCodeDec().getDeclaringType();
    }
    
    public Expr makeTargetExpr() {
        return super.makeThisExpr();
    }

    public Type makeThisExprType() { return null; }
    public Expr makeThisExpr() {
        return null;
    }
    public Exprs makeArgsExprs() {
        return getCodeFormals().makeExprs();
    }

    public ASTObject getSourceLocation() { return null; }
    public ASTObject getTargetNode() { return getCodeDec(); }
    public Type getTargetType() {
        return getRootDeclaringType(getTargetSO(), getDeclaringType());
    }

    Type getRootDeclaringType(SemanticObject so, Type inType) {
        SemanticObject matchingSO = inType.findMatchingSemanticObject(so);
        if (matchingSO == null) return null;
        
        if (so.isStatic()) {
            if (matchingSO != so) return null;
        }

        //??? for fields, static methods and constructors this can probably be optimized
        //??? to only go up the class hierarchy ignoring interfaces
        for (Iterator i = inType.getDirectSuperTypes().iterator(); i.hasNext(); ) {
            Type ret = getRootDeclaringType(so, (Type)i.next());
            //!!! this doesn't handle ambiguity at all
            if (ret != null) return ret;
        }

        return inType;
    }    

    public boolean isSynthetic() {
        return getCodeDec().getDeclaringType() != typeDec.getType();
    }

    public MethodDec getBaseMethodDec() {
        if (isSynthetic()) {
            //methodDec = getMethodDec().makeSuperCaller();
        }
        return (MethodDec)getCodeDec();
    }
    
    public String getPostDispatchMethodName() {
        return getAST().makeGeneratedName(getCodeDec().getId() + "$ajcPostCall");
    }
    
    private Formals codeFormals = null;
    public Formals getCodeFormals() {
        if (codeFormals == null) {
            codeFormals = (Formals)getBaseMethodDec().getFormals().copy();
        }
        return codeFormals;
    }

    
    public MethodDec getPostMethodDec() {
        if (postDispatchMethod == null) {
            Formals formals = getBaseMethodDec().getFormals();
            Type resultType = getBaseMethodDec().getResultType();
            String id = getPostDispatchMethodName();
            Modifiers modifiers = (Modifiers)getBaseMethodDec().getModifiers().copy();
            postDispatchMethod =
                getAST().makeMethod(modifiers, resultType, id, formals, null);
            TypeDs _throws = getBaseMethodDec().getThrows();
            if (_throws != null) {
                postDispatchMethod.setThrows((TypeDs)_throws.copy());
            }
            postDispatchMethod.owner = getBaseMethodDec();
            getBaseMethodDec().getBytecodeTypeDec().getBody().add(postDispatchMethod);
        }
        return postDispatchMethod;
    }
        
    
    MethodDec postDispatchMethod = null;
    boolean madeInnerMethod = false;
    public Stmts getStmts() {
        if (!madeInnerMethod) {
            madeInnerMethod = true;
            CodeBody body = getBaseMethodDec().getBody();
            Type resultType = getBaseMethodDec().getResultType();
            
            getBaseMethodDec().setBody(null);
            
            getPostMethodDec();
            
            getBaseMethodDec().setFormals(getCodeFormals());
            getBaseMethodDec().setResultTypeD(resultType.makeTypeD());
            

            postDispatchMethod.setBody(body);
            
            Exprs args = getCodeFormals().makeExprs();
            Expr thisExpr = makeThisExprOrType();
            Expr callExpr = getAST().makeCall(postDispatchMethod, thisExpr, args);
            callExpr.setSourceLocation(getTargetNode().getSourceLocation());
            Stmt stmt = null;
            if (resultType.isVoid()) {
                stmt = getAST().makeStmt(callExpr);
            } else {
                stmt = getAST().makeReturn(callExpr);
            }
            getBaseMethodDec().setBody(getAST().makeBlock(stmt));
        }
        return getBaseMethodDec().getBody().getStmts();
    }
}
