/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the compiler and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */

package org.aspectj.compiler.crosscuts.joinpoints;

import org.aspectj.compiler.crosscuts.ast.*;
import org.aspectj.compiler.base.ast.*;

import org.aspectj.compiler.base.CompilerObject;
import org.aspectj.compiler.base.JavaCompiler;

import java.util.*;

/**
 */

public class InstanceTest extends CompilerObject {
    private List andBlocks = new ArrayList();

    private InstanceTest(JavaCompiler compiler) {
        this(compiler,new ArrayList());
    }

    private InstanceTest(JavaCompiler compiler, Type type, boolean isTrue) {
        this(compiler);
        addOr(type, isTrue);
        //System.out.println("new InstanceTest(" + type + ": " + isTrue);
    }

    private InstanceTest(JavaCompiler compiler, List andBlocks) {
        super(compiler);
        this.andBlocks = andBlocks;
    }

    public int size() { return andBlocks.size(); }

    public String toString() {
        StringBuffer buf = new StringBuffer("ITest(");
        for (int i=0; i < andBlocks.size(); i++) {
            AndBlock block1 = (AndBlock)andBlocks.get(i);
            if (i > 0) buf.append("  || ");
            buf.append(block1.toString());
        }
        buf.append(")");
        return buf.toString();
    }


    public void addOr(Type typeDec, boolean isTrue) {
        //System.out.println("add: " + typeDec + " to " + andBlocks.size());
        andBlocks.add(new AndBlock(getCompiler(),typeDec, isTrue));
    }

    public InstanceTest intersect(InstanceTest other) {
        List newBlocks = new ArrayList();
        for (int i=0; i < andBlocks.size(); i++) {
            AndBlock block1 = (AndBlock)andBlocks.get(i);
            for (int j=0; j < other.andBlocks.size(); j++) {
                AndBlock block2 = (AndBlock)other.andBlocks.get(j);
                newBlocks.add(block1.intersect(block2));
            }
        }
        return new InstanceTest(getCompiler(),newBlocks);
    }

    public InstanceTest union(InstanceTest other) {
        List newBlocks = new ArrayList(andBlocks);
        newBlocks.addAll(other.andBlocks);
        return new InstanceTest(getCompiler(),newBlocks);
    }

    public InstanceTest complement() {
        InstanceTest ret = null;
        for (int i=0; i < andBlocks.size(); i++) {
            AndBlock block = (AndBlock)andBlocks.get(i);
            InstanceTest test = block.complement();
            if (ret == null) {
                ret = test;
            } else {
                ret = ret.intersect(test);
            }
        }
        return ret;
    }

    private void reduce(Type staticType) {
        if (staticType == null) {
            System.out.println("troubled..." + this);
            return;
        }

        List newBlocks = new ArrayList();

        for (int i=0; i < andBlocks.size(); i++) {
            AndBlock block = (AndBlock)andBlocks.get(i);
            block = block.reduce(staticType);
            if (block != null) {
                newBlocks.add(block);
            }
        }

        this.andBlocks = newBlocks;
    }

    public Expr makeExpr(Expr thisExpr) {
        Type type = thisExpr.getType();
        //System.out.println("  instance test: " + type + ", " + this);

        reduce(type);

        //System.out.println("      -> " + this);


        Expr ret = null;
        for (int i=0; i < andBlocks.size(); i++) {
            AndBlock block = (AndBlock)andBlocks.get(i);
            Expr blockExpr = block.makeExpr(thisExpr);
            if (ret == null) {
                ret = blockExpr;
            } else {
                ret = getAST().makeBinop("||", ret, blockExpr);
            }
        }
        return ret;
    }

    private static class AndBlock extends CompilerObject {
        private List trueTypes = new ArrayList();
        private List falseTypes = new ArrayList();

        public AndBlock(JavaCompiler compiler, Type typeDec, boolean isTrue) {
            super(compiler);
            if (isTrue) {
                trueTypes.add(typeDec);
            } else {
                falseTypes.add(typeDec);
            }
        }

        public String toString() {
            StringBuffer buf = new StringBuffer();
            int i;
            for (i = 0; i < trueTypes.size(); i++) {
                if (i > 0) buf.append(" && ");
                buf.append(((Type)trueTypes.get(i)).toShortString());
            }

            for (i = 0; i < falseTypes.size(); i++) {
                if (i > 0 || trueTypes.size() > 0) buf.append(" && ");
                buf.append("!");
                buf.append(((Type)falseTypes.get(i)).toShortString());
            }
            return buf.toString();
        }

        public InstanceTest complement() {
            List newBlocks = new ArrayList(trueTypes.size() + falseTypes.size());
            int i;
            for (i = 0; i < trueTypes.size(); i++) {
                newBlocks.add(new AndBlock(getCompiler(),(Type)trueTypes.get(i), false));
            }

            for (i = 0; i < falseTypes.size(); i++) {
                newBlocks.add(new AndBlock(getCompiler(),(Type)falseTypes.get(i), true));
            }
            return new InstanceTest(getCompiler(),newBlocks);
        }

        public AndBlock intersect(AndBlock other) {
            trueTypes.addAll(other.trueTypes);
            falseTypes.addAll(other.falseTypes);
            //XXX destructive
            return this;
        }

        public AndBlock reduce(Type staticType) {
            {
                for (Iterator i = trueTypes.iterator(); i.hasNext(); ) {
                    Type trueType = (Type)i.next();
                    if (staticType.isSubtypeOf(trueType)) i.remove();
                    //??? add a check that the other way is possible
                }
            }

            {
                for (Iterator i = falseTypes.iterator(); i.hasNext(); ) {
                    Type falseType = (Type)i.next();
                    if (!falseType.isSubtypeOf(staticType) &&
                        !staticType.isSubtypeOf(falseType))
                    {
                        i.remove();
                    }
                }
            }

            if (trueTypes.size() == 0 && falseTypes.size() == 0) return null;
            //??? destructive
            return this;
        }

        private Expr makeInstanceOfExpr(Expr expr, Type checkType) {
            return getAST().makeParen(getAST().makeInstanceof(expr, checkType));
        }

        private Expr intersectExpr(Expr e1, Expr e2) {
            if (e1 == null) return e2;
            return getAST().makeBinop("&&", e1, e2);
        }


        public Expr makeExpr(Expr thisExpr) {
            Expr ret = null;
            int i;
            for (i = 0; i < trueTypes.size(); i++) {
                Expr testExpr = makeInstanceOfExpr(thisExpr, (Type)trueTypes.get(i));
                ret = intersectExpr(ret, testExpr);
            }

            for (i = 0; i < falseTypes.size(); i++) {
                Expr testExpr = makeInstanceOfExpr(thisExpr, (Type)falseTypes.get(i));
                testExpr = getAST().makeUnop("!", testExpr);
                ret = intersectExpr(ret, testExpr);
            }

            if (ret instanceof UnopExpr) return ret;
            return getAST().makeParen(ret);
        }
    }
}
