/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the IDE support for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 *
 * Contributor(s):
 */


package org.aspectj.util;

import java.util.*;
import java.io.*;

/**
 * @deprecated  use <CODE>ConfigParser</CODE> instead.
 */
public class ConfigFileUtil {

	/** Represents the current working dir - per System.properties("user.dir") */
	private static final File USERDIR;
	/** Represents the "." dir - used for USERDIR when some Throwable prevents its initialization */
	private static final File DOTDIR;
	/** Filter that accepts any file name ending with ".java" */
	private static final FileFilter SOURCEFILTER;
	static {
		File userDir = null;
		File dotDir = null;
		try {
			dotDir = new File(".");
			userDir = new File(System.getProperty("user.dir")).getCanonicalFile();
		} catch (Throwable t) {
			if (null == userDir) userDir = dotDir; // still might be null...
		} finally {
			DOTDIR = dotDir;
			USERDIR = userDir;
		}
		SOURCEFILTER = new FileFilter() {
				public boolean accept(File f) {
					return ((f != null) &&
						(f.getName().endsWith(".java")));
				}
			};
	}

    /**
     * Retrieves all of the .lst files in a directory, replacing any '\' characters with '/'.
     *
     * @returns an empty List if no files were found
     *
     * @deprecated  use <CODE>ConfigParser</CODE> instead.
     */
    public static List getLstFilesInDir(String dirPath) { // todo: why not use a FileFilter?
        List configs = new ArrayList();
        File f = new File(dirPath);
        File[] dirContents = f.listFiles();
        for (int j = 0; j < dirContents.length; j++) {
            if (dirContents[j].isDirectory()) {
                configs.addAll(getLstFilesInDir(dirContents[j].getAbsolutePath()));
            } else if (dirContents[j].getName().endsWith(".lst")) {
                configs.add(dirContents[j].getAbsolutePath().replace('\\', '/'));
            }
        }
        return configs;
    }

    /**
     * Render filename as a canonical absolute File, using relativeDir if necessary.
     * Does guarantee that the result is valid using the expensive
     * <code>{File}.getCanonicalFile()</code>.
	 * If the filename is not absolute, then resolve using currentWorkingDir.
	 * If currentWorkingDir is null, then use
	 *  <code>new File(System.getProperties("user.dir"))</code>.
     * See also org.aspectj.tools.ajc.Main.qualifiedFile(..).
     * @param filename the String to render as a qualified File (absolute preferred)
	 * @param currentWorkingDir the File directory base for the name (may be null)
     * @return the File represented by filename, rendered absolute and canonical
	 * @throws IOException if the input cannot represent a valid file on this filesystem
     *
     * @deprecated  use <CODE>ConfigParser</CODE> instead.
     */
    public static File qualifiedFile(String name, File currentWorkingDir)
		throws IOException {
        name = name.replace('/', File.separatorChar); // uncertain how this helps...
        File file = new File(name);
        if (!file.isAbsolute()) {
			if (null == currentWorkingDir) {
				currentWorkingDir = USERDIR;
			}
			file = new File(currentWorkingDir, name).getAbsoluteFile();
		}
        return file.getCanonicalFile();
    }

    /**
	 * Same as <code>expandLstFile(filename, null)<code>
	 * @see #expandLstFile(String, File)
     * @param filename must be resolvable by qualifiedFile(filename, null)
     *        i.e., either absolute or relative to user.dir
	 *              System property (current working dir)
     *
     * @deprecated  use <CODE>ConfigParser</CODE> instead.
     */
    public static List expandLstFile(String filename) throws IOException, LstFileEntryInvalidException {
        return expandLstFile(filename, null);
    }

    /**
	 * Create the list of java files reachable via the input filename,
	 * taken as a file listing (per line) .java files or other list files
	 * either of which may be absolute or relative to the parent directory
	 * of the containing list file.
	 * Prefer this API over <code>expandLstFile(String)</code>.
     * @param filename must be resolvable by qualifiedFile(filename, relativeTo)
     *        i.e., either absolute or relative to "."
	 * @return list of files reachable from filename
	 * @throws IOException same as expandAtFile
	 * @throws LstFileEntryInvalidException same as expandAtFile
     *
     * @deprecated  use <CODE>ConfigParser</CODE> instead.
     */
    public static List expandLstFile(String filename, File relativeTo)
        throws IOException, LstFileEntryInvalidException {
        List fileContents = new ArrayList();
        expandAtFile(qualifiedFile(filename, relativeTo), fileContents, null);
        return fileContents;
    }

    /**
	 * Implement <code>List expandLstFile(String, File)</code>.
     * This differs from org.aspectj.tools.ajc.Main in not relying on
	 * a parseArgs(..) method to do recursive calls.
     * Instead it tracks in atFiles the stack of atFiles called,
     * rendering each in absolute canonical form and interning the string
     * to detect differences in naming.
	 * @param file the File to expand - should be readable and contain list
	 * @param args the List of files reachable from earlier calls (results go here)
	 *             - not null
	 * @param atFiles the List of list files traversed
	 *         - should be null from any external caller unless you want to exclude a file
     */
    private static void expandAtFile(File file, List args, List atFiles)
		throws IOException, LstFileEntryInvalidException {
        BufferedReader in = new BufferedReader(new FileReader(file));
        final File parentDir = (null == file? null : file.getParentFile());
        String line;
        while ((line = in.readLine()) != null) {
            if (line == null || line.length() < 1) continue;
            line = line.trim();
            if (line.startsWith("//")) continue;

            int indexStarJava = line.lastIndexOf("*.java");
            int indexStar = line.lastIndexOf("*");
            String dirName = null;
            if (indexStarJava != -1) {
                dirName = line.substring(0, indexStarJava);
            } else if (indexStar != -1) {
                dirName = line.substring(0, indexStar);
            }
            if (dirName != null) {
                File newDirFile = qualifiedFile(dirName, parentDir);
                File[] javafiles = newDirFile.listFiles(SOURCEFILTER);
                if (javafiles == null) {
                    cantResolve(newDirFile, "unable to file java files in wildcard directory "
                                + newDirFile.getPath());
                } else {
                    for (int i = 0; i < javafiles.length; i++) {
                        if (!maybeAdd(javafiles[i], args)) {
                            cantResolve(javafiles[i], "unable to add wildcard to collection");
                        }
                    }
                }
            } else if (line.startsWith("@")) {
                File newfile = qualifiedFile(line.substring(1).trim(), parentDir);
				if (!newfile.exists()) {
					cantResolve(newfile, "@file does not exist");
				} else { // recursive call unless cycle detected
					String filePath = newfile.getCanonicalPath().intern(); // exists, s.b. no IOException
					if (null == atFiles) {
						atFiles = new ArrayList();
						atFiles.add(file.getCanonicalPath().intern());
					}
					if (!atFiles.contains(filePath)) {
						atFiles.add(filePath);
						expandAtFile(newfile,  args, atFiles);
					} // else ignore newfile when cycle detected
				}
            } else {
                File newfile = qualifiedFile(line, parentDir);
                if (newfile.exists()) {
                    if (!maybeAdd(newfile, args)) {
                        cantResolve(newfile, "unable to add to collection");
                    }
                } else if (!(newfile.getName().equals("*.java") ||
                           newfile.getName().equals("*"))) {
                    cantResolve(newfile, "expecting @{listFile}, *, *.java, or {path}.java");
                } else { // maybe *.java or ...
//                    File newFileParentDir = newfile.getParentFile();
//                    if (null == newFileParentDir) {
//                        cantResolve(newfile);
//                    } else {
//                        File[] javafiles = newFileParentDir.listFiles(SOURCEFILTER);
//                        if (javafiles == null) {
                            cantResolve(newfile, "unable to find file "
										+ newfile);
//                        } else {
//                            for (int i = 0; i < javafiles.length; i++) {
//                                if (!maybeAdd(javafiles[i], args)) {
//                                    cantResolve(javafiles[i], "unable to add wildcard to collection");
//                                }
//                            }
//                        }
//                    }
                }
            }
        }
        in.close();
    }

    public static class LstFileEntryInvalidException extends Exception {
        public final String lstFile;
        public final int lineNumber;

        public LstFileEntryInvalidException(String lstFile, int lineNumber) {
			this(lstFile, lineNumber, "");
        }
        public LstFileEntryInvalidException(String lstFile, int lineNumber, String reason) {
			super("" + lstFile + ":" + lineNumber + ": " + reason
				  + " (invalid entry - expecting *.java or {path}.java or @{listFile})");
            this.lstFile = lstFile;
            this.lineNumber = lineNumber;
		}
    }

    /**
     * @todo:   make LstFileEntryInvalidException put in the right line numbers
     */
    private static void cantResolve(File file, String reason) throws LstFileEntryInvalidException{
        throw new LstFileEntryInvalidException(file.getAbsolutePath(), -1, reason);
	}
    private static void cantResolve(File file) throws LstFileEntryInvalidException{
        throw new LstFileEntryInvalidException(file.getAbsolutePath(), -1);
    }

    private static boolean maybeAdd(File file, Collection files) {
        if (isJavaFile(file)) {
            files.add(file.getAbsolutePath());
            return true;
        }
        return false;
    }

    private static boolean isJavaFile(File file) {
        return file != null && file.exists() && !file.isDirectory()
            && file.getName().endsWith(".java");
    }
}
